//===---- Floating.cpp - Support for floating point values ------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#include "Floating.h"

namespace clang {
namespace interp {

llvm::raw_ostream &operator<<(llvm::raw_ostream &OS, Floating F) {
  F.print(OS);
  return OS;
}

Floating getSwappedBytes(Floating F) { return F; }

} // namespace interp
} // namespace clang
