#include <cstring>
#include <limits>

#include "CompletionModel.h"
#include "Quality.h"
#include "clang/Sema/CodeCompleteConsumer.h"
#include "llvm/ADT/bit.h"

#define BIT(X) (1LL << X)

namespace clang {
namespace clangd {

using SymbolCategory_type = clang::clangd::SymbolQualitySignals::SymbolCategory;
using ContextKind_type = clang::CodeCompletionContext::Kind;
using Scope_type = clang::clangd::SymbolRelevanceSignals::AccessibleScope;

uint32_t Example::OrderEncode(float F) {
  static_assert(std::numeric_limits<float>::is_iec559, "");
  constexpr uint32_t TopBit = ~(~uint32_t{0} >> 1);

  // Get the bits of the float. Endianness is the same as for integers.
  uint32_t U = llvm::bit_cast<uint32_t>(F);
  std::memcpy(&U, &F, sizeof(U));
  // IEEE 754 floats compare like sign-magnitude integers.
  if (U & TopBit)    // Negative float.
    return 0 - U;    // Map onto the low half of integers, order reversed.
  return U + TopBit; // Positive floats map onto the high half of integers.
}

namespace {
LLVM_ATTRIBUTE_NOINLINE float EvaluateTree0(const Example& E) {
  t0_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t0_n244;
  t0_n1: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t0_n237;
  t0_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t0_n88;
  t0_n3: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t0_n65;
  t0_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t0_n40;
  t0_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t0_n15;
  t0_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t0_n10;
  t0_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t0_n9;
  t0_n8: return 0.04381459951400757f;
  t0_n9: return -0.07934795320034027f;
  t0_n10: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t0_n14;
  t0_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n13;
  t0_n12: return 0.08100750297307968f;
  t0_n13: return 0.16864584386348724f;
  t0_n14: return 0.1547447293996811f;
  t0_n15: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t0_n31;
  t0_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t0_n22;
  t0_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n19;
  t0_n18: return -0.17401087284088135f;
  t0_n19: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t0_n21;
  t0_n20: return 0.058513566851615906f;
  t0_n21: return -0.08097416162490845f;
  t0_n22: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t0_n28;
  t0_n23: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t0_n25;
  t0_n24: return -0.06604892760515213f;
  t0_n25: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t0_n27;
  t0_n26: return 0.01976003311574459f;
  t0_n27: return 0.14826877415180206f;
  t0_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t0_n30;
  t0_n29: return -0.15115617215633392f;
  t0_n30: return -0.07294262945652008f;
  t0_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n37;
  t0_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t0_n36;
  t0_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t0_n35;
  t0_n34: return -0.035924121737480164f;
  t0_n35: return 0.10752751678228378f;
  t0_n36: return -0.09618733078241348f;
  t0_n37: if (E.getNumReferences() >= 3332892672 /*21492.0*/) goto t0_n39;
  t0_n38: return 0.12352735549211502f;
  t0_n39: return 0.17493310570716858f;
  t0_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t0_n48;
  t0_n41: if (E.getNumReferences() >= 3307436032 /*2615.5*/) goto t0_n47;
  t0_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t0_n44;
  t0_n43: return -0.12128043919801712f;
  t0_n44: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t0_n46;
  t0_n45: return 0.14435838162899017f;
  t0_n46: return -0.015689551830291748f;
  t0_n47: return 0.14758464694023132f;
  t0_n48: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t0_n60;
  t0_n49: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t0_n59;
  t0_n50: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t0_n54;
  t0_n51: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t0_n53;
  t0_n52: return 0.0826253741979599f;
  t0_n53: return 0.1444838047027588f;
  t0_n54: if (E.getNumReferences() >= 3332389376 /*20509.0*/) goto t0_n58;
  t0_n55: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t0_n57;
  t0_n56: return 0.0436660572886467f;
  t0_n57: return -0.036370277404785156f;
  t0_n58: return 0.1600474864244461f;
  t0_n59: return -0.1922399252653122f;
  t0_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t0_n64;
  t0_n61: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t0_n63;
  t0_n62: return 0.17966283857822418f;
  t0_n63: return 0.09564231336116791f;
  t0_n64: return 0.17903363704681396f;
  t0_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t0_n81;
  t0_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n78;
  t0_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t0_n75;
  t0_n68: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t0_n72;
  t0_n69: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t0_n71;
  t0_n70: return 0.16343006491661072f;
  t0_n71: return 0.059975720942020416f;
  t0_n72: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t0_n74;
  t0_n73: return 0.1975041776895523f;
  t0_n74: return 0.17443816363811493f;
  t0_n75: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t0_n77;
  t0_n76: return 0.039421748369932175f;
  t0_n77: return 0.19089017808437347f;
  t0_n78: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t0_n80;
  t0_n79: return 0.09394174814224243f;
  t0_n80: return 0.1935829222202301f;
  t0_n81: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t0_n83;
  t0_n82: return 0.11955298483371735f;
  t0_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t0_n87;
  t0_n84: if (E.getNumReferences() >= 3352594624 /*108817.5*/) goto t0_n86;
  t0_n85: return 0.17330411076545715f;
  t0_n86: return 0.1923811435699463f;
  t0_n87: return 0.196611225605011f;
  t0_n88: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t0_n158;
  t0_n89: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t0_n141;
  t0_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t0_n122;
  t0_n91: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t0_n115;
  t0_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t0_n104;
  t0_n93: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t0_n101;
  t0_n94: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t0_n98;
  t0_n95: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n97;
  t0_n96: return 0.15734350681304932f;
  t0_n97: return 0.1247122511267662f;
  t0_n98: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n100;
  t0_n99: return 0.18153995275497437f;
  t0_n100: return 0.1647741198539734f;
  t0_n101: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n103;
  t0_n102: return 0.11501937359571457f;
  t0_n103: return 0.18159659206867218f;
  t0_n104: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n110;
  t0_n105: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t0_n109;
  t0_n106: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t0_n108;
  t0_n107: return 0.18379583954811096f;
  t0_n108: return 0.1939573884010315f;
  t0_n109: return 0.19570596516132355f;
  t0_n110: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t0_n112;
  t0_n111: return 0.0991196259856224f;
  t0_n112: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t0_n114;
  t0_n113: return 0.18978017568588257f;
  t0_n114: return -0.20000000298023224f;
  t0_n115: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t0_n121;
  t0_n116: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n118;
  t0_n117: return 0.10526332259178162f;
  t0_n118: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n120;
  t0_n119: return 0.18847665190696716f;
  t0_n120: return 0.17364506423473358f;
  t0_n121: return 0.19471041858196259f;
  t0_n122: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t0_n132;
  t0_n123: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t0_n125;
  t0_n124: return 0.17943941056728363f;
  t0_n125: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t0_n129;
  t0_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t0_n128;
  t0_n127: return 0.15801025927066803f;
  t0_n128: return 0.19034187495708466f;
  t0_n129: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t0_n131;
  t0_n130: return 0.19942399859428406f;
  t0_n131: return 0.1800059676170349f;
  t0_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t0_n138;
  t0_n133: if (E.getNumReferences() >= 3332434432 /*20597.0*/) goto t0_n137;
  t0_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t0_n136;
  t0_n135: return 0.1020292416214943f;
  t0_n136: return 0.18616393208503723f;
  t0_n137: return 0.19419915974140167f;
  t0_n138: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t0_n140;
  t0_n139: return 0.18478761613368988f;
  t0_n140: return 0.19306901097297668f;
  t0_n141: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t0_n157;
  t0_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n154;
  t0_n143: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t0_n149;
  t0_n144: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t0_n146;
  t0_n145: return 0.06863135099411011f;
  t0_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t0_n148;
  t0_n147: return 0.09177519381046295f;
  t0_n148: return 0.16046667098999023f;
  t0_n149: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t0_n153;
  t0_n150: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t0_n152;
  t0_n151: return 0.15973368287086487f;
  t0_n152: return 0.18605412542819977f;
  t0_n153: return -0.20000000298023224f;
  t0_n154: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t0_n156;
  t0_n155: return 0.1514042168855667f;
  t0_n156: return 0.1834273636341095f;
  t0_n157: return 0.19761313498020172f;
  t0_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t0_n194;
  t0_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t0_n161;
  t0_n160: return -0.1932513415813446f;
  t0_n161: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n177;
  t0_n162: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t0_n166;
  t0_n163: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n165;
  t0_n164: return -0.048965923488140106f;
  t0_n165: return 0.06573343276977539f;
  t0_n166: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n176;
  t0_n167: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t0_n173;
  t0_n168: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t0_n170;
  t0_n169: return 0.03787053003907204f;
  t0_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t0_n172;
  t0_n171: return 0.07951510697603226f;
  t0_n172: return 0.14633552730083466f;
  t0_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t0_n175;
  t0_n174: return 0.1372806280851364f;
  t0_n175: return 0.19920119643211365f;
  t0_n176: return 0.14596906304359436f;
  t0_n177: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t0_n193;
  t0_n178: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t0_n188;
  t0_n179: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t0_n181;
  t0_n180: return 0.06330600380897522f;
  t0_n181: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n187;
  t0_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n186;
  t0_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t0_n185;
  t0_n184: return 0.14506708085536957f;
  t0_n185: return 0.1656939536333084f;
  t0_n186: return 0.10698114335536957f;
  t0_n187: return 0.1838480830192566f;
  t0_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n192;
  t0_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t0_n191;
  t0_n190: return 0.1837356984615326f;
  t0_n191: return 0.13489769399166107f;
  t0_n192: return 0.1969956010580063f;
  t0_n193: return -0.19370104372501373f;
  t0_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t0_n208;
  t0_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t0_n205;
  t0_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t0_n200;
  t0_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t0_n199;
  t0_n198: return -0.19911359250545502f;
  t0_n199: return 0.008813080377876759f;
  t0_n200: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t0_n202;
  t0_n201: return 0.04461604356765747f;
  t0_n202: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t0_n204;
  t0_n203: return 0.19722077250480652f;
  t0_n204: return -0.15700146555900574f;
  t0_n205: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t0_n207;
  t0_n206: return 0.15436846017837524f;
  t0_n207: return 0.19539889693260193f;
  t0_n208: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t0_n232;
  t0_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t0_n229;
  t0_n210: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n216;
  t0_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t0_n213;
  t0_n212: return -0.04543353617191315f;
  t0_n213: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t0_n215;
  t0_n214: return 0.09861121326684952f;
  t0_n215: return 0.17618303000926971f;
  t0_n216: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n228;
  t0_n217: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t0_n219;
  t0_n218: return 0.13875418901443481f;
  t0_n219: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t0_n221;
  t0_n220: return -0.13945826888084412f;
  t0_n221: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t0_n223;
  t0_n222: return -8.339667692780495e-05f;
  t0_n223: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n227;
  t0_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t0_n226;
  t0_n225: return 0.06480171531438828f;
  t0_n226: return 0.16269299387931824f;
  t0_n227: return 0.09486537426710129f;
  t0_n228: return 0.17932546138763428f;
  t0_n229: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t0_n231;
  t0_n230: return 0.14165668189525604f;
  t0_n231: return 0.18670964241027832f;
  t0_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t0_n236;
  t0_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t0_n235;
  t0_n234: return 0.17782676219940186f;
  t0_n235: return 0.1604052633047104f;
  t0_n236: return 0.18952670693397522f;
  t0_n237: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t0_n243;
  t0_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t0_n240;
  t0_n239: return 0.15890617668628693f;
  t0_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t0_n242;
  t0_n241: return 0.1965150088071823f;
  t0_n242: return 0.14942441880702972f;
  t0_n243: return 0.19869807362556458f;
  t0_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t0_n246;
  t0_n245: return 0.006307594478130341f;
  t0_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t0_n254;
  t0_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t0_n253;
  t0_n248: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t0_n250;
  t0_n249: return 0.19164785742759705f;
  t0_n250: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t0_n252;
  t0_n251: return 0.19516006112098694f;
  t0_n252: return 0.19298601150512695f;
  t0_n253: return -0.10492532700300217f;
  t0_n254: return 0.1970273107290268f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree1(const Example& E) {
  t1_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t1_n244;
  t1_n1: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t1_n237;
  t1_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t1_n88;
  t1_n3: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t1_n65;
  t1_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t1_n36;
  t1_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t1_n15;
  t1_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n10;
  t1_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t1_n9;
  t1_n8: return 0.05864229425787926f;
  t1_n9: return -0.10569842904806137f;
  t1_n10: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t1_n14;
  t1_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n13;
  t1_n12: return 0.09925279021263123f;
  t1_n13: return 0.15963847935199738f;
  t1_n14: return 0.1503359079360962f;
  t1_n15: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n31;
  t1_n16: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t1_n22;
  t1_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n21;
  t1_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t1_n20;
  t1_n19: return -0.16126444935798645f;
  t1_n20: return -0.09174440056085587f;
  t1_n21: return 0.05892568081617355f;
  t1_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t1_n26;
  t1_n23: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n25;
  t1_n24: return -0.10753640532493591f;
  t1_n25: return 0.12494514882564545f;
  t1_n26: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n30;
  t1_n27: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t1_n29;
  t1_n28: return 0.028882766142487526f;
  t1_n29: return 0.1169862225651741f;
  t1_n30: return 0.15312351286411285f;
  t1_n31: if (E.getNumReferences() >= 3290931200 /*670.5*/) goto t1_n33;
  t1_n32: return -0.15785080194473267f;
  t1_n33: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n35;
  t1_n34: return -0.11045386642217636f;
  t1_n35: return 0.14530658721923828f;
  t1_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t1_n44;
  t1_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t1_n39;
  t1_n38: return -0.13928842544555664f;
  t1_n39: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t1_n41;
  t1_n40: return 0.14791369438171387f;
  t1_n41: if (E.getNumReferences() >= 3307436032 /*2615.5*/) goto t1_n43;
  t1_n42: return 0.010207350365817547f;
  t1_n43: return 0.1492609828710556f;
  t1_n44: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n60;
  t1_n45: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t1_n59;
  t1_n46: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n50;
  t1_n47: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t1_n49;
  t1_n48: return 0.0861087366938591f;
  t1_n49: return 0.1448994129896164f;
  t1_n50: if (E.getNumReferences() >= 3332389376 /*20509.0*/) goto t1_n58;
  t1_n51: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t1_n53;
  t1_n52: return 0.014164520427584648f;
  t1_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t1_n55;
  t1_n54: return -0.0015365395229309797f;
  t1_n55: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t1_n57;
  t1_n56: return 0.17775259912014008f;
  t1_n57: return 0.15263903141021729f;
  t1_n58: return 0.15500111877918243f;
  t1_n59: return -0.1689857840538025f;
  t1_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t1_n64;
  t1_n61: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n63;
  t1_n62: return 0.1693732887506485f;
  t1_n63: return 0.11021314561367035f;
  t1_n64: return 0.1651707887649536f;
  t1_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t1_n81;
  t1_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n78;
  t1_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t1_n73;
  t1_n68: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t1_n72;
  t1_n69: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t1_n71;
  t1_n70: return 0.15608295798301697f;
  t1_n71: return 0.06868483871221542f;
  t1_n72: return 0.16499853134155273f;
  t1_n73: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n77;
  t1_n74: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t1_n76;
  t1_n75: return -0.02576316148042679f;
  t1_n76: return 0.09417004138231277f;
  t1_n77: return 0.16653694212436676f;
  t1_n78: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t1_n80;
  t1_n79: return 0.10624748468399048f;
  t1_n80: return 0.17000022530555725f;
  t1_n81: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t1_n83;
  t1_n82: return 0.12199915200471878f;
  t1_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t1_n87;
  t1_n84: if (E.getNumReferences() >= 3352594624 /*108817.5*/) goto t1_n86;
  t1_n85: return 0.1616566777229309f;
  t1_n86: return 0.16865934431552887f;
  t1_n87: return 0.1725209802389145f;
  t1_n88: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t1_n170;
  t1_n89: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t1_n153;
  t1_n90: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t1_n130;
  t1_n91: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t1_n105;
  t1_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t1_n102;
  t1_n93: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t1_n101;
  t1_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t1_n98;
  t1_n95: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t1_n97;
  t1_n96: return 0.14982688426971436f;
  t1_n97: return 0.13161303102970123f;
  t1_n98: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n100;
  t1_n99: return 0.16417934000492096f;
  t1_n100: return 0.1307990849018097f;
  t1_n101: return 0.1607930213212967f;
  t1_n102: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t1_n104;
  t1_n103: return 0.16888529062271118f;
  t1_n104: return 0.1604202538728714f;
  t1_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t1_n125;
  t1_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t1_n112;
  t1_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n109;
  t1_n108: return 0.12190522253513336f;
  t1_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t1_n111;
  t1_n110: return 0.04777759313583374f;
  t1_n111: return 0.16579733788967133f;
  t1_n112: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t1_n120;
  t1_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n119;
  t1_n114: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t1_n116;
  t1_n115: return 0.1560467630624771f;
  t1_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t1_n118;
  t1_n117: return 0.15392252802848816f;
  t1_n118: return 0.16860151290893555f;
  t1_n119: return 0.17147164046764374f;
  t1_n120: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t1_n124;
  t1_n121: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t1_n123;
  t1_n122: return 0.1692265123128891f;
  t1_n123: return 0.1711229532957077f;
  t1_n124: return 0.16357260942459106f;
  t1_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n127;
  t1_n126: return 0.15169982612133026f;
  t1_n127: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t1_n129;
  t1_n128: return 0.16702626645565033f;
  t1_n129: return -0.175074964761734f;
  t1_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t1_n150;
  t1_n131: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t1_n149;
  t1_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t1_n144;
  t1_n133: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n141;
  t1_n134: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t1_n138;
  t1_n135: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t1_n137;
  t1_n136: return 0.16471685469150543f;
  t1_n137: return 0.15898099541664124f;
  t1_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t1_n140;
  t1_n139: return 0.17276112735271454f;
  t1_n140: return 0.16352130472660065f;
  t1_n141: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t1_n143;
  t1_n142: return 0.13886868953704834f;
  t1_n143: return 0.16449801623821259f;
  t1_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t1_n146;
  t1_n145: return 0.17154660820960999f;
  t1_n146: if (E.getNumReferences() >= 3316067328 /*5349.5*/) goto t1_n148;
  t1_n147: return 0.1671976000070572f;
  t1_n148: return 0.1694512814283371f;
  t1_n149: return 0.1682843267917633f;
  t1_n150: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t1_n152;
  t1_n151: return 0.16732017695903778f;
  t1_n152: return 0.1706983745098114f;
  t1_n153: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t1_n169;
  t1_n154: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t1_n160;
  t1_n155: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t1_n159;
  t1_n156: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n158;
  t1_n157: return 0.08604030311107635f;
  t1_n158: return 0.14936451613903046f;
  t1_n159: return 0.1562318503856659f;
  t1_n160: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n168;
  t1_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t1_n163;
  t1_n162: return 0.10557645559310913f;
  t1_n163: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t1_n167;
  t1_n164: if (E.getNumReferences() >= 3318322176 /*6450.5*/) goto t1_n166;
  t1_n165: return 0.152414470911026f;
  t1_n166: return 0.16552692651748657f;
  t1_n167: return -0.16779406368732452f;
  t1_n168: return 0.16632699966430664f;
  t1_n169: return 0.17045937478542328f;
  t1_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t1_n208;
  t1_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t1_n173;
  t1_n172: return -0.16870714724063873f;
  t1_n173: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n191;
  t1_n174: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t1_n178;
  t1_n175: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n177;
  t1_n176: return -0.07798636704683304f;
  t1_n177: return 0.08636608719825745f;
  t1_n178: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n188;
  t1_n179: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t1_n185;
  t1_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n182;
  t1_n181: return 0.04533401504158974f;
  t1_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t1_n184;
  t1_n183: return 0.10128161311149597f;
  t1_n184: return 0.1477314680814743f;
  t1_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t1_n187;
  t1_n186: return 0.14149019122123718f;
  t1_n187: return 0.16968144476413727f;
  t1_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t1_n190;
  t1_n189: return 0.151018425822258f;
  t1_n190: return 0.16072402894496918f;
  t1_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t1_n207;
  t1_n192: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t1_n202;
  t1_n193: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t1_n195;
  t1_n194: return 0.08147435635328293f;
  t1_n195: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n201;
  t1_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t1_n200;
  t1_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t1_n199;
  t1_n198: return 0.14504404366016388f;
  t1_n199: return 0.15743094682693481f;
  t1_n200: return 0.12245187908411026f;
  t1_n201: return 0.16542880237102509f;
  t1_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n206;
  t1_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t1_n205;
  t1_n204: return 0.15487730503082275f;
  t1_n205: return 0.16630683839321136f;
  t1_n206: return 0.16949501633644104f;
  t1_n207: return -0.1698521226644516f;
  t1_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t1_n220;
  t1_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t1_n217;
  t1_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t1_n212;
  t1_n211: return -0.09946514666080475f;
  t1_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t1_n214;
  t1_n213: return 0.03577881306409836f;
  t1_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t1_n216;
  t1_n215: return 0.16909195482730865f;
  t1_n216: return -0.15543821454048157f;
  t1_n217: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t1_n219;
  t1_n218: return 0.13251914083957672f;
  t1_n219: return 0.16929800808429718f;
  t1_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t1_n232;
  t1_n221: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t1_n225;
  t1_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t1_n224;
  t1_n223: return 0.1044926792383194f;
  t1_n224: return 0.14694420993328094f;
  t1_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t1_n231;
  t1_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n228;
  t1_n227: return -0.044721461832523346f;
  t1_n228: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n230;
  t1_n229: return 0.13720722496509552f;
  t1_n230: return 0.16553179919719696f;
  t1_n231: return 0.16303226351737976f;
  t1_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t1_n236;
  t1_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t1_n235;
  t1_n234: return 0.16340629756450653f;
  t1_n235: return 0.15127478539943695f;
  t1_n236: return 0.16704440116882324f;
  t1_n237: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t1_n243;
  t1_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t1_n240;
  t1_n239: return 0.15209516882896423f;
  t1_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t1_n242;
  t1_n241: return 0.1707134246826172f;
  t1_n242: return 0.1531279981136322f;
  t1_n243: return 0.1713443547487259f;
  t1_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t1_n246;
  t1_n245: return 0.02058301866054535f;
  t1_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t1_n254;
  t1_n247: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t1_n251;
  t1_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t1_n250;
  t1_n249: return 0.16857142746448517f;
  t1_n250: return -0.13122344017028809f;
  t1_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t1_n253;
  t1_n252: return 0.16965043544769287f;
  t1_n253: return 0.16884683072566986f;
  t1_n254: return 0.17080119252204895f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree2(const Example& E) {
  t2_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t2_n244;
  t2_n1: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t2_n237;
  t2_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t2_n92;
  t2_n3: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t2_n69;
  t2_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t2_n36;
  t2_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t2_n15;
  t2_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n10;
  t2_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t2_n9;
  t2_n8: return 0.046585436910390854f;
  t2_n9: return -0.08026197552680969f;
  t2_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t2_n14;
  t2_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n13;
  t2_n12: return 0.07915130257606506f;
  t2_n13: return 0.13851292431354523f;
  t2_n14: return 0.12804220616817474f;
  t2_n15: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t2_n27;
  t2_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n26;
  t2_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t2_n21;
  t2_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n20;
  t2_n19: return -0.14198435842990875f;
  t2_n20: return 0.05377653241157532f;
  t2_n21: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t2_n23;
  t2_n22: return -0.07909336686134338f;
  t2_n23: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n25;
  t2_n24: return 0.026040226221084595f;
  t2_n25: return 0.13231100142002106f;
  t2_n26: return -0.14052613079547882f;
  t2_n27: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n33;
  t2_n28: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n32;
  t2_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t2_n31;
  t2_n30: return -0.056663040071725845f;
  t2_n31: return 0.10488138347864151f;
  t2_n32: return -0.0789332166314125f;
  t2_n33: if (E.getNumReferences() >= 3332892672 /*21492.0*/) goto t2_n35;
  t2_n34: return 0.1106741651892662f;
  t2_n35: return 0.14635606110095978f;
  t2_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t2_n44;
  t2_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t2_n39;
  t2_n38: return -0.12101104110479355f;
  t2_n39: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t2_n41;
  t2_n40: return 0.12934809923171997f;
  t2_n41: if (E.getNumReferences() >= 3307436032 /*2615.5*/) goto t2_n43;
  t2_n42: return 0.0067842998541891575f;
  t2_n43: return 0.13125552237033844f;
  t2_n44: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n64;
  t2_n45: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t2_n53;
  t2_n46: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t2_n48;
  t2_n47: return 0.006239415146410465f;
  t2_n48: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t2_n50;
  t2_n49: return 0.15282219648361206f;
  t2_n50: if (E.getNumReferences() >= 3332389376 /*20509.0*/) goto t2_n52;
  t2_n51: return 0.07543952763080597f;
  t2_n52: return 0.1397043913602829f;
  t2_n53: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t2_n57;
  t2_n54: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t2_n56;
  t2_n55: return 0.1287938803434372f;
  t2_n56: return -0.1471850723028183f;
  t2_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t2_n61;
  t2_n58: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t2_n60;
  t2_n59: return 0.09584422409534454f;
  t2_n60: return 0.005734036210924387f;
  t2_n61: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t2_n63;
  t2_n62: return -0.05343735218048096f;
  t2_n63: return 0.15033230185508728f;
  t2_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t2_n68;
  t2_n65: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n67;
  t2_n66: return 0.15294736623764038f;
  t2_n67: return 0.09703253954648972f;
  t2_n68: return 0.14814265072345734f;
  t2_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t2_n85;
  t2_n70: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n82;
  t2_n71: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t2_n77;
  t2_n72: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t2_n76;
  t2_n73: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t2_n75;
  t2_n74: return 0.13854600489139557f;
  t2_n75: return 0.05823113024234772f;
  t2_n76: return 0.1471627652645111f;
  t2_n77: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n81;
  t2_n78: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t2_n80;
  t2_n79: return -0.028328537940979004f;
  t2_n80: return 0.08673255890607834f;
  t2_n81: return 0.14768977463245392f;
  t2_n82: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t2_n84;
  t2_n83: return 0.08953730762004852f;
  t2_n84: return 0.15202684700489044f;
  t2_n85: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t2_n87;
  t2_n86: return 0.10249080508947372f;
  t2_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t2_n91;
  t2_n88: if (E.getNumReferences() >= 3352594624 /*108817.5*/) goto t2_n90;
  t2_n89: return 0.14352184534072876f;
  t2_n90: return 0.1506611406803131f;
  t2_n91: return 0.15508578717708588f;
  t2_n92: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t2_n170;
  t2_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t2_n153;
  t2_n94: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t2_n130;
  t2_n95: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t2_n107;
  t2_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t2_n104;
  t2_n97: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t2_n103;
  t2_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t2_n100;
  t2_n99: return 0.12745007872581482f;
  t2_n100: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n102;
  t2_n101: return 0.1455935537815094f;
  t2_n102: return 0.11362037062644958f;
  t2_n103: return 0.14222349226474762f;
  t2_n104: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t2_n106;
  t2_n105: return 0.15041328966617584f;
  t2_n106: return 0.14176741242408752f;
  t2_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t2_n127;
  t2_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t2_n114;
  t2_n109: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n111;
  t2_n110: return 0.10420667380094528f;
  t2_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t2_n113;
  t2_n112: return 0.021743804216384888f;
  t2_n113: return 0.14680303633213043f;
  t2_n114: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t2_n122;
  t2_n115: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n121;
  t2_n116: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t2_n118;
  t2_n117: return 0.13810108602046967f;
  t2_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t2_n120;
  t2_n119: return 0.1365198940038681f;
  t2_n120: return 0.15012964606285095f;
  t2_n121: return 0.151300311088562f;
  t2_n122: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t2_n126;
  t2_n123: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t2_n125;
  t2_n124: return 0.1500474214553833f;
  t2_n125: return 0.15244720876216888f;
  t2_n126: return 0.144512340426445f;
  t2_n127: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n129;
  t2_n128: return 0.13484029471874237f;
  t2_n129: return 0.14781226217746735f;
  t2_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t2_n150;
  t2_n131: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t2_n149;
  t2_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t2_n144;
  t2_n133: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n141;
  t2_n134: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t2_n138;
  t2_n135: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t2_n137;
  t2_n136: return 0.1464795470237732f;
  t2_n137: return 0.1409185230731964f;
  t2_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t2_n140;
  t2_n139: return 0.15495337545871735f;
  t2_n140: return 0.14671483635902405f;
  t2_n141: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t2_n143;
  t2_n142: return 0.12058647722005844f;
  t2_n143: return 0.14545537531375885f;
  t2_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t2_n146;
  t2_n145: return 0.15395553410053253f;
  t2_n146: if (E.getNumReferences() >= 3316067328 /*5349.5*/) goto t2_n148;
  t2_n147: return 0.14815135300159454f;
  t2_n148: return 0.15047454833984375f;
  t2_n149: return 0.14929066598415375f;
  t2_n150: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t2_n152;
  t2_n151: return 0.14835670590400696f;
  t2_n152: return 0.1516115516424179f;
  t2_n153: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t2_n169;
  t2_n154: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t2_n160;
  t2_n155: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t2_n159;
  t2_n156: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n158;
  t2_n157: return 0.07251094281673431f;
  t2_n158: return 0.13211779296398163f;
  t2_n159: return 0.13767579197883606f;
  t2_n160: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n168;
  t2_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t2_n163;
  t2_n162: return 0.09247835725545883f;
  t2_n163: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t2_n167;
  t2_n164: if (E.getNumReferences() >= 3318322176 /*6450.5*/) goto t2_n166;
  t2_n165: return 0.1341649442911148f;
  t2_n166: return 0.14664432406425476f;
  t2_n167: return -0.1486622542142868f;
  t2_n168: return 0.14859846234321594f;
  t2_n169: return 0.15215343236923218f;
  t2_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t2_n208;
  t2_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t2_n173;
  t2_n172: return -0.14995400607585907f;
  t2_n173: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n191;
  t2_n174: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t2_n178;
  t2_n175: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n177;
  t2_n176: return -0.06677867472171783f;
  t2_n177: return 0.07573500275611877f;
  t2_n178: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n188;
  t2_n179: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t2_n185;
  t2_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n182;
  t2_n181: return 0.04241118207573891f;
  t2_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t2_n184;
  t2_n183: return 0.0906544104218483f;
  t2_n184: return 0.13016828894615173f;
  t2_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t2_n187;
  t2_n186: return 0.1250816136598587f;
  t2_n187: return 0.15090744197368622f;
  t2_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t2_n190;
  t2_n189: return 0.1378934383392334f;
  t2_n190: return 0.1413843184709549f;
  t2_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t2_n207;
  t2_n192: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t2_n202;
  t2_n193: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t2_n195;
  t2_n194: return 0.06974995136260986f;
  t2_n195: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n201;
  t2_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t2_n200;
  t2_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t2_n199;
  t2_n198: return 0.12677161395549774f;
  t2_n199: return 0.13914361596107483f;
  t2_n200: return 0.10487604886293411f;
  t2_n201: return 0.14650315046310425f;
  t2_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n206;
  t2_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t2_n205;
  t2_n204: return 0.13627324998378754f;
  t2_n205: return 0.14791250228881836f;
  t2_n206: return 0.15074218809604645f;
  t2_n207: return -0.1516505777835846f;
  t2_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t2_n220;
  t2_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t2_n217;
  t2_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t2_n212;
  t2_n211: return -0.09102858603000641f;
  t2_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t2_n214;
  t2_n213: return 0.02203204482793808f;
  t2_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t2_n216;
  t2_n215: return 0.15040287375450134f;
  t2_n216: return -0.14173327386379242f;
  t2_n217: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t2_n219;
  t2_n218: return 0.11469417065382004f;
  t2_n219: return 0.1504640132188797f;
  t2_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t2_n232;
  t2_n221: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t2_n225;
  t2_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t2_n224;
  t2_n223: return 0.09050043672323227f;
  t2_n224: return 0.1292455792427063f;
  t2_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t2_n231;
  t2_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n228;
  t2_n227: return -0.04138435050845146f;
  t2_n228: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n230;
  t2_n229: return 0.12220322340726852f;
  t2_n230: return 0.14674684405326843f;
  t2_n231: return 0.14500021934509277f;
  t2_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t2_n236;
  t2_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t2_n235;
  t2_n234: return 0.14494916796684265f;
  t2_n235: return 0.13270336389541626f;
  t2_n236: return 0.1485733985900879f;
  t2_n237: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t2_n243;
  t2_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t2_n240;
  t2_n239: return 0.13365139067173004f;
  t2_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t2_n242;
  t2_n241: return 0.15242359042167664f;
  t2_n242: return 0.1332457959651947f;
  t2_n243: return 0.15301726758480072f;
  t2_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t2_n246;
  t2_n245: return 0.0189053937792778f;
  t2_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t2_n254;
  t2_n247: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t2_n251;
  t2_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t2_n250;
  t2_n249: return 0.1502094715833664f;
  t2_n250: return -0.11232098937034607f;
  t2_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t2_n253;
  t2_n252: return 0.15126243233680725f;
  t2_n253: return 0.15046553313732147f;
  t2_n254: return 0.15229535102844238f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree3(const Example& E) {
  t3_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t3_n244;
  t3_n1: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t3_n237;
  t3_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t3_n98;
  t3_n3: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t3_n75;
  t3_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t3_n36;
  t3_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t3_n13;
  t3_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n10;
  t3_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n9;
  t3_n8: return 0.03923378139734268f;
  t3_n9: return -0.06926216185092926f;
  t3_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t3_n12;
  t3_n11: return 0.07563339918851852f;
  t3_n12: return 0.11524554342031479f;
  t3_n13: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t3_n25;
  t3_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t3_n18;
  t3_n15: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n17;
  t3_n16: return -0.1375555396080017f;
  t3_n17: return -0.047991134226322174f;
  t3_n18: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n24;
  t3_n19: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t3_n21;
  t3_n20: return -0.06922228634357452f;
  t3_n21: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n23;
  t3_n22: return 0.024068892002105713f;
  t3_n23: return 0.11967655271291733f;
  t3_n24: return -0.11717651039361954f;
  t3_n25: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n33;
  t3_n26: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n32;
  t3_n27: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n31;
  t3_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t3_n30;
  t3_n29: return -0.04981289058923721f;
  t3_n30: return 0.090534508228302f;
  t3_n31: return -0.07225030660629272f;
  t3_n32: return 0.13320574164390564f;
  t3_n33: if (E.getNumReferences() >= 3332892672 /*21492.0*/) goto t3_n35;
  t3_n34: return 0.09868115186691284f;
  t3_n35: return 0.13387064635753632f;
  t3_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t3_n44;
  t3_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t3_n39;
  t3_n38: return -0.10744839906692505f;
  t3_n39: if (E.getNumReferences() >= 3307436032 /*2615.5*/) goto t3_n43;
  t3_n40: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t3_n42;
  t3_n41: return 0.11521003395318985f;
  t3_n42: return 0.005204482469707727f;
  t3_n43: return 0.1198052391409874f;
  t3_n44: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n66;
  t3_n45: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t3_n55;
  t3_n46: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t3_n52;
  t3_n47: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t3_n51;
  t3_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n50;
  t3_n49: return 0.06424921751022339f;
  t3_n50: return -0.0016738876001909375f;
  t3_n51: return -0.1457492709159851f;
  t3_n52: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t3_n54;
  t3_n53: return 0.13960492610931396f;
  t3_n54: return 0.09588909894227982f;
  t3_n55: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n59;
  t3_n56: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t3_n58;
  t3_n57: return 0.11726795136928558f;
  t3_n58: return -0.13461798429489136f;
  t3_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t3_n63;
  t3_n60: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t3_n62;
  t3_n61: return 0.08398263901472092f;
  t3_n62: return -0.00013325239706318825f;
  t3_n63: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t3_n65;
  t3_n64: return -0.05175580084323883f;
  t3_n65: return 0.13521228730678558f;
  t3_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t3_n72;
  t3_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n69;
  t3_n68: return 0.14159318804740906f;
  t3_n69: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t3_n71;
  t3_n70: return 0.13354851305484772f;
  t3_n71: return 0.07254661619663239f;
  t3_n72: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t3_n74;
  t3_n73: return 0.11944756656885147f;
  t3_n74: return 0.1405000239610672f;
  t3_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t3_n91;
  t3_n76: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n88;
  t3_n77: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n83;
  t3_n78: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t3_n82;
  t3_n79: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t3_n81;
  t3_n80: return 0.1259843409061432f;
  t3_n81: return 0.05297345295548439f;
  t3_n82: return 0.134999617934227f;
  t3_n83: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n87;
  t3_n84: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t3_n86;
  t3_n85: return -0.028413616120815277f;
  t3_n86: return 0.0757773369550705f;
  t3_n87: return 0.1352328360080719f;
  t3_n88: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t3_n90;
  t3_n89: return 0.0798695757985115f;
  t3_n90: return 0.14022508263587952f;
  t3_n91: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t3_n93;
  t3_n92: return 0.09002068638801575f;
  t3_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t3_n97;
  t3_n94: if (E.getNumReferences() >= 3352594624 /*108817.5*/) goto t3_n96;
  t3_n95: return 0.1312221884727478f;
  t3_n96: return 0.13870957493782043f;
  t3_n97: return 0.14349539577960968f;
  t3_n98: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t3_n172;
  t3_n99: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t3_n155;
  t3_n100: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t3_n132;
  t3_n101: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t3_n109;
  t3_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t3_n106;
  t3_n103: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t3_n105;
  t3_n104: return 0.11798160523176193f;
  t3_n105: return 0.12928789854049683f;
  t3_n106: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n108;
  t3_n107: return 0.13794782757759094f;
  t3_n108: return 0.12899604439735413f;
  t3_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t3_n129;
  t3_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t3_n116;
  t3_n111: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t3_n113;
  t3_n112: return 0.09215444326400757f;
  t3_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t3_n115;
  t3_n114: return 0.018928291276097298f;
  t3_n115: return 0.13420186936855316f;
  t3_n116: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t3_n124;
  t3_n117: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t3_n123;
  t3_n118: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t3_n120;
  t3_n119: return 0.12529897689819336f;
  t3_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t3_n122;
  t3_n121: return 0.12508316338062286f;
  t3_n122: return 0.13786378502845764f;
  t3_n123: return 0.13811492919921875f;
  t3_n124: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t3_n128;
  t3_n125: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t3_n127;
  t3_n126: return 0.13723386824131012f;
  t3_n127: return 0.1401970386505127f;
  t3_n128: return 0.13127586245536804f;
  t3_n129: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n131;
  t3_n130: return 0.12205123901367188f;
  t3_n131: return 0.13521887362003326f;
  t3_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t3_n152;
  t3_n133: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t3_n151;
  t3_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t3_n146;
  t3_n135: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n143;
  t3_n136: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t3_n140;
  t3_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t3_n139;
  t3_n138: return 0.1341373324394226f;
  t3_n139: return 0.12846247851848602f;
  t3_n140: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t3_n142;
  t3_n141: return 0.143375962972641f;
  t3_n142: return 0.13550786674022675f;
  t3_n143: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t3_n145;
  t3_n144: return 0.10716212540864944f;
  t3_n145: return 0.13256163895130157f;
  t3_n146: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t3_n148;
  t3_n147: return 0.14199858903884888f;
  t3_n148: if (E.getNumReferences() >= 3317256192 /*5930.0*/) goto t3_n150;
  t3_n149: return 0.13554522395133972f;
  t3_n150: return 0.13807246088981628f;
  t3_n151: return 0.13682542741298676f;
  t3_n152: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t3_n154;
  t3_n153: return 0.13570816814899445f;
  t3_n154: return 0.13902103900909424f;
  t3_n155: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t3_n171;
  t3_n156: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t3_n162;
  t3_n157: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t3_n161;
  t3_n158: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n160;
  t3_n159: return 0.06314817816019058f;
  t3_n160: return 0.11961252987384796f;
  t3_n161: return 0.12444785982370377f;
  t3_n162: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n170;
  t3_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t3_n165;
  t3_n164: return 0.08089909702539444f;
  t3_n165: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t3_n169;
  t3_n166: if (E.getNumReferences() >= 3318322176 /*6450.5*/) goto t3_n168;
  t3_n167: return 0.12116461247205734f;
  t3_n168: return 0.1339026838541031f;
  t3_n169: return -0.1364395171403885f;
  t3_n170: return 0.13692551851272583f;
  t3_n171: return 0.14038771390914917f;
  t3_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t3_n208;
  t3_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t3_n175;
  t3_n174: return -0.13776852190494537f;
  t3_n175: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t3_n191;
  t3_n176: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t3_n180;
  t3_n177: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n179;
  t3_n178: return -0.058771226555109024f;
  t3_n179: return 0.06741133332252502f;
  t3_n180: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n190;
  t3_n181: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t3_n187;
  t3_n182: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n184;
  t3_n183: return 0.03722731024026871f;
  t3_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t3_n186;
  t3_n185: return 0.08181355893611908f;
  t3_n186: return 0.11761044710874557f;
  t3_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t3_n189;
  t3_n188: return 0.08910182863473892f;
  t3_n189: return 0.1396014243364334f;
  t3_n190: return 0.12793539464473724f;
  t3_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t3_n207;
  t3_n192: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t3_n202;
  t3_n193: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t3_n195;
  t3_n194: return 0.0601702556014061f;
  t3_n195: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n201;
  t3_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t3_n200;
  t3_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t3_n199;
  t3_n198: return 0.11347981542348862f;
  t3_n199: return 0.12639573216438293f;
  t3_n200: return 0.0928204357624054f;
  t3_n201: return 0.13379442691802979f;
  t3_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n206;
  t3_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t3_n205;
  t3_n204: return 0.12333247810602188f;
  t3_n205: return 0.13546068966388702f;
  t3_n206: return 0.13851186633110046f;
  t3_n207: return -0.13949863612651825f;
  t3_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t3_n220;
  t3_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t3_n217;
  t3_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t3_n212;
  t3_n211: return -0.08323707431554794f;
  t3_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t3_n214;
  t3_n213: return 0.02088886871933937f;
  t3_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t3_n216;
  t3_n215: return 0.1381932944059372f;
  t3_n216: return -0.13304990530014038f;
  t3_n217: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t3_n219;
  t3_n218: return 0.10189694166183472f;
  t3_n219: return 0.1381164938211441f;
  t3_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t3_n232;
  t3_n221: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t3_n225;
  t3_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t3_n224;
  t3_n223: return 0.07897063344717026f;
  t3_n224: return 0.11651987582445145f;
  t3_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t3_n231;
  t3_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t3_n228;
  t3_n227: return -0.03606373816728592f;
  t3_n228: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n230;
  t3_n229: return 0.10936389118432999f;
  t3_n230: return 0.13415446877479553f;
  t3_n231: return 0.13259655237197876f;
  t3_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t3_n236;
  t3_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t3_n235;
  t3_n234: return 0.13253413140773773f;
  t3_n235: return 0.1200137659907341f;
  t3_n236: return 0.136396124958992f;
  t3_n237: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t3_n243;
  t3_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t3_n240;
  t3_n239: return 0.12059909105300903f;
  t3_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t3_n242;
  t3_n241: return 0.14043724536895752f;
  t3_n242: return 0.12053101509809494f;
  t3_n243: return 0.14106890559196472f;
  t3_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t3_n246;
  t3_n245: return 0.017146291211247444f;
  t3_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t3_n254;
  t3_n247: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t3_n251;
  t3_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t3_n250;
  t3_n249: return 0.13809612393379211f;
  t3_n250: return -0.09687822312116623f;
  t3_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t3_n253;
  t3_n252: return 0.1392662674188614f;
  t3_n253: return 0.1384020298719406f;
  t3_n254: return 0.14027617871761322f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree4(const Example& E) {
  t4_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t4_n244;
  t4_n1: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t4_n237;
  t4_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t4_n100;
  t4_n3: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t4_n77;
  t4_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t4_n42;
  t4_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t4_n17;
  t4_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n14;
  t4_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n13;
  t4_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t4_n10;
  t4_n9: return -0.0039026245940476656f;
  t4_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t4_n12;
  t4_n11: return 0.0292171873152256f;
  t4_n12: return 0.1009465679526329f;
  t4_n13: return -0.0604998841881752f;
  t4_n14: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t4_n16;
  t4_n15: return 0.06604607403278351f;
  t4_n16: return 0.10512399673461914f;
  t4_n17: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t4_n31;
  t4_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t4_n24;
  t4_n19: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n21;
  t4_n20: return -0.12901060283184052f;
  t4_n21: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n23;
  t4_n22: return 0.048817507922649384f;
  t4_n23: return -0.061059415340423584f;
  t4_n24: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t4_n30;
  t4_n25: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t4_n27;
  t4_n26: return -0.0613226443529129f;
  t4_n27: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n29;
  t4_n28: return 0.020388251170516014f;
  t4_n29: return 0.10966802388429642f;
  t4_n30: return -0.10738255828619003f;
  t4_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n39;
  t4_n32: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n38;
  t4_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t4_n35;
  t4_n34: return -0.06717780977487564f;
  t4_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t4_n37;
  t4_n36: return -0.03535696119070053f;
  t4_n37: return 0.0649673193693161f;
  t4_n38: return 0.12601225078105927f;
  t4_n39: if (E.getNumReferences() >= 3332892672 /*21492.0*/) goto t4_n41;
  t4_n40: return 0.08869346976280212f;
  t4_n41: return 0.1248442754149437f;
  t4_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t4_n50;
  t4_n43: if (E.getNumReferences() >= 3307436032 /*2615.5*/) goto t4_n49;
  t4_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t4_n46;
  t4_n45: return -0.09852996468544006f;
  t4_n46: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t4_n48;
  t4_n47: return 0.09921697527170181f;
  t4_n48: return -0.0021524596959352493f;
  t4_n49: return 0.11546836793422699f;
  t4_n50: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n68;
  t4_n51: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t4_n57;
  t4_n52: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t4_n54;
  t4_n53: return 0.000680825556628406f;
  t4_n54: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t4_n56;
  t4_n55: return 0.12783223390579224f;
  t4_n56: return 0.08461324870586395f;
  t4_n57: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n61;
  t4_n58: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t4_n60;
  t4_n59: return 0.10816715657711029f;
  t4_n60: return -0.12507084012031555f;
  t4_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t4_n65;
  t4_n62: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t4_n64;
  t4_n63: return 0.07378796488046646f;
  t4_n64: return 0.010963642038404942f;
  t4_n65: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t4_n67;
  t4_n66: return -0.04847811162471771f;
  t4_n67: return 0.12381131947040558f;
  t4_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t4_n74;
  t4_n69: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n71;
  t4_n70: return 0.13345526158809662f;
  t4_n71: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t4_n73;
  t4_n72: return 0.12494406849145889f;
  t4_n73: return 0.06688376516103745f;
  t4_n74: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t4_n76;
  t4_n75: return 0.10966937243938446f;
  t4_n76: return 0.131972536444664f;
  t4_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t4_n93;
  t4_n78: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n90;
  t4_n79: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n85;
  t4_n80: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t4_n84;
  t4_n81: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t4_n83;
  t4_n82: return 0.11612967401742935f;
  t4_n83: return 0.046420253813266754f;
  t4_n84: return 0.12601648271083832f;
  t4_n85: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n89;
  t4_n86: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t4_n88;
  t4_n87: return -0.024922605603933334f;
  t4_n88: return 0.06578359007835388f;
  t4_n89: return 0.1261170506477356f;
  t4_n90: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t4_n92;
  t4_n91: return 0.07471520453691483f;
  t4_n92: return 0.1319267302751541f;
  t4_n93: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t4_n95;
  t4_n94: return 0.07883468270301819f;
  t4_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t4_n99;
  t4_n96: if (E.getNumReferences() >= 3364454096 /*281590.5*/) goto t4_n98;
  t4_n97: return 0.12681595981121063f;
  t4_n98: return 0.13269546627998352f;
  t4_n99: return 0.13991065323352814f;
  t4_n100: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t4_n172;
  t4_n101: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t4_n155;
  t4_n102: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t4_n134;
  t4_n103: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t4_n111;
  t4_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t4_n108;
  t4_n105: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t4_n107;
  t4_n106: return 0.10796602070331573f;
  t4_n107: return 0.1202448382973671f;
  t4_n108: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n110;
  t4_n109: return 0.1290421485900879f;
  t4_n110: return 0.11950986087322235f;
  t4_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t4_n131;
  t4_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t4_n118;
  t4_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t4_n115;
  t4_n114: return 0.0817975178360939f;
  t4_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t4_n117;
  t4_n116: return -0.22471219301223755f;
  t4_n117: return 0.12516458332538605f;
  t4_n118: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t4_n126;
  t4_n119: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t4_n125;
  t4_n120: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t4_n122;
  t4_n121: return 0.11542798578739166f;
  t4_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t4_n124;
  t4_n123: return 0.11523205041885376f;
  t4_n124: return 0.12905703485012054f;
  t4_n125: return 0.1289408951997757f;
  t4_n126: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n130;
  t4_n127: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t4_n129;
  t4_n128: return 0.1279810667037964f;
  t4_n129: return 0.13146726787090302f;
  t4_n130: return 0.1215856671333313f;
  t4_n131: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t4_n133;
  t4_n132: return 0.11235614120960236f;
  t4_n133: return 0.1261405348777771f;
  t4_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t4_n152;
  t4_n135: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t4_n151;
  t4_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t4_n146;
  t4_n137: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t4_n143;
  t4_n138: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t4_n142;
  t4_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t4_n141;
  t4_n140: return 0.12506374716758728f;
  t4_n141: return 0.11906753480434418f;
  t4_n142: return 0.13390423357486725f;
  t4_n143: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t4_n145;
  t4_n144: return 0.09634672850370407f;
  t4_n145: return 0.12310579419136047f;
  t4_n146: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t4_n148;
  t4_n147: return 0.1334737092256546f;
  t4_n148: if (E.getNumReferences() >= 3317256192 /*5930.0*/) goto t4_n150;
  t4_n149: return 0.12648600339889526f;
  t4_n150: return 0.12938199937343597f;
  t4_n151: return 0.12807393074035645f;
  t4_n152: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t4_n154;
  t4_n153: return 0.13225343823432922f;
  t4_n154: return 0.12733213603496552f;
  t4_n155: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t4_n171;
  t4_n156: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t4_n162;
  t4_n157: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t4_n161;
  t4_n158: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n160;
  t4_n159: return 0.053574319928884506f;
  t4_n160: return 0.11103085428476334f;
  t4_n161: return 0.11490809172391891f;
  t4_n162: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n170;
  t4_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t4_n165;
  t4_n164: return 0.07258813083171844f;
  t4_n165: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t4_n169;
  t4_n166: if (E.getNumReferences() >= 3318322176 /*6450.5*/) goto t4_n168;
  t4_n167: return 0.11101918667554855f;
  t4_n168: return 0.12466850876808167f;
  t4_n169: return -0.12790387868881226f;
  t4_n170: return 0.12822479009628296f;
  t4_n171: return 0.13216887414455414f;
  t4_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t4_n208;
  t4_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t4_n175;
  t4_n174: return -0.12925612926483154f;
  t4_n175: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t4_n191;
  t4_n176: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t4_n180;
  t4_n177: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n179;
  t4_n178: return -0.05188321694731712f;
  t4_n179: return 0.06102824956178665f;
  t4_n180: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n190;
  t4_n181: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t4_n187;
  t4_n182: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n184;
  t4_n183: return 0.03326258808374405f;
  t4_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t4_n186;
  t4_n185: return 0.07453133910894394f;
  t4_n186: return 0.10729032009840012f;
  t4_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t4_n189;
  t4_n188: return 0.07936863601207733f;
  t4_n189: return 0.13099399209022522f;
  t4_n190: return 0.1198098435997963f;
  t4_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t4_n207;
  t4_n192: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t4_n202;
  t4_n193: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t4_n195;
  t4_n194: return 0.05169079452753067f;
  t4_n195: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n201;
  t4_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t4_n200;
  t4_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t4_n199;
  t4_n198: return 0.10297778248786926f;
  t4_n199: return 0.11692527681589127f;
  t4_n200: return 0.08279073983430862f;
  t4_n201: return 0.12500230967998505f;
  t4_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n206;
  t4_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t4_n205;
  t4_n204: return 0.1134251058101654f;
  t4_n205: return 0.12667934596538544f;
  t4_n206: return 0.13009746372699738f;
  t4_n207: return -0.13110211491584778f;
  t4_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t4_n220;
  t4_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t4_n217;
  t4_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t4_n212;
  t4_n211: return -0.08033233880996704f;
  t4_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t4_n214;
  t4_n213: return 0.019855936989188194f;
  t4_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t4_n216;
  t4_n215: return 0.12961725890636444f;
  t4_n216: return -0.12008111923933029f;
  t4_n217: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t4_n219;
  t4_n218: return 0.09035993367433548f;
  t4_n219: return 0.12928181886672974f;
  t4_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t4_n232;
  t4_n221: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t4_n225;
  t4_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t4_n224;
  t4_n223: return 0.06944558024406433f;
  t4_n224: return 0.10598810017108917f;
  t4_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t4_n231;
  t4_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t4_n228;
  t4_n227: return -0.029426006600260735f;
  t4_n228: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n230;
  t4_n229: return 0.09922802448272705f;
  t4_n230: return 0.12568232417106628f;
  t4_n231: return 0.12354519218206406f;
  t4_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t4_n236;
  t4_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t4_n235;
  t4_n234: return 0.12349254637956619f;
  t4_n235: return 0.11018656194210052f;
  t4_n236: return 0.1277197301387787f;
  t4_n237: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t4_n243;
  t4_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t4_n240;
  t4_n239: return 0.11145352572202682f;
  t4_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t4_n242;
  t4_n241: return 0.1320129930973053f;
  t4_n242: return 0.11011315137147903f;
  t4_n243: return 0.13273310661315918f;
  t4_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t4_n246;
  t4_n245: return 0.01475172396749258f;
  t4_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t4_n254;
  t4_n247: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t4_n251;
  t4_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t4_n250;
  t4_n249: return 0.12964262068271637f;
  t4_n250: return -0.0848785787820816f;
  t4_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t4_n253;
  t4_n252: return 0.130961611866951f;
  t4_n253: return 0.1300591379404068f;
  t4_n254: return 0.13218621909618378f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree5(const Example& E) {
  t5_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t5_n244;
  t5_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t5_n229;
  t5_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t5_n86;
  t5_n3: if (E.getNumReferences() >= 3332525568 /*20775.0*/) goto t5_n71;
  t5_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t5_n36;
  t5_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t5_n17;
  t5_n6: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t5_n14;
  t5_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t5_n9;
  t5_n8: return -0.03867683932185173f;
  t5_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n11;
  t5_n10: return 0.012622897513210773f;
  t5_n11: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t5_n13;
  t5_n12: return 0.09230829775333405f;
  t5_n13: return 0.03323188051581383f;
  t5_n14: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t5_n16;
  t5_n15: return 0.058409433811903f;
  t5_n16: return 0.09626113623380661f;
  t5_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t5_n25;
  t5_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n20;
  t5_n19: return -0.12005627155303955f;
  t5_n20: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t5_n24;
  t5_n21: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t5_n23;
  t5_n22: return -0.05616729333996773f;
  t5_n23: return 0.04615847393870354f;
  t5_n24: return 0.09073173999786377f;
  t5_n25: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t5_n35;
  t5_n26: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t5_n32;
  t5_n27: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t5_n29;
  t5_n28: return -0.053000837564468384f;
  t5_n29: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t5_n31;
  t5_n30: return 0.005759749095886946f;
  t5_n31: return 0.09917016327381134f;
  t5_n32: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t5_n34;
  t5_n33: return 0.053068507462739944f;
  t5_n34: return 0.10478129237890244f;
  t5_n35: return -0.09577184915542603f;
  t5_n36: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t5_n64;
  t5_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t5_n45;
  t5_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t5_n40;
  t5_n39: return -0.09242434799671173f;
  t5_n40: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t5_n42;
  t5_n41: return 0.0917908325791359f;
  t5_n42: if (E.getNumReferences() >= 3307290624 /*2580.0*/) goto t5_n44;
  t5_n43: return -0.005003747530281544f;
  t5_n44: return 0.09531768411397934f;
  t5_n45: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t5_n53;
  t5_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t5_n52;
  t5_n47: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t5_n49;
  t5_n48: return 0.05712578445672989f;
  t5_n49: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t5_n51;
  t5_n50: return 0.015538726933300495f;
  t5_n51: return -0.07188466936349869f;
  t5_n52: return -0.13263936340808868f;
  t5_n53: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t5_n57;
  t5_n54: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t5_n56;
  t5_n55: return 0.10010846704244614f;
  t5_n56: return -0.11828798055648804f;
  t5_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t5_n61;
  t5_n58: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t5_n60;
  t5_n59: return 0.06585148721933365f;
  t5_n60: return 0.007898946292698383f;
  t5_n61: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t5_n63;
  t5_n62: return -0.04686686396598816f;
  t5_n63: return 0.11538015305995941f;
  t5_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t5_n68;
  t5_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t5_n67;
  t5_n66: return 0.06468965113162994f;
  t5_n67: return 0.13672983646392822f;
  t5_n68: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t5_n70;
  t5_n69: return 0.10344748944044113f;
  t5_n70: return 0.1247379332780838f;
  t5_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t5_n81;
  t5_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n80;
  t5_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t5_n77;
  t5_n74: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t5_n76;
  t5_n75: return -0.005176777485758066f;
  t5_n76: return 0.09518218040466309f;
  t5_n77: if (E.getNumReferences() >= 3344974336 /*57410.0*/) goto t5_n79;
  t5_n78: return 0.06327199935913086f;
  t5_n79: return 0.100596584379673f;
  t5_n80: return 0.1220708116889f;
  t5_n81: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t5_n83;
  t5_n82: return 0.10937154293060303f;
  t5_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t5_n85;
  t5_n84: return 0.11830742657184601f;
  t5_n85: return 0.1310049146413803f;
  t5_n86: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t5_n158;
  t5_n87: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t5_n141;
  t5_n88: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t5_n120;
  t5_n89: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t5_n97;
  t5_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n94;
  t5_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t5_n93;
  t5_n92: return 0.09966325014829636f;
  t5_n93: return 0.1128840520977974f;
  t5_n94: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t5_n96;
  t5_n95: return 0.1226276233792305f;
  t5_n96: return 0.11189378798007965f;
  t5_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t5_n117;
  t5_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t5_n104;
  t5_n99: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t5_n101;
  t5_n100: return 0.06253930181264877f;
  t5_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n103;
  t5_n102: return -0.197727233171463f;
  t5_n103: return 0.11849427968263626f;
  t5_n104: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t5_n112;
  t5_n105: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t5_n111;
  t5_n106: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t5_n110;
  t5_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t5_n109;
  t5_n108: return 0.10919875651597977f;
  t5_n109: return 0.12216728180646896f;
  t5_n110: return 0.12197421491146088f;
  t5_n111: return 0.09808104485273361f;
  t5_n112: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t5_n116;
  t5_n113: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t5_n115;
  t5_n114: return 0.12112551927566528f;
  t5_n115: return 0.12514518201351166f;
  t5_n116: return 0.12131305038928986f;
  t5_n117: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t5_n119;
  t5_n118: return 0.10453781485557556f;
  t5_n119: return 0.11929851770401001f;
  t5_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n138;
  t5_n121: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t5_n137;
  t5_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t5_n134;
  t5_n123: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t5_n131;
  t5_n124: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t5_n130;
  t5_n125: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t5_n127;
  t5_n126: return 0.10889916121959686f;
  t5_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t5_n129;
  t5_n128: return 0.10037854313850403f;
  t5_n129: return 0.11950811743736267f;
  t5_n130: return 0.1271876096725464f;
  t5_n131: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t5_n133;
  t5_n132: return 0.09116452932357788f;
  t5_n133: return 0.11597698926925659f;
  t5_n134: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t5_n136;
  t5_n135: return 0.12727753818035126f;
  t5_n136: return 0.12167051434516907f;
  t5_n137: return 0.12161250412464142f;
  t5_n138: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t5_n140;
  t5_n139: return 0.12598246335983276f;
  t5_n140: return 0.1206311360001564f;
  t5_n141: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t5_n157;
  t5_n142: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t5_n148;
  t5_n143: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t5_n147;
  t5_n144: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n146;
  t5_n145: return 0.046348053961992264f;
  t5_n146: return 0.10321896523237228f;
  t5_n147: return 0.10691967606544495f;
  t5_n148: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n156;
  t5_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t5_n151;
  t5_n150: return 0.0658111572265625f;
  t5_n151: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t5_n155;
  t5_n152: if (E.getNumReferences() >= 3318322176 /*6450.5*/) goto t5_n154;
  t5_n153: return 0.10296797752380371f;
  t5_n154: return 0.11768489331007004f;
  t5_n155: return -0.1219232901930809f;
  t5_n156: return 0.1219838336110115f;
  t5_n157: return 0.12637147307395935f;
  t5_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t5_n200;
  t5_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t5_n161;
  t5_n160: return -0.12311381846666336f;
  t5_n161: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t5_n185;
  t5_n162: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t5_n174;
  t5_n163: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n173;
  t5_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n166;
  t5_n165: return -0.06090119853615761f;
  t5_n166: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t5_n170;
  t5_n167: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t5_n169;
  t5_n168: return 0.06055036187171936f;
  t5_n169: return -0.1448560357093811f;
  t5_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t5_n172;
  t5_n171: return 0.052775271236896515f;
  t5_n172: return 0.16989587247371674f;
  t5_n173: return 0.05509134382009506f;
  t5_n174: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n184;
  t5_n175: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t5_n181;
  t5_n176: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t5_n178;
  t5_n177: return 0.03006095066666603f;
  t5_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t5_n180;
  t5_n179: return 0.06876885890960693f;
  t5_n180: return 0.09961062669754028f;
  t5_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t5_n183;
  t5_n182: return 0.07118076831102371f;
  t5_n183: return 0.1242576539516449f;
  t5_n184: return 0.11258705705404282f;
  t5_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t5_n199;
  t5_n186: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t5_n198;
  t5_n187: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t5_n191;
  t5_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n190;
  t5_n189: return 0.022949177771806717f;
  t5_n190: return 0.1128898486495018f;
  t5_n191: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n197;
  t5_n192: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t5_n196;
  t5_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t5_n195;
  t5_n194: return 0.09456247836351395f;
  t5_n195: return 0.10921084880828857f;
  t5_n196: return 0.07377047091722488f;
  t5_n197: return 0.11815059930086136f;
  t5_n198: return 0.11629904806613922f;
  t5_n199: return -0.12502261996269226f;
  t5_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t5_n212;
  t5_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t5_n209;
  t5_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t5_n204;
  t5_n203: return -0.07612906396389008f;
  t5_n204: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t5_n206;
  t5_n205: return 0.019841529428958893f;
  t5_n206: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t5_n208;
  t5_n207: return 0.123448446393013f;
  t5_n208: return -0.11843422800302505f;
  t5_n209: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t5_n211;
  t5_n210: return 0.08187846094369888f;
  t5_n211: return 0.12301783263683319f;
  t5_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t5_n224;
  t5_n213: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t5_n217;
  t5_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t5_n216;
  t5_n215: return 0.061413273215293884f;
  t5_n216: return 0.09825242310762405f;
  t5_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t5_n223;
  t5_n218: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t5_n220;
  t5_n219: return -0.02601718157529831f;
  t5_n220: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n222;
  t5_n221: return 0.09026645869016647f;
  t5_n222: return 0.11904612928628922f;
  t5_n223: return 0.11661208420991898f;
  t5_n224: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t5_n228;
  t5_n225: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t5_n227;
  t5_n226: return 0.11664862185716629f;
  t5_n227: return 0.10152636468410492f;
  t5_n228: return 0.12133148312568665f;
  t5_n229: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t5_n237;
  t5_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t5_n236;
  t5_n231: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t5_n235;
  t5_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t5_n234;
  t5_n233: return 0.1122320368885994f;
  t5_n234: return 0.1274237185716629f;
  t5_n235: return 0.041670821607112885f;
  t5_n236: return 0.12684358656406403f;
  t5_n237: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t5_n243;
  t5_n238: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t5_n240;
  t5_n239: return 0.10269111394882202f;
  t5_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t5_n242;
  t5_n241: return 0.12587934732437134f;
  t5_n242: return 0.10176095366477966f;
  t5_n243: return 0.12677042186260223f;
  t5_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t5_n246;
  t5_n245: return 0.013267513364553452f;
  t5_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t5_n254;
  t5_n247: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t5_n251;
  t5_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t5_n250;
  t5_n249: return 0.12337252497673035f;
  t5_n250: return -0.07567253708839417f;
  t5_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t5_n253;
  t5_n252: return 0.12492799013853073f;
  t5_n253: return 0.12394796311855316f;
  t5_n254: return 0.12633147835731506f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree6(const Example& E) {
  t6_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t6_n246;
  t6_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t6_n231;
  t6_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t6_n92;
  t6_n3: if (E.getNumReferences() >= 3332525568 /*20775.0*/) goto t6_n71;
  t6_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t6_n48;
  t6_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t6_n25;
  t6_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t6_n14;
  t6_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t6_n11;
  t6_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n10;
  t6_n9: return -0.05275452509522438f;
  t6_n10: return 0.06824968755245209f;
  t6_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n13;
  t6_n12: return -0.005396819673478603f;
  t6_n13: return 0.06953218579292297f;
  t6_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t6_n20;
  t6_n15: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n17;
  t6_n16: return -0.11611585319042206f;
  t6_n17: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t6_n19;
  t6_n18: return -0.03576504439115524f;
  t6_n19: return 0.08532412350177765f;
  t6_n20: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t6_n24;
  t6_n21: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t6_n23;
  t6_n22: return -0.04978378117084503f;
  t6_n23: return 0.06300879269838333f;
  t6_n24: return -0.0918155387043953f;
  t6_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t6_n31;
  t6_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t6_n28;
  t6_n27: return -0.08499255031347275f;
  t6_n28: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t6_n30;
  t6_n29: return 0.08536478877067566f;
  t6_n30: return 0.0009990063263103366f;
  t6_n31: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t6_n39;
  t6_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t6_n34;
  t6_n33: return -0.006745863705873489f;
  t6_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t6_n36;
  t6_n35: return -0.015131774358451366f;
  t6_n36: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t6_n38;
  t6_n37: return 0.12490229308605194f;
  t6_n38: return 0.09432626515626907f;
  t6_n39: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t6_n43;
  t6_n40: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t6_n42;
  t6_n41: return 0.0940069928765297f;
  t6_n42: return -0.11254339665174484f;
  t6_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t6_n45;
  t6_n44: return 0.029358793050050735f;
  t6_n45: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t6_n47;
  t6_n46: return -0.041714493185281754f;
  t6_n47: return 0.10808639228343964f;
  t6_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t6_n64;
  t6_n49: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t6_n59;
  t6_n50: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t6_n56;
  t6_n51: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t6_n55;
  t6_n52: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n54;
  t6_n53: return 0.0658392459154129f;
  t6_n54: return 0.1148306354880333f;
  t6_n55: return 0.010585331358015537f;
  t6_n56: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t6_n58;
  t6_n57: return 0.08605583012104034f;
  t6_n58: return 0.11804854869842529f;
  t6_n59: if (E.getNumReferences() >= 3293208576 /*809.5*/) goto t6_n63;
  t6_n60: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n62;
  t6_n61: return -0.053400669246912f;
  t6_n62: return 0.08736086636781693f;
  t6_n63: return 0.11483436077833176f;
  t6_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t6_n68;
  t6_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t6_n67;
  t6_n66: return 0.0729563981294632f;
  t6_n67: return 0.13361084461212158f;
  t6_n68: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t6_n70;
  t6_n69: return 0.09484327584505081f;
  t6_n70: return 0.12096944451332092f;
  t6_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t6_n87;
  t6_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n86;
  t6_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t6_n77;
  t6_n74: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t6_n76;
  t6_n75: return -0.0013135613407939672f;
  t6_n76: return 0.09381265193223953f;
  t6_n77: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t6_n83;
  t6_n78: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t6_n82;
  t6_n79: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t6_n81;
  t6_n80: return 0.0891069695353508f;
  t6_n81: return 0.10602197796106339f;
  t6_n82: return 0.03912288323044777f;
  t6_n83: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t6_n85;
  t6_n84: return 0.11842267215251923f;
  t6_n85: return 0.06042105332016945f;
  t6_n86: return 0.11670512706041336f;
  t6_n87: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t6_n89;
  t6_n88: return 0.10246049612760544f;
  t6_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t6_n91;
  t6_n90: return 0.11454898864030838f;
  t6_n91: return 0.13171100616455078f;
  t6_n92: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t6_n156;
  t6_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t6_n141;
  t6_n94: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t6_n124;
  t6_n95: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t6_n103;
  t6_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n100;
  t6_n97: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t6_n99;
  t6_n98: return 0.09268879890441895f;
  t6_n99: return 0.10699354857206345f;
  t6_n100: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n102;
  t6_n101: return 0.11477108299732208f;
  t6_n102: return 0.09493549913167953f;
  t6_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t6_n121;
  t6_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t6_n110;
  t6_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n107;
  t6_n106: return 0.05461925268173218f;
  t6_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n109;
  t6_n108: return -0.18433918058872223f;
  t6_n109: return 0.11318471282720566f;
  t6_n110: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t6_n118;
  t6_n111: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t6_n117;
  t6_n112: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t6_n116;
  t6_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n115;
  t6_n114: return 0.102376788854599f;
  t6_n115: return 0.11699502915143967f;
  t6_n116: return 0.11664564907550812f;
  t6_n117: return 0.09175629168748856f;
  t6_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t6_n120;
  t6_n119: return 0.10767804831266403f;
  t6_n120: return 0.12100794166326523f;
  t6_n121: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t6_n123;
  t6_n122: return 0.09821224957704544f;
  t6_n123: return 0.11392194777727127f;
  t6_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n138;
  t6_n125: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t6_n137;
  t6_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t6_n134;
  t6_n127: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t6_n131;
  t6_n128: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t6_n130;
  t6_n129: return 0.11085257679224014f;
  t6_n130: return 0.12212371081113815f;
  t6_n131: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t6_n133;
  t6_n132: return 0.08356772363185883f;
  t6_n133: return 0.1101066917181015f;
  t6_n134: if (E.getNumReferences() >= 3308503040 /*2876.0*/) goto t6_n136;
  t6_n135: return 0.1180632933974266f;
  t6_n136: return 0.11918322741985321f;
  t6_n137: return 0.11669336259365082f;
  t6_n138: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t6_n140;
  t6_n139: return 0.12129069864749908f;
  t6_n140: return 0.11526037752628326f;
  t6_n141: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t6_n155;
  t6_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n152;
  t6_n143: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t6_n147;
  t6_n144: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t6_n146;
  t6_n145: return 0.03734089434146881f;
  t6_n146: return 0.08507206290960312f;
  t6_n147: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t6_n151;
  t6_n148: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t6_n150;
  t6_n149: return 0.09955451637506485f;
  t6_n150: return 0.11300400644540787f;
  t6_n151: return -0.1176115870475769f;
  t6_n152: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t6_n154;
  t6_n153: return 0.09763405472040176f;
  t6_n154: return 0.1164829358458519f;
  t6_n155: return 0.12192479521036148f;
  t6_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t6_n202;
  t6_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t6_n159;
  t6_n158: return -0.1185428723692894f;
  t6_n159: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t6_n179;
  t6_n160: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n172;
  t6_n161: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n171;
  t6_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n164;
  t6_n163: return -0.05446433648467064f;
  t6_n164: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t6_n168;
  t6_n165: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t6_n167;
  t6_n166: return 0.06412170082330704f;
  t6_n167: return -0.139354407787323f;
  t6_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t6_n170;
  t6_n169: return 0.04789343848824501f;
  t6_n170: return 0.15410876274108887f;
  t6_n171: return 0.05050477012991905f;
  t6_n172: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t6_n178;
  t6_n173: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t6_n177;
  t6_n174: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t6_n176;
  t6_n175: return 0.023448312655091286f;
  t6_n176: return 0.08734987676143646f;
  t6_n177: return 0.10842875391244888f;
  t6_n178: return -0.1209210529923439f;
  t6_n179: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n199;
  t6_n180: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t6_n198;
  t6_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n189;
  t6_n182: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t6_n186;
  t6_n183: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t6_n185;
  t6_n184: return 0.027010638266801834f;
  t6_n185: return 0.06716262549161911f;
  t6_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t6_n188;
  t6_n187: return 0.06454640626907349f;
  t6_n188: return 0.11917995661497116f;
  t6_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t6_n197;
  t6_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t6_n196;
  t6_n191: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t6_n195;
  t6_n192: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t6_n194;
  t6_n193: return 0.054265812039375305f;
  t6_n194: return 0.09520072489976883f;
  t6_n195: return 0.11294608563184738f;
  t6_n196: return 0.11434587836265564f;
  t6_n197: return 0.048782337456941605f;
  t6_n198: return -0.11461447179317474f;
  t6_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t6_n201;
  t6_n200: return 0.10715899616479874f;
  t6_n201: return 0.11213543266057968f;
  t6_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t6_n214;
  t6_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t6_n211;
  t6_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t6_n206;
  t6_n205: return -0.07179585844278336f;
  t6_n206: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t6_n208;
  t6_n207: return 0.0189349427819252f;
  t6_n208: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t6_n210;
  t6_n209: return 0.11877691000699997f;
  t6_n210: return -0.10939878225326538f;
  t6_n211: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t6_n213;
  t6_n212: return 0.07542522996664047f;
  t6_n213: return 0.11818340420722961f;
  t6_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t6_n226;
  t6_n215: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t6_n219;
  t6_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t6_n218;
  t6_n217: return 0.05498604476451874f;
  t6_n218: return 0.09124597162008286f;
  t6_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t6_n225;
  t6_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n222;
  t6_n221: return -0.020818689838051796f;
  t6_n222: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n224;
  t6_n223: return 0.08322066813707352f;
  t6_n224: return 0.11383268237113953f;
  t6_n225: return 0.11093156784772873f;
  t6_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t6_n230;
  t6_n227: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t6_n229;
  t6_n228: return 0.11093226075172424f;
  t6_n229: return 0.09452306479215622f;
  t6_n230: return 0.11629190295934677f;
  t6_n231: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t6_n239;
  t6_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t6_n238;
  t6_n233: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t6_n237;
  t6_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t6_n236;
  t6_n235: return 0.10691746324300766f;
  t6_n236: return 0.12213660776615143f;
  t6_n237: return 0.037479568272829056f;
  t6_n238: return 0.12231266498565674f;
  t6_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t6_n245;
  t6_n240: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t6_n242;
  t6_n241: return 0.09548179060220718f;
  t6_n242: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t6_n244;
  t6_n243: return 0.12106719613075256f;
  t6_n244: return 0.09497188031673431f;
  t6_n245: return 0.12219524383544922f;
  t6_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t6_n248;
  t6_n247: return 0.012289801612496376f;
  t6_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t6_n254;
  t6_n249: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t6_n251;
  t6_n250: return 0.11853868514299393f;
  t6_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t6_n253;
  t6_n252: return 0.12038695812225342f;
  t6_n253: return 0.11935454607009888f;
  t6_n254: return 0.12200726568698883f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree7(const Example& E) {
  t7_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t7_n244;
  t7_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t7_n231;
  t7_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t7_n104;
  t7_n3: if (E.getNumReferences() >= 3332525568 /*20775.0*/) goto t7_n79;
  t7_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t7_n54;
  t7_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t7_n31;
  t7_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t7_n16;
  t7_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t7_n13;
  t7_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n10;
  t7_n9: return -0.048218775540590286f;
  t7_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t7_n12;
  t7_n11: return 0.039144642651081085f;
  t7_n12: return 0.1184224858880043f;
  t7_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t7_n15;
  t7_n14: return -0.004730020184069872f;
  t7_n15: return 0.06400088965892792f;
  t7_n16: if (E.getNumReferences() >= 3282321408 /*328.5*/) goto t7_n28;
  t7_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t7_n21;
  t7_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n20;
  t7_n19: return -0.11559450626373291f;
  t7_n20: return -0.03226667270064354f;
  t7_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t7_n27;
  t7_n22: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t7_n26;
  t7_n23: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t7_n25;
  t7_n24: return -0.04431630298495293f;
  t7_n25: return 0.03070029616355896f;
  t7_n26: return -0.08222733438014984f;
  t7_n27: return -0.1107359379529953f;
  t7_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n30;
  t7_n29: return -0.039909571409225464f;
  t7_n30: return 0.0674118921160698f;
  t7_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t7_n37;
  t7_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t7_n34;
  t7_n33: return -0.07826213538646698f;
  t7_n34: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t7_n36;
  t7_n35: return 0.07567574083805084f;
  t7_n36: return -0.003465155838057399f;
  t7_n37: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t7_n45;
  t7_n38: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t7_n44;
  t7_n39: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t7_n41;
  t7_n40: return 0.049766071140766144f;
  t7_n41: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t7_n43;
  t7_n42: return 0.011183121241629124f;
  t7_n43: return -0.06567926704883575f;
  t7_n44: return -0.12409932166337967f;
  t7_n45: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t7_n49;
  t7_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t7_n48;
  t7_n47: return 0.08850919455289841f;
  t7_n48: return -0.10773701220750809f;
  t7_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t7_n51;
  t7_n50: return 0.00837669800966978f;
  t7_n51: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t7_n53;
  t7_n52: return -0.01096459198743105f;
  t7_n53: return 0.06651464104652405f;
  t7_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t7_n72;
  t7_n55: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t7_n65;
  t7_n56: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t7_n62;
  t7_n57: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t7_n61;
  t7_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n60;
  t7_n59: return 0.059674911201000214f;
  t7_n60: return 0.11096393316984177f;
  t7_n61: return 0.017550913617014885f;
  t7_n62: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t7_n64;
  t7_n63: return 0.0836709588766098f;
  t7_n64: return 0.11643993854522705f;
  t7_n65: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n71;
  t7_n66: if (E.getNumReferences() >= 3292758016 /*782.0*/) goto t7_n70;
  t7_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t7_n69;
  t7_n68: return -0.07261008769273758f;
  t7_n69: return 0.018843552097678185f;
  t7_n70: return 0.10434840619564056f;
  t7_n71: return 0.09103723615407944f;
  t7_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t7_n76;
  t7_n73: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t7_n75;
  t7_n74: return 0.07183738052845001f;
  t7_n75: return 0.12662257254123688f;
  t7_n76: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t7_n78;
  t7_n77: return 0.0892573744058609f;
  t7_n78: return 0.11743985116481781f;
  t7_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t7_n95;
  t7_n80: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n94;
  t7_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t7_n85;
  t7_n82: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t7_n84;
  t7_n83: return -7.469265983672813e-05f;
  t7_n84: return 0.08635824918746948f;
  t7_n85: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t7_n89;
  t7_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t7_n88;
  t7_n87: return 0.08925529569387436f;
  t7_n88: return 0.03221891447901726f;
  t7_n89: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t7_n93;
  t7_n90: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t7_n92;
  t7_n91: return 0.12250697612762451f;
  t7_n92: return 0.10770054161548615f;
  t7_n93: return 0.05158914625644684f;
  t7_n94: return 0.11251743137836456f;
  t7_n95: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t7_n97;
  t7_n96: return 0.09561421722173691f;
  t7_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t7_n99;
  t7_n98: return 0.11031404882669449f;
  t7_n99: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t7_n103;
  t7_n100: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t7_n102;
  t7_n101: return 0.11386752873659134f;
  t7_n102: return 0.13595525920391083f;
  t7_n103: return 0.023185523226857185f;
  t7_n104: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t7_n158;
  t7_n105: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t7_n141;
  t7_n106: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t7_n126;
  t7_n107: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t7_n113;
  t7_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t7_n110;
  t7_n109: return 0.08810285478830338f;
  t7_n110: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n112;
  t7_n111: return 0.11043140292167664f;
  t7_n112: return 0.08854340761899948f;
  t7_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t7_n123;
  t7_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t7_n118;
  t7_n115: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n117;
  t7_n116: return 0.04769017547369003f;
  t7_n117: return 0.10808912664651871f;
  t7_n118: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t7_n120;
  t7_n119: return 0.10318810492753983f;
  t7_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t7_n122;
  t7_n121: return 0.10203713923692703f;
  t7_n122: return 0.11723780632019043f;
  t7_n123: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t7_n125;
  t7_n124: return 0.09205096960067749f;
  t7_n125: return 0.10961083322763443f;
  t7_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t7_n140;
  t7_n127: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t7_n139;
  t7_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t7_n136;
  t7_n129: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t7_n133;
  t7_n130: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t7_n132;
  t7_n131: return 0.1060887798666954f;
  t7_n132: return 0.11826707422733307f;
  t7_n133: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t7_n135;
  t7_n134: return 0.0770397037267685f;
  t7_n135: return 0.1055036261677742f;
  t7_n136: if (E.getNumReferences() >= 3308503040 /*2876.0*/) goto t7_n138;
  t7_n137: return 0.11365600675344467f;
  t7_n138: return 0.11522064357995987f;
  t7_n139: return 0.11277537047863007f;
  t7_n140: return 0.1143830195069313f;
  t7_n141: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t7_n157;
  t7_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n154;
  t7_n143: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t7_n149;
  t7_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t7_n148;
  t7_n145: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t7_n147;
  t7_n146: return 0.05284883454442024f;
  t7_n147: return 0.08966568112373352f;
  t7_n148: return 0.0009452869999222457f;
  t7_n149: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t7_n153;
  t7_n150: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t7_n152;
  t7_n151: return 0.09368475526571274f;
  t7_n152: return 0.10851357877254486f;
  t7_n153: return -0.11405528336763382f;
  t7_n154: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t7_n156;
  t7_n155: return 0.09252406656742096f;
  t7_n156: return 0.11224037408828735f;
  t7_n157: return 0.11866112053394318f;
  t7_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t7_n202;
  t7_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t7_n161;
  t7_n160: return -0.11501043289899826f;
  t7_n161: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t7_n179;
  t7_n162: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n172;
  t7_n163: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n171;
  t7_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t7_n166;
  t7_n165: return -0.05700121819972992f;
  t7_n166: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t7_n168;
  t7_n167: return -0.00970158725976944f;
  t7_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t7_n170;
  t7_n169: return 0.010265559889376163f;
  t7_n170: return 0.14247189462184906f;
  t7_n171: return 0.04503646120429039f;
  t7_n172: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t7_n178;
  t7_n173: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t7_n177;
  t7_n174: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t7_n176;
  t7_n175: return 0.020855357870459557f;
  t7_n176: return 0.08062005788087845f;
  t7_n177: return 0.10316433757543564f;
  t7_n178: return -0.11739140003919601f;
  t7_n179: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n199;
  t7_n180: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t7_n198;
  t7_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n189;
  t7_n182: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t7_n186;
  t7_n183: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t7_n185;
  t7_n184: return 0.024755064398050308f;
  t7_n185: return 0.06211823597550392f;
  t7_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t7_n188;
  t7_n187: return 0.060432568192481995f;
  t7_n188: return 0.11497634649276733f;
  t7_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t7_n197;
  t7_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t7_n196;
  t7_n191: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t7_n195;
  t7_n192: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t7_n194;
  t7_n193: return 0.05089675635099411f;
  t7_n194: return 0.0896000862121582f;
  t7_n195: return 0.10853725671768188f;
  t7_n196: return 0.10984078794717789f;
  t7_n197: return 0.04241855442523956f;
  t7_n198: return -0.11038874834775925f;
  t7_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t7_n201;
  t7_n200: return 0.10171238332986832f;
  t7_n201: return 0.10743233561515808f;
  t7_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t7_n214;
  t7_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t7_n211;
  t7_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t7_n206;
  t7_n205: return -0.06647530943155289f;
  t7_n206: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t7_n208;
  t7_n207: return 0.017961274832487106f;
  t7_n208: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t7_n210;
  t7_n209: return 0.11499027162790298f;
  t7_n210: return -0.1012706309556961f;
  t7_n211: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t7_n213;
  t7_n212: return 0.06798683851957321f;
  t7_n213: return 0.11436542123556137f;
  t7_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t7_n226;
  t7_n215: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t7_n219;
  t7_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t7_n218;
  t7_n217: return 0.04834944009780884f;
  t7_n218: return 0.08528122305870056f;
  t7_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t7_n225;
  t7_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n222;
  t7_n221: return -0.01868138462305069f;
  t7_n222: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n224;
  t7_n223: return 0.0769311785697937f;
  t7_n224: return 0.10946986824274063f;
  t7_n225: return 0.1060892716050148f;
  t7_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t7_n230;
  t7_n227: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t7_n229;
  t7_n228: return 0.10621825605630875f;
  t7_n229: return 0.08815775066614151f;
  t7_n230: return 0.11219702661037445f;
  t7_n231: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t7_n239;
  t7_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t7_n238;
  t7_n233: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t7_n237;
  t7_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t7_n236;
  t7_n235: return 0.10024236142635345f;
  t7_n236: return 0.11775520443916321f;
  t7_n237: return 0.032140545547008514f;
  t7_n238: return 0.11879107356071472f;
  t7_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t7_n243;
  t7_n240: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t7_n242;
  t7_n241: return 0.08890262991189957f;
  t7_n242: return 0.11662090569734573f;
  t7_n243: return 0.11873926967382431f;
  t7_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t7_n246;
  t7_n245: return 0.011305419728159904f;
  t7_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t7_n254;
  t7_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t7_n253;
  t7_n248: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t7_n250;
  t7_n249: return 0.11472802609205246f;
  t7_n250: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t7_n252;
  t7_n251: return 0.11689451336860657f;
  t7_n252: return 0.11583173274993896f;
  t7_n253: return -0.09141149371862411f;
  t7_n254: return 0.11875754594802856f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree8(const Example& E) {
  t8_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t8_n244;
  t8_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t8_n231;
  t8_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t8_n82;
  t8_n3: if (E.getNumReferences() >= 3331404288 /*18585.0*/) goto t8_n63;
  t8_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t8_n36;
  t8_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t8_n21;
  t8_n6: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t8_n12;
  t8_n7: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t8_n9;
  t8_n8: return -0.02406221628189087f;
  t8_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n11;
  t8_n10: return 0.03608322516083717f;
  t8_n11: return 0.1155393049120903f;
  t8_n12: if (E.getNumReferences() >= 3282321408 /*328.5*/) goto t8_n18;
  t8_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t8_n17;
  t8_n14: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n16;
  t8_n15: return -0.1125863716006279f;
  t8_n16: return -0.029584020376205444f;
  t8_n17: return -0.07290451973676682f;
  t8_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n20;
  t8_n19: return -0.03490583598613739f;
  t8_n20: return 0.0627601370215416f;
  t8_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t8_n25;
  t8_n22: if (E.getNumReferences() >= 3291488256 /*704.5*/) goto t8_n24;
  t8_n23: return -0.051254455000162125f;
  t8_n24: return 0.07630355656147003f;
  t8_n25: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t8_n27;
  t8_n26: return 0.00021914299577474594f;
  t8_n27: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t8_n31;
  t8_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t8_n30;
  t8_n29: return 0.0834164023399353f;
  t8_n30: return -0.10392892360687256f;
  t8_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t8_n33;
  t8_n32: return 0.02310517244040966f;
  t8_n33: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t8_n35;
  t8_n34: return -0.04150616005063057f;
  t8_n35: return 0.09861452132463455f;
  t8_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t8_n56;
  t8_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t8_n47;
  t8_n38: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t8_n44;
  t8_n39: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t8_n43;
  t8_n40: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n42;
  t8_n41: return 0.05722584202885628f;
  t8_n42: return 0.10707052052021027f;
  t8_n43: return 0.01860024221241474f;
  t8_n44: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t8_n46;
  t8_n45: return 0.08262083679437637f;
  t8_n46: return 0.11385874450206757f;
  t8_n47: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n55;
  t8_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t8_n50;
  t8_n49: return -0.06372235715389252f;
  t8_n50: if (E.getNumReferences() >= 3293061120 /*800.5*/) goto t8_n54;
  t8_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n53;
  t8_n52: return -0.0026584379374980927f;
  t8_n53: return 0.07407286018133163f;
  t8_n54: return 0.10765106230974197f;
  t8_n55: return 0.0830330178141594f;
  t8_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t8_n60;
  t8_n57: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t8_n59;
  t8_n58: return 0.07141368836164474f;
  t8_n59: return 0.11972648650407791f;
  t8_n60: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t8_n62;
  t8_n61: return 0.08380730450153351f;
  t8_n62: return 0.11368228495121002f;
  t8_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t8_n73;
  t8_n64: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n72;
  t8_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t8_n69;
  t8_n66: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t8_n68;
  t8_n67: return -0.0011934232898056507f;
  t8_n68: return 0.0793924406170845f;
  t8_n69: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t8_n71;
  t8_n70: return 0.06925846636295319f;
  t8_n71: return 0.10912179201841354f;
  t8_n72: return 0.1090693548321724f;
  t8_n73: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t8_n75;
  t8_n74: return 0.08798039704561234f;
  t8_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t8_n77;
  t8_n76: return 0.10661464929580688f;
  t8_n77: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t8_n81;
  t8_n78: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t8_n80;
  t8_n79: return 0.11109252274036407f;
  t8_n80: return 0.133000910282135f;
  t8_n81: return 0.016738926991820335f;
  t8_n82: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t8_n144;
  t8_n83: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t8_n129;
  t8_n84: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t8_n110;
  t8_n85: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t8_n99;
  t8_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n98;
  t8_n87: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t8_n89;
  t8_n88: return 0.08039990812540054f;
  t8_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t8_n97;
  t8_n90: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n96;
  t8_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t8_n93;
  t8_n92: return 0.04086792469024658f;
  t8_n93: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t8_n95;
  t8_n94: return 0.08607320487499237f;
  t8_n95: return 0.11825152486562729f;
  t8_n96: return 0.10292874276638031f;
  t8_n97: return 0.1031741127371788f;
  t8_n98: return 0.10252160578966141f;
  t8_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n105;
  t8_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t8_n104;
  t8_n101: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n103;
  t8_n102: return 0.045262016355991364f;
  t8_n103: return 0.1039481982588768f;
  t8_n104: return 0.10517904907464981f;
  t8_n105: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t8_n107;
  t8_n106: return 0.10349558293819427f;
  t8_n107: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t8_n109;
  t8_n108: return 0.11371488869190216f;
  t8_n109: return 0.09562668949365616f;
  t8_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n128;
  t8_n111: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t8_n123;
  t8_n112: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t8_n122;
  t8_n113: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t8_n117;
  t8_n114: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t8_n116;
  t8_n115: return 0.09338921308517456f;
  t8_n116: return 0.10547604411840439f;
  t8_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t8_n121;
  t8_n118: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t8_n120;
  t8_n119: return 0.05293648689985275f;
  t8_n120: return 0.10512237995862961f;
  t8_n121: return 0.10575203597545624f;
  t8_n122: return 0.10881774127483368f;
  t8_n123: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t8_n125;
  t8_n124: return 0.11282245069742203f;
  t8_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t8_n127;
  t8_n126: return 0.06752335280179977f;
  t8_n127: return 0.11250010877847672f;
  t8_n128: return 0.11112851649522781f;
  t8_n129: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t8_n143;
  t8_n130: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n140;
  t8_n131: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t8_n135;
  t8_n132: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t8_n134;
  t8_n133: return 0.02793951705098152f;
  t8_n134: return 0.07405000925064087f;
  t8_n135: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t8_n139;
  t8_n136: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t8_n138;
  t8_n137: return 0.0880219116806984f;
  t8_n138: return 0.1046171560883522f;
  t8_n139: return -0.11154373735189438f;
  t8_n140: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t8_n142;
  t8_n141: return 0.0868542268872261f;
  t8_n142: return 0.1084330826997757f;
  t8_n143: return 0.1162441223859787f;
  t8_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t8_n204;
  t8_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t8_n147;
  t8_n146: return -0.1120966300368309f;
  t8_n147: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t8_n175;
  t8_n148: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n168;
  t8_n149: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t8_n165;
  t8_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n152;
  t8_n151: return -0.06267209351062775f;
  t8_n152: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t8_n162;
  t8_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t8_n157;
  t8_n154: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n156;
  t8_n155: return -0.014679930172860622f;
  t8_n156: return 0.05581405386328697f;
  t8_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n159;
  t8_n158: return -0.015689874067902565f;
  t8_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t8_n161;
  t8_n160: return 0.12675103545188904f;
  t8_n161: return -0.002790498547255993f;
  t8_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t8_n164;
  t8_n163: return 0.059732913970947266f;
  t8_n164: return 0.13417595624923706f;
  t8_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n167;
  t8_n166: return -0.09906323999166489f;
  t8_n167: return 0.07240574806928635f;
  t8_n168: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t8_n174;
  t8_n169: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t8_n173;
  t8_n170: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t8_n172;
  t8_n171: return 0.018281370401382446f;
  t8_n172: return 0.07467237114906311f;
  t8_n173: return 0.09866443276405334f;
  t8_n174: return -0.1143074780702591f;
  t8_n175: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n201;
  t8_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n192;
  t8_n177: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t8_n191;
  t8_n178: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n186;
  t8_n179: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t8_n185;
  t8_n180: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t8_n182;
  t8_n181: return 0.017367146909236908f;
  t8_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t8_n184;
  t8_n183: return 0.030954236164689064f;
  t8_n184: return 0.11229630559682846f;
  t8_n185: return 0.054676514118909836f;
  t8_n186: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t8_n190;
  t8_n187: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t8_n189;
  t8_n188: return -0.02728566899895668f;
  t8_n189: return 0.0905829519033432f;
  t8_n190: return 0.07921449095010757f;
  t8_n191: return -0.10598009079694748f;
  t8_n192: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t8_n200;
  t8_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n199;
  t8_n194: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n196;
  t8_n195: return 0.06347216665744781f;
  t8_n196: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t8_n198;
  t8_n197: return 0.08440050482749939f;
  t8_n198: return 0.10836010426282883f;
  t8_n199: return 0.10489145666360855f;
  t8_n200: return -0.11123615503311157f;
  t8_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t8_n203;
  t8_n202: return 0.09722563624382019f;
  t8_n203: return 0.10322580486536026f;
  t8_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n216;
  t8_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t8_n213;
  t8_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t8_n208;
  t8_n207: return -0.06265988200902939f;
  t8_n208: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t8_n210;
  t8_n209: return 0.017924780026078224f;
  t8_n210: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t8_n212;
  t8_n211: return 0.11217091977596283f;
  t8_n212: return -0.0964498519897461f;
  t8_n213: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t8_n215;
  t8_n214: return 0.06159818172454834f;
  t8_n215: return 0.11158299446105957f;
  t8_n216: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t8_n228;
  t8_n217: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t8_n221;
  t8_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t8_n220;
  t8_n219: return 0.04335322231054306f;
  t8_n220: return 0.07972396165132523f;
  t8_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t8_n227;
  t8_n222: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n224;
  t8_n223: return -0.01504459697753191f;
  t8_n224: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n226;
  t8_n225: return 0.07079628854990005f;
  t8_n226: return 0.1062440276145935f;
  t8_n227: return 0.10175447165966034f;
  t8_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t8_n230;
  t8_n229: return 0.09997715801000595f;
  t8_n230: return 0.10891495645046234f;
  t8_n231: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t8_n239;
  t8_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t8_n238;
  t8_n233: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t8_n237;
  t8_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t8_n236;
  t8_n235: return 0.09358306229114532f;
  t8_n236: return 0.11413650959730148f;
  t8_n237: return 0.0283666905015707f;
  t8_n238: return 0.1160721406340599f;
  t8_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t8_n243;
  t8_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t8_n242;
  t8_n241: return 0.11309314519166946f;
  t8_n242: return 0.061868198215961456f;
  t8_n243: return 0.11610662937164307f;
  t8_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t8_n246;
  t8_n245: return 0.009700912050902843f;
  t8_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t8_n254;
  t8_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t8_n253;
  t8_n248: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t8_n250;
  t8_n249: return 0.11154477298259735f;
  t8_n250: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t8_n252;
  t8_n251: return 0.11416026949882507f;
  t8_n252: return 0.11301549524068832f;
  t8_n253: return -0.08490672707557678f;
  t8_n254: return 0.11622261255979538f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree9(const Example& E) {
  t9_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t9_n244;
  t9_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t9_n231;
  t9_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t9_n78;
  t9_n3: if (E.getNumReferences() >= 3331404288 /*18585.0*/) goto t9_n59;
  t9_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t9_n36;
  t9_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t9_n25;
  t9_n6: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t9_n16;
  t9_n7: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t9_n9;
  t9_n8: return -0.030893277376890182f;
  t9_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n15;
  t9_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t9_n14;
  t9_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t9_n13;
  t9_n12: return -0.004655881319195032f;
  t9_n13: return 0.060269929468631744f;
  t9_n14: return -0.029615288600325584f;
  t9_n15: return 0.11454739421606064f;
  t9_n16: if (E.getNumReferences() >= 3282321408 /*328.5*/) goto t9_n22;
  t9_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n21;
  t9_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t9_n20;
  t9_n19: return -0.1101599857211113f;
  t9_n20: return -0.07365848869085312f;
  t9_n21: return -0.02609551139175892f;
  t9_n22: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n24;
  t9_n23: return -0.0320189967751503f;
  t9_n24: return 0.05762016773223877f;
  t9_n25: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t9_n33;
  t9_n26: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t9_n30;
  t9_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t9_n29;
  t9_n28: return -0.05365622043609619f;
  t9_n29: return 0.004898638930171728f;
  t9_n30: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t9_n32;
  t9_n31: return 0.102329783141613f;
  t9_n32: return 0.039892133325338364f;
  t9_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t9_n35;
  t9_n34: return 0.014834370464086533f;
  t9_n35: return 0.04769543185830116f;
  t9_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t9_n54;
  t9_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t9_n47;
  t9_n38: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t9_n44;
  t9_n39: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n43;
  t9_n40: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t9_n42;
  t9_n41: return 0.051889970898628235f;
  t9_n42: return 0.0118778757750988f;
  t9_n43: return 0.0949343666434288f;
  t9_n44: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t9_n46;
  t9_n45: return 0.07967337220907211f;
  t9_n46: return 0.10968664288520813f;
  t9_n47: if (E.getNumReferences() >= 3273687040 /*160.5*/) goto t9_n53;
  t9_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n52;
  t9_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t9_n51;
  t9_n50: return -0.06096794828772545f;
  t9_n51: return 0.018718749284744263f;
  t9_n52: return 0.07365833222866058f;
  t9_n53: return 0.08917273581027985f;
  t9_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t9_n58;
  t9_n55: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t9_n57;
  t9_n56: return 0.06824296712875366f;
  t9_n57: return 0.11313112080097198f;
  t9_n58: return 0.107003353536129f;
  t9_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t9_n69;
  t9_n60: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n68;
  t9_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t9_n65;
  t9_n62: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t9_n64;
  t9_n63: return -0.0006972107803449035f;
  t9_n64: return 0.07372178137302399f;
  t9_n65: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t9_n67;
  t9_n66: return 0.0659128874540329f;
  t9_n67: return 0.1054917424917221f;
  t9_n68: return 0.10622762888669968f;
  t9_n69: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t9_n71;
  t9_n70: return 0.08347876369953156f;
  t9_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t9_n73;
  t9_n72: return 0.10328451544046402f;
  t9_n73: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t9_n77;
  t9_n74: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t9_n76;
  t9_n75: return 0.10872294008731842f;
  t9_n76: return 0.1306021362543106f;
  t9_n77: return 0.013410818763077259f;
  t9_n78: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t9_n134;
  t9_n79: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t9_n119;
  t9_n80: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t9_n104;
  t9_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t9_n95;
  t9_n82: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t9_n84;
  t9_n83: return 0.07429377734661102f;
  t9_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t9_n94;
  t9_n85: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n93;
  t9_n86: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t9_n88;
  t9_n87: return 0.001502872328273952f;
  t9_n88: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t9_n92;
  t9_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t9_n91;
  t9_n90: return 0.08672243356704712f;
  t9_n91: return 0.13770084083080292f;
  t9_n92: return 0.06668078154325485f;
  t9_n93: return 0.10183198750019073f;
  t9_n94: return 0.1008056253194809f;
  t9_n95: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t9_n101;
  t9_n96: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t9_n100;
  t9_n97: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t9_n99;
  t9_n98: return 0.09664366394281387f;
  t9_n99: return 0.10439250618219376f;
  t9_n100: return 0.08290252834558487f;
  t9_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t9_n103;
  t9_n102: return 0.0803273543715477f;
  t9_n103: return 0.1090412512421608f;
  t9_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t9_n116;
  t9_n105: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t9_n115;
  t9_n106: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t9_n114;
  t9_n107: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t9_n111;
  t9_n108: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t9_n110;
  t9_n109: return 0.08890587091445923f;
  t9_n110: return 0.10156642645597458f;
  t9_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t9_n113;
  t9_n112: return 0.0614229179918766f;
  t9_n113: return 0.10230270773172379f;
  t9_n114: return 0.1060125008225441f;
  t9_n115: return 0.10783354192972183f;
  t9_n116: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t9_n118;
  t9_n117: return 0.11123453080654144f;
  t9_n118: return 0.10043582320213318f;
  t9_n119: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t9_n125;
  t9_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n124;
  t9_n121: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t9_n123;
  t9_n122: return 0.02797657996416092f;
  t9_n123: return 0.08374501764774323f;
  t9_n124: return 0.08299500495195389f;
  t9_n125: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t9_n133;
  t9_n126: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n132;
  t9_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t9_n129;
  t9_n128: return 0.044679801911115646f;
  t9_n129: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t9_n131;
  t9_n130: return 0.08984068781137466f;
  t9_n131: return -0.10945164412260056f;
  t9_n132: return 0.10749109834432602f;
  t9_n133: return 0.11450736224651337f;
  t9_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t9_n196;
  t9_n135: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n189;
  t9_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t9_n138;
  t9_n137: return -0.10968128591775894f;
  t9_n138: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t9_n162;
  t9_n139: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n153;
  t9_n140: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t9_n150;
  t9_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t9_n143;
  t9_n142: return -0.0617765448987484f;
  t9_n143: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t9_n147;
  t9_n144: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t9_n146;
  t9_n145: return 0.02266213856637478f;
  t9_n146: return -0.04334687441587448f;
  t9_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t9_n149;
  t9_n148: return 0.061460502445697784f;
  t9_n149: return 0.1272536963224411f;
  t9_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t9_n152;
  t9_n151: return -0.11367207765579224f;
  t9_n152: return 0.06890672445297241f;
  t9_n153: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t9_n161;
  t9_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t9_n160;
  t9_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t9_n159;
  t9_n156: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t9_n158;
  t9_n157: return 0.05288698151707649f;
  t9_n158: return 0.09719330072402954f;
  t9_n159: return 0.08831929415464401f;
  t9_n160: return 0.0038880787324160337f;
  t9_n161: return -0.11208412051200867f;
  t9_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t9_n180;
  t9_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t9_n179;
  t9_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t9_n178;
  t9_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t9_n171;
  t9_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t9_n168;
  t9_n167: return -0.050740569829940796f;
  t9_n168: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t9_n170;
  t9_n169: return 0.03452669829130173f;
  t9_n170: return 0.06578367203474045f;
  t9_n171: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t9_n175;
  t9_n172: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n174;
  t9_n173: return 0.04058697447180748f;
  t9_n174: return 0.08950041234493256f;
  t9_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t9_n177;
  t9_n176: return 0.0807945653796196f;
  t9_n177: return 0.1294819861650467f;
  t9_n178: return -0.09901988506317139f;
  t9_n179: return 0.006869764998555183f;
  t9_n180: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t9_n188;
  t9_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t9_n187;
  t9_n182: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n184;
  t9_n183: return 0.0626433864235878f;
  t9_n184: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t9_n186;
  t9_n185: return 0.07907576113939285f;
  t9_n186: return 0.10540197044610977f;
  t9_n187: return 0.10206498205661774f;
  t9_n188: return -0.10823880881071091f;
  t9_n189: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t9_n193;
  t9_n190: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n192;
  t9_n191: return 0.04014963656663895f;
  t9_n192: return 0.09552031010389328f;
  t9_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n195;
  t9_n194: return 0.091588094830513f;
  t9_n195: return 0.10231655836105347f;
  t9_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t9_n202;
  t9_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t9_n199;
  t9_n198: return -0.11162908375263214f;
  t9_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t9_n201;
  t9_n200: return 0.01428006961941719f;
  t9_n201: return 0.10641128569841385f;
  t9_n202: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t9_n224;
  t9_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t9_n215;
  t9_n204: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n206;
  t9_n205: return 0.00621330551803112f;
  t9_n206: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n214;
  t9_n207: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t9_n209;
  t9_n208: return 0.06925924122333527f;
  t9_n209: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t9_n211;
  t9_n210: return -0.1357751488685608f;
  t9_n211: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t9_n213;
  t9_n212: return 0.014852204360067844f;
  t9_n213: return 0.08203447610139847f;
  t9_n214: return 0.1050766259431839f;
  t9_n215: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t9_n221;
  t9_n216: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t9_n220;
  t9_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t9_n219;
  t9_n218: return 0.036798905581235886f;
  t9_n219: return 0.0685330182313919f;
  t9_n220: return 0.11675655096769333f;
  t9_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t9_n223;
  t9_n222: return 0.09466202557086945f;
  t9_n223: return 0.10958206653594971f;
  t9_n224: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t9_n230;
  t9_n225: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t9_n229;
  t9_n226: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t9_n228;
  t9_n227: return 0.09952612221240997f;
  t9_n228: return 0.07675500959157944f;
  t9_n229: return -0.08924910426139832f;
  t9_n230: return 0.1060781478881836f;
  t9_n231: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t9_n239;
  t9_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t9_n238;
  t9_n233: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t9_n237;
  t9_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t9_n236;
  t9_n235: return 0.08844324946403503f;
  t9_n236: return 0.11130587011575699f;
  t9_n237: return 0.026080220937728882f;
  t9_n238: return 0.11386457085609436f;
  t9_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t9_n243;
  t9_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t9_n242;
  t9_n241: return 0.11019829660654068f;
  t9_n242: return 0.05768797919154167f;
  t9_n243: return 0.11392705142498016f;
  t9_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t9_n246;
  t9_n245: return 0.009753049351274967f;
  t9_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t9_n254;
  t9_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t9_n253;
  t9_n248: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t9_n250;
  t9_n249: return 0.10888456553220749f;
  t9_n250: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t9_n252;
  t9_n251: return 0.11190256476402283f;
  t9_n252: return 0.11065909266471863f;
  t9_n253: return -0.07946090400218964f;
  t9_n254: return 0.11430513858795166f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree10(const Example& E) {
  t10_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t10_n244;
  t10_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t10_n231;
  t10_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t10_n92;
  t10_n3: if (E.getNumReferences() >= 3329785856 /*15904.0*/) goto t10_n65;
  t10_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t10_n40;
  t10_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t10_n31;
  t10_n6: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t10_n16;
  t10_n7: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t10_n9;
  t10_n8: return -0.02648966573178768f;
  t10_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n15;
  t10_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t10_n14;
  t10_n11: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t10_n13;
  t10_n12: return 0.06134309247136116f;
  t10_n13: return 0.003277688752859831f;
  t10_n14: return -0.019568990916013718f;
  t10_n15: return 0.10875918716192245f;
  t10_n16: if (E.getNumReferences() >= 3282321408 /*328.5*/) goto t10_n24;
  t10_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n21;
  t10_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t10_n20;
  t10_n19: return -0.10787810385227203f;
  t10_n20: return -0.06978398561477661f;
  t10_n21: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t10_n23;
  t10_n22: return -0.03514175862073898f;
  t10_n23: return 0.04454370215535164f;
  t10_n24: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t10_n30;
  t10_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t10_n27;
  t10_n26: return -0.026969511061906815f;
  t10_n27: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t10_n29;
  t10_n28: return 0.07796205580234528f;
  t10_n29: return 0.025505777448415756f;
  t10_n30: return -0.044934190809726715f;
  t10_n31: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t10_n39;
  t10_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t10_n34;
  t10_n33: return -0.04985201731324196f;
  t10_n34: if (E.getNumReferences() >= 3279519744 /*249.5*/) goto t10_n36;
  t10_n35: return 7.695138629060239e-05f;
  t10_n36: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t10_n38;
  t10_n37: return 0.11505196243524551f;
  t10_n38: return 0.06704962253570557f;
  t10_n39: return 0.026118911802768707f;
  t10_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t10_n58;
  t10_n41: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t10_n51;
  t10_n42: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t10_n46;
  t10_n43: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n45;
  t10_n44: return -0.012644424103200436f;
  t10_n45: return 0.06476689875125885f;
  t10_n46: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n50;
  t10_n47: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t10_n49;
  t10_n48: return 0.06469162553548813f;
  t10_n49: return 0.03029419109225273f;
  t10_n50: return 0.10255438089370728f;
  t10_n51: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t10_n53;
  t10_n52: return 0.09568463265895844f;
  t10_n53: if (E.getNumReferences() >= 3273424896 /*156.5*/) goto t10_n57;
  t10_n54: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t10_n56;
  t10_n55: return 0.07436498254537582f;
  t10_n56: return -0.018085381016135216f;
  t10_n57: return 0.08927732706069946f;
  t10_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t10_n62;
  t10_n59: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t10_n61;
  t10_n60: return 0.06403456628322601f;
  t10_n61: return 0.10893741250038147f;
  t10_n62: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t10_n64;
  t10_n63: return 0.07833915203809738f;
  t10_n64: return 0.10779207944869995f;
  t10_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t10_n83;
  t10_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n82;
  t10_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t10_n71;
  t10_n68: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t10_n70;
  t10_n69: return -0.0027607118245214224f;
  t10_n70: return 0.06615544855594635f;
  t10_n71: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t10_n79;
  t10_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t10_n74;
  t10_n73: return 0.07193787395954132f;
  t10_n74: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t10_n76;
  t10_n75: return 0.00977353099733591f;
  t10_n76: if (E.getNumReferences() >= 3359911200 /*200884.5*/) goto t10_n78;
  t10_n77: return 0.1157560646533966f;
  t10_n78: return 0.06915208697319031f;
  t10_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t10_n81;
  t10_n80: return 0.08591532707214355f;
  t10_n81: return 0.11120492219924927f;
  t10_n82: return 0.10261329263448715f;
  t10_n83: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t10_n85;
  t10_n84: return 0.07893253117799759f;
  t10_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t10_n87;
  t10_n86: return 0.10017193853855133f;
  t10_n87: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t10_n89;
  t10_n88: return 0.10490443557500839f;
  t10_n89: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t10_n91;
  t10_n90: return 0.12820766866207123f;
  t10_n91: return -0.027602175250649452f;
  t10_n92: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t10_n144;
  t10_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t10_n129;
  t10_n94: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t10_n114;
  t10_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t10_n109;
  t10_n96: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t10_n98;
  t10_n97: return 0.06929711252450943f;
  t10_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t10_n108;
  t10_n99: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t10_n107;
  t10_n100: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t10_n102;
  t10_n101: return -0.001991261262446642f;
  t10_n102: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t10_n106;
  t10_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t10_n105;
  t10_n104: return 0.08530456572771072f;
  t10_n105: return 0.1288270801305771f;
  t10_n106: return 0.060881201177835464f;
  t10_n107: return 0.09891026467084885f;
  t10_n108: return 0.09750732034444809f;
  t10_n109: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t10_n111;
  t10_n110: return 0.09277079254388809f;
  t10_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t10_n113;
  t10_n112: return 0.07423285394906998f;
  t10_n113: return 0.1063150018453598f;
  t10_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t10_n126;
  t10_n115: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t10_n125;
  t10_n116: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t10_n124;
  t10_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t10_n119;
  t10_n118: return 0.0955369621515274f;
  t10_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t10_n123;
  t10_n120: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t10_n122;
  t10_n121: return 0.0949033722281456f;
  t10_n122: return 0.03150058165192604f;
  t10_n123: return 0.09929127246141434f;
  t10_n124: return 0.10364969819784164f;
  t10_n125: return 0.1052030399441719f;
  t10_n126: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t10_n128;
  t10_n127: return 0.10893315076828003f;
  t10_n128: return 0.09716806560754776f;
  t10_n129: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t10_n135;
  t10_n130: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n134;
  t10_n131: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t10_n133;
  t10_n132: return 0.025403501465916634f;
  t10_n133: return 0.07872946560382843f;
  t10_n134: return 0.07977598905563354f;
  t10_n135: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t10_n143;
  t10_n136: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n142;
  t10_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t10_n139;
  t10_n138: return 0.043038249015808105f;
  t10_n139: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t10_n141;
  t10_n140: return 0.08673278242349625f;
  t10_n141: return -0.10813603550195694f;
  t10_n142: return 0.10516422986984253f;
  t10_n143: return 0.11302406340837479f;
  t10_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t10_n204;
  t10_n145: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n195;
  t10_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t10_n148;
  t10_n147: return -0.10772937536239624f;
  t10_n148: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t10_n172;
  t10_n149: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t10_n163;
  t10_n150: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t10_n160;
  t10_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t10_n153;
  t10_n152: return -0.057586438953876495f;
  t10_n153: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t10_n157;
  t10_n154: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t10_n156;
  t10_n155: return 0.021838398650288582f;
  t10_n156: return -0.0397360622882843f;
  t10_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t10_n159;
  t10_n158: return 0.06033444404602051f;
  t10_n159: return 0.12209545075893402f;
  t10_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t10_n162;
  t10_n161: return -0.1277645379304886f;
  t10_n162: return 0.0269073024392128f;
  t10_n163: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t10_n171;
  t10_n164: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t10_n170;
  t10_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t10_n169;
  t10_n166: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t10_n168;
  t10_n167: return 0.04757379740476608f;
  t10_n168: return 0.09363672137260437f;
  t10_n169: return 0.08357780426740646f;
  t10_n170: return 0.002292738063260913f;
  t10_n171: return -0.11023149639368057f;
  t10_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t10_n188;
  t10_n173: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t10_n187;
  t10_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t10_n186;
  t10_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t10_n181;
  t10_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t10_n178;
  t10_n177: return -0.05487443134188652f;
  t10_n178: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t10_n180;
  t10_n179: return 0.03098118305206299f;
  t10_n180: return 0.0613538883626461f;
  t10_n181: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t10_n183;
  t10_n182: return 0.0641285628080368f;
  t10_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t10_n185;
  t10_n184: return 0.0945146456360817f;
  t10_n185: return 0.1422143131494522f;
  t10_n186: return -0.09546016901731491f;
  t10_n187: return 0.006409863941371441f;
  t10_n188: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t10_n194;
  t10_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t10_n193;
  t10_n190: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t10_n192;
  t10_n191: return 0.05794383957982063f;
  t10_n192: return 0.08050299435853958f;
  t10_n193: return 0.09778542816638947f;
  t10_n194: return -0.10550480335950851f;
  t10_n195: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t10_n201;
  t10_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t10_n198;
  t10_n197: return 0.029872991144657135f;
  t10_n198: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t10_n200;
  t10_n199: return 0.010023502632975578f;
  t10_n200: return 0.09154534339904785f;
  t10_n201: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t10_n203;
  t10_n202: return 0.07931246608495712f;
  t10_n203: return 0.10385134071111679f;
  t10_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t10_n210;
  t10_n205: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t10_n207;
  t10_n206: return -0.10566776245832443f;
  t10_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t10_n209;
  t10_n208: return 0.014873840846121311f;
  t10_n209: return 0.10341809689998627f;
  t10_n210: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t10_n224;
  t10_n211: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t10_n215;
  t10_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t10_n214;
  t10_n213: return 0.0321175679564476f;
  t10_n214: return 0.07178671658039093f;
  t10_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t10_n221;
  t10_n216: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t10_n218;
  t10_n217: return 0.028087567538022995f;
  t10_n218: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n220;
  t10_n219: return 0.06044412776827812f;
  t10_n220: return 0.09975612908601761f;
  t10_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t10_n223;
  t10_n222: return 0.09039555490016937f;
  t10_n223: return 0.10682743787765503f;
  t10_n224: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t10_n230;
  t10_n225: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t10_n229;
  t10_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t10_n228;
  t10_n227: return 0.09221696108579636f;
  t10_n228: return 0.11143869906663895f;
  t10_n229: return -0.08478569984436035f;
  t10_n230: return 0.10340819507837296f;
  t10_n231: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t10_n239;
  t10_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t10_n238;
  t10_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t10_n237;
  t10_n234: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t10_n236;
  t10_n235: return 0.08353859186172485f;
  t10_n236: return -0.0344865620136261f;
  t10_n237: return 0.10445360839366913f;
  t10_n238: return 0.11200720816850662f;
  t10_n239: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t10_n243;
  t10_n240: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t10_n242;
  t10_n241: return 0.07126577198505402f;
  t10_n242: return 0.10863076150417328f;
  t10_n243: return 0.11212217062711716f;
  t10_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t10_n246;
  t10_n245: return 0.008864154107868671f;
  t10_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t10_n254;
  t10_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t10_n253;
  t10_n248: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t10_n250;
  t10_n249: return 0.10645624995231628f;
  t10_n250: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t10_n252;
  t10_n251: return 0.10999232530593872f;
  t10_n252: return 0.10868818312883377f;
  t10_n253: return -0.0717880055308342f;
  t10_n254: return 0.11272518336772919f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree11(const Example& E) {
  t11_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t11_n248;
  t11_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t11_n235;
  t11_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t11_n94;
  t11_n3: if (E.getNumReferences() >= 3329785856 /*15904.0*/) goto t11_n69;
  t11_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t11_n46;
  t11_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t11_n31;
  t11_n6: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t11_n16;
  t11_n7: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t11_n9;
  t11_n8: return -0.02368917688727379f;
  t11_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n15;
  t11_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t11_n14;
  t11_n11: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t11_n13;
  t11_n12: return 0.034899380058050156f;
  t11_n13: return 0.07802258431911469f;
  t11_n14: return -0.01683119311928749f;
  t11_n15: return 0.10161592811346054f;
  t11_n16: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t11_n22;
  t11_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n21;
  t11_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t11_n20;
  t11_n19: return -0.1055774837732315f;
  t11_n20: return -0.06589969992637634f;
  t11_n21: return -0.019587771967053413f;
  t11_n22: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t11_n28;
  t11_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t11_n25;
  t11_n24: return -0.02444872446358204f;
  t11_n25: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t11_n27;
  t11_n26: return 0.07097446918487549f;
  t11_n27: return 0.02139517106115818f;
  t11_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n30;
  t11_n29: return -0.049424003809690475f;
  t11_n30: return 0.05827069655060768f;
  t11_n31: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t11_n45;
  t11_n32: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t11_n42;
  t11_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t11_n35;
  t11_n34: return -0.04614725708961487f;
  t11_n35: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t11_n41;
  t11_n36: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t11_n38;
  t11_n37: return 0.05528421327471733f;
  t11_n38: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t11_n40;
  t11_n39: return 0.009403818286955357f;
  t11_n40: return -0.05886528640985489f;
  t11_n41: return -0.11724168807268143f;
  t11_n42: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t11_n44;
  t11_n43: return 0.09655074775218964f;
  t11_n44: return 0.035531528294086456f;
  t11_n45: return 0.023836692795157433f;
  t11_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t11_n62;
  t11_n47: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t11_n57;
  t11_n48: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t11_n52;
  t11_n49: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n51;
  t11_n50: return -0.009774947538971901f;
  t11_n51: return 0.05862367898225784f;
  t11_n52: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t11_n54;
  t11_n53: return 0.0879574567079544f;
  t11_n54: if (E.getNumReferences() >= 3273424896 /*156.5*/) goto t11_n56;
  t11_n55: return 0.00437132827937603f;
  t11_n56: return 0.07622218132019043f;
  t11_n57: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t11_n61;
  t11_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n60;
  t11_n59: return 0.05524801090359688f;
  t11_n60: return 0.10013271123170853f;
  t11_n61: return 0.10294249653816223f;
  t11_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t11_n66;
  t11_n63: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t11_n65;
  t11_n64: return 0.060833532363176346f;
  t11_n65: return 0.10428833961486816f;
  t11_n66: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t11_n68;
  t11_n67: return 0.07064104080200195f;
  t11_n68: return 0.10497681051492691f;
  t11_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t11_n85;
  t11_n70: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n84;
  t11_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t11_n73;
  t11_n72: return 0.004285367205739021f;
  t11_n73: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t11_n81;
  t11_n74: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t11_n76;
  t11_n75: return 0.06755990535020828f;
  t11_n76: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t11_n78;
  t11_n77: return 0.010863068513572216f;
  t11_n78: if (E.getNumReferences() >= 3359911200 /*200884.5*/) goto t11_n80;
  t11_n79: return 0.11303015053272247f;
  t11_n80: return 0.06442137062549591f;
  t11_n81: if (E.getNumReferences() >= 3330296064 /*16420.5*/) goto t11_n83;
  t11_n82: return 0.142336905002594f;
  t11_n83: return 0.09601662307977676f;
  t11_n84: return 0.09947749227285385f;
  t11_n85: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t11_n87;
  t11_n86: return 0.07426085323095322f;
  t11_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t11_n89;
  t11_n88: return 0.09765138477087021f;
  t11_n89: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t11_n91;
  t11_n90: return 0.10270489752292633f;
  t11_n91: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t11_n93;
  t11_n92: return 0.11512576043605804f;
  t11_n93: return 0.13655909895896912f;
  t11_n94: if (E.getFileProximityDistanceCost() >= 3256745984 /*39.5*/) goto t11_n134;
  t11_n95: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t11_n123;
  t11_n96: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t11_n108;
  t11_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n105;
  t11_n98: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t11_n100;
  t11_n99: return 0.06456999480724335f;
  t11_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t11_n104;
  t11_n101: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t11_n103;
  t11_n102: return 0.03649178892374039f;
  t11_n103: return 0.09596733748912811f;
  t11_n104: return 0.0952289029955864f;
  t11_n105: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t11_n107;
  t11_n106: return 0.08887141197919846f;
  t11_n107: return 0.10093063116073608f;
  t11_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n120;
  t11_n109: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t11_n119;
  t11_n110: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t11_n118;
  t11_n111: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t11_n113;
  t11_n112: return 0.09198067337274551f;
  t11_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t11_n117;
  t11_n114: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t11_n116;
  t11_n115: return 0.035309065133333206f;
  t11_n116: return 0.09741976112127304f;
  t11_n117: return 0.09659988433122635f;
  t11_n118: return 0.1013430655002594f;
  t11_n119: return 0.102765753865242f;
  t11_n120: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t11_n122;
  t11_n121: return 0.10685274004936218f;
  t11_n122: return 0.09358696639537811f;
  t11_n123: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t11_n129;
  t11_n124: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n128;
  t11_n125: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t11_n127;
  t11_n126: return 0.022222939878702164f;
  t11_n127: return 0.07425316423177719f;
  t11_n128: return 0.07514062523841858f;
  t11_n129: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t11_n133;
  t11_n130: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n132;
  t11_n131: return 0.06979169696569443f;
  t11_n132: return 0.1023588627576828f;
  t11_n133: return 0.11179395020008087f;
  t11_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t11_n200;
  t11_n135: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n189;
  t11_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t11_n138;
  t11_n137: return -0.10596225410699844f;
  t11_n138: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t11_n168;
  t11_n139: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t11_n161;
  t11_n140: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t11_n158;
  t11_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n145;
  t11_n142: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t11_n144;
  t11_n143: return -0.062219876796007156f;
  t11_n144: return 0.06333066523075104f;
  t11_n145: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t11_n155;
  t11_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t11_n148;
  t11_n147: return 0.0005079416441731155f;
  t11_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n152;
  t11_n149: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t11_n151;
  t11_n150: return -0.009939918294548988f;
  t11_n151: return 0.14686870574951172f;
  t11_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t11_n154;
  t11_n153: return 0.11052460968494415f;
  t11_n154: return -0.006499826908111572f;
  t11_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t11_n157;
  t11_n156: return 0.06370719522237778f;
  t11_n157: return 0.11806726455688477f;
  t11_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n160;
  t11_n159: return -0.12119340151548386f;
  t11_n160: return 0.027118127793073654f;
  t11_n161: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t11_n167;
  t11_n162: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t11_n166;
  t11_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n165;
  t11_n164: return 0.05729557201266289f;
  t11_n165: return 0.08837217837572098f;
  t11_n166: return -0.02627704106271267f;
  t11_n167: return -0.10857166349887848f;
  t11_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t11_n184;
  t11_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t11_n183;
  t11_n170: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t11_n182;
  t11_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t11_n177;
  t11_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n174;
  t11_n173: return -0.051800522953271866f;
  t11_n174: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t11_n176;
  t11_n175: return 0.02786341868340969f;
  t11_n176: return 0.05744291841983795f;
  t11_n177: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t11_n179;
  t11_n178: return 0.059250608086586f;
  t11_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t11_n181;
  t11_n180: return 0.08944039046764374f;
  t11_n181: return 0.13422466814517975f;
  t11_n182: return -0.09150108695030212f;
  t11_n183: return 0.005437145009636879f;
  t11_n184: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t11_n188;
  t11_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t11_n187;
  t11_n186: return 0.06641644984483719f;
  t11_n187: return 0.0938851609826088f;
  t11_n188: return -0.10257112234830856f;
  t11_n189: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t11_n197;
  t11_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t11_n194;
  t11_n191: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t11_n193;
  t11_n192: return 0.04741901159286499f;
  t11_n193: return -0.0735492929816246f;
  t11_n194: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t11_n196;
  t11_n195: return 0.007975660264492035f;
  t11_n196: return 0.08747051656246185f;
  t11_n197: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t11_n199;
  t11_n198: return 0.07277096062898636f;
  t11_n199: return 0.1001618504524231f;
  t11_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n204;
  t11_n201: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t11_n203;
  t11_n202: return -0.1293577402830124f;
  t11_n203: return 0.014565762132406235f;
  t11_n204: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t11_n228;
  t11_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t11_n219;
  t11_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t11_n210;
  t11_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t11_n209;
  t11_n208: return -0.05082540214061737f;
  t11_n209: return 0.03311498835682869f;
  t11_n210: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n218;
  t11_n211: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t11_n213;
  t11_n212: return 0.06078295037150383f;
  t11_n213: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t11_n215;
  t11_n214: return -0.1234467402100563f;
  t11_n215: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t11_n217;
  t11_n216: return 0.005823640618473291f;
  t11_n217: return 0.07560564577579498f;
  t11_n218: return 0.10004030168056488f;
  t11_n219: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t11_n225;
  t11_n220: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t11_n224;
  t11_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t11_n223;
  t11_n222: return 0.03659534081816673f;
  t11_n223: return 0.06718665361404419f;
  t11_n224: return 0.11222179234027863f;
  t11_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t11_n227;
  t11_n226: return 0.08704851567745209f;
  t11_n227: return 0.10489346086978912f;
  t11_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t11_n234;
  t11_n229: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t11_n233;
  t11_n230: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t11_n232;
  t11_n231: return 0.09286525100469589f;
  t11_n232: return 0.06717807799577713f;
  t11_n233: return -0.07979966700077057f;
  t11_n234: return 0.10135096311569214f;
  t11_n235: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t11_n243;
  t11_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t11_n242;
  t11_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t11_n241;
  t11_n238: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t11_n240;
  t11_n239: return 0.07869399338960648f;
  t11_n240: return -0.03269942104816437f;
  t11_n241: return 0.1014917865395546f;
  t11_n242: return 0.11039459705352783f;
  t11_n243: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t11_n247;
  t11_n244: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t11_n246;
  t11_n245: return 0.10572925955057144f;
  t11_n246: return 0.043018776923418045f;
  t11_n247: return 0.11072034388780594f;
  t11_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t11_n250;
  t11_n249: return 0.008406274020671844f;
  t11_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t11_n254;
  t11_n251: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t11_n253;
  t11_n252: return 0.10650491714477539f;
  t11_n253: return -0.06472498923540115f;
  t11_n254: return 0.11138951778411865f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree12(const Example& E) {
  t12_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t12_n246;
  t12_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t12_n233;
  t12_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t12_n94;
  t12_n3: if (E.getNumReferences() >= 3329785856 /*15904.0*/) goto t12_n71;
  t12_n4: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n48;
  t12_n5: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t12_n27;
  t12_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t12_n24;
  t12_n7: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t12_n21;
  t12_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t12_n14;
  t12_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n11;
  t12_n10: return -0.05559629946947098f;
  t12_n11: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t12_n13;
  t12_n12: return 0.07537300884723663f;
  t12_n13: return 0.0015853744698688388f;
  t12_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n16;
  t12_n15: return -0.011718052439391613f;
  t12_n16: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t12_n20;
  t12_n17: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t12_n19;
  t12_n18: return 0.07121307402849197f;
  t12_n19: return 0.021258560940623283f;
  t12_n20: return 0.005377792753279209f;
  t12_n21: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t12_n23;
  t12_n22: return 0.02675715647637844f;
  t12_n23: return 0.07664120942354202f;
  t12_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t12_n26;
  t12_n25: return 0.07987955212593079f;
  t12_n26: return 0.11176452785730362f;
  t12_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t12_n35;
  t12_n28: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t12_n30;
  t12_n29: return -0.07848601043224335f;
  t12_n30: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t12_n34;
  t12_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n33;
  t12_n32: return -0.059782493859529495f;
  t12_n33: return 0.042156048119068146f;
  t12_n34: return 0.06217409297823906f;
  t12_n35: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t12_n47;
  t12_n36: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t12_n46;
  t12_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t12_n41;
  t12_n38: if (E.getNumReferences() >= 3278274560 /*230.5*/) goto t12_n40;
  t12_n39: return -0.03546789288520813f;
  t12_n40: return 0.051749009639024734f;
  t12_n41: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t12_n43;
  t12_n42: return 0.05278395116329193f;
  t12_n43: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t12_n45;
  t12_n44: return 0.00518062524497509f;
  t12_n45: return -0.06268112361431122f;
  t12_n46: return -0.11319289356470108f;
  t12_n47: return 0.019565405324101448f;
  t12_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t12_n64;
  t12_n49: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t12_n59;
  t12_n50: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t12_n54;
  t12_n51: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t12_n53;
  t12_n52: return 0.027992937713861465f;
  t12_n53: return -0.02019602246582508f;
  t12_n54: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t12_n56;
  t12_n55: return 0.08644679933786392f;
  t12_n56: if (E.getNumReferences() >= 3276701696 /*206.5*/) goto t12_n58;
  t12_n57: return 0.01732068881392479f;
  t12_n58: return 0.07431712001562119f;
  t12_n59: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t12_n63;
  t12_n60: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n62;
  t12_n61: return 0.05281434580683708f;
  t12_n62: return 0.09519091993570328f;
  t12_n63: return 0.1008511558175087f;
  t12_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t12_n68;
  t12_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t12_n67;
  t12_n66: return 0.059131499379873276f;
  t12_n67: return 0.09990186989307404f;
  t12_n68: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t12_n70;
  t12_n69: return 0.07008624821901321f;
  t12_n70: return 0.10248027741909027f;
  t12_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t12_n83;
  t12_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n82;
  t12_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t12_n77;
  t12_n74: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n76;
  t12_n75: return 0.002299513667821884f;
  t12_n76: return 0.09741201996803284f;
  t12_n77: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t12_n79;
  t12_n78: return 0.050682444125413895f;
  t12_n79: if (E.getNumReferences() >= 3330296064 /*16420.5*/) goto t12_n81;
  t12_n80: return 0.1352178454399109f;
  t12_n81: return 0.09233834594488144f;
  t12_n82: return 0.0967678353190422f;
  t12_n83: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n89;
  t12_n84: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t12_n86;
  t12_n85: return 0.0667901411652565f;
  t12_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t12_n88;
  t12_n87: return 0.08878261595964432f;
  t12_n88: return 0.10673645883798599f;
  t12_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t12_n91;
  t12_n90: return 0.11338748037815094f;
  t12_n91: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t12_n93;
  t12_n92: return 0.10494104772806168f;
  t12_n93: return 0.13436424732208252f;
  t12_n94: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t12_n150;
  t12_n95: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t12_n135;
  t12_n96: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t12_n120;
  t12_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n115;
  t12_n98: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t12_n104;
  t12_n99: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t12_n101;
  t12_n100: return 0.047153957188129425f;
  t12_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t12_n103;
  t12_n102: return 0.06536468118429184f;
  t12_n103: return 0.09178458154201508f;
  t12_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t12_n114;
  t12_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n113;
  t12_n106: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t12_n108;
  t12_n107: return -0.01133006066083908f;
  t12_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t12_n110;
  t12_n109: return 0.02885100431740284f;
  t12_n110: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t12_n112;
  t12_n111: return 0.1117219477891922f;
  t12_n112: return 0.08398330956697464f;
  t12_n113: return 0.09314404428005219f;
  t12_n114: return 0.09219783544540405f;
  t12_n115: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t12_n117;
  t12_n116: return 0.08485722541809082f;
  t12_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t12_n119;
  t12_n118: return 0.061167389154434204f;
  t12_n119: return 0.10149285942316055f;
  t12_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n132;
  t12_n121: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t12_n131;
  t12_n122: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t12_n128;
  t12_n123: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t12_n127;
  t12_n124: if (E.getNumReferences() >= 3318892544 /*6729.0*/) goto t12_n126;
  t12_n125: return 0.08826514333486557f;
  t12_n126: return -0.015111394226551056f;
  t12_n127: return 0.0839146301150322f;
  t12_n128: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t12_n130;
  t12_n129: return 0.08492014557123184f;
  t12_n130: return 0.09856394678354263f;
  t12_n131: return 0.10048643499612808f;
  t12_n132: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t12_n134;
  t12_n133: return 0.10489821434020996f;
  t12_n134: return 0.09038610011339188f;
  t12_n135: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t12_n145;
  t12_n136: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n144;
  t12_n137: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t12_n143;
  t12_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n142;
  t12_n139: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t12_n141;
  t12_n140: return 0.062020882964134216f;
  t12_n141: return -0.009826294146478176f;
  t12_n142: return -0.04071251302957535f;
  t12_n143: return 0.06935008615255356f;
  t12_n144: return 0.07103943824768066f;
  t12_n145: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t12_n149;
  t12_n146: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n148;
  t12_n147: return 0.06559499353170395f;
  t12_n148: return 0.10013443231582642f;
  t12_n149: return 0.11082863062620163f;
  t12_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t12_n200;
  t12_n151: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n193;
  t12_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t12_n154;
  t12_n153: return -0.1042642593383789f;
  t12_n154: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t12_n176;
  t12_n155: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n169;
  t12_n156: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t12_n166;
  t12_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n159;
  t12_n158: return -0.04899092763662338f;
  t12_n159: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t12_n163;
  t12_n160: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t12_n162;
  t12_n161: return 0.01907278783619404f;
  t12_n162: return -0.03642919659614563f;
  t12_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t12_n165;
  t12_n164: return 0.06436183303594589f;
  t12_n165: return 0.11479420959949493f;
  t12_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n168;
  t12_n167: return -0.11545311659574509f;
  t12_n168: return 0.03223631531000137f;
  t12_n169: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t12_n175;
  t12_n170: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t12_n174;
  t12_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n173;
  t12_n172: return 0.05253491923213005f;
  t12_n173: return 0.08378932625055313f;
  t12_n174: return -0.026445893570780754f;
  t12_n175: return -0.10686219483613968f;
  t12_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t12_n188;
  t12_n177: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t12_n187;
  t12_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t12_n186;
  t12_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n185;
  t12_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n184;
  t12_n181: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t12_n183;
  t12_n182: return 0.026890954002738f;
  t12_n183: return -0.06177801266312599f;
  t12_n184: return 0.0491022951900959f;
  t12_n185: return 0.06423056870698929f;
  t12_n186: return -0.09437628090381622f;
  t12_n187: return -0.0020628604106605053f;
  t12_n188: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t12_n192;
  t12_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t12_n191;
  t12_n190: return 0.06207434460520744f;
  t12_n191: return 0.09040419012308121f;
  t12_n192: return -0.10231799632310867f;
  t12_n193: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n199;
  t12_n194: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n198;
  t12_n195: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t12_n197;
  t12_n196: return 0.028677072376012802f;
  t12_n197: return 0.06302237510681152f;
  t12_n198: return 0.0872410386800766f;
  t12_n199: return 0.09671885520219803f;
  t12_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n204;
  t12_n201: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t12_n203;
  t12_n202: return -0.1258220672607422f;
  t12_n203: return 0.014546224847435951f;
  t12_n204: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t12_n226;
  t12_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t12_n219;
  t12_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n210;
  t12_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t12_n209;
  t12_n208: return -0.045668650418519974f;
  t12_n209: return 0.031128516420722008f;
  t12_n210: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n218;
  t12_n211: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t12_n213;
  t12_n212: return 0.05558444932103157f;
  t12_n213: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t12_n215;
  t12_n214: return -0.11540687084197998f;
  t12_n215: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t12_n217;
  t12_n216: return -0.010843856260180473f;
  t12_n217: return 0.06576400995254517f;
  t12_n218: return 0.09698328375816345f;
  t12_n219: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t12_n225;
  t12_n220: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t12_n224;
  t12_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t12_n223;
  t12_n222: return 0.026429543271660805f;
  t12_n223: return 0.05687207356095314f;
  t12_n224: return 0.10937661677598953f;
  t12_n225: return 0.09075017273426056f;
  t12_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t12_n232;
  t12_n227: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t12_n231;
  t12_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t12_n230;
  t12_n229: return 0.08541548997163773f;
  t12_n230: return 0.10904605686664581f;
  t12_n231: return -0.07820446789264679f;
  t12_n232: return 0.09905092418193817f;
  t12_n233: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t12_n241;
  t12_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t12_n240;
  t12_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t12_n239;
  t12_n236: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t12_n238;
  t12_n237: return 0.07429125159978867f;
  t12_n238: return -0.02978007309138775f;
  t12_n239: return 0.09862742573022842f;
  t12_n240: return 0.10901837795972824f;
  t12_n241: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t12_n245;
  t12_n242: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t12_n244;
  t12_n243: return 0.1037362664937973f;
  t12_n244: return 0.03973373398184776f;
  t12_n245: return 0.10949138551950455f;
  t12_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t12_n248;
  t12_n247: return 0.007632675115019083f;
  t12_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t12_n254;
  t12_n249: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t12_n251;
  t12_n250: return 0.10198338329792023f;
  t12_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t12_n253;
  t12_n252: return 0.10679315030574799f;
  t12_n253: return 0.10522998124361038f;
  t12_n254: return 0.11027605086565018f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree13(const Example& E) {
  t13_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t13_n250;
  t13_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t13_n239;
  t13_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t13_n106;
  t13_n3: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t13_n59;
  t13_n4: if (E.getNumReferences() >= 3328224256 /*14379.0*/) goto t13_n40;
  t13_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t13_n11;
  t13_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n8;
  t13_n7: return -0.09262233972549438f;
  t13_n8: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t13_n10;
  t13_n9: return -0.013731924816966057f;
  t13_n10: return 0.05114782229065895f;
  t13_n11: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t13_n19;
  t13_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t13_n16;
  t13_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n15;
  t13_n14: return 0.0059635229408741f;
  t13_n15: return 0.05250248685479164f;
  t13_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t13_n18;
  t13_n17: return 0.08141354471445084f;
  t13_n18: return 0.11015407741069794f;
  t13_n19: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t13_n37;
  t13_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t13_n32;
  t13_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t13_n27;
  t13_n22: if (E.getNumReferences() >= 3279781888 /*253.5*/) goto t13_n24;
  t13_n23: return -0.06933102011680603f;
  t13_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n26;
  t13_n25: return -0.059248264878988266f;
  t13_n26: return 0.03997389227151871f;
  t13_n27: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t13_n31;
  t13_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t13_n30;
  t13_n29: return 0.030864017084240913f;
  t13_n30: return -0.11218022555112839f;
  t13_n31: return -0.03209961578249931f;
  t13_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t13_n36;
  t13_n33: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t13_n35;
  t13_n34: return 0.05410117283463478f;
  t13_n35: return -0.014553462155163288f;
  t13_n36: return 0.07547227293252945f;
  t13_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t13_n39;
  t13_n38: return -0.01871749758720398f;
  t13_n39: return 0.017038024961948395f;
  t13_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t13_n52;
  t13_n41: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n51;
  t13_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t13_n44;
  t13_n43: return 0.0013722338480874896f;
  t13_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t13_n46;
  t13_n45: return 0.06315737217664719f;
  t13_n46: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t13_n48;
  t13_n47: return 0.011640964075922966f;
  t13_n48: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t13_n50;
  t13_n49: return 0.10239474475383759f;
  t13_n50: return -0.10464171320199966f;
  t13_n51: return 0.09329398721456528f;
  t13_n52: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t13_n56;
  t13_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t13_n55;
  t13_n54: return 0.04100467637181282f;
  t13_n55: return 0.093720942735672f;
  t13_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t13_n58;
  t13_n57: return 0.09071315079927444f;
  t13_n58: return 0.10760438442230225f;
  t13_n59: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t13_n71;
  t13_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t13_n68;
  t13_n61: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t13_n67;
  t13_n62: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t13_n66;
  t13_n63: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n65;
  t13_n64: return 0.046716321259737015f;
  t13_n65: return 0.09881110489368439f;
  t13_n66: return 0.009823686443269253f;
  t13_n67: return 0.07751832902431488f;
  t13_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t13_n70;
  t13_n69: return 0.09719161689281464f;
  t13_n70: return 0.09814955294132233f;
  t13_n71: if (E.getNumReferences() >= 3326663680 /*12855.0*/) goto t13_n93;
  t13_n72: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t13_n86;
  t13_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t13_n81;
  t13_n74: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n80;
  t13_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t13_n77;
  t13_n76: return -0.06395627558231354f;
  t13_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t13_n79;
  t13_n78: return -0.01693691685795784f;
  t13_n79: return 0.048845790326595306f;
  t13_n80: return 0.05281538888812065f;
  t13_n81: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t13_n83;
  t13_n82: return 0.017344964668154716f;
  t13_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t13_n85;
  t13_n84: return 0.03910089656710625f;
  t13_n85: return 0.08933230489492416f;
  t13_n86: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t13_n92;
  t13_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t13_n89;
  t13_n88: return 0.05158054456114769f;
  t13_n89: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t13_n91;
  t13_n90: return 0.033883895725011826f;
  t13_n91: return 0.10468999296426773f;
  t13_n92: return 0.12411945313215256f;
  t13_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t13_n97;
  t13_n94: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t13_n96;
  t13_n95: return 0.05460035800933838f;
  t13_n96: return 0.10375940054655075f;
  t13_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t13_n99;
  t13_n98: return 0.059352029114961624f;
  t13_n99: if (E.getNumReferences() >= 3348479616 /*76669.0*/) goto t13_n103;
  t13_n100: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t13_n102;
  t13_n101: return 0.09809787571430206f;
  t13_n102: return 0.12856654822826385f;
  t13_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t13_n105;
  t13_n104: return 0.10479376465082169f;
  t13_n105: return 0.13152927160263062f;
  t13_n106: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t13_n148;
  t13_n107: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t13_n137;
  t13_n108: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t13_n124;
  t13_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n121;
  t13_n110: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t13_n116;
  t13_n111: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t13_n113;
  t13_n112: return 0.04216604307293892f;
  t13_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t13_n115;
  t13_n114: return 0.06159448251128197f;
  t13_n115: return 0.08863931894302368f;
  t13_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t13_n120;
  t13_n117: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t13_n119;
  t13_n118: return 0.02764124982059002f;
  t13_n119: return 0.09051144868135452f;
  t13_n120: return 0.08921173959970474f;
  t13_n121: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t13_n123;
  t13_n122: return 0.08121202886104584f;
  t13_n123: return 0.09560829401016235f;
  t13_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n134;
  t13_n125: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t13_n133;
  t13_n126: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t13_n130;
  t13_n127: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t13_n129;
  t13_n128: return 0.035610854625701904f;
  t13_n129: return 0.08003649860620499f;
  t13_n130: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t13_n132;
  t13_n131: return 0.080368772149086f;
  t13_n132: return 0.09542851895093918f;
  t13_n133: return 0.0980902910232544f;
  t13_n134: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t13_n136;
  t13_n135: return 0.10326110571622849f;
  t13_n136: return 0.08743086457252502f;
  t13_n137: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t13_n143;
  t13_n138: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n142;
  t13_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t13_n141;
  t13_n140: return 0.04132414609193802f;
  t13_n141: return -0.008135493844747543f;
  t13_n142: return 0.06715595722198486f;
  t13_n143: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t13_n147;
  t13_n144: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n146;
  t13_n145: return 0.06155102327466011f;
  t13_n146: return 0.09840735793113708f;
  t13_n147: return 0.11004095524549484f;
  t13_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t13_n212;
  t13_n149: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n203;
  t13_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t13_n152;
  t13_n151: return -0.10278166085481644f;
  t13_n152: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t13_n178;
  t13_n153: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t13_n173;
  t13_n154: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t13_n170;
  t13_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n157;
  t13_n156: return -0.04392024502158165f;
  t13_n157: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t13_n167;
  t13_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t13_n160;
  t13_n159: return -0.0009354445501230657f;
  t13_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n164;
  t13_n161: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t13_n163;
  t13_n162: return -0.024524129927158356f;
  t13_n163: return 0.13823598623275757f;
  t13_n164: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t13_n166;
  t13_n165: return 0.10239008814096451f;
  t13_n166: return -0.006537392269819975f;
  t13_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t13_n169;
  t13_n168: return 0.0616023950278759f;
  t13_n169: return 0.11208570748567581f;
  t13_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n172;
  t13_n171: return -0.11070620268583298f;
  t13_n172: return 0.03137695789337158f;
  t13_n173: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t13_n177;
  t13_n174: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t13_n176;
  t13_n175: return 0.052884042263031006f;
  t13_n176: return -0.028905075043439865f;
  t13_n177: return -0.10558448731899261f;
  t13_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t13_n198;
  t13_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t13_n197;
  t13_n180: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t13_n196;
  t13_n181: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t13_n195;
  t13_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t13_n188;
  t13_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n187;
  t13_n184: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t13_n186;
  t13_n185: return -0.09186765551567078f;
  t13_n186: return 0.09290814399719238f;
  t13_n187: return 0.023742061108350754f;
  t13_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t13_n190;
  t13_n189: return 0.02316410467028618f;
  t13_n190: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t13_n194;
  t13_n191: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t13_n193;
  t13_n192: return 0.09037838131189346f;
  t13_n193: return -0.21891961991786957f;
  t13_n194: return 0.11824428290128708f;
  t13_n195: return 0.04981320723891258f;
  t13_n196: return -0.09690210223197937f;
  t13_n197: return -0.0047273277305066586f;
  t13_n198: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t13_n202;
  t13_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t13_n201;
  t13_n200: return 0.05816428363323212f;
  t13_n201: return 0.08639510720968246f;
  t13_n202: return -0.10004466027021408f;
  t13_n203: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t13_n211;
  t13_n204: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t13_n210;
  t13_n205: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t13_n207;
  t13_n206: return 0.049999676644802094f;
  t13_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n209;
  t13_n208: return -0.056144893169403076f;
  t13_n209: return 0.07693495601415634f;
  t13_n210: return 0.08353525400161743f;
  t13_n211: return 0.09304194152355194f;
  t13_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t13_n216;
  t13_n213: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t13_n215;
  t13_n214: return -0.12331246584653854f;
  t13_n215: return 0.014662949368357658f;
  t13_n216: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t13_n234;
  t13_n217: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t13_n233;
  t13_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t13_n228;
  t13_n219: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t13_n221;
  t13_n220: return 0.0063010710291564465f;
  t13_n221: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n227;
  t13_n222: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t13_n226;
  t13_n223: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t13_n225;
  t13_n224: return 0.044518955051898956f;
  t13_n225: return -0.021241430193185806f;
  t13_n226: return 0.0587582103908062f;
  t13_n227: return 0.09403622895479202f;
  t13_n228: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t13_n232;
  t13_n229: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t13_n231;
  t13_n230: return 0.03935341536998749f;
  t13_n231: return 0.10666700452566147f;
  t13_n232: return 0.08747990429401398f;
  t13_n233: return -0.137372687458992f;
  t13_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t13_n238;
  t13_n235: if (E.getFractionNameInContext() >= 3175960628 /*0.05012531578540802*/) goto t13_n237;
  t13_n236: return 0.07309982180595398f;
  t13_n237: return 0.09031807631254196f;
  t13_n238: return 0.09669604152441025f;
  t13_n239: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t13_n245;
  t13_n240: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t13_n244;
  t13_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t13_n243;
  t13_n242: return 0.03225594013929367f;
  t13_n243: return 0.09532467275857925f;
  t13_n244: return 0.10765745490789413f;
  t13_n245: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t13_n249;
  t13_n246: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t13_n248;
  t13_n247: return 0.0548374317586422f;
  t13_n248: return 0.10297077149152756f;
  t13_n249: return 0.10833121091127396f;
  t13_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t13_n252;
  t13_n251: return 0.007940222509205341f;
  t13_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t13_n254;
  t13_n253: return 0.1029914915561676f;
  t13_n254: return 0.10931379348039627f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree14(const Example& E) {
  t14_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t14_n248;
  t14_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t14_n237;
  t14_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t14_n98;
  t14_n3: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t14_n43;
  t14_n4: if (E.getNumReferences() >= 3328224256 /*14379.0*/) goto t14_n32;
  t14_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t14_n23;
  t14_n6: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t14_n12;
  t14_n7: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t14_n9;
  t14_n8: return -0.0337882861495018f;
  t14_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n11;
  t14_n10: return 0.026844382286071777f;
  t14_n11: return 0.09536668658256531f;
  t14_n12: if (E.getNumReferences() >= 3301085184 /*1556.5*/) goto t14_n22;
  t14_n13: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t14_n15;
  t14_n14: return -0.07278461754322052f;
  t14_n15: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t14_n17;
  t14_n16: return -0.10537638515233994f;
  t14_n17: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t14_n21;
  t14_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t14_n20;
  t14_n19: return -0.05935904383659363f;
  t14_n20: return 0.03485177457332611f;
  t14_n21: return -0.08464941382408142f;
  t14_n22: return -0.009926775470376015f;
  t14_n23: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t14_n31;
  t14_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t14_n26;
  t14_n25: return -0.022947324439883232f;
  t14_n26: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t14_n28;
  t14_n27: return 0.04634543135762215f;
  t14_n28: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t14_n30;
  t14_n29: return 0.00815068930387497f;
  t14_n30: return -0.043038949370384216f;
  t14_n31: return -0.1104554682970047f;
  t14_n32: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n42;
  t14_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t14_n37;
  t14_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t14_n36;
  t14_n35: return 0.0008149142959155142f;
  t14_n36: return 0.0473465733230114f;
  t14_n37: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t14_n41;
  t14_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t14_n40;
  t14_n39: return 0.037005748599767685f;
  t14_n40: return 0.09031926840543747f;
  t14_n41: return 0.09156546741724014f;
  t14_n42: return 0.09039755910634995f;
  t14_n43: if (E.getNumReferences() >= 3303284736 /*1825.0*/) goto t14_n87;
  t14_n44: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t14_n64;
  t14_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t14_n61;
  t14_n46: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t14_n54;
  t14_n47: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t14_n53;
  t14_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n52;
  t14_n49: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t14_n51;
  t14_n50: return 0.02735799364745617f;
  t14_n51: return 0.058644942939281464f;
  t14_n52: return 0.09457922726869583f;
  t14_n53: return -0.003013188485056162f;
  t14_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t14_n58;
  t14_n55: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t14_n57;
  t14_n56: return 0.060237590223550797f;
  t14_n57: return 0.09105715900659561f;
  t14_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t14_n60;
  t14_n59: return -0.1768762320280075f;
  t14_n60: return 0.10029668360948563f;
  t14_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t14_n63;
  t14_n62: return 0.07979532331228256f;
  t14_n63: return 0.09093733131885529f;
  t14_n64: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t14_n80;
  t14_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t14_n73;
  t14_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n72;
  t14_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t14_n69;
  t14_n68: return -0.07094211876392365f;
  t14_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t14_n71;
  t14_n70: return -0.029025427997112274f;
  t14_n71: return 0.03610840439796448f;
  t14_n72: return 0.04161053150892258f;
  t14_n73: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t14_n75;
  t14_n74: return 0.013189781457185745f;
  t14_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t14_n79;
  t14_n76: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t14_n78;
  t14_n77: return 0.08243356645107269f;
  t14_n78: return 0.008293933235108852f;
  t14_n79: return 0.08574043959379196f;
  t14_n80: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t14_n86;
  t14_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t14_n83;
  t14_n82: return 0.0484749898314476f;
  t14_n83: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t14_n85;
  t14_n84: return 0.03134290874004364f;
  t14_n85: return 0.10526005178689957f;
  t14_n86: return 0.12123949080705643f;
  t14_n87: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t14_n93;
  t14_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t14_n90;
  t14_n89: return 0.056422773748636246f;
  t14_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t14_n92;
  t14_n91: return 0.07199957221746445f;
  t14_n92: return 0.09739232808351517f;
  t14_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t14_n95;
  t14_n94: return 0.0890495553612709f;
  t14_n95: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t14_n97;
  t14_n96: return 0.10035528242588043f;
  t14_n97: return 0.12886305153369904f;
  t14_n98: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t14_n146;
  t14_n99: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t14_n127;
  t14_n100: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t14_n122;
  t14_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n113;
  t14_n102: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t14_n112;
  t14_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t14_n107;
  t14_n104: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t14_n106;
  t14_n105: return 0.05449357628822327f;
  t14_n106: return -0.0015944893239066005f;
  t14_n107: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t14_n111;
  t14_n108: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t14_n110;
  t14_n109: return 0.09067186713218689f;
  t14_n110: return -0.04569403454661369f;
  t14_n111: return 0.0838189572095871f;
  t14_n112: return 0.08294099569320679f;
  t14_n113: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t14_n119;
  t14_n114: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t14_n118;
  t14_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t14_n117;
  t14_n116: return 0.08595910668373108f;
  t14_n117: return 0.1081356331706047f;
  t14_n118: return -0.012317130342125893f;
  t14_n119: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t14_n121;
  t14_n120: return 0.0697946771979332f;
  t14_n121: return 0.08895917236804962f;
  t14_n122: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n126;
  t14_n123: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t14_n125;
  t14_n124: return 0.014386285096406937f;
  t14_n125: return 0.06317558884620667f;
  t14_n126: return 0.06305764615535736f;
  t14_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n145;
  t14_n128: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t14_n144;
  t14_n129: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t14_n139;
  t14_n130: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t14_n136;
  t14_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t14_n135;
  t14_n132: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t14_n134;
  t14_n133: return 0.07399631291627884f;
  t14_n134: return 0.03227456286549568f;
  t14_n135: return 0.09150465577840805f;
  t14_n136: if (E.getNumReferences() >= 3352108224 /*105017.5*/) goto t14_n138;
  t14_n137: return 0.09529575705528259f;
  t14_n138: return -0.09803763031959534f;
  t14_n139: if (E.getNumReferences() >= 3332163840 /*20068.5*/) goto t14_n143;
  t14_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t14_n142;
  t14_n141: return 0.06623811274766922f;
  t14_n142: return 0.08802172541618347f;
  t14_n143: return 0.1029152050614357f;
  t14_n144: return 0.10383038967847824f;
  t14_n145: return 0.09434624761343002f;
  t14_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t14_n214;
  t14_n147: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n205;
  t14_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t14_n150;
  t14_n149: return -0.10120445489883423f;
  t14_n150: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t14_n186;
  t14_n151: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t14_n177;
  t14_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t14_n176;
  t14_n153: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t14_n171;
  t14_n154: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t14_n158;
  t14_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n157;
  t14_n156: return -0.04695713520050049f;
  t14_n157: return 0.06361015886068344f;
  t14_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t14_n164;
  t14_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n161;
  t14_n160: return -0.06782558560371399f;
  t14_n161: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t14_n163;
  t14_n162: return -0.11956846714019775f;
  t14_n163: return 0.025550849735736847f;
  t14_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t14_n168;
  t14_n165: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t14_n167;
  t14_n166: return 0.03468400612473488f;
  t14_n167: return 0.13511237502098083f;
  t14_n168: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t14_n170;
  t14_n169: return 0.1254875808954239f;
  t14_n170: return 0.10935458540916443f;
  t14_n171: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t14_n175;
  t14_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t14_n174;
  t14_n173: return 0.0298733152449131f;
  t14_n174: return 0.06930091232061386f;
  t14_n175: return -0.10294905304908752f;
  t14_n176: return -0.028260396793484688f;
  t14_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n179;
  t14_n178: return -0.10712731629610062f;
  t14_n179: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t14_n183;
  t14_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t14_n182;
  t14_n181: return -0.05711158365011215f;
  t14_n182: return 0.1213301494717598f;
  t14_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t14_n185;
  t14_n184: return 0.03682245314121246f;
  t14_n185: return 0.07917479425668716f;
  t14_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t14_n204;
  t14_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t14_n195;
  t14_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n192;
  t14_n189: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t14_n191;
  t14_n190: return 0.03474780544638634f;
  t14_n191: return -0.12481459230184555f;
  t14_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t14_n194;
  t14_n193: return 0.028782416135072708f;
  t14_n194: return 0.06257394701242447f;
  t14_n195: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t14_n201;
  t14_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t14_n200;
  t14_n197: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t14_n199;
  t14_n198: return 0.047610461711883545f;
  t14_n199: return -0.24455896019935608f;
  t14_n200: return 0.0824362114071846f;
  t14_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol))) goto t14_n203;
  t14_n202: return 0.07607262581586838f;
  t14_n203: return 0.11907033622264862f;
  t14_n204: return -0.09496340155601501f;
  t14_n205: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t14_n211;
  t14_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t14_n210;
  t14_n207: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t14_n209;
  t14_n208: return 0.0345756858587265f;
  t14_n209: return -0.06949637085199356f;
  t14_n210: return 0.06082748994231224f;
  t14_n211: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t14_n213;
  t14_n212: return 0.061605457216501236f;
  t14_n213: return 0.08910807222127914f;
  t14_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t14_n218;
  t14_n215: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t14_n217;
  t14_n216: return -0.12100532650947571f;
  t14_n217: return 0.013620343990623951f;
  t14_n218: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t14_n232;
  t14_n219: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t14_n231;
  t14_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t14_n230;
  t14_n221: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t14_n223;
  t14_n222: return 0.022218333557248116f;
  t14_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t14_n229;
  t14_n224: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t14_n226;
  t14_n225: return 0.028450047597289085f;
  t14_n226: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n228;
  t14_n227: return 0.046204887330532074f;
  t14_n228: return 0.0883917510509491f;
  t14_n229: return 0.08392440527677536f;
  t14_n230: return 0.06115452200174332f;
  t14_n231: return -0.129404217004776f;
  t14_n232: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t14_n236;
  t14_n233: if (E.getFractionNameInContext() >= 3175960628 /*0.05012531578540802*/) goto t14_n235;
  t14_n234: return 0.06735528260469437f;
  t14_n235: return 0.08810502290725708f;
  t14_n236: return 0.09451346844434738f;
  t14_n237: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t14_n245;
  t14_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t14_n242;
  t14_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t14_n241;
  t14_n240: return 0.03711801767349243f;
  t14_n241: return 0.10502982884645462f;
  t14_n242: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t14_n244;
  t14_n243: return 0.09249448776245117f;
  t14_n244: return 0.10773327946662903f;
  t14_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t14_n247;
  t14_n246: return 0.10247695446014404f;
  t14_n247: return 0.1091419905424118f;
  t14_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t14_n250;
  t14_n249: return 0.0072029950097203255f;
  t14_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t14_n254;
  t14_n251: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t14_n253;
  t14_n252: return 0.09782491624355316f;
  t14_n253: return 0.1032983735203743f;
  t14_n254: return 0.10852354764938354f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree15(const Example& E) {
  t15_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t15_n248;
  t15_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t15_n239;
  t15_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t15_n104;
  t15_n3: if (E.getFractionNameInContext() >= 3184099116 /*0.09838709235191345*/) goto t15_n47;
  t15_n4: if (E.getNumReferences() >= 3328224256 /*14379.0*/) goto t15_n34;
  t15_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t15_n11;
  t15_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n8;
  t15_n7: return -0.0911315530538559f;
  t15_n8: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t15_n10;
  t15_n9: return 0.03515787422657013f;
  t15_n10: return -0.016008706763386726f;
  t15_n11: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t15_n17;
  t15_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t15_n16;
  t15_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n15;
  t15_n14: return 0.002784733660519123f;
  t15_n15: return 0.04697931930422783f;
  t15_n16: return 0.08893464505672455f;
  t15_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t15_n19;
  t15_n18: return -0.06244746595621109f;
  t15_n19: if (E.getNumReferences() >= 3294044160 /*860.5*/) goto t15_n33;
  t15_n20: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t15_n26;
  t15_n21: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t15_n25;
  t15_n22: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t15_n24;
  t15_n23: return -0.013002175837755203f;
  t15_n24: return 0.03448950871825218f;
  t15_n25: return -0.12437527626752853f;
  t15_n26: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t15_n28;
  t15_n27: return -0.08890199661254883f;
  t15_n28: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t15_n30;
  t15_n29: return 0.025974851101636887f;
  t15_n30: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t15_n32;
  t15_n31: return 0.06979309022426605f;
  t15_n32: return -0.07731132954359055f;
  t15_n33: return 0.03164903074502945f;
  t15_n34: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n46;
  t15_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t15_n43;
  t15_n36: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t15_n42;
  t15_n37: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t15_n39;
  t15_n38: return 0.0409991517663002f;
  t15_n39: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t15_n41;
  t15_n40: return 0.09662064909934998f;
  t15_n41: return 0.0435328483581543f;
  t15_n42: return 0.0016195903299376369f;
  t15_n43: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t15_n45;
  t15_n44: return 0.05231987312436104f;
  t15_n45: return 0.08904615789651871f;
  t15_n46: return 0.0866396501660347f;
  t15_n47: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t15_n93;
  t15_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t15_n68;
  t15_n49: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t15_n63;
  t15_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t15_n58;
  t15_n51: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t15_n55;
  t15_n52: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n54;
  t15_n53: return 0.03572545573115349f;
  t15_n54: return 0.09131273627281189f;
  t15_n55: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t15_n57;
  t15_n56: return -0.011268144473433495f;
  t15_n57: return 0.05544014647603035f;
  t15_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t15_n60;
  t15_n59: return 0.06772947311401367f;
  t15_n60: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t15_n62;
  t15_n61: return 0.0733523815870285f;
  t15_n62: return 0.10387175530195236f;
  t15_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t15_n67;
  t15_n64: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t15_n66;
  t15_n65: return 0.05829944461584091f;
  t15_n66: return 0.08731140196323395f;
  t15_n67: return 0.09294410794973373f;
  t15_n68: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t15_n86;
  t15_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t15_n77;
  t15_n70: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n76;
  t15_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t15_n73;
  t15_n72: return -0.07100468873977661f;
  t15_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t15_n75;
  t15_n74: return -0.028808599337935448f;
  t15_n75: return 0.034276243299245834f;
  t15_n76: return 0.03953590989112854f;
  t15_n77: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t15_n83;
  t15_n78: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t15_n82;
  t15_n79: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t15_n81;
  t15_n80: return 0.044608231633901596f;
  t15_n81: return 0.0974775180220604f;
  t15_n82: return 0.001525506959296763f;
  t15_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t15_n85;
  t15_n84: return 0.03503930941224098f;
  t15_n85: return 0.08171176165342331f;
  t15_n86: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t15_n92;
  t15_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t15_n89;
  t15_n88: return 0.045305829495191574f;
  t15_n89: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t15_n91;
  t15_n90: return 0.031257905066013336f;
  t15_n91: return 0.09844062477350235f;
  t15_n92: return 0.11242365092039108f;
  t15_n93: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t15_n99;
  t15_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t15_n96;
  t15_n95: return 0.0552511066198349f;
  t15_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t15_n98;
  t15_n97: return 0.07061053067445755f;
  t15_n98: return 0.09359394013881683f;
  t15_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t15_n101;
  t15_n100: return 0.08555863052606583f;
  t15_n101: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t15_n103;
  t15_n102: return 0.09798385947942734f;
  t15_n103: return 0.12616249918937683f;
  t15_n104: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t15_n148;
  t15_n105: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t15_n139;
  t15_n106: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t15_n130;
  t15_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n121;
  t15_n108: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t15_n120;
  t15_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t15_n115;
  t15_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t15_n114;
  t15_n111: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t15_n113;
  t15_n112: return 0.045023318380117416f;
  t15_n113: return 0.07593867182731628f;
  t15_n114: return -0.0029398594051599503f;
  t15_n115: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t15_n119;
  t15_n116: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n118;
  t15_n117: return 0.08726586401462555f;
  t15_n118: return -0.04171106219291687f;
  t15_n119: return 0.08053300529718399f;
  t15_n120: return 0.07992244511842728f;
  t15_n121: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t15_n127;
  t15_n122: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n126;
  t15_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t15_n125;
  t15_n124: return 0.08214253187179565f;
  t15_n125: return 0.10651647299528122f;
  t15_n126: return -0.014508834108710289f;
  t15_n127: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t15_n129;
  t15_n128: return 0.06571376323699951f;
  t15_n129: return 0.08564937859773636f;
  t15_n130: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n138;
  t15_n131: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t15_n137;
  t15_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n136;
  t15_n133: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t15_n135;
  t15_n134: return 0.05087350681424141f;
  t15_n135: return -0.018062081187963486f;
  t15_n136: return -0.05625011771917343f;
  t15_n137: return 0.060641098767519f;
  t15_n138: return 0.05848066136240959f;
  t15_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n147;
  t15_n140: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t15_n146;
  t15_n141: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t15_n143;
  t15_n142: return 0.06722456216812134f;
  t15_n143: if (E.getNumReferences() >= 3332163840 /*20068.5*/) goto t15_n145;
  t15_n144: return 0.0806601271033287f;
  t15_n145: return 0.10265659540891647f;
  t15_n146: return 0.10232718288898468f;
  t15_n147: return 0.0926566943526268f;
  t15_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t15_n214;
  t15_n149: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n205;
  t15_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t15_n152;
  t15_n151: return -0.09958402067422867f;
  t15_n152: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t15_n188;
  t15_n153: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t15_n179;
  t15_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t15_n178;
  t15_n155: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n173;
  t15_n156: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t15_n160;
  t15_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n159;
  t15_n158: return -0.04230114817619324f;
  t15_n159: return 0.05907713994383812f;
  t15_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t15_n166;
  t15_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n163;
  t15_n162: return -0.06496978551149368f;
  t15_n163: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t15_n165;
  t15_n164: return -0.1123284250497818f;
  t15_n165: return 0.023184198886156082f;
  t15_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t15_n170;
  t15_n167: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t15_n169;
  t15_n168: return 0.036197058856487274f;
  t15_n169: return 0.12934742867946625f;
  t15_n170: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t15_n172;
  t15_n171: return 0.11672291904687881f;
  t15_n172: return 0.1076982393860817f;
  t15_n173: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t15_n177;
  t15_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t15_n176;
  t15_n175: return 0.0261534433811903f;
  t15_n176: return 0.06582117825746536f;
  t15_n177: return -0.10146594047546387f;
  t15_n178: return -0.026385655626654625f;
  t15_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n181;
  t15_n180: return -0.10274237394332886f;
  t15_n181: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t15_n185;
  t15_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t15_n184;
  t15_n183: return -0.0555889867246151f;
  t15_n184: return 0.11551547050476074f;
  t15_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t15_n187;
  t15_n186: return 0.03358772024512291f;
  t15_n187: return 0.07377468049526215f;
  t15_n188: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t15_n204;
  t15_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t15_n197;
  t15_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n194;
  t15_n191: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t15_n193;
  t15_n192: return 0.0307061318308115f;
  t15_n193: return -0.11306288838386536f;
  t15_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t15_n196;
  t15_n195: return 0.027072608470916748f;
  t15_n196: return 0.05921868979930878f;
  t15_n197: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t15_n201;
  t15_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t15_n200;
  t15_n199: return 0.04163969308137894f;
  t15_n200: return 0.08194580674171448f;
  t15_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol))) goto t15_n203;
  t15_n202: return 0.07372362166643143f;
  t15_n203: return 0.11270631104707718f;
  t15_n204: return -0.09284079819917679f;
  t15_n205: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t15_n211;
  t15_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t15_n210;
  t15_n207: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t15_n209;
  t15_n208: return 0.03238915652036667f;
  t15_n209: return -0.07166169583797455f;
  t15_n210: return 0.05676628649234772f;
  t15_n211: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t15_n213;
  t15_n212: return 0.058037903159856796f;
  t15_n213: return 0.08564498275518417f;
  t15_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n216;
  t15_n215: return -0.03788677603006363f;
  t15_n216: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t15_n234;
  t15_n217: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t15_n229;
  t15_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t15_n224;
  t15_n219: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n221;
  t15_n220: return -0.02381359227001667f;
  t15_n221: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n223;
  t15_n222: return 0.023788420483469963f;
  t15_n223: return 0.0937814861536026f;
  t15_n224: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t15_n228;
  t15_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t15_n227;
  t15_n226: return 0.018626442179083824f;
  t15_n227: return 0.04711455851793289f;
  t15_n228: return 0.10346589982509613f;
  t15_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t15_n233;
  t15_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t15_n232;
  t15_n231: return -0.038424428552389145f;
  t15_n232: return 0.051702357828617096f;
  t15_n233: return 0.08116085082292557f;
  t15_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t15_n238;
  t15_n235: if (E.getFractionNameInContext() >= 3175960628 /*0.05012531578540802*/) goto t15_n237;
  t15_n236: return 0.06374350190162659f;
  t15_n237: return 0.08540398627519608f;
  t15_n238: return 0.092014379799366f;
  t15_n239: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t15_n247;
  t15_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t15_n244;
  t15_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t15_n243;
  t15_n242: return 0.03399127721786499f;
  t15_n243: return 0.10344266146421432f;
  t15_n244: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t15_n246;
  t15_n245: return 0.09022273123264313f;
  t15_n246: return 0.10644551366567612f;
  t15_n247: return 0.10491655021905899f;
  t15_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t15_n250;
  t15_n249: return 0.0007781512686051428f;
  t15_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t15_n254;
  t15_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t15_n253;
  t15_n252: return 0.09560608863830566f;
  t15_n253: return 0.10188520699739456f;
  t15_n254: return 0.10766410827636719f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree16(const Example& E) {
  t16_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t16_n248;
  t16_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t16_n241;
  t16_n2: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t16_n104;
  t16_n3: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t16_n59;
  t16_n4: if (E.getNumReferences() >= 3301085184 /*1556.5*/) goto t16_n46;
  t16_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t16_n19;
  t16_n6: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t16_n12;
  t16_n7: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t16_n9;
  t16_n8: return -0.030070947483181953f;
  t16_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n11;
  t16_n10: return 0.028386536985635757f;
  t16_n11: return 0.09603055566549301f;
  t16_n12: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t16_n14;
  t16_n13: return -0.06722275912761688f;
  t16_n14: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t16_n16;
  t16_n15: return -0.10547428578138351f;
  t16_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t16_n18;
  t16_n17: return 0.0013825841015204787f;
  t16_n18: return -0.08178726583719254f;
  t16_n19: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t16_n45;
  t16_n20: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t16_n36;
  t16_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t16_n27;
  t16_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t16_n24;
  t16_n23: return -0.007045823615044355f;
  t16_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t16_n26;
  t16_n25: return 0.061797674745321274f;
  t16_n26: return 0.09994997829198837f;
  t16_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t16_n29;
  t16_n28: return -0.025571385398507118f;
  t16_n29: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t16_n31;
  t16_n30: return 0.0424933023750782f;
  t16_n31: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t16_n35;
  t16_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t16_n34;
  t16_n33: return -0.08138182759284973f;
  t16_n34: return 0.012796858325600624f;
  t16_n35: return -0.05747435986995697f;
  t16_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t16_n38;
  t16_n37: return -0.017809823155403137f;
  t16_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t16_n40;
  t16_n39: return -0.18460504710674286f;
  t16_n40: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t16_n42;
  t16_n41: return -0.0012865957105532289f;
  t16_n42: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t16_n44;
  t16_n43: return 0.08325820416212082f;
  t16_n44: return 0.054554883390665054f;
  t16_n45: return -0.1077837273478508f;
  t16_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t16_n54;
  t16_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t16_n51;
  t16_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n50;
  t16_n49: return -0.013270474970340729f;
  t16_n50: return 0.057303622364997864f;
  t16_n51: if (E.getNumReferences() >= 3338664448 /*32765.0*/) goto t16_n53;
  t16_n52: return 0.04779139906167984f;
  t16_n53: return 0.08678412437438965f;
  t16_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t16_n58;
  t16_n55: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n57;
  t16_n56: return 0.03450498729944229f;
  t16_n57: return 0.08551628887653351f;
  t16_n58: return 0.08653297275304794f;
  t16_n59: if (E.getNumReferences() >= 3309346816 /*3082.0*/) goto t16_n93;
  t16_n60: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t16_n74;
  t16_n61: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t16_n69;
  t16_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n68;
  t16_n63: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t16_n67;
  t16_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t16_n66;
  t16_n65: return 0.010164092294871807f;
  t16_n66: return 0.07780852913856506f;
  t16_n67: return 0.05323250591754913f;
  t16_n68: return 0.07281036674976349f;
  t16_n69: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t16_n73;
  t16_n70: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n72;
  t16_n71: return 0.06313673406839371f;
  t16_n72: return 0.09780406951904297f;
  t16_n73: return 0.09863559156656265f;
  t16_n74: if (E.getFileProximityDistanceCost() >= 3255173120 /*33.5*/) goto t16_n88;
  t16_n75: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t16_n85;
  t16_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t16_n80;
  t16_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n79;
  t16_n78: return -0.07145851105451584f;
  t16_n79: return 0.03248584270477295f;
  t16_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t16_n82;
  t16_n81: return -0.004843064583837986f;
  t16_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t16_n84;
  t16_n83: return 0.02618560753762722f;
  t16_n84: return 0.05786756053566933f;
  t16_n85: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t16_n87;
  t16_n86: return 0.041096899658441544f;
  t16_n87: return 0.08808881789445877f;
  t16_n88: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t16_n92;
  t16_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t16_n91;
  t16_n90: return 0.01701187528669834f;
  t16_n91: return 0.07517288625240326f;
  t16_n92: return 0.09053497016429901f;
  t16_n93: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t16_n99;
  t16_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t16_n96;
  t16_n95: return 0.053896088153123856f;
  t16_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t16_n98;
  t16_n97: return 0.06847387552261353f;
  t16_n98: return 0.09343913197517395f;
  t16_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t16_n101;
  t16_n100: return 0.08566107600927353f;
  t16_n101: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t16_n103;
  t16_n102: return 0.09748547524213791f;
  t16_n103: return 0.1242254450917244f;
  t16_n104: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t16_n152;
  t16_n105: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t16_n141;
  t16_n106: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t16_n128;
  t16_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n119;
  t16_n108: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t16_n118;
  t16_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t16_n113;
  t16_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t16_n112;
  t16_n111: return 0.04376782104372978f;
  t16_n112: return -0.00594095466658473f;
  t16_n113: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t16_n117;
  t16_n114: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n116;
  t16_n115: return 0.08432701975107193f;
  t16_n116: return -0.03770134225487709f;
  t16_n117: return 0.07714498043060303f;
  t16_n118: return 0.07673054188489914f;
  t16_n119: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t16_n127;
  t16_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n122;
  t16_n121: return 0.06182459369301796f;
  t16_n122: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t16_n126;
  t16_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t16_n125;
  t16_n124: return 0.07572401314973831f;
  t16_n125: return 0.10476820170879364f;
  t16_n126: return 0.06255875527858734f;
  t16_n127: return 0.08897951245307922f;
  t16_n128: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t16_n138;
  t16_n129: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n135;
  t16_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n132;
  t16_n131: return -0.031519822776317596f;
  t16_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t16_n134;
  t16_n133: return -0.1677902489900589f;
  t16_n134: return 0.048004455864429474f;
  t16_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t16_n137;
  t16_n136: return 0.03745410963892937f;
  t16_n137: return 0.08606702834367752f;
  t16_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n140;
  t16_n139: return 0.05551770329475403f;
  t16_n140: return -0.08636679500341415f;
  t16_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n149;
  t16_n142: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t16_n148;
  t16_n143: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t16_n145;
  t16_n144: return 0.06254161894321442f;
  t16_n145: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t16_n147;
  t16_n146: return 0.0691598579287529f;
  t16_n147: return 0.08890411257743835f;
  t16_n148: return 0.1008118987083435f;
  t16_n149: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t16_n151;
  t16_n150: return 0.1005592942237854f;
  t16_n151: return 0.0842849537730217f;
  t16_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t16_n214;
  t16_n153: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n205;
  t16_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t16_n156;
  t16_n155: return -0.09785819053649902f;
  t16_n156: if (E.getFractionNameInContext() >= 3172203529 /*0.03612903133034706*/) goto t16_n178;
  t16_n157: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t16_n175;
  t16_n158: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t16_n174;
  t16_n159: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n173;
  t16_n160: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t16_n164;
  t16_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n163;
  t16_n162: return -0.04560280218720436f;
  t16_n163: return 0.04565580561757088f;
  t16_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t16_n168;
  t16_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n167;
  t16_n166: return -0.06183901056647301f;
  t16_n167: return 0.016799502074718475f;
  t16_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t16_n172;
  t16_n169: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t16_n171;
  t16_n170: return 0.031966179609298706f;
  t16_n171: return 0.12959136068820953f;
  t16_n172: return 0.11515393108129501f;
  t16_n173: return 0.035079196095466614f;
  t16_n174: return -0.031169962137937546f;
  t16_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n177;
  t16_n176: return -0.09876450151205063f;
  t16_n177: return 0.03112604096531868f;
  t16_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t16_n204;
  t16_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n197;
  t16_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t16_n192;
  t16_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n183;
  t16_n182: return -0.038443900644779205f;
  t16_n183: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t16_n189;
  t16_n184: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t16_n186;
  t16_n185: return -0.06762784719467163f;
  t16_n186: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t16_n188;
  t16_n187: return 0.022398030385375023f;
  t16_n188: return 0.04340213164687157f;
  t16_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n191;
  t16_n190: return -0.0837133377790451f;
  t16_n191: return 0.05053173378109932f;
  t16_n192: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t16_n194;
  t16_n193: return 0.02572273276746273f;
  t16_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t16_n196;
  t16_n195: return 0.06292790174484253f;
  t16_n196: return 0.10970078408718109f;
  t16_n197: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t16_n203;
  t16_n198: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t16_n202;
  t16_n199: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t16_n201;
  t16_n200: return 0.04629683867096901f;
  t16_n201: return -0.009459925815463066f;
  t16_n202: return 0.08325625211000443f;
  t16_n203: return 0.06546196341514587f;
  t16_n204: return -0.08932628482580185f;
  t16_n205: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t16_n211;
  t16_n206: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t16_n208;
  t16_n207: return 0.03805574029684067f;
  t16_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n210;
  t16_n209: return -0.07340645045042038f;
  t16_n210: return 0.06972993165254593f;
  t16_n211: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t16_n213;
  t16_n212: return 0.05477450415492058f;
  t16_n213: return 0.0824725478887558f;
  t16_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n218;
  t16_n215: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t16_n217;
  t16_n216: return -0.11831093579530716f;
  t16_n217: return 0.014257458969950676f;
  t16_n218: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t16_n236;
  t16_n219: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t16_n231;
  t16_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t16_n228;
  t16_n221: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n223;
  t16_n222: return -0.027682101354002953f;
  t16_n223: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n227;
  t16_n224: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t16_n226;
  t16_n225: return -0.043397169560194016f;
  t16_n226: return 0.07010643184185028f;
  t16_n227: return 0.09153735637664795f;
  t16_n228: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t16_n230;
  t16_n229: return 0.029521768912672997f;
  t16_n230: return 0.10135728120803833f;
  t16_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t16_n235;
  t16_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t16_n234;
  t16_n233: return -0.03432818874716759f;
  t16_n234: return 0.04870620742440224f;
  t16_n235: return 0.07804930955171585f;
  t16_n236: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t16_n240;
  t16_n237: if (E.getFractionNameInContext() >= 3175960628 /*0.05012531578540802*/) goto t16_n239;
  t16_n238: return 0.05979437753558159f;
  t16_n239: return 0.08284495770931244f;
  t16_n240: return 0.08972368389368057f;
  t16_n241: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t16_n247;
  t16_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t16_n244;
  t16_n243: return 0.039618007838726044f;
  t16_n244: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t16_n246;
  t16_n245: return 0.08658834546804428f;
  t16_n246: return 0.10497349500656128f;
  t16_n247: return 0.10392142087221146f;
  t16_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t16_n250;
  t16_n249: return 0.0008542459108866751f;
  t16_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t16_n254;
  t16_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t16_n253;
  t16_n252: return 0.09347398579120636f;
  t16_n253: return 0.10036642104387283f;
  t16_n254: return 0.10697859525680542f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree17(const Example& E) {
  t17_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t17_n248;
  t17_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t17_n239;
  t17_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t17_n112;
  t17_n3: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t17_n61;
  t17_n4: if (E.getNumReferences() >= 3326663680 /*12855.0*/) goto t17_n54;
  t17_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t17_n13;
  t17_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n10;
  t17_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t17_n9;
  t17_n8: return -0.1226615309715271f;
  t17_n9: return -0.08322560787200928f;
  t17_n10: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t17_n12;
  t17_n11: return 0.018006235361099243f;
  t17_n12: return -0.0465724878013134f;
  t17_n13: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t17_n19;
  t17_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t17_n18;
  t17_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n17;
  t17_n16: return 0.0026393099687993526f;
  t17_n17: return 0.04339185729622841f;
  t17_n18: return 0.08146128058433533f;
  t17_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t17_n25;
  t17_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t17_n22;
  t17_n21: return -0.05984886735677719f;
  t17_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n24;
  t17_n23: return -0.0016724448651075363f;
  t17_n24: return 0.06404628604650497f;
  t17_n25: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t17_n53;
  t17_n26: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t17_n52;
  t17_n27: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t17_n51;
  t17_n28: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t17_n44;
  t17_n29: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t17_n33;
  t17_n30: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t17_n32;
  t17_n31: return -0.00153436535038054f;
  t17_n32: return -0.05664217844605446f;
  t17_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t17_n41;
  t17_n34: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t17_n36;
  t17_n35: return -0.024636687710881233f;
  t17_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n38;
  t17_n37: return -0.04680269584059715f;
  t17_n38: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t17_n40;
  t17_n39: return 0.09517379105091095f;
  t17_n40: return 0.04054529592394829f;
  t17_n41: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t17_n43;
  t17_n42: return -0.005242961924523115f;
  t17_n43: return 0.05113357678055763f;
  t17_n44: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t17_n46;
  t17_n45: return -0.06729865819215775f;
  t17_n46: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t17_n50;
  t17_n47: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t17_n49;
  t17_n48: return 0.05984092131257057f;
  t17_n49: return 0.026130683720111847f;
  t17_n50: return -0.03836371377110481f;
  t17_n51: return -0.08655684441328049f;
  t17_n52: return 0.031095607206225395f;
  t17_n53: return -0.12292716652154922f;
  t17_n54: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t17_n58;
  t17_n55: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t17_n57;
  t17_n56: return 0.037275463342666626f;
  t17_n57: return 0.07258249074220657f;
  t17_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t17_n60;
  t17_n59: return 0.002953986171633005f;
  t17_n60: return 0.07656170427799225f;
  t17_n61: if (E.getNumReferences() >= 3309346816 /*3082.0*/) goto t17_n101;
  t17_n62: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t17_n74;
  t17_n63: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t17_n71;
  t17_n64: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n70;
  t17_n65: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t17_n69;
  t17_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t17_n68;
  t17_n67: return 0.008692118339240551f;
  t17_n68: return 0.05480740964412689f;
  t17_n69: return 0.05119713768362999f;
  t17_n70: return 0.06859373301267624f;
  t17_n71: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t17_n73;
  t17_n72: return 0.06275281310081482f;
  t17_n73: return 0.09612111002206802f;
  t17_n74: if (E.getFileProximityDistanceCost() >= 3255173120 /*33.5*/) goto t17_n96;
  t17_n75: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t17_n89;
  t17_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t17_n80;
  t17_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n79;
  t17_n78: return -0.06846483796834946f;
  t17_n79: return 0.030569203197956085f;
  t17_n80: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t17_n82;
  t17_n81: return -0.0049885110929608345f;
  t17_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t17_n86;
  t17_n83: if (E.getNumReferences() >= 3269918720 /*115.5*/) goto t17_n85;
  t17_n84: return 0.0143196489661932f;
  t17_n85: return 0.07171657681465149f;
  t17_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t17_n88;
  t17_n87: return 0.02163633145391941f;
  t17_n88: return 0.08570580929517746f;
  t17_n89: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t17_n95;
  t17_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t17_n92;
  t17_n91: return 0.03374047949910164f;
  t17_n92: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t17_n94;
  t17_n93: return 0.011772441677749157f;
  t17_n94: return 0.09992841631174088f;
  t17_n95: return 0.08466055989265442f;
  t17_n96: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t17_n100;
  t17_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t17_n99;
  t17_n98: return 0.018790749832987785f;
  t17_n99: return 0.07144811749458313f;
  t17_n100: return 0.08746077865362167f;
  t17_n101: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t17_n107;
  t17_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t17_n104;
  t17_n103: return 0.05038420855998993f;
  t17_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t17_n106;
  t17_n105: return 0.06616242229938507f;
  t17_n106: return 0.09092336893081665f;
  t17_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t17_n109;
  t17_n108: return 0.08226369321346283f;
  t17_n109: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t17_n111;
  t17_n110: return 0.09531722962856293f;
  t17_n111: return 0.12177357822656631f;
  t17_n112: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t17_n156;
  t17_n113: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t17_n139;
  t17_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t17_n124;
  t17_n115: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t17_n121;
  t17_n116: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t17_n120;
  t17_n117: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n119;
  t17_n118: return 0.014907168224453926f;
  t17_n119: return 0.05461858585476875f;
  t17_n120: return 0.07058615982532501f;
  t17_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t17_n123;
  t17_n122: return 0.05139973759651184f;
  t17_n123: return -0.0782516822218895f;
  t17_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n132;
  t17_n125: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t17_n127;
  t17_n126: return 0.03720085322856903f;
  t17_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t17_n131;
  t17_n128: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t17_n130;
  t17_n129: return 0.045197390019893646f;
  t17_n130: return 0.07720883935689926f;
  t17_n131: return 0.08425101637840271f;
  t17_n132: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t17_n138;
  t17_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n135;
  t17_n134: return 0.05906052514910698f;
  t17_n135: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t17_n137;
  t17_n136: return 0.08878657966852188f;
  t17_n137: return 0.05880400165915489f;
  t17_n138: return 0.08651160448789597f;
  t17_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n153;
  t17_n140: if (E.getFractionNameInContext() >= 3173545706 /*0.04112903028726578*/) goto t17_n144;
  t17_n141: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t17_n143;
  t17_n142: return 0.058705661445856094f;
  t17_n143: return 0.0974545031785965f;
  t17_n144: if (E.getNumReferences() >= 3332163840 /*20068.5*/) goto t17_n152;
  t17_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t17_n151;
  t17_n146: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n150;
  t17_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n149;
  t17_n148: return -0.03806950896978378f;
  t17_n149: return 0.06076180934906006f;
  t17_n150: return 0.0904197245836258f;
  t17_n151: return 0.07894042134284973f;
  t17_n152: return 0.10041303187608719f;
  t17_n153: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t17_n155;
  t17_n154: return 0.09626863151788712f;
  t17_n155: return 0.0742807611823082f;
  t17_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t17_n188;
  t17_n157: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n179;
  t17_n158: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t17_n176;
  t17_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t17_n161;
  t17_n160: return -0.09589508920907974f;
  t17_n161: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t17_n173;
  t17_n162: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t17_n166;
  t17_n163: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t17_n165;
  t17_n164: return 0.015276617370545864f;
  t17_n165: return -0.012827591970562935f;
  t17_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n168;
  t17_n167: return -0.0933203473687172f;
  t17_n168: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t17_n170;
  t17_n169: return -0.07782524824142456f;
  t17_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t17_n172;
  t17_n171: return 0.03029704838991165f;
  t17_n172: return 0.06768301129341125f;
  t17_n173: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t17_n175;
  t17_n174: return 0.0446418821811676f;
  t17_n175: return 0.01946314051747322f;
  t17_n176: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t17_n178;
  t17_n177: return 0.0073483227752149105f;
  t17_n178: return -0.10805855691432953f;
  t17_n179: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t17_n185;
  t17_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t17_n184;
  t17_n181: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t17_n183;
  t17_n182: return 0.02554067224264145f;
  t17_n183: return -0.06783122569322586f;
  t17_n184: return 0.052094608545303345f;
  t17_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t17_n187;
  t17_n186: return 0.05202171579003334f;
  t17_n187: return 0.08039624989032745f;
  t17_n188: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t17_n232;
  t17_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n201;
  t17_n190: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t17_n196;
  t17_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t17_n193;
  t17_n192: return -0.03944515436887741f;
  t17_n193: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t17_n195;
  t17_n194: return 0.023298248648643494f;
  t17_n195: return 0.06658948212862015f;
  t17_n196: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t17_n198;
  t17_n197: return 0.05201687663793564f;
  t17_n198: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n200;
  t17_n199: return -0.015641018748283386f;
  t17_n200: return 0.08695971965789795f;
  t17_n201: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t17_n219;
  t17_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t17_n206;
  t17_n203: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n205;
  t17_n204: return -0.05488847941160202f;
  t17_n205: return 0.055185817182064056f;
  t17_n206: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t17_n214;
  t17_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t17_n209;
  t17_n208: return 0.03365994989871979f;
  t17_n209: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t17_n213;
  t17_n210: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t17_n212;
  t17_n211: return 0.06000091880559921f;
  t17_n212: return -0.18807312846183777f;
  t17_n213: return 0.10088670253753662f;
  t17_n214: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t17_n218;
  t17_n215: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t17_n217;
  t17_n216: return 0.1141337975859642f;
  t17_n217: return 0.09379938989877701f;
  t17_n218: return -0.000959669123403728f;
  t17_n219: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t17_n231;
  t17_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t17_n226;
  t17_n221: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n223;
  t17_n222: return 0.04730633646249771f;
  t17_n223: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t17_n225;
  t17_n224: return 0.033188726752996445f;
  t17_n225: return 0.11371500045061111f;
  t17_n226: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t17_n230;
  t17_n227: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t17_n229;
  t17_n228: return 0.07704175263643265f;
  t17_n229: return -0.2138231098651886f;
  t17_n230: return 0.10114303976297379f;
  t17_n231: return -0.07287584990262985f;
  t17_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t17_n234;
  t17_n233: return 0.014156310819089413f;
  t17_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t17_n238;
  t17_n235: if (E.getFractionNameInContext() >= 3175960628 /*0.05012531578540802*/) goto t17_n237;
  t17_n236: return 0.056297071278095245f;
  t17_n237: return 0.07991767674684525f;
  t17_n238: return 0.08726968616247177f;
  t17_n239: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t17_n247;
  t17_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t17_n244;
  t17_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t17_n243;
  t17_n242: return 0.029828136786818504f;
  t17_n243: return 0.10077531635761261f;
  t17_n244: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t17_n246;
  t17_n245: return 0.08545621484518051f;
  t17_n246: return 0.1042909100651741f;
  t17_n247: return 0.10270808637142181f;
  t17_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t17_n250;
  t17_n249: return 0.0007001198828220367f;
  t17_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t17_n254;
  t17_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t17_n253;
  t17_n252: return 0.09120485186576843f;
  t17_n253: return 0.09890985488891602f;
  t17_n254: return 0.10619451105594635f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree18(const Example& E) {
  t18_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t18_n248;
  t18_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t18_n241;
  t18_n2: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t18_n128;
  t18_n3: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t18_n69;
  t18_n4: if (E.getNumReferences() >= 3301085184 /*1556.5*/) goto t18_n58;
  t18_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t18_n11;
  t18_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n10;
  t18_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t18_n9;
  t18_n8: return -0.1219959557056427f;
  t18_n9: return -0.08494135737419128f;
  t18_n10: return -0.001919018686749041f;
  t18_n11: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t18_n23;
  t18_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t18_n20;
  t18_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t18_n15;
  t18_n14: return -0.01566029153764248f;
  t18_n15: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t18_n19;
  t18_n16: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t18_n18;
  t18_n17: return 0.04712570086121559f;
  t18_n18: return 0.024693936109542847f;
  t18_n19: return 0.005919786635786295f;
  t18_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t18_n22;
  t18_n21: return 0.05547989904880524f;
  t18_n22: return 0.09353113174438477f;
  t18_n23: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t18_n53;
  t18_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t18_n30;
  t18_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t18_n27;
  t18_n26: return -0.04875052347779274f;
  t18_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t18_n29;
  t18_n28: return 0.00037049909587949514f;
  t18_n29: return 0.05796661227941513f;
  t18_n30: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t18_n40;
  t18_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t18_n33;
  t18_n32: return -0.04051407799124718f;
  t18_n33: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t18_n35;
  t18_n34: return 0.005281738005578518f;
  t18_n35: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t18_n37;
  t18_n36: return -0.027800654992461205f;
  t18_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n39;
  t18_n38: return -0.013929623179137707f;
  t18_n39: return 0.07919907569885254f;
  t18_n40: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t18_n46;
  t18_n41: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t18_n45;
  t18_n42: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t18_n44;
  t18_n43: return 0.03135722130537033f;
  t18_n44: return -0.10378704965114594f;
  t18_n45: return -0.049095846712589264f;
  t18_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t18_n52;
  t18_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t18_n49;
  t18_n48: return -0.09459899365901947f;
  t18_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t18_n51;
  t18_n50: return 0.07554405182600021f;
  t18_n51: return 0.11510644108057022f;
  t18_n52: return -0.1388096809387207f;
  t18_n53: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t18_n55;
  t18_n54: return -0.10354798287153244f;
  t18_n55: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t18_n57;
  t18_n56: return 0.017056234180927277f;
  t18_n57: return -0.07829947024583817f;
  t18_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t18_n64;
  t18_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t18_n63;
  t18_n60: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n62;
  t18_n61: return -0.010785254649817944f;
  t18_n62: return 0.05185173824429512f;
  t18_n63: return 0.07250222563743591f;
  t18_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t18_n68;
  t18_n65: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n67;
  t18_n66: return 0.03168477490544319f;
  t18_n67: return 0.08109401911497116f;
  t18_n68: return 0.08134656399488449f;
  t18_n69: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t18_n93;
  t18_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n82;
  t18_n71: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t18_n79;
  t18_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t18_n76;
  t18_n73: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t18_n75;
  t18_n74: return 0.04373518377542496f;
  t18_n75: return -0.0057802097871899605f;
  t18_n76: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t18_n78;
  t18_n77: return 0.06945833563804626f;
  t18_n78: return 0.009705744683742523f;
  t18_n79: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t18_n81;
  t18_n80: return -0.010627057403326035f;
  t18_n81: return 0.07780719548463821f;
  t18_n82: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t18_n92;
  t18_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t18_n87;
  t18_n84: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n86;
  t18_n85: return -0.002222369657829404f;
  t18_n86: return 0.07766464352607727f;
  t18_n87: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t18_n91;
  t18_n88: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t18_n90;
  t18_n89: return 0.09434252977371216f;
  t18_n90: return -0.0223320871591568f;
  t18_n91: return 0.06133650243282318f;
  t18_n92: return 0.09888317435979843f;
  t18_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t18_n103;
  t18_n94: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t18_n98;
  t18_n95: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n97;
  t18_n96: return -0.02149050123989582f;
  t18_n97: return 0.03177015855908394f;
  t18_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n100;
  t18_n99: return -0.08893885463476181f;
  t18_n100: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n102;
  t18_n101: return -0.06727046519517899f;
  t18_n102: return 0.04369499906897545f;
  t18_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n117;
  t18_n104: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t18_n112;
  t18_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t18_n111;
  t18_n106: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n108;
  t18_n107: return -0.047959815710783005f;
  t18_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n110;
  t18_n109: return -0.021145159378647804f;
  t18_n110: return 0.05077805742621422f;
  t18_n111: return 0.018862247467041016f;
  t18_n112: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t18_n116;
  t18_n113: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t18_n115;
  t18_n114: return 0.04656370356678963f;
  t18_n115: return 0.08152807503938675f;
  t18_n116: return -0.04959354177117348f;
  t18_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t18_n123;
  t18_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n122;
  t18_n119: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t18_n121;
  t18_n120: return 0.08077791333198547f;
  t18_n121: return -0.05068482831120491f;
  t18_n122: return 0.050685618072748184f;
  t18_n123: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t18_n125;
  t18_n124: return 0.04012814164161682f;
  t18_n125: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t18_n127;
  t18_n126: return 0.10557582974433899f;
  t18_n127: return -0.007807585876435041f;
  t18_n128: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t18_n176;
  t18_n129: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t18_n165;
  t18_n130: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t18_n142;
  t18_n131: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t18_n139;
  t18_n132: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t18_n138;
  t18_n133: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n137;
  t18_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t18_n136;
  t18_n135: return 0.010496031492948532f;
  t18_n136: return 0.05428953468799591f;
  t18_n137: return 0.06236223503947258f;
  t18_n138: return 0.0539453960955143f;
  t18_n139: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t18_n141;
  t18_n140: return 0.06225244700908661f;
  t18_n141: return 0.09232728183269501f;
  t18_n142: if (E.getFileProximityDistanceCost() >= 3255173120 /*33.5*/) goto t18_n160;
  t18_n143: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t18_n157;
  t18_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t18_n148;
  t18_n145: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n147;
  t18_n146: return -0.06688377261161804f;
  t18_n147: return 0.029356295242905617f;
  t18_n148: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t18_n150;
  t18_n149: return -0.004293648060411215f;
  t18_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t18_n154;
  t18_n151: if (E.getNumReferences() >= 3269918720 /*115.5*/) goto t18_n153;
  t18_n152: return 0.014663553796708584f;
  t18_n153: return 0.06835760176181793f;
  t18_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t18_n156;
  t18_n155: return 0.020803578197956085f;
  t18_n156: return 0.08237989991903305f;
  t18_n157: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t18_n159;
  t18_n158: return 0.03721620887517929f;
  t18_n159: return 0.08055759221315384f;
  t18_n160: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t18_n164;
  t18_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t18_n163;
  t18_n162: return 0.014592533931136131f;
  t18_n163: return 0.06785143166780472f;
  t18_n164: return 0.08402469009160995f;
  t18_n165: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t18_n169;
  t18_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t18_n168;
  t18_n167: return 0.04529193788766861f;
  t18_n168: return 0.07574821263551712f;
  t18_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t18_n171;
  t18_n170: return 0.07130883634090424f;
  t18_n171: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t18_n173;
  t18_n172: return 0.09733043611049652f;
  t18_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t18_n175;
  t18_n174: return 0.11189819127321243f;
  t18_n175: return 0.12056370079517365f;
  t18_n176: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t18_n192;
  t18_n177: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t18_n189;
  t18_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t18_n182;
  t18_n179: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n181;
  t18_n180: return 0.02803107164800167f;
  t18_n181: return 0.06364888697862625f;
  t18_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t18_n188;
  t18_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t18_n187;
  t18_n184: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t18_n186;
  t18_n185: return 0.06080296263098717f;
  t18_n186: return 0.033169347792863846f;
  t18_n187: return 0.08713843673467636f;
  t18_n188: return 0.0775546208024025f;
  t18_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t18_n191;
  t18_n190: return 0.08108855783939362f;
  t18_n191: return 0.09566812962293625f;
  t18_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t18_n194;
  t18_n193: return -0.07493442296981812f;
  t18_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n232;
  t18_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t18_n231;
  t18_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t18_n222;
  t18_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t18_n215;
  t18_n198: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t18_n212;
  t18_n199: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t18_n211;
  t18_n200: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t18_n210;
  t18_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t18_n205;
  t18_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n204;
  t18_n203: return -0.0553092323243618f;
  t18_n204: return 0.015397938899695873f;
  t18_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t18_n207;
  t18_n206: return 0.0176982618868351f;
  t18_n207: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t18_n209;
  t18_n208: return 0.046846356242895126f;
  t18_n209: return 0.09728778153657913f;
  t18_n210: return 0.033404089510440826f;
  t18_n211: return -0.08371195942163467f;
  t18_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t18_n214;
  t18_n213: return -0.0801248848438263f;
  t18_n214: return 0.031051723286509514f;
  t18_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t18_n219;
  t18_n216: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n218;
  t18_n217: return -0.0863114446401596f;
  t18_n218: return 0.05628510192036629f;
  t18_n219: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n221;
  t18_n220: return 0.04117821529507637f;
  t18_n221: return 0.07199114561080933f;
  t18_n222: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t18_n230;
  t18_n223: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n227;
  t18_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t18_n226;
  t18_n225: return 0.02880493365228176f;
  t18_n226: return 0.07182341068983078f;
  t18_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t18_n229;
  t18_n228: return 0.057220082730054855f;
  t18_n229: return 0.0930497944355011f;
  t18_n230: return -0.07258857786655426f;
  t18_n231: return 0.07175608724355698f;
  t18_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t18_n236;
  t18_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t18_n235;
  t18_n234: return -0.045385248959064484f;
  t18_n235: return 0.07859831303358078f;
  t18_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t18_n240;
  t18_n237: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t18_n239;
  t18_n238: return -0.11743790656328201f;
  t18_n239: return 0.06027140095829964f;
  t18_n240: return 0.08097516745328903f;
  t18_n241: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t18_n247;
  t18_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t18_n244;
  t18_n243: return 0.04600481316447258f;
  t18_n244: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t18_n246;
  t18_n245: return 0.09790767729282379f;
  t18_n246: return 0.03755580633878708f;
  t18_n247: return 0.10164259374141693f;
  t18_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t18_n250;
  t18_n249: return 0.0007574164774268866f;
  t18_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t18_n254;
  t18_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t18_n253;
  t18_n252: return 0.08881326019763947f;
  t18_n253: return 0.09730231761932373f;
  t18_n254: return 0.10551128536462784f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree19(const Example& E) {
  t19_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t19_n248;
  t19_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t19_n241;
  t19_n2: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t19_n114;
  t19_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t19_n63;
  t19_n4: if (E.getNumReferences() >= 3301085184 /*1556.5*/) goto t19_n52;
  t19_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t19_n13;
  t19_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n8;
  t19_n7: return -0.08682798594236374f;
  t19_n8: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t19_n12;
  t19_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t19_n11;
  t19_n10: return -0.048379138112068176f;
  t19_n11: return 0.02912013791501522f;
  t19_n12: return -0.04636610299348831f;
  t19_n13: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t19_n23;
  t19_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n20;
  t19_n15: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t19_n17;
  t19_n16: return -0.005774271674454212f;
  t19_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t19_n19;
  t19_n18: return 0.04237106814980507f;
  t19_n19: return -0.028414275497198105f;
  t19_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t19_n22;
  t19_n21: return 0.0515919104218483f;
  t19_n22: return 0.08911124616861343f;
  t19_n23: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t19_n47;
  t19_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t19_n26;
  t19_n25: return -0.03908855840563774f;
  t19_n26: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t19_n36;
  t19_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t19_n29;
  t19_n28: return -0.03799903765320778f;
  t19_n29: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t19_n31;
  t19_n30: return 0.004415302071720362f;
  t19_n31: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t19_n33;
  t19_n32: return -0.028351357206702232f;
  t19_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t19_n35;
  t19_n34: return -0.012778478674590588f;
  t19_n35: return 0.07462165504693985f;
  t19_n36: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t19_n42;
  t19_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t19_n41;
  t19_n38: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t19_n40;
  t19_n39: return 0.029577964916825294f;
  t19_n40: return -0.10257237404584885f;
  t19_n41: return -0.04561291262507439f;
  t19_n42: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t19_n46;
  t19_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t19_n45;
  t19_n44: return 0.06294618546962738f;
  t19_n45: return 0.09882063418626785f;
  t19_n46: return -0.13384157419204712f;
  t19_n47: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t19_n49;
  t19_n48: return -0.10050737112760544f;
  t19_n49: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t19_n51;
  t19_n50: return 0.013645577244460583f;
  t19_n51: return -0.07545478641986847f;
  t19_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t19_n56;
  t19_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t19_n55;
  t19_n54: return -0.007627623155713081f;
  t19_n55: return 0.06900252401828766f;
  t19_n56: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n60;
  t19_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n59;
  t19_n58: return 0.030377361923456192f;
  t19_n59: return 0.07750282436609268f;
  t19_n60: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t19_n62;
  t19_n61: return 0.013637538999319077f;
  t19_n62: return 0.09247134625911713f;
  t19_n63: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t19_n87;
  t19_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t19_n76;
  t19_n65: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t19_n73;
  t19_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n70;
  t19_n67: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t19_n69;
  t19_n68: return 0.04013225436210632f;
  t19_n69: return -0.0063885655254125595f;
  t19_n70: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n72;
  t19_n71: return 0.06571948528289795f;
  t19_n72: return 0.0077853430993855f;
  t19_n73: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n75;
  t19_n74: return -0.01385275274515152f;
  t19_n75: return 0.07478225231170654f;
  t19_n76: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t19_n86;
  t19_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t19_n81;
  t19_n78: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n80;
  t19_n79: return -0.004803450778126717f;
  t19_n80: return 0.07400713115930557f;
  t19_n81: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t19_n85;
  t19_n82: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n84;
  t19_n83: return 0.09184002131223679f;
  t19_n84: return -0.0219985693693161f;
  t19_n85: return 0.05722232908010483f;
  t19_n86: return 0.09793921560049057f;
  t19_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n113;
  t19_n88: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n104;
  t19_n89: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t19_n101;
  t19_n90: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n100;
  t19_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t19_n93;
  t19_n92: return -0.06314115971326828f;
  t19_n93: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t19_n99;
  t19_n94: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t19_n96;
  t19_n95: return 0.005445700138807297f;
  t19_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t19_n98;
  t19_n97: return 0.04121624678373337f;
  t19_n98: return 0.10404003411531448f;
  t19_n99: return -0.028338758274912834f;
  t19_n100: return 0.029162472113966942f;
  t19_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t19_n103;
  t19_n102: return -0.08810549229383469f;
  t19_n103: return 0.01606176421046257f;
  t19_n104: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t19_n112;
  t19_n105: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n111;
  t19_n106: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t19_n110;
  t19_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t19_n109;
  t19_n108: return 0.0022376247216016054f;
  t19_n109: return 0.040294017642736435f;
  t19_n110: return -0.08077210187911987f;
  t19_n111: return 0.07044421881437302f;
  t19_n112: return 0.07291650027036667f;
  t19_n113: return 0.02459515631198883f;
  t19_n114: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t19_n186;
  t19_n115: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t19_n141;
  t19_n116: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t19_n132;
  t19_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n119;
  t19_n118: return -0.014167491346597672f;
  t19_n119: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t19_n127;
  t19_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t19_n124;
  t19_n121: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n123;
  t19_n122: return 0.01799074560403824f;
  t19_n123: return 0.07024584710597992f;
  t19_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t19_n126;
  t19_n125: return 0.03917007893323898f;
  t19_n126: return 0.07078143209218979f;
  t19_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t19_n131;
  t19_n128: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n130;
  t19_n129: return 0.04600479453802109f;
  t19_n130: return 0.0903640165925026f;
  t19_n131: return 0.0907847210764885f;
  t19_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n138;
  t19_n133: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t19_n135;
  t19_n134: return 0.04435529187321663f;
  t19_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t19_n137;
  t19_n136: return 0.04929511249065399f;
  t19_n137: return 0.07743098586797714f;
  t19_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t19_n140;
  t19_n139: return -0.15977664291858673f;
  t19_n140: return 0.08989333361387253f;
  t19_n141: if (E.getNumReferences() >= 3306285056 /*2334.5*/) goto t19_n177;
  t19_n142: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t19_n174;
  t19_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t19_n163;
  t19_n144: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n160;
  t19_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t19_n147;
  t19_n146: return -0.060406964272260666f;
  t19_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t19_n149;
  t19_n148: return -0.021124571561813354f;
  t19_n149: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t19_n151;
  t19_n150: return 0.02193238027393818f;
  t19_n151: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t19_n159;
  t19_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n154;
  t19_n153: return 0.009504484012722969f;
  t19_n154: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t19_n158;
  t19_n155: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t19_n157;
  t19_n156: return -0.05598604306578636f;
  t19_n157: return 0.1302400678396225f;
  t19_n158: return -0.05344446375966072f;
  t19_n159: return -0.07027741521596909f;
  t19_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t19_n162;
  t19_n161: return -0.03616395220160484f;
  t19_n162: return 0.04844525083899498f;
  t19_n163: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t19_n169;
  t19_n164: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t19_n168;
  t19_n165: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t19_n167;
  t19_n166: return 0.03784359246492386f;
  t19_n167: return 0.09050175547599792f;
  t19_n168: return 0.002325026784092188f;
  t19_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t19_n173;
  t19_n170: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t19_n172;
  t19_n171: return 0.06365711241960526f;
  t19_n172: return 0.0023474683985114098f;
  t19_n173: return 0.07130961865186691f;
  t19_n174: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t19_n176;
  t19_n175: return 0.04247064143419266f;
  t19_n176: return 0.10458920896053314f;
  t19_n177: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t19_n179;
  t19_n178: return 0.047326624393463135f;
  t19_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t19_n181;
  t19_n180: return 0.06774117052555084f;
  t19_n181: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t19_n183;
  t19_n182: return 0.09510940313339233f;
  t19_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t19_n185;
  t19_n184: return 0.11148997396230698f;
  t19_n185: return 0.1188788115978241f;
  t19_n186: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t19_n200;
  t19_n187: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t19_n197;
  t19_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t19_n192;
  t19_n189: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n191;
  t19_n190: return 0.026394285261631012f;
  t19_n191: return 0.06177189573645592f;
  t19_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n196;
  t19_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n195;
  t19_n194: return 0.05059698969125748f;
  t19_n195: return 0.08456900715827942f;
  t19_n196: return 0.07470441609621048f;
  t19_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n199;
  t19_n198: return 0.07849942147731781f;
  t19_n199: return 0.0956626608967781f;
  t19_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t19_n202;
  t19_n201: return -0.07124411314725876f;
  t19_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n228;
  t19_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t19_n227;
  t19_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t19_n220;
  t19_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n217;
  t19_n206: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t19_n216;
  t19_n207: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t19_n215;
  t19_n208: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n214;
  t19_n209: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t19_n213;
  t19_n210: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t19_n212;
  t19_n211: return 0.011244315654039383f;
  t19_n212: return -0.03444498032331467f;
  t19_n213: return 0.029098059982061386f;
  t19_n214: return 0.04340164363384247f;
  t19_n215: return -0.08445042371749878f;
  t19_n216: return -0.11181472986936569f;
  t19_n217: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t19_n219;
  t19_n218: return 0.06898421794176102f;
  t19_n219: return -0.0616443008184433f;
  t19_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t19_n224;
  t19_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t19_n223;
  t19_n222: return -0.05231807753443718f;
  t19_n223: return 0.038282256573438644f;
  t19_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t19_n226;
  t19_n225: return 0.06495742499828339f;
  t19_n226: return 0.09100478142499924f;
  t19_n227: return 0.06837449967861176f;
  t19_n228: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t19_n232;
  t19_n229: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t19_n231;
  t19_n230: return -0.03999538719654083f;
  t19_n231: return 0.07601036131381989f;
  t19_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t19_n238;
  t19_n233: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t19_n235;
  t19_n234: return -0.10224807262420654f;
  t19_n235: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t19_n237;
  t19_n236: return 0.04375891014933586f;
  t19_n237: return 0.08155979961156845f;
  t19_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t19_n240;
  t19_n239: return 0.06552579998970032f;
  t19_n240: return 0.09656500071287155f;
  t19_n241: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t19_n247;
  t19_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t19_n244;
  t19_n243: return 0.02664380520582199f;
  t19_n244: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t19_n246;
  t19_n245: return 0.07736501842737198f;
  t19_n246: return 0.10133399069309235f;
  t19_n247: return 0.1004239171743393f;
  t19_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t19_n250;
  t19_n249: return 0.0005538350087590516f;
  t19_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t19_n254;
  t19_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t19_n253;
  t19_n252: return 0.08641254156827927f;
  t19_n253: return 0.09575098752975464f;
  t19_n254: return 0.10472235828638077f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree20(const Example& E) {
  t20_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t20_n250;
  t20_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t20_n243;
  t20_n2: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t20_n130;
  t20_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t20_n65;
  t20_n4: if (E.getNumReferences() >= 3294208000 /*870.5*/) goto t20_n48;
  t20_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t20_n9;
  t20_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n8;
  t20_n7: return -0.08446692675352097f;
  t20_n8: return 0.00097454380011186f;
  t20_n9: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t20_n15;
  t20_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n14;
  t20_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n13;
  t20_n12: return -0.007407733239233494f;
  t20_n13: return 0.029866216704249382f;
  t20_n14: return 0.06931301951408386f;
  t20_n15: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t20_n43;
  t20_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t20_n28;
  t20_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t20_n19;
  t20_n18: return -0.04288862273097038f;
  t20_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t20_n27;
  t20_n20: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t20_n22;
  t20_n21: return -0.05304865911602974f;
  t20_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n24;
  t20_n23: return -0.008867346681654453f;
  t20_n24: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t20_n26;
  t20_n25: return 0.09764218330383301f;
  t20_n26: return 0.11693631112575531f;
  t20_n27: return 0.0549137108027935f;
  t20_n28: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t20_n32;
  t20_n29: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t20_n31;
  t20_n30: return 0.00355226406827569f;
  t20_n31: return -0.036602653563022614f;
  t20_n32: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t20_n38;
  t20_n33: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t20_n37;
  t20_n34: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t20_n36;
  t20_n35: return 0.029248477891087532f;
  t20_n36: return -0.10091078281402588f;
  t20_n37: return -0.04293544590473175f;
  t20_n38: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t20_n42;
  t20_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t20_n41;
  t20_n40: return 0.05772420018911362f;
  t20_n41: return 0.09015408903360367f;
  t20_n42: return -0.12972143292427063f;
  t20_n43: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t20_n47;
  t20_n44: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t20_n46;
  t20_n45: return -0.08221940696239471f;
  t20_n46: return 0.02045132964849472f;
  t20_n47: return -0.07184974104166031f;
  t20_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t20_n56;
  t20_n49: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t20_n53;
  t20_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t20_n52;
  t20_n51: return 0.018698906525969505f;
  t20_n52: return 0.04623384401202202f;
  t20_n53: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t20_n55;
  t20_n54: return 0.09809751808643341f;
  t20_n55: return 0.052636682987213135f;
  t20_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t20_n64;
  t20_n57: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t20_n59;
  t20_n58: return -0.051967985928058624f;
  t20_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t20_n63;
  t20_n60: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n62;
  t20_n61: return -0.002490680431947112f;
  t20_n62: return 0.05914066731929779f;
  t20_n63: return 0.05223003774881363f;
  t20_n64: return 0.06477952003479004f;
  t20_n65: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t20_n89;
  t20_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n78;
  t20_n67: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t20_n75;
  t20_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n72;
  t20_n69: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t20_n71;
  t20_n70: return 0.03656499460339546f;
  t20_n71: return -0.0066644735634326935f;
  t20_n72: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t20_n74;
  t20_n73: return 0.06139461323618889f;
  t20_n74: return 0.007321399170905352f;
  t20_n75: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t20_n77;
  t20_n76: return -0.014286796562373638f;
  t20_n77: return 0.07164452224969864f;
  t20_n78: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t20_n88;
  t20_n79: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t20_n83;
  t20_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t20_n82;
  t20_n81: return -0.005573098547756672f;
  t20_n82: return 0.08950471878051758f;
  t20_n83: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t20_n85;
  t20_n84: return 0.05239318311214447f;
  t20_n85: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n87;
  t20_n86: return -0.02284478209912777f;
  t20_n87: return 0.07205484807491302f;
  t20_n88: return 0.0958346575498581f;
  t20_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n129;
  t20_n90: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t20_n120;
  t20_n91: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t20_n117;
  t20_n92: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t20_n94;
  t20_n93: return -0.05190487205982208f;
  t20_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t20_n110;
  t20_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n107;
  t20_n96: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n106;
  t20_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t20_n99;
  t20_n98: return -0.051081329584121704f;
  t20_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n103;
  t20_n100: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n102;
  t20_n101: return 0.012832926586270332f;
  t20_n102: return -0.1044711098074913f;
  t20_n103: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n105;
  t20_n104: return -0.14434757828712463f;
  t20_n105: return 0.00934070535004139f;
  t20_n106: return 0.02930881641805172f;
  t20_n107: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n109;
  t20_n108: return -0.038993336260318756f;
  t20_n109: return 0.060413748025894165f;
  t20_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t20_n114;
  t20_n111: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t20_n113;
  t20_n112: return 0.021879035979509354f;
  t20_n113: return 0.12797299027442932f;
  t20_n114: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t20_n116;
  t20_n115: return 0.10001148283481598f;
  t20_n116: return 0.027437085285782814f;
  t20_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n119;
  t20_n118: return -0.0850173830986023f;
  t20_n119: return 0.01434915978461504f;
  t20_n120: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t20_n128;
  t20_n121: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n127;
  t20_n122: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t20_n126;
  t20_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t20_n125;
  t20_n124: return 0.0015232909936457872f;
  t20_n125: return 0.037290751934051514f;
  t20_n126: return -0.07696015387773514f;
  t20_n127: return 0.0696488693356514f;
  t20_n128: return 0.06982062011957169f;
  t20_n129: return 0.02321135252714157f;
  t20_n130: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t20_n190;
  t20_n131: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t20_n151;
  t20_n132: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t20_n144;
  t20_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t20_n135;
  t20_n134: return -0.0126145975664258f;
  t20_n135: if (E.getFractionNameInContext() >= 3193057410 /*0.2052631676197052*/) goto t20_n137;
  t20_n136: return 0.024738823994994164f;
  t20_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t20_n143;
  t20_n138: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n142;
  t20_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t20_n141;
  t20_n140: return 0.04791749268770218f;
  t20_n141: return 0.0048164064064621925f;
  t20_n142: return 0.08252574503421783f;
  t20_n143: return 0.08207780867815018f;
  t20_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t20_n150;
  t20_n145: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t20_n147;
  t20_n146: return 0.04207991808652878f;
  t20_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t20_n149;
  t20_n148: return 0.04851240664720535f;
  t20_n149: return 0.07518533617258072f;
  t20_n150: return 0.08282148092985153f;
  t20_n151: if (E.getNumReferences() >= 3306285056 /*2334.5*/) goto t20_n185;
  t20_n152: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t20_n172;
  t20_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n169;
  t20_n154: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n168;
  t20_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t20_n157;
  t20_n156: return -0.0652155727148056f;
  t20_n157: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t20_n159;
  t20_n158: return -0.014505937695503235f;
  t20_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t20_n161;
  t20_n160: return 0.021951941773295403f;
  t20_n161: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t20_n163;
  t20_n162: return -0.039148781448602676f;
  t20_n163: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t20_n167;
  t20_n164: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t20_n166;
  t20_n165: return 0.12677302956581116f;
  t20_n166: return -0.07420652359724045f;
  t20_n167: return -0.15558141469955444f;
  t20_n168: return 0.029806705191731453f;
  t20_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t20_n171;
  t20_n170: return 0.00819002278149128f;
  t20_n171: return 0.076438307762146f;
  t20_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t20_n182;
  t20_n173: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t20_n177;
  t20_n174: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t20_n176;
  t20_n175: return 0.0050080013461411f;
  t20_n176: return 0.06769081950187683f;
  t20_n177: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t20_n179;
  t20_n178: return 0.11467179656028748f;
  t20_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t20_n181;
  t20_n180: return 0.04071382060647011f;
  t20_n181: return 0.1094680204987526f;
  t20_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t20_n184;
  t20_n183: return 0.04229774326086044f;
  t20_n184: return 0.0705614760518074f;
  t20_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t20_n187;
  t20_n186: return 0.045066237449645996f;
  t20_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t20_n189;
  t20_n188: return 0.06431018561124802f;
  t20_n189: return 0.10903210937976837f;
  t20_n190: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t20_n202;
  t20_n191: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t20_n201;
  t20_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t20_n196;
  t20_n193: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n195;
  t20_n194: return 0.023748906329274178f;
  t20_n195: return 0.059338998049497604f;
  t20_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t20_n200;
  t20_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n199;
  t20_n198: return 0.04861718788743019f;
  t20_n199: return 0.08216684311628342f;
  t20_n200: return 0.07276076823472977f;
  t20_n201: return 0.07949776202440262f;
  t20_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t20_n204;
  t20_n203: return -0.06702902913093567f;
  t20_n204: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t20_n232;
  t20_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t20_n231;
  t20_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t20_n218;
  t20_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t20_n211;
  t20_n208: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n210;
  t20_n209: return 0.03195973485708237f;
  t20_n210: return -0.10640867799520493f;
  t20_n211: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t20_n215;
  t20_n212: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t20_n214;
  t20_n213: return 0.015176868997514248f;
  t20_n214: return -0.10777612030506134f;
  t20_n215: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t20_n217;
  t20_n216: return 0.06208501756191254f;
  t20_n217: return 0.01857876218855381f;
  t20_n218: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n228;
  t20_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t20_n221;
  t20_n220: return 0.018400469794869423f;
  t20_n221: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t20_n223;
  t20_n222: return 0.030649183318018913f;
  t20_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t20_n225;
  t20_n224: return 0.03667152300477028f;
  t20_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t20_n227;
  t20_n226: return 0.09160792827606201f;
  t20_n227: return 0.0531383715569973f;
  t20_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t20_n230;
  t20_n229: return 0.06108298897743225f;
  t20_n230: return 0.08859045803546906f;
  t20_n231: return 0.06578335911035538f;
  t20_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t20_n236;
  t20_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t20_n235;
  t20_n234: return -0.04178015887737274f;
  t20_n235: return 0.0730225145816803f;
  t20_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t20_n242;
  t20_n237: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t20_n239;
  t20_n238: return -0.09311359375715256f;
  t20_n239: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t20_n241;
  t20_n240: return 0.03755008429288864f;
  t20_n241: return 0.07981447875499725f;
  t20_n242: return 0.07642342150211334f;
  t20_n243: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t20_n249;
  t20_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t20_n246;
  t20_n245: return 0.038466062396764755f;
  t20_n246: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t20_n248;
  t20_n247: return 0.09462285786867142f;
  t20_n248: return 0.0268978513777256f;
  t20_n249: return 0.09923490136861801f;
  t20_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t20_n252;
  t20_n251: return 0.000464608718175441f;
  t20_n252: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t20_n254;
  t20_n253: return 0.08772865682840347f;
  t20_n254: return 0.09641122817993164f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree21(const Example& E) {
  t21_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t21_n248;
  t21_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t21_n241;
  t21_n2: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t21_n108;
  t21_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t21_n61;
  t21_n4: if (E.getNumReferences() >= 3303288832 /*1825.5*/) goto t21_n50;
  t21_n5: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t21_n13;
  t21_n6: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t21_n10;
  t21_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t21_n9;
  t21_n8: return -0.017660589888691902f;
  t21_n9: return 0.04810502380132675f;
  t21_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n12;
  t21_n11: return 0.03256858140230179f;
  t21_n12: return 0.0893988385796547f;
  t21_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t21_n39;
  t21_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t21_n34;
  t21_n15: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t21_n33;
  t21_n16: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n32;
  t21_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t21_n19;
  t21_n18: return -0.0852886214852333f;
  t21_n19: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t21_n21;
  t21_n20: return -0.04189859330654144f;
  t21_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n23;
  t21_n22: return -0.03138996288180351f;
  t21_n23: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t21_n29;
  t21_n24: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t21_n28;
  t21_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t21_n27;
  t21_n26: return 0.04107252508401871f;
  t21_n27: return -0.13073670864105225f;
  t21_n28: return -0.09258755296468735f;
  t21_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t21_n31;
  t21_n30: return 0.056583043187856674f;
  t21_n31: return 0.09075275808572769f;
  t21_n32: return 0.0037409812211990356f;
  t21_n33: return -0.13122060894966125f;
  t21_n34: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t21_n36;
  t21_n35: return -0.11074119806289673f;
  t21_n36: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t21_n38;
  t21_n37: return -0.003941063303500414f;
  t21_n38: return -0.08654595911502838f;
  t21_n39: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t21_n41;
  t21_n40: return -0.012079334817826748f;
  t21_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t21_n43;
  t21_n42: return -0.13587448000907898f;
  t21_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t21_n45;
  t21_n44: return -0.017871158197522163f;
  t21_n45: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t21_n47;
  t21_n46: return -0.03711516037583351f;
  t21_n47: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t21_n49;
  t21_n48: return 0.06018409505486488f;
  t21_n49: return -0.10311909765005112f;
  t21_n50: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t21_n58;
  t21_n51: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t21_n53;
  t21_n52: return 0.0328422449529171f;
  t21_n53: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t21_n57;
  t21_n54: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n56;
  t21_n55: return 0.08798015117645264f;
  t21_n56: return 0.1320185661315918f;
  t21_n57: return 0.05107228457927704f;
  t21_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t21_n60;
  t21_n59: return -0.004882383160293102f;
  t21_n60: return 0.0621512345969677f;
  t21_n61: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t21_n77;
  t21_n62: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t21_n70;
  t21_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n67;
  t21_n64: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t21_n66;
  t21_n65: return 0.026404617354273796f;
  t21_n66: return -0.05003299191594124f;
  t21_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t21_n69;
  t21_n68: return 0.0875263437628746f;
  t21_n69: return 0.03027326613664627f;
  t21_n70: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t21_n74;
  t21_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n73;
  t21_n72: return 0.049134548753499985f;
  t21_n73: return 0.07478822022676468f;
  t21_n74: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t21_n76;
  t21_n75: return 0.1207113042473793f;
  t21_n76: return 0.08426836133003235f;
  t21_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t21_n95;
  t21_n78: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n92;
  t21_n79: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t21_n91;
  t21_n80: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t21_n88;
  t21_n81: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t21_n85;
  t21_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t21_n84;
  t21_n83: return -0.060715582221746445f;
  t21_n84: return 0.0038725873455405235f;
  t21_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n87;
  t21_n86: return -0.06607094407081604f;
  t21_n87: return 0.026456739753484726f;
  t21_n88: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t21_n90;
  t21_n89: return 0.002559500513598323f;
  t21_n90: return 0.06194464862346649f;
  t21_n91: return -0.05008108541369438f;
  t21_n92: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t21_n94;
  t21_n93: return 0.017606602981686592f;
  t21_n94: return 0.0705142691731453f;
  t21_n95: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t21_n103;
  t21_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n102;
  t21_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t21_n99;
  t21_n98: return -0.019845962524414062f;
  t21_n99: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t21_n101;
  t21_n100: return 0.013994164764881134f;
  t21_n101: return 0.09776913374662399f;
  t21_n102: return 0.036195699125528336f;
  t21_n103: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t21_n107;
  t21_n104: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t21_n106;
  t21_n105: return 0.09925658255815506f;
  t21_n106: return 0.075357586145401f;
  t21_n107: return 0.024015840142965317f;
  t21_n108: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t21_n166;
  t21_n109: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t21_n157;
  t21_n110: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t21_n124;
  t21_n111: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t21_n119;
  t21_n112: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t21_n118;
  t21_n113: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n117;
  t21_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t21_n116;
  t21_n115: return 0.011326676234602928f;
  t21_n116: return 0.06660959124565125f;
  t21_n117: return 0.05186133459210396f;
  t21_n118: return 0.04757707938551903f;
  t21_n119: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t21_n123;
  t21_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n122;
  t21_n121: return 0.052768297493457794f;
  t21_n122: return 0.09037275612354279f;
  t21_n123: return 0.08538372069597244f;
  t21_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t21_n146;
  t21_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t21_n129;
  t21_n126: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n128;
  t21_n127: return -0.0556110180914402f;
  t21_n128: return 0.027215968817472458f;
  t21_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n131;
  t21_n130: return -0.00507610896602273f;
  t21_n131: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t21_n135;
  t21_n132: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n134;
  t21_n133: return 0.006369420327246189f;
  t21_n134: return 0.0593663789331913f;
  t21_n135: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t21_n145;
  t21_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n138;
  t21_n137: return 0.013723080977797508f;
  t21_n138: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t21_n144;
  t21_n139: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t21_n141;
  t21_n140: return -0.011388116516172886f;
  t21_n141: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n143;
  t21_n142: return 0.1187608391046524f;
  t21_n143: return 0.027430858463048935f;
  t21_n144: return -0.06163299083709717f;
  t21_n145: return -0.06084693223237991f;
  t21_n146: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n152;
  t21_n147: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t21_n149;
  t21_n148: return 0.006135711912065744f;
  t21_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t21_n151;
  t21_n150: return 0.035351768136024475f;
  t21_n151: return 0.061584826558828354f;
  t21_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement))) goto t21_n156;
  t21_n153: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t21_n155;
  t21_n154: return 0.038981880992650986f;
  t21_n155: return 0.07669363170862198f;
  t21_n156: return 0.07439883053302765f;
  t21_n157: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n163;
  t21_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t21_n160;
  t21_n159: return 0.0382685512304306f;
  t21_n160: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t21_n162;
  t21_n161: return 0.06035589054226875f;
  t21_n162: return 0.08764096349477768f;
  t21_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t21_n165;
  t21_n164: return 0.06254567205905914f;
  t21_n165: return 0.10777954012155533f;
  t21_n166: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t21_n182;
  t21_n167: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t21_n179;
  t21_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t21_n174;
  t21_n169: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n173;
  t21_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n172;
  t21_n171: return -0.03798958286643028f;
  t21_n172: return 0.03351264074444771f;
  t21_n173: return 0.05793866887688637f;
  t21_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n178;
  t21_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t21_n177;
  t21_n176: return 0.04589905962347984f;
  t21_n177: return 0.07961735129356384f;
  t21_n178: return 0.0695846751332283f;
  t21_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n181;
  t21_n180: return 0.07373574376106262f;
  t21_n181: return 0.09342595189809799f;
  t21_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t21_n184;
  t21_n183: return -0.0630820021033287f;
  t21_n184: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n226;
  t21_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t21_n225;
  t21_n186: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n210;
  t21_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n207;
  t21_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t21_n198;
  t21_n189: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t21_n195;
  t21_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n194;
  t21_n191: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t21_n193;
  t21_n192: return 0.025784090161323547f;
  t21_n193: return -0.10833871364593506f;
  t21_n194: return 0.010522305965423584f;
  t21_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n197;
  t21_n196: return -0.08941081166267395f;
  t21_n197: return 0.019006893038749695f;
  t21_n198: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t21_n206;
  t21_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t21_n201;
  t21_n200: return 0.005262511782348156f;
  t21_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t21_n203;
  t21_n202: return 0.01171373575925827f;
  t21_n203: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t21_n205;
  t21_n204: return 0.03103834018111229f;
  t21_n205: return 0.0881718173623085f;
  t21_n206: return 0.0559178926050663f;
  t21_n207: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t21_n209;
  t21_n208: return 0.04902995377779007f;
  t21_n209: return -0.05985572561621666f;
  t21_n210: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t21_n216;
  t21_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t21_n213;
  t21_n212: return 0.03487525135278702f;
  t21_n213: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t21_n215;
  t21_n214: return 0.04582371190190315f;
  t21_n215: return 0.08009430021047592f;
  t21_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t21_n218;
  t21_n217: return -0.10312946140766144f;
  t21_n218: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t21_n220;
  t21_n219: return -0.0709339901804924f;
  t21_n220: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t21_n224;
  t21_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t21_n223;
  t21_n222: return -0.031155699864029884f;
  t21_n223: return 0.07745353877544403f;
  t21_n224: return -0.0016023337375372648f;
  t21_n225: return 0.06224945932626724f;
  t21_n226: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t21_n230;
  t21_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t21_n229;
  t21_n228: return -0.04016260802745819f;
  t21_n229: return 0.0707937628030777f;
  t21_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t21_n236;
  t21_n231: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t21_n233;
  t21_n232: return -0.08701765537261963f;
  t21_n233: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t21_n235;
  t21_n234: return 0.037114132195711136f;
  t21_n235: return 0.07737923413515091f;
  t21_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t21_n238;
  t21_n237: return 0.059987083077430725f;
  t21_n238: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t21_n240;
  t21_n239: return 0.09281321614980698f;
  t21_n240: return -0.10566782206296921f;
  t21_n241: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t21_n247;
  t21_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t21_n246;
  t21_n243: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t21_n245;
  t21_n244: return 0.09765352308750153f;
  t21_n245: return 0.023405618965625763f;
  t21_n246: return 0.09649825096130371f;
  t21_n247: return 0.09792269766330719f;
  t21_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t21_n254;
  t21_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t21_n251;
  t21_n250: return -0.00020703079644590616f;
  t21_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t21_n253;
  t21_n252: return 0.08144746720790863f;
  t21_n253: return 0.0923483818769455f;
  t21_n254: return 0.10317377001047134f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree22(const Example& E) {
  t22_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t22_n252;
  t22_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t22_n111;
  t22_n2: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t22_n110;
  t22_n3: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t22_n85;
  t22_n4: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t22_n44;
  t22_n5: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t22_n15;
  t22_n6: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t22_n12;
  t22_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t22_n9;
  t22_n8: return -0.039822421967983246f;
  t22_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t22_n11;
  t22_n10: return 0.025359272956848145f;
  t22_n11: return -0.011726466938853264f;
  t22_n12: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t22_n14;
  t22_n13: return 0.03598053753376007f;
  t22_n14: return 0.008059509098529816f;
  t22_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t22_n41;
  t22_n16: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t22_n18;
  t22_n17: return -0.08492957055568695f;
  t22_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n38;
  t22_n19: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t22_n25;
  t22_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t22_n22;
  t22_n21: return -0.06562621891498566f;
  t22_n22: if (E.getNumReferences() >= 3268608000 /*105.5*/) goto t22_n24;
  t22_n23: return -0.01498045027256012f;
  t22_n24: return 0.06268271058797836f;
  t22_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t22_n27;
  t22_n26: return -0.07605123519897461f;
  t22_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t22_n33;
  t22_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t22_n32;
  t22_n29: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t22_n31;
  t22_n30: return 0.025273969396948814f;
  t22_n31: return -0.13237541913986206f;
  t22_n32: return -0.09356242418289185f;
  t22_n33: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t22_n37;
  t22_n34: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t22_n36;
  t22_n35: return 0.06796815246343613f;
  t22_n36: return -0.043968044221401215f;
  t22_n37: return -0.12704205513000488f;
  t22_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t22_n40;
  t22_n39: return -0.08898263424634933f;
  t22_n40: return 0.03700250759720802f;
  t22_n41: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t22_n43;
  t22_n42: return 4.428053580340929e-05f;
  t22_n43: return -0.050931207835674286f;
  t22_n44: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t22_n62;
  t22_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t22_n59;
  t22_n46: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t22_n56;
  t22_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t22_n55;
  t22_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t22_n54;
  t22_n49: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t22_n51;
  t22_n50: return 0.04536445066332817f;
  t22_n51: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n53;
  t22_n52: return -0.02131679654121399f;
  t22_n53: return 0.054524026811122894f;
  t22_n54: return -0.014721457846462727f;
  t22_n55: return 0.04267771914601326f;
  t22_n56: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t22_n58;
  t22_n57: return -0.023372258991003036f;
  t22_n58: return 0.06468883901834488f;
  t22_n59: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t22_n61;
  t22_n60: return 0.08837075531482697f;
  t22_n61: return 0.033792346715927124f;
  t22_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t22_n64;
  t22_n63: return -0.018502069637179375f;
  t22_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t22_n72;
  t22_n65: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t22_n67;
  t22_n66: return -0.015438049100339413f;
  t22_n67: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n71;
  t22_n68: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t22_n70;
  t22_n69: return 0.03301067277789116f;
  t22_n70: return -0.047911666333675385f;
  t22_n71: return 0.08380847424268723f;
  t22_n72: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t22_n84;
  t22_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t22_n77;
  t22_n74: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n76;
  t22_n75: return -0.040663834661245346f;
  t22_n76: return 0.042868565768003464f;
  t22_n77: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t22_n79;
  t22_n78: return 0.02583738975226879f;
  t22_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t22_n83;
  t22_n80: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t22_n82;
  t22_n81: return 0.09285453706979752f;
  t22_n82: return 0.07189871370792389f;
  t22_n83: return -0.0015539037995040417f;
  t22_n84: return 0.06820765882730484f;
  t22_n85: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t22_n93;
  t22_n86: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t22_n88;
  t22_n87: return 0.04412587732076645f;
  t22_n88: if (E.getNumReferences() >= 3361959232 /*232885.0*/) goto t22_n92;
  t22_n89: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n91;
  t22_n90: return 0.09050342440605164f;
  t22_n91: return 0.13227252662181854f;
  t22_n92: return 0.07205983251333237f;
  t22_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t22_n107;
  t22_n94: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t22_n106;
  t22_n95: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t22_n97;
  t22_n96: return 0.009341474622488022f;
  t22_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t22_n103;
  t22_n98: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t22_n102;
  t22_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t22_n101;
  t22_n100: return -0.13732948899269104f;
  t22_n101: return 0.09972120821475983f;
  t22_n102: return 0.027830373495817184f;
  t22_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t22_n105;
  t22_n104: return -0.16001495718955994f;
  t22_n105: return 0.12360203266143799f;
  t22_n106: return -0.04293643683195114f;
  t22_n107: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t22_n109;
  t22_n108: return 0.019297409802675247f;
  t22_n109: return 0.08325894922018051f;
  t22_n110: return 0.090275339782238f;
  t22_n111: if (E.getNumReferences() >= 3326314496 /*12514.0*/) goto t22_n239;
  t22_n112: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t22_n172;
  t22_n113: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t22_n165;
  t22_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t22_n130;
  t22_n115: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t22_n127;
  t22_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t22_n122;
  t22_n117: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t22_n121;
  t22_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n120;
  t22_n119: return 0.024193083867430687f;
  t22_n120: return 0.07236923277378082f;
  t22_n121: return -0.005528525914996862f;
  t22_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t22_n126;
  t22_n123: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t22_n125;
  t22_n124: return 0.01674993522465229f;
  t22_n125: return 0.07396739721298218f;
  t22_n126: return 0.0648023784160614f;
  t22_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t22_n129;
  t22_n128: return 0.05316375568509102f;
  t22_n129: return 0.08152072876691818f;
  t22_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t22_n136;
  t22_n131: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n133;
  t22_n132: return -0.050565704703330994f;
  t22_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t22_n135;
  t22_n134: return -0.03524129092693329f;
  t22_n135: return 0.04407075420022011f;
  t22_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t22_n158;
  t22_n137: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t22_n153;
  t22_n138: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t22_n140;
  t22_n139: return -0.012651240453124046f;
  t22_n140: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t22_n152;
  t22_n141: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t22_n151;
  t22_n142: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t22_n144;
  t22_n143: return 0.011491547338664532f;
  t22_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t22_n150;
  t22_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t22_n147;
  t22_n146: return -0.09656865894794464f;
  t22_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t22_n149;
  t22_n148: return 0.11515773087739944f;
  t22_n149: return 0.11253242939710617f;
  t22_n150: return -0.033081088215112686f;
  t22_n151: return -0.14245766401290894f;
  t22_n152: return -0.08601906895637512f;
  t22_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t22_n155;
  t22_n154: return 0.02104014903306961f;
  t22_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t22_n157;
  t22_n156: return 0.0359388068318367f;
  t22_n157: return 0.0651618093252182f;
  t22_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t22_n162;
  t22_n159: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t22_n161;
  t22_n160: return -0.029158124700188637f;
  t22_n161: return 0.042659319937229156f;
  t22_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t22_n164;
  t22_n163: return -0.008541225455701351f;
  t22_n164: return 0.07240114361047745f;
  t22_n165: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t22_n171;
  t22_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t22_n170;
  t22_n167: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n169;
  t22_n168: return 0.03532523289322853f;
  t22_n169: return 0.07741449028253555f;
  t22_n170: return 0.07431796193122864f;
  t22_n171: return 0.07664944231510162f;
  t22_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t22_n196;
  t22_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n191;
  t22_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t22_n176;
  t22_n175: return -0.057857438921928406f;
  t22_n176: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t22_n186;
  t22_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t22_n183;
  t22_n178: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t22_n182;
  t22_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t22_n181;
  t22_n180: return 0.037165217101573944f;
  t22_n181: return -0.03865008428692818f;
  t22_n182: return -0.10390646755695343f;
  t22_n183: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t22_n185;
  t22_n184: return 0.013178643770515919f;
  t22_n185: return -0.09295402467250824f;
  t22_n186: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t22_n190;
  t22_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Statement))) goto t22_n189;
  t22_n188: return 0.03755037486553192f;
  t22_n189: return 0.06751663982868195f;
  t22_n190: return 0.01613052375614643f;
  t22_n191: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t22_n195;
  t22_n192: if (E.getNumReferences() >= 3315746816 /*5193.0*/) goto t22_n194;
  t22_n193: return 0.04075400531291962f;
  t22_n194: return 0.1343366503715515f;
  t22_n195: return 0.06828513741493225f;
  t22_n196: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t22_n232;
  t22_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t22_n209;
  t22_n198: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t22_n208;
  t22_n199: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t22_n201;
  t22_n200: return 0.04738261178135872f;
  t22_n201: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t22_n205;
  t22_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t22_n204;
  t22_n203: return -0.013774477876722813f;
  t22_n204: return 0.04819831997156143f;
  t22_n205: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n207;
  t22_n206: return 0.04972014203667641f;
  t22_n207: return 0.09827175736427307f;
  t22_n208: return 0.0583675391972065f;
  t22_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t22_n231;
  t22_n210: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t22_n230;
  t22_n211: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t22_n229;
  t22_n212: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t22_n218;
  t22_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t22_n217;
  t22_n214: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t22_n216;
  t22_n215: return 0.01655348762869835f;
  t22_n216: return 0.06383705139160156f;
  t22_n217: return 0.0766592025756836f;
  t22_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t22_n226;
  t22_n219: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t22_n225;
  t22_n220: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t22_n222;
  t22_n221: return -0.0044402084313333035f;
  t22_n222: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t22_n224;
  t22_n223: return -0.0016367328353226185f;
  t22_n224: return 0.08321958035230637f;
  t22_n225: return 0.08889246731996536f;
  t22_n226: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t22_n228;
  t22_n227: return 0.06799590587615967f;
  t22_n228: return 0.08665550500154495f;
  t22_n229: return 0.08114530891180038f;
  t22_n230: return 0.08481054753065109f;
  t22_n231: return 0.06889069080352783f;
  t22_n232: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t22_n234;
  t22_n233: return -0.08651486039161682f;
  t22_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t22_n236;
  t22_n235: return -0.09247734397649765f;
  t22_n236: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t22_n238;
  t22_n237: return 0.029512256383895874f;
  t22_n238: return 0.0684371143579483f;
  t22_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t22_n251;
  t22_n240: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t22_n250;
  t22_n241: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t22_n243;
  t22_n242: return 0.06811776757240295f;
  t22_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t22_n245;
  t22_n244: return 0.08835053443908691f;
  t22_n245: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t22_n247;
  t22_n246: return 0.09274859726428986f;
  t22_n247: if (E.getNumReferences() >= 3349544768 /*84990.5*/) goto t22_n249;
  t22_n248: return 0.11544239521026611f;
  t22_n249: return 0.10740295052528381f;
  t22_n250: return 0.01823018305003643f;
  t22_n251: return 0.09711265563964844f;
  t22_n252: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t22_n254;
  t22_n253: return 0.08272507786750793f;
  t22_n254: return 0.09312154352664948f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree23(const Example& E) {
  t23_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t23_n252;
  t23_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t23_n131;
  t23_n2: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t23_n94;
  t23_n3: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t23_n57;
  t23_n4: if (E.getNumReferences() >= 3289473024 /*581.5*/) goto t23_n50;
  t23_n5: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t23_n13;
  t23_n6: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t23_n10;
  t23_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t23_n9;
  t23_n8: return -0.015163918025791645f;
  t23_n9: return 0.044111788272857666f;
  t23_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n12;
  t23_n11: return 0.029083790257573128f;
  t23_n12: return 0.08392807096242905f;
  t23_n13: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t23_n49;
  t23_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t23_n18;
  t23_n15: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n17;
  t23_n16: return -0.08171065151691437f;
  t23_n17: return 0.0014842142118141055f;
  t23_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t23_n20;
  t23_n19: return -0.03167363256216049f;
  t23_n20: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t23_n26;
  t23_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t23_n23;
  t23_n22: return -0.043290577828884125f;
  t23_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n25;
  t23_n24: return -0.017333853989839554f;
  t23_n25: return 0.01772993430495262f;
  t23_n26: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t23_n48;
  t23_n27: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t23_n29;
  t23_n28: return 0.03146202489733696f;
  t23_n29: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t23_n41;
  t23_n30: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t23_n32;
  t23_n31: return -0.022962380200624466f;
  t23_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n34;
  t23_n33: return -0.03708374500274658f;
  t23_n34: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t23_n36;
  t23_n35: return 0.08569660782814026f;
  t23_n36: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t23_n38;
  t23_n37: return -0.1997450739145279f;
  t23_n38: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t23_n40;
  t23_n39: return 0.08334456384181976f;
  t23_n40: return 0.03100009448826313f;
  t23_n41: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t23_n43;
  t23_n42: return 0.006435176823288202f;
  t23_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t23_n45;
  t23_n44: return -0.08190710842609406f;
  t23_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t23_n47;
  t23_n46: return 0.045556604862213135f;
  t23_n47: return 0.07606726884841919f;
  t23_n48: return -0.10938204824924469f;
  t23_n49: return -0.07592273503541946f;
  t23_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n54;
  t23_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t23_n53;
  t23_n52: return -0.010733982548117638f;
  t23_n53: return 0.056104764342308044f;
  t23_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t23_n56;
  t23_n55: return 0.02162468060851097f;
  t23_n56: return 0.06922274827957153f;
  t23_n57: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t23_n73;
  t23_n58: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t23_n68;
  t23_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n65;
  t23_n60: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t23_n62;
  t23_n61: return 0.004607690032571554f;
  t23_n62: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t23_n64;
  t23_n63: return 0.019056949764490128f;
  t23_n64: return 0.06292029470205307f;
  t23_n65: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t23_n67;
  t23_n66: return 0.08231966942548752f;
  t23_n67: return 0.02403952367603779f;
  t23_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n70;
  t23_n69: return 0.03724489361047745f;
  t23_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t23_n72;
  t23_n71: return 0.0695817619562149f;
  t23_n72: return 0.022720307111740112f;
  t23_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t23_n93;
  t23_n74: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n86;
  t23_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n85;
  t23_n76: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t23_n82;
  t23_n77: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t23_n79;
  t23_n78: return -0.013753273524343967f;
  t23_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n81;
  t23_n80: return -0.08316756039857864f;
  t23_n81: return 0.008341067470610142f;
  t23_n82: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t23_n84;
  t23_n83: return 0.005011792294681072f;
  t23_n84: return 0.05805373936891556f;
  t23_n85: return 0.03126411512494087f;
  t23_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t23_n90;
  t23_n87: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t23_n89;
  t23_n88: return 0.020519107580184937f;
  t23_n89: return -0.046285390853881836f;
  t23_n90: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t23_n92;
  t23_n91: return -0.006130884867161512f;
  t23_n92: return 0.05735174939036369f;
  t23_n93: return 0.020877689123153687f;
  t23_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t23_n110;
  t23_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t23_n97;
  t23_n96: return -0.03093758411705494f;
  t23_n97: if (E.getNumReferences() >= 3363896688 /*264171.5*/) goto t23_n103;
  t23_n98: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t23_n102;
  t23_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t23_n101;
  t23_n100: return 0.015854839235544205f;
  t23_n101: return 0.09744954109191895f;
  t23_n102: return -0.061177629977464676f;
  t23_n103: if (E.getNumReferences() >= 3364180016 /*273025.5*/) goto t23_n109;
  t23_n104: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t23_n106;
  t23_n105: return 0.04856429994106293f;
  t23_n106: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t23_n108;
  t23_n107: return 0.08128876239061356f;
  t23_n108: return 0.11204017698764801f;
  t23_n109: return 0.08293403685092926f;
  t23_n110: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t23_n120;
  t23_n111: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n119;
  t23_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t23_n116;
  t23_n113: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t23_n115;
  t23_n114: return 0.08814340829849243f;
  t23_n115: return -0.13570636510849f;
  t23_n116: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t23_n118;
  t23_n117: return 0.08932498842477798f;
  t23_n118: return 0.13540054857730865f;
  t23_n119: return 0.13015133142471313f;
  t23_n120: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t23_n130;
  t23_n121: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t23_n127;
  t23_n122: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t23_n124;
  t23_n123: return -0.0023750122636556625f;
  t23_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t23_n126;
  t23_n125: return 0.09164793789386749f;
  t23_n126: return 0.12941700220108032f;
  t23_n127: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t23_n129;
  t23_n128: return -0.11321079730987549f;
  t23_n129: return 0.0616714246571064f;
  t23_n130: return 0.09983295947313309f;
  t23_n131: if (E.getNumReferences() >= 3320626176 /*7575.5*/) goto t23_n237;
  t23_n132: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t23_n182;
  t23_n133: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t23_n147;
  t23_n134: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t23_n144;
  t23_n135: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t23_n143;
  t23_n136: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n142;
  t23_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t23_n139;
  t23_n138: return 0.013387382961809635f;
  t23_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t23_n141;
  t23_n140: return 0.04259112477302551f;
  t23_n141: return 0.10189071297645569f;
  t23_n142: return 0.05100608617067337f;
  t23_n143: return 0.04301051050424576f;
  t23_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t23_n146;
  t23_n145: return 0.06714751571416855f;
  t23_n146: return 0.08277486264705658f;
  t23_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t23_n151;
  t23_n148: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n150;
  t23_n149: return -0.0476553700864315f;
  t23_n150: return 0.025023505091667175f;
  t23_n151: if (E.getFileProximityDistanceCost() >= 3255173120 /*33.5*/) goto t23_n179;
  t23_n152: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t23_n176;
  t23_n153: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t23_n159;
  t23_n154: if (E.getNumReferences() >= 3278405632 /*232.5*/) goto t23_n156;
  t23_n155: return -0.007588032633066177f;
  t23_n156: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t23_n158;
  t23_n157: return 0.07571226358413696f;
  t23_n158: return 0.02212350070476532f;
  t23_n159: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t23_n161;
  t23_n160: return -0.014698009006679058f;
  t23_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t23_n169;
  t23_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t23_n168;
  t23_n163: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t23_n167;
  t23_n164: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t23_n166;
  t23_n165: return 0.08841720223426819f;
  t23_n166: return -0.04706868529319763f;
  t23_n167: return -0.1415119767189026f;
  t23_n168: return -0.04361078888177872f;
  t23_n169: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t23_n175;
  t23_n170: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t23_n174;
  t23_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t23_n173;
  t23_n172: return -0.0792955830693245f;
  t23_n173: return 0.10404191166162491f;
  t23_n174: return -0.039785128086805344f;
  t23_n175: return -0.13794420659542084f;
  t23_n176: if (E.getNumReferences() >= 3284254720 /*387.5*/) goto t23_n178;
  t23_n177: return 0.03454116731882095f;
  t23_n178: return 0.08126267790794373f;
  t23_n179: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t23_n181;
  t23_n180: return 0.03493443876504898f;
  t23_n181: return 0.06922168284654617f;
  t23_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t23_n204;
  t23_n183: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n199;
  t23_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t23_n186;
  t23_n185: return -0.053795844316482544f;
  t23_n186: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t23_n194;
  t23_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n191;
  t23_n188: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t23_n190;
  t23_n189: return 0.002416216768324375f;
  t23_n190: return -0.09857532382011414f;
  t23_n191: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t23_n193;
  t23_n192: return 0.012297623790800571f;
  t23_n193: return -0.08739450573921204f;
  t23_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t23_n198;
  t23_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Statement))) goto t23_n197;
  t23_n196: return 0.0369831807911396f;
  t23_n197: return 0.06341297924518585f;
  t23_n198: return 0.017075568437576294f;
  t23_n199: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t23_n203;
  t23_n200: if (E.getNumReferences() >= 3315746816 /*5193.0*/) goto t23_n202;
  t23_n201: return 0.037340447306632996f;
  t23_n202: return 0.1310865432024002f;
  t23_n203: return 0.06586812436580658f;
  t23_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t23_n230;
  t23_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t23_n229;
  t23_n206: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t23_n226;
  t23_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t23_n209;
  t23_n208: return -0.11006338149309158f;
  t23_n209: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t23_n217;
  t23_n210: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t23_n216;
  t23_n211: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t23_n215;
  t23_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t23_n214;
  t23_n213: return -0.009917452931404114f;
  t23_n214: return 0.03406090661883354f;
  t23_n215: return 0.04354720190167427f;
  t23_n216: return 0.07571464776992798f;
  t23_n217: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t23_n221;
  t23_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t23_n220;
  t23_n219: return 0.032060544937849045f;
  t23_n220: return 0.07570477575063705f;
  t23_n221: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t23_n225;
  t23_n222: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t23_n224;
  t23_n223: return 0.06619442254304886f;
  t23_n224: return -0.037024322897195816f;
  t23_n225: return 0.08888519555330276f;
  t23_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t23_n228;
  t23_n227: return 0.04848765581846237f;
  t23_n228: return 0.0696754977107048f;
  t23_n229: return 0.06619539111852646f;
  t23_n230: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t23_n232;
  t23_n231: return -0.08138225227594376f;
  t23_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t23_n234;
  t23_n233: return -0.08989682793617249f;
  t23_n234: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t23_n236;
  t23_n235: return 0.026207393035292625f;
  t23_n236: return 0.06650588661432266f;
  t23_n237: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t23_n251;
  t23_n238: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t23_n248;
  t23_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t23_n243;
  t23_n240: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t23_n242;
  t23_n241: return 0.05002440884709358f;
  t23_n242: return 0.08616528660058975f;
  t23_n243: if (E.getNumReferences() >= 3348085888 /*73593.0*/) goto t23_n245;
  t23_n244: return 0.06938593834638596f;
  t23_n245: if (E.getNumReferences() >= 3349544768 /*84990.5*/) goto t23_n247;
  t23_n246: return 0.11141793429851532f;
  t23_n247: return 0.0896790474653244f;
  t23_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t23_n250;
  t23_n249: return 0.007308934349566698f;
  t23_n250: return 0.08896294236183167f;
  t23_n251: return 0.09479282051324844f;
  t23_n252: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t23_n254;
  t23_n253: return 0.08007028698921204f;
  t23_n254: return 0.09135644137859344f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree24(const Example& E) {
  t24_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t24_n250;
  t24_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t24_n131;
  t24_n2: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t24_n110;
  t24_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t24_n63;
  t24_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t24_n34;
  t24_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n25;
  t24_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n16;
  t24_n7: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t24_n15;
  t24_n8: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t24_n14;
  t24_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n11;
  t24_n10: return -0.1274133175611496f;
  t24_n11: if (E.getNumReferences() >= 3308015616 /*2757.0*/) goto t24_n13;
  t24_n12: return -0.05180923268198967f;
  t24_n13: return 0.01934683695435524f;
  t24_n14: return 0.0025255673099309206f;
  t24_n15: return -0.07652298361063004f;
  t24_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t24_n18;
  t24_n17: return -0.10098420083522797f;
  t24_n18: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t24_n20;
  t24_n19: return 0.004954515025019646f;
  t24_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t24_n24;
  t24_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t24_n23;
  t24_n22: return 0.0007864386425353587f;
  t24_n23: return 0.048143502324819565f;
  t24_n24: return 0.059236474335193634f;
  t24_n25: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t24_n29;
  t24_n26: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t24_n28;
  t24_n27: return 0.017048699781298637f;
  t24_n28: return 0.06257680058479309f;
  t24_n29: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t24_n31;
  t24_n30: return 0.13223616778850555f;
  t24_n31: if (E.getNumReferences() >= 3343006592 /*49723.5*/) goto t24_n33;
  t24_n32: return 0.08089102059602737f;
  t24_n33: return -0.13882485032081604f;
  t24_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n46;
  t24_n35: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t24_n43;
  t24_n36: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t24_n42;
  t24_n37: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t24_n41;
  t24_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n40;
  t24_n39: return -0.00470251077786088f;
  t24_n40: return -0.05716763809323311f;
  t24_n41: return 0.0093355318531394f;
  t24_n42: return -0.11906106770038605f;
  t24_n43: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n45;
  t24_n44: return 0.03959982097148895f;
  t24_n45: return -0.01148206740617752f;
  t24_n46: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t24_n54;
  t24_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t24_n49;
  t24_n48: return 0.027624404057860374f;
  t24_n49: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t24_n51;
  t24_n50: return 0.04586068168282509f;
  t24_n51: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t24_n53;
  t24_n52: return 0.09117156267166138f;
  t24_n53: return -0.17214460670948029f;
  t24_n54: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t24_n62;
  t24_n55: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t24_n61;
  t24_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t24_n60;
  t24_n57: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t24_n59;
  t24_n58: return 0.015615345910191536f;
  t24_n59: return 0.06180159002542496f;
  t24_n60: return 0.029589621350169182f;
  t24_n61: return 0.09589429199695587f;
  t24_n62: return -0.10800918191671371f;
  t24_n63: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t24_n105;
  t24_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t24_n92;
  t24_n65: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n89;
  t24_n66: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t24_n88;
  t24_n67: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t24_n87;
  t24_n68: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t24_n78;
  t24_n69: if (E.getNumReferences() >= 3295354880 /*940.5*/) goto t24_n77;
  t24_n70: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t24_n74;
  t24_n71: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t24_n73;
  t24_n72: return -0.024015512317419052f;
  t24_n73: return -0.07375454157590866f;
  t24_n74: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t24_n76;
  t24_n75: return -0.004346480593085289f;
  t24_n76: return -0.08970794081687927f;
  t24_n77: return 0.005134996492415667f;
  t24_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t24_n80;
  t24_n79: return 0.0011257156729698181f;
  t24_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n82;
  t24_n81: return -0.04201535880565643f;
  t24_n82: if (E.getNumReferences() >= 3316109312 /*5370.0*/) goto t24_n86;
  t24_n83: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n85;
  t24_n84: return 0.07281772792339325f;
  t24_n85: return 0.023397235199809074f;
  t24_n86: return -0.012499156408011913f;
  t24_n87: return -0.08928748965263367f;
  t24_n88: return -0.081760473549366f;
  t24_n89: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t24_n91;
  t24_n90: return -0.04138358682394028f;
  t24_n91: return 0.025786008685827255f;
  t24_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t24_n98;
  t24_n93: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t24_n97;
  t24_n94: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t24_n96;
  t24_n95: return 0.07791799306869507f;
  t24_n96: return -0.03218597173690796f;
  t24_n97: return 0.046262115240097046f;
  t24_n98: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t24_n104;
  t24_n99: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t24_n103;
  t24_n100: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t24_n102;
  t24_n101: return 0.0905221700668335f;
  t24_n102: return 0.10901685059070587f;
  t24_n103: return 0.04626083746552467f;
  t24_n104: return 0.04501616954803467f;
  t24_n105: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t24_n109;
  t24_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t24_n108;
  t24_n107: return -0.05682460218667984f;
  t24_n108: return 0.06129080057144165f;
  t24_n109: return 0.05544688180088997f;
  t24_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t24_n116;
  t24_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t24_n115;
  t24_n112: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n114;
  t24_n113: return -0.003137121209874749f;
  t24_n114: return 0.08331993967294693f;
  t24_n115: return 0.08110979944467545f;
  t24_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t24_n122;
  t24_n117: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n121;
  t24_n118: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t24_n120;
  t24_n119: return 0.029967542737722397f;
  t24_n120: return -0.15728268027305603f;
  t24_n121: return 0.09351178258657455f;
  t24_n122: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n130;
  t24_n123: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t24_n127;
  t24_n124: if (E.getNumReferences() >= 3359752544 /*198405.5*/) goto t24_n126;
  t24_n125: return 0.08111914992332458f;
  t24_n126: return 0.10496819019317627f;
  t24_n127: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t24_n129;
  t24_n128: return -0.10193636268377304f;
  t24_n129: return 0.08945541083812714f;
  t24_n130: return -0.015474632382392883f;
  t24_n131: if (E.getNumReferences() >= 3320626176 /*7575.5*/) goto t24_n239;
  t24_n132: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t24_n188;
  t24_n133: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t24_n179;
  t24_n134: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t24_n144;
  t24_n135: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t24_n141;
  t24_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t24_n140;
  t24_n137: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n139;
  t24_n138: return 0.013157844543457031f;
  t24_n139: return 0.0485570915043354f;
  t24_n140: return 0.04729313403367996f;
  t24_n141: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t24_n143;
  t24_n142: return 0.05778088793158531f;
  t24_n143: return 0.08511848002672195f;
  t24_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t24_n170;
  t24_n145: if (E.getNumReferences() >= 3278405632 /*232.5*/) goto t24_n165;
  t24_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t24_n160;
  t24_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t24_n151;
  t24_n148: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n150;
  t24_n149: return -0.04667884483933449f;
  t24_n150: return 0.024628279730677605f;
  t24_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t24_n153;
  t24_n152: return 0.003667350159958005f;
  t24_n153: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t24_n155;
  t24_n154: return 0.03153389319777489f;
  t24_n155: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t24_n159;
  t24_n156: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t24_n158;
  t24_n157: return -0.009438318200409412f;
  t24_n158: return 0.09772106260061264f;
  t24_n159: return -0.04428441822528839f;
  t24_n160: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t24_n162;
  t24_n161: return -0.09723427146673203f;
  t24_n162: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t24_n164;
  t24_n163: return 0.042777229100465775f;
  t24_n164: return -0.052950672805309296f;
  t24_n165: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t24_n169;
  t24_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t24_n168;
  t24_n167: return 0.0285759549587965f;
  t24_n168: return 0.06794261187314987f;
  t24_n169: return 0.020437201485037804f;
  t24_n170: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t24_n176;
  t24_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t24_n173;
  t24_n172: return -0.01924499124288559f;
  t24_n173: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t24_n175;
  t24_n174: return 0.0462612584233284f;
  t24_n175: return 0.010783005505800247f;
  t24_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t24_n178;
  t24_n177: return 0.03550266847014427f;
  t24_n178: return 0.06686954945325851f;
  t24_n179: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t24_n185;
  t24_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t24_n184;
  t24_n181: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n183;
  t24_n182: return 0.03280389681458473f;
  t24_n183: return 0.07271486520767212f;
  t24_n184: return 0.06921873986721039f;
  t24_n185: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t24_n187;
  t24_n186: return 0.06994906067848206f;
  t24_n187: return 0.1162165030837059f;
  t24_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t24_n204;
  t24_n189: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n201;
  t24_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n192;
  t24_n191: return -0.03598126024007797f;
  t24_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t24_n194;
  t24_n193: return -0.039104361087083817f;
  t24_n194: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t24_n198;
  t24_n195: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t24_n197;
  t24_n196: return 0.004919133614748716f;
  t24_n197: return 0.022040395066142082f;
  t24_n198: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t24_n200;
  t24_n199: return 0.056548845022916794f;
  t24_n200: return 0.020556598901748657f;
  t24_n201: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t24_n203;
  t24_n202: return 0.03681119903922081f;
  t24_n203: return 0.06319963186979294f;
  t24_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t24_n236;
  t24_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t24_n235;
  t24_n206: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t24_n230;
  t24_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n209;
  t24_n208: return -0.10918457806110382f;
  t24_n209: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t24_n221;
  t24_n210: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t24_n220;
  t24_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t24_n219;
  t24_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t24_n218;
  t24_n213: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t24_n217;
  t24_n214: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t24_n216;
  t24_n215: return 0.033598076552152634f;
  t24_n216: return -0.0018233559094369411f;
  t24_n217: return 0.05820286646485329f;
  t24_n218: return 0.04204033687710762f;
  t24_n219: return 0.053393542766571045f;
  t24_n220: return 0.06960007548332214f;
  t24_n221: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t24_n225;
  t24_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t24_n224;
  t24_n223: return 0.02886701561510563f;
  t24_n224: return 0.07283500581979752f;
  t24_n225: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t24_n229;
  t24_n226: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t24_n228;
  t24_n227: return 0.06463898718357086f;
  t24_n228: return -0.044710878282785416f;
  t24_n229: return 0.08710283786058426f;
  t24_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t24_n234;
  t24_n231: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n233;
  t24_n232: return 0.04861126467585564f;
  t24_n233: return 0.09342345595359802f;
  t24_n234: return 0.08036766201257706f;
  t24_n235: return 0.06343098729848862f;
  t24_n236: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n238;
  t24_n237: return 0.0027830321341753006f;
  t24_n238: return 0.07563214004039764f;
  t24_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t24_n243;
  t24_n240: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t24_n242;
  t24_n241: return 0.046058539301157f;
  t24_n242: return 0.09019778668880463f;
  t24_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t24_n249;
  t24_n244: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t24_n248;
  t24_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t24_n247;
  t24_n246: return -0.17707450687885284f;
  t24_n247: return 0.06652458757162094f;
  t24_n248: return 0.09310548007488251f;
  t24_n249: return 0.09284372627735138f;
  t24_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t24_n254;
  t24_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t24_n253;
  t24_n252: return 0.07291499525308609f;
  t24_n253: return 0.08627073466777802f;
  t24_n254: return 0.10067874938249588f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree25(const Example& E) {
  t25_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t25_n252;
  t25_n1: if (E.getFractionNameInContext() >= 3187806340 /*0.1270161271095276*/) goto t25_n177;
  t25_n2: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t25_n176;
  t25_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t25_n89;
  t25_n4: if (E.getNumReferences() >= 3289473024 /*581.5*/) goto t25_n62;
  t25_n5: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t25_n19;
  t25_n6: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t25_n16;
  t25_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t25_n13;
  t25_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t25_n12;
  t25_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t25_n11;
  t25_n10: return -0.04457830637693405f;
  t25_n11: return 0.015509710647165775f;
  t25_n12: return -0.02502615936100483f;
  t25_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t25_n15;
  t25_n14: return 0.032302066683769226f;
  t25_n15: return 0.08106954395771027f;
  t25_n16: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n18;
  t25_n17: return 0.030924739316105843f;
  t25_n18: return 0.07979777455329895f;
  t25_n19: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t25_n57;
  t25_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t25_n28;
  t25_n21: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n23;
  t25_n22: return -0.07472296804189682f;
  t25_n23: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t25_n25;
  t25_n24: return -0.02660578489303589f;
  t25_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t25_n27;
  t25_n26: return -0.08614985644817352f;
  t25_n27: return 0.039446283131837845f;
  t25_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t25_n30;
  t25_n29: return -0.02687297761440277f;
  t25_n30: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t25_n40;
  t25_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t25_n33;
  t25_n32: return -0.03862855210900307f;
  t25_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n37;
  t25_n34: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t25_n36;
  t25_n35: return -0.007625964470207691f;
  t25_n36: return -0.13782262802124023f;
  t25_n37: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t25_n39;
  t25_n38: return 0.009463856928050518f;
  t25_n39: return 0.046005502343177795f;
  t25_n40: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t25_n56;
  t25_n41: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t25_n55;
  t25_n42: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t25_n48;
  t25_n43: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t25_n47;
  t25_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t25_n46;
  t25_n45: return -0.012954932637512684f;
  t25_n46: return 0.036466922610998154f;
  t25_n47: return -0.02071189135313034f;
  t25_n48: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t25_n50;
  t25_n49: return 0.009433649480342865f;
  t25_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n52;
  t25_n51: return 0.03299416974186897f;
  t25_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t25_n54;
  t25_n53: return -0.08199266344308853f;
  t25_n54: return 0.0617600753903389f;
  t25_n55: return 0.0356283038854599f;
  t25_n56: return -0.1068122461438179f;
  t25_n57: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t25_n59;
  t25_n58: return -0.10662547498941422f;
  t25_n59: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t25_n61;
  t25_n60: return 0.007835101336240768f;
  t25_n61: return -0.08111248910427094f;
  t25_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n86;
  t25_n63: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t25_n83;
  t25_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n72;
  t25_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t25_n69;
  t25_n66: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t25_n68;
  t25_n67: return -0.0563492551445961f;
  t25_n68: return 0.01070721261203289f;
  t25_n69: if (E.getNumReferences() >= 3338664448 /*32765.0*/) goto t25_n71;
  t25_n70: return -0.0013012569397687912f;
  t25_n71: return 0.06899567693471909f;
  t25_n72: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t25_n74;
  t25_n73: return 0.026508890092372894f;
  t25_n74: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t25_n82;
  t25_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t25_n77;
  t25_n76: return 0.05875775218009949f;
  t25_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t25_n79;
  t25_n78: return -0.13516467809677124f;
  t25_n79: if (E.getNumReferences() >= 3359032704 /*187158.0*/) goto t25_n81;
  t25_n80: return 0.08303230255842209f;
  t25_n81: return 0.11304394900798798f;
  t25_n82: return -0.008844533003866673f;
  t25_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n85;
  t25_n84: return -0.13421818614006042f;
  t25_n85: return -0.00757240317761898f;
  t25_n86: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t25_n88;
  t25_n87: return 0.03936528414487839f;
  t25_n88: return 0.09682779014110565f;
  t25_n89: if (E.getFileProximityDistanceCost() >= 3255173120 /*33.5*/) goto t25_n115;
  t25_n90: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t25_n106;
  t25_n91: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t25_n101;
  t25_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n94;
  t25_n93: return 0.022219710052013397f;
  t25_n94: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t25_n98;
  t25_n95: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t25_n97;
  t25_n96: return 0.08101535588502884f;
  t25_n97: return -0.07509507983922958f;
  t25_n98: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t25_n100;
  t25_n99: return 0.028504744172096252f;
  t25_n100: return 0.0735241025686264f;
  t25_n101: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n103;
  t25_n102: return -0.012877598404884338f;
  t25_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t25_n105;
  t25_n104: return 0.01358069572597742f;
  t25_n105: return 0.07777697592973709f;
  t25_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n110;
  t25_n107: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t25_n109;
  t25_n108: return 0.0030367558356374502f;
  t25_n109: return 0.049544937908649445f;
  t25_n110: if (E.getNumReferences() >= 3332434432 /*20597.0*/) goto t25_n114;
  t25_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t25_n113;
  t25_n112: return -0.011487503536045551f;
  t25_n113: return 0.06489517539739609f;
  t25_n114: return 0.08147579431533813f;
  t25_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t25_n143;
  t25_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n138;
  t25_n117: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t25_n133;
  t25_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t25_n120;
  t25_n119: return -0.09524787962436676f;
  t25_n120: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t25_n130;
  t25_n121: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t25_n129;
  t25_n122: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t25_n128;
  t25_n123: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t25_n125;
  t25_n124: return 0.009905827231705189f;
  t25_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t25_n127;
  t25_n126: return 0.04435458406805992f;
  t25_n127: return 0.11598225682973862f;
  t25_n128: return -0.010364354588091373f;
  t25_n129: return -0.16902890801429749f;
  t25_n130: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t25_n132;
  t25_n131: return 0.018392672762274742f;
  t25_n132: return -0.08850383758544922f;
  t25_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n135;
  t25_n134: return -0.0753120556473732f;
  t25_n135: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t25_n137;
  t25_n136: return -0.07598181068897247f;
  t25_n137: return 0.03217291086912155f;
  t25_n138: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t25_n140;
  t25_n139: return 0.02620515413582325f;
  t25_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n142;
  t25_n141: return -0.049851033836603165f;
  t25_n142: return 0.057847291231155396f;
  t25_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n157;
  t25_n144: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t25_n154;
  t25_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t25_n151;
  t25_n146: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t25_n148;
  t25_n147: return -0.03554097190499306f;
  t25_n148: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t25_n150;
  t25_n149: return 0.028965743258595467f;
  t25_n150: return -0.031674522906541824f;
  t25_n151: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t25_n153;
  t25_n152: return 0.01435299776494503f;
  t25_n153: return 0.09575559198856354f;
  t25_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n156;
  t25_n155: return -0.013857407495379448f;
  t25_n156: return 0.055724211037158966f;
  t25_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t25_n165;
  t25_n158: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t25_n162;
  t25_n159: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n161;
  t25_n160: return -0.023274598643183708f;
  t25_n161: return 0.040045369416475296f;
  t25_n162: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t25_n164;
  t25_n163: return 0.09303568303585052f;
  t25_n164: return -0.09786032140254974f;
  t25_n165: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t25_n173;
  t25_n166: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t25_n172;
  t25_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n169;
  t25_n168: return 0.02850550413131714f;
  t25_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t25_n171;
  t25_n170: return 0.06606845557689667f;
  t25_n171: return -0.0035564282443374395f;
  t25_n172: return -0.19050925970077515f;
  t25_n173: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t25_n175;
  t25_n174: return 0.08227699995040894f;
  t25_n175: return 0.0056931376457214355f;
  t25_n176: return 0.08774168789386749f;
  t25_n177: if (E.getNumReferences() >= 3303129088 /*1806.0*/) goto t25_n247;
  t25_n178: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t25_n236;
  t25_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t25_n225;
  t25_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t25_n182;
  t25_n181: return -0.08098731189966202f;
  t25_n182: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t25_n208;
  t25_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t25_n205;
  t25_n184: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t25_n188;
  t25_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t25_n187;
  t25_n186: return 0.03340846300125122f;
  t25_n187: return 0.06654965877532959f;
  t25_n188: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t25_n196;
  t25_n189: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t25_n195;
  t25_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t25_n194;
  t25_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n193;
  t25_n192: return 0.010025756433606148f;
  t25_n193: return 0.05797497555613518f;
  t25_n194: return 0.056078147143125534f;
  t25_n195: return 0.06761649250984192f;
  t25_n196: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t25_n200;
  t25_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t25_n199;
  t25_n198: return -0.00725484499707818f;
  t25_n199: return 0.04099170118570328f;
  t25_n200: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t25_n204;
  t25_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n203;
  t25_n202: return 0.036195915192365646f;
  t25_n203: return 0.07510485500097275f;
  t25_n204: return -0.08625160157680511f;
  t25_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t25_n207;
  t25_n206: return 0.040482331067323685f;
  t25_n207: return 0.06984925270080566f;
  t25_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n218;
  t25_n209: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t25_n215;
  t25_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t25_n214;
  t25_n211: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t25_n213;
  t25_n212: return 0.021809836849570274f;
  t25_n213: return 0.07975336909294128f;
  t25_n214: return 0.035743337124586105f;
  t25_n215: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t25_n217;
  t25_n216: return 0.057269856333732605f;
  t25_n217: return 0.0789833813905716f;
  t25_n218: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t25_n224;
  t25_n219: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t25_n223;
  t25_n220: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t25_n222;
  t25_n221: return 0.04949215054512024f;
  t25_n222: return 0.07354723662137985f;
  t25_n223: return 0.043819598853588104f;
  t25_n224: return -0.0666096955537796f;
  t25_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t25_n227;
  t25_n226: return -0.04329584911465645f;
  t25_n227: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t25_n233;
  t25_n228: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t25_n232;
  t25_n229: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t25_n231;
  t25_n230: return 0.033339258283376694f;
  t25_n231: return -0.05207139253616333f;
  t25_n232: return 0.04983522742986679f;
  t25_n233: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t25_n235;
  t25_n234: return 0.020032761618494987f;
  t25_n235: return -0.02814585715532303f;
  t25_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t25_n242;
  t25_n237: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t25_n241;
  t25_n238: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t25_n240;
  t25_n239: return 0.045474469661712646f;
  t25_n240: return 0.09455171227455139f;
  t25_n241: return 0.009529179893434048f;
  t25_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t25_n244;
  t25_n243: return 0.051557548344135284f;
  t25_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t25_n246;
  t25_n245: return 0.05412360653281212f;
  t25_n246: return 0.09424003213644028f;
  t25_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t25_n249;
  t25_n248: return 0.03760486841201782f;
  t25_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t25_n251;
  t25_n250: return 0.06425967067480087f;
  t25_n251: return 0.08851215988397598f;
  t25_n252: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t25_n254;
  t25_n253: return 0.075346939265728f;
  t25_n254: return 0.08789907395839691f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree26(const Example& E) {
  t26_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t26_n250;
  t26_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t26_n119;
  t26_n2: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t26_n90;
  t26_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t26_n51;
  t26_n4: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t26_n46;
  t26_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n19;
  t26_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t26_n18;
  t26_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t26_n9;
  t26_n8: return -0.056688521057367325f;
  t26_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t26_n17;
  t26_n10: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t26_n12;
  t26_n11: return -0.023729657754302025f;
  t26_n12: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t26_n16;
  t26_n13: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t26_n15;
  t26_n14: return 0.006781045347452164f;
  t26_n15: return -0.09365928918123245f;
  t26_n16: return -0.16892889142036438f;
  t26_n17: return -0.06345485895872116f;
  t26_n18: return 0.07908763736486435f;
  t26_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t26_n21;
  t26_n20: return -0.09939649701118469f;
  t26_n21: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t26_n41;
  t26_n22: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t26_n40;
  t26_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t26_n37;
  t26_n24: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t26_n30;
  t26_n25: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t26_n27;
  t26_n26: return 0.01960531249642372f;
  t26_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n29;
  t26_n28: return 0.021737249568104744f;
  t26_n29: return 0.08705512434244156f;
  t26_n30: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t26_n32;
  t26_n31: return -0.005929052364081144f;
  t26_n32: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t26_n36;
  t26_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n35;
  t26_n34: return 0.025309709832072258f;
  t26_n35: return 0.057003047317266464f;
  t26_n36: return -0.04394466429948807f;
  t26_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n39;
  t26_n38: return 0.01698054000735283f;
  t26_n39: return 0.030693987384438515f;
  t26_n40: return -0.1082284152507782f;
  t26_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t26_n45;
  t26_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t26_n44;
  t26_n43: return 0.009102663956582546f;
  t26_n44: return 0.05209973081946373f;
  t26_n45: return 0.052531417459249496f;
  t26_n46: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t26_n50;
  t26_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t26_n49;
  t26_n48: return 0.020128609612584114f;
  t26_n49: return 0.05985894054174423f;
  t26_n50: return 0.05462200567126274f;
  t26_n51: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t26_n85;
  t26_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t26_n70;
  t26_n53: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n67;
  t26_n54: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t26_n60;
  t26_n55: if (E.getNumReferences() >= 3338715264 /*32960.5*/) goto t26_n57;
  t26_n56: return -0.02547462098300457f;
  t26_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n59;
  t26_n58: return 0.009552362374961376f;
  t26_n59: return 0.04664462432265282f;
  t26_n60: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t26_n66;
  t26_n61: if (E.getNumReferences() >= 3354001600 /*119809.5*/) goto t26_n63;
  t26_n62: return -0.0770590528845787f;
  t26_n63: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t26_n65;
  t26_n64: return 0.12277239561080933f;
  t26_n65: return -0.10059624165296555f;
  t26_n66: return -0.15363211929798126f;
  t26_n67: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t26_n69;
  t26_n68: return -0.03874290734529495f;
  t26_n69: return 0.02340787649154663f;
  t26_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t26_n78;
  t26_n71: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t26_n77;
  t26_n72: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t26_n74;
  t26_n73: return 0.036649495363235474f;
  t26_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n76;
  t26_n75: return -0.07359009236097336f;
  t26_n76: return -0.0145797124132514f;
  t26_n77: return 0.040203243494033813f;
  t26_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t26_n84;
  t26_n79: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t26_n83;
  t26_n80: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t26_n82;
  t26_n81: return 0.08636943995952606f;
  t26_n82: return 0.10637534409761429f;
  t26_n83: return 0.04249753803014755f;
  t26_n84: return 0.04685540124773979f;
  t26_n85: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t26_n87;
  t26_n86: return -0.024783611297607422f;
  t26_n87: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t26_n89;
  t26_n88: return 0.05497560650110245f;
  t26_n89: return -0.1648312211036682f;
  t26_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n106;
  t26_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t26_n93;
  t26_n92: return -0.031708575785160065f;
  t26_n93: if (E.getNumReferences() >= 3363896688 /*264171.5*/) goto t26_n99;
  t26_n94: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t26_n98;
  t26_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t26_n97;
  t26_n96: return 0.005990504752844572f;
  t26_n97: return 0.094392329454422f;
  t26_n98: return -0.0650925412774086f;
  t26_n99: if (E.getNumReferences() >= 3364180016 /*273025.5*/) goto t26_n105;
  t26_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t26_n102;
  t26_n101: return 0.04257171228528023f;
  t26_n102: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t26_n104;
  t26_n103: return 0.004607091192156076f;
  t26_n104: return 0.1096184253692627f;
  t26_n105: return 0.07584714144468307f;
  t26_n106: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t26_n116;
  t26_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n115;
  t26_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t26_n112;
  t26_n109: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t26_n111;
  t26_n110: return 0.08579452335834503f;
  t26_n111: return -0.11999232321977615f;
  t26_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n114;
  t26_n113: return 0.08615061640739441f;
  t26_n114: return 0.1256972849369049f;
  t26_n115: return 0.12398365885019302f;
  t26_n116: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t26_n118;
  t26_n117: return 0.05438857153058052f;
  t26_n118: return 0.09871447831392288f;
  t26_n119: if (E.getNumReferences() >= 3320626176 /*7575.5*/) goto t26_n241;
  t26_n120: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t26_n182;
  t26_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t26_n177;
  t26_n122: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n172;
  t26_n123: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t26_n143;
  t26_n124: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t26_n136;
  t26_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n127;
  t26_n126: return 0.005012127570807934f;
  t26_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t26_n129;
  t26_n128: return -0.1055876836180687f;
  t26_n129: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t26_n135;
  t26_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t26_n134;
  t26_n131: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t26_n133;
  t26_n132: return -0.015522364526987076f;
  t26_n133: return 0.08618791401386261f;
  t26_n134: return 0.03929045423865318f;
  t26_n135: return 0.05641775205731392f;
  t26_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n142;
  t26_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t26_n139;
  t26_n138: return 0.023860899731516838f;
  t26_n139: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t26_n141;
  t26_n140: return 0.03104669786989689f;
  t26_n141: return 0.062197159975767136f;
  t26_n142: return 0.07944057136774063f;
  t26_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t26_n147;
  t26_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n146;
  t26_n145: return -0.05640244483947754f;
  t26_n146: return -0.03367221727967262f;
  t26_n147: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t26_n169;
  t26_n148: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t26_n166;
  t26_n149: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t26_n153;
  t26_n150: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t26_n152;
  t26_n151: return 0.039818186312913895f;
  t26_n152: return -0.009758470579981804f;
  t26_n153: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t26_n155;
  t26_n154: return -0.015806525945663452f;
  t26_n155: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t26_n165;
  t26_n156: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t26_n164;
  t26_n157: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t26_n163;
  t26_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n160;
  t26_n159: return -0.045858778059482574f;
  t26_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n162;
  t26_n161: return 0.10030711442232132f;
  t26_n162: return 0.09123774617910385f;
  t26_n163: return -0.0469030886888504f;
  t26_n164: return -0.13580265641212463f;
  t26_n165: return -0.052153367549180984f;
  t26_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t26_n168;
  t26_n167: return -0.5574522614479065f;
  t26_n168: return 0.06758418679237366f;
  t26_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n171;
  t26_n170: return -0.013119873590767384f;
  t26_n171: return 0.07168685644865036f;
  t26_n172: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t26_n176;
  t26_n173: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t26_n175;
  t26_n174: return 0.024543261155486107f;
  t26_n175: return 0.03901701793074608f;
  t26_n176: return 0.1193774864077568f;
  t26_n177: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t26_n181;
  t26_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t26_n180;
  t26_n179: return 0.01607251539826393f;
  t26_n180: return 0.0669250413775444f;
  t26_n181: return 0.06157805025577545f;
  t26_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t26_n184;
  t26_n183: return -0.04662651568651199f;
  t26_n184: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t26_n208;
  t26_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t26_n203;
  t26_n186: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t26_n200;
  t26_n187: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t26_n197;
  t26_n188: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t26_n194;
  t26_n189: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t26_n193;
  t26_n190: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t26_n192;
  t26_n191: return 0.031680621206760406f;
  t26_n192: return -0.008399727754294872f;
  t26_n193: return 0.049269162118434906f;
  t26_n194: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t26_n196;
  t26_n195: return 0.017658470198512077f;
  t26_n196: return -0.0341850109398365f;
  t26_n197: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t26_n199;
  t26_n198: return 0.08074765652418137f;
  t26_n199: return 0.05413491651415825f;
  t26_n200: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t26_n202;
  t26_n201: return 0.05590153858065605f;
  t26_n202: return 0.08988682925701141f;
  t26_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n205;
  t26_n204: return -0.03723074495792389f;
  t26_n205: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t26_n207;
  t26_n206: return 0.03928451985120773f;
  t26_n207: return 0.06419087201356888f;
  t26_n208: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t26_n240;
  t26_n209: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t26_n235;
  t26_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t26_n232;
  t26_n211: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t26_n221;
  t26_n212: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t26_n218;
  t26_n213: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t26_n217;
  t26_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t26_n216;
  t26_n215: return -0.016564732417464256f;
  t26_n216: return 0.031144898384809494f;
  t26_n217: return 0.04685072973370552f;
  t26_n218: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n220;
  t26_n219: return 0.03022221103310585f;
  t26_n220: return 0.06411162763834f;
  t26_n221: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t26_n231;
  t26_n222: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t26_n226;
  t26_n223: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t26_n225;
  t26_n224: return 0.048363812267780304f;
  t26_n225: return -0.035306110978126526f;
  t26_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t26_n228;
  t26_n227: return 0.04887983575463295f;
  t26_n228: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n230;
  t26_n229: return 0.07032612711191177f;
  t26_n230: return 0.10641731321811676f;
  t26_n231: return -0.062209319323301315f;
  t26_n232: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n234;
  t26_n233: return 0.041038207709789276f;
  t26_n234: return 0.08414869010448456f;
  t26_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t26_n237;
  t26_n236: return -0.09231913834810257f;
  t26_n237: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t26_n239;
  t26_n238: return -0.048040490597486496f;
  t26_n239: return 0.04741872847080231f;
  t26_n240: return 0.06487281620502472f;
  t26_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t26_n247;
  t26_n242: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t26_n246;
  t26_n243: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t26_n245;
  t26_n244: return 0.030162474140524864f;
  t26_n245: return 0.07643233984708786f;
  t26_n246: return 0.0873364731669426f;
  t26_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t26_n249;
  t26_n248: return 0.06968195736408234f;
  t26_n249: return 0.08998654037714005f;
  t26_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t26_n254;
  t26_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t26_n253;
  t26_n252: return 0.06736098229885101f;
  t26_n253: return 0.08197091519832611f;
  t26_n254: return 0.09872359037399292f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree27(const Example& E) {
  t27_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t27_n250;
  t27_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t27_n155;
  t27_n2: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t27_n138;
  t27_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t27_n59;
  t27_n4: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t27_n54;
  t27_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n21;
  t27_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t27_n20;
  t27_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t27_n9;
  t27_n8: return -0.054896578192710876f;
  t27_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t27_n19;
  t27_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n18;
  t27_n11: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t27_n13;
  t27_n12: return -0.028897222131490707f;
  t27_n13: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t27_n17;
  t27_n14: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t27_n16;
  t27_n15: return 0.010109794326126575f;
  t27_n16: return -0.024139219895005226f;
  t27_n17: return -0.14115308225154877f;
  t27_n18: return 0.02221117913722992f;
  t27_n19: return -0.06101098656654358f;
  t27_n20: return 0.07650573551654816f;
  t27_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t27_n23;
  t27_n22: return -0.09822528809309006f;
  t27_n23: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t27_n49;
  t27_n24: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t27_n48;
  t27_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t27_n45;
  t27_n26: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t27_n32;
  t27_n27: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t27_n29;
  t27_n28: return 0.01842333935201168f;
  t27_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n31;
  t27_n30: return 0.019983593374490738f;
  t27_n31: return 0.08498378098011017f;
  t27_n32: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t27_n34;
  t27_n33: return -0.006136552896350622f;
  t27_n34: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t27_n44;
  t27_n35: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t27_n37;
  t27_n36: return 0.01393858902156353f;
  t27_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n41;
  t27_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t27_n40;
  t27_n39: return -0.12956923246383667f;
  t27_n40: return 0.050061728805303574f;
  t27_n41: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t27_n43;
  t27_n42: return 0.061443161219358444f;
  t27_n43: return 0.017721718177199364f;
  t27_n44: return -0.04078193008899689f;
  t27_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n47;
  t27_n46: return 0.01650068536400795f;
  t27_n47: return 0.02866552583873272f;
  t27_n48: return -0.10647056251764297f;
  t27_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t27_n53;
  t27_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t27_n52;
  t27_n51: return 0.008381363935768604f;
  t27_n52: return 0.04853443428874016f;
  t27_n53: return 0.04899131879210472f;
  t27_n54: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t27_n58;
  t27_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t27_n57;
  t27_n56: return 0.01674993708729744f;
  t27_n57: return 0.051899079233407974f;
  t27_n58: return 0.0467047281563282f;
  t27_n59: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t27_n133;
  t27_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t27_n118;
  t27_n61: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t27_n117;
  t27_n62: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t27_n106;
  t27_n63: if (E.getSymbolScopeDistanceCost() >= 3252158464 /*27.0*/) goto t27_n75;
  t27_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t27_n70;
  t27_n65: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t27_n69;
  t27_n66: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t27_n68;
  t27_n67: return -0.05032522976398468f;
  t27_n68: return 0.020904025062918663f;
  t27_n69: return -0.09601348638534546f;
  t27_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t27_n72;
  t27_n71: return -0.13665585219860077f;
  t27_n72: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t27_n74;
  t27_n73: return 0.0068383910693228245f;
  t27_n74: return -0.043921686708927155f;
  t27_n75: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n103;
  t27_n76: if (E.getNumReferences() >= 3290030080 /*615.5*/) goto t27_n100;
  t27_n77: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t27_n81;
  t27_n78: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t27_n80;
  t27_n79: return 0.044391948729753494f;
  t27_n80: return -0.06803583353757858f;
  t27_n81: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t27_n99;
  t27_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t27_n84;
  t27_n83: return -0.060227613896131516f;
  t27_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t27_n94;
  t27_n85: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t27_n93;
  t27_n86: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t27_n88;
  t27_n87: return -0.02556418627500534f;
  t27_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n90;
  t27_n89: return -0.06838186830282211f;
  t27_n90: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t27_n92;
  t27_n91: return 0.041117437183856964f;
  t27_n92: return -0.1068190485239029f;
  t27_n93: return -0.09131370484828949f;
  t27_n94: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t27_n98;
  t27_n95: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t27_n97;
  t27_n96: return 0.03580121695995331f;
  t27_n97: return 0.04931050166487694f;
  t27_n98: return -0.09011422097682953f;
  t27_n99: return -0.1297176033258438f;
  t27_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t27_n102;
  t27_n101: return -0.0033012637868523598f;
  t27_n102: return 0.049270398914813995f;
  t27_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t27_n105;
  t27_n104: return -0.05955085530877113f;
  t27_n105: return 0.029164733365178108f;
  t27_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t27_n110;
  t27_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n109;
  t27_n108: return -0.026718884706497192f;
  t27_n109: return 0.042006563395261765f;
  t27_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n112;
  t27_n111: return -0.03709835931658745f;
  t27_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t27_n114;
  t27_n113: return 0.02485499531030655f;
  t27_n114: if (E.getNumReferences() >= 3316109312 /*5370.0*/) goto t27_n116;
  t27_n115: return 0.061465781182050705f;
  t27_n116: return -0.018037322908639908f;
  t27_n117: return -0.09335577487945557f;
  t27_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t27_n126;
  t27_n119: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t27_n125;
  t27_n120: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t27_n122;
  t27_n121: return 0.033681388944387436f;
  t27_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n124;
  t27_n123: return -0.06970115005970001f;
  t27_n124: return -0.014903333969414234f;
  t27_n125: return 0.03612849861383438f;
  t27_n126: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t27_n132;
  t27_n127: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t27_n131;
  t27_n128: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t27_n130;
  t27_n129: return 0.08181336522102356f;
  t27_n130: return 0.10313776880502701f;
  t27_n131: return 0.03824583441019058f;
  t27_n132: return 0.037565089762210846f;
  t27_n133: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t27_n137;
  t27_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t27_n136;
  t27_n135: return -0.046401247382164f;
  t27_n136: return 0.05929142236709595f;
  t27_n137: return 0.04947618767619133f;
  t27_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n142;
  t27_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t27_n141;
  t27_n140: return 0.000743978307582438f;
  t27_n141: return 0.07459289580583572f;
  t27_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t27_n146;
  t27_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t27_n145;
  t27_n144: return -0.006551298778504133f;
  t27_n145: return 0.08843601495027542f;
  t27_n146: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t27_n154;
  t27_n147: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t27_n151;
  t27_n148: if (E.getNumReferences() >= 3359752544 /*198405.5*/) goto t27_n150;
  t27_n149: return 0.07586498558521271f;
  t27_n150: return 0.09589798748493195f;
  t27_n151: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t27_n153;
  t27_n152: return -0.09076095372438431f;
  t27_n153: return 0.08527660369873047f;
  t27_n154: return -0.023993564769625664f;
  t27_n155: if (E.getNumReferences() >= 3309340672 /*3080.5*/) goto t27_n235;
  t27_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t27_n200;
  t27_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t27_n195;
  t27_n158: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n194;
  t27_n159: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t27_n189;
  t27_n160: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t27_n164;
  t27_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n163;
  t27_n162: return 0.025141548365354538f;
  t27_n163: return 0.057466693222522736f;
  t27_n164: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t27_n176;
  t27_n165: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t27_n173;
  t27_n166: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t27_n172;
  t27_n167: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t27_n171;
  t27_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t27_n170;
  t27_n169: return -0.00040069565875455737f;
  t27_n170: return 0.05645941197872162f;
  t27_n171: return 0.0642884150147438f;
  t27_n172: return -0.010735388845205307f;
  t27_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t27_n175;
  t27_n174: return 0.020888676866889f;
  t27_n175: return 0.058669835329055786f;
  t27_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t27_n178;
  t27_n177: return 0.0016626284923404455f;
  t27_n178: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t27_n188;
  t27_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t27_n187;
  t27_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t27_n182;
  t27_n181: return 0.030581381171941757f;
  t27_n182: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t27_n186;
  t27_n183: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t27_n185;
  t27_n184: return 0.023291969671845436f;
  t27_n185: return 0.06533820182085037f;
  t27_n186: return 0.08996883034706116f;
  t27_n187: return -0.029236869886517525f;
  t27_n188: return -0.10151847451925278f;
  t27_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t27_n191;
  t27_n190: return -0.042949531227350235f;
  t27_n191: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t27_n193;
  t27_n192: return 0.015112093649804592f;
  t27_n193: return -0.03528653085231781f;
  t27_n194: return 0.03181548789143562f;
  t27_n195: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t27_n199;
  t27_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t27_n198;
  t27_n197: return 0.014889768324792385f;
  t27_n198: return 0.06327126175165176f;
  t27_n199: return 0.0584188736975193f;
  t27_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t27_n202;
  t27_n201: return -0.04203947260975838f;
  t27_n202: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t27_n222;
  t27_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t27_n219;
  t27_n204: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t27_n216;
  t27_n205: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t27_n211;
  t27_n206: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t27_n210;
  t27_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t27_n209;
  t27_n208: return -0.006779302377253771f;
  t27_n209: return 0.02927275002002716f;
  t27_n210: return -0.01563681662082672f;
  t27_n211: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t27_n213;
  t27_n212: return 0.059307217597961426f;
  t27_n213: if (E.getNumReferences() >= 3270443008 /*119.5*/) goto t27_n215;
  t27_n214: return 0.02371937967836857f;
  t27_n215: return 0.05773753300309181f;
  t27_n216: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t27_n218;
  t27_n217: return 0.05526881292462349f;
  t27_n218: return 0.08832202851772308f;
  t27_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n221;
  t27_n220: return -0.03433993458747864f;
  t27_n221: return 0.052609220147132874f;
  t27_n222: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t27_n234;
  t27_n223: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t27_n229;
  t27_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t27_n226;
  t27_n225: return 0.031185993924736977f;
  t27_n226: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n228;
  t27_n227: return 0.040069084614515305f;
  t27_n228: return 0.08449019491672516f;
  t27_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t27_n231;
  t27_n230: return -0.08958901464939117f;
  t27_n231: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t27_n233;
  t27_n232: return -0.04513317719101906f;
  t27_n233: return 0.04506154730916023f;
  t27_n234: return 0.06201847270131111f;
  t27_n235: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t27_n249;
  t27_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t27_n238;
  t27_n237: return 0.04058661311864853f;
  t27_n238: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t27_n240;
  t27_n239: return 0.0579235665500164f;
  t27_n240: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t27_n246;
  t27_n241: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t27_n245;
  t27_n242: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t27_n244;
  t27_n243: return 0.05048217251896858f;
  t27_n244: return 0.12794436514377594f;
  t27_n245: return 0.07343357801437378f;
  t27_n246: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t27_n248;
  t27_n247: return 0.07897485047578812f;
  t27_n248: return 0.1088867336511612f;
  t27_n249: return 0.09491527825593948f;
  t27_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t27_n254;
  t27_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t27_n253;
  t27_n252: return 0.06483777612447739f;
  t27_n253: return 0.07989729195833206f;
  t27_n254: return 0.09731945395469666f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree28(const Example& E) {
  t28_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t28_n252;
  t28_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t28_n155;
  t28_n2: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t28_n112;
  t28_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t28_n79;
  t28_n4: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t28_n68;
  t28_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n35;
  t28_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t28_n34;
  t28_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t28_n9;
  t28_n8: return -0.05170422047376633f;
  t28_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t28_n33;
  t28_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n32;
  t28_n11: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t28_n13;
  t28_n12: return -0.02584465965628624f;
  t28_n13: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t28_n31;
  t28_n14: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t28_n30;
  t28_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t28_n29;
  t28_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t28_n26;
  t28_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t28_n25;
  t28_n18: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t28_n22;
  t28_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n21;
  t28_n20: return -0.1480434685945511f;
  t28_n21: return -0.02080627717077732f;
  t28_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n24;
  t28_n23: return -0.11473193764686584f;
  t28_n24: return 0.012380711734294891f;
  t28_n25: return 0.04241477698087692f;
  t28_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n28;
  t28_n27: return -0.0891285166144371f;
  t28_n28: return 0.09208983182907104f;
  t28_n29: return -0.017084330320358276f;
  t28_n30: return -0.14392858743667603f;
  t28_n31: return -0.09781856834888458f;
  t28_n32: return 0.02049054019153118f;
  t28_n33: return -0.05839083716273308f;
  t28_n34: return 0.0717950388789177f;
  t28_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t28_n37;
  t28_n36: return -0.09693103283643723f;
  t28_n37: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t28_n63;
  t28_n38: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t28_n62;
  t28_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t28_n59;
  t28_n40: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t28_n46;
  t28_n41: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n43;
  t28_n42: return 0.016220683231949806f;
  t28_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t28_n45;
  t28_n44: return 0.01872973144054413f;
  t28_n45: return 0.0823250338435173f;
  t28_n46: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t28_n48;
  t28_n47: return -0.005524680949747562f;
  t28_n48: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t28_n58;
  t28_n49: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t28_n51;
  t28_n50: return 0.012886445969343185f;
  t28_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n55;
  t28_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n54;
  t28_n53: return -0.11264675855636597f;
  t28_n54: return 0.0444624200463295f;
  t28_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t28_n57;
  t28_n56: return 0.0588812418282032f;
  t28_n57: return 0.017069444060325623f;
  t28_n58: return -0.03809359669685364f;
  t28_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t28_n61;
  t28_n60: return 0.015391561202704906f;
  t28_n61: return 0.027166632935404778f;
  t28_n62: return -0.10535284131765366f;
  t28_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t28_n67;
  t28_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t28_n66;
  t28_n65: return 0.00874739047139883f;
  t28_n66: return 0.04615236073732376f;
  t28_n67: return 0.046604227274656296f;
  t28_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n78;
  t28_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n77;
  t28_n70: if (E.getNumReferences() >= 3317207040 /*5906.0*/) goto t28_n72;
  t28_n71: return 0.008270022459328175f;
  t28_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n74;
  t28_n73: return 0.027814514935016632f;
  t28_n74: if (E.getNumReferences() >= 3342947200 /*49491.5*/) goto t28_n76;
  t28_n75: return 0.08596788346767426f;
  t28_n76: return -0.12068384140729904f;
  t28_n77: return 0.050527047365903854f;
  t28_n78: return 0.053674109280109406f;
  t28_n79: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t28_n107;
  t28_n80: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t28_n106;
  t28_n81: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t28_n95;
  t28_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t28_n84;
  t28_n83: return -0.03569694980978966f;
  t28_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t28_n92;
  t28_n85: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t28_n91;
  t28_n86: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t28_n88;
  t28_n87: return 0.030452914535999298f;
  t28_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n90;
  t28_n89: return -0.06569499522447586f;
  t28_n90: return -0.015425696037709713f;
  t28_n91: return 0.03260684758424759f;
  t28_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t28_n94;
  t28_n93: return 0.05018160492181778f;
  t28_n94: return 0.10281327366828918f;
  t28_n95: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t28_n105;
  t28_n96: if (E.getNumReferences() >= 3318942720 /*6753.5*/) goto t28_n104;
  t28_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t28_n101;
  t28_n98: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n100;
  t28_n99: return 0.001044596079736948f;
  t28_n100: return 0.04197731241583824f;
  t28_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t28_n103;
  t28_n102: return -0.014195054769515991f;
  t28_n103: return 0.05808885768055916f;
  t28_n104: return -0.05140319839119911f;
  t28_n105: return -0.1030004620552063f;
  t28_n106: return -0.09096032381057739f;
  t28_n107: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t28_n111;
  t28_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t28_n110;
  t28_n109: return -0.03565765172243118f;
  t28_n110: return 0.05696915090084076f;
  t28_n111: return 0.05106373876333237f;
  t28_n112: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t28_n118;
  t28_n113: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n117;
  t28_n114: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n116;
  t28_n115: return 0.05508884787559509f;
  t28_n116: return 0.106698177754879f;
  t28_n117: return 0.0956726148724556f;
  t28_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t28_n154;
  t28_n119: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t28_n149;
  t28_n120: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t28_n140;
  t28_n121: if (E.getNumReferences() >= 3355687776 /*134893.5*/) goto t28_n139;
  t28_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n124;
  t28_n123: return -0.008301276713609695f;
  t28_n124: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t28_n132;
  t28_n125: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t28_n127;
  t28_n126: return -0.029141129925847054f;
  t28_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n129;
  t28_n128: return -0.15278786420822144f;
  t28_n129: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n131;
  t28_n130: return 0.13182079792022705f;
  t28_n131: return 7.61427654651925e-05f;
  t28_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n134;
  t28_n133: return -0.15168526768684387f;
  t28_n134: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t28_n136;
  t28_n135: return -0.0038981360848993063f;
  t28_n136: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t28_n138;
  t28_n137: return -0.04383093863725662f;
  t28_n138: return 0.09349572658538818f;
  t28_n139: return -0.07351695746183395f;
  t28_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n146;
  t28_n141: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t28_n145;
  t28_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t28_n144;
  t28_n143: return -0.1464146226644516f;
  t28_n144: return 0.09135929495096207f;
  t28_n145: return 0.022298621013760567f;
  t28_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n148;
  t28_n147: return -0.1359197050333023f;
  t28_n148: return 0.101224385201931f;
  t28_n149: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t28_n151;
  t28_n150: return -0.04754885286092758f;
  t28_n151: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t28_n153;
  t28_n152: return -0.08591583371162415f;
  t28_n153: return 0.06578417867422104f;
  t28_n154: return 0.07604724168777466f;
  t28_n155: if (E.getNumReferences() >= 3309340672 /*3080.5*/) goto t28_n241;
  t28_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t28_n202;
  t28_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t28_n197;
  t28_n158: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n194;
  t28_n159: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t28_n189;
  t28_n160: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t28_n164;
  t28_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t28_n163;
  t28_n162: return 0.023507457226514816f;
  t28_n163: return 0.054460346698760986f;
  t28_n164: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t28_n176;
  t28_n165: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n173;
  t28_n166: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t28_n172;
  t28_n167: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t28_n171;
  t28_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n170;
  t28_n169: return -0.0007008301909081638f;
  t28_n170: return 0.05376634746789932f;
  t28_n171: return 0.0613047257065773f;
  t28_n172: return -0.00954741332679987f;
  t28_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n175;
  t28_n174: return 0.019876891747117043f;
  t28_n175: return 0.05591952055692673f;
  t28_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t28_n178;
  t28_n177: return 0.0015481149312108755f;
  t28_n178: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t28_n188;
  t28_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t28_n187;
  t28_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n182;
  t28_n181: return 0.028315464034676552f;
  t28_n182: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t28_n186;
  t28_n183: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t28_n185;
  t28_n184: return 0.02231230027973652f;
  t28_n185: return 0.06319055706262589f;
  t28_n186: return 0.08556722849607468f;
  t28_n187: return -0.025079328566789627f;
  t28_n188: return -0.0947052389383316f;
  t28_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t28_n191;
  t28_n190: return -0.040689680725336075f;
  t28_n191: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t28_n193;
  t28_n192: return 0.01412515714764595f;
  t28_n193: return -0.03278148174285889f;
  t28_n194: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n196;
  t28_n195: return 0.020780935883522034f;
  t28_n196: return 0.034839991480112076f;
  t28_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n201;
  t28_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t28_n200;
  t28_n199: return 0.014263676479458809f;
  t28_n200: return 0.06001740321516991f;
  t28_n201: return 0.05514563247561455f;
  t28_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t28_n204;
  t28_n203: return -0.03746871277689934f;
  t28_n204: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t28_n226;
  t28_n205: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t28_n217;
  t28_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t28_n212;
  t28_n207: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t28_n209;
  t28_n208: return 0.007017638999968767f;
  t28_n209: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t28_n211;
  t28_n210: return 0.05579717457294464f;
  t28_n211: return 0.029081353917717934f;
  t28_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t28_n214;
  t28_n213: return 0.0101176667958498f;
  t28_n214: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t28_n216;
  t28_n215: return 0.08755938708782196f;
  t28_n216: return 0.03385409712791443f;
  t28_n217: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t28_n225;
  t28_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t28_n224;
  t28_n219: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t28_n223;
  t28_n220: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t28_n222;
  t28_n221: return 0.036695368587970734f;
  t28_n222: return 0.022527368739247322f;
  t28_n223: return 0.05962337926030159f;
  t28_n224: return 0.05748143047094345f;
  t28_n225: return 0.08031335473060608f;
  t28_n226: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t28_n238;
  t28_n227: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t28_n233;
  t28_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n230;
  t28_n229: return 0.030194353312253952f;
  t28_n230: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n232;
  t28_n231: return 0.037732359021902084f;
  t28_n232: return 0.0800907164812088f;
  t28_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t28_n235;
  t28_n234: return -0.08646124601364136f;
  t28_n235: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t28_n237;
  t28_n236: return -0.05672710761427879f;
  t28_n237: return 0.04210464656352997f;
  t28_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n240;
  t28_n239: return 0.027698509395122528f;
  t28_n240: return 0.06629891693592072f;
  t28_n241: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t28_n251;
  t28_n242: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t28_n248;
  t28_n243: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t28_n245;
  t28_n244: return 0.052835676819086075f;
  t28_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t28_n247;
  t28_n246: return 0.05418791621923447f;
  t28_n247: return 0.10078409314155579f;
  t28_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t28_n250;
  t28_n249: return 0.0037530993577092886f;
  t28_n250: return 0.07438463717699051f;
  t28_n251: return 0.08236880600452423f;
  t28_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t28_n254;
  t28_n253: return 0.06793849170207977f;
  t28_n254: return 0.0819200873374939f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree29(const Example& E) {
  t29_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t29_n250;
  t29_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t29_n111;
  t29_n2: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t29_n74;
  t29_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t29_n57;
  t29_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t29_n34;
  t29_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n25;
  t29_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n16;
  t29_n7: if (E.getNumReferences() >= 3308015616 /*2757.0*/) goto t29_n15;
  t29_n8: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t29_n12;
  t29_n9: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t29_n11;
  t29_n10: return -0.11948397010564804f;
  t29_n11: return -0.04391788691282272f;
  t29_n12: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t29_n14;
  t29_n13: return 0.0011944841826334596f;
  t29_n14: return -0.05773455649614334f;
  t29_n15: return 0.02320624515414238f;
  t29_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t29_n18;
  t29_n17: return -0.09492268413305283f;
  t29_n18: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t29_n22;
  t29_n19: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t29_n21;
  t29_n20: return -0.00041819503530859947f;
  t29_n21: return 0.08922731876373291f;
  t29_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t29_n24;
  t29_n23: return 0.014613030478358269f;
  t29_n24: return 0.04488014802336693f;
  t29_n25: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t29_n31;
  t29_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n28;
  t29_n27: return -0.0036310043651610613f;
  t29_n28: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t29_n30;
  t29_n29: return 0.023544035851955414f;
  t29_n30: return 0.05892447754740715f;
  t29_n31: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t29_n33;
  t29_n32: return 0.13324998319149017f;
  t29_n33: return 0.07493004947900772f;
  t29_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n46;
  t29_n35: if (E.getNumReferences() >= 3273424896 /*156.5*/) goto t29_n41;
  t29_n36: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t29_n38;
  t29_n37: return -0.007110360078513622f;
  t29_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t29_n40;
  t29_n39: return 0.016654687002301216f;
  t29_n40: return 0.08329803496599197f;
  t29_n41: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t29_n45;
  t29_n42: if (E.getNumReferences() >= 3324926976 /*11159.0*/) goto t29_n44;
  t29_n43: return 0.032231204211711884f;
  t29_n44: return 0.06725713610649109f;
  t29_n45: return -0.013688432052731514f;
  t29_n46: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t29_n56;
  t29_n47: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t29_n51;
  t29_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t29_n50;
  t29_n49: return 0.018271317705512047f;
  t29_n50: return 0.0505048893392086f;
  t29_n51: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t29_n55;
  t29_n52: if (E.getNumReferences() >= 3275784192 /*192.5*/) goto t29_n54;
  t29_n53: return 0.018821455538272858f;
  t29_n54: return 0.0565386563539505f;
  t29_n55: return -0.10009641200304031f;
  t29_n56: return 0.09512044489383698f;
  t29_n57: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t29_n71;
  t29_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n68;
  t29_n59: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t29_n67;
  t29_n60: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t29_n66;
  t29_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t29_n65;
  t29_n62: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t29_n64;
  t29_n63: return 0.0022506634704768658f;
  t29_n64: return 0.05887269973754883f;
  t29_n65: return 0.04255631938576698f;
  t29_n66: return -0.021750079467892647f;
  t29_n67: return -0.07344096899032593f;
  t29_n68: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t29_n70;
  t29_n69: return -0.03035743348300457f;
  t29_n70: return 0.02006218023598194f;
  t29_n71: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t29_n73;
  t29_n72: return -0.024592354893684387f;
  t29_n73: return 0.03996999189257622f;
  t29_n74: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t29_n108;
  t29_n75: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t29_n81;
  t29_n76: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t29_n78;
  t29_n77: return 0.03150460869073868f;
  t29_n78: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t29_n80;
  t29_n79: return 0.08677578717470169f;
  t29_n80: return 0.06805512309074402f;
  t29_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t29_n107;
  t29_n82: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t29_n84;
  t29_n83: return -0.05671990290284157f;
  t29_n84: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t29_n86;
  t29_n85: return -0.03932544216513634f;
  t29_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t29_n104;
  t29_n87: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t29_n95;
  t29_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t29_n90;
  t29_n89: return 0.03527369350194931f;
  t29_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t29_n94;
  t29_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n93;
  t29_n92: return -0.16222429275512695f;
  t29_n93: return 0.07080846279859543f;
  t29_n94: return 0.08109865337610245f;
  t29_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t29_n99;
  t29_n96: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t29_n98;
  t29_n97: return -0.1511598378419876f;
  t29_n98: return 0.04221218824386597f;
  t29_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n101;
  t29_n100: return -0.1697389781475067f;
  t29_n101: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t29_n103;
  t29_n102: return -0.17863242328166962f;
  t29_n103: return 0.09741964936256409f;
  t29_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n106;
  t29_n105: return -0.10902635008096695f;
  t29_n106: return 0.022055240347981453f;
  t29_n107: return 0.05804700404405594f;
  t29_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t29_n110;
  t29_n109: return 0.05871719866991043f;
  t29_n110: return 0.088137686252594f;
  t29_n111: if (E.getNumReferences() >= 3309340672 /*3080.5*/) goto t29_n237;
  t29_n112: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t29_n170;
  t29_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t29_n165;
  t29_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t29_n130;
  t29_n115: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t29_n127;
  t29_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n126;
  t29_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t29_n119;
  t29_n118: return 0.004673061426728964f;
  t29_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t29_n121;
  t29_n120: return -0.10472628474235535f;
  t29_n121: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t29_n125;
  t29_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t29_n124;
  t29_n123: return -0.005279206205159426f;
  t29_n124: return 0.03410521149635315f;
  t29_n125: return 0.053005170077085495f;
  t29_n126: return 0.03136134892702103f;
  t29_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t29_n129;
  t29_n128: return 0.04409226402640343f;
  t29_n129: return 0.07363655418157578f;
  t29_n130: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t29_n164;
  t29_n131: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t29_n161;
  t29_n132: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t29_n160;
  t29_n133: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t29_n157;
  t29_n134: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t29_n138;
  t29_n135: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t29_n137;
  t29_n136: return 0.015586393885314465f;
  t29_n137: return 0.07335933297872543f;
  t29_n138: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t29_n156;
  t29_n139: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t29_n141;
  t29_n140: return -0.008568771183490753f;
  t29_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t29_n143;
  t29_n142: return -0.011811408214271069f;
  t29_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t29_n155;
  t29_n144: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t29_n154;
  t29_n145: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t29_n147;
  t29_n146: return -0.08139266818761826f;
  t29_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n149;
  t29_n148: return -0.05085267871618271f;
  t29_n149: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t29_n153;
  t29_n150: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t29_n152;
  t29_n151: return 0.09153381735086441f;
  t29_n152: return -0.02692231722176075f;
  t29_n153: return 0.09592506289482117f;
  t29_n154: return -0.07798359543085098f;
  t29_n155: return -0.047748863697052f;
  t29_n156: return -0.05163893476128578f;
  t29_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t29_n159;
  t29_n158: return 0.012837531976401806f;
  t29_n159: return 0.0652552992105484f;
  t29_n160: return -0.09455323964357376f;
  t29_n161: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t29_n163;
  t29_n162: return -0.03325660899281502f;
  t29_n163: return 0.0512549951672554f;
  t29_n164: return 0.046577177941799164f;
  t29_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t29_n169;
  t29_n166: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t29_n168;
  t29_n167: return 0.012979437597095966f;
  t29_n168: return 0.04966020956635475f;
  t29_n169: return 0.06325966119766235f;
  t29_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t29_n172;
  t29_n171: return -0.03365720435976982f;
  t29_n172: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t29_n198;
  t29_n173: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t29_n183;
  t29_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t29_n178;
  t29_n175: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t29_n177;
  t29_n176: return 0.007564837113022804f;
  t29_n177: return 0.038253989070653915f;
  t29_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t29_n180;
  t29_n179: return 0.009848499670624733f;
  t29_n180: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t29_n182;
  t29_n181: return 0.0853831022977829f;
  t29_n182: return 0.032702866941690445f;
  t29_n183: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t29_n197;
  t29_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t29_n196;
  t29_n185: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t29_n195;
  t29_n186: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t29_n190;
  t29_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t29_n189;
  t29_n188: return 0.028777752071619034f;
  t29_n189: return 0.06279182434082031f;
  t29_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n192;
  t29_n191: return -0.10883091390132904f;
  t29_n192: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n194;
  t29_n193: return 0.0208722036331892f;
  t29_n194: return 0.05230061709880829f;
  t29_n195: return 0.0575391948223114f;
  t29_n196: return 0.0541723407804966f;
  t29_n197: return 0.07774142920970917f;
  t29_n198: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t29_n236;
  t29_n199: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t29_n231;
  t29_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t29_n230;
  t29_n201: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n217;
  t29_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t29_n208;
  t29_n203: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t29_n205;
  t29_n204: return 0.007197953760623932f;
  t29_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t29_n207;
  t29_n206: return 0.012714296579360962f;
  t29_n207: return 0.04648563638329506f;
  t29_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t29_n210;
  t29_n209: return 0.019863486289978027f;
  t29_n210: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t29_n214;
  t29_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n213;
  t29_n212: return -0.025589825585484505f;
  t29_n213: return 0.038998425006866455f;
  t29_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t29_n216;
  t29_n215: return 0.05105379968881607f;
  t29_n216: return 0.06700385361909866f;
  t29_n217: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t29_n223;
  t29_n218: if (E.getFileProximityDistanceCost() >= 3252158464 /*27.0*/) goto t29_n222;
  t29_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t29_n221;
  t29_n220: return 0.055803753435611725f;
  t29_n221: return -0.1089334785938263f;
  t29_n222: return -0.05485520511865616f;
  t29_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t29_n225;
  t29_n224: return -0.06263580173254013f;
  t29_n225: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t29_n227;
  t29_n226: return 0.06254716962575912f;
  t29_n227: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t29_n229;
  t29_n228: return 0.0966532900929451f;
  t29_n229: return -0.2579922080039978f;
  t29_n230: return 0.04729384928941727f;
  t29_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t29_n233;
  t29_n232: return -0.0831596851348877f;
  t29_n233: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t29_n235;
  t29_n234: return -0.05373558774590492f;
  t29_n235: return 0.040914759039878845f;
  t29_n236: return 0.05748714506626129f;
  t29_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t29_n241;
  t29_n238: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t29_n240;
  t29_n239: return 0.0360809825360775f;
  t29_n240: return 0.08088003098964691f;
  t29_n241: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t29_n243;
  t29_n242: return 0.05415734648704529f;
  t29_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t29_n249;
  t29_n244: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t29_n248;
  t29_n245: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t29_n247;
  t29_n246: return 0.05137704685330391f;
  t29_n247: return 0.12378296256065369f;
  t29_n248: return 0.08566766232252121f;
  t29_n249: return 0.10481078922748566f;
  t29_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t29_n254;
  t29_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t29_n253;
  t29_n252: return 0.059359967708587646f;
  t29_n253: return 0.07541389763355255f;
  t29_n254: return 0.09523913264274597f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree30(const Example& E) {
  t30_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t30_n252;
  t30_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t30_n115;
  t30_n2: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t30_n98;
  t30_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t30_n43;
  t30_n4: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t30_n40;
  t30_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t30_n19;
  t30_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t30_n18;
  t30_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t30_n9;
  t30_n8: return -0.0488622710108757f;
  t30_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t30_n17;
  t30_n10: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t30_n12;
  t30_n11: return -0.021522454917430878f;
  t30_n12: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t30_n16;
  t30_n13: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n15;
  t30_n14: return 0.005322855897247791f;
  t30_n15: return -0.15789274871349335f;
  t30_n16: return -0.08782973140478134f;
  t30_n17: return -0.05377275124192238f;
  t30_n18: return 0.06687027961015701f;
  t30_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t30_n21;
  t30_n20: return -0.09438060224056244f;
  t30_n21: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t30_n37;
  t30_n22: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n36;
  t30_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t30_n33;
  t30_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t30_n32;
  t30_n25: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t30_n31;
  t30_n26: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t30_n28;
  t30_n27: return 0.01892719604074955f;
  t30_n28: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t30_n30;
  t30_n29: return 0.057284947484731674f;
  t30_n30: return -0.07780008763074875f;
  t30_n31: return 0.0065612285397946835f;
  t30_n32: return -0.01826765388250351f;
  t30_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t30_n35;
  t30_n34: return 0.015888450667262077f;
  t30_n35: return 0.024427957832813263f;
  t30_n36: return -0.10285010188817978f;
  t30_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t30_n39;
  t30_n38: return 0.022345244884490967f;
  t30_n39: return 0.04307738319039345f;
  t30_n40: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t30_n42;
  t30_n41: return 0.017517006024718285f;
  t30_n42: return 0.04244013875722885f;
  t30_n43: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t30_n93;
  t30_n44: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t30_n90;
  t30_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t30_n85;
  t30_n46: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t30_n52;
  t30_n47: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t30_n51;
  t30_n48: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t30_n50;
  t30_n49: return -0.03450384363532066f;
  t30_n50: return 0.04120306670665741f;
  t30_n51: return -0.08732997626066208f;
  t30_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t30_n78;
  t30_n53: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t30_n55;
  t30_n54: return -0.05753720551729202f;
  t30_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t30_n59;
  t30_n56: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t30_n58;
  t30_n57: return -0.0445745587348938f;
  t30_n58: return -0.08384106308221817f;
  t30_n59: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t30_n75;
  t30_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t30_n72;
  t30_n61: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t30_n71;
  t30_n62: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t30_n64;
  t30_n63: return -0.024512236937880516f;
  t30_n64: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n70;
  t30_n65: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t30_n69;
  t30_n66: if (E.getFileProximityDistanceCost() >= 3254910976 /*32.5*/) goto t30_n68;
  t30_n67: return 0.03700254112482071f;
  t30_n68: return -0.16985902190208435f;
  t30_n69: return -0.06421791017055511f;
  t30_n70: return -0.12890276312828064f;
  t30_n71: return -0.08839187771081924f;
  t30_n72: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n74;
  t30_n73: return 0.03209618851542473f;
  t30_n74: return -0.12238617241382599f;
  t30_n75: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t30_n77;
  t30_n76: return 0.06493328511714935f;
  t30_n77: return 0.005209289025515318f;
  t30_n78: if (E.getNumReferences() >= 3268608000 /*105.5*/) goto t30_n82;
  t30_n79: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t30_n81;
  t30_n80: return 0.06875486671924591f;
  t30_n81: return -0.009659246541559696f;
  t30_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t30_n84;
  t30_n83: return -0.020121801644563675f;
  t30_n84: return 0.057003505527973175f;
  t30_n85: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t30_n89;
  t30_n86: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t30_n88;
  t30_n87: return 0.0005694928113371134f;
  t30_n88: return 0.029253382235765457f;
  t30_n89: return -0.03643738105893135f;
  t30_n90: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t30_n92;
  t30_n91: return -0.027217797935009003f;
  t30_n92: return 0.019154459238052368f;
  t30_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t30_n97;
  t30_n94: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t30_n96;
  t30_n95: return -0.03396669775247574f;
  t30_n96: return 0.03614475205540657f;
  t30_n97: return 0.05774068459868431f;
  t30_n98: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t30_n112;
  t30_n99: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t30_n107;
  t30_n100: if (E.getNumReferences() >= 3358114336 /*172808.5*/) goto t30_n106;
  t30_n101: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t30_n105;
  t30_n102: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t30_n104;
  t30_n103: return 0.062447551637887955f;
  t30_n104: return -0.014186302199959755f;
  t30_n105: return 0.06274707615375519f;
  t30_n106: return -0.15338076651096344f;
  t30_n107: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t30_n111;
  t30_n108: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t30_n110;
  t30_n109: return 0.08069968223571777f;
  t30_n110: return 0.11871843039989471f;
  t30_n111: return 0.080286405980587f;
  t30_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t30_n114;
  t30_n113: return 0.005712434649467468f;
  t30_n114: return 0.06244051456451416f;
  t30_n115: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t30_n235;
  t30_n116: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t30_n190;
  t30_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t30_n185;
  t30_n118: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t30_n150;
  t30_n119: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t30_n147;
  t30_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t30_n146;
  t30_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t30_n135;
  t30_n122: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t30_n134;
  t30_n123: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t30_n133;
  t30_n124: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t30_n132;
  t30_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t30_n129;
  t30_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t30_n128;
  t30_n127: return -0.1882205605506897f;
  t30_n128: return -0.1101563423871994f;
  t30_n129: if (E.getFileProximityDistanceCost() >= 3255042048 /*33.0*/) goto t30_n131;
  t30_n130: return -0.012955175712704659f;
  t30_n131: return 0.049105238169431686f;
  t30_n132: return 0.008464979007840157f;
  t30_n133: return 0.08424864709377289f;
  t30_n134: return -0.08545787632465363f;
  t30_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t30_n137;
  t30_n136: return -0.03904174268245697f;
  t30_n137: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t30_n143;
  t30_n138: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t30_n142;
  t30_n139: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t30_n141;
  t30_n140: return 0.026649633422493935f;
  t30_n141: return -0.0013432442210614681f;
  t30_n142: return -0.027742043137550354f;
  t30_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t30_n145;
  t30_n144: return 0.018663667142391205f;
  t30_n145: return 0.05669742450118065f;
  t30_n146: return 0.029992137104272842f;
  t30_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t30_n149;
  t30_n148: return 0.04232913628220558f;
  t30_n149: return 0.06652430444955826f;
  t30_n150: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t30_n184;
  t30_n151: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t30_n183;
  t30_n152: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t30_n182;
  t30_n153: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t30_n175;
  t30_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t30_n156;
  t30_n155: return -0.02493319660425186f;
  t30_n156: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t30_n158;
  t30_n157: return 0.0034075237344950438f;
  t30_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t30_n166;
  t30_n159: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n165;
  t30_n160: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t30_n162;
  t30_n161: return 0.004257092718034983f;
  t30_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t30_n164;
  t30_n163: return 0.07544683665037155f;
  t30_n164: return -0.01079794205725193f;
  t30_n165: return -0.12918312847614288f;
  t30_n166: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t30_n168;
  t30_n167: return 0.005628400016576052f;
  t30_n168: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n174;
  t30_n169: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t30_n173;
  t30_n170: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t30_n172;
  t30_n171: return 0.0676712915301323f;
  t30_n172: return -0.02139037288725376f;
  t30_n173: return 0.08903664350509644f;
  t30_n174: return -0.1171005442738533f;
  t30_n175: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t30_n181;
  t30_n176: if (E.getNumReferences() >= 3256745984 /*39.5*/) goto t30_n178;
  t30_n177: return 0.005949538666754961f;
  t30_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t30_n180;
  t30_n179: return 0.09309937059879303f;
  t30_n180: return 0.16495999693870544f;
  t30_n181: return -0.06646249443292618f;
  t30_n182: return 0.01722104847431183f;
  t30_n183: return -0.08822080492973328f;
  t30_n184: return 0.04177059978246689f;
  t30_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t30_n189;
  t30_n186: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t30_n188;
  t30_n187: return 0.01230979710817337f;
  t30_n188: return 0.046972423791885376f;
  t30_n189: return 0.06093563511967659f;
  t30_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t30_n192;
  t30_n191: return -0.029613470658659935f;
  t30_n192: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t30_n226;
  t30_n193: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t30_n221;
  t30_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t30_n206;
  t30_n195: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t30_n205;
  t30_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t30_n202;
  t30_n197: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t30_n201;
  t30_n198: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t30_n200;
  t30_n199: return 0.015508972108364105f;
  t30_n200: return 0.02468775399029255f;
  t30_n201: return -0.06396942585706711f;
  t30_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t30_n204;
  t30_n203: return 0.010766744613647461f;
  t30_n204: return 0.042043790221214294f;
  t30_n205: return 0.0569855198264122f;
  t30_n206: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t30_n220;
  t30_n207: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t30_n211;
  t30_n208: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t30_n210;
  t30_n209: return 0.05021703988313675f;
  t30_n210: return 0.07034517824649811f;
  t30_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t30_n217;
  t30_n212: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t30_n214;
  t30_n213: return 0.027568282559514046f;
  t30_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type))) goto t30_n216;
  t30_n215: return 0.03677414357662201f;
  t30_n216: return 0.054188091307878494f;
  t30_n217: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t30_n219;
  t30_n218: return 0.05509111285209656f;
  t30_n219: return -0.11555109918117523f;
  t30_n220: return -0.06553547084331512f;
  t30_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t30_n223;
  t30_n222: return -0.14799654483795166f;
  t30_n223: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t30_n225;
  t30_n224: return 0.007183616980910301f;
  t30_n225: return 0.032969165593385696f;
  t30_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t30_n234;
  t30_n227: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t30_n229;
  t30_n228: return 0.05526537820696831f;
  t30_n229: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t30_n233;
  t30_n230: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t30_n232;
  t30_n231: return -0.0036995471455156803f;
  t30_n232: return 0.1182938888669014f;
  t30_n233: return 0.06582283973693848f;
  t30_n234: return 0.06192893162369728f;
  t30_n235: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t30_n245;
  t30_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t30_n238;
  t30_n237: return 0.025752494111657143f;
  t30_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t30_n242;
  t30_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t30_n241;
  t30_n240: return -0.06514526158571243f;
  t30_n241: return 0.04849109426140785f;
  t30_n242: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t30_n244;
  t30_n243: return 0.057397447526454926f;
  t30_n244: return 0.09401583671569824f;
  t30_n245: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t30_n251;
  t30_n246: if (E.getNumReferences() >= 3337910528 /*31292.5*/) goto t30_n248;
  t30_n247: return 0.0452331118285656f;
  t30_n248: if (E.getNumReferences() >= 3338286080 /*32026.0*/) goto t30_n250;
  t30_n249: return 0.1367819756269455f;
  t30_n250: return 0.09052231907844543f;
  t30_n251: return 0.11291104555130005f;
  t30_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t30_n254;
  t30_n253: return 0.06283742934465408f;
  t30_n254: return 0.07781297713518143f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree31(const Example& E) {
  t31_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t31_n250;
  t31_n1: if (E.getFractionNameInContext() >= 3189477818 /*0.15192309021949768*/) goto t31_n163;
  t31_n2: if (E.getNumReferences() >= 3369019008 /*424244.0*/) goto t31_n162;
  t31_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t31_n83;
  t31_n4: if (E.getNumReferences() >= 3317122048 /*5864.5*/) goto t31_n68;
  t31_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n57;
  t31_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t31_n20;
  t31_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n17;
  t31_n8: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t31_n16;
  t31_n9: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t31_n11;
  t31_n10: return -0.015871193259954453f;
  t31_n11: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t31_n13;
  t31_n12: return 0.03794604912400246f;
  t31_n13: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t31_n15;
  t31_n14: return 0.07060610502958298f;
  t31_n15: return 0.11146526038646698f;
  t31_n16: return -0.04300956428050995f;
  t31_n17: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t31_n19;
  t31_n18: return -0.004992357920855284f;
  t31_n19: return 0.07271485775709152f;
  t31_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n44;
  t31_n21: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t31_n41;
  t31_n22: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t31_n36;
  t31_n23: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t31_n35;
  t31_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t31_n34;
  t31_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t31_n29;
  t31_n26: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t31_n28;
  t31_n27: return 0.0043676006607711315f;
  t31_n28: return -0.15059080719947815f;
  t31_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n31;
  t31_n30: return -0.07946015894412994f;
  t31_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t31_n33;
  t31_n32: return -0.04446970298886299f;
  t31_n33: return 0.012537873350083828f;
  t31_n34: return 0.03861755132675171f;
  t31_n35: return -0.11752240359783173f;
  t31_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t31_n40;
  t31_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t31_n39;
  t31_n38: return 0.0119274090975523f;
  t31_n39: return 0.09675638377666473f;
  t31_n40: return 0.08080832660198212f;
  t31_n41: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t31_n43;
  t31_n42: return 0.018453532829880714f;
  t31_n43: return 0.053265154361724854f;
  t31_n44: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t31_n56;
  t31_n45: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t31_n55;
  t31_n46: if (E.getNumReferences() >= 3275784192 /*192.5*/) goto t31_n54;
  t31_n47: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t31_n53;
  t31_n48: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t31_n52;
  t31_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t31_n51;
  t31_n50: return 0.019129851832985878f;
  t31_n51: return 0.047907087951898575f;
  t31_n52: return -0.1262829303741455f;
  t31_n53: return 0.01871340535581112f;
  t31_n54: return 0.05809534713625908f;
  t31_n55: return -0.09731214493513107f;
  t31_n56: return 0.0862656682729721f;
  t31_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t31_n63;
  t31_n58: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t31_n60;
  t31_n59: return -0.3265455961227417f;
  t31_n60: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t31_n62;
  t31_n61: return 0.016937721520662308f;
  t31_n62: return -0.04993107542395592f;
  t31_n63: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t31_n67;
  t31_n64: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t31_n66;
  t31_n65: return 0.05896938964724541f;
  t31_n66: return 0.02242685668170452f;
  t31_n67: return 0.05818074196577072f;
  t31_n68: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n78;
  t31_n69: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t31_n73;
  t31_n70: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t31_n72;
  t31_n71: return 0.04437659680843353f;
  t31_n72: return 0.07439224421977997f;
  t31_n73: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t31_n75;
  t31_n74: return 0.014671692624688148f;
  t31_n75: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t31_n77;
  t31_n76: return 0.08386264741420746f;
  t31_n77: return -0.08215655386447906f;
  t31_n78: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t31_n82;
  t31_n79: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t31_n81;
  t31_n80: return 0.023732542991638184f;
  t31_n81: return 0.13343532383441925f;
  t31_n82: return 0.08271971344947815f;
  t31_n83: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t31_n151;
  t31_n84: if (E.getNumReferences() >= 3292913664 /*791.5*/) goto t31_n124;
  t31_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t31_n113;
  t31_n86: if (E.getFractionNameInContext() >= 3146676277 /*0.004347825888544321*/) goto t31_n100;
  t31_n87: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t31_n99;
  t31_n88: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t31_n94;
  t31_n89: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t31_n93;
  t31_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t31_n92;
  t31_n91: return -0.08483614772558212f;
  t31_n92: return -0.02033904567360878f;
  t31_n93: return -0.12143848091363907f;
  t31_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t31_n98;
  t31_n95: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n97;
  t31_n96: return -0.019721131771802902f;
  t31_n97: return 0.03308212757110596f;
  t31_n98: return 0.030078331008553505f;
  t31_n99: return -0.08336486667394638f;
  t31_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t31_n104;
  t31_n101: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n103;
  t31_n102: return -0.04365561902523041f;
  t31_n103: return 0.02177293412387371f;
  t31_n104: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t31_n112;
  t31_n105: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t31_n107;
  t31_n106: return -0.0021798715461045504f;
  t31_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t31_n109;
  t31_n108: return 0.015940839424729347f;
  t31_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t31_n111;
  t31_n110: return -0.06260258704423904f;
  t31_n111: return 0.03996725380420685f;
  t31_n112: return 0.06589046865701675f;
  t31_n113: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t31_n123;
  t31_n114: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t31_n116;
  t31_n115: return -0.01608338952064514f;
  t31_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n120;
  t31_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t31_n119;
  t31_n118: return 0.04505482316017151f;
  t31_n119: return 0.11402006447315216f;
  t31_n120: if (E.getNumReferences() >= 3287793664 /*495.5*/) goto t31_n122;
  t31_n121: return 0.09573016315698624f;
  t31_n122: return 0.13804307579994202f;
  t31_n123: return -0.08380501717329025f;
  t31_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t31_n144;
  t31_n125: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t31_n127;
  t31_n126: return -0.02841862291097641f;
  t31_n127: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n143;
  t31_n128: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t31_n140;
  t31_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t31_n131;
  t31_n130: return 0.009568772278726101f;
  t31_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t31_n135;
  t31_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n134;
  t31_n133: return -0.14580246806144714f;
  t31_n134: return 0.051132939755916595f;
  t31_n135: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t31_n139;
  t31_n136: if (E.getNumReferences() >= 3346303232 /*62601.0*/) goto t31_n138;
  t31_n137: return 0.03952091559767723f;
  t31_n138: return 0.07927732169628143f;
  t31_n139: return -0.039945363998413086f;
  t31_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n142;
  t31_n141: return -0.066285140812397f;
  t31_n142: return 0.005361567251384258f;
  t31_n143: return 0.05120770260691643f;
  t31_n144: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t31_n150;
  t31_n145: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t31_n149;
  t31_n146: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t31_n148;
  t31_n147: return 0.06675709784030914f;
  t31_n148: return 0.012902600690722466f;
  t31_n149: return -0.5112839341163635f;
  t31_n150: return 0.06158776581287384f;
  t31_n151: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t31_n155;
  t31_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t31_n154;
  t31_n153: return -0.0303091611713171f;
  t31_n154: return 0.04935561493039131f;
  t31_n155: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t31_n161;
  t31_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t31_n160;
  t31_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n159;
  t31_n158: return -0.10197562724351883f;
  t31_n159: return 0.04735410585999489f;
  t31_n160: return 0.059180181473493576f;
  t31_n161: return -0.129667267203331f;
  t31_n162: return 0.07850899547338486f;
  t31_n163: if (E.getNumReferences() >= 3278405632 /*232.5*/) goto t31_n241;
  t31_n164: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n230;
  t31_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t31_n175;
  t31_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t31_n168;
  t31_n167: return -0.03968057408928871f;
  t31_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t31_n170;
  t31_n169: return -0.02287527173757553f;
  t31_n170: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t31_n172;
  t31_n171: return 0.009697560220956802f;
  t31_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t31_n174;
  t31_n173: return 0.016153480857610703f;
  t31_n174: return 0.04505108669400215f;
  t31_n175: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t31_n209;
  t31_n176: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t31_n196;
  t31_n177: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t31_n195;
  t31_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t31_n180;
  t31_n179: return -0.023675406351685524f;
  t31_n180: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t31_n186;
  t31_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t31_n185;
  t31_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t31_n184;
  t31_n183: return 0.008630923926830292f;
  t31_n184: return 0.06475615501403809f;
  t31_n185: return 0.042475562542676926f;
  t31_n186: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t31_n188;
  t31_n187: return 0.020342355594038963f;
  t31_n188: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t31_n190;
  t31_n189: return -0.0896872729063034f;
  t31_n190: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t31_n192;
  t31_n191: return 0.03805159032344818f;
  t31_n192: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t31_n194;
  t31_n193: return 0.059022389352321625f;
  t31_n194: return -0.026219729334115982f;
  t31_n195: return 0.04962872713804245f;
  t31_n196: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t31_n202;
  t31_n197: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t31_n201;
  t31_n198: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t31_n200;
  t31_n199: return 0.0281737819314003f;
  t31_n200: return 0.05755442753434181f;
  t31_n201: return 0.08649983257055283f;
  t31_n202: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t31_n204;
  t31_n203: return 0.06251999735832214f;
  t31_n204: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t31_n206;
  t31_n205: return 0.10993747413158417f;
  t31_n206: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t31_n208;
  t31_n207: return 0.09678182750940323f;
  t31_n208: return 0.13064809143543243f;
  t31_n209: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t31_n229;
  t31_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t31_n222;
  t31_n211: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t31_n213;
  t31_n212: return 0.013315653428435326f;
  t31_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t31_n217;
  t31_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t31_n216;
  t31_n215: return -0.12063969671726227f;
  t31_n216: return 0.04155013710260391f;
  t31_n217: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t31_n221;
  t31_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t31_n220;
  t31_n219: return 0.026346173137426376f;
  t31_n220: return 0.06716059148311615f;
  t31_n221: return 0.08126100152730942f;
  t31_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t31_n228;
  t31_n223: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t31_n227;
  t31_n224: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t31_n226;
  t31_n225: return 0.04257833585143089f;
  t31_n226: return 0.024861782789230347f;
  t31_n227: return 0.05660329386591911f;
  t31_n228: return 0.05070158839225769f;
  t31_n229: return 0.0071503836661577225f;
  t31_n230: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t31_n236;
  t31_n231: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t31_n235;
  t31_n232: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t31_n234;
  t31_n233: return 0.039511919021606445f;
  t31_n234: return 0.08900243788957596f;
  t31_n235: return 0.0005142244626767933f;
  t31_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t31_n238;
  t31_n237: return 0.042724400758743286f;
  t31_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t31_n240;
  t31_n239: return 0.044405922293663025f;
  t31_n240: return 0.08095543831586838f;
  t31_n241: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t31_n247;
  t31_n242: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t31_n244;
  t31_n243: return 0.06755217909812927f;
  t31_n244: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t31_n246;
  t31_n245: return 0.02956334315240383f;
  t31_n246: return 0.08490217477083206f;
  t31_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t31_n249;
  t31_n248: return 0.04284221678972244f;
  t31_n249: return 0.07675749063491821f;
  t31_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t31_n254;
  t31_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t31_n253;
  t31_n252: return 0.05431217700242996f;
  t31_n253: return 0.0709424614906311f;
  t31_n254: return 0.09287775307893753f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree32(const Example& E) {
  t32_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t32_n252;
  t32_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t32_n127;
  t32_n2: if (E.getNumReferences() >= 3326663680 /*12855.0*/) goto t32_n112;
  t32_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t32_n41;
  t32_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t32_n16;
  t32_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n9;
  t32_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n8;
  t32_n7: return -0.025433171540498734f;
  t32_n8: return -0.007574731949716806f;
  t32_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n11;
  t32_n10: return -0.002793812658637762f;
  t32_n11: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t32_n13;
  t32_n12: return 0.0262871365994215f;
  t32_n13: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t32_n15;
  t32_n14: return 0.1323920339345932f;
  t32_n15: return 0.06142183765769005f;
  t32_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n30;
  t32_n17: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t32_n23;
  t32_n18: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t32_n20;
  t32_n19: return -0.005967851262539625f;
  t32_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t32_n22;
  t32_n21: return 0.014607377350330353f;
  t32_n22: return 0.06884945183992386f;
  t32_n23: if (E.getNumReferences() >= 3324926976 /*11159.0*/) goto t32_n29;
  t32_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t32_n28;
  t32_n25: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t32_n27;
  t32_n26: return 0.05138380080461502f;
  t32_n27: return 0.019728241488337517f;
  t32_n28: return -0.020006824284791946f;
  t32_n29: return 0.1063934788107872f;
  t32_n30: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t32_n40;
  t32_n31: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t32_n39;
  t32_n32: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t32_n38;
  t32_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t32_n37;
  t32_n34: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t32_n36;
  t32_n35: return 0.025251390412449837f;
  t32_n36: return 0.06277299672365189f;
  t32_n37: return 0.02055121399462223f;
  t32_n38: return -0.10052908957004547f;
  t32_n39: return 0.08812910318374634f;
  t32_n40: return -0.09211175888776779f;
  t32_n41: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n111;
  t32_n42: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t32_n110;
  t32_n43: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t32_n47;
  t32_n44: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t32_n46;
  t32_n45: return 0.003224417567253113f;
  t32_n46: return 0.05489467456936836f;
  t32_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t32_n109;
  t32_n48: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t32_n106;
  t32_n49: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t32_n105;
  t32_n50: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t32_n68;
  t32_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t32_n53;
  t32_n52: return -0.04693494737148285f;
  t32_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n61;
  t32_n54: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t32_n60;
  t32_n55: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t32_n57;
  t32_n56: return -0.030010119080543518f;
  t32_n57: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t32_n59;
  t32_n58: return 0.003950499929487705f;
  t32_n59: return 0.06728505343198776f;
  t32_n60: return -0.08490835875272751f;
  t32_n61: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t32_n63;
  t32_n62: return -0.0016225090948864818f;
  t32_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t32_n65;
  t32_n64: return -0.005540388636291027f;
  t32_n65: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t32_n67;
  t32_n66: return 0.02479381673038006f;
  t32_n67: return 0.06375989317893982f;
  t32_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t32_n90;
  t32_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t32_n73;
  t32_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t32_n72;
  t32_n71: return -0.1574506014585495f;
  t32_n72: return -0.02173517644405365f;
  t32_n73: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t32_n89;
  t32_n74: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t32_n86;
  t32_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t32_n79;
  t32_n76: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t32_n78;
  t32_n77: return -0.0009745062561705709f;
  t32_n78: return -0.08526100963354111f;
  t32_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t32_n81;
  t32_n80: return 0.002393584232777357f;
  t32_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t32_n83;
  t32_n82: return 0.1318207085132599f;
  t32_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t32_n85;
  t32_n84: return 0.03713025152683258f;
  t32_n85: return 0.09563842415809631f;
  t32_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t32_n88;
  t32_n87: return 0.016307257115840912f;
  t32_n88: return 0.04876011237502098f;
  t32_n89: return -0.0651320368051529f;
  t32_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t32_n92;
  t32_n91: return -0.049697376787662506f;
  t32_n92: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t32_n104;
  t32_n93: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t32_n95;
  t32_n94: return -0.030485359951853752f;
  t32_n95: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t32_n99;
  t32_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t32_n98;
  t32_n97: return -0.11572805047035217f;
  t32_n98: return 0.08790314942598343f;
  t32_n99: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t32_n101;
  t32_n100: return -0.18547776341438293f;
  t32_n101: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t32_n103;
  t32_n102: return 0.07626546919345856f;
  t32_n103: return 0.032237708568573f;
  t32_n104: return 0.03727389872074127f;
  t32_n105: return -0.12174712121486664f;
  t32_n106: if (E.getNumReferences() >= 3288997888 /*552.5*/) goto t32_n108;
  t32_n107: return -0.0704488530755043f;
  t32_n108: return 0.00013735098764300346f;
  t32_n109: return 0.035673242062330246f;
  t32_n110: return -0.06792878359556198f;
  t32_n111: return 0.005053394939750433f;
  t32_n112: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t32_n122;
  t32_n113: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t32_n117;
  t32_n114: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t32_n116;
  t32_n115: return 0.03069232404232025f;
  t32_n116: return 0.0676826685667038f;
  t32_n117: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t32_n121;
  t32_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n120;
  t32_n119: return 0.07420697808265686f;
  t32_n120: return 0.11696656793355942f;
  t32_n121: return 0.0774284154176712f;
  t32_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t32_n126;
  t32_n123: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t32_n125;
  t32_n124: return 0.0017446177080273628f;
  t32_n125: return 0.06838768720626831f;
  t32_n126: return 0.052662745118141174f;
  t32_n127: if (E.getNumReferences() >= 3302404096 /*1717.5*/) goto t32_n239;
  t32_n128: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t32_n176;
  t32_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t32_n171;
  t32_n130: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n168;
  t32_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t32_n163;
  t32_n132: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t32_n136;
  t32_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t32_n135;
  t32_n134: return 0.02116810716688633f;
  t32_n135: return 0.04782012850046158f;
  t32_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t32_n140;
  t32_n137: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t32_n139;
  t32_n138: return 0.004547476302832365f;
  t32_n139: return -0.08342328667640686f;
  t32_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t32_n142;
  t32_n141: return -0.10680737346410751f;
  t32_n142: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t32_n146;
  t32_n143: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t32_n145;
  t32_n144: return 0.030850153416395187f;
  t32_n145: return 0.075289785861969f;
  t32_n146: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t32_n162;
  t32_n147: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t32_n149;
  t32_n148: return -0.0023279651068150997f;
  t32_n149: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t32_n161;
  t32_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t32_n152;
  t32_n151: return 0.007910139858722687f;
  t32_n152: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t32_n156;
  t32_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion))) goto t32_n155;
  t32_n154: return 0.03544469550251961f;
  t32_n155: return 0.08271973580121994f;
  t32_n156: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t32_n160;
  t32_n157: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t32_n159;
  t32_n158: return -0.0013123888056725264f;
  t32_n159: return 0.08458434045314789f;
  t32_n160: return -0.09985116869211197f;
  t32_n161: return -0.0800049677491188f;
  t32_n162: return 0.0744016245007515f;
  t32_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n165;
  t32_n164: return -0.04602662846446037f;
  t32_n165: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t32_n167;
  t32_n166: return 0.022960321977734566f;
  t32_n167: return -0.012764731422066689f;
  t32_n168: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t32_n170;
  t32_n169: return 0.02633214369416237f;
  t32_n170: return -0.10444667190313339f;
  t32_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t32_n175;
  t32_n172: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t32_n174;
  t32_n173: return 0.011815586127340794f;
  t32_n174: return 0.04415002837777138f;
  t32_n175: return 0.05819724500179291f;
  t32_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t32_n178;
  t32_n177: return -0.022255809977650642f;
  t32_n178: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t32_n198;
  t32_n179: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t32_n187;
  t32_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t32_n182;
  t32_n181: return 0.011223805136978626f;
  t32_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t32_n184;
  t32_n183: return 0.004582082852721214f;
  t32_n184: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t32_n186;
  t32_n185: return 0.08025812357664108f;
  t32_n186: return 0.028621966019272804f;
  t32_n187: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t32_n195;
  t32_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t32_n194;
  t32_n189: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t32_n193;
  t32_n190: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t32_n192;
  t32_n191: return 0.03200753033161163f;
  t32_n192: return -0.006367719732224941f;
  t32_n193: return 0.04678735136985779f;
  t32_n194: return 0.04996670410037041f;
  t32_n195: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t32_n197;
  t32_n196: return 0.05350644886493683f;
  t32_n197: return 0.08684835582971573f;
  t32_n198: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t32_n238;
  t32_n199: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t32_n233;
  t32_n200: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t32_n232;
  t32_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t32_n231;
  t32_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n220;
  t32_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t32_n209;
  t32_n204: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t32_n206;
  t32_n205: return 0.006248227320611477f;
  t32_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t32_n208;
  t32_n207: return 0.01640472374856472f;
  t32_n208: return 0.040019452571868896f;
  t32_n209: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t32_n219;
  t32_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t32_n218;
  t32_n211: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t32_n213;
  t32_n212: return -0.02199694514274597f;
  t32_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t32_n215;
  t32_n214: return -0.014408871531486511f;
  t32_n215: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t32_n217;
  t32_n216: return 0.046629346907138824f;
  t32_n217: return 0.08078368008136749f;
  t32_n218: return 0.03707560896873474f;
  t32_n219: return 0.05319524556398392f;
  t32_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t32_n224;
  t32_n221: if (E.getFileProximityDistanceCost() >= 3252158464 /*27.0*/) goto t32_n223;
  t32_n222: return 0.042507730424404144f;
  t32_n223: return -0.050040796399116516f;
  t32_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n226;
  t32_n225: return -0.05597410351037979f;
  t32_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t32_n228;
  t32_n227: return 0.04865969344973564f;
  t32_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t32_n230;
  t32_n229: return 0.07758935540914536f;
  t32_n230: return 0.09569396078586578f;
  t32_n231: return 0.04438488185405731f;
  t32_n232: return -0.10627920925617218f;
  t32_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t32_n235;
  t32_n234: return -0.07579398155212402f;
  t32_n235: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t32_n237;
  t32_n236: return -0.03778025135397911f;
  t32_n237: return 0.03919453173875809f;
  t32_n238: return 0.05950223654508591f;
  t32_n239: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t32_n249;
  t32_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t32_n242;
  t32_n241: return 0.02255702018737793f;
  t32_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t32_n246;
  t32_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t32_n245;
  t32_n244: return -0.05873020738363266f;
  t32_n245: return 0.04534343630075455f;
  t32_n246: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t32_n248;
  t32_n247: return 0.0541410893201828f;
  t32_n248: return 0.09004807472229004f;
  t32_n249: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n251;
  t32_n250: return 0.07162114977836609f;
  t32_n251: return 0.11082888394594193f;
  t32_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t32_n254;
  t32_n253: return 0.06238105893135071f;
  t32_n254: return 0.0911487489938736f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree33(const Example& E) {
  t33_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t33_n252;
  t33_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t33_n113;
  t33_n2: if (E.getNumReferences() >= 3326663680 /*12855.0*/) goto t33_n90;
  t33_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t33_n67;
  t33_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t33_n42;
  t33_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t33_n35;
  t33_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n24;
  t33_n7: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t33_n23;
  t33_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t33_n18;
  t33_n9: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n13;
  t33_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n12;
  t33_n11: return -0.12175574153661728f;
  t33_n12: return -0.04076904430985451f;
  t33_n13: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t33_n17;
  t33_n14: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t33_n16;
  t33_n15: return 0.005826680455356836f;
  t33_n16: return -0.15612958371639252f;
  t33_n17: return -0.04758225008845329f;
  t33_n18: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t33_n20;
  t33_n19: return 0.03595619648694992f;
  t33_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t33_n22;
  t33_n21: return 0.0579693429172039f;
  t33_n22: return 0.11195779591798782f;
  t33_n23: return -0.03634306415915489f;
  t33_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t33_n26;
  t33_n25: return -0.0912558063864708f;
  t33_n26: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n30;
  t33_n27: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t33_n29;
  t33_n28: return 2.6864781830227003e-06f;
  t33_n29: return 0.07785079628229141f;
  t33_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t33_n34;
  t33_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t33_n33;
  t33_n32: return -0.008677436038851738f;
  t33_n33: return 0.03165026754140854f;
  t33_n34: return 0.04124071076512337f;
  t33_n35: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t33_n39;
  t33_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n38;
  t33_n37: return -0.002155431779101491f;
  t33_n38: return 0.02432139962911606f;
  t33_n39: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t33_n41;
  t33_n40: return 0.13143736124038696f;
  t33_n41: return 0.062478289008140564f;
  t33_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n58;
  t33_n43: if (E.getNumReferences() >= 3324926976 /*11159.0*/) goto t33_n57;
  t33_n44: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t33_n54;
  t33_n45: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t33_n51;
  t33_n46: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t33_n50;
  t33_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n49;
  t33_n48: return -0.01202970091253519f;
  t33_n49: return 0.015089817345142365f;
  t33_n50: return -0.05465606227517128f;
  t33_n51: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t33_n53;
  t33_n52: return 0.009805376641452312f;
  t33_n53: return -0.11056502908468246f;
  t33_n54: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t33_n56;
  t33_n55: return 0.02547796443104744f;
  t33_n56: return -0.02018129825592041f;
  t33_n57: return 0.1046493649482727f;
  t33_n58: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t33_n66;
  t33_n59: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t33_n65;
  t33_n60: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t33_n64;
  t33_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n63;
  t33_n62: return 0.016295718029141426f;
  t33_n63: return 0.026506666094064713f;
  t33_n64: return -0.10100885480642319f;
  t33_n65: return -0.09039565175771713f;
  t33_n66: return 0.08224193006753922f;
  t33_n67: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t33_n83;
  t33_n68: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t33_n74;
  t33_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t33_n71;
  t33_n70: return -0.09097080677747726f;
  t33_n71: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t33_n73;
  t33_n72: return -0.01818476989865303f;
  t33_n73: return -0.11667662858963013f;
  t33_n74: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t33_n80;
  t33_n75: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t33_n77;
  t33_n76: return -0.06181420758366585f;
  t33_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n79;
  t33_n78: return -0.0006392320501618087f;
  t33_n79: return 0.07585887610912323f;
  t33_n80: if (E.getNumReferences() >= 3293372416 /*819.5*/) goto t33_n82;
  t33_n81: return -0.0897437110543251f;
  t33_n82: return -0.012227202765643597f;
  t33_n83: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n89;
  t33_n84: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t33_n88;
  t33_n85: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t33_n87;
  t33_n86: return 0.014405899681150913f;
  t33_n87: return -0.09925343841314316f;
  t33_n88: return -0.10219266265630722f;
  t33_n89: return -0.07968942075967789f;
  t33_n90: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t33_n106;
  t33_n91: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t33_n105;
  t33_n92: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t33_n104;
  t33_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t33_n103;
  t33_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t33_n100;
  t33_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t33_n99;
  t33_n96: if (E.getNumReferences() >= 3329227264 /*15358.5*/) goto t33_n98;
  t33_n97: return 0.13145677745342255f;
  t33_n98: return 0.015942417085170746f;
  t33_n99: return 0.06519461423158646f;
  t33_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t33_n102;
  t33_n101: return -0.11253321915864944f;
  t33_n102: return 0.0504574216902256f;
  t33_n103: return 0.012387664057314396f;
  t33_n104: return 0.07812011241912842f;
  t33_n105: return 0.08100694417953491f;
  t33_n106: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t33_n112;
  t33_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t33_n111;
  t33_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t33_n110;
  t33_n109: return 3.11071053147316e-05f;
  t33_n110: return 0.04695973917841911f;
  t33_n111: return 0.06597568094730377f;
  t33_n112: return 0.06702791899442673f;
  t33_n113: if (E.getNumReferences() >= 3292798976 /*784.5*/) goto t33_n243;
  t33_n114: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t33_n192;
  t33_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t33_n185;
  t33_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t33_n154;
  t33_n117: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t33_n151;
  t33_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t33_n148;
  t33_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t33_n135;
  t33_n120: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t33_n134;
  t33_n121: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t33_n133;
  t33_n122: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n130;
  t33_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n127;
  t33_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t33_n126;
  t33_n125: return -0.17409683763980865f;
  t33_n126: return -0.10949891060590744f;
  t33_n127: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t33_n129;
  t33_n128: return -0.005879647098481655f;
  t33_n129: return 0.04265758395195007f;
  t33_n130: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t33_n132;
  t33_n131: return 0.008275886997580528f;
  t33_n132: return -0.08490685373544693f;
  t33_n133: return 0.07703462243080139f;
  t33_n134: return -0.08199869096279144f;
  t33_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t33_n137;
  t33_n136: return -0.028339318931102753f;
  t33_n137: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n145;
  t33_n138: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t33_n144;
  t33_n139: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t33_n141;
  t33_n140: return 0.01928507536649704f;
  t33_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t33_n143;
  t33_n142: return 0.020267551764845848f;
  t33_n143: return 0.04688533768057823f;
  t33_n144: return -0.024786505848169327f;
  t33_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t33_n147;
  t33_n146: return 0.01810513436794281f;
  t33_n147: return 0.05461912229657173f;
  t33_n148: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t33_n150;
  t33_n149: return 0.027314554899930954f;
  t33_n150: return -0.10234981030225754f;
  t33_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t33_n153;
  t33_n152: return 0.03916846215724945f;
  t33_n153: return 0.05445679649710655f;
  t33_n154: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t33_n184;
  t33_n155: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n183;
  t33_n156: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t33_n176;
  t33_n157: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t33_n175;
  t33_n158: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t33_n160;
  t33_n159: return -0.003240457270294428f;
  t33_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t33_n162;
  t33_n161: return 0.0008328019175678492f;
  t33_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t33_n174;
  t33_n163: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t33_n173;
  t33_n164: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t33_n166;
  t33_n165: return -0.07714113593101501f;
  t33_n166: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t33_n170;
  t33_n167: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t33_n169;
  t33_n168: return 0.07171369343996048f;
  t33_n169: return -0.03564215078949928f;
  t33_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t33_n172;
  t33_n171: return 0.06983990222215652f;
  t33_n172: return 0.07971078157424927f;
  t33_n173: return -0.0732477605342865f;
  t33_n174: return -0.03878169134259224f;
  t33_n175: return 0.04654564708471298f;
  t33_n176: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t33_n182;
  t33_n177: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t33_n179;
  t33_n178: return -0.01975932903587818f;
  t33_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t33_n181;
  t33_n180: return 0.06166255101561546f;
  t33_n181: return 0.1297113448381424f;
  t33_n182: return -0.06427007913589478f;
  t33_n183: return -0.0820629894733429f;
  t33_n184: return 0.03810277581214905f;
  t33_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t33_n191;
  t33_n186: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t33_n190;
  t33_n187: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t33_n189;
  t33_n188: return 0.04763873666524887f;
  t33_n189: return -0.0022695926018059254f;
  t33_n190: return 0.040864668786525726f;
  t33_n191: return 0.05600574612617493f;
  t33_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t33_n194;
  t33_n193: return -0.0191775131970644f;
  t33_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t33_n230;
  t33_n195: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t33_n225;
  t33_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t33_n204;
  t33_n197: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t33_n203;
  t33_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t33_n200;
  t33_n199: return 0.017872275784611702f;
  t33_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t33_n202;
  t33_n201: return 0.009535695426166058f;
  t33_n202: return 0.03730694204568863f;
  t33_n203: return 0.05151539668440819f;
  t33_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t33_n224;
  t33_n205: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t33_n209;
  t33_n206: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t33_n208;
  t33_n207: return 0.0451677180826664f;
  t33_n208: return 0.06493972986936569f;
  t33_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t33_n215;
  t33_n210: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t33_n212;
  t33_n211: return 0.024302620440721512f;
  t33_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type))) goto t33_n214;
  t33_n213: return 0.03178567439317703f;
  t33_n214: return 0.0478428453207016f;
  t33_n215: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n223;
  t33_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t33_n218;
  t33_n217: return -0.020066142082214355f;
  t33_n218: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t33_n220;
  t33_n219: return 0.07011142373085022f;
  t33_n220: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t33_n222;
  t33_n221: return 0.01788417063653469f;
  t33_n222: return 0.048015691339969635f;
  t33_n223: return 0.07333295792341232f;
  t33_n224: return -0.059132251888513565f;
  t33_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t33_n227;
  t33_n226: return -0.14348652958869934f;
  t33_n227: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t33_n229;
  t33_n228: return 0.0054922266863286495f;
  t33_n229: return 0.03075670450925827f;
  t33_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t33_n240;
  t33_n231: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t33_n233;
  t33_n232: return 0.05044933408498764f;
  t33_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t33_n239;
  t33_n234: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t33_n238;
  t33_n235: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t33_n237;
  t33_n236: return -0.0338645875453949f;
  t33_n237: return 0.05384417995810509f;
  t33_n238: return 0.11450212448835373f;
  t33_n239: return 0.059937722980976105f;
  t33_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t33_n242;
  t33_n241: return 0.04596750810742378f;
  t33_n242: return 0.07501328736543655f;
  t33_n243: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t33_n251;
  t33_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t33_n246;
  t33_n245: return 0.022834522649645805f;
  t33_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t33_n248;
  t33_n247: return 0.02454250305891037f;
  t33_n248: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t33_n250;
  t33_n249: return 0.051524706184864044f;
  t33_n250: return 0.08442248404026031f;
  t33_n251: return 0.07035399228334427f;
  t33_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t33_n254;
  t33_n253: return 0.05507612228393555f;
  t33_n254: return 0.07183859497308731f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree34(const Example& E) {
  t34_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t34_n252;
  t34_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t34_n117;
  t34_n2: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t34_n98;
  t34_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t34_n89;
  t34_n4: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t34_n84;
  t34_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n33;
  t34_n6: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t34_n32;
  t34_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n23;
  t34_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n12;
  t34_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t34_n11;
  t34_n10: return -0.13979223370552063f;
  t34_n11: return -0.10978157818317413f;
  t34_n12: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t34_n22;
  t34_n13: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t34_n19;
  t34_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t34_n16;
  t34_n15: return -0.057705070823431015f;
  t34_n16: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n18;
  t34_n17: return -0.022495103999972343f;
  t34_n18: return 0.0239657424390316f;
  t34_n19: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t34_n21;
  t34_n20: return 0.061083417385816574f;
  t34_n21: return 0.0008640990126878023f;
  t34_n22: return 0.04695865139365196f;
  t34_n23: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t34_n31;
  t34_n24: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t34_n30;
  t34_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n29;
  t34_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t34_n28;
  t34_n27: return -0.11897297203540802f;
  t34_n28: return 0.11270076036453247f;
  t34_n29: return 0.008437794633209705f;
  t34_n30: return -0.1467464566230774f;
  t34_n31: return -0.0811101421713829f;
  t34_n32: return -0.05143105983734131f;
  t34_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t34_n35;
  t34_n34: return -0.09050822257995605f;
  t34_n35: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n75;
  t34_n36: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t34_n74;
  t34_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t34_n69;
  t34_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t34_n62;
  t34_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t34_n41;
  t34_n40: return -0.026268677785992622f;
  t34_n41: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t34_n47;
  t34_n42: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t34_n44;
  t34_n43: return 0.018174001947045326f;
  t34_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t34_n46;
  t34_n45: return 0.028556067496538162f;
  t34_n46: return 0.08816750347614288f;
  t34_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n53;
  t34_n48: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t34_n50;
  t34_n49: return 0.0047701275907456875f;
  t34_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t34_n52;
  t34_n51: return 0.02348346635699272f;
  t34_n52: return 0.06210104003548622f;
  t34_n53: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t34_n61;
  t34_n54: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t34_n56;
  t34_n55: return 0.001589955179952085f;
  t34_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n60;
  t34_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t34_n59;
  t34_n58: return -0.09493313729763031f;
  t34_n59: return 0.04297884181141853f;
  t34_n60: return 0.058596547693014145f;
  t34_n61: return -0.0848858654499054f;
  t34_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n64;
  t34_n63: return -0.022209366783499718f;
  t34_n64: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t34_n68;
  t34_n65: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n67;
  t34_n66: return -0.006056469865143299f;
  t34_n67: return 0.04985922574996948f;
  t34_n68: return 0.07092997431755066f;
  t34_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n71;
  t34_n70: return -0.08001022040843964f;
  t34_n71: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t34_n73;
  t34_n72: return 0.020191194489598274f;
  t34_n73: return 0.08829309791326523f;
  t34_n74: return -0.09762012958526611f;
  t34_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t34_n79;
  t34_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t34_n78;
  t34_n77: return 0.0028617980424314737f;
  t34_n78: return 0.03628191351890564f;
  t34_n79: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t34_n81;
  t34_n80: return 0.05009978264570236f;
  t34_n81: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n83;
  t34_n82: return -0.02947591245174408f;
  t34_n83: return 0.04609984904527664f;
  t34_n84: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n88;
  t34_n85: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t34_n87;
  t34_n86: return 0.04286125302314758f;
  t34_n87: return 0.015704365447163582f;
  t34_n88: return 0.06271465867757797f;
  t34_n89: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t34_n91;
  t34_n90: return -0.026277726516127586f;
  t34_n91: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n97;
  t34_n92: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t34_n96;
  t34_n93: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t34_n95;
  t34_n94: return 0.013408105820417404f;
  t34_n95: return -0.08287939429283142f;
  t34_n96: return -0.09439411759376526f;
  t34_n97: return -0.07635243982076645f;
  t34_n98: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t34_n104;
  t34_n99: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t34_n103;
  t34_n100: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n102;
  t34_n101: return 0.07198544591665268f;
  t34_n102: return 0.11539303511381149f;
  t34_n103: return 0.07414313405752182f;
  t34_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t34_n116;
  t34_n105: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t34_n107;
  t34_n106: return 0.05622943118214607f;
  t34_n107: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t34_n109;
  t34_n108: return -0.1549025923013687f;
  t34_n109: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t34_n115;
  t34_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n112;
  t34_n111: return -0.14888006448745728f;
  t34_n112: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t34_n114;
  t34_n113: return 0.0723169595003128f;
  t34_n114: return 0.10509346425533295f;
  t34_n115: return -0.012238109484314919f;
  t34_n116: return 0.06720864027738571f;
  t34_n117: if (E.getNumReferences() >= 3292798976 /*784.5*/) goto t34_n243;
  t34_n118: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t34_n200;
  t34_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t34_n195;
  t34_n120: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t34_n160;
  t34_n121: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t34_n159;
  t34_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t34_n142;
  t34_n123: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t34_n141;
  t34_n124: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t34_n140;
  t34_n125: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n133;
  t34_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n130;
  t34_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t34_n129;
  t34_n128: return -0.18208061158657074f;
  t34_n129: return -0.10577019304037094f;
  t34_n130: if (E.getFileProximityDistanceCost() >= 3255042048 /*33.0*/) goto t34_n132;
  t34_n131: return -0.007594272959977388f;
  t34_n132: return 0.046160101890563965f;
  t34_n133: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t34_n139;
  t34_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n138;
  t34_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t34_n137;
  t34_n136: return -0.11642000079154968f;
  t34_n137: return 0.09949857741594315f;
  t34_n138: return 0.011085662059485912f;
  t34_n139: return -0.0789208710193634f;
  t34_n140: return 0.07792055606842041f;
  t34_n141: return -0.0794069841504097f;
  t34_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t34_n144;
  t34_n143: return -0.028236130252480507f;
  t34_n144: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n156;
  t34_n145: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t34_n155;
  t34_n146: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t34_n154;
  t34_n147: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n151;
  t34_n148: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t34_n150;
  t34_n149: return 0.017896903678774834f;
  t34_n150: return 0.03983793035149574f;
  t34_n151: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t34_n153;
  t34_n152: return 0.022466203197836876f;
  t34_n153: return 0.04899481311440468f;
  t34_n154: return -0.018794404342770576f;
  t34_n155: return -0.00839778408408165f;
  t34_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t34_n158;
  t34_n157: return 0.020696556195616722f;
  t34_n158: return 0.05510292574763298f;
  t34_n159: return 0.03983599692583084f;
  t34_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t34_n168;
  t34_n161: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n165;
  t34_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n164;
  t34_n163: return -0.0444624163210392f;
  t34_n164: return -0.02159486711025238f;
  t34_n165: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n167;
  t34_n166: return 0.020078329369425774f;
  t34_n167: return -0.23920340836048126f;
  t34_n168: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t34_n192;
  t34_n169: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n189;
  t34_n170: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t34_n174;
  t34_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t34_n173;
  t34_n172: return -0.013202021829783916f;
  t34_n173: return 0.012304096482694149f;
  t34_n174: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t34_n188;
  t34_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n177;
  t34_n176: return -0.0752284973859787f;
  t34_n177: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t34_n187;
  t34_n178: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t34_n180;
  t34_n179: return 0.008831041865050793f;
  t34_n180: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t34_n186;
  t34_n181: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t34_n185;
  t34_n182: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t34_n184;
  t34_n183: return 0.071070596575737f;
  t34_n184: return -0.03201868757605553f;
  t34_n185: return 0.07930006831884384f;
  t34_n186: return -0.007254859898239374f;
  t34_n187: return -0.08448418974876404f;
  t34_n188: return -0.08270088583230972f;
  t34_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n191;
  t34_n190: return -0.4506841003894806f;
  t34_n191: return 0.05737049877643585f;
  t34_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t34_n194;
  t34_n193: return -0.008243480697274208f;
  t34_n194: return 0.05522102117538452f;
  t34_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t34_n199;
  t34_n196: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t34_n198;
  t34_n197: return 0.01002114824950695f;
  t34_n198: return 0.039024803787469864f;
  t34_n199: return 0.05393816903233528f;
  t34_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t34_n202;
  t34_n201: return -0.015002305619418621f;
  t34_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t34_n232;
  t34_n203: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t34_n227;
  t34_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t34_n218;
  t34_n205: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t34_n211;
  t34_n206: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t34_n208;
  t34_n207: return 0.011836130172014236f;
  t34_n208: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t34_n210;
  t34_n209: return 0.017811404541134834f;
  t34_n210: return 0.06709031760692596f;
  t34_n211: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t34_n217;
  t34_n212: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t34_n216;
  t34_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t34_n215;
  t34_n214: return 0.024412008002400398f;
  t34_n215: return 0.06612809747457504f;
  t34_n216: return 0.05278582498431206f;
  t34_n217: return 0.04044632986187935f;
  t34_n218: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t34_n226;
  t34_n219: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t34_n221;
  t34_n220: return 0.04616950824856758f;
  t34_n221: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t34_n225;
  t34_n222: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t34_n224;
  t34_n223: return 0.030580587685108185f;
  t34_n224: return 0.07730386406183243f;
  t34_n225: return 0.068732351064682f;
  t34_n226: return -0.057809196412563324f;
  t34_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t34_n229;
  t34_n228: return -0.14042073488235474f;
  t34_n229: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t34_n231;
  t34_n230: return 0.005672418978065252f;
  t34_n231: return 0.030063925310969353f;
  t34_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t34_n240;
  t34_n233: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t34_n235;
  t34_n234: return 0.048850830644369125f;
  t34_n235: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t34_n239;
  t34_n236: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t34_n238;
  t34_n237: return -0.0016680924454703927f;
  t34_n238: return 0.10931283235549927f;
  t34_n239: return 0.057586535811424255f;
  t34_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t34_n242;
  t34_n241: return 0.04272296652197838f;
  t34_n242: return 0.07494264096021652f;
  t34_n243: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t34_n251;
  t34_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t34_n246;
  t34_n245: return 0.021434186026453972f;
  t34_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t34_n248;
  t34_n247: return 0.023309938609600067f;
  t34_n248: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t34_n250;
  t34_n249: return 0.04909480735659599f;
  t34_n250: return 0.08214148879051208f;
  t34_n251: return 0.06867145001888275f;
  t34_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t34_n254;
  t34_n253: return 0.052966732531785965f;
  t34_n254: return 0.06974685937166214f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree35(const Example& E) {
  t35_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t35_n250;
  t35_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t35_n109;
  t35_n2: if (E.getNumReferences() >= 3325353472 /*11575.5*/) goto t35_n94;
  t35_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t35_n71;
  t35_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n24;
  t35_n5: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t35_n21;
  t35_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t35_n8;
  t35_n7: return -0.04010966420173645f;
  t35_n8: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t35_n20;
  t35_n9: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t35_n19;
  t35_n10: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t35_n18;
  t35_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t35_n17;
  t35_n12: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t35_n16;
  t35_n13: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n15;
  t35_n14: return -0.01916283182799816f;
  t35_n15: return 0.02767273783683777f;
  t35_n16: return 0.004888914525508881f;
  t35_n17: return -0.04320479556918144f;
  t35_n18: return -0.13595302402973175f;
  t35_n19: return -0.08454163372516632f;
  t35_n20: return 0.06486214697360992f;
  t35_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t35_n23;
  t35_n22: return 0.06714434921741486f;
  t35_n23: return 0.012616646476089954f;
  t35_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t35_n26;
  t35_n25: return -0.08887531608343124f;
  t35_n26: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t35_n60;
  t35_n27: if (E.getNumReferences() >= 3271098368 /*124.5*/) goto t35_n59;
  t35_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t35_n58;
  t35_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t35_n51;
  t35_n30: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t35_n50;
  t35_n31: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t35_n49;
  t35_n32: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n42;
  t35_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t35_n37;
  t35_n34: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t35_n36;
  t35_n35: return 0.01378543395549059f;
  t35_n36: return 0.04978173226118088f;
  t35_n37: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t35_n39;
  t35_n38: return -0.0003435408289078623f;
  t35_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t35_n41;
  t35_n40: return 0.015557238832116127f;
  t35_n41: return 0.0596640408039093f;
  t35_n42: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t35_n44;
  t35_n43: return 0.0022975034080445766f;
  t35_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n48;
  t35_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t35_n47;
  t35_n46: return -0.09560805559158325f;
  t35_n47: return 0.03957536444067955f;
  t35_n48: return 0.052630916237831116f;
  t35_n49: return -0.08436673134565353f;
  t35_n50: return -0.014264368452131748f;
  t35_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n53;
  t35_n52: return -0.06481462717056274f;
  t35_n53: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t35_n57;
  t35_n54: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n56;
  t35_n55: return 0.01961001753807068f;
  t35_n56: return -0.07740701735019684f;
  t35_n57: return 0.08710569143295288f;
  t35_n58: return -0.09554161876440048f;
  t35_n59: return 0.038215138018131256f;
  t35_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t35_n66;
  t35_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t35_n65;
  t35_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n64;
  t35_n63: return -0.008571881800889969f;
  t35_n64: return 0.047986168414354324f;
  t35_n65: return 0.03434176370501518f;
  t35_n66: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t35_n68;
  t35_n67: return 0.04714515432715416f;
  t35_n68: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n70;
  t35_n69: return -0.027877984568476677f;
  t35_n70: return 0.04307309538125992f;
  t35_n71: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n83;
  t35_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t35_n80;
  t35_n73: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t35_n77;
  t35_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n76;
  t35_n75: return 0.008847581222653389f;
  t35_n76: return 0.047878507524728775f;
  t35_n77: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n79;
  t35_n78: return -0.016998395323753357f;
  t35_n79: return 0.03085201047360897f;
  t35_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t35_n82;
  t35_n81: return -0.11212589591741562f;
  t35_n82: return -0.057734835892915726f;
  t35_n83: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t35_n85;
  t35_n84: return -0.023410527035593987f;
  t35_n85: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t35_n93;
  t35_n86: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t35_n92;
  t35_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t35_n89;
  t35_n88: return -0.04774480313062668f;
  t35_n89: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t35_n91;
  t35_n90: return 0.0378069244325161f;
  t35_n91: return -0.04214123263955116f;
  t35_n92: return 0.033993568271398544f;
  t35_n93: return -0.15291865170001984f;
  t35_n94: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t35_n102;
  t35_n95: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t35_n101;
  t35_n96: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n100;
  t35_n97: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t35_n99;
  t35_n98: return 0.043832309544086456f;
  t35_n99: return 0.014322961680591106f;
  t35_n100: return 0.07470240443944931f;
  t35_n101: return 0.0782558023929596f;
  t35_n102: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t35_n108;
  t35_n103: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n107;
  t35_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t35_n106;
  t35_n105: return -0.0012069129152223468f;
  t35_n106: return 0.04005962237715721f;
  t35_n107: return 0.0632399320602417f;
  t35_n108: return 0.06253102421760559f;
  t35_n109: if (E.getNumReferences() >= 3292798976 /*784.5*/) goto t35_n241;
  t35_n110: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t35_n182;
  t35_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t35_n175;
  t35_n112: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n168;
  t35_n113: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t35_n161;
  t35_n114: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t35_n122;
  t35_n115: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t35_n119;
  t35_n116: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t35_n118;
  t35_n117: return 0.02520519308745861f;
  t35_n118: return 0.046497661620378494f;
  t35_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t35_n121;
  t35_n120: return -0.023274756968021393f;
  t35_n121: return 0.08426450937986374f;
  t35_n122: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t35_n130;
  t35_n123: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t35_n127;
  t35_n124: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t35_n126;
  t35_n125: return 0.006289051845669746f;
  t35_n126: return 0.049158502370119095f;
  t35_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n129;
  t35_n128: return -0.016206473112106323f;
  t35_n129: return 0.046308208256959915f;
  t35_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t35_n132;
  t35_n131: return -0.07154319435358047f;
  t35_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t35_n146;
  t35_n133: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t35_n145;
  t35_n134: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t35_n144;
  t35_n135: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t35_n139;
  t35_n136: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t35_n138;
  t35_n137: return 0.012877125293016434f;
  t35_n138: return -0.02081962861120701f;
  t35_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n141;
  t35_n140: return -0.07754314690828323f;
  t35_n141: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n143;
  t35_n142: return 0.01372556947171688f;
  t35_n143: return 0.07481469959020615f;
  t35_n144: return -0.12337859719991684f;
  t35_n145: return -0.08518067747354507f;
  t35_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t35_n160;
  t35_n147: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t35_n159;
  t35_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t35_n154;
  t35_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n151;
  t35_n150: return -0.02096167951822281f;
  t35_n151: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t35_n153;
  t35_n152: return 0.031088268384337425f;
  t35_n153: return 0.01123654842376709f;
  t35_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t35_n158;
  t35_n155: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t35_n157;
  t35_n156: return 0.02657010778784752f;
  t35_n157: return 0.05572423338890076f;
  t35_n158: return 0.05288035050034523f;
  t35_n159: return -0.07837304472923279f;
  t35_n160: return -0.056444983929395676f;
  t35_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n163;
  t35_n162: return -0.11612223833799362f;
  t35_n163: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t35_n167;
  t35_n164: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t35_n166;
  t35_n165: return 0.013739049434661865f;
  t35_n166: return 0.07410223037004471f;
  t35_n167: return -0.015405667014420033f;
  t35_n168: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t35_n174;
  t35_n169: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n173;
  t35_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t35_n172;
  t35_n171: return 0.015111400745809078f;
  t35_n172: return 0.028132285922765732f;
  t35_n173: return 0.05159249156713486f;
  t35_n174: return -0.08315654844045639f;
  t35_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t35_n181;
  t35_n176: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t35_n180;
  t35_n177: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t35_n179;
  t35_n178: return 0.0443236269056797f;
  t35_n179: return -0.002906846348196268f;
  t35_n180: return 0.03707905486226082f;
  t35_n181: return 0.051149580627679825f;
  t35_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t35_n184;
  t35_n183: return -0.012929604388773441f;
  t35_n184: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n228;
  t35_n185: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t35_n221;
  t35_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t35_n212;
  t35_n187: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t35_n199;
  t35_n188: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t35_n198;
  t35_n189: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t35_n191;
  t35_n190: return 0.009634477086365223f;
  t35_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t35_n195;
  t35_n192: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t35_n194;
  t35_n193: return 0.0028946406673640013f;
  t35_n194: return 0.06960751861333847f;
  t35_n195: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t35_n197;
  t35_n196: return 0.03190566599369049f;
  t35_n197: return -0.0016052030259743333f;
  t35_n198: return 0.051015160977840424f;
  t35_n199: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t35_n211;
  t35_n200: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t35_n210;
  t35_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t35_n203;
  t35_n202: return 0.018675383180379868f;
  t35_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n205;
  t35_n204: return -0.06677711755037308f;
  t35_n205: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t35_n209;
  t35_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n208;
  t35_n207: return 0.12572512030601501f;
  t35_n208: return 0.001336421468295157f;
  t35_n209: return 0.05506046861410141f;
  t35_n210: return 0.05081329867243767f;
  t35_n211: return 0.04043731093406677f;
  t35_n212: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t35_n220;
  t35_n213: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t35_n217;
  t35_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t35_n216;
  t35_n215: return 0.042269375175237656f;
  t35_n216: return 0.06676039844751358f;
  t35_n217: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t35_n219;
  t35_n218: return 0.030898701399564743f;
  t35_n219: return 0.06587851792573929f;
  t35_n220: return -0.052271559834480286f;
  t35_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t35_n225;
  t35_n222: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t35_n224;
  t35_n223: return -0.03171607851982117f;
  t35_n224: return 0.044511470943689346f;
  t35_n225: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t35_n227;
  t35_n226: return 0.005873534362763166f;
  t35_n227: return 0.02977369725704193f;
  t35_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t35_n238;
  t35_n229: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t35_n231;
  t35_n230: return 0.04544701427221298f;
  t35_n231: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t35_n237;
  t35_n232: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t35_n236;
  t35_n233: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t35_n235;
  t35_n234: return -0.05589699745178223f;
  t35_n235: return 0.0992833822965622f;
  t35_n236: return 0.05365251004695892f;
  t35_n237: return 0.055192410945892334f;
  t35_n238: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t35_n240;
  t35_n239: return 0.040795039385557175f;
  t35_n240: return 0.07280713319778442f;
  t35_n241: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t35_n247;
  t35_n242: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t35_n244;
  t35_n243: return 0.04617500305175781f;
  t35_n244: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t35_n246;
  t35_n245: return 0.11261248588562012f;
  t35_n246: return 0.08506985008716583f;
  t35_n247: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t35_n249;
  t35_n248: return 0.025195058435201645f;
  t35_n249: return 0.0729842483997345f;
  t35_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t35_n254;
  t35_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t35_n253;
  t35_n252: return 0.04381062462925911f;
  t35_n253: return 0.06225614249706268f;
  t35_n254: return 0.08753515779972076f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree36(const Example& E) {
  t36_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t36_n252;
  t36_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t36_n103;
  t36_n2: if (E.getNumReferences() >= 3303280640 /*1824.5*/) goto t36_n92;
  t36_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t36_n75;
  t36_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t36_n54;
  t36_n5: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t36_n17;
  t36_n6: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t36_n12;
  t36_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t36_n11;
  t36_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n10;
  t36_n9: return -0.04501238837838173f;
  t36_n10: return -0.011010309681296349f;
  t36_n11: return 0.033996839076280594f;
  t36_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n16;
  t36_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n15;
  t36_n14: return 0.014569074846804142f;
  t36_n15: return 0.04108286276459694f;
  t36_n16: return 0.07896331697702408f;
  t36_n17: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t36_n49;
  t36_n18: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t36_n48;
  t36_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t36_n21;
  t36_n20: return -0.05346804857254028f;
  t36_n21: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n45;
  t36_n22: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t36_n24;
  t36_n23: return -0.039143435657024384f;
  t36_n24: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t36_n44;
  t36_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t36_n27;
  t36_n26: return -0.007715218234807253f;
  t36_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t36_n31;
  t36_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t36_n30;
  t36_n29: return -0.03454345837235451f;
  t36_n30: return 0.09975942969322205f;
  t36_n31: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t36_n43;
  t36_n32: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t36_n42;
  t36_n33: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t36_n37;
  t36_n34: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t36_n36;
  t36_n35: return 0.023310508579015732f;
  t36_n36: return -0.047776635736227036f;
  t36_n37: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t36_n39;
  t36_n38: return 0.03002280183136463f;
  t36_n39: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t36_n41;
  t36_n40: return 0.0498538576066494f;
  t36_n41: return 0.0011421693488955498f;
  t36_n42: return -0.1219191774725914f;
  t36_n43: return -0.12058460712432861f;
  t36_n44: return -0.06326289474964142f;
  t36_n45: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t36_n47;
  t36_n46: return 0.02508951723575592f;
  t36_n47: return -0.052073199301958084f;
  t36_n48: return -0.12421020120382309f;
  t36_n49: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t36_n51;
  t36_n50: return -0.09790179878473282f;
  t36_n51: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t36_n53;
  t36_n52: return 0.0042633493430912495f;
  t36_n53: return -0.06106987223029137f;
  t36_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t36_n58;
  t36_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n57;
  t36_n56: return -0.15756341814994812f;
  t36_n57: return -0.0798916295170784f;
  t36_n58: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t36_n64;
  t36_n59: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t36_n63;
  t36_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n62;
  t36_n61: return -0.026333119720220566f;
  t36_n62: return 0.009144352748990059f;
  t36_n63: return 0.03428799659013748f;
  t36_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t36_n66;
  t36_n65: return -0.0033951217774301767f;
  t36_n66: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t36_n72;
  t36_n67: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t36_n71;
  t36_n68: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t36_n70;
  t36_n69: return 0.011525182984769344f;
  t36_n70: return 0.1250440925359726f;
  t36_n71: return -0.026287438347935677f;
  t36_n72: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t36_n74;
  t36_n73: return 0.052236296236515045f;
  t36_n74: return -0.08399802446365356f;
  t36_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n87;
  t36_n76: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t36_n84;
  t36_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n81;
  t36_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t36_n80;
  t36_n79: return -0.0021084577310830355f;
  t36_n80: return 0.01358719076961279f;
  t36_n81: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t36_n83;
  t36_n82: return 0.019254444167017937f;
  t36_n83: return -0.087277352809906f;
  t36_n84: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t36_n86;
  t36_n85: return -0.035290297120809555f;
  t36_n86: return 0.04103948175907135f;
  t36_n87: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t36_n91;
  t36_n88: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t36_n90;
  t36_n89: return 0.024120446294546127f;
  t36_n90: return -0.007206630427390337f;
  t36_n91: return 0.06259796023368835f;
  t36_n92: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t36_n98;
  t36_n93: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t36_n95;
  t36_n94: return 0.0312366783618927f;
  t36_n95: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n97;
  t36_n96: return 0.05834106355905533f;
  t36_n97: return 0.10293307900428772f;
  t36_n98: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t36_n100;
  t36_n99: return -2.9740635000052862e-05f;
  t36_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t36_n102;
  t36_n101: return 0.002971682930365205f;
  t36_n102: return 0.06333186477422714f;
  t36_n103: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t36_n177;
  t36_n104: if (E.getNumReferences() >= 3300708352 /*1510.5*/) goto t36_n170;
  t36_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t36_n165;
  t36_n106: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t36_n136;
  t36_n107: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t36_n131;
  t36_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t36_n124;
  t36_n109: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t36_n123;
  t36_n110: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t36_n122;
  t36_n111: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t36_n121;
  t36_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n116;
  t36_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t36_n115;
  t36_n114: return -0.1712620109319687f;
  t36_n115: return -0.10805050283670425f;
  t36_n116: if (E.getFileProximityDistanceCost() >= 3255042048 /*33.0*/) goto t36_n118;
  t36_n117: return -0.0059481277130544186f;
  t36_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n120;
  t36_n119: return 0.03480498492717743f;
  t36_n120: return 0.11627006530761719f;
  t36_n121: return 0.00782120693475008f;
  t36_n122: return 0.07173950225114822f;
  t36_n123: return -0.07427900284528732f;
  t36_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t36_n126;
  t36_n125: return -0.017099512740969658f;
  t36_n126: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t36_n128;
  t36_n127: return 0.016776634380221367f;
  t36_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t36_n130;
  t36_n129: return 0.019525792449712753f;
  t36_n130: return 0.05264142155647278f;
  t36_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t36_n133;
  t36_n132: return 0.024195579811930656f;
  t36_n133: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t36_n135;
  t36_n134: return 0.0304782222956419f;
  t36_n135: return 0.053811706602573395f;
  t36_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t36_n142;
  t36_n137: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n139;
  t36_n138: return -0.02341323532164097f;
  t36_n139: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t36_n141;
  t36_n140: return 0.01720242388546467f;
  t36_n141: return -0.21697553992271423f;
  t36_n142: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t36_n164;
  t36_n143: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t36_n161;
  t36_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t36_n146;
  t36_n145: return -0.0062955026514828205f;
  t36_n146: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t36_n148;
  t36_n147: return 0.004527022130787373f;
  t36_n148: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t36_n150;
  t36_n149: return 0.0008886450086720288f;
  t36_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t36_n160;
  t36_n151: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t36_n159;
  t36_n152: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t36_n158;
  t36_n153: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t36_n155;
  t36_n154: return 0.0403033122420311f;
  t36_n155: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t36_n157;
  t36_n156: return 0.0674615278840065f;
  t36_n157: return -0.02585485950112343f;
  t36_n158: return -0.13168084621429443f;
  t36_n159: return -0.11867154389619827f;
  t36_n160: return 0.07246478646993637f;
  t36_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n163;
  t36_n162: return -0.44007933139801025f;
  t36_n163: return 0.05428774654865265f;
  t36_n164: return 0.04072977975010872f;
  t36_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t36_n169;
  t36_n166: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t36_n168;
  t36_n167: return 0.010132981464266777f;
  t36_n168: return 0.03686661273241043f;
  t36_n169: return 0.048216186463832855f;
  t36_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t36_n174;
  t36_n171: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t36_n173;
  t36_n172: return 0.018091129139065742f;
  t36_n173: return 0.0676976665854454f;
  t36_n174: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t36_n176;
  t36_n175: return 0.04830782115459442f;
  t36_n176: return 0.08420298993587494f;
  t36_n177: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t36_n245;
  t36_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t36_n180;
  t36_n179: return -0.010508334264159203f;
  t36_n180: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t36_n204;
  t36_n181: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t36_n195;
  t36_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t36_n190;
  t36_n183: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t36_n189;
  t36_n184: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t36_n188;
  t36_n185: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t36_n187;
  t36_n186: return 0.02294936217367649f;
  t36_n187: return -0.01894136518239975f;
  t36_n188: return 0.037484075874090195f;
  t36_n189: return -0.00023479803348891437f;
  t36_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t36_n192;
  t36_n191: return 0.014435481280088425f;
  t36_n192: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t36_n194;
  t36_n193: return 0.06925618648529053f;
  t36_n194: return 0.029947523027658463f;
  t36_n195: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t36_n203;
  t36_n196: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t36_n198;
  t36_n197: return 0.02677667886018753f;
  t36_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t36_n200;
  t36_n199: return 0.012570898979902267f;
  t36_n200: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t36_n202;
  t36_n201: return 0.052517011761665344f;
  t36_n202: return 0.08033675700426102f;
  t36_n203: return 0.08181554824113846f;
  t36_n204: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t36_n222;
  t36_n205: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t36_n215;
  t36_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n208;
  t36_n207: return -0.03410341963171959f;
  t36_n208: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t36_n214;
  t36_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t36_n213;
  t36_n210: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t36_n212;
  t36_n211: return 0.039171572774648666f;
  t36_n212: return -0.011363762430846691f;
  t36_n213: return 0.09241504967212677f;
  t36_n214: return 0.016225995495915413f;
  t36_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t36_n221;
  t36_n216: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t36_n218;
  t36_n217: return 0.012164687737822533f;
  t36_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n220;
  t36_n219: return -0.017078857868909836f;
  t36_n220: return 0.04076504707336426f;
  t36_n221: return 0.057034846395254135f;
  t36_n222: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t36_n244;
  t36_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t36_n227;
  t36_n224: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t36_n226;
  t36_n225: return 0.04048306867480278f;
  t36_n226: return -0.6174516081809998f;
  t36_n227: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t36_n237;
  t36_n228: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t36_n232;
  t36_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t36_n231;
  t36_n230: return 0.029912736266851425f;
  t36_n231: return 0.061675213277339935f;
  t36_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t36_n234;
  t36_n233: return -0.20576806366443634f;
  t36_n234: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t36_n236;
  t36_n235: return -0.0015060023870319128f;
  t36_n236: return 0.10538676381111145f;
  t36_n237: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t36_n243;
  t36_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n242;
  t36_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t36_n241;
  t36_n240: return 0.03264966234564781f;
  t36_n241: return 0.0690142959356308f;
  t36_n242: return 0.073092982172966f;
  t36_n243: return 0.09806692600250244f;
  t36_n244: return -0.06875789910554886f;
  t36_n245: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t36_n247;
  t36_n246: return 0.07266838103532791f;
  t36_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t36_n251;
  t36_n248: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t36_n250;
  t36_n249: return 0.04226282611489296f;
  t36_n250: return 0.06363706290721893f;
  t36_n251: return 0.09786312282085419f;
  t36_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t36_n254;
  t36_n253: return 0.05338817834854126f;
  t36_n254: return 0.08580446243286133f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree37(const Example& E) {
  t37_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t37_n252;
  t37_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t37_n105;
  t37_n2: if (E.getNumReferences() >= 3303288832 /*1825.5*/) goto t37_n84;
  t37_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t37_n63;
  t37_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n30;
  t37_n5: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t37_n29;
  t37_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t37_n10;
  t37_n7: if (E.getNumReferences() >= 3268739072 /*106.5*/) goto t37_n9;
  t37_n8: return -0.038498926907777786f;
  t37_n9: return 0.05295413359999657f;
  t37_n10: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t37_n28;
  t37_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n27;
  t37_n12: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t37_n16;
  t37_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t37_n15;
  t37_n14: return -0.00869380496442318f;
  t37_n15: return -0.16136790812015533f;
  t37_n16: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t37_n22;
  t37_n17: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n19;
  t37_n18: return -0.13522489368915558f;
  t37_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n21;
  t37_n20: return -0.12560275197029114f;
  t37_n21: return 0.00856646429747343f;
  t37_n22: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t37_n26;
  t37_n23: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n25;
  t37_n24: return -0.008902688510715961f;
  t37_n25: return 0.05689748376607895f;
  t37_n26: return 0.045470334589481354f;
  t37_n27: return -0.03806084766983986f;
  t37_n28: return -0.13125889003276825f;
  t37_n29: return -0.08030454814434052f;
  t37_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t37_n32;
  t37_n31: return -0.08671397715806961f;
  t37_n32: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n56;
  t37_n33: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t37_n55;
  t37_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t37_n50;
  t37_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t37_n49;
  t37_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t37_n42;
  t37_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n39;
  t37_n38: return -0.08459402620792389f;
  t37_n39: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t37_n41;
  t37_n40: return -0.002413123147562146f;
  t37_n41: return 0.09945756942033768f;
  t37_n42: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t37_n48;
  t37_n43: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t37_n45;
  t37_n44: return 0.015849880874156952f;
  t37_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t37_n47;
  t37_n46: return 0.02363646775484085f;
  t37_n47: return 0.08405490964651108f;
  t37_n48: return 0.011819890700280666f;
  t37_n49: return -0.010178141295909882f;
  t37_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n52;
  t37_n51: return -0.07085960358381271f;
  t37_n52: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t37_n54;
  t37_n53: return 0.018612874671816826f;
  t37_n54: return 0.0837734043598175f;
  t37_n55: return -0.09308470785617828f;
  t37_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t37_n60;
  t37_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t37_n59;
  t37_n58: return 0.003569123102352023f;
  t37_n59: return 0.033059872686862946f;
  t37_n60: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n62;
  t37_n61: return 0.04565304517745972f;
  t37_n62: return 0.01596042886376381f;
  t37_n63: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t37_n73;
  t37_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t37_n66;
  t37_n65: return -0.05910969525575638f;
  t37_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n68;
  t37_n67: return -0.016004079952836037f;
  t37_n68: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t37_n70;
  t37_n69: return -0.035416379570961f;
  t37_n70: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n72;
  t37_n71: return 0.03023000992834568f;
  t37_n72: return -0.14942173659801483f;
  t37_n73: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t37_n83;
  t37_n74: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t37_n76;
  t37_n75: return -0.04749844968318939f;
  t37_n76: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n82;
  t37_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n81;
  t37_n78: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t37_n80;
  t37_n79: return 0.006286883261054754f;
  t37_n80: return 0.03360774740576744f;
  t37_n81: return 0.06583207845687866f;
  t37_n82: return 0.0710177943110466f;
  t37_n83: return -0.07143907994031906f;
  t37_n84: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t37_n98;
  t37_n85: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t37_n95;
  t37_n86: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t37_n94;
  t37_n87: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n89;
  t37_n88: return 0.03132378309965134f;
  t37_n89: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t37_n91;
  t37_n90: return 0.020219236612319946f;
  t37_n91: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t37_n93;
  t37_n92: return 0.13302487134933472f;
  t37_n93: return 0.06329736858606339f;
  t37_n94: return -0.042654283344745636f;
  t37_n95: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n97;
  t37_n96: return 0.05616361275315285f;
  t37_n97: return 0.10142152011394501f;
  t37_n98: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t37_n100;
  t37_n99: return -0.000894030905328691f;
  t37_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t37_n104;
  t37_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t37_n103;
  t37_n102: return -0.02166450023651123f;
  t37_n103: return 0.038476962596178055f;
  t37_n104: return 0.0601351298391819f;
  t37_n105: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t37_n217;
  t37_n106: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t37_n188;
  t37_n107: if (E.getNumReferences() >= 3292798976 /*784.5*/) goto t37_n181;
  t37_n108: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t37_n126;
  t37_n109: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t37_n121;
  t37_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t37_n118;
  t37_n111: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t37_n115;
  t37_n112: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t37_n114;
  t37_n113: return 0.018618110567331314f;
  t37_n114: return -0.011358898133039474f;
  t37_n115: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t37_n117;
  t37_n116: return 0.028140418231487274f;
  t37_n117: return 0.06517250835895538f;
  t37_n118: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t37_n120;
  t37_n119: return 0.05345406383275986f;
  t37_n120: return 0.034162748605012894f;
  t37_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t37_n125;
  t37_n122: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t37_n124;
  t37_n123: return 0.031701501458883286f;
  t37_n124: return 0.05832554027438164f;
  t37_n125: return 0.06130078807473183f;
  t37_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t37_n168;
  t37_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t37_n155;
  t37_n128: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t37_n154;
  t37_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t37_n143;
  t37_n130: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t37_n140;
  t37_n131: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n137;
  t37_n132: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t37_n134;
  t37_n133: return 0.0030599881429225206f;
  t37_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t37_n136;
  t37_n135: return -0.0015490028308704495f;
  t37_n136: return 0.07746588438749313f;
  t37_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n139;
  t37_n138: return -0.0668819397687912f;
  t37_n139: return 0.01321684755384922f;
  t37_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t37_n142;
  t37_n141: return 0.020104117691516876f;
  t37_n142: return 0.03539818525314331f;
  t37_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n147;
  t37_n144: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t37_n146;
  t37_n145: return -0.35923004150390625f;
  t37_n146: return 0.04731789603829384f;
  t37_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t37_n149;
  t37_n148: return 0.05317017808556557f;
  t37_n149: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t37_n153;
  t37_n150: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n152;
  t37_n151: return 0.008725282736122608f;
  t37_n152: return 0.07000444829463959f;
  t37_n153: return 0.04590742662549019f;
  t37_n154: return -0.08343073725700378f;
  t37_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t37_n157;
  t37_n156: return -0.13385847210884094f;
  t37_n157: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t37_n165;
  t37_n158: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t37_n162;
  t37_n159: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t37_n161;
  t37_n160: return 0.0098398607224226f;
  t37_n161: return 0.05190228670835495f;
  t37_n162: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t37_n164;
  t37_n163: return 0.04380863904953003f;
  t37_n164: return -0.04167594760656357f;
  t37_n165: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t37_n167;
  t37_n166: return 0.04871078208088875f;
  t37_n167: return -0.03510476276278496f;
  t37_n168: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n178;
  t37_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t37_n171;
  t37_n170: return -0.02431558631360531f;
  t37_n171: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t37_n177;
  t37_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t37_n176;
  t37_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t37_n175;
  t37_n174: return 0.0013504633679986f;
  t37_n175: return 0.01765197142958641f;
  t37_n176: return 0.04275256022810936f;
  t37_n177: return 0.030633466318249702f;
  t37_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t37_n180;
  t37_n179: return 0.019410761073231697f;
  t37_n180: return 0.05487745255231857f;
  t37_n181: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t37_n183;
  t37_n182: return 0.06527573615312576f;
  t37_n183: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t37_n185;
  t37_n184: return 0.038043826818466187f;
  t37_n185: if (E.getNumReferences() >= 3325638144 /*11853.5*/) goto t37_n187;
  t37_n186: return 0.03752902150154114f;
  t37_n187: return 0.08570234477519989f;
  t37_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n190;
  t37_n189: return -0.12269246578216553f;
  t37_n190: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t37_n196;
  t37_n191: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t37_n195;
  t37_n192: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t37_n194;
  t37_n193: return 0.012603227980434895f;
  t37_n194: return 0.031360555440187454f;
  t37_n195: return 0.06459841877222061f;
  t37_n196: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t37_n214;
  t37_n197: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t37_n211;
  t37_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t37_n206;
  t37_n199: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n205;
  t37_n200: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n204;
  t37_n201: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t37_n203;
  t37_n202: return -0.00610036076977849f;
  t37_n203: return -0.07709122449159622f;
  t37_n204: return 0.010693681426346302f;
  t37_n205: return -0.07932504266500473f;
  t37_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t37_n208;
  t37_n207: return 0.02873181365430355f;
  t37_n208: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t37_n210;
  t37_n209: return 0.03726251423358917f;
  t37_n210: return -0.01130911335349083f;
  t37_n211: if (E.getNumReferences() >= 3344720384 /*56418.0*/) goto t37_n213;
  t37_n212: return 0.10453484952449799f;
  t37_n213: return 0.03627874329686165f;
  t37_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t37_n216;
  t37_n215: return -0.026800353080034256f;
  t37_n216: return -0.11981663852930069f;
  t37_n217: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t37_n251;
  t37_n218: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t37_n240;
  t37_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t37_n229;
  t37_n220: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t37_n228;
  t37_n221: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t37_n225;
  t37_n222: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t37_n224;
  t37_n223: return 0.035180870443582535f;
  t37_n224: return 0.14821581542491913f;
  t37_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t37_n227;
  t37_n226: return -0.029572846367955208f;
  t37_n227: return 0.05043608695268631f;
  t37_n228: return 0.07605192065238953f;
  t37_n229: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t37_n233;
  t37_n230: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t37_n232;
  t37_n231: return 0.04933791235089302f;
  t37_n232: return 0.022320736199617386f;
  t37_n233: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t37_n237;
  t37_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t37_n236;
  t37_n235: return 0.04983368143439293f;
  t37_n236: return 0.08598984032869339f;
  t37_n237: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t37_n239;
  t37_n238: return -0.16354183852672577f;
  t37_n239: return 0.04094022884964943f;
  t37_n240: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t37_n248;
  t37_n241: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t37_n247;
  t37_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t37_n246;
  t37_n243: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t37_n245;
  t37_n244: return 0.01153060607612133f;
  t37_n245: return -0.07793349027633667f;
  t37_n246: return 0.03521772474050522f;
  t37_n247: return 0.05671711266040802f;
  t37_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t37_n250;
  t37_n249: return -0.08133767545223236f;
  t37_n250: return 0.05934561416506767f;
  t37_n251: return 0.09552828967571259f;
  t37_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t37_n254;
  t37_n253: return 0.046028994023799896f;
  t37_n254: return 0.06409329175949097f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree38(const Example& E) {
  t38_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t38_n250;
  t38_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t38_n145;
  t38_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t38_n84;
  t38_n3: if (E.getNumReferences() >= 3317122048 /*5864.5*/) goto t38_n69;
  t38_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n34;
  t38_n5: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t38_n31;
  t38_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t38_n30;
  t38_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t38_n9;
  t38_n8: return -0.03884991258382797f;
  t38_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t38_n29;
  t38_n10: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t38_n14;
  t38_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n13;
  t38_n12: return -0.01421373337507248f;
  t38_n13: return 0.019956080242991447f;
  t38_n14: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t38_n24;
  t38_n15: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t38_n17;
  t38_n16: return -0.13453985750675201f;
  t38_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n19;
  t38_n18: return -0.12506277859210968f;
  t38_n19: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t38_n23;
  t38_n20: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t38_n22;
  t38_n21: return 0.00805643666535616f;
  t38_n22: return -0.14688478410243988f;
  t38_n23: return -0.06543360650539398f;
  t38_n24: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t38_n28;
  t38_n25: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n27;
  t38_n26: return -0.018623190000653267f;
  t38_n27: return 0.061966825276613235f;
  t38_n28: return 0.04302186518907547f;
  t38_n29: return -0.04073196277022362f;
  t38_n30: return -0.06423905491828918f;
  t38_n31: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t38_n33;
  t38_n32: return 0.047681279480457306f;
  t38_n33: return 0.00248690671287477f;
  t38_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t38_n36;
  t38_n35: return -0.08481971174478531f;
  t38_n36: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t38_n64;
  t38_n37: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t38_n63;
  t38_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t38_n60;
  t38_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t38_n51;
  t38_n40: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t38_n50;
  t38_n41: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t38_n43;
  t38_n42: return 0.010956703685224056f;
  t38_n43: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t38_n45;
  t38_n44: return 0.0011003657709807158f;
  t38_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n49;
  t38_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t38_n48;
  t38_n47: return -0.08988677710294724f;
  t38_n48: return 0.038092296570539474f;
  t38_n49: return 0.04996141418814659f;
  t38_n50: return -0.07749388366937637f;
  t38_n51: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n53;
  t38_n52: return -0.01840163581073284f;
  t38_n53: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t38_n59;
  t38_n54: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t38_n58;
  t38_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t38_n57;
  t38_n56: return -0.03539295494556427f;
  t38_n57: return 0.03569317236542702f;
  t38_n58: return 0.0485004298388958f;
  t38_n59: return 0.06714963912963867f;
  t38_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n62;
  t38_n61: return -0.06478123366832733f;
  t38_n62: return 0.018734056502580643f;
  t38_n63: return -0.09131789952516556f;
  t38_n64: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n68;
  t38_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t38_n67;
  t38_n66: return 0.0077520194463431835f;
  t38_n67: return 0.03377729654312134f;
  t38_n68: return 0.05871393904089928f;
  t38_n69: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n83;
  t38_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t38_n78;
  t38_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t38_n75;
  t38_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t38_n74;
  t38_n73: return 0.026577072218060493f;
  t38_n74: return 0.06496082246303558f;
  t38_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t38_n77;
  t38_n76: return -0.06431648880243301f;
  t38_n77: return 0.053434453904628754f;
  t38_n78: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t38_n80;
  t38_n79: return 0.014023357070982456f;
  t38_n80: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t38_n82;
  t38_n81: return 0.07400906085968018f;
  t38_n82: return -0.0773804560303688f;
  t38_n83: return 0.07590950280427933f;
  t38_n84: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t38_n144;
  t38_n85: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t38_n137;
  t38_n86: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t38_n94;
  t38_n87: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t38_n89;
  t38_n88: return -0.0016964994138106704f;
  t38_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t38_n91;
  t38_n90: return 0.006409105379134417f;
  t38_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t38_n93;
  t38_n92: return 0.04249434545636177f;
  t38_n93: return 0.11286884546279907f;
  t38_n94: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t38_n104;
  t38_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t38_n97;
  t38_n96: return -0.021395951509475708f;
  t38_n97: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t38_n99;
  t38_n98: return -0.10063536465167999f;
  t38_n99: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t38_n103;
  t38_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t38_n102;
  t38_n101: return 0.005233341362327337f;
  t38_n102: return 0.0686052143573761f;
  t38_n103: return -0.06922652572393417f;
  t38_n104: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n136;
  t38_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t38_n135;
  t38_n106: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t38_n118;
  t38_n107: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t38_n117;
  t38_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t38_n110;
  t38_n109: return -0.03608093410730362f;
  t38_n110: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t38_n112;
  t38_n111: return 0.03441697731614113f;
  t38_n112: if (E.getNumReferences() >= 3355447552 /*131140.0*/) goto t38_n116;
  t38_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n115;
  t38_n114: return -0.1381109356880188f;
  t38_n115: return 0.13254736363887787f;
  t38_n116: return -0.03540786728262901f;
  t38_n117: return -0.06928054988384247f;
  t38_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t38_n128;
  t38_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t38_n127;
  t38_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n122;
  t38_n121: return -0.08337754756212234f;
  t38_n122: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t38_n124;
  t38_n123: return 0.005749038886278868f;
  t38_n124: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t38_n126;
  t38_n125: return 0.00442351121455431f;
  t38_n126: return 0.09484084695577621f;
  t38_n127: return 0.013005226850509644f;
  t38_n128: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t38_n134;
  t38_n129: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t38_n133;
  t38_n130: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t38_n132;
  t38_n131: return 0.05725967884063721f;
  t38_n132: return 0.0005052855121903121f;
  t38_n133: return 0.15880252420902252f;
  t38_n134: return -0.08449260890483856f;
  t38_n135: return 0.040493473410606384f;
  t38_n136: return 0.04017328470945358f;
  t38_n137: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t38_n143;
  t38_n138: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t38_n142;
  t38_n139: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t38_n141;
  t38_n140: return 0.012566056102514267f;
  t38_n141: return -0.07502108067274094f;
  t38_n142: return -0.09376545250415802f;
  t38_n143: return -0.07367213070392609f;
  t38_n144: return 0.05621909350156784f;
  t38_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t38_n203;
  t38_n146: if (E.getNumReferences() >= 3300708352 /*1510.5*/) goto t38_n196;
  t38_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t38_n185;
  t38_n148: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n182;
  t38_n149: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t38_n161;
  t38_n150: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t38_n158;
  t38_n151: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t38_n157;
  t38_n152: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t38_n156;
  t38_n153: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t38_n155;
  t38_n154: return 0.01382796373218298f;
  t38_n155: return 0.027203841134905815f;
  t38_n156: return -0.01917012594640255f;
  t38_n157: return 0.032489076256752014f;
  t38_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t38_n160;
  t38_n159: return 0.04765276238322258f;
  t38_n160: return 0.07253538817167282f;
  t38_n161: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t38_n175;
  t38_n162: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t38_n172;
  t38_n163: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t38_n169;
  t38_n164: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t38_n166;
  t38_n165: return -0.013257583603262901f;
  t38_n166: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t38_n168;
  t38_n167: return 0.0054583800956606865f;
  t38_n168: return -0.07966621965169907f;
  t38_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t38_n171;
  t38_n170: return 0.0064859893172979355f;
  t38_n171: return 0.06557302176952362f;
  t38_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n174;
  t38_n173: return -0.07426615804433823f;
  t38_n174: return 0.0004963764222338796f;
  t38_n175: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t38_n181;
  t38_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t38_n180;
  t38_n177: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t38_n179;
  t38_n178: return 0.02392416261136532f;
  t38_n179: return 0.05837130919098854f;
  t38_n180: return 0.061356693506240845f;
  t38_n181: return -0.03213227167725563f;
  t38_n182: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t38_n184;
  t38_n183: return 0.020056985318660736f;
  t38_n184: return -0.07941450923681259f;
  t38_n185: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t38_n195;
  t38_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t38_n188;
  t38_n187: return -0.011714396998286247f;
  t38_n188: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t38_n192;
  t38_n189: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t38_n191;
  t38_n190: return 0.03460879251360893f;
  t38_n191: return 0.09154312312602997f;
  t38_n192: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t38_n194;
  t38_n193: return 0.06460341811180115f;
  t38_n194: return 0.008970405906438828f;
  t38_n195: return 0.036492228507995605f;
  t38_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t38_n200;
  t38_n197: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t38_n199;
  t38_n198: return 0.015908170491456985f;
  t38_n199: return 0.06453903764486313f;
  t38_n200: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t38_n202;
  t38_n201: return 0.04649034142494202f;
  t38_n202: return 0.08139998465776443f;
  t38_n203: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t38_n243;
  t38_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t38_n206;
  t38_n205: return -0.006345563102513552f;
  t38_n206: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n228;
  t38_n207: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t38_n223;
  t38_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t38_n218;
  t38_n209: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t38_n211;
  t38_n210: return 0.013589619658887386f;
  t38_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t38_n215;
  t38_n212: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t38_n214;
  t38_n213: return 0.015925699844956398f;
  t38_n214: return 0.04402054101228714f;
  t38_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n217;
  t38_n216: return -0.03767475113272667f;
  t38_n217: return 0.03590850904583931f;
  t38_n218: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t38_n222;
  t38_n219: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t38_n221;
  t38_n220: return 0.03947022184729576f;
  t38_n221: return 0.028909223154187202f;
  t38_n222: return -0.048969075083732605f;
  t38_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t38_n225;
  t38_n224: return -0.024382688105106354f;
  t38_n225: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t38_n227;
  t38_n226: return 0.004447092767804861f;
  t38_n227: return 0.026364294812083244f;
  t38_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t38_n240;
  t38_n229: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t38_n233;
  t38_n230: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t38_n232;
  t38_n231: return 0.030706441029906273f;
  t38_n232: return 0.06051429733633995f;
  t38_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t38_n239;
  t38_n234: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t38_n238;
  t38_n235: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t38_n237;
  t38_n236: return -0.047823574393987656f;
  t38_n237: return 0.09255746752023697f;
  t38_n238: return 0.048624783754348755f;
  t38_n239: return 0.051367636770009995f;
  t38_n240: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t38_n242;
  t38_n241: return 0.04034016653895378f;
  t38_n242: return 0.09508908540010452f;
  t38_n243: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t38_n245;
  t38_n244: return 0.06925416737794876f;
  t38_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t38_n249;
  t38_n246: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t38_n248;
  t38_n247: return 0.0399092435836792f;
  t38_n248: return 0.06090112775564194f;
  t38_n249: return 0.09495051205158234f;
  t38_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t38_n254;
  t38_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t38_n253;
  t38_n252: return 0.03720444068312645f;
  t38_n253: return 0.056517988443374634f;
  t38_n254: return 0.0836578980088234f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree39(const Example& E) {
  t39_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t39_n252;
  t39_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t39_n141;
  t39_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t39_n64;
  t39_n3: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t39_n51;
  t39_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n28;
  t39_n5: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t39_n27;
  t39_n6: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t39_n24;
  t39_n7: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t39_n23;
  t39_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t39_n10;
  t39_n9: return -0.03526604548096657f;
  t39_n10: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t39_n22;
  t39_n11: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t39_n13;
  t39_n12: return -0.010294876992702484f;
  t39_n13: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t39_n17;
  t39_n14: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n16;
  t39_n15: return -0.12812012434005737f;
  t39_n16: return 0.004592789337038994f;
  t39_n17: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t39_n21;
  t39_n18: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t39_n20;
  t39_n19: return -0.018118005245923996f;
  t39_n20: return 0.059398770332336426f;
  t39_n21: return 0.04036540538072586f;
  t39_n22: return -0.038427457213401794f;
  t39_n23: return -0.06054059788584709f;
  t39_n24: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t39_n26;
  t39_n25: return 0.04571901634335518f;
  t39_n26: return 1.2111920113966335e-05f;
  t39_n27: return -0.0898296907544136f;
  t39_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t39_n30;
  t39_n29: return -0.08271756023168564f;
  t39_n30: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n48;
  t39_n31: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t39_n47;
  t39_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t39_n46;
  t39_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t39_n45;
  t39_n34: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t39_n44;
  t39_n35: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t39_n37;
  t39_n36: return 0.010174395516514778f;
  t39_n37: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t39_n39;
  t39_n38: return 0.0029658018611371517f;
  t39_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n43;
  t39_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t39_n42;
  t39_n41: return -0.08022761344909668f;
  t39_n42: return 0.03578754514455795f;
  t39_n43: return 0.046851422637701035f;
  t39_n44: return -0.07152987271547318f;
  t39_n45: return -0.0084593640640378f;
  t39_n46: return 0.016579121351242065f;
  t39_n47: return -0.08978376537561417f;
  t39_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t39_n50;
  t39_n49: return 0.014360107481479645f;
  t39_n50: return 0.03421521931886673f;
  t39_n51: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t39_n59;
  t39_n52: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t39_n58;
  t39_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t39_n55;
  t39_n54: return 0.02335548773407936f;
  t39_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t39_n57;
  t39_n56: return 0.05117752030491829f;
  t39_n57: return -0.05622373893857002f;
  t39_n58: return 0.06309192627668381f;
  t39_n59: if (E.getNumReferences() >= 3309029376 /*3004.5*/) goto t39_n63;
  t39_n60: if (E.getNumReferences() >= 3308949504 /*2985.0*/) goto t39_n62;
  t39_n61: return 0.07466094940900803f;
  t39_n62: return 0.15750205516815186f;
  t39_n63: return 0.06731688976287842f;
  t39_n64: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t39_n140;
  t39_n65: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t39_n135;
  t39_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t39_n118;
  t39_n67: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t39_n113;
  t39_n68: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t39_n82;
  t39_n69: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t39_n81;
  t39_n70: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t39_n72;
  t39_n71: return -0.0271470807492733f;
  t39_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n74;
  t39_n73: return 0.0011697731679305434f;
  t39_n74: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t39_n76;
  t39_n75: return 0.036753199994564056f;
  t39_n76: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t39_n80;
  t39_n77: if (E.getNumReferences() >= 3355438208 /*131033.0*/) goto t39_n79;
  t39_n78: return 0.12643268704414368f;
  t39_n79: return 0.06972407549619675f;
  t39_n80: return 0.02875063382089138f;
  t39_n81: return -0.07676992565393448f;
  t39_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t39_n106;
  t39_n83: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t39_n85;
  t39_n84: return -0.035866137593984604f;
  t39_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t39_n89;
  t39_n86: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t39_n88;
  t39_n87: return -0.02971099317073822f;
  t39_n88: return 0.08394132554531097f;
  t39_n89: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t39_n105;
  t39_n90: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t39_n104;
  t39_n91: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t39_n103;
  t39_n92: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t39_n94;
  t39_n93: return 0.04393762722611427f;
  t39_n94: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t39_n96;
  t39_n95: return -0.05588605999946594f;
  t39_n96: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t39_n98;
  t39_n97: return 0.05272720381617546f;
  t39_n98: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t39_n100;
  t39_n99: return -0.07395267486572266f;
  t39_n100: if (E.getFileProximityDistanceCost() >= 3254910976 /*32.5*/) goto t39_n102;
  t39_n101: return 0.03665580227971077f;
  t39_n102: return -0.16934047639369965f;
  t39_n103: return -0.11760948598384857f;
  t39_n104: return 0.048723042011260986f;
  t39_n105: return -0.07347164303064346f;
  t39_n106: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t39_n110;
  t39_n107: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t39_n109;
  t39_n108: return 0.058998316526412964f;
  t39_n109: return -0.002007434843108058f;
  t39_n110: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t39_n112;
  t39_n111: return 0.044063229113817215f;
  t39_n112: return -0.011059951037168503f;
  t39_n113: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t39_n115;
  t39_n114: return -0.04486379399895668f;
  t39_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n117;
  t39_n116: return -0.08739254623651505f;
  t39_n117: return 0.0406353659927845f;
  t39_n118: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t39_n134;
  t39_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t39_n125;
  t39_n120: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t39_n122;
  t39_n121: return -0.06077628955245018f;
  t39_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n124;
  t39_n123: return -0.01847780868411064f;
  t39_n124: return 0.03544619679450989f;
  t39_n125: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t39_n127;
  t39_n126: return 0.010149464942514896f;
  t39_n127: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t39_n129;
  t39_n128: return 0.05155903100967407f;
  t39_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n131;
  t39_n130: return -0.06859272718429565f;
  t39_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n133;
  t39_n132: return -0.012415091507136822f;
  t39_n133: return 0.0913599282503128f;
  t39_n134: return -0.031216898933053017f;
  t39_n135: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t39_n139;
  t39_n136: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t39_n138;
  t39_n137: return -0.0238693468272686f;
  t39_n138: return 0.018544064834713936f;
  t39_n139: return 0.0486990250647068f;
  t39_n140: return 0.054135192185640335f;
  t39_n141: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t39_n191;
  t39_n142: if (E.getNumReferences() >= 3282124800 /*322.5*/) goto t39_n182;
  t39_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t39_n177;
  t39_n144: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t39_n172;
  t39_n145: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t39_n171;
  t39_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t39_n162;
  t39_n147: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t39_n161;
  t39_n148: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n160;
  t39_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n153;
  t39_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t39_n152;
  t39_n151: return -0.1565038412809372f;
  t39_n152: return -0.10734419524669647f;
  t39_n153: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t39_n155;
  t39_n154: return -0.011349521577358246f;
  t39_n155: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t39_n159;
  t39_n156: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t39_n158;
  t39_n157: return -0.016877995803952217f;
  t39_n158: return 0.04358560964465141f;
  t39_n159: return 0.0808640569448471f;
  t39_n160: return 0.006598726846277714f;
  t39_n161: return -0.06340576708316803f;
  t39_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t39_n164;
  t39_n163: return -0.01275179535150528f;
  t39_n164: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n168;
  t39_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t39_n167;
  t39_n166: return 0.011189262382686138f;
  t39_n167: return 0.03383777290582657f;
  t39_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t39_n170;
  t39_n169: return 0.017992673441767693f;
  t39_n170: return 0.04886111244559288f;
  t39_n171: return 0.03268910199403763f;
  t39_n172: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n176;
  t39_n173: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t39_n175;
  t39_n174: return -0.005921557080000639f;
  t39_n175: return 0.013937298208475113f;
  t39_n176: return -0.07227492332458496f;
  t39_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t39_n181;
  t39_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t39_n180;
  t39_n179: return -0.003170951269567013f;
  t39_n180: return 0.01930972747504711f;
  t39_n181: return 0.045413244515657425f;
  t39_n182: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t39_n188;
  t39_n183: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t39_n185;
  t39_n184: return 0.04770907759666443f;
  t39_n185: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t39_n187;
  t39_n186: return 0.10869460552930832f;
  t39_n187: return 0.08061840385198593f;
  t39_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t39_n190;
  t39_n189: return 0.016347920522093773f;
  t39_n190: return 0.054645538330078125f;
  t39_n191: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t39_n245;
  t39_n192: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t39_n230;
  t39_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t39_n199;
  t39_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t39_n196;
  t39_n195: return -0.028782276436686516f;
  t39_n196: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t39_n198;
  t39_n197: return 0.016017325222492218f;
  t39_n198: return -0.08688339591026306f;
  t39_n199: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t39_n229;
  t39_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t39_n228;
  t39_n201: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t39_n207;
  t39_n202: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t39_n206;
  t39_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t39_n205;
  t39_n204: return 0.02485649287700653f;
  t39_n205: return -0.015349939465522766f;
  t39_n206: return 0.03624889627099037f;
  t39_n207: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t39_n227;
  t39_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t39_n226;
  t39_n209: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t39_n211;
  t39_n210: return -0.01823943294584751f;
  t39_n211: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t39_n223;
  t39_n212: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t39_n216;
  t39_n213: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t39_n215;
  t39_n214: return 0.025144588202238083f;
  t39_n215: return -0.07287686318159103f;
  t39_n216: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t39_n218;
  t39_n217: return 0.021756896749138832f;
  t39_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t39_n220;
  t39_n219: return 0.016382012516260147f;
  t39_n220: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t39_n222;
  t39_n221: return 0.1518246978521347f;
  t39_n222: return 0.0585801899433136f;
  t39_n223: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t39_n225;
  t39_n224: return 0.038906972855329514f;
  t39_n225: return 0.09546500444412231f;
  t39_n226: return 0.024029048159718513f;
  t39_n227: return -0.054661866277456284f;
  t39_n228: return 0.03246527537703514f;
  t39_n229: return 0.0470459908246994f;
  t39_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t39_n242;
  t39_n231: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t39_n239;
  t39_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t39_n234;
  t39_n233: return 0.02543589659035206f;
  t39_n234: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t39_n238;
  t39_n235: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t39_n237;
  t39_n236: return 0.0706193596124649f;
  t39_n237: return 0.14147700369358063f;
  t39_n238: return 0.02345248870551586f;
  t39_n239: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t39_n241;
  t39_n240: return -0.017506413161754608f;
  t39_n241: return 0.04783428832888603f;
  t39_n242: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t39_n244;
  t39_n243: return 0.0382075160741806f;
  t39_n244: return 0.09184388816356659f;
  t39_n245: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t39_n247;
  t39_n246: return 0.05079101398587227f;
  t39_n247: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t39_n249;
  t39_n248: return 0.02153118886053562f;
  t39_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t39_n251;
  t39_n250: return 0.02528710849583149f;
  t39_n251: return 0.060351308435201645f;
  t39_n252: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t39_n254;
  t39_n253: return 0.042553916573524475f;
  t39_n254: return 0.060221780091524124f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree40(const Example& E) {
  t40_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t40_n252;
  t40_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t40_n87;
  t40_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t40_n70;
  t40_n3: if (E.getNumReferences() >= 3317122048 /*5864.5*/) goto t40_n57;
  t40_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n26;
  t40_n5: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t40_n23;
  t40_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n20;
  t40_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n9;
  t40_n8: return -0.11629708856344223f;
  t40_n9: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t40_n13;
  t40_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n12;
  t40_n11: return -0.021429024636745453f;
  t40_n12: return 0.016698410734534264f;
  t40_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t40_n15;
  t40_n14: return 0.017982356250286102f;
  t40_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t40_n17;
  t40_n16: return 0.020975932478904724f;
  t40_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t40_n19;
  t40_n18: return 0.019945969805121422f;
  t40_n19: return 0.10193752497434616f;
  t40_n20: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t40_n22;
  t40_n21: return 0.0033889024052768946f;
  t40_n22: return -0.03548935055732727f;
  t40_n23: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t40_n25;
  t40_n24: return 0.04186753183603287f;
  t40_n25: return 0.0014080532127991319f;
  t40_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t40_n28;
  t40_n27: return -0.08058590441942215f;
  t40_n28: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n54;
  t40_n29: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t40_n53;
  t40_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t40_n50;
  t40_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n39;
  t40_n32: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t40_n38;
  t40_n33: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t40_n35;
  t40_n34: return -0.007863521575927734f;
  t40_n35: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t40_n37;
  t40_n36: return 0.0193842314183712f;
  t40_n37: return -0.020749235525727272f;
  t40_n38: return 0.020912908017635345f;
  t40_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t40_n41;
  t40_n40: return -0.04704952985048294f;
  t40_n41: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t40_n49;
  t40_n42: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t40_n46;
  t40_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n45;
  t40_n44: return -0.04969378188252449f;
  t40_n45: return 0.031721893697977066f;
  t40_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t40_n48;
  t40_n47: return -0.05773472785949707f;
  t40_n48: return 0.06628631800413132f;
  t40_n49: return -0.011938164941966534f;
  t40_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n52;
  t40_n51: return -0.05967314541339874f;
  t40_n52: return 0.01745135709643364f;
  t40_n53: return -0.08582422882318497f;
  t40_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t40_n56;
  t40_n55: return 0.013795590028166771f;
  t40_n56: return 0.03214019909501076f;
  t40_n57: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t40_n65;
  t40_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n64;
  t40_n59: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t40_n61;
  t40_n60: return 0.038307659327983856f;
  t40_n61: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t40_n63;
  t40_n62: return 0.023317668586969376f;
  t40_n63: return -0.13696540892124176f;
  t40_n64: return 0.0633639469742775f;
  t40_n65: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t40_n69;
  t40_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n68;
  t40_n67: return 0.06572314351797104f;
  t40_n68: return 0.11302852630615234f;
  t40_n69: return 0.06624189764261246f;
  t40_n70: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t40_n80;
  t40_n71: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t40_n79;
  t40_n72: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t40_n74;
  t40_n73: return -0.0010286795441061258f;
  t40_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t40_n76;
  t40_n75: return 0.006141636520624161f;
  t40_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t40_n78;
  t40_n77: return 0.03717099130153656f;
  t40_n78: return 0.10589969158172607f;
  t40_n79: return -0.022082403302192688f;
  t40_n80: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n86;
  t40_n81: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t40_n85;
  t40_n82: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t40_n84;
  t40_n83: return 0.008071331307291985f;
  t40_n84: return 0.030522853136062622f;
  t40_n85: return -0.09456981718540192f;
  t40_n86: return -0.06982482969760895f;
  t40_n87: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t40_n169;
  t40_n88: if (E.getNumReferences() >= 3282124800 /*322.5*/) goto t40_n154;
  t40_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t40_n147;
  t40_n90: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n146;
  t40_n91: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t40_n97;
  t40_n92: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t40_n96;
  t40_n93: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t40_n95;
  t40_n94: return 0.01886199228465557f;
  t40_n95: return -0.008867981843650341f;
  t40_n96: return 0.03668927401304245f;
  t40_n97: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t40_n139;
  t40_n98: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t40_n136;
  t40_n99: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t40_n135;
  t40_n100: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t40_n134;
  t40_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t40_n133;
  t40_n102: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t40_n108;
  t40_n103: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t40_n107;
  t40_n104: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t40_n106;
  t40_n105: return 0.008436043746769428f;
  t40_n106: return 0.05241475999355316f;
  t40_n107: return -0.01803339459002018f;
  t40_n108: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t40_n110;
  t40_n109: return -1.533644535811618e-05f;
  t40_n110: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t40_n132;
  t40_n111: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t40_n131;
  t40_n112: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n130;
  t40_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t40_n127;
  t40_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t40_n116;
  t40_n115: return 0.008227440528571606f;
  t40_n116: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t40_n124;
  t40_n117: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t40_n123;
  t40_n118: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t40_n122;
  t40_n119: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t40_n121;
  t40_n120: return 0.07503126561641693f;
  t40_n121: return 0.007305358070880175f;
  t40_n122: return 0.0612969845533371f;
  t40_n123: return -0.03825628384947777f;
  t40_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t40_n126;
  t40_n125: return 0.059452105313539505f;
  t40_n126: return 0.08682917803525925f;
  t40_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t40_n129;
  t40_n128: return -0.0239487886428833f;
  t40_n129: return 0.1136905625462532f;
  t40_n130: return -0.044714413583278656f;
  t40_n131: return -0.06434720754623413f;
  t40_n132: return -0.13996392488479614f;
  t40_n133: return 0.03900574520230293f;
  t40_n134: return -0.016195230185985565f;
  t40_n135: return 0.03226340934634209f;
  t40_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n138;
  t40_n137: return -0.06691122055053711f;
  t40_n138: return 0.0007991707534529269f;
  t40_n139: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t40_n145;
  t40_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t40_n144;
  t40_n141: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t40_n143;
  t40_n142: return 0.021413566544651985f;
  t40_n143: return 0.0543600395321846f;
  t40_n144: return 0.05757106468081474f;
  t40_n145: return -0.02920818328857422f;
  t40_n146: return 0.017608605325222015f;
  t40_n147: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t40_n153;
  t40_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n150;
  t40_n149: return -0.013582654297351837f;
  t40_n150: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t40_n152;
  t40_n151: return 0.032912831753492355f;
  t40_n152: return 0.009580472484230995f;
  t40_n153: return 0.032856494188308716f;
  t40_n154: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t40_n160;
  t40_n155: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t40_n159;
  t40_n156: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t40_n158;
  t40_n157: return 0.0588628426194191f;
  t40_n158: return 0.037967268377542496f;
  t40_n159: return 0.08108053356409073f;
  t40_n160: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t40_n168;
  t40_n161: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t40_n165;
  t40_n162: if (E.getNumReferences() >= 3356314496 /*144686.0*/) goto t40_n164;
  t40_n163: return -0.03413311392068863f;
  t40_n164: return 0.05824405327439308f;
  t40_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t40_n167;
  t40_n166: return 0.024874171242117882f;
  t40_n167: return 0.06945423781871796f;
  t40_n168: return 0.05264727398753166f;
  t40_n169: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t40_n241;
  t40_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t40_n172;
  t40_n171: return -0.0002237919979961589f;
  t40_n172: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t40_n198;
  t40_n173: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t40_n185;
  t40_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t40_n182;
  t40_n175: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t40_n179;
  t40_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n178;
  t40_n177: return -0.012690775096416473f;
  t40_n178: return 0.021571990102529526f;
  t40_n179: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t40_n181;
  t40_n180: return -0.03988223150372505f;
  t40_n181: return 0.01787077635526657f;
  t40_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n184;
  t40_n183: return -0.10879069566726685f;
  t40_n184: return 0.03198476508259773f;
  t40_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t40_n191;
  t40_n186: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n190;
  t40_n187: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t40_n189;
  t40_n188: return 0.0052154590375721455f;
  t40_n189: return -0.05339672416448593f;
  t40_n190: return 0.032727357000112534f;
  t40_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t40_n197;
  t40_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t40_n194;
  t40_n193: return 0.01908930577337742f;
  t40_n194: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n196;
  t40_n195: return 0.030692407861351967f;
  t40_n196: return 0.05779999867081642f;
  t40_n197: return -0.0033603881020098925f;
  t40_n198: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t40_n236;
  t40_n199: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t40_n207;
  t40_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t40_n202;
  t40_n201: return 0.011276130564510822f;
  t40_n202: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t40_n204;
  t40_n203: return 0.04437137022614479f;
  t40_n204: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n206;
  t40_n205: return 0.009570405818521976f;
  t40_n206: return 0.0555516816675663f;
  t40_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t40_n211;
  t40_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t40_n210;
  t40_n209: return -0.02591586299240589f;
  t40_n210: return 0.018404025584459305f;
  t40_n211: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t40_n227;
  t40_n212: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t40_n218;
  t40_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t40_n215;
  t40_n214: return 0.014211038127541542f;
  t40_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t40_n217;
  t40_n216: return 0.032622575759887695f;
  t40_n217: return 0.09438423812389374f;
  t40_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t40_n220;
  t40_n219: return 0.03757208585739136f;
  t40_n220: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t40_n222;
  t40_n221: return 0.061686765402555466f;
  t40_n222: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t40_n226;
  t40_n223: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t40_n225;
  t40_n224: return 0.11031480878591537f;
  t40_n225: return 0.14848968386650085f;
  t40_n226: return 0.0816197469830513f;
  t40_n227: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t40_n229;
  t40_n228: return 0.058607008308172226f;
  t40_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t40_n231;
  t40_n230: return 0.081576868891716f;
  t40_n231: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n233;
  t40_n232: return 0.09705815464258194f;
  t40_n233: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t40_n235;
  t40_n234: return 0.14058150351047516f;
  t40_n235: return 0.06317593157291412f;
  t40_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t40_n238;
  t40_n237: return 0.014909236691892147f;
  t40_n238: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t40_n240;
  t40_n239: return 0.03355460241436958f;
  t40_n240: return 0.06140381097793579f;
  t40_n241: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t40_n243;
  t40_n242: return 0.054394517093896866f;
  t40_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t40_n251;
  t40_n244: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n250;
  t40_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t40_n247;
  t40_n246: return 0.014223232865333557f;
  t40_n247: if (E.getNumReferences() >= 3273359360 /*155.5*/) goto t40_n249;
  t40_n248: return 0.027914749458432198f;
  t40_n249: return 0.053342465311288834f;
  t40_n250: return 0.0708228051662445f;
  t40_n251: return 0.09247854351997375f;
  t40_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t40_n254;
  t40_n253: return 0.04527038708329201f;
  t40_n254: return 0.08088694512844086f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree41(const Example& E) {
  t41_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t41_n252;
  t41_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t41_n131;
  t41_n2: if (E.getNumReferences() >= 3284844544 /*405.5*/) goto t41_n122;
  t41_n3: if (E.getFileProximityDistanceCost() >= 3257139200 /*41.0*/) goto t41_n109;
  t41_n4: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t41_n92;
  t41_n5: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t41_n9;
  t41_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t41_n8;
  t41_n7: return 0.004293951205909252f;
  t41_n8: return 0.03974546492099762f;
  t41_n9: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t41_n83;
  t41_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t41_n12;
  t41_n11: return -0.08404054492712021f;
  t41_n12: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t41_n82;
  t41_n13: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n79;
  t41_n14: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t41_n78;
  t41_n15: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t41_n77;
  t41_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t41_n38;
  t41_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n33;
  t41_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t41_n22;
  t41_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n21;
  t41_n20: return -0.1223851889371872f;
  t41_n21: return -0.022835006937384605f;
  t41_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t41_n28;
  t41_n23: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t41_n27;
  t41_n24: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t41_n26;
  t41_n25: return -0.006472584325820208f;
  t41_n26: return 0.019535329192876816f;
  t41_n27: return -0.007512901443988085f;
  t41_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n32;
  t41_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n31;
  t41_n30: return -0.08977425843477249f;
  t41_n31: return -0.07180455327033997f;
  t41_n32: return 0.020372774451971054f;
  t41_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t41_n37;
  t41_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n36;
  t41_n35: return -0.07328976690769196f;
  t41_n36: return 0.012053352780640125f;
  t41_n37: return 0.05993553623557091f;
  t41_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n40;
  t41_n39: return -0.05327226221561432f;
  t41_n40: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t41_n54;
  t41_n41: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t41_n51;
  t41_n42: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t41_n46;
  t41_n43: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t41_n45;
  t41_n44: return 0.023206740617752075f;
  t41_n45: return -0.02330644801259041f;
  t41_n46: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t41_n48;
  t41_n47: return 0.01900104247033596f;
  t41_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t41_n50;
  t41_n49: return 0.08372855186462402f;
  t41_n50: return 0.07760735601186752f;
  t41_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n53;
  t41_n52: return -0.1335676610469818f;
  t41_n53: return -0.02306448668241501f;
  t41_n54: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t41_n72;
  t41_n55: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t41_n59;
  t41_n56: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t41_n58;
  t41_n57: return 0.028493527323007584f;
  t41_n58: return -0.050133559852838516f;
  t41_n59: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t41_n63;
  t41_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n62;
  t41_n61: return -0.11593843996524811f;
  t41_n62: return 0.0862189456820488f;
  t41_n63: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t41_n65;
  t41_n64: return -0.18768759071826935f;
  t41_n65: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t41_n67;
  t41_n66: return 0.07334847748279572f;
  t41_n67: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t41_n69;
  t41_n68: return -0.15344323217868805f;
  t41_n69: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t41_n71;
  t41_n70: return 0.07697252929210663f;
  t41_n71: return 0.019401676952838898f;
  t41_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t41_n74;
  t41_n73: return 0.025678280740976334f;
  t41_n74: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t41_n76;
  t41_n75: return 0.03785388544201851f;
  t41_n76: return 0.05834886059165001f;
  t41_n77: return -0.11316832900047302f;
  t41_n78: return -0.0499640554189682f;
  t41_n79: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t41_n81;
  t41_n80: return -0.010602391324937344f;
  t41_n81: return 0.029938040301203728f;
  t41_n82: return -0.11909741163253784f;
  t41_n83: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t41_n85;
  t41_n84: return -0.10010933130979538f;
  t41_n85: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t41_n91;
  t41_n86: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n88;
  t41_n87: return 0.00549098988994956f;
  t41_n88: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t41_n90;
  t41_n89: return 0.06365174055099487f;
  t41_n90: return 0.141992449760437f;
  t41_n91: return -0.04935052618384361f;
  t41_n92: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t41_n102;
  t41_n93: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t41_n99;
  t41_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n96;
  t41_n95: return 0.016931844875216484f;
  t41_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t41_n98;
  t41_n97: return 0.03878118470311165f;
  t41_n98: return 0.07149403542280197f;
  t41_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n101;
  t41_n100: return -0.19812782108783722f;
  t41_n101: return -0.030021555721759796f;
  t41_n102: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t41_n106;
  t41_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n105;
  t41_n104: return -0.05302973464131355f;
  t41_n105: return 0.013508794829249382f;
  t41_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n108;
  t41_n107: return -0.14870744943618774f;
  t41_n108: return 0.024752601981163025f;
  t41_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t41_n117;
  t41_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n116;
  t41_n111: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n113;
  t41_n112: return -0.005724295508116484f;
  t41_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t41_n115;
  t41_n114: return 0.006956339813768864f;
  t41_n115: return 0.04830893129110336f;
  t41_n116: return 0.058319319039583206f;
  t41_n117: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t41_n121;
  t41_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n120;
  t41_n119: return -0.07392647117376328f;
  t41_n120: return 0.015348617918789387f;
  t41_n121: return -0.08001692593097687f;
  t41_n122: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t41_n128;
  t41_n123: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t41_n127;
  t41_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t41_n126;
  t41_n125: return 0.022774750366806984f;
  t41_n126: return 0.053899794816970825f;
  t41_n127: return 0.055069807916879654f;
  t41_n128: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t41_n130;
  t41_n129: return -0.001491633476689458f;
  t41_n130: return 0.030158013105392456f;
  t41_n131: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t41_n185;
  t41_n132: if (E.getNumReferences() >= 3275718656 /*191.5*/) goto t41_n174;
  t41_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t41_n169;
  t41_n134: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n168;
  t41_n135: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t41_n141;
  t41_n136: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t41_n140;
  t41_n137: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t41_n139;
  t41_n138: return 0.018569378182291985f;
  t41_n139: return -0.012604273855686188f;
  t41_n140: return 0.029264051467180252f;
  t41_n141: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t41_n159;
  t41_n142: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t41_n156;
  t41_n143: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t41_n155;
  t41_n144: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t41_n148;
  t41_n145: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t41_n147;
  t41_n146: return 0.0061858962289988995f;
  t41_n147: return -0.010098634287714958f;
  t41_n148: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t41_n150;
  t41_n149: return -0.09675932675600052f;
  t41_n150: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t41_n152;
  t41_n151: return 0.03426700085401535f;
  t41_n152: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t41_n154;
  t41_n153: return 0.04701763391494751f;
  t41_n154: return -0.05271638184785843f;
  t41_n155: return 0.030747106298804283f;
  t41_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n158;
  t41_n157: return -0.06483130902051926f;
  t41_n158: return 0.0006218614289537072f;
  t41_n159: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t41_n167;
  t41_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t41_n164;
  t41_n161: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t41_n163;
  t41_n162: return 0.016858655959367752f;
  t41_n163: return 0.05108210816979408f;
  t41_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t41_n166;
  t41_n165: return -0.19395852088928223f;
  t41_n166: return 0.040911026298999786f;
  t41_n167: return -0.028432579711079597f;
  t41_n168: return 0.01672441139817238f;
  t41_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t41_n173;
  t41_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t41_n172;
  t41_n171: return -0.0038174453657120466f;
  t41_n172: return 0.01770881377160549f;
  t41_n173: return 0.041394416242837906f;
  t41_n174: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t41_n180;
  t41_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t41_n179;
  t41_n176: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t41_n178;
  t41_n177: return 0.02688945084810257f;
  t41_n178: return 0.06017206609249115f;
  t41_n179: return 0.06182088702917099f;
  t41_n180: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t41_n182;
  t41_n181: return 0.07545313984155655f;
  t41_n182: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t41_n184;
  t41_n183: return 0.012904117815196514f;
  t41_n184: return 0.06394483149051666f;
  t41_n185: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t41_n247;
  t41_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t41_n188;
  t41_n187: return 0.0020503797568380833f;
  t41_n188: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t41_n200;
  t41_n189: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t41_n193;
  t41_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t41_n192;
  t41_n191: return 0.009166265837848186f;
  t41_n192: return 0.02758074551820755f;
  t41_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t41_n197;
  t41_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n196;
  t41_n195: return 0.0036075946409255266f;
  t41_n196: return 0.03203694522380829f;
  t41_n197: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t41_n199;
  t41_n198: return 0.03062661550939083f;
  t41_n199: return -0.0011833257740363479f;
  t41_n200: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t41_n242;
  t41_n201: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t41_n207;
  t41_n202: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t41_n206;
  t41_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t41_n205;
  t41_n204: return 0.002108566462993622f;
  t41_n205: return 0.03442618250846863f;
  t41_n206: return 0.04604951664805412f;
  t41_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t41_n221;
  t41_n208: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t41_n210;
  t41_n209: return -0.08435504138469696f;
  t41_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t41_n212;
  t41_n211: return -0.001991087105125189f;
  t41_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t41_n220;
  t41_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t41_n219;
  t41_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t41_n216;
  t41_n215: return -0.02703542448580265f;
  t41_n216: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t41_n218;
  t41_n217: return 0.07348038256168365f;
  t41_n218: return -0.04658140242099762f;
  t41_n219: return 0.10421523451805115f;
  t41_n220: return 0.034524306654930115f;
  t41_n221: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t41_n233;
  t41_n222: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t41_n228;
  t41_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t41_n225;
  t41_n224: return 0.013923595659434795f;
  t41_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t41_n227;
  t41_n226: return 0.03491311892867088f;
  t41_n227: return 0.0922655388712883f;
  t41_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t41_n230;
  t41_n229: return 0.03559430316090584f;
  t41_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t41_n232;
  t41_n231: return 0.0676032304763794f;
  t41_n232: return 0.1015326976776123f;
  t41_n233: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t41_n235;
  t41_n234: return 0.05624141916632652f;
  t41_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t41_n237;
  t41_n236: return 0.08137117326259613f;
  t41_n237: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t41_n239;
  t41_n238: return 0.08897016197443008f;
  t41_n239: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t41_n241;
  t41_n240: return 0.13480313122272491f;
  t41_n241: return 0.10245257616043091f;
  t41_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t41_n244;
  t41_n243: return 0.01641356758773327f;
  t41_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t41_n246;
  t41_n245: return 0.031370192766189575f;
  t41_n246: return 0.05939238518476486f;
  t41_n247: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t41_n249;
  t41_n248: return 0.061280228197574615f;
  t41_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t41_n251;
  t41_n250: return 0.03846067190170288f;
  t41_n251: return 0.09008751809597015f;
  t41_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t41_n254;
  t41_n253: return 0.0383763462305069f;
  t41_n254: return 0.056582771241664886f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree42(const Example& E) {
  t42_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t42_n252;
  t42_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t42_n119;
  t42_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t42_n52;
  t42_n3: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t42_n43;
  t42_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n22;
  t42_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n19;
  t42_n6: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t42_n18;
  t42_n7: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t42_n17;
  t42_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t42_n10;
  t42_n9: return -0.11440881341695786f;
  t42_n10: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t42_n16;
  t42_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t42_n15;
  t42_n12: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t42_n14;
  t42_n13: return 0.0024396658409386873f;
  t42_n14: return -0.09061284363269806f;
  t42_n15: return -0.034083347767591476f;
  t42_n16: return -0.12529128789901733f;
  t42_n17: return 0.04578683525323868f;
  t42_n18: return -0.02737162448465824f;
  t42_n19: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t42_n21;
  t42_n20: return 0.01363146398216486f;
  t42_n21: return -0.02193453162908554f;
  t42_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t42_n24;
  t42_n23: return -0.07799830287694931f;
  t42_n24: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n42;
  t42_n25: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t42_n41;
  t42_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n32;
  t42_n27: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t42_n29;
  t42_n28: return 0.004220736678689718f;
  t42_n29: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t42_n31;
  t42_n30: return 0.00670249667018652f;
  t42_n31: return 0.03280336782336235f;
  t42_n32: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t42_n40;
  t42_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t42_n35;
  t42_n34: return -0.01807071454823017f;
  t42_n35: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t42_n37;
  t42_n36: return 0.014596058055758476f;
  t42_n37: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t42_n39;
  t42_n38: return 0.11986848711967468f;
  t42_n39: return 0.038290102034807205f;
  t42_n40: return 0.029941553249955177f;
  t42_n41: return -0.04882432520389557f;
  t42_n42: return 0.028898194432258606f;
  t42_n43: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t42_n47;
  t42_n44: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n46;
  t42_n45: return 0.028305046260356903f;
  t42_n46: return 0.059811633080244064f;
  t42_n47: if (E.getNumReferences() >= 3309029376 /*3004.5*/) goto t42_n51;
  t42_n48: if (E.getNumReferences() >= 3308949504 /*2985.0*/) goto t42_n50;
  t42_n49: return 0.06550242006778717f;
  t42_n50: return 0.15305960178375244f;
  t42_n51: return 0.06397901475429535f;
  t42_n52: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t42_n118;
  t42_n53: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n113;
  t42_n54: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t42_n110;
  t42_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n83;
  t42_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t42_n70;
  t42_n57: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t42_n67;
  t42_n58: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t42_n66;
  t42_n59: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t42_n65;
  t42_n60: if (E.getNumReferences() >= 3359032704 /*187158.0*/) goto t42_n62;
  t42_n61: return -0.017816107720136642f;
  t42_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t42_n64;
  t42_n63: return 0.0009674985776655376f;
  t42_n64: return 0.09374208748340607f;
  t42_n65: return -0.07079055160284042f;
  t42_n66: return -0.06777811050415039f;
  t42_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n69;
  t42_n68: return -0.35184887051582336f;
  t42_n69: return 0.04439201205968857f;
  t42_n70: if (E.getNumReferences() >= 3363896688 /*264171.5*/) goto t42_n78;
  t42_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t42_n77;
  t42_n72: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t42_n74;
  t42_n73: return -0.0004405655199661851f;
  t42_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t42_n76;
  t42_n75: return -0.04995359107851982f;
  t42_n76: return -0.0023943136911839247f;
  t42_n77: return 0.01749139465391636f;
  t42_n78: if (E.getNumReferences() >= 3364180016 /*273025.5*/) goto t42_n82;
  t42_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n81;
  t42_n80: return -0.127621591091156f;
  t42_n81: return 0.09695468097925186f;
  t42_n82: return 0.06353171914815903f;
  t42_n83: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t42_n95;
  t42_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t42_n94;
  t42_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n87;
  t42_n86: return -0.04668719694018364f;
  t42_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t42_n89;
  t42_n88: return -0.10596001148223877f;
  t42_n89: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t42_n91;
  t42_n90: return 0.09342598915100098f;
  t42_n91: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t42_n93;
  t42_n92: return 0.03155611455440521f;
  t42_n93: return 0.14130692183971405f;
  t42_n94: return 0.007711453828960657f;
  t42_n95: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t42_n109;
  t42_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t42_n98;
  t42_n97: return -0.11020558327436447f;
  t42_n98: if (E.getNumReferences() >= 3359032704 /*187158.0*/) goto t42_n100;
  t42_n99: return 0.06103358417749405f;
  t42_n100: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t42_n104;
  t42_n101: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t42_n103;
  t42_n102: return 0.10905906558036804f;
  t42_n103: return 0.1381366103887558f;
  t42_n104: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t42_n108;
  t42_n105: if (E.getNumReferences() >= 3361394368 /*224059.0*/) goto t42_n107;
  t42_n106: return 0.06750861555337906f;
  t42_n107: return -0.2574765980243683f;
  t42_n108: return 0.11200857907533646f;
  t42_n109: return -0.01355533953756094f;
  t42_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t42_n112;
  t42_n111: return -0.11234042048454285f;
  t42_n112: return -0.04062904417514801f;
  t42_n113: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t42_n117;
  t42_n114: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t42_n116;
  t42_n115: return -0.021861586719751358f;
  t42_n116: return 0.015200980007648468f;
  t42_n117: return 0.04641721025109291f;
  t42_n118: return 0.049207307398319244f;
  t42_n119: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t42_n201;
  t42_n120: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t42_n182;
  t42_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t42_n171;
  t42_n122: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n166;
  t42_n123: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t42_n139;
  t42_n124: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t42_n138;
  t42_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t42_n129;
  t42_n126: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t42_n128;
  t42_n127: return 0.0007710344507358968f;
  t42_n128: return -0.0530339851975441f;
  t42_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t42_n131;
  t42_n130: return -0.007538147270679474f;
  t42_n131: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t42_n135;
  t42_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t42_n134;
  t42_n133: return 0.010754481889307499f;
  t42_n134: return 0.029473554342985153f;
  t42_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t42_n137;
  t42_n136: return 0.012968970462679863f;
  t42_n137: return 0.04501194879412651f;
  t42_n138: return 0.03956266865134239f;
  t42_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t42_n141;
  t42_n140: return -0.01877599023282528f;
  t42_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t42_n161;
  t42_n142: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t42_n160;
  t42_n143: if (E.getFileProximityDistanceCost() >= 3256221696 /*37.5*/) goto t42_n159;
  t42_n144: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t42_n148;
  t42_n145: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t42_n147;
  t42_n146: return -0.00882644485682249f;
  t42_n147: return 0.039479129016399384f;
  t42_n148: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t42_n150;
  t42_n149: return -0.08299977332353592f;
  t42_n150: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t42_n158;
  t42_n151: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t42_n157;
  t42_n152: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t42_n156;
  t42_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n155;
  t42_n154: return -0.021077042445540428f;
  t42_n155: return 0.05656436085700989f;
  t42_n156: return -0.059632714837789536f;
  t42_n157: return -0.11760767549276352f;
  t42_n158: return -0.022641927003860474f;
  t42_n159: return -0.20077253878116608f;
  t42_n160: return 0.03287321329116821f;
  t42_n161: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t42_n165;
  t42_n162: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t42_n164;
  t42_n163: return 0.027070598676800728f;
  t42_n164: return 0.05584423989057541f;
  t42_n165: return -0.013477452099323273f;
  t42_n166: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t42_n170;
  t42_n167: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t42_n169;
  t42_n168: return 0.04357719048857689f;
  t42_n169: return 0.009565450251102448f;
  t42_n170: return 0.04588083177804947f;
  t42_n171: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t42_n177;
  t42_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t42_n174;
  t42_n173: return -0.008417780511081219f;
  t42_n174: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t42_n176;
  t42_n175: return 0.03192295506596565f;
  t42_n176: return 0.008283640258014202f;
  t42_n177: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t42_n181;
  t42_n178: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t42_n180;
  t42_n179: return 0.02462034858763218f;
  t42_n180: return -0.0749669000506401f;
  t42_n181: return 0.05357442423701286f;
  t42_n182: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t42_n194;
  t42_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t42_n193;
  t42_n184: if (E.getNumReferences() >= 3314739200 /*4701.0*/) goto t42_n188;
  t42_n185: if (E.getNumReferences() >= 3275784192 /*192.5*/) goto t42_n187;
  t42_n186: return 0.20917820930480957f;
  t42_n187: return 0.01375352218747139f;
  t42_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n192;
  t42_n189: if (E.getNumReferences() >= 3315203072 /*4927.5*/) goto t42_n191;
  t42_n190: return 0.13171693682670593f;
  t42_n191: return 0.0331350639462471f;
  t42_n192: return 0.10167955607175827f;
  t42_n193: return 0.059500742703676224f;
  t42_n194: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t42_n196;
  t42_n195: return 0.07163655757904053f;
  t42_n196: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t42_n200;
  t42_n197: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n199;
  t42_n198: return 0.00955402385443449f;
  t42_n199: return 0.06298813968896866f;
  t42_n200: return 0.06167752668261528f;
  t42_n201: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t42_n247;
  t42_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n242;
  t42_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t42_n209;
  t42_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t42_n206;
  t42_n205: return -0.025320902466773987f;
  t42_n206: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t42_n208;
  t42_n207: return 0.014634810388088226f;
  t42_n208: return -0.08233275264501572f;
  t42_n209: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t42_n239;
  t42_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t42_n238;
  t42_n211: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t42_n217;
  t42_n212: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t42_n216;
  t42_n213: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t42_n215;
  t42_n214: return 0.023588914424180984f;
  t42_n215: return -0.01361081749200821f;
  t42_n216: return 0.03327969089150429f;
  t42_n217: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t42_n237;
  t42_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t42_n236;
  t42_n219: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t42_n221;
  t42_n220: return -0.015993401408195496f;
  t42_n221: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t42_n233;
  t42_n222: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t42_n226;
  t42_n223: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t42_n225;
  t42_n224: return 0.02184978872537613f;
  t42_n225: return -0.06821814924478531f;
  t42_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t42_n228;
  t42_n227: return 0.002862169872969389f;
  t42_n228: if (E.getFractionNameInContext() >= 3206096018 /*0.5982142686843872*/) goto t42_n230;
  t42_n229: return 0.05058157071471214f;
  t42_n230: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t42_n232;
  t42_n231: return 0.12027976661920547f;
  t42_n232: return 0.056220170110464096f;
  t42_n233: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t42_n235;
  t42_n234: return 0.03546970337629318f;
  t42_n235: return 0.08491899818181992f;
  t42_n236: return 0.02195402793586254f;
  t42_n237: return -0.04927009344100952f;
  t42_n238: return 0.029972819611430168f;
  t42_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t42_n241;
  t42_n240: return 0.03482617810368538f;
  t42_n241: return 0.07139187306165695f;
  t42_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t42_n244;
  t42_n243: return 0.025290701538324356f;
  t42_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t42_n246;
  t42_n245: return 0.04217825457453728f;
  t42_n246: return 0.09150570631027222f;
  t42_n247: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t42_n249;
  t42_n248: return 0.046156592667102814f;
  t42_n249: if (E.getNumReferences() >= 3278405632 /*232.5*/) goto t42_n251;
  t42_n250: return 0.021390579640865326f;
  t42_n251: return 0.044871728867292404f;
  t42_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t42_n254;
  t42_n253: return 0.041573360562324524f;
  t42_n254: return 0.07842329144477844f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree43(const Example& E) {
  t43_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t43_n252;
  t43_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t43_n115;
  t43_n2: if (E.getNumReferences() >= 3284844544 /*405.5*/) goto t43_n74;
  t43_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t43_n49;
  t43_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n24;
  t43_n5: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t43_n9;
  t43_n6: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t43_n8;
  t43_n7: return -0.020184842869639397f;
  t43_n8: return 0.0046147010289132595f;
  t43_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t43_n15;
  t43_n10: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t43_n14;
  t43_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n13;
  t43_n12: return -0.065738745033741f;
  t43_n13: return 0.0011773010483011603f;
  t43_n14: return -0.036426905542612076f;
  t43_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t43_n23;
  t43_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t43_n20;
  t43_n17: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t43_n19;
  t43_n18: return 0.014923159033060074f;
  t43_n19: return 0.06910571455955505f;
  t43_n20: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t43_n22;
  t43_n21: return -0.05319187045097351f;
  t43_n22: return 0.0712495818734169f;
  t43_n23: return -0.08220823854207993f;
  t43_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t43_n32;
  t43_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n27;
  t43_n26: return -0.013038529083132744f;
  t43_n27: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t43_n31;
  t43_n28: if (E.getNumReferences() >= 3275784192 /*192.5*/) goto t43_n30;
  t43_n29: return 0.015722934156656265f;
  t43_n30: return 0.07379202544689178f;
  t43_n31: return 0.07379497587680817f;
  t43_n32: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t43_n48;
  t43_n33: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t43_n35;
  t43_n34: return 0.015206053853034973f;
  t43_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t43_n39;
  t43_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n38;
  t43_n37: return -0.12360826879739761f;
  t43_n38: return 0.003715314669534564f;
  t43_n39: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t43_n47;
  t43_n40: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t43_n44;
  t43_n41: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t43_n43;
  t43_n42: return -0.006544902920722961f;
  t43_n43: return 0.04252394288778305f;
  t43_n44: if (E.getNumReferences() >= 3282141184 /*323.0*/) goto t43_n46;
  t43_n45: return 0.05761278048157692f;
  t43_n46: return 0.15599288046360016f;
  t43_n47: return 0.05216791853308678f;
  t43_n48: return -0.07899399846792221f;
  t43_n49: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t43_n61;
  t43_n50: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t43_n54;
  t43_n51: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t43_n53;
  t43_n52: return -0.014528091065585613f;
  t43_n53: return -0.11243508011102676f;
  t43_n54: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t43_n58;
  t43_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t43_n57;
  t43_n56: return -0.0417327880859375f;
  t43_n57: return 0.03583855554461479f;
  t43_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n60;
  t43_n59: return -0.04495887830853462f;
  t43_n60: return 0.03670765459537506f;
  t43_n61: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t43_n73;
  t43_n62: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t43_n64;
  t43_n63: return -0.025302479043602943f;
  t43_n64: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t43_n68;
  t43_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t43_n67;
  t43_n66: return 0.01702168397605419f;
  t43_n67: return 0.06261057406663895f;
  t43_n68: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t43_n70;
  t43_n69: return 0.05948223918676376f;
  t43_n70: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t43_n72;
  t43_n71: return 0.15799188613891602f;
  t43_n72: return 0.07200003415346146f;
  t43_n73: return -0.0712391659617424f;
  t43_n74: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t43_n80;
  t43_n75: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t43_n77;
  t43_n76: return 0.02417183294892311f;
  t43_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t43_n79;
  t43_n78: return 0.04801947623491287f;
  t43_n79: return 0.0972171425819397f;
  t43_n80: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t43_n114;
  t43_n81: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t43_n85;
  t43_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t43_n84;
  t43_n83: return -0.03568938001990318f;
  t43_n84: return 0.03681467846035957f;
  t43_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t43_n103;
  t43_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n90;
  t43_n87: if (E.getNumReferences() >= 3344808448 /*56762.0*/) goto t43_n89;
  t43_n88: return -0.05170828476548195f;
  t43_n89: return -0.15655069053173065f;
  t43_n90: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t43_n92;
  t43_n91: return 0.008970317430794239f;
  t43_n92: if (E.getNumReferences() >= 3361088608 /*219281.5*/) goto t43_n96;
  t43_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t43_n95;
  t43_n94: return -0.24054081737995148f;
  t43_n95: return 0.09699857980012894f;
  t43_n96: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t43_n98;
  t43_n97: return -0.061111148446798325f;
  t43_n98: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t43_n100;
  t43_n99: return -0.07268907874822617f;
  t43_n100: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t43_n102;
  t43_n101: return 0.08979954570531845f;
  t43_n102: return -0.17751248180866241f;
  t43_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t43_n105;
  t43_n104: return 0.005696565378457308f;
  t43_n105: if (E.getNumReferences() >= 3346546048 /*63549.5*/) goto t43_n109;
  t43_n106: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t43_n108;
  t43_n107: return 0.03943123668432236f;
  t43_n108: return -0.06436473876237869f;
  t43_n109: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t43_n113;
  t43_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t43_n112;
  t43_n111: return 0.010400199331343174f;
  t43_n112: return 0.08333856612443924f;
  t43_n113: return 0.0005596401751972735f;
  t43_n114: return 0.02773132547736168f;
  t43_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t43_n177;
  t43_n116: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t43_n168;
  t43_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t43_n163;
  t43_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t43_n160;
  t43_n119: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t43_n133;
  t43_n120: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t43_n130;
  t43_n121: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t43_n125;
  t43_n122: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t43_n124;
  t43_n123: return 0.013100171461701393f;
  t43_n124: return 0.0256027989089489f;
  t43_n125: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t43_n127;
  t43_n126: return -0.07205265015363693f;
  t43_n127: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t43_n129;
  t43_n128: return 0.024108482524752617f;
  t43_n129: return -0.040832217782735825f;
  t43_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t43_n132;
  t43_n131: return 0.029482562094926834f;
  t43_n132: return 0.04861900210380554f;
  t43_n133: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t43_n159;
  t43_n134: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t43_n156;
  t43_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t43_n153;
  t43_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t43_n138;
  t43_n137: return -0.0735803097486496f;
  t43_n138: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t43_n140;
  t43_n139: return -0.0046876175329089165f;
  t43_n140: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t43_n152;
  t43_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t43_n145;
  t43_n142: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t43_n144;
  t43_n143: return 0.004417781252413988f;
  t43_n144: return -0.047685299068689346f;
  t43_n145: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t43_n151;
  t43_n146: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t43_n148;
  t43_n147: return 0.02754506655037403f;
  t43_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t43_n150;
  t43_n149: return -0.012047077529132366f;
  t43_n150: return 0.0786629244685173f;
  t43_n151: return -0.024203669279813766f;
  t43_n152: return -0.11371316015720367f;
  t43_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t43_n155;
  t43_n154: return -0.12653855979442596f;
  t43_n155: return 0.053060054779052734f;
  t43_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n158;
  t43_n157: return -0.04125671461224556f;
  t43_n158: return -0.00048645713832229376f;
  t43_n159: return -0.037620335817337036f;
  t43_n160: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t43_n162;
  t43_n161: return 0.012112072668969631f;
  t43_n162: return 0.04469988867640495f;
  t43_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t43_n167;
  t43_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t43_n166;
  t43_n165: return -0.00270156585611403f;
  t43_n166: return 0.016339482739567757f;
  t43_n167: return 0.03783193603157997f;
  t43_n168: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t43_n172;
  t43_n169: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t43_n171;
  t43_n170: return 0.0424567349255085f;
  t43_n171: return 0.07781830430030823f;
  t43_n172: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t43_n174;
  t43_n173: return 0.06865949183702469f;
  t43_n174: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t43_n176;
  t43_n175: return 0.011888988316059113f;
  t43_n176: return 0.05882170796394348f;
  t43_n177: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t43_n243;
  t43_n178: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t43_n226;
  t43_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t43_n185;
  t43_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n182;
  t43_n181: return -0.022766271606087685f;
  t43_n182: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t43_n184;
  t43_n183: return 0.01395344827324152f;
  t43_n184: return -0.08237181603908539f;
  t43_n185: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t43_n223;
  t43_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t43_n222;
  t43_n187: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t43_n199;
  t43_n188: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t43_n198;
  t43_n189: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t43_n197;
  t43_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n192;
  t43_n191: return -0.0067428844049572945f;
  t43_n192: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t43_n196;
  t43_n193: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t43_n195;
  t43_n194: return 0.03406781703233719f;
  t43_n195: return 0.013970296829938889f;
  t43_n196: return 0.03796696662902832f;
  t43_n197: return -0.0010246318997815251f;
  t43_n198: return 0.039972517639398575f;
  t43_n199: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t43_n221;
  t43_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t43_n208;
  t43_n201: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t43_n203;
  t43_n202: return -0.014954203739762306f;
  t43_n203: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t43_n205;
  t43_n204: return 0.0018611741252243519f;
  t43_n205: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t43_n207;
  t43_n206: return 0.032550498843193054f;
  t43_n207: return 0.08356624841690063f;
  t43_n208: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t43_n220;
  t43_n209: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t43_n215;
  t43_n210: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t43_n214;
  t43_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t43_n213;
  t43_n212: return -0.015591593459248543f;
  t43_n213: return 0.024816883727908134f;
  t43_n214: return 0.07276642322540283f;
  t43_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t43_n217;
  t43_n216: return -0.21058300137519836f;
  t43_n217: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t43_n219;
  t43_n218: return 0.022947175428271294f;
  t43_n219: return 0.05346151813864708f;
  t43_n220: return 0.005785374902188778f;
  t43_n221: return -0.04506896808743477f;
  t43_n222: return 0.028497571125626564f;
  t43_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t43_n225;
  t43_n224: return 0.032930709421634674f;
  t43_n225: return 0.054128605872392654f;
  t43_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t43_n228;
  t43_n227: return 0.02448778599500656f;
  t43_n228: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t43_n230;
  t43_n229: return 0.03276113420724869f;
  t43_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t43_n232;
  t43_n231: return 0.04890785366296768f;
  t43_n232: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t43_n234;
  t43_n233: return 0.023465260863304138f;
  t43_n234: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t43_n236;
  t43_n235: return 0.0767349973320961f;
  t43_n236: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t43_n242;
  t43_n237: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t43_n239;
  t43_n238: return -0.20422066748142242f;
  t43_n239: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t43_n241;
  t43_n240: return 0.1392282098531723f;
  t43_n241: return 0.07145605236291885f;
  t43_n242: return 0.04755246266722679f;
  t43_n243: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t43_n245;
  t43_n244: return 0.04460899904370308f;
  t43_n245: if (E.getNumReferences() >= 3325673472 /*11888.0*/) goto t43_n251;
  t43_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t43_n250;
  t43_n247: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t43_n249;
  t43_n248: return 0.004659419413655996f;
  t43_n249: return 0.054664235562086105f;
  t43_n250: return 0.029625078663229942f;
  t43_n251: return 0.05948670953512192f;
  t43_n252: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t43_n254;
  t43_n253: return 0.03451225534081459f;
  t43_n254: return 0.05317814648151398f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree44(const Example& E) {
  t44_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t44_n252;
  t44_n1: if (E.getFractionNameInContext() >= 3189477818 /*0.15192309021949768*/) goto t44_n181;
  t44_n2: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t44_n174;
  t44_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t44_n107;
  t44_n4: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t44_n94;
  t44_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n43;
  t44_n6: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t44_n8;
  t44_n7: return -0.016183285042643547f;
  t44_n8: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t44_n42;
  t44_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t44_n39;
  t44_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n32;
  t44_n11: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t44_n23;
  t44_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n14;
  t44_n13: return -0.09685542434453964f;
  t44_n14: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t44_n18;
  t44_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t44_n17;
  t44_n16: return 0.03007105551660061f;
  t44_n17: return -0.03145531937479973f;
  t44_n18: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t44_n20;
  t44_n19: return 0.020232701674103737f;
  t44_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t44_n22;
  t44_n21: return 0.017553487792611122f;
  t44_n22: return 0.09331438690423965f;
  t44_n23: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t44_n31;
  t44_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n28;
  t44_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t44_n27;
  t44_n26: return -0.11925447732210159f;
  t44_n27: return 0.11543954908847809f;
  t44_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t44_n30;
  t44_n29: return 0.005867535248398781f;
  t44_n30: return -0.14434584975242615f;
  t44_n31: return -0.056146908551454544f;
  t44_n32: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t44_n36;
  t44_n33: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t44_n35;
  t44_n34: return -0.211867094039917f;
  t44_n35: return 0.014361249282956123f;
  t44_n36: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t44_n38;
  t44_n37: return 0.059675607830286026f;
  t44_n38: return 0.11579413712024689f;
  t44_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t44_n41;
  t44_n40: return 0.03687319904565811f;
  t44_n41: return -0.06618261337280273f;
  t44_n42: return -0.03950027376413345f;
  t44_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t44_n45;
  t44_n44: return -0.06458237767219543f;
  t44_n45: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t44_n85;
  t44_n46: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t44_n78;
  t44_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t44_n55;
  t44_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n50;
  t44_n49: return -0.07183375954627991f;
  t44_n50: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t44_n52;
  t44_n51: return -0.01841561496257782f;
  t44_n52: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n54;
  t44_n53: return 0.006771992892026901f;
  t44_n54: return 0.049639321863651276f;
  t44_n55: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t44_n59;
  t44_n56: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t44_n58;
  t44_n57: return 0.025262825191020966f;
  t44_n58: return -0.08730253577232361f;
  t44_n59: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t44_n65;
  t44_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t44_n62;
  t44_n61: return -0.005390367936342955f;
  t44_n62: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t44_n64;
  t44_n63: return 0.07933919876813889f;
  t44_n64: return 0.012273700907826424f;
  t44_n65: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t44_n77;
  t44_n66: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t44_n70;
  t44_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t44_n69;
  t44_n68: return 0.001407311880029738f;
  t44_n69: return 0.08570375293493271f;
  t44_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t44_n72;
  t44_n71: return -0.06735393404960632f;
  t44_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n76;
  t44_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t44_n75;
  t44_n74: return -0.0654696673154831f;
  t44_n75: return 0.041306979954242706f;
  t44_n76: return 0.04526599124073982f;
  t44_n77: return -0.09400768578052521f;
  t44_n78: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t44_n80;
  t44_n79: return -0.09180091321468353f;
  t44_n80: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t44_n84;
  t44_n81: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n83;
  t44_n82: return -0.003749942174181342f;
  t44_n83: return 0.0617092102766037f;
  t44_n84: return 0.03139190375804901f;
  t44_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t44_n89;
  t44_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t44_n88;
  t44_n87: return -0.00264080916531384f;
  t44_n88: return 0.02538362704217434f;
  t44_n89: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t44_n91;
  t44_n90: return 0.042187757790088654f;
  t44_n91: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t44_n93;
  t44_n92: return -0.023703856393694878f;
  t44_n93: return 0.03894994780421257f;
  t44_n94: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n100;
  t44_n95: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t44_n99;
  t44_n96: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t44_n98;
  t44_n97: return 0.034543491899967194f;
  t44_n98: return 0.009715214371681213f;
  t44_n99: return 0.014787374064326286f;
  t44_n100: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t44_n102;
  t44_n101: return 0.028725987300276756f;
  t44_n102: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t44_n104;
  t44_n103: return 0.1365642249584198f;
  t44_n104: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t44_n106;
  t44_n105: return 0.07121019810438156f;
  t44_n106: return -0.1618807017803192f;
  t44_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t44_n163;
  t44_n108: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t44_n120;
  t44_n109: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t44_n119;
  t44_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t44_n114;
  t44_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t44_n113;
  t44_n112: return -0.011264830827713013f;
  t44_n113: return 0.01840541698038578f;
  t44_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t44_n116;
  t44_n115: return -0.06617020070552826f;
  t44_n116: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t44_n118;
  t44_n117: return 0.06355299055576324f;
  t44_n118: return 0.04291694611310959f;
  t44_n119: return 0.047771379351615906f;
  t44_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n152;
  t44_n121: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t44_n123;
  t44_n122: return -0.023120615631341934f;
  t44_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t44_n125;
  t44_n124: return -0.03514333814382553f;
  t44_n125: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t44_n127;
  t44_n126: return -0.00795077346265316f;
  t44_n127: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t44_n151;
  t44_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t44_n130;
  t44_n129: return 0.011910250410437584f;
  t44_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t44_n136;
  t44_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n135;
  t44_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t44_n134;
  t44_n133: return -0.1017933115363121f;
  t44_n134: return 0.008531571365892887f;
  t44_n135: return 0.04174318164587021f;
  t44_n136: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t44_n150;
  t44_n137: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t44_n139;
  t44_n138: return 0.08230838924646378f;
  t44_n139: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t44_n141;
  t44_n140: return -0.18108922243118286f;
  t44_n141: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t44_n145;
  t44_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n144;
  t44_n143: return -0.09758633375167847f;
  t44_n144: return 0.07797374576330185f;
  t44_n145: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t44_n147;
  t44_n146: return -0.14249774813652039f;
  t44_n147: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t44_n149;
  t44_n148: return 0.07546789199113846f;
  t44_n149: return 0.007416322361677885f;
  t44_n150: return -0.029646828770637512f;
  t44_n151: return -0.053270597010850906f;
  t44_n152: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t44_n154;
  t44_n153: return -0.02991386130452156f;
  t44_n154: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t44_n162;
  t44_n155: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t44_n161;
  t44_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t44_n158;
  t44_n157: return -0.03633427619934082f;
  t44_n158: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t44_n160;
  t44_n159: return 0.0328681617975235f;
  t44_n160: return -0.0405697263777256f;
  t44_n161: return 0.02697233483195305f;
  t44_n162: return -0.13909004628658295f;
  t44_n163: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t44_n167;
  t44_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t44_n166;
  t44_n165: return -0.025389650836586952f;
  t44_n166: return 0.03701276704668999f;
  t44_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t44_n171;
  t44_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n170;
  t44_n169: return -0.04761481657624245f;
  t44_n170: return 0.03547021001577377f;
  t44_n171: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t44_n173;
  t44_n172: return 0.028703702613711357f;
  t44_n173: return 0.09597110003232956f;
  t44_n174: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t44_n178;
  t44_n175: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t44_n177;
  t44_n176: return 0.034410636872053146f;
  t44_n177: return 0.07357162237167358f;
  t44_n178: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t44_n180;
  t44_n179: return -0.03624780848622322f;
  t44_n180: return 0.01582440920174122f;
  t44_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t44_n195;
  t44_n182: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n190;
  t44_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n185;
  t44_n184: return -0.023441849276423454f;
  t44_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t44_n189;
  t44_n186: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t44_n188;
  t44_n187: return 0.011107103899121284f;
  t44_n188: return -0.06820987910032272f;
  t44_n189: return -0.07860057801008224f;
  t44_n190: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t44_n192;
  t44_n191: return 0.010057525709271431f;
  t44_n192: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t44_n194;
  t44_n193: return 0.03032219596207142f;
  t44_n194: return 0.07722873985767365f;
  t44_n195: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t44_n249;
  t44_n196: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t44_n230;
  t44_n197: if (E.getNumReferences() >= 3292749824 /*781.5*/) goto t44_n227;
  t44_n198: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t44_n210;
  t44_n199: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t44_n209;
  t44_n200: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t44_n206;
  t44_n201: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t44_n205;
  t44_n202: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t44_n204;
  t44_n203: return 0.021105077117681503f;
  t44_n204: return 0.033670421689748764f;
  t44_n205: return 0.03747597709298134f;
  t44_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t44_n208;
  t44_n207: return -0.0029395841993391514f;
  t44_n208: return 0.06678520143032074f;
  t44_n209: return 0.05210459232330322f;
  t44_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t44_n220;
  t44_n211: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t44_n215;
  t44_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t44_n214;
  t44_n213: return 0.012968111783266068f;
  t44_n214: return 0.02533457987010479f;
  t44_n215: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t44_n217;
  t44_n216: return -0.08463281393051147f;
  t44_n217: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t44_n219;
  t44_n218: return 0.02862749807536602f;
  t44_n219: return -0.0061782547272741795f;
  t44_n220: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t44_n222;
  t44_n221: return 0.03551669046282768f;
  t44_n222: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t44_n226;
  t44_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n225;
  t44_n224: return -0.03314676135778427f;
  t44_n225: return 0.01671665720641613f;
  t44_n226: return 0.059564363211393356f;
  t44_n227: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t44_n229;
  t44_n228: return 0.06587834656238556f;
  t44_n229: return 0.039465054869651794f;
  t44_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t44_n232;
  t44_n231: return -0.08099988847970963f;
  t44_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t44_n246;
  t44_n233: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t44_n245;
  t44_n234: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t44_n238;
  t44_n235: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t44_n237;
  t44_n236: return 0.018252283334732056f;
  t44_n237: return 0.04963616654276848f;
  t44_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n242;
  t44_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t44_n241;
  t44_n240: return 0.14751183986663818f;
  t44_n241: return 0.059584084898233414f;
  t44_n242: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t44_n244;
  t44_n243: return 0.16105139255523682f;
  t44_n244: return 0.12980742752552032f;
  t44_n245: return -0.011793110519647598f;
  t44_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n248;
  t44_n247: return 0.01716713234782219f;
  t44_n248: return 0.0480586513876915f;
  t44_n249: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n251;
  t44_n250: return 0.03506721183657646f;
  t44_n251: return 0.07019275426864624f;
  t44_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t44_n254;
  t44_n253: return 0.03807228058576584f;
  t44_n254: return 0.07550982385873795f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree45(const Example& E) {
  t45_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t45_n252;
  t45_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t45_n151;
  t45_n2: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t45_n136;
  t45_n3: if (E.getFileProximityDistanceCost() >= 3257139200 /*41.0*/) goto t45_n125;
  t45_n4: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t45_n106;
  t45_n5: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t45_n7;
  t45_n6: return 0.005116590298712254f;
  t45_n7: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t45_n99;
  t45_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t45_n10;
  t45_n9: return -0.07963337749242783f;
  t45_n10: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t45_n98;
  t45_n11: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t45_n97;
  t45_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t45_n94;
  t45_n13: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t45_n93;
  t45_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t45_n48;
  t45_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n37;
  t45_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t45_n18;
  t45_n17: return -0.015294873155653477f;
  t45_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n24;
  t45_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n21;
  t45_n20: return -0.05066600814461708f;
  t45_n21: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t45_n23;
  t45_n22: return 0.09208918362855911f;
  t45_n23: return 0.13403953611850739f;
  t45_n24: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t45_n28;
  t45_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n27;
  t45_n26: return -0.02677016891539097f;
  t45_n27: return 0.005805009510368109f;
  t45_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n32;
  t45_n29: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t45_n31;
  t45_n30: return -0.02834298647940159f;
  t45_n31: return 0.015716630965471268f;
  t45_n32: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t45_n36;
  t45_n33: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t45_n35;
  t45_n34: return 0.018641386181116104f;
  t45_n35: return 0.05039697885513306f;
  t45_n36: return 0.0015199421904981136f;
  t45_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t45_n41;
  t45_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n40;
  t45_n39: return -0.10078100860118866f;
  t45_n40: return 0.0219353549182415f;
  t45_n41: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t45_n47;
  t45_n42: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t45_n44;
  t45_n43: return 0.03374004736542702f;
  t45_n44: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t45_n46;
  t45_n45: return -0.11464961618185043f;
  t45_n46: return 0.15138640999794006f;
  t45_n47: return 0.03232250362634659f;
  t45_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n52;
  t45_n49: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t45_n51;
  t45_n50: return 0.02886613830924034f;
  t45_n51: return -0.07650461792945862f;
  t45_n52: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t45_n70;
  t45_n53: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t45_n67;
  t45_n54: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t45_n62;
  t45_n55: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t45_n59;
  t45_n56: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t45_n58;
  t45_n57: return 0.021659398451447487f;
  t45_n58: return -0.0109062809497118f;
  t45_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n61;
  t45_n60: return -0.11859598755836487f;
  t45_n61: return -0.05594908446073532f;
  t45_n62: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t45_n64;
  t45_n63: return 0.02190416492521763f;
  t45_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t45_n66;
  t45_n65: return 0.07968355715274811f;
  t45_n66: return 0.07484111934900284f;
  t45_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n69;
  t45_n68: return -0.12284502387046814f;
  t45_n69: return -0.023027723655104637f;
  t45_n70: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t45_n88;
  t45_n71: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t45_n75;
  t45_n72: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t45_n74;
  t45_n73: return 0.030124090611934662f;
  t45_n74: return -0.04740636423230171f;
  t45_n75: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t45_n79;
  t45_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n78;
  t45_n77: return -0.11504478752613068f;
  t45_n78: return 0.0790986716747284f;
  t45_n79: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t45_n81;
  t45_n80: return -0.1538708657026291f;
  t45_n81: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t45_n83;
  t45_n82: return 0.06732077151536942f;
  t45_n83: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t45_n85;
  t45_n84: return -0.12743166089057922f;
  t45_n85: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t45_n87;
  t45_n86: return 0.06937959790229797f;
  t45_n87: return 0.014984079636633396f;
  t45_n88: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t45_n92;
  t45_n89: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t45_n91;
  t45_n90: return 0.02579193376004696f;
  t45_n91: return 0.04245873540639877f;
  t45_n92: return 0.008608483709394932f;
  t45_n93: return -0.042838387191295624f;
  t45_n94: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t45_n96;
  t45_n95: return 0.018871912732720375f;
  t45_n96: return -0.045135825872421265f;
  t45_n97: return -0.10512825101613998f;
  t45_n98: return -0.11492576450109482f;
  t45_n99: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t45_n101;
  t45_n100: return -0.0966622531414032f;
  t45_n101: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t45_n105;
  t45_n102: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t45_n104;
  t45_n103: return 0.004940178710967302f;
  t45_n104: return 0.06200229004025459f;
  t45_n105: return -0.04385679215192795f;
  t45_n106: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t45_n112;
  t45_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t45_n111;
  t45_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n110;
  t45_n109: return 0.0022234914358705282f;
  t45_n110: return 0.032346196472644806f;
  t45_n111: return -0.03248673304915428f;
  t45_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n114;
  t45_n113: return -0.16359049081802368f;
  t45_n114: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t45_n124;
  t45_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n121;
  t45_n116: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t45_n118;
  t45_n117: return -0.06506279110908508f;
  t45_n118: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t45_n120;
  t45_n119: return 0.03656671941280365f;
  t45_n120: return -0.04742514714598656f;
  t45_n121: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t45_n123;
  t45_n122: return 0.005832510534673929f;
  t45_n123: return 0.05676725134253502f;
  t45_n124: return 0.0209646075963974f;
  t45_n125: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t45_n133;
  t45_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t45_n128;
  t45_n127: return -0.07890838384628296f;
  t45_n128: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t45_n132;
  t45_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n131;
  t45_n130: return 0.002938284073024988f;
  t45_n131: return 0.011079584248363972f;
  t45_n132: return -0.05855005234479904f;
  t45_n133: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t45_n135;
  t45_n134: return -0.029647981747984886f;
  t45_n135: return 0.03156422823667526f;
  t45_n136: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t45_n140;
  t45_n137: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t45_n139;
  t45_n138: return 0.02326243557035923f;
  t45_n139: return 0.0499725416302681f;
  t45_n140: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t45_n146;
  t45_n141: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t45_n145;
  t45_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t45_n144;
  t45_n143: return -0.03358859941363335f;
  t45_n144: return 0.03485250473022461f;
  t45_n145: return 0.004736734554171562f;
  t45_n146: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t45_n148;
  t45_n147: return 0.05684206634759903f;
  t45_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t45_n150;
  t45_n149: return -0.005999754182994366f;
  t45_n150: return 0.04614538699388504f;
  t45_n151: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t45_n205;
  t45_n152: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t45_n190;
  t45_n153: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t45_n159;
  t45_n154: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t45_n158;
  t45_n155: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t45_n157;
  t45_n156: return 0.018651120364665985f;
  t45_n157: return -0.04022315889596939f;
  t45_n158: return 0.05870567634701729f;
  t45_n159: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t45_n177;
  t45_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t45_n162;
  t45_n161: return -0.041841715574264526f;
  t45_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t45_n168;
  t45_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t45_n165;
  t45_n164: return 0.004029351286590099f;
  t45_n165: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t45_n167;
  t45_n166: return 0.015582697466015816f;
  t45_n167: return 0.03145310655236244f;
  t45_n168: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t45_n170;
  t45_n169: return -0.08163149654865265f;
  t45_n170: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t45_n174;
  t45_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t45_n173;
  t45_n172: return 0.03145914897322655f;
  t45_n173: return 0.053544528782367706f;
  t45_n174: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t45_n176;
  t45_n175: return 0.02358301915228367f;
  t45_n176: return -0.04982702061533928f;
  t45_n177: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t45_n179;
  t45_n178: return 0.026770321652293205f;
  t45_n179: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t45_n189;
  t45_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t45_n182;
  t45_n181: return 0.003431019838899374f;
  t45_n182: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t45_n186;
  t45_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t45_n185;
  t45_n184: return 0.001824265462346375f;
  t45_n185: return 0.04331672191619873f;
  t45_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t45_n188;
  t45_n187: return 0.004587230272591114f;
  t45_n188: return 0.03980124369263649f;
  t45_n189: return -0.07350879907608032f;
  t45_n190: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t45_n194;
  t45_n191: if (E.getNumReferences() >= 3322594816 /*8881.5*/) goto t45_n193;
  t45_n192: return 0.03922422230243683f;
  t45_n193: return 0.062476616352796555f;
  t45_n194: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t45_n204;
  t45_n195: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t45_n201;
  t45_n196: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t45_n198;
  t45_n197: return 0.06613047420978546f;
  t45_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t45_n200;
  t45_n199: return -0.06645026803016663f;
  t45_n200: return 0.03736865520477295f;
  t45_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t45_n203;
  t45_n202: return 0.02065277472138405f;
  t45_n203: return 0.06851338595151901f;
  t45_n204: return 0.046848442405462265f;
  t45_n205: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t45_n243;
  t45_n206: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t45_n228;
  t45_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t45_n213;
  t45_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t45_n212;
  t45_n209: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t45_n211;
  t45_n210: return -0.19239746034145355f;
  t45_n211: return -0.11159166693687439f;
  t45_n212: return 0.009101544506847858f;
  t45_n213: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t45_n227;
  t45_n214: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t45_n222;
  t45_n215: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t45_n219;
  t45_n216: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t45_n218;
  t45_n217: return 0.022263912484049797f;
  t45_n218: return -0.002389879198744893f;
  t45_n219: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t45_n221;
  t45_n220: return 0.02785448543727398f;
  t45_n221: return 0.04270098730921745f;
  t45_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t45_n224;
  t45_n223: return 0.009971879422664642f;
  t45_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t45_n226;
  t45_n225: return 0.00703191664069891f;
  t45_n226: return 0.03648074343800545f;
  t45_n227: return 0.037520881742239f;
  t45_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t45_n230;
  t45_n229: return 0.02365925908088684f;
  t45_n230: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t45_n232;
  t45_n231: return 0.02410312369465828f;
  t45_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t45_n234;
  t45_n233: return 0.04140869528055191f;
  t45_n234: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t45_n236;
  t45_n235: return 0.01953882910311222f;
  t45_n236: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t45_n238;
  t45_n237: return 0.07244444638490677f;
  t45_n238: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t45_n242;
  t45_n239: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t45_n241;
  t45_n240: return -0.1480574607849121f;
  t45_n241: return 0.13172481954097748f;
  t45_n242: return 0.062344275414943695f;
  t45_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t45_n251;
  t45_n244: if (E.getNumReferences() >= 3347421312 /*68401.0*/) goto t45_n248;
  t45_n245: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t45_n247;
  t45_n246: return 0.05231805890798569f;
  t45_n247: return 0.02524024061858654f;
  t45_n248: if (E.getNumReferences() >= 3351977984 /*104000.0*/) goto t45_n250;
  t45_n249: return 0.09863559901714325f;
  t45_n250: return 0.05867299437522888f;
  t45_n251: return 0.05862895026803017f;
  t45_n252: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t45_n254;
  t45_n253: return 0.031239956617355347f;
  t45_n254: return 0.049956660717725754f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree46(const Example& E) {
  t46_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t46_n250;
  t46_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t46_n81;
  t46_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t46_n54;
  t46_n3: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t46_n35;
  t46_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n28;
  t46_n5: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t46_n25;
  t46_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t46_n22;
  t46_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n9;
  t46_n8: return -0.11444664746522903f;
  t46_n9: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t46_n15;
  t46_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t46_n12;
  t46_n11: return -0.05688409134745598f;
  t46_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n14;
  t46_n13: return -0.013113302178680897f;
  t46_n14: return 0.017112860456109047f;
  t46_n15: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t46_n21;
  t46_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t46_n20;
  t46_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t46_n19;
  t46_n18: return -0.020547989755868912f;
  t46_n19: return -0.12265463918447495f;
  t46_n20: return 0.009758383966982365f;
  t46_n21: return 0.0210165623575449f;
  t46_n22: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t46_n24;
  t46_n23: return 0.0038890521973371506f;
  t46_n24: return -0.02973208948969841f;
  t46_n25: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t46_n27;
  t46_n26: return 0.04309898987412453f;
  t46_n27: return 0.0072482796385884285f;
  t46_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t46_n30;
  t46_n29: return -0.07026010751724243f;
  t46_n30: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n32;
  t46_n31: return 0.007586566265672445f;
  t46_n32: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t46_n34;
  t46_n33: return 0.015236495062708855f;
  t46_n34: return 0.049572307616472244f;
  t46_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n53;
  t46_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t46_n44;
  t46_n37: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t46_n41;
  t46_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t46_n40;
  t46_n39: return 0.02336304262280464f;
  t46_n40: return 0.05590163916349411f;
  t46_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t46_n43;
  t46_n42: return -0.06495781242847443f;
  t46_n43: return 0.04518408328294754f;
  t46_n44: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t46_n50;
  t46_n45: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t46_n49;
  t46_n46: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t46_n48;
  t46_n47: return 0.019014764577150345f;
  t46_n48: return 0.1589963287115097f;
  t46_n49: return -0.13516323268413544f;
  t46_n50: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t46_n52;
  t46_n51: return 0.06598098576068878f;
  t46_n52: return -0.08803544193506241f;
  t46_n53: return 0.06871910393238068f;
  t46_n54: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t46_n64;
  t46_n55: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t46_n63;
  t46_n56: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t46_n58;
  t46_n57: return -0.0006445946637541056f;
  t46_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t46_n60;
  t46_n59: return 0.025755340233445168f;
  t46_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t46_n62;
  t46_n61: return 0.03452155739068985f;
  t46_n62: return 0.10597719252109528f;
  t46_n63: return -0.018202826380729675f;
  t46_n64: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t46_n80;
  t46_n65: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t46_n79;
  t46_n66: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t46_n78;
  t46_n67: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t46_n75;
  t46_n68: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n70;
  t46_n69: return 0.010832443833351135f;
  t46_n70: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t46_n72;
  t46_n71: return 0.0203749667853117f;
  t46_n72: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t46_n74;
  t46_n73: return 0.17090508341789246f;
  t46_n74: return 0.09024835377931595f;
  t46_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t46_n77;
  t46_n76: return -0.15692877769470215f;
  t46_n77: return 0.016853345558047295f;
  t46_n78: return -0.06949755549430847f;
  t46_n79: return -0.0856887623667717f;
  t46_n80: return -0.0668870061635971f;
  t46_n81: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t46_n155;
  t46_n82: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t46_n150;
  t46_n83: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n127;
  t46_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t46_n124;
  t46_n85: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t46_n121;
  t46_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t46_n96;
  t46_n87: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t46_n91;
  t46_n88: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t46_n90;
  t46_n89: return 0.0023695784620940685f;
  t46_n90: return -0.03430650383234024f;
  t46_n91: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t46_n95;
  t46_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n94;
  t46_n93: return -0.1952972412109375f;
  t46_n94: return 0.03139905259013176f;
  t46_n95: return -0.02130764164030552f;
  t46_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t46_n98;
  t46_n97: return -0.01534914132207632f;
  t46_n98: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t46_n104;
  t46_n99: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t46_n103;
  t46_n100: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t46_n102;
  t46_n101: return 0.023837324231863022f;
  t46_n102: return -0.06227140501141548f;
  t46_n103: return 0.056347813457250595f;
  t46_n104: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t46_n120;
  t46_n105: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t46_n107;
  t46_n106: return 0.002903748070821166f;
  t46_n107: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t46_n111;
  t46_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t46_n110;
  t46_n109: return -0.05641935393214226f;
  t46_n110: return 0.012478293851017952f;
  t46_n111: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t46_n119;
  t46_n112: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t46_n118;
  t46_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t46_n117;
  t46_n114: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t46_n116;
  t46_n115: return 0.03452964499592781f;
  t46_n116: return 0.0557672381401062f;
  t46_n117: return 0.06277238577604294f;
  t46_n118: return -0.1301860809326172f;
  t46_n119: return -0.08886168152093887f;
  t46_n120: return 0.06259544938802719f;
  t46_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n123;
  t46_n122: return -0.11685195565223694f;
  t46_n123: return -0.0031392041128128767f;
  t46_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t46_n126;
  t46_n125: return 0.01898307539522648f;
  t46_n126: return 0.04514315724372864f;
  t46_n127: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t46_n147;
  t46_n128: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t46_n146;
  t46_n129: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t46_n139;
  t46_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t46_n134;
  t46_n131: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t46_n133;
  t46_n132: return -0.01608211360871792f;
  t46_n133: return 0.08168568462133408f;
  t46_n134: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t46_n136;
  t46_n135: return 0.02121652662754059f;
  t46_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t46_n138;
  t46_n137: return 0.021250471472740173f;
  t46_n138: return 0.06954123079776764f;
  t46_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t46_n145;
  t46_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t46_n144;
  t46_n141: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t46_n143;
  t46_n142: return 0.00803927518427372f;
  t46_n143: return -0.08879710733890533f;
  t46_n144: return 0.03429022431373596f;
  t46_n145: return 0.045765768736600876f;
  t46_n146: return -0.06297170370817184f;
  t46_n147: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t46_n149;
  t46_n148: return 0.1299510896205902f;
  t46_n149: return 0.06179330497980118f;
  t46_n150: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t46_n154;
  t46_n151: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t46_n153;
  t46_n152: return 0.11810944229364395f;
  t46_n153: return 0.056676629930734634f;
  t46_n154: return 0.005503247492015362f;
  t46_n155: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t46_n181;
  t46_n156: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t46_n180;
  t46_n157: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t46_n171;
  t46_n158: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t46_n160;
  t46_n159: return 0.018078207969665527f;
  t46_n160: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t46_n166;
  t46_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t46_n165;
  t46_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n164;
  t46_n163: return -0.01748245768249035f;
  t46_n164: return 0.030304675921797752f;
  t46_n165: return 0.07193909585475922f;
  t46_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t46_n168;
  t46_n167: return -0.04583126679062843f;
  t46_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t46_n170;
  t46_n169: return -0.5559026598930359f;
  t46_n170: return 0.048471469432115555f;
  t46_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t46_n173;
  t46_n172: return 0.011800632812082767f;
  t46_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n175;
  t46_n174: return 0.039447031915187836f;
  t46_n175: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t46_n177;
  t46_n176: return 0.06029360741376877f;
  t46_n177: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t46_n179;
  t46_n178: return 0.14859816431999207f;
  t46_n179: return 0.11651960015296936f;
  t46_n180: return 0.04869041219353676f;
  t46_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t46_n243;
  t46_n182: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t46_n220;
  t46_n183: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t46_n217;
  t46_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t46_n186;
  t46_n185: return -0.005392987746745348f;
  t46_n186: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t46_n214;
  t46_n187: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t46_n205;
  t46_n188: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t46_n198;
  t46_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t46_n197;
  t46_n190: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t46_n196;
  t46_n191: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t46_n193;
  t46_n192: return -0.009413576684892178f;
  t46_n193: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t46_n195;
  t46_n194: return 0.029009006917476654f;
  t46_n195: return 0.094138003885746f;
  t46_n196: return -0.04647063463926315f;
  t46_n197: return 0.03131360933184624f;
  t46_n198: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t46_n204;
  t46_n199: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t46_n201;
  t46_n200: return 0.02737342193722725f;
  t46_n201: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t46_n203;
  t46_n202: return 0.04918183758854866f;
  t46_n203: return 0.0917113795876503f;
  t46_n204: return 0.028153391554951668f;
  t46_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t46_n207;
  t46_n206: return -0.09054682403802872f;
  t46_n207: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t46_n213;
  t46_n208: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t46_n210;
  t46_n209: return -0.06440003961324692f;
  t46_n210: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t46_n212;
  t46_n211: return 0.020384814590215683f;
  t46_n212: return 0.06649159640073776f;
  t46_n213: return -0.006282659247517586f;
  t46_n214: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t46_n216;
  t46_n215: return -0.07434076070785522f;
  t46_n216: return 0.013886158354580402f;
  t46_n217: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t46_n219;
  t46_n218: return 0.028077594935894012f;
  t46_n219: return 0.05980170890688896f;
  t46_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t46_n232;
  t46_n221: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t46_n225;
  t46_n222: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t46_n224;
  t46_n223: return 0.002394067822024226f;
  t46_n224: return -0.09620208293199539f;
  t46_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t46_n227;
  t46_n226: return -0.0003018011921085417f;
  t46_n227: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t46_n229;
  t46_n228: return 0.02115793526172638f;
  t46_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t46_n231;
  t46_n230: return 0.03882298991084099f;
  t46_n231: return 0.06551193445920944f;
  t46_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t46_n238;
  t46_n233: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t46_n235;
  t46_n234: return -0.09845935553312302f;
  t46_n235: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n237;
  t46_n236: return 0.019884604960680008f;
  t46_n237: return 0.08342903852462769f;
  t46_n238: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t46_n242;
  t46_n239: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n241;
  t46_n240: return 0.03774401172995567f;
  t46_n241: return 0.06337829679250717f;
  t46_n242: return -0.14230987429618835f;
  t46_n243: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t46_n245;
  t46_n244: return 0.03988618403673172f;
  t46_n245: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t46_n247;
  t46_n246: return 0.015439625829458237f;
  t46_n247: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t46_n249;
  t46_n248: return 0.10387355834245682f;
  t46_n249: return 0.0781632587313652f;
  t46_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t46_n254;
  t46_n251: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t46_n253;
  t46_n252: return 0.022963320836424828f;
  t46_n253: return 0.04229736328125f;
  t46_n254: return 0.07247808575630188f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree47(const Example& E) {
  t47_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t47_n250;
  t47_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t47_n161;
  t47_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t47_n44;
  t47_n3: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t47_n35;
  t47_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n26;
  t47_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t47_n23;
  t47_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t47_n20;
  t47_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n9;
  t47_n8: return -0.1134573444724083f;
  t47_n9: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t47_n15;
  t47_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t47_n12;
  t47_n11: return -0.05595478415489197f;
  t47_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n14;
  t47_n13: return -0.013204323127865791f;
  t47_n14: return 0.017127344384789467f;
  t47_n15: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n19;
  t47_n16: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t47_n18;
  t47_n17: return -0.01577862538397312f;
  t47_n18: return 0.018092239275574684f;
  t47_n19: return 0.05389392375946045f;
  t47_n20: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t47_n22;
  t47_n21: return 0.0033973208628594875f;
  t47_n22: return -0.027813805267214775f;
  t47_n23: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t47_n25;
  t47_n24: return 0.03295128419995308f;
  t47_n25: return 0.0019363068277016282f;
  t47_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t47_n28;
  t47_n27: return -0.0677555724978447f;
  t47_n28: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t47_n30;
  t47_n29: return 0.006550418678671122f;
  t47_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t47_n32;
  t47_n31: return 0.00912558939307928f;
  t47_n32: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t47_n34;
  t47_n33: return 0.036980897188186646f;
  t47_n34: return 0.00841135811060667f;
  t47_n35: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t47_n39;
  t47_n36: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n38;
  t47_n37: return 0.02514943666756153f;
  t47_n38: return 0.05603346228599548f;
  t47_n39: if (E.getNumReferences() >= 3309029376 /*3004.5*/) goto t47_n43;
  t47_n40: if (E.getNumReferences() >= 3308949504 /*2985.0*/) goto t47_n42;
  t47_n41: return 0.05705125629901886f;
  t47_n42: return 0.1470845490694046f;
  t47_n43: return 0.059422995895147324f;
  t47_n44: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t47_n158;
  t47_n45: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n155;
  t47_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t47_n94;
  t47_n47: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t47_n67;
  t47_n48: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t47_n64;
  t47_n49: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t47_n51;
  t47_n50: return -0.019217321649193764f;
  t47_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n53;
  t47_n52: return 0.005973609630018473f;
  t47_n53: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t47_n55;
  t47_n54: return 0.03834804520010948f;
  t47_n55: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t47_n63;
  t47_n56: if (E.getNumReferences() >= 3355438208 /*131033.0*/) goto t47_n62;
  t47_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n59;
  t47_n58: return -0.13976381719112396f;
  t47_n59: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t47_n61;
  t47_n60: return 0.1287935972213745f;
  t47_n61: return -0.0931992381811142f;
  t47_n62: return 0.06622854620218277f;
  t47_n63: return 0.028041573241353035f;
  t47_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t47_n66;
  t47_n65: return -0.12164463847875595f;
  t47_n66: return -0.06614356487989426f;
  t47_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t47_n85;
  t47_n68: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t47_n84;
  t47_n69: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t47_n71;
  t47_n70: return -0.035097382962703705f;
  t47_n71: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t47_n83;
  t47_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t47_n74;
  t47_n73: return -0.043260250240564346f;
  t47_n74: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t47_n82;
  t47_n75: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t47_n79;
  t47_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n78;
  t47_n77: return -0.01609387993812561f;
  t47_n78: return 0.016190404072403908f;
  t47_n79: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t47_n81;
  t47_n80: return 0.018150243908166885f;
  t47_n81: return -0.04380802437663078f;
  t47_n82: return -0.11747873574495316f;
  t47_n83: return -0.042230024933815f;
  t47_n84: return 0.025441892445087433f;
  t47_n85: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t47_n87;
  t47_n86: return 0.0037377553526312113f;
  t47_n87: if (E.getNumReferences() >= 3352491648 /*108013.0*/) goto t47_n93;
  t47_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t47_n92;
  t47_n89: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t47_n91;
  t47_n90: return -0.0022924940567463636f;
  t47_n91: return 0.10424511879682541f;
  t47_n92: return 0.04573701694607735f;
  t47_n93: return -0.04801083356142044f;
  t47_n94: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t47_n140;
  t47_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t47_n99;
  t47_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t47_n98;
  t47_n97: return -0.1452011615037918f;
  t47_n98: return 0.014868457801640034f;
  t47_n99: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t47_n135;
  t47_n100: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t47_n104;
  t47_n101: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t47_n103;
  t47_n102: return 0.0052618649788200855f;
  t47_n103: return -0.040982648730278015f;
  t47_n104: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t47_n112;
  t47_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t47_n107;
  t47_n106: return -0.016047563403844833f;
  t47_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t47_n109;
  t47_n108: return 0.0040825288742780685f;
  t47_n109: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t47_n111;
  t47_n110: return 0.08402464538812637f;
  t47_n111: return 0.052692949771881104f;
  t47_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t47_n124;
  t47_n113: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t47_n123;
  t47_n114: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t47_n120;
  t47_n115: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t47_n119;
  t47_n116: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t47_n118;
  t47_n117: return 0.010748284868896008f;
  t47_n118: return 0.044199179857969284f;
  t47_n119: return 0.0016557982889935374f;
  t47_n120: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t47_n122;
  t47_n121: return 0.021583879366517067f;
  t47_n122: return 0.09247621148824692f;
  t47_n123: return 0.0015814144862815738f;
  t47_n124: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t47_n126;
  t47_n125: return -0.07128319144248962f;
  t47_n126: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t47_n134;
  t47_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t47_n133;
  t47_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t47_n130;
  t47_n129: return -0.01413983665406704f;
  t47_n130: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t47_n132;
  t47_n131: return 0.029470393434166908f;
  t47_n132: return 0.057381726801395416f;
  t47_n133: return 0.0990758016705513f;
  t47_n134: return -0.05937083438038826f;
  t47_n135: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t47_n137;
  t47_n136: return 0.045294713228940964f;
  t47_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n139;
  t47_n138: return -0.05838087946176529f;
  t47_n139: return -0.01565297693014145f;
  t47_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t47_n148;
  t47_n141: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t47_n143;
  t47_n142: return -0.046917881816625595f;
  t47_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n145;
  t47_n144: return -0.032559726387262344f;
  t47_n145: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t47_n147;
  t47_n146: return 0.07142004370689392f;
  t47_n147: return 0.007960844784975052f;
  t47_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t47_n150;
  t47_n149: return -0.0028432549443095922f;
  t47_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t47_n152;
  t47_n151: return 0.06968940794467926f;
  t47_n152: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t47_n154;
  t47_n153: return 0.04326716437935829f;
  t47_n154: return 0.1370518058538437f;
  t47_n155: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t47_n157;
  t47_n156: return 0.004859868437051773f;
  t47_n157: return 0.04340120404958725f;
  t47_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n160;
  t47_n159: return -0.40594643354415894f;
  t47_n160: return 0.04255858436226845f;
  t47_n161: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t47_n213;
  t47_n162: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t47_n206;
  t47_n163: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t47_n167;
  t47_n164: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t47_n166;
  t47_n165: return 0.014884349890053272f;
  t47_n166: return 0.053906120359897614f;
  t47_n167: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t47_n195;
  t47_n168: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t47_n192;
  t47_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t47_n191;
  t47_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t47_n184;
  t47_n171: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t47_n179;
  t47_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t47_n174;
  t47_n173: return -0.02983848936855793f;
  t47_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t47_n176;
  t47_n175: return 0.0006414134986698627f;
  t47_n176: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t47_n178;
  t47_n177: return -0.013855344615876675f;
  t47_n178: return 0.03230806812644005f;
  t47_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t47_n183;
  t47_n180: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t47_n182;
  t47_n181: return 0.003405379131436348f;
  t47_n182: return -0.06493663787841797f;
  t47_n183: return 0.03665409982204437f;
  t47_n184: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t47_n190;
  t47_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n187;
  t47_n186: return -0.009328465908765793f;
  t47_n187: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t47_n189;
  t47_n188: return 0.025556445121765137f;
  t47_n189: return 0.005804743152111769f;
  t47_n190: return 0.026249827817082405f;
  t47_n191: return 0.027172280475497246f;
  t47_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n194;
  t47_n193: return -0.05452028661966324f;
  t47_n194: return 0.004376117140054703f;
  t47_n195: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t47_n205;
  t47_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t47_n198;
  t47_n197: return 0.02441820316016674f;
  t47_n198: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t47_n200;
  t47_n199: return 0.009338932111859322f;
  t47_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n204;
  t47_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t47_n203;
  t47_n202: return -0.26822876930236816f;
  t47_n203: return 0.019780393689870834f;
  t47_n204: return 0.04106542095541954f;
  t47_n205: return -0.02065844088792801f;
  t47_n206: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t47_n210;
  t47_n207: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t47_n209;
  t47_n208: return 0.03884514421224594f;
  t47_n209: return 0.07252568751573563f;
  t47_n210: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t47_n212;
  t47_n211: return 0.060232244431972504f;
  t47_n212: return 0.014363494701683521f;
  t47_n213: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t47_n245;
  t47_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n230;
  t47_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t47_n219;
  t47_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t47_n218;
  t47_n217: return -0.146667018532753f;
  t47_n218: return 0.008779007010161877f;
  t47_n219: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t47_n229;
  t47_n220: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t47_n224;
  t47_n221: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t47_n223;
  t47_n222: return 0.016039978712797165f;
  t47_n223: return 0.02949659712612629f;
  t47_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t47_n226;
  t47_n225: return 0.009151963517069817f;
  t47_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t47_n228;
  t47_n227: return 0.00597270205616951f;
  t47_n228: return 0.034266263246536255f;
  t47_n229: return 0.03499872609972954f;
  t47_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t47_n232;
  t47_n231: return 0.022298451513051987f;
  t47_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t47_n234;
  t47_n233: return 0.029948949813842773f;
  t47_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t47_n236;
  t47_n235: return 0.04757456108927727f;
  t47_n236: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t47_n238;
  t47_n237: return 0.021351773291826248f;
  t47_n238: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t47_n240;
  t47_n239: return 0.07552655041217804f;
  t47_n240: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t47_n244;
  t47_n241: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t47_n243;
  t47_n242: return -0.12317384779453278f;
  t47_n243: return 0.12479980289936066f;
  t47_n244: return 0.036234866827726364f;
  t47_n245: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t47_n249;
  t47_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t47_n248;
  t47_n247: return 0.032560765743255615f;
  t47_n248: return 0.08346764743328094f;
  t47_n249: return 0.05568421259522438f;
  t47_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t47_n254;
  t47_n251: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t47_n253;
  t47_n252: return 0.033864576369524f;
  t47_n253: return 0.07086992263793945f;
  t47_n254: return -0.22431062161922455f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree48(const Example& E) {
  t48_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t48_n252;
  t48_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t48_n77;
  t48_n2: if (E.getNumReferences() >= 3325332480 /*11555.0*/) goto t48_n72;
  t48_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t48_n37;
  t48_n4: if (E.getNumReferences() >= 3271229440 /*125.5*/) goto t48_n34;
  t48_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n19;
  t48_n6: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t48_n18;
  t48_n7: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t48_n11;
  t48_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t48_n10;
  t48_n9: return -0.015840673819184303f;
  t48_n10: return 0.015196210704743862f;
  t48_n11: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t48_n13;
  t48_n12: return 0.052369412034749985f;
  t48_n13: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t48_n17;
  t48_n14: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t48_n16;
  t48_n15: return 0.0025677175726741552f;
  t48_n16: return -0.14016827940940857f;
  t48_n17: return -0.027786871418356895f;
  t48_n18: return -0.06932717561721802f;
  t48_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t48_n31;
  t48_n20: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t48_n22;
  t48_n21: return -0.010355884209275246f;
  t48_n22: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t48_n30;
  t48_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t48_n27;
  t48_n24: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t48_n26;
  t48_n25: return 0.015550102107226849f;
  t48_n26: return 0.04829734191298485f;
  t48_n27: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t48_n29;
  t48_n28: return 0.05650855228304863f;
  t48_n29: return -0.015615857206285f;
  t48_n30: return -0.10543631762266159f;
  t48_n31: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t48_n33;
  t48_n32: return 0.012128656730055809f;
  t48_n33: return -0.07580244541168213f;
  t48_n34: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t48_n36;
  t48_n35: return 0.029913943260908127f;
  t48_n36: return 0.013015956617891788f;
  t48_n37: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t48_n47;
  t48_n38: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t48_n46;
  t48_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t48_n43;
  t48_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t48_n42;
  t48_n41: return -0.014100838452577591f;
  t48_n42: return 0.010972058400511742f;
  t48_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t48_n45;
  t48_n44: return -0.06321455538272858f;
  t48_n45: return 0.041829317808151245f;
  t48_n46: return 0.03772295266389847f;
  t48_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t48_n69;
  t48_n48: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t48_n60;
  t48_n49: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t48_n51;
  t48_n50: return -0.02803546003997326f;
  t48_n51: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t48_n59;
  t48_n52: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t48_n54;
  t48_n53: return -0.012265325523912907f;
  t48_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t48_n58;
  t48_n55: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t48_n57;
  t48_n56: return -0.015991639345884323f;
  t48_n57: return 0.022664451971650124f;
  t48_n58: return 0.019651861861348152f;
  t48_n59: return -0.06112951040267944f;
  t48_n60: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t48_n62;
  t48_n61: return -0.08885055780410767f;
  t48_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n64;
  t48_n63: return -0.024403762072324753f;
  t48_n64: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t48_n66;
  t48_n65: return -0.08192294090986252f;
  t48_n66: if (E.getNumReferences() >= 3316463616 /*5543.0*/) goto t48_n68;
  t48_n67: return 0.07227473706007004f;
  t48_n68: return -0.007692262995988131f;
  t48_n69: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t48_n71;
  t48_n70: return 0.025392672047019005f;
  t48_n71: return -0.14424121379852295f;
  t48_n72: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t48_n76;
  t48_n73: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t48_n75;
  t48_n74: return 0.027663839980959892f;
  t48_n75: return 0.06455206871032715f;
  t48_n76: return 0.0058952756226062775f;
  t48_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t48_n203;
  t48_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t48_n156;
  t48_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n81;
  t48_n80: return -0.12300314754247665f;
  t48_n81: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t48_n135;
  t48_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t48_n120;
  t48_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t48_n109;
  t48_n84: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t48_n108;
  t48_n85: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t48_n107;
  t48_n86: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t48_n90;
  t48_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t48_n89;
  t48_n88: return -0.1639118492603302f;
  t48_n89: return -0.11451734602451324f;
  t48_n90: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t48_n98;
  t48_n91: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t48_n93;
  t48_n92: return -0.016564158722758293f;
  t48_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t48_n95;
  t48_n94: return 0.017207542434334755f;
  t48_n95: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t48_n97;
  t48_n96: return 0.0721433237195015f;
  t48_n97: return 0.14757053554058075f;
  t48_n98: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t48_n104;
  t48_n99: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t48_n101;
  t48_n100: return 0.06776675581932068f;
  t48_n101: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t48_n103;
  t48_n102: return -0.10050492733716965f;
  t48_n103: return 0.15314218401908875f;
  t48_n104: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t48_n106;
  t48_n105: return 0.10414213687181473f;
  t48_n106: return 0.043520599603652954f;
  t48_n107: return 0.0038071805611252785f;
  t48_n108: return -0.0483546182513237f;
  t48_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t48_n111;
  t48_n110: return 0.03140143305063248f;
  t48_n111: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t48_n115;
  t48_n112: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t48_n114;
  t48_n113: return 0.013571943156421185f;
  t48_n114: return -0.06423155218362808f;
  t48_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t48_n119;
  t48_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n118;
  t48_n117: return -0.010780276730656624f;
  t48_n118: return 0.03479913994669914f;
  t48_n119: return 0.040541794151067734f;
  t48_n120: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t48_n134;
  t48_n121: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t48_n127;
  t48_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t48_n124;
  t48_n123: return -0.11762502789497375f;
  t48_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t48_n126;
  t48_n125: return -0.08282597362995148f;
  t48_n126: return 0.0299917571246624f;
  t48_n127: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t48_n133;
  t48_n128: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t48_n130;
  t48_n129: return 0.033806346356868744f;
  t48_n130: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t48_n132;
  t48_n131: return -0.022840527817606926f;
  t48_n132: return 0.117274209856987f;
  t48_n133: return -0.09265671670436859f;
  t48_n134: return -0.06932314485311508f;
  t48_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t48_n137;
  t48_n136: return -0.0015695919282734394f;
  t48_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t48_n153;
  t48_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n140;
  t48_n139: return -0.017080157995224f;
  t48_n140: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t48_n142;
  t48_n141: return 0.020820874720811844f;
  t48_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t48_n148;
  t48_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t48_n145;
  t48_n144: return 0.027068739756941795f;
  t48_n145: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t48_n147;
  t48_n146: return 0.06878990679979324f;
  t48_n147: return -0.31822389364242554f;
  t48_n148: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t48_n152;
  t48_n149: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t48_n151;
  t48_n150: return 0.07591750472784042f;
  t48_n151: return 0.008419751189649105f;
  t48_n152: return 0.08442479372024536f;
  t48_n153: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t48_n155;
  t48_n154: return 0.005834463518112898f;
  t48_n155: return 0.046517807990312576f;
  t48_n156: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t48_n192;
  t48_n157: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t48_n177;
  t48_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t48_n164;
  t48_n159: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t48_n161;
  t48_n160: return -0.001244599581696093f;
  t48_n161: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t48_n163;
  t48_n162: return 0.03893870860338211f;
  t48_n163: return 0.006339734420180321f;
  t48_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t48_n176;
  t48_n165: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t48_n171;
  t48_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n168;
  t48_n167: return -0.03599412366747856f;
  t48_n168: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t48_n170;
  t48_n169: return 0.02681775763630867f;
  t48_n170: return 0.055968984961509705f;
  t48_n171: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t48_n173;
  t48_n172: return 0.011294256895780563f;
  t48_n173: if (E.getNumReferences() >= 3349612032 /*85516.0*/) goto t48_n175;
  t48_n174: return 0.09883259981870651f;
  t48_n175: return 0.04319296032190323f;
  t48_n176: return -0.10581055283546448f;
  t48_n177: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t48_n183;
  t48_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t48_n180;
  t48_n179: return 0.017023956403136253f;
  t48_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n182;
  t48_n181: return -0.0542471818625927f;
  t48_n182: return 0.033492133021354675f;
  t48_n183: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t48_n189;
  t48_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t48_n188;
  t48_n185: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t48_n187;
  t48_n186: return 0.009764686226844788f;
  t48_n187: return 0.08280552923679352f;
  t48_n188: return 0.043584030121564865f;
  t48_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t48_n191;
  t48_n190: return 0.046510111540555954f;
  t48_n191: return 0.0668119490146637f;
  t48_n192: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t48_n196;
  t48_n193: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t48_n195;
  t48_n194: return 0.030440399423241615f;
  t48_n195: return 0.10764706134796143f;
  t48_n196: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t48_n202;
  t48_n197: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t48_n201;
  t48_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n200;
  t48_n199: return -0.0014934749342501163f;
  t48_n200: return 0.02167442813515663f;
  t48_n201: return -0.1994766891002655f;
  t48_n202: return 0.05073351785540581f;
  t48_n203: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t48_n251;
  t48_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t48_n240;
  t48_n205: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t48_n233;
  t48_n206: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t48_n212;
  t48_n207: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t48_n211;
  t48_n208: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t48_n210;
  t48_n209: return 0.019982578232884407f;
  t48_n210: return 0.054437048733234406f;
  t48_n211: return 0.06985568255186081f;
  t48_n212: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t48_n232;
  t48_n213: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t48_n227;
  t48_n214: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t48_n224;
  t48_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t48_n223;
  t48_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t48_n218;
  t48_n217: return -0.026248635724186897f;
  t48_n218: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t48_n222;
  t48_n219: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t48_n221;
  t48_n220: return 0.18767723441123962f;
  t48_n221: return 0.08510170131921768f;
  t48_n222: return 0.06555041670799255f;
  t48_n223: return 0.027716834098100662f;
  t48_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t48_n226;
  t48_n225: return -0.27425068616867065f;
  t48_n226: return -0.046210210770368576f;
  t48_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t48_n231;
  t48_n228: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t48_n230;
  t48_n229: return 0.0001955212646862492f;
  t48_n230: return 0.042302634567022324f;
  t48_n231: return 0.06172364205121994f;
  t48_n232: return 0.07964762300252914f;
  t48_n233: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t48_n239;
  t48_n234: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t48_n236;
  t48_n235: return 0.04285315051674843f;
  t48_n236: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t48_n238;
  t48_n237: return -0.20333009958267212f;
  t48_n238: return 0.03290332481265068f;
  t48_n239: return 0.0867115929722786f;
  t48_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t48_n250;
  t48_n241: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t48_n249;
  t48_n242: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t48_n248;
  t48_n243: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t48_n245;
  t48_n244: return -0.00664533581584692f;
  t48_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t48_n247;
  t48_n246: return 0.01930636167526245f;
  t48_n247: return 0.08523011952638626f;
  t48_n248: return 0.02980971895158291f;
  t48_n249: return -0.08171381056308746f;
  t48_n250: return 0.05796465650200844f;
  t48_n251: return 0.08980147540569305f;
  t48_n252: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t48_n254;
  t48_n253: return 0.027044471353292465f;
  t48_n254: return 0.04522956907749176f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree49(const Example& E) {
  t49_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t49_n252;
  t49_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t49_n73;
  t49_n2: if (E.getNumReferences() >= 3325332480 /*11555.0*/) goto t49_n62;
  t49_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t49_n45;
  t49_n4: if (E.getNumReferences() >= 3271229440 /*125.5*/) goto t49_n40;
  t49_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n25;
  t49_n6: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t49_n20;
  t49_n7: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t49_n9;
  t49_n8: return 0.02267487905919552f;
  t49_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t49_n13;
  t49_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t49_n12;
  t49_n11: return 0.0386803038418293f;
  t49_n12: return -0.04542795568704605f;
  t49_n13: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t49_n15;
  t49_n14: return -0.01841607317328453f;
  t49_n15: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t49_n19;
  t49_n16: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t49_n18;
  t49_n17: return 0.0042813499458134174f;
  t49_n18: return -0.060992155224084854f;
  t49_n19: return -0.12229486554861069f;
  t49_n20: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n22;
  t49_n21: return -0.02488432079553604f;
  t49_n22: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t49_n24;
  t49_n23: return 0.012840572744607925f;
  t49_n24: return -0.03266623988747597f;
  t49_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t49_n37;
  t49_n26: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t49_n28;
  t49_n27: return -0.005906434264034033f;
  t49_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t49_n34;
  t49_n29: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t49_n31;
  t49_n30: return -0.05511634424328804f;
  t49_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t49_n33;
  t49_n32: return -0.09592384845018387f;
  t49_n33: return 0.02658708393573761f;
  t49_n34: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t49_n36;
  t49_n35: return 0.12504084408283234f;
  t49_n36: return 0.041560493409633636f;
  t49_n37: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t49_n39;
  t49_n38: return 0.011441342532634735f;
  t49_n39: return -0.0741983950138092f;
  t49_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t49_n44;
  t49_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t49_n43;
  t49_n42: return 0.01422168780118227f;
  t49_n43: return 0.053862057626247406f;
  t49_n44: return 0.042265623807907104f;
  t49_n45: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t49_n53;
  t49_n46: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t49_n52;
  t49_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n49;
  t49_n48: return -0.005573473870754242f;
  t49_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t49_n51;
  t49_n50: return -0.05798754468560219f;
  t49_n51: return 0.03940298780798912f;
  t49_n52: return 0.03788820281624794f;
  t49_n53: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t49_n59;
  t49_n54: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t49_n56;
  t49_n55: return -0.01431447546929121f;
  t49_n56: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t49_n58;
  t49_n57: return -0.08670307695865631f;
  t49_n58: return -0.019041357561945915f;
  t49_n59: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t49_n61;
  t49_n60: return 0.023774372413754463f;
  t49_n61: return -0.13480712473392487f;
  t49_n62: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t49_n72;
  t49_n63: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t49_n67;
  t49_n64: if (E.getNumReferences() >= 3358114336 /*172808.5*/) goto t49_n66;
  t49_n65: return 0.028387049213051796f;
  t49_n66: return -0.14920929074287415f;
  t49_n67: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t49_n71;
  t49_n68: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n70;
  t49_n69: return 0.0556669682264328f;
  t49_n70: return 0.11124683916568756f;
  t49_n71: return 0.05696090683341026f;
  t49_n72: return 0.004969597794115543f;
  t49_n73: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t49_n155;
  t49_n74: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t49_n140;
  t49_n75: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t49_n83;
  t49_n76: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t49_n82;
  t49_n77: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t49_n81;
  t49_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t49_n80;
  t49_n79: return 0.012768616899847984f;
  t49_n80: return 0.03477254882454872f;
  t49_n81: return -0.037916913628578186f;
  t49_n82: return 0.030817026272416115f;
  t49_n83: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t49_n131;
  t49_n84: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t49_n130;
  t49_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t49_n123;
  t49_n86: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t49_n122;
  t49_n87: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t49_n119;
  t49_n88: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t49_n94;
  t49_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t49_n91;
  t49_n90: return -0.0020655766129493713f;
  t49_n91: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t49_n93;
  t49_n92: return 0.05937311798334122f;
  t49_n93: return 0.013061432167887688f;
  t49_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t49_n100;
  t49_n95: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n97;
  t49_n96: return 0.0023467272985726595f;
  t49_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t49_n99;
  t49_n98: return 0.0041969395242631435f;
  t49_n99: return 0.023567114025354385f;
  t49_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t49_n118;
  t49_n101: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t49_n117;
  t49_n102: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t49_n106;
  t49_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t49_n105;
  t49_n104: return -0.0088344756513834f;
  t49_n105: return 0.011580881662666798f;
  t49_n106: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t49_n116;
  t49_n107: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t49_n113;
  t49_n108: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t49_n112;
  t49_n109: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t49_n111;
  t49_n110: return 0.028603818267583847f;
  t49_n111: return 0.04457353428006172f;
  t49_n112: return -0.04183942824602127f;
  t49_n113: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t49_n115;
  t49_n114: return 0.054122645407915115f;
  t49_n115: return -0.20827646553516388f;
  t49_n116: return -0.016864430159330368f;
  t49_n117: return 0.03592968359589577f;
  t49_n118: return 0.04296235367655754f;
  t49_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n121;
  t49_n120: return -0.045515596866607666f;
  t49_n121: return 0.005407707765698433f;
  t49_n122: return -0.1117236539721489f;
  t49_n123: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t49_n125;
  t49_n124: return -0.08648519963026047f;
  t49_n125: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t49_n127;
  t49_n126: return 0.03178318962454796f;
  t49_n127: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t49_n129;
  t49_n128: return 0.035225141793489456f;
  t49_n129: return -0.051491789519786835f;
  t49_n130: return -0.06676259636878967f;
  t49_n131: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t49_n139;
  t49_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t49_n136;
  t49_n133: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t49_n135;
  t49_n134: return 0.017133360728621483f;
  t49_n135: return 0.048133574426174164f;
  t49_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n138;
  t49_n137: return -0.03626039996743202f;
  t49_n138: return 0.037532564252614975f;
  t49_n139: return -0.020205404609441757f;
  t49_n140: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t49_n144;
  t49_n141: if (E.getNumReferences() >= 3321306112 /*7907.5*/) goto t49_n143;
  t49_n142: return 0.03477279096841812f;
  t49_n143: return 0.05794639512896538f;
  t49_n144: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t49_n154;
  t49_n145: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t49_n151;
  t49_n146: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t49_n148;
  t49_n147: return 0.06022511422634125f;
  t49_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t49_n150;
  t49_n149: return -0.06407027691602707f;
  t49_n150: return 0.0346432588994503f;
  t49_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t49_n153;
  t49_n152: return 0.018786072731018066f;
  t49_n153: return 0.06330488622188568f;
  t49_n154: return 0.040941108018159866f;
  t49_n155: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t49_n237;
  t49_n156: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t49_n230;
  t49_n157: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t49_n229;
  t49_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n176;
  t49_n159: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t49_n165;
  t49_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t49_n164;
  t49_n161: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t49_n163;
  t49_n162: return -0.016115913167595863f;
  t49_n163: return -0.08316665142774582f;
  t49_n164: return -0.0007818465819582343f;
  t49_n165: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t49_n173;
  t49_n166: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t49_n168;
  t49_n167: return 0.03080286644399166f;
  t49_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t49_n170;
  t49_n169: return 0.06959839165210724f;
  t49_n170: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t49_n172;
  t49_n171: return 0.10185915976762772f;
  t49_n172: return 0.06878120452165604f;
  t49_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n175;
  t49_n174: return -0.036183297634124756f;
  t49_n175: return -0.2713219225406647f;
  t49_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t49_n212;
  t49_n177: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t49_n207;
  t49_n178: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t49_n204;
  t49_n179: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t49_n197;
  t49_n180: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n194;
  t49_n181: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t49_n191;
  t49_n182: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t49_n184;
  t49_n183: return 0.0139527078717947f;
  t49_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t49_n186;
  t49_n185: return -0.010915368795394897f;
  t49_n186: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t49_n190;
  t49_n187: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t49_n189;
  t49_n188: return 0.07970771938562393f;
  t49_n189: return 0.17553994059562683f;
  t49_n190: return -0.05370551720261574f;
  t49_n191: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t49_n193;
  t49_n192: return 0.05921027436852455f;
  t49_n193: return 0.009583202190697193f;
  t49_n194: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t49_n196;
  t49_n195: return 0.03109283745288849f;
  t49_n196: return 0.0840630754828453f;
  t49_n197: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t49_n203;
  t49_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t49_n202;
  t49_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n201;
  t49_n200: return -0.27573978900909424f;
  t49_n201: return -0.030204910784959793f;
  t49_n202: return -0.07302569597959518f;
  t49_n203: return 0.029167721047997475f;
  t49_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t49_n206;
  t49_n205: return 0.014775962568819523f;
  t49_n206: return 0.03852150961756706f;
  t49_n207: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t49_n211;
  t49_n208: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t49_n210;
  t49_n209: return 0.0051263771019876f;
  t49_n210: return 0.07759105414152145f;
  t49_n211: return 0.032798539847135544f;
  t49_n212: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t49_n220;
  t49_n213: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t49_n217;
  t49_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t49_n216;
  t49_n215: return 0.029885049909353256f;
  t49_n216: return 0.061612773686647415f;
  t49_n217: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t49_n219;
  t49_n218: return 0.023041047155857086f;
  t49_n219: return -0.019837889820337296f;
  t49_n220: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t49_n224;
  t49_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n223;
  t49_n222: return -0.010588455945253372f;
  t49_n223: return 0.026256851851940155f;
  t49_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t49_n228;
  t49_n225: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t49_n227;
  t49_n226: return 0.02556304819881916f;
  t49_n227: return -0.35038477182388306f;
  t49_n228: return 0.05255172401666641f;
  t49_n229: return -0.008378199301660061f;
  t49_n230: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t49_n232;
  t49_n231: return -0.06442414224147797f;
  t49_n232: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t49_n236;
  t49_n233: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t49_n235;
  t49_n234: return 0.01765774004161358f;
  t49_n235: return 0.048130180686712265f;
  t49_n236: return -0.02852802351117134f;
  t49_n237: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t49_n241;
  t49_n238: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t49_n240;
  t49_n239: return 0.005519367754459381f;
  t49_n240: return 0.04299835115671158f;
  t49_n241: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t49_n243;
  t49_n242: return 0.04126228392124176f;
  t49_n243: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n251;
  t49_n244: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t49_n250;
  t49_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t49_n247;
  t49_n246: return 0.005026265978813171f;
  t49_n247: if (E.getNumReferences() >= 3298291712 /*1215.5*/) goto t49_n249;
  t49_n248: return 0.02766120247542858f;
  t49_n249: return 0.06439794600009918f;
  t49_n250: return 0.12438005208969116f;
  t49_n251: return 0.06944481283426285f;
  t49_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t49_n254;
  t49_n253: return 0.03057914786040783f;
  t49_n254: return 0.06844270974397659f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree50(const Example& E) {
  t50_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t50_n254;
  t50_n1: if (E.getFractionNameInContext() >= 3191375101 /*0.18019481003284454*/) goto t50_n171;
  t50_n2: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t50_n156;
  t50_n3: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t50_n153;
  t50_n4: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t50_n150;
  t50_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t50_n79;
  t50_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t50_n70;
  t50_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n31;
  t50_n8: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t50_n26;
  t50_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t50_n13;
  t50_n10: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t50_n12;
  t50_n11: return 0.007797681260854006f;
  t50_n12: return 0.037701621651649475f;
  t50_n13: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t50_n25;
  t50_n14: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t50_n20;
  t50_n15: if (E.getFractionNameInContext() >= 3174441576 /*0.04446640610694885*/) goto t50_n19;
  t50_n16: if (E.getNumReferences() >= 3342302208 /*46972.0*/) goto t50_n18;
  t50_n17: return -0.011831644922494888f;
  t50_n18: return -0.16538503766059875f;
  t50_n19: return 0.0071297260001301765f;
  t50_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n22;
  t50_n21: return -0.024938838556408882f;
  t50_n22: if (E.getNumReferences() >= 3345303552 /*58696.0*/) goto t50_n24;
  t50_n23: return 0.10796714574098587f;
  t50_n24: return 0.05135684460401535f;
  t50_n25: return -0.14639674127101898f;
  t50_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t50_n28;
  t50_n27: return -0.08591160178184509f;
  t50_n28: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t50_n30;
  t50_n29: return 0.026224059984087944f;
  t50_n30: return 0.05040516331791878f;
  t50_n31: if (E.getNumReferences() >= 3334946304 /*25503.0*/) goto t50_n61;
  t50_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t50_n46;
  t50_n33: if (E.getNumReferences() >= 3268345856 /*103.5*/) goto t50_n45;
  t50_n34: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t50_n42;
  t50_n35: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t50_n37;
  t50_n36: return 0.020417936146259308f;
  t50_n37: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t50_n39;
  t50_n38: return -0.018117211759090424f;
  t50_n39: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t50_n41;
  t50_n40: return 0.04537419602274895f;
  t50_n41: return -0.10819622874259949f;
  t50_n42: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t50_n44;
  t50_n43: return 0.04422618821263313f;
  t50_n44: return -0.024767369031906128f;
  t50_n45: return 0.06024285405874252f;
  t50_n46: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t50_n60;
  t50_n47: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t50_n49;
  t50_n48: return -0.001802551792934537f;
  t50_n49: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t50_n51;
  t50_n50: return -0.06354043632745743f;
  t50_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t50_n53;
  t50_n52: return -0.10776608437299728f;
  t50_n53: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t50_n59;
  t50_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t50_n56;
  t50_n55: return 0.018547700718045235f;
  t50_n56: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t50_n58;
  t50_n57: return 0.037703126668930054f;
  t50_n58: return 0.058399152010679245f;
  t50_n59: return -0.11318811774253845f;
  t50_n60: return -0.022460779175162315f;
  t50_n61: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t50_n69;
  t50_n62: if (E.getNumReferences() >= 3359032704 /*187158.0*/) goto t50_n64;
  t50_n63: return 0.0490572415292263f;
  t50_n64: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t50_n68;
  t50_n65: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t50_n67;
  t50_n66: return 0.10454949736595154f;
  t50_n67: return 0.1309090256690979f;
  t50_n68: return 0.0675610825419426f;
  t50_n69: return -0.09979459643363953f;
  t50_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n72;
  t50_n71: return -0.0649753212928772f;
  t50_n72: if (E.getNumReferences() >= 3343995264 /*53585.5*/) goto t50_n78;
  t50_n73: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t50_n75;
  t50_n74: return -0.04648430645465851f;
  t50_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t50_n77;
  t50_n76: return 0.026993196457624435f;
  t50_n77: return 0.13850609958171844f;
  t50_n78: return -0.11108022183179855f;
  t50_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t50_n139;
  t50_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t50_n110;
  t50_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t50_n89;
  t50_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t50_n86;
  t50_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n85;
  t50_n84: return -0.1456599235534668f;
  t50_n85: return -0.06524734944105148f;
  t50_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n88;
  t50_n87: return -0.016673775389790535f;
  t50_n88: return 0.06765304505825043f;
  t50_n89: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t50_n93;
  t50_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n92;
  t50_n91: return -0.017480632290244102f;
  t50_n92: return 0.0006613216828554869f;
  t50_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t50_n97;
  t50_n94: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t50_n96;
  t50_n95: return 0.1851627379655838f;
  t50_n96: return 0.0028744458686560392f;
  t50_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t50_n99;
  t50_n98: return -0.11691557615995407f;
  t50_n99: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t50_n109;
  t50_n100: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t50_n106;
  t50_n101: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t50_n105;
  t50_n102: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t50_n104;
  t50_n103: return 0.004160770680755377f;
  t50_n104: return 0.11693311482667923f;
  t50_n105: return -0.019272392615675926f;
  t50_n106: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t50_n108;
  t50_n107: return 0.04252493008971214f;
  t50_n108: return -0.06302369385957718f;
  t50_n109: return -0.0021030153147876263f;
  t50_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t50_n116;
  t50_n111: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t50_n115;
  t50_n112: if (E.getNumReferences() >= 3325781504 /*11993.5*/) goto t50_n114;
  t50_n113: return -0.021119533106684685f;
  t50_n114: return -0.1544654816389084f;
  t50_n115: return 0.005439593456685543f;
  t50_n116: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t50_n134;
  t50_n117: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t50_n119;
  t50_n118: return 0.05072196573019028f;
  t50_n119: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t50_n123;
  t50_n120: if (E.getNumReferences() >= 3279519744 /*249.5*/) goto t50_n122;
  t50_n121: return -0.011888813227415085f;
  t50_n122: return 0.048372749239206314f;
  t50_n123: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t50_n131;
  t50_n124: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t50_n130;
  t50_n125: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t50_n129;
  t50_n126: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t50_n128;
  t50_n127: return 0.06951343268156052f;
  t50_n128: return 0.09596192836761475f;
  t50_n129: return 0.0560961589217186f;
  t50_n130: return 0.034947820007801056f;
  t50_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t50_n133;
  t50_n132: return 0.003852015594020486f;
  t50_n133: return 0.08682464063167572f;
  t50_n134: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t50_n138;
  t50_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t50_n137;
  t50_n136: return 0.047230180352926254f;
  t50_n137: return 0.11366388201713562f;
  t50_n138: return -0.06443848460912704f;
  t50_n139: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t50_n149;
  t50_n140: if (E.getNumReferences() >= 3364003440 /*267507.5*/) goto t50_n148;
  t50_n141: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t50_n147;
  t50_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t50_n146;
  t50_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t50_n145;
  t50_n144: return 0.006405639927834272f;
  t50_n145: return 0.05508217588067055f;
  t50_n146: return 0.018271170556545258f;
  t50_n147: return -0.08471459150314331f;
  t50_n148: return 0.08807612955570221f;
  t50_n149: return 0.049695998430252075f;
  t50_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t50_n152;
  t50_n151: return 0.02656453661620617f;
  t50_n152: return 0.07688898593187332f;
  t50_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t50_n155;
  t50_n154: return -0.07023446261882782f;
  t50_n155: return -0.018455447629094124f;
  t50_n156: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t50_n166;
  t50_n157: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t50_n165;
  t50_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t50_n164;
  t50_n159: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t50_n163;
  t50_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t50_n162;
  t50_n161: return -0.01102107297629118f;
  t50_n162: return 0.024639664217829704f;
  t50_n163: return -0.008121650665998459f;
  t50_n164: return 0.016295064240694046f;
  t50_n165: return 0.03864997252821922f;
  t50_n166: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t50_n170;
  t50_n167: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t50_n169;
  t50_n168: return 0.033119674772024155f;
  t50_n169: return 0.08352499455213547f;
  t50_n170: return 0.08319980651140213f;
  t50_n171: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t50_n187;
  t50_n172: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t50_n186;
  t50_n173: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t50_n185;
  t50_n174: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t50_n176;
  t50_n175: return 0.014447606168687344f;
  t50_n176: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t50_n180;
  t50_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t50_n179;
  t50_n178: return 0.021891387179493904f;
  t50_n179: return 0.06699047237634659f;
  t50_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t50_n182;
  t50_n181: return -0.04451196640729904f;
  t50_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n184;
  t50_n183: return -0.36595916748046875f;
  t50_n184: return 0.04170706123113632f;
  t50_n185: return 0.03215717896819115f;
  t50_n186: return 0.040750645101070404f;
  t50_n187: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t50_n253;
  t50_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t50_n248;
  t50_n189: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t50_n241;
  t50_n190: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t50_n200;
  t50_n191: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t50_n195;
  t50_n192: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t50_n194;
  t50_n193: return 0.011771881952881813f;
  t50_n194: return -0.018701346591114998f;
  t50_n195: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t50_n197;
  t50_n196: return 0.03341034799814224f;
  t50_n197: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t50_n199;
  t50_n198: return 0.07638969272375107f;
  t50_n199: return 0.12537957727909088f;
  t50_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t50_n228;
  t50_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n223;
  t50_n202: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t50_n220;
  t50_n203: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t50_n219;
  t50_n204: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t50_n212;
  t50_n205: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t50_n211;
  t50_n206: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t50_n208;
  t50_n207: return 0.005384459625929594f;
  t50_n208: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t50_n210;
  t50_n209: return -0.10727567970752716f;
  t50_n210: return -0.02865408919751644f;
  t50_n211: return -0.0448652058839798f;
  t50_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t50_n214;
  t50_n213: return 0.010226404294371605f;
  t50_n214: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t50_n216;
  t50_n215: return 0.013012873940169811f;
  t50_n216: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t50_n218;
  t50_n217: return 0.12144908308982849f;
  t50_n218: return -0.047638483345508575f;
  t50_n219: return -0.23273202776908875f;
  t50_n220: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t50_n222;
  t50_n221: return 0.036743782460689545f;
  t50_n222: return -0.05439222604036331f;
  t50_n223: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t50_n225;
  t50_n224: return 0.043171122670173645f;
  t50_n225: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t50_n227;
  t50_n226: return 0.011188757605850697f;
  t50_n227: return -0.0778205469250679f;
  t50_n228: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t50_n240;
  t50_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t50_n231;
  t50_n230: return -0.029534265398979187f;
  t50_n231: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t50_n233;
  t50_n232: return 0.013155731372535229f;
  t50_n233: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t50_n235;
  t50_n234: return 0.030074821785092354f;
  t50_n235: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t50_n239;
  t50_n236: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t50_n238;
  t50_n237: return 0.14368587732315063f;
  t50_n238: return -0.29582011699676514f;
  t50_n239: return 0.07356708496809006f;
  t50_n240: return 0.029183320701122284f;
  t50_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t50_n245;
  t50_n242: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t50_n244;
  t50_n243: return 0.008184760808944702f;
  t50_n244: return 0.048115115612745285f;
  t50_n245: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t50_n247;
  t50_n246: return 0.02504635974764824f;
  t50_n247: return 0.06159157678484917f;
  t50_n248: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t50_n250;
  t50_n249: return 0.02295607514679432f;
  t50_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t50_n252;
  t50_n251: return -0.15470243990421295f;
  t50_n252: return -0.009993448853492737f;
  t50_n253: return -0.08046498894691467f;
  t50_n254: return 0.035357747226953506f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree51(const Example& E) {
  t51_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t51_n116;
  t51_n1: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t51_n107;
  t51_n2: if (E.getNumReferences() >= 3303288832 /*1825.5*/) goto t51_n92;
  t51_n3: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t51_n49;
  t51_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n30;
  t51_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t51_n21;
  t51_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t51_n18;
  t51_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n9;
  t51_n8: return -0.1121119037270546f;
  t51_n9: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t51_n13;
  t51_n10: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n12;
  t51_n11: return -0.01535834837704897f;
  t51_n12: return -0.12356028705835342f;
  t51_n13: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n15;
  t51_n14: return 0.011383644305169582f;
  t51_n15: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t51_n17;
  t51_n16: return 0.06140308082103729f;
  t51_n17: return -0.3595663607120514f;
  t51_n18: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t51_n20;
  t51_n19: return 0.0030881878919899464f;
  t51_n20: return -0.025087304413318634f;
  t51_n21: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t51_n29;
  t51_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n26;
  t51_n23: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t51_n25;
  t51_n24: return -0.09582530707120895f;
  t51_n25: return 0.062103476375341415f;
  t51_n26: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n28;
  t51_n27: return 0.02964862249791622f;
  t51_n28: return 0.07873047888278961f;
  t51_n29: return 0.002554848324507475f;
  t51_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t51_n46;
  t51_n31: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t51_n33;
  t51_n32: return -0.007229749578982592f;
  t51_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t51_n43;
  t51_n34: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t51_n36;
  t51_n35: return -0.04857299104332924f;
  t51_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t51_n38;
  t51_n37: return -0.09411177039146423f;
  t51_n38: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t51_n40;
  t51_n39: return 0.03453751280903816f;
  t51_n40: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t51_n42;
  t51_n41: return -0.02573755942285061f;
  t51_n42: return 0.02655174396932125f;
  t51_n43: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t51_n45;
  t51_n44: return 0.11624029278755188f;
  t51_n45: return 0.03558894246816635f;
  t51_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t51_n48;
  t51_n47: return 0.011587811633944511f;
  t51_n48: return -0.07105395197868347f;
  t51_n49: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t51_n57;
  t51_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t51_n54;
  t51_n51: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t51_n53;
  t51_n52: return -0.0062433574348688126f;
  t51_n53: return 0.0344853550195694f;
  t51_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t51_n56;
  t51_n55: return -0.04530901461839676f;
  t51_n56: return 0.04026065021753311f;
  t51_n57: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t51_n91;
  t51_n58: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t51_n88;
  t51_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t51_n61;
  t51_n60: return -0.11694733053445816f;
  t51_n61: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t51_n63;
  t51_n62: return -0.025715501978993416f;
  t51_n63: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t51_n83;
  t51_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t51_n80;
  t51_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t51_n67;
  t51_n66: return -0.1274336278438568f;
  t51_n67: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t51_n79;
  t51_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t51_n70;
  t51_n69: return -0.05170639976859093f;
  t51_n70: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t51_n72;
  t51_n71: return -0.011473488062620163f;
  t51_n72: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t51_n76;
  t51_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t51_n75;
  t51_n74: return 0.009849580004811287f;
  t51_n75: return 0.046416718512773514f;
  t51_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n78;
  t51_n77: return -0.1584433764219284f;
  t51_n78: return 0.061891861259937286f;
  t51_n79: return -0.037672702223062515f;
  t51_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t51_n82;
  t51_n81: return 0.011802656576037407f;
  t51_n82: return 0.0796661302447319f;
  t51_n83: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t51_n87;
  t51_n84: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n86;
  t51_n85: return -0.0048095169477164745f;
  t51_n86: return 0.018036291003227234f;
  t51_n87: return -0.11809152364730835f;
  t51_n88: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n90;
  t51_n89: return 0.023266324773430824f;
  t51_n90: return -0.1745585799217224f;
  t51_n91: return -0.0549892783164978f;
  t51_n92: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t51_n104;
  t51_n93: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n103;
  t51_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t51_n100;
  t51_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t51_n99;
  t51_n96: if (E.getNumReferences() >= 3303448576 /*1845.0*/) goto t51_n98;
  t51_n97: return 0.15211866796016693f;
  t51_n98: return 0.021513618528842926f;
  t51_n99: return 0.04283977299928665f;
  t51_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t51_n102;
  t51_n101: return -0.06255916506052017f;
  t51_n102: return 0.01973395235836506f;
  t51_n103: return 0.05879303812980652f;
  t51_n104: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t51_n106;
  t51_n105: return -0.0004214816144667566f;
  t51_n106: return 0.031586967408657074f;
  t51_n107: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n115;
  t51_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t51_n114;
  t51_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t51_n113;
  t51_n110: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n112;
  t51_n111: return 0.010395677760243416f;
  t51_n112: return 0.04789023846387863f;
  t51_n113: return 0.08358123153448105f;
  t51_n114: return 0.03570019081234932f;
  t51_n115: return -0.08906982839107513f;
  t51_n116: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t51_n254;
  t51_n117: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t51_n183;
  t51_n118: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t51_n168;
  t51_n119: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t51_n137;
  t51_n120: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t51_n132;
  t51_n121: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t51_n127;
  t51_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t51_n124;
  t51_n123: return -0.006883667781949043f;
  t51_n124: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n126;
  t51_n125: return 0.020924126729369164f;
  t51_n126: return -0.11056222766637802f;
  t51_n127: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t51_n129;
  t51_n128: return -0.06534408777952194f;
  t51_n129: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t51_n131;
  t51_n130: return 0.02244790457189083f;
  t51_n131: return -0.007334337569773197f;
  t51_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t51_n136;
  t51_n133: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n135;
  t51_n134: return 0.026704687625169754f;
  t51_n135: return 0.07162024825811386f;
  t51_n136: return 0.051126834005117416f;
  t51_n137: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t51_n163;
  t51_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t51_n154;
  t51_n139: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t51_n151;
  t51_n140: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t51_n150;
  t51_n141: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t51_n143;
  t51_n142: return -0.002617946360260248f;
  t51_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t51_n145;
  t51_n144: return 0.0030540956649929285f;
  t51_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t51_n149;
  t51_n146: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t51_n148;
  t51_n147: return 0.0117153599858284f;
  t51_n148: return 0.031990811228752136f;
  t51_n149: return 0.04035293683409691f;
  t51_n150: return -0.10840378701686859f;
  t51_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n153;
  t51_n152: return -0.04594539478421211f;
  t51_n153: return 0.005932931322604418f;
  t51_n154: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t51_n158;
  t51_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t51_n157;
  t51_n156: return 0.0346776619553566f;
  t51_n157: return 0.08065859973430634f;
  t51_n158: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t51_n162;
  t51_n159: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n161;
  t51_n160: return 0.016760069876909256f;
  t51_n161: return -0.019028954207897186f;
  t51_n162: return 0.04713717848062515f;
  t51_n163: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t51_n165;
  t51_n164: return -0.08978009968996048f;
  t51_n165: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t51_n167;
  t51_n166: return 0.024615367874503136f;
  t51_n167: return -0.033927690237760544f;
  t51_n168: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t51_n172;
  t51_n169: if (E.getNumReferences() >= 3321306112 /*7907.5*/) goto t51_n171;
  t51_n170: return 0.03396828845143318f;
  t51_n171: return 0.05562294274568558f;
  t51_n172: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t51_n174;
  t51_n173: return 0.048495542258024216f;
  t51_n174: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t51_n182;
  t51_n175: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t51_n177;
  t51_n176: return -0.031240813434123993f;
  t51_n177: if (E.getNumReferences() >= 3355264064 /*129672.5*/) goto t51_n181;
  t51_n178: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t51_n180;
  t51_n179: return 0.02859473042190075f;
  t51_n180: return 0.100898377597332f;
  t51_n181: return -0.14148636162281036f;
  t51_n182: return 0.03939212113618851f;
  t51_n183: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t51_n249;
  t51_n184: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t51_n220;
  t51_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t51_n195;
  t51_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n188;
  t51_n187: return -0.01794356480240822f;
  t51_n188: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t51_n194;
  t51_n189: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t51_n191;
  t51_n190: return 0.01360731478780508f;
  t51_n191: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t51_n193;
  t51_n192: return -0.08247840404510498f;
  t51_n193: return 0.010681849904358387f;
  t51_n194: return -0.07604867964982986f;
  t51_n195: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t51_n217;
  t51_n196: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t51_n204;
  t51_n197: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t51_n203;
  t51_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t51_n202;
  t51_n199: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t51_n201;
  t51_n200: return 0.0012086721835657954f;
  t51_n201: return 0.016547854989767075f;
  t51_n202: return 0.02017570100724697f;
  t51_n203: return -0.010345007292926311f;
  t51_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t51_n206;
  t51_n205: return 0.009219408966600895f;
  t51_n206: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t51_n212;
  t51_n207: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t51_n211;
  t51_n208: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t51_n210;
  t51_n209: return 0.034208573400974274f;
  t51_n210: return 0.07306509464979172f;
  t51_n211: return 0.02321823313832283f;
  t51_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t51_n214;
  t51_n213: return 0.013796490617096424f;
  t51_n214: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t51_n216;
  t51_n215: return 0.05766594782471657f;
  t51_n216: return 0.16484467685222626f;
  t51_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t51_n219;
  t51_n218: return 0.02356981299817562f;
  t51_n219: return 0.060863468796014786f;
  t51_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t51_n236;
  t51_n221: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t51_n223;
  t51_n222: return 0.03588617965579033f;
  t51_n223: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t51_n231;
  t51_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t51_n230;
  t51_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t51_n227;
  t51_n226: return -0.03051225282251835f;
  t51_n227: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t51_n229;
  t51_n228: return 0.1509227454662323f;
  t51_n229: return 0.039059117436409f;
  t51_n230: return 0.0314410962164402f;
  t51_n231: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t51_n235;
  t51_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t51_n234;
  t51_n233: return -0.06267203390598297f;
  t51_n234: return 0.027325917035341263f;
  t51_n235: return 0.11078296601772308f;
  t51_n236: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t51_n248;
  t51_n237: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t51_n239;
  t51_n238: return 0.004549737554043531f;
  t51_n239: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t51_n241;
  t51_n240: return 0.03854887932538986f;
  t51_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t51_n243;
  t51_n242: return 0.07431341707706451f;
  t51_n243: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t51_n247;
  t51_n244: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t51_n246;
  t51_n245: return -0.11147429794073105f;
  t51_n246: return 0.11775468289852142f;
  t51_n247: return 0.03333276882767677f;
  t51_n248: return 0.08893795311450958f;
  t51_n249: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t51_n251;
  t51_n250: return 0.0500020869076252f;
  t51_n251: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t51_n253;
  t51_n252: return 0.028445115312933922f;
  t51_n253: return 0.08169223368167877f;
  t51_n254: return 0.04121404141187668f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree52(const Example& E) {
  t52_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t52_n108;
  t52_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t52_n81;
  t52_n2: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t52_n72;
  t52_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n53;
  t52_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t52_n42;
  t52_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t52_n39;
  t52_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t52_n34;
  t52_n7: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t52_n33;
  t52_n8: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t52_n32;
  t52_n9: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t52_n25;
  t52_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t52_n12;
  t52_n11: return -0.05206241458654404f;
  t52_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t52_n24;
  t52_n13: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t52_n17;
  t52_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n16;
  t52_n15: return -0.020445672795176506f;
  t52_n16: return 0.020920775830745697f;
  t52_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t52_n19;
  t52_n18: return -0.01028459146618843f;
  t52_n19: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t52_n21;
  t52_n20: return -0.09739593416452408f;
  t52_n21: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t52_n23;
  t52_n22: return 0.06993667781352997f;
  t52_n23: return 0.20145928859710693f;
  t52_n24: return 0.02286115102469921f;
  t52_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t52_n31;
  t52_n26: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t52_n30;
  t52_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t52_n29;
  t52_n28: return -0.11672241985797882f;
  t52_n29: return 0.011427977122366428f;
  t52_n30: return 0.004757683724164963f;
  t52_n31: return 0.03993361443281174f;
  t52_n32: return -0.020645540207624435f;
  t52_n33: return -0.12147493660449982f;
  t52_n34: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t52_n36;
  t52_n35: return -0.03601416200399399f;
  t52_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n38;
  t52_n37: return -0.09160660952329636f;
  t52_n38: return 0.046323176473379135f;
  t52_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t52_n41;
  t52_n40: return 0.007371734827756882f;
  t52_n41: return 0.04735911637544632f;
  t52_n42: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t52_n46;
  t52_n43: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t52_n45;
  t52_n44: return 0.049612171947956085f;
  t52_n45: return -0.16132418811321259f;
  t52_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n50;
  t52_n47: if (E.getNumReferences() >= 3273818112 /*162.5*/) goto t52_n49;
  t52_n48: return -0.03224799409508705f;
  t52_n49: return 0.030369816347956657f;
  t52_n50: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t52_n52;
  t52_n51: return 0.01224122941493988f;
  t52_n52: return -0.06343905627727509f;
  t52_n53: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t52_n71;
  t52_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t52_n68;
  t52_n55: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t52_n59;
  t52_n56: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t52_n58;
  t52_n57: return -0.030459614470601082f;
  t52_n58: return 0.025140436366200447f;
  t52_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t52_n63;
  t52_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t52_n62;
  t52_n61: return -0.06976261734962463f;
  t52_n62: return -0.04922432079911232f;
  t52_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t52_n65;
  t52_n64: return 0.17986536026000977f;
  t52_n65: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t52_n67;
  t52_n66: return 0.05974738299846649f;
  t52_n67: return -0.0037876428104937077f;
  t52_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n70;
  t52_n69: return 0.021360518410801888f;
  t52_n70: return 0.06438565254211426f;
  t52_n71: return -0.03726881742477417f;
  t52_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t52_n80;
  t52_n73: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t52_n75;
  t52_n74: return 0.02436167374253273f;
  t52_n75: if (E.getNumReferences() >= 3309029376 /*3004.5*/) goto t52_n79;
  t52_n76: if (E.getNumReferences() >= 3308949504 /*2985.0*/) goto t52_n78;
  t52_n77: return 0.0704355239868164f;
  t52_n78: return 0.14047473669052124f;
  t52_n79: return 0.05041685327887535f;
  t52_n80: return 0.05870278179645538f;
  t52_n81: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t52_n95;
  t52_n82: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t52_n92;
  t52_n83: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t52_n91;
  t52_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t52_n88;
  t52_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t52_n87;
  t52_n86: return -0.02430083602666855f;
  t52_n87: return 0.0008532438077963889f;
  t52_n88: if (E.getNumReferences() >= 3349284736 /*82959.0*/) goto t52_n90;
  t52_n89: return 0.002880425425246358f;
  t52_n90: return 0.04708936810493469f;
  t52_n91: return -0.043254341930150986f;
  t52_n92: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t52_n94;
  t52_n93: return 0.003146946895867586f;
  t52_n94: return 0.07311218231916428f;
  t52_n95: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t52_n107;
  t52_n96: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t52_n106;
  t52_n97: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t52_n105;
  t52_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n102;
  t52_n99: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t52_n101;
  t52_n100: return 0.01317803654819727f;
  t52_n101: return -0.17322753369808197f;
  t52_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t52_n104;
  t52_n103: return -0.1862047016620636f;
  t52_n104: return 0.011726953089237213f;
  t52_n105: return -0.06092318892478943f;
  t52_n106: return -0.09135466068983078f;
  t52_n107: return -0.05795746669173241f;
  t52_n108: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t52_n254;
  t52_n109: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t52_n217;
  t52_n110: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t52_n166;
  t52_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t52_n153;
  t52_n112: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t52_n152;
  t52_n113: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t52_n129;
  t52_n114: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t52_n128;
  t52_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n119;
  t52_n116: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t52_n118;
  t52_n117: return -0.00012632129073608667f;
  t52_n118: return -0.03663361817598343f;
  t52_n119: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t52_n127;
  t52_n120: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n122;
  t52_n121: return 0.0126774488016963f;
  t52_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t52_n124;
  t52_n123: return 0.022259561344981194f;
  t52_n124: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t52_n126;
  t52_n125: return 0.002499293303117156f;
  t52_n126: return 0.04519396275281906f;
  t52_n127: return -0.019268754869699478f;
  t52_n128: return 0.029877537861466408f;
  t52_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n149;
  t52_n130: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t52_n148;
  t52_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t52_n143;
  t52_n132: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t52_n140;
  t52_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n135;
  t52_n134: return -0.007698419503867626f;
  t52_n135: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t52_n137;
  t52_n136: return 0.007601069752126932f;
  t52_n137: if (E.getNumReferences() >= 3349612032 /*85516.0*/) goto t52_n139;
  t52_n138: return 0.10010956227779388f;
  t52_n139: return 0.036197781562805176f;
  t52_n140: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t52_n142;
  t52_n141: return -0.023584390059113503f;
  t52_n142: return 0.02404148317873478f;
  t52_n143: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t52_n147;
  t52_n144: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t52_n146;
  t52_n145: return -0.01856055110692978f;
  t52_n146: return 0.10291920602321625f;
  t52_n147: return -0.10161961615085602f;
  t52_n148: return -0.06020655483007431f;
  t52_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n151;
  t52_n150: return -0.37451475858688354f;
  t52_n151: return 0.030675845220685005f;
  t52_n152: return 0.08824768662452698f;
  t52_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t52_n165;
  t52_n154: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t52_n160;
  t52_n155: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t52_n157;
  t52_n156: return 0.03977624699473381f;
  t52_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n159;
  t52_n158: return -0.09776680171489716f;
  t52_n159: return 0.0023523583076894283f;
  t52_n160: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t52_n164;
  t52_n161: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n163;
  t52_n162: return 0.01617927849292755f;
  t52_n163: return -0.06677845120429993f;
  t52_n164: return 0.05135336145758629f;
  t52_n165: return 0.0410303995013237f;
  t52_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t52_n182;
  t52_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t52_n169;
  t52_n168: return -0.01181876566261053f;
  t52_n169: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t52_n171;
  t52_n170: return 0.008149090223014355f;
  t52_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t52_n173;
  t52_n172: return 0.02149559184908867f;
  t52_n173: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t52_n181;
  t52_n174: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t52_n180;
  t52_n175: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t52_n177;
  t52_n176: return 0.02059670351445675f;
  t52_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t52_n179;
  t52_n178: return 0.03346888720989227f;
  t52_n179: return 0.06777115166187286f;
  t52_n180: return -0.12827402353286743f;
  t52_n181: return 0.012001363560557365f;
  t52_n182: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t52_n216;
  t52_n183: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t52_n215;
  t52_n184: if (E.getNumReferences() >= 3284680704 /*400.5*/) goto t52_n212;
  t52_n185: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t52_n199;
  t52_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type))) goto t52_n196;
  t52_n187: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t52_n195;
  t52_n188: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t52_n194;
  t52_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n191;
  t52_n190: return -0.014480187557637691f;
  t52_n191: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t52_n193;
  t52_n192: return 0.023838387802243233f;
  t52_n193: return -0.013762615621089935f;
  t52_n194: return 0.029537487775087357f;
  t52_n195: return 0.05024109408259392f;
  t52_n196: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t52_n198;
  t52_n197: return 0.046168290078639984f;
  t52_n198: return 0.13185422122478485f;
  t52_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t52_n209;
  t52_n200: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n202;
  t52_n201: return 0.009141246788203716f;
  t52_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n204;
  t52_n203: return -0.12245313078165054f;
  t52_n204: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t52_n208;
  t52_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n207;
  t52_n206: return -0.014985806308686733f;
  t52_n207: return 0.058891747146844864f;
  t52_n208: return 0.044497787952423096f;
  t52_n209: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t52_n211;
  t52_n210: return 0.03266851231455803f;
  t52_n211: return 0.012877172790467739f;
  t52_n212: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t52_n214;
  t52_n213: return 0.03274466469883919f;
  t52_n214: return 0.05372639000415802f;
  t52_n215: return 0.053304273635149f;
  t52_n216: return -0.06098366528749466f;
  t52_n217: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t52_n251;
  t52_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t52_n248;
  t52_n219: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t52_n227;
  t52_n220: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t52_n226;
  t52_n221: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t52_n225;
  t52_n222: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t52_n224;
  t52_n223: return 0.030619075521826744f;
  t52_n224: return 0.13551577925682068f;
  t52_n225: return -0.07210544496774673f;
  t52_n226: return 0.05410680174827576f;
  t52_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t52_n247;
  t52_n228: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t52_n246;
  t52_n229: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t52_n241;
  t52_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t52_n236;
  t52_n231: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t52_n233;
  t52_n232: return -0.03722427785396576f;
  t52_n233: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t52_n235;
  t52_n234: return 0.11094774305820465f;
  t52_n235: return 0.029548177495598793f;
  t52_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t52_n238;
  t52_n237: return -0.22940821945667267f;
  t52_n238: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t52_n240;
  t52_n239: return 0.019549399614334106f;
  t52_n240: return 0.055610425770282745f;
  t52_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t52_n245;
  t52_n242: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t52_n244;
  t52_n243: return -0.0006564091891050339f;
  t52_n244: return 0.03751993179321289f;
  t52_n245: return 0.061094556003808975f;
  t52_n246: return 0.07241897284984589f;
  t52_n247: return 0.039364688098430634f;
  t52_n248: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t52_n250;
  t52_n249: return 0.01952507346868515f;
  t52_n250: return 0.05130638927221298f;
  t52_n251: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t52_n253;
  t52_n252: return 0.12261559814214706f;
  t52_n253: return 0.07485251128673553f;
  t52_n254: return 0.039628565311431885f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree53(const Example& E) {
  t53_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t53_n104;
  t53_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t53_n81;
  t53_n2: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t53_n62;
  t53_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n45;
  t53_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t53_n34;
  t53_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t53_n29;
  t53_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t53_n26;
  t53_n7: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t53_n25;
  t53_n8: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t53_n24;
  t53_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t53_n23;
  t53_n10: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t53_n20;
  t53_n11: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t53_n17;
  t53_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n14;
  t53_n13: return -0.018664300441741943f;
  t53_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t53_n16;
  t53_n15: return -0.027931593358516693f;
  t53_n16: return 0.02135797217488289f;
  t53_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t53_n19;
  t53_n18: return -0.00430144090205431f;
  t53_n19: return 0.04334818571805954f;
  t53_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t53_n22;
  t53_n21: return 0.010027051903307438f;
  t53_n22: return 0.03926096856594086f;
  t53_n23: return -0.0570080429315567f;
  t53_n24: return -0.06285563111305237f;
  t53_n25: return -0.11917109042406082f;
  t53_n26: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t53_n28;
  t53_n27: return -0.052376601845026016f;
  t53_n28: return -0.0027190253604203463f;
  t53_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t53_n33;
  t53_n30: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t53_n32;
  t53_n31: return 0.012943951413035393f;
  t53_n32: return -0.012631678953766823f;
  t53_n33: return 0.04266943782567978f;
  t53_n34: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t53_n38;
  t53_n35: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t53_n37;
  t53_n36: return 0.042590927332639694f;
  t53_n37: return -0.134795680642128f;
  t53_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n42;
  t53_n39: if (E.getNumReferences() >= 3273818112 /*162.5*/) goto t53_n41;
  t53_n40: return -0.03429033234715462f;
  t53_n41: return 0.028079137206077576f;
  t53_n42: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t53_n44;
  t53_n43: return 0.011823197826743126f;
  t53_n44: return -0.059744272381067276f;
  t53_n45: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t53_n61;
  t53_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t53_n60;
  t53_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n59;
  t53_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t53_n58;
  t53_n49: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t53_n51;
  t53_n50: return -0.022292710840702057f;
  t53_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n53;
  t53_n52: return -0.06514354795217514f;
  t53_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t53_n55;
  t53_n54: return 0.1458011418581009f;
  t53_n55: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t53_n57;
  t53_n56: return 0.05615190416574478f;
  t53_n57: return -0.0014266662765294313f;
  t53_n58: return 0.016367489472031593f;
  t53_n59: return 0.061675939708948135f;
  t53_n60: return -0.04392411187291145f;
  t53_n61: return 0.045113641768693924f;
  t53_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t53_n76;
  t53_n63: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t53_n75;
  t53_n64: if (E.getNumReferences() >= 3318742016 /*6655.5*/) goto t53_n66;
  t53_n65: return 0.07965770363807678f;
  t53_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t53_n72;
  t53_n67: if (E.getNumReferences() >= 3338991488 /*34039.5*/) goto t53_n69;
  t53_n68: return 0.024268943816423416f;
  t53_n69: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t53_n71;
  t53_n70: return 0.14296773076057434f;
  t53_n71: return 0.05245080217719078f;
  t53_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t53_n74;
  t53_n73: return -0.055077798664569855f;
  t53_n74: return 0.036914873868227005f;
  t53_n75: return 0.01791957952082157f;
  t53_n76: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t53_n80;
  t53_n77: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t53_n79;
  t53_n78: return 0.001299406518228352f;
  t53_n79: return 0.1324758380651474f;
  t53_n80: return 0.05716140568256378f;
  t53_n81: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t53_n101;
  t53_n82: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t53_n96;
  t53_n83: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t53_n93;
  t53_n84: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t53_n92;
  t53_n85: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n89;
  t53_n86: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t53_n88;
  t53_n87: return -0.0008928679162636399f;
  t53_n88: return 0.024484071880578995f;
  t53_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t53_n91;
  t53_n90: return -0.05040958151221275f;
  t53_n91: return 0.034062061458826065f;
  t53_n92: return 0.03752703219652176f;
  t53_n93: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n95;
  t53_n94: return -0.01579350046813488f;
  t53_n95: return 0.020739633589982986f;
  t53_n96: if (E.getNumReferences() >= 3267428352 /*96.5*/) goto t53_n100;
  t53_n97: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t53_n99;
  t53_n98: return -0.015854109078645706f;
  t53_n99: return 0.013372012414038181f;
  t53_n100: return 0.037926290184259415f;
  t53_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n103;
  t53_n102: return -0.29537901282310486f;
  t53_n103: return 0.04192867875099182f;
  t53_n104: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t53_n252;
  t53_n105: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t53_n217;
  t53_n106: if (E.getFractionNameInContext() >= 3196560022 /*0.26491230726242065*/) goto t53_n178;
  t53_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t53_n165;
  t53_n108: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t53_n158;
  t53_n109: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t53_n131;
  t53_n110: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t53_n128;
  t53_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t53_n115;
  t53_n112: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t53_n114;
  t53_n113: return 0.0004336383135523647f;
  t53_n114: return -0.03808339312672615f;
  t53_n115: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t53_n127;
  t53_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t53_n118;
  t53_n117: return 0.01661137491464615f;
  t53_n118: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t53_n126;
  t53_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t53_n121;
  t53_n120: return 0.01046082191169262f;
  t53_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t53_n125;
  t53_n122: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t53_n124;
  t53_n123: return -0.042241744697093964f;
  t53_n124: return 0.0529007688164711f;
  t53_n125: return 0.026785816997289658f;
  t53_n126: return 0.038625724613666534f;
  t53_n127: return -0.019645772874355316f;
  t53_n128: if (E.getNumReferences() >= 3358114336 /*172808.5*/) goto t53_n130;
  t53_n129: return 0.029163077473640442f;
  t53_n130: return -0.1726091355085373f;
  t53_n131: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n153;
  t53_n132: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t53_n152;
  t53_n133: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t53_n149;
  t53_n134: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t53_n138;
  t53_n135: if (E.getNumReferences() >= 3334782720 /*25183.5*/) goto t53_n137;
  t53_n136: return 0.0018808842869475484f;
  t53_n137: return 0.035376500338315964f;
  t53_n138: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t53_n140;
  t53_n139: return -0.022494107484817505f;
  t53_n140: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t53_n148;
  t53_n141: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t53_n147;
  t53_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t53_n144;
  t53_n143: return 0.0007395740831270814f;
  t53_n144: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t53_n146;
  t53_n145: return 0.03746248781681061f;
  t53_n146: return -0.015643680468201637f;
  t53_n147: return -0.12734007835388184f;
  t53_n148: return -0.15299668908119202f;
  t53_n149: if (E.getNumReferences() >= 3352808320 /*110487.0*/) goto t53_n151;
  t53_n150: return -0.016615090891718864f;
  t53_n151: return -0.23480072617530823f;
  t53_n152: return -0.06104101240634918f;
  t53_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n155;
  t53_n154: return -0.3736592233181f;
  t53_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t53_n157;
  t53_n156: return 0.005674996413290501f;
  t53_n157: return 0.04393601417541504f;
  t53_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t53_n162;
  t53_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t53_n161;
  t53_n160: return -0.023244263604283333f;
  t53_n161: return 0.05732831358909607f;
  t53_n162: if (E.getNumReferences() >= 3359963264 /*201698.0*/) goto t53_n164;
  t53_n163: return 0.12620994448661804f;
  t53_n164: return 0.07739368081092834f;
  t53_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t53_n177;
  t53_n166: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t53_n172;
  t53_n167: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t53_n169;
  t53_n168: return 0.036299996078014374f;
  t53_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n171;
  t53_n170: return -0.06081931293010712f;
  t53_n171: return 0.007608082611113787f;
  t53_n172: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t53_n176;
  t53_n173: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n175;
  t53_n174: return 0.01831021159887314f;
  t53_n175: return -0.057153597474098206f;
  t53_n176: return 0.04977113753557205f;
  t53_n177: return 0.03869310021400452f;
  t53_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t53_n184;
  t53_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n181;
  t53_n180: return -0.025312472134828568f;
  t53_n181: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t53_n183;
  t53_n182: return 0.013577399775385857f;
  t53_n183: return -0.06971538066864014f;
  t53_n184: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t53_n208;
  t53_n185: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t53_n207;
  t53_n186: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t53_n188;
  t53_n187: return 0.020734965801239014f;
  t53_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t53_n204;
  t53_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t53_n195;
  t53_n190: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t53_n192;
  t53_n191: return 0.030979221686720848f;
  t53_n192: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t53_n194;
  t53_n193: return 0.10678412020206451f;
  t53_n194: return 0.1603301614522934f;
  t53_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n197;
  t53_n196: return -0.07104276865720749f;
  t53_n197: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t53_n203;
  t53_n198: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t53_n200;
  t53_n199: return 0.04211189225316048f;
  t53_n200: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t53_n202;
  t53_n201: return 0.12603415548801422f;
  t53_n202: return -0.003980227280408144f;
  t53_n203: return 0.13704590499401093f;
  t53_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n206;
  t53_n205: return -0.13809114694595337f;
  t53_n206: return 0.029460569843649864f;
  t53_n207: return 0.052662625908851624f;
  t53_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t53_n214;
  t53_n209: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t53_n213;
  t53_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t53_n212;
  t53_n211: return 0.0049098036251962185f;
  t53_n212: return 0.020303405821323395f;
  t53_n213: return 0.024631574749946594f;
  t53_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n216;
  t53_n215: return -0.09223974496126175f;
  t53_n216: return 0.028368866071105003f;
  t53_n217: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t53_n249;
  t53_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t53_n246;
  t53_n219: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t53_n227;
  t53_n220: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t53_n226;
  t53_n221: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t53_n225;
  t53_n222: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t53_n224;
  t53_n223: return 0.029350707307457924f;
  t53_n224: return 0.13100633025169373f;
  t53_n225: return -0.06440527737140656f;
  t53_n226: return 0.051510151475667953f;
  t53_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t53_n245;
  t53_n228: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t53_n244;
  t53_n229: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t53_n239;
  t53_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t53_n234;
  t53_n231: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t53_n233;
  t53_n232: return -0.03546256944537163f;
  t53_n233: return 0.07334059476852417f;
  t53_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t53_n236;
  t53_n235: return -0.20090675354003906f;
  t53_n236: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t53_n238;
  t53_n237: return 0.05447905510663986f;
  t53_n238: return 0.01790224015712738f;
  t53_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t53_n243;
  t53_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t53_n242;
  t53_n241: return 0.00019737420370802283f;
  t53_n242: return 0.03540948033332825f;
  t53_n243: return 0.0585143156349659f;
  t53_n244: return 0.0672619417309761f;
  t53_n245: return 0.03726689890027046f;
  t53_n246: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t53_n248;
  t53_n247: return 0.019270852208137512f;
  t53_n248: return 0.05050789564847946f;
  t53_n249: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t53_n251;
  t53_n250: return 0.11995746940374374f;
  t53_n251: return 0.07278937846422195f;
  t53_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t53_n254;
  t53_n253: return 0.03201992064714432f;
  t53_n254: return 0.07154560089111328f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree54(const Example& E) {
  t54_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t54_n88;
  t54_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t54_n75;
  t54_n2: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t54_n66;
  t54_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t54_n49;
  t54_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t54_n36;
  t54_n5: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t54_n33;
  t54_n6: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t54_n32;
  t54_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t54_n23;
  t54_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t54_n10;
  t54_n9: return -0.05080345645546913f;
  t54_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t54_n12;
  t54_n11: return -0.11599605530500412f;
  t54_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t54_n20;
  t54_n13: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t54_n17;
  t54_n14: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t54_n16;
  t54_n15: return -0.005968765821307898f;
  t54_n16: return -0.10995398461818695f;
  t54_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t54_n19;
  t54_n18: return -0.003622206626459956f;
  t54_n19: return 0.03840448707342148f;
  t54_n20: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t54_n22;
  t54_n21: return 0.06724175810813904f;
  t54_n22: return 0.017320074141025543f;
  t54_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t54_n27;
  t54_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n26;
  t54_n25: return -0.12289755791425705f;
  t54_n26: return 0.001024981727823615f;
  t54_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t54_n31;
  t54_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t54_n30;
  t54_n29: return 0.00580636179074645f;
  t54_n30: return 0.05437557399272919f;
  t54_n31: return 0.11355340480804443f;
  t54_n32: return -0.06296782195568085f;
  t54_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t54_n35;
  t54_n34: return 0.0075763980858027935f;
  t54_n35: return 0.05118473991751671f;
  t54_n36: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t54_n42;
  t54_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n41;
  t54_n38: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t54_n40;
  t54_n39: return -0.12347486615180969f;
  t54_n40: return 0.08930535614490509f;
  t54_n41: return 0.041300129145383835f;
  t54_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n46;
  t54_n43: if (E.getNumReferences() >= 3273818112 /*162.5*/) goto t54_n45;
  t54_n44: return -0.030063508078455925f;
  t54_n45: return 0.030769163742661476f;
  t54_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t54_n48;
  t54_n47: return 0.011324777267873287f;
  t54_n48: return -0.05801750719547272f;
  t54_n49: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t54_n65;
  t54_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t54_n62;
  t54_n51: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t54_n55;
  t54_n52: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t54_n54;
  t54_n53: return -0.026530183851718903f;
  t54_n54: return 0.025037188082933426f;
  t54_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n57;
  t54_n56: return -0.06551887094974518f;
  t54_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t54_n59;
  t54_n58: return 0.11156368255615234f;
  t54_n59: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t54_n61;
  t54_n60: return 0.052457209676504135f;
  t54_n61: return -0.0013063555816188455f;
  t54_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n64;
  t54_n63: return 0.01975894346833229f;
  t54_n64: return 0.056143369525671005f;
  t54_n65: return -0.03216199949383736f;
  t54_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t54_n74;
  t54_n67: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t54_n69;
  t54_n68: return 0.023585204035043716f;
  t54_n69: if (E.getNumReferences() >= 3309029376 /*3004.5*/) goto t54_n73;
  t54_n70: if (E.getNumReferences() >= 3308949504 /*2985.0*/) goto t54_n72;
  t54_n71: return 0.08421223610639572f;
  t54_n72: return 0.1311950832605362f;
  t54_n73: return 0.048770464956760406f;
  t54_n74: return 0.056441884487867355f;
  t54_n75: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t54_n81;
  t54_n76: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t54_n80;
  t54_n77: if (E.getNumReferences() >= 3266904064 /*92.5*/) goto t54_n79;
  t54_n78: return 0.001689922297373414f;
  t54_n79: return 0.04728587716817856f;
  t54_n80: return -0.014859492890536785f;
  t54_n81: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t54_n87;
  t54_n82: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t54_n86;
  t54_n83: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t54_n85;
  t54_n84: return 0.009427635930478573f;
  t54_n85: return -0.053225159645080566f;
  t54_n86: return -0.08748576790094376f;
  t54_n87: return -0.05545097216963768f;
  t54_n88: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t54_n254;
  t54_n89: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t54_n221;
  t54_n90: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t54_n176;
  t54_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t54_n173;
  t54_n92: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t54_n172;
  t54_n93: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t54_n163;
  t54_n94: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t54_n114;
  t54_n95: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t54_n111;
  t54_n96: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t54_n102;
  t54_n97: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t54_n101;
  t54_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n100;
  t54_n99: return -0.022313570603728294f;
  t54_n100: return 0.013957766816020012f;
  t54_n101: return -0.018024884164333344f;
  t54_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t54_n104;
  t54_n103: return 0.020706692710518837f;
  t54_n104: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t54_n106;
  t54_n105: return 0.0321887731552124f;
  t54_n106: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t54_n108;
  t54_n107: return 0.0768178328871727f;
  t54_n108: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t54_n110;
  t54_n109: return 0.22357122600078583f;
  t54_n110: return 0.11552947759628296f;
  t54_n111: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t54_n113;
  t54_n112: return 0.030295582488179207f;
  t54_n113: return -0.001021072268486023f;
  t54_n114: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t54_n162;
  t54_n115: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t54_n161;
  t54_n116: if (E.getNumReferences() >= 3320709120 /*7616.0*/) goto t54_n158;
  t54_n117: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t54_n119;
  t54_n118: return -0.00467847753316164f;
  t54_n119: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t54_n133;
  t54_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n122;
  t54_n121: return -0.052310481667518616f;
  t54_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t54_n128;
  t54_n123: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t54_n125;
  t54_n124: return 0.0012950343079864979f;
  t54_n125: if (E.getNumReferences() >= 3319284736 /*6920.5*/) goto t54_n127;
  t54_n126: return -0.07838287204504013f;
  t54_n127: return -0.5154335498809814f;
  t54_n128: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t54_n130;
  t54_n129: return -0.0025176811031997204f;
  t54_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t54_n132;
  t54_n131: return 0.00824067648500204f;
  t54_n132: return 0.02215557172894478f;
  t54_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n135;
  t54_n134: return -0.00414573447778821f;
  t54_n135: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t54_n153;
  t54_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t54_n150;
  t54_n137: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t54_n139;
  t54_n138: return -0.0019526747055351734f;
  t54_n139: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t54_n149;
  t54_n140: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t54_n148;
  t54_n141: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t54_n147;
  t54_n142: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t54_n146;
  t54_n143: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t54_n145;
  t54_n144: return 0.05067116767168045f;
  t54_n145: return -0.008058750070631504f;
  t54_n146: return 0.04781473055481911f;
  t54_n147: return -0.02813979610800743f;
  t54_n148: return 0.05696980282664299f;
  t54_n149: return -0.0666801780462265f;
  t54_n150: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t54_n152;
  t54_n151: return 0.0814903974533081f;
  t54_n152: return 0.0020463308319449425f;
  t54_n153: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t54_n155;
  t54_n154: return 0.07852856069803238f;
  t54_n155: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t54_n157;
  t54_n156: return -0.07820173352956772f;
  t54_n157: return 0.07336243242025375f;
  t54_n158: if (E.getNumReferences() >= 3320842240 /*7681.0*/) goto t54_n160;
  t54_n159: return 0.12262655794620514f;
  t54_n160: return 0.023364940658211708f;
  t54_n161: return -0.009994427673518658f;
  t54_n162: return -0.1156669482588768f;
  t54_n163: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t54_n171;
  t54_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t54_n166;
  t54_n165: return 0.018002042546868324f;
  t54_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n168;
  t54_n167: return -0.1682370901107788f;
  t54_n168: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t54_n170;
  t54_n169: return 0.001974282553419471f;
  t54_n170: return 0.04271538183093071f;
  t54_n171: return -0.016634756699204445f;
  t54_n172: return 0.08452052623033524f;
  t54_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t54_n175;
  t54_n174: return 0.016228923574090004f;
  t54_n175: return 0.03620756417512894f;
  t54_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t54_n184;
  t54_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n179;
  t54_n178: return -0.022066760808229446f;
  t54_n179: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t54_n183;
  t54_n180: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t54_n182;
  t54_n181: return 0.01158017199486494f;
  t54_n182: return -0.07124026864767075f;
  t54_n183: return -0.06953221559524536f;
  t54_n184: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t54_n218;
  t54_n185: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t54_n193;
  t54_n186: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t54_n192;
  t54_n187: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t54_n191;
  t54_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t54_n190;
  t54_n189: return 0.015724370256066322f;
  t54_n190: return 0.04051338881254196f;
  t54_n191: return 0.032030798494815826f;
  t54_n192: return 0.04295707121491432f;
  t54_n193: if (E.getNumReferences() >= 3298648064 /*1259.0*/) goto t54_n217;
  t54_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t54_n208;
  t54_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t54_n197;
  t54_n196: return 0.0010924249654635787f;
  t54_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t54_n203;
  t54_n198: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t54_n200;
  t54_n199: return -0.0007119461661204696f;
  t54_n200: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t54_n202;
  t54_n201: return 0.025021839886903763f;
  t54_n202: return 0.05094030499458313f;
  t54_n203: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t54_n207;
  t54_n204: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t54_n206;
  t54_n205: return 0.03690503537654877f;
  t54_n206: return 0.06329875439405441f;
  t54_n207: return 0.01634577475488186f;
  t54_n208: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t54_n210;
  t54_n209: return 0.0268325824290514f;
  t54_n210: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t54_n214;
  t54_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t54_n213;
  t54_n212: return -0.02656221203505993f;
  t54_n213: return 0.014652823098003864f;
  t54_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t54_n216;
  t54_n215: return 0.01959332451224327f;
  t54_n216: return 0.08418316394090652f;
  t54_n217: return 0.04159155115485191f;
  t54_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t54_n220;
  t54_n219: return 0.018482523038983345f;
  t54_n220: return 0.051940616220235825f;
  t54_n221: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t54_n251;
  t54_n222: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t54_n250;
  t54_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t54_n247;
  t54_n224: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t54_n226;
  t54_n225: return 0.02895142324268818f;
  t54_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t54_n246;
  t54_n227: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t54_n245;
  t54_n228: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t54_n242;
  t54_n229: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t54_n241;
  t54_n230: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t54_n238;
  t54_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t54_n237;
  t54_n232: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t54_n234;
  t54_n233: return -0.044673770666122437f;
  t54_n234: if (E.getNumReferences() >= 3259236352 /*49.0*/) goto t54_n236;
  t54_n235: return 0.013259077444672585f;
  t54_n236: return 0.09562364965677261f;
  t54_n237: return 0.022368617355823517f;
  t54_n238: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t54_n240;
  t54_n239: return 0.0950370505452156f;
  t54_n240: return 0.0015500321751460433f;
  t54_n241: return -0.04130653664469719f;
  t54_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t54_n244;
  t54_n243: return 0.0053681558929383755f;
  t54_n244: return 0.05691034719347954f;
  t54_n245: return 0.06664188951253891f;
  t54_n246: return 0.03604026138782501f;
  t54_n247: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t54_n249;
  t54_n248: return 0.020442664623260498f;
  t54_n249: return 0.049471043050289154f;
  t54_n250: return -0.06110817566514015f;
  t54_n251: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t54_n253;
  t54_n252: return 0.059410661458969116f;
  t54_n253: return 0.11777876317501068f;
  t54_n254: return 0.03687473386526108f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree55(const Example& E) {
  t55_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t55_n144;
  t55_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t55_n31;
  t55_n2: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t55_n24;
  t55_n3: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t55_n19;
  t55_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n16;
  t55_n5: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t55_n13;
  t55_n6: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t55_n12;
  t55_n7: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t55_n11;
  t55_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t55_n10;
  t55_n9: return 0.01758917048573494f;
  t55_n10: return 0.07820496708154678f;
  t55_n11: return -0.004185371100902557f;
  t55_n12: return -0.028883157297968864f;
  t55_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t55_n15;
  t55_n14: return 0.012617564760148525f;
  t55_n15: return -0.28382036089897156f;
  t55_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t55_n18;
  t55_n17: return 3.2727351936046034e-05f;
  t55_n18: return 0.008634945377707481f;
  t55_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n23;
  t55_n20: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t55_n22;
  t55_n21: return 0.003999761305749416f;
  t55_n22: return -0.03943977504968643f;
  t55_n23: return 0.02195758745074272f;
  t55_n24: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t55_n26;
  t55_n25: return 0.027933040633797646f;
  t55_n26: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t55_n30;
  t55_n27: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t55_n29;
  t55_n28: return 0.001223216182552278f;
  t55_n29: return 0.12748722732067108f;
  t55_n30: return 0.05376705154776573f;
  t55_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t55_n141;
  t55_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t55_n104;
  t55_n33: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t55_n53;
  t55_n34: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t55_n52;
  t55_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n37;
  t55_n36: return -0.02974041737616062f;
  t55_n37: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t55_n45;
  t55_n38: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t55_n44;
  t55_n39: if (E.getNumReferences() >= 3268345856 /*103.5*/) goto t55_n41;
  t55_n40: return 0.0013270790223032236f;
  t55_n41: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t55_n43;
  t55_n42: return 0.14086639881134033f;
  t55_n43: return 0.043295081704854965f;
  t55_n44: return 0.03116259165108204f;
  t55_n45: if (E.getNumReferences() >= 3355438208 /*131033.0*/) goto t55_n51;
  t55_n46: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t55_n50;
  t55_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n49;
  t55_n48: return -0.1292627602815628f;
  t55_n49: return 0.13527557253837585f;
  t55_n50: return -0.09709123522043228f;
  t55_n51: return 0.02042931690812111f;
  t55_n52: return -0.06102798879146576f;
  t55_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t55_n101;
  t55_n54: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t55_n74;
  t55_n55: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t55_n57;
  t55_n56: return 0.040843475610017776f;
  t55_n57: if (E.getNumReferences() >= 3286286336 /*449.5*/) goto t55_n59;
  t55_n58: return -0.04045327380299568f;
  t55_n59: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t55_n73;
  t55_n60: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t55_n68;
  t55_n61: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t55_n67;
  t55_n62: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t55_n64;
  t55_n63: return 0.008403521962463856f;
  t55_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n66;
  t55_n65: return -0.20015175640583038f;
  t55_n66: return 0.12336838245391846f;
  t55_n67: return -0.03247803449630737f;
  t55_n68: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t55_n72;
  t55_n69: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t55_n71;
  t55_n70: return 0.010247703641653061f;
  t55_n71: return 0.10693728923797607f;
  t55_n72: return 0.0323248952627182f;
  t55_n73: return -0.06017237529158592f;
  t55_n74: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t55_n96;
  t55_n75: if (E.getNumReferences() >= 3363138304 /*251308.0*/) goto t55_n95;
  t55_n76: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t55_n90;
  t55_n77: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t55_n87;
  t55_n78: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t55_n86;
  t55_n79: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t55_n81;
  t55_n80: return -0.00872083380818367f;
  t55_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t55_n83;
  t55_n82: return -0.1114482507109642f;
  t55_n83: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t55_n85;
  t55_n84: return 0.014440487138926983f;
  t55_n85: return -0.11363319307565689f;
  t55_n86: return -0.1014399304986f;
  t55_n87: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t55_n89;
  t55_n88: return -0.03389706090092659f;
  t55_n89: return 0.1632252037525177f;
  t55_n90: if (E.getNumReferences() >= 3361098688 /*219439.0*/) goto t55_n92;
  t55_n91: return 0.10730250924825668f;
  t55_n92: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t55_n94;
  t55_n93: return -0.25227174162864685f;
  t55_n94: return 0.08884807676076889f;
  t55_n95: return -0.20719008147716522f;
  t55_n96: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t55_n98;
  t55_n97: return 0.09240256994962692f;
  t55_n98: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t55_n100;
  t55_n99: return -0.09691861271858215f;
  t55_n100: return 0.05213538184762001f;
  t55_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t55_n103;
  t55_n102: return -0.07113787531852722f;
  t55_n103: return 0.013864162378013134f;
  t55_n104: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t55_n140;
  t55_n105: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t55_n135;
  t55_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t55_n108;
  t55_n107: return -0.05338485166430473f;
  t55_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t55_n120;
  t55_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t55_n111;
  t55_n110: return -0.022135311737656593f;
  t55_n111: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t55_n113;
  t55_n112: return 0.005627820733934641f;
  t55_n113: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t55_n115;
  t55_n114: return 0.08026840537786484f;
  t55_n115: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t55_n117;
  t55_n116: return -0.04051728919148445f;
  t55_n117: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t55_n119;
  t55_n118: return 0.0789041742682457f;
  t55_n119: return 0.01563647948205471f;
  t55_n120: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t55_n130;
  t55_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t55_n123;
  t55_n122: return -0.01145399734377861f;
  t55_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n127;
  t55_n124: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t55_n126;
  t55_n125: return 0.008077234029769897f;
  t55_n126: return -0.08627963811159134f;
  t55_n127: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t55_n129;
  t55_n128: return 0.06923598796129227f;
  t55_n129: return 0.047625742852687836f;
  t55_n130: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t55_n134;
  t55_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t55_n133;
  t55_n132: return -0.03845735266804695f;
  t55_n133: return 0.05273844674229622f;
  t55_n134: return 0.014411182142794132f;
  t55_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n139;
  t55_n136: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t55_n138;
  t55_n137: return 0.04168299585580826f;
  t55_n138: return -0.02853589691221714f;
  t55_n139: return 0.02367737889289856f;
  t55_n140: return -0.021845536306500435f;
  t55_n141: if (E.getNumReferences() >= 3267428352 /*96.5*/) goto t55_n143;
  t55_n142: return 0.005207906011492014f;
  t55_n143: return 0.035928044468164444f;
  t55_n144: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t55_n254;
  t55_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t55_n201;
  t55_n146: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t55_n198;
  t55_n147: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t55_n171;
  t55_n148: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t55_n170;
  t55_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t55_n161;
  t55_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t55_n158;
  t55_n151: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t55_n157;
  t55_n152: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t55_n156;
  t55_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n155;
  t55_n154: return -0.06057935580611229f;
  t55_n155: return -0.005514855030924082f;
  t55_n156: return 0.0026295206043869257f;
  t55_n157: return -0.04097340628504753f;
  t55_n158: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t55_n160;
  t55_n159: return 0.0015336191281676292f;
  t55_n160: return 0.027299052104353905f;
  t55_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t55_n163;
  t55_n162: return 0.006574918981641531f;
  t55_n163: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t55_n169;
  t55_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t55_n168;
  t55_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n167;
  t55_n166: return -0.044010844081640244f;
  t55_n167: return 0.009808981791138649f;
  t55_n168: return 0.02131134644150734f;
  t55_n169: return 0.03146350756287575f;
  t55_n170: return 0.031050406396389008f;
  t55_n171: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t55_n173;
  t55_n172: return 0.03365100920200348f;
  t55_n173: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t55_n197;
  t55_n174: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t55_n196;
  t55_n175: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t55_n195;
  t55_n176: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t55_n178;
  t55_n177: return -0.005872584413737059f;
  t55_n178: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t55_n190;
  t55_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t55_n181;
  t55_n180: return -0.05645410344004631f;
  t55_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t55_n183;
  t55_n182: return -0.053770169615745544f;
  t55_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t55_n185;
  t55_n184: return 0.00971019547432661f;
  t55_n185: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t55_n187;
  t55_n186: return 0.03884709253907204f;
  t55_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t55_n189;
  t55_n188: return -0.3452548086643219f;
  t55_n189: return 0.15168949961662292f;
  t55_n190: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t55_n194;
  t55_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t55_n193;
  t55_n192: return -0.003086724551394582f;
  t55_n193: return 0.025148676708340645f;
  t55_n194: return -0.128494530916214f;
  t55_n195: return -0.061835940927267075f;
  t55_n196: return -0.04372608661651611f;
  t55_n197: return 0.03089383989572525f;
  t55_n198: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t55_n200;
  t55_n199: return 0.040843281894922256f;
  t55_n200: return 0.013554221950471401f;
  t55_n201: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t55_n249;
  t55_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t55_n240;
  t55_n203: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t55_n239;
  t55_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t55_n224;
  t55_n205: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t55_n223;
  t55_n206: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t55_n208;
  t55_n207: return -0.003407885553315282f;
  t55_n208: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t55_n216;
  t55_n209: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t55_n211;
  t55_n210: return -0.01800018548965454f;
  t55_n211: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t55_n215;
  t55_n212: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t55_n214;
  t55_n213: return 0.05864959955215454f;
  t55_n214: return 0.1081058531999588f;
  t55_n215: return -0.055420707911252975f;
  t55_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t55_n218;
  t55_n217: return 0.018681610003113747f;
  t55_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t55_n220;
  t55_n219: return 0.017478160560131073f;
  t55_n220: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t55_n222;
  t55_n221: return 0.05926065146923065f;
  t55_n222: return 0.20910464227199554f;
  t55_n223: return -0.06700340658426285f;
  t55_n224: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t55_n232;
  t55_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t55_n231;
  t55_n226: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t55_n230;
  t55_n227: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t55_n229;
  t55_n228: return 0.019563624635338783f;
  t55_n229: return 0.041803471744060516f;
  t55_n230: return 0.031007084995508194f;
  t55_n231: return 0.02250717766582966f;
  t55_n232: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t55_n234;
  t55_n233: return -0.06273901462554932f;
  t55_n234: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t55_n238;
  t55_n235: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t55_n237;
  t55_n236: return -0.011134245432913303f;
  t55_n237: return 0.04552208632230759f;
  t55_n238: return 0.03372155874967575f;
  t55_n239: return -0.007461160887032747f;
  t55_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t55_n242;
  t55_n241: return 0.018340976908802986f;
  t55_n242: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t55_n244;
  t55_n243: return 0.002278057159855962f;
  t55_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t55_n246;
  t55_n245: return 0.028138292953372f;
  t55_n246: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t55_n248;
  t55_n247: return 0.04110164940357208f;
  t55_n248: return 0.09853839129209518f;
  t55_n249: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t55_n253;
  t55_n250: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t55_n252;
  t55_n251: return 0.004941465798765421f;
  t55_n252: return 0.04028542712330818f;
  t55_n253: return 0.03326434642076492f;
  t55_n254: return 0.035685401409864426f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree56(const Example& E) {
  t56_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t56_n100;
  t56_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t56_n75;
  t56_n2: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t56_n66;
  t56_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t56_n49;
  t56_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t56_n42;
  t56_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t56_n41;
  t56_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t56_n34;
  t56_n7: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t56_n33;
  t56_n8: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t56_n32;
  t56_n9: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t56_n25;
  t56_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t56_n12;
  t56_n11: return -0.04406058415770531f;
  t56_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n24;
  t56_n13: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t56_n17;
  t56_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n16;
  t56_n15: return -0.020336687564849854f;
  t56_n16: return 0.016627321019768715f;
  t56_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t56_n19;
  t56_n18: return -0.009646838530898094f;
  t56_n19: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t56_n21;
  t56_n20: return -0.11145104467868805f;
  t56_n21: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t56_n23;
  t56_n22: return 0.05866702273488045f;
  t56_n23: return 0.1629796177148819f;
  t56_n24: return 0.020264185965061188f;
  t56_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t56_n31;
  t56_n26: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t56_n30;
  t56_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t56_n29;
  t56_n28: return -0.11366816610097885f;
  t56_n29: return 0.009720046073198318f;
  t56_n30: return 0.004493555054068565f;
  t56_n31: return 0.03566965460777283f;
  t56_n32: return -0.018749859184026718f;
  t56_n33: return -0.11770032346248627f;
  t56_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t56_n36;
  t56_n35: return -0.07708680629730225f;
  t56_n36: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t56_n40;
  t56_n37: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t56_n39;
  t56_n38: return 0.007444296032190323f;
  t56_n39: return 0.04987248778343201f;
  t56_n40: return -0.03944094851613045f;
  t56_n41: return 0.008793504908680916f;
  t56_n42: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t56_n44;
  t56_n43: return 0.03316080570220947f;
  t56_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n46;
  t56_n45: return -0.017559457570314407f;
  t56_n46: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t56_n48;
  t56_n47: return 0.011362566612660885f;
  t56_n48: return -0.05429839715361595f;
  t56_n49: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n65;
  t56_n50: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t56_n64;
  t56_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n63;
  t56_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t56_n62;
  t56_n53: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t56_n55;
  t56_n54: return -0.018488705158233643f;
  t56_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t56_n57;
  t56_n56: return -0.05401912331581116f;
  t56_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t56_n59;
  t56_n58: return 0.10162442922592163f;
  t56_n59: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t56_n61;
  t56_n60: return 0.05098012834787369f;
  t56_n61: return 0.00022807021741755307f;
  t56_n62: return 0.014583921059966087f;
  t56_n63: return 0.05742741376161575f;
  t56_n64: return -0.039623841643333435f;
  t56_n65: return 0.04235963150858879f;
  t56_n66: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t56_n70;
  t56_n67: if (E.getNumReferences() >= 3358114336 /*172808.5*/) goto t56_n69;
  t56_n68: return 0.024985317140817642f;
  t56_n69: return -0.15091651678085327f;
  t56_n70: if (E.getNumReferences() >= 3360228064 /*205835.5*/) goto t56_n74;
  t56_n71: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n73;
  t56_n72: return 0.05333747714757919f;
  t56_n73: return 0.109255850315094f;
  t56_n74: return 0.04982677102088928f;
  t56_n75: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t56_n95;
  t56_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t56_n78;
  t56_n77: return -0.05975513905286789f;
  t56_n78: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n90;
  t56_n79: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t56_n89;
  t56_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n82;
  t56_n81: return -0.00538007915019989f;
  t56_n82: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t56_n84;
  t56_n83: return 0.014085836708545685f;
  t56_n84: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t56_n88;
  t56_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t56_n87;
  t56_n86: return 0.03496629744768143f;
  t56_n87: return 0.09665914624929428f;
  t56_n88: return 0.03297645226120949f;
  t56_n89: return -0.012068454176187515f;
  t56_n90: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t56_n92;
  t56_n91: return -0.01967122219502926f;
  t56_n92: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t56_n94;
  t56_n93: return 0.028451869264245033f;
  t56_n94: return -0.040169116109609604f;
  t56_n95: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t56_n99;
  t56_n96: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t56_n98;
  t56_n97: return 0.008634760044515133f;
  t56_n98: return -0.0793365091085434f;
  t56_n99: return -0.053845301270484924f;
  t56_n100: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t56_n252;
  t56_n101: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n235;
  t56_n102: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t56_n186;
  t56_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t56_n161;
  t56_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n106;
  t56_n105: return -0.11955055594444275f;
  t56_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t56_n108;
  t56_n107: return -0.1160259023308754f;
  t56_n108: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t56_n160;
  t56_n109: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t56_n123;
  t56_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t56_n114;
  t56_n111: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t56_n113;
  t56_n112: return 0.0007111341692507267f;
  t56_n113: return -0.03249980881810188f;
  t56_n114: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t56_n116;
  t56_n115: return 0.016282638534903526f;
  t56_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Variable))) goto t56_n122;
  t56_n117: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t56_n119;
  t56_n118: return -0.07086755335330963f;
  t56_n119: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t56_n121;
  t56_n120: return -0.017819514498114586f;
  t56_n121: return 0.030990544706583023f;
  t56_n122: return 0.03343024104833603f;
  t56_n123: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t56_n151;
  t56_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t56_n126;
  t56_n125: return -0.004271252546459436f;
  t56_n126: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t56_n136;
  t56_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t56_n135;
  t56_n128: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t56_n130;
  t56_n129: return 0.022368090227246284f;
  t56_n130: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t56_n134;
  t56_n131: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t56_n133;
  t56_n132: return 0.07975340634584427f;
  t56_n133: return 0.11658254265785217f;
  t56_n134: return 0.05079374834895134f;
  t56_n135: return 0.14148801565170288f;
  t56_n136: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t56_n138;
  t56_n137: return 0.07462615519762039f;
  t56_n138: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t56_n150;
  t56_n139: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t56_n143;
  t56_n140: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t56_n142;
  t56_n141: return 0.13528235256671906f;
  t56_n142: return 0.020996907725930214f;
  t56_n143: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t56_n145;
  t56_n144: return 0.12850533425807953f;
  t56_n145: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t56_n149;
  t56_n146: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t56_n148;
  t56_n147: return 0.13901621103286743f;
  t56_n148: return 0.0829499363899231f;
  t56_n149: return 0.10939390957355499f;
  t56_n150: return 0.05604904890060425f;
  t56_n151: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t56_n153;
  t56_n152: return 0.03046819567680359f;
  t56_n153: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t56_n159;
  t56_n154: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t56_n158;
  t56_n155: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t56_n157;
  t56_n156: return -0.016574569046497345f;
  t56_n157: return -0.11398434638977051f;
  t56_n158: return 0.11240945756435394f;
  t56_n159: return -0.07498485594987869f;
  t56_n160: return -0.0508747436106205f;
  t56_n161: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t56_n177;
  t56_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t56_n164;
  t56_n163: return 0.003176626982167363f;
  t56_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t56_n176;
  t56_n165: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t56_n171;
  t56_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n170;
  t56_n167: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t56_n169;
  t56_n168: return -0.013898156583309174f;
  t56_n169: return -0.1962067037820816f;
  t56_n170: return 0.02359485626220703f;
  t56_n171: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t56_n173;
  t56_n172: return 0.00967680849134922f;
  t56_n173: if (E.getNumReferences() >= 3349612032 /*85516.0*/) goto t56_n175;
  t56_n174: return 0.09589584916830063f;
  t56_n175: return 0.03408270329236984f;
  t56_n176: return -0.09340158104896545f;
  t56_n177: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t56_n179;
  t56_n178: return 0.025863733142614365f;
  t56_n179: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t56_n185;
  t56_n180: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t56_n184;
  t56_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n183;
  t56_n182: return -0.002300758846104145f;
  t56_n183: return 0.0189548097550869f;
  t56_n184: return -0.1811893880367279f;
  t56_n185: return 0.03885679319500923f;
  t56_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t56_n188;
  t56_n187: return 0.011932929046452045f;
  t56_n188: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t56_n234;
  t56_n189: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t56_n227;
  t56_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t56_n198;
  t56_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t56_n193;
  t56_n192: return -0.02819809876382351f;
  t56_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t56_n197;
  t56_n194: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t56_n196;
  t56_n195: return -0.0003722311230376363f;
  t56_n196: return 0.0794118270277977f;
  t56_n197: return 0.0775042399764061f;
  t56_n198: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t56_n204;
  t56_n199: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t56_n201;
  t56_n200: return 0.026545090600848198f;
  t56_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t56_n203;
  t56_n202: return 0.00267965835519135f;
  t56_n203: return 0.07987607270479202f;
  t56_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t56_n208;
  t56_n205: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t56_n207;
  t56_n206: return -0.00781386997550726f;
  t56_n207: return 0.028506113216280937f;
  t56_n208: if (E.getFractionNameInContext() >= 3193337030 /*0.20942983031272888*/) goto t56_n210;
  t56_n209: return -0.06900983303785324f;
  t56_n210: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t56_n218;
  t56_n211: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t56_n213;
  t56_n212: return 0.00045403034891933203f;
  t56_n213: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t56_n215;
  t56_n214: return 0.05007411539554596f;
  t56_n215: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t56_n217;
  t56_n216: return 0.059536710381507874f;
  t56_n217: return 0.1322205364704132f;
  t56_n218: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t56_n222;
  t56_n219: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t56_n221;
  t56_n220: return -0.06860565394163132f;
  t56_n221: return 0.02862616255879402f;
  t56_n222: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t56_n226;
  t56_n223: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t56_n225;
  t56_n224: return -0.04413975402712822f;
  t56_n225: return 0.0423538014292717f;
  t56_n226: return 0.0722176805138588f;
  t56_n227: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t56_n233;
  t56_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t56_n230;
  t56_n229: return 0.0063331774435937405f;
  t56_n230: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t56_n232;
  t56_n231: return 0.018597029149532318f;
  t56_n232: return 0.03895554319024086f;
  t56_n233: return -0.16493552923202515f;
  t56_n234: return 0.03268662095069885f;
  t56_n235: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t56_n249;
  t56_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t56_n244;
  t56_n237: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t56_n239;
  t56_n238: return 0.027096141129732132f;
  t56_n239: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t56_n243;
  t56_n240: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t56_n242;
  t56_n241: return 0.004265814553946257f;
  t56_n242: return 0.0622943751513958f;
  t56_n243: return 0.03422358259558678f;
  t56_n244: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t56_n246;
  t56_n245: return 0.01651277393102646f;
  t56_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t56_n248;
  t56_n247: return 0.026603875681757927f;
  t56_n248: return 0.049814727157354355f;
  t56_n249: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t56_n251;
  t56_n250: return 0.12320278584957123f;
  t56_n251: return 0.06838694959878922f;
  t56_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t56_n254;
  t56_n253: return 0.02801465429365635f;
  t56_n254: return 0.06813579797744751f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree57(const Example& E) {
  t57_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t57_n102;
  t57_n1: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t57_n61;
  t57_n2: if (E.getNumReferences() >= 3271229440 /*125.5*/) goto t57_n30;
  t57_n3: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t57_n27;
  t57_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n16;
  t57_n5: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t57_n11;
  t57_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n8;
  t57_n7: return -0.048902593553066254f;
  t57_n8: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t57_n10;
  t57_n9: return -0.009535564109683037f;
  t57_n10: return -0.0991370752453804f;
  t57_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t57_n15;
  t57_n12: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t57_n14;
  t57_n13: return 0.04281124100089073f;
  t57_n14: return 0.001884403987787664f;
  t57_n15: return -0.024900181218981743f;
  t57_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t57_n26;
  t57_n17: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t57_n19;
  t57_n18: return -0.004781418479979038f;
  t57_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t57_n23;
  t57_n20: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t57_n22;
  t57_n21: return -0.04670057073235512f;
  t57_n22: return 0.02107248641550541f;
  t57_n23: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t57_n25;
  t57_n24: return 0.10573001205921173f;
  t57_n25: return 0.036969587206840515f;
  t57_n26: return 0.008465414866805077f;
  t57_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n29;
  t57_n28: return -0.005364524200558662f;
  t57_n29: return 0.02648601494729519f;
  t57_n30: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t57_n58;
  t57_n31: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t57_n57;
  t57_n32: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t57_n52;
  t57_n33: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t57_n51;
  t57_n34: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t57_n42;
  t57_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t57_n41;
  t57_n36: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n40;
  t57_n37: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t57_n39;
  t57_n38: return 0.029147429391741753f;
  t57_n39: return 0.019173603504896164f;
  t57_n40: return -0.004746914841234684f;
  t57_n41: return 0.04196176677942276f;
  t57_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n46;
  t57_n43: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t57_n45;
  t57_n44: return 0.05445060506463051f;
  t57_n45: return 0.1237448900938034f;
  t57_n46: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t57_n50;
  t57_n47: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t57_n49;
  t57_n48: return 0.11868966370820999f;
  t57_n49: return 0.24084053933620453f;
  t57_n50: return 0.08025704324245453f;
  t57_n51: return -0.08690131455659866f;
  t57_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n56;
  t57_n53: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t57_n55;
  t57_n54: return 0.11003678292036057f;
  t57_n55: return 0.003800041740760207f;
  t57_n56: return -0.2537919580936432f;
  t57_n57: return -0.06511369347572327f;
  t57_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t57_n60;
  t57_n59: return 0.036362174898386f;
  t57_n60: return 0.08775845170021057f;
  t57_n61: if (E.getNumReferences() >= 3334963968 /*25537.5*/) goto t57_n91;
  t57_n62: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t57_n70;
  t57_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t57_n67;
  t57_n64: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t57_n66;
  t57_n65: return -0.0036556199193000793f;
  t57_n66: return 0.02728436514735222f;
  t57_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t57_n69;
  t57_n68: return -0.040632568299770355f;
  t57_n69: return 0.03613352030515671f;
  t57_n70: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t57_n88;
  t57_n71: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t57_n87;
  t57_n72: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t57_n74;
  t57_n73: return -0.010811444371938705f;
  t57_n74: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t57_n76;
  t57_n75: return -0.07274466007947922f;
  t57_n76: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t57_n78;
  t57_n77: return -0.08242456614971161f;
  t57_n78: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t57_n82;
  t57_n79: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t57_n81;
  t57_n80: return 0.024170711636543274f;
  t57_n81: return -0.05347376689314842f;
  t57_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t57_n84;
  t57_n83: return -0.0054914215579628944f;
  t57_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t57_n86;
  t57_n85: return 0.009180961176753044f;
  t57_n86: return 0.06020916998386383f;
  t57_n87: return -0.09083235263824463f;
  t57_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n90;
  t57_n89: return -0.04446394369006157f;
  t57_n90: return 0.023578768596053123f;
  t57_n91: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t57_n101;
  t57_n92: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t57_n94;
  t57_n93: return -0.039300452917814255f;
  t57_n94: if (E.getNumReferences() >= 3338301440 /*32056.0*/) goto t57_n100;
  t57_n95: if (E.getNumReferences() >= 3337825536 /*31126.5*/) goto t57_n99;
  t57_n96: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t57_n98;
  t57_n97: return 0.07501694560050964f;
  t57_n98: return 0.14372803270816803f;
  t57_n99: return 0.13503733277320862f;
  t57_n100: return 0.02778218500316143f;
  t57_n101: return 0.012503293342888355f;
  t57_n102: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t57_n252;
  t57_n103: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t57_n135;
  t57_n104: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t57_n130;
  t57_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t57_n125;
  t57_n106: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t57_n124;
  t57_n107: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t57_n115;
  t57_n108: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t57_n112;
  t57_n109: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t57_n111;
  t57_n110: return 0.012184330262243748f;
  t57_n111: return -0.021122626960277557f;
  t57_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t57_n114;
  t57_n113: return 0.015822671353816986f;
  t57_n114: return 0.03939232602715492f;
  t57_n115: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t57_n123;
  t57_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t57_n118;
  t57_n117: return 0.0011619820725172758f;
  t57_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t57_n120;
  t57_n119: return -0.11348192393779755f;
  t57_n120: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t57_n122;
  t57_n121: return -0.004864725284278393f;
  t57_n122: return 0.04595034569501877f;
  t57_n123: return -0.03333396092057228f;
  t57_n124: return 0.009983204305171967f;
  t57_n125: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n129;
  t57_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t57_n128;
  t57_n127: return -0.005435336381196976f;
  t57_n128: return 0.009875200688838959f;
  t57_n129: return 0.030750662088394165f;
  t57_n130: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t57_n132;
  t57_n131: return 0.03792433813214302f;
  t57_n132: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t57_n134;
  t57_n133: return 0.056168049573898315f;
  t57_n134: return 0.007344111334532499f;
  t57_n135: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t57_n217;
  t57_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t57_n144;
  t57_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n141;
  t57_n138: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t57_n140;
  t57_n139: return -0.004517893772572279f;
  t57_n140: return -0.0877976045012474f;
  t57_n141: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t57_n143;
  t57_n142: return 0.013385028578341007f;
  t57_n143: return -0.07458390295505524f;
  t57_n144: if (E.getNumReferences() >= 3267035136 /*93.5*/) goto t57_n214;
  t57_n145: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t57_n165;
  t57_n146: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t57_n164;
  t57_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t57_n149;
  t57_n148: return -0.004246353637427092f;
  t57_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n163;
  t57_n150: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t57_n162;
  t57_n151: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t57_n153;
  t57_n152: return 0.020474858582019806f;
  t57_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t57_n155;
  t57_n154: return 0.0008877521031536162f;
  t57_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t57_n159;
  t57_n156: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t57_n158;
  t57_n157: return 0.15998151898384094f;
  t57_n158: return 0.011222305707633495f;
  t57_n159: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t57_n161;
  t57_n160: return 0.05192353203892708f;
  t57_n161: return 0.01834951899945736f;
  t57_n162: return 0.05088033527135849f;
  t57_n163: return -0.002623109146952629f;
  t57_n164: return 0.027514120563864708f;
  t57_n165: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t57_n169;
  t57_n166: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t57_n168;
  t57_n167: return 0.02034856379032135f;
  t57_n168: return 0.03135654330253601f;
  t57_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t57_n213;
  t57_n170: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t57_n206;
  t57_n171: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t57_n187;
  t57_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n186;
  t57_n173: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t57_n183;
  t57_n174: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t57_n178;
  t57_n175: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t57_n177;
  t57_n176: return -0.012107286602258682f;
  t57_n177: return 0.042502593249082565f;
  t57_n178: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t57_n182;
  t57_n179: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t57_n181;
  t57_n180: return -0.12930066883563995f;
  t57_n181: return 0.11649633198976517f;
  t57_n182: return 0.1496858447790146f;
  t57_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n185;
  t57_n184: return -0.0774768814444542f;
  t57_n185: return -0.05581148713827133f;
  t57_n186: return 0.014929343946278095f;
  t57_n187: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t57_n205;
  t57_n188: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t57_n190;
  t57_n189: return 0.12353364378213882f;
  t57_n190: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t57_n202;
  t57_n191: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t57_n193;
  t57_n192: return 0.005878875497728586f;
  t57_n193: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t57_n195;
  t57_n194: return 0.024331461638212204f;
  t57_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t57_n197;
  t57_n196: return 0.0373389907181263f;
  t57_n197: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t57_n199;
  t57_n198: return 0.10972605645656586f;
  t57_n199: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t57_n201;
  t57_n200: return 0.012294663116335869f;
  t57_n201: return 0.15016289055347443f;
  t57_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t57_n204;
  t57_n203: return -0.2675694525241852f;
  t57_n204: return 0.06899367272853851f;
  t57_n205: return 0.038568805903196335f;
  t57_n206: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t57_n208;
  t57_n207: return 0.01662321761250496f;
  t57_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t57_n210;
  t57_n209: return 0.04475810378789902f;
  t57_n210: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t57_n212;
  t57_n211: return 0.11204632371664047f;
  t57_n212: return -0.13315695524215698f;
  t57_n213: return 0.030032774433493614f;
  t57_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t57_n216;
  t57_n215: return 0.02931865304708481f;
  t57_n216: return 0.046480901539325714f;
  t57_n217: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t57_n251;
  t57_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t57_n240;
  t57_n219: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t57_n233;
  t57_n220: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t57_n222;
  t57_n221: return 0.018088987097144127f;
  t57_n222: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t57_n230;
  t57_n223: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t57_n225;
  t57_n224: return 0.05576397478580475f;
  t57_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t57_n227;
  t57_n226: return 0.012176544405519962f;
  t57_n227: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t57_n229;
  t57_n228: return 0.12551741302013397f;
  t57_n229: return 0.007986792363226414f;
  t57_n230: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t57_n232;
  t57_n231: return -0.09628885984420776f;
  t57_n232: return 0.04185245558619499f;
  t57_n233: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t57_n239;
  t57_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t57_n238;
  t57_n235: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t57_n237;
  t57_n236: return -0.07656855881214142f;
  t57_n237: return 0.04526727646589279f;
  t57_n238: return 0.07099661976099014f;
  t57_n239: return 0.019536521285772324f;
  t57_n240: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t57_n242;
  t57_n241: return 0.0001861182099673897f;
  t57_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t57_n244;
  t57_n243: return 0.02656443789601326f;
  t57_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t57_n246;
  t57_n245: return 0.03865004703402519f;
  t57_n246: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t57_n250;
  t57_n247: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t57_n249;
  t57_n248: return -0.10626735538244247f;
  t57_n249: return 0.10717461258172989f;
  t57_n250: return 0.025830265134572983f;
  t57_n251: return 0.06771646440029144f;
  t57_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t57_n254;
  t57_n253: return 0.027007006108760834f;
  t57_n254: return 0.06599516421556473f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree58(const Example& E) {
  t58_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t58_n144;
  t58_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t58_n131;
  t58_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t58_n116;
  t58_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t58_n97;
  t58_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t58_n94;
  t58_n5: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t58_n67;
  t58_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t58_n40;
  t58_n7: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t58_n35;
  t58_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t58_n14;
  t58_n9: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t58_n11;
  t58_n10: return 0.005745461210608482f;
  t58_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t58_n13;
  t58_n12: return -0.0032711524982005358f;
  t58_n13: return 0.05087712034583092f;
  t58_n14: if (E.getNumReferences() >= 3342302208 /*46972.0*/) goto t58_n34;
  t58_n15: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t58_n33;
  t58_n16: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t58_n18;
  t58_n17: return -0.009297071024775505f;
  t58_n18: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t58_n22;
  t58_n19: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t58_n21;
  t58_n20: return 0.03125448897480965f;
  t58_n21: return -0.001828620908781886f;
  t58_n22: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t58_n24;
  t58_n23: return -0.038213711231946945f;
  t58_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t58_n26;
  t58_n25: return 0.011072822846472263f;
  t58_n26: if (E.getNumReferences() >= 3285336064 /*420.5*/) goto t58_n28;
  t58_n27: return 0.12720438838005066f;
  t58_n28: if (E.getNumReferences() >= 3300208640 /*1449.5*/) goto t58_n32;
  t58_n29: if (E.getNumReferences() >= 3300044800 /*1429.5*/) goto t58_n31;
  t58_n30: return 0.07943032681941986f;
  t58_n31: return 0.22994694113731384f;
  t58_n32: return 0.02105853334069252f;
  t58_n33: return -0.10589955002069473f;
  t58_n34: return -0.15487782657146454f;
  t58_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t58_n37;
  t58_n36: return -0.07416082918643951f;
  t58_n37: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t58_n39;
  t58_n38: return 0.0180609580129385f;
  t58_n39: return 0.04177280142903328f;
  t58_n40: if (E.getNumReferences() >= 3334946304 /*25503.0*/) goto t58_n66;
  t58_n41: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n45;
  t58_n42: if (E.getNumReferences() >= 3275194368 /*183.5*/) goto t58_n44;
  t58_n43: return 0.014190821908414364f;
  t58_n44: return 0.061464861035346985f;
  t58_n45: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t58_n63;
  t58_n46: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t58_n48;
  t58_n47: return -0.019724467769265175f;
  t58_n48: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t58_n50;
  t58_n49: return 0.0031622766982764006f;
  t58_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t58_n52;
  t58_n51: return 0.010989846661686897f;
  t58_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t58_n54;
  t58_n53: return -0.10573021322488785f;
  t58_n54: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t58_n60;
  t58_n55: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t58_n59;
  t58_n56: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t58_n58;
  t58_n57: return 0.025386903434991837f;
  t58_n58: return 0.048729512840509415f;
  t58_n59: return -0.057623717933893204f;
  t58_n60: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t58_n62;
  t58_n61: return 0.057236652821302414f;
  t58_n62: return -0.10807570070028305f;
  t58_n63: if (E.getNumReferences() >= 3321944576 /*8246.5*/) goto t58_n65;
  t58_n64: return -0.030439872294664383f;
  t58_n65: return 0.03284737095236778f;
  t58_n66: return 0.06254015862941742f;
  t58_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t58_n71;
  t58_n68: if (E.getNumReferences() >= 3343995264 /*53585.5*/) goto t58_n70;
  t58_n69: return 0.09710567444562912f;
  t58_n70: return -0.029718928039073944f;
  t58_n71: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t58_n91;
  t58_n72: if (E.getNumReferences() >= 3359032704 /*187158.0*/) goto t58_n80;
  t58_n73: if (E.getNumReferences() >= 3356904640 /*153907.0*/) goto t58_n79;
  t58_n74: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t58_n76;
  t58_n75: return 0.028909100219607353f;
  t58_n76: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t58_n78;
  t58_n77: return -0.01826605573296547f;
  t58_n78: return 0.07268374413251877f;
  t58_n79: return -0.2670902907848358f;
  t58_n80: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n82;
  t58_n81: return 0.05000597611069679f;
  t58_n82: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t58_n86;
  t58_n83: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t58_n85;
  t58_n84: return 0.10010747611522675f;
  t58_n85: return 0.13634440302848816f;
  t58_n86: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t58_n90;
  t58_n87: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t58_n89;
  t58_n88: return 0.07293318212032318f;
  t58_n89: return -0.22774949669837952f;
  t58_n90: return 0.10413713753223419f;
  t58_n91: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t58_n93;
  t58_n92: return -0.11478741466999054f;
  t58_n93: return 0.04614109545946121f;
  t58_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t58_n96;
  t58_n95: return -0.06930915266275406f;
  t58_n96: return -0.04900979995727539f;
  t58_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t58_n103;
  t58_n98: if (E.getNumReferences() >= 3364454096 /*281590.5*/) goto t58_n102;
  t58_n99: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t58_n101;
  t58_n100: return -0.0007539942744188011f;
  t58_n101: return 0.01611262373626232f;
  t58_n102: return 0.040802519768476486f;
  t58_n103: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t58_n115;
  t58_n104: if (E.getNumReferences() >= 3333689088 /*23047.5*/) goto t58_n112;
  t58_n105: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n111;
  t58_n106: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t58_n110;
  t58_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t58_n109;
  t58_n108: return -0.0002297559258295223f;
  t58_n109: return 0.05099557712674141f;
  t58_n110: return -0.061519112437963486f;
  t58_n111: return 0.010283700190484524f;
  t58_n112: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t58_n114;
  t58_n113: return 0.08052412420511246f;
  t58_n114: return 0.036998897790908813f;
  t58_n115: return 0.04598081856966019f;
  t58_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t58_n118;
  t58_n117: return -0.07637163996696472f;
  t58_n118: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t58_n120;
  t58_n119: return -0.021487630903720856f;
  t58_n120: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n130;
  t58_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t58_n123;
  t58_n122: return 0.016235224902629852f;
  t58_n123: if (E.getNumReferences() >= 3354001600 /*119809.5*/) goto t58_n125;
  t58_n124: return 0.08586341142654419f;
  t58_n125: if (E.getNumReferences() >= 3355438208 /*131033.0*/) goto t58_n129;
  t58_n126: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t58_n128;
  t58_n127: return 0.13499245047569275f;
  t58_n128: return -0.10264934599399567f;
  t58_n129: return 0.08689755201339722f;
  t58_n130: return -0.007247872184962034f;
  t58_n131: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t58_n139;
  t58_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t58_n138;
  t58_n133: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t58_n135;
  t58_n134: return 0.04319315403699875f;
  t58_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t58_n137;
  t58_n136: return -0.0001255378156201914f;
  t58_n137: return 0.012436351738870144f;
  t58_n138: return 0.038417719304561615f;
  t58_n139: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t58_n141;
  t58_n140: return 0.03415834531188011f;
  t58_n141: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t58_n143;
  t58_n142: return 0.11088424175977707f;
  t58_n143: return 0.07635984569787979f;
  t58_n144: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t58_n254;
  t58_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t58_n193;
  t58_n146: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t58_n192;
  t58_n147: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n179;
  t58_n148: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t58_n166;
  t58_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t58_n155;
  t58_n150: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t58_n154;
  t58_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t58_n153;
  t58_n152: return -0.04298705980181694f;
  t58_n153: return 0.0023011721204966307f;
  t58_n154: return -0.033080894500017166f;
  t58_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t58_n157;
  t58_n156: return 0.005671329330652952f;
  t58_n157: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t58_n163;
  t58_n158: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t58_n162;
  t58_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t58_n161;
  t58_n160: return 0.001895543304271996f;
  t58_n161: return 0.017014803364872932f;
  t58_n162: return -0.06268217414617538f;
  t58_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t58_n165;
  t58_n164: return 0.011737114749848843f;
  t58_n165: return 0.03531753271818161f;
  t58_n166: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t58_n176;
  t58_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t58_n173;
  t58_n168: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t58_n170;
  t58_n169: return 0.009502037428319454f;
  t58_n170: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t58_n172;
  t58_n171: return 0.07351760566234589f;
  t58_n172: return 0.15864048898220062f;
  t58_n173: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t58_n175;
  t58_n174: return 0.0432850606739521f;
  t58_n175: return 0.028206879273056984f;
  t58_n176: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t58_n178;
  t58_n177: return 0.029845360666513443f;
  t58_n178: return 0.09760750085115433f;
  t58_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t58_n191;
  t58_n180: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t58_n182;
  t58_n181: return 0.03967219963669777f;
  t58_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t58_n184;
  t58_n183: return -0.003356190398335457f;
  t58_n184: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t58_n188;
  t58_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t58_n187;
  t58_n186: return -0.11523276567459106f;
  t58_n187: return -0.04059493541717529f;
  t58_n188: if (E.getNumReferences() >= 3266904064 /*92.5*/) goto t58_n190;
  t58_n189: return 0.023630831390619278f;
  t58_n190: return 0.04714222252368927f;
  t58_n191: return 0.023455718532204628f;
  t58_n192: return 0.08054807782173157f;
  t58_n193: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t58_n251;
  t58_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t58_n196;
  t58_n195: return 0.005623173899948597f;
  t58_n196: if (E.getNumReferences() >= 3273228288 /*153.5*/) goto t58_n250;
  t58_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t58_n247;
  t58_n198: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t58_n216;
  t58_n199: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t58_n203;
  t58_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type))) goto t58_n202;
  t58_n201: return 0.013286140747368336f;
  t58_n202: return 0.03890569135546684f;
  t58_n203: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t58_n205;
  t58_n204: return 0.025091493502259254f;
  t58_n205: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t58_n215;
  t58_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t58_n208;
  t58_n207: return 0.0522083044052124f;
  t58_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t58_n210;
  t58_n209: return -0.008078569546341896f;
  t58_n210: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t58_n214;
  t58_n211: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t58_n213;
  t58_n212: return 0.04851605370640755f;
  t58_n213: return 0.12661920487880707f;
  t58_n214: return 0.12964653968811035f;
  t58_n215: return 0.015169814229011536f;
  t58_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t58_n240;
  t58_n217: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t58_n227;
  t58_n218: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t58_n220;
  t58_n219: return 0.0010288914199918509f;
  t58_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t58_n222;
  t58_n221: return -0.0047188797034323215f;
  t58_n222: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t58_n224;
  t58_n223: return 0.03518938645720482f;
  t58_n224: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t58_n226;
  t58_n225: return 0.09443113207817078f;
  t58_n226: return -0.01169491745531559f;
  t58_n227: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t58_n237;
  t58_n228: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t58_n232;
  t58_n229: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t58_n231;
  t58_n230: return -0.07062769681215286f;
  t58_n231: return 0.1686832308769226f;
  t58_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t58_n234;
  t58_n233: return -0.04389209300279617f;
  t58_n234: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t58_n236;
  t58_n235: return 0.014953614212572575f;
  t58_n236: return 0.04885664954781532f;
  t58_n237: if (E.getFileProximityDistanceCost() >= 3249012736 /*21.0*/) goto t58_n239;
  t58_n238: return 0.11317102611064911f;
  t58_n239: return -0.0019799857400357723f;
  t58_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t58_n246;
  t58_n241: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t58_n243;
  t58_n242: return -0.000658760720398277f;
  t58_n243: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t58_n245;
  t58_n244: return 0.07622232288122177f;
  t58_n245: return 0.19462355971336365f;
  t58_n246: return 0.02684725821018219f;
  t58_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t58_n249;
  t58_n248: return 0.01857583038508892f;
  t58_n249: return 0.054491881281137466f;
  t58_n250: return 0.03832670673727989f;
  t58_n251: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t58_n253;
  t58_n252: return 0.01964903064072132f;
  t58_n253: return 0.06600696593523026f;
  t58_n254: return 0.03185277432203293f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree59(const Example& E) {
  t59_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t59_n94;
  t59_n1: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t59_n21;
  t59_n2: if (E.getNumReferences() >= 3301818368 /*1646.0*/) goto t59_n12;
  t59_n3: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t59_n9;
  t59_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n6;
  t59_n5: return -0.003877968993037939f;
  t59_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t59_n8;
  t59_n7: return -0.06107764691114426f;
  t59_n8: return 0.005834126845002174f;
  t59_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t59_n11;
  t59_n10: return -0.0056140064261853695f;
  t59_n11: return 0.02550564333796501f;
  t59_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n20;
  t59_n13: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t59_n17;
  t59_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t59_n16;
  t59_n15: return 0.0202465932816267f;
  t59_n16: return 0.03924234211444855f;
  t59_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t59_n19;
  t59_n18: return -0.043568436056375504f;
  t59_n19: return 0.013557597994804382f;
  t59_n20: return 0.05328943952918053f;
  t59_n21: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t59_n33;
  t59_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n30;
  t59_n23: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t59_n25;
  t59_n24: return -0.0033102771267294884f;
  t59_n25: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t59_n27;
  t59_n26: return 0.03299844264984131f;
  t59_n27: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t59_n29;
  t59_n28: return 0.0721064954996109f;
  t59_n29: return 0.12277809530496597f;
  t59_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t59_n32;
  t59_n31: return -0.07009876519441605f;
  t59_n32: return 0.02579347975552082f;
  t59_n33: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t59_n91;
  t59_n34: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t59_n70;
  t59_n35: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t59_n69;
  t59_n36: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t59_n66;
  t59_n37: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t59_n65;
  t59_n38: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n40;
  t59_n39: return -0.013660753145813942f;
  t59_n40: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t59_n56;
  t59_n41: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t59_n43;
  t59_n42: return -0.027926357463002205f;
  t59_n43: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t59_n45;
  t59_n44: return -0.006634158082306385f;
  t59_n45: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t59_n55;
  t59_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t59_n48;
  t59_n47: return 0.07267844676971436f;
  t59_n48: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t59_n54;
  t59_n49: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t59_n51;
  t59_n50: return 0.09704960882663727f;
  t59_n51: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t59_n53;
  t59_n52: return 0.12355966866016388f;
  t59_n53: return -0.012012451887130737f;
  t59_n54: return 0.10639522969722748f;
  t59_n55: return 0.031210025772452354f;
  t59_n56: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t59_n64;
  t59_n57: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t59_n63;
  t59_n58: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t59_n62;
  t59_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t59_n61;
  t59_n60: return -0.034680746495723724f;
  t59_n61: return 0.024791307747364044f;
  t59_n62: return -0.03804098814725876f;
  t59_n63: return 0.02093491703271866f;
  t59_n64: return -0.1226152554154396f;
  t59_n65: return -0.2461247593164444f;
  t59_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n68;
  t59_n67: return -0.20061367750167847f;
  t59_n68: return 0.09594564139842987f;
  t59_n69: return -0.13669416308403015f;
  t59_n70: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t59_n80;
  t59_n71: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t59_n73;
  t59_n72: return 0.026228733360767365f;
  t59_n73: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t59_n77;
  t59_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t59_n76;
  t59_n75: return -0.03518777713179588f;
  t59_n76: return 0.09432277828454971f;
  t59_n77: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t59_n79;
  t59_n78: return 0.015169422142207623f;
  t59_n79: return 0.09863307327032089f;
  t59_n80: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t59_n88;
  t59_n81: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t59_n87;
  t59_n82: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t59_n84;
  t59_n83: return -0.15944910049438477f;
  t59_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t59_n86;
  t59_n85: return 0.04362988844513893f;
  t59_n86: return 0.10159620642662048f;
  t59_n87: return -0.08969251066446304f;
  t59_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n90;
  t59_n89: return -0.13655290007591248f;
  t59_n90: return 0.020588364452123642f;
  t59_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n93;
  t59_n92: return -0.035194799304008484f;
  t59_n93: return 0.023634951561689377f;
  t59_n94: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t59_n250;
  t59_n95: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t59_n175;
  t59_n96: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t59_n162;
  t59_n97: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t59_n147;
  t59_n98: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t59_n146;
  t59_n99: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t59_n139;
  t59_n100: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t59_n104;
  t59_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t59_n103;
  t59_n102: return 0.014455555006861687f;
  t59_n103: return 0.029116343706846237f;
  t59_n104: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t59_n138;
  t59_n105: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t59_n137;
  t59_n106: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t59_n136;
  t59_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t59_n115;
  t59_n108: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t59_n110;
  t59_n109: return -0.02599988505244255f;
  t59_n110: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n114;
  t59_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t59_n113;
  t59_n112: return -0.0056568896397948265f;
  t59_n113: return 0.031794220209121704f;
  t59_n114: return 0.03310787305235863f;
  t59_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t59_n135;
  t59_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n134;
  t59_n117: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t59_n133;
  t59_n118: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t59_n120;
  t59_n119: return -0.0008184568141587079f;
  t59_n120: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t59_n122;
  t59_n121: return 0.003354314947500825f;
  t59_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t59_n124;
  t59_n123: return -0.03904210403561592f;
  t59_n124: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t59_n130;
  t59_n125: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t59_n129;
  t59_n126: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t59_n128;
  t59_n127: return 0.037206508219242096f;
  t59_n128: return 0.039573170244693756f;
  t59_n129: return -0.026939550414681435f;
  t59_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t59_n132;
  t59_n131: return -0.019952718168497086f;
  t59_n132: return 0.07626311480998993f;
  t59_n133: return -0.014350627548992634f;
  t59_n134: return 0.01575460471212864f;
  t59_n135: return 0.008937377482652664f;
  t59_n136: return -0.047735750675201416f;
  t59_n137: return -0.10121551901102066f;
  t59_n138: return -0.017890850082039833f;
  t59_n139: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t59_n141;
  t59_n140: return -0.06938375532627106f;
  t59_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t59_n145;
  t59_n142: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t59_n144;
  t59_n143: return 0.012748331762850285f;
  t59_n144: return 0.03869986534118652f;
  t59_n145: return 0.02579071931540966f;
  t59_n146: return -0.02853870764374733f;
  t59_n147: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t59_n149;
  t59_n148: return -0.05435968190431595f;
  t59_n149: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t59_n155;
  t59_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t59_n154;
  t59_n151: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n153;
  t59_n152: return 0.012755039148032665f;
  t59_n153: return 0.04822720214724541f;
  t59_n154: return 0.03404625505208969f;
  t59_n155: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t59_n161;
  t59_n156: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t59_n158;
  t59_n157: return 0.014629373326897621f;
  t59_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t59_n160;
  t59_n159: return 0.0910377949476242f;
  t59_n160: return 0.15357670187950134f;
  t59_n161: return -0.04947579652070999f;
  t59_n162: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t59_n172;
  t59_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t59_n171;
  t59_n164: if (E.getNumReferences() >= 3314693120 /*4678.5*/) goto t59_n166;
  t59_n165: return 0.020958367735147476f;
  t59_n166: if (E.getNumReferences() >= 3315203072 /*4927.5*/) goto t59_n168;
  t59_n167: return 0.11769755184650421f;
  t59_n168: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n170;
  t59_n169: return 0.03995514288544655f;
  t59_n170: return 0.0880064070224762f;
  t59_n171: return 0.07968740165233612f;
  t59_n172: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t59_n174;
  t59_n173: return 0.042179640382528305f;
  t59_n174: return 0.0052430047653615475f;
  t59_n175: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t59_n231;
  t59_n176: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n206;
  t59_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t59_n185;
  t59_n178: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t59_n184;
  t59_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t59_n181;
  t59_n180: return -0.017612449824810028f;
  t59_n181: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t59_n183;
  t59_n182: return -0.0004533436440397054f;
  t59_n183: return 0.01652233488857746f;
  t59_n184: return -0.0622582770884037f;
  t59_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t59_n205;
  t59_n186: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t59_n196;
  t59_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t59_n193;
  t59_n188: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t59_n192;
  t59_n189: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t59_n191;
  t59_n190: return 0.01796489953994751f;
  t59_n191: return 0.038323551416397095f;
  t59_n192: return 0.02848539873957634f;
  t59_n193: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t59_n195;
  t59_n194: return 0.02121574804186821f;
  t59_n195: return -0.15921242535114288f;
  t59_n196: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t59_n198;
  t59_n197: return -0.058783359825611115f;
  t59_n198: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t59_n202;
  t59_n199: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t59_n201;
  t59_n200: return 0.00887239445000887f;
  t59_n201: return -0.060445405542850494f;
  t59_n202: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t59_n204;
  t59_n203: return 0.1358916014432907f;
  t59_n204: return 0.039423320442438126f;
  t59_n205: return -0.0059119779616594315f;
  t59_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t59_n228;
  t59_n207: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t59_n221;
  t59_n208: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t59_n210;
  t59_n209: return 0.016027072444558144f;
  t59_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t59_n218;
  t59_n211: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t59_n213;
  t59_n212: return 0.05555806681513786f;
  t59_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t59_n215;
  t59_n214: return 0.014948275871574879f;
  t59_n215: if (E.getFractionNameInContext() >= 3207524079 /*0.6833333373069763*/) goto t59_n217;
  t59_n216: return 0.1241416484117508f;
  t59_n217: return -0.023492805659770966f;
  t59_n218: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t59_n220;
  t59_n219: return -0.08224816620349884f;
  t59_n220: return 0.04017689451575279f;
  t59_n221: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t59_n227;
  t59_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t59_n226;
  t59_n223: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t59_n225;
  t59_n224: return -0.07331603765487671f;
  t59_n225: return 0.04153907671570778f;
  t59_n226: return 0.06672044098377228f;
  t59_n227: return 0.018170326948165894f;
  t59_n228: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t59_n230;
  t59_n229: return 0.010993768461048603f;
  t59_n230: return 0.03824469819664955f;
  t59_n231: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t59_n233;
  t59_n232: return 0.030502518638968468f;
  t59_n233: if (E.getNumReferences() >= 3347421312 /*68401.0*/) goto t59_n243;
  t59_n234: if (E.getNumReferences() >= 3346253184 /*62405.5*/) goto t59_n240;
  t59_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t59_n239;
  t59_n236: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t59_n238;
  t59_n237: return -0.002068500267341733f;
  t59_n238: return 0.04230945557355881f;
  t59_n239: return 0.02252013236284256f;
  t59_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t59_n242;
  t59_n241: return -0.6988016963005066f;
  t59_n242: return -0.07712960988283157f;
  t59_n243: if (E.getNumReferences() >= 3351977984 /*104000.0*/) goto t59_n249;
  t59_n244: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t59_n248;
  t59_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t59_n247;
  t59_n246: return 0.057254306972026825f;
  t59_n247: return 0.10149210691452026f;
  t59_n248: return 0.12990790605545044f;
  t59_n249: return 0.044759973883628845f;
  t59_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t59_n254;
  t59_n251: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t59_n253;
  t59_n252: return 0.024936603382229805f;
  t59_n253: return 0.06332670897245407f;
  t59_n254: return -0.3150322735309601f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree60(const Example& E) {
  t60_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t60_n50;
  t60_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t60_n37;
  t60_n2: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t60_n30;
  t60_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t60_n21;
  t60_n4: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t60_n20;
  t60_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t60_n15;
  t60_n6: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t60_n14;
  t60_n7: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t60_n11;
  t60_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t60_n10;
  t60_n9: return -0.002538092201575637f;
  t60_n10: return 0.012585564516484737f;
  t60_n11: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t60_n13;
  t60_n12: return -0.046939488500356674f;
  t60_n13: return -0.0001051498475135304f;
  t60_n14: return -0.11476488411426544f;
  t60_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t60_n17;
  t60_n16: return 0.0013842800399288535f;
  t60_n17: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t60_n19;
  t60_n18: return 0.06640665978193283f;
  t60_n19: return -0.038415566086769104f;
  t60_n20: return 0.011636472307145596f;
  t60_n21: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t60_n29;
  t60_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t60_n26;
  t60_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n25;
  t60_n24: return -0.0837666466832161f;
  t60_n25: return 0.004213890992105007f;
  t60_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n28;
  t60_n27: return 0.017305100336670876f;
  t60_n28: return 0.05191721022129059f;
  t60_n29: return -0.028422033414244652f;
  t60_n30: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t60_n32;
  t60_n31: return 0.026066744700074196f;
  t60_n32: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t60_n36;
  t60_n33: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t60_n35;
  t60_n34: return -0.009893939830362797f;
  t60_n35: return 0.1261860430240631f;
  t60_n36: return 0.04976096749305725f;
  t60_n37: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t60_n39;
  t60_n38: return -0.011415326036512852f;
  t60_n39: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t60_n49;
  t60_n40: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t60_n48;
  t60_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n47;
  t60_n42: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t60_n46;
  t60_n43: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t60_n45;
  t60_n44: return 0.06153779849410057f;
  t60_n45: return -0.04387462139129639f;
  t60_n46: return -0.08954038470983505f;
  t60_n47: return 0.008489402011036873f;
  t60_n48: return -0.07492697238922119f;
  t60_n49: return -0.04830457270145416f;
  t60_n50: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t60_n254;
  t60_n51: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t60_n193;
  t60_n52: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t60_n180;
  t60_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t60_n169;
  t60_n54: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t60_n108;
  t60_n55: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t60_n101;
  t60_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n80;
  t60_n57: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t60_n79;
  t60_n58: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t60_n76;
  t60_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n65;
  t60_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t60_n64;
  t60_n61: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t60_n63;
  t60_n62: return -0.12301970273256302f;
  t60_n63: return -0.24329298734664917f;
  t60_n64: return -0.10823965072631836f;
  t60_n65: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t60_n67;
  t60_n66: return -0.0072914049960672855f;
  t60_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t60_n69;
  t60_n68: return 0.0011201559100300074f;
  t60_n69: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t60_n75;
  t60_n70: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t60_n72;
  t60_n71: return 0.054042525589466095f;
  t60_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t60_n74;
  t60_n73: return 0.09501705318689346f;
  t60_n74: return -0.10346771776676178f;
  t60_n75: return -0.12171506136655807f;
  t60_n76: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t60_n78;
  t60_n77: return 0.1077246442437172f;
  t60_n78: return 0.0021168862003833055f;
  t60_n79: return -0.03250298276543617f;
  t60_n80: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t60_n100;
  t60_n81: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t60_n97;
  t60_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t60_n84;
  t60_n83: return 0.019791506230831146f;
  t60_n84: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t60_n94;
  t60_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t60_n93;
  t60_n86: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t60_n88;
  t60_n87: return 0.006986898835748434f;
  t60_n88: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t60_n92;
  t60_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n91;
  t60_n90: return -0.11930187046527863f;
  t60_n91: return 0.15733471512794495f;
  t60_n92: return 0.01858740672469139f;
  t60_n93: return 0.029666204005479813f;
  t60_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t60_n96;
  t60_n95: return 0.028346847742795944f;
  t60_n96: return 0.06399893760681152f;
  t60_n97: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t60_n99;
  t60_n98: return -0.05812980607151985f;
  t60_n99: return 0.005006530322134495f;
  t60_n100: return -0.016828861087560654f;
  t60_n101: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t60_n107;
  t60_n102: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t60_n106;
  t60_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n105;
  t60_n104: return -0.09243499487638474f;
  t60_n105: return 0.023839155212044716f;
  t60_n106: return 0.06741086393594742f;
  t60_n107: return -0.002877861959859729f;
  t60_n108: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t60_n120;
  t60_n109: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t60_n119;
  t60_n110: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t60_n116;
  t60_n111: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t60_n113;
  t60_n112: return 0.00027067691553384066f;
  t60_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t60_n115;
  t60_n114: return 0.011899604462087154f;
  t60_n115: return 0.07667280733585358f;
  t60_n116: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t60_n118;
  t60_n117: return 0.05108213797211647f;
  t60_n118: return 0.10001002252101898f;
  t60_n119: return -0.049823593348264694f;
  t60_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t60_n158;
  t60_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t60_n157;
  t60_n122: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t60_n156;
  t60_n123: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t60_n137;
  t60_n124: if (E.getNumReferences() >= 3354956672 /*127271.0*/) goto t60_n132;
  t60_n125: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t60_n131;
  t60_n126: if (E.getNumReferences() >= 3269001216 /*108.5*/) goto t60_n128;
  t60_n127: return -0.002471942687407136f;
  t60_n128: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t60_n130;
  t60_n129: return 0.14441601932048798f;
  t60_n130: return 0.019918520003557205f;
  t60_n131: return 0.11685050278902054f;
  t60_n132: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t60_n134;
  t60_n133: return -0.2493424266576767f;
  t60_n134: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t60_n136;
  t60_n135: return 0.09685645252466202f;
  t60_n136: return -0.03230232745409012f;
  t60_n137: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t60_n155;
  t60_n138: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t60_n154;
  t60_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t60_n145;
  t60_n140: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t60_n144;
  t60_n141: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t60_n143;
  t60_n142: return 0.049695517867803574f;
  t60_n143: return 0.008216231130063534f;
  t60_n144: return -0.03682680055499077f;
  t60_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t60_n147;
  t60_n146: return 0.03562617301940918f;
  t60_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t60_n149;
  t60_n148: return 0.02116917073726654f;
  t60_n149: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t60_n153;
  t60_n150: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t60_n152;
  t60_n151: return 0.12648095190525055f;
  t60_n152: return -0.10154607892036438f;
  t60_n153: return 0.04816155880689621f;
  t60_n154: return -0.0028927207458764315f;
  t60_n155: return -0.15447598695755005f;
  t60_n156: return -0.0579410195350647f;
  t60_n157: return 0.028327608481049538f;
  t60_n158: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t60_n160;
  t60_n159: return 0.01830585114657879f;
  t60_n160: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t60_n164;
  t60_n161: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t60_n163;
  t60_n162: return 0.16380886733531952f;
  t60_n163: return 0.08133696019649506f;
  t60_n164: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t60_n166;
  t60_n165: return 0.0160425566136837f;
  t60_n166: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t60_n168;
  t60_n167: return 0.14471328258514404f;
  t60_n168: return 0.06846655905246735f;
  t60_n169: if (E.getNumReferences() >= 3346232448 /*62324.5*/) goto t60_n179;
  t60_n170: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t60_n176;
  t60_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n175;
  t60_n172: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t60_n174;
  t60_n173: return -0.008996978402137756f;
  t60_n174: return 0.07242672890424728f;
  t60_n175: return 0.010294867679476738f;
  t60_n176: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t60_n178;
  t60_n177: return 0.02477002702653408f;
  t60_n178: return 0.04508022964000702f;
  t60_n179: return 0.061268400400877f;
  t60_n180: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t60_n192;
  t60_n181: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t60_n191;
  t60_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t60_n190;
  t60_n183: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t60_n189;
  t60_n184: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t60_n186;
  t60_n185: return 0.027901340276002884f;
  t60_n186: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t60_n188;
  t60_n187: return 0.0035881693474948406f;
  t60_n188: return 0.05607203021645546f;
  t60_n189: return 0.030633170157670975f;
  t60_n190: return 0.016866371035575867f;
  t60_n191: return -0.051290638744831085f;
  t60_n192: return 0.07260030508041382f;
  t60_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t60_n195;
  t60_n194: return 0.009931347332894802f;
  t60_n195: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t60_n249;
  t60_n196: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t60_n240;
  t60_n197: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t60_n211;
  t60_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n204;
  t60_n199: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t60_n201;
  t60_n200: return 0.013470067642629147f;
  t60_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n203;
  t60_n202: return -0.04788099601864815f;
  t60_n203: return -0.03974420577287674f;
  t60_n204: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t60_n208;
  t60_n205: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t60_n207;
  t60_n206: return -0.0019335818942636251f;
  t60_n207: return 0.028729598969221115f;
  t60_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t60_n210;
  t60_n209: return -0.0009523615008220077f;
  t60_n210: return 0.08503548055887222f;
  t60_n211: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t60_n233;
  t60_n212: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t60_n214;
  t60_n213: return -0.0027233115397393703f;
  t60_n214: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t60_n216;
  t60_n215: return 0.019716866314411163f;
  t60_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t60_n220;
  t60_n217: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t60_n219;
  t60_n218: return -0.06371011584997177f;
  t60_n219: return 0.04432837292551994f;
  t60_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t60_n222;
  t60_n221: return 0.02459508739411831f;
  t60_n222: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t60_n224;
  t60_n223: return 0.06781888008117676f;
  t60_n224: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t60_n232;
  t60_n225: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t60_n231;
  t60_n226: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t60_n230;
  t60_n227: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t60_n229;
  t60_n228: return 0.12622664868831635f;
  t60_n229: return -0.12324393540620804f;
  t60_n230: return 0.15971289575099945f;
  t60_n231: return 0.11328479647636414f;
  t60_n232: return -0.05472857505083084f;
  t60_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n235;
  t60_n234: return -0.14366477727890015f;
  t60_n235: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t60_n237;
  t60_n236: return 0.033942870795726776f;
  t60_n237: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t60_n239;
  t60_n238: return 0.04826856032013893f;
  t60_n239: return 0.06980464607477188f;
  t60_n240: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t60_n248;
  t60_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t60_n245;
  t60_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t60_n244;
  t60_n243: return -0.32438424229621887f;
  t60_n244: return 0.011622094549238682f;
  t60_n245: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t60_n247;
  t60_n246: return 0.01878349669277668f;
  t60_n247: return 0.03894997388124466f;
  t60_n248: return -0.14524239301681519f;
  t60_n249: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t60_n253;
  t60_n250: if (E.getNumReferences() >= 3318212608 /*6397.0*/) goto t60_n252;
  t60_n251: return 0.027322864159941673f;
  t60_n252: return 0.08168542385101318f;
  t60_n253: return 0.14891254901885986f;
  t60_n254: return 0.029518358409404755f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree61(const Example& E) {
  t61_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t61_n166;
  t61_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t61_n37;
  t61_n2: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t61_n34;
  t61_n3: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t61_n29;
  t61_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n26;
  t61_n5: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t61_n23;
  t61_n6: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t61_n22;
  t61_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t61_n11;
  t61_n8: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t61_n10;
  t61_n9: return 0.04686436802148819f;
  t61_n10: return -0.036336783319711685f;
  t61_n11: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t61_n17;
  t61_n12: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t61_n16;
  t61_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n15;
  t61_n14: return -0.013805372640490532f;
  t61_n15: return 0.03590245917439461f;
  t61_n16: return -0.021463114768266678f;
  t61_n17: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t61_n21;
  t61_n18: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t61_n20;
  t61_n19: return 0.006185272708535194f;
  t61_n20: return -0.1103881374001503f;
  t61_n21: return -0.03504638373851776f;
  t61_n22: return -0.025234537199139595f;
  t61_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t61_n25;
  t61_n24: return 0.01349903829395771f;
  t61_n25: return -0.022819219157099724f;
  t61_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t61_n28;
  t61_n27: return -0.058023251593112946f;
  t61_n28: return 0.004612463992089033f;
  t61_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n33;
  t61_n30: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t61_n32;
  t61_n31: return 0.004299085587263107f;
  t61_n32: return -0.03631434217095375f;
  t61_n33: return 0.019493643194437027f;
  t61_n34: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t61_n36;
  t61_n35: return 0.022923942655324936f;
  t61_n36: return 0.05189673602581024f;
  t61_n37: if (E.getNumReferences() >= 3328217600 /*14372.5*/) goto t61_n163;
  t61_n38: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t61_n56;
  t61_n39: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t61_n53;
  t61_n40: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t61_n50;
  t61_n41: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t61_n49;
  t61_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t61_n44;
  t61_n43: return -0.00805275235325098f;
  t61_n44: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t61_n46;
  t61_n45: return 0.001588897779583931f;
  t61_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t61_n48;
  t61_n47: return -0.002097941702231765f;
  t61_n48: return 0.044099707156419754f;
  t61_n49: return 0.01988386921584606f;
  t61_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t61_n52;
  t61_n51: return -0.04158875346183777f;
  t61_n52: return 0.030236436054110527f;
  t61_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t61_n55;
  t61_n54: return 0.017013942822813988f;
  t61_n55: return 0.04890824481844902f;
  t61_n56: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t61_n156;
  t61_n57: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t61_n155;
  t61_n58: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t61_n60;
  t61_n59: return -0.019608668982982635f;
  t61_n60: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t61_n138;
  t61_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t61_n127;
  t61_n62: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t61_n104;
  t61_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t61_n67;
  t61_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t61_n66;
  t61_n65: return -0.1482430398464203f;
  t61_n66: return -0.10080592334270477f;
  t61_n67: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t61_n103;
  t61_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t61_n74;
  t61_n69: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t61_n71;
  t61_n70: return -0.04154708608984947f;
  t61_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t61_n73;
  t61_n72: return 0.010267677716910839f;
  t61_n73: return 0.06970490515232086f;
  t61_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t61_n86;
  t61_n75: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t61_n85;
  t61_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n78;
  t61_n77: return -0.10517902672290802f;
  t61_n78: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t61_n82;
  t61_n79: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t61_n81;
  t61_n80: return 0.015734855085611343f;
  t61_n81: return -0.10916602611541748f;
  t61_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t61_n84;
  t61_n83: return 0.0065474798902869225f;
  t61_n84: return 0.06404954195022583f;
  t61_n85: return -0.13172608613967896f;
  t61_n86: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t61_n98;
  t61_n87: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t61_n97;
  t61_n88: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t61_n92;
  t61_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t61_n91;
  t61_n90: return -0.005964307114481926f;
  t61_n91: return 0.08588092029094696f;
  t61_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n94;
  t61_n93: return -0.14184752106666565f;
  t61_n94: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t61_n96;
  t61_n95: return 0.0387892872095108f;
  t61_n96: return 0.05213730409741402f;
  t61_n97: return -0.001928027835674584f;
  t61_n98: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t61_n100;
  t61_n99: return 0.004877576604485512f;
  t61_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n102;
  t61_n101: return -0.15787701308727264f;
  t61_n102: return 0.05348433926701546f;
  t61_n103: return -0.03655645251274109f;
  t61_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t61_n106;
  t61_n105: return -0.03877343609929085f;
  t61_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n108;
  t61_n107: return -0.019737938418984413f;
  t61_n108: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t61_n126;
  t61_n109: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t61_n113;
  t61_n110: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t61_n112;
  t61_n111: return 0.0663054883480072f;
  t61_n112: return -0.1252143532037735f;
  t61_n113: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t61_n115;
  t61_n114: return 0.07440772652626038f;
  t61_n115: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t61_n117;
  t61_n116: return -0.1450224220752716f;
  t61_n117: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t61_n119;
  t61_n118: return 0.07321043312549591f;
  t61_n119: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t61_n121;
  t61_n120: return -0.10733359307050705f;
  t61_n121: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t61_n123;
  t61_n122: return 0.07532592117786407f;
  t61_n123: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t61_n125;
  t61_n124: return -0.11914955079555511f;
  t61_n125: return 0.06339535862207413f;
  t61_n126: return -0.005479310639202595f;
  t61_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t61_n133;
  t61_n128: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t61_n132;
  t61_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t61_n131;
  t61_n130: return -0.01234228815883398f;
  t61_n131: return 0.03827163949608803f;
  t61_n132: return 0.04367470741271973f;
  t61_n133: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t61_n135;
  t61_n134: return 0.06070978194475174f;
  t61_n135: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t61_n137;
  t61_n136: return 0.09478320926427841f;
  t61_n137: return 0.0697815790772438f;
  t61_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type))) goto t61_n142;
  t61_n139: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t61_n141;
  t61_n140: return -0.02778046764433384f;
  t61_n141: return 0.12403631210327148f;
  t61_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t61_n152;
  t61_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t61_n151;
  t61_n144: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t61_n146;
  t61_n145: return -0.01061770599335432f;
  t61_n146: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t61_n150;
  t61_n147: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t61_n149;
  t61_n148: return 0.018069762736558914f;
  t61_n149: return -0.02728543058037758f;
  t61_n150: return -0.04056255519390106f;
  t61_n151: return 0.016250479966402054f;
  t61_n152: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t61_n154;
  t61_n153: return 0.04237207770347595f;
  t61_n154: return 0.007422917056828737f;
  t61_n155: return -0.11163546144962311f;
  t61_n156: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t61_n158;
  t61_n157: return -0.08618396520614624f;
  t61_n158: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t61_n162;
  t61_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t61_n161;
  t61_n160: return -0.0014521604171022773f;
  t61_n161: return 0.03654273599386215f;
  t61_n162: return -0.0352800078690052f;
  t61_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t61_n165;
  t61_n164: return 0.003168744733557105f;
  t61_n165: return 0.03354918956756592f;
  t61_n166: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t61_n254;
  t61_n167: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t61_n215;
  t61_n168: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t61_n206;
  t61_n169: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t61_n201;
  t61_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t61_n172;
  t61_n171: return -0.11562870442867279f;
  t61_n172: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t61_n200;
  t61_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t61_n177;
  t61_n174: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t61_n176;
  t61_n175: return -0.00010789276711875573f;
  t61_n176: return -0.04037828743457794f;
  t61_n177: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t61_n181;
  t61_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t61_n180;
  t61_n179: return 0.010400710627436638f;
  t61_n180: return -0.01550176553428173f;
  t61_n181: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t61_n183;
  t61_n182: return -0.0012235015165060759f;
  t61_n183: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t61_n197;
  t61_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t61_n194;
  t61_n185: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t61_n191;
  t61_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n190;
  t61_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n189;
  t61_n188: return -0.14729806780815125f;
  t61_n189: return -0.03194013610482216f;
  t61_n190: return 0.011349315755069256f;
  t61_n191: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t61_n193;
  t61_n192: return 0.019577806815505028f;
  t61_n193: return 0.055406127125024796f;
  t61_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t61_n196;
  t61_n195: return -0.050078898668289185f;
  t61_n196: return 0.044851236045360565f;
  t61_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t61_n199;
  t61_n198: return -0.015523991547524929f;
  t61_n199: return 0.04290610924363136f;
  t61_n200: return -0.09386811405420303f;
  t61_n201: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t61_n203;
  t61_n202: return -0.049540966749191284f;
  t61_n203: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t61_n205;
  t61_n204: return 0.016601761803030968f;
  t61_n205: return -0.0205409936606884f;
  t61_n206: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t61_n212;
  t61_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t61_n211;
  t61_n208: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t61_n210;
  t61_n209: return 0.02349972352385521f;
  t61_n210: return 0.0514121912419796f;
  t61_n211: return 0.07682221382856369f;
  t61_n212: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t61_n214;
  t61_n213: return 0.036363016813993454f;
  t61_n214: return 0.0013180304085835814f;
  t61_n215: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t61_n251;
  t61_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t61_n222;
  t61_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t61_n221;
  t61_n218: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t61_n220;
  t61_n219: return -0.00431204866617918f;
  t61_n220: return -0.07469969242811203f;
  t61_n221: return 0.010293903760612011f;
  t61_n222: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t61_n250;
  t61_n223: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t61_n249;
  t61_n224: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t61_n228;
  t61_n225: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t61_n227;
  t61_n226: return 0.009233581833541393f;
  t61_n227: return -0.009048271924257278f;
  t61_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t61_n242;
  t61_n229: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t61_n231;
  t61_n230: return 0.0003032396489288658f;
  t61_n231: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t61_n239;
  t61_n232: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t61_n234;
  t61_n233: return 0.029740985482931137f;
  t61_n234: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t61_n236;
  t61_n235: return 0.07768227159976959f;
  t61_n236: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t61_n238;
  t61_n237: return -0.33251050114631653f;
  t61_n238: return 0.14567887783050537f;
  t61_n239: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t61_n241;
  t61_n240: return -0.03352377563714981f;
  t61_n241: return 0.07792744040489197f;
  t61_n242: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t61_n248;
  t61_n243: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t61_n245;
  t61_n244: return 0.028194133192300797f;
  t61_n245: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t61_n247;
  t61_n246: return 0.012396966107189655f;
  t61_n247: return 0.058560531586408615f;
  t61_n248: return 0.0166033748537302f;
  t61_n249: return 0.022797834128141403f;
  t61_n250: return 0.03279786929488182f;
  t61_n251: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t61_n253;
  t61_n252: return 0.018236296251416206f;
  t61_n253: return 0.06363079696893692f;
  t61_n254: return 0.028344156220555305f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree62(const Example& E) {
  t62_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t62_n204;
  t62_n1: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t62_n155;
  t62_n2: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t62_n102;
  t62_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t62_n67;
  t62_n4: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t62_n66;
  t62_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n61;
  t62_n6: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t62_n56;
  t62_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t62_n49;
  t62_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n10;
  t62_n9: return -0.10950612276792526f;
  t62_n10: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t62_n48;
  t62_n11: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t62_n33;
  t62_n12: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t62_n28;
  t62_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t62_n25;
  t62_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t62_n16;
  t62_n15: return -0.010897117666900158f;
  t62_n16: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t62_n22;
  t62_n17: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t62_n19;
  t62_n18: return -0.0595037005841732f;
  t62_n19: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n21;
  t62_n20: return 0.002362153260037303f;
  t62_n21: return -0.04574383795261383f;
  t62_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n24;
  t62_n23: return -0.03757384419441223f;
  t62_n24: return 0.032936859875917435f;
  t62_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t62_n27;
  t62_n26: return -0.09039995074272156f;
  t62_n27: return -0.0704164057970047f;
  t62_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t62_n30;
  t62_n29: return -0.01015009917318821f;
  t62_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t62_n32;
  t62_n31: return 0.015654820948839188f;
  t62_n32: return 0.09589219838380814f;
  t62_n33: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t62_n35;
  t62_n34: return -0.002597517566755414f;
  t62_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t62_n47;
  t62_n36: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t62_n38;
  t62_n37: return 0.024480648338794708f;
  t62_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t62_n40;
  t62_n39: return -0.0038216484244912863f;
  t62_n40: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n46;
  t62_n41: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t62_n45;
  t62_n42: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t62_n44;
  t62_n43: return 0.044035404920578f;
  t62_n44: return -0.019359517842531204f;
  t62_n45: return 0.05794771760702133f;
  t62_n46: return 0.0038683153688907623f;
  t62_n47: return -0.007405574433505535f;
  t62_n48: return -0.06066427752375603f;
  t62_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n55;
  t62_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t62_n52;
  t62_n51: return -0.07087605446577072f;
  t62_n52: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t62_n54;
  t62_n53: return 0.11044403165578842f;
  t62_n54: return -0.05268391966819763f;
  t62_n55: return 0.003321306314319372f;
  t62_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t62_n60;
  t62_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t62_n59;
  t62_n58: return -0.21813078224658966f;
  t62_n59: return 0.010257530957460403f;
  t62_n60: return 0.03554157167673111f;
  t62_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t62_n65;
  t62_n62: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t62_n64;
  t62_n63: return 0.03212911635637283f;
  t62_n64: return 0.0016188351437449455f;
  t62_n65: return 0.013181084766983986f;
  t62_n66: return -0.06209050118923187f;
  t62_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t62_n81;
  t62_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t62_n80;
  t62_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t62_n73;
  t62_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n72;
  t62_n71: return -0.13241954147815704f;
  t62_n72: return -0.03710594400763512f;
  t62_n73: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t62_n79;
  t62_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t62_n76;
  t62_n75: return 0.007051027845591307f;
  t62_n76: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t62_n78;
  t62_n77: return 0.055987659841775894f;
  t62_n78: return 0.007825520820915699f;
  t62_n79: return 0.004056458827108145f;
  t62_n80: return 0.016412384808063507f;
  t62_n81: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t62_n93;
  t62_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n88;
  t62_n83: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t62_n85;
  t62_n84: return 0.008135441690683365f;
  t62_n85: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t62_n87;
  t62_n86: return -8.707261440576985e-05f;
  t62_n87: return -0.04987650364637375f;
  t62_n88: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t62_n90;
  t62_n89: return 0.042083073407411575f;
  t62_n90: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t62_n92;
  t62_n91: return 0.013709502294659615f;
  t62_n92: return -0.010601761750876904f;
  t62_n93: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t62_n95;
  t62_n94: return 0.06286618113517761f;
  t62_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n97;
  t62_n96: return -0.05363034829497337f;
  t62_n97: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t62_n99;
  t62_n98: return 0.07646258175373077f;
  t62_n99: if (E.getFractionNameInContext() >= 3174706828 /*0.04545454680919647*/) goto t62_n101;
  t62_n100: return 0.12330633401870728f;
  t62_n101: return 0.12698349356651306f;
  t62_n102: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t62_n116;
  t62_n103: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n115;
  t62_n104: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t62_n114;
  t62_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t62_n109;
  t62_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t62_n108;
  t62_n107: return 0.018161825835704803f;
  t62_n108: return 0.11847671121358871f;
  t62_n109: if (E.getNumReferences() >= 3339043712 /*34243.5*/) goto t62_n113;
  t62_n110: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t62_n112;
  t62_n111: return 0.023205429315567017f;
  t62_n112: return -0.2671273946762085f;
  t62_n113: return 0.06037047505378723f;
  t62_n114: return -0.00019587029237300158f;
  t62_n115: return 0.048603348433971405f;
  t62_n116: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t62_n118;
  t62_n117: return 0.050979118794202805f;
  t62_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t62_n154;
  t62_n119: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t62_n139;
  t62_n120: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t62_n122;
  t62_n121: return 0.04571026191115379f;
  t62_n122: if (E.getNumReferences() >= 3350798592 /*94786.0*/) goto t62_n126;
  t62_n123: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n125;
  t62_n124: return -0.0019068651599809527f;
  t62_n125: return -0.06181695684790611f;
  t62_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n128;
  t62_n127: return -0.026684043928980827f;
  t62_n128: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t62_n130;
  t62_n129: return 0.027493707835674286f;
  t62_n130: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n138;
  t62_n131: if (E.getNumReferences() >= 3356586400 /*148934.5*/) goto t62_n137;
  t62_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n134;
  t62_n133: return -0.14598602056503296f;
  t62_n134: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t62_n136;
  t62_n135: return 0.13507506251335144f;
  t62_n136: return -0.06434082984924316f;
  t62_n137: return 0.0474570170044899f;
  t62_n138: return -0.017008939757943153f;
  t62_n139: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t62_n143;
  t62_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t62_n142;
  t62_n141: return 0.12069874256849289f;
  t62_n142: return 0.18509477376937866f;
  t62_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t62_n145;
  t62_n144: return 0.0024732269812375307f;
  t62_n145: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t62_n153;
  t62_n146: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t62_n150;
  t62_n147: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t62_n149;
  t62_n148: return 0.04250997304916382f;
  t62_n149: return 0.004422527272254229f;
  t62_n150: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t62_n152;
  t62_n151: return 0.1680920124053955f;
  t62_n152: return 0.12336883693933487f;
  t62_n153: return -0.07301373779773712f;
  t62_n154: return 0.030407723039388657f;
  t62_n155: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t62_n171;
  t62_n156: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t62_n162;
  t62_n157: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t62_n159;
  t62_n158: return 0.01430212240666151f;
  t62_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t62_n161;
  t62_n160: return 0.015008090063929558f;
  t62_n161: return 0.049892719835042953f;
  t62_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t62_n164;
  t62_n163: return -0.0025395285338163376f;
  t62_n164: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t62_n168;
  t62_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t62_n167;
  t62_n166: return -0.02348782867193222f;
  t62_n167: return 0.010718714445829391f;
  t62_n168: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n170;
  t62_n169: return 0.017092561349272728f;
  t62_n170: return 0.04932120814919472f;
  t62_n171: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t62_n203;
  t62_n172: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t62_n202;
  t62_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t62_n195;
  t62_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t62_n194;
  t62_n175: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t62_n179;
  t62_n176: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n178;
  t62_n177: return 0.01576550304889679f;
  t62_n178: return 0.04132842272520065f;
  t62_n179: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n181;
  t62_n180: return -0.002752893604338169f;
  t62_n181: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t62_n191;
  t62_n182: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n190;
  t62_n183: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t62_n185;
  t62_n184: return 0.002516498789191246f;
  t62_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t62_n187;
  t62_n186: return 0.0028835050761699677f;
  t62_n187: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t62_n189;
  t62_n188: return 0.1245587021112442f;
  t62_n189: return 0.07148976624011993f;
  t62_n190: return 0.05374142900109291f;
  t62_n191: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t62_n193;
  t62_n192: return 0.05093694478273392f;
  t62_n193: return -0.0003040516749024391f;
  t62_n194: return 0.03443927690386772f;
  t62_n195: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t62_n201;
  t62_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t62_n200;
  t62_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t62_n199;
  t62_n198: return -0.09674172848463058f;
  t62_n199: return 0.013458352535963058f;
  t62_n200: return 0.03909771516919136f;
  t62_n201: return 0.0431709922850132f;
  t62_n202: return 0.057689227163791656f;
  t62_n203: return 0.04360782727599144f;
  t62_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t62_n246;
  t62_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t62_n227;
  t62_n206: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t62_n210;
  t62_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n209;
  t62_n208: return -0.13369305431842804f;
  t62_n209: return -0.034680552780628204f;
  t62_n210: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t62_n214;
  t62_n211: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t62_n213;
  t62_n212: return -0.04340199753642082f;
  t62_n213: return 0.037665825337171555f;
  t62_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t62_n226;
  t62_n215: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t62_n225;
  t62_n216: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t62_n220;
  t62_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n219;
  t62_n218: return -0.11976847797632217f;
  t62_n219: return -0.010040149092674255f;
  t62_n220: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t62_n222;
  t62_n221: return -0.01814989186823368f;
  t62_n222: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t62_n224;
  t62_n223: return 0.0459587387740612f;
  t62_n224: return 0.0988524779677391f;
  t62_n225: return -0.15000559389591217f;
  t62_n226: return 0.03345167264342308f;
  t62_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n231;
  t62_n228: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t62_n230;
  t62_n229: return -0.12358443439006805f;
  t62_n230: return 0.014727617613971233f;
  t62_n231: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t62_n237;
  t62_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t62_n236;
  t62_n233: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t62_n235;
  t62_n234: return -0.02645217254757881f;
  t62_n235: return 0.005861897487193346f;
  t62_n236: return 0.02513471618294716f;
  t62_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n241;
  t62_n238: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t62_n240;
  t62_n239: return 0.012180114164948463f;
  t62_n240: return 0.029734313488006592f;
  t62_n241: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t62_n243;
  t62_n242: return 0.04660370200872421f;
  t62_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n245;
  t62_n244: return -0.013645214028656483f;
  t62_n245: return 0.08368158340454102f;
  t62_n246: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t62_n250;
  t62_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t62_n249;
  t62_n248: return -0.1023520976305008f;
  t62_n249: return -0.08056818693876266f;
  t62_n250: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t62_n252;
  t62_n251: return -0.10688316822052002f;
  t62_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t62_n254;
  t62_n253: return -0.03156250715255737f;
  t62_n254: return 0.03237351030111313f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree63(const Example& E) {
  t63_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t63_n108;
  t63_n1: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t63_n37;
  t63_n2: if (E.getNumReferences() >= 3277553664 /*219.5*/) goto t63_n18;
  t63_n3: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t63_n17;
  t63_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n6;
  t63_n5: return -0.0038710832595825195f;
  t63_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t63_n16;
  t63_n7: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t63_n9;
  t63_n8: return -0.003583386540412903f;
  t63_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t63_n13;
  t63_n10: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t63_n12;
  t63_n11: return -0.042077500373125076f;
  t63_n12: return 0.02038237638771534f;
  t63_n13: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t63_n15;
  t63_n14: return 0.09366658329963684f;
  t63_n15: return 0.03511970862746239f;
  t63_n16: return 0.007197351660579443f;
  t63_n17: return 0.016673343256115913f;
  t63_n18: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t63_n36;
  t63_n19: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t63_n35;
  t63_n20: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t63_n30;
  t63_n21: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t63_n29;
  t63_n22: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t63_n24;
  t63_n23: return 0.018052857369184494f;
  t63_n24: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t63_n28;
  t63_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n27;
  t63_n26: return 0.05138392001390457f;
  t63_n27: return 0.09508539736270905f;
  t63_n28: return 0.11948675662279129f;
  t63_n29: return -0.0792083591222763f;
  t63_n30: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n34;
  t63_n31: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t63_n33;
  t63_n32: return 0.1029396504163742f;
  t63_n33: return 0.006874884478747845f;
  t63_n34: return -0.22045603394508362f;
  t63_n35: return -0.05729000270366669f;
  t63_n36: return 0.036840926855802536f;
  t63_n37: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t63_n45;
  t63_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n42;
  t63_n39: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t63_n41;
  t63_n40: return -0.0021293330937623978f;
  t63_n41: return 0.036659616976976395f;
  t63_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t63_n44;
  t63_n43: return -0.06199798360466957f;
  t63_n44: return 0.02477280981838703f;
  t63_n45: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t63_n107;
  t63_n46: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t63_n106;
  t63_n47: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t63_n105;
  t63_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t63_n72;
  t63_n49: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t63_n55;
  t63_n50: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t63_n54;
  t63_n51: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t63_n53;
  t63_n52: return -0.022272970527410507f;
  t63_n53: return 0.031042736023664474f;
  t63_n54: return -0.061278048902750015f;
  t63_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t63_n69;
  t63_n56: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t63_n58;
  t63_n57: return -0.026610445231199265f;
  t63_n58: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t63_n66;
  t63_n59: if (E.getNumReferences() >= 3363138304 /*251308.0*/) goto t63_n65;
  t63_n60: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t63_n62;
  t63_n61: return -0.004722291138023138f;
  t63_n62: if (E.getNumReferences() >= 3361098688 /*219439.0*/) goto t63_n64;
  t63_n63: return 0.098707415163517f;
  t63_n64: return 0.0013906712410971522f;
  t63_n65: return -0.19359195232391357f;
  t63_n66: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t63_n68;
  t63_n67: return 0.0931207537651062f;
  t63_n68: return 0.023722810670733452f;
  t63_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t63_n71;
  t63_n70: return -0.0681920200586319f;
  t63_n71: return 0.011067919433116913f;
  t63_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t63_n76;
  t63_n73: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n75;
  t63_n74: return -0.02673337794840336f;
  t63_n75: return 0.005856631323695183f;
  t63_n76: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t63_n100;
  t63_n77: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t63_n95;
  t63_n78: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t63_n92;
  t63_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t63_n87;
  t63_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n82;
  t63_n81: return -0.029958106577396393f;
  t63_n82: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t63_n84;
  t63_n83: return 0.02575511857867241f;
  t63_n84: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t63_n86;
  t63_n85: return -0.013085567392408848f;
  t63_n86: return 0.10117487609386444f;
  t63_n87: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t63_n91;
  t63_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t63_n90;
  t63_n89: return 0.021605174988508224f;
  t63_n90: return 0.07171360403299332f;
  t63_n91: return 0.02715340629220009f;
  t63_n92: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t63_n94;
  t63_n93: return 0.012955508194863796f;
  t63_n94: return -1.202111005783081f;
  t63_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t63_n97;
  t63_n96: return -0.030643971636891365f;
  t63_n97: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t63_n99;
  t63_n98: return 0.08307991921901703f;
  t63_n99: return -0.20648279786109924f;
  t63_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n102;
  t63_n101: return -0.03982770815491676f;
  t63_n102: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t63_n104;
  t63_n103: return 0.08412431925535202f;
  t63_n104: return 0.12089831382036209f;
  t63_n105: return 0.005360722541809082f;
  t63_n106: return 0.032017387449741364f;
  t63_n107: return 0.016838645562529564f;
  t63_n108: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t63_n212;
  t63_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t63_n137;
  t63_n110: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t63_n134;
  t63_n111: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t63_n133;
  t63_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t63_n128;
  t63_n113: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t63_n127;
  t63_n114: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t63_n126;
  t63_n115: if (E.getNumReferences() >= 3354956672 /*127271.0*/) goto t63_n125;
  t63_n116: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t63_n124;
  t63_n117: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t63_n123;
  t63_n118: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t63_n122;
  t63_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t63_n121;
  t63_n120: return -0.0011192617239430547f;
  t63_n121: return 0.0105823939666152f;
  t63_n122: return 0.0230843685567379f;
  t63_n123: return -0.0019179219380021095f;
  t63_n124: return 0.08877373486757278f;
  t63_n125: return -0.08487830311059952f;
  t63_n126: return 0.06907937675714493f;
  t63_n127: return -0.07446784526109695f;
  t63_n128: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n132;
  t63_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t63_n131;
  t63_n130: return -0.0035717375576496124f;
  t63_n131: return 0.010089797899127007f;
  t63_n132: return 0.026817260310053825f;
  t63_n133: return 0.07706760615110397f;
  t63_n134: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t63_n136;
  t63_n135: return 0.022442752495408058f;
  t63_n136: return -0.2125382125377655f;
  t63_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t63_n141;
  t63_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n140;
  t63_n139: return -0.13991934061050415f;
  t63_n140: return 0.006093696691095829f;
  t63_n141: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t63_n211;
  t63_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t63_n174;
  t63_n143: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t63_n163;
  t63_n144: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t63_n158;
  t63_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t63_n157;
  t63_n146: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t63_n148;
  t63_n147: return -0.013035263866186142f;
  t63_n148: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t63_n150;
  t63_n149: return 0.024119161069393158f;
  t63_n150: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t63_n152;
  t63_n151: return -0.050047166645526886f;
  t63_n152: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t63_n156;
  t63_n153: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t63_n155;
  t63_n154: return -0.09691590815782547f;
  t63_n155: return 0.11090321838855743f;
  t63_n156: return -0.02415419928729534f;
  t63_n157: return 0.05133965611457825f;
  t63_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n160;
  t63_n159: return -0.009689218364655972f;
  t63_n160: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t63_n162;
  t63_n161: return 0.002263030968606472f;
  t63_n162: return 0.03354231268167496f;
  t63_n163: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t63_n171;
  t63_n164: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t63_n166;
  t63_n165: return 0.03212624043226242f;
  t63_n166: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t63_n168;
  t63_n167: return 0.0790838897228241f;
  t63_n168: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t63_n170;
  t63_n169: return 0.1528455913066864f;
  t63_n170: return 0.0009494354599155486f;
  t63_n171: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n173;
  t63_n172: return -0.0515170656144619f;
  t63_n173: return 0.05012550577521324f;
  t63_n174: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t63_n208;
  t63_n175: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t63_n193;
  t63_n176: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t63_n182;
  t63_n177: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t63_n179;
  t63_n178: return 0.14421014487743378f;
  t63_n179: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t63_n181;
  t63_n180: return -0.007513819728046656f;
  t63_n181: return 0.053574852645397186f;
  t63_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n184;
  t63_n183: return 0.019542088732123375f;
  t63_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t63_n188;
  t63_n185: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t63_n187;
  t63_n186: return -0.015673860907554626f;
  t63_n187: return 0.01925099827349186f;
  t63_n188: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t63_n190;
  t63_n189: return 0.03419611603021622f;
  t63_n190: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t63_n192;
  t63_n191: return -0.10535126179456711f;
  t63_n192: return 0.1488601118326187f;
  t63_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n195;
  t63_n194: return -0.09666159749031067f;
  t63_n195: if (E.getFractionNameInContext() >= 3202467612 /*0.4409722089767456*/) goto t63_n197;
  t63_n196: return 0.00693721417337656f;
  t63_n197: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t63_n207;
  t63_n198: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t63_n200;
  t63_n199: return 0.035129401832818985f;
  t63_n200: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t63_n206;
  t63_n201: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t63_n205;
  t63_n202: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t63_n204;
  t63_n203: return -0.034959692507982254f;
  t63_n204: return 0.13191694021224976f;
  t63_n205: return 0.036056723445653915f;
  t63_n206: return 0.024755820631980896f;
  t63_n207: return 0.014731482602655888f;
  t63_n208: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t63_n210;
  t63_n209: return 0.027867961674928665f;
  t63_n210: return -0.3131951689720154f;
  t63_n211: return 0.03095298260450363f;
  t63_n212: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t63_n252;
  t63_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t63_n243;
  t63_n214: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t63_n232;
  t63_n215: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t63_n219;
  t63_n216: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t63_n218;
  t63_n217: return 0.010440630838274956f;
  t63_n218: return 0.04480677470564842f;
  t63_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t63_n225;
  t63_n220: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t63_n222;
  t63_n221: return 0.03212253376841545f;
  t63_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t63_n224;
  t63_n223: return 0.11698681116104126f;
  t63_n224: return 0.13315151631832123f;
  t63_n225: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t63_n227;
  t63_n226: return -0.10706824064254761f;
  t63_n227: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t63_n231;
  t63_n228: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t63_n230;
  t63_n229: return 0.11398869752883911f;
  t63_n230: return 0.13217797875404358f;
  t63_n231: return 0.026307646185159683f;
  t63_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n234;
  t63_n233: return -0.023896707221865654f;
  t63_n234: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t63_n236;
  t63_n235: return -0.008879323489964008f;
  t63_n236: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t63_n238;
  t63_n237: return 0.12922966480255127f;
  t63_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t63_n240;
  t63_n239: return 0.021321117877960205f;
  t63_n240: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t63_n242;
  t63_n241: return 0.039436157792806625f;
  t63_n242: return 0.11157327145338058f;
  t63_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t63_n245;
  t63_n244: return 0.006129433400928974f;
  t63_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t63_n249;
  t63_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t63_n248;
  t63_n247: return -0.06823897361755371f;
  t63_n248: return 0.05232113599777222f;
  t63_n249: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t63_n251;
  t63_n250: return 0.03558103367686272f;
  t63_n251: return 0.0972369834780693f;
  t63_n252: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t63_n254;
  t63_n253: return 0.1194356307387352f;
  t63_n254: return 0.05944008380174637f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree64(const Example& E) {
  t64_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t64_n124;
  t64_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t64_n41;
  t64_n2: if (E.getNumReferences() >= 3271720960 /*130.5*/) goto t64_n26;
  t64_n3: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t64_n21;
  t64_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n20;
  t64_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t64_n7;
  t64_n6: return -0.10259103775024414f;
  t64_n7: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t64_n17;
  t64_n8: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t64_n16;
  t64_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t64_n13;
  t64_n10: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t64_n12;
  t64_n11: return 0.020411254838109016f;
  t64_n12: return -0.0017978358082473278f;
  t64_n13: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t64_n15;
  t64_n14: return 0.047174010425806046f;
  t64_n15: return 0.11106449365615845f;
  t64_n16: return -0.02582438476383686f;
  t64_n17: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t64_n19;
  t64_n18: return 0.01251264289021492f;
  t64_n19: return 0.10748737305402756f;
  t64_n20: return 0.0004297455889172852f;
  t64_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t64_n25;
  t64_n22: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t64_n24;
  t64_n23: return 0.012509100139141083f;
  t64_n24: return -0.08490655571222305f;
  t64_n25: return 0.07099331170320511f;
  t64_n26: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t64_n40;
  t64_n27: if (E.getNumReferences() >= 3382768696 /*1318535.0*/) goto t64_n39;
  t64_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t64_n32;
  t64_n29: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t64_n31;
  t64_n30: return 0.018201172351837158f;
  t64_n31: return -0.019387833774089813f;
  t64_n32: if (E.getNumReferences() >= 3354076032 /*120391.0*/) goto t64_n34;
  t64_n33: return 0.032555364072322845f;
  t64_n34: if (E.getNumReferences() >= 3354764736 /*125771.5*/) goto t64_n38;
  t64_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n37;
  t64_n36: return -0.12310326844453812f;
  t64_n37: return 0.13276895880699158f;
  t64_n38: return -0.026484476402401924f;
  t64_n39: return -0.4588862657546997f;
  t64_n40: return 0.03419599309563637f;
  t64_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t64_n83;
  t64_n42: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t64_n46;
  t64_n43: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t64_n45;
  t64_n44: return -0.0417940728366375f;
  t64_n45: return 0.04103514924645424f;
  t64_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t64_n70;
  t64_n47: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t64_n69;
  t64_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t64_n50;
  t64_n49: return -0.03288566693663597f;
  t64_n50: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t64_n54;
  t64_n51: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t64_n53;
  t64_n52: return 0.03602071478962898f;
  t64_n53: return -0.023223668336868286f;
  t64_n54: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t64_n66;
  t64_n55: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t64_n63;
  t64_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t64_n60;
  t64_n57: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t64_n59;
  t64_n58: return 0.0025919058825820684f;
  t64_n59: return -0.08885039389133453f;
  t64_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t64_n62;
  t64_n61: return -0.11536179482936859f;
  t64_n62: return 0.03183393552899361f;
  t64_n63: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t64_n65;
  t64_n64: return -0.02863086760044098f;
  t64_n65: return 0.1616353839635849f;
  t64_n66: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t64_n68;
  t64_n67: return 0.08858969807624817f;
  t64_n68: return 0.03540997952222824f;
  t64_n69: return 0.012586679309606552f;
  t64_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t64_n78;
  t64_n71: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t64_n77;
  t64_n72: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t64_n74;
  t64_n73: return -0.0025724531151354313f;
  t64_n74: if (E.getNumReferences() >= 3351182400 /*97784.5*/) goto t64_n76;
  t64_n75: return 0.04748992994427681f;
  t64_n76: return -0.09935300797224045f;
  t64_n77: return -0.05542775243520737f;
  t64_n78: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t64_n80;
  t64_n79: return 0.04292575269937515f;
  t64_n80: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t64_n82;
  t64_n81: return 0.08558264374732971f;
  t64_n82: return 0.02587965503334999f;
  t64_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n87;
  t64_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t64_n86;
  t64_n85: return -0.0653892382979393f;
  t64_n86: return 0.002354951109737158f;
  t64_n87: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t64_n115;
  t64_n88: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t64_n114;
  t64_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n97;
  t64_n90: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t64_n94;
  t64_n91: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t64_n93;
  t64_n92: return 0.0358036607503891f;
  t64_n93: return -0.05193039029836655f;
  t64_n94: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t64_n96;
  t64_n95: return 0.04539550468325615f;
  t64_n96: return -0.11761781573295593f;
  t64_n97: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t64_n107;
  t64_n98: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t64_n106;
  t64_n99: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t64_n105;
  t64_n100: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t64_n104;
  t64_n101: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t64_n103;
  t64_n102: return 0.03635337948799133f;
  t64_n103: return 0.0018274649046361446f;
  t64_n104: return 0.047498419880867004f;
  t64_n105: return 0.0040156044997274876f;
  t64_n106: return 0.03916854038834572f;
  t64_n107: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t64_n113;
  t64_n108: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t64_n112;
  t64_n109: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t64_n111;
  t64_n110: return 0.0062378463335335255f;
  t64_n111: return 0.03406664729118347f;
  t64_n112: return -0.00497979624196887f;
  t64_n113: return 0.038940537720918655f;
  t64_n114: return -0.010200460441410542f;
  t64_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n117;
  t64_n116: return -0.0006706796702928841f;
  t64_n117: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t64_n123;
  t64_n118: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t64_n120;
  t64_n119: return 0.10516437888145447f;
  t64_n120: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t64_n122;
  t64_n121: return 0.05047669634222984f;
  t64_n122: return -0.04690547287464142f;
  t64_n123: return 0.04563390091061592f;
  t64_n124: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t64_n222;
  t64_n125: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t64_n177;
  t64_n126: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t64_n160;
  t64_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t64_n131;
  t64_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n130;
  t64_n129: return -0.09342163056135178f;
  t64_n130: return -0.006523359566926956f;
  t64_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t64_n157;
  t64_n132: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t64_n156;
  t64_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t64_n143;
  t64_n134: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t64_n136;
  t64_n135: return 0.0019239264074712992f;
  t64_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t64_n140;
  t64_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t64_n139;
  t64_n138: return 0.014543615281581879f;
  t64_n139: return -0.014216596260666847f;
  t64_n140: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t64_n142;
  t64_n141: return 0.015198469161987305f;
  t64_n142: return 0.04817112535238266f;
  t64_n143: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t64_n145;
  t64_n144: return 0.020467065274715424f;
  t64_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t64_n147;
  t64_n146: return 0.05542948096990585f;
  t64_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n149;
  t64_n148: return -0.04620324820280075f;
  t64_n149: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t64_n155;
  t64_n150: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t64_n154;
  t64_n151: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t64_n153;
  t64_n152: return 0.1325548142194748f;
  t64_n153: return -0.06384696066379547f;
  t64_n154: return 0.11973614990711212f;
  t64_n155: return 0.046642303466796875f;
  t64_n156: return -0.014511502347886562f;
  t64_n157: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t64_n159;
  t64_n158: return 0.02117861621081829f;
  t64_n159: return 0.007307675201445818f;
  t64_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t64_n162;
  t64_n161: return -0.014658037573099136f;
  t64_n162: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t64_n174;
  t64_n163: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t64_n167;
  t64_n164: if (E.getFractionNameInContext() >= 3202118087 /*0.4305555522441864*/) goto t64_n166;
  t64_n165: return 0.01320596132427454f;
  t64_n166: return 0.04259919747710228f;
  t64_n167: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t64_n171;
  t64_n168: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t64_n170;
  t64_n169: return -0.055701449513435364f;
  t64_n170: return 0.12891879677772522f;
  t64_n171: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t64_n173;
  t64_n172: return 0.10235652327537537f;
  t64_n173: return 0.1443425565958023f;
  t64_n174: if (E.getNumReferences() >= 3261726720 /*58.5*/) goto t64_n176;
  t64_n175: return 0.1209552064538002f;
  t64_n176: return 0.09281805157661438f;
  t64_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t64_n179;
  t64_n178: return 0.0024616767186671495f;
  t64_n179: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t64_n197;
  t64_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t64_n196;
  t64_n181: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t64_n191;
  t64_n182: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t64_n184;
  t64_n183: return 0.031496912240982056f;
  t64_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t64_n190;
  t64_n185: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t64_n189;
  t64_n186: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t64_n188;
  t64_n187: return -0.0506594181060791f;
  t64_n188: return 0.0638909712433815f;
  t64_n189: return 0.03105495497584343f;
  t64_n190: return 0.07927114516496658f;
  t64_n191: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t64_n193;
  t64_n192: return 0.21168190240859985f;
  t64_n193: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t64_n195;
  t64_n194: return -0.18622885644435883f;
  t64_n195: return 0.09673745930194855f;
  t64_n196: return 0.022033268585801125f;
  t64_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t64_n215;
  t64_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t64_n200;
  t64_n199: return -0.003434103215113282f;
  t64_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t64_n204;
  t64_n201: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t64_n203;
  t64_n202: return 0.0011964760487899184f;
  t64_n203: return 0.06574603170156479f;
  t64_n204: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t64_n208;
  t64_n205: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t64_n207;
  t64_n206: return 0.020198684185743332f;
  t64_n207: return -0.08498905599117279f;
  t64_n208: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t64_n210;
  t64_n209: return 0.00032069714507088065f;
  t64_n210: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t64_n214;
  t64_n211: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t64_n213;
  t64_n212: return 0.03508324921131134f;
  t64_n213: return 0.06883756071329117f;
  t64_n214: return 0.008124840445816517f;
  t64_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t64_n221;
  t64_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t64_n218;
  t64_n217: return -0.002449724590405822f;
  t64_n218: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t64_n220;
  t64_n219: return 0.09195660799741745f;
  t64_n220: return 0.05592609941959381f;
  t64_n221: return 0.02465202473104f;
  t64_n222: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t64_n252;
  t64_n223: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t64_n239;
  t64_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t64_n232;
  t64_n225: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t64_n231;
  t64_n226: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t64_n228;
  t64_n227: return 0.025568567216396332f;
  t64_n228: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t64_n230;
  t64_n229: return 0.13968174159526825f;
  t64_n230: return 0.1051780954003334f;
  t64_n231: return -0.014807409606873989f;
  t64_n232: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t64_n234;
  t64_n233: return 0.013398370705544949f;
  t64_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t64_n236;
  t64_n235: return 0.010174103081226349f;
  t64_n236: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t64_n238;
  t64_n237: return 0.05836672708392143f;
  t64_n238: return 0.11159086227416992f;
  t64_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t64_n245;
  t64_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n244;
  t64_n241: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t64_n243;
  t64_n242: return 0.00204429286532104f;
  t64_n243: return -0.06054336577653885f;
  t64_n244: return 0.019815636798739433f;
  t64_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t64_n249;
  t64_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t64_n248;
  t64_n247: return -0.06007838249206543f;
  t64_n248: return 0.0499747097492218f;
  t64_n249: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t64_n251;
  t64_n250: return 0.032984815537929535f;
  t64_n251: return 0.0951598733663559f;
  t64_n252: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t64_n254;
  t64_n253: return 0.03807879611849785f;
  t64_n254: return 0.07783982157707214f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree65(const Example& E) {
  t65_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t65_n108;
  t65_n1: if (E.getNumReferences() >= 3303288832 /*1825.5*/) goto t65_n101;
  t65_n2: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t65_n70;
  t65_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t65_n35;
  t65_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t65_n6;
  t65_n5: return -0.11044760048389435f;
  t65_n6: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t65_n34;
  t65_n7: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t65_n31;
  t65_n8: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t65_n28;
  t65_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n19;
  t65_n10: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t65_n12;
  t65_n11: return -0.019550932571291924f;
  t65_n12: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t65_n14;
  t65_n13: return -0.08117755502462387f;
  t65_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t65_n18;
  t65_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t65_n17;
  t65_n16: return -0.04046831652522087f;
  t65_n17: return 0.022472260519862175f;
  t65_n18: return -0.06541253626346588f;
  t65_n19: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t65_n27;
  t65_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t65_n24;
  t65_n21: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t65_n23;
  t65_n22: return 0.005724205169826746f;
  t65_n23: return -0.11759180575609207f;
  t65_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t65_n26;
  t65_n25: return -0.0657368078827858f;
  t65_n26: return 0.04353662580251694f;
  t65_n27: return -0.085345059633255f;
  t65_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t65_n30;
  t65_n29: return -0.10955038666725159f;
  t65_n30: return -0.022447532042860985f;
  t65_n31: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t65_n33;
  t65_n32: return 0.04297995939850807f;
  t65_n33: return 0.004169804975390434f;
  t65_n34: return -0.10816860944032669f;
  t65_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t65_n37;
  t65_n36: return -0.11185437440872192f;
  t65_n37: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t65_n59;
  t65_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t65_n54;
  t65_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n41;
  t65_n40: return -0.018340392038226128f;
  t65_n41: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t65_n49;
  t65_n42: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t65_n44;
  t65_n43: return -0.011300098150968552f;
  t65_n44: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t65_n46;
  t65_n45: return 0.033955007791519165f;
  t65_n46: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t65_n48;
  t65_n47: return 0.012483504600822926f;
  t65_n48: return 0.030286086723208427f;
  t65_n49: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t65_n51;
  t65_n50: return 0.015024877153337002f;
  t65_n51: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t65_n53;
  t65_n52: return -0.058337513357400894f;
  t65_n53: return -0.0019415990682318807f;
  t65_n54: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t65_n58;
  t65_n55: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t65_n57;
  t65_n56: return 0.04942285269498825f;
  t65_n57: return -0.04477526620030403f;
  t65_n58: return 0.0012367902090772986f;
  t65_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n61;
  t65_n60: return -0.01830599643290043f;
  t65_n61: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t65_n67;
  t65_n62: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t65_n66;
  t65_n63: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t65_n65;
  t65_n64: return 0.01332151424139738f;
  t65_n65: return 0.10159973055124283f;
  t65_n66: return -0.010453733615577221f;
  t65_n67: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t65_n69;
  t65_n68: return 0.036706238985061646f;
  t65_n69: return -0.006184176541864872f;
  t65_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n94;
  t65_n71: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t65_n91;
  t65_n72: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t65_n88;
  t65_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t65_n85;
  t65_n74: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t65_n76;
  t65_n75: return 0.014809241518378258f;
  t65_n76: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t65_n84;
  t65_n77: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t65_n81;
  t65_n78: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t65_n80;
  t65_n79: return -0.0206089336425066f;
  t65_n80: return 0.07247715443372726f;
  t65_n81: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t65_n83;
  t65_n82: return -0.07135161757469177f;
  t65_n83: return 0.02906283736228943f;
  t65_n84: return 0.01632048189640045f;
  t65_n85: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t65_n87;
  t65_n86: return 0.0037009681109339f;
  t65_n87: return -0.10939280688762665f;
  t65_n88: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t65_n90;
  t65_n89: return 0.012883846648037434f;
  t65_n90: return -0.0593155212700367f;
  t65_n91: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t65_n93;
  t65_n92: return -0.0199350006878376f;
  t65_n93: return 0.021541893482208252f;
  t65_n94: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t65_n100;
  t65_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n99;
  t65_n96: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t65_n98;
  t65_n97: return 0.04407354071736336f;
  t65_n98: return 0.005108109209686518f;
  t65_n99: return 0.024014370515942574f;
  t65_n100: return -0.014805309474468231f;
  t65_n101: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t65_n105;
  t65_n102: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t65_n104;
  t65_n103: return 0.02025272138416767f;
  t65_n104: return 0.04525928944349289f;
  t65_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t65_n107;
  t65_n106: return 0.001313940272666514f;
  t65_n107: return 0.031243907287716866f;
  t65_n108: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t65_n232;
  t65_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t65_n155;
  t65_n110: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t65_n154;
  t65_n111: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t65_n153;
  t65_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t65_n148;
  t65_n113: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t65_n147;
  t65_n114: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t65_n146;
  t65_n115: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t65_n145;
  t65_n116: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t65_n140;
  t65_n117: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t65_n123;
  t65_n118: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t65_n122;
  t65_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t65_n121;
  t65_n120: return -0.0011082966811954975f;
  t65_n121: return 0.009300579316914082f;
  t65_n122: return 0.019477667286992073f;
  t65_n123: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t65_n137;
  t65_n124: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t65_n132;
  t65_n125: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t65_n127;
  t65_n126: return -0.000688007683493197f;
  t65_n127: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t65_n129;
  t65_n128: return 0.03356080874800682f;
  t65_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t65_n131;
  t65_n130: return 0.10160478204488754f;
  t65_n131: return 0.11884935945272446f;
  t65_n132: if (E.getNumReferences() >= 3318897664 /*6731.5*/) goto t65_n134;
  t65_n133: return -0.04578372836112976f;
  t65_n134: if (E.getNumReferences() >= 3319636992 /*7092.5*/) goto t65_n136;
  t65_n135: return -0.46331706643104553f;
  t65_n136: return -0.04056865721940994f;
  t65_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t65_n139;
  t65_n138: return 0.0006749511230736971f;
  t65_n139: return 0.022866886109113693f;
  t65_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t65_n142;
  t65_n141: return 0.061968397349119186f;
  t65_n142: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t65_n144;
  t65_n143: return 0.1392991840839386f;
  t65_n144: return 0.053854454308748245f;
  t65_n145: return -0.07392913848161697f;
  t65_n146: return 0.06465093791484833f;
  t65_n147: return -0.06783748418092728f;
  t65_n148: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t65_n152;
  t65_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t65_n151;
  t65_n150: return -0.0033290074206888676f;
  t65_n151: return 0.00925609190016985f;
  t65_n152: return 0.024711906909942627f;
  t65_n153: return 0.07705967128276825f;
  t65_n154: return 0.020959777757525444f;
  t65_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t65_n161;
  t65_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n160;
  t65_n157: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t65_n159;
  t65_n158: return -0.0035449874121695757f;
  t65_n159: return -0.07666249573230743f;
  t65_n160: return 0.009748408570885658f;
  t65_n161: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t65_n231;
  t65_n162: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t65_n230;
  t65_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t65_n189;
  t65_n164: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t65_n176;
  t65_n165: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t65_n173;
  t65_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t65_n172;
  t65_n167: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t65_n169;
  t65_n168: return -0.011730240657925606f;
  t65_n169: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t65_n171;
  t65_n170: return 0.020258933305740356f;
  t65_n171: return -0.04106384143233299f;
  t65_n172: return 0.046984802931547165f;
  t65_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n175;
  t65_n174: return -0.006759664975106716f;
  t65_n175: return 0.02225402183830738f;
  t65_n176: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t65_n184;
  t65_n177: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t65_n179;
  t65_n178: return 0.027716418728232384f;
  t65_n179: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t65_n181;
  t65_n180: return 0.0750664472579956f;
  t65_n181: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t65_n183;
  t65_n182: return 0.1385258287191391f;
  t65_n183: return -0.024843955412507057f;
  t65_n184: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t65_n188;
  t65_n185: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t65_n187;
  t65_n186: return -0.08814743906259537f;
  t65_n187: return -0.006230470724403858f;
  t65_n188: return 0.05160613730549812f;
  t65_n189: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t65_n229;
  t65_n190: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t65_n206;
  t65_n191: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t65_n197;
  t65_n192: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t65_n194;
  t65_n193: return 0.13910794258117676f;
  t65_n194: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t65_n196;
  t65_n195: return -0.0044740475714206696f;
  t65_n196: return 0.046760305762290955f;
  t65_n197: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t65_n199;
  t65_n198: return 0.018435880541801453f;
  t65_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t65_n201;
  t65_n200: return 0.002415245398879051f;
  t65_n201: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t65_n203;
  t65_n202: return 0.03205668926239014f;
  t65_n203: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t65_n205;
  t65_n204: return -0.09464501589536667f;
  t65_n205: return 0.14089737832546234f;
  t65_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t65_n208;
  t65_n207: return -0.08434988558292389f;
  t65_n208: if (E.getFractionNameInContext() >= 3202467612 /*0.4409722089767456*/) goto t65_n210;
  t65_n209: return 0.007629083003848791f;
  t65_n210: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t65_n212;
  t65_n211: return 0.008076044730842113f;
  t65_n212: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t65_n214;
  t65_n213: return 0.04588470235466957f;
  t65_n214: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t65_n228;
  t65_n215: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t65_n227;
  t65_n216: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t65_n226;
  t65_n217: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t65_n225;
  t65_n218: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t65_n224;
  t65_n219: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t65_n221;
  t65_n220: return 0.1226639598608017f;
  t65_n221: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t65_n223;
  t65_n222: return 0.0774165689945221f;
  t65_n223: return 0.17464402318000793f;
  t65_n224: return 0.0019929443951696157f;
  t65_n225: return 0.12325357645750046f;
  t65_n226: return 0.07173657417297363f;
  t65_n227: return -0.1365300863981247f;
  t65_n228: return 0.003690059296786785f;
  t65_n229: return 0.025323254987597466f;
  t65_n230: return -0.1457304209470749f;
  t65_n231: return 0.03097316063940525f;
  t65_n232: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t65_n254;
  t65_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t65_n253;
  t65_n234: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t65_n238;
  t65_n235: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t65_n237;
  t65_n236: return 0.03858266770839691f;
  t65_n237: return 0.00955228041857481f;
  t65_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t65_n242;
  t65_n239: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t65_n241;
  t65_n240: return 0.00250087957829237f;
  t65_n241: return 0.05471723899245262f;
  t65_n242: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t65_n248;
  t65_n243: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t65_n245;
  t65_n244: return -0.02081209234893322f;
  t65_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t65_n247;
  t65_n246: return 0.016043605282902718f;
  t65_n247: return 0.0562322624027729f;
  t65_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t65_n252;
  t65_n249: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t65_n251;
  t65_n250: return 0.019064778462052345f;
  t65_n251: return 0.07070837914943695f;
  t65_n252: return -0.2046610563993454f;
  t65_n253: return 0.0403277650475502f;
  t65_n254: return 0.0669485330581665f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree66(const Example& E) {
  t66_n0: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t66_n170;
  t66_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t66_n57;
  t66_n2: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t66_n26;
  t66_n3: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t66_n23;
  t66_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n22;
  t66_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t66_n19;
  t66_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t66_n8;
  t66_n7: return -0.09009432792663574f;
  t66_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t66_n10;
  t66_n9: return -0.008105181157588959f;
  t66_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t66_n14;
  t66_n11: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t66_n13;
  t66_n12: return 0.010494804941117764f;
  t66_n13: return -0.13479238748550415f;
  t66_n14: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t66_n18;
  t66_n15: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t66_n17;
  t66_n16: return 0.013010665774345398f;
  t66_n17: return -0.18700574338436127f;
  t66_n18: return 0.10888850688934326f;
  t66_n19: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t66_n21;
  t66_n20: return 0.011810597032308578f;
  t66_n21: return 0.10531524568796158f;
  t66_n22: return 0.0009485561749897897f;
  t66_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n25;
  t66_n24: return -0.0078004891984164715f;
  t66_n25: return 0.022736577317118645f;
  t66_n26: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t66_n56;
  t66_n27: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t66_n53;
  t66_n28: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t66_n34;
  t66_n29: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t66_n33;
  t66_n30: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t66_n32;
  t66_n31: return 0.02025422267615795f;
  t66_n32: return 0.11395049095153809f;
  t66_n33: return -0.12229518592357635f;
  t66_n34: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t66_n36;
  t66_n35: return 0.0979456827044487f;
  t66_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n38;
  t66_n37: return -0.07660795748233795f;
  t66_n38: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t66_n40;
  t66_n39: return -0.028784077614545822f;
  t66_n40: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t66_n50;
  t66_n41: if (E.getNumReferences() >= 3354076032 /*120391.0*/) goto t66_n43;
  t66_n42: return 0.0334947407245636f;
  t66_n43: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t66_n49;
  t66_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n46;
  t66_n45: return -0.1141667515039444f;
  t66_n46: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t66_n48;
  t66_n47: return 0.1322462409734726f;
  t66_n48: return -0.09394924342632294f;
  t66_n49: return 0.060498230159282684f;
  t66_n50: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t66_n52;
  t66_n51: return -0.14767305552959442f;
  t66_n52: return 0.041943054646253586f;
  t66_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t66_n55;
  t66_n54: return -0.044913433492183685f;
  t66_n55: return 0.011078440584242344f;
  t66_n56: return 0.042738176882267f;
  t66_n57: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t66_n59;
  t66_n58: return 0.01769552007317543f;
  t66_n59: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t66_n167;
  t66_n60: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t66_n156;
  t66_n61: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t66_n155;
  t66_n62: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t66_n152;
  t66_n63: if (E.getNumReferences() >= 3357252288 /*159339.0*/) goto t66_n151;
  t66_n64: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t66_n106;
  t66_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t66_n67;
  t66_n66: return -0.04479892551898956f;
  t66_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n71;
  t66_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t66_n70;
  t66_n69: return -0.06079039350152016f;
  t66_n70: return 0.0011287499219179153f;
  t66_n71: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t66_n105;
  t66_n72: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t66_n84;
  t66_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n77;
  t66_n74: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t66_n76;
  t66_n75: return -0.07174067944288254f;
  t66_n76: return 0.0119941933080554f;
  t66_n77: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t66_n83;
  t66_n78: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t66_n82;
  t66_n79: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t66_n81;
  t66_n80: return 0.03233451023697853f;
  t66_n81: return 0.009300119243562222f;
  t66_n82: return 0.036434393376111984f;
  t66_n83: return 0.005981202702969313f;
  t66_n84: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t66_n88;
  t66_n85: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t66_n87;
  t66_n86: return 0.0519048236310482f;
  t66_n87: return -0.029925454407930374f;
  t66_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t66_n90;
  t66_n89: return 0.0005432674661278725f;
  t66_n90: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t66_n104;
  t66_n91: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t66_n103;
  t66_n92: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t66_n102;
  t66_n93: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t66_n101;
  t66_n94: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t66_n100;
  t66_n95: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t66_n99;
  t66_n96: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t66_n98;
  t66_n97: return 0.06176620349287987f;
  t66_n98: return -0.119837187230587f;
  t66_n99: return 0.06424812972545624f;
  t66_n100: return -0.1484692543745041f;
  t66_n101: return 0.06177370995283127f;
  t66_n102: return -0.09611944109201431f;
  t66_n103: return 0.06467700004577637f;
  t66_n104: return -0.11848071217536926f;
  t66_n105: return -0.016912449151277542f;
  t66_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t66_n138;
  t66_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t66_n137;
  t66_n108: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t66_n116;
  t66_n109: if (E.getNumReferences() >= 3286253568 /*448.5*/) goto t66_n111;
  t66_n110: return -0.02460254170000553f;
  t66_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t66_n115;
  t66_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n114;
  t66_n113: return -0.06157669052481651f;
  t66_n114: return 0.006494444794952869f;
  t66_n115: return 0.018615808337926865f;
  t66_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type))) goto t66_n118;
  t66_n117: return -0.012988271191716194f;
  t66_n118: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t66_n136;
  t66_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t66_n135;
  t66_n120: if (E.getNumReferences() >= 3338812928 /*33342.0*/) goto t66_n134;
  t66_n121: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t66_n133;
  t66_n122: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t66_n132;
  t66_n123: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t66_n125;
  t66_n124: return 0.031476955860853195f;
  t66_n125: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t66_n127;
  t66_n126: return -0.07355809956789017f;
  t66_n127: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t66_n129;
  t66_n128: return 0.0390082448720932f;
  t66_n129: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t66_n131;
  t66_n130: return -0.09361021965742111f;
  t66_n131: return 0.025996359065175056f;
  t66_n132: return -0.10947704315185547f;
  t66_n133: return -0.04720829054713249f;
  t66_n134: return 0.06103077903389931f;
  t66_n135: return 0.034858766943216324f;
  t66_n136: return -0.056307993829250336f;
  t66_n137: return 0.010526537895202637f;
  t66_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t66_n144;
  t66_n139: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t66_n143;
  t66_n140: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t66_n142;
  t66_n141: return 0.0003286873979959637f;
  t66_n142: return -0.03751187026500702f;
  t66_n143: return 0.04007554426789284f;
  t66_n144: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t66_n146;
  t66_n145: return 0.040303487330675125f;
  t66_n146: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t66_n148;
  t66_n147: return 0.027822067961096764f;
  t66_n148: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t66_n150;
  t66_n149: return 0.13626264035701752f;
  t66_n150: return -0.20566076040267944f;
  t66_n151: return -0.22856277227401733f;
  t66_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n154;
  t66_n153: return -0.19713540375232697f;
  t66_n154: return 0.09398183971643448f;
  t66_n155: return -0.1444423794746399f;
  t66_n156: if (E.getNumReferences() >= 3360662464 /*212623.0*/) goto t66_n158;
  t66_n157: return 0.06475590914487839f;
  t66_n158: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t66_n166;
  t66_n159: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t66_n165;
  t66_n160: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t66_n162;
  t66_n161: return -0.13818757236003876f;
  t66_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t66_n164;
  t66_n163: return 0.04339785873889923f;
  t66_n164: return 0.1012294590473175f;
  t66_n165: return -0.08540664613246918f;
  t66_n166: return 0.018013272434473038f;
  t66_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n169;
  t66_n168: return -0.05731528624892235f;
  t66_n169: return 0.022246865555644035f;
  t66_n170: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t66_n232;
  t66_n171: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t66_n187;
  t66_n172: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t66_n186;
  t66_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t66_n185;
  t66_n174: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t66_n184;
  t66_n175: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t66_n183;
  t66_n176: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t66_n182;
  t66_n177: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t66_n181;
  t66_n178: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t66_n180;
  t66_n179: return 0.0016890473198145628f;
  t66_n180: return 0.08618050068616867f;
  t66_n181: return -0.06911660730838776f;
  t66_n182: return 0.05948089808225632f;
  t66_n183: return -0.09156626462936401f;
  t66_n184: return 0.06500323116779327f;
  t66_n185: return 0.00886403489857912f;
  t66_n186: return 0.01991737075150013f;
  t66_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t66_n189;
  t66_n188: return 0.004695722367614508f;
  t66_n189: if (E.getNumReferences() >= 3323824640 /*10082.5*/) goto t66_n231;
  t66_n190: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t66_n230;
  t66_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t66_n229;
  t66_n192: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t66_n210;
  t66_n193: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t66_n197;
  t66_n194: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t66_n196;
  t66_n195: return 0.0022733230143785477f;
  t66_n196: return 0.013631048612296581f;
  t66_n197: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t66_n207;
  t66_n198: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t66_n200;
  t66_n199: return 0.029831886291503906f;
  t66_n200: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t66_n206;
  t66_n201: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t66_n205;
  t66_n202: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t66_n204;
  t66_n203: return -0.06502658873796463f;
  t66_n204: return 0.125129833817482f;
  t66_n205: return 0.09499769657850266f;
  t66_n206: return 0.13914854824543f;
  t66_n207: if (E.getNumReferences() >= 3261726720 /*58.5*/) goto t66_n209;
  t66_n208: return 0.11915308237075806f;
  t66_n209: return 0.08483608812093735f;
  t66_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t66_n226;
  t66_n211: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t66_n213;
  t66_n212: return -0.0126465680077672f;
  t66_n213: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t66_n223;
  t66_n214: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t66_n218;
  t66_n215: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t66_n217;
  t66_n216: return 0.03905278816819191f;
  t66_n217: return -0.04632173478603363f;
  t66_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t66_n220;
  t66_n219: return 0.008796686306595802f;
  t66_n220: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t66_n222;
  t66_n221: return 0.05990315601229668f;
  t66_n222: return -0.027594275772571564f;
  t66_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t66_n225;
  t66_n224: return 0.016718652099370956f;
  t66_n225: return 0.05938597023487091f;
  t66_n226: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t66_n228;
  t66_n227: return 0.025309456512331963f;
  t66_n228: return 0.01750040240585804f;
  t66_n229: return 0.00021719647338613868f;
  t66_n230: return -0.13729296624660492f;
  t66_n231: return 0.05263073742389679f;
  t66_n232: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t66_n254;
  t66_n233: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t66_n245;
  t66_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t66_n244;
  t66_n235: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t66_n243;
  t66_n236: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t66_n240;
  t66_n237: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t66_n239;
  t66_n238: return 0.006820891052484512f;
  t66_n239: return 0.04631711170077324f;
  t66_n240: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t66_n242;
  t66_n241: return 0.13683739304542542f;
  t66_n242: return 0.10091196745634079f;
  t66_n243: return -0.015522509813308716f;
  t66_n244: return 0.016607005149126053f;
  t66_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t66_n247;
  t66_n246: return 0.004917940590530634f;
  t66_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t66_n251;
  t66_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t66_n250;
  t66_n249: return -0.060666874051094055f;
  t66_n250: return 0.04614033177495003f;
  t66_n251: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t66_n253;
  t66_n252: return 0.030004555359482765f;
  t66_n253: return 0.06537231057882309f;
  t66_n254: return 0.0434676818549633f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree67(const Example& E) {
  t67_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t67_n212;
  t67_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t67_n187;
  t67_n2: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t67_n144;
  t67_n3: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t67_n27;
  t67_n4: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t67_n16;
  t67_n5: if (E.getSymbolScopeDistanceCost() >= 3250061312 /*23.0*/) goto t67_n11;
  t67_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n8;
  t67_n7: return 0.0014898446388542652f;
  t67_n8: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t67_n10;
  t67_n9: return 0.041829053312540054f;
  t67_n10: return 0.014388681389391422f;
  t67_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t67_n13;
  t67_n12: return 0.01352621614933014f;
  t67_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t67_n15;
  t67_n14: return 0.03848912939429283f;
  t67_n15: return 0.0815599113702774f;
  t67_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t67_n26;
  t67_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t67_n21;
  t67_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n20;
  t67_n19: return -0.008779989555478096f;
  t67_n20: return 0.014501474797725677f;
  t67_n21: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t67_n23;
  t67_n22: return 0.0488976426422596f;
  t67_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n25;
  t67_n24: return -0.01907969079911709f;
  t67_n25: return 0.045772671699523926f;
  t67_n26: return 0.041110701858997345f;
  t67_n27: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t67_n127;
  t67_n28: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t67_n126;
  t67_n29: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t67_n45;
  t67_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t67_n36;
  t67_n31: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t67_n33;
  t67_n32: return -0.04011697694659233f;
  t67_n33: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n35;
  t67_n34: return 0.008272189646959305f;
  t67_n35: return 0.04691476374864578f;
  t67_n36: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t67_n44;
  t67_n37: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t67_n41;
  t67_n38: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t67_n40;
  t67_n39: return -0.0016921875067055225f;
  t67_n40: return -0.09758062660694122f;
  t67_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t67_n43;
  t67_n42: return -0.003758589504286647f;
  t67_n43: return 0.0348626933991909f;
  t67_n44: return -0.07607553154230118f;
  t67_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t67_n97;
  t67_n46: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t67_n50;
  t67_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t67_n49;
  t67_n48: return -0.03105163387954235f;
  t67_n49: return -0.05354788526892662f;
  t67_n50: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t67_n92;
  t67_n51: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t67_n57;
  t67_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t67_n54;
  t67_n53: return -0.002357656368985772f;
  t67_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t67_n56;
  t67_n55: return -0.07177163660526276f;
  t67_n56: return 0.04803824797272682f;
  t67_n57: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t67_n73;
  t67_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t67_n60;
  t67_n59: return -0.12644533812999725f;
  t67_n60: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t67_n66;
  t67_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t67_n63;
  t67_n62: return 0.0299016572535038f;
  t67_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t67_n65;
  t67_n64: return -0.04822525009512901f;
  t67_n65: return 0.0516928993165493f;
  t67_n66: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t67_n68;
  t67_n67: return 0.022085458040237427f;
  t67_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t67_n70;
  t67_n69: return -0.0021904688328504562f;
  t67_n70: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t67_n72;
  t67_n71: return 0.023621493950486183f;
  t67_n72: return 0.0958361029624939f;
  t67_n73: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t67_n91;
  t67_n74: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t67_n84;
  t67_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n79;
  t67_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t67_n78;
  t67_n77: return -0.15531226992607117f;
  t67_n78: return -0.10410244017839432f;
  t67_n79: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t67_n81;
  t67_n80: return 0.0009624421363696456f;
  t67_n81: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t67_n83;
  t67_n82: return 0.009075114503502846f;
  t67_n83: return 0.041748981922864914f;
  t67_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t67_n90;
  t67_n85: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t67_n87;
  t67_n86: return 0.008451576344668865f;
  t67_n87: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n89;
  t67_n88: return -0.11235108971595764f;
  t67_n89: return -0.20880979299545288f;
  t67_n90: return 0.031052321195602417f;
  t67_n91: return 0.039276912808418274f;
  t67_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n94;
  t67_n93: return -0.03451082482933998f;
  t67_n94: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t67_n96;
  t67_n95: return -0.06440457701683044f;
  t67_n96: return 0.012791267596185207f;
  t67_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel))) goto t67_n107;
  t67_n98: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t67_n100;
  t67_n99: return 0.003305082907900214f;
  t67_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t67_n104;
  t67_n101: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t67_n103;
  t67_n102: return 0.03371778875589371f;
  t67_n103: return -0.013453341089189053f;
  t67_n104: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t67_n106;
  t67_n105: return 0.11053574830293655f;
  t67_n106: return 0.06804139167070389f;
  t67_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t67_n113;
  t67_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t67_n110;
  t67_n109: return 0.00851849652826786f;
  t67_n110: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t67_n112;
  t67_n111: return 0.030609816312789917f;
  t67_n112: return 0.10085245221853256f;
  t67_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t67_n115;
  t67_n114: return -0.0062629482708871365f;
  t67_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n117;
  t67_n116: return -0.03813124820590019f;
  t67_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t67_n121;
  t67_n118: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t67_n120;
  t67_n119: return 0.025984404608607292f;
  t67_n120: return 0.00022969962446950376f;
  t67_n121: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t67_n123;
  t67_n122: return 0.03637619689106941f;
  t67_n123: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t67_n125;
  t67_n124: return 0.12906663119792938f;
  t67_n125: return -0.05776900425553322f;
  t67_n126: return -0.05749422311782837f;
  t67_n127: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t67_n131;
  t67_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t67_n130;
  t67_n129: return -0.08775011450052261f;
  t67_n130: return -0.07518445700407028f;
  t67_n131: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t67_n141;
  t67_n132: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n140;
  t67_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t67_n139;
  t67_n134: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t67_n136;
  t67_n135: return 0.0025846934877336025f;
  t67_n136: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t67_n138;
  t67_n137: return 0.10685832053422928f;
  t67_n138: return 0.19044767320156097f;
  t67_n139: return 0.021174969151616096f;
  t67_n140: return 0.044468484818935394f;
  t67_n141: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t67_n143;
  t67_n142: return 0.007061059586703777f;
  t67_n143: return -0.055574558675289154f;
  t67_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t67_n164;
  t67_n145: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n163;
  t67_n146: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t67_n162;
  t67_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t67_n157;
  t67_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t67_n150;
  t67_n149: return 0.017023714259266853f;
  t67_n150: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t67_n152;
  t67_n151: return 0.06596371531486511f;
  t67_n152: if (E.getNumReferences() >= 3299737600 /*1392.0*/) goto t67_n156;
  t67_n153: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t67_n155;
  t67_n154: return 0.18355749547481537f;
  t67_n155: return 0.1438538134098053f;
  t67_n156: return -0.05439665913581848f;
  t67_n157: if (E.getNumReferences() >= 3339043712 /*34243.5*/) goto t67_n161;
  t67_n158: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t67_n160;
  t67_n159: return 0.020706042647361755f;
  t67_n160: return -0.22162650525569916f;
  t67_n161: return 0.056882500648498535f;
  t67_n162: return 0.0005204029148444533f;
  t67_n163: return 0.044552549719810486f;
  t67_n164: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t67_n166;
  t67_n165: return 0.04662083089351654f;
  t67_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t67_n186;
  t67_n167: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t67_n173;
  t67_n168: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t67_n170;
  t67_n169: return 0.03914244845509529f;
  t67_n170: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t67_n172;
  t67_n171: return -0.038581185042858124f;
  t67_n172: return 0.01048929151147604f;
  t67_n173: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t67_n177;
  t67_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t67_n176;
  t67_n175: return 0.11564847826957703f;
  t67_n176: return 0.16105873882770538f;
  t67_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t67_n179;
  t67_n178: return 0.0023054652847349644f;
  t67_n179: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t67_n185;
  t67_n180: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t67_n184;
  t67_n181: if (E.getNumReferences() >= 3315225600 /*4938.5*/) goto t67_n183;
  t67_n182: return 0.052912041544914246f;
  t67_n183: return 0.00610078452154994f;
  t67_n184: return 0.12603884935379028f;
  t67_n185: return -0.06078767776489258f;
  t67_n186: return 0.03009229339659214f;
  t67_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t67_n203;
  t67_n188: if (E.getNumReferences() >= 3298988032 /*1300.5*/) goto t67_n202;
  t67_n189: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t67_n201;
  t67_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t67_n200;
  t67_n191: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t67_n195;
  t67_n192: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n194;
  t67_n193: return 0.012033442966639996f;
  t67_n194: return 0.03892077878117561f;
  t67_n195: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t67_n197;
  t67_n196: return -0.005724189803004265f;
  t67_n197: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t67_n199;
  t67_n198: return 0.00880042091012001f;
  t67_n199: return 0.04574919119477272f;
  t67_n200: return 0.03024807572364807f;
  t67_n201: return 0.03639160469174385f;
  t67_n202: return 0.06818202137947083f;
  t67_n203: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n211;
  t67_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t67_n210;
  t67_n205: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t67_n207;
  t67_n206: return -0.06203550472855568f;
  t67_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t67_n209;
  t67_n208: return -0.07423635572195053f;
  t67_n209: return 0.036790307611227036f;
  t67_n210: return 0.018790144473314285f;
  t67_n211: return 0.04008049517869949f;
  t67_n212: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t67_n250;
  t67_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t67_n231;
  t67_n214: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t67_n220;
  t67_n215: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t67_n219;
  t67_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n218;
  t67_n217: return -0.1290377378463745f;
  t67_n218: return -0.04965889826416969f;
  t67_n219: return 0.004418191034346819f;
  t67_n220: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t67_n226;
  t67_n221: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t67_n223;
  t67_n222: return -0.04166598990559578f;
  t67_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t67_n225;
  t67_n224: return -0.11760041862726212f;
  t67_n225: return 0.03747669234871864f;
  t67_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n230;
  t67_n227: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t67_n229;
  t67_n228: return -0.009049896150827408f;
  t67_n229: return -0.14725276827812195f;
  t67_n230: return 0.031486593186855316f;
  t67_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n235;
  t67_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t67_n234;
  t67_n233: return -0.11612685769796371f;
  t67_n234: return 0.010767867788672447f;
  t67_n235: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t67_n243;
  t67_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t67_n240;
  t67_n237: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t67_n239;
  t67_n238: return -0.025029011070728302f;
  t67_n239: return 0.003959497436881065f;
  t67_n240: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n242;
  t67_n241: return 0.016872631385922432f;
  t67_n242: return 0.0463605560362339f;
  t67_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n245;
  t67_n244: return 0.020236322656273842f;
  t67_n245: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t67_n247;
  t67_n246: return 0.0424962043762207f;
  t67_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t67_n249;
  t67_n248: return -0.014637026935815811f;
  t67_n249: return 0.08283902704715729f;
  t67_n250: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t67_n252;
  t67_n251: return -0.17604584991931915f;
  t67_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t67_n254;
  t67_n253: return -0.017231881618499756f;
  t67_n254: return 0.04105702042579651f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree68(const Example& E) {
  t68_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t68_n68;
  t68_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t68_n53;
  t68_n2: if (E.getNumReferences() >= 3314787328 /*4724.5*/) goto t68_n40;
  t68_n3: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t68_n33;
  t68_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t68_n18;
  t68_n5: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t68_n15;
  t68_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n8;
  t68_n7: return -0.009205595590174198f;
  t68_n8: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t68_n12;
  t68_n9: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t68_n11;
  t68_n10: return 0.012244309298694134f;
  t68_n11: return -9.687299461802468e-05f;
  t68_n12: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t68_n14;
  t68_n13: return -0.07885318249464035f;
  t68_n14: return -0.004089669790118933f;
  t68_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n17;
  t68_n16: return -0.08339560776948929f;
  t68_n17: return 0.010013054125010967f;
  t68_n18: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t68_n28;
  t68_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n21;
  t68_n20: return -0.032235078513622284f;
  t68_n21: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t68_n27;
  t68_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t68_n26;
  t68_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n25;
  t68_n24: return -0.01470847986638546f;
  t68_n25: return 0.014449289999902248f;
  t68_n26: return 0.05611506476998329f;
  t68_n27: return -0.03934461250901222f;
  t68_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n30;
  t68_n29: return -0.04359908401966095f;
  t68_n30: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t68_n32;
  t68_n31: return 0.03982485830783844f;
  t68_n32: return 0.07934954762458801f;
  t68_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n37;
  t68_n34: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t68_n36;
  t68_n35: return 0.002927349181845784f;
  t68_n36: return -0.03204500675201416f;
  t68_n37: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t68_n39;
  t68_n38: return 0.018006939440965652f;
  t68_n39: return -0.14507848024368286f;
  t68_n40: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t68_n46;
  t68_n41: if (E.getNumReferences() >= 3314894848 /*4777.0*/) goto t68_n43;
  t68_n42: return 0.09156343340873718f;
  t68_n43: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t68_n45;
  t68_n44: return 0.065458282828331f;
  t68_n45: return 0.2394275814294815f;
  t68_n46: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t68_n52;
  t68_n47: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t68_n51;
  t68_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t68_n50;
  t68_n49: return 0.019695132970809937f;
  t68_n50: return 0.037892721593379974f;
  t68_n51: return 0.0024646404199302197f;
  t68_n52: return 0.04788602516055107f;
  t68_n53: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n55;
  t68_n54: return -0.009458140470087528f;
  t68_n55: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t68_n67;
  t68_n56: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t68_n66;
  t68_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n61;
  t68_n58: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t68_n60;
  t68_n59: return 0.03278103098273277f;
  t68_n60: return -0.07555735111236572f;
  t68_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t68_n63;
  t68_n62: return -0.174080491065979f;
  t68_n63: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t68_n65;
  t68_n64: return 0.009603488259017467f;
  t68_n65: return -0.0454011894762516f;
  t68_n66: return -0.06764782965183258f;
  t68_n67: return -0.04324723407626152f;
  t68_n68: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t68_n196;
  t68_n69: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t68_n107;
  t68_n70: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t68_n106;
  t68_n71: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t68_n87;
  t68_n72: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t68_n80;
  t68_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t68_n75;
  t68_n74: return -0.003931946121156216f;
  t68_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t68_n77;
  t68_n76: return -0.11621180921792984f;
  t68_n77: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t68_n79;
  t68_n78: return 0.014250781387090683f;
  t68_n79: return 0.0038619970437139273f;
  t68_n80: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t68_n82;
  t68_n81: return -0.046918727457523346f;
  t68_n82: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t68_n86;
  t68_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t68_n85;
  t68_n84: return 0.0068213921040296555f;
  t68_n85: return 0.028289271518588066f;
  t68_n86: return -0.020688043907284737f;
  t68_n87: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t68_n99;
  t68_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t68_n98;
  t68_n89: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t68_n93;
  t68_n90: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n92;
  t68_n91: return 0.02584080770611763f;
  t68_n92: return -0.014859527349472046f;
  t68_n93: if (E.getNumReferences() >= 3335170304 /*25940.5*/) goto t68_n97;
  t68_n94: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t68_n96;
  t68_n95: return 0.04644254222512245f;
  t68_n96: return 0.10088250786066055f;
  t68_n97: return 0.04225064814090729f;
  t68_n98: return 0.07317130267620087f;
  t68_n99: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t68_n105;
  t68_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t68_n102;
  t68_n101: return 0.029012572020292282f;
  t68_n102: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t68_n104;
  t68_n103: return 0.009430029429495335f;
  t68_n104: return 0.09051787108182907f;
  t68_n105: return -0.00011117298708995804f;
  t68_n106: return 0.018687991425395012f;
  t68_n107: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t68_n181;
  t68_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t68_n122;
  t68_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t68_n113;
  t68_n110: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t68_n112;
  t68_n111: return -0.0015981891192495823f;
  t68_n112: return 0.029006032273173332f;
  t68_n113: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t68_n121;
  t68_n114: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t68_n120;
  t68_n115: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t68_n117;
  t68_n116: return 0.00015539051673840731f;
  t68_n117: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t68_n119;
  t68_n118: return 0.05745235085487366f;
  t68_n119: return 0.10105389356613159f;
  t68_n120: return 0.007617449853569269f;
  t68_n121: return 0.046873509883880615f;
  t68_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t68_n134;
  t68_n123: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t68_n129;
  t68_n124: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t68_n126;
  t68_n125: return -0.016114210709929466f;
  t68_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n128;
  t68_n127: return -0.004223562777042389f;
  t68_n128: return 0.019557319581508636f;
  t68_n129: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t68_n131;
  t68_n130: return 0.022404244169592857f;
  t68_n131: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n133;
  t68_n132: return 0.09271413832902908f;
  t68_n133: return -0.03752858564257622f;
  t68_n134: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n166;
  t68_n135: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t68_n149;
  t68_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t68_n142;
  t68_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n141;
  t68_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t68_n140;
  t68_n139: return -0.11170926690101624f;
  t68_n140: return -0.13369978964328766f;
  t68_n141: return -0.0016145480331033468f;
  t68_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t68_n144;
  t68_n143: return 0.006835853215306997f;
  t68_n144: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t68_n148;
  t68_n145: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t68_n147;
  t68_n146: return 0.02109207585453987f;
  t68_n147: return 0.09317293018102646f;
  t68_n148: return 0.002756634261459112f;
  t68_n149: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t68_n151;
  t68_n150: return 0.022527875378727913f;
  t68_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t68_n153;
  t68_n152: return 0.03646276891231537f;
  t68_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t68_n155;
  t68_n154: return -0.006926950067281723f;
  t68_n155: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t68_n161;
  t68_n156: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t68_n160;
  t68_n157: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t68_n159;
  t68_n158: return 0.06862872838973999f;
  t68_n159: return 0.15943121910095215f;
  t68_n160: return 0.06042564660310745f;
  t68_n161: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t68_n165;
  t68_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t68_n164;
  t68_n163: return 0.14363425970077515f;
  t68_n164: return 0.15171365439891815f;
  t68_n165: return 0.0745910257101059f;
  t68_n166: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t68_n172;
  t68_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t68_n169;
  t68_n168: return 0.019840501248836517f;
  t68_n169: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t68_n171;
  t68_n170: return 0.026128679513931274f;
  t68_n171: return 0.0782860815525055f;
  t68_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n178;
  t68_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t68_n177;
  t68_n174: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t68_n176;
  t68_n175: return -0.18863269686698914f;
  t68_n176: return -0.5155418515205383f;
  t68_n177: return 0.03848155215382576f;
  t68_n178: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t68_n180;
  t68_n179: return 0.02811427041888237f;
  t68_n180: return 0.045406240969896317f;
  t68_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t68_n183;
  t68_n182: return -0.08557277172803879f;
  t68_n183: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t68_n189;
  t68_n184: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t68_n186;
  t68_n185: return -0.053463615477085114f;
  t68_n186: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t68_n188;
  t68_n187: return 0.008373741991817951f;
  t68_n188: return 0.2692127823829651f;
  t68_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t68_n195;
  t68_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression))) goto t68_n192;
  t68_n191: return -0.0293552465736866f;
  t68_n192: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t68_n194;
  t68_n193: return 0.03143089637160301f;
  t68_n194: return 0.06696493923664093f;
  t68_n195: return -0.0018375503132119775f;
  t68_n196: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t68_n252;
  t68_n197: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t68_n235;
  t68_n198: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t68_n224;
  t68_n199: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t68_n211;
  t68_n200: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t68_n204;
  t68_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t68_n203;
  t68_n202: return -0.014244114980101585f;
  t68_n203: return 0.012564416043460369f;
  t68_n204: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t68_n210;
  t68_n205: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t68_n207;
  t68_n206: return 0.001915031811222434f;
  t68_n207: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t68_n209;
  t68_n208: return 0.08082020282745361f;
  t68_n209: return -0.012276604771614075f;
  t68_n210: return 0.04723542556166649f;
  t68_n211: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t68_n221;
  t68_n212: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t68_n214;
  t68_n213: return 0.038844313472509384f;
  t68_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t68_n220;
  t68_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t68_n217;
  t68_n216: return 0.06763997673988342f;
  t68_n217: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t68_n219;
  t68_n218: return 0.12695612013339996f;
  t68_n219: return 0.07578017562627792f;
  t68_n220: return -0.04033005237579346f;
  t68_n221: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t68_n223;
  t68_n222: return -0.09175467491149902f;
  t68_n223: return 0.03692564740777016f;
  t68_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n228;
  t68_n225: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t68_n227;
  t68_n226: return -0.040675681084394455f;
  t68_n227: return 0.03918440267443657f;
  t68_n228: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n230;
  t68_n229: return -0.04094868153333664f;
  t68_n230: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t68_n232;
  t68_n231: return 0.12263157963752747f;
  t68_n232: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t68_n234;
  t68_n233: return 0.07884956896305084f;
  t68_n234: return 0.0186779722571373f;
  t68_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t68_n245;
  t68_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n244;
  t68_n237: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t68_n243;
  t68_n238: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t68_n242;
  t68_n239: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t68_n241;
  t68_n240: return 0.003025791374966502f;
  t68_n241: return -0.03745052218437195f;
  t68_n242: return 0.04867655783891678f;
  t68_n243: return -0.05364804342389107f;
  t68_n244: return 0.017743553966283798f;
  t68_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t68_n249;
  t68_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t68_n248;
  t68_n247: return -0.05464918911457062f;
  t68_n248: return 0.043559178709983826f;
  t68_n249: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t68_n251;
  t68_n250: return 0.026925433427095413f;
  t68_n251: return 0.060764748603105545f;
  t68_n252: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t68_n254;
  t68_n253: return 0.032875906676054f;
  t68_n254: return 0.07206141948699951f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree69(const Example& E) {
  t69_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t69_n204;
  t69_n1: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t69_n189;
  t69_n2: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t69_n146;
  t69_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t69_n63;
  t69_n4: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t69_n24;
  t69_n5: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t69_n21;
  t69_n6: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t69_n18;
  t69_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t69_n17;
  t69_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t69_n10;
  t69_n9: return -0.024585364386439323f;
  t69_n10: if (E.getSymbolScopeDistanceCost() >= 3250061312 /*23.0*/) goto t69_n12;
  t69_n11: return 0.0066382382065057755f;
  t69_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t69_n14;
  t69_n13: return -0.061416566371917725f;
  t69_n14: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t69_n16;
  t69_n15: return 0.03238349407911301f;
  t69_n16: return 0.06351539492607117f;
  t69_n17: return -0.021131327375769615f;
  t69_n18: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t69_n20;
  t69_n19: return 0.020371265709400177f;
  t69_n20: return 0.05261973664164543f;
  t69_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t69_n23;
  t69_n22: return 0.009346842765808105f;
  t69_n23: return 0.03667404502630234f;
  t69_n24: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t69_n48;
  t69_n25: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t69_n47;
  t69_n26: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t69_n46;
  t69_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t69_n29;
  t69_n28: return -0.1118781790137291f;
  t69_n29: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t69_n43;
  t69_n30: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t69_n34;
  t69_n31: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t69_n33;
  t69_n32: return -0.0076941680163145065f;
  t69_n33: return 0.003149392083287239f;
  t69_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t69_n42;
  t69_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t69_n39;
  t69_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t69_n38;
  t69_n37: return -0.08970291167497635f;
  t69_n38: return 0.002195828128606081f;
  t69_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t69_n41;
  t69_n40: return 0.004565492272377014f;
  t69_n41: return 0.027369698509573936f;
  t69_n42: return 0.025490373373031616f;
  t69_n43: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t69_n45;
  t69_n44: return -0.059432677924633026f;
  t69_n45: return -0.010385612957179546f;
  t69_n46: return -0.11120700091123581f;
  t69_n47: return -0.05342399701476097f;
  t69_n48: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t69_n56;
  t69_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n51;
  t69_n50: return -0.07334505766630173f;
  t69_n51: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t69_n53;
  t69_n52: return -0.06943868845701218f;
  t69_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n55;
  t69_n54: return -0.2259327471256256f;
  t69_n55: return -0.029802381992340088f;
  t69_n56: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t69_n62;
  t69_n57: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t69_n61;
  t69_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t69_n60;
  t69_n59: return 0.002760627306997776f;
  t69_n60: return 0.019407017156481743f;
  t69_n61: return 0.041848763823509216f;
  t69_n62: return -0.027199266478419304f;
  t69_n63: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t69_n73;
  t69_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t69_n66;
  t69_n65: return 0.007872170768678188f;
  t69_n66: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t69_n70;
  t69_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t69_n69;
  t69_n68: return 0.025305328890681267f;
  t69_n69: return 0.053334906697273254f;
  t69_n70: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t69_n72;
  t69_n71: return -0.045360397547483444f;
  t69_n72: return 0.02593200094997883f;
  t69_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t69_n95;
  t69_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n76;
  t69_n75: return -0.12004073709249496f;
  t69_n76: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t69_n84;
  t69_n77: if (E.getFileProximityDistanceCost() >= 3255304192 /*34.0*/) goto t69_n81;
  t69_n78: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t69_n80;
  t69_n79: return 0.014091034419834614f;
  t69_n80: return -0.0495871976017952f;
  t69_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n83;
  t69_n82: return -0.025140108540654182f;
  t69_n83: return 0.022854400798678398f;
  t69_n84: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t69_n92;
  t69_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t69_n91;
  t69_n86: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t69_n88;
  t69_n87: return 0.052195243537425995f;
  t69_n88: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t69_n90;
  t69_n89: return 0.08480794727802277f;
  t69_n90: return 0.057350169867277145f;
  t69_n91: return -0.04941180720925331f;
  t69_n92: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t69_n94;
  t69_n93: return -0.0059385341592133045f;
  t69_n94: return 0.031125228852033615f;
  t69_n95: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t69_n141;
  t69_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t69_n100;
  t69_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t69_n99;
  t69_n98: return -0.016170555725693703f;
  t69_n99: return 0.01827268861234188f;
  t69_n100: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t69_n104;
  t69_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n103;
  t69_n102: return -0.07177291065454483f;
  t69_n103: return 0.0025727611500769854f;
  t69_n104: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t69_n118;
  t69_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t69_n115;
  t69_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n108;
  t69_n107: return -0.01151289977133274f;
  t69_n108: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t69_n110;
  t69_n109: return 0.011251511983573437f;
  t69_n110: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t69_n112;
  t69_n111: return 0.01252078264951706f;
  t69_n112: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t69_n114;
  t69_n113: return -0.05969751626253128f;
  t69_n114: return -0.002310876501724124f;
  t69_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t69_n117;
  t69_n116: return 0.005768734961748123f;
  t69_n117: return 0.023827601224184036f;
  t69_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t69_n122;
  t69_n119: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t69_n121;
  t69_n120: return -0.014134456403553486f;
  t69_n121: return 0.009846163913607597f;
  t69_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n124;
  t69_n123: return -0.020927734673023224f;
  t69_n124: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t69_n132;
  t69_n125: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t69_n131;
  t69_n126: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t69_n128;
  t69_n127: return -0.01422639936208725f;
  t69_n128: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t69_n130;
  t69_n129: return 0.054393839091062546f;
  t69_n130: return 0.02387552708387375f;
  t69_n131: return -0.09819086641073227f;
  t69_n132: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t69_n140;
  t69_n133: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t69_n135;
  t69_n134: return 0.035658638924360275f;
  t69_n135: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t69_n137;
  t69_n136: return -0.02498564124107361f;
  t69_n137: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t69_n139;
  t69_n138: return 0.019202303141355515f;
  t69_n139: return 0.11570818722248077f;
  t69_n140: return -0.03637463226914406f;
  t69_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n143;
  t69_n142: return -0.028027435764670372f;
  t69_n143: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t69_n145;
  t69_n144: return 0.06927484273910522f;
  t69_n145: return 0.11810559779405594f;
  t69_n146: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t69_n162;
  t69_n147: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t69_n161;
  t69_n148: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t69_n158;
  t69_n149: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t69_n151;
  t69_n150: return 0.019500402733683586f;
  t69_n151: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t69_n155;
  t69_n152: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t69_n154;
  t69_n153: return 0.09063219279050827f;
  t69_n154: return 0.10972978174686432f;
  t69_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t69_n157;
  t69_n156: return 0.002600056119263172f;
  t69_n157: return 0.055320028215646744f;
  t69_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n160;
  t69_n159: return -0.02553577721118927f;
  t69_n160: return 0.012104342691600323f;
  t69_n161: return 0.04267645254731178f;
  t69_n162: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t69_n188;
  t69_n163: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t69_n165;
  t69_n164: return 0.04161845147609711f;
  t69_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t69_n187;
  t69_n166: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t69_n170;
  t69_n167: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t69_n169;
  t69_n168: return 0.03770177438855171f;
  t69_n169: return -0.02359423227608204f;
  t69_n170: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t69_n172;
  t69_n171: return 0.11655662208795547f;
  t69_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t69_n180;
  t69_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n177;
  t69_n174: if (E.getNumReferences() >= 3344808448 /*56762.0*/) goto t69_n176;
  t69_n175: return -0.036265868693590164f;
  t69_n176: return -0.157640278339386f;
  t69_n177: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t69_n179;
  t69_n178: return 0.005402117967605591f;
  t69_n179: return 0.06490904092788696f;
  t69_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t69_n182;
  t69_n181: return 0.0042724586091935635f;
  t69_n182: if (E.getNumReferences() >= 3346546432 /*63551.0*/) goto t69_n184;
  t69_n183: return 0.015330738388001919f;
  t69_n184: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t69_n186;
  t69_n185: return 0.059568874537944794f;
  t69_n186: return 0.00409740349277854f;
  t69_n187: return 0.025965576991438866f;
  t69_n188: return 0.028956713154911995f;
  t69_n189: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t69_n193;
  t69_n190: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t69_n192;
  t69_n191: return 0.0133379390463233f;
  t69_n192: return 0.002782319439575076f;
  t69_n193: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t69_n199;
  t69_n194: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t69_n198;
  t69_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t69_n197;
  t69_n196: return 0.012307027354836464f;
  t69_n197: return 0.015098980627954006f;
  t69_n198: return 0.050652436912059784f;
  t69_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t69_n201;
  t69_n200: return 0.030248722061514854f;
  t69_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t69_n203;
  t69_n202: return 0.061760928481817245f;
  t69_n203: return 0.1889885812997818f;
  t69_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t69_n244;
  t69_n205: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t69_n211;
  t69_n206: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t69_n210;
  t69_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t69_n209;
  t69_n208: return -0.08642549812793732f;
  t69_n209: return 0.02926739677786827f;
  t69_n210: return 0.013706988655030727f;
  t69_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t69_n221;
  t69_n212: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t69_n216;
  t69_n213: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t69_n215;
  t69_n214: return -0.03605586662888527f;
  t69_n215: return 0.03152866289019585f;
  t69_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n220;
  t69_n217: if (E.getNumReferences() >= 3321914880 /*8217.5*/) goto t69_n219;
  t69_n218: return -0.01658075861632824f;
  t69_n219: return -0.26123306155204773f;
  t69_n220: return 0.029456064105033875f;
  t69_n221: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t69_n229;
  t69_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t69_n226;
  t69_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t69_n225;
  t69_n224: return -0.048129767179489136f;
  t69_n225: return 0.00028935709269717336f;
  t69_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t69_n228;
  t69_n227: return 0.021067675203084946f;
  t69_n228: return 0.07600966095924377f;
  t69_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n233;
  t69_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n232;
  t69_n231: return -0.11699681729078293f;
  t69_n232: return 0.022511428222060204f;
  t69_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n235;
  t69_n234: return 0.01765330135822296f;
  t69_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n239;
  t69_n236: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t69_n238;
  t69_n237: return -0.09946931898593903f;
  t69_n238: return 0.018027449026703835f;
  t69_n239: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t69_n241;
  t69_n240: return 0.03993644565343857f;
  t69_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n243;
  t69_n242: return -0.012243119068443775f;
  t69_n243: return 0.07840599864721298f;
  t69_n244: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t69_n246;
  t69_n245: return -0.16199244558811188f;
  t69_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t69_n248;
  t69_n247: return -0.10104726254940033f;
  t69_n248: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t69_n252;
  t69_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t69_n251;
  t69_n250: return -0.0852026492357254f;
  t69_n251: return 0.12346014380455017f;
  t69_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t69_n254;
  t69_n253: return -0.029897339642047882f;
  t69_n254: return 0.027976421639323235f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree70(const Example& E) {
  t70_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t70_n96;
  t70_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t70_n87;
  t70_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t70_n72;
  t70_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t70_n53;
  t70_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t70_n46;
  t70_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n15;
  t70_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t70_n10;
  t70_n7: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t70_n9;
  t70_n8: return 0.0037180916406214237f;
  t70_n9: return 0.02874496579170227f;
  t70_n10: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t70_n12;
  t70_n11: return -0.009784668684005737f;
  t70_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t70_n14;
  t70_n13: return 0.012447888031601906f;
  t70_n14: return 0.06809727102518082f;
  t70_n15: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t70_n45;
  t70_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t70_n18;
  t70_n17: return -0.1054556667804718f;
  t70_n18: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t70_n42;
  t70_n19: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t70_n41;
  t70_n20: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t70_n40;
  t70_n21: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t70_n35;
  t70_n22: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t70_n28;
  t70_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n25;
  t70_n24: return -0.014374143444001675f;
  t70_n25: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t70_n27;
  t70_n26: return 0.025860359892249107f;
  t70_n27: return 0.001964410999789834f;
  t70_n28: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t70_n34;
  t70_n29: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t70_n31;
  t70_n30: return 0.030736656859517097f;
  t70_n31: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t70_n33;
  t70_n32: return -0.012850752100348473f;
  t70_n33: return 0.012542026117444038f;
  t70_n34: return -0.14277130365371704f;
  t70_n35: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t70_n37;
  t70_n36: return -0.05601263418793678f;
  t70_n37: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t70_n39;
  t70_n38: return 0.015264918096363544f;
  t70_n39: return -0.018787328153848648f;
  t70_n40: return -0.09826666116714478f;
  t70_n41: return -0.06374657899141312f;
  t70_n42: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t70_n44;
  t70_n43: return 0.021531937643885612f;
  t70_n44: return 0.08339829742908478f;
  t70_n45: return 0.03779643774032593f;
  t70_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n48;
  t70_n47: return -0.06333062797784805f;
  t70_n48: if (E.getNumReferences() >= 3365173872 /*304083.5*/) goto t70_n52;
  t70_n49: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t70_n51;
  t70_n50: return -0.041913945227861404f;
  t70_n51: return 0.0805462971329689f;
  t70_n52: return -0.12658096849918365f;
  t70_n53: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t70_n55;
  t70_n54: return 0.0002596431295387447f;
  t70_n55: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t70_n71;
  t70_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t70_n58;
  t70_n57: return 0.011492914520204067f;
  t70_n58: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t70_n64;
  t70_n59: if (E.getNumReferences() >= 3333689088 /*23047.5*/) goto t70_n63;
  t70_n60: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t70_n62;
  t70_n61: return 0.037082038819789886f;
  t70_n62: return -0.006534911692142487f;
  t70_n63: return 0.03951925039291382f;
  t70_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t70_n68;
  t70_n65: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t70_n67;
  t70_n66: return -9.488161595072597e-05f;
  t70_n67: return 0.07135593146085739f;
  t70_n68: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t70_n70;
  t70_n69: return 0.042500194162130356f;
  t70_n70: return 0.09854082763195038f;
  t70_n71: return 0.040223628282547f;
  t70_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t70_n74;
  t70_n73: return -0.06772133708000183f;
  t70_n74: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t70_n76;
  t70_n75: return -0.01536569558084011f;
  t70_n76: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t70_n86;
  t70_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t70_n79;
  t70_n78: return 0.015170847065746784f;
  t70_n79: if (E.getNumReferences() >= 3354001600 /*119809.5*/) goto t70_n81;
  t70_n80: return 0.07781877368688583f;
  t70_n81: if (E.getNumReferences() >= 3355438208 /*131033.0*/) goto t70_n85;
  t70_n82: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t70_n84;
  t70_n83: return 0.13240568339824677f;
  t70_n84: return -0.09073320031166077f;
  t70_n85: return 0.08590472489595413f;
  t70_n86: return -0.005416477099061012f;
  t70_n87: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t70_n95;
  t70_n88: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t70_n94;
  t70_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t70_n93;
  t70_n90: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t70_n92;
  t70_n91: return 0.011875782161951065f;
  t70_n92: return -0.009918025694787502f;
  t70_n93: return 0.011676819995045662f;
  t70_n94: return 0.0333237498998642f;
  t70_n95: return 0.07425707578659058f;
  t70_n96: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t70_n240;
  t70_n97: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t70_n169;
  t70_n98: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t70_n158;
  t70_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t70_n143;
  t70_n100: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t70_n142;
  t70_n101: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t70_n129;
  t70_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t70_n108;
  t70_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t70_n107;
  t70_n104: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t70_n106;
  t70_n105: return -0.11019354313611984f;
  t70_n106: return -0.21296614408493042f;
  t70_n107: return -0.0782366544008255f;
  t70_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace))) goto t70_n126;
  t70_n109: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t70_n115;
  t70_n110: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t70_n112;
  t70_n111: return 0.0044263978488743305f;
  t70_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t70_n114;
  t70_n113: return 0.001242998638190329f;
  t70_n114: return 0.08153422921895981f;
  t70_n115: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t70_n125;
  t70_n116: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t70_n118;
  t70_n117: return -0.010800966061651707f;
  t70_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t70_n120;
  t70_n119: return -0.0063761938363313675f;
  t70_n120: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t70_n124;
  t70_n121: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t70_n123;
  t70_n122: return 0.046299047768116f;
  t70_n123: return 0.08389978110790253f;
  t70_n124: return -0.05662062391638756f;
  t70_n125: return 0.11433184146881104f;
  t70_n126: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t70_n128;
  t70_n127: return 0.004402089398354292f;
  t70_n128: return 0.05708376318216324f;
  t70_n129: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t70_n131;
  t70_n130: return 0.10093560069799423f;
  t70_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n135;
  t70_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t70_n134;
  t70_n133: return -0.11752552539110184f;
  t70_n134: return 0.09046617895364761f;
  t70_n135: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t70_n137;
  t70_n136: return 0.00253682560287416f;
  t70_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t70_n139;
  t70_n138: return 0.010681167244911194f;
  t70_n139: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t70_n141;
  t70_n140: return 0.04646321386098862f;
  t70_n141: return -0.20978260040283203f;
  t70_n142: return -0.028603676706552505f;
  t70_n143: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t70_n157;
  t70_n144: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t70_n150;
  t70_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t70_n147;
  t70_n146: return -0.11947037279605865f;
  t70_n147: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t70_n149;
  t70_n148: return 0.010185997001826763f;
  t70_n149: return 0.016707489266991615f;
  t70_n150: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t70_n152;
  t70_n151: return -0.04952637106180191f;
  t70_n152: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t70_n154;
  t70_n153: return 0.0017726122168824077f;
  t70_n154: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t70_n156;
  t70_n155: return 0.1203204020857811f;
  t70_n156: return 0.03093569166958332f;
  t70_n157: return -0.011331886053085327f;
  t70_n158: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t70_n168;
  t70_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n163;
  t70_n160: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t70_n162;
  t70_n161: return -0.08320096135139465f;
  t70_n162: return 0.0443752221763134f;
  t70_n163: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t70_n165;
  t70_n164: return 0.024446187540888786f;
  t70_n165: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t70_n167;
  t70_n166: return 0.0967518538236618f;
  t70_n167: return 0.05660240724682808f;
  t70_n168: return -0.0003423729503992945f;
  t70_n169: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t70_n235;
  t70_n170: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t70_n230;
  t70_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t70_n207;
  t70_n172: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t70_n206;
  t70_n173: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t70_n205;
  t70_n174: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t70_n200;
  t70_n175: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t70_n199;
  t70_n176: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t70_n194;
  t70_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t70_n179;
  t70_n178: return -0.0005192339303903282f;
  t70_n179: if (E.getNumReferences() >= 3356745728 /*151424.0*/) goto t70_n193;
  t70_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t70_n182;
  t70_n181: return 0.007682160008698702f;
  t70_n182: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t70_n184;
  t70_n183: return 0.016656072810292244f;
  t70_n184: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t70_n186;
  t70_n185: return -0.02530699595808983f;
  t70_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n188;
  t70_n187: return -0.187574103474617f;
  t70_n188: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t70_n192;
  t70_n189: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t70_n191;
  t70_n190: return 0.11232347041368484f;
  t70_n191: return -0.24603410065174103f;
  t70_n192: return 0.05973834544420242f;
  t70_n193: return 0.06287229061126709f;
  t70_n194: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t70_n198;
  t70_n195: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t70_n197;
  t70_n196: return -0.01897764392197132f;
  t70_n197: return 0.10820161551237106f;
  t70_n198: return -0.10594604909420013f;
  t70_n199: return -0.052139174193143845f;
  t70_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t70_n202;
  t70_n201: return -0.003364401403814554f;
  t70_n202: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t70_n204;
  t70_n203: return 0.01499840896576643f;
  t70_n204: return 0.04589444771409035f;
  t70_n205: return -0.14595571160316467f;
  t70_n206: return -0.049314409494400024f;
  t70_n207: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t70_n211;
  t70_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t70_n210;
  t70_n209: return -0.07496010512113571f;
  t70_n210: return -0.021299047395586967f;
  t70_n211: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t70_n215;
  t70_n212: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t70_n214;
  t70_n213: return 0.016623605042696f;
  t70_n214: return -0.4121217727661133f;
  t70_n215: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t70_n221;
  t70_n216: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t70_n220;
  t70_n217: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t70_n219;
  t70_n218: return 0.10258150100708008f;
  t70_n219: return 0.1672564595937729f;
  t70_n220: return 0.07408419251441956f;
  t70_n221: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t70_n225;
  t70_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n224;
  t70_n223: return -0.06041523069143295f;
  t70_n224: return 0.038887783885002136f;
  t70_n225: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t70_n229;
  t70_n226: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t70_n228;
  t70_n227: return 0.2015175074338913f;
  t70_n228: return 0.10212890058755875f;
  t70_n229: return 0.06098787859082222f;
  t70_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t70_n232;
  t70_n231: return 0.014904329553246498f;
  t70_n232: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t70_n234;
  t70_n233: return 0.03418472409248352f;
  t70_n234: return 0.07485323399305344f;
  t70_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t70_n239;
  t70_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t70_n238;
  t70_n237: return -0.17262287437915802f;
  t70_n238: return 0.04385679215192795f;
  t70_n239: return 0.030494684353470802f;
  t70_n240: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t70_n254;
  t70_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t70_n243;
  t70_n242: return -0.16796159744262695f;
  t70_n243: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t70_n253;
  t70_n244: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t70_n248;
  t70_n245: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t70_n247;
  t70_n246: return 0.03496182709932327f;
  t70_n247: return 0.006414040457457304f;
  t70_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t70_n250;
  t70_n249: return 0.003299789736047387f;
  t70_n250: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t70_n252;
  t70_n251: return 0.011135446839034557f;
  t70_n252: return 0.0367005281150341f;
  t70_n253: return 0.03589833155274391f;
  t70_n254: return 0.059999845921993256f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree71(const Example& E) {
  t71_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t71_n218;
  t71_n1: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t71_n217;
  t71_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t71_n98;
  t71_n3: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t71_n81;
  t71_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n36;
  t71_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n31;
  t71_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t71_n20;
  t71_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t71_n9;
  t71_n8: return -0.028751235455274582f;
  t71_n9: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t71_n19;
  t71_n10: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t71_n16;
  t71_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t71_n13;
  t71_n12: return -0.11275365948677063f;
  t71_n13: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t71_n15;
  t71_n14: return 0.03063090331852436f;
  t71_n15: return 0.06677465885877609f;
  t71_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t71_n18;
  t71_n17: return -0.12146227061748505f;
  t71_n18: return 0.09175892919301987f;
  t71_n19: return -0.007807630114257336f;
  t71_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t71_n30;
  t71_n21: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t71_n29;
  t71_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n28;
  t71_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t71_n27;
  t71_n24: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t71_n26;
  t71_n25: return 0.007388814352452755f;
  t71_n26: return -0.08325912803411484f;
  t71_n27: return 0.036826737225055695f;
  t71_n28: return 0.003342939307913184f;
  t71_n29: return 0.014144610613584518f;
  t71_n30: return -0.05384319648146629f;
  t71_n31: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t71_n33;
  t71_n32: return 0.00529444869607687f;
  t71_n33: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t71_n35;
  t71_n34: return 0.056727562099695206f;
  t71_n35: return -0.11349531263113022f;
  t71_n36: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t71_n78;
  t71_n37: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t71_n69;
  t71_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t71_n48;
  t71_n39: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t71_n41;
  t71_n40: return -0.016853084787726402f;
  t71_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t71_n43;
  t71_n42: return -0.01511345710605383f;
  t71_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t71_n47;
  t71_n44: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n46;
  t71_n45: return 0.004053792450577021f;
  t71_n46: return 0.03730740770697594f;
  t71_n47: return 0.029492583125829697f;
  t71_n48: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t71_n68;
  t71_n49: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t71_n57;
  t71_n50: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t71_n56;
  t71_n51: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t71_n53;
  t71_n52: return 0.02291863225400448f;
  t71_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t71_n55;
  t71_n54: return 0.009696713648736477f;
  t71_n55: return 0.06553984433412552f;
  t71_n56: return -0.10010524094104767f;
  t71_n57: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t71_n59;
  t71_n58: return 0.005653791129589081f;
  t71_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t71_n61;
  t71_n60: return 0.002504076110199094f;
  t71_n61: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t71_n63;
  t71_n62: return 0.045853883028030396f;
  t71_n63: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t71_n65;
  t71_n64: return 0.07790952175855637f;
  t71_n65: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t71_n67;
  t71_n66: return 0.11020470410585403f;
  t71_n67: return 0.10390473157167435f;
  t71_n68: return 0.002382950158789754f;
  t71_n69: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t71_n73;
  t71_n70: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t71_n72;
  t71_n71: return -0.05785110220313072f;
  t71_n72: return -0.18666324019432068f;
  t71_n73: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n75;
  t71_n74: return -0.0013333630049601197f;
  t71_n75: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t71_n77;
  t71_n76: return 0.04553498327732086f;
  t71_n77: return 0.008277458138763905f;
  t71_n78: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t71_n80;
  t71_n79: return -0.013426230289041996f;
  t71_n80: return 0.022462183609604836f;
  t71_n81: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n97;
  t71_n82: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t71_n94;
  t71_n83: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t71_n89;
  t71_n84: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t71_n88;
  t71_n85: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t71_n87;
  t71_n86: return 0.017832661047577858f;
  t71_n87: return 0.11005241423845291f;
  t71_n88: return -0.11292045563459396f;
  t71_n89: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t71_n93;
  t71_n90: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t71_n92;
  t71_n91: return 0.08574328571557999f;
  t71_n92: return 0.10696668177843094f;
  t71_n93: return 0.0323273167014122f;
  t71_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t71_n96;
  t71_n95: return -0.026646453887224197f;
  t71_n96: return 0.010735615156590939f;
  t71_n97: return 0.03745754063129425f;
  t71_n98: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t71_n106;
  t71_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t71_n103;
  t71_n100: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t71_n102;
  t71_n101: return 0.0015722744865342975f;
  t71_n102: return 0.03845896199345589f;
  t71_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t71_n105;
  t71_n104: return 0.025647277012467384f;
  t71_n105: return 0.051098089665174484f;
  t71_n106: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t71_n214;
  t71_n107: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t71_n213;
  t71_n108: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t71_n132;
  t71_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n123;
  t71_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n114;
  t71_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t71_n113;
  t71_n112: return -0.10419312864542007f;
  t71_n113: return -0.07519478350877762f;
  t71_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t71_n116;
  t71_n115: return -0.04208843782544136f;
  t71_n116: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t71_n122;
  t71_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n119;
  t71_n118: return -0.026992756873369217f;
  t71_n119: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t71_n121;
  t71_n120: return 0.025821158662438393f;
  t71_n121: return 0.13329261541366577f;
  t71_n122: return 0.023461678996682167f;
  t71_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t71_n127;
  t71_n124: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t71_n126;
  t71_n125: return -0.027863604947924614f;
  t71_n126: return -0.5279580354690552f;
  t71_n127: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t71_n131;
  t71_n128: if (E.getNumReferences() >= 3358501856 /*178863.5*/) goto t71_n130;
  t71_n129: return 0.010308504104614258f;
  t71_n130: return 0.052865561097860336f;
  t71_n131: return -0.013665090315043926f;
  t71_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t71_n178;
  t71_n133: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t71_n139;
  t71_n134: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t71_n138;
  t71_n135: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t71_n137;
  t71_n136: return -0.022784464061260223f;
  t71_n137: return 0.049740228801965714f;
  t71_n138: return -0.07263775914907455f;
  t71_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t71_n141;
  t71_n140: return -0.035509321838617325f;
  t71_n141: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n175;
  t71_n142: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t71_n172;
  t71_n143: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t71_n155;
  t71_n144: if (E.getNumReferences() >= 3338441472 /*32329.5*/) goto t71_n150;
  t71_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n147;
  t71_n146: return -0.02320118434727192f;
  t71_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t71_n149;
  t71_n148: return -0.1126018762588501f;
  t71_n149: return 0.014335177838802338f;
  t71_n150: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t71_n154;
  t71_n151: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t71_n153;
  t71_n152: return 0.05731713026762009f;
  t71_n153: return 0.010549576953053474f;
  t71_n154: return -0.014783789403736591f;
  t71_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t71_n167;
  t71_n156: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t71_n166;
  t71_n157: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t71_n165;
  t71_n158: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t71_n162;
  t71_n159: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t71_n161;
  t71_n160: return 0.012293297797441483f;
  t71_n161: return -0.07851066440343857f;
  t71_n162: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t71_n164;
  t71_n163: return -0.02563866786658764f;
  t71_n164: return 0.024327926337718964f;
  t71_n165: return -0.11397617310285568f;
  t71_n166: return -0.08566278964281082f;
  t71_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t71_n169;
  t71_n168: return -0.11537347733974457f;
  t71_n169: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t71_n171;
  t71_n170: return 0.032572221010923386f;
  t71_n171: return -0.11066985130310059f;
  t71_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n174;
  t71_n173: return -0.05039924383163452f;
  t71_n174: return -0.011254052631556988f;
  t71_n175: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t71_n177;
  t71_n176: return 0.013730434700846672f;
  t71_n177: return -0.13439612090587616f;
  t71_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t71_n190;
  t71_n179: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t71_n183;
  t71_n180: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t71_n182;
  t71_n181: return 0.0026453742757439613f;
  t71_n182: return -0.03366788849234581f;
  t71_n183: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t71_n189;
  t71_n184: if (E.getNumReferences() >= 3331706880 /*19176.0*/) goto t71_n188;
  t71_n185: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t71_n187;
  t71_n186: return 0.13448169827461243f;
  t71_n187: return 0.042087361216545105f;
  t71_n188: return 0.10350263118743896f;
  t71_n189: return -0.03823171183466911f;
  t71_n190: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t71_n212;
  t71_n191: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t71_n193;
  t71_n192: return 0.01079937070608139f;
  t71_n193: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t71_n209;
  t71_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t71_n196;
  t71_n195: return 0.007691866718232632f;
  t71_n196: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t71_n208;
  t71_n197: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t71_n205;
  t71_n198: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t71_n202;
  t71_n199: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t71_n201;
  t71_n200: return 0.1021914929151535f;
  t71_n201: return 0.13169491291046143f;
  t71_n202: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t71_n204;
  t71_n203: return 0.25252798199653625f;
  t71_n204: return 0.18569549918174744f;
  t71_n205: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t71_n207;
  t71_n206: return 0.0064409272745251656f;
  t71_n207: return 0.09967310726642609f;
  t71_n208: return 0.02566862851381302f;
  t71_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t71_n211;
  t71_n210: return 0.004576199222356081f;
  t71_n211: return 0.04575693607330322f;
  t71_n212: return 0.029757201671600342f;
  t71_n213: return 0.0364827997982502f;
  t71_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t71_n216;
  t71_n215: return -0.05458897724747658f;
  t71_n216: return 0.021749896928668022f;
  t71_n217: return 0.017435694113373756f;
  t71_n218: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t71_n252;
  t71_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t71_n237;
  t71_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t71_n236;
  t71_n221: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t71_n235;
  t71_n222: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t71_n224;
  t71_n223: return -0.016966544091701508f;
  t71_n224: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t71_n234;
  t71_n225: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t71_n229;
  t71_n226: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t71_n228;
  t71_n227: return 0.10590396076440811f;
  t71_n228: return 0.0012015588581562042f;
  t71_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t71_n233;
  t71_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t71_n232;
  t71_n231: return 0.0007580540841445327f;
  t71_n232: return 0.03383727744221687f;
  t71_n233: return 0.05213085561990738f;
  t71_n234: return -0.1962215006351471f;
  t71_n235: return 0.016562214121222496f;
  t71_n236: return -0.04905340075492859f;
  t71_n237: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t71_n251;
  t71_n238: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t71_n244;
  t71_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t71_n241;
  t71_n240: return -0.12003493309020996f;
  t71_n241: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t71_n243;
  t71_n242: return 0.016175219789147377f;
  t71_n243: return 0.030848465859889984f;
  t71_n244: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t71_n246;
  t71_n245: return -0.015428265556693077f;
  t71_n246: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t71_n248;
  t71_n247: return 0.009011354297399521f;
  t71_n248: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t71_n250;
  t71_n249: return 0.09278161078691483f;
  t71_n250: return 0.029951946809887886f;
  t71_n251: return -0.0028569456189870834f;
  t71_n252: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t71_n254;
  t71_n253: return 0.018780697137117386f;
  t71_n254: return 0.05655878409743309f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree72(const Example& E) {
  t72_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t72_n218;
  t72_n1: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t72_n201;
  t72_n2: if (E.getNumReferences() >= 3325315072 /*11538.0*/) goto t72_n186;
  t72_n3: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t72_n51;
  t72_n4: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t72_n48;
  t72_n5: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t72_n45;
  t72_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t72_n30;
  t72_n7: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n27;
  t72_n8: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t72_n16;
  t72_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t72_n11;
  t72_n10: return -0.022945456206798553f;
  t72_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t72_n13;
  t72_n12: return -0.027031080797314644f;
  t72_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t72_n15;
  t72_n14: return 0.010798439383506775f;
  t72_n15: return -0.08937552571296692f;
  t72_n16: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t72_n18;
  t72_n17: return -0.05955107510089874f;
  t72_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t72_n24;
  t72_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t72_n21;
  t72_n20: return -0.011949531733989716f;
  t72_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n23;
  t72_n22: return -0.12106898427009583f;
  t72_n23: return 0.01795179583132267f;
  t72_n24: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t72_n26;
  t72_n25: return 0.04249892756342888f;
  t72_n26: return 0.006423247512429953f;
  t72_n27: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t72_n29;
  t72_n28: return 0.008481742814183235f;
  t72_n29: return 0.04492722079157829f;
  t72_n30: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t72_n34;
  t72_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t72_n33;
  t72_n32: return 0.011723071336746216f;
  t72_n33: return 0.046009719371795654f;
  t72_n34: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t72_n36;
  t72_n35: return 0.043239302933216095f;
  t72_n36: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t72_n38;
  t72_n37: return -0.08849408477544785f;
  t72_n38: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t72_n44;
  t72_n39: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t72_n41;
  t72_n40: return 0.05790461227297783f;
  t72_n41: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t72_n43;
  t72_n42: return 0.13262230157852173f;
  t72_n43: return 0.08327508717775345f;
  t72_n44: return 0.006255957297980785f;
  t72_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t72_n47;
  t72_n46: return 0.011897680349647999f;
  t72_n47: return 0.034137625247240067f;
  t72_n48: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t72_n50;
  t72_n49: return 0.018477724865078926f;
  t72_n50: return -0.05849084258079529f;
  t72_n51: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t72_n97;
  t72_n52: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t72_n94;
  t72_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t72_n79;
  t72_n54: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t72_n66;
  t72_n55: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t72_n65;
  t72_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t72_n62;
  t72_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t72_n61;
  t72_n58: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t72_n60;
  t72_n59: return -0.07707279920578003f;
  t72_n60: return -0.0992853045463562f;
  t72_n61: return -0.08698058873414993f;
  t72_n62: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t72_n64;
  t72_n63: return 0.08119005709886551f;
  t72_n64: return 0.02678476646542549f;
  t72_n65: return -0.003860791679471731f;
  t72_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t72_n70;
  t72_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n69;
  t72_n68: return -0.12067995220422745f;
  t72_n69: return 0.0017220383742824197f;
  t72_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t72_n78;
  t72_n71: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n77;
  t72_n72: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t72_n74;
  t72_n73: return 0.00988228153437376f;
  t72_n74: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t72_n76;
  t72_n75: return 0.13014563918113708f;
  t72_n76: return -0.18151047825813293f;
  t72_n77: return 0.038360845297575f;
  t72_n78: return 0.10550690442323685f;
  t72_n79: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t72_n93;
  t72_n80: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t72_n92;
  t72_n81: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t72_n85;
  t72_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n84;
  t72_n83: return -0.020558562129735947f;
  t72_n84: return 0.012851179577410221f;
  t72_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t72_n87;
  t72_n86: return 0.0014959834516048431f;
  t72_n87: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t72_n89;
  t72_n88: return 0.034874726086854935f;
  t72_n89: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t72_n91;
  t72_n90: return -0.017712559551000595f;
  t72_n91: return 0.059374988079071045f;
  t72_n92: return 0.04545494541525841f;
  t72_n93: return -0.04328424111008644f;
  t72_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t72_n96;
  t72_n95: return 0.01045547891408205f;
  t72_n96: return 0.041611939668655396f;
  t72_n97: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t72_n173;
  t72_n98: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t72_n100;
  t72_n99: return -0.014745493419468403f;
  t72_n100: if (E.getSymbolScopeDistanceCost() >= 3247964160 /*19.0*/) goto t72_n128;
  t72_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t72_n119;
  t72_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t72_n104;
  t72_n103: return -0.03903406858444214f;
  t72_n104: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t72_n118;
  t72_n105: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t72_n117;
  t72_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t72_n108;
  t72_n107: return -0.006857056636363268f;
  t72_n108: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t72_n110;
  t72_n109: return -0.0037047695368528366f;
  t72_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t72_n112;
  t72_n111: return 0.009888457134366035f;
  t72_n112: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t72_n116;
  t72_n113: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t72_n115;
  t72_n114: return 0.0416906476020813f;
  t72_n115: return -0.09836450964212418f;
  t72_n116: return 0.05587027221918106f;
  t72_n117: return 0.05594070628285408f;
  t72_n118: return -0.009100121445953846f;
  t72_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t72_n123;
  t72_n120: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t72_n122;
  t72_n121: return 0.00937761552631855f;
  t72_n122: return 0.10288578271865845f;
  t72_n123: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t72_n125;
  t72_n124: return 0.046487826853990555f;
  t72_n125: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t72_n127;
  t72_n126: return 0.09381868690252304f;
  t72_n127: return 0.05445033311843872f;
  t72_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type))) goto t72_n130;
  t72_n129: return -0.016633426770567894f;
  t72_n130: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t72_n172;
  t72_n131: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t72_n171;
  t72_n132: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t72_n170;
  t72_n133: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t72_n169;
  t72_n134: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t72_n168;
  t72_n135: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n165;
  t72_n136: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t72_n156;
  t72_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t72_n145;
  t72_n138: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t72_n144;
  t72_n139: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t72_n141;
  t72_n140: return -0.002940180478617549f;
  t72_n141: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t72_n143;
  t72_n142: return 0.02631577104330063f;
  t72_n143: return -0.06785233318805695f;
  t72_n144: return -0.044524185359478f;
  t72_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t72_n147;
  t72_n146: return -0.005051585845649242f;
  t72_n147: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t72_n153;
  t72_n148: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t72_n152;
  t72_n149: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t72_n151;
  t72_n150: return 0.08426354080438614f;
  t72_n151: return 0.09795699268579483f;
  t72_n152: return 0.040962908416986465f;
  t72_n153: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t72_n155;
  t72_n154: return 0.054033324122428894f;
  t72_n155: return 0.15269342064857483f;
  t72_n156: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t72_n164;
  t72_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n159;
  t72_n158: return -0.05418839305639267f;
  t72_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t72_n161;
  t72_n160: return 0.026996441185474396f;
  t72_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t72_n163;
  t72_n162: return -0.11484655737876892f;
  t72_n163: return 0.05825858563184738f;
  t72_n164: return -0.04990806430578232f;
  t72_n165: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t72_n167;
  t72_n166: return 0.03787078708410263f;
  t72_n167: return -0.00022021248878445476f;
  t72_n168: return -0.1347620040178299f;
  t72_n169: return 0.013937159441411495f;
  t72_n170: return -0.0397600457072258f;
  t72_n171: return -0.11560741066932678f;
  t72_n172: return -0.0648217648267746f;
  t72_n173: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t72_n177;
  t72_n174: if (E.getNumReferences() >= 3318942720 /*6753.5*/) goto t72_n176;
  t72_n175: return 0.028713738545775414f;
  t72_n176: return -0.23702946305274963f;
  t72_n177: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t72_n179;
  t72_n178: return -0.07913464307785034f;
  t72_n179: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t72_n185;
  t72_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t72_n182;
  t72_n181: return 0.018648376688361168f;
  t72_n182: if (E.getNumReferences() >= 3298836480 /*1282.0*/) goto t72_n184;
  t72_n183: return 0.09643704444169998f;
  t72_n184: return -0.036759428679943085f;
  t72_n185: return -0.01752423495054245f;
  t72_n186: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t72_n194;
  t72_n187: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t72_n193;
  t72_n188: if (E.getNumReferences() >= 3327362048 /*13537.0*/) goto t72_n192;
  t72_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t72_n191;
  t72_n190: return 0.06992003321647644f;
  t72_n191: return 0.09357676655054092f;
  t72_n192: return 0.017076730728149414f;
  t72_n193: return 0.05206622555851936f;
  t72_n194: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t72_n196;
  t72_n195: return -0.06044550985097885f;
  t72_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t72_n198;
  t72_n197: return 0.002767222933471203f;
  t72_n198: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t72_n200;
  t72_n199: return 0.03290214389562607f;
  t72_n200: return -0.027728626504540443f;
  t72_n201: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t72_n211;
  t72_n202: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t72_n204;
  t72_n203: return 0.01186963077634573f;
  t72_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t72_n206;
  t72_n205: return -0.003968730568885803f;
  t72_n206: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t72_n208;
  t72_n207: return 0.004695393610745668f;
  t72_n208: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n210;
  t72_n209: return 0.011789160780608654f;
  t72_n210: return 0.04082502797245979f;
  t72_n211: if (E.getNumReferences() >= 3318212608 /*6397.0*/) goto t72_n217;
  t72_n212: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t72_n214;
  t72_n213: return 0.012286447919905186f;
  t72_n214: if (E.getFractionNameInContext() >= 3210856220 /*0.8819444179534912*/) goto t72_n216;
  t72_n215: return 0.03875812143087387f;
  t72_n216: return -0.295769602060318f;
  t72_n217: return 0.08002658188343048f;
  t72_n218: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t72_n254;
  t72_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t72_n237;
  t72_n220: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t72_n224;
  t72_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n223;
  t72_n222: return -0.08893156051635742f;
  t72_n223: return -0.020714046433568f;
  t72_n224: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t72_n232;
  t72_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t72_n231;
  t72_n226: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t72_n230;
  t72_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n229;
  t72_n228: return -0.05458785593509674f;
  t72_n229: return -0.012219890020787716f;
  t72_n230: return 0.028905903920531273f;
  t72_n231: return 0.07005158066749573f;
  t72_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n236;
  t72_n233: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t72_n235;
  t72_n234: return -0.03323420137166977f;
  t72_n235: return -0.14538906514644623f;
  t72_n236: return 0.02704518660902977f;
  t72_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t72_n241;
  t72_n238: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t72_n240;
  t72_n239: return -0.19233117997646332f;
  t72_n240: return -0.021651336923241615f;
  t72_n241: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t72_n245;
  t72_n242: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n244;
  t72_n243: return 0.014901879243552685f;
  t72_n244: return 0.04136919230222702f;
  t72_n245: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t72_n249;
  t72_n246: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t72_n248;
  t72_n247: return 0.034685373306274414f;
  t72_n248: return -0.011581739410758018f;
  t72_n249: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t72_n251;
  t72_n250: return 0.04605426639318466f;
  t72_n251: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t72_n253;
  t72_n252: return 0.14880114793777466f;
  t72_n253: return 0.09074285626411438f;
  t72_n254: return -0.01150002982467413f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree73(const Example& E) {
  t73_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t73_n106;
  t73_n1: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t73_n73;
  t73_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t73_n70;
  t73_n3: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t73_n23;
  t73_n4: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t73_n16;
  t73_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t73_n7;
  t73_n6: return -0.04981663450598717f;
  t73_n7: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t73_n13;
  t73_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t73_n12;
  t73_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t73_n11;
  t73_n10: return -0.01755828782916069f;
  t73_n11: return 0.00898686982691288f;
  t73_n12: return 0.027196288108825684f;
  t73_n13: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t73_n15;
  t73_n14: return -0.0060792709700763226f;
  t73_n15: return -0.09657695144414902f;
  t73_n16: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t73_n20;
  t73_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t73_n19;
  t73_n18: return -0.06890421360731125f;
  t73_n19: return -0.04347182810306549f;
  t73_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n22;
  t73_n21: return -0.02031085267663002f;
  t73_n22: return 0.007861959747970104f;
  t73_n23: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t73_n25;
  t73_n24: return 0.014233840629458427f;
  t73_n25: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t73_n49;
  t73_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n28;
  t73_n27: return -0.04354546591639519f;
  t73_n28: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t73_n48;
  t73_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t73_n31;
  t73_n30: return -0.12335032224655151f;
  t73_n31: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t73_n33;
  t73_n32: return 0.5003851652145386f;
  t73_n33: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t73_n35;
  t73_n34: return 0.6098985075950623f;
  t73_n35: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t73_n47;
  t73_n36: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t73_n40;
  t73_n37: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t73_n39;
  t73_n38: return -0.10083610564470291f;
  t73_n39: return 1.1761845350265503f;
  t73_n40: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t73_n42;
  t73_n41: return -0.10194680839776993f;
  t73_n42: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t73_n44;
  t73_n43: return 0.8737509846687317f;
  t73_n44: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t73_n46;
  t73_n45: return -0.10086706280708313f;
  t73_n46: return 0.9996175169944763f;
  t73_n47: return -0.12995260953903198f;
  t73_n48: return -0.07243475317955017f;
  t73_n49: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t73_n69;
  t73_n50: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t73_n56;
  t73_n51: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t73_n53;
  t73_n52: return 0.004347482230514288f;
  t73_n53: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t73_n55;
  t73_n54: return -0.054193705320358276f;
  t73_n55: return 0.00016910943668335676f;
  t73_n56: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t73_n60;
  t73_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t73_n59;
  t73_n58: return -0.10144715756177902f;
  t73_n59: return 0.10357367247343063f;
  t73_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n62;
  t73_n61: return -0.02303328923881054f;
  t73_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t73_n64;
  t73_n63: return -0.059840548783540726f;
  t73_n64: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t73_n68;
  t73_n65: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t73_n67;
  t73_n66: return 0.08766338974237442f;
  t73_n67: return 0.12414827942848206f;
  t73_n68: return 0.05748986452817917f;
  t73_n69: return -0.028611522167921066f;
  t73_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t73_n72;
  t73_n71: return -0.10441387444734573f;
  t73_n72: return 0.0048584407195448875f;
  t73_n73: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t73_n93;
  t73_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t73_n82;
  t73_n75: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t73_n81;
  t73_n76: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t73_n78;
  t73_n77: return 0.006447385996580124f;
  t73_n78: if (E.getNumReferences() >= 3334835456 /*25286.5*/) goto t73_n80;
  t73_n79: return 0.02334633469581604f;
  t73_n80: return -0.1117279902100563f;
  t73_n81: return -0.0009640800999477506f;
  t73_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n84;
  t73_n83: return -0.07968689501285553f;
  t73_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t73_n90;
  t73_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n87;
  t73_n86: return -0.006042721681296825f;
  t73_n87: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t73_n89;
  t73_n88: return -0.03186482936143875f;
  t73_n89: return 0.014039518311619759f;
  t73_n90: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t73_n92;
  t73_n91: return 0.04302852228283882f;
  t73_n92: return 0.0009137805318459868f;
  t73_n93: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t73_n95;
  t73_n94: return 0.008432230912148952f;
  t73_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t73_n99;
  t73_n96: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t73_n98;
  t73_n97: return 0.01761571504175663f;
  t73_n98: return 0.0598205104470253f;
  t73_n99: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t73_n101;
  t73_n100: return 0.03310148045420647f;
  t73_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t73_n103;
  t73_n102: return -0.02186942845582962f;
  t73_n103: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t73_n105;
  t73_n104: return 0.08744215965270996f;
  t73_n105: return -0.05060736462473869f;
  t73_n106: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t73_n208;
  t73_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t73_n177;
  t73_n108: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t73_n170;
  t73_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n125;
  t73_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t73_n124;
  t73_n111: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t73_n119;
  t73_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t73_n114;
  t73_n113: return -0.11183249950408936f;
  t73_n114: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t73_n116;
  t73_n115: return -0.018268423154950142f;
  t73_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t73_n118;
  t73_n117: return -0.23711994290351868f;
  t73_n118: return -0.041107501834630966f;
  t73_n119: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t73_n123;
  t73_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t73_n122;
  t73_n121: return 0.0267327968031168f;
  t73_n122: return 0.10475239157676697f;
  t73_n123: return -0.08740564435720444f;
  t73_n124: return 0.04564954340457916f;
  t73_n125: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t73_n155;
  t73_n126: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t73_n146;
  t73_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t73_n145;
  t73_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n136;
  t73_n129: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t73_n133;
  t73_n130: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t73_n132;
  t73_n131: return 0.00613003084436059f;
  t73_n132: return 0.03188901022076607f;
  t73_n133: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t73_n135;
  t73_n134: return -0.020232729613780975f;
  t73_n135: return 0.04827321320772171f;
  t73_n136: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t73_n142;
  t73_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n139;
  t73_n138: return 0.007003704085946083f;
  t73_n139: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t73_n141;
  t73_n140: return 0.06474508345127106f;
  t73_n141: return -0.025998283177614212f;
  t73_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n144;
  t73_n143: return -0.21477921307086945f;
  t73_n144: return 0.0551726296544075f;
  t73_n145: return 0.028911828994750977f;
  t73_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n148;
  t73_n147: return -0.11660299450159073f;
  t73_n148: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t73_n152;
  t73_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t73_n151;
  t73_n150: return -0.0322967953979969f;
  t73_n151: return 0.039697274565696716f;
  t73_n152: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t73_n154;
  t73_n153: return -0.11633839458227158f;
  t73_n154: return 0.08033385872840881f;
  t73_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n157;
  t73_n156: return -0.10950282216072083f;
  t73_n157: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t73_n169;
  t73_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t73_n166;
  t73_n159: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t73_n163;
  t73_n160: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t73_n162;
  t73_n161: return 0.013008498586714268f;
  t73_n162: return -0.0301316287368536f;
  t73_n163: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t73_n165;
  t73_n164: return 0.060490258038043976f;
  t73_n165: return 0.01111049484461546f;
  t73_n166: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t73_n168;
  t73_n167: return -0.010955184698104858f;
  t73_n168: return 0.026429973542690277f;
  t73_n169: return 0.03186328709125519f;
  t73_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t73_n172;
  t73_n171: return 0.02134108357131481f;
  t73_n172: if (E.getNumReferences() >= 3334111488 /*23872.5*/) goto t73_n176;
  t73_n173: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t73_n175;
  t73_n174: return 0.07250925153493881f;
  t73_n175: return 0.14764437079429626f;
  t73_n176: return 0.050364911556243896f;
  t73_n177: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t73_n191;
  t73_n178: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t73_n190;
  t73_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t73_n183;
  t73_n180: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t73_n182;
  t73_n181: return 0.011132343672215939f;
  t73_n182: return 0.0556306354701519f;
  t73_n183: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t73_n185;
  t73_n184: return -0.12219560146331787f;
  t73_n185: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t73_n187;
  t73_n186: return 0.025369731709361076f;
  t73_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n189;
  t73_n188: return -0.46777379512786865f;
  t73_n189: return -0.02828373946249485f;
  t73_n190: return 0.06132253259420395f;
  t73_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t73_n207;
  t73_n192: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t73_n196;
  t73_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t73_n195;
  t73_n194: return -0.01265869289636612f;
  t73_n195: return 0.05471886321902275f;
  t73_n196: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t73_n206;
  t73_n197: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t73_n205;
  t73_n198: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t73_n200;
  t73_n199: return 0.08173603564500809f;
  t73_n200: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t73_n204;
  t73_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t73_n203;
  t73_n202: return 0.14287889003753662f;
  t73_n203: return 0.12303251028060913f;
  t73_n204: return 0.021020036190748215f;
  t73_n205: return 0.0909789651632309f;
  t73_n206: return -0.008483229205012321f;
  t73_n207: return 0.024250537157058716f;
  t73_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t73_n218;
  t73_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n215;
  t73_n210: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t73_n212;
  t73_n211: return 0.0011995723471045494f;
  t73_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t73_n214;
  t73_n213: return -0.014214739203453064f;
  t73_n214: return 0.027429649606347084f;
  t73_n215: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t73_n217;
  t73_n216: return 0.020844873040914536f;
  t73_n217: return -0.008493813686072826f;
  t73_n218: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t73_n238;
  t73_n219: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t73_n231;
  t73_n220: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t73_n230;
  t73_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t73_n225;
  t73_n222: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t73_n224;
  t73_n223: return -0.041864726692438126f;
  t73_n224: return 0.05602660030126572f;
  t73_n225: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t73_n229;
  t73_n226: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t73_n228;
  t73_n227: return -0.004310849588364363f;
  t73_n228: return 0.017483724281191826f;
  t73_n229: return 0.05247248336672783f;
  t73_n230: return -0.1010725200176239f;
  t73_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t73_n233;
  t73_n232: return -0.003610124345868826f;
  t73_n233: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t73_n237;
  t73_n234: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t73_n236;
  t73_n235: return 0.019364919513463974f;
  t73_n236: return -0.07655194401741028f;
  t73_n237: return 0.02570102922618389f;
  t73_n238: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t73_n252;
  t73_n239: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t73_n251;
  t73_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t73_n244;
  t73_n241: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t73_n243;
  t73_n242: return 0.012314686551690102f;
  t73_n243: return -0.1334502249956131f;
  t73_n244: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t73_n250;
  t73_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t73_n247;
  t73_n246: return 0.041912298649549484f;
  t73_n247: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t73_n249;
  t73_n248: return 0.10683734714984894f;
  t73_n249: return 0.08793220669031143f;
  t73_n250: return -0.10955479741096497f;
  t73_n251: return 0.043713755905628204f;
  t73_n252: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t73_n254;
  t73_n253: return 0.08992193639278412f;
  t73_n254: return -0.1672011762857437f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree74(const Example& E) {
  t74_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t74_n48;
  t74_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t74_n43;
  t74_n2: if (E.getNumReferences() >= 3317190656 /*5898.0*/) goto t74_n34;
  t74_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t74_n5;
  t74_n4: return -0.001062791678123176f;
  t74_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t74_n29;
  t74_n6: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t74_n12;
  t74_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n11;
  t74_n8: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t74_n10;
  t74_n9: return -0.2716217637062073f;
  t74_n10: return -0.06708414107561111f;
  t74_n11: return -0.017519088461995125f;
  t74_n12: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t74_n28;
  t74_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n15;
  t74_n14: return -0.025313911959528923f;
  t74_n15: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t74_n21;
  t74_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n18;
  t74_n17: return -0.020801961421966553f;
  t74_n18: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t74_n20;
  t74_n19: return 0.012742121703922749f;
  t74_n20: return -0.032995790243148804f;
  t74_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t74_n23;
  t74_n22: return -0.03507456183433533f;
  t74_n23: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t74_n27;
  t74_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t74_n26;
  t74_n25: return 0.025076471269130707f;
  t74_n26: return 0.03377765789628029f;
  t74_n27: return 0.09605629742145538f;
  t74_n28: return 0.04214794188737869f;
  t74_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t74_n33;
  t74_n30: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t74_n32;
  t74_n31: return 0.009704730473458767f;
  t74_n32: return 0.0432092547416687f;
  t74_n33: return 0.10490783303976059f;
  t74_n34: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t74_n38;
  t74_n35: if (E.getNumReferences() >= 3317212160 /*5908.5*/) goto t74_n37;
  t74_n36: return 0.1135588064789772f;
  t74_n37: return 0.021066877990961075f;
  t74_n38: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t74_n42;
  t74_n39: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t74_n41;
  t74_n40: return -0.027902089059352875f;
  t74_n41: return 0.12364301830530167f;
  t74_n42: return 0.04103269800543785f;
  t74_n43: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t74_n45;
  t74_n44: return -0.008310186676681042f;
  t74_n45: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t74_n47;
  t74_n46: return 0.006346292328089476f;
  t74_n47: return -0.037633925676345825f;
  t74_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t74_n236;
  t74_n49: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t74_n233;
  t74_n50: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t74_n162;
  t74_n51: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t74_n161;
  t74_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t74_n108;
  t74_n53: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t74_n83;
  t74_n54: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t74_n80;
  t74_n55: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t74_n67;
  t74_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t74_n62;
  t74_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t74_n61;
  t74_n58: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t74_n60;
  t74_n59: return -0.10219791531562805f;
  t74_n60: return -0.19298090040683746f;
  t74_n61: return -0.07677242159843445f;
  t74_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t74_n64;
  t74_n63: return -0.007038779091089964f;
  t74_n64: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t74_n66;
  t74_n65: return 0.004467518534511328f;
  t74_n66: return 0.05228208377957344f;
  t74_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n71;
  t74_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t74_n70;
  t74_n69: return -0.1167207732796669f;
  t74_n70: return 0.08625168353319168f;
  t74_n71: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t74_n75;
  t74_n72: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t74_n74;
  t74_n73: return 0.0951092541217804f;
  t74_n74: return 0.0015239089261740446f;
  t74_n75: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t74_n79;
  t74_n76: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t74_n78;
  t74_n77: return 0.013986271806061268f;
  t74_n78: return -0.13015176355838776f;
  t74_n79: return -0.03352372720837593f;
  t74_n80: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t74_n82;
  t74_n81: return -0.014868082478642464f;
  t74_n82: return -0.045336224138736725f;
  t74_n83: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t74_n107;
  t74_n84: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t74_n104;
  t74_n85: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t74_n97;
  t74_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n90;
  t74_n87: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t74_n89;
  t74_n88: return -0.11369480937719345f;
  t74_n89: return -0.017870379611849785f;
  t74_n90: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t74_n92;
  t74_n91: return 0.03224949911236763f;
  t74_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t74_n94;
  t74_n93: return -0.004632439464330673f;
  t74_n94: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t74_n96;
  t74_n95: return 0.016066985204815865f;
  t74_n96: return 0.05330895632505417f;
  t74_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t74_n99;
  t74_n98: return -0.07967641204595566f;
  t74_n99: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t74_n101;
  t74_n100: return 0.024262169376015663f;
  t74_n101: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t74_n103;
  t74_n102: return -0.0843718871474266f;
  t74_n103: return -0.012711979448795319f;
  t74_n104: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t74_n106;
  t74_n105: return 0.08805015683174133f;
  t74_n106: return -0.008665794506669044f;
  t74_n107: return 0.14751030504703522f;
  t74_n108: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t74_n148;
  t74_n109: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t74_n113;
  t74_n110: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t74_n112;
  t74_n111: return 0.00588969187811017f;
  t74_n112: return 0.015564528293907642f;
  t74_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t74_n141;
  t74_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t74_n116;
  t74_n115: return -0.007708538789302111f;
  t74_n116: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t74_n136;
  t74_n117: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t74_n119;
  t74_n118: return 0.005752846598625183f;
  t74_n119: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t74_n121;
  t74_n120: return 0.025963973253965378f;
  t74_n121: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t74_n135;
  t74_n122: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t74_n124;
  t74_n123: return -0.003448744770139456f;
  t74_n124: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t74_n126;
  t74_n125: return 0.057987280189991f;
  t74_n126: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t74_n134;
  t74_n127: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t74_n133;
  t74_n128: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t74_n130;
  t74_n129: return 0.07018215209245682f;
  t74_n130: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t74_n132;
  t74_n131: return 0.15720374882221222f;
  t74_n132: return 0.10515835881233215f;
  t74_n133: return 0.06723824888467789f;
  t74_n134: return 0.13322322070598602f;
  t74_n135: return 0.10334174335002899f;
  t74_n136: if (E.getNumReferences() >= 3271720960 /*130.5*/) goto t74_n140;
  t74_n137: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t74_n139;
  t74_n138: return 0.24817320704460144f;
  t74_n139: return 0.024737613275647163f;
  t74_n140: return 0.15507842600345612f;
  t74_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t74_n145;
  t74_n142: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t74_n144;
  t74_n143: return 0.014411559328436852f;
  t74_n144: return 0.018094312399625778f;
  t74_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t74_n147;
  t74_n146: return -0.04473263770341873f;
  t74_n147: return 0.027671249583363533f;
  t74_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t74_n150;
  t74_n149: return -0.08009305596351624f;
  t74_n150: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t74_n154;
  t74_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t74_n153;
  t74_n152: return -0.07005836814641953f;
  t74_n153: return 0.120979443192482f;
  t74_n154: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t74_n160;
  t74_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n157;
  t74_n156: return -0.024629177525639534f;
  t74_n157: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t74_n159;
  t74_n158: return 0.030323948711156845f;
  t74_n159: return 0.06386128813028336f;
  t74_n160: return -0.003168532159179449f;
  t74_n161: return 0.032731931656599045f;
  t74_n162: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t74_n230;
  t74_n163: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t74_n219;
  t74_n164: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t74_n216;
  t74_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t74_n189;
  t74_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t74_n174;
  t74_n167: if (E.getNumReferences() >= 3354093056 /*120524.0*/) goto t74_n173;
  t74_n168: if (E.getNumReferences() >= 3345303552 /*58696.0*/) goto t74_n172;
  t74_n169: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t74_n171;
  t74_n170: return -0.04222094267606735f;
  t74_n171: return -0.00011024565901607275f;
  t74_n172: return 0.07718737423419952f;
  t74_n173: return -0.043901000171899796f;
  t74_n174: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t74_n176;
  t74_n175: return 0.010156487114727497f;
  t74_n176: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t74_n178;
  t74_n177: return 0.07686594873666763f;
  t74_n178: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t74_n180;
  t74_n179: return -0.04433530941605568f;
  t74_n180: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t74_n188;
  t74_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t74_n183;
  t74_n182: return 0.10935366898775101f;
  t74_n183: if (E.getNumReferences() >= 3261857792 /*59.0*/) goto t74_n185;
  t74_n184: return -0.10155407339334488f;
  t74_n185: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t74_n187;
  t74_n186: return 0.3214721977710724f;
  t74_n187: return -0.14219559729099274f;
  t74_n188: return 0.03780032694339752f;
  t74_n189: if (E.getNumReferences() >= 3356822208 /*152619.0*/) goto t74_n213;
  t74_n190: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t74_n212;
  t74_n191: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t74_n195;
  t74_n192: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t74_n194;
  t74_n193: return 0.03665198013186455f;
  t74_n194: return 0.005755984224379063f;
  t74_n195: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t74_n201;
  t74_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t74_n198;
  t74_n197: return 0.0010719775455072522f;
  t74_n198: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t74_n200;
  t74_n199: return 0.011987020261585712f;
  t74_n200: return 0.04338904470205307f;
  t74_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t74_n203;
  t74_n202: return -0.03384548798203468f;
  t74_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t74_n205;
  t74_n204: return 0.026946716010570526f;
  t74_n205: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t74_n211;
  t74_n206: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t74_n210;
  t74_n207: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t74_n209;
  t74_n208: return 0.12213048338890076f;
  t74_n209: return 0.029371406883001328f;
  t74_n210: return 0.10730478167533875f;
  t74_n211: return 0.04080953076481819f;
  t74_n212: return -0.0329459086060524f;
  t74_n213: if (E.getNumReferences() >= 3365016304 /*299159.5*/) goto t74_n215;
  t74_n214: return 0.06916239857673645f;
  t74_n215: return 0.03814125433564186f;
  t74_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n218;
  t74_n217: return -0.08804195374250412f;
  t74_n218: return 0.03014920838177204f;
  t74_n219: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t74_n229;
  t74_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n222;
  t74_n221: return -0.016423214226961136f;
  t74_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t74_n224;
  t74_n223: return -0.05105886235833168f;
  t74_n224: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t74_n226;
  t74_n225: return -0.01601145602762699f;
  t74_n226: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t74_n228;
  t74_n227: return 0.08133838325738907f;
  t74_n228: return 0.010217494331300259f;
  t74_n229: return -0.05516062676906586f;
  t74_n230: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t74_n232;
  t74_n231: return 0.02443871460855007f;
  t74_n232: return 0.051204320043325424f;
  t74_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n235;
  t74_n234: return -0.16204077005386353f;
  t74_n235: return -0.09116345643997192f;
  t74_n236: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t74_n254;
  t74_n237: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t74_n253;
  t74_n238: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t74_n242;
  t74_n239: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t74_n241;
  t74_n240: return 0.0035548999439924955f;
  t74_n241: return 0.0348152294754982f;
  t74_n242: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t74_n248;
  t74_n243: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t74_n245;
  t74_n244: return 0.03156503289937973f;
  t74_n245: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t74_n247;
  t74_n246: return -0.0851164162158966f;
  t74_n247: return 0.017789533361792564f;
  t74_n248: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t74_n252;
  t74_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t74_n251;
  t74_n250: return -0.027054650709033012f;
  t74_n251: return 0.033104829490184784f;
  t74_n252: return 0.016255537047982216f;
  t74_n253: return -0.038852300494909286f;
  t74_n254: return 0.057434018701314926f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree75(const Example& E) {
  t75_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t75_n146;
  t75_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t75_n135;
  t75_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t75_n16;
  t75_n3: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t75_n7;
  t75_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n6;
  t75_n5: return -0.0067369830794632435f;
  t75_n6: return 0.0014050399186089635f;
  t75_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n11;
  t75_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t75_n10;
  t75_n9: return -0.05197471007704735f;
  t75_n10: return 0.01812909170985222f;
  t75_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t75_n15;
  t75_n12: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t75_n14;
  t75_n13: return 0.01999494805932045f;
  t75_n14: return 0.03488783910870552f;
  t75_n15: return -0.003999622538685799f;
  t75_n16: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t75_n134;
  t75_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t75_n93;
  t75_n18: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t75_n30;
  t75_n19: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t75_n23;
  t75_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t75_n22;
  t75_n21: return -0.02306055836379528f;
  t75_n22: return 0.016854479908943176f;
  t75_n23: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t75_n25;
  t75_n24: return -0.051213983446359634f;
  t75_n25: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t75_n27;
  t75_n26: return 0.01937681995332241f;
  t75_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t75_n29;
  t75_n28: return -0.004379733465611935f;
  t75_n29: return 0.11715234071016312f;
  t75_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t75_n60;
  t75_n31: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t75_n35;
  t75_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t75_n34;
  t75_n33: return 0.015654228627681732f;
  t75_n34: return 0.060219865292310715f;
  t75_n35: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t75_n39;
  t75_n36: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t75_n38;
  t75_n37: return -0.022954484447836876f;
  t75_n38: return -0.107254758477211f;
  t75_n39: if (E.getNumReferences() >= 3338505216 /*32454.0*/) goto t75_n41;
  t75_n40: return -0.006885842885822058f;
  t75_n41: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t75_n49;
  t75_n42: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t75_n46;
  t75_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n45;
  t75_n44: return -0.06366857141256332f;
  t75_n45: return 0.05867268517613411f;
  t75_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n48;
  t75_n47: return -0.21990960836410522f;
  t75_n48: return 0.13435152173042297f;
  t75_n49: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t75_n51;
  t75_n50: return 0.021794775500893593f;
  t75_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n53;
  t75_n52: return -0.16379031538963318f;
  t75_n53: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t75_n55;
  t75_n54: return -0.03299134597182274f;
  t75_n55: if (E.getNumReferences() >= 3362781216 /*245728.5*/) goto t75_n59;
  t75_n56: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t75_n58;
  t75_n57: return 0.1165243536233902f;
  t75_n58: return 0.08792002499103546f;
  t75_n59: return -0.22694531083106995f;
  t75_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t75_n68;
  t75_n61: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t75_n67;
  t75_n62: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t75_n64;
  t75_n63: return 0.0032961349934339523f;
  t75_n64: if (E.getNumReferences() >= 3316470784 /*5546.5*/) goto t75_n66;
  t75_n65: return 0.05090288072824478f;
  t75_n66: return -0.01710537262260914f;
  t75_n67: return -0.02568986639380455f;
  t75_n68: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t75_n86;
  t75_n69: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t75_n71;
  t75_n70: return 0.01776563748717308f;
  t75_n71: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t75_n77;
  t75_n72: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t75_n76;
  t75_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t75_n75;
  t75_n74: return 0.0003872342058457434f;
  t75_n75: return 0.08723459392786026f;
  t75_n76: return 0.005975387524813414f;
  t75_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t75_n83;
  t75_n78: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t75_n82;
  t75_n79: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t75_n81;
  t75_n80: return -0.008605851791799068f;
  t75_n81: return 0.05277803912758827f;
  t75_n82: return -0.004327040631324053f;
  t75_n83: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t75_n85;
  t75_n84: return 0.03298678249120712f;
  t75_n85: return 0.15083342790603638f;
  t75_n86: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t75_n92;
  t75_n87: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t75_n91;
  t75_n88: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t75_n90;
  t75_n89: return -0.10828885436058044f;
  t75_n90: return 0.15550115704536438f;
  t75_n91: return 0.06278681755065918f;
  t75_n92: return 0.013306746259331703f;
  t75_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t75_n97;
  t75_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t75_n96;
  t75_n95: return -0.05425821989774704f;
  t75_n96: return 0.00027869187761098146f;
  t75_n97: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t75_n99;
  t75_n98: return 0.038305673748254776f;
  t75_n99: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t75_n133;
  t75_n100: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t75_n106;
  t75_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n103;
  t75_n102: return -0.04775720834732056f;
  t75_n103: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t75_n105;
  t75_n104: return 0.02392776869237423f;
  t75_n105: return -0.00586438924074173f;
  t75_n106: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t75_n112;
  t75_n107: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t75_n111;
  t75_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n110;
  t75_n109: return -0.10881736874580383f;
  t75_n110: return 0.04611080512404442f;
  t75_n111: return 0.00841303076595068f;
  t75_n112: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t75_n116;
  t75_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n115;
  t75_n114: return -0.09036267548799515f;
  t75_n115: return 0.05255269259214401f;
  t75_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t75_n118;
  t75_n117: return 0.00252702459692955f;
  t75_n118: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t75_n132;
  t75_n119: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t75_n131;
  t75_n120: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t75_n130;
  t75_n121: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t75_n129;
  t75_n122: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t75_n128;
  t75_n123: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t75_n127;
  t75_n124: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t75_n126;
  t75_n125: return 0.050660502165555954f;
  t75_n126: return -0.09428320825099945f;
  t75_n127: return 0.050864268094301224f;
  t75_n128: return -0.14571061730384827f;
  t75_n129: return 0.056520864367485046f;
  t75_n130: return -0.10747475177049637f;
  t75_n131: return 0.05687014386057854f;
  t75_n132: return -0.13151085376739502f;
  t75_n133: return -0.012871936894953251f;
  t75_n134: return 0.03425312042236328f;
  t75_n135: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t75_n145;
  t75_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n138;
  t75_n137: return -0.0072065419517457485f;
  t75_n138: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t75_n144;
  t75_n139: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t75_n143;
  t75_n140: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t75_n142;
  t75_n141: return 0.01310131698846817f;
  t75_n142: return 0.023834899067878723f;
  t75_n143: return -0.024438854306936264f;
  t75_n144: return 0.018274059519171715f;
  t75_n145: return 0.07152165472507477f;
  t75_n146: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t75_n246;
  t75_n147: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t75_n201;
  t75_n148: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t75_n198;
  t75_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t75_n181;
  t75_n150: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t75_n180;
  t75_n151: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t75_n165;
  t75_n152: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t75_n162;
  t75_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace))) goto t75_n161;
  t75_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n160;
  t75_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t75_n159;
  t75_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t75_n158;
  t75_n157: return -0.10673105716705322f;
  t75_n158: return -0.18647116422653198f;
  t75_n159: return -0.07351981103420258f;
  t75_n160: return -0.00781989935785532f;
  t75_n161: return 0.02147924154996872f;
  t75_n162: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t75_n164;
  t75_n163: return 0.0012396987294778228f;
  t75_n164: return -0.14867597818374634f;
  t75_n165: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t75_n179;
  t75_n166: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t75_n174;
  t75_n167: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t75_n173;
  t75_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n170;
  t75_n169: return -0.06931557506322861f;
  t75_n170: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t75_n172;
  t75_n171: return 0.0426197275519371f;
  t75_n172: return 0.018671521916985512f;
  t75_n173: return 0.06579270213842392f;
  t75_n174: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t75_n176;
  t75_n175: return 0.034743424504995346f;
  t75_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t75_n178;
  t75_n177: return -0.17527104914188385f;
  t75_n178: return -0.02683376893401146f;
  t75_n179: return 0.1341850459575653f;
  t75_n180: return -0.044988781213760376f;
  t75_n181: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t75_n197;
  t75_n182: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t75_n188;
  t75_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t75_n185;
  t75_n184: return -0.11556033790111542f;
  t75_n185: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t75_n187;
  t75_n186: return 0.008844412863254547f;
  t75_n187: return 0.014236782677471638f;
  t75_n188: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t75_n190;
  t75_n189: return -0.03985830023884773f;
  t75_n190: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t75_n192;
  t75_n191: return 0.003929443191736937f;
  t75_n192: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t75_n196;
  t75_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n195;
  t75_n194: return -0.45666950941085815f;
  t75_n195: return 0.12256933003664017f;
  t75_n196: return 0.028210068121552467f;
  t75_n197: return -0.008213505148887634f;
  t75_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t75_n200;
  t75_n199: return 0.004686519503593445f;
  t75_n200: return 0.025965284556150436f;
  t75_n201: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t75_n245;
  t75_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t75_n242;
  t75_n203: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t75_n241;
  t75_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t75_n234;
  t75_n205: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t75_n229;
  t75_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t75_n208;
  t75_n207: return -0.0002517095999792218f;
  t75_n208: if (E.getNumReferences() >= 3357343040 /*160757.0*/) goto t75_n228;
  t75_n209: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t75_n211;
  t75_n210: return 0.007134441286325455f;
  t75_n211: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t75_n215;
  t75_n212: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t75_n214;
  t75_n213: return -0.006001174449920654f;
  t75_n214: return 0.03369489312171936f;
  t75_n215: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t75_n217;
  t75_n216: return -0.01177812460809946f;
  t75_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t75_n219;
  t75_n218: return 0.027875376865267754f;
  t75_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t75_n227;
  t75_n220: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t75_n224;
  t75_n221: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t75_n223;
  t75_n222: return 0.11808086186647415f;
  t75_n223: return -0.04078303277492523f;
  t75_n224: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t75_n226;
  t75_n225: return 0.10475604236125946f;
  t75_n226: return -0.7709452509880066f;
  t75_n227: return 0.03697383031249046f;
  t75_n228: return 0.05943522974848747f;
  t75_n229: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t75_n233;
  t75_n230: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t75_n232;
  t75_n231: return -0.014134753495454788f;
  t75_n232: return 0.10850478708744049f;
  t75_n233: return -0.09078659117221832f;
  t75_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t75_n236;
  t75_n235: return -0.061199914664030075f;
  t75_n236: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t75_n240;
  t75_n237: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t75_n239;
  t75_n238: return 0.02986270934343338f;
  t75_n239: return 0.07009503245353699f;
  t75_n240: return 0.01313652005046606f;
  t75_n241: return -0.047842178493738174f;
  t75_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t75_n244;
  t75_n243: return -0.08639080822467804f;
  t75_n244: return 0.023469913750886917f;
  t75_n245: return 0.028739577159285545f;
  t75_n246: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t75_n252;
  t75_n247: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t75_n251;
  t75_n248: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t75_n250;
  t75_n249: return 0.020171338692307472f;
  t75_n250: return 0.004728218074887991f;
  t75_n251: return 0.03237592428922653f;
  t75_n252: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t75_n254;
  t75_n253: return 0.044719770550727844f;
  t75_n254: return 0.022424008697271347f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree76(const Example& E) {
  t76_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t76_n58;
  t76_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t76_n19;
  t76_n2: if (E.getNumReferences() >= 3301818368 /*1646.0*/) goto t76_n16;
  t76_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t76_n5;
  t76_n4: return -0.0017215897096320987f;
  t76_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t76_n11;
  t76_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n10;
  t76_n7: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t76_n9;
  t76_n8: return -0.0259876549243927f;
  t76_n9: return 0.025782162323594093f;
  t76_n10: return 0.006532543804496527f;
  t76_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t76_n15;
  t76_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t76_n14;
  t76_n13: return 0.007974939420819283f;
  t76_n14: return 0.03997628390789032f;
  t76_n15: return 0.10356408357620239f;
  t76_n16: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t76_n18;
  t76_n17: return 0.017796602100133896f;
  t76_n18: return 0.03967147320508957f;
  t76_n19: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t76_n23;
  t76_n20: if (E.getFileProximityDistanceCost() >= 3255697408 /*35.5*/) goto t76_n22;
  t76_n21: return -0.0077065653167665005f;
  t76_n22: return -0.13945895433425903f;
  t76_n23: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t76_n57;
  t76_n24: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t76_n56;
  t76_n25: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t76_n55;
  t76_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t76_n28;
  t76_n27: return -0.16630050539970398f;
  t76_n28: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t76_n54;
  t76_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n51;
  t76_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t76_n42;
  t76_n31: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t76_n33;
  t76_n32: return 0.014748971909284592f;
  t76_n33: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t76_n39;
  t76_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n36;
  t76_n35: return -0.17481648921966553f;
  t76_n36: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t76_n38;
  t76_n37: return 0.05999472737312317f;
  t76_n38: return -0.019553298130631447f;
  t76_n39: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t76_n41;
  t76_n40: return 0.00828997790813446f;
  t76_n41: return 0.06639424711465836f;
  t76_n42: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t76_n48;
  t76_n43: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t76_n45;
  t76_n44: return 0.0184167493134737f;
  t76_n45: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t76_n47;
  t76_n46: return -0.05382019281387329f;
  t76_n47: return 0.19035032391548157f;
  t76_n48: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t76_n50;
  t76_n49: return 0.011627529747784138f;
  t76_n50: return -0.02110833302140236f;
  t76_n51: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t76_n53;
  t76_n52: return 0.013956933282315731f;
  t76_n53: return 0.04487402364611626f;
  t76_n54: return -0.04083048179745674f;
  t76_n55: return -0.21717113256454468f;
  t76_n56: return -0.06484151631593704f;
  t76_n57: return -0.035975903272628784f;
  t76_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t76_n232;
  t76_n59: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t76_n159;
  t76_n60: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t76_n158;
  t76_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t76_n125;
  t76_n62: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t76_n116;
  t76_n63: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t76_n113;
  t76_n64: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t76_n78;
  t76_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace))) goto t76_n73;
  t76_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t76_n72;
  t76_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t76_n71;
  t76_n68: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t76_n70;
  t76_n69: return -0.09809662401676178f;
  t76_n70: return -0.17867478728294373f;
  t76_n71: return -0.10473156720399857f;
  t76_n72: return -0.006899653002619743f;
  t76_n73: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t76_n75;
  t76_n74: return 0.0019996571354568005f;
  t76_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t76_n77;
  t76_n76: return 0.0475919172167778f;
  t76_n77: return 0.1021021157503128f;
  t76_n78: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t76_n112;
  t76_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n83;
  t76_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t76_n82;
  t76_n81: return -0.11640388518571854f;
  t76_n82: return 0.08442378044128418f;
  t76_n83: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t76_n95;
  t76_n84: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t76_n86;
  t76_n85: return 0.09408837556838989f;
  t76_n86: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t76_n90;
  t76_n87: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t76_n89;
  t76_n88: return 0.002535431180149317f;
  t76_n89: return 0.013029581867158413f;
  t76_n90: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t76_n92;
  t76_n91: return -0.0104170897975564f;
  t76_n92: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t76_n94;
  t76_n93: return 0.4648086428642273f;
  t76_n94: return -0.1651851087808609f;
  t76_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t76_n99;
  t76_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n98;
  t76_n97: return -0.0009271121816709638f;
  t76_n98: return 0.03139038383960724f;
  t76_n99: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t76_n111;
  t76_n100: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t76_n108;
  t76_n101: if (E.getFractionNameInContext() >= 3197626751 /*0.2967033088207245*/) goto t76_n103;
  t76_n102: return 0.082688108086586f;
  t76_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t76_n105;
  t76_n104: return 0.028908319771289825f;
  t76_n105: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t76_n107;
  t76_n106: return 0.1118769571185112f;
  t76_n107: return -0.07020515203475952f;
  t76_n108: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t76_n110;
  t76_n109: return 0.08748947083950043f;
  t76_n110: return 0.2269800901412964f;
  t76_n111: return 0.029593417420983315f;
  t76_n112: return -0.1432214379310608f;
  t76_n113: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t76_n115;
  t76_n114: return -0.014560898765921593f;
  t76_n115: return -0.04224705696105957f;
  t76_n116: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t76_n122;
  t76_n117: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t76_n119;
  t76_n118: return 0.011939614079892635f;
  t76_n119: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t76_n121;
  t76_n120: return 0.08225000649690628f;
  t76_n121: return -0.0016895747976377606f;
  t76_n122: if (E.getFractionNameInContext() >= 3188603108 /*0.1388888955116272*/) goto t76_n124;
  t76_n123: return 0.06423050165176392f;
  t76_n124: return 0.14293965697288513f;
  t76_n125: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t76_n137;
  t76_n126: if (E.getFractionNameInContext() >= 3193337030 /*0.20942983031272888*/) goto t76_n130;
  t76_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t76_n129;
  t76_n128: return 0.0056691849604249f;
  t76_n129: return 0.0139167420566082f;
  t76_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t76_n136;
  t76_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t76_n133;
  t76_n132: return -0.007829999551177025f;
  t76_n133: if (E.getNumReferences() >= 3282419712 /*331.5*/) goto t76_n135;
  t76_n134: return 0.02246239222586155f;
  t76_n135: return 0.20851945877075195f;
  t76_n136: return 0.014768454246222973f;
  t76_n137: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t76_n155;
  t76_n138: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t76_n140;
  t76_n139: return -0.20475341379642487f;
  t76_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t76_n142;
  t76_n141: return -0.06915709376335144f;
  t76_n142: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t76_n144;
  t76_n143: return -0.06129121407866478f;
  t76_n144: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t76_n152;
  t76_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n147;
  t76_n146: return -0.020040828734636307f;
  t76_n147: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t76_n151;
  t76_n148: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t76_n150;
  t76_n149: return 0.11397317051887512f;
  t76_n150: return 0.026751095429062843f;
  t76_n151: return 0.059119343757629395f;
  t76_n152: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t76_n154;
  t76_n153: return 0.03354388102889061f;
  t76_n154: return -0.030647272244095802f;
  t76_n155: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t76_n157;
  t76_n156: return 0.1191881000995636f;
  t76_n157: return 0.0948372483253479f;
  t76_n158: return 0.030676977708935738f;
  t76_n159: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t76_n223;
  t76_n160: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t76_n170;
  t76_n161: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t76_n165;
  t76_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t76_n164;
  t76_n163: return 0.007859385572373867f;
  t76_n164: return 0.03133697807788849f;
  t76_n165: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t76_n167;
  t76_n166: return 0.04299791529774666f;
  t76_n167: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t76_n169;
  t76_n168: return 0.1257946938276291f;
  t76_n169: return 0.06860829144716263f;
  t76_n170: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t76_n222;
  t76_n171: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t76_n215;
  t76_n172: if (E.getFractionNameInContext() >= 3173466330 /*0.040833331644535065*/) goto t76_n174;
  t76_n173: return 2.45273756980896f;
  t76_n174: if (E.getNumReferences() >= 3320626176 /*7575.5*/) goto t76_n212;
  t76_n175: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t76_n179;
  t76_n176: if (E.getNumReferences() >= 3319284736 /*6920.5*/) goto t76_n178;
  t76_n177: return -0.020933346822857857f;
  t76_n178: return -0.42749086022377014f;
  t76_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t76_n203;
  t76_n180: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t76_n184;
  t76_n181: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t76_n183;
  t76_n182: return -0.002706253668293357f;
  t76_n183: return 0.02211897261440754f;
  t76_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion))) goto t76_n200;
  t76_n185: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t76_n199;
  t76_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t76_n188;
  t76_n187: return 0.005316158756613731f;
  t76_n188: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t76_n190;
  t76_n189: return 0.028287876397371292f;
  t76_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t76_n192;
  t76_n191: return 0.02213222347199917f;
  t76_n192: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t76_n194;
  t76_n193: return 0.031121309846639633f;
  t76_n194: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t76_n198;
  t76_n195: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t76_n197;
  t76_n196: return 0.1098584458231926f;
  t76_n197: return 0.04565230384469032f;
  t76_n198: return 0.09785666316747665f;
  t76_n199: return -0.028046857565641403f;
  t76_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t76_n202;
  t76_n201: return -0.03374385088682175f;
  t76_n202: return 0.03322446718811989f;
  t76_n203: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t76_n205;
  t76_n204: return -0.05595921725034714f;
  t76_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t76_n207;
  t76_n206: return 0.005689991172403097f;
  t76_n207: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t76_n211;
  t76_n208: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t76_n210;
  t76_n209: return 0.1454775482416153f;
  t76_n210: return 0.0905475988984108f;
  t76_n211: return 0.026396222412586212f;
  t76_n212: if (E.getNumReferences() >= 3320842240 /*7681.0*/) goto t76_n214;
  t76_n213: return 0.11717686802148819f;
  t76_n214: return 0.016784798353910446f;
  t76_n215: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t76_n221;
  t76_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n218;
  t76_n217: return -0.021242186427116394f;
  t76_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t76_n220;
  t76_n219: return -0.04944739490747452f;
  t76_n220: return 0.043821655213832855f;
  t76_n221: return -0.055472198873758316f;
  t76_n222: return 0.026709062978625298f;
  t76_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n227;
  t76_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t76_n226;
  t76_n225: return -0.15390627086162567f;
  t76_n226: return 0.008450840599834919f;
  t76_n227: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t76_n231;
  t76_n228: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t76_n230;
  t76_n229: return -0.022414401173591614f;
  t76_n230: return 0.023823466151952744f;
  t76_n231: return 0.04386962950229645f;
  t76_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t76_n238;
  t76_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t76_n235;
  t76_n234: return 0.002167745493352413f;
  t76_n235: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t76_n237;
  t76_n236: return -0.046541906893253326f;
  t76_n237: return 0.06877851486206055f;
  t76_n238: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t76_n254;
  t76_n239: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t76_n251;
  t76_n240: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t76_n244;
  t76_n241: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t76_n243;
  t76_n242: return 0.01082882471382618f;
  t76_n243: return 0.03535402938723564f;
  t76_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t76_n246;
  t76_n245: return 0.01744028367102146f;
  t76_n246: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t76_n250;
  t76_n247: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t76_n249;
  t76_n248: return 0.12081676721572876f;
  t76_n249: return 0.12286108732223511f;
  t76_n250: return 0.0020318888127803802f;
  t76_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t76_n253;
  t76_n252: return -0.017406081780791283f;
  t76_n253: return 0.016875959932804108f;
  t76_n254: return 0.023393960669636726f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree77(const Example& E) {
  t77_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t77_n106;
  t77_n1: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t77_n69;
  t77_n2: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t77_n64;
  t77_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t77_n11;
  t77_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t77_n8;
  t77_n5: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t77_n7;
  t77_n6: return 0.0011320925550535321f;
  t77_n7: return -0.010112243704497814f;
  t77_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t77_n10;
  t77_n9: return -0.06773241609334946f;
  t77_n10: return -0.05162927135825157f;
  t77_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t77_n13;
  t77_n12: return -0.09926263988018036f;
  t77_n13: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t77_n21;
  t77_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t77_n16;
  t77_n15: return 0.1345115751028061f;
  t77_n16: if (E.getFractionNameInContext() >= 3182307328 /*0.0850372314453125*/) goto t77_n18;
  t77_n17: return 0.16555023193359375f;
  t77_n18: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t77_n20;
  t77_n19: return -0.4973312318325043f;
  t77_n20: return 0.23901787400245667f;
  t77_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t77_n49;
  t77_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n24;
  t77_n23: return -0.022196628153324127f;
  t77_n24: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t77_n40;
  t77_n25: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t77_n37;
  t77_n26: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t77_n30;
  t77_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n29;
  t77_n28: return -0.020620975643396378f;
  t77_n29: return 0.011183620430529118f;
  t77_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n32;
  t77_n31: return -0.001610844163224101f;
  t77_n32: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t77_n36;
  t77_n33: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t77_n35;
  t77_n34: return 0.0302076768130064f;
  t77_n35: return -0.01943589188158512f;
  t77_n36: return 0.03447171673178673f;
  t77_n37: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t77_n39;
  t77_n38: return -0.0009269930887967348f;
  t77_n39: return -0.05150406062602997f;
  t77_n40: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t77_n42;
  t77_n41: return 0.03504427149891853f;
  t77_n42: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t77_n48;
  t77_n43: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t77_n47;
  t77_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n46;
  t77_n45: return -0.04722519963979721f;
  t77_n46: return 0.009179881773889065f;
  t77_n47: return 0.010057748295366764f;
  t77_n48: return -0.013708489947021008f;
  t77_n49: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t77_n53;
  t77_n50: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t77_n52;
  t77_n51: return 0.03676402196288109f;
  t77_n52: return -0.029881784692406654f;
  t77_n53: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t77_n55;
  t77_n54: return -0.05747877061367035f;
  t77_n55: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t77_n63;
  t77_n56: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t77_n62;
  t77_n57: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t77_n61;
  t77_n58: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t77_n60;
  t77_n59: return 0.011204355396330357f;
  t77_n60: return 0.0776289775967598f;
  t77_n61: return -0.1170605942606926f;
  t77_n62: return 0.08301091194152832f;
  t77_n63: return -0.019634705036878586f;
  t77_n64: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t77_n68;
  t77_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t77_n67;
  t77_n66: return 0.014200911857187748f;
  t77_n67: return 0.05358985438942909f;
  t77_n68: return -0.0003616895410232246f;
  t77_n69: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t77_n97;
  t77_n70: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t77_n90;
  t77_n71: if (E.getNumReferences() >= 3275718656 /*191.5*/) goto t77_n89;
  t77_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t77_n86;
  t77_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n75;
  t77_n74: return -0.04409098997712135f;
  t77_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t77_n77;
  t77_n76: return -0.0031058844178915024f;
  t77_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t77_n79;
  t77_n78: return 0.012547432444989681f;
  t77_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t77_n81;
  t77_n80: return -0.03378303349018097f;
  t77_n81: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t77_n83;
  t77_n82: return 0.0162554569542408f;
  t77_n83: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t77_n85;
  t77_n84: return 0.2142341285943985f;
  t77_n85: return 0.07414814084768295f;
  t77_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t77_n88;
  t77_n87: return -0.0117747588083148f;
  t77_n88: return 0.051925960928201675f;
  t77_n89: return 0.031524259597063065f;
  t77_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t77_n94;
  t77_n91: if (E.getNumReferences() >= 3336405248 /*28352.5*/) goto t77_n93;
  t77_n92: return 0.0014445974957197905f;
  t77_n93: return -0.041085392236709595f;
  t77_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n96;
  t77_n95: return -0.0901099145412445f;
  t77_n96: return 0.01268976554274559f;
  t77_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t77_n99;
  t77_n98: return 0.0167329553514719f;
  t77_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t77_n101;
  t77_n100: return -0.020435206592082977f;
  t77_n101: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t77_n103;
  t77_n102: return 0.0654030591249466f;
  t77_n103: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t77_n105;
  t77_n104: return 0.08693089336156845f;
  t77_n105: return 0.12841925024986267f;
  t77_n106: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t77_n214;
  t77_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t77_n185;
  t77_n108: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t77_n176;
  t77_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n121;
  t77_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t77_n120;
  t77_n111: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t77_n115;
  t77_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t77_n114;
  t77_n113: return -0.09188435971736908f;
  t77_n114: return -0.01821650005877018f;
  t77_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t77_n119;
  t77_n116: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t77_n118;
  t77_n117: return 0.02077125944197178f;
  t77_n118: return -0.14434446394443512f;
  t77_n119: return 0.0918223038315773f;
  t77_n120: return 0.033078309148550034f;
  t77_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t77_n155;
  t77_n122: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t77_n140;
  t77_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t77_n137;
  t77_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n130;
  t77_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t77_n129;
  t77_n126: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t77_n128;
  t77_n127: return 0.004863514099270105f;
  t77_n128: return 0.028418395668268204f;
  t77_n129: return -0.013527999632060528f;
  t77_n130: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t77_n134;
  t77_n131: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t77_n133;
  t77_n132: return 0.008867972530424595f;
  t77_n133: return 0.039285674691200256f;
  t77_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n136;
  t77_n135: return -0.23388324677944183f;
  t77_n136: return 0.04768522456288338f;
  t77_n137: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t77_n139;
  t77_n138: return 0.027601057663559914f;
  t77_n139: return -0.02011210098862648f;
  t77_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t77_n154;
  t77_n141: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t77_n153;
  t77_n142: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t77_n152;
  t77_n143: if (E.getNumReferences() >= 3287695360 /*492.5*/) goto t77_n145;
  t77_n144: return 0.001350689330138266f;
  t77_n145: if (E.getNumReferences() >= 3295567872 /*953.5*/) goto t77_n151;
  t77_n146: if (E.getNumReferences() >= 3294674944 /*899.0*/) goto t77_n148;
  t77_n147: return 0.07596437633037567f;
  t77_n148: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t77_n150;
  t77_n149: return 0.1274406462907791f;
  t77_n150: return 0.11301657557487488f;
  t77_n151: return 0.025492805987596512f;
  t77_n152: return -0.025354042649269104f;
  t77_n153: return 0.030087871477007866f;
  t77_n154: return 0.016344774514436722f;
  t77_n155: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t77_n161;
  t77_n156: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t77_n160;
  t77_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t77_n159;
  t77_n158: return -0.022969355806708336f;
  t77_n159: return 0.114238440990448f;
  t77_n160: return -0.06535903364419937f;
  t77_n161: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t77_n167;
  t77_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t77_n164;
  t77_n163: return -0.010816991329193115f;
  t77_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n166;
  t77_n165: return -0.256021648645401f;
  t77_n166: return 0.02604440040886402f;
  t77_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t77_n169;
  t77_n168: return 0.014037338085472584f;
  t77_n169: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t77_n171;
  t77_n170: return 0.0266929492354393f;
  t77_n171: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t77_n175;
  t77_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t77_n174;
  t77_n173: return -0.034395504742860794f;
  t77_n174: return 0.11352191865444183f;
  t77_n175: return 0.07489890605211258f;
  t77_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t77_n178;
  t77_n177: return 0.01960446871817112f;
  t77_n178: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t77_n184;
  t77_n179: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t77_n181;
  t77_n180: return 0.070024773478508f;
  t77_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t77_n183;
  t77_n182: return 0.10399763286113739f;
  t77_n183: return 0.14377610385417938f;
  t77_n184: return 0.046530842781066895f;
  t77_n185: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t77_n197;
  t77_n186: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t77_n196;
  t77_n187: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t77_n193;
  t77_n188: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t77_n192;
  t77_n189: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t77_n191;
  t77_n190: return 0.004414119757711887f;
  t77_n191: return 0.08140316605567932f;
  t77_n192: return 0.05479481816291809f;
  t77_n193: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t77_n195;
  t77_n194: return -0.10837003588676453f;
  t77_n195: return 0.010731061920523643f;
  t77_n196: return 0.05734270438551903f;
  t77_n197: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t77_n213;
  t77_n198: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t77_n202;
  t77_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t77_n201;
  t77_n200: return -0.013118745759129524f;
  t77_n201: return 0.051093120127916336f;
  t77_n202: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t77_n212;
  t77_n203: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t77_n211;
  t77_n204: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t77_n206;
  t77_n205: return 0.07992631196975708f;
  t77_n206: if (E.getFractionNameInContext() >= 3175926989 /*0.05000000074505806*/) goto t77_n210;
  t77_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t77_n209;
  t77_n208: return 0.14498525857925415f;
  t77_n209: return 0.11678534746170044f;
  t77_n210: return -0.01973530650138855f;
  t77_n211: return 0.08576362580060959f;
  t77_n212: return -0.011226153932511806f;
  t77_n213: return 0.022056538611650467f;
  t77_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t77_n224;
  t77_n215: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t77_n217;
  t77_n216: return 0.001496477983891964f;
  t77_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n219;
  t77_n218: return -0.01685783639550209f;
  t77_n219: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t77_n223;
  t77_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t77_n222;
  t77_n221: return -0.045790836215019226f;
  t77_n222: return 0.058251406997442245f;
  t77_n223: return 0.022338656708598137f;
  t77_n224: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t77_n242;
  t77_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t77_n237;
  t77_n226: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t77_n236;
  t77_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t77_n229;
  t77_n228: return -0.033195603638887405f;
  t77_n229: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t77_n231;
  t77_n230: return -0.0032376472372561693f;
  t77_n231: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t77_n235;
  t77_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t77_n234;
  t77_n233: return -0.006243009120225906f;
  t77_n234: return 0.02097887545824051f;
  t77_n235: return 0.044078245759010315f;
  t77_n236: return -0.09702187776565552f;
  t77_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t77_n239;
  t77_n238: return -0.0032454964239150286f;
  t77_n239: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t77_n241;
  t77_n240: return 0.012934810481965542f;
  t77_n241: return 0.037578415125608444f;
  t77_n242: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t77_n254;
  t77_n243: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t77_n253;
  t77_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t77_n248;
  t77_n245: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t77_n247;
  t77_n246: return 0.011607781052589417f;
  t77_n247: return -0.11592119187116623f;
  t77_n248: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t77_n252;
  t77_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t77_n251;
  t77_n250: return 0.0374724417924881f;
  t77_n251: return 0.09540975093841553f;
  t77_n252: return -0.0973752811551094f;
  t77_n253: return -0.11112568527460098f;
  t77_n254: return 0.037703968584537506f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree78(const Example& E) {
  t78_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t78_n218;
  t78_n1: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t78_n213;
  t78_n2: if (E.getNumReferences() >= 3303133184 /*1806.5*/) goto t78_n206;
  t78_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t78_n163;
  t78_n4: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t78_n154;
  t78_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t78_n61;
  t78_n6: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t78_n52;
  t78_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n9;
  t78_n8: return -0.00729366997256875f;
  t78_n9: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t78_n45;
  t78_n10: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t78_n36;
  t78_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n33;
  t78_n12: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t78_n20;
  t78_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t78_n15;
  t78_n14: return -0.022188350558280945f;
  t78_n15: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t78_n17;
  t78_n16: return 0.019678814336657524f;
  t78_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t78_n19;
  t78_n18: return -0.0017223469913005829f;
  t78_n19: return 0.04021833464503288f;
  t78_n20: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t78_n26;
  t78_n21: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t78_n23;
  t78_n22: return 0.002327321795746684f;
  t78_n23: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t78_n25;
  t78_n24: return 0.03570609539747238f;
  t78_n25: return 0.008606716059148312f;
  t78_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t78_n30;
  t78_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n29;
  t78_n28: return -0.08549552410840988f;
  t78_n29: return 0.055571917444467545f;
  t78_n30: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t78_n32;
  t78_n31: return -0.14629556238651276f;
  t78_n32: return 0.12992767989635468f;
  t78_n33: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t78_n35;
  t78_n34: return -0.015574995428323746f;
  t78_n35: return 0.05150683969259262f;
  t78_n36: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t78_n40;
  t78_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n39;
  t78_n38: return -0.10111788660287857f;
  t78_n39: return -0.004113609902560711f;
  t78_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t78_n42;
  t78_n41: return 0.0010728162014856935f;
  t78_n42: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t78_n44;
  t78_n43: return 0.010138248093426228f;
  t78_n44: return -0.1445903182029724f;
  t78_n45: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t78_n47;
  t78_n46: return -0.06036724895238876f;
  t78_n47: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t78_n51;
  t78_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n50;
  t78_n49: return 0.014611756429076195f;
  t78_n50: return -0.02442571334540844f;
  t78_n51: return -0.02336900867521763f;
  t78_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n58;
  t78_n53: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t78_n57;
  t78_n54: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t78_n56;
  t78_n55: return -0.1306351274251938f;
  t78_n56: return 0.045280177146196365f;
  t78_n57: return -0.009833641350269318f;
  t78_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n60;
  t78_n59: return 0.014400629326701164f;
  t78_n60: return -0.010885135270655155f;
  t78_n61: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t78_n67;
  t78_n62: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t78_n66;
  t78_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t78_n65;
  t78_n64: return -0.025753788650035858f;
  t78_n65: return 0.02329234965145588f;
  t78_n66: return -0.065094493329525f;
  t78_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t78_n123;
  t78_n68: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t78_n72;
  t78_n69: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t78_n71;
  t78_n70: return 0.028584623709321022f;
  t78_n71: return -0.021799376234412193f;
  t78_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t78_n120;
  t78_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t78_n115;
  t78_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t78_n76;
  t78_n75: return -0.022448386996984482f;
  t78_n76: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t78_n100;
  t78_n77: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t78_n99;
  t78_n78: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t78_n98;
  t78_n79: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t78_n97;
  t78_n80: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t78_n84;
  t78_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n83;
  t78_n82: return -0.019059419631958008f;
  t78_n83: return 0.020476100966334343f;
  t78_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n94;
  t78_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t78_n87;
  t78_n86: return -0.037728626281023026f;
  t78_n87: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t78_n93;
  t78_n88: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t78_n92;
  t78_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t78_n91;
  t78_n90: return 0.029463229700922966f;
  t78_n91: return 0.0413113608956337f;
  t78_n92: return -0.11305700987577438f;
  t78_n93: return -0.054160311818122864f;
  t78_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t78_n96;
  t78_n95: return -0.06130432337522507f;
  t78_n96: return 0.04808421805500984f;
  t78_n97: return -0.032085686922073364f;
  t78_n98: return 0.019633343443274498f;
  t78_n99: return -0.05026007816195488f;
  t78_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t78_n102;
  t78_n101: return -0.00017939400277100503f;
  t78_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t78_n112;
  t78_n103: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n111;
  t78_n104: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t78_n108;
  t78_n105: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t78_n107;
  t78_n106: return 0.04691395163536072f;
  t78_n107: return -0.10864917933940887f;
  t78_n108: if (E.getNumReferences() >= 3302715392 /*1755.5*/) goto t78_n110;
  t78_n109: return -0.074998639523983f;
  t78_n110: return 0.13299337029457092f;
  t78_n111: return -0.042947422713041306f;
  t78_n112: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n114;
  t78_n113: return 0.03696579858660698f;
  t78_n114: return 0.07322956621646881f;
  t78_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t78_n117;
  t78_n116: return -0.1082642674446106f;
  t78_n117: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t78_n119;
  t78_n118: return 0.020669657737016678f;
  t78_n119: return 0.036372411996126175f;
  t78_n120: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t78_n122;
  t78_n121: return -0.020751407369971275f;
  t78_n122: return 0.14898401498794556f;
  t78_n123: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t78_n151;
  t78_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t78_n128;
  t78_n125: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t78_n127;
  t78_n126: return 0.007028191816061735f;
  t78_n127: return -0.028081152588129044f;
  t78_n128: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t78_n140;
  t78_n129: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t78_n139;
  t78_n130: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t78_n132;
  t78_n131: return 0.04827757179737091f;
  t78_n132: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t78_n136;
  t78_n133: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t78_n135;
  t78_n134: return 0.08340681344270706f;
  t78_n135: return -0.10548856854438782f;
  t78_n136: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t78_n138;
  t78_n137: return -0.026015067473053932f;
  t78_n138: return 0.05832184851169586f;
  t78_n139: return -0.014552833512425423f;
  t78_n140: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t78_n146;
  t78_n141: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t78_n145;
  t78_n142: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t78_n144;
  t78_n143: return 0.09895777702331543f;
  t78_n144: return 0.141835555434227f;
  t78_n145: return 0.052176833152770996f;
  t78_n146: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t78_n148;
  t78_n147: return -0.00505052087828517f;
  t78_n148: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t78_n150;
  t78_n149: return 0.03620534762740135f;
  t78_n150: return 0.08782786875963211f;
  t78_n151: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t78_n153;
  t78_n152: return 0.15624995529651642f;
  t78_n153: return 0.05027791112661362f;
  t78_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n158;
  t78_n155: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t78_n157;
  t78_n156: return -0.0018493842799216509f;
  t78_n157: return -0.02761734277009964f;
  t78_n158: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t78_n162;
  t78_n159: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t78_n161;
  t78_n160: return 0.013393268920481205f;
  t78_n161: return -0.012261577881872654f;
  t78_n162: return 0.011916846968233585f;
  t78_n163: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t78_n205;
  t78_n164: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t78_n204;
  t78_n165: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t78_n175;
  t78_n166: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t78_n172;
  t78_n167: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t78_n171;
  t78_n168: if (E.getNumReferences() >= 3288932352 /*548.5*/) goto t78_n170;
  t78_n169: return 0.009680529125034809f;
  t78_n170: return 0.05706578865647316f;
  t78_n171: return 0.039825279265642166f;
  t78_n172: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t78_n174;
  t78_n173: return 0.06492865085601807f;
  t78_n174: return -0.02818717621266842f;
  t78_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t78_n195;
  t78_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t78_n178;
  t78_n177: return -0.10685154050588608f;
  t78_n178: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t78_n190;
  t78_n179: if (E.getFileProximityDistanceCost() >= 3255304192 /*34.0*/) goto t78_n187;
  t78_n180: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t78_n186;
  t78_n181: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t78_n183;
  t78_n182: return 0.007745806127786636f;
  t78_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n185;
  t78_n184: return -0.04413534700870514f;
  t78_n185: return 0.20476816594600677f;
  t78_n186: return 0.009545119479298592f;
  t78_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n189;
  t78_n188: return -0.021484250202775f;
  t78_n189: return 0.02104836143553257f;
  t78_n190: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t78_n194;
  t78_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t78_n193;
  t78_n192: return 0.03265639767050743f;
  t78_n193: return -0.0095536969602108f;
  t78_n194: return 0.03842472285032272f;
  t78_n195: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t78_n197;
  t78_n196: return 0.0044036447070539f;
  t78_n197: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t78_n203;
  t78_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t78_n200;
  t78_n199: return -0.02266327477991581f;
  t78_n200: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t78_n202;
  t78_n201: return 0.05332442373037338f;
  t78_n202: return 0.12653906643390656f;
  t78_n203: return 0.059495311230421066f;
  t78_n204: return 0.06307826191186905f;
  t78_n205: return -0.2025275081396103f;
  t78_n206: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t78_n212;
  t78_n207: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t78_n211;
  t78_n208: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n210;
  t78_n209: return 0.02631285786628723f;
  t78_n210: return 0.001807592692784965f;
  t78_n211: return 0.04386616870760918f;
  t78_n212: return 0.0024464258458465338f;
  t78_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t78_n215;
  t78_n214: return -0.0026117151137441397f;
  t78_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t78_n217;
  t78_n216: return -0.14983415603637695f;
  t78_n217: return 0.019925298169255257f;
  t78_n218: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t78_n252;
  t78_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t78_n231;
  t78_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t78_n230;
  t78_n221: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t78_n229;
  t78_n222: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t78_n224;
  t78_n223: return -0.01579108089208603f;
  t78_n224: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t78_n228;
  t78_n225: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t78_n227;
  t78_n226: return 0.10093600302934647f;
  t78_n227: return 0.0026911080349236727f;
  t78_n228: return -0.12980788946151733f;
  t78_n229: return 0.014521687291562557f;
  t78_n230: return -0.03798620402812958f;
  t78_n231: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t78_n251;
  t78_n232: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t78_n238;
  t78_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t78_n235;
  t78_n234: return -0.11564687639474869f;
  t78_n235: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t78_n237;
  t78_n236: return 0.013715528883039951f;
  t78_n237: return 0.02748003974556923f;
  t78_n238: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t78_n242;
  t78_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t78_n241;
  t78_n240: return -0.03072693571448326f;
  t78_n241: return 0.01367452833801508f;
  t78_n242: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t78_n250;
  t78_n243: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t78_n247;
  t78_n244: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t78_n246;
  t78_n245: return 0.07031583040952682f;
  t78_n246: return 0.11026262491941452f;
  t78_n247: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t78_n249;
  t78_n248: return 0.007140217814594507f;
  t78_n249: return 0.02399919554591179f;
  t78_n250: return -0.010713696479797363f;
  t78_n251: return -0.0009421068243682384f;
  t78_n252: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t78_n254;
  t78_n253: return 0.01661154441535473f;
  t78_n254: return 0.05389002710580826f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree79(const Example& E) {
  t79_n0: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t79_n176;
  t79_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t79_n153;
  t79_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t79_n122;
  t79_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t79_n25;
  t79_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t79_n22;
  t79_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n11;
  t79_n6: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t79_n8;
  t79_n7: return -0.0069734458811581135f;
  t79_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t79_n10;
  t79_n9: return 0.01340915821492672f;
  t79_n10: return 0.0635608583688736f;
  t79_n11: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t79_n21;
  t79_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t79_n14;
  t79_n13: return -0.10433860123157501f;
  t79_n14: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t79_n20;
  t79_n15: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t79_n19;
  t79_n16: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t79_n18;
  t79_n17: return 0.007471795193850994f;
  t79_n18: return -0.05338224396109581f;
  t79_n19: return -0.09443136304616928f;
  t79_n20: return 0.021931633353233337f;
  t79_n21: return 0.035254091024398804f;
  t79_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n24;
  t79_n23: return -0.05418563634157181f;
  t79_n24: return -0.03755057975649834f;
  t79_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t79_n107;
  t79_n26: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t79_n94;
  t79_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t79_n67;
  t79_n28: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t79_n62;
  t79_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t79_n31;
  t79_n30: return -0.011626899242401123f;
  t79_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t79_n35;
  t79_n32: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t79_n34;
  t79_n33: return -0.0036835509818047285f;
  t79_n34: return -0.09349554777145386f;
  t79_n35: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t79_n45;
  t79_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n40;
  t79_n37: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t79_n39;
  t79_n38: return -0.009654349647462368f;
  t79_n39: return 0.09880121797323227f;
  t79_n40: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t79_n44;
  t79_n41: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t79_n43;
  t79_n42: return -0.010381274856626987f;
  t79_n43: return 0.013703184202313423f;
  t79_n44: return 0.0012204822851344943f;
  t79_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n49;
  t79_n46: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t79_n48;
  t79_n47: return 0.049854837357997894f;
  t79_n48: return -0.01836632750928402f;
  t79_n49: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t79_n59;
  t79_n50: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t79_n54;
  t79_n51: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t79_n53;
  t79_n52: return 0.024162065237760544f;
  t79_n53: return 0.09411738067865372f;
  t79_n54: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t79_n58;
  t79_n55: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t79_n57;
  t79_n56: return -0.011026114225387573f;
  t79_n57: return 0.04436080530285835f;
  t79_n58: return -0.08389542251825333f;
  t79_n59: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t79_n61;
  t79_n60: return 0.028848430141806602f;
  t79_n61: return -0.01606716401875019f;
  t79_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n66;
  t79_n63: if (E.getNumReferences() >= 3330284288 /*16397.5*/) goto t79_n65;
  t79_n64: return -0.05342891067266464f;
  t79_n65: return -0.15759122371673584f;
  t79_n66: return 0.007694465108215809f;
  t79_n67: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t79_n77;
  t79_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t79_n70;
  t79_n69: return -0.07661954313516617f;
  t79_n70: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t79_n76;
  t79_n71: if (E.getNumReferences() >= 3316316160 /*5471.0*/) goto t79_n73;
  t79_n72: return 0.03637154772877693f;
  t79_n73: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t79_n75;
  t79_n74: return 0.14457537233829498f;
  t79_n75: return 0.08207491785287857f;
  t79_n76: return -0.03168300911784172f;
  t79_n77: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t79_n89;
  t79_n78: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t79_n80;
  t79_n79: return 0.01817588321864605f;
  t79_n80: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t79_n86;
  t79_n81: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t79_n85;
  t79_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t79_n84;
  t79_n83: return 0.0021965873893350363f;
  t79_n84: return 0.07719503343105316f;
  t79_n85: return 0.008189036510884762f;
  t79_n86: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t79_n88;
  t79_n87: return 0.01385490596294403f;
  t79_n88: return 0.04177385941147804f;
  t79_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t79_n91;
  t79_n90: return -0.0020857553463429213f;
  t79_n91: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t79_n93;
  t79_n92: return 0.038734205067157745f;
  t79_n93: return 0.12834221124649048f;
  t79_n94: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t79_n104;
  t79_n95: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t79_n99;
  t79_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t79_n98;
  t79_n97: return 0.003711898811161518f;
  t79_n98: return 0.014617936685681343f;
  t79_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n103;
  t79_n100: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t79_n102;
  t79_n101: return -0.5336731672286987f;
  t79_n102: return -0.02539672516286373f;
  t79_n103: return 0.003412415971979499f;
  t79_n104: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t79_n106;
  t79_n105: return -0.0791308805346489f;
  t79_n106: return -0.0013080986682325602f;
  t79_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t79_n109;
  t79_n108: return 0.010012668557465076f;
  t79_n109: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t79_n115;
  t79_n110: if (E.getNumReferences() >= 3356993440 /*155294.5*/) goto t79_n114;
  t79_n111: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t79_n113;
  t79_n112: return 0.03096671961247921f;
  t79_n113: return 0.004766432568430901f;
  t79_n114: return 0.05361124128103256f;
  t79_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t79_n117;
  t79_n116: return 0.04867080971598625f;
  t79_n117: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t79_n119;
  t79_n118: return 0.03378962725400925f;
  t79_n119: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t79_n121;
  t79_n120: return 0.03166515752673149f;
  t79_n121: return 0.09107646346092224f;
  t79_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t79_n124;
  t79_n123: return -0.06041985750198364f;
  t79_n124: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t79_n146;
  t79_n125: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t79_n145;
  t79_n126: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t79_n132;
  t79_n127: if (E.getNumReferences() >= 3350072960 /*89117.0*/) goto t79_n131;
  t79_n128: if (E.getNumReferences() >= 3325608960 /*11825.0*/) goto t79_n130;
  t79_n129: return -0.011054432019591331f;
  t79_n130: return 0.015304919332265854f;
  t79_n131: return -0.1886807680130005f;
  t79_n132: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t79_n142;
  t79_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t79_n139;
  t79_n134: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t79_n138;
  t79_n135: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t79_n137;
  t79_n136: return 0.06310753524303436f;
  t79_n137: return 0.1465735286474228f;
  t79_n138: return -0.13530421257019043f;
  t79_n139: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t79_n141;
  t79_n140: return 0.1312291920185089f;
  t79_n141: return -0.10281020402908325f;
  t79_n142: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t79_n144;
  t79_n143: return 0.05343719571828842f;
  t79_n144: return -0.10992328077554703f;
  t79_n145: return -0.1320609748363495f;
  t79_n146: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t79_n152;
  t79_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t79_n149;
  t79_n148: return -0.14986754953861237f;
  t79_n149: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t79_n151;
  t79_n150: return 0.07140126079320908f;
  t79_n151: return 0.08887732028961182f;
  t79_n152: return -0.0979294404387474f;
  t79_n153: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t79_n159;
  t79_n154: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t79_n158;
  t79_n155: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t79_n157;
  t79_n156: return 0.030268298462033272f;
  t79_n157: return 0.0025212205946445465f;
  t79_n158: return 0.028499556705355644f;
  t79_n159: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t79_n173;
  t79_n160: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t79_n168;
  t79_n161: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t79_n165;
  t79_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t79_n164;
  t79_n163: return 0.0172275397926569f;
  t79_n164: return 0.0653286799788475f;
  t79_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t79_n167;
  t79_n166: return 0.12094463407993317f;
  t79_n167: return 0.11173605918884277f;
  t79_n168: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t79_n170;
  t79_n169: return -0.032124899327754974f;
  t79_n170: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t79_n172;
  t79_n171: return 0.03712190315127373f;
  t79_n172: return -0.274260014295578f;
  t79_n173: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t79_n175;
  t79_n174: return 0.10486199706792831f;
  t79_n175: return 0.065703384578228f;
  t79_n176: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t79_n248;
  t79_n177: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t79_n189;
  t79_n178: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t79_n188;
  t79_n179: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t79_n187;
  t79_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t79_n186;
  t79_n181: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t79_n185;
  t79_n182: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t79_n184;
  t79_n183: return 0.001318494789302349f;
  t79_n184: return 0.054322198033332825f;
  t79_n185: return -0.057340752333402634f;
  t79_n186: return 0.005731872748583555f;
  t79_n187: return 0.013531404547393322f;
  t79_n188: return 0.06573652476072311f;
  t79_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t79_n195;
  t79_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t79_n194;
  t79_n191: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t79_n193;
  t79_n192: return -0.0023224493488669395f;
  t79_n193: return 0.02722748927772045f;
  t79_n194: return 0.018797187134623528f;
  t79_n195: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t79_n247;
  t79_n196: if (E.getNumReferences() >= 3323824640 /*10082.5*/) goto t79_n246;
  t79_n197: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t79_n227;
  t79_n198: if (E.getFractionNameInContext() >= 3197498870 /*0.2928921580314636*/) goto t79_n204;
  t79_n199: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t79_n201;
  t79_n200: return 0.004922868683934212f;
  t79_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n203;
  t79_n202: return -0.12539565563201904f;
  t79_n203: return -0.14975404739379883f;
  t79_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t79_n206;
  t79_n205: return 0.004028852563351393f;
  t79_n206: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t79_n210;
  t79_n207: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t79_n209;
  t79_n208: return 0.015208262950181961f;
  t79_n209: return 0.03928826004266739f;
  t79_n210: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t79_n224;
  t79_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t79_n223;
  t79_n212: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t79_n222;
  t79_n213: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t79_n215;
  t79_n214: return -0.08619474619626999f;
  t79_n215: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t79_n221;
  t79_n216: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t79_n218;
  t79_n217: return -0.29389631748199463f;
  t79_n218: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t79_n220;
  t79_n219: return 0.17608506977558136f;
  t79_n220: return -1.1275469064712524f;
  t79_n221: return -0.9991827607154846f;
  t79_n222: return 0.03632489964365959f;
  t79_n223: return 0.04074614495038986f;
  t79_n224: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t79_n226;
  t79_n225: return 0.005341779440641403f;
  t79_n226: return 0.07413214445114136f;
  t79_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n229;
  t79_n228: return -0.04645320400595665f;
  t79_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t79_n243;
  t79_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t79_n234;
  t79_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t79_n233;
  t79_n232: return -0.30611133575439453f;
  t79_n233: return -0.05192257836461067f;
  t79_n234: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t79_n240;
  t79_n235: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t79_n239;
  t79_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t79_n238;
  t79_n237: return 0.06590430438518524f;
  t79_n238: return 0.13344214856624603f;
  t79_n239: return 0.010658026672899723f;
  t79_n240: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t79_n242;
  t79_n241: return 0.02829638123512268f;
  t79_n242: return 0.05510761961340904f;
  t79_n243: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t79_n245;
  t79_n244: return 0.046852242201566696f;
  t79_n245: return 0.10973736643791199f;
  t79_n246: return 0.04692764952778816f;
  t79_n247: return -0.002293347381055355f;
  t79_n248: if (E.getNumReferences() >= 3316158464 /*5394.0*/) goto t79_n254;
  t79_n249: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t79_n253;
  t79_n250: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t79_n252;
  t79_n251: return 0.018755847588181496f;
  t79_n252: return 0.0029181058052927256f;
  t79_n253: return 0.016847502440214157f;
  t79_n254: return 0.05257115140557289f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree80(const Example& E) {
  t80_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t80_n176;
  t80_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t80_n29;
  t80_n2: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t80_n26;
  t80_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t80_n5;
  t80_n4: return -0.0010287887416779995f;
  t80_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t80_n21;
  t80_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n16;
  t80_n7: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t80_n9;
  t80_n8: return -0.07920090109109879f;
  t80_n9: if (E.getNumReferences() >= 3272966144 /*149.5*/) goto t80_n11;
  t80_n10: return -0.0062149278819561005f;
  t80_n11: if (E.getNumReferences() >= 3274145792 /*167.5*/) goto t80_n15;
  t80_n12: if (E.getNumReferences() >= 3273883648 /*163.5*/) goto t80_n14;
  t80_n13: return 0.09011197090148926f;
  t80_n14: return 0.16898752748966217f;
  t80_n15: return 0.059739600867033005f;
  t80_n16: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t80_n20;
  t80_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n19;
  t80_n18: return -0.018561558797955513f;
  t80_n19: return 0.012814342975616455f;
  t80_n20: return 0.026598481461405754f;
  t80_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t80_n25;
  t80_n22: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t80_n24;
  t80_n23: return 0.007839121855795383f;
  t80_n24: return 0.03496941179037094f;
  t80_n25: return 0.10058116912841797f;
  t80_n26: if (E.getNumReferences() >= 3308726272 /*2930.5*/) goto t80_n28;
  t80_n27: return 0.09760720282793045f;
  t80_n28: return 0.020184597000479698f;
  t80_n29: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t80_n173;
  t80_n30: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t80_n148;
  t80_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t80_n109;
  t80_n32: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t80_n108;
  t80_n33: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t80_n95;
  t80_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t80_n72;
  t80_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t80_n71;
  t80_n36: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t80_n48;
  t80_n37: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t80_n43;
  t80_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t80_n40;
  t80_n39: return -0.011002857238054276f;
  t80_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t80_n42;
  t80_n41: return -0.0509253665804863f;
  t80_n42: return 0.025974847376346588f;
  t80_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t80_n45;
  t80_n44: return -0.04743655398488045f;
  t80_n45: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t80_n47;
  t80_n46: return -0.11958706378936768f;
  t80_n47: return 0.13344666361808777f;
  t80_n48: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t80_n68;
  t80_n49: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t80_n63;
  t80_n50: if (E.getNumReferences() >= 3341362816 /*43302.5*/) goto t80_n52;
  t80_n51: return -0.0154331149533391f;
  t80_n52: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t80_n56;
  t80_n53: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t80_n55;
  t80_n54: return -0.019811924546957016f;
  t80_n55: return 0.12912259995937347f;
  t80_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t80_n58;
  t80_n57: return 0.008291922509670258f;
  t80_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t80_n62;
  t80_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t80_n61;
  t80_n60: return -0.009445380419492722f;
  t80_n61: return 0.12543590366840363f;
  t80_n62: return -0.14461465179920197f;
  t80_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n65;
  t80_n64: return -0.17685678601264954f;
  t80_n65: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t80_n67;
  t80_n66: return 0.09961593151092529f;
  t80_n67: return 0.1374175250530243f;
  t80_n68: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t80_n70;
  t80_n69: return -0.09903965890407562f;
  t80_n70: return -0.015101034194231033f;
  t80_n71: return 0.00767225818708539f;
  t80_n72: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t80_n78;
  t80_n73: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t80_n77;
  t80_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t80_n76;
  t80_n75: return 0.0006840861169621348f;
  t80_n76: return 0.049701027572155f;
  t80_n77: return -0.06591739505529404f;
  t80_n78: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t80_n94;
  t80_n79: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t80_n93;
  t80_n80: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t80_n90;
  t80_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t80_n83;
  t80_n82: return -0.03660484775900841f;
  t80_n83: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t80_n89;
  t80_n84: if (E.getNumReferences() >= 3270443008 /*119.5*/) goto t80_n86;
  t80_n85: return 0.05188361555337906f;
  t80_n86: if (E.getNumReferences() >= 3277389824 /*217.0*/) goto t80_n88;
  t80_n87: return 0.2559397220611572f;
  t80_n88: return 0.06773949414491653f;
  t80_n89: return 0.013110388070344925f;
  t80_n90: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t80_n92;
  t80_n91: return 0.1320265382528305f;
  t80_n92: return 0.011729948222637177f;
  t80_n93: return -0.12230633199214935f;
  t80_n94: return 0.08088885992765427f;
  t80_n95: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t80_n107;
  t80_n96: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t80_n104;
  t80_n97: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t80_n99;
  t80_n98: return -0.08925441652536392f;
  t80_n99: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t80_n101;
  t80_n100: return -0.018664013594388962f;
  t80_n101: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t80_n103;
  t80_n102: return 0.06850039213895798f;
  t80_n103: return 0.10397768020629883f;
  t80_n104: if (E.getNumReferences() >= 3360267424 /*206450.5*/) goto t80_n106;
  t80_n105: return -0.18578751385211945f;
  t80_n106: return 0.13072368502616882f;
  t80_n107: return 0.02450009435415268f;
  t80_n108: return -0.03761561959981918f;
  t80_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t80_n111;
  t80_n110: return -0.08337195217609406f;
  t80_n111: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t80_n115;
  t80_n112: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t80_n114;
  t80_n113: return 0.098673515021801f;
  t80_n114: return 0.15275947749614716f;
  t80_n115: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t80_n119;
  t80_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n118;
  t80_n117: return -0.008828331716358662f;
  t80_n118: return 0.00593704404309392f;
  t80_n119: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t80_n143;
  t80_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n134;
  t80_n121: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t80_n131;
  t80_n122: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t80_n126;
  t80_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n125;
  t80_n124: return -0.09995283931493759f;
  t80_n125: return 0.026054291054606438f;
  t80_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n128;
  t80_n127: return -0.039180148392915726f;
  t80_n128: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t80_n130;
  t80_n129: return 0.015724580734968185f;
  t80_n130: return 0.04944298043847084f;
  t80_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n133;
  t80_n132: return -0.14412179589271545f;
  t80_n133: return 0.0006851644720882177f;
  t80_n134: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t80_n138;
  t80_n135: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t80_n137;
  t80_n136: return -1.743262146192137e-05f;
  t80_n137: return 0.029695868492126465f;
  t80_n138: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t80_n140;
  t80_n139: return 0.002745185513049364f;
  t80_n140: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t80_n142;
  t80_n141: return -0.03618878126144409f;
  t80_n142: return 0.023823196068406105f;
  t80_n143: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t80_n145;
  t80_n144: return 0.07456130534410477f;
  t80_n145: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t80_n147;
  t80_n146: return -0.08151654899120331f;
  t80_n147: return 0.0766424611210823f;
  t80_n148: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t80_n166;
  t80_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t80_n151;
  t80_n150: return -0.02915174700319767f;
  t80_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t80_n153;
  t80_n152: return -0.02959008514881134f;
  t80_n153: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t80_n165;
  t80_n154: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t80_n164;
  t80_n155: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t80_n157;
  t80_n156: return 0.07500051707029343f;
  t80_n157: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t80_n159;
  t80_n158: return 0.1212310716509819f;
  t80_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n161;
  t80_n160: return -0.13487526774406433f;
  t80_n161: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t80_n163;
  t80_n162: return 0.18267673254013062f;
  t80_n163: return 0.1073329970240593f;
  t80_n164: return -0.15887193381786346f;
  t80_n165: return 0.02101953513920307f;
  t80_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n168;
  t80_n167: return -0.016346968710422516f;
  t80_n168: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t80_n170;
  t80_n169: return 0.0021524406038224697f;
  t80_n170: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t80_n172;
  t80_n171: return 0.08512801676988602f;
  t80_n172: return 0.035844288766384125f;
  t80_n173: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t80_n175;
  t80_n174: return 0.00554298609495163f;
  t80_n175: return -0.03485541790723801f;
  t80_n176: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t80_n196;
  t80_n177: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t80_n185;
  t80_n178: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t80_n182;
  t80_n179: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t80_n181;
  t80_n180: return 0.0013450686819851398f;
  t80_n181: return 0.012848018668591976f;
  t80_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t80_n184;
  t80_n183: return 0.002380102640017867f;
  t80_n184: return 0.02308555692434311f;
  t80_n185: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t80_n193;
  t80_n186: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t80_n190;
  t80_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t80_n189;
  t80_n188: return -0.11455990374088287f;
  t80_n189: return 0.007086066994816065f;
  t80_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n192;
  t80_n191: return -0.12261240184307098f;
  t80_n192: return -0.09995914995670319f;
  t80_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n195;
  t80_n194: return -0.09713544696569443f;
  t80_n195: return -0.029708851128816605f;
  t80_n196: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t80_n254;
  t80_n197: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t80_n253;
  t80_n198: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t80_n202;
  t80_n199: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t80_n201;
  t80_n200: return 0.031248630955815315f;
  t80_n201: return 0.001681619556620717f;
  t80_n202: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t80_n204;
  t80_n203: return -0.004918423481285572f;
  t80_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t80_n240;
  t80_n205: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t80_n215;
  t80_n206: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t80_n210;
  t80_n207: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t80_n209;
  t80_n208: return 0.000497597036883235f;
  t80_n209: return 0.03832210972905159f;
  t80_n210: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t80_n214;
  t80_n211: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t80_n213;
  t80_n212: return 0.08858536928892136f;
  t80_n213: return 0.14802014827728271f;
  t80_n214: return 0.014502884820103645f;
  t80_n215: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t80_n229;
  t80_n216: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t80_n226;
  t80_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t80_n219;
  t80_n218: return -0.17936980724334717f;
  t80_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t80_n225;
  t80_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t80_n222;
  t80_n221: return -0.04459937661886215f;
  t80_n222: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t80_n224;
  t80_n223: return -0.09224991500377655f;
  t80_n224: return 0.15958909690380096f;
  t80_n225: return 0.05881275609135628f;
  t80_n226: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t80_n228;
  t80_n227: return -0.0625460222363472f;
  t80_n228: return 0.1679379791021347f;
  t80_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t80_n237;
  t80_n230: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t80_n236;
  t80_n231: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t80_n235;
  t80_n232: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t80_n234;
  t80_n233: return 0.0004872353165410459f;
  t80_n234: return -0.18796497583389282f;
  t80_n235: return -0.11501879245042801f;
  t80_n236: return 0.05542002245783806f;
  t80_n237: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t80_n239;
  t80_n238: return 0.019655516371130943f;
  t80_n239: return 0.05341653525829315f;
  t80_n240: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t80_n252;
  t80_n241: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t80_n251;
  t80_n242: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t80_n248;
  t80_n243: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t80_n247;
  t80_n244: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t80_n246;
  t80_n245: return 0.07842948287725449f;
  t80_n246: return 0.12070177495479584f;
  t80_n247: return 0.021908462047576904f;
  t80_n248: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t80_n250;
  t80_n249: return 0.012560467235744f;
  t80_n250: return 0.04021080583333969f;
  t80_n251: return -0.36966753005981445f;
  t80_n252: return -0.19982561469078064f;
  t80_n253: return 0.02892700955271721f;
  t80_n254: return 0.03336801752448082f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree81(const Example& E) {
  t81_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t81_n32;
  t81_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t81_n25;
  t81_n2: if (E.getNumReferences() >= 3301396480 /*1594.5*/) goto t81_n24;
  t81_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t81_n5;
  t81_n4: return -0.0011111331405118108f;
  t81_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t81_n19;
  t81_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n14;
  t81_n7: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t81_n9;
  t81_n8: return -0.06980973482131958f;
  t81_n9: if (E.getNumReferences() >= 3269787648 /*114.5*/) goto t81_n13;
  t81_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t81_n12;
  t81_n11: return -0.02102404087781906f;
  t81_n12: return 0.006539011374115944f;
  t81_n13: return 0.05900314450263977f;
  t81_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t81_n18;
  t81_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n17;
  t81_n16: return -0.01730477437376976f;
  t81_n17: return 0.011852575466036797f;
  t81_n18: return 0.02512171119451523f;
  t81_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t81_n23;
  t81_n20: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t81_n22;
  t81_n21: return 0.00805793609470129f;
  t81_n22: return 0.030573133379220963f;
  t81_n23: return 0.09843491017818451f;
  t81_n24: return 0.019149744883179665f;
  t81_n25: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t81_n29;
  t81_n26: if (E.getFileProximityDistanceCost() >= 3255697408 /*35.5*/) goto t81_n28;
  t81_n27: return -0.006784459576010704f;
  t81_n28: return -0.1384374499320984f;
  t81_n29: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t81_n31;
  t81_n30: return 0.005108349956572056f;
  t81_n31: return -0.03163838014006615f;
  t81_n32: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t81_n234;
  t81_n33: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t81_n233;
  t81_n34: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t81_n232;
  t81_n35: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t81_n227;
  t81_n36: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t81_n92;
  t81_n37: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t81_n85;
  t81_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n42;
  t81_n39: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t81_n41;
  t81_n40: return -0.0017638576682657003f;
  t81_n41: return -0.13911615312099457f;
  t81_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t81_n68;
  t81_n43: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t81_n65;
  t81_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t81_n48;
  t81_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n47;
  t81_n46: return -0.03077421709895134f;
  t81_n47: return 0.009602587670087814f;
  t81_n48: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t81_n54;
  t81_n49: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t81_n53;
  t81_n50: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t81_n52;
  t81_n51: return 0.022916045039892197f;
  t81_n52: return -0.08984849601984024f;
  t81_n53: return 0.014077137224376202f;
  t81_n54: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t81_n64;
  t81_n55: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t81_n57;
  t81_n56: return -0.020114624872803688f;
  t81_n57: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t81_n61;
  t81_n58: if (E.getSymbolScopeDistanceCost() >= 3244294144 /*14.0*/) goto t81_n60;
  t81_n59: return 0.015800783410668373f;
  t81_n60: return 0.05794263631105423f;
  t81_n61: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t81_n63;
  t81_n62: return 0.2085186243057251f;
  t81_n63: return 0.1210172027349472f;
  t81_n64: return -0.01729605533182621f;
  t81_n65: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t81_n67;
  t81_n66: return 0.01979639381170273f;
  t81_n67: return -0.0608525313436985f;
  t81_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t81_n80;
  t81_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n73;
  t81_n70: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t81_n72;
  t81_n71: return -0.021495943889021873f;
  t81_n72: return -0.0006895192200317979f;
  t81_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t81_n77;
  t81_n74: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t81_n76;
  t81_n75: return 0.006783305201679468f;
  t81_n76: return 0.09296872466802597f;
  t81_n77: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t81_n79;
  t81_n78: return 0.10416655987501144f;
  t81_n79: return 0.007279614452272654f;
  t81_n80: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t81_n82;
  t81_n81: return 0.06319000571966171f;
  t81_n82: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t81_n84;
  t81_n83: return -0.019785692915320396f;
  t81_n84: return 0.03890516236424446f;
  t81_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t81_n91;
  t81_n86: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t81_n88;
  t81_n87: return -0.003307042410597205f;
  t81_n88: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t81_n90;
  t81_n89: return 0.0408795140683651f;
  t81_n90: return 0.009743837639689445f;
  t81_n91: return 0.02855687029659748f;
  t81_n92: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t81_n210;
  t81_n93: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t81_n187;
  t81_n94: if (E.getNumReferences() >= 3358036832 /*171597.5*/) goto t81_n186;
  t81_n95: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t81_n185;
  t81_n96: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t81_n182;
  t81_n97: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t81_n157;
  t81_n98: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t81_n146;
  t81_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t81_n103;
  t81_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t81_n102;
  t81_n101: return -0.11509784311056137f;
  t81_n102: return -0.1019820123910904f;
  t81_n103: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t81_n119;
  t81_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t81_n112;
  t81_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n107;
  t81_n106: return -0.024652574211359024f;
  t81_n107: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t81_n109;
  t81_n108: return -0.0007741269655525684f;
  t81_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n111;
  t81_n110: return -0.11525961756706238f;
  t81_n111: return 0.08025916665792465f;
  t81_n112: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t81_n114;
  t81_n113: return -0.0054041980765759945f;
  t81_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n118;
  t81_n115: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t81_n117;
  t81_n116: return -0.11554163694381714f;
  t81_n117: return 0.02947958931326866f;
  t81_n118: return 0.010177935473620892f;
  t81_n119: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t81_n121;
  t81_n120: return -0.004443799611181021f;
  t81_n121: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t81_n145;
  t81_n122: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t81_n132;
  t81_n123: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t81_n125;
  t81_n124: return -0.006131582427769899f;
  t81_n125: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t81_n131;
  t81_n126: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t81_n128;
  t81_n127: return 0.02103416807949543f;
  t81_n128: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t81_n130;
  t81_n129: return 0.07295087724924088f;
  t81_n130: return 0.012071932666003704f;
  t81_n131: return 0.13510216772556305f;
  t81_n132: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t81_n142;
  t81_n133: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t81_n141;
  t81_n134: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t81_n138;
  t81_n135: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t81_n137;
  t81_n136: return 0.01697707735002041f;
  t81_n137: return -0.04535214230418205f;
  t81_n138: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t81_n140;
  t81_n139: return 0.042339716106653214f;
  t81_n140: return -0.04898098111152649f;
  t81_n141: return -0.1767125427722931f;
  t81_n142: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t81_n144;
  t81_n143: return 0.01046844758093357f;
  t81_n144: return 0.035465024411678314f;
  t81_n145: return -0.141665518283844f;
  t81_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n150;
  t81_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n149;
  t81_n148: return -0.2763291895389557f;
  t81_n149: return -0.052044790238142014f;
  t81_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t81_n152;
  t81_n151: return 0.0034893383271992207f;
  t81_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t81_n156;
  t81_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t81_n155;
  t81_n154: return 0.001280491822399199f;
  t81_n155: return 0.03133952245116234f;
  t81_n156: return 0.02728590928018093f;
  t81_n157: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t81_n167;
  t81_n158: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t81_n164;
  t81_n159: if (E.getNumReferences() >= 3350520768 /*92615.5*/) goto t81_n163;
  t81_n160: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t81_n162;
  t81_n161: return 0.02679525315761566f;
  t81_n162: return 0.06976068019866943f;
  t81_n163: return -0.09873219579458237f;
  t81_n164: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t81_n166;
  t81_n165: return 0.1332324594259262f;
  t81_n166: return 0.025444433093070984f;
  t81_n167: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t81_n179;
  t81_n168: if (E.getNumReferences() >= 3357701696 /*166361.0*/) goto t81_n178;
  t81_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t81_n171;
  t81_n170: return -0.021405456587672234f;
  t81_n171: if (E.getNumReferences() >= 3345968640 /*61294.0*/) goto t81_n175;
  t81_n172: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t81_n174;
  t81_n173: return 0.010266601108014584f;
  t81_n174: return -0.3381246030330658f;
  t81_n175: if (E.getNumReferences() >= 3352087680 /*104857.0*/) goto t81_n177;
  t81_n176: return 0.06197340413928032f;
  t81_n177: return 0.009349036030471325f;
  t81_n178: return -0.5944486260414124f;
  t81_n179: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t81_n181;
  t81_n180: return 0.28016918897628784f;
  t81_n181: return 0.08144306391477585f;
  t81_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n184;
  t81_n183: return -0.12897586822509766f;
  t81_n184: return -0.0935416966676712f;
  t81_n185: return -0.04023100435733795f;
  t81_n186: return -0.18365494906902313f;
  t81_n187: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t81_n205;
  t81_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t81_n194;
  t81_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t81_n191;
  t81_n190: return -0.022641560062766075f;
  t81_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n193;
  t81_n192: return -0.16797341406345367f;
  t81_n193: return 0.03853869065642357f;
  t81_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t81_n198;
  t81_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n197;
  t81_n196: return -0.3300004005432129f;
  t81_n197: return 0.00974472239613533f;
  t81_n198: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t81_n200;
  t81_n199: return 0.03880804032087326f;
  t81_n200: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t81_n202;
  t81_n201: return 0.04517754167318344f;
  t81_n202: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t81_n204;
  t81_n203: return 0.2697811424732208f;
  t81_n204: return -0.10128290951251984f;
  t81_n205: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t81_n207;
  t81_n206: return 0.03821028023958206f;
  t81_n207: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t81_n209;
  t81_n208: return 0.197668194770813f;
  t81_n209: return 0.07921392470598221f;
  t81_n210: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t81_n218;
  t81_n211: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t81_n217;
  t81_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t81_n216;
  t81_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n215;
  t81_n214: return -0.03915754333138466f;
  t81_n215: return -0.005429718177765608f;
  t81_n216: return 0.07139638066291809f;
  t81_n217: return 0.022745613008737564f;
  t81_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n220;
  t81_n219: return -0.06437505036592484f;
  t81_n220: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t81_n224;
  t81_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t81_n223;
  t81_n222: return -0.1808777153491974f;
  t81_n223: return 0.016079600900411606f;
  t81_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t81_n226;
  t81_n225: return 0.022670842707157135f;
  t81_n226: return 0.07413999736309052f;
  t81_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t81_n231;
  t81_n228: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t81_n230;
  t81_n229: return 0.10434967279434204f;
  t81_n230: return 0.018446490168571472f;
  t81_n231: return 0.10282666981220245f;
  t81_n232: return -0.09284695982933044f;
  t81_n233: return 0.04702698066830635f;
  t81_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t81_n238;
  t81_n235: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t81_n237;
  t81_n236: return 0.002259809523820877f;
  t81_n237: return -0.03444904088973999f;
  t81_n238: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t81_n254;
  t81_n239: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t81_n249;
  t81_n240: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t81_n244;
  t81_n241: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t81_n243;
  t81_n242: return 0.006612319964915514f;
  t81_n243: return 0.03876720368862152f;
  t81_n244: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t81_n248;
  t81_n245: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t81_n247;
  t81_n246: return 0.008551066741347313f;
  t81_n247: return 0.07738550007343292f;
  t81_n248: return 0.03912090137600899f;
  t81_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t81_n253;
  t81_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t81_n252;
  t81_n251: return -0.030289342626929283f;
  t81_n252: return 0.031566739082336426f;
  t81_n253: return 0.01756334863603115f;
  t81_n254: return 0.02118379808962345f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree82(const Example& E) {
  t82_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t82_n182;
  t82_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t82_n157;
  t82_n2: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t82_n68;
  t82_n3: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t82_n31;
  t82_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t82_n8;
  t82_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n7;
  t82_n6: return -0.06944719702005386f;
  t82_n7: return -0.019179673865437508f;
  t82_n8: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t82_n20;
  t82_n9: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t82_n19;
  t82_n10: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t82_n18;
  t82_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t82_n15;
  t82_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t82_n14;
  t82_n13: return -0.01735859364271164f;
  t82_n14: return 0.006868240889161825f;
  t82_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t82_n17;
  t82_n16: return 0.0027968892827630043f;
  t82_n17: return 0.039693813771009445f;
  t82_n18: return -0.00326314615085721f;
  t82_n19: return -0.09429626911878586f;
  t82_n20: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t82_n30;
  t82_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t82_n23;
  t82_n22: return -0.004610279109328985f;
  t82_n23: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t82_n27;
  t82_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t82_n26;
  t82_n25: return 0.011254929006099701f;
  t82_n26: return 0.02033267728984356f;
  t82_n27: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t82_n29;
  t82_n28: return 0.00510060042142868f;
  t82_n29: return -0.18240302801132202f;
  t82_n30: return -0.05693056434392929f;
  t82_n31: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t82_n45;
  t82_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t82_n34;
  t82_n33: return -0.06104453280568123f;
  t82_n34: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t82_n44;
  t82_n35: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t82_n37;
  t82_n36: return -0.02175920642912388f;
  t82_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n39;
  t82_n38: return -0.1352405548095703f;
  t82_n39: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t82_n41;
  t82_n40: return 0.03260636329650879f;
  t82_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t82_n43;
  t82_n42: return 0.06529324501752853f;
  t82_n43: return 0.11924902349710464f;
  t82_n44: return -0.06927471607923508f;
  t82_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n65;
  t82_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t82_n62;
  t82_n47: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t82_n61;
  t82_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t82_n60;
  t82_n49: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t82_n55;
  t82_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n54;
  t82_n51: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t82_n53;
  t82_n52: return 0.01267862319946289f;
  t82_n53: return -0.08270730823278427f;
  t82_n54: return -0.006765530910342932f;
  t82_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t82_n59;
  t82_n56: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t82_n58;
  t82_n57: return 0.029986973851919174f;
  t82_n58: return -0.014199379831552505f;
  t82_n59: return 0.04488654062151909f;
  t82_n60: return 0.04101734608411789f;
  t82_n61: return -0.05363592877984047f;
  t82_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n64;
  t82_n63: return -0.13671426475048065f;
  t82_n64: return 0.1150699034333229f;
  t82_n65: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t82_n67;
  t82_n66: return 0.03088025189936161f;
  t82_n67: return 0.009908009320497513f;
  t82_n68: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t82_n76;
  t82_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n71;
  t82_n70: return -0.12450756132602692f;
  t82_n71: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t82_n75;
  t82_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t82_n74;
  t82_n73: return 0.01325142104178667f;
  t82_n74: return 0.04447650909423828f;
  t82_n75: return -0.1829228550195694f;
  t82_n76: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t82_n100;
  t82_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t82_n79;
  t82_n78: return -0.011999523267149925f;
  t82_n79: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t82_n99;
  t82_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t82_n82;
  t82_n81: return -0.12082025408744812f;
  t82_n82: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t82_n98;
  t82_n83: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t82_n85;
  t82_n84: return 0.40326613187789917f;
  t82_n85: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t82_n87;
  t82_n86: return 0.32386863231658936f;
  t82_n87: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t82_n91;
  t82_n88: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t82_n90;
  t82_n89: return -0.10066814720630646f;
  t82_n90: return 0.4091842770576477f;
  t82_n91: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t82_n93;
  t82_n92: return -0.1015121191740036f;
  t82_n93: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t82_n95;
  t82_n94: return 0.3460303246974945f;
  t82_n95: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t82_n97;
  t82_n96: return -0.10063597559928894f;
  t82_n97: return 0.3715114891529083f;
  t82_n98: return -0.14548887312412262f;
  t82_n99: return -0.07150088995695114f;
  t82_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t82_n148;
  t82_n101: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t82_n105;
  t82_n102: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t82_n104;
  t82_n103: return 0.015893323346972466f;
  t82_n104: return -0.03964640572667122f;
  t82_n105: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t82_n113;
  t82_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t82_n112;
  t82_n107: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t82_n111;
  t82_n108: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t82_n110;
  t82_n109: return -0.10902591049671173f;
  t82_n110: return 0.25025245547294617f;
  t82_n111: return 0.10081485658884048f;
  t82_n112: return 0.13072209060192108f;
  t82_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t82_n141;
  t82_n114: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t82_n136;
  t82_n115: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t82_n135;
  t82_n116: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t82_n126;
  t82_n117: if (E.getNumReferences() >= 3355687776 /*134893.5*/) goto t82_n119;
  t82_n118: return 0.0016037514433264732f;
  t82_n119: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t82_n121;
  t82_n120: return -0.20240333676338196f;
  t82_n121: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t82_n125;
  t82_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n124;
  t82_n123: return -0.1474010944366455f;
  t82_n124: return 0.08961518108844757f;
  t82_n125: return -0.11284840852022171f;
  t82_n126: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t82_n130;
  t82_n127: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t82_n129;
  t82_n128: return -0.1855808049440384f;
  t82_n129: return 0.08229514211416245f;
  t82_n130: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t82_n134;
  t82_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n133;
  t82_n132: return -0.21203510463237762f;
  t82_n133: return 0.10063176602125168f;
  t82_n134: return 0.052385300397872925f;
  t82_n135: return -0.08219332993030548f;
  t82_n136: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t82_n140;
  t82_n137: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t82_n139;
  t82_n138: return 0.037464529275894165f;
  t82_n139: return -0.06637982279062271f;
  t82_n140: return 0.037969425320625305f;
  t82_n141: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t82_n147;
  t82_n142: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t82_n146;
  t82_n143: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t82_n145;
  t82_n144: return 0.027748720720410347f;
  t82_n145: return -0.007910286076366901f;
  t82_n146: return 0.11099489033222198f;
  t82_n147: return -0.055211834609508514f;
  t82_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n150;
  t82_n149: return -0.03992133587598801f;
  t82_n150: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t82_n152;
  t82_n151: return 0.01025482825934887f;
  t82_n152: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t82_n154;
  t82_n153: return 0.09361086040735245f;
  t82_n154: if (E.getNumReferences() >= 3358501856 /*178863.5*/) goto t82_n156;
  t82_n155: return 0.00858441460877657f;
  t82_n156: return 0.04720024764537811f;
  t82_n157: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t82_n169;
  t82_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t82_n160;
  t82_n159: return -0.003663829294964671f;
  t82_n160: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t82_n166;
  t82_n161: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t82_n165;
  t82_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n164;
  t82_n163: return -0.012877063825726509f;
  t82_n164: return 0.020924506708979607f;
  t82_n165: return 0.03360068425536156f;
  t82_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t82_n168;
  t82_n167: return -0.026530671864748f;
  t82_n168: return 0.049125995486974716f;
  t82_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t82_n173;
  t82_n170: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t82_n172;
  t82_n171: return 0.008031471632421017f;
  t82_n172: return 0.03655495122075081f;
  t82_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t82_n175;
  t82_n174: return -0.07980187237262726f;
  t82_n175: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t82_n179;
  t82_n176: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t82_n178;
  t82_n177: return 0.05260862782597542f;
  t82_n178: return 0.00013326479529496282f;
  t82_n179: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t82_n181;
  t82_n180: return 0.08138563483953476f;
  t82_n181: return 0.1273084580898285f;
  t82_n182: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t82_n248;
  t82_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t82_n203;
  t82_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n194;
  t82_n185: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t82_n187;
  t82_n186: return 0.0014825002290308475f;
  t82_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n189;
  t82_n188: return -0.015898829326033592f;
  t82_n189: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t82_n193;
  t82_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t82_n192;
  t82_n191: return -0.036996692419052124f;
  t82_n192: return 0.05253734439611435f;
  t82_n193: return 0.021829208359122276f;
  t82_n194: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t82_n202;
  t82_n195: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t82_n199;
  t82_n196: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t82_n198;
  t82_n197: return -0.19614361226558685f;
  t82_n198: return 0.06804531067609787f;
  t82_n199: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t82_n201;
  t82_n200: return 0.05002228915691376f;
  t82_n201: return -0.017628222703933716f;
  t82_n202: return -0.21177905797958374f;
  t82_n203: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t82_n239;
  t82_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t82_n206;
  t82_n205: return -0.005424083210527897f;
  t82_n206: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t82_n238;
  t82_n207: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t82_n237;
  t82_n208: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t82_n212;
  t82_n209: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t82_n211;
  t82_n210: return -0.006611641030758619f;
  t82_n211: return 0.00816433783620596f;
  t82_n212: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t82_n230;
  t82_n213: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t82_n229;
  t82_n214: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t82_n228;
  t82_n215: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t82_n225;
  t82_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n218;
  t82_n217: return -0.10178021341562271f;
  t82_n218: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t82_n222;
  t82_n219: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t82_n221;
  t82_n220: return 0.012434907257556915f;
  t82_n221: return 0.04787728190422058f;
  t82_n222: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t82_n224;
  t82_n223: return -0.01409226842224598f;
  t82_n224: return 0.11042477935552597f;
  t82_n225: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t82_n227;
  t82_n226: return 0.11874180287122726f;
  t82_n227: return -0.12948010861873627f;
  t82_n228: return -0.021576927974820137f;
  t82_n229: return 0.03839277848601341f;
  t82_n230: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t82_n234;
  t82_n231: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t82_n233;
  t82_n232: return 0.034621693193912506f;
  t82_n233: return -0.06113150715827942f;
  t82_n234: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t82_n236;
  t82_n235: return 0.021171916276216507f;
  t82_n236: return 0.07685965299606323f;
  t82_n237: return -0.1043529063463211f;
  t82_n238: return -0.08370309323072433f;
  t82_n239: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t82_n247;
  t82_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t82_n242;
  t82_n241: return 0.009909510612487793f;
  t82_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t82_n246;
  t82_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t82_n245;
  t82_n244: return 0.03679344430565834f;
  t82_n245: return 0.09088081121444702f;
  t82_n246: return -0.09024187922477722f;
  t82_n247: return 0.033897221088409424f;
  t82_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t82_n250;
  t82_n249: return 0.016281334683299065f;
  t82_n250: if (E.getNumReferences() >= 3315766272 /*5202.5*/) goto t82_n254;
  t82_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t82_n253;
  t82_n252: return -0.033267565071582794f;
  t82_n253: return 0.04322522133588791f;
  t82_n254: return 0.0854635164141655f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree83(const Example& E) {
  t83_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t83_n114;
  t83_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t83_n87;
  t83_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t83_n78;
  t83_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t83_n45;
  t83_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t83_n42;
  t83_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n29;
  t83_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t83_n12;
  t83_n7: if (E.getNumReferences() >= 3301126144 /*1561.5*/) goto t83_n9;
  t83_n8: return 0.008970703929662704f;
  t83_n9: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t83_n11;
  t83_n10: return 0.07300388067960739f;
  t83_n11: return 0.2549280822277069f;
  t83_n12: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t83_n26;
  t83_n13: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t83_n25;
  t83_n14: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t83_n24;
  t83_n15: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t83_n17;
  t83_n16: return -0.008901252411305904f;
  t83_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t83_n19;
  t83_n18: return -0.04057161137461662f;
  t83_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t83_n21;
  t83_n20: return 0.00900606531649828f;
  t83_n21: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t83_n23;
  t83_n22: return 0.04059014841914177f;
  t83_n23: return -0.07362847030162811f;
  t83_n24: return -0.05145223066210747f;
  t83_n25: return 0.04767593368887901f;
  t83_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t83_n28;
  t83_n27: return 0.0107052531093359f;
  t83_n28: return 0.05928286910057068f;
  t83_n29: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t83_n41;
  t83_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t83_n32;
  t83_n31: return -0.10358050465583801f;
  t83_n32: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t83_n38;
  t83_n33: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t83_n37;
  t83_n34: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t83_n36;
  t83_n35: return 0.006845065392553806f;
  t83_n36: return -0.0942729040980339f;
  t83_n37: return -0.05178190767765045f;
  t83_n38: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t83_n40;
  t83_n39: return 0.016016948968172073f;
  t83_n40: return 0.07813804596662521f;
  t83_n41: return 0.03288101777434349f;
  t83_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n44;
  t83_n43: return -0.05371199920773506f;
  t83_n44: return -0.03446461632847786f;
  t83_n45: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t83_n61;
  t83_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t83_n56;
  t83_n47: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t83_n55;
  t83_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t83_n52;
  t83_n49: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t83_n51;
  t83_n50: return 9.496194979874417e-05f;
  t83_n51: return -0.022660640999674797f;
  t83_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t83_n54;
  t83_n53: return -0.07026584446430206f;
  t83_n54: return 0.023612122982740402f;
  t83_n55: return 0.002532492857426405f;
  t83_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n58;
  t83_n57: return -0.0622633695602417f;
  t83_n58: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t83_n60;
  t83_n59: return 0.04310940206050873f;
  t83_n60: return 0.24217751622200012f;
  t83_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t83_n65;
  t83_n62: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t83_n64;
  t83_n63: return 0.13017645478248596f;
  t83_n64: return 0.007751097436994314f;
  t83_n65: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t83_n71;
  t83_n66: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t83_n70;
  t83_n67: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t83_n69;
  t83_n68: return 0.027611946687102318f;
  t83_n69: return 0.007145080715417862f;
  t83_n70: return 0.0711415708065033f;
  t83_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t83_n73;
  t83_n72: return 0.04534980654716492f;
  t83_n73: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t83_n75;
  t83_n74: return 0.029685016721487045f;
  t83_n75: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t83_n77;
  t83_n76: return 0.02831205353140831f;
  t83_n77: return 0.08951093256473541f;
  t83_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t83_n80;
  t83_n79: return -0.05899740383028984f;
  t83_n80: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t83_n82;
  t83_n81: return -0.0024249290581792593f;
  t83_n82: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t83_n84;
  t83_n83: return -0.03618599474430084f;
  t83_n84: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t83_n86;
  t83_n85: return 0.0033386140130460262f;
  t83_n86: return -0.03362831100821495f;
  t83_n87: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t83_n101;
  t83_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n90;
  t83_n89: return -0.006730993278324604f;
  t83_n90: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t83_n94;
  t83_n91: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t83_n93;
  t83_n92: return 0.011751055717468262f;
  t83_n93: return -0.024717923253774643f;
  t83_n94: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t83_n98;
  t83_n95: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t83_n97;
  t83_n96: return -0.08275651931762695f;
  t83_n97: return 0.015126006677746773f;
  t83_n98: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t83_n100;
  t83_n99: return 0.14757303893566132f;
  t83_n100: return -0.24836410582065582f;
  t83_n101: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t83_n113;
  t83_n102: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t83_n112;
  t83_n103: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t83_n111;
  t83_n104: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t83_n110;
  t83_n105: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t83_n109;
  t83_n106: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t83_n108;
  t83_n107: return 0.052437879145145416f;
  t83_n108: return 0.014133669435977936f;
  t83_n109: return 0.11276019364595413f;
  t83_n110: return -0.0067199235782027245f;
  t83_n111: return 0.06491909176111221f;
  t83_n112: return 0.05946284532546997f;
  t83_n113: return 0.19986701011657715f;
  t83_n114: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t83_n252;
  t83_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t83_n133;
  t83_n116: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t83_n130;
  t83_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t83_n119;
  t83_n118: return -0.002050945768132806f;
  t83_n119: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t83_n123;
  t83_n120: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t83_n122;
  t83_n121: return 0.004775497131049633f;
  t83_n122: return -0.010488021187484264f;
  t83_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t83_n127;
  t83_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t83_n126;
  t83_n125: return -0.004685217514634132f;
  t83_n126: return 0.01412256434559822f;
  t83_n127: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t83_n129;
  t83_n128: return -0.1467495858669281f;
  t83_n129: return 0.06565733253955841f;
  t83_n130: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t83_n132;
  t83_n131: return 0.023887919262051582f;
  t83_n132: return 0.006988656707108021f;
  t83_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t83_n141;
  t83_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t83_n138;
  t83_n135: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t83_n137;
  t83_n136: return -0.0022877855226397514f;
  t83_n137: return 0.025283047929406166f;
  t83_n138: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t83_n140;
  t83_n139: return -0.13411898910999298f;
  t83_n140: return 0.017517227679491043f;
  t83_n141: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t83_n251;
  t83_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t83_n170;
  t83_n143: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t83_n163;
  t83_n144: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t83_n148;
  t83_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t83_n147;
  t83_n146: return -0.01743997260928154f;
  t83_n147: return 0.04203725978732109f;
  t83_n148: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t83_n162;
  t83_n149: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t83_n155;
  t83_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t83_n152;
  t83_n151: return 0.0023612468503415585f;
  t83_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t83_n154;
  t83_n153: return 0.04725786671042442f;
  t83_n154: return -0.0228930301964283f;
  t83_n155: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t83_n157;
  t83_n156: return 0.029999947175383568f;
  t83_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t83_n159;
  t83_n158: return 0.04610593616962433f;
  t83_n159: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t83_n161;
  t83_n160: return 0.10053494572639465f;
  t83_n161: return -0.07904684543609619f;
  t83_n162: return -0.1186777651309967f;
  t83_n163: if (E.getNumReferences() >= 3294511104 /*889.0*/) goto t83_n169;
  t83_n164: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t83_n166;
  t83_n165: return 0.00898111891001463f;
  t83_n166: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t83_n168;
  t83_n167: return 0.06053818389773369f;
  t83_n168: return -0.022898487746715546f;
  t83_n169: return 0.051330920308828354f;
  t83_n170: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t83_n228;
  t83_n171: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t83_n221;
  t83_n172: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t83_n182;
  t83_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t83_n177;
  t83_n174: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t83_n176;
  t83_n175: return 0.010547547601163387f;
  t83_n176: return -0.03570061922073364f;
  t83_n177: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t83_n181;
  t83_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n180;
  t83_n179: return -0.011276300996541977f;
  t83_n180: return 0.05629713460803032f;
  t83_n181: return 0.045920971781015396f;
  t83_n182: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t83_n184;
  t83_n183: return -0.022269414737820625f;
  t83_n184: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t83_n220;
  t83_n185: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t83_n191;
  t83_n186: if (E.getNumReferences() >= 3348641152 /*77931.0*/) goto t83_n188;
  t83_n187: return 0.005686512682586908f;
  t83_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n190;
  t83_n189: return -0.10843993723392487f;
  t83_n190: return 0.08894794434309006f;
  t83_n191: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t83_n193;
  t83_n192: return 0.017747310921549797f;
  t83_n193: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t83_n197;
  t83_n194: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t83_n196;
  t83_n195: return -0.013844934292137623f;
  t83_n196: return 0.1383884847164154f;
  t83_n197: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t83_n219;
  t83_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t83_n204;
  t83_n199: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t83_n203;
  t83_n200: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t83_n202;
  t83_n201: return 0.13890905678272247f;
  t83_n202: return -0.14748333394527435f;
  t83_n203: return 0.02806595340371132f;
  t83_n204: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t83_n216;
  t83_n205: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t83_n215;
  t83_n206: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t83_n208;
  t83_n207: return 0.06765784323215485f;
  t83_n208: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t83_n212;
  t83_n209: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t83_n211;
  t83_n210: return 0.15471041202545166f;
  t83_n211: return 0.14732632040977478f;
  t83_n212: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t83_n214;
  t83_n213: return -1.4114047189650591e-05f;
  t83_n214: return 0.15579555928707123f;
  t83_n215: return 0.05448848754167557f;
  t83_n216: if (E.getNumReferences() >= 3265921024 /*85.0*/) goto t83_n218;
  t83_n217: return 0.03741149976849556f;
  t83_n218: return 0.12221258878707886f;
  t83_n219: return 0.014774609357118607f;
  t83_n220: return -0.034016016870737076f;
  t83_n221: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t83_n227;
  t83_n222: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t83_n224;
  t83_n223: return 0.03057035803794861f;
  t83_n224: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t83_n226;
  t83_n225: return 0.057191915810108185f;
  t83_n226: return 0.10487981885671616f;
  t83_n227: return -0.04261041432619095f;
  t83_n228: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t83_n236;
  t83_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n235;
  t83_n230: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t83_n232;
  t83_n231: return 0.002202004427090287f;
  t83_n232: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t83_n234;
  t83_n233: return -0.049727533012628555f;
  t83_n234: return 0.05471126735210419f;
  t83_n235: return 0.016986027359962463f;
  t83_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t83_n238;
  t83_n237: return -0.04963710159063339f;
  t83_n238: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t83_n240;
  t83_n239: return 0.018771300092339516f;
  t83_n240: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t83_n250;
  t83_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t83_n249;
  t83_n242: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t83_n244;
  t83_n243: return 0.06593524664640427f;
  t83_n244: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t83_n248;
  t83_n245: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t83_n247;
  t83_n246: return 0.1481810361146927f;
  t83_n247: return 0.022845063358545303f;
  t83_n248: return 0.07337544858455658f;
  t83_n249: return 0.008258026093244553f;
  t83_n250: return 0.029201913625001907f;
  t83_n251: return -0.0023685165215283632f;
  t83_n252: if (E.getNumReferences() >= 3279126528 /*243.5*/) goto t83_n254;
  t83_n253: return 0.009448782540857792f;
  t83_n254: return 0.03585617616772652f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree84(const Example& E) {
  t84_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t84_n150;
  t84_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t84_n129;
  t84_n2: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t84_n94;
  t84_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t84_n55;
  t84_n4: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t84_n6;
  t84_n5: return 0.0013822587206959724f;
  t84_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t84_n42;
  t84_n7: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t84_n9;
  t84_n8: return -0.01226037461310625f;
  t84_n9: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t84_n11;
  t84_n10: return -0.017206517979502678f;
  t84_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t84_n13;
  t84_n12: return -0.02226472645998001f;
  t84_n13: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t84_n41;
  t84_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t84_n40;
  t84_n15: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t84_n37;
  t84_n16: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t84_n26;
  t84_n17: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t84_n23;
  t84_n18: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t84_n20;
  t84_n19: return -0.0012763517443090677f;
  t84_n20: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t84_n22;
  t84_n21: return 0.01965894177556038f;
  t84_n22: return -0.05591943487524986f;
  t84_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t84_n25;
  t84_n24: return -0.05352277308702469f;
  t84_n25: return -0.005794072523713112f;
  t84_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t84_n36;
  t84_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t84_n35;
  t84_n28: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t84_n32;
  t84_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t84_n31;
  t84_n30: return -0.08610742539167404f;
  t84_n31: return 0.02001538872718811f;
  t84_n32: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t84_n34;
  t84_n33: return -0.09113404899835587f;
  t84_n34: return 0.00016094897000584751f;
  t84_n35: return 0.03810109570622444f;
  t84_n36: return -0.05535196140408516f;
  t84_n37: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t84_n39;
  t84_n38: return 0.029078306630253792f;
  t84_n39: return -0.023605598136782646f;
  t84_n40: return 0.01283484697341919f;
  t84_n41: return -0.11444713920354843f;
  t84_n42: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t84_n44;
  t84_n43: return -0.08179159462451935f;
  t84_n44: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t84_n46;
  t84_n45: return 0.010915175080299377f;
  t84_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t84_n48;
  t84_n47: return -0.04792900010943413f;
  t84_n48: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t84_n50;
  t84_n49: return 0.2755584716796875f;
  t84_n50: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t84_n54;
  t84_n51: if (E.getNumReferences() >= 3249537024 /*22.0*/) goto t84_n53;
  t84_n52: return 0.15661346912384033f;
  t84_n53: return -0.10048084706068039f;
  t84_n54: return 0.14748428761959076f;
  t84_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t84_n57;
  t84_n56: return -0.0904945507645607f;
  t84_n57: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t84_n67;
  t84_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t84_n66;
  t84_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t84_n61;
  t84_n60: return -0.01409638486802578f;
  t84_n61: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t84_n63;
  t84_n62: return -0.0007305468316189945f;
  t84_n63: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t84_n65;
  t84_n64: return 0.012326414696872234f;
  t84_n65: return -0.00098345463629812f;
  t84_n66: return 0.017937159165740013f;
  t84_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t84_n79;
  t84_n68: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t84_n76;
  t84_n69: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t84_n75;
  t84_n70: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t84_n74;
  t84_n71: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t84_n73;
  t84_n72: return 0.17613060772418976f;
  t84_n73: return -0.14404554665088654f;
  t84_n74: return 0.010715101845562458f;
  t84_n75: return 0.05151320993900299f;
  t84_n76: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t84_n78;
  t84_n77: return -0.0071322559379041195f;
  t84_n78: return 0.006261791102588177f;
  t84_n79: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t84_n83;
  t84_n80: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t84_n82;
  t84_n81: return 0.02189517579972744f;
  t84_n82: return -0.01870841160416603f;
  t84_n83: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t84_n89;
  t84_n84: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t84_n86;
  t84_n85: return 0.020212313160300255f;
  t84_n86: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t84_n88;
  t84_n87: return -0.027563177049160004f;
  t84_n88: return 0.03186932951211929f;
  t84_n89: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t84_n91;
  t84_n90: return 0.02847764454782009f;
  t84_n91: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t84_n93;
  t84_n92: return 0.11435346305370331f;
  t84_n93: return -0.06066490709781647f;
  t84_n94: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t84_n96;
  t84_n95: return 0.013418289832770824f;
  t84_n96: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t84_n120;
  t84_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t84_n99;
  t84_n98: return -0.010921857319772243f;
  t84_n99: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t84_n119;
  t84_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t84_n102;
  t84_n101: return -0.11934822797775269f;
  t84_n102: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t84_n118;
  t84_n103: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t84_n105;
  t84_n104: return 0.31834638118743896f;
  t84_n105: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t84_n107;
  t84_n106: return 0.22801586985588074f;
  t84_n107: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t84_n111;
  t84_n108: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t84_n110;
  t84_n109: return -0.10065057128667831f;
  t84_n110: return 0.2253081202507019f;
  t84_n111: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t84_n113;
  t84_n112: return -0.10137733817100525f;
  t84_n113: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t84_n115;
  t84_n114: return 0.26234516501426697f;
  t84_n115: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t84_n117;
  t84_n116: return 0.3020428717136383f;
  t84_n117: return 0.34273582696914673f;
  t84_n118: return -0.12573587894439697f;
  t84_n119: return -0.06972651928663254f;
  t84_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t84_n128;
  t84_n121: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t84_n125;
  t84_n122: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t84_n124;
  t84_n123: return 0.01421463955193758f;
  t84_n124: return -0.03722155839204788f;
  t84_n125: if (E.getSymbolScopeDistanceCost() >= 3250061312 /*23.0*/) goto t84_n127;
  t84_n126: return 0.09961158037185669f;
  t84_n127: return 0.002067916328087449f;
  t84_n128: return 0.023924557492136955f;
  t84_n129: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t84_n137;
  t84_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t84_n132;
  t84_n131: return -0.020158931612968445f;
  t84_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t84_n136;
  t84_n133: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t84_n135;
  t84_n134: return 0.003131677396595478f;
  t84_n135: return -0.07652521133422852f;
  t84_n136: return 0.048989187926054f;
  t84_n137: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t84_n147;
  t84_n138: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t84_n144;
  t84_n139: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t84_n141;
  t84_n140: return 0.029377946630120277f;
  t84_n141: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t84_n143;
  t84_n142: return 0.0033680929336696863f;
  t84_n143: return 0.0390484519302845f;
  t84_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t84_n146;
  t84_n145: return 0.010199769400060177f;
  t84_n146: return 0.037942368537187576f;
  t84_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t84_n149;
  t84_n148: return 0.03127995878458023f;
  t84_n149: return 0.06924059242010117f;
  t84_n150: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t84_n234;
  t84_n151: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t84_n217;
  t84_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t84_n188;
  t84_n153: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t84_n187;
  t84_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t84_n162;
  t84_n155: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t84_n159;
  t84_n156: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t84_n158;
  t84_n157: return -0.023960283026099205f;
  t84_n158: return -0.16456758975982666f;
  t84_n159: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t84_n161;
  t84_n160: return 0.04957108944654465f;
  t84_n161: return 0.1301698535680771f;
  t84_n162: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t84_n186;
  t84_n163: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t84_n177;
  t84_n164: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t84_n176;
  t84_n165: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t84_n175;
  t84_n166: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t84_n174;
  t84_n167: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t84_n169;
  t84_n168: return 0.004281159024685621f;
  t84_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t84_n173;
  t84_n170: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t84_n172;
  t84_n171: return -0.031103530898690224f;
  t84_n172: return -0.08499240130186081f;
  t84_n173: return -0.012862415984272957f;
  t84_n174: return 0.040917910635471344f;
  t84_n175: return -0.07594611495733261f;
  t84_n176: return 0.024942860007286072f;
  t84_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t84_n183;
  t84_n178: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t84_n180;
  t84_n179: return -0.024187346920371056f;
  t84_n180: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t84_n182;
  t84_n181: return -0.07664822041988373f;
  t84_n182: return 0.017648188397288322f;
  t84_n183: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t84_n185;
  t84_n184: return -0.03312776982784271f;
  t84_n185: return 0.05263914540410042f;
  t84_n186: return 0.007829184643924236f;
  t84_n187: return 0.023872971534729004f;
  t84_n188: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t84_n204;
  t84_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t84_n193;
  t84_n190: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t84_n192;
  t84_n191: return 0.01909453235566616f;
  t84_n192: return 0.001786035019904375f;
  t84_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t84_n195;
  t84_n194: return -0.09044087678194046f;
  t84_n195: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t84_n199;
  t84_n196: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t84_n198;
  t84_n197: return 0.03913275524973869f;
  t84_n198: return -0.02477591671049595f;
  t84_n199: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t84_n203;
  t84_n200: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t84_n202;
  t84_n201: return 0.07400188595056534f;
  t84_n202: return 0.16240811347961426f;
  t84_n203: return 0.07614489644765854f;
  t84_n204: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t84_n216;
  t84_n205: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t84_n207;
  t84_n206: return 0.008063708432018757f;
  t84_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t84_n209;
  t84_n208: return 0.01746373251080513f;
  t84_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t84_n211;
  t84_n210: return -0.10333796590566635f;
  t84_n211: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t84_n213;
  t84_n212: return 0.024172160774469376f;
  t84_n213: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t84_n215;
  t84_n214: return 0.08320572227239609f;
  t84_n215: return 0.0433095283806324f;
  t84_n216: return 0.009449227713048458f;
  t84_n217: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t84_n231;
  t84_n218: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t84_n230;
  t84_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t84_n229;
  t84_n220: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t84_n222;
  t84_n221: return 0.010090354830026627f;
  t84_n222: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t84_n226;
  t84_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t84_n225;
  t84_n224: return 0.051800210028886795f;
  t84_n225: return 0.10365224629640579f;
  t84_n226: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t84_n228;
  t84_n227: return -0.088736392557621f;
  t84_n228: return 0.08472659438848495f;
  t84_n229: return -0.0351925827562809f;
  t84_n230: return 0.04038051888346672f;
  t84_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t84_n233;
  t84_n232: return 0.012622755020856857f;
  t84_n233: return 0.04903365299105644f;
  t84_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t84_n240;
  t84_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t84_n237;
  t84_n236: return -1.615786095499061e-05f;
  t84_n237: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t84_n239;
  t84_n238: return 0.01700015179812908f;
  t84_n239: return -0.004032541997730732f;
  t84_n240: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t84_n246;
  t84_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t84_n243;
  t84_n242: return -0.008363483473658562f;
  t84_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t84_n245;
  t84_n244: return -0.0023495673667639494f;
  t84_n245: return 0.014137259684503078f;
  t84_n246: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t84_n254;
  t84_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t84_n253;
  t84_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t84_n250;
  t84_n249: return 0.009950903244316578f;
  t84_n250: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t84_n252;
  t84_n251: return 0.060928646475076675f;
  t84_n252: return -0.08259764313697815f;
  t84_n253: return -0.11819673329591751f;
  t84_n254: return 0.032161224633455276f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree85(const Example& E) {
  t85_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t85_n112;
  t85_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t85_n95;
  t85_n2: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t85_n94;
  t85_n3: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t85_n93;
  t85_n4: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t85_n76;
  t85_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t85_n59;
  t85_n6: if (E.getNumReferences() >= 3274932224 /*179.5*/) goto t85_n16;
  t85_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n15;
  t85_n8: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t85_n10;
  t85_n9: return -0.007725180592387915f;
  t85_n10: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t85_n12;
  t85_n11: return 0.019590388983488083f;
  t85_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n14;
  t85_n13: return -0.0401458740234375f;
  t85_n14: return 0.005137166939675808f;
  t85_n15: return 0.0011315536685287952f;
  t85_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t85_n28;
  t85_n17: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t85_n27;
  t85_n18: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t85_n26;
  t85_n19: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t85_n25;
  t85_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t85_n24;
  t85_n21: if (E.getNumReferences() >= 3355264064 /*129672.5*/) goto t85_n23;
  t85_n22: return 0.017635254189372063f;
  t85_n23: return 0.08090194314718246f;
  t85_n24: return -0.0028419988229870796f;
  t85_n25: return -0.20427066087722778f;
  t85_n26: return -0.012475001625716686f;
  t85_n27: return -0.15765027701854706f;
  t85_n28: if (E.getNumReferences() >= 3354076032 /*120391.0*/) goto t85_n54;
  t85_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t85_n31;
  t85_n30: return -0.03835533931851387f;
  t85_n31: if (E.getNumReferences() >= 3340697088 /*40702.0*/) goto t85_n49;
  t85_n32: if (E.getNumReferences() >= 3338715264 /*32960.5*/) goto t85_n46;
  t85_n33: if (E.getNumReferences() >= 3332434944 /*20598.0*/) goto t85_n45;
  t85_n34: if (E.getNumReferences() >= 3332116224 /*19975.5*/) goto t85_n36;
  t85_n35: return 0.03399019315838814f;
  t85_n36: if (E.getNumReferences() >= 3332216832 /*20172.0*/) goto t85_n44;
  t85_n37: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t85_n43;
  t85_n38: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t85_n42;
  t85_n39: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t85_n41;
  t85_n40: return 0.11149975657463074f;
  t85_n41: return 0.18029408156871796f;
  t85_n42: return 0.08498172461986542f;
  t85_n43: return 0.14316995441913605f;
  t85_n44: return -0.004819488152861595f;
  t85_n45: return -0.08922110497951508f;
  t85_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t85_n48;
  t85_n47: return 0.03227987140417099f;
  t85_n48: return 0.11651206761598587f;
  t85_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t85_n51;
  t85_n50: return -0.1389978528022766f;
  t85_n51: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t85_n53;
  t85_n52: return -0.038067542016506195f;
  t85_n53: return 0.07429900020360947f;
  t85_n54: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t85_n56;
  t85_n55: return -0.08228801935911179f;
  t85_n56: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t85_n58;
  t85_n57: return 0.12947122752666473f;
  t85_n58: return 0.056811973452568054f;
  t85_n59: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t85_n61;
  t85_n60: return 0.01310733426362276f;
  t85_n61: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t85_n75;
  t85_n62: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t85_n70;
  t85_n63: if (E.getNumReferences() >= 3357252288 /*159339.0*/) goto t85_n69;
  t85_n64: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t85_n66;
  t85_n65: return -0.008171482011675835f;
  t85_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n68;
  t85_n67: return -0.04417373985052109f;
  t85_n68: return 0.01322454959154129f;
  t85_n69: return -0.2048458456993103f;
  t85_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n72;
  t85_n71: return -0.20300015807151794f;
  t85_n72: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t85_n74;
  t85_n73: return 0.010861390270292759f;
  t85_n74: return 0.0910612940788269f;
  t85_n75: return -0.12101595103740692f;
  t85_n76: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t85_n86;
  t85_n77: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t85_n81;
  t85_n78: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t85_n80;
  t85_n79: return -0.01778268814086914f;
  t85_n80: return 0.044713884592056274f;
  t85_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t85_n83;
  t85_n82: return -0.16966216266155243f;
  t85_n83: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t85_n85;
  t85_n84: return 0.0603070892393589f;
  t85_n85: return 0.12649019062519073f;
  t85_n86: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t85_n88;
  t85_n87: return -0.08391908556222916f;
  t85_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t85_n90;
  t85_n89: return 0.004373996518552303f;
  t85_n90: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t85_n92;
  t85_n91: return 0.0793502926826477f;
  t85_n92: return 0.11366226524114609f;
  t85_n93: return -0.07622995227575302f;
  t85_n94: return 0.02154582366347313f;
  t85_n95: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t85_n103;
  t85_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n98;
  t85_n97: return -0.004569431766867638f;
  t85_n98: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t85_n102;
  t85_n99: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t85_n101;
  t85_n100: return 0.011900787241756916f;
  t85_n101: return -0.020048517733812332f;
  t85_n102: return 0.019897904247045517f;
  t85_n103: if (E.getNumReferences() >= 3331945984 /*19643.0*/) goto t85_n105;
  t85_n104: return 0.1278916299343109f;
  t85_n105: if (E.getNumReferences() >= 3362781216 /*245728.5*/) goto t85_n111;
  t85_n106: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t85_n108;
  t85_n107: return 0.028393277898430824f;
  t85_n108: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t85_n110;
  t85_n109: return 0.10084301233291626f;
  t85_n110: return 0.11356241255998611f;
  t85_n111: return -0.34685859084129333f;
  t85_n112: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t85_n210;
  t85_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t85_n115;
  t85_n114: return -0.1094171479344368f;
  t85_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t85_n143;
  t85_n116: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t85_n142;
  t85_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t85_n127;
  t85_n118: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t85_n124;
  t85_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t85_n121;
  t85_n120: return -0.00024438006221316755f;
  t85_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t85_n123;
  t85_n122: return 0.0045108734630048275f;
  t85_n123: return 0.013296789489686489f;
  t85_n124: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t85_n126;
  t85_n125: return 0.019972560927271843f;
  t85_n126: return -0.0024161965120583773f;
  t85_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t85_n139;
  t85_n128: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t85_n134;
  t85_n129: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t85_n131;
  t85_n130: return 0.012533902190625668f;
  t85_n131: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t85_n133;
  t85_n132: return -0.032236725091934204f;
  t85_n133: return 0.05438368767499924f;
  t85_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t85_n136;
  t85_n135: return -0.010508357547223568f;
  t85_n136: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t85_n138;
  t85_n137: return 0.004119907505810261f;
  t85_n138: return 0.04077116772532463f;
  t85_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t85_n141;
  t85_n140: return -0.06502529233694077f;
  t85_n141: return 0.007228025700896978f;
  t85_n142: return 0.06533077359199524f;
  t85_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t85_n183;
  t85_n144: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t85_n182;
  t85_n145: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t85_n181;
  t85_n146: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t85_n164;
  t85_n147: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t85_n149;
  t85_n148: return -0.0097345020622015f;
  t85_n149: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t85_n155;
  t85_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t85_n154;
  t85_n151: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t85_n153;
  t85_n152: return 0.045828476548194885f;
  t85_n153: return -0.25013110041618347f;
  t85_n154: return -0.01034832838922739f;
  t85_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t85_n163;
  t85_n156: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t85_n160;
  t85_n157: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t85_n159;
  t85_n158: return -0.7414761781692505f;
  t85_n159: return -0.1442653238773346f;
  t85_n160: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t85_n162;
  t85_n161: return 0.13616101443767548f;
  t85_n162: return -0.10833220183849335f;
  t85_n163: return -0.04409041255712509f;
  t85_n164: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t85_n166;
  t85_n165: return 0.09779977053403854f;
  t85_n166: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t85_n180;
  t85_n167: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t85_n171;
  t85_n168: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t85_n170;
  t85_n169: return 0.003205025801435113f;
  t85_n170: return -0.014007609337568283f;
  t85_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t85_n179;
  t85_n172: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t85_n174;
  t85_n173: return -0.004399505909532309f;
  t85_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t85_n176;
  t85_n175: return -0.07138168066740036f;
  t85_n176: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t85_n178;
  t85_n177: return 0.04047752916812897f;
  t85_n178: return 0.13338881731033325f;
  t85_n179: return 0.037018924951553345f;
  t85_n180: return -0.11568104475736618f;
  t85_n181: return 0.012051108293235302f;
  t85_n182: return -0.03624146431684494f;
  t85_n183: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t85_n209;
  t85_n184: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t85_n208;
  t85_n185: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t85_n187;
  t85_n186: return 0.011584657244384289f;
  t85_n187: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t85_n205;
  t85_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t85_n204;
  t85_n189: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t85_n191;
  t85_n190: return -0.026947205886244774f;
  t85_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t85_n193;
  t85_n192: return 0.0015972215915098786f;
  t85_n193: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t85_n195;
  t85_n194: return -0.11738459765911102f;
  t85_n195: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t85_n203;
  t85_n196: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t85_n198;
  t85_n197: return 0.038870666176080704f;
  t85_n198: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t85_n202;
  t85_n199: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t85_n201;
  t85_n200: return -0.07724855840206146f;
  t85_n201: return 0.11346445232629776f;
  t85_n202: return 0.04872498661279678f;
  t85_n203: return 0.13342420756816864f;
  t85_n204: return 0.060826774686574936f;
  t85_n205: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t85_n207;
  t85_n206: return 0.0837743729352951f;
  t85_n207: return 0.013817919418215752f;
  t85_n208: return 0.0005926173180341721f;
  t85_n209: return 0.023321617394685745f;
  t85_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t85_n212;
  t85_n211: return -0.15507401525974274f;
  t85_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t85_n220;
  t85_n213: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t85_n215;
  t85_n214: return 0.001540403114631772f;
  t85_n215: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t85_n217;
  t85_n216: return -0.040542252361774445f;
  t85_n217: if (E.getFractionNameInContext() >= 3206445544 /*0.6190476417541504*/) goto t85_n219;
  t85_n218: return 0.11330535262823105f;
  t85_n219: return -0.19511747360229492f;
  t85_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t85_n252;
  t85_n221: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t85_n241;
  t85_n222: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t85_n224;
  t85_n223: return 0.015558954328298569f;
  t85_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t85_n226;
  t85_n225: return -0.007319084834307432f;
  t85_n226: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t85_n234;
  t85_n227: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t85_n229;
  t85_n228: return -0.00926700048148632f;
  t85_n229: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t85_n233;
  t85_n230: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t85_n232;
  t85_n231: return 0.06037178263068199f;
  t85_n232: return 0.0011183461174368858f;
  t85_n233: return -0.001096114399842918f;
  t85_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t85_n240;
  t85_n235: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t85_n239;
  t85_n236: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t85_n238;
  t85_n237: return 0.06298453360795975f;
  t85_n238: return -0.046533796936273575f;
  t85_n239: return -0.006552278995513916f;
  t85_n240: return 0.014444495551288128f;
  t85_n241: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t85_n251;
  t85_n242: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t85_n244;
  t85_n243: return 0.09606640785932541f;
  t85_n244: if (E.getNumReferences() >= 3339206144 /*34878.0*/) goto t85_n248;
  t85_n245: if (E.getNumReferences() >= 3338675456 /*32805.0*/) goto t85_n247;
  t85_n246: return 0.031260766088962555f;
  t85_n247: return -0.15410682559013367f;
  t85_n248: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t85_n250;
  t85_n249: return 0.12767764925956726f;
  t85_n250: return 0.11857467144727707f;
  t85_n251: return 0.017944924533367157f;
  t85_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t85_n254;
  t85_n253: return 0.02054690383374691f;
  t85_n254: return -0.15967817604541779f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree86(const Example& E) {
  t86_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t86_n132;
  t86_n1: if (E.getNumReferences() >= 3325315072 /*11538.0*/) goto t86_n99;
  t86_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t86_n74;
  t86_n3: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t86_n71;
  t86_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n12;
  t86_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t86_n7;
  t86_n6: return -0.007033057045191526f;
  t86_n7: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t86_n11;
  t86_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t86_n10;
  t86_n9: return 0.0724867507815361f;
  t86_n10: return 0.008029882796108723f;
  t86_n11: return 0.10300213098526001f;
  t86_n12: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t86_n66;
  t86_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t86_n15;
  t86_n14: return -0.05199434980750084f;
  t86_n15: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t86_n65;
  t86_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t86_n20;
  t86_n17: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t86_n19;
  t86_n18: return 0.004520036280155182f;
  t86_n19: return -0.05601249262690544f;
  t86_n20: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t86_n22;
  t86_n21: return -0.0035490719601511955f;
  t86_n22: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t86_n28;
  t86_n23: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t86_n25;
  t86_n24: return -0.01605827920138836f;
  t86_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n27;
  t86_n26: return -0.052547622472047806f;
  t86_n27: return 0.03139093890786171f;
  t86_n28: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t86_n44;
  t86_n29: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t86_n31;
  t86_n30: return -0.004967608954757452f;
  t86_n31: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t86_n41;
  t86_n32: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t86_n38;
  t86_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t86_n37;
  t86_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n36;
  t86_n35: return -0.01899518445134163f;
  t86_n36: return 0.050504233688116074f;
  t86_n37: return 0.06610040366649628f;
  t86_n38: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t86_n40;
  t86_n39: return 0.02566816844046116f;
  t86_n40: return 0.11763552576303482f;
  t86_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t86_n43;
  t86_n42: return -0.03376239910721779f;
  t86_n43: return 0.04931572824716568f;
  t86_n44: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t86_n62;
  t86_n45: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t86_n47;
  t86_n46: return -0.04887311905622482f;
  t86_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t86_n53;
  t86_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t86_n50;
  t86_n49: return 0.07932280749082565f;
  t86_n50: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t86_n52;
  t86_n51: return 0.17048682272434235f;
  t86_n52: return 0.10224788635969162f;
  t86_n53: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t86_n55;
  t86_n54: return 0.06801079213619232f;
  t86_n55: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t86_n59;
  t86_n56: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t86_n58;
  t86_n57: return -0.14897125959396362f;
  t86_n58: return 0.11616657674312592f;
  t86_n59: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t86_n61;
  t86_n60: return 0.14838865399360657f;
  t86_n61: return 0.0991385355591774f;
  t86_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t86_n64;
  t86_n63: return 0.03040413372218609f;
  t86_n64: return 0.11840635538101196f;
  t86_n65: return -0.007818339392542839f;
  t86_n66: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t86_n70;
  t86_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t86_n69;
  t86_n68: return -0.10123062133789062f;
  t86_n69: return -0.010660517029464245f;
  t86_n70: return -0.05892222002148628f;
  t86_n71: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t86_n73;
  t86_n72: return 0.008043603971600533f;
  t86_n73: return 0.03018505685031414f;
  t86_n74: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t86_n86;
  t86_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t86_n79;
  t86_n76: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t86_n78;
  t86_n77: return -0.0025596774648875f;
  t86_n78: return 0.03534915670752525f;
  t86_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t86_n85;
  t86_n80: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t86_n84;
  t86_n81: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t86_n83;
  t86_n82: return -0.10722766816616058f;
  t86_n83: return 0.12409104406833649f;
  t86_n84: return 0.011946827173233032f;
  t86_n85: return 0.04016444832086563f;
  t86_n86: if (E.getNumReferences() >= 3324400128 /*10644.5*/) goto t86_n98;
  t86_n87: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t86_n91;
  t86_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t86_n90;
  t86_n89: return -0.11297149211168289f;
  t86_n90: return -0.003964717499911785f;
  t86_n91: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t86_n95;
  t86_n92: if (E.getNumReferences() >= 3319284736 /*6920.5*/) goto t86_n94;
  t86_n93: return 0.007080379873514175f;
  t86_n94: return -0.24926874041557312f;
  t86_n95: if (E.getNumReferences() >= 3305279488 /*2089.0*/) goto t86_n97;
  t86_n96: return -0.034774601459503174f;
  t86_n97: return 0.010982728563249111f;
  t86_n98: return -0.07866649329662323f;
  t86_n99: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t86_n105;
  t86_n100: if (E.getNumReferences() >= 3325491712 /*11710.5*/) goto t86_n104;
  t86_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t86_n103;
  t86_n102: return 0.09321552515029907f;
  t86_n103: return 0.10584980994462967f;
  t86_n104: return 0.019755670800805092f;
  t86_n105: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t86_n107;
  t86_n106: return -0.055820051580667496f;
  t86_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t86_n109;
  t86_n108: return 0.0029887985438108444f;
  t86_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t86_n129;
  t86_n110: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t86_n116;
  t86_n111: if (E.getNumReferences() >= 3365435344 /*312254.5*/) goto t86_n115;
  t86_n112: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t86_n114;
  t86_n113: return 0.013345605693757534f;
  t86_n114: return -0.9577898383140564f;
  t86_n115: return 0.06652317941188812f;
  t86_n116: if (E.getNumReferences() >= 3333346816 /*22379.0*/) goto t86_n122;
  t86_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t86_n119;
  t86_n118: return 0.10735572129487991f;
  t86_n119: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t86_n121;
  t86_n120: return 0.11964883655309677f;
  t86_n121: return 0.18991278111934662f;
  t86_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n124;
  t86_n123: return -0.10796017944812775f;
  t86_n124: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t86_n128;
  t86_n125: if (E.getNumReferences() >= 3340838912 /*41256.0*/) goto t86_n127;
  t86_n126: return 0.03581313416361809f;
  t86_n127: return 0.12491215020418167f;
  t86_n128: return -0.015022977255284786f;
  t86_n129: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t86_n131;
  t86_n130: return -0.03834113851189613f;
  t86_n131: return 0.09952682256698608f;
  t86_n132: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t86_n252;
  t86_n133: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t86_n225;
  t86_n134: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t86_n208;
  t86_n135: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t86_n143;
  t86_n136: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t86_n138;
  t86_n137: return 0.004823712632060051f;
  t86_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t86_n142;
  t86_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t86_n141;
  t86_n140: return 0.001727903145365417f;
  t86_n141: return 0.028043262660503387f;
  t86_n142: return 0.02559327520430088f;
  t86_n143: if (E.getNumReferences() >= 3336376576 /*28296.5*/) goto t86_n205;
  t86_n144: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t86_n204;
  t86_n145: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t86_n193;
  t86_n146: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t86_n192;
  t86_n147: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t86_n167;
  t86_n148: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t86_n156;
  t86_n149: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t86_n151;
  t86_n150: return 0.0010526153491809964f;
  t86_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n155;
  t86_n152: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t86_n154;
  t86_n153: return -0.016553277149796486f;
  t86_n154: return -0.042357105761766434f;
  t86_n155: return 0.006424546707421541f;
  t86_n156: if (E.getNumReferences() >= 3319284736 /*6920.5*/) goto t86_n164;
  t86_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n163;
  t86_n158: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t86_n162;
  t86_n159: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t86_n161;
  t86_n160: return -0.01260618306696415f;
  t86_n161: return 0.06859584152698517f;
  t86_n162: return -0.03868495672941208f;
  t86_n163: return 0.02450486645102501f;
  t86_n164: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t86_n166;
  t86_n165: return -0.33684366941452026f;
  t86_n166: return -0.015851732343435287f;
  t86_n167: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t86_n191;
  t86_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t86_n180;
  t86_n169: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t86_n173;
  t86_n170: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t86_n172;
  t86_n171: return -0.0021201998461037874f;
  t86_n172: return 0.022123103961348534f;
  t86_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t86_n179;
  t86_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t86_n176;
  t86_n175: return -0.05718852952122688f;
  t86_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n178;
  t86_n177: return -0.11301922053098679f;
  t86_n178: return 0.00996681209653616f;
  t86_n179: return 0.03271286189556122f;
  t86_n180: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t86_n184;
  t86_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t86_n183;
  t86_n182: return 0.023859789595007896f;
  t86_n183: return 0.08314190804958344f;
  t86_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t86_n186;
  t86_n185: return 0.013755827210843563f;
  t86_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t86_n188;
  t86_n187: return 0.07701089978218079f;
  t86_n188: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t86_n190;
  t86_n189: return 0.14222735166549683f;
  t86_n190: return 0.08766666054725647f;
  t86_n191: return -0.013657203875482082f;
  t86_n192: return -0.029795106500387192f;
  t86_n193: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t86_n199;
  t86_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t86_n196;
  t86_n195: return 0.056926533579826355f;
  t86_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t86_n198;
  t86_n197: return 0.017642587423324585f;
  t86_n198: return 0.10496234893798828f;
  t86_n199: if (E.getNumReferences() >= 3334782720 /*25183.5*/) goto t86_n203;
  t86_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t86_n202;
  t86_n201: return -0.1831531524658203f;
  t86_n202: return 0.09730623662471771f;
  t86_n203: return 0.04466910660266876f;
  t86_n204: return -0.08932241797447205f;
  t86_n205: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t86_n207;
  t86_n206: return 0.022628435865044594f;
  t86_n207: return -0.12884798645973206f;
  t86_n208: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t86_n216;
  t86_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t86_n213;
  t86_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n212;
  t86_n211: return -0.5845083594322205f;
  t86_n212: return 0.1455819457769394f;
  t86_n213: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t86_n215;
  t86_n214: return 0.1456572413444519f;
  t86_n215: return 0.10872510075569153f;
  t86_n216: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t86_n224;
  t86_n217: if (E.getNumReferences() >= 3339148928 /*34654.5*/) goto t86_n219;
  t86_n218: return -0.09109344333410263f;
  t86_n219: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t86_n221;
  t86_n220: return 0.03592453896999359f;
  t86_n221: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t86_n223;
  t86_n222: return 0.12193930149078369f;
  t86_n223: return 0.1175723746418953f;
  t86_n224: return -0.0076970551162958145f;
  t86_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t86_n239;
  t86_n226: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t86_n236;
  t86_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t86_n229;
  t86_n228: return 0.010267307050526142f;
  t86_n229: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t86_n231;
  t86_n230: return 0.03198930621147156f;
  t86_n231: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t86_n235;
  t86_n232: if (E.getFractionNameInContext() >= 3205668418 /*0.5727273225784302*/) goto t86_n234;
  t86_n233: return 0.12171182781457901f;
  t86_n234: return 0.13856908679008484f;
  t86_n235: return -0.07103818655014038f;
  t86_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n238;
  t86_n237: return -0.031857579946517944f;
  t86_n238: return 0.010273991152644157f;
  t86_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n249;
  t86_n240: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t86_n244;
  t86_n241: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t86_n243;
  t86_n242: return 0.2067624181509018f;
  t86_n243: return -0.08155786991119385f;
  t86_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t86_n246;
  t86_n245: return -0.11641596257686615f;
  t86_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t86_n248;
  t86_n247: return -0.32104435563087463f;
  t86_n248: return 0.06993451714515686f;
  t86_n249: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t86_n251;
  t86_n250: return 0.011945654638111591f;
  t86_n251: return -0.10798835009336472f;
  t86_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t86_n254;
  t86_n253: return 0.01491138618439436f;
  t86_n254: return -0.18811452388763428f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree87(const Example& E) {
  t87_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t87_n142;
  t87_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t87_n23;
  t87_n2: if (E.getNumReferences() >= 3314254848 /*4464.5*/) goto t87_n22;
  t87_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t87_n5;
  t87_n4: return -0.0007731445366516709f;
  t87_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t87_n19;
  t87_n6: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t87_n16;
  t87_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n15;
  t87_n8: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t87_n10;
  t87_n9: return -0.051820412278175354f;
  t87_n10: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t87_n12;
  t87_n11: return -1.4061232805252075f;
  t87_n12: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n14;
  t87_n13: return -0.5520821809768677f;
  t87_n14: return 0.043808408081531525f;
  t87_n15: return -0.016089800745248795f;
  t87_n16: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t87_n18;
  t87_n17: return 0.005960213951766491f;
  t87_n18: return 0.03793603554368019f;
  t87_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t87_n21;
  t87_n20: return 0.009937891736626625f;
  t87_n21: return 0.09393381327390671f;
  t87_n22: return 0.019464081153273582f;
  t87_n23: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t87_n103;
  t87_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t87_n70;
  t87_n25: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t87_n69;
  t87_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t87_n42;
  t87_n27: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t87_n29;
  t87_n28: return -0.009595268405973911f;
  t87_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n39;
  t87_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t87_n36;
  t87_n31: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t87_n35;
  t87_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t87_n34;
  t87_n33: return -0.05496186763048172f;
  t87_n34: return 0.010881693102419376f;
  t87_n35: return 0.0008411778253503144f;
  t87_n36: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t87_n38;
  t87_n37: return 0.04300696775317192f;
  t87_n38: return -0.024767711758613586f;
  t87_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t87_n41;
  t87_n40: return -0.060501065105199814f;
  t87_n41: return 0.014830326661467552f;
  t87_n42: if (E.getNumReferences() >= 3300700160 /*1509.5*/) goto t87_n68;
  t87_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t87_n53;
  t87_n44: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t87_n48;
  t87_n45: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t87_n47;
  t87_n46: return -0.33415067195892334f;
  t87_n47: return -0.007007909938693047f;
  t87_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t87_n50;
  t87_n49: return -0.06865096837282181f;
  t87_n50: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t87_n52;
  t87_n51: return 0.047864094376564026f;
  t87_n52: return 0.01762910932302475f;
  t87_n53: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t87_n67;
  t87_n54: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t87_n58;
  t87_n55: if (E.getNumReferences() >= 3300491264 /*1484.0*/) goto t87_n57;
  t87_n56: return 0.015880534425377846f;
  t87_n57: return 0.13880570232868195f;
  t87_n58: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t87_n66;
  t87_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t87_n61;
  t87_n60: return 0.004211821593344212f;
  t87_n61: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t87_n63;
  t87_n62: return -0.05737824738025665f;
  t87_n63: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t87_n65;
  t87_n64: return 0.07099709659814835f;
  t87_n65: return 0.09274351596832275f;
  t87_n66: return -0.010463079437613487f;
  t87_n67: return 0.016880223527550697f;
  t87_n68: return -0.022117627784609795f;
  t87_n69: return 0.00733533687889576f;
  t87_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t87_n72;
  t87_n71: return -0.07524513453245163f;
  t87_n72: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t87_n82;
  t87_n73: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t87_n77;
  t87_n74: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t87_n76;
  t87_n75: return 0.09043637663125992f;
  t87_n76: return 0.14110088348388672f;
  t87_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n81;
  t87_n78: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t87_n80;
  t87_n79: return 0.007692310493439436f;
  t87_n80: return -0.0385291688144207f;
  t87_n81: return 0.0049448395147919655f;
  t87_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n90;
  t87_n83: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t87_n85;
  t87_n84: return -0.09083044528961182f;
  t87_n85: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t87_n89;
  t87_n86: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t87_n88;
  t87_n87: return 0.008715602569282055f;
  t87_n88: return 0.04466992989182472f;
  t87_n89: return -0.1243726909160614f;
  t87_n90: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t87_n98;
  t87_n91: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n97;
  t87_n92: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t87_n94;
  t87_n93: return 0.0028318024706095457f;
  t87_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t87_n96;
  t87_n95: return 0.031784575432538986f;
  t87_n96: return 0.02840675227344036f;
  t87_n97: return 0.07360292226076126f;
  t87_n98: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t87_n100;
  t87_n99: return -0.0025053932331502438f;
  t87_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n102;
  t87_n101: return 0.011733313091099262f;
  t87_n102: return 0.07094579935073853f;
  t87_n103: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t87_n121;
  t87_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t87_n106;
  t87_n105: return -0.025837574154138565f;
  t87_n106: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t87_n120;
  t87_n107: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t87_n119;
  t87_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t87_n110;
  t87_n109: return 0.020558902993798256f;
  t87_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t87_n118;
  t87_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t87_n113;
  t87_n112: return 0.07153207808732986f;
  t87_n113: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t87_n115;
  t87_n114: return 0.09182991087436676f;
  t87_n115: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t87_n117;
  t87_n116: return 0.15548424422740936f;
  t87_n117: return -0.12688162922859192f;
  t87_n118: return 0.04852423444390297f;
  t87_n119: return -0.06306911259889603f;
  t87_n120: return -0.014459578320384026f;
  t87_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n123;
  t87_n122: return -0.01528242789208889f;
  t87_n123: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t87_n139;
  t87_n124: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t87_n138;
  t87_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword))) goto t87_n131;
  t87_n126: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n130;
  t87_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t87_n129;
  t87_n128: return 0.002140187891200185f;
  t87_n129: return 0.018864938989281654f;
  t87_n130: return -0.027676889672875404f;
  t87_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t87_n137;
  t87_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n134;
  t87_n133: return 0.012732942588627338f;
  t87_n134: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t87_n136;
  t87_n135: return -0.010858393274247646f;
  t87_n136: return 0.050372157245874405f;
  t87_n137: return -0.0052198320627212524f;
  t87_n138: return -0.07966089248657227f;
  t87_n139: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t87_n141;
  t87_n140: return 0.08558987081050873f;
  t87_n141: return 0.02971397526562214f;
  t87_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t87_n224;
  t87_n143: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t87_n223;
  t87_n144: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t87_n220;
  t87_n145: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t87_n219;
  t87_n146: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t87_n218;
  t87_n147: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t87_n217;
  t87_n148: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t87_n158;
  t87_n149: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t87_n151;
  t87_n150: return 0.004670929629355669f;
  t87_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t87_n157;
  t87_n152: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t87_n154;
  t87_n153: return -0.00282033858820796f;
  t87_n154: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t87_n156;
  t87_n155: return 0.0363561250269413f;
  t87_n156: return 0.00353665160946548f;
  t87_n157: return 0.026831036433577538f;
  t87_n158: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t87_n216;
  t87_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t87_n187;
  t87_n160: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t87_n184;
  t87_n161: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t87_n183;
  t87_n162: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t87_n178;
  t87_n163: if (E.getNumReferences() >= 3350490112 /*92376.0*/) goto t87_n177;
  t87_n164: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t87_n170;
  t87_n165: if (E.getNumReferences() >= 3336376576 /*28296.5*/) goto t87_n169;
  t87_n166: if (E.getNumReferences() >= 3335302144 /*26198.0*/) goto t87_n168;
  t87_n167: return -0.00038240113644860685f;
  t87_n168: return 0.06371857225894928f;
  t87_n169: return -0.06380550563335419f;
  t87_n170: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t87_n174;
  t87_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n173;
  t87_n172: return -0.4746703505516052f;
  t87_n173: return 0.1351272612810135f;
  t87_n174: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t87_n176;
  t87_n175: return -0.005420486442744732f;
  t87_n176: return 0.057325396686792374f;
  t87_n177: return -0.10604406893253326f;
  t87_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t87_n180;
  t87_n179: return 0.06096310541033745f;
  t87_n180: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t87_n182;
  t87_n181: return 0.04750867560505867f;
  t87_n182: return 0.12352985888719559f;
  t87_n183: return -0.151254802942276f;
  t87_n184: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n186;
  t87_n185: return 0.1319596767425537f;
  t87_n186: return 0.003301295917481184f;
  t87_n187: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t87_n215;
  t87_n188: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t87_n202;
  t87_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n193;
  t87_n190: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t87_n192;
  t87_n191: return -0.0919807180762291f;
  t87_n192: return 0.016020767390727997f;
  t87_n193: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t87_n201;
  t87_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n200;
  t87_n195: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t87_n199;
  t87_n196: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t87_n198;
  t87_n197: return 0.032376594841480255f;
  t87_n198: return 0.11540348082780838f;
  t87_n199: return -0.041310109198093414f;
  t87_n200: return 0.010939548723399639f;
  t87_n201: return 0.0004024860681965947f;
  t87_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n206;
  t87_n203: if (E.getNumReferences() >= 3322870784 /*9151.0*/) goto t87_n205;
  t87_n204: return -0.05767333507537842f;
  t87_n205: return -0.3273204267024994f;
  t87_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n212;
  t87_n207: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t87_n209;
  t87_n208: return 0.00756241986528039f;
  t87_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n211;
  t87_n210: return -0.08844252675771713f;
  t87_n211: return -0.10253700613975525f;
  t87_n212: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t87_n214;
  t87_n213: return -0.016471005976200104f;
  t87_n214: return 0.03197932988405228f;
  t87_n215: return 0.02986367978155613f;
  t87_n216: return -0.16849803924560547f;
  t87_n217: return 0.05106929689645767f;
  t87_n218: return -0.07705572992563248f;
  t87_n219: return 0.040645062923431396f;
  t87_n220: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t87_n222;
  t87_n221: return 0.012339410372078419f;
  t87_n222: return -0.195091113448143f;
  t87_n223: return 0.008084023371338844f;
  t87_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t87_n230;
  t87_n225: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t87_n229;
  t87_n226: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t87_n228;
  t87_n227: return -0.0026393213775008917f;
  t87_n228: return 0.025259193032979965f;
  t87_n229: return -0.058048591017723083f;
  t87_n230: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t87_n254;
  t87_n231: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t87_n249;
  t87_n232: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t87_n236;
  t87_n233: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t87_n235;
  t87_n234: return 0.006075229961425066f;
  t87_n235: return 0.03133990615606308f;
  t87_n236: if (E.getNumReferences() >= 3315746816 /*5193.0*/) goto t87_n248;
  t87_n237: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t87_n239;
  t87_n238: return 0.022929854691028595f;
  t87_n239: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t87_n243;
  t87_n240: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t87_n242;
  t87_n241: return -0.0860566720366478f;
  t87_n242: return -1.4662744998931885f;
  t87_n243: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t87_n247;
  t87_n244: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t87_n246;
  t87_n245: return 0.008429781533777714f;
  t87_n246: return 0.13025017082691193f;
  t87_n247: return 0.010288864374160767f;
  t87_n248: return 0.09674538671970367f;
  t87_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t87_n251;
  t87_n250: return -0.018496131524443626f;
  t87_n251: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t87_n253;
  t87_n252: return 0.10158893465995789f;
  t87_n253: return 0.011716393753886223f;
  t87_n254: return 0.018745534121990204f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree88(const Example& E) {
  t88_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t88_n198;
  t88_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t88_n53;
  t88_n2: if (E.getNumReferences() >= 3301306368 /*1583.5*/) goto t88_n48;
  t88_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t88_n29;
  t88_n4: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t88_n28;
  t88_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t88_n21;
  t88_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n18;
  t88_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t88_n17;
  t88_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t88_n12;
  t88_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n11;
  t88_n10: return -0.14114035665988922f;
  t88_n11: return -0.0013560501392930746f;
  t88_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n16;
  t88_n13: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t88_n15;
  t88_n14: return -0.0032646621111780405f;
  t88_n15: return 0.06705498695373535f;
  t88_n16: return 0.015576758421957493f;
  t88_n17: return 0.01673591136932373f;
  t88_n18: if (E.getNumReferences() >= 3281715200 /*310.0*/) goto t88_n20;
  t88_n19: return -0.013654001988470554f;
  t88_n20: return 0.11117951571941376f;
  t88_n21: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t88_n27;
  t88_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t88_n24;
  t88_n23: return -0.021216757595539093f;
  t88_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t88_n26;
  t88_n25: return 0.026607699692249298f;
  t88_n26: return 0.09075856953859329f;
  t88_n27: return 0.04400138929486275f;
  t88_n28: return -0.0008528773905709386f;
  t88_n29: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t88_n43;
  t88_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t88_n40;
  t88_n31: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t88_n39;
  t88_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t88_n38;
  t88_n33: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t88_n37;
  t88_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t88_n36;
  t88_n35: return -0.04805494472384453f;
  t88_n36: return -0.22747796773910522f;
  t88_n37: return -0.017667042091488838f;
  t88_n38: return 0.003686883021146059f;
  t88_n39: return 0.036307938396930695f;
  t88_n40: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t88_n42;
  t88_n41: return 0.012424909509718418f;
  t88_n42: return -0.0521734319627285f;
  t88_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t88_n47;
  t88_n44: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t88_n46;
  t88_n45: return 0.009322870522737503f;
  t88_n46: return 0.028042633086442947f;
  t88_n47: return 0.08842961490154266f;
  t88_n48: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t88_n52;
  t88_n49: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n51;
  t88_n50: return 0.023992590606212616f;
  t88_n51: return -0.00023979567049536854f;
  t88_n52: return 0.041713371872901917f;
  t88_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t88_n139;
  t88_n54: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t88_n70;
  t88_n55: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t88_n67;
  t88_n56: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t88_n66;
  t88_n57: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t88_n59;
  t88_n58: return -0.0058787367306649685f;
  t88_n59: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t88_n61;
  t88_n60: return 0.0065578920766711235f;
  t88_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t88_n65;
  t88_n62: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t88_n64;
  t88_n63: return 0.036934930831193924f;
  t88_n64: return -0.10258577764034271f;
  t88_n65: return 0.0841694101691246f;
  t88_n66: return 0.013914979994297028f;
  t88_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t88_n69;
  t88_n68: return -0.11168347299098969f;
  t88_n69: return -0.043751224875450134f;
  t88_n70: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t88_n78;
  t88_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n73;
  t88_n72: return -0.07282332330942154f;
  t88_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t88_n75;
  t88_n74: return 0.018525592982769012f;
  t88_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t88_n77;
  t88_n76: return 0.04483906924724579f;
  t88_n77: return 0.07200555503368378f;
  t88_n78: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t88_n82;
  t88_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t88_n81;
  t88_n80: return -0.016763882711529732f;
  t88_n81: return 0.009263107553124428f;
  t88_n82: if (E.getNumReferences() >= 3338505216 /*32454.0*/) goto t88_n120;
  t88_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t88_n89;
  t88_n84: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t88_n86;
  t88_n85: return -0.0033380864188075066f;
  t88_n86: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t88_n88;
  t88_n87: return 0.010174326598644257f;
  t88_n88: return -0.12842456996440887f;
  t88_n89: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t88_n111;
  t88_n90: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t88_n110;
  t88_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t88_n93;
  t88_n92: return 0.01146095059812069f;
  t88_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n95;
  t88_n94: return -0.11171592026948929f;
  t88_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t88_n105;
  t88_n96: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t88_n102;
  t88_n97: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t88_n101;
  t88_n98: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t88_n100;
  t88_n99: return 0.07116273790597916f;
  t88_n100: return 0.010029460303485394f;
  t88_n101: return 0.07064397633075714f;
  t88_n102: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t88_n104;
  t88_n103: return -0.028399348258972168f;
  t88_n104: return 0.05244474112987518f;
  t88_n105: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t88_n107;
  t88_n106: return -0.0003974813153035939f;
  t88_n107: if (E.getNumReferences() >= 3337714176 /*30909.0*/) goto t88_n109;
  t88_n108: return 0.11736501008272171f;
  t88_n109: return -0.21650001406669617f;
  t88_n110: return -0.001035704743117094f;
  t88_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n115;
  t88_n112: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t88_n114;
  t88_n113: return -0.03345801681280136f;
  t88_n114: return 0.01903625950217247f;
  t88_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t88_n117;
  t88_n116: return -0.10769357532262802f;
  t88_n117: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t88_n119;
  t88_n118: return 0.020271185785531998f;
  t88_n119: return -0.12313579022884369f;
  t88_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n136;
  t88_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t88_n123;
  t88_n122: return -0.10139559954404831f;
  t88_n123: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t88_n133;
  t88_n124: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t88_n126;
  t88_n125: return -0.03716392442584038f;
  t88_n126: if (E.getNumReferences() >= 3345303552 /*58696.0*/) goto t88_n130;
  t88_n127: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n129;
  t88_n128: return 0.10130998492240906f;
  t88_n129: return 0.147953599691391f;
  t88_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n132;
  t88_n131: return -0.18103419244289398f;
  t88_n132: return 0.07112796604633331f;
  t88_n133: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t88_n135;
  t88_n134: return -0.12166988104581833f;
  t88_n135: return 0.04670814797282219f;
  t88_n136: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t88_n138;
  t88_n137: return 0.05575472488999367f;
  t88_n138: return 0.01400893647223711f;
  t88_n139: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t88_n197;
  t88_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t88_n160;
  t88_n141: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n155;
  t88_n142: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t88_n150;
  t88_n143: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t88_n149;
  t88_n144: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t88_n146;
  t88_n145: return 0.0023682289756834507f;
  t88_n146: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t88_n148;
  t88_n147: return -0.07828529924154282f;
  t88_n148: return 0.127557173371315f;
  t88_n149: return -0.04014982655644417f;
  t88_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t88_n152;
  t88_n151: return -0.12768635153770447f;
  t88_n152: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t88_n154;
  t88_n153: return 0.009785322472453117f;
  t88_n154: return -0.05290937051177025f;
  t88_n155: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t88_n159;
  t88_n156: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t88_n158;
  t88_n157: return 0.020308323204517365f;
  t88_n158: return 0.038712598383426666f;
  t88_n159: return 0.0043721869587898254f;
  t88_n160: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t88_n188;
  t88_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t88_n177;
  t88_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n164;
  t88_n163: return -0.016020361334085464f;
  t88_n164: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t88_n172;
  t88_n165: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t88_n167;
  t88_n166: return -0.006868850439786911f;
  t88_n167: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t88_n169;
  t88_n168: return 0.046711113303899765f;
  t88_n169: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t88_n171;
  t88_n170: return -0.06489653885364532f;
  t88_n171: return 0.03708909824490547f;
  t88_n172: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t88_n174;
  t88_n173: return -0.08188101649284363f;
  t88_n174: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t88_n176;
  t88_n175: return -0.010356818325817585f;
  t88_n176: return 0.09581215679645538f;
  t88_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n187;
  t88_n178: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t88_n182;
  t88_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t88_n181;
  t88_n180: return 0.019467052072286606f;
  t88_n181: return -0.0594579242169857f;
  t88_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n186;
  t88_n183: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t88_n185;
  t88_n184: return -0.12093167006969452f;
  t88_n185: return -0.25126421451568604f;
  t88_n186: return -0.011516558937728405f;
  t88_n187: return 0.039115916937589645f;
  t88_n188: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t88_n190;
  t88_n189: return 0.008945289067924023f;
  t88_n190: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t88_n192;
  t88_n191: return 0.043196987360715866f;
  t88_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t88_n196;
  t88_n193: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t88_n195;
  t88_n194: return -0.012824454344809055f;
  t88_n195: return -0.060668639838695526f;
  t88_n196: return 0.07170777767896652f;
  t88_n197: return -0.02426709607243538f;
  t88_n198: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t88_n254;
  t88_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t88_n203;
  t88_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t88_n202;
  t88_n201: return -0.001802724553272128f;
  t88_n202: return 0.015073047950863838f;
  t88_n203: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t88_n253;
  t88_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t88_n224;
  t88_n205: if (E.getNumReferences() >= 3286646784 /*460.5*/) goto t88_n223;
  t88_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t88_n212;
  t88_n207: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t88_n211;
  t88_n208: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t88_n210;
  t88_n209: return -0.00856267660856247f;
  t88_n210: return -0.1411672681570053f;
  t88_n211: return 0.028963908553123474f;
  t88_n212: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t88_n216;
  t88_n213: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t88_n215;
  t88_n214: return 0.002414318732917309f;
  t88_n215: return 0.027781251817941666f;
  t88_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t88_n218;
  t88_n217: return 0.04757899045944214f;
  t88_n218: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t88_n220;
  t88_n219: return 0.1507842242717743f;
  t88_n220: if (E.getNumReferences() >= 3260809216 /*55.0*/) goto t88_n222;
  t88_n221: return -0.04883907735347748f;
  t88_n222: return 0.1016862541437149f;
  t88_n223: return 0.0434211827814579f;
  t88_n224: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t88_n246;
  t88_n225: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t88_n237;
  t88_n226: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t88_n236;
  t88_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t88_n233;
  t88_n228: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n232;
  t88_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t88_n231;
  t88_n230: return -0.015017258003354073f;
  t88_n231: return 0.013953439891338348f;
  t88_n232: return -0.013469048775732517f;
  t88_n233: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t88_n235;
  t88_n234: return 0.010053029283881187f;
  t88_n235: return 0.042387705296278f;
  t88_n236: return -0.013556457124650478f;
  t88_n237: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t88_n239;
  t88_n238: return 0.014809302054345608f;
  t88_n239: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t88_n241;
  t88_n240: return -0.013683252036571503f;
  t88_n241: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t88_n245;
  t88_n242: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t88_n244;
  t88_n243: return 0.09283415228128433f;
  t88_n244: return 0.12503387033939362f;
  t88_n245: return 0.046388670802116394f;
  t88_n246: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t88_n250;
  t88_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n249;
  t88_n248: return -0.01416861917823553f;
  t88_n249: return 0.014580233953893185f;
  t88_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t88_n252;
  t88_n251: return -0.046720318496227264f;
  t88_n252: return 0.02258707769215107f;
  t88_n253: return -0.0035362013150006533f;
  t88_n254: return 0.014755213633179665f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree89(const Example& E) {
  t89_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t89_n252;
  t89_n1: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t89_n227;
  t89_n2: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t89_n226;
  t89_n3: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t89_n225;
  t89_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t89_n222;
  t89_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t89_n31;
  t89_n6: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t89_n28;
  t89_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t89_n23;
  t89_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t89_n22;
  t89_n9: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t89_n19;
  t89_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t89_n12;
  t89_n11: return 0.003668989287689328f;
  t89_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t89_n18;
  t89_n13: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t89_n17;
  t89_n14: if (E.getSymbolScopeDistanceCost() >= 3246391296 /*16.0*/) goto t89_n16;
  t89_n15: return 0.021785784512758255f;
  t89_n16: return 0.06372461467981339f;
  t89_n17: return 0.0005754795856773853f;
  t89_n18: return 0.03775356709957123f;
  t89_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t89_n21;
  t89_n20: return -0.014361321926116943f;
  t89_n21: return 0.03083324432373047f;
  t89_n22: return 0.022897345945239067f;
  t89_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n27;
  t89_n24: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n26;
  t89_n25: return -0.15625980496406555f;
  t89_n26: return 0.058722373098134995f;
  t89_n27: return 0.03942409157752991f;
  t89_n28: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t89_n30;
  t89_n29: return -0.056422699242830276f;
  t89_n30: return 0.004318200517445803f;
  t89_n31: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t89_n199;
  t89_n32: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t89_n164;
  t89_n33: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t89_n163;
  t89_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t89_n40;
  t89_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n39;
  t89_n36: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t89_n38;
  t89_n37: return -0.13374567031860352f;
  t89_n38: return -0.02602129429578781f;
  t89_n39: return -0.0036090330686420202f;
  t89_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t89_n42;
  t89_n41: return -0.0009412377257831395f;
  t89_n42: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t89_n146;
  t89_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t89_n85;
  t89_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n52;
  t89_n45: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n49;
  t89_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n48;
  t89_n47: return -0.04407394677400589f;
  t89_n48: return 0.05081285163760185f;
  t89_n49: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t89_n51;
  t89_n50: return -0.042301423847675323f;
  t89_n51: return -0.041711218655109406f;
  t89_n52: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t89_n80;
  t89_n53: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t89_n57;
  t89_n54: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n56;
  t89_n55: return 0.032251402735710144f;
  t89_n56: return -0.04284908249974251f;
  t89_n57: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t89_n77;
  t89_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t89_n64;
  t89_n59: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t89_n61;
  t89_n60: return 0.04147406667470932f;
  t89_n61: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t89_n63;
  t89_n62: return 0.08917877078056335f;
  t89_n63: return -0.005512541625648737f;
  t89_n64: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t89_n66;
  t89_n65: return 0.02475220523774624f;
  t89_n66: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t89_n74;
  t89_n67: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t89_n71;
  t89_n68: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t89_n70;
  t89_n69: return 0.0675654485821724f;
  t89_n70: return -0.001522206817753613f;
  t89_n71: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t89_n73;
  t89_n72: return 0.0645519271492958f;
  t89_n73: return 0.16081278026103973f;
  t89_n74: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t89_n76;
  t89_n75: return -0.015479492954909801f;
  t89_n76: return 0.050395939499139786f;
  t89_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n79;
  t89_n78: return -0.011270103044807911f;
  t89_n79: return 0.038995109498500824f;
  t89_n80: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t89_n82;
  t89_n81: return -0.05803660303354263f;
  t89_n82: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t89_n84;
  t89_n83: return -0.00021558637672569603f;
  t89_n84: return 0.050407033413648605f;
  t89_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n91;
  t89_n86: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t89_n90;
  t89_n87: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t89_n89;
  t89_n88: return -0.05459388718008995f;
  t89_n89: return 0.006285178009420633f;
  t89_n90: return -0.04696742817759514f;
  t89_n91: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t89_n145;
  t89_n92: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t89_n94;
  t89_n93: return 0.019883031025528908f;
  t89_n94: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n114;
  t89_n95: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t89_n113;
  t89_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n98;
  t89_n97: return -0.006994909141212702f;
  t89_n98: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t89_n108;
  t89_n99: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t89_n107;
  t89_n100: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t89_n106;
  t89_n101: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t89_n103;
  t89_n102: return 0.003795197233557701f;
  t89_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n105;
  t89_n104: return -0.009598239324986935f;
  t89_n105: return 0.04719113931059837f;
  t89_n106: return -0.03897154703736305f;
  t89_n107: return 0.035857249051332474f;
  t89_n108: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t89_n112;
  t89_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n111;
  t89_n110: return -0.09032844752073288f;
  t89_n111: return 0.03026205115020275f;
  t89_n112: return 0.0007863484206609428f;
  t89_n113: return -0.03180081397294998f;
  t89_n114: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t89_n116;
  t89_n115: return 0.0032130414620041847f;
  t89_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t89_n126;
  t89_n117: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t89_n125;
  t89_n118: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t89_n124;
  t89_n119: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t89_n123;
  t89_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t89_n122;
  t89_n121: return 0.00031033437699079514f;
  t89_n122: return 0.007040611933916807f;
  t89_n123: return 0.03204520791769028f;
  t89_n124: return -0.008153073489665985f;
  t89_n125: return 0.021981950849294662f;
  t89_n126: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t89_n142;
  t89_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t89_n129;
  t89_n128: return 0.021871745586395264f;
  t89_n129: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t89_n137;
  t89_n130: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t89_n136;
  t89_n131: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t89_n135;
  t89_n132: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t89_n134;
  t89_n133: return 0.03939174860715866f;
  t89_n134: return -0.08149538934230804f;
  t89_n135: return 0.042711447924375534f;
  t89_n136: return -0.1389949470758438f;
  t89_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t89_n139;
  t89_n138: return 0.023961568251252174f;
  t89_n139: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t89_n141;
  t89_n140: return 0.06008746474981308f;
  t89_n141: return 0.16663356125354767f;
  t89_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n144;
  t89_n143: return -0.11656070500612259f;
  t89_n144: return -0.004143510013818741f;
  t89_n145: return 0.0004199556424282491f;
  t89_n146: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t89_n152;
  t89_n147: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t89_n149;
  t89_n148: return 0.0070940051227808f;
  t89_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n151;
  t89_n150: return -0.18910528719425201f;
  t89_n151: return 0.026399245485663414f;
  t89_n152: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n162;
  t89_n153: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t89_n161;
  t89_n154: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t89_n160;
  t89_n155: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t89_n159;
  t89_n156: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t89_n158;
  t89_n157: return 0.0800795778632164f;
  t89_n158: return 0.12097886204719543f;
  t89_n159: return 0.04005362465977669f;
  t89_n160: return -0.007998653687536716f;
  t89_n161: return 0.11261116713285446f;
  t89_n162: return 0.02473258413374424f;
  t89_n163: return -0.08560168743133545f;
  t89_n164: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t89_n168;
  t89_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n167;
  t89_n166: return -0.07553716748952866f;
  t89_n167: return -0.05814704671502113f;
  t89_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t89_n188;
  t89_n169: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t89_n173;
  t89_n170: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t89_n172;
  t89_n171: return 0.00028346601175144315f;
  t89_n172: return 0.03336242213845253f;
  t89_n173: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t89_n187;
  t89_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t89_n182;
  t89_n175: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t89_n177;
  t89_n176: return 0.013786264695227146f;
  t89_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t89_n179;
  t89_n178: return 0.15989628434181213f;
  t89_n179: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t89_n181;
  t89_n180: return 0.1481451541185379f;
  t89_n181: return 0.2029617577791214f;
  t89_n182: if (E.getNumReferences() >= 3277684736 /*221.5*/) goto t89_n184;
  t89_n183: return 0.09968364238739014f;
  t89_n184: if (E.getNumReferences() >= 3278143488 /*228.5*/) goto t89_n186;
  t89_n185: return 0.1975339651107788f;
  t89_n186: return 0.11003074049949646f;
  t89_n187: return -0.05172128230333328f;
  t89_n188: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t89_n194;
  t89_n189: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t89_n191;
  t89_n190: return 0.010549299418926239f;
  t89_n191: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t89_n193;
  t89_n192: return 0.03133122995495796f;
  t89_n193: return 0.07357668876647949f;
  t89_n194: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t89_n198;
  t89_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t89_n197;
  t89_n196: return -0.019985215738415718f;
  t89_n197: return 0.015079349279403687f;
  t89_n198: return -0.04804385080933571f;
  t89_n199: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t89_n213;
  t89_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t89_n208;
  t89_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t89_n203;
  t89_n202: return -0.050933316349983215f;
  t89_n203: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t89_n207;
  t89_n204: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t89_n206;
  t89_n205: return -0.01564321294426918f;
  t89_n206: return 0.09130256623029709f;
  t89_n207: return 0.019572794437408447f;
  t89_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t89_n212;
  t89_n209: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t89_n211;
  t89_n210: return 0.132169708609581f;
  t89_n211: return 0.009236705489456654f;
  t89_n212: return 0.06976436823606491f;
  t89_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t89_n217;
  t89_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n216;
  t89_n215: return -0.05683663487434387f;
  t89_n216: return 0.011884156614542007f;
  t89_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t89_n219;
  t89_n218: return 0.01966335065662861f;
  t89_n219: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t89_n221;
  t89_n220: return 0.00847698375582695f;
  t89_n221: return 0.030304402112960815f;
  t89_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t89_n224;
  t89_n223: return -0.1881731152534485f;
  t89_n224: return -0.028970953077077866f;
  t89_n225: return -0.043145690113306046f;
  t89_n226: return 0.010787543840706348f;
  t89_n227: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t89_n251;
  t89_n228: if (E.getNumReferences() >= 3341306880 /*43084.0*/) goto t89_n232;
  t89_n229: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t89_n231;
  t89_n230: return 0.013593879528343678f;
  t89_n231: return -0.21170715987682343f;
  t89_n232: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t89_n242;
  t89_n233: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t89_n237;
  t89_n234: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t89_n236;
  t89_n235: return 0.03540550917387009f;
  t89_n236: return 0.10567139089107513f;
  t89_n237: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t89_n241;
  t89_n238: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t89_n240;
  t89_n239: return 0.11770883202552795f;
  t89_n240: return 0.16293148696422577f;
  t89_n241: return 0.06766586750745773f;
  t89_n242: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t89_n244;
  t89_n243: return -0.06973444670438766f;
  t89_n244: if (E.getNumReferences() >= 3344543232 /*55726.0*/) goto t89_n248;
  t89_n245: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t89_n247;
  t89_n246: return 0.10436280816793442f;
  t89_n247: return -0.16535289585590363f;
  t89_n248: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t89_n250;
  t89_n249: return -0.02296728640794754f;
  t89_n250: return 0.0303787924349308f;
  t89_n251: return 0.002024532062932849f;
  t89_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t89_n254;
  t89_n253: return 0.005178213119506836f;
  t89_n254: return 0.01425581518560648f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree90(const Example& E) {
  t90_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t90_n68;
  t90_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t90_n49;
  t90_n2: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t90_n42;
  t90_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t90_n17;
  t90_n4: if (E.getNumReferences() >= 3274932224 /*179.5*/) goto t90_n10;
  t90_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n9;
  t90_n6: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t90_n8;
  t90_n7: return -0.00702136242762208f;
  t90_n8: return 0.004472318105399609f;
  t90_n9: return 0.0010444365907460451f;
  t90_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t90_n16;
  t90_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n15;
  t90_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n14;
  t90_n13: return -0.057674191892147064f;
  t90_n14: return 0.022944170981645584f;
  t90_n15: return 0.01836114190518856f;
  t90_n16: return -0.005380936432629824f;
  t90_n17: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t90_n33;
  t90_n18: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t90_n32;
  t90_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t90_n23;
  t90_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t90_n22;
  t90_n21: return 0.0001257062394870445f;
  t90_n22: return 0.02121172472834587f;
  t90_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n25;
  t90_n24: return -0.023642495274543762f;
  t90_n25: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t90_n27;
  t90_n26: return 0.022602157667279243f;
  t90_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n29;
  t90_n28: return -0.016550812870264053f;
  t90_n29: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t90_n31;
  t90_n30: return 0.08806394040584564f;
  t90_n31: return 0.03967178985476494f;
  t90_n32: return 0.05423593893647194f;
  t90_n33: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t90_n35;
  t90_n34: return -0.007514883764088154f;
  t90_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n37;
  t90_n36: return -0.04711653292179108f;
  t90_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t90_n39;
  t90_n38: return 0.011488661170005798f;
  t90_n39: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t90_n41;
  t90_n40: return -0.03820032626390457f;
  t90_n41: return 0.14032095670700073f;
  t90_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n46;
  t90_n43: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t90_n45;
  t90_n44: return 0.06910491734743118f;
  t90_n45: return -0.28803879022598267f;
  t90_n46: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t90_n48;
  t90_n47: return 0.042474448680877686f;
  t90_n48: return -0.028826015070080757f;
  t90_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t90_n55;
  t90_n50: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t90_n54;
  t90_n51: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t90_n53;
  t90_n52: return 0.009832062758505344f;
  t90_n53: return -0.004437899217009544f;
  t90_n54: return 0.02464934252202511f;
  t90_n55: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t90_n67;
  t90_n56: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t90_n64;
  t90_n57: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t90_n59;
  t90_n58: return 0.06161326915025711f;
  t90_n59: if (E.getNumReferences() >= 3357605312 /*164855.0*/) goto t90_n61;
  t90_n60: return 0.01333630084991455f;
  t90_n61: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t90_n63;
  t90_n62: return 0.10080588608980179f;
  t90_n63: return 0.09782674908638f;
  t90_n64: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t90_n66;
  t90_n65: return 0.13386985659599304f;
  t90_n66: return -0.0585850290954113f;
  t90_n67: return -0.10174842923879623f;
  t90_n68: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t90_n218;
  t90_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t90_n71;
  t90_n70: return -0.10864800214767456f;
  t90_n71: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t90_n215;
  t90_n72: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t90_n166;
  t90_n73: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t90_n163;
  t90_n74: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t90_n152;
  t90_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t90_n133;
  t90_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t90_n92;
  t90_n77: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t90_n83;
  t90_n78: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t90_n80;
  t90_n79: return -0.0027201937045902014f;
  t90_n80: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t90_n82;
  t90_n81: return -0.018152344971895218f;
  t90_n82: return -0.06897823512554169f;
  t90_n83: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t90_n91;
  t90_n84: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t90_n86;
  t90_n85: return -0.1289609968662262f;
  t90_n86: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t90_n88;
  t90_n87: return 0.023970618844032288f;
  t90_n88: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t90_n90;
  t90_n89: return 0.01568400114774704f;
  t90_n90: return -0.027035420760512352f;
  t90_n91: return 0.13360927999019623f;
  t90_n92: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t90_n122;
  t90_n93: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t90_n111;
  t90_n94: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t90_n110;
  t90_n95: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t90_n109;
  t90_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t90_n98;
  t90_n97: return -0.013141111470758915f;
  t90_n98: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t90_n104;
  t90_n99: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t90_n103;
  t90_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t90_n102;
  t90_n101: return 0.010492943227291107f;
  t90_n102: return 0.02413085289299488f;
  t90_n103: return -0.08944530785083771f;
  t90_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t90_n108;
  t90_n105: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t90_n107;
  t90_n106: return 0.12065327167510986f;
  t90_n107: return 0.03593728318810463f;
  t90_n108: return 0.2045746147632599f;
  t90_n109: return -0.0014283645432442427f;
  t90_n110: return 0.005932191852480173f;
  t90_n111: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t90_n117;
  t90_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n116;
  t90_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n115;
  t90_n114: return -0.10086487978696823f;
  t90_n115: return -0.10443249344825745f;
  t90_n116: return 0.002169782994315028f;
  t90_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t90_n119;
  t90_n118: return 0.004786317236721516f;
  t90_n119: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t90_n121;
  t90_n120: return 0.11551351100206375f;
  t90_n121: return 0.011989310383796692f;
  t90_n122: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t90_n126;
  t90_n123: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t90_n125;
  t90_n124: return -0.023226484656333923f;
  t90_n125: return 0.03195019066333771f;
  t90_n126: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t90_n130;
  t90_n127: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t90_n129;
  t90_n128: return 0.12191399931907654f;
  t90_n129: return -0.09507991373538971f;
  t90_n130: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t90_n132;
  t90_n131: return 0.007941552437841892f;
  t90_n132: return -0.01891305111348629f;
  t90_n133: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t90_n149;
  t90_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t90_n136;
  t90_n135: return -0.014620053581893444f;
  t90_n136: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t90_n146;
  t90_n137: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t90_n139;
  t90_n138: return 0.01865449734032154f;
  t90_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n141;
  t90_n140: return -0.11274001002311707f;
  t90_n141: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t90_n145;
  t90_n142: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t90_n144;
  t90_n143: return 0.09457843005657196f;
  t90_n144: return 0.02824416384100914f;
  t90_n145: return 0.22258400917053223f;
  t90_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t90_n148;
  t90_n147: return 0.01921379752457142f;
  t90_n148: return -0.04272192344069481f;
  t90_n149: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t90_n151;
  t90_n150: return 0.06568203866481781f;
  t90_n151: return 0.030600951984524727f;
  t90_n152: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t90_n162;
  t90_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t90_n161;
  t90_n154: if (E.getNumReferences() >= 3314693120 /*4678.5*/) goto t90_n156;
  t90_n155: return 0.007882173173129559f;
  t90_n156: if (E.getNumReferences() >= 3315159040 /*4906.0*/) goto t90_n160;
  t90_n157: if (E.getNumReferences() >= 3315124224 /*4889.0*/) goto t90_n159;
  t90_n158: return 0.09916096180677414f;
  t90_n159: return 0.13652774691581726f;
  t90_n160: return 0.029549742117524147f;
  t90_n161: return 0.06825108081102371f;
  t90_n162: return 0.0061157275922596455f;
  t90_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n165;
  t90_n164: return -0.12317875772714615f;
  t90_n165: return 0.011358998715877533f;
  t90_n166: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t90_n214;
  t90_n167: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t90_n185;
  t90_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n174;
  t90_n169: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t90_n173;
  t90_n170: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t90_n172;
  t90_n171: return 0.18848766386508942f;
  t90_n172: return -0.04822220280766487f;
  t90_n173: return 0.01729612611234188f;
  t90_n174: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t90_n184;
  t90_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n177;
  t90_n176: return -0.03700031712651253f;
  t90_n177: if (E.getFractionNameInContext() >= 3210856220 /*0.8819444179534912*/) goto t90_n183;
  t90_n178: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t90_n180;
  t90_n179: return 0.016722584143280983f;
  t90_n180: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t90_n182;
  t90_n181: return 0.09202414751052856f;
  t90_n182: return 0.119891457259655f;
  t90_n183: return -0.04786965996026993f;
  t90_n184: return -0.001207019668072462f;
  t90_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n187;
  t90_n186: return -0.05139691010117531f;
  t90_n187: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t90_n199;
  t90_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n198;
  t90_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n197;
  t90_n190: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t90_n192;
  t90_n191: return -0.002369816182181239f;
  t90_n192: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t90_n196;
  t90_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t90_n195;
  t90_n194: return -0.08402875065803528f;
  t90_n195: return 0.03339826688170433f;
  t90_n196: return -0.003004560014232993f;
  t90_n197: return 0.001777934841811657f;
  t90_n198: return 0.009201224893331528f;
  t90_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t90_n203;
  t90_n200: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t90_n202;
  t90_n201: return -0.07075788825750351f;
  t90_n202: return 0.0056999605149030685f;
  t90_n203: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t90_n211;
  t90_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t90_n210;
  t90_n205: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t90_n207;
  t90_n206: return -0.04767272621393204f;
  t90_n207: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t90_n209;
  t90_n208: return 0.006884966976940632f;
  t90_n209: return -0.18807078897953033f;
  t90_n210: return 0.030015533789992332f;
  t90_n211: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t90_n213;
  t90_n212: return 0.03430671989917755f;
  t90_n213: return 0.13109450042247772f;
  t90_n214: return 0.031386636197566986f;
  t90_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n217;
  t90_n216: return -0.12226783484220505f;
  t90_n217: return -0.07557827234268188f;
  t90_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t90_n224;
  t90_n219: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t90_n221;
  t90_n220: return -0.00011926521256100386f;
  t90_n221: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t90_n223;
  t90_n222: return -0.033236924558877945f;
  t90_n223: return 0.07417259365320206f;
  t90_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t90_n226;
  t90_n225: return -0.1519993245601654f;
  t90_n226: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t90_n254;
  t90_n227: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t90_n247;
  t90_n228: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t90_n230;
  t90_n229: return 0.015035419724881649f;
  t90_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t90_n232;
  t90_n231: return -0.009769129566848278f;
  t90_n232: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t90_n240;
  t90_n233: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t90_n235;
  t90_n234: return -0.007314735557883978f;
  t90_n235: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t90_n239;
  t90_n236: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t90_n238;
  t90_n237: return 0.05592447519302368f;
  t90_n238: return 0.0013066353276371956f;
  t90_n239: return -0.003286333754658699f;
  t90_n240: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t90_n246;
  t90_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t90_n245;
  t90_n242: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t90_n244;
  t90_n243: return 0.04418351128697395f;
  t90_n244: return 0.08723479509353638f;
  t90_n245: return 0.020811716094613075f;
  t90_n246: return -0.030946029350161552f;
  t90_n247: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t90_n253;
  t90_n248: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t90_n252;
  t90_n249: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t90_n251;
  t90_n250: return 0.08398985862731934f;
  t90_n251: return 0.14106066524982452f;
  t90_n252: return 0.033242400735616684f;
  t90_n253: return 0.015788594260811806f;
  t90_n254: return 0.01732724718749523f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree91(const Example& E) {
  t91_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t91_n50;
  t91_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t91_n41;
  t91_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t91_n36;
  t91_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t91_n21;
  t91_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t91_n18;
  t91_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n9;
  t91_n6: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t91_n8;
  t91_n7: return -0.006066375412046909f;
  t91_n8: return 0.012732000090181828f;
  t91_n9: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t91_n17;
  t91_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t91_n12;
  t91_n11: return -0.10284116864204407f;
  t91_n12: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t91_n14;
  t91_n13: return 0.005430974066257477f;
  t91_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t91_n16;
  t91_n15: return 0.013931278139352798f;
  t91_n16: return 0.0734473392367363f;
  t91_n17: return 0.03008180856704712f;
  t91_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n20;
  t91_n19: return -0.0495726615190506f;
  t91_n20: return -0.030863862484693527f;
  t91_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t91_n23;
  t91_n22: return 0.0001552869944134727f;
  t91_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t91_n25;
  t91_n24: return 0.006649862974882126f;
  t91_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t91_n31;
  t91_n26: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t91_n28;
  t91_n27: return 0.02343027852475643f;
  t91_n28: if (E.getNumReferences() >= 3364539984 /*284274.5*/) goto t91_n30;
  t91_n29: return 0.09470801055431366f;
  t91_n30: return 0.031211649999022484f;
  t91_n31: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t91_n33;
  t91_n32: return 0.022641262039542198f;
  t91_n33: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t91_n35;
  t91_n34: return 0.02587977424263954f;
  t91_n35: return 0.08668377995491028f;
  t91_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t91_n38;
  t91_n37: return -0.10873707383871078f;
  t91_n38: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t91_n40;
  t91_n39: return -0.0013579309452325106f;
  t91_n40: return -0.013285081833600998f;
  t91_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t91_n47;
  t91_n42: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t91_n46;
  t91_n43: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t91_n45;
  t91_n44: return 0.009626702405512333f;
  t91_n45: return -0.004868409130722284f;
  t91_n46: return 0.02318853698670864f;
  t91_n47: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t91_n49;
  t91_n48: return 0.013786458410322666f;
  t91_n49: return -0.0913430005311966f;
  t91_n50: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t91_n210;
  t91_n51: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t91_n119;
  t91_n52: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t91_n116;
  t91_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t91_n113;
  t91_n54: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t91_n82;
  t91_n55: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t91_n81;
  t91_n56: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t91_n76;
  t91_n57: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t91_n59;
  t91_n58: return -0.004985510837286711f;
  t91_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t91_n75;
  t91_n60: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t91_n62;
  t91_n61: return -0.02329990267753601f;
  t91_n62: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t91_n72;
  t91_n63: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t91_n65;
  t91_n64: return -0.07679757475852966f;
  t91_n65: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t91_n69;
  t91_n66: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t91_n68;
  t91_n67: return 0.016893090680241585f;
  t91_n68: return -0.23244701325893402f;
  t91_n69: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t91_n71;
  t91_n70: return 0.13087333738803864f;
  t91_n71: return -0.33562615513801575f;
  t91_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t91_n74;
  t91_n73: return 0.1032668948173523f;
  t91_n74: return -0.044094666838645935f;
  t91_n75: return -0.009605399332940578f;
  t91_n76: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t91_n80;
  t91_n77: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t91_n79;
  t91_n78: return -0.007793457247316837f;
  t91_n79: return 0.08394336700439453f;
  t91_n80: return 0.0010751566151157022f;
  t91_n81: return -0.018727155402302742f;
  t91_n82: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t91_n110;
  t91_n83: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t91_n85;
  t91_n84: return 0.026148920878767967f;
  t91_n85: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t91_n101;
  t91_n86: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t91_n90;
  t91_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n89;
  t91_n88: return -0.08051349967718124f;
  t91_n89: return 0.013931548222899437f;
  t91_n90: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t91_n92;
  t91_n91: return 0.04159756749868393f;
  t91_n92: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t91_n94;
  t91_n93: return 0.09863395988941193f;
  t91_n94: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t91_n96;
  t91_n95: return 0.03778330981731415f;
  t91_n96: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t91_n100;
  t91_n97: if (E.getNumReferences() >= 3244294144 /*14.0*/) goto t91_n99;
  t91_n98: return 0.10144384205341339f;
  t91_n99: return 0.1410428136587143f;
  t91_n100: return 0.0688990131020546f;
  t91_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t91_n109;
  t91_n102: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t91_n106;
  t91_n103: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t91_n105;
  t91_n104: return -0.06434502452611923f;
  t91_n105: return -0.46241071820259094f;
  t91_n106: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t91_n108;
  t91_n107: return -0.0007955495966598392f;
  t91_n108: return -0.0473887175321579f;
  t91_n109: return 0.07073530554771423f;
  t91_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t91_n112;
  t91_n111: return 0.034296903759241104f;
  t91_n112: return 0.14105182886123657f;
  t91_n113: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t91_n115;
  t91_n114: return 0.007483551278710365f;
  t91_n115: return -0.006884882692247629f;
  t91_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t91_n118;
  t91_n117: return -0.013862216845154762f;
  t91_n118: return 0.03115040622651577f;
  t91_n119: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t91_n209;
  t91_n120: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t91_n208;
  t91_n121: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t91_n199;
  t91_n122: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t91_n134;
  t91_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n131;
  t91_n124: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t91_n126;
  t91_n125: return 0.010067205876111984f;
  t91_n126: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t91_n130;
  t91_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t91_n129;
  t91_n128: return -0.09588013589382172f;
  t91_n129: return -0.06925887614488602f;
  t91_n130: return -0.029379917308688164f;
  t91_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t91_n133;
  t91_n132: return -0.024829497560858727f;
  t91_n133: return 0.010200856253504753f;
  t91_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t91_n162;
  t91_n135: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t91_n137;
  t91_n136: return -0.003646690398454666f;
  t91_n137: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t91_n161;
  t91_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t91_n156;
  t91_n139: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t91_n145;
  t91_n140: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t91_n144;
  t91_n141: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t91_n143;
  t91_n142: return 0.009450618177652359f;
  t91_n143: return 0.030666198581457138f;
  t91_n144: return -0.05366210639476776f;
  t91_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t91_n149;
  t91_n146: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t91_n148;
  t91_n147: return 0.030482254922389984f;
  t91_n148: return -0.10962497442960739f;
  t91_n149: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t91_n151;
  t91_n150: return 0.072381392121315f;
  t91_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n153;
  t91_n152: return -0.1869858354330063f;
  t91_n153: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t91_n155;
  t91_n154: return 0.09830009192228317f;
  t91_n155: return 0.04452107846736908f;
  t91_n156: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t91_n160;
  t91_n157: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t91_n159;
  t91_n158: return 0.12345577031373978f;
  t91_n159: return 0.04871748015284538f;
  t91_n160: return -0.0005154531681910157f;
  t91_n161: return -0.008729182183742523f;
  t91_n162: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t91_n190;
  t91_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t91_n167;
  t91_n164: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t91_n166;
  t91_n165: return 0.003169064410030842f;
  t91_n166: return -0.03371331840753555f;
  t91_n167: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t91_n189;
  t91_n168: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t91_n178;
  t91_n169: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t91_n171;
  t91_n170: return 0.052743829786777496f;
  t91_n171: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t91_n173;
  t91_n172: return 0.07235773652791977f;
  t91_n173: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t91_n175;
  t91_n174: return 0.06554093956947327f;
  t91_n175: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t91_n177;
  t91_n176: return 0.14047661423683167f;
  t91_n177: return 0.05824709311127663f;
  t91_n178: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t91_n188;
  t91_n179: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t91_n181;
  t91_n180: return 0.030990561470389366f;
  t91_n181: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t91_n183;
  t91_n182: return 0.05525876209139824f;
  t91_n183: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t91_n185;
  t91_n184: return 0.052142299711704254f;
  t91_n185: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t91_n187;
  t91_n186: return 0.22009101510047913f;
  t91_n187: return 0.11111140996217728f;
  t91_n188: return -0.09074851125478745f;
  t91_n189: return -0.031400881707668304f;
  t91_n190: if (E.getNumReferences() >= 3361098688 /*219439.0*/) goto t91_n198;
  t91_n191: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t91_n193;
  t91_n192: return 0.042852967977523804f;
  t91_n193: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t91_n195;
  t91_n194: return 0.12161882221698761f;
  t91_n195: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t91_n197;
  t91_n196: return -0.011619333177804947f;
  t91_n197: return 0.08505307137966156f;
  t91_n198: return -0.34099113941192627f;
  t91_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t91_n201;
  t91_n200: return -0.08591709285974503f;
  t91_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t91_n207;
  t91_n202: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t91_n206;
  t91_n203: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t91_n205;
  t91_n204: return 0.0001496284967288375f;
  t91_n205: return 0.01849564164876938f;
  t91_n206: return -0.01435636542737484f;
  t91_n207: return 0.018125241622328758f;
  t91_n208: return -0.04179996997117996f;
  t91_n209: return 0.020601915195584297f;
  t91_n210: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t91_n254;
  t91_n211: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t91_n213;
  t91_n212: return 0.002894065575674176f;
  t91_n213: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t91_n219;
  t91_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t91_n218;
  t91_n215: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t91_n217;
  t91_n216: return 0.028277216479182243f;
  t91_n217: return -0.059215035289525986f;
  t91_n218: return 0.030553143471479416f;
  t91_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t91_n221;
  t91_n220: return 0.012445562519133091f;
  t91_n221: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t91_n223;
  t91_n222: return 0.2150525599718094f;
  t91_n223: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t91_n225;
  t91_n224: return -0.07565518468618393f;
  t91_n225: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t91_n231;
  t91_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t91_n230;
  t91_n227: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t91_n229;
  t91_n228: return 0.008939382620155811f;
  t91_n229: return 0.13463707268238068f;
  t91_n230: return 0.10835163295269012f;
  t91_n231: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t91_n235;
  t91_n232: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t91_n234;
  t91_n233: return -0.05194497108459473f;
  t91_n234: return 0.12740568816661835f;
  t91_n235: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t91_n239;
  t91_n236: if (E.getFractionNameInContext() >= 3207689309 /*0.6931818127632141*/) goto t91_n238;
  t91_n237: return 0.10651465505361557f;
  t91_n238: return 0.36156877875328064f;
  t91_n239: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t91_n249;
  t91_n240: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t91_n242;
  t91_n241: return -0.015647171065211296f;
  t91_n242: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t91_n244;
  t91_n243: return 0.14546489715576172f;
  t91_n244: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t91_n248;
  t91_n245: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t91_n247;
  t91_n246: return 0.023373667150735855f;
  t91_n247: return 0.10662500560283661f;
  t91_n248: return 0.01115521602332592f;
  t91_n249: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t91_n253;
  t91_n250: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t91_n252;
  t91_n251: return 0.04513125866651535f;
  t91_n252: return 0.14007534086704254f;
  t91_n253: return 0.05724178999662399f;
  t91_n254: return 0.02501949481666088f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree92(const Example& E) {
  t92_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t92_n196;
  t92_n1: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t92_n175;
  t92_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t92_n124;
  t92_n3: if (E.getFractionNameInContext() >= 3201569812 /*0.41421568393707275*/) goto t92_n83;
  t92_n4: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t92_n82;
  t92_n5: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t92_n61;
  t92_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t92_n12;
  t92_n7: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t92_n9;
  t92_n8: return 0.5883225798606873f;
  t92_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n11;
  t92_n10: return -0.06188279390335083f;
  t92_n11: return -0.012061062268912792f;
  t92_n12: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n60;
  t92_n13: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t92_n23;
  t92_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t92_n18;
  t92_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t92_n17;
  t92_n16: return -0.012521952390670776f;
  t92_n17: return 0.008755907416343689f;
  t92_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t92_n20;
  t92_n19: return 0.009065148420631886f;
  t92_n20: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t92_n22;
  t92_n21: return 0.025512108579277992f;
  t92_n22: return 0.07276234775781631f;
  t92_n23: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t92_n31;
  t92_n24: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t92_n30;
  t92_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t92_n27;
  t92_n26: return -0.0041884249076247215f;
  t92_n27: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t92_n29;
  t92_n28: return 0.035929467529058456f;
  t92_n29: return -0.012484870851039886f;
  t92_n30: return -0.04058610647916794f;
  t92_n31: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t92_n33;
  t92_n32: return -0.012899116612970829f;
  t92_n33: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t92_n59;
  t92_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t92_n58;
  t92_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t92_n55;
  t92_n36: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t92_n42;
  t92_n37: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t92_n39;
  t92_n38: return 0.0018339904490858316f;
  t92_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t92_n41;
  t92_n40: return -0.04849061742424965f;
  t92_n41: return -0.0028530689887702465f;
  t92_n42: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t92_n54;
  t92_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t92_n45;
  t92_n44: return -0.000307566428091377f;
  t92_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t92_n51;
  t92_n46: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t92_n50;
  t92_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t92_n49;
  t92_n48: return -0.042264483869075775f;
  t92_n49: return 0.03777040168642998f;
  t92_n50: return -0.03180178254842758f;
  t92_n51: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t92_n53;
  t92_n52: return 0.05027761310338974f;
  t92_n53: return 0.09279248863458633f;
  t92_n54: return -0.05416381359100342f;
  t92_n55: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t92_n57;
  t92_n56: return 0.02587432600557804f;
  t92_n57: return -0.02021077275276184f;
  t92_n58: return 0.011174716055393219f;
  t92_n59: return -0.11270277202129364f;
  t92_n60: return -0.06529580056667328f;
  t92_n61: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t92_n65;
  t92_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t92_n64;
  t92_n63: return -0.05413823202252388f;
  t92_n64: return -0.031239312142133713f;
  t92_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol))) goto t92_n71;
  t92_n66: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n68;
  t92_n67: return -0.011816535145044327f;
  t92_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n70;
  t92_n69: return -0.1391560435295105f;
  t92_n70: return 0.10420223325490952f;
  t92_n71: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t92_n73;
  t92_n72: return 0.02885681577026844f;
  t92_n73: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t92_n81;
  t92_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t92_n76;
  t92_n75: return -0.012657486833631992f;
  t92_n76: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t92_n80;
  t92_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t92_n79;
  t92_n78: return 0.018839092925190926f;
  t92_n79: return -0.030914850533008575f;
  t92_n80: return 0.04889073967933655f;
  t92_n81: return 0.10970107465982437f;
  t92_n82: return -0.06766953319311142f;
  t92_n83: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t92_n85;
  t92_n84: return 0.008709842339158058f;
  t92_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t92_n87;
  t92_n86: return 0.016646046191453934f;
  t92_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n89;
  t92_n88: return -0.10267875343561172f;
  t92_n89: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t92_n109;
  t92_n90: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t92_n106;
  t92_n91: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t92_n93;
  t92_n92: return 0.009727605618536472f;
  t92_n93: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t92_n105;
  t92_n94: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t92_n102;
  t92_n95: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t92_n97;
  t92_n96: return 0.08724918961524963f;
  t92_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n99;
  t92_n98: return -0.2504003345966339f;
  t92_n99: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n101;
  t92_n100: return 0.12677238881587982f;
  t92_n101: return 0.09871667623519897f;
  t92_n102: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n104;
  t92_n103: return 0.03697112202644348f;
  t92_n104: return 0.10521619021892548f;
  t92_n105: return 0.014111923985183239f;
  t92_n106: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t92_n108;
  t92_n107: return 0.14067880809307098f;
  t92_n108: return 0.07036279141902924f;
  t92_n109: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t92_n123;
  t92_n110: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t92_n116;
  t92_n111: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t92_n113;
  t92_n112: return 0.1058596819639206f;
  t92_n113: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t92_n115;
  t92_n114: return -0.11542295664548874f;
  t92_n115: return 0.10412836074829102f;
  t92_n116: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t92_n120;
  t92_n117: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t92_n119;
  t92_n118: return 0.05671226978302002f;
  t92_n119: return 0.10154279321432114f;
  t92_n120: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t92_n122;
  t92_n121: return 0.12980735301971436f;
  t92_n122: return 0.11684510856866837f;
  t92_n123: return -0.3141418993473053f;
  t92_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t92_n126;
  t92_n125: return -0.04698794707655907f;
  t92_n126: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t92_n128;
  t92_n127: return 0.001835957053117454f;
  t92_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t92_n140;
  t92_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t92_n135;
  t92_n130: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t92_n134;
  t92_n131: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t92_n133;
  t92_n132: return 0.15318997204303741f;
  t92_n133: return -0.05169650539755821f;
  t92_n134: return -0.00041898590279743075f;
  t92_n135: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t92_n139;
  t92_n136: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t92_n138;
  t92_n137: return 0.04909256473183632f;
  t92_n138: return 0.016252435743808746f;
  t92_n139: return 0.1059308648109436f;
  t92_n140: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t92_n160;
  t92_n141: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t92_n159;
  t92_n142: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t92_n150;
  t92_n143: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t92_n149;
  t92_n144: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t92_n148;
  t92_n145: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t92_n147;
  t92_n146: return 0.008972829207777977f;
  t92_n147: return -0.055055733770132065f;
  t92_n148: return 0.029978057369589806f;
  t92_n149: return -0.051339518278837204f;
  t92_n150: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t92_n152;
  t92_n151: return 0.0051327915862202644f;
  t92_n152: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t92_n154;
  t92_n153: return 0.036832429468631744f;
  t92_n154: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t92_n156;
  t92_n155: return 0.1201106607913971f;
  t92_n156: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t92_n158;
  t92_n157: return -0.021765049546957016f;
  t92_n158: return 0.08213617652654648f;
  t92_n159: return -0.11616449058055878f;
  t92_n160: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t92_n162;
  t92_n161: return 0.014450070448219776f;
  t92_n162: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t92_n164;
  t92_n163: return 0.05563390254974365f;
  t92_n164: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t92_n166;
  t92_n165: return -0.10093981772661209f;
  t92_n166: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t92_n170;
  t92_n167: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t92_n169;
  t92_n168: return 0.04946333169937134f;
  t92_n169: return 0.16366060078144073f;
  t92_n170: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t92_n172;
  t92_n171: return -0.11388997733592987f;
  t92_n172: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t92_n174;
  t92_n173: return 0.05441373959183693f;
  t92_n174: return 0.002473287982866168f;
  t92_n175: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t92_n181;
  t92_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n178;
  t92_n177: return -0.023534663021564484f;
  t92_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t92_n180;
  t92_n179: return 0.012036727741360664f;
  t92_n180: return 0.04021640494465828f;
  t92_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t92_n195;
  t92_n182: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t92_n194;
  t92_n183: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t92_n193;
  t92_n184: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t92_n188;
  t92_n185: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t92_n187;
  t92_n186: return 0.009415137581527233f;
  t92_n187: return -0.01834196038544178f;
  t92_n188: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t92_n190;
  t92_n189: return -0.021592773497104645f;
  t92_n190: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t92_n192;
  t92_n191: return 0.07352925837039948f;
  t92_n192: return 0.03595365211367607f;
  t92_n193: return -0.040039677172899246f;
  t92_n194: return -0.009391742758452892f;
  t92_n195: return 0.02190319262444973f;
  t92_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t92_n222;
  t92_n197: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t92_n211;
  t92_n198: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t92_n210;
  t92_n199: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t92_n209;
  t92_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n208;
  t92_n201: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t92_n203;
  t92_n202: return -0.00519915483891964f;
  t92_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t92_n207;
  t92_n204: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n206;
  t92_n205: return -0.008098652586340904f;
  t92_n206: return 0.19801150262355804f;
  t92_n207: return 0.022960403934121132f;
  t92_n208: return 0.018588604405522346f;
  t92_n209: return -6.111132825026289e-05f;
  t92_n210: return 5.273303031572141e-05f;
  t92_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t92_n217;
  t92_n212: if (E.getNumReferences() >= 3365215632 /*305388.5*/) goto t92_n216;
  t92_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n215;
  t92_n214: return -0.44789424538612366f;
  t92_n215: return 0.018744727596640587f;
  t92_n216: return -0.052955929189920425f;
  t92_n217: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t92_n221;
  t92_n218: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t92_n220;
  t92_n219: return 0.07285614311695099f;
  t92_n220: return 0.14198297262191772f;
  t92_n221: return 0.041103314608335495f;
  t92_n222: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t92_n238;
  t92_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t92_n227;
  t92_n224: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t92_n226;
  t92_n225: return 0.1592935472726822f;
  t92_n226: return -0.005853250622749329f;
  t92_n227: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t92_n237;
  t92_n228: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t92_n230;
  t92_n229: return 0.0063949935138225555f;
  t92_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t92_n232;
  t92_n231: return 0.03028133139014244f;
  t92_n232: if (E.getNumReferences() >= 3264675840 /*75.5*/) goto t92_n234;
  t92_n233: return 0.006183510646224022f;
  t92_n234: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t92_n236;
  t92_n235: return 0.10455463081598282f;
  t92_n236: return 0.06666435301303864f;
  t92_n237: return -0.0754176452755928f;
  t92_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t92_n254;
  t92_n239: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t92_n249;
  t92_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t92_n242;
  t92_n241: return 0.008094174787402153f;
  t92_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t92_n248;
  t92_n243: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t92_n245;
  t92_n244: return 0.03483371064066887f;
  t92_n245: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t92_n247;
  t92_n246: return 0.0447511151432991f;
  t92_n247: return 0.16296909749507904f;
  t92_n248: return -0.077614925801754f;
  t92_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t92_n251;
  t92_n250: return -0.036934759467840195f;
  t92_n251: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t92_n253;
  t92_n252: return 0.012578320689499378f;
  t92_n253: return -0.27369052171707153f;
  t92_n254: return 0.0287431962788105f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree93(const Example& E) {
  t93_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t93_n120;
  t93_n1: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t93_n83;
  t93_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t93_n80;
  t93_n3: if (E.getNumReferences() >= 3274801152 /*177.5*/) goto t93_n73;
  t93_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n12;
  t93_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t93_n7;
  t93_n6: return -0.005277651362121105f;
  t93_n7: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t93_n11;
  t93_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t93_n10;
  t93_n9: return 0.06424903869628906f;
  t93_n10: return 0.00852784514427185f;
  t93_n11: return 0.1019224300980568f;
  t93_n12: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t93_n72;
  t93_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t93_n21;
  t93_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n16;
  t93_n15: return -0.00607087230309844f;
  t93_n16: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t93_n20;
  t93_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t93_n19;
  t93_n18: return -0.07012919336557388f;
  t93_n19: return 0.018287163227796555f;
  t93_n20: return -0.012851305305957794f;
  t93_n21: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t93_n65;
  t93_n22: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t93_n26;
  t93_n23: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t93_n25;
  t93_n24: return -0.0701591968536377f;
  t93_n25: return -0.11082833260297775f;
  t93_n26: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t93_n28;
  t93_n27: return 0.0053887758404016495f;
  t93_n28: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t93_n60;
  t93_n29: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t93_n33;
  t93_n30: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t93_n32;
  t93_n31: return 0.36976486444473267f;
  t93_n32: return -0.001280908240005374f;
  t93_n33: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t93_n57;
  t93_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t93_n40;
  t93_n35: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t93_n37;
  t93_n36: return 0.04539719223976135f;
  t93_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t93_n39;
  t93_n38: return 0.08571508526802063f;
  t93_n39: return -0.043848875910043716f;
  t93_n40: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t93_n44;
  t93_n41: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t93_n43;
  t93_n42: return 0.019099876284599304f;
  t93_n43: return -0.2584359049797058f;
  t93_n44: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t93_n56;
  t93_n45: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t93_n55;
  t93_n46: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t93_n48;
  t93_n47: return -0.008611688390374184f;
  t93_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t93_n50;
  t93_n49: return 0.011540954932570457f;
  t93_n50: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t93_n52;
  t93_n51: return 0.13056643307209015f;
  t93_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t93_n54;
  t93_n53: return 0.14293570816516876f;
  t93_n54: return 0.11097671091556549f;
  t93_n55: return -0.7177134156227112f;
  t93_n56: return 0.06746775656938553f;
  t93_n57: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t93_n59;
  t93_n58: return 0.10237827152013779f;
  t93_n59: return -0.02051510103046894f;
  t93_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n62;
  t93_n61: return -0.1186760887503624f;
  t93_n62: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t93_n64;
  t93_n63: return -0.03677191212773323f;
  t93_n64: return 0.056593701243400574f;
  t93_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t93_n71;
  t93_n66: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t93_n68;
  t93_n67: return 0.00380739220418036f;
  t93_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t93_n70;
  t93_n69: return -0.06448348611593246f;
  t93_n70: return 0.0029043867252767086f;
  t93_n71: return 0.02468111738562584f;
  t93_n72: return -0.008535140193998814f;
  t93_n73: if (E.getNumReferences() >= 3350889856 /*95499.0*/) goto t93_n79;
  t93_n74: if (E.getNumReferences() >= 3350665472 /*93746.0*/) goto t93_n78;
  t93_n75: if (E.getNumReferences() >= 3350538496 /*92754.0*/) goto t93_n77;
  t93_n76: return 0.01118648611009121f;
  t93_n77: return -0.17595823109149933f;
  t93_n78: return 0.12659890949726105f;
  t93_n79: return -0.17308039963245392f;
  t93_n80: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t93_n82;
  t93_n81: return 0.011592968367040157f;
  t93_n82: return -0.00615523150190711f;
  t93_n83: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t93_n87;
  t93_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n86;
  t93_n85: return -0.2171567678451538f;
  t93_n86: return 0.10017462819814682f;
  t93_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t93_n103;
  t93_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n102;
  t93_n89: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t93_n99;
  t93_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n92;
  t93_n91: return -0.26703983545303345f;
  t93_n92: if (E.getNumReferences() >= 3351224064 /*98110.0*/) goto t93_n94;
  t93_n93: return 0.003347752382978797f;
  t93_n94: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t93_n96;
  t93_n95: return 0.12375367432832718f;
  t93_n96: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t93_n98;
  t93_n97: return -0.0010431889677420259f;
  t93_n98: return 0.1253684163093567f;
  t93_n99: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t93_n101;
  t93_n100: return -0.062471579760313034f;
  t93_n101: return 0.08958058059215546f;
  t93_n102: return 0.005135675426572561f;
  t93_n103: if (E.getNumReferences() >= 3352190592 /*105661.0*/) goto t93_n105;
  t93_n104: return 0.05964316055178642f;
  t93_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t93_n107;
  t93_n106: return -0.00920926220715046f;
  t93_n107: if (E.getNumReferences() >= 3360662464 /*212623.0*/) goto t93_n117;
  t93_n108: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t93_n110;
  t93_n109: return 0.03359503298997879f;
  t93_n110: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t93_n112;
  t93_n111: return 0.05272378772497177f;
  t93_n112: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t93_n116;
  t93_n113: if (E.getNumReferences() >= 3356575264 /*148760.5*/) goto t93_n115;
  t93_n114: return 0.245758056640625f;
  t93_n115: return 0.1194014623761177f;
  t93_n116: return 0.09903764724731445f;
  t93_n117: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t93_n119;
  t93_n118: return -0.1784844845533371f;
  t93_n119: return 0.030356992036104202f;
  t93_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t93_n198;
  t93_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t93_n123;
  t93_n122: return -0.10652047395706177f;
  t93_n123: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t93_n127;
  t93_n124: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t93_n126;
  t93_n125: return 0.0021736633498221636f;
  t93_n126: return 0.05507505685091019f;
  t93_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t93_n153;
  t93_n128: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t93_n152;
  t93_n129: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t93_n131;
  t93_n130: return 0.11122811585664749f;
  t93_n131: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t93_n151;
  t93_n132: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t93_n150;
  t93_n133: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t93_n135;
  t93_n134: return -0.009446886368095875f;
  t93_n135: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t93_n149;
  t93_n136: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t93_n148;
  t93_n137: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t93_n139;
  t93_n138: return -0.006860444322228432f;
  t93_n139: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t93_n141;
  t93_n140: return 0.01097889058291912f;
  t93_n141: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t93_n145;
  t93_n142: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t93_n144;
  t93_n143: return -0.10064353793859482f;
  t93_n144: return 0.08074560761451721f;
  t93_n145: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t93_n147;
  t93_n146: return -0.2754988372325897f;
  t93_n147: return 0.06281174719333649f;
  t93_n148: return 0.07525892555713654f;
  t93_n149: return -0.06889712065458298f;
  t93_n150: return 0.0016504661180078983f;
  t93_n151: return 0.010167698375880718f;
  t93_n152: return -0.03224068135023117f;
  t93_n153: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t93_n197;
  t93_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t93_n190;
  t93_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n157;
  t93_n156: return -0.09422214329242706f;
  t93_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t93_n183;
  t93_n158: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t93_n182;
  t93_n159: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t93_n179;
  t93_n160: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t93_n178;
  t93_n161: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t93_n163;
  t93_n162: return 0.008295300416648388f;
  t93_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t93_n167;
  t93_n164: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t93_n166;
  t93_n165: return -0.23850344121456146f;
  t93_n166: return 0.06731349974870682f;
  t93_n167: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t93_n177;
  t93_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t93_n170;
  t93_n169: return -0.14636799693107605f;
  t93_n170: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t93_n176;
  t93_n171: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t93_n175;
  t93_n172: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t93_n174;
  t93_n173: return 0.17226101458072662f;
  t93_n174: return 0.10131369531154633f;
  t93_n175: return 0.005560129415243864f;
  t93_n176: return 0.08802050352096558f;
  t93_n177: return 0.04400884360074997f;
  t93_n178: return -0.0019932882860302925f;
  t93_n179: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t93_n181;
  t93_n180: return -0.035987488925457f;
  t93_n181: return 0.015796368941664696f;
  t93_n182: return 0.026497047394514084f;
  t93_n183: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t93_n185;
  t93_n184: return 0.047004327178001404f;
  t93_n185: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t93_n187;
  t93_n186: return -0.004187775310128927f;
  t93_n187: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t93_n189;
  t93_n188: return -0.09045977890491486f;
  t93_n189: return 0.12927474081516266f;
  t93_n190: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t93_n196;
  t93_n191: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t93_n193;
  t93_n192: return 0.006786986254155636f;
  t93_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t93_n195;
  t93_n194: return 0.036946851760149f;
  t93_n195: return 0.0932910218834877f;
  t93_n196: return -0.001295138499699533f;
  t93_n197: return 0.021352669224143028f;
  t93_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t93_n230;
  t93_n199: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t93_n229;
  t93_n200: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t93_n214;
  t93_n201: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t93_n211;
  t93_n202: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t93_n206;
  t93_n203: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t93_n205;
  t93_n204: return -0.0002493478241376579f;
  t93_n205: return -0.050025176256895065f;
  t93_n206: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t93_n210;
  t93_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t93_n209;
  t93_n208: return -0.03214125335216522f;
  t93_n209: return 0.10692787915468216f;
  t93_n210: return -1.0179097652435303f;
  t93_n211: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t93_n213;
  t93_n212: return -0.012881585396826267f;
  t93_n213: return 0.06476243585348129f;
  t93_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t93_n218;
  t93_n215: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t93_n217;
  t93_n216: return -0.03635311871767044f;
  t93_n217: return 0.1536838710308075f;
  t93_n218: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t93_n228;
  t93_n219: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t93_n227;
  t93_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t93_n222;
  t93_n221: return 0.018542708829045296f;
  t93_n222: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t93_n224;
  t93_n223: return -1.1006075143814087f;
  t93_n224: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t93_n226;
  t93_n225: return 0.1348617672920227f;
  t93_n226: return 0.07550366967916489f;
  t93_n227: return -0.2545028328895569f;
  t93_n228: return 0.060935989022254944f;
  t93_n229: return -0.14592330157756805f;
  t93_n230: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t93_n254;
  t93_n231: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t93_n253;
  t93_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t93_n240;
  t93_n233: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t93_n239;
  t93_n234: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t93_n236;
  t93_n235: return 0.019147392362356186f;
  t93_n236: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t93_n238;
  t93_n237: return 0.13455833494663239f;
  t93_n238: return 0.08851335197687149f;
  t93_n239: return -0.008500328287482262f;
  t93_n240: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t93_n250;
  t93_n241: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t93_n243;
  t93_n242: return 0.00888303853571415f;
  t93_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t93_n245;
  t93_n244: return 0.02189423143863678f;
  t93_n245: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t93_n249;
  t93_n246: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t93_n248;
  t93_n247: return 0.06899039447307587f;
  t93_n248: return 0.11496682465076447f;
  t93_n249: return 0.031069176271557808f;
  t93_n250: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t93_n252;
  t93_n251: return -0.06778721511363983f;
  t93_n252: return 0.008835120126605034f;
  t93_n253: return 0.027721155434846878f;
  t93_n254: return 0.02259867452085018f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree94(const Example& E) {
  t94_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t94_n210;
  t94_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t94_n169;
  t94_n2: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t94_n92;
  t94_n3: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t94_n45;
  t94_n4: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t94_n16;
  t94_n5: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t94_n9;
  t94_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t94_n8;
  t94_n7: return -0.31455761194229126f;
  t94_n8: return 0.20321063697338104f;
  t94_n9: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t94_n11;
  t94_n10: return 0.0010968983406201005f;
  t94_n11: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t94_n15;
  t94_n12: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t94_n14;
  t94_n13: return 0.02140156552195549f;
  t94_n14: return -0.003104310715571046f;
  t94_n15: return 0.03394646570086479f;
  t94_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t94_n32;
  t94_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t94_n21;
  t94_n18: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t94_n20;
  t94_n19: return -0.005709023214876652f;
  t94_n20: return -0.09853562712669373f;
  t94_n21: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t94_n27;
  t94_n22: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t94_n24;
  t94_n23: return 0.03233574330806732f;
  t94_n24: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t94_n26;
  t94_n25: return 0.07977107912302017f;
  t94_n26: return 0.16877470910549164f;
  t94_n27: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t94_n29;
  t94_n28: return -0.08096592873334885f;
  t94_n29: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t94_n31;
  t94_n30: return 0.14588254690170288f;
  t94_n31: return 0.013907641172409058f;
  t94_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t94_n38;
  t94_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t94_n35;
  t94_n34: return 0.014803607016801834f;
  t94_n35: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t94_n37;
  t94_n36: return 0.04676531255245209f;
  t94_n37: return -0.16233943402767181f;
  t94_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t94_n44;
  t94_n39: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t94_n43;
  t94_n40: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t94_n42;
  t94_n41: return -0.060599155724048615f;
  t94_n42: return -0.741658091545105f;
  t94_n43: return 0.014173523522913456f;
  t94_n44: return 0.013257818296551704f;
  t94_n45: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t94_n91;
  t94_n46: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t94_n86;
  t94_n47: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t94_n77;
  t94_n48: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t94_n74;
  t94_n49: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t94_n57;
  t94_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t94_n54;
  t94_n51: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t94_n53;
  t94_n52: return -0.033757418394088745f;
  t94_n53: return -0.01952856034040451f;
  t94_n54: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t94_n56;
  t94_n55: return -0.002819733927026391f;
  t94_n56: return 0.003111344063654542f;
  t94_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t94_n71;
  t94_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t94_n70;
  t94_n59: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t94_n61;
  t94_n60: return 0.002779860282316804f;
  t94_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t94_n63;
  t94_n62: return -0.0027858742978423834f;
  t94_n63: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t94_n69;
  t94_n64: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t94_n66;
  t94_n65: return -0.009959799237549305f;
  t94_n66: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t94_n68;
  t94_n67: return 0.014662530273199081f;
  t94_n68: return -0.017989544197916985f;
  t94_n69: return 0.01983397826552391f;
  t94_n70: return 0.02357977069914341f;
  t94_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t94_n73;
  t94_n72: return 0.009244048036634922f;
  t94_n73: return 0.09024184197187424f;
  t94_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t94_n76;
  t94_n75: return -0.011593934148550034f;
  t94_n76: return 0.04291282594203949f;
  t94_n77: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t94_n79;
  t94_n78: return -0.06759187579154968f;
  t94_n79: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t94_n81;
  t94_n80: return 0.008418635465204716f;
  t94_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t94_n83;
  t94_n82: return -0.038527555763721466f;
  t94_n83: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t94_n85;
  t94_n84: return 0.08059608936309814f;
  t94_n85: return 0.009334837086498737f;
  t94_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t94_n90;
  t94_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t94_n89;
  t94_n88: return -0.20243114233016968f;
  t94_n89: return 0.009202924557030201f;
  t94_n90: return 0.05726732686161995f;
  t94_n91: return -0.03982849419116974f;
  t94_n92: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t94_n160;
  t94_n93: if (E.getNumReferences() >= 3338991488 /*34039.5*/) goto t94_n127;
  t94_n94: if (E.getNumReferences() >= 3337447936 /*30389.0*/) goto t94_n126;
  t94_n95: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t94_n103;
  t94_n96: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t94_n102;
  t94_n97: if (E.getNumReferences() >= 3334015488 /*23685.0*/) goto t94_n101;
  t94_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t94_n100;
  t94_n99: return 0.010387470945715904f;
  t94_n100: return 0.02912597730755806f;
  t94_n101: return -0.2001335471868515f;
  t94_n102: return -0.19420255720615387f;
  t94_n103: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t94_n117;
  t94_n104: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t94_n110;
  t94_n105: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t94_n107;
  t94_n106: return 0.08739427477121353f;
  t94_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t94_n109;
  t94_n108: return 0.10324269533157349f;
  t94_n109: return 0.13922041654586792f;
  t94_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t94_n116;
  t94_n111: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t94_n113;
  t94_n112: return 0.10520049929618835f;
  t94_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t94_n115;
  t94_n114: return 0.10903574526309967f;
  t94_n115: return 0.13145111501216888f;
  t94_n116: return -0.006423346698284149f;
  t94_n117: if (E.getNumReferences() >= 3336021248 /*27602.5*/) goto t94_n121;
  t94_n118: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t94_n120;
  t94_n119: return -0.034552477300167084f;
  t94_n120: return 0.10884043574333191f;
  t94_n121: if (E.getNumReferences() >= 3336353536 /*28251.5*/) goto t94_n123;
  t94_n122: return 0.1220279112458229f;
  t94_n123: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t94_n125;
  t94_n124: return 0.0016626229044049978f;
  t94_n125: return 0.12439364939928055f;
  t94_n126: return -0.04525703191757202f;
  t94_n127: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t94_n129;
  t94_n128: return 0.14315585792064667f;
  t94_n129: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t94_n143;
  t94_n130: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t94_n134;
  t94_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t94_n133;
  t94_n132: return -0.06090891733765602f;
  t94_n133: return 0.0655534565448761f;
  t94_n134: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t94_n140;
  t94_n135: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t94_n137;
  t94_n136: return 0.0845475047826767f;
  t94_n137: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t94_n139;
  t94_n138: return 0.10025106370449066f;
  t94_n139: return 0.14475366473197937f;
  t94_n140: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t94_n142;
  t94_n141: return 0.050444457679986954f;
  t94_n142: return 0.09599698334932327f;
  t94_n143: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t94_n145;
  t94_n144: return -0.11351057887077332f;
  t94_n145: if (E.getNumReferences() >= 3347572416 /*69581.5*/) goto t94_n159;
  t94_n146: if (E.getNumReferences() >= 3346692864 /*64123.0*/) goto t94_n156;
  t94_n147: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t94_n155;
  t94_n148: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t94_n154;
  t94_n149: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t94_n151;
  t94_n150: return -0.10176660865545273f;
  t94_n151: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t94_n153;
  t94_n152: return 0.11508582532405853f;
  t94_n153: return 0.06745079159736633f;
  t94_n154: return -0.004875146318227053f;
  t94_n155: return -0.05001409351825714f;
  t94_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t94_n158;
  t94_n157: return -0.1333942413330078f;
  t94_n158: return -0.4347483813762665f;
  t94_n159: return 0.028761498630046844f;
  t94_n160: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t94_n162;
  t94_n161: return 0.04693283513188362f;
  t94_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t94_n164;
  t94_n163: return -0.00019551867444533855f;
  t94_n164: if (E.getNumReferences() >= 3364450048 /*281464.0*/) goto t94_n166;
  t94_n165: return 0.012234300374984741f;
  t94_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t94_n168;
  t94_n167: return -0.18287669122219086f;
  t94_n168: return 0.04640227183699608f;
  t94_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t94_n201;
  t94_n170: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t94_n200;
  t94_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t94_n185;
  t94_n172: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t94_n178;
  t94_n173: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t94_n175;
  t94_n174: return -0.041940558701753616f;
  t94_n175: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t94_n177;
  t94_n176: return 0.04633393883705139f;
  t94_n177: return -0.006627410184592009f;
  t94_n178: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t94_n182;
  t94_n179: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t94_n181;
  t94_n180: return -0.06276480853557587f;
  t94_n181: return 0.06603088229894638f;
  t94_n182: if (E.getFractionNameInContext() >= 3207384269 /*0.675000011920929*/) goto t94_n184;
  t94_n183: return 0.06221460923552513f;
  t94_n184: return -1.503665804862976f;
  t94_n185: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t94_n199;
  t94_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t94_n198;
  t94_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t94_n189;
  t94_n188: return 0.00795654021203518f;
  t94_n189: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t94_n195;
  t94_n190: if (E.getNumReferences() >= 3259760640 /*51.0*/) goto t94_n192;
  t94_n191: return 0.018561966717243195f;
  t94_n192: if (E.getNumReferences() >= 3262644224 /*62.0*/) goto t94_n194;
  t94_n193: return 0.1190069168806076f;
  t94_n194: return 0.11238162219524384f;
  t94_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t94_n197;
  t94_n196: return -0.7253004312515259f;
  t94_n197: return 0.09575732797384262f;
  t94_n198: return 0.047449685633182526f;
  t94_n199: return 0.00633340934291482f;
  t94_n200: return 0.0371045283973217f;
  t94_n201: if (E.getFileProximityDistanceCost() >= 3254255616 /*31.0*/) goto t94_n209;
  t94_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t94_n206;
  t94_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t94_n205;
  t94_n204: return -0.01617267169058323f;
  t94_n205: return 0.03971102833747864f;
  t94_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t94_n208;
  t94_n207: return 0.05582265928387642f;
  t94_n208: return 0.22609694302082062f;
  t94_n209: return 0.05834672227501869f;
  t94_n210: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t94_n238;
  t94_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t94_n237;
  t94_n212: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t94_n236;
  t94_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t94_n223;
  t94_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t94_n216;
  t94_n215: return -0.04963718354701996f;
  t94_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t94_n220;
  t94_n217: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t94_n219;
  t94_n218: return -0.00807760376483202f;
  t94_n219: return -0.027806356549263f;
  t94_n220: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t94_n222;
  t94_n221: return 0.06484037637710571f;
  t94_n222: return -0.8504762649536133f;
  t94_n223: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t94_n233;
  t94_n224: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t94_n226;
  t94_n225: return 0.10725107043981552f;
  t94_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t94_n228;
  t94_n227: return 0.00584561750292778f;
  t94_n228: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t94_n232;
  t94_n229: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t94_n231;
  t94_n230: return 0.06923581659793854f;
  t94_n231: return 0.14099855720996857f;
  t94_n232: return 0.029256649315357208f;
  t94_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t94_n235;
  t94_n234: return -0.0712246522307396f;
  t94_n235: return -0.10927987843751907f;
  t94_n236: return 0.017643630504608154f;
  t94_n237: return 0.06605341285467148f;
  t94_n238: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t94_n242;
  t94_n239: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t94_n241;
  t94_n240: return -0.042231932282447815f;
  t94_n241: return 0.004006099421530962f;
  t94_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t94_n248;
  t94_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t94_n247;
  t94_n244: if (E.getNumReferences() >= 3322022400 /*8322.5*/) goto t94_n246;
  t94_n245: return -0.041329484432935715f;
  t94_n246: return -0.23328797519207f;
  t94_n247: return 0.0035329663660377264f;
  t94_n248: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t94_n250;
  t94_n249: return 0.016490411013364792f;
  t94_n250: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t94_n254;
  t94_n251: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t94_n253;
  t94_n252: return -0.08065760880708694f;
  t94_n253: return 0.01768592931330204f;
  t94_n254: return 0.03050772286951542f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree95(const Example& E) {
  t95_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t95_n158;
  t95_n1: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t95_n71;
  t95_n2: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t95_n4;
  t95_n3: return 0.18072214722633362f;
  t95_n4: if (E.getNumReferences() >= 3306391552 /*2360.5*/) goto t95_n12;
  t95_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t95_n9;
  t95_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t95_n8;
  t95_n7: return -0.006827930919826031f;
  t95_n8: return 0.0008365457179024816f;
  t95_n9: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t95_n11;
  t95_n10: return 0.010160339996218681f;
  t95_n11: return -0.012966041453182697f;
  t95_n12: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t95_n14;
  t95_n13: return 0.014816807582974434f;
  t95_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t95_n26;
  t95_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n17;
  t95_n16: return -0.10846714675426483f;
  t95_n17: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t95_n25;
  t95_n18: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t95_n20;
  t95_n19: return -0.0014113072538748384f;
  t95_n20: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t95_n24;
  t95_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n23;
  t95_n22: return -0.14906372129917145f;
  t95_n23: return 0.06934139132499695f;
  t95_n24: return -0.17653732001781464f;
  t95_n25: return -0.18510372936725616f;
  t95_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n34;
  t95_n27: if (E.getNumReferences() >= 3343775872 /*52728.5*/) goto t95_n33;
  t95_n28: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t95_n30;
  t95_n29: return -0.001080815796740353f;
  t95_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t95_n32;
  t95_n31: return 0.08107807487249374f;
  t95_n32: return 0.14958572387695312f;
  t95_n33: return -0.034542717039585114f;
  t95_n34: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t95_n36;
  t95_n35: return 0.3148078918457031f;
  t95_n36: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t95_n68;
  t95_n37: if (E.getNumReferences() >= 3360198688 /*205376.5*/) goto t95_n61;
  t95_n38: if (E.getNumReferences() >= 3349284736 /*82959.0*/) goto t95_n40;
  t95_n39: return 0.022652875632047653f;
  t95_n40: if (E.getNumReferences() >= 3352190592 /*105661.0*/) goto t95_n50;
  t95_n41: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t95_n45;
  t95_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n44;
  t95_n43: return -0.007618836592882872f;
  t95_n44: return 0.07541248202323914f;
  t95_n45: if (E.getNumReferences() >= 3350654080 /*93657.0*/) goto t95_n47;
  t95_n46: return 0.05369029939174652f;
  t95_n47: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t95_n49;
  t95_n48: return -0.0829242393374443f;
  t95_n49: return 0.09332183748483658f;
  t95_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n52;
  t95_n51: return -0.008381547406315804f;
  t95_n52: if (E.getNumReferences() >= 3356455232 /*146885.0*/) goto t95_n60;
  t95_n53: if (E.getNumReferences() >= 3354093056 /*120524.0*/) goto t95_n55;
  t95_n54: return -0.06641971319913864f;
  t95_n55: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t95_n57;
  t95_n56: return 0.08314071595668793f;
  t95_n57: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t95_n59;
  t95_n58: return 0.2574633061885834f;
  t95_n59: return 0.09729256480932236f;
  t95_n60: return -0.24692372977733612f;
  t95_n61: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t95_n65;
  t95_n62: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t95_n64;
  t95_n63: return 0.06799785792827606f;
  t95_n64: return 0.11596444994211197f;
  t95_n65: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t95_n67;
  t95_n66: return 0.006209371145814657f;
  t95_n67: return 0.07675979286432266f;
  t95_n68: if (E.getNumReferences() >= 3364253712 /*275328.5*/) goto t95_n70;
  t95_n69: return -0.1348223090171814f;
  t95_n70: return 0.023634983226656914f;
  t95_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t95_n155;
  t95_n72: if (E.getFractionNameInContext() >= 3201569812 /*0.41421568393707275*/) goto t95_n126;
  t95_n73: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t95_n107;
  t95_n74: if (E.getNumReferences() >= 3281272832 /*296.5*/) goto t95_n106;
  t95_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t95_n97;
  t95_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n86;
  t95_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t95_n79;
  t95_n78: return -0.21539179980754852f;
  t95_n79: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t95_n85;
  t95_n80: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t95_n82;
  t95_n81: return 0.023987608030438423f;
  t95_n82: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t95_n84;
  t95_n83: return -0.1114472970366478f;
  t95_n84: return -1.1872715950012207f;
  t95_n85: return -0.12225360423326492f;
  t95_n86: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t95_n92;
  t95_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t95_n89;
  t95_n88: return -0.0056367190554738045f;
  t95_n89: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t95_n91;
  t95_n90: return 0.010258487425744534f;
  t95_n91: return -0.07594702392816544f;
  t95_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t95_n94;
  t95_n93: return -0.03255371376872063f;
  t95_n94: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t95_n96;
  t95_n95: return 0.04504862427711487f;
  t95_n96: return 0.0030767356511205435f;
  t95_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t95_n101;
  t95_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n100;
  t95_n99: return -0.024053171277046204f;
  t95_n100: return 0.03342912718653679f;
  t95_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n103;
  t95_n102: return -0.03636186942458153f;
  t95_n103: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t95_n105;
  t95_n104: return 0.075019471347332f;
  t95_n105: return 0.25790467858314514f;
  t95_n106: return 0.02588023990392685f;
  t95_n107: if (E.getNumReferences() >= 3336376576 /*28296.5*/) goto t95_n125;
  t95_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t95_n112;
  t95_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t95_n111;
  t95_n110: return -0.07314224541187286f;
  t95_n111: return -0.01575971581041813f;
  t95_n112: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t95_n122;
  t95_n113: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t95_n115;
  t95_n114: return 0.004678755067288876f;
  t95_n115: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t95_n117;
  t95_n116: return -0.02537819929420948f;
  t95_n117: if (E.getNumReferences() >= 3292520448 /*767.5*/) goto t95_n121;
  t95_n118: if (E.getNumReferences() >= 3292192768 /*747.5*/) goto t95_n120;
  t95_n119: return 0.03660227730870247f;
  t95_n120: return 0.4002256393432617f;
  t95_n121: return -0.059970129281282425f;
  t95_n122: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t95_n124;
  t95_n123: return -0.022457623854279518f;
  t95_n124: return 0.0034562605433166027f;
  t95_n125: return -0.02894885092973709f;
  t95_n126: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t95_n130;
  t95_n127: if (E.getNumReferences() >= 3341955584 /*45618.0*/) goto t95_n129;
  t95_n128: return 0.00953617226332426f;
  t95_n129: return -0.15990397334098816f;
  t95_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t95_n132;
  t95_n131: return 0.016088731586933136f;
  t95_n132: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t95_n136;
  t95_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n135;
  t95_n134: return -0.0980946496129036f;
  t95_n135: return 0.0393402986228466f;
  t95_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n138;
  t95_n137: return -0.14976198971271515f;
  t95_n138: if (E.getFractionNameInContext() >= 3205668418 /*0.5727273225784302*/) goto t95_n144;
  t95_n139: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t95_n141;
  t95_n140: return 0.09843943268060684f;
  t95_n141: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t95_n143;
  t95_n142: return -0.09615065157413483f;
  t95_n143: return 0.0990917906165123f;
  t95_n144: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t95_n154;
  t95_n145: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t95_n153;
  t95_n146: if (E.getFractionNameInContext() >= 3207524079 /*0.6833333373069763*/) goto t95_n152;
  t95_n147: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t95_n151;
  t95_n148: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t95_n150;
  t95_n149: return 0.10087648034095764f;
  t95_n150: return 0.12358830124139786f;
  t95_n151: return 0.07719288766384125f;
  t95_n152: return 0.08514173328876495f;
  t95_n153: return 0.029627803713083267f;
  t95_n154: return 0.13146771490573883f;
  t95_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n157;
  t95_n156: return -0.05740050971508026f;
  t95_n157: return 0.008613122627139091f;
  t95_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t95_n222;
  t95_n159: if (E.getNumReferences() >= 3332163840 /*20068.5*/) goto t95_n203;
  t95_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n188;
  t95_n161: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t95_n185;
  t95_n162: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t95_n170;
  t95_n163: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t95_n169;
  t95_n164: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t95_n166;
  t95_n165: return -0.002584183355793357f;
  t95_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n168;
  t95_n167: return -0.16015243530273438f;
  t95_n168: return 0.028049595654010773f;
  t95_n169: return 0.020300550386309624f;
  t95_n170: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t95_n178;
  t95_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t95_n175;
  t95_n172: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t95_n174;
  t95_n173: return -0.013004312291741371f;
  t95_n174: return 0.026897160336375237f;
  t95_n175: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t95_n177;
  t95_n176: return 0.03716646879911423f;
  t95_n177: return 0.010883591137826443f;
  t95_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t95_n184;
  t95_n179: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t95_n183;
  t95_n180: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t95_n182;
  t95_n181: return 0.007976497523486614f;
  t95_n182: return -0.043349117040634155f;
  t95_n183: return 0.022361399605870247f;
  t95_n184: return 0.01990555413067341f;
  t95_n185: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t95_n187;
  t95_n186: return -0.08914448320865631f;
  t95_n187: return -5.6181153922807425e-05f;
  t95_n188: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t95_n202;
  t95_n189: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t95_n193;
  t95_n190: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t95_n192;
  t95_n191: return -0.18092148005962372f;
  t95_n192: return 0.05985524132847786f;
  t95_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t95_n201;
  t95_n194: if (E.getFractionNameInContext() >= 3177464468 /*0.055727556347846985*/) goto t95_n196;
  t95_n195: return 0.03927094489336014f;
  t95_n196: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t95_n198;
  t95_n197: return 0.05602092295885086f;
  t95_n198: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t95_n200;
  t95_n199: return 0.13575251400470734f;
  t95_n200: return 0.05482390895485878f;
  t95_n201: return -0.021675029769539833f;
  t95_n202: return -0.19750411808490753f;
  t95_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t95_n213;
  t95_n204: if (E.getNumReferences() >= 3365215632 /*305388.5*/) goto t95_n212;
  t95_n205: if (E.getNumReferences() >= 3337910528 /*31292.5*/) goto t95_n209;
  t95_n206: if (E.getNumReferences() >= 3333241856 /*22174.0*/) goto t95_n208;
  t95_n207: return 0.07455262541770935f;
  t95_n208: return -0.06325392425060272f;
  t95_n209: if (E.getNumReferences() >= 3338286080 /*32026.0*/) goto t95_n211;
  t95_n210: return 0.11976706981658936f;
  t95_n211: return 0.018854398280382156f;
  t95_n212: return -0.04495999962091446f;
  t95_n213: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t95_n221;
  t95_n214: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t95_n220;
  t95_n215: if (E.getNumReferences() >= 3334007808 /*23670.0*/) goto t95_n219;
  t95_n216: if (E.getNumReferences() >= 3333976064 /*23608.0*/) goto t95_n218;
  t95_n217: return 0.08170301467180252f;
  t95_n218: return 0.11639399081468582f;
  t95_n219: return -0.1867942214012146f;
  t95_n220: return 0.13679583370685577f;
  t95_n221: return 0.038105469197034836f;
  t95_n222: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t95_n244;
  t95_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t95_n227;
  t95_n224: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t95_n226;
  t95_n225: return -0.050149932503700256f;
  t95_n226: return 0.0014965091831982136f;
  t95_n227: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t95_n243;
  t95_n228: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t95_n242;
  t95_n229: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t95_n231;
  t95_n230: return 0.006289143580943346f;
  t95_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t95_n233;
  t95_n232: return 0.027666712179780006f;
  t95_n233: if (E.getNumReferences() >= 3263496192 /*66.5*/) goto t95_n239;
  t95_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t95_n236;
  t95_n235: return -0.09205795079469681f;
  t95_n236: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t95_n238;
  t95_n237: return 0.050727616995573044f;
  t95_n238: return -0.0018685113172978163f;
  t95_n239: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t95_n241;
  t95_n240: return 0.10182081907987595f;
  t95_n241: return 0.0634288415312767f;
  t95_n242: return -0.09271757304668427f;
  t95_n243: return -0.06682749092578888f;
  t95_n244: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t95_n254;
  t95_n245: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t95_n253;
  t95_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t95_n248;
  t95_n247: return 0.00849141739308834f;
  t95_n248: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t95_n252;
  t95_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t95_n251;
  t95_n250: return 0.027023743838071823f;
  t95_n251: return 0.08546628803014755f;
  t95_n252: return -0.07523422688245773f;
  t95_n253: return -0.09684915095567703f;
  t95_n254: return 0.027416806668043137f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree96(const Example& E) {
  t96_n0: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t96_n184;
  t96_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t96_n163;
  t96_n2: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t96_n4;
  t96_n3: return 0.17851263284683228f;
  t96_n4: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t96_n102;
  t96_n5: if (E.getNumReferences() >= 3279519744 /*249.5*/) goto t96_n71;
  t96_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n70;
  t96_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t96_n23;
  t96_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t96_n10;
  t96_n9: return -0.024151166900992393f;
  t96_n10: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t96_n20;
  t96_n11: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t96_n15;
  t96_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t96_n14;
  t96_n13: return -0.10748805850744247f;
  t96_n14: return 0.023236047476530075f;
  t96_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t96_n17;
  t96_n16: return -0.10293897986412048f;
  t96_n17: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t96_n19;
  t96_n18: return 0.08441146463155746f;
  t96_n19: return -0.25159549713134766f;
  t96_n20: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t96_n22;
  t96_n21: return -0.17551691830158234f;
  t96_n22: return 0.004277938045561314f;
  t96_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t96_n69;
  t96_n24: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t96_n68;
  t96_n25: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t96_n59;
  t96_n26: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t96_n56;
  t96_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t96_n43;
  t96_n28: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t96_n30;
  t96_n29: return -0.007207246031612158f;
  t96_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n40;
  t96_n31: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t96_n39;
  t96_n32: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t96_n36;
  t96_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t96_n35;
  t96_n34: return 0.0522710382938385f;
  t96_n35: return -0.048805493861436844f;
  t96_n36: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t96_n38;
  t96_n37: return -0.05951560661196709f;
  t96_n38: return 0.0022918609902262688f;
  t96_n39: return 0.048666853457689285f;
  t96_n40: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t96_n42;
  t96_n41: return 0.014306494034826756f;
  t96_n42: return -0.026341356337070465f;
  t96_n43: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t96_n53;
  t96_n44: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t96_n46;
  t96_n45: return 0.020410750061273575f;
  t96_n46: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t96_n48;
  t96_n47: return -0.08252322673797607f;
  t96_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t96_n50;
  t96_n49: return 0.0010180105455219746f;
  t96_n50: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t96_n52;
  t96_n51: return 0.23955141007900238f;
  t96_n52: return 0.04476485401391983f;
  t96_n53: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t96_n55;
  t96_n54: return 0.02761908993124962f;
  t96_n55: return 0.07734659314155579f;
  t96_n56: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t96_n58;
  t96_n57: return 0.009084484539926052f;
  t96_n58: return -0.09766903519630432f;
  t96_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t96_n61;
  t96_n60: return -0.02864576317369938f;
  t96_n61: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t96_n65;
  t96_n62: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t96_n64;
  t96_n63: return 0.0300071369856596f;
  t96_n64: return 0.04786424711346626f;
  t96_n65: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t96_n67;
  t96_n66: return -0.09838173538446426f;
  t96_n67: return 0.7815695405006409f;
  t96_n68: return 0.06922069936990738f;
  t96_n69: return -0.04129716753959656f;
  t96_n70: return 0.0010924666421487927f;
  t96_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n77;
  t96_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t96_n74;
  t96_n73: return -0.042061947286129f;
  t96_n74: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t96_n76;
  t96_n75: return 0.0842382088303566f;
  t96_n76: return 0.005824567750096321f;
  t96_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t96_n97;
  t96_n78: if (E.getNumReferences() >= 3331489536 /*18751.5*/) goto t96_n82;
  t96_n79: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t96_n81;
  t96_n80: return 0.021809201687574387f;
  t96_n81: return 0.011354969814419746f;
  t96_n82: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t96_n96;
  t96_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t96_n87;
  t96_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t96_n86;
  t96_n85: return 0.02206784300506115f;
  t96_n86: return 0.09567487239837646f;
  t96_n87: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t96_n95;
  t96_n88: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t96_n94;
  t96_n89: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t96_n91;
  t96_n90: return 0.11238517612218857f;
  t96_n91: if (E.getNumReferences() >= 3331815680 /*19388.5*/) goto t96_n93;
  t96_n92: return -0.1100204810500145f;
  t96_n93: return 0.17558102309703827f;
  t96_n94: return 0.08954524248838425f;
  t96_n95: return 0.13218098878860474f;
  t96_n96: return 0.02058609202504158f;
  t96_n97: if (E.getNumReferences() >= 3279716352 /*252.5*/) goto t96_n101;
  t96_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t96_n100;
  t96_n99: return -0.3054352402687073f;
  t96_n100: return 0.6691856980323792f;
  t96_n101: return -0.005010206252336502f;
  t96_n102: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t96_n146;
  t96_n103: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t96_n145;
  t96_n104: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t96_n124;
  t96_n105: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t96_n111;
  t96_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t96_n108;
  t96_n107: return -0.0002581348526291549f;
  t96_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t96_n110;
  t96_n109: return 0.005418163258582354f;
  t96_n110: return 0.029887618497014046f;
  t96_n111: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t96_n123;
  t96_n112: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t96_n116;
  t96_n113: if (E.getNumReferences() >= 3357252288 /*159339.0*/) goto t96_n115;
  t96_n114: return -0.006633158307522535f;
  t96_n115: return -0.19805808365345f;
  t96_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t96_n118;
  t96_n117: return -0.19648417830467224f;
  t96_n118: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t96_n120;
  t96_n119: return 0.01361667737364769f;
  t96_n120: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t96_n122;
  t96_n121: return 0.08327198028564453f;
  t96_n122: return 0.1310962736606598f;
  t96_n123: return -0.1154707819223404f;
  t96_n124: if (E.getNumReferences() >= 3360662464 /*212623.0*/) goto t96_n138;
  t96_n125: if (E.getNumReferences() >= 3360267424 /*206450.5*/) goto t96_n135;
  t96_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n128;
  t96_n127: return -0.0309059489518404f;
  t96_n128: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t96_n134;
  t96_n129: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t96_n131;
  t96_n130: return -0.13071845471858978f;
  t96_n131: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t96_n133;
  t96_n132: return 0.07651253789663315f;
  t96_n133: return 0.08289399743080139f;
  t96_n134: return -0.16371774673461914f;
  t96_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t96_n137;
  t96_n136: return 0.03218565881252289f;
  t96_n137: return 0.07411736249923706f;
  t96_n138: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t96_n142;
  t96_n139: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t96_n141;
  t96_n140: return -0.10724736750125885f;
  t96_n141: return -1.2643908262252808f;
  t96_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t96_n144;
  t96_n143: return 0.01661098189651966f;
  t96_n144: return 0.07735917717218399f;
  t96_n145: return -0.06745707988739014f;
  t96_n146: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t96_n154;
  t96_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t96_n151;
  t96_n148: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t96_n150;
  t96_n149: return 0.03668008744716644f;
  t96_n150: return -0.050126343965530396f;
  t96_n151: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t96_n153;
  t96_n152: return 0.05915237218141556f;
  t96_n153: return 0.09221746772527695f;
  t96_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t96_n162;
  t96_n155: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t96_n157;
  t96_n156: return -0.03891346603631973f;
  t96_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n161;
  t96_n158: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t96_n160;
  t96_n159: return 0.06974883377552032f;
  t96_n160: return -0.298102468252182f;
  t96_n161: return 0.025691328570246696f;
  t96_n162: return 0.06487032771110535f;
  t96_n163: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t96_n175;
  t96_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n170;
  t96_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t96_n167;
  t96_n166: return -0.0066405488178133965f;
  t96_n167: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t96_n169;
  t96_n168: return -0.022821582853794098f;
  t96_n169: return 0.024497995153069496f;
  t96_n170: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t96_n174;
  t96_n171: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t96_n173;
  t96_n172: return 0.009542739950120449f;
  t96_n173: return -0.010718626901507378f;
  t96_n174: return 0.020407386124134064f;
  t96_n175: if (E.getNumReferences() >= 3338172416 /*31804.0*/) goto t96_n177;
  t96_n176: return 0.09013888239860535f;
  t96_n177: if (E.getNumReferences() >= 3341699072 /*44616.0*/) goto t96_n179;
  t96_n178: return -0.04248563572764397f;
  t96_n179: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t96_n181;
  t96_n180: return 0.08388206362724304f;
  t96_n181: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t96_n183;
  t96_n182: return -0.05584423243999481f;
  t96_n183: return 0.06156519427895546f;
  t96_n184: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t96_n224;
  t96_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t96_n207;
  t96_n186: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t96_n204;
  t96_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t96_n199;
  t96_n188: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t96_n198;
  t96_n189: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t96_n191;
  t96_n190: return -0.0017523706192150712f;
  t96_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t96_n195;
  t96_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t96_n194;
  t96_n193: return -0.18007881939411163f;
  t96_n194: return 0.02639971487224102f;
  t96_n195: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t96_n197;
  t96_n196: return 0.06080184504389763f;
  t96_n197: return 0.21275509893894196f;
  t96_n198: return -0.02137056365609169f;
  t96_n199: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t96_n201;
  t96_n200: return 0.004823196213692427f;
  t96_n201: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t96_n203;
  t96_n202: return -0.010687639005482197f;
  t96_n203: return 0.035746511071920395f;
  t96_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t96_n206;
  t96_n205: return 0.004586603958159685f;
  t96_n206: return 0.019715458154678345f;
  t96_n207: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t96_n223;
  t96_n208: if (E.getNumReferences() >= 3325673472 /*11888.0*/) goto t96_n214;
  t96_n209: if (E.getNumReferences() >= 3325138432 /*11365.5*/) goto t96_n211;
  t96_n210: return 0.004446706734597683f;
  t96_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n213;
  t96_n212: return -0.46834924817085266f;
  t96_n213: return -0.17365705966949463f;
  t96_n214: if (E.getNumReferences() >= 3325918208 /*12127.0*/) goto t96_n218;
  t96_n215: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t96_n217;
  t96_n216: return 0.1632569134235382f;
  t96_n217: return 0.12933683395385742f;
  t96_n218: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t96_n220;
  t96_n219: return 0.040046803653240204f;
  t96_n220: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t96_n222;
  t96_n221: return -0.0033282481599599123f;
  t96_n222: return -0.2198020964860916f;
  t96_n223: return -0.035632263869047165f;
  t96_n224: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t96_n236;
  t96_n225: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t96_n227;
  t96_n226: return 0.007211220450699329f;
  t96_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t96_n231;
  t96_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t96_n230;
  t96_n229: return -0.034171730279922485f;
  t96_n230: return -0.01165702473372221f;
  t96_n231: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t96_n233;
  t96_n232: return 0.1027960404753685f;
  t96_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t96_n235;
  t96_n234: return 0.0013032727874815464f;
  t96_n235: return 0.04759044572710991f;
  t96_n236: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t96_n250;
  t96_n237: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t96_n241;
  t96_n238: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t96_n240;
  t96_n239: return 0.011486397124826908f;
  t96_n240: return -0.01770889200270176f;
  t96_n241: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t96_n249;
  t96_n242: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t96_n246;
  t96_n243: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t96_n245;
  t96_n244: return 0.023092882707715034f;
  t96_n245: return -0.05407016724348068f;
  t96_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t96_n248;
  t96_n247: return 0.01362394168972969f;
  t96_n248: return 0.05829544737935066f;
  t96_n249: return -0.23635819554328918f;
  t96_n250: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t96_n254;
  t96_n251: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t96_n253;
  t96_n252: return 0.04797273501753807f;
  t96_n253: return 0.1277008205652237f;
  t96_n254: return 0.007949513383209705f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree97(const Example& E) {
  t97_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t97_n170;
  t97_n1: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t97_n73;
  t97_n2: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t97_n4;
  t97_n3: return 0.16101592779159546f;
  t97_n4: if (E.getNumReferences() >= 3302502400 /*1729.5*/) goto t97_n70;
  t97_n5: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t97_n11;
  t97_n6: if (E.getNumReferences() >= 3260678144 /*54.5*/) goto t97_n8;
  t97_n7: return 0.0035562252160161734f;
  t97_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t97_n10;
  t97_n9: return 0.011723414063453674f;
  t97_n10: return 0.049515776336193085f;
  t97_n11: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t97_n69;
  t97_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t97_n22;
  t97_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t97_n17;
  t97_n14: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t97_n16;
  t97_n15: return -0.011005700565874577f;
  t97_n16: return 0.0047079757787287235f;
  t97_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t97_n19;
  t97_n18: return -0.01150431577116251f;
  t97_n19: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t97_n21;
  t97_n20: return 0.011324938386678696f;
  t97_n21: return -0.003309202380478382f;
  t97_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t97_n30;
  t97_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t97_n29;
  t97_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n26;
  t97_n25: return -0.01675291545689106f;
  t97_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t97_n28;
  t97_n27: return 0.001236168434843421f;
  t97_n28: return 0.015849323943257332f;
  t97_n29: return 0.0019039789913222194f;
  t97_n30: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t97_n62;
  t97_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n35;
  t97_n32: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t97_n34;
  t97_n33: return 0.024189069867134094f;
  t97_n34: return -0.011107193306088448f;
  t97_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t97_n55;
  t97_n36: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t97_n52;
  t97_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t97_n49;
  t97_n38: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t97_n40;
  t97_n39: return 0.05700492113828659f;
  t97_n40: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t97_n44;
  t97_n41: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t97_n43;
  t97_n42: return 0.423521488904953f;
  t97_n43: return -0.10123776644468307f;
  t97_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t97_n46;
  t97_n45: return 0.10793120414018631f;
  t97_n46: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t97_n48;
  t97_n47: return -0.08508610725402832f;
  t97_n48: return 0.18548601865768433f;
  t97_n49: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t97_n51;
  t97_n50: return -0.007162777241319418f;
  t97_n51: return -0.07453616708517075f;
  t97_n52: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t97_n54;
  t97_n53: return 0.025850029662251472f;
  t97_n54: return -0.046688735485076904f;
  t97_n55: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t97_n57;
  t97_n56: return 0.021130425855517387f;
  t97_n57: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t97_n59;
  t97_n58: return 0.02197456546127796f;
  t97_n59: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t97_n61;
  t97_n60: return -0.10281925648450851f;
  t97_n61: return 0.08124734461307526f;
  t97_n62: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t97_n68;
  t97_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n67;
  t97_n64: if (E.getSymbolScopeDistanceCost() >= 3253207040 /*29.0*/) goto t97_n66;
  t97_n65: return -0.014923522248864174f;
  t97_n66: return 0.007508645765483379f;
  t97_n67: return 0.005839773453772068f;
  t97_n68: return -0.01941458322107792f;
  t97_n69: return -0.04233932867646217f;
  t97_n70: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t97_n72;
  t97_n71: return 0.016070464625954628f;
  t97_n72: return 0.0016709562623873353f;
  t97_n73: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t97_n151;
  t97_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t97_n116;
  t97_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t97_n95;
  t97_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n86;
  t97_n77: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t97_n85;
  t97_n78: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t97_n80;
  t97_n79: return -0.1942182034254074f;
  t97_n80: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t97_n84;
  t97_n81: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t97_n83;
  t97_n82: return -0.057730454951524734f;
  t97_n83: return -0.4578004479408264f;
  t97_n84: return 0.06909125298261642f;
  t97_n85: return 0.0024757578503340483f;
  t97_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t97_n88;
  t97_n87: return -0.005574618000537157f;
  t97_n88: if (E.getNumReferences() >= 3354093056 /*120524.0*/) goto t97_n90;
  t97_n89: return 0.0017509621102362871f;
  t97_n90: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t97_n94;
  t97_n91: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t97_n93;
  t97_n92: return 0.11703398078680038f;
  t97_n93: return 0.02313051000237465f;
  t97_n94: return 0.004229046404361725f;
  t97_n95: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t97_n105;
  t97_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t97_n102;
  t97_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t97_n99;
  t97_n98: return 0.008387850597500801f;
  t97_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t97_n101;
  t97_n100: return -0.01457869540899992f;
  t97_n101: return 0.006348934955894947f;
  t97_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n104;
  t97_n103: return -0.012185383588075638f;
  t97_n104: return 0.03698568791151047f;
  t97_n105: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t97_n109;
  t97_n106: if (E.getFractionNameInContext() >= 3173036146 /*0.039230771362781525*/) goto t97_n108;
  t97_n107: return 0.42525744438171387f;
  t97_n108: return 0.0004620076797436923f;
  t97_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t97_n115;
  t97_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n114;
  t97_n111: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t97_n113;
  t97_n112: return 0.0009034452377818525f;
  t97_n113: return -0.09910125285387039f;
  t97_n114: return 0.02359752357006073f;
  t97_n115: return 0.05079220235347748f;
  t97_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t97_n122;
  t97_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n119;
  t97_n118: return -0.050546303391456604f;
  t97_n119: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t97_n121;
  t97_n120: return -0.005124207586050034f;
  t97_n121: return 0.006749980617314577f;
  t97_n122: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t97_n134;
  t97_n123: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t97_n131;
  t97_n124: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t97_n126;
  t97_n125: return -0.0005204575136303902f;
  t97_n126: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t97_n130;
  t97_n127: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t97_n129;
  t97_n128: return 0.07915069162845612f;
  t97_n129: return -0.011415638960897923f;
  t97_n130: return 0.022863171994686127f;
  t97_n131: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t97_n133;
  t97_n132: return 0.01564497873187065f;
  t97_n133: return -0.043692708015441895f;
  t97_n134: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t97_n140;
  t97_n135: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t97_n139;
  t97_n136: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t97_n138;
  t97_n137: return 0.016190743073821068f;
  t97_n138: return -0.02253161370754242f;
  t97_n139: return 0.03121960535645485f;
  t97_n140: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t97_n142;
  t97_n141: return 0.026241229847073555f;
  t97_n142: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t97_n150;
  t97_n143: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t97_n147;
  t97_n144: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t97_n146;
  t97_n145: return 0.07839792966842651f;
  t97_n146: return -0.1855749785900116f;
  t97_n147: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t97_n149;
  t97_n148: return 0.14716219902038574f;
  t97_n149: return -0.2663700580596924f;
  t97_n150: return -0.06250077486038208f;
  t97_n151: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t97_n161;
  t97_n152: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t97_n160;
  t97_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n159;
  t97_n154: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t97_n156;
  t97_n155: return -0.04071967303752899f;
  t97_n156: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t97_n158;
  t97_n157: return 0.018049612641334534f;
  t97_n158: return -0.03778444975614548f;
  t97_n159: return 0.0074679856188595295f;
  t97_n160: return 0.034454431384801865f;
  t97_n161: if (E.getFileProximityDistanceCost() >= 3254255616 /*31.0*/) goto t97_n169;
  t97_n162: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t97_n164;
  t97_n163: return -0.021324053406715393f;
  t97_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t97_n166;
  t97_n165: return 0.031752973794937134f;
  t97_n166: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t97_n168;
  t97_n167: return -0.10202611982822418f;
  t97_n168: return 0.22961674630641937f;
  t97_n169: return 0.05429450422525406f;
  t97_n170: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t97_n252;
  t97_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t97_n215;
  t97_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n190;
  t97_n173: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t97_n181;
  t97_n174: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t97_n180;
  t97_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n179;
  t97_n176: if (E.getNumReferences() >= 3372375264 /*533966.0*/) goto t97_n178;
  t97_n177: return -0.10832076519727707f;
  t97_n178: return -0.670333743095398f;
  t97_n179: return -0.04445447400212288f;
  t97_n180: return 0.05770185589790344f;
  t97_n181: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t97_n183;
  t97_n182: return -0.038145873695611954f;
  t97_n183: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t97_n185;
  t97_n184: return 0.02609059028327465f;
  t97_n185: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t97_n189;
  t97_n186: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t97_n188;
  t97_n187: return 0.11581505835056305f;
  t97_n188: return 0.06955432146787643f;
  t97_n189: return 0.059029314666986465f;
  t97_n190: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t97_n192;
  t97_n191: return -0.026207756251096725f;
  t97_n192: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t97_n196;
  t97_n193: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t97_n195;
  t97_n194: return -0.0284926425665617f;
  t97_n195: return 0.02144555002450943f;
  t97_n196: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t97_n212;
  t97_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n203;
  t97_n198: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t97_n200;
  t97_n199: return -0.007495576981455088f;
  t97_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n202;
  t97_n201: return -0.13293226063251495f;
  t97_n202: return 0.07946357876062393f;
  t97_n203: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t97_n209;
  t97_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n208;
  t97_n205: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t97_n207;
  t97_n206: return -0.02343798615038395f;
  t97_n207: return 0.011334008537232876f;
  t97_n208: return 0.0662168338894844f;
  t97_n209: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t97_n211;
  t97_n210: return -0.003408993361517787f;
  t97_n211: return 0.02096804976463318f;
  t97_n212: if (E.getFractionNameInContext() >= 3208871340 /*0.7636363506317139*/) goto t97_n214;
  t97_n213: return -0.12452608346939087f;
  t97_n214: return 0.17006327211856842f;
  t97_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n227;
  t97_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n226;
  t97_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n223;
  t97_n218: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t97_n222;
  t97_n219: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t97_n221;
  t97_n220: return -0.24828746914863586f;
  t97_n221: return 0.07274428755044937f;
  t97_n222: return 0.01656860113143921f;
  t97_n223: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t97_n225;
  t97_n224: return -0.05903225392103195f;
  t97_n225: return 0.04933987557888031f;
  t97_n226: return -0.003991343080997467f;
  t97_n227: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t97_n239;
  t97_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t97_n234;
  t97_n229: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t97_n233;
  t97_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t97_n232;
  t97_n231: return -0.041276197880506516f;
  t97_n232: return -0.009531931020319462f;
  t97_n233: return 0.08012118190526962f;
  t97_n234: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t97_n238;
  t97_n235: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t97_n237;
  t97_n236: return 0.006648014299571514f;
  t97_n237: return 0.029377182945609093f;
  t97_n238: return 0.10461415350437164f;
  t97_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t97_n241;
  t97_n240: return 0.02172056958079338f;
  t97_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t97_n247;
  t97_n242: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t97_n244;
  t97_n243: return 0.01488487422466278f;
  t97_n244: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t97_n246;
  t97_n245: return -0.19821158051490784f;
  t97_n246: return 0.07224271446466446f;
  t97_n247: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t97_n251;
  t97_n248: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t97_n250;
  t97_n249: return 0.0387859046459198f;
  t97_n250: return 0.03659095987677574f;
  t97_n251: return 0.0189859326928854f;
  t97_n252: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t97_n254;
  t97_n253: return -0.14764481782913208f;
  t97_n254: return -0.012204686179757118f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree98(const Example& E) {
  t98_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t98_n192;
  t98_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t98_n71;
  t98_n2: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t98_n4;
  t98_n3: return 0.16273769736289978f;
  t98_n4: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t98_n70;
  t98_n5: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t98_n69;
  t98_n6: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t98_n62;
  t98_n7: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t98_n59;
  t98_n8: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t98_n28;
  t98_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n11;
  t98_n10: return -0.005993878003209829f;
  t98_n11: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t98_n13;
  t98_n12: return 0.0014734049327671528f;
  t98_n13: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t98_n23;
  t98_n14: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t98_n16;
  t98_n15: return 0.03679376840591431f;
  t98_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t98_n18;
  t98_n17: return -0.28253814578056335f;
  t98_n18: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t98_n20;
  t98_n19: return 0.07914584130048752f;
  t98_n20: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t98_n22;
  t98_n21: return 0.17333410680294037f;
  t98_n22: return 0.08467022329568863f;
  t98_n23: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t98_n25;
  t98_n24: return -0.06241277977824211f;
  t98_n25: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t98_n27;
  t98_n26: return 0.007148402277380228f;
  t98_n27: return -0.019436774775385857f;
  t98_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t98_n58;
  t98_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n37;
  t98_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t98_n32;
  t98_n31: return -0.057973407208919525f;
  t98_n32: if (E.getNumReferences() >= 3315876864 /*5256.5*/) goto t98_n34;
  t98_n33: return 0.022339800372719765f;
  t98_n34: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t98_n36;
  t98_n35: return 0.14002327620983124f;
  t98_n36: return 0.07870712876319885f;
  t98_n37: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t98_n41;
  t98_n38: if (E.getNumReferences() >= 3350889856 /*95499.0*/) goto t98_n40;
  t98_n39: return 0.01141002681106329f;
  t98_n40: return -0.16455969214439392f;
  t98_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t98_n51;
  t98_n42: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t98_n48;
  t98_n43: if (E.getNumReferences() >= 3351224064 /*98110.0*/) goto t98_n45;
  t98_n44: return 0.060188110917806625f;
  t98_n45: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t98_n47;
  t98_n46: return 0.1350860446691513f;
  t98_n47: return 0.08922827988862991f;
  t98_n48: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t98_n50;
  t98_n49: return 0.04020088538527489f;
  t98_n50: return -0.14669936895370483f;
  t98_n51: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t98_n55;
  t98_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n54;
  t98_n53: return -0.1252622753381729f;
  t98_n54: return 0.10912034660577774f;
  t98_n55: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t98_n57;
  t98_n56: return 0.1243007555603981f;
  t98_n57: return 0.07268614321947098f;
  t98_n58: return -0.005515407305210829f;
  t98_n59: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t98_n61;
  t98_n60: return 0.006396143697202206f;
  t98_n61: return -0.006376660894602537f;
  t98_n62: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t98_n64;
  t98_n63: return 0.036900635808706284f;
  t98_n64: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t98_n66;
  t98_n65: return -0.06616999953985214f;
  t98_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t98_n68;
  t98_n67: return 0.00021914941316936165f;
  t98_n68: return 0.07003740966320038f;
  t98_n69: return -0.06701651215553284f;
  t98_n70: return 0.01752297952771187f;
  t98_n71: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t98_n187;
  t98_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t98_n124;
  t98_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t98_n113;
  t98_n74: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t98_n112;
  t98_n75: if (E.getNumReferences() >= 3270443008 /*119.5*/) goto t98_n109;
  t98_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t98_n84;
  t98_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t98_n79;
  t98_n78: return 0.003634636290371418f;
  t98_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n81;
  t98_n80: return -0.0357145331799984f;
  t98_n81: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t98_n83;
  t98_n82: return -0.014546426944434643f;
  t98_n83: return 0.07605734467506409f;
  t98_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t98_n90;
  t98_n85: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t98_n87;
  t98_n86: return -0.001218349440023303f;
  t98_n87: if (E.getFractionNameInContext() >= 3208578032 /*0.7461538314819336*/) goto t98_n89;
  t98_n88: return -0.12325272709131241f;
  t98_n89: return 0.16126760840415955f;
  t98_n90: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t98_n104;
  t98_n91: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t98_n93;
  t98_n92: return 0.047689612954854965f;
  t98_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t98_n95;
  t98_n94: return 0.035352643579244614f;
  t98_n95: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t98_n101;
  t98_n96: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t98_n98;
  t98_n97: return 0.09510122984647751f;
  t98_n98: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t98_n100;
  t98_n99: return -0.01836196705698967f;
  t98_n100: return 0.11236590892076492f;
  t98_n101: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t98_n103;
  t98_n102: return 0.03823576495051384f;
  t98_n103: return 0.09715348482131958f;
  t98_n104: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t98_n106;
  t98_n105: return 0.016169507056474686f;
  t98_n106: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t98_n108;
  t98_n107: return 0.053951479494571686f;
  t98_n108: return 0.07260750979185104f;
  t98_n109: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t98_n111;
  t98_n110: return 0.017878752201795578f;
  t98_n111: return 0.00887833721935749f;
  t98_n112: return -0.008155182003974915f;
  t98_n113: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t98_n117;
  t98_n114: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t98_n116;
  t98_n115: return 0.010377918370068073f;
  t98_n116: return 0.025318479165434837f;
  t98_n117: if (E.getNumReferences() >= 3362853824 /*246863.0*/) goto t98_n121;
  t98_n118: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t98_n120;
  t98_n119: return 0.002435346134006977f;
  t98_n120: return 0.06432661414146423f;
  t98_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n123;
  t98_n122: return -0.5383580327033997f;
  t98_n123: return -0.06953823566436768f;
  t98_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t98_n142;
  t98_n125: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t98_n127;
  t98_n126: return -0.03136254847049713f;
  t98_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t98_n131;
  t98_n128: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t98_n130;
  t98_n129: return -0.030801502987742424f;
  t98_n130: return 0.004054146818816662f;
  t98_n131: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t98_n133;
  t98_n132: return 0.0034355297684669495f;
  t98_n133: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t98_n135;
  t98_n134: return -0.1203552857041359f;
  t98_n135: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t98_n137;
  t98_n136: return 0.01874522678554058f;
  t98_n137: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t98_n139;
  t98_n138: return -0.034037210047245026f;
  t98_n139: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t98_n141;
  t98_n140: return 0.03131600841879845f;
  t98_n141: return -1.0893442630767822f;
  t98_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t98_n166;
  t98_n143: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t98_n165;
  t98_n144: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t98_n164;
  t98_n145: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t98_n155;
  t98_n146: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t98_n148;
  t98_n147: return 0.007990137673914433f;
  t98_n148: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t98_n154;
  t98_n149: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t98_n151;
  t98_n150: return -0.16469794511795044f;
  t98_n151: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t98_n153;
  t98_n152: return -0.010616338811814785f;
  t98_n153: return 0.11535055190324783f;
  t98_n154: return -0.5359131693840027f;
  t98_n155: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t98_n157;
  t98_n156: return 0.03206474334001541f;
  t98_n157: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t98_n159;
  t98_n158: return 0.03476346284151077f;
  t98_n159: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t98_n161;
  t98_n160: return 0.12208137661218643f;
  t98_n161: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t98_n163;
  t98_n162: return 0.03444336727261543f;
  t98_n163: return 0.1154833510518074f;
  t98_n164: return -0.009136226959526539f;
  t98_n165: return 0.01692260056734085f;
  t98_n166: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t98_n186;
  t98_n167: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t98_n179;
  t98_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t98_n170;
  t98_n169: return -0.029258323833346367f;
  t98_n170: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t98_n176;
  t98_n171: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t98_n173;
  t98_n172: return 0.01190209574997425f;
  t98_n173: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t98_n175;
  t98_n174: return 0.05571133643388748f;
  t98_n175: return 0.10790915787220001f;
  t98_n176: if (E.getNumReferences() >= 3326432256 /*12629.0*/) goto t98_n178;
  t98_n177: return -0.0035147424787282944f;
  t98_n178: return -0.056591324508190155f;
  t98_n179: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t98_n185;
  t98_n180: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t98_n184;
  t98_n181: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t98_n183;
  t98_n182: return 0.09002981334924698f;
  t98_n183: return 0.1621781438589096f;
  t98_n184: return 0.022270865738391876f;
  t98_n185: return -0.10776519030332565f;
  t98_n186: return 0.02044973522424698f;
  t98_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t98_n191;
  t98_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n190;
  t98_n189: return -0.12518255412578583f;
  t98_n190: return -0.07110611349344254f;
  t98_n191: return 0.21487721800804138f;
  t98_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t98_n232;
  t98_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t98_n225;
  t98_n194: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t98_n224;
  t98_n195: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t98_n217;
  t98_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t98_n200;
  t98_n197: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t98_n199;
  t98_n198: return 0.06453418731689453f;
  t98_n199: return -0.04629013314843178f;
  t98_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t98_n216;
  t98_n201: if (E.getNumReferences() >= 3258056704 /*44.5*/) goto t98_n213;
  t98_n202: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t98_n212;
  t98_n203: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t98_n211;
  t98_n204: if (E.getNumReferences() >= 3247964160 /*19.0*/) goto t98_n210;
  t98_n205: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t98_n209;
  t98_n206: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t98_n208;
  t98_n207: return 0.08532910794019699f;
  t98_n208: return -0.2848532497882843f;
  t98_n209: return -0.12932612001895905f;
  t98_n210: return 0.12019854784011841f;
  t98_n211: return 0.026880457997322083f;
  t98_n212: return -0.015484199859201908f;
  t98_n213: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t98_n215;
  t98_n214: return 0.02552773617208004f;
  t98_n215: return 0.0636511817574501f;
  t98_n216: return 0.013243752531707287f;
  t98_n217: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t98_n221;
  t98_n218: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t98_n220;
  t98_n219: return 0.2395303100347519f;
  t98_n220: return 0.015455815009772778f;
  t98_n221: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t98_n223;
  t98_n222: return 0.055840905755758286f;
  t98_n223: return -0.01497582159936428f;
  t98_n224: return -0.0021405937150120735f;
  t98_n225: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t98_n231;
  t98_n226: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t98_n228;
  t98_n227: return 0.02337981015443802f;
  t98_n228: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t98_n230;
  t98_n229: return -0.17499062418937683f;
  t98_n230: return 0.013202507048845291f;
  t98_n231: return 0.05970511585474014f;
  t98_n232: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t98_n252;
  t98_n233: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t98_n249;
  t98_n234: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t98_n238;
  t98_n235: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t98_n237;
  t98_n236: return 0.00954117625951767f;
  t98_n237: return -0.21513068675994873f;
  t98_n238: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t98_n242;
  t98_n239: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t98_n241;
  t98_n240: return 0.007850822992622852f;
  t98_n241: return 0.12301826477050781f;
  t98_n242: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t98_n248;
  t98_n243: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t98_n245;
  t98_n244: return 0.036022499203681946f;
  t98_n245: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t98_n247;
  t98_n246: return 0.10416563600301743f;
  t98_n247: return 0.05675714090466499f;
  t98_n248: return 0.016572317108511925f;
  t98_n249: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t98_n251;
  t98_n250: return 0.11716398596763611f;
  t98_n251: return 0.03724822402000427f;
  t98_n252: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t98_n254;
  t98_n253: return -0.10033774375915527f;
  t98_n254: return -0.24308207631111145f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree99(const Example& E) {
  t99_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t99_n198;
  t99_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t99_n151;
  t99_n2: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t99_n132;
  t99_n3: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t99_n33;
  t99_n4: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t99_n6;
  t99_n5: return 0.0015444798627868295f;
  t99_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n20;
  t99_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t99_n13;
  t99_n8: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t99_n10;
  t99_n9: return -0.005059742834419012f;
  t99_n10: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t99_n12;
  t99_n11: return -0.0322234183549881f;
  t99_n12: return -0.11054477840662003f;
  t99_n13: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t99_n15;
  t99_n14: return 0.029519526287913322f;
  t99_n15: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t99_n17;
  t99_n16: return -0.0653611347079277f;
  t99_n17: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t99_n19;
  t99_n18: return 0.13274972140789032f;
  t99_n19: return 0.0010349774966016412f;
  t99_n20: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t99_n26;
  t99_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t99_n23;
  t99_n22: return 0.014226777479052544f;
  t99_n23: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t99_n25;
  t99_n24: return 0.04305749759078026f;
  t99_n25: return -0.17615503072738647f;
  t99_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n32;
  t99_n27: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n31;
  t99_n28: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t99_n30;
  t99_n29: return -0.05742491036653519f;
  t99_n30: return -0.5629098415374756f;
  t99_n31: return 0.016671760007739067f;
  t99_n32: return 0.011722024530172348f;
  t99_n33: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t99_n127;
  t99_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t99_n78;
  t99_n35: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t99_n77;
  t99_n36: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t99_n60;
  t99_n37: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t99_n43;
  t99_n38: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t99_n42;
  t99_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t99_n41;
  t99_n40: return -0.10941994935274124f;
  t99_n41: return -0.004443175625056028f;
  t99_n42: return -0.06718079745769501f;
  t99_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t99_n59;
  t99_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t99_n56;
  t99_n45: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t99_n47;
  t99_n46: return 0.0016469251131638885f;
  t99_n47: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t99_n55;
  t99_n48: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t99_n50;
  t99_n49: return -0.009630979038774967f;
  t99_n50: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t99_n54;
  t99_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t99_n53;
  t99_n52: return -0.1145457774400711f;
  t99_n53: return 0.27792420983314514f;
  t99_n54: return 0.047054875642061234f;
  t99_n55: return 0.004654848016798496f;
  t99_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t99_n58;
  t99_n57: return 0.0073224278166890144f;
  t99_n58: return 0.09730590134859085f;
  t99_n59: return 0.023251699283719063f;
  t99_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t99_n76;
  t99_n61: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t99_n69;
  t99_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n68;
  t99_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n67;
  t99_n64: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t99_n66;
  t99_n65: return -0.07521215081214905f;
  t99_n66: return -0.1311160922050476f;
  t99_n67: return -0.029215581715106964f;
  t99_n68: return 0.06569109112024307f;
  t99_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n73;
  t99_n70: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n72;
  t99_n71: return -0.012868156656622887f;
  t99_n72: return -0.019898656755685806f;
  t99_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t99_n75;
  t99_n74: return -0.02157612331211567f;
  t99_n75: return 0.013022792525589466f;
  t99_n76: return 0.042078375816345215f;
  t99_n77: return -0.09955935180187225f;
  t99_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t99_n94;
  t99_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t99_n89;
  t99_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n88;
  t99_n81: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t99_n87;
  t99_n82: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t99_n86;
  t99_n83: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t99_n85;
  t99_n84: return -0.015078632161021233f;
  t99_n85: return -0.065934918820858f;
  t99_n86: return 0.09079433232545853f;
  t99_n87: return -0.06779919564723969f;
  t99_n88: return 0.0010198509553447366f;
  t99_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n93;
  t99_n90: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t99_n92;
  t99_n91: return -0.058985836803913116f;
  t99_n92: return 0.0012231194414198399f;
  t99_n93: return 0.007814143784344196f;
  t99_n94: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t99_n118;
  t99_n95: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t99_n97;
  t99_n96: return 0.018722452223300934f;
  t99_n97: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t99_n107;
  t99_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n106;
  t99_n99: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t99_n105;
  t99_n100: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t99_n104;
  t99_n101: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t99_n103;
  t99_n102: return 0.10765405744314194f;
  t99_n103: return -0.1518353521823883f;
  t99_n104: return -0.2082284688949585f;
  t99_n105: return -0.06889740377664566f;
  t99_n106: return -0.002982890931889415f;
  t99_n107: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t99_n113;
  t99_n108: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t99_n110;
  t99_n109: return 0.007288878783583641f;
  t99_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t99_n112;
  t99_n111: return 0.0014144539600238204f;
  t99_n112: return 0.007840381935238838f;
  t99_n113: if (E.getNumReferences() >= 3261857792 /*59.0*/) goto t99_n117;
  t99_n114: if (E.getNumReferences() >= 3244294144 /*14.0*/) goto t99_n116;
  t99_n115: return 0.15013937652111053f;
  t99_n116: return 0.6597793698310852f;
  t99_n117: return -0.17874301970005035f;
  t99_n118: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t99_n124;
  t99_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t99_n121;
  t99_n120: return -0.010830181650817394f;
  t99_n121: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t99_n123;
  t99_n122: return 0.05010885372757912f;
  t99_n123: return 0.1305907666683197f;
  t99_n124: if (E.getNumReferences() >= 3278471168 /*233.5*/) goto t99_n126;
  t99_n125: return -0.05166163668036461f;
  t99_n126: return 0.12046488374471664f;
  t99_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t99_n131;
  t99_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t99_n130;
  t99_n129: return -0.19320763647556305f;
  t99_n130: return 0.00821281224489212f;
  t99_n131: return 0.052210960537195206f;
  t99_n132: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t99_n144;
  t99_n133: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n143;
  t99_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t99_n142;
  t99_n135: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t99_n139;
  t99_n136: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t99_n138;
  t99_n137: return 0.013999994844198227f;
  t99_n138: return -0.1838485300540924f;
  t99_n139: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t99_n141;
  t99_n140: return 0.10970006883144379f;
  t99_n141: return 0.022536104544997215f;
  t99_n142: return -0.0019393296679481864f;
  t99_n143: return 0.03376977890729904f;
  t99_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t99_n146;
  t99_n145: return -0.0012338601518422365f;
  t99_n146: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t99_n148;
  t99_n147: return 0.01917070522904396f;
  t99_n148: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t99_n150;
  t99_n149: return 0.13134025037288666f;
  t99_n150: return 0.05359535291790962f;
  t99_n151: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t99_n167;
  t99_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n166;
  t99_n153: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t99_n157;
  t99_n154: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t99_n156;
  t99_n155: return -0.004085042048245668f;
  t99_n156: return 0.06299248337745667f;
  t99_n157: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t99_n163;
  t99_n158: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t99_n162;
  t99_n159: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t99_n161;
  t99_n160: return 0.08013243228197098f;
  t99_n161: return -0.06580398231744766f;
  t99_n162: return 0.03010997548699379f;
  t99_n163: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t99_n165;
  t99_n164: return 0.06590551882982254f;
  t99_n165: return -0.23520757257938385f;
  t99_n166: return 0.0076425038278102875f;
  t99_n167: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t99_n195;
  t99_n168: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t99_n170;
  t99_n169: return 0.0003143441863358021f;
  t99_n170: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t99_n174;
  t99_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t99_n173;
  t99_n172: return 0.07702848315238953f;
  t99_n173: return 0.11993374675512314f;
  t99_n174: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t99_n186;
  t99_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t99_n177;
  t99_n176: return -0.05880605801939964f;
  t99_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t99_n179;
  t99_n178: return 0.008715328760445118f;
  t99_n179: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t99_n181;
  t99_n180: return 0.022907624021172523f;
  t99_n181: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t99_n185;
  t99_n182: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t99_n184;
  t99_n183: return 0.13905145227909088f;
  t99_n184: return 0.12272751331329346f;
  t99_n185: return 0.08664382994174957f;
  t99_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t99_n188;
  t99_n187: return 0.0025321936700493097f;
  t99_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n190;
  t99_n189: return 0.08682474493980408f;
  t99_n190: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t99_n192;
  t99_n191: return -0.10027185827493668f;
  t99_n192: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t99_n194;
  t99_n193: return 0.3104957640171051f;
  t99_n194: return -0.04834887012839317f;
  t99_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t99_n197;
  t99_n196: return 0.021231885999441147f;
  t99_n197: return 0.1858130842447281f;
  t99_n198: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t99_n222;
  t99_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t99_n221;
  t99_n200: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t99_n216;
  t99_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n207;
  t99_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t99_n204;
  t99_n203: return -0.03760313615202904f;
  t99_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t99_n206;
  t99_n205: return -0.011003481224179268f;
  t99_n206: return 0.05130103975534439f;
  t99_n207: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t99_n213;
  t99_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t99_n212;
  t99_n209: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t99_n211;
  t99_n210: return 0.09960036724805832f;
  t99_n211: return 0.006168649531900883f;
  t99_n212: return 0.06833048164844513f;
  t99_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t99_n215;
  t99_n214: return -0.06519705057144165f;
  t99_n215: return -0.10861965268850327f;
  t99_n216: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t99_n220;
  t99_n217: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t99_n219;
  t99_n218: return 0.016042612493038177f;
  t99_n219: return -0.01813739724457264f;
  t99_n220: return 0.07377640157938004f;
  t99_n221: return 0.059790872037410736f;
  t99_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t99_n240;
  t99_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n237;
  t99_n224: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t99_n234;
  t99_n225: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t99_n229;
  t99_n226: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t99_n228;
  t99_n227: return 0.07130005955696106f;
  t99_n228: return -0.021084409207105637f;
  t99_n229: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t99_n231;
  t99_n230: return -0.11645497381687164f;
  t99_n231: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t99_n233;
  t99_n232: return 0.09516090154647827f;
  t99_n233: return 0.1661548912525177f;
  t99_n234: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t99_n236;
  t99_n235: return -0.038089603185653687f;
  t99_n236: return -0.11321917921304703f;
  t99_n237: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t99_n239;
  t99_n238: return 0.12040425091981888f;
  t99_n239: return 0.002454552333801985f;
  t99_n240: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t99_n246;
  t99_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t99_n245;
  t99_n242: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t99_n244;
  t99_n243: return 0.11681193858385086f;
  t99_n244: return -0.06372161954641342f;
  t99_n245: return -0.006897261366248131f;
  t99_n246: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t99_n248;
  t99_n247: return 0.014691525138914585f;
  t99_n248: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t99_n254;
  t99_n249: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t99_n253;
  t99_n250: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t99_n252;
  t99_n251: return 0.10301855206489563f;
  t99_n252: return -0.0943586677312851f;
  t99_n253: return 0.016246341168880463f;
  t99_n254: return 0.027959097176790237f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree100(const Example& E) {
  t100_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t100_n226;
  t100_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t100_n121;
  t100_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t100_n16;
  t100_n3: if (E.getNumReferences() >= 3274932224 /*179.5*/) goto t100_n9;
  t100_n4: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t100_n6;
  t100_n5: return 0.14852163195610046f;
  t100_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n8;
  t100_n7: return -0.004314564634114504f;
  t100_n8: return 0.0009069134830497205f;
  t100_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n13;
  t100_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t100_n12;
  t100_n11: return -0.014442251063883305f;
  t100_n12: return 0.4228764474391937f;
  t100_n13: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t100_n15;
  t100_n14: return 0.01675611175596714f;
  t100_n15: return -0.004480951465666294f;
  t100_n16: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t100_n98;
  t100_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n31;
  t100_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword))) goto t100_n20;
  t100_n19: return -0.015250122174620628f;
  t100_n20: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t100_n26;
  t100_n21: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t100_n25;
  t100_n22: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t100_n24;
  t100_n23: return -0.0010465983068570495f;
  t100_n24: return 0.020679809153079987f;
  t100_n25: return -0.10824573785066605f;
  t100_n26: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t100_n28;
  t100_n27: return -0.17988315224647522f;
  t100_n28: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t100_n30;
  t100_n29: return 0.13181203603744507f;
  t100_n30: return 0.07660937309265137f;
  t100_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t100_n63;
  t100_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n38;
  t100_n33: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t100_n37;
  t100_n34: if (E.getNumReferences() >= 3325242368 /*11467.0*/) goto t100_n36;
  t100_n35: return -0.09868740290403366f;
  t100_n36: return -0.14804381132125854f;
  t100_n37: return -0.18318983912467957f;
  t100_n38: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t100_n62;
  t100_n39: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t100_n47;
  t100_n40: if (E.getNumReferences() >= 3277717504 /*222.0*/) goto t100_n46;
  t100_n41: if (E.getNumReferences() >= 3274342400 /*170.5*/) goto t100_n43;
  t100_n42: return 0.0004634713986888528f;
  t100_n43: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t100_n45;
  t100_n44: return 0.22840243577957153f;
  t100_n45: return -0.10457402467727661f;
  t100_n46: return -0.09234851598739624f;
  t100_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t100_n49;
  t100_n48: return -0.10572408884763718f;
  t100_n49: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t100_n61;
  t100_n50: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t100_n52;
  t100_n51: return -0.11094490438699722f;
  t100_n52: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t100_n58;
  t100_n53: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t100_n57;
  t100_n54: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t100_n56;
  t100_n55: return 0.0570329874753952f;
  t100_n56: return -0.01367179024964571f;
  t100_n57: return 0.05743873491883278f;
  t100_n58: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t100_n60;
  t100_n59: return -0.01148964837193489f;
  t100_n60: return 0.04374879226088524f;
  t100_n61: return 0.01624974235892296f;
  t100_n62: return -0.01144713256508112f;
  t100_n63: if (E.getNumReferences() >= 3347421312 /*68401.0*/) goto t100_n87;
  t100_n64: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t100_n86;
  t100_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t100_n81;
  t100_n66: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t100_n78;
  t100_n67: if (E.getNumReferences() >= 3344267904 /*54650.5*/) goto t100_n75;
  t100_n68: if (E.getNumReferences() >= 3342301952 /*46971.0*/) goto t100_n74;
  t100_n69: if (E.getNumReferences() >= 3342115200 /*46241.5*/) goto t100_n73;
  t100_n70: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t100_n72;
  t100_n71: return 0.0034833326935768127f;
  t100_n72: return -0.0478060208261013f;
  t100_n73: return 0.08846301585435867f;
  t100_n74: return -0.07909993082284927f;
  t100_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n77;
  t100_n76: return -0.10427838563919067f;
  t100_n77: return 0.11046022176742554f;
  t100_n78: if (E.getNumReferences() >= 3345968640 /*61294.0*/) goto t100_n80;
  t100_n79: return -0.4507174491882324f;
  t100_n80: return -0.05568827688694f;
  t100_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n83;
  t100_n82: return -0.11211608350276947f;
  t100_n83: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t100_n85;
  t100_n84: return 0.023784564808011055f;
  t100_n85: return 0.006437968462705612f;
  t100_n86: return -0.026146072894334793f;
  t100_n87: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t100_n95;
  t100_n88: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t100_n92;
  t100_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n91;
  t100_n90: return -0.13137024641036987f;
  t100_n91: return 0.02431873045861721f;
  t100_n92: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t100_n94;
  t100_n93: return -0.19164299964904785f;
  t100_n94: return 0.0526309497654438f;
  t100_n95: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t100_n97;
  t100_n96: return -0.12510475516319275f;
  t100_n97: return 0.017933281138539314f;
  t100_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n116;
  t100_n99: if (E.getNumReferences() >= 3344576000 /*55854.0*/) goto t100_n115;
  t100_n100: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t100_n110;
  t100_n101: if (E.getNumReferences() >= 3334309120 /*24258.5*/) goto t100_n109;
  t100_n102: if (E.getNumReferences() >= 3333992192 /*23639.5*/) goto t100_n104;
  t100_n103: return -0.0436314195394516f;
  t100_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n106;
  t100_n105: return -0.1212577298283577f;
  t100_n106: if (E.getNumReferences() >= 3334143744 /*23935.5*/) goto t100_n108;
  t100_n107: return 0.21428674459457397f;
  t100_n108: return 0.09385866671800613f;
  t100_n109: return -0.13757283985614777f;
  t100_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n112;
  t100_n111: return -0.11167199164628983f;
  t100_n112: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t100_n114;
  t100_n113: return 0.0805797129869461f;
  t100_n114: return 0.15519489347934723f;
  t100_n115: return -0.11080110818147659f;
  t100_n116: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t100_n118;
  t100_n117: return -0.06923390179872513f;
  t100_n118: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t100_n120;
  t100_n119: return 0.0018581333570182323f;
  t100_n120: return 0.12113368511199951f;
  t100_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t100_n211;
  t100_n122: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t100_n202;
  t100_n123: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t100_n165;
  t100_n124: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t100_n164;
  t100_n125: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t100_n163;
  t100_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t100_n136;
  t100_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t100_n133;
  t100_n128: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t100_n130;
  t100_n129: return -0.03556276485323906f;
  t100_n130: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t100_n132;
  t100_n131: return -0.02535202167928219f;
  t100_n132: return -0.17909114062786102f;
  t100_n133: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t100_n135;
  t100_n134: return 0.003916676621884108f;
  t100_n135: return -0.10898133367300034f;
  t100_n136: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t100_n138;
  t100_n137: return 0.009950281120836735f;
  t100_n138: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t100_n146;
  t100_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t100_n143;
  t100_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t100_n142;
  t100_n141: return 0.013604857958853245f;
  t100_n142: return 0.06603100895881653f;
  t100_n143: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t100_n145;
  t100_n144: return 0.07670482993125916f;
  t100_n145: return -0.07871101051568985f;
  t100_n146: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t100_n162;
  t100_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t100_n161;
  t100_n148: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t100_n160;
  t100_n149: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t100_n157;
  t100_n150: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t100_n154;
  t100_n151: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t100_n153;
  t100_n152: return -0.3373510539531708f;
  t100_n153: return -0.032392941415309906f;
  t100_n154: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t100_n156;
  t100_n155: return 0.11513679474592209f;
  t100_n156: return 0.005568983033299446f;
  t100_n157: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t100_n159;
  t100_n158: return 0.14142197370529175f;
  t100_n159: return 0.08310264348983765f;
  t100_n160: return -0.32429245114326477f;
  t100_n161: return 0.015557819977402687f;
  t100_n162: return -0.23981168866157532f;
  t100_n163: return -0.005154765211045742f;
  t100_n164: return -0.00045541964937001467f;
  t100_n165: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t100_n195;
  t100_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t100_n194;
  t100_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t100_n179;
  t100_n168: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t100_n178;
  t100_n169: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t100_n171;
  t100_n170: return 0.012812839820981026f;
  t100_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t100_n173;
  t100_n172: return 0.022538647055625916f;
  t100_n173: if (E.getFractionNameInContext() >= 3206431018 /*0.6181818246841431*/) goto t100_n177;
  t100_n174: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t100_n176;
  t100_n175: return 0.07879756391048431f;
  t100_n176: return 0.11885593086481094f;
  t100_n177: return 0.0704910159111023f;
  t100_n178: return -0.019666172564029694f;
  t100_n179: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t100_n189;
  t100_n180: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t100_n182;
  t100_n181: return -0.024560170248150826f;
  t100_n182: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t100_n184;
  t100_n183: return 0.014211814850568771f;
  t100_n184: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t100_n186;
  t100_n185: return 0.0618579164147377f;
  t100_n186: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t100_n188;
  t100_n187: return 0.11598203331232071f;
  t100_n188: return -0.31250908970832825f;
  t100_n189: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t100_n191;
  t100_n190: return 0.0354391485452652f;
  t100_n191: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t100_n193;
  t100_n192: return 0.13052335381507874f;
  t100_n193: return -0.004047596361488104f;
  t100_n194: return 0.03167496994137764f;
  t100_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t100_n197;
  t100_n196: return 0.05771949887275696f;
  t100_n197: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t100_n199;
  t100_n198: return 0.06214025244116783f;
  t100_n199: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t100_n201;
  t100_n200: return 0.2697189152240753f;
  t100_n201: return 0.13512983918190002f;
  t100_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n204;
  t100_n203: return -0.0420088917016983f;
  t100_n204: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t100_n206;
  t100_n205: return 0.005038122646510601f;
  t100_n206: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t100_n210;
  t100_n207: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t100_n209;
  t100_n208: return 0.021284958347678185f;
  t100_n209: return 0.04455047473311424f;
  t100_n210: return -0.0029630365315824747f;
  t100_n211: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t100_n215;
  t100_n212: if (E.getNumReferences() >= 3291389952 /*698.5*/) goto t100_n214;
  t100_n213: return 0.001918526366353035f;
  t100_n214: return 0.026152795180678368f;
  t100_n215: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t100_n221;
  t100_n216: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t100_n220;
  t100_n217: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t100_n219;
  t100_n218: return 0.05883749574422836f;
  t100_n219: return -0.007273177150636911f;
  t100_n220: return -0.06722814589738846f;
  t100_n221: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t100_n225;
  t100_n222: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t100_n224;
  t100_n223: return 0.01757732965052128f;
  t100_n224: return 0.05247131362557411f;
  t100_n225: return -0.11024702340364456f;
  t100_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t100_n244;
  t100_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t100_n237;
  t100_n228: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t100_n236;
  t100_n229: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t100_n233;
  t100_n230: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t100_n232;
  t100_n231: return 0.013578586280345917f;
  t100_n232: return 0.03930527716875076f;
  t100_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t100_n235;
  t100_n234: return -0.03696705773472786f;
  t100_n235: return 0.008251311257481575f;
  t100_n236: return -0.0019442526390776038f;
  t100_n237: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t100_n243;
  t100_n238: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t100_n240;
  t100_n239: return 0.02209644578397274f;
  t100_n240: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t100_n242;
  t100_n241: return -0.1550694853067398f;
  t100_n242: return 0.017209317535161972f;
  t100_n243: return 0.05723072588443756f;
  t100_n244: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t100_n254;
  t100_n245: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t100_n253;
  t100_n246: if (E.getNumReferences() >= 3317853184 /*6221.5*/) goto t100_n250;
  t100_n247: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t100_n249;
  t100_n248: return 0.009669322520494461f;
  t100_n249: return -0.01725599355995655f;
  t100_n250: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t100_n252;
  t100_n251: return 0.1152254268527031f;
  t100_n252: return 0.034028586000204086f;
  t100_n253: return 0.024989960715174675f;
  t100_n254: return -0.12545153498649597f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree101(const Example& E) {
  t101_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t101_n138;
  t101_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t101_n41;
  t101_n2: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t101_n38;
  t101_n3: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t101_n35;
  t101_n4: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t101_n14;
  t101_n5: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t101_n9;
  t101_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n8;
  t101_n7: return -0.003968219272792339f;
  t101_n8: return 0.0011410254519432783f;
  t101_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n11;
  t101_n10: return -0.01028823759406805f;
  t101_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t101_n13;
  t101_n12: return 0.015523781068623066f;
  t101_n13: return -0.0040281591936945915f;
  t101_n14: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t101_n16;
  t101_n15: return 0.007345923222601414f;
  t101_n16: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t101_n18;
  t101_n17: return -0.005597065202891827f;
  t101_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t101_n24;
  t101_n19: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t101_n21;
  t101_n20: return -0.043716445565223694f;
  t101_n21: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t101_n23;
  t101_n22: return 0.010196666233241558f;
  t101_n23: return -0.2649160623550415f;
  t101_n24: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t101_n28;
  t101_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n27;
  t101_n26: return -0.08424253016710281f;
  t101_n27: return 0.028276793658733368f;
  t101_n28: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t101_n30;
  t101_n29: return -0.1257839798927307f;
  t101_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t101_n32;
  t101_n31: return 0.02276376448571682f;
  t101_n32: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t101_n34;
  t101_n33: return 0.11250777542591095f;
  t101_n34: return 0.039444323629140854f;
  t101_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t101_n37;
  t101_n36: return -0.18453891575336456f;
  t101_n37: return -0.03664669021964073f;
  t101_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t101_n40;
  t101_n39: return 0.041934918612241745f;
  t101_n40: return -0.07095885276794434f;
  t101_n41: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t101_n43;
  t101_n42: return 0.0014509694883599877f;
  t101_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t101_n107;
  t101_n44: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t101_n106;
  t101_n45: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t101_n105;
  t101_n46: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t101_n82;
  t101_n47: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t101_n53;
  t101_n48: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t101_n50;
  t101_n49: return -0.011553792282938957f;
  t101_n50: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t101_n52;
  t101_n51: return -0.07822980731725693f;
  t101_n52: return 0.11413540691137314f;
  t101_n53: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t101_n75;
  t101_n54: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t101_n74;
  t101_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t101_n69;
  t101_n56: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t101_n58;
  t101_n57: return 0.016651315614581108f;
  t101_n58: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t101_n62;
  t101_n59: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t101_n61;
  t101_n60: return -0.027953004464507103f;
  t101_n61: return 0.07188111543655396f;
  t101_n62: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t101_n68;
  t101_n63: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t101_n67;
  t101_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t101_n66;
  t101_n65: return -0.14091888070106506f;
  t101_n66: return -0.03565693274140358f;
  t101_n67: return -0.4128842055797577f;
  t101_n68: return 0.08196958899497986f;
  t101_n69: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t101_n73;
  t101_n70: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t101_n72;
  t101_n71: return -0.22204874455928802f;
  t101_n72: return 0.007588895969092846f;
  t101_n73: return 0.08577319234609604f;
  t101_n74: return 0.07364681363105774f;
  t101_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t101_n81;
  t101_n76: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t101_n80;
  t101_n77: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t101_n79;
  t101_n78: return -0.3069593906402588f;
  t101_n79: return 0.08134343475103378f;
  t101_n80: return -0.09227194637060165f;
  t101_n81: return -0.030675522983074188f;
  t101_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t101_n84;
  t101_n83: return -0.027374254539608955f;
  t101_n84: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t101_n104;
  t101_n85: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t101_n89;
  t101_n86: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t101_n88;
  t101_n87: return -0.009214445017278194f;
  t101_n88: return 0.0029835456516593695f;
  t101_n89: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t101_n95;
  t101_n90: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t101_n94;
  t101_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t101_n93;
  t101_n92: return 0.07456804066896439f;
  t101_n93: return 0.09793902933597565f;
  t101_n94: return -0.04783545807003975f;
  t101_n95: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t101_n99;
  t101_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t101_n98;
  t101_n97: return -0.031004715710878372f;
  t101_n98: return 0.07182075083255768f;
  t101_n99: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t101_n103;
  t101_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t101_n102;
  t101_n101: return -0.059904795140028f;
  t101_n102: return 0.10095316916704178f;
  t101_n103: return -0.007855425588786602f;
  t101_n104: return 0.023685602471232414f;
  t101_n105: return 0.00952492468059063f;
  t101_n106: return -0.055010583251714706f;
  t101_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t101_n129;
  t101_n108: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t101_n126;
  t101_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n123;
  t101_n110: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t101_n112;
  t101_n111: return -0.020175283774733543f;
  t101_n112: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t101_n114;
  t101_n113: return 0.0015154278371483088f;
  t101_n114: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t101_n122;
  t101_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t101_n117;
  t101_n116: return -0.10070914030075073f;
  t101_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n119;
  t101_n118: return -0.10774274170398712f;
  t101_n119: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t101_n121;
  t101_n120: return 0.08308343589305878f;
  t101_n121: return 0.04746146872639656f;
  t101_n122: return -0.009227133356034756f;
  t101_n123: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t101_n125;
  t101_n124: return 0.0981799066066742f;
  t101_n125: return 0.003887983039021492f;
  t101_n126: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t101_n128;
  t101_n127: return 0.021279916167259216f;
  t101_n128: return 0.14140349626541138f;
  t101_n129: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t101_n137;
  t101_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro))) goto t101_n132;
  t101_n131: return 0.0059258961118757725f;
  t101_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n134;
  t101_n133: return -0.17007794976234436f;
  t101_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t101_n136;
  t101_n135: return 0.016996214166283607f;
  t101_n136: return 0.07382015138864517f;
  t101_n137: return -0.001415493432432413f;
  t101_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t101_n216;
  t101_n139: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t101_n209;
  t101_n140: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t101_n192;
  t101_n141: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t101_n161;
  t101_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t101_n144;
  t101_n143: return -0.01809687912464142f;
  t101_n144: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t101_n152;
  t101_n145: if (E.getNumReferences() >= 3348289088 /*75180.5*/) goto t101_n151;
  t101_n146: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t101_n148;
  t101_n147: return -0.006900365464389324f;
  t101_n148: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t101_n150;
  t101_n149: return 0.03847435116767883f;
  t101_n150: return -0.006577826105058193f;
  t101_n151: return -1.329437494277954f;
  t101_n152: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t101_n160;
  t101_n153: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t101_n155;
  t101_n154: return 0.08538952469825745f;
  t101_n155: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t101_n159;
  t101_n156: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t101_n158;
  t101_n157: return 0.006288611795753241f;
  t101_n158: return 0.030628327280282974f;
  t101_n159: return -0.01855795830488205f;
  t101_n160: return 0.03964176028966904f;
  t101_n161: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t101_n189;
  t101_n162: if (E.getNumReferences() >= 3265200128 /*79.5*/) goto t101_n184;
  t101_n163: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t101_n181;
  t101_n164: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t101_n166;
  t101_n165: return 0.010136722587049007f;
  t101_n166: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t101_n178;
  t101_n167: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t101_n177;
  t101_n168: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t101_n176;
  t101_n169: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t101_n171;
  t101_n170: return 0.041477449238300323f;
  t101_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t101_n173;
  t101_n172: return 0.15086637437343597f;
  t101_n173: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t101_n175;
  t101_n174: return 0.1080087423324585f;
  t101_n175: return -0.011953279376029968f;
  t101_n176: return 0.05904840677976608f;
  t101_n177: return 0.03160075098276138f;
  t101_n178: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t101_n180;
  t101_n179: return -0.5440248847007751f;
  t101_n180: return 0.04400142282247543f;
  t101_n181: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t101_n183;
  t101_n182: return -0.9101483225822449f;
  t101_n183: return -0.011285699903964996f;
  t101_n184: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t101_n188;
  t101_n185: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t101_n187;
  t101_n186: return 0.08244926482439041f;
  t101_n187: return 0.13809196650981903f;
  t101_n188: return 0.054996557533741f;
  t101_n189: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t101_n191;
  t101_n190: return 0.04806533828377724f;
  t101_n191: return -0.025148194283246994f;
  t101_n192: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t101_n200;
  t101_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t101_n197;
  t101_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t101_n196;
  t101_n195: return -0.16440638899803162f;
  t101_n196: return -0.0221351757645607f;
  t101_n197: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t101_n199;
  t101_n198: return -0.012766079977154732f;
  t101_n199: return 0.0784851536154747f;
  t101_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t101_n206;
  t101_n201: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t101_n205;
  t101_n202: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t101_n204;
  t101_n203: return -0.0007188949966803193f;
  t101_n204: return -0.09688608348369598f;
  t101_n205: return 0.04625699669122696f;
  t101_n206: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t101_n208;
  t101_n207: return 0.01843268610537052f;
  t101_n208: return 0.04462238773703575f;
  t101_n209: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t101_n215;
  t101_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t101_n212;
  t101_n211: return 0.021213239058852196f;
  t101_n212: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t101_n214;
  t101_n213: return -0.1378934234380722f;
  t101_n214: return 0.01290257927030325f;
  t101_n215: return 0.05578085407614708f;
  t101_n216: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t101_n254;
  t101_n217: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t101_n253;
  t101_n218: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t101_n230;
  t101_n219: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t101_n223;
  t101_n220: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t101_n222;
  t101_n221: return 0.003155966755002737f;
  t101_n222: return -0.5345289707183838f;
  t101_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t101_n227;
  t101_n224: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t101_n226;
  t101_n225: return 0.02027585543692112f;
  t101_n226: return 0.1294293850660324f;
  t101_n227: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t101_n229;
  t101_n228: return 0.042854249477386475f;
  t101_n229: return 0.12762829661369324f;
  t101_n230: if (E.getNumReferences() >= 3315875840 /*5256.0*/) goto t101_n252;
  t101_n231: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t101_n245;
  t101_n232: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t101_n242;
  t101_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t101_n235;
  t101_n234: return -0.06343138962984085f;
  t101_n235: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t101_n237;
  t101_n236: return 0.03788582235574722f;
  t101_n237: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t101_n241;
  t101_n238: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t101_n240;
  t101_n239: return 0.10931340605020523f;
  t101_n240: return 0.11584204435348511f;
  t101_n241: return -0.041494280099868774f;
  t101_n242: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t101_n244;
  t101_n243: return 0.009120991453528404f;
  t101_n244: return 0.03376396745443344f;
  t101_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t101_n251;
  t101_n246: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t101_n248;
  t101_n247: return -0.21818727254867554f;
  t101_n248: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t101_n250;
  t101_n249: return 0.032844554632902145f;
  t101_n250: return 0.11310575902462006f;
  t101_n251: return -0.020992334932088852f;
  t101_n252: return 0.08173222094774246f;
  t101_n253: return -0.037225715816020966f;
  t101_n254: return -0.11775253713130951f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree102(const Example& E) {
  t102_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t102_n184;
  t102_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t102_n89;
  t102_n2: if (E.getNumReferences() >= 3325315072 /*11538.0*/) goto t102_n34;
  t102_n3: if (E.getNumReferences() >= 3324264448 /*10512.0*/) goto t102_n33;
  t102_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t102_n30;
  t102_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t102_n27;
  t102_n6: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t102_n22;
  t102_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n9;
  t102_n8: return -0.005237276665866375f;
  t102_n9: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t102_n13;
  t102_n10: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t102_n12;
  t102_n11: return 0.1434841752052307f;
  t102_n12: return 0.0016473681898787618f;
  t102_n13: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t102_n19;
  t102_n14: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t102_n16;
  t102_n15: return 0.03229350224137306f;
  t102_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t102_n18;
  t102_n17: return -0.23303759098052979f;
  t102_n18: return 0.13567525148391724f;
  t102_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t102_n21;
  t102_n20: return -0.08300478011369705f;
  t102_n21: return -0.012994172982871532f;
  t102_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n26;
  t102_n23: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t102_n25;
  t102_n24: return -0.02547421306371689f;
  t102_n25: return 0.22142647206783295f;
  t102_n26: return 0.008111448027193546f;
  t102_n27: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t102_n29;
  t102_n28: return 0.007153610698878765f;
  t102_n29: return -0.00570911169052124f;
  t102_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t102_n32;
  t102_n31: return -0.17842306196689606f;
  t102_n32: return -0.034582991153001785f;
  t102_n33: return -0.0602964386343956f;
  t102_n34: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t102_n60;
  t102_n35: if (E.getNumReferences() >= 3327478272 /*13650.5*/) goto t102_n45;
  t102_n36: if (E.getNumReferences() >= 3327457280 /*13630.0*/) goto t102_n42;
  t102_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t102_n39;
  t102_n38: return 0.06011873483657837f;
  t102_n39: if (E.getNumReferences() >= 3325893632 /*12103.0*/) goto t102_n41;
  t102_n40: return 0.10146581381559372f;
  t102_n41: return 0.05283303186297417f;
  t102_n42: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t102_n44;
  t102_n43: return 0.3506263196468353f;
  t102_n44: return 0.04106186702847481f;
  t102_n45: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t102_n53;
  t102_n46: if (E.getNumReferences() >= 3328655872 /*14800.5*/) goto t102_n48;
  t102_n47: return -0.08303817361593246f;
  t102_n48: if (E.getNumReferences() >= 3328821248 /*14962.0*/) goto t102_n52;
  t102_n49: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t102_n51;
  t102_n50: return 0.09034867584705353f;
  t102_n51: return 0.19405195116996765f;
  t102_n52: return 0.020085761323571205f;
  t102_n53: if (E.getNumReferences() >= 3333053952 /*21807.0*/) goto t102_n55;
  t102_n54: return -0.07059618830680847f;
  t102_n55: if (E.getNumReferences() >= 3333079040 /*21856.0*/) goto t102_n57;
  t102_n56: return 0.14911870658397675f;
  t102_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t102_n59;
  t102_n58: return -0.05253211781382561f;
  t102_n59: return 0.011132213287055492f;
  t102_n60: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t102_n62;
  t102_n61: return -0.054273754358291626f;
  t102_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t102_n64;
  t102_n63: return 0.002409586915746331f;
  t102_n64: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t102_n86;
  t102_n65: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t102_n69;
  t102_n66: if (E.getNumReferences() >= 3365435344 /*312254.5*/) goto t102_n68;
  t102_n67: return 0.006990643218159676f;
  t102_n68: return 0.05337650701403618f;
  t102_n69: if (E.getNumReferences() >= 3333346816 /*22379.0*/) goto t102_n79;
  t102_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t102_n76;
  t102_n71: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t102_n75;
  t102_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n74;
  t102_n73: return -0.13188661634922028f;
  t102_n74: return 0.16251713037490845f;
  t102_n75: return -0.07761536538600922f;
  t102_n76: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t102_n78;
  t102_n77: return 0.11750603467226028f;
  t102_n78: return 0.1744450181722641f;
  t102_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n81;
  t102_n80: return -0.1014116033911705f;
  t102_n81: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t102_n85;
  t102_n82: if (E.getNumReferences() >= 3340838912 /*41256.0*/) goto t102_n84;
  t102_n83: return 0.03142766281962395f;
  t102_n84: return 0.12301843613386154f;
  t102_n85: return -0.027674561366438866f;
  t102_n86: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t102_n88;
  t102_n87: return -0.03977631777524948f;
  t102_n88: return 0.10175154358148575f;
  t102_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t102_n107;
  t102_n90: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t102_n96;
  t102_n91: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t102_n93;
  t102_n92: return -0.002703191014006734f;
  t102_n93: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t102_n95;
  t102_n94: return -0.011970192193984985f;
  t102_n95: return -0.06130187213420868f;
  t102_n96: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t102_n106;
  t102_n97: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t102_n105;
  t102_n98: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t102_n102;
  t102_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t102_n101;
  t102_n100: return -0.012162487022578716f;
  t102_n101: return 0.011822287924587727f;
  t102_n102: if (E.getNumReferences() >= 3299520512 /*1365.5*/) goto t102_n104;
  t102_n103: return 0.07202711701393127f;
  t102_n104: return 0.028081992641091347f;
  t102_n105: return -0.012034536339342594f;
  t102_n106: return 0.13922296464443207f;
  t102_n107: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t102_n111;
  t102_n108: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t102_n110;
  t102_n109: return 0.0016730029601603746f;
  t102_n110: return 0.009448036551475525f;
  t102_n111: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t102_n177;
  t102_n112: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t102_n172;
  t102_n113: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t102_n145;
  t102_n114: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t102_n118;
  t102_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t102_n117;
  t102_n116: return -0.006015770137310028f;
  t102_n117: return 0.030998272821307182f;
  t102_n118: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t102_n126;
  t102_n119: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t102_n121;
  t102_n120: return 0.006034792400896549f;
  t102_n121: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t102_n123;
  t102_n122: return 0.03508313372731209f;
  t102_n123: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t102_n125;
  t102_n124: return 0.24915441870689392f;
  t102_n125: return 0.27237066626548767f;
  t102_n126: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t102_n138;
  t102_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t102_n131;
  t102_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t102_n130;
  t102_n129: return -0.09119199961423874f;
  t102_n130: return 0.06325513869524002f;
  t102_n131: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t102_n137;
  t102_n132: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t102_n136;
  t102_n133: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t102_n135;
  t102_n134: return 0.1405186504125595f;
  t102_n135: return 0.14748407900333405f;
  t102_n136: return 0.06969595700502396f;
  t102_n137: return 0.07928015291690826f;
  t102_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t102_n140;
  t102_n139: return -0.08145684748888016f;
  t102_n140: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t102_n142;
  t102_n141: return 0.02641754038631916f;
  t102_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t102_n144;
  t102_n143: return 0.08043230324983597f;
  t102_n144: return 0.2016856074333191f;
  t102_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement))) goto t102_n157;
  t102_n146: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t102_n156;
  t102_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t102_n151;
  t102_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n150;
  t102_n149: return -0.09998320043087006f;
  t102_n150: return 0.012237191200256348f;
  t102_n151: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t102_n153;
  t102_n152: return 0.10056024044752121f;
  t102_n153: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t102_n155;
  t102_n154: return 0.01119908131659031f;
  t102_n155: return 0.05371599271893501f;
  t102_n156: return 0.14649087190628052f;
  t102_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t102_n159;
  t102_n158: return -0.20365940034389496f;
  t102_n159: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t102_n161;
  t102_n160: return 0.018910054117441177f;
  t102_n161: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t102_n165;
  t102_n162: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t102_n164;
  t102_n163: return 0.038303520530462265f;
  t102_n164: return 0.11923179775476456f;
  t102_n165: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t102_n169;
  t102_n166: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t102_n168;
  t102_n167: return 0.10053116083145142f;
  t102_n168: return -2.3144984245300293f;
  t102_n169: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t102_n171;
  t102_n170: return 0.1431606262922287f;
  t102_n171: return 0.07105457037687302f;
  t102_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t102_n176;
  t102_n173: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t102_n175;
  t102_n174: return 0.002178802154958248f;
  t102_n175: return 0.13678152859210968f;
  t102_n176: return 0.034821417182683945f;
  t102_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t102_n181;
  t102_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t102_n180;
  t102_n179: return -0.009111599065363407f;
  t102_n180: return 0.02105584554374218f;
  t102_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n183;
  t102_n182: return -0.2447758913040161f;
  t102_n183: return 0.005072144791483879f;
  t102_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t102_n202;
  t102_n185: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t102_n199;
  t102_n186: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t102_n194;
  t102_n187: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t102_n191;
  t102_n188: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t102_n190;
  t102_n189: return 0.024701440706849098f;
  t102_n190: return -0.0014736552257090807f;
  t102_n191: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t102_n193;
  t102_n192: return 0.1242145523428917f;
  t102_n193: return 0.05500654876232147f;
  t102_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t102_n196;
  t102_n195: return -0.034770745784044266f;
  t102_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t102_n198;
  t102_n197: return -0.1406310796737671f;
  t102_n198: return 0.002376747550442815f;
  t102_n199: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t102_n201;
  t102_n200: return 0.01843707263469696f;
  t102_n201: return 0.05402519926428795f;
  t102_n202: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t102_n254;
  t102_n203: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t102_n217;
  t102_n204: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t102_n212;
  t102_n205: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t102_n211;
  t102_n206: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t102_n208;
  t102_n207: return 0.032370708882808685f;
  t102_n208: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t102_n210;
  t102_n209: return 0.10899682343006134f;
  t102_n210: return 0.020256130024790764f;
  t102_n211: return -1.7258145817322657e-05f;
  t102_n212: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t102_n216;
  t102_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t102_n215;
  t102_n214: return 0.01899629272520542f;
  t102_n215: return 0.0383763425052166f;
  t102_n216: return 0.1330050379037857f;
  t102_n217: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t102_n229;
  t102_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t102_n226;
  t102_n219: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t102_n223;
  t102_n220: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t102_n222;
  t102_n221: return -0.14523102343082428f;
  t102_n222: return 0.015755832195281982f;
  t102_n223: if (E.getFractionNameInContext() >= 3188638956 /*0.13942307233810425*/) goto t102_n225;
  t102_n224: return 0.04870747774839401f;
  t102_n225: return 0.12053646892309189f;
  t102_n226: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t102_n228;
  t102_n227: return 0.04011359065771103f;
  t102_n228: return 0.09806733578443527f;
  t102_n229: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t102_n241;
  t102_n230: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t102_n240;
  t102_n231: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t102_n239;
  t102_n232: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t102_n234;
  t102_n233: return -0.015945937484502792f;
  t102_n234: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t102_n236;
  t102_n235: return 0.03370105102658272f;
  t102_n236: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t102_n238;
  t102_n237: return 0.06500061601400375f;
  t102_n238: return 0.1482083648443222f;
  t102_n239: return -0.05463426187634468f;
  t102_n240: return 0.01128912903368473f;
  t102_n241: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t102_n253;
  t102_n242: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t102_n244;
  t102_n243: return 0.03208881989121437f;
  t102_n244: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t102_n246;
  t102_n245: return 0.10339941829442978f;
  t102_n246: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t102_n252;
  t102_n247: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t102_n249;
  t102_n248: return 0.0313291922211647f;
  t102_n249: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t102_n251;
  t102_n250: return 0.07460319250822067f;
  t102_n251: return 0.09418966621160507f;
  t102_n252: return -0.2695453464984894f;
  t102_n253: return 0.012250352650880814f;
  t102_n254: return -0.11742560565471649f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree103(const Example& E) {
  t103_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t103_n220;
  t103_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t103_n219;
  t103_n2: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t103_n152;
  t103_n3: if (E.getNumReferences() >= 3338797696 /*33282.5*/) goto t103_n151;
  t103_n4: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t103_n88;
  t103_n5: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t103_n55;
  t103_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t103_n38;
  t103_n7: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t103_n9;
  t103_n8: return -0.0007164083071984351f;
  t103_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n37;
  t103_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t103_n20;
  t103_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n13;
  t103_n12: return -0.10133307427167892f;
  t103_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n15;
  t103_n14: return -0.01566147990524769f;
  t103_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n17;
  t103_n16: return -0.007949457503855228f;
  t103_n17: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t103_n19;
  t103_n18: return 0.09923777729272842f;
  t103_n19: return 0.016979290172457695f;
  t103_n20: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t103_n34;
  t103_n21: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t103_n33;
  t103_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t103_n24;
  t103_n23: return 0.022902701050043106f;
  t103_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t103_n26;
  t103_n25: return -0.027376526966691017f;
  t103_n26: if (E.getNumReferences() >= 3272245248 /*138.5*/) goto t103_n28;
  t103_n27: return 0.04160699248313904f;
  t103_n28: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t103_n32;
  t103_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t103_n31;
  t103_n30: return -0.1025783121585846f;
  t103_n31: return 0.147847980260849f;
  t103_n32: return 0.10422220081090927f;
  t103_n33: return 0.04655415937304497f;
  t103_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n36;
  t103_n35: return -0.010627935640513897f;
  t103_n36: return 0.032648857682943344f;
  t103_n37: return 0.01897953636944294f;
  t103_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t103_n54;
  t103_n39: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t103_n53;
  t103_n40: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t103_n48;
  t103_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n43;
  t103_n42: return -0.09066368639469147f;
  t103_n43: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t103_n47;
  t103_n44: if (E.getNumReferences() >= 3281928192 /*316.5*/) goto t103_n46;
  t103_n45: return 0.01012533437460661f;
  t103_n46: return 0.10454218834638596f;
  t103_n47: return -0.04143953323364258f;
  t103_n48: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t103_n52;
  t103_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n51;
  t103_n50: return -0.06939012557268143f;
  t103_n51: return 0.07266426831483841f;
  t103_n52: return 0.007215029560029507f;
  t103_n53: return 0.027208779007196426f;
  t103_n54: return 0.030032021924853325f;
  t103_n55: if (E.getNumReferences() >= 3336999680 /*29513.5*/) goto t103_n81;
  t103_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t103_n80;
  t103_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n63;
  t103_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t103_n60;
  t103_n59: return -0.05860449746251106f;
  t103_n60: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t103_n62;
  t103_n61: return -0.004337509162724018f;
  t103_n62: return 0.02199523337185383f;
  t103_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n77;
  t103_n64: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t103_n66;
  t103_n65: return 0.03559262305498123f;
  t103_n66: if (E.getNumReferences() >= 3317257216 /*5930.5*/) goto t103_n74;
  t103_n67: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t103_n71;
  t103_n68: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t103_n70;
  t103_n69: return -0.08507218956947327f;
  t103_n70: return -0.31024354696273804f;
  t103_n71: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t103_n73;
  t103_n72: return -0.02213924378156662f;
  t103_n73: return -0.3015333414077759f;
  t103_n74: if (E.getNumReferences() >= 3322536448 /*8824.5*/) goto t103_n76;
  t103_n75: return 0.13271211087703705f;
  t103_n76: return -0.0023790965788066387f;
  t103_n77: if (E.getNumReferences() >= 3289833472 /*603.5*/) goto t103_n79;
  t103_n78: return 0.012560545466840267f;
  t103_n79: return 0.04856371134519577f;
  t103_n80: return 0.031385235488414764f;
  t103_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t103_n85;
  t103_n82: if (E.getNumReferences() >= 3338617856 /*32674.0*/) goto t103_n84;
  t103_n83: return 0.12029379606246948f;
  t103_n84: return -0.24529050290584564f;
  t103_n85: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t103_n87;
  t103_n86: return 0.12056175619363785f;
  t103_n87: return 0.02353524975478649f;
  t103_n88: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t103_n150;
  t103_n89: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t103_n135;
  t103_n90: if (E.getNumReferences() >= 3303165952 /*1810.5*/) goto t103_n126;
  t103_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t103_n93;
  t103_n92: return -0.0055610500276088715f;
  t103_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t103_n99;
  t103_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t103_n96;
  t103_n95: return -0.008997456170618534f;
  t103_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n98;
  t103_n97: return -0.04151296615600586f;
  t103_n98: return 0.007157231215387583f;
  t103_n99: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t103_n115;
  t103_n100: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t103_n106;
  t103_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n105;
  t103_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n104;
  t103_n103: return -0.0754447728395462f;
  t103_n104: return 0.030834905803203583f;
  t103_n105: return -0.002083604456856847f;
  t103_n106: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t103_n114;
  t103_n107: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t103_n109;
  t103_n108: return 0.0065895747393369675f;
  t103_n109: if (E.getNumReferences() >= 3261857792 /*59.0*/) goto t103_n113;
  t103_n110: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t103_n112;
  t103_n111: return 0.15477317571640015f;
  t103_n112: return 0.4047374427318573f;
  t103_n113: return -0.08820430934429169f;
  t103_n114: return 0.0025730340275913477f;
  t103_n115: if (E.getNumReferences() >= 3297341440 /*1099.5*/) goto t103_n121;
  t103_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t103_n118;
  t103_n117: return -0.02078280970454216f;
  t103_n118: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t103_n120;
  t103_n119: return 0.05083552375435829f;
  t103_n120: return 0.1269753873348236f;
  t103_n121: if (E.getNumReferences() >= 3297697792 /*1143.0*/) goto t103_n125;
  t103_n122: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t103_n124;
  t103_n123: return 0.08685962855815887f;
  t103_n124: return 0.17058230936527252f;
  t103_n125: return 0.08097898960113525f;
  t103_n126: if (E.getNumReferences() >= 3303182336 /*1812.5*/) goto t103_n130;
  t103_n127: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t103_n129;
  t103_n128: return 0.22751101851463318f;
  t103_n129: return 0.0875716507434845f;
  t103_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t103_n132;
  t103_n131: return -0.002249696757644415f;
  t103_n132: if (E.getNumReferences() >= 3337447424 /*30388.0*/) goto t103_n134;
  t103_n133: return 0.014345051720738411f;
  t103_n134: return -0.05180966481566429f;
  t103_n135: if (E.getNumReferences() >= 3318942720 /*6753.5*/) goto t103_n145;
  t103_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t103_n140;
  t103_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t103_n139;
  t103_n138: return -0.07154784351587296f;
  t103_n139: return 0.001537512755021453f;
  t103_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t103_n144;
  t103_n141: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t103_n143;
  t103_n142: return -0.094384104013443f;
  t103_n143: return 0.014496893621981144f;
  t103_n144: return 0.020955881103873253f;
  t103_n145: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t103_n147;
  t103_n146: return -0.2766907811164856f;
  t103_n147: if (E.getNumReferences() >= 3320151040 /*7343.5*/) goto t103_n149;
  t103_n148: return 0.09909486025571823f;
  t103_n149: return -0.029836267232894897f;
  t103_n150: return -0.010434195399284363f;
  t103_n151: return -0.14495708048343658f;
  t103_n152: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t103_n158;
  t103_n153: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t103_n157;
  t103_n154: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t103_n156;
  t103_n155: return 0.05650724470615387f;
  t103_n156: return 0.12918904423713684f;
  t103_n157: return 0.02233750931918621f;
  t103_n158: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t103_n160;
  t103_n159: return -0.03388098627328873f;
  t103_n160: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t103_n176;
  t103_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n163;
  t103_n162: return -0.010710105299949646f;
  t103_n163: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t103_n165;
  t103_n164: return 0.008200312964618206f;
  t103_n165: if (E.getNumReferences() >= 3341362816 /*43302.5*/) goto t103_n169;
  t103_n166: if (E.getNumReferences() >= 3339196032 /*34838.5*/) goto t103_n168;
  t103_n167: return 0.0719035416841507f;
  t103_n168: return -0.02866765856742859f;
  t103_n169: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t103_n175;
  t103_n170: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t103_n172;
  t103_n171: return 0.004206627607345581f;
  t103_n172: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t103_n174;
  t103_n173: return 0.12760493159294128f;
  t103_n174: return 0.011043976061046124f;
  t103_n175: return 0.06235511973500252f;
  t103_n176: if (E.getNumReferences() >= 3343039104 /*49850.5*/) goto t103_n178;
  t103_n177: return -0.11065971851348877f;
  t103_n178: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t103_n192;
  t103_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t103_n181;
  t103_n180: return -0.016766704618930817f;
  t103_n181: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t103_n189;
  t103_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t103_n184;
  t103_n183: return 0.00021931776427663863f;
  t103_n184: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t103_n188;
  t103_n185: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t103_n187;
  t103_n186: return 0.15508303046226501f;
  t103_n187: return 0.1444264054298401f;
  t103_n188: return -0.1077931597828865f;
  t103_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n191;
  t103_n190: return -0.12211994081735611f;
  t103_n191: return 0.13712617754936218f;
  t103_n192: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t103_n204;
  t103_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n203;
  t103_n194: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t103_n196;
  t103_n195: return 0.08206233382225037f;
  t103_n196: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t103_n202;
  t103_n197: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t103_n201;
  t103_n198: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t103_n200;
  t103_n199: return -0.24001890420913696f;
  t103_n200: return -1.6263939142227173f;
  t103_n201: return -0.01686592772603035f;
  t103_n202: return 0.03028150089085102f;
  t103_n203: return -0.029390716925263405f;
  t103_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n206;
  t103_n205: return -0.0030754199251532555f;
  t103_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t103_n212;
  t103_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t103_n211;
  t103_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n210;
  t103_n209: return -0.13337238132953644f;
  t103_n210: return 0.10526464134454727f;
  t103_n211: return -0.10991929471492767f;
  t103_n212: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t103_n216;
  t103_n213: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t103_n215;
  t103_n214: return 0.023529931902885437f;
  t103_n215: return 0.05365116521716118f;
  t103_n216: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t103_n218;
  t103_n217: return -0.13372902572155f;
  t103_n218: return 0.01802639104425907f;
  t103_n219: return 0.006503457203507423f;
  t103_n220: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t103_n234;
  t103_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t103_n233;
  t103_n222: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t103_n232;
  t103_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n227;
  t103_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t103_n226;
  t103_n225: return -0.044963061809539795f;
  t103_n226: return -0.010446294210851192f;
  t103_n227: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t103_n231;
  t103_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t103_n230;
  t103_n229: return 0.006064692512154579f;
  t103_n230: return 0.0637378916144371f;
  t103_n231: return -0.07555222511291504f;
  t103_n232: return 0.013259323313832283f;
  t103_n233: return 0.055792130529880524f;
  t103_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n238;
  t103_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t103_n237;
  t103_n236: return -0.029323119670152664f;
  t103_n237: return 0.0021787865553051233f;
  t103_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t103_n252;
  t103_n239: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t103_n241;
  t103_n240: return -0.06916818767786026f;
  t103_n241: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t103_n243;
  t103_n242: return -0.02725161798298359f;
  t103_n243: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t103_n247;
  t103_n244: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t103_n246;
  t103_n245: return -0.05072987452149391f;
  t103_n246: return 0.05766996741294861f;
  t103_n247: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t103_n251;
  t103_n248: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t103_n250;
  t103_n249: return 0.10702459514141083f;
  t103_n250: return 0.03930184990167618f;
  t103_n251: return 0.06268864125013351f;
  t103_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t103_n254;
  t103_n253: return 0.014272834174335003f;
  t103_n254: return 0.03836056962609291f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree104(const Example& E) {
  t104_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t104_n212;
  t104_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t104_n67;
  t104_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t104_n8;
  t104_n3: if (E.getNumReferences() >= 3313733632 /*4210.0*/) goto t104_n7;
  t104_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t104_n6;
  t104_n5: return -0.0006570644327439368f;
  t104_n6: return 0.007673335261642933f;
  t104_n7: return 0.014548463746905327f;
  t104_n8: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t104_n64;
  t104_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t104_n11;
  t104_n10: return -0.0057435710914433f;
  t104_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t104_n31;
  t104_n12: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t104_n14;
  t104_n13: return -0.00916411355137825f;
  t104_n14: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t104_n30;
  t104_n15: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t104_n29;
  t104_n16: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t104_n28;
  t104_n17: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t104_n19;
  t104_n18: return 0.00082089111674577f;
  t104_n19: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t104_n21;
  t104_n20: return 0.054053209722042084f;
  t104_n21: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t104_n23;
  t104_n22: return -0.03896297514438629f;
  t104_n23: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t104_n25;
  t104_n24: return 0.05372988432645798f;
  t104_n25: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t104_n27;
  t104_n26: return -0.060820601880550385f;
  t104_n27: return 0.05037027969956398f;
  t104_n28: return -0.08384273946285248f;
  t104_n29: return 0.06695590168237686f;
  t104_n30: return -0.019849885255098343f;
  t104_n31: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t104_n55;
  t104_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n36;
  t104_n33: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t104_n35;
  t104_n34: return 0.016627812758088112f;
  t104_n35: return -0.01897479221224785f;
  t104_n36: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t104_n52;
  t104_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t104_n45;
  t104_n38: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t104_n44;
  t104_n39: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t104_n41;
  t104_n40: return -0.020569484680891037f;
  t104_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t104_n43;
  t104_n42: return -0.062363509088754654f;
  t104_n43: return 0.024834193289279938f;
  t104_n44: return -0.018072133883833885f;
  t104_n45: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t104_n49;
  t104_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n48;
  t104_n47: return -0.1535986065864563f;
  t104_n48: return 0.038211870938539505f;
  t104_n49: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t104_n51;
  t104_n50: return 0.04398593679070473f;
  t104_n51: return 0.01814485341310501f;
  t104_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n54;
  t104_n53: return -0.12210673838853836f;
  t104_n54: return -0.018161363899707794f;
  t104_n55: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t104_n57;
  t104_n56: return 0.007357367780059576f;
  t104_n57: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t104_n63;
  t104_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t104_n62;
  t104_n59: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t104_n61;
  t104_n60: return 0.03874421864748001f;
  t104_n61: return -0.006177866365760565f;
  t104_n62: return 0.07143484055995941f;
  t104_n63: return -0.06656037271022797f;
  t104_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t104_n66;
  t104_n65: return -0.033182911574840546f;
  t104_n66: return -0.007410347927361727f;
  t104_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t104_n69;
  t104_n68: return -0.1042509451508522f;
  t104_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t104_n97;
  t104_n70: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t104_n90;
  t104_n71: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t104_n87;
  t104_n72: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t104_n76;
  t104_n73: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t104_n75;
  t104_n74: return -0.004065978806465864f;
  t104_n75: return -0.014872802421450615f;
  t104_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t104_n84;
  t104_n77: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t104_n79;
  t104_n78: return 0.00015599436301272362f;
  t104_n79: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t104_n81;
  t104_n80: return 0.015297961421310902f;
  t104_n81: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t104_n83;
  t104_n82: return 0.19386328756809235f;
  t104_n83: return 0.03670255094766617f;
  t104_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t104_n86;
  t104_n85: return 0.023701002821326256f;
  t104_n86: return -0.025825142860412598f;
  t104_n87: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t104_n89;
  t104_n88: return -0.01210570428520441f;
  t104_n89: return -0.03674231842160225f;
  t104_n90: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t104_n96;
  t104_n91: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t104_n95;
  t104_n92: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t104_n94;
  t104_n93: return 0.01235543005168438f;
  t104_n94: return 0.03389647230505943f;
  t104_n95: return -0.005439247936010361f;
  t104_n96: return -0.010276935994625092f;
  t104_n97: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t104_n177;
  t104_n98: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t104_n134;
  t104_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t104_n101;
  t104_n100: return 0.003843475366011262f;
  t104_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t104_n103;
  t104_n102: return -0.02838926576077938f;
  t104_n103: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t104_n127;
  t104_n104: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t104_n108;
  t104_n105: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t104_n107;
  t104_n106: return 0.12824171781539917f;
  t104_n107: return 0.031114405021071434f;
  t104_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t104_n114;
  t104_n109: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t104_n113;
  t104_n110: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t104_n112;
  t104_n111: return 0.15641094744205475f;
  t104_n112: return 0.44026005268096924f;
  t104_n113: return 0.03452230617403984f;
  t104_n114: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t104_n116;
  t104_n115: return 0.07176932692527771f;
  t104_n116: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t104_n124;
  t104_n117: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t104_n121;
  t104_n118: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t104_n120;
  t104_n119: return 0.15348608791828156f;
  t104_n120: return 0.16495126485824585f;
  t104_n121: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t104_n123;
  t104_n122: return -0.1863253116607666f;
  t104_n123: return 0.29089441895484924f;
  t104_n124: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t104_n126;
  t104_n125: return 0.3576142489910126f;
  t104_n126: return 0.24342550337314606f;
  t104_n127: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t104_n131;
  t104_n128: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t104_n130;
  t104_n129: return -0.0347432866692543f;
  t104_n130: return 0.15683665871620178f;
  t104_n131: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t104_n133;
  t104_n132: return 0.052786216139793396f;
  t104_n133: return 0.010174715891480446f;
  t104_n134: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t104_n174;
  t104_n135: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t104_n137;
  t104_n136: return -0.03003944829106331f;
  t104_n137: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t104_n171;
  t104_n138: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t104_n144;
  t104_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n141;
  t104_n140: return -0.08537530899047852f;
  t104_n141: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t104_n143;
  t104_n142: return 0.0030817040242254734f;
  t104_n143: return -0.017166618257761f;
  t104_n144: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t104_n168;
  t104_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t104_n147;
  t104_n146: return 0.025838138535618782f;
  t104_n147: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t104_n149;
  t104_n148: return 0.03814978152513504f;
  t104_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t104_n151;
  t104_n150: return -0.07054924219846725f;
  t104_n151: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t104_n165;
  t104_n152: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t104_n164;
  t104_n153: if (E.getNumReferences() >= 3256090624 /*37.0*/) goto t104_n159;
  t104_n154: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t104_n158;
  t104_n155: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t104_n157;
  t104_n156: return 0.13996782898902893f;
  t104_n157: return 0.01003751065582037f;
  t104_n158: return 0.15103167295455933f;
  t104_n159: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t104_n161;
  t104_n160: return 0.08927620947360992f;
  t104_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t104_n163;
  t104_n162: return 0.18562179803848267f;
  t104_n163: return 0.5151102542877197f;
  t104_n164: return 0.1982344686985016f;
  t104_n165: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t104_n167;
  t104_n166: return 0.07297158241271973f;
  t104_n167: return 0.2347271740436554f;
  t104_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t104_n170;
  t104_n169: return 0.015325590968132019f;
  t104_n170: return 0.036018967628479004f;
  t104_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t104_n173;
  t104_n172: return -0.009367240592837334f;
  t104_n173: return 0.02636527456343174f;
  t104_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n176;
  t104_n175: return -0.09529998153448105f;
  t104_n176: return 0.07706666737794876f;
  t104_n177: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t104_n205;
  t104_n178: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t104_n202;
  t104_n179: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t104_n193;
  t104_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n182;
  t104_n181: return -0.015726298093795776f;
  t104_n182: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t104_n184;
  t104_n183: return 0.0030438106041401625f;
  t104_n184: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t104_n192;
  t104_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t104_n191;
  t104_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t104_n188;
  t104_n187: return 0.017917128279805183f;
  t104_n188: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t104_n190;
  t104_n189: return 0.11298755556344986f;
  t104_n190: return 0.35354313254356384f;
  t104_n191: return 0.05930068716406822f;
  t104_n192: return -0.017165595665574074f;
  t104_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement))) goto t104_n195;
  t104_n194: return 0.013513179495930672f;
  t104_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t104_n197;
  t104_n196: return -0.19003801047801971f;
  t104_n197: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t104_n199;
  t104_n198: return 0.01670554280281067f;
  t104_n199: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t104_n201;
  t104_n200: return 0.03643299266695976f;
  t104_n201: return 0.09534335136413574f;
  t104_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t104_n204;
  t104_n203: return 0.001986064715310931f;
  t104_n204: return 0.03257580101490021f;
  t104_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t104_n211;
  t104_n206: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t104_n210;
  t104_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t104_n209;
  t104_n208: return -0.010416646488010883f;
  t104_n209: return 0.018825281411409378f;
  t104_n210: return 0.05583038181066513f;
  t104_n211: return 0.008993654511868954f;
  t104_n212: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t104_n250;
  t104_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t104_n237;
  t104_n214: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t104_n234;
  t104_n215: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t104_n221;
  t104_n216: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t104_n218;
  t104_n217: return 0.005880700424313545f;
  t104_n218: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t104_n220;
  t104_n219: return 0.033347237855196f;
  t104_n220: return -0.0036092414520680904f;
  t104_n221: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t104_n229;
  t104_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t104_n226;
  t104_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t104_n225;
  t104_n224: return -0.15902364253997803f;
  t104_n225: return -0.01860998198390007f;
  t104_n226: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t104_n228;
  t104_n227: return -0.01985032483935356f;
  t104_n228: return 0.07328388839960098f;
  t104_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t104_n233;
  t104_n230: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t104_n232;
  t104_n231: return -0.0012294433545321226f;
  t104_n232: return 0.04134339094161987f;
  t104_n233: return 0.024899156764149666f;
  t104_n234: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t104_n236;
  t104_n235: return 0.017447026446461678f;
  t104_n236: return 0.048906322568655014f;
  t104_n237: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t104_n249;
  t104_n238: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t104_n240;
  t104_n239: return 0.006045724265277386f;
  t104_n240: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t104_n246;
  t104_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t104_n243;
  t104_n242: return -0.06571466475725174f;
  t104_n243: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t104_n245;
  t104_n244: return 0.036022838205099106f;
  t104_n245: return 0.09446413069963455f;
  t104_n246: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t104_n248;
  t104_n247: return 0.007869083434343338f;
  t104_n248: return 0.02664043940603733f;
  t104_n249: return -0.1016312912106514f;
  t104_n250: if (E.getNumReferences() >= 3352414592 /*107411.0*/) goto t104_n252;
  t104_n251: return 0.09406065195798874f;
  t104_n252: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t104_n254;
  t104_n253: return -0.1608031988143921f;
  t104_n254: return 0.06551875919103622f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree105(const Example& E) {
  t105_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t105_n228;
  t105_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t105_n85;
  t105_n2: if (E.getNumReferences() >= 3302502400 /*1729.5*/) goto t105_n68;
  t105_n3: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t105_n67;
  t105_n4: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t105_n42;
  t105_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t105_n11;
  t105_n6: if (E.getNumReferences() >= 3302428672 /*1720.5*/) goto t105_n10;
  t105_n7: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t105_n9;
  t105_n8: return 0.14018437266349792f;
  t105_n9: return -0.0033430575858801603f;
  t105_n10: return -0.18607711791992188f;
  t105_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t105_n13;
  t105_n12: return -0.06383644789457321f;
  t105_n13: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t105_n25;
  t105_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t105_n20;
  t105_n15: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t105_n19;
  t105_n16: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t105_n18;
  t105_n17: return 0.05809086188673973f;
  t105_n18: return 0.13017335534095764f;
  t105_n19: return 0.0005100931157357991f;
  t105_n20: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t105_n24;
  t105_n21: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t105_n23;
  t105_n22: return 0.023463604971766472f;
  t105_n23: return -0.1596931666135788f;
  t105_n24: return 0.08147115260362625f;
  t105_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n33;
  t105_n26: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t105_n28;
  t105_n27: return -0.08016212284564972f;
  t105_n28: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t105_n32;
  t105_n29: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n31;
  t105_n30: return 0.0009517524158582091f;
  t105_n31: return 0.03694695979356766f;
  t105_n32: return -0.11140325665473938f;
  t105_n33: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t105_n37;
  t105_n34: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n36;
  t105_n35: return 0.02520575001835823f;
  t105_n36: return 0.033578358590602875f;
  t105_n37: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t105_n39;
  t105_n38: return -0.004745318088680506f;
  t105_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t105_n41;
  t105_n40: return 0.011068829335272312f;
  t105_n41: return 0.060214973986148834f;
  t105_n42: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t105_n64;
  t105_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n49;
  t105_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t105_n46;
  t105_n45: return -0.02645988017320633f;
  t105_n46: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t105_n48;
  t105_n47: return 0.026681384071707726f;
  t105_n48: return -0.14460881054401398f;
  t105_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t105_n51;
  t105_n50: return -0.017378447577357292f;
  t105_n51: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t105_n55;
  t105_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t105_n54;
  t105_n53: return 0.0036630043759942055f;
  t105_n54: return -0.062206804752349854f;
  t105_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t105_n57;
  t105_n56: return 0.024492856115102768f;
  t105_n57: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t105_n59;
  t105_n58: return 0.044786352664232254f;
  t105_n59: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t105_n61;
  t105_n60: return -0.08935634791851044f;
  t105_n61: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n63;
  t105_n62: return -0.06619405746459961f;
  t105_n63: return 0.032817985862493515f;
  t105_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n66;
  t105_n65: return -0.01282538939267397f;
  t105_n66: return 0.003516799071803689f;
  t105_n67: return 0.000648310873657465f;
  t105_n68: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t105_n74;
  t105_n69: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t105_n73;
  t105_n70: if (E.getNumReferences() >= 3302559744 /*1736.5*/) goto t105_n72;
  t105_n71: return 0.12378621846437454f;
  t105_n72: return 0.009889931418001652f;
  t105_n73: return 0.0763811469078064f;
  t105_n74: if (E.getNumReferences() >= 3302584320 /*1739.5*/) goto t105_n84;
  t105_n75: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t105_n79;
  t105_n76: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t105_n78;
  t105_n77: return -0.055990803986787796f;
  t105_n78: return 0.23532702028751373f;
  t105_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t105_n81;
  t105_n80: return -0.01952563039958477f;
  t105_n81: if (E.getNumReferences() >= 3302551552 /*1735.5*/) goto t105_n83;
  t105_n82: return -0.17452742159366608f;
  t105_n83: return 0.12928348779678345f;
  t105_n84: return 0.0012801876291632652f;
  t105_n85: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t105_n227;
  t105_n86: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t105_n224;
  t105_n87: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t105_n219;
  t105_n88: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t105_n218;
  t105_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t105_n141;
  t105_n90: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t105_n128;
  t105_n91: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t105_n125;
  t105_n92: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t105_n96;
  t105_n93: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t105_n95;
  t105_n94: return -0.004054568242281675f;
  t105_n95: return -0.014033220708370209f;
  t105_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t105_n112;
  t105_n97: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t105_n99;
  t105_n98: return 0.0001386706717312336f;
  t105_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t105_n109;
  t105_n100: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t105_n106;
  t105_n101: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t105_n105;
  t105_n102: if (E.getFractionNameInContext() >= 3197626751 /*0.2967033088207245*/) goto t105_n104;
  t105_n103: return 0.04524790123105049f;
  t105_n104: return -0.025339456275105476f;
  t105_n105: return 0.031758278608322144f;
  t105_n106: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t105_n108;
  t105_n107: return -0.02705065906047821f;
  t105_n108: return -0.3118859827518463f;
  t105_n109: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t105_n111;
  t105_n110: return 0.024542473256587982f;
  t105_n111: return 0.13874761760234833f;
  t105_n112: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t105_n124;
  t105_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t105_n123;
  t105_n114: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t105_n118;
  t105_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t105_n117;
  t105_n116: return 0.016902897506952286f;
  t105_n117: return 0.04147854074835777f;
  t105_n118: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t105_n120;
  t105_n119: return 0.04369184747338295f;
  t105_n120: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t105_n122;
  t105_n121: return 0.16183307766914368f;
  t105_n122: return 0.17862142622470856f;
  t105_n123: return -0.016426945105195045f;
  t105_n124: return -0.02266625128686428f;
  t105_n125: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t105_n127;
  t105_n126: return -0.009908699430525303f;
  t105_n127: return -0.03329924866557121f;
  t105_n128: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t105_n140;
  t105_n129: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n139;
  t105_n130: if (E.getNumReferences() >= 3314605056 /*4635.5*/) goto t105_n134;
  t105_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t105_n133;
  t105_n132: return -0.005703335627913475f;
  t105_n133: return 0.012126389890909195f;
  t105_n134: if (E.getNumReferences() >= 3315159040 /*4906.0*/) goto t105_n138;
  t105_n135: if (E.getNumReferences() >= 3315124224 /*4889.0*/) goto t105_n137;
  t105_n136: return 0.08423493057489395f;
  t105_n137: return 0.1512250155210495f;
  t105_n138: return 0.02702256105840206f;
  t105_n139: return -0.014583667740225792f;
  t105_n140: return 0.13161011040210724f;
  t105_n141: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t105_n191;
  t105_n142: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t105_n190;
  t105_n143: if (E.getNumReferences() >= 3357701696 /*166361.0*/) goto t105_n189;
  t105_n144: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t105_n146;
  t105_n145: return 0.005527477245777845f;
  t105_n146: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t105_n160;
  t105_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n151;
  t105_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t105_n150;
  t105_n149: return -0.10188250243663788f;
  t105_n150: return -0.0890144556760788f;
  t105_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n159;
  t105_n152: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t105_n156;
  t105_n153: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n155;
  t105_n154: return 0.002713269554078579f;
  t105_n155: return -0.04812955856323242f;
  t105_n156: if (E.getFractionNameInContext() >= 3206291208 /*0.6098484992980957*/) goto t105_n158;
  t105_n157: return -0.013058673590421677f;
  t105_n158: return 0.11465717852115631f;
  t105_n159: return 0.004522939678281546f;
  t105_n160: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t105_n178;
  t105_n161: if (E.getNumReferences() >= 3269525504 /*112.5*/) goto t105_n165;
  t105_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t105_n164;
  t105_n163: return 0.005268112290650606f;
  t105_n164: return -0.03345625102519989f;
  t105_n165: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t105_n177;
  t105_n166: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t105_n172;
  t105_n167: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t105_n171;
  t105_n168: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t105_n170;
  t105_n169: return 0.16880419850349426f;
  t105_n170: return 0.237132266163826f;
  t105_n171: return -0.1016468033194542f;
  t105_n172: if (E.getNumReferences() >= 3301007360 /*1547.0*/) goto t105_n174;
  t105_n173: return 0.08258265256881714f;
  t105_n174: if (E.getNumReferences() >= 3303546880 /*1857.0*/) goto t105_n176;
  t105_n175: return 0.13049174845218658f;
  t105_n176: return 0.08707670867443085f;
  t105_n177: return 0.018196510151028633f;
  t105_n178: if (E.getFractionNameInContext() >= 3211159142 /*0.8999999761581421*/) goto t105_n184;
  t105_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t105_n183;
  t105_n180: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t105_n182;
  t105_n181: return 0.04484878480434418f;
  t105_n182: return 0.0033676265738904476f;
  t105_n183: return -0.016919391229748726f;
  t105_n184: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t105_n186;
  t105_n185: return 0.0930338129401207f;
  t105_n186: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t105_n188;
  t105_n187: return 0.13571864366531372f;
  t105_n188: return 0.16789238154888153f;
  t105_n189: return -0.3431384563446045f;
  t105_n190: return 0.08120985329151154f;
  t105_n191: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t105_n199;
  t105_n192: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t105_n196;
  t105_n193: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t105_n195;
  t105_n194: return 0.01142480131238699f;
  t105_n195: return 0.11791151762008667f;
  t105_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t105_n198;
  t105_n197: return 0.0005178184364922345f;
  t105_n198: return 0.029510440304875374f;
  t105_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Variable))) goto t105_n201;
  t105_n200: return -0.002676095347851515f;
  t105_n201: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t105_n211;
  t105_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t105_n210;
  t105_n203: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t105_n209;
  t105_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t105_n206;
  t105_n205: return -0.15513741970062256f;
  t105_n206: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t105_n208;
  t105_n207: return 0.02961326576769352f;
  t105_n208: return 0.06691025197505951f;
  t105_n209: return 0.14481455087661743f;
  t105_n210: return -0.008767632767558098f;
  t105_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t105_n217;
  t105_n212: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t105_n214;
  t105_n213: return -0.004565402399748564f;
  t105_n214: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t105_n216;
  t105_n215: return 0.023015303537249565f;
  t105_n216: return -0.01583467796444893f;
  t105_n217: return 0.016650177538394928f;
  t105_n218: return -0.17039409279823303f;
  t105_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t105_n223;
  t105_n220: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t105_n222;
  t105_n221: return 0.09862381964921951f;
  t105_n222: return 0.006676640827208757f;
  t105_n223: return 0.10047262161970139f;
  t105_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t105_n226;
  t105_n225: return -0.13560175895690918f;
  t105_n226: return 0.017277469858527184f;
  t105_n227: return 0.034781564027071f;
  t105_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t105_n252;
  t105_n229: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t105_n247;
  t105_n230: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t105_n246;
  t105_n231: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t105_n235;
  t105_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t105_n234;
  t105_n233: return 0.019382836297154427f;
  t105_n234: return 0.0675024464726448f;
  t105_n235: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t105_n243;
  t105_n236: if (E.getNumReferences() >= 3289595904 /*589.0*/) goto t105_n242;
  t105_n237: if (E.getNumReferences() >= 3287793664 /*495.5*/) goto t105_n241;
  t105_n238: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t105_n240;
  t105_n239: return -0.01179727166891098f;
  t105_n240: return 0.02446347288787365f;
  t105_n241: return 0.12736201286315918f;
  t105_n242: return -0.4147593677043915f;
  t105_n243: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t105_n245;
  t105_n244: return 0.12325403839349747f;
  t105_n245: return 0.03683220222592354f;
  t105_n246: return 0.0006564109935425222f;
  t105_n247: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t105_n251;
  t105_n248: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t105_n250;
  t105_n249: return 0.01851683296263218f;
  t105_n250: return -0.026006262749433517f;
  t105_n251: return 0.05012624338269234f;
  t105_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t105_n254;
  t105_n253: return 0.010587597265839577f;
  t105_n254: return -0.10607317835092545f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree106(const Example& E) {
  t106_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t106_n246;
  t106_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t106_n123;
  t106_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t106_n114;
  t106_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n57;
  t106_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t106_n38;
  t106_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t106_n35;
  t106_n6: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t106_n28;
  t106_n7: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t106_n19;
  t106_n8: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t106_n18;
  t106_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t106_n17;
  t106_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t106_n12;
  t106_n11: return 0.0008496815571561456f;
  t106_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t106_n16;
  t106_n13: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t106_n15;
  t106_n14: return 0.0008007544674910605f;
  t106_n15: return 0.05386703833937645f;
  t106_n16: return 0.03236154839396477f;
  t106_n17: return -0.0028639461379498243f;
  t106_n18: return -0.07411293685436249f;
  t106_n19: if (E.getNumReferences() >= 3334339584 /*24318.0*/) goto t106_n23;
  t106_n20: if (E.getNumReferences() >= 3292127232 /*743.5*/) goto t106_n22;
  t106_n21: return 0.011972370557487011f;
  t106_n22: return -0.07305188477039337f;
  t106_n23: if (E.getNumReferences() >= 3364353728 /*278454.0*/) goto t106_n25;
  t106_n24: return 0.050437554717063904f;
  t106_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n27;
  t106_n26: return -0.04606360197067261f;
  t106_n27: return 0.08662454783916473f;
  t106_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t106_n30;
  t106_n29: return -0.03655668720602989f;
  t106_n30: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t106_n32;
  t106_n31: return 0.03872811794281006f;
  t106_n32: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t106_n34;
  t106_n33: return 0.003569579916074872f;
  t106_n34: return 0.0577877014875412f;
  t106_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n37;
  t106_n36: return -0.03646465390920639f;
  t106_n37: return -0.018153652548789978f;
  t106_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t106_n44;
  t106_n39: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t106_n41;
  t106_n40: return 0.0004949116846546531f;
  t106_n41: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t106_n43;
  t106_n42: return 0.03326010704040527f;
  t106_n43: return -0.01705021969974041f;
  t106_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t106_n50;
  t106_n45: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t106_n47;
  t106_n46: return 0.016562560573220253f;
  t106_n47: if (E.getNumReferences() >= 3364365712 /*278828.5*/) goto t106_n49;
  t106_n48: return 0.09224867075681686f;
  t106_n49: return 0.02480519749224186f;
  t106_n50: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t106_n52;
  t106_n51: return 0.017566651105880737f;
  t106_n52: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t106_n54;
  t106_n53: return 0.02514495886862278f;
  t106_n54: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t106_n56;
  t106_n55: return 0.07666753977537155f;
  t106_n56: return -0.10240250080823898f;
  t106_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t106_n59;
  t106_n58: return -0.11575208604335785f;
  t106_n59: if (E.getNumReferences() >= 3322147840 /*8445.0*/) goto t106_n113;
  t106_n60: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t106_n98;
  t106_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t106_n77;
  t106_n62: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t106_n64;
  t106_n63: return -0.008579765446484089f;
  t106_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t106_n66;
  t106_n65: return -0.10228163748979568f;
  t106_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n68;
  t106_n67: return -0.049973152577877045f;
  t106_n68: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t106_n74;
  t106_n69: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t106_n73;
  t106_n70: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t106_n72;
  t106_n71: return 0.01744016259908676f;
  t106_n72: return 0.06269548088312149f;
  t106_n73: return -0.0010800492018461227f;
  t106_n74: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t106_n76;
  t106_n75: return 0.021836446598172188f;
  t106_n76: return -0.0025554706808179617f;
  t106_n77: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t106_n85;
  t106_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t106_n82;
  t106_n79: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t106_n81;
  t106_n80: return 0.009688322432339191f;
  t106_n81: return 0.022865135222673416f;
  t106_n82: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t106_n84;
  t106_n83: return 0.0010893727885559201f;
  t106_n84: return 0.016009269282221794f;
  t106_n85: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t106_n95;
  t106_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t106_n94;
  t106_n87: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t106_n89;
  t106_n88: return -0.1289084106683731f;
  t106_n89: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t106_n91;
  t106_n90: return -0.052150096744298935f;
  t106_n91: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t106_n93;
  t106_n92: return 0.0540149062871933f;
  t106_n93: return 0.10625941306352615f;
  t106_n94: return 0.02881530113518238f;
  t106_n95: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t106_n97;
  t106_n96: return -0.018555963411927223f;
  t106_n97: return 0.13047417998313904f;
  t106_n98: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t106_n112;
  t106_n99: if (E.getNumReferences() >= 3275259904 /*184.5*/) goto t106_n103;
  t106_n100: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t106_n102;
  t106_n101: return 0.013768617063760757f;
  t106_n102: return -0.014367845840752125f;
  t106_n103: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t106_n105;
  t106_n104: return 0.04969869554042816f;
  t106_n105: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t106_n111;
  t106_n106: if (E.getNumReferences() >= 3321282560 /*7896.0*/) goto t106_n108;
  t106_n107: return 0.03772655501961708f;
  t106_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n110;
  t106_n109: return -0.22184127569198608f;
  t106_n110: return 0.12730515003204346f;
  t106_n111: return -0.003060691524296999f;
  t106_n112: return -0.033927105367183685f;
  t106_n113: return 0.022927651181817055f;
  t106_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t106_n116;
  t106_n115: return -0.04938668757677078f;
  t106_n116: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t106_n118;
  t106_n117: return 0.0005672489642165601f;
  t106_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t106_n120;
  t106_n119: return -0.016735335811972618f;
  t106_n120: if (E.getNumReferences() >= 3329239040 /*15370.0*/) goto t106_n122;
  t106_n121: return -0.0013564075343310833f;
  t106_n122: return 0.07524527609348297f;
  t106_n123: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t106_n125;
  t106_n124: return 0.001249150955118239f;
  t106_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t106_n201;
  t106_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t106_n198;
  t106_n127: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t106_n197;
  t106_n128: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t106_n180;
  t106_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t106_n173;
  t106_n130: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t106_n168;
  t106_n131: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t106_n133;
  t106_n132: return -0.0026878025382757187f;
  t106_n133: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t106_n135;
  t106_n134: return -0.00985756702721119f;
  t106_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t106_n167;
  t106_n136: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t106_n138;
  t106_n137: return 0.025604333728551865f;
  t106_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t106_n164;
  t106_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n161;
  t106_n140: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t106_n160;
  t106_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t106_n145;
  t106_n142: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t106_n144;
  t106_n143: return -0.2870461046695709f;
  t106_n144: return 0.0263997670263052f;
  t106_n145: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t106_n159;
  t106_n146: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t106_n154;
  t106_n147: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t106_n153;
  t106_n148: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t106_n152;
  t106_n149: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t106_n151;
  t106_n150: return -0.07954902946949005f;
  t106_n151: return 0.10431447625160217f;
  t106_n152: return 0.021594082936644554f;
  t106_n153: return 0.02536378987133503f;
  t106_n154: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t106_n156;
  t106_n155: return -0.033847626298666f;
  t106_n156: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t106_n158;
  t106_n157: return -0.11246966570615768f;
  t106_n158: return 0.07861234992742538f;
  t106_n159: return -0.10028315335512161f;
  t106_n160: return -0.02890835888683796f;
  t106_n161: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t106_n163;
  t106_n162: return -0.10093548893928528f;
  t106_n163: return 0.07478407770395279f;
  t106_n164: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t106_n166;
  t106_n165: return -0.06793887168169022f;
  t106_n166: return 0.06674191355705261f;
  t106_n167: return -0.10739414393901825f;
  t106_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t106_n170;
  t106_n169: return 0.010688421316444874f;
  t106_n170: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t106_n172;
  t106_n171: return 0.09130840748548508f;
  t106_n172: return 0.030606146901845932f;
  t106_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t106_n175;
  t106_n174: return -0.01294037513434887f;
  t106_n175: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t106_n177;
  t106_n176: return -0.0027728762943297625f;
  t106_n177: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t106_n179;
  t106_n178: return 0.09162057936191559f;
  t106_n179: return -0.1220201700925827f;
  t106_n180: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t106_n196;
  t106_n181: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t106_n189;
  t106_n182: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t106_n186;
  t106_n183: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t106_n185;
  t106_n184: return 0.08808452636003494f;
  t106_n185: return 0.001999557251110673f;
  t106_n186: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t106_n188;
  t106_n187: return 0.00270060240291059f;
  t106_n188: return -0.04994339123368263f;
  t106_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t106_n195;
  t106_n190: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t106_n194;
  t106_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n193;
  t106_n192: return -0.02796102687716484f;
  t106_n193: return -0.0044983201660215855f;
  t106_n194: return 0.0035699422005563974f;
  t106_n195: return 0.07653890550136566f;
  t106_n196: return 0.014895040541887283f;
  t106_n197: return 0.008647706359624863f;
  t106_n198: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t106_n200;
  t106_n199: return 0.02577994018793106f;
  t106_n200: return 0.07665345072746277f;
  t106_n201: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t106_n241;
  t106_n202: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t106_n240;
  t106_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t106_n239;
  t106_n204: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t106_n208;
  t106_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n207;
  t106_n206: return -0.07376847416162491f;
  t106_n207: return 0.003228484420105815f;
  t106_n208: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t106_n218;
  t106_n209: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t106_n217;
  t106_n210: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t106_n214;
  t106_n211: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t106_n213;
  t106_n212: return -0.13349956274032593f;
  t106_n213: return -0.00999013427644968f;
  t106_n214: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t106_n216;
  t106_n215: return 0.45125794410705566f;
  t106_n216: return -0.08328159153461456f;
  t106_n217: return 0.02276334911584854f;
  t106_n218: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t106_n236;
  t106_n219: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t106_n221;
  t106_n220: return 0.0250605046749115f;
  t106_n221: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t106_n235;
  t106_n222: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t106_n232;
  t106_n223: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t106_n229;
  t106_n224: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t106_n226;
  t106_n225: return -0.14590668678283691f;
  t106_n226: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t106_n228;
  t106_n227: return 0.004937760066241026f;
  t106_n228: return 0.09171440452337265f;
  t106_n229: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t106_n231;
  t106_n230: return 0.04448496177792549f;
  t106_n231: return 0.11076552420854568f;
  t106_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t106_n234;
  t106_n233: return -0.04055386781692505f;
  t106_n234: return 0.04892435669898987f;
  t106_n235: return 0.09088490903377533f;
  t106_n236: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t106_n238;
  t106_n237: return 0.03831087797880173f;
  t106_n238: return 0.1414455771446228f;
  t106_n239: return 0.013296288438141346f;
  t106_n240: return -0.003300473326817155f;
  t106_n241: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t106_n245;
  t106_n242: if (E.getNumReferences() >= 3349388864 /*83772.5*/) goto t106_n244;
  t106_n243: return -0.007507277652621269f;
  t106_n244: return 0.04661192744970322f;
  t106_n245: return 0.009939833544194698f;
  t106_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t106_n252;
  t106_n247: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t106_n251;
  t106_n248: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t106_n250;
  t106_n249: return -0.13871534168720245f;
  t106_n250: return 0.002383654471486807f;
  t106_n251: return 0.01700861193239689f;
  t106_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t106_n254;
  t106_n253: return 0.010494677349925041f;
  t106_n254: return -0.10417904704809189f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree107(const Example& E) {
  t107_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t107_n90;
  t107_n1: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t107_n81;
  t107_n2: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t107_n74;
  t107_n3: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t107_n45;
  t107_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t107_n6;
  t107_n5: return -0.02774161286652088f;
  t107_n6: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t107_n44;
  t107_n7: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t107_n27;
  t107_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t107_n12;
  t107_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t107_n11;
  t107_n10: return -0.011664016172289848f;
  t107_n11: return 0.006482105236500502f;
  t107_n12: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t107_n26;
  t107_n13: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t107_n17;
  t107_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t107_n16;
  t107_n15: return -0.014753763563930988f;
  t107_n16: return 0.033569540828466415f;
  t107_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t107_n21;
  t107_n18: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t107_n20;
  t107_n19: return 0.009347033686935902f;
  t107_n20: return 0.09594632685184479f;
  t107_n21: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t107_n25;
  t107_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t107_n24;
  t107_n23: return -0.22158195078372955f;
  t107_n24: return 0.19492214918136597f;
  t107_n25: return 0.05527687445282936f;
  t107_n26: return 0.05187229812145233f;
  t107_n27: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t107_n43;
  t107_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t107_n42;
  t107_n29: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t107_n41;
  t107_n30: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t107_n32;
  t107_n31: return -0.004317150451242924f;
  t107_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t107_n40;
  t107_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t107_n35;
  t107_n34: return 0.00010239481343887746f;
  t107_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n37;
  t107_n36: return -0.08763545006513596f;
  t107_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t107_n39;
  t107_n38: return 0.020406249910593033f;
  t107_n39: return 0.03545013815164566f;
  t107_n40: return 0.007881822995841503f;
  t107_n41: return -0.006080781575292349f;
  t107_n42: return -0.07997789978981018f;
  t107_n43: return -0.06593925505876541f;
  t107_n44: return -0.05673381686210632f;
  t107_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n65;
  t107_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t107_n56;
  t107_n47: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t107_n49;
  t107_n48: return -0.06151343509554863f;
  t107_n49: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t107_n53;
  t107_n50: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t107_n52;
  t107_n51: return 0.0033576434943825006f;
  t107_n52: return 0.033470865339040756f;
  t107_n53: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t107_n55;
  t107_n54: return -0.04189135506749153f;
  t107_n55: return -0.010733111761510372f;
  t107_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n58;
  t107_n57: return -0.14049264788627625f;
  t107_n58: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t107_n60;
  t107_n59: return 0.06606739014387131f;
  t107_n60: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t107_n64;
  t107_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t107_n63;
  t107_n62: return 0.1157316267490387f;
  t107_n63: return 0.16600050032138824f;
  t107_n64: return 0.0895526334643364f;
  t107_n65: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t107_n67;
  t107_n66: return 0.03163856640458107f;
  t107_n67: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t107_n69;
  t107_n68: return -0.04416752979159355f;
  t107_n69: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t107_n71;
  t107_n70: return 0.010219857096672058f;
  t107_n71: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t107_n73;
  t107_n72: return 0.0735868439078331f;
  t107_n73: return 0.12811262905597687f;
  t107_n74: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t107_n80;
  t107_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n77;
  t107_n76: return -0.023856118321418762f;
  t107_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t107_n79;
  t107_n78: return 0.010515226051211357f;
  t107_n79: return 0.036411453038454056f;
  t107_n80: return 1.5782765331096016e-05f;
  t107_n81: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t107_n83;
  t107_n82: return 0.0002954349620267749f;
  t107_n83: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t107_n89;
  t107_n84: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t107_n86;
  t107_n85: return 0.012889022007584572f;
  t107_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n88;
  t107_n87: return -0.051673464477062225f;
  t107_n88: return 0.04841994121670723f;
  t107_n89: return -0.020550696179270744f;
  t107_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t107_n236;
  t107_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n223;
  t107_n92: if (E.getNumReferences() >= 3319913472 /*7227.5*/) goto t107_n220;
  t107_n93: if (E.getNumReferences() >= 3319290880 /*6923.5*/) goto t107_n219;
  t107_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n142;
  t107_n95: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t107_n141;
  t107_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t107_n108;
  t107_n97: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t107_n105;
  t107_n98: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n102;
  t107_n99: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t107_n101;
  t107_n100: return -0.012669957242906094f;
  t107_n101: return -0.15567825734615326f;
  t107_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n104;
  t107_n103: return -0.1063784807920456f;
  t107_n104: return 0.021190516650676727f;
  t107_n105: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t107_n107;
  t107_n106: return -0.04563822224736214f;
  t107_n107: return 0.01720043085515499f;
  t107_n108: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t107_n120;
  t107_n109: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t107_n119;
  t107_n110: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t107_n114;
  t107_n111: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t107_n113;
  t107_n112: return -0.009661892428994179f;
  t107_n113: return 0.12685610353946686f;
  t107_n114: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t107_n116;
  t107_n115: return -0.0690465047955513f;
  t107_n116: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t107_n118;
  t107_n117: return 0.11933333426713943f;
  t107_n118: return -0.23723511397838593f;
  t107_n119: return 0.03143477067351341f;
  t107_n120: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t107_n136;
  t107_n121: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t107_n133;
  t107_n122: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t107_n132;
  t107_n123: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t107_n125;
  t107_n124: return 0.010569934733211994f;
  t107_n125: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n131;
  t107_n126: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t107_n130;
  t107_n127: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t107_n129;
  t107_n128: return 0.08209819346666336f;
  t107_n129: return 0.11721231788396835f;
  t107_n130: return 0.013453890569508076f;
  t107_n131: return -0.055464860051870346f;
  t107_n132: return -0.09213649481534958f;
  t107_n133: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n135;
  t107_n134: return 0.0799606516957283f;
  t107_n135: return -0.11444742232561111f;
  t107_n136: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t107_n140;
  t107_n137: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t107_n139;
  t107_n138: return 0.011241345666348934f;
  t107_n139: return 0.03686467930674553f;
  t107_n140: return -0.08619311451911926f;
  t107_n141: return -0.058842167258262634f;
  t107_n142: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t107_n218;
  t107_n143: if (E.getNumReferences() >= 3318440960 /*6508.5*/) goto t107_n213;
  t107_n144: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t107_n212;
  t107_n145: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t107_n151;
  t107_n146: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t107_n148;
  t107_n147: return -0.0040702600963413715f;
  t107_n148: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t107_n150;
  t107_n149: return 0.01728646270930767f;
  t107_n150: return 0.11361652612686157f;
  t107_n151: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t107_n153;
  t107_n152: return 0.019226698204874992f;
  t107_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t107_n183;
  t107_n154: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t107_n160;
  t107_n155: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t107_n159;
  t107_n156: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t107_n158;
  t107_n157: return 0.0048158117569983006f;
  t107_n158: return -0.027345309033989906f;
  t107_n159: return 0.04295137897133827f;
  t107_n160: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t107_n180;
  t107_n161: if (E.getNumReferences() >= 3290292224 /*631.5*/) goto t107_n165;
  t107_n162: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n164;
  t107_n163: return -0.011683549731969833f;
  t107_n164: return 0.009777765721082687f;
  t107_n165: if (E.getNumReferences() >= 3305728000 /*2198.5*/) goto t107_n179;
  t107_n166: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t107_n174;
  t107_n167: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n173;
  t107_n168: if (E.getNumReferences() >= 3295109120 /*925.5*/) goto t107_n172;
  t107_n169: if (E.getNumReferences() >= 3294535680 /*890.5*/) goto t107_n171;
  t107_n170: return 0.010594361461699009f;
  t107_n171: return 0.12189789861440659f;
  t107_n172: return 0.08158987015485764f;
  t107_n173: return -0.0025620588567107916f;
  t107_n174: if (E.getNumReferences() >= 3291062272 /*678.5*/) goto t107_n178;
  t107_n175: if (E.getNumReferences() >= 3290898432 /*668.5*/) goto t107_n177;
  t107_n176: return 0.10289348661899567f;
  t107_n177: return 0.14098012447357178f;
  t107_n178: return 0.0550684854388237f;
  t107_n179: return 0.000882403866853565f;
  t107_n180: if (E.getNumReferences() >= 3314368512 /*4520.0*/) goto t107_n182;
  t107_n181: return 0.040227826684713364f;
  t107_n182: return -0.479594886302948f;
  t107_n183: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t107_n187;
  t107_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t107_n186;
  t107_n185: return -0.1558103859424591f;
  t107_n186: return 0.010071251541376114f;
  t107_n187: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n211;
  t107_n188: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t107_n208;
  t107_n189: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t107_n203;
  t107_n190: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t107_n200;
  t107_n191: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t107_n195;
  t107_n192: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t107_n194;
  t107_n193: return 0.04956824332475662f;
  t107_n194: return 0.11981700360774994f;
  t107_n195: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t107_n197;
  t107_n196: return 0.14560110867023468f;
  t107_n197: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t107_n199;
  t107_n198: return -0.028809309005737305f;
  t107_n199: return 0.10623329877853394f;
  t107_n200: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t107_n202;
  t107_n201: return -0.2730390131473541f;
  t107_n202: return 0.11671760678291321f;
  t107_n203: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t107_n205;
  t107_n204: return -0.23769129812717438f;
  t107_n205: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t107_n207;
  t107_n206: return 0.13995669782161713f;
  t107_n207: return 0.0272601880133152f;
  t107_n208: if (E.getNumReferences() >= 3258056704 /*44.5*/) goto t107_n210;
  t107_n209: return -0.05060233548283577f;
  t107_n210: return 0.052727118134498596f;
  t107_n211: return -0.022678233683109283f;
  t107_n212: return 0.017390115186572075f;
  t107_n213: if (E.getNumReferences() >= 3318799360 /*6683.5*/) goto t107_n217;
  t107_n214: if (E.getNumReferences() >= 3318701056 /*6635.5*/) goto t107_n216;
  t107_n215: return 0.07753836363554001f;
  t107_n216: return 0.12691755592823029f;
  t107_n217: return -0.025763271376490593f;
  t107_n218: return 0.0003209246788173914f;
  t107_n219: return -0.2163248509168625f;
  t107_n220: if (E.getNumReferences() >= 3322916352 /*9195.5*/) goto t107_n222;
  t107_n221: return 0.07792152464389801f;
  t107_n222: return 0.016521744430065155f;
  t107_n223: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t107_n235;
  t107_n224: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t107_n228;
  t107_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n227;
  t107_n226: return -0.16396211087703705f;
  t107_n227: return 0.051444850862026215f;
  t107_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t107_n234;
  t107_n229: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t107_n233;
  t107_n230: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t107_n232;
  t107_n231: return 0.05020829290151596f;
  t107_n232: return -0.035559188574552536f;
  t107_n233: return 0.04855404794216156f;
  t107_n234: return -0.013737502507865429f;
  t107_n235: return -0.184476837515831f;
  t107_n236: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t107_n242;
  t107_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t107_n239;
  t107_n238: return -0.005741122178733349f;
  t107_n239: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t107_n241;
  t107_n240: return 0.005655296146869659f;
  t107_n241: return -0.061741288751363754f;
  t107_n242: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t107_n254;
  t107_n243: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t107_n251;
  t107_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t107_n246;
  t107_n245: return 0.00703436229377985f;
  t107_n246: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t107_n250;
  t107_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t107_n249;
  t107_n248: return 0.021747535094618797f;
  t107_n249: return 0.08137889206409454f;
  t107_n250: return -0.0716225728392601f;
  t107_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t107_n253;
  t107_n252: return -0.0360398106276989f;
  t107_n253: return 0.008902299217879772f;
  t107_n254: return 0.023760279640555382f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree108(const Example& E) {
  t108_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t108_n232;
  t108_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t108_n67;
  t108_n2: if (E.getNumReferences() >= 3300470784 /*1481.5*/) goto t108_n66;
  t108_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t108_n33;
  t108_n4: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t108_n32;
  t108_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t108_n27;
  t108_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t108_n22;
  t108_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t108_n11;
  t108_n8: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t108_n10;
  t108_n9: return -0.005388978868722916f;
  t108_n10: return 0.06082634627819061f;
  t108_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t108_n21;
  t108_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n14;
  t108_n13: return -0.1642722636461258f;
  t108_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t108_n16;
  t108_n15: return 0.036114390939474106f;
  t108_n16: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t108_n18;
  t108_n17: return -0.06580280512571335f;
  t108_n18: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t108_n20;
  t108_n19: return 0.024649078026413918f;
  t108_n20: return 0.44147035479545593f;
  t108_n21: return 0.014001207426190376f;
  t108_n22: if (E.getNumReferences() >= 3281715200 /*310.0*/) goto t108_n24;
  t108_n23: return -0.011792570352554321f;
  t108_n24: if (E.getNumReferences() >= 3282468864 /*333.0*/) goto t108_n26;
  t108_n25: return 0.13065968453884125f;
  t108_n26: return 0.06732656061649323f;
  t108_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t108_n31;
  t108_n28: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t108_n30;
  t108_n29: return -0.017739811912178993f;
  t108_n30: return 0.028102219104766846f;
  t108_n31: return 0.04543454945087433f;
  t108_n32: return -0.0006980219040997326f;
  t108_n33: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t108_n59;
  t108_n34: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t108_n38;
  t108_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n37;
  t108_n36: return -0.11901484429836273f;
  t108_n37: return -0.016515251249074936f;
  t108_n38: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t108_n46;
  t108_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t108_n41;
  t108_n40: return -0.001106737065128982f;
  t108_n41: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t108_n43;
  t108_n42: return 0.00929641630500555f;
  t108_n43: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t108_n45;
  t108_n44: return 0.005710008088499308f;
  t108_n45: return 0.043072670698165894f;
  t108_n46: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t108_n52;
  t108_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n49;
  t108_n48: return -0.13584406673908234f;
  t108_n49: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t108_n51;
  t108_n50: return 0.03592971712350845f;
  t108_n51: return 0.17396041750907898f;
  t108_n52: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t108_n54;
  t108_n53: return 0.048379458487033844f;
  t108_n54: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t108_n56;
  t108_n55: return -0.1566549837589264f;
  t108_n56: if (E.getNumReferences() >= 3254255616 /*31.0*/) goto t108_n58;
  t108_n57: return 0.1290012151002884f;
  t108_n58: return 0.08502009510993958f;
  t108_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t108_n65;
  t108_n60: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t108_n64;
  t108_n61: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t108_n63;
  t108_n62: return 0.011030896566808224f;
  t108_n63: return -0.0014717400772497058f;
  t108_n64: return -0.12107828259468079f;
  t108_n65: return 0.08207952231168747f;
  t108_n66: return 0.01618322730064392f;
  t108_n67: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t108_n183;
  t108_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t108_n158;
  t108_n69: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t108_n157;
  t108_n70: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t108_n154;
  t108_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t108_n73;
  t108_n72: return -0.0042183892801404f;
  t108_n73: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t108_n87;
  t108_n74: if (E.getNumReferences() >= 3287154688 /*476.0*/) goto t108_n86;
  t108_n75: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t108_n77;
  t108_n76: return -0.05790819227695465f;
  t108_n77: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t108_n81;
  t108_n78: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t108_n80;
  t108_n79: return 0.017251357436180115f;
  t108_n80: return 0.0755322203040123f;
  t108_n81: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t108_n83;
  t108_n82: return -0.07209143787622452f;
  t108_n83: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t108_n85;
  t108_n84: return -0.7128198146820068f;
  t108_n85: return -0.32480838894844055f;
  t108_n86: return -0.13533590734004974f;
  t108_n87: if (E.getNumReferences() >= 3355687776 /*134893.5*/) goto t108_n153;
  t108_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t108_n110;
  t108_n89: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t108_n101;
  t108_n90: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t108_n94;
  t108_n91: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t108_n93;
  t108_n92: return 0.013408327475190163f;
  t108_n93: return -0.051237478852272034f;
  t108_n94: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t108_n96;
  t108_n95: return -0.04355105012655258f;
  t108_n96: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t108_n100;
  t108_n97: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t108_n99;
  t108_n98: return 0.012079327367246151f;
  t108_n99: return 0.07132600247859955f;
  t108_n100: return -0.026299333199858665f;
  t108_n101: if (E.getNumReferences() >= 3316470784 /*5546.5*/) goto t108_n105;
  t108_n102: if (E.getNumReferences() >= 3268739072 /*106.5*/) goto t108_n104;
  t108_n103: return 0.03058595582842827f;
  t108_n104: return 0.040099773555994034f;
  t108_n105: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t108_n109;
  t108_n106: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t108_n108;
  t108_n107: return -0.013282150030136108f;
  t108_n108: return 0.04540615901350975f;
  t108_n109: return -0.19866593182086945f;
  t108_n110: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t108_n140;
  t108_n111: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t108_n115;
  t108_n112: if (E.getNumReferences() >= 3344816512 /*56793.5*/) goto t108_n114;
  t108_n113: return 0.01107652485370636f;
  t108_n114: return -0.180574432015419f;
  t108_n115: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t108_n139;
  t108_n116: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t108_n130;
  t108_n117: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t108_n123;
  t108_n118: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t108_n122;
  t108_n119: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t108_n121;
  t108_n120: return 0.051337577402591705f;
  t108_n121: return -0.00801513995975256f;
  t108_n122: return 0.051478028297424316f;
  t108_n123: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t108_n127;
  t108_n124: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t108_n126;
  t108_n125: return 0.844743549823761f;
  t108_n126: return 0.11853162199258804f;
  t108_n127: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t108_n129;
  t108_n128: return 0.09003283828496933f;
  t108_n129: return 0.15918254852294922f;
  t108_n130: if (E.getNumReferences() >= 3334173440 /*23993.5*/) goto t108_n138;
  t108_n131: if (E.getNumReferences() >= 3329227264 /*15358.5*/) goto t108_n137;
  t108_n132: if (E.getNumReferences() >= 3326295040 /*12495.0*/) goto t108_n134;
  t108_n133: return 0.002963882638141513f;
  t108_n134: if (E.getNumReferences() >= 3327545344 /*13716.0*/) goto t108_n136;
  t108_n135: return 0.13055862486362457f;
  t108_n136: return -0.01551778707653284f;
  t108_n137: return -0.0996985211968422f;
  t108_n138: return 0.05375099554657936f;
  t108_n139: return -0.015488079749047756f;
  t108_n140: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t108_n152;
  t108_n141: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t108_n147;
  t108_n142: if (E.getNumReferences() >= 3346330624 /*62708.0*/) goto t108_n146;
  t108_n143: if (E.getNumReferences() >= 3345018368 /*57582.0*/) goto t108_n145;
  t108_n144: return 0.006072735879570246f;
  t108_n145: return 0.14730755984783173f;
  t108_n146: return -0.1649361103773117f;
  t108_n147: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t108_n151;
  t108_n148: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t108_n150;
  t108_n149: return 0.1164599061012268f;
  t108_n150: return 0.018820220604538918f;
  t108_n151: return -0.12379266321659088f;
  t108_n152: return 0.08577702194452286f;
  t108_n153: return -0.17787474393844604f;
  t108_n154: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t108_n156;
  t108_n155: return -0.1414077877998352f;
  t108_n156: return 0.03716345876455307f;
  t108_n157: return -0.029431771486997604f;
  t108_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement))) goto t108_n176;
  t108_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t108_n161;
  t108_n160: return -0.0017825613031163812f;
  t108_n161: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t108_n173;
  t108_n162: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t108_n170;
  t108_n163: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t108_n167;
  t108_n164: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t108_n166;
  t108_n165: return 0.002646968699991703f;
  t108_n166: return 0.03547050058841705f;
  t108_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t108_n169;
  t108_n168: return 0.006468089763075113f;
  t108_n169: return 0.05837412178516388f;
  t108_n170: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t108_n172;
  t108_n171: return -0.0647786483168602f;
  t108_n172: return 0.04347199946641922f;
  t108_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t108_n175;
  t108_n174: return -0.0029212129302322865f;
  t108_n175: return 0.05709487572312355f;
  t108_n176: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t108_n178;
  t108_n177: return -0.00959071610122919f;
  t108_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t108_n182;
  t108_n179: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t108_n181;
  t108_n180: return -0.012700064107775688f;
  t108_n181: return 0.014506499283015728f;
  t108_n182: return 0.009211575612425804f;
  t108_n183: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t108_n225;
  t108_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t108_n188;
  t108_n185: if (E.getFileProximityDistanceCost() >= 3255697408 /*35.5*/) goto t108_n187;
  t108_n186: return -0.01893048733472824f;
  t108_n187: return -0.13908398151397705f;
  t108_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t108_n206;
  t108_n189: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t108_n205;
  t108_n190: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t108_n202;
  t108_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t108_n193;
  t108_n192: return 0.018280690535902977f;
  t108_n193: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t108_n201;
  t108_n194: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t108_n198;
  t108_n195: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t108_n197;
  t108_n196: return -0.025673430413007736f;
  t108_n197: return 0.12796661257743835f;
  t108_n198: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t108_n200;
  t108_n199: return 0.1508338451385498f;
  t108_n200: return 0.0777600109577179f;
  t108_n201: return 0.02993147261440754f;
  t108_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n204;
  t108_n203: return -0.06285692006349564f;
  t108_n204: return 0.016298260539770126f;
  t108_n205: return -0.030411774292588234f;
  t108_n206: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t108_n224;
  t108_n207: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t108_n223;
  t108_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t108_n210;
  t108_n209: return 0.0663851946592331f;
  t108_n210: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t108_n214;
  t108_n211: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t108_n213;
  t108_n212: return -0.5416795015335083f;
  t108_n213: return 0.08003285527229309f;
  t108_n214: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t108_n220;
  t108_n215: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t108_n219;
  t108_n216: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t108_n218;
  t108_n217: return 0.14689266681671143f;
  t108_n218: return 0.14525499939918518f;
  t108_n219: return -0.1896539032459259f;
  t108_n220: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t108_n222;
  t108_n221: return 0.2665605843067169f;
  t108_n222: return 0.21774624288082123f;
  t108_n223: return 0.04654970020055771f;
  t108_n224: return 0.0426330603659153f;
  t108_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n227;
  t108_n226: return -0.0056077586486935616f;
  t108_n227: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t108_n229;
  t108_n228: return 0.004907046910375357f;
  t108_n229: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t108_n231;
  t108_n230: return 0.07949905097484589f;
  t108_n231: return 0.025098126381635666f;
  t108_n232: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t108_n246;
  t108_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n245;
  t108_n234: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t108_n244;
  t108_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t108_n241;
  t108_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t108_n240;
  t108_n237: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t108_n239;
  t108_n238: return -0.010335627943277359f;
  t108_n239: return -0.04126114398241043f;
  t108_n240: return 0.015292681753635406f;
  t108_n241: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t108_n243;
  t108_n242: return 0.00046336278319358826f;
  t108_n243: return -0.055100686848163605f;
  t108_n244: return 0.008682993240654469f;
  t108_n245: return 0.005084828473627567f;
  t108_n246: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t108_n250;
  t108_n247: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t108_n249;
  t108_n248: return 0.011912763118743896f;
  t108_n249: return 0.04273753985762596f;
  t108_n250: if (E.getNumReferences() >= 3303321600 /*1829.5*/) goto t108_n254;
  t108_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t108_n253;
  t108_n252: return -0.055717822164297104f;
  t108_n253: return -0.0024598175659775734f;
  t108_n254: return -4.634953498840332f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree109(const Example& E) {
  t109_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t109_n202;
  t109_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t109_n57;
  t109_n2: if (E.getNumReferences() >= 3314717696 /*4690.5*/) goto t109_n50;
  t109_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t109_n17;
  t109_n4: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t109_n14;
  t109_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t109_n9;
  t109_n6: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t109_n8;
  t109_n7: return 0.0017607779009267688f;
  t109_n8: return 0.02119499072432518f;
  t109_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t109_n13;
  t109_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t109_n12;
  t109_n11: return -0.01617603562772274f;
  t109_n12: return 0.026566995307803154f;
  t109_n13: return 0.04058045893907547f;
  t109_n14: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t109_n16;
  t109_n15: return -0.00032948266016319394f;
  t109_n16: return -0.10627840459346771f;
  t109_n17: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t109_n43;
  t109_n18: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t109_n24;
  t109_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n21;
  t109_n20: return -0.10229040682315826f;
  t109_n21: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t109_n23;
  t109_n22: return -0.011309255845844746f;
  t109_n23: return -0.09862598031759262f;
  t109_n24: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t109_n36;
  t109_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t109_n27;
  t109_n26: return -0.0013377105351537466f;
  t109_n27: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t109_n29;
  t109_n28: return 0.008762271143496037f;
  t109_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t109_n31;
  t109_n30: return 0.005314672831445932f;
  t109_n31: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t109_n35;
  t109_n32: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t109_n34;
  t109_n33: return 0.04131330922245979f;
  t109_n34: return 0.22628705203533173f;
  t109_n35: return -0.13578951358795166f;
  t109_n36: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t109_n38;
  t109_n37: return 0.029710667207837105f;
  t109_n38: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t109_n42;
  t109_n39: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t109_n41;
  t109_n40: return 0.10105780512094498f;
  t109_n41: return 0.1183251291513443f;
  t109_n42: return 0.05921067297458649f;
  t109_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t109_n49;
  t109_n44: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t109_n46;
  t109_n45: return 0.010432113893330097f;
  t109_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n48;
  t109_n47: return -0.041135773062705994f;
  t109_n48: return 0.010989969596266747f;
  t109_n49: return 0.08122789114713669f;
  t109_n50: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t109_n56;
  t109_n51: if (E.getNumReferences() >= 3314894848 /*4777.0*/) goto t109_n53;
  t109_n52: return 0.07934325933456421f;
  t109_n53: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t109_n55;
  t109_n54: return 0.15138055384159088f;
  t109_n55: return 0.24942560493946075f;
  t109_n56: return 0.016610851511359215f;
  t109_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t109_n59;
  t109_n58: return -0.109060138463974f;
  t109_n59: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t109_n163;
  t109_n60: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t109_n162;
  t109_n61: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t109_n77;
  t109_n62: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t109_n68;
  t109_n63: if (E.getNumReferences() >= 3289202688 /*565.0*/) goto t109_n65;
  t109_n64: return 0.0027583418413996696f;
  t109_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t109_n67;
  t109_n66: return 0.0629386454820633f;
  t109_n67: return 0.14890623092651367f;
  t109_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t109_n70;
  t109_n69: return -0.04326305538415909f;
  t109_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t109_n74;
  t109_n71: if (E.getNumReferences() >= 3305136128 /*2054.0*/) goto t109_n73;
  t109_n72: return 0.021732255816459656f;
  t109_n73: return 0.44890105724334717f;
  t109_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t109_n76;
  t109_n75: return 0.03412312641739845f;
  t109_n76: return 0.08141899108886719f;
  t109_n77: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t109_n79;
  t109_n78: return -0.0015172064304351807f;
  t109_n79: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t109_n85;
  t109_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t109_n84;
  t109_n81: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t109_n83;
  t109_n82: return -0.1238020732998848f;
  t109_n83: return -0.06020626053214073f;
  t109_n84: return -0.049486685544252396f;
  t109_n85: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t109_n117;
  t109_n86: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t109_n116;
  t109_n87: if (E.getNumReferences() >= 3360067360 /*203324.5*/) goto t109_n111;
  t109_n88: if (E.getSymbolScopeDistanceCost() >= 3247964160 /*19.0*/) goto t109_n100;
  t109_n89: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t109_n97;
  t109_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t109_n96;
  t109_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n95;
  t109_n92: if (E.getNumReferences() >= 3329498112 /*15623.0*/) goto t109_n94;
  t109_n93: return -0.0444360077381134f;
  t109_n94: return -0.11710644513368607f;
  t109_n95: return 0.014284453354775906f;
  t109_n96: return 0.01554173231124878f;
  t109_n97: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t109_n99;
  t109_n98: return 0.07498030364513397f;
  t109_n99: return 0.10871107131242752f;
  t109_n100: if (E.getNumReferences() >= 3355687776 /*134893.5*/) goto t109_n110;
  t109_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t109_n105;
  t109_n102: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t109_n104;
  t109_n103: return -0.0035046017728745937f;
  t109_n104: return 0.036144860088825226f;
  t109_n105: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t109_n109;
  t109_n106: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t109_n108;
  t109_n107: return 0.27495455741882324f;
  t109_n108: return 0.0590028315782547f;
  t109_n109: return 0.016398832201957703f;
  t109_n110: return -0.06508945673704147f;
  t109_n111: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t109_n115;
  t109_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n114;
  t109_n113: return -0.12918752431869507f;
  t109_n114: return 0.055010031908750534f;
  t109_n115: return 0.004769369959831238f;
  t109_n116: return -0.029609184712171555f;
  t109_n117: if (E.getNumReferences() >= 3363974896 /*266615.5*/) goto t109_n153;
  t109_n118: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t109_n152;
  t109_n119: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t109_n127;
  t109_n120: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t109_n124;
  t109_n121: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t109_n123;
  t109_n122: return 0.031777866184711456f;
  t109_n123: return 0.1664263904094696f;
  t109_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t109_n126;
  t109_n125: return 0.07853681594133377f;
  t109_n126: return 0.18543384969234467f;
  t109_n127: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t109_n135;
  t109_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t109_n134;
  t109_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n133;
  t109_n130: if (E.getNumReferences() >= 3311218688 /*3539.0*/) goto t109_n132;
  t109_n131: return -0.09747802466154099f;
  t109_n132: return -0.13944998383522034f;
  t109_n133: return 0.0182370375841856f;
  t109_n134: return -0.034811586141586304f;
  t109_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t109_n137;
  t109_n136: return -0.0048890127800405025f;
  t109_n137: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t109_n151;
  t109_n138: if (E.getNumReferences() >= 3350349056 /*91274.0*/) goto t109_n150;
  t109_n139: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t109_n147;
  t109_n140: if (E.getNumReferences() >= 3337447936 /*30389.0*/) goto t109_n146;
  t109_n141: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t109_n145;
  t109_n142: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t109_n144;
  t109_n143: return 0.02228519506752491f;
  t109_n144: return -0.095067098736763f;
  t109_n145: return 0.10299456864595413f;
  t109_n146: return -0.1463741809129715f;
  t109_n147: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t109_n149;
  t109_n148: return 0.1209503784775734f;
  t109_n149: return 0.03849272429943085f;
  t109_n150: return -0.08923058956861496f;
  t109_n151: return 0.0624992810189724f;
  t109_n152: return -0.1575719118118286f;
  t109_n153: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t109_n157;
  t109_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t109_n156;
  t109_n155: return 0.026910340413451195f;
  t109_n156: return 0.09384317696094513f;
  t109_n157: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t109_n159;
  t109_n158: return -0.2076113075017929f;
  t109_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t109_n161;
  t109_n160: return 0.028053410351276398f;
  t109_n161: return -0.037541359663009644f;
  t109_n162: return -0.13324585556983948f;
  t109_n163: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t109_n201;
  t109_n164: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t109_n200;
  t109_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t109_n187;
  t109_n166: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t109_n186;
  t109_n167: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t109_n179;
  t109_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n178;
  t109_n169: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t109_n171;
  t109_n170: return 0.04208815470337868f;
  t109_n171: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t109_n177;
  t109_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t109_n176;
  t109_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n175;
  t109_n174: return -0.12777520716190338f;
  t109_n175: return -0.02069544419646263f;
  t109_n176: return -0.0015889592468738556f;
  t109_n177: return -0.002130104461684823f;
  t109_n178: return 0.003941577859222889f;
  t109_n179: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t109_n183;
  t109_n180: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t109_n182;
  t109_n181: return -0.027428502216935158f;
  t109_n182: return 0.008817837573587894f;
  t109_n183: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t109_n185;
  t109_n184: return 0.1388411819934845f;
  t109_n185: return 0.06300324201583862f;
  t109_n186: return -0.060149721801280975f;
  t109_n187: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t109_n193;
  t109_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t109_n190;
  t109_n189: return -0.019121157005429268f;
  t109_n190: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t109_n192;
  t109_n191: return -0.038629017770290375f;
  t109_n192: return 0.01679743267595768f;
  t109_n193: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t109_n197;
  t109_n194: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t109_n196;
  t109_n195: return 0.012081571854650974f;
  t109_n196: return -0.09772136807441711f;
  t109_n197: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t109_n199;
  t109_n198: return 0.09237640351057053f;
  t109_n199: return 0.0047182063572108746f;
  t109_n200: return -0.05949893593788147f;
  t109_n201: return -0.0275914017111063f;
  t109_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t109_n242;
  t109_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t109_n241;
  t109_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t109_n238;
  t109_n205: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t109_n237;
  t109_n206: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t109_n208;
  t109_n207: return -0.010482804849743843f;
  t109_n208: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t109_n236;
  t109_n209: if (E.getFractionNameInContext() >= 3203243738 /*0.464102566242218*/) goto t109_n227;
  t109_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t109_n212;
  t109_n211: return -0.02720329351723194f;
  t109_n212: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t109_n218;
  t109_n213: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t109_n215;
  t109_n214: return 0.00010076625767396763f;
  t109_n215: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t109_n217;
  t109_n216: return -0.005765360780060291f;
  t109_n217: return 0.013147509656846523f;
  t109_n218: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t109_n222;
  t109_n219: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t109_n221;
  t109_n220: return 0.08506004512310028f;
  t109_n221: return -0.02563495561480522f;
  t109_n222: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t109_n224;
  t109_n223: return -0.013924078084528446f;
  t109_n224: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t109_n226;
  t109_n225: return 0.09406585246324539f;
  t109_n226: return -0.011903461068868637f;
  t109_n227: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t109_n231;
  t109_n228: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t109_n230;
  t109_n229: return -0.023221999406814575f;
  t109_n230: return 0.0006703255930915475f;
  t109_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t109_n233;
  t109_n232: return 0.004217496141791344f;
  t109_n233: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t109_n235;
  t109_n234: return 0.04273262247443199f;
  t109_n235: return -0.026468835771083832f;
  t109_n236: return -0.020058367401361465f;
  t109_n237: return 0.007652761414647102f;
  t109_n238: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t109_n240;
  t109_n239: return 0.022426867857575417f;
  t109_n240: return 0.0679284855723381f;
  t109_n241: return 0.005855710245668888f;
  t109_n242: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t109_n254;
  t109_n243: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t109_n245;
  t109_n244: return 0.007162673398852348f;
  t109_n245: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t109_n253;
  t109_n246: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t109_n250;
  t109_n247: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t109_n249;
  t109_n248: return 0.018310245126485825f;
  t109_n249: return -0.07312905788421631f;
  t109_n250: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t109_n252;
  t109_n251: return 0.26413530111312866f;
  t109_n252: return 0.03131091222167015f;
  t109_n253: return -0.12704390287399292f;
  t109_n254: return 0.04201345518231392f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree110(const Example& E) {
  t110_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t110_n178;
  t110_n1: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t110_n19;
  t110_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t110_n18;
  t110_n3: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t110_n7;
  t110_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n6;
  t110_n5: return -0.005201318766921759f;
  t110_n6: return 0.000772632600273937f;
  t110_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t110_n17;
  t110_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n16;
  t110_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t110_n11;
  t110_n10: return -0.048979271203279495f;
  t110_n11: if (E.getNumReferences() >= 3315876864 /*5256.5*/) goto t110_n13;
  t110_n12: return 0.024601953104138374f;
  t110_n13: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t110_n15;
  t110_n14: return 0.16212406754493713f;
  t110_n15: return 0.07498686015605927f;
  t110_n16: return 0.011135285720229149f;
  t110_n17: return -0.00559559790417552f;
  t110_n18: return -0.004292536061257124f;
  t110_n19: if (E.getFractionNameInContext() >= 3205809565 /*0.5811403393745422*/) goto t110_n147;
  t110_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t110_n136;
  t110_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t110_n121;
  t110_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t110_n50;
  t110_n23: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t110_n49;
  t110_n24: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t110_n48;
  t110_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t110_n43;
  t110_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t110_n42;
  t110_n27: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t110_n29;
  t110_n28: return -0.037886105477809906f;
  t110_n29: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t110_n41;
  t110_n30: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t110_n32;
  t110_n31: return -0.1269204318523407f;
  t110_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t110_n40;
  t110_n33: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t110_n35;
  t110_n34: return 0.025398489087820053f;
  t110_n35: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t110_n37;
  t110_n36: return 0.003884962759912014f;
  t110_n37: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t110_n39;
  t110_n38: return 0.19217799603939056f;
  t110_n39: return 0.04455575346946716f;
  t110_n40: return -0.06402710825204849f;
  t110_n41: return -0.14911553263664246f;
  t110_n42: return 0.0028875269927084446f;
  t110_n43: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t110_n45;
  t110_n44: return 0.00887866597622633f;
  t110_n45: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t110_n47;
  t110_n46: return -0.009708845056593418f;
  t110_n47: return 0.01660260371863842f;
  t110_n48: return -0.0001324318873230368f;
  t110_n49: return 0.010473774746060371f;
  t110_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t110_n56;
  t110_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n53;
  t110_n52: return -0.032100826501846313f;
  t110_n53: if (E.getNumReferences() >= 3339846528 /*37379.5*/) goto t110_n55;
  t110_n54: return -0.010631367564201355f;
  t110_n55: return -0.1045786514878273f;
  t110_n56: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t110_n96;
  t110_n57: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t110_n91;
  t110_n58: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t110_n72;
  t110_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t110_n63;
  t110_n60: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t110_n62;
  t110_n61: return 0.009659088216722012f;
  t110_n62: return 0.07572650909423828f;
  t110_n63: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t110_n65;
  t110_n64: return 0.04914816841483116f;
  t110_n65: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t110_n67;
  t110_n66: return 0.035169269889593124f;
  t110_n67: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t110_n69;
  t110_n68: return 0.06616559624671936f;
  t110_n69: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t110_n71;
  t110_n70: return 0.1227182000875473f;
  t110_n71: return 0.10001029074192047f;
  t110_n72: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t110_n80;
  t110_n73: if (E.getNumReferences() >= 3325131776 /*11359.0*/) goto t110_n79;
  t110_n74: if (E.getNumReferences() >= 3315668992 /*5155.0*/) goto t110_n76;
  t110_n75: return 0.0022597392089664936f;
  t110_n76: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t110_n78;
  t110_n77: return 0.011646865867078304f;
  t110_n78: return 0.14437167346477509f;
  t110_n79: return -0.18712837994098663f;
  t110_n80: if (E.getNumReferences() >= 3332673792 /*21064.5*/) goto t110_n82;
  t110_n81: return 0.14999404549598694f;
  t110_n82: if (E.getNumReferences() >= 3334637824 /*24900.5*/) goto t110_n90;
  t110_n83: if (E.getNumReferences() >= 3334173440 /*23993.5*/) goto t110_n85;
  t110_n84: return -0.018145399168133736f;
  t110_n85: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t110_n89;
  t110_n86: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t110_n88;
  t110_n87: return 0.14630375802516937f;
  t110_n88: return -0.12873192131519318f;
  t110_n89: return 0.12576410174369812f;
  t110_n90: return -0.0023635674733668566f;
  t110_n91: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t110_n93;
  t110_n92: return 0.020525388419628143f;
  t110_n93: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t110_n95;
  t110_n94: return 0.047386642545461655f;
  t110_n95: return 0.06238992512226105f;
  t110_n96: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t110_n120;
  t110_n97: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t110_n99;
  t110_n98: return -0.0019679339602589607f;
  t110_n99: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t110_n103;
  t110_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n102;
  t110_n101: return -0.08866115659475327f;
  t110_n102: return 0.009947589598596096f;
  t110_n103: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t110_n119;
  t110_n104: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t110_n114;
  t110_n105: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t110_n107;
  t110_n106: return 0.04656270518898964f;
  t110_n107: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t110_n113;
  t110_n108: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t110_n112;
  t110_n109: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t110_n111;
  t110_n110: return 0.028024528175592422f;
  t110_n111: return 0.13608752191066742f;
  t110_n112: return 0.1309078335762024f;
  t110_n113: return -0.09873697906732559f;
  t110_n114: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t110_n116;
  t110_n115: return -0.04578611999750137f;
  t110_n116: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t110_n118;
  t110_n117: return 0.07571617513895035f;
  t110_n118: return 0.025126254186034203f;
  t110_n119: return 0.11163483560085297f;
  t110_n120: return -0.030864287167787552f;
  t110_n121: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t110_n133;
  t110_n122: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t110_n126;
  t110_n123: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t110_n125;
  t110_n124: return 0.029262421652674675f;
  t110_n125: return 0.004001713823527098f;
  t110_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t110_n128;
  t110_n127: return 0.028956396505236626f;
  t110_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n130;
  t110_n129: return -0.11694604158401489f;
  t110_n130: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t110_n132;
  t110_n131: return 0.059059642255306244f;
  t110_n132: return 0.1704065054655075f;
  t110_n133: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t110_n135;
  t110_n134: return 0.004066715016961098f;
  t110_n135: return 0.048205312341451645f;
  t110_n136: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t110_n140;
  t110_n137: if (E.getNumReferences() >= 3291389952 /*698.5*/) goto t110_n139;
  t110_n138: return 0.0013323230668902397f;
  t110_n139: return 0.024413486942648888f;
  t110_n140: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t110_n146;
  t110_n141: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t110_n145;
  t110_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t110_n144;
  t110_n143: return 0.05652255192399025f;
  t110_n144: return -0.00610149372369051f;
  t110_n145: return -0.05324860289692879f;
  t110_n146: return 0.023574406281113625f;
  t110_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t110_n149;
  t110_n148: return 0.0042008887976408005f;
  t110_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n151;
  t110_n150: return -0.06326977908611298f;
  t110_n151: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t110_n153;
  t110_n152: return 0.012571688741445541f;
  t110_n153: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t110_n165;
  t110_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t110_n156;
  t110_n155: return 0.013612275943160057f;
  t110_n156: if (E.getFractionNameInContext() >= 3207524079 /*0.6833333373069763*/) goto t110_n164;
  t110_n157: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t110_n159;
  t110_n158: return 0.10315993428230286f;
  t110_n159: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t110_n161;
  t110_n160: return -0.19591674208641052f;
  t110_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t110_n163;
  t110_n162: return 0.01905515417456627f;
  t110_n163: return 0.05950808525085449f;
  t110_n164: return 0.011036409065127373f;
  t110_n165: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t110_n169;
  t110_n166: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t110_n168;
  t110_n167: return 0.024613700807094574f;
  t110_n168: return 0.16168177127838135f;
  t110_n169: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t110_n177;
  t110_n170: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t110_n176;
  t110_n171: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t110_n173;
  t110_n172: return 0.10340838134288788f;
  t110_n173: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t110_n175;
  t110_n174: return 0.13737869262695312f;
  t110_n175: return 0.049175161868333817f;
  t110_n176: return -0.10541453212499619f;
  t110_n177: return 0.028076456859707832f;
  t110_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n200;
  t110_n179: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t110_n199;
  t110_n180: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t110_n188;
  t110_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n185;
  t110_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t110_n184;
  t110_n183: return -0.14768949151039124f;
  t110_n184: return -0.10560250282287598f;
  t110_n185: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t110_n187;
  t110_n186: return -1.782733678817749f;
  t110_n187: return -0.02395460568368435f;
  t110_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t110_n196;
  t110_n189: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t110_n195;
  t110_n190: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t110_n192;
  t110_n191: return 0.0014709606766700745f;
  t110_n192: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t110_n194;
  t110_n193: return -0.05607473477721214f;
  t110_n194: return -0.0005930791376158595f;
  t110_n195: return 0.041047919541597366f;
  t110_n196: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t110_n198;
  t110_n197: return 0.017637787386775017f;
  t110_n198: return 0.04821731522679329f;
  t110_n199: return -0.018087642267346382f;
  t110_n200: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t110_n250;
  t110_n201: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t110_n249;
  t110_n202: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t110_n236;
  t110_n203: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t110_n213;
  t110_n204: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t110_n212;
  t110_n205: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t110_n207;
  t110_n206: return 0.007376855704933405f;
  t110_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t110_n209;
  t110_n208: return 0.11661893129348755f;
  t110_n209: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t110_n211;
  t110_n210: return 0.1096789613366127f;
  t110_n211: return 0.10876727849245071f;
  t110_n212: return -0.030670268461108208f;
  t110_n213: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t110_n217;
  t110_n214: if (E.getNumReferences() >= 3265200128 /*79.5*/) goto t110_n216;
  t110_n215: return 0.0032325093634426594f;
  t110_n216: return 0.0680447518825531f;
  t110_n217: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t110_n235;
  t110_n218: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t110_n232;
  t110_n219: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t110_n231;
  t110_n220: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t110_n226;
  t110_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t110_n225;
  t110_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t110_n224;
  t110_n223: return -0.10892304033041f;
  t110_n224: return 0.011389249935746193f;
  t110_n225: return 0.037663668394088745f;
  t110_n226: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t110_n228;
  t110_n227: return 0.082058846950531f;
  t110_n228: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t110_n230;
  t110_n229: return 0.08836867660284042f;
  t110_n230: return 0.013132060877978802f;
  t110_n231: return 0.1360996812582016f;
  t110_n232: if (E.getNumReferences() >= 3290218496 /*627.0*/) goto t110_n234;
  t110_n233: return -0.06380024552345276f;
  t110_n234: return 0.05639408901333809f;
  t110_n235: return 0.0025341357104480267f;
  t110_n236: if (E.getNumReferences() >= 3338172416 /*31804.0*/) goto t110_n238;
  t110_n237: return 0.09328605234622955f;
  t110_n238: if (E.getNumReferences() >= 3341699072 /*44616.0*/) goto t110_n240;
  t110_n239: return -0.0398903451859951f;
  t110_n240: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t110_n242;
  t110_n241: return 0.08381330966949463f;
  t110_n242: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t110_n244;
  t110_n243: return -0.05135186389088631f;
  t110_n244: if (E.getNumReferences() >= 3352414592 /*107411.0*/) goto t110_n246;
  t110_n245: return 0.08741586655378342f;
  t110_n246: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t110_n248;
  t110_n247: return -0.15742532908916473f;
  t110_n248: return 0.06486315280199051f;
  t110_n249: return -0.007691806647926569f;
  t110_n250: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t110_n252;
  t110_n251: return 0.014712159521877766f;
  t110_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t110_n254;
  t110_n253: return 0.04431169852614403f;
  t110_n254: return -0.11859950423240662f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree111(const Example& E) {
  t111_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t111_n234;
  t111_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t111_n111;
  t111_n2: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t111_n38;
  t111_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t111_n37;
  t111_n4: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t111_n8;
  t111_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t111_n7;
  t111_n6: return -0.004955016076564789f;
  t111_n7: return 0.000826278468593955f;
  t111_n8: if (E.getNumReferences() >= 3337393664 /*30283.0*/) goto t111_n36;
  t111_n9: if (E.getNumReferences() >= 3336021248 /*27602.5*/) goto t111_n27;
  t111_n10: if (E.getNumReferences() >= 3334350336 /*24339.0*/) goto t111_n26;
  t111_n11: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t111_n21;
  t111_n12: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t111_n14;
  t111_n13: return 0.0063614314422011375f;
  t111_n14: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t111_n18;
  t111_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n17;
  t111_n16: return 0.10779708623886108f;
  t111_n17: return -0.10548136383295059f;
  t111_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t111_n20;
  t111_n19: return 0.10085421800613403f;
  t111_n20: return 0.1366836577653885f;
  t111_n21: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t111_n23;
  t111_n22: return 0.06885496526956558f;
  t111_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t111_n25;
  t111_n24: return 0.10174094140529633f;
  t111_n25: return 0.10774408280849457f;
  t111_n26: return -0.06010580435395241f;
  t111_n27: if (E.getNumReferences() >= 3336316928 /*28180.0*/) goto t111_n31;
  t111_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t111_n30;
  t111_n29: return 0.065726138651371f;
  t111_n30: return 0.13473552465438843f;
  t111_n31: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t111_n33;
  t111_n32: return -0.013164343312382698f;
  t111_n33: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t111_n35;
  t111_n34: return 0.10706564784049988f;
  t111_n35: return 0.05805261805653572f;
  t111_n36: return -0.05943670496344566f;
  t111_n37: return -0.004689315799623728f;
  t111_n38: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t111_n40;
  t111_n39: return -0.02883552573621273f;
  t111_n40: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t111_n76;
  t111_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t111_n43;
  t111_n42: return -0.006900402717292309f;
  t111_n43: if (E.getNumReferences() >= 3341553664 /*44048.0*/) goto t111_n47;
  t111_n44: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t111_n46;
  t111_n45: return 0.07817194610834122f;
  t111_n46: return 0.001840852783061564f;
  t111_n47: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t111_n55;
  t111_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n50;
  t111_n49: return 0.008189341053366661f;
  t111_n50: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t111_n52;
  t111_n51: return 0.13258741796016693f;
  t111_n52: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t111_n54;
  t111_n53: return 0.1279686987400055f;
  t111_n54: return 0.05667230114340782f;
  t111_n55: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t111_n57;
  t111_n56: return -0.1815255731344223f;
  t111_n57: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t111_n69;
  t111_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n62;
  t111_n59: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t111_n61;
  t111_n60: return 0.06044379994273186f;
  t111_n61: return 0.13826905190944672f;
  t111_n62: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t111_n68;
  t111_n63: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t111_n67;
  t111_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t111_n66;
  t111_n65: return 0.10559412091970444f;
  t111_n66: return 0.1722072809934616f;
  t111_n67: return 0.10800279676914215f;
  t111_n68: return 0.13055048882961273f;
  t111_n69: if (E.getNumReferences() >= 3342417792 /*47423.5*/) goto t111_n73;
  t111_n70: if (E.getNumReferences() >= 3341955328 /*45617.0*/) goto t111_n72;
  t111_n71: return -0.19024299085140228f;
  t111_n72: return 0.043240323662757874f;
  t111_n73: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n75;
  t111_n74: return 0.08812731504440308f;
  t111_n75: return -0.1462177038192749f;
  t111_n76: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t111_n78;
  t111_n77: return -0.11915245652198792f;
  t111_n78: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t111_n86;
  t111_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t111_n83;
  t111_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t111_n82;
  t111_n81: return -0.19928276538848877f;
  t111_n82: return 0.09180538356304169f;
  t111_n83: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t111_n85;
  t111_n84: return 0.15263313055038452f;
  t111_n85: return 0.1414436250925064f;
  t111_n86: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n108;
  t111_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t111_n89;
  t111_n88: return -0.0019284250447526574f;
  t111_n89: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t111_n103;
  t111_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t111_n94;
  t111_n91: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t111_n93;
  t111_n92: return -0.007338542956858873f;
  t111_n93: return 0.06006458029150963f;
  t111_n94: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t111_n96;
  t111_n95: return 0.01893451437354088f;
  t111_n96: if (E.getNumReferences() >= 3345303552 /*58696.0*/) goto t111_n100;
  t111_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t111_n99;
  t111_n98: return -0.18563233315944672f;
  t111_n99: return 0.11563024669885635f;
  t111_n100: if (E.getNumReferences() >= 3346303232 /*62601.0*/) goto t111_n102;
  t111_n101: return -0.15741361677646637f;
  t111_n102: return 0.050433505326509476f;
  t111_n103: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t111_n107;
  t111_n104: if (E.getNumReferences() >= 3364253712 /*275328.5*/) goto t111_n106;
  t111_n105: return -0.12067398428916931f;
  t111_n106: return 0.007159675471484661f;
  t111_n107: return 0.06293699890375137f;
  t111_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t111_n110;
  t111_n109: return -0.049956176429986954f;
  t111_n110: return 0.007095606066286564f;
  t111_n111: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t111_n231;
  t111_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t111_n210;
  t111_n113: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t111_n209;
  t111_n114: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t111_n204;
  t111_n115: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t111_n185;
  t111_n116: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t111_n184;
  t111_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t111_n131;
  t111_n118: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n130;
  t111_n119: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t111_n125;
  t111_n120: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t111_n122;
  t111_n121: return -0.0007093716412782669f;
  t111_n122: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t111_n124;
  t111_n123: return 0.018022002652287483f;
  t111_n124: return 0.1327665150165558f;
  t111_n125: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t111_n127;
  t111_n126: return 0.0039385040290653706f;
  t111_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t111_n129;
  t111_n128: return -0.02731527015566826f;
  t111_n129: return 0.008955840021371841f;
  t111_n130: return -0.03690498322248459f;
  t111_n131: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t111_n155;
  t111_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n152;
  t111_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t111_n135;
  t111_n134: return -0.02030360698699951f;
  t111_n135: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t111_n137;
  t111_n136: return -0.0163728017359972f;
  t111_n137: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t111_n143;
  t111_n138: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t111_n140;
  t111_n139: return -0.003385034156963229f;
  t111_n140: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t111_n142;
  t111_n141: return 0.047649916261434555f;
  t111_n142: return 0.012838131748139858f;
  t111_n143: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t111_n145;
  t111_n144: return 0.006634129211306572f;
  t111_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t111_n149;
  t111_n146: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t111_n148;
  t111_n147: return 0.069037064909935f;
  t111_n148: return 0.09759863466024399f;
  t111_n149: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t111_n151;
  t111_n150: return -0.108615942299366f;
  t111_n151: return 0.0798036977648735f;
  t111_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t111_n154;
  t111_n153: return -0.052090879529714584f;
  t111_n154: return 0.055302344262599945f;
  t111_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t111_n159;
  t111_n156: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t111_n158;
  t111_n157: return 0.033522069454193115f;
  t111_n158: return -0.04321013391017914f;
  t111_n159: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t111_n179;
  t111_n160: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t111_n162;
  t111_n161: return 0.12857647240161896f;
  t111_n162: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t111_n172;
  t111_n163: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t111_n167;
  t111_n164: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t111_n166;
  t111_n165: return 0.053599920123815536f;
  t111_n166: return -0.18444295227527618f;
  t111_n167: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t111_n171;
  t111_n168: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n170;
  t111_n169: return -0.8458218574523926f;
  t111_n170: return 0.014802183024585247f;
  t111_n171: return 0.011732845567166805f;
  t111_n172: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t111_n178;
  t111_n173: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t111_n175;
  t111_n174: return -1.8350533246994019f;
  t111_n175: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t111_n177;
  t111_n176: return 0.033834949135780334f;
  t111_n177: return -1.6359494924545288f;
  t111_n178: return 0.07878253608942032f;
  t111_n179: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n181;
  t111_n180: return -0.11120189726352692f;
  t111_n181: if (E.getFileProximityDistanceCost() >= 3250585600 /*24.0*/) goto t111_n183;
  t111_n182: return -0.13557736575603485f;
  t111_n183: return 0.12035846710205078f;
  t111_n184: return -0.01513928547501564f;
  t111_n185: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t111_n195;
  t111_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t111_n192;
  t111_n187: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t111_n189;
  t111_n188: return -0.04813265800476074f;
  t111_n189: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t111_n191;
  t111_n190: return 0.14426150918006897f;
  t111_n191: return -0.22892571985721588f;
  t111_n192: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t111_n194;
  t111_n193: return 0.000561899971216917f;
  t111_n194: return 0.10358958691358566f;
  t111_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t111_n203;
  t111_n196: if (E.getFractionNameInContext() >= 3209807644 /*0.8194444179534912*/) goto t111_n202;
  t111_n197: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t111_n201;
  t111_n198: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t111_n200;
  t111_n199: return -0.22476808726787567f;
  t111_n200: return 0.08979512751102448f;
  t111_n201: return -0.13583654165267944f;
  t111_n202: return 0.11597324162721634f;
  t111_n203: return -0.026627520099282265f;
  t111_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t111_n206;
  t111_n205: return -0.003253932110965252f;
  t111_n206: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t111_n208;
  t111_n207: return 0.0077056242153048515f;
  t111_n208: return 0.02502637729048729f;
  t111_n209: return -0.04917037859559059f;
  t111_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t111_n224;
  t111_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t111_n213;
  t111_n212: return 0.002679849276319146f;
  t111_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t111_n219;
  t111_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t111_n218;
  t111_n215: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t111_n217;
  t111_n216: return -0.030540751293301582f;
  t111_n217: return -0.21052636206150055f;
  t111_n218: return 0.018304849043488503f;
  t111_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t111_n221;
  t111_n220: return 0.009496494196355343f;
  t111_n221: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t111_n223;
  t111_n222: return 0.021467693150043488f;
  t111_n223: return -0.042139939963817596f;
  t111_n224: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t111_n228;
  t111_n225: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t111_n227;
  t111_n226: return -0.0069471788592636585f;
  t111_n227: return 0.023852964863181114f;
  t111_n228: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t111_n230;
  t111_n229: return -0.020510630682110786f;
  t111_n230: return 0.025955861434340477f;
  t111_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t111_n233;
  t111_n232: return 0.0037670712918043137f;
  t111_n233: return 0.016378769651055336f;
  t111_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t111_n252;
  t111_n235: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t111_n251;
  t111_n236: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t111_n250;
  t111_n237: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t111_n239;
  t111_n238: return 0.005742757115513086f;
  t111_n239: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t111_n249;
  t111_n240: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t111_n242;
  t111_n241: return 0.026107823476195335f;
  t111_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t111_n248;
  t111_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t111_n245;
  t111_n244: return 0.04449603334069252f;
  t111_n245: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t111_n247;
  t111_n246: return -0.0007271672948263586f;
  t111_n247: return 0.10758641362190247f;
  t111_n248: return -0.07010538130998611f;
  t111_n249: return 5.031563341617584e-05f;
  t111_n250: return -0.002219980349764228f;
  t111_n251: return 0.015432794578373432f;
  t111_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t111_n254;
  t111_n253: return 0.00943811796605587f;
  t111_n254: return -0.08874285966157913f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree112(const Example& E) {
  t112_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t112_n180;
  t112_n1: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t112_n159;
  t112_n2: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t112_n142;
  t112_n3: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t112_n141;
  t112_n4: if (E.getNumReferences() >= 3359435680 /*193454.5*/) goto t112_n130;
  t112_n5: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t112_n129;
  t112_n6: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t112_n126;
  t112_n7: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t112_n125;
  t112_n8: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t112_n106;
  t112_n9: if (E.getNumReferences() >= 3350789120 /*94712.0*/) goto t112_n103;
  t112_n10: if (E.getNumReferences() >= 3350665472 /*93746.0*/) goto t112_n100;
  t112_n11: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t112_n85;
  t112_n12: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t112_n70;
  t112_n13: if (E.getNumReferences() >= 3341141504 /*42438.0*/) goto t112_n41;
  t112_n14: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t112_n16;
  t112_n15: return -0.0023849690333008766f;
  t112_n16: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t112_n40;
  t112_n17: if (E.getNumReferences() >= 3275718656 /*191.5*/) goto t112_n39;
  t112_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n24;
  t112_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t112_n21;
  t112_n20: return -0.19448454678058624f;
  t112_n21: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t112_n23;
  t112_n22: return 0.01807326450943947f;
  t112_n23: return -0.11874990165233612f;
  t112_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t112_n36;
  t112_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t112_n27;
  t112_n26: return -0.003167852759361267f;
  t112_n27: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n29;
  t112_n28: return 0.008746090345084667f;
  t112_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t112_n31;
  t112_n30: return -0.023566216230392456f;
  t112_n31: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t112_n33;
  t112_n32: return 0.009924876503646374f;
  t112_n33: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t112_n35;
  t112_n34: return 0.18983140587806702f;
  t112_n35: return 0.0673995241522789f;
  t112_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t112_n38;
  t112_n37: return -0.01124120969325304f;
  t112_n38: return 0.03520095720887184f;
  t112_n39: return 0.025804560631513596f;
  t112_n40: return 0.0009646415710449219f;
  t112_n41: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n47;
  t112_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t112_n44;
  t112_n43: return -0.015110882930457592f;
  t112_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t112_n46;
  t112_n45: return 0.04973830655217171f;
  t112_n46: return 0.11504503339529037f;
  t112_n47: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t112_n51;
  t112_n48: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t112_n50;
  t112_n49: return 0.12600040435791016f;
  t112_n50: return 0.02077779546380043f;
  t112_n51: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t112_n61;
  t112_n52: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t112_n54;
  t112_n53: return -0.19816693663597107f;
  t112_n54: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t112_n56;
  t112_n55: return -0.10563648492097855f;
  t112_n56: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t112_n58;
  t112_n57: return 0.11640513688325882f;
  t112_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t112_n60;
  t112_n59: return 0.1279112994670868f;
  t112_n60: return 0.10950107127428055f;
  t112_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t112_n65;
  t112_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t112_n64;
  t112_n63: return -0.07876444607973099f;
  t112_n64: return 0.04404352232813835f;
  t112_n65: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t112_n67;
  t112_n66: return -0.13544999063014984f;
  t112_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t112_n69;
  t112_n68: return -0.1047116369009018f;
  t112_n69: return 0.1426888406276703f;
  t112_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t112_n76;
  t112_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n73;
  t112_n72: return -0.14855962991714478f;
  t112_n73: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n75;
  t112_n74: return 0.06696828454732895f;
  t112_n75: return -0.1473298817873001f;
  t112_n76: if (E.getNumReferences() >= 3344543232 /*55726.0*/) goto t112_n82;
  t112_n77: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t112_n81;
  t112_n78: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t112_n80;
  t112_n79: return 0.07216425240039825f;
  t112_n80: return 0.13513118028640747f;
  t112_n81: return 0.14805133640766144f;
  t112_n82: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t112_n84;
  t112_n83: return 0.1071108803153038f;
  t112_n84: return -0.158321812748909f;
  t112_n85: if (E.getNumReferences() >= 3347793152 /*71306.0*/) goto t112_n87;
  t112_n86: return -0.04350103437900543f;
  t112_n87: if (E.getNumReferences() >= 3350448512 /*92051.0*/) goto t112_n99;
  t112_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n90;
  t112_n89: return -0.012949111871421337f;
  t112_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t112_n98;
  t112_n91: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t112_n97;
  t112_n92: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t112_n94;
  t112_n93: return 0.020436495542526245f;
  t112_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t112_n96;
  t112_n95: return -0.03233746811747551f;
  t112_n96: return 0.0887322649359703f;
  t112_n97: return 0.006305673159658909f;
  t112_n98: return 0.0831417590379715f;
  t112_n99: return -0.07152204215526581f;
  t112_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n102;
  t112_n101: return -0.1136685311794281f;
  t112_n102: return 0.10507894307374954f;
  t112_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n105;
  t112_n104: return -0.1608457714319229f;
  t112_n105: return -0.1610443890094757f;
  t112_n106: if (E.getNumReferences() >= 3352190592 /*105661.0*/) goto t112_n112;
  t112_n107: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t112_n109;
  t112_n108: return 0.04834260791540146f;
  t112_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t112_n111;
  t112_n110: return 0.07310564070940018f;
  t112_n111: return 0.1158093735575676f;
  t112_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n114;
  t112_n113: return -0.047660406678915024f;
  t112_n114: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t112_n118;
  t112_n115: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t112_n117;
  t112_n116: return -0.036595944315195084f;
  t112_n117: return 0.12101174145936966f;
  t112_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t112_n120;
  t112_n119: return -0.12226519733667374f;
  t112_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t112_n122;
  t112_n121: return 0.0024548075161874294f;
  t112_n122: if (E.getNumReferences() >= 3354093056 /*120524.0*/) goto t112_n124;
  t112_n123: return -0.052903637290000916f;
  t112_n124: return 0.09483189880847931f;
  t112_n125: return -0.21154667437076569f;
  t112_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t112_n128;
  t112_n127: return -0.004151434171944857f;
  t112_n128: return 0.07701951265335083f;
  t112_n129: return -0.10678945481777191f;
  t112_n130: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t112_n138;
  t112_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t112_n135;
  t112_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n134;
  t112_n133: return -0.19551482796669006f;
  t112_n134: return 0.018218522891402245f;
  t112_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t112_n137;
  t112_n136: return -0.10418860614299774f;
  t112_n137: return 0.060730576515197754f;
  t112_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t112_n140;
  t112_n139: return -0.007157705258578062f;
  t112_n140: return 0.036033034324645996f;
  t112_n141: return -0.06558343023061752f;
  t112_n142: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t112_n154;
  t112_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t112_n147;
  t112_n144: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t112_n146;
  t112_n145: return 0.051700327545404434f;
  t112_n146: return -0.04010331258177757f;
  t112_n147: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t112_n153;
  t112_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t112_n150;
  t112_n149: return 0.04062872380018234f;
  t112_n150: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n152;
  t112_n151: return 0.06674748659133911f;
  t112_n152: return 0.09630940854549408f;
  t112_n153: return 0.005430266261100769f;
  t112_n154: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t112_n156;
  t112_n155: return -0.11206169426441193f;
  t112_n156: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n158;
  t112_n157: return 0.024901138618588448f;
  t112_n158: return -0.07014524191617966f;
  t112_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t112_n173;
  t112_n160: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t112_n162;
  t112_n161: return 0.0023272880353033543f;
  t112_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t112_n164;
  t112_n163: return -0.1795036643743515f;
  t112_n164: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t112_n166;
  t112_n165: return 0.13323666155338287f;
  t112_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t112_n168;
  t112_n167: return -0.06873956322669983f;
  t112_n168: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t112_n172;
  t112_n169: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t112_n171;
  t112_n170: return 0.12298658490180969f;
  t112_n171: return 0.1412690281867981f;
  t112_n172: return -0.00645489664748311f;
  t112_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t112_n177;
  t112_n174: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t112_n176;
  t112_n175: return -0.011959961615502834f;
  t112_n176: return 0.07825152575969696f;
  t112_n177: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t112_n179;
  t112_n178: return 0.06990277022123337f;
  t112_n179: return -0.024079032242298126f;
  t112_n180: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t112_n250;
  t112_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t112_n233;
  t112_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n226;
  t112_n183: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t112_n185;
  t112_n184: return -0.07937192916870117f;
  t112_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n221;
  t112_n186: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t112_n220;
  t112_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t112_n199;
  t112_n188: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t112_n196;
  t112_n189: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t112_n193;
  t112_n190: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t112_n192;
  t112_n191: return -0.012338201515376568f;
  t112_n192: return -0.10322044044733047f;
  t112_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n195;
  t112_n194: return -0.09802231192588806f;
  t112_n195: return 0.021011944860219955f;
  t112_n196: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t112_n198;
  t112_n197: return -0.04651069641113281f;
  t112_n198: return 0.015389588661491871f;
  t112_n199: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t112_n211;
  t112_n200: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t112_n204;
  t112_n201: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t112_n203;
  t112_n202: return -0.009046518243849277f;
  t112_n203: return -0.08020038157701492f;
  t112_n204: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t112_n210;
  t112_n205: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t112_n209;
  t112_n206: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t112_n208;
  t112_n207: return 0.007945477962493896f;
  t112_n208: return 0.08274711668491364f;
  t112_n209: return -0.12105555087327957f;
  t112_n210: return 0.13492102921009064f;
  t112_n211: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t112_n219;
  t112_n212: if (E.getFileProximityDistanceCost() >= 3251634176 /*26.0*/) goto t112_n216;
  t112_n213: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t112_n215;
  t112_n214: return 0.11479691416025162f;
  t112_n215: return 0.021103786304593086f;
  t112_n216: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t112_n218;
  t112_n217: return 0.07067462801933289f;
  t112_n218: return 0.0447857566177845f;
  t112_n219: return 0.014854333363473415f;
  t112_n220: return -0.05503895878791809f;
  t112_n221: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t112_n225;
  t112_n222: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t112_n224;
  t112_n223: return 0.004244479816406965f;
  t112_n224: return 0.015994500368833542f;
  t112_n225: return 0.0003350320621393621f;
  t112_n226: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t112_n232;
  t112_n227: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t112_n231;
  t112_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t112_n230;
  t112_n229: return -0.16267384588718414f;
  t112_n230: return 0.04341832175850868f;
  t112_n231: return 0.031241632997989655f;
  t112_n232: return -0.14461608231067657f;
  t112_n233: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t112_n245;
  t112_n234: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t112_n244;
  t112_n235: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t112_n237;
  t112_n236: return -0.015341459773480892f;
  t112_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t112_n241;
  t112_n238: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t112_n240;
  t112_n239: return -0.07055865228176117f;
  t112_n240: return 0.08062306046485901f;
  t112_n241: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t112_n243;
  t112_n242: return 0.013163436204195023f;
  t112_n243: return 0.06989337503910065f;
  t112_n244: return 0.003954735584557056f;
  t112_n245: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t112_n247;
  t112_n246: return 0.007924298755824566f;
  t112_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t112_n249;
  t112_n248: return 0.01477009616792202f;
  t112_n249: return 0.03531655669212341f;
  t112_n250: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t112_n252;
  t112_n251: return 0.03133270889520645f;
  t112_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t112_n254;
  t112_n253: return 0.012311493046581745f;
  t112_n254: return 0.060262344777584076f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree113(const Example& E) {
  t113_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t113_n224;
  t113_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t113_n47;
  t113_n2: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t113_n40;
  t113_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n39;
  t113_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t113_n6;
  t113_n5: return -0.005736449267715216f;
  t113_n6: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t113_n18;
  t113_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t113_n13;
  t113_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t113_n10;
  t113_n9: return -0.15314045548439026f;
  t113_n10: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t113_n12;
  t113_n11: return 0.04460678622126579f;
  t113_n12: return 0.005170593969523907f;
  t113_n13: if (E.getNumReferences() >= 3274211328 /*168.5*/) goto t113_n15;
  t113_n14: return 0.028400635346770287f;
  t113_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t113_n17;
  t113_n16: return 0.05272459611296654f;
  t113_n17: return 0.15599551796913147f;
  t113_n18: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t113_n28;
  t113_n19: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t113_n27;
  t113_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n24;
  t113_n21: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t113_n23;
  t113_n22: return -0.07757987082004547f;
  t113_n23: return -0.004447129555046558f;
  t113_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t113_n26;
  t113_n25: return 0.01173095777630806f;
  t113_n26: return -0.020030686631798744f;
  t113_n27: return 0.049537114799022675f;
  t113_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t113_n36;
  t113_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t113_n35;
  t113_n30: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t113_n34;
  t113_n31: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t113_n33;
  t113_n32: return 0.01877022162079811f;
  t113_n33: return -0.19058580696582794f;
  t113_n34: return -0.015294366516172886f;
  t113_n35: return 0.01781558245420456f;
  t113_n36: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t113_n38;
  t113_n37: return -0.0064754793420434f;
  t113_n38: return -0.06319129467010498f;
  t113_n39: return 0.0015845184680074453f;
  t113_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n44;
  t113_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n43;
  t113_n42: return -0.03415511175990105f;
  t113_n43: return 0.014492071233689785f;
  t113_n44: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t113_n46;
  t113_n45: return 0.01612858474254608f;
  t113_n46: return -0.0022018258459866047f;
  t113_n47: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t113_n205;
  t113_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n76;
  t113_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t113_n53;
  t113_n50: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t113_n52;
  t113_n51: return -0.009993808344006538f;
  t113_n52: return 0.014980337582528591f;
  t113_n53: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t113_n73;
  t113_n54: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t113_n72;
  t113_n55: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t113_n71;
  t113_n56: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t113_n66;
  t113_n57: if (E.getNumReferences() >= 3337697792 /*30877.0*/) goto t113_n59;
  t113_n58: return -0.004279077053070068f;
  t113_n59: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t113_n65;
  t113_n60: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t113_n64;
  t113_n61: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t113_n63;
  t113_n62: return 0.10955122858285904f;
  t113_n63: return 0.07590501755475998f;
  t113_n64: return 0.11521683633327484f;
  t113_n65: return 0.02020682580769062f;
  t113_n66: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t113_n68;
  t113_n67: return 0.018251823261380196f;
  t113_n68: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t113_n70;
  t113_n69: return 0.04807957634329796f;
  t113_n70: return -0.011117805726826191f;
  t113_n71: return -0.09366310387849808f;
  t113_n72: return 0.06835304945707321f;
  t113_n73: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t113_n75;
  t113_n74: return 0.10994027554988861f;
  t113_n75: return 0.15738284587860107f;
  t113_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t113_n114;
  t113_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n83;
  t113_n78: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t113_n82;
  t113_n79: if (E.getNumReferences() >= 3325242368 /*11467.0*/) goto t113_n81;
  t113_n80: return -0.09625978767871857f;
  t113_n81: return -0.14505699276924133f;
  t113_n82: return -0.15975455939769745f;
  t113_n83: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t113_n113;
  t113_n84: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t113_n92;
  t113_n85: if (E.getNumReferences() >= 3277717504 /*222.0*/) goto t113_n91;
  t113_n86: if (E.getNumReferences() >= 3274342400 /*170.5*/) goto t113_n88;
  t113_n87: return 0.005235723685473204f;
  t113_n88: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t113_n90;
  t113_n89: return 0.21377377212047577f;
  t113_n90: return -0.10380604863166809f;
  t113_n91: return -0.08397913724184036f;
  t113_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t113_n94;
  t113_n93: return -0.10461385548114777f;
  t113_n94: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t113_n106;
  t113_n95: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t113_n103;
  t113_n96: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t113_n98;
  t113_n97: return -0.10961317270994186f;
  t113_n98: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t113_n102;
  t113_n99: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t113_n101;
  t113_n100: return 0.054279062896966934f;
  t113_n101: return 0.0022679090034216642f;
  t113_n102: return 0.05108409747481346f;
  t113_n103: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t113_n105;
  t113_n104: return -0.015399891883134842f;
  t113_n105: return 0.03845791146159172f;
  t113_n106: if (E.getNumReferences() >= 3356993440 /*155294.5*/) goto t113_n112;
  t113_n107: if (E.getNumReferences() >= 3334173440 /*23993.5*/) goto t113_n111;
  t113_n108: if (E.getNumReferences() >= 3332691712 /*21099.5*/) goto t113_n110;
  t113_n109: return 0.009852546267211437f;
  t113_n110: return -0.14561587572097778f;
  t113_n111: return 0.03945377841591835f;
  t113_n112: return -0.3533863127231598f;
  t113_n113: return -0.010045250877737999f;
  t113_n114: if (E.getNumReferences() >= 3322245120 /*8540.0*/) goto t113_n204;
  t113_n115: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t113_n195;
  t113_n116: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t113_n192;
  t113_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n123;
  t113_n118: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t113_n122;
  t113_n119: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t113_n121;
  t113_n120: return 0.010376626625657082f;
  t113_n121: return -0.05268038809299469f;
  t113_n122: return -0.01215886976569891f;
  t113_n123: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t113_n191;
  t113_n124: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t113_n184;
  t113_n125: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t113_n177;
  t113_n126: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t113_n162;
  t113_n127: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t113_n157;
  t113_n128: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t113_n156;
  t113_n129: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t113_n131;
  t113_n130: return 0.01687428168952465f;
  t113_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n135;
  t113_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n134;
  t113_n133: return -0.06434766948223114f;
  t113_n134: return 0.00745679996907711f;
  t113_n135: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t113_n137;
  t113_n136: return -0.0020542277488857508f;
  t113_n137: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t113_n155;
  t113_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t113_n140;
  t113_n139: return -0.010905220173299313f;
  t113_n140: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t113_n154;
  t113_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t113_n151;
  t113_n142: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t113_n144;
  t113_n143: return 0.02017829194664955f;
  t113_n144: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t113_n146;
  t113_n145: return 0.018443066626787186f;
  t113_n146: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t113_n150;
  t113_n147: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t113_n149;
  t113_n148: return 0.08824146538972855f;
  t113_n149: return 0.12959101796150208f;
  t113_n150: return 0.0741359144449234f;
  t113_n151: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t113_n153;
  t113_n152: return 0.04097279533743858f;
  t113_n153: return -0.001584047800861299f;
  t113_n154: return 0.09590199589729309f;
  t113_n155: return 0.25080764293670654f;
  t113_n156: return -0.006366312503814697f;
  t113_n157: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t113_n161;
  t113_n158: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t113_n160;
  t113_n159: return 0.009541863575577736f;
  t113_n160: return -0.10865360498428345f;
  t113_n161: return -0.04353565722703934f;
  t113_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n168;
  t113_n163: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t113_n165;
  t113_n164: return -0.04794768616557121f;
  t113_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t113_n167;
  t113_n166: return 0.02377360686659813f;
  t113_n167: return 0.11069890856742859f;
  t113_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n170;
  t113_n169: return -0.006415022071450949f;
  t113_n170: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t113_n176;
  t113_n171: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t113_n173;
  t113_n172: return 0.03656414896249771f;
  t113_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n175;
  t113_n174: return -0.07679085433483124f;
  t113_n175: return 0.12282291054725647f;
  t113_n176: return -0.13411185145378113f;
  t113_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t113_n181;
  t113_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n180;
  t113_n179: return -0.12125115841627121f;
  t113_n180: return 0.010065978392958641f;
  t113_n181: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t113_n183;
  t113_n182: return 0.029850685968995094f;
  t113_n183: return -0.1085430160164833f;
  t113_n184: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t113_n190;
  t113_n185: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t113_n187;
  t113_n186: return -0.0041124094277620316f;
  t113_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n189;
  t113_n188: return -0.14206045866012573f;
  t113_n189: return 0.16816876828670502f;
  t113_n190: return -0.0709267258644104f;
  t113_n191: return 0.031762395054101944f;
  t113_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t113_n194;
  t113_n193: return -0.025313880294561386f;
  t113_n194: return 0.04761645197868347f;
  t113_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n197;
  t113_n196: return -0.018836205825209618f;
  t113_n197: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t113_n199;
  t113_n198: return -0.05830562114715576f;
  t113_n199: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t113_n201;
  t113_n200: return 0.036187686026096344f;
  t113_n201: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t113_n203;
  t113_n202: return -0.07545486837625504f;
  t113_n203: return 0.024591386318206787f;
  t113_n204: return 0.01653948612511158f;
  t113_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n207;
  t113_n206: return -0.052355311810970306f;
  t113_n207: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t113_n211;
  t113_n208: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t113_n210;
  t113_n209: return -0.04934771731495857f;
  t113_n210: return 0.05107763037085533f;
  t113_n211: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t113_n217;
  t113_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t113_n214;
  t113_n213: return 0.016958091408014297f;
  t113_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t113_n216;
  t113_n215: return 0.011041529476642609f;
  t113_n216: return 0.05908632278442383f;
  t113_n217: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t113_n223;
  t113_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t113_n220;
  t113_n219: return -0.005157408770173788f;
  t113_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t113_n222;
  t113_n221: return -0.032967835664749146f;
  t113_n222: return 0.02216389775276184f;
  t113_n223: return 0.099570132791996f;
  t113_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n236;
  t113_n225: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t113_n235;
  t113_n226: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t113_n230;
  t113_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t113_n229;
  t113_n228: return -0.13641303777694702f;
  t113_n229: return -0.023168860003352165f;
  t113_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t113_n232;
  t113_n231: return 0.0006326998118311167f;
  t113_n232: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t113_n234;
  t113_n233: return 0.014072928577661514f;
  t113_n234: return 0.049114253371953964f;
  t113_n235: return -0.015303835272789001f;
  t113_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t113_n252;
  t113_n237: if (E.getNumReferences() >= 3273097216 /*151.5*/) goto t113_n251;
  t113_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t113_n244;
  t113_n239: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t113_n243;
  t113_n240: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t113_n242;
  t113_n241: return 0.011862154118716717f;
  t113_n242: return 0.1160079836845398f;
  t113_n243: return -0.01063183881342411f;
  t113_n244: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t113_n248;
  t113_n245: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t113_n247;
  t113_n246: return -0.0008188047213479877f;
  t113_n247: return 0.014364889822900295f;
  t113_n248: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t113_n250;
  t113_n249: return 0.03190087154507637f;
  t113_n250: return -0.002867949428036809f;
  t113_n251: return 0.02288883551955223f;
  t113_n252: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t113_n254;
  t113_n253: return 0.01383031252771616f;
  t113_n254: return 0.03937830403447151f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree114(const Example& E) {
  t114_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t114_n236;
  t114_n1: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t114_n229;
  t114_n2: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t114_n200;
  t114_n3: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t114_n163;
  t114_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t114_n6;
  t114_n5: return -0.02597106620669365f;
  t114_n6: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t114_n162;
  t114_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t114_n9;
  t114_n8: return -0.0016677845269441605f;
  t114_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n51;
  t114_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n18;
  t114_n11: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t114_n13;
  t114_n12: return -0.09702319651842117f;
  t114_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n15;
  t114_n14: return -0.10725616663694382f;
  t114_n15: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n17;
  t114_n16: return 0.03694186732172966f;
  t114_n17: return -0.011563270352780819f;
  t114_n18: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t114_n20;
  t114_n19: return -0.04756084457039833f;
  t114_n20: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t114_n40;
  t114_n21: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t114_n39;
  t114_n22: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t114_n24;
  t114_n23: return 0.020110873505473137f;
  t114_n24: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t114_n36;
  t114_n25: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t114_n31;
  t114_n26: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t114_n30;
  t114_n27: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t114_n29;
  t114_n28: return 0.04622703790664673f;
  t114_n29: return 0.0010175658389925957f;
  t114_n30: return 0.04564996808767319f;
  t114_n31: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t114_n35;
  t114_n32: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t114_n34;
  t114_n33: return 0.32339146733283997f;
  t114_n34: return 0.1134098470211029f;
  t114_n35: return 0.1046086773276329f;
  t114_n36: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t114_n38;
  t114_n37: return -0.011282003484666348f;
  t114_n38: return 0.03235044330358505f;
  t114_n39: return -0.010113955475389957f;
  t114_n40: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t114_n44;
  t114_n41: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t114_n43;
  t114_n42: return 0.04646436870098114f;
  t114_n43: return -0.017687736079096794f;
  t114_n44: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t114_n50;
  t114_n45: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t114_n49;
  t114_n46: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t114_n48;
  t114_n47: return -0.10080470889806747f;
  t114_n48: return 1.1375268697738647f;
  t114_n49: return -0.121660977602005f;
  t114_n50: return 0.036628589034080505f;
  t114_n51: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t114_n139;
  t114_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t114_n90;
  t114_n53: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n63;
  t114_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n56;
  t114_n55: return -0.01062625739723444f;
  t114_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n62;
  t114_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n59;
  t114_n58: return -0.03699127584695816f;
  t114_n59: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t114_n61;
  t114_n60: return 0.023422284051775932f;
  t114_n61: return 0.0898444876074791f;
  t114_n62: return 0.01103063765913248f;
  t114_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t114_n65;
  t114_n64: return 0.003067909274250269f;
  t114_n65: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t114_n79;
  t114_n66: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t114_n74;
  t114_n67: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t114_n73;
  t114_n68: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t114_n70;
  t114_n69: return 0.01479382999241352f;
  t114_n70: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t114_n72;
  t114_n71: return 0.13410280644893646f;
  t114_n72: return 0.028671475127339363f;
  t114_n73: return -0.008147931657731533f;
  t114_n74: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t114_n78;
  t114_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n77;
  t114_n76: return -0.11725073307752609f;
  t114_n77: return 0.0530594065785408f;
  t114_n78: return 0.0008862720569595695f;
  t114_n79: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t114_n83;
  t114_n80: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t114_n82;
  t114_n81: return -0.011154674924910069f;
  t114_n82: return 0.011976159177720547f;
  t114_n83: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t114_n87;
  t114_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n86;
  t114_n85: return -0.13691692054271698f;
  t114_n86: return 0.021371927112340927f;
  t114_n87: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t114_n89;
  t114_n88: return 0.002408769214525819f;
  t114_n89: return 0.03779341280460358f;
  t114_n90: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t114_n128;
  t114_n91: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t114_n95;
  t114_n92: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t114_n94;
  t114_n93: return 0.018770210444927216f;
  t114_n94: return 0.19309678673744202f;
  t114_n95: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t114_n117;
  t114_n96: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t114_n110;
  t114_n97: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n109;
  t114_n98: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t114_n100;
  t114_n99: return 0.0023998164106160402f;
  t114_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t114_n102;
  t114_n101: return 0.0019114919705316424f;
  t114_n102: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t114_n104;
  t114_n103: return 0.04119708389043808f;
  t114_n104: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t114_n108;
  t114_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n107;
  t114_n106: return -0.04145418480038643f;
  t114_n107: return 0.1207703948020935f;
  t114_n108: return 0.030698511749505997f;
  t114_n109: return 0.003674475010484457f;
  t114_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n116;
  t114_n111: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t114_n113;
  t114_n112: return -0.08204032480716705f;
  t114_n113: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t114_n115;
  t114_n114: return 0.08833962678909302f;
  t114_n115: return -0.1011025533080101f;
  t114_n116: return -0.029504209756851196f;
  t114_n117: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t114_n119;
  t114_n118: return 0.027057886123657227f;
  t114_n119: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t114_n123;
  t114_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n122;
  t114_n121: return -0.06605532020330429f;
  t114_n122: return -0.014937447383999825f;
  t114_n123: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t114_n125;
  t114_n124: return 0.021775750443339348f;
  t114_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n127;
  t114_n126: return -0.08555255085229874f;
  t114_n127: return 0.016993306577205658f;
  t114_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t114_n132;
  t114_n129: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t114_n131;
  t114_n130: return -0.005984407849609852f;
  t114_n131: return 0.04057050868868828f;
  t114_n132: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n138;
  t114_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n135;
  t114_n134: return -0.002298480598255992f;
  t114_n135: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t114_n137;
  t114_n136: return 0.06547489017248154f;
  t114_n137: return 0.15109992027282715f;
  t114_n138: return 0.04875184968113899f;
  t114_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t114_n143;
  t114_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n142;
  t114_n141: return -0.04219229891896248f;
  t114_n142: return 0.00019750147475861013f;
  t114_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n145;
  t114_n144: return -0.018161285668611526f;
  t114_n145: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t114_n159;
  t114_n146: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t114_n156;
  t114_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n149;
  t114_n148: return -0.0318593755364418f;
  t114_n149: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t114_n151;
  t114_n150: return -0.21094465255737305f;
  t114_n151: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t114_n153;
  t114_n152: return 0.09473061561584473f;
  t114_n153: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t114_n155;
  t114_n154: return 0.12256628274917603f;
  t114_n155: return 0.22358624637126923f;
  t114_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n158;
  t114_n157: return -0.031249115243554115f;
  t114_n158: return 0.15033948421478271f;
  t114_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n161;
  t114_n160: return -0.0006794232176616788f;
  t114_n161: return 0.03829061612486839f;
  t114_n162: return -0.054001711308956146f;
  t114_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n187;
  t114_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t114_n166;
  t114_n165: return -0.017104696482419968f;
  t114_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n168;
  t114_n167: return -0.13662543892860413f;
  t114_n168: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t114_n170;
  t114_n169: return 0.05665671452879906f;
  t114_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t114_n180;
  t114_n171: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t114_n173;
  t114_n172: return 0.06339586526155472f;
  t114_n173: if (E.getNumReferences() >= 3249537024 /*22.0*/) goto t114_n179;
  t114_n174: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t114_n178;
  t114_n175: if (E.getFractionNameInContext() >= 3188603108 /*0.1388888955116272*/) goto t114_n177;
  t114_n176: return 0.7254771590232849f;
  t114_n177: return 0.13920453190803528f;
  t114_n178: return -0.1440895050764084f;
  t114_n179: return -0.5038118362426758f;
  t114_n180: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t114_n186;
  t114_n181: if (E.getFractionNameInContext() >= 3175926989 /*0.05000000074505806*/) goto t114_n185;
  t114_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t114_n184;
  t114_n183: return 0.07195068895816803f;
  t114_n184: return 0.13812875747680664f;
  t114_n185: return 0.13460847735404968f;
  t114_n186: return 0.07064228504896164f;
  t114_n187: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t114_n189;
  t114_n188: return 0.028749734163284302f;
  t114_n189: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t114_n191;
  t114_n190: return -0.03998810797929764f;
  t114_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t114_n193;
  t114_n192: return -0.01072782464325428f;
  t114_n193: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t114_n199;
  t114_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t114_n198;
  t114_n195: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t114_n197;
  t114_n196: return 0.01735427789390087f;
  t114_n197: return -0.029029766097664833f;
  t114_n198: return 0.03580562397837639f;
  t114_n199: return 0.10044964402914047f;
  t114_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t114_n226;
  t114_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t114_n225;
  t114_n202: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t114_n216;
  t114_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n211;
  t114_n204: if (E.getNumReferences() >= 3329498112 /*15623.0*/) goto t114_n206;
  t114_n205: return -0.056640662252902985f;
  t114_n206: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n208;
  t114_n207: return -0.3559429943561554f;
  t114_n208: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t114_n210;
  t114_n209: return -0.11744114011526108f;
  t114_n210: return 0.027973128482699394f;
  t114_n211: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n215;
  t114_n212: if (E.getNumReferences() >= 3316316160 /*5471.0*/) goto t114_n214;
  t114_n213: return 0.020294783636927605f;
  t114_n214: return 0.07662207633256912f;
  t114_n215: return -0.02746506594121456f;
  t114_n216: if (E.getNumReferences() >= 3325781504 /*11993.5*/) goto t114_n222;
  t114_n217: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t114_n221;
  t114_n218: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t114_n220;
  t114_n219: return -0.1599729061126709f;
  t114_n220: return 0.26133042573928833f;
  t114_n221: return 0.025714918971061707f;
  t114_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n224;
  t114_n223: return -0.06829214096069336f;
  t114_n224: return 0.04428427666425705f;
  t114_n225: return 0.00051396299386397f;
  t114_n226: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t114_n228;
  t114_n227: return 0.02611999399960041f;
  t114_n228: return 0.009349498897790909f;
  t114_n229: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t114_n231;
  t114_n230: return 0.0063604977913200855f;
  t114_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t114_n233;
  t114_n232: return 0.00956257339566946f;
  t114_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n235;
  t114_n234: return -0.04893653839826584f;
  t114_n235: return 0.054953351616859436f;
  t114_n236: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t114_n252;
  t114_n237: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t114_n247;
  t114_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n240;
  t114_n239: return 6.0952839703531936e-05f;
  t114_n240: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t114_n246;
  t114_n241: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t114_n245;
  t114_n242: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t114_n244;
  t114_n243: return -0.16083800792694092f;
  t114_n244: return 0.04655851423740387f;
  t114_n245: return 0.029921263456344604f;
  t114_n246: return -0.12882667779922485f;
  t114_n247: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t114_n251;
  t114_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t114_n250;
  t114_n249: return -0.04717494919896126f;
  t114_n250: return 0.0034555683378130198f;
  t114_n251: return 0.008995169773697853f;
  t114_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t114_n254;
  t114_n253: return 0.009516747668385506f;
  t114_n254: return 0.0323210246860981f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree115(const Example& E) {
  t115_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t115_n232;
  t115_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t115_n41;
  t115_n2: if (E.getNumReferences() >= 3300470784 /*1481.5*/) goto t115_n32;
  t115_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t115_n13;
  t115_n4: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t115_n12;
  t115_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t115_n7;
  t115_n6: return 0.002375636249780655f;
  t115_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t115_n11;
  t115_n8: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t115_n10;
  t115_n9: return -0.015430856496095657f;
  t115_n10: return 0.022858064621686935f;
  t115_n11: return 0.04063224419951439f;
  t115_n12: return -0.0007057770271785557f;
  t115_n13: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t115_n25;
  t115_n14: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t115_n18;
  t115_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n17;
  t115_n16: return -0.09361381083726883f;
  t115_n17: return -0.01408380176872015f;
  t115_n18: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t115_n20;
  t115_n19: return 0.004307369235903025f;
  t115_n20: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t115_n22;
  t115_n21: return 0.027120618149638176f;
  t115_n22: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t115_n24;
  t115_n23: return 0.06815558671951294f;
  t115_n24: return 0.10535949468612671f;
  t115_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t115_n31;
  t115_n26: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t115_n28;
  t115_n27: return 0.009480170905590057f;
  t115_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n30;
  t115_n29: return -0.03890591487288475f;
  t115_n30: return 0.009704632684588432f;
  t115_n31: return 0.07963550090789795f;
  t115_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n34;
  t115_n33: return -0.0607120655477047f;
  t115_n34: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t115_n40;
  t115_n35: if (E.getNumReferences() >= 3301531648 /*1611.0*/) goto t115_n39;
  t115_n36: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t115_n38;
  t115_n37: return 0.14802607893943787f;
  t115_n38: return -0.02901131473481655f;
  t115_n39: return 0.0431952066719532f;
  t115_n40: return 0.01531428936868906f;
  t115_n41: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t115_n157;
  t115_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t115_n86;
  t115_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t115_n45;
  t115_n44: return -0.003593018278479576f;
  t115_n45: if (E.getNumReferences() >= 3303006208 /*1791.0*/) goto t115_n85;
  t115_n46: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t115_n48;
  t115_n47: return -0.011921098455786705f;
  t115_n48: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t115_n84;
  t115_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t115_n61;
  t115_n50: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t115_n58;
  t115_n51: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t115_n55;
  t115_n52: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t115_n54;
  t115_n53: return 0.01284706499427557f;
  t115_n54: return -0.035403963178396225f;
  t115_n55: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t115_n57;
  t115_n56: return 0.017312267795205116f;
  t115_n57: return 0.06237030401825905f;
  t115_n58: if (E.getSymbolScopeDistanceCost() >= 3249012736 /*21.0*/) goto t115_n60;
  t115_n59: return 0.17564891278743744f;
  t115_n60: return -0.027649523690342903f;
  t115_n61: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t115_n75;
  t115_n62: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t115_n66;
  t115_n63: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t115_n65;
  t115_n64: return 0.09217070788145065f;
  t115_n65: return 0.012677743099629879f;
  t115_n66: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t115_n74;
  t115_n67: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t115_n73;
  t115_n68: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t115_n72;
  t115_n69: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t115_n71;
  t115_n70: return 0.04028934985399246f;
  t115_n71: return -0.012201766483485699f;
  t115_n72: return 0.04092585667967796f;
  t115_n73: return 0.11648958176374435f;
  t115_n74: return -0.015560224652290344f;
  t115_n75: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t115_n77;
  t115_n76: return 0.006832376588135958f;
  t115_n77: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t115_n79;
  t115_n78: return 0.0762142539024353f;
  t115_n79: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t115_n83;
  t115_n80: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t115_n82;
  t115_n81: return 0.1010182574391365f;
  t115_n82: return -0.11038711667060852f;
  t115_n83: return -0.03018435835838318f;
  t115_n84: return 0.03897171467542648f;
  t115_n85: return -0.023287680000066757f;
  t115_n86: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t115_n110;
  t115_n87: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t115_n103;
  t115_n88: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t115_n98;
  t115_n89: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t115_n91;
  t115_n90: return 0.0069238608703017235f;
  t115_n91: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t115_n97;
  t115_n92: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t115_n96;
  t115_n93: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t115_n95;
  t115_n94: return 0.1013672724366188f;
  t115_n95: return 0.028467442840337753f;
  t115_n96: return 0.05321776866912842f;
  t115_n97: return -0.021114930510520935f;
  t115_n98: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t115_n102;
  t115_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t115_n101;
  t115_n100: return -0.008188579231500626f;
  t115_n101: return 0.0021085471380501986f;
  t115_n102: return 0.013798504136502743f;
  t115_n103: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t115_n107;
  t115_n104: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t115_n106;
  t115_n105: return -0.043813496828079224f;
  t115_n106: return -0.2199534922838211f;
  t115_n107: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t115_n109;
  t115_n108: return -0.002037421567365527f;
  t115_n109: return 0.18500150740146637f;
  t115_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement))) goto t115_n130;
  t115_n111: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t115_n115;
  t115_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n114;
  t115_n113: return -0.046012453734874725f;
  t115_n114: return -0.0035494500771164894f;
  t115_n115: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t115_n127;
  t115_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t115_n118;
  t115_n117: return -0.0067421323619782925f;
  t115_n118: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t115_n120;
  t115_n119: return 0.0033003126736730337f;
  t115_n120: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t115_n126;
  t115_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t115_n125;
  t115_n122: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t115_n124;
  t115_n123: return 0.03695892170071602f;
  t115_n124: return -0.0025316195096820593f;
  t115_n125: return 0.07207994908094406f;
  t115_n126: return -0.01564568653702736f;
  t115_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n129;
  t115_n128: return -0.07104837149381638f;
  t115_n129: return -0.008099970407783985f;
  t115_n130: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t115_n144;
  t115_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t115_n137;
  t115_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t115_n136;
  t115_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n135;
  t115_n134: return -0.051688168197870255f;
  t115_n135: return 0.004972155671566725f;
  t115_n136: return 0.07533760368824005f;
  t115_n137: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t115_n141;
  t115_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t115_n140;
  t115_n139: return 0.018217161297798157f;
  t115_n140: return 0.056297607719898224f;
  t115_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t115_n143;
  t115_n142: return 0.04254290089011192f;
  t115_n143: return 0.1260591745376587f;
  t115_n144: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t115_n148;
  t115_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n147;
  t115_n146: return -0.026624485850334167f;
  t115_n147: return 0.0038211725186556578f;
  t115_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t115_n152;
  t115_n149: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t115_n151;
  t115_n150: return -0.00816869456321001f;
  t115_n151: return 0.06717725843191147f;
  t115_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t115_n156;
  t115_n153: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t115_n155;
  t115_n154: return 0.00946497917175293f;
  t115_n155: return 0.09087306261062622f;
  t115_n156: return 0.05480930209159851f;
  t115_n157: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t115_n173;
  t115_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t115_n160;
  t115_n159: return -0.018337048590183258f;
  t115_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t115_n162;
  t115_n161: return 0.005078161600977182f;
  t115_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t115_n172;
  t115_n163: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t115_n171;
  t115_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t115_n166;
  t115_n165: return 0.06486096978187561f;
  t115_n166: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t115_n170;
  t115_n167: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t115_n169;
  t115_n168: return -0.3104464113712311f;
  t115_n169: return 0.07604237645864487f;
  t115_n170: return 0.15025708079338074f;
  t115_n171: return 0.04142647981643677f;
  t115_n172: return 0.0386316180229187f;
  t115_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n175;
  t115_n174: return -0.004853751976042986f;
  t115_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t115_n177;
  t115_n176: return -0.03260929882526398f;
  t115_n177: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t115_n231;
  t115_n178: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t115_n226;
  t115_n179: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t115_n187;
  t115_n180: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t115_n182;
  t115_n181: return 0.44688060879707336f;
  t115_n182: if (E.getNumReferences() >= 3325837312 /*12048.0*/) goto t115_n186;
  t115_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n185;
  t115_n184: return -0.02427101694047451f;
  t115_n185: return 0.03043292835354805f;
  t115_n186: return 0.03700198978185654f;
  t115_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t115_n189;
  t115_n188: return 0.01171769481152296f;
  t115_n189: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t115_n225;
  t115_n190: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t115_n224;
  t115_n191: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t115_n209;
  t115_n192: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t115_n208;
  t115_n193: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t115_n205;
  t115_n194: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t115_n204;
  t115_n195: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t115_n203;
  t115_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t115_n202;
  t115_n197: if (E.getFractionNameInContext() >= 3179644278 /*0.0651959627866745*/) goto t115_n199;
  t115_n198: return 0.002115968381986022f;
  t115_n199: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t115_n201;
  t115_n200: return -0.21110863983631134f;
  t115_n201: return -0.032492995262145996f;
  t115_n202: return 0.04251403361558914f;
  t115_n203: return 0.1144401803612709f;
  t115_n204: return -0.045273538678884506f;
  t115_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t115_n207;
  t115_n206: return 0.03987306356430054f;
  t115_n207: return 0.05748530477285385f;
  t115_n208: return -0.07149477303028107f;
  t115_n209: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t115_n215;
  t115_n210: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t115_n214;
  t115_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t115_n213;
  t115_n212: return 0.05580928549170494f;
  t115_n213: return 0.06876808404922485f;
  t115_n214: return 0.11072784662246704f;
  t115_n215: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t115_n217;
  t115_n216: return -0.0841180831193924f;
  t115_n217: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t115_n221;
  t115_n218: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t115_n220;
  t115_n219: return 0.06669612973928452f;
  t115_n220: return 0.09330270439386368f;
  t115_n221: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t115_n223;
  t115_n222: return -0.16630074381828308f;
  t115_n223: return 0.05195528641343117f;
  t115_n224: return -0.10641330480575562f;
  t115_n225: return -0.008940648287534714f;
  t115_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t115_n230;
  t115_n227: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t115_n229;
  t115_n228: return -0.653803825378418f;
  t115_n229: return -0.04089927300810814f;
  t115_n230: return 0.03144896775484085f;
  t115_n231: return -0.010434816591441631f;
  t115_n232: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t115_n234;
  t115_n233: return 0.003266445128247142f;
  t115_n234: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t115_n254;
  t115_n235: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t115_n237;
  t115_n236: return 0.006169200409203768f;
  t115_n237: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t115_n253;
  t115_n238: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t115_n250;
  t115_n239: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t115_n243;
  t115_n240: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t115_n242;
  t115_n241: return 0.028964724391698837f;
  t115_n242: return 0.054281558841466904f;
  t115_n243: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t115_n245;
  t115_n244: return 0.0007276833639480174f;
  t115_n245: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t115_n249;
  t115_n246: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t115_n248;
  t115_n247: return 0.1383521556854248f;
  t115_n248: return 0.04151758924126625f;
  t115_n249: return 0.02709398791193962f;
  t115_n250: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t115_n252;
  t115_n251: return -0.2663571536540985f;
  t115_n252: return 0.021261323243379593f;
  t115_n253: return 0.030068762600421906f;
  t115_n254: return 0.038564249873161316f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree116(const Example& E) {
  t116_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t116_n170;
  t116_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t116_n163;
  t116_n2: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t116_n154;
  t116_n3: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t116_n123;
  t116_n4: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t116_n112;
  t116_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t116_n53;
  t116_n6: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t116_n20;
  t116_n7: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t116_n19;
  t116_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t116_n18;
  t116_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n17;
  t116_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t116_n12;
  t116_n11: return -0.11984774470329285f;
  t116_n12: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t116_n16;
  t116_n13: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t116_n15;
  t116_n14: return 0.006277382839471102f;
  t116_n15: return 0.23454400897026062f;
  t116_n16: return -0.037439025938510895f;
  t116_n17: return 0.0056553916074335575f;
  t116_n18: return -0.009560362435877323f;
  t116_n19: return 0.011509270407259464f;
  t116_n20: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t116_n46;
  t116_n21: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t116_n23;
  t116_n22: return -0.00632026931270957f;
  t116_n23: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t116_n43;
  t116_n24: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t116_n26;
  t116_n25: return -0.009108525700867176f;
  t116_n26: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t116_n42;
  t116_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t116_n29;
  t116_n28: return -0.027257971465587616f;
  t116_n29: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t116_n39;
  t116_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t116_n38;
  t116_n31: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t116_n37;
  t116_n32: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t116_n34;
  t116_n33: return 0.003286819439381361f;
  t116_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n36;
  t116_n35: return -0.08061004430055618f;
  t116_n36: return 0.009072203189134598f;
  t116_n37: return 0.017459888011217117f;
  t116_n38: return 0.00878145545721054f;
  t116_n39: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t116_n41;
  t116_n40: return 0.020065436139702797f;
  t116_n41: return -0.013459459878504276f;
  t116_n42: return -0.11125019937753677f;
  t116_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n45;
  t116_n44: return -0.10627637058496475f;
  t116_n45: return -0.0035379056353121996f;
  t116_n46: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t116_n52;
  t116_n47: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t116_n51;
  t116_n48: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t116_n50;
  t116_n49: return -0.0272140484303236f;
  t116_n50: return -0.06065600365400314f;
  t116_n51: return 0.010125871747732162f;
  t116_n52: return -0.04966650530695915f;
  t116_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t116_n75;
  t116_n54: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t116_n62;
  t116_n55: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t116_n61;
  t116_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t116_n60;
  t116_n57: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t116_n59;
  t116_n58: return -0.00508191529661417f;
  t116_n59: return 0.02574605494737625f;
  t116_n60: return 0.0661061480641365f;
  t116_n61: return 0.03723578900098801f;
  t116_n62: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t116_n72;
  t116_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n71;
  t116_n64: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t116_n70;
  t116_n65: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t116_n69;
  t116_n66: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t116_n68;
  t116_n67: return -0.011195512488484383f;
  t116_n68: return 0.08839742094278336f;
  t116_n69: return -0.05645196884870529f;
  t116_n70: return -0.048508137464523315f;
  t116_n71: return 0.001427146140486002f;
  t116_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n74;
  t116_n73: return -0.020923377946019173f;
  t116_n74: return 0.007041463162750006f;
  t116_n75: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t116_n97;
  t116_n76: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t116_n96;
  t116_n77: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t116_n95;
  t116_n78: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t116_n82;
  t116_n79: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t116_n81;
  t116_n80: return -0.01359616033732891f;
  t116_n81: return -0.05222376435995102f;
  t116_n82: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t116_n94;
  t116_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t116_n85;
  t116_n84: return -0.001596741029061377f;
  t116_n85: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t116_n89;
  t116_n86: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t116_n88;
  t116_n87: return 0.039442721754312515f;
  t116_n88: return -0.18361438810825348f;
  t116_n89: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t116_n93;
  t116_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n92;
  t116_n91: return -0.048958275467157364f;
  t116_n92: return 0.07837552577257156f;
  t116_n93: return 0.028526129201054573f;
  t116_n94: return -0.004671327769756317f;
  t116_n95: return -0.034814368933439255f;
  t116_n96: return 0.003790962975472212f;
  t116_n97: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t116_n109;
  t116_n98: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t116_n106;
  t116_n99: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t116_n103;
  t116_n100: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t116_n102;
  t116_n101: return 0.0066881682723760605f;
  t116_n102: return -0.024188650771975517f;
  t116_n103: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t116_n105;
  t116_n104: return 0.10314779728651047f;
  t116_n105: return 0.14444266259670258f;
  t116_n106: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t116_n108;
  t116_n107: return 0.03338591754436493f;
  t116_n108: return -0.01573362573981285f;
  t116_n109: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t116_n111;
  t116_n110: return 0.09570655971765518f;
  t116_n111: return 0.5769132971763611f;
  t116_n112: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t116_n120;
  t116_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n115;
  t116_n114: return -0.021552059799432755f;
  t116_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t116_n119;
  t116_n116: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t116_n118;
  t116_n117: return 0.010975666344165802f;
  t116_n118: return -0.1357920616865158f;
  t116_n119: return 0.28027990460395813f;
  t116_n120: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t116_n122;
  t116_n121: return 0.0037326014135032892f;
  t116_n122: return -0.010705679655075073f;
  t116_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t116_n145;
  t116_n124: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t116_n128;
  t116_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t116_n127;
  t116_n126: return -0.046103715896606445f;
  t116_n127: return 0.0016946722753345966f;
  t116_n128: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t116_n132;
  t116_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n131;
  t116_n130: return -0.010841659270226955f;
  t116_n131: return -0.0005582079757004976f;
  t116_n132: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t116_n136;
  t116_n133: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t116_n135;
  t116_n134: return 0.09732616692781448f;
  t116_n135: return 0.014116169884800911f;
  t116_n136: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t116_n144;
  t116_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n143;
  t116_n138: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t116_n140;
  t116_n139: return -0.008958615362644196f;
  t116_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n142;
  t116_n141: return -0.056343816220760345f;
  t116_n142: return 0.03670073673129082f;
  t116_n143: return 0.015673911198973656f;
  t116_n144: return 0.02308502234518528f;
  t116_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t116_n151;
  t116_n146: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t116_n150;
  t116_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n149;
  t116_n148: return -0.15067651867866516f;
  t116_n149: return 0.024025319144129753f;
  t116_n150: return 0.006072433665394783f;
  t116_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t116_n153;
  t116_n152: return 0.011985131539404392f;
  t116_n153: return 0.04789121821522713f;
  t116_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t116_n162;
  t116_n155: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t116_n161;
  t116_n156: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t116_n160;
  t116_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n159;
  t116_n158: return -0.07510533928871155f;
  t116_n159: return -0.026261569932103157f;
  t116_n160: return 0.0520847924053669f;
  t116_n161: return -0.005660246592015028f;
  t116_n162: return 0.0378614105284214f;
  t116_n163: if (E.getNumReferences() >= 3318212608 /*6397.0*/) goto t116_n167;
  t116_n164: if (E.getNumReferences() >= 3316799488 /*5707.0*/) goto t116_n166;
  t116_n165: return 0.0051532625220716f;
  t116_n166: return -0.7200466990470886f;
  t116_n167: if (E.getNumReferences() >= 3338303744 /*32060.5*/) goto t116_n169;
  t116_n168: return 0.07083117216825485f;
  t116_n169: return -0.15985314548015594f;
  t116_n170: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t116_n252;
  t116_n171: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t116_n219;
  t116_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n188;
  t116_n173: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t116_n177;
  t116_n174: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t116_n176;
  t116_n175: return -0.04984248802065849f;
  t116_n176: return 0.044253457337617874f;
  t116_n177: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t116_n181;
  t116_n178: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t116_n180;
  t116_n179: return -0.02417709492146969f;
  t116_n180: return -0.4075770676136017f;
  t116_n181: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t116_n183;
  t116_n182: return 0.02341504395008087f;
  t116_n183: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t116_n187;
  t116_n184: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t116_n186;
  t116_n185: return 0.10166887938976288f;
  t116_n186: return 0.06298190355300903f;
  t116_n187: return 0.046441446989774704f;
  t116_n188: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t116_n190;
  t116_n189: return -0.019803214818239212f;
  t116_n190: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t116_n194;
  t116_n191: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t116_n193;
  t116_n192: return -0.02587788738310337f;
  t116_n193: return 0.017395202070474625f;
  t116_n194: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t116_n216;
  t116_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n199;
  t116_n196: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t116_n198;
  t116_n197: return -0.025959137827157974f;
  t116_n198: return 0.005285854917019606f;
  t116_n199: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t116_n207;
  t116_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n206;
  t116_n201: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t116_n205;
  t116_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n204;
  t116_n203: return -0.02408573031425476f;
  t116_n204: return 0.08805765956640244f;
  t116_n205: return 0.01062039379030466f;
  t116_n206: return 0.05736422911286354f;
  t116_n207: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t116_n215;
  t116_n208: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t116_n214;
  t116_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t116_n213;
  t116_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n212;
  t116_n211: return -0.09287606924772263f;
  t116_n212: return 0.05256623029708862f;
  t116_n213: return 0.02909817360341549f;
  t116_n214: return 0.013451624661684036f;
  t116_n215: return -0.0018055540276691318f;
  t116_n216: if (E.getFractionNameInContext() >= 3208871340 /*0.7636363506317139*/) goto t116_n218;
  t116_n217: return -0.11970432847738266f;
  t116_n218: return 0.14101910591125488f;
  t116_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n249;
  t116_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n222;
  t116_n221: return -0.02883034385740757f;
  t116_n222: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t116_n236;
  t116_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t116_n233;
  t116_n224: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t116_n232;
  t116_n225: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t116_n227;
  t116_n226: return -0.00036499870475381613f;
  t116_n227: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t116_n229;
  t116_n228: return -0.12404294312000275f;
  t116_n229: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t116_n231;
  t116_n230: return -0.002662813290953636f;
  t116_n231: return -0.11296822875738144f;
  t116_n232: return 0.09515631943941116f;
  t116_n233: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t116_n235;
  t116_n234: return 0.006080983206629753f;
  t116_n235: return 0.17104394733905792f;
  t116_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t116_n238;
  t116_n237: return 0.01682494953274727f;
  t116_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t116_n244;
  t116_n239: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t116_n241;
  t116_n240: return 0.012864925898611546f;
  t116_n241: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t116_n243;
  t116_n242: return -0.16797183454036713f;
  t116_n243: return 0.06566726416349411f;
  t116_n244: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t116_n248;
  t116_n245: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t116_n247;
  t116_n246: return 0.03391556441783905f;
  t116_n247: return 0.01177766639739275f;
  t116_n248: return 0.017442788928747177f;
  t116_n249: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t116_n251;
  t116_n250: return 0.03056401014328003f;
  t116_n251: return 0.10005796700716019f;
  t116_n252: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t116_n254;
  t116_n253: return -0.13522975146770477f;
  t116_n254: return -0.013435802422463894f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree117(const Example& E) {
  t117_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t117_n178;
  t117_n1: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t117_n161;
  t117_n2: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t117_n160;
  t117_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t117_n69;
  t117_n4: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t117_n54;
  t117_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t117_n41;
  t117_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t117_n10;
  t117_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t117_n9;
  t117_n8: return -0.11161623895168304f;
  t117_n9: return 0.07646918296813965f;
  t117_n10: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t117_n26;
  t117_n11: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t117_n17;
  t117_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t117_n14;
  t117_n13: return -0.022950466722249985f;
  t117_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t117_n16;
  t117_n15: return 0.002171476371586323f;
  t117_n16: return 0.024981966242194176f;
  t117_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t117_n25;
  t117_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t117_n22;
  t117_n19: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t117_n21;
  t117_n20: return -0.09755988419055939f;
  t117_n21: return -0.04432089999318123f;
  t117_n22: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t117_n24;
  t117_n23: return 0.05024867504835129f;
  t117_n24: return 0.010452774353325367f;
  t117_n25: return -0.004132767673581839f;
  t117_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t117_n36;
  t117_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t117_n35;
  t117_n28: if (E.getFileProximityDistanceCost() >= 3255828480 /*36.0*/) goto t117_n34;
  t117_n29: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t117_n33;
  t117_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t117_n32;
  t117_n31: return -0.10196114331483841f;
  t117_n32: return 0.15513460338115692f;
  t117_n33: return -0.12110237032175064f;
  t117_n34: return -0.08166028559207916f;
  t117_n35: return 0.008745999075472355f;
  t117_n36: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t117_n38;
  t117_n37: return 0.009259535931050777f;
  t117_n38: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t117_n40;
  t117_n39: return 0.07497917860746384f;
  t117_n40: return -0.1747518926858902f;
  t117_n41: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t117_n43;
  t117_n42: return 0.0014692749828100204f;
  t117_n43: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t117_n53;
  t117_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t117_n46;
  t117_n45: return -0.02101106382906437f;
  t117_n46: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t117_n52;
  t117_n47: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t117_n51;
  t117_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t117_n50;
  t117_n49: return -0.12317882478237152f;
  t117_n50: return 0.07303038984537125f;
  t117_n51: return -0.07157295197248459f;
  t117_n52: return 0.3065928518772125f;
  t117_n53: return -0.11476002633571625f;
  t117_n54: if (E.getNumReferences() >= 3337393664 /*30283.0*/) goto t117_n68;
  t117_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t117_n67;
  t117_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t117_n60;
  t117_n57: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t117_n59;
  t117_n58: return -0.02888481505215168f;
  t117_n59: return 0.22726672887802124f;
  t117_n60: if (E.getNumReferences() >= 3315876864 /*5256.5*/) goto t117_n62;
  t117_n61: return 0.01478923112154007f;
  t117_n62: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t117_n66;
  t117_n63: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t117_n65;
  t117_n64: return 0.10425281524658203f;
  t117_n65: return 0.13895611464977264f;
  t117_n66: return 0.08130759000778198f;
  t117_n67: return 0.008086415007710457f;
  t117_n68: return -0.04286159202456474f;
  t117_n69: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t117_n81;
  t117_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t117_n72;
  t117_n71: return 0.0037840669974684715f;
  t117_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n80;
  t117_n73: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t117_n79;
  t117_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t117_n78;
  t117_n75: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t117_n77;
  t117_n76: return -0.11325494945049286f;
  t117_n77: return 0.121756911277771f;
  t117_n78: return 0.09739834815263748f;
  t117_n79: return 0.015795178711414337f;
  t117_n80: return 0.03303470462560654f;
  t117_n81: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t117_n83;
  t117_n82: return -0.0015229604905471206f;
  t117_n83: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t117_n89;
  t117_n84: if (E.getNumReferences() >= 3316092928 /*5362.0*/) goto t117_n86;
  t117_n85: return 0.03673520311713219f;
  t117_n86: if (E.getNumReferences() >= 3330759936 /*17326.5*/) goto t117_n88;
  t117_n87: return -0.14537020027637482f;
  t117_n88: return 0.007889414206147194f;
  t117_n89: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t117_n115;
  t117_n90: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t117_n102;
  t117_n91: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t117_n93;
  t117_n92: return -0.0580938458442688f;
  t117_n93: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t117_n101;
  t117_n94: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t117_n96;
  t117_n95: return 0.13727393746376038f;
  t117_n96: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t117_n100;
  t117_n97: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t117_n99;
  t117_n98: return 0.4255756437778473f;
  t117_n99: return 0.1617254912853241f;
  t117_n100: return -0.5440766215324402f;
  t117_n101: return 0.13623186945915222f;
  t117_n102: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t117_n114;
  t117_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n111;
  t117_n104: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t117_n108;
  t117_n105: if (E.getNumReferences() >= 3337319680 /*30138.5*/) goto t117_n107;
  t117_n106: return 0.037219054996967316f;
  t117_n107: return 0.11208869516849518f;
  t117_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t117_n110;
  t117_n109: return -0.0003146169474348426f;
  t117_n110: return 0.04003649204969406f;
  t117_n111: if (E.getNumReferences() >= 3298836480 /*1282.0*/) goto t117_n113;
  t117_n112: return 0.05179620906710625f;
  t117_n113: return -0.015965070575475693f;
  t117_n114: return -0.009835162200033665f;
  t117_n115: if (E.getNumReferences() >= 3334364160 /*24366.0*/) goto t117_n147;
  t117_n116: if (E.getNumReferences() >= 3290177536 /*624.5*/) goto t117_n132;
  t117_n117: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t117_n119;
  t117_n118: return -0.02539379708468914f;
  t117_n119: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t117_n131;
  t117_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t117_n128;
  t117_n121: if (E.getNumReferences() >= 3281338368 /*298.5*/) goto t117_n127;
  t117_n122: if (E.getNumReferences() >= 3267100672 /*94.0*/) goto t117_n124;
  t117_n123: return 0.03079182095825672f;
  t117_n124: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t117_n126;
  t117_n125: return 0.14943742752075195f;
  t117_n126: return 0.11261320114135742f;
  t117_n127: return -0.09304868429899216f;
  t117_n128: if (E.getNumReferences() >= 3283304448 /*358.5*/) goto t117_n130;
  t117_n129: return 0.11662428081035614f;
  t117_n130: return 0.15576785802841187f;
  t117_n131: return 0.009076167829334736f;
  t117_n132: if (E.getNumReferences() >= 3294699520 /*900.5*/) goto t117_n146;
  t117_n133: if (E.getNumReferences() >= 3294683136 /*899.5*/) goto t117_n145;
  t117_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t117_n136;
  t117_n135: return 0.028134338557720184f;
  t117_n136: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t117_n138;
  t117_n137: return -0.004440354648977518f;
  t117_n138: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t117_n144;
  t117_n139: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t117_n141;
  t117_n140: return 0.14131474494934082f;
  t117_n141: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t117_n143;
  t117_n142: return 0.11614417284727097f;
  t117_n143: return -0.14976342022418976f;
  t117_n144: return 0.09157435595989227f;
  t117_n145: return 0.11554060131311417f;
  t117_n146: return 0.03369787335395813f;
  t117_n147: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t117_n155;
  t117_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t117_n150;
  t117_n149: return 0.008990845642983913f;
  t117_n150: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t117_n152;
  t117_n151: return 0.11834882199764252f;
  t117_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t117_n154;
  t117_n153: return 0.11186284571886063f;
  t117_n154: return 0.1098862737417221f;
  t117_n155: if (E.getNumReferences() >= 3334782720 /*25183.5*/) goto t117_n157;
  t117_n156: return -0.22537539899349213f;
  t117_n157: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t117_n159;
  t117_n158: return 0.019069727510213852f;
  t117_n159: return -0.10854774713516235f;
  t117_n160: return -0.029997318983078003f;
  t117_n161: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t117_n173;
  t117_n162: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t117_n168;
  t117_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n165;
  t117_n164: return 0.05861743167042732f;
  t117_n165: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t117_n167;
  t117_n166: return -0.116947703063488f;
  t117_n167: return 0.1266273558139801f;
  t117_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n170;
  t117_n169: return 0.1253153532743454f;
  t117_n170: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t117_n172;
  t117_n171: return 0.1140662357211113f;
  t117_n172: return 0.13228189945220947f;
  t117_n173: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t117_n175;
  t117_n174: return 0.019435415044426918f;
  t117_n175: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t117_n177;
  t117_n176: return -0.02965734153985977f;
  t117_n177: return 0.00557591812685132f;
  t117_n178: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t117_n254;
  t117_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t117_n231;
  t117_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n188;
  t117_n181: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t117_n183;
  t117_n182: return -0.0026789435651153326f;
  t117_n183: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t117_n185;
  t117_n184: return 0.018888289108872414f;
  t117_n185: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t117_n187;
  t117_n186: return 0.1228613629937172f;
  t117_n187: return 0.1329728215932846f;
  t117_n188: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t117_n194;
  t117_n189: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t117_n191;
  t117_n190: return -0.014946049079298973f;
  t117_n191: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t117_n193;
  t117_n192: return -0.28786662220954895f;
  t117_n193: return 0.14320723712444305f;
  t117_n194: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t117_n226;
  t117_n195: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t117_n197;
  t117_n196: return 0.0013410650426521897f;
  t117_n197: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t117_n219;
  t117_n198: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t117_n218;
  t117_n199: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t117_n211;
  t117_n200: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t117_n208;
  t117_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t117_n203;
  t117_n202: return 0.05806298181414604f;
  t117_n203: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t117_n205;
  t117_n204: return -0.2723483145236969f;
  t117_n205: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t117_n207;
  t117_n206: return 0.19867251813411713f;
  t117_n207: return 0.09329487383365631f;
  t117_n208: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t117_n210;
  t117_n209: return 0.13496190309524536f;
  t117_n210: return 0.02161310985684395f;
  t117_n211: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t117_n217;
  t117_n212: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t117_n214;
  t117_n213: return 0.2674146890640259f;
  t117_n214: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t117_n216;
  t117_n215: return -0.040771082043647766f;
  t117_n216: return 0.10777938365936279f;
  t117_n217: return 0.214747816324234f;
  t117_n218: return 0.009748296812176704f;
  t117_n219: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t117_n225;
  t117_n220: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t117_n224;
  t117_n221: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t117_n223;
  t117_n222: return 0.1309615969657898f;
  t117_n223: return 0.14946967363357544f;
  t117_n224: return -0.6009328961372375f;
  t117_n225: return 0.01995094120502472f;
  t117_n226: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t117_n228;
  t117_n227: return -0.12690071761608124f;
  t117_n228: if (E.getNumReferences() >= 3331587584 /*18943.0*/) goto t117_n230;
  t117_n229: return 0.013897327706217766f;
  t117_n230: return -0.3198107182979584f;
  t117_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t117_n253;
  t117_n232: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t117_n236;
  t117_n233: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t117_n235;
  t117_n234: return -0.035557106137275696f;
  t117_n235: return 0.1295512467622757f;
  t117_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t117_n252;
  t117_n237: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t117_n239;
  t117_n238: return -0.006584179122000933f;
  t117_n239: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t117_n243;
  t117_n240: if (E.getNumReferences() >= 3294420992 /*883.5*/) goto t117_n242;
  t117_n241: return 0.002858088817447424f;
  t117_n242: return 0.02860281802713871f;
  t117_n243: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t117_n245;
  t117_n244: return 0.04239572584629059f;
  t117_n245: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t117_n251;
  t117_n246: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t117_n250;
  t117_n247: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t117_n249;
  t117_n248: return 0.07547067850828171f;
  t117_n249: return 0.13320571184158325f;
  t117_n250: return -0.03337264433503151f;
  t117_n251: return -0.017634477466344833f;
  t117_n252: return 0.008583194576203823f;
  t117_n253: return 0.007611863315105438f;
  t117_n254: return 0.009543932043015957f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree118(const Example& E) {
  t118_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t118_n188;
  t118_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t118_n51;
  t118_n2: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t118_n36;
  t118_n3: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t118_n35;
  t118_n4: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t118_n22;
  t118_n5: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t118_n21;
  t118_n6: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t118_n8;
  t118_n7: return 0.13790592551231384f;
  t118_n8: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t118_n20;
  t118_n9: if (E.getNumReferences() >= 3354076032 /*120391.0*/) goto t118_n13;
  t118_n10: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t118_n12;
  t118_n11: return -0.0003491016977932304f;
  t118_n12: return 0.007268888875842094f;
  t118_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t118_n15;
  t118_n14: return 0.018732575699687004f;
  t118_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t118_n17;
  t118_n16: return -0.05204692482948303f;
  t118_n17: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t118_n19;
  t118_n18: return 0.11707659810781479f;
  t118_n19: return -0.0036064675077795982f;
  t118_n20: return -0.003960867412388325f;
  t118_n21: return -0.14554139971733093f;
  t118_n22: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t118_n24;
  t118_n23: return 0.02931390330195427f;
  t118_n24: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t118_n26;
  t118_n25: return -0.12291167676448822f;
  t118_n26: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t118_n34;
  t118_n27: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t118_n29;
  t118_n28: return 0.021490395069122314f;
  t118_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t118_n31;
  t118_n30: return 0.06286702305078506f;
  t118_n31: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t118_n33;
  t118_n32: return 0.09152686595916748f;
  t118_n33: return 0.16154368221759796f;
  t118_n34: return -0.15382398664951324f;
  t118_n35: return -0.05679338425397873f;
  t118_n36: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t118_n40;
  t118_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t118_n39;
  t118_n38: return 0.0017929807072505355f;
  t118_n39: return 0.061067450791597366f;
  t118_n40: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t118_n42;
  t118_n41: return -0.09290846437215805f;
  t118_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t118_n50;
  t118_n43: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t118_n45;
  t118_n44: return -0.04522987827658653f;
  t118_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t118_n49;
  t118_n46: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t118_n48;
  t118_n47: return 0.08401937782764435f;
  t118_n48: return -0.23522576689720154f;
  t118_n49: return 0.016717860475182533f;
  t118_n50: return 0.0352255217730999f;
  t118_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t118_n69;
  t118_n52: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t118_n54;
  t118_n53: return -0.001498684287071228f;
  t118_n54: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t118_n68;
  t118_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t118_n63;
  t118_n56: if (E.getNumReferences() >= 3314693120 /*4678.5*/) goto t118_n58;
  t118_n57: return 0.005529620219022036f;
  t118_n58: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t118_n62;
  t118_n59: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t118_n61;
  t118_n60: return 0.11274608969688416f;
  t118_n61: return -0.005587033927440643f;
  t118_n62: return 0.02531440742313862f;
  t118_n63: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t118_n65;
  t118_n64: return 0.043110478669404984f;
  t118_n65: if (E.getNumReferences() >= 3248488448 /*20.0*/) goto t118_n67;
  t118_n66: return 0.14458730816841125f;
  t118_n67: return 0.00613754615187645f;
  t118_n68: return -0.0005358494236133993f;
  t118_n69: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t118_n121;
  t118_n70: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t118_n74;
  t118_n71: if (E.getNumReferences() >= 3293822976 /*847.0*/) goto t118_n73;
  t118_n72: return 0.005741099826991558f;
  t118_n73: return -0.03564049303531647f;
  t118_n74: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t118_n96;
  t118_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t118_n89;
  t118_n76: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t118_n84;
  t118_n77: if (E.getNumReferences() >= 3289104384 /*559.0*/) goto t118_n83;
  t118_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t118_n80;
  t118_n79: return -0.07851005345582962f;
  t118_n80: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t118_n82;
  t118_n81: return 0.0025163383688777685f;
  t118_n82: return -0.04110982269048691f;
  t118_n83: return -0.07608204334974289f;
  t118_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t118_n88;
  t118_n85: if (E.getFractionNameInContext() >= 3208032480 /*0.7136363983154297*/) goto t118_n87;
  t118_n86: return 0.02860211580991745f;
  t118_n87: return 0.14762859046459198f;
  t118_n88: return 0.10896484553813934f;
  t118_n89: if (E.getNumReferences() >= 3355784000 /*136397.0*/) goto t118_n91;
  t118_n90: return 0.0037608256097882986f;
  t118_n91: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t118_n93;
  t118_n92: return 0.056426744908094406f;
  t118_n93: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t118_n95;
  t118_n94: return -0.07368141412734985f;
  t118_n95: return 0.10354543477296829f;
  t118_n96: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t118_n102;
  t118_n97: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t118_n101;
  t118_n98: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t118_n100;
  t118_n99: return -0.006368858739733696f;
  t118_n100: return 0.16128666698932648f;
  t118_n101: return -0.0725456029176712f;
  t118_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t118_n104;
  t118_n103: return 0.0072293588891625404f;
  t118_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t118_n108;
  t118_n105: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t118_n107;
  t118_n106: return 0.3622494637966156f;
  t118_n107: return -0.04442960396409035f;
  t118_n108: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t118_n116;
  t118_n109: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t118_n113;
  t118_n110: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t118_n112;
  t118_n111: return 0.05609910190105438f;
  t118_n112: return 0.12185179442167282f;
  t118_n113: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t118_n115;
  t118_n114: return 0.06387511640787125f;
  t118_n115: return 0.027674494311213493f;
  t118_n116: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t118_n120;
  t118_n117: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t118_n119;
  t118_n118: return -0.40407535433769226f;
  t118_n119: return -0.029834842309355736f;
  t118_n120: return 0.023969655856490135f;
  t118_n121: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t118_n185;
  t118_n122: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t118_n160;
  t118_n123: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t118_n149;
  t118_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t118_n130;
  t118_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t118_n127;
  t118_n126: return -0.021582426503300667f;
  t118_n127: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t118_n129;
  t118_n128: return -0.20217740535736084f;
  t118_n129: return 0.00510704331099987f;
  t118_n130: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t118_n132;
  t118_n131: return -0.011956813745200634f;
  t118_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t118_n148;
  t118_n133: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t118_n143;
  t118_n134: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t118_n140;
  t118_n135: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t118_n137;
  t118_n136: return 0.07311877608299255f;
  t118_n137: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t118_n139;
  t118_n138: return 0.027315225452184677f;
  t118_n139: return -0.20463843643665314f;
  t118_n140: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t118_n142;
  t118_n141: return 0.04544103890657425f;
  t118_n142: return 0.17540882527828217f;
  t118_n143: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t118_n147;
  t118_n144: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t118_n146;
  t118_n145: return 0.1299007534980774f;
  t118_n146: return 0.026401720941066742f;
  t118_n147: return 0.14680978655815125f;
  t118_n148: return -0.019660573452711105f;
  t118_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement))) goto t118_n155;
  t118_n150: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t118_n152;
  t118_n151: return 0.2535754442214966f;
  t118_n152: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t118_n154;
  t118_n153: return 0.011456739157438278f;
  t118_n154: return 0.1193501278758049f;
  t118_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t118_n157;
  t118_n156: return -0.17195908725261688f;
  t118_n157: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t118_n159;
  t118_n158: return -0.019758492708206177f;
  t118_n159: return 0.02700350433588028f;
  t118_n160: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t118_n182;
  t118_n161: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t118_n181;
  t118_n162: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t118_n176;
  t118_n163: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t118_n165;
  t118_n164: return 0.020540615543723106f;
  t118_n165: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t118_n171;
  t118_n166: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t118_n168;
  t118_n167: return -0.10041139274835587f;
  t118_n168: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t118_n170;
  t118_n169: return 0.15504467487335205f;
  t118_n170: return -0.018913695588707924f;
  t118_n171: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t118_n175;
  t118_n172: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t118_n174;
  t118_n173: return 0.2504871189594269f;
  t118_n174: return -0.18055163323879242f;
  t118_n175: return 0.08937636762857437f;
  t118_n176: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t118_n180;
  t118_n177: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t118_n179;
  t118_n178: return -0.00675966776907444f;
  t118_n179: return -0.23978909850120544f;
  t118_n180: return 0.09382099658250809f;
  t118_n181: return -0.11237724125385284f;
  t118_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t118_n184;
  t118_n183: return -0.0010205277940258384f;
  t118_n184: return 0.015321588143706322f;
  t118_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t118_n187;
  t118_n186: return -0.005809381604194641f;
  t118_n187: return 0.006905480287969112f;
  t118_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t118_n244;
  t118_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t118_n191;
  t118_n190: return -0.14249439537525177f;
  t118_n191: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t118_n241;
  t118_n192: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t118_n226;
  t118_n193: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t118_n197;
  t118_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t118_n196;
  t118_n195: return -0.004428786225616932f;
  t118_n196: return 0.009597317315638065f;
  t118_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t118_n203;
  t118_n198: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t118_n202;
  t118_n199: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t118_n201;
  t118_n200: return 0.03821362927556038f;
  t118_n201: return -0.2146500200033188f;
  t118_n202: return -0.27643823623657227f;
  t118_n203: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t118_n211;
  t118_n204: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t118_n208;
  t118_n205: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t118_n207;
  t118_n206: return 0.06235335022211075f;
  t118_n207: return 0.10244576632976532f;
  t118_n208: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t118_n210;
  t118_n209: return 0.13373003900051117f;
  t118_n210: return 0.05804726481437683f;
  t118_n211: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t118_n225;
  t118_n212: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t118_n218;
  t118_n213: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t118_n217;
  t118_n214: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t118_n216;
  t118_n215: return -0.5818341374397278f;
  t118_n216: return -0.07452434301376343f;
  t118_n217: return -0.0023343663197010756f;
  t118_n218: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t118_n224;
  t118_n219: if (E.getFractionNameInContext() >= 3205613340 /*0.5694444179534912*/) goto t118_n223;
  t118_n220: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t118_n222;
  t118_n221: return 0.11314737051725388f;
  t118_n222: return 0.014830718748271465f;
  t118_n223: return 0.035629913210868835f;
  t118_n224: return -0.15276655554771423f;
  t118_n225: return 0.1367075890302658f;
  t118_n226: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t118_n228;
  t118_n227: return -0.0015508814249187708f;
  t118_n228: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t118_n230;
  t118_n229: return 0.005889330059289932f;
  t118_n230: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t118_n236;
  t118_n231: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t118_n233;
  t118_n232: return 0.09129638224840164f;
  t118_n233: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t118_n235;
  t118_n234: return 0.4264606833457947f;
  t118_n235: return 0.1014535054564476f;
  t118_n236: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t118_n238;
  t118_n237: return -0.04473527893424034f;
  t118_n238: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t118_n240;
  t118_n239: return 0.26591816544532776f;
  t118_n240: return 0.023290302604436874f;
  t118_n241: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t118_n243;
  t118_n242: return 0.014848681166768074f;
  t118_n243: return 0.1632160246372223f;
  t118_n244: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t118_n254;
  t118_n245: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t118_n247;
  t118_n246: return 0.0070661697536706924f;
  t118_n247: if (E.getNumReferences() >= 3306696704 /*2435.0*/) goto t118_n249;
  t118_n248: return 0.11818220466375351f;
  t118_n249: if (E.getNumReferences() >= 3309611008 /*3146.5*/) goto t118_n253;
  t118_n250: if (E.getNumReferences() >= 3309430784 /*3102.5*/) goto t118_n252;
  t118_n251: return 0.0767441913485527f;
  t118_n252: return 0.12950442731380463f;
  t118_n253: return 0.031312234699726105f;
  t118_n254: return -0.07932544499635696f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree119(const Example& E) {
  t119_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t119_n200;
  t119_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t119_n119;
  t119_n2: if (E.getNumReferences() >= 3305713664 /*2195.0*/) goto t119_n42;
  t119_n3: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t119_n33;
  t119_n4: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t119_n8;
  t119_n5: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t119_n7;
  t119_n6: return 0.13489805161952972f;
  t119_n7: return -0.0016057188622653484f;
  t119_n8: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t119_n32;
  t119_n9: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t119_n29;
  t119_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n12;
  t119_n11: return -0.03987665846943855f;
  t119_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t119_n14;
  t119_n13: return -0.01896427944302559f;
  t119_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t119_n26;
  t119_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n21;
  t119_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t119_n20;
  t119_n17: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t119_n19;
  t119_n18: return -0.013607860542833805f;
  t119_n19: return 0.013254859484732151f;
  t119_n20: return -0.0029287291690707207f;
  t119_n21: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t119_n23;
  t119_n22: return -0.04808923974633217f;
  t119_n23: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t119_n25;
  t119_n24: return -0.1455874890089035f;
  t119_n25: return 0.03538259118795395f;
  t119_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n28;
  t119_n27: return -0.017716599628329277f;
  t119_n28: return 0.023761725053191185f;
  t119_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n31;
  t119_n30: return -0.008986982516944408f;
  t119_n31: return 0.0035631044302135706f;
  t119_n32: return -0.12443819642066956f;
  t119_n33: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t119_n41;
  t119_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t119_n40;
  t119_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n37;
  t119_n36: return -0.007310532499104738f;
  t119_n37: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t119_n39;
  t119_n38: return 0.024424491450190544f;
  t119_n39: return 0.11467655003070831f;
  t119_n40: return 0.0025168994907289743f;
  t119_n41: return -0.007896837778389454f;
  t119_n42: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t119_n48;
  t119_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n45;
  t119_n44: return -0.09799113869667053f;
  t119_n45: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n47;
  t119_n46: return 0.016401216387748718f;
  t119_n47: return -0.0029453637544065714f;
  t119_n48: if (E.getNumReferences() >= 3316280320 /*5453.5*/) goto t119_n88;
  t119_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t119_n51;
  t119_n50: return -0.0343179889023304f;
  t119_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t119_n53;
  t119_n52: return 0.018736520782113075f;
  t119_n53: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t119_n63;
  t119_n54: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t119_n56;
  t119_n55: return 0.17129550874233246f;
  t119_n56: if (E.getNumReferences() >= 3306795008 /*2459.0*/) goto t119_n62;
  t119_n57: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t119_n59;
  t119_n58: return 0.04814969003200531f;
  t119_n59: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t119_n61;
  t119_n60: return 0.1811632215976715f;
  t119_n61: return 0.1127946600317955f;
  t119_n62: return 0.01605847105383873f;
  t119_n63: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t119_n65;
  t119_n64: return 0.16216635704040527f;
  t119_n65: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t119_n87;
  t119_n66: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t119_n68;
  t119_n67: return 0.025991687551140785f;
  t119_n68: if (E.getNumReferences() >= 3313659904 /*4174.0*/) goto t119_n72;
  t119_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t119_n71;
  t119_n70: return 0.10520297288894653f;
  t119_n71: return 0.17494872212409973f;
  t119_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t119_n78;
  t119_n73: if (E.getNumReferences() >= 3313923072 /*4302.5*/) goto t119_n77;
  t119_n74: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t119_n76;
  t119_n75: return 0.14764991402626038f;
  t119_n76: return 0.1424681395292282f;
  t119_n77: return 0.06765639036893845f;
  t119_n78: if (E.getNumReferences() >= 3315931136 /*5283.0*/) goto t119_n84;
  t119_n79: if (E.getNumReferences() >= 3314120704 /*4399.0*/) goto t119_n81;
  t119_n80: return -0.15222792327404022f;
  t119_n81: if (E.getNumReferences() >= 3314141184 /*4409.0*/) goto t119_n83;
  t119_n82: return 0.23694580793380737f;
  t119_n83: return 0.0861191600561142f;
  t119_n84: if (E.getNumReferences() >= 3316045824 /*5339.0*/) goto t119_n86;
  t119_n85: return 0.13327419757843018f;
  t119_n86: return -0.048991765826940536f;
  t119_n87: return 0.037627581506967545f;
  t119_n88: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t119_n90;
  t119_n89: return -0.04705028980970383f;
  t119_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t119_n92;
  t119_n91: return -0.0008097323006950319f;
  t119_n92: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n118;
  t119_n93: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t119_n95;
  t119_n94: return 0.010975620709359646f;
  t119_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t119_n97;
  t119_n96: return -0.048882707953453064f;
  t119_n97: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t119_n117;
  t119_n98: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t119_n100;
  t119_n99: return 0.055542539805173874f;
  t119_n100: if (E.getNumReferences() >= 3333346816 /*22379.0*/) goto t119_n112;
  t119_n101: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t119_n111;
  t119_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t119_n106;
  t119_n103: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t119_n105;
  t119_n104: return 0.16111785173416138f;
  t119_n105: return -0.07220607250928879f;
  t119_n106: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t119_n108;
  t119_n107: return 0.10764356702566147f;
  t119_n108: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t119_n110;
  t119_n109: return 0.22853614389896393f;
  t119_n110: return 0.14972756803035736f;
  t119_n111: return 0.15869903564453125f;
  t119_n112: if (E.getNumReferences() >= 3340838912 /*41256.0*/) goto t119_n114;
  t119_n113: return -0.004856247454881668f;
  t119_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n116;
  t119_n115: return -0.10907115787267685f;
  t119_n116: return 0.11032595485448837f;
  t119_n117: return -0.036545705050230026f;
  t119_n118: return -0.016514940187335014f;
  t119_n119: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t119_n199;
  t119_n120: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t119_n198;
  t119_n121: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t119_n193;
  t119_n122: if (E.getNumReferences() >= 3358036832 /*171597.5*/) goto t119_n192;
  t119_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t119_n177;
  t119_n124: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t119_n164;
  t119_n125: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t119_n127;
  t119_n126: return -0.001262929174117744f;
  t119_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t119_n135;
  t119_n128: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t119_n134;
  t119_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n131;
  t119_n130: return -0.04248042404651642f;
  t119_n131: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t119_n133;
  t119_n132: return 0.145375058054924f;
  t119_n133: return -0.18598118424415588f;
  t119_n134: return -0.32088735699653625f;
  t119_n135: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t119_n163;
  t119_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t119_n162;
  t119_n137: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t119_n157;
  t119_n138: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t119_n148;
  t119_n139: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t119_n147;
  t119_n140: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t119_n146;
  t119_n141: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t119_n145;
  t119_n142: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t119_n144;
  t119_n143: return 0.011728363111615181f;
  t119_n144: return -0.22896546125411987f;
  t119_n145: return 0.13734161853790283f;
  t119_n146: return -0.04241473227739334f;
  t119_n147: return 0.10945569723844528f;
  t119_n148: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t119_n154;
  t119_n149: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t119_n153;
  t119_n150: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t119_n152;
  t119_n151: return -0.2658828794956207f;
  t119_n152: return -0.13810862600803375f;
  t119_n153: return -0.023536983877420425f;
  t119_n154: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t119_n156;
  t119_n155: return -0.01857573725283146f;
  t119_n156: return -0.6945720314979553f;
  t119_n157: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t119_n159;
  t119_n158: return -0.006141533609479666f;
  t119_n159: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t119_n161;
  t119_n160: return 0.1259857565164566f;
  t119_n161: return -0.2445489764213562f;
  t119_n162: return -0.001003385172225535f;
  t119_n163: return 0.0052934433333575726f;
  t119_n164: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t119_n176;
  t119_n165: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t119_n171;
  t119_n166: if (E.getNumReferences() >= 3331078656 /*17949.0*/) goto t119_n168;
  t119_n167: return 0.010960215702652931f;
  t119_n168: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n170;
  t119_n169: return -0.02468270994722843f;
  t119_n170: return -0.19413258135318756f;
  t119_n171: if (E.getNumReferences() >= 3332287488 /*20310.0*/) goto t119_n175;
  t119_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n174;
  t119_n173: return -0.1251096874475479f;
  t119_n174: return 0.13890071213245392f;
  t119_n175: return 0.027628496289253235f;
  t119_n176: return -0.011414100416004658f;
  t119_n177: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t119_n187;
  t119_n178: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t119_n182;
  t119_n179: if (E.getNumReferences() >= 3357252288 /*159339.0*/) goto t119_n181;
  t119_n180: return 0.0016955184983089566f;
  t119_n181: return -0.23360785841941833f;
  t119_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t119_n186;
  t119_n183: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t119_n185;
  t119_n184: return 0.22724463045597076f;
  t119_n185: return 0.12574242055416107f;
  t119_n186: return 0.007077918853610754f;
  t119_n187: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t119_n189;
  t119_n188: return 0.004790626000612974f;
  t119_n189: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t119_n191;
  t119_n190: return -0.015247398987412453f;
  t119_n191: return -0.10248398780822754f;
  t119_n192: return -0.17459654808044434f;
  t119_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t119_n197;
  t119_n194: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t119_n196;
  t119_n195: return 0.08971341699361801f;
  t119_n196: return 0.005859398283064365f;
  t119_n197: return 0.09782407432794571f;
  t119_n198: return -0.05690528079867363f;
  t119_n199: return 0.031871963292360306f;
  t119_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t119_n236;
  t119_n201: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t119_n233;
  t119_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t119_n204;
  t119_n203: return -0.13798761367797852f;
  t119_n204: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t119_n230;
  t119_n205: if (E.getNumReferences() >= 3288121344 /*505.5*/) goto t119_n229;
  t119_n206: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t119_n228;
  t119_n207: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t119_n211;
  t119_n208: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t119_n210;
  t119_n209: return 0.009465603157877922f;
  t119_n210: return -0.03919774666428566f;
  t119_n211: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t119_n223;
  t119_n212: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t119_n214;
  t119_n213: return 0.02768954075872898f;
  t119_n214: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t119_n222;
  t119_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t119_n217;
  t119_n216: return 0.044455695897340775f;
  t119_n217: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t119_n219;
  t119_n218: return 0.10289989411830902f;
  t119_n219: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t119_n221;
  t119_n220: return 0.11293839663267136f;
  t119_n221: return 0.1635269969701767f;
  t119_n222: return 0.0010709043126553297f;
  t119_n223: if (E.getFractionNameInContext() >= 3168244790 /*0.02631578966975212*/) goto t119_n225;
  t119_n224: return -0.04391387850046158f;
  t119_n225: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t119_n227;
  t119_n226: return 0.11619990319013596f;
  t119_n227: return 0.023931834846735f;
  t119_n228: return 0.05299295485019684f;
  t119_n229: return 0.13589170575141907f;
  t119_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t119_n232;
  t119_n231: return -0.027567902579903603f;
  t119_n232: return 0.0017791838617995381f;
  t119_n233: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t119_n235;
  t119_n234: return 0.017157770693302155f;
  t119_n235: return 0.14944793283939362f;
  t119_n236: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t119_n254;
  t119_n237: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t119_n239;
  t119_n238: return 0.007190174423158169f;
  t119_n239: if (E.getNumReferences() >= 3306696704 /*2435.0*/) goto t119_n241;
  t119_n240: return 0.11735771596431732f;
  t119_n241: if (E.getNumReferences() >= 3309623296 /*3149.5*/) goto t119_n253;
  t119_n242: if (E.getNumReferences() >= 3309430784 /*3102.5*/) goto t119_n252;
  t119_n243: if (E.getNumReferences() >= 3307098112 /*2533.0*/) goto t119_n245;
  t119_n244: return -0.047284360975027084f;
  t119_n245: if (E.getNumReferences() >= 3307692032 /*2678.0*/) goto t119_n251;
  t119_n246: if (E.getNumReferences() >= 3307503616 /*2632.0*/) goto t119_n248;
  t119_n247: return 0.10478711128234863f;
  t119_n248: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t119_n250;
  t119_n249: return 0.11754748225212097f;
  t119_n250: return 0.07072458416223526f;
  t119_n251: return 0.07078348100185394f;
  t119_n252: return 0.12186964601278305f;
  t119_n253: return 0.030291467905044556f;
  t119_n254: return -0.07271980494260788f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree120(const Example& E) {
  t120_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t120_n222;
  t120_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t120_n121;
  t120_n2: if (E.getNumReferences() >= 3325315072 /*11538.0*/) goto t120_n52;
  t120_n3: if (E.getNumReferences() >= 3324264448 /*10512.0*/) goto t120_n51;
  t120_n4: if (E.getNumReferences() >= 3306622976 /*2417.0*/) goto t120_n42;
  t120_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t120_n7;
  t120_n6: return -0.0026942575350403786f;
  t120_n7: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t120_n41;
  t120_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t120_n20;
  t120_n9: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t120_n19;
  t120_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t120_n14;
  t120_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t120_n13;
  t120_n12: return 0.008150971494615078f;
  t120_n13: return 0.030892062932252884f;
  t120_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t120_n16;
  t120_n15: return -0.013298619538545609f;
  t120_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n18;
  t120_n17: return -0.0037101132329553366f;
  t120_n18: return 0.011938800103962421f;
  t120_n19: return -0.009309097193181515f;
  t120_n20: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t120_n34;
  t120_n21: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t120_n33;
  t120_n22: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t120_n32;
  t120_n23: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t120_n25;
  t120_n24: return -0.014060160145163536f;
  t120_n25: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t120_n31;
  t120_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t120_n28;
  t120_n27: return -0.004696503281593323f;
  t120_n28: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t120_n30;
  t120_n29: return 0.02583887055516243f;
  t120_n30: return -0.0945664718747139f;
  t120_n31: return -0.00806677807122469f;
  t120_n32: return -0.03540974482893944f;
  t120_n33: return 0.0039001659024506807f;
  t120_n34: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t120_n40;
  t120_n35: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t120_n37;
  t120_n36: return 0.10408622026443481f;
  t120_n37: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t120_n39;
  t120_n38: return 0.00542305875569582f;
  t120_n39: return 0.0439685694873333f;
  t120_n40: return -0.023827973753213882f;
  t120_n41: return -0.1741335391998291f;
  t120_n42: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t120_n50;
  t120_n43: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t120_n47;
  t120_n44: if (E.getNumReferences() >= 3316353024 /*5489.0*/) goto t120_n46;
  t120_n45: return 0.0874442458152771f;
  t120_n46: return -0.19661694765090942f;
  t120_n47: if (E.getNumReferences() >= 3320151040 /*7343.5*/) goto t120_n49;
  t120_n48: return 0.09689395874738693f;
  t120_n49: return 0.06683764606714249f;
  t120_n50: return 0.010606556199491024f;
  t120_n51: return -0.057923078536987305f;
  t120_n52: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t120_n120;
  t120_n53: if (E.getNumReferences() >= 3327478272 /*13650.5*/) goto t120_n63;
  t120_n54: if (E.getNumReferences() >= 3327457280 /*13630.0*/) goto t120_n60;
  t120_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t120_n57;
  t120_n56: return 0.05193081125617027f;
  t120_n57: if (E.getNumReferences() >= 3325893632 /*12103.0*/) goto t120_n59;
  t120_n58: return 0.09570205956697464f;
  t120_n59: return 0.044678039848804474f;
  t120_n60: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t120_n62;
  t120_n61: return 0.3135356903076172f;
  t120_n62: return 0.03707519918680191f;
  t120_n63: if (E.getNumReferences() >= 3332993536 /*21689.0*/) goto t120_n65;
  t120_n64: return -0.013790400698781013f;
  t120_n65: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t120_n85;
  t120_n66: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t120_n82;
  t120_n67: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t120_n73;
  t120_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t120_n70;
  t120_n69: return 0.022694366052746773f;
  t120_n70: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n72;
  t120_n71: return 0.06121528148651123f;
  t120_n72: return 0.15241092443466187f;
  t120_n73: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t120_n79;
  t120_n74: if (E.getNumReferences() >= 3334007808 /*23670.0*/) goto t120_n78;
  t120_n75: if (E.getNumReferences() >= 3333936384 /*23530.5*/) goto t120_n77;
  t120_n76: return 0.06684472411870956f;
  t120_n77: return 0.11293838173151016f;
  t120_n78: return -0.17256002128124237f;
  t120_n79: if (E.getNumReferences() >= 3333992192 /*23639.5*/) goto t120_n81;
  t120_n80: return 0.1308707445859909f;
  t120_n81: return 0.13562996685504913f;
  t120_n82: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t120_n84;
  t120_n83: return 0.060994792729616165f;
  t120_n84: return 0.10343771427869797f;
  t120_n85: if (E.getNumReferences() >= 3338991488 /*34039.5*/) goto t120_n87;
  t120_n86: return -0.01714162901043892f;
  t120_n87: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t120_n89;
  t120_n88: return 0.14591781795024872f;
  t120_n89: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t120_n103;
  t120_n90: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t120_n94;
  t120_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t120_n93;
  t120_n92: return -0.07253166288137436f;
  t120_n93: return 0.055260706692934036f;
  t120_n94: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t120_n100;
  t120_n95: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t120_n97;
  t120_n96: return 0.08861199021339417f;
  t120_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n99;
  t120_n98: return 0.026025809347629547f;
  t120_n99: return 0.14853766560554504f;
  t120_n100: if (E.getNumReferences() >= 3342540672 /*47903.5*/) goto t120_n102;
  t120_n101: return 0.041568320244550705f;
  t120_n102: return 0.08384064584970474f;
  t120_n103: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t120_n107;
  t120_n104: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t120_n106;
  t120_n105: return -0.04304707050323486f;
  t120_n106: return -0.27682042121887207f;
  t120_n107: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n111;
  t120_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t120_n110;
  t120_n109: return 0.0014056472573429346f;
  t120_n110: return 0.03263849765062332f;
  t120_n111: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t120_n113;
  t120_n112: return 0.17819459736347198f;
  t120_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t120_n115;
  t120_n114: return -0.07985755056142807f;
  t120_n115: if (E.getNumReferences() >= 3349064384 /*81237.5*/) goto t120_n119;
  t120_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t120_n118;
  t120_n117: return -0.05908498913049698f;
  t120_n118: return 0.15357714891433716f;
  t120_n119: return 0.02071131393313408f;
  t120_n120: return 0.0015353065682575107f;
  t120_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t120_n203;
  t120_n122: if (E.getFractionNameInContext() >= 3205809565 /*0.5811403393745422*/) goto t120_n182;
  t120_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t120_n165;
  t120_n124: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t120_n158;
  t120_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t120_n157;
  t120_n126: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t120_n154;
  t120_n127: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t120_n141;
  t120_n128: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t120_n130;
  t120_n129: return -0.0038999649696052074f;
  t120_n130: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t120_n132;
  t120_n131: return 0.0020820770878344774f;
  t120_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n140;
  t120_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t120_n139;
  t120_n134: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t120_n138;
  t120_n135: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t120_n137;
  t120_n136: return 0.0019871785771101713f;
  t120_n137: return -0.05720027908682823f;
  t120_n138: return -0.015362963080406189f;
  t120_n139: return -0.003132858080789447f;
  t120_n140: return 0.020735034719109535f;
  t120_n141: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t120_n143;
  t120_n142: return 0.0014616678236052394f;
  t120_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t120_n153;
  t120_n144: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t120_n150;
  t120_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t120_n147;
  t120_n146: return -0.01491548866033554f;
  t120_n147: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t120_n149;
  t120_n148: return 0.012774158269166946f;
  t120_n149: return 0.1512681096792221f;
  t120_n150: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t120_n152;
  t120_n151: return 0.11102037876844406f;
  t120_n152: return 0.03170159086585045f;
  t120_n153: return 0.031604018062353134f;
  t120_n154: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t120_n156;
  t120_n155: return -0.036265552043914795f;
  t120_n156: return 0.12860222160816193f;
  t120_n157: return 0.0008502574055455625f;
  t120_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t120_n162;
  t120_n159: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t120_n161;
  t120_n160: return 0.0019774187821894884f;
  t120_n161: return 0.039876118302345276f;
  t120_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n164;
  t120_n163: return 0.01316104270517826f;
  t120_n164: return -0.007926573045551777f;
  t120_n165: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t120_n177;
  t120_n166: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n168;
  t120_n167: return 0.005906339269131422f;
  t120_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t120_n170;
  t120_n169: return 0.024063006043434143f;
  t120_n170: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t120_n176;
  t120_n171: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t120_n175;
  t120_n172: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t120_n174;
  t120_n173: return 0.10288197547197342f;
  t120_n174: return -0.13758307695388794f;
  t120_n175: return 0.04192972183227539f;
  t120_n176: return 0.16171684861183167f;
  t120_n177: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t120_n181;
  t120_n178: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t120_n180;
  t120_n179: return 0.011669686995446682f;
  t120_n180: return -0.02742583304643631f;
  t120_n181: return 0.04277527704834938f;
  t120_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t120_n186;
  t120_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t120_n185;
  t120_n184: return -0.005367468111217022f;
  t120_n185: return 0.011560242623090744f;
  t120_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t120_n194;
  t120_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t120_n193;
  t120_n188: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t120_n192;
  t120_n189: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t120_n191;
  t120_n190: return 0.11850801855325699f;
  t120_n191: return -0.0017599978018552065f;
  t120_n192: return 0.06907923519611359f;
  t120_n193: return 0.03720709681510925f;
  t120_n194: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t120_n202;
  t120_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t120_n197;
  t120_n196: return 0.030104711651802063f;
  t120_n197: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t120_n201;
  t120_n198: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t120_n200;
  t120_n199: return -0.12493003159761429f;
  t120_n200: return 0.12832720577716827f;
  t120_n201: return 0.45086240768432617f;
  t120_n202: return 0.21303798258304596f;
  t120_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t120_n205;
  t120_n204: return 0.00182180549018085f;
  t120_n205: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t120_n215;
  t120_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t120_n214;
  t120_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t120_n213;
  t120_n208: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t120_n210;
  t120_n209: return 0.04906149208545685f;
  t120_n210: if (E.getFractionNameInContext() >= 3206445544 /*0.6190476417541504*/) goto t120_n212;
  t120_n211: return 0.0464773029088974f;
  t120_n212: return 0.0802665501832962f;
  t120_n213: return -0.005792181473225355f;
  t120_n214: return -0.0451512336730957f;
  t120_n215: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t120_n221;
  t120_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t120_n218;
  t120_n217: return -0.04533904790878296f;
  t120_n218: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t120_n220;
  t120_n219: return 0.0175915714353323f;
  t120_n220: return 0.04742194712162018f;
  t120_n221: return -0.10982761532068253f;
  t120_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t120_n252;
  t120_n223: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t120_n239;
  t120_n224: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t120_n238;
  t120_n225: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t120_n231;
  t120_n226: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t120_n228;
  t120_n227: return 0.010098605416715145f;
  t120_n228: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t120_n230;
  t120_n229: return 0.09667878597974777f;
  t120_n230: return 0.03151776269078255f;
  t120_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t120_n237;
  t120_n232: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t120_n236;
  t120_n233: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t120_n235;
  t120_n234: return 0.0991765707731247f;
  t120_n235: return -0.07550010830163956f;
  t120_n236: return -0.03518424555659294f;
  t120_n237: return 0.006073031574487686f;
  t120_n238: return -0.0020142316352576017f;
  t120_n239: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t120_n251;
  t120_n240: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t120_n250;
  t120_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t120_n249;
  t120_n242: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t120_n248;
  t120_n243: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t120_n245;
  t120_n244: return -0.014230316504836082f;
  t120_n245: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t120_n247;
  t120_n246: return 0.08036689460277557f;
  t120_n247: return -0.334110289812088f;
  t120_n248: return 0.07107304781675339f;
  t120_n249: return 0.01914840377867222f;
  t120_n250: return 0.04183298721909523f;
  t120_n251: return 0.13736312091350555f;
  t120_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t120_n254;
  t120_n253: return 0.00846092402935028f;
  t120_n254: return -0.07963182032108307f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree121(const Example& E) {
  t121_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t121_n4;
  t121_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t121_n3;
  t121_n2: return -0.10024348646402359f;
  t121_n3: return -0.0009124461212195456f;
  t121_n4: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t121_n86;
  t121_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t121_n69;
  t121_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n68;
  t121_n7: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t121_n67;
  t121_n8: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t121_n52;
  t121_n9: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t121_n51;
  t121_n10: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t121_n34;
  t121_n11: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t121_n33;
  t121_n12: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t121_n28;
  t121_n13: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t121_n17;
  t121_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t121_n16;
  t121_n15: return -0.00421023927628994f;
  t121_n16: return 0.0468035452067852f;
  t121_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t121_n27;
  t121_n18: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t121_n20;
  t121_n19: return -0.056531667709350586f;
  t121_n20: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t121_n26;
  t121_n21: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t121_n25;
  t121_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n24;
  t121_n23: return 0.03760319575667381f;
  t121_n24: return -0.2844904959201813f;
  t121_n25: return 0.410255491733551f;
  t121_n26: return -0.039646197110414505f;
  t121_n27: return -0.00925513170659542f;
  t121_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t121_n32;
  t121_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t121_n31;
  t121_n30: return -0.15573719143867493f;
  t121_n31: return 0.044586386531591415f;
  t121_n32: return 0.1248648539185524f;
  t121_n33: return -0.07266387343406677f;
  t121_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t121_n38;
  t121_n35: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t121_n37;
  t121_n36: return -0.030194537714123726f;
  t121_n37: return -0.3345746695995331f;
  t121_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n42;
  t121_n39: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t121_n41;
  t121_n40: return -0.01646747998893261f;
  t121_n41: return -0.0503368079662323f;
  t121_n42: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t121_n50;
  t121_n43: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t121_n45;
  t121_n44: return -0.5227890014648438f;
  t121_n45: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t121_n47;
  t121_n46: return -0.15484528243541718f;
  t121_n47: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t121_n49;
  t121_n48: return 0.04730264097452164f;
  t121_n49: return -0.06776309013366699f;
  t121_n50: return 0.04749784246087074f;
  t121_n51: return 0.0022940603084862232f;
  t121_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t121_n60;
  t121_n53: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t121_n57;
  t121_n54: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t121_n56;
  t121_n55: return 0.04280773177742958f;
  t121_n56: return -0.18449170887470245f;
  t121_n57: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t121_n59;
  t121_n58: return -0.07236646860837936f;
  t121_n59: return 0.01509571261703968f;
  t121_n60: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t121_n66;
  t121_n61: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t121_n63;
  t121_n62: return 0.09862184524536133f;
  t121_n63: if (E.getFractionNameInContext() >= 3174706828 /*0.04545454680919647*/) goto t121_n65;
  t121_n64: return -0.8512701392173767f;
  t121_n65: return 0.06309676170349121f;
  t121_n66: return 0.133380725979805f;
  t121_n67: return 0.0151128601282835f;
  t121_n68: return 0.006541426759213209f;
  t121_n69: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t121_n73;
  t121_n70: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t121_n72;
  t121_n71: return 0.0018497334094718099f;
  t121_n72: return 0.03546380251646042f;
  t121_n73: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t121_n85;
  t121_n74: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t121_n82;
  t121_n75: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t121_n77;
  t121_n76: return 0.025654487311840057f;
  t121_n77: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t121_n79;
  t121_n78: return 0.07787908613681793f;
  t121_n79: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t121_n81;
  t121_n80: return 0.023752516135573387f;
  t121_n81: return 0.10737966001033783f;
  t121_n82: if (E.getNumReferences() >= 3290218496 /*627.0*/) goto t121_n84;
  t121_n83: return -0.05972527712583542f;
  t121_n84: return 0.05603818967938423f;
  t121_n85: return 0.005757858045399189f;
  t121_n86: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t121_n234;
  t121_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t121_n91;
  t121_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n90;
  t121_n89: return -0.0018824003636837006f;
  t121_n90: return 0.020773131400346756f;
  t121_n91: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t121_n233;
  t121_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t121_n200;
  t121_n93: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t121_n189;
  t121_n94: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t121_n186;
  t121_n95: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t121_n133;
  t121_n96: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t121_n114;
  t121_n97: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t121_n101;
  t121_n98: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t121_n100;
  t121_n99: return -0.011629240587353706f;
  t121_n100: return -0.0328415222465992f;
  t121_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t121_n109;
  t121_n102: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t121_n104;
  t121_n103: return -0.09881503880023956f;
  t121_n104: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t121_n106;
  t121_n105: return -0.053410883992910385f;
  t121_n106: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t121_n108;
  t121_n107: return 0.11015525460243225f;
  t121_n108: return 0.038680627942085266f;
  t121_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n113;
  t121_n110: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n112;
  t121_n111: return -0.06895601004362106f;
  t121_n112: return -0.0299227274954319f;
  t121_n113: return -0.07219880074262619f;
  t121_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t121_n130;
  t121_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t121_n125;
  t121_n116: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t121_n124;
  t121_n117: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n123;
  t121_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n120;
  t121_n119: return -0.08326348662376404f;
  t121_n120: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t121_n122;
  t121_n121: return 0.01584498956799507f;
  t121_n122: return 0.15239909291267395f;
  t121_n123: return -0.021945124492049217f;
  t121_n124: return 0.15077897906303406f;
  t121_n125: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t121_n127;
  t121_n126: return 0.03901751711964607f;
  t121_n127: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t121_n129;
  t121_n128: return -0.05166623368859291f;
  t121_n129: return -0.3568873107433319f;
  t121_n130: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t121_n132;
  t121_n131: return 0.026816265657544136f;
  t121_n132: return -0.0032335103023797274f;
  t121_n133: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t121_n179;
  t121_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t121_n138;
  t121_n135: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t121_n137;
  t121_n136: return 0.003165645757690072f;
  t121_n137: return -0.0009449964854866266f;
  t121_n138: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t121_n152;
  t121_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n151;
  t121_n140: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t121_n150;
  t121_n141: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t121_n145;
  t121_n142: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t121_n144;
  t121_n143: return 0.004928826354444027f;
  t121_n144: return -0.03157328814268112f;
  t121_n145: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t121_n147;
  t121_n146: return 0.0027737983036786318f;
  t121_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t121_n149;
  t121_n148: return 0.06257830560207367f;
  t121_n149: return 0.11338242143392563f;
  t121_n150: return -0.15625737607479095f;
  t121_n151: return 0.015137880109250546f;
  t121_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t121_n168;
  t121_n153: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t121_n157;
  t121_n154: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t121_n156;
  t121_n155: return -0.2254643440246582f;
  t121_n156: return 0.0012624308001250029f;
  t121_n157: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t121_n161;
  t121_n158: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t121_n160;
  t121_n159: return -0.00785698089748621f;
  t121_n160: return 0.12335193157196045f;
  t121_n161: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t121_n167;
  t121_n162: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t121_n164;
  t121_n163: return 0.04884497821331024f;
  t121_n164: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t121_n166;
  t121_n165: return 0.1924591213464737f;
  t121_n166: return 0.06838716566562653f;
  t121_n167: return -0.04174426570534706f;
  t121_n168: if (E.getFractionNameInContext() >= 3196938455 /*0.27619048953056335*/) goto t121_n170;
  t121_n169: return 0.08620313555002213f;
  t121_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t121_n174;
  t121_n171: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n173;
  t121_n172: return 0.01055699959397316f;
  t121_n173: return 0.09166831523180008f;
  t121_n174: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t121_n178;
  t121_n175: if (E.getFractionNameInContext() >= 3200482732 /*0.38181817531585693*/) goto t121_n177;
  t121_n176: return 0.11174001544713974f;
  t121_n177: return 0.07858690619468689f;
  t121_n178: return -0.3653782904148102f;
  t121_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t121_n181;
  t121_n180: return -0.04378349334001541f;
  t121_n181: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t121_n183;
  t121_n182: return 0.006529053673148155f;
  t121_n183: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t121_n185;
  t121_n184: return -0.05184756591916084f;
  t121_n185: return 0.056790441274642944f;
  t121_n186: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t121_n188;
  t121_n187: return 0.007530285511165857f;
  t121_n188: return 0.13078071177005768f;
  t121_n189: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t121_n193;
  t121_n190: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t121_n192;
  t121_n191: return -0.08911450207233429f;
  t121_n192: return 0.019215742126107216f;
  t121_n193: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t121_n195;
  t121_n194: return 0.020424598827958107f;
  t121_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t121_n199;
  t121_n196: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t121_n198;
  t121_n197: return -0.06605146825313568f;
  t121_n198: return 0.12042967975139618f;
  t121_n199: return 0.16714222729206085f;
  t121_n200: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t121_n218;
  t121_n201: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t121_n207;
  t121_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n204;
  t121_n203: return -0.17340068519115448f;
  t121_n204: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t121_n206;
  t121_n205: return 0.0007627106388099492f;
  t121_n206: return 0.01037480216473341f;
  t121_n207: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t121_n213;
  t121_n208: if (E.getFractionNameInContext() >= 3212137814 /*0.9583333730697632*/) goto t121_n210;
  t121_n209: return 0.04754066467285156f;
  t121_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t121_n212;
  t121_n211: return -0.14591169357299805f;
  t121_n212: return 0.16220280528068542f;
  t121_n213: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t121_n217;
  t121_n214: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n216;
  t121_n215: return 0.10641317814588547f;
  t121_n216: return -0.23344279825687408f;
  t121_n217: return 0.13250766694545746f;
  t121_n218: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t121_n232;
  t121_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t121_n221;
  t121_n220: return 0.005777824204415083f;
  t121_n221: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t121_n231;
  t121_n222: if (E.getFractionNameInContext() >= 3211159142 /*0.8999999761581421*/) goto t121_n230;
  t121_n223: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t121_n229;
  t121_n224: if (E.getFractionNameInContext() >= 3177111263 /*0.054411765187978745*/) goto t121_n226;
  t121_n225: return 0.033868152648210526f;
  t121_n226: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t121_n228;
  t121_n227: return 0.07426635175943375f;
  t121_n228: return -0.01554314885288477f;
  t121_n229: return -0.1291540116071701f;
  t121_n230: return 0.1169532835483551f;
  t121_n231: return 0.0705345943570137f;
  t121_n232: return -0.02837296947836876f;
  t121_n233: return -0.07829426229000092f;
  t121_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n238;
  t121_n235: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t121_n237;
  t121_n236: return -0.013300401158630848f;
  t121_n237: return -0.02615475095808506f;
  t121_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t121_n254;
  t121_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t121_n251;
  t121_n240: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t121_n242;
  t121_n241: return -0.041594043374061584f;
  t121_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t121_n246;
  t121_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t121_n245;
  t121_n244: return -0.029332948848605156f;
  t121_n245: return 0.003040298353880644f;
  t121_n246: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t121_n250;
  t121_n247: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t121_n249;
  t121_n248: return 0.015572352334856987f;
  t121_n249: return 0.04154224693775177f;
  t121_n250: return -0.009189972653985023f;
  t121_n251: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t121_n253;
  t121_n252: return 0.030207233503460884f;
  t121_n253: return 0.08283963054418564f;
  t121_n254: return 0.03979991748929024f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree122(const Example& E) {
  t122_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t122_n250;
  t122_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t122_n47;
  t122_n2: if (E.getNumReferences() >= 3289325568 /*572.5*/) goto t122_n42;
  t122_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t122_n7;
  t122_n4: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t122_n6;
  t122_n5: return -0.00013713593943975866f;
  t122_n6: return -0.10300993919372559f;
  t122_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t122_n39;
  t122_n8: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t122_n20;
  t122_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n19;
  t122_n10: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t122_n14;
  t122_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n13;
  t122_n12: return -0.20100924372673035f;
  t122_n13: return 0.0777101069688797f;
  t122_n14: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t122_n18;
  t122_n15: if (E.getNumReferences() >= 3256745984 /*39.5*/) goto t122_n17;
  t122_n16: return -0.2630074918270111f;
  t122_n17: return 0.10625836253166199f;
  t122_n18: return -0.1272727996110916f;
  t122_n19: return -0.012498794123530388f;
  t122_n20: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t122_n34;
  t122_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t122_n27;
  t122_n22: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t122_n26;
  t122_n23: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t122_n25;
  t122_n24: return -0.028776157647371292f;
  t122_n25: return -0.0748089849948883f;
  t122_n26: return 0.001138879219070077f;
  t122_n27: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t122_n29;
  t122_n28: return 0.007258948870003223f;
  t122_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n31;
  t122_n30: return 0.002881062915548682f;
  t122_n31: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t122_n33;
  t122_n32: return 0.03911150246858597f;
  t122_n33: return -0.11251983791589737f;
  t122_n34: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t122_n38;
  t122_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n37;
  t122_n36: return -0.12196841835975647f;
  t122_n37: return 0.0272446908056736f;
  t122_n38: return 0.13942863047122955f;
  t122_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t122_n41;
  t122_n40: return 0.007070333696901798f;
  t122_n41: return 0.07632800191640854f;
  t122_n42: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t122_n46;
  t122_n43: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t122_n45;
  t122_n44: return 0.012157647870481014f;
  t122_n45: return 0.09644411504268646f;
  t122_n46: return -0.18423551321029663f;
  t122_n47: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t122_n227;
  t122_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t122_n212;
  t122_n49: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t122_n89;
  t122_n50: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t122_n82;
  t122_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t122_n53;
  t122_n52: return -0.004189439117908478f;
  t122_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t122_n57;
  t122_n54: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t122_n56;
  t122_n55: return -0.21706445515155792f;
  t122_n56: return -0.09088798612356186f;
  t122_n57: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t122_n61;
  t122_n58: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t122_n60;
  t122_n59: return 0.010470184497535229f;
  t122_n60: return -0.02006070502102375f;
  t122_n61: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t122_n81;
  t122_n62: if (E.getNumReferences() >= 3353316032 /*114453.5*/) goto t122_n78;
  t122_n63: if (E.getNumReferences() >= 3341397376 /*43437.5*/) goto t122_n77;
  t122_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t122_n70;
  t122_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n67;
  t122_n66: return -0.10983797907829285f;
  t122_n67: if (E.getNumReferences() >= 3267166208 /*94.5*/) goto t122_n69;
  t122_n68: return 0.01572110876441002f;
  t122_n69: return 0.06050478667020798f;
  t122_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t122_n74;
  t122_n71: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t122_n73;
  t122_n72: return 0.021836567670106888f;
  t122_n73: return -0.020907845348119736f;
  t122_n74: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t122_n76;
  t122_n75: return 0.03908837214112282f;
  t122_n76: return 0.10959862917661667f;
  t122_n77: return -0.16425654292106628f;
  t122_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n80;
  t122_n79: return -0.02786710113286972f;
  t122_n80: return 0.11017565429210663f;
  t122_n81: return 0.011035148985683918f;
  t122_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t122_n88;
  t122_n83: if (E.getNumReferences() >= 3364639120 /*287372.5*/) goto t122_n87;
  t122_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t122_n86;
  t122_n85: return -0.1025344505906105f;
  t122_n86: return -0.03306536376476288f;
  t122_n87: return -0.05266396328806877f;
  t122_n88: return 0.09915849566459656f;
  t122_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t122_n177;
  t122_n90: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t122_n164;
  t122_n91: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t122_n95;
  t122_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n94;
  t122_n93: return -0.05712660774588585f;
  t122_n94: return 0.030106905847787857f;
  t122_n95: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t122_n107;
  t122_n96: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t122_n104;
  t122_n97: if (E.getNumReferences() >= 3277619200 /*220.5*/) goto t122_n99;
  t122_n98: return 0.02196711301803589f;
  t122_n99: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t122_n103;
  t122_n100: if (E.getNumReferences() >= 3283615744 /*368.0*/) goto t122_n102;
  t122_n101: return 0.2995036542415619f;
  t122_n102: return -0.10544906556606293f;
  t122_n103: return 0.10369031131267548f;
  t122_n104: if (E.getNumReferences() >= 3367700464 /*383039.5*/) goto t122_n106;
  t122_n105: return -0.012170521542429924f;
  t122_n106: return -0.052961524575948715f;
  t122_n107: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t122_n157;
  t122_n108: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t122_n152;
  t122_n109: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t122_n151;
  t122_n110: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t122_n144;
  t122_n111: if (E.getNumReferences() >= 3353301504 /*114340.0*/) goto t122_n143;
  t122_n112: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t122_n140;
  t122_n113: if (E.getNumReferences() >= 3338535680 /*32513.5*/) goto t122_n123;
  t122_n114: if (E.getNumReferences() >= 3336381440 /*28306.0*/) goto t122_n122;
  t122_n115: if (E.getNumReferences() >= 3333976064 /*23608.0*/) goto t122_n117;
  t122_n116: return 0.0002523923176340759f;
  t122_n117: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t122_n121;
  t122_n118: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t122_n120;
  t122_n119: return 0.1029563620686531f;
  t122_n120: return 0.12183696031570435f;
  t122_n121: return 0.0257573202252388f;
  t122_n122: return -0.06691945344209671f;
  t122_n123: if (E.getNumReferences() >= 3339196032 /*34838.5*/) goto t122_n137;
  t122_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t122_n128;
  t122_n125: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t122_n127;
  t122_n126: return -0.04293940216302872f;
  t122_n127: return 0.08973729610443115f;
  t122_n128: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t122_n134;
  t122_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n133;
  t122_n130: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t122_n132;
  t122_n131: return 0.1996934413909912f;
  t122_n132: return 0.09559276700019836f;
  t122_n133: return 0.0016729211201891303f;
  t122_n134: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t122_n136;
  t122_n135: return -0.053108248859643936f;
  t122_n136: return 0.08412274718284607f;
  t122_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n139;
  t122_n138: return -0.0032793078571558f;
  t122_n139: return 0.025673111900687218f;
  t122_n140: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n142;
  t122_n141: return 0.07903455942869186f;
  t122_n142: return -0.04285065457224846f;
  t122_n143: return -0.0490911640226841f;
  t122_n144: if (E.getNumReferences() >= 3361098688 /*219439.0*/) goto t122_n148;
  t122_n145: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n147;
  t122_n146: return 0.04673252999782562f;
  t122_n147: return 0.09400129318237305f;
  t122_n148: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t122_n150;
  t122_n149: return -0.08915743231773376f;
  t122_n150: return 0.0535912849009037f;
  t122_n151: return -0.16197113692760468f;
  t122_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n156;
  t122_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n155;
  t122_n154: return -0.06893464177846909f;
  t122_n155: return -0.026122869923710823f;
  t122_n156: return -0.008531076833605766f;
  t122_n157: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t122_n161;
  t122_n158: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n160;
  t122_n159: return 0.024748986586928368f;
  t122_n160: return 0.09633392840623856f;
  t122_n161: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t122_n163;
  t122_n162: return -0.05210632085800171f;
  t122_n163: return 0.028262227773666382f;
  t122_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t122_n166;
  t122_n165: return -0.007567909080535173f;
  t122_n166: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t122_n176;
  t122_n167: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t122_n175;
  t122_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t122_n174;
  t122_n169: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t122_n173;
  t122_n170: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t122_n172;
  t122_n171: return 0.023069268092513084f;
  t122_n172: return 0.03335797041654587f;
  t122_n173: return -0.0019205717835575342f;
  t122_n174: return 0.05098547413945198f;
  t122_n175: return -0.10401364415884018f;
  t122_n176: return -0.014318152330815792f;
  t122_n177: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t122_n205;
  t122_n178: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t122_n204;
  t122_n179: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t122_n201;
  t122_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t122_n184;
  t122_n181: if (E.getNumReferences() >= 3347831872 /*71608.5*/) goto t122_n183;
  t122_n182: return -0.02338491380214691f;
  t122_n183: return 0.16520154476165771f;
  t122_n184: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t122_n198;
  t122_n185: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t122_n197;
  t122_n186: if (E.getNumReferences() >= 3267493888 /*97.0*/) goto t122_n188;
  t122_n187: return 0.03137388080358505f;
  t122_n188: if (E.getNumReferences() >= 3277389824 /*217.0*/) goto t122_n196;
  t122_n189: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t122_n191;
  t122_n190: return 0.07150455564260483f;
  t122_n191: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t122_n195;
  t122_n192: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t122_n194;
  t122_n193: return 0.24134685099124908f;
  t122_n194: return 0.2193797379732132f;
  t122_n195: return -0.10946819931268692f;
  t122_n196: return 0.009434820152819157f;
  t122_n197: return 0.14811460673809052f;
  t122_n198: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t122_n200;
  t122_n199: return 0.011377205140888691f;
  t122_n200: return -0.19936606287956238f;
  t122_n201: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t122_n203;
  t122_n202: return 0.10780216753482819f;
  t122_n203: return 0.02545340359210968f;
  t122_n204: return -0.05432966351509094f;
  t122_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n207;
  t122_n206: return -0.029932662844657898f;
  t122_n207: if (E.getNumReferences() >= 3331470848 /*18715.0*/) goto t122_n209;
  t122_n208: return 0.006772107444703579f;
  t122_n209: if (E.getNumReferences() >= 3331793408 /*19345.0*/) goto t122_n211;
  t122_n210: return 0.11695563048124313f;
  t122_n211: return 0.04059597849845886f;
  t122_n212: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t122_n222;
  t122_n213: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t122_n215;
  t122_n214: return 0.0024448304902762175f;
  t122_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t122_n217;
  t122_n216: return -0.019264139235019684f;
  t122_n217: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t122_n221;
  t122_n218: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t122_n220;
  t122_n219: return 0.026562046259641647f;
  t122_n220: return 0.08861265331506729f;
  t122_n221: return -0.08669006824493408f;
  t122_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n226;
  t122_n223: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t122_n225;
  t122_n224: return -0.06548918038606644f;
  t122_n225: return 0.07599932700395584f;
  t122_n226: return 0.03673877939581871f;
  t122_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n229;
  t122_n228: return -0.02653176710009575f;
  t122_n229: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t122_n233;
  t122_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t122_n232;
  t122_n231: return -0.05342138558626175f;
  t122_n232: return -0.021131940186023712f;
  t122_n233: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t122_n249;
  t122_n234: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t122_n248;
  t122_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t122_n237;
  t122_n236: return -0.03841579705476761f;
  t122_n237: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t122_n239;
  t122_n238: return 0.024973243474960327f;
  t122_n239: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t122_n245;
  t122_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t122_n242;
  t122_n241: return -0.05177057161927223f;
  t122_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t122_n244;
  t122_n243: return 0.21085265278816223f;
  t122_n244: return -0.14191485941410065f;
  t122_n245: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t122_n247;
  t122_n246: return 0.004261022433638573f;
  t122_n247: return -0.04052131995558739f;
  t122_n248: return -0.007245237939059734f;
  t122_n249: return -0.07097875326871872f;
  t122_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t122_n252;
  t122_n251: return 0.0026784008368849754f;
  t122_n252: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t122_n254;
  t122_n253: return 0.010598448105156422f;
  t122_n254: return -0.013376638293266296f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree123(const Example& E) {
  t123_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t123_n180;
  t123_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t123_n157;
  t123_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t123_n150;
  t123_n3: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t123_n71;
  t123_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n6;
  t123_n5: return -0.00452780444175005f;
  t123_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t123_n50;
  t123_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t123_n13;
  t123_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n10;
  t123_n9: return -0.01753406971693039f;
  t123_n10: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t123_n12;
  t123_n11: return -0.003096699947491288f;
  t123_n12: return 0.008553776890039444f;
  t123_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t123_n27;
  t123_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t123_n22;
  t123_n15: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t123_n19;
  t123_n16: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t123_n18;
  t123_n17: return 0.012771343812346458f;
  t123_n18: return -0.0008670383831486106f;
  t123_n19: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t123_n21;
  t123_n20: return -0.05108867958188057f;
  t123_n21: return -0.02002062276005745f;
  t123_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t123_n24;
  t123_n23: return -0.005766354966908693f;
  t123_n24: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t123_n26;
  t123_n25: return 0.02329920418560505f;
  t123_n26: return -0.011376204900443554f;
  t123_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t123_n47;
  t123_n28: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t123_n30;
  t123_n29: return 0.0035661691799759865f;
  t123_n30: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t123_n46;
  t123_n31: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t123_n43;
  t123_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n36;
  t123_n33: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t123_n35;
  t123_n34: return 0.015102088451385498f;
  t123_n35: return -0.02761068381369114f;
  t123_n36: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t123_n38;
  t123_n37: return -0.003015610622242093f;
  t123_n38: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t123_n40;
  t123_n39: return 0.035082846879959106f;
  t123_n40: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t123_n42;
  t123_n41: return -0.04452761262655258f;
  t123_n42: return 0.18548007309436798f;
  t123_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t123_n45;
  t123_n44: return 0.004183686804026365f;
  t123_n45: return 0.029507778584957123f;
  t123_n46: return -0.018538309261202812f;
  t123_n47: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t123_n49;
  t123_n48: return 0.008961247280240059f;
  t123_n49: return 0.015180389396846294f;
  t123_n50: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t123_n62;
  t123_n51: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t123_n53;
  t123_n52: return 0.04215243458747864f;
  t123_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t123_n55;
  t123_n54: return -0.19710898399353027f;
  t123_n55: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t123_n59;
  t123_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n58;
  t123_n57: return -0.42625296115875244f;
  t123_n58: return 0.12533694505691528f;
  t123_n59: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t123_n61;
  t123_n60: return 0.1546953320503235f;
  t123_n61: return 0.06489672511816025f;
  t123_n62: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t123_n70;
  t123_n63: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t123_n69;
  t123_n64: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t123_n66;
  t123_n65: return -0.0501626618206501f;
  t123_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t123_n68;
  t123_n67: return -0.037150535732507706f;
  t123_n68: return 0.1635202020406723f;
  t123_n69: return -0.11809579282999039f;
  t123_n70: return -0.0031761308200657368f;
  t123_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t123_n149;
  t123_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n142;
  t123_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t123_n129;
  t123_n74: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t123_n126;
  t123_n75: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t123_n99;
  t123_n76: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t123_n98;
  t123_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n97;
  t123_n78: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t123_n96;
  t123_n79: if (E.getNumReferences() >= 3260678144 /*54.5*/) goto t123_n81;
  t123_n80: return -0.012423207052052021f;
  t123_n81: if (E.getNumReferences() >= 3282616320 /*337.5*/) goto t123_n95;
  t123_n82: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t123_n94;
  t123_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n85;
  t123_n84: return -0.06315770000219345f;
  t123_n85: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t123_n91;
  t123_n86: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t123_n88;
  t123_n87: return 0.03315587714314461f;
  t123_n88: if (E.getNumReferences() >= 3267887104 /*100.0*/) goto t123_n90;
  t123_n89: return -0.13621628284454346f;
  t123_n90: return 0.13763968646526337f;
  t123_n91: if (E.getNumReferences() >= 3261726720 /*58.5*/) goto t123_n93;
  t123_n92: return 0.20067091286182404f;
  t123_n93: return 0.08220399916172028f;
  t123_n94: return -0.07975684851408005f;
  t123_n95: return -0.14742957055568695f;
  t123_n96: return -0.034332744777202606f;
  t123_n97: return -0.004861864261329174f;
  t123_n98: return 0.011061073280870914f;
  t123_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t123_n103;
  t123_n100: if (E.getNumReferences() >= 3326525952 /*12720.5*/) goto t123_n102;
  t123_n101: return 0.0004337524878792465f;
  t123_n102: return -0.052730195224285126f;
  t123_n103: if (E.getNumReferences() >= 3340753024 /*40920.5*/) goto t123_n123;
  t123_n104: if (E.getNumReferences() >= 3338736000 /*33041.5*/) goto t123_n120;
  t123_n105: if (E.getNumReferences() >= 3275194368 /*183.5*/) goto t123_n107;
  t123_n106: return 0.01620151475071907f;
  t123_n107: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t123_n119;
  t123_n108: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t123_n118;
  t123_n109: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t123_n115;
  t123_n110: if (E.getNumReferences() >= 3278340096 /*231.5*/) goto t123_n112;
  t123_n111: return 0.19040973484516144f;
  t123_n112: if (E.getNumReferences() >= 3287007232 /*471.5*/) goto t123_n114;
  t123_n113: return -0.07030696421861649f;
  t123_n114: return 0.11152742058038712f;
  t123_n115: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t123_n117;
  t123_n116: return 0.009154224768280983f;
  t123_n117: return 0.132423534989357f;
  t123_n118: return 0.014836492948234081f;
  t123_n119: return 0.03707120195031166f;
  t123_n120: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t123_n122;
  t123_n121: return 0.026254350319504738f;
  t123_n122: return 0.14003463089466095f;
  t123_n123: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t123_n125;
  t123_n124: return -0.023276273161172867f;
  t123_n125: return -0.5697132349014282f;
  t123_n126: if (E.getNumReferences() >= 3244294144 /*14.0*/) goto t123_n128;
  t123_n127: return 0.3675655424594879f;
  t123_n128: return 0.1669570803642273f;
  t123_n129: if (E.getNumReferences() >= 3349388864 /*83772.5*/) goto t123_n133;
  t123_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t123_n132;
  t123_n131: return -0.0425051674246788f;
  t123_n132: return 0.03263818100094795f;
  t123_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n135;
  t123_n134: return -0.157002255320549f;
  t123_n135: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t123_n139;
  t123_n136: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t123_n138;
  t123_n137: return 0.09346015006303787f;
  t123_n138: return -0.02318812534213066f;
  t123_n139: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t123_n141;
  t123_n140: return 0.10440874844789505f;
  t123_n141: return 0.033595696091651917f;
  t123_n142: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t123_n144;
  t123_n143: return 0.025032317265868187f;
  t123_n144: if (E.getNumReferences() >= 3331353088 /*18485.0*/) goto t123_n146;
  t123_n145: return 0.01091484073549509f;
  t123_n146: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t123_n148;
  t123_n147: return 0.1062101274728775f;
  t123_n148: return 0.01728680357336998f;
  t123_n149: return -0.005686541553586721f;
  t123_n150: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t123_n156;
  t123_n151: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t123_n153;
  t123_n152: return 0.005751962307840586f;
  t123_n153: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t123_n155;
  t123_n154: return 0.057258397340774536f;
  t123_n155: return -0.11684975773096085f;
  t123_n156: return -0.003975861705839634f;
  t123_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t123_n171;
  t123_n158: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t123_n168;
  t123_n159: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t123_n161;
  t123_n160: return -0.0007499621133320034f;
  t123_n161: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t123_n163;
  t123_n162: return 0.07326993346214294f;
  t123_n163: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t123_n165;
  t123_n164: return -0.07524699717760086f;
  t123_n165: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t123_n167;
  t123_n166: return -0.0009591796551831067f;
  t123_n167: return -0.02411074750125408f;
  t123_n168: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t123_n170;
  t123_n169: return 0.018201498314738274f;
  t123_n170: return 0.05725253373384476f;
  t123_n171: if (E.getFractionNameInContext() >= 3204006750 /*0.4868420958518982*/) goto t123_n179;
  t123_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace))) goto t123_n174;
  t123_n173: return 0.0016390540404245257f;
  t123_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n176;
  t123_n175: return -0.027673494070768356f;
  t123_n176: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t123_n178;
  t123_n177: return 0.025476140901446342f;
  t123_n178: return -0.11191882193088531f;
  t123_n179: return 0.00838513020426035f;
  t123_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t123_n226;
  t123_n181: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t123_n225;
  t123_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t123_n184;
  t123_n183: return -0.13124299049377441f;
  t123_n184: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t123_n224;
  t123_n185: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t123_n217;
  t123_n186: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t123_n216;
  t123_n187: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t123_n215;
  t123_n188: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t123_n194;
  t123_n189: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t123_n193;
  t123_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t123_n192;
  t123_n191: return -0.0035382313653826714f;
  t123_n192: return 0.029733896255493164f;
  t123_n193: return 0.03734799101948738f;
  t123_n194: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t123_n214;
  t123_n195: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t123_n205;
  t123_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t123_n198;
  t123_n197: return -0.09894584119319916f;
  t123_n198: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t123_n204;
  t123_n199: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t123_n201;
  t123_n200: return -0.0030434238724410534f;
  t123_n201: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t123_n203;
  t123_n202: return -0.21593500673770905f;
  t123_n203: return -0.03691395744681358f;
  t123_n204: return 0.026983272284269333f;
  t123_n205: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t123_n207;
  t123_n206: return -0.46245768666267395f;
  t123_n207: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t123_n213;
  t123_n208: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t123_n210;
  t123_n209: return -0.011951551772654057f;
  t123_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t123_n212;
  t123_n211: return -0.6589498519897461f;
  t123_n212: return -0.06141754239797592f;
  t123_n213: return 0.10043363273143768f;
  t123_n214: return 0.030327312648296356f;
  t123_n215: return 0.0387357659637928f;
  t123_n216: return -0.03495293855667114f;
  t123_n217: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t123_n223;
  t123_n218: if (E.getNumReferences() >= 3292471296 /*764.5*/) goto t123_n222;
  t123_n219: if (E.getNumReferences() >= 3287826432 /*496.5*/) goto t123_n221;
  t123_n220: return 0.022603238001465797f;
  t123_n221: return 0.12463806569576263f;
  t123_n222: return -0.10447246581315994f;
  t123_n223: return -0.19923536479473114f;
  t123_n224: return 0.0009132710401900113f;
  t123_n225: return 0.015746375545859337f;
  t123_n226: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t123_n254;
  t123_n227: if (E.getNumReferences() >= 3349646912 /*85788.5*/) goto t123_n251;
  t123_n228: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t123_n230;
  t123_n229: return 0.0065354714170098305f;
  t123_n230: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t123_n246;
  t123_n231: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t123_n239;
  t123_n232: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t123_n238;
  t123_n233: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t123_n235;
  t123_n234: return 0.03711097687482834f;
  t123_n235: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t123_n237;
  t123_n236: return -0.3518964946269989f;
  t123_n237: return -0.02157897874712944f;
  t123_n238: return 0.05623637139797211f;
  t123_n239: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t123_n241;
  t123_n240: return 0.024599619209766388f;
  t123_n241: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t123_n245;
  t123_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t123_n244;
  t123_n243: return 0.016398636624217033f;
  t123_n244: return 0.09122700244188309f;
  t123_n245: return 0.054197363555431366f;
  t123_n246: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t123_n248;
  t123_n247: return 0.004633437376469374f;
  t123_n248: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t123_n250;
  t123_n249: return 0.10808533430099487f;
  t123_n250: return 0.04461992159485817f;
  t123_n251: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t123_n253;
  t123_n252: return 0.0918859913945198f;
  t123_n253: return 0.05634203925728798f;
  t123_n254: return -0.07258450984954834f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree124(const Example& E) {
  t124_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t124_n252;
  t124_n1: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t124_n129;
  t124_n2: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t124_n114;
  t124_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t124_n77;
  t124_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t124_n28;
  t124_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t124_n23;
  t124_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t124_n8;
  t124_n7: return -0.002327068941667676f;
  t124_n8: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t124_n20;
  t124_n9: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t124_n17;
  t124_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t124_n16;
  t124_n11: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t124_n13;
  t124_n12: return -0.0168082807213068f;
  t124_n13: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t124_n15;
  t124_n14: return -0.09146628528833389f;
  t124_n15: return 0.605949878692627f;
  t124_n16: return 0.034929338842630386f;
  t124_n17: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t124_n19;
  t124_n18: return 0.010657954961061478f;
  t124_n19: return -0.002407502382993698f;
  t124_n20: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t124_n22;
  t124_n21: return 0.027313342317938805f;
  t124_n22: return 0.011744608171284199f;
  t124_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n27;
  t124_n24: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t124_n26;
  t124_n25: return -0.03678084909915924f;
  t124_n26: return 0.09884210675954819f;
  t124_n27: return -0.021303845569491386f;
  t124_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t124_n46;
  t124_n29: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t124_n33;
  t124_n30: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t124_n32;
  t124_n31: return -0.004449688363820314f;
  t124_n32: return -0.02913655899465084f;
  t124_n33: if (E.getNumReferences() >= 3363896688 /*264171.5*/) goto t124_n43;
  t124_n34: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t124_n42;
  t124_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n37;
  t124_n36: return -0.006544506177306175f;
  t124_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n41;
  t124_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n40;
  t124_n39: return -0.08947605639696121f;
  t124_n40: return 0.01932956464588642f;
  t124_n41: return 0.005613049957901239f;
  t124_n42: return -0.036198753863573074f;
  t124_n43: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t124_n45;
  t124_n44: return -0.15468454360961914f;
  t124_n45: return 0.038875989615917206f;
  t124_n46: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t124_n76;
  t124_n47: if (E.getNumReferences() >= 3363896688 /*264171.5*/) goto t124_n73;
  t124_n48: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t124_n64;
  t124_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t124_n51;
  t124_n50: return 0.009416268207132816f;
  t124_n51: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t124_n59;
  t124_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t124_n54;
  t124_n53: return 0.024253148585557938f;
  t124_n54: if (E.getNumReferences() >= 3257663488 /*43.0*/) goto t124_n58;
  t124_n55: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t124_n57;
  t124_n56: return 0.10555897653102875f;
  t124_n57: return 0.34462350606918335f;
  t124_n58: return 0.2117031067609787f;
  t124_n59: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t124_n61;
  t124_n60: return 0.010658477433025837f;
  t124_n61: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t124_n63;
  t124_n62: return 0.03161128982901573f;
  t124_n63: return 0.07376376539468765f;
  t124_n64: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t124_n72;
  t124_n65: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t124_n71;
  t124_n66: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t124_n68;
  t124_n67: return 0.048825353384017944f;
  t124_n68: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t124_n70;
  t124_n69: return -0.02666870318353176f;
  t124_n70: return 0.10033959150314331f;
  t124_n71: return 0.024375759065151215f;
  t124_n72: return 0.00947010237723589f;
  t124_n73: if (E.getNumReferences() >= 3364365712 /*278828.5*/) goto t124_n75;
  t124_n74: return 0.07443199306726456f;
  t124_n75: return 0.017328433692455292f;
  t124_n76: return 0.0959937646985054f;
  t124_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t124_n81;
  t124_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t124_n80;
  t124_n79: return -0.018704036250710487f;
  t124_n80: return -0.057483647018671036f;
  t124_n81: if (E.getNumReferences() >= 3322147840 /*8445.0*/) goto t124_n109;
  t124_n82: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t124_n108;
  t124_n83: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t124_n107;
  t124_n84: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t124_n86;
  t124_n85: return 0.016793668270111084f;
  t124_n86: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t124_n96;
  t124_n87: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t124_n89;
  t124_n88: return 0.008042150177061558f;
  t124_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n95;
  t124_n90: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t124_n92;
  t124_n91: return -0.07437096536159515f;
  t124_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t124_n94;
  t124_n93: return -0.0077415392734110355f;
  t124_n94: return 0.08692494034767151f;
  t124_n95: return -0.023569855839014053f;
  t124_n96: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t124_n98;
  t124_n97: return 0.025675181299448013f;
  t124_n98: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t124_n102;
  t124_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n101;
  t124_n100: return -0.057800356298685074f;
  t124_n101: return -0.01057498436421156f;
  t124_n102: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t124_n106;
  t124_n103: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t124_n105;
  t124_n104: return 0.015474531799554825f;
  t124_n105: return 0.08696188032627106f;
  t124_n106: return -0.07606072723865509f;
  t124_n107: return -0.0014870587037876248f;
  t124_n108: return -0.0018315829802304506f;
  t124_n109: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t124_n111;
  t124_n110: return 0.028769806027412415f;
  t124_n111: if (E.getNumReferences() >= 3331140352 /*18069.5*/) goto t124_n113;
  t124_n112: return -0.10635702311992645f;
  t124_n113: return 0.21943867206573486f;
  t124_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t124_n116;
  t124_n115: return -0.10892795771360397f;
  t124_n116: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t124_n128;
  t124_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t124_n123;
  t124_n118: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t124_n120;
  t124_n119: return 0.002337029902264476f;
  t124_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t124_n122;
  t124_n121: return 0.016223499551415443f;
  t124_n122: return 0.0708024725317955f;
  t124_n123: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t124_n127;
  t124_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t124_n126;
  t124_n125: return -0.09885875135660172f;
  t124_n126: return 0.08879311382770538f;
  t124_n127: return 0.03361710533499718f;
  t124_n128: return -0.024230919778347015f;
  t124_n129: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t124_n251;
  t124_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n248;
  t124_n131: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t124_n173;
  t124_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t124_n158;
  t124_n133: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t124_n137;
  t124_n134: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t124_n136;
  t124_n135: return 4.711574365501292e-05f;
  t124_n136: return 0.035535380244255066f;
  t124_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t124_n143;
  t124_n138: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t124_n142;
  t124_n139: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t124_n141;
  t124_n140: return -0.10592134296894073f;
  t124_n141: return 0.002466740319505334f;
  t124_n142: return -0.3880917727947235f;
  t124_n143: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t124_n151;
  t124_n144: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t124_n146;
  t124_n145: return -0.028751591220498085f;
  t124_n146: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t124_n150;
  t124_n147: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t124_n149;
  t124_n148: return -0.1341622918844223f;
  t124_n149: return 0.017547788098454475f;
  t124_n150: return -0.04495736211538315f;
  t124_n151: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t124_n155;
  t124_n152: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t124_n154;
  t124_n153: return -0.01213759370148182f;
  t124_n154: return 0.10832519829273224f;
  t124_n155: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t124_n157;
  t124_n156: return 0.061121661216020584f;
  t124_n157: return 0.0025917303282767534f;
  t124_n158: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t124_n172;
  t124_n159: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t124_n171;
  t124_n160: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t124_n166;
  t124_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n163;
  t124_n162: return -0.06618612259626389f;
  t124_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n165;
  t124_n164: return -0.00981801189482212f;
  t124_n165: return 0.04134007543325424f;
  t124_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t124_n168;
  t124_n167: return 0.007498632185161114f;
  t124_n168: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t124_n170;
  t124_n169: return 0.06864845007658005f;
  t124_n170: return 0.033303868025541306f;
  t124_n171: return -0.016259023919701576f;
  t124_n172: return -0.0942402109503746f;
  t124_n173: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t124_n243;
  t124_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t124_n220;
  t124_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t124_n219;
  t124_n176: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t124_n212;
  t124_n177: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t124_n209;
  t124_n178: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t124_n204;
  t124_n179: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t124_n189;
  t124_n180: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t124_n184;
  t124_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t124_n183;
  t124_n182: return -0.014555301517248154f;
  t124_n183: return 0.03167393058538437f;
  t124_n184: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t124_n188;
  t124_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t124_n187;
  t124_n186: return -0.35373765230178833f;
  t124_n187: return -0.04297233372926712f;
  t124_n188: return 0.03846336901187897f;
  t124_n189: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t124_n197;
  t124_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n192;
  t124_n191: return -0.14409135282039642f;
  t124_n192: if (E.getFractionNameInContext() >= 3208762398 /*0.7571429014205933*/) goto t124_n194;
  t124_n193: return -0.0019239538814872503f;
  t124_n194: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t124_n196;
  t124_n195: return 0.08705304563045502f;
  t124_n196: return 0.011498223058879375f;
  t124_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t124_n201;
  t124_n198: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t124_n200;
  t124_n199: return 0.0008922623237594962f;
  t124_n200: return -0.13772675395011902f;
  t124_n201: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t124_n203;
  t124_n202: return 0.011244762688875198f;
  t124_n203: return 0.0597819983959198f;
  t124_n204: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t124_n208;
  t124_n205: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t124_n207;
  t124_n206: return 0.03727913647890091f;
  t124_n207: return -0.5149034857749939f;
  t124_n208: return -0.033513378351926804f;
  t124_n209: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t124_n211;
  t124_n210: return -0.04780978709459305f;
  t124_n211: return -0.0003918668953701854f;
  t124_n212: if (E.getFractionNameInContext() >= 3196336196 /*0.25824177265167236*/) goto t124_n218;
  t124_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n215;
  t124_n214: return -0.1177716925740242f;
  t124_n215: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t124_n217;
  t124_n216: return 0.10717064887285233f;
  t124_n217: return -0.11763907968997955f;
  t124_n218: return -0.14213517308235168f;
  t124_n219: return -0.014426414854824543f;
  t124_n220: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t124_n232;
  t124_n221: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t124_n231;
  t124_n222: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t124_n228;
  t124_n223: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t124_n227;
  t124_n224: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t124_n226;
  t124_n225: return -0.017891934141516685f;
  t124_n226: return 0.08110619336366653f;
  t124_n227: return -0.5625141859054565f;
  t124_n228: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t124_n230;
  t124_n229: return 0.01151618454605341f;
  t124_n230: return 0.1255285143852234f;
  t124_n231: return -0.10935372859239578f;
  t124_n232: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t124_n242;
  t124_n233: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t124_n235;
  t124_n234: return -0.0009468266507610679f;
  t124_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t124_n237;
  t124_n236: return 0.012249866500496864f;
  t124_n237: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t124_n239;
  t124_n238: return 0.0653151273727417f;
  t124_n239: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t124_n241;
  t124_n240: return 0.03393387794494629f;
  t124_n241: return 0.38723108172416687f;
  t124_n242: return 0.025721775367856026f;
  t124_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t124_n245;
  t124_n244: return -0.052553366869688034f;
  t124_n245: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t124_n247;
  t124_n246: return 0.07282716035842896f;
  t124_n247: return -0.08092911541461945f;
  t124_n248: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t124_n250;
  t124_n249: return 0.00534850126132369f;
  t124_n250: return 0.0009456591797061265f;
  t124_n251: return 0.012649516575038433f;
  t124_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t124_n254;
  t124_n253: return 0.0031684329733252525f;
  t124_n254: return 0.00763728329911828f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree125(const Example& E) {
  t125_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t125_n178;
  t125_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t125_n169;
  t125_n2: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t125_n164;
  t125_n3: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t125_n163;
  t125_n4: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t125_n150;
  t125_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t125_n7;
  t125_n6: return -0.10107261687517166f;
  t125_n7: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t125_n125;
  t125_n8: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t125_n92;
  t125_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t125_n63;
  t125_n10: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t125_n28;
  t125_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n17;
  t125_n12: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t125_n16;
  t125_n13: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t125_n15;
  t125_n14: return 0.0007730999495834112f;
  t125_n15: return 0.019076203927397728f;
  t125_n16: return -0.013817992992699146f;
  t125_n17: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t125_n19;
  t125_n18: return 0.003403385868296027f;
  t125_n19: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t125_n27;
  t125_n20: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t125_n26;
  t125_n21: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t125_n23;
  t125_n22: return -0.01557657215744257f;
  t125_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t125_n25;
  t125_n24: return 0.012503103353083134f;
  t125_n25: return -0.008057907223701477f;
  t125_n26: return 0.0388454794883728f;
  t125_n27: return -0.02502502128481865f;
  t125_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t125_n52;
  t125_n29: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t125_n37;
  t125_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n32;
  t125_n31: return -0.014054653234779835f;
  t125_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t125_n34;
  t125_n33: return 0.0007515914039686322f;
  t125_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t125_n36;
  t125_n35: return 0.07785360515117645f;
  t125_n36: return 0.017687972635030746f;
  t125_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t125_n51;
  t125_n38: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t125_n50;
  t125_n39: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t125_n41;
  t125_n40: return -0.006171678192913532f;
  t125_n41: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t125_n43;
  t125_n42: return 0.03170068934559822f;
  t125_n43: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t125_n47;
  t125_n44: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t125_n46;
  t125_n45: return 0.045737460255622864f;
  t125_n46: return 0.06998492032289505f;
  t125_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t125_n49;
  t125_n48: return -0.11560098081827164f;
  t125_n49: return -0.0006458893185481429f;
  t125_n50: return -0.002057549776509404f;
  t125_n51: return 0.03264964371919632f;
  t125_n52: if (E.getNumReferences() >= 3301359616 /*1590.0*/) goto t125_n62;
  t125_n53: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t125_n55;
  t125_n54: return 0.022442366927862167f;
  t125_n55: if (E.getNumReferences() >= 3293691904 /*839.0*/) goto t125_n59;
  t125_n56: if (E.getNumReferences() >= 3293364224 /*819.0*/) goto t125_n58;
  t125_n57: return 0.0854945182800293f;
  t125_n58: return 0.19883014261722565f;
  t125_n59: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t125_n61;
  t125_n60: return -0.6152868866920471f;
  t125_n61: return 0.017520271241664886f;
  t125_n62: return -0.17719106376171112f;
  t125_n63: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t125_n89;
  t125_n64: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t125_n82;
  t125_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n67;
  t125_n66: return -0.04714240878820419f;
  t125_n67: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t125_n81;
  t125_n68: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t125_n80;
  t125_n69: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t125_n79;
  t125_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t125_n78;
  t125_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t125_n73;
  t125_n72: return -0.03928442671895027f;
  t125_n73: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t125_n77;
  t125_n74: if (E.getNumReferences() >= 3281518592 /*304.0*/) goto t125_n76;
  t125_n75: return 0.05575950816273689f;
  t125_n76: return 0.13117925822734833f;
  t125_n77: return -0.06106317415833473f;
  t125_n78: return 0.05416394770145416f;
  t125_n79: return -0.018319953233003616f;
  t125_n80: return 0.03769192472100258f;
  t125_n81: return -0.04012180119752884f;
  t125_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t125_n84;
  t125_n83: return 0.015766914933919907f;
  t125_n84: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t125_n88;
  t125_n85: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t125_n87;
  t125_n86: return 0.1919398307800293f;
  t125_n87: return -0.048677559942007065f;
  t125_n88: return 0.04530778527259827f;
  t125_n89: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t125_n91;
  t125_n90: return 0.05079275369644165f;
  t125_n91: return 0.02350427769124508f;
  t125_n92: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t125_n94;
  t125_n93: return -0.00015315190830733627f;
  t125_n94: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t125_n114;
  t125_n95: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t125_n113;
  t125_n96: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t125_n98;
  t125_n97: return -0.011522210203111172f;
  t125_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n112;
  t125_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t125_n105;
  t125_n100: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t125_n104;
  t125_n101: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t125_n103;
  t125_n102: return 0.20169775187969208f;
  t125_n103: return -0.3830605447292328f;
  t125_n104: return -0.4954158067703247f;
  t125_n105: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t125_n111;
  t125_n106: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t125_n108;
  t125_n107: return 0.04541449621319771f;
  t125_n108: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t125_n110;
  t125_n109: return 0.1811959594488144f;
  t125_n110: return -0.028752600774168968f;
  t125_n111: return -0.09853939712047577f;
  t125_n112: return 0.03583687171339989f;
  t125_n113: return 0.008890224620699883f;
  t125_n114: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t125_n124;
  t125_n115: if (E.getNumReferences() >= 3305914368 /*2244.0*/) goto t125_n117;
  t125_n116: return -0.0031359035056084394f;
  t125_n117: if (E.getNumReferences() >= 3305973760 /*2258.5*/) goto t125_n123;
  t125_n118: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t125_n122;
  t125_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t125_n121;
  t125_n120: return 0.08904335647821426f;
  t125_n121: return 0.24983547627925873f;
  t125_n122: return -0.19252680242061615f;
  t125_n123: return 0.06961623579263687f;
  t125_n124: return -0.037794049829244614f;
  t125_n125: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t125_n127;
  t125_n126: return 0.012629233300685883f;
  t125_n127: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t125_n129;
  t125_n128: return 0.14637233316898346f;
  t125_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t125_n139;
  t125_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n138;
  t125_n131: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t125_n137;
  t125_n132: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t125_n134;
  t125_n133: return -0.09433291107416153f;
  t125_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n136;
  t125_n135: return -0.4741767346858978f;
  t125_n136: return 0.016636637970805168f;
  t125_n137: return 0.039939068257808685f;
  t125_n138: return -0.00047201229608617723f;
  t125_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n143;
  t125_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n142;
  t125_n141: return -0.03168054670095444f;
  t125_n142: return -0.00247525074519217f;
  t125_n143: if (E.getNumReferences() >= 3306864640 /*2476.0*/) goto t125_n149;
  t125_n144: if (E.getNumReferences() >= 3306846208 /*2471.5*/) goto t125_n146;
  t125_n145: return 0.08263277262449265f;
  t125_n146: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t125_n148;
  t125_n147: return 0.1202641949057579f;
  t125_n148: return 0.20148883759975433f;
  t125_n149: return 0.017760321497917175f;
  t125_n150: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t125_n162;
  t125_n151: if (E.getNumReferences() >= 3298988032 /*1300.5*/) goto t125_n159;
  t125_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n158;
  t125_n153: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t125_n155;
  t125_n154: return 0.2203696370124817f;
  t125_n155: if (E.getNumReferences() >= 3288735744 /*536.5*/) goto t125_n157;
  t125_n156: return -0.008415140211582184f;
  t125_n157: return 0.10865337401628494f;
  t125_n158: return 0.006100551690906286f;
  t125_n159: if (E.getNumReferences() >= 3299151872 /*1320.5*/) goto t125_n161;
  t125_n160: return 0.12720909714698792f;
  t125_n161: return 0.04043447971343994f;
  t125_n162: return -0.03695666044950485f;
  t125_n163: return -0.07280039042234421f;
  t125_n164: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t125_n166;
  t125_n165: return -0.022721100598573685f;
  t125_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n168;
  t125_n167: return -0.07722233235836029f;
  t125_n168: return -0.03192640468478203f;
  t125_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t125_n177;
  t125_n170: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t125_n174;
  t125_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n173;
  t125_n172: return -0.1483989655971527f;
  t125_n173: return 0.005968847777694464f;
  t125_n174: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t125_n176;
  t125_n175: return -0.028539301827549934f;
  t125_n176: return -0.4870975613594055f;
  t125_n177: return 0.04020145535469055f;
  t125_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t125_n254;
  t125_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t125_n211;
  t125_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n192;
  t125_n181: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t125_n183;
  t125_n182: return -0.04130930453538895f;
  t125_n183: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t125_n185;
  t125_n184: return -0.02647601068019867f;
  t125_n185: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t125_n187;
  t125_n186: return 0.02351347543299198f;
  t125_n187: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t125_n191;
  t125_n188: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t125_n190;
  t125_n189: return 0.09682033210992813f;
  t125_n190: return 0.05628706514835358f;
  t125_n191: return -0.003882678924128413f;
  t125_n192: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t125_n194;
  t125_n193: return -0.018368801102042198f;
  t125_n194: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t125_n198;
  t125_n195: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t125_n197;
  t125_n196: return -0.0226959940046072f;
  t125_n197: return 0.013899204321205616f;
  t125_n198: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t125_n208;
  t125_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n207;
  t125_n200: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t125_n202;
  t125_n201: return -0.005562970414757729f;
  t125_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n204;
  t125_n203: return -0.1022999957203865f;
  t125_n204: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t125_n206;
  t125_n205: return 0.17856843769550323f;
  t125_n206: return 0.05218985304236412f;
  t125_n207: return 0.009302161633968353f;
  t125_n208: if (E.getFractionNameInContext() >= 3208871340 /*0.7636363506317139*/) goto t125_n210;
  t125_n209: return -0.1185946986079216f;
  t125_n210: return 0.13227279484272003f;
  t125_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n213;
  t125_n212: return -0.03173323720693588f;
  t125_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t125_n221;
  t125_n214: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t125_n220;
  t125_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t125_n219;
  t125_n216: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t125_n218;
  t125_n217: return -0.03291019797325134f;
  t125_n218: return 0.09008649736642838f;
  t125_n219: return 0.005141392350196838f;
  t125_n220: return 0.10437838733196259f;
  t125_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t125_n243;
  t125_n222: if (E.getFractionNameInContext() >= 3211998003 /*0.949999988079071*/) goto t125_n224;
  t125_n223: return 0.015595413744449615f;
  t125_n224: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t125_n242;
  t125_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t125_n239;
  t125_n226: if (E.getFileProximityDistanceCost() >= 3238002688 /*8.0*/) goto t125_n228;
  t125_n227: return -0.2928027808666229f;
  t125_n228: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t125_n238;
  t125_n229: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t125_n237;
  t125_n230: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t125_n236;
  t125_n231: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t125_n233;
  t125_n232: return 0.16829155385494232f;
  t125_n233: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t125_n235;
  t125_n234: return 0.1426202356815338f;
  t125_n235: return 0.19156986474990845f;
  t125_n236: return 0.0627661719918251f;
  t125_n237: return 0.010323403403162956f;
  t125_n238: return 0.06674587726593018f;
  t125_n239: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t125_n241;
  t125_n240: return -0.16547133028507233f;
  t125_n241: return 0.09863774478435516f;
  t125_n242: return 0.13433782756328583f;
  t125_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t125_n249;
  t125_n244: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t125_n246;
  t125_n245: return 0.012676683254539967f;
  t125_n246: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t125_n248;
  t125_n247: return -0.15509632229804993f;
  t125_n248: return 0.060462746769189835f;
  t125_n249: if (E.getFractionNameInContext() >= 3176353077 /*0.0515873022377491*/) goto t125_n253;
  t125_n250: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t125_n252;
  t125_n251: return 0.03358397260308266f;
  t125_n252: return 0.027090756222605705f;
  t125_n253: return 0.013358557596802711f;
  t125_n254: return -0.014023395255208015f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree126(const Example& E) {
  t126_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t126_n180;
  t126_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t126_n75;
  t126_n2: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t126_n4;
  t126_n3: return -0.0016624314012005925f;
  t126_n4: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t126_n20;
  t126_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t126_n19;
  t126_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n16;
  t126_n7: if (E.getNumReferences() >= 3309969408 /*3234.0*/) goto t126_n11;
  t126_n8: if (E.getNumReferences() >= 3309606912 /*3145.5*/) goto t126_n10;
  t126_n9: return -0.018572473898530006f;
  t126_n10: return 0.1268555372953415f;
  t126_n11: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t126_n13;
  t126_n12: return -0.23152711987495422f;
  t126_n13: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t126_n15;
  t126_n14: return 0.12897491455078125f;
  t126_n15: return -0.09710296243429184f;
  t126_n16: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t126_n18;
  t126_n17: return 0.055497460067272186f;
  t126_n18: return 0.014898300170898438f;
  t126_n19: return -0.0020708227530121803f;
  t126_n20: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t126_n74;
  t126_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t126_n23;
  t126_n22: return -0.0010744292521849275f;
  t126_n23: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t126_n27;
  t126_n24: if (E.getNumReferences() >= 3333772544 /*23210.5*/) goto t126_n26;
  t126_n25: return -0.04297608882188797f;
  t126_n26: return 0.014186015352606773f;
  t126_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t126_n29;
  t126_n28: return -0.05037490278482437f;
  t126_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t126_n73;
  t126_n30: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t126_n72;
  t126_n31: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t126_n45;
  t126_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n34;
  t126_n33: return -0.13670773804187775f;
  t126_n34: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t126_n36;
  t126_n35: return 0.13736645877361298f;
  t126_n36: if (E.getNumReferences() >= 3308046336 /*2764.5*/) goto t126_n44;
  t126_n37: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t126_n39;
  t126_n38: return 0.11052005738019943f;
  t126_n39: if (E.getNumReferences() >= 3307317248 /*2586.5*/) goto t126_n41;
  t126_n40: return 0.14428988099098206f;
  t126_n41: if (E.getNumReferences() >= 3307757568 /*2694.0*/) goto t126_n43;
  t126_n42: return -0.11410806328058243f;
  t126_n43: return 0.14415474236011505f;
  t126_n44: return 0.04715441167354584f;
  t126_n45: if (E.getNumReferences() >= 3333346816 /*22379.0*/) goto t126_n67;
  t126_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t126_n50;
  t126_n47: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t126_n49;
  t126_n48: return 0.15468481183052063f;
  t126_n49: return -0.06696020811796188f;
  t126_n50: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t126_n52;
  t126_n51: return 0.09820949286222458f;
  t126_n52: if (E.getNumReferences() >= 3326347776 /*12546.5*/) goto t126_n60;
  t126_n53: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t126_n55;
  t126_n54: return 0.3571297526359558f;
  t126_n55: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t126_n57;
  t126_n56: return -0.10237313061952591f;
  t126_n57: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t126_n59;
  t126_n58: return 0.13512086868286133f;
  t126_n59: return -0.13063950836658478f;
  t126_n60: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t126_n66;
  t126_n61: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t126_n65;
  t126_n62: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t126_n64;
  t126_n63: return 0.13576605916023254f;
  t126_n64: return 0.15415269136428833f;
  t126_n65: return -0.16655570268630981f;
  t126_n66: return 0.14657340943813324f;
  t126_n67: if (E.getNumReferences() >= 3340838912 /*41256.0*/) goto t126_n69;
  t126_n68: return -0.005441601388156414f;
  t126_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n71;
  t126_n70: return -0.10696125775575638f;
  t126_n71: return 0.10755780339241028f;
  t126_n72: return -0.03888782486319542f;
  t126_n73: return 0.006984378211200237f;
  t126_n74: return 0.19392986595630646f;
  t126_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t126_n171;
  t126_n76: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t126_n152;
  t126_n77: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t126_n149;
  t126_n78: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t126_n148;
  t126_n79: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t126_n81;
  t126_n80: return -0.001990227960050106f;
  t126_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t126_n123;
  t126_n82: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t126_n84;
  t126_n83: return -0.02885497361421585f;
  t126_n84: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t126_n122;
  t126_n85: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t126_n97;
  t126_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n88;
  t126_n87: return -0.004519737791270018f;
  t126_n88: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t126_n92;
  t126_n89: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t126_n91;
  t126_n90: return -0.04726670682430267f;
  t126_n91: return 0.03371777385473251f;
  t126_n92: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t126_n94;
  t126_n93: return -0.12438582628965378f;
  t126_n94: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t126_n96;
  t126_n95: return 0.12362448871135712f;
  t126_n96: return 0.11637271195650101f;
  t126_n97: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t126_n121;
  t126_n98: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t126_n118;
  t126_n99: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t126_n113;
  t126_n100: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t126_n110;
  t126_n101: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t126_n109;
  t126_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t126_n108;
  t126_n103: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t126_n105;
  t126_n104: return -0.008523523807525635f;
  t126_n105: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t126_n107;
  t126_n106: return 0.11129654198884964f;
  t126_n107: return 0.0488070473074913f;
  t126_n108: return 0.08495768904685974f;
  t126_n109: return -0.3647865355014801f;
  t126_n110: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t126_n112;
  t126_n111: return -0.379285991191864f;
  t126_n112: return -0.06276345252990723f;
  t126_n113: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t126_n117;
  t126_n114: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t126_n116;
  t126_n115: return 0.11746573448181152f;
  t126_n116: return 0.03870555758476257f;
  t126_n117: return 0.03324376791715622f;
  t126_n118: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t126_n120;
  t126_n119: return -0.051565755158662796f;
  t126_n120: return 0.10267037153244019f;
  t126_n121: return -0.3757169842720032f;
  t126_n122: return -0.035019855946302414f;
  t126_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t126_n127;
  t126_n124: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t126_n126;
  t126_n125: return 0.21126312017440796f;
  t126_n126: return 0.008630182594060898f;
  t126_n127: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t126_n141;
  t126_n128: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t126_n140;
  t126_n129: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t126_n139;
  t126_n130: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t126_n132;
  t126_n131: return 0.010281368158757687f;
  t126_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n134;
  t126_n133: return -0.2568714916706085f;
  t126_n134: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t126_n138;
  t126_n135: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t126_n137;
  t126_n136: return 0.1341576725244522f;
  t126_n137: return -0.2884790003299713f;
  t126_n138: return 0.14004527032375336f;
  t126_n139: return -0.05477839708328247f;
  t126_n140: return 0.08144064992666245f;
  t126_n141: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t126_n145;
  t126_n142: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t126_n144;
  t126_n143: return 0.07907228916883469f;
  t126_n144: return -0.5230880379676819f;
  t126_n145: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t126_n147;
  t126_n146: return -0.03400474414229393f;
  t126_n147: return 0.08872359991073608f;
  t126_n148: return -0.009437011554837227f;
  t126_n149: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t126_n151;
  t126_n150: return 0.00476189935579896f;
  t126_n151: return 0.1229759156703949f;
  t126_n152: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t126_n170;
  t126_n153: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t126_n165;
  t126_n154: if (E.getNumReferences() >= 3299520512 /*1365.5*/) goto t126_n164;
  t126_n155: if (E.getNumReferences() >= 3299295232 /*1338.0*/) goto t126_n157;
  t126_n156: return 0.04369893670082092f;
  t126_n157: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t126_n159;
  t126_n158: return 0.13037189841270447f;
  t126_n159: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t126_n163;
  t126_n160: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t126_n162;
  t126_n161: return 0.12705089151859283f;
  t126_n162: return -0.19981591403484344f;
  t126_n163: return -0.04668193310499191f;
  t126_n164: return 0.020675111562013626f;
  t126_n165: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t126_n167;
  t126_n166: return -0.01559891365468502f;
  t126_n167: if (E.getNumReferences() >= 3308009472 /*2755.5*/) goto t126_n169;
  t126_n168: return 0.11995040625333786f;
  t126_n169: return 0.10674373060464859f;
  t126_n170: return 0.05111749842762947f;
  t126_n171: if (E.getFractionNameInContext() >= 3204006750 /*0.4868420958518982*/) goto t126_n179;
  t126_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace))) goto t126_n174;
  t126_n173: return 0.0014802623772993684f;
  t126_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t126_n176;
  t126_n175: return -0.025657739490270615f;
  t126_n176: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t126_n178;
  t126_n177: return 0.023955637589097023f;
  t126_n178: return -0.1118156686425209f;
  t126_n179: return 0.00749246496707201f;
  t126_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t126_n248;
  t126_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t126_n243;
  t126_n182: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t126_n210;
  t126_n183: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t126_n185;
  t126_n184: return 0.005932012107223272f;
  t126_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t126_n191;
  t126_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t126_n190;
  t126_n187: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t126_n189;
  t126_n188: return 0.033427558839321136f;
  t126_n189: return -0.1933455765247345f;
  t126_n190: return -0.22850088775157928f;
  t126_n191: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t126_n201;
  t126_n192: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t126_n198;
  t126_n193: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t126_n197;
  t126_n194: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t126_n196;
  t126_n195: return 0.04493116959929466f;
  t126_n196: return 0.10368930548429489f;
  t126_n197: return 0.09658420830965042f;
  t126_n198: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t126_n200;
  t126_n199: return 0.12887097895145416f;
  t126_n200: return 0.003896588459610939f;
  t126_n201: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t126_n205;
  t126_n202: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t126_n204;
  t126_n203: return -0.05351674184203148f;
  t126_n204: return 0.08442805707454681f;
  t126_n205: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t126_n207;
  t126_n206: return 0.11821084469556808f;
  t126_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t126_n209;
  t126_n208: return -0.4770536720752716f;
  t126_n209: return 0.08464516699314117f;
  t126_n210: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t126_n228;
  t126_n211: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t126_n213;
  t126_n212: return -0.029881024733185768f;
  t126_n213: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t126_n225;
  t126_n214: if (E.getFractionNameInContext() >= 3182766931 /*0.08846154063940048*/) goto t126_n224;
  t126_n215: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t126_n219;
  t126_n216: if (E.getFractionNameInContext() >= 3172829657 /*0.03846153989434242*/) goto t126_n218;
  t126_n217: return -0.05383998155593872f;
  t126_n218: return 0.11546443402767181f;
  t126_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t126_n221;
  t126_n220: return 0.14772655069828033f;
  t126_n221: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t126_n223;
  t126_n222: return 0.09142053872346878f;
  t126_n223: return -0.01157040148973465f;
  t126_n224: return 0.004644074477255344f;
  t126_n225: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t126_n227;
  t126_n226: return -0.01240998599678278f;
  t126_n227: return 0.08869067579507828f;
  t126_n228: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t126_n236;
  t126_n229: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t126_n233;
  t126_n230: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t126_n232;
  t126_n231: return -0.03833260014653206f;
  t126_n232: return 0.09035608172416687f;
  t126_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t126_n235;
  t126_n234: return -0.034839726984500885f;
  t126_n235: return 0.03538915142416954f;
  t126_n236: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t126_n238;
  t126_n237: return -0.0016990706790238619f;
  t126_n238: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t126_n240;
  t126_n239: return 0.02448938973248005f;
  t126_n240: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t126_n242;
  t126_n241: return 0.2459680587053299f;
  t126_n242: return 0.037849243730306625f;
  t126_n243: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t126_n247;
  t126_n244: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t126_n246;
  t126_n245: return 0.01702594943344593f;
  t126_n246: return -0.008689844980835915f;
  t126_n247: return 0.12029630690813065f;
  t126_n248: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t126_n252;
  t126_n249: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t126_n251;
  t126_n250: return 0.006226181983947754f;
  t126_n251: return -0.0617864690721035f;
  t126_n252: if (E.getNumReferences() >= 3306696704 /*2435.0*/) goto t126_n254;
  t126_n253: return 0.11483565717935562f;
  t126_n254: return 0.032201603055000305f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree127(const Example& E) {
  t127_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t127_n252;
  t127_n1: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t127_n231;
  t127_n2: if (E.getNumReferences() >= 3350789120 /*94712.0*/) goto t127_n226;
  t127_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t127_n13;
  t127_n4: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t127_n12;
  t127_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n11;
  t127_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t127_n10;
  t127_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t127_n9;
  t127_n8: return -0.12522552907466888f;
  t127_n9: return 0.05767285078763962f;
  t127_n10: return -0.002975654788315296f;
  t127_n11: return 0.0009965458884835243f;
  t127_n12: return 0.0060614896938204765f;
  t127_n13: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t127_n15;
  t127_n14: return 0.00735706090927124f;
  t127_n15: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t127_n119;
  t127_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n32;
  t127_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t127_n27;
  t127_n18: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t127_n20;
  t127_n19: return -0.034750133752822876f;
  t127_n20: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t127_n24;
  t127_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n23;
  t127_n22: return -0.49226096272468567f;
  t127_n23: return 0.22518962621688843f;
  t127_n24: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t127_n26;
  t127_n25: return -0.1119447872042656f;
  t127_n26: return 0.07593152672052383f;
  t127_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t127_n29;
  t127_n28: return -0.019270317628979683f;
  t127_n29: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t127_n31;
  t127_n30: return 0.03039899282157421f;
  t127_n31: return -0.02967793121933937f;
  t127_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t127_n38;
  t127_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n35;
  t127_n34: return -0.019972670823335648f;
  t127_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t127_n37;
  t127_n36: return 0.23504610359668732f;
  t127_n37: return 0.1194930225610733f;
  t127_n38: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t127_n118;
  t127_n39: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t127_n117;
  t127_n40: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t127_n52;
  t127_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n47;
  t127_n42: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t127_n46;
  t127_n43: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t127_n45;
  t127_n44: return 0.03630492836236954f;
  t127_n45: return -0.04951898381114006f;
  t127_n46: return 0.024041756987571716f;
  t127_n47: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t127_n51;
  t127_n48: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t127_n50;
  t127_n49: return -0.0320163331925869f;
  t127_n50: return 0.01724916510283947f;
  t127_n51: return -0.004893117118626833f;
  t127_n52: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t127_n60;
  t127_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n55;
  t127_n54: return -0.04657662659883499f;
  t127_n55: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t127_n59;
  t127_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n58;
  t127_n57: return -0.034317828714847565f;
  t127_n58: return 0.04637395218014717f;
  t127_n59: return 0.00412809057161212f;
  t127_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t127_n84;
  t127_n61: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t127_n65;
  t127_n62: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t127_n64;
  t127_n63: return -0.010772663168609142f;
  t127_n64: return 0.011592444963753223f;
  t127_n65: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t127_n69;
  t127_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n68;
  t127_n67: return -0.0251948032528162f;
  t127_n68: return 0.01896762102842331f;
  t127_n69: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t127_n83;
  t127_n70: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t127_n82;
  t127_n71: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t127_n79;
  t127_n72: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t127_n78;
  t127_n73: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t127_n77;
  t127_n74: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t127_n76;
  t127_n75: return 0.005663914140313864f;
  t127_n76: return -0.002095353789627552f;
  t127_n77: return 0.0314520001411438f;
  t127_n78: return -0.006056106649339199f;
  t127_n79: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t127_n81;
  t127_n80: return 0.04049789533019066f;
  t127_n81: return -0.09823276102542877f;
  t127_n82: return -0.016152985394001007f;
  t127_n83: return 0.047187164425849915f;
  t127_n84: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t127_n100;
  t127_n85: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t127_n95;
  t127_n86: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t127_n92;
  t127_n87: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t127_n91;
  t127_n88: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t127_n90;
  t127_n89: return 0.021728480234742165f;
  t127_n90: return -0.028964461758732796f;
  t127_n91: return 0.030207747593522072f;
  t127_n92: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t127_n94;
  t127_n93: return -0.0840420126914978f;
  t127_n94: return 0.05658329650759697f;
  t127_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n97;
  t127_n96: return -0.07628229260444641f;
  t127_n97: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t127_n99;
  t127_n98: return 0.046296052634716034f;
  t127_n99: return 0.11195947974920273f;
  t127_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n108;
  t127_n101: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t127_n103;
  t127_n102: return -0.10195588320493698f;
  t127_n103: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t127_n107;
  t127_n104: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t127_n106;
  t127_n105: return -0.01674666255712509f;
  t127_n106: return 0.02548973634839058f;
  t127_n107: return 0.12529131770133972f;
  t127_n108: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t127_n116;
  t127_n109: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t127_n113;
  t127_n110: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t127_n112;
  t127_n111: return 0.05776024982333183f;
  t127_n112: return -4.078963279724121f;
  t127_n113: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t127_n115;
  t127_n114: return 0.07871928066015244f;
  t127_n115: return -0.34873783588409424f;
  t127_n116: return 0.12906621396541595f;
  t127_n117: return 0.03552021458745003f;
  t127_n118: return -0.021520335227251053f;
  t127_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t127_n209;
  t127_n120: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t127_n128;
  t127_n121: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t127_n127;
  t127_n122: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t127_n126;
  t127_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t127_n125;
  t127_n124: return 0.001352529739961028f;
  t127_n125: return 0.07269033044576645f;
  t127_n126: return 0.09820578992366791f;
  t127_n127: return -0.011772729456424713f;
  t127_n128: if (E.getNumReferences() >= 3338535680 /*32513.5*/) goto t127_n208;
  t127_n129: if (E.getNumReferences() >= 3336381440 /*28306.0*/) goto t127_n207;
  t127_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n132;
  t127_n131: return -0.0034663155674934387f;
  t127_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator))) goto t127_n194;
  t127_n133: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t127_n189;
  t127_n134: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t127_n160;
  t127_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t127_n145;
  t127_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t127_n138;
  t127_n137: return -0.014261007308959961f;
  t127_n138: if (E.getNumReferences() >= 3331854592 /*19464.5*/) goto t127_n144;
  t127_n139: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t127_n143;
  t127_n140: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t127_n142;
  t127_n141: return 0.08857140690088272f;
  t127_n142: return 0.2097034752368927f;
  t127_n143: return 0.025473980233073235f;
  t127_n144: return 0.11872857064008713f;
  t127_n145: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t127_n153;
  t127_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t127_n148;
  t127_n147: return 0.006418443750590086f;
  t127_n148: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t127_n150;
  t127_n149: return 0.036181047558784485f;
  t127_n150: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t127_n152;
  t127_n151: return 0.05683870613574982f;
  t127_n152: return 0.1568448692560196f;
  t127_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t127_n155;
  t127_n154: return 0.010399160906672478f;
  t127_n155: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t127_n159;
  t127_n156: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t127_n158;
  t127_n157: return 0.04520072042942047f;
  t127_n158: return 0.14123205840587616f;
  t127_n159: return 0.0339239239692688f;
  t127_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t127_n162;
  t127_n161: return -0.022695478051900864f;
  t127_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t127_n188;
  t127_n163: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t127_n187;
  t127_n164: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t127_n176;
  t127_n165: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t127_n175;
  t127_n166: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t127_n174;
  t127_n167: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t127_n173;
  t127_n168: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t127_n172;
  t127_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n171;
  t127_n170: return -0.007051641587167978f;
  t127_n171: return 0.037026308476924896f;
  t127_n172: return 0.11083546280860901f;
  t127_n173: return -0.04150198772549629f;
  t127_n174: return 0.03470772132277489f;
  t127_n175: return -0.05864673852920532f;
  t127_n176: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t127_n180;
  t127_n177: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t127_n179;
  t127_n178: return 0.04815496876835823f;
  t127_n179: return 0.10496639460325241f;
  t127_n180: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t127_n182;
  t127_n181: return -0.07368117570877075f;
  t127_n182: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t127_n184;
  t127_n183: return 0.06419387459754944f;
  t127_n184: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t127_n186;
  t127_n185: return -0.15529769659042358f;
  t127_n186: return 0.04564548656344414f;
  t127_n187: return -0.10463807731866837f;
  t127_n188: return -0.029267923906445503f;
  t127_n189: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t127_n191;
  t127_n190: return -0.03012293390929699f;
  t127_n191: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t127_n193;
  t127_n192: return -0.03923351317644119f;
  t127_n193: return 0.06954441964626312f;
  t127_n194: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t127_n196;
  t127_n195: return 0.019501956179738045f;
  t127_n196: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t127_n206;
  t127_n197: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t127_n199;
  t127_n198: return 0.013275510631501675f;
  t127_n199: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t127_n205;
  t127_n200: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t127_n204;
  t127_n201: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t127_n203;
  t127_n202: return 0.030957257375121117f;
  t127_n203: return -0.021731168031692505f;
  t127_n204: return 0.03656211867928505f;
  t127_n205: return -0.12010848522186279f;
  t127_n206: return -0.01495064701884985f;
  t127_n207: return -0.059707093983888626f;
  t127_n208: return 0.017153823748230934f;
  t127_n209: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t127_n221;
  t127_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t127_n220;
  t127_n211: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t127_n213;
  t127_n212: return -0.0015570458490401506f;
  t127_n213: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t127_n217;
  t127_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n216;
  t127_n215: return -0.0457843653857708f;
  t127_n216: return -0.06871774047613144f;
  t127_n217: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t127_n219;
  t127_n218: return 0.056678447872400284f;
  t127_n219: return 0.14841961860656738f;
  t127_n220: return 0.011446679010987282f;
  t127_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t127_n225;
  t127_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t127_n224;
  t127_n223: return -0.027332454919815063f;
  t127_n224: return 0.0023757184389978647f;
  t127_n225: return 0.029452266171574593f;
  t127_n226: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t127_n230;
  t127_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n229;
  t127_n228: return -0.1506715714931488f;
  t127_n229: return -0.1545838564634323f;
  t127_n230: return -0.2765735685825348f;
  t127_n231: if (E.getNumReferences() >= 3351669952 /*101593.5*/) goto t127_n247;
  t127_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t127_n236;
  t127_n233: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t127_n235;
  t127_n234: return 0.08888457715511322f;
  t127_n235: return 0.033183347433805466f;
  t127_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t127_n240;
  t127_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t127_n239;
  t127_n238: return -0.005546011961996555f;
  t127_n239: return 0.10753721743822098f;
  t127_n240: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t127_n242;
  t127_n241: return 0.06083684787154198f;
  t127_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n244;
  t127_n243: return -0.11833161115646362f;
  t127_n244: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t127_n246;
  t127_n245: return 0.11477165669202805f;
  t127_n246: return 0.09683886170387268f;
  t127_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t127_n251;
  t127_n248: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t127_n250;
  t127_n249: return -0.0437840074300766f;
  t127_n250: return 0.0788964331150055f;
  t127_n251: return 0.007307353895157576f;
  t127_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t127_n254;
  t127_n253: return 0.002513736952096224f;
  t127_n254: return 0.008136741816997528f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree128(const Example& E) {
  t128_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t128_n184;
  t128_n1: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t128_n181;
  t128_n2: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t128_n88;
  t128_n3: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t128_n69;
  t128_n4: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t128_n6;
  t128_n5: return -0.002749080304056406f;
  t128_n6: if (E.getFractionNameInContext() >= 3173036146 /*0.039230771362781525*/) goto t128_n8;
  t128_n7: return 0.41247108578681946f;
  t128_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro))) goto t128_n20;
  t128_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t128_n11;
  t128_n10: return -0.0029140650294721127f;
  t128_n11: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t128_n15;
  t128_n12: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n14;
  t128_n13: return 0.007861419580876827f;
  t128_n14: return -0.001390197197906673f;
  t128_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace))) goto t128_n17;
  t128_n16: return -0.0014802414225414395f;
  t128_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t128_n19;
  t128_n18: return -0.10552618652582169f;
  t128_n19: return 0.01786086894571781f;
  t128_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t128_n38;
  t128_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n33;
  t128_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n26;
  t128_n23: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t128_n25;
  t128_n24: return -0.03960204869508743f;
  t128_n25: return -0.17784470319747925f;
  t128_n26: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t128_n30;
  t128_n27: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t128_n29;
  t128_n28: return 0.1420530378818512f;
  t128_n29: return -0.05909198895096779f;
  t128_n30: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t128_n32;
  t128_n31: return -0.010153657756745815f;
  t128_n32: return 0.002136986004188657f;
  t128_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t128_n37;
  t128_n34: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t128_n36;
  t128_n35: return 0.036244772374629974f;
  t128_n36: return 0.06919911503791809f;
  t128_n37: return 0.01845489628612995f;
  t128_n38: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t128_n48;
  t128_n39: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t128_n47;
  t128_n40: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t128_n44;
  t128_n41: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t128_n43;
  t128_n42: return 0.013858671300113201f;
  t128_n43: return 0.0695246160030365f;
  t128_n44: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t128_n46;
  t128_n45: return -0.002336552133783698f;
  t128_n46: return 0.018576625734567642f;
  t128_n47: return -0.014136821031570435f;
  t128_n48: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t128_n50;
  t128_n49: return -0.012623968534171581f;
  t128_n50: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t128_n66;
  t128_n51: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t128_n53;
  t128_n52: return -0.033140093088150024f;
  t128_n53: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t128_n59;
  t128_n54: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t128_n56;
  t128_n55: return 0.0011690022656694055f;
  t128_n56: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t128_n58;
  t128_n57: return 0.01760183647274971f;
  t128_n58: return 0.04095606878399849f;
  t128_n59: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t128_n61;
  t128_n60: return 0.01116863451898098f;
  t128_n61: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t128_n63;
  t128_n62: return 0.0784081295132637f;
  t128_n63: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t128_n65;
  t128_n64: return -0.0982820987701416f;
  t128_n65: return 0.07329436391592026f;
  t128_n66: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t128_n68;
  t128_n67: return -0.03965172916650772f;
  t128_n68: return 0.0599638931453228f;
  t128_n69: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t128_n77;
  t128_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t128_n76;
  t128_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n73;
  t128_n72: return -0.01701691746711731f;
  t128_n73: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t128_n75;
  t128_n74: return 0.009878815151751041f;
  t128_n75: return -0.1311606913805008f;
  t128_n76: return 0.25732576847076416f;
  t128_n77: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t128_n79;
  t128_n78: return 0.000523740251082927f;
  t128_n79: if (E.getNumReferences() >= 3268476928 /*104.5*/) goto t128_n87;
  t128_n80: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t128_n82;
  t128_n81: return 0.11178727447986603f;
  t128_n82: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n86;
  t128_n83: if (E.getNumReferences() >= 3265855488 /*84.5*/) goto t128_n85;
  t128_n84: return 0.08278275281190872f;
  t128_n85: return 0.1363927125930786f;
  t128_n86: return -0.11328443139791489f;
  t128_n87: return 0.030663302168250084f;
  t128_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t128_n166;
  t128_n89: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t128_n109;
  t128_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n100;
  t128_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t128_n99;
  t128_n92: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t128_n96;
  t128_n93: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t128_n95;
  t128_n94: return -0.07767779380083084f;
  t128_n95: return -0.14348240196704865f;
  t128_n96: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t128_n98;
  t128_n97: return 0.07809469103813171f;
  t128_n98: return -0.10741609334945679f;
  t128_n99: return -0.09277409315109253f;
  t128_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t128_n102;
  t128_n101: return 0.0007229075417853892f;
  t128_n102: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t128_n108;
  t128_n103: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t128_n105;
  t128_n104: return 0.008119088597595692f;
  t128_n105: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t128_n107;
  t128_n106: return 0.07717171311378479f;
  t128_n107: return -0.013557438738644123f;
  t128_n108: return 0.030090264976024628f;
  t128_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n113;
  t128_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t128_n112;
  t128_n111: return -0.036952536553144455f;
  t128_n112: return 0.04638173058629036f;
  t128_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Variable))) goto t128_n147;
  t128_n114: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t128_n118;
  t128_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t128_n117;
  t128_n116: return 0.0029050633311271667f;
  t128_n117: return 0.03127264976501465f;
  t128_n118: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t128_n138;
  t128_n119: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t128_n127;
  t128_n120: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t128_n122;
  t128_n121: return -0.00025957004982046783f;
  t128_n122: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t128_n126;
  t128_n123: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n125;
  t128_n124: return -0.008304610848426819f;
  t128_n125: return 0.008293534629046917f;
  t128_n126: return -0.026825908571481705f;
  t128_n127: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t128_n131;
  t128_n128: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t128_n130;
  t128_n129: return 0.007527480833232403f;
  t128_n130: return -0.09242419898509979f;
  t128_n131: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t128_n133;
  t128_n132: return -0.017055222764611244f;
  t128_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t128_n135;
  t128_n134: return 0.012404259294271469f;
  t128_n135: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t128_n137;
  t128_n136: return 0.08540590852499008f;
  t128_n137: return 0.005511674098670483f;
  t128_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n144;
  t128_n139: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t128_n141;
  t128_n140: return 0.11527787894010544f;
  t128_n141: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t128_n143;
  t128_n142: return 0.14077213406562805f;
  t128_n143: return 0.02295619063079357f;
  t128_n144: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t128_n146;
  t128_n145: return -0.2198437750339508f;
  t128_n146: return 0.10089290142059326f;
  t128_n147: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n157;
  t128_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n156;
  t128_n149: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t128_n151;
  t128_n150: return -0.010808801278471947f;
  t128_n151: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t128_n153;
  t128_n152: return 0.027767790481448174f;
  t128_n153: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t128_n155;
  t128_n154: return 0.17735838890075684f;
  t128_n155: return 0.07565616816282272f;
  t128_n156: return 0.0067160180769860744f;
  t128_n157: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t128_n159;
  t128_n158: return 0.012111708521842957f;
  t128_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t128_n161;
  t128_n160: return 0.01810985989868641f;
  t128_n161: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t128_n165;
  t128_n162: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t128_n164;
  t128_n163: return 0.03760522976517677f;
  t128_n164: return 0.09635115414857864f;
  t128_n165: return -0.1314740628004074f;
  t128_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t128_n180;
  t128_n167: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t128_n179;
  t128_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n170;
  t128_n169: return -0.1408228576183319f;
  t128_n170: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t128_n172;
  t128_n171: return 0.011040838435292244f;
  t128_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t128_n176;
  t128_n173: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t128_n175;
  t128_n174: return 0.0014805186074227095f;
  t128_n175: return 0.024581633508205414f;
  t128_n176: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t128_n178;
  t128_n177: return 0.0943220779299736f;
  t128_n178: return 0.027825117111206055f;
  t128_n179: return -0.011474181897938251f;
  t128_n180: return 0.044125039130449295f;
  t128_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t128_n183;
  t128_n182: return -0.008336059749126434f;
  t128_n183: return 0.03878917545080185f;
  t128_n184: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t128_n254;
  t128_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t128_n223;
  t128_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n194;
  t128_n187: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t128_n191;
  t128_n188: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t128_n190;
  t128_n189: return -0.0437786728143692f;
  t128_n190: return 0.040251292288303375f;
  t128_n191: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t128_n193;
  t128_n192: return -0.039847567677497864f;
  t128_n193: return 0.037286754697561264f;
  t128_n194: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t128_n200;
  t128_n195: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t128_n197;
  t128_n196: return -0.03839225322008133f;
  t128_n197: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t128_n199;
  t128_n198: return 0.0025961408391594887f;
  t128_n199: return 0.08236151933670044f;
  t128_n200: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t128_n208;
  t128_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n207;
  t128_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t128_n206;
  t128_n203: if (E.getFractionNameInContext() >= 3164058662 /*0.018518518656492233*/) goto t128_n205;
  t128_n204: return -0.029613174498081207f;
  t128_n205: return -0.004428735934197903f;
  t128_n206: return 0.051225464791059494f;
  t128_n207: return 0.007796879857778549f;
  t128_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n214;
  t128_n209: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t128_n213;
  t128_n210: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t128_n212;
  t128_n211: return -0.0394623763859272f;
  t128_n212: return 0.0228638406842947f;
  t128_n213: return -0.0955670177936554f;
  t128_n214: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t128_n222;
  t128_n215: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t128_n221;
  t128_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t128_n220;
  t128_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n219;
  t128_n218: return -0.07719141244888306f;
  t128_n219: return 0.04547983035445213f;
  t128_n220: return 0.02511112578213215f;
  t128_n221: return 0.01162667479366064f;
  t128_n222: return -0.0040410407818853855f;
  t128_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n225;
  t128_n224: return -0.028723297640681267f;
  t128_n225: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t128_n237;
  t128_n226: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t128_n236;
  t128_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t128_n235;
  t128_n228: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t128_n232;
  t128_n229: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t128_n231;
  t128_n230: return -0.004018100444227457f;
  t128_n231: return -0.1055484265089035f;
  t128_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t128_n234;
  t128_n233: return 0.020131388679146767f;
  t128_n234: return -0.04078677296638489f;
  t128_n235: return 0.004991411231458187f;
  t128_n236: return 0.0961185172200203f;
  t128_n237: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t128_n247;
  t128_n238: if (E.getFractionNameInContext() >= 3211998003 /*0.949999988079071*/) goto t128_n240;
  t128_n239: return 0.014973265118896961f;
  t128_n240: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t128_n246;
  t128_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t128_n243;
  t128_n242: return 0.0510077141225338f;
  t128_n243: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t128_n245;
  t128_n244: return -0.19874314963817596f;
  t128_n245: return 0.09807255119085312f;
  t128_n246: return 0.1258709728717804f;
  t128_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t128_n249;
  t128_n248: return 0.010740390047430992f;
  t128_n249: if (E.getFractionNameInContext() >= 3176353077 /*0.0515873022377491*/) goto t128_n253;
  t128_n250: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t128_n252;
  t128_n251: return 0.030299127101898193f;
  t128_n252: return 0.025301268324255943f;
  t128_n253: return 0.012234548106789589f;
  t128_n254: return -0.01303115114569664f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree129(const Example& E) {
  t129_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t129_n44;
  t129_n1: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t129_n3;
  t129_n2: return -6.209079583641142e-05f;
  t129_n3: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n43;
  t129_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t129_n8;
  t129_n5: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t129_n7;
  t129_n6: return -0.0021601542830467224f;
  t129_n7: return 0.026769356802105904f;
  t129_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t129_n28;
  t129_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n11;
  t129_n10: return -0.0226267259567976f;
  t129_n11: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n15;
  t129_n12: if (E.getNumReferences() >= 3338441472 /*32329.5*/) goto t129_n14;
  t129_n13: return 0.013392185792326927f;
  t129_n14: return 0.04427262768149376f;
  t129_n15: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t129_n27;
  t129_n16: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t129_n24;
  t129_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t129_n19;
  t129_n18: return 0.06587015092372894f;
  t129_n19: if (E.getFractionNameInContext() >= 3190545458 /*0.1678321659564972*/) goto t129_n21;
  t129_n20: return -0.1428874284029007f;
  t129_n21: if (E.getFractionNameInContext() >= 3193736649 /*0.2153846174478531*/) goto t129_n23;
  t129_n22: return 0.3073946237564087f;
  t129_n23: return -0.0036813935730606318f;
  t129_n24: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t129_n26;
  t129_n25: return 0.10349048674106598f;
  t129_n26: return 0.01346964854747057f;
  t129_n27: return 0.024136018007993698f;
  t129_n28: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t129_n30;
  t129_n29: return -0.008118954487144947f;
  t129_n30: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n32;
  t129_n31: return 0.029000993818044662f;
  t129_n32: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t129_n38;
  t129_n33: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t129_n37;
  t129_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t129_n36;
  t129_n35: return 0.11275172233581543f;
  t129_n36: return 0.021481746807694435f;
  t129_n37: return 0.09182284027338028f;
  t129_n38: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t129_n42;
  t129_n39: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t129_n41;
  t129_n40: return 0.025629190728068352f;
  t129_n41: return 0.08012394607067108f;
  t129_n42: return -0.6265851259231567f;
  t129_n43: return -0.02349306456744671f;
  t129_n44: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t129_n182;
  t129_n45: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t129_n165;
  t129_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n164;
  t129_n47: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t129_n153;
  t129_n48: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t129_n142;
  t129_n49: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t129_n141;
  t129_n50: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t129_n108;
  t129_n51: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t129_n103;
  t129_n52: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t129_n98;
  t129_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t129_n83;
  t129_n54: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t129_n74;
  t129_n55: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t129_n65;
  t129_n56: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t129_n58;
  t129_n57: return -0.025387190282344818f;
  t129_n58: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t129_n62;
  t129_n59: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t129_n61;
  t129_n60: return 0.11665213853120804f;
  t129_n61: return -0.151570662856102f;
  t129_n62: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t129_n64;
  t129_n63: return -0.014544609934091568f;
  t129_n64: return 0.1155441403388977f;
  t129_n65: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t129_n71;
  t129_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t129_n68;
  t129_n67: return -0.043849095702171326f;
  t129_n68: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t129_n70;
  t129_n69: return -0.10818687826395035f;
  t129_n70: return 0.003725592512637377f;
  t129_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t129_n73;
  t129_n72: return -0.9258720278739929f;
  t129_n73: return -0.09534350037574768f;
  t129_n74: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t129_n80;
  t129_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n77;
  t129_n76: return 0.023834899067878723f;
  t129_n77: if (E.getFractionNameInContext() >= 3180013008 /*0.06794321537017822*/) goto t129_n79;
  t129_n78: return -0.2608596682548523f;
  t129_n79: return 0.21923542022705078f;
  t129_n80: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t129_n82;
  t129_n81: return -0.06712199002504349f;
  t129_n82: return -0.007628526538610458f;
  t129_n83: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t129_n97;
  t129_n84: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t129_n88;
  t129_n85: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t129_n87;
  t129_n86: return 0.021612467244267464f;
  t129_n87: return -0.02743743360042572f;
  t129_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t129_n90;
  t129_n89: return 0.018574628978967667f;
  t129_n90: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n92;
  t129_n91: return 0.06777944415807724f;
  t129_n92: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t129_n94;
  t129_n93: return 0.02310495637357235f;
  t129_n94: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t129_n96;
  t129_n95: return 0.15349778532981873f;
  t129_n96: return 0.0457572266459465f;
  t129_n97: return -0.01640508510172367f;
  t129_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t129_n102;
  t129_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t129_n101;
  t129_n100: return -0.12362933158874512f;
  t129_n101: return 0.044001512229442596f;
  t129_n102: return 0.12227088958024979f;
  t129_n103: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t129_n107;
  t129_n104: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t129_n106;
  t129_n105: return -0.07016465812921524f;
  t129_n106: return 0.07578161358833313f;
  t129_n107: return -0.9256091117858887f;
  t129_n108: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t129_n136;
  t129_n109: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t129_n117;
  t129_n110: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n112;
  t129_n111: return 0.03559790179133415f;
  t129_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n114;
  t129_n113: return -0.042472995817661285f;
  t129_n114: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t129_n116;
  t129_n115: return 0.0021411709021776915f;
  t129_n116: return -0.3106423616409302f;
  t129_n117: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t129_n133;
  t129_n118: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t129_n132;
  t129_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t129_n131;
  t129_n120: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t129_n126;
  t129_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t129_n123;
  t129_n122: return -0.5548510551452637f;
  t129_n123: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t129_n125;
  t129_n124: return -0.04872964322566986f;
  t129_n125: return 0.09038715809583664f;
  t129_n126: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n130;
  t129_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t129_n129;
  t129_n128: return -2.5616343021392822f;
  t129_n129: return -0.344493567943573f;
  t129_n130: return -0.06699799746274948f;
  t129_n131: return -0.025708360597491264f;
  t129_n132: return -0.014857595786452293f;
  t129_n133: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t129_n135;
  t129_n134: return -0.016013454645872116f;
  t129_n135: return 0.06531166285276413f;
  t129_n136: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t129_n138;
  t129_n137: return -0.13625440001487732f;
  t129_n138: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n140;
  t129_n139: return -0.05662216618657112f;
  t129_n140: return 0.026021623983979225f;
  t129_n141: return 0.001461015548557043f;
  t129_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t129_n148;
  t129_n143: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t129_n145;
  t129_n144: return 0.03891408443450928f;
  t129_n145: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t129_n147;
  t129_n146: return -0.06383675336837769f;
  t129_n147: return 0.01483512856066227f;
  t129_n148: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t129_n152;
  t129_n149: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t129_n151;
  t129_n150: return 0.10309440642595291f;
  t129_n151: return 0.01984906941652298f;
  t129_n152: return 0.1318715512752533f;
  t129_n153: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t129_n155;
  t129_n154: return 0.029071133583784103f;
  t129_n155: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t129_n157;
  t129_n156: return -0.036922987550497055f;
  t129_n157: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t129_n161;
  t129_n158: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t129_n160;
  t129_n159: return 0.010950912721455097f;
  t129_n160: return 0.0409083291888237f;
  t129_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t129_n163;
  t129_n162: return -0.09426949918270111f;
  t129_n163: return -0.014481354504823685f;
  t129_n164: return 0.005824715830385685f;
  t129_n165: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t129_n169;
  t129_n166: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t129_n168;
  t129_n167: return 0.0013326467014849186f;
  t129_n168: return 0.03389984369277954f;
  t129_n169: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t129_n181;
  t129_n170: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t129_n178;
  t129_n171: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t129_n173;
  t129_n172: return 0.025957927107810974f;
  t129_n173: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t129_n175;
  t129_n174: return 0.07346714287996292f;
  t129_n175: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t129_n177;
  t129_n176: return 0.08977698534727097f;
  t129_n177: return 0.015384029597043991f;
  t129_n178: if (E.getNumReferences() >= 3315734528 /*5187.0*/) goto t129_n180;
  t129_n179: return -0.028424011543393135f;
  t129_n180: return 0.07261147350072861f;
  t129_n181: return 0.006935046054422855f;
  t129_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t129_n190;
  t129_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n189;
  t129_n184: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t129_n188;
  t129_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t129_n187;
  t129_n186: return -0.20701996982097626f;
  t129_n187: return -0.03794875368475914f;
  t129_n188: return -2.612753087305464e-05f;
  t129_n189: return 0.02277120016515255f;
  t129_n190: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t129_n240;
  t129_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n193;
  t129_n192: return -0.006533915642648935f;
  t129_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t129_n239;
  t129_n194: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t129_n232;
  t129_n195: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t129_n219;
  t129_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n204;
  t129_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t129_n201;
  t129_n198: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t129_n200;
  t129_n199: return -0.029132375493645668f;
  t129_n200: return 0.04178120940923691f;
  t129_n201: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t129_n203;
  t129_n202: return -0.0015550142852589488f;
  t129_n203: return -0.07206682115793228f;
  t129_n204: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t129_n214;
  t129_n205: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t129_n209;
  t129_n206: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n208;
  t129_n207: return 0.007869773544371128f;
  t129_n208: return 0.11233007907867432f;
  t129_n209: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t129_n211;
  t129_n210: return -0.004675008822232485f;
  t129_n211: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t129_n213;
  t129_n212: return 0.008731959387660027f;
  t129_n213: return -0.09483659267425537f;
  t129_n214: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t129_n218;
  t129_n215: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t129_n217;
  t129_n216: return -0.09970366209745407f;
  t129_n217: return 0.027194920927286148f;
  t129_n218: return 0.15256302058696747f;
  t129_n219: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t129_n227;
  t129_n220: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n224;
  t129_n221: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t129_n223;
  t129_n222: return -0.1392795890569687f;
  t129_n223: return 0.08070877939462662f;
  t129_n224: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t129_n226;
  t129_n225: return -0.1001325249671936f;
  t129_n226: return 0.15622597932815552f;
  t129_n227: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t129_n229;
  t129_n228: return 0.011515296064317226f;
  t129_n229: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n231;
  t129_n230: return 0.05128052458167076f;
  t129_n231: return -0.026077790185809135f;
  t129_n232: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t129_n238;
  t129_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n237;
  t129_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t129_n236;
  t129_n235: return -0.21593999862670898f;
  t129_n236: return -0.02891029603779316f;
  t129_n237: return 0.007795094978064299f;
  t129_n238: return -0.037362027913331985f;
  t129_n239: return 0.012310375459492207f;
  t129_n240: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t129_n252;
  t129_n241: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t129_n243;
  t129_n242: return 0.006319113075733185f;
  t129_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n245;
  t129_n244: return -0.624359667301178f;
  t129_n245: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t129_n247;
  t129_n246: return -0.10871422290802002f;
  t129_n247: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t129_n251;
  t129_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t129_n250;
  t129_n249: return -0.061002109199762344f;
  t129_n250: return 0.17846496403217316f;
  t129_n251: return 0.1974688172340393f;
  t129_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t129_n254;
  t129_n253: return 0.013352826237678528f;
  t129_n254: return 0.030647119507193565f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree130(const Example& E) {
  t130_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t130_n246;
  t130_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t130_n201;
  t130_n2: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t130_n192;
  t130_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t130_n69;
  t130_n4: if (E.getNumReferences() >= 3288801280 /*540.5*/) goto t130_n68;
  t130_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t130_n51;
  t130_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n10;
  t130_n7: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t130_n9;
  t130_n8: return -0.1911434382200241f;
  t130_n9: return -0.06753343343734741f;
  t130_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t130_n22;
  t130_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n19;
  t130_n12: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t130_n16;
  t130_n13: if (E.getNumReferences() >= 3288719360 /*535.5*/) goto t130_n15;
  t130_n14: return -0.015611420385539532f;
  t130_n15: return -0.16668598353862762f;
  t130_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n18;
  t130_n17: return -0.10452818870544434f;
  t130_n18: return 0.09636766463518143f;
  t130_n19: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t130_n21;
  t130_n20: return -0.2540135085582733f;
  t130_n21: return -0.0004944982356391847f;
  t130_n22: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t130_n48;
  t130_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n29;
  t130_n24: if (E.getNumReferences() >= 3281354752 /*299.0*/) goto t130_n26;
  t130_n25: return -0.0019852265249937773f;
  t130_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t130_n28;
  t130_n27: return 0.241974875330925f;
  t130_n28: return 0.08027859777212143f;
  t130_n29: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t130_n47;
  t130_n30: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t130_n38;
  t130_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t130_n33;
  t130_n32: return -0.10624470561742783f;
  t130_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t130_n35;
  t130_n34: return 0.016765983775258064f;
  t130_n35: if (E.getNumReferences() >= 3257663488 /*43.0*/) goto t130_n37;
  t130_n36: return 0.19031000137329102f;
  t130_n37: return 0.21544446051120758f;
  t130_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t130_n40;
  t130_n39: return 0.01621556282043457f;
  t130_n40: if (E.getFileProximityDistanceCost() >= 3251634176 /*26.0*/) goto t130_n46;
  t130_n41: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t130_n45;
  t130_n42: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t130_n44;
  t130_n43: return -0.4906547963619232f;
  t130_n44: return 0.044552553445100784f;
  t130_n45: return -0.02641955576837063f;
  t130_n46: return 0.05887341871857643f;
  t130_n47: return 0.004840652924031019f;
  t130_n48: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t130_n50;
  t130_n49: return 0.0759374275803566f;
  t130_n50: return -0.02646273747086525f;
  t130_n51: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t130_n61;
  t130_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t130_n54;
  t130_n53: return 0.0022785011678934097f;
  t130_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n58;
  t130_n55: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t130_n57;
  t130_n56: return -0.0361865796148777f;
  t130_n57: return -0.03931949660181999f;
  t130_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t130_n60;
  t130_n59: return -0.0017848012503236532f;
  t130_n60: return 0.021740401163697243f;
  t130_n61: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t130_n63;
  t130_n62: return 0.008818382397294044f;
  t130_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n65;
  t130_n64: return -0.03285114839673042f;
  t130_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t130_n67;
  t130_n66: return 0.002137869829311967f;
  t130_n67: return 0.1076432541012764f;
  t130_n68: return 0.010642889887094498f;
  t130_n69: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t130_n189;
  t130_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t130_n72;
  t130_n71: return -0.0036463786382228136f;
  t130_n72: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t130_n186;
  t130_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t130_n87;
  t130_n74: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t130_n86;
  t130_n75: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t130_n77;
  t130_n76: return -0.006007716525346041f;
  t130_n77: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t130_n85;
  t130_n78: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t130_n84;
  t130_n79: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t130_n81;
  t130_n80: return 0.002722848439589143f;
  t130_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n83;
  t130_n82: return -0.0176907517015934f;
  t130_n83: return 0.04321363568305969f;
  t130_n84: return -0.04112030565738678f;
  t130_n85: return 0.04048401862382889f;
  t130_n86: return -0.06399650126695633f;
  t130_n87: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t130_n115;
  t130_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n92;
  t130_n89: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t130_n91;
  t130_n90: return 0.016271665692329407f;
  t130_n91: return -0.015913430601358414f;
  t130_n92: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t130_n114;
  t130_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n105;
  t130_n94: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t130_n104;
  t130_n95: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t130_n97;
  t130_n96: return -0.010689353570342064f;
  t130_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t130_n99;
  t130_n98: return -0.06747405230998993f;
  t130_n99: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t130_n101;
  t130_n100: return 0.03048279881477356f;
  t130_n101: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t130_n103;
  t130_n102: return -0.04309597983956337f;
  t130_n103: return 0.024350283667445183f;
  t130_n104: return -0.013562212698161602f;
  t130_n105: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t130_n111;
  t130_n106: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t130_n108;
  t130_n107: return -0.142654687166214f;
  t130_n108: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t130_n110;
  t130_n109: return 0.04356860741972923f;
  t130_n110: return -0.053595222532749176f;
  t130_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n113;
  t130_n112: return -0.05791706591844559f;
  t130_n113: return 0.025796253234148026f;
  t130_n114: return -0.05623910576105118f;
  t130_n115: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t130_n117;
  t130_n116: return 0.09927359968423843f;
  t130_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n137;
  t130_n118: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t130_n126;
  t130_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n121;
  t130_n120: return -0.006392304319888353f;
  t130_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t130_n123;
  t130_n122: return 0.015804274007678032f;
  t130_n123: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t130_n125;
  t130_n124: return 0.053770359605550766f;
  t130_n125: return 0.1310732513666153f;
  t130_n126: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t130_n128;
  t130_n127: return -0.09671631455421448f;
  t130_n128: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t130_n130;
  t130_n129: return -0.0341719314455986f;
  t130_n130: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t130_n132;
  t130_n131: return -0.11580859869718552f;
  t130_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t130_n134;
  t130_n133: return -0.0007187004666775465f;
  t130_n134: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t130_n136;
  t130_n135: return 0.0945146381855011f;
  t130_n136: return 0.057025786489248276f;
  t130_n137: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t130_n139;
  t130_n138: return 0.015931516885757446f;
  t130_n139: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t130_n155;
  t130_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n148;
  t130_n141: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t130_n143;
  t130_n142: return 0.05950302630662918f;
  t130_n143: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t130_n145;
  t130_n144: return -0.07905872911214828f;
  t130_n145: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t130_n147;
  t130_n146: return -0.040269531309604645f;
  t130_n147: return 0.027277609333395958f;
  t130_n148: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t130_n152;
  t130_n149: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t130_n151;
  t130_n150: return -0.04056936502456665f;
  t130_n151: return 0.01890442706644535f;
  t130_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n154;
  t130_n153: return -0.009805711917579174f;
  t130_n154: return 0.0718260109424591f;
  t130_n155: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t130_n161;
  t130_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n158;
  t130_n157: return -0.07729021459817886f;
  t130_n158: if (E.getNumReferences() >= 3316233216 /*5430.5*/) goto t130_n160;
  t130_n159: return 0.044926755130290985f;
  t130_n160: return -0.002282290253788233f;
  t130_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t130_n173;
  t130_n162: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t130_n166;
  t130_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n165;
  t130_n164: return -0.01022137701511383f;
  t130_n165: return 0.05899861454963684f;
  t130_n166: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t130_n170;
  t130_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n169;
  t130_n168: return -0.11091025918722153f;
  t130_n169: return 0.018412195146083832f;
  t130_n170: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t130_n172;
  t130_n171: return 0.003613121574744582f;
  t130_n172: return 0.05842473730444908f;
  t130_n173: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t130_n183;
  t130_n174: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t130_n182;
  t130_n175: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t130_n181;
  t130_n176: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t130_n180;
  t130_n177: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t130_n179;
  t130_n178: return 0.01963561214506626f;
  t130_n179: return -0.037497058510780334f;
  t130_n180: return 0.027636181563138962f;
  t130_n181: return -0.08760417997837067f;
  t130_n182: return 0.04151955246925354f;
  t130_n183: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t130_n185;
  t130_n184: return -0.09890418499708176f;
  t130_n185: return 0.023519551381468773f;
  t130_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t130_n188;
  t130_n187: return -0.03348756581544876f;
  t130_n188: return 0.03386535868048668f;
  t130_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n191;
  t130_n190: return -0.02628377266228199f;
  t130_n191: return -0.005775230936706066f;
  t130_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t130_n196;
  t130_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n195;
  t130_n194: return -0.165236234664917f;
  t130_n195: return -0.030181847512722015f;
  t130_n196: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t130_n200;
  t130_n197: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t130_n199;
  t130_n198: return 0.04702974855899811f;
  t130_n199: return -0.09601641446352005f;
  t130_n200: return 0.21337024867534637f;
  t130_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t130_n245;
  t130_n202: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t130_n204;
  t130_n203: return -0.0022881212644279003f;
  t130_n204: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t130_n244;
  t130_n205: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t130_n243;
  t130_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t130_n242;
  t130_n207: if (E.getNumReferences() >= 3314693120 /*4678.5*/) goto t130_n237;
  t130_n208: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t130_n212;
  t130_n209: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t130_n211;
  t130_n210: return 0.011613945476710796f;
  t130_n211: return -0.01897215098142624f;
  t130_n212: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t130_n232;
  t130_n213: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t130_n231;
  t130_n214: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t130_n230;
  t130_n215: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t130_n225;
  t130_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t130_n222;
  t130_n217: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t130_n221;
  t130_n218: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t130_n220;
  t130_n219: return -0.010839094407856464f;
  t130_n220: return 0.1061517521739006f;
  t130_n221: return -0.1409737467765808f;
  t130_n222: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t130_n224;
  t130_n223: return 0.04328459873795509f;
  t130_n224: return 0.10324717313051224f;
  t130_n225: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t130_n229;
  t130_n226: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t130_n228;
  t130_n227: return -0.2845565378665924f;
  t130_n228: return -1.016629934310913f;
  t130_n229: return -0.06605380773544312f;
  t130_n230: return 0.07535487413406372f;
  t130_n231: return 0.019846899434924126f;
  t130_n232: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t130_n234;
  t130_n233: return -0.02984795533120632f;
  t130_n234: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t130_n236;
  t130_n235: return -0.434749037027359f;
  t130_n236: return -0.017109179869294167f;
  t130_n237: if (E.getNumReferences() >= 3315159040 /*4906.0*/) goto t130_n241;
  t130_n238: if (E.getNumReferences() >= 3315124224 /*4889.0*/) goto t130_n240;
  t130_n239: return 0.07278066873550415f;
  t130_n240: return 0.14455154538154602f;
  t130_n241: return 0.02139548398554325f;
  t130_n242: return 0.028103098273277283f;
  t130_n243: return -0.004715561866760254f;
  t130_n244: return 0.12126026302576065f;
  t130_n245: return 0.00226655718870461f;
  t130_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t130_n254;
  t130_n247: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t130_n249;
  t130_n248: return 0.0014412200544029474f;
  t130_n249: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t130_n253;
  t130_n250: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t130_n252;
  t130_n251: return 0.016425030305981636f;
  t130_n252: return -0.009049694053828716f;
  t130_n253: return 0.11225781589746475f;
  t130_n254: return 0.007434238214045763f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree131(const Example& E) {
  t131_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t131_n252;
  t131_n1: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t131_n3;
  t131_n2: return -0.001284936210140586f;
  t131_n3: if (E.getNumReferences() >= 3335283968 /*26162.5*/) goto t131_n251;
  t131_n4: if (E.getNumReferences() >= 3334364160 /*24366.0*/) goto t131_n248;
  t131_n5: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t131_n245;
  t131_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t131_n238;
  t131_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t131_n101;
  t131_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t131_n100;
  t131_n9: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t131_n55;
  t131_n10: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t131_n54;
  t131_n11: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t131_n53;
  t131_n12: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t131_n50;
  t131_n13: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t131_n31;
  t131_n14: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t131_n24;
  t131_n15: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t131_n19;
  t131_n16: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t131_n18;
  t131_n17: return -0.023187613114714622f;
  t131_n18: return -0.004143225494772196f;
  t131_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t131_n23;
  t131_n20: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t131_n22;
  t131_n21: return -0.02226981893181801f;
  t131_n22: return 0.05552222579717636f;
  t131_n23: return 0.04470019415020943f;
  t131_n24: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t131_n26;
  t131_n25: return 0.03189944103360176f;
  t131_n26: if (E.getNumReferences() >= 3276701696 /*206.5*/) goto t131_n30;
  t131_n27: if (E.getNumReferences() >= 3276636160 /*205.5*/) goto t131_n29;
  t131_n28: return -0.019677285104990005f;
  t131_n29: return -1.2313841581344604f;
  t131_n30: return 0.020996129140257835f;
  t131_n31: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t131_n39;
  t131_n32: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t131_n36;
  t131_n33: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t131_n35;
  t131_n34: return -0.025515953078866005f;
  t131_n35: return 0.08280010521411896f;
  t131_n36: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t131_n38;
  t131_n37: return 0.026298368349671364f;
  t131_n38: return -0.09925834089517593f;
  t131_n39: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t131_n49;
  t131_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t131_n44;
  t131_n41: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t131_n43;
  t131_n42: return -0.07791505008935928f;
  t131_n43: return -0.004668282810598612f;
  t131_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t131_n48;
  t131_n45: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t131_n47;
  t131_n46: return 0.046103496104478836f;
  t131_n47: return -0.023266151547431946f;
  t131_n48: return -0.02558889426290989f;
  t131_n49: return 0.07539957761764526f;
  t131_n50: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t131_n52;
  t131_n51: return 0.03125736489892006f;
  t131_n52: return -0.02806226909160614f;
  t131_n53: return -0.044731445610523224f;
  t131_n54: return 0.030098503455519676f;
  t131_n55: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t131_n99;
  t131_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t131_n62;
  t131_n57: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t131_n61;
  t131_n58: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n60;
  t131_n59: return -0.008206627331674099f;
  t131_n60: return -0.1287834495306015f;
  t131_n61: return 0.012411405332386494f;
  t131_n62: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t131_n78;
  t131_n63: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t131_n75;
  t131_n64: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t131_n74;
  t131_n65: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t131_n69;
  t131_n66: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t131_n68;
  t131_n67: return 0.00017668871441856027f;
  t131_n68: return 0.09674631059169769f;
  t131_n69: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t131_n71;
  t131_n70: return 0.14274831116199493f;
  t131_n71: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t131_n73;
  t131_n72: return 0.014915743842720985f;
  t131_n73: return -0.1452728807926178f;
  t131_n74: return -0.10572179406881332f;
  t131_n75: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t131_n77;
  t131_n76: return 0.055377766489982605f;
  t131_n77: return 0.16854271292686462f;
  t131_n78: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t131_n92;
  t131_n79: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n87;
  t131_n80: if (E.getNumReferences() >= 3308029952 /*2760.5*/) goto t131_n82;
  t131_n81: return -0.0120706120505929f;
  t131_n82: if (E.getNumReferences() >= 3310786560 /*3433.5*/) goto t131_n84;
  t131_n83: return 0.1273631900548935f;
  t131_n84: if (E.getNumReferences() >= 3332669952 /*21057.0*/) goto t131_n86;
  t131_n85: return -0.23650947213172913f;
  t131_n86: return 0.1132504940032959f;
  t131_n87: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n91;
  t131_n88: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t131_n90;
  t131_n89: return 0.05451355502009392f;
  t131_n90: return 0.081944040954113f;
  t131_n91: return 0.044836048036813736f;
  t131_n92: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t131_n94;
  t131_n93: return 0.033974118530750275f;
  t131_n94: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t131_n96;
  t131_n95: return -0.02675449103116989f;
  t131_n96: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n98;
  t131_n97: return 0.12212302535772324f;
  t131_n98: return 0.035423826426267624f;
  t131_n99: return -0.048054322600364685f;
  t131_n100: return 0.0006347025046125054f;
  t131_n101: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t131_n213;
  t131_n102: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t131_n104;
  t131_n103: return -0.012961633503437042f;
  t131_n104: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t131_n114;
  t131_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t131_n107;
  t131_n106: return 0.0019549650605767965f;
  t131_n107: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t131_n109;
  t131_n108: return -0.050648804754018784f;
  t131_n109: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t131_n111;
  t131_n110: return 0.014894271269440651f;
  t131_n111: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t131_n113;
  t131_n112: return -0.6771196722984314f;
  t131_n113: return -0.14055949449539185f;
  t131_n114: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t131_n188;
  t131_n115: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t131_n183;
  t131_n116: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t131_n164;
  t131_n117: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t131_n155;
  t131_n118: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t131_n154;
  t131_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t131_n153;
  t131_n120: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t131_n122;
  t131_n121: return -0.01181051880121231f;
  t131_n122: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t131_n142;
  t131_n123: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t131_n129;
  t131_n124: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t131_n126;
  t131_n125: return -0.033600013703107834f;
  t131_n126: if (E.getNumReferences() >= 3264610304 /*75.0*/) goto t131_n128;
  t131_n127: return 0.08262555301189423f;
  t131_n128: return 0.4880698025226593f;
  t131_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t131_n141;
  t131_n130: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t131_n136;
  t131_n131: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t131_n133;
  t131_n132: return 0.032075319439172745f;
  t131_n133: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t131_n135;
  t131_n134: return 0.10920540988445282f;
  t131_n135: return 0.051811687648296356f;
  t131_n136: if (E.getNumReferences() >= 3262119936 /*60.0*/) goto t131_n138;
  t131_n137: return 0.014239626936614513f;
  t131_n138: if (E.getNumReferences() >= 3268149248 /*102.0*/) goto t131_n140;
  t131_n139: return 0.14308907091617584f;
  t131_n140: return 0.09266984462738037f;
  t131_n141: return -0.02378462813794613f;
  t131_n142: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t131_n152;
  t131_n143: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t131_n149;
  t131_n144: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t131_n148;
  t131_n145: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t131_n147;
  t131_n146: return -0.4940255284309387f;
  t131_n147: return -0.04772954806685448f;
  t131_n148: return 0.0736498311161995f;
  t131_n149: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t131_n151;
  t131_n150: return 0.014685514383018017f;
  t131_n151: return -0.05048086866736412f;
  t131_n152: return 0.06439736485481262f;
  t131_n153: return 0.012981066480278969f;
  t131_n154: return -0.039081644266843796f;
  t131_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t131_n163;
  t131_n156: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t131_n160;
  t131_n157: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t131_n159;
  t131_n158: return 0.11623505502939224f;
  t131_n159: return -0.10178817808628082f;
  t131_n160: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t131_n162;
  t131_n161: return -0.1187591552734375f;
  t131_n162: return -0.9287672638893127f;
  t131_n163: return 0.04883275181055069f;
  t131_n164: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t131_n180;
  t131_n165: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t131_n179;
  t131_n166: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t131_n170;
  t131_n167: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t131_n169;
  t131_n168: return -0.2216339111328125f;
  t131_n169: return -0.030948176980018616f;
  t131_n170: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t131_n178;
  t131_n171: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t131_n175;
  t131_n172: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t131_n174;
  t131_n173: return 0.09728285670280457f;
  t131_n174: return -1.1598280668258667f;
  t131_n175: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t131_n177;
  t131_n176: return -4.047545433044434f;
  t131_n177: return -0.18940284848213196f;
  t131_n178: return 0.006964227184653282f;
  t131_n179: return 0.12320045381784439f;
  t131_n180: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n182;
  t131_n181: return 0.12000033259391785f;
  t131_n182: return 0.12022153288125992f;
  t131_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t131_n185;
  t131_n184: return 0.0030314817558974028f;
  t131_n185: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t131_n187;
  t131_n186: return 0.06227882578969002f;
  t131_n187: return -0.02601204439997673f;
  t131_n188: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t131_n194;
  t131_n189: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t131_n191;
  t131_n190: return -0.016720078885555267f;
  t131_n191: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t131_n193;
  t131_n192: return 0.02247944474220276f;
  t131_n193: return 0.056365203112363815f;
  t131_n194: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t131_n210;
  t131_n195: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t131_n205;
  t131_n196: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t131_n198;
  t131_n197: return 0.06111820787191391f;
  t131_n198: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t131_n204;
  t131_n199: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t131_n201;
  t131_n200: return 0.05729968473315239f;
  t131_n201: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t131_n203;
  t131_n202: return 0.13016176223754883f;
  t131_n203: return 0.1486217975616455f;
  t131_n204: return -0.06217343360185623f;
  t131_n205: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t131_n209;
  t131_n206: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t131_n208;
  t131_n207: return 0.03895705193281174f;
  t131_n208: return 0.09113797545433044f;
  t131_n209: return -0.0013454812578856945f;
  t131_n210: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t131_n212;
  t131_n211: return 0.14786049723625183f;
  t131_n212: return -0.03564148768782616f;
  t131_n213: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t131_n231;
  t131_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t131_n222;
  t131_n215: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t131_n219;
  t131_n216: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t131_n218;
  t131_n217: return 0.006586747709661722f;
  t131_n218: return 0.1038617193698883f;
  t131_n219: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t131_n221;
  t131_n220: return -0.022827213630080223f;
  t131_n221: return 0.01278875395655632f;
  t131_n222: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t131_n228;
  t131_n223: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t131_n225;
  t131_n224: return 0.01948917657136917f;
  t131_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t131_n227;
  t131_n226: return -0.0208810493350029f;
  t131_n227: return -0.32818618416786194f;
  t131_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t131_n230;
  t131_n229: return 0.024211199954152107f;
  t131_n230: return 0.13232548534870148f;
  t131_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t131_n233;
  t131_n232: return -0.024645131081342697f;
  t131_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t131_n235;
  t131_n234: return -0.00019360630540177226f;
  t131_n235: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t131_n237;
  t131_n236: return 0.047167375683784485f;
  t131_n237: return -0.2072698473930359f;
  t131_n238: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t131_n244;
  t131_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t131_n243;
  t131_n240: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t131_n242;
  t131_n241: return -0.01743943616747856f;
  t131_n242: return 0.02581457793712616f;
  t131_n243: return 0.023707790300250053f;
  t131_n244: return -0.05133046954870224f;
  t131_n245: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t131_n247;
  t131_n246: return 0.05682500824332237f;
  t131_n247: return 0.13187740743160248f;
  t131_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t131_n250;
  t131_n249: return -0.3154350519180298f;
  t131_n250: return -0.07398166507482529f;
  t131_n251: return 0.01612645946443081f;
  t131_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t131_n254;
  t131_n253: return 0.00288125011138618f;
  t131_n254: return 0.007081192918121815f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree132(const Example& E) {
  t132_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t132_n182;
  t132_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t132_n71;
  t132_n2: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t132_n60;
  t132_n3: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t132_n5;
  t132_n4: return -0.0008076915983110666f;
  t132_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t132_n41;
  t132_n6: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t132_n16;
  t132_n7: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t132_n9;
  t132_n8: return 0.05425123870372772f;
  t132_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t132_n15;
  t132_n10: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t132_n14;
  t132_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t132_n13;
  t132_n12: return -0.18989971280097961f;
  t132_n13: return 0.18649962544441223f;
  t132_n14: return -0.21408015489578247f;
  t132_n15: return 0.10494678467512131f;
  t132_n16: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t132_n36;
  t132_n17: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t132_n19;
  t132_n18: return 0.004221324808895588f;
  t132_n19: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t132_n33;
  t132_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t132_n24;
  t132_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t132_n23;
  t132_n22: return -0.036594126373529434f;
  t132_n23: return -0.053657375276088715f;
  t132_n24: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t132_n26;
  t132_n25: return -0.0445370152592659f;
  t132_n26: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t132_n32;
  t132_n27: if (E.getSymbolScopeDistanceCost() >= 3242196992 /*12.0*/) goto t132_n31;
  t132_n28: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t132_n30;
  t132_n29: return 0.016821181401610374f;
  t132_n30: return -0.11920542269945145f;
  t132_n31: return 0.08110211044549942f;
  t132_n32: return 0.16998125612735748f;
  t132_n33: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t132_n35;
  t132_n34: return 0.003230134956538677f;
  t132_n35: return -0.03296702727675438f;
  t132_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t132_n40;
  t132_n37: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t132_n39;
  t132_n38: return -0.043688274919986725f;
  t132_n39: return 0.15048490464687347f;
  t132_n40: return 0.013998083770275116f;
  t132_n41: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t132_n49;
  t132_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t132_n48;
  t132_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t132_n45;
  t132_n44: return -0.05059971287846565f;
  t132_n45: if (E.getNumReferences() >= 3282550784 /*335.5*/) goto t132_n47;
  t132_n46: return 0.01767268404364586f;
  t132_n47: return 0.1650572270154953f;
  t132_n48: return 0.030816150829195976f;
  t132_n49: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t132_n55;
  t132_n50: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t132_n54;
  t132_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t132_n53;
  t132_n52: return -0.2807748019695282f;
  t132_n53: return 0.4166176915168762f;
  t132_n54: return -0.04681948944926262f;
  t132_n55: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t132_n59;
  t132_n56: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t132_n58;
  t132_n57: return 0.006700173020362854f;
  t132_n58: return 0.035302940756082535f;
  t132_n59: return 0.003241482423618436f;
  t132_n60: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t132_n68;
  t132_n61: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t132_n67;
  t132_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t132_n64;
  t132_n63: return -0.0266535934060812f;
  t132_n64: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t132_n66;
  t132_n65: return 0.0510273277759552f;
  t132_n66: return 0.014401315711438656f;
  t132_n67: return -0.0020736465230584145f;
  t132_n68: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t132_n70;
  t132_n69: return 0.00011914357310160995f;
  t132_n70: return 0.15307879447937012f;
  t132_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t132_n159;
  t132_n72: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t132_n104;
  t132_n73: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t132_n103;
  t132_n74: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t132_n100;
  t132_n75: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t132_n77;
  t132_n76: return -0.00025061643100343645f;
  t132_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t132_n93;
  t132_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t132_n92;
  t132_n79: if (E.getFractionNameInContext() >= 3211904796 /*0.9444444179534912*/) goto t132_n89;
  t132_n80: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t132_n82;
  t132_n81: return 0.2877407968044281f;
  t132_n82: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t132_n84;
  t132_n83: return -0.009613612666726112f;
  t132_n84: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t132_n86;
  t132_n85: return -0.004057413898408413f;
  t132_n86: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t132_n88;
  t132_n87: return -0.036749038845300674f;
  t132_n88: return 0.1303565949201584f;
  t132_n89: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t132_n91;
  t132_n90: return 0.021108919754624367f;
  t132_n91: return 0.14474473893642426f;
  t132_n92: return -0.0035514100454747677f;
  t132_n93: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t132_n95;
  t132_n94: return 0.007603168953210115f;
  t132_n95: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t132_n99;
  t132_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t132_n98;
  t132_n97: return -0.1193254292011261f;
  t132_n98: return 0.1417694389820099f;
  t132_n99: return -0.023802975192666054f;
  t132_n100: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t132_n102;
  t132_n101: return -0.015153972432017326f;
  t132_n102: return 0.03867601975798607f;
  t132_n103: return -0.012804374098777771f;
  t132_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t132_n156;
  t132_n105: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t132_n151;
  t132_n106: if (E.getNumReferences() >= 3346296960 /*62576.5*/) goto t132_n148;
  t132_n107: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t132_n139;
  t132_n108: if (E.getNumReferences() >= 3337423104 /*30340.5*/) goto t132_n138;
  t132_n109: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t132_n129;
  t132_n110: if (E.getNumReferences() >= 3326878720 /*13065.0*/) goto t132_n124;
  t132_n111: if (E.getNumReferences() >= 3325608960 /*11825.0*/) goto t132_n117;
  t132_n112: if (E.getNumReferences() >= 3323745792 /*10005.5*/) goto t132_n114;
  t132_n113: return -0.0004597318184096366f;
  t132_n114: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t132_n116;
  t132_n115: return -0.23923088610172272f;
  t132_n116: return -0.03666738420724869f;
  t132_n117: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t132_n119;
  t132_n118: return 0.08760282397270203f;
  t132_n119: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t132_n121;
  t132_n120: return -0.15224216878414154f;
  t132_n121: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t132_n123;
  t132_n122: return 0.18206609785556793f;
  t132_n123: return 0.11624765396118164f;
  t132_n124: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t132_n128;
  t132_n125: if (E.getNumReferences() >= 3329200128 /*15332.0*/) goto t132_n127;
  t132_n126: return -0.9851589202880859f;
  t132_n127: return -0.16280946135520935f;
  t132_n128: return -0.03359891101717949f;
  t132_n129: if (E.getNumReferences() >= 3335106048 /*25815.0*/) goto t132_n137;
  t132_n130: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t132_n134;
  t132_n131: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t132_n133;
  t132_n132: return -0.01734205149114132f;
  t132_n133: return 0.08851295709609985f;
  t132_n134: if (E.getNumReferences() >= 3332417792 /*20564.5*/) goto t132_n136;
  t132_n135: return 0.13376763463020325f;
  t132_n136: return 0.060023099184036255f;
  t132_n137: return -0.025030015036463737f;
  t132_n138: return -0.15594829618930817f;
  t132_n139: if (E.getNumReferences() >= 3338968320 /*33949.0*/) goto t132_n143;
  t132_n140: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t132_n142;
  t132_n141: return 0.12658020853996277f;
  t132_n142: return 0.07550415396690369f;
  t132_n143: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t132_n147;
  t132_n144: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t132_n146;
  t132_n145: return -0.06455236673355103f;
  t132_n146: return 0.06920372694730759f;
  t132_n147: return -0.960381031036377f;
  t132_n148: if (E.getNumReferences() >= 3347118592 /*66036.0*/) goto t132_n150;
  t132_n149: return -0.26687929034233093f;
  t132_n150: return -0.07570146769285202f;
  t132_n151: if (E.getNumReferences() >= 3352808320 /*110487.0*/) goto t132_n153;
  t132_n152: return 0.10476235300302505f;
  t132_n153: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t132_n155;
  t132_n154: return 0.009627556428313255f;
  t132_n155: return -1.3344062566757202f;
  t132_n156: if (E.getFileProximityDistanceCost() >= 3255042048 /*33.0*/) goto t132_n158;
  t132_n157: return 0.00980337243527174f;
  t132_n158: return 0.10146772861480713f;
  t132_n159: if (E.getFractionNameInContext() >= 3204006750 /*0.4868420958518982*/) goto t132_n181;
  t132_n160: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t132_n166;
  t132_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t132_n165;
  t132_n162: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t132_n164;
  t132_n163: return -0.005439239554107189f;
  t132_n164: return 0.0032844385132193565f;
  t132_n165: return 0.012789590284228325f;
  t132_n166: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t132_n176;
  t132_n167: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t132_n175;
  t132_n168: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t132_n174;
  t132_n169: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t132_n173;
  t132_n170: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t132_n172;
  t132_n171: return -0.04025162756443024f;
  t132_n172: return 0.21169571578502655f;
  t132_n173: return -0.03442415967583656f;
  t132_n174: return -0.12153556942939758f;
  t132_n175: return 0.0004688366025220603f;
  t132_n176: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t132_n180;
  t132_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t132_n179;
  t132_n178: return -0.012555580586194992f;
  t132_n179: return 0.05394105985760689f;
  t132_n180: return -0.033500056713819504f;
  t132_n181: return 0.008053292520344257f;
  t132_n182: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t132_n238;
  t132_n183: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t132_n237;
  t132_n184: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t132_n230;
  t132_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t132_n213;
  t132_n186: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t132_n212;
  t132_n187: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t132_n193;
  t132_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t132_n192;
  t132_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t132_n191;
  t132_n190: return -0.11698000133037567f;
  t132_n191: return -0.10347466170787811f;
  t132_n192: return -0.01704997755587101f;
  t132_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t132_n207;
  t132_n194: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t132_n196;
  t132_n195: return 0.002200815826654434f;
  t132_n196: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t132_n206;
  t132_n197: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t132_n205;
  t132_n198: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t132_n202;
  t132_n199: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t132_n201;
  t132_n200: return -0.020502114668488503f;
  t132_n201: return -0.06846587359905243f;
  t132_n202: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t132_n204;
  t132_n203: return 0.0008919649990275502f;
  t132_n204: return -0.02061012014746666f;
  t132_n205: return -0.18548713624477386f;
  t132_n206: return 0.034468874335289f;
  t132_n207: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t132_n209;
  t132_n208: return 0.01490599662065506f;
  t132_n209: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t132_n211;
  t132_n210: return 0.04641656205058098f;
  t132_n211: return -0.10632477700710297f;
  t132_n212: return -0.013599907979369164f;
  t132_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t132_n225;
  t132_n214: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t132_n216;
  t132_n215: return 0.005879159085452557f;
  t132_n216: if (E.getNumReferences() >= 3317866496 /*6228.0*/) goto t132_n224;
  t132_n217: if (E.getNumReferences() >= 3278143488 /*228.5*/) goto t132_n219;
  t132_n218: return -0.0076774610206484795f;
  t132_n219: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t132_n223;
  t132_n220: if (E.getNumReferences() >= 3278274560 /*230.5*/) goto t132_n222;
  t132_n221: return 0.08896856009960175f;
  t132_n222: return 0.052111294120550156f;
  t132_n223: return 0.1938953697681427f;
  t132_n224: return -0.6128130555152893f;
  t132_n225: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t132_n227;
  t132_n226: return 0.011526976712048054f;
  t132_n227: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t132_n229;
  t132_n228: return 0.05855172500014305f;
  t132_n229: return 0.03138341009616852f;
  t132_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t132_n232;
  t132_n231: return 0.11071084439754486f;
  t132_n232: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t132_n236;
  t132_n233: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t132_n235;
  t132_n234: return 0.10890088230371475f;
  t132_n235: return 0.10675381869077682f;
  t132_n236: return 0.10897216945886612f;
  t132_n237: return -0.027133308351039886f;
  t132_n238: if (E.getNumReferences() >= 3338172416 /*31804.0*/) goto t132_n240;
  t132_n239: return 0.08149267733097076f;
  t132_n240: if (E.getNumReferences() >= 3341699072 /*44616.0*/) goto t132_n244;
  t132_n241: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t132_n243;
  t132_n242: return 0.034298233687877655f;
  t132_n243: return -0.07564513385295868f;
  t132_n244: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t132_n246;
  t132_n245: return 0.07720731943845749f;
  t132_n246: if (E.getNumReferences() >= 3363497856 /*256926.0*/) goto t132_n254;
  t132_n247: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t132_n249;
  t132_n248: return -0.04450218006968498f;
  t132_n249: if (E.getNumReferences() >= 3352414592 /*107411.0*/) goto t132_n251;
  t132_n250: return 0.07798721641302109f;
  t132_n251: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t132_n253;
  t132_n252: return -0.15270227193832397f;
  t132_n253: return 0.0641995221376419f;
  t132_n254: return -0.3237709403038025f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree133(const Example& E) {
  t133_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t133_n182;
  t133_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t133_n43;
  t133_n2: if (E.getNumReferences() >= 3314787328 /*4724.5*/) goto t133_n38;
  t133_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t133_n21;
  t133_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t133_n6;
  t133_n5: return -1.6424253772129305e-05f;
  t133_n6: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t133_n20;
  t133_n7: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t133_n19;
  t133_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t133_n14;
  t133_n9: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t133_n11;
  t133_n10: return -0.05426008254289627f;
  t133_n11: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t133_n13;
  t133_n12: return 0.11916641891002655f;
  t133_n13: return -0.0760713517665863f;
  t133_n14: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t133_n18;
  t133_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t133_n17;
  t133_n16: return 0.053786247968673706f;
  t133_n17: return -0.08828332275152206f;
  t133_n18: return 0.17748208343982697f;
  t133_n19: return -0.10784900188446045f;
  t133_n20: return -0.14632746577262878f;
  t133_n21: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t133_n31;
  t133_n22: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t133_n24;
  t133_n23: return -0.01222881581634283f;
  t133_n24: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t133_n26;
  t133_n25: return 0.0036628604866564274f;
  t133_n26: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t133_n28;
  t133_n27: return 0.02322736196219921f;
  t133_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t133_n30;
  t133_n29: return 0.06543811410665512f;
  t133_n30: return 0.11293602734804153f;
  t133_n31: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t133_n35;
  t133_n32: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t133_n34;
  t133_n33: return 0.009203627705574036f;
  t133_n34: return -0.029851600527763367f;
  t133_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n37;
  t133_n36: return -0.026556165888905525f;
  t133_n37: return 0.0062300353311002254f;
  t133_n38: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t133_n42;
  t133_n39: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t133_n41;
  t133_n40: return 0.09042683243751526f;
  t133_n41: return 0.22141319513320923f;
  t133_n42: return 0.013752676546573639f;
  t133_n43: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t133_n149;
  t133_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t133_n148;
  t133_n45: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t133_n147;
  t133_n46: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t133_n146;
  t133_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t133_n49;
  t133_n48: return -0.003258779877796769f;
  t133_n49: if (E.getNumReferences() >= 3290701824 /*656.5*/) goto t133_n145;
  t133_n50: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t133_n132;
  t133_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t133_n53;
  t133_n52: return -0.038339853286743164f;
  t133_n53: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t133_n55;
  t133_n54: return -0.04532129690051079f;
  t133_n55: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t133_n63;
  t133_n56: if (E.getNumReferences() >= 3289120768 /*560.0*/) goto t133_n58;
  t133_n57: return 0.020344272255897522f;
  t133_n58: if (E.getNumReferences() >= 3289710592 /*596.0*/) goto t133_n62;
  t133_n59: if (E.getNumReferences() >= 3289268224 /*569.0*/) goto t133_n61;
  t133_n60: return 0.09804705530405045f;
  t133_n61: return 0.1354023516178131f;
  t133_n62: return 0.08257932960987091f;
  t133_n63: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t133_n83;
  t133_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n68;
  t133_n65: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t133_n67;
  t133_n66: return -0.027516013011336327f;
  t133_n67: return -0.325901061296463f;
  t133_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t133_n70;
  t133_n69: return 0.008130851201713085f;
  t133_n70: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t133_n80;
  t133_n71: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t133_n75;
  t133_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t133_n74;
  t133_n73: return -0.019036533311009407f;
  t133_n74: return 0.15470652282238007f;
  t133_n75: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t133_n79;
  t133_n76: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t133_n78;
  t133_n77: return 0.08227872103452682f;
  t133_n78: return 0.19835662841796875f;
  t133_n79: return 0.0649571344256401f;
  t133_n80: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t133_n82;
  t133_n81: return 0.055761221796274185f;
  t133_n82: return 0.28348326683044434f;
  t133_n83: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t133_n121;
  t133_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t133_n86;
  t133_n85: return 0.0040780361741781235f;
  t133_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t133_n104;
  t133_n87: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t133_n89;
  t133_n88: return 0.012865160591900349f;
  t133_n89: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t133_n101;
  t133_n90: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t133_n98;
  t133_n91: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t133_n97;
  t133_n92: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t133_n94;
  t133_n93: return 0.0362430103123188f;
  t133_n94: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t133_n96;
  t133_n95: return 0.28126034140586853f;
  t133_n96: return 0.11067378520965576f;
  t133_n97: return -0.004008475225418806f;
  t133_n98: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t133_n100;
  t133_n99: return 0.03629761561751366f;
  t133_n100: return 0.11726465821266174f;
  t133_n101: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t133_n103;
  t133_n102: return -0.02189992554485798f;
  t133_n103: return 0.02881796844303608f;
  t133_n104: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t133_n106;
  t133_n105: return -0.0183044895529747f;
  t133_n106: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t133_n114;
  t133_n107: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t133_n109;
  t133_n108: return 0.04735817015171051f;
  t133_n109: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t133_n111;
  t133_n110: return 0.11654902249574661f;
  t133_n111: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t133_n113;
  t133_n112: return -0.061565693467855453f;
  t133_n113: return 0.11006142944097519f;
  t133_n114: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t133_n116;
  t133_n115: return -0.03475746512413025f;
  t133_n116: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t133_n118;
  t133_n117: return 0.4000227451324463f;
  t133_n118: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t133_n120;
  t133_n119: return 0.06954678893089294f;
  t133_n120: return -0.010636524297297001f;
  t133_n121: if (E.getNumReferences() >= 3265331200 /*80.5*/) goto t133_n131;
  t133_n122: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t133_n130;
  t133_n123: if (E.getNumReferences() >= 3265069056 /*78.5*/) goto t133_n129;
  t133_n124: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t133_n128;
  t133_n125: if (E.getNumReferences() >= 3264937984 /*77.5*/) goto t133_n127;
  t133_n126: return 0.8609305024147034f;
  t133_n127: return -0.2796136736869812f;
  t133_n128: return -0.13600489497184753f;
  t133_n129: return 0.5105454325675964f;
  t133_n130: return 0.03397351875901222f;
  t133_n131: return 0.03971867635846138f;
  t133_n132: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t133_n134;
  t133_n133: return -0.051343098282814026f;
  t133_n134: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t133_n136;
  t133_n135: return 0.011854995973408222f;
  t133_n136: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t133_n138;
  t133_n137: return -0.11026690155267715f;
  t133_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t133_n140;
  t133_n139: return -0.11512011289596558f;
  t133_n140: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t133_n144;
  t133_n141: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t133_n143;
  t133_n142: return 0.17540815472602844f;
  t133_n143: return -0.43990230560302734f;
  t133_n144: return -0.130198135972023f;
  t133_n145: return -0.014513703994452953f;
  t133_n146: return 0.02459239959716797f;
  t133_n147: return -0.021125925704836845f;
  t133_n148: return 0.0020274079870432615f;
  t133_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t133_n155;
  t133_n150: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t133_n152;
  t133_n151: return -0.021052774041891098f;
  t133_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n154;
  t133_n153: return -0.01119181327521801f;
  t133_n154: return 0.002293128753080964f;
  t133_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t133_n177;
  t133_n156: if (E.getFractionNameInContext() >= 3176672643 /*0.05277777835726738*/) goto t133_n176;
  t133_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t133_n163;
  t133_n158: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t133_n162;
  t133_n159: if (E.getNumReferences() >= 3332163840 /*20068.5*/) goto t133_n161;
  t133_n160: return -0.028521491214632988f;
  t133_n161: return -0.3485015034675598f;
  t133_n162: return 0.04280893877148628f;
  t133_n163: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t133_n175;
  t133_n164: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t133_n174;
  t133_n165: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t133_n173;
  t133_n166: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t133_n170;
  t133_n167: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t133_n169;
  t133_n168: return 0.11422640085220337f;
  t133_n169: return 0.031166445463895798f;
  t133_n170: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t133_n172;
  t133_n171: return 0.12437940388917923f;
  t133_n172: return 0.1000467836856842f;
  t133_n173: return -0.0736483708024025f;
  t133_n174: return 0.06805077195167542f;
  t133_n175: return 0.013325969688594341f;
  t133_n176: return -0.042906321585178375f;
  t133_n177: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t133_n181;
  t133_n178: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t133_n180;
  t133_n179: return 0.049379270523786545f;
  t133_n180: return 0.022409895434975624f;
  t133_n181: return 0.0035641142167150974f;
  t133_n182: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t133_n226;
  t133_n183: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t133_n215;
  t133_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n188;
  t133_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t133_n187;
  t133_n186: return -0.05719872564077377f;
  t133_n187: return -0.014270126819610596f;
  t133_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t133_n208;
  t133_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t133_n207;
  t133_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n202;
  t133_n191: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t133_n197;
  t133_n192: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t133_n194;
  t133_n193: return -0.015745334327220917f;
  t133_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n196;
  t133_n195: return -0.0009317589574493468f;
  t133_n196: return 0.03158142790198326f;
  t133_n197: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t133_n199;
  t133_n198: return -0.04410915821790695f;
  t133_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t133_n201;
  t133_n200: return 0.006694428622722626f;
  t133_n201: return 0.10933035612106323f;
  t133_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t133_n204;
  t133_n203: return -0.038665637373924255f;
  t133_n204: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t133_n206;
  t133_n205: return 0.0312652513384819f;
  t133_n206: return 0.12337850779294968f;
  t133_n207: return 0.03102671355009079f;
  t133_n208: if (E.getFractionNameInContext() >= 3199584274 /*0.355042040348053*/) goto t133_n214;
  t133_n209: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t133_n211;
  t133_n210: return 0.0008999896235764027f;
  t133_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t133_n213;
  t133_n212: return 0.049368515610694885f;
  t133_n213: return 0.10781511664390564f;
  t133_n214: return 0.025639701634645462f;
  t133_n215: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t133_n217;
  t133_n216: return 0.00928430911153555f;
  t133_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t133_n225;
  t133_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace))) goto t133_n220;
  t133_n219: return -7.276600081240758e-05f;
  t133_n220: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t133_n224;
  t133_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t133_n223;
  t133_n222: return -0.017492705956101418f;
  t133_n223: return 0.0855901911854744f;
  t133_n224: return -0.10625940561294556f;
  t133_n225: return 0.01201393362134695f;
  t133_n226: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t133_n254;
  t133_n227: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t133_n229;
  t133_n228: return 0.0052569047547876835f;
  t133_n229: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t133_n253;
  t133_n230: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t133_n232;
  t133_n231: return 0.010791255161166191f;
  t133_n232: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t133_n234;
  t133_n233: return 0.2249753177165985f;
  t133_n234: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t133_n252;
  t133_n235: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t133_n251;
  t133_n236: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t133_n250;
  t133_n237: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t133_n247;
  t133_n238: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t133_n242;
  t133_n239: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t133_n241;
  t133_n240: return 0.1648392230272293f;
  t133_n241: return -0.29267844557762146f;
  t133_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t133_n246;
  t133_n243: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t133_n245;
  t133_n244: return -0.040490809828042984f;
  t133_n245: return 0.08474448323249817f;
  t133_n246: return 0.2060401886701584f;
  t133_n247: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t133_n249;
  t133_n248: return 0.12349305301904678f;
  t133_n249: return 0.1131838709115982f;
  t133_n250: return -0.027822459116578102f;
  t133_n251: return 0.10335563123226166f;
  t133_n252: return 0.018286192789673805f;
  t133_n253: return 0.13223795592784882f;
  t133_n254: return 0.033986277878284454f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree134(const Example& E) {
  t134_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t134_n252;
  t134_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t134_n235;
  t134_n2: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t134_n234;
  t134_n3: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t134_n225;
  t134_n4: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t134_n224;
  t134_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t134_n43;
  t134_n6: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t134_n22;
  t134_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n21;
  t134_n8: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t134_n10;
  t134_n9: return -0.004712421912699938f;
  t134_n10: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t134_n20;
  t134_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t134_n15;
  t134_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t134_n14;
  t134_n13: return -0.14016197621822357f;
  t134_n14: return 0.017599718645215034f;
  t134_n15: if (E.getNumReferences() >= 3274211328 /*168.5*/) goto t134_n17;
  t134_n16: return 0.022893289104104042f;
  t134_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t134_n19;
  t134_n18: return 0.042930010706186295f;
  t134_n19: return 0.14399966597557068f;
  t134_n20: return 0.003902002004906535f;
  t134_n21: return 0.0011827511480078101f;
  t134_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t134_n42;
  t134_n23: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t134_n27;
  t134_n24: if (E.getNumReferences() >= 3350889856 /*95499.0*/) goto t134_n26;
  t134_n25: return 0.010032344609498978f;
  t134_n26: return -0.1321980506181717f;
  t134_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t134_n31;
  t134_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t134_n30;
  t134_n29: return 0.00852958019822836f;
  t134_n30: return 0.10565686970949173f;
  t134_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t134_n39;
  t134_n32: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t134_n36;
  t134_n33: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t134_n35;
  t134_n34: return 0.056860532611608505f;
  t134_n35: return 0.10277403146028519f;
  t134_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t134_n38;
  t134_n37: return 0.07462351024150848f;
  t134_n38: return -0.124207504093647f;
  t134_n39: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t134_n41;
  t134_n40: return 0.10996513068675995f;
  t134_n41: return -0.06271757930517197f;
  t134_n42: return -0.004558957181870937f;
  t134_n43: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t134_n45;
  t134_n44: return 0.00758777279406786f;
  t134_n45: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t134_n169;
  t134_n46: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t134_n166;
  t134_n47: if (E.getNumReferences() >= 3357335872 /*160645.0*/) goto t134_n165;
  t134_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n64;
  t134_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t134_n51;
  t134_n50: return -0.00837474875152111f;
  t134_n51: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t134_n61;
  t134_n52: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t134_n60;
  t134_n53: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t134_n55;
  t134_n54: return 0.00020478229271247983f;
  t134_n55: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t134_n57;
  t134_n56: return 0.014635924249887466f;
  t134_n57: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t134_n59;
  t134_n58: return 0.04184703528881073f;
  t134_n59: return -0.011967801488935947f;
  t134_n60: return -0.08268596231937408f;
  t134_n61: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t134_n63;
  t134_n62: return 0.09720033407211304f;
  t134_n63: return 0.12642629444599152f;
  t134_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t134_n82;
  t134_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n71;
  t134_n66: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t134_n70;
  t134_n67: if (E.getNumReferences() >= 3325242368 /*11467.0*/) goto t134_n69;
  t134_n68: return -0.08684992045164108f;
  t134_n69: return -0.13882435858249664f;
  t134_n70: return -0.14758314192295074f;
  t134_n71: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t134_n81;
  t134_n72: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t134_n76;
  t134_n73: if (E.getNumReferences() >= 3277717504 /*222.0*/) goto t134_n75;
  t134_n74: return 0.007650325540453196f;
  t134_n75: return -0.07410300523042679f;
  t134_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t134_n78;
  t134_n77: return -0.10280642658472061f;
  t134_n78: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t134_n80;
  t134_n79: return 0.024591373279690742f;
  t134_n80: return 0.09186425805091858f;
  t134_n81: return -0.012639855034649372f;
  t134_n82: if (E.getNumReferences() >= 3347421312 /*68401.0*/) goto t134_n160;
  t134_n83: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t134_n157;
  t134_n84: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t134_n118;
  t134_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n91;
  t134_n86: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t134_n90;
  t134_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t134_n89;
  t134_n88: return -0.012490826658904552f;
  t134_n89: return 0.015383767895400524f;
  t134_n90: return -0.009677935391664505f;
  t134_n91: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t134_n117;
  t134_n92: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t134_n110;
  t134_n93: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t134_n107;
  t134_n94: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t134_n100;
  t134_n95: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t134_n97;
  t134_n96: return 0.0057676793076097965f;
  t134_n97: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t134_n99;
  t134_n98: return -0.004962095990777016f;
  t134_n99: return 0.008502139709889889f;
  t134_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t134_n106;
  t134_n101: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t134_n103;
  t134_n102: return -0.039633672684431076f;
  t134_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t134_n105;
  t134_n104: return 0.03224535658955574f;
  t134_n105: return 0.10629629343748093f;
  t134_n106: return 0.002116968622431159f;
  t134_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n109;
  t134_n108: return -0.002707627136260271f;
  t134_n109: return 0.0225498266518116f;
  t134_n110: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t134_n116;
  t134_n111: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t134_n113;
  t134_n112: return -0.001528955646790564f;
  t134_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n115;
  t134_n114: return -0.10630203038454056f;
  t134_n115: return 0.1431623101234436f;
  t134_n116: return -0.05667947977781296f;
  t134_n117: return 0.02249608188867569f;
  t134_n118: if (E.getNumReferences() >= 3322180096 /*8476.5*/) goto t134_n134;
  t134_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n123;
  t134_n120: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t134_n122;
  t134_n121: return -0.1172778382897377f;
  t134_n122: return -0.016525056213140488f;
  t134_n123: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t134_n125;
  t134_n124: return -0.08996648341417313f;
  t134_n125: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t134_n129;
  t134_n126: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t134_n128;
  t134_n127: return 0.06688954681158066f;
  t134_n128: return 0.02178775519132614f;
  t134_n129: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t134_n131;
  t134_n130: return -0.0663188248872757f;
  t134_n131: if (E.getNumReferences() >= 3320417280 /*7473.5*/) goto t134_n133;
  t134_n132: return 0.025562740862369537f;
  t134_n133: return -0.1215212494134903f;
  t134_n134: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t134_n152;
  t134_n135: if (E.getNumReferences() >= 3344048512 /*53793.5*/) goto t134_n147;
  t134_n136: if (E.getNumReferences() >= 3342301952 /*46971.0*/) goto t134_n146;
  t134_n137: if (E.getNumReferences() >= 3342115200 /*46241.5*/) goto t134_n145;
  t134_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t134_n144;
  t134_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n143;
  t134_n140: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t134_n142;
  t134_n141: return -0.10338429361581802f;
  t134_n142: return 0.02654184214770794f;
  t134_n143: return 0.035159796476364136f;
  t134_n144: return 0.021756984293460846f;
  t134_n145: return 0.07676829397678375f;
  t134_n146: return -0.06588133424520493f;
  t134_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n149;
  t134_n148: return -0.07026449590921402f;
  t134_n149: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t134_n151;
  t134_n150: return 0.04762813076376915f;
  t134_n151: return 0.10067971795797348f;
  t134_n152: if (E.getNumReferences() >= 3345968640 /*61294.0*/) goto t134_n156;
  t134_n153: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t134_n155;
  t134_n154: return -0.06662780791521072f;
  t134_n155: return -0.3873879909515381f;
  t134_n156: return -0.05512602627277374f;
  t134_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t134_n159;
  t134_n158: return -0.022048654034733772f;
  t134_n159: return 0.0433371476829052f;
  t134_n160: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t134_n162;
  t134_n161: return 0.008066191338002682f;
  t134_n162: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t134_n164;
  t134_n163: return -0.0409347303211689f;
  t134_n164: return 0.05106541886925697f;
  t134_n165: return -0.20275366306304932f;
  t134_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n168;
  t134_n167: return -0.18289320170879364f;
  t134_n168: return 0.07416559755802155f;
  t134_n169: if (E.getNumReferences() >= 3344576000 /*55854.0*/) goto t134_n223;
  t134_n170: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t134_n216;
  t134_n171: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t134_n173;
  t134_n172: return -0.05465766042470932f;
  t134_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n187;
  t134_n174: if (E.getNumReferences() >= 3334472448 /*24577.5*/) goto t134_n186;
  t134_n175: if (E.getNumReferences() >= 3333992192 /*23639.5*/) goto t134_n177;
  t134_n176: return -0.024104753509163857f;
  t134_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n179;
  t134_n178: return -0.11697892844676971f;
  t134_n179: if (E.getNumReferences() >= 3334143744 /*23935.5*/) goto t134_n185;
  t134_n180: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t134_n184;
  t134_n181: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t134_n183;
  t134_n182: return 0.11590353399515152f;
  t134_n183: return 0.19421051442623138f;
  t134_n184: return 0.10352880507707596f;
  t134_n185: return 0.09532508254051208f;
  t134_n186: return -0.1376655399799347f;
  t134_n187: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t134_n211;
  t134_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t134_n190;
  t134_n189: return 0.0018011487554758787f;
  t134_n190: if (E.getNumReferences() >= 3327147008 /*13327.0*/) goto t134_n196;
  t134_n191: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t134_n195;
  t134_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t134_n194;
  t134_n193: return -0.09904047101736069f;
  t134_n194: return 0.0800977498292923f;
  t134_n195: return -0.013666884042322636f;
  t134_n196: if (E.getNumReferences() >= 3327212032 /*13390.5*/) goto t134_n198;
  t134_n197: return 0.4124438464641571f;
  t134_n198: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t134_n210;
  t134_n199: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t134_n209;
  t134_n200: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t134_n208;
  t134_n201: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t134_n207;
  t134_n202: if (E.getNumReferences() >= 3331260928 /*18305.0*/) goto t134_n204;
  t134_n203: return 0.07781140506267548f;
  t134_n204: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t134_n206;
  t134_n205: return 0.15764276683330536f;
  t134_n206: return 0.09583566337823868f;
  t134_n207: return 0.07009807974100113f;
  t134_n208: return 0.19992277026176453f;
  t134_n209: return -0.13246724009513855f;
  t134_n210: return -0.04343261197209358f;
  t134_n211: if (E.getNumReferences() >= 3342583296 /*48070.0*/) goto t134_n215;
  t134_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n214;
  t134_n213: return -0.15596041083335876f;
  t134_n214: return 0.09274778515100479f;
  t134_n215: return -0.030554134398698807f;
  t134_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n218;
  t134_n217: return -0.18326330184936523f;
  t134_n218: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t134_n222;
  t134_n219: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t134_n221;
  t134_n220: return 0.06905090063810349f;
  t134_n221: return 0.1355765163898468f;
  t134_n222: return 0.14571157097816467f;
  t134_n223: return -0.04003315791487694f;
  t134_n224: return -0.14958524703979492f;
  t134_n225: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t134_n231;
  t134_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t134_n228;
  t134_n227: return -0.012955172918736935f;
  t134_n228: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t134_n230;
  t134_n229: return -0.03800826519727707f;
  t134_n230: return 0.05335920304059982f;
  t134_n231: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t134_n233;
  t134_n232: return 0.017543915659189224f;
  t134_n233: return -0.06972047686576843f;
  t134_n234: return -0.05459415912628174f;
  t134_n235: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t134_n247;
  t134_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t134_n240;
  t134_n237: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t134_n239;
  t134_n238: return 0.023458095267415047f;
  t134_n239: return -0.052185337990522385f;
  t134_n240: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t134_n246;
  t134_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t134_n243;
  t134_n242: return 0.02666710689663887f;
  t134_n243: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t134_n245;
  t134_n244: return 0.059591226279735565f;
  t134_n245: return 0.09180008620023727f;
  t134_n246: return 0.002480890369042754f;
  t134_n247: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t134_n249;
  t134_n248: return -0.08489398658275604f;
  t134_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t134_n251;
  t134_n250: return -0.0001880819909274578f;
  t134_n251: return 0.05310757830739021f;
  t134_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t134_n254;
  t134_n253: return 0.0026029536966234446f;
  t134_n254: return 0.00675848126411438f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree135(const Example& E) {
  t135_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t135_n252;
  t135_n1: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t135_n211;
  t135_n2: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t135_n210;
  t135_n3: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t135_n209;
  t135_n4: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t135_n180;
  t135_n5: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t135_n179;
  t135_n6: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t135_n132;
  t135_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t135_n13;
  t135_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n12;
  t135_n9: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t135_n11;
  t135_n10: return -0.07614192366600037f;
  t135_n11: return -0.020704131573438644f;
  t135_n12: return 0.0008905512513592839f;
  t135_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t135_n15;
  t135_n14: return -0.0005744961090385914f;
  t135_n15: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t135_n93;
  t135_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n40;
  t135_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n31;
  t135_n18: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t135_n20;
  t135_n19: return -0.0914255902171135f;
  t135_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n22;
  t135_n21: return -0.10700166970491409f;
  t135_n22: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t135_n30;
  t135_n23: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t135_n25;
  t135_n24: return 0.002085469663143158f;
  t135_n25: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t135_n29;
  t135_n26: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t135_n28;
  t135_n27: return 0.00897774938493967f;
  t135_n28: return 0.09290006011724472f;
  t135_n29: return -0.011520959436893463f;
  t135_n30: return -0.008547231554985046f;
  t135_n31: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t135_n33;
  t135_n32: return -0.04483933374285698f;
  t135_n33: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t135_n39;
  t135_n34: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t135_n38;
  t135_n35: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t135_n37;
  t135_n36: return -0.11177810281515121f;
  t135_n37: return 0.11526324599981308f;
  t135_n38: return 0.10714627057313919f;
  t135_n39: return 0.019261864945292473f;
  t135_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n62;
  t135_n41: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t135_n43;
  t135_n42: return -0.021843209862709045f;
  t135_n43: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t135_n45;
  t135_n44: return 0.009112276136875153f;
  t135_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n57;
  t135_n46: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t135_n48;
  t135_n47: return -0.034781619906425476f;
  t135_n48: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t135_n52;
  t135_n49: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t135_n51;
  t135_n50: return 0.014979771338403225f;
  t135_n51: return 0.07276316732168198f;
  t135_n52: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t135_n56;
  t135_n53: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t135_n55;
  t135_n54: return 0.1658807396888733f;
  t135_n55: return 0.09197455644607544f;
  t135_n56: return -0.014627712778747082f;
  t135_n57: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t135_n61;
  t135_n58: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t135_n60;
  t135_n59: return 0.09469389915466309f;
  t135_n60: return 0.003858044045045972f;
  t135_n61: return 0.027735400944948196f;
  t135_n62: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t135_n92;
  t135_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n81;
  t135_n64: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t135_n80;
  t135_n65: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t135_n73;
  t135_n66: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t135_n72;
  t135_n67: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t135_n69;
  t135_n68: return -0.0986628457903862f;
  t135_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n71;
  t135_n70: return -0.03922317922115326f;
  t135_n71: return 0.09765610843896866f;
  t135_n72: return 0.039141733199357986f;
  t135_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n75;
  t135_n74: return -0.028326673433184624f;
  t135_n75: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t135_n79;
  t135_n76: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t135_n78;
  t135_n77: return 0.08183860033750534f;
  t135_n78: return -0.4306332468986511f;
  t135_n79: return 0.00885771494358778f;
  t135_n80: return -0.031342294067144394f;
  t135_n81: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t135_n87;
  t135_n82: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t135_n84;
  t135_n83: return 0.005748257972300053f;
  t135_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n86;
  t135_n85: return -0.003770761890336871f;
  t135_n86: return 0.011218525469303131f;
  t135_n87: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t135_n91;
  t135_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n90;
  t135_n89: return -0.08167380094528198f;
  t135_n90: return 0.00554429879412055f;
  t135_n91: return 0.027484215795993805f;
  t135_n92: return 0.0010362752946093678f;
  t135_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t135_n109;
  t135_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t135_n96;
  t135_n95: return 0.0037050058599561453f;
  t135_n96: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t135_n102;
  t135_n97: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t135_n99;
  t135_n98: return -0.044720426201820374f;
  t135_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n101;
  t135_n100: return -0.15300437808036804f;
  t135_n101: return 0.017649006098508835f;
  t135_n102: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t135_n108;
  t135_n103: if (E.getFractionNameInContext() >= 3177130029 /*0.05448167398571968*/) goto t135_n107;
  t135_n104: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t135_n106;
  t135_n105: return -0.22504429519176483f;
  t135_n106: return 0.05857612565159798f;
  t135_n107: return 0.08509765565395355f;
  t135_n108: return 0.1147279143333435f;
  t135_n109: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t135_n113;
  t135_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t135_n112;
  t135_n111: return 0.0054090130142867565f;
  t135_n112: return -0.2303236573934555f;
  t135_n113: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t135_n115;
  t135_n114: return -0.10730710625648499f;
  t135_n115: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t135_n129;
  t135_n116: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t135_n124;
  t135_n117: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t135_n123;
  t135_n118: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t135_n120;
  t135_n119: return 0.049957748502492905f;
  t135_n120: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t135_n122;
  t135_n121: return 0.10331213474273682f;
  t135_n122: return -0.05318252742290497f;
  t135_n123: return -0.002391108078882098f;
  t135_n124: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t135_n126;
  t135_n125: return 0.10504040122032166f;
  t135_n126: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t135_n128;
  t135_n127: return -0.002517029643058777f;
  t135_n128: return 0.11262232810258865f;
  t135_n129: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t135_n131;
  t135_n130: return -0.025379842147231102f;
  t135_n131: return 0.12047511339187622f;
  t135_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n136;
  t135_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n135;
  t135_n134: return -0.21304729580879211f;
  t135_n135: return -0.01265617273747921f;
  t135_n136: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t135_n146;
  t135_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n141;
  t135_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t135_n140;
  t135_n139: return -0.017383024096488953f;
  t135_n140: return 0.004109319765120745f;
  t135_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t135_n145;
  t135_n142: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t135_n144;
  t135_n143: return 0.11671609431505203f;
  t135_n144: return 0.005200193263590336f;
  t135_n145: return 0.04428980126976967f;
  t135_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t135_n168;
  t135_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t135_n151;
  t135_n148: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t135_n150;
  t135_n149: return -0.039351195096969604f;
  t135_n150: return 0.11862979829311371f;
  t135_n151: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t135_n155;
  t135_n152: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t135_n154;
  t135_n153: return 0.10355139523744583f;
  t135_n154: return -0.033183977007865906f;
  t135_n155: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t135_n163;
  t135_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n162;
  t135_n157: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t135_n161;
  t135_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t135_n160;
  t135_n159: return -0.02920595183968544f;
  t135_n160: return 0.05443519353866577f;
  t135_n161: return -0.10831350833177567f;
  t135_n162: return 0.08451537042856216f;
  t135_n163: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t135_n167;
  t135_n164: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t135_n166;
  t135_n165: return 0.002957037417218089f;
  t135_n166: return -0.7489467263221741f;
  t135_n167: return 0.03718357905745506f;
  t135_n168: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t135_n178;
  t135_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t135_n175;
  t135_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t135_n174;
  t135_n171: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t135_n173;
  t135_n172: return 0.21738301217556f;
  t135_n173: return -0.03115501068532467f;
  t135_n174: return 0.015947308391332626f;
  t135_n175: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t135_n177;
  t135_n176: return 0.02058020420372486f;
  t135_n177: return -0.12391205877065659f;
  t135_n178: return 0.004039331804960966f;
  t135_n179: return -0.02048558183014393f;
  t135_n180: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t135_n194;
  t135_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t135_n187;
  t135_n182: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t135_n184;
  t135_n183: return -0.048718564212322235f;
  t135_n184: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t135_n186;
  t135_n185: return 0.07834428548812866f;
  t135_n186: return -0.05231255665421486f;
  t135_n187: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t135_n189;
  t135_n188: return -0.12980826199054718f;
  t135_n189: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t135_n193;
  t135_n190: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t135_n192;
  t135_n191: return 0.0004342622123658657f;
  t135_n192: return 0.11377774924039841f;
  t135_n193: return 0.23571592569351196f;
  t135_n194: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t135_n198;
  t135_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t135_n197;
  t135_n196: return 0.006236442364752293f;
  t135_n197: return 0.024762321263551712f;
  t135_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t135_n208;
  t135_n199: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t135_n201;
  t135_n200: return -0.013286944478750229f;
  t135_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t135_n205;
  t135_n202: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t135_n204;
  t135_n203: return 0.04545416682958603f;
  t135_n204: return 0.123537078499794f;
  t135_n205: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t135_n207;
  t135_n206: return 0.18059031665325165f;
  t135_n207: return -0.11236737668514252f;
  t135_n208: return 0.00471850298345089f;
  t135_n209: return -0.06608335673809052f;
  t135_n210: return -0.024745650589466095f;
  t135_n211: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t135_n245;
  t135_n212: if (E.getNumReferences() >= 3332993536 /*21689.0*/) goto t135_n222;
  t135_n213: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t135_n221;
  t135_n214: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t135_n218;
  t135_n215: if (E.getNumReferences() >= 3332205568 /*20150.0*/) goto t135_n217;
  t135_n216: return 0.008292204700410366f;
  t135_n217: return -0.23891590535640717f;
  t135_n218: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t135_n220;
  t135_n219: return 0.062258388847112656f;
  t135_n220: return 0.10877031087875366f;
  t135_n221: return -0.06066863238811493f;
  t135_n222: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t135_n244;
  t135_n223: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t135_n231;
  t135_n224: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t135_n230;
  t135_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t135_n227;
  t135_n226: return 0.02049807645380497f;
  t135_n227: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t135_n229;
  t135_n228: return 0.03121931478381157f;
  t135_n229: return 0.1560927927494049f;
  t135_n230: return 0.09673146158456802f;
  t135_n231: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t135_n243;
  t135_n232: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t135_n240;
  t135_n233: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t135_n235;
  t135_n234: return -0.09778054803609848f;
  t135_n235: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t135_n239;
  t135_n236: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t135_n238;
  t135_n237: return 0.09116168320178986f;
  t135_n238: return 0.13274243474006653f;
  t135_n239: return 0.02380344085395336f;
  t135_n240: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t135_n242;
  t135_n241: return 0.09761015325784683f;
  t135_n242: return 0.12216737866401672f;
  t135_n243: return -0.04407130181789398f;
  t135_n244: return 0.013747102580964565f;
  t135_n245: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t135_n247;
  t135_n246: return 0.0004018302424810827f;
  t135_n247: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t135_n249;
  t135_n248: return 0.017679976299405098f;
  t135_n249: if (E.getNumReferences() >= 3283550208 /*366.0*/) goto t135_n251;
  t135_n250: return 0.1504603773355484f;
  t135_n251: return 0.13852377235889435f;
  t135_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t135_n254;
  t135_n253: return 0.0022291792556643486f;
  t135_n254: return 0.007662678137421608f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree136(const Example& E) {
  t136_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t136_n104;
  t136_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t136_n7;
  t136_n2: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t136_n6;
  t136_n3: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t136_n5;
  t136_n4: return 0.1312214434146881f;
  t136_n5: return -0.0009327465668320656f;
  t136_n6: return -0.03266708925366402f;
  t136_n7: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t136_n95;
  t136_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t136_n44;
  t136_n9: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t136_n43;
  t136_n10: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t136_n14;
  t136_n11: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t136_n13;
  t136_n12: return -0.001810825546272099f;
  t136_n13: return -0.0795307382941246f;
  t136_n14: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t136_n16;
  t136_n15: return 0.000618130317889154f;
  t136_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t136_n38;
  t136_n17: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t136_n25;
  t136_n18: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t136_n24;
  t136_n19: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t136_n23;
  t136_n20: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t136_n22;
  t136_n21: return 0.020027635619044304f;
  t136_n22: return 0.0029452084563672543f;
  t136_n23: return 0.14236512780189514f;
  t136_n24: return -0.21682295203208923f;
  t136_n25: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t136_n29;
  t136_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t136_n28;
  t136_n27: return 0.03044774942100048f;
  t136_n28: return 0.08657484501600266f;
  t136_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t136_n33;
  t136_n30: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t136_n32;
  t136_n31: return 0.007316057104617357f;
  t136_n32: return 0.07290057092905045f;
  t136_n33: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t136_n37;
  t136_n34: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t136_n36;
  t136_n35: return 0.08310915529727936f;
  t136_n36: return -0.23156195878982544f;
  t136_n37: return 0.04456363990902901f;
  t136_n38: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t136_n40;
  t136_n39: return -0.04200071841478348f;
  t136_n40: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t136_n42;
  t136_n41: return -0.07059430330991745f;
  t136_n42: return 0.015318037010729313f;
  t136_n43: return -0.04133536294102669f;
  t136_n44: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t136_n52;
  t136_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t136_n47;
  t136_n46: return 0.0018960082670673728f;
  t136_n47: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t136_n51;
  t136_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t136_n50;
  t136_n49: return 0.00028308844775892794f;
  t136_n50: return 0.01354444120079279f;
  t136_n51: return 0.04851605370640755f;
  t136_n52: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t136_n94;
  t136_n53: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t136_n75;
  t136_n54: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t136_n72;
  t136_n55: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t136_n71;
  t136_n56: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t136_n58;
  t136_n57: return -0.03554273024201393f;
  t136_n58: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t136_n70;
  t136_n59: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t136_n61;
  t136_n60: return -0.028206706047058105f;
  t136_n61: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t136_n69;
  t136_n62: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t136_n66;
  t136_n63: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t136_n65;
  t136_n64: return 0.05411040782928467f;
  t136_n65: return -0.23308321833610535f;
  t136_n66: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t136_n68;
  t136_n67: return -0.07617483288049698f;
  t136_n68: return 0.0825815424323082f;
  t136_n69: return 0.17693574726581573f;
  t136_n70: return -0.06762527674436569f;
  t136_n71: return -0.1059688925743103f;
  t136_n72: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t136_n74;
  t136_n73: return -0.0016749135684221983f;
  t136_n74: return 0.019402096047997475f;
  t136_n75: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t136_n81;
  t136_n76: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t136_n78;
  t136_n77: return 0.03252272680401802f;
  t136_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t136_n80;
  t136_n79: return 0.052236009389162064f;
  t136_n80: return 0.13615930080413818f;
  t136_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t136_n83;
  t136_n82: return -0.14159312844276428f;
  t136_n83: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t136_n91;
  t136_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t136_n90;
  t136_n85: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t136_n87;
  t136_n86: return -0.15968668460845947f;
  t136_n87: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t136_n89;
  t136_n88: return 0.09791503101587296f;
  t136_n89: return -0.070843905210495f;
  t136_n90: return 0.04107140749692917f;
  t136_n91: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t136_n93;
  t136_n92: return 0.06992881745100021f;
  t136_n93: return -0.02965729683637619f;
  t136_n94: return 0.12204079329967499f;
  t136_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t136_n103;
  t136_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t136_n98;
  t136_n97: return 0.0026517438236624002f;
  t136_n98: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t136_n100;
  t136_n99: return 0.02898760698735714f;
  t136_n100: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t136_n102;
  t136_n101: return -0.09085630625486374f;
  t136_n102: return 0.07737422734498978f;
  t136_n103: return 0.012463361956179142f;
  t136_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t136_n194;
  t136_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t136_n183;
  t136_n106: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t136_n182;
  t136_n107: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t136_n125;
  t136_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t136_n124;
  t136_n109: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t136_n111;
  t136_n110: return 0.005205557215958834f;
  t136_n111: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t136_n115;
  t136_n112: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t136_n114;
  t136_n113: return 0.02086947113275528f;
  t136_n114: return -0.11782119423151016f;
  t136_n115: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t136_n123;
  t136_n116: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t136_n122;
  t136_n117: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t136_n119;
  t136_n118: return 0.07546158879995346f;
  t136_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t136_n121;
  t136_n120: return 0.0433497354388237f;
  t136_n121: return 0.23361150920391083f;
  t136_n122: return -0.028110170736908913f;
  t136_n123: return 0.09132546186447144f;
  t136_n124: return 0.059705376625061035f;
  t136_n125: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t136_n163;
  t136_n126: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t136_n138;
  t136_n127: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t136_n133;
  t136_n128: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t136_n130;
  t136_n129: return -0.047034647315740585f;
  t136_n130: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t136_n132;
  t136_n131: return -0.20780743658542633f;
  t136_n132: return -0.8331498503684998f;
  t136_n133: if (E.getNumReferences() >= 3252158464 /*27.0*/) goto t136_n135;
  t136_n134: return 0.10499934107065201f;
  t136_n135: if (E.getNumReferences() >= 3262251008 /*60.5*/) goto t136_n137;
  t136_n136: return -0.3092290461063385f;
  t136_n137: return 0.06289072334766388f;
  t136_n138: if (E.getNumReferences() >= 3291635712 /*713.5*/) goto t136_n160;
  t136_n139: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t136_n141;
  t136_n140: return -0.008060585707426071f;
  t136_n141: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t136_n157;
  t136_n142: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t136_n154;
  t136_n143: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t136_n145;
  t136_n144: return 0.018850700929760933f;
  t136_n145: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t136_n153;
  t136_n146: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t136_n148;
  t136_n147: return 0.05686461552977562f;
  t136_n148: if (E.getFractionNameInContext() >= 3180960154 /*0.07500000298023224*/) goto t136_n152;
  t136_n149: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t136_n151;
  t136_n150: return 0.11059262603521347f;
  t136_n151: return 0.13237453997135162f;
  t136_n152: return 0.13543486595153809f;
  t136_n153: return -0.007829384878277779f;
  t136_n154: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t136_n156;
  t136_n155: return 0.1532379537820816f;
  t136_n156: return -0.13489627838134766f;
  t136_n157: if (E.getNumReferences() >= 3288612864 /*529.0*/) goto t136_n159;
  t136_n158: return -0.011137607507407665f;
  t136_n159: return 0.12119866162538528f;
  t136_n160: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t136_n162;
  t136_n161: return -0.3724445700645447f;
  t136_n162: return -0.003912226762622595f;
  t136_n163: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t136_n165;
  t136_n164: return 0.12106246501207352f;
  t136_n165: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t136_n179;
  t136_n166: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t136_n178;
  t136_n167: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t136_n173;
  t136_n168: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t136_n172;
  t136_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t136_n171;
  t136_n170: return 0.05419934540987015f;
  t136_n171: return 0.12218141555786133f;
  t136_n172: return -0.11053638160228729f;
  t136_n173: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t136_n177;
  t136_n174: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t136_n176;
  t136_n175: return 0.15605005621910095f;
  t136_n176: return 0.07912294566631317f;
  t136_n177: return -0.0730074793100357f;
  t136_n178: return 0.22587206959724426f;
  t136_n179: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t136_n181;
  t136_n180: return -0.49768930673599243f;
  t136_n181: return 0.01082985382527113f;
  t136_n182: return 8.564466406824067e-05f;
  t136_n183: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t136_n193;
  t136_n184: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t136_n192;
  t136_n185: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t136_n191;
  t136_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t136_n190;
  t136_n187: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t136_n189;
  t136_n188: return -0.03421202674508095f;
  t136_n189: return 0.04557691514492035f;
  t136_n190: return 0.011266784742474556f;
  t136_n191: return 0.01330705638974905f;
  t136_n192: return 0.039101045578718185f;
  t136_n193: return 0.10851065814495087f;
  t136_n194: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t136_n254;
  t136_n195: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t136_n253;
  t136_n196: if (E.getNumReferences() >= 3283845120 /*375.0*/) goto t136_n246;
  t136_n197: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t136_n209;
  t136_n198: if (E.getFileProximityDistanceCost() >= 3254910976 /*32.5*/) goto t136_n208;
  t136_n199: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t136_n207;
  t136_n200: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t136_n202;
  t136_n201: return 0.02164728380739689f;
  t136_n202: if (E.getNumReferences() >= 3283075072 /*351.5*/) goto t136_n204;
  t136_n203: return 0.0006879388820379972f;
  t136_n204: if (E.getNumReferences() >= 3283140608 /*353.5*/) goto t136_n206;
  t136_n205: return -2.3250505924224854f;
  t136_n206: return -0.07604531943798065f;
  t136_n207: return 0.10637155920267105f;
  t136_n208: return -1.1635589599609375f;
  t136_n209: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t136_n243;
  t136_n210: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t136_n228;
  t136_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t136_n221;
  t136_n212: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t136_n218;
  t136_n213: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t136_n215;
  t136_n214: return -0.025182468816637993f;
  t136_n215: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t136_n217;
  t136_n216: return -0.1263628453016281f;
  t136_n217: return 0.10171560198068619f;
  t136_n218: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t136_n220;
  t136_n219: return 0.05437062308192253f;
  t136_n220: return -0.2442377358675003f;
  t136_n221: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t136_n223;
  t136_n222: return 0.02982574701309204f;
  t136_n223: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t136_n227;
  t136_n224: if (E.getFractionNameInContext() >= 3170893824 /*0.03125*/) goto t136_n226;
  t136_n225: return 0.09660680592060089f;
  t136_n226: return 0.1144624650478363f;
  t136_n227: return -0.039260413497686386f;
  t136_n228: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t136_n232;
  t136_n229: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t136_n231;
  t136_n230: return 0.005302328150719404f;
  t136_n231: return -0.22390438616275787f;
  t136_n232: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t136_n240;
  t136_n233: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t136_n235;
  t136_n234: return 0.019519178196787834f;
  t136_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t136_n237;
  t136_n236: return -0.024944700300693512f;
  t136_n237: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t136_n239;
  t136_n238: return 0.06479702889919281f;
  t136_n239: return 0.11718155443668365f;
  t136_n240: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t136_n242;
  t136_n241: return -0.2191244661808014f;
  t136_n242: return -0.011606038548052311f;
  t136_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t136_n245;
  t136_n244: return -0.1258513182401657f;
  t136_n245: return -0.03998662903904915f;
  t136_n246: if (E.getNumReferences() >= 3284238336 /*387.0*/) goto t136_n250;
  t136_n247: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t136_n249;
  t136_n248: return 0.11907932907342911f;
  t136_n249: return 0.14140792191028595f;
  t136_n250: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t136_n252;
  t136_n251: return 0.027518780902028084f;
  t136_n252: return -0.19327495992183685f;
  t136_n253: return -0.18876557052135468f;
  t136_n254: return -0.06903892755508423f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree137(const Example& E) {
  t137_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t137_n246;
  t137_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t137_n3;
  t137_n2: return -0.1005130484700203f;
  t137_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t137_n59;
  t137_n4: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t137_n52;
  t137_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n51;
  t137_n6: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t137_n8;
  t137_n7: return -0.004194421228021383f;
  t137_n8: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t137_n20;
  t137_n9: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t137_n13;
  t137_n10: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t137_n12;
  t137_n11: return 0.027401555329561234f;
  t137_n12: return 0.049354929476976395f;
  t137_n13: if (E.getNumReferences() >= 3274211328 /*168.5*/) goto t137_n15;
  t137_n14: return 0.004256966523826122f;
  t137_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t137_n19;
  t137_n16: if (E.getNumReferences() >= 3275456512 /*187.5*/) goto t137_n18;
  t137_n17: return 0.10313858091831207f;
  t137_n18: return -0.3750849962234497f;
  t137_n19: return 0.1585739105939865f;
  t137_n20: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t137_n32;
  t137_n21: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t137_n29;
  t137_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n26;
  t137_n23: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t137_n25;
  t137_n24: return -0.06901273131370544f;
  t137_n25: return -0.0014686771901324391f;
  t137_n26: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n28;
  t137_n27: return 0.010416402481496334f;
  t137_n28: return -0.01842416450381279f;
  t137_n29: if (E.getNumReferences() >= 3268214784 /*102.5*/) goto t137_n31;
  t137_n30: return 0.04373729228973389f;
  t137_n31: return -0.30609211325645447f;
  t137_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t137_n50;
  t137_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n43;
  t137_n34: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t137_n40;
  t137_n35: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t137_n39;
  t137_n36: if (E.getFractionNameInContext() >= 3198995661 /*0.3375000059604645*/) goto t137_n38;
  t137_n37: return -0.19676955044269562f;
  t137_n38: return 0.11175564676523209f;
  t137_n39: return 0.11311852931976318f;
  t137_n40: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t137_n42;
  t137_n41: return -0.14458364248275757f;
  t137_n42: return -0.024945441633462906f;
  t137_n43: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t137_n45;
  t137_n44: return 0.0034490302205085754f;
  t137_n45: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t137_n47;
  t137_n46: return -0.08027882128953934f;
  t137_n47: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t137_n49;
  t137_n48: return 0.10964417457580566f;
  t137_n49: return 0.09284064173698425f;
  t137_n50: return -0.011228806339204311f;
  t137_n51: return 0.0012880584690719843f;
  t137_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n56;
  t137_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t137_n55;
  t137_n54: return -0.03187592327594757f;
  t137_n55: return 0.010543291456997395f;
  t137_n56: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n58;
  t137_n57: return 0.013279748149216175f;
  t137_n58: return -0.0033902886789292097f;
  t137_n59: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t137_n61;
  t137_n60: return 0.007638061419129372f;
  t137_n61: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t137_n243;
  t137_n62: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t137_n196;
  t137_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n65;
  t137_n64: return -0.04216315969824791f;
  t137_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t137_n155;
  t137_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n96;
  t137_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t137_n69;
  t137_n68: return -0.007080084644258022f;
  t137_n69: if (E.getNumReferences() >= 3363872672 /*263421.0*/) goto t137_n93;
  t137_n70: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t137_n92;
  t137_n71: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t137_n91;
  t137_n72: if (E.getNumReferences() >= 3337697792 /*30877.0*/) goto t137_n84;
  t137_n73: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t137_n75;
  t137_n74: return 0.00017395502072758973f;
  t137_n75: if (E.getNumReferences() >= 3317536768 /*6067.0*/) goto t137_n81;
  t137_n76: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t137_n80;
  t137_n77: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t137_n79;
  t137_n78: return -0.05824361369013786f;
  t137_n79: return 0.029379069805145264f;
  t137_n80: return -0.1905510127544403f;
  t137_n81: if (E.getNumReferences() >= 3324282368 /*10529.5*/) goto t137_n83;
  t137_n82: return 0.1205582395195961f;
  t137_n83: return 0.0756959542632103f;
  t137_n84: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t137_n90;
  t137_n85: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t137_n89;
  t137_n86: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t137_n88;
  t137_n87: return 0.10910484939813614f;
  t137_n88: return 0.06858232617378235f;
  t137_n89: return 0.017412761226296425f;
  t137_n90: return 0.11244361102581024f;
  t137_n91: return 0.014701871201395988f;
  t137_n92: return -0.08017046749591827f;
  t137_n93: if (E.getNumReferences() >= 3364539984 /*284274.5*/) goto t137_n95;
  t137_n94: return 0.07591350376605988f;
  t137_n95: return 0.03683525323867798f;
  t137_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t137_n112;
  t137_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n101;
  t137_n98: if (E.getNumReferences() >= 3325242368 /*11467.0*/) goto t137_n100;
  t137_n99: return -0.08146468549966812f;
  t137_n100: return -0.13189347088336945f;
  t137_n101: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t137_n107;
  t137_n102: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t137_n104;
  t137_n103: return -0.00726274074986577f;
  t137_n104: if (E.getNumReferences() >= 3349616960 /*85554.5*/) goto t137_n106;
  t137_n105: return 0.0228851530700922f;
  t137_n106: return -0.26024141907691956f;
  t137_n107: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t137_n111;
  t137_n108: if (E.getNumReferences() >= 3336206080 /*27963.5*/) goto t137_n110;
  t137_n109: return -0.01769796386361122f;
  t137_n110: return 0.05131985992193222f;
  t137_n111: return 0.11786990612745285f;
  t137_n112: if (E.getNumReferences() >= 3322245120 /*8540.0*/) goto t137_n150;
  t137_n113: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t137_n117;
  t137_n114: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t137_n116;
  t137_n115: return 0.0034690052270889282f;
  t137_n116: return -0.04478370398283005f;
  t137_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n123;
  t137_n118: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t137_n122;
  t137_n119: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t137_n121;
  t137_n120: return -0.12033907324075699f;
  t137_n121: return 0.13334432244300842f;
  t137_n122: return -0.01611350290477276f;
  t137_n123: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t137_n125;
  t137_n124: return -0.05301975831389427f;
  t137_n125: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t137_n135;
  t137_n126: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t137_n134;
  t137_n127: if (E.getNumReferences() >= 3289554944 /*586.5*/) goto t137_n133;
  t137_n128: if (E.getNumReferences() >= 3289104384 /*559.0*/) goto t137_n132;
  t137_n129: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t137_n131;
  t137_n130: return -0.04330899193882942f;
  t137_n131: return 0.10234995931386948f;
  t137_n132: return 0.13602204620838165f;
  t137_n133: return -0.037844061851501465f;
  t137_n134: return 0.020129607990384102f;
  t137_n135: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t137_n137;
  t137_n136: return -0.06070975586771965f;
  t137_n137: if (E.getNumReferences() >= 3320337408 /*7434.5*/) goto t137_n149;
  t137_n138: if (E.getNumReferences() >= 3316756480 /*5686.0*/) goto t137_n140;
  t137_n139: return 0.023061631247401237f;
  t137_n140: if (E.getNumReferences() >= 3316938752 /*5775.0*/) goto t137_n144;
  t137_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t137_n143;
  t137_n142: return 0.07394558191299438f;
  t137_n143: return 0.19874288141727448f;
  t137_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t137_n146;
  t137_n145: return 0.025234615430235863f;
  t137_n146: if (E.getNumReferences() >= 3317367808 /*5984.5*/) goto t137_n148;
  t137_n147: return 0.1178746223449707f;
  t137_n148: return 0.09978839755058289f;
  t137_n149: return -0.08782931417226791f;
  t137_n150: if (E.getNumReferences() >= 3324619776 /*10859.0*/) goto t137_n152;
  t137_n151: return 0.05254797264933586f;
  t137_n152: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t137_n154;
  t137_n153: return 0.00906816590577364f;
  t137_n154: return 0.36258992552757263f;
  t137_n155: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t137_n175;
  t137_n156: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t137_n172;
  t137_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t137_n159;
  t137_n158: return 0.01993720419704914f;
  t137_n159: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t137_n171;
  t137_n160: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t137_n170;
  t137_n161: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t137_n169;
  t137_n162: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t137_n166;
  t137_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t137_n165;
  t137_n164: return -0.018311945721507072f;
  t137_n165: return 0.06973990797996521f;
  t137_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t137_n168;
  t137_n167: return 0.0700298324227333f;
  t137_n168: return 0.11522802710533142f;
  t137_n169: return -0.051403388381004333f;
  t137_n170: return -0.6834306716918945f;
  t137_n171: return 0.1160333901643753f;
  t137_n172: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t137_n174;
  t137_n173: return -0.0014711321564391255f;
  t137_n174: return 0.0533621683716774f;
  t137_n175: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t137_n177;
  t137_n176: return -0.07185915112495422f;
  t137_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n195;
  t137_n178: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t137_n194;
  t137_n179: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t137_n193;
  t137_n180: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t137_n186;
  t137_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t137_n183;
  t137_n182: return 0.05075100436806679f;
  t137_n183: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t137_n185;
  t137_n184: return 0.08994629979133606f;
  t137_n185: return 0.07947777956724167f;
  t137_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t137_n192;
  t137_n187: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t137_n191;
  t137_n188: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t137_n190;
  t137_n189: return -0.029181823134422302f;
  t137_n190: return 0.06092042848467827f;
  t137_n191: return -0.025570079684257507f;
  t137_n192: return 0.055521395057439804f;
  t137_n193: return 0.009506402537226677f;
  t137_n194: return 0.07364392280578613f;
  t137_n195: return -0.15023371577262878f;
  t137_n196: if (E.getNumReferences() >= 3344576000 /*55854.0*/) goto t137_n242;
  t137_n197: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t137_n239;
  t137_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n204;
  t137_n199: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t137_n201;
  t137_n200: return -0.04056360945105553f;
  t137_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t137_n203;
  t137_n202: return -0.1037021353840828f;
  t137_n203: return 0.12214494496583939f;
  t137_n204: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t137_n206;
  t137_n205: return -0.048676829785108566f;
  t137_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t137_n214;
  t137_n207: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t137_n209;
  t137_n208: return -0.001574713154695928f;
  t137_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n211;
  t137_n210: return -0.1229916512966156f;
  t137_n211: if (E.getNumReferences() >= 3342583296 /*48070.0*/) goto t137_n213;
  t137_n212: return 0.0870227962732315f;
  t137_n213: return -0.008573790080845356f;
  t137_n214: if (E.getNumReferences() >= 3327147008 /*13327.0*/) goto t137_n222;
  t137_n215: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n221;
  t137_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t137_n218;
  t137_n217: return -0.09269993752241135f;
  t137_n218: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t137_n220;
  t137_n219: return 0.0847490057349205f;
  t137_n220: return -0.026375558227300644f;
  t137_n221: return -0.01297700684517622f;
  t137_n222: if (E.getNumReferences() >= 3327212032 /*13390.5*/) goto t137_n224;
  t137_n223: return 0.293490469455719f;
  t137_n224: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t137_n238;
  t137_n225: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t137_n237;
  t137_n226: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t137_n234;
  t137_n227: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t137_n233;
  t137_n228: if (E.getNumReferences() >= 3331260928 /*18305.0*/) goto t137_n230;
  t137_n229: return 0.07219183444976807f;
  t137_n230: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t137_n232;
  t137_n231: return 0.15834835171699524f;
  t137_n232: return 0.10646718740463257f;
  t137_n233: return 0.05788595229387283f;
  t137_n234: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t137_n236;
  t137_n235: return 0.13805517554283142f;
  t137_n236: return 0.227375328540802f;
  t137_n237: return -0.12419701367616653f;
  t137_n238: return -0.04429379850625992f;
  t137_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n241;
  t137_n240: return -0.1807805597782135f;
  t137_n241: return 0.13581161201000214f;
  t137_n242: return -0.030573207885026932f;
  t137_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t137_n245;
  t137_n244: return -0.03444811329245567f;
  t137_n245: return 0.010589491575956345f;
  t137_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t137_n248;
  t137_n247: return 0.002406321233138442f;
  t137_n248: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t137_n250;
  t137_n249: return 0.0064253732562065125f;
  t137_n250: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t137_n252;
  t137_n251: return -0.20898273587226868f;
  t137_n252: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t137_n254;
  t137_n253: return 0.1253359615802765f;
  t137_n254: return -0.17447154223918915f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree138(const Example& E) {
  t138_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t138_n40;
  t138_n1: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t138_n31;
  t138_n2: if (E.getNumReferences() >= 3350789120 /*94712.0*/) goto t138_n28;
  t138_n3: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t138_n27;
  t138_n4: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t138_n26;
  t138_n5: if (E.getNumReferences() >= 3289325568 /*572.5*/) goto t138_n25;
  t138_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t138_n8;
  t138_n7: return -0.00023076735669746995f;
  t138_n8: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t138_n22;
  t138_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n17;
  t138_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t138_n14;
  t138_n11: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t138_n13;
  t138_n12: return -0.046108145266771317f;
  t138_n13: return -0.012899668887257576f;
  t138_n14: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t138_n16;
  t138_n15: return 0.0029286008793860674f;
  t138_n16: return -0.026966288685798645f;
  t138_n17: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t138_n19;
  t138_n18: return 0.004517955239862204f;
  t138_n19: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t138_n21;
  t138_n20: return 0.02262958325445652f;
  t138_n21: return 0.13924053311347961f;
  t138_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t138_n24;
  t138_n23: return 0.006079592276364565f;
  t138_n24: return 0.07055094093084335f;
  t138_n25: return 0.010104534216225147f;
  t138_n26: return -0.001465342123992741f;
  t138_n27: return -0.06446468085050583f;
  t138_n28: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t138_n30;
  t138_n29: return -0.14219799637794495f;
  t138_n30: return -0.258483350276947f;
  t138_n31: if (E.getNumReferences() >= 3351669952 /*101593.5*/) goto t138_n37;
  t138_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t138_n36;
  t138_n33: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t138_n35;
  t138_n34: return 0.0882512629032135f;
  t138_n35: return 0.03290465101599693f;
  t138_n36: return 0.06227174401283264f;
  t138_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n39;
  t138_n38: return -0.029995478689670563f;
  t138_n39: return 0.006124514155089855f;
  t138_n40: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t138_n252;
  t138_n41: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t138_n65;
  t138_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n44;
  t138_n43: return -0.026877202093601227f;
  t138_n44: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t138_n58;
  t138_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n53;
  t138_n46: if (E.getNumReferences() >= 3308226560 /*2808.5*/) goto t138_n52;
  t138_n47: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t138_n51;
  t138_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t138_n50;
  t138_n49: return -0.018626980483531952f;
  t138_n50: return -0.0005071241757832468f;
  t138_n51: return -0.03050564043223858f;
  t138_n52: return 0.12154007703065872f;
  t138_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t138_n55;
  t138_n54: return -0.022729087620973587f;
  t138_n55: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t138_n57;
  t138_n56: return 0.0292589720338583f;
  t138_n57: return 0.12285906076431274f;
  t138_n58: if (E.getFractionNameInContext() >= 3197124868 /*0.2817460298538208*/) goto t138_n64;
  t138_n59: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t138_n61;
  t138_n60: return -0.007882929407060146f;
  t138_n61: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t138_n63;
  t138_n62: return 0.06904774159193039f;
  t138_n63: return -0.010031374171376228f;
  t138_n64: return 0.012740939855575562f;
  t138_n65: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t138_n247;
  t138_n66: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t138_n140;
  t138_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t138_n101;
  t138_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t138_n84;
  t138_n69: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t138_n77;
  t138_n70: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t138_n72;
  t138_n71: return -0.2350432425737381f;
  t138_n72: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t138_n74;
  t138_n73: return 0.11794918775558472f;
  t138_n74: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t138_n76;
  t138_n75: return -5.436509132385254f;
  t138_n76: return 0.014286871068179607f;
  t138_n77: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t138_n79;
  t138_n78: return 0.0029683923348784447f;
  t138_n79: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t138_n81;
  t138_n80: return -0.1615784615278244f;
  t138_n81: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t138_n83;
  t138_n82: return -0.6505182981491089f;
  t138_n83: return -0.0211399607360363f;
  t138_n84: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t138_n92;
  t138_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n89;
  t138_n86: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t138_n88;
  t138_n87: return -0.30879589915275574f;
  t138_n88: return -0.08436036854982376f;
  t138_n89: if (E.getNumReferences() >= 3315802112 /*5220.0*/) goto t138_n91;
  t138_n90: return -0.0028249421156942844f;
  t138_n91: return 0.04958905279636383f;
  t138_n92: if (E.getNumReferences() >= 3303993344 /*1911.5*/) goto t138_n100;
  t138_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t138_n95;
  t138_n94: return 0.006418746896088123f;
  t138_n95: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t138_n99;
  t138_n96: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t138_n98;
  t138_n97: return 0.16769254207611084f;
  t138_n98: return 0.06254786998033524f;
  t138_n99: return 0.02211977355182171f;
  t138_n100: return -0.12703926861286163f;
  t138_n101: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t138_n137;
  t138_n102: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t138_n104;
  t138_n103: return 0.31625476479530334f;
  t138_n104: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t138_n136;
  t138_n105: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t138_n131;
  t138_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t138_n108;
  t138_n107: return 0.02844737470149994f;
  t138_n108: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t138_n110;
  t138_n109: return 0.02076040767133236f;
  t138_n110: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t138_n130;
  t138_n111: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t138_n119;
  t138_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t138_n114;
  t138_n113: return -0.10571790486574173f;
  t138_n114: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t138_n118;
  t138_n115: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t138_n117;
  t138_n116: return 0.15663959085941315f;
  t138_n117: return 0.12967509031295776f;
  t138_n118: return 0.12289970368146896f;
  t138_n119: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t138_n129;
  t138_n120: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t138_n124;
  t138_n121: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t138_n123;
  t138_n122: return 0.017465168610215187f;
  t138_n123: return -0.44401416182518005f;
  t138_n124: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t138_n126;
  t138_n125: return 0.05755028501152992f;
  t138_n126: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t138_n128;
  t138_n127: return 0.18815037608146667f;
  t138_n128: return 0.13436342775821686f;
  t138_n129: return 0.11246451735496521f;
  t138_n130: return 0.15485771000385284f;
  t138_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t138_n135;
  t138_n132: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t138_n134;
  t138_n133: return 0.09721492230892181f;
  t138_n134: return -0.034711889922618866f;
  t138_n135: return 0.008682827465236187f;
  t138_n136: return 0.012583635747432709f;
  t138_n137: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t138_n139;
  t138_n138: return -0.02184257283806801f;
  t138_n139: return 0.010978012345731258f;
  t138_n140: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t138_n244;
  t138_n141: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t138_n211;
  t138_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t138_n182;
  t138_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t138_n181;
  t138_n144: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t138_n156;
  t138_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n147;
  t138_n146: return -0.06127467751502991f;
  t138_n147: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t138_n153;
  t138_n148: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t138_n150;
  t138_n149: return 0.006406267173588276f;
  t138_n150: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t138_n152;
  t138_n151: return 0.07092749327421188f;
  t138_n152: return -0.04635564610362053f;
  t138_n153: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t138_n155;
  t138_n154: return 0.12601983547210693f;
  t138_n155: return -0.14482198655605316f;
  t138_n156: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t138_n158;
  t138_n157: return 0.006710313260555267f;
  t138_n158: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t138_n176;
  t138_n159: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t138_n165;
  t138_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t138_n162;
  t138_n161: return -0.02601083368062973f;
  t138_n162: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t138_n164;
  t138_n163: return 0.0785675123333931f;
  t138_n164: return 0.3005591928958893f;
  t138_n165: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t138_n169;
  t138_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n168;
  t138_n167: return -0.16116905212402344f;
  t138_n168: return 0.143637016415596f;
  t138_n169: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t138_n173;
  t138_n170: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t138_n172;
  t138_n171: return -0.10508759319782257f;
  t138_n172: return 0.1548110991716385f;
  t138_n173: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t138_n175;
  t138_n174: return 0.14234095811843872f;
  t138_n175: return 0.11227946728467941f;
  t138_n176: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t138_n178;
  t138_n177: return 0.010482681915163994f;
  t138_n178: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t138_n180;
  t138_n179: return -0.15488237142562866f;
  t138_n180: return 0.10426025837659836f;
  t138_n181: return -0.08676311373710632f;
  t138_n182: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t138_n188;
  t138_n183: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t138_n185;
  t138_n184: return 0.03539048507809639f;
  t138_n185: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t138_n187;
  t138_n186: return 0.05102038010954857f;
  t138_n187: return 0.1315193474292755f;
  t138_n188: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t138_n194;
  t138_n189: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t138_n193;
  t138_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t138_n192;
  t138_n191: return 0.08938037604093552f;
  t138_n192: return 0.14039431512355804f;
  t138_n193: return 0.09925686568021774f;
  t138_n194: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t138_n208;
  t138_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t138_n205;
  t138_n196: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t138_n202;
  t138_n197: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t138_n199;
  t138_n198: return -0.0731213316321373f;
  t138_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t138_n201;
  t138_n200: return 0.06397344917058945f;
  t138_n201: return 0.3108110725879669f;
  t138_n202: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t138_n204;
  t138_n203: return 0.08037754148244858f;
  t138_n204: return 0.12304417788982391f;
  t138_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t138_n207;
  t138_n206: return 0.19625794887542725f;
  t138_n207: return 0.11858433485031128f;
  t138_n208: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t138_n210;
  t138_n209: return 0.07524441182613373f;
  t138_n210: return -0.042202338576316833f;
  t138_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t138_n239;
  t138_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t138_n214;
  t138_n213: return -0.11427459865808487f;
  t138_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t138_n232;
  t138_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t138_n223;
  t138_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n220;
  t138_n217: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t138_n219;
  t138_n218: return -0.17346659302711487f;
  t138_n219: return 0.05940373241901398f;
  t138_n220: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t138_n222;
  t138_n221: return 0.06719682365655899f;
  t138_n222: return -0.15873828530311584f;
  t138_n223: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t138_n231;
  t138_n224: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t138_n228;
  t138_n225: if (E.getNumReferences() >= 3281813504 /*313.0*/) goto t138_n227;
  t138_n226: return -0.016573602333664894f;
  t138_n227: return 0.1505288928747177f;
  t138_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol))) goto t138_n230;
  t138_n229: return 0.05905424803495407f;
  t138_n230: return 0.11893906444311142f;
  t138_n231: return 0.06158499792218208f;
  t138_n232: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t138_n234;
  t138_n233: return -0.03311637043952942f;
  t138_n234: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t138_n236;
  t138_n235: return 0.0655474066734314f;
  t138_n236: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t138_n238;
  t138_n237: return 0.12728698551654816f;
  t138_n238: return -0.06261385232210159f;
  t138_n239: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t138_n243;
  t138_n240: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t138_n242;
  t138_n241: return 0.04521150514483452f;
  t138_n242: return -0.19599920511245728f;
  t138_n243: return 0.11079079657793045f;
  t138_n244: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t138_n246;
  t138_n245: return -0.052702728658914566f;
  t138_n246: return 0.044176895171403885f;
  t138_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t138_n251;
  t138_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t138_n250;
  t138_n249: return -0.000732414482627064f;
  t138_n250: return 0.02987961657345295f;
  t138_n251: return 0.006258454639464617f;
  t138_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t138_n254;
  t138_n253: return -0.0689462348818779f;
  t138_n254: return -0.01884234882891178f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree139(const Example& E) {
  t139_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t139_n72;
  t139_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t139_n3;
  t139_n2: return -0.001000265241600573f;
  t139_n3: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t139_n69;
  t139_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t139_n32;
  t139_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t139_n31;
  t139_n6: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t139_n20;
  t139_n7: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t139_n11;
  t139_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t139_n10;
  t139_n9: return -0.0026326023507863283f;
  t139_n10: return 0.005214035976678133f;
  t139_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t139_n17;
  t139_n12: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t139_n16;
  t139_n13: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t139_n15;
  t139_n14: return -0.04412682726979256f;
  t139_n15: return 0.0808163434267044f;
  t139_n16: return -0.17513158917427063f;
  t139_n17: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t139_n19;
  t139_n18: return -0.17431491613388062f;
  t139_n19: return -0.0019559585489332676f;
  t139_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t139_n28;
  t139_n21: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t139_n23;
  t139_n22: return -0.0022951303981244564f;
  t139_n23: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t139_n25;
  t139_n24: return -0.049109719693660736f;
  t139_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t139_n27;
  t139_n26: return 0.05964742973446846f;
  t139_n27: return 0.10342270880937576f;
  t139_n28: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t139_n30;
  t139_n29: return 0.004376998636871576f;
  t139_n30: return 0.0182834193110466f;
  t139_n31: return -0.03970372676849365f;
  t139_n32: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t139_n40;
  t139_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t139_n39;
  t139_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t139_n38;
  t139_n35: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t139_n37;
  t139_n36: return -0.0856037363409996f;
  t139_n37: return -0.15104852616786957f;
  t139_n38: return -0.09884420037269592f;
  t139_n39: return 0.0015165304066613317f;
  t139_n40: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t139_n42;
  t139_n41: return 0.08371944725513458f;
  t139_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t139_n44;
  t139_n43: return 0.0010733278468251228f;
  t139_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t139_n46;
  t139_n45: return 0.03433609381318092f;
  t139_n46: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t139_n64;
  t139_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t139_n61;
  t139_n48: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t139_n52;
  t139_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t139_n51;
  t139_n50: return -0.10143958777189255f;
  t139_n51: return 0.07388443499803543f;
  t139_n52: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t139_n60;
  t139_n53: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t139_n55;
  t139_n54: return -0.013644049875438213f;
  t139_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t139_n59;
  t139_n56: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t139_n58;
  t139_n57: return 0.012916956096887589f;
  t139_n58: return -0.00851268321275711f;
  t139_n59: return 0.0048583876341581345f;
  t139_n60: return 0.08150480687618256f;
  t139_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t139_n63;
  t139_n62: return 0.013510202057659626f;
  t139_n63: return 0.027910621836781502f;
  t139_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t139_n66;
  t139_n65: return -0.02160157822072506f;
  t139_n66: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t139_n68;
  t139_n67: return 0.12495635449886322f;
  t139_n68: return 0.058875977993011475f;
  t139_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t139_n71;
  t139_n70: return 0.0033554963301867247f;
  t139_n71: return 0.013085946440696716f;
  t139_n72: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t139_n188;
  t139_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t139_n109;
  t139_n74: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t139_n104;
  t139_n75: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t139_n89;
  t139_n76: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t139_n78;
  t139_n77: return -0.07622195780277252f;
  t139_n78: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t139_n88;
  t139_n79: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t139_n85;
  t139_n80: if (E.getFileProximityDistanceCost() >= 3251634176 /*26.0*/) goto t139_n84;
  t139_n81: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t139_n83;
  t139_n82: return 0.05959423631429672f;
  t139_n83: return 0.13978974521160126f;
  t139_n84: return -0.04642820730805397f;
  t139_n85: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t139_n87;
  t139_n86: return 0.04480474069714546f;
  t139_n87: return 0.08963460475206375f;
  t139_n88: return -0.03164438530802727f;
  t139_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t139_n101;
  t139_n90: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t139_n92;
  t139_n91: return 0.0028861030004918575f;
  t139_n92: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t139_n98;
  t139_n93: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t139_n97;
  t139_n94: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t139_n96;
  t139_n95: return -0.010121136903762817f;
  t139_n96: return -0.04530204087495804f;
  t139_n97: return 0.0025529307313263416f;
  t139_n98: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t139_n100;
  t139_n99: return 0.09190256893634796f;
  t139_n100: return 0.03768543154001236f;
  t139_n101: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t139_n103;
  t139_n102: return 0.0232494305819273f;
  t139_n103: return 0.04779177904129028f;
  t139_n104: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t139_n106;
  t139_n105: return -0.01911839284002781f;
  t139_n106: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t139_n108;
  t139_n107: return 0.09627258777618408f;
  t139_n108: return -0.024494143202900887f;
  t139_n109: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t139_n185;
  t139_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t139_n144;
  t139_n111: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t139_n135;
  t139_n112: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t139_n120;
  t139_n113: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t139_n119;
  t139_n114: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t139_n116;
  t139_n115: return -0.08592511713504791f;
  t139_n116: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t139_n118;
  t139_n117: return 0.05620286613702774f;
  t139_n118: return -0.0921274796128273f;
  t139_n119: return 0.0092519735917449f;
  t139_n120: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t139_n134;
  t139_n121: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t139_n129;
  t139_n122: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t139_n128;
  t139_n123: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t139_n125;
  t139_n124: return 0.18496745824813843f;
  t139_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t139_n127;
  t139_n126: return 0.09676535427570343f;
  t139_n127: return -0.06547930836677551f;
  t139_n128: return -0.2038363367319107f;
  t139_n129: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t139_n131;
  t139_n130: return 0.08082938194274902f;
  t139_n131: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t139_n133;
  t139_n132: return 0.009092549793422222f;
  t139_n133: return 0.10088600218296051f;
  t139_n134: return 0.14180408418178558f;
  t139_n135: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t139_n137;
  t139_n136: return -0.015201829373836517f;
  t139_n137: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t139_n141;
  t139_n138: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t139_n140;
  t139_n139: return 0.01414478663355112f;
  t139_n140: return -1.019425392150879f;
  t139_n141: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t139_n143;
  t139_n142: return -0.1044415682554245f;
  t139_n143: return 0.11355728656053543f;
  t139_n144: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t139_n170;
  t139_n145: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t139_n147;
  t139_n146: return -0.0019975751638412476f;
  t139_n147: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t139_n167;
  t139_n148: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t139_n154;
  t139_n149: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t139_n151;
  t139_n150: return -0.04262319207191467f;
  t139_n151: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t139_n153;
  t139_n152: return 0.08218427747488022f;
  t139_n153: return 0.00444413349032402f;
  t139_n154: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t139_n166;
  t139_n155: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t139_n165;
  t139_n156: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t139_n162;
  t139_n157: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t139_n161;
  t139_n158: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t139_n160;
  t139_n159: return 0.020460963249206543f;
  t139_n160: return -0.014491213485598564f;
  t139_n161: return 0.03181585296988487f;
  t139_n162: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t139_n164;
  t139_n163: return -0.03169717267155647f;
  t139_n164: return 0.1082855612039566f;
  t139_n165: return 0.051561035215854645f;
  t139_n166: return -0.007970111444592476f;
  t139_n167: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t139_n169;
  t139_n168: return 0.09525588899850845f;
  t139_n169: return -0.09625297039747238f;
  t139_n170: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t139_n182;
  t139_n171: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t139_n173;
  t139_n172: return 0.029886361211538315f;
  t139_n173: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t139_n179;
  t139_n174: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t139_n176;
  t139_n175: return 0.13278372585773468f;
  t139_n176: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t139_n178;
  t139_n177: return 0.033732011914253235f;
  t139_n178: return -3.138275623321533f;
  t139_n179: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t139_n181;
  t139_n180: return -0.14243075251579285f;
  t139_n181: return -0.008894701488316059f;
  t139_n182: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t139_n184;
  t139_n183: return -0.011043772101402283f;
  t139_n184: return 0.036165982484817505f;
  t139_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t139_n187;
  t139_n186: return 0.008291048929095268f;
  t139_n187: return -0.06961002945899963f;
  t139_n188: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t139_n218;
  t139_n189: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t139_n209;
  t139_n190: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t139_n192;
  t139_n191: return 0.03294890746474266f;
  t139_n192: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t139_n208;
  t139_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t139_n195;
  t139_n194: return 0.04769042879343033f;
  t139_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t139_n199;
  t139_n196: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t139_n198;
  t139_n197: return -0.03264419361948967f;
  t139_n198: return 0.09207066893577576f;
  t139_n199: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t139_n207;
  t139_n200: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t139_n206;
  t139_n201: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t139_n205;
  t139_n202: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t139_n204;
  t139_n203: return 0.19493316113948822f;
  t139_n204: return 0.08745478838682175f;
  t139_n205: return 0.12752953171730042f;
  t139_n206: return 0.03747221454977989f;
  t139_n207: return -0.24180491268634796f;
  t139_n208: return 0.11970344185829163f;
  t139_n209: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t139_n211;
  t139_n210: return -0.07540217787027359f;
  t139_n211: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t139_n217;
  t139_n212: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t139_n214;
  t139_n213: return 0.04187074303627014f;
  t139_n214: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t139_n216;
  t139_n215: return 0.01816469058394432f;
  t139_n216: return 0.11363176256418228f;
  t139_n217: return -0.16415943205356598f;
  t139_n218: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t139_n252;
  t139_n219: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t139_n221;
  t139_n220: return 0.0038251630030572414f;
  t139_n221: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t139_n247;
  t139_n222: if (E.getNumReferences() >= 3297013760 /*1059.5*/) goto t139_n244;
  t139_n223: if (E.getNumReferences() >= 3273064448 /*151.0*/) goto t139_n235;
  t139_n224: if (E.getNumReferences() >= 3272179712 /*137.5*/) goto t139_n232;
  t139_n225: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t139_n227;
  t139_n226: return 0.06049615517258644f;
  t139_n227: if (E.getNumReferences() >= 3256745984 /*39.5*/) goto t139_n229;
  t139_n228: return -0.019171638414263725f;
  t139_n229: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t139_n231;
  t139_n230: return 0.14594367146492004f;
  t139_n231: return 0.05140113830566406f;
  t139_n232: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t139_n234;
  t139_n233: return -0.21225276589393616f;
  t139_n234: return -0.18139711022377014f;
  t139_n235: if (E.getNumReferences() >= 3273490432 /*157.5*/) goto t139_n237;
  t139_n236: return 0.18691031634807587f;
  t139_n237: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t139_n239;
  t139_n238: return -0.5400559306144714f;
  t139_n239: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t139_n243;
  t139_n240: if (E.getNumReferences() >= 3275259904 /*184.5*/) goto t139_n242;
  t139_n241: return 0.06053773686289787f;
  t139_n242: return 0.1370556801557541f;
  t139_n243: return 0.07610651850700378f;
  t139_n244: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t139_n246;
  t139_n245: return -0.33091920614242554f;
  t139_n246: return 0.06141047179698944f;
  t139_n247: if (E.getNumReferences() >= 3266641920 /*90.5*/) goto t139_n249;
  t139_n248: return -0.08843503147363663f;
  t139_n249: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t139_n251;
  t139_n250: return 0.12341978400945663f;
  t139_n251: return 0.026808157563209534f;
  t139_n252: if (E.getNumReferences() >= 3267100672 /*94.0*/) goto t139_n254;
  t139_n253: return -0.10019040107727051f;
  t139_n254: return 0.21912941336631775f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree140(const Example& E) {
  t140_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t140_n184;
  t140_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t140_n109;
  t140_n2: if (E.getNumReferences() >= 3306622976 /*2417.0*/) goto t140_n26;
  t140_n3: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t140_n9;
  t140_n4: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t140_n8;
  t140_n5: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t140_n7;
  t140_n6: return 0.12792585790157318f;
  t140_n7: return -0.0011056270450353622f;
  t140_n8: return -0.006262804381549358f;
  t140_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t140_n11;
  t140_n10: return -0.04613441228866577f;
  t140_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t140_n25;
  t140_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t140_n24;
  t140_n13: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t140_n19;
  t140_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t140_n16;
  t140_n15: return -0.003777633188292384f;
  t140_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t140_n18;
  t140_n17: return 0.002811406273394823f;
  t140_n18: return 0.014515137299895287f;
  t140_n19: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t140_n23;
  t140_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t140_n22;
  t140_n21: return -0.006406429223716259f;
  t140_n22: return 0.01093329954892397f;
  t140_n23: return 0.006620696280151606f;
  t140_n24: return 0.0036683068610727787f;
  t140_n25: return -0.00607148976996541f;
  t140_n26: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t140_n44;
  t140_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t140_n29;
  t140_n28: return -0.08871682733297348f;
  t140_n29: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t140_n31;
  t140_n30: return 0.028466634452342987f;
  t140_n31: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t140_n43;
  t140_n32: if (E.getNumReferences() >= 3331498752 /*18769.5*/) goto t140_n34;
  t140_n33: return 0.006123276427388191f;
  t140_n34: if (E.getNumReferences() >= 3332023808 /*19795.0*/) goto t140_n42;
  t140_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t140_n41;
  t140_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t140_n38;
  t140_n37: return -0.1112200990319252f;
  t140_n38: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t140_n40;
  t140_n39: return 0.09126196801662445f;
  t140_n40: return 0.14359325170516968f;
  t140_n41: return 0.11296894401311874f;
  t140_n42: return 0.016637912020087242f;
  t140_n43: return -0.004349765367805958f;
  t140_n44: if (E.getNumReferences() >= 3316280320 /*5453.5*/) goto t140_n102;
  t140_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t140_n47;
  t140_n46: return -0.007869117893278599f;
  t140_n47: if (E.getNumReferences() >= 3313587200 /*4138.5*/) goto t140_n57;
  t140_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t140_n50;
  t140_n49: return 0.01458230335265398f;
  t140_n50: if (E.getNumReferences() >= 3307108352 /*2535.5*/) goto t140_n54;
  t140_n51: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t140_n53;
  t140_n52: return -0.1004374623298645f;
  t140_n53: return 0.13940057158470154f;
  t140_n54: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t140_n56;
  t140_n55: return 0.21827292442321777f;
  t140_n56: return 0.027855057269334793f;
  t140_n57: if (E.getNumReferences() >= 3313608704 /*4149.0*/) goto t140_n61;
  t140_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t140_n60;
  t140_n59: return 0.13482347130775452f;
  t140_n60: return 0.16990762948989868f;
  t140_n61: if (E.getNumReferences() >= 3315954688 /*5294.5*/) goto t140_n91;
  t140_n62: if (E.getNumReferences() >= 3315423232 /*5035.0*/) goto t140_n90;
  t140_n63: if (E.getNumReferences() >= 3315402752 /*5025.0*/) goto t140_n89;
  t140_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t140_n66;
  t140_n65: return -0.0921175479888916f;
  t140_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t140_n68;
  t140_n67: return 0.06922157108783722f;
  t140_n68: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t140_n76;
  t140_n69: if (E.getNumReferences() >= 3313894400 /*4288.5*/) goto t140_n75;
  t140_n70: if (E.getNumReferences() >= 3313889280 /*4286.0*/) goto t140_n74;
  t140_n71: if (E.getNumReferences() >= 3313837056 /*4260.5*/) goto t140_n73;
  t140_n72: return -0.019754594191908836f;
  t140_n73: return 0.14415371417999268f;
  t140_n74: return 0.14192205667495728f;
  t140_n75: return 0.07355893403291702f;
  t140_n76: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t140_n84;
  t140_n77: if (E.getNumReferences() >= 3313659904 /*4174.0*/) goto t140_n79;
  t140_n78: return 0.12582279741764069f;
  t140_n79: if (E.getNumReferences() >= 3314128896 /*4403.0*/) goto t140_n81;
  t140_n80: return -0.12146233767271042f;
  t140_n81: if (E.getNumReferences() >= 3314141184 /*4409.0*/) goto t140_n83;
  t140_n82: return 0.23081611096858978f;
  t140_n83: return 0.09168330579996109f;
  t140_n84: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t140_n86;
  t140_n85: return -0.008629785850644112f;
  t140_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t140_n88;
  t140_n87: return -0.10087522119283676f;
  t140_n88: return 0.3826945126056671f;
  t140_n89: return 0.5713917016983032f;
  t140_n90: return -0.025919411331415176f;
  t140_n91: if (E.getNumReferences() >= 3316060160 /*5346.0*/) goto t140_n101;
  t140_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t140_n98;
  t140_n93: if (E.getNumReferences() >= 3316038656 /*5335.5*/) goto t140_n95;
  t140_n94: return -0.11706365644931793f;
  t140_n95: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t140_n97;
  t140_n96: return 0.13567768037319183f;
  t140_n97: return -0.10030588507652283f;
  t140_n98: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t140_n100;
  t140_n99: return 0.09633695334196091f;
  t140_n100: return 0.12714016437530518f;
  t140_n101: return 0.04840794578194618f;
  t140_n102: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t140_n106;
  t140_n103: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t140_n105;
  t140_n104: return -0.04850275442004204f;
  t140_n105: return 0.0810413509607315f;
  t140_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t140_n108;
  t140_n107: return -0.0009132958948612213f;
  t140_n108: return 0.017551440745592117f;
  t140_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t140_n123;
  t140_n110: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t140_n120;
  t140_n111: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t140_n119;
  t140_n112: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t140_n114;
  t140_n113: return -0.004281728062778711f;
  t140_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t140_n116;
  t140_n115: return 0.028319839388132095f;
  t140_n116: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t140_n118;
  t140_n117: return 0.15326933562755585f;
  t140_n118: return 0.06070902198553085f;
  t140_n119: return -0.0013197107473388314f;
  t140_n120: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t140_n122;
  t140_n121: return 0.006797345355153084f;
  t140_n122: return 0.09213710576295853f;
  t140_n123: if (E.getFractionNameInContext() >= 3204006750 /*0.4868420958518982*/) goto t140_n125;
  t140_n124: return 0.0021573270205408335f;
  t140_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t140_n127;
  t140_n126: return 0.004863560199737549f;
  t140_n127: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t140_n129;
  t140_n128: return 0.006261396687477827f;
  t140_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t140_n163;
  t140_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t140_n132;
  t140_n131: return -0.0126012172549963f;
  t140_n132: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t140_n152;
  t140_n133: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t140_n137;
  t140_n134: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t140_n136;
  t140_n135: return -0.012517422437667847f;
  t140_n136: return -0.23898175358772278f;
  t140_n137: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t140_n149;
  t140_n138: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t140_n148;
  t140_n139: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t140_n147;
  t140_n140: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t140_n146;
  t140_n141: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t140_n145;
  t140_n142: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t140_n144;
  t140_n143: return 0.13670116662979126f;
  t140_n144: return 0.31703823804855347f;
  t140_n145: return 0.08313687145709991f;
  t140_n146: return 0.11612428724765778f;
  t140_n147: return 0.07082173973321915f;
  t140_n148: return -0.16712616384029388f;
  t140_n149: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t140_n151;
  t140_n150: return 0.03591202199459076f;
  t140_n151: return 0.13052232563495636f;
  t140_n152: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t140_n160;
  t140_n153: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t140_n155;
  t140_n154: return -0.012935427017509937f;
  t140_n155: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t140_n157;
  t140_n156: return -0.05489085242152214f;
  t140_n157: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t140_n159;
  t140_n158: return 0.10257194936275482f;
  t140_n159: return 0.10693827271461487f;
  t140_n160: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t140_n162;
  t140_n161: return 0.12134808301925659f;
  t140_n162: return 0.07787971943616867f;
  t140_n163: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t140_n165;
  t140_n164: return 0.029705043882131577f;
  t140_n165: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t140_n175;
  t140_n166: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t140_n172;
  t140_n167: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t140_n169;
  t140_n168: return -0.08905252814292908f;
  t140_n169: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t140_n171;
  t140_n170: return 0.5364012122154236f;
  t140_n171: return -0.10822359472513199f;
  t140_n172: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t140_n174;
  t140_n173: return 0.1475527137517929f;
  t140_n174: return -0.27337750792503357f;
  t140_n175: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t140_n183;
  t140_n176: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t140_n182;
  t140_n177: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t140_n179;
  t140_n178: return 0.06356314569711685f;
  t140_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t140_n181;
  t140_n180: return 0.1373838186264038f;
  t140_n181: return 0.1665639728307724f;
  t140_n182: return 0.00411083921790123f;
  t140_n183: return 0.12383914738893509f;
  t140_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t140_n204;
  t140_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t140_n197;
  t140_n186: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t140_n196;
  t140_n187: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t140_n195;
  t140_n188: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t140_n190;
  t140_n189: return -0.0017210900550708175f;
  t140_n190: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t140_n192;
  t140_n191: return -0.013438774272799492f;
  t140_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t140_n194;
  t140_n193: return -0.1435384303331375f;
  t140_n194: return 0.10019761323928833f;
  t140_n195: return -0.06402675062417984f;
  t140_n196: return 0.041754961013793945f;
  t140_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t140_n199;
  t140_n198: return -0.03504322096705437f;
  t140_n199: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t140_n201;
  t140_n200: return 0.013315213844180107f;
  t140_n201: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t140_n203;
  t140_n202: return 0.04705152288079262f;
  t140_n203: return 0.0013905902160331607f;
  t140_n204: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t140_n252;
  t140_n205: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t140_n239;
  t140_n206: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t140_n238;
  t140_n207: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t140_n231;
  t140_n208: if (E.getNumReferences() >= 3323748864 /*10008.5*/) goto t140_n230;
  t140_n209: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t140_n215;
  t140_n210: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t140_n212;
  t140_n211: return 0.005793331656605005f;
  t140_n212: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t140_n214;
  t140_n213: return -0.25150495767593384f;
  t140_n214: return -0.02181182987987995f;
  t140_n215: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t140_n223;
  t140_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t140_n218;
  t140_n217: return 0.1096448227763176f;
  t140_n218: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t140_n222;
  t140_n219: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t140_n221;
  t140_n220: return 0.10834860801696777f;
  t140_n221: return 0.1048721969127655f;
  t140_n222: return 0.1078253909945488f;
  t140_n223: if (E.getNumReferences() >= 3321964544 /*8266.0*/) goto t140_n225;
  t140_n224: return -0.06443855911493301f;
  t140_n225: if (E.getNumReferences() >= 3322671104 /*8956.0*/) goto t140_n227;
  t140_n226: return 0.09438508749008179f;
  t140_n227: if (E.getNumReferences() >= 3323247616 /*9519.0*/) goto t140_n229;
  t140_n228: return -0.09768746048212051f;
  t140_n229: return 0.05104856565594673f;
  t140_n230: return -0.037455908954143524f;
  t140_n231: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t140_n233;
  t140_n232: return -0.0062326532788574696f;
  t140_n233: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t140_n237;
  t140_n234: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t140_n236;
  t140_n235: return 0.06843973696231842f;
  t140_n236: return 0.1583748161792755f;
  t140_n237: return -0.043151579797267914f;
  t140_n238: return 0.010101201012730598f;
  t140_n239: if (E.getNumReferences() >= 3338172416 /*31804.0*/) goto t140_n241;
  t140_n240: return 0.07966428250074387f;
  t140_n241: if (E.getNumReferences() >= 3341699072 /*44616.0*/) goto t140_n245;
  t140_n242: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t140_n244;
  t140_n243: return 0.03497028723359108f;
  t140_n244: return -0.06831853836774826f;
  t140_n245: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t140_n247;
  t140_n246: return 0.07379161566495895f;
  t140_n247: if (E.getNumReferences() >= 3363497856 /*256926.0*/) goto t140_n251;
  t140_n248: if (E.getNumReferences() >= 3348712704 /*78490.0*/) goto t140_n250;
  t140_n249: return -0.03778917342424393f;
  t140_n250: return 0.05219210684299469f;
  t140_n251: return -0.32495638728141785f;
  t140_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t140_n254;
  t140_n253: return 0.05676625669002533f;
  t140_n254: return 0.02679959312081337f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree141(const Example& E) {
  t141_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t141_n30;
  t141_n1: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t141_n29;
  t141_n2: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t141_n4;
  t141_n3: return -0.000963354017585516f;
  t141_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t141_n12;
  t141_n5: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t141_n7;
  t141_n6: return -0.003060924122110009f;
  t141_n7: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t141_n11;
  t141_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t141_n10;
  t141_n9: return -0.09193436801433563f;
  t141_n10: return 0.12329122424125671f;
  t141_n11: return 0.05643361061811447f;
  t141_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t141_n14;
  t141_n13: return 0.0085002975538373f;
  t141_n14: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t141_n28;
  t141_n15: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t141_n19;
  t141_n16: if (E.getFractionNameInContext() >= 3209574628 /*0.8055555820465088*/) goto t141_n18;
  t141_n17: return 0.054487936198711395f;
  t141_n18: return -0.13406065106391907f;
  t141_n19: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t141_n23;
  t141_n20: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t141_n22;
  t141_n21: return 0.12224147468805313f;
  t141_n22: return 0.020499004051089287f;
  t141_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t141_n27;
  t141_n24: if (E.getFractionNameInContext() >= 3208152317 /*0.7207792401313782*/) goto t141_n26;
  t141_n25: return 0.013885103166103363f;
  t141_n26: return 0.1375766545534134f;
  t141_n27: return 0.12800541520118713f;
  t141_n28: return -0.016789229586720467f;
  t141_n29: return 0.024061866104602814f;
  t141_n30: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t141_n172;
  t141_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t141_n137;
  t141_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n136;
  t141_n33: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t141_n131;
  t141_n34: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t141_n112;
  t141_n35: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t141_n111;
  t141_n36: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t141_n68;
  t141_n37: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t141_n65;
  t141_n38: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t141_n62;
  t141_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t141_n55;
  t141_n40: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t141_n46;
  t141_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n43;
  t141_n42: return -0.037788160145282745f;
  t141_n43: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t141_n45;
  t141_n44: return -0.02987137995660305f;
  t141_n45: return 0.025927264243364334f;
  t141_n46: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t141_n54;
  t141_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t141_n51;
  t141_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t141_n50;
  t141_n49: return 0.026778094470500946f;
  t141_n50: return -0.12607605755329132f;
  t141_n51: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t141_n53;
  t141_n52: return -0.2445608228445053f;
  t141_n53: return 0.2090713381767273f;
  t141_n54: return -0.014511732384562492f;
  t141_n55: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t141_n61;
  t141_n56: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t141_n60;
  t141_n57: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t141_n59;
  t141_n58: return 0.019351603463292122f;
  t141_n59: return -0.023274097591638565f;
  t141_n60: return 0.044225599616765976f;
  t141_n61: return -0.014792616479098797f;
  t141_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t141_n64;
  t141_n63: return 0.031673628836870193f;
  t141_n64: return 0.12174052745103836f;
  t141_n65: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t141_n67;
  t141_n66: return -0.06551635265350342f;
  t141_n67: return 0.0743539109826088f;
  t141_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t141_n104;
  t141_n69: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t141_n79;
  t141_n70: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t141_n74;
  t141_n71: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t141_n73;
  t141_n72: return 0.00307001662440598f;
  t141_n73: return -0.07858976721763611f;
  t141_n74: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t141_n76;
  t141_n75: return -0.1367419809103012f;
  t141_n76: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t141_n78;
  t141_n77: return -0.009246227331459522f;
  t141_n78: return 0.12744635343551636f;
  t141_n79: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t141_n101;
  t141_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t141_n100;
  t141_n81: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t141_n93;
  t141_n82: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t141_n84;
  t141_n83: return -0.04750727489590645f;
  t141_n84: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t141_n92;
  t141_n85: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t141_n87;
  t141_n86: return -0.022232338786125183f;
  t141_n87: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n91;
  t141_n88: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t141_n90;
  t141_n89: return -0.7838250398635864f;
  t141_n90: return -0.25857576727867126f;
  t141_n91: return 0.05914948508143425f;
  t141_n92: return 0.08716340363025665f;
  t141_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t141_n99;
  t141_n94: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t141_n96;
  t141_n95: return -0.00956329982727766f;
  t141_n96: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t141_n98;
  t141_n97: return -0.6233419179916382f;
  t141_n98: return -0.3601115643978119f;
  t141_n99: return 0.001456090365536511f;
  t141_n100: return -0.0186707004904747f;
  t141_n101: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t141_n103;
  t141_n102: return -0.00836645346134901f;
  t141_n103: return 0.05574364587664604f;
  t141_n104: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t141_n106;
  t141_n105: return -0.12062415480613708f;
  t141_n106: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t141_n108;
  t141_n107: return -0.01416418794542551f;
  t141_n108: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t141_n110;
  t141_n109: return 0.04197140783071518f;
  t141_n110: return -0.14158515632152557f;
  t141_n111: return 0.0012983463238924742f;
  t141_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t141_n128;
  t141_n113: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t141_n115;
  t141_n114: return 0.03374820575118065f;
  t141_n115: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t141_n117;
  t141_n116: return -0.0554649755358696f;
  t141_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t141_n123;
  t141_n118: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t141_n122;
  t141_n119: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t141_n121;
  t141_n120: return -0.04277658089995384f;
  t141_n121: return 0.016926012933254242f;
  t141_n122: return -2.305982828140259f;
  t141_n123: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t141_n125;
  t141_n124: return 0.02897232584655285f;
  t141_n125: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t141_n127;
  t141_n126: return 0.06092006713151932f;
  t141_n127: return -0.4019031524658203f;
  t141_n128: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t141_n130;
  t141_n129: return 0.061329495161771774f;
  t141_n130: return 0.13117322325706482f;
  t141_n131: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t141_n133;
  t141_n132: return 0.023729827255010605f;
  t141_n133: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t141_n135;
  t141_n134: return -0.029594814404845238f;
  t141_n135: return 0.009738091379404068f;
  t141_n136: return 0.0052124266512691975f;
  t141_n137: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t141_n143;
  t141_n138: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t141_n140;
  t141_n139: return 0.001744244946166873f;
  t141_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n142;
  t141_n141: return -0.38768503069877625f;
  t141_n142: return 0.05644087493419647f;
  t141_n143: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t141_n163;
  t141_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t141_n160;
  t141_n145: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t141_n157;
  t141_n146: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t141_n152;
  t141_n147: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t141_n149;
  t141_n148: return 0.02066691406071186f;
  t141_n149: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t141_n151;
  t141_n150: return 0.13859061896800995f;
  t141_n151: return -0.0002584746398497373f;
  t141_n152: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t141_n154;
  t141_n153: return 0.06620695441961288f;
  t141_n154: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t141_n156;
  t141_n155: return 0.08355496823787689f;
  t141_n156: return 0.016973011195659637f;
  t141_n157: if (E.getNumReferences() >= 3315734528 /*5187.0*/) goto t141_n159;
  t141_n158: return -0.021764885634183884f;
  t141_n159: return 0.06844186037778854f;
  t141_n160: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t141_n162;
  t141_n161: return 0.045659612864255905f;
  t141_n162: return 0.45518791675567627f;
  t141_n163: if (E.getFractionNameInContext() >= 3173614803 /*0.0413864366710186*/) goto t141_n169;
  t141_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t141_n168;
  t141_n165: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t141_n167;
  t141_n166: return -0.030413903295993805f;
  t141_n167: return 0.07693253457546234f;
  t141_n168: return 0.03731456771492958f;
  t141_n169: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t141_n171;
  t141_n170: return 0.03648601472377777f;
  t141_n171: return -0.041189126670360565f;
  t141_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t141_n178;
  t141_n173: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t141_n175;
  t141_n174: return -0.071306973695755f;
  t141_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n177;
  t141_n176: return -0.0001210196569445543f;
  t141_n177: return 0.016787046566605568f;
  t141_n178: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t141_n240;
  t141_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n181;
  t141_n180: return -0.006049427669495344f;
  t141_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t141_n239;
  t141_n182: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t141_n234;
  t141_n183: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t141_n219;
  t141_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n192;
  t141_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t141_n189;
  t141_n186: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t141_n188;
  t141_n187: return -0.02784167416393757f;
  t141_n188: return 0.042368434369564056f;
  t141_n189: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t141_n191;
  t141_n190: return -0.0011592177907004952f;
  t141_n191: return -0.06491433084011078f;
  t141_n192: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t141_n218;
  t141_n193: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t141_n217;
  t141_n194: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t141_n202;
  t141_n195: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t141_n201;
  t141_n196: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t141_n200;
  t141_n197: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n199;
  t141_n198: return 0.006853026803582907f;
  t141_n199: return 0.10716832429170609f;
  t141_n200: return 0.0004421839548740536f;
  t141_n201: return -0.15352505445480347f;
  t141_n202: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t141_n204;
  t141_n203: return -0.08949687331914902f;
  t141_n204: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t141_n212;
  t141_n205: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n211;
  t141_n206: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t141_n208;
  t141_n207: return 0.04841115325689316f;
  t141_n208: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t141_n210;
  t141_n209: return 0.08946053683757782f;
  t141_n210: return 0.07733814418315887f;
  t141_n211: return 0.03428886830806732f;
  t141_n212: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t141_n216;
  t141_n213: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t141_n215;
  t141_n214: return -0.016851868480443954f;
  t141_n215: return 0.0358511283993721f;
  t141_n216: return -0.0005216404679231346f;
  t141_n217: return -0.12423336505889893f;
  t141_n218: return 0.13028587400913239f;
  t141_n219: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t141_n227;
  t141_n220: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n224;
  t141_n221: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t141_n223;
  t141_n222: return -0.12865674495697021f;
  t141_n223: return 0.07929416745901108f;
  t141_n224: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t141_n226;
  t141_n225: return -0.09517481923103333f;
  t141_n226: return 0.1286287158727646f;
  t141_n227: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t141_n229;
  t141_n228: return 0.010528373531997204f;
  t141_n229: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n233;
  t141_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n232;
  t141_n231: return -0.8532565832138062f;
  t141_n232: return 0.049674250185489655f;
  t141_n233: return -0.0230389516800642f;
  t141_n234: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t141_n238;
  t141_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n237;
  t141_n236: return -0.06789752840995789f;
  t141_n237: return 0.006707730703055859f;
  t141_n238: return -0.03602274879813194f;
  t141_n239: return 0.010858174413442612f;
  t141_n240: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t141_n246;
  t141_n241: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t141_n243;
  t141_n242: return 0.0056824227795004845f;
  t141_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t141_n245;
  t141_n244: return 0.010000930167734623f;
  t141_n245: return 0.030873432755470276f;
  t141_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n248;
  t141_n247: return -0.4615822732448578f;
  t141_n248: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t141_n250;
  t141_n249: return -0.09872683137655258f;
  t141_n250: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t141_n254;
  t141_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t141_n253;
  t141_n252: return -0.06969372928142548f;
  t141_n253: return 0.17390114068984985f;
  t141_n254: return 0.1990634649991989f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree142(const Example& E) {
  t142_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t142_n182;
  t142_n1: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t142_n117;
  t142_n2: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t142_n76;
  t142_n3: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t142_n35;
  t142_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t142_n6;
  t142_n5: return -0.004003587178885937f;
  t142_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t142_n10;
  t142_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n9;
  t142_n8: return -0.07690492272377014f;
  t142_n9: return -0.002329412382096052f;
  t142_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t142_n32;
  t142_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n13;
  t142_n12: return -0.03550126776099205f;
  t142_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t142_n15;
  t142_n14: return 0.0031092825811356306f;
  t142_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t142_n17;
  t142_n16: return 0.0065895067527890205f;
  t142_n17: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t142_n29;
  t142_n18: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t142_n20;
  t142_n19: return 0.011856216937303543f;
  t142_n20: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t142_n22;
  t142_n21: return 0.06411807239055634f;
  t142_n22: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t142_n26;
  t142_n23: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t142_n25;
  t142_n24: return -0.01262547355145216f;
  t142_n25: return 0.035609833896160126f;
  t142_n26: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t142_n28;
  t142_n27: return 0.1903916895389557f;
  t142_n28: return -0.05432111769914627f;
  t142_n29: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t142_n31;
  t142_n30: return 0.18521656095981598f;
  t142_n31: return -0.12901490926742554f;
  t142_n32: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t142_n34;
  t142_n33: return -1.882174728962127e-05f;
  t142_n34: return -0.13830742239952087f;
  t142_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n63;
  t142_n36: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t142_n56;
  t142_n37: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t142_n55;
  t142_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t142_n42;
  t142_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t142_n41;
  t142_n40: return -0.08847583830356598f;
  t142_n41: return -0.03112710826098919f;
  t142_n42: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t142_n54;
  t142_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n45;
  t142_n44: return -0.07082828879356384f;
  t142_n45: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t142_n51;
  t142_n46: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t142_n50;
  t142_n47: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t142_n49;
  t142_n48: return 0.12620006501674652f;
  t142_n49: return 0.14181208610534668f;
  t142_n50: return -0.012003724463284016f;
  t142_n51: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t142_n53;
  t142_n52: return 0.0410507395863533f;
  t142_n53: return 0.0020289719104766846f;
  t142_n54: return -0.051099993288517f;
  t142_n55: return -0.005255431868135929f;
  t142_n56: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t142_n60;
  t142_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n59;
  t142_n58: return -0.13901470601558685f;
  t142_n59: return 0.06753375381231308f;
  t142_n60: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t142_n62;
  t142_n61: return 0.1268429309129715f;
  t142_n62: return -0.008216533809900284f;
  t142_n63: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t142_n65;
  t142_n64: return 0.021218188107013702f;
  t142_n65: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t142_n67;
  t142_n66: return -0.030446331948041916f;
  t142_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t142_n69;
  t142_n68: return -0.008742094971239567f;
  t142_n69: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t142_n73;
  t142_n70: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t142_n72;
  t142_n71: return 0.024462776258587837f;
  t142_n72: return -0.012587741948664188f;
  t142_n73: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t142_n75;
  t142_n74: return -0.0065872156992554665f;
  t142_n75: return 0.05282379686832428f;
  t142_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t142_n108;
  t142_n77: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t142_n105;
  t142_n78: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t142_n104;
  t142_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Unknown))) goto t142_n87;
  t142_n80: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t142_n82;
  t142_n81: return 0.011916355229914188f;
  t142_n82: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t142_n84;
  t142_n83: return -0.043562449514865875f;
  t142_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n86;
  t142_n85: return -0.030420050024986267f;
  t142_n86: return -0.0024446938186883926f;
  t142_n87: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t142_n89;
  t142_n88: return 0.0009514577104710042f;
  t142_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t142_n91;
  t142_n90: return 0.0031758337281644344f;
  t142_n91: if (E.getNumReferences() >= 3305209856 /*2072.0*/) goto t142_n103;
  t142_n92: if (E.getNumReferences() >= 3281420288 /*301.0*/) goto t142_n100;
  t142_n93: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t142_n95;
  t142_n94: return 0.013352543115615845f;
  t142_n95: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t142_n99;
  t142_n96: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t142_n98;
  t142_n97: return 0.5213337540626526f;
  t142_n98: return -0.21362696588039398f;
  t142_n99: return 0.132751002907753f;
  t142_n100: if (E.getNumReferences() >= 3282468864 /*333.0*/) goto t142_n102;
  t142_n101: return 0.2075267881155014f;
  t142_n102: return 0.13355639576911926f;
  t142_n103: return -0.12897932529449463f;
  t142_n104: return 0.004013698548078537f;
  t142_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t142_n107;
  t142_n106: return -0.04626351594924927f;
  t142_n107: return -0.008927793242037296f;
  t142_n108: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t142_n110;
  t142_n109: return 0.008009660989046097f;
  t142_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t142_n114;
  t142_n111: if (E.getNumReferences() >= 3293667328 /*837.5*/) goto t142_n113;
  t142_n112: return 0.02200702577829361f;
  t142_n113: return 0.09827306866645813f;
  t142_n114: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t142_n116;
  t142_n115: return 0.12114188075065613f;
  t142_n116: return 0.09777884930372238f;
  t142_n117: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t142_n119;
  t142_n118: return 0.004122056066989899f;
  t142_n119: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t142_n121;
  t142_n120: return 0.0005839176592417061f;
  t142_n121: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t142_n179;
  t142_n122: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t142_n132;
  t142_n123: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t142_n131;
  t142_n124: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t142_n126;
  t142_n125: return 0.0057124122977256775f;
  t142_n126: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t142_n128;
  t142_n127: return 0.019424540922045708f;
  t142_n128: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t142_n130;
  t142_n129: return 0.10922787338495255f;
  t142_n130: return 0.12740308046340942f;
  t142_n131: return 0.04273701459169388f;
  t142_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t142_n134;
  t142_n133: return -0.005756421014666557f;
  t142_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t142_n176;
  t142_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t142_n145;
  t142_n136: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t142_n142;
  t142_n137: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t142_n141;
  t142_n138: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t142_n140;
  t142_n139: return 0.09529896825551987f;
  t142_n140: return 0.16440436244010925f;
  t142_n141: return -0.00932131428271532f;
  t142_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n144;
  t142_n143: return -0.13658902049064636f;
  t142_n144: return 0.08646457642316818f;
  t142_n145: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t142_n147;
  t142_n146: return -0.018136993050575256f;
  t142_n147: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t142_n175;
  t142_n148: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t142_n174;
  t142_n149: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t142_n155;
  t142_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t142_n152;
  t142_n151: return -0.011772498488426208f;
  t142_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n154;
  t142_n153: return -0.05370846763253212f;
  t142_n154: return 0.20660525560379028f;
  t142_n155: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t142_n161;
  t142_n156: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t142_n160;
  t142_n157: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t142_n159;
  t142_n158: return 0.1400085985660553f;
  t142_n159: return 0.1032269299030304f;
  t142_n160: return 0.020010972395539284f;
  t142_n161: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t142_n173;
  t142_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t142_n172;
  t142_n163: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t142_n171;
  t142_n164: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t142_n166;
  t142_n165: return 0.05795374885201454f;
  t142_n166: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t142_n170;
  t142_n167: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t142_n169;
  t142_n168: return 0.14847853779792786f;
  t142_n169: return 0.14597585797309875f;
  t142_n170: return 0.09901244193315506f;
  t142_n171: return -0.15521758794784546f;
  t142_n172: return 0.09567059576511383f;
  t142_n173: return -0.036835748702287674f;
  t142_n174: return 0.13480369746685028f;
  t142_n175: return 0.008841034956276417f;
  t142_n176: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t142_n178;
  t142_n177: return 0.01099391933530569f;
  t142_n178: return 0.08264424651861191f;
  t142_n179: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t142_n181;
  t142_n180: return 0.008874772116541862f;
  t142_n181: return -0.0918494462966919f;
  t142_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t142_n198;
  t142_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n195;
  t142_n184: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t142_n192;
  t142_n185: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t142_n191;
  t142_n186: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t142_n188;
  t142_n187: return -0.057503439486026764f;
  t142_n188: if (E.getNumReferences() >= 3342258560 /*46801.5*/) goto t142_n190;
  t142_n189: return 5.087775207357481e-05f;
  t142_n190: return 0.08428759127855301f;
  t142_n191: return -0.1268555223941803f;
  t142_n192: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t142_n194;
  t142_n193: return 0.08886577934026718f;
  t142_n194: return 0.019340068101882935f;
  t142_n195: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t142_n197;
  t142_n196: return 0.0213353019207716f;
  t142_n197: return -0.12703688442707062f;
  t142_n198: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t142_n222;
  t142_n199: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t142_n201;
  t142_n200: return -0.006881034933030605f;
  t142_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t142_n221;
  t142_n202: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t142_n216;
  t142_n203: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t142_n205;
  t142_n204: return -0.031520526856184006f;
  t142_n205: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t142_n215;
  t142_n206: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t142_n214;
  t142_n207: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t142_n211;
  t142_n208: if (E.getNumReferences() >= 3288023040 /*502.5*/) goto t142_n210;
  t142_n209: return 0.041958317160606384f;
  t142_n210: return -1.574643850326538f;
  t142_n211: if (E.getNumReferences() >= 3273654272 /*160.0*/) goto t142_n213;
  t142_n212: return 0.07603171467781067f;
  t142_n213: return 0.11369326710700989f;
  t142_n214: return 0.01396672148257494f;
  t142_n215: return -0.0007734504761174321f;
  t142_n216: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t142_n218;
  t142_n217: return 0.12719184160232544f;
  t142_n218: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t142_n220;
  t142_n219: return -0.04754770174622536f;
  t142_n220: return -0.21872912347316742f;
  t142_n221: return 0.004527879413217306f;
  t142_n222: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t142_n250;
  t142_n223: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t142_n225;
  t142_n224: return 0.0074455211870372295f;
  t142_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t142_n245;
  t142_n226: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t142_n242;
  t142_n227: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t142_n229;
  t142_n228: return -0.008408273570239544f;
  t142_n229: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t142_n241;
  t142_n230: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t142_n240;
  t142_n231: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t142_n239;
  t142_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t142_n238;
  t142_n233: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t142_n235;
  t142_n234: return 0.029676493257284164f;
  t142_n235: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t142_n237;
  t142_n236: return 0.12114787846803665f;
  t142_n237: return -0.006994599010795355f;
  t142_n238: return 0.12900835275650024f;
  t142_n239: return 0.07420612871646881f;
  t142_n240: return -0.03230193629860878f;
  t142_n241: return -0.09871806204319f;
  t142_n242: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t142_n244;
  t142_n243: return -0.19752386212348938f;
  t142_n244: return 0.1398364156484604f;
  t142_n245: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t142_n249;
  t142_n246: if (E.getFractionNameInContext() >= 3196938455 /*0.27619048953056335*/) goto t142_n248;
  t142_n247: return 0.08245448023080826f;
  t142_n248: return 0.0467301644384861f;
  t142_n249: return -0.11215700954198837f;
  t142_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t142_n252;
  t142_n251: return -0.019701818004250526f;
  t142_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t142_n254;
  t142_n253: return 0.0007020220509730279f;
  t142_n254: return 0.0365595817565918f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree143(const Example& E) {
  t143_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t143_n32;
  t143_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t143_n31;
  t143_n2: if (E.getNumReferences() >= 3289325568 /*572.5*/) goto t143_n18;
  t143_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t143_n5;
  t143_n4: return -0.0003055269771721214f;
  t143_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t143_n15;
  t143_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n8;
  t143_n7: return -0.09539466351270676f;
  t143_n8: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t143_n10;
  t143_n9: return 0.002458833158016205f;
  t143_n10: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t143_n14;
  t143_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n13;
  t143_n12: return -0.12204346805810928f;
  t143_n13: return 0.022185808047652245f;
  t143_n14: return 0.13472315669059753f;
  t143_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t143_n17;
  t143_n16: return 0.005467576440423727f;
  t143_n17: return 0.06917224824428558f;
  t143_n18: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t143_n30;
  t143_n19: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t143_n25;
  t143_n20: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t143_n24;
  t143_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n23;
  t143_n22: return -0.04013590142130852f;
  t143_n23: return 0.011347491294145584f;
  t143_n24: return 0.031490202993154526f;
  t143_n25: if (E.getNumReferences() >= 3362860128 /*246961.5*/) goto t143_n27;
  t143_n26: return 0.008079431019723415f;
  t143_n27: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t143_n29;
  t143_n28: return -0.071917325258255f;
  t143_n29: return 0.02363217994570732f;
  t143_n30: return 0.07263598591089249f;
  t143_n31: return -0.0015772880287840962f;
  t143_n32: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t143_n250;
  t143_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t143_n159;
  t143_n34: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t143_n142;
  t143_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t143_n93;
  t143_n36: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t143_n66;
  t143_n37: if (E.getNumReferences() >= 3352681216 /*109494.0*/) goto t143_n65;
  t143_n38: if (E.getNumReferences() >= 3349877760 /*87592.0*/) goto t143_n64;
  t143_n39: if (E.getNumReferences() >= 3346296960 /*62576.5*/) goto t143_n63;
  t143_n40: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t143_n60;
  t143_n41: if (E.getNumReferences() >= 3324916736 /*11149.0*/) goto t143_n55;
  t143_n42: if (E.getNumReferences() >= 3315782656 /*5210.5*/) goto t143_n44;
  t143_n43: return -0.0036856664810329676f;
  t143_n44: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t143_n48;
  t143_n45: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t143_n47;
  t143_n46: return 0.10166192799806595f;
  t143_n47: return 0.011896884068846703f;
  t143_n48: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t143_n54;
  t143_n49: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t143_n53;
  t143_n50: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t143_n52;
  t143_n51: return 0.20376785099506378f;
  t143_n52: return 0.10610590130090714f;
  t143_n53: return 0.12002640962600708f;
  t143_n54: return 0.04635327309370041f;
  t143_n55: if (E.getNumReferences() >= 3325593088 /*11809.5*/) goto t143_n57;
  t143_n56: return -0.2769474983215332f;
  t143_n57: if (E.getNumReferences() >= 3326107136 /*12311.5*/) goto t143_n59;
  t143_n58: return 0.15188761055469513f;
  t143_n59: return -0.0797615647315979f;
  t143_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t143_n62;
  t143_n61: return 0.04063396528363228f;
  t143_n62: return 0.1065555214881897f;
  t143_n63: return -0.22593332827091217f;
  t143_n64: return 0.10736199468374252f;
  t143_n65: return -0.24777601659297943f;
  t143_n66: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t143_n92;
  t143_n67: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t143_n89;
  t143_n68: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t143_n70;
  t143_n69: return -0.018056031316518784f;
  t143_n70: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t143_n88;
  t143_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t143_n73;
  t143_n72: return 0.02112584002315998f;
  t143_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t143_n75;
  t143_n74: return 0.014315054751932621f;
  t143_n75: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t143_n83;
  t143_n76: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t143_n82;
  t143_n77: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t143_n79;
  t143_n78: return 0.04888812452554703f;
  t143_n79: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t143_n81;
  t143_n80: return 0.17605917155742645f;
  t143_n81: return 0.08225174248218536f;
  t143_n82: return -0.04367274045944214f;
  t143_n83: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t143_n87;
  t143_n84: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t143_n86;
  t143_n85: return 0.07380317151546478f;
  t143_n86: return -0.3319551646709442f;
  t143_n87: return 0.04486040771007538f;
  t143_n88: return -0.012123527005314827f;
  t143_n89: if (E.getNumNameInContext() >= 3235905536 /*7.0*/) goto t143_n91;
  t143_n90: return 0.10050387680530548f;
  t143_n91: return -0.1287812441587448f;
  t143_n92: return -0.0032935324124991894f;
  t143_n93: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t143_n111;
  t143_n94: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t143_n98;
  t143_n95: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t143_n97;
  t143_n96: return 0.06643149256706238f;
  t143_n97: return 0.12808865308761597f;
  t143_n98: if (E.getSymbolScopeDistanceCost() >= 3242196992 /*12.0*/) goto t143_n102;
  t143_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n101;
  t143_n100: return -0.141658216714859f;
  t143_n101: return -0.016499163582921028f;
  t143_n102: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t143_n110;
  t143_n103: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t143_n105;
  t143_n104: return 0.031987566500902176f;
  t143_n105: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t143_n107;
  t143_n106: return 0.08088060468435287f;
  t143_n107: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t143_n109;
  t143_n108: return 0.14749839901924133f;
  t143_n109: return -0.03783364221453667f;
  t143_n110: return -0.02513645775616169f;
  t143_n111: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t143_n139;
  t143_n112: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t143_n114;
  t143_n113: return -0.002473875181749463f;
  t143_n114: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t143_n138;
  t143_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t143_n131;
  t143_n116: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t143_n126;
  t143_n117: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t143_n125;
  t143_n118: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t143_n122;
  t143_n119: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t143_n121;
  t143_n120: return -0.036614853888750076f;
  t143_n121: return 0.06815636903047562f;
  t143_n122: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t143_n124;
  t143_n123: return 0.16076846420764923f;
  t143_n124: return 0.03256663680076599f;
  t143_n125: return -0.01940220594406128f;
  t143_n126: if (E.getNumReferences() >= 3250585600 /*24.0*/) goto t143_n130;
  t143_n127: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t143_n129;
  t143_n128: return 0.12837131321430206f;
  t143_n129: return -0.559484601020813f;
  t143_n130: return 0.0221323873847723f;
  t143_n131: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t143_n133;
  t143_n132: return -0.206013485789299f;
  t143_n133: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t143_n137;
  t143_n134: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t143_n136;
  t143_n135: return -0.13839398324489594f;
  t143_n136: return 0.11380842328071594f;
  t143_n137: return 0.10081978887319565f;
  t143_n138: return 0.12463614344596863f;
  t143_n139: if (E.getNumReferences() >= 3343679232 /*52351.0*/) goto t143_n141;
  t143_n140: return 0.01077981572598219f;
  t143_n141: return -0.6597424745559692f;
  t143_n142: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t143_n148;
  t143_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n145;
  t143_n144: return -0.15461452305316925f;
  t143_n145: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t143_n147;
  t143_n146: return 0.00992329977452755f;
  t143_n147: return -0.0850663036108017f;
  t143_n148: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t143_n156;
  t143_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t143_n155;
  t143_n150: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t143_n152;
  t143_n151: return 0.062735415995121f;
  t143_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t143_n154;
  t143_n153: return 0.060327012091875076f;
  t143_n154: return 0.13100728392601013f;
  t143_n155: return -0.3897840082645416f;
  t143_n156: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t143_n158;
  t143_n157: return 0.03171736001968384f;
  t143_n158: return 0.10440531373023987f;
  t143_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t143_n187;
  t143_n160: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t143_n164;
  t143_n161: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t143_n163;
  t143_n162: return -0.04499287158250809f;
  t143_n163: return 0.12377656251192093f;
  t143_n164: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t143_n176;
  t143_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t143_n175;
  t143_n166: if (E.getNumReferences() >= 3294486528 /*887.5*/) goto t143_n168;
  t143_n167: return -0.004128517583012581f;
  t143_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t143_n174;
  t143_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t143_n173;
  t143_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t143_n172;
  t143_n171: return -0.3641900420188904f;
  t143_n172: return -0.039940156042575836f;
  t143_n173: return 0.058195728808641434f;
  t143_n174: return 0.03917231410741806f;
  t143_n175: return 0.022186055779457092f;
  t143_n176: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t143_n180;
  t143_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t143_n179;
  t143_n178: return 0.010914502665400505f;
  t143_n179: return 0.03290357440710068f;
  t143_n180: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t143_n186;
  t143_n181: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t143_n185;
  t143_n182: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t143_n184;
  t143_n183: return 0.015664078295230865f;
  t143_n184: return -0.0652126595377922f;
  t143_n185: return 0.010528039187192917f;
  t143_n186: return 0.0940544381737709f;
  t143_n187: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t143_n233;
  t143_n188: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t143_n220;
  t143_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n199;
  t143_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t143_n196;
  t143_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n195;
  t143_n192: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t143_n194;
  t143_n193: return -0.020484253764152527f;
  t143_n194: return 0.039555929601192474f;
  t143_n195: return -0.002672286471351981f;
  t143_n196: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t143_n198;
  t143_n197: return -0.24585478007793427f;
  t143_n198: return 0.0267016664147377f;
  t143_n199: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t143_n215;
  t143_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t143_n202;
  t143_n201: return -0.0571664460003376f;
  t143_n202: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t143_n210;
  t143_n203: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t143_n209;
  t143_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t143_n206;
  t143_n205: return 0.07355272769927979f;
  t143_n206: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t143_n208;
  t143_n207: return 0.10172897577285767f;
  t143_n208: return 0.15868164598941803f;
  t143_n209: return -0.006229140795767307f;
  t143_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t143_n214;
  t143_n211: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t143_n213;
  t143_n212: return -0.05272066965699196f;
  t143_n213: return 0.01636551506817341f;
  t143_n214: return 0.03857787325978279f;
  t143_n215: if (E.getFractionNameInContext() >= 3208223130 /*0.7250000238418579*/) goto t143_n219;
  t143_n216: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t143_n218;
  t143_n217: return -0.009158443659543991f;
  t143_n218: return 0.14032918214797974f;
  t143_n219: return 0.11916091293096542f;
  t143_n220: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t143_n228;
  t143_n221: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t143_n223;
  t143_n222: return 0.32477250695228577f;
  t143_n223: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t143_n225;
  t143_n224: return 0.01898655854165554f;
  t143_n225: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t143_n227;
  t143_n226: return -0.015437116846442223f;
  t143_n227: return 0.01570088416337967f;
  t143_n228: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t143_n232;
  t143_n229: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t143_n231;
  t143_n230: return -2.6872670650482178f;
  t143_n231: return 0.005631542298942804f;
  t143_n232: return 0.12924063205718994f;
  t143_n233: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t143_n245;
  t143_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n244;
  t143_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t143_n243;
  t143_n236: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t143_n242;
  t143_n237: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t143_n241;
  t143_n238: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t143_n240;
  t143_n239: return -0.0025651874020695686f;
  t143_n240: return -0.061847835779190063f;
  t143_n241: return 0.0005863608093932271f;
  t143_n242: return 0.10966435074806213f;
  t143_n243: return 0.0005538416444323957f;
  t143_n244: return 0.01011747121810913f;
  t143_n245: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t143_n249;
  t143_n246: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t143_n248;
  t143_n247: return 0.026994207873940468f;
  t143_n248: return -0.10813795030117035f;
  t143_n249: return -0.09491339325904846f;
  t143_n250: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t143_n252;
  t143_n251: return 0.008595187216997147f;
  t143_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t143_n254;
  t143_n253: return -0.034982457756996155f;
  t143_n254: return -0.0064215101301670074f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree144(const Example& E) {
  t144_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t144_n98;
  t144_n1: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t144_n3;
  t144_n2: return 5.004140257369727e-05f;
  t144_n3: if (E.getNumReferences() >= 3364003440 /*267507.5*/) goto t144_n97;
  t144_n4: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t144_n96;
  t144_n5: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t144_n95;
  t144_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Operator))) goto t144_n66;
  t144_n7: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t144_n23;
  t144_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n10;
  t144_n9: return -0.014520823024213314f;
  t144_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t144_n16;
  t144_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t144_n13;
  t144_n12: return 0.006812186446040869f;
  t144_n13: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t144_n15;
  t144_n14: return 0.13162052631378174f;
  t144_n15: return 0.030399303883314133f;
  t144_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n18;
  t144_n17: return -0.007710539735853672f;
  t144_n18: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t144_n22;
  t144_n19: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t144_n21;
  t144_n20: return 0.008597863838076591f;
  t144_n21: return -0.26500117778778076f;
  t144_n22: return 0.03989306837320328f;
  t144_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n29;
  t144_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t144_n26;
  t144_n25: return -0.029172630980610847f;
  t144_n26: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t144_n28;
  t144_n27: return -0.027851751074194908f;
  t144_n28: return 0.02485487423837185f;
  t144_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t144_n33;
  t144_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t144_n32;
  t144_n31: return -0.027240298688411713f;
  t144_n32: return 0.07763063907623291f;
  t144_n33: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t144_n53;
  t144_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n38;
  t144_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t144_n37;
  t144_n36: return -0.10039796680212021f;
  t144_n37: return -0.009848442859947681f;
  t144_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t144_n52;
  t144_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t144_n51;
  t144_n40: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t144_n42;
  t144_n41: return 0.005518535617738962f;
  t144_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t144_n44;
  t144_n43: return -0.10501861572265625f;
  t144_n44: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t144_n50;
  t144_n45: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t144_n49;
  t144_n46: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t144_n48;
  t144_n47: return 0.02636292576789856f;
  t144_n48: return -0.014190574176609516f;
  t144_n49: return -0.10342826694250107f;
  t144_n50: return -0.08109612762928009f;
  t144_n51: return -0.03590983897447586f;
  t144_n52: return 0.02784261666238308f;
  t144_n53: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t144_n59;
  t144_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n56;
  t144_n55: return -0.025938794016838074f;
  t144_n56: if (E.getFractionNameInContext() >= 3175316908 /*0.04772727191448212*/) goto t144_n58;
  t144_n57: return -0.006516226101666689f;
  t144_n58: return 0.022869747132062912f;
  t144_n59: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t144_n63;
  t144_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t144_n62;
  t144_n61: return -0.0012823722790926695f;
  t144_n62: return 0.0227559395134449f;
  t144_n63: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t144_n65;
  t144_n64: return -0.03406034782528877f;
  t144_n65: return 0.0011273471172899008f;
  t144_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t144_n76;
  t144_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n69;
  t144_n68: return -0.016297662630677223f;
  t144_n69: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t144_n71;
  t144_n70: return 0.013218212872743607f;
  t144_n71: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t144_n73;
  t144_n72: return 0.053308289498090744f;
  t144_n73: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t144_n75;
  t144_n74: return 0.09972955286502838f;
  t144_n75: return 0.06671057641506195f;
  t144_n76: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t144_n78;
  t144_n77: return -0.013447746634483337f;
  t144_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t144_n94;
  t144_n79: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t144_n89;
  t144_n80: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t144_n88;
  t144_n81: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t144_n83;
  t144_n82: return 0.03371023014187813f;
  t144_n83: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t144_n87;
  t144_n84: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t144_n86;
  t144_n85: return 0.1181865930557251f;
  t144_n86: return 0.11505729705095291f;
  t144_n87: return -0.11087486892938614f;
  t144_n88: return -0.10673413425683975f;
  t144_n89: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t144_n91;
  t144_n90: return 0.032031506299972534f;
  t144_n91: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t144_n93;
  t144_n92: return 0.13458660244941711f;
  t144_n93: return 0.057098936289548874f;
  t144_n94: return -0.015433134511113167f;
  t144_n95: return -0.019839659333229065f;
  t144_n96: return -0.11525487154722214f;
  t144_n97: return 0.021177267655730247f;
  t144_n98: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t144_n236;
  t144_n99: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t144_n217;
  t144_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n216;
  t144_n101: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t144_n211;
  t144_n102: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t144_n198;
  t144_n103: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t144_n197;
  t144_n104: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t144_n178;
  t144_n105: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t144_n177;
  t144_n106: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t144_n172;
  t144_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t144_n141;
  t144_n108: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t144_n134;
  t144_n109: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t144_n133;
  t144_n110: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t144_n124;
  t144_n111: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t144_n115;
  t144_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n114;
  t144_n113: return -0.23701885342597961f;
  t144_n114: return 0.02221003547310829f;
  t144_n115: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t144_n119;
  t144_n116: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t144_n118;
  t144_n117: return -0.025037214159965515f;
  t144_n118: return -0.1844075471162796f;
  t144_n119: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t144_n121;
  t144_n120: return 0.01605342887341976f;
  t144_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n123;
  t144_n122: return -0.09695165604352951f;
  t144_n123: return 0.06553646177053452f;
  t144_n124: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t144_n128;
  t144_n125: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t144_n127;
  t144_n126: return -0.007599124684929848f;
  t144_n127: return 0.12198180705308914f;
  t144_n128: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t144_n130;
  t144_n129: return 0.11760992556810379f;
  t144_n130: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t144_n132;
  t144_n131: return -0.3755583167076111f;
  t144_n132: return 0.10292435437440872f;
  t144_n133: return -0.03918236866593361f;
  t144_n134: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t144_n140;
  t144_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t144_n137;
  t144_n136: return 0.0220626313239336f;
  t144_n137: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t144_n139;
  t144_n138: return -0.4459063708782196f;
  t144_n139: return 0.19895318150520325f;
  t144_n140: return -0.013356195762753487f;
  t144_n141: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t144_n165;
  t144_n142: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t144_n164;
  t144_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t144_n163;
  t144_n144: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t144_n160;
  t144_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t144_n147;
  t144_n146: return 0.0013265142915770411f;
  t144_n147: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t144_n157;
  t144_n148: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t144_n150;
  t144_n149: return -0.6308029890060425f;
  t144_n150: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t144_n156;
  t144_n151: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t144_n155;
  t144_n152: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t144_n154;
  t144_n153: return 0.15378226339817047f;
  t144_n154: return -0.056433793157339096f;
  t144_n155: return 0.06165379658341408f;
  t144_n156: return -0.12725485861301422f;
  t144_n157: if (E.getFractionNameInContext() >= 3174706828 /*0.04545454680919647*/) goto t144_n159;
  t144_n158: return 0.08105622231960297f;
  t144_n159: return 0.11520858854055405f;
  t144_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t144_n162;
  t144_n161: return -0.1871601939201355f;
  t144_n162: return -0.046378243714571f;
  t144_n163: return 0.04168562963604927f;
  t144_n164: return -0.009333549998700619f;
  t144_n165: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t144_n167;
  t144_n166: return -0.013516178354620934f;
  t144_n167: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t144_n171;
  t144_n168: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t144_n170;
  t144_n169: return 0.045029085129499435f;
  t144_n170: return 0.14443641901016235f;
  t144_n171: return -0.021032089367508888f;
  t144_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t144_n176;
  t144_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t144_n175;
  t144_n174: return -0.1214531734585762f;
  t144_n175: return 0.034545864909887314f;
  t144_n176: return 0.1229785680770874f;
  t144_n177: return -0.04854540154337883f;
  t144_n178: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t144_n186;
  t144_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t144_n185;
  t144_n180: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t144_n182;
  t144_n181: return -0.0406004823744297f;
  t144_n182: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t144_n184;
  t144_n183: return -0.3585803508758545f;
  t144_n184: return -0.1381608545780182f;
  t144_n185: return -0.01965940184891224f;
  t144_n186: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t144_n190;
  t144_n187: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t144_n189;
  t144_n188: return -0.21429283916950226f;
  t144_n189: return -0.020854227244853973f;
  t144_n190: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t144_n192;
  t144_n191: return 0.022570235654711723f;
  t144_n192: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t144_n194;
  t144_n193: return 0.048412300646305084f;
  t144_n194: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t144_n196;
  t144_n195: return 0.12223375588655472f;
  t144_n196: return 0.06363322585821152f;
  t144_n197: return 0.0011380879441276193f;
  t144_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t144_n208;
  t144_n199: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t144_n201;
  t144_n200: return 0.031111836433410645f;
  t144_n201: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t144_n203;
  t144_n202: return -0.052601851522922516f;
  t144_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t144_n207;
  t144_n204: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t144_n206;
  t144_n205: return -0.047851212322711945f;
  t144_n206: return 0.01503789983689785f;
  t144_n207: return 0.022789841517806053f;
  t144_n208: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t144_n210;
  t144_n209: return 0.05741705372929573f;
  t144_n210: return 0.13051019608974457f;
  t144_n211: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t144_n213;
  t144_n212: return 0.022710371762514114f;
  t144_n213: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t144_n215;
  t144_n214: return -0.025147275999188423f;
  t144_n215: return 0.008555847220122814f;
  t144_n216: return 0.005064033903181553f;
  t144_n217: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t144_n219;
  t144_n218: return 0.002772195963189006f;
  t144_n219: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t144_n227;
  t144_n220: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t144_n226;
  t144_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t144_n223;
  t144_n222: return 0.02565554343163967f;
  t144_n223: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t144_n225;
  t144_n224: return 0.038066886365413666f;
  t144_n225: return 0.46634531021118164f;
  t144_n226: return 0.08896146714687347f;
  t144_n227: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t144_n231;
  t144_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t144_n230;
  t144_n229: return -0.024859096854925156f;
  t144_n230: return 0.037706974893808365f;
  t144_n231: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t144_n233;
  t144_n232: return 0.04188825190067291f;
  t144_n233: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t144_n235;
  t144_n234: return 0.1235916018486023f;
  t144_n235: return -0.04296969622373581f;
  t144_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t144_n246;
  t144_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n239;
  t144_n238: return -0.0002479067479725927f;
  t144_n239: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t144_n241;
  t144_n240: return 0.015986083075404167f;
  t144_n241: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t144_n245;
  t144_n242: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t144_n244;
  t144_n243: return -0.025082221254706383f;
  t144_n244: return 0.002453000284731388f;
  t144_n245: return 0.051845256239175797f;
  t144_n246: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t144_n250;
  t144_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t144_n249;
  t144_n248: return -0.004900412634015083f;
  t144_n249: return 0.002606191672384739f;
  t144_n250: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t144_n252;
  t144_n251: return 0.004449937026947737f;
  t144_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t144_n254;
  t144_n253: return 0.009633165784180164f;
  t144_n254: return 0.03035796992480755f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree145(const Example& E) {
  t145_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t145_n248;
  t145_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t145_n225;
  t145_n2: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t145_n84;
  t145_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n83;
  t145_n4: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t145_n10;
  t145_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n7;
  t145_n6: return -0.016962097957730293f;
  t145_n7: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t145_n9;
  t145_n8: return 0.004234594292938709f;
  t145_n9: return 0.033516477793455124f;
  t145_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t145_n28;
  t145_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t145_n25;
  t145_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n24;
  t145_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n23;
  t145_n14: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t145_n20;
  t145_n15: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t145_n19;
  t145_n16: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t145_n18;
  t145_n17: return -0.8377662301063538f;
  t145_n18: return -0.09500553458929062f;
  t145_n19: return -0.024822358042001724f;
  t145_n20: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t145_n22;
  t145_n21: return -0.33434394001960754f;
  t145_n22: return 0.2697595953941345f;
  t145_n23: return 0.018381264060735703f;
  t145_n24: return -0.00634567579254508f;
  t145_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t145_n27;
  t145_n26: return 0.026646971702575684f;
  t145_n27: return 0.049291614443063736f;
  t145_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n44;
  t145_n29: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t145_n43;
  t145_n30: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t145_n40;
  t145_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n37;
  t145_n32: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t145_n36;
  t145_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n35;
  t145_n34: return -0.01304718665778637f;
  t145_n35: return -0.07577437907457352f;
  t145_n36: return -0.1152137741446495f;
  t145_n37: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t145_n39;
  t145_n38: return -0.1599292904138565f;
  t145_n39: return 0.20704591274261475f;
  t145_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n42;
  t145_n41: return -0.08702253550291061f;
  t145_n42: return 0.03475483879446983f;
  t145_n43: return -0.00707037840038538f;
  t145_n44: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t145_n80;
  t145_n45: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t145_n47;
  t145_n46: return -0.0022849428933113813f;
  t145_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t145_n65;
  t145_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t145_n50;
  t145_n49: return -0.01027411688119173f;
  t145_n50: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t145_n60;
  t145_n51: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t145_n59;
  t145_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n54;
  t145_n53: return 0.020067773759365082f;
  t145_n54: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t145_n58;
  t145_n55: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t145_n57;
  t145_n56: return -0.027464965358376503f;
  t145_n57: return -0.2167072743177414f;
  t145_n58: return 0.050394196063280106f;
  t145_n59: return 0.0936138704419136f;
  t145_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n64;
  t145_n61: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t145_n63;
  t145_n62: return -0.1756063997745514f;
  t145_n63: return -0.12718980014324188f;
  t145_n64: return 0.007242790888994932f;
  t145_n65: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n73;
  t145_n66: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t145_n68;
  t145_n67: return 0.005534606520086527f;
  t145_n68: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t145_n70;
  t145_n69: return 0.09279580414295197f;
  t145_n70: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t145_n72;
  t145_n71: return -0.14643923938274384f;
  t145_n72: return 0.22983244061470032f;
  t145_n73: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t145_n79;
  t145_n74: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t145_n76;
  t145_n75: return 0.061968155205249786f;
  t145_n76: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t145_n78;
  t145_n77: return -0.14146800339221954f;
  t145_n78: return 0.12076234072446823f;
  t145_n79: return -0.16282415390014648f;
  t145_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n82;
  t145_n81: return -0.12437945604324341f;
  t145_n82: return 0.06778945028781891f;
  t145_n83: return 0.0008389283902943134f;
  t145_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t145_n114;
  t145_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n109;
  t145_n86: if (E.getNumReferences() >= 3340598528 /*40317.0*/) goto t145_n108;
  t145_n87: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t145_n91;
  t145_n88: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t145_n90;
  t145_n89: return -0.0026395930908620358f;
  t145_n90: return -0.04266232252120972f;
  t145_n91: if (E.getNumReferences() >= 3303100416 /*1802.5*/) goto t145_n103;
  t145_n92: if (E.getNumReferences() >= 3277881344 /*224.5*/) goto t145_n94;
  t145_n93: return 0.022673651576042175f;
  t145_n94: if (E.getNumReferences() >= 3280404480 /*270.0*/) goto t145_n96;
  t145_n95: return 0.13020797073841095f;
  t145_n96: if (E.getFractionNameInContext() >= 3177875683 /*0.057259451597929*/) goto t145_n100;
  t145_n97: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t145_n99;
  t145_n98: return -0.11821195483207703f;
  t145_n99: return 0.08210229873657227f;
  t145_n100: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t145_n102;
  t145_n101: return 0.13724294304847717f;
  t145_n102: return 0.07560966908931732f;
  t145_n103: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t145_n107;
  t145_n104: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t145_n106;
  t145_n105: return -0.11287709325551987f;
  t145_n106: return -2.2254202365875244f;
  t145_n107: return -0.04757709056138992f;
  t145_n108: return -0.05026332661509514f;
  t145_n109: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t145_n111;
  t145_n110: return 0.025097331032156944f;
  t145_n111: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t145_n113;
  t145_n112: return 0.1330629587173462f;
  t145_n113: return 0.20815689861774445f;
  t145_n114: if (E.getNumReferences() >= 3331513856 /*18799.0*/) goto t145_n150;
  t145_n115: if (E.getNumReferences() >= 3330547456 /*16911.5*/) goto t145_n147;
  t145_n116: if (E.getNumReferences() >= 3324873728 /*11107.0*/) goto t145_n128;
  t145_n117: if (E.getNumReferences() >= 3322937344 /*9216.0*/) goto t145_n121;
  t145_n118: if (E.getNumReferences() >= 3322597376 /*8884.0*/) goto t145_n120;
  t145_n119: return 0.007139839231967926f;
  t145_n120: return 0.08143482357263565f;
  t145_n121: if (E.getNumReferences() >= 3323005952 /*9283.0*/) goto t145_n127;
  t145_n122: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t145_n126;
  t145_n123: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t145_n125;
  t145_n124: return -0.12806160748004913f;
  t145_n125: return -0.33987849950790405f;
  t145_n126: return -1.5594873428344727f;
  t145_n127: return -0.017081493511795998f;
  t145_n128: if (E.getNumReferences() >= 3327478272 /*13650.5*/) goto t145_n144;
  t145_n129: if (E.getNumReferences() >= 3327454720 /*13627.5*/) goto t145_n139;
  t145_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t145_n132;
  t145_n131: return 0.03445060923695564f;
  t145_n132: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t145_n138;
  t145_n133: if (E.getNumReferences() >= 3326979584 /*13163.5*/) goto t145_n137;
  t145_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n136;
  t145_n135: return 0.06774517893791199f;
  t145_n136: return 0.1272001713514328f;
  t145_n137: return 0.03280264511704445f;
  t145_n138: return -0.012766459956765175f;
  t145_n139: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t145_n143;
  t145_n140: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t145_n142;
  t145_n141: return 0.22872474789619446f;
  t145_n142: return 0.11832745373249054f;
  t145_n143: return 0.10187754034996033f;
  t145_n144: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t145_n146;
  t145_n145: return -0.006250785663723946f;
  t145_n146: return 0.221157044172287f;
  t145_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t145_n149;
  t145_n148: return -0.05254361778497696f;
  t145_n149: return 0.1577230840921402f;
  t145_n150: if (E.getNumReferences() >= 3334350336 /*24339.0*/) goto t145_n216;
  t145_n151: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t145_n209;
  t145_n152: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t145_n202;
  t145_n153: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t145_n201;
  t145_n154: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t145_n184;
  t145_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t145_n177;
  t145_n156: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t145_n172;
  t145_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n169;
  t145_n158: if (E.getNumReferences() >= 3331859712 /*19474.5*/) goto t145_n168;
  t145_n159: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t145_n161;
  t145_n160: return 0.04579807072877884f;
  t145_n161: if (E.getNumReferences() >= 3331773952 /*19307.0*/) goto t145_n165;
  t145_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n164;
  t145_n163: return 0.13790558278560638f;
  t145_n164: return 0.07482150942087173f;
  t145_n165: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t145_n167;
  t145_n166: return 0.07425963878631592f;
  t145_n167: return -0.04975201189517975f;
  t145_n168: return -0.13726498186588287f;
  t145_n169: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t145_n171;
  t145_n170: return 0.08823006600141525f;
  t145_n171: return -0.3923184871673584f;
  t145_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n174;
  t145_n173: return 0.06480859220027924f;
  t145_n174: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t145_n176;
  t145_n175: return -0.07422570139169693f;
  t145_n176: return 0.11263154447078705f;
  t145_n177: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t145_n183;
  t145_n178: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t145_n182;
  t145_n179: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t145_n181;
  t145_n180: return 0.09839775413274765f;
  t145_n181: return 0.1746002435684204f;
  t145_n182: return 0.10506655275821686f;
  t145_n183: return 0.05263441056013107f;
  t145_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n194;
  t145_n185: if (E.getNumReferences() >= 3333030400 /*21761.0*/) goto t145_n187;
  t145_n186: return -0.047116246074438095f;
  t145_n187: if (E.getNumReferences() >= 3333079040 /*21856.0*/) goto t145_n193;
  t145_n188: if (E.getNumReferences() >= 3333053952 /*21807.0*/) goto t145_n190;
  t145_n189: return 0.047979287803173065f;
  t145_n190: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n192;
  t145_n191: return 0.07883195579051971f;
  t145_n192: return 0.15400809049606323f;
  t145_n193: return 0.007943528704345226f;
  t145_n194: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t145_n196;
  t145_n195: return -0.03499477356672287f;
  t145_n196: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t145_n200;
  t145_n197: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t145_n199;
  t145_n198: return 0.42342546582221985f;
  t145_n199: return 0.12726934254169464f;
  t145_n200: return 0.11540830880403519f;
  t145_n201: return -0.13576450943946838f;
  t145_n202: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n208;
  t145_n203: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t145_n205;
  t145_n204: return 0.10605553537607193f;
  t145_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t145_n207;
  t145_n206: return 0.10122273862361908f;
  t145_n207: return 0.1312047690153122f;
  t145_n208: return -0.09958752989768982f;
  t145_n209: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t145_n215;
  t145_n210: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t145_n212;
  t145_n211: return 0.08716028183698654f;
  t145_n212: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t145_n214;
  t145_n213: return 0.09562017023563385f;
  t145_n214: return 0.12091520428657532f;
  t145_n215: return -0.04197731614112854f;
  t145_n216: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t145_n220;
  t145_n217: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t145_n219;
  t145_n218: return -0.014510270208120346f;
  t145_n219: return 0.11185857653617859f;
  t145_n220: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t145_n224;
  t145_n221: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t145_n223;
  t145_n222: return 0.04317587614059448f;
  t145_n223: return 0.1164681538939476f;
  t145_n224: return 0.01418856531381607f;
  t145_n225: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t145_n233;
  t145_n226: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t145_n228;
  t145_n227: return 0.006135544739663601f;
  t145_n228: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t145_n232;
  t145_n229: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t145_n231;
  t145_n230: return 0.06171811744570732f;
  t145_n231: return 0.0033046375028789043f;
  t145_n232: return -0.07343307882547379f;
  t145_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t145_n235;
  t145_n234: return -0.0025483083445578814f;
  t145_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t145_n237;
  t145_n236: return -0.028362667188048363f;
  t145_n237: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t145_n241;
  t145_n238: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t145_n240;
  t145_n239: return 0.023692132905125618f;
  t145_n240: return -0.3442424535751343f;
  t145_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t145_n243;
  t145_n242: return -0.9593660831451416f;
  t145_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t145_n247;
  t145_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t145_n246;
  t145_n245: return 0.0040854415856301785f;
  t145_n246: return 0.05347691848874092f;
  t145_n247: return 0.05188920348882675f;
  t145_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t145_n250;
  t145_n249: return 0.0019772143568843603f;
  t145_n250: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t145_n252;
  t145_n251: return 0.02928861975669861f;
  t145_n252: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t145_n254;
  t145_n253: return 0.004752991255372763f;
  t145_n254: return 0.037738487124443054f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree146(const Example& E) {
  t146_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t146_n90;
  t146_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t146_n3;
  t146_n2: return -0.0009238710044883192f;
  t146_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t146_n87;
  t146_n4: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t146_n86;
  t146_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t146_n77;
  t146_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t146_n8;
  t146_n7: return -0.001711043412797153f;
  t146_n8: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t146_n72;
  t146_n9: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t146_n27;
  t146_n10: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t146_n20;
  t146_n11: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t146_n13;
  t146_n12: return 0.03058682009577751f;
  t146_n13: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t146_n15;
  t146_n14: return 0.008509528823196888f;
  t146_n15: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t146_n19;
  t146_n16: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t146_n18;
  t146_n17: return -0.17928457260131836f;
  t146_n18: return 0.12200561910867691f;
  t146_n19: return -0.007446837145835161f;
  t146_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t146_n26;
  t146_n21: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t146_n23;
  t146_n22: return -0.1752908080816269f;
  t146_n23: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t146_n25;
  t146_n24: return 0.012897800654172897f;
  t146_n25: return -0.14279571175575256f;
  t146_n26: return -0.2027616947889328f;
  t146_n27: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t146_n31;
  t146_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t146_n30;
  t146_n29: return 0.04138743877410889f;
  t146_n30: return 0.0984049141407013f;
  t146_n31: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t146_n45;
  t146_n32: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t146_n38;
  t146_n33: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t146_n35;
  t146_n34: return 0.031989794224500656f;
  t146_n35: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t146_n37;
  t146_n36: return 0.10793500393629074f;
  t146_n37: return 0.012507105246186256f;
  t146_n38: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t146_n42;
  t146_n39: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t146_n41;
  t146_n40: return 0.059771828353405f;
  t146_n41: return -0.00915819127112627f;
  t146_n42: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t146_n44;
  t146_n43: return -0.05532153695821762f;
  t146_n44: return -1.1909608840942383f;
  t146_n45: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t146_n59;
  t146_n46: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t146_n48;
  t146_n47: return -0.031206930056214333f;
  t146_n48: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t146_n50;
  t146_n49: return -0.036027874797582626f;
  t146_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t146_n52;
  t146_n51: return 0.04719020426273346f;
  t146_n52: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t146_n56;
  t146_n53: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t146_n55;
  t146_n54: return -0.20584243535995483f;
  t146_n55: return 0.11808928847312927f;
  t146_n56: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t146_n58;
  t146_n57: return 0.10966759920120239f;
  t146_n58: return 0.13335467875003815f;
  t146_n59: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t146_n69;
  t146_n60: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t146_n66;
  t146_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n63;
  t146_n62: return -0.16849231719970703f;
  t146_n63: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t146_n65;
  t146_n64: return 0.09303026646375656f;
  t146_n65: return 0.0021276185289025307f;
  t146_n66: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t146_n68;
  t146_n67: return -0.42625102400779724f;
  t146_n68: return 0.027409752830863f;
  t146_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n71;
  t146_n70: return -0.27249160408973694f;
  t146_n71: return 0.005664349067956209f;
  t146_n72: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t146_n76;
  t146_n73: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t146_n75;
  t146_n74: return 0.14902038872241974f;
  t146_n75: return 0.0625850260257721f;
  t146_n76: return 0.02587289735674858f;
  t146_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n83;
  t146_n78: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t146_n82;
  t146_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t146_n81;
  t146_n80: return 0.01651046611368656f;
  t146_n81: return 0.08771350234746933f;
  t146_n82: return -0.050532061606645584f;
  t146_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t146_n85;
  t146_n84: return -0.05336328595876694f;
  t146_n85: return -0.001470765215344727f;
  t146_n86: return 0.006407003849744797f;
  t146_n87: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t146_n89;
  t146_n88: return 0.0011650307569652796f;
  t146_n89: return 0.0035498780198395252f;
  t146_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n128;
  t146_n91: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t146_n119;
  t146_n92: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t146_n98;
  t146_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n97;
  t146_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t146_n96;
  t146_n95: return -0.11538241058588028f;
  t146_n96: return -0.10297432541847229f;
  t146_n97: return -0.017229577526450157f;
  t146_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t146_n116;
  t146_n99: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t146_n101;
  t146_n100: return 0.0018068355275318027f;
  t146_n101: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t146_n105;
  t146_n102: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t146_n104;
  t146_n103: return -0.02094257064163685f;
  t146_n104: return -0.06715506315231323f;
  t146_n105: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t146_n115;
  t146_n106: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t146_n114;
  t146_n107: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t146_n111;
  t146_n108: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t146_n110;
  t146_n109: return -0.0025679138489067554f;
  t146_n110: return -0.03965217247605324f;
  t146_n111: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t146_n113;
  t146_n112: return -0.282495379447937f;
  t146_n113: return 0.06830496340990067f;
  t146_n114: return 0.03059319406747818f;
  t146_n115: return -0.39589232206344604f;
  t146_n116: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t146_n118;
  t146_n117: return 0.015150396153330803f;
  t146_n118: return 0.05457284301519394f;
  t146_n119: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t146_n125;
  t146_n120: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t146_n122;
  t146_n121: return -0.01054100226610899f;
  t146_n122: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t146_n124;
  t146_n123: return -0.028245875611901283f;
  t146_n124: return -0.26589760184288025f;
  t146_n125: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t146_n127;
  t146_n126: return -0.5111668109893799f;
  t146_n127: return 0.07784746587276459f;
  t146_n128: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t146_n252;
  t146_n129: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t146_n249;
  t146_n130: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t146_n248;
  t146_n131: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t146_n247;
  t146_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t146_n168;
  t146_n133: if (E.getNumReferences() >= 3329390080 /*15517.5*/) goto t146_n167;
  t146_n134: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t146_n166;
  t146_n135: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t146_n155;
  t146_n136: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t146_n142;
  t146_n137: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t146_n139;
  t146_n138: return 0.006268741562962532f;
  t146_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t146_n141;
  t146_n140: return -0.7038456201553345f;
  t146_n141: return -0.061807781457901f;
  t146_n142: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t146_n144;
  t146_n143: return -0.05907009541988373f;
  t146_n144: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t146_n152;
  t146_n145: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t146_n149;
  t146_n146: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t146_n148;
  t146_n147: return 0.04271235316991806f;
  t146_n148: return -0.026608886197209358f;
  t146_n149: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t146_n151;
  t146_n150: return 0.11920171976089478f;
  t146_n151: return 0.06155640631914139f;
  t146_n152: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t146_n154;
  t146_n153: return -0.5291914343833923f;
  t146_n154: return -0.11093154549598694f;
  t146_n155: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t146_n157;
  t146_n156: return -0.012648673728108406f;
  t146_n157: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t146_n165;
  t146_n158: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t146_n160;
  t146_n159: return 0.011763175018131733f;
  t146_n160: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t146_n162;
  t146_n161: return 0.13866780698299408f;
  t146_n162: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t146_n164;
  t146_n163: return -0.08909390866756439f;
  t146_n164: return 0.11078698188066483f;
  t146_n165: return -0.3332878351211548f;
  t146_n166: return 0.02366303652524948f;
  t146_n167: return 0.2953280806541443f;
  t146_n168: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t146_n212;
  t146_n169: if (E.getNumReferences() >= 3323748864 /*10008.5*/) goto t146_n211;
  t146_n170: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t146_n208;
  t146_n171: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t146_n195;
  t146_n172: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t146_n174;
  t146_n173: return -0.0031144279055297375f;
  t146_n174: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t146_n192;
  t146_n175: if (E.getNumReferences() >= 3310125056 /*3272.0*/) goto t146_n189;
  t146_n176: if (E.getNumReferences() >= 3277881344 /*224.5*/) goto t146_n178;
  t146_n177: return 0.02199121192097664f;
  t146_n178: if (E.getNumReferences() >= 3293478912 /*826.0*/) goto t146_n186;
  t146_n179: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t146_n181;
  t146_n180: return 0.08205527067184448f;
  t146_n181: if (E.getNumReferences() >= 3284533248 /*396.0*/) goto t146_n183;
  t146_n182: return 0.12564368546009064f;
  t146_n183: if (E.getNumReferences() >= 3287695360 /*492.5*/) goto t146_n185;
  t146_n184: return -0.07692187279462814f;
  t146_n185: return 0.10859159380197525f;
  t146_n186: if (E.getNumReferences() >= 3300478976 /*1482.5*/) goto t146_n188;
  t146_n187: return -0.14827753603458405f;
  t146_n188: return 0.07698275148868561f;
  t146_n189: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t146_n191;
  t146_n190: return -0.595031201839447f;
  t146_n191: return -0.03512046858668327f;
  t146_n192: if (E.getNumReferences() >= 3323708928 /*9969.5*/) goto t146_n194;
  t146_n193: return 0.004720330238342285f;
  t146_n194: return 0.1359364241361618f;
  t146_n195: if (E.getNumReferences() >= 3318066176 /*6325.5*/) goto t146_n199;
  t146_n196: if (E.getNumReferences() >= 3299115008 /*1316.0*/) goto t146_n198;
  t146_n197: return 0.01151252444833517f;
  t146_n198: return -0.07156732678413391f;
  t146_n199: if (E.getNumReferences() >= 3318227968 /*6404.5*/) goto t146_n205;
  t146_n200: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t146_n204;
  t146_n201: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t146_n203;
  t146_n202: return 0.1046668142080307f;
  t146_n203: return 0.10244882106781006f;
  t146_n204: return 0.10801809281110764f;
  t146_n205: if (E.getNumReferences() >= 3321964544 /*8266.0*/) goto t146_n207;
  t146_n206: return -0.09054631739854813f;
  t146_n207: return 0.04829219728708267f;
  t146_n208: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t146_n210;
  t146_n209: return 0.13627557456493378f;
  t146_n210: return -0.21608789265155792f;
  t146_n211: return -0.045744504779577255f;
  t146_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t146_n218;
  t146_n213: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t146_n217;
  t146_n214: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t146_n216;
  t146_n215: return 0.011783055029809475f;
  t146_n216: return -0.16713227331638336f;
  t146_n217: return 0.03751340135931969f;
  t146_n218: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t146_n230;
  t146_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t146_n225;
  t146_n220: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t146_n222;
  t146_n221: return 0.010144381783902645f;
  t146_n222: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t146_n224;
  t146_n223: return -0.3150927722454071f;
  t146_n224: return -0.014375285245478153f;
  t146_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t146_n229;
  t146_n226: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t146_n228;
  t146_n227: return -0.04721715301275253f;
  t146_n228: return -0.1394423544406891f;
  t146_n229: return 0.013490051031112671f;
  t146_n230: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t146_n238;
  t146_n231: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t146_n235;
  t146_n232: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t146_n234;
  t146_n233: return 0.07447401434183121f;
  t146_n234: return 0.1369399130344391f;
  t146_n235: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t146_n237;
  t146_n236: return -0.06273282319307327f;
  t146_n237: return 0.09281507879495621f;
  t146_n238: if (E.getNumReferences() >= 3272867840 /*148.0*/) goto t146_n244;
  t146_n239: if (E.getNumReferences() >= 3272245248 /*138.5*/) goto t146_n243;
  t146_n240: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t146_n242;
  t146_n241: return -0.6680701971054077f;
  t146_n242: return 0.09047350287437439f;
  t146_n243: return -0.16451895236968994f;
  t146_n244: if (E.getNumReferences() >= 3275849728 /*193.5*/) goto t146_n246;
  t146_n245: return 0.16123709082603455f;
  t146_n246: return 0.07623837888240814f;
  t146_n247: return 0.10286859422922134f;
  t146_n248: return 0.009618686512112617f;
  t146_n249: if (E.getNumReferences() >= 3335315200 /*26223.5*/) goto t146_n251;
  t146_n250: return 0.11018887907266617f;
  t146_n251: return 0.026629513129591942f;
  t146_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t146_n254;
  t146_n253: return 0.032244209200143814f;
  t146_n254: return -0.11106459051370621f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree147(const Example& E) {
  t147_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t147_n94;
  t147_n1: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t147_n93;
  t147_n2: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t147_n34;
  t147_n3: if (E.getFileProximityDistanceCost() >= 3257139200 /*41.0*/) goto t147_n33;
  t147_n4: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t147_n8;
  t147_n5: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t147_n7;
  t147_n6: return -0.0016682147979736328f;
  t147_n7: return -0.06428799778223038f;
  t147_n8: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t147_n12;
  t147_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t147_n11;
  t147_n10: return -0.04712493345141411f;
  t147_n11: return -0.01900039054453373f;
  t147_n12: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t147_n32;
  t147_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t147_n15;
  t147_n14: return -0.017216481268405914f;
  t147_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n17;
  t147_n16: return -0.021153125911951065f;
  t147_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t147_n31;
  t147_n18: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t147_n30;
  t147_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t147_n21;
  t147_n20: return 0.05175265669822693f;
  t147_n21: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t147_n23;
  t147_n22: return 0.032139282673597336f;
  t147_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t147_n29;
  t147_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t147_n26;
  t147_n25: return 0.040508121252059937f;
  t147_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n28;
  t147_n27: return -0.14268441498279572f;
  t147_n28: return 0.11701265722513199f;
  t147_n29: return 0.16256065666675568f;
  t147_n30: return -0.10799255967140198f;
  t147_n31: return -0.030654016882181168f;
  t147_n32: return -0.019258705899119377f;
  t147_n33: return 0.0004998938529752195f;
  t147_n34: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t147_n48;
  t147_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t147_n47;
  t147_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n44;
  t147_n37: if (E.getNumReferences() >= 3309969408 /*3234.0*/) goto t147_n41;
  t147_n38: if (E.getNumReferences() >= 3309606912 /*3145.5*/) goto t147_n40;
  t147_n39: return -0.013988068327307701f;
  t147_n40: return 0.12069647014141083f;
  t147_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n43;
  t147_n42: return -0.15285564959049225f;
  t147_n43: return 0.1088394895195961f;
  t147_n44: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t147_n46;
  t147_n45: return 0.04523777589201927f;
  t147_n46: return 0.013154146261513233f;
  t147_n47: return -0.0019917774479836226f;
  t147_n48: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t147_n92;
  t147_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t147_n51;
  t147_n50: return -0.0009978251764550805f;
  t147_n51: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t147_n55;
  t147_n52: if (E.getNumReferences() >= 3332389888 /*20510.0*/) goto t147_n54;
  t147_n53: return -0.04722010716795921f;
  t147_n54: return 0.011571364477276802f;
  t147_n55: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t147_n79;
  t147_n56: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t147_n78;
  t147_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t147_n59;
  t147_n58: return -0.0043238201178610325f;
  t147_n59: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t147_n77;
  t147_n60: if (E.getNumReferences() >= 3340838912 /*41256.0*/) goto t147_n76;
  t147_n61: if (E.getNumReferences() >= 3335799808 /*27170.0*/) goto t147_n75;
  t147_n62: if (E.getNumReferences() >= 3316636672 /*5627.5*/) goto t147_n72;
  t147_n63: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t147_n71;
  t147_n64: if (E.getNumReferences() >= 3305613312 /*2170.5*/) goto t147_n66;
  t147_n65: return 0.1035098284482956f;
  t147_n66: if (E.getNumReferences() >= 3307247616 /*2569.5*/) goto t147_n68;
  t147_n67: return 0.12376464158296585f;
  t147_n68: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t147_n70;
  t147_n69: return 0.16971030831336975f;
  t147_n70: return -0.11065590381622314f;
  t147_n71: return 0.09712809324264526f;
  t147_n72: if (E.getNumReferences() >= 3331977984 /*19705.5*/) goto t147_n74;
  t147_n73: return 0.05314154177904129f;
  t147_n74: return 0.11205075681209564f;
  t147_n75: return -0.11589395254850388f;
  t147_n76: return 0.09960704296827316f;
  t147_n77: return -0.06253518909215927f;
  t147_n78: return -0.01354311965405941f;
  t147_n79: if (E.getNumReferences() >= 3344186496 /*54332.5*/) goto t147_n91;
  t147_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t147_n82;
  t147_n81: return 0.040357351303100586f;
  t147_n82: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t147_n84;
  t147_n83: return -0.2274387776851654f;
  t147_n84: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t147_n90;
  t147_n85: if (E.getNumReferences() >= 3328774656 /*14916.5*/) goto t147_n87;
  t147_n86: return 0.13568644225597382f;
  t147_n87: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t147_n89;
  t147_n88: return -0.18557246029376984f;
  t147_n89: return 0.14420653879642487f;
  t147_n90: return -0.12022777646780014f;
  t147_n91: return 0.10971905291080475f;
  t147_n92: return 0.141806498169899f;
  t147_n93: return 0.001121508190408349f;
  t147_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n122;
  t147_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t147_n117;
  t147_n96: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t147_n112;
  t147_n97: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t147_n105;
  t147_n98: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t147_n100;
  t147_n99: return -0.0016710602212697268f;
  t147_n100: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t147_n102;
  t147_n101: return -0.009154300205409527f;
  t147_n102: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t147_n104;
  t147_n103: return -0.03534110635519028f;
  t147_n104: return -0.19822916388511658f;
  t147_n105: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t147_n111;
  t147_n106: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t147_n110;
  t147_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t147_n109;
  t147_n108: return -0.03483160585165024f;
  t147_n109: return -0.27916088700294495f;
  t147_n110: return 0.04110053926706314f;
  t147_n111: return 0.0929543673992157f;
  t147_n112: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t147_n116;
  t147_n113: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t147_n115;
  t147_n114: return 0.020146287977695465f;
  t147_n115: return 0.12419816106557846f;
  t147_n116: return -0.019468728452920914f;
  t147_n117: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t147_n119;
  t147_n118: return -0.031116008758544922f;
  t147_n119: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t147_n121;
  t147_n120: return 0.013496862724423409f;
  t147_n121: return 0.035050004720687866f;
  t147_n122: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t147_n250;
  t147_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t147_n191;
  t147_n124: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t147_n176;
  t147_n125: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t147_n173;
  t147_n126: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t147_n154;
  t147_n127: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t147_n131;
  t147_n128: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t147_n130;
  t147_n129: return 0.06089054420590401f;
  t147_n130: return -0.2508236765861511f;
  t147_n131: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t147_n147;
  t147_n132: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t147_n144;
  t147_n133: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t147_n143;
  t147_n134: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t147_n136;
  t147_n135: return -0.17708393931388855f;
  t147_n136: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t147_n138;
  t147_n137: return 0.04756613448262215f;
  t147_n138: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t147_n142;
  t147_n139: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t147_n141;
  t147_n140: return -0.08216382563114166f;
  t147_n141: return -0.6517001986503601f;
  t147_n142: return 0.06651616096496582f;
  t147_n143: return -0.378934383392334f;
  t147_n144: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t147_n146;
  t147_n145: return -0.0004918013582937419f;
  t147_n146: return -0.6920337677001953f;
  t147_n147: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t147_n149;
  t147_n148: return 0.09325874596834183f;
  t147_n149: if (E.getNumReferences() >= 3261333504 /*57.0*/) goto t147_n153;
  t147_n150: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t147_n152;
  t147_n151: return -0.7749258875846863f;
  t147_n152: return 0.06944756209850311f;
  t147_n153: return 0.0866810604929924f;
  t147_n154: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t147_n156;
  t147_n155: return 0.009110266342759132f;
  t147_n156: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t147_n172;
  t147_n157: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t147_n169;
  t147_n158: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t147_n164;
  t147_n159: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t147_n163;
  t147_n160: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t147_n162;
  t147_n161: return 0.16063012182712555f;
  t147_n162: return 0.03620489686727524f;
  t147_n163: return -0.4160708487033844f;
  t147_n164: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t147_n166;
  t147_n165: return 0.0724511593580246f;
  t147_n166: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t147_n168;
  t147_n167: return 0.01691935397684574f;
  t147_n168: return 0.15734019875526428f;
  t147_n169: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t147_n171;
  t147_n170: return 0.1363849937915802f;
  t147_n171: return 0.05822819471359253f;
  t147_n172: return -0.10596578568220139f;
  t147_n173: if (E.getNumReferences() >= 3280617472 /*276.5*/) goto t147_n175;
  t147_n174: return -0.009356453083455563f;
  t147_n175: return -0.4938565194606781f;
  t147_n176: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t147_n188;
  t147_n177: if (E.getNumReferences() >= 3301273600 /*1579.5*/) goto t147_n187;
  t147_n178: if (E.getNumReferences() >= 3300638720 /*1502.0*/) goto t147_n186;
  t147_n179: if (E.getNumReferences() >= 3280781312 /*281.5*/) goto t147_n181;
  t147_n180: return 0.13338540494441986f;
  t147_n181: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t147_n183;
  t147_n182: return 0.04120105877518654f;
  t147_n183: if (E.getNumReferences() >= 3282436096 /*332.0*/) goto t147_n185;
  t147_n184: return 0.13647425174713135f;
  t147_n185: return 0.016254311427474022f;
  t147_n186: return 0.11747866123914719f;
  t147_n187: return -0.009714770130813122f;
  t147_n188: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t147_n190;
  t147_n189: return 0.27383172512054443f;
  t147_n190: return 0.0948072001338005f;
  t147_n191: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t147_n215;
  t147_n192: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t147_n214;
  t147_n193: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t147_n197;
  t147_n194: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t147_n196;
  t147_n195: return -0.0030248481780290604f;
  t147_n196: return 0.12083381414413452f;
  t147_n197: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t147_n213;
  t147_n198: if (E.getNumReferences() >= 3302543360 /*1734.5*/) goto t147_n212;
  t147_n199: if (E.getNumReferences() >= 3300020224 /*1426.5*/) goto t147_n211;
  t147_n200: if (E.getNumReferences() >= 3277881344 /*224.5*/) goto t147_n202;
  t147_n201: return 0.021735427901148796f;
  t147_n202: if (E.getNumReferences() >= 3289767936 /*599.5*/) goto t147_n210;
  t147_n203: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t147_n205;
  t147_n204: return 0.0801914632320404f;
  t147_n205: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t147_n209;
  t147_n206: if (E.getNumReferences() >= 3285745664 /*433.0*/) goto t147_n208;
  t147_n207: return -0.24231795966625214f;
  t147_n208: return 0.11575203388929367f;
  t147_n209: return 0.1229216605424881f;
  t147_n210: return -0.24516814947128296f;
  t147_n211: return 0.20718707144260406f;
  t147_n212: return -0.13975507020950317f;
  t147_n213: return 0.00830183457583189f;
  t147_n214: return 0.010259544476866722f;
  t147_n215: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t147_n229;
  t147_n216: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t147_n228;
  t147_n217: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t147_n219;
  t147_n218: return 0.014597962610423565f;
  t147_n219: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t147_n221;
  t147_n220: return 0.03210273012518883f;
  t147_n221: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t147_n223;
  t147_n222: return 0.042028218507766724f;
  t147_n223: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t147_n225;
  t147_n224: return 0.07716909050941467f;
  t147_n225: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t147_n227;
  t147_n226: return 0.09742885828018188f;
  t147_n227: return 0.028043542057275772f;
  t147_n228: return -0.01601843535900116f;
  t147_n229: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t147_n239;
  t147_n230: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t147_n232;
  t147_n231: return 0.003269449109211564f;
  t147_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t147_n234;
  t147_n233: return -0.27305006980895996f;
  t147_n234: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t147_n238;
  t147_n235: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t147_n237;
  t147_n236: return -0.22223569452762604f;
  t147_n237: return 0.008399776183068752f;
  t147_n238: return 0.006282859481871128f;
  t147_n239: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t147_n249;
  t147_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t147_n248;
  t147_n241: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t147_n247;
  t147_n242: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t147_n244;
  t147_n243: return 0.09175769984722137f;
  t147_n244: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t147_n246;
  t147_n245: return 0.11525119841098785f;
  t147_n246: return 0.14617490768432617f;
  t147_n247: return -0.07815969735383987f;
  t147_n248: return 0.16584645211696625f;
  t147_n249: return 0.03228059783577919f;
  t147_n250: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t147_n252;
  t147_n251: return 0.009223620407283306f;
  t147_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t147_n254;
  t147_n253: return 0.05330098047852516f;
  t147_n254: return 0.02269108220934868f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree148(const Example& E) {
  t148_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t148_n250;
  t148_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t148_n3;
  t148_n2: return -0.09967417269945145f;
  t148_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t148_n69;
  t148_n4: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t148_n62;
  t148_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n61;
  t148_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t148_n16;
  t148_n7: if (E.getNumReferences() >= 3274211328 /*168.5*/) goto t148_n13;
  t148_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t148_n12;
  t148_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t148_n11;
  t148_n10: return 0.002069602720439434f;
  t148_n11: return 0.0427560955286026f;
  t148_n12: return 0.02144080214202404f;
  t148_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t148_n15;
  t148_n14: return 0.05910192430019379f;
  t148_n15: return 0.14706678688526154f;
  t148_n16: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t148_n32;
  t148_n17: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t148_n31;
  t148_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t148_n28;
  t148_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t148_n27;
  t148_n20: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t148_n22;
  t148_n21: return -0.022164637222886086f;
  t148_n22: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t148_n24;
  t148_n23: return -0.37612494826316833f;
  t148_n24: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t148_n26;
  t148_n25: return 0.21402689814567566f;
  t148_n26: return 0.4053155779838562f;
  t148_n27: return -0.003577129216864705f;
  t148_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t148_n30;
  t148_n29: return 0.0252196304500103f;
  t148_n30: return 0.04558214917778969f;
  t148_n31: return -0.0031068939715623856f;
  t148_n32: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t148_n44;
  t148_n33: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t148_n43;
  t148_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n40;
  t148_n35: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t148_n37;
  t148_n36: return -0.06288930773735046f;
  t148_n37: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t148_n39;
  t148_n38: return 0.08215714246034622f;
  t148_n39: return -0.021580971777439117f;
  t148_n40: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t148_n42;
  t148_n41: return 0.010827062651515007f;
  t148_n42: return -0.014591047540307045f;
  t148_n43: return 0.03858748450875282f;
  t148_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t148_n60;
  t148_n45: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t148_n49;
  t148_n46: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t148_n48;
  t148_n47: return -0.13187934458255768f;
  t148_n48: return -0.017289847135543823f;
  t148_n49: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t148_n59;
  t148_n50: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t148_n56;
  t148_n51: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t148_n53;
  t148_n52: return -0.344868540763855f;
  t148_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t148_n55;
  t148_n54: return -0.02484997920691967f;
  t148_n55: return 0.15740245580673218f;
  t148_n56: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t148_n58;
  t148_n57: return 0.16303503513336182f;
  t148_n58: return -0.01275172084569931f;
  t148_n59: return 0.08198849856853485f;
  t148_n60: return -0.006581307854503393f;
  t148_n61: return 0.0011182877933606505f;
  t148_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n66;
  t148_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t148_n65;
  t148_n64: return -0.028071871027350426f;
  t148_n65: return 0.01058058999478817f;
  t148_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t148_n68;
  t148_n67: return 0.0123204430565238f;
  t148_n68: return -0.0037493810523301363f;
  t148_n69: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t148_n71;
  t148_n70: return 0.0055158804170787334f;
  t148_n71: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t148_n247;
  t148_n72: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t148_n212;
  t148_n73: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t148_n209;
  t148_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n82;
  t148_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t148_n77;
  t148_n76: return -0.008885440416634083f;
  t148_n77: if (E.getNumReferences() >= 3356472608 /*147156.5*/) goto t148_n79;
  t148_n78: return 0.007201476022601128f;
  t148_n79: if (E.getNumReferences() >= 3359621216 /*196353.5*/) goto t148_n81;
  t148_n80: return 0.07202785462141037f;
  t148_n81: return 0.019609130918979645f;
  t148_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t148_n92;
  t148_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n85;
  t148_n84: return -0.08487731963396072f;
  t148_n85: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t148_n91;
  t148_n86: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t148_n88;
  t148_n87: return -0.018595224246382713f;
  t148_n88: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t148_n90;
  t148_n89: return 0.019376607611775398f;
  t148_n90: return 0.08542213588953018f;
  t148_n91: return -0.011557084508240223f;
  t148_n92: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t148_n208;
  t148_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t148_n149;
  t148_n94: if (E.getNumReferences() >= 3319640064 /*7094.0*/) goto t148_n134;
  t148_n95: if (E.getNumReferences() >= 3319071744 /*6816.5*/) goto t148_n133;
  t148_n96: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t148_n116;
  t148_n97: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t148_n113;
  t148_n98: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t148_n106;
  t148_n99: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n101;
  t148_n100: return -0.0012308405712246895f;
  t148_n101: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t148_n105;
  t148_n102: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t148_n104;
  t148_n103: return 0.010369255207479f;
  t148_n104: return -0.01923431269824505f;
  t148_n105: return 0.01536529790610075f;
  t148_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n108;
  t148_n107: return -0.016458973288536072f;
  t148_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t148_n112;
  t148_n109: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t148_n111;
  t148_n110: return -0.01589646376669407f;
  t148_n111: return -0.28443872928619385f;
  t148_n112: return 0.023363469168543816f;
  t148_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n115;
  t148_n114: return -0.004564268514513969f;
  t148_n115: return 0.009639609605073929f;
  t148_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n118;
  t148_n117: return -0.01804637350142002f;
  t148_n118: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t148_n120;
  t148_n119: return -0.07796660810709f;
  t148_n120: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t148_n130;
  t148_n121: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t148_n129;
  t148_n122: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t148_n124;
  t148_n123: return 0.07675370573997498f;
  t148_n124: if (E.getNumReferences() >= 3270508544 /*120.0*/) goto t148_n128;
  t148_n125: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t148_n127;
  t148_n126: return 1.096040964126587f;
  t148_n127: return 0.029678992927074432f;
  t148_n128: return 0.4864813983440399f;
  t148_n129: return 0.020828744396567345f;
  t148_n130: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t148_n132;
  t148_n131: return -0.05952337756752968f;
  t148_n132: return 0.020030977204442024f;
  t148_n133: return -0.13123847544193268f;
  t148_n134: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t148_n142;
  t148_n135: if (E.getNumReferences() >= 3347329664 /*67685.0*/) goto t148_n141;
  t148_n136: if (E.getNumReferences() >= 3342301952 /*46971.0*/) goto t148_n140;
  t148_n137: if (E.getNumReferences() >= 3342115200 /*46241.5*/) goto t148_n139;
  t148_n138: return 0.014521090313792229f;
  t148_n139: return 0.073260098695755f;
  t148_n140: return -0.04067988321185112f;
  t148_n141: return 0.02745814435184002f;
  t148_n142: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t148_n144;
  t148_n143: return -0.04026688262820244f;
  t148_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n148;
  t148_n145: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t148_n147;
  t148_n146: return 0.08281929045915604f;
  t148_n147: return -0.22779594361782074f;
  t148_n148: return 0.02021307870745659f;
  t148_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n151;
  t148_n150: return -0.11793400347232819f;
  t148_n151: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t148_n157;
  t148_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n156;
  t148_n153: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t148_n155;
  t148_n154: return 0.04417034983634949f;
  t148_n155: return -0.04662192612886429f;
  t148_n156: return 0.0023176283575594425f;
  t148_n157: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t148_n167;
  t148_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n162;
  t148_n159: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t148_n161;
  t148_n160: return -0.07264979183673859f;
  t148_n161: return 0.04731893911957741f;
  t148_n162: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t148_n166;
  t148_n163: if (E.getNumReferences() >= 3316233216 /*5430.5*/) goto t148_n165;
  t148_n164: return 0.03885957598686218f;
  t148_n165: return -0.010742568410933018f;
  t148_n166: return -0.0014783326769247651f;
  t148_n167: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t148_n203;
  t148_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t148_n186;
  t148_n169: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t148_n185;
  t148_n170: if (E.getNumReferences() >= 3340176384 /*38668.0*/) goto t148_n184;
  t148_n171: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t148_n179;
  t148_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n176;
  t148_n173: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t148_n175;
  t148_n174: return -0.0097710145637393f;
  t148_n175: return 0.01059426087886095f;
  t148_n176: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t148_n178;
  t148_n177: return 0.058495599776506424f;
  t148_n178: return -0.015586654655635357f;
  t148_n179: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t148_n183;
  t148_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n182;
  t148_n181: return -0.1107129231095314f;
  t148_n182: return 0.014786962419748306f;
  t148_n183: return 0.003453355748206377f;
  t148_n184: return 0.023525036871433258f;
  t148_n185: return 0.03780077025294304f;
  t148_n186: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t148_n202;
  t148_n187: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t148_n197;
  t148_n188: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t148_n196;
  t148_n189: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t148_n195;
  t148_n190: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t148_n194;
  t148_n191: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t148_n193;
  t148_n192: return 0.017403723672032356f;
  t148_n193: return -0.02895928919315338f;
  t148_n194: return 0.02328232303261757f;
  t148_n195: return -0.0774620771408081f;
  t148_n196: return 0.039677515625953674f;
  t148_n197: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t148_n199;
  t148_n198: return -0.07974626868963242f;
  t148_n199: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t148_n201;
  t148_n200: return -0.01160360500216484f;
  t148_n201: return 0.024418644607067108f;
  t148_n202: return 0.07092142105102539f;
  t148_n203: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t148_n207;
  t148_n204: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t148_n206;
  t148_n205: return -1.4247578382492065f;
  t148_n206: return -0.011300225742161274f;
  t148_n207: return 0.024614909663796425f;
  t148_n208: return -0.01782984845340252f;
  t148_n209: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t148_n211;
  t148_n210: return 0.059919573366642f;
  t148_n211: return 0.10073927789926529f;
  t148_n212: if (E.getNumReferences() >= 3344576000 /*55854.0*/) goto t148_n246;
  t148_n213: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t148_n243;
  t148_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n220;
  t148_n215: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t148_n217;
  t148_n216: return -0.038834888488054276f;
  t148_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t148_n219;
  t148_n218: return -0.10324253886938095f;
  t148_n219: return 0.10496927052736282f;
  t148_n220: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t148_n222;
  t148_n221: return -0.048116810619831085f;
  t148_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t148_n224;
  t148_n223: return -0.0005919380346313119f;
  t148_n224: if (E.getNumReferences() >= 3327147008 /*13327.0*/) goto t148_n228;
  t148_n225: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t148_n227;
  t148_n226: return 0.0613262765109539f;
  t148_n227: return -0.01737404242157936f;
  t148_n228: if (E.getNumReferences() >= 3327212032 /*13390.5*/) goto t148_n230;
  t148_n229: return 0.2127271294593811f;
  t148_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t148_n242;
  t148_n231: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t148_n241;
  t148_n232: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t148_n240;
  t148_n233: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t148_n239;
  t148_n234: if (E.getNumReferences() >= 3331260928 /*18305.0*/) goto t148_n236;
  t148_n235: return 0.0657215416431427f;
  t148_n236: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t148_n238;
  t148_n237: return 0.15651339292526245f;
  t148_n238: return 0.09267889708280563f;
  t148_n239: return 0.05937838926911354f;
  t148_n240: return 0.18126371502876282f;
  t148_n241: return -0.12184717506170273f;
  t148_n242: return -0.04131479188799858f;
  t148_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n245;
  t148_n244: return -0.17302605509757996f;
  t148_n245: return 0.13246478140354156f;
  t148_n246: return -0.027100956067442894f;
  t148_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t148_n249;
  t148_n248: return -0.030201684683561325f;
  t148_n249: return 0.010369128547608852f;
  t148_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t148_n252;
  t148_n251: return 0.0022606831043958664f;
  t148_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t148_n254;
  t148_n253: return 0.006299193482846022f;
  t148_n254: return -0.07027756422758102f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree149(const Example& E) {
  t149_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t149_n230;
  t149_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t149_n195;
  t149_n2: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t149_n190;
  t149_n3: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t149_n175;
  t149_n4: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t149_n172;
  t149_n5: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t149_n159;
  t149_n6: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t149_n158;
  t149_n7: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t149_n9;
  t149_n8: return -0.0009195612510666251f;
  t149_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t149_n157;
  t149_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t149_n108;
  t149_n11: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t149_n101;
  t149_n12: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t149_n82;
  t149_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t149_n15;
  t149_n14: return -0.00222615129314363f;
  t149_n15: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t149_n31;
  t149_n16: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t149_n30;
  t149_n17: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t149_n27;
  t149_n18: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t149_n26;
  t149_n19: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t149_n21;
  t149_n20: return -0.026678450405597687f;
  t149_n21: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t149_n25;
  t149_n22: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t149_n24;
  t149_n23: return -0.03686126321554184f;
  t149_n24: return 0.10651059448719025f;
  t149_n25: return 0.010693583637475967f;
  t149_n26: return -0.06479347497224808f;
  t149_n27: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t149_n29;
  t149_n28: return 0.0038421000353991985f;
  t149_n29: return -0.07001537084579468f;
  t149_n30: return 0.019531385973095894f;
  t149_n31: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t149_n81;
  t149_n32: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t149_n64;
  t149_n33: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t149_n35;
  t149_n34: return -0.26415854692459106f;
  t149_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t149_n39;
  t149_n36: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t149_n38;
  t149_n37: return -0.17430199682712555f;
  t149_n38: return 0.022766586393117905f;
  t149_n39: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t149_n45;
  t149_n40: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t149_n44;
  t149_n41: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t149_n43;
  t149_n42: return 0.009476685896515846f;
  t149_n43: return -0.08745656907558441f;
  t149_n44: return -0.09093592315912247f;
  t149_n45: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t149_n63;
  t149_n46: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t149_n56;
  t149_n47: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t149_n51;
  t149_n48: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t149_n50;
  t149_n49: return -0.07738576084375381f;
  t149_n50: return 0.033066824078559875f;
  t149_n51: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t149_n55;
  t149_n52: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t149_n54;
  t149_n53: return 0.14007917046546936f;
  t149_n54: return 0.25796881318092346f;
  t149_n55: return 0.026861920952796936f;
  t149_n56: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t149_n60;
  t149_n57: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t149_n59;
  t149_n58: return -0.0019431646214798093f;
  t149_n59: return -0.12488644570112228f;
  t149_n60: if (E.getFractionNameInContext() >= 3200453681 /*0.380952388048172*/) goto t149_n62;
  t149_n61: return 0.06086626648902893f;
  t149_n62: return 0.12536020576953888f;
  t149_n63: return 0.050757091492414474f;
  t149_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t149_n76;
  t149_n65: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t149_n67;
  t149_n66: return -0.3679274618625641f;
  t149_n67: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t149_n71;
  t149_n68: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t149_n70;
  t149_n69: return -0.26061660051345825f;
  t149_n70: return 0.07054900377988815f;
  t149_n71: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t149_n75;
  t149_n72: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t149_n74;
  t149_n73: return -0.17314067482948303f;
  t149_n74: return 0.13829660415649414f;
  t149_n75: return -0.09401166439056396f;
  t149_n76: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t149_n80;
  t149_n77: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t149_n79;
  t149_n78: return 0.13262681663036346f;
  t149_n79: return 0.0771440863609314f;
  t149_n80: return -0.006013157777488232f;
  t149_n81: return 0.03395109251141548f;
  t149_n82: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t149_n94;
  t149_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t149_n89;
  t149_n84: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t149_n88;
  t149_n85: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t149_n87;
  t149_n86: return 0.17793823778629303f;
  t149_n87: return -0.06073678657412529f;
  t149_n88: return -0.3171694874763489f;
  t149_n89: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t149_n91;
  t149_n90: return 0.050160326063632965f;
  t149_n91: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t149_n93;
  t149_n92: return 0.01554340124130249f;
  t149_n93: return -0.09362412244081497f;
  t149_n94: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t149_n96;
  t149_n95: return -0.277998149394989f;
  t149_n96: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t149_n100;
  t149_n97: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t149_n99;
  t149_n98: return 0.0928230732679367f;
  t149_n99: return 0.15917040407657623f;
  t149_n100: return 0.026064207777380943f;
  t149_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t149_n107;
  t149_n102: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t149_n104;
  t149_n103: return -0.018566399812698364f;
  t149_n104: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t149_n106;
  t149_n105: return -0.11717040091753006f;
  t149_n106: return 0.06946773827075958f;
  t149_n107: return -0.006111096125096083f;
  t149_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t149_n122;
  t149_n109: if (E.getNumReferences() >= 3349388864 /*83772.5*/) goto t149_n121;
  t149_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t149_n120;
  t149_n111: if (E.getNumReferences() >= 3324926976 /*11159.0*/) goto t149_n119;
  t149_n112: if (E.getNumReferences() >= 3324764160 /*11000.0*/) goto t149_n118;
  t149_n113: if (E.getNumReferences() >= 3324302336 /*10549.0*/) goto t149_n115;
  t149_n114: return -0.013822401873767376f;
  t149_n115: if (E.getNumReferences() >= 3324478976 /*10721.5*/) goto t149_n117;
  t149_n116: return 0.20249083638191223f;
  t149_n117: return -0.04559123516082764f;
  t149_n118: return 0.1230374276638031f;
  t149_n119: return -0.06605308502912521f;
  t149_n120: return 0.01570790633559227f;
  t149_n121: return 0.06677619367837906f;
  t149_n122: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t149_n126;
  t149_n123: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t149_n125;
  t149_n124: return -1.3456543684005737f;
  t149_n125: return 0.012637274339795113f;
  t149_n126: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t149_n134;
  t149_n127: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t149_n129;
  t149_n128: return 0.0062681264244019985f;
  t149_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t149_n131;
  t149_n130: return -0.06784844398498535f;
  t149_n131: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t149_n133;
  t149_n132: return -0.00863630324602127f;
  t149_n133: return -0.08802974224090576f;
  t149_n134: if (E.getNumReferences() >= 3299512320 /*1364.5*/) goto t149_n156;
  t149_n135: if (E.getNumReferences() >= 3299278848 /*1336.0*/) goto t149_n151;
  t149_n136: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t149_n138;
  t149_n137: return -0.03257418051362038f;
  t149_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t149_n146;
  t149_n139: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t149_n143;
  t149_n140: if (E.getNumReferences() >= 3298889728 /*1288.5*/) goto t149_n142;
  t149_n141: return 0.08781860023736954f;
  t149_n142: return -0.10358240455389023f;
  t149_n143: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t149_n145;
  t149_n144: return 0.12507176399230957f;
  t149_n145: return -0.0010570194572210312f;
  t149_n146: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t149_n148;
  t149_n147: return 0.039183177053928375f;
  t149_n148: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t149_n150;
  t149_n149: return 0.32659679651260376f;
  t149_n150: return 0.10175497829914093f;
  t149_n151: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t149_n153;
  t149_n152: return 0.12174196541309357f;
  t149_n153: if (E.getNumReferences() >= 3299307520 /*1339.5*/) goto t149_n155;
  t149_n154: return 0.14516830444335938f;
  t149_n155: return 0.059716127812862396f;
  t149_n156: return 0.013017409481108189f;
  t149_n157: return 0.0020165604073554277f;
  t149_n158: return -0.1794997602701187f;
  t149_n159: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t149_n163;
  t149_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t149_n162;
  t149_n161: return 0.0284457728266716f;
  t149_n162: return 0.06537684798240662f;
  t149_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t149_n165;
  t149_n164: return -0.16265924274921417f;
  t149_n165: if (E.getNumReferences() >= 3358250528 /*174936.5*/) goto t149_n167;
  t149_n166: return -0.15658767521381378f;
  t149_n167: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t149_n169;
  t149_n168: return 0.12414222955703735f;
  t149_n169: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t149_n171;
  t149_n170: return 0.21214967966079712f;
  t149_n171: return 0.15678027272224426f;
  t149_n172: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t149_n174;
  t149_n173: return -0.14915108680725098f;
  t149_n174: return -2.9813332557678223f;
  t149_n175: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t149_n177;
  t149_n176: return 0.025747383013367653f;
  t149_n177: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t149_n185;
  t149_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t149_n182;
  t149_n179: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t149_n181;
  t149_n180: return -0.18643558025360107f;
  t149_n181: return -0.9263914227485657f;
  t149_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t149_n184;
  t149_n183: return -0.055788129568099976f;
  t149_n184: return 0.08303964138031006f;
  t149_n185: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t149_n189;
  t149_n186: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t149_n188;
  t149_n187: return 0.00706545589491725f;
  t149_n188: return 0.04894235357642174f;
  t149_n189: return -0.10898908972740173f;
  t149_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t149_n194;
  t149_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t149_n193;
  t149_n192: return -0.6126480102539062f;
  t149_n193: return -0.09925256669521332f;
  t149_n194: return -0.049297481775283813f;
  t149_n195: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t149_n217;
  t149_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t149_n206;
  t149_n197: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t149_n199;
  t149_n198: return 0.040179044008255005f;
  t149_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t149_n201;
  t149_n200: return -0.4677855372428894f;
  t149_n201: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t149_n205;
  t149_n202: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t149_n204;
  t149_n203: return -0.004233391024172306f;
  t149_n204: return -0.12290338426828384f;
  t149_n205: return 0.05127359926700592f;
  t149_n206: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t149_n216;
  t149_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t149_n209;
  t149_n208: return 0.023807864636182785f;
  t149_n209: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t149_n211;
  t149_n210: return 0.05584844574332237f;
  t149_n211: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t149_n213;
  t149_n212: return 0.07241406291723251f;
  t149_n213: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t149_n215;
  t149_n214: return 0.11657792329788208f;
  t149_n215: return 0.0728551372885704f;
  t149_n216: return 0.004097227938473225f;
  t149_n217: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t149_n219;
  t149_n218: return -0.07762499153614044f;
  t149_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t149_n229;
  t149_n220: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t149_n226;
  t149_n221: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t149_n225;
  t149_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t149_n224;
  t149_n223: return -0.4425405263900757f;
  t149_n224: return -0.07861223071813583f;
  t149_n225: return -0.011501367203891277f;
  t149_n226: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t149_n228;
  t149_n227: return -0.05973733589053154f;
  t149_n228: return 0.010438401252031326f;
  t149_n229: return 0.04431630298495293f;
  t149_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t149_n250;
  t149_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t149_n233;
  t149_n232: return -0.12830102443695068f;
  t149_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t149_n249;
  t149_n234: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t149_n248;
  t149_n235: if (E.getNumReferences() >= 3288121344 /*505.5*/) goto t149_n247;
  t149_n236: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t149_n242;
  t149_n237: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t149_n239;
  t149_n238: return 0.011680318973958492f;
  t149_n239: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t149_n241;
  t149_n240: return 0.12136049568653107f;
  t149_n241: return 0.06400255858898163f;
  t149_n242: if (E.getNumReferences() >= 3265265664 /*80.0*/) goto t149_n246;
  t149_n243: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t149_n245;
  t149_n244: return -0.07041449099779129f;
  t149_n245: return -0.3481118679046631f;
  t149_n246: return 0.05834675207734108f;
  t149_n247: return 0.12747251987457275f;
  t149_n248: return 0.00040866131894290447f;
  t149_n249: return 0.010374044068157673f;
  t149_n250: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t149_n254;
  t149_n251: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t149_n253;
  t149_n252: return 0.00623029051348567f;
  t149_n253: return -0.05569540336728096f;
  t149_n254: return -0.18095622956752777f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree150(const Example& E) {
  t150_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t150_n164;
  t150_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t150_n161;
  t150_n2: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t150_n108;
  t150_n3: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t150_n99;
  t150_n4: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t150_n78;
  t150_n5: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t150_n7;
  t150_n6: return -0.0012676608748733997f;
  t150_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t150_n73;
  t150_n8: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t150_n44;
  t150_n9: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t150_n29;
  t150_n10: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t150_n16;
  t150_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t150_n15;
  t150_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t150_n14;
  t150_n13: return -0.09203261882066727f;
  t150_n14: return 0.08449176698923111f;
  t150_n15: return -0.09505528956651688f;
  t150_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t150_n18;
  t150_n17: return 0.00030285396496765316f;
  t150_n18: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t150_n26;
  t150_n19: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t150_n21;
  t150_n20: return 0.0011332364520058036f;
  t150_n21: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t150_n23;
  t150_n22: return 0.033499203622341156f;
  t150_n23: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t150_n25;
  t150_n24: return 0.0038577918894588947f;
  t150_n25: return -0.043419815599918365f;
  t150_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t150_n28;
  t150_n27: return 0.0887235477566719f;
  t150_n28: return -0.0008691111579537392f;
  t150_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n31;
  t150_n30: return -0.024708887562155724f;
  t150_n31: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t150_n41;
  t150_n32: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t150_n34;
  t150_n33: return 0.002396391471847892f;
  t150_n34: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t150_n38;
  t150_n35: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t150_n37;
  t150_n36: return 0.04368405044078827f;
  t150_n37: return -0.043639909476041794f;
  t150_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n40;
  t150_n39: return -0.10623679310083389f;
  t150_n40: return 0.08937536925077438f;
  t150_n41: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t150_n43;
  t150_n42: return 0.08615967631340027f;
  t150_n43: return 0.005401114467531443f;
  t150_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t150_n66;
  t150_n45: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t150_n47;
  t150_n46: return -0.0024607849773019552f;
  t150_n47: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t150_n55;
  t150_n48: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t150_n54;
  t150_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t150_n51;
  t150_n50: return 0.05015811324119568f;
  t150_n51: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t150_n53;
  t150_n52: return 0.13418011367321014f;
  t150_n53: return 0.16313962638378143f;
  t150_n54: return 0.13117073476314545f;
  t150_n55: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t150_n57;
  t150_n56: return -0.1703130006790161f;
  t150_n57: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t150_n63;
  t150_n58: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t150_n62;
  t150_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n61;
  t150_n60: return -0.12238463759422302f;
  t150_n61: return 0.15374647080898285f;
  t150_n62: return 0.12952488660812378f;
  t150_n63: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t150_n65;
  t150_n64: return 0.13854220509529114f;
  t150_n65: return -0.3373187482357025f;
  t150_n66: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t150_n68;
  t150_n67: return -0.0016282513970509171f;
  t150_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t150_n70;
  t150_n69: return -0.00015802263806108385f;
  t150_n70: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t150_n72;
  t150_n71: return 0.011164703406393528f;
  t150_n72: return -0.07187952846288681f;
  t150_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t150_n75;
  t150_n74: return 0.002899224404245615f;
  t150_n75: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t150_n77;
  t150_n76: return 0.08430734276771545f;
  t150_n77: return 0.04010468348860741f;
  t150_n78: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t150_n90;
  t150_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n89;
  t150_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t150_n84;
  t150_n81: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t150_n83;
  t150_n82: return -0.05428055673837662f;
  t150_n83: return -0.08644241839647293f;
  t150_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n86;
  t150_n85: return -0.05994950979948044f;
  t150_n86: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t150_n88;
  t150_n87: return 0.021164949983358383f;
  t150_n88: return 0.08574125915765762f;
  t150_n89: return 0.03679153323173523f;
  t150_n90: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t150_n96;
  t150_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n93;
  t150_n92: return -0.017819661647081375f;
  t150_n93: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t150_n95;
  t150_n94: return 0.009522300213575363f;
  t150_n95: return 0.06265182793140411f;
  t150_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n98;
  t150_n97: return -0.023341331630945206f;
  t150_n98: return -0.0023001430090516806f;
  t150_n99: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t150_n105;
  t150_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t150_n104;
  t150_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n103;
  t150_n102: return -0.15600553154945374f;
  t150_n103: return 0.005190331023186445f;
  t150_n104: return 0.03207132965326309f;
  t150_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n107;
  t150_n106: return -0.2430875599384308f;
  t150_n107: return 0.06214151531457901f;
  t150_n108: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t150_n118;
  t150_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t150_n117;
  t150_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n114;
  t150_n111: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t150_n113;
  t150_n112: return 0.06565792113542557f;
  t150_n113: return -0.01786581613123417f;
  t150_n114: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t150_n116;
  t150_n115: return 0.009173772297799587f;
  t150_n116: return -0.12479384988546371f;
  t150_n117: return 0.24085503816604614f;
  t150_n118: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t150_n122;
  t150_n119: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t150_n121;
  t150_n120: return 0.005370166152715683f;
  t150_n121: return -0.016254259273409843f;
  t150_n122: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t150_n124;
  t150_n123: return 0.045905593782663345f;
  t150_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t150_n154;
  t150_n125: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t150_n151;
  t150_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n128;
  t150_n127: return -0.03764724358916283f;
  t150_n128: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t150_n148;
  t150_n129: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t150_n131;
  t150_n130: return 0.028022460639476776f;
  t150_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t150_n133;
  t150_n132: return -0.10443184524774551f;
  t150_n133: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t150_n147;
  t150_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n136;
  t150_n135: return -0.10066299140453339f;
  t150_n136: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t150_n138;
  t150_n137: return -0.028586482629179955f;
  t150_n138: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t150_n140;
  t150_n139: return 0.2852506935596466f;
  t150_n140: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t150_n144;
  t150_n141: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t150_n143;
  t150_n142: return -0.10041157901287079f;
  t150_n143: return 0.25701844692230225f;
  t150_n144: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t150_n146;
  t150_n145: return 0.50511234998703f;
  t150_n146: return 0.17227940261363983f;
  t150_n147: return -0.1339842826128006f;
  t150_n148: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t150_n150;
  t150_n149: return -0.005081541370600462f;
  t150_n150: return 0.3645973801612854f;
  t150_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t150_n153;
  t150_n152: return -0.002702693920582533f;
  t150_n153: return 0.014270873740315437f;
  t150_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t150_n156;
  t150_n155: return 0.013897761702537537f;
  t150_n156: if (E.getNumReferences() >= 3334782208 /*25182.5*/) goto t150_n160;
  t150_n157: if (E.getNumReferences() >= 3334246656 /*24136.5*/) goto t150_n159;
  t150_n158: return 0.1022026389837265f;
  t150_n159: return 0.10371380299329758f;
  t150_n160: return -0.14148543775081635f;
  t150_n161: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t150_n163;
  t150_n162: return 0.004290095996111631f;
  t150_n163: return -0.030434058979153633f;
  t150_n164: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t150_n208;
  t150_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t150_n205;
  t150_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t150_n192;
  t150_n167: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t150_n179;
  t150_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t150_n176;
  t150_n169: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t150_n171;
  t150_n170: return -0.006331433542072773f;
  t150_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n175;
  t150_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t150_n174;
  t150_n173: return -0.05898601561784744f;
  t150_n174: return -0.024681296199560165f;
  t150_n175: return 0.0029712512623518705f;
  t150_n176: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t150_n178;
  t150_n177: return 0.07878232002258301f;
  t150_n178: return -0.001845017890445888f;
  t150_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t150_n183;
  t150_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n182;
  t150_n181: return -0.01963634043931961f;
  t150_n182: return 0.007469095289707184f;
  t150_n183: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t150_n189;
  t150_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t150_n188;
  t150_n185: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t150_n187;
  t150_n186: return 0.00828090775758028f;
  t150_n187: return -0.06531242281198502f;
  t150_n188: return 0.09862474352121353f;
  t150_n189: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t150_n191;
  t150_n190: return 0.07499300688505173f;
  t150_n191: return 0.02520165592432022f;
  t150_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t150_n194;
  t150_n193: return 0.010201769880950451f;
  t150_n194: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t150_n196;
  t150_n195: return 0.041928380727767944f;
  t150_n196: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t150_n204;
  t150_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t150_n203;
  t150_n198: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t150_n202;
  t150_n199: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t150_n201;
  t150_n200: return 0.1575450450181961f;
  t150_n201: return 0.16306638717651367f;
  t150_n202: return 0.006924634333699942f;
  t150_n203: return 0.07657919079065323f;
  t150_n204: return 0.05268778279423714f;
  t150_n205: if (E.getFileProximityDistanceCost() >= 3238002688 /*8.0*/) goto t150_n207;
  t150_n206: return 0.1827218383550644f;
  t150_n207: return 0.04074990004301071f;
  t150_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t150_n224;
  t150_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t150_n223;
  t150_n210: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t150_n222;
  t150_n211: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t150_n219;
  t150_n212: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t150_n218;
  t150_n213: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t150_n217;
  t150_n214: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t150_n216;
  t150_n215: return 0.139981210231781f;
  t150_n216: return 0.0030543310567736626f;
  t150_n217: return -0.008875560015439987f;
  t150_n218: return -0.007845213636755943f;
  t150_n219: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t150_n221;
  t150_n220: return -0.11357630789279938f;
  t150_n221: return 0.11328276991844177f;
  t150_n222: return -0.032982807606458664f;
  t150_n223: return 0.0008352893637493253f;
  t150_n224: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t150_n242;
  t150_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t150_n227;
  t150_n226: return -0.017664430662989616f;
  t150_n227: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t150_n235;
  t150_n228: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t150_n232;
  t150_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t150_n231;
  t150_n230: return -0.06604741513729095f;
  t150_n231: return 0.040696825832128525f;
  t150_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t150_n234;
  t150_n233: return 0.07232290506362915f;
  t150_n234: return 0.1310470849275589f;
  t150_n235: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t150_n241;
  t150_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t150_n240;
  t150_n237: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t150_n239;
  t150_n238: return -0.18157629668712616f;
  t150_n239: return -0.3995767831802368f;
  t150_n240: return -0.0816180482506752f;
  t150_n241: return 0.020054398104548454f;
  t150_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t150_n244;
  t150_n243: return 0.008648165501654148f;
  t150_n244: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t150_n246;
  t150_n245: return 0.008009188808500767f;
  t150_n246: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t150_n248;
  t150_n247: return 0.023681387305259705f;
  t150_n248: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t150_n252;
  t150_n249: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t150_n251;
  t150_n250: return 0.000481467111967504f;
  t150_n251: return -0.09614996612071991f;
  t150_n252: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t150_n254;
  t150_n253: return 0.027495920658111572f;
  t150_n254: return 0.09286915510892868f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree151(const Example& E) {
  t151_n0: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t151_n254;
  t151_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t151_n235;
  t151_n2: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t151_n220;
  t151_n3: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t151_n219;
  t151_n4: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t151_n208;
  t151_n5: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t151_n207;
  t151_n6: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t151_n196;
  t151_n7: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t151_n195;
  t151_n8: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t151_n58;
  t151_n9: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t151_n25;
  t151_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n24;
  t151_n11: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t151_n13;
  t151_n12: return -0.0036205248907208443f;
  t151_n13: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t151_n21;
  t151_n14: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t151_n20;
  t151_n15: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t151_n19;
  t151_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t151_n18;
  t151_n17: return 0.018394984304904938f;
  t151_n18: return 0.08605500310659409f;
  t151_n19: return 0.04318183660507202f;
  t151_n20: return 0.0029480610974133015f;
  t151_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n23;
  t151_n22: return -0.031976114958524704f;
  t151_n23: return 0.005231150425970554f;
  t151_n24: return 0.0010435703443363309f;
  t151_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t151_n47;
  t151_n26: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t151_n32;
  t151_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n31;
  t151_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t151_n30;
  t151_n29: return -0.009450162760913372f;
  t151_n30: return 0.17252491414546967f;
  t151_n31: return 0.010923215188086033f;
  t151_n32: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t151_n38;
  t151_n33: if (E.getNumReferences() >= 3351224064 /*98110.0*/) goto t151_n35;
  t151_n34: return 0.05048010125756264f;
  t151_n35: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t151_n37;
  t151_n36: return 0.11790791153907776f;
  t151_n37: return 0.05871541053056717f;
  t151_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t151_n40;
  t151_n39: return 0.004510982893407345f;
  t151_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t151_n44;
  t151_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n43;
  t151_n42: return -0.0863095372915268f;
  t151_n43: return 0.10387858748435974f;
  t151_n44: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t151_n46;
  t151_n45: return 0.10943222045898438f;
  t151_n46: return -0.41116416454315186f;
  t151_n47: if (E.getNumReferences() >= 3344297088 /*54764.5*/) goto t151_n57;
  t151_n48: if (E.getNumReferences() >= 3341367040 /*43319.0*/) goto t151_n50;
  t151_n49: return -0.003674954641610384f;
  t151_n50: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t151_n56;
  t151_n51: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t151_n53;
  t151_n52: return 0.09507540613412857f;
  t151_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t151_n55;
  t151_n54: return 0.10074900835752487f;
  t151_n55: return 0.147505521774292f;
  t151_n56: return 0.02398291788995266f;
  t151_n57: return -0.047405071556568146f;
  t151_n58: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t151_n82;
  t151_n59: if (E.getNumReferences() >= 3318784000 /*6676.0*/) goto t151_n81;
  t151_n60: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t151_n80;
  t151_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t151_n73;
  t151_n62: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t151_n64;
  t151_n63: return 0.0062373545952141285f;
  t151_n64: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t151_n72;
  t151_n65: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t151_n67;
  t151_n66: return -0.012102169916033745f;
  t151_n67: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t151_n71;
  t151_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t151_n70;
  t151_n69: return 0.049269113689661026f;
  t151_n70: return 0.17559465765953064f;
  t151_n71: return -0.34527185559272766f;
  t151_n72: return 0.11808965355157852f;
  t151_n73: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t151_n79;
  t151_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t151_n78;
  t151_n75: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t151_n77;
  t151_n76: return 0.008362199179828167f;
  t151_n77: return 0.06486136466264725f;
  t151_n78: return 0.039023395627737045f;
  t151_n79: return -0.029695043340325356f;
  t151_n80: return 0.021248159930109978f;
  t151_n81: return -0.046208929270505905f;
  t151_n82: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t151_n192;
  t151_n83: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t151_n171;
  t151_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n102;
  t151_n85: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t151_n87;
  t151_n86: return -0.029727621003985405f;
  t151_n87: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t151_n89;
  t151_n88: return -0.016756586730480194f;
  t151_n89: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t151_n91;
  t151_n90: return -0.03008529543876648f;
  t151_n91: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t151_n101;
  t151_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t151_n94;
  t151_n93: return -0.004776382818818092f;
  t151_n94: if (E.getNumReferences() >= 3328893952 /*15033.0*/) goto t151_n100;
  t151_n95: if (E.getNumReferences() >= 3293454336 /*824.5*/) goto t151_n97;
  t151_n96: return 0.09894561022520065f;
  t151_n97: if (E.getNumReferences() >= 3295109120 /*925.5*/) goto t151_n99;
  t151_n98: return 0.11485476791858673f;
  t151_n99: return 0.09754913300275803f;
  t151_n100: return -0.3102653920650482f;
  t151_n101: return 0.019037822261452675f;
  t151_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t151_n112;
  t151_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t151_n105;
  t151_n104: return -0.015272408723831177f;
  t151_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t151_n107;
  t151_n106: return 0.2428063303232193f;
  t151_n107: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t151_n109;
  t151_n108: return 0.0920775905251503f;
  t151_n109: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t151_n111;
  t151_n110: return 0.21009798347949982f;
  t151_n111: return 0.03866586461663246f;
  t151_n112: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t151_n120;
  t151_n113: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t151_n119;
  t151_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n116;
  t151_n115: return -0.03879965469241142f;
  t151_n116: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t151_n118;
  t151_n117: return 0.011979651637375355f;
  t151_n118: return -0.008712214417755604f;
  t151_n119: return 0.03078865446150303f;
  t151_n120: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t151_n134;
  t151_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n123;
  t151_n122: return -0.0367475263774395f;
  t151_n123: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t151_n129;
  t151_n124: if (E.getNumReferences() >= 3289530368 /*585.0*/) goto t151_n128;
  t151_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t151_n127;
  t151_n126: return 0.0033825356513261795f;
  t151_n127: return 0.036123186349868774f;
  t151_n128: return -0.05425964668393135f;
  t151_n129: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t151_n133;
  t151_n130: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t151_n132;
  t151_n131: return 0.004922173451632261f;
  t151_n132: return -0.1935093104839325f;
  t151_n133: return 0.0999378189444542f;
  t151_n134: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t151_n170;
  t151_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t151_n147;
  t151_n136: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t151_n144;
  t151_n137: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t151_n141;
  t151_n138: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t151_n140;
  t151_n139: return -0.008839498274028301f;
  t151_n140: return 0.011647670529782772f;
  t151_n141: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t151_n143;
  t151_n142: return 0.013937020674347878f;
  t151_n143: return 0.003662084462121129f;
  t151_n144: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t151_n146;
  t151_n145: return 0.023978471755981445f;
  t151_n146: return -1.1428340673446655f;
  t151_n147: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t151_n157;
  t151_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t151_n150;
  t151_n149: return 0.0037569531705230474f;
  t151_n150: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t151_n152;
  t151_n151: return 0.028455030173063278f;
  t151_n152: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t151_n156;
  t151_n153: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t151_n155;
  t151_n154: return 0.1274782419204712f;
  t151_n155: return 0.12966497242450714f;
  t151_n156: return -0.03592551127076149f;
  t151_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t151_n161;
  t151_n158: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t151_n160;
  t151_n159: return -0.030836978927254677f;
  t151_n160: return -0.3838501572608948f;
  t151_n161: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t151_n167;
  t151_n162: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t151_n166;
  t151_n163: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t151_n165;
  t151_n164: return 0.0169691052287817f;
  t151_n165: return -0.10757601261138916f;
  t151_n166: return 0.03503278270363808f;
  t151_n167: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t151_n169;
  t151_n168: return 0.07986775785684586f;
  t151_n169: return 0.03613264113664627f;
  t151_n170: return -0.051836807280778885f;
  t151_n171: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t151_n179;
  t151_n172: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t151_n178;
  t151_n173: if (E.getNumReferences() >= 3265200128 /*79.5*/) goto t151_n177;
  t151_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t151_n176;
  t151_n175: return 0.025739239528775215f;
  t151_n176: return 0.08900832384824753f;
  t151_n177: return 0.09892338514328003f;
  t151_n178: return -0.009838273748755455f;
  t151_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t151_n191;
  t151_n180: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t151_n184;
  t151_n181: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t151_n183;
  t151_n182: return -0.12631574273109436f;
  t151_n183: return -8.06388197815977e-05f;
  t151_n184: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t151_n190;
  t151_n185: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t151_n189;
  t151_n186: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t151_n188;
  t151_n187: return -0.10804992914199829f;
  t151_n188: return 0.14625266194343567f;
  t151_n189: return 0.012668165378272533f;
  t151_n190: return 0.09980862587690353f;
  t151_n191: return 0.005708801560103893f;
  t151_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n194;
  t151_n193: return -0.027843095362186432f;
  t151_n194: return 0.009192013181746006f;
  t151_n195: return -0.17112767696380615f;
  t151_n196: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t151_n200;
  t151_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t151_n199;
  t151_n198: return 0.026721222326159477f;
  t151_n199: return 0.062156956642866135f;
  t151_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n202;
  t151_n201: return -0.15334628522396088f;
  t151_n202: if (E.getNumReferences() >= 3358250528 /*174936.5*/) goto t151_n204;
  t151_n203: return -0.15699340403079987f;
  t151_n204: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t151_n206;
  t151_n205: return 0.11996787786483765f;
  t151_n206: return 0.1854521781206131f;
  t151_n207: return -0.14830505847930908f;
  t151_n208: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t151_n216;
  t151_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n211;
  t151_n210: return -0.014199627563357353f;
  t151_n211: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t151_n215;
  t151_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t151_n214;
  t151_n213: return -0.2281925529241562f;
  t151_n214: return 0.09567258507013321f;
  t151_n215: return 0.04848827049136162f;
  t151_n216: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t151_n218;
  t151_n217: return 0.014387715607881546f;
  t151_n218: return -0.062454935163259506f;
  t151_n219: return -0.045755449682474136f;
  t151_n220: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t151_n232;
  t151_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t151_n223;
  t151_n222: return 0.008649935945868492f;
  t151_n223: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t151_n231;
  t151_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t151_n226;
  t151_n225: return 0.019029825925827026f;
  t151_n226: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t151_n228;
  t151_n227: return -0.3918980062007904f;
  t151_n228: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t151_n230;
  t151_n229: return 0.05600713565945625f;
  t151_n230: return 0.08800080418586731f;
  t151_n231: return 0.0055870171636343f;
  t151_n232: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t151_n234;
  t151_n233: return -0.06752876192331314f;
  t151_n234: return 0.012525311671197414f;
  t151_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t151_n247;
  t151_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t151_n242;
  t151_n237: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t151_n241;
  t151_n238: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t151_n240;
  t151_n239: return -0.0030483293812721968f;
  t151_n240: return -0.02583404816687107f;
  t151_n241: return 0.039825793355703354f;
  t151_n242: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t151_n244;
  t151_n243: return -0.029805628582835197f;
  t151_n244: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t151_n246;
  t151_n245: return 0.01252059917896986f;
  t151_n246: return 0.03520376607775688f;
  t151_n247: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t151_n253;
  t151_n248: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t151_n252;
  t151_n249: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t151_n251;
  t151_n250: return 0.005121679976582527f;
  t151_n251: return 0.03282593935728073f;
  t151_n252: return -0.008171170018613338f;
  t151_n253: return 0.009641560725867748f;
  t151_n254: return -0.06323254108428955f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree152(const Example& E) {
  t152_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t152_n158;
  t152_n1: if (E.getNumReferences() >= 3325613056 /*11829.0*/) goto t152_n95;
  t152_n2: if (E.getNumReferences() >= 3324264448 /*10512.0*/) goto t152_n68;
  t152_n3: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t152_n5;
  t152_n4: return -0.0009379096445627511f;
  t152_n5: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t152_n11;
  t152_n6: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t152_n10;
  t152_n7: if (E.getNumReferences() >= 3316353024 /*5489.0*/) goto t152_n9;
  t152_n8: return 0.08883367478847504f;
  t152_n9: return -0.15188157558441162f;
  t152_n10: return 0.07483997195959091f;
  t152_n11: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t152_n15;
  t152_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t152_n14;
  t152_n13: return 0.09806791692972183f;
  t152_n14: return 0.16272012889385223f;
  t152_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n17;
  t152_n16: return -0.018801579251885414f;
  t152_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t152_n27;
  t152_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n20;
  t152_n19: return -0.028517000377178192f;
  t152_n20: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t152_n22;
  t152_n21: return 0.02186526358127594f;
  t152_n22: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t152_n26;
  t152_n23: if (E.getNumReferences() >= 3311079424 /*3505.0*/) goto t152_n25;
  t152_n24: return 0.09739246964454651f;
  t152_n25: return -0.08502474427223206f;
  t152_n26: return 0.13567376136779785f;
  t152_n27: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t152_n29;
  t152_n28: return 0.0132911941036582f;
  t152_n29: if (E.getNumReferences() >= 3318440960 /*6508.5*/) goto t152_n49;
  t152_n30: if (E.getNumReferences() >= 3309787136 /*3189.5*/) goto t152_n40;
  t152_n31: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t152_n39;
  t152_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t152_n34;
  t152_n33: return 0.06445629894733429f;
  t152_n34: if (E.getNumReferences() >= 3308810240 /*2951.0*/) goto t152_n36;
  t152_n35: return -0.028056994080543518f;
  t152_n36: if (E.getNumReferences() >= 3308972032 /*2990.5*/) goto t152_n38;
  t152_n37: return 0.11825201660394669f;
  t152_n38: return 0.18474960327148438f;
  t152_n39: return 0.00896071270108223f;
  t152_n40: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t152_n42;
  t152_n41: return 0.11855712532997131f;
  t152_n42: if (E.getNumReferences() >= 3313593344 /*4141.5*/) goto t152_n48;
  t152_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n45;
  t152_n44: return -0.809259831905365f;
  t152_n45: if (E.getNumReferences() >= 3313477632 /*4090.5*/) goto t152_n47;
  t152_n46: return -0.038219280540943146f;
  t152_n47: return -0.7415831685066223f;
  t152_n48: return 0.009899797849357128f;
  t152_n49: if (E.getNumReferences() >= 3318784000 /*6676.0*/) goto t152_n53;
  t152_n50: if (E.getNumReferences() >= 3318701056 /*6635.5*/) goto t152_n52;
  t152_n51: return 0.08733866363763809f;
  t152_n52: return 0.12045002728700638f;
  t152_n53: if (E.getNumReferences() >= 3321156608 /*7834.5*/) goto t152_n57;
  t152_n54: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t152_n56;
  t152_n55: return 0.028165610507130623f;
  t152_n56: return -0.1551261693239212f;
  t152_n57: if (E.getNumReferences() >= 3321181184 /*7846.5*/) goto t152_n59;
  t152_n58: return 0.1718910187482834f;
  t152_n59: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t152_n61;
  t152_n60: return -0.024335039779543877f;
  t152_n61: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t152_n65;
  t152_n62: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t152_n64;
  t152_n63: return 0.10333261638879776f;
  t152_n64: return -0.2825421392917633f;
  t152_n65: if (E.getNumReferences() >= 3322180096 /*8476.5*/) goto t152_n67;
  t152_n66: return 0.15283505618572235f;
  t152_n67: return -0.07034026831388474f;
  t152_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t152_n70;
  t152_n69: return -0.08307166397571564f;
  t152_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n72;
  t152_n71: return -0.07144328206777573f;
  t152_n72: if (E.getNumReferences() >= 3325015040 /*11245.0*/) goto t152_n74;
  t152_n73: return -0.008823790587484837f;
  t152_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t152_n86;
  t152_n75: if (E.getNumReferences() >= 3325406720 /*11627.5*/) goto t152_n85;
  t152_n76: if (E.getNumReferences() >= 3325383680 /*11605.0*/) goto t152_n80;
  t152_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t152_n79;
  t152_n78: return 0.0600319504737854f;
  t152_n79: return 0.11323276162147522f;
  t152_n80: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t152_n82;
  t152_n81: return 0.13916945457458496f;
  t152_n82: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t152_n84;
  t152_n83: return -0.13515648245811462f;
  t152_n84: return 0.12827938795089722f;
  t152_n85: return -0.14138123393058777f;
  t152_n86: if (E.getNumReferences() >= 3325511680 /*11730.0*/) goto t152_n94;
  t152_n87: if (E.getNumReferences() >= 3325349888 /*11572.0*/) goto t152_n89;
  t152_n88: return 0.08909127116203308f;
  t152_n89: if (E.getNumReferences() >= 3325383680 /*11605.0*/) goto t152_n93;
  t152_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t152_n92;
  t152_n91: return 0.11998668313026428f;
  t152_n92: return 0.185525044798851f;
  t152_n93: return 0.07665982842445374f;
  t152_n94: return -0.2786402106285095f;
  t152_n95: if (E.getNumReferences() >= 3325964800 /*12172.5*/) goto t152_n119;
  t152_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t152_n98;
  t152_n97: return 0.0906510278582573f;
  t152_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t152_n108;
  t152_n99: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t152_n103;
  t152_n100: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t152_n102;
  t152_n101: return 0.0875566229224205f;
  t152_n102: return -0.1575687974691391f;
  t152_n103: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t152_n107;
  t152_n104: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t152_n106;
  t152_n105: return 0.3932572305202484f;
  t152_n106: return -0.15014860033988953f;
  t152_n107: return -0.10222488641738892f;
  t152_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n110;
  t152_n109: return -0.07808146625757217f;
  t152_n110: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t152_n118;
  t152_n111: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t152_n113;
  t152_n112: return 0.14037269353866577f;
  t152_n113: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t152_n115;
  t152_n114: return 0.28247615694999695f;
  t152_n115: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t152_n117;
  t152_n116: return -0.10187295824289322f;
  t152_n117: return 0.12738056480884552f;
  t152_n118: return -0.13488057255744934f;
  t152_n119: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t152_n129;
  t152_n120: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t152_n126;
  t152_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t152_n123;
  t152_n122: return -0.061221491545438766f;
  t152_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t152_n125;
  t152_n124: return 0.025636153295636177f;
  t152_n125: return 0.11376088857650757f;
  t152_n126: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t152_n128;
  t152_n127: return 0.041772305965423584f;
  t152_n128: return 0.12934063374996185f;
  t152_n129: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t152_n157;
  t152_n130: if (E.getNumReferences() >= 3327478272 /*13650.5*/) goto t152_n136;
  t152_n131: if (E.getNumReferences() >= 3327457280 /*13630.0*/) goto t152_n133;
  t152_n132: return 0.04938134551048279f;
  t152_n133: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t152_n135;
  t152_n134: return 0.2521989643573761f;
  t152_n135: return 0.016085416078567505f;
  t152_n136: if (E.getNumReferences() >= 3331353088 /*18485.0*/) goto t152_n154;
  t152_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t152_n153;
  t152_n138: if (E.getNumReferences() >= 3330547456 /*16911.5*/) goto t152_n152;
  t152_n139: if (E.getNumReferences() >= 3328655872 /*14800.5*/) goto t152_n141;
  t152_n140: return -0.0720902755856514f;
  t152_n141: if (E.getNumReferences() >= 3328821248 /*14962.0*/) goto t152_n145;
  t152_n142: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t152_n144;
  t152_n143: return 0.08810246735811234f;
  t152_n144: return 0.20131392776966095f;
  t152_n145: if (E.getNumReferences() >= 3329227264 /*15358.5*/) goto t152_n149;
  t152_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t152_n148;
  t152_n147: return -0.27325403690338135f;
  t152_n148: return 0.08142557740211487f;
  t152_n149: if (E.getNumReferences() >= 3329321472 /*15450.5*/) goto t152_n151;
  t152_n150: return -0.13568419218063354f;
  t152_n151: return 0.03510923311114311f;
  t152_n152: return -0.07780119776725769f;
  t152_n153: return 0.19343028962612152f;
  t152_n154: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t152_n156;
  t152_n155: return 0.11481445282697678f;
  t152_n156: return 0.011814051307737827f;
  t152_n157: return 0.002054353477433324f;
  t152_n158: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t152_n190;
  t152_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t152_n177;
  t152_n160: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t152_n174;
  t152_n161: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t152_n163;
  t152_n162: return -0.00014682057371828705f;
  t152_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t152_n173;
  t152_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t152_n166;
  t152_n165: return -0.05250578746199608f;
  t152_n166: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t152_n168;
  t152_n167: return -0.03203093260526657f;
  t152_n168: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t152_n172;
  t152_n169: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t152_n171;
  t152_n170: return 0.1041010394692421f;
  t152_n171: return -0.2500298321247101f;
  t152_n172: return 0.1231025978922844f;
  t152_n173: return 0.0534038320183754f;
  t152_n174: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t152_n176;
  t152_n175: return 0.01648169942200184f;
  t152_n176: return -0.005571965128183365f;
  t152_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace))) goto t152_n179;
  t152_n178: return 0.0015367245068773627f;
  t152_n179: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t152_n189;
  t152_n180: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t152_n182;
  t152_n181: return 0.011465790681540966f;
  t152_n182: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t152_n188;
  t152_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t152_n185;
  t152_n184: return 0.0685146227478981f;
  t152_n185: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t152_n187;
  t152_n186: return 0.01676822267472744f;
  t152_n187: return 0.06748930364847183f;
  t152_n188: return -0.024325914680957794f;
  t152_n189: return -0.1097826138138771f;
  t152_n190: if (E.getFractionNameInContext() >= 3185148150 /*0.10620300471782684*/) goto t152_n194;
  t152_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t152_n193;
  t152_n192: return 0.011055567301809788f;
  t152_n193: return 0.01669999025762081f;
  t152_n194: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t152_n234;
  t152_n195: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t152_n199;
  t152_n196: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t152_n198;
  t152_n197: return -0.28885239362716675f;
  t152_n198: return 0.001687230309471488f;
  t152_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t152_n201;
  t152_n200: return -0.014968397095799446f;
  t152_n201: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t152_n213;
  t152_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t152_n212;
  t152_n203: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t152_n211;
  t152_n204: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t152_n210;
  t152_n205: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t152_n207;
  t152_n206: return 0.024154115468263626f;
  t152_n207: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t152_n209;
  t152_n208: return 0.10923245549201965f;
  t152_n209: return 0.04886525496840477f;
  t152_n210: return 0.009989823214709759f;
  t152_n211: return 0.004479991737753153f;
  t152_n212: return 0.1009778380393982f;
  t152_n213: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t152_n217;
  t152_n214: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t152_n216;
  t152_n215: return -0.08909133076667786f;
  t152_n216: return -0.01051932293921709f;
  t152_n217: if (E.getNumReferences() >= 3275718656 /*191.5*/) goto t152_n231;
  t152_n218: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t152_n230;
  t152_n219: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t152_n227;
  t152_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t152_n222;
  t152_n221: return 0.018124481663107872f;
  t152_n222: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t152_n224;
  t152_n223: return 0.040103986859321594f;
  t152_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t152_n226;
  t152_n225: return 0.030140165239572525f;
  t152_n226: return 0.09122568368911743f;
  t152_n227: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t152_n229;
  t152_n228: return 0.08350656181573868f;
  t152_n229: return -0.019295549020171165f;
  t152_n230: return -0.14173394441604614f;
  t152_n231: if (E.getNumReferences() >= 3297206272 /*1083.0*/) goto t152_n233;
  t152_n232: return 0.11972405761480331f;
  t152_n233: return -0.10600259900093079f;
  t152_n234: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t152_n248;
  t152_n235: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t152_n247;
  t152_n236: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t152_n244;
  t152_n237: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t152_n243;
  t152_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t152_n242;
  t152_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t152_n241;
  t152_n240: return -0.3617297410964966f;
  t152_n241: return -0.10035045444965363f;
  t152_n242: return -0.0648089051246643f;
  t152_n243: return 0.028130805119872093f;
  t152_n244: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t152_n246;
  t152_n245: return 0.03306911513209343f;
  t152_n246: return 0.051270533353090286f;
  t152_n247: return 0.00473210122436285f;
  t152_n248: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t152_n254;
  t152_n249: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t152_n251;
  t152_n250: return 0.03728919476270676f;
  t152_n251: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t152_n253;
  t152_n252: return 0.16641996800899506f;
  t152_n253: return 0.07578504830598831f;
  t152_n254: return -0.009012590162456036f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree153(const Example& E) {
  t153_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t153_n180;
  t153_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t153_n49;
  t153_n2: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t153_n4;
  t153_n3: return -0.0007129230070859194f;
  t153_n4: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t153_n42;
  t153_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t153_n11;
  t153_n6: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t153_n8;
  t153_n7: return -0.008967039175331593f;
  t153_n8: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t153_n10;
  t153_n9: return 0.1205822229385376f;
  t153_n10: return 0.07914672791957855f;
  t153_n11: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t153_n41;
  t153_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t153_n14;
  t153_n13: return -0.02891518548130989f;
  t153_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t153_n40;
  t153_n15: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t153_n23;
  t153_n16: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t153_n18;
  t153_n17: return 0.0005499468534253538f;
  t153_n18: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t153_n22;
  t153_n19: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t153_n21;
  t153_n20: return 0.4417293667793274f;
  t153_n21: return -0.2688146233558655f;
  t153_n22: return 0.6688527464866638f;
  t153_n23: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t153_n25;
  t153_n24: return 0.20050984621047974f;
  t153_n25: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t153_n29;
  t153_n26: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t153_n28;
  t153_n27: return -0.10045656561851501f;
  t153_n28: return 0.23429518938064575f;
  t153_n29: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t153_n31;
  t153_n30: return -0.026346305385231972f;
  t153_n31: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t153_n33;
  t153_n32: return 0.2620154321193695f;
  t153_n33: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t153_n35;
  t153_n34: return -0.10036366432905197f;
  t153_n35: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t153_n37;
  t153_n36: return 0.25364044308662415f;
  t153_n37: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t153_n39;
  t153_n38: return 0.2365456223487854f;
  t153_n39: return 0.1661195009946823f;
  t153_n40: return -0.05340857803821564f;
  t153_n41: return -0.06043237820267677f;
  t153_n42: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t153_n48;
  t153_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n45;
  t153_n44: return -0.017111506313085556f;
  t153_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t153_n47;
  t153_n46: return 0.008839649148285389f;
  t153_n47: return 0.15001502633094788f;
  t153_n48: return -0.0003334670327603817f;
  t153_n49: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t153_n119;
  t153_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t153_n54;
  t153_n51: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t153_n53;
  t153_n52: return 0.002292707795277238f;
  t153_n53: return 0.02707427553832531f;
  t153_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t153_n104;
  t153_n55: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t153_n97;
  t153_n56: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t153_n60;
  t153_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n59;
  t153_n58: return -0.14297640323638916f;
  t153_n59: return -0.12271413952112198f;
  t153_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t153_n84;
  t153_n61: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t153_n83;
  t153_n62: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t153_n80;
  t153_n63: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t153_n73;
  t153_n64: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t153_n70;
  t153_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n69;
  t153_n66: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n68;
  t153_n67: return -0.028893083333969116f;
  t153_n68: return 0.07945450395345688f;
  t153_n69: return 0.0345071405172348f;
  t153_n70: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t153_n72;
  t153_n71: return -0.021756792441010475f;
  t153_n72: return -0.23137183487415314f;
  t153_n73: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t153_n75;
  t153_n74: return 0.02776613086462021f;
  t153_n75: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t153_n77;
  t153_n76: return -0.5832440257072449f;
  t153_n77: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t153_n79;
  t153_n78: return 0.11753039062023163f;
  t153_n79: return 0.09166688472032547f;
  t153_n80: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t153_n82;
  t153_n81: return 0.025738254189491272f;
  t153_n82: return 0.005720091983675957f;
  t153_n83: return -0.08424089848995209f;
  t153_n84: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t153_n96;
  t153_n85: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t153_n93;
  t153_n86: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t153_n88;
  t153_n87: return 0.013242166489362717f;
  t153_n88: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n90;
  t153_n89: return -0.1066398173570633f;
  t153_n90: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t153_n92;
  t153_n91: return 0.0405302494764328f;
  t153_n92: return -0.20130555331707f;
  t153_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t153_n95;
  t153_n94: return 0.0009720955858938396f;
  t153_n95: return 0.09051977843046188f;
  t153_n96: return 0.04993785172700882f;
  t153_n97: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t153_n101;
  t153_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n100;
  t153_n99: return -0.08773264288902283f;
  t153_n100: return -0.024139923974871635f;
  t153_n101: if (E.getFractionNameInContext() >= 3208803880 /*0.759615421295166*/) goto t153_n103;
  t153_n102: return 0.15009905397891998f;
  t153_n103: return 0.01793220080435276f;
  t153_n104: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t153_n116;
  t153_n105: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t153_n109;
  t153_n106: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t153_n108;
  t153_n107: return 0.041704609990119934f;
  t153_n108: return 0.14975062012672424f;
  t153_n109: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t153_n111;
  t153_n110: return -0.39861780405044556f;
  t153_n111: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t153_n115;
  t153_n112: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t153_n114;
  t153_n113: return 0.30108898878097534f;
  t153_n114: return -0.19048035144805908f;
  t153_n115: return -0.031040946021676064f;
  t153_n116: if (E.getFractionNameInContext() >= 3200453681 /*0.380952388048172*/) goto t153_n118;
  t153_n117: return 0.021554289385676384f;
  t153_n118: return 0.14470094442367554f;
  t153_n119: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t153_n127;
  t153_n120: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n122;
  t153_n121: return -0.02911890670657158f;
  t153_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t153_n126;
  t153_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n125;
  t153_n124: return -0.0641317218542099f;
  t153_n125: return 0.053425226360559464f;
  t153_n126: return 0.13151676952838898f;
  t153_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t153_n175;
  t153_n128: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t153_n174;
  t153_n129: if (E.getNumReferences() >= 3276341248 /*201.0*/) goto t153_n173;
  t153_n130: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t153_n170;
  t153_n131: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t153_n151;
  t153_n132: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t153_n136;
  t153_n133: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n135;
  t153_n134: return 0.05615615099668503f;
  t153_n135: return 0.11374147981405258f;
  t153_n136: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n138;
  t153_n137: return -0.01979777216911316f;
  t153_n138: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t153_n140;
  t153_n139: return -0.03676852956414223f;
  t153_n140: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t153_n150;
  t153_n141: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t153_n149;
  t153_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t153_n144;
  t153_n143: return 0.07663854956626892f;
  t153_n144: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n148;
  t153_n145: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t153_n147;
  t153_n146: return 0.12871095538139343f;
  t153_n147: return 0.09976619482040405f;
  t153_n148: return 0.08117610216140747f;
  t153_n149: return -0.24157482385635376f;
  t153_n150: return -0.05182238668203354f;
  t153_n151: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t153_n157;
  t153_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n154;
  t153_n153: return -0.36563900113105774f;
  t153_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n156;
  t153_n155: return 0.08581510931253433f;
  t153_n156: return -0.16525180637836456f;
  t153_n157: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t153_n159;
  t153_n158: return 0.1474395990371704f;
  t153_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t153_n161;
  t153_n160: return 0.010132948867976665f;
  t153_n161: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n163;
  t153_n162: return 0.14262089133262634f;
  t153_n163: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t153_n169;
  t153_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n166;
  t153_n165: return -0.004493872169405222f;
  t153_n166: if (E.getFractionNameInContext() >= 3206096018 /*0.5982142686843872*/) goto t153_n168;
  t153_n167: return 0.11145889759063721f;
  t153_n168: return 0.06824865192174911f;
  t153_n169: return -0.12318069487810135f;
  t153_n170: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n172;
  t153_n171: return 0.1250338852405548f;
  t153_n172: return 0.00035260836011730134f;
  t153_n173: return -0.9693260192871094f;
  t153_n174: return -0.1619635969400406f;
  t153_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t153_n179;
  t153_n176: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n178;
  t153_n177: return 0.11070432513952255f;
  t153_n178: return -0.4277738332748413f;
  t153_n179: return 0.13226400315761566f;
  t153_n180: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t153_n200;
  t153_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t153_n199;
  t153_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t153_n196;
  t153_n183: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t153_n195;
  t153_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n192;
  t153_n185: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t153_n187;
  t153_n186: return -0.005978669039905071f;
  t153_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n191;
  t153_n188: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t153_n190;
  t153_n189: return -0.05351943522691727f;
  t153_n190: return -0.022814467549324036f;
  t153_n191: return 0.0018510966328904033f;
  t153_n192: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t153_n194;
  t153_n193: return 0.07546260952949524f;
  t153_n194: return -0.0021697371266782284f;
  t153_n195: return 0.006353740114718676f;
  t153_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t153_n198;
  t153_n197: return 0.009852803312242031f;
  t153_n198: return 0.05007888376712799f;
  t153_n199: return 0.04298819601535797f;
  t153_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n224;
  t153_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n209;
  t153_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n204;
  t153_n203: return -0.019357632845640182f;
  t153_n204: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t153_n208;
  t153_n205: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t153_n207;
  t153_n206: return 0.006587623618543148f;
  t153_n207: return -0.08859608322381973f;
  t153_n208: return 0.08456120640039444f;
  t153_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n211;
  t153_n210: return -0.035753872245550156f;
  t153_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t153_n213;
  t153_n212: return -0.001152299577370286f;
  t153_n213: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t153_n223;
  t153_n214: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t153_n218;
  t153_n215: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t153_n217;
  t153_n216: return 0.01676233857870102f;
  t153_n217: return 0.008346511982381344f;
  t153_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t153_n222;
  t153_n219: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t153_n221;
  t153_n220: return 0.012298489920794964f;
  t153_n221: return -0.1150633916258812f;
  t153_n222: return 0.03618570417165756f;
  t153_n223: return -0.08049625903367996f;
  t153_n224: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t153_n242;
  t153_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t153_n227;
  t153_n226: return -0.01556497160345316f;
  t153_n227: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t153_n237;
  t153_n228: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t153_n234;
  t153_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t153_n233;
  t153_n230: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t153_n232;
  t153_n231: return -0.014377137646079063f;
  t153_n232: return -0.3627298176288605f;
  t153_n233: return 0.037353310734033585f;
  t153_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t153_n236;
  t153_n235: return 0.05854054167866707f;
  t153_n236: return 0.12594151496887207f;
  t153_n237: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t153_n241;
  t153_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t153_n240;
  t153_n239: return -0.21537581086158752f;
  t153_n240: return -0.07795758545398712f;
  t153_n241: return 0.017873259261250496f;
  t153_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t153_n244;
  t153_n243: return 0.008146225474774837f;
  t153_n244: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t153_n246;
  t153_n245: return 0.007241897284984589f;
  t153_n246: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t153_n248;
  t153_n247: return 0.022379331290721893f;
  t153_n248: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t153_n252;
  t153_n249: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t153_n251;
  t153_n250: return -0.00216707494109869f;
  t153_n251: return -0.08184418827295303f;
  t153_n252: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t153_n254;
  t153_n253: return 0.028280165046453476f;
  t153_n254: return 0.093626469373703f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree154(const Example& E) {
  t154_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t154_n184;
  t154_n1: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t154_n3;
  t154_n2: return -0.0008327829418703914f;
  t154_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t154_n127;
  t154_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t154_n12;
  t154_n5: if (E.getFractionNameInContext() >= 3174441576 /*0.04446640610694885*/) goto t154_n7;
  t154_n6: return 0.0976591408252716f;
  t154_n7: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t154_n9;
  t154_n8: return -0.00401657959446311f;
  t154_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n11;
  t154_n10: return -0.003987138159573078f;
  t154_n11: return 0.008504435420036316f;
  t154_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t154_n86;
  t154_n13: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t154_n71;
  t154_n14: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t154_n64;
  t154_n15: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t154_n59;
  t154_n16: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t154_n40;
  t154_n17: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t154_n39;
  t154_n18: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t154_n24;
  t154_n19: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t154_n21;
  t154_n20: return 0.0012367627350613475f;
  t154_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t154_n23;
  t154_n22: return -0.0128306420519948f;
  t154_n23: return 0.010414487682282925f;
  t154_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t154_n26;
  t154_n25: return -0.01634824089705944f;
  t154_n26: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t154_n36;
  t154_n27: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t154_n29;
  t154_n28: return 0.03929777815937996f;
  t154_n29: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t154_n31;
  t154_n30: return 0.04590274766087532f;
  t154_n31: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t154_n35;
  t154_n32: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t154_n34;
  t154_n33: return 0.11281303316354752f;
  t154_n34: return 0.1149551123380661f;
  t154_n35: return 0.07011692970991135f;
  t154_n36: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t154_n38;
  t154_n37: return 0.13845087587833405f;
  t154_n38: return 0.0002822509268298745f;
  t154_n39: return -0.004216144327074289f;
  t154_n40: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t154_n56;
  t154_n41: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t154_n45;
  t154_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n44;
  t154_n43: return -0.09095434099435806f;
  t154_n44: return 0.10001669824123383f;
  t154_n45: if (E.getFractionNameInContext() >= 3194215998 /*0.22252747416496277*/) goto t154_n51;
  t154_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n48;
  t154_n47: return -0.09909647703170776f;
  t154_n48: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t154_n50;
  t154_n49: return 0.045554932206869125f;
  t154_n50: return -0.3545428216457367f;
  t154_n51: if (E.getNumReferences() >= 3299045376 /*1307.5*/) goto t154_n55;
  t154_n52: if (E.getNumReferences() >= 3295617024 /*956.5*/) goto t154_n54;
  t154_n53: return -0.02478114701807499f;
  t154_n54: return -0.3074185848236084f;
  t154_n55: return 0.0878395214676857f;
  t154_n56: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t154_n58;
  t154_n57: return 0.22765497863292694f;
  t154_n58: return 0.015540018677711487f;
  t154_n59: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t154_n63;
  t154_n60: if (E.getFractionNameInContext() >= 3200304895 /*0.37651821970939636*/) goto t154_n62;
  t154_n61: return -0.00478961830958724f;
  t154_n62: return -0.09118859469890594f;
  t154_n63: return 0.12425476312637329f;
  t154_n64: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t154_n70;
  t154_n65: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t154_n69;
  t154_n66: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t154_n68;
  t154_n67: return 0.029902903363108635f;
  t154_n68: return -0.04564666375517845f;
  t154_n69: return 0.06777402758598328f;
  t154_n70: return 0.009635462425649166f;
  t154_n71: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t154_n83;
  t154_n72: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t154_n82;
  t154_n73: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t154_n75;
  t154_n74: return -0.013043341226875782f;
  t154_n75: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t154_n79;
  t154_n76: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t154_n78;
  t154_n77: return 0.002735552843660116f;
  t154_n78: return 0.05843352526426315f;
  t154_n79: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t154_n81;
  t154_n80: return 0.025934861972928047f;
  t154_n81: return 0.10031972825527191f;
  t154_n82: return -0.14630739390850067f;
  t154_n83: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t154_n85;
  t154_n84: return 0.20541535317897797f;
  t154_n85: return 0.13497887551784515f;
  t154_n86: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t154_n104;
  t154_n87: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t154_n101;
  t154_n88: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t154_n90;
  t154_n89: return 0.00564919039607048f;
  t154_n90: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t154_n98;
  t154_n91: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t154_n95;
  t154_n92: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t154_n94;
  t154_n93: return -0.04971541464328766f;
  t154_n94: return -0.021705687046051025f;
  t154_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t154_n97;
  t154_n96: return 0.10885797441005707f;
  t154_n97: return -0.18944041430950165f;
  t154_n98: if (E.getNumReferences() >= 3280601088 /*276.0*/) goto t154_n100;
  t154_n99: return 0.1295892894268036f;
  t154_n100: return -6.421474972739816e-05f;
  t154_n101: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t154_n103;
  t154_n102: return 0.06459620594978333f;
  t154_n103: return -0.05664996802806854f;
  t154_n104: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t154_n112;
  t154_n105: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t154_n107;
  t154_n106: return 0.009030444547533989f;
  t154_n107: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t154_n111;
  t154_n108: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t154_n110;
  t154_n109: return -0.09927154332399368f;
  t154_n110: return -0.00047763067414052784f;
  t154_n111: return 0.011845405213534832f;
  t154_n112: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t154_n124;
  t154_n113: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t154_n123;
  t154_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t154_n116;
  t154_n115: return 0.04608972743153572f;
  t154_n116: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t154_n120;
  t154_n117: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t154_n119;
  t154_n118: return -0.151677206158638f;
  t154_n119: return 0.11992846429347992f;
  t154_n120: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t154_n122;
  t154_n121: return 0.1128256618976593f;
  t154_n122: return -0.08957874029874802f;
  t154_n123: return 0.011594762094318867f;
  t154_n124: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t154_n126;
  t154_n125: return 0.6290426254272461f;
  t154_n126: return 0.08849387615919113f;
  t154_n127: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t154_n183;
  t154_n128: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t154_n172;
  t154_n129: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t154_n169;
  t154_n130: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t154_n142;
  t154_n131: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t154_n141;
  t154_n132: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t154_n134;
  t154_n133: return 0.012916183099150658f;
  t154_n134: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t154_n136;
  t154_n135: return 0.002250427845865488f;
  t154_n136: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t154_n138;
  t154_n137: return 0.012132161296904087f;
  t154_n138: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t154_n140;
  t154_n139: return 0.37391752004623413f;
  t154_n140: return 0.1809931993484497f;
  t154_n141: return 0.0011096142698079348f;
  t154_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n168;
  t154_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t154_n147;
  t154_n144: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t154_n146;
  t154_n145: return -0.021694660186767578f;
  t154_n146: return 2.9682785680051893e-05f;
  t154_n147: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t154_n167;
  t154_n148: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t154_n150;
  t154_n149: return 0.043267473578453064f;
  t154_n150: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t154_n152;
  t154_n151: return -0.01606784760951996f;
  t154_n152: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t154_n160;
  t154_n153: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t154_n155;
  t154_n154: return 0.06471025198698044f;
  t154_n155: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t154_n159;
  t154_n156: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t154_n158;
  t154_n157: return -0.005585966631770134f;
  t154_n158: return -0.5995609164237976f;
  t154_n159: return 0.04514555633068085f;
  t154_n160: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t154_n166;
  t154_n161: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t154_n165;
  t154_n162: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t154_n164;
  t154_n163: return 0.11865591257810593f;
  t154_n164: return -0.09325748682022095f;
  t154_n165: return -0.18990787863731384f;
  t154_n166: return 0.08387797325849533f;
  t154_n167: return -0.030550338327884674f;
  t154_n168: return 0.027407174929976463f;
  t154_n169: if (E.getFractionNameInContext() >= 3212237678 /*0.9642857313156128*/) goto t154_n171;
  t154_n170: return -0.004219105001538992f;
  t154_n171: return 0.029889285564422607f;
  t154_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t154_n174;
  t154_n173: return -0.14194801449775696f;
  t154_n174: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t154_n176;
  t154_n175: return 0.00034316390519961715f;
  t154_n176: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t154_n180;
  t154_n177: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t154_n179;
  t154_n178: return 0.019408751279115677f;
  t154_n179: return -0.00250615063123405f;
  t154_n180: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t154_n182;
  t154_n181: return 0.36885330080986023f;
  t154_n182: return 0.0021035431418567896f;
  t154_n183: return 0.007233025971800089f;
  t154_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n202;
  t154_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t154_n197;
  t154_n186: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t154_n196;
  t154_n187: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t154_n195;
  t154_n188: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t154_n190;
  t154_n189: return -0.0016799484146758914f;
  t154_n190: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t154_n192;
  t154_n191: return -0.012826674617826939f;
  t154_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t154_n194;
  t154_n193: return -0.12819336354732513f;
  t154_n194: return 0.09676289558410645f;
  t154_n195: return -0.03720804676413536f;
  t154_n196: return 0.03574002534151077f;
  t154_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t154_n199;
  t154_n198: return -0.029769698157906532f;
  t154_n199: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t154_n201;
  t154_n200: return 0.016364237293601036f;
  t154_n201: return 0.04717300087213516f;
  t154_n202: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t154_n250;
  t154_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t154_n217;
  t154_n204: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t154_n216;
  t154_n205: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t154_n209;
  t154_n206: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t154_n208;
  t154_n207: return 0.007701096590608358f;
  t154_n208: return 0.07909845560789108f;
  t154_n209: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t154_n211;
  t154_n210: return -0.012627852149307728f;
  t154_n211: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t154_n213;
  t154_n212: return 0.011696655303239822f;
  t154_n213: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t154_n215;
  t154_n214: return -0.07501830905675888f;
  t154_n215: return 0.10299888998270035f;
  t154_n216: return 0.02305787429213524f;
  t154_n217: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t154_n231;
  t154_n218: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t154_n230;
  t154_n219: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t154_n221;
  t154_n220: return -0.002870124764740467f;
  t154_n221: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t154_n229;
  t154_n222: if (E.getNumReferences() >= 3310100480 /*3266.0*/) goto t154_n228;
  t154_n223: if (E.getNumReferences() >= 3275915264 /*194.5*/) goto t154_n227;
  t154_n224: if (E.getNumReferences() >= 3274178560 /*168.0*/) goto t154_n226;
  t154_n225: return 0.020858528092503548f;
  t154_n226: return -0.2686040699481964f;
  t154_n227: return 0.06333497166633606f;
  t154_n228: return -0.21597322821617126f;
  t154_n229: return 0.007768656127154827f;
  t154_n230: return 0.009693634696304798f;
  t154_n231: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t154_n237;
  t154_n232: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t154_n236;
  t154_n233: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t154_n235;
  t154_n234: return 0.012584757059812546f;
  t154_n235: return 0.03736163675785065f;
  t154_n236: return -0.013011770322918892f;
  t154_n237: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t154_n239;
  t154_n238: return -0.00463229650631547f;
  t154_n239: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t154_n249;
  t154_n240: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t154_n248;
  t154_n241: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t154_n243;
  t154_n242: return 0.08814200013875961f;
  t154_n243: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t154_n247;
  t154_n244: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t154_n246;
  t154_n245: return 0.1507526934146881f;
  t154_n246: return 0.10058129578828812f;
  t154_n247: return 0.1403297632932663f;
  t154_n248: return -0.0388345830142498f;
  t154_n249: return 0.03196931257843971f;
  t154_n250: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t154_n252;
  t154_n251: return 0.008722607977688313f;
  t154_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t154_n254;
  t154_n253: return 0.04813183471560478f;
  t154_n254: return 0.021289465948939323f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree155(const Example& E) {
  t155_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t155_n252;
  t155_n1: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t155_n241;
  t155_n2: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t155_n230;
  t155_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t155_n171;
  t155_n4: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t155_n16;
  t155_n5: if (E.getNumReferences() >= 3272310784 /*139.5*/) goto t155_n13;
  t155_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n8;
  t155_n7: return 0.0021744119003415108f;
  t155_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t155_n12;
  t155_n9: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t155_n11;
  t155_n10: return -0.006550749763846397f;
  t155_n11: return 0.015910757705569267f;
  t155_n12: return 0.028884708881378174f;
  t155_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t155_n15;
  t155_n14: return 0.05876857042312622f;
  t155_n15: return 0.21323180198669434f;
  t155_n16: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t155_n138;
  t155_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t155_n27;
  t155_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n26;
  t155_n19: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t155_n25;
  t155_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n24;
  t155_n21: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t155_n23;
  t155_n22: return 0.10925175994634628f;
  t155_n23: return -0.2638489007949829f;
  t155_n24: return -0.07194487005472183f;
  t155_n25: return -0.018628571182489395f;
  t155_n26: return 0.0005709349061362445f;
  t155_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t155_n29;
  t155_n28: return -0.0008896253420971334f;
  t155_n29: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t155_n91;
  t155_n30: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t155_n72;
  t155_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t155_n67;
  t155_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t155_n58;
  t155_n33: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t155_n55;
  t155_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n44;
  t155_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n37;
  t155_n36: return -0.10961497575044632f;
  t155_n37: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t155_n39;
  t155_n38: return -0.07583041489124298f;
  t155_n39: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t155_n43;
  t155_n40: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t155_n42;
  t155_n41: return 0.06903199851512909f;
  t155_n42: return 0.021848097443580627f;
  t155_n43: return -0.01869354583323002f;
  t155_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t155_n46;
  t155_n45: return -0.04269365221261978f;
  t155_n46: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t155_n50;
  t155_n47: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t155_n49;
  t155_n48: return 0.013602769002318382f;
  t155_n49: return 0.08096402883529663f;
  t155_n50: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t155_n54;
  t155_n51: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t155_n53;
  t155_n52: return 0.018803153187036514f;
  t155_n53: return 0.1010453924536705f;
  t155_n54: return 0.03618680685758591f;
  t155_n55: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t155_n57;
  t155_n56: return -0.029007496312260628f;
  t155_n57: return -0.0167599655687809f;
  t155_n58: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t155_n62;
  t155_n59: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t155_n61;
  t155_n60: return -0.03489379957318306f;
  t155_n61: return 0.05244787409901619f;
  t155_n62: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t155_n64;
  t155_n63: return 0.00037907843943685293f;
  t155_n64: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t155_n66;
  t155_n65: return -0.03777686506509781f;
  t155_n66: return 0.02153114043176174f;
  t155_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n71;
  t155_n68: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t155_n70;
  t155_n69: return -0.12039345502853394f;
  t155_n70: return 0.0011330160778015852f;
  t155_n71: return 0.0052872877568006516f;
  t155_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t155_n74;
  t155_n73: return -0.0022044179495424032f;
  t155_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t155_n76;
  t155_n75: return 0.015702946111559868f;
  t155_n76: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t155_n90;
  t155_n77: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t155_n81;
  t155_n78: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t155_n80;
  t155_n79: return -0.26436570286750793f;
  t155_n80: return 0.07254991680383682f;
  t155_n81: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t155_n87;
  t155_n82: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t155_n86;
  t155_n83: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t155_n85;
  t155_n84: return 0.1330985426902771f;
  t155_n85: return 0.13560718297958374f;
  t155_n86: return -0.180008664727211f;
  t155_n87: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t155_n89;
  t155_n88: return 0.182562917470932f;
  t155_n89: return 0.1376953125f;
  t155_n90: return 0.04738541319966316f;
  t155_n91: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t155_n111;
  t155_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t155_n94;
  t155_n93: return 0.0031228982843458652f;
  t155_n94: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t155_n106;
  t155_n95: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t155_n99;
  t155_n96: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t155_n98;
  t155_n97: return -0.22175000607967377f;
  t155_n98: return 0.03696814924478531f;
  t155_n99: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t155_n103;
  t155_n100: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t155_n102;
  t155_n101: return 0.04934520274400711f;
  t155_n102: return -0.02663453482091427f;
  t155_n103: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t155_n105;
  t155_n104: return 0.07221674174070358f;
  t155_n105: return 0.11485811322927475f;
  t155_n106: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t155_n110;
  t155_n107: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t155_n109;
  t155_n108: return -0.03880688175559044f;
  t155_n109: return 0.1219949722290039f;
  t155_n110: return 0.011007539927959442f;
  t155_n111: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t155_n115;
  t155_n112: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t155_n114;
  t155_n113: return -0.06201751530170441f;
  t155_n114: return -6.965250492095947f;
  t155_n115: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t155_n135;
  t155_n116: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t155_n124;
  t155_n117: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t155_n123;
  t155_n118: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t155_n120;
  t155_n119: return 0.04947160184383392f;
  t155_n120: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t155_n122;
  t155_n121: return 0.09884340316057205f;
  t155_n122: return -0.03920314833521843f;
  t155_n123: return -0.00428219698369503f;
  t155_n124: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t155_n128;
  t155_n125: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t155_n127;
  t155_n126: return 0.13155803084373474f;
  t155_n127: return 0.05462288111448288f;
  t155_n128: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t155_n132;
  t155_n129: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t155_n131;
  t155_n130: return 0.00023891753517091274f;
  t155_n131: return 0.0983048677444458f;
  t155_n132: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t155_n134;
  t155_n133: return -0.06846068054437637f;
  t155_n134: return 0.09289288520812988f;
  t155_n135: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t155_n137;
  t155_n136: return -0.017926055938005447f;
  t155_n137: return 0.12793363630771637f;
  t155_n138: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t155_n144;
  t155_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n143;
  t155_n140: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t155_n142;
  t155_n141: return -0.05539511516690254f;
  t155_n142: return -0.05518792197108269f;
  t155_n143: return -0.04034961387515068f;
  t155_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t155_n146;
  t155_n145: return -0.006768969818949699f;
  t155_n146: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t155_n170;
  t155_n147: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t155_n169;
  t155_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t155_n168;
  t155_n149: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t155_n167;
  t155_n150: if (E.getNumReferences() >= 3269787648 /*114.5*/) goto t155_n156;
  t155_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n155;
  t155_n152: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t155_n154;
  t155_n153: return -0.06072014570236206f;
  t155_n154: return 0.028927866369485855f;
  t155_n155: return 0.030609123408794403f;
  t155_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t155_n160;
  t155_n157: if (E.getNumReferences() >= 3271589888 /*128.5*/) goto t155_n159;
  t155_n158: return 0.10215329378843307f;
  t155_n159: return -0.057466521859169006f;
  t155_n160: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t155_n166;
  t155_n161: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t155_n163;
  t155_n162: return -0.11767412722110748f;
  t155_n163: if (E.getNumReferences() >= 3271753728 /*131.0*/) goto t155_n165;
  t155_n164: return 0.23384127020835876f;
  t155_n165: return 0.20604580640792847f;
  t155_n166: return -0.06731051206588745f;
  t155_n167: return -0.01102792751044035f;
  t155_n168: return 0.012678139843046665f;
  t155_n169: return -0.011225437745451927f;
  t155_n170: return -0.016218550503253937f;
  t155_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n179;
  t155_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n178;
  t155_n173: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t155_n177;
  t155_n174: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t155_n176;
  t155_n175: return -0.32528579235076904f;
  t155_n176: return -0.15989114344120026f;
  t155_n177: return 0.08313433825969696f;
  t155_n178: return -0.009997456334531307f;
  t155_n179: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t155_n223;
  t155_n180: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t155_n186;
  t155_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n185;
  t155_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t155_n184;
  t155_n183: return -0.013509847223758698f;
  t155_n184: return 0.004172727465629578f;
  t155_n185: return 0.004197319038212299f;
  t155_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t155_n202;
  t155_n187: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t155_n195;
  t155_n188: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t155_n194;
  t155_n189: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t155_n193;
  t155_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t155_n192;
  t155_n191: return 0.058344610035419464f;
  t155_n192: return 0.15295076370239258f;
  t155_n193: return -0.007514358032494783f;
  t155_n194: return 0.04278147965669632f;
  t155_n195: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t155_n201;
  t155_n196: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t155_n198;
  t155_n197: return -0.029716484248638153f;
  t155_n198: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t155_n200;
  t155_n199: return -0.7323216199874878f;
  t155_n200: return -0.07304318249225616f;
  t155_n201: return 0.1142905205488205f;
  t155_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n210;
  t155_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n209;
  t155_n204: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t155_n206;
  t155_n205: return 0.11219681799411774f;
  t155_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n208;
  t155_n207: return -0.028111878782510757f;
  t155_n208: return -0.027929259464144707f;
  t155_n209: return 0.012947232462465763f;
  t155_n210: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t155_n218;
  t155_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t155_n217;
  t155_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n214;
  t155_n213: return 0.023792706429958344f;
  t155_n214: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t155_n216;
  t155_n215: return 0.025247899815440178f;
  t155_n216: return 0.0012846138561144471f;
  t155_n217: return 0.021007856354117393f;
  t155_n218: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t155_n220;
  t155_n219: return 0.00817959662526846f;
  t155_n220: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t155_n222;
  t155_n221: return -0.03513643518090248f;
  t155_n222: return 0.009758652187883854f;
  t155_n223: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t155_n225;
  t155_n224: return -0.11026198416948318f;
  t155_n225: if (E.getNumReferences() >= 3272441856 /*141.5*/) goto t155_n227;
  t155_n226: return 0.0012079450534656644f;
  t155_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t155_n229;
  t155_n228: return -0.1857578605413437f;
  t155_n229: return -0.05684657394886017f;
  t155_n230: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t155_n240;
  t155_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t155_n235;
  t155_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t155_n234;
  t155_n233: return -0.16078294813632965f;
  t155_n234: return -0.0163123719394207f;
  t155_n235: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t155_n239;
  t155_n236: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t155_n238;
  t155_n237: return 0.0379500538110733f;
  t155_n238: return -0.06313807517290115f;
  t155_n239: return 0.16139113903045654f;
  t155_n240: return -0.10071814060211182f;
  t155_n241: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t155_n245;
  t155_n242: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t155_n244;
  t155_n243: return 0.009257437661290169f;
  t155_n244: return -0.1617567241191864f;
  t155_n245: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t155_n247;
  t155_n246: return 0.0002622609317768365f;
  t155_n247: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t155_n249;
  t155_n248: return 0.023511674255132675f;
  t155_n249: if (E.getNumReferences() >= 3283550208 /*366.0*/) goto t155_n251;
  t155_n250: return 0.1423938274383545f;
  t155_n251: return 0.13011257350444794f;
  t155_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t155_n254;
  t155_n253: return 0.0015812547644600272f;
  t155_n254: return 0.006493689492344856f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree156(const Example& E) {
  t156_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t156_n248;
  t156_n1: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t156_n233;
  t156_n2: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t156_n232;
  t156_n3: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t156_n229;
  t156_n4: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t156_n228;
  t156_n5: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t156_n217;
  t156_n6: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t156_n216;
  t156_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t156_n9;
  t156_n8: return -0.09824693202972412f;
  t156_n9: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t156_n215;
  t156_n10: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t156_n104;
  t156_n11: if (E.getNumReferences() >= 3314787328 /*4724.5*/) goto t156_n97;
  t156_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n96;
  t156_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t156_n15;
  t156_n14: return -0.095174141228199f;
  t156_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t156_n77;
  t156_n16: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t156_n34;
  t156_n17: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t156_n19;
  t156_n18: return 0.0010559739312157035f;
  t156_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t156_n31;
  t156_n20: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t156_n28;
  t156_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t156_n25;
  t156_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n24;
  t156_n23: return -0.00872237142175436f;
  t156_n24: return 0.017736777663230896f;
  t156_n25: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t156_n27;
  t156_n26: return 0.03519447520375252f;
  t156_n27: return 0.09373614192008972f;
  t156_n28: if (E.getNumReferences() >= 3309178880 /*3041.0*/) goto t156_n30;
  t156_n29: return 0.023389771580696106f;
  t156_n30: return 0.12599650025367737f;
  t156_n31: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t156_n33;
  t156_n32: return 0.01956038363277912f;
  t156_n33: return -0.13525941967964172f;
  t156_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n76;
  t156_n35: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t156_n45;
  t156_n36: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t156_n38;
  t156_n37: return 0.009843421168625355f;
  t156_n38: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t156_n42;
  t156_n39: if (E.getNumReferences() >= 3293265920 /*813.0*/) goto t156_n41;
  t156_n40: return -0.00936864223331213f;
  t156_n41: return 0.04042644053697586f;
  t156_n42: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t156_n44;
  t156_n43: return 0.012605036608874798f;
  t156_n44: return 0.1261850744485855f;
  t156_n45: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t156_n67;
  t156_n46: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t156_n66;
  t156_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t156_n49;
  t156_n48: return 0.0036875875666737556f;
  t156_n49: if (E.getFractionNameInContext() >= 3200224622 /*0.3741258978843689*/) goto t156_n53;
  t156_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t156_n52;
  t156_n51: return 0.04939659684896469f;
  t156_n52: return 0.1037105992436409f;
  t156_n53: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t156_n59;
  t156_n54: if (E.getFractionNameInContext() >= 3202467612 /*0.4409722089767456*/) goto t156_n56;
  t156_n55: return 0.013338809832930565f;
  t156_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t156_n58;
  t156_n57: return -0.24691852927207947f;
  t156_n58: return 0.10068166255950928f;
  t156_n59: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t156_n63;
  t156_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t156_n62;
  t156_n61: return -0.056975219398736954f;
  t156_n62: return 0.06513593345880508f;
  t156_n63: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t156_n65;
  t156_n64: return -0.08733619004487991f;
  t156_n65: return 0.029227694496512413f;
  t156_n66: return 0.0649876669049263f;
  t156_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t156_n71;
  t156_n68: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t156_n70;
  t156_n69: return -0.3691386580467224f;
  t156_n70: return -0.046059537678956985f;
  t156_n71: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t156_n73;
  t156_n72: return -0.04812595993280411f;
  t156_n73: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t156_n75;
  t156_n74: return 0.1183452233672142f;
  t156_n75: return -0.34167471528053284f;
  t156_n76: return 0.003309168154373765f;
  t156_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n81;
  t156_n78: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t156_n80;
  t156_n79: return -0.15954633057117462f;
  t156_n80: return -0.07794781029224396f;
  t156_n81: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t156_n83;
  t156_n82: return -0.003472334938123822f;
  t156_n83: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t156_n85;
  t156_n84: return 0.012696435675024986f;
  t156_n85: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t156_n89;
  t156_n86: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t156_n88;
  t156_n87: return -0.1621713787317276f;
  t156_n88: return 0.14159025251865387f;
  t156_n89: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t156_n93;
  t156_n90: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t156_n92;
  t156_n91: return 0.023423193022608757f;
  t156_n92: return 0.12104736268520355f;
  t156_n93: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t156_n95;
  t156_n94: return -0.6220019459724426f;
  t156_n95: return -0.06237965449690819f;
  t156_n96: return 0.0027836300432682037f;
  t156_n97: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t156_n103;
  t156_n98: if (E.getNumReferences() >= 3314894848 /*4777.0*/) goto t156_n100;
  t156_n99: return 0.07433426380157471f;
  t156_n100: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t156_n102;
  t156_n101: return 0.06837885826826096f;
  t156_n102: return 0.14969991147518158f;
  t156_n103: return 0.009852765128016472f;
  t156_n104: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t156_n156;
  t156_n105: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t156_n141;
  t156_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t156_n108;
  t156_n107: return -0.0018115566344931722f;
  t156_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t156_n110;
  t156_n109: return -0.054987262934446335f;
  t156_n110: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t156_n112;
  t156_n111: return 0.00153782544657588f;
  t156_n112: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t156_n128;
  t156_n113: if (E.getNumReferences() >= 3352190016 /*105656.5*/) goto t156_n125;
  t156_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t156_n116;
  t156_n115: return -0.003320480464026332f;
  t156_n116: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t156_n118;
  t156_n117: return 0.038809869438409805f;
  t156_n118: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t156_n120;
  t156_n119: return 0.059659335762262344f;
  t156_n120: if (E.getNumReferences() >= 3272376320 /*140.5*/) goto t156_n122;
  t156_n121: return 0.3736547529697418f;
  t156_n122: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t156_n124;
  t156_n123: return 0.20703816413879395f;
  t156_n124: return 0.10507892817258835f;
  t156_n125: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t156_n127;
  t156_n126: return 0.10520221292972565f;
  t156_n127: return -0.3911207616329193f;
  t156_n128: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t156_n130;
  t156_n129: return 0.05010247975587845f;
  t156_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t156_n132;
  t156_n131: return 0.0015173029387369752f;
  t156_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t156_n140;
  t156_n133: if (E.getNumReferences() >= 3282567168 /*336.0*/) goto t156_n135;
  t156_n134: return 0.05949005112051964f;
  t156_n135: if (E.getNumReferences() >= 3282649088 /*338.5*/) goto t156_n137;
  t156_n136: return 0.1629762202501297f;
  t156_n137: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t156_n139;
  t156_n138: return 0.12765023112297058f;
  t156_n139: return 0.04624702408909798f;
  t156_n140: return -0.00231960229575634f;
  t156_n141: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t156_n153;
  t156_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t156_n144;
  t156_n143: return -0.013065896928310394f;
  t156_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t156_n146;
  t156_n145: return 0.0018873723456636071f;
  t156_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t156_n152;
  t156_n147: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t156_n151;
  t156_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t156_n150;
  t156_n149: return 0.06178741529583931f;
  t156_n150: return 0.09809743613004684f;
  t156_n151: return 0.039496440440416336f;
  t156_n152: return 0.023411495611071587f;
  t156_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n155;
  t156_n154: return -0.004080628044903278f;
  t156_n155: return 0.004640114028006792f;
  t156_n156: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t156_n190;
  t156_n157: if (E.getNumReferences() >= 3260678144 /*54.5*/) goto t156_n185;
  t156_n158: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t156_n160;
  t156_n159: return -0.02012339048087597f;
  t156_n160: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t156_n176;
  t156_n161: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t156_n167;
  t156_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t156_n164;
  t156_n163: return -0.05973496660590172f;
  t156_n164: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t156_n166;
  t156_n165: return -0.47525709867477417f;
  t156_n166: return 0.03830215707421303f;
  t156_n167: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t156_n171;
  t156_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n170;
  t156_n169: return -0.49180540442466736f;
  t156_n170: return 0.05147877708077431f;
  t156_n171: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t156_n175;
  t156_n172: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t156_n174;
  t156_n173: return -0.008471393026411533f;
  t156_n174: return 0.051053356379270554f;
  t156_n175: return -0.15491770207881927f;
  t156_n176: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t156_n184;
  t156_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t156_n181;
  t156_n178: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t156_n180;
  t156_n179: return 0.012567562982439995f;
  t156_n180: return -1.4701908826828003f;
  t156_n181: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t156_n183;
  t156_n182: return 0.03465738892555237f;
  t156_n183: return 0.10372719168663025f;
  t156_n184: return -0.10392909497022629f;
  t156_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t156_n187;
  t156_n186: return 0.03386314958333969f;
  t156_n187: if (E.getNumReferences() >= 3266510848 /*89.5*/) goto t156_n189;
  t156_n188: return 0.11340374499559402f;
  t156_n189: return 0.11517967283725739f;
  t156_n190: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t156_n212;
  t156_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n201;
  t156_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t156_n200;
  t156_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t156_n199;
  t156_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n198;
  t156_n195: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t156_n197;
  t156_n196: return -0.13439875841140747f;
  t156_n197: return -1.5612088441848755f;
  t156_n198: return -0.058834925293922424f;
  t156_n199: return 0.08117244392633438f;
  t156_n200: return 0.06152771785855293f;
  t156_n201: if (E.getFractionNameInContext() >= 3211205746 /*0.9027777910232544*/) goto t156_n211;
  t156_n202: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t156_n204;
  t156_n203: return 0.136910542845726f;
  t156_n204: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t156_n206;
  t156_n205: return 0.022180788218975067f;
  t156_n206: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t156_n208;
  t156_n207: return 0.02247055061161518f;
  t156_n208: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t156_n210;
  t156_n209: return 0.13503971695899963f;
  t156_n210: return 0.05373595654964447f;
  t156_n211: return -0.18732111155986786f;
  t156_n212: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t156_n214;
  t156_n213: return 0.15004171431064606f;
  t156_n214: return 0.056966252624988556f;
  t156_n215: return -0.057196229696273804f;
  t156_n216: return -0.16323375701904297f;
  t156_n217: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t156_n221;
  t156_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t156_n220;
  t156_n219: return 0.025385640561580658f;
  t156_n220: return 0.05720817670226097f;
  t156_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t156_n223;
  t156_n222: return -0.14876192808151245f;
  t156_n223: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t156_n227;
  t156_n224: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t156_n226;
  t156_n225: return 0.11659363657236099f;
  t156_n226: return 0.17889167368412018f;
  t156_n227: return -0.1548699587583542f;
  t156_n228: return -0.1499992460012436f;
  t156_n229: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t156_n231;
  t156_n230: return 0.02188146486878395f;
  t156_n231: return -0.004616320598870516f;
  t156_n232: return -0.042543381452560425f;
  t156_n233: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t156_n243;
  t156_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t156_n242;
  t156_n235: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t156_n237;
  t156_n236: return 0.019331850111484528f;
  t156_n237: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t156_n241;
  t156_n238: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t156_n240;
  t156_n239: return -0.08280866593122482f;
  t156_n240: return 0.22081619501113892f;
  t156_n241: return -0.054580021649599075f;
  t156_n242: return 0.04144161567091942f;
  t156_n243: if (E.getNumReferences() >= 3365790608 /*323356.5*/) goto t156_n245;
  t156_n244: return -0.06254804879426956f;
  t156_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t156_n247;
  t156_n246: return -0.000800419831648469f;
  t156_n247: return 0.04144292324781418f;
  t156_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t156_n250;
  t156_n249: return 0.0018558921292424202f;
  t156_n250: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t156_n254;
  t156_n251: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t156_n253;
  t156_n252: return 0.0060021583922207355f;
  t156_n253: return -0.06320279091596603f;
  t156_n254: return -0.17419840395450592f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree157(const Example& E) {
  t157_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t157_n54;
  t157_n1: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t157_n53;
  t157_n2: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t157_n44;
  t157_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t157_n5;
  t157_n4: return -0.0006382806459441781f;
  t157_n5: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t157_n11;
  t157_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t157_n8;
  t157_n7: return -0.024104272946715355f;
  t157_n8: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t157_n10;
  t157_n9: return -0.005608795676380396f;
  t157_n10: return 0.014468175359070301f;
  t157_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n13;
  t157_n12: return -0.035968415439128876f;
  t157_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n43;
  t157_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t157_n34;
  t157_n15: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t157_n25;
  t157_n16: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t157_n20;
  t157_n17: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t157_n19;
  t157_n18: return 0.05746909976005554f;
  t157_n19: return 0.012452660128474236f;
  t157_n20: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t157_n24;
  t157_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t157_n23;
  t157_n22: return 0.16336943209171295f;
  t157_n23: return 0.12709949910640717f;
  t157_n24: return 0.06274736672639847f;
  t157_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t157_n33;
  t157_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t157_n30;
  t157_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t157_n29;
  t157_n28: return -0.026144856587052345f;
  t157_n29: return -0.030505545437335968f;
  t157_n30: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t157_n32;
  t157_n31: return 0.014407742768526077f;
  t157_n32: return -0.06274145841598511f;
  t157_n33: return 0.014824877493083477f;
  t157_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t157_n42;
  t157_n35: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t157_n41;
  t157_n36: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t157_n40;
  t157_n37: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t157_n39;
  t157_n38: return 0.12681223452091217f;
  t157_n39: return -0.004076449200510979f;
  t157_n40: return 0.06484860181808472f;
  t157_n41: return -0.09732396900653839f;
  t157_n42: return 0.010471196845173836f;
  t157_n43: return 0.028144896030426025f;
  t157_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t157_n46;
  t157_n45: return -0.04062999039888382f;
  t157_n46: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t157_n50;
  t157_n47: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t157_n49;
  t157_n48: return 0.053805604577064514f;
  t157_n49: return 0.11288735270500183f;
  t157_n50: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t157_n52;
  t157_n51: return -0.05807052552700043f;
  t157_n52: return 0.1379023641347885f;
  t157_n53: return 0.03728695586323738f;
  t157_n54: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t157_n238;
  t157_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t157_n131;
  t157_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t157_n60;
  t157_n57: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t157_n59;
  t157_n58: return -0.05812457203865051f;
  t157_n59: return -0.016309447586536407f;
  t157_n60: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t157_n90;
  t157_n61: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t157_n89;
  t157_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t157_n80;
  t157_n63: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t157_n71;
  t157_n64: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t157_n70;
  t157_n65: if (E.getNumReferences() >= 3285090304 /*413.0*/) goto t157_n67;
  t157_n66: return 0.012638699263334274f;
  t157_n67: if (E.getNumReferences() >= 3326212096 /*12414.0*/) goto t157_n69;
  t157_n68: return -0.47656476497650146f;
  t157_n69: return 0.25358492136001587f;
  t157_n70: return -0.06941306591033936f;
  t157_n71: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t157_n79;
  t157_n72: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t157_n78;
  t157_n73: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t157_n75;
  t157_n74: return -0.02597801946103573f;
  t157_n75: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t157_n77;
  t157_n76: return 0.0062348204664886f;
  t157_n77: return -0.01782154105603695f;
  t157_n78: return -0.18867909908294678f;
  t157_n79: return 0.009252536110579967f;
  t157_n80: if (E.getNumReferences() >= 3306575872 /*2405.5*/) goto t157_n88;
  t157_n81: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t157_n85;
  t157_n82: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t157_n84;
  t157_n83: return 0.02873258665204048f;
  t157_n84: return -0.04123006388545036f;
  t157_n85: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t157_n87;
  t157_n86: return 0.12228383123874664f;
  t157_n87: return -0.02400975488126278f;
  t157_n88: return 0.06455560773611069f;
  t157_n89: return 0.0030620230827480555f;
  t157_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n100;
  t157_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t157_n93;
  t157_n92: return -0.03906036168336868f;
  t157_n93: if (E.getNumReferences() >= 3309658112 /*3158.0*/) goto t157_n99;
  t157_n94: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t157_n98;
  t157_n95: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t157_n97;
  t157_n96: return 0.005434026475995779f;
  t157_n97: return -0.03397883474826813f;
  t157_n98: return 0.047558773308992386f;
  t157_n99: return 0.042060788720846176f;
  t157_n100: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t157_n130;
  t157_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t157_n103;
  t157_n102: return -0.004049049690365791f;
  t157_n103: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t157_n125;
  t157_n104: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t157_n114;
  t157_n105: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t157_n109;
  t157_n106: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t157_n108;
  t157_n107: return 0.010464057326316833f;
  t157_n108: return -0.00899704173207283f;
  t157_n109: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t157_n113;
  t157_n110: if (E.getNumReferences() >= 3316233216 /*5430.5*/) goto t157_n112;
  t157_n111: return 0.03470873460173607f;
  t157_n112: return 0.0002369472204009071f;
  t157_n113: return 0.006185967940837145f;
  t157_n114: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t157_n122;
  t157_n115: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t157_n119;
  t157_n116: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t157_n118;
  t157_n117: return -0.008008151315152645f;
  t157_n118: return 0.009834044612944126f;
  t157_n119: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t157_n121;
  t157_n120: return 0.013336907140910625f;
  t157_n121: return 0.004023315850645304f;
  t157_n122: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t157_n124;
  t157_n123: return -0.19520336389541626f;
  t157_n124: return 0.024261794984340668f;
  t157_n125: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t157_n129;
  t157_n126: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t157_n128;
  t157_n127: return 0.04364785924553871f;
  t157_n128: return 0.10568385571241379f;
  t157_n129: return -0.08443521708250046f;
  t157_n130: return -0.01939358003437519f;
  t157_n131: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t157_n175;
  t157_n132: if (E.getFractionNameInContext() >= 3199184816 /*0.343137264251709*/) goto t157_n156;
  t157_n133: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t157_n155;
  t157_n134: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t157_n154;
  t157_n135: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t157_n137;
  t157_n136: return 0.1260966658592224f;
  t157_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t157_n139;
  t157_n138: return 0.005529902875423431f;
  t157_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t157_n153;
  t157_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n152;
  t157_n141: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t157_n149;
  t157_n142: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t157_n148;
  t157_n143: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t157_n145;
  t157_n144: return -0.07365074008703232f;
  t157_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n147;
  t157_n146: return -0.546181857585907f;
  t157_n147: return 0.011451204307377338f;
  t157_n148: return 0.011427718214690685f;
  t157_n149: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t157_n151;
  t157_n150: return 0.026072870939970016f;
  t157_n151: return -0.008853365667164326f;
  t157_n152: return 0.0357554629445076f;
  t157_n153: return 0.03109312243759632f;
  t157_n154: return -0.014806232415139675f;
  t157_n155: return 0.018463239073753357f;
  t157_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t157_n158;
  t157_n157: return -0.03001166321337223f;
  t157_n158: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t157_n172;
  t157_n159: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t157_n161;
  t157_n160: return 0.020548637956380844f;
  t157_n161: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t157_n165;
  t157_n162: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t157_n164;
  t157_n163: return -0.08439627289772034f;
  t157_n164: return -0.015862487256526947f;
  t157_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t157_n167;
  t157_n166: return 0.017200559377670288f;
  t157_n167: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t157_n171;
  t157_n168: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t157_n170;
  t157_n169: return 0.12042906880378723f;
  t157_n170: return 0.09656748920679092f;
  t157_n171: return 0.04035573825240135f;
  t157_n172: if (E.getNumReferences() >= 3261726720 /*58.5*/) goto t157_n174;
  t157_n173: return 0.11923610419034958f;
  t157_n174: return 0.06378896534442902f;
  t157_n175: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t157_n189;
  t157_n176: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t157_n178;
  t157_n177: return -0.0002539501874707639f;
  t157_n178: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t157_n180;
  t157_n179: return 0.0213338490575552f;
  t157_n180: if (E.getFractionNameInContext() >= 3207459552 /*0.6794872283935547*/) goto t157_n184;
  t157_n181: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t157_n183;
  t157_n182: return 0.11033952236175537f;
  t157_n183: return -0.5113992691040039f;
  t157_n184: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t157_n188;
  t157_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t157_n187;
  t157_n186: return 0.034450046718120575f;
  t157_n187: return 0.12936371564865112f;
  t157_n188: return -0.009781556203961372f;
  t157_n189: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t157_n237;
  t157_n190: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t157_n208;
  t157_n191: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t157_n197;
  t157_n192: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t157_n196;
  t157_n193: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t157_n195;
  t157_n194: return -0.0221867598593235f;
  t157_n195: return 0.004112062510102987f;
  t157_n196: return 0.01863601803779602f;
  t157_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t157_n199;
  t157_n198: return 0.025652481243014336f;
  t157_n199: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t157_n201;
  t157_n200: return -0.019652515649795532f;
  t157_n201: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t157_n207;
  t157_n202: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t157_n204;
  t157_n203: return 0.12181895226240158f;
  t157_n204: if (E.getFractionNameInContext() >= 3197027564 /*0.2788461446762085*/) goto t157_n206;
  t157_n205: return -0.05501777306199074f;
  t157_n206: return 0.12329541146755219f;
  t157_n207: return 0.03802942484617233f;
  t157_n208: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t157_n236;
  t157_n209: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t157_n235;
  t157_n210: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t157_n232;
  t157_n211: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t157_n213;
  t157_n212: return 0.020942602306604385f;
  t157_n213: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t157_n229;
  t157_n214: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t157_n216;
  t157_n215: return -0.021803461015224457f;
  t157_n216: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t157_n218;
  t157_n217: return -0.0736427754163742f;
  t157_n218: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t157_n220;
  t157_n219: return 0.04565109312534332f;
  t157_n220: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t157_n226;
  t157_n221: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t157_n225;
  t157_n222: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t157_n224;
  t157_n223: return 0.13959041237831116f;
  t157_n224: return -0.019141942262649536f;
  t157_n225: return 0.12011604011058807f;
  t157_n226: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t157_n228;
  t157_n227: return -1.1286765336990356f;
  t157_n228: return -0.03338524326682091f;
  t157_n229: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t157_n231;
  t157_n230: return -0.19727978110313416f;
  t157_n231: return -0.07526816427707672f;
  t157_n232: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t157_n234;
  t157_n233: return 0.09643866866827011f;
  t157_n234: return 0.06319216638803482f;
  t157_n235: return 0.04815775528550148f;
  t157_n236: return -0.06825366616249084f;
  t157_n237: return 0.0018834793008863926f;
  t157_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t157_n244;
  t157_n239: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t157_n243;
  t157_n240: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t157_n242;
  t157_n241: return -0.01637124828994274f;
  t157_n242: return 0.004017098806798458f;
  t157_n243: return 0.007105841767042875f;
  t157_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t157_n246;
  t157_n245: return 0.007089162245392799f;
  t157_n246: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t157_n252;
  t157_n247: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t157_n251;
  t157_n248: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t157_n250;
  t157_n249: return 0.02872108481824398f;
  t157_n250: return -0.07166194170713425f;
  t157_n251: return 0.07653362303972244f;
  t157_n252: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t157_n254;
  t157_n253: return 0.020094478502869606f;
  t157_n254: return 0.11238609254360199f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree158(const Example& E) {
  t158_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t158_n240;
  t158_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t158_n39;
  t158_n2: if (E.getNumReferences() >= 3314787328 /*4724.5*/) goto t158_n32;
  t158_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t158_n5;
  t158_n4: return -0.0002226712240371853f;
  t158_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t158_n17;
  t158_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n10;
  t158_n7: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t158_n9;
  t158_n8: return -0.12366608530282974f;
  t158_n9: return -0.0037011802196502686f;
  t158_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t158_n12;
  t158_n11: return 0.003324672346934676f;
  t158_n12: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t158_n16;
  t158_n13: if (E.getNumReferences() >= 3310284800 /*3311.0*/) goto t158_n15;
  t158_n14: return 0.021114332601428032f;
  t158_n15: return -0.22690457105636597f;
  t158_n16: return 0.12837038934230804f;
  t158_n17: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t158_n25;
  t158_n18: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t158_n22;
  t158_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t158_n21;
  t158_n20: return 0.007218272890895605f;
  t158_n21: return 0.06591590493917465f;
  t158_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n24;
  t158_n23: return -0.12230315804481506f;
  t158_n24: return -0.012718942947685719f;
  t158_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n29;
  t158_n26: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t158_n28;
  t158_n27: return -0.023788336664438248f;
  t158_n28: return 0.13271848857402802f;
  t158_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n31;
  t158_n30: return -0.0008582552545703948f;
  t158_n31: return 0.030999984592199326f;
  t158_n32: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t158_n38;
  t158_n33: if (E.getNumReferences() >= 3314894848 /*4777.0*/) goto t158_n35;
  t158_n34: return 0.07540139555931091f;
  t158_n35: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t158_n37;
  t158_n36: return 0.10395807772874832f;
  t158_n37: return 0.17665530741214752f;
  t158_n38: return 0.011661228723824024f;
  t158_n39: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t158_n239;
  t158_n40: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t158_n54;
  t158_n41: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t158_n53;
  t158_n42: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t158_n46;
  t158_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t158_n45;
  t158_n44: return -0.14956089854240417f;
  t158_n45: return 0.12222307175397873f;
  t158_n46: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t158_n52;
  t158_n47: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t158_n49;
  t158_n48: return 0.0019824530463665724f;
  t158_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t158_n51;
  t158_n50: return -0.009473097510635853f;
  t158_n51: return 0.03411813825368881f;
  t158_n52: return -0.0019802870228886604f;
  t158_n53: return -0.12839217483997345f;
  t158_n54: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t158_n222;
  t158_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t158_n165;
  t158_n56: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t158_n74;
  t158_n57: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t158_n59;
  t158_n58: return 0.0035067100543528795f;
  t158_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t158_n69;
  t158_n60: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t158_n64;
  t158_n61: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n63;
  t158_n62: return 0.0474553108215332f;
  t158_n63: return -0.008199287578463554f;
  t158_n64: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t158_n68;
  t158_n65: if (E.getNumReferences() >= 3255042048 /*33.0*/) goto t158_n67;
  t158_n66: return 0.18434949219226837f;
  t158_n67: return 0.09675250947475433f;
  t158_n68: return 0.04475310444831848f;
  t158_n69: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t158_n73;
  t158_n70: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t158_n72;
  t158_n71: return 0.1328619122505188f;
  t158_n72: return 0.10667972266674042f;
  t158_n73: return -0.2796494662761688f;
  t158_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n124;
  t158_n75: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t158_n85;
  t158_n76: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t158_n78;
  t158_n77: return 0.013663643039762974f;
  t158_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n84;
  t158_n79: if (E.getNumReferences() >= 3265855488 /*84.5*/) goto t158_n81;
  t158_n80: return -0.0913417637348175f;
  t158_n81: if (E.getNumReferences() >= 3285073920 /*412.5*/) goto t158_n83;
  t158_n82: return -0.6995878219604492f;
  t158_n83: return 0.006405914667993784f;
  t158_n84: return -0.008710064925253391f;
  t158_n85: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t158_n101;
  t158_n86: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t158_n90;
  t158_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n89;
  t158_n88: return -0.3387572169303894f;
  t158_n89: return -0.07410869747400284f;
  t158_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n100;
  t158_n91: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t158_n95;
  t158_n92: if (E.getNumReferences() >= 3260284928 /*53.0*/) goto t158_n94;
  t158_n93: return -0.10695858299732208f;
  t158_n94: return -0.02569579891860485f;
  t158_n95: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n97;
  t158_n96: return 0.009737773798406124f;
  t158_n97: if (E.getNumReferences() >= 3323816960 /*10075.0*/) goto t158_n99;
  t158_n98: return 0.12502330541610718f;
  t158_n99: return -0.05798531323671341f;
  t158_n100: return 0.02726205810904503f;
  t158_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t158_n115;
  t158_n102: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t158_n112;
  t158_n103: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t158_n107;
  t158_n104: if (E.getNumReferences() >= 3339744256 /*36980.0*/) goto t158_n106;
  t158_n105: return -0.059100281447172165f;
  t158_n106: return 0.02139766700565815f;
  t158_n107: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t158_n109;
  t158_n108: return 0.000803205359261483f;
  t158_n109: if (E.getNumReferences() >= 3345826304 /*60738.0*/) goto t158_n111;
  t158_n110: return 0.03469937667250633f;
  t158_n111: return 0.14883174002170563f;
  t158_n112: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t158_n114;
  t158_n113: return -0.11327558010816574f;
  t158_n114: return -0.022358234971761703f;
  t158_n115: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t158_n117;
  t158_n116: return 0.005801152903586626f;
  t158_n117: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t158_n123;
  t158_n118: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t158_n120;
  t158_n119: return -0.015244640409946442f;
  t158_n120: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t158_n122;
  t158_n121: return -0.36023908853530884f;
  t158_n122: return -0.1636299043893814f;
  t158_n123: return 0.0010315966792404652f;
  t158_n124: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t158_n140;
  t158_n125: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t158_n133;
  t158_n126: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t158_n132;
  t158_n127: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t158_n131;
  t158_n128: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n130;
  t158_n129: return -0.002700967714190483f;
  t158_n130: return 0.06926180422306061f;
  t158_n131: return -0.02501862868666649f;
  t158_n132: return 0.01203928142786026f;
  t158_n133: if (E.getNumReferences() >= 3320255488 /*7394.5*/) goto t158_n139;
  t158_n134: if (E.getNumReferences() >= 3316911104 /*5761.5*/) goto t158_n136;
  t158_n135: return -0.058809760957956314f;
  t158_n136: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t158_n138;
  t158_n137: return -0.55006343126297f;
  t158_n138: return -1.0780378580093384f;
  t158_n139: return 0.03306485712528229f;
  t158_n140: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t158_n164;
  t158_n141: if (E.getNumReferences() >= 3265921024 /*85.0*/) goto t158_n161;
  t158_n142: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t158_n158;
  t158_n143: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t158_n145;
  t158_n144: return -0.05471741408109665f;
  t158_n145: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t158_n157;
  t158_n146: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t158_n150;
  t158_n147: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t158_n149;
  t158_n148: return 0.06637561321258545f;
  t158_n149: return 0.34812799096107483f;
  t158_n150: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t158_n152;
  t158_n151: return 0.004829050973057747f;
  t158_n152: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t158_n156;
  t158_n153: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t158_n155;
  t158_n154: return 0.1323307305574417f;
  t158_n155: return 0.0031802672892808914f;
  t158_n156: return -0.105026014149189f;
  t158_n157: return 0.009391154162585735f;
  t158_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t158_n160;
  t158_n159: return 0.05055908486247063f;
  t158_n160: return 0.06096611171960831f;
  t158_n161: if (E.getNumReferences() >= 3293503488 /*827.5*/) goto t158_n163;
  t158_n162: return -0.10742127895355225f;
  t158_n163: return 0.01930350251495838f;
  t158_n164: return -0.0036225328221917152f;
  t158_n165: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t158_n219;
  t158_n166: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t158_n206;
  t158_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t158_n181;
  t158_n168: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t158_n180;
  t158_n169: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t158_n171;
  t158_n170: return -0.011651336215436459f;
  t158_n171: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t158_n175;
  t158_n172: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t158_n174;
  t158_n173: return -0.01345216017216444f;
  t158_n174: return 0.105338454246521f;
  t158_n175: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t158_n179;
  t158_n176: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t158_n178;
  t158_n177: return 0.06592462211847305f;
  t158_n178: return 0.12133433669805527f;
  t158_n179: return -0.018853511661291122f;
  t158_n180: return -0.03760036826133728f;
  t158_n181: if (E.getNumReferences() >= 3277062144 /*212.0*/) goto t158_n205;
  t158_n182: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t158_n204;
  t158_n183: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t158_n203;
  t158_n184: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t158_n190;
  t158_n185: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n187;
  t158_n186: return 0.0067311543971300125f;
  t158_n187: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t158_n189;
  t158_n188: return 0.062158603221178055f;
  t158_n189: return -0.14181700348854065f;
  t158_n190: if (E.getFractionNameInContext() >= 3181279720 /*0.0773809552192688*/) goto t158_n198;
  t158_n191: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t158_n193;
  t158_n192: return -0.11117088049650192f;
  t158_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n197;
  t158_n194: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t158_n196;
  t158_n195: return 0.14867845177650452f;
  t158_n196: return -0.13769520819187164f;
  t158_n197: return -0.0010617828229442239f;
  t158_n198: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t158_n200;
  t158_n199: return -0.08141852170228958f;
  t158_n200: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n202;
  t158_n201: return 0.005490880459547043f;
  t158_n202: return 0.10356160998344421f;
  t158_n203: return -0.04864117503166199f;
  t158_n204: return -0.02042856626212597f;
  t158_n205: return 0.13083603978157043f;
  t158_n206: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t158_n218;
  t158_n207: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t158_n209;
  t158_n208: return 0.029557837173342705f;
  t158_n209: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t158_n213;
  t158_n210: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n212;
  t158_n211: return -0.34893688559532166f;
  t158_n212: return 0.07143149524927139f;
  t158_n213: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t158_n217;
  t158_n214: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t158_n216;
  t158_n215: return 0.12306094169616699f;
  t158_n216: return 0.07737801223993301f;
  t158_n217: return -0.08300302922725677f;
  t158_n218: return 0.004962845705449581f;
  t158_n219: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n221;
  t158_n220: return 0.004121131729334593f;
  t158_n221: return -0.08612290024757385f;
  t158_n222: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t158_n236;
  t158_n223: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t158_n225;
  t158_n224: return -0.06888286024332047f;
  t158_n225: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t158_n233;
  t158_n226: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t158_n232;
  t158_n227: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n231;
  t158_n228: if (E.getFractionNameInContext() >= 3182156850 /*0.0839160829782486*/) goto t158_n230;
  t158_n229: return 0.07015257328748703f;
  t158_n230: return 0.11879970878362656f;
  t158_n231: return -0.17199723422527313f;
  t158_n232: return 0.019264815375208855f;
  t158_n233: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t158_n235;
  t158_n234: return -0.15241718292236328f;
  t158_n235: return 0.06430355459451675f;
  t158_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t158_n238;
  t158_n237: return -0.08763261884450912f;
  t158_n238: return 0.028703797608613968f;
  t158_n239: return -0.06283995509147644f;
  t158_n240: if (E.getNumReferences() >= 3349388864 /*83772.5*/) goto t158_n246;
  t158_n241: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t158_n245;
  t158_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_TopLevel))) goto t158_n244;
  t158_n243: return 0.045549388974905014f;
  t158_n244: return 0.1566171646118164f;
  t158_n245: return 0.0019470562692731619f;
  t158_n246: if (E.getNumReferences() >= 3351977984 /*104000.0*/) goto t158_n250;
  t158_n247: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t158_n249;
  t158_n248: return 0.07233943790197372f;
  t158_n249: return 0.09810435026884079f;
  t158_n250: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t158_n252;
  t158_n251: return 0.03885587677359581f;
  t158_n252: if (E.getNumReferences() >= 3356244608 /*143594.0*/) goto t158_n254;
  t158_n253: return -0.5737907886505127f;
  t158_n254: return -0.06335800886154175f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree159(const Example& E) {
  t159_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t159_n246;
  t159_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t159_n59;
  t159_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t159_n52;
  t159_n3: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t159_n25;
  t159_n4: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t159_n6;
  t159_n5: return -0.0009959019953384995f;
  t159_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t159_n20;
  t159_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t159_n19;
  t159_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t159_n12;
  t159_n9: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t159_n11;
  t159_n10: return -0.07918111234903336f;
  t159_n11: return -0.16652703285217285f;
  t159_n12: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t159_n16;
  t159_n13: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t159_n15;
  t159_n14: return 0.03813846781849861f;
  t159_n15: return -0.03373274952173233f;
  t159_n16: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t159_n18;
  t159_n17: return 0.07504748553037643f;
  t159_n18: return -0.2217353731393814f;
  t159_n19: return -0.02041168138384819f;
  t159_n20: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t159_n24;
  t159_n21: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t159_n23;
  t159_n22: return -0.0006842185975983739f;
  t159_n23: return 0.028208432719111443f;
  t159_n24: return -0.035366930067539215f;
  t159_n25: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t159_n29;
  t159_n26: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t159_n28;
  t159_n27: return -0.010409476235508919f;
  t159_n28: return 0.012657398357987404f;
  t159_n29: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t159_n47;
  t159_n30: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t159_n36;
  t159_n31: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t159_n33;
  t159_n32: return 0.007984207011759281f;
  t159_n33: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t159_n35;
  t159_n34: return 0.0241637472063303f;
  t159_n35: return -0.003796959761530161f;
  t159_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t159_n40;
  t159_n37: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t159_n39;
  t159_n38: return -0.2068641632795334f;
  t159_n39: return -0.025057082995772362f;
  t159_n40: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t159_n44;
  t159_n41: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t159_n43;
  t159_n42: return 0.022854700684547424f;
  t159_n43: return -0.01720256730914116f;
  t159_n44: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t159_n46;
  t159_n45: return 0.03868945688009262f;
  t159_n46: return -0.02151675894856453f;
  t159_n47: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t159_n51;
  t159_n48: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t159_n50;
  t159_n49: return 0.10820269584655762f;
  t159_n50: return 0.07005009800195694f;
  t159_n51: return 0.017215432599186897f;
  t159_n52: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t159_n54;
  t159_n53: return 0.00018971998360939324f;
  t159_n54: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t159_n58;
  t159_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t159_n57;
  t159_n56: return -0.1546860933303833f;
  t159_n57: return 0.011027421802282333f;
  t159_n58: return -0.0034218274522572756f;
  t159_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t159_n173;
  t159_n60: if (E.getNumReferences() >= 3324678656 /*10916.5*/) goto t159_n160;
  t159_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t159_n63;
  t159_n62: return -0.008206377737224102f;
  t159_n63: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t159_n159;
  t159_n64: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t159_n122;
  t159_n65: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t159_n121;
  t159_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t159_n92;
  t159_n67: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t159_n77;
  t159_n68: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t159_n72;
  t159_n69: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t159_n71;
  t159_n70: return 0.01117137260735035f;
  t159_n71: return -0.0553273968398571f;
  t159_n72: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t159_n74;
  t159_n73: return -0.030037201941013336f;
  t159_n74: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t159_n76;
  t159_n75: return -0.0045267450623214245f;
  t159_n76: return -0.040496814996004105f;
  t159_n77: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t159_n87;
  t159_n78: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t159_n80;
  t159_n79: return -0.07721266895532608f;
  t159_n80: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t159_n82;
  t159_n81: return 0.05060591921210289f;
  t159_n82: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t159_n86;
  t159_n83: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t159_n85;
  t159_n84: return -0.37373000383377075f;
  t159_n85: return 0.05967872962355614f;
  t159_n86: return -0.3263503313064575f;
  t159_n87: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t159_n89;
  t159_n88: return -0.2147361785173416f;
  t159_n89: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t159_n91;
  t159_n90: return 0.09650299698114395f;
  t159_n91: return 0.013539246283471584f;
  t159_n92: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t159_n94;
  t159_n93: return 0.022459451109170914f;
  t159_n94: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t159_n114;
  t159_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t159_n109;
  t159_n96: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t159_n102;
  t159_n97: if (E.getNumReferences() >= 3307630592 /*2663.0*/) goto t159_n101;
  t159_n98: if (E.getNumReferences() >= 3307091968 /*2531.5*/) goto t159_n100;
  t159_n99: return 0.0047057741321623325f;
  t159_n100: return 0.13039375841617584f;
  t159_n101: return -0.12736015021800995f;
  t159_n102: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t159_n104;
  t159_n103: return -0.09541778266429901f;
  t159_n104: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t159_n108;
  t159_n105: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t159_n107;
  t159_n106: return -0.014357154257595539f;
  t159_n107: return -0.10162149369716644f;
  t159_n108: return 0.0028077003080397844f;
  t159_n109: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t159_n113;
  t159_n110: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t159_n112;
  t159_n111: return 0.022244293242692947f;
  t159_n112: return 0.12524735927581787f;
  t159_n113: return 0.1937878131866455f;
  t159_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t159_n116;
  t159_n115: return -0.06459184736013412f;
  t159_n116: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t159_n120;
  t159_n117: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t159_n119;
  t159_n118: return 0.026097353547811508f;
  t159_n119: return 0.0602777823805809f;
  t159_n120: return -0.006091568153351545f;
  t159_n121: return 0.0023805592209100723f;
  t159_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t159_n124;
  t159_n123: return -0.010986566543579102f;
  t159_n124: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t159_n158;
  t159_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t159_n127;
  t159_n126: return -0.008765424601733685f;
  t159_n127: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t159_n139;
  t159_n128: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t159_n136;
  t159_n129: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t159_n131;
  t159_n130: return 0.010399299673736095f;
  t159_n131: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t159_n133;
  t159_n132: return -0.005921339616179466f;
  t159_n133: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t159_n135;
  t159_n134: return -1.0194538831710815f;
  t159_n135: return -0.05853233486413956f;
  t159_n136: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t159_n138;
  t159_n137: return 0.028658321127295494f;
  t159_n138: return 0.00857981014996767f;
  t159_n139: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t159_n143;
  t159_n140: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t159_n142;
  t159_n141: return -0.007153646554797888f;
  t159_n142: return 0.01178657915443182f;
  t159_n143: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t159_n149;
  t159_n144: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t159_n148;
  t159_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t159_n147;
  t159_n146: return 0.011562252417206764f;
  t159_n147: return -0.09389780461788177f;
  t159_n148: return 0.06879006326198578f;
  t159_n149: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t159_n157;
  t159_n150: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t159_n152;
  t159_n151: return 0.003545230720192194f;
  t159_n152: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t159_n156;
  t159_n153: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t159_n155;
  t159_n154: return 0.03879566863179207f;
  t159_n155: return 0.0929967612028122f;
  t159_n156: return -0.07914910465478897f;
  t159_n157: return 0.024547496810555458f;
  t159_n158: return -0.021546095609664917f;
  t159_n159: return 0.006078989710658789f;
  t159_n160: if (E.getNumReferences() >= 3331016704 /*17828.0*/) goto t159_n172;
  t159_n161: if (E.getNumReferences() >= 3330960128 /*17717.5*/) goto t159_n165;
  t159_n162: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t159_n164;
  t159_n163: return 0.0016144493129104376f;
  t159_n164: return 0.08278954029083252f;
  t159_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t159_n167;
  t159_n166: return 0.15717773139476776f;
  t159_n167: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t159_n171;
  t159_n168: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t159_n170;
  t159_n169: return 0.15544553101062775f;
  t159_n170: return -0.15653443336486816f;
  t159_n171: return 0.13802367448806763f;
  t159_n172: return 0.010050454176962376f;
  t159_n173: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t159_n175;
  t159_n174: return 0.11969831585884094f;
  t159_n175: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t159_n245;
  t159_n176: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t159_n180;
  t159_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t159_n179;
  t159_n178: return -0.0007048065308481455f;
  t159_n179: return 0.010212486609816551f;
  t159_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t159_n198;
  t159_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t159_n191;
  t159_n182: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t159_n190;
  t159_n183: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t159_n187;
  t159_n184: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t159_n186;
  t159_n185: return -0.139473557472229f;
  t159_n186: return -0.013316758908331394f;
  t159_n187: if (E.getNumReferences() >= 3373331040 /*593702.0*/) goto t159_n189;
  t159_n188: return 0.02113582193851471f;
  t159_n189: return -0.3295174837112427f;
  t159_n190: return -0.0921393409371376f;
  t159_n191: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t159_n197;
  t159_n192: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t159_n194;
  t159_n193: return 0.0488244891166687f;
  t159_n194: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t159_n196;
  t159_n195: return -0.015059729106724262f;
  t159_n196: return 0.0334184467792511f;
  t159_n197: return -0.01267552375793457f;
  t159_n198: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t159_n236;
  t159_n199: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t159_n235;
  t159_n200: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t159_n204;
  t159_n201: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t159_n203;
  t159_n202: return -0.017373323440551758f;
  t159_n203: return -6.2094645500183105f;
  t159_n204: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t159_n212;
  t159_n205: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t159_n211;
  t159_n206: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t159_n210;
  t159_n207: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t159_n209;
  t159_n208: return 0.06402098387479782f;
  t159_n209: return -0.05990537256002426f;
  t159_n210: return 0.13993531465530396f;
  t159_n211: return -0.1661318838596344f;
  t159_n212: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t159_n228;
  t159_n213: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t159_n223;
  t159_n214: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t159_n220;
  t159_n215: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t159_n219;
  t159_n216: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t159_n218;
  t159_n217: return -0.03726831451058388f;
  t159_n218: return -0.3267916440963745f;
  t159_n219: return 0.07672062516212463f;
  t159_n220: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t159_n222;
  t159_n221: return 0.05689319595694542f;
  t159_n222: return 0.13307377696037292f;
  t159_n223: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t159_n227;
  t159_n224: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t159_n226;
  t159_n225: return -0.2695808410644531f;
  t159_n226: return 0.09031153470277786f;
  t159_n227: return -0.3117411434650421f;
  t159_n228: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t159_n234;
  t159_n229: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t159_n231;
  t159_n230: return 0.04733788222074509f;
  t159_n231: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t159_n233;
  t159_n232: return 0.12108208984136581f;
  t159_n233: return -0.10946565121412277f;
  t159_n234: return -0.008471250534057617f;
  t159_n235: return 0.0019696007948368788f;
  t159_n236: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t159_n242;
  t159_n237: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t159_n241;
  t159_n238: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t159_n240;
  t159_n239: return 0.004166717641055584f;
  t159_n240: return -0.22358287870883942f;
  t159_n241: return 0.01993325725197792f;
  t159_n242: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t159_n244;
  t159_n243: return 0.008322419598698616f;
  t159_n244: return 0.016888229176402092f;
  t159_n245: return -0.0001779713056748733f;
  t159_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t159_n248;
  t159_n247: return 0.001988952048122883f;
  t159_n248: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t159_n250;
  t159_n249: return 0.00519005861133337f;
  t159_n250: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t159_n254;
  t159_n251: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t159_n253;
  t159_n252: return -0.17915482819080353f;
  t159_n253: return 0.10785268247127533f;
  t159_n254: return -0.1632150262594223f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree160(const Example& E) {
  t160_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t160_n254;
  t160_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t160_n251;
  t160_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t160_n32;
  t160_n3: if (E.getNumReferences() >= 3275587584 /*189.5*/) goto t160_n23;
  t160_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n22;
  t160_n5: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t160_n7;
  t160_n6: return -0.003526815678924322f;
  t160_n7: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t160_n17;
  t160_n8: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t160_n14;
  t160_n9: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t160_n13;
  t160_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t160_n12;
  t160_n11: return 0.017152559012174606f;
  t160_n12: return 0.0796082392334938f;
  t160_n13: return 0.041929785162210464f;
  t160_n14: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t160_n16;
  t160_n15: return -0.001930848229676485f;
  t160_n16: return 0.029067346826195717f;
  t160_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n21;
  t160_n18: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t160_n20;
  t160_n19: return -0.04304214194417f;
  t160_n20: return 0.022180335596203804f;
  t160_n21: return 0.005245938431471586f;
  t160_n22: return 0.0009674333268776536f;
  t160_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t160_n31;
  t160_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n30;
  t160_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t160_n29;
  t160_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t160_n28;
  t160_n27: return -0.027941862121224403f;
  t160_n28: return 0.011780048720538616f;
  t160_n29: return 0.16022180020809174f;
  t160_n30: return 0.011132658459246159f;
  t160_n31: return -0.00452166423201561f;
  t160_n32: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t160_n50;
  t160_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n49;
  t160_n34: if (E.getNumReferences() >= 3267428352 /*96.5*/) goto t160_n46;
  t160_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t160_n37;
  t160_n36: return -0.037347257137298584f;
  t160_n37: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t160_n41;
  t160_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n40;
  t160_n39: return -0.12608793377876282f;
  t160_n40: return 0.037675537168979645f;
  t160_n41: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t160_n43;
  t160_n42: return -0.03276602178812027f;
  t160_n43: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t160_n45;
  t160_n44: return 0.1375616043806076f;
  t160_n45: return 0.054626915603876114f;
  t160_n46: if (E.getNumReferences() >= 3267559424 /*97.5*/) goto t160_n48;
  t160_n47: return 0.8495182394981384f;
  t160_n48: return 0.05443872883915901f;
  t160_n49: return 0.008501260541379452f;
  t160_n50: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t160_n212;
  t160_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n53;
  t160_n52: return -0.036937493830919266f;
  t160_n53: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t160_n205;
  t160_n54: if (E.getNumReferences() >= 3334963968 /*25537.5*/) goto t160_n192;
  t160_n55: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t160_n189;
  t160_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n66;
  t160_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword))) goto t160_n59;
  t160_n58: return -0.006534113083034754f;
  t160_n59: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t160_n61;
  t160_n60: return -0.0012939547887071967f;
  t160_n61: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t160_n65;
  t160_n62: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t160_n64;
  t160_n63: return 0.011862474493682384f;
  t160_n64: return 0.09868917614221573f;
  t160_n65: return -0.07441926747560501f;
  t160_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t160_n94;
  t160_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n71;
  t160_n68: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t160_n70;
  t160_n69: return -0.08017508685588837f;
  t160_n70: return -0.16546322405338287f;
  t160_n71: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t160_n93;
  t160_n72: if (E.getNumReferences() >= 3329227264 /*15358.5*/) goto t160_n88;
  t160_n73: if (E.getNumReferences() >= 3327209984 /*13388.5*/) goto t160_n83;
  t160_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t160_n76;
  t160_n75: return -0.10239163041114807f;
  t160_n76: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t160_n78;
  t160_n77: return 0.0016327921766787767f;
  t160_n78: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t160_n82;
  t160_n79: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t160_n81;
  t160_n80: return 0.019022397696971893f;
  t160_n81: return 0.03073755092918873f;
  t160_n82: return 0.006640593055635691f;
  t160_n83: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t160_n87;
  t160_n84: if (E.getNumReferences() >= 3327545344 /*13716.0*/) goto t160_n86;
  t160_n85: return 0.11895716935396194f;
  t160_n86: return -0.05614138022065163f;
  t160_n87: return 0.11604379117488861f;
  t160_n88: if (E.getNumReferences() >= 3334173440 /*23993.5*/) goto t160_n92;
  t160_n89: if (E.getFractionNameInContext() >= 3187240856 /*0.12179487943649292*/) goto t160_n91;
  t160_n90: return -0.05820425972342491f;
  t160_n91: return -0.6336929798126221f;
  t160_n92: return 0.08457411825656891f;
  t160_n93: return -0.015490192919969559f;
  t160_n94: if (E.getNumReferences() >= 3329648640 /*15770.0*/) goto t160_n166;
  t160_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t160_n155;
  t160_n96: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t160_n150;
  t160_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t160_n111;
  t160_n98: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t160_n100;
  t160_n99: return 0.01795278675854206f;
  t160_n100: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t160_n102;
  t160_n101: return -0.06641858071088791f;
  t160_n102: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t160_n110;
  t160_n103: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t160_n105;
  t160_n104: return 0.07917038351297379f;
  t160_n105: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t160_n107;
  t160_n106: return -0.028822265565395355f;
  t160_n107: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t160_n109;
  t160_n108: return 0.062471531331539154f;
  t160_n109: return 0.11812596768140793f;
  t160_n110: return -0.004134381655603647f;
  t160_n111: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t160_n113;
  t160_n112: return 0.004498590249568224f;
  t160_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t160_n121;
  t160_n114: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t160_n116;
  t160_n115: return 0.0007600862300023437f;
  t160_n116: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t160_n118;
  t160_n117: return -0.025313127785921097f;
  t160_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t160_n120;
  t160_n119: return 0.14108020067214966f;
  t160_n120: return 0.22298552095890045f;
  t160_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t160_n147;
  t160_n122: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t160_n146;
  t160_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t160_n125;
  t160_n124: return -0.019769325852394104f;
  t160_n125: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t160_n145;
  t160_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n128;
  t160_n127: return -0.000875027384608984f;
  t160_n128: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t160_n138;
  t160_n129: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t160_n133;
  t160_n130: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t160_n132;
  t160_n131: return 0.022606998682022095f;
  t160_n132: return 0.11122715473175049f;
  t160_n133: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t160_n135;
  t160_n134: return -0.03458452597260475f;
  t160_n135: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t160_n137;
  t160_n136: return 0.11436613649129868f;
  t160_n137: return 0.01604459062218666f;
  t160_n138: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t160_n140;
  t160_n139: return -0.016599353402853012f;
  t160_n140: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t160_n142;
  t160_n141: return 0.01395431999117136f;
  t160_n142: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t160_n144;
  t160_n143: return -0.14551423490047455f;
  t160_n144: return 0.07245714962482452f;
  t160_n145: return 0.021214904263615608f;
  t160_n146: return -0.0026659034192562103f;
  t160_n147: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t160_n149;
  t160_n148: return 0.009490189142525196f;
  t160_n149: return 0.03499143198132515f;
  t160_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t160_n152;
  t160_n151: return -0.04433957859873772f;
  t160_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n154;
  t160_n153: return -0.22097456455230713f;
  t160_n154: return 0.10846593230962753f;
  t160_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n163;
  t160_n156: if (E.getNumReferences() >= 3329101824 /*15236.0*/) goto t160_n160;
  t160_n157: if (E.getNumReferences() >= 3328489984 /*14638.5*/) goto t160_n159;
  t160_n158: return -0.01208017859607935f;
  t160_n159: return 0.09856541454792023f;
  t160_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n162;
  t160_n161: return -0.14094847440719604f;
  t160_n162: return 0.071615070104599f;
  t160_n163: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t160_n165;
  t160_n164: return -0.04516790062189102f;
  t160_n165: return 0.016056234017014503f;
  t160_n166: if (E.getNumReferences() >= 3330190848 /*16299.5*/) goto t160_n186;
  t160_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t160_n173;
  t160_n168: if (E.getNumReferences() >= 3330140160 /*16250.0*/) goto t160_n170;
  t160_n169: return 0.03284329175949097f;
  t160_n170: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t160_n172;
  t160_n171: return 0.02940535731613636f;
  t160_n172: return 0.13321052491664886f;
  t160_n173: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t160_n181;
  t160_n174: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t160_n176;
  t160_n175: return 0.05975617840886116f;
  t160_n176: if (E.getFractionNameInContext() >= 3183239395 /*0.0919816717505455*/) goto t160_n180;
  t160_n177: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t160_n179;
  t160_n178: return 0.12051026523113251f;
  t160_n179: return 0.1358967274427414f;
  t160_n180: return 0.12834876775741577f;
  t160_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t160_n183;
  t160_n182: return 0.024222994223237038f;
  t160_n183: if (E.getNumReferences() >= 3330029056 /*16141.5*/) goto t160_n185;
  t160_n184: return 0.1676546037197113f;
  t160_n185: return 0.005021404474973679f;
  t160_n186: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t160_n188;
  t160_n187: return 0.018633026629686356f;
  t160_n188: return 0.2834759056568146f;
  t160_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n191;
  t160_n190: return -0.1438155472278595f;
  t160_n191: return -0.04825301840901375f;
  t160_n192: if (E.getNumReferences() >= 3341720320 /*44699.0*/) goto t160_n204;
  t160_n193: if (E.getNumReferences() >= 3341141504 /*42438.0*/) goto t160_n195;
  t160_n194: return 0.016629885882139206f;
  t160_n195: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t160_n199;
  t160_n196: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n198;
  t160_n197: return -0.19776439666748047f;
  t160_n198: return 0.05500492453575134f;
  t160_n199: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t160_n203;
  t160_n200: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t160_n202;
  t160_n201: return 0.12437921017408371f;
  t160_n202: return -0.11735007911920547f;
  t160_n203: return -0.0006312868208624423f;
  t160_n204: return 0.0039511569775640965f;
  t160_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t160_n207;
  t160_n206: return 0.008504044264554977f;
  t160_n207: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t160_n209;
  t160_n208: return -0.04409629851579666f;
  t160_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n211;
  t160_n210: return -0.15463842451572418f;
  t160_n211: return 0.03520826995372772f;
  t160_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t160_n214;
  t160_n213: return -0.10584384948015213f;
  t160_n214: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t160_n216;
  t160_n215: return -0.04019346833229065f;
  t160_n216: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t160_n250;
  t160_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t160_n219;
  t160_n218: return -0.002566104056313634f;
  t160_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t160_n223;
  t160_n220: if (E.getNumReferences() >= 3347831872 /*71608.5*/) goto t160_n222;
  t160_n221: return -0.022291814908385277f;
  t160_n222: return 0.19589857757091522f;
  t160_n223: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t160_n249;
  t160_n224: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t160_n226;
  t160_n225: return 0.014446130953729153f;
  t160_n226: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t160_n248;
  t160_n227: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t160_n239;
  t160_n228: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t160_n234;
  t160_n229: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t160_n233;
  t160_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t160_n232;
  t160_n231: return -0.18201252818107605f;
  t160_n232: return 0.16840165853500366f;
  t160_n233: return 0.08814404159784317f;
  t160_n234: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t160_n236;
  t160_n235: return -0.08300977945327759f;
  t160_n236: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t160_n238;
  t160_n237: return 0.06850868463516235f;
  t160_n238: return -0.0019344831816852093f;
  t160_n239: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t160_n241;
  t160_n240: return -0.046636223793029785f;
  t160_n241: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t160_n245;
  t160_n242: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t160_n244;
  t160_n243: return 0.07687977701425552f;
  t160_n244: return -0.07420490682125092f;
  t160_n245: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t160_n247;
  t160_n246: return 0.08235105872154236f;
  t160_n247: return 0.11726447194814682f;
  t160_n248: return -0.0034615076147019863f;
  t160_n249: return -0.043348006904125214f;
  t160_n250: return 0.08740735054016113f;
  t160_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t160_n253;
  t160_n252: return 0.001756542013026774f;
  t160_n253: return 0.005444488488137722f;
  t160_n254: return -0.019924568012356758f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree161(const Example& E) {
  t161_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t161_n196;
  t161_n1: if (E.getNumReferences() >= 3285270528 /*418.5*/) goto t161_n75;
  t161_n2: if (E.getFileProximityDistanceCost() >= 3257139200 /*41.0*/) goto t161_n40;
  t161_n3: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t161_n13;
  t161_n4: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t161_n8;
  t161_n5: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t161_n7;
  t161_n6: return -0.0009731304016895592f;
  t161_n7: return -0.053195130079984665f;
  t161_n8: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t161_n10;
  t161_n9: return 0.10444305837154388f;
  t161_n10: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t161_n12;
  t161_n11: return -0.08114007860422134f;
  t161_n12: return -0.14451591670513153f;
  t161_n13: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t161_n17;
  t161_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t161_n16;
  t161_n15: return -0.03211551159620285f;
  t161_n16: return -0.01045304536819458f;
  t161_n17: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t161_n39;
  t161_n18: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t161_n38;
  t161_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n21;
  t161_n20: return -0.045851461589336395f;
  t161_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t161_n35;
  t161_n22: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t161_n34;
  t161_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t161_n33;
  t161_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n26;
  t161_n25: return -0.11456941068172455f;
  t161_n26: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t161_n32;
  t161_n27: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t161_n29;
  t161_n28: return -0.3063596785068512f;
  t161_n29: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t161_n31;
  t161_n30: return 0.12267619371414185f;
  t161_n31: return -0.11889081448316574f;
  t161_n32: return -0.014002813957631588f;
  t161_n33: return -0.10419133305549622f;
  t161_n34: return 0.10223463177680969f;
  t161_n35: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t161_n37;
  t161_n36: return 0.018988804891705513f;
  t161_n37: return -0.10342592746019363f;
  t161_n38: return -0.046952590346336365f;
  t161_n39: return -0.015950394794344902f;
  t161_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t161_n58;
  t161_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n55;
  t161_n42: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t161_n46;
  t161_n43: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t161_n45;
  t161_n44: return -0.011486164294183254f;
  t161_n45: return -0.15315760672092438f;
  t161_n46: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t161_n54;
  t161_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t161_n49;
  t161_n48: return 0.009303353726863861f;
  t161_n49: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t161_n51;
  t161_n50: return 0.03604292869567871f;
  t161_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t161_n53;
  t161_n52: return 0.037828296422958374f;
  t161_n53: return 0.11464841663837433f;
  t161_n54: return -0.01416569110006094f;
  t161_n55: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t161_n57;
  t161_n56: return 0.016454769298434258f;
  t161_n57: return 0.1021660640835762f;
  t161_n58: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t161_n60;
  t161_n59: return 0.002005181508138776f;
  t161_n60: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t161_n70;
  t161_n61: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t161_n69;
  t161_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n64;
  t161_n63: return -0.07047723978757858f;
  t161_n64: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t161_n68;
  t161_n65: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t161_n67;
  t161_n66: return -0.04567405954003334f;
  t161_n67: return 0.08235999196767807f;
  t161_n68: return 0.12408193945884705f;
  t161_n69: return 0.049842942506074905f;
  t161_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t161_n74;
  t161_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n73;
  t161_n72: return -0.019406959414482117f;
  t161_n73: return 0.008962557651102543f;
  t161_n74: return -0.013954345136880875f;
  t161_n75: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t161_n77;
  t161_n76: return 0.00877426564693451f;
  t161_n77: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t161_n195;
  t161_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n154;
  t161_n79: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t161_n151;
  t161_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t161_n118;
  t161_n81: if (E.getNumReferences() >= 3285303296 /*419.5*/) goto t161_n83;
  t161_n82: return 0.11931760609149933f;
  t161_n83: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t161_n117;
  t161_n84: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t161_n112;
  t161_n85: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t161_n87;
  t161_n86: return -0.03671186417341232f;
  t161_n87: if (E.getNumReferences() >= 3350496896 /*92429.0*/) goto t161_n109;
  t161_n88: if (E.getNumReferences() >= 3348207616 /*74544.0*/) goto t161_n104;
  t161_n89: if (E.getNumReferences() >= 3346697088 /*64139.5*/) goto t161_n103;
  t161_n90: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t161_n94;
  t161_n91: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t161_n93;
  t161_n92: return -0.004735884722322226f;
  t161_n93: return -0.151162788271904f;
  t161_n94: if (E.getNumReferences() >= 3345444992 /*59248.5*/) goto t161_n100;
  t161_n95: if (E.getNumReferences() >= 3345041664 /*57673.0*/) goto t161_n99;
  t161_n96: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t161_n98;
  t161_n97: return 0.07569319009780884f;
  t161_n98: return -0.21607472002506256f;
  t161_n99: return 0.10328861325979233f;
  t161_n100: if (E.getNumReferences() >= 3346171648 /*62087.0*/) goto t161_n102;
  t161_n101: return -0.3315441906452179f;
  t161_n102: return 0.03146623075008392f;
  t161_n103: return -0.1585596352815628f;
  t161_n104: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t161_n106;
  t161_n105: return 0.08110585063695908f;
  t161_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t161_n108;
  t161_n107: return -0.008979281410574913f;
  t161_n108: return 0.219569131731987f;
  t161_n109: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t161_n111;
  t161_n110: return 0.08145228773355484f;
  t161_n111: return -0.21042916178703308f;
  t161_n112: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t161_n116;
  t161_n113: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t161_n115;
  t161_n114: return 0.10396444797515869f;
  t161_n115: return -0.01234421692788601f;
  t161_n116: return 0.11708075553178787f;
  t161_n117: return -0.05580344796180725f;
  t161_n118: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t161_n126;
  t161_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t161_n121;
  t161_n120: return 0.03582831099629402f;
  t161_n121: if (E.getNumReferences() >= 3294461952 /*886.0*/) goto t161_n125;
  t161_n122: if (E.getNumReferences() >= 3294257152 /*873.5*/) goto t161_n124;
  t161_n123: return 0.0926065668463707f;
  t161_n124: return 0.15045525133609772f;
  t161_n125: return 0.054309215396642685f;
  t161_n126: if (E.getNumReferences() >= 3314463744 /*4566.5*/) goto t161_n130;
  t161_n127: if (E.getNumReferences() >= 3314287616 /*4480.5*/) goto t161_n129;
  t161_n128: return -0.00046464873594231904f;
  t161_n129: return -0.3491741418838501f;
  t161_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t161_n132;
  t161_n131: return 0.042792435735464096f;
  t161_n132: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t161_n146;
  t161_n133: if (E.getNumReferences() >= 3317335040 /*5968.5*/) goto t161_n137;
  t161_n134: if (E.getNumReferences() >= 3316881408 /*5747.0*/) goto t161_n136;
  t161_n135: return 0.0336347296833992f;
  t161_n136: return 0.13033324480056763f;
  t161_n137: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t161_n139;
  t161_n138: return 0.0977618619799614f;
  t161_n139: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t161_n145;
  t161_n140: if (E.getNumReferences() >= 3347359424 /*67917.5*/) goto t161_n144;
  t161_n141: if (E.getNumReferences() >= 3339934592 /*37723.5*/) goto t161_n143;
  t161_n142: return -0.0038731112144887447f;
  t161_n143: return -0.4245408773422241f;
  t161_n144: return 0.08268754929304123f;
  t161_n145: return 0.049302082508802414f;
  t161_n146: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t161_n150;
  t161_n147: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t161_n149;
  t161_n148: return 0.12050960212945938f;
  t161_n149: return 0.106528140604496f;
  t161_n150: return -0.218390554189682f;
  t161_n151: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t161_n153;
  t161_n152: return -0.05470995977520943f;
  t161_n153: return -0.5589450001716614f;
  t161_n154: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t161_n194;
  t161_n155: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t161_n193;
  t161_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel))) goto t161_n170;
  t161_n157: if (E.getNumReferences() >= 3291652096 /*714.5*/) goto t161_n163;
  t161_n158: if (E.getNumReferences() >= 3291619328 /*712.5*/) goto t161_n162;
  t161_n159: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t161_n161;
  t161_n160: return 0.028459379449486732f;
  t161_n161: return -0.14523033797740936f;
  t161_n162: return 0.22311906516551971f;
  t161_n163: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t161_n165;
  t161_n164: return -0.03540891781449318f;
  t161_n165: if (E.getNumReferences() >= 3306031104 /*2272.5*/) goto t161_n169;
  t161_n166: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t161_n168;
  t161_n167: return -0.4664923846721649f;
  t161_n168: return -0.25549817085266113f;
  t161_n169: return -0.07232446223497391f;
  t161_n170: if (E.getNumReferences() >= 3316056064 /*5344.0*/) goto t161_n192;
  t161_n171: if (E.getNumReferences() >= 3315686400 /*5163.5*/) goto t161_n191;
  t161_n172: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t161_n190;
  t161_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t161_n175;
  t161_n174: return -0.01601664163172245f;
  t161_n175: if (E.getNumReferences() >= 3287023616 /*472.0*/) goto t161_n179;
  t161_n176: if (E.getNumReferences() >= 3286253568 /*448.5*/) goto t161_n178;
  t161_n177: return 0.12120308727025986f;
  t161_n178: return 0.1855584681034088f;
  t161_n179: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t161_n181;
  t161_n180: return -0.008911682292819023f;
  t161_n181: if (E.getNumReferences() >= 3308804096 /*2949.5*/) goto t161_n187;
  t161_n182: if (E.getNumReferences() >= 3296534528 /*1012.5*/) goto t161_n186;
  t161_n183: if (E.getNumReferences() >= 3296067584 /*984.0*/) goto t161_n185;
  t161_n184: return 0.12706151604652405f;
  t161_n185: return 0.26403024792671204f;
  t161_n186: return 0.04989210143685341f;
  t161_n187: if (E.getNumReferences() >= 3312951296 /*3962.0*/) goto t161_n189;
  t161_n188: return 0.4923761785030365f;
  t161_n189: return -0.10560549050569534f;
  t161_n190: return 0.015158567577600479f;
  t161_n191: return 0.13784611225128174f;
  t161_n192: return -0.03213989734649658f;
  t161_n193: return 0.003692059777677059f;
  t161_n194: return -0.018723422661423683f;
  t161_n195: return 0.13531869649887085f;
  t161_n196: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t161_n198;
  t161_n197: return 0.0009606350213289261f;
  t161_n198: if (E.getFractionNameInContext() >= 3185148150 /*0.10620300471782684*/) goto t161_n202;
  t161_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t161_n201;
  t161_n200: return 0.011089012026786804f;
  t161_n201: return 0.015327390283346176f;
  t161_n202: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t161_n220;
  t161_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n209;
  t161_n204: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t161_n206;
  t161_n205: return -0.007991892285645008f;
  t161_n206: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t161_n208;
  t161_n207: return -0.11166754364967346f;
  t161_n208: return 0.061816196888685226f;
  t161_n209: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t161_n215;
  t161_n210: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t161_n212;
  t161_n211: return 0.012242303229868412f;
  t161_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t161_n214;
  t161_n213: return -0.030140511691570282f;
  t161_n214: return -0.003631716826930642f;
  t161_n215: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t161_n219;
  t161_n216: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t161_n218;
  t161_n217: return -0.25548338890075684f;
  t161_n218: return 0.008698823861777782f;
  t161_n219: return -0.014425798319280148f;
  t161_n220: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t161_n248;
  t161_n221: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t161_n247;
  t161_n222: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t161_n230;
  t161_n223: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t161_n229;
  t161_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n228;
  t161_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t161_n227;
  t161_n226: return -0.2696499228477478f;
  t161_n227: return -0.09351889044046402f;
  t161_n228: return -0.058538321405649185f;
  t161_n229: return 0.011328881606459618f;
  t161_n230: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t161_n246;
  t161_n231: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t161_n235;
  t161_n232: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t161_n234;
  t161_n233: return 0.024201365187764168f;
  t161_n234: return -0.973320722579956f;
  t161_n235: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t161_n245;
  t161_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t161_n238;
  t161_n237: return -1.1254489421844482f;
  t161_n238: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t161_n240;
  t161_n239: return -0.11098535358905792f;
  t161_n240: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t161_n244;
  t161_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t161_n243;
  t161_n242: return -0.009295891970396042f;
  t161_n243: return 0.10253646969795227f;
  t161_n244: return 0.20532527565956116f;
  t161_n245: return 0.025241145864129066f;
  t161_n246: return 0.0520591102540493f;
  t161_n247: return 0.0038483545649796724f;
  t161_n248: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t161_n254;
  t161_n249: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t161_n251;
  t161_n250: return 0.036103829741477966f;
  t161_n251: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t161_n253;
  t161_n252: return 0.16392378509044647f;
  t161_n253: return 0.07092894613742828f;
  t161_n254: return -0.009950865991413593f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree162(const Example& E) {
  t162_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t162_n4;
  t162_n1: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t162_n3;
  t162_n2: return 0.00015007056936156005f;
  t162_n3: return -0.00299236667342484f;
  t162_n4: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t162_n226;
  t162_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n225;
  t162_n6: if (E.getNumReferences() >= 3292954624 /*794.0*/) goto t162_n222;
  t162_n7: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t162_n47;
  t162_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t162_n28;
  t162_n9: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n11;
  t162_n10: return -0.05862662196159363f;
  t162_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t162_n17;
  t162_n12: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t162_n16;
  t162_n13: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t162_n15;
  t162_n14: return 0.1106216087937355f;
  t162_n15: return 0.07421110570430756f;
  t162_n16: return 0.01073241326957941f;
  t162_n17: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t162_n23;
  t162_n18: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t162_n20;
  t162_n19: return -0.03951922804117203f;
  t162_n20: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t162_n22;
  t162_n21: return -0.25722968578338623f;
  t162_n22: return 0.09141994267702103f;
  t162_n23: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t162_n27;
  t162_n24: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t162_n26;
  t162_n25: return 0.11188214272260666f;
  t162_n26: return -0.1249396875500679f;
  t162_n27: return -0.2884851396083832f;
  t162_n28: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t162_n38;
  t162_n29: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t162_n31;
  t162_n30: return -0.0012465600157156587f;
  t162_n31: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t162_n37;
  t162_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t162_n36;
  t162_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n35;
  t162_n34: return -0.09379430115222931f;
  t162_n35: return 0.09494142234325409f;
  t162_n36: return 0.12432879954576492f;
  t162_n37: return -0.057646751403808594f;
  t162_n38: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t162_n46;
  t162_n39: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t162_n41;
  t162_n40: return 0.024143394082784653f;
  t162_n41: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t162_n45;
  t162_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol))) goto t162_n44;
  t162_n43: return 0.04122631624341011f;
  t162_n44: return 0.11743814498186111f;
  t162_n45: return -0.1724073737859726f;
  t162_n46: return 0.009818640537559986f;
  t162_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t162_n173;
  t162_n48: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t162_n142;
  t162_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t162_n75;
  t162_n50: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t162_n66;
  t162_n51: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t162_n61;
  t162_n52: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t162_n54;
  t162_n53: return -0.8278540968894958f;
  t162_n54: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t162_n60;
  t162_n55: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t162_n57;
  t162_n56: return -0.13153468072414398f;
  t162_n57: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t162_n59;
  t162_n58: return 0.1495865136384964f;
  t162_n59: return -0.1509907990694046f;
  t162_n60: return -0.07325667142868042f;
  t162_n61: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t162_n63;
  t162_n62: return 0.1259842962026596f;
  t162_n63: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t162_n65;
  t162_n64: return 0.163358673453331f;
  t162_n65: return -0.4548207223415375f;
  t162_n66: if (E.getNumReferences() >= 3261595648 /*58.0*/) goto t162_n70;
  t162_n67: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t162_n69;
  t162_n68: return -0.20188798010349274f;
  t162_n69: return 0.006001830101013184f;
  t162_n70: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t162_n74;
  t162_n71: if (E.getNumReferences() >= 3279421440 /*248.0*/) goto t162_n73;
  t162_n72: return 0.17514605820178986f;
  t162_n73: return -0.2971140146255493f;
  t162_n74: return 0.09169206768274307f;
  t162_n75: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t162_n115;
  t162_n76: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t162_n114;
  t162_n77: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t162_n79;
  t162_n78: return -0.023781845346093178f;
  t162_n79: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t162_n113;
  t162_n80: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t162_n86;
  t162_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n85;
  t162_n82: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t162_n84;
  t162_n83: return -0.07398708164691925f;
  t162_n84: return -0.13737940788269043f;
  t162_n85: return -0.01669752411544323f;
  t162_n86: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t162_n100;
  t162_n87: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t162_n97;
  t162_n88: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t162_n92;
  t162_n89: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t162_n91;
  t162_n90: return -0.045584939420223236f;
  t162_n91: return 0.04975680261850357f;
  t162_n92: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t162_n94;
  t162_n93: return -0.024004295468330383f;
  t162_n94: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t162_n96;
  t162_n95: return -0.4319038987159729f;
  t162_n96: return -0.09511703252792358f;
  t162_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n99;
  t162_n98: return 0.04639521241188049f;
  t162_n99: return -0.179730623960495f;
  t162_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n110;
  t162_n101: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t162_n107;
  t162_n102: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t162_n104;
  t162_n103: return 0.07138483226299286f;
  t162_n104: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t162_n106;
  t162_n105: return 0.007608278188854456f;
  t162_n106: return -0.3660590350627899f;
  t162_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n109;
  t162_n108: return -0.22046029567718506f;
  t162_n109: return 0.03419656306505203f;
  t162_n110: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t162_n112;
  t162_n111: return -0.0249750018119812f;
  t162_n112: return 0.12158222496509552f;
  t162_n113: return 0.10117416828870773f;
  t162_n114: return -0.00845332257449627f;
  t162_n115: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t162_n117;
  t162_n116: return 0.07162181288003922f;
  t162_n117: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t162_n141;
  t162_n118: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t162_n126;
  t162_n119: if (E.getFractionNameInContext() >= 3207384269 /*0.675000011920929*/) goto t162_n125;
  t162_n120: if (E.getNumReferences() >= 3277979648 /*226.0*/) goto t162_n122;
  t162_n121: return 0.00980627816170454f;
  t162_n122: if (E.getNumReferences() >= 3280322560 /*267.5*/) goto t162_n124;
  t162_n123: return -0.8583166599273682f;
  t162_n124: return -0.06488372385501862f;
  t162_n125: return -2.001957654953003f;
  t162_n126: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t162_n132;
  t162_n127: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t162_n131;
  t162_n128: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t162_n130;
  t162_n129: return 0.10730528831481934f;
  t162_n130: return 0.015040501020848751f;
  t162_n131: return 0.053085628896951675f;
  t162_n132: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t162_n136;
  t162_n133: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t162_n135;
  t162_n134: return -0.22007018327713013f;
  t162_n135: return 0.05572273209691048f;
  t162_n136: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t162_n138;
  t162_n137: return 0.10059461742639542f;
  t162_n138: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t162_n140;
  t162_n139: return -0.11654406785964966f;
  t162_n140: return 0.10154390335083008f;
  t162_n141: return -0.07988971471786499f;
  t162_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t162_n168;
  t162_n143: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t162_n165;
  t162_n144: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t162_n164;
  t162_n145: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t162_n161;
  t162_n146: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t162_n152;
  t162_n147: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t162_n149;
  t162_n148: return -0.00831122137606144f;
  t162_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n151;
  t162_n150: return -0.10104109346866608f;
  t162_n151: return 0.09831125289201736f;
  t162_n152: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t162_n160;
  t162_n153: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t162_n157;
  t162_n154: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t162_n156;
  t162_n155: return 0.11793742328882217f;
  t162_n156: return 0.03095085918903351f;
  t162_n157: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t162_n159;
  t162_n158: return 0.02442062832415104f;
  t162_n159: return -0.28257009387016296f;
  t162_n160: return 0.11885558068752289f;
  t162_n161: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t162_n163;
  t162_n162: return 0.06058623641729355f;
  t162_n163: return -0.03087019920349121f;
  t162_n164: return 0.16199618577957153f;
  t162_n165: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t162_n167;
  t162_n166: return -0.034514252096414566f;
  t162_n167: return -0.37986403703689575f;
  t162_n168: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t162_n172;
  t162_n169: if (E.getNumReferences() >= 3267035136 /*93.5*/) goto t162_n171;
  t162_n170: return 0.049332983791828156f;
  t162_n171: return -0.6091328859329224f;
  t162_n172: return 0.12442383170127869f;
  t162_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t162_n221;
  t162_n174: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t162_n218;
  t162_n175: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t162_n217;
  t162_n176: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t162_n204;
  t162_n177: if (E.getNumReferences() >= 3292717056 /*779.5*/) goto t162_n203;
  t162_n178: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t162_n188;
  t162_n179: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t162_n187;
  t162_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t162_n182;
  t162_n181: return -0.01710810698568821f;
  t162_n182: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t162_n184;
  t162_n183: return -0.05532904341816902f;
  t162_n184: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t162_n186;
  t162_n185: return 0.1335337907075882f;
  t162_n186: return -0.04941777139902115f;
  t162_n187: return 0.03676449507474899f;
  t162_n188: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t162_n190;
  t162_n189: return -0.03388389199972153f;
  t162_n190: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t162_n202;
  t162_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t162_n193;
  t162_n192: return -0.025339974090456963f;
  t162_n193: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t162_n201;
  t162_n194: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t162_n200;
  t162_n195: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n199;
  t162_n196: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t162_n198;
  t162_n197: return 0.07774719595909119f;
  t162_n198: return 0.11330850422382355f;
  t162_n199: return 0.07178269326686859f;
  t162_n200: return -0.25303900241851807f;
  t162_n201: return 0.019023992121219635f;
  t162_n202: return -0.012753689661622047f;
  t162_n203: return -0.5816779732704163f;
  t162_n204: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t162_n206;
  t162_n205: return 0.022354256361722946f;
  t162_n206: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t162_n216;
  t162_n207: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t162_n211;
  t162_n208: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t162_n210;
  t162_n209: return -0.42645570635795593f;
  t162_n210: return 0.03946143016219139f;
  t162_n211: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t162_n213;
  t162_n212: return -0.043478239327669144f;
  t162_n213: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n215;
  t162_n214: return -0.5904929041862488f;
  t162_n215: return 0.044461000710725784f;
  t162_n216: return 0.05600515007972717f;
  t162_n217: return -0.041289493441581726f;
  t162_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t162_n220;
  t162_n219: return 0.008239028044044971f;
  t162_n220: return 0.12440373748540878f;
  t162_n221: return 0.02035490982234478f;
  t162_n222: if (E.getNumReferences() >= 3293044736 /*799.5*/) goto t162_n224;
  t162_n223: return 0.2597789764404297f;
  t162_n224: return 0.014063028618693352f;
  t162_n225: return 0.005096385721117258f;
  t162_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t162_n230;
  t162_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n229;
  t162_n228: return -0.0002471075567882508f;
  t162_n229: return 0.014115513302385807f;
  t162_n230: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t162_n236;
  t162_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t162_n233;
  t162_n232: return -0.04560206085443497f;
  t162_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t162_n235;
  t162_n234: return 0.0012664112728089094f;
  t162_n235: return 0.009078862145543098f;
  t162_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t162_n244;
  t162_n237: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t162_n243;
  t162_n238: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t162_n240;
  t162_n239: return -0.01540613453835249f;
  t162_n240: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t162_n242;
  t162_n241: return -0.016818543896079063f;
  t162_n242: return 0.01335116010159254f;
  t162_n243: return 0.006831742823123932f;
  t162_n244: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t162_n254;
  t162_n245: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t162_n247;
  t162_n246: return 0.005083155352622271f;
  t162_n247: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t162_n249;
  t162_n248: return 0.011475278995931149f;
  t162_n249: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t162_n251;
  t162_n250: return 0.02352317050099373f;
  t162_n251: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t162_n253;
  t162_n252: return -0.2785453200340271f;
  t162_n253: return -0.37486693263053894f;
  t162_n254: return -0.08709463477134705f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree163(const Example& E) {
  t163_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t163_n214;
  t163_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t163_n181;
  t163_n2: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t163_n88;
  t163_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t163_n5;
  t163_n4: return -0.0011515496298670769f;
  t163_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t163_n29;
  t163_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t163_n22;
  t163_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t163_n21;
  t163_n8: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t163_n10;
  t163_n9: return -0.017032604664564133f;
  t163_n10: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t163_n20;
  t163_n11: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t163_n19;
  t163_n12: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t163_n18;
  t163_n13: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t163_n15;
  t163_n14: return -0.0197950080037117f;
  t163_n15: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t163_n17;
  t163_n16: return 0.1756085455417633f;
  t163_n17: return 0.18690700829029083f;
  t163_n18: return -0.044440340250730515f;
  t163_n19: return 0.05046186223626137f;
  t163_n20: return -0.04785984754562378f;
  t163_n21: return 0.0017866941634565592f;
  t163_n22: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t163_n26;
  t163_n23: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t163_n25;
  t163_n24: return 0.031170634552836418f;
  t163_n25: return -0.021755913272500038f;
  t163_n26: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t163_n28;
  t163_n27: return 0.004908669274300337f;
  t163_n28: return 0.09491557627916336f;
  t163_n29: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t163_n45;
  t163_n30: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t163_n44;
  t163_n31: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t163_n43;
  t163_n32: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t163_n34;
  t163_n33: return -0.012230955064296722f;
  t163_n34: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t163_n38;
  t163_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t163_n37;
  t163_n36: return -0.0014797864714637399f;
  t163_n37: return 0.01955225132405758f;
  t163_n38: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t163_n40;
  t163_n39: return -0.0035080364905297756f;
  t163_n40: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t163_n42;
  t163_n41: return -0.22634616494178772f;
  t163_n42: return -0.060994382947683334f;
  t163_n43: return -0.030013786628842354f;
  t163_n44: return 0.002791696460917592f;
  t163_n45: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t163_n47;
  t163_n46: return 0.013042382895946503f;
  t163_n47: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t163_n87;
  t163_n48: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t163_n84;
  t163_n49: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t163_n53;
  t163_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n52;
  t163_n51: return -0.0234467014670372f;
  t163_n52: return 0.0026747470255941153f;
  t163_n53: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t163_n61;
  t163_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n58;
  t163_n55: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t163_n57;
  t163_n56: return -0.04551436007022858f;
  t163_n57: return 0.03716893494129181f;
  t163_n58: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t163_n60;
  t163_n59: return 0.026159752160310745f;
  t163_n60: return 0.0014098044484853745f;
  t163_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t163_n63;
  t163_n62: return 0.0031277555972337723f;
  t163_n63: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t163_n83;
  t163_n64: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t163_n76;
  t163_n65: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t163_n71;
  t163_n66: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t163_n70;
  t163_n67: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t163_n69;
  t163_n68: return 0.014634215272963047f;
  t163_n69: return -0.02687644027173519f;
  t163_n70: return 0.024409782141447067f;
  t163_n71: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t163_n75;
  t163_n72: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t163_n74;
  t163_n73: return -0.06937258690595627f;
  t163_n74: return 0.059086065739393234f;
  t163_n75: return -0.29210636019706726f;
  t163_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t163_n78;
  t163_n77: return -0.060533154755830765f;
  t163_n78: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t163_n80;
  t163_n79: return 0.03477173298597336f;
  t163_n80: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t163_n82;
  t163_n81: return 0.09900365769863129f;
  t163_n82: return -0.007931957952678204f;
  t163_n83: return 0.005391839426010847f;
  t163_n84: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t163_n86;
  t163_n85: return 0.07795707881450653f;
  t163_n86: return 0.13133011758327484f;
  t163_n87: return -0.009709379635751247f;
  t163_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n180;
  t163_n89: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t163_n177;
  t163_n90: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t163_n136;
  t163_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t163_n105;
  t163_n92: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t163_n98;
  t163_n93: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t163_n97;
  t163_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n96;
  t163_n95: return -0.05094258859753609f;
  t163_n96: return 0.001688548014499247f;
  t163_n97: return 0.017673524096608162f;
  t163_n98: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t163_n100;
  t163_n99: return -0.028838247060775757f;
  t163_n100: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t163_n104;
  t163_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n103;
  t163_n102: return -0.031081613153219223f;
  t163_n103: return -0.005328363738954067f;
  t163_n104: return 0.016612987965345383f;
  t163_n105: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t163_n107;
  t163_n106: return -0.028737422078847885f;
  t163_n107: if (E.getNumReferences() >= 3308879872 /*2968.0*/) goto t163_n135;
  t163_n108: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t163_n110;
  t163_n109: return 0.04789895564317703f;
  t163_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t163_n112;
  t163_n111: return 0.03498581796884537f;
  t163_n112: if (E.getNumReferences() >= 3260022784 /*52.0*/) goto t163_n114;
  t163_n113: return 0.005799164529889822f;
  t163_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t163_n116;
  t163_n115: return 0.037904396653175354f;
  t163_n116: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t163_n118;
  t163_n117: return 0.4811939001083374f;
  t163_n118: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t163_n134;
  t163_n119: if (E.getNumReferences() >= 3277619200 /*220.5*/) goto t163_n131;
  t163_n120: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t163_n126;
  t163_n121: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t163_n125;
  t163_n122: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t163_n124;
  t163_n123: return 0.1480054408311844f;
  t163_n124: return -0.015408394858241081f;
  t163_n125: return -0.04410463571548462f;
  t163_n126: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t163_n130;
  t163_n127: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t163_n129;
  t163_n128: return 0.634691596031189f;
  t163_n129: return 0.09124476462602615f;
  t163_n130: return 0.07929942011833191f;
  t163_n131: if (E.getNumReferences() >= 3277914112 /*225.0*/) goto t163_n133;
  t163_n132: return 0.23047441244125366f;
  t163_n133: return 0.14297796785831451f;
  t163_n134: return -0.13112877309322357f;
  t163_n135: return -0.14310809969902039f;
  t163_n136: if (E.getSymbolScopeDistanceCost() >= 3247964160 /*19.0*/) goto t163_n166;
  t163_n137: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t163_n157;
  t163_n138: if (E.getNumReferences() >= 3264675840 /*75.5*/) goto t163_n156;
  t163_n139: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t163_n155;
  t163_n140: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t163_n142;
  t163_n141: return 0.04497115686535835f;
  t163_n142: if (E.getNumReferences() >= 3262119936 /*60.0*/) goto t163_n154;
  t163_n143: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t163_n153;
  t163_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t163_n146;
  t163_n145: return -0.13686220347881317f;
  t163_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t163_n148;
  t163_n147: return -0.1868647187948227f;
  t163_n148: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t163_n150;
  t163_n149: return 0.08443348109722137f;
  t163_n150: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t163_n152;
  t163_n151: return 0.16258899867534637f;
  t163_n152: return 0.027936367318034172f;
  t163_n153: return -0.16401000320911407f;
  t163_n154: return 0.11337094753980637f;
  t163_n155: return 0.04128766059875488f;
  t163_n156: return 0.005237955134361982f;
  t163_n157: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t163_n163;
  t163_n158: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t163_n160;
  t163_n159: return -0.10127148777246475f;
  t163_n160: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t163_n162;
  t163_n161: return 0.2850206196308136f;
  t163_n162: return 0.06915047019720078f;
  t163_n163: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t163_n165;
  t163_n164: return 0.04949408769607544f;
  t163_n165: return 0.0832405686378479f;
  t163_n166: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t163_n168;
  t163_n167: return 0.0012819173280149698f;
  t163_n168: if (E.getNumReferences() >= 3292766208 /*782.5*/) goto t163_n176;
  t163_n169: if (E.getNumReferences() >= 3287875584 /*498.0*/) goto t163_n173;
  t163_n170: if (E.getNumReferences() >= 3284156416 /*384.5*/) goto t163_n172;
  t163_n171: return 0.07876419275999069f;
  t163_n172: return -0.20037351548671722f;
  t163_n173: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t163_n175;
  t163_n174: return 0.11636481434106827f;
  t163_n175: return 0.05488942191004753f;
  t163_n176: return 0.031760986894369125f;
  t163_n177: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t163_n179;
  t163_n178: return 0.03197875991463661f;
  t163_n179: return -0.006821207236498594f;
  t163_n180: return 0.007134600542485714f;
  t163_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t163_n213;
  t163_n182: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t163_n184;
  t163_n183: return 0.002737049711868167f;
  t163_n184: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t163_n186;
  t163_n185: return 0.13142816722393036f;
  t163_n186: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t163_n212;
  t163_n187: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t163_n201;
  t163_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t163_n194;
  t163_n189: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t163_n191;
  t163_n190: return -0.1838751882314682f;
  t163_n191: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t163_n193;
  t163_n192: return 0.0365845188498497f;
  t163_n193: return 0.1331464946269989f;
  t163_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t163_n196;
  t163_n195: return 0.035387661308050156f;
  t163_n196: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t163_n200;
  t163_n197: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t163_n199;
  t163_n198: return 0.1260194629430771f;
  t163_n199: return 0.1102413460612297f;
  t163_n200: return 0.07534182816743851f;
  t163_n201: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t163_n209;
  t163_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t163_n208;
  t163_n203: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t163_n205;
  t163_n204: return 0.10010156780481339f;
  t163_n205: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t163_n207;
  t163_n206: return -0.3070422112941742f;
  t163_n207: return 0.07902512699365616f;
  t163_n208: return 0.08655393868684769f;
  t163_n209: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t163_n211;
  t163_n210: return -0.6132172346115112f;
  t163_n211: return -0.027464887127280235f;
  t163_n212: return -0.0025943215005099773f;
  t163_n213: return 0.029956592246890068f;
  t163_n214: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t163_n252;
  t163_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t163_n221;
  t163_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n220;
  t163_n217: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t163_n219;
  t163_n218: return -0.019968319684267044f;
  t163_n219: return -0.2195812463760376f;
  t163_n220: return -4.0654085751157254e-05f;
  t163_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n249;
  t163_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n232;
  t163_n223: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t163_n225;
  t163_n224: return -0.028918074443936348f;
  t163_n225: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t163_n227;
  t163_n226: return -0.19645927846431732f;
  t163_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t163_n229;
  t163_n228: return 0.027373531833291054f;
  t163_n229: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t163_n231;
  t163_n230: return 0.060735929757356644f;
  t163_n231: return 0.18159542977809906f;
  t163_n232: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t163_n248;
  t163_n233: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t163_n247;
  t163_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t163_n246;
  t163_n235: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t163_n245;
  t163_n236: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t163_n244;
  t163_n237: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t163_n241;
  t163_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n240;
  t163_n239: return -0.02779024839401245f;
  t163_n240: return 0.008565192110836506f;
  t163_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t163_n243;
  t163_n242: return -0.0912223681807518f;
  t163_n243: return -0.021894121542572975f;
  t163_n244: return 0.008059795945882797f;
  t163_n245: return -0.03209381178021431f;
  t163_n246: return 0.003953801468014717f;
  t163_n247: return 0.08631224930286407f;
  t163_n248: return 0.010236471891403198f;
  t163_n249: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t163_n251;
  t163_n250: return 0.02221059985458851f;
  t163_n251: return 0.09579655528068542f;
  t163_n252: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t163_n254;
  t163_n253: return -0.09833000600337982f;
  t163_n254: return -0.01076436135917902f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree164(const Example& E) {
  t164_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t164_n224;
  t164_n1: if (E.getNumReferences() >= 3325613056 /*11829.0*/) goto t164_n61;
  t164_n2: if (E.getNumReferences() >= 3324400128 /*10644.5*/) goto t164_n44;
  t164_n3: if (E.getNumReferences() >= 3324340736 /*10586.5*/) goto t164_n43;
  t164_n4: if (E.getNumReferences() >= 3321325568 /*7917.0*/) goto t164_n6;
  t164_n5: return -0.0003284394624643028f;
  t164_n6: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t164_n18;
  t164_n7: if (E.getNumReferences() >= 3321606144 /*8054.0*/) goto t164_n17;
  t164_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t164_n10;
  t164_n9: return 0.04202892258763313f;
  t164_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t164_n12;
  t164_n11: return -0.1285262405872345f;
  t164_n12: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t164_n16;
  t164_n13: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t164_n15;
  t164_n14: return 0.08877543359994888f;
  t164_n15: return 0.21349941194057465f;
  t164_n16: return 0.05041343718767166f;
  t164_n17: return 0.010783053003251553f;
  t164_n18: if (E.getNumReferences() >= 3322322432 /*8615.5*/) goto t164_n32;
  t164_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t164_n25;
  t164_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t164_n22;
  t164_n21: return -0.12314485758543015f;
  t164_n22: if (E.getNumReferences() >= 3322056192 /*8355.5*/) goto t164_n24;
  t164_n23: return 0.06409318000078201f;
  t164_n24: return 0.1443203240633011f;
  t164_n25: if (E.getNumReferences() >= 3322022400 /*8322.5*/) goto t164_n31;
  t164_n26: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t164_n28;
  t164_n27: return 0.09831883758306503f;
  t164_n28: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t164_n30;
  t164_n29: return 0.14734730124473572f;
  t164_n30: return 0.11275098472833633f;
  t164_n31: return 0.1196984052658081f;
  t164_n32: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t164_n42;
  t164_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t164_n41;
  t164_n34: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t164_n38;
  t164_n35: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t164_n37;
  t164_n36: return 0.010050983168184757f;
  t164_n37: return -0.3721829652786255f;
  t164_n38: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t164_n40;
  t164_n39: return 0.17852069437503815f;
  t164_n40: return 0.04894942790269852f;
  t164_n41: return 0.0908871591091156f;
  t164_n42: return -0.06192508712410927f;
  t164_n43: return 0.17575672268867493f;
  t164_n44: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t164_n60;
  t164_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t164_n47;
  t164_n46: return -0.07793063670396805f;
  t164_n47: if (E.getNumReferences() >= 3325015040 /*11245.0*/) goto t164_n49;
  t164_n48: return -0.016946271061897278f;
  t164_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t164_n53;
  t164_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t164_n52;
  t164_n51: return -0.05428366735577583f;
  t164_n52: return 0.07345432043075562f;
  t164_n53: if (E.getNumReferences() >= 3325349888 /*11572.0*/) goto t164_n55;
  t164_n54: return 0.07376116514205933f;
  t164_n55: if (E.getNumReferences() >= 3325406720 /*11627.5*/) goto t164_n59;
  t164_n56: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t164_n58;
  t164_n57: return 0.18815183639526367f;
  t164_n58: return 0.09974271059036255f;
  t164_n59: return 0.015488316304981709f;
  t164_n60: return -0.11025380343198776f;
  t164_n61: if (E.getNumReferences() >= 3325976064 /*12183.5*/) goto t164_n89;
  t164_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t164_n64;
  t164_n63: return 0.0897490382194519f;
  t164_n64: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t164_n84;
  t164_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t164_n73;
  t164_n66: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t164_n72;
  t164_n67: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t164_n69;
  t164_n68: return 0.082343190908432f;
  t164_n69: if (E.getNumReferences() >= 3325743104 /*11956.0*/) goto t164_n71;
  t164_n70: return -0.04568734019994736f;
  t164_n71: return 0.31309694051742554f;
  t164_n72: return -0.13660572469234467f;
  t164_n73: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t164_n75;
  t164_n74: return -0.07486861199140549f;
  t164_n75: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t164_n77;
  t164_n76: return 0.14060282707214355f;
  t164_n77: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t164_n79;
  t164_n78: return 0.27950823307037354f;
  t164_n79: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t164_n81;
  t164_n80: return -0.10148020833730698f;
  t164_n81: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t164_n83;
  t164_n82: return 0.13164769113063812f;
  t164_n83: return 0.08311842381954193f;
  t164_n84: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t164_n88;
  t164_n85: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t164_n87;
  t164_n86: return -0.12937533855438232f;
  t164_n87: return 0.17518337070941925f;
  t164_n88: return 0.12405309081077576f;
  t164_n89: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t164_n93;
  t164_n90: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t164_n92;
  t164_n91: return 0.015540807507932186f;
  t164_n92: return 0.0026081139221787453f;
  t164_n93: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t164_n101;
  t164_n94: if (E.getNumReferences() >= 3329094144 /*15228.5*/) goto t164_n96;
  t164_n95: return -0.000953526294324547f;
  t164_n96: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t164_n100;
  t164_n97: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t164_n99;
  t164_n98: return -0.08852890878915787f;
  t164_n99: return -0.06731687486171722f;
  t164_n100: return -0.22510245442390442f;
  t164_n101: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t164_n107;
  t164_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t164_n104;
  t164_n103: return 0.06511852890253067f;
  t164_n104: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t164_n106;
  t164_n105: return -0.021963873878121376f;
  t164_n106: return 0.11087043583393097f;
  t164_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t164_n131;
  t164_n108: if (E.getNumReferences() >= 3343775872 /*52728.5*/) goto t164_n126;
  t164_n109: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t164_n121;
  t164_n110: if (E.getNumReferences() >= 3341720320 /*44699.0*/) goto t164_n120;
  t164_n111: if (E.getNumReferences() >= 3341367040 /*43319.0*/) goto t164_n113;
  t164_n112: return -0.011077930219471455f;
  t164_n113: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t164_n119;
  t164_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t164_n116;
  t164_n115: return 0.1332833170890808f;
  t164_n116: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t164_n118;
  t164_n117: return 0.12155592441558838f;
  t164_n118: return 0.051908425986766815f;
  t164_n119: return -0.5131214261054993f;
  t164_n120: return -0.14496566355228424f;
  t164_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t164_n123;
  t164_n122: return 0.09669787436723709f;
  t164_n123: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t164_n125;
  t164_n124: return 0.15079280734062195f;
  t164_n125: return 0.14457480609416962f;
  t164_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t164_n130;
  t164_n127: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t164_n129;
  t164_n128: return -0.095041923224926f;
  t164_n129: return 0.04611114040017128f;
  t164_n130: return -0.037403371185064316f;
  t164_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t164_n171;
  t164_n132: if (E.getNumReferences() >= 3333689088 /*23047.5*/) goto t164_n134;
  t164_n133: return -0.08724944293498993f;
  t164_n134: if (E.getNumReferences() >= 3335344128 /*26280.0*/) goto t164_n150;
  t164_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t164_n137;
  t164_n136: return 0.0549769401550293f;
  t164_n137: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t164_n147;
  t164_n138: if (E.getNumReferences() >= 3334260992 /*24164.5*/) goto t164_n144;
  t164_n139: if (E.getNumReferences() >= 3334015488 /*23685.0*/) goto t164_n143;
  t164_n140: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t164_n142;
  t164_n141: return 0.023211108520627022f;
  t164_n142: return 0.09652931988239288f;
  t164_n143: return -0.1427999883890152f;
  t164_n144: if (E.getNumReferences() >= 3334591232 /*24809.5*/) goto t164_n146;
  t164_n145: return 0.09137316793203354f;
  t164_n146: return 0.06982459872961044f;
  t164_n147: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t164_n149;
  t164_n148: return 0.13314154744148254f;
  t164_n149: return 0.07341023534536362f;
  t164_n150: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t164_n170;
  t164_n151: if (E.getNumReferences() >= 3342343040 /*47131.5*/) goto t164_n161;
  t164_n152: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t164_n154;
  t164_n153: return -0.05006397143006325f;
  t164_n154: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t164_n160;
  t164_n155: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t164_n157;
  t164_n156: return 0.07747294008731842f;
  t164_n157: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t164_n159;
  t164_n158: return 0.11580663919448853f;
  t164_n159: return 0.11254103481769562f;
  t164_n160: return 0.015823611989617348f;
  t164_n161: if (E.getNumReferences() >= 3342384384 /*47293.0*/) goto t164_n163;
  t164_n162: return 0.08358746767044067f;
  t164_n163: if (E.getNumReferences() >= 3364003440 /*267507.5*/) goto t164_n165;
  t164_n164: return 0.005508963018655777f;
  t164_n165: if (E.getNumReferences() >= 3364184064 /*273152.0*/) goto t164_n169;
  t164_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t164_n168;
  t164_n167: return 0.008725414983928204f;
  t164_n168: return 0.0846422091126442f;
  t164_n169: return 0.04366319999098778f;
  t164_n170: return -0.07135505974292755f;
  t164_n171: if (E.getNumReferences() >= 3332703232 /*21122.0*/) goto t164_n183;
  t164_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t164_n176;
  t164_n173: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t164_n175;
  t164_n174: return -0.09131752699613571f;
  t164_n175: return 0.08845420181751251f;
  t164_n176: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t164_n178;
  t164_n177: return 0.22692404687404633f;
  t164_n178: if (E.getNumReferences() >= 3332493312 /*20712.0*/) goto t164_n180;
  t164_n179: return -0.11119348555803299f;
  t164_n180: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t164_n182;
  t164_n181: return 0.11839532852172852f;
  t164_n182: return 0.3456929326057434f;
  t164_n183: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t164_n189;
  t164_n184: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t164_n188;
  t164_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t164_n187;
  t164_n186: return -0.07339950650930405f;
  t164_n187: return 0.11448441445827484f;
  t164_n188: return -0.1516655832529068f;
  t164_n189: if (E.getNumReferences() >= 3339033984 /*34205.5*/) goto t164_n193;
  t164_n190: if (E.getFractionNameInContext() >= 3179881618 /*0.0669642835855484*/) goto t164_n192;
  t164_n191: return 0.09004462510347366f;
  t164_n192: return 0.129476860165596f;
  t164_n193: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t164_n213;
  t164_n194: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t164_n196;
  t164_n195: return -0.0677686408162117f;
  t164_n196: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t164_n208;
  t164_n197: if (E.getNumReferences() >= 3342038016 /*45940.0*/) goto t164_n207;
  t164_n198: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t164_n200;
  t164_n199: return -0.12097560614347458f;
  t164_n200: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t164_n204;
  t164_n201: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t164_n203;
  t164_n202: return 0.07629207521677017f;
  t164_n203: return 0.1512058973312378f;
  t164_n204: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t164_n206;
  t164_n205: return -2.7026960849761963f;
  t164_n206: return 0.09590760618448257f;
  t164_n207: return -0.016114776954054832f;
  t164_n208: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t164_n212;
  t164_n209: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t164_n211;
  t164_n210: return 0.11607395112514496f;
  t164_n211: return 0.14664864540100098f;
  t164_n212: return -0.12122758477926254f;
  t164_n213: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t164_n215;
  t164_n214: return -0.08626097440719604f;
  t164_n215: if (E.getNumReferences() >= 3349417664 /*83997.5*/) goto t164_n223;
  t164_n216: if (E.getNumReferences() >= 3349106816 /*81569.0*/) goto t164_n222;
  t164_n217: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t164_n219;
  t164_n218: return -0.037479110062122345f;
  t164_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t164_n221;
  t164_n220: return -0.33047473430633545f;
  t164_n221: return 0.10163382440805435f;
  t164_n222: return 0.0767359733581543f;
  t164_n223: return -0.016912158578634262f;
  t164_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t164_n252;
  t164_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t164_n227;
  t164_n226: return -0.1280938684940338f;
  t164_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t164_n245;
  t164_n228: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t164_n244;
  t164_n229: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t164_n231;
  t164_n230: return 0.015307937748730183f;
  t164_n231: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t164_n233;
  t164_n232: return -0.005023008678108454f;
  t164_n233: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t164_n243;
  t164_n234: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t164_n242;
  t164_n235: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t164_n237;
  t164_n236: return 0.047590356320142746f;
  t164_n237: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t164_n241;
  t164_n238: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t164_n240;
  t164_n239: return 0.14917634427547455f;
  t164_n240: return 0.07190883159637451f;
  t164_n241: return -0.0749419555068016f;
  t164_n242: return 0.21437005698680878f;
  t164_n243: return 0.004750311374664307f;
  t164_n244: return -0.00014085489965509623f;
  t164_n245: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t164_n247;
  t164_n246: return 0.013055004179477692f;
  t164_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t164_n249;
  t164_n248: return -0.011395012959837914f;
  t164_n249: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t164_n251;
  t164_n250: return 0.01806180737912655f;
  t164_n251: return 0.10190045833587646f;
  t164_n252: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t164_n254;
  t164_n253: return 0.005306783597916365f;
  t164_n254: return -0.16531166434288025f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree165(const Example& E) {
  t165_n0: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t165_n244;
  t165_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t165_n243;
  t165_n2: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t165_n198;
  t165_n3: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t165_n197;
  t165_n4: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t165_n52;
  t165_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n51;
  t165_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t165_n50;
  t165_n7: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t165_n9;
  t165_n8: return -0.0029236311092972755f;
  t165_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t165_n15;
  t165_n10: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t165_n12;
  t165_n11: return -0.00782578531652689f;
  t165_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t165_n14;
  t165_n13: return 0.0013109466526657343f;
  t165_n14: return 0.03154264762997627f;
  t165_n15: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t165_n27;
  t165_n16: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t165_n24;
  t165_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t165_n21;
  t165_n18: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t165_n20;
  t165_n19: return 0.0015419968403875828f;
  t165_n20: return -0.09449896961450577f;
  t165_n21: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t165_n23;
  t165_n22: return 0.02644496224820614f;
  t165_n23: return 0.0016411170363426208f;
  t165_n24: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t165_n26;
  t165_n25: return 0.05477086082100868f;
  t165_n26: return -0.057522233575582504f;
  t165_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t165_n29;
  t165_n28: return 0.007085234858095646f;
  t165_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t165_n41;
  t165_n30: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t165_n38;
  t165_n31: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t165_n33;
  t165_n32: return 0.01491264346987009f;
  t165_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t165_n35;
  t165_n34: return 0.040466003119945526f;
  t165_n35: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t165_n37;
  t165_n36: return 0.11966708302497864f;
  t165_n37: return 0.02260548807680607f;
  t165_n38: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t165_n40;
  t165_n39: return 0.16297154128551483f;
  t165_n40: return -0.14246726036071777f;
  t165_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t165_n43;
  t165_n42: return 0.010947824455797672f;
  t165_n43: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t165_n49;
  t165_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t165_n46;
  t165_n45: return 0.05851330980658531f;
  t165_n46: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t165_n48;
  t165_n47: return 0.07524897903203964f;
  t165_n48: return -0.11211925745010376f;
  t165_n49: return -0.1894121915102005f;
  t165_n50: return -0.030015137046575546f;
  t165_n51: return 0.0004713015223387629f;
  t165_n52: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t165_n64;
  t165_n53: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t165_n61;
  t165_n54: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t165_n58;
  t165_n55: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t165_n57;
  t165_n56: return 0.050842516124248505f;
  t165_n57: return 0.20600363612174988f;
  t165_n58: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t165_n60;
  t165_n59: return -0.03927111253142357f;
  t165_n60: return 0.03816383332014084f;
  t165_n61: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t165_n63;
  t165_n62: return -0.09783690422773361f;
  t165_n63: return 0.1325547844171524f;
  t165_n64: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t165_n124;
  t165_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t165_n123;
  t165_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t165_n122;
  t165_n67: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t165_n75;
  t165_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t165_n70;
  t165_n69: return -0.002743548946455121f;
  t165_n70: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t165_n72;
  t165_n71: return 0.013859341852366924f;
  t165_n72: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t165_n74;
  t165_n73: return 0.048605769872665405f;
  t165_n74: return -0.055877286940813065f;
  t165_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t165_n77;
  t165_n76: return 0.0037117749452590942f;
  t165_n77: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t165_n79;
  t165_n78: return -0.04243753105401993f;
  t165_n79: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t165_n121;
  t165_n80: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t165_n98;
  t165_n81: if (E.getNumReferences() >= 3264413696 /*73.5*/) goto t165_n83;
  t165_n82: return -0.00912838988006115f;
  t165_n83: if (E.getNumReferences() >= 3264544768 /*74.5*/) goto t165_n85;
  t165_n84: return 0.13791610300540924f;
  t165_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n87;
  t165_n86: return -0.08513475209474564f;
  t165_n87: if (E.getNumReferences() >= 3276439552 /*202.5*/) goto t165_n97;
  t165_n88: if (E.getNumReferences() >= 3274899456 /*179.0*/) goto t165_n96;
  t165_n89: if (E.getNumReferences() >= 3267166208 /*94.5*/) goto t165_n93;
  t165_n90: if (E.getNumReferences() >= 3266052096 /*86.0*/) goto t165_n92;
  t165_n91: return -0.06323353201150894f;
  t165_n92: return 0.1624913066625595f;
  t165_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t165_n95;
  t165_n94: return -0.006716185715049505f;
  t165_n95: return 0.2042207419872284f;
  t165_n96: return 0.13276629149913788f;
  t165_n97: return -0.2870790958404541f;
  t165_n98: if (E.getNumReferences() >= 3261595648 /*58.0*/) goto t165_n120;
  t165_n99: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t165_n119;
  t165_n100: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t165_n112;
  t165_n101: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t165_n111;
  t165_n102: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t165_n108;
  t165_n103: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t165_n107;
  t165_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t165_n106;
  t165_n105: return 0.016153112053871155f;
  t165_n106: return 0.16467882692813873f;
  t165_n107: return 0.03664998337626457f;
  t165_n108: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t165_n110;
  t165_n109: return -0.07335028797388077f;
  t165_n110: return 0.10054195672273636f;
  t165_n111: return 0.10657127946615219f;
  t165_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t165_n116;
  t165_n113: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t165_n115;
  t165_n114: return 0.13005435466766357f;
  t165_n115: return -0.09961184859275818f;
  t165_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t165_n118;
  t165_n117: return 0.12648239731788635f;
  t165_n118: return 0.1346416473388672f;
  t165_n119: return -0.2741384506225586f;
  t165_n120: return 0.11389990150928497f;
  t165_n121: return -0.06169206276535988f;
  t165_n122: return -0.0056699542328715324f;
  t165_n123: return 0.017432061955332756f;
  t165_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t165_n196;
  t165_n125: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t165_n191;
  t165_n126: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t165_n128;
  t165_n127: return -0.03681168332695961f;
  t165_n128: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t165_n158;
  t165_n129: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t165_n149;
  t165_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n134;
  t165_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t165_n133;
  t165_n132: return -0.07522919774055481f;
  t165_n133: return -0.013570532202720642f;
  t165_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t165_n138;
  t165_n135: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t165_n137;
  t165_n136: return 0.0011707685189321637f;
  t165_n137: return -0.024497443810105324f;
  t165_n138: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t165_n140;
  t165_n139: return 0.07563535124063492f;
  t165_n140: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t165_n142;
  t165_n141: return 0.002397926291450858f;
  t165_n142: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t165_n144;
  t165_n143: return 0.02505251206457615f;
  t165_n144: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t165_n146;
  t165_n145: return 0.4509213864803314f;
  t165_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t165_n148;
  t165_n147: return 0.07333848625421524f;
  t165_n148: return 0.15410110354423523f;
  t165_n149: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t165_n157;
  t165_n150: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t165_n154;
  t165_n151: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t165_n153;
  t165_n152: return 0.03527294099330902f;
  t165_n153: return 0.06279470771551132f;
  t165_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t165_n156;
  t165_n155: return -0.04713326320052147f;
  t165_n156: return -0.007949150167405605f;
  t165_n157: return -0.007519048638641834f;
  t165_n158: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t165_n190;
  t165_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t165_n161;
  t165_n160: return 0.014390924945473671f;
  t165_n161: if (E.getNumReferences() >= 3275128832 /*182.5*/) goto t165_n179;
  t165_n162: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t165_n164;
  t165_n163: return 0.06161431223154068f;
  t165_n164: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t165_n166;
  t165_n165: return 0.3212369680404663f;
  t165_n166: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t165_n178;
  t165_n167: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t165_n177;
  t165_n168: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t165_n174;
  t165_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t165_n173;
  t165_n170: if (E.getNumReferences() >= 3241148416 /*11.0*/) goto t165_n172;
  t165_n171: return -0.10740084946155548f;
  t165_n172: return 0.1461893618106842f;
  t165_n173: return -0.09539907425642014f;
  t165_n174: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t165_n176;
  t165_n175: return 0.24949371814727783f;
  t165_n176: return 0.10335320979356766f;
  t165_n177: return 0.01681343838572502f;
  t165_n178: return 0.026649825274944305f;
  t165_n179: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t165_n181;
  t165_n180: return -0.1537436991930008f;
  t165_n181: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t165_n183;
  t165_n182: return 0.05525236204266548f;
  t165_n183: if (E.getNumReferences() >= 3275948032 /*195.0*/) goto t165_n185;
  t165_n184: return 0.3712356686592102f;
  t165_n185: if (E.getNumReferences() >= 3277684736 /*221.5*/) goto t165_n187;
  t165_n186: return -0.05420220270752907f;
  t165_n187: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t165_n189;
  t165_n188: return 0.22138772904872894f;
  t165_n189: return 0.20905743539333344f;
  t165_n190: return -0.030083253979682922f;
  t165_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t165_n193;
  t165_n192: return 0.04624368995428085f;
  t165_n193: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t165_n195;
  t165_n194: return 0.14881858229637146f;
  t165_n195: return -0.1042012870311737f;
  t165_n196: return 0.004832659848034382f;
  t165_n197: return -0.05687146633863449f;
  t165_n198: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t165_n206;
  t165_n199: if (E.getNumReferences() >= 3279716352 /*252.5*/) goto t165_n205;
  t165_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t165_n202;
  t165_n201: return 0.04921003058552742f;
  t165_n202: if (E.getNumReferences() >= 3279650816 /*251.5*/) goto t165_n204;
  t165_n203: return 0.1112370640039444f;
  t165_n204: return 0.1560727059841156f;
  t165_n205: return 0.009084844961762428f;
  t165_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t165_n208;
  t165_n207: return -0.0008619402069598436f;
  t165_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n216;
  t165_n209: if (E.getNumReferences() >= 3347950656 /*72536.5*/) goto t165_n211;
  t165_n210: return -0.04985562339425087f;
  t165_n211: if (E.getNumReferences() >= 3350874624 /*95380.0*/) goto t165_n215;
  t165_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t165_n214;
  t165_n213: return 0.07878267019987106f;
  t165_n214: return 0.19879691302776337f;
  t165_n215: return -0.09399720281362534f;
  t165_n216: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t165_n222;
  t165_n217: if (E.getNumReferences() >= 3356745728 /*151424.0*/) goto t165_n219;
  t165_n218: return 0.0005735942395403981f;
  t165_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n221;
  t165_n220: return -0.1509617567062378f;
  t165_n221: return 0.021592946723103523f;
  t165_n222: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t165_n242;
  t165_n223: if (E.getNumReferences() >= 3333346816 /*22379.0*/) goto t165_n235;
  t165_n224: if (E.getNumReferences() >= 3329239040 /*15370.0*/) goto t165_n226;
  t165_n225: return 0.05329960957169533f;
  t165_n226: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t165_n234;
  t165_n227: if (E.getNumReferences() >= 3331047168 /*17887.5*/) goto t165_n231;
  t165_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n230;
  t165_n229: return -0.12332458794116974f;
  t165_n230: return 0.14959093928337097f;
  t165_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t165_n233;
  t165_n232: return -0.13405710458755493f;
  t165_n233: return 0.11847910284996033f;
  t165_n234: return -0.14816436171531677f;
  t165_n235: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t165_n241;
  t165_n236: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t165_n240;
  t165_n237: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t165_n239;
  t165_n238: return -0.036215826869010925f;
  t165_n239: return 0.051045410335063934f;
  t165_n240: return 0.07447464764118195f;
  t165_n241: return -0.1290968358516693f;
  t165_n242: return 0.09330088645219803f;
  t165_n243: return 0.003437994746491313f;
  t165_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t165_n246;
  t165_n245: return -0.17975138127803802f;
  t165_n246: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t165_n254;
  t165_n247: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t165_n249;
  t165_n248: return 0.00512775219976902f;
  t165_n249: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t165_n253;
  t165_n250: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t165_n252;
  t165_n251: return 0.13386794924736023f;
  t165_n252: return -0.07915378361940384f;
  t165_n253: return -0.008529760874807835f;
  t165_n254: return -0.08972468227148056f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree166(const Example& E) {
  t166_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t166_n2;
  t166_n1: return -0.0005644413759000599f;
  t166_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t166_n110;
  t166_n3: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t166_n87;
  t166_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t166_n50;
  t166_n5: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t166_n19;
  t166_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t166_n8;
  t166_n7: return 0.012694804929196835f;
  t166_n8: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t166_n12;
  t166_n9: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t166_n11;
  t166_n10: return -0.06138390302658081f;
  t166_n11: return 0.032986756414175034f;
  t166_n12: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t166_n18;
  t166_n13: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t166_n15;
  t166_n14: return 0.03628988191485405f;
  t166_n15: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t166_n17;
  t166_n16: return 0.16737397015094757f;
  t166_n17: return -0.10091403871774673f;
  t166_n18: return 0.15127640962600708f;
  t166_n19: if (E.getNumReferences() >= 3297988608 /*1178.5*/) goto t166_n45;
  t166_n20: if (E.getNumReferences() >= 3297841152 /*1160.5*/) goto t166_n40;
  t166_n21: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t166_n23;
  t166_n22: return 4.960523801855743e-05f;
  t166_n23: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t166_n25;
  t166_n24: return -0.07289483398199081f;
  t166_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t166_n35;
  t166_n26: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n30;
  t166_n27: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t166_n29;
  t166_n28: return -0.04335103556513786f;
  t166_n29: return 0.007670744322240353f;
  t166_n30: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t166_n34;
  t166_n31: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t166_n33;
  t166_n32: return 0.33099475502967834f;
  t166_n33: return 0.15088237822055817f;
  t166_n34: return -0.016327138990163803f;
  t166_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t166_n37;
  t166_n36: return -0.02640494517982006f;
  t166_n37: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t166_n39;
  t166_n38: return 0.10250990092754364f;
  t166_n39: return 0.013912368565797806f;
  t166_n40: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t166_n42;
  t166_n41: return 0.06913410872220993f;
  t166_n42: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t166_n44;
  t166_n43: return 0.1524650752544403f;
  t166_n44: return 0.16069020330905914f;
  t166_n45: if (E.getNumReferences() >= 3298123776 /*1195.0*/) goto t166_n49;
  t166_n46: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t166_n48;
  t166_n47: return -0.10519301891326904f;
  t166_n48: return -0.920111894607544f;
  t166_n49: return -0.059487566351890564f;
  t166_n50: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t166_n56;
  t166_n51: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n53;
  t166_n52: return -0.059675682336091995f;
  t166_n53: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t166_n55;
  t166_n54: return -0.10965836048126221f;
  t166_n55: return -0.21537521481513977f;
  t166_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t166_n58;
  t166_n57: return -0.044676586985588074f;
  t166_n58: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t166_n86;
  t166_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t166_n67;
  t166_n60: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t166_n62;
  t166_n61: return -0.0547635480761528f;
  t166_n62: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t166_n66;
  t166_n63: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t166_n65;
  t166_n64: return 0.10147929191589355f;
  t166_n65: return -0.23978307843208313f;
  t166_n66: return 0.12189212441444397f;
  t166_n67: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t166_n85;
  t166_n68: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t166_n70;
  t166_n69: return -0.06692530959844589f;
  t166_n70: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n78;
  t166_n71: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t166_n77;
  t166_n72: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t166_n74;
  t166_n73: return 0.10617517679929733f;
  t166_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t166_n76;
  t166_n75: return -0.12747624516487122f;
  t166_n76: return 0.1810598224401474f;
  t166_n77: return -0.0007444637012667954f;
  t166_n78: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t166_n80;
  t166_n79: return -0.13348956406116486f;
  t166_n80: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t166_n84;
  t166_n81: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t166_n83;
  t166_n82: return 0.13613012433052063f;
  t166_n83: return 0.04108451306819916f;
  t166_n84: return -0.2642333209514618f;
  t166_n85: return 0.13345323503017426f;
  t166_n86: return -0.1092379093170166f;
  t166_n87: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t166_n99;
  t166_n88: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t166_n92;
  t166_n89: if (E.getNumReferences() >= 3298697216 /*1265.0*/) goto t166_n91;
  t166_n90: return 0.0949530303478241f;
  t166_n91: return 0.015406324528157711f;
  t166_n92: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t166_n98;
  t166_n93: if (E.getNumReferences() >= 3321480192 /*7992.5*/) goto t166_n95;
  t166_n94: return -0.0807984471321106f;
  t166_n95: if (E.getNumReferences() >= 3328018944 /*14178.5*/) goto t166_n97;
  t166_n96: return 0.13409149646759033f;
  t166_n97: return 0.0667891874909401f;
  t166_n98: return 0.06676560640335083f;
  t166_n99: if (E.getNumReferences() >= 3324926976 /*11159.0*/) goto t166_n109;
  t166_n100: if (E.getNumReferences() >= 3324868096 /*11101.5*/) goto t166_n108;
  t166_n101: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t166_n107;
  t166_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t166_n104;
  t166_n103: return -0.022632095962762833f;
  t166_n104: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t166_n106;
  t166_n105: return 0.10191993415355682f;
  t166_n106: return 0.03444529324769974f;
  t166_n107: return 0.0368274562060833f;
  t166_n108: return 0.09583530575037003f;
  t166_n109: return -0.022099394351243973f;
  t166_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t166_n250;
  t166_n111: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t166_n249;
  t166_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t166_n142;
  t166_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n123;
  t166_n114: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t166_n122;
  t166_n115: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t166_n121;
  t166_n116: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t166_n118;
  t166_n117: return 0.009596109390258789f;
  t166_n118: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t166_n120;
  t166_n119: return -0.20440399646759033f;
  t166_n120: return -0.020043130964040756f;
  t166_n121: return -0.005907696671783924f;
  t166_n122: return 0.009299027733504772f;
  t166_n123: if (E.getNumReferences() >= 3363951184 /*265874.5*/) goto t166_n141;
  t166_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t166_n130;
  t166_n125: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t166_n129;
  t166_n126: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t166_n128;
  t166_n127: return -0.015022020787000656f;
  t166_n128: return 0.03339194878935814f;
  t166_n129: return 0.003103866707533598f;
  t166_n130: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t166_n140;
  t166_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t166_n135;
  t166_n132: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t166_n134;
  t166_n133: return 0.03994268178939819f;
  t166_n134: return -0.04921839013695717f;
  t166_n135: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t166_n139;
  t166_n136: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t166_n138;
  t166_n137: return 0.0825822576880455f;
  t166_n138: return 0.04419848695397377f;
  t166_n139: return -0.12590725719928741f;
  t166_n140: return -0.010072006843984127f;
  t166_n141: return 0.1160593330860138f;
  t166_n142: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t166_n144;
  t166_n143: return 0.007008376065641642f;
  t166_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t166_n200;
  t166_n145: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n157;
  t166_n146: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t166_n148;
  t166_n147: return 0.004102547653019428f;
  t166_n148: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t166_n156;
  t166_n149: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t166_n151;
  t166_n150: return 0.017122454941272736f;
  t166_n151: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t166_n155;
  t166_n152: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t166_n154;
  t166_n153: return -0.11643616855144501f;
  t166_n154: return -0.2134949117898941f;
  t166_n155: return -0.0149850370362401f;
  t166_n156: return 0.031500112265348434f;
  t166_n157: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t166_n197;
  t166_n158: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t166_n192;
  t166_n159: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t166_n185;
  t166_n160: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t166_n172;
  t166_n161: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t166_n165;
  t166_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t166_n164;
  t166_n163: return -0.12491902709007263f;
  t166_n164: return 0.1003219410777092f;
  t166_n165: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t166_n169;
  t166_n166: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t166_n168;
  t166_n167: return 0.07867147773504257f;
  t166_n168: return -0.03408712148666382f;
  t166_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t166_n171;
  t166_n170: return 0.04888079687952995f;
  t166_n171: return 0.08340130001306534f;
  t166_n172: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t166_n178;
  t166_n173: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t166_n175;
  t166_n174: return 0.07690365612506866f;
  t166_n175: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t166_n177;
  t166_n176: return 0.1253814995288849f;
  t166_n177: return -0.002691572764888406f;
  t166_n178: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t166_n182;
  t166_n179: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t166_n181;
  t166_n180: return 0.013391800224781036f;
  t166_n181: return -0.22725126147270203f;
  t166_n182: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t166_n184;
  t166_n183: return -0.046266455203294754f;
  t166_n184: return 0.13538815081119537f;
  t166_n185: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t166_n187;
  t166_n186: return -0.10433700680732727f;
  t166_n187: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t166_n191;
  t166_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t166_n190;
  t166_n189: return 0.11417494714260101f;
  t166_n190: return 0.06816904991865158f;
  t166_n191: return -0.018009142950177193f;
  t166_n192: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t166_n196;
  t166_n193: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t166_n195;
  t166_n194: return 0.14026787877082825f;
  t166_n195: return 0.030864005908370018f;
  t166_n196: return -0.18348775804042816f;
  t166_n197: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t166_n199;
  t166_n198: return -0.31345614790916443f;
  t166_n199: return 0.12176647037267685f;
  t166_n200: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t166_n244;
  t166_n201: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t166_n243;
  t166_n202: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t166_n204;
  t166_n203: return 0.025157608091831207f;
  t166_n204: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t166_n230;
  t166_n205: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t166_n207;
  t166_n206: return 0.06335695832967758f;
  t166_n207: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t166_n215;
  t166_n208: if (E.getNumReferences() >= 3242196992 /*12.0*/) goto t166_n214;
  t166_n209: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t166_n211;
  t166_n210: return -0.09507425874471664f;
  t166_n211: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t166_n213;
  t166_n212: return 0.08269830793142319f;
  t166_n213: return -0.01017773523926735f;
  t166_n214: return 0.08550434559583664f;
  t166_n215: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t166_n225;
  t166_n216: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t166_n224;
  t166_n217: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t166_n219;
  t166_n218: return 0.06567375361919403f;
  t166_n219: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t166_n223;
  t166_n220: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t166_n222;
  t166_n221: return 0.06955226510763168f;
  t166_n222: return 0.12423200905323029f;
  t166_n223: return 0.04533703252673149f;
  t166_n224: return -0.286262184381485f;
  t166_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t166_n229;
  t166_n226: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t166_n228;
  t166_n227: return -0.5252321362495422f;
  t166_n228: return -0.01672186143696308f;
  t166_n229: return 0.10539969801902771f;
  t166_n230: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t166_n242;
  t166_n231: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t166_n239;
  t166_n232: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t166_n234;
  t166_n233: return 0.019004113972187042f;
  t166_n234: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t166_n238;
  t166_n235: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t166_n237;
  t166_n236: return 0.20015385746955872f;
  t166_n237: return 0.211528941988945f;
  t166_n238: return 0.03864600136876106f;
  t166_n239: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t166_n241;
  t166_n240: return -0.10283330082893372f;
  t166_n241: return 0.10506977140903473f;
  t166_n242: return 0.1211220771074295f;
  t166_n243: return 0.015028651803731918f;
  t166_n244: if (E.getFractionNameInContext() >= 3206464748 /*0.620192289352417*/) goto t166_n246;
  t166_n245: return 0.01963602937757969f;
  t166_n246: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t166_n248;
  t166_n247: return -0.1925618052482605f;
  t166_n248: return 0.0717591643333435f;
  t166_n249: return 0.000995972310192883f;
  t166_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t166_n254;
  t166_n251: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t166_n253;
  t166_n252: return 0.055777862668037415f;
  t166_n253: return 0.13414043188095093f;
  t166_n254: return -0.10885307192802429f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree167(const Example& E) {
  t167_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t167_n218;
  t167_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t167_n177;
  t167_n2: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t167_n138;
  t167_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t167_n75;
  t167_n4: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t167_n54;
  t167_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t167_n19;
  t167_n6: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t167_n8;
  t167_n7: return 0.0012257241178303957f;
  t167_n8: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t167_n10;
  t167_n9: return -0.03986909240484238f;
  t167_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t167_n16;
  t167_n11: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t167_n13;
  t167_n12: return 0.028675558045506477f;
  t167_n13: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t167_n15;
  t167_n14: return 0.2647033929824829f;
  t167_n15: return 0.0633503869175911f;
  t167_n16: if (E.getFractionNameInContext() >= 3190187622 /*0.16249999403953552*/) goto t167_n18;
  t167_n17: return 0.12386880069971085f;
  t167_n18: return 0.09897810220718384f;
  t167_n19: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t167_n25;
  t167_n20: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t167_n22;
  t167_n21: return 0.010302100330591202f;
  t167_n22: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t167_n24;
  t167_n23: return -0.015463192947208881f;
  t167_n24: return 0.10935923457145691f;
  t167_n25: if (E.getNumReferences() >= 3338508288 /*32460.0*/) goto t167_n53;
  t167_n26: if (E.getNumReferences() >= 3337906944 /*31285.5*/) goto t167_n48;
  t167_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t167_n29;
  t167_n28: return 0.014061744324862957f;
  t167_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t167_n31;
  t167_n30: return 0.03992382809519768f;
  t167_n31: if (E.getNumReferences() >= 3273818112 /*162.5*/) goto t167_n33;
  t167_n32: return 0.03947185352444649f;
  t167_n33: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t167_n47;
  t167_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t167_n38;
  t167_n35: if (E.getNumReferences() >= 3278471168 /*233.5*/) goto t167_n37;
  t167_n36: return 0.06887650489807129f;
  t167_n37: return 0.23694458603858948f;
  t167_n38: if (E.getNumReferences() >= 3276374016 /*201.5*/) goto t167_n46;
  t167_n39: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t167_n45;
  t167_n40: if (E.getNumReferences() >= 3275423744 /*187.0*/) goto t167_n44;
  t167_n41: if (E.getNumReferences() >= 3274473472 /*172.5*/) goto t167_n43;
  t167_n42: return 0.17016522586345673f;
  t167_n43: return -0.12050798535346985f;
  t167_n44: return 0.27583709359169006f;
  t167_n45: return -0.11360412836074829f;
  t167_n46: return 0.015348920598626137f;
  t167_n47: return 0.07256753742694855f;
  t167_n48: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t167_n52;
  t167_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n51;
  t167_n50: return -0.10069107264280319f;
  t167_n51: return 0.10891922563314438f;
  t167_n52: return 0.11082372069358826f;
  t167_n53: return -0.048094552010297775f;
  t167_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n56;
  t167_n55: return -0.03009001910686493f;
  t167_n56: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t167_n74;
  t167_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t167_n69;
  t167_n58: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t167_n66;
  t167_n59: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t167_n61;
  t167_n60: return -7.095365435816348e-05f;
  t167_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n65;
  t167_n62: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t167_n64;
  t167_n63: return -0.18525268137454987f;
  t167_n64: return 0.07645636796951294f;
  t167_n65: return -0.042981475591659546f;
  t167_n66: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t167_n68;
  t167_n67: return 0.04849327355623245f;
  t167_n68: return 0.006348276510834694f;
  t167_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t167_n71;
  t167_n70: return 0.014854248613119125f;
  t167_n71: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t167_n73;
  t167_n72: return 0.06574555486440659f;
  t167_n73: return 0.06545058637857437f;
  t167_n74: return -0.011908996850252151f;
  t167_n75: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t167_n131;
  t167_n76: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t167_n86;
  t167_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t167_n85;
  t167_n78: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t167_n84;
  t167_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n81;
  t167_n80: return -0.04977096617221832f;
  t167_n81: if (E.getNumReferences() >= 3281518592 /*304.0*/) goto t167_n83;
  t167_n82: return 0.0189673975110054f;
  t167_n83: return 0.12112802267074585f;
  t167_n84: return 0.019752662628889084f;
  t167_n85: return 0.0444716215133667f;
  t167_n86: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t167_n102;
  t167_n87: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t167_n99;
  t167_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n96;
  t167_n89: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t167_n93;
  t167_n90: if (E.getNumReferences() >= 3289423872 /*578.5*/) goto t167_n92;
  t167_n91: return -0.05810143053531647f;
  t167_n92: return 0.049621496349573135f;
  t167_n93: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t167_n95;
  t167_n94: return 0.0077498978935182095f;
  t167_n95: return -0.05708300322294235f;
  t167_n96: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t167_n98;
  t167_n97: return 0.004453714936971664f;
  t167_n98: return -0.09801670163869858f;
  t167_n99: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t167_n101;
  t167_n100: return 0.02861766144633293f;
  t167_n101: return -0.028774093836545944f;
  t167_n102: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t167_n118;
  t167_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t167_n105;
  t167_n104: return 0.011645583435893059f;
  t167_n105: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t167_n115;
  t167_n106: if (E.getFractionNameInContext() >= 3179644278 /*0.0651959627866745*/) goto t167_n114;
  t167_n107: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t167_n111;
  t167_n108: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t167_n110;
  t167_n109: return 0.05792824923992157f;
  t167_n110: return -0.1300962269306183f;
  t167_n111: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t167_n113;
  t167_n112: return 0.0772150456905365f;
  t167_n113: return 0.037819672375917435f;
  t167_n114: return 0.15555641055107117f;
  t167_n115: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t167_n117;
  t167_n116: return 0.060688138008117676f;
  t167_n117: return -0.1699964702129364f;
  t167_n118: if (E.getNumReferences() >= 3282452480 /*332.5*/) goto t167_n126;
  t167_n119: if (E.getNumReferences() >= 3277094912 /*212.5*/) goto t167_n123;
  t167_n120: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t167_n122;
  t167_n121: return -0.06963741779327393f;
  t167_n122: return 0.04260503873229027f;
  t167_n123: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n125;
  t167_n124: return -0.31327787041664124f;
  t167_n125: return 0.1300140619277954f;
  t167_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t167_n128;
  t167_n127: return 0.017332671210169792f;
  t167_n128: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t167_n130;
  t167_n129: return 0.12607227265834808f;
  t167_n130: return 0.06080387160181999f;
  t167_n131: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t167_n137;
  t167_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n134;
  t167_n133: return -0.05842669680714607f;
  t167_n134: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t167_n136;
  t167_n135: return 0.0589633546769619f;
  t167_n136: return 0.033550016582012177f;
  t167_n137: return 0.01855562813580036f;
  t167_n138: if (E.getNumReferences() >= 3301855232 /*1650.5*/) goto t167_n140;
  t167_n139: return -0.0010049949632957578f;
  t167_n140: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t167_n144;
  t167_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t167_n143;
  t167_n142: return 0.00881852488964796f;
  t167_n143: return 0.11963656544685364f;
  t167_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t167_n148;
  t167_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n147;
  t167_n146: return -0.05795470252633095f;
  t167_n147: return -0.0004998232470825315f;
  t167_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t167_n176;
  t167_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n151;
  t167_n150: return -0.025277290493249893f;
  t167_n151: if (E.getNumReferences() >= 3346600320 /*63761.5*/) goto t167_n175;
  t167_n152: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t167_n154;
  t167_n153: return 0.006580500863492489f;
  t167_n154: if (E.getNumReferences() >= 3343775872 /*52728.5*/) goto t167_n162;
  t167_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t167_n159;
  t167_n156: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n158;
  t167_n157: return 0.08935976028442383f;
  t167_n158: return 0.0171610526740551f;
  t167_n159: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t167_n161;
  t167_n160: return 0.15049029886722565f;
  t167_n161: return 0.1405654102563858f;
  t167_n162: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t167_n174;
  t167_n163: if (E.getNumReferences() >= 3345974912 /*61318.5*/) goto t167_n165;
  t167_n164: return -0.16463127732276917f;
  t167_n165: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t167_n167;
  t167_n166: return 0.0012461986625567079f;
  t167_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t167_n169;
  t167_n168: return 0.02974359504878521f;
  t167_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n173;
  t167_n170: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t167_n172;
  t167_n171: return 0.14271233975887299f;
  t167_n172: return 0.12542220950126648f;
  t167_n173: return -0.11386696249246597f;
  t167_n174: return 0.07338747382164001f;
  t167_n175: return -0.01969371736049652f;
  t167_n176: return 0.0147407166659832f;
  t167_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t167_n211;
  t167_n178: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t167_n188;
  t167_n179: if (E.getNumReferences() >= 3364626832 /*286988.5*/) goto t167_n187;
  t167_n180: if (E.getNumReferences() >= 3352044608 /*104520.5*/) goto t167_n186;
  t167_n181: if (E.getNumReferences() >= 3313734656 /*4210.5*/) goto t167_n183;
  t167_n182: return 0.002568311057984829f;
  t167_n183: if (E.getNumReferences() >= 3330759936 /*17326.5*/) goto t167_n185;
  t167_n184: return -0.09838132560253143f;
  t167_n185: return 0.003076181747019291f;
  t167_n186: return 0.04967779293656349f;
  t167_n187: return -0.04431472346186638f;
  t167_n188: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t167_n210;
  t167_n189: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t167_n199;
  t167_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t167_n196;
  t167_n191: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t167_n193;
  t167_n192: return -0.1527663916349411f;
  t167_n193: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t167_n195;
  t167_n194: return 0.04278524965047836f;
  t167_n195: return 0.118968166410923f;
  t167_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n198;
  t167_n197: return 0.03358161449432373f;
  t167_n198: return 0.09614847600460052f;
  t167_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t167_n201;
  t167_n200: return 0.04450786113739014f;
  t167_n201: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t167_n209;
  t167_n202: if (E.getNumReferences() >= 3262119936 /*60.0*/) goto t167_n208;
  t167_n203: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t167_n207;
  t167_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t167_n206;
  t167_n205: return -0.7392513751983643f;
  t167_n206: return 0.09713813662528992f;
  t167_n207: return 0.13121989369392395f;
  t167_n208: return -0.3427818715572357f;
  t167_n209: return 0.04883909970521927f;
  t167_n210: return 0.0017474597552791238f;
  t167_n211: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t167_n217;
  t167_n212: if (E.getNumReferences() >= 3301859328 /*1651.0*/) goto t167_n216;
  t167_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n215;
  t167_n214: return -0.04668530076742172f;
  t167_n215: return 0.01544148102402687f;
  t167_n216: return 0.055696144700050354f;
  t167_n217: return 0.044080596417188644f;
  t167_n218: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t167_n226;
  t167_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t167_n225;
  t167_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t167_n222;
  t167_n221: return -0.0036808561999350786f;
  t167_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t167_n224;
  t167_n223: return 0.010121431201696396f;
  t167_n224: return 0.048699140548706055f;
  t167_n225: return 0.03864504024386406f;
  t167_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t167_n252;
  t167_n227: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t167_n245;
  t167_n228: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t167_n238;
  t167_n229: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t167_n233;
  t167_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t167_n232;
  t167_n231: return 0.04093291610479355f;
  t167_n232: return 0.16822411119937897f;
  t167_n233: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t167_n237;
  t167_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n236;
  t167_n235: return -0.017411096021533012f;
  t167_n236: return 0.018246904015541077f;
  t167_n237: return -0.12130056321620941f;
  t167_n238: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t167_n242;
  t167_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement))) goto t167_n241;
  t167_n240: return -0.008308108896017075f;
  t167_n241: return 0.03550827130675316f;
  t167_n242: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t167_n244;
  t167_n243: return -0.1771153062582016f;
  t167_n244: return 0.06393175572156906f;
  t167_n245: if (E.getFractionNameInContext() >= 3177483234 /*0.05579746514558792*/) goto t167_n249;
  t167_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n248;
  t167_n247: return -0.044739868491888046f;
  t167_n248: return 0.04362919554114342f;
  t167_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n251;
  t167_n250: return -0.14287208020687103f;
  t167_n251: return -0.07046035677194595f;
  t167_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t167_n254;
  t167_n253: return -0.028716817498207092f;
  t167_n254: return 0.005597393959760666f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree168(const Example& E) {
  t168_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t168_n254;
  t168_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t168_n205;
  t168_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t168_n184;
  t168_n3: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t168_n35;
  t168_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t168_n34;
  t168_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t168_n33;
  t168_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t168_n26;
  t168_n7: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t168_n9;
  t168_n8: return -0.004165472462773323f;
  t168_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t168_n21;
  t168_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t168_n16;
  t168_n11: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t168_n15;
  t168_n12: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t168_n14;
  t168_n13: return -0.005893779452890158f;
  t168_n14: return -0.21936126053333282f;
  t168_n15: return 0.1189756989479065f;
  t168_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t168_n20;
  t168_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t168_n19;
  t168_n18: return -0.09546974301338196f;
  t168_n19: return 0.02905449829995632f;
  t168_n20: return 0.046686965972185135f;
  t168_n21: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t168_n25;
  t168_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t168_n24;
  t168_n23: return -0.06817266345024109f;
  t168_n24: return 0.11865009367465973f;
  t168_n25: return -0.013141202740371227f;
  t168_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t168_n32;
  t168_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t168_n29;
  t168_n28: return -0.12660625576972961f;
  t168_n29: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t168_n31;
  t168_n30: return 0.022647887468338013f;
  t168_n31: return 0.07577241957187653f;
  t168_n32: return 0.038609765470027924f;
  t168_n33: return -0.00370642333291471f;
  t168_n34: return 0.0008963476284407079f;
  t168_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t168_n175;
  t168_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t168_n136;
  t168_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t168_n111;
  t168_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t168_n86;
  t168_n39: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t168_n73;
  t168_n40: if (E.getNumReferences() >= 3310442496 /*3349.5*/) goto t168_n70;
  t168_n41: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t168_n47;
  t168_n42: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t168_n46;
  t168_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t168_n45;
  t168_n44: return -0.02989555150270462f;
  t168_n45: return 0.01902512088418007f;
  t168_n46: return -0.018523938953876495f;
  t168_n47: if (E.getNumReferences() >= 3309832192 /*3200.5*/) goto t168_n67;
  t168_n48: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t168_n64;
  t168_n49: if (E.getNumReferences() >= 3295412224 /*944.0*/) goto t168_n63;
  t168_n50: if (E.getNumReferences() >= 3294912512 /*913.5*/) goto t168_n60;
  t168_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t168_n55;
  t168_n52: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t168_n54;
  t168_n53: return -0.06627847254276276f;
  t168_n54: return -0.000989198568277061f;
  t168_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t168_n59;
  t168_n56: if (E.getNumReferences() >= 3289858048 /*605.0*/) goto t168_n58;
  t168_n57: return -0.07576112449169159f;
  t168_n58: return 0.06255248188972473f;
  t168_n59: return -0.011627067811787128f;
  t168_n60: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t168_n62;
  t168_n61: return 0.04985763877630234f;
  t168_n62: return 0.19352230429649353f;
  t168_n63: return -0.06416691839694977f;
  t168_n64: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t168_n66;
  t168_n65: return 0.01221637986600399f;
  t168_n66: return -0.01571207493543625f;
  t168_n67: if (E.getNumReferences() >= 3309864960 /*3208.5*/) goto t168_n69;
  t168_n68: return 0.10235190391540527f;
  t168_n69: return -0.11800247430801392f;
  t168_n70: if (E.getNumReferences() >= 3311284224 /*3555.0*/) goto t168_n72;
  t168_n71: return -0.38128575682640076f;
  t168_n72: return -0.07679466158151627f;
  t168_n73: if (E.getNumReferences() >= 3263627264 /*67.5*/) goto t168_n75;
  t168_n74: return 0.015224739909172058f;
  t168_n75: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t168_n77;
  t168_n76: return 0.031317323446273804f;
  t168_n77: if (E.getNumReferences() >= 3269132288 /*109.5*/) goto t168_n85;
  t168_n78: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t168_n84;
  t168_n79: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t168_n83;
  t168_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t168_n82;
  t168_n81: return 0.07382597774267197f;
  t168_n82: return 0.16413600742816925f;
  t168_n83: return 0.13096794486045837f;
  t168_n84: return 0.023097213357686996f;
  t168_n85: return 0.051656465977430344f;
  t168_n86: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t168_n104;
  t168_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t168_n103;
  t168_n88: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t168_n94;
  t168_n89: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t168_n93;
  t168_n90: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t168_n92;
  t168_n91: return 0.0008732673595659435f;
  t168_n92: return 0.021677277982234955f;
  t168_n93: return -0.011693492531776428f;
  t168_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t168_n96;
  t168_n95: return 0.012222942896187305f;
  t168_n96: if (E.getNumReferences() >= 3265462272 /*81.5*/) goto t168_n98;
  t168_n97: return 0.048476628959178925f;
  t168_n98: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t168_n100;
  t168_n99: return 0.3824116587638855f;
  t168_n100: if (E.getNumReferences() >= 3266379776 /*88.5*/) goto t168_n102;
  t168_n101: return 0.12087533622980118f;
  t168_n102: return 0.050883639603853226f;
  t168_n103: return 0.15274415910243988f;
  t168_n104: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t168_n110;
  t168_n105: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t168_n109;
  t168_n106: if (E.getNumReferences() >= 3246391296 /*16.0*/) goto t168_n108;
  t168_n107: return 0.20402514934539795f;
  t168_n108: return 0.3629390597343445f;
  t168_n109: return 0.07777874171733856f;
  t168_n110: return -0.017230883240699768f;
  t168_n111: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t168_n121;
  t168_n112: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t168_n116;
  t168_n113: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t168_n115;
  t168_n114: return 0.06532544642686844f;
  t168_n115: return 0.1146358996629715f;
  t168_n116: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t168_n120;
  t168_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t168_n119;
  t168_n118: return -0.054480139166116714f;
  t168_n119: return 0.020532706752419472f;
  t168_n120: return 0.02237461879849434f;
  t168_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t168_n123;
  t168_n122: return -0.02160889282822609f;
  t168_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t168_n133;
  t168_n124: if (E.getNumReferences() >= 3329498112 /*15623.0*/) goto t168_n126;
  t168_n125: return -0.02755855768918991f;
  t168_n126: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t168_n132;
  t168_n127: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t168_n129;
  t168_n128: return -0.15744435787200928f;
  t168_n129: if (E.getNumReferences() >= 3335233024 /*26063.0*/) goto t168_n131;
  t168_n130: return -0.7436381578445435f;
  t168_n131: return -0.3553686738014221f;
  t168_n132: return -0.07504813373088837f;
  t168_n133: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t168_n135;
  t168_n134: return 0.019426610320806503f;
  t168_n135: return 0.06979985535144806f;
  t168_n136: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t168_n168;
  t168_n137: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t168_n167;
  t168_n138: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t168_n140;
  t168_n139: return 0.008545049466192722f;
  t168_n140: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t168_n162;
  t168_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t168_n145;
  t168_n142: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t168_n144;
  t168_n143: return -1.0353567600250244f;
  t168_n144: return -0.026491817086935043f;
  t168_n145: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t168_n159;
  t168_n146: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t168_n148;
  t168_n147: return -0.030410373583436012f;
  t168_n148: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t168_n158;
  t168_n149: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t168_n155;
  t168_n150: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t168_n152;
  t168_n151: return 0.03368053957819939f;
  t168_n152: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t168_n154;
  t168_n153: return 0.10674668848514557f;
  t168_n154: return -0.08325496315956116f;
  t168_n155: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t168_n157;
  t168_n156: return 0.11177898943424225f;
  t168_n157: return 0.012271793559193611f;
  t168_n158: return 0.03834247216582298f;
  t168_n159: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t168_n161;
  t168_n160: return 0.023201677948236465f;
  t168_n161: return -0.13749870657920837f;
  t168_n162: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t168_n166;
  t168_n163: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t168_n165;
  t168_n164: return 0.11094898730516434f;
  t168_n165: return 0.09185923635959625f;
  t168_n166: return 0.06661698967218399f;
  t168_n167: return -0.13003607094287872f;
  t168_n168: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t168_n172;
  t168_n169: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t168_n171;
  t168_n170: return 0.05379575490951538f;
  t168_n171: return 0.09466330707073212f;
  t168_n172: if (E.getNumReferences() >= 3335283968 /*26162.5*/) goto t168_n174;
  t168_n173: return -0.052204232662916183f;
  t168_n174: return 0.015254128724336624f;
  t168_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t168_n183;
  t168_n176: if (E.getNumReferences() >= 3341720320 /*44699.0*/) goto t168_n182;
  t168_n177: if (E.getNumReferences() >= 3341362816 /*43302.5*/) goto t168_n181;
  t168_n178: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t168_n180;
  t168_n179: return 0.1293143481016159f;
  t168_n180: return -0.016902947798371315f;
  t168_n181: return 0.12332268059253693f;
  t168_n182: return -0.07289230823516846f;
  t168_n183: return 0.0031754393130540848f;
  t168_n184: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t168_n200;
  t168_n185: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t168_n191;
  t168_n186: if (E.getNumReferences() >= 3330058752 /*16170.5*/) goto t168_n188;
  t168_n187: return 0.004625796806067228f;
  t168_n188: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t168_n190;
  t168_n189: return -0.05346252769231796f;
  t168_n190: return 0.12655051052570343f;
  t168_n191: if (E.getNumReferences() >= 3328232960 /*14387.5*/) goto t168_n193;
  t168_n192: return -0.002255527302622795f;
  t168_n193: if (E.getNumReferences() >= 3329021952 /*15158.0*/) goto t168_n199;
  t168_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t168_n196;
  t168_n195: return 0.0576048269867897f;
  t168_n196: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t168_n198;
  t168_n197: return 0.07964672148227692f;
  t168_n198: return 0.11580841988325119f;
  t168_n199: return 0.00301585765555501f;
  t168_n200: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t168_n204;
  t168_n201: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t168_n203;
  t168_n202: return -0.043514396995306015f;
  t168_n203: return 0.051557399332523346f;
  t168_n204: return 0.092947818338871f;
  t168_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t168_n213;
  t168_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t168_n208;
  t168_n207: return 0.0004901962238363922f;
  t168_n208: if (E.getNumNameInContext() >= 3239575552 /*9.5*/) goto t168_n212;
  t168_n209: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t168_n211;
  t168_n210: return 0.012068959884345531f;
  t168_n211: return -0.007987309247255325f;
  t168_n212: return 0.09861753135919571f;
  t168_n213: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t168_n217;
  t168_n214: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t168_n216;
  t168_n215: return 0.0036068479530513287f;
  t168_n216: return -0.11241675168275833f;
  t168_n217: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t168_n249;
  t168_n218: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t168_n234;
  t168_n219: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t168_n231;
  t168_n220: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t168_n230;
  t168_n221: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t168_n227;
  t168_n222: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t168_n226;
  t168_n223: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t168_n225;
  t168_n224: return 0.16243422031402588f;
  t168_n225: return 0.01851595565676689f;
  t168_n226: return 0.053483158349990845f;
  t168_n227: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t168_n229;
  t168_n228: return -0.02191561833024025f;
  t168_n229: return 0.13651777803897858f;
  t168_n230: return 0.02010670304298401f;
  t168_n231: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t168_n233;
  t168_n232: return 0.00726708397269249f;
  t168_n233: return -0.3215225338935852f;
  t168_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t168_n240;
  t168_n235: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t168_n237;
  t168_n236: return -0.14204750955104828f;
  t168_n237: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t168_n239;
  t168_n238: return 0.1290321797132492f;
  t168_n239: return 0.041485633701086044f;
  t168_n240: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t168_n244;
  t168_n241: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t168_n243;
  t168_n242: return 0.10570266097784042f;
  t168_n243: return 0.12407169491052628f;
  t168_n244: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t168_n248;
  t168_n245: if (E.getFractionNameInContext() >= 3174706828 /*0.04545454680919647*/) goto t168_n247;
  t168_n246: return 0.06100369989871979f;
  t168_n247: return 0.12490072846412659f;
  t168_n248: return -0.08853339403867722f;
  t168_n249: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t168_n253;
  t168_n250: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t168_n252;
  t168_n251: return -0.17736971378326416f;
  t168_n252: return 0.05194623023271561f;
  t168_n253: return 0.01795024238526821f;
  t168_n254: return -0.018576262518763542f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree169(const Example& E) {
  t169_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t169_n246;
  t169_n1: if (E.getNumReferences() >= 3376956888 /*820317.5*/) goto t169_n241;
  t169_n2: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t169_n168;
  t169_n3: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t169_n17;
  t169_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n16;
  t169_n5: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t169_n15;
  t169_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t169_n8;
  t169_n7: return 0.0003880017320625484f;
  t169_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t169_n14;
  t169_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t169_n11;
  t169_n10: return -0.11506088823080063f;
  t169_n11: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t169_n13;
  t169_n12: return 0.023554252460598946f;
  t169_n13: return 0.07260299474000931f;
  t169_n14: return 0.03864649310708046f;
  t169_n15: return -0.0037627536803483963f;
  t169_n16: return 0.0008221819880418479f;
  t169_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t169_n159;
  t169_n18: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t169_n24;
  t169_n19: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t169_n23;
  t169_n20: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t169_n22;
  t169_n21: return 0.09675705432891846f;
  t169_n22: return 0.0039424775168299675f;
  t169_n23: return 0.11195548623800278f;
  t169_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n36;
  t169_n25: if (E.getNumReferences() >= 3366869888 /*357084.0*/) goto t169_n35;
  t169_n26: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t169_n34;
  t169_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t169_n29;
  t169_n28: return -0.010601954534649849f;
  t169_n29: if (E.getNumReferences() >= 3315876864 /*5256.5*/) goto t169_n31;
  t169_n30: return 0.009481553919613361f;
  t169_n31: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t169_n33;
  t169_n32: return 0.12843111157417297f;
  t169_n33: return 0.06402251869440079f;
  t169_n34: return -0.03905877843499184f;
  t169_n35: return 0.09778138250112534f;
  t169_n36: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t169_n132;
  t169_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t169_n131;
  t169_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t169_n44;
  t169_n39: if (E.getNumReferences() >= 3339230592 /*34973.5*/) goto t169_n41;
  t169_n40: return -0.02844611555337906f;
  t169_n41: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t169_n43;
  t169_n42: return -0.3128858208656311f;
  t169_n43: return -0.07743780314922333f;
  t169_n44: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t169_n68;
  t169_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t169_n67;
  t169_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n56;
  t169_n47: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t169_n55;
  t169_n48: if (E.getNumReferences() >= 3310442496 /*3349.5*/) goto t169_n54;
  t169_n49: if (E.getNumReferences() >= 3309752320 /*3181.0*/) goto t169_n51;
  t169_n50: return -0.007519051898270845f;
  t169_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n53;
  t169_n52: return -0.1428452879190445f;
  t169_n53: return 0.12238451093435287f;
  t169_n54: return -0.08703386783599854f;
  t169_n55: return 0.03868988901376724f;
  t169_n56: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t169_n58;
  t169_n57: return -0.0011120563140138984f;
  t169_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t169_n60;
  t169_n59: return 0.013227731920778751f;
  t169_n60: if (E.getNumReferences() >= 3265462272 /*81.5*/) goto t169_n62;
  t169_n61: return 0.04760739952325821f;
  t169_n62: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t169_n64;
  t169_n63: return 0.38572537899017334f;
  t169_n64: if (E.getNumReferences() >= 3266379776 /*88.5*/) goto t169_n66;
  t169_n65: return 0.11632321029901505f;
  t169_n66: return 0.05057952180504799f;
  t169_n67: return 0.14275099337100983f;
  t169_n68: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t169_n94;
  t169_n69: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t169_n89;
  t169_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n88;
  t169_n71: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t169_n85;
  t169_n72: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t169_n84;
  t169_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t169_n75;
  t169_n74: return -0.3201814293861389f;
  t169_n75: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t169_n79;
  t169_n76: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t169_n78;
  t169_n77: return -0.009128941223025322f;
  t169_n78: return 0.11357089877128601f;
  t169_n79: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t169_n81;
  t169_n80: return 0.1213240846991539f;
  t169_n81: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t169_n83;
  t169_n82: return -0.11590030789375305f;
  t169_n83: return 0.09710223972797394f;
  t169_n84: return -0.3430927097797394f;
  t169_n85: if (E.getNumReferences() >= 3274375168 /*171.0*/) goto t169_n87;
  t169_n86: return -0.15126024186611176f;
  t169_n87: return -0.040251292288303375f;
  t169_n88: return 0.010135028511285782f;
  t169_n89: if (E.getNumReferences() >= 3315605504 /*5124.0*/) goto t169_n91;
  t169_n90: return 0.026526080444455147f;
  t169_n91: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t169_n93;
  t169_n92: return -0.01487996056675911f;
  t169_n93: return 0.09095362573862076f;
  t169_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t169_n96;
  t169_n95: return -0.016020307317376137f;
  t169_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t169_n98;
  t169_n97: return -0.0021081557497382164f;
  t169_n98: if (E.getNumReferences() >= 3267428352 /*96.5*/) goto t169_n106;
  t169_n99: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t169_n101;
  t169_n100: return 0.015842238441109657f;
  t169_n101: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t169_n103;
  t169_n102: return -0.1701178401708603f;
  t169_n103: if (E.getNumReferences() >= 3256352768 /*38.0*/) goto t169_n105;
  t169_n104: return 0.6117788553237915f;
  t169_n105: return 0.11926278471946716f;
  t169_n106: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t169_n122;
  t169_n107: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t169_n109;
  t169_n108: return -0.017576750367879868f;
  t169_n109: if (E.getNumReferences() >= 3331159552 /*18107.0*/) goto t169_n111;
  t169_n110: return 0.07513030618429184f;
  t169_n111: if (E.getNumReferences() >= 3332281088 /*20297.5*/) goto t169_n117;
  t169_n112: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t169_n116;
  t169_n113: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t169_n115;
  t169_n114: return 0.16852180659770966f;
  t169_n115: return 0.09701929241418839f;
  t169_n116: return 0.07961179316043854f;
  t169_n117: if (E.getNumReferences() >= 3349332224 /*83330.0*/) goto t169_n119;
  t169_n118: return -0.16185490787029266f;
  t169_n119: if (E.getNumReferences() >= 3356586400 /*148934.5*/) goto t169_n121;
  t169_n120: return 0.1112024113535881f;
  t169_n121: return 0.03427521139383316f;
  t169_n122: if (E.getNumReferences() >= 3272179712 /*137.5*/) goto t169_n126;
  t169_n123: if (E.getNumReferences() >= 3270049792 /*116.5*/) goto t169_n125;
  t169_n124: return 0.0825294777750969f;
  t169_n125: return 0.39032793045043945f;
  t169_n126: if (E.getNumReferences() >= 3298713600 /*1267.0*/) goto t169_n130;
  t169_n127: if (E.getNumReferences() >= 3285090304 /*413.0*/) goto t169_n129;
  t169_n128: return 0.09355554729700089f;
  t169_n129: return 0.20247018337249756f;
  t169_n130: return 0.03014937974512577f;
  t169_n131: return 0.009117284789681435f;
  t169_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t169_n134;
  t169_n133: return -0.004993515554815531f;
  t169_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t169_n148;
  t169_n135: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t169_n147;
  t169_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n138;
  t169_n137: return -0.011696899309754372f;
  t169_n138: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t169_n142;
  t169_n139: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t169_n141;
  t169_n140: return 0.33512747287750244f;
  t169_n141: return 0.06360994279384613f;
  t169_n142: if (E.getNumReferences() >= 3284336640 /*390.0*/) goto t169_n144;
  t169_n143: return 0.46541666984558105f;
  t169_n144: if (E.getNumReferences() >= 3325082624 /*11311.0*/) goto t169_n146;
  t169_n145: return -0.34166550636291504f;
  t169_n146: return 0.18790341913700104f;
  t169_n147: return 0.03452778235077858f;
  t169_n148: if (E.getNumReferences() >= 3286859776 /*467.0*/) goto t169_n158;
  t169_n149: if (E.getFractionNameInContext() >= 3179644278 /*0.0651959627866745*/) goto t169_n157;
  t169_n150: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t169_n156;
  t169_n151: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t169_n155;
  t169_n152: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t169_n154;
  t169_n153: return 0.22512315213680267f;
  t169_n154: return 0.17015622556209564f;
  t169_n155: return 0.07274753600358963f;
  t169_n156: return 0.2221590280532837f;
  t169_n157: return 0.05494113266468048f;
  t169_n158: return -0.5395686030387878f;
  t169_n159: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t169_n167;
  t169_n160: if (E.getNumReferences() >= 3341720320 /*44699.0*/) goto t169_n166;
  t169_n161: if (E.getNumReferences() >= 3341362816 /*43302.5*/) goto t169_n165;
  t169_n162: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t169_n164;
  t169_n163: return 0.13126812875270844f;
  t169_n164: return -0.01570022851228714f;
  t169_n165: return 0.12055278569459915f;
  t169_n166: return -0.07461482286453247f;
  t169_n167: return 0.0030904519371688366f;
  t169_n168: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t169_n238;
  t169_n169: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t169_n223;
  t169_n170: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t169_n206;
  t169_n171: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t169_n197;
  t169_n172: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t169_n174;
  t169_n173: return -0.0015932375099509954f;
  t169_n174: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t169_n192;
  t169_n175: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t169_n191;
  t169_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t169_n184;
  t169_n177: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t169_n183;
  t169_n178: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t169_n180;
  t169_n179: return 0.05181571841239929f;
  t169_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t169_n182;
  t169_n181: return -0.1043664738535881f;
  t169_n182: return 0.1289389580488205f;
  t169_n183: return -0.04709141328930855f;
  t169_n184: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t169_n188;
  t169_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t169_n187;
  t169_n186: return 0.020327530801296234f;
  t169_n187: return 0.11985870450735092f;
  t169_n188: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t169_n190;
  t169_n189: return 0.12263194471597672f;
  t169_n190: return -0.11195763200521469f;
  t169_n191: return -0.013346182182431221f;
  t169_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t169_n194;
  t169_n193: return -0.18880046904087067f;
  t169_n194: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t169_n196;
  t169_n195: return 0.06808136403560638f;
  t169_n196: return -0.016961224377155304f;
  t169_n197: if (E.getNumReferences() >= 3361426240 /*224557.0*/) goto t169_n199;
  t169_n198: return 0.02798936888575554f;
  t169_n199: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t169_n205;
  t169_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n204;
  t169_n201: if (E.getSemaFileProximityScore() >= 3204448256 /*0.5*/) goto t169_n203;
  t169_n202: return -0.16707926988601685f;
  t169_n203: return -0.5877678990364075f;
  t169_n204: return 0.05866231024265289f;
  t169_n205: return 0.016047213226556778f;
  t169_n206: if (E.getNumReferences() >= 3363672576 /*259656.0*/) goto t169_n210;
  t169_n207: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t169_n209;
  t169_n208: return 0.23682549595832825f;
  t169_n209: return -0.1427675038576126f;
  t169_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t169_n212;
  t169_n211: return -0.05351985618472099f;
  t169_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t169_n214;
  t169_n213: return -0.033136751502752304f;
  t169_n214: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t169_n218;
  t169_n215: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t169_n217;
  t169_n216: return 0.10967852920293808f;
  t169_n217: return 0.13408291339874268f;
  t169_n218: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t169_n220;
  t169_n219: return -0.21466846764087677f;
  t169_n220: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t169_n222;
  t169_n221: return 0.08261919021606445f;
  t169_n222: return 0.06358624994754791f;
  t169_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t169_n227;
  t169_n224: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t169_n226;
  t169_n225: return 0.0337834507226944f;
  t169_n226: return -0.04327010363340378f;
  t169_n227: if (E.getFractionNameInContext() >= 3177875683 /*0.057259451597929*/) goto t169_n237;
  t169_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t169_n230;
  t169_n229: return 0.01802166923880577f;
  t169_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t169_n236;
  t169_n231: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t169_n235;
  t169_n232: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t169_n234;
  t169_n233: return -0.024664245545864105f;
  t169_n234: return 0.12020498514175415f;
  t169_n235: return 0.036793071776628494f;
  t169_n236: return 0.08051632344722748f;
  t169_n237: return 0.002327092457562685f;
  t169_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t169_n240;
  t169_n239: return -0.0286415945738554f;
  t169_n240: return 0.02690838649868965f;
  t169_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t169_n245;
  t169_n242: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t169_n244;
  t169_n243: return 0.06757804751396179f;
  t169_n244: return -0.16557542979717255f;
  t169_n245: return 0.020105797797441483f;
  t169_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t169_n248;
  t169_n247: return 0.001671666163019836f;
  t169_n248: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t169_n250;
  t169_n249: return 0.004713143687695265f;
  t169_n250: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t169_n254;
  t169_n251: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t169_n253;
  t169_n252: return -0.1724746823310852f;
  t169_n253: return 0.10449246317148209f;
  t169_n254: return -0.1548391431570053f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree170(const Example& E) {
  t170_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t170_n234;
  t170_n1: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t170_n223;
  t170_n2: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t170_n222;
  t170_n3: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t170_n199;
  t170_n4: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t170_n198;
  t170_n5: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t170_n155;
  t170_n6: if (E.getNumReferences() >= 3350448512 /*92051.0*/) goto t170_n138;
  t170_n7: if (E.getNumReferences() >= 3350016768 /*88678.0*/) goto t170_n135;
  t170_n8: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t170_n92;
  t170_n9: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t170_n69;
  t170_n10: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t170_n48;
  t170_n11: if (E.getNumReferences() >= 3341141504 /*42438.0*/) goto t170_n13;
  t170_n12: return -0.00025789858773350716f;
  t170_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t170_n15;
  t170_n14: return -0.032772570848464966f;
  t170_n15: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t170_n25;
  t170_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n18;
  t170_n17: return 0.04583512246608734f;
  t170_n18: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t170_n20;
  t170_n19: return 0.11737606674432755f;
  t170_n20: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t170_n24;
  t170_n21: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t170_n23;
  t170_n22: return 0.11940700560808182f;
  t170_n23: return -0.10929617285728455f;
  t170_n24: return 0.015077335759997368f;
  t170_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t170_n31;
  t170_n26: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t170_n28;
  t170_n27: return -0.03383323922753334f;
  t170_n28: if (E.getNumReferences() >= 3341955328 /*45617.0*/) goto t170_n30;
  t170_n29: return -0.44468069076538086f;
  t170_n30: return -0.06754403561353683f;
  t170_n31: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t170_n33;
  t170_n32: return -0.1865263134241104f;
  t170_n33: if (E.getNumReferences() >= 3341956992 /*45623.5*/) goto t170_n45;
  t170_n34: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n36;
  t170_n35: return 0.06681079417467117f;
  t170_n36: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t170_n38;
  t170_n37: return -0.11797633767127991f;
  t170_n38: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t170_n44;
  t170_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t170_n41;
  t170_n40: return 0.06568653881549835f;
  t170_n41: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t170_n43;
  t170_n42: return 0.14250628650188446f;
  t170_n43: return 0.11317802220582962f;
  t170_n44: return 0.12047077715396881f;
  t170_n45: if (E.getNumReferences() >= 3342343040 /*47131.5*/) goto t170_n47;
  t170_n46: return 0.0280819870531559f;
  t170_n47: return 0.0634288415312767f;
  t170_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t170_n60;
  t170_n49: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t170_n53;
  t170_n50: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t170_n52;
  t170_n51: return -0.08401543647050858f;
  t170_n52: return -0.1378137171268463f;
  t170_n53: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t170_n57;
  t170_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n56;
  t170_n55: return -0.20896604657173157f;
  t170_n56: return 0.08368027955293655f;
  t170_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n59;
  t170_n58: return -0.09338006377220154f;
  t170_n59: return 0.005164545029401779f;
  t170_n60: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n62;
  t170_n61: return -0.14009593427181244f;
  t170_n62: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t170_n64;
  t170_n63: return -0.2504730522632599f;
  t170_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t170_n66;
  t170_n65: return -0.1042235791683197f;
  t170_n66: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t170_n68;
  t170_n67: return 0.14864718914031982f;
  t170_n68: return 0.14256510138511658f;
  t170_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t170_n87;
  t170_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n74;
  t170_n71: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t170_n73;
  t170_n72: return -0.0666348785161972f;
  t170_n73: return -0.44866588711738586f;
  t170_n74: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n86;
  t170_n75: if (E.getNumReferences() >= 3344543232 /*55726.0*/) goto t170_n81;
  t170_n76: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t170_n80;
  t170_n77: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t170_n79;
  t170_n78: return 0.13790446519851685f;
  t170_n79: return 0.09472601860761642f;
  t170_n80: return -0.34104904532432556f;
  t170_n81: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t170_n83;
  t170_n82: return 0.0889262706041336f;
  t170_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n85;
  t170_n84: return -0.003600657917559147f;
  t170_n85: return 0.09046938270330429f;
  t170_n86: return -0.13453316688537598f;
  t170_n87: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t170_n91;
  t170_n88: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t170_n90;
  t170_n89: return 0.11576544493436813f;
  t170_n90: return 0.19369907677173615f;
  t170_n91: return -0.18403728306293488f;
  t170_n92: if (E.getNumReferences() >= 3344974336 /*57410.0*/) goto t170_n94;
  t170_n93: return -0.15447503328323364f;
  t170_n94: if (E.getNumReferences() >= 3345503360 /*59476.5*/) goto t170_n110;
  t170_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t170_n101;
  t170_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n98;
  t170_n97: return -0.12611457705497742f;
  t170_n98: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n100;
  t170_n99: return 0.08881248533725739f;
  t170_n100: return -0.17366725206375122f;
  t170_n101: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t170_n107;
  t170_n102: if (E.getNumReferences() >= 3345144064 /*58073.0*/) goto t170_n106;
  t170_n103: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t170_n105;
  t170_n104: return 0.12087009847164154f;
  t170_n105: return 0.06758396327495575f;
  t170_n106: return -0.015327426604926586f;
  t170_n107: if (E.getNumReferences() >= 3345041664 /*57673.0*/) goto t170_n109;
  t170_n108: return -0.9590389132499695f;
  t170_n109: return 0.06758628785610199f;
  t170_n110: if (E.getNumReferences() >= 3345933184 /*61155.5*/) goto t170_n124;
  t170_n111: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n123;
  t170_n112: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t170_n114;
  t170_n113: return -0.1706763207912445f;
  t170_n114: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t170_n122;
  t170_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n121;
  t170_n116: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t170_n120;
  t170_n117: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t170_n119;
  t170_n118: return -0.8834623694419861f;
  t170_n119: return -0.24588775634765625f;
  t170_n120: return -0.2079637050628662f;
  t170_n121: return -0.18249528110027313f;
  t170_n122: return 0.1039172112941742f;
  t170_n123: return 0.04217028245329857f;
  t170_n124: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t170_n130;
  t170_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t170_n129;
  t170_n126: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t170_n128;
  t170_n127: return -0.0573328398168087f;
  t170_n128: return -0.005959104746580124f;
  t170_n129: return 0.042646750807762146f;
  t170_n130: if (E.getNumReferences() >= 3349802816 /*87006.5*/) goto t170_n134;
  t170_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t170_n133;
  t170_n132: return -0.0011603570310398936f;
  t170_n133: return 0.05282732471823692f;
  t170_n134: return 0.11426528543233871f;
  t170_n135: if (E.getNumReferences() >= 3350198272 /*90096.0*/) goto t170_n137;
  t170_n136: return 0.04789954796433449f;
  t170_n137: return -0.004293324425816536f;
  t170_n138: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t170_n150;
  t170_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n141;
  t170_n140: return -0.03748243674635887f;
  t170_n141: if (E.getNumReferences() >= 3350654080 /*93657.0*/) goto t170_n143;
  t170_n142: return 0.005702515598386526f;
  t170_n143: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t170_n147;
  t170_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n146;
  t170_n145: return 0.10217230021953583f;
  t170_n146: return 0.20210714638233185f;
  t170_n147: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t170_n149;
  t170_n148: return 0.07396252453327179f;
  t170_n149: return -0.5465219020843506f;
  t170_n150: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t170_n154;
  t170_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n153;
  t170_n152: return -0.13374216854572296f;
  t170_n153: return -0.15045997500419617f;
  t170_n154: return -0.2723238170146942f;
  t170_n155: if (E.getNumReferences() >= 3351669952 /*101593.5*/) goto t170_n173;
  t170_n156: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t170_n170;
  t170_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t170_n161;
  t170_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t170_n160;
  t170_n159: return -0.1298128068447113f;
  t170_n160: return 0.05234197527170181f;
  t170_n161: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t170_n163;
  t170_n162: return 0.01750778779387474f;
  t170_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t170_n165;
  t170_n164: return 0.0780504122376442f;
  t170_n165: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t170_n167;
  t170_n166: return 0.09709756821393967f;
  t170_n167: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t170_n169;
  t170_n168: return -0.007679884321987629f;
  t170_n169: return 0.10053142160177231f;
  t170_n170: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t170_n172;
  t170_n171: return 0.08886713534593582f;
  t170_n172: return -0.06973635405302048f;
  t170_n173: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t170_n183;
  t170_n174: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t170_n180;
  t170_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t170_n177;
  t170_n176: return -0.0037300819531083107f;
  t170_n177: if (E.getNumReferences() >= 3356105632 /*141422.5*/) goto t170_n179;
  t170_n178: return 0.0509081706404686f;
  t170_n179: return -0.09034676849842072f;
  t170_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t170_n182;
  t170_n181: return 0.022366318851709366f;
  t170_n182: return 0.054075345396995544f;
  t170_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n185;
  t170_n184: return -0.11419269442558289f;
  t170_n185: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t170_n197;
  t170_n186: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t170_n194;
  t170_n187: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t170_n193;
  t170_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t170_n192;
  t170_n189: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t170_n191;
  t170_n190: return 0.09947433322668076f;
  t170_n191: return -0.029114559292793274f;
  t170_n192: return -0.09609373658895493f;
  t170_n193: return 0.12644672393798828f;
  t170_n194: if (E.getNumReferences() >= 3357156000 /*157834.5*/) goto t170_n196;
  t170_n195: return -0.08264260739088058f;
  t170_n196: return 0.26984649896621704f;
  t170_n197: return -0.12760749459266663f;
  t170_n198: return -0.14675436913967133f;
  t170_n199: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t170_n207;
  t170_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n202;
  t170_n201: return -0.017601795494556427f;
  t170_n202: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t170_n206;
  t170_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t170_n205;
  t170_n204: return -0.2281647026538849f;
  t170_n205: return 0.0883350595831871f;
  t170_n206: return 0.041373006999492645f;
  t170_n207: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t170_n209;
  t170_n208: return -0.003305550431832671f;
  t170_n209: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t170_n219;
  t170_n210: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t170_n216;
  t170_n211: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t170_n215;
  t170_n212: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t170_n214;
  t170_n213: return 0.16831807792186737f;
  t170_n214: return 0.08954138308763504f;
  t170_n215: return 0.029313594102859497f;
  t170_n216: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t170_n218;
  t170_n217: return -0.11217544972896576f;
  t170_n218: return 0.0849418044090271f;
  t170_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t170_n221;
  t170_n220: return -0.022917436435818672f;
  t170_n221: return 0.08453966677188873f;
  t170_n222: return -0.05278874561190605f;
  t170_n223: if (E.getNumReferences() >= 3364286400 /*276350.0*/) goto t170_n233;
  t170_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t170_n230;
  t170_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t170_n227;
  t170_n226: return -0.008872113190591335f;
  t170_n227: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t170_n229;
  t170_n228: return 0.1471727192401886f;
  t170_n229: return 0.14698158204555511f;
  t170_n230: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t170_n232;
  t170_n231: return 0.06393841654062271f;
  t170_n232: return 0.1580905169248581f;
  t170_n233: return 0.007989158853888512f;
  t170_n234: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t170_n252;
  t170_n235: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t170_n249;
  t170_n236: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t170_n238;
  t170_n237: return 0.010352497920393944f;
  t170_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t170_n240;
  t170_n239: return -0.12919044494628906f;
  t170_n240: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t170_n242;
  t170_n241: return 0.00185781205072999f;
  t170_n242: if (E.getNumReferences() >= 3352414592 /*107411.0*/) goto t170_n244;
  t170_n243: return 0.07388949394226074f;
  t170_n244: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t170_n248;
  t170_n245: if (E.getNumReferences() >= 3354001600 /*119809.5*/) goto t170_n247;
  t170_n246: return -0.12631435692310333f;
  t170_n247: return -0.6211646199226379f;
  t170_n248: return 0.12335284799337387f;
  t170_n249: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t170_n251;
  t170_n250: return -0.24916322529315948f;
  t170_n251: return -0.686982274055481f;
  t170_n252: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t170_n254;
  t170_n253: return 0.08195562660694122f;
  t170_n254: return 0.042762015014886856f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree171(const Example& E) {
  t171_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t171_n2;
  t171_n1: return -0.00023543900169897825f;
  t171_n2: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t171_n56;
  t171_n3: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t171_n27;
  t171_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n22;
  t171_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n7;
  t171_n6: return -0.02917776070535183f;
  t171_n7: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t171_n19;
  t171_n8: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t171_n10;
  t171_n9: return -0.0008875893545337021f;
  t171_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t171_n14;
  t171_n11: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t171_n13;
  t171_n12: return -0.01968207210302353f;
  t171_n13: return -0.05206817761063576f;
  t171_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n18;
  t171_n15: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t171_n17;
  t171_n16: return -0.06379786878824234f;
  t171_n17: return 0.023644840344786644f;
  t171_n18: return 0.030047481879591942f;
  t171_n19: if (E.getNumReferences() >= 3308226560 /*2808.5*/) goto t171_n21;
  t171_n20: return -0.028016062453389168f;
  t171_n21: return 0.11141648888587952f;
  t171_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t171_n24;
  t171_n23: return -0.02294413559138775f;
  t171_n24: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t171_n26;
  t171_n25: return 0.026283755898475647f;
  t171_n26: return 0.11699432879686356f;
  t171_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n47;
  t171_n28: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t171_n32;
  t171_n29: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t171_n31;
  t171_n30: return 0.08376150578260422f;
  t171_n31: return -0.019554898142814636f;
  t171_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t171_n46;
  t171_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t171_n37;
  t171_n34: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t171_n36;
  t171_n35: return 0.020844224840402603f;
  t171_n36: return 0.1067313477396965f;
  t171_n37: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t171_n43;
  t171_n38: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t171_n42;
  t171_n39: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t171_n41;
  t171_n40: return -0.1261087954044342f;
  t171_n41: return 0.13367019593715668f;
  t171_n42: return -0.09946663677692413f;
  t171_n43: if (E.getFractionNameInContext() >= 3199175417 /*0.34285715222358704*/) goto t171_n45;
  t171_n44: return 0.11867813766002655f;
  t171_n45: return 0.10158222168684006f;
  t171_n46: return -0.031065762042999268f;
  t171_n47: if (E.getFractionNameInContext() >= 3197399006 /*0.2899159789085388*/) goto t171_n55;
  t171_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t171_n50;
  t171_n49: return -0.00760948471724987f;
  t171_n50: if (E.getNumReferences() >= 3263561728 /*67.0*/) goto t171_n54;
  t171_n51: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t171_n53;
  t171_n52: return 0.02415584959089756f;
  t171_n53: return 0.11803723871707916f;
  t171_n54: return 0.12723077833652496f;
  t171_n55: return 0.01941792294383049f;
  t171_n56: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t171_n222;
  t171_n57: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t171_n217;
  t171_n58: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t171_n144;
  t171_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t171_n95;
  t171_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t171_n78;
  t171_n61: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t171_n69;
  t171_n62: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t171_n68;
  t171_n63: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t171_n67;
  t171_n64: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t171_n66;
  t171_n65: return -0.1423199623823166f;
  t171_n66: return -1.9753613471984863f;
  t171_n67: return 0.0159112811088562f;
  t171_n68: return -0.04848131164908409f;
  t171_n69: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t171_n77;
  t171_n70: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t171_n74;
  t171_n71: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t171_n73;
  t171_n72: return 0.21692292392253876f;
  t171_n73: return 0.14028723537921906f;
  t171_n74: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t171_n76;
  t171_n75: return -0.15394020080566406f;
  t171_n76: return 0.10964030772447586f;
  t171_n77: return -0.05505024641752243f;
  t171_n78: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t171_n82;
  t171_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n81;
  t171_n80: return -0.08088815957307816f;
  t171_n81: return -0.0017625265754759312f;
  t171_n82: if (E.getNumReferences() >= 3308613632 /*2903.0*/) goto t171_n94;
  t171_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t171_n89;
  t171_n84: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t171_n88;
  t171_n85: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t171_n87;
  t171_n86: return 0.06792492419481277f;
  t171_n87: return 0.13439059257507324f;
  t171_n88: return 0.003715021535754204f;
  t171_n89: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t171_n93;
  t171_n90: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t171_n92;
  t171_n91: return 0.17495658993721008f;
  t171_n92: return 0.054823022335767746f;
  t171_n93: return 0.0187400933355093f;
  t171_n94: return -0.11428085714578629f;
  t171_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t171_n143;
  t171_n96: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t171_n98;
  t171_n97: return 0.2493181675672531f;
  t171_n98: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t171_n100;
  t171_n99: return 0.01843033730983734f;
  t171_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t171_n102;
  t171_n101: return 0.009188038296997547f;
  t171_n102: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t171_n106;
  t171_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t171_n105;
  t171_n104: return 0.03440793603658676f;
  t171_n105: return 0.12686727941036224f;
  t171_n106: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t171_n136;
  t171_n107: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t171_n111;
  t171_n108: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t171_n110;
  t171_n109: return 0.1577555239200592f;
  t171_n110: return 0.05402425304055214f;
  t171_n111: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t171_n115;
  t171_n112: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t171_n114;
  t171_n113: return -0.11646389961242676f;
  t171_n114: return 0.10030192881822586f;
  t171_n115: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t171_n119;
  t171_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t171_n118;
  t171_n117: return 0.1394059658050537f;
  t171_n118: return 0.10815604031085968f;
  t171_n119: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t171_n133;
  t171_n120: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t171_n128;
  t171_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t171_n123;
  t171_n122: return -0.012769154272973537f;
  t171_n123: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t171_n127;
  t171_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n126;
  t171_n125: return -0.100754514336586f;
  t171_n126: return 0.1930171698331833f;
  t171_n127: return -0.11109907180070877f;
  t171_n128: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t171_n130;
  t171_n129: return 0.17717880010604858f;
  t171_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t171_n132;
  t171_n131: return -0.041285037994384766f;
  t171_n132: return 0.635258674621582f;
  t171_n133: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t171_n135;
  t171_n134: return -0.3626664876937866f;
  t171_n135: return 0.07224049419164658f;
  t171_n136: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t171_n138;
  t171_n137: return 0.008782957680523396f;
  t171_n138: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t171_n140;
  t171_n139: return -0.14110101759433746f;
  t171_n140: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t171_n142;
  t171_n141: return -0.10185319185256958f;
  t171_n142: return 0.6094319224357605f;
  t171_n143: return 0.001984178088605404f;
  t171_n144: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t171_n216;
  t171_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n147;
  t171_n146: return -0.037530843168497086f;
  t171_n147: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t171_n199;
  t171_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t171_n166;
  t171_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t171_n165;
  t171_n150: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t171_n156;
  t171_n151: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t171_n153;
  t171_n152: return 0.07351186126470566f;
  t171_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n155;
  t171_n154: return -0.11212236434221268f;
  t171_n155: return 0.07420738786458969f;
  t171_n156: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t171_n160;
  t171_n157: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t171_n159;
  t171_n158: return 0.020512670278549194f;
  t171_n159: return -0.2259305864572525f;
  t171_n160: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t171_n162;
  t171_n161: return 0.0421726256608963f;
  t171_n162: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t171_n164;
  t171_n163: return 0.12259607017040253f;
  t171_n164: return -0.07115583121776581f;
  t171_n165: return -0.08020734786987305f;
  t171_n166: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t171_n178;
  t171_n167: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t171_n175;
  t171_n168: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t171_n172;
  t171_n169: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t171_n171;
  t171_n170: return -0.07240293174982071f;
  t171_n171: return 0.04977016523480415f;
  t171_n172: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t171_n174;
  t171_n173: return 0.021999020129442215f;
  t171_n174: return -0.11299125105142593f;
  t171_n175: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t171_n177;
  t171_n176: return 0.0466352179646492f;
  t171_n177: return 0.12275102734565735f;
  t171_n178: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t171_n180;
  t171_n179: return 0.09525593370199203f;
  t171_n180: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t171_n198;
  t171_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t171_n195;
  t171_n182: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t171_n192;
  t171_n183: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t171_n185;
  t171_n184: return -0.0872269794344902f;
  t171_n185: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t171_n187;
  t171_n186: return 0.027943428605794907f;
  t171_n187: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t171_n189;
  t171_n188: return 0.03772507235407829f;
  t171_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t171_n191;
  t171_n190: return 0.1696886271238327f;
  t171_n191: return 0.13347645103931427f;
  t171_n192: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t171_n194;
  t171_n193: return 0.07386458665132523f;
  t171_n194: return 0.11560514569282532f;
  t171_n195: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t171_n197;
  t171_n196: return 0.011921885423362255f;
  t171_n197: return 0.12108666449785233f;
  t171_n198: return 0.012890866957604885f;
  t171_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t171_n211;
  t171_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t171_n202;
  t171_n201: return -0.08577315509319305f;
  t171_n202: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t171_n204;
  t171_n203: return -0.0013827894581481814f;
  t171_n204: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t171_n206;
  t171_n205: return -0.03183579444885254f;
  t171_n206: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t171_n208;
  t171_n207: return 0.06208306923508644f;
  t171_n208: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t171_n210;
  t171_n209: return 0.12136168032884598f;
  t171_n210: return -0.019660381600260735f;
  t171_n211: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t171_n215;
  t171_n212: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t171_n214;
  t171_n213: return 0.0404927134513855f;
  t171_n214: return -0.16845348477363586f;
  t171_n215: return 0.10364461690187454f;
  t171_n216: return -0.024547211825847626f;
  t171_n217: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t171_n219;
  t171_n218: return 0.13398650288581848f;
  t171_n219: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t171_n221;
  t171_n220: return -0.1799926906824112f;
  t171_n221: return 0.09210403263568878f;
  t171_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Variable))) goto t171_n254;
  t171_n223: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t171_n253;
  t171_n224: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t171_n228;
  t171_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t171_n227;
  t171_n226: return -0.061479922384023666f;
  t171_n227: return -0.006430324167013168f;
  t171_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t171_n244;
  t171_n229: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t171_n231;
  t171_n230: return 0.003518004436045885f;
  t171_n231: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t171_n243;
  t171_n232: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t171_n234;
  t171_n233: return 0.0045126816257834435f;
  t171_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t171_n236;
  t171_n235: return 0.049110911786556244f;
  t171_n236: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t171_n238;
  t171_n237: return -0.11728993058204651f;
  t171_n238: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t171_n242;
  t171_n239: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t171_n241;
  t171_n240: return 0.1474202275276184f;
  t171_n241: return 0.16026411950588226f;
  t171_n242: return -0.06124419718980789f;
  t171_n243: return 0.12700162827968597f;
  t171_n244: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t171_n252;
  t171_n245: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t171_n247;
  t171_n246: return 0.03440210595726967f;
  t171_n247: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t171_n251;
  t171_n248: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t171_n250;
  t171_n249: return 0.13235364854335785f;
  t171_n250: return 0.37690240144729614f;
  t171_n251: return 0.0715414360165596f;
  t171_n252: return 0.014588505029678345f;
  t171_n253: return -0.0017169448547065258f;
  t171_n254: return 0.00480049941688776f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree172(const Example& E) {
  t172_n0: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t172_n186;
  t172_n1: if (E.getNumReferences() >= 3289325568 /*572.5*/) goto t172_n185;
  t172_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t172_n4;
  t172_n3: return -0.0001344587217317894f;
  t172_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t172_n168;
  t172_n5: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t172_n49;
  t172_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t172_n18;
  t172_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n9;
  t172_n8: return -0.15755771100521088f;
  t172_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t172_n17;
  t172_n10: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t172_n14;
  t172_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t172_n13;
  t172_n12: return 0.05944523960351944f;
  t172_n13: return 0.09136252850294113f;
  t172_n14: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t172_n16;
  t172_n15: return -0.06931321322917938f;
  t172_n16: return 0.03956470638513565f;
  t172_n17: return -0.23523584008216858f;
  t172_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n26;
  t172_n19: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t172_n23;
  t172_n20: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t172_n22;
  t172_n21: return -0.10664734244346619f;
  t172_n22: return -0.03520607575774193f;
  t172_n23: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t172_n25;
  t172_n24: return 0.06609175354242325f;
  t172_n25: return -0.29698508977890015f;
  t172_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t172_n28;
  t172_n27: return -0.016630331054329872f;
  t172_n28: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t172_n30;
  t172_n29: return -0.036483753472566605f;
  t172_n30: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t172_n32;
  t172_n31: return 0.007403599098324776f;
  t172_n32: if (E.getFractionNameInContext() >= 3198656122 /*0.3273809552192688*/) goto t172_n34;
  t172_n33: return -0.20278124511241913f;
  t172_n34: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t172_n48;
  t172_n35: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t172_n43;
  t172_n36: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t172_n42;
  t172_n37: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t172_n39;
  t172_n38: return 0.09031543135643005f;
  t172_n39: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t172_n41;
  t172_n40: return 0.13841785490512848f;
  t172_n41: return 0.1274796724319458f;
  t172_n42: return -0.06337094306945801f;
  t172_n43: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t172_n47;
  t172_n44: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t172_n46;
  t172_n45: return 0.08637482672929764f;
  t172_n46: return 0.13964195549488068f;
  t172_n47: return 0.1589689999818802f;
  t172_n48: return 0.0009047677740454674f;
  t172_n49: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t172_n97;
  t172_n50: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t172_n52;
  t172_n51: return 0.0015726088313385844f;
  t172_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n58;
  t172_n53: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t172_n57;
  t172_n54: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t172_n56;
  t172_n55: return -0.024386078119277954f;
  t172_n56: return 0.05436760559678078f;
  t172_n57: return -0.12230388075113297f;
  t172_n58: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t172_n62;
  t172_n59: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t172_n61;
  t172_n60: return 0.03502603620290756f;
  t172_n61: return -0.12015175074338913f;
  t172_n62: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t172_n84;
  t172_n63: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t172_n79;
  t172_n64: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t172_n70;
  t172_n65: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t172_n67;
  t172_n66: return 0.004730757791548967f;
  t172_n67: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t172_n69;
  t172_n68: return 0.12156404554843903f;
  t172_n69: return -0.06425700336694717f;
  t172_n70: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t172_n78;
  t172_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n75;
  t172_n72: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t172_n74;
  t172_n73: return 0.13831348717212677f;
  t172_n74: return -0.07102663815021515f;
  t172_n75: if (E.getFractionNameInContext() >= 3193736649 /*0.2153846174478531*/) goto t172_n77;
  t172_n76: return 0.15397141873836517f;
  t172_n77: return 0.1794823259115219f;
  t172_n78: return 0.049844272434711456f;
  t172_n79: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t172_n81;
  t172_n80: return 0.10154091566801071f;
  t172_n81: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t172_n83;
  t172_n82: return 0.19425779581069946f;
  t172_n83: return -0.4371388256549835f;
  t172_n84: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t172_n86;
  t172_n85: return 0.008025349117815495f;
  t172_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t172_n92;
  t172_n87: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t172_n91;
  t172_n88: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t172_n90;
  t172_n89: return 0.048625729978084564f;
  t172_n90: return 0.12983573973178864f;
  t172_n91: return 0.002613233169540763f;
  t172_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n94;
  t172_n93: return -0.03137988597154617f;
  t172_n94: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t172_n96;
  t172_n95: return 0.14553427696228027f;
  t172_n96: return -0.11837832629680634f;
  t172_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n99;
  t172_n98: return -0.10725169628858566f;
  t172_n99: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t172_n167;
  t172_n100: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t172_n102;
  t172_n101: return 0.018306156620383263f;
  t172_n102: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t172_n110;
  t172_n103: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t172_n105;
  t172_n104: return 0.037132300436496735f;
  t172_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t172_n107;
  t172_n106: return -0.01071099005639553f;
  t172_n107: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t172_n109;
  t172_n108: return 0.15459363162517548f;
  t172_n109: return -0.007604220416396856f;
  t172_n110: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t172_n112;
  t172_n111: return -0.006971422117203474f;
  t172_n112: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t172_n138;
  t172_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n133;
  t172_n114: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t172_n116;
  t172_n115: return 0.008133349940180779f;
  t172_n116: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t172_n132;
  t172_n117: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t172_n129;
  t172_n118: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t172_n126;
  t172_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n121;
  t172_n120: return -0.16066418588161469f;
  t172_n121: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t172_n123;
  t172_n122: return 0.08903435617685318f;
  t172_n123: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t172_n125;
  t172_n124: return 0.17901955544948578f;
  t172_n125: return 0.1295304298400879f;
  t172_n126: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t172_n128;
  t172_n127: return -0.13549548387527466f;
  t172_n128: return 0.12687215209007263f;
  t172_n129: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t172_n131;
  t172_n130: return 0.11436537653207779f;
  t172_n131: return 0.1525953859090805f;
  t172_n132: return -0.11824687570333481f;
  t172_n133: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t172_n135;
  t172_n134: return -0.018571622669696808f;
  t172_n135: if (E.getNumReferences() >= 3266445312 /*89.0*/) goto t172_n137;
  t172_n136: return -0.2857263386249542f;
  t172_n137: return 0.1255621314048767f;
  t172_n138: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t172_n140;
  t172_n139: return 0.061354801058769226f;
  t172_n140: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t172_n142;
  t172_n141: return -0.07214939594268799f;
  t172_n142: if (E.getNumReferences() >= 3268214784 /*102.5*/) goto t172_n152;
  t172_n143: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t172_n147;
  t172_n144: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t172_n146;
  t172_n145: return 0.046224918216466904f;
  t172_n146: return 0.10464855283498764f;
  t172_n147: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t172_n149;
  t172_n148: return 0.12302698940038681f;
  t172_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n151;
  t172_n150: return 0.02693101577460766f;
  t172_n151: return 0.11406100541353226f;
  t172_n152: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t172_n166;
  t172_n153: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t172_n163;
  t172_n154: if (E.getNumReferences() >= 3274735616 /*176.5*/) goto t172_n162;
  t172_n155: if (E.getNumReferences() >= 3270115328 /*117.0*/) goto t172_n157;
  t172_n156: return -0.20810995995998383f;
  t172_n157: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t172_n159;
  t172_n158: return 0.1455736607313156f;
  t172_n159: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t172_n161;
  t172_n160: return 0.1857789307832718f;
  t172_n161: return -0.00495120557025075f;
  t172_n162: return -0.008305778726935387f;
  t172_n163: if (E.getNumReferences() >= 3273064448 /*151.0*/) goto t172_n165;
  t172_n164: return 0.016108030453324318f;
  t172_n165: return 0.14056231081485748f;
  t172_n166: return -0.07638148218393326f;
  t172_n167: return 0.11438813805580139f;
  t172_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t172_n184;
  t172_n169: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t172_n171;
  t172_n170: return 0.006461850833147764f;
  t172_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n173;
  t172_n172: return -0.006424418184906244f;
  t172_n173: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t172_n183;
  t172_n174: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t172_n180;
  t172_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n179;
  t172_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t172_n178;
  t172_n177: return 0.020835870876908302f;
  t172_n178: return 0.16122490167617798f;
  t172_n179: return -0.04586823657155037f;
  t172_n180: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t172_n182;
  t172_n181: return 0.1341940015554428f;
  t172_n182: return 0.04851152002811432f;
  t172_n183: return -0.15332196652889252f;
  t172_n184: return 0.06812125444412231f;
  t172_n185: return 0.010004013776779175f;
  t172_n186: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t172_n202;
  t172_n187: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t172_n189;
  t172_n188: return -0.001460812403820455f;
  t172_n189: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t172_n201;
  t172_n190: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t172_n192;
  t172_n191: return 0.0022746282629668713f;
  t172_n192: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t172_n200;
  t172_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t172_n199;
  t172_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t172_n196;
  t172_n195: return 0.04164956510066986f;
  t172_n196: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t172_n198;
  t172_n197: return 0.11337349563837051f;
  t172_n198: return 0.06359583139419556f;
  t172_n199: return 0.015966122969985008f;
  t172_n200: return 0.007739981636404991f;
  t172_n201: return -0.021099580451846123f;
  t172_n202: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t172_n204;
  t172_n203: return 0.14867909252643585f;
  t172_n204: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t172_n242;
  t172_n205: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t172_n241;
  t172_n206: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t172_n210;
  t172_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t172_n209;
  t172_n208: return -0.0149640878662467f;
  t172_n209: return 0.1063314825296402f;
  t172_n210: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t172_n222;
  t172_n211: if (E.getNumReferences() >= 3280551936 /*274.5*/) goto t172_n221;
  t172_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n220;
  t172_n213: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t172_n217;
  t172_n214: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t172_n216;
  t172_n215: return -0.10199904441833496f;
  t172_n216: return 0.004201027099043131f;
  t172_n217: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t172_n219;
  t172_n218: return 0.11400813609361649f;
  t172_n219: return -0.3171706795692444f;
  t172_n220: return 0.060277972370386124f;
  t172_n221: return 0.1428626924753189f;
  t172_n222: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t172_n240;
  t172_n223: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t172_n237;
  t172_n224: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t172_n232;
  t172_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n231;
  t172_n226: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t172_n228;
  t172_n227: return -0.2689448893070221f;
  t172_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t172_n230;
  t172_n229: return -0.3041895925998688f;
  t172_n230: return 0.09980588406324387f;
  t172_n231: return 0.048243798315525055f;
  t172_n232: if (E.getFractionNameInContext() >= 3202870911 /*0.45299145579338074*/) goto t172_n236;
  t172_n233: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t172_n235;
  t172_n234: return -0.1042928397655487f;
  t172_n235: return 0.1478797197341919f;
  t172_n236: return 0.010637234896421432f;
  t172_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n239;
  t172_n238: return -0.3273825943470001f;
  t172_n239: return -0.013657866977155209f;
  t172_n240: return 0.08325850963592529f;
  t172_n241: return -0.013567346148192883f;
  t172_n242: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t172_n252;
  t172_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t172_n251;
  t172_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t172_n250;
  t172_n245: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t172_n249;
  t172_n246: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t172_n248;
  t172_n247: return -0.1277683973312378f;
  t172_n248: return -0.02354576624929905f;
  t172_n249: return 0.07830731570720673f;
  t172_n250: return 0.05570063367486f;
  t172_n251: return 0.021871337667107582f;
  t172_n252: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t172_n254;
  t172_n253: return 0.14550505578517914f;
  t172_n254: return 0.034662190824747086f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree173(const Example& E) {
  t173_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t173_n32;
  t173_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t173_n31;
  t173_n2: if (E.getNumReferences() >= 3288801280 /*540.5*/) goto t173_n20;
  t173_n3: if (E.getNumReferences() >= 3288719360 /*535.5*/) goto t173_n17;
  t173_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t173_n6;
  t173_n5: return -0.0008917553350329399f;
  t173_n6: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t173_n8;
  t173_n7: return 0.0009145672665908933f;
  t173_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t173_n16;
  t173_n9: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t173_n13;
  t173_n10: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t173_n12;
  t173_n11: return 0.005959755275398493f;
  t173_n12: return -0.04682578891515732f;
  t173_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t173_n15;
  t173_n14: return -0.026464831084012985f;
  t173_n15: return -0.05544942989945412f;
  t173_n16: return 0.03088502585887909f;
  t173_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t173_n19;
  t173_n18: return -0.11450574547052383f;
  t173_n19: return -0.15524917840957642f;
  t173_n20: if (E.getNumReferences() >= 3288866816 /*544.5*/) goto t173_n30;
  t173_n21: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t173_n23;
  t173_n22: return -0.09578945487737656f;
  t173_n23: if (E.getNumReferences() >= 3288834048 /*542.5*/) goto t173_n25;
  t173_n24: return 0.09488687664270401f;
  t173_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t173_n29;
  t173_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t173_n28;
  t173_n27: return -0.1546860784292221f;
  t173_n28: return 0.14642727375030518f;
  t173_n29: return 1.3412247896194458f;
  t173_n30: return 0.008381959050893784f;
  t173_n31: return -0.00170743593480438f;
  t173_n32: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t173_n50;
  t173_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t173_n49;
  t173_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t173_n36;
  t173_n35: return -0.0016668552998453379f;
  t173_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t173_n44;
  t173_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t173_n41;
  t173_n38: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t173_n40;
  t173_n39: return -0.08500047028064728f;
  t173_n40: return -0.10804849863052368f;
  t173_n41: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t173_n43;
  t173_n42: return -0.012499036267399788f;
  t173_n43: return 0.011422811076045036f;
  t173_n44: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t173_n48;
  t173_n45: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t173_n47;
  t173_n46: return 0.007388721685856581f;
  t173_n47: return -0.08130119740962982f;
  t173_n48: return 0.040158435702323914f;
  t173_n49: return 0.001618824782781303f;
  t173_n50: if (E.getFractionNameInContext() >= 3185148150 /*0.10620300471782684*/) goto t173_n52;
  t173_n51: return 0.011201594024896622f;
  t173_n52: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t173_n68;
  t173_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t173_n67;
  t173_n54: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t173_n62;
  t173_n55: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t173_n57;
  t173_n56: return -0.006953389849513769f;
  t173_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t173_n59;
  t173_n58: return -0.09715051203966141f;
  t173_n59: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t173_n61;
  t173_n60: return -0.02702506072819233f;
  t173_n61: return 0.07468590140342712f;
  t173_n62: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t173_n66;
  t173_n63: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t173_n65;
  t173_n64: return 0.028259048238396645f;
  t173_n65: return -0.16917258501052856f;
  t173_n66: return 0.05911281704902649f;
  t173_n67: return 0.0011032222537323833f;
  t173_n68: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t173_n212;
  t173_n69: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t173_n169;
  t173_n70: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t173_n88;
  t173_n71: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t173_n77;
  t173_n72: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t173_n74;
  t173_n73: return 0.005140598397701979f;
  t173_n74: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t173_n76;
  t173_n75: return -0.11203208565711975f;
  t173_n76: return 0.14906463027000427f;
  t173_n77: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t173_n79;
  t173_n78: return -0.008285683579742908f;
  t173_n79: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t173_n87;
  t173_n80: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t173_n84;
  t173_n81: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t173_n83;
  t173_n82: return 0.0866333395242691f;
  t173_n83: return -0.2020006626844406f;
  t173_n84: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t173_n86;
  t173_n85: return -0.0997808650135994f;
  t173_n86: return -0.9783933162689209f;
  t173_n87: return -0.05998018756508827f;
  t173_n88: if (E.getNumReferences() >= 3278077952 /*227.5*/) goto t173_n168;
  t173_n89: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t173_n157;
  t173_n90: if (E.getNumReferences() >= 3269918720 /*115.5*/) goto t173_n156;
  t173_n91: if (E.getFractionNameInContext() >= 3197124868 /*0.2817460298538208*/) goto t173_n155;
  t173_n92: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t173_n142;
  t173_n93: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t173_n95;
  t173_n94: return -0.13393384218215942f;
  t173_n95: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t173_n141;
  t173_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t173_n112;
  t173_n97: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t173_n109;
  t173_n98: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t173_n104;
  t173_n99: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t173_n103;
  t173_n100: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t173_n102;
  t173_n101: return -0.02404257468879223f;
  t173_n102: return 0.072862908244133f;
  t173_n103: return -0.4506838917732239f;
  t173_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t173_n106;
  t173_n105: return 0.0115790581330657f;
  t173_n106: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t173_n108;
  t173_n107: return 0.15422536432743073f;
  t173_n108: return 0.12909886240959167f;
  t173_n109: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t173_n111;
  t173_n110: return -0.22489187121391296f;
  t173_n111: return -0.07152915745973587f;
  t173_n112: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t173_n128;
  t173_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t173_n119;
  t173_n114: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t173_n118;
  t173_n115: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t173_n117;
  t173_n116: return -0.1460408866405487f;
  t173_n117: return 0.07376865297555923f;
  t173_n118: return 0.05761319771409035f;
  t173_n119: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t173_n123;
  t173_n120: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t173_n122;
  t173_n121: return 0.019356463104486465f;
  t173_n122: return 0.24398525059223175f;
  t173_n123: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t173_n127;
  t173_n124: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t173_n126;
  t173_n125: return 0.01394380908459425f;
  t173_n126: return 0.09901214390993118f;
  t173_n127: return 0.016309097409248352f;
  t173_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t173_n138;
  t173_n129: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t173_n137;
  t173_n130: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t173_n134;
  t173_n131: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t173_n133;
  t173_n132: return 0.0894433781504631f;
  t173_n133: return -0.004540731199085712f;
  t173_n134: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t173_n136;
  t173_n135: return 0.1397862732410431f;
  t173_n136: return -0.06743987649679184f;
  t173_n137: return 0.169659823179245f;
  t173_n138: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t173_n140;
  t173_n139: return 0.12497296184301376f;
  t173_n140: return 0.029506072402000427f;
  t173_n141: return -0.010138723999261856f;
  t173_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t173_n150;
  t173_n143: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t173_n145;
  t173_n144: return -0.04046115279197693f;
  t173_n145: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t173_n149;
  t173_n146: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t173_n148;
  t173_n147: return -0.9736359715461731f;
  t173_n148: return 0.10014889389276505f;
  t173_n149: return 0.12598353624343872f;
  t173_n150: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t173_n154;
  t173_n151: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t173_n153;
  t173_n152: return 0.16838940978050232f;
  t173_n153: return 0.11930166929960251f;
  t173_n154: return 0.004981586243957281f;
  t173_n155: return 0.008970575407147408f;
  t173_n156: return -0.14635734260082245f;
  t173_n157: if (E.getFractionNameInContext() >= 3209341610 /*0.7916666269302368*/) goto t173_n167;
  t173_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t173_n160;
  t173_n159: return -0.04375835508108139f;
  t173_n160: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t173_n162;
  t173_n161: return -0.004009436350315809f;
  t173_n162: if (E.getNumReferences() >= 3252158464 /*27.0*/) goto t173_n166;
  t173_n163: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t173_n165;
  t173_n164: return 0.09098970144987106f;
  t173_n165: return 0.1092524304986f;
  t173_n166: return -0.01447319332510233f;
  t173_n167: return 0.11796087771654129f;
  t173_n168: return 0.06905807554721832f;
  t173_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t173_n203;
  t173_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t173_n172;
  t173_n171: return 0.013645939528942108f;
  t173_n172: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t173_n202;
  t173_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t173_n187;
  t173_n174: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t173_n178;
  t173_n175: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t173_n177;
  t173_n176: return -0.29997316002845764f;
  t173_n177: return 0.04882584512233734f;
  t173_n178: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t173_n186;
  t173_n179: if (E.getFractionNameInContext() >= 3189577542 /*0.15340909361839294*/) goto t173_n181;
  t173_n180: return -0.10429446399211884f;
  t173_n181: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t173_n185;
  t173_n182: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t173_n184;
  t173_n183: return 0.1406739503145218f;
  t173_n184: return 0.08532875776290894f;
  t173_n185: return -0.08415746688842773f;
  t173_n186: return -0.02048199065029621f;
  t173_n187: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t173_n195;
  t173_n188: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t173_n194;
  t173_n189: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t173_n193;
  t173_n190: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t173_n192;
  t173_n191: return 0.11425051093101501f;
  t173_n192: return -0.12858018279075623f;
  t173_n193: return 0.12851077318191528f;
  t173_n194: return 0.11765944957733154f;
  t173_n195: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t173_n197;
  t173_n196: return -0.11836320161819458f;
  t173_n197: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t173_n199;
  t173_n198: return 0.10852464288473129f;
  t173_n199: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t173_n201;
  t173_n200: return 0.09660116583108902f;
  t173_n201: return 0.03755218908190727f;
  t173_n202: return 0.0165037140250206f;
  t173_n203: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t173_n205;
  t173_n204: return 0.25607162714004517f;
  t173_n205: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t173_n207;
  t173_n206: return -0.014391105622053146f;
  t173_n207: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t173_n211;
  t173_n208: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t173_n210;
  t173_n209: return 0.09831968694925308f;
  t173_n210: return 0.18022115528583527f;
  t173_n211: return 0.10507132858037949f;
  t173_n212: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t173_n220;
  t173_n213: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t173_n219;
  t173_n214: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t173_n216;
  t173_n215: return 0.028055939823389053f;
  t173_n216: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t173_n218;
  t173_n217: return -0.23729431629180908f;
  t173_n218: return -0.041886184364557266f;
  t173_n219: return 0.09564023464918137f;
  t173_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t173_n236;
  t173_n221: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t173_n227;
  t173_n222: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t173_n226;
  t173_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t173_n225;
  t173_n224: return -0.42102760076522827f;
  t173_n225: return 0.18545038998126984f;
  t173_n226: return 0.1309794932603836f;
  t173_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t173_n229;
  t173_n228: return -0.03051151894032955f;
  t173_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t173_n233;
  t173_n230: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t173_n232;
  t173_n231: return 0.07110686600208282f;
  t173_n232: return -0.005393794272094965f;
  t173_n233: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t173_n235;
  t173_n234: return 0.1276245266199112f;
  t173_n235: return 0.13002346456050873f;
  t173_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t173_n244;
  t173_n237: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t173_n241;
  t173_n238: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t173_n240;
  t173_n239: return -0.005825710482895374f;
  t173_n240: return 0.020785963162779808f;
  t173_n241: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t173_n243;
  t173_n242: return 0.02211940661072731f;
  t173_n243: return -0.35123008489608765f;
  t173_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t173_n248;
  t173_n245: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t173_n247;
  t173_n246: return 0.011920137330889702f;
  t173_n247: return 0.0889667347073555f;
  t173_n248: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t173_n252;
  t173_n249: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t173_n251;
  t173_n250: return 0.016563082113862038f;
  t173_n251: return -0.34465640783309937f;
  t173_n252: if (E.getFractionNameInContext() >= 3206096018 /*0.5982142686843872*/) goto t173_n254;
  t173_n253: return 0.0738590881228447f;
  t173_n254: return -0.053838782012462616f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree174(const Example& E) {
  t174_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t174_n2;
  t174_n1: return -0.096468985080719f;
  t174_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t174_n168;
  t174_n3: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t174_n167;
  t174_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t174_n54;
  t174_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n7;
  t174_n6: return -0.09328766167163849f;
  t174_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t174_n41;
  t174_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t174_n14;
  t174_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n13;
  t174_n10: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t174_n12;
  t174_n11: return -0.09359577298164368f;
  t174_n12: return -0.1324118822813034f;
  t174_n13: return -0.09227432310581207f;
  t174_n14: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t174_n16;
  t174_n15: return -0.0015402418794110417f;
  t174_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n32;
  t174_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t174_n19;
  t174_n18: return 0.0006354946526698768f;
  t174_n19: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t174_n29;
  t174_n20: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t174_n22;
  t174_n21: return 0.013221094384789467f;
  t174_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t174_n26;
  t174_n23: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t174_n25;
  t174_n24: return 0.008081657811999321f;
  t174_n25: return 0.1243395134806633f;
  t174_n26: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t174_n28;
  t174_n27: return -0.1985408067703247f;
  t174_n28: return 0.12555186450481415f;
  t174_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t174_n31;
  t174_n30: return -0.03144324570894241f;
  t174_n31: return 0.0404713936150074f;
  t174_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t174_n34;
  t174_n33: return -0.008223721757531166f;
  t174_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t174_n38;
  t174_n35: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t174_n37;
  t174_n36: return 0.050663527101278305f;
  t174_n37: return 0.0731855034828186f;
  t174_n38: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t174_n40;
  t174_n39: return 0.061508528888225555f;
  t174_n40: return -0.09210903942584991f;
  t174_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t174_n53;
  t174_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t174_n52;
  t174_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t174_n45;
  t174_n44: return -0.12185027450323105f;
  t174_n45: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t174_n51;
  t174_n46: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t174_n48;
  t174_n47: return 0.0008789560524746776f;
  t174_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n50;
  t174_n49: return 0.048796575516462326f;
  t174_n50: return -0.03541671112179756f;
  t174_n51: return -0.006868894211947918f;
  t174_n52: return 0.011746163479983807f;
  t174_n53: return 0.007453237194567919f;
  t174_n54: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t174_n160;
  t174_n55: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t174_n99;
  t174_n56: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t174_n62;
  t174_n57: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t174_n59;
  t174_n58: return 0.006396080367267132f;
  t174_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t174_n61;
  t174_n60: return 0.05682593211531639f;
  t174_n61: return 0.1515507698059082f;
  t174_n62: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t174_n84;
  t174_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t174_n79;
  t174_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n78;
  t174_n65: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t174_n69;
  t174_n66: if (E.getNumReferences() >= 3280060416 /*259.5*/) goto t174_n68;
  t174_n67: return -0.00823164265602827f;
  t174_n68: return -0.220645472407341f;
  t174_n69: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t174_n75;
  t174_n70: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t174_n74;
  t174_n71: if (E.getNumReferences() >= 3278077952 /*227.5*/) goto t174_n73;
  t174_n72: return 0.027409400790929794f;
  t174_n73: return -0.20657527446746826f;
  t174_n74: return 0.003312034299597144f;
  t174_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t174_n77;
  t174_n76: return -0.1981055736541748f;
  t174_n77: return -0.026970770210027695f;
  t174_n78: return 0.0148649662733078f;
  t174_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t174_n81;
  t174_n80: return -0.007732229772955179f;
  t174_n81: if (E.getNumReferences() >= 3266248704 /*87.5*/) goto t174_n83;
  t174_n82: return 0.061644576489925385f;
  t174_n83: return 0.1632453203201294f;
  t174_n84: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t174_n96;
  t174_n85: if (E.getNumReferences() >= 3250585600 /*24.0*/) goto t174_n93;
  t174_n86: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t174_n90;
  t174_n87: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n89;
  t174_n88: return -0.07275156676769257f;
  t174_n89: return 0.11475202441215515f;
  t174_n90: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t174_n92;
  t174_n91: return 0.13976353406906128f;
  t174_n92: return 0.1445479691028595f;
  t174_n93: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t174_n95;
  t174_n94: return 0.11831990629434586f;
  t174_n95: return 0.030548501759767532f;
  t174_n96: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t174_n98;
  t174_n97: return 0.0030882006976753473f;
  t174_n98: return -1.2317357063293457f;
  t174_n99: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t174_n109;
  t174_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t174_n108;
  t174_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t174_n103;
  t174_n102: return 0.014739397913217545f;
  t174_n103: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t174_n105;
  t174_n104: return 0.07874274998903275f;
  t174_n105: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t174_n107;
  t174_n106: return -0.09033771604299545f;
  t174_n107: return 0.17504650354385376f;
  t174_n108: return 0.05626197159290314f;
  t174_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n155;
  t174_n110: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t174_n150;
  t174_n111: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t174_n115;
  t174_n112: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t174_n114;
  t174_n113: return -0.15834327042102814f;
  t174_n114: return 0.06547964364290237f;
  t174_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t174_n143;
  t174_n116: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t174_n128;
  t174_n117: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t174_n123;
  t174_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t174_n122;
  t174_n119: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t174_n121;
  t174_n120: return 0.007536131422966719f;
  t174_n121: return -0.21418820321559906f;
  t174_n122: return 0.06980299949645996f;
  t174_n123: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t174_n127;
  t174_n124: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t174_n126;
  t174_n125: return 0.09006668627262115f;
  t174_n126: return 0.10612396895885468f;
  t174_n127: return -0.00015264614194165915f;
  t174_n128: if (E.getNumReferences() >= 3262251008 /*60.5*/) goto t174_n142;
  t174_n129: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t174_n141;
  t174_n130: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t174_n138;
  t174_n131: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t174_n137;
  t174_n132: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t174_n136;
  t174_n133: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t174_n135;
  t174_n134: return 0.5627285242080688f;
  t174_n135: return 0.1840452253818512f;
  t174_n136: return 0.09022029489278793f;
  t174_n137: return 0.11514876782894135f;
  t174_n138: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t174_n140;
  t174_n139: return 0.09563468396663666f;
  t174_n140: return -0.0058653042651712894f;
  t174_n141: return 0.028439583256840706f;
  t174_n142: return 0.12768830358982086f;
  t174_n143: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t174_n145;
  t174_n144: return 0.014290818013250828f;
  t174_n145: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t174_n149;
  t174_n146: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t174_n148;
  t174_n147: return 0.0773436650633812f;
  t174_n148: return -0.22843775153160095f;
  t174_n149: return -0.16712330281734467f;
  t174_n150: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t174_n152;
  t174_n151: return 0.03232301399111748f;
  t174_n152: if (E.getFractionNameInContext() >= 3198110197 /*0.31111112236976624*/) goto t174_n154;
  t174_n153: return 0.23633630573749542f;
  t174_n154: return 0.6824629306793213f;
  t174_n155: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t174_n157;
  t174_n156: return 0.00978131778538227f;
  t174_n157: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t174_n159;
  t174_n158: return 0.1596284955739975f;
  t174_n159: return -0.04999454692006111f;
  t174_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n166;
  t174_n161: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t174_n163;
  t174_n162: return -0.0025289850309491158f;
  t174_n163: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t174_n165;
  t174_n164: return 0.025420241057872772f;
  t174_n165: return -0.00230037746950984f;
  t174_n166: return 0.004225763957947493f;
  t174_n167: return 0.008968564681708813f;
  t174_n168: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t174_n170;
  t174_n169: return -0.0007587576983496547f;
  t174_n170: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t174_n172;
  t174_n171: return 0.14408545196056366f;
  t174_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t174_n180;
  t174_n173: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t174_n177;
  t174_n174: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t174_n176;
  t174_n175: return -0.049539271742105484f;
  t174_n176: return 0.088730089366436f;
  t174_n177: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t174_n179;
  t174_n178: return 0.07049840688705444f;
  t174_n179: return 0.10489281266927719f;
  t174_n180: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t174_n250;
  t174_n181: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t174_n249;
  t174_n182: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t174_n196;
  t174_n183: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t174_n185;
  t174_n184: return -0.016728632152080536f;
  t174_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t174_n193;
  t174_n186: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n192;
  t174_n187: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t174_n189;
  t174_n188: return 0.05902782455086708f;
  t174_n189: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t174_n191;
  t174_n190: return 0.1297723948955536f;
  t174_n191: return -0.12421491742134094f;
  t174_n192: return -0.09086594730615616f;
  t174_n193: if (E.getFractionNameInContext() >= 3200482732 /*0.38181817531585693*/) goto t174_n195;
  t174_n194: return 0.14076916873455048f;
  t174_n195: return 0.12259640544652939f;
  t174_n196: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t174_n198;
  t174_n197: return -0.002354390686377883f;
  t174_n198: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t174_n248;
  t174_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t174_n201;
  t174_n200: return 0.014256265945732594f;
  t174_n201: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t174_n245;
  t174_n202: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t174_n204;
  t174_n203: return -0.014852338470518589f;
  t174_n204: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t174_n242;
  t174_n205: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t174_n223;
  t174_n206: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t174_n214;
  t174_n207: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t174_n213;
  t174_n208: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t174_n210;
  t174_n209: return 0.09075935930013657f;
  t174_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t174_n212;
  t174_n211: return 0.005248914938420057f;
  t174_n212: return 0.21828004717826843f;
  t174_n213: return -0.2520126700401306f;
  t174_n214: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t174_n222;
  t174_n215: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t174_n217;
  t174_n216: return 0.08332827687263489f;
  t174_n217: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t174_n221;
  t174_n218: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t174_n220;
  t174_n219: return 0.1264118105173111f;
  t174_n220: return 0.09497799724340439f;
  t174_n221: return 0.11268199235200882f;
  t174_n222: return 0.12595391273498535f;
  t174_n223: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t174_n231;
  t174_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n226;
  t174_n225: return -0.1587333083152771f;
  t174_n226: if (E.getFractionNameInContext() >= 3207294391 /*0.6696428656578064*/) goto t174_n230;
  t174_n227: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t174_n229;
  t174_n228: return 0.1469915360212326f;
  t174_n229: return -0.5173037052154541f;
  t174_n230: return 0.12664978206157684f;
  t174_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t174_n239;
  t174_n232: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t174_n238;
  t174_n233: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t174_n237;
  t174_n234: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t174_n236;
  t174_n235: return 0.02975708618760109f;
  t174_n236: return -0.4920058846473694f;
  t174_n237: return 0.07965316623449326f;
  t174_n238: return -0.43944665789604187f;
  t174_n239: if (E.getFractionNameInContext() >= 3206096018 /*0.5982142686843872*/) goto t174_n241;
  t174_n240: return 0.1563151776790619f;
  t174_n241: return 0.015891041606664658f;
  t174_n242: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t174_n244;
  t174_n243: return 0.023978129029273987f;
  t174_n244: return 0.09207173436880112f;
  t174_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t174_n247;
  t174_n246: return -0.07274977117776871f;
  t174_n247: return 0.06307583302259445f;
  t174_n248: return 0.020288940519094467f;
  t174_n249: return -0.09171800315380096f;
  t174_n250: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t174_n252;
  t174_n251: return 0.018920212984085083f;
  t174_n252: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t174_n254;
  t174_n253: return 0.13931237161159515f;
  t174_n254: return 0.0656556785106659f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree175(const Example& E) {
  t175_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t175_n98;
  t175_n1: if (E.getNumReferences() >= 3370747520 /*478260.0*/) goto t175_n93;
  t175_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t175_n20;
  t175_n3: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t175_n19;
  t175_n4: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t175_n6;
  t175_n5: return -0.0005238158046267927f;
  t175_n6: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t175_n16;
  t175_n7: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t175_n9;
  t175_n8: return -0.020942864939570427f;
  t175_n9: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t175_n13;
  t175_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t175_n12;
  t175_n11: return 0.09495895355939865f;
  t175_n12: return 0.14127270877361298f;
  t175_n13: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t175_n15;
  t175_n14: return -0.40470048785209656f;
  t175_n15: return 0.11661657691001892f;
  t175_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n18;
  t175_n17: return -0.05277301371097565f;
  t175_n18: return 0.025434531271457672f;
  t175_n19: return -0.03952590376138687f;
  t175_n20: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t175_n92;
  t175_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t175_n89;
  t175_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t175_n30;
  t175_n23: if (E.getNumReferences() >= 3367213520 /*367822.5*/) goto t175_n25;
  t175_n24: return -0.005794932134449482f;
  t175_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t175_n27;
  t175_n26: return 0.04714657738804817f;
  t175_n27: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t175_n29;
  t175_n28: return 0.10251879692077637f;
  t175_n29: return 0.07564283162355423f;
  t175_n30: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t175_n36;
  t175_n31: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t175_n35;
  t175_n32: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t175_n34;
  t175_n33: return -0.0017643598839640617f;
  t175_n34: return -0.022313173860311508f;
  t175_n35: return 0.00502645131200552f;
  t175_n36: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t175_n84;
  t175_n37: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t175_n81;
  t175_n38: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t175_n80;
  t175_n39: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t175_n69;
  t175_n40: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t175_n66;
  t175_n41: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t175_n51;
  t175_n42: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t175_n50;
  t175_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t175_n49;
  t175_n44: if (E.getNumReferences() >= 3283419136 /*362.0*/) goto t175_n46;
  t175_n45: return -0.02194165624678135f;
  t175_n46: if (E.getNumReferences() >= 3284385792 /*391.5*/) goto t175_n48;
  t175_n47: return 0.1677432656288147f;
  t175_n48: return 0.03442801162600517f;
  t175_n49: return 0.00902818888425827f;
  t175_n50: return -0.007091346196830273f;
  t175_n51: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t175_n65;
  t175_n52: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t175_n64;
  t175_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n59;
  t175_n54: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t175_n56;
  t175_n55: return -0.033444225788116455f;
  t175_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t175_n58;
  t175_n57: return -0.10542137920856476f;
  t175_n58: return 0.08117066323757172f;
  t175_n59: if (E.getNumReferences() >= 3289530368 /*585.0*/) goto t175_n63;
  t175_n60: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t175_n62;
  t175_n61: return 0.02387826330959797f;
  t175_n62: return 0.006240928079932928f;
  t175_n63: return -0.018461979925632477f;
  t175_n64: return 0.004701706580817699f;
  t175_n65: return 0.035030487924814224f;
  t175_n66: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t175_n68;
  t175_n67: return 0.05669768154621124f;
  t175_n68: return 0.12565825879573822f;
  t175_n69: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t175_n77;
  t175_n70: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t175_n72;
  t175_n71: return 0.00020849051361437887f;
  t175_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t175_n74;
  t175_n73: return 0.033298883587121964f;
  t175_n74: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t175_n76;
  t175_n75: return 0.13875281810760498f;
  t175_n76: return 0.0929722785949707f;
  t175_n77: if (E.getNumReferences() >= 3325781504 /*11993.5*/) goto t175_n79;
  t175_n78: return -0.020175281912088394f;
  t175_n79: return -0.12997455894947052f;
  t175_n80: return -0.0636647567152977f;
  t175_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n83;
  t175_n82: return -0.08691340684890747f;
  t175_n83: return 0.043892838060855865f;
  t175_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n88;
  t175_n85: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t175_n87;
  t175_n86: return -0.9828953742980957f;
  t175_n87: return -0.31387341022491455f;
  t175_n88: return -0.035022132098674774f;
  t175_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n91;
  t175_n90: return -0.013607353903353214f;
  t175_n91: return 0.007226993329823017f;
  t175_n92: return -0.16004447638988495f;
  t175_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n97;
  t175_n94: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t175_n96;
  t175_n95: return 0.09344089776277542f;
  t175_n96: return -0.1008334830403328f;
  t175_n97: return 0.017562206834554672f;
  t175_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n112;
  t175_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t175_n105;
  t175_n100: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t175_n104;
  t175_n101: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t175_n103;
  t175_n102: return -0.0030609529931098223f;
  t175_n103: return -0.03669579327106476f;
  t175_n104: return 0.030601518228650093f;
  t175_n105: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t175_n107;
  t175_n106: return -0.025231000036001205f;
  t175_n107: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t175_n109;
  t175_n108: return 0.011075994931161404f;
  t175_n109: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t175_n111;
  t175_n110: return 0.030951853841543198f;
  t175_n111: return -0.07899049669504166f;
  t175_n112: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t175_n250;
  t175_n113: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t175_n249;
  t175_n114: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t175_n244;
  t175_n115: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t175_n239;
  t175_n116: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t175_n232;
  t175_n117: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t175_n135;
  t175_n118: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t175_n126;
  t175_n119: if (E.getNumReferences() >= 3310100480 /*3266.0*/) goto t175_n125;
  t175_n120: if (E.getNumReferences() >= 3306506240 /*2388.5*/) goto t175_n124;
  t175_n121: if (E.getNumReferences() >= 3289882624 /*606.5*/) goto t175_n123;
  t175_n122: return 0.005752937868237495f;
  t175_n123: return -0.20777231454849243f;
  t175_n124: return 0.12830030918121338f;
  t175_n125: return -0.35862496495246887f;
  t175_n126: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t175_n134;
  t175_n127: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t175_n129;
  t175_n128: return 0.046977393329143524f;
  t175_n129: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t175_n133;
  t175_n130: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t175_n132;
  t175_n131: return 0.13077950477600098f;
  t175_n132: return 0.030782021582126617f;
  t175_n133: return 0.07942523807287216f;
  t175_n134: return 0.012414990924298763f;
  t175_n135: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t175_n231;
  t175_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t175_n178;
  t175_n137: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t175_n157;
  t175_n138: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t175_n140;
  t175_n139: return -0.010976498946547508f;
  t175_n140: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t175_n146;
  t175_n141: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t175_n143;
  t175_n142: return -0.155215784907341f;
  t175_n143: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t175_n145;
  t175_n144: return -0.062453433871269226f;
  t175_n145: return 0.10024499148130417f;
  t175_n146: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t175_n152;
  t175_n147: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t175_n149;
  t175_n148: return -0.26152631640434265f;
  t175_n149: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t175_n151;
  t175_n150: return -0.042923785746097565f;
  t175_n151: return 0.09962230920791626f;
  t175_n152: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t175_n154;
  t175_n153: return -0.3057248592376709f;
  t175_n154: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t175_n156;
  t175_n155: return 0.13996493816375732f;
  t175_n156: return -0.05157845467329025f;
  t175_n157: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t175_n167;
  t175_n158: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t175_n164;
  t175_n159: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t175_n161;
  t175_n160: return -0.03420880809426308f;
  t175_n161: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t175_n163;
  t175_n162: return 0.18335136771202087f;
  t175_n163: return -0.12483739107847214f;
  t175_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t175_n166;
  t175_n165: return -0.1551242172718048f;
  t175_n166: return 0.15679173171520233f;
  t175_n167: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t175_n171;
  t175_n168: if (E.getNumReferences() >= 3280617472 /*276.5*/) goto t175_n170;
  t175_n169: return 0.012193044647574425f;
  t175_n170: return -0.356397807598114f;
  t175_n171: if (E.getNumReferences() >= 3280863232 /*284.0*/) goto t175_n173;
  t175_n172: return 0.13274633884429932f;
  t175_n173: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t175_n175;
  t175_n174: return 0.033646516501903534f;
  t175_n175: if (E.getNumReferences() >= 3282534400 /*335.0*/) goto t175_n177;
  t175_n176: return 0.1298300176858902f;
  t175_n177: return -0.007591265253722668f;
  t175_n178: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t175_n210;
  t175_n179: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t175_n181;
  t175_n180: return -0.0023559273686259985f;
  t175_n181: if (E.getNumReferences() >= 3299115008 /*1316.0*/) goto t175_n209;
  t175_n182: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t175_n184;
  t175_n183: return -0.02276906743645668f;
  t175_n184: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t175_n204;
  t175_n185: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t175_n203;
  t175_n186: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t175_n202;
  t175_n187: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t175_n197;
  t175_n188: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t175_n194;
  t175_n189: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t175_n193;
  t175_n190: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t175_n192;
  t175_n191: return 0.016575712710618973f;
  t175_n192: return -0.4075877368450165f;
  t175_n193: return -0.01600530929863453f;
  t175_n194: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t175_n196;
  t175_n195: return 0.03227127343416214f;
  t175_n196: return -0.05018428713083267f;
  t175_n197: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t175_n199;
  t175_n198: return -0.0353311151266098f;
  t175_n199: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t175_n201;
  t175_n200: return 0.11380530893802643f;
  t175_n201: return 0.06511328369379044f;
  t175_n202: return 0.049788765609264374f;
  t175_n203: return -0.00679179560393095f;
  t175_n204: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t175_n206;
  t175_n205: return -0.039464071393013f;
  t175_n206: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t175_n208;
  t175_n207: return 0.16453799605369568f;
  t175_n208: return -0.10568825155496597f;
  t175_n209: return -0.06884437799453735f;
  t175_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t175_n224;
  t175_n211: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t175_n223;
  t175_n212: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t175_n222;
  t175_n213: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t175_n215;
  t175_n214: return 0.01997394487261772f;
  t175_n215: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t175_n217;
  t175_n216: return 0.12390165776014328f;
  t175_n217: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t175_n219;
  t175_n218: return 0.040822722017765045f;
  t175_n219: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t175_n221;
  t175_n220: return -0.09409262984991074f;
  t175_n221: return 0.15394443273544312f;
  t175_n222: return -0.039821475744247437f;
  t175_n223: return -0.0033725479152053595f;
  t175_n224: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t175_n230;
  t175_n225: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t175_n227;
  t175_n226: return -0.08896152675151825f;
  t175_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t175_n229;
  t175_n228: return -1.0292984247207642f;
  t175_n229: return -0.105103500187397f;
  t175_n230: return -0.00227805832400918f;
  t175_n231: return 0.006642133928835392f;
  t175_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t175_n234;
  t175_n233: return 0.10927402973175049f;
  t175_n234: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t175_n238;
  t175_n235: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t175_n237;
  t175_n236: return 0.1049899086356163f;
  t175_n237: return 0.10063917934894562f;
  t175_n238: return 0.10779812186956406f;
  t175_n239: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t175_n243;
  t175_n240: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t175_n242;
  t175_n241: return -0.06626220047473907f;
  t175_n242: return 0.09052487462759018f;
  t175_n243: return -0.004057141486555338f;
  t175_n244: if (E.getNumReferences() >= 3335315200 /*26223.5*/) goto t175_n246;
  t175_n245: return 0.10254884511232376f;
  t175_n246: if (E.getNumReferences() >= 3362781216 /*245728.5*/) goto t175_n248;
  t175_n247: return 0.02566828764975071f;
  t175_n248: return -0.2399723082780838f;
  t175_n249: return 0.02366735227406025f;
  t175_n250: if (E.getFractionNameInContext() >= 3200515136 /*0.3827838897705078*/) goto t175_n252;
  t175_n251: return 0.1202285885810852f;
  t175_n252: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t175_n254;
  t175_n253: return 0.0056070731952786446f;
  t175_n254: return 0.11880363523960114f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree176(const Example& E) {
  t176_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t176_n246;
  t176_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t176_n175;
  t176_n2: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t176_n30;
  t176_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n17;
  t176_n4: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t176_n16;
  t176_n5: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t176_n7;
  t176_n6: return 0.0002941511629614979f;
  t176_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t176_n15;
  t176_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t176_n12;
  t176_n9: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t176_n11;
  t176_n10: return -0.09611846506595612f;
  t176_n11: return -0.5746138691902161f;
  t176_n12: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t176_n14;
  t176_n13: return 0.017851008102297783f;
  t176_n14: return 0.065803162753582f;
  t176_n15: return 0.03704368695616722f;
  t176_n16: return -0.0036476741079241037f;
  t176_n17: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t176_n29;
  t176_n18: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t176_n28;
  t176_n19: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t176_n21;
  t176_n20: return -0.0022524830419570208f;
  t176_n21: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t176_n23;
  t176_n22: return 0.005379570182412863f;
  t176_n23: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t176_n25;
  t176_n24: return 0.016327310353517532f;
  t176_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t176_n27;
  t176_n26: return -0.006036695092916489f;
  t176_n27: return 0.019711680710315704f;
  t176_n28: return 0.0003335924120619893f;
  t176_n29: return -0.009782921522855759f;
  t176_n30: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t176_n168;
  t176_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t176_n151;
  t176_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t176_n72;
  t176_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t176_n69;
  t176_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n60;
  t176_n35: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t176_n59;
  t176_n36: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t176_n54;
  t176_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t176_n39;
  t176_n38: return -0.09871011227369308f;
  t176_n39: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t176_n53;
  t176_n40: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t176_n48;
  t176_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t176_n47;
  t176_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n44;
  t176_n43: return -0.04469956085085869f;
  t176_n44: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t176_n46;
  t176_n45: return 0.08496797829866409f;
  t176_n46: return 0.12162208557128906f;
  t176_n47: return 0.024123333394527435f;
  t176_n48: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t176_n50;
  t176_n49: return -0.011448289267718792f;
  t176_n50: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t176_n52;
  t176_n51: return 0.2992783188819885f;
  t176_n52: return -0.08126470446586609f;
  t176_n53: return -0.012491502799093723f;
  t176_n54: if (E.getNumReferences() >= 3295412224 /*944.0*/) goto t176_n58;
  t176_n55: if (E.getNumReferences() >= 3294912512 /*913.5*/) goto t176_n57;
  t176_n56: return -0.011829490773379803f;
  t176_n57: return 0.10479770600795746f;
  t176_n58: return -0.04183437302708626f;
  t176_n59: return 0.025580963119864464f;
  t176_n60: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t176_n62;
  t176_n61: return 0.0034011227544397116f;
  t176_n62: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t176_n68;
  t176_n63: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t176_n67;
  t176_n64: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t176_n66;
  t176_n65: return -0.10021896660327911f;
  t176_n66: return 0.21472430229187012f;
  t176_n67: return 0.10716748982667923f;
  t176_n68: return -0.012739277444779873f;
  t176_n69: if (E.getNumReferences() >= 3233808384 /*6.0*/) goto t176_n71;
  t176_n70: return -0.10002818703651428f;
  t176_n71: return 0.14106352627277374f;
  t176_n72: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t176_n82;
  t176_n73: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t176_n77;
  t176_n74: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t176_n76;
  t176_n75: return 0.06200198084115982f;
  t176_n76: return 0.09073145687580109f;
  t176_n77: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t176_n81;
  t176_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n80;
  t176_n79: return -0.04287879914045334f;
  t176_n80: return 0.019255267456173897f;
  t176_n81: return 0.019505508244037628f;
  t176_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t176_n84;
  t176_n83: return -0.018287012353539467f;
  t176_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t176_n86;
  t176_n85: return -0.06928757578134537f;
  t176_n86: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t176_n88;
  t176_n87: return 0.017104003578424454f;
  t176_n88: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t176_n146;
  t176_n89: if (E.getNumReferences() >= 3262251008 /*60.5*/) goto t176_n101;
  t176_n90: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t176_n92;
  t176_n91: return 0.1057049110531807f;
  t176_n92: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t176_n94;
  t176_n93: return 0.09883683919906616f;
  t176_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t176_n96;
  t176_n95: return -1.0155678987503052f;
  t176_n96: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t176_n100;
  t176_n97: if (E.getNumReferences() >= 3250061312 /*23.0*/) goto t176_n99;
  t176_n98: return -0.14320701360702515f;
  t176_n99: return -0.9717908501625061f;
  t176_n100: return 0.03694210201501846f;
  t176_n101: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t176_n115;
  t176_n102: if (E.getNumReferences() >= 3262513152 /*61.5*/) goto t176_n104;
  t176_n103: return 0.4734632968902588f;
  t176_n104: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t176_n114;
  t176_n105: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t176_n107;
  t176_n106: return -0.06921509653329849f;
  t176_n107: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t176_n113;
  t176_n108: if (E.getNumReferences() >= 3284287488 /*388.5*/) goto t176_n110;
  t176_n109: return 0.05364115536212921f;
  t176_n110: if (E.getNumReferences() >= 3308017664 /*2757.5*/) goto t176_n112;
  t176_n111: return 0.1683986485004425f;
  t176_n112: return 0.2434215247631073f;
  t176_n113: return 0.0621248334646225f;
  t176_n114: return 0.017985502257943153f;
  t176_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t176_n137;
  t176_n116: if (E.getNumReferences() >= 3320181760 /*7358.5*/) goto t176_n134;
  t176_n117: if (E.getNumReferences() >= 3281125376 /*292.0*/) goto t176_n123;
  t176_n118: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t176_n122;
  t176_n119: if (E.getNumReferences() >= 3264217088 /*72.0*/) goto t176_n121;
  t176_n120: return 0.1774294525384903f;
  t176_n121: return -0.30205947160720825f;
  t176_n122: return 0.0932229682803154f;
  t176_n123: if (E.getNumReferences() >= 3282976768 /*348.5*/) goto t176_n125;
  t176_n124: return 0.32392001152038574f;
  t176_n125: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t176_n127;
  t176_n126: return 0.0002062775893136859f;
  t176_n127: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t176_n133;
  t176_n128: if (E.getNumReferences() >= 3288104960 /*505.0*/) goto t176_n130;
  t176_n129: return -0.10365458577871323f;
  t176_n130: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t176_n132;
  t176_n131: return 0.2940492331981659f;
  t176_n132: return 0.16495630145072937f;
  t176_n133: return -0.06900142878293991f;
  t176_n134: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t176_n136;
  t176_n135: return -0.47750529646873474f;
  t176_n136: return 0.12525807321071625f;
  t176_n137: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t176_n141;
  t176_n138: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t176_n140;
  t176_n139: return 0.30538225173950195f;
  t176_n140: return 0.08431611955165863f;
  t176_n141: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t176_n143;
  t176_n142: return 0.05262285843491554f;
  t176_n143: if (E.getNumReferences() >= 3272310784 /*139.5*/) goto t176_n145;
  t176_n144: return 0.1834474354982376f;
  t176_n145: return 0.11473242193460464f;
  t176_n146: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t176_n148;
  t176_n147: return -0.7803411483764648f;
  t176_n148: if (E.getNumReferences() >= 3355493120 /*131852.0*/) goto t176_n150;
  t176_n149: return -0.18576671183109283f;
  t176_n150: return 0.04980821907520294f;
  t176_n151: if (E.getNumReferences() >= 3331353088 /*18485.0*/) goto t176_n163;
  t176_n152: if (E.getNumReferences() >= 3331035648 /*17865.0*/) goto t176_n162;
  t176_n153: if (E.getNumReferences() >= 3330932480 /*17663.5*/) goto t176_n157;
  t176_n154: if (E.getNumReferences() >= 3330792960 /*17391.0*/) goto t176_n156;
  t176_n155: return 0.008290073834359646f;
  t176_n156: return -0.15056908130645752f;
  t176_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t176_n161;
  t176_n158: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t176_n160;
  t176_n159: return -0.02736555226147175f;
  t176_n160: return 0.13350608944892883f;
  t176_n161: return 0.17683443427085876f;
  t176_n162: return -0.06181267648935318f;
  t176_n163: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t176_n167;
  t176_n164: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t176_n166;
  t176_n165: return 0.09925219416618347f;
  t176_n166: return 0.14270353317260742f;
  t176_n167: return 0.014744539745151997f;
  t176_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t176_n174;
  t176_n169: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t176_n171;
  t176_n170: return 0.1281937062740326f;
  t176_n171: if (E.getNumReferences() >= 3335613952 /*26807.0*/) goto t176_n173;
  t176_n172: return -0.013939378783106804f;
  t176_n173: return -0.05221690237522125f;
  t176_n174: return 0.0021785434801131487f;
  t176_n175: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t176_n241;
  t176_n176: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t176_n226;
  t176_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t176_n183;
  t176_n178: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t176_n180;
  t176_n179: return -0.01806742697954178f;
  t176_n180: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t176_n182;
  t176_n181: return -0.050067752599716187f;
  t176_n182: return 0.07505249977111816f;
  t176_n183: if (E.getNumReferences() >= 3363141184 /*251353.0*/) goto t176_n223;
  t176_n184: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t176_n218;
  t176_n185: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t176_n211;
  t176_n186: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t176_n200;
  t176_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t176_n197;
  t176_n188: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t176_n194;
  t176_n189: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t176_n191;
  t176_n190: return 0.012784565798938274f;
  t176_n191: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t176_n193;
  t176_n192: return -0.018069887533783913f;
  t176_n193: return 0.11876127868890762f;
  t176_n194: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t176_n196;
  t176_n195: return 0.046527646481990814f;
  t176_n196: return -0.023192070424556732f;
  t176_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t176_n199;
  t176_n198: return -0.049029525369405746f;
  t176_n199: return 0.028535140678286552f;
  t176_n200: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t176_n208;
  t176_n201: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t176_n207;
  t176_n202: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t176_n206;
  t176_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t176_n205;
  t176_n204: return 0.0022574402391910553f;
  t176_n205: return 0.13375243544578552f;
  t176_n206: return 0.10941572487354279f;
  t176_n207: return -0.1448262482881546f;
  t176_n208: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t176_n210;
  t176_n209: return 0.11296875029802322f;
  t176_n210: return -0.010480662807822227f;
  t176_n211: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t176_n215;
  t176_n212: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t176_n214;
  t176_n213: return -0.008565587922930717f;
  t176_n214: return 0.1121235340833664f;
  t176_n215: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t176_n217;
  t176_n216: return 0.06655385345220566f;
  t176_n217: return -0.5723102688789368f;
  t176_n218: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t176_n222;
  t176_n219: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t176_n221;
  t176_n220: return 0.06067918241024017f;
  t176_n221: return 0.02988412417471409f;
  t176_n222: return -0.016348151490092278f;
  t176_n223: if (E.getNumReferences() >= 3364013584 /*267824.5*/) goto t176_n225;
  t176_n224: return 0.21996672451496124f;
  t176_n225: return 0.038564737886190414f;
  t176_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t176_n234;
  t176_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n229;
  t176_n228: return -0.034600432962179184f;
  t176_n229: if (E.getNumReferences() >= 3292405760 /*760.5*/) goto t176_n233;
  t176_n230: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t176_n232;
  t176_n231: return 0.0007636004011146724f;
  t176_n232: return 0.026343638077378273f;
  t176_n233: return -0.05131208524107933f;
  t176_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t176_n240;
  t176_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n239;
  t176_n236: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t176_n238;
  t176_n237: return -0.27936238050460815f;
  t176_n238: return -0.01855996996164322f;
  t176_n239: return 0.01606757752597332f;
  t176_n240: return 0.06304711848497391f;
  t176_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t176_n243;
  t176_n242: return -0.001968995202332735f;
  t176_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t176_n245;
  t176_n244: return -0.02867327444255352f;
  t176_n245: return 0.007788061164319515f;
  t176_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t176_n248;
  t176_n247: return 0.0014850205043330789f;
  t176_n248: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t176_n250;
  t176_n249: return 0.004518900066614151f;
  t176_n250: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t176_n254;
  t176_n251: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t176_n253;
  t176_n252: return -0.14863747358322144f;
  t176_n253: return 0.09962848573923111f;
  t176_n254: return -0.14569377899169922f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree177(const Example& E) {
  t177_n0: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t177_n254;
  t177_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t177_n199;
  t177_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t177_n4;
  t177_n3: return -0.09535988420248032f;
  t177_n4: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t177_n8;
  t177_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n7;
  t177_n6: return -0.10226359963417053f;
  t177_n7: return 0.11808482557535172f;
  t177_n8: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t177_n16;
  t177_n9: if (E.getNumReferences() >= 3314787328 /*4724.5*/) goto t177_n11;
  t177_n10: return 0.0005527889006771147f;
  t177_n11: if (E.getNumReferences() >= 3314960384 /*4809.0*/) goto t177_n15;
  t177_n12: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t177_n14;
  t177_n13: return 0.06484153121709824f;
  t177_n14: return 0.1450112760066986f;
  t177_n15: return 0.010014809668064117f;
  t177_n16: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t177_n196;
  t177_n17: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t177_n193;
  t177_n18: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t177_n20;
  t177_n19: return -0.0006641882937401533f;
  t177_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t177_n64;
  t177_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t177_n31;
  t177_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n30;
  t177_n23: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t177_n25;
  t177_n24: return -0.002784009324386716f;
  t177_n25: if (E.getNumReferences() >= 3335474944 /*26535.5*/) goto t177_n27;
  t177_n26: return -0.0290871299803257f;
  t177_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t177_n29;
  t177_n28: return 0.02644176594913006f;
  t177_n29: return 0.09155304729938507f;
  t177_n30: return 0.008703641593456268f;
  t177_n31: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t177_n61;
  t177_n32: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t177_n58;
  t177_n33: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t177_n37;
  t177_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n36;
  t177_n35: return -0.0579846054315567f;
  t177_n36: return 0.020758051425218582f;
  t177_n37: if (E.getNumReferences() >= 3265527808 /*82.0*/) goto t177_n57;
  t177_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t177_n52;
  t177_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t177_n51;
  t177_n40: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t177_n46;
  t177_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t177_n45;
  t177_n42: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t177_n44;
  t177_n43: return -0.01411597989499569f;
  t177_n44: return 0.16390562057495117f;
  t177_n45: return 0.06605442613363266f;
  t177_n46: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t177_n50;
  t177_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t177_n49;
  t177_n48: return 0.13149411976337433f;
  t177_n49: return 0.19765189290046692f;
  t177_n50: return 0.06347891688346863f;
  t177_n51: return 0.015494955703616142f;
  t177_n52: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t177_n54;
  t177_n53: return 0.02689986489713192f;
  t177_n54: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t177_n56;
  t177_n55: return 0.13503018021583557f;
  t177_n56: return 0.004395286086946726f;
  t177_n57: return 3.0736919143237174e-05f;
  t177_n58: if (E.getNumReferences() >= 3329617408 /*15739.5*/) goto t177_n60;
  t177_n59: return -0.049353621900081635f;
  t177_n60: return -0.8960301280021667f;
  t177_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n63;
  t177_n62: return -0.007662306074053049f;
  t177_n63: return 0.00593505660071969f;
  t177_n64: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t177_n190;
  t177_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n85;
  t177_n66: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t177_n70;
  t177_n67: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t177_n69;
  t177_n68: return -0.03830653056502342f;
  t177_n69: return -0.23837910592556f;
  t177_n70: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t177_n80;
  t177_n71: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t177_n73;
  t177_n72: return -0.10442738980054855f;
  t177_n73: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t177_n79;
  t177_n74: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t177_n76;
  t177_n75: return 0.14999516308307648f;
  t177_n76: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t177_n78;
  t177_n77: return 0.2815078794956207f;
  t177_n78: return 0.21946834027767181f;
  t177_n79: return -0.2523823082447052f;
  t177_n80: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t177_n82;
  t177_n81: return 0.10971372574567795f;
  t177_n82: if (E.getNumReferences() >= 3259891712 /*51.5*/) goto t177_n84;
  t177_n83: return -0.1416226178407669f;
  t177_n84: return 0.07085999846458435f;
  t177_n85: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t177_n133;
  t177_n86: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t177_n130;
  t177_n87: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t177_n99;
  t177_n88: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t177_n90;
  t177_n89: return -0.026385167613625526f;
  t177_n90: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t177_n98;
  t177_n91: if (E.getNumReferences() >= 3271819264 /*132.0*/) goto t177_n97;
  t177_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n94;
  t177_n93: return -0.05201803892850876f;
  t177_n94: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t177_n96;
  t177_n95: return 0.08479157090187073f;
  t177_n96: return -0.0065581039525568485f;
  t177_n97: return 0.1527671366930008f;
  t177_n98: return -0.05873231962323189f;
  t177_n99: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t177_n129;
  t177_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t177_n124;
  t177_n101: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t177_n119;
  t177_n102: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t177_n116;
  t177_n103: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t177_n113;
  t177_n104: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t177_n112;
  t177_n105: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t177_n107;
  t177_n106: return 0.01811033859848976f;
  t177_n107: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t177_n111;
  t177_n108: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t177_n110;
  t177_n109: return 0.07087352126836777f;
  t177_n110: return 0.09299182891845703f;
  t177_n111: return 0.045227162539958954f;
  t177_n112: return 0.13282901048660278f;
  t177_n113: if (E.getNumReferences() >= 3275489280 /*188.0*/) goto t177_n115;
  t177_n114: return -0.06443235278129578f;
  t177_n115: return 0.09297393262386322f;
  t177_n116: if (E.getNumReferences() >= 3246915584 /*17.0*/) goto t177_n118;
  t177_n117: return 0.16013430058956146f;
  t177_n118: return 0.002750653773546219f;
  t177_n119: if (E.getNumReferences() >= 3266314240 /*88.0*/) goto t177_n121;
  t177_n120: return 0.0101302620023489f;
  t177_n121: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t177_n123;
  t177_n122: return 0.0781799927353859f;
  t177_n123: return 0.15511928498744965f;
  t177_n124: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t177_n126;
  t177_n125: return 0.07252775877714157f;
  t177_n126: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t177_n128;
  t177_n127: return 0.19230514764785767f;
  t177_n128: return 0.2163838893175125f;
  t177_n129: return 0.27573221921920776f;
  t177_n130: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t177_n132;
  t177_n131: return -0.053252559155225754f;
  t177_n132: return 0.09940724074840546f;
  t177_n133: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t177_n159;
  t177_n134: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t177_n142;
  t177_n135: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t177_n137;
  t177_n136: return 0.004336277022957802f;
  t177_n137: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t177_n139;
  t177_n138: return 0.04863325506448746f;
  t177_n139: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t177_n141;
  t177_n140: return 0.05020361766219139f;
  t177_n141: return 0.13271284103393555f;
  t177_n142: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t177_n156;
  t177_n143: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t177_n155;
  t177_n144: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t177_n148;
  t177_n145: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t177_n147;
  t177_n146: return 0.14748287200927734f;
  t177_n147: return -0.12136676162481308f;
  t177_n148: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t177_n152;
  t177_n149: if (E.getNumReferences() >= 3249012736 /*21.0*/) goto t177_n151;
  t177_n150: return -0.015160782262682915f;
  t177_n151: return 0.10447061061859131f;
  t177_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t177_n154;
  t177_n153: return 0.15126657485961914f;
  t177_n154: return 0.22783175110816956f;
  t177_n155: return -0.01326843537390232f;
  t177_n156: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t177_n158;
  t177_n157: return -0.0669851005077362f;
  t177_n158: return -0.2805939316749573f;
  t177_n159: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t177_n185;
  t177_n160: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t177_n174;
  t177_n161: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t177_n165;
  t177_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t177_n164;
  t177_n163: return 0.12244725972414017f;
  t177_n164: return 0.1171283945441246f;
  t177_n165: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t177_n173;
  t177_n166: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t177_n170;
  t177_n167: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t177_n169;
  t177_n168: return -0.00916599202901125f;
  t177_n169: return 0.07234552502632141f;
  t177_n170: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t177_n172;
  t177_n171: return 0.08173488825559616f;
  t177_n172: return 0.07061293721199036f;
  t177_n173: return -0.08142352104187012f;
  t177_n174: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t177_n178;
  t177_n175: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t177_n177;
  t177_n176: return -0.0019000038737431169f;
  t177_n177: return -0.18573302030563354f;
  t177_n178: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t177_n184;
  t177_n179: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t177_n181;
  t177_n180: return 0.04669911414384842f;
  t177_n181: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t177_n183;
  t177_n182: return 0.14613664150238037f;
  t177_n183: return -0.008056914433836937f;
  t177_n184: return -0.014299876987934113f;
  t177_n185: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t177_n187;
  t177_n186: return 0.07272762805223465f;
  t177_n187: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t177_n189;
  t177_n188: return 0.13624900579452515f;
  t177_n189: return 0.08884573727846146f;
  t177_n190: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t177_n192;
  t177_n191: return -0.026868781074881554f;
  t177_n192: return 0.035389598459005356f;
  t177_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t177_n195;
  t177_n194: return -0.018239153549075127f;
  t177_n195: return -0.0026173803489655256f;
  t177_n196: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t177_n198;
  t177_n197: return 0.13938331604003906f;
  t177_n198: return 0.009985058568418026f;
  t177_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t177_n243;
  t177_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t177_n236;
  t177_n201: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t177_n213;
  t177_n202: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t177_n210;
  t177_n203: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t177_n205;
  t177_n204: return -0.0015464057214558125f;
  t177_n205: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t177_n209;
  t177_n206: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t177_n208;
  t177_n207: return -0.010356243699789047f;
  t177_n208: return 0.02838524989783764f;
  t177_n209: return -0.027588162571191788f;
  t177_n210: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t177_n212;
  t177_n211: return -0.0254798736423254f;
  t177_n212: return 0.09604150801897049f;
  t177_n213: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t177_n235;
  t177_n214: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t177_n234;
  t177_n215: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t177_n229;
  t177_n216: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t177_n224;
  t177_n217: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t177_n219;
  t177_n218: return 0.010250476188957691f;
  t177_n219: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t177_n221;
  t177_n220: return -0.5246350169181824f;
  t177_n221: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t177_n223;
  t177_n222: return 0.10827133059501648f;
  t177_n223: return 0.08049481362104416f;
  t177_n224: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t177_n228;
  t177_n225: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t177_n227;
  t177_n226: return -0.7910801768302917f;
  t177_n227: return -0.08721933513879776f;
  t177_n228: return 0.06202736869454384f;
  t177_n229: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t177_n231;
  t177_n230: return 0.023292699828743935f;
  t177_n231: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t177_n233;
  t177_n232: return 0.10883251577615738f;
  t177_n233: return 0.1275295466184616f;
  t177_n234: return 0.12088329344987869f;
  t177_n235: return -0.01784696616232395f;
  t177_n236: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t177_n240;
  t177_n237: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t177_n239;
  t177_n238: return -0.030130142346024513f;
  t177_n239: return 0.10231555253267288f;
  t177_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t177_n242;
  t177_n241: return 0.013276598416268826f;
  t177_n242: return 0.04543843865394592f;
  t177_n243: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t177_n249;
  t177_n244: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t177_n248;
  t177_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t177_n247;
  t177_n246: return 0.0030151018872857094f;
  t177_n247: return 0.0048087118193507195f;
  t177_n248: return 0.021828781813383102f;
  t177_n249: if (E.getFractionNameInContext() >= 3200515136 /*0.3827838897705078*/) goto t177_n251;
  t177_n250: return 0.1128876656293869f;
  t177_n251: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t177_n253;
  t177_n252: return -0.01151646114885807f;
  t177_n253: return 0.11269547045230865f;
  t177_n254: return -0.05492208153009415f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree178(const Example& E) {
  t178_n0: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t178_n2;
  t178_n1: return -0.0005100686103105545f;
  t178_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t178_n32;
  t178_n3: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t178_n19;
  t178_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t178_n6;
  t178_n5: return -6.960803148103878e-05f;
  t178_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t178_n8;
  t178_n7: return -0.043084193021059036f;
  t178_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t178_n18;
  t178_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t178_n17;
  t178_n10: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t178_n12;
  t178_n11: return -0.030684439465403557f;
  t178_n12: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t178_n16;
  t178_n13: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t178_n15;
  t178_n14: return 0.09917616099119186f;
  t178_n15: return -0.2081533521413803f;
  t178_n16: return 0.12056265026330948f;
  t178_n17: return 0.062019556760787964f;
  t178_n18: return -0.09669066965579987f;
  t178_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t178_n21;
  t178_n20: return 0.0013002667110413313f;
  t178_n21: if (E.getNumReferences() >= 3299512320 /*1364.5*/) goto t178_n31;
  t178_n22: if (E.getNumReferences() >= 3299295232 /*1338.0*/) goto t178_n26;
  t178_n23: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t178_n25;
  t178_n24: return -0.02401590906083584f;
  t178_n25: return 0.0670204907655716f;
  t178_n26: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t178_n28;
  t178_n27: return 0.14400792121887207f;
  t178_n28: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t178_n30;
  t178_n29: return 0.12588824331760406f;
  t178_n30: return 0.00556566845625639f;
  t178_n31: return 0.017298555001616478f;
  t178_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t178_n250;
  t178_n33: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t178_n39;
  t178_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t178_n36;
  t178_n35: return 0.0021876555401831865f;
  t178_n36: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t178_n38;
  t178_n37: return 0.009361300617456436f;
  t178_n38: return 0.049685217440128326f;
  t178_n39: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t178_n175;
  t178_n40: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t178_n132;
  t178_n41: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t178_n61;
  t178_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t178_n58;
  t178_n43: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t178_n51;
  t178_n44: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t178_n46;
  t178_n45: return 0.03248167783021927f;
  t178_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t178_n48;
  t178_n47: return 0.08646634221076965f;
  t178_n48: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t178_n50;
  t178_n49: return 0.18322204053401947f;
  t178_n50: return 0.07389295846223831f;
  t178_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t178_n55;
  t178_n52: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t178_n54;
  t178_n53: return -0.09912729263305664f;
  t178_n54: return -0.9552164077758789f;
  t178_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t178_n57;
  t178_n56: return 0.017153212800621986f;
  t178_n57: return -0.06405722349882126f;
  t178_n58: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t178_n60;
  t178_n59: return 0.14339470863342285f;
  t178_n60: return 0.07936793565750122f;
  t178_n61: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t178_n131;
  t178_n62: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t178_n88;
  t178_n63: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t178_n87;
  t178_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t178_n84;
  t178_n65: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t178_n73;
  t178_n66: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t178_n72;
  t178_n67: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t178_n69;
  t178_n68: return 0.11552626639604568f;
  t178_n69: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t178_n71;
  t178_n70: return -0.1669202744960785f;
  t178_n71: return -0.7937946915626526f;
  t178_n72: return 0.09356258809566498f;
  t178_n73: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t178_n83;
  t178_n74: if (E.getFractionNameInContext() >= 3184451170 /*0.10101009905338287*/) goto t178_n76;
  t178_n75: return -2.5841033458709717f;
  t178_n76: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t178_n78;
  t178_n77: return -1.0110386610031128f;
  t178_n78: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t178_n80;
  t178_n79: return -0.22262850403785706f;
  t178_n80: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t178_n82;
  t178_n81: return 0.1234322115778923f;
  t178_n82: return -0.1267455518245697f;
  t178_n83: return 0.017712444067001343f;
  t178_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t178_n86;
  t178_n85: return -0.022658390924334526f;
  t178_n86: return 0.6228850483894348f;
  t178_n87: return -0.00701581547036767f;
  t178_n88: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t178_n98;
  t178_n89: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t178_n97;
  t178_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t178_n92;
  t178_n91: return -0.09297393262386322f;
  t178_n92: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t178_n96;
  t178_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t178_n95;
  t178_n94: return 0.14103424549102783f;
  t178_n95: return 0.09557675570249557f;
  t178_n96: return -0.05979524552822113f;
  t178_n97: return 0.1099986732006073f;
  t178_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t178_n102;
  t178_n99: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t178_n101;
  t178_n100: return -0.003742254339158535f;
  t178_n101: return -0.16642794013023376f;
  t178_n102: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t178_n108;
  t178_n103: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t178_n105;
  t178_n104: return -0.021261516958475113f;
  t178_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t178_n107;
  t178_n106: return -0.09590594470500946f;
  t178_n107: return 0.07009618729352951f;
  t178_n108: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t178_n126;
  t178_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t178_n125;
  t178_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t178_n122;
  t178_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t178_n117;
  t178_n112: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t178_n116;
  t178_n113: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t178_n115;
  t178_n114: return 0.19172808527946472f;
  t178_n115: return -0.13028579950332642f;
  t178_n116: return 0.07303783297538757f;
  t178_n117: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t178_n121;
  t178_n118: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t178_n120;
  t178_n119: return 0.41816896200180054f;
  t178_n120: return 0.2023322880268097f;
  t178_n121: return -0.1334366500377655f;
  t178_n122: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t178_n124;
  t178_n123: return -0.30031922459602356f;
  t178_n124: return 0.07864696532487869f;
  t178_n125: return 0.018743246793746948f;
  t178_n126: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t178_n128;
  t178_n127: return 0.04689643904566765f;
  t178_n128: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t178_n130;
  t178_n129: return 0.12440567463636398f;
  t178_n130: return -0.8952358961105347f;
  t178_n131: return 0.036704275757074356f;
  t178_n132: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t178_n134;
  t178_n133: return 0.0001776784920366481f;
  t178_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace))) goto t178_n148;
  t178_n135: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t178_n139;
  t178_n136: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t178_n138;
  t178_n137: return -0.07106351107358932f;
  t178_n138: return -0.0007429310935549438f;
  t178_n139: if (E.getNumReferences() >= 3343692416 /*52402.5*/) goto t178_n147;
  t178_n140: if (E.getNumReferences() >= 3338280960 /*32016.0*/) goto t178_n142;
  t178_n141: return -0.028750592842698097f;
  t178_n142: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t178_n144;
  t178_n143: return 0.02718137949705124f;
  t178_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t178_n146;
  t178_n145: return 0.10132163763046265f;
  t178_n146: return 0.10763930529356003f;
  t178_n147: return -0.15670719742774963f;
  t178_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t178_n158;
  t178_n149: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t178_n151;
  t178_n150: return 0.02109311893582344f;
  t178_n151: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t178_n157;
  t178_n152: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t178_n156;
  t178_n153: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t178_n155;
  t178_n154: return -0.7869927287101746f;
  t178_n155: return -0.03048560954630375f;
  t178_n156: return -0.47887852787971497f;
  t178_n157: return -0.03961500525474548f;
  t178_n158: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t178_n168;
  t178_n159: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t178_n167;
  t178_n160: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t178_n166;
  t178_n161: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t178_n165;
  t178_n162: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t178_n164;
  t178_n163: return 0.09798986464738846f;
  t178_n164: return 0.12627819180488586f;
  t178_n165: return -0.2999638020992279f;
  t178_n166: return 0.0022684449795633554f;
  t178_n167: return -0.03589881584048271f;
  t178_n168: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t178_n172;
  t178_n169: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t178_n171;
  t178_n170: return 0.15006311237812042f;
  t178_n171: return 0.12107402831315994f;
  t178_n172: if (E.getFileProximityDistanceCost() >= 3250061312 /*23.0*/) goto t178_n174;
  t178_n173: return -0.15271419286727905f;
  t178_n174: return 0.08243975043296814f;
  t178_n175: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t178_n201;
  t178_n176: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t178_n184;
  t178_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t178_n179;
  t178_n178: return 0.038208071142435074f;
  t178_n179: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t178_n181;
  t178_n180: return -0.04580643028020859f;
  t178_n181: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t178_n183;
  t178_n182: return 0.13902521133422852f;
  t178_n183: return 0.1928253024816513f;
  t178_n184: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t178_n190;
  t178_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t178_n187;
  t178_n186: return 0.13588358461856842f;
  t178_n187: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t178_n189;
  t178_n188: return 0.10640201717615128f;
  t178_n189: return 0.27521586418151855f;
  t178_n190: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t178_n196;
  t178_n191: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t178_n193;
  t178_n192: return -0.04968775436282158f;
  t178_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t178_n195;
  t178_n194: return -0.8070321083068848f;
  t178_n195: return -0.10505561530590057f;
  t178_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t178_n198;
  t178_n197: return -0.06606924533843994f;
  t178_n198: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t178_n200;
  t178_n199: return 0.09526514261960983f;
  t178_n200: return 0.15734028816223145f;
  t178_n201: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t178_n215;
  t178_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t178_n214;
  t178_n203: if (E.getNumReferences() >= 3305187328 /*2066.5*/) goto t178_n213;
  t178_n204: if (E.getNumReferences() >= 3303534592 /*1855.5*/) goto t178_n210;
  t178_n205: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t178_n207;
  t178_n206: return -0.040857333689928055f;
  t178_n207: if (E.getNumReferences() >= 3241148416 /*11.0*/) goto t178_n209;
  t178_n208: return 0.15982122719287872f;
  t178_n209: return -0.13746324181556702f;
  t178_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t178_n212;
  t178_n211: return -0.10096287727355957f;
  t178_n212: return 0.13904599845409393f;
  t178_n213: return -0.12020796537399292f;
  t178_n214: return -0.0025466899387538433f;
  t178_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t178_n217;
  t178_n216: return 0.007338539697229862f;
  t178_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t178_n219;
  t178_n218: return -0.023864945396780968f;
  t178_n219: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t178_n239;
  t178_n220: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t178_n234;
  t178_n221: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t178_n229;
  t178_n222: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t178_n226;
  t178_n223: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t178_n225;
  t178_n224: return 0.08780362457036972f;
  t178_n225: return 0.029775960370898247f;
  t178_n226: if (E.getNumReferences() >= 3327046144 /*13228.5*/) goto t178_n228;
  t178_n227: return -0.10848844051361084f;
  t178_n228: return 0.08552569150924683f;
  t178_n229: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t178_n233;
  t178_n230: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t178_n232;
  t178_n231: return -0.060783758759498596f;
  t178_n232: return -0.4393450915813446f;
  t178_n233: return 0.034917574375867844f;
  t178_n234: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t178_n236;
  t178_n235: return 0.09311555325984955f;
  t178_n236: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t178_n238;
  t178_n237: return -0.01751863956451416f;
  t178_n238: return 0.05050090700387955f;
  t178_n239: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t178_n241;
  t178_n240: return -0.03493228927254677f;
  t178_n241: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t178_n245;
  t178_n242: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t178_n244;
  t178_n243: return 0.06273441761732101f;
  t178_n244: return -0.07772652059793472f;
  t178_n245: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t178_n247;
  t178_n246: return -0.027615725994110107f;
  t178_n247: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t178_n249;
  t178_n248: return -0.1614069789648056f;
  t178_n249: return 0.023193202912807465f;
  t178_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t178_n254;
  t178_n251: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t178_n253;
  t178_n252: return 0.05301434546709061f;
  t178_n253: return 0.11822260916233063f;
  t178_n254: return -0.10892389714717865f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree179(const Example& E) {
  t179_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t179_n204;
  t179_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t179_n149;
  t179_n2: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t179_n146;
  t179_n3: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t179_n137;
  t179_n4: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t179_n136;
  t179_n5: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t179_n7;
  t179_n6: return 0.00016226922161877155f;
  t179_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n81;
  t179_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t179_n74;
  t179_n9: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t179_n17;
  t179_n10: if (E.getNumReferences() >= 3276374016 /*201.5*/) goto t179_n16;
  t179_n11: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t179_n15;
  t179_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n14;
  t179_n13: return -0.4079519808292389f;
  t179_n14: return -0.008202120661735535f;
  t179_n15: return 0.019501203671097755f;
  t179_n16: return 0.038438569754362106f;
  t179_n17: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t179_n73;
  t179_n18: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t179_n64;
  t179_n19: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t179_n53;
  t179_n20: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t179_n34;
  t179_n21: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t179_n23;
  t179_n22: return -0.032017920166254044f;
  t179_n23: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t179_n31;
  t179_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n30;
  t179_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t179_n29;
  t179_n26: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t179_n28;
  t179_n27: return 0.013192238286137581f;
  t179_n28: return 0.07222909480333328f;
  t179_n29: return 0.25113698840141296f;
  t179_n30: return -0.20234814286231995f;
  t179_n31: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t179_n33;
  t179_n32: return -0.00795096904039383f;
  t179_n33: return -0.09432873129844666f;
  t179_n34: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t179_n52;
  t179_n35: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t179_n43;
  t179_n36: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t179_n40;
  t179_n37: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n39;
  t179_n38: return -0.0604020357131958f;
  t179_n39: return 0.030251210555434227f;
  t179_n40: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t179_n42;
  t179_n41: return -0.1131797805428505f;
  t179_n42: return 0.05949778854846954f;
  t179_n43: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t179_n51;
  t179_n44: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t179_n48;
  t179_n45: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t179_n47;
  t179_n46: return 0.0019509608391672373f;
  t179_n47: return -0.2458588182926178f;
  t179_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t179_n50;
  t179_n49: return -0.056215718388557434f;
  t179_n50: return 0.09234841912984848f;
  t179_n51: return 0.03117714263498783f;
  t179_n52: return -0.34352847933769226f;
  t179_n53: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t179_n63;
  t179_n54: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t179_n58;
  t179_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t179_n57;
  t179_n56: return -0.2348085641860962f;
  t179_n57: return 0.027092689648270607f;
  t179_n58: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t179_n62;
  t179_n59: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t179_n61;
  t179_n60: return 0.09832858294248581f;
  t179_n61: return 0.09611072391271591f;
  t179_n62: return -0.038136232644319534f;
  t179_n63: return -0.04703669622540474f;
  t179_n64: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t179_n68;
  t179_n65: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t179_n67;
  t179_n66: return 0.00507644098252058f;
  t179_n67: return -0.2539588510990143f;
  t179_n68: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t179_n72;
  t179_n69: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t179_n71;
  t179_n70: return 0.08030194044113159f;
  t179_n71: return 0.12493940442800522f;
  t179_n72: return -0.14519327878952026f;
  t179_n73: return -0.010081583634018898f;
  t179_n74: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t179_n80;
  t179_n75: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t179_n77;
  t179_n76: return 0.00540500320494175f;
  t179_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t179_n79;
  t179_n78: return -0.04353351145982742f;
  t179_n79: return 0.05110758915543556f;
  t179_n80: return 0.02647249773144722f;
  t179_n81: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t179_n131;
  t179_n82: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t179_n126;
  t179_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n85;
  t179_n84: return -0.018537169322371483f;
  t179_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t179_n119;
  t179_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t179_n114;
  t179_n87: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t179_n111;
  t179_n88: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t179_n98;
  t179_n89: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t179_n91;
  t179_n90: return -0.0049418495036661625f;
  t179_n91: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t179_n95;
  t179_n92: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t179_n94;
  t179_n93: return 0.031375300139188766f;
  t179_n94: return 0.28423357009887695f;
  t179_n95: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t179_n97;
  t179_n96: return 0.15809345245361328f;
  t179_n97: return 0.09760113805532455f;
  t179_n98: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t179_n102;
  t179_n99: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t179_n101;
  t179_n100: return -0.25154778361320496f;
  t179_n101: return 0.02515174262225628f;
  t179_n102: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n110;
  t179_n103: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t179_n105;
  t179_n104: return 0.010832361876964569f;
  t179_n105: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t179_n109;
  t179_n106: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t179_n108;
  t179_n107: return 0.10365407168865204f;
  t179_n108: return 0.11714857816696167f;
  t179_n109: return 0.029517602175474167f;
  t179_n110: return 0.031029099598526955f;
  t179_n111: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t179_n113;
  t179_n112: return 0.027291355654597282f;
  t179_n113: return -0.014034941792488098f;
  t179_n114: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t179_n116;
  t179_n115: return 0.03841174766421318f;
  t179_n116: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t179_n118;
  t179_n117: return -0.031086860224604607f;
  t179_n118: return -0.23258869349956512f;
  t179_n119: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n125;
  t179_n120: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t179_n124;
  t179_n121: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t179_n123;
  t179_n122: return 0.041735634207725525f;
  t179_n123: return 0.0902668908238411f;
  t179_n124: return -0.14190496504306793f;
  t179_n125: return -0.015682971104979515f;
  t179_n126: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t179_n130;
  t179_n127: if (E.getNumReferences() >= 3333705984 /*23080.5*/) goto t179_n129;
  t179_n128: return 0.01788323186337948f;
  t179_n129: return -0.2505795657634735f;
  t179_n130: return -0.005352239590138197f;
  t179_n131: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n135;
  t179_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n134;
  t179_n133: return -0.009086435660719872f;
  t179_n134: return 0.004717185627669096f;
  t179_n135: return 0.008756190538406372f;
  t179_n136: return -0.0023325751535594463f;
  t179_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t179_n141;
  t179_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n140;
  t179_n139: return -0.16664007306098938f;
  t179_n140: return 0.0008247844525612891f;
  t179_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t179_n143;
  t179_n142: return 0.0027417882811278105f;
  t179_n143: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t179_n145;
  t179_n144: return 0.08001482486724854f;
  t179_n145: return 0.03488713130354881f;
  t179_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t179_n148;
  t179_n147: return -0.0053966413252055645f;
  t179_n148: return 0.03249695897102356f;
  t179_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t179_n197;
  t179_n150: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t179_n152;
  t179_n151: return 0.0005845198757015169f;
  t179_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t179_n196;
  t179_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n155;
  t179_n154: return -0.023135531693696976f;
  t179_n155: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t179_n179;
  t179_n156: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t179_n166;
  t179_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n165;
  t179_n158: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t179_n162;
  t179_n159: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t179_n161;
  t179_n160: return -0.06755095720291138f;
  t179_n161: return 0.08439026027917862f;
  t179_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t179_n164;
  t179_n163: return -0.9589225053787231f;
  t179_n164: return -0.15856745839118958f;
  t179_n165: return 0.04434896633028984f;
  t179_n166: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t179_n172;
  t179_n167: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t179_n171;
  t179_n168: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t179_n170;
  t179_n169: return 0.1097937524318695f;
  t179_n170: return 0.11902729421854019f;
  t179_n171: return -0.189262256026268f;
  t179_n172: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t179_n178;
  t179_n173: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t179_n175;
  t179_n174: return 0.039004433900117874f;
  t179_n175: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t179_n177;
  t179_n176: return -0.13049030303955078f;
  t179_n177: return 0.009226232767105103f;
  t179_n178: return 0.10310691595077515f;
  t179_n179: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t179_n187;
  t179_n180: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t179_n184;
  t179_n181: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t179_n183;
  t179_n182: return -0.03057483211159706f;
  t179_n183: return 0.05056481063365936f;
  t179_n184: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t179_n186;
  t179_n185: return -0.8645451664924622f;
  t179_n186: return -0.00163967942353338f;
  t179_n187: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t179_n195;
  t179_n188: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t179_n192;
  t179_n189: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n191;
  t179_n190: return 0.05515288561582565f;
  t179_n191: return 0.003598278621211648f;
  t179_n192: if (E.getNumReferences() >= 3276734464 /*207.0*/) goto t179_n194;
  t179_n193: return 0.07963226735591888f;
  t179_n194: return -0.4548892676830292f;
  t179_n195: return 0.01613902859389782f;
  t179_n196: return 0.06719109416007996f;
  t179_n197: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t179_n203;
  t179_n198: if (E.getNumReferences() >= 3301859328 /*1651.0*/) goto t179_n202;
  t179_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n201;
  t179_n200: return -0.041076209396123886f;
  t179_n201: return 0.017220964655280113f;
  t179_n202: return 0.05297677963972092f;
  t179_n203: return 0.039815664291381836f;
  t179_n204: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t179_n214;
  t179_n205: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t179_n207;
  t179_n206: return 0.09068308025598526f;
  t179_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t179_n211;
  t179_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t179_n210;
  t179_n209: return -0.0034042755141854286f;
  t179_n210: return 0.03100557252764702f;
  t179_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t179_n213;
  t179_n212: return 0.010432545095682144f;
  t179_n213: return 0.04136763513088226f;
  t179_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t179_n228;
  t179_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t179_n227;
  t179_n216: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t179_n218;
  t179_n217: return 0.004255400504916906f;
  t179_n218: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t179_n224;
  t179_n219: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t179_n221;
  t179_n220: return -0.025147901847958565f;
  t179_n221: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t179_n223;
  t179_n222: return 0.06209517642855644f;
  t179_n223: return 0.106470488011837f;
  t179_n224: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t179_n226;
  t179_n225: return -0.13277308642864227f;
  t179_n226: return 0.011295419186353683f;
  t179_n227: return 0.0011927274754270911f;
  t179_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t179_n246;
  t179_n229: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t179_n233;
  t179_n230: if (E.getNumReferences() >= 3283042304 /*350.5*/) goto t179_n232;
  t179_n231: return -0.03176485374569893f;
  t179_n232: return -0.15406611561775208f;
  t179_n233: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t179_n245;
  t179_n234: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t179_n236;
  t179_n235: return -0.011153844185173512f;
  t179_n236: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n244;
  t179_n237: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t179_n239;
  t179_n238: return -0.04795226827263832f;
  t179_n239: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t179_n243;
  t179_n240: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t179_n242;
  t179_n241: return 0.09282737225294113f;
  t179_n242: return 0.056664250791072845f;
  t179_n243: return 0.009305290877819061f;
  t179_n244: return 0.054377760738134384f;
  t179_n245: return -0.24705249071121216f;
  t179_n246: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t179_n248;
  t179_n247: return 0.008498306386172771f;
  t179_n248: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t179_n252;
  t179_n249: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t179_n251;
  t179_n250: return -0.002184867626056075f;
  t179_n251: return 0.01059628650546074f;
  t179_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t179_n254;
  t179_n253: return 0.09919720888137817f;
  t179_n254: return -0.10599198192358017f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree180(const Example& E) {
  t180_n0: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t180_n24;
  t180_n1: if (E.getNumReferences() >= 3331449088 /*18672.5*/) goto t180_n3;
  t180_n2: return -0.00065999774960801f;
  t180_n3: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t180_n5;
  t180_n4: return -0.014742686413228512f;
  t180_n5: if (E.getNumReferences() >= 3331859712 /*19474.5*/) goto t180_n9;
  t180_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t180_n8;
  t180_n7: return 0.04257889837026596f;
  t180_n8: return 0.09071499854326248f;
  t180_n9: if (E.getNumReferences() >= 3331912192 /*19577.0*/) goto t180_n11;
  t180_n10: return -0.1952407956123352f;
  t180_n11: if (E.getNumReferences() >= 3332022528 /*19792.5*/) goto t180_n23;
  t180_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n14;
  t180_n13: return -0.22078531980514526f;
  t180_n14: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n22;
  t180_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t180_n17;
  t180_n16: return 0.1027805507183075f;
  t180_n17: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t180_n19;
  t180_n18: return 0.0970834419131279f;
  t180_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t180_n21;
  t180_n20: return 0.1159689724445343f;
  t180_n21: return 0.13063180446624756f;
  t180_n22: return -0.08012320846319199f;
  t180_n23: return 0.00518193282186985f;
  t180_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t180_n254;
  t180_n25: if (E.getNumReferences() >= 3298365440 /*1224.5*/) goto t180_n251;
  t180_n26: if (E.getFractionNameInContext() >= 3203243738 /*0.464102566242218*/) goto t180_n148;
  t180_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t180_n69;
  t180_n28: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t180_n40;
  t180_n29: if (E.getFractionNameInContext() >= 3182321783 /*0.08514492958784103*/) goto t180_n39;
  t180_n30: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n38;
  t180_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t180_n33;
  t180_n32: return 0.008704504929482937f;
  t180_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n37;
  t180_n34: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t180_n36;
  t180_n35: return -0.10899347811937332f;
  t180_n36: return 0.01950531266629696f;
  t180_n37: return 0.06600549817085266f;
  t180_n38: return -0.0158533938229084f;
  t180_n39: return -0.0010152297327294946f;
  t180_n40: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n66;
  t180_n41: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t180_n43;
  t180_n42: return -0.0729648768901825f;
  t180_n43: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t180_n63;
  t180_n44: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t180_n52;
  t180_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t180_n47;
  t180_n46: return 0.009782093577086926f;
  t180_n47: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t180_n51;
  t180_n48: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t180_n50;
  t180_n49: return -0.16429808735847473f;
  t180_n50: return 0.13996140658855438f;
  t180_n51: return -0.015950413420796394f;
  t180_n52: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t180_n62;
  t180_n53: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t180_n57;
  t180_n54: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t180_n56;
  t180_n55: return 0.05143182352185249f;
  t180_n56: return -0.0466998815536499f;
  t180_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t180_n61;
  t180_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n60;
  t180_n59: return -0.04629715532064438f;
  t180_n60: return 0.061347976326942444f;
  t180_n61: return 0.14060276746749878f;
  t180_n62: return 0.09268483519554138f;
  t180_n63: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t180_n65;
  t180_n64: return -0.12301360070705414f;
  t180_n65: return -0.01113086473196745f;
  t180_n66: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t180_n68;
  t180_n67: return -0.02480808086693287f;
  t180_n68: return 0.11621081084012985f;
  t180_n69: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t180_n119;
  t180_n70: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t180_n72;
  t180_n71: return -6.552321428898722e-05f;
  t180_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t180_n78;
  t180_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t180_n75;
  t180_n74: return 0.0058492072857916355f;
  t180_n75: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t180_n77;
  t180_n76: return -0.09570226818323135f;
  t180_n77: return 0.07629524916410446f;
  t180_n78: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t180_n80;
  t180_n79: return -0.06257551908493042f;
  t180_n80: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t180_n92;
  t180_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n83;
  t180_n82: return -0.05739797651767731f;
  t180_n83: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n85;
  t180_n84: return 0.09187004715204239f;
  t180_n85: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n87;
  t180_n86: return -0.041406095027923584f;
  t180_n87: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t180_n91;
  t180_n88: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t180_n90;
  t180_n89: return 0.09019263088703156f;
  t180_n90: return -0.06514861434698105f;
  t180_n91: return 0.08569924533367157f;
  t180_n92: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t180_n96;
  t180_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t180_n95;
  t180_n94: return -0.02477770857512951f;
  t180_n95: return 0.0699712485074997f;
  t180_n96: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t180_n110;
  t180_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n101;
  t180_n98: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t180_n100;
  t180_n99: return -0.11861295998096466f;
  t180_n100: return 0.1545436531305313f;
  t180_n101: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n103;
  t180_n102: return 0.03454658016562462f;
  t180_n103: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t180_n107;
  t180_n104: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t180_n106;
  t180_n105: return 0.03891140595078468f;
  t180_n106: return 0.07345255464315414f;
  t180_n107: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t180_n109;
  t180_n108: return 0.08216946572065353f;
  t180_n109: return -0.21767094731330872f;
  t180_n110: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t180_n114;
  t180_n111: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n113;
  t180_n112: return -0.05810777470469475f;
  t180_n113: return 0.03928137570619583f;
  t180_n114: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n116;
  t180_n115: return 0.05997772142291069f;
  t180_n116: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t180_n118;
  t180_n117: return -0.02211846597492695f;
  t180_n118: return -0.6577861309051514f;
  t180_n119: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n121;
  t180_n120: return 0.0046291230246424675f;
  t180_n121: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t180_n147;
  t180_n122: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t180_n138;
  t180_n123: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t180_n137;
  t180_n124: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t180_n126;
  t180_n125: return 0.0038503550458699465f;
  t180_n126: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n136;
  t180_n127: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t180_n135;
  t180_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t180_n132;
  t180_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t180_n131;
  t180_n130: return 0.08324556052684784f;
  t180_n131: return -0.06471014022827148f;
  t180_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t180_n134;
  t180_n133: return 0.10944059491157532f;
  t180_n134: return 0.12062802910804749f;
  t180_n135: return -0.04794185981154442f;
  t180_n136: return -0.025613509118556976f;
  t180_n137: return -0.0073595475405454636f;
  t180_n138: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t180_n142;
  t180_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t180_n141;
  t180_n140: return 0.06786924600601196f;
  t180_n141: return 0.11509273946285248f;
  t180_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n144;
  t180_n143: return -0.005795067176222801f;
  t180_n144: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t180_n146;
  t180_n145: return 0.11843878775835037f;
  t180_n146: return 0.09484029561281204f;
  t180_n147: return 0.029006537050008774f;
  t180_n148: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t180_n238;
  t180_n149: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t180_n165;
  t180_n150: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t180_n164;
  t180_n151: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n153;
  t180_n152: return 0.005864245351403952f;
  t180_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t180_n159;
  t180_n154: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t180_n156;
  t180_n155: return -0.024317026138305664f;
  t180_n156: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t180_n158;
  t180_n157: return 0.20823580026626587f;
  t180_n158: return -0.252867728471756f;
  t180_n159: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t180_n163;
  t180_n160: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n162;
  t180_n161: return 0.13700860738754272f;
  t180_n162: return 0.10127896815538406f;
  t180_n163: return 0.04740794748067856f;
  t180_n164: return -0.08074300736188889f;
  t180_n165: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t180_n199;
  t180_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t180_n178;
  t180_n167: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t180_n173;
  t180_n168: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n172;
  t180_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n171;
  t180_n170: return -0.06576202064752579f;
  t180_n171: return -0.02448451705276966f;
  t180_n172: return -0.00324549013748765f;
  t180_n173: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t180_n177;
  t180_n174: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n176;
  t180_n175: return -0.03424564376473427f;
  t180_n176: return 0.17899100482463837f;
  t180_n177: return -0.11440259218215942f;
  t180_n178: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n182;
  t180_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t180_n181;
  t180_n180: return 0.011059487238526344f;
  t180_n181: return 0.019864613190293312f;
  t180_n182: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t180_n198;
  t180_n183: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t180_n197;
  t180_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n190;
  t180_n185: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n187;
  t180_n186: return -0.061976365745067596f;
  t180_n187: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t180_n189;
  t180_n188: return -0.012933227233588696f;
  t180_n189: return 0.06364748626947403f;
  t180_n190: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t180_n196;
  t180_n191: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t180_n193;
  t180_n192: return -0.06349991261959076f;
  t180_n193: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t180_n195;
  t180_n194: return 0.05138997733592987f;
  t180_n195: return -0.05077631399035454f;
  t180_n196: return -0.10068222880363464f;
  t180_n197: return -0.3651650547981262f;
  t180_n198: return 0.3011763393878937f;
  t180_n199: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t180_n235;
  t180_n200: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t180_n234;
  t180_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t180_n227;
  t180_n202: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t180_n220;
  t180_n203: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t180_n217;
  t180_n204: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t180_n214;
  t180_n205: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t180_n213;
  t180_n206: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t180_n208;
  t180_n207: return -0.08989942818880081f;
  t180_n208: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t180_n210;
  t180_n209: return -0.2447628378868103f;
  t180_n210: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t180_n212;
  t180_n211: return -0.2203143835067749f;
  t180_n212: return 0.050873901695013046f;
  t180_n213: return -0.024652604013681412f;
  t180_n214: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t180_n216;
  t180_n215: return 0.019478969275951385f;
  t180_n216: return -0.6022821068763733f;
  t180_n217: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t180_n219;
  t180_n218: return 0.11968585848808289f;
  t180_n219: return -0.005590181797742844f;
  t180_n220: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t180_n224;
  t180_n221: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t180_n223;
  t180_n222: return -0.010157149285078049f;
  t180_n223: return 0.07594645023345947f;
  t180_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n226;
  t180_n225: return -0.12475567311048508f;
  t180_n226: return 0.1340324878692627f;
  t180_n227: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t180_n229;
  t180_n228: return -0.015436580404639244f;
  t180_n229: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t180_n233;
  t180_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t180_n232;
  t180_n231: return -0.23353266716003418f;
  t180_n232: return 0.08659645915031433f;
  t180_n233: return 0.09353069961071014f;
  t180_n234: return -0.006893177982419729f;
  t180_n235: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t180_n237;
  t180_n236: return 0.012378011830151081f;
  t180_n237: return -0.027063410729169846f;
  t180_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t180_n250;
  t180_n239: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t180_n241;
  t180_n240: return 0.0009096846915781498f;
  t180_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t180_n243;
  t180_n242: return 0.01737179048359394f;
  t180_n243: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t180_n249;
  t180_n244: if (E.getNumReferences() >= 3244294144 /*14.0*/) goto t180_n248;
  t180_n245: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t180_n247;
  t180_n246: return -0.09490416198968887f;
  t180_n247: return 0.11715614795684814f;
  t180_n248: return 0.13307616114616394f;
  t180_n249: return -0.031860291957855225f;
  t180_n250: return 0.025388484820723534f;
  t180_n251: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t180_n253;
  t180_n252: return 0.011246765963733196f;
  t180_n253: return 0.04411713778972626f;
  t180_n254: return 0.002141169970855117f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree181(const Example& E) {
  t181_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t181_n76;
  t181_n1: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t181_n73;
  t181_n2: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t181_n62;
  t181_n3: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t181_n23;
  t181_n4: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t181_n22;
  t181_n5: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t181_n21;
  t181_n6: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t181_n20;
  t181_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n19;
  t181_n8: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t181_n10;
  t181_n9: return -0.0010581054957583547f;
  t181_n10: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t181_n12;
  t181_n11: return 0.004529754165560007f;
  t181_n12: if (E.getNumReferences() >= 3267166208 /*94.5*/) goto t181_n14;
  t181_n13: return 0.028017038479447365f;
  t181_n14: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t181_n18;
  t181_n15: if (E.getNumReferences() >= 3270377472 /*119.0*/) goto t181_n17;
  t181_n16: return 0.30754318833351135f;
  t181_n17: return -0.104608453810215f;
  t181_n18: return 0.08396939933300018f;
  t181_n19: return 0.006979840341955423f;
  t181_n20: return -0.00689384900033474f;
  t181_n21: return 0.006008528638631105f;
  t181_n22: return -0.0013387203216552734f;
  t181_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n31;
  t181_n24: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t181_n26;
  t181_n25: return -0.024380140006542206f;
  t181_n26: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t181_n28;
  t181_n27: return -0.1832636296749115f;
  t181_n28: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t181_n30;
  t181_n29: return 0.10395606607198715f;
  t181_n30: return 0.02833181992173195f;
  t181_n31: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t181_n41;
  t181_n32: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t181_n38;
  t181_n33: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t181_n37;
  t181_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t181_n36;
  t181_n35: return -0.18022198975086212f;
  t181_n36: return -0.031980205327272415f;
  t181_n37: return -0.004524445626884699f;
  t181_n38: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t181_n40;
  t181_n39: return 0.01320897787809372f;
  t181_n40: return 0.0664108619093895f;
  t181_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n61;
  t181_n42: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t181_n52;
  t181_n43: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t181_n51;
  t181_n44: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t181_n48;
  t181_n45: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t181_n47;
  t181_n46: return 0.0500173382461071f;
  t181_n47: return 0.010038836859166622f;
  t181_n48: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t181_n50;
  t181_n49: return 0.11533814668655396f;
  t181_n50: return 0.037799324840307236f;
  t181_n51: return 0.010085963644087315f;
  t181_n52: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t181_n60;
  t181_n53: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t181_n59;
  t181_n54: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t181_n58;
  t181_n55: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t181_n57;
  t181_n56: return 0.1265457570552826f;
  t181_n57: return -0.010307964868843555f;
  t181_n58: return 0.06303707510232925f;
  t181_n59: return -0.08472958952188492f;
  t181_n60: return 0.00911420863121748f;
  t181_n61: return 0.024906432256102562f;
  t181_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t181_n64;
  t181_n63: return -0.03604986518621445f;
  t181_n64: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t181_n68;
  t181_n65: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t181_n67;
  t181_n66: return 0.04403195157647133f;
  t181_n67: return 0.09783259779214859f;
  t181_n68: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t181_n70;
  t181_n69: return -0.061156272888183594f;
  t181_n70: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t181_n72;
  t181_n71: return 0.10553991794586182f;
  t181_n72: return 0.1341058760881424f;
  t181_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n75;
  t181_n74: return -0.04588254541158676f;
  t181_n75: return 0.04291743040084839f;
  t181_n76: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t181_n230;
  t181_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t181_n155;
  t181_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t181_n82;
  t181_n79: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t181_n81;
  t181_n80: return -0.05184672772884369f;
  t181_n81: return -0.01434377022087574f;
  t181_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t181_n122;
  t181_n83: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t181_n119;
  t181_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t181_n102;
  t181_n85: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t181_n93;
  t181_n86: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t181_n92;
  t181_n87: if (E.getNumReferences() >= 3285090304 /*413.0*/) goto t181_n89;
  t181_n88: return 0.013088559731841087f;
  t181_n89: if (E.getNumReferences() >= 3326212096 /*12414.0*/) goto t181_n91;
  t181_n90: return -0.3006678819656372f;
  t181_n91: return 0.15888351202011108f;
  t181_n92: return -0.04210877791047096f;
  t181_n93: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t181_n101;
  t181_n94: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t181_n100;
  t181_n95: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t181_n97;
  t181_n96: return -0.022681020200252533f;
  t181_n97: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t181_n99;
  t181_n98: return 0.007223787717521191f;
  t181_n99: return -0.016238199546933174f;
  t181_n100: return -0.1602701097726822f;
  t181_n101: return 0.008382860571146011f;
  t181_n102: if (E.getNumReferences() >= 3315876864 /*5256.5*/) goto t181_n114;
  t181_n103: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t181_n111;
  t181_n104: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t181_n110;
  t181_n105: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t181_n109;
  t181_n106: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t181_n108;
  t181_n107: return -0.05971542000770569f;
  t181_n108: return 0.015739304944872856f;
  t181_n109: return 0.03676745295524597f;
  t181_n110: return -0.032401349395513535f;
  t181_n111: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t181_n113;
  t181_n112: return 0.06366453319787979f;
  t181_n113: return -0.024086156859993935f;
  t181_n114: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t181_n118;
  t181_n115: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t181_n117;
  t181_n116: return 0.10412414371967316f;
  t181_n117: return 0.13150648772716522f;
  t181_n118: return 0.058165792375802994f;
  t181_n119: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t181_n121;
  t181_n120: return 0.004211776424199343f;
  t181_n121: return -0.007097771856933832f;
  t181_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n134;
  t181_n123: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t181_n125;
  t181_n124: return -0.03312718868255615f;
  t181_n125: if (E.getNumReferences() >= 3309658112 /*3158.0*/) goto t181_n133;
  t181_n126: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t181_n132;
  t181_n127: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t181_n129;
  t181_n128: return 0.0051159546710550785f;
  t181_n129: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t181_n131;
  t181_n130: return -0.03991421312093735f;
  t181_n131: return 0.05248638242483139f;
  t181_n132: return 0.04627446457743645f;
  t181_n133: return 0.04025118052959442f;
  t181_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t181_n136;
  t181_n135: return -0.0033412768971174955f;
  t181_n136: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t181_n154;
  t181_n137: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t181_n149;
  t181_n138: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t181_n148;
  t181_n139: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t181_n145;
  t181_n140: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t181_n142;
  t181_n141: return 0.005362818483263254f;
  t181_n142: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t181_n144;
  t181_n143: return -0.0006327320006676018f;
  t181_n144: return 0.02905306965112686f;
  t181_n145: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t181_n147;
  t181_n146: return 0.024691395461559296f;
  t181_n147: return -0.25143399834632874f;
  t181_n148: return -5.1228100346634164e-05f;
  t181_n149: if (E.getFractionNameInContext() >= 3179644278 /*0.0651959627866745*/) goto t181_n153;
  t181_n150: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t181_n152;
  t181_n151: return 0.036113373935222626f;
  t181_n152: return 0.09257243573665619f;
  t181_n153: return -0.07880876213312149f;
  t181_n154: return -0.02044200897216797f;
  t181_n155: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t181_n167;
  t181_n156: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t181_n162;
  t181_n157: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t181_n159;
  t181_n158: return 0.11595156788825989f;
  t181_n159: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t181_n161;
  t181_n160: return 0.004518447909504175f;
  t181_n161: return 0.01470186561346054f;
  t181_n162: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t181_n164;
  t181_n163: return 0.015744240954518318f;
  t181_n164: if (E.getNumReferences() >= 3269525504 /*112.5*/) goto t181_n166;
  t181_n165: return 0.18734273314476013f;
  t181_n166: return 0.08123613893985748f;
  t181_n167: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t181_n169;
  t181_n168: return 0.0010039054322987795f;
  t181_n169: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t181_n205;
  t181_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n196;
  t181_n171: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t181_n191;
  t181_n172: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t181_n184;
  t181_n173: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t181_n183;
  t181_n174: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t181_n178;
  t181_n175: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t181_n177;
  t181_n176: return 0.10231778770685196f;
  t181_n177: return 0.13139772415161133f;
  t181_n178: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t181_n180;
  t181_n179: return 0.053057145327329636f;
  t181_n180: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t181_n182;
  t181_n181: return 0.09718929976224899f;
  t181_n182: return -0.11611230671405792f;
  t181_n183: return -0.19177962839603424f;
  t181_n184: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n188;
  t181_n185: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t181_n187;
  t181_n186: return -0.10194674134254456f;
  t181_n187: return -0.022579673677682877f;
  t181_n188: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t181_n190;
  t181_n189: return 0.09307758510112762f;
  t181_n190: return -0.029095491394400597f;
  t181_n191: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t181_n193;
  t181_n192: return 0.039976805448532104f;
  t181_n193: if (E.getFractionNameInContext() >= 3209574628 /*0.8055555820465088*/) goto t181_n195;
  t181_n194: return -0.05202426761388779f;
  t181_n195: return 0.16022884845733643f;
  t181_n196: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n198;
  t181_n197: return 0.018101686611771584f;
  t181_n198: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t181_n204;
  t181_n199: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t181_n203;
  t181_n200: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t181_n202;
  t181_n201: return 0.10099556297063828f;
  t181_n202: return -0.1798289269208908f;
  t181_n203: return 0.11454004049301147f;
  t181_n204: return 0.04015068709850311f;
  t181_n205: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t181_n229;
  t181_n206: if (E.getFractionNameInContext() >= 3199399908 /*0.34954750537872314*/) goto t181_n208;
  t181_n207: return -0.16931861639022827f;
  t181_n208: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t181_n228;
  t181_n209: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t181_n213;
  t181_n210: if (E.getFractionNameInContext() >= 3199954359 /*0.3660714328289032*/) goto t181_n212;
  t181_n211: return 0.11340714991092682f;
  t181_n212: return -0.008279876783490181f;
  t181_n213: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t181_n219;
  t181_n214: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n216;
  t181_n215: return 0.12278192490339279f;
  t181_n216: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t181_n218;
  t181_n217: return 0.08369173109531403f;
  t181_n218: return 0.1324503868818283f;
  t181_n219: if (E.getFractionNameInContext() >= 3209574628 /*0.8055555820465088*/) goto t181_n223;
  t181_n220: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t181_n222;
  t181_n221: return 0.04522503539919853f;
  t181_n222: return -0.39501702785491943f;
  t181_n223: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n227;
  t181_n224: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t181_n226;
  t181_n225: return 0.1278865933418274f;
  t181_n226: return 0.10917524993419647f;
  t181_n227: return -0.020361049100756645f;
  t181_n228: return 0.1247820034623146f;
  t181_n229: return 0.020253712311387062f;
  t181_n230: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t181_n252;
  t181_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t181_n237;
  t181_n232: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n234;
  t181_n233: return -0.012115576304495335f;
  t181_n234: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t181_n236;
  t181_n235: return 0.009326636791229248f;
  t181_n236: return 0.07961700111627579f;
  t181_n237: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t181_n239;
  t181_n238: return 0.007309332489967346f;
  t181_n239: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t181_n245;
  t181_n240: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t181_n242;
  t181_n241: return 0.010955072939395905f;
  t181_n242: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t181_n244;
  t181_n243: return 0.043167226016521454f;
  t181_n244: return 0.10134225338697433f;
  t181_n245: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t181_n251;
  t181_n246: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t181_n250;
  t181_n247: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t181_n249;
  t181_n248: return 0.07517065852880478f;
  t181_n249: return 0.1479707956314087f;
  t181_n250: return -0.2140721082687378f;
  t181_n251: return 0.06096900999546051f;
  t181_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t181_n254;
  t181_n253: return -0.015817271545529366f;
  t181_n254: return 0.0021935501135885715f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree182(const Example& E) {
  t182_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t182_n254;
  t182_n1: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t182_n3;
  t182_n2: return -0.0003913131367880851f;
  t182_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t182_n5;
  t182_n4: return 0.009567167609930038f;
  t182_n5: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t182_n9;
  t182_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n8;
  t182_n7: return -0.13377737998962402f;
  t182_n8: return 0.13042590022087097f;
  t182_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n15;
  t182_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t182_n12;
  t182_n11: return -0.019109368324279785f;
  t182_n12: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t182_n14;
  t182_n13: return 0.03796868398785591f;
  t182_n14: return -0.12260942906141281f;
  t182_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t182_n85;
  t182_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n84;
  t182_n17: if (E.getNumReferences() >= 3322564096 /*8851.5*/) goto t182_n81;
  t182_n18: if (E.getNumReferences() >= 3322543616 /*8831.5*/) goto t182_n76;
  t182_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n21;
  t182_n20: return -0.05741146579384804f;
  t182_n21: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t182_n23;
  t182_n22: return -0.04509132727980614f;
  t182_n23: if (E.getNumReferences() >= 3306690560 /*2433.5*/) goto t182_n29;
  t182_n24: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t182_n28;
  t182_n25: if (E.getNumReferences() >= 3306643456 /*2422.0*/) goto t182_n27;
  t182_n26: return 0.10913856327533722f;
  t182_n27: return 0.2382049560546875f;
  t182_n28: return 0.08716267347335815f;
  t182_n29: if (E.getNumReferences() >= 3311417344 /*3587.5*/) goto t182_n31;
  t182_n30: return -0.07997392863035202f;
  t182_n31: if (E.getNumReferences() >= 3317616640 /*6106.0*/) goto t182_n69;
  t182_n32: if (E.getNumReferences() >= 3317563392 /*6080.0*/) goto t182_n54;
  t182_n33: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t182_n53;
  t182_n34: if (E.getNumReferences() >= 3314141184 /*4409.0*/) goto t182_n52;
  t182_n35: if (E.getNumReferences() >= 3314128896 /*4403.0*/) goto t182_n51;
  t182_n36: if (E.getNumReferences() >= 3313586176 /*4138.0*/) goto t182_n44;
  t182_n37: if (E.getNumReferences() >= 3313168384 /*4015.0*/) goto t182_n43;
  t182_n38: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t182_n40;
  t182_n39: return 0.12268286198377609f;
  t182_n40: if (E.getNumReferences() >= 3311990784 /*3727.5*/) goto t182_n42;
  t182_n41: return 0.13652141392230988f;
  t182_n42: return -0.12533320486545563f;
  t182_n43: return -0.12407654523849487f;
  t182_n44: if (E.getNumReferences() >= 3313673216 /*4180.5*/) goto t182_n46;
  t182_n45: return 0.1476014107465744f;
  t182_n46: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t182_n50;
  t182_n47: if (E.getNumReferences() >= 3313934336 /*4308.0*/) goto t182_n49;
  t182_n48: return 0.11451612412929535f;
  t182_n49: return -0.06223750114440918f;
  t182_n50: return -0.07059740275144577f;
  t182_n51: return 0.16606608033180237f;
  t182_n52: return 0.0558781772851944f;
  t182_n53: return -0.008018720895051956f;
  t182_n54: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t182_n66;
  t182_n55: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t182_n57;
  t182_n56: return -0.10544838756322861f;
  t182_n57: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t182_n59;
  t182_n58: return 0.3622540235519409f;
  t182_n59: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t182_n65;
  t182_n60: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t182_n64;
  t182_n61: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t182_n63;
  t182_n62: return -0.10785926878452301f;
  t182_n63: return 0.16530495882034302f;
  t182_n64: return 0.7965317964553833f;
  t182_n65: return -0.11962717771530151f;
  t182_n66: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t182_n68;
  t182_n67: return 0.11972498893737793f;
  t182_n68: return 0.16436079144477844f;
  t182_n69: if (E.getNumReferences() >= 3319041024 /*6801.5*/) goto t182_n71;
  t182_n70: return -0.10107363015413284f;
  t182_n71: if (E.getNumReferences() >= 3319080960 /*6821.0*/) goto t182_n75;
  t182_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t182_n74;
  t182_n73: return 0.21302706003189087f;
  t182_n74: return 0.2534198462963104f;
  t182_n75: return 0.04482494667172432f;
  t182_n76: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t182_n78;
  t182_n77: return -0.11082567274570465f;
  t182_n78: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t182_n80;
  t182_n79: return 0.15374435484409332f;
  t182_n80: return 0.22999724745750427f;
  t182_n81: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t182_n83;
  t182_n82: return 0.007475765887647867f;
  t182_n83: return -0.07305506616830826f;
  t182_n84: return 0.0027351484168320894f;
  t182_n85: if (E.getNumReferences() >= 3316164608 /*5397.0*/) goto t182_n165;
  t182_n86: if (E.getNumReferences() >= 3315997696 /*5315.5*/) goto t182_n160;
  t182_n87: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t182_n89;
  t182_n88: return 0.1608429253101349f;
  t182_n89: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t182_n135;
  t182_n90: if (E.getNumReferences() >= 3312912384 /*3952.5*/) goto t182_n130;
  t182_n91: if (E.getNumReferences() >= 3310782464 /*3432.5*/) goto t182_n101;
  t182_n92: if (E.getNumReferences() >= 3306967040 /*2501.0*/) goto t182_n100;
  t182_n93: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t182_n99;
  t182_n94: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t182_n98;
  t182_n95: if (E.getNumReferences() >= 3306846208 /*2471.5*/) goto t182_n97;
  t182_n96: return 0.12374068051576614f;
  t182_n97: return 0.11720442771911621f;
  t182_n98: return 0.013177946209907532f;
  t182_n99: return 0.13112479448318481f;
  t182_n100: return 0.00012710060400422662f;
  t182_n101: if (E.getNumReferences() >= 3312549888 /*3864.0*/) goto t182_n129;
  t182_n102: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t182_n114;
  t182_n103: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t182_n109;
  t182_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t182_n106;
  t182_n105: return 0.04164163023233414f;
  t182_n106: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t182_n108;
  t182_n107: return 0.39083996415138245f;
  t182_n108: return -0.011308778077363968f;
  t182_n109: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t182_n111;
  t182_n110: return 0.17429755628108978f;
  t182_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n113;
  t182_n112: return -0.5117800235748291f;
  t182_n113: return 0.1388203501701355f;
  t182_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t182_n116;
  t182_n115: return 0.06393085420131683f;
  t182_n116: if (E.getNumReferences() >= 3311859712 /*3695.5*/) goto t182_n128;
  t182_n117: if (E.getNumReferences() >= 3311824896 /*3687.0*/) goto t182_n127;
  t182_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t182_n126;
  t182_n119: if (E.getNumReferences() >= 3311169536 /*3527.0*/) goto t182_n125;
  t182_n120: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t182_n124;
  t182_n121: if (E.getFractionNameInContext() >= 3185993318 /*0.11249999701976776*/) goto t182_n123;
  t182_n122: return 0.3291146457195282f;
  t182_n123: return 0.13048957288265228f;
  t182_n124: return 0.12637624144554138f;
  t182_n125: return -0.01014040969312191f;
  t182_n126: return 0.1441708356142044f;
  t182_n127: return 0.15062099695205688f;
  t182_n128: return 0.06367719173431396f;
  t182_n129: return -0.0399707555770874f;
  t182_n130: if (E.getNumReferences() >= 3313016832 /*3978.0*/) goto t182_n132;
  t182_n131: return -0.22241167724132538f;
  t182_n132: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t182_n134;
  t182_n133: return 0.014675647020339966f;
  t182_n134: return -0.39005568623542786f;
  t182_n135: if (E.getNumReferences() >= 3315294208 /*4972.0*/) goto t182_n159;
  t182_n136: if (E.getNumReferences() >= 3315181568 /*4917.0*/) goto t182_n156;
  t182_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t182_n139;
  t182_n138: return 0.06066347658634186f;
  t182_n139: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t182_n143;
  t182_n140: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t182_n142;
  t182_n141: return 0.004165453836321831f;
  t182_n142: return 0.15703009068965912f;
  t182_n143: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t182_n153;
  t182_n144: if (E.getNumReferences() >= 3314460672 /*4565.0*/) goto t182_n146;
  t182_n145: return 0.04937734082341194f;
  t182_n146: if (E.getNumReferences() >= 3314592768 /*4629.5*/) goto t182_n152;
  t182_n147: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t182_n151;
  t182_n148: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t182_n150;
  t182_n149: return 0.10648380219936371f;
  t182_n150: return 0.13811659812927246f;
  t182_n151: return 0.045170027762651443f;
  t182_n152: return 0.0877143070101738f;
  t182_n153: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t182_n155;
  t182_n154: return 0.1643393337726593f;
  t182_n155: return -0.27603012323379517f;
  t182_n156: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t182_n158;
  t182_n157: return 0.14552117884159088f;
  t182_n158: return 0.10348332673311234f;
  t182_n159: return -0.012874899432063103f;
  t182_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t182_n162;
  t182_n161: return 0.10674437135457993f;
  t182_n162: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t182_n164;
  t182_n163: return 0.12790629267692566f;
  t182_n164: return -0.10092984884977341f;
  t182_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n167;
  t182_n166: return -0.003072714665904641f;
  t182_n167: if (E.getNumReferences() >= 3316531200 /*5576.0*/) goto t182_n169;
  t182_n168: return -0.16860315203666687f;
  t182_n169: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t182_n253;
  t182_n170: if (E.getNumReferences() >= 3349284736 /*82959.0*/) goto t182_n242;
  t182_n171: if (E.getNumReferences() >= 3340753024 /*40920.5*/) goto t182_n241;
  t182_n172: if (E.getNumReferences() >= 3339042560 /*34239.0*/) goto t182_n236;
  t182_n173: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t182_n179;
  t182_n174: if (E.getNumReferences() >= 3336251648 /*28052.5*/) goto t182_n176;
  t182_n175: return -0.04435020312666893f;
  t182_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t182_n178;
  t182_n177: return 0.10490526258945465f;
  t182_n178: return 0.13177964091300964f;
  t182_n179: if (E.getNumReferences() >= 3332389888 /*20510.0*/) goto t182_n219;
  t182_n180: if (E.getNumReferences() >= 3329713152 /*15833.0*/) goto t182_n184;
  t182_n181: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t182_n183;
  t182_n182: return 0.029360469430685043f;
  t182_n183: return -0.009850375354290009f;
  t182_n184: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t182_n218;
  t182_n185: if (E.getNumReferences() >= 3331040000 /*17873.5*/) goto t182_n205;
  t182_n186: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t182_n202;
  t182_n187: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t182_n189;
  t182_n188: return 0.0216293353587389f;
  t182_n189: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t182_n197;
  t182_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n192;
  t182_n191: return -0.19826802611351013f;
  t182_n192: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t182_n196;
  t182_n193: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t182_n195;
  t182_n194: return 0.11848154664039612f;
  t182_n195: return 0.1322007179260254f;
  t182_n196: return 0.1271970421075821f;
  t182_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t182_n199;
  t182_n198: return 0.04631010815501213f;
  t182_n199: if (E.getNumReferences() >= 3330071040 /*16182.5*/) goto t182_n201;
  t182_n200: return 0.1393010914325714f;
  t182_n201: return 0.10330194979906082f;
  t182_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n204;
  t182_n203: return -0.10277081280946732f;
  t182_n204: return 0.14106380939483643f;
  t182_n205: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t182_n217;
  t182_n206: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t182_n208;
  t182_n207: return 0.0189013984054327f;
  t182_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t182_n210;
  t182_n209: return -0.04394647851586342f;
  t182_n210: if (E.getNumReferences() >= 3331443968 /*18662.5*/) goto t182_n212;
  t182_n211: return 0.02588443085551262f;
  t182_n212: if (E.getNumReferences() >= 3331489536 /*18751.5*/) goto t182_n214;
  t182_n213: return 0.19101057946681976f;
  t182_n214: if (E.getNumReferences() >= 3332227072 /*20192.0*/) goto t182_n216;
  t182_n215: return 0.05733737722039223f;
  t182_n216: return 0.13662715256214142f;
  t182_n217: return 0.013822384178638458f;
  t182_n218: return 0.21404483914375305f;
  t182_n219: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t182_n231;
  t182_n220: if (E.getNumReferences() >= 3335347712 /*26287.0*/) goto t182_n222;
  t182_n221: return -0.03573218733072281f;
  t182_n222: if (E.getNumReferences() >= 3335779328 /*27130.0*/) goto t182_n228;
  t182_n223: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t182_n225;
  t182_n224: return -0.005060578230768442f;
  t182_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t182_n227;
  t182_n226: return 0.05902621150016785f;
  t182_n227: return 0.12923820316791534f;
  t182_n228: if (E.getNumReferences() >= 3338878720 /*33599.0*/) goto t182_n230;
  t182_n229: return 0.032432038336992264f;
  t182_n230: return -0.061570726335048676f;
  t182_n231: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t182_n233;
  t182_n232: return -0.1314811408519745f;
  t182_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t182_n235;
  t182_n234: return 0.12106561660766602f;
  t182_n235: return -0.003595477668568492f;
  t182_n236: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t182_n238;
  t182_n237: return -0.6152410507202148f;
  t182_n238: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t182_n240;
  t182_n239: return 0.04516613855957985f;
  t182_n240: return 0.08176279067993164f;
  t182_n241: return -0.014532545581459999f;
  t182_n242: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t182_n244;
  t182_n243: return -0.13563276827335358f;
  t182_n244: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t182_n246;
  t182_n245: return 0.00932010542601347f;
  t182_n246: if (E.getNumReferences() >= 3352491648 /*108013.0*/) goto t182_n252;
  t182_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t182_n249;
  t182_n248: return 0.013976935297250748f;
  t182_n249: if (E.getNumReferences() >= 3350654080 /*93657.0*/) goto t182_n251;
  t182_n250: return 0.04799894988536835f;
  t182_n251: return 0.08618427813053131f;
  t182_n252: return 0.024868043139576912f;
  t182_n253: return -0.004807665478438139f;
  t182_n254: return 0.0015979873714968562f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree183(const Example& E) {
  t183_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t183_n192;
  t183_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t183_n137;
  t183_n2: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t183_n116;
  t183_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t183_n5;
  t183_n4: return -0.0009163310169242322f;
  t183_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t183_n91;
  t183_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t183_n20;
  t183_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n19;
  t183_n8: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t183_n10;
  t183_n9: return -0.018060006201267242f;
  t183_n10: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t183_n12;
  t183_n11: return -0.08645346760749817f;
  t183_n12: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n18;
  t183_n13: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t183_n17;
  t183_n14: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t183_n16;
  t183_n15: return 0.12486733496189117f;
  t183_n16: return 0.07915075868368149f;
  t183_n17: return 0.11495240777730942f;
  t183_n18: return -0.03020651452243328f;
  t183_n19: return 0.0014847440179437399f;
  t183_n20: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t183_n76;
  t183_n21: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t183_n75;
  t183_n22: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t183_n74;
  t183_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n33;
  t183_n24: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n26;
  t183_n25: return -0.00632390845566988f;
  t183_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n32;
  t183_n27: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t183_n31;
  t183_n28: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t183_n30;
  t183_n29: return -0.019138844683766365f;
  t183_n30: return -0.09516719728708267f;
  t183_n31: return 0.027088375762104988f;
  t183_n32: return 0.030522974207997322f;
  t183_n33: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t183_n35;
  t183_n34: return -0.029634321108460426f;
  t183_n35: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t183_n55;
  t183_n36: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t183_n48;
  t183_n37: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t183_n45;
  t183_n38: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t183_n42;
  t183_n39: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t183_n41;
  t183_n40: return -0.05832929164171219f;
  t183_n41: return 0.010159398429095745f;
  t183_n42: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t183_n44;
  t183_n43: return 0.030305059626698494f;
  t183_n44: return -0.149762824177742f;
  t183_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n47;
  t183_n46: return -0.04672861844301224f;
  t183_n47: return 0.04600830376148224f;
  t183_n48: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t183_n54;
  t183_n49: if (E.getNumReferences() >= 3260547072 /*54.0*/) goto t183_n53;
  t183_n50: if (E.getNumReferences() >= 3247964160 /*19.0*/) goto t183_n52;
  t183_n51: return 0.020610973238945007f;
  t183_n52: return 0.14344090223312378f;
  t183_n53: return -1.5098541975021362f;
  t183_n54: return 0.07659611105918884f;
  t183_n55: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t183_n71;
  t183_n56: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t183_n62;
  t183_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n59;
  t183_n58: return -0.06350986659526825f;
  t183_n59: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t183_n61;
  t183_n60: return -0.28119245171546936f;
  t183_n61: return 0.07766437530517578f;
  t183_n62: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n64;
  t183_n63: return -0.004270222503691912f;
  t183_n64: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t183_n66;
  t183_n65: return -0.6448265314102173f;
  t183_n66: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t183_n70;
  t183_n67: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t183_n69;
  t183_n68: return 0.057095061987638474f;
  t183_n69: return 0.02622651681303978f;
  t183_n70: return 0.0025858357548713684f;
  t183_n71: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t183_n73;
  t183_n72: return -0.2072470337152481f;
  t183_n73: return -0.055284932255744934f;
  t183_n74: return -0.0287215244024992f;
  t183_n75: return 0.0023672578390687704f;
  t183_n76: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t183_n84;
  t183_n77: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t183_n81;
  t183_n78: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t183_n80;
  t183_n79: return 0.004149146378040314f;
  t183_n80: return -0.07022861391305923f;
  t183_n81: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t183_n83;
  t183_n82: return 0.07267820090055466f;
  t183_n83: return 0.12293075770139694f;
  t183_n84: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t183_n86;
  t183_n85: return 0.026273507624864578f;
  t183_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace))) goto t183_n88;
  t183_n87: return -0.011107471771538258f;
  t183_n88: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t183_n90;
  t183_n89: return 0.039380788803100586f;
  t183_n90: return 0.12004078179597855f;
  t183_n91: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t183_n95;
  t183_n92: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t183_n94;
  t183_n93: return 0.02859276719391346f;
  t183_n94: return -0.01234542578458786f;
  t183_n95: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t183_n115;
  t183_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n98;
  t183_n97: return -0.024802224710583687f;
  t183_n98: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t183_n104;
  t183_n99: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t183_n101;
  t183_n100: return 0.007166189141571522f;
  t183_n101: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t183_n103;
  t183_n102: return -0.5028961300849915f;
  t183_n103: return 0.1092158854007721f;
  t183_n104: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t183_n106;
  t183_n105: return -0.016295308247208595f;
  t183_n106: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t183_n114;
  t183_n107: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t183_n113;
  t183_n108: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t183_n112;
  t183_n109: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t183_n111;
  t183_n110: return 0.12986735999584198f;
  t183_n111: return 0.19555439054965973f;
  t183_n112: return -0.18002960085868835f;
  t183_n113: return -0.5256078243255615f;
  t183_n114: return 0.05898451805114746f;
  t183_n115: return 0.0889640673995018f;
  t183_n116: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t183_n136;
  t183_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n119;
  t183_n118: return -0.014782804995775223f;
  t183_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n121;
  t183_n120: return 0.006811150349676609f;
  t183_n121: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t183_n135;
  t183_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t183_n124;
  t183_n123: return 0.007954234257340431f;
  t183_n124: if (E.getNumReferences() >= 3274145792 /*167.5*/) goto t183_n134;
  t183_n125: if (E.getNumReferences() >= 3271786496 /*131.5*/) goto t183_n129;
  t183_n126: if (E.getNumReferences() >= 3271491584 /*127.5*/) goto t183_n128;
  t183_n127: return 0.1814536452293396f;
  t183_n128: return -0.13023020327091217f;
  t183_n129: if (E.getNumReferences() >= 3273326592 /*155.0*/) goto t183_n133;
  t183_n130: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t183_n132;
  t183_n131: return 0.15358686447143555f;
  t183_n132: return 0.14179562032222748f;
  t183_n133: return 0.1623256653547287f;
  t183_n134: return 0.05421781167387962f;
  t183_n135: return 0.019176477566361427f;
  t183_n136: return 0.0003629112907219678f;
  t183_n137: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t183_n177;
  t183_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t183_n164;
  t183_n139: if (E.getNumReferences() >= 3364626832 /*286988.5*/) goto t183_n163;
  t183_n140: if (E.getNumReferences() >= 3352044608 /*104520.5*/) goto t183_n156;
  t183_n141: if (E.getNumReferences() >= 3318942720 /*6753.5*/) goto t183_n145;
  t183_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n144;
  t183_n143: return -0.14275701344013214f;
  t183_n144: return 0.003444362198933959f;
  t183_n145: if (E.getNumReferences() >= 3330759936 /*17326.5*/) goto t183_n147;
  t183_n146: return -0.09011740237474442f;
  t183_n147: if (E.getNumReferences() >= 3331140352 /*18069.5*/) goto t183_n155;
  t183_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t183_n150;
  t183_n149: return 0.03778243064880371f;
  t183_n150: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t183_n154;
  t183_n151: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t183_n153;
  t183_n152: return 0.11611520498991013f;
  t183_n153: return 0.19437475502490997f;
  t183_n154: return 0.07959244400262833f;
  t183_n155: return -0.0037595811299979687f;
  t183_n156: if (E.getNumReferences() >= 3354468928 /*123460.5*/) goto t183_n162;
  t183_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t183_n159;
  t183_n158: return 0.05241771042346954f;
  t183_n159: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t183_n161;
  t183_n160: return 0.13493719696998596f;
  t183_n161: return 0.08965781331062317f;
  t183_n162: return 0.03457871079444885f;
  t183_n163: return -0.03731849044561386f;
  t183_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t183_n174;
  t183_n165: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t183_n171;
  t183_n166: if (E.getNumReferences() >= 3266838528 /*92.0*/) goto t183_n170;
  t183_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n169;
  t183_n168: return -0.04463495686650276f;
  t183_n169: return 0.011049703694880009f;
  t183_n170: return 0.04226791113615036f;
  t183_n171: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t183_n173;
  t183_n172: return 0.0392163060605526f;
  t183_n173: return -0.07544487714767456f;
  t183_n174: if (E.getNumReferences() >= 3264937984 /*77.5*/) goto t183_n176;
  t183_n175: return 0.10323992371559143f;
  t183_n176: return -0.08351227641105652f;
  t183_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t183_n185;
  t183_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n184;
  t183_n179: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t183_n183;
  t183_n180: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t183_n182;
  t183_n181: return -0.06683365255594254f;
  t183_n182: return 0.029147321358323097f;
  t183_n183: return 0.07724546641111374f;
  t183_n184: return 0.024035153910517693f;
  t183_n185: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t183_n189;
  t183_n186: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t183_n188;
  t183_n187: return 0.11098016053438187f;
  t183_n188: return 0.10252304375171661f;
  t183_n189: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n191;
  t183_n190: return -0.024001026526093483f;
  t183_n191: return 0.08193831890821457f;
  t183_n192: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t183_n242;
  t183_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n237;
  t183_n194: if (E.getFractionNameInContext() >= 3211904796 /*0.9444444179534912*/) goto t183_n226;
  t183_n195: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t183_n219;
  t183_n196: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t183_n216;
  t183_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n215;
  t183_n198: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t183_n204;
  t183_n199: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t183_n203;
  t183_n200: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t183_n202;
  t183_n201: return -0.3473508358001709f;
  t183_n202: return 0.025674227625131607f;
  t183_n203: return -0.032818764448165894f;
  t183_n204: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t183_n212;
  t183_n205: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t183_n211;
  t183_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t183_n210;
  t183_n207: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t183_n209;
  t183_n208: return 0.2453717589378357f;
  t183_n209: return 0.0412106066942215f;
  t183_n210: return 0.09238862246274948f;
  t183_n211: return -0.09139084815979004f;
  t183_n212: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t183_n214;
  t183_n213: return -0.059857457876205444f;
  t183_n214: return 0.0037456329446285963f;
  t183_n215: return 0.000561980064958334f;
  t183_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n218;
  t183_n217: return -0.014873871579766273f;
  t183_n218: return 0.25927630066871643f;
  t183_n219: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t183_n221;
  t183_n220: return -0.24635915458202362f;
  t183_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n225;
  t183_n222: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t183_n224;
  t183_n223: return -0.06665343791246414f;
  t183_n224: return -0.21254700422286987f;
  t183_n225: return 0.03985106572508812f;
  t183_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t183_n228;
  t183_n227: return -0.3002622127532959f;
  t183_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t183_n234;
  t183_n229: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n233;
  t183_n230: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t183_n232;
  t183_n231: return 0.10538511723279953f;
  t183_n232: return -0.3049219250679016f;
  t183_n233: return -0.12812277674674988f;
  t183_n234: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t183_n236;
  t183_n235: return -0.08353931456804276f;
  t183_n236: return 0.09479136765003204f;
  t183_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t183_n239;
  t183_n238: return 0.003246947657316923f;
  t183_n239: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t183_n241;
  t183_n240: return 0.015534039586782455f;
  t183_n241: return 0.08978629112243652f;
  t183_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t183_n244;
  t183_n243: return -0.011190575547516346f;
  t183_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t183_n248;
  t183_n245: if (E.getFractionNameInContext() >= 3211788288 /*0.9375*/) goto t183_n247;
  t183_n246: return 0.004667395260185003f;
  t183_n247: return 0.2208116501569748f;
  t183_n248: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t183_n252;
  t183_n249: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t183_n251;
  t183_n250: return 0.20289431512355804f;
  t183_n251: return 0.48565346002578735f;
  t183_n252: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t183_n254;
  t183_n253: return 0.07976502925157547f;
  t183_n254: return 0.13074585795402527f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree184(const Example& E) {
  t184_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t184_n248;
  t184_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t184_n137;
  t184_n2: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t184_n134;
  t184_n3: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t184_n11;
  t184_n4: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t184_n10;
  t184_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t184_n7;
  t184_n6: return -0.16788801550865173f;
  t184_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n9;
  t184_n8: return -0.029082896187901497f;
  t184_n9: return 0.07706061750650406f;
  t184_n10: return -0.16393227875232697f;
  t184_n11: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t184_n83;
  t184_n12: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t184_n14;
  t184_n13: return -0.0007148605072870851f;
  t184_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n32;
  t184_n15: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t184_n29;
  t184_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t184_n28;
  t184_n17: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t184_n27;
  t184_n18: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t184_n22;
  t184_n19: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t184_n21;
  t184_n20: return -0.2972467541694641f;
  t184_n21: return -0.10268815606832504f;
  t184_n22: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t184_n24;
  t184_n23: return -0.1490233987569809f;
  t184_n24: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t184_n26;
  t184_n25: return 0.07186176627874374f;
  t184_n26: return -0.18546825647354126f;
  t184_n27: return 0.08296477794647217f;
  t184_n28: return -0.019842762500047684f;
  t184_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t184_n31;
  t184_n30: return 0.06367804855108261f;
  t184_n31: return 0.12137775123119354f;
  t184_n32: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t184_n42;
  t184_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t184_n41;
  t184_n34: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t184_n36;
  t184_n35: return -0.0039840759709477425f;
  t184_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t184_n40;
  t184_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Type))) goto t184_n39;
  t184_n38: return 0.014104899950325489f;
  t184_n39: return 0.07476359605789185f;
  t184_n40: return -0.02633053809404373f;
  t184_n41: return 0.03908887133002281f;
  t184_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n80;
  t184_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t184_n67;
  t184_n44: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t184_n66;
  t184_n45: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t184_n63;
  t184_n46: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t184_n56;
  t184_n47: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t184_n55;
  t184_n48: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t184_n54;
  t184_n49: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t184_n51;
  t184_n50: return 0.12075051665306091f;
  t184_n51: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t184_n53;
  t184_n52: return -0.29477810859680176f;
  t184_n53: return 0.1527562141418457f;
  t184_n54: return -0.019006209447979927f;
  t184_n55: return -0.007800794672220945f;
  t184_n56: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t184_n58;
  t184_n57: return -0.03462577238678932f;
  t184_n58: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t184_n62;
  t184_n59: if (E.getFractionNameInContext() >= 3189821965 /*0.15705128014087677*/) goto t184_n61;
  t184_n60: return 0.16080836951732635f;
  t184_n61: return 0.03744153678417206f;
  t184_n62: return -0.012362181209027767f;
  t184_n63: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t184_n65;
  t184_n64: return -0.06150364130735397f;
  t184_n65: return 0.17827127873897552f;
  t184_n66: return -0.030796485021710396f;
  t184_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n73;
  t184_n68: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t184_n70;
  t184_n69: return 0.008776303380727768f;
  t184_n70: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t184_n72;
  t184_n71: return -0.07877999544143677f;
  t184_n72: return -0.2576081156730652f;
  t184_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t184_n75;
  t184_n74: return -0.00941185001283884f;
  t184_n75: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t184_n79;
  t184_n76: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t184_n78;
  t184_n77: return 0.01561480388045311f;
  t184_n78: return 0.004332670010626316f;
  t184_n79: return -0.04487578943371773f;
  t184_n80: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t184_n82;
  t184_n81: return -0.011781317181885242f;
  t184_n82: return 0.012218723073601723f;
  t184_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t184_n119;
  t184_n84: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t184_n86;
  t184_n85: return -0.037711091339588165f;
  t184_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t184_n114;
  t184_n87: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t184_n105;
  t184_n88: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t184_n102;
  t184_n89: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t184_n101;
  t184_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t184_n92;
  t184_n91: return -0.04297204688191414f;
  t184_n92: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t184_n100;
  t184_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n95;
  t184_n94: return -0.10995370894670486f;
  t184_n95: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t184_n99;
  t184_n96: if (E.getNumReferences() >= 3246915584 /*17.0*/) goto t184_n98;
  t184_n97: return 0.14579333364963531f;
  t184_n98: return 0.12175016850233078f;
  t184_n99: return 0.011962142772972584f;
  t184_n100: return 0.03992854803800583f;
  t184_n101: return 0.10591371357440948f;
  t184_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n104;
  t184_n103: return -0.10213665664196014f;
  t184_n104: return 0.2873989939689636f;
  t184_n105: if (E.getNumReferences() >= 3283959808 /*378.5*/) goto t184_n113;
  t184_n106: if (E.getNumReferences() >= 3279028224 /*242.0*/) goto t184_n112;
  t184_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t184_n109;
  t184_n108: return -0.05373983457684517f;
  t184_n109: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t184_n111;
  t184_n110: return 0.14474600553512573f;
  t184_n111: return 0.17541109025478363f;
  t184_n112: return 0.24033907055854797f;
  t184_n113: return -0.14531870186328888f;
  t184_n114: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t184_n118;
  t184_n115: if (E.getFractionNameInContext() >= 3190187622 /*0.16249999403953552*/) goto t184_n117;
  t184_n116: return 0.14331556856632233f;
  t184_n117: return 0.08573367446660995f;
  t184_n118: return 0.08450685441493988f;
  t184_n119: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t184_n129;
  t184_n120: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t184_n124;
  t184_n121: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t184_n123;
  t184_n122: return 0.047882117331027985f;
  t184_n123: return 0.07626249641180038f;
  t184_n124: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t184_n128;
  t184_n125: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t184_n127;
  t184_n126: return -0.12946146726608276f;
  t184_n127: return 0.026392018422484398f;
  t184_n128: return 0.10387639701366425f;
  t184_n129: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t184_n131;
  t184_n130: return -0.11497031152248383f;
  t184_n131: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t184_n133;
  t184_n132: return 0.10142381489276886f;
  t184_n133: return -0.007353014312684536f;
  t184_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n136;
  t184_n135: return -0.03505460545420647f;
  t184_n136: return 0.041561566293239594f;
  t184_n137: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t184_n175;
  t184_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t184_n172;
  t184_n139: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t184_n141;
  t184_n140: return -0.056727346032857895f;
  t184_n141: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t184_n157;
  t184_n142: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t184_n150;
  t184_n143: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t184_n149;
  t184_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t184_n148;
  t184_n145: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t184_n147;
  t184_n146: return -0.027160700410604477f;
  t184_n147: return 0.03754635900259018f;
  t184_n148: return 0.017579879611730576f;
  t184_n149: return 0.030161447823047638f;
  t184_n150: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t184_n154;
  t184_n151: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t184_n153;
  t184_n152: return -0.05401746556162834f;
  t184_n153: return -0.48748621344566345f;
  t184_n154: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t184_n156;
  t184_n155: return -0.21958978474140167f;
  t184_n156: return 0.02476189285516739f;
  t184_n157: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t184_n171;
  t184_n158: if (E.getNumReferences() >= 3305873408 /*2234.0*/) goto t184_n168;
  t184_n159: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t184_n165;
  t184_n160: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t184_n164;
  t184_n161: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t184_n163;
  t184_n162: return 0.0065347058698534966f;
  t184_n163: return -0.07923172414302826f;
  t184_n164: return 0.04956696182489395f;
  t184_n165: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t184_n167;
  t184_n166: return -0.020709095522761345f;
  t184_n167: return -0.08756747841835022f;
  t184_n168: if (E.getNumReferences() >= 3335566848 /*26715.0*/) goto t184_n170;
  t184_n169: return 0.05536834895610809f;
  t184_n170: return 0.0014034819323569536f;
  t184_n171: return 0.006461313925683498f;
  t184_n172: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t184_n174;
  t184_n173: return 0.1144125759601593f;
  t184_n174: return 0.0009978400776162744f;
  t184_n175: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t184_n245;
  t184_n176: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t184_n222;
  t184_n177: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t184_n179;
  t184_n178: return 0.006292636040598154f;
  t184_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n185;
  t184_n180: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t184_n182;
  t184_n181: return -0.013572700321674347f;
  t184_n182: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t184_n184;
  t184_n183: return 0.20288997888565063f;
  t184_n184: return -0.048298876732587814f;
  t184_n185: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t184_n213;
  t184_n186: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t184_n210;
  t184_n187: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t184_n195;
  t184_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n194;
  t184_n189: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t184_n191;
  t184_n190: return -0.008073712699115276f;
  t184_n191: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t184_n193;
  t184_n192: return -0.1382657289505005f;
  t184_n193: return 0.02220441959798336f;
  t184_n194: return 0.00913221575319767f;
  t184_n195: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t184_n203;
  t184_n196: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t184_n202;
  t184_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t184_n199;
  t184_n198: return 0.019298475235700607f;
  t184_n199: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t184_n201;
  t184_n200: return 0.03936021402478218f;
  t184_n201: return 0.08401931077241898f;
  t184_n202: return -0.011398477479815483f;
  t184_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t184_n205;
  t184_n204: return 0.03714175149798393f;
  t184_n205: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t184_n209;
  t184_n206: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t184_n208;
  t184_n207: return 0.07862874120473862f;
  t184_n208: return 0.0006466511404141784f;
  t184_n209: return -0.16369082033634186f;
  t184_n210: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t184_n212;
  t184_n211: return -0.035783786326646805f;
  t184_n212: return -0.6031245589256287f;
  t184_n213: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t184_n221;
  t184_n214: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t184_n216;
  t184_n215: return 0.0990787222981453f;
  t184_n216: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t184_n218;
  t184_n217: return -0.039828915148973465f;
  t184_n218: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t184_n220;
  t184_n219: return 0.07814985513687134f;
  t184_n220: return -0.0402148999273777f;
  t184_n221: return -0.012247410602867603f;
  t184_n222: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t184_n244;
  t184_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n225;
  t184_n224: return -0.08600186556577682f;
  t184_n225: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t184_n243;
  t184_n226: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t184_n242;
  t184_n227: if (E.getFractionNameInContext() >= 3209042018 /*0.773809552192688*/) goto t184_n241;
  t184_n228: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t184_n234;
  t184_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t184_n233;
  t184_n230: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t184_n232;
  t184_n231: return -0.008341665379703045f;
  t184_n232: return 0.11735739558935165f;
  t184_n233: return 0.11980392783880234f;
  t184_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t184_n236;
  t184_n235: return -0.04340657964348793f;
  t184_n236: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t184_n240;
  t184_n237: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t184_n239;
  t184_n238: return 0.003375125350430608f;
  t184_n239: return 0.14139395952224731f;
  t184_n240: return -0.12419968098402023f;
  t184_n241: return 0.11620714515447617f;
  t184_n242: return -0.027973897755146027f;
  t184_n243: return -0.04388224706053734f;
  t184_n244: return 0.11303509771823883f;
  t184_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t184_n247;
  t184_n246: return -0.006180211436003447f;
  t184_n247: return 0.030585123226046562f;
  t184_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t184_n252;
  t184_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t184_n251;
  t184_n250: return -0.12600643932819366f;
  t184_n251: return 0.0017277721781283617f;
  t184_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t184_n254;
  t184_n253: return 0.004801740869879723f;
  t184_n254: return -0.056234266608953476f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree185(const Example& E) {
  t185_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t185_n2;
  t185_n1: return -0.0002046758309006691f;
  t185_n2: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t185_n34;
  t185_n3: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t185_n29;
  t185_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n6;
  t185_n5: return -0.012178156524896622f;
  t185_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t185_n28;
  t185_n7: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t185_n25;
  t185_n8: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t185_n18;
  t185_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t185_n13;
  t185_n10: if (E.getNumReferences() >= 3291774976 /*722.0*/) goto t185_n12;
  t185_n11: return -0.0020751941483467817f;
  t185_n12: return 0.12589512765407562f;
  t185_n13: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t185_n17;
  t185_n14: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t185_n16;
  t185_n15: return 0.031154653057456017f;
  t185_n16: return 0.11548329144716263f;
  t185_n17: return -0.019876427948474884f;
  t185_n18: if (E.getNumReferences() >= 3264610304 /*75.0*/) goto t185_n22;
  t185_n19: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t185_n21;
  t185_n20: return -0.03227384760975838f;
  t185_n21: return -0.004547934979200363f;
  t185_n22: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t185_n24;
  t185_n23: return -0.5459371209144592f;
  t185_n24: return -0.0718044564127922f;
  t185_n25: if (E.getFractionNameInContext() >= 3197399006 /*0.2899159789085388*/) goto t185_n27;
  t185_n26: return -0.0007165158749558032f;
  t185_n27: return 0.015678392723202705f;
  t185_n28: return 0.028581973165273666f;
  t185_n29: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t185_n31;
  t185_n30: return 0.005942496471107006f;
  t185_n31: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t185_n33;
  t185_n32: return 0.13166119158267975f;
  t185_n33: return 0.028614837676286697f;
  t185_n34: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t185_n254;
  t185_n35: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t185_n211;
  t185_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t185_n210;
  t185_n37: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t185_n39;
  t185_n38: return 0.10652821511030197f;
  t185_n39: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t185_n163;
  t185_n40: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t185_n98;
  t185_n41: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t185_n43;
  t185_n42: return -0.010126291774213314f;
  t185_n43: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t185_n91;
  t185_n44: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t185_n80;
  t185_n45: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t185_n79;
  t185_n46: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t185_n64;
  t185_n47: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t185_n63;
  t185_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t185_n58;
  t185_n49: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t185_n53;
  t185_n50: if (E.getNumReferences() >= 3292209152 /*748.5*/) goto t185_n52;
  t185_n51: return 0.014340858906507492f;
  t185_n52: return 0.11876988410949707f;
  t185_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t185_n55;
  t185_n54: return -0.8975696563720703f;
  t185_n55: if (E.getNumReferences() >= 3277193216 /*214.0*/) goto t185_n57;
  t185_n56: return -0.13104234635829926f;
  t185_n57: return 0.10331058502197266f;
  t185_n58: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t185_n62;
  t185_n59: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t185_n61;
  t185_n60: return 0.1387518048286438f;
  t185_n61: return 0.019940895959734917f;
  t185_n62: return 0.08981940150260925f;
  t185_n63: return -0.15201358497142792f;
  t185_n64: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t185_n78;
  t185_n65: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t185_n75;
  t185_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t185_n72;
  t185_n67: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t185_n69;
  t185_n68: return 0.029559282585978508f;
  t185_n69: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t185_n71;
  t185_n70: return 0.14021414518356323f;
  t185_n71: return 0.08560609072446823f;
  t185_n72: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t185_n74;
  t185_n73: return 0.1437310129404068f;
  t185_n74: return 0.00027138125733472407f;
  t185_n75: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t185_n77;
  t185_n76: return 1.0000792741775513f;
  t185_n77: return 0.04038035124540329f;
  t185_n78: return 0.010807296261191368f;
  t185_n79: return 0.5067926645278931f;
  t185_n80: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t185_n82;
  t185_n81: return 0.0047861868515610695f;
  t185_n82: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t185_n90;
  t185_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t185_n87;
  t185_n84: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t185_n86;
  t185_n85: return 0.053610485047101974f;
  t185_n86: return 0.11443153023719788f;
  t185_n87: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t185_n89;
  t185_n88: return 0.08489301055669785f;
  t185_n89: return 0.025600124150514603f;
  t185_n90: return 0.0019698147661983967f;
  t185_n91: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t185_n95;
  t185_n92: if (E.getNumReferences() >= 3299532800 /*1367.0*/) goto t185_n94;
  t185_n93: return -0.0086420439183712f;
  t185_n94: return 0.07476498186588287f;
  t185_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t185_n97;
  t185_n96: return 0.028951305896043777f;
  t185_n97: return 0.17456229031085968f;
  t185_n98: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t185_n146;
  t185_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t185_n119;
  t185_n100: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t185_n106;
  t185_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n105;
  t185_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n104;
  t185_n103: return -0.21317225694656372f;
  t185_n104: return -0.07385082542896271f;
  t185_n105: return 0.22342386841773987f;
  t185_n106: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t185_n114;
  t185_n107: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t185_n109;
  t185_n108: return -0.13014084100723267f;
  t185_n109: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t185_n113;
  t185_n110: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t185_n112;
  t185_n111: return 0.07666988670825958f;
  t185_n112: return 0.10673949122428894f;
  t185_n113: return 0.0417344905436039f;
  t185_n114: if (E.getFractionNameInContext() >= 3208871340 /*0.7636363506317139*/) goto t185_n116;
  t185_n115: return -0.01704687811434269f;
  t185_n116: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t185_n118;
  t185_n117: return 0.08073791861534119f;
  t185_n118: return 0.00744564738124609f;
  t185_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n135;
  t185_n120: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t185_n134;
  t185_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t185_n125;
  t185_n122: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t185_n124;
  t185_n123: return 0.0775410383939743f;
  t185_n124: return -0.011625470593571663f;
  t185_n125: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t185_n131;
  t185_n126: if (E.getNumReferences() >= 3355841184 /*137290.5*/) goto t185_n130;
  t185_n127: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t185_n129;
  t185_n128: return 0.06114405766129494f;
  t185_n129: return -0.1641869693994522f;
  t185_n130: return -0.49694231152534485f;
  t185_n131: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t185_n133;
  t185_n132: return 0.10404308140277863f;
  t185_n133: return 0.1608085334300995f;
  t185_n134: return -0.046432413160800934f;
  t185_n135: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t185_n143;
  t185_n136: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t185_n140;
  t185_n137: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t185_n139;
  t185_n138: return 0.05527637153863907f;
  t185_n139: return -0.0850517600774765f;
  t185_n140: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t185_n142;
  t185_n141: return -0.002405168954282999f;
  t185_n142: return 0.17117144167423248f;
  t185_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t185_n145;
  t185_n144: return 0.01656518504023552f;
  t185_n145: return 0.10835020989179611f;
  t185_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t185_n158;
  t185_n147: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t185_n155;
  t185_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t185_n152;
  t185_n149: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t185_n151;
  t185_n150: return 0.22915464639663696f;
  t185_n151: return 0.07609434425830841f;
  t185_n152: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t185_n154;
  t185_n153: return -0.13236863911151886f;
  t185_n154: return 0.5534843802452087f;
  t185_n155: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t185_n157;
  t185_n156: return -0.006836156360805035f;
  t185_n157: return -0.6246297359466553f;
  t185_n158: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t185_n160;
  t185_n159: return -0.10061338543891907f;
  t185_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n162;
  t185_n161: return -0.10202254354953766f;
  t185_n162: return 0.7386455535888672f;
  t185_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t185_n205;
  t185_n164: if (E.getFractionNameInContext() >= 3198211988 /*0.31414473056793213*/) goto t185_n200;
  t185_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n167;
  t185_n166: return -0.023880887776613235f;
  t185_n167: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t185_n195;
  t185_n168: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t185_n192;
  t185_n169: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t185_n173;
  t185_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t185_n172;
  t185_n171: return 0.009115898981690407f;
  t185_n172: return 0.03837272524833679f;
  t185_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t185_n191;
  t185_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t185_n186;
  t185_n175: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t185_n177;
  t185_n176: return -0.023246070370078087f;
  t185_n177: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t185_n185;
  t185_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t185_n180;
  t185_n179: return -0.023181995376944542f;
  t185_n180: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t185_n182;
  t185_n181: return 0.026169592514634132f;
  t185_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t185_n184;
  t185_n183: return 0.04870214685797691f;
  t185_n184: return 0.1340450644493103f;
  t185_n185: return -0.002993740141391754f;
  t185_n186: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t185_n190;
  t185_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n189;
  t185_n188: return -0.10701951384544373f;
  t185_n189: return 0.182301864027977f;
  t185_n190: return 0.04335082694888115f;
  t185_n191: return 0.015124385245144367f;
  t185_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n194;
  t185_n193: return -0.05465197190642357f;
  t185_n194: return 0.03799409419298172f;
  t185_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t185_n197;
  t185_n196: return 0.0033052770886570215f;
  t185_n197: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t185_n199;
  t185_n198: return 0.022148804739117622f;
  t185_n199: return -0.004963950254023075f;
  t185_n200: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t185_n204;
  t185_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n203;
  t185_n202: return -0.010834651999175549f;
  t185_n203: return 0.002757829846814275f;
  t185_n204: return 0.0021178731694817543f;
  t185_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t185_n209;
  t185_n206: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t185_n208;
  t185_n207: return -0.1111777052283287f;
  t185_n208: return 0.020774394273757935f;
  t185_n209: return 0.028805293142795563f;
  t185_n210: return 0.007234177086502314f;
  t185_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t185_n217;
  t185_n212: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t185_n216;
  t185_n213: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t185_n215;
  t185_n214: return 0.031115388497710228f;
  t185_n215: return 0.11165789514780045f;
  t185_n216: return -0.006883664987981319f;
  t185_n217: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t185_n235;
  t185_n218: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t185_n224;
  t185_n219: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t185_n223;
  t185_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t185_n222;
  t185_n221: return 0.0004104448016732931f;
  t185_n222: return 0.021111780777573586f;
  t185_n223: return 0.139961838722229f;
  t185_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t185_n226;
  t185_n225: return 0.002253043930977583f;
  t185_n226: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t185_n234;
  t185_n227: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t185_n229;
  t185_n228: return 0.06363695859909058f;
  t185_n229: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t185_n233;
  t185_n230: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t185_n232;
  t185_n231: return 0.1096896082162857f;
  t185_n232: return 0.11741377413272858f;
  t185_n233: return -0.10291150957345963f;
  t185_n234: return 0.038563214242458344f;
  t185_n235: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t185_n243;
  t185_n236: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t185_n242;
  t185_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t185_n239;
  t185_n238: return 0.018593810498714447f;
  t185_n239: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t185_n241;
  t185_n240: return 0.15503273904323578f;
  t185_n241: return -0.13776761293411255f;
  t185_n242: return -0.0064310841262340546f;
  t185_n243: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t185_n245;
  t185_n244: return -0.12821130454540253f;
  t185_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t185_n253;
  t185_n246: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t185_n248;
  t185_n247: return -0.025218965485692024f;
  t185_n248: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t185_n252;
  t185_n249: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t185_n251;
  t185_n250: return 0.08844064176082611f;
  t185_n251: return -0.08953575789928436f;
  t185_n252: return 0.10830318927764893f;
  t185_n253: return 0.029396727681159973f;
  t185_n254: return 0.003929317928850651f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree186(const Example& E) {
  t186_n0: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t186_n204;
  t186_n1: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t186_n203;
  t186_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t186_n4;
  t186_n3: return -0.00016824259364511818f;
  t186_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t186_n176;
  t186_n5: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t186_n73;
  t186_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t186_n16;
  t186_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t186_n9;
  t186_n8: return -0.1445317566394806f;
  t186_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t186_n15;
  t186_n10: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t186_n14;
  t186_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t186_n13;
  t186_n12: return 0.05421604588627815f;
  t186_n13: return 0.09649734944105148f;
  t186_n14: return -0.011246743611991405f;
  t186_n15: return -0.21053220331668854f;
  t186_n16: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t186_n18;
  t186_n17: return -0.01568264700472355f;
  t186_n18: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t186_n20;
  t186_n19: return -0.032052718102931976f;
  t186_n20: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t186_n60;
  t186_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n27;
  t186_n22: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t186_n26;
  t186_n23: if (E.getFractionNameInContext() >= 3172829657 /*0.03846153989434242*/) goto t186_n25;
  t186_n24: return -0.21737226843833923f;
  t186_n25: return -0.02171824499964714f;
  t186_n26: return 0.06602652370929718f;
  t186_n27: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t186_n51;
  t186_n28: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t186_n30;
  t186_n29: return 0.0095867570489645f;
  t186_n30: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t186_n48;
  t186_n31: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t186_n37;
  t186_n32: if (E.getNumReferences() >= 3247439872 /*18.0*/) goto t186_n36;
  t186_n33: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t186_n35;
  t186_n34: return 0.20275484025478363f;
  t186_n35: return 0.014548365026712418f;
  t186_n36: return 0.1029948890209198f;
  t186_n37: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t186_n39;
  t186_n38: return 0.10329683870077133f;
  t186_n39: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t186_n45;
  t186_n40: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t186_n44;
  t186_n41: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t186_n43;
  t186_n42: return -0.4566442668437958f;
  t186_n43: return -1.3562793731689453f;
  t186_n44: return -0.06924724578857422f;
  t186_n45: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t186_n47;
  t186_n46: return -0.03487095236778259f;
  t186_n47: return 0.07626072317361832f;
  t186_n48: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t186_n50;
  t186_n49: return 0.04704824835062027f;
  t186_n50: return -0.19189706444740295f;
  t186_n51: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t186_n53;
  t186_n52: return 0.01572612300515175f;
  t186_n53: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t186_n55;
  t186_n54: return -0.01293243933469057f;
  t186_n55: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t186_n57;
  t186_n56: return -0.35536104440689087f;
  t186_n57: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t186_n59;
  t186_n58: return 0.1255563646554947f;
  t186_n59: return 0.009119099006056786f;
  t186_n60: if (E.getFractionNameInContext() >= 3199399908 /*0.34954750537872314*/) goto t186_n62;
  t186_n61: return -0.1511625200510025f;
  t186_n62: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t186_n68;
  t186_n63: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t186_n65;
  t186_n64: return 0.09684425592422485f;
  t186_n65: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t186_n67;
  t186_n66: return 0.11014027893543243f;
  t186_n67: return 0.13133934140205383f;
  t186_n68: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t186_n70;
  t186_n69: return -0.007961797527968884f;
  t186_n70: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t186_n72;
  t186_n71: return 0.11041259765625f;
  t186_n72: return -0.0035836088936775923f;
  t186_n73: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t186_n127;
  t186_n74: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t186_n92;
  t186_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t186_n89;
  t186_n76: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t186_n78;
  t186_n77: return 0.0036732966545969248f;
  t186_n78: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t186_n86;
  t186_n79: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t186_n85;
  t186_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n84;
  t186_n81: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t186_n83;
  t186_n82: return 0.080511175096035f;
  t186_n83: return -0.17445355653762817f;
  t186_n84: return 0.11588393151760101f;
  t186_n85: return -0.02145909145474434f;
  t186_n86: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t186_n88;
  t186_n87: return -0.015843383967876434f;
  t186_n88: return 0.09737277030944824f;
  t186_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n91;
  t186_n90: return -0.10983658581972122f;
  t186_n91: return -0.0013757947599515319f;
  t186_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n98;
  t186_n93: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t186_n97;
  t186_n94: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t186_n96;
  t186_n95: return -0.2946198880672455f;
  t186_n96: return -0.0038729142397642136f;
  t186_n97: return -0.10112779587507248f;
  t186_n98: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t186_n102;
  t186_n99: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t186_n101;
  t186_n100: return 0.01035606674849987f;
  t186_n101: return -0.11330585926771164f;
  t186_n102: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t186_n110;
  t186_n103: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t186_n105;
  t186_n104: return 0.07179907709360123f;
  t186_n105: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t186_n107;
  t186_n106: return 0.06966546177864075f;
  t186_n107: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t186_n109;
  t186_n108: return 0.1753871887922287f;
  t186_n109: return -0.3483048677444458f;
  t186_n110: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t186_n112;
  t186_n111: return 0.007504134904593229f;
  t186_n112: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t186_n118;
  t186_n113: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t186_n117;
  t186_n114: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t186_n116;
  t186_n115: return 0.0444200225174427f;
  t186_n116: return 0.1275949329137802f;
  t186_n117: return -0.008670806884765625f;
  t186_n118: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t186_n124;
  t186_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n121;
  t186_n120: return -0.03439178690314293f;
  t186_n121: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t186_n123;
  t186_n122: return 0.13930968940258026f;
  t186_n123: return -0.25635331869125366f;
  t186_n124: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t186_n126;
  t186_n125: return -0.05249154940247536f;
  t186_n126: return 0.12529586255550385f;
  t186_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n129;
  t186_n128: return -0.09981405735015869f;
  t186_n129: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t186_n167;
  t186_n130: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t186_n166;
  t186_n131: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t186_n133;
  t186_n132: return 0.015050469897687435f;
  t186_n133: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t186_n157;
  t186_n134: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t186_n136;
  t186_n135: return 0.03422760218381882f;
  t186_n136: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t186_n142;
  t186_n137: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t186_n141;
  t186_n138: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t186_n140;
  t186_n139: return 0.012362221255898476f;
  t186_n140: return -0.0327286571264267f;
  t186_n141: return -0.0973803922533989f;
  t186_n142: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t186_n156;
  t186_n143: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t186_n155;
  t186_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t186_n146;
  t186_n145: return 0.00670017721131444f;
  t186_n146: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t186_n148;
  t186_n147: return -0.02896944247186184f;
  t186_n148: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t186_n154;
  t186_n149: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t186_n153;
  t186_n150: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t186_n152;
  t186_n151: return -0.4076443314552307f;
  t186_n152: return 0.07909619808197021f;
  t186_n153: return -1.1064932346343994f;
  t186_n154: return -0.008523316122591496f;
  t186_n155: return 0.0387575589120388f;
  t186_n156: return 0.1178889200091362f;
  t186_n157: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t186_n165;
  t186_n158: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t186_n160;
  t186_n159: return -0.1703128218650818f;
  t186_n160: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t186_n164;
  t186_n161: if (E.getNumReferences() >= 3265396736 /*81.0*/) goto t186_n163;
  t186_n162: return -0.07738343626260757f;
  t186_n163: return 0.10543346405029297f;
  t186_n164: return 0.13214826583862305f;
  t186_n165: return 0.03667289763689041f;
  t186_n166: return 0.10782712697982788f;
  t186_n167: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t186_n171;
  t186_n168: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t186_n170;
  t186_n169: return 0.0689590722322464f;
  t186_n170: return -0.05243372544646263f;
  t186_n171: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t186_n175;
  t186_n172: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t186_n174;
  t186_n173: return 0.13251537084579468f;
  t186_n174: return 0.11336417496204376f;
  t186_n175: return -0.06244735047221184f;
  t186_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t186_n202;
  t186_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t186_n183;
  t186_n178: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t186_n182;
  t186_n179: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t186_n181;
  t186_n180: return 0.0044757165014743805f;
  t186_n181: return 0.04955926164984703f;
  t186_n182: return -0.00500637898221612f;
  t186_n183: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t186_n187;
  t186_n184: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t186_n186;
  t186_n185: return 0.007247111760079861f;
  t186_n186: return -0.014022949151694775f;
  t186_n187: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t186_n201;
  t186_n188: if (E.getFractionNameInContext() >= 3196938455 /*0.27619048953056335*/) goto t186_n190;
  t186_n189: return 0.08579074591398239f;
  t186_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t186_n198;
  t186_n191: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t186_n197;
  t186_n192: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t186_n196;
  t186_n193: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t186_n195;
  t186_n194: return 0.014382507652044296f;
  t186_n195: return 0.08494676649570465f;
  t186_n196: return -0.06048901006579399f;
  t186_n197: return 0.08465549349784851f;
  t186_n198: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t186_n200;
  t186_n199: return 0.06252453476190567f;
  t186_n200: return -0.04359906166791916f;
  t186_n201: return -0.10879149287939072f;
  t186_n202: return 0.0650196522474289f;
  t186_n203: return 0.008488246239721775f;
  t186_n204: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t186_n206;
  t186_n205: return -0.0009251984884031117f;
  t186_n206: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t186_n208;
  t186_n207: return 0.13593211770057678f;
  t186_n208: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t186_n242;
  t186_n209: if (E.getFractionNameInContext() >= 3206790530 /*0.6396104097366333*/) goto t186_n227;
  t186_n210: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t186_n222;
  t186_n211: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t186_n213;
  t186_n212: return 0.06503430753946304f;
  t186_n213: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t186_n215;
  t186_n214: return 0.01229721400886774f;
  t186_n215: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t186_n221;
  t186_n216: if (E.getSymbolScopeDistanceCost() >= 3249012736 /*21.0*/) goto t186_n220;
  t186_n217: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t186_n219;
  t186_n218: return -1.6320977210998535f;
  t186_n219: return -0.15286149084568024f;
  t186_n220: return 0.005269252695143223f;
  t186_n221: return -0.04375699162483215f;
  t186_n222: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t186_n224;
  t186_n223: return -0.1517723649740219f;
  t186_n224: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t186_n226;
  t186_n225: return 0.13308575749397278f;
  t186_n226: return 0.06638423353433609f;
  t186_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t186_n231;
  t186_n228: if (E.getNumReferences() >= 3238002688 /*8.0*/) goto t186_n230;
  t186_n229: return -0.1449034959077835f;
  t186_n230: return 0.10397332906723022f;
  t186_n231: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t186_n241;
  t186_n232: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t186_n234;
  t186_n233: return 0.037357673048973083f;
  t186_n234: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t186_n240;
  t186_n235: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t186_n239;
  t186_n236: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t186_n238;
  t186_n237: return 0.05311711132526398f;
  t186_n238: return 0.11151005327701569f;
  t186_n239: return -0.14708136022090912f;
  t186_n240: return 0.09928534179925919f;
  t186_n241: return -0.04733576625585556f;
  t186_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t186_n244;
  t186_n243: return 0.008148283697664738f;
  t186_n244: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t186_n254;
  t186_n245: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t186_n247;
  t186_n246: return -0.047440432012081146f;
  t186_n247: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t186_n253;
  t186_n248: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t186_n252;
  t186_n249: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t186_n251;
  t186_n250: return 0.06572017818689346f;
  t186_n251: return 0.10739511996507645f;
  t186_n252: return 0.012323279865086079f;
  t186_n253: return 0.065489262342453f;
  t186_n254: return 0.10394368320703506f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree187(const Example& E) {
  t187_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t187_n216;
  t187_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t187_n3;
  t187_n2: return -0.000626898545306176f;
  t187_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t187_n215;
  t187_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace))) goto t187_n106;
  t187_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t187_n7;
  t187_n6: return -0.009341657161712646f;
  t187_n7: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t187_n47;
  t187_n8: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t187_n46;
  t187_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t187_n15;
  t187_n10: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t187_n14;
  t187_n11: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t187_n13;
  t187_n12: return 0.01056885626167059f;
  t187_n13: return -0.0369470939040184f;
  t187_n14: return -0.012541241943836212f;
  t187_n15: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t187_n17;
  t187_n16: return 0.009052521549165249f;
  t187_n17: if (E.getNumReferences() >= 3274211328 /*168.5*/) goto t187_n41;
  t187_n18: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t187_n30;
  t187_n19: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t187_n21;
  t187_n20: return 0.010426012799143791f;
  t187_n21: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t187_n23;
  t187_n22: return -0.1071741133928299f;
  t187_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n29;
  t187_n24: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t187_n28;
  t187_n25: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t187_n27;
  t187_n26: return -0.015737542882561684f;
  t187_n27: return -0.0940052717924118f;
  t187_n28: return -0.002722698962315917f;
  t187_n29: return 0.1621546447277069f;
  t187_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t187_n32;
  t187_n31: return -0.01412682794034481f;
  t187_n32: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t187_n40;
  t187_n33: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t187_n37;
  t187_n34: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t187_n36;
  t187_n35: return 0.027265820652246475f;
  t187_n36: return 0.05463296175003052f;
  t187_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n39;
  t187_n38: return -0.053772538900375366f;
  t187_n39: return 0.023162806406617165f;
  t187_n40: return -0.07585597038269043f;
  t187_n41: if (E.getNumReferences() >= 3274276864 /*169.5*/) goto t187_n43;
  t187_n42: return 0.2364569902420044f;
  t187_n43: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t187_n45;
  t187_n44: return 0.019890792667865753f;
  t187_n45: return -0.1436290293931961f;
  t187_n46: return 0.002126717008650303f;
  t187_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n49;
  t187_n48: return -0.008111299015581608f;
  t187_n49: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t187_n97;
  t187_n50: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t187_n54;
  t187_n51: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t187_n53;
  t187_n52: return 0.006053392309695482f;
  t187_n53: return -0.0072609782218933105f;
  t187_n54: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t187_n68;
  t187_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n57;
  t187_n56: return -0.019569069147109985f;
  t187_n57: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t187_n65;
  t187_n58: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t187_n64;
  t187_n59: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t187_n63;
  t187_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t187_n62;
  t187_n61: return -0.004654192365705967f;
  t187_n62: return 0.024326622486114502f;
  t187_n63: return 0.008020474575459957f;
  t187_n64: return -0.014528150670230389f;
  t187_n65: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t187_n67;
  t187_n66: return 0.11258355528116226f;
  t187_n67: return 1.159951090812683f;
  t187_n68: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t187_n96;
  t187_n69: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t187_n75;
  t187_n70: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t187_n74;
  t187_n71: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t187_n73;
  t187_n72: return -0.004934391938149929f;
  t187_n73: return 0.15600506961345673f;
  t187_n74: return 0.004574578255414963f;
  t187_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t187_n79;
  t187_n76: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t187_n78;
  t187_n77: return -0.01870655082166195f;
  t187_n78: return -0.1393333226442337f;
  t187_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t187_n83;
  t187_n80: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t187_n82;
  t187_n81: return 0.008775672875344753f;
  t187_n82: return -0.002127789193764329f;
  t187_n83: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t187_n93;
  t187_n84: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t187_n92;
  t187_n85: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t187_n89;
  t187_n86: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t187_n88;
  t187_n87: return 0.01548696681857109f;
  t187_n88: return -0.1164667010307312f;
  t187_n89: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t187_n91;
  t187_n90: return 0.046812281012535095f;
  t187_n91: return 0.009222683496773243f;
  t187_n92: return -0.385146826505661f;
  t187_n93: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t187_n95;
  t187_n94: return 0.0696633830666542f;
  t187_n95: return 0.02829798124730587f;
  t187_n96: return 0.04009879380464554f;
  t187_n97: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t187_n105;
  t187_n98: if (E.getNumReferences() >= 3281108992 /*291.5*/) goto t187_n100;
  t187_n99: return -0.0008960621198639274f;
  t187_n100: if (E.getNumReferences() >= 3310174208 /*3284.0*/) goto t187_n104;
  t187_n101: if (E.getNumReferences() >= 3309481984 /*3115.0*/) goto t187_n103;
  t187_n102: return 0.07256380468606949f;
  t187_n103: return 0.18607504665851593f;
  t187_n104: return -0.02266940288245678f;
  t187_n105: return -0.02657654881477356f;
  t187_n106: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t187_n158;
  t187_n107: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t187_n157;
  t187_n108: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t187_n126;
  t187_n109: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t187_n111;
  t187_n110: return 0.008957941085100174f;
  t187_n111: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t187_n125;
  t187_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t187_n120;
  t187_n113: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t187_n117;
  t187_n114: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t187_n116;
  t187_n115: return -0.005984005983918905f;
  t187_n116: return 0.24951060116291046f;
  t187_n117: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t187_n119;
  t187_n118: return 0.9447497129440308f;
  t187_n119: return 0.053756505250930786f;
  t187_n120: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t187_n122;
  t187_n121: return 0.10772950202226639f;
  t187_n122: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t187_n124;
  t187_n123: return 0.1461026668548584f;
  t187_n124: return 0.08643222600221634f;
  t187_n125: return 0.02524532563984394f;
  t187_n126: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t187_n146;
  t187_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t187_n143;
  t187_n128: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t187_n136;
  t187_n129: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t187_n131;
  t187_n130: return 0.004961137659847736f;
  t187_n131: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n135;
  t187_n132: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t187_n134;
  t187_n133: return -0.2742350399494171f;
  t187_n134: return 0.07083526998758316f;
  t187_n135: return 0.10932599753141403f;
  t187_n136: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t187_n140;
  t187_n137: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t187_n139;
  t187_n138: return -0.10295528918504715f;
  t187_n139: return 0.010244583711028099f;
  t187_n140: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t187_n142;
  t187_n141: return 0.03609880432486534f;
  t187_n142: return -0.2879229485988617f;
  t187_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n145;
  t187_n144: return -0.052994586527347565f;
  t187_n145: return 0.11199787259101868f;
  t187_n146: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t187_n152;
  t187_n147: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t187_n151;
  t187_n148: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t187_n150;
  t187_n149: return 0.09659457951784134f;
  t187_n150: return 0.05545355752110481f;
  t187_n151: return 0.02643992193043232f;
  t187_n152: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t187_n154;
  t187_n153: return 0.04275393486022949f;
  t187_n154: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t187_n156;
  t187_n155: return 0.05344979465007782f;
  t187_n156: return -0.25734227895736694f;
  t187_n157: return 0.001685313880443573f;
  t187_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t187_n194;
  t187_n159: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t187_n193;
  t187_n160: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t187_n162;
  t187_n161: return -0.03507978469133377f;
  t187_n162: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t187_n166;
  t187_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n165;
  t187_n164: return 0.02400834672152996f;
  t187_n165: return 0.07159819453954697f;
  t187_n166: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t187_n192;
  t187_n167: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t187_n187;
  t187_n168: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t187_n186;
  t187_n169: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t187_n175;
  t187_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n172;
  t187_n171: return -0.06632421165704727f;
  t187_n172: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t187_n174;
  t187_n173: return 0.062398217618465424f;
  t187_n174: return -0.005866926163434982f;
  t187_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n177;
  t187_n176: return -0.10839705914258957f;
  t187_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n185;
  t187_n178: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t187_n180;
  t187_n179: return 0.043028172105550766f;
  t187_n180: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t187_n184;
  t187_n181: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t187_n183;
  t187_n182: return -0.20762614905834198f;
  t187_n183: return -0.16377781331539154f;
  t187_n184: return 0.0858042910695076f;
  t187_n185: return 0.05115978792309761f;
  t187_n186: return -0.043471258133649826f;
  t187_n187: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n191;
  t187_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n190;
  t187_n189: return -0.12581698596477509f;
  t187_n190: return 0.030424481257796288f;
  t187_n191: return 0.05024043098092079f;
  t187_n192: return -0.026410112157464027f;
  t187_n193: return -0.12158595770597458f;
  t187_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t187_n212;
  t187_n195: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t187_n203;
  t187_n196: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t187_n198;
  t187_n197: return -0.005929804407060146f;
  t187_n198: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t187_n200;
  t187_n199: return 0.0011841666419059038f;
  t187_n200: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t187_n202;
  t187_n201: return 0.7085154056549072f;
  t187_n202: return 0.045739173889160156f;
  t187_n203: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t187_n205;
  t187_n204: return 0.015081940218806267f;
  t187_n205: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t187_n211;
  t187_n206: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t187_n208;
  t187_n207: return 0.053863443434238434f;
  t187_n208: if (E.getNumReferences() >= 3326905856 /*13091.5*/) goto t187_n210;
  t187_n209: return 0.01241949014365673f;
  t187_n210: return -0.11406544595956802f;
  t187_n211: return 0.12348063290119171f;
  t187_n212: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t187_n214;
  t187_n213: return -0.010972399264574051f;
  t187_n214: return 0.012741615995764732f;
  t187_n215: return 0.0037964603397995234f;
  t187_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t187_n242;
  t187_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t187_n237;
  t187_n218: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t187_n228;
  t187_n219: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t187_n221;
  t187_n220: return -0.002522902563214302f;
  t187_n221: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t187_n225;
  t187_n222: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t187_n224;
  t187_n223: return -0.030150914564728737f;
  t187_n224: return 0.09323671460151672f;
  t187_n225: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t187_n227;
  t187_n226: return 0.012373130768537521f;
  t187_n227: return -0.03715955838561058f;
  t187_n228: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t187_n234;
  t187_n229: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t187_n231;
  t187_n230: return 0.016087589785456657f;
  t187_n231: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t187_n233;
  t187_n232: return 0.12102971225976944f;
  t187_n233: return -0.04187949001789093f;
  t187_n234: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t187_n236;
  t187_n235: return 0.07744558155536652f;
  t187_n236: return 0.13025163114070892f;
  t187_n237: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t187_n239;
  t187_n238: return -0.022099629044532776f;
  t187_n239: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t187_n241;
  t187_n240: return 0.011603191494941711f;
  t187_n241: return 0.02848798967897892f;
  t187_n242: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t187_n254;
  t187_n243: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t187_n245;
  t187_n244: return 0.002973940223455429f;
  t187_n245: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t187_n247;
  t187_n246: return 0.0692133679986f;
  t187_n247: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t187_n251;
  t187_n248: if (E.getNumReferences() >= 3356814208 /*152494.0*/) goto t187_n250;
  t187_n249: return 0.01562727428972721f;
  t187_n250: return -0.3003333806991577f;
  t187_n251: if (E.getNumReferences() >= 3361141760 /*220112.0*/) goto t187_n253;
  t187_n252: return 0.07819665968418121f;
  t187_n253: return 0.04793303832411766f;
  t187_n254: return 0.02400130406022072f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree188(const Example& E) {
  t188_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t188_n254;
  t188_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t188_n249;
  t188_n2: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t188_n186;
  t188_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t188_n121;
  t188_n4: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t188_n114;
  t188_n5: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t188_n85;
  t188_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t188_n32;
  t188_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n27;
  t188_n8: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t188_n26;
  t188_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n11;
  t188_n10: return -0.10707085579633713f;
  t188_n11: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t188_n21;
  t188_n12: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t188_n20;
  t188_n13: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t188_n15;
  t188_n14: return 0.023111119866371155f;
  t188_n15: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n19;
  t188_n16: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t188_n18;
  t188_n17: return 0.12008335441350937f;
  t188_n18: return 0.11581774055957794f;
  t188_n19: return 0.06319020688533783f;
  t188_n20: return 0.018351895734667778f;
  t188_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t188_n23;
  t188_n22: return -0.12692321836948395f;
  t188_n23: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n25;
  t188_n24: return 0.797773540019989f;
  t188_n25: return 0.12769454717636108f;
  t188_n26: return -0.010928072966635227f;
  t188_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t188_n29;
  t188_n28: return -0.06617840379476547f;
  t188_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t188_n31;
  t188_n30: return 0.0002680967445485294f;
  t188_n31: return 0.006378361023962498f;
  t188_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t188_n48;
  t188_n33: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t188_n41;
  t188_n34: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t188_n36;
  t188_n35: return -0.0057405708357691765f;
  t188_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n40;
  t188_n37: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n39;
  t188_n38: return 0.015560587868094444f;
  t188_n39: return 0.300592839717865f;
  t188_n40: return 0.004237506538629532f;
  t188_n41: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n47;
  t188_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t188_n44;
  t188_n43: return 0.056889958679676056f;
  t188_n44: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t188_n46;
  t188_n45: return 0.2448912113904953f;
  t188_n46: return -0.14172135293483734f;
  t188_n47: return -0.013762333430349827f;
  t188_n48: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t188_n52;
  t188_n49: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t188_n51;
  t188_n50: return 0.07050983607769012f;
  t188_n51: return -0.04150928184390068f;
  t188_n52: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n70;
  t188_n53: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t188_n67;
  t188_n54: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t188_n66;
  t188_n55: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t188_n65;
  t188_n56: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t188_n64;
  t188_n57: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t188_n61;
  t188_n58: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t188_n60;
  t188_n59: return 0.04172290861606598f;
  t188_n60: return 0.009426930919289589f;
  t188_n61: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t188_n63;
  t188_n62: return -0.5251564979553223f;
  t188_n63: return 0.04032896086573601f;
  t188_n64: return 0.023188743740320206f;
  t188_n65: return -0.10362043976783752f;
  t188_n66: return -0.12394887208938599f;
  t188_n67: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n69;
  t188_n68: return 0.05120014399290085f;
  t188_n69: return -0.18924929201602936f;
  t188_n70: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t188_n72;
  t188_n71: return -0.13298185169696808f;
  t188_n72: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t188_n84;
  t188_n73: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t188_n75;
  t188_n74: return 0.011874448508024216f;
  t188_n75: if (E.getFileProximityDistanceCost() >= 3254910976 /*32.5*/) goto t188_n83;
  t188_n76: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t188_n80;
  t188_n77: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t188_n79;
  t188_n78: return -0.06951580941677094f;
  t188_n79: return -0.1035563126206398f;
  t188_n80: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t188_n82;
  t188_n81: return -0.02630799449980259f;
  t188_n82: return 0.027569517493247986f;
  t188_n83: return -0.4542044401168823f;
  t188_n84: return 0.06281653046607971f;
  t188_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n93;
  t188_n86: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t188_n88;
  t188_n87: return -0.3066834807395935f;
  t188_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t188_n92;
  t188_n89: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t188_n91;
  t188_n90: return -0.0435347743332386f;
  t188_n91: return 0.5664917826652527f;
  t188_n92: return -0.10128659754991531f;
  t188_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t188_n95;
  t188_n94: return -0.011791698634624481f;
  t188_n95: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t188_n109;
  t188_n96: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t188_n106;
  t188_n97: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t188_n103;
  t188_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n102;
  t188_n99: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t188_n101;
  t188_n100: return -0.042878296226263046f;
  t188_n101: return 0.20257768034934998f;
  t188_n102: return -0.0040351832285523415f;
  t188_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n105;
  t188_n104: return -0.10483184456825256f;
  t188_n105: return 0.06327665597200394f;
  t188_n106: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n108;
  t188_n107: return 0.022749394178390503f;
  t188_n108: return 0.07446007430553436f;
  t188_n109: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t188_n111;
  t188_n110: return 0.02618134208023548f;
  t188_n111: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n113;
  t188_n112: return 0.17825835943222046f;
  t188_n113: return -0.008004069328308105f;
  t188_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n116;
  t188_n115: return -0.048694759607315063f;
  t188_n116: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t188_n120;
  t188_n117: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t188_n119;
  t188_n118: return 0.12583594024181366f;
  t188_n119: return 0.2748804986476898f;
  t188_n120: return -0.006298573222011328f;
  t188_n121: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t188_n155;
  t188_n122: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t188_n152;
  t188_n123: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n137;
  t188_n124: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t188_n126;
  t188_n125: return -0.000640995625872165f;
  t188_n126: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t188_n132;
  t188_n127: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t188_n131;
  t188_n128: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t188_n130;
  t188_n129: return 4.164293204667047e-05f;
  t188_n130: return -0.027691425755620003f;
  t188_n131: return 0.05294949561357498f;
  t188_n132: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t188_n134;
  t188_n133: return -0.13782308995723724f;
  t188_n134: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t188_n136;
  t188_n135: return 0.16177920997142792f;
  t188_n136: return -0.11262723803520203f;
  t188_n137: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t188_n149;
  t188_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n146;
  t188_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n145;
  t188_n140: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t188_n144;
  t188_n141: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t188_n143;
  t188_n142: return -0.10354144126176834f;
  t188_n143: return -0.10620223730802536f;
  t188_n144: return 0.08496406674385071f;
  t188_n145: return 0.0030946361366659403f;
  t188_n146: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t188_n148;
  t188_n147: return 0.017983974888920784f;
  t188_n148: return -0.0015383388381451368f;
  t188_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t188_n151;
  t188_n150: return 0.06319790333509445f;
  t188_n151: return 0.11266156286001205f;
  t188_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t188_n154;
  t188_n153: return 0.0013938592746853828f;
  t188_n154: return 0.013953527435660362f;
  t188_n155: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n173;
  t188_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t188_n158;
  t188_n157: return 0.008305622264742851f;
  t188_n158: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t188_n164;
  t188_n159: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t188_n163;
  t188_n160: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t188_n162;
  t188_n161: return 0.03656627982854843f;
  t188_n162: return 0.011429959908127785f;
  t188_n163: return -0.10548123717308044f;
  t188_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t188_n166;
  t188_n165: return -1.0812541246414185f;
  t188_n166: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t188_n168;
  t188_n167: return -0.02848714031279087f;
  t188_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t188_n170;
  t188_n169: return 0.08084532618522644f;
  t188_n170: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t188_n172;
  t188_n171: return 0.10896947234869003f;
  t188_n172: return -0.10054897516965866f;
  t188_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n185;
  t188_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t188_n180;
  t188_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n179;
  t188_n176: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t188_n178;
  t188_n177: return -0.10181386768817902f;
  t188_n178: return 0.123222216963768f;
  t188_n179: return -0.024056240916252136f;
  t188_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n182;
  t188_n181: return -0.07891983538866043f;
  t188_n182: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t188_n184;
  t188_n183: return -0.12183432281017303f;
  t188_n184: return 4.294602870941162f;
  t188_n185: return 0.011561665683984756f;
  t188_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t188_n218;
  t188_n187: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t188_n189;
  t188_n188: return -0.01624353975057602f;
  t188_n189: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t188_n191;
  t188_n190: return -0.12290152907371521f;
  t188_n191: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t188_n201;
  t188_n192: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t188_n194;
  t188_n193: return -0.003974169492721558f;
  t188_n194: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t188_n196;
  t188_n195: return 0.03497075289487839f;
  t188_n196: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t188_n200;
  t188_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t188_n199;
  t188_n198: return 0.03649419546127319f;
  t188_n199: return 0.11975449323654175f;
  t188_n200: return 0.07228299230337143f;
  t188_n201: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t188_n205;
  t188_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n204;
  t188_n203: return -0.027713080868124962f;
  t188_n204: return 0.11742200702428818f;
  t188_n205: if (E.getNumReferences() >= 3263627264 /*67.5*/) goto t188_n211;
  t188_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t188_n208;
  t188_n207: return 0.01282954216003418f;
  t188_n208: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t188_n210;
  t188_n209: return 0.04958618804812431f;
  t188_n210: return 0.1303745061159134f;
  t188_n211: if (E.getNumReferences() >= 3269591040 /*113.0*/) goto t188_n215;
  t188_n212: if (E.getNumReferences() >= 3267821568 /*99.5*/) goto t188_n214;
  t188_n213: return -0.024965571239590645f;
  t188_n214: return -0.3843179941177368f;
  t188_n215: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t188_n217;
  t188_n216: return 0.06176484003663063f;
  t188_n217: return -0.05859784409403801f;
  t188_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t188_n248;
  t188_n219: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t188_n225;
  t188_n220: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t188_n224;
  t188_n221: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t188_n223;
  t188_n222: return 0.02279553934931755f;
  t188_n223: return -0.2039349228143692f;
  t188_n224: return 0.27494168281555176f;
  t188_n225: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t188_n229;
  t188_n226: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t188_n228;
  t188_n227: return 0.007363366894423962f;
  t188_n228: return -0.003590163541957736f;
  t188_n229: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t188_n231;
  t188_n230: return 0.03708415850996971f;
  t188_n231: if (E.getNumReferences() >= 3292815360 /*785.5*/) goto t188_n239;
  t188_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t188_n238;
  t188_n233: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t188_n237;
  t188_n234: if (E.getNumReferences() >= 3288268800 /*510.0*/) goto t188_n236;
  t188_n235: return -0.012101538479328156f;
  t188_n236: return 0.2303500473499298f;
  t188_n237: return -0.11227750778198242f;
  t188_n238: return 0.009359976276755333f;
  t188_n239: if (E.getNumReferences() >= 3292995584 /*796.5*/) goto t188_n241;
  t188_n240: return 0.2540584206581116f;
  t188_n241: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t188_n247;
  t188_n242: if (E.getNumReferences() >= 3293544448 /*830.0*/) goto t188_n246;
  t188_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t188_n245;
  t188_n244: return 0.052616946399211884f;
  t188_n245: return 0.1380709558725357f;
  t188_n246: return 0.020611271262168884f;
  t188_n247: return -0.5959507822990417f;
  t188_n248: return 0.13872672617435455f;
  t188_n249: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t188_n251;
  t188_n250: return -0.0011493518250063062f;
  t188_n251: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t188_n253;
  t188_n252: return -0.019533796235919f;
  t188_n253: return -0.0032434524036943913f;
  t188_n254: return 0.000978067866526544f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree189(const Example& E) {
  t189_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t189_n2;
  t189_n1: return -0.0004899075720459223f;
  t189_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t189_n188;
  t189_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t189_n5;
  t189_n4: return -4.553383405436762e-05f;
  t189_n5: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t189_n53;
  t189_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t189_n42;
  t189_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t189_n9;
  t189_n8: return -0.028937658295035362f;
  t189_n9: if (E.getNumReferences() >= 3304468480 /*1969.5*/) goto t189_n41;
  t189_n10: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t189_n20;
  t189_n11: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t189_n19;
  t189_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t189_n14;
  t189_n13: return -0.07223943620920181f;
  t189_n14: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t189_n18;
  t189_n15: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t189_n17;
  t189_n16: return -0.0731995552778244f;
  t189_n17: return 0.12387137115001678f;
  t189_n18: return 0.04874753579497337f;
  t189_n19: return 0.005114032421261072f;
  t189_n20: if (E.getNumReferences() >= 3277684736 /*221.5*/) goto t189_n34;
  t189_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t189_n27;
  t189_n22: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t189_n24;
  t189_n23: return -0.2439032346010208f;
  t189_n24: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t189_n26;
  t189_n25: return 0.001721831620670855f;
  t189_n26: return -0.26556196808815f;
  t189_n27: if (E.getNumReferences() >= 3269394432 /*111.5*/) goto t189_n31;
  t189_n28: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t189_n30;
  t189_n29: return -0.015952639281749725f;
  t189_n30: return -0.903538703918457f;
  t189_n31: if (E.getNumReferences() >= 3271983104 /*134.5*/) goto t189_n33;
  t189_n32: return 0.1622217893600464f;
  t189_n33: return -0.01671701855957508f;
  t189_n34: if (E.getNumReferences() >= 3289382912 /*576.0*/) goto t189_n38;
  t189_n35: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t189_n37;
  t189_n36: return 0.09760361164808273f;
  t189_n37: return 0.021903708577156067f;
  t189_n38: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t189_n40;
  t189_n39: return -0.5119736790657043f;
  t189_n40: return -0.040404144674539566f;
  t189_n41: return 0.1159946545958519f;
  t189_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t189_n46;
  t189_n43: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t189_n45;
  t189_n44: return -0.031134815886616707f;
  t189_n45: return 0.03704385086894035f;
  t189_n46: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t189_n50;
  t189_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t189_n49;
  t189_n48: return 0.059392355382442474f;
  t189_n49: return 0.025252221152186394f;
  t189_n50: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t189_n52;
  t189_n51: return -0.038920432329177856f;
  t189_n52: return 0.14340101182460785f;
  t189_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t189_n55;
  t189_n54: return -0.01977589540183544f;
  t189_n55: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t189_n121;
  t189_n56: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t189_n108;
  t189_n57: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t189_n103;
  t189_n58: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t189_n82;
  t189_n59: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t189_n81;
  t189_n60: if (E.getNumReferences() >= 3286974464 /*470.5*/) goto t189_n80;
  t189_n61: if (E.getNumReferences() >= 3271819264 /*132.0*/) goto t189_n77;
  t189_n62: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t189_n68;
  t189_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t189_n67;
  t189_n64: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t189_n66;
  t189_n65: return -0.0057829865254461765f;
  t189_n66: return -0.6198936700820923f;
  t189_n67: return 0.056325480341911316f;
  t189_n68: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t189_n72;
  t189_n69: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t189_n71;
  t189_n70: return -0.12796230614185333f;
  t189_n71: return -0.5979392528533936f;
  t189_n72: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t189_n76;
  t189_n73: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t189_n75;
  t189_n74: return -0.9128319621086121f;
  t189_n75: return -0.06264018267393112f;
  t189_n76: return -0.001140847452916205f;
  t189_n77: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t189_n79;
  t189_n78: return 0.16185446083545685f;
  t189_n79: return 0.10387786477804184f;
  t189_n80: return -0.3308092951774597f;
  t189_n81: return -0.010798170231282711f;
  t189_n82: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t189_n94;
  t189_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t189_n91;
  t189_n84: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t189_n86;
  t189_n85: return -0.09446509182453156f;
  t189_n86: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t189_n88;
  t189_n87: return 0.0706552192568779f;
  t189_n88: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t189_n90;
  t189_n89: return 0.22119590640068054f;
  t189_n90: return 0.11129564791917801f;
  t189_n91: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t189_n93;
  t189_n92: return 0.10374109447002411f;
  t189_n93: return 0.12286043912172318f;
  t189_n94: if (E.getNumReferences() >= 3262513152 /*61.5*/) goto t189_n100;
  t189_n95: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t189_n99;
  t189_n96: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t189_n98;
  t189_n97: return -0.02542676031589508f;
  t189_n98: return 0.10469765961170197f;
  t189_n99: return -0.10537387430667877f;
  t189_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t189_n102;
  t189_n101: return 0.005036272574216127f;
  t189_n102: return 0.5208917856216431f;
  t189_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type))) goto t189_n105;
  t189_n104: return -0.0067269387654960155f;
  t189_n105: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t189_n107;
  t189_n106: return 0.1374935805797577f;
  t189_n107: return 0.027505138888955116f;
  t189_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t189_n110;
  t189_n109: return 0.14655284583568573f;
  t189_n110: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t189_n114;
  t189_n111: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t189_n113;
  t189_n112: return 0.10925130546092987f;
  t189_n113: return -0.0036366076674312353f;
  t189_n114: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t189_n118;
  t189_n115: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t189_n117;
  t189_n116: return 0.11007504910230637f;
  t189_n117: return -1.3002445786725048e-09f;
  t189_n118: if (E.getFractionNameInContext() >= 3193263446 /*0.2083333432674408*/) goto t189_n120;
  t189_n119: return 0.16370955109596252f;
  t189_n120: return 0.12770727276802063f;
  t189_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t189_n127;
  t189_n122: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t189_n124;
  t189_n123: return -0.043691884726285934f;
  t189_n124: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t189_n126;
  t189_n125: return 0.13444991409778595f;
  t189_n126: return 0.191347137093544f;
  t189_n127: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t189_n153;
  t189_n128: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t189_n130;
  t189_n129: return 0.024082956835627556f;
  t189_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t189_n142;
  t189_n131: if (E.getNumReferences() >= 3297681408 /*1141.0*/) goto t189_n139;
  t189_n132: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t189_n136;
  t189_n133: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t189_n135;
  t189_n134: return -0.0651792511343956f;
  t189_n135: return -0.42579522728919983f;
  t189_n136: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t189_n138;
  t189_n137: return -0.005449510645121336f;
  t189_n138: return 0.13884972035884857f;
  t189_n139: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t189_n141;
  t189_n140: return -0.42899444699287415f;
  t189_n141: return -0.09039556235074997f;
  t189_n142: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t189_n144;
  t189_n143: return 0.0804295688867569f;
  t189_n144: if (E.getNumReferences() >= 3283468288 /*363.5*/) goto t189_n146;
  t189_n145: return -0.02056458592414856f;
  t189_n146: if (E.getNumReferences() >= 3284172800 /*385.0*/) goto t189_n148;
  t189_n147: return 0.15292787551879883f;
  t189_n148: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t189_n150;
  t189_n149: return 0.007501464337110519f;
  t189_n150: if (E.getNumReferences() >= 3301195776 /*1570.0*/) goto t189_n152;
  t189_n151: return 0.05693446472287178f;
  t189_n152: return 0.21896860003471375f;
  t189_n153: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t189_n187;
  t189_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t189_n158;
  t189_n155: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t189_n157;
  t189_n156: return 0.18413367867469788f;
  t189_n157: return 0.006742463447153568f;
  t189_n158: if (E.getNumReferences() >= 3328401408 /*14552.0*/) goto t189_n186;
  t189_n159: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t189_n185;
  t189_n160: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t189_n174;
  t189_n161: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t189_n171;
  t189_n162: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t189_n168;
  t189_n163: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t189_n165;
  t189_n164: return 0.07029920816421509f;
  t189_n165: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t189_n167;
  t189_n166: return 0.27775922417640686f;
  t189_n167: return -0.054955609142780304f;
  t189_n168: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t189_n170;
  t189_n169: return -0.38497471809387207f;
  t189_n170: return 0.08120295405387878f;
  t189_n171: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t189_n173;
  t189_n172: return -0.036322422325611115f;
  t189_n173: return 0.02668612077832222f;
  t189_n174: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t189_n176;
  t189_n175: return -0.021776815876364708f;
  t189_n176: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t189_n180;
  t189_n177: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t189_n179;
  t189_n178: return 0.06485901027917862f;
  t189_n179: return -0.0938490554690361f;
  t189_n180: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t189_n184;
  t189_n181: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t189_n183;
  t189_n182: return -0.15940602123737335f;
  t189_n183: return -0.019732270389795303f;
  t189_n184: return 0.018852531909942627f;
  t189_n185: return -0.01893812231719494f;
  t189_n186: return 0.0723816454410553f;
  t189_n187: return -0.12788823246955872f;
  t189_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Namespace))) goto t189_n214;
  t189_n189: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t189_n211;
  t189_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t189_n198;
  t189_n191: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t189_n193;
  t189_n192: return -0.01835550367832184f;
  t189_n193: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t189_n197;
  t189_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t189_n196;
  t189_n195: return -0.013461374677717686f;
  t189_n196: return 0.001433235825970769f;
  t189_n197: return 0.0036323312669992447f;
  t189_n198: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t189_n200;
  t189_n199: return 0.00430239038541913f;
  t189_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t189_n202;
  t189_n201: return -0.13042831420898438f;
  t189_n202: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t189_n204;
  t189_n203: return -0.0030043472070246935f;
  t189_n204: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t189_n208;
  t189_n205: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t189_n207;
  t189_n206: return 0.2579324245452881f;
  t189_n207: return -0.10770323872566223f;
  t189_n208: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t189_n210;
  t189_n209: return -0.06062892824411392f;
  t189_n210: return 0.0065246582962572575f;
  t189_n211: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t189_n213;
  t189_n212: return 0.037397485226392746f;
  t189_n213: return 0.006733829155564308f;
  t189_n214: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t189_n216;
  t189_n215: return -0.029688207432627678f;
  t189_n216: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t189_n226;
  t189_n217: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t189_n225;
  t189_n218: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t189_n220;
  t189_n219: return -0.012835810892283916f;
  t189_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t189_n222;
  t189_n221: return 0.04213515669107437f;
  t189_n222: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t189_n224;
  t189_n223: return 0.1481594741344452f;
  t189_n224: return -0.1359235942363739f;
  t189_n225: return -0.1472999006509781f;
  t189_n226: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t189_n238;
  t189_n227: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t189_n235;
  t189_n228: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t189_n234;
  t189_n229: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t189_n231;
  t189_n230: return 0.04291671887040138f;
  t189_n231: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t189_n233;
  t189_n232: return 0.047551434487104416f;
  t189_n233: return -0.035731468349695206f;
  t189_n234: return 0.12865129113197327f;
  t189_n235: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t189_n237;
  t189_n236: return 0.0029972258489578962f;
  t189_n237: return -0.2728275656700134f;
  t189_n238: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t189_n240;
  t189_n239: return 0.002672858303412795f;
  t189_n240: if (E.getFractionNameInContext() >= 3195014913 /*0.23443223536014557*/) goto t189_n246;
  t189_n241: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t189_n245;
  t189_n242: if (E.getFractionNameInContext() >= 3194215998 /*0.22252747416496277*/) goto t189_n244;
  t189_n243: return 0.08070756494998932f;
  t189_n244: return 0.11044221371412277f;
  t189_n245: return -0.10811302810907364f;
  t189_n246: if (E.getFractionNameInContext() >= 3198561514 /*0.3245614171028137*/) goto t189_n250;
  t189_n247: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t189_n249;
  t189_n248: return -0.07437805831432343f;
  t189_n249: return 0.08837592601776123f;
  t189_n250: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t189_n254;
  t189_n251: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t189_n253;
  t189_n252: return 0.07532765716314316f;
  t189_n253: return -0.17321671545505524f;
  t189_n254: return 0.021701300516724586f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree190(const Example& E) {
  t190_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t190_n104;
  t190_n1: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t190_n3;
  t190_n2: return -0.00020805506210308522f;
  t190_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t190_n5;
  t190_n4: return -0.15779680013656616f;
  t190_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t190_n17;
  t190_n6: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n8;
  t190_n7: return -0.003377165412530303f;
  t190_n8: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t190_n14;
  t190_n9: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t190_n13;
  t190_n10: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t190_n12;
  t190_n11: return 0.04830247536301613f;
  t190_n12: return 0.0019230461912229657f;
  t190_n13: return -0.0011839502258226275f;
  t190_n14: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t190_n16;
  t190_n15: return -0.1060049757361412f;
  t190_n16: return 0.13547760248184204f;
  t190_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t190_n19;
  t190_n18: return -0.009821811690926552f;
  t190_n19: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t190_n103;
  t190_n20: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t190_n28;
  t190_n21: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t190_n23;
  t190_n22: return -0.01352731604129076f;
  t190_n23: if (E.getFractionNameInContext() >= 3208152317 /*0.7207792401313782*/) goto t190_n27;
  t190_n24: if (E.getFractionNameInContext() >= 3207044730 /*0.6547619104385376*/) goto t190_n26;
  t190_n25: return 0.002878654282540083f;
  t190_n26: return 0.10247937589883804f;
  t190_n27: return -0.1344747245311737f;
  t190_n28: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t190_n66;
  t190_n29: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t190_n63;
  t190_n30: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t190_n62;
  t190_n31: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t190_n37;
  t190_n32: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t190_n34;
  t190_n33: return 0.018530819565057755f;
  t190_n34: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t190_n36;
  t190_n35: return -0.624600350856781f;
  t190_n36: return 0.10831921547651291f;
  t190_n37: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t190_n59;
  t190_n38: if (E.getNumReferences() >= 3266969600 /*93.0*/) goto t190_n56;
  t190_n39: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t190_n55;
  t190_n40: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t190_n48;
  t190_n41: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n47;
  t190_n42: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n46;
  t190_n43: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t190_n45;
  t190_n44: return 0.19074001908302307f;
  t190_n45: return 0.16117152571678162f;
  t190_n46: return 0.005244896747171879f;
  t190_n47: return 0.07763738185167313f;
  t190_n48: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t190_n52;
  t190_n49: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t190_n51;
  t190_n50: return 0.08723682910203934f;
  t190_n51: return -0.16681337356567383f;
  t190_n52: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t190_n54;
  t190_n53: return 0.11749223619699478f;
  t190_n54: return 0.050411406904459f;
  t190_n55: return 0.10429447144269943f;
  t190_n56: if (E.getNumReferences() >= 3278602240 /*235.5*/) goto t190_n58;
  t190_n57: return 0.16682876646518707f;
  t190_n58: return -0.4037652313709259f;
  t190_n59: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t190_n61;
  t190_n60: return -0.10744447261095047f;
  t190_n61: return 0.08533571660518646f;
  t190_n62: return 0.04933692887425423f;
  t190_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t190_n65;
  t190_n64: return -0.019727516919374466f;
  t190_n65: return 0.05905630812048912f;
  t190_n66: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t190_n68;
  t190_n67: return 0.007854443043470383f;
  t190_n68: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t190_n70;
  t190_n69: return -0.009430931881070137f;
  t190_n70: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t190_n102;
  t190_n71: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t190_n101;
  t190_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t190_n76;
  t190_n73: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t190_n75;
  t190_n74: return 0.05122501775622368f;
  t190_n75: return -0.09266756474971771f;
  t190_n76: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t190_n86;
  t190_n77: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t190_n81;
  t190_n78: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t190_n80;
  t190_n79: return 0.04208461940288544f;
  t190_n80: return 0.08539815992116928f;
  t190_n81: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t190_n83;
  t190_n82: return 0.07986506819725037f;
  t190_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n85;
  t190_n84: return -0.09147863835096359f;
  t190_n85: return 0.018193062394857407f;
  t190_n86: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t190_n88;
  t190_n87: return -0.015459845773875713f;
  t190_n88: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t190_n90;
  t190_n89: return 0.08686311542987823f;
  t190_n90: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t190_n100;
  t190_n91: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t190_n99;
  t190_n92: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t190_n98;
  t190_n93: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t190_n95;
  t190_n94: return -0.27925071120262146f;
  t190_n95: if (E.getFractionNameInContext() >= 3211159142 /*0.8999999761581421*/) goto t190_n97;
  t190_n96: return 0.11011667549610138f;
  t190_n97: return 0.14308123290538788f;
  t190_n98: return 0.11222895234823227f;
  t190_n99: return 0.1081698015332222f;
  t190_n100: return 0.0027204900979995728f;
  t190_n101: return 0.11757225543260574f;
  t190_n102: return 0.03250735253095627f;
  t190_n103: return -0.01274335477501154f;
  t190_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t190_n254;
  t190_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t190_n243;
  t190_n106: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t190_n194;
  t190_n107: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t190_n109;
  t190_n108: return 0.0032226149924099445f;
  t190_n109: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t190_n111;
  t190_n110: return 0.16452838480472565f;
  t190_n111: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t190_n177;
  t190_n112: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t190_n176;
  t190_n113: if (E.getFractionNameInContext() >= 3197124868 /*0.2817460298538208*/) goto t190_n129;
  t190_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n116;
  t190_n115: return -0.10025741904973984f;
  t190_n116: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n128;
  t190_n117: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n119;
  t190_n118: return 0.023896194994449615f;
  t190_n119: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t190_n121;
  t190_n120: return -0.20963941514492035f;
  t190_n121: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t190_n127;
  t190_n122: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t190_n126;
  t190_n123: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t190_n125;
  t190_n124: return 0.11707542091608047f;
  t190_n125: return 0.1761387586593628f;
  t190_n126: return 0.07288256287574768f;
  t190_n127: return 0.06452850252389908f;
  t190_n128: return 0.017680613324046135f;
  t190_n129: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t190_n143;
  t190_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n138;
  t190_n131: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t190_n135;
  t190_n132: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t190_n134;
  t190_n133: return -0.14211629331111908f;
  t190_n134: return 0.002412287052720785f;
  t190_n135: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t190_n137;
  t190_n136: return -0.12132951617240906f;
  t190_n137: return -1.5582008361816406f;
  t190_n138: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t190_n142;
  t190_n139: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n141;
  t190_n140: return -0.002174093620851636f;
  t190_n141: return -0.08819503337144852f;
  t190_n142: return 0.13752193748950958f;
  t190_n143: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t190_n155;
  t190_n144: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n146;
  t190_n145: return 0.033671244978904724f;
  t190_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n154;
  t190_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n153;
  t190_n148: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t190_n152;
  t190_n149: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t190_n151;
  t190_n150: return -0.23735365271568298f;
  t190_n151: return -0.3136734664440155f;
  t190_n152: return -0.011314923875033855f;
  t190_n153: return -0.04585929587483406f;
  t190_n154: return 0.033578060567379f;
  t190_n155: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n175;
  t190_n156: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t190_n166;
  t190_n157: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t190_n165;
  t190_n158: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n160;
  t190_n159: return 0.018465818837285042f;
  t190_n160: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t190_n164;
  t190_n161: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t190_n163;
  t190_n162: return 0.1600523740053177f;
  t190_n163: return 0.07144642621278763f;
  t190_n164: return -0.036602310836315155f;
  t190_n165: return 0.3968706727027893f;
  t190_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t190_n174;
  t190_n167: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t190_n173;
  t190_n168: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t190_n172;
  t190_n169: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t190_n171;
  t190_n170: return -0.23877067863941193f;
  t190_n171: return 0.03706388175487518f;
  t190_n172: return 0.0323745496571064f;
  t190_n173: return 0.0772232860326767f;
  t190_n174: return 0.1319665163755417f;
  t190_n175: return -0.06924717873334885f;
  t190_n176: return 0.20108191668987274f;
  t190_n177: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t190_n193;
  t190_n178: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t190_n184;
  t190_n179: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t190_n181;
  t190_n180: return 0.041724275797605515f;
  t190_n181: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t190_n183;
  t190_n182: return 0.05725352093577385f;
  t190_n183: return -0.19046108424663544f;
  t190_n184: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t190_n192;
  t190_n185: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t190_n189;
  t190_n186: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t190_n188;
  t190_n187: return 0.15200310945510864f;
  t190_n188: return 0.14931921660900116f;
  t190_n189: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t190_n191;
  t190_n190: return 0.11960706114768982f;
  t190_n191: return -0.1233731359243393f;
  t190_n192: return 0.15080232918262482f;
  t190_n193: return -0.08056733757257462f;
  t190_n194: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t190_n228;
  t190_n195: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t190_n197;
  t190_n196: return -0.019825609400868416f;
  t190_n197: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t190_n227;
  t190_n198: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t190_n210;
  t190_n199: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t190_n201;
  t190_n200: return -0.03871886804699898f;
  t190_n201: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t190_n209;
  t190_n202: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t190_n206;
  t190_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n205;
  t190_n204: return -0.015973111614584923f;
  t190_n205: return 0.18123094737529755f;
  t190_n206: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t190_n208;
  t190_n207: return 0.01791105419397354f;
  t190_n208: return 0.1216145008802414f;
  t190_n209: return 0.049779925495386124f;
  t190_n210: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t190_n224;
  t190_n211: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t190_n219;
  t190_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n218;
  t190_n213: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t190_n215;
  t190_n214: return -0.030998365953564644f;
  t190_n215: if (E.getNumReferences() >= 3258187776 /*45.0*/) goto t190_n217;
  t190_n216: return 0.2278376668691635f;
  t190_n217: return 0.08739056438207626f;
  t190_n218: return 0.025309402495622635f;
  t190_n219: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t190_n223;
  t190_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t190_n222;
  t190_n221: return -0.014317609369754791f;
  t190_n222: return 0.07930239289999008f;
  t190_n223: return 0.09818076342344284f;
  t190_n224: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t190_n226;
  t190_n225: return -0.029455378651618958f;
  t190_n226: return -0.11426593363285065f;
  t190_n227: return -0.001736943842843175f;
  t190_n228: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t190_n230;
  t190_n229: return 0.008947998285293579f;
  t190_n230: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t190_n236;
  t190_n231: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t190_n233;
  t190_n232: return 0.0832013189792633f;
  t190_n233: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t190_n235;
  t190_n234: return 0.2510433793067932f;
  t190_n235: return 0.08809497952461243f;
  t190_n236: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t190_n238;
  t190_n237: return -0.036227114498615265f;
  t190_n238: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t190_n242;
  t190_n239: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t190_n241;
  t190_n240: return 0.16461774706840515f;
  t190_n241: return 0.1506742686033249f;
  t190_n242: return 0.030635716393589973f;
  t190_n243: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t190_n253;
  t190_n244: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t190_n246;
  t190_n245: return 0.010016510263085365f;
  t190_n246: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t190_n250;
  t190_n247: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t190_n249;
  t190_n248: return -0.012505742721259594f;
  t190_n249: return 0.03555859252810478f;
  t190_n250: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t190_n252;
  t190_n251: return 0.10041848570108414f;
  t190_n252: return -0.13708403706550598f;
  t190_n253: return 0.030807975679636f;
  t190_n254: return 0.0043790824711322784f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree191(const Example& E) {
  t191_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t191_n2;
  t191_n1: return -0.0004717946576420218f;
  t191_n2: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t191_n252;
  t191_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t191_n133;
  t191_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t191_n18;
  t191_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t191_n9;
  t191_n6: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t191_n8;
  t191_n7: return -0.07756617665290833f;
  t191_n8: return -0.04734949395060539f;
  t191_n9: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t191_n11;
  t191_n10: return 0.10335025936365128f;
  t191_n11: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t191_n13;
  t191_n12: return 0.0018472514348104596f;
  t191_n13: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t191_n15;
  t191_n14: return 0.024465134367346764f;
  t191_n15: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t191_n17;
  t191_n16: return 0.06307130306959152f;
  t191_n17: return 0.15638306736946106f;
  t191_n18: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t191_n46;
  t191_n19: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t191_n45;
  t191_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t191_n30;
  t191_n21: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t191_n25;
  t191_n22: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t191_n24;
  t191_n23: return 0.01065438985824585f;
  t191_n24: return -0.03302072361111641f;
  t191_n25: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t191_n29;
  t191_n26: if (E.getNumReferences() >= 3368194704 /*398484.5*/) goto t191_n28;
  t191_n27: return -0.013261091895401478f;
  t191_n28: return -0.19659751653671265f;
  t191_n29: return 0.01309114322066307f;
  t191_n30: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t191_n38;
  t191_n31: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t191_n37;
  t191_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t191_n34;
  t191_n33: return -0.052164968103170395f;
  t191_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n36;
  t191_n35: return -0.03150990977883339f;
  t191_n36: return 0.01382086519151926f;
  t191_n37: return 0.01261135283857584f;
  t191_n38: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t191_n40;
  t191_n39: return -0.007074030116200447f;
  t191_n40: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t191_n44;
  t191_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n43;
  t191_n42: return -0.14382703602313995f;
  t191_n43: return 0.023999273777008057f;
  t191_n44: return 0.054038338363170624f;
  t191_n45: return 0.0013941786019131541f;
  t191_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n54;
  t191_n47: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t191_n49;
  t191_n48: return -0.1002093255519867f;
  t191_n49: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t191_n51;
  t191_n50: return 0.016109062358736992f;
  t191_n51: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t191_n53;
  t191_n52: return -0.002513653365895152f;
  t191_n53: return 0.05191337689757347f;
  t191_n54: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t191_n132;
  t191_n55: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t191_n97;
  t191_n56: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t191_n70;
  t191_n57: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t191_n61;
  t191_n58: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t191_n60;
  t191_n59: return 0.005387096665799618f;
  t191_n60: return -0.0367160364985466f;
  t191_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t191_n69;
  t191_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n68;
  t191_n63: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t191_n67;
  t191_n64: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t191_n66;
  t191_n65: return -0.08556331694126129f;
  t191_n66: return -0.011954578571021557f;
  t191_n67: return -0.30579322576522827f;
  t191_n68: return -0.0018588851671665907f;
  t191_n69: return 0.012491095811128616f;
  t191_n70: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t191_n80;
  t191_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n73;
  t191_n72: return -0.01713901199400425f;
  t191_n73: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t191_n79;
  t191_n74: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t191_n78;
  t191_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t191_n77;
  t191_n76: return -0.004271680489182472f;
  t191_n77: return 0.0223537664860487f;
  t191_n78: return 0.006006719544529915f;
  t191_n79: return -0.014982050284743309f;
  t191_n80: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t191_n96;
  t191_n81: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t191_n93;
  t191_n82: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t191_n90;
  t191_n83: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t191_n89;
  t191_n84: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t191_n88;
  t191_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n87;
  t191_n86: return -0.0051134261302649975f;
  t191_n87: return 0.009686306118965149f;
  t191_n88: return -0.36199480295181274f;
  t191_n89: return 0.20250539481639862f;
  t191_n90: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t191_n92;
  t191_n91: return 0.0051690018735826015f;
  t191_n92: return -0.010953841730952263f;
  t191_n93: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t191_n95;
  t191_n94: return 0.008459032513201237f;
  t191_n95: return -0.24016883969306946f;
  t191_n96: return 0.0416802354156971f;
  t191_n97: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t191_n99;
  t191_n98: return 0.021609660238027573f;
  t191_n99: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t191_n131;
  t191_n100: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t191_n108;
  t191_n101: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t191_n103;
  t191_n102: return -0.010253913700580597f;
  t191_n103: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t191_n107;
  t191_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t191_n106;
  t191_n105: return 0.0930323675274849f;
  t191_n106: return 0.776103675365448f;
  t191_n107: return 0.061743494123220444f;
  t191_n108: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t191_n130;
  t191_n109: if (E.getNumReferences() >= 3310114816 /*3269.5*/) goto t191_n129;
  t191_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t191_n118;
  t191_n111: if (E.getNumReferences() >= 3297271808 /*1091.0*/) goto t191_n113;
  t191_n112: return 0.02722848393023014f;
  t191_n113: if (E.getNumReferences() >= 3297775616 /*1152.5*/) goto t191_n115;
  t191_n114: return 0.2054782658815384f;
  t191_n115: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t191_n117;
  t191_n116: return 0.07510538399219513f;
  t191_n117: return 0.11512981355190277f;
  t191_n118: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t191_n120;
  t191_n119: return 0.32517534494400024f;
  t191_n120: if (E.getNumReferences() >= 3298951168 /*1296.0*/) goto t191_n128;
  t191_n121: if (E.getNumReferences() >= 3276242944 /*199.5*/) goto t191_n123;
  t191_n122: return -0.0673159658908844f;
  t191_n123: if (E.getNumReferences() >= 3282452480 /*332.5*/) goto t191_n125;
  t191_n124: return 0.3360001742839813f;
  t191_n125: if (E.getNumReferences() >= 3297456128 /*1113.5*/) goto t191_n127;
  t191_n126: return 0.07565334439277649f;
  t191_n127: return 0.112408846616745f;
  t191_n128: return -0.11592110246419907f;
  t191_n129: return 0.00835324265062809f;
  t191_n130: return -0.08444616943597794f;
  t191_n131: return -0.033536147326231f;
  t191_n132: return 0.03835337981581688f;
  t191_n133: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t191_n251;
  t191_n134: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t191_n156;
  t191_n135: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t191_n153;
  t191_n136: if (E.getNumReferences() >= 3265462272 /*81.5*/) goto t191_n138;
  t191_n137: return 0.007551718037575483f;
  t191_n138: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t191_n152;
  t191_n139: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t191_n147;
  t191_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t191_n142;
  t191_n141: return 0.06504737585783005f;
  t191_n142: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t191_n144;
  t191_n143: return 0.33949145674705505f;
  t191_n144: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t191_n146;
  t191_n145: return 0.02469533309340477f;
  t191_n146: return 0.1036999449133873f;
  t191_n147: if (E.getSymbolScopeDistanceCost() >= 3242196992 /*12.0*/) goto t191_n149;
  t191_n148: return -0.19052083790302277f;
  t191_n149: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t191_n151;
  t191_n150: return -0.13374362885951996f;
  t191_n151: return 0.5570861101150513f;
  t191_n152: return 0.02276836335659027f;
  t191_n153: if (E.getNumReferences() >= 3343124992 /*50186.0*/) goto t191_n155;
  t191_n154: return -0.008519125171005726f;
  t191_n155: return -0.27487632632255554f;
  t191_n156: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t191_n180;
  t191_n157: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t191_n177;
  t191_n158: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t191_n176;
  t191_n159: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t191_n161;
  t191_n160: return 0.019542384892702103f;
  t191_n161: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t191_n163;
  t191_n162: return -0.02672327682375908f;
  t191_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t191_n175;
  t191_n164: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t191_n172;
  t191_n165: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t191_n171;
  t191_n166: if (E.getFractionNameInContext() >= 3207294391 /*0.6696428656578064*/) goto t191_n170;
  t191_n167: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t191_n169;
  t191_n168: return 0.05059902369976044f;
  t191_n169: return -0.15329496562480927f;
  t191_n170: return 0.10303480923175812f;
  t191_n171: return -0.025237826630473137f;
  t191_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t191_n174;
  t191_n173: return 0.10767263174057007f;
  t191_n174: return 0.05186007544398308f;
  t191_n175: return 0.10110106319189072f;
  t191_n176: return 0.04930591210722923f;
  t191_n177: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t191_n179;
  t191_n178: return 0.11147382855415344f;
  t191_n179: return -0.09838996082544327f;
  t191_n180: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t191_n246;
  t191_n181: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t191_n229;
  t191_n182: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t191_n222;
  t191_n183: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t191_n199;
  t191_n184: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t191_n188;
  t191_n185: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t191_n187;
  t191_n186: return 0.2507694363594055f;
  t191_n187: return 0.07164964079856873f;
  t191_n188: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t191_n198;
  t191_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t191_n191;
  t191_n190: return -0.010654940269887447f;
  t191_n191: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t191_n197;
  t191_n192: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t191_n194;
  t191_n193: return 0.005204018205404282f;
  t191_n194: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t191_n196;
  t191_n195: return 0.12181830406188965f;
  t191_n196: return 0.10642983764410019f;
  t191_n197: return 0.1594514101743698f;
  t191_n198: return 0.021581850945949554f;
  t191_n199: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t191_n211;
  t191_n200: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t191_n208;
  t191_n201: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t191_n203;
  t191_n202: return 0.015324401669204235f;
  t191_n203: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t191_n207;
  t191_n204: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t191_n206;
  t191_n205: return 0.12396284192800522f;
  t191_n206: return -0.1770140528678894f;
  t191_n207: return -0.12998969852924347f;
  t191_n208: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t191_n210;
  t191_n209: return 0.06221044063568115f;
  t191_n210: return -0.008946791291236877f;
  t191_n211: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t191_n213;
  t191_n212: return -0.23969663679599762f;
  t191_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t191_n221;
  t191_n214: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t191_n216;
  t191_n215: return -0.01927945576608181f;
  t191_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t191_n218;
  t191_n217: return 0.09938210994005203f;
  t191_n218: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t191_n220;
  t191_n219: return -0.06616475433111191f;
  t191_n220: return 0.10008955001831055f;
  t191_n221: return 0.09622704237699509f;
  t191_n222: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t191_n224;
  t191_n223: return 0.027801718562841415f;
  t191_n224: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t191_n226;
  t191_n225: return 0.12017948180437088f;
  t191_n226: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t191_n228;
  t191_n227: return 0.08166270703077316f;
  t191_n228: return -0.25574013590812683f;
  t191_n229: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t191_n245;
  t191_n230: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t191_n244;
  t191_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t191_n233;
  t191_n232: return 0.02204681746661663f;
  t191_n233: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t191_n243;
  t191_n234: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t191_n242;
  t191_n235: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t191_n237;
  t191_n236: return 0.15264949202537537f;
  t191_n237: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t191_n239;
  t191_n238: return -0.0005086538149043918f;
  t191_n239: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t191_n241;
  t191_n240: return 0.14088085293769836f;
  t191_n241: return 0.08891268074512482f;
  t191_n242: return 0.11708618700504303f;
  t191_n243: return -0.08780594170093536f;
  t191_n244: return -0.004684825427830219f;
  t191_n245: return -0.05571684241294861f;
  t191_n246: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t191_n250;
  t191_n247: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t191_n249;
  t191_n248: return 0.06743043661117554f;
  t191_n249: return -0.03488261252641678f;
  t191_n250: return -0.06064142286777496f;
  t191_n251: return 0.0012364051071926951f;
  t191_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t191_n254;
  t191_n253: return 0.0015288389986380935f;
  t191_n254: return 0.007191034499555826f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree192(const Example& E) {
  t192_n0: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t192_n244;
  t192_n1: if (E.getNumReferences() >= 3314254848 /*4464.5*/) goto t192_n243;
  t192_n2: if (E.getNumReferences() >= 3313252352 /*4035.5*/) goto t192_n238;
  t192_n3: if (E.getNumReferences() >= 3313096704 /*3997.5*/) goto t192_n235;
  t192_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t192_n212;
  t192_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n99;
  t192_n6: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t192_n20;
  t192_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t192_n19;
  t192_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t192_n14;
  t192_n9: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t192_n11;
  t192_n10: return 0.0031107845716178417f;
  t192_n11: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t192_n13;
  t192_n12: return 0.11282654851675034f;
  t192_n13: return -0.21648608148097992f;
  t192_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t192_n18;
  t192_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n17;
  t192_n16: return -0.01848263293504715f;
  t192_n17: return 0.01823270693421364f;
  t192_n18: return -0.0008264699135906994f;
  t192_n19: return 0.023092560470104218f;
  t192_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t192_n54;
  t192_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t192_n23;
  t192_n22: return -0.013626650907099247f;
  t192_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t192_n25;
  t192_n24: return -0.05266523361206055f;
  t192_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t192_n53;
  t192_n26: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t192_n50;
  t192_n27: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t192_n29;
  t192_n28: return 0.01742546632885933f;
  t192_n29: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t192_n49;
  t192_n30: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t192_n36;
  t192_n31: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t192_n35;
  t192_n32: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t192_n34;
  t192_n33: return 0.13068798184394836f;
  t192_n34: return 0.5006436109542847f;
  t192_n35: return 0.06553755700588226f;
  t192_n36: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t192_n42;
  t192_n37: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t192_n41;
  t192_n38: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t192_n40;
  t192_n39: return 0.05575646832585335f;
  t192_n40: return 0.10889049619436264f;
  t192_n41: return 0.02639535441994667f;
  t192_n42: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t192_n44;
  t192_n43: return 0.05797027796506882f;
  t192_n44: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t192_n48;
  t192_n45: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t192_n47;
  t192_n46: return 0.1073957085609436f;
  t192_n47: return 0.11494578421115875f;
  t192_n48: return 0.05738890916109085f;
  t192_n49: return 0.10236963629722595f;
  t192_n50: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t192_n52;
  t192_n51: return 0.04777558892965317f;
  t192_n52: return 0.07603845000267029f;
  t192_n53: return 0.006993937771767378f;
  t192_n54: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t192_n98;
  t192_n55: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t192_n93;
  t192_n56: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t192_n76;
  t192_n57: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t192_n75;
  t192_n58: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t192_n68;
  t192_n59: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t192_n61;
  t192_n60: return 0.013232005760073662f;
  t192_n61: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t192_n65;
  t192_n62: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t192_n64;
  t192_n63: return 0.21664948761463165f;
  t192_n64: return 0.05892309546470642f;
  t192_n65: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t192_n67;
  t192_n66: return 0.0040025063790380955f;
  t192_n67: return 0.2903074026107788f;
  t192_n68: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t192_n74;
  t192_n69: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t192_n73;
  t192_n70: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n72;
  t192_n71: return -0.02405446395277977f;
  t192_n72: return 0.11857564747333527f;
  t192_n73: return -0.5030319094657898f;
  t192_n74: return 0.09652301669120789f;
  t192_n75: return -0.011068685911595821f;
  t192_n76: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t192_n86;
  t192_n77: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t192_n85;
  t192_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t192_n82;
  t192_n79: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t192_n81;
  t192_n80: return 0.024132974445819855f;
  t192_n81: return 0.09657640755176544f;
  t192_n82: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n84;
  t192_n83: return 0.025606026872992516f;
  t192_n84: return -0.21139048039913177f;
  t192_n85: return -0.061473239213228226f;
  t192_n86: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t192_n90;
  t192_n87: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t192_n89;
  t192_n88: return -0.01661432720720768f;
  t192_n89: return 0.07359884679317474f;
  t192_n90: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t192_n92;
  t192_n91: return -0.2596028745174408f;
  t192_n92: return 0.08139261603355408f;
  t192_n93: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t192_n95;
  t192_n94: return -0.01984761655330658f;
  t192_n95: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t192_n97;
  t192_n96: return 0.003323004115372896f;
  t192_n97: return 0.06086413562297821f;
  t192_n98: return -0.025360403582453728f;
  t192_n99: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t192_n211;
  t192_n100: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t192_n210;
  t192_n101: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t192_n163;
  t192_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n126;
  t192_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t192_n107;
  t192_n104: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t192_n106;
  t192_n105: return -0.04003902152180672f;
  t192_n106: return -0.09047207236289978f;
  t192_n107: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t192_n113;
  t192_n108: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n110;
  t192_n109: return 0.0038328361697494984f;
  t192_n110: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t192_n112;
  t192_n111: return -0.11522255092859268f;
  t192_n112: return 0.048147521913051605f;
  t192_n113: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t192_n125;
  t192_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n124;
  t192_n115: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t192_n117;
  t192_n116: return -0.03041635826230049f;
  t192_n117: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t192_n119;
  t192_n118: return 0.01870795153081417f;
  t192_n119: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t192_n121;
  t192_n120: return -0.0157580878585577f;
  t192_n121: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t192_n123;
  t192_n122: return -0.01727977767586708f;
  t192_n123: return 0.15458731353282928f;
  t192_n124: return -0.07725842297077179f;
  t192_n125: return 0.03822210803627968f;
  t192_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown))) goto t192_n142;
  t192_n127: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t192_n141;
  t192_n128: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t192_n138;
  t192_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t192_n133;
  t192_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t192_n132;
  t192_n131: return -0.15146182477474213f;
  t192_n132: return -0.02403353340923786f;
  t192_n133: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t192_n137;
  t192_n134: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t192_n136;
  t192_n135: return -0.02343680150806904f;
  t192_n136: return 0.07472047954797745f;
  t192_n137: return -0.1936255246400833f;
  t192_n138: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t192_n140;
  t192_n139: return -0.023434320464730263f;
  t192_n140: return 0.09192901849746704f;
  t192_n141: return 0.0006269702571444213f;
  t192_n142: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t192_n160;
  t192_n143: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t192_n157;
  t192_n144: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t192_n152;
  t192_n145: if (E.getFractionNameInContext() >= 3211159142 /*0.8999999761581421*/) goto t192_n151;
  t192_n146: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t192_n150;
  t192_n147: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t192_n149;
  t192_n148: return 0.032342396676540375f;
  t192_n149: return -0.0038914713077247143f;
  t192_n150: return 0.10747905820608139f;
  t192_n151: return 0.2516767978668213f;
  t192_n152: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t192_n156;
  t192_n153: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t192_n155;
  t192_n154: return 0.042351581156253815f;
  t192_n155: return 0.08878026902675629f;
  t192_n156: return -0.2103716880083084f;
  t192_n157: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t192_n159;
  t192_n158: return 0.15751448273658752f;
  t192_n159: return 0.0594918318092823f;
  t192_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t192_n162;
  t192_n161: return 0.000703765545040369f;
  t192_n162: return 0.06370773166418076f;
  t192_n163: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t192_n195;
  t192_n164: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t192_n168;
  t192_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t192_n167;
  t192_n166: return -0.012287947349250317f;
  t192_n167: return 0.009890188463032246f;
  t192_n168: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t192_n194;
  t192_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t192_n171;
  t192_n170: return 0.017292805016040802f;
  t192_n171: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t192_n193;
  t192_n172: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t192_n174;
  t192_n173: return 0.007936708629131317f;
  t192_n174: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t192_n190;
  t192_n175: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n187;
  t192_n176: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t192_n178;
  t192_n177: return -0.02150803431868553f;
  t192_n178: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t192_n184;
  t192_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t192_n181;
  t192_n180: return -0.1048215925693512f;
  t192_n181: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t192_n183;
  t192_n182: return 0.16027319431304932f;
  t192_n183: return -0.0509113147854805f;
  t192_n184: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t192_n186;
  t192_n185: return 0.03454143553972244f;
  t192_n186: return 0.09618552029132843f;
  t192_n187: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t192_n189;
  t192_n188: return -0.03410370647907257f;
  t192_n189: return -0.523491382598877f;
  t192_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t192_n192;
  t192_n191: return -0.102369524538517f;
  t192_n192: return 0.10172583907842636f;
  t192_n193: return 0.24503645300865173f;
  t192_n194: return 0.18894246220588684f;
  t192_n195: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t192_n197;
  t192_n196: return 0.023051610216498375f;
  t192_n197: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t192_n207;
  t192_n198: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t192_n200;
  t192_n199: return -0.5403812527656555f;
  t192_n200: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t192_n206;
  t192_n201: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t192_n203;
  t192_n202: return 0.009141098707914352f;
  t192_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t192_n205;
  t192_n204: return 0.02366524003446102f;
  t192_n205: return 0.10770638287067413f;
  t192_n206: return -0.19442018866539001f;
  t192_n207: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t192_n209;
  t192_n208: return -0.02884571999311447f;
  t192_n209: return 0.1341688632965088f;
  t192_n210: return -0.02033214457333088f;
  t192_n211: return 0.0004922975786030293f;
  t192_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n222;
  t192_n213: if (E.getNumReferences() >= 3312590848 /*3874.0*/) goto t192_n221;
  t192_n214: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t192_n220;
  t192_n215: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t192_n217;
  t192_n216: return 0.000696358853019774f;
  t192_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n219;
  t192_n218: return -0.06203873082995415f;
  t192_n219: return -0.007753666955977678f;
  t192_n220: return 0.007949844934046268f;
  t192_n221: return 0.12241819500923157f;
  t192_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t192_n232;
  t192_n223: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t192_n225;
  t192_n224: return 0.007065048441290855f;
  t192_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t192_n227;
  t192_n226: return -0.007124327588826418f;
  t192_n227: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t192_n231;
  t192_n228: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t192_n230;
  t192_n229: return 0.04803675040602684f;
  t192_n230: return 0.10036325454711914f;
  t192_n231: return 0.1342354565858841f;
  t192_n232: if (E.getNumReferences() >= 3310284800 /*3311.0*/) goto t192_n234;
  t192_n233: return 0.017413429915905f;
  t192_n234: return -0.21833547949790955f;
  t192_n235: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t192_n237;
  t192_n236: return 0.11248493194580078f;
  t192_n237: return -0.3177371919155121f;
  t192_n238: if (E.getNumReferences() >= 3313733632 /*4210.0*/) goto t192_n240;
  t192_n239: return -0.1432752162218094f;
  t192_n240: if (E.getNumReferences() >= 3313820672 /*4252.5*/) goto t192_n242;
  t192_n241: return 0.11711981147527695f;
  t192_n242: return -0.02593368850648403f;
  t192_n243: return 0.010247895494103432f;
  t192_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t192_n246;
  t192_n245: return -0.0009224828681908548f;
  t192_n246: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t192_n248;
  t192_n247: return 0.13257476687431335f;
  t192_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t192_n250;
  t192_n249: return 0.009203577414155006f;
  t192_n250: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t192_n252;
  t192_n251: return 0.0036220799665898085f;
  t192_n252: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t192_n254;
  t192_n253: return 0.03698424622416496f;
  t192_n254: return 0.10177680104970932f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree193(const Example& E) {
  t193_n0: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t193_n254;
  t193_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t193_n69;
  t193_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t193_n4;
  t193_n3: return -0.0004869515832979232f;
  t193_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t193_n68;
  t193_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t193_n39;
  t193_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t193_n8;
  t193_n7: return -0.01828569918870926f;
  t193_n8: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t193_n32;
  t193_n9: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t193_n31;
  t193_n10: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t193_n30;
  t193_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t193_n13;
  t193_n12: return -0.0072114490903913975f;
  t193_n13: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t193_n15;
  t193_n14: return -0.02390456199645996f;
  t193_n15: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t193_n17;
  t193_n16: return 0.01127170491963625f;
  t193_n17: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t193_n27;
  t193_n18: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t193_n26;
  t193_n19: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t193_n25;
  t193_n20: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t193_n24;
  t193_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t193_n23;
  t193_n22: return -0.05817655101418495f;
  t193_n23: return 0.017820855602622032f;
  t193_n24: return -0.0008456968935206532f;
  t193_n25: return 0.0834072157740593f;
  t193_n26: return -0.05610714852809906f;
  t193_n27: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t193_n29;
  t193_n28: return -0.033365171402692795f;
  t193_n29: return -0.3122960031032562f;
  t193_n30: return -0.0406905822455883f;
  t193_n31: return 0.0009847148321568966f;
  t193_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n34;
  t193_n33: return -0.0062074060551822186f;
  t193_n34: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t193_n36;
  t193_n35: return 0.003866686951369047f;
  t193_n36: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t193_n38;
  t193_n37: return 0.029123982414603233f;
  t193_n38: return -0.008355313912034035f;
  t193_n39: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t193_n43;
  t193_n40: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t193_n42;
  t193_n41: return 0.008883552625775337f;
  t193_n42: return 0.002546918112784624f;
  t193_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t193_n63;
  t193_n44: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t193_n46;
  t193_n45: return -0.03437727317214012f;
  t193_n46: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t193_n52;
  t193_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n49;
  t193_n48: return -0.0757371261715889f;
  t193_n49: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t193_n51;
  t193_n50: return 0.06737309694290161f;
  t193_n51: return 0.0045654913410544395f;
  t193_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t193_n54;
  t193_n53: return -0.027068641036748886f;
  t193_n54: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t193_n56;
  t193_n55: return -0.05920550599694252f;
  t193_n56: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t193_n58;
  t193_n57: return 0.04663396626710892f;
  t193_n58: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t193_n60;
  t193_n59: return -0.10780718177556992f;
  t193_n60: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t193_n62;
  t193_n61: return 0.04727755859494209f;
  t193_n62: return -0.004188746213912964f;
  t193_n63: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t193_n67;
  t193_n64: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t193_n66;
  t193_n65: return 0.007057482842355967f;
  t193_n66: return -0.005482710897922516f;
  t193_n67: return -0.12799513339996338f;
  t193_n68: return 0.0038496372289955616f;
  t193_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n85;
  t193_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t193_n80;
  t193_n71: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t193_n79;
  t193_n72: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t193_n74;
  t193_n73: return -0.002645618049427867f;
  t193_n74: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t193_n78;
  t193_n75: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t193_n77;
  t193_n76: return -0.02430642955005169f;
  t193_n77: return 0.03360642492771149f;
  t193_n78: return 0.0823514387011528f;
  t193_n79: return 0.011082030832767487f;
  t193_n80: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t193_n82;
  t193_n81: return -0.01964961178600788f;
  t193_n82: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t193_n84;
  t193_n83: return 0.010805357247591019f;
  t193_n84: return 0.027793429791927338f;
  t193_n85: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t193_n253;
  t193_n86: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t193_n250;
  t193_n87: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t193_n101;
  t193_n88: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t193_n100;
  t193_n89: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t193_n99;
  t193_n90: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t193_n94;
  t193_n91: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t193_n93;
  t193_n92: return 0.07414309680461884f;
  t193_n93: return -0.0060271527618169785f;
  t193_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t193_n96;
  t193_n95: return -0.05287497490644455f;
  t193_n96: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t193_n98;
  t193_n97: return 0.07280150055885315f;
  t193_n98: return -0.1210918128490448f;
  t193_n99: return -0.18906204402446747f;
  t193_n100: return 0.008781728334724903f;
  t193_n101: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t193_n249;
  t193_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t193_n234;
  t193_n103: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t193_n197;
  t193_n104: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t193_n114;
  t193_n105: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t193_n107;
  t193_n106: return -0.026585204526782036f;
  t193_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t193_n113;
  t193_n108: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t193_n110;
  t193_n109: return 0.005800396669656038f;
  t193_n110: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t193_n112;
  t193_n111: return 0.06323083490133286f;
  t193_n112: return 0.149654820561409f;
  t193_n113: return 0.015750279650092125f;
  t193_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t193_n184;
  t193_n115: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t193_n181;
  t193_n116: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t193_n118;
  t193_n117: return -0.014297941699624062f;
  t193_n118: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t193_n130;
  t193_n119: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t193_n127;
  t193_n120: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t193_n126;
  t193_n121: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t193_n123;
  t193_n122: return 0.11079856008291245f;
  t193_n123: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t193_n125;
  t193_n124: return 0.12538614869117737f;
  t193_n125: return 0.04516371339559555f;
  t193_n126: return -0.037078410387039185f;
  t193_n127: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t193_n129;
  t193_n128: return 0.02090710587799549f;
  t193_n129: return 0.10194102674722672f;
  t193_n130: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t193_n136;
  t193_n131: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t193_n133;
  t193_n132: return -0.004938047379255295f;
  t193_n133: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t193_n135;
  t193_n134: return -0.4479513466358185f;
  t193_n135: return -0.14045922458171844f;
  t193_n136: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t193_n142;
  t193_n137: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t193_n139;
  t193_n138: return 0.043702755123376846f;
  t193_n139: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t193_n141;
  t193_n140: return 0.09513063728809357f;
  t193_n141: return 0.10597009211778641f;
  t193_n142: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t193_n158;
  t193_n143: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t193_n145;
  t193_n144: return -0.00688102887943387f;
  t193_n145: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t193_n151;
  t193_n146: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t193_n150;
  t193_n147: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t193_n149;
  t193_n148: return 0.09731817245483398f;
  t193_n149: return 0.12514643371105194f;
  t193_n150: return 0.02320677973330021f;
  t193_n151: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t193_n153;
  t193_n152: return 0.0009545695502310991f;
  t193_n153: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t193_n157;
  t193_n154: if (E.getFractionNameInContext() >= 3197839050 /*0.3030303120613098*/) goto t193_n156;
  t193_n155: return 0.11735201627016068f;
  t193_n156: return 0.11349374055862427f;
  t193_n157: return -0.9150100946426392f;
  t193_n158: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t193_n162;
  t193_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n161;
  t193_n160: return -0.04997211694717407f;
  t193_n161: return 0.0028899714816361666f;
  t193_n162: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t193_n174;
  t193_n163: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t193_n165;
  t193_n164: return -0.06482527405023575f;
  t193_n165: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t193_n171;
  t193_n166: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t193_n170;
  t193_n167: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t193_n169;
  t193_n168: return 0.031130582094192505f;
  t193_n169: return 0.13208049535751343f;
  t193_n170: return 0.021319909021258354f;
  t193_n171: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t193_n173;
  t193_n172: return 0.039564620703458786f;
  t193_n173: return 0.07614981383085251f;
  t193_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t193_n180;
  t193_n175: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t193_n177;
  t193_n176: return 0.06957592070102692f;
  t193_n177: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t193_n179;
  t193_n178: return -0.5156329274177551f;
  t193_n179: return -0.057770732790231705f;
  t193_n180: return 0.0796804279088974f;
  t193_n181: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t193_n183;
  t193_n182: return -0.17219488322734833f;
  t193_n183: return -0.09164302051067352f;
  t193_n184: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t193_n190;
  t193_n185: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t193_n187;
  t193_n186: return 0.04538460448384285f;
  t193_n187: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t193_n189;
  t193_n188: return 0.1120653748512268f;
  t193_n189: return 0.07302875816822052f;
  t193_n190: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t193_n194;
  t193_n191: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t193_n193;
  t193_n192: return 0.053728535771369934f;
  t193_n193: return -0.1364806890487671f;
  t193_n194: if (E.getFractionNameInContext() >= 3202118087 /*0.4305555522441864*/) goto t193_n196;
  t193_n195: return -0.0975569412112236f;
  t193_n196: return 0.06963296979665756f;
  t193_n197: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t193_n233;
  t193_n198: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t193_n232;
  t193_n199: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t193_n219;
  t193_n200: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t193_n218;
  t193_n201: if (E.getNumReferences() >= 3259891712 /*51.5*/) goto t193_n217;
  t193_n202: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t193_n216;
  t193_n203: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t193_n215;
  t193_n204: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t193_n212;
  t193_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n207;
  t193_n206: return -0.09957689046859741f;
  t193_n207: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t193_n211;
  t193_n208: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t193_n210;
  t193_n209: return 0.2430713027715683f;
  t193_n210: return -0.11181192845106125f;
  t193_n211: return 0.15858472883701324f;
  t193_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t193_n214;
  t193_n213: return 0.2573169767856598f;
  t193_n214: return 0.30288854241371155f;
  t193_n215: return 0.094693623483181f;
  t193_n216: return 0.03283674642443657f;
  t193_n217: return 0.2755756080150604f;
  t193_n218: return -0.03345348313450813f;
  t193_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t193_n221;
  t193_n220: return -0.13437318801879883f;
  t193_n221: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t193_n229;
  t193_n222: if (E.getNumReferences() >= 3259236352 /*49.0*/) goto t193_n224;
  t193_n223: return -0.017295369878411293f;
  t193_n224: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t193_n226;
  t193_n225: return -0.005435221828520298f;
  t193_n226: if (E.getNumReferences() >= 3263430656 /*66.0*/) goto t193_n228;
  t193_n227: return 0.11961319297552109f;
  t193_n228: return 0.08198021352291107f;
  t193_n229: if (E.getNumReferences() >= 3246391296 /*16.0*/) goto t193_n231;
  t193_n230: return 0.12412726879119873f;
  t193_n231: return -0.03768032416701317f;
  t193_n232: return -0.00024659206974320114f;
  t193_n233: return -0.018779195845127106f;
  t193_n234: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t193_n242;
  t193_n235: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t193_n237;
  t193_n236: return -0.004296578001230955f;
  t193_n237: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t193_n241;
  t193_n238: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t193_n240;
  t193_n239: return -0.024699296802282333f;
  t193_n240: return 0.011857856996357441f;
  t193_n241: return -0.052023448050022125f;
  t193_n242: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t193_n246;
  t193_n243: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t193_n245;
  t193_n244: return 0.02568429708480835f;
  t193_n245: return -0.04814942181110382f;
  t193_n246: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t193_n248;
  t193_n247: return -0.03507460281252861f;
  t193_n248: return 0.022808266803622246f;
  t193_n249: return 0.006376327946782112f;
  t193_n250: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t193_n252;
  t193_n251: return 0.012796763330698013f;
  t193_n252: return 0.1561160534620285f;
  t193_n253: return 0.021538201719522476f;
  t193_n254: return -0.05134141817688942f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree194(const Example& E) {
  t194_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t194_n252;
  t194_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t194_n3;
  t194_n2: return -0.0006425904575735331f;
  t194_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t194_n225;
  t194_n4: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t194_n6;
  t194_n5: return 0.1130964532494545f;
  t194_n6: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t194_n8;
  t194_n7: return -0.007448021322488785f;
  t194_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t194_n222;
  t194_n9: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t194_n219;
  t194_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace))) goto t194_n144;
  t194_n11: if (E.getNumReferences() >= 3346517376 /*63437.5*/) goto t194_n141;
  t194_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t194_n34;
  t194_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n33;
  t194_n14: if (E.getNumReferences() >= 3333870848 /*23402.5*/) goto t194_n30;
  t194_n15: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t194_n17;
  t194_n16: return -0.05225350335240364f;
  t194_n17: if (E.getNumReferences() >= 3330932480 /*17663.5*/) goto t194_n25;
  t194_n18: if (E.getNumReferences() >= 3329498112 /*15623.0*/) goto t194_n24;
  t194_n19: if (E.getNumReferences() >= 3328390656 /*14541.5*/) goto t194_n21;
  t194_n20: return -0.021753421053290367f;
  t194_n21: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t194_n23;
  t194_n22: return 0.23682402074337006f;
  t194_n23: return 0.08197347819805145f;
  t194_n24: return -0.12381922453641891f;
  t194_n25: if (E.getNumReferences() >= 3331016704 /*17828.0*/) goto t194_n29;
  t194_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t194_n28;
  t194_n27: return -0.016145089641213417f;
  t194_n28: return 0.138849675655365f;
  t194_n29: return -0.10713367164134979f;
  t194_n30: if (E.getNumReferences() >= 3333898752 /*23457.0*/) goto t194_n32;
  t194_n31: return -0.38677382469177246f;
  t194_n32: return -0.07829026877880096f;
  t194_n33: return 0.0005175556871108711f;
  t194_n34: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t194_n68;
  t194_n35: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t194_n37;
  t194_n36: return 0.0007118361536413431f;
  t194_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t194_n59;
  t194_n38: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t194_n42;
  t194_n39: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t194_n41;
  t194_n40: return 0.00577399879693985f;
  t194_n41: return -0.011347253806889057f;
  t194_n42: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t194_n56;
  t194_n43: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t194_n47;
  t194_n44: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t194_n46;
  t194_n45: return -0.04844105616211891f;
  t194_n46: return 0.006458897143602371f;
  t194_n47: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t194_n51;
  t194_n48: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t194_n50;
  t194_n49: return -0.2919263541698456f;
  t194_n50: return -0.014154493808746338f;
  t194_n51: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t194_n53;
  t194_n52: return -0.9942768812179565f;
  t194_n53: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t194_n55;
  t194_n54: return 0.032953955233097076f;
  t194_n55: return -0.06055543199181557f;
  t194_n56: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t194_n58;
  t194_n57: return 0.04969668760895729f;
  t194_n58: return -0.048340052366256714f;
  t194_n59: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t194_n65;
  t194_n60: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t194_n62;
  t194_n61: return -0.005168418399989605f;
  t194_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n64;
  t194_n63: return -0.056358180940151215f;
  t194_n64: return 0.01529211737215519f;
  t194_n65: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t194_n67;
  t194_n66: return -0.011684948578476906f;
  t194_n67: return -0.07155384123325348f;
  t194_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n70;
  t194_n69: return -0.0940387174487114f;
  t194_n70: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t194_n94;
  t194_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n81;
  t194_n72: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t194_n76;
  t194_n73: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t194_n75;
  t194_n74: return 0.044086210429668427f;
  t194_n75: return -0.05317151919007301f;
  t194_n76: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t194_n80;
  t194_n77: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t194_n79;
  t194_n78: return -0.0018560588359832764f;
  t194_n79: return 0.040395841002464294f;
  t194_n80: return -0.027490384876728058f;
  t194_n81: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t194_n85;
  t194_n82: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t194_n84;
  t194_n83: return -0.017220495268702507f;
  t194_n84: return 0.01408112421631813f;
  t194_n85: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t194_n87;
  t194_n86: return -0.007997567765414715f;
  t194_n87: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t194_n89;
  t194_n88: return 0.0199980940669775f;
  t194_n89: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t194_n91;
  t194_n90: return -0.005251976661384106f;
  t194_n91: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t194_n93;
  t194_n92: return 0.01217347290366888f;
  t194_n93: return 0.0030040843412280083f;
  t194_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t194_n120;
  t194_n95: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t194_n97;
  t194_n96: return -0.0064571453258395195f;
  t194_n97: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t194_n111;
  t194_n98: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t194_n110;
  t194_n99: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t194_n103;
  t194_n100: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t194_n102;
  t194_n101: return -0.08143047988414764f;
  t194_n102: return 0.21242661774158478f;
  t194_n103: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t194_n107;
  t194_n104: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t194_n106;
  t194_n105: return 0.05310693383216858f;
  t194_n106: return -0.017967846244573593f;
  t194_n107: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t194_n109;
  t194_n108: return 0.11242944002151489f;
  t194_n109: return 0.03243285417556763f;
  t194_n110: return 0.00620965426787734f;
  t194_n111: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t194_n117;
  t194_n112: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t194_n114;
  t194_n113: return -0.006069194059818983f;
  t194_n114: if (E.getFileProximityDistanceCost() >= 3250061312 /*23.0*/) goto t194_n116;
  t194_n115: return 0.07539788633584976f;
  t194_n116: return 0.1718989759683609f;
  t194_n117: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t194_n119;
  t194_n118: return -0.0462660938501358f;
  t194_n119: return -1.12400484085083f;
  t194_n120: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t194_n130;
  t194_n121: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t194_n127;
  t194_n122: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t194_n126;
  t194_n123: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t194_n125;
  t194_n124: return 0.04896952956914902f;
  t194_n125: return -0.1327010691165924f;
  t194_n126: return 0.08419071137905121f;
  t194_n127: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t194_n129;
  t194_n128: return -0.061548423022031784f;
  t194_n129: return 0.09436729550361633f;
  t194_n130: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t194_n132;
  t194_n131: return 0.003117562737315893f;
  t194_n132: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t194_n134;
  t194_n133: return 0.01476798765361309f;
  t194_n134: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t194_n138;
  t194_n135: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t194_n137;
  t194_n136: return 0.062234945595264435f;
  t194_n137: return -0.2215394377708435f;
  t194_n138: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t194_n140;
  t194_n139: return -0.0731235146522522f;
  t194_n140: return 0.06615065038204193f;
  t194_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t194_n143;
  t194_n142: return 0.014919940382242203f;
  t194_n143: return 0.07810612767934799f;
  t194_n144: if (E.getNumReferences() >= 3361840416 /*231028.5*/) goto t194_n218;
  t194_n145: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t194_n151;
  t194_n146: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t194_n148;
  t194_n147: return 0.00776870409026742f;
  t194_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n150;
  t194_n149: return -0.07187674939632416f;
  t194_n150: return 0.016057845205068588f;
  t194_n151: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t194_n209;
  t194_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t194_n208;
  t194_n153: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t194_n173;
  t194_n154: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t194_n172;
  t194_n155: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t194_n157;
  t194_n156: return -0.05848480388522148f;
  t194_n157: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t194_n171;
  t194_n158: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t194_n160;
  t194_n159: return 0.029871230944991112f;
  t194_n160: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t194_n168;
  t194_n161: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t194_n167;
  t194_n162: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t194_n164;
  t194_n163: return 0.104869045317173f;
  t194_n164: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t194_n166;
  t194_n165: return -0.10551732778549194f;
  t194_n166: return 0.09194763004779816f;
  t194_n167: return 0.18427735567092896f;
  t194_n168: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t194_n170;
  t194_n169: return 0.003162878332659602f;
  t194_n170: return 0.06883955001831055f;
  t194_n171: return 0.005201449617743492f;
  t194_n172: return -0.0055602132342755795f;
  t194_n173: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t194_n199;
  t194_n174: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t194_n184;
  t194_n175: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t194_n181;
  t194_n176: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t194_n178;
  t194_n177: return -0.025665948167443275f;
  t194_n178: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t194_n180;
  t194_n179: return 0.41833874583244324f;
  t194_n180: return 0.0672958642244339f;
  t194_n181: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t194_n183;
  t194_n182: return 0.018177242949604988f;
  t194_n183: return -0.09004238247871399f;
  t194_n184: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t194_n198;
  t194_n185: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t194_n195;
  t194_n186: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t194_n194;
  t194_n187: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t194_n189;
  t194_n188: return 0.006357143633067608f;
  t194_n189: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t194_n193;
  t194_n190: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t194_n192;
  t194_n191: return 0.16386863589286804f;
  t194_n192: return 0.26340988278388977f;
  t194_n193: return 0.06669547408819199f;
  t194_n194: return 0.1467401087284088f;
  t194_n195: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t194_n197;
  t194_n196: return 0.23028583824634552f;
  t194_n197: return -0.16238345205783844f;
  t194_n198: return 0.028773004189133644f;
  t194_n199: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t194_n207;
  t194_n200: if (E.getFractionNameInContext() >= 3168244790 /*0.02631578966975212*/) goto t194_n202;
  t194_n201: return 0.056820787489414215f;
  t194_n202: if (E.getNumReferences() >= 3324623872 /*10863.0*/) goto t194_n204;
  t194_n203: return 0.04920356720685959f;
  t194_n204: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t194_n206;
  t194_n205: return -0.06671672314405441f;
  t194_n206: return -0.4762827455997467f;
  t194_n207: return 0.11926163733005524f;
  t194_n208: return -0.005674379877746105f;
  t194_n209: if (E.getNumReferences() >= 3340972032 /*41776.0*/) goto t194_n215;
  t194_n210: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t194_n214;
  t194_n211: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t194_n213;
  t194_n212: return -0.019569043070077896f;
  t194_n213: return -0.09901886433362961f;
  t194_n214: return -0.009594209492206573f;
  t194_n215: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t194_n217;
  t194_n216: return 0.08121444284915924f;
  t194_n217: return -0.14154264330863953f;
  t194_n218: return -0.19606399536132812f;
  t194_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t194_n221;
  t194_n220: return 0.0033819081727415323f;
  t194_n221: return 0.038657527416944504f;
  t194_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t194_n224;
  t194_n223: return 0.0013479480985552073f;
  t194_n224: return 0.009702064096927643f;
  t194_n225: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t194_n247;
  t194_n226: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t194_n228;
  t194_n227: return 0.004472353961318731f;
  t194_n228: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t194_n246;
  t194_n229: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t194_n245;
  t194_n230: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t194_n244;
  t194_n231: if (E.getFractionNameInContext() >= 3209042018 /*0.773809552192688*/) goto t194_n243;
  t194_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t194_n234;
  t194_n233: return -0.009214970283210278f;
  t194_n234: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t194_n242;
  t194_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t194_n241;
  t194_n236: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t194_n238;
  t194_n237: return 0.023165788501501083f;
  t194_n238: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t194_n240;
  t194_n239: return 0.11703576892614365f;
  t194_n240: return 0.002053510397672653f;
  t194_n241: return 0.11061833053827286f;
  t194_n242: return 0.043306462466716766f;
  t194_n243: return 0.11093796044588089f;
  t194_n244: return -0.006633542012423277f;
  t194_n245: return -0.047525495290756226f;
  t194_n246: return 0.10841835290193558f;
  t194_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t194_n251;
  t194_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t194_n250;
  t194_n249: return -0.015812719240784645f;
  t194_n250: return -0.0008995907846838236f;
  t194_n251: return 0.04457341507077217f;
  t194_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t194_n254;
  t194_n253: return 0.0013701822608709335f;
  t194_n254: return 0.004258192144334316f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree195(const Example& E) {
  t195_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t195_n188;
  t195_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t195_n157;
  t195_n2: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t195_n4;
  t195_n3: return -0.0005291019333526492f;
  t195_n4: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t195_n42;
  t195_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t195_n11;
  t195_n6: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t195_n8;
  t195_n7: return -0.011448034085333347f;
  t195_n8: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t195_n10;
  t195_n9: return 0.11844799667596817f;
  t195_n10: return 0.08641844987869263f;
  t195_n11: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t195_n41;
  t195_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t195_n14;
  t195_n13: return -0.022365011274814606f;
  t195_n14: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t195_n40;
  t195_n15: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t195_n23;
  t195_n16: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t195_n18;
  t195_n17: return 0.012313726358115673f;
  t195_n18: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t195_n22;
  t195_n19: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t195_n21;
  t195_n20: return 0.26608192920684814f;
  t195_n21: return -0.20962746441364288f;
  t195_n22: return 0.35819005966186523f;
  t195_n23: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t195_n25;
  t195_n24: return 0.17835475504398346f;
  t195_n25: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t195_n29;
  t195_n26: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t195_n28;
  t195_n27: return -0.10043896734714508f;
  t195_n28: return 0.18493026494979858f;
  t195_n29: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t195_n31;
  t195_n30: return -0.10064045339822769f;
  t195_n31: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t195_n33;
  t195_n32: return 0.2107526808977127f;
  t195_n33: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t195_n35;
  t195_n34: return -0.10030188411474228f;
  t195_n35: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t195_n37;
  t195_n36: return 0.2273554652929306f;
  t195_n37: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t195_n39;
  t195_n38: return 0.21685607731342316f;
  t195_n39: return 0.150767982006073f;
  t195_n40: return -0.05036154016852379f;
  t195_n41: return -0.06587959080934525f;
  t195_n42: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t195_n48;
  t195_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t195_n47;
  t195_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n46;
  t195_n45: return -0.010623293928802013f;
  t195_n46: return 0.006675906013697386f;
  t195_n47: return 0.12304722517728806f;
  t195_n48: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t195_n50;
  t195_n49: return 0.002814581850543618f;
  t195_n50: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t195_n90;
  t195_n51: if (E.getNumReferences() >= 3300507648 /*1486.0*/) goto t195_n89;
  t195_n52: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t195_n66;
  t195_n53: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t195_n59;
  t195_n54: if (E.getNumReferences() >= 3277783040 /*223.0*/) goto t195_n58;
  t195_n55: if (E.getNumReferences() >= 3275816960 /*193.0*/) goto t195_n57;
  t195_n56: return 0.04601756110787392f;
  t195_n57: return 0.1406306028366089f;
  t195_n58: return -0.08252658694982529f;
  t195_n59: if (E.getNumReferences() >= 3284205568 /*386.0*/) goto t195_n63;
  t195_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t195_n62;
  t195_n61: return -0.30402612686157227f;
  t195_n62: return 0.14024138450622559f;
  t195_n63: if (E.getNumReferences() >= 3287580672 /*489.0*/) goto t195_n65;
  t195_n64: return 0.31258076429367065f;
  t195_n65: return -0.1505764275789261f;
  t195_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t195_n74;
  t195_n67: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t195_n69;
  t195_n68: return -0.06741736084222794f;
  t195_n69: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t195_n71;
  t195_n70: return 0.04971923306584358f;
  t195_n71: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t195_n73;
  t195_n72: return 0.14029347896575928f;
  t195_n73: return 0.18660195171833038f;
  t195_n74: if (E.getNumReferences() >= 3275096064 /*182.0*/) goto t195_n76;
  t195_n75: return 0.0034538975451141596f;
  t195_n76: if (E.getNumReferences() >= 3279847424 /*254.5*/) goto t195_n82;
  t195_n77: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t195_n79;
  t195_n78: return -0.020769542083144188f;
  t195_n79: if (E.getNumReferences() >= 3276931072 /*210.0*/) goto t195_n81;
  t195_n80: return 0.1724250316619873f;
  t195_n81: return 0.28222861886024475f;
  t195_n82: if (E.getNumReferences() >= 3290071040 /*618.0*/) goto t195_n84;
  t195_n83: return -0.04260357469320297f;
  t195_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t195_n86;
  t195_n85: return 0.20203207433223724f;
  t195_n86: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t195_n88;
  t195_n87: return 0.20745441317558289f;
  t195_n88: return 0.1074204072356224f;
  t195_n89: return -0.0477195680141449f;
  t195_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t195_n150;
  t195_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t195_n93;
  t195_n92: return -0.02421608567237854f;
  t195_n93: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t195_n145;
  t195_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t195_n102;
  t195_n95: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t195_n97;
  t195_n96: return -0.006985208950936794f;
  t195_n97: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t195_n101;
  t195_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n100;
  t195_n99: return -0.10799015313386917f;
  t195_n100: return 0.07862184941768646f;
  t195_n101: return 0.010093514807522297f;
  t195_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n104;
  t195_n103: return -0.10215632617473602f;
  t195_n104: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t195_n138;
  t195_n105: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t195_n137;
  t195_n106: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t195_n136;
  t195_n107: if (E.getNumReferences() >= 3325292544 /*11516.0*/) goto t195_n115;
  t195_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t195_n110;
  t195_n109: return 0.018119895830750465f;
  t195_n110: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t195_n114;
  t195_n111: if (E.getNumReferences() >= 3317206016 /*5905.5*/) goto t195_n113;
  t195_n112: return 0.06453955173492432f;
  t195_n113: return 0.15713514387607574f;
  t195_n114: return 0.02416124939918518f;
  t195_n115: if (E.getNumReferences() >= 3332617216 /*20954.0*/) goto t195_n135;
  t195_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t195_n118;
  t195_n117: return 0.03728553280234337f;
  t195_n118: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t195_n120;
  t195_n119: return -0.22745950520038605f;
  t195_n120: if (E.getNumReferences() >= 3326347776 /*12546.5*/) goto t195_n128;
  t195_n121: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t195_n123;
  t195_n122: return 0.2348020374774933f;
  t195_n123: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t195_n125;
  t195_n124: return -0.1012069508433342f;
  t195_n125: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t195_n127;
  t195_n126: return 0.1275615692138672f;
  t195_n127: return 0.08688686043024063f;
  t195_n128: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t195_n134;
  t195_n129: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t195_n133;
  t195_n130: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t195_n132;
  t195_n131: return 0.1183847188949585f;
  t195_n132: return 0.14935339987277985f;
  t195_n133: return -0.13604667782783508f;
  t195_n134: return 0.1335977464914322f;
  t195_n135: return 0.00350101082585752f;
  t195_n136: return 0.13820073008537292f;
  t195_n137: return -0.026842813938856125f;
  t195_n138: if (E.getNumReferences() >= 3342293120 /*46936.5*/) goto t195_n144;
  t195_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n141;
  t195_n140: return -0.32118070125579834f;
  t195_n141: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t195_n143;
  t195_n142: return 0.1647445112466812f;
  t195_n143: return 0.045422300696372986f;
  t195_n144: return 0.049044281244277954f;
  t195_n145: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t195_n149;
  t195_n146: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t195_n148;
  t195_n147: return 0.11767003685235977f;
  t195_n148: return -2.3410415649414062f;
  t195_n149: return 0.018768873065710068f;
  t195_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t195_n152;
  t195_n151: return 0.012004474177956581f;
  t195_n152: if (E.getNumReferences() >= 3334782208 /*25182.5*/) goto t195_n156;
  t195_n153: if (E.getNumReferences() >= 3334246656 /*24136.5*/) goto t195_n155;
  t195_n154: return 0.10040543228387833f;
  t195_n155: return 0.10386563092470169f;
  t195_n156: return -0.12863294780254364f;
  t195_n157: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t195_n165;
  t195_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t195_n164;
  t195_n159: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t195_n161;
  t195_n160: return 0.0016188487643375993f;
  t195_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n163;
  t195_n162: return -0.11549711227416992f;
  t195_n163: return 0.0578371025621891f;
  t195_n164: return 0.02157369814813137f;
  t195_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t195_n177;
  t195_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t195_n170;
  t195_n167: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t195_n169;
  t195_n168: return -0.02415715530514717f;
  t195_n169: return 0.06276464462280273f;
  t195_n170: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t195_n174;
  t195_n171: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t195_n173;
  t195_n172: return 0.025410538539290428f;
  t195_n173: return -0.02094120718538761f;
  t195_n174: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t195_n176;
  t195_n175: return -0.002479705959558487f;
  t195_n176: return 0.04394986107945442f;
  t195_n177: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t195_n183;
  t195_n178: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t195_n182;
  t195_n179: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t195_n181;
  t195_n180: return 0.08728593587875366f;
  t195_n181: return 0.11702805012464523f;
  t195_n182: return -0.2449132651090622f;
  t195_n183: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t195_n185;
  t195_n184: return -0.0202270969748497f;
  t195_n185: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t195_n187;
  t195_n186: return 0.0837491974234581f;
  t195_n187: return -0.4797428548336029f;
  t195_n188: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t195_n234;
  t195_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n217;
  t195_n190: if (E.getFractionNameInContext() >= 3211904796 /*0.9444444179534912*/) goto t195_n210;
  t195_n191: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t195_n209;
  t195_n192: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t195_n206;
  t195_n193: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t195_n205;
  t195_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t195_n196;
  t195_n195: return -0.010100883431732655f;
  t195_n196: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t195_n204;
  t195_n197: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t195_n203;
  t195_n198: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t195_n202;
  t195_n199: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t195_n201;
  t195_n200: return 0.0040865372866392136f;
  t195_n201: return 0.053423333913087845f;
  t195_n202: return -0.11523899435997009f;
  t195_n203: return 0.10204746574163437f;
  t195_n204: return -0.4338230490684509f;
  t195_n205: return 0.03055795468389988f;
  t195_n206: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t195_n208;
  t195_n207: return -0.05425574257969856f;
  t195_n208: return 0.22839020192623138f;
  t195_n209: return -0.07832223176956177f;
  t195_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t195_n212;
  t195_n211: return -0.25577443838119507f;
  t195_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t195_n214;
  t195_n213: return 0.06768383085727692f;
  t195_n214: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t195_n216;
  t195_n215: return -0.13693971931934357f;
  t195_n216: return 0.08995359390974045f;
  t195_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Type))) goto t195_n219;
  t195_n218: return 0.002105038845911622f;
  t195_n219: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t195_n233;
  t195_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t195_n224;
  t195_n221: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t195_n223;
  t195_n222: return -0.02843489870429039f;
  t195_n223: return -0.23310591280460358f;
  t195_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t195_n230;
  t195_n225: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t195_n229;
  t195_n226: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t195_n228;
  t195_n227: return 0.0048200348392128944f;
  t195_n228: return 0.07791294902563095f;
  t195_n229: return -0.7702137231826782f;
  t195_n230: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t195_n232;
  t195_n231: return 0.018175238743424416f;
  t195_n232: return 0.08728780597448349f;
  t195_n233: return 0.06760736554861069f;
  t195_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t195_n238;
  t195_n235: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t195_n237;
  t195_n236: return -0.08496279269456863f;
  t195_n237: return -0.010405119508504868f;
  t195_n238: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t195_n250;
  t195_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t195_n247;
  t195_n240: if (E.getFractionNameInContext() >= 3211788288 /*0.9375*/) goto t195_n246;
  t195_n241: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t195_n243;
  t195_n242: return -0.07157766073942184f;
  t195_n243: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t195_n245;
  t195_n244: return 0.22159311175346375f;
  t195_n245: return 0.00897032581269741f;
  t195_n246: return 0.2144937366247177f;
  t195_n247: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t195_n249;
  t195_n248: return 0.21263571083545685f;
  t195_n249: return 0.3564177453517914f;
  t195_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t195_n252;
  t195_n251: return 0.04554670676589012f;
  t195_n252: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t195_n254;
  t195_n253: return 0.07435541599988937f;
  t195_n254: return 0.12963318824768066f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree196(const Example& E) {
  t196_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t196_n4;
  t196_n1: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t196_n3;
  t196_n2: return -0.0003412017540540546f;
  t196_n3: return -0.07664259523153305f;
  t196_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t196_n250;
  t196_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t196_n183;
  t196_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t196_n56;
  t196_n7: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t196_n33;
  t196_n8: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t196_n10;
  t196_n9: return -0.072549968957901f;
  t196_n10: if (E.getNumReferences() >= 3331977984 /*19705.5*/) goto t196_n32;
  t196_n11: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t196_n13;
  t196_n12: return -0.049240678548812866f;
  t196_n13: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t196_n15;
  t196_n14: return -0.11106114834547043f;
  t196_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n17;
  t196_n16: return -0.07388277351856232f;
  t196_n17: if (E.getNumReferences() >= 3329239040 /*15370.0*/) goto t196_n29;
  t196_n18: if (E.getNumReferences() >= 3326347776 /*12546.5*/) goto t196_n28;
  t196_n19: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t196_n21;
  t196_n20: return 0.13176195323467255f;
  t196_n21: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t196_n23;
  t196_n22: return 0.24048036336898804f;
  t196_n23: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t196_n25;
  t196_n24: return -0.10108854621648788f;
  t196_n25: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t196_n27;
  t196_n26: return 0.1239938959479332f;
  t196_n27: return 0.08304569125175476f;
  t196_n28: return -0.15692491829395294f;
  t196_n29: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t196_n31;
  t196_n30: return 0.13613219559192657f;
  t196_n31: return 0.12217681854963303f;
  t196_n32: return -0.10852864384651184f;
  t196_n33: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t196_n35;
  t196_n34: return 0.09836909919977188f;
  t196_n35: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t196_n49;
  t196_n36: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t196_n38;
  t196_n37: return 0.0024569081142544746f;
  t196_n38: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t196_n44;
  t196_n39: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t196_n41;
  t196_n40: return -0.2810101807117462f;
  t196_n41: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t196_n43;
  t196_n42: return 0.14584773778915405f;
  t196_n43: return -0.22354725003242493f;
  t196_n44: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t196_n46;
  t196_n45: return 0.11744293570518494f;
  t196_n46: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t196_n48;
  t196_n47: return -0.11257395893335342f;
  t196_n48: return 0.07986519485712051f;
  t196_n49: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t196_n51;
  t196_n50: return 0.023756757378578186f;
  t196_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t196_n55;
  t196_n52: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t196_n54;
  t196_n53: return 0.020094627514481544f;
  t196_n54: return 0.1333635300397873f;
  t196_n55: return 0.25004976987838745f;
  t196_n56: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t196_n80;
  t196_n57: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t196_n79;
  t196_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t196_n60;
  t196_n59: return -0.008764359168708324f;
  t196_n60: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t196_n70;
  t196_n61: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t196_n69;
  t196_n62: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t196_n64;
  t196_n63: return -0.05176571384072304f;
  t196_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n66;
  t196_n65: return -0.022735202684998512f;
  t196_n66: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t196_n68;
  t196_n67: return 0.00939659122377634f;
  t196_n68: return 0.09384074062108994f;
  t196_n69: return 0.009213489480316639f;
  t196_n70: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t196_n76;
  t196_n71: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t196_n75;
  t196_n72: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t196_n74;
  t196_n73: return -0.050015754997730255f;
  t196_n74: return -0.008028790354728699f;
  t196_n75: return 0.06720636785030365f;
  t196_n76: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t196_n78;
  t196_n77: return -0.1083933413028717f;
  t196_n78: return -0.33739984035491943f;
  t196_n79: return 0.0011440083617344499f;
  t196_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n118;
  t196_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t196_n83;
  t196_n82: return -0.09283813834190369f;
  t196_n83: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t196_n85;
  t196_n84: return 0.015784626826643944f;
  t196_n85: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t196_n91;
  t196_n86: if (E.getNumReferences() >= 3306027008 /*2271.5*/) goto t196_n90;
  t196_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n89;
  t196_n88: return -0.02132917381823063f;
  t196_n89: return 0.03159011900424957f;
  t196_n90: return 0.02241710014641285f;
  t196_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t196_n117;
  t196_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n116;
  t196_n93: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t196_n105;
  t196_n94: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t196_n104;
  t196_n95: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t196_n99;
  t196_n96: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t196_n98;
  t196_n97: return -0.0519026480615139f;
  t196_n98: return -0.6758073568344116f;
  t196_n99: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t196_n103;
  t196_n100: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t196_n102;
  t196_n101: return 0.0660930871963501f;
  t196_n102: return -0.015155814588069916f;
  t196_n103: return -0.049881983548402786f;
  t196_n104: return 0.12374754995107651f;
  t196_n105: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t196_n111;
  t196_n106: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t196_n108;
  t196_n107: return -0.05538380891084671f;
  t196_n108: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t196_n110;
  t196_n109: return 0.12171981483697891f;
  t196_n110: return -0.04405752569437027f;
  t196_n111: if (E.getNumReferences() >= 3379731776 /*993748.0*/) goto t196_n115;
  t196_n112: if (E.getNumReferences() >= 3301904384 /*1656.5*/) goto t196_n114;
  t196_n113: return -0.1288416087627411f;
  t196_n114: return 0.03237038105726242f;
  t196_n115: return -0.4539128839969635f;
  t196_n116: return 0.021404284983873367f;
  t196_n117: return 0.008649135008454323f;
  t196_n118: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t196_n182;
  t196_n119: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t196_n181;
  t196_n120: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t196_n134;
  t196_n121: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t196_n127;
  t196_n122: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t196_n124;
  t196_n123: return 0.006112335715442896f;
  t196_n124: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t196_n126;
  t196_n125: return -0.04956245794892311f;
  t196_n126: return 0.06889665871858597f;
  t196_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n133;
  t196_n128: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t196_n132;
  t196_n129: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t196_n131;
  t196_n130: return -0.031674616038799286f;
  t196_n131: return -0.20448505878448486f;
  t196_n132: return 0.033095233142375946f;
  t196_n133: return -0.0012646353570744395f;
  t196_n134: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t196_n162;
  t196_n135: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t196_n161;
  t196_n136: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t196_n150;
  t196_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n143;
  t196_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t196_n140;
  t196_n139: return -0.0315510630607605f;
  t196_n140: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t196_n142;
  t196_n141: return 0.05336759239435196f;
  t196_n142: return -0.0894051343202591f;
  t196_n143: if (E.getNumReferences() >= 3289530368 /*585.0*/) goto t196_n147;
  t196_n144: if (E.getNumReferences() >= 3285073920 /*412.5*/) goto t196_n146;
  t196_n145: return 0.01550234667956829f;
  t196_n146: return 0.13351145386695862f;
  t196_n147: if (E.getNumReferences() >= 3341778816 /*44927.5*/) goto t196_n149;
  t196_n148: return -0.09197713434696198f;
  t196_n149: return 0.020888635888695717f;
  t196_n150: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t196_n160;
  t196_n151: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t196_n159;
  t196_n152: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t196_n158;
  t196_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n157;
  t196_n154: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t196_n156;
  t196_n155: return -0.005045725032687187f;
  t196_n156: return 0.02992892451584339f;
  t196_n157: return 0.009335840120911598f;
  t196_n158: return -0.2903726398944855f;
  t196_n159: return 0.17990966141223907f;
  t196_n160: return 0.004442295059561729f;
  t196_n161: return 0.008491302840411663f;
  t196_n162: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t196_n178;
  t196_n163: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t196_n165;
  t196_n164: return -0.006418031640350819f;
  t196_n165: if (E.getNumReferences() >= 3327328256 /*13504.0*/) goto t196_n177;
  t196_n166: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t196_n176;
  t196_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t196_n169;
  t196_n168: return -0.09987008571624756f;
  t196_n169: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t196_n173;
  t196_n170: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t196_n172;
  t196_n171: return 0.5803961753845215f;
  t196_n172: return -0.11199715733528137f;
  t196_n173: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t196_n175;
  t196_n174: return -0.1776985228061676f;
  t196_n175: return 0.12770229578018188f;
  t196_n176: return -0.1868026703596115f;
  t196_n177: return -0.311826229095459f;
  t196_n178: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t196_n180;
  t196_n179: return 2.1433277130126953f;
  t196_n180: return -0.12712863087654114f;
  t196_n181: return 0.03401981294155121f;
  t196_n182: return -0.06184544414281845f;
  t196_n183: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t196_n191;
  t196_n184: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t196_n188;
  t196_n185: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t196_n187;
  t196_n186: return 0.007839771918952465f;
  t196_n187: return -0.009517744183540344f;
  t196_n188: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t196_n190;
  t196_n189: return 0.025145338848233223f;
  t196_n190: return 0.00653375219553709f;
  t196_n191: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t196_n193;
  t196_n192: return 0.0010247612372040749f;
  t196_n193: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t196_n223;
  t196_n194: if (E.getNumReferences() >= 3264741376 /*76.0*/) goto t196_n220;
  t196_n195: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t196_n219;
  t196_n196: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t196_n214;
  t196_n197: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t196_n209;
  t196_n198: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t196_n208;
  t196_n199: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t196_n205;
  t196_n200: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t196_n202;
  t196_n201: return 0.02313743159174919f;
  t196_n202: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t196_n204;
  t196_n203: return 0.056153759360313416f;
  t196_n204: return 0.10695923864841461f;
  t196_n205: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t196_n207;
  t196_n206: return 0.0675349161028862f;
  t196_n207: return -0.34560415148735046f;
  t196_n208: return -0.034074146300554276f;
  t196_n209: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t196_n211;
  t196_n210: return 0.03388367220759392f;
  t196_n211: if (E.getFractionNameInContext() >= 3209574628 /*0.8055555820465088*/) goto t196_n213;
  t196_n212: return -0.035254642367362976f;
  t196_n213: return 0.14070969820022583f;
  t196_n214: if (E.getFractionNameInContext() >= 3206814274 /*0.6410256624221802*/) goto t196_n216;
  t196_n215: return 0.07677700370550156f;
  t196_n216: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t196_n218;
  t196_n217: return -0.1460229605436325f;
  t196_n218: return 0.1276484876871109f;
  t196_n219: return -0.1676350235939026f;
  t196_n220: if (E.getNumReferences() >= 3281321984 /*298.0*/) goto t196_n222;
  t196_n221: return 0.11455193161964417f;
  t196_n222: return -0.018512627109885216f;
  t196_n223: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t196_n249;
  t196_n224: if (E.getFractionNameInContext() >= 3205017619 /*0.5339366793632507*/) goto t196_n228;
  t196_n225: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t196_n227;
  t196_n226: return -0.005726521369069815f;
  t196_n227: return 0.0704798549413681f;
  t196_n228: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t196_n238;
  t196_n229: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t196_n233;
  t196_n230: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t196_n232;
  t196_n231: return -0.005052131600677967f;
  t196_n232: return 0.10036672651767731f;
  t196_n233: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t196_n235;
  t196_n234: return 0.1321340948343277f;
  t196_n235: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t196_n237;
  t196_n236: return -0.26420271396636963f;
  t196_n237: return 0.1314634084701538f;
  t196_n238: if (E.getFractionNameInContext() >= 3209913561 /*0.8257575631141663*/) goto t196_n240;
  t196_n239: return -0.0050130621530115604f;
  t196_n240: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t196_n248;
  t196_n241: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t196_n247;
  t196_n242: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t196_n244;
  t196_n243: return 0.02587844803929329f;
  t196_n244: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t196_n246;
  t196_n245: return 0.11027451604604721f;
  t196_n246: return 0.15978623926639557f;
  t196_n247: return -0.1025056317448616f;
  t196_n248: return -0.17261044681072235f;
  t196_n249: return 0.05879093334078789f;
  t196_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t196_n252;
  t196_n251: return 0.0007689058547839522f;
  t196_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t196_n254;
  t196_n253: return 0.005542821250855923f;
  t196_n254: return 0.015165651217103004f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree197(const Example& E) {
  t197_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t197_n242;
  t197_n1: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t197_n227;
  t197_n2: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t197_n218;
  t197_n3: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t197_n191;
  t197_n4: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t197_n190;
  t197_n5: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t197_n183;
  t197_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t197_n166;
  t197_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t197_n119;
  t197_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t197_n34;
  t197_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t197_n11;
  t197_n10: return -0.11174117028713226f;
  t197_n11: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t197_n33;
  t197_n12: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t197_n32;
  t197_n13: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t197_n31;
  t197_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n18;
  t197_n15: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t197_n17;
  t197_n16: return -0.15061555802822113f;
  t197_n17: return 0.033484622836112976f;
  t197_n18: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t197_n30;
  t197_n19: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n29;
  t197_n20: if (E.getFractionNameInContext() >= 3211788288 /*0.9375*/) goto t197_n28;
  t197_n21: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t197_n25;
  t197_n22: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t197_n24;
  t197_n23: return 0.07940766215324402f;
  t197_n24: return 0.31952351331710815f;
  t197_n25: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t197_n27;
  t197_n26: return 0.016624048352241516f;
  t197_n27: return -0.05857019126415253f;
  t197_n28: return 0.13256074488162994f;
  t197_n29: return 0.038395680487155914f;
  t197_n30: return -0.004492184147238731f;
  t197_n31: return -0.004551617428660393f;
  t197_n32: return 0.01942620798945427f;
  t197_n33: return -0.0197600144892931f;
  t197_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n114;
  t197_n35: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t197_n101;
  t197_n36: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n100;
  t197_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t197_n97;
  t197_n38: if (E.getNumReferences() >= 3322621952 /*8908.0*/) goto t197_n94;
  t197_n39: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t197_n41;
  t197_n40: return 0.006337812636047602f;
  t197_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n47;
  t197_n42: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t197_n44;
  t197_n43: return -0.07612387090921402f;
  t197_n44: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t197_n46;
  t197_n45: return 0.11577123403549194f;
  t197_n46: return -0.28758206963539124f;
  t197_n47: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t197_n93;
  t197_n48: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t197_n92;
  t197_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n85;
  t197_n50: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n76;
  t197_n51: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t197_n57;
  t197_n52: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t197_n54;
  t197_n53: return -0.008691680617630482f;
  t197_n54: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t197_n56;
  t197_n55: return -0.06958567351102829f;
  t197_n56: return 0.21779631078243256f;
  t197_n57: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t197_n75;
  t197_n58: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t197_n66;
  t197_n59: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t197_n61;
  t197_n60: return 0.018016787245869637f;
  t197_n61: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t197_n65;
  t197_n62: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t197_n64;
  t197_n63: return 0.05693502351641655f;
  t197_n64: return 0.022212818264961243f;
  t197_n65: return -0.12473561614751816f;
  t197_n66: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t197_n68;
  t197_n67: return 0.19342568516731262f;
  t197_n68: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t197_n74;
  t197_n69: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t197_n71;
  t197_n70: return -0.015648603439331055f;
  t197_n71: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t197_n73;
  t197_n72: return -0.20803222060203552f;
  t197_n73: return -0.012991799041628838f;
  t197_n74: return 0.09969896823167801f;
  t197_n75: return -0.010564245283603668f;
  t197_n76: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t197_n84;
  t197_n77: if (E.getFractionNameInContext() >= 3187191692 /*0.12142857909202576*/) goto t197_n79;
  t197_n78: return -0.2773752510547638f;
  t197_n79: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t197_n81;
  t197_n80: return 0.0990620106458664f;
  t197_n81: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t197_n83;
  t197_n82: return -0.2477700263261795f;
  t197_n83: return 0.08846729248762131f;
  t197_n84: return -0.014623641967773438f;
  t197_n85: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t197_n91;
  t197_n86: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t197_n88;
  t197_n87: return 0.0029122577980160713f;
  t197_n88: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t197_n90;
  t197_n89: return 0.02058115229010582f;
  t197_n90: return 0.1220150738954544f;
  t197_n91: return 0.02034006640315056f;
  t197_n92: return 0.01694534532725811f;
  t197_n93: return -0.006402449682354927f;
  t197_n94: if (E.getNumReferences() >= 3324270080 /*10517.5*/) goto t197_n96;
  t197_n95: return 0.09211058169603348f;
  t197_n96: return 0.02555183134973049f;
  t197_n97: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t197_n99;
  t197_n98: return -0.018848232924938202f;
  t197_n99: return -0.13472598791122437f;
  t197_n100: return 0.018238835036754608f;
  t197_n101: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t197_n113;
  t197_n102: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t197_n112;
  t197_n103: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n111;
  t197_n104: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t197_n108;
  t197_n105: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t197_n107;
  t197_n106: return 0.023557601496577263f;
  t197_n107: return 0.05088384449481964f;
  t197_n108: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t197_n110;
  t197_n109: return -0.013265029527246952f;
  t197_n110: return -0.23807848989963531f;
  t197_n111: return -0.011575077660381794f;
  t197_n112: return 0.0025533856824040413f;
  t197_n113: return -0.02276710234582424f;
  t197_n114: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t197_n116;
  t197_n115: return 0.008120348677039146f;
  t197_n116: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t197_n118;
  t197_n117: return 0.30036795139312744f;
  t197_n118: return -0.10713932663202286f;
  t197_n119: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t197_n121;
  t197_n120: return -0.01863454096019268f;
  t197_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t197_n137;
  t197_n122: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n124;
  t197_n123: return -0.016740083694458008f;
  t197_n124: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t197_n126;
  t197_n125: return 0.007234441116452217f;
  t197_n126: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t197_n136;
  t197_n127: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t197_n135;
  t197_n128: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t197_n130;
  t197_n129: return 0.055795978754758835f;
  t197_n130: if (E.getNumReferences() >= 3295920128 /*975.0*/) goto t197_n134;
  t197_n131: if (E.getNumReferences() >= 3293454336 /*824.5*/) goto t197_n133;
  t197_n132: return 0.09348203241825104f;
  t197_n133: return 0.11444955319166183f;
  t197_n134: return 0.04668626934289932f;
  t197_n135: return 0.004245678428560495f;
  t197_n136: return 0.017922282218933105f;
  t197_n137: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t197_n165;
  t197_n138: if (E.getNumReferences() >= 3318124544 /*6354.0*/) goto t197_n148;
  t197_n139: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n143;
  t197_n140: if (E.getNumReferences() >= 3311583232 /*3628.0*/) goto t197_n142;
  t197_n141: return 0.034252241253852844f;
  t197_n142: return -0.0784505233168602f;
  t197_n143: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n145;
  t197_n144: return -0.007443028502166271f;
  t197_n145: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t197_n147;
  t197_n146: return -0.05220261588692665f;
  t197_n147: return -0.2462652623653412f;
  t197_n148: if (E.getNumReferences() >= 3323090944 /*9366.0*/) goto t197_n154;
  t197_n149: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t197_n153;
  t197_n150: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n152;
  t197_n151: return 0.11755277216434479f;
  t197_n152: return 0.10601307451725006f;
  t197_n153: return 0.13561825454235077f;
  t197_n154: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n160;
  t197_n155: if (E.getNumReferences() >= 3331587072 /*18942.0*/) goto t197_n157;
  t197_n156: return -0.09057671576738358f;
  t197_n157: if (E.getNumReferences() >= 3331859712 /*19474.5*/) goto t197_n159;
  t197_n158: return 0.0762505903840065f;
  t197_n159: return 0.014221047051250935f;
  t197_n160: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t197_n162;
  t197_n161: return 0.12618905305862427f;
  t197_n162: if (E.getNumReferences() >= 3333771776 /*23209.0*/) goto t197_n164;
  t197_n163: return 0.11251845210790634f;
  t197_n164: return 0.10961496829986572f;
  t197_n165: return -0.001530256005935371f;
  t197_n166: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t197_n168;
  t197_n167: return 0.0018790953326970339f;
  t197_n168: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n182;
  t197_n169: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t197_n171;
  t197_n170: return 0.007516471669077873f;
  t197_n171: if (E.getNumReferences() >= 3352190016 /*105656.5*/) goto t197_n181;
  t197_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t197_n174;
  t197_n173: return -0.009061217308044434f;
  t197_n174: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t197_n180;
  t197_n175: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t197_n177;
  t197_n176: return 0.013753573410212994f;
  t197_n177: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t197_n179;
  t197_n178: return -0.10145102441310883f;
  t197_n179: return 0.10919713973999023f;
  t197_n180: return 0.06899033486843109f;
  t197_n181: return 0.0851869210600853f;
  t197_n182: return 0.014284585602581501f;
  t197_n183: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t197_n189;
  t197_n184: if (E.getNumReferences() >= 3367700464 /*383039.5*/) goto t197_n188;
  t197_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t197_n187;
  t197_n186: return -0.01018136739730835f;
  t197_n187: return 0.005932672414928675f;
  t197_n188: return -0.05498312786221504f;
  t197_n189: return 0.007370860781520605f;
  t197_n190: return 0.005846699699759483f;
  t197_n191: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t197_n205;
  t197_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n196;
  t197_n193: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t197_n195;
  t197_n194: return -0.024056408554315567f;
  t197_n195: return 0.11844649165868759f;
  t197_n196: if (E.getNumReferences() >= 3307055104 /*2522.5*/) goto t197_n198;
  t197_n197: return -0.0021278795320540667f;
  t197_n198: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t197_n204;
  t197_n199: if (E.getNumReferences() >= 3327250432 /*13428.0*/) goto t197_n203;
  t197_n200: if (E.getNumReferences() >= 3319356416 /*6955.5*/) goto t197_n202;
  t197_n201: return 0.03134620934724808f;
  t197_n202: return 0.08822289854288101f;
  t197_n203: return 0.01618955470621586f;
  t197_n204: return 0.00792042538523674f;
  t197_n205: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t197_n215;
  t197_n206: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t197_n208;
  t197_n207: return -0.0004433112917467952f;
  t197_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n210;
  t197_n209: return -0.04357446730136871f;
  t197_n210: if (E.getNumReferences() >= 3294396416 /*882.0*/) goto t197_n212;
  t197_n211: return 0.08218182623386383f;
  t197_n212: if (E.getNumReferences() >= 3306844160 /*2471.0*/) goto t197_n214;
  t197_n213: return 0.13776296377182007f;
  t197_n214: return 0.06867194920778275f;
  t197_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n217;
  t197_n216: return -0.002464119577780366f;
  t197_n217: return 0.005950771737843752f;
  t197_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t197_n222;
  t197_n219: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t197_n221;
  t197_n220: return -0.020479997619986534f;
  t197_n221: return 0.05841001868247986f;
  t197_n222: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t197_n226;
  t197_n223: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t197_n225;
  t197_n224: return -0.004336223471909761f;
  t197_n225: return 0.011134203523397446f;
  t197_n226: return 0.008105127140879631f;
  t197_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t197_n229;
  t197_n228: return -0.03249881789088249f;
  t197_n229: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t197_n237;
  t197_n230: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t197_n234;
  t197_n231: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t197_n233;
  t197_n232: return 0.03555655479431152f;
  t197_n233: return 0.05626047030091286f;
  t197_n234: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t197_n236;
  t197_n235: return 0.011476516723632812f;
  t197_n236: return 0.09445960074663162f;
  t197_n237: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t197_n239;
  t197_n238: return -0.036123842000961304f;
  t197_n239: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t197_n241;
  t197_n240: return 0.09477530419826508f;
  t197_n241: return 0.06981803476810455f;
  t197_n242: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t197_n246;
  t197_n243: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t197_n245;
  t197_n244: return 0.11178795993328094f;
  t197_n245: return 0.0007466818206012249f;
  t197_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t197_n248;
  t197_n247: return 0.001025406294502318f;
  t197_n248: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t197_n250;
  t197_n249: return 0.006502903066575527f;
  t197_n250: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t197_n254;
  t197_n251: if (E.getFractionNameInContext() >= 3196938455 /*0.27619048953056335*/) goto t197_n253;
  t197_n252: return 0.08227382600307465f;
  t197_n253: return 0.03248634934425354f;
  t197_n254: return -0.10696577280759811f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree198(const Example& E) {
  t198_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t198_n254;
  t198_n1: if (E.getNumReferences() >= 3306514432 /*2390.5*/) goto t198_n9;
  t198_n2: if (E.getNumReferences() >= 3306424320 /*2368.5*/) goto t198_n4;
  t198_n3: return -0.00038719255826435983f;
  t198_n4: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t198_n6;
  t198_n5: return -0.08040565997362137f;
  t198_n6: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t198_n8;
  t198_n7: return -0.11440043896436691f;
  t198_n8: return -1.0417593717575073f;
  t198_n9: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t198_n29;
  t198_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n12;
  t198_n11: return -0.048923444002866745f;
  t198_n12: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t198_n28;
  t198_n13: if (E.getNumReferences() >= 3321825280 /*8161.0*/) goto t198_n27;
  t198_n14: if (E.getNumReferences() >= 3320384512 /*7457.5*/) goto t198_n16;
  t198_n15: return 0.03265361860394478f;
  t198_n16: if (E.getNumReferences() >= 3321112576 /*7813.0*/) goto t198_n24;
  t198_n17: if (E.getNumReferences() >= 3320945664 /*7731.5*/) goto t198_n23;
  t198_n18: if (E.getNumReferences() >= 3320802304 /*7661.5*/) goto t198_n22;
  t198_n19: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t198_n21;
  t198_n20: return 0.14546900987625122f;
  t198_n21: return 0.07790983468294144f;
  t198_n22: return -0.1093016192317009f;
  t198_n23: return 0.16355739533901215f;
  t198_n24: if (E.getNumReferences() >= 3321802752 /*8150.0*/) goto t198_n26;
  t198_n25: return -0.015856178477406502f;
  t198_n26: return 0.15589509904384613f;
  t198_n27: return 0.00896135251969099f;
  t198_n28: return 0.007887957617640495f;
  t198_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n35;
  t198_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t198_n32;
  t198_n31: return -0.018210219219326973f;
  t198_n32: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t198_n34;
  t198_n33: return 0.032688457518815994f;
  t198_n34: return -0.08667563647031784f;
  t198_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t198_n133;
  t198_n36: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t198_n40;
  t198_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n39;
  t198_n38: return -0.010306079871952534f;
  t198_n39: return 0.12648726999759674f;
  t198_n40: if (E.getNumReferences() >= 3322891776 /*9171.5*/) goto t198_n80;
  t198_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t198_n67;
  t198_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t198_n44;
  t198_n43: return -0.01848246157169342f;
  t198_n44: if (E.getNumReferences() >= 3321030656 /*7773.0*/) goto t198_n56;
  t198_n45: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t198_n55;
  t198_n46: if (E.getNumReferences() >= 3318628352 /*6600.0*/) goto t198_n48;
  t198_n47: return 0.02167152240872383f;
  t198_n48: if (E.getNumReferences() >= 3318811648 /*6689.5*/) goto t198_n54;
  t198_n49: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t198_n53;
  t198_n50: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t198_n52;
  t198_n51: return 0.10250318050384521f;
  t198_n52: return 0.1467350423336029f;
  t198_n53: return 0.11480661481618881f;
  t198_n54: return 0.16525477170944214f;
  t198_n55: return -0.009750226512551308f;
  t198_n56: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t198_n66;
  t198_n57: if (E.getNumReferences() >= 3322669568 /*8954.5*/) goto t198_n65;
  t198_n58: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t198_n64;
  t198_n59: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t198_n61;
  t198_n60: return 0.09486217796802521f;
  t198_n61: if (E.getNumReferences() >= 3322096640 /*8395.0*/) goto t198_n63;
  t198_n62: return 0.18050158023834229f;
  t198_n63: return 0.19578291475772858f;
  t198_n64: return 0.12290336191654205f;
  t198_n65: return -0.1716231405735016f;
  t198_n66: return 0.03529202938079834f;
  t198_n67: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t198_n69;
  t198_n68: return -0.01858900487422943f;
  t198_n69: if (E.getNumReferences() >= 3317616640 /*6106.0*/) goto t198_n79;
  t198_n70: if (E.getNumReferences() >= 3317563392 /*6080.0*/) goto t198_n72;
  t198_n71: return 0.039408739656209946f;
  t198_n72: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t198_n76;
  t198_n73: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t198_n75;
  t198_n74: return -0.10473596304655075f;
  t198_n75: return 0.23747633397579193f;
  t198_n76: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t198_n78;
  t198_n77: return 0.11831367015838623f;
  t198_n78: return 0.15356779098510742f;
  t198_n79: return 0.02752767503261566f;
  t198_n80: if (E.getNumReferences() >= 3330825472 /*17454.5*/) goto t198_n82;
  t198_n81: return -0.03749651834368706f;
  t198_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n130;
  t198_n83: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t198_n115;
  t198_n84: if (E.getNumReferences() >= 3332353792 /*20439.5*/) goto t198_n96;
  t198_n85: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n95;
  t198_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t198_n94;
  t198_n87: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t198_n93;
  t198_n88: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t198_n92;
  t198_n89: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t198_n91;
  t198_n90: return 0.0866374745965004f;
  t198_n91: return 0.1662793755531311f;
  t198_n92: return 0.06265241652727127f;
  t198_n93: return 0.07087390124797821f;
  t198_n94: return 0.14954879879951477f;
  t198_n95: return 0.11707014590501785f;
  t198_n96: if (E.getNumReferences() >= 3345992448 /*61387.0*/) goto t198_n102;
  t198_n97: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t198_n99;
  t198_n98: return -0.032317519187927246f;
  t198_n99: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t198_n101;
  t198_n100: return -0.08065730333328247f;
  t198_n101: return -0.21920792758464813f;
  t198_n102: if (E.getNumReferences() >= 3358377760 /*176924.5*/) goto t198_n114;
  t198_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t198_n105;
  t198_n104: return 0.01279636099934578f;
  t198_n105: if (E.getNumReferences() >= 3357529824 /*163675.5*/) goto t198_n111;
  t198_n106: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n110;
  t198_n107: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t198_n109;
  t198_n108: return 0.09480782598257065f;
  t198_n109: return -0.12401001900434494f;
  t198_n110: return 0.008423037827014923f;
  t198_n111: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t198_n113;
  t198_n112: return 0.10637466609477997f;
  t198_n113: return 0.1445990800857544f;
  t198_n114: return 0.0066303289495408535f;
  t198_n115: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t198_n129;
  t198_n116: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t198_n118;
  t198_n117: return -0.08512052893638611f;
  t198_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t198_n122;
  t198_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t198_n121;
  t198_n120: return -0.1201157346367836f;
  t198_n121: return 0.1351759284734726f;
  t198_n122: if (E.getNumReferences() >= 3334111488 /*23872.5*/) goto t198_n128;
  t198_n123: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t198_n127;
  t198_n124: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t198_n126;
  t198_n125: return 0.11936190724372864f;
  t198_n126: return 0.18732967972755432f;
  t198_n127: return 0.09989219158887863f;
  t198_n128: return 0.08846797049045563f;
  t198_n129: return -0.08295423537492752f;
  t198_n130: if (E.getNumReferences() >= 3331031296 /*17856.5*/) goto t198_n132;
  t198_n131: return 0.09020231664180756f;
  t198_n132: return 0.005795105826109648f;
  t198_n133: if (E.getNumReferences() >= 3316164608 /*5397.0*/) goto t198_n225;
  t198_n134: if (E.getNumReferences() >= 3315997696 /*5315.5*/) goto t198_n220;
  t198_n135: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t198_n137;
  t198_n136: return 0.1563563048839569f;
  t198_n137: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t198_n187;
  t198_n138: if (E.getNumReferences() >= 3312912384 /*3952.5*/) goto t198_n184;
  t198_n139: if (E.getNumReferences() >= 3310782464 /*3432.5*/) goto t198_n155;
  t198_n140: if (E.getNumReferences() >= 3306967040 /*2501.0*/) goto t198_n148;
  t198_n141: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t198_n147;
  t198_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n146;
  t198_n143: if (E.getNumReferences() >= 3306846208 /*2471.5*/) goto t198_n145;
  t198_n144: return 0.09355437755584717f;
  t198_n145: return 0.11581064760684967f;
  t198_n146: return 0.017291897907853127f;
  t198_n147: return 0.128237783908844f;
  t198_n148: if (E.getNumReferences() >= 3307317248 /*2586.5*/) goto t198_n150;
  t198_n149: return -0.0911623165011406f;
  t198_n150: if (E.getNumReferences() >= 3308132352 /*2785.5*/) goto t198_n154;
  t198_n151: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t198_n153;
  t198_n152: return 0.06330819427967072f;
  t198_n153: return 0.11186978965997696f;
  t198_n154: return -0.00509650819003582f;
  t198_n155: if (E.getNumReferences() >= 3312549888 /*3864.0*/) goto t198_n181;
  t198_n156: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n172;
  t198_n157: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t198_n163;
  t198_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t198_n160;
  t198_n159: return -0.006341014988720417f;
  t198_n160: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t198_n162;
  t198_n161: return 0.2348732203245163f;
  t198_n162: return -0.012575551867485046f;
  t198_n163: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t198_n169;
  t198_n164: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t198_n166;
  t198_n165: return 0.16444604098796844f;
  t198_n166: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t198_n168;
  t198_n167: return 0.16834519803524017f;
  t198_n168: return 0.12988042831420898f;
  t198_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t198_n171;
  t198_n170: return -0.3822127878665924f;
  t198_n171: return 0.15990985929965973f;
  t198_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t198_n174;
  t198_n173: return 0.05527827888727188f;
  t198_n174: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t198_n180;
  t198_n175: if (E.getNumReferences() >= 3310870528 /*3454.0*/) goto t198_n179;
  t198_n176: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t198_n178;
  t198_n177: return 0.4778553545475006f;
  t198_n178: return 0.12792301177978516f;
  t198_n179: return 0.09834177792072296f;
  t198_n180: return 0.14428597688674927f;
  t198_n181: if (E.getNumReferences() >= 3312828416 /*3932.0*/) goto t198_n183;
  t198_n182: return -0.08995255082845688f;
  t198_n183: return 0.092605821788311f;
  t198_n184: if (E.getNumReferences() >= 3313016832 /*3978.0*/) goto t198_n186;
  t198_n185: return -0.1922687590122223f;
  t198_n186: return -0.03154255449771881f;
  t198_n187: if (E.getNumReferences() >= 3315294208 /*4972.0*/) goto t198_n211;
  t198_n188: if (E.getNumReferences() >= 3315181568 /*4917.0*/) goto t198_n208;
  t198_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t198_n191;
  t198_n190: return 0.05731843411922455f;
  t198_n191: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t198_n195;
  t198_n192: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n194;
  t198_n193: return 0.0028048246167600155f;
  t198_n194: return 0.1558714509010315f;
  t198_n195: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t198_n205;
  t198_n196: if (E.getNumReferences() >= 3314460672 /*4565.0*/) goto t198_n198;
  t198_n197: return 0.0429706871509552f;
  t198_n198: if (E.getNumReferences() >= 3314592768 /*4629.5*/) goto t198_n204;
  t198_n199: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t198_n203;
  t198_n200: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t198_n202;
  t198_n201: return 0.10356330126523972f;
  t198_n202: return 0.13366329669952393f;
  t198_n203: return 0.04757817089557648f;
  t198_n204: return 0.08242396265268326f;
  t198_n205: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t198_n207;
  t198_n206: return 0.15958184003829956f;
  t198_n207: return -0.2231290638446808f;
  t198_n208: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t198_n210;
  t198_n209: return 0.13754142820835114f;
  t198_n210: return 0.09901484847068787f;
  t198_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t198_n217;
  t198_n212: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n216;
  t198_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t198_n215;
  t198_n214: return -0.007780204527080059f;
  t198_n215: return 0.223271444439888f;
  t198_n216: return -0.05114929378032684f;
  t198_n217: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t198_n219;
  t198_n218: return 0.06064114719629288f;
  t198_n219: return 0.14213119447231293f;
  t198_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t198_n222;
  t198_n221: return 0.09812329709529877f;
  t198_n222: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t198_n224;
  t198_n223: return 0.12269385159015656f;
  t198_n224: return -0.10083211958408356f;
  t198_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n227;
  t198_n226: return -0.002526323078200221f;
  t198_n227: if (E.getNumReferences() >= 3319705600 /*7126.0*/) goto t198_n231;
  t198_n228: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t198_n230;
  t198_n229: return -0.12497439980506897f;
  t198_n230: return -0.010261944495141506f;
  t198_n231: if (E.getNumReferences() >= 3319743488 /*7144.5*/) goto t198_n237;
  t198_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t198_n234;
  t198_n233: return 0.02043198049068451f;
  t198_n234: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t198_n236;
  t198_n235: return 0.0942467600107193f;
  t198_n236: return 0.17037084698677063f;
  t198_n237: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t198_n253;
  t198_n238: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t198_n242;
  t198_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n241;
  t198_n240: return -0.03684042766690254f;
  t198_n241: return 0.013723365031182766f;
  t198_n242: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t198_n252;
  t198_n243: if (E.getNumReferences() >= 3350496896 /*92429.0*/) goto t198_n245;
  t198_n244: return 0.015537606552243233f;
  t198_n245: if (E.getNumReferences() >= 3351669952 /*101593.5*/) goto t198_n249;
  t198_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t198_n248;
  t198_n247: return -0.020430048927664757f;
  t198_n248: return 0.08287185430526733f;
  t198_n249: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t198_n251;
  t198_n250: return -0.1639132797718048f;
  t198_n251: return 0.019880080595612526f;
  t198_n252: return 0.16007797420024872f;
  t198_n253: return -0.003885374404489994f;
  t198_n254: return 0.0015791459009051323f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree199(const Example& E) {
  t199_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t199_n182;
  t199_n1: if (E.getNumReferences() >= 3325613056 /*11829.0*/) goto t199_n59;
  t199_n2: if (E.getNumReferences() >= 3324152832 /*10403.0*/) goto t199_n44;
  t199_n3: if (E.getNumReferences() >= 3321408512 /*7957.5*/) goto t199_n5;
  t199_n4: return -0.0002648808585945517f;
  t199_n5: if (E.getNumReferences() >= 3321606144 /*8054.0*/) goto t199_n13;
  t199_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t199_n8;
  t199_n7: return 0.06913832575082779f;
  t199_n8: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t199_n12;
  t199_n9: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t199_n11;
  t199_n10: return 0.10684827715158463f;
  t199_n11: return 0.13752435147762299f;
  t199_n12: return 0.073671855032444f;
  t199_n13: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t199_n23;
  t199_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t199_n16;
  t199_n15: return 0.009429759345948696f;
  t199_n16: if (E.getNumReferences() >= 3324042752 /*10295.5*/) goto t199_n20;
  t199_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t199_n19;
  t199_n18: return -0.09221523255109787f;
  t199_n19: return 0.10978186130523682f;
  t199_n20: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t199_n22;
  t199_n21: return -0.10732836276292801f;
  t199_n22: return 0.30097508430480957f;
  t199_n23: if (E.getNumReferences() >= 3322180096 /*8476.5*/) goto t199_n31;
  t199_n24: if (E.getNumReferences() >= 3322037760 /*8337.5*/) goto t199_n26;
  t199_n25: return 0.1155930683016777f;
  t199_n26: if (E.getNumReferences() >= 3322075136 /*8374.0*/) goto t199_n28;
  t199_n27: return 0.11521108448505402f;
  t199_n28: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t199_n30;
  t199_n29: return 0.09400704503059387f;
  t199_n30: return 0.18127788603305817f;
  t199_n31: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t199_n39;
  t199_n32: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t199_n34;
  t199_n33: return -1.113555908203125f;
  t199_n34: if (E.getNumReferences() >= 3323953664 /*10208.5*/) goto t199_n38;
  t199_n35: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t199_n37;
  t199_n36: return 0.07021412253379822f;
  t199_n37: return -0.09048451483249664f;
  t199_n38: return -0.094937264919281f;
  t199_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t199_n41;
  t199_n40: return -0.025881938636302948f;
  t199_n41: if (E.getNumReferences() >= 3322985984 /*9263.5*/) goto t199_n43;
  t199_n42: return -0.038273077458143234f;
  t199_n43: return -0.41546547412872314f;
  t199_n44: if (E.getNumReferences() >= 3325485056 /*11704.0*/) goto t199_n58;
  t199_n45: if (E.getNumReferences() >= 3325383680 /*11605.0*/) goto t199_n47;
  t199_n46: return -0.0336308628320694f;
  t199_n47: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t199_n53;
  t199_n48: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t199_n52;
  t199_n49: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t199_n51;
  t199_n50: return 0.08183471113443375f;
  t199_n51: return 0.15889781713485718f;
  t199_n52: return -0.057440005242824554f;
  t199_n53: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t199_n57;
  t199_n54: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t199_n56;
  t199_n55: return 0.11211181432008743f;
  t199_n56: return -0.12207634001970291f;
  t199_n57: return 0.12377313524484634f;
  t199_n58: return -0.14716526865959167f;
  t199_n59: if (E.getNumReferences() >= 3325976064 /*12183.5*/) goto t199_n69;
  t199_n60: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t199_n64;
  t199_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t199_n63;
  t199_n62: return -0.0498972162604332f;
  t199_n63: return 0.0945366844534874f;
  t199_n64: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t199_n68;
  t199_n65: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t199_n67;
  t199_n66: return -0.08778355270624161f;
  t199_n67: return 0.16751603782176971f;
  t199_n68: return 0.10261396318674088f;
  t199_n69: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t199_n73;
  t199_n70: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t199_n72;
  t199_n71: return 0.012547843158245087f;
  t199_n72: return 0.002845478942617774f;
  t199_n73: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t199_n87;
  t199_n74: if (E.getNumReferences() >= 3328254464 /*14408.5*/) goto t199_n80;
  t199_n75: if (E.getNumReferences() >= 3328224256 /*14379.0*/) goto t199_n79;
  t199_n76: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t199_n78;
  t199_n77: return 0.06344368308782578f;
  t199_n78: return -0.03571648895740509f;
  t199_n79: return 0.16816666722297668f;
  t199_n80: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t199_n84;
  t199_n81: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t199_n83;
  t199_n82: return -0.0831647589802742f;
  t199_n83: return -0.05870756134390831f;
  t199_n84: if (E.getNumReferences() >= 3329200128 /*15332.0*/) goto t199_n86;
  t199_n85: return 0.08540216088294983f;
  t199_n86: return -0.20384210348129272f;
  t199_n87: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t199_n93;
  t199_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t199_n90;
  t199_n89: return 0.048390600830316544f;
  t199_n90: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t199_n92;
  t199_n91: return -0.026476679369807243f;
  t199_n92: return 0.10111124813556671f;
  t199_n93: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t199_n167;
  t199_n94: if (E.getNumReferences() >= 3341367040 /*43319.0*/) goto t199_n118;
  t199_n95: if (E.getNumReferences() >= 3334334464 /*24308.0*/) goto t199_n117;
  t199_n96: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t199_n108;
  t199_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t199_n99;
  t199_n98: return -0.11579279601573944f;
  t199_n99: if (E.getNumReferences() >= 3334015488 /*23685.0*/) goto t199_n107;
  t199_n100: if (E.getNumReferences() >= 3333976064 /*23608.0*/) goto t199_n102;
  t199_n101: return 0.01880032941699028f;
  t199_n102: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t199_n106;
  t199_n103: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t199_n105;
  t199_n104: return 0.1252559870481491f;
  t199_n105: return 0.12296354025602341f;
  t199_n106: return -0.0004912205040454865f;
  t199_n107: return -0.13921914994716644f;
  t199_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t199_n110;
  t199_n109: return 0.0747639611363411f;
  t199_n110: if (E.getFractionNameInContext() >= 3168244790 /*0.02631578966975212*/) goto t199_n116;
  t199_n111: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t199_n115;
  t199_n112: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t199_n114;
  t199_n113: return 0.0981503576040268f;
  t199_n114: return -0.05979914590716362f;
  t199_n115: return 0.11896085739135742f;
  t199_n116: return 0.0678127259016037f;
  t199_n117: return -0.017559872940182686f;
  t199_n118: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t199_n140;
  t199_n119: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t199_n123;
  t199_n120: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t199_n122;
  t199_n121: return 0.13895678520202637f;
  t199_n122: return -0.20655345916748047f;
  t199_n123: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t199_n127;
  t199_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t199_n126;
  t199_n125: return 0.04454958438873291f;
  t199_n126: return 0.12163547426462173f;
  t199_n127: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t199_n137;
  t199_n128: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t199_n136;
  t199_n129: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t199_n135;
  t199_n130: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t199_n132;
  t199_n131: return 0.13141347467899323f;
  t199_n132: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t199_n134;
  t199_n133: return 0.11747270077466965f;
  t199_n134: return 0.04602586477994919f;
  t199_n135: return -0.5052392482757568f;
  t199_n136: return -0.21366189420223236f;
  t199_n137: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t199_n139;
  t199_n138: return 0.08685965090990067f;
  t199_n139: return 0.11847605556249619f;
  t199_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t199_n160;
  t199_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t199_n143;
  t199_n142: return -0.0779196098446846f;
  t199_n143: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t199_n153;
  t199_n144: if (E.getNumReferences() >= 3342343040 /*47131.5*/) goto t199_n146;
  t199_n145: return -0.18109968304634094f;
  t199_n146: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t199_n148;
  t199_n147: return -0.019218483939766884f;
  t199_n148: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t199_n152;
  t199_n149: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t199_n151;
  t199_n150: return 0.05389577895402908f;
  t199_n151: return 0.11171698570251465f;
  t199_n152: return 0.059555284678936005f;
  t199_n153: if (E.getNumReferences() >= 3344025984 /*53705.5*/) goto t199_n155;
  t199_n154: return -0.04541151598095894f;
  t199_n155: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t199_n157;
  t199_n156: return -0.09887386858463287f;
  t199_n157: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t199_n159;
  t199_n158: return 0.0956440418958664f;
  t199_n159: return 0.1348256915807724f;
  t199_n160: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t199_n162;
  t199_n161: return -0.17402631044387817f;
  t199_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t199_n164;
  t199_n163: return -0.10337964445352554f;
  t199_n164: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t199_n166;
  t199_n165: return 0.14729495346546173f;
  t199_n166: return 0.1426984816789627f;
  t199_n167: if (E.getNumReferences() >= 3345144064 /*58073.0*/) goto t199_n169;
  t199_n168: return -0.16256868839263916f;
  t199_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t199_n171;
  t199_n170: return -0.03204693645238876f;
  t199_n171: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t199_n181;
  t199_n172: if (E.getNumReferences() >= 3364003440 /*267507.5*/) goto t199_n178;
  t199_n173: if (E.getNumReferences() >= 3346497664 /*63360.5*/) goto t199_n177;
  t199_n174: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t199_n176;
  t199_n175: return 0.017071129754185677f;
  t199_n176: return 0.0741877630352974f;
  t199_n177: return -0.005761079955846071f;
  t199_n178: if (E.getNumReferences() >= 3364184064 /*273152.0*/) goto t199_n180;
  t199_n179: return 0.07143931090831757f;
  t199_n180: return 0.028869060799479485f;
  t199_n181: return -0.0672452375292778f;
  t199_n182: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t199_n250;
  t199_n183: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t199_n241;
  t199_n184: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t199_n216;
  t199_n185: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t199_n213;
  t199_n186: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t199_n212;
  t199_n187: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t199_n191;
  t199_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t199_n190;
  t199_n189: return 0.02077498286962509f;
  t199_n190: return 0.07568582147359848f;
  t199_n191: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t199_n203;
  t199_n192: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t199_n196;
  t199_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t199_n195;
  t199_n194: return -0.16502133011817932f;
  t199_n195: return 0.022670915350317955f;
  t199_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t199_n202;
  t199_n197: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t199_n199;
  t199_n198: return 0.09508075565099716f;
  t199_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t199_n201;
  t199_n200: return -0.18911436200141907f;
  t199_n201: return 0.10095015913248062f;
  t199_n202: return -0.03541313484311104f;
  t199_n203: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t199_n211;
  t199_n204: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t199_n210;
  t199_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t199_n209;
  t199_n206: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t199_n208;
  t199_n207: return -0.04681047797203064f;
  t199_n208: return 0.10921111702919006f;
  t199_n209: return 0.045622024685144424f;
  t199_n210: return 0.11923570185899734f;
  t199_n211: return -0.35645127296447754f;
  t199_n212: return 0.03935063257813454f;
  t199_n213: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t199_n215;
  t199_n214: return -0.006105945445597172f;
  t199_n215: return 0.027182137593626976f;
  t199_n216: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t199_n220;
  t199_n217: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t199_n219;
  t199_n218: return 0.0052520208992064f;
  t199_n219: return -0.14129363000392914f;
  t199_n220: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t199_n240;
  t199_n221: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t199_n239;
  t199_n222: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t199_n234;
  t199_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t199_n229;
  t199_n224: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t199_n228;
  t199_n225: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t199_n227;
  t199_n226: return -0.05573830008506775f;
  t199_n227: return -0.30166271328926086f;
  t199_n228: return 0.0721949115395546f;
  t199_n229: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t199_n233;
  t199_n230: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t199_n232;
  t199_n231: return 0.041522808372974396f;
  t199_n232: return 0.2354302853345871f;
  t199_n233: return -0.22896620631217957f;
  t199_n234: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t199_n238;
  t199_n235: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t199_n237;
  t199_n236: return 0.09895510226488113f;
  t199_n237: return 0.10430239886045456f;
  t199_n238: return -0.04587729647755623f;
  t199_n239: return -0.06624897569417953f;
  t199_n240: return 0.019355857744812965f;
  t199_n241: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t199_n243;
  t199_n242: return 0.11702247709035873f;
  t199_n243: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t199_n249;
  t199_n244: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t199_n248;
  t199_n245: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t199_n247;
  t199_n246: return 0.06680310517549515f;
  t199_n247: return -0.04577428475022316f;
  t199_n248: return 0.15737968683242798f;
  t199_n249: return -0.002159544499590993f;
  t199_n250: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t199_n252;
  t199_n251: return 0.0013880640035495162f;
  t199_n252: if (E.getNumReferences() >= 3362781216 /*245728.5*/) goto t199_n254;
  t199_n253: return 0.04731578752398491f;
  t199_n254: return -0.25325247645378113f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree200(const Example& E) {
  t200_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t200_n214;
  t200_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t200_n211;
  t200_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t200_n10;
  t200_n3: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t200_n5;
  t200_n4: return -0.0005493570351973176f;
  t200_n5: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t200_n9;
  t200_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n8;
  t200_n7: return -0.3743409216403961f;
  t200_n8: return -0.056388139724731445f;
  t200_n9: return 0.030685530975461006f;
  t200_n10: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t200_n184;
  t200_n11: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t200_n13;
  t200_n12: return -0.00752274738624692f;
  t200_n13: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t200_n61;
  t200_n14: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t200_n18;
  t200_n15: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t200_n17;
  t200_n16: return 0.11074768751859665f;
  t200_n17: return 0.0003199976054020226f;
  t200_n18: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t200_n52;
  t200_n19: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n43;
  t200_n20: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t200_n40;
  t200_n21: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t200_n35;
  t200_n22: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t200_n26;
  t200_n23: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t200_n25;
  t200_n24: return -0.10961279273033142f;
  t200_n25: return 0.09257571399211884f;
  t200_n26: if (E.getFractionNameInContext() >= 3207459552 /*0.6794872283935547*/) goto t200_n34;
  t200_n27: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t200_n29;
  t200_n28: return -0.04952898994088173f;
  t200_n29: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t200_n31;
  t200_n30: return 0.13630810379981995f;
  t200_n31: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t200_n33;
  t200_n32: return -0.016616294160485268f;
  t200_n33: return 0.10757581889629364f;
  t200_n34: return -0.010180031880736351f;
  t200_n35: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t200_n39;
  t200_n36: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t200_n38;
  t200_n37: return 0.09520462900400162f;
  t200_n38: return 0.11181201785802841f;
  t200_n39: return -0.0004221016715746373f;
  t200_n40: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t200_n42;
  t200_n41: return -1.1208853721618652f;
  t200_n42: return -0.13454459607601166f;
  t200_n43: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t200_n51;
  t200_n44: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t200_n48;
  t200_n45: if (E.getFractionNameInContext() >= 3209341610 /*0.7916666269302368*/) goto t200_n47;
  t200_n46: return -0.0038869346026331186f;
  t200_n47: return 0.11032743006944656f;
  t200_n48: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t200_n50;
  t200_n49: return 0.06657810509204865f;
  t200_n50: return 0.12589582800865173f;
  t200_n51: return 0.19459116458892822f;
  t200_n52: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t200_n56;
  t200_n53: if (E.getFractionNameInContext() >= 3200115678 /*0.3708791136741638*/) goto t200_n55;
  t200_n54: return 0.10209432244300842f;
  t200_n55: return 0.012197231873869896f;
  t200_n56: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t200_n60;
  t200_n57: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n59;
  t200_n58: return 0.10671738535165787f;
  t200_n59: return -0.0705072358250618f;
  t200_n60: return -0.30222517251968384f;
  t200_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t200_n183;
  t200_n62: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t200_n64;
  t200_n63: return 0.0005501427222043276f;
  t200_n64: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t200_n146;
  t200_n65: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t200_n143;
  t200_n66: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t200_n136;
  t200_n67: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t200_n131;
  t200_n68: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t200_n72;
  t200_n69: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t200_n71;
  t200_n70: return -0.0017490320606157184f;
  t200_n71: return -0.056157249957323074f;
  t200_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n90;
  t200_n73: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t200_n89;
  t200_n74: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t200_n80;
  t200_n75: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t200_n77;
  t200_n76: return -0.06546413153409958f;
  t200_n77: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t200_n79;
  t200_n78: return -0.2856169641017914f;
  t200_n79: return 0.10782059282064438f;
  t200_n80: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t200_n88;
  t200_n81: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t200_n87;
  t200_n82: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t200_n86;
  t200_n83: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t200_n85;
  t200_n84: return 0.10291864722967148f;
  t200_n85: return 0.059281863272190094f;
  t200_n86: return 0.017465075477957726f;
  t200_n87: return -0.0030733859166502953f;
  t200_n88: return 0.09509793668985367f;
  t200_n89: return 0.03780914470553398f;
  t200_n90: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t200_n114;
  t200_n91: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t200_n101;
  t200_n92: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t200_n96;
  t200_n93: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t200_n95;
  t200_n94: return -0.019312238320708275f;
  t200_n95: return -0.13438718020915985f;
  t200_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n98;
  t200_n97: return -0.49414944648742676f;
  t200_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t200_n100;
  t200_n99: return -0.3637961447238922f;
  t200_n100: return 0.00463513471186161f;
  t200_n101: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t200_n105;
  t200_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t200_n104;
  t200_n103: return 0.005959650501608849f;
  t200_n104: return 0.10119857639074326f;
  t200_n105: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t200_n113;
  t200_n106: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t200_n108;
  t200_n107: return -0.01725548505783081f;
  t200_n108: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t200_n110;
  t200_n109: return -0.11192993819713593f;
  t200_n110: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t200_n112;
  t200_n111: return -0.7524662017822266f;
  t200_n112: return 0.03887379914522171f;
  t200_n113: return 0.14409375190734863f;
  t200_n114: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t200_n130;
  t200_n115: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t200_n129;
  t200_n116: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t200_n126;
  t200_n117: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t200_n119;
  t200_n118: return 0.1134374663233757f;
  t200_n119: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t200_n121;
  t200_n120: return 0.014012754894793034f;
  t200_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n123;
  t200_n122: return -0.13704897463321686f;
  t200_n123: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t200_n125;
  t200_n124: return -0.033271752297878265f;
  t200_n125: return 0.20824487507343292f;
  t200_n126: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t200_n128;
  t200_n127: return 0.10165153443813324f;
  t200_n128: return 0.159428209066391f;
  t200_n129: return -0.03562885522842407f;
  t200_n130: return 0.003711908357217908f;
  t200_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n133;
  t200_n132: return -0.3132103383541107f;
  t200_n133: if (E.getNumReferences() >= 3296780288 /*1031.0*/) goto t200_n135;
  t200_n134: return 0.05710725486278534f;
  t200_n135: return 6.827642937423661e-05f;
  t200_n136: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t200_n138;
  t200_n137: return -0.05629972368478775f;
  t200_n138: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t200_n142;
  t200_n139: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n141;
  t200_n140: return 0.2852608561515808f;
  t200_n141: return 0.08195924013853073f;
  t200_n142: return -0.22006605565547943f;
  t200_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n145;
  t200_n144: return -0.01406060066074133f;
  t200_n145: return 0.059578388929367065f;
  t200_n146: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t200_n152;
  t200_n147: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t200_n149;
  t200_n148: return -0.021166017279028893f;
  t200_n149: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t200_n151;
  t200_n150: return 0.02434641495347023f;
  t200_n151: return -0.05263493210077286f;
  t200_n152: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t200_n176;
  t200_n153: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t200_n163;
  t200_n154: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t200_n162;
  t200_n155: if (E.getNumReferences() >= 3276570624 /*204.5*/) goto t200_n161;
  t200_n156: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t200_n160;
  t200_n157: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t200_n159;
  t200_n158: return 0.08102752268314362f;
  t200_n159: return -0.029610443860292435f;
  t200_n160: return 0.01161388773471117f;
  t200_n161: return 0.14779268205165863f;
  t200_n162: return 0.035920754075050354f;
  t200_n163: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t200_n165;
  t200_n164: return 0.021948980167508125f;
  t200_n165: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t200_n169;
  t200_n166: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t200_n168;
  t200_n167: return 0.12631572782993317f;
  t200_n168: return 0.11576010286808014f;
  t200_n169: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t200_n175;
  t200_n170: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n174;
  t200_n171: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t200_n173;
  t200_n172: return 0.05415722727775574f;
  t200_n173: return 0.1336681842803955f;
  t200_n174: return 0.1267240047454834f;
  t200_n175: return -0.06268058717250824f;
  t200_n176: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t200_n178;
  t200_n177: return 0.024048442021012306f;
  t200_n178: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t200_n182;
  t200_n179: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t200_n181;
  t200_n180: return -0.04317934811115265f;
  t200_n181: return 0.08695921301841736f;
  t200_n182: return -0.14613917469978333f;
  t200_n183: return 0.009446049109101295f;
  t200_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n208;
  t200_n185: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t200_n199;
  t200_n186: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t200_n194;
  t200_n187: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t200_n189;
  t200_n188: return -0.01451924443244934f;
  t200_n189: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t200_n191;
  t200_n190: return -0.006200640462338924f;
  t200_n191: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t200_n193;
  t200_n192: return 0.1466081291437149f;
  t200_n193: return 0.04251366853713989f;
  t200_n194: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t200_n198;
  t200_n195: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t200_n197;
  t200_n196: return -0.004462672863155603f;
  t200_n197: return -0.0446007214486599f;
  t200_n198: return 0.012297092005610466f;
  t200_n199: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t200_n205;
  t200_n200: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t200_n202;
  t200_n201: return 0.00930256862193346f;
  t200_n202: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t200_n204;
  t200_n203: return 0.14038321375846863f;
  t200_n204: return 0.10086607933044434f;
  t200_n205: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t200_n207;
  t200_n206: return 0.007124442141503096f;
  t200_n207: return -0.028870288282632828f;
  t200_n208: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t200_n210;
  t200_n209: return 0.0053776102140545845f;
  t200_n210: return -0.07752399146556854f;
  t200_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t200_n213;
  t200_n212: return 0.0011546446476131678f;
  t200_n213: return 0.004394302144646645f;
  t200_n214: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t200_n216;
  t200_n215: return -0.012494313530623913f;
  t200_n216: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t200_n246;
  t200_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t200_n245;
  t200_n218: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t200_n236;
  t200_n219: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t200_n221;
  t200_n220: return -0.16646458208560944f;
  t200_n221: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t200_n223;
  t200_n222: return -0.1250087022781372f;
  t200_n223: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t200_n235;
  t200_n224: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t200_n230;
  t200_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t200_n227;
  t200_n226: return -0.07515055686235428f;
  t200_n227: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t200_n229;
  t200_n228: return 0.1592903733253479f;
  t200_n229: return -0.12241296470165253f;
  t200_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t200_n234;
  t200_n231: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t200_n233;
  t200_n232: return 0.017167646437883377f;
  t200_n233: return 0.8112882375717163f;
  t200_n234: return 0.11096876859664917f;
  t200_n235: return -0.08497529476881027f;
  t200_n236: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t200_n238;
  t200_n237: return -0.10944394022226334f;
  t200_n238: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t200_n240;
  t200_n239: return -0.13921838998794556f;
  t200_n240: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t200_n244;
  t200_n241: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t200_n243;
  t200_n242: return -0.18564774096012115f;
  t200_n243: return 0.1365806758403778f;
  t200_n244: return 0.0503632053732872f;
  t200_n245: return -0.027745869010686874f;
  t200_n246: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t200_n248;
  t200_n247: return -0.14714016020298004f;
  t200_n248: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t200_n250;
  t200_n249: return -0.14075420796871185f;
  t200_n250: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t200_n252;
  t200_n251: return 1.6656439304351807f;
  t200_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t200_n254;
  t200_n253: return 0.05830458924174309f;
  t200_n254: return 0.1628972291946411f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree201(const Example& E) {
  t201_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t201_n80;
  t201_n1: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t201_n69;
  t201_n2: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t201_n4;
  t201_n3: return -3.514939089654945e-05f;
  t201_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t201_n32;
  t201_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t201_n31;
  t201_n6: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t201_n8;
  t201_n7: return -0.037399422377347946f;
  t201_n8: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t201_n28;
  t201_n9: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t201_n11;
  t201_n10: return 0.01855573058128357f;
  t201_n11: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t201_n25;
  t201_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t201_n16;
  t201_n13: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n15;
  t201_n14: return -0.045600973069667816f;
  t201_n15: return 0.0535174235701561f;
  t201_n16: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t201_n18;
  t201_n17: return 0.1526464968919754f;
  t201_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t201_n20;
  t201_n19: return 0.07211495190858841f;
  t201_n20: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t201_n22;
  t201_n21: return 0.08553757518529892f;
  t201_n22: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t201_n24;
  t201_n23: return 0.1294133961200714f;
  t201_n24: return 0.09683210402727127f;
  t201_n25: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n27;
  t201_n26: return 0.1783190220594406f;
  t201_n27: return -0.13180071115493774f;
  t201_n28: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t201_n30;
  t201_n29: return -0.039516616612672806f;
  t201_n30: return 0.1325504183769226f;
  t201_n31: return -0.07131873816251755f;
  t201_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t201_n58;
  t201_n33: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t201_n57;
  t201_n34: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t201_n54;
  t201_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t201_n51;
  t201_n36: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t201_n40;
  t201_n37: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t201_n39;
  t201_n38: return -0.01188009325414896f;
  t201_n39: return 0.010161833837628365f;
  t201_n40: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t201_n48;
  t201_n41: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t201_n45;
  t201_n42: if (E.getNumReferences() >= 3266904064 /*92.5*/) goto t201_n44;
  t201_n43: return 0.01755615510046482f;
  t201_n44: return -0.08583423495292664f;
  t201_n45: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t201_n47;
  t201_n46: return -0.46494078636169434f;
  t201_n47: return 0.08090190589427948f;
  t201_n48: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t201_n50;
  t201_n49: return 0.11580952256917953f;
  t201_n50: return -0.11453621834516525f;
  t201_n51: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t201_n53;
  t201_n52: return 0.006287842057645321f;
  t201_n53: return 0.018327578902244568f;
  t201_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t201_n56;
  t201_n55: return -0.017558054998517036f;
  t201_n56: return 0.003889070823788643f;
  t201_n57: return 0.01131226122379303f;
  t201_n58: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t201_n68;
  t201_n59: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t201_n61;
  t201_n60: return -0.01871034875512123f;
  t201_n61: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t201_n63;
  t201_n62: return 0.035271111875772476f;
  t201_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t201_n65;
  t201_n64: return -0.0038263630121946335f;
  t201_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t201_n67;
  t201_n66: return 0.019370021298527718f;
  t201_n67: return 0.0647948607802391f;
  t201_n68: return -0.040437985211610794f;
  t201_n69: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t201_n71;
  t201_n70: return 0.006580534391105175f;
  t201_n71: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t201_n75;
  t201_n72: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t201_n74;
  t201_n73: return -0.3806741237640381f;
  t201_n74: return 0.0004267723415978253f;
  t201_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t201_n77;
  t201_n76: return -0.09880881756544113f;
  t201_n77: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t201_n79;
  t201_n78: return 0.07111292332410812f;
  t201_n79: return 0.16464751958847046f;
  t201_n80: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t201_n206;
  t201_n81: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t201_n197;
  t201_n82: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t201_n130;
  t201_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_TopLevel))) goto t201_n127;
  t201_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t201_n120;
  t201_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t201_n107;
  t201_n86: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t201_n98;
  t201_n87: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t201_n93;
  t201_n88: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n90;
  t201_n89: return -0.16693739593029022f;
  t201_n90: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t201_n92;
  t201_n91: return -0.03637691214680672f;
  t201_n92: return -0.24863113462924957f;
  t201_n93: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t201_n95;
  t201_n94: return -0.027835015207529068f;
  t201_n95: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t201_n97;
  t201_n96: return -0.15500158071517944f;
  t201_n97: return 0.10764473676681519f;
  t201_n98: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t201_n100;
  t201_n99: return 0.008115834556519985f;
  t201_n100: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t201_n106;
  t201_n101: if (E.getNumReferences() >= 3225419776 /*3.0*/) goto t201_n105;
  t201_n102: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n104;
  t201_n103: return 0.08945570886135101f;
  t201_n104: return 0.016480306163430214f;
  t201_n105: return -0.05889769643545151f;
  t201_n106: return 0.10580545663833618f;
  t201_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t201_n115;
  t201_n108: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n110;
  t201_n109: return -0.09032976627349854f;
  t201_n110: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t201_n114;
  t201_n111: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t201_n113;
  t201_n112: return 0.004353020805865526f;
  t201_n113: return 0.1739915907382965f;
  t201_n114: return -0.22048144042491913f;
  t201_n115: if (E.getNumReferences() >= 3335198464 /*25995.5*/) goto t201_n119;
  t201_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t201_n118;
  t201_n117: return 0.021525869145989418f;
  t201_n118: return -0.003998941741883755f;
  t201_n119: return 0.10760220885276794f;
  t201_n120: if (E.getNumReferences() >= 3274473472 /*172.5*/) goto t201_n122;
  t201_n121: return 0.01582501269876957f;
  t201_n122: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t201_n126;
  t201_n123: if (E.getNumReferences() >= 3288793088 /*540.0*/) goto t201_n125;
  t201_n124: return 0.1411786675453186f;
  t201_n125: return 0.00034906304790638387f;
  t201_n126: return -0.04580859839916229f;
  t201_n127: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t201_n129;
  t201_n128: return 0.018705639988183975f;
  t201_n129: return 0.05686352401971817f;
  t201_n130: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t201_n132;
  t201_n131: return -0.010867305099964142f;
  t201_n132: if (E.getNumReferences() >= 3343861504 /*53063.0*/) goto t201_n194;
  t201_n133: if (E.getNumReferences() >= 3319284736 /*6920.5*/) goto t201_n187;
  t201_n134: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t201_n168;
  t201_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t201_n163;
  t201_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t201_n144;
  t201_n137: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t201_n141;
  t201_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t201_n140;
  t201_n139: return -0.17180581390857697f;
  t201_n140: return -0.027672860771417618f;
  t201_n141: if (E.getNumReferences() >= 3286794240 /*465.0*/) goto t201_n143;
  t201_n142: return 0.014474011026322842f;
  t201_n143: return 0.12683117389678955f;
  t201_n144: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t201_n156;
  t201_n145: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t201_n151;
  t201_n146: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t201_n148;
  t201_n147: return 0.018169421702623367f;
  t201_n148: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t201_n150;
  t201_n149: return 0.09078745543956757f;
  t201_n150: return 0.0f;
  t201_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t201_n153;
  t201_n152: return -0.1380155384540558f;
  t201_n153: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t201_n155;
  t201_n154: return -0.0622495599091053f;
  t201_n155: return 0.09547124058008194f;
  t201_n156: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t201_n158;
  t201_n157: return 0.06604573130607605f;
  t201_n158: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t201_n160;
  t201_n159: return 0.02348477952182293f;
  t201_n160: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t201_n162;
  t201_n161: return -7.526709079742432f;
  t201_n162: return 0.06784225255250931f;
  t201_n163: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t201_n165;
  t201_n164: return 0.04879254847764969f;
  t201_n165: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t201_n167;
  t201_n166: return -0.2790871262550354f;
  t201_n167: return 0.05421411618590355f;
  t201_n168: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t201_n176;
  t201_n169: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t201_n175;
  t201_n170: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t201_n172;
  t201_n171: return 0.04461875185370445f;
  t201_n172: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t201_n174;
  t201_n173: return 0.14083704352378845f;
  t201_n174: return 0.08263815939426422f;
  t201_n175: return 0.0026732091791927814f;
  t201_n176: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t201_n178;
  t201_n177: return -0.033644210547208786f;
  t201_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t201_n186;
  t201_n179: if (E.getNumReferences() >= 3251109888 /*25.0*/) goto t201_n185;
  t201_n180: if (E.getFractionNameInContext() >= 3187804193 /*0.1269841343164444*/) goto t201_n182;
  t201_n181: return 0.17298929393291473f;
  t201_n182: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t201_n184;
  t201_n183: return -0.17143061757087708f;
  t201_n184: return 0.07919805496931076f;
  t201_n185: return -0.005728548392653465f;
  t201_n186: return 0.12633377313613892f;
  t201_n187: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t201_n189;
  t201_n188: return -0.14443090558052063f;
  t201_n189: if (E.getNumReferences() >= 3320151040 /*7343.5*/) goto t201_n191;
  t201_n190: return 0.0975448414683342f;
  t201_n191: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t201_n193;
  t201_n192: return 0.051849015057086945f;
  t201_n193: return -0.02284466102719307f;
  t201_n194: if (E.getNumReferences() >= 3364353728 /*278454.0*/) goto t201_n196;
  t201_n195: return 0.0375521257519722f;
  t201_n196: return 0.021024011075496674f;
  t201_n197: if (E.getFractionNameInContext() >= 3200163166 /*0.3722943663597107*/) goto t201_n199;
  t201_n198: return -0.20189595222473145f;
  t201_n199: if (E.getFractionNameInContext() >= 3203710796 /*0.4780219793319702*/) goto t201_n201;
  t201_n200: return 0.12612506747245789f;
  t201_n201: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t201_n203;
  t201_n202: return 0.008523405529558659f;
  t201_n203: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t201_n205;
  t201_n204: return 0.12125435471534729f;
  t201_n205: return -0.1288488358259201f;
  t201_n206: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t201_n254;
  t201_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t201_n217;
  t201_n208: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t201_n210;
  t201_n209: return -0.08307340741157532f;
  t201_n210: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t201_n212;
  t201_n211: return -0.03882857784628868f;
  t201_n212: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n214;
  t201_n213: return 0.1189127266407013f;
  t201_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t201_n216;
  t201_n215: return 0.024885915219783783f;
  t201_n216: return 0.08948712795972824f;
  t201_n217: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t201_n237;
  t201_n218: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t201_n220;
  t201_n219: return 0.00398485129699111f;
  t201_n220: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t201_n222;
  t201_n221: return 0.41988539695739746f;
  t201_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t201_n232;
  t201_n223: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t201_n227;
  t201_n224: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t201_n226;
  t201_n225: return 0.1191425547003746f;
  t201_n226: return 0.0002647716028150171f;
  t201_n227: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t201_n231;
  t201_n228: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t201_n230;
  t201_n229: return -0.316211074590683f;
  t201_n230: return 0.09187445789575577f;
  t201_n231: return -0.2978511452674866f;
  t201_n232: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t201_n234;
  t201_n233: return 0.07245475053787231f;
  t201_n234: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t201_n236;
  t201_n235: return 0.18971829116344452f;
  t201_n236: return 0.18304668366909027f;
  t201_n237: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t201_n239;
  t201_n238: return -0.07686062902212143f;
  t201_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t201_n247;
  t201_n240: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t201_n246;
  t201_n241: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t201_n243;
  t201_n242: return -0.11586343497037888f;
  t201_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t201_n245;
  t201_n244: return 0.20260323584079742f;
  t201_n245: return -0.07510426640510559f;
  t201_n246: return 0.22152991592884064f;
  t201_n247: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t201_n253;
  t201_n248: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t201_n252;
  t201_n249: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t201_n251;
  t201_n250: return 0.2224234789609909f;
  t201_n251: return 1.9181666374206543f;
  t201_n252: return -0.1738778054714203f;
  t201_n253: return 0.2845557928085327f;
  t201_n254: return 0.08226734399795532f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree202(const Example& E) {
  t202_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t202_n2;
  t202_n1: return -0.0004503066884353757f;
  t202_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t202_n254;
  t202_n3: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t202_n145;
  t202_n4: if (E.getFractionNameInContext() >= 3203243738 /*0.464102566242218*/) goto t202_n130;
  t202_n5: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t202_n35;
  t202_n6: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t202_n16;
  t202_n7: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t202_n15;
  t202_n8: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t202_n10;
  t202_n9: return -0.009027051739394665f;
  t202_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t202_n12;
  t202_n11: return 0.02643747255206108f;
  t202_n12: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t202_n14;
  t202_n13: return 0.15214383602142334f;
  t202_n14: return 0.05003092437982559f;
  t202_n15: return -0.0006612091674469411f;
  t202_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t202_n34;
  t202_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t202_n23;
  t202_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t202_n22;
  t202_n19: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t202_n21;
  t202_n20: return -0.012058894149959087f;
  t202_n21: return -0.01889335736632347f;
  t202_n22: return -0.0009996771113947034f;
  t202_n23: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t202_n31;
  t202_n24: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t202_n26;
  t202_n25: return -0.04289716109633446f;
  t202_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t202_n28;
  t202_n27: return 0.08160076290369034f;
  t202_n28: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t202_n30;
  t202_n29: return 0.09451586753129959f;
  t202_n30: return 0.03870204836130142f;
  t202_n31: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t202_n33;
  t202_n32: return -0.40602612495422363f;
  t202_n33: return 0.004646541550755501f;
  t202_n34: return 0.12440714240074158f;
  t202_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t202_n45;
  t202_n36: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t202_n44;
  t202_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t202_n41;
  t202_n38: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t202_n40;
  t202_n39: return -0.0954047217965126f;
  t202_n40: return 0.06038466840982437f;
  t202_n41: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t202_n43;
  t202_n42: return -0.31672027707099915f;
  t202_n43: return 0.21895720064640045f;
  t202_n44: return -0.009074127301573753f;
  t202_n45: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t202_n85;
  t202_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t202_n50;
  t202_n47: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t202_n49;
  t202_n48: return -0.005604174919426441f;
  t202_n49: return -0.19730572402477264f;
  t202_n50: if (E.getFractionNameInContext() >= 3196560022 /*0.26491230726242065*/) goto t202_n52;
  t202_n51: return -0.0005773149314336479f;
  t202_n52: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t202_n64;
  t202_n53: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t202_n63;
  t202_n54: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t202_n60;
  t202_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t202_n59;
  t202_n56: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t202_n58;
  t202_n57: return -0.28658631443977356f;
  t202_n58: return -0.03306048735976219f;
  t202_n59: return 0.019786635413765907f;
  t202_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t202_n62;
  t202_n61: return 0.011286339722573757f;
  t202_n62: return 0.11696869879961014f;
  t202_n63: return 0.0862356573343277f;
  t202_n64: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t202_n70;
  t202_n65: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t202_n69;
  t202_n66: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t202_n68;
  t202_n67: return -0.015928832814097404f;
  t202_n68: return 0.08843564987182617f;
  t202_n69: return -0.368541419506073f;
  t202_n70: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t202_n72;
  t202_n71: return 0.11899885535240173f;
  t202_n72: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t202_n82;
  t202_n73: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t202_n81;
  t202_n74: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t202_n80;
  t202_n75: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t202_n79;
  t202_n76: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t202_n78;
  t202_n77: return 0.16393856704235077f;
  t202_n78: return 0.15897442400455475f;
  t202_n79: return 0.09320497512817383f;
  t202_n80: return -0.03556406870484352f;
  t202_n81: return -0.18013787269592285f;
  t202_n82: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t202_n84;
  t202_n83: return 1.2139091491699219f;
  t202_n84: return 0.06307024508714676f;
  t202_n85: if (E.getFractionNameInContext() >= 3197399006 /*0.2899159789085388*/) goto t202_n91;
  t202_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t202_n88;
  t202_n87: return -0.016324257478117943f;
  t202_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t202_n90;
  t202_n89: return 0.07294964790344238f;
  t202_n90: return 0.10639622807502747f;
  t202_n91: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t202_n93;
  t202_n92: return 0.12191174179315567f;
  t202_n93: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t202_n103;
  t202_n94: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t202_n98;
  t202_n95: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t202_n97;
  t202_n96: return 0.014177682809531689f;
  t202_n97: return 0.09833134710788727f;
  t202_n98: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t202_n100;
  t202_n99: return -0.06110970303416252f;
  t202_n100: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t202_n102;
  t202_n101: return -0.3475892245769501f;
  t202_n102: return 0.08483901619911194f;
  t202_n103: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t202_n113;
  t202_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t202_n106;
  t202_n105: return -0.07084082812070847f;
  t202_n106: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t202_n112;
  t202_n107: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t202_n109;
  t202_n108: return 0.07551127672195435f;
  t202_n109: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t202_n111;
  t202_n110: return 0.21335381269454956f;
  t202_n111: return 0.10915037244558334f;
  t202_n112: return -0.12990206480026245f;
  t202_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t202_n121;
  t202_n114: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t202_n116;
  t202_n115: return -0.015079222619533539f;
  t202_n116: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t202_n118;
  t202_n117: return -0.030741654336452484f;
  t202_n118: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t202_n120;
  t202_n119: return -0.12959636747837067f;
  t202_n120: return -0.9742901921272278f;
  t202_n121: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t202_n125;
  t202_n122: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t202_n124;
  t202_n123: return 0.011165488511323929f;
  t202_n124: return 0.0962359756231308f;
  t202_n125: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t202_n127;
  t202_n126: return -0.13879233598709106f;
  t202_n127: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t202_n129;
  t202_n128: return -0.15321199595928192f;
  t202_n129: return 0.12324308604001999f;
  t202_n130: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t202_n142;
  t202_n131: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t202_n141;
  t202_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t202_n134;
  t202_n133: return -0.03788170963525772f;
  t202_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t202_n136;
  t202_n135: return 0.013042803853750229f;
  t202_n136: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t202_n140;
  t202_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t202_n139;
  t202_n138: return 0.032254960387945175f;
  t202_n139: return 0.10097155719995499f;
  t202_n140: return 0.0331585556268692f;
  t202_n141: return -0.01016057189553976f;
  t202_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t202_n144;
  t202_n143: return 0.00013957603368908167f;
  t202_n144: return 0.027059823274612427f;
  t202_n145: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t202_n253;
  t202_n146: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t202_n252;
  t202_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t202_n187;
  t202_n148: if (E.getNumReferences() >= 3314693120 /*4678.5*/) goto t202_n176;
  t202_n149: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t202_n155;
  t202_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t202_n154;
  t202_n151: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t202_n153;
  t202_n152: return -0.05982527881860733f;
  t202_n153: return 0.12431174516677856f;
  t202_n154: return 0.00522504560649395f;
  t202_n155: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t202_n175;
  t202_n156: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t202_n174;
  t202_n157: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t202_n167;
  t202_n158: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t202_n160;
  t202_n159: return -0.0013095642207190394f;
  t202_n160: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t202_n166;
  t202_n161: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t202_n163;
  t202_n162: return -0.060602568089962006f;
  t202_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t202_n165;
  t202_n164: return -0.4261806905269623f;
  t202_n165: return 0.058216799050569534f;
  t202_n166: return -0.061688102781772614f;
  t202_n167: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t202_n173;
  t202_n168: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t202_n172;
  t202_n169: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t202_n171;
  t202_n170: return 0.13634788990020752f;
  t202_n171: return 0.08405160903930664f;
  t202_n172: return 0.10678664594888687f;
  t202_n173: return 0.009662303142249584f;
  t202_n174: return 0.011700581759214401f;
  t202_n175: return -0.025330036878585815f;
  t202_n176: if (E.getNumReferences() >= 3315159040 /*4906.0*/) goto t202_n186;
  t202_n177: if (E.getNumReferences() >= 3315124224 /*4889.0*/) goto t202_n185;
  t202_n178: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t202_n180;
  t202_n179: return -0.07561205327510834f;
  t202_n180: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t202_n182;
  t202_n181: return 0.016907520592212677f;
  t202_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t202_n184;
  t202_n183: return 0.04612603783607483f;
  t202_n184: return 0.12122182548046112f;
  t202_n185: return 0.1373916119337082f;
  t202_n186: return 0.01661170832812786f;
  t202_n187: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t202_n221;
  t202_n188: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t202_n218;
  t202_n189: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t202_n191;
  t202_n190: return -0.03724037483334541f;
  t202_n191: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t202_n209;
  t202_n192: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t202_n198;
  t202_n193: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t202_n197;
  t202_n194: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t202_n196;
  t202_n195: return -0.10013549029827118f;
  t202_n196: return 0.9207162857055664f;
  t202_n197: return 0.034455589950084686f;
  t202_n198: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t202_n204;
  t202_n199: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t202_n201;
  t202_n200: return -0.09129447489976883f;
  t202_n201: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t202_n203;
  t202_n202: return 0.09373608976602554f;
  t202_n203: return -0.0725422203540802f;
  t202_n204: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t202_n208;
  t202_n205: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t202_n207;
  t202_n206: return 0.09281592816114426f;
  t202_n207: return 0.1303878128528595f;
  t202_n208: return 0.06545967608690262f;
  t202_n209: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t202_n217;
  t202_n210: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t202_n216;
  t202_n211: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t202_n213;
  t202_n212: return -0.16011196374893188f;
  t202_n213: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t202_n215;
  t202_n214: return -0.10020695626735687f;
  t202_n215: return 0.1882621794939041f;
  t202_n216: return -0.024703316390514374f;
  t202_n217: return -0.10805249959230423f;
  t202_n218: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t202_n220;
  t202_n219: return -0.3462420403957367f;
  t202_n220: return -0.008050516247749329f;
  t202_n221: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t202_n223;
  t202_n222: return 0.0327092707157135f;
  t202_n223: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t202_n251;
  t202_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t202_n250;
  t202_n225: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t202_n227;
  t202_n226: return 0.13083255290985107f;
  t202_n227: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t202_n229;
  t202_n228: return 0.030444545671343803f;
  t202_n229: if (E.getNumReferences() >= 3254255616 /*31.0*/) goto t202_n237;
  t202_n230: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t202_n234;
  t202_n231: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t202_n233;
  t202_n232: return -0.2494281530380249f;
  t202_n233: return 0.14083489775657654f;
  t202_n234: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t202_n236;
  t202_n235: return 0.09033436328172684f;
  t202_n236: return -0.09914827346801758f;
  t202_n237: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t202_n239;
  t202_n238: return 0.12516309320926666f;
  t202_n239: if (E.getNumReferences() >= 3265200128 /*79.5*/) goto t202_n243;
  t202_n240: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t202_n242;
  t202_n241: return 0.10314234346151352f;
  t202_n242: return 0.00809276383370161f;
  t202_n243: if (E.getNumReferences() >= 3268935680 /*108.0*/) goto t202_n245;
  t202_n244: return 0.12728846073150635f;
  t202_n245: if (E.getNumReferences() >= 3277946880 /*225.5*/) goto t202_n247;
  t202_n246: return -0.16373887658119202f;
  t202_n247: if (E.getNumReferences() >= 3283599360 /*367.5*/) goto t202_n249;
  t202_n248: return 0.13296179473400116f;
  t202_n249: return -0.015241309069097042f;
  t202_n250: return 0.1178647130727768f;
  t202_n251: return 0.02718961611390114f;
  t202_n252: return -0.0013000506442040205f;
  t202_n253: return 0.08183973282575607f;
  t202_n254: return 0.0019622324034571648f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree203(const Example& E) {
  t203_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t203_n252;
  t203_n1: if (E.getNumReferences() >= 3265855488 /*84.5*/) goto t203_n205;
  t203_n2: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t203_n6;
  t203_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t203_n5;
  t203_n4: return -0.002118696691468358f;
  t203_n5: return 0.0002878682571463287f;
  t203_n6: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t203_n50;
  t203_n7: if (E.getNumReferences() >= 3264675840 /*75.5*/) goto t203_n49;
  t203_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t203_n38;
  t203_n9: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t203_n29;
  t203_n10: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t203_n28;
  t203_n11: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t203_n27;
  t203_n12: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t203_n20;
  t203_n13: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t203_n15;
  t203_n14: return -0.0017763368086889386f;
  t203_n15: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t203_n17;
  t203_n16: return -0.10080849379301071f;
  t203_n17: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t203_n19;
  t203_n18: return 0.1143479198217392f;
  t203_n19: return -0.05325932055711746f;
  t203_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t203_n26;
  t203_n21: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t203_n23;
  t203_n22: return 0.12916474044322968f;
  t203_n23: if (E.getFractionNameInContext() >= 3179644278 /*0.0651959627866745*/) goto t203_n25;
  t203_n24: return -0.163718119263649f;
  t203_n25: return 0.135684996843338f;
  t203_n26: return 0.09911011159420013f;
  t203_n27: return -0.08279967308044434f;
  t203_n28: return 0.0003773478383664042f;
  t203_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t203_n31;
  t203_n30: return -0.07317391782999039f;
  t203_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t203_n33;
  t203_n32: return -0.009845281019806862f;
  t203_n33: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t203_n35;
  t203_n34: return 0.04184844717383385f;
  t203_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t203_n37;
  t203_n36: return 0.06813764572143555f;
  t203_n37: return 0.32145923376083374f;
  t203_n38: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t203_n46;
  t203_n39: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t203_n41;
  t203_n40: return 0.02117915078997612f;
  t203_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t203_n45;
  t203_n42: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t203_n44;
  t203_n43: return -0.560924232006073f;
  t203_n44: return -0.10320945084095001f;
  t203_n45: return -0.02085546776652336f;
  t203_n46: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t203_n48;
  t203_n47: return -0.04702761024236679f;
  t203_n48: return 0.005094265099614859f;
  t203_n49: return -0.09330887347459793f;
  t203_n50: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t203_n56;
  t203_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t203_n55;
  t203_n52: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t203_n54;
  t203_n53: return -0.04738247022032738f;
  t203_n54: return -0.04761342331767082f;
  t203_n55: return -0.03063422441482544f;
  t203_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t203_n168;
  t203_n57: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t203_n159;
  t203_n58: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t203_n132;
  t203_n59: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t203_n131;
  t203_n60: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t203_n94;
  t203_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t203_n67;
  t203_n62: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t203_n64;
  t203_n63: return -0.04624509438872337f;
  t203_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t203_n66;
  t203_n65: return -0.5732429027557373f;
  t203_n66: return -0.07325921207666397f;
  t203_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t203_n93;
  t203_n68: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t203_n92;
  t203_n69: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t203_n91;
  t203_n70: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t203_n72;
  t203_n71: return -0.05236360803246498f;
  t203_n72: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t203_n76;
  t203_n73: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t203_n75;
  t203_n74: return 0.12757772207260132f;
  t203_n75: return -0.16356733441352844f;
  t203_n76: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t203_n82;
  t203_n77: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t203_n79;
  t203_n78: return -0.04753834754228592f;
  t203_n79: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t203_n81;
  t203_n80: return 0.07566382735967636f;
  t203_n81: return -0.025037109851837158f;
  t203_n82: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t203_n90;
  t203_n83: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t203_n87;
  t203_n84: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t203_n86;
  t203_n85: return 0.10694526880979538f;
  t203_n86: return 0.10802284628152847f;
  t203_n87: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t203_n89;
  t203_n88: return 0.07812429219484329f;
  t203_n89: return -0.41770097613334656f;
  t203_n90: return -0.03842854127287865f;
  t203_n91: return 0.06830970197916031f;
  t203_n92: return -0.025712210685014725f;
  t203_n93: return -0.001377115142531693f;
  t203_n94: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t203_n110;
  t203_n95: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t203_n107;
  t203_n96: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t203_n98;
  t203_n97: return 0.008000737987458706f;
  t203_n98: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t203_n106;
  t203_n99: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t203_n101;
  t203_n100: return 0.0640990287065506f;
  t203_n101: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t203_n103;
  t203_n102: return 0.19009891152381897f;
  t203_n103: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t203_n105;
  t203_n104: return 0.16276989877223969f;
  t203_n105: return 0.01881643757224083f;
  t203_n106: return 0.023117467761039734f;
  t203_n107: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t203_n109;
  t203_n108: return 0.04576680436730385f;
  t203_n109: return 0.046745460480451584f;
  t203_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t203_n126;
  t203_n111: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t203_n115;
  t203_n112: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t203_n114;
  t203_n113: return -0.33380764722824097f;
  t203_n114: return -0.043763283640146255f;
  t203_n115: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t203_n117;
  t203_n116: return -0.0039512524381279945f;
  t203_n117: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t203_n119;
  t203_n118: return -0.16178536415100098f;
  t203_n119: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t203_n121;
  t203_n120: return -4.203240871429443f;
  t203_n121: if (E.getNumReferences() >= 3251634176 /*26.0*/) goto t203_n123;
  t203_n122: return -0.09494395554065704f;
  t203_n123: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t203_n125;
  t203_n124: return -1.647055745124817f;
  t203_n125: return -0.2592405080795288f;
  t203_n126: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t203_n128;
  t203_n127: return 0.002524977782741189f;
  t203_n128: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t203_n130;
  t203_n129: return -0.21624267101287842f;
  t203_n130: return 0.01053405087441206f;
  t203_n131: return -0.035299524664878845f;
  t203_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t203_n134;
  t203_n133: return -0.027129309251904488f;
  t203_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t203_n156;
  t203_n135: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t203_n137;
  t203_n136: return -0.04087620973587036f;
  t203_n137: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t203_n149;
  t203_n138: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t203_n142;
  t203_n139: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t203_n141;
  t203_n140: return 0.15546204149723053f;
  t203_n141: return -0.11393596976995468f;
  t203_n142: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t203_n144;
  t203_n143: return 0.052141908556222916f;
  t203_n144: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t203_n148;
  t203_n145: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t203_n147;
  t203_n146: return 0.21617302298545837f;
  t203_n147: return 0.1434461623430252f;
  t203_n148: return -0.1060100793838501f;
  t203_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t203_n151;
  t203_n150: return 0.03151136264204979f;
  t203_n151: if (E.getNumReferences() >= 3262775296 /*62.5*/) goto t203_n155;
  t203_n152: if (E.getNumReferences() >= 3257925632 /*44.0*/) goto t203_n154;
  t203_n153: return 0.11077111959457397f;
  t203_n154: return 0.15857046842575073f;
  t203_n155: return 0.036774322390556335f;
  t203_n156: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t203_n158;
  t203_n157: return 0.12329006940126419f;
  t203_n158: return -0.1011156290769577f;
  t203_n159: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t203_n167;
  t203_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t203_n164;
  t203_n161: if (E.getNumReferences() >= 3264544768 /*74.5*/) goto t203_n163;
  t203_n162: return 0.03532861918210983f;
  t203_n163: return 0.16994202136993408f;
  t203_n164: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t203_n166;
  t203_n165: return 0.13898922502994537f;
  t203_n166: return -0.11924096941947937f;
  t203_n167: return 0.11620106548070908f;
  t203_n168: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t203_n204;
  t203_n169: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t203_n183;
  t203_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t203_n172;
  t203_n171: return 0.004619233775883913f;
  t203_n172: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t203_n182;
  t203_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t203_n175;
  t203_n174: return 0.008328322321176529f;
  t203_n175: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t203_n181;
  t203_n176: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t203_n178;
  t203_n177: return 0.017156032845377922f;
  t203_n178: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t203_n180;
  t203_n179: return 0.07372309267520905f;
  t203_n180: return 0.032625868916511536f;
  t203_n181: return 0.0978415384888649f;
  t203_n182: return 0.0035778486635535955f;
  t203_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t203_n201;
  t203_n184: if (E.getNumReferences() >= 3264675840 /*75.5*/) goto t203_n198;
  t203_n185: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t203_n191;
  t203_n186: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t203_n190;
  t203_n187: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t203_n189;
  t203_n188: return -0.22670748829841614f;
  t203_n189: return 0.29392507672309875f;
  t203_n190: return -0.5139454007148743f;
  t203_n191: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t203_n195;
  t203_n192: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t203_n194;
  t203_n193: return -0.11450950056314468f;
  t203_n194: return 0.12003965675830841f;
  t203_n195: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t203_n197;
  t203_n196: return -0.2684653103351593f;
  t203_n197: return -0.06200921908020973f;
  t203_n198: if (E.getNumReferences() >= 3265396736 /*81.0*/) goto t203_n200;
  t203_n199: return 0.11104032397270203f;
  t203_n200: return -0.24949190020561218f;
  t203_n201: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t203_n203;
  t203_n202: return 0.05875037610530853f;
  t203_n203: return -0.01417490653693676f;
  t203_n204: return -0.01254106406122446f;
  t203_n205: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t203_n219;
  t203_n206: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t203_n218;
  t203_n207: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t203_n213;
  t203_n208: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t203_n212;
  t203_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t203_n211;
  t203_n210: return 0.05974334478378296f;
  t203_n211: return 0.15806150436401367f;
  t203_n212: return 0.016094675287604332f;
  t203_n213: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t203_n217;
  t203_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t203_n216;
  t203_n215: return -0.02187894843518734f;
  t203_n216: return 0.03629221394658089f;
  t203_n217: return 0.007710252422839403f;
  t203_n218: return -0.14347989857196808f;
  t203_n219: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t203_n247;
  t203_n220: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t203_n246;
  t203_n221: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t203_n225;
  t203_n222: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t203_n224;
  t203_n223: return 0.13290774822235107f;
  t203_n224: return -0.014908390119671822f;
  t203_n225: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t203_n235;
  t203_n226: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t203_n234;
  t203_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t203_n229;
  t203_n228: return -0.17810429632663727f;
  t203_n229: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t203_n233;
  t203_n230: if (E.getNumReferences() >= 3272638464 /*144.5*/) goto t203_n232;
  t203_n231: return 0.1431225836277008f;
  t203_n232: return 0.08235512673854828f;
  t203_n233: return -0.08266789466142654f;
  t203_n234: return 0.06212332472205162f;
  t203_n235: if (E.getNumReferences() >= 3267624960 /*98.0*/) goto t203_n243;
  t203_n236: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t203_n242;
  t203_n237: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t203_n239;
  t203_n238: return -0.04584111273288727f;
  t203_n239: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t203_n241;
  t203_n240: return 0.13835372030735016f;
  t203_n241: return -0.024531971663236618f;
  t203_n242: return 0.10452169924974442f;
  t203_n243: if (E.getNumReferences() >= 3326988288 /*13172.0*/) goto t203_n245;
  t203_n244: return -0.02945902571082115f;
  t203_n245: return 0.10961351543664932f;
  t203_n246: return -0.0002196401183027774f;
  t203_n247: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t203_n249;
  t203_n248: return -0.01659565605223179f;
  t203_n249: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t203_n251;
  t203_n250: return 0.11473853141069412f;
  t203_n251: return 0.18805551528930664f;
  t203_n252: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t203_n254;
  t203_n253: return -0.026169227436184883f;
  t203_n254: return 0.0015698514180257916f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree204(const Example& E) {
  t204_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t204_n2;
  t204_n1: return -0.0005431039608083665f;
  t204_n2: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t204_n254;
  t204_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n209;
  t204_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t204_n6;
  t204_n5: return -0.0002672868431545794f;
  t204_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t204_n206;
  t204_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t204_n205;
  t204_n8: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t204_n204;
  t204_n9: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t204_n181;
  t204_n10: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t204_n164;
  t204_n11: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t204_n159;
  t204_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n22;
  t204_n13: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t204_n15;
  t204_n14: return -0.002386899432167411f;
  t204_n15: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t204_n21;
  t204_n16: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t204_n18;
  t204_n17: return -0.07939958572387695f;
  t204_n18: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t204_n20;
  t204_n19: return 0.0939270481467247f;
  t204_n20: return -0.038662102073431015f;
  t204_n21: return -0.11312220990657806f;
  t204_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t204_n24;
  t204_n23: return 0.0033262467477470636f;
  t204_n24: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t204_n122;
  t204_n25: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t204_n35;
  t204_n26: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t204_n28;
  t204_n27: return -0.02411036193370819f;
  t204_n28: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t204_n34;
  t204_n29: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t204_n31;
  t204_n30: return -0.028346432372927666f;
  t204_n31: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t204_n33;
  t204_n32: return 0.03350884094834328f;
  t204_n33: return 0.12494479864835739f;
  t204_n34: return -0.0057175192050635815f;
  t204_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t204_n43;
  t204_n36: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t204_n38;
  t204_n37: return 0.021673614159226418f;
  t204_n38: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t204_n42;
  t204_n39: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t204_n41;
  t204_n40: return 0.09269392490386963f;
  t204_n41: return -0.08037032932043076f;
  t204_n42: return 0.04637090861797333f;
  t204_n43: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t204_n115;
  t204_n44: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t204_n54;
  t204_n45: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t204_n53;
  t204_n46: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t204_n52;
  t204_n47: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t204_n49;
  t204_n48: return -0.01822209171950817f;
  t204_n49: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t204_n51;
  t204_n50: return 0.05715037137269974f;
  t204_n51: return 0.19238796830177307f;
  t204_n52: return -0.2095276117324829f;
  t204_n53: return -0.23563332855701447f;
  t204_n54: if (E.getNumReferences() >= 3282649088 /*338.5*/) goto t204_n112;
  t204_n55: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t204_n61;
  t204_n56: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t204_n58;
  t204_n57: return 0.003606738056987524f;
  t204_n58: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t204_n60;
  t204_n59: return -0.03961320221424103f;
  t204_n60: return -0.5137034058570862f;
  t204_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t204_n111;
  t204_n62: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t204_n74;
  t204_n63: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t204_n71;
  t204_n64: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t204_n70;
  t204_n65: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t204_n67;
  t204_n66: return 0.016778966411948204f;
  t204_n67: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t204_n69;
  t204_n68: return 0.11551302671432495f;
  t204_n69: return -0.011940712109208107f;
  t204_n70: return 0.5226296782493591f;
  t204_n71: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t204_n73;
  t204_n72: return 0.13130705058574677f;
  t204_n73: return 0.05152406916022301f;
  t204_n74: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t204_n92;
  t204_n75: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t204_n87;
  t204_n76: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t204_n80;
  t204_n77: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t204_n79;
  t204_n78: return -0.14048103988170624f;
  t204_n79: return 0.13508643209934235f;
  t204_n80: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t204_n86;
  t204_n81: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t204_n83;
  t204_n82: return 0.20018112659454346f;
  t204_n83: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t204_n85;
  t204_n84: return 0.02736043557524681f;
  t204_n85: return -0.22771266102790833f;
  t204_n86: return -0.18726889789104462f;
  t204_n87: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t204_n91;
  t204_n88: if (E.getNumReferences() >= 3258712064 /*47.0*/) goto t204_n90;
  t204_n89: return -0.28456056118011475f;
  t204_n90: return 0.11649978905916214f;
  t204_n91: return 0.11832991987466812f;
  t204_n92: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t204_n110;
  t204_n93: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t204_n103;
  t204_n94: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t204_n102;
  t204_n95: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t204_n97;
  t204_n96: return 0.06455555558204651f;
  t204_n97: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t204_n101;
  t204_n98: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t204_n100;
  t204_n99: return 0.10491137206554413f;
  t204_n100: return -0.1082359328866005f;
  t204_n101: return 0.06432846933603287f;
  t204_n102: return 0.011628526262938976f;
  t204_n103: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t204_n109;
  t204_n104: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t204_n108;
  t204_n105: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t204_n107;
  t204_n106: return -0.19655434787273407f;
  t204_n107: return -0.027904583141207695f;
  t204_n108: return 0.08714047074317932f;
  t204_n109: return 0.019315699115395546f;
  t204_n110: return -0.11441827565431595f;
  t204_n111: return 0.09835702180862427f;
  t204_n112: if (E.getNumReferences() >= 3284844544 /*405.5*/) goto t204_n114;
  t204_n113: return 0.12214314937591553f;
  t204_n114: return 0.10332449525594711f;
  t204_n115: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t204_n117;
  t204_n116: return 0.021408826112747192f;
  t204_n117: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t204_n121;
  t204_n118: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t204_n120;
  t204_n119: return 0.10506048798561096f;
  t204_n120: return -0.09777303040027618f;
  t204_n121: return -0.04412982612848282f;
  t204_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n126;
  t204_n123: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t204_n125;
  t204_n124: return -0.07324158400297165f;
  t204_n125: return -1.538496494293213f;
  t204_n126: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t204_n156;
  t204_n127: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t204_n153;
  t204_n128: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t204_n134;
  t204_n129: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t204_n131;
  t204_n130: return -0.015628686174750328f;
  t204_n131: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t204_n133;
  t204_n132: return 0.15448905527591705f;
  t204_n133: return 0.040273964405059814f;
  t204_n134: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t204_n146;
  t204_n135: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t204_n141;
  t204_n136: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t204_n138;
  t204_n137: return -0.004219743888825178f;
  t204_n138: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t204_n140;
  t204_n139: return 0.06917709857225418f;
  t204_n140: return 0.10098337382078171f;
  t204_n141: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t204_n145;
  t204_n142: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t204_n144;
  t204_n143: return 0.1520043909549713f;
  t204_n144: return -1.231143593788147f;
  t204_n145: return 0.06909868866205215f;
  t204_n146: if (E.getNumReferences() >= 3265789952 /*84.0*/) goto t204_n150;
  t204_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t204_n149;
  t204_n148: return 0.015189955942332745f;
  t204_n149: return 0.1932443380355835f;
  t204_n150: if (E.getNumReferences() >= 3274637312 /*175.0*/) goto t204_n152;
  t204_n151: return 0.13511787354946136f;
  t204_n152: return -0.027896951884031296f;
  t204_n153: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t204_n155;
  t204_n154: return -0.5537791848182678f;
  t204_n155: return -0.05958883836865425f;
  t204_n156: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t204_n158;
  t204_n157: return 0.11433979868888855f;
  t204_n158: return -0.03002716414630413f;
  t204_n159: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t204_n163;
  t204_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n162;
  t204_n161: return -0.12230260670185089f;
  t204_n162: return -0.029829930514097214f;
  t204_n163: return 0.013785628601908684f;
  t204_n164: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t204_n166;
  t204_n165: return -0.15441212058067322f;
  t204_n166: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t204_n180;
  t204_n167: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t204_n169;
  t204_n168: return -0.05516073852777481f;
  t204_n169: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t204_n179;
  t204_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n176;
  t204_n171: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t204_n175;
  t204_n172: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t204_n174;
  t204_n173: return 0.12423679232597351f;
  t204_n174: return -0.4887821674346924f;
  t204_n175: return -0.6505228281021118f;
  t204_n176: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t204_n178;
  t204_n177: return -0.01637306995689869f;
  t204_n178: return 0.056660089641809464f;
  t204_n179: return 0.03399323299527168f;
  t204_n180: return 0.024696946144104004f;
  t204_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t204_n185;
  t204_n182: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t204_n184;
  t204_n183: return -0.07403559982776642f;
  t204_n184: return 0.037481747567653656f;
  t204_n185: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t204_n189;
  t204_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t204_n188;
  t204_n187: return 0.07721609622240067f;
  t204_n188: return -0.3174856901168823f;
  t204_n189: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t204_n203;
  t204_n190: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t204_n194;
  t204_n191: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t204_n193;
  t204_n192: return 0.026388294994831085f;
  t204_n193: return 0.14669975638389587f;
  t204_n194: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t204_n198;
  t204_n195: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t204_n197;
  t204_n196: return 0.12639740109443665f;
  t204_n197: return 0.08913086354732513f;
  t204_n198: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t204_n202;
  t204_n199: if (E.getFractionNameInContext() >= 3194861276 /*0.2321428656578064*/) goto t204_n201;
  t204_n200: return -0.48633357882499695f;
  t204_n201: return 0.12731197476387024f;
  t204_n202: return 0.005869329907000065f;
  t204_n203: return -0.22746212780475616f;
  t204_n204: return 0.0003716560313478112f;
  t204_n205: return 0.010492376983165741f;
  t204_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t204_n208;
  t204_n207: return 0.0008542333380319178f;
  t204_n208: return 0.0067710913717746735f;
  t204_n209: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t204_n211;
  t204_n210: return 0.0071859355084598064f;
  t204_n211: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t204_n235;
  t204_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t204_n234;
  t204_n213: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t204_n219;
  t204_n214: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t204_n216;
  t204_n215: return -0.0033449504990130663f;
  t204_n216: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t204_n218;
  t204_n217: return -0.14208245277404785f;
  t204_n218: return -0.007352417800575495f;
  t204_n219: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t204_n233;
  t204_n220: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t204_n232;
  t204_n221: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t204_n227;
  t204_n222: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t204_n224;
  t204_n223: return 0.01473263930529356f;
  t204_n224: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t204_n226;
  t204_n225: return -0.04561280831694603f;
  t204_n226: return -0.32694023847579956f;
  t204_n227: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t204_n231;
  t204_n228: if (E.getNumReferences() >= 3231711232 /*5.0*/) goto t204_n230;
  t204_n229: return 0.09308794140815735f;
  t204_n230: return 0.10804522037506104f;
  t204_n231: return -0.007008852902799845f;
  t204_n232: return -0.2590634822845459f;
  t204_n233: return 0.04222717881202698f;
  t204_n234: return 0.022573197260499f;
  t204_n235: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t204_n253;
  t204_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t204_n242;
  t204_n237: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t204_n239;
  t204_n238: return -0.12296272069215775f;
  t204_n239: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t204_n241;
  t204_n240: return -0.039642687886953354f;
  t204_n241: return 0.10438854247331619f;
  t204_n242: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t204_n244;
  t204_n243: return -0.06760859489440918f;
  t204_n244: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t204_n252;
  t204_n245: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t204_n247;
  t204_n246: return 0.08023257553577423f;
  t204_n247: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t204_n249;
  t204_n248: return -0.19908030331134796f;
  t204_n249: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t204_n251;
  t204_n250: return 0.06907465308904648f;
  t204_n251: return -0.054733019322156906f;
  t204_n252: return 0.07419348508119583f;
  t204_n253: return -0.005718646105378866f;
  t204_n254: return 0.0093486737459898f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree205(const Example& E) {
  t205_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t205_n2;
  t205_n1: return -0.0004315527621656656f;
  t205_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t205_n254;
  t205_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t205_n85;
  t205_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t205_n84;
  t205_n5: if (E.getNumReferences() >= 3309334528 /*3079.0*/) goto t205_n83;
  t205_n6: if (E.getNumReferences() >= 3308642304 /*2910.0*/) goto t205_n80;
  t205_n7: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t205_n79;
  t205_n8: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t205_n72;
  t205_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t205_n63;
  t205_n10: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t205_n62;
  t205_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t205_n49;
  t205_n12: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t205_n30;
  t205_n13: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t205_n25;
  t205_n14: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t205_n16;
  t205_n15: return 0.005684926640242338f;
  t205_n16: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t205_n24;
  t205_n17: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t205_n21;
  t205_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t205_n20;
  t205_n19: return 0.0470300056040287f;
  t205_n20: return 0.11378508806228638f;
  t205_n21: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t205_n23;
  t205_n22: return 0.017074966803193092f;
  t205_n23: return 0.06091887876391411f;
  t205_n24: return -0.04247039183974266f;
  t205_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t205_n29;
  t205_n26: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t205_n28;
  t205_n27: return 0.13189734518527985f;
  t205_n28: return -0.08539123088121414f;
  t205_n29: return 0.0012439857237040997f;
  t205_n30: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t205_n46;
  t205_n31: if (E.getNumReferences() >= 3279912960 /*255.5*/) goto t205_n43;
  t205_n32: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t205_n42;
  t205_n33: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t205_n41;
  t205_n34: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t205_n36;
  t205_n35: return 0.00030028310720808804f;
  t205_n36: if (E.getFractionNameInContext() >= 3208762398 /*0.7571429014205933*/) goto t205_n38;
  t205_n37: return 0.045102231204509735f;
  t205_n38: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t205_n40;
  t205_n39: return -0.032597772777080536f;
  t205_n40: return -0.5032300353050232f;
  t205_n41: return -0.1384194791316986f;
  t205_n42: return -0.2464544028043747f;
  t205_n43: if (E.getNumReferences() >= 3281649664 /*308.0*/) goto t205_n45;
  t205_n44: return 0.11992069333791733f;
  t205_n45: return 0.04750756174325943f;
  t205_n46: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t205_n48;
  t205_n47: return 0.12364352494478226f;
  t205_n48: return -0.06941196322441101f;
  t205_n49: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t205_n61;
  t205_n50: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t205_n54;
  t205_n51: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t205_n53;
  t205_n52: return -0.06510742008686066f;
  t205_n53: return 0.28385597467422485f;
  t205_n54: if (E.getNumReferences() >= 3288711168 /*535.0*/) goto t205_n58;
  t205_n55: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t205_n57;
  t205_n56: return 0.018677376210689545f;
  t205_n57: return -0.0018703830428421497f;
  t205_n58: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t205_n60;
  t205_n59: return 0.0929400697350502f;
  t205_n60: return -0.15918132662773132f;
  t205_n61: return 0.17376309633255005f;
  t205_n62: return 0.11553031951189041f;
  t205_n63: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t205_n65;
  t205_n64: return -0.11624858528375626f;
  t205_n65: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t205_n71;
  t205_n66: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t205_n68;
  t205_n67: return 0.24977746605873108f;
  t205_n68: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t205_n70;
  t205_n69: return 1.1575425863265991f;
  t205_n70: return -0.10037349164485931f;
  t205_n71: return -0.10649062693119049f;
  t205_n72: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t205_n78;
  t205_n73: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t205_n75;
  t205_n74: return -0.024480240419507027f;
  t205_n75: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t205_n77;
  t205_n76: return 0.11987397819757462f;
  t205_n77: return 0.06949739158153534f;
  t205_n78: return -0.003170775016769767f;
  t205_n79: return -0.006850854493677616f;
  t205_n80: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t205_n82;
  t205_n81: return -0.5340169072151184f;
  t205_n82: return -0.1533389389514923f;
  t205_n83: return 0.014766136184334755f;
  t205_n84: return -0.02388046309351921f;
  t205_n85: if (E.getFractionNameInContext() >= 3191965208 /*0.1889880895614624*/) goto t205_n89;
  t205_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t205_n88;
  t205_n87: return -0.001190840033814311f;
  t205_n88: return 0.0031150963623076677f;
  t205_n89: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t205_n167;
  t205_n90: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t205_n92;
  t205_n91: return -0.0022301606368273497f;
  t205_n92: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t205_n116;
  t205_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t205_n115;
  t205_n94: if (E.getNumReferences() >= 3322161152 /*8458.0*/) goto t205_n100;
  t205_n95: if (E.getNumReferences() >= 3318116352 /*6350.0*/) goto t205_n97;
  t205_n96: return -0.02361833117902279f;
  t205_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t205_n99;
  t205_n98: return -0.10031209141016006f;
  t205_n99: return 0.18595488369464874f;
  t205_n100: if (E.getNumReferences() >= 3325662720 /*11877.5*/) goto t205_n102;
  t205_n101: return -0.21825702488422394f;
  t205_n102: if (E.getNumReferences() >= 3349616960 /*85554.5*/) goto t205_n114;
  t205_n103: if (E.getNumReferences() >= 3346296960 /*62576.5*/) goto t205_n113;
  t205_n104: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t205_n110;
  t205_n105: if (E.getNumReferences() >= 3326428672 /*12625.5*/) goto t205_n109;
  t205_n106: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t205_n108;
  t205_n107: return -0.028555629774928093f;
  t205_n108: return 0.15358221530914307f;
  t205_n109: return -0.12792618572711945f;
  t205_n110: if (E.getNumReferences() >= 3332080128 /*19905.0*/) goto t205_n112;
  t205_n111: return -0.11660310626029968f;
  t205_n112: return 0.10819415748119354f;
  t205_n113: return -0.19067633152008057f;
  t205_n114: return 0.09891252219676971f;
  t205_n115: return 0.007886955514550209f;
  t205_n116: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t205_n166;
  t205_n117: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t205_n165;
  t205_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t205_n120;
  t205_n119: return 0.027160540223121643f;
  t205_n120: if (E.getNumReferences() >= 3321043968 /*7779.5*/) goto t205_n164;
  t205_n121: if (E.getNumReferences() >= 3281125376 /*292.0*/) goto t205_n153;
  t205_n122: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t205_n124;
  t205_n123: return -0.03416929021477699f;
  t205_n124: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t205_n138;
  t205_n125: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t205_n133;
  t205_n126: if (E.getNumReferences() >= 3271098368 /*124.5*/) goto t205_n132;
  t205_n127: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t205_n131;
  t205_n128: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t205_n130;
  t205_n129: return -0.024659277871251106f;
  t205_n130: return 0.14147382974624634f;
  t205_n131: return -0.03447794169187546f;
  t205_n132: return 0.10072046518325806f;
  t205_n133: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t205_n135;
  t205_n134: return 0.023826545104384422f;
  t205_n135: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t205_n137;
  t205_n136: return 0.20120231807231903f;
  t205_n137: return 0.11096891760826111f;
  t205_n138: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t205_n146;
  t205_n139: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t205_n145;
  t205_n140: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t205_n144;
  t205_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t205_n143;
  t205_n142: return -0.10010270029306412f;
  t205_n143: return 0.21830515563488007f;
  t205_n144: return 0.18837563693523407f;
  t205_n145: return -0.05081554502248764f;
  t205_n146: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t205_n148;
  t205_n147: return 0.21033698320388794f;
  t205_n148: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t205_n152;
  t205_n149: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t205_n151;
  t205_n150: return -0.14175710082054138f;
  t205_n151: return 0.6473594903945923f;
  t205_n152: return 0.14306530356407166f;
  t205_n153: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t205_n155;
  t205_n154: return 0.027354011312127113f;
  t205_n155: if (E.getNumReferences() >= 3282976768 /*348.5*/) goto t205_n157;
  t205_n156: return 0.19652284681797028f;
  t205_n157: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t205_n163;
  t205_n158: if (E.getNumReferences() >= 3288104960 /*505.0*/) goto t205_n160;
  t205_n159: return 0.037500396370887756f;
  t205_n160: if (E.getNumReferences() >= 3294224384 /*871.5*/) goto t205_n162;
  t205_n161: return 0.23363518714904785f;
  t205_n162: return 0.09162599593400955f;
  t205_n163: return -0.11960968375205994f;
  t205_n164: return -0.26649343967437744f;
  t205_n165: return -0.024918444454669952f;
  t205_n166: return 0.0037298761308193207f;
  t205_n167: if (E.getFractionNameInContext() >= 3203243738 /*0.464102566242218*/) goto t205_n211;
  t205_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t205_n210;
  t205_n169: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t205_n175;
  t205_n170: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t205_n174;
  t205_n171: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t205_n173;
  t205_n172: return -0.11956819891929626f;
  t205_n173: return 0.21246100962162018f;
  t205_n174: return -0.016110779717564583f;
  t205_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t205_n187;
  t205_n176: if (E.getNumReferences() >= 3314067456 /*4373.0*/) goto t205_n186;
  t205_n177: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t205_n179;
  t205_n178: return 0.0022420636378228664f;
  t205_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t205_n181;
  t205_n180: return -0.0502685010433197f;
  t205_n181: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t205_n183;
  t205_n182: return -0.16059565544128418f;
  t205_n183: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t205_n185;
  t205_n184: return 0.1579316258430481f;
  t205_n185: return 0.05725986883044243f;
  t205_n186: return 0.07765993475914001f;
  t205_n187: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t205_n207;
  t205_n188: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t205_n204;
  t205_n189: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t205_n193;
  t205_n190: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t205_n192;
  t205_n191: return -0.01980600878596306f;
  t205_n192: return 0.11010280251502991f;
  t205_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t205_n195;
  t205_n194: return -0.04605161026120186f;
  t205_n195: if (E.getFractionNameInContext() >= 3194215998 /*0.22252747416496277*/) goto t205_n203;
  t205_n196: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t205_n198;
  t205_n197: return -0.002637619851157069f;
  t205_n198: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t205_n202;
  t205_n199: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t205_n201;
  t205_n200: return -0.13075435161590576f;
  t205_n201: return 0.1284123808145523f;
  t205_n202: return -0.04469350352883339f;
  t205_n203: return -0.02349153906106949f;
  t205_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t205_n206;
  t205_n205: return 0.00876912847161293f;
  t205_n206: return 0.028656721115112305f;
  t205_n207: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t205_n209;
  t205_n208: return -0.012845069169998169f;
  t205_n209: return 0.109808050096035f;
  t205_n210: return 0.03213135153055191f;
  t205_n211: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t205_n223;
  t205_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t205_n216;
  t205_n213: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t205_n215;
  t205_n214: return -0.019308820366859436f;
  t205_n215: return -0.006672906689345837f;
  t205_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t205_n218;
  t205_n217: return -0.016757875680923462f;
  t205_n218: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t205_n222;
  t205_n219: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t205_n221;
  t205_n220: return -0.03369014337658882f;
  t205_n221: return 0.11769061535596848f;
  t205_n222: return 0.03394465520977974f;
  t205_n223: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t205_n235;
  t205_n224: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t205_n234;
  t205_n225: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t205_n231;
  t205_n226: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t205_n228;
  t205_n227: return -0.04172493517398834f;
  t205_n228: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t205_n230;
  t205_n229: return 0.11057206988334656f;
  t205_n230: return -0.01665065810084343f;
  t205_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t205_n233;
  t205_n232: return -0.009869984351098537f;
  t205_n233: return 0.05380208417773247f;
  t205_n234: return 0.017251862213015556f;
  t205_n235: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t205_n253;
  t205_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t205_n238;
  t205_n237: return 0.01061435230076313f;
  t205_n238: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t205_n252;
  t205_n239: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t205_n241;
  t205_n240: return 0.05027381330728531f;
  t205_n241: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t205_n247;
  t205_n242: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t205_n244;
  t205_n243: return 0.015046442858874798f;
  t205_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t205_n246;
  t205_n245: return -0.4659261405467987f;
  t205_n246: return 0.05581042170524597f;
  t205_n247: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t205_n249;
  t205_n248: return 0.03481624275445938f;
  t205_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t205_n251;
  t205_n250: return 0.01287981029599905f;
  t205_n251: return 0.087323859333992f;
  t205_n252: return 0.01262893807142973f;
  t205_n253: return -0.2908034920692444f;
  t205_n254: return 0.0023998143151402473f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree206(const Example& E) {
  t206_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t206_n96;
  t206_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t206_n95;
  t206_n2: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t206_n6;
  t206_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t206_n5;
  t206_n4: return -0.0005485553992912173f;
  t206_n5: return 0.0021510941442102194f;
  t206_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n30;
  t206_n7: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t206_n9;
  t206_n8: return -0.012716282159090042f;
  t206_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t206_n29;
  t206_n10: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t206_n12;
  t206_n11: return -0.10268067568540573f;
  t206_n12: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t206_n14;
  t206_n13: return 0.047397416085004807f;
  t206_n14: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t206_n18;
  t206_n15: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t206_n17;
  t206_n16: return -0.02544553019106388f;
  t206_n17: return 0.11082595586776733f;
  t206_n18: if (E.getNumReferences() >= 3263627264 /*67.5*/) goto t206_n26;
  t206_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t206_n21;
  t206_n20: return 0.012876680120825768f;
  t206_n21: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t206_n23;
  t206_n22: return 0.08310799300670624f;
  t206_n23: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t206_n25;
  t206_n24: return 0.1594735085964203f;
  t206_n25: return 0.13277655839920044f;
  t206_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t206_n28;
  t206_n27: return -0.002794134197756648f;
  t206_n28: return 0.10795573145151138f;
  t206_n29: return 0.09176483750343323f;
  t206_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t206_n94;
  t206_n31: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t206_n61;
  t206_n32: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t206_n40;
  t206_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t206_n39;
  t206_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t206_n36;
  t206_n35: return 0.06281812489032745f;
  t206_n36: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t206_n38;
  t206_n37: return 0.12615419924259186f;
  t206_n38: return -0.12578079104423523f;
  t206_n39: return 0.011557512916624546f;
  t206_n40: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t206_n52;
  t206_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t206_n43;
  t206_n42: return 0.015390807762742043f;
  t206_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t206_n51;
  t206_n44: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t206_n48;
  t206_n45: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t206_n47;
  t206_n46: return 0.07988572120666504f;
  t206_n47: return 0.206670343875885f;
  t206_n48: if (E.getNumReferences() >= 3272671232 /*145.0*/) goto t206_n50;
  t206_n49: return -0.02349531091749668f;
  t206_n50: return 0.07579604536294937f;
  t206_n51: return 0.17237882316112518f;
  t206_n52: if (E.getNumReferences() >= 3303284736 /*1825.0*/) goto t206_n58;
  t206_n53: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t206_n55;
  t206_n54: return -0.02023705467581749f;
  t206_n55: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t206_n57;
  t206_n56: return 0.11516879498958588f;
  t206_n57: return -1.0649813475538394e-08f;
  t206_n58: if (E.getNumReferences() >= 3314014208 /*4347.0*/) goto t206_n60;
  t206_n59: return 0.36678585410118103f;
  t206_n60: return -0.11774760484695435f;
  t206_n61: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n71;
  t206_n62: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t206_n66;
  t206_n63: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t206_n65;
  t206_n64: return 0.02129141241312027f;
  t206_n65: return 0.2284097820520401f;
  t206_n66: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t206_n70;
  t206_n67: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t206_n69;
  t206_n68: return -0.015209892764687538f;
  t206_n69: return 0.10214408487081528f;
  t206_n70: return 0.008816235698759556f;
  t206_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t206_n73;
  t206_n72: return -0.013912294059991837f;
  t206_n73: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t206_n75;
  t206_n74: return 0.029393691569566727f;
  t206_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n77;
  t206_n76: return -0.0007034283480606973f;
  t206_n77: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t206_n91;
  t206_n78: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t206_n88;
  t206_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t206_n81;
  t206_n80: return -0.038340017199516296f;
  t206_n81: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t206_n83;
  t206_n82: return -0.38608500361442566f;
  t206_n83: if (E.getNumReferences() >= 3273457664 /*157.0*/) goto t206_n85;
  t206_n84: return 0.1173536479473114f;
  t206_n85: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t206_n87;
  t206_n86: return 0.2706958055496216f;
  t206_n87: return -0.11872942000627518f;
  t206_n88: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t206_n90;
  t206_n89: return 0.17415539920330048f;
  t206_n90: return -0.03472286835312843f;
  t206_n91: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t206_n93;
  t206_n92: return 0.1906302273273468f;
  t206_n93: return 0.03146045282483101f;
  t206_n94: return 0.1286102682352066f;
  t206_n95: return -0.0014401260996237397f;
  t206_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t206_n98;
  t206_n97: return -4.0113209252012894e-05f;
  t206_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t206_n250;
  t206_n99: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t206_n201;
  t206_n100: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t206_n112;
  t206_n101: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t206_n103;
  t206_n102: return 0.004238346591591835f;
  t206_n103: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n111;
  t206_n104: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t206_n106;
  t206_n105: return 0.033448051661252975f;
  t206_n106: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t206_n108;
  t206_n107: return 0.011670217849314213f;
  t206_n108: if (E.getFractionNameInContext() >= 3198110197 /*0.31111112236976624*/) goto t206_n110;
  t206_n109: return 0.13360968232154846f;
  t206_n110: return 0.3247864842414856f;
  t206_n111: return 0.017764346674084663f;
  t206_n112: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t206_n184;
  t206_n113: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t206_n115;
  t206_n114: return -0.0033991269301623106f;
  t206_n115: if (E.getFractionNameInContext() >= 3172460927 /*0.03708791360259056*/) goto t206_n117;
  t206_n116: return 0.2574666142463684f;
  t206_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t206_n119;
  t206_n118: return -0.12597864866256714f;
  t206_n119: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t206_n173;
  t206_n120: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t206_n138;
  t206_n121: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t206_n127;
  t206_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n126;
  t206_n123: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t206_n125;
  t206_n124: return -1.0598613023757935f;
  t206_n125: return -0.36476844549179077f;
  t206_n126: return -0.0006484161713160574f;
  t206_n127: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t206_n137;
  t206_n128: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t206_n132;
  t206_n129: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t206_n131;
  t206_n130: return 0.06167158484458923f;
  t206_n131: return 0.27871835231781006f;
  t206_n132: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t206_n134;
  t206_n133: return 0.06856631487607956f;
  t206_n134: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t206_n136;
  t206_n135: return 0.1739487648010254f;
  t206_n136: return -0.037381310015916824f;
  t206_n137: return 0.017890769988298416f;
  t206_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t206_n156;
  t206_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n155;
  t206_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t206_n142;
  t206_n141: return -0.03874490037560463f;
  t206_n142: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t206_n144;
  t206_n143: return 0.01960700936615467f;
  t206_n144: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t206_n148;
  t206_n145: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t206_n147;
  t206_n146: return -0.05551521107554436f;
  t206_n147: return 0.03704103082418442f;
  t206_n148: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t206_n154;
  t206_n149: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t206_n151;
  t206_n150: return 0.02833867073059082f;
  t206_n151: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t206_n153;
  t206_n152: return 0.1321757286787033f;
  t206_n153: return 0.01496877707540989f;
  t206_n154: return 0.02646353468298912f;
  t206_n155: return -0.006919031962752342f;
  t206_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t206_n160;
  t206_n157: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t206_n159;
  t206_n158: return 0.02629994787275791f;
  t206_n159: return 0.11260402202606201f;
  t206_n160: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t206_n164;
  t206_n161: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t206_n163;
  t206_n162: return 0.04666462913155556f;
  t206_n163: return 0.08501588553190231f;
  t206_n164: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t206_n172;
  t206_n165: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t206_n171;
  t206_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n168;
  t206_n167: return -0.10010881721973419f;
  t206_n168: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n170;
  t206_n169: return 0.0983470231294632f;
  t206_n170: return 0.0696931853890419f;
  t206_n171: return 0.05257096141576767f;
  t206_n172: return 0.026345031335949898f;
  t206_n173: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t206_n179;
  t206_n174: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t206_n176;
  t206_n175: return 9.733585102367215e-07f;
  t206_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n178;
  t206_n177: return -0.07729370146989822f;
  t206_n178: return 0.12202149629592896f;
  t206_n179: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t206_n181;
  t206_n180: return -0.126961812376976f;
  t206_n181: if (E.getNumReferences() >= 3339033984 /*34205.5*/) goto t206_n183;
  t206_n182: return 0.12404869496822357f;
  t206_n183: return -0.02623867802321911f;
  t206_n184: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t206_n188;
  t206_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t206_n187;
  t206_n186: return -0.0009468593052588403f;
  t206_n187: return 0.01349787600338459f;
  t206_n188: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t206_n190;
  t206_n189: return 0.012301044538617134f;
  t206_n190: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t206_n200;
  t206_n191: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t206_n197;
  t206_n192: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t206_n194;
  t206_n193: return 0.07059439271688461f;
  t206_n194: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t206_n196;
  t206_n195: return 0.12230200320482254f;
  t206_n196: return 0.02305373176932335f;
  t206_n197: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t206_n199;
  t206_n198: return -0.024939442053437233f;
  t206_n199: return 0.09251265972852707f;
  t206_n200: return 0.03063778392970562f;
  t206_n201: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t206_n249;
  t206_n202: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t206_n212;
  t206_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n207;
  t206_n204: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t206_n206;
  t206_n205: return -0.010015549138188362f;
  t206_n206: return -0.01750786416232586f;
  t206_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t206_n211;
  t206_n208: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t206_n210;
  t206_n209: return -0.017793135717511177f;
  t206_n210: return 0.1369817554950714f;
  t206_n211: return 0.0025685816071927547f;
  t206_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t206_n228;
  t206_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t206_n225;
  t206_n214: if (E.getFractionNameInContext() >= 3205753151 /*0.5777778029441833*/) goto t206_n222;
  t206_n215: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t206_n219;
  t206_n216: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t206_n218;
  t206_n217: return 0.22161033749580383f;
  t206_n218: return 0.0723554864525795f;
  t206_n219: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t206_n221;
  t206_n220: return -0.10029854625463486f;
  t206_n221: return 0.15066276490688324f;
  t206_n222: if (E.getNumNameInContext() >= 3229614080 /*4.0*/) goto t206_n224;
  t206_n223: return 0.2176845818758011f;
  t206_n224: return 0.1388341635465622f;
  t206_n225: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t206_n227;
  t206_n226: return -0.15823285281658173f;
  t206_n227: return 0.3246951401233673f;
  t206_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t206_n230;
  t206_n229: return -0.13042153418064117f;
  t206_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t206_n236;
  t206_n231: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t206_n235;
  t206_n232: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t206_n234;
  t206_n233: return 0.013923813588917255f;
  t206_n234: return -0.013248772360384464f;
  t206_n235: return 0.0288372989743948f;
  t206_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t206_n244;
  t206_n237: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t206_n241;
  t206_n238: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t206_n240;
  t206_n239: return 0.017016345635056496f;
  t206_n240: return -0.10055999457836151f;
  t206_n241: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t206_n243;
  t206_n242: return -0.037897732108831406f;
  t206_n243: return 0.11858274787664413f;
  t206_n244: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t206_n248;
  t206_n245: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t206_n247;
  t206_n246: return 0.09006232023239136f;
  t206_n247: return 0.2279384285211563f;
  t206_n248: return 0.05206291377544403f;
  t206_n249: return 0.0008722975035198033f;
  t206_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t206_n254;
  t206_n251: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t206_n253;
  t206_n252: return 0.019606221467256546f;
  t206_n253: return 0.06922072917222977f;
  t206_n254: return -0.10722141712903976f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree207(const Example& E) {
  t207_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t207_n108;
  t207_n1: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t207_n3;
  t207_n2: return 0.00019438250456005335f;
  t207_n3: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t207_n107;
  t207_n4: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t207_n30;
  t207_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n21;
  t207_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n20;
  t207_n7: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t207_n19;
  t207_n8: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t207_n10;
  t207_n9: return -0.019162192940711975f;
  t207_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n18;
  t207_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n17;
  t207_n12: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t207_n14;
  t207_n13: return -0.130849227309227f;
  t207_n14: if (E.getFractionNameInContext() >= 3197279810 /*0.2863636612892151*/) goto t207_n16;
  t207_n15: return -7.0407304763793945f;
  t207_n16: return -0.008026424795389175f;
  t207_n17: return 0.003847751533612609f;
  t207_n18: return 0.009667964652180672f;
  t207_n19: return -0.0020104157738387585f;
  t207_n20: return 0.0028881465550512075f;
  t207_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t207_n23;
  t207_n22: return 0.008006243966519833f;
  t207_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t207_n27;
  t207_n24: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t207_n26;
  t207_n25: return -0.0037945357616990805f;
  t207_n26: return 0.11608416587114334f;
  t207_n27: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t207_n29;
  t207_n28: return 0.007286466658115387f;
  t207_n29: return 0.028369508683681488f;
  t207_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n34;
  t207_n31: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t207_n33;
  t207_n32: return -0.02250467613339424f;
  t207_n33: return 0.01546049490571022f;
  t207_n34: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t207_n70;
  t207_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n41;
  t207_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n38;
  t207_n37: return -0.08878757059574127f;
  t207_n38: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t207_n40;
  t207_n39: return -0.0064981915056705475f;
  t207_n40: return -0.10699359327554703f;
  t207_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t207_n55;
  t207_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n44;
  t207_n43: return -0.02540510706603527f;
  t207_n44: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t207_n48;
  t207_n45: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t207_n47;
  t207_n46: return 0.001060498645529151f;
  t207_n47: return 0.12089184671640396f;
  t207_n48: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t207_n50;
  t207_n49: return 0.013169505633413792f;
  t207_n50: if (E.getSymbolScopeDistanceCost() >= 3247964160 /*19.0*/) goto t207_n54;
  t207_n51: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t207_n53;
  t207_n52: return -0.12876762449741364f;
  t207_n53: return 0.3678223788738251f;
  t207_n54: return 0.027495307847857475f;
  t207_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t207_n61;
  t207_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t207_n58;
  t207_n57: return -0.003273008856922388f;
  t207_n58: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t207_n60;
  t207_n59: return -0.028288112953305244f;
  t207_n60: return 0.021425215527415276f;
  t207_n61: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t207_n63;
  t207_n62: return 0.06221161037683487f;
  t207_n63: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t207_n65;
  t207_n64: return 0.08773145079612732f;
  t207_n65: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t207_n69;
  t207_n66: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t207_n68;
  t207_n67: return -0.29439520835876465f;
  t207_n68: return 0.029230089858174324f;
  t207_n69: return -0.963269054889679f;
  t207_n70: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t207_n76;
  t207_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n73;
  t207_n72: return -0.020655812695622444f;
  t207_n73: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t207_n75;
  t207_n74: return -0.001280769007280469f;
  t207_n75: return 0.042740676552057266f;
  t207_n76: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t207_n86;
  t207_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t207_n79;
  t207_n78: return -0.0014432307798415422f;
  t207_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t207_n81;
  t207_n80: return -0.038815878331661224f;
  t207_n81: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t207_n85;
  t207_n82: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t207_n84;
  t207_n83: return 0.024447649717330933f;
  t207_n84: return -0.08956863731145859f;
  t207_n85: return -0.01911989040672779f;
  t207_n86: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t207_n96;
  t207_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t207_n89;
  t207_n88: return -0.02544759027659893f;
  t207_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t207_n91;
  t207_n90: return 0.022598231211304665f;
  t207_n91: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t207_n95;
  t207_n92: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t207_n94;
  t207_n93: return -0.04446372017264366f;
  t207_n94: return 0.12472458928823471f;
  t207_n95: return 0.1420649141073227f;
  t207_n96: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t207_n100;
  t207_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n99;
  t207_n98: return -0.048724401742219925f;
  t207_n99: return 0.017415866255760193f;
  t207_n100: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t207_n106;
  t207_n101: if (E.getSymbolScopeDistanceCost() >= 3250585600 /*24.0*/) goto t207_n105;
  t207_n102: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t207_n104;
  t207_n103: return -0.01254893559962511f;
  t207_n104: return 0.10236725211143494f;
  t207_n105: return -0.06996390968561172f;
  t207_n106: return 0.0010975097538903356f;
  t207_n107: return -0.01616525836288929f;
  t207_n108: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t207_n210;
  t207_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n163;
  t207_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t207_n112;
  t207_n111: return -0.0002473389613442123f;
  t207_n112: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t207_n158;
  t207_n113: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t207_n157;
  t207_n114: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t207_n124;
  t207_n115: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t207_n121;
  t207_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t207_n118;
  t207_n117: return -0.0913688912987709f;
  t207_n118: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t207_n120;
  t207_n119: return 0.15352216362953186f;
  t207_n120: return -0.019997959956526756f;
  t207_n121: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t207_n123;
  t207_n122: return -0.00394678907468915f;
  t207_n123: return -0.06685402244329453f;
  t207_n124: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t207_n126;
  t207_n125: return -0.0032904697582125664f;
  t207_n126: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t207_n140;
  t207_n127: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t207_n139;
  t207_n128: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t207_n132;
  t207_n129: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t207_n131;
  t207_n130: return 0.021155986934900284f;
  t207_n131: return -0.12828077375888824f;
  t207_n132: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t207_n136;
  t207_n133: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t207_n135;
  t207_n134: return -0.06774980574846268f;
  t207_n135: return 0.09111828356981277f;
  t207_n136: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t207_n138;
  t207_n137: return -0.1588008999824524f;
  t207_n138: return 0.07116677612066269f;
  t207_n139: return 0.004153701476752758f;
  t207_n140: if (E.getFractionNameInContext() >= 3199399908 /*0.34954750537872314*/) goto t207_n142;
  t207_n141: return -0.13599367439746857f;
  t207_n142: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t207_n148;
  t207_n143: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t207_n145;
  t207_n144: return 0.09113278239965439f;
  t207_n145: if (E.getFractionNameInContext() >= 3200353816 /*0.3779761791229248*/) goto t207_n147;
  t207_n146: return 0.10947487503290176f;
  t207_n147: return 0.11343957483768463f;
  t207_n148: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t207_n156;
  t207_n149: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t207_n153;
  t207_n150: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t207_n152;
  t207_n151: return 0.009300321340560913f;
  t207_n152: return 0.09716729074716568f;
  t207_n153: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t207_n155;
  t207_n154: return 0.08581309020519257f;
  t207_n155: return 0.13229314982891083f;
  t207_n156: return -0.010142644867300987f;
  t207_n157: return 0.002101888647302985f;
  t207_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t207_n160;
  t207_n159: return 0.0004773197288159281f;
  t207_n160: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t207_n162;
  t207_n161: return 0.005106438882648945f;
  t207_n162: return 0.014226988889276981f;
  t207_n163: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t207_n165;
  t207_n164: return 0.0067806788720190525f;
  t207_n165: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t207_n173;
  t207_n166: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t207_n172;
  t207_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n169;
  t207_n168: return -0.007060345262289047f;
  t207_n169: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t207_n171;
  t207_n170: return 0.024379590526223183f;
  t207_n171: return 0.09258376061916351f;
  t207_n172: return -0.08441556245088577f;
  t207_n173: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t207_n185;
  t207_n174: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t207_n176;
  t207_n175: return -0.032093875110149384f;
  t207_n176: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t207_n180;
  t207_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n179;
  t207_n178: return -0.03718927502632141f;
  t207_n179: return 0.04188990220427513f;
  t207_n180: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t207_n184;
  t207_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t207_n183;
  t207_n182: return 0.14689743518829346f;
  t207_n183: return 0.04655719920992851f;
  t207_n184: return 0.038592398166656494f;
  t207_n185: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t207_n207;
  t207_n186: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t207_n192;
  t207_n187: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t207_n189;
  t207_n188: return 0.010116344317793846f;
  t207_n189: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t207_n191;
  t207_n190: return 0.003699900582432747f;
  t207_n191: return -0.06877932697534561f;
  t207_n192: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t207_n194;
  t207_n193: return -0.00965686235576868f;
  t207_n194: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t207_n206;
  t207_n195: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t207_n201;
  t207_n196: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t207_n198;
  t207_n197: return 0.002371876733377576f;
  t207_n198: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t207_n200;
  t207_n199: return -1.439780831336975f;
  t207_n200: return -0.24182891845703125f;
  t207_n201: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t207_n205;
  t207_n202: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t207_n204;
  t207_n203: return 0.11228181421756744f;
  t207_n204: return 0.029294369742274284f;
  t207_n205: return 0.029865456745028496f;
  t207_n206: return 0.1594974249601364f;
  t207_n207: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t207_n209;
  t207_n208: return 0.038555800914764404f;
  t207_n209: return -0.049726974219083786f;
  t207_n210: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t207_n254;
  t207_n211: if (E.getNumReferences() >= 3306108928 /*2291.5*/) goto t207_n251;
  t207_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t207_n230;
  t207_n213: if (E.getNumReferences() >= 3290923008 /*670.0*/) goto t207_n221;
  t207_n214: if (E.getNumReferences() >= 3290873856 /*667.0*/) goto t207_n220;
  t207_n215: if (E.getNumReferences() >= 3289956352 /*611.0*/) goto t207_n217;
  t207_n216: return -0.29959163069725037f;
  t207_n217: if (E.getNumReferences() >= 3290038272 /*616.0*/) goto t207_n219;
  t207_n218: return 0.190138041973114f;
  t207_n219: return 0.06155373901128769f;
  t207_n220: return 0.13603109121322632f;
  t207_n221: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t207_n225;
  t207_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t207_n224;
  t207_n223: return -0.28979921340942383f;
  t207_n224: return 0.010976889170706272f;
  t207_n225: if (E.getNumReferences() >= 3292921856 /*792.0*/) goto t207_n229;
  t207_n226: if (E.getNumReferences() >= 3292749824 /*781.5*/) goto t207_n228;
  t207_n227: return -0.09986767172813416f;
  t207_n228: return -0.7282330989837646f;
  t207_n229: return 0.003486678237095475f;
  t207_n230: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t207_n242;
  t207_n231: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t207_n233;
  t207_n232: return 0.034910134971141815f;
  t207_n233: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t207_n235;
  t207_n234: return 0.021420419216156006f;
  t207_n235: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t207_n241;
  t207_n236: if (E.getNumReferences() >= 3294715904 /*901.5*/) goto t207_n240;
  t207_n237: if (E.getNumReferences() >= 3293921280 /*853.0*/) goto t207_n239;
  t207_n238: return 0.07821109890937805f;
  t207_n239: return 0.11001583933830261f;
  t207_n240: return 0.028440823778510094f;
  t207_n241: return 0.08454901725053787f;
  t207_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t207_n246;
  t207_n243: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t207_n245;
  t207_n244: return 0.002666893880814314f;
  t207_n245: return -0.2322511374950409f;
  t207_n246: if (E.getNumReferences() >= 3290685440 /*655.5*/) goto t207_n250;
  t207_n247: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t207_n249;
  t207_n248: return 0.11430465430021286f;
  t207_n249: return -0.13532456755638123f;
  t207_n250: return 0.03968232870101929f;
  t207_n251: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t207_n253;
  t207_n252: return 0.005740271415561438f;
  t207_n253: return 0.05845693126320839f;
  t207_n254: return -0.2902987599372864f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree208(const Example& E) {
  t208_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t208_n254;
  t208_n1: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t208_n3;
  t208_n2: return -0.0003427371848374605f;
  t208_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t208_n5;
  t208_n4: return 0.008145561441779137f;
  t208_n5: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t208_n7;
  t208_n6: return 0.11587418615818024f;
  t208_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t208_n143;
  t208_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t208_n140;
  t208_n9: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t208_n11;
  t208_n10: return -0.11734238266944885f;
  t208_n11: if (E.getNumReferences() >= 3327053312 /*13235.5*/) goto t208_n23;
  t208_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t208_n14;
  t208_n13: return -0.05251917243003845f;
  t208_n14: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t208_n16;
  t208_n15: return -0.013002108782529831f;
  t208_n16: if (E.getNumReferences() >= 3319603200 /*7076.0*/) goto t208_n20;
  t208_n17: if (E.getNumReferences() >= 3318611968 /*6592.0*/) goto t208_n19;
  t208_n18: return 0.003919155802577734f;
  t208_n19: return -0.5982512831687927f;
  t208_n20: if (E.getNumReferences() >= 3319744512 /*7145.0*/) goto t208_n22;
  t208_n21: return 0.22944845259189606f;
  t208_n22: return 0.04345818981528282f;
  t208_n23: if (E.getNumReferences() >= 3327144448 /*13324.5*/) goto t208_n29;
  t208_n24: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t208_n26;
  t208_n25: return -0.11029400676488876f;
  t208_n26: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t208_n28;
  t208_n27: return 0.22586886584758759f;
  t208_n28: return 0.2307373285293579f;
  t208_n29: if (E.getNumReferences() >= 3330499328 /*16817.5*/) goto t208_n57;
  t208_n30: if (E.getNumReferences() >= 3327636992 /*13805.5*/) goto t208_n32;
  t208_n31: return -0.035530418157577515f;
  t208_n32: if (E.getNumReferences() >= 3327717888 /*13884.5*/) goto t208_n40;
  t208_n33: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t208_n35;
  t208_n34: return -0.03725532442331314f;
  t208_n35: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t208_n39;
  t208_n36: if (E.getNumReferences() >= 3327674368 /*13842.0*/) goto t208_n38;
  t208_n37: return 0.10213132947683334f;
  t208_n38: return 0.3380003571510315f;
  t208_n39: return 0.03663242980837822f;
  t208_n40: if (E.getNumReferences() >= 3330487040 /*16793.5*/) goto t208_n54;
  t208_n41: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t208_n49;
  t208_n42: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t208_n44;
  t208_n43: return 0.00012432194489520043f;
  t208_n44: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t208_n48;
  t208_n45: if (E.getNumReferences() >= 3328821248 /*14962.0*/) goto t208_n47;
  t208_n46: return -0.13319674134254456f;
  t208_n47: return 0.13298079371452332f;
  t208_n48: return 0.11915253847837448f;
  t208_n49: if (E.getNumReferences() >= 3329438720 /*15565.0*/) goto t208_n51;
  t208_n50: return -0.059744879603385925f;
  t208_n51: if (E.getNumReferences() >= 3330296064 /*16420.5*/) goto t208_n53;
  t208_n52: return 0.08203519135713577f;
  t208_n53: return -0.1653883457183838f;
  t208_n54: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t208_n56;
  t208_n55: return -0.04325033724308014f;
  t208_n56: return 0.14321352541446686f;
  t208_n57: if (E.getNumReferences() >= 3331498752 /*18769.5*/) goto t208_n59;
  t208_n58: return -0.14406967163085938f;
  t208_n59: if (E.getNumReferences() >= 3331793408 /*19345.0*/) goto t208_n69;
  t208_n60: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t208_n66;
  t208_n61: if (E.getNumReferences() >= 3331557888 /*18885.0*/) goto t208_n65;
  t208_n62: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t208_n64;
  t208_n63: return -0.01277067419141531f;
  t208_n64: return 0.11597904562950134f;
  t208_n65: return -0.18262363970279694f;
  t208_n66: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t208_n68;
  t208_n67: return 0.16661019623279572f;
  t208_n68: return 0.12706327438354492f;
  t208_n69: if (E.getNumReferences() >= 3336160512 /*27874.5*/) goto t208_n81;
  t208_n70: if (E.getNumReferences() >= 3332305408 /*20345.0*/) goto t208_n80;
  t208_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t208_n75;
  t208_n72: if (E.getNumReferences() >= 3332112384 /*19968.0*/) goto t208_n74;
  t208_n73: return -0.20758554339408875f;
  t208_n74: return 0.04260384291410446f;
  t208_n75: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t208_n79;
  t208_n76: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t208_n78;
  t208_n77: return 0.06562275439500809f;
  t208_n78: return -0.28540584444999695f;
  t208_n79: return 0.12158913165330887f;
  t208_n80: return -0.07535776495933533f;
  t208_n81: if (E.getNumReferences() >= 3336764160 /*29053.5*/) goto t208_n89;
  t208_n82: if (E.getNumReferences() >= 3336534016 /*28604.0*/) goto t208_n84;
  t208_n83: return 0.027139395475387573f;
  t208_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t208_n86;
  t208_n85: return 0.07815387845039368f;
  t208_n86: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t208_n88;
  t208_n87: return 0.1635686159133911f;
  t208_n88: return 0.1620003879070282f;
  t208_n89: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t208_n91;
  t208_n90: return -0.39616864919662476f;
  t208_n91: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t208_n93;
  t208_n92: return 0.09719782322645187f;
  t208_n93: if (E.getNumReferences() >= 3338616320 /*32671.0*/) goto t208_n97;
  t208_n94: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t208_n96;
  t208_n95: return -0.12629002332687378f;
  t208_n96: return -0.45813247561454773f;
  t208_n97: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t208_n107;
  t208_n98: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t208_n102;
  t208_n99: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t208_n101;
  t208_n100: return 0.0062875947915017605f;
  t208_n101: return -0.5516344904899597f;
  t208_n102: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t208_n106;
  t208_n103: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t208_n105;
  t208_n104: return 0.1330050826072693f;
  t208_n105: return 0.11127164214849472f;
  t208_n106: return 0.046756528317928314f;
  t208_n107: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t208_n139;
  t208_n108: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t208_n134;
  t208_n109: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t208_n133;
  t208_n110: if (E.getNumReferences() >= 3339443456 /*35805.0*/) goto t208_n112;
  t208_n111: return -0.22883881628513336f;
  t208_n112: if (E.getNumReferences() >= 3340902912 /*41506.0*/) goto t208_n114;
  t208_n113: return 0.11415217071771622f;
  t208_n114: if (E.getNumReferences() >= 3348072384 /*73487.5*/) goto t208_n122;
  t208_n115: if (E.getNumReferences() >= 3345400960 /*59076.5*/) goto t208_n121;
  t208_n116: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t208_n120;
  t208_n117: if (E.getNumReferences() >= 3343545472 /*51828.5*/) goto t208_n119;
  t208_n118: return 0.011288527399301529f;
  t208_n119: return -0.2911071479320526f;
  t208_n120: return 0.07146653532981873f;
  t208_n121: return -0.0787549614906311f;
  t208_n122: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t208_n124;
  t208_n123: return 0.07699470967054367f;
  t208_n124: if (E.getNumReferences() >= 3349883520 /*87637.0*/) goto t208_n128;
  t208_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t208_n127;
  t208_n126: return -0.26826030015945435f;
  t208_n127: return 0.036568280309438705f;
  t208_n128: if (E.getNumReferences() >= 3350198912 /*90101.0*/) goto t208_n130;
  t208_n129: return 0.08356720209121704f;
  t208_n130: if (E.getNumReferences() >= 3350654080 /*93657.0*/) goto t208_n132;
  t208_n131: return -0.20857661962509155f;
  t208_n132: return 0.05569455027580261f;
  t208_n133: return -0.0549299381673336f;
  t208_n134: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t208_n138;
  t208_n135: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t208_n137;
  t208_n136: return 0.09585598856210709f;
  t208_n137: return 0.039391230791807175f;
  t208_n138: return 0.12086441367864609f;
  t208_n139: return -0.03215525671839714f;
  t208_n140: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t208_n142;
  t208_n141: return 0.030161846429109573f;
  t208_n142: return -0.07126197218894958f;
  t208_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t208_n147;
  t208_n144: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t208_n146;
  t208_n145: return -0.0023377605248242617f;
  t208_n146: return -0.06050194427371025f;
  t208_n147: if (E.getNumReferences() >= 3316164608 /*5397.0*/) goto t208_n223;
  t208_n148: if (E.getNumReferences() >= 3315997696 /*5315.5*/) goto t208_n218;
  t208_n149: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t208_n151;
  t208_n150: return 0.14582201838493347f;
  t208_n151: if (E.getNumReferences() >= 3310782464 /*3432.5*/) goto t208_n167;
  t208_n152: if (E.getNumReferences() >= 3306967040 /*2501.0*/) goto t208_n162;
  t208_n153: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t208_n161;
  t208_n154: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t208_n160;
  t208_n155: if (E.getNumReferences() >= 3306846208 /*2471.5*/) goto t208_n159;
  t208_n156: if (E.getNumReferences() >= 3306639360 /*2421.0*/) goto t208_n158;
  t208_n157: return 0.1336616575717926f;
  t208_n158: return 0.11266788095235825f;
  t208_n159: return 0.11419027298688889f;
  t208_n160: return 0.00650007463991642f;
  t208_n161: return 0.127282053232193f;
  t208_n162: if (E.getNumReferences() >= 3307317248 /*2586.5*/) goto t208_n164;
  t208_n163: return -0.08732689917087555f;
  t208_n164: if (E.getNumReferences() >= 3308132352 /*2785.5*/) goto t208_n166;
  t208_n165: return 0.06312263011932373f;
  t208_n166: return -0.006022676359862089f;
  t208_n167: if (E.getNumReferences() >= 3312549888 /*3864.0*/) goto t208_n189;
  t208_n168: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t208_n178;
  t208_n169: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t208_n171;
  t208_n170: return 0.10043483227491379f;
  t208_n171: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t208_n175;
  t208_n172: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t208_n174;
  t208_n173: return 0.15297982096672058f;
  t208_n174: return 0.10215838253498077f;
  t208_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t208_n177;
  t208_n176: return -0.25018584728240967f;
  t208_n177: return 0.16390782594680786f;
  t208_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t208_n180;
  t208_n179: return 0.050107311457395554f;
  t208_n180: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t208_n188;
  t208_n181: if (E.getNumReferences() >= 3311169536 /*3527.0*/) goto t208_n187;
  t208_n182: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t208_n184;
  t208_n183: return 0.10868111997842789f;
  t208_n184: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t208_n186;
  t208_n185: return 0.17572832107543945f;
  t208_n186: return 0.1551651805639267f;
  t208_n187: return 0.08722924441099167f;
  t208_n188: return 0.14009234309196472f;
  t208_n189: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t208_n191;
  t208_n190: return -0.06585178524255753f;
  t208_n191: if (E.getNumReferences() >= 3315294208 /*4972.0*/) goto t208_n217;
  t208_n192: if (E.getNumReferences() >= 3315181568 /*4917.0*/) goto t208_n214;
  t208_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t208_n195;
  t208_n194: return 0.05290678143501282f;
  t208_n195: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t208_n199;
  t208_n196: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t208_n198;
  t208_n197: return 0.0013326747575774789f;
  t208_n198: return 0.15471284091472626f;
  t208_n199: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t208_n211;
  t208_n200: if (E.getNumReferences() >= 3314460672 /*4565.0*/) goto t208_n202;
  t208_n201: return 0.036550045013427734f;
  t208_n202: if (E.getNumReferences() >= 3314841600 /*4751.0*/) goto t208_n210;
  t208_n203: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t208_n209;
  t208_n204: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t208_n206;
  t208_n205: return 0.11105719208717346f;
  t208_n206: if (E.getNumReferences() >= 3314592768 /*4629.5*/) goto t208_n208;
  t208_n207: return 0.12966518104076385f;
  t208_n208: return 0.07806669175624847f;
  t208_n209: return 0.022776488214731216f;
  t208_n210: return -0.0007179211243055761f;
  t208_n211: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t208_n213;
  t208_n212: return 0.156686469912529f;
  t208_n213: return -0.22438882291316986f;
  t208_n214: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t208_n216;
  t208_n215: return 0.13112927973270416f;
  t208_n216: return 0.09477139264345169f;
  t208_n217: return -0.009655253030359745f;
  t208_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t208_n220;
  t208_n219: return 0.09639054536819458f;
  t208_n220: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t208_n222;
  t208_n221: return 0.11793170869350433f;
  t208_n222: return -0.10075896233320236f;
  t208_n223: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t208_n225;
  t208_n224: return -0.0178639218211174f;
  t208_n225: if (E.getNumReferences() >= 3321896960 /*8200.0*/) goto t208_n229;
  t208_n226: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t208_n228;
  t208_n227: return -0.09551037847995758f;
  t208_n228: return -0.012202988378703594f;
  t208_n229: if (E.getNumReferences() >= 3322147840 /*8445.0*/) goto t208_n241;
  t208_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t208_n232;
  t208_n231: return 0.010154003277420998f;
  t208_n232: if (E.getNumReferences() >= 3322048000 /*8347.5*/) goto t208_n234;
  t208_n233: return 0.17830395698547363f;
  t208_n234: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t208_n236;
  t208_n235: return 0.31481093168258667f;
  t208_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t208_n238;
  t208_n237: return -0.10189343243837357f;
  t208_n238: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t208_n240;
  t208_n239: return 0.223653182387352f;
  t208_n240: return -0.10033552348613739f;
  t208_n241: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t208_n253;
  t208_n242: if (E.getNumReferences() >= 3361153472 /*220295.0*/) goto t208_n246;
  t208_n243: if (E.getNumReferences() >= 3360198688 /*205376.5*/) goto t208_n245;
  t208_n244: return 0.014987103641033173f;
  t208_n245: return 0.032146524637937546f;
  t208_n246: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t208_n250;
  t208_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t208_n249;
  t208_n248: return -0.20146392285823822f;
  t208_n249: return 0.06757905334234238f;
  t208_n250: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t208_n252;
  t208_n251: return -0.01616618037223816f;
  t208_n252: return 0.013323524966835976f;
  t208_n253: return -0.010578149929642677f;
  t208_n254: return 0.001424298039637506f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree209(const Example& E) {
  t209_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t209_n2;
  t209_n1: return -0.00014565260789822787f;
  t209_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t209_n34;
  t209_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t209_n19;
  t209_n4: if (E.getNumReferences() >= 3319705600 /*7126.0*/) goto t209_n18;
  t209_n5: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t209_n9;
  t209_n6: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t209_n8;
  t209_n7: return -0.0025835814885795116f;
  t209_n8: return 0.02590089850127697f;
  t209_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t209_n11;
  t209_n10: return -0.30017155408859253f;
  t209_n11: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t209_n13;
  t209_n12: return 0.02201133593916893f;
  t209_n13: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t209_n17;
  t209_n14: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t209_n16;
  t209_n15: return -0.08138682693243027f;
  t209_n16: return -0.3253145217895508f;
  t209_n17: return 0.00396716920658946f;
  t209_n18: return -0.15128396451473236f;
  t209_n19: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t209_n31;
  t209_n20: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t209_n24;
  t209_n21: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t209_n23;
  t209_n22: return -0.036956798285245895f;
  t209_n23: return 0.0998588502407074f;
  t209_n24: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t209_n28;
  t209_n25: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t209_n27;
  t209_n26: return 0.06959941238164902f;
  t209_n27: return -0.008326399140059948f;
  t209_n28: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t209_n30;
  t209_n29: return 0.12393384426832199f;
  t209_n30: return 0.08219082653522491f;
  t209_n31: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t209_n33;
  t209_n32: return 0.009986631572246552f;
  t209_n33: return 0.025451533496379852f;
  t209_n34: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t209_n250;
  t209_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t209_n163;
  t209_n36: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t209_n158;
  t209_n37: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t209_n109;
  t209_n38: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t209_n108;
  t209_n39: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t209_n49;
  t209_n40: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t209_n46;
  t209_n41: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t209_n43;
  t209_n42: return 0.0027278230991214514f;
  t209_n43: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t209_n45;
  t209_n44: return 0.09393750876188278f;
  t209_n45: return -0.006432475056499243f;
  t209_n46: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t209_n48;
  t209_n47: return -2.334617853164673f;
  t209_n48: return -0.08630499243736267f;
  t209_n49: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t209_n65;
  t209_n50: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t209_n52;
  t209_n51: return 0.05588735640048981f;
  t209_n52: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t209_n62;
  t209_n53: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t209_n55;
  t209_n54: return 0.07594712823629379f;
  t209_n55: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t209_n61;
  t209_n56: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t209_n58;
  t209_n57: return 0.018067626282572746f;
  t209_n58: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t209_n60;
  t209_n59: return 0.05861598625779152f;
  t209_n60: return 0.2268056571483612f;
  t209_n61: return -0.3673512041568756f;
  t209_n62: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t209_n64;
  t209_n63: return -0.07121063768863678f;
  t209_n64: return 0.14197291433811188f;
  t209_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t209_n77;
  t209_n66: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t209_n76;
  t209_n67: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t209_n75;
  t209_n68: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t209_n70;
  t209_n69: return 0.06614407896995544f;
  t209_n70: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t209_n72;
  t209_n71: return -0.00919303297996521f;
  t209_n72: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t209_n74;
  t209_n73: return -1.8488991260528564f;
  t209_n74: return -0.10645414888858795f;
  t209_n75: return -0.031045766547322273f;
  t209_n76: return -0.06432821601629257f;
  t209_n77: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t209_n87;
  t209_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t209_n80;
  t209_n79: return -0.06305299699306488f;
  t209_n80: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t209_n82;
  t209_n81: return -0.026994189247488976f;
  t209_n82: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t209_n84;
  t209_n83: return 0.1515788733959198f;
  t209_n84: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t209_n86;
  t209_n85: return 0.12689325213432312f;
  t209_n86: return 0.010052516125142574f;
  t209_n87: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t209_n93;
  t209_n88: if (E.getFractionNameInContext() >= 3207689309 /*0.6931818127632141*/) goto t209_n92;
  t209_n89: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t209_n91;
  t209_n90: return 0.13423453271389008f;
  t209_n91: return 0.07031866908073425f;
  t209_n92: return 0.19227783381938934f;
  t209_n93: if (E.getNumReferences() >= 3239051264 /*9.0*/) goto t209_n107;
  t209_n94: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t209_n98;
  t209_n95: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t209_n97;
  t209_n96: return -0.15522056818008423f;
  t209_n97: return -0.42121633887290955f;
  t209_n98: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t209_n102;
  t209_n99: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t209_n101;
  t209_n100: return -0.12284883111715317f;
  t209_n101: return 0.06155204027891159f;
  t209_n102: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t209_n106;
  t209_n103: if (E.getFractionNameInContext() >= 3201314051 /*0.4065934121608734*/) goto t209_n105;
  t209_n104: return 0.0933539867401123f;
  t209_n105: return 0.2263505607843399f;
  t209_n106: return -0.2238750457763672f;
  t209_n107: return 0.13462136685848236f;
  t209_n108: return -0.00960214901715517f;
  t209_n109: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t209_n125;
  t209_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t209_n122;
  t209_n111: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t209_n121;
  t209_n112: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t209_n120;
  t209_n113: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t209_n119;
  t209_n114: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t209_n116;
  t209_n115: return -0.057575371116399765f;
  t209_n116: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t209_n118;
  t209_n117: return 0.17922407388687134f;
  t209_n118: return -0.22337552905082703f;
  t209_n119: return 0.17913556098937988f;
  t209_n120: return -0.061085741966962814f;
  t209_n121: return -0.2867049276828766f;
  t209_n122: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t209_n124;
  t209_n123: return -0.007692300248891115f;
  t209_n124: return 0.13993403315544128f;
  t209_n125: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t209_n143;
  t209_n126: if (E.getNumReferences() >= 3322480128 /*8769.5*/) goto t209_n140;
  t209_n127: if (E.getNumReferences() >= 3315065856 /*4860.5*/) goto t209_n139;
  t209_n128: if (E.getNumReferences() >= 3283468288 /*363.5*/) goto t209_n130;
  t209_n129: return 0.014859075658023357f;
  t209_n130: if (E.getNumReferences() >= 3285172224 /*415.5*/) goto t209_n138;
  t209_n131: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t209_n137;
  t209_n132: if (E.getNumReferences() >= 3284615168 /*398.5*/) goto t209_n136;
  t209_n133: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t209_n135;
  t209_n134: return 0.10954996198415756f;
  t209_n135: return 0.037383630871772766f;
  t209_n136: return 0.12623871862888336f;
  t209_n137: return -0.034020159393548965f;
  t209_n138: return 0.02662748284637928f;
  t209_n139: return -0.11600128561258316f;
  t209_n140: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t209_n142;
  t209_n141: return 0.05325585603713989f;
  t209_n142: return 0.23654349148273468f;
  t209_n143: if (E.getNumReferences() >= 3281190912 /*294.0*/) goto t209_n151;
  t209_n144: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t209_n146;
  t209_n145: return -0.107187919318676f;
  t209_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t209_n148;
  t209_n147: return 0.0011163550661876798f;
  t209_n148: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t209_n150;
  t209_n149: return 0.036711689084768295f;
  t209_n150: return 0.11548103392124176f;
  t209_n151: if (E.getNumReferences() >= 3282436096 /*332.0*/) goto t209_n153;
  t209_n152: return 0.12211597710847855f;
  t209_n153: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t209_n155;
  t209_n154: return -0.11193759739398956f;
  t209_n155: if (E.getNumReferences() >= 3291611136 /*712.0*/) goto t209_n157;
  t209_n156: return 0.11434526741504669f;
  t209_n157: return 0.043758392333984375f;
  t209_n158: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t209_n162;
  t209_n159: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t209_n161;
  t209_n160: return 0.13127565383911133f;
  t209_n161: return -0.01636001281440258f;
  t209_n162: return -0.012101301923394203f;
  t209_n163: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t209_n207;
  t209_n164: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t209_n206;
  t209_n165: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t209_n175;
  t209_n166: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t209_n172;
  t209_n167: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t209_n169;
  t209_n168: return -0.003060518065467477f;
  t209_n169: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t209_n171;
  t209_n170: return -0.03189654275774956f;
  t209_n171: return -0.29634934663772583f;
  t209_n172: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t209_n174;
  t209_n173: return 0.01991972140967846f;
  t209_n174: return 0.004226201679557562f;
  t209_n175: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t209_n177;
  t209_n176: return 0.06798003613948822f;
  t209_n177: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t209_n183;
  t209_n178: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t209_n182;
  t209_n179: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t209_n181;
  t209_n180: return 0.16898994147777557f;
  t209_n181: return -0.025998856872320175f;
  t209_n182: return 0.07766610383987427f;
  t209_n183: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t209_n199;
  t209_n184: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t209_n196;
  t209_n185: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t209_n195;
  t209_n186: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t209_n194;
  t209_n187: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t209_n193;
  t209_n188: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t209_n190;
  t209_n189: return 0.010968301445245743f;
  t209_n190: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t209_n192;
  t209_n191: return -0.03130703046917915f;
  t209_n192: return -0.25192469358444214f;
  t209_n193: return 0.04538140073418617f;
  t209_n194: return -0.012564141303300858f;
  t209_n195: return 0.03392840921878815f;
  t209_n196: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t209_n198;
  t209_n197: return 0.25759631395339966f;
  t209_n198: return 0.1219634860754013f;
  t209_n199: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t209_n201;
  t209_n200: return -0.013584508560597897f;
  t209_n201: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t209_n203;
  t209_n202: return 0.034705471247434616f;
  t209_n203: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t209_n205;
  t209_n204: return 0.16739048063755035f;
  t209_n205: return -0.19098521769046783f;
  t209_n206: return 0.020652292296290398f;
  t209_n207: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t209_n219;
  t209_n208: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t209_n210;
  t209_n209: return 0.02331049181520939f;
  t209_n210: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t209_n216;
  t209_n211: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t209_n213;
  t209_n212: return 0.022038957104086876f;
  t209_n213: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t209_n215;
  t209_n214: return -0.14917245507240295f;
  t209_n215: return 0.02292700670659542f;
  t209_n216: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t209_n218;
  t209_n217: return -0.7118356823921204f;
  t209_n218: return -0.01576012186706066f;
  t209_n219: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t209_n239;
  t209_n220: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t209_n238;
  t209_n221: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t209_n229;
  t209_n222: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t209_n228;
  t209_n223: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t209_n227;
  t209_n224: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t209_n226;
  t209_n225: return -0.029241040349006653f;
  t209_n226: return 0.002690085442736745f;
  t209_n227: return 0.027973970398306847f;
  t209_n228: return 0.03299345076084137f;
  t209_n229: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t209_n235;
  t209_n230: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t209_n232;
  t209_n231: return -0.021054374054074287f;
  t209_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t209_n234;
  t209_n233: return -0.8177648782730103f;
  t209_n234: return 0.0033736564218997955f;
  t209_n235: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t209_n237;
  t209_n236: return -0.6606543064117432f;
  t209_n237: return -0.15064765512943268f;
  t209_n238: return -0.03940398991107941f;
  t209_n239: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t209_n249;
  t209_n240: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t209_n248;
  t209_n241: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t209_n243;
  t209_n242: return 0.07677490264177322f;
  t209_n243: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t209_n247;
  t209_n244: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t209_n246;
  t209_n245: return 0.14356133341789246f;
  t209_n246: return 0.10039757192134857f;
  t209_n247: return 0.13467256724834442f;
  t209_n248: return -0.04400177672505379f;
  t209_n249: return 0.027996502816677094f;
  t209_n250: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t209_n252;
  t209_n251: return 0.0066510653123259544f;
  t209_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t209_n254;
  t209_n253: return 0.042243070900440216f;
  t209_n254: return 0.016761943697929382f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree210(const Example& E) {
  t210_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t210_n168;
  t210_n1: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t210_n163;
  t210_n2: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t210_n160;
  t210_n3: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t210_n17;
  t210_n4: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t210_n14;
  t210_n5: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t210_n13;
  t210_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t210_n10;
  t210_n7: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t210_n9;
  t210_n8: return -0.000916011689696461f;
  t210_n9: return 0.005012028850615025f;
  t210_n10: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t210_n12;
  t210_n11: return 0.0017189803766086698f;
  t210_n12: return 0.01600510999560356f;
  t210_n13: return -0.00555049953982234f;
  t210_n14: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t210_n16;
  t210_n15: return 0.005004150792956352f;
  t210_n16: return 0.08188340067863464f;
  t210_n17: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t210_n51;
  t210_n18: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t210_n48;
  t210_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n21;
  t210_n20: return -0.020449763163924217f;
  t210_n21: if (E.getNumReferences() >= 3295404032 /*943.5*/) goto t210_n41;
  t210_n22: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t210_n30;
  t210_n23: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t210_n25;
  t210_n24: return -0.009163454174995422f;
  t210_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t210_n27;
  t210_n26: return 0.02041831612586975f;
  t210_n27: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t210_n29;
  t210_n28: return 0.3106181025505066f;
  t210_n29: return 0.6586479544639587f;
  t210_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t210_n32;
  t210_n31: return -0.001678325468674302f;
  t210_n32: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t210_n34;
  t210_n33: return 0.03545185551047325f;
  t210_n34: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t210_n40;
  t210_n35: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n39;
  t210_n36: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t210_n38;
  t210_n37: return 0.057631660252809525f;
  t210_n38: return 0.09941305965185165f;
  t210_n39: return -0.03198067843914032f;
  t210_n40: return 0.019075820222496986f;
  t210_n41: if (E.getNumReferences() >= 3295584256 /*954.5*/) goto t210_n47;
  t210_n42: if (E.getNumReferences() >= 3295551488 /*952.5*/) goto t210_n46;
  t210_n43: if (E.getNumReferences() >= 3295477760 /*948.0*/) goto t210_n45;
  t210_n44: return 0.10147099941968918f;
  t210_n45: return -0.1390133798122406f;
  t210_n46: return 0.18322046101093292f;
  t210_n47: return 0.01548346970230341f;
  t210_n48: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t210_n50;
  t210_n49: return 0.034550610929727554f;
  t210_n50: return 0.12304327636957169f;
  t210_n51: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t210_n61;
  t210_n52: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t210_n54;
  t210_n53: return -0.0002716723829507828f;
  t210_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n56;
  t210_n55: return -0.03900943696498871f;
  t210_n56: if (E.getNumReferences() >= 3294396416 /*882.0*/) goto t210_n58;
  t210_n57: return 0.07980038225650787f;
  t210_n58: if (E.getNumReferences() >= 3306844160 /*2471.0*/) goto t210_n60;
  t210_n59: return 0.13559606671333313f;
  t210_n60: return 0.06913898140192032f;
  t210_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n63;
  t210_n62: return -0.002612990327179432f;
  t210_n63: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t210_n157;
  t210_n64: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t210_n156;
  t210_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t210_n139;
  t210_n66: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t210_n136;
  t210_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n85;
  t210_n68: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t210_n80;
  t210_n69: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n73;
  t210_n70: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t210_n72;
  t210_n71: return 0.005093300249427557f;
  t210_n72: return 0.0666014701128006f;
  t210_n73: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t210_n77;
  t210_n74: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t210_n76;
  t210_n75: return -0.03739553689956665f;
  t210_n76: return -0.18595875799655914f;
  t210_n77: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t210_n79;
  t210_n78: return 0.05262508988380432f;
  t210_n79: return -0.22859834134578705f;
  t210_n80: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t210_n82;
  t210_n81: return -0.01248097326606512f;
  t210_n82: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t210_n84;
  t210_n83: return -0.09652026742696762f;
  t210_n84: return 0.09599999338388443f;
  t210_n85: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t210_n115;
  t210_n86: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t210_n114;
  t210_n87: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t210_n103;
  t210_n88: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t210_n102;
  t210_n89: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t210_n97;
  t210_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n92;
  t210_n91: return -0.014796852134168148f;
  t210_n92: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t210_n94;
  t210_n93: return -0.001314240973442793f;
  t210_n94: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t210_n96;
  t210_n95: return 0.08872316777706146f;
  t210_n96: return 0.036691825836896896f;
  t210_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t210_n101;
  t210_n98: if (E.getFractionNameInContext() >= 3210972729 /*0.8888888955116272*/) goto t210_n100;
  t210_n99: return 0.02107827551662922f;
  t210_n100: return -0.46564728021621704f;
  t210_n101: return 0.045815609395504f;
  t210_n102: return -0.0195343978703022f;
  t210_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t210_n105;
  t210_n104: return -0.10478688776493073f;
  t210_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n107;
  t210_n106: return -0.007087789010256529f;
  t210_n107: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t210_n113;
  t210_n108: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n110;
  t210_n109: return 0.030171891674399376f;
  t210_n110: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t210_n112;
  t210_n111: return -0.154652938246727f;
  t210_n112: return 0.0205735694617033f;
  t210_n113: return 0.10365908592939377f;
  t210_n114: return -0.03249618038535118f;
  t210_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t210_n117;
  t210_n116: return -0.10515747219324112f;
  t210_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t210_n133;
  t210_n118: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t210_n132;
  t210_n119: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t210_n131;
  t210_n120: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t210_n128;
  t210_n121: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t210_n127;
  t210_n122: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t210_n124;
  t210_n123: return 0.02539801225066185f;
  t210_n124: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t210_n126;
  t210_n125: return 0.12593522667884827f;
  t210_n126: return 0.06133405491709709f;
  t210_n127: return -0.04831327125430107f;
  t210_n128: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t210_n130;
  t210_n129: return 0.03565562888979912f;
  t210_n130: return 0.08037745952606201f;
  t210_n131: return -0.06434717029333115f;
  t210_n132: return 0.04556003585457802f;
  t210_n133: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n135;
  t210_n134: return 0.03434860706329346f;
  t210_n135: return -0.16217255592346191f;
  t210_n136: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t210_n138;
  t210_n137: return -0.022914614528417587f;
  t210_n138: return 0.1367885023355484f;
  t210_n139: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t210_n147;
  t210_n140: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n142;
  t210_n141: return 0.01821235939860344f;
  t210_n142: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t210_n144;
  t210_n143: return 0.020040661096572876f;
  t210_n144: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t210_n146;
  t210_n145: return 0.06684665381908417f;
  t210_n146: return -0.062694251537323f;
  t210_n147: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t210_n151;
  t210_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n150;
  t210_n149: return -0.03799315169453621f;
  t210_n150: return 0.03334979712963104f;
  t210_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n153;
  t210_n152: return -0.00822021160274744f;
  t210_n153: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t210_n155;
  t210_n154: return 0.05974849686026573f;
  t210_n155: return 0.01998378522694111f;
  t210_n156: return -0.004695728421211243f;
  t210_n157: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t210_n159;
  t210_n158: return 0.06719866394996643f;
  t210_n159: return 0.015306545421481133f;
  t210_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n162;
  t210_n161: return -0.0318160243332386f;
  t210_n162: return 0.03732950612902641f;
  t210_n163: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t210_n165;
  t210_n164: return -0.07361036539077759f;
  t210_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t210_n167;
  t210_n166: return -0.19409850239753723f;
  t210_n167: return -0.4698115885257721f;
  t210_n168: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t210_n208;
  t210_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t210_n207;
  t210_n170: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t210_n174;
  t210_n171: if (E.getNumReferences() >= 3347831872 /*71608.5*/) goto t210_n173;
  t210_n172: return -0.04719759151339531f;
  t210_n173: return 0.1824890375137329f;
  t210_n174: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t210_n202;
  t210_n175: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t210_n199;
  t210_n176: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t210_n198;
  t210_n177: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t210_n185;
  t210_n178: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t210_n182;
  t210_n179: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t210_n181;
  t210_n180: return -0.03113177791237831f;
  t210_n181: return 0.04495181143283844f;
  t210_n182: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t210_n184;
  t210_n183: return 0.06397317349910736f;
  t210_n184: return -0.036018241196870804f;
  t210_n185: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t210_n187;
  t210_n186: return 0.03474026545882225f;
  t210_n187: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t210_n191;
  t210_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t210_n190;
  t210_n189: return -0.10283182561397552f;
  t210_n190: return 0.008108402602374554f;
  t210_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t210_n195;
  t210_n192: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n194;
  t210_n193: return 0.0038376119919121265f;
  t210_n194: return 0.09786312282085419f;
  t210_n195: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t210_n197;
  t210_n196: return 0.09175543487071991f;
  t210_n197: return -0.01976037584245205f;
  t210_n198: return 0.027053387835621834f;
  t210_n199: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t210_n201;
  t210_n200: return -0.12587980926036835f;
  t210_n201: return 0.0057458942756056786f;
  t210_n202: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t210_n206;
  t210_n203: if (E.getNumReferences() >= 3305873408 /*2234.0*/) goto t210_n205;
  t210_n204: return -0.01777484081685543f;
  t210_n205: return 0.02301556058228016f;
  t210_n206: return 0.004539038520306349f;
  t210_n207: return 0.0009566816152073443f;
  t210_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t210_n210;
  t210_n209: return 0.0007693638908676803f;
  t210_n210: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t210_n252;
  t210_n211: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t210_n223;
  t210_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t210_n222;
  t210_n213: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t210_n217;
  t210_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t210_n216;
  t210_n215: return 0.005069504026323557f;
  t210_n216: return 0.01374448649585247f;
  t210_n217: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t210_n221;
  t210_n218: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t210_n220;
  t210_n219: return -0.11099335551261902f;
  t210_n220: return -0.02309909090399742f;
  t210_n221: return 0.06824678182601929f;
  t210_n222: return 0.06050966680049896f;
  t210_n223: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t210_n243;
  t210_n224: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t210_n242;
  t210_n225: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t210_n239;
  t210_n226: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t210_n238;
  t210_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t210_n233;
  t210_n228: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t210_n230;
  t210_n229: return 0.038685645908117294f;
  t210_n230: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t210_n232;
  t210_n231: return -0.18921582400798798f;
  t210_n232: return 0.003520180704072118f;
  t210_n233: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t210_n235;
  t210_n234: return -0.16894470155239105f;
  t210_n235: if (E.getFractionNameInContext() >= 3200482732 /*0.38181817531585693*/) goto t210_n237;
  t210_n236: return 0.09410201013088226f;
  t210_n237: return 0.055116210132837296f;
  t210_n238: return -0.0722692459821701f;
  t210_n239: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t210_n241;
  t210_n240: return 0.11903754621744156f;
  t210_n241: return -0.3117418885231018f;
  t210_n242: return -0.31188902258872986f;
  t210_n243: if (E.getFractionNameInContext() >= 3206096018 /*0.5982142686843872*/) goto t210_n251;
  t210_n244: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t210_n250;
  t210_n245: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t210_n249;
  t210_n246: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t210_n248;
  t210_n247: return 0.08150483667850494f;
  t210_n248: return -0.12299036234617233f;
  t210_n249: return 0.12589839100837708f;
  t210_n250: return 0.0682876780629158f;
  t210_n251: return -0.07421154528856277f;
  t210_n252: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t210_n254;
  t210_n253: return -0.05306994169950485f;
  t210_n254: return -0.19097135961055756f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree211(const Example& E) {
  t211_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t211_n174;
  t211_n1: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t211_n163;
  t211_n2: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t211_n56;
  t211_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t211_n55;
  t211_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t211_n54;
  t211_n5: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t211_n7;
  t211_n6: return -0.0026514455676078796f;
  t211_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t211_n9;
  t211_n8: return -0.0015645847888663411f;
  t211_n9: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t211_n41;
  t211_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t211_n12;
  t211_n11: return 0.006505582481622696f;
  t211_n12: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n28;
  t211_n13: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t211_n23;
  t211_n14: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t211_n16;
  t211_n15: return 0.013871636241674423f;
  t211_n16: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t211_n18;
  t211_n17: return 0.18539486825466156f;
  t211_n18: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t211_n22;
  t211_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t211_n21;
  t211_n20: return 0.05699494108557701f;
  t211_n21: return 0.09006315469741821f;
  t211_n22: return 0.033524688333272934f;
  t211_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t211_n27;
  t211_n24: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t211_n26;
  t211_n25: return -0.1255345344543457f;
  t211_n26: return 0.1168825700879097f;
  t211_n27: return 0.10845571011304855f;
  t211_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t211_n30;
  t211_n29: return 0.017934713512659073f;
  t211_n30: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t211_n40;
  t211_n31: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t211_n37;
  t211_n32: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t211_n34;
  t211_n33: return 0.04632994905114174f;
  t211_n34: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t211_n36;
  t211_n35: return -0.0357479453086853f;
  t211_n36: return 0.07113759219646454f;
  t211_n37: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t211_n39;
  t211_n38: return 0.05914769321680069f;
  t211_n39: return -0.09150722622871399f;
  t211_n40: return -0.1697850227355957f;
  t211_n41: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t211_n53;
  t211_n42: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t211_n52;
  t211_n43: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t211_n51;
  t211_n44: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t211_n46;
  t211_n45: return 0.0056886812672019005f;
  t211_n46: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t211_n50;
  t211_n47: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n49;
  t211_n48: return 0.051461149007081985f;
  t211_n49: return 0.24238087236881256f;
  t211_n50: return -0.008147562853991985f;
  t211_n51: return -0.16398277878761292f;
  t211_n52: return 0.024650702252984047f;
  t211_n53: return -0.06444449722766876f;
  t211_n54: return -0.02623855322599411f;
  t211_n55: return 0.00034445500932633877f;
  t211_n56: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t211_n70;
  t211_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t211_n69;
  t211_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n66;
  t211_n59: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t211_n61;
  t211_n60: return 0.000395089213270694f;
  t211_n61: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t211_n63;
  t211_n62: return 0.017098117619752884f;
  t211_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t211_n65;
  t211_n64: return -0.048393070697784424f;
  t211_n65: return 0.01223739329725504f;
  t211_n66: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t211_n68;
  t211_n67: return -0.017379572615027428f;
  t211_n68: return 0.010896947234869003f;
  t211_n69: return 0.01687496155500412f;
  t211_n70: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t211_n106;
  t211_n71: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t211_n103;
  t211_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t211_n74;
  t211_n73: return -0.04163951054215431f;
  t211_n74: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t211_n78;
  t211_n75: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t211_n77;
  t211_n76: return 0.005907642189413309f;
  t211_n77: return -0.07970612496137619f;
  t211_n78: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t211_n80;
  t211_n79: return -0.09785770624876022f;
  t211_n80: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t211_n96;
  t211_n81: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t211_n93;
  t211_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t211_n92;
  t211_n83: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t211_n87;
  t211_n84: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t211_n86;
  t211_n85: return 0.043826013803482056f;
  t211_n86: return -0.13931049406528473f;
  t211_n87: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t211_n89;
  t211_n88: return 0.05780541151762009f;
  t211_n89: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n91;
  t211_n90: return 0.1626531034708023f;
  t211_n91: return 0.07950972765684128f;
  t211_n92: return 0.12483696639537811f;
  t211_n93: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t211_n95;
  t211_n94: return -0.07576968520879745f;
  t211_n95: return 0.08766436576843262f;
  t211_n96: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t211_n98;
  t211_n97: return -0.09576379507780075f;
  t211_n98: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t211_n100;
  t211_n99: return -0.015414971858263016f;
  t211_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t211_n102;
  t211_n101: return 0.05447147786617279f;
  t211_n102: return 0.11081758886575699f;
  t211_n103: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t211_n105;
  t211_n104: return 0.019410137087106705f;
  t211_n105: return 0.08993998914957047f;
  t211_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t211_n144;
  t211_n107: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t211_n109;
  t211_n108: return -0.008316421881318092f;
  t211_n109: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t211_n117;
  t211_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t211_n112;
  t211_n111: return -0.017143461853265762f;
  t211_n112: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t211_n116;
  t211_n113: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t211_n115;
  t211_n114: return 0.12097529321908951f;
  t211_n115: return 0.1356637179851532f;
  t211_n116: return 0.1433713734149933f;
  t211_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t211_n119;
  t211_n118: return -0.05705644562840462f;
  t211_n119: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t211_n137;
  t211_n120: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t211_n136;
  t211_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t211_n129;
  t211_n122: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t211_n124;
  t211_n123: return 0.07268539816141129f;
  t211_n124: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t211_n126;
  t211_n125: return -0.10087589919567108f;
  t211_n126: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t211_n128;
  t211_n127: return 0.12809839844703674f;
  t211_n128: return 0.12529560923576355f;
  t211_n129: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t211_n135;
  t211_n130: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t211_n134;
  t211_n131: if (E.getNumReferences() >= 3235905536 /*7.0*/) goto t211_n133;
  t211_n132: return 0.10978518426418304f;
  t211_n133: return 0.30623140931129456f;
  t211_n134: return -0.03627915307879448f;
  t211_n135: return -0.0966123566031456f;
  t211_n136: return 0.1233164593577385f;
  t211_n137: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t211_n143;
  t211_n138: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t211_n140;
  t211_n139: return 0.004007006995379925f;
  t211_n140: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t211_n142;
  t211_n141: return -0.7990759611129761f;
  t211_n142: return -0.042494118213653564f;
  t211_n143: return 0.0956818088889122f;
  t211_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n162;
  t211_n145: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t211_n157;
  t211_n146: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t211_n150;
  t211_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t211_n149;
  t211_n148: return -0.0061476221308112144f;
  t211_n149: return 0.011122805997729301f;
  t211_n150: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t211_n156;
  t211_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t211_n153;
  t211_n152: return 0.07146166265010834f;
  t211_n153: if (E.getNumReferences() >= 3233808384 /*6.0*/) goto t211_n155;
  t211_n154: return 0.09714923053979874f;
  t211_n155: return 0.1191338300704956f;
  t211_n156: return 0.026684341952204704f;
  t211_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t211_n159;
  t211_n158: return -0.11859222501516342f;
  t211_n159: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t211_n161;
  t211_n160: return 0.001931900973431766f;
  t211_n161: return -0.04852517321705818f;
  t211_n162: return -0.012290406972169876f;
  t211_n163: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t211_n173;
  t211_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t211_n172;
  t211_n165: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t211_n167;
  t211_n166: return -0.007542268838733435f;
  t211_n167: if (E.getNumReferences() >= 3314966528 /*4812.0*/) goto t211_n171;
  t211_n168: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t211_n170;
  t211_n169: return -0.06275684386491776f;
  t211_n170: return 0.06201707944273949f;
  t211_n171: return -1.7572919130325317f;
  t211_n172: return 0.005849302746355534f;
  t211_n173: return 0.0004489492566790432f;
  t211_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t211_n228;
  t211_n175: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t211_n219;
  t211_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t211_n218;
  t211_n177: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t211_n179;
  t211_n178: return 0.0024254960007965565f;
  t211_n179: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t211_n199;
  t211_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t211_n192;
  t211_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t211_n189;
  t211_n182: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t211_n186;
  t211_n183: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t211_n185;
  t211_n184: return -0.14511673152446747f;
  t211_n185: return 0.10421282798051834f;
  t211_n186: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t211_n188;
  t211_n187: return -0.13700830936431885f;
  t211_n188: return -0.022749921306967735f;
  t211_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t211_n191;
  t211_n190: return -0.1474505513906479f;
  t211_n191: return 0.005949998740106821f;
  t211_n192: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t211_n196;
  t211_n193: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t211_n195;
  t211_n194: return -0.0023458204232156277f;
  t211_n195: return 0.09664233773946762f;
  t211_n196: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t211_n198;
  t211_n197: return 0.03245511278510094f;
  t211_n198: return -0.043490976095199585f;
  t211_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t211_n209;
  t211_n200: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t211_n202;
  t211_n201: return -0.15886029601097107f;
  t211_n202: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t211_n206;
  t211_n203: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t211_n205;
  t211_n204: return 0.10017135739326477f;
  t211_n205: return 0.059805311262607574f;
  t211_n206: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t211_n208;
  t211_n207: return 0.02006513439118862f;
  t211_n208: return -0.341708779335022f;
  t211_n209: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t211_n215;
  t211_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t211_n214;
  t211_n211: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t211_n213;
  t211_n212: return -0.06148221716284752f;
  t211_n213: return 0.10395807772874832f;
  t211_n214: return 0.062369950115680695f;
  t211_n215: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t211_n217;
  t211_n216: return 0.11881951987743378f;
  t211_n217: return 0.08275974541902542f;
  t211_n218: return 0.0274175014346838f;
  t211_n219: if (E.getFractionNameInContext() >= 3200163166 /*0.3722943663597107*/) goto t211_n221;
  t211_n220: return -0.1175939291715622f;
  t211_n221: if (E.getFractionNameInContext() >= 3203710796 /*0.4780219793319702*/) goto t211_n223;
  t211_n222: return 0.10543958842754364f;
  t211_n223: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t211_n225;
  t211_n224: return 0.00947555247694254f;
  t211_n225: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t211_n227;
  t211_n226: return 0.11846151202917099f;
  t211_n227: return -0.12584461271762848f;
  t211_n228: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t211_n240;
  t211_n229: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t211_n231;
  t211_n230: return -0.16027575731277466f;
  t211_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t211_n237;
  t211_n232: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t211_n234;
  t211_n233: return -0.06593437492847443f;
  t211_n234: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t211_n236;
  t211_n235: return 0.006286268122494221f;
  t211_n236: return 0.09147374331951141f;
  t211_n237: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t211_n239;
  t211_n238: return -0.030954455956816673f;
  t211_n239: return 0.126866415143013f;
  t211_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t211_n242;
  t211_n241: return 0.005768056493252516f;
  t211_n242: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t211_n254;
  t211_n243: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t211_n253;
  t211_n244: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t211_n246;
  t211_n245: return -0.6405038237571716f;
  t211_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t211_n250;
  t211_n247: if (E.getFileProximityDistanceCost() >= 3249012736 /*21.0*/) goto t211_n249;
  t211_n248: return 0.02224036119878292f;
  t211_n249: return 0.10878775268793106f;
  t211_n250: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t211_n252;
  t211_n251: return 0.10723060369491577f;
  t211_n252: return 0.07018791139125824f;
  t211_n253: return 0.11173947900533676f;
  t211_n254: return 0.016069352626800537f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree212(const Example& E) {
  t212_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t212_n2;
  t212_n1: return -0.0004561660753097385f;
  t212_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t212_n254;
  t212_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t212_n91;
  t212_n4: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t212_n66;
  t212_n5: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t212_n9;
  t212_n6: if (E.getFractionNameInContext() >= 3197638680 /*0.2970588207244873*/) goto t212_n8;
  t212_n7: return -1.1286030712653883e-05f;
  t212_n8: return 0.014478473924100399f;
  t212_n9: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t212_n17;
  t212_n10: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t212_n16;
  t212_n11: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t212_n15;
  t212_n12: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t212_n14;
  t212_n13: return -0.16192422807216644f;
  t212_n14: return 0.16685901582241058f;
  t212_n15: return 0.1447104662656784f;
  t212_n16: return -0.11704318970441818f;
  t212_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t212_n65;
  t212_n18: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t212_n64;
  t212_n19: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t212_n29;
  t212_n20: if (E.getNumReferences() >= 3280715776 /*279.5*/) goto t212_n26;
  t212_n21: if (E.getNumReferences() >= 3267887104 /*100.0*/) goto t212_n23;
  t212_n22: return 0.011804426088929176f;
  t212_n23: if (E.getNumReferences() >= 3270836224 /*122.5*/) goto t212_n25;
  t212_n24: return -0.4456690847873688f;
  t212_n25: return -0.050303418189287186f;
  t212_n26: if (E.getNumReferences() >= 3282108416 /*322.0*/) goto t212_n28;
  t212_n27: return 0.12738628685474396f;
  t212_n28: return 0.08001777529716492f;
  t212_n29: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t212_n57;
  t212_n30: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t212_n54;
  t212_n31: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t212_n53;
  t212_n32: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t212_n48;
  t212_n33: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t212_n47;
  t212_n34: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t212_n36;
  t212_n35: return -0.12095711380243301f;
  t212_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t212_n42;
  t212_n37: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t212_n39;
  t212_n38: return -0.10021282732486725f;
  t212_n39: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t212_n41;
  t212_n40: return -0.1248604878783226f;
  t212_n41: return 0.10635976493358612f;
  t212_n42: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t212_n44;
  t212_n43: return -0.23103654384613037f;
  t212_n44: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t212_n46;
  t212_n45: return 0.14646023511886597f;
  t212_n46: return -0.10933471471071243f;
  t212_n47: return 0.08606245368719101f;
  t212_n48: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t212_n52;
  t212_n49: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t212_n51;
  t212_n50: return 0.09400618821382523f;
  t212_n51: return 0.12731434404850006f;
  t212_n52: return 0.005376963876187801f;
  t212_n53: return -0.04010014608502388f;
  t212_n54: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t212_n56;
  t212_n55: return -0.015902671962976456f;
  t212_n56: return -0.13747663795948029f;
  t212_n57: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t212_n63;
  t212_n58: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t212_n60;
  t212_n59: return -0.5150572657585144f;
  t212_n60: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t212_n62;
  t212_n61: return -0.24018049240112305f;
  t212_n62: return -0.005583953578025103f;
  t212_n63: return 0.08578184247016907f;
  t212_n64: return -0.004769188817590475f;
  t212_n65: return -0.0007932678563520312f;
  t212_n66: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t212_n78;
  t212_n67: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t212_n75;
  t212_n68: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t212_n70;
  t212_n69: return -0.09243419021368027f;
  t212_n70: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t212_n74;
  t212_n71: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t212_n73;
  t212_n72: return -0.09725379198789597f;
  t212_n73: return 0.12257714569568634f;
  t212_n74: return -0.017709804698824883f;
  t212_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t212_n77;
  t212_n76: return -0.0014166005421429873f;
  t212_n77: return 0.2358483523130417f;
  t212_n78: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t212_n90;
  t212_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t212_n85;
  t212_n80: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t212_n84;
  t212_n81: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t212_n83;
  t212_n82: return 0.05659661442041397f;
  t212_n83: return -0.05640764907002449f;
  t212_n84: return 0.006539280992001295f;
  t212_n85: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t212_n89;
  t212_n86: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t212_n88;
  t212_n87: return 0.2206984907388687f;
  t212_n88: return 0.13676293194293976f;
  t212_n89: return -0.005928004160523415f;
  t212_n90: return -0.022959206253290176f;
  t212_n91: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t212_n235;
  t212_n92: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t212_n104;
  t212_n93: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t212_n95;
  t212_n94: return -0.019721388816833496f;
  t212_n95: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t212_n103;
  t212_n96: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t212_n100;
  t212_n97: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t212_n99;
  t212_n98: return 0.04047465696930885f;
  t212_n99: return -0.2132037878036499f;
  t212_n100: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t212_n102;
  t212_n101: return 0.05889340862631798f;
  t212_n102: return 0.12137919664382935f;
  t212_n103: return -0.07627254724502563f;
  t212_n104: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t212_n150;
  t212_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t212_n113;
  t212_n106: if (E.getNumReferences() >= 3338093824 /*31650.5*/) goto t212_n110;
  t212_n107: if (E.getNumNameInContext() >= 3238526976 /*8.5*/) goto t212_n109;
  t212_n108: return -0.0032268427312374115f;
  t212_n109: return 0.10989920049905777f;
  t212_n110: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t212_n112;
  t212_n111: return -0.11412588506937027f;
  t212_n112: return -0.8027863502502441f;
  t212_n113: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t212_n115;
  t212_n114: return -0.4186485707759857f;
  t212_n115: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t212_n129;
  t212_n116: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t212_n118;
  t212_n117: return 0.004516717977821827f;
  t212_n118: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t212_n122;
  t212_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t212_n121;
  t212_n120: return -0.07668623328208923f;
  t212_n121: return -0.015611646696925163f;
  t212_n122: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t212_n124;
  t212_n123: return 0.002023829845711589f;
  t212_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t212_n126;
  t212_n125: return -0.04002372920513153f;
  t212_n126: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t212_n128;
  t212_n127: return -0.10109812766313553f;
  t212_n128: return 0.14747647941112518f;
  t212_n129: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t212_n149;
  t212_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t212_n148;
  t212_n131: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t212_n139;
  t212_n132: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t212_n134;
  t212_n133: return 0.08569202572107315f;
  t212_n134: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t212_n136;
  t212_n135: return -0.21437352895736694f;
  t212_n136: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t212_n138;
  t212_n137: return -0.4895264804363251f;
  t212_n138: return -0.15488660335540771f;
  t212_n139: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t212_n147;
  t212_n140: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t212_n146;
  t212_n141: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t212_n143;
  t212_n142: return -0.017575932666659355f;
  t212_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t212_n145;
  t212_n144: return -0.18098771572113037f;
  t212_n145: return -0.023346323519945145f;
  t212_n146: return 0.00395863875746727f;
  t212_n147: return 0.02225196361541748f;
  t212_n148: return 0.005144220311194658f;
  t212_n149: return 0.012463482096791267f;
  t212_n150: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t212_n154;
  t212_n151: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t212_n153;
  t212_n152: return -0.07169853150844574f;
  t212_n153: return 0.11707065999507904f;
  t212_n154: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t212_n204;
  t212_n155: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t212_n173;
  t212_n156: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t212_n166;
  t212_n157: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t212_n161;
  t212_n158: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t212_n160;
  t212_n159: return 0.13018269836902618f;
  t212_n160: return -0.06198849529027939f;
  t212_n161: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t212_n163;
  t212_n162: return 0.10854696482419968f;
  t212_n163: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t212_n165;
  t212_n164: return 0.02838829904794693f;
  t212_n165: return 0.14949187636375427f;
  t212_n166: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t212_n168;
  t212_n167: return -0.014298654161393642f;
  t212_n168: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t212_n170;
  t212_n169: return -0.4372553825378418f;
  t212_n170: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t212_n172;
  t212_n171: return -0.01868933066725731f;
  t212_n172: return -0.14853426814079285f;
  t212_n173: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t212_n201;
  t212_n174: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t212_n198;
  t212_n175: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t212_n187;
  t212_n176: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t212_n182;
  t212_n177: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t212_n181;
  t212_n178: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t212_n180;
  t212_n179: return -0.04900076985359192f;
  t212_n180: return 0.08784998953342438f;
  t212_n181: return 0.014474408701062202f;
  t212_n182: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t212_n184;
  t212_n183: return 0.008142282254993916f;
  t212_n184: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t212_n186;
  t212_n185: return 0.05406845360994339f;
  t212_n186: return 0.01718827523291111f;
  t212_n187: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t212_n193;
  t212_n188: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t212_n190;
  t212_n189: return 0.010603108443319798f;
  t212_n190: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t212_n192;
  t212_n191: return 0.05266236513853073f;
  t212_n192: return 0.08563650399446487f;
  t212_n193: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t212_n197;
  t212_n194: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t212_n196;
  t212_n195: return 0.09823383390903473f;
  t212_n196: return -0.052840519696474075f;
  t212_n197: return 0.03787824511528015f;
  t212_n198: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t212_n200;
  t212_n199: return -0.034212980419397354f;
  t212_n200: return 0.010959500446915627f;
  t212_n201: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t212_n203;
  t212_n202: return -0.18994572758674622f;
  t212_n203: return 0.056313976645469666f;
  t212_n204: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t212_n216;
  t212_n205: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t212_n213;
  t212_n206: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t212_n212;
  t212_n207: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t212_n209;
  t212_n208: return 0.06650491803884506f;
  t212_n209: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t212_n211;
  t212_n210: return 0.13762690126895905f;
  t212_n211: return 0.09643833339214325f;
  t212_n212: return -0.035833511501550674f;
  t212_n213: if (E.getFractionNameInContext() >= 3199175417 /*0.34285715222358704*/) goto t212_n215;
  t212_n214: return 0.0019977441988885403f;
  t212_n215: return 0.08122458308935165f;
  t212_n216: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t212_n218;
  t212_n217: return -0.35606396198272705f;
  t212_n218: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t212_n232;
  t212_n219: if (E.getNumReferences() >= 3273359360 /*155.5*/) goto t212_n229;
  t212_n220: if (E.getNumReferences() >= 3269132288 /*109.5*/) goto t212_n228;
  t212_n221: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t212_n223;
  t212_n222: return 0.09427805989980698f;
  t212_n223: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t212_n225;
  t212_n224: return -0.07802461832761765f;
  t212_n225: if (E.getNumReferences() >= 3255304192 /*34.0*/) goto t212_n227;
  t212_n226: return 0.13345789909362793f;
  t212_n227: return -0.03215835615992546f;
  t212_n228: return -0.32620036602020264f;
  t212_n229: if (E.getNumReferences() >= 3274735616 /*176.5*/) goto t212_n231;
  t212_n230: return 0.17697270214557648f;
  t212_n231: return 0.03469441458582878f;
  t212_n232: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t212_n234;
  t212_n233: return -0.025019844993948936f;
  t212_n234: return 0.11210723966360092f;
  t212_n235: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t212_n245;
  t212_n236: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t212_n240;
  t212_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t212_n239;
  t212_n238: return 0.08090243488550186f;
  t212_n239: return 0.11714442074298859f;
  t212_n240: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t212_n242;
  t212_n241: return -0.09477794915437698f;
  t212_n242: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t212_n244;
  t212_n243: return 0.004839787259697914f;
  t212_n244: return -0.12014688551425934f;
  t212_n245: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t212_n249;
  t212_n246: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t212_n248;
  t212_n247: return 0.014972803182899952f;
  t212_n248: return 0.07393289357423782f;
  t212_n249: if (E.getFractionNameInContext() >= 3180529969 /*0.07179487496614456*/) goto t212_n251;
  t212_n250: return 0.12542009353637695f;
  t212_n251: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t212_n253;
  t212_n252: return 0.09761406481266022f;
  t212_n253: return 0.0036352071911096573f;
  t212_n254: return 0.0023936734069138765f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree213(const Example& E) {
  t213_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t213_n2;
  t213_n1: return -0.09406394511461258f;
  t213_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t213_n236;
  t213_n3: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t213_n235;
  t213_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t213_n34;
  t213_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t213_n7;
  t213_n6: return -0.09124784916639328f;
  t213_n7: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t213_n9;
  t213_n8: return 0.0005870084860362113f;
  t213_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t213_n33;
  t213_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t213_n24;
  t213_n11: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t213_n15;
  t213_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t213_n14;
  t213_n13: return -0.11171837151050568f;
  t213_n14: return -0.00803742092102766f;
  t213_n15: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t213_n17;
  t213_n16: return 0.01722889207303524f;
  t213_n17: if (E.getNumReferences() >= 3264282624 /*72.5*/) goto t213_n21;
  t213_n18: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t213_n20;
  t213_n19: return 0.061176691204309464f;
  t213_n20: return 0.09126845002174377f;
  t213_n21: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t213_n23;
  t213_n22: return -0.025112763047218323f;
  t213_n23: return -0.2179282009601593f;
  t213_n24: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t213_n26;
  t213_n25: return 0.0014296574518084526f;
  t213_n26: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t213_n28;
  t213_n27: return -0.136802539229393f;
  t213_n28: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t213_n32;
  t213_n29: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t213_n31;
  t213_n30: return 0.09684329479932785f;
  t213_n31: return -0.004827295895665884f;
  t213_n32: return 0.134978249669075f;
  t213_n33: return -0.016633305698633194f;
  t213_n34: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t213_n234;
  t213_n35: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t213_n171;
  t213_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t213_n158;
  t213_n37: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t213_n115;
  t213_n38: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t213_n112;
  t213_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t213_n89;
  t213_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Expression))) goto t213_n78;
  t213_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t213_n45;
  t213_n42: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t213_n44;
  t213_n43: return -0.09843675792217255f;
  t213_n44: return 0.18030481040477753f;
  t213_n45: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t213_n77;
  t213_n46: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t213_n54;
  t213_n47: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t213_n49;
  t213_n48: return -0.000348642555763945f;
  t213_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t213_n53;
  t213_n50: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t213_n52;
  t213_n51: return -0.17122940719127655f;
  t213_n52: return -0.10416953265666962f;
  t213_n53: return -0.04774262756109238f;
  t213_n54: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t213_n70;
  t213_n55: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t213_n67;
  t213_n56: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t213_n58;
  t213_n57: return 0.0011509853648021817f;
  t213_n58: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t213_n60;
  t213_n59: return 0.11771223694086075f;
  t213_n60: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t213_n64;
  t213_n61: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t213_n63;
  t213_n62: return -0.20364828407764435f;
  t213_n63: return 0.08494862914085388f;
  t213_n64: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t213_n66;
  t213_n65: return 0.09057392179965973f;
  t213_n66: return -0.01910940371453762f;
  t213_n67: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t213_n69;
  t213_n68: return 0.01778516359627247f;
  t213_n69: return -0.12793231010437012f;
  t213_n70: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t213_n74;
  t213_n71: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t213_n73;
  t213_n72: return -0.0061818077228963375f;
  t213_n73: return 0.07349462062120438f;
  t213_n74: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t213_n76;
  t213_n75: return 0.10136879235506058f;
  t213_n76: return -0.03753180429339409f;
  t213_n77: return 0.14971162378787994f;
  t213_n78: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t213_n88;
  t213_n79: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t213_n81;
  t213_n80: return -0.05467453971505165f;
  t213_n81: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t213_n85;
  t213_n82: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t213_n84;
  t213_n83: return 0.09808187931776047f;
  t213_n84: return 0.015458962880074978f;
  t213_n85: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t213_n87;
  t213_n86: return -0.20295628905296326f;
  t213_n87: return 0.10218507051467896f;
  t213_n88: return 0.010162760503590107f;
  t213_n89: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t213_n91;
  t213_n90: return -0.011328726075589657f;
  t213_n91: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t213_n93;
  t213_n92: return 0.004019299056380987f;
  t213_n93: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t213_n95;
  t213_n94: return 0.029982034116983414f;
  t213_n95: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t213_n111;
  t213_n96: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t213_n102;
  t213_n97: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t213_n99;
  t213_n98: return 0.009425912983715534f;
  t213_n99: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t213_n101;
  t213_n100: return 0.13605773448944092f;
  t213_n101: return 0.06350620090961456f;
  t213_n102: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t213_n108;
  t213_n103: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t213_n107;
  t213_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t213_n106;
  t213_n105: return -0.6300734877586365f;
  t213_n106: return -0.16842664778232574f;
  t213_n107: return -0.12327428162097931f;
  t213_n108: if (E.getFractionNameInContext() >= 3188638956 /*0.13942307233810425*/) goto t213_n110;
  t213_n109: return 0.1265980303287506f;
  t213_n110: return -0.05240829661488533f;
  t213_n111: return -3.192453145980835f;
  t213_n112: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t213_n114;
  t213_n113: return 0.00802908930927515f;
  t213_n114: return 0.029213888570666313f;
  t213_n115: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t213_n149;
  t213_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t213_n148;
  t213_n117: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t213_n121;
  t213_n118: if (E.getNumReferences() >= 3280060416 /*259.5*/) goto t213_n120;
  t213_n119: return -0.006630588788539171f;
  t213_n120: return -0.20367278158664703f;
  t213_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t213_n141;
  t213_n122: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t213_n130;
  t213_n123: if (E.getNumReferences() >= 3278077952 /*227.5*/) goto t213_n125;
  t213_n124: return 0.02190626785159111f;
  t213_n125: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t213_n127;
  t213_n126: return -0.6885370016098022f;
  t213_n127: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t213_n129;
  t213_n128: return 0.033303938806056976f;
  t213_n129: return -0.46434369683265686f;
  t213_n130: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t213_n132;
  t213_n131: return -0.9297444224357605f;
  t213_n132: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t213_n140;
  t213_n133: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t213_n135;
  t213_n134: return -0.004108244553208351f;
  t213_n135: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t213_n137;
  t213_n136: return -0.1471971869468689f;
  t213_n137: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t213_n139;
  t213_n138: return 0.021025720983743668f;
  t213_n139: return 0.07927703857421875f;
  t213_n140: return -0.11135905236005783f;
  t213_n141: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t213_n147;
  t213_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t213_n146;
  t213_n143: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t213_n145;
  t213_n144: return -0.017638683319091797f;
  t213_n145: return -0.32009291648864746f;
  t213_n146: return -0.02695496194064617f;
  t213_n147: return 0.08942238241434097f;
  t213_n148: return 0.011541541665792465f;
  t213_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t213_n153;
  t213_n150: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t213_n152;
  t213_n151: return -0.01435721293091774f;
  t213_n152: return -0.44257381558418274f;
  t213_n153: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t213_n157;
  t213_n154: if (E.getNumReferences() >= 3250585600 /*24.0*/) goto t213_n156;
  t213_n155: return 0.1016918420791626f;
  t213_n156: return 0.10776345431804657f;
  t213_n157: return 0.1292262226343155f;
  t213_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t213_n160;
  t213_n159: return -0.009936350397765636f;
  t213_n160: if (E.getNumReferences() >= 3266248704 /*87.5*/) goto t213_n170;
  t213_n161: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t213_n163;
  t213_n162: return 0.04375426098704338f;
  t213_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t213_n165;
  t213_n164: return 0.1503009796142578f;
  t213_n165: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t213_n169;
  t213_n166: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t213_n168;
  t213_n167: return -0.02440602146089077f;
  t213_n168: return 0.20195864140987396f;
  t213_n169: return 0.07325243949890137f;
  t213_n170: return 0.15063460171222687f;
  t213_n171: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t213_n225;
  t213_n172: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t213_n194;
  t213_n173: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t213_n193;
  t213_n174: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t213_n176;
  t213_n175: return 0.014969650655984879f;
  t213_n176: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t213_n192;
  t213_n177: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t213_n191;
  t213_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t213_n180;
  t213_n179: return 0.05112472176551819f;
  t213_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t213_n190;
  t213_n181: if (E.getNumReferences() >= 3261071360 /*56.0*/) goto t213_n189;
  t213_n182: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t213_n184;
  t213_n183: return 0.06183479353785515f;
  t213_n184: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t213_n186;
  t213_n185: return -0.0892113745212555f;
  t213_n186: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t213_n188;
  t213_n187: return 0.12610803544521332f;
  t213_n188: return 0.05536012724041939f;
  t213_n189: return -0.049125563353300095f;
  t213_n190: return 0.11599990725517273f;
  t213_n191: return 0.013177214190363884f;
  t213_n192: return -0.08916940540075302f;
  t213_n193: return 0.07042773067951202f;
  t213_n194: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t213_n218;
  t213_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t213_n197;
  t213_n196: return 0.03033219464123249f;
  t213_n197: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t213_n199;
  t213_n198: return 0.01492992788553238f;
  t213_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t213_n201;
  t213_n200: return -0.17619606852531433f;
  t213_n201: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t213_n215;
  t213_n202: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t213_n212;
  t213_n203: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t213_n207;
  t213_n204: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t213_n206;
  t213_n205: return -0.22034068405628204f;
  t213_n206: return 0.2940122187137604f;
  t213_n207: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t213_n211;
  t213_n208: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t213_n210;
  t213_n209: return 0.2116803526878357f;
  t213_n210: return 0.07760509103536606f;
  t213_n211: return -0.11428362131118774f;
  t213_n212: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t213_n214;
  t213_n213: return 0.3516521453857422f;
  t213_n214: return -0.10217153280973434f;
  t213_n215: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t213_n217;
  t213_n216: return 0.06629685312509537f;
  t213_n217: return -0.1559167504310608f;
  t213_n218: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t213_n222;
  t213_n219: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t213_n221;
  t213_n220: return -0.10556041449308395f;
  t213_n221: return 0.14265114068984985f;
  t213_n222: if (E.getFractionNameInContext() >= 3198110197 /*0.31111112236976624*/) goto t213_n224;
  t213_n223: return 0.1264914870262146f;
  t213_n224: return 0.21243254840373993f;
  t213_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t213_n227;
  t213_n226: return 0.005149543285369873f;
  t213_n227: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t213_n233;
  t213_n228: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t213_n232;
  t213_n229: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t213_n231;
  t213_n230: return 0.13099414110183716f;
  t213_n231: return 0.08307185024023056f;
  t213_n232: return 0.1269424855709076f;
  t213_n233: return -0.1052994579076767f;
  t213_n234: return 0.0011918003438040614f;
  t213_n235: return 0.007620325777679682f;
  t213_n236: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t213_n238;
  t213_n237: return -0.000581488711759448f;
  t213_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t213_n252;
  t213_n239: if (E.getFractionNameInContext() >= 3191774558 /*0.18614718317985535*/) goto t213_n241;
  t213_n240: return 0.12845173478126526f;
  t213_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t213_n249;
  t213_n242: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t213_n248;
  t213_n243: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t213_n245;
  t213_n244: return -0.042201776057481766f;
  t213_n245: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t213_n247;
  t213_n246: return 0.08997037261724472f;
  t213_n247: return 0.03609589859843254f;
  t213_n248: return 0.08349606394767761f;
  t213_n249: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t213_n251;
  t213_n250: return 0.08431673794984818f;
  t213_n251: return 0.008987843990325928f;
  t213_n252: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t213_n254;
  t213_n253: return 0.009533368051052094f;
  t213_n254: return 0.040935419499874115f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree214(const Example& E) {
  t214_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t214_n2;
  t214_n1: return -0.00014373898738995194f;
  t214_n2: if (E.getNumReferences() >= 3359549600 /*195234.5*/) goto t214_n252;
  t214_n3: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t214_n249;
  t214_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t214_n6;
  t214_n5: return -0.12611067295074463f;
  t214_n6: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t214_n244;
  t214_n7: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t214_n21;
  t214_n8: if (E.getNumReferences() >= 3285975040 /*440.0*/) goto t214_n20;
  t214_n9: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t214_n17;
  t214_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t214_n14;
  t214_n11: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t214_n13;
  t214_n12: return 0.02054184302687645f;
  t214_n13: return 0.0029440410435199738f;
  t214_n14: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t214_n16;
  t214_n15: return 0.008346988819539547f;
  t214_n16: return -0.015389584004878998f;
  t214_n17: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t214_n19;
  t214_n18: return 0.04887514188885689f;
  t214_n19: return -0.03205511346459389f;
  t214_n20: return 0.10429558157920837f;
  t214_n21: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t214_n27;
  t214_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t214_n26;
  t214_n23: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t214_n25;
  t214_n24: return -0.02258406952023506f;
  t214_n25: return 0.014171992428600788f;
  t214_n26: return 0.0016521790530532598f;
  t214_n27: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t214_n207;
  t214_n28: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t214_n100;
  t214_n29: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t214_n71;
  t214_n30: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t214_n60;
  t214_n31: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t214_n33;
  t214_n32: return 0.0070806704461574554f;
  t214_n33: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t214_n55;
  t214_n34: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t214_n54;
  t214_n35: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t214_n47;
  t214_n36: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t214_n46;
  t214_n37: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t214_n41;
  t214_n38: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t214_n40;
  t214_n39: return 0.0004054139426443726f;
  t214_n40: return -0.16313885152339935f;
  t214_n41: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t214_n45;
  t214_n42: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t214_n44;
  t214_n43: return -0.12505871057510376f;
  t214_n44: return -0.006500485818833113f;
  t214_n45: return -0.14202336966991425f;
  t214_n46: return 0.09639633446931839f;
  t214_n47: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t214_n49;
  t214_n48: return -0.2590993046760559f;
  t214_n49: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t214_n51;
  t214_n50: return -0.14125990867614746f;
  t214_n51: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t214_n53;
  t214_n52: return -0.10264649242162704f;
  t214_n53: return 0.12705089151859283f;
  t214_n54: return -0.003659683745354414f;
  t214_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t214_n57;
  t214_n56: return 0.03717067092657089f;
  t214_n57: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t214_n59;
  t214_n58: return 0.10741671919822693f;
  t214_n59: return 0.12042666226625443f;
  t214_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t214_n62;
  t214_n61: return -0.12278439849615097f;
  t214_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t214_n64;
  t214_n63: return -0.05774294584989548f;
  t214_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t214_n70;
  t214_n65: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t214_n69;
  t214_n66: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t214_n68;
  t214_n67: return 0.14912641048431396f;
  t214_n68: return -0.32367175817489624f;
  t214_n69: return -0.007781712803989649f;
  t214_n70: return 0.06441908329725266f;
  t214_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t214_n73;
  t214_n72: return -0.21585331857204437f;
  t214_n73: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t214_n91;
  t214_n74: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t214_n86;
  t214_n75: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t214_n81;
  t214_n76: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t214_n80;
  t214_n77: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t214_n79;
  t214_n78: return 0.015184520743787289f;
  t214_n79: return 0.15589961409568787f;
  t214_n80: return 0.11571872234344482f;
  t214_n81: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t214_n83;
  t214_n82: return 0.12950149178504944f;
  t214_n83: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t214_n85;
  t214_n84: return -0.6201540231704712f;
  t214_n85: return 0.03894590586423874f;
  t214_n86: if (E.getNumReferences() >= 3263430656 /*66.0*/) goto t214_n88;
  t214_n87: return 0.00619496451690793f;
  t214_n88: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t214_n90;
  t214_n89: return 0.10726804286241531f;
  t214_n90: return 0.06333131343126297f;
  t214_n91: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t214_n99;
  t214_n92: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t214_n98;
  t214_n93: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t214_n97;
  t214_n94: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t214_n96;
  t214_n95: return -0.6398957371711731f;
  t214_n96: return -0.1490122675895691f;
  t214_n97: return 0.05691658332943916f;
  t214_n98: return 0.14279456436634064f;
  t214_n99: return 0.02546328492462635f;
  t214_n100: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t214_n174;
  t214_n101: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t214_n137;
  t214_n102: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t214_n114;
  t214_n103: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t214_n107;
  t214_n104: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t214_n106;
  t214_n105: return -0.6314622759819031f;
  t214_n106: return 0.011200747452676296f;
  t214_n107: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t214_n111;
  t214_n108: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t214_n110;
  t214_n109: return 0.054465651512145996f;
  t214_n110: return 0.11990141123533249f;
  t214_n111: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t214_n113;
  t214_n112: return -0.3477479815483093f;
  t214_n113: return 0.06740076094865799f;
  t214_n114: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t214_n134;
  t214_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t214_n129;
  t214_n116: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t214_n118;
  t214_n117: return -0.7012214660644531f;
  t214_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t214_n126;
  t214_n119: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t214_n125;
  t214_n120: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t214_n122;
  t214_n121: return 0.06771861016750336f;
  t214_n122: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t214_n124;
  t214_n123: return -0.12388940900564194f;
  t214_n124: return 0.08775847405195236f;
  t214_n125: return -0.14989382028579712f;
  t214_n126: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t214_n128;
  t214_n127: return 0.05026036128401756f;
  t214_n128: return -0.09409070014953613f;
  t214_n129: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t214_n131;
  t214_n130: return 0.01104403380304575f;
  t214_n131: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t214_n133;
  t214_n132: return -0.11939654499292374f;
  t214_n133: return 0.16240674257278442f;
  t214_n134: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t214_n136;
  t214_n135: return 0.02954268828034401f;
  t214_n136: return 0.09459325671195984f;
  t214_n137: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t214_n153;
  t214_n138: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t214_n144;
  t214_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t214_n141;
  t214_n140: return 0.04006754979491234f;
  t214_n141: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t214_n143;
  t214_n142: return 0.03537433221936226f;
  t214_n143: return 0.30208542943000793f;
  t214_n144: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t214_n146;
  t214_n145: return 0.021356020122766495f;
  t214_n146: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t214_n152;
  t214_n147: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t214_n149;
  t214_n148: return -0.16595707833766937f;
  t214_n149: if (E.getNumReferences() >= 3233808384 /*6.0*/) goto t214_n151;
  t214_n150: return -2.181929111480713f;
  t214_n151: return -0.1796569526195526f;
  t214_n152: return -0.16799312829971313f;
  t214_n153: if (E.getFractionNameInContext() >= 3170893824 /*0.03125*/) goto t214_n173;
  t214_n154: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t214_n162;
  t214_n155: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t214_n161;
  t214_n156: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t214_n158;
  t214_n157: return 0.09336844831705093f;
  t214_n158: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t214_n160;
  t214_n159: return 0.08987148851156235f;
  t214_n160: return -0.01442671101540327f;
  t214_n161: return 0.059822868555784225f;
  t214_n162: if (E.getNumReferences() >= 3264413696 /*73.5*/) goto t214_n168;
  t214_n163: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t214_n165;
  t214_n164: return 0.09213025122880936f;
  t214_n165: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t214_n167;
  t214_n166: return -0.0931900292634964f;
  t214_n167: return 0.04569248855113983f;
  t214_n168: if (E.getNumReferences() >= 3297480704 /*1116.5*/) goto t214_n172;
  t214_n169: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t214_n171;
  t214_n170: return 0.1298830360174179f;
  t214_n171: return 0.10049274563789368f;
  t214_n172: return 0.012241164222359657f;
  t214_n173: return 0.018734483048319817f;
  t214_n174: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t214_n184;
  t214_n175: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t214_n183;
  t214_n176: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t214_n180;
  t214_n177: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t214_n179;
  t214_n178: return -0.034798718988895416f;
  t214_n179: return 0.008589666336774826f;
  t214_n180: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t214_n182;
  t214_n181: return 0.07591740787029266f;
  t214_n182: return -0.2712242901325226f;
  t214_n183: return 0.024337144568562508f;
  t214_n184: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t214_n204;
  t214_n185: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t214_n197;
  t214_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t214_n196;
  t214_n187: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t214_n195;
  t214_n188: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t214_n194;
  t214_n189: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t214_n193;
  t214_n190: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t214_n192;
  t214_n191: return 0.09124241769313812f;
  t214_n192: return -0.061930395662784576f;
  t214_n193: return 0.11014396697282791f;
  t214_n194: return 0.0024414199870079756f;
  t214_n195: return 0.061307068914175034f;
  t214_n196: return 0.15581050515174866f;
  t214_n197: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t214_n201;
  t214_n198: if (E.getNumReferences() >= 3270377472 /*119.0*/) goto t214_n200;
  t214_n199: return 0.019620392471551895f;
  t214_n200: return -0.13865923881530762f;
  t214_n201: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t214_n203;
  t214_n202: return 0.1354673206806183f;
  t214_n203: return -0.6169354915618896f;
  t214_n204: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t214_n206;
  t214_n205: return 0.10854913294315338f;
  t214_n206: return -0.037610866129398346f;
  t214_n207: if (E.getFractionNameInContext() >= 3196906982 /*0.2752525210380554*/) goto t214_n209;
  t214_n208: return 0.002679207129403949f;
  t214_n209: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t214_n231;
  t214_n210: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t214_n214;
  t214_n211: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t214_n213;
  t214_n212: return -0.027221238240599632f;
  t214_n213: return -0.010547208599746227f;
  t214_n214: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t214_n224;
  t214_n215: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t214_n223;
  t214_n216: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t214_n220;
  t214_n217: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t214_n219;
  t214_n218: return -0.024425992742180824f;
  t214_n219: return 0.0817093625664711f;
  t214_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t214_n222;
  t214_n221: return -0.1156025156378746f;
  t214_n222: return -0.12856066226959229f;
  t214_n223: return -0.00046089591342024505f;
  t214_n224: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t214_n230;
  t214_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t214_n227;
  t214_n226: return -0.36074888706207275f;
  t214_n227: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t214_n229;
  t214_n228: return -0.029935065656900406f;
  t214_n229: return 0.13381344079971313f;
  t214_n230: return 0.05438344180583954f;
  t214_n231: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t214_n237;
  t214_n232: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t214_n236;
  t214_n233: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t214_n235;
  t214_n234: return 0.014389673247933388f;
  t214_n235: return 0.07431089133024216f;
  t214_n236: return 0.052454087883234024f;
  t214_n237: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t214_n241;
  t214_n238: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t214_n240;
  t214_n239: return -0.010647068731486797f;
  t214_n240: return -0.3623816668987274f;
  t214_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t214_n243;
  t214_n242: return -0.021474678069353104f;
  t214_n243: return 0.04646608978509903f;
  t214_n244: if (E.getNumReferences() >= 3352414592 /*107411.0*/) goto t214_n246;
  t214_n245: return 0.06537142395973206f;
  t214_n246: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t214_n248;
  t214_n247: return -0.13638193905353546f;
  t214_n248: return 0.12539398670196533f;
  t214_n249: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t214_n251;
  t214_n250: return -0.25537431240081787f;
  t214_n251: return -0.7532874345779419f;
  t214_n252: if (E.getNumReferences() >= 3362781216 /*245728.5*/) goto t214_n254;
  t214_n253: return 0.08226261287927628f;
  t214_n254: return -0.23462611436843872f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree215(const Example& E) {
  t215_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t215_n20;
  t215_n1: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t215_n3;
  t215_n2: return -8.324447117047384e-05f;
  t215_n3: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t215_n5;
  t215_n4: return -0.012310924008488655f;
  t215_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t215_n15;
  t215_n6: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t215_n8;
  t215_n7: return -0.1115562915802002f;
  t215_n8: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t215_n10;
  t215_n9: return -0.13418367505073547f;
  t215_n10: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t215_n14;
  t215_n11: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t215_n13;
  t215_n12: return -0.183660626411438f;
  t215_n13: return 0.1283283233642578f;
  t215_n14: return 0.05482359975576401f;
  t215_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t215_n17;
  t215_n16: return -0.02861224301159382f;
  t215_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t215_n19;
  t215_n18: return 0.05756356567144394f;
  t215_n19: return 0.16677415370941162f;
  t215_n20: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t215_n206;
  t215_n21: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t215_n205;
  t215_n22: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t215_n62;
  t215_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t215_n61;
  t215_n24: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t215_n50;
  t215_n25: if (E.getNumReferences() >= 3274637312 /*175.0*/) goto t215_n49;
  t215_n26: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t215_n48;
  t215_n27: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t215_n29;
  t215_n28: return 0.007395100314170122f;
  t215_n29: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t215_n47;
  t215_n30: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t215_n46;
  t215_n31: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t215_n37;
  t215_n32: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t215_n34;
  t215_n33: return -0.11890233308076859f;
  t215_n34: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t215_n36;
  t215_n35: return -0.013235644437372684f;
  t215_n36: return 0.14356952905654907f;
  t215_n37: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t215_n45;
  t215_n38: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n40;
  t215_n39: return -0.007209479808807373f;
  t215_n40: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t215_n44;
  t215_n41: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t215_n43;
  t215_n42: return -0.05546201765537262f;
  t215_n43: return 0.1544123888015747f;
  t215_n44: return 0.1328791379928589f;
  t215_n45: return -0.24619022011756897f;
  t215_n46: return -0.25123390555381775f;
  t215_n47: return 0.00021425083104986697f;
  t215_n48: return -0.28476592898368835f;
  t215_n49: return 0.048342470079660416f;
  t215_n50: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t215_n60;
  t215_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t215_n57;
  t215_n52: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t215_n54;
  t215_n53: return 0.025565672665834427f;
  t215_n54: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n56;
  t215_n55: return 0.04144538566470146f;
  t215_n56: return 0.10203276574611664f;
  t215_n57: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t215_n59;
  t215_n58: return 0.12173520773649216f;
  t215_n59: return 0.09774956852197647f;
  t215_n60: return -0.6561558842658997f;
  t215_n61: return 0.02534254640340805f;
  t215_n62: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t215_n154;
  t215_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t215_n89;
  t215_n64: if (E.getNumReferences() >= 3365173872 /*304083.5*/) goto t215_n82;
  t215_n65: if (E.getNumReferences() >= 3330759936 /*17326.5*/) goto t215_n81;
  t215_n66: if (E.getNumReferences() >= 3330320896 /*16469.0*/) goto t215_n80;
  t215_n67: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t215_n73;
  t215_n68: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t215_n70;
  t215_n69: return -0.021765490993857384f;
  t215_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t215_n72;
  t215_n71: return -0.4337118864059448f;
  t215_n72: return -0.13188302516937256f;
  t215_n73: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t215_n79;
  t215_n74: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t215_n78;
  t215_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t215_n77;
  t215_n76: return 0.10772676020860672f;
  t215_n77: return 0.11490678042173386f;
  t215_n78: return 0.004922092892229557f;
  t215_n79: return -0.06307145208120346f;
  t215_n80: return -0.7482934594154358f;
  t215_n81: return 0.017972704023122787f;
  t215_n82: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t215_n84;
  t215_n83: return -0.03134969249367714f;
  t215_n84: if (E.getNumReferences() >= 3376125336 /*768345.5*/) goto t215_n88;
  t215_n85: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t215_n87;
  t215_n86: return 0.01276054885238409f;
  t215_n87: return -0.36379528045654297f;
  t215_n88: return 0.11918766796588898f;
  t215_n89: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t215_n111;
  t215_n90: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t215_n108;
  t215_n91: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t215_n95;
  t215_n92: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t215_n94;
  t215_n93: return -0.10580018907785416f;
  t215_n94: return -0.6229572296142578f;
  t215_n95: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t215_n105;
  t215_n96: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t215_n100;
  t215_n97: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t215_n99;
  t215_n98: return -0.11010566353797913f;
  t215_n99: return 0.125234454870224f;
  t215_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t215_n104;
  t215_n101: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n103;
  t215_n102: return -0.010832400992512703f;
  t215_n103: return 0.11038614064455032f;
  t215_n104: return 0.02650037221610546f;
  t215_n105: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t215_n107;
  t215_n106: return -0.18253254890441895f;
  t215_n107: return 0.11654140800237656f;
  t215_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t215_n110;
  t215_n109: return 0.035655491054058075f;
  t215_n110: return 0.08369428664445877f;
  t215_n111: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t215_n131;
  t215_n112: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t215_n124;
  t215_n113: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t215_n115;
  t215_n114: return 0.003055540844798088f;
  t215_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t215_n121;
  t215_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n120;
  t215_n117: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t215_n119;
  t215_n118: return 0.03221553564071655f;
  t215_n119: return 0.1292629837989807f;
  t215_n120: return -0.06692445278167725f;
  t215_n121: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t215_n123;
  t215_n122: return 0.13772785663604736f;
  t215_n123: return -0.01629951223731041f;
  t215_n124: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t215_n128;
  t215_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n127;
  t215_n126: return -0.005075254011899233f;
  t215_n127: return -0.11965915560722351f;
  t215_n128: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t215_n130;
  t215_n129: return 0.1501011997461319f;
  t215_n130: return 0.01811552606523037f;
  t215_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t215_n143;
  t215_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t215_n138;
  t215_n133: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t215_n135;
  t215_n134: return 0.07081438601016998f;
  t215_n135: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t215_n137;
  t215_n136: return -0.9224581122398376f;
  t215_n137: return -0.060934290289878845f;
  t215_n138: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t215_n140;
  t215_n139: return -0.03384235501289368f;
  t215_n140: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t215_n142;
  t215_n141: return 0.13757382333278656f;
  t215_n142: return 0.11030615866184235f;
  t215_n143: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t215_n151;
  t215_n144: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t215_n150;
  t215_n145: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t215_n149;
  t215_n146: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t215_n148;
  t215_n147: return 0.10741616040468216f;
  t215_n148: return 0.14558200538158417f;
  t215_n149: return 0.0637090876698494f;
  t215_n150: return -0.3410343825817108f;
  t215_n151: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t215_n153;
  t215_n152: return 0.03595633804798126f;
  t215_n153: return -0.17590242624282837f;
  t215_n154: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t215_n186;
  t215_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t215_n177;
  t215_n156: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t215_n160;
  t215_n157: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t215_n159;
  t215_n158: return 0.0011402418604120612f;
  t215_n159: return 0.06900434195995331f;
  t215_n160: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n166;
  t215_n161: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t215_n163;
  t215_n162: return 0.04120289161801338f;
  t215_n163: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t215_n165;
  t215_n164: return -0.7759310603141785f;
  t215_n165: return -0.1144154742360115f;
  t215_n166: if (E.getNumReferences() >= 3310524416 /*3369.5*/) goto t215_n176;
  t215_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t215_n169;
  t215_n168: return -0.6634150147438049f;
  t215_n169: if (E.getFractionNameInContext() >= 3180529969 /*0.07179487496614456*/) goto t215_n171;
  t215_n170: return 0.06820198148488998f;
  t215_n171: if (E.getNumReferences() >= 3273195520 /*153.0*/) goto t215_n175;
  t215_n172: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t215_n174;
  t215_n173: return -0.09509819000959396f;
  t215_n174: return 0.046524111181497574f;
  t215_n175: return 0.19229644536972046f;
  t215_n176: return 0.02676590345799923f;
  t215_n177: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t215_n185;
  t215_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t215_n180;
  t215_n179: return 0.034630514681339264f;
  t215_n180: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t215_n184;
  t215_n181: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t215_n183;
  t215_n182: return 0.04220842942595482f;
  t215_n183: return 0.07839348912239075f;
  t215_n184: return -0.06703820824623108f;
  t215_n185: return 0.016036611050367355f;
  t215_n186: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t215_n194;
  t215_n187: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t215_n193;
  t215_n188: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t215_n190;
  t215_n189: return 0.04460955038666725f;
  t215_n190: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t215_n192;
  t215_n191: return 0.12532153725624084f;
  t215_n192: return 0.07934573292732239f;
  t215_n193: return 0.0023251702077686787f;
  t215_n194: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t215_n200;
  t215_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t215_n197;
  t215_n196: return -0.1122840940952301f;
  t215_n197: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t215_n199;
  t215_n198: return -0.06785425543785095f;
  t215_n199: return 0.0902586579322815f;
  t215_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t215_n204;
  t215_n201: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t215_n203;
  t215_n202: return 0.014405385591089725f;
  t215_n203: return -0.039640121161937714f;
  t215_n204: return 0.05776000767946243f;
  t215_n205: return 0.08302593231201172f;
  t215_n206: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t215_n254;
  t215_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t215_n213;
  t215_n208: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t215_n210;
  t215_n209: return -0.06328992545604706f;
  t215_n210: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t215_n212;
  t215_n211: return 0.004547647200524807f;
  t215_n212: return 0.08380327373743057f;
  t215_n213: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t215_n235;
  t215_n214: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t215_n218;
  t215_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t215_n217;
  t215_n216: return -0.021237842738628387f;
  t215_n217: return 0.076621413230896f;
  t215_n218: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t215_n220;
  t215_n219: return 0.3755891025066376f;
  t215_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t215_n230;
  t215_n221: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t215_n223;
  t215_n222: return -0.31775882840156555f;
  t215_n223: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t215_n225;
  t215_n224: return 0.09250286221504211f;
  t215_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t215_n227;
  t215_n226: return -0.5813613533973694f;
  t215_n227: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t215_n229;
  t215_n228: return -0.12976481020450592f;
  t215_n229: return 0.11255067586898804f;
  t215_n230: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t215_n234;
  t215_n231: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t215_n233;
  t215_n232: return 0.07640504837036133f;
  t215_n233: return 0.1803392469882965f;
  t215_n234: return 0.04718611389398575f;
  t215_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t215_n239;
  t215_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t215_n238;
  t215_n237: return -0.07479357719421387f;
  t215_n238: return 0.12023665010929108f;
  t215_n239: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t215_n253;
  t215_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t215_n250;
  t215_n241: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t215_n243;
  t215_n242: return -0.10986989736557007f;
  t215_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t215_n249;
  t215_n244: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t215_n246;
  t215_n245: return 0.0664755254983902f;
  t215_n246: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t215_n248;
  t215_n247: return 0.2203027904033661f;
  t215_n248: return 0.17433467507362366f;
  t215_n249: return -0.06976243108510971f;
  t215_n250: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t215_n252;
  t215_n251: return 0.21368153393268585f;
  t215_n252: return 0.23443906009197235f;
  t215_n253: return 0.18319617211818695f;
  t215_n254: return 0.07387746125459671f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree216(const Example& E) {
  t216_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t216_n218;
  t216_n1: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t216_n3;
  t216_n2: return -0.00022224019630812109f;
  t216_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t216_n133;
  t216_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t216_n56;
  t216_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t216_n7;
  t216_n6: return -0.004202173091471195f;
  t216_n7: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t216_n11;
  t216_n8: if (E.getNumReferences() >= 3312187392 /*3775.5*/) goto t216_n10;
  t216_n9: return 0.0025992807932198048f;
  t216_n10: return 0.0698801800608635f;
  t216_n11: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t216_n43;
  t216_n12: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t216_n22;
  t216_n13: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t216_n15;
  t216_n14: return -0.02593507617712021f;
  t216_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t216_n21;
  t216_n16: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t216_n18;
  t216_n17: return -0.016465436667203903f;
  t216_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t216_n20;
  t216_n19: return -0.08759929984807968f;
  t216_n20: return 0.2663816213607788f;
  t216_n21: return 0.023401794955134392f;
  t216_n22: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t216_n42;
  t216_n23: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t216_n27;
  t216_n24: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t216_n26;
  t216_n25: return 0.03691767156124115f;
  t216_n26: return -0.03730209916830063f;
  t216_n27: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t216_n41;
  t216_n28: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t216_n32;
  t216_n29: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t216_n31;
  t216_n30: return 0.1008872240781784f;
  t216_n31: return 0.011411257088184357f;
  t216_n32: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t216_n40;
  t216_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t216_n35;
  t216_n34: return -0.03598840534687042f;
  t216_n35: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t216_n39;
  t216_n36: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t216_n38;
  t216_n37: return 0.16840475797653198f;
  t216_n38: return 0.04176581650972366f;
  t216_n39: return -0.12490499019622803f;
  t216_n40: return 0.08091767877340317f;
  t216_n41: return -0.06061749532818794f;
  t216_n42: return -0.011716301552951336f;
  t216_n43: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t216_n55;
  t216_n44: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t216_n52;
  t216_n45: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t216_n47;
  t216_n46: return 0.05738869681954384f;
  t216_n47: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t216_n49;
  t216_n48: return 0.018756110221147537f;
  t216_n49: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t216_n51;
  t216_n50: return 0.1244947537779808f;
  t216_n51: return 0.0776815265417099f;
  t216_n52: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t216_n54;
  t216_n53: return 0.036055877804756165f;
  t216_n54: return -0.031625788658857346f;
  t216_n55: return 0.009508082643151283f;
  t216_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t216_n132;
  t216_n57: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t216_n119;
  t216_n58: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t216_n60;
  t216_n59: return -0.014850818552076817f;
  t216_n60: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t216_n106;
  t216_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t216_n75;
  t216_n62: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t216_n68;
  t216_n63: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t216_n65;
  t216_n64: return -0.004220801405608654f;
  t216_n65: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t216_n67;
  t216_n66: return 0.04296271875500679f;
  t216_n67: return 0.11786769330501556f;
  t216_n68: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t216_n74;
  t216_n69: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t216_n71;
  t216_n70: return 0.048496875911951065f;
  t216_n71: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t216_n73;
  t216_n72: return -0.21277792751789093f;
  t216_n73: return 0.07606672495603561f;
  t216_n74: return 0.17179693281650543f;
  t216_n75: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t216_n101;
  t216_n76: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t216_n90;
  t216_n77: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t216_n85;
  t216_n78: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t216_n80;
  t216_n79: return 0.06756901741027832f;
  t216_n80: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t216_n82;
  t216_n81: return 0.11699933558702469f;
  t216_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t216_n84;
  t216_n83: return -0.22372175753116608f;
  t216_n84: return 0.06349124014377594f;
  t216_n85: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t216_n89;
  t216_n86: if (E.getFractionNameInContext() >= 3202118087 /*0.4305555522441864*/) goto t216_n88;
  t216_n87: return -0.03305544704198837f;
  t216_n88: return 0.07605664432048798f;
  t216_n89: return -0.4389355182647705f;
  t216_n90: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t216_n94;
  t216_n91: if (E.getFractionNameInContext() >= 3208223130 /*0.7250000238418579*/) goto t216_n93;
  t216_n92: return -0.025935404002666473f;
  t216_n93: return -1.63064706325531f;
  t216_n94: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t216_n100;
  t216_n95: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t216_n99;
  t216_n96: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t216_n98;
  t216_n97: return 0.11988359689712524f;
  t216_n98: return 0.07335149496793747f;
  t216_n99: return -0.04366285726428032f;
  t216_n100: return 0.1537722498178482f;
  t216_n101: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t216_n105;
  t216_n102: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t216_n104;
  t216_n103: return 0.033964380621910095f;
  t216_n104: return -0.17342890799045563f;
  t216_n105: return 0.10789617151021957f;
  t216_n106: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t216_n114;
  t216_n107: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t216_n109;
  t216_n108: return -0.042123597115278244f;
  t216_n109: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t216_n111;
  t216_n110: return -0.0022501349449157715f;
  t216_n111: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t216_n113;
  t216_n112: return 0.10772497206926346f;
  t216_n113: return 0.0584319531917572f;
  t216_n114: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t216_n116;
  t216_n115: return -0.0038909141439944506f;
  t216_n116: if (E.getFractionNameInContext() >= 3191495258 /*0.1819852888584137*/) goto t216_n118;
  t216_n117: return -0.5126950740814209f;
  t216_n118: return 0.02934025414288044f;
  t216_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t216_n131;
  t216_n120: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t216_n124;
  t216_n121: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t216_n123;
  t216_n122: return -0.05212235450744629f;
  t216_n123: return 0.11646933853626251f;
  t216_n124: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t216_n128;
  t216_n125: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t216_n127;
  t216_n126: return 0.0426308810710907f;
  t216_n127: return -0.06052621826529503f;
  t216_n128: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t216_n130;
  t216_n129: return -0.18356336653232574f;
  t216_n130: return 0.037757519632577896f;
  t216_n131: return -0.02015146054327488f;
  t216_n132: return 0.03348710760474205f;
  t216_n133: if (E.getFractionNameInContext() >= 3205438578 /*0.5590277910232544*/) goto t216_n207;
  t216_n134: if (E.getFractionNameInContext() >= 3196906982 /*0.2752525210380554*/) goto t216_n164;
  t216_n135: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t216_n139;
  t216_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t216_n138;
  t216_n137: return -0.009482743218541145f;
  t216_n138: return 0.0357237383723259f;
  t216_n139: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t216_n159;
  t216_n140: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t216_n158;
  t216_n141: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t216_n155;
  t216_n142: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t216_n148;
  t216_n143: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t216_n145;
  t216_n144: return -0.017778819426894188f;
  t216_n145: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t216_n147;
  t216_n146: return -0.20629161596298218f;
  t216_n147: return 0.09403527528047562f;
  t216_n148: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t216_n150;
  t216_n149: return -0.30694714188575745f;
  t216_n150: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t216_n152;
  t216_n151: return 0.08225294947624207f;
  t216_n152: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t216_n154;
  t216_n153: return 0.19766934216022491f;
  t216_n154: return 0.055365778505802155f;
  t216_n155: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t216_n157;
  t216_n156: return 0.07117893546819687f;
  t216_n157: return -0.19021378457546234f;
  t216_n158: return 0.01912982389330864f;
  t216_n159: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t216_n161;
  t216_n160: return 0.020340178161859512f;
  t216_n161: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t216_n163;
  t216_n162: return 0.11385831236839294f;
  t216_n163: return -0.15045495331287384f;
  t216_n164: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t216_n204;
  t216_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t216_n167;
  t216_n166: return -0.02872585877776146f;
  t216_n167: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t216_n201;
  t216_n168: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t216_n178;
  t216_n169: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t216_n177;
  t216_n170: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t216_n174;
  t216_n171: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t216_n173;
  t216_n172: return 0.08567014336585999f;
  t216_n173: return 0.006581497378647327f;
  t216_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t216_n176;
  t216_n175: return -0.10116057842969894f;
  t216_n176: return 0.21855397522449493f;
  t216_n177: return 0.002293594414368272f;
  t216_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t216_n180;
  t216_n179: return 0.037079401314258575f;
  t216_n180: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t216_n200;
  t216_n181: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t216_n197;
  t216_n182: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t216_n196;
  t216_n183: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t216_n193;
  t216_n184: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t216_n192;
  t216_n185: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t216_n191;
  t216_n186: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t216_n188;
  t216_n187: return 0.03321918845176697f;
  t216_n188: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t216_n190;
  t216_n189: return 0.12276171147823334f;
  t216_n190: return 0.005376605782657862f;
  t216_n191: return 0.18852218985557556f;
  t216_n192: return -0.06476277858018875f;
  t216_n193: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t216_n195;
  t216_n194: return 0.16267284750938416f;
  t216_n195: return 0.10887037962675095f;
  t216_n196: return 0.12555323541164398f;
  t216_n197: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t216_n199;
  t216_n198: return -0.2746085524559021f;
  t216_n199: return 0.1032308042049408f;
  t216_n200: return -0.007009149994701147f;
  t216_n201: if (E.getNumReferences() >= 3265331200 /*80.5*/) goto t216_n203;
  t216_n202: return -0.01582574099302292f;
  t216_n203: return 0.10982731729745865f;
  t216_n204: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t216_n206;
  t216_n205: return 0.00905695278197527f;
  t216_n206: return -0.09638497978448868f;
  t216_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t216_n215;
  t216_n208: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t216_n214;
  t216_n209: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t216_n211;
  t216_n210: return 0.02785797044634819f;
  t216_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t216_n213;
  t216_n212: return -0.1048375815153122f;
  t216_n213: return 0.12079812586307526f;
  t216_n214: return -0.012149897404015064f;
  t216_n215: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t216_n217;
  t216_n216: return 0.14124920964241028f;
  t216_n217: return -0.07493839412927628f;
  t216_n218: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t216_n222;
  t216_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t216_n221;
  t216_n220: return 0.0024406127631664276f;
  t216_n221: return 0.024985849857330322f;
  t216_n222: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t216_n254;
  t216_n223: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t216_n243;
  t216_n224: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t216_n228;
  t216_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t216_n227;
  t216_n226: return -0.019344504922628403f;
  t216_n227: return 0.06770483404397964f;
  t216_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t216_n238;
  t216_n229: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t216_n237;
  t216_n230: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t216_n236;
  t216_n231: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t216_n235;
  t216_n232: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t216_n234;
  t216_n233: return -0.22274288535118103f;
  t216_n234: return 0.11332927644252777f;
  t216_n235: return -0.04092175513505936f;
  t216_n236: return -0.01610437221825123f;
  t216_n237: return 0.10835906118154526f;
  t216_n238: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t216_n240;
  t216_n239: return 0.04487187787890434f;
  t216_n240: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t216_n242;
  t216_n241: return 0.19563716650009155f;
  t216_n242: return 0.18211033940315247f;
  t216_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t216_n245;
  t216_n244: return -0.05145297944545746f;
  t216_n245: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t216_n253;
  t216_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t216_n252;
  t216_n247: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t216_n249;
  t216_n248: return -0.06461894512176514f;
  t216_n249: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t216_n251;
  t216_n250: return 0.1894560307264328f;
  t216_n251: return 0.0028417420107871294f;
  t216_n252: return 0.15150751173496246f;
  t216_n253: return 0.18521267175674438f;
  t216_n254: return 0.07128550112247467f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree217(const Example& E) {
  t217_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t217_n248;
  t217_n1: if (E.getNumReferences() >= 3306514432 /*2390.5*/) goto t217_n11;
  t217_n2: if (E.getNumReferences() >= 3306424320 /*2368.5*/) goto t217_n4;
  t217_n3: return -0.0003398969420231879f;
  t217_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t217_n10;
  t217_n5: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t217_n7;
  t217_n6: return -0.09294426441192627f;
  t217_n7: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t217_n9;
  t217_n8: return -0.10664258897304535f;
  t217_n9: return -0.47089213132858276f;
  t217_n10: return 0.12450715154409409f;
  t217_n11: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t217_n33;
  t217_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t217_n16;
  t217_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t217_n15;
  t217_n14: return -0.08134359866380692f;
  t217_n15: return 0.03886666148900986f;
  t217_n16: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t217_n32;
  t217_n17: if (E.getNumReferences() >= 3321825280 /*8161.0*/) goto t217_n31;
  t217_n18: if (E.getNumReferences() >= 3320384512 /*7457.5*/) goto t217_n20;
  t217_n19: return 0.030124586075544357f;
  t217_n20: if (E.getNumReferences() >= 3321112576 /*7813.0*/) goto t217_n28;
  t217_n21: if (E.getNumReferences() >= 3320945664 /*7731.5*/) goto t217_n27;
  t217_n22: if (E.getNumReferences() >= 3320802304 /*7661.5*/) goto t217_n26;
  t217_n23: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t217_n25;
  t217_n24: return 0.1383035033941269f;
  t217_n25: return 0.07621867954730988f;
  t217_n26: return -0.0995996817946434f;
  t217_n27: return 0.15032565593719482f;
  t217_n28: if (E.getNumReferences() >= 3321802752 /*8150.0*/) goto t217_n30;
  t217_n29: return -0.01630363240838051f;
  t217_n30: return 0.15158778429031372f;
  t217_n31: return 0.009343365207314491f;
  t217_n32: return 0.00697467103600502f;
  t217_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t217_n121;
  t217_n34: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t217_n118;
  t217_n35: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t217_n37;
  t217_n36: return -0.11279362440109253f;
  t217_n37: if (E.getNumReferences() >= 3327053312 /*13235.5*/) goto t217_n41;
  t217_n38: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t217_n40;
  t217_n39: return -0.0033267736434936523f;
  t217_n40: return -0.06810490787029266f;
  t217_n41: if (E.getNumReferences() >= 3327144448 /*13324.5*/) goto t217_n47;
  t217_n42: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t217_n44;
  t217_n43: return -0.10892517864704132f;
  t217_n44: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t217_n46;
  t217_n45: return 0.22005346417427063f;
  t217_n46: return 0.16906419396400452f;
  t217_n47: if (E.getNumReferences() >= 3330499328 /*16817.5*/) goto t217_n71;
  t217_n48: if (E.getNumReferences() >= 3327636992 /*13805.5*/) goto t217_n50;
  t217_n49: return -0.03721986711025238f;
  t217_n50: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t217_n60;
  t217_n51: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t217_n55;
  t217_n52: if (E.getNumReferences() >= 3328016384 /*14176.0*/) goto t217_n54;
  t217_n53: return 0.09888190776109695f;
  t217_n54: return -0.1866273283958435f;
  t217_n55: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t217_n59;
  t217_n56: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t217_n58;
  t217_n57: return -0.13093720376491547f;
  t217_n58: return 0.11553268134593964f;
  t217_n59: return 0.11736533045768738f;
  t217_n60: if (E.getNumReferences() >= 3329438720 /*15565.0*/) goto t217_n62;
  t217_n61: return -0.056149598211050034f;
  t217_n62: if (E.getNumReferences() >= 3329564160 /*15687.5*/) goto t217_n64;
  t217_n63: return 0.08950873464345932f;
  t217_n64: if (E.getNumReferences() >= 3330487040 /*16793.5*/) goto t217_n68;
  t217_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t217_n67;
  t217_n66: return -0.13674096763134003f;
  t217_n67: return 0.07771372050046921f;
  t217_n68: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t217_n70;
  t217_n69: return -0.040926188230514526f;
  t217_n70: return 0.13649584352970123f;
  t217_n71: if (E.getNumReferences() >= 3331498752 /*18769.5*/) goto t217_n73;
  t217_n72: return -0.1381049007177353f;
  t217_n73: if (E.getNumReferences() >= 3331793408 /*19345.0*/) goto t217_n83;
  t217_n74: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t217_n80;
  t217_n75: if (E.getNumReferences() >= 3331557888 /*18885.0*/) goto t217_n79;
  t217_n76: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t217_n78;
  t217_n77: return -0.011831436306238174f;
  t217_n78: return 0.11293952167034149f;
  t217_n79: return -0.17799849808216095f;
  t217_n80: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t217_n82;
  t217_n81: return 0.16761577129364014f;
  t217_n82: return 0.12650124728679657f;
  t217_n83: if (E.getNumReferences() >= 3336160512 /*27874.5*/) goto t217_n91;
  t217_n84: if (E.getNumReferences() >= 3332305408 /*20345.0*/) goto t217_n90;
  t217_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t217_n87;
  t217_n86: return -0.003887808881700039f;
  t217_n87: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t217_n89;
  t217_n88: return 0.03971268981695175f;
  t217_n89: return 0.1200193390250206f;
  t217_n90: return -0.06815442442893982f;
  t217_n91: if (E.getNumReferences() >= 3336764160 /*29053.5*/) goto t217_n99;
  t217_n92: if (E.getNumReferences() >= 3336534016 /*28604.0*/) goto t217_n94;
  t217_n93: return 0.023858169093728065f;
  t217_n94: if (E.getNumReferences() >= 3336705792 /*28939.5*/) goto t217_n98;
  t217_n95: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t217_n97;
  t217_n96: return 0.15265198051929474f;
  t217_n97: return 0.15568190813064575f;
  t217_n98: return 0.08176694065332413f;
  t217_n99: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t217_n101;
  t217_n100: return -0.36767733097076416f;
  t217_n101: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t217_n103;
  t217_n102: return 0.09323474019765854f;
  t217_n103: if (E.getNumReferences() >= 3338616320 /*32671.0*/) goto t217_n105;
  t217_n104: return -0.1377364993095398f;
  t217_n105: if (E.getNumReferences() >= 3340748416 /*40902.5*/) goto t217_n117;
  t217_n106: if (E.getNumReferences() >= 3340537472 /*40078.5*/) goto t217_n116;
  t217_n107: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t217_n115;
  t217_n108: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t217_n110;
  t217_n109: return -0.021310560405254364f;
  t217_n110: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t217_n114;
  t217_n111: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t217_n113;
  t217_n112: return 0.14446015655994415f;
  t217_n113: return 0.10637932270765305f;
  t217_n114: return 0.04385150969028473f;
  t217_n115: return -0.18423229455947876f;
  t217_n116: return 0.11102434992790222f;
  t217_n117: return -0.006687142886221409f;
  t217_n118: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t217_n120;
  t217_n119: return 0.026772823184728622f;
  t217_n120: return -0.06747034937143326f;
  t217_n121: if (E.getNumReferences() >= 3318849536 /*6708.0*/) goto t217_n247;
  t217_n122: if (E.getNumReferences() >= 3318411264 /*6494.0*/) goto t217_n224;
  t217_n123: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t217_n199;
  t217_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t217_n140;
  t217_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t217_n127;
  t217_n126: return -0.004486845340579748f;
  t217_n127: if (E.getNumReferences() >= 3308777472 /*2943.0*/) goto t217_n139;
  t217_n128: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t217_n130;
  t217_n129: return 0.14758066833019257f;
  t217_n130: if (E.getNumReferences() >= 3307108352 /*2535.5*/) goto t217_n134;
  t217_n131: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t217_n133;
  t217_n132: return -0.10010379552841187f;
  t217_n133: return 0.12645301222801208f;
  t217_n134: if (E.getNumReferences() >= 3307757568 /*2694.0*/) goto t217_n136;
  t217_n135: return -0.07253601402044296f;
  t217_n136: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t217_n138;
  t217_n137: return 0.14271579682826996f;
  t217_n138: return 0.1436345875263214f;
  t217_n139: return 0.04109447821974754f;
  t217_n140: if (E.getNumReferences() >= 3306864640 /*2476.0*/) goto t217_n146;
  t217_n141: if (E.getNumReferences() >= 3306856448 /*2474.0*/) goto t217_n145;
  t217_n142: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t217_n144;
  t217_n143: return -0.005811640061438084f;
  t217_n144: return 0.10314103960990906f;
  t217_n145: return 0.1367071568965912f;
  t217_n146: if (E.getNumReferences() >= 3310528512 /*3370.5*/) goto t217_n158;
  t217_n147: if (E.getNumReferences() >= 3310450688 /*3351.5*/) goto t217_n157;
  t217_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t217_n150;
  t217_n149: return -0.027818752452731133f;
  t217_n150: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t217_n152;
  t217_n151: return 0.06881655007600784f;
  t217_n152: if (E.getNumReferences() >= 3307169792 /*2550.5*/) goto t217_n154;
  t217_n153: return 0.12668997049331665f;
  t217_n154: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t217_n156;
  t217_n155: return 0.8549372553825378f;
  t217_n156: return 0.04129726439714432f;
  t217_n157: return -0.27632394433021545f;
  t217_n158: if (E.getNumReferences() >= 3316045824 /*5339.0*/) goto t217_n180;
  t217_n159: if (E.getNumReferences() >= 3315954688 /*5294.5*/) goto t217_n175;
  t217_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t217_n174;
  t217_n161: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t217_n173;
  t217_n162: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t217_n164;
  t217_n163: return -0.014595367945730686f;
  t217_n164: if (E.getNumReferences() >= 3313557504 /*4124.0*/) goto t217_n166;
  t217_n165: return 0.016482418403029442f;
  t217_n166: if (E.getNumReferences() >= 3313644544 /*4166.5*/) goto t217_n168;
  t217_n167: return 0.14535902440547943f;
  t217_n168: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t217_n170;
  t217_n169: return -0.021071521565318108f;
  t217_n170: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t217_n172;
  t217_n171: return 0.10068019479513168f;
  t217_n172: return -0.1571211814880371f;
  t217_n173: return -0.05150702968239784f;
  t217_n174: return 0.05397460609674454f;
  t217_n175: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t217_n179;
  t217_n176: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t217_n178;
  t217_n177: return 0.10366339981555939f;
  t217_n178: return 0.1297450065612793f;
  t217_n179: return -0.1006784588098526f;
  t217_n180: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t217_n184;
  t217_n181: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t217_n183;
  t217_n182: return -2.1979475021362305f;
  t217_n183: return 0.011228172108530998f;
  t217_n184: if (E.getNumReferences() >= 3317416960 /*6008.5*/) goto t217_n190;
  t217_n185: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t217_n187;
  t217_n186: return 0.12981726229190826f;
  t217_n187: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t217_n189;
  t217_n188: return 0.17535671591758728f;
  t217_n189: return -0.1406605988740921f;
  t217_n190: if (E.getNumReferences() >= 3317616640 /*6106.0*/) goto t217_n194;
  t217_n191: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t217_n193;
  t217_n192: return 0.11265795677900314f;
  t217_n193: return 0.1437888741493225f;
  t217_n194: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t217_n198;
  t217_n195: if (E.getNumReferences() >= 3318166528 /*6374.5*/) goto t217_n197;
  t217_n196: return 0.13799159228801727f;
  t217_n197: return -0.10791786015033722f;
  t217_n198: return -0.04797689989209175f;
  t217_n199: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t217_n219;
  t217_n200: if (E.getNumReferences() >= 3307098112 /*2533.0*/) goto t217_n204;
  t217_n201: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t217_n203;
  t217_n202: return 0.12270418554544449f;
  t217_n203: return 0.07146667689085007f;
  t217_n204: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t217_n208;
  t217_n205: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t217_n207;
  t217_n206: return -0.35621580481529236f;
  t217_n207: return 0.03330662474036217f;
  t217_n208: if (E.getNumReferences() >= 3308249088 /*2814.0*/) goto t217_n218;
  t217_n209: if (E.getFractionNameInContext() >= 3184451170 /*0.10101009905338287*/) goto t217_n217;
  t217_n210: if (E.getNumReferences() >= 3307874304 /*2722.5*/) goto t217_n212;
  t217_n211: return -0.07326314598321915f;
  t217_n212: if (E.getNumReferences() >= 3307921408 /*2734.0*/) goto t217_n214;
  t217_n213: return -2.478522777557373f;
  t217_n214: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t217_n216;
  t217_n215: return -1.7968758344650269f;
  t217_n216: return 0.021351903676986694f;
  t217_n217: return 0.06989220529794693f;
  t217_n218: return -0.05716109648346901f;
  t217_n219: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t217_n223;
  t217_n220: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t217_n222;
  t217_n221: return 0.1039302796125412f;
  t217_n222: return 0.09062378853559494f;
  t217_n223: return 0.09090439975261688f;
  t217_n224: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t217_n240;
  t217_n225: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t217_n233;
  t217_n226: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t217_n228;
  t217_n227: return 0.05486714094877243f;
  t217_n228: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t217_n232;
  t217_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t217_n231;
  t217_n230: return 0.06328912824392319f;
  t217_n231: return 0.3358222544193268f;
  t217_n232: return -0.6605071425437927f;
  t217_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t217_n239;
  t217_n234: if (E.getNumReferences() >= 3318735872 /*6652.5*/) goto t217_n236;
  t217_n235: return -0.05327189341187477f;
  t217_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t217_n238;
  t217_n237: return 0.13054975867271423f;
  t217_n238: return 0.1423722803592682f;
  t217_n239: return 0.1337660253047943f;
  t217_n240: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t217_n244;
  t217_n241: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t217_n243;
  t217_n242: return 0.06227399781346321f;
  t217_n243: return 0.11392082273960114f;
  t217_n244: if (E.getNumReferences() >= 3318637568 /*6604.5*/) goto t217_n246;
  t217_n245: return 0.1367977410554886f;
  t217_n246: return 0.11426930129528046f;
  t217_n247: return -3.874232334055705e-06f;
  t217_n248: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t217_n252;
  t217_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_TopLevel))) goto t217_n251;
  t217_n250: return 0.02707040309906006f;
  t217_n251: return 0.13543497025966644f;
  t217_n252: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t217_n254;
  t217_n253: return -0.02775740623474121f;
  t217_n254: return 0.0015182921197265387f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree218(const Example& E) {
  t218_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t218_n148;
  t218_n1: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t218_n109;
  t218_n2: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t218_n108;
  t218_n3: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t218_n5;
  t218_n4: return 0.00023052006145007908f;
  t218_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n25;
  t218_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t218_n24;
  t218_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n9;
  t218_n8: return -0.009737391956150532f;
  t218_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t218_n15;
  t218_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n12;
  t218_n11: return -0.022734418511390686f;
  t218_n12: if (E.getFractionNameInContext() >= 3177875683 /*0.057259451597929*/) goto t218_n14;
  t218_n13: return -0.03025197796523571f;
  t218_n14: return 0.004842444323003292f;
  t218_n15: if (E.getNumReferences() >= 3333761024 /*23188.0*/) goto t218_n21;
  t218_n16: if (E.getNumReferences() >= 3333127936 /*21951.5*/) goto t218_n20;
  t218_n17: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t218_n19;
  t218_n18: return 0.012698935344815254f;
  t218_n19: return -0.0015964774647727609f;
  t218_n20: return 0.12034722417593002f;
  t218_n21: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n23;
  t218_n22: return -0.14167682826519012f;
  t218_n23: return -0.009427987970411777f;
  t218_n24: return 0.00848524272441864f;
  t218_n25: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t218_n27;
  t218_n26: return 0.010166878812015057f;
  t218_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t218_n57;
  t218_n28: if (E.getNumReferences() >= 3315822592 /*5230.0*/) goto t218_n48;
  t218_n29: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t218_n45;
  t218_n30: if (E.getNumReferences() >= 3280371712 /*269.0*/) goto t218_n44;
  t218_n31: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t218_n43;
  t218_n32: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t218_n38;
  t218_n33: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t218_n37;
  t218_n34: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t218_n36;
  t218_n35: return -0.06119612976908684f;
  t218_n36: return -0.003977784421294928f;
  t218_n37: return 0.5984421968460083f;
  t218_n38: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t218_n40;
  t218_n39: return -0.07313496619462967f;
  t218_n40: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t218_n42;
  t218_n41: return -0.5668542385101318f;
  t218_n42: return -0.31674981117248535f;
  t218_n43: return -0.0008764421218074858f;
  t218_n44: return -0.2010090947151184f;
  t218_n45: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t218_n47;
  t218_n46: return 0.03755487501621246f;
  t218_n47: return -0.013118123635649681f;
  t218_n48: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t218_n56;
  t218_n49: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t218_n55;
  t218_n50: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t218_n52;
  t218_n51: return 0.11662889271974564f;
  t218_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t218_n54;
  t218_n53: return 0.10743087530136108f;
  t218_n54: return 0.10628576576709747f;
  t218_n55: return 0.022534204646945f;
  t218_n56: return 0.013848211616277695f;
  t218_n57: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t218_n107;
  t218_n58: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t218_n90;
  t218_n59: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t218_n61;
  t218_n60: return 0.025468936190009117f;
  t218_n61: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t218_n75;
  t218_n62: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t218_n72;
  t218_n63: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t218_n65;
  t218_n64: return -0.03316589072346687f;
  t218_n65: if (E.getNumReferences() >= 3330095104 /*16206.0*/) goto t218_n69;
  t218_n66: if (E.getNumReferences() >= 3321384960 /*7946.0*/) goto t218_n68;
  t218_n67: return 0.0055386065505445f;
  t218_n68: return 0.08419424295425415f;
  t218_n69: if (E.getNumReferences() >= 3350798592 /*94786.0*/) goto t218_n71;
  t218_n70: return -0.10097400844097137f;
  t218_n71: return 0.0014118080725893378f;
  t218_n72: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t218_n74;
  t218_n73: return 0.02790191024541855f;
  t218_n74: return 0.14212140440940857f;
  t218_n75: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t218_n85;
  t218_n76: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t218_n84;
  t218_n77: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t218_n81;
  t218_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t218_n80;
  t218_n79: return 0.10882753878831863f;
  t218_n80: return 0.038555167615413666f;
  t218_n81: if (E.getFractionNameInContext() >= 3182156850 /*0.0839160829782486*/) goto t218_n83;
  t218_n82: return -0.057741858065128326f;
  t218_n83: return 0.3084905445575714f;
  t218_n84: return 0.018892919644713402f;
  t218_n85: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t218_n89;
  t218_n86: if (E.getNumReferences() >= 3254255616 /*31.0*/) goto t218_n88;
  t218_n87: return -0.08112737536430359f;
  t218_n88: return -0.6216436624526978f;
  t218_n89: return -0.013038813136518002f;
  t218_n90: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t218_n106;
  t218_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t218_n93;
  t218_n92: return 0.018457861617207527f;
  t218_n93: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t218_n105;
  t218_n94: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t218_n96;
  t218_n95: return -0.027369387447834015f;
  t218_n96: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t218_n100;
  t218_n97: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t218_n99;
  t218_n98: return 0.05871729180216789f;
  t218_n99: return -0.04338842257857323f;
  t218_n100: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t218_n102;
  t218_n101: return 0.04808541387319565f;
  t218_n102: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t218_n104;
  t218_n103: return 0.0148008419200778f;
  t218_n104: return -0.27932724356651306f;
  t218_n105: return 0.05604347586631775f;
  t218_n106: return -0.06712494790554047f;
  t218_n107: return 0.036549389362335205f;
  t218_n108: return -0.0009173642029054463f;
  t218_n109: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t218_n115;
  t218_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n112;
  t218_n111: return -0.1438738852739334f;
  t218_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t218_n114;
  t218_n113: return 0.002945809857919812f;
  t218_n114: return 0.019829943776130676f;
  t218_n115: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t218_n117;
  t218_n116: return 0.10969644784927368f;
  t218_n117: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t218_n141;
  t218_n118: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t218_n132;
  t218_n119: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t218_n121;
  t218_n120: return 0.016734670847654343f;
  t218_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t218_n123;
  t218_n122: return 0.005270833615213633f;
  t218_n123: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t218_n125;
  t218_n124: return 0.087054543197155f;
  t218_n125: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t218_n131;
  t218_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t218_n128;
  t218_n127: return -0.028459656983613968f;
  t218_n128: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t218_n130;
  t218_n129: return 0.08710368722677231f;
  t218_n130: return 0.03617284446954727f;
  t218_n131: return -0.12269314378499985f;
  t218_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n140;
  t218_n133: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t218_n137;
  t218_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t218_n136;
  t218_n135: return -0.08900338411331177f;
  t218_n136: return 0.11427266895771027f;
  t218_n137: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t218_n139;
  t218_n138: return -0.25414830446243286f;
  t218_n139: return 0.10312934964895248f;
  t218_n140: return 0.0775347650051117f;
  t218_n141: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t218_n143;
  t218_n142: return 0.03718556836247444f;
  t218_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n145;
  t218_n144: return -0.42723432183265686f;
  t218_n145: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t218_n147;
  t218_n146: return -0.06460016220808029f;
  t218_n147: return 0.08363334834575653f;
  t218_n148: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t218_n218;
  t218_n149: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t218_n151;
  t218_n150: return 0.08598975837230682f;
  t218_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t218_n213;
  t218_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t218_n176;
  t218_n153: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t218_n155;
  t218_n154: return -0.020593145862221718f;
  t218_n155: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t218_n165;
  t218_n156: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t218_n164;
  t218_n157: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t218_n159;
  t218_n158: return -0.02546512894332409f;
  t218_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t218_n161;
  t218_n160: return -0.054789211601018906f;
  t218_n161: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t218_n163;
  t218_n162: return 0.10810854285955429f;
  t218_n163: return 0.06560314446687698f;
  t218_n164: return 0.019271621480584145f;
  t218_n165: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t218_n169;
  t218_n166: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t218_n168;
  t218_n167: return 0.10513680428266525f;
  t218_n168: return 0.00948649737983942f;
  t218_n169: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t218_n175;
  t218_n170: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t218_n172;
  t218_n171: return 0.05000218749046326f;
  t218_n172: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t218_n174;
  t218_n173: return 0.10615776479244232f;
  t218_n174: return 0.06540066003799438f;
  t218_n175: return -0.07735264301300049f;
  t218_n176: if (E.getFractionNameInContext() >= 3164058662 /*0.018518518656492233*/) goto t218_n192;
  t218_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_TopLevel))) goto t218_n191;
  t218_n178: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t218_n186;
  t218_n179: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t218_n185;
  t218_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t218_n184;
  t218_n181: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t218_n183;
  t218_n182: return -0.012093956582248211f;
  t218_n183: return 0.0981588289141655f;
  t218_n184: return 0.07316364347934723f;
  t218_n185: return 0.016102192923426628f;
  t218_n186: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t218_n190;
  t218_n187: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t218_n189;
  t218_n188: return 0.020272042602300644f;
  t218_n189: return -0.07124879956245422f;
  t218_n190: return -0.001902248477563262f;
  t218_n191: return 0.028370654210448265f;
  t218_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n212;
  t218_n193: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t218_n211;
  t218_n194: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t218_n210;
  t218_n195: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t218_n209;
  t218_n196: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t218_n198;
  t218_n197: return -0.07001741230487823f;
  t218_n198: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t218_n208;
  t218_n199: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t218_n205;
  t218_n200: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t218_n204;
  t218_n201: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t218_n203;
  t218_n202: return -0.08560650795698166f;
  t218_n203: return 0.042431846261024475f;
  t218_n204: return 0.05316401645541191f;
  t218_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n207;
  t218_n206: return -0.06384143233299255f;
  t218_n207: return -0.005799292586743832f;
  t218_n208: return 0.10275030136108398f;
  t218_n209: return -0.10356276482343674f;
  t218_n210: return 0.047779109328985214f;
  t218_n211: return -0.05972979590296745f;
  t218_n212: return -0.005238381680101156f;
  t218_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t218_n217;
  t218_n214: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t218_n216;
  t218_n215: return -0.008105081506073475f;
  t218_n216: return -0.3682255744934082f;
  t218_n217: return 0.006928531918674707f;
  t218_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t218_n250;
  t218_n219: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t218_n247;
  t218_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n238;
  t218_n221: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t218_n231;
  t218_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n224;
  t218_n223: return -0.2446300983428955f;
  t218_n224: if (E.getNumReferences() >= 3283533824 /*365.5*/) goto t218_n226;
  t218_n225: return -0.015675639733672142f;
  t218_n226: if (E.getNumReferences() >= 3284189184 /*385.5*/) goto t218_n228;
  t218_n227: return -1.0542540550231934f;
  t218_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n230;
  t218_n229: return -0.13985460996627808f;
  t218_n230: return 0.08776383846998215f;
  t218_n231: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t218_n237;
  t218_n232: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t218_n234;
  t218_n233: return 0.12282302975654602f;
  t218_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n236;
  t218_n235: return -0.13939835131168365f;
  t218_n236: return -0.023904383182525635f;
  t218_n237: return -0.10666576772928238f;
  t218_n238: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t218_n242;
  t218_n239: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t218_n241;
  t218_n240: return -0.06338576227426529f;
  t218_n241: return 0.03383218124508858f;
  t218_n242: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t218_n244;
  t218_n243: return 0.12158642709255219f;
  t218_n244: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t218_n246;
  t218_n245: return 0.0784241259098053f;
  t218_n246: return -0.8485381603240967f;
  t218_n247: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t218_n249;
  t218_n248: return -0.0027056222315877676f;
  t218_n249: return 0.027394382283091545f;
  t218_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n252;
  t218_n251: return -0.021863743662834167f;
  t218_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t218_n254;
  t218_n253: return 0.0040427870117127895f;
  t218_n254: return 0.01936095952987671f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree219(const Example& E) {
  t219_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t219_n240;
  t219_n1: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t219_n3;
  t219_n2: return -0.0001821554178604856f;
  t219_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t219_n53;
  t219_n4: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t219_n10;
  t219_n5: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t219_n7;
  t219_n6: return -0.0034516144078224897f;
  t219_n7: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t219_n9;
  t219_n8: return 0.12791599333286285f;
  t219_n9: return -0.10115126520395279f;
  t219_n10: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t219_n14;
  t219_n11: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t219_n13;
  t219_n12: return 0.024970028549432755f;
  t219_n13: return -0.00726738478988409f;
  t219_n14: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t219_n38;
  t219_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t219_n25;
  t219_n16: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t219_n20;
  t219_n17: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t219_n19;
  t219_n18: return 0.04061947390437126f;
  t219_n19: return 0.11304660886526108f;
  t219_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t219_n24;
  t219_n21: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t219_n23;
  t219_n22: return 0.02484029345214367f;
  t219_n23: return 0.10700327903032303f;
  t219_n24: return 0.13226298987865448f;
  t219_n25: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t219_n37;
  t219_n26: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t219_n28;
  t219_n27: return 0.031953904777765274f;
  t219_n28: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t219_n34;
  t219_n29: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t219_n33;
  t219_n30: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t219_n32;
  t219_n31: return 0.06252136826515198f;
  t219_n32: return 0.09989733248949051f;
  t219_n33: return 0.11357545107603073f;
  t219_n34: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t219_n36;
  t219_n35: return -0.2491654008626938f;
  t219_n36: return 0.1528625786304474f;
  t219_n37: return 0.04021335020661354f;
  t219_n38: if (E.getNumReferences() >= 3264086016 /*71.0*/) goto t219_n48;
  t219_n39: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t219_n45;
  t219_n40: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t219_n44;
  t219_n41: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t219_n43;
  t219_n42: return -0.03235962241888046f;
  t219_n43: return -0.8637441396713257f;
  t219_n44: return 0.017222750931978226f;
  t219_n45: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t219_n47;
  t219_n46: return -0.10024746507406235f;
  t219_n47: return -0.820609986782074f;
  t219_n48: if (E.getNumReferences() >= 3280814080 /*282.5*/) goto t219_n52;
  t219_n49: if (E.getNumReferences() >= 3280715776 /*279.5*/) goto t219_n51;
  t219_n50: return 0.07704594731330872f;
  t219_n51: return 0.11179936677217484f;
  t219_n52: return -0.018672650679945946f;
  t219_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t219_n55;
  t219_n54: return -0.00875724758952856f;
  t219_n55: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t219_n231;
  t219_n56: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t219_n102;
  t219_n57: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t219_n73;
  t219_n58: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t219_n68;
  t219_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t219_n67;
  t219_n60: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t219_n64;
  t219_n61: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t219_n63;
  t219_n62: return 0.11842133849859238f;
  t219_n63: return -0.03764206916093826f;
  t219_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t219_n66;
  t219_n65: return -0.18901698291301727f;
  t219_n66: return 0.035983163863420486f;
  t219_n67: return 0.1590130627155304f;
  t219_n68: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t219_n72;
  t219_n69: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t219_n71;
  t219_n70: return -0.024048859253525734f;
  t219_n71: return 0.10743884742259979f;
  t219_n72: return -0.15759088099002838f;
  t219_n73: if (E.getFractionNameInContext() >= 3208152317 /*0.7207792401313782*/) goto t219_n99;
  t219_n74: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t219_n80;
  t219_n75: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t219_n79;
  t219_n76: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t219_n78;
  t219_n77: return -0.08994075655937195f;
  t219_n78: return 0.07066556066274643f;
  t219_n79: return 0.11506112664937973f;
  t219_n80: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t219_n98;
  t219_n81: if (E.getNumNameInContext() >= 3225419776 /*3.0*/) goto t219_n91;
  t219_n82: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t219_n84;
  t219_n83: return -0.2636973261833191f;
  t219_n84: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t219_n88;
  t219_n85: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t219_n87;
  t219_n86: return -0.11416883766651154f;
  t219_n87: return 0.1182965561747551f;
  t219_n88: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t219_n90;
  t219_n89: return -0.17406749725341797f;
  t219_n90: return 0.09260845184326172f;
  t219_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t219_n95;
  t219_n92: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t219_n94;
  t219_n93: return 0.0026234148535877466f;
  t219_n94: return 1.864938735961914f;
  t219_n95: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t219_n97;
  t219_n96: return 0.17300787568092346f;
  t219_n97: return 0.0863301157951355f;
  t219_n98: return -0.4493461549282074f;
  t219_n99: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t219_n101;
  t219_n100: return -0.06498347967863083f;
  t219_n101: return -0.5274680852890015f;
  t219_n102: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t219_n192;
  t219_n103: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t219_n191;
  t219_n104: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t219_n180;
  t219_n105: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t219_n177;
  t219_n106: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t219_n132;
  t219_n107: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t219_n109;
  t219_n108: return -0.09005290269851685f;
  t219_n109: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t219_n127;
  t219_n110: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t219_n118;
  t219_n111: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t219_n117;
  t219_n112: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t219_n116;
  t219_n113: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t219_n115;
  t219_n114: return -0.03514654561877251f;
  t219_n115: return 0.08834719657897949f;
  t219_n116: return 0.122235007584095f;
  t219_n117: return 0.12396177649497986f;
  t219_n118: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t219_n126;
  t219_n119: if (E.getNumReferences() >= 3247964160 /*19.0*/) goto t219_n123;
  t219_n120: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t219_n122;
  t219_n121: return -0.004970814101397991f;
  t219_n122: return -0.4942832589149475f;
  t219_n123: if (E.getNumReferences() >= 3262382080 /*61.0*/) goto t219_n125;
  t219_n124: return 0.3533018231391907f;
  t219_n125: return 0.01903945952653885f;
  t219_n126: return 0.12315621227025986f;
  t219_n127: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t219_n131;
  t219_n128: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t219_n130;
  t219_n129: return 0.11122771352529526f;
  t219_n130: return 0.03969987854361534f;
  t219_n131: return 0.04994199424982071f;
  t219_n132: if (E.getNumReferences() >= 3266969600 /*93.0*/) goto t219_n174;
  t219_n133: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t219_n141;
  t219_n134: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t219_n140;
  t219_n135: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t219_n139;
  t219_n136: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t219_n138;
  t219_n137: return 0.18188031017780304f;
  t219_n138: return 0.13469667732715607f;
  t219_n139: return 0.0072936262004077435f;
  t219_n140: return 0.058480747044086456f;
  t219_n141: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t219_n157;
  t219_n142: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t219_n156;
  t219_n143: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t219_n153;
  t219_n144: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t219_n152;
  t219_n145: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t219_n149;
  t219_n146: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t219_n148;
  t219_n147: return -0.19220396876335144f;
  t219_n148: return 0.12025395035743713f;
  t219_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t219_n151;
  t219_n150: return -0.26199033856391907f;
  t219_n151: return 0.11355345696210861f;
  t219_n152: return 0.011006336659193039f;
  t219_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t219_n155;
  t219_n154: return -0.22721299529075623f;
  t219_n155: return 0.11164649575948715f;
  t219_n156: return 0.12477650493383408f;
  t219_n157: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t219_n165;
  t219_n158: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t219_n164;
  t219_n159: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t219_n161;
  t219_n160: return -0.08792800456285477f;
  t219_n161: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t219_n163;
  t219_n162: return 0.11058959364891052f;
  t219_n163: return -0.005888373591005802f;
  t219_n164: return 0.07338651269674301f;
  t219_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t219_n167;
  t219_n166: return 0.05268118530511856f;
  t219_n167: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t219_n173;
  t219_n168: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t219_n172;
  t219_n169: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t219_n171;
  t219_n170: return 0.28828829526901245f;
  t219_n171: return -0.03282530978322029f;
  t219_n172: return -0.18033123016357422f;
  t219_n173: return 0.112153559923172f;
  t219_n174: if (E.getNumReferences() >= 3278602240 /*235.5*/) goto t219_n176;
  t219_n175: return 0.16186118125915527f;
  t219_n176: return -0.3228762447834015f;
  t219_n177: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t219_n179;
  t219_n178: return -0.06051163002848625f;
  t219_n179: return 0.08450291305780411f;
  t219_n180: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t219_n184;
  t219_n181: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t219_n183;
  t219_n182: return -0.03710944205522537f;
  t219_n183: return 0.057244379073381424f;
  t219_n184: if (E.getNumReferences() >= 3256090624 /*37.0*/) goto t219_n190;
  t219_n185: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t219_n189;
  t219_n186: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t219_n188;
  t219_n187: return 0.10188979655504227f;
  t219_n188: return 0.15654230117797852f;
  t219_n189: return 0.39183929562568665f;
  t219_n190: return -0.016224386170506477f;
  t219_n191: return 0.15803325176239014f;
  t219_n192: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t219_n208;
  t219_n193: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t219_n197;
  t219_n194: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t219_n196;
  t219_n195: return 0.10547942668199539f;
  t219_n196: return 0.00895602535456419f;
  t219_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t219_n207;
  t219_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t219_n202;
  t219_n199: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t219_n201;
  t219_n200: return 0.01232109684497118f;
  t219_n201: return -0.08836278319358826f;
  t219_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t219_n204;
  t219_n203: return 0.032361067831516266f;
  t219_n204: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t219_n206;
  t219_n205: return -0.06348571181297302f;
  t219_n206: return -0.00571161974221468f;
  t219_n207: return 0.04980679973959923f;
  t219_n208: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t219_n210;
  t219_n209: return -0.005034881643950939f;
  t219_n210: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t219_n230;
  t219_n211: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t219_n221;
  t219_n212: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t219_n214;
  t219_n213: return -0.18581613898277283f;
  t219_n214: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t219_n220;
  t219_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t219_n217;
  t219_n216: return 0.0007513374439440668f;
  t219_n217: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t219_n219;
  t219_n218: return 0.11148256808519363f;
  t219_n219: return 0.035485148429870605f;
  t219_n220: return 0.028150243684649467f;
  t219_n221: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t219_n229;
  t219_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t219_n224;
  t219_n223: return -0.007021393161267042f;
  t219_n224: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t219_n228;
  t219_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t219_n227;
  t219_n226: return 0.07314834743738174f;
  t219_n227: return 0.10464126616716385f;
  t219_n228: return 0.06871715188026428f;
  t219_n229: return 0.0008462637197226286f;
  t219_n230: return 0.015543865971267223f;
  t219_n231: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t219_n233;
  t219_n232: return -0.023045750334858894f;
  t219_n233: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t219_n237;
  t219_n234: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t219_n236;
  t219_n235: return -0.008111266419291496f;
  t219_n236: return 0.09154843538999557f;
  t219_n237: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t219_n239;
  t219_n238: return 0.050165604799985886f;
  t219_n239: return -0.23697957396507263f;
  t219_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t219_n252;
  t219_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t219_n243;
  t219_n242: return 0.00017717760056257248f;
  t219_n243: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t219_n251;
  t219_n244: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t219_n246;
  t219_n245: return 0.009251846000552177f;
  t219_n246: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t219_n248;
  t219_n247: return -0.01056070625782013f;
  t219_n248: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t219_n250;
  t219_n249: return 0.029902417212724686f;
  t219_n250: return 0.13893242180347443f;
  t219_n251: return 0.028144868090748787f;
  t219_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t219_n254;
  t219_n253: return 0.0041206455789506435f;
  t219_n254: return -0.04926171153783798f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree220(const Example& E) {
  t220_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t220_n2;
  t220_n1: return -0.00039792247116565704f;
  t220_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t220_n254;
  t220_n3: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t220_n251;
  t220_n4: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t220_n214;
  t220_n5: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t220_n211;
  t220_n6: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t220_n54;
  t220_n7: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t220_n11;
  t220_n8: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t220_n10;
  t220_n9: return -0.0026838532648980618f;
  t220_n10: return -0.11409082263708115f;
  t220_n11: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t220_n53;
  t220_n12: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t220_n52;
  t220_n13: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t220_n49;
  t220_n14: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t220_n18;
  t220_n15: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t220_n17;
  t220_n16: return 0.008295510895550251f;
  t220_n17: return -0.14574690163135529f;
  t220_n18: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t220_n46;
  t220_n19: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t220_n37;
  t220_n20: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t220_n22;
  t220_n21: return -0.005648232996463776f;
  t220_n22: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t220_n28;
  t220_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t220_n25;
  t220_n24: return -0.17879435420036316f;
  t220_n25: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t220_n27;
  t220_n26: return 0.30434170365333557f;
  t220_n27: return -0.2572804391384125f;
  t220_n28: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t220_n32;
  t220_n29: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t220_n31;
  t220_n30: return -0.00603916822001338f;
  t220_n31: return 0.08098165690898895f;
  t220_n32: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t220_n36;
  t220_n33: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t220_n35;
  t220_n34: return -0.04495048150420189f;
  t220_n35: return -0.1602644920349121f;
  t220_n36: return 0.02225261740386486f;
  t220_n37: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t220_n45;
  t220_n38: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t220_n40;
  t220_n39: return 0.008279803209006786f;
  t220_n40: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t220_n42;
  t220_n41: return 0.054717451333999634f;
  t220_n42: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t220_n44;
  t220_n43: return -0.10674639791250229f;
  t220_n44: return 0.1334517002105713f;
  t220_n45: return -0.03418433293700218f;
  t220_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t220_n48;
  t220_n47: return -0.31169945001602173f;
  t220_n48: return -0.004659230820834637f;
  t220_n49: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t220_n51;
  t220_n50: return 0.1709892898797989f;
  t220_n51: return 0.034074001014232635f;
  t220_n52: return 0.10354278981685638f;
  t220_n53: return -0.23692895472049713f;
  t220_n54: if (E.getFractionNameInContext() >= 3174441576 /*0.04446640610694885*/) goto t220_n56;
  t220_n55: return 0.10381477326154709f;
  t220_n56: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t220_n66;
  t220_n57: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t220_n59;
  t220_n58: return 0.002135467017069459f;
  t220_n59: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t220_n61;
  t220_n60: return 0.032920271158218384f;
  t220_n61: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t220_n65;
  t220_n62: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n64;
  t220_n63: return 0.0932723805308342f;
  t220_n64: return 0.9119345545768738f;
  t220_n65: return -0.6637743711471558f;
  t220_n66: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t220_n112;
  t220_n67: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t220_n71;
  t220_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t220_n70;
  t220_n69: return -0.052114252001047134f;
  t220_n70: return 0.03449907526373863f;
  t220_n71: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t220_n97;
  t220_n72: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t220_n96;
  t220_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t220_n79;
  t220_n74: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n78;
  t220_n75: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t220_n77;
  t220_n76: return -0.020783016458153725f;
  t220_n77: return -0.06999318301677704f;
  t220_n78: return -0.31771567463874817f;
  t220_n79: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n89;
  t220_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t220_n88;
  t220_n81: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t220_n87;
  t220_n82: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t220_n86;
  t220_n83: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t220_n85;
  t220_n84: return -0.2776235342025757f;
  t220_n85: return -1.4751695394515991f;
  t220_n86: return -0.08745601028203964f;
  t220_n87: return 0.06770291179418564f;
  t220_n88: return -0.0713869258761406f;
  t220_n89: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t220_n93;
  t220_n90: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t220_n92;
  t220_n91: return 0.17777684330940247f;
  t220_n92: return 0.03699979558587074f;
  t220_n93: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t220_n95;
  t220_n94: return -0.08057555556297302f;
  t220_n95: return 0.025061286985874176f;
  t220_n96: return 0.011392625980079174f;
  t220_n97: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t220_n111;
  t220_n98: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t220_n100;
  t220_n99: return -0.004708685912191868f;
  t220_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t220_n108;
  t220_n101: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t220_n105;
  t220_n102: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t220_n104;
  t220_n103: return 0.032141491770744324f;
  t220_n104: return -0.3371149003505707f;
  t220_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t220_n107;
  t220_n106: return -0.017873302102088928f;
  t220_n107: return 0.2165490984916687f;
  t220_n108: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t220_n110;
  t220_n109: return -0.0025177858769893646f;
  t220_n110: return 0.11568940430879593f;
  t220_n111: return -7.288752385647967e-05f;
  t220_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t220_n122;
  t220_n113: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t220_n115;
  t220_n114: return -0.009212293662130833f;
  t220_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n119;
  t220_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t220_n118;
  t220_n117: return 0.02660413645207882f;
  t220_n118: return -0.48822909593582153f;
  t220_n119: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t220_n121;
  t220_n120: return 0.10296011716127396f;
  t220_n121: return 0.1865645796060562f;
  t220_n122: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t220_n136;
  t220_n123: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t220_n129;
  t220_n124: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t220_n128;
  t220_n125: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t220_n127;
  t220_n126: return 0.012266159988939762f;
  t220_n127: return -0.22249174118041992f;
  t220_n128: return 0.27322259545326233f;
  t220_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t220_n133;
  t220_n130: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t220_n132;
  t220_n131: return -0.13957525789737701f;
  t220_n132: return 0.08689767122268677f;
  t220_n133: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t220_n135;
  t220_n134: return 0.11738239973783493f;
  t220_n135: return 0.14584662020206451f;
  t220_n136: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t220_n210;
  t220_n137: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t220_n175;
  t220_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n144;
  t220_n139: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t220_n141;
  t220_n140: return -0.27075400948524475f;
  t220_n141: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t220_n143;
  t220_n142: return -0.3023747503757477f;
  t220_n143: return 0.10601159930229187f;
  t220_n144: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t220_n174;
  t220_n145: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t220_n153;
  t220_n146: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t220_n148;
  t220_n147: return -0.0016729553462937474f;
  t220_n148: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t220_n150;
  t220_n149: return 0.008107904344797134f;
  t220_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t220_n152;
  t220_n151: return 0.02342240698635578f;
  t220_n152: return 0.12559647858142853f;
  t220_n153: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t220_n169;
  t220_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t220_n162;
  t220_n155: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t220_n161;
  t220_n156: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t220_n160;
  t220_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n159;
  t220_n158: return -0.18318387866020203f;
  t220_n159: return 0.07615115493535995f;
  t220_n160: return 0.1003166064620018f;
  t220_n161: return 0.01786096766591072f;
  t220_n162: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t220_n166;
  t220_n163: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t220_n165;
  t220_n164: return 0.009153751656413078f;
  t220_n165: return -0.27910611033439636f;
  t220_n166: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n168;
  t220_n167: return 0.06621900200843811f;
  t220_n168: return -0.08021058142185211f;
  t220_n169: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t220_n173;
  t220_n170: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t220_n172;
  t220_n171: return 0.01373088825494051f;
  t220_n172: return 0.11001362651586533f;
  t220_n173: return -0.0312834307551384f;
  t220_n174: return -0.0023098227102309465f;
  t220_n175: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t220_n181;
  t220_n176: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t220_n180;
  t220_n177: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n179;
  t220_n178: return 0.08507951349020004f;
  t220_n179: return 0.13689199090003967f;
  t220_n180: return -0.09404734522104263f;
  t220_n181: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t220_n199;
  t220_n182: if (E.getFractionNameInContext() >= 3194215998 /*0.22252747416496277*/) goto t220_n186;
  t220_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n185;
  t220_n184: return -0.04907083138823509f;
  t220_n185: return 0.03464636206626892f;
  t220_n186: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t220_n190;
  t220_n187: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t220_n189;
  t220_n188: return 0.06590726226568222f;
  t220_n189: return 0.1210516095161438f;
  t220_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t220_n192;
  t220_n191: return -0.023656781762838364f;
  t220_n192: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t220_n194;
  t220_n193: return -0.13618016242980957f;
  t220_n194: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t220_n198;
  t220_n195: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::GlobalScope))) goto t220_n197;
  t220_n196: return -0.19035956263542175f;
  t220_n197: return 0.034897707402706146f;
  t220_n198: return 0.04904879629611969f;
  t220_n199: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t220_n203;
  t220_n200: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t220_n202;
  t220_n201: return -0.29061537981033325f;
  t220_n202: return 0.05720806121826172f;
  t220_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_TopLevel))) goto t220_n205;
  t220_n204: return 0.009540022350847721f;
  t220_n205: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t220_n209;
  t220_n206: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t220_n208;
  t220_n207: return 0.02348393015563488f;
  t220_n208: return 0.13256993889808655f;
  t220_n209: return -0.08418741822242737f;
  t220_n210: return 0.09974436461925507f;
  t220_n211: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n213;
  t220_n212: return -0.0013942912919446826f;
  t220_n213: return -0.02302723191678524f;
  t220_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n216;
  t220_n215: return -0.14467893540859222f;
  t220_n216: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t220_n226;
  t220_n217: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t220_n219;
  t220_n218: return 0.0006179400952532887f;
  t220_n219: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t220_n225;
  t220_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t220_n224;
  t220_n221: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t220_n223;
  t220_n222: return -0.37766242027282715f;
  t220_n223: return 0.17491638660430908f;
  t220_n224: return 0.09803344309329987f;
  t220_n225: return 0.015516201965510845f;
  t220_n226: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t220_n246;
  t220_n227: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t220_n245;
  t220_n228: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t220_n236;
  t220_n229: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t220_n231;
  t220_n230: return -0.00448828749358654f;
  t220_n231: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t220_n233;
  t220_n232: return 0.14297643303871155f;
  t220_n233: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t220_n235;
  t220_n234: return -0.5691443681716919f;
  t220_n235: return 0.09608425945043564f;
  t220_n236: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t220_n242;
  t220_n237: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t220_n241;
  t220_n238: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t220_n240;
  t220_n239: return -0.030014030635356903f;
  t220_n240: return -0.012472604401409626f;
  t220_n241: return 0.012828816659748554f;
  t220_n242: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t220_n244;
  t220_n243: return 0.011372246779501438f;
  t220_n244: return -0.031184367835521698f;
  t220_n245: return 0.0011110324412584305f;
  t220_n246: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t220_n250;
  t220_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t220_n249;
  t220_n248: return 0.040323346853256226f;
  t220_n249: return 0.1516694277524948f;
  t220_n250: return 0.0025434517301619053f;
  t220_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t220_n253;
  t220_n252: return -0.006142050959169865f;
  t220_n253: return 0.0067800129763782024f;
  t220_n254: return 0.0015335313510149717f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree221(const Example& E) {
  t221_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t221_n60;
  t221_n1: if (E.getNumReferences() >= 3364277024 /*276057.0*/) goto t221_n47;
  t221_n2: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t221_n46;
  t221_n3: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t221_n23;
  t221_n4: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t221_n22;
  t221_n5: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t221_n9;
  t221_n6: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t221_n8;
  t221_n7: return -0.0001653348299441859f;
  t221_n8: return -0.15426090359687805f;
  t221_n9: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t221_n15;
  t221_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t221_n14;
  t221_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n13;
  t221_n12: return -0.3185185492038727f;
  t221_n13: return 0.02793322317302227f;
  t221_n14: return 0.05531566962599754f;
  t221_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n17;
  t221_n16: return -0.14886420965194702f;
  t221_n17: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t221_n21;
  t221_n18: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t221_n20;
  t221_n19: return 0.10316061228513718f;
  t221_n20: return 0.13187500834465027f;
  t221_n21: return -0.15309591591358185f;
  t221_n22: return -0.15487496554851532f;
  t221_n23: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t221_n25;
  t221_n24: return -0.026221467182040215f;
  t221_n25: if (E.getNumReferences() >= 3360630592 /*212125.0*/) goto t221_n27;
  t221_n26: return 0.020543979480862617f;
  t221_n27: if (E.getNumReferences() >= 3362054368 /*234371.5*/) goto t221_n35;
  t221_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n32;
  t221_n29: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t221_n31;
  t221_n30: return -0.16487832367420197f;
  t221_n31: return -0.37464889883995056f;
  t221_n32: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t221_n34;
  t221_n33: return -0.07690413296222687f;
  t221_n34: return 0.050962481647729874f;
  t221_n35: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t221_n43;
  t221_n36: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t221_n42;
  t221_n37: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t221_n39;
  t221_n38: return 0.024895524606108665f;
  t221_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n41;
  t221_n40: return -0.13308651745319366f;
  t221_n41: return 0.12054913491010666f;
  t221_n42: return -0.046330761164426804f;
  t221_n43: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t221_n45;
  t221_n44: return -0.05824539065361023f;
  t221_n45: return -0.34977343678474426f;
  t221_n46: return -0.053247928619384766f;
  t221_n47: if (E.getNumReferences() >= 3364286400 /*276350.0*/) goto t221_n59;
  t221_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t221_n56;
  t221_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n53;
  t221_n50: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t221_n52;
  t221_n51: return -0.11040276288986206f;
  t221_n52: return 1.6554880142211914f;
  t221_n53: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t221_n55;
  t221_n54: return 0.14888836443424225f;
  t221_n55: return 0.13903680443763733f;
  t221_n56: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t221_n58;
  t221_n57: return 0.0685037299990654f;
  t221_n58: return 0.15359334647655487f;
  t221_n59: return 0.006048195064067841f;
  t221_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t221_n250;
  t221_n61: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t221_n249;
  t221_n62: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t221_n248;
  t221_n63: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t221_n243;
  t221_n64: if (E.getNumReferences() >= 3318066176 /*6325.5*/) goto t221_n242;
  t221_n65: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t221_n215;
  t221_n66: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t221_n190;
  t221_n67: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t221_n129;
  t221_n68: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t221_n100;
  t221_n69: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t221_n81;
  t221_n70: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t221_n80;
  t221_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t221_n73;
  t221_n72: return 0.008394908159971237f;
  t221_n73: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t221_n79;
  t221_n74: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t221_n76;
  t221_n75: return -0.008262685500085354f;
  t221_n76: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t221_n78;
  t221_n77: return 0.10384111106395721f;
  t221_n78: return 0.13106904923915863f;
  t221_n79: return 0.10106290876865387f;
  t221_n80: return 0.10435730963945389f;
  t221_n81: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t221_n85;
  t221_n82: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t221_n84;
  t221_n83: return -0.00901998020708561f;
  t221_n84: return 0.09171377867460251f;
  t221_n85: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t221_n91;
  t221_n86: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t221_n90;
  t221_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t221_n89;
  t221_n88: return -0.9241102933883667f;
  t221_n89: return -0.01634824648499489f;
  t221_n90: return -0.06938845664262772f;
  t221_n91: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t221_n93;
  t221_n92: return 0.011974548920989037f;
  t221_n93: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t221_n95;
  t221_n94: return -0.18180090188980103f;
  t221_n95: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t221_n99;
  t221_n96: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t221_n98;
  t221_n97: return -0.008902976289391518f;
  t221_n98: return 0.22102059423923492f;
  t221_n99: return -0.2182898372411728f;
  t221_n100: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t221_n106;
  t221_n101: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t221_n103;
  t221_n102: return -0.00610053725540638f;
  t221_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t221_n105;
  t221_n104: return -0.07731970399618149f;
  t221_n105: return 0.02396160550415516f;
  t221_n106: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t221_n128;
  t221_n107: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t221_n117;
  t221_n108: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t221_n110;
  t221_n109: return -0.02276836521923542f;
  t221_n110: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t221_n116;
  t221_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t221_n113;
  t221_n112: return -0.033079467713832855f;
  t221_n113: if (E.getFractionNameInContext() >= 3196898509 /*0.2750000059604645*/) goto t221_n115;
  t221_n114: return 0.13690167665481567f;
  t221_n115: return 0.12183929979801178f;
  t221_n116: return -0.3112029433250427f;
  t221_n117: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t221_n123;
  t221_n118: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t221_n120;
  t221_n119: return -0.10507190227508545f;
  t221_n120: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t221_n122;
  t221_n121: return -0.060545533895492554f;
  t221_n122: return 0.10721100121736526f;
  t221_n123: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t221_n125;
  t221_n124: return -0.310007780790329f;
  t221_n125: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t221_n127;
  t221_n126: return 0.10025772452354431f;
  t221_n127: return -0.05257324129343033f;
  t221_n128: return 0.018622759729623795f;
  t221_n129: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t221_n185;
  t221_n130: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t221_n152;
  t221_n131: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t221_n141;
  t221_n132: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t221_n134;
  t221_n133: return -0.09582401067018509f;
  t221_n134: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t221_n140;
  t221_n135: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t221_n139;
  t221_n136: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t221_n138;
  t221_n137: return 0.06776463240385056f;
  t221_n138: return 0.16969332098960876f;
  t221_n139: return 0.02778102084994316f;
  t221_n140: return -0.11192207038402557f;
  t221_n141: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t221_n143;
  t221_n142: return 0.012464029714465141f;
  t221_n143: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t221_n145;
  t221_n144: return -0.018082033842802048f;
  t221_n145: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t221_n151;
  t221_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t221_n150;
  t221_n147: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t221_n149;
  t221_n148: return -0.495918333530426f;
  t221_n149: return -0.22914309799671173f;
  t221_n150: return -0.03150416910648346f;
  t221_n151: return -0.05742708593606949f;
  t221_n152: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t221_n154;
  t221_n153: return 0.1447019726037979f;
  t221_n154: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t221_n178;
  t221_n155: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t221_n175;
  t221_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t221_n164;
  t221_n157: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t221_n159;
  t221_n158: return 0.10813159495592117f;
  t221_n159: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t221_n163;
  t221_n160: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t221_n162;
  t221_n161: return -0.19065000116825104f;
  t221_n162: return -0.30224278569221497f;
  t221_n163: return 0.0004051426949445158f;
  t221_n164: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t221_n166;
  t221_n165: return 0.14346112310886383f;
  t221_n166: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t221_n174;
  t221_n167: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t221_n173;
  t221_n168: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t221_n170;
  t221_n169: return 0.1015269085764885f;
  t221_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t221_n172;
  t221_n171: return 0.1344384253025055f;
  t221_n172: return 0.4022846519947052f;
  t221_n173: return -0.11797884106636047f;
  t221_n174: return -0.06516928970813751f;
  t221_n175: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t221_n177;
  t221_n176: return -0.025377152487635612f;
  t221_n177: return 0.11668708920478821f;
  t221_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t221_n180;
  t221_n179: return -0.011576078832149506f;
  t221_n180: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t221_n184;
  t221_n181: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t221_n183;
  t221_n182: return -0.10234245657920837f;
  t221_n183: return 0.1618717759847641f;
  t221_n184: return -0.12514440715312958f;
  t221_n185: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t221_n189;
  t221_n186: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t221_n188;
  t221_n187: return 0.0572141632437706f;
  t221_n188: return 0.12617388367652893f;
  t221_n189: return 0.06637255102396011f;
  t221_n190: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t221_n208;
  t221_n191: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t221_n207;
  t221_n192: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t221_n200;
  t221_n193: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t221_n195;
  t221_n194: return 0.05767625942826271f;
  t221_n195: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t221_n197;
  t221_n196: return 0.03369816020131111f;
  t221_n197: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t221_n199;
  t221_n198: return 0.12473277002573013f;
  t221_n199: return -0.02738378383219242f;
  t221_n200: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t221_n206;
  t221_n201: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t221_n203;
  t221_n202: return -0.174307718873024f;
  t221_n203: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t221_n205;
  t221_n204: return 0.06534843891859055f;
  t221_n205: return -0.37985387444496155f;
  t221_n206: return 0.09506898373365402f;
  t221_n207: return 0.0887003168463707f;
  t221_n208: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t221_n210;
  t221_n209: return 0.01852632500231266f;
  t221_n210: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t221_n214;
  t221_n211: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t221_n213;
  t221_n212: return -0.005323866847902536f;
  t221_n213: return -0.21379709243774414f;
  t221_n214: return 0.1545107364654541f;
  t221_n215: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t221_n219;
  t221_n216: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t221_n218;
  t221_n217: return 0.01643209159374237f;
  t221_n218: return -0.007350473199039698f;
  t221_n219: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t221_n231;
  t221_n220: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t221_n230;
  t221_n221: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t221_n229;
  t221_n222: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t221_n226;
  t221_n223: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t221_n225;
  t221_n224: return 0.19743916392326355f;
  t221_n225: return 0.08353351056575775f;
  t221_n226: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t221_n228;
  t221_n227: return -0.08187117427587509f;
  t221_n228: return 0.10632085055112839f;
  t221_n229: return 0.009704835712909698f;
  t221_n230: return 0.18137910962104797f;
  t221_n231: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t221_n239;
  t221_n232: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t221_n236;
  t221_n233: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t221_n235;
  t221_n234: return -0.1532406508922577f;
  t221_n235: return 0.046513963490724564f;
  t221_n236: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t221_n238;
  t221_n237: return -0.14649564027786255f;
  t221_n238: return 0.11621727794408798f;
  t221_n239: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t221_n241;
  t221_n240: return 0.11520954221487045f;
  t221_n241: return -0.162383571267128f;
  t221_n242: return 0.07627541571855545f;
  t221_n243: if (E.getNumReferences() >= 3331493888 /*18760.0*/) goto t221_n247;
  t221_n244: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t221_n246;
  t221_n245: return -0.1269865185022354f;
  t221_n246: return 0.1320735365152359f;
  t221_n247: return 0.001969671342521906f;
  t221_n248: return -0.0009088979568332434f;
  t221_n249: return 0.007191415876150131f;
  t221_n250: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t221_n252;
  t221_n251: return 0.004077967721968889f;
  t221_n252: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t221_n254;
  t221_n253: return -0.024038221687078476f;
  t221_n254: return -0.14730314910411835f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree222(const Example& E) {
  t222_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t222_n6;
  t222_n1: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t222_n3;
  t222_n2: return 0.0002755201421678066f;
  t222_n3: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t222_n5;
  t222_n4: return -0.00233064335770905f;
  t222_n5: return -0.10764269530773163f;
  t222_n6: if (E.getNumReferences() >= 3284484096 /*394.5*/) goto t222_n246;
  t222_n7: if (E.getNumReferences() >= 3282845696 /*344.5*/) goto t222_n237;
  t222_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n182;
  t222_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t222_n11;
  t222_n10: return -0.00023406533000525087f;
  t222_n11: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t222_n155;
  t222_n12: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t222_n148;
  t222_n13: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t222_n147;
  t222_n14: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t222_n138;
  t222_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n21;
  t222_n16: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t222_n20;
  t222_n17: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t222_n19;
  t222_n18: return 0.02349381521344185f;
  t222_n19: return -0.05622875317931175f;
  t222_n20: return 0.01499202474951744f;
  t222_n21: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t222_n123;
  t222_n22: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t222_n116;
  t222_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t222_n69;
  t222_n24: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t222_n26;
  t222_n25: return -0.016615211963653564f;
  t222_n26: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t222_n68;
  t222_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n39;
  t222_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t222_n38;
  t222_n29: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t222_n35;
  t222_n30: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t222_n34;
  t222_n31: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t222_n33;
  t222_n32: return -0.0314890518784523f;
  t222_n33: return -0.20683911442756653f;
  t222_n34: return 0.04381876066327095f;
  t222_n35: if (E.getFractionNameInContext() >= 3196336196 /*0.25824177265167236*/) goto t222_n37;
  t222_n36: return 0.12049593031406403f;
  t222_n37: return -0.06326846778392792f;
  t222_n38: return 0.004926701076328754f;
  t222_n39: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t222_n67;
  t222_n40: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t222_n62;
  t222_n41: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t222_n43;
  t222_n42: return 0.009414917789399624f;
  t222_n43: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t222_n57;
  t222_n44: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t222_n48;
  t222_n45: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t222_n47;
  t222_n46: return 0.02509445883333683f;
  t222_n47: return 0.14255599677562714f;
  t222_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t222_n50;
  t222_n49: return 0.03303971141576767f;
  t222_n50: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t222_n52;
  t222_n51: return 0.1308843493461609f;
  t222_n52: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t222_n54;
  t222_n53: return -0.08438923209905624f;
  t222_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n56;
  t222_n55: return -0.10478118062019348f;
  t222_n56: return 0.11675456911325455f;
  t222_n57: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t222_n59;
  t222_n58: return -0.09992201626300812f;
  t222_n59: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t222_n61;
  t222_n60: return 0.036688707768917084f;
  t222_n61: return -0.015503969974815845f;
  t222_n62: if (E.getFractionNameInContext() >= 3200224622 /*0.3741258978843689*/) goto t222_n66;
  t222_n63: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t222_n65;
  t222_n64: return -0.1517750769853592f;
  t222_n65: return 0.10892336070537567f;
  t222_n66: return -0.08261110633611679f;
  t222_n67: return 0.11365781724452972f;
  t222_n68: return -0.005111553706228733f;
  t222_n69: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t222_n109;
  t222_n70: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t222_n72;
  t222_n71: return 0.0007099146605469286f;
  t222_n72: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t222_n106;
  t222_n73: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t222_n81;
  t222_n74: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t222_n80;
  t222_n75: if (E.getNumReferences() >= 3275358208 /*186.0*/) goto t222_n77;
  t222_n76: return 0.010984467342495918f;
  t222_n77: if (E.getNumReferences() >= 3275948032 /*195.0*/) goto t222_n79;
  t222_n78: return 0.11988359689712524f;
  t222_n79: return 0.07615648210048676f;
  t222_n80: return -0.018661586567759514f;
  t222_n81: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t222_n87;
  t222_n82: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t222_n86;
  t222_n83: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t222_n85;
  t222_n84: return 0.04894096776843071f;
  t222_n85: return 0.005601688288152218f;
  t222_n86: return 0.04553278908133507f;
  t222_n87: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t222_n105;
  t222_n88: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t222_n104;
  t222_n89: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t222_n93;
  t222_n90: if (E.getFractionNameInContext() >= 3190176728 /*0.16233766078948975*/) goto t222_n92;
  t222_n91: return 0.06594423949718475f;
  t222_n92: return -0.19707775115966797f;
  t222_n93: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t222_n95;
  t222_n94: return 0.017461225390434265f;
  t222_n95: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t222_n101;
  t222_n96: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t222_n98;
  t222_n97: return 0.086665578186512f;
  t222_n98: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t222_n100;
  t222_n99: return 0.09338954091072083f;
  t222_n100: return -0.01715177297592163f;
  t222_n101: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t222_n103;
  t222_n102: return 0.07820523530244827f;
  t222_n103: return -0.012931782752275467f;
  t222_n104: return 0.031060298904776573f;
  t222_n105: return -0.012953672558069229f;
  t222_n106: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t222_n108;
  t222_n107: return 0.025585094466805458f;
  t222_n108: return 0.1406758427619934f;
  t222_n109: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t222_n111;
  t222_n110: return 0.024442991241812706f;
  t222_n111: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t222_n115;
  t222_n112: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t222_n114;
  t222_n113: return 0.10812868922948837f;
  t222_n114: return 0.038240786641836166f;
  t222_n115: return -0.12457948178052902f;
  t222_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n118;
  t222_n117: return -0.14854592084884644f;
  t222_n118: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t222_n120;
  t222_n119: return -0.033778995275497437f;
  t222_n120: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t222_n122;
  t222_n121: return 0.039669305086135864f;
  t222_n122: return -0.13795116543769836f;
  t222_n123: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t222_n129;
  t222_n124: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t222_n128;
  t222_n125: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t222_n127;
  t222_n126: return -0.27998170256614685f;
  t222_n127: return 0.0937797948718071f;
  t222_n128: return -0.6411069631576538f;
  t222_n129: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t222_n137;
  t222_n130: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t222_n136;
  t222_n131: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t222_n133;
  t222_n132: return -0.01307238545268774f;
  t222_n133: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t222_n135;
  t222_n134: return -0.16572889685630798f;
  t222_n135: return 0.025549283251166344f;
  t222_n136: return 0.06507062166929245f;
  t222_n137: return 0.024324947968125343f;
  t222_n138: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t222_n142;
  t222_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t222_n141;
  t222_n140: return 0.012230987660586834f;
  t222_n141: return 0.04295452684164047f;
  t222_n142: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t222_n144;
  t222_n143: return -0.04241050034761429f;
  t222_n144: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t222_n146;
  t222_n145: return 0.10735981166362762f;
  t222_n146: return 0.008717245422303677f;
  t222_n147: return 0.06878174841403961f;
  t222_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t222_n150;
  t222_n149: return 0.00020979452528990805f;
  t222_n150: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t222_n154;
  t222_n151: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t222_n153;
  t222_n152: return 0.00883433036506176f;
  t222_n153: return -0.03811460733413696f;
  t222_n154: return 0.05215366929769516f;
  t222_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t222_n181;
  t222_n156: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t222_n180;
  t222_n157: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t222_n159;
  t222_n158: return 0.0027560272719711065f;
  t222_n159: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t222_n179;
  t222_n160: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t222_n178;
  t222_n161: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t222_n165;
  t222_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t222_n164;
  t222_n163: return -0.024048687890172005f;
  t222_n164: return 0.10044620931148529f;
  t222_n165: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t222_n177;
  t222_n166: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t222_n176;
  t222_n167: if (E.getFractionNameInContext() >= 3206195882 /*0.6041666269302368*/) goto t222_n173;
  t222_n168: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t222_n172;
  t222_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n171;
  t222_n170: return -0.009338184259831905f;
  t222_n171: return 0.10947418212890625f;
  t222_n172: return -0.03423375263810158f;
  t222_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n175;
  t222_n174: return -0.006278427317738533f;
  t222_n175: return 0.13594599068164825f;
  t222_n176: return -0.01841380074620247f;
  t222_n177: return -0.019720595329999924f;
  t222_n178: return -0.10260768979787827f;
  t222_n179: return 0.11187566816806793f;
  t222_n180: return -0.005639359354972839f;
  t222_n181: return 0.006306375376880169f;
  t222_n182: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t222_n184;
  t222_n183: return 0.004580333828926086f;
  t222_n184: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t222_n232;
  t222_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n187;
  t222_n186: return -0.03282254561781883f;
  t222_n187: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t222_n231;
  t222_n188: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t222_n228;
  t222_n189: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t222_n191;
  t222_n190: return 0.009918938390910625f;
  t222_n191: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t222_n211;
  t222_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t222_n196;
  t222_n193: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t222_n195;
  t222_n194: return -0.08820035308599472f;
  t222_n195: return 0.02383938990533352f;
  t222_n196: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t222_n198;
  t222_n197: return -0.06015649437904358f;
  t222_n198: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t222_n200;
  t222_n199: return 0.0704190731048584f;
  t222_n200: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t222_n208;
  t222_n201: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t222_n203;
  t222_n202: return -0.15376847982406616f;
  t222_n203: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t222_n205;
  t222_n204: return 0.05575910210609436f;
  t222_n205: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t222_n207;
  t222_n206: return -0.2453080415725708f;
  t222_n207: return -0.023289648815989494f;
  t222_n208: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t222_n210;
  t222_n209: return 0.09125611931085587f;
  t222_n210: return 0.04643759876489639f;
  t222_n211: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t222_n213;
  t222_n212: return -0.04431487247347832f;
  t222_n213: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t222_n215;
  t222_n214: return 0.09544757753610611f;
  t222_n215: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t222_n219;
  t222_n216: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t222_n218;
  t222_n217: return 0.02043445222079754f;
  t222_n218: return -0.02877577766776085f;
  t222_n219: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t222_n227;
  t222_n220: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t222_n226;
  t222_n221: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t222_n225;
  t222_n222: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t222_n224;
  t222_n223: return 0.06259645521640778f;
  t222_n224: return 0.11273790150880814f;
  t222_n225: return 0.050521012395620346f;
  t222_n226: return -0.1938258856534958f;
  t222_n227: return -0.04166071489453316f;
  t222_n228: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t222_n230;
  t222_n229: return -0.25593486428260803f;
  t222_n230: return 0.06559047847986221f;
  t222_n231: return 0.09119356423616409f;
  t222_n232: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t222_n236;
  t222_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n235;
  t222_n234: return -0.3122004568576813f;
  t222_n235: return 0.05514898523688316f;
  t222_n236: return -0.014190156944096088f;
  t222_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t222_n243;
  t222_n238: if (E.getNumReferences() >= 3283140608 /*353.5*/) goto t222_n242;
  t222_n239: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t222_n241;
  t222_n240: return -0.23541228473186493f;
  t222_n241: return -0.7637856602668762f;
  t222_n242: return -0.056589335203170776f;
  t222_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t222_n245;
  t222_n244: return -0.03405805304646492f;
  t222_n245: return 0.10946132242679596f;
  t222_n246: if (E.getNumReferences() >= 3284516864 /*395.5*/) goto t222_n250;
  t222_n247: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t222_n249;
  t222_n248: return 0.11958757787942886f;
  t222_n249: return 0.01531969290226698f;
  t222_n250: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t222_n252;
  t222_n251: return 0.008205443620681763f;
  t222_n252: if (E.getNumReferences() >= 3330413824 /*16650.5*/) goto t222_n254;
  t222_n253: return -0.890437126159668f;
  t222_n254: return -0.13600538671016693f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree223(const Example& E) {
  t223_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t223_n2;
  t223_n1: return -0.00013553023745771497f;
  t223_n2: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t223_n4;
  t223_n3: return 0.008415651507675648f;
  t223_n4: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t223_n254;
  t223_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t223_n59;
  t223_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t223_n56;
  t223_n7: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t223_n31;
  t223_n8: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t223_n20;
  t223_n9: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t223_n15;
  t223_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t223_n14;
  t223_n11: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t223_n13;
  t223_n12: return -0.11891920864582062f;
  t223_n13: return 0.11799807101488113f;
  t223_n14: return -0.020498398691415787f;
  t223_n15: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t223_n17;
  t223_n16: return 0.0006893305690027773f;
  t223_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t223_n19;
  t223_n18: return -0.012132704257965088f;
  t223_n19: return 0.007243538741022348f;
  t223_n20: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t223_n30;
  t223_n21: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t223_n25;
  t223_n22: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t223_n24;
  t223_n23: return -0.14933660626411438f;
  t223_n24: return -0.009115448221564293f;
  t223_n25: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t223_n27;
  t223_n26: return 0.006078374106436968f;
  t223_n27: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t223_n29;
  t223_n28: return -0.21049636602401733f;
  t223_n29: return 0.0753159373998642f;
  t223_n30: return -0.12810997664928436f;
  t223_n31: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t223_n55;
  t223_n32: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t223_n54;
  t223_n33: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t223_n45;
  t223_n34: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t223_n42;
  t223_n35: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t223_n37;
  t223_n36: return 0.056508053094148636f;
  t223_n37: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t223_n39;
  t223_n38: return -0.07043551653623581f;
  t223_n39: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t223_n41;
  t223_n40: return 0.008885283954441547f;
  t223_n41: return -0.20324669778347015f;
  t223_n42: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t223_n44;
  t223_n43: return -0.4125797152519226f;
  t223_n44: return 0.07459061592817307f;
  t223_n45: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t223_n53;
  t223_n46: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t223_n48;
  t223_n47: return -0.07111575454473495f;
  t223_n48: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t223_n52;
  t223_n49: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t223_n51;
  t223_n50: return 0.1595248132944107f;
  t223_n51: return 0.12479205429553986f;
  t223_n52: return 0.09208323061466217f;
  t223_n53: return -0.0884842723608017f;
  t223_n54: return 0.11835812777280807f;
  t223_n55: return -0.018446223810315132f;
  t223_n56: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t223_n58;
  t223_n57: return -0.01981070078909397f;
  t223_n58: return 0.01511180866509676f;
  t223_n59: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t223_n253;
  t223_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t223_n154;
  t223_n61: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t223_n115;
  t223_n62: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t223_n78;
  t223_n63: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t223_n77;
  t223_n64: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t223_n76;
  t223_n65: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t223_n71;
  t223_n66: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t223_n70;
  t223_n67: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t223_n69;
  t223_n68: return -0.02195751667022705f;
  t223_n69: return 0.042820077389478683f;
  t223_n70: return -0.12897692620754242f;
  t223_n71: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t223_n75;
  t223_n72: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t223_n74;
  t223_n73: return 0.1582372635602951f;
  t223_n74: return 0.04735303297638893f;
  t223_n75: return 0.004817671608179808f;
  t223_n76: return -0.02411726862192154f;
  t223_n77: return 0.019665353000164032f;
  t223_n78: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t223_n112;
  t223_n79: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t223_n81;
  t223_n80: return -0.017386645078659058f;
  t223_n81: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t223_n83;
  t223_n82: return -0.015478917397558689f;
  t223_n83: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t223_n99;
  t223_n84: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t223_n94;
  t223_n85: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t223_n87;
  t223_n86: return -0.034144509583711624f;
  t223_n87: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t223_n93;
  t223_n88: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t223_n92;
  t223_n89: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t223_n91;
  t223_n90: return 0.087359219789505f;
  t223_n91: return 0.02312704548239708f;
  t223_n92: return 0.08504582941532135f;
  t223_n93: return -0.004078572615981102f;
  t223_n94: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t223_n96;
  t223_n95: return 0.045254696160554886f;
  t223_n96: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t223_n98;
  t223_n97: return 0.006869286764413118f;
  t223_n98: return 0.09170322865247726f;
  t223_n99: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t223_n111;
  t223_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t223_n106;
  t223_n101: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t223_n103;
  t223_n102: return -0.048218440264463425f;
  t223_n103: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t223_n105;
  t223_n104: return 0.03916395083069801f;
  t223_n105: return -0.02644231542944908f;
  t223_n106: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t223_n110;
  t223_n107: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t223_n109;
  t223_n108: return 0.001713335164822638f;
  t223_n109: return -0.11236933618783951f;
  t223_n110: return 0.025281723588705063f;
  t223_n111: return -0.09268968552350998f;
  t223_n112: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t223_n114;
  t223_n113: return 0.06234940513968468f;
  t223_n114: return 0.008132504299283028f;
  t223_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t223_n125;
  t223_n116: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t223_n118;
  t223_n117: return -0.0074704429134726524f;
  t223_n118: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t223_n122;
  t223_n119: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t223_n121;
  t223_n120: return 0.015046852640807629f;
  t223_n121: return -0.03418722003698349f;
  t223_n122: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t223_n124;
  t223_n123: return 0.12665051221847534f;
  t223_n124: return 0.053943078964948654f;
  t223_n125: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t223_n143;
  t223_n126: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t223_n142;
  t223_n127: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t223_n133;
  t223_n128: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t223_n130;
  t223_n129: return -0.14697310328483582f;
  t223_n130: if (E.getFractionNameInContext() >= 3172829657 /*0.03846153989434242*/) goto t223_n132;
  t223_n131: return 0.34135445952415466f;
  t223_n132: return -0.10866141319274902f;
  t223_n133: if (E.getFractionNameInContext() >= 3181279720 /*0.0773809552192688*/) goto t223_n139;
  t223_n134: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t223_n138;
  t223_n135: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t223_n137;
  t223_n136: return 0.1935759037733078f;
  t223_n137: return -0.2135554552078247f;
  t223_n138: return -0.10596663504838943f;
  t223_n139: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t223_n141;
  t223_n140: return -0.11457612365484238f;
  t223_n141: return 0.10491056740283966f;
  t223_n142: return 0.015477171167731285f;
  t223_n143: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t223_n145;
  t223_n144: return 0.017307598143815994f;
  t223_n145: if (E.getNumReferences() >= 3283009536 /*349.5*/) goto t223_n151;
  t223_n146: if (E.getNumReferences() >= 3257663488 /*43.0*/) goto t223_n148;
  t223_n147: return -0.06359393149614334f;
  t223_n148: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t223_n150;
  t223_n149: return -0.2567616403102875f;
  t223_n150: return -0.5687853693962097f;
  t223_n151: if (E.getFractionNameInContext() >= 3188638956 /*0.13942307233810425*/) goto t223_n153;
  t223_n152: return 0.05628342181444168f;
  t223_n153: return 0.10856898128986359f;
  t223_n154: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t223_n220;
  t223_n155: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t223_n163;
  t223_n156: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t223_n158;
  t223_n157: return -0.0016479261685162783f;
  t223_n158: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t223_n160;
  t223_n159: return -0.02279423363506794f;
  t223_n160: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t223_n162;
  t223_n161: return 0.009845202788710594f;
  t223_n162: return -0.022364990785717964f;
  t223_n163: if (E.getFractionNameInContext() >= 3185414456 /*0.10818713903427124*/) goto t223_n165;
  t223_n164: return 0.2820756137371063f;
  t223_n165: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t223_n197;
  t223_n166: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t223_n190;
  t223_n167: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t223_n169;
  t223_n168: return -0.03573235496878624f;
  t223_n169: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t223_n189;
  t223_n170: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t223_n178;
  t223_n171: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t223_n177;
  t223_n172: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t223_n174;
  t223_n173: return 0.012085122056305408f;
  t223_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t223_n176;
  t223_n175: return 0.05887996405363083f;
  t223_n176: return 0.1161474958062172f;
  t223_n177: return 0.00236877822317183f;
  t223_n178: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t223_n184;
  t223_n179: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t223_n181;
  t223_n180: return 0.035585250705480576f;
  t223_n181: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t223_n183;
  t223_n182: return 0.13530349731445312f;
  t223_n183: return 0.05890602245926857f;
  t223_n184: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t223_n188;
  t223_n185: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t223_n187;
  t223_n186: return 0.1159546747803688f;
  t223_n187: return 0.08535636961460114f;
  t223_n188: return 0.025359109044075012f;
  t223_n189: return 0.13258279860019684f;
  t223_n190: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t223_n196;
  t223_n191: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t223_n195;
  t223_n192: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t223_n194;
  t223_n193: return 0.1471613049507141f;
  t223_n194: return 0.11613444983959198f;
  t223_n195: return 0.04770829901099205f;
  t223_n196: return 0.050935354083776474f;
  t223_n197: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t223_n213;
  t223_n198: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t223_n208;
  t223_n199: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t223_n205;
  t223_n200: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t223_n204;
  t223_n201: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t223_n203;
  t223_n202: return -0.049837253987789154f;
  t223_n203: return 0.0761389210820198f;
  t223_n204: return -0.23657365143299103f;
  t223_n205: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t223_n207;
  t223_n206: return 0.007671264465898275f;
  t223_n207: return -0.5490806698799133f;
  t223_n208: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t223_n210;
  t223_n209: return 0.031262923032045364f;
  t223_n210: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t223_n212;
  t223_n211: return 0.02079875022172928f;
  t223_n212: return 0.16973024606704712f;
  t223_n213: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t223_n219;
  t223_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t223_n216;
  t223_n215: return 0.05719251558184624f;
  t223_n216: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t223_n218;
  t223_n217: return 0.1566375494003296f;
  t223_n218: return -0.06038927659392357f;
  t223_n219: return -0.03993406146764755f;
  t223_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t223_n246;
  t223_n221: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t223_n235;
  t223_n222: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t223_n224;
  t223_n223: return 0.0173647478222847f;
  t223_n224: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t223_n234;
  t223_n225: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t223_n229;
  t223_n226: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t223_n228;
  t223_n227: return 0.039008911699056625f;
  t223_n228: return -0.028209509328007698f;
  t223_n229: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t223_n231;
  t223_n230: return 0.25201791524887085f;
  t223_n231: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t223_n233;
  t223_n232: return -0.10123596340417862f;
  t223_n233: return 0.14974401891231537f;
  t223_n234: return 0.08287204802036285f;
  t223_n235: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t223_n243;
  t223_n236: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t223_n242;
  t223_n237: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t223_n241;
  t223_n238: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t223_n240;
  t223_n239: return -0.4965924024581909f;
  t223_n240: return -0.12436933070421219f;
  t223_n241: return -1.5305638313293457f;
  t223_n242: return 0.055399827659130096f;
  t223_n243: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t223_n245;
  t223_n244: return 0.015212073922157288f;
  t223_n245: return -0.08948317915201187f;
  t223_n246: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t223_n252;
  t223_n247: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t223_n249;
  t223_n248: return -0.02301814965903759f;
  t223_n249: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t223_n251;
  t223_n250: return -0.46864551305770874f;
  t223_n251: return -0.22789709270000458f;
  t223_n252: return 0.022583208978176117f;
  t223_n253: return 0.0068273586221039295f;
  t223_n254: return -0.10102644562721252f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree224(const Example& E) {
  t224_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t224_n2;
  t224_n1: return -0.00041693137609399855f;
  t224_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t224_n20;
  t224_n3: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t224_n5;
  t224_n4: return -0.00044204553705640137f;
  t224_n5: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t224_n15;
  t224_n6: if (E.getNumReferences() >= 3272769536 /*146.5*/) goto t224_n14;
  t224_n7: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t224_n13;
  t224_n8: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t224_n12;
  t224_n9: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t224_n11;
  t224_n10: return 0.30356279015541077f;
  t224_n11: return 0.09528201073408127f;
  t224_n12: return 0.08238273859024048f;
  t224_n13: return 0.1296083629131317f;
  t224_n14: return 0.010980186983942986f;
  t224_n15: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t224_n19;
  t224_n16: if (E.getNumReferences() >= 3280322560 /*267.5*/) goto t224_n18;
  t224_n17: return 0.3564110994338989f;
  t224_n18: return 0.07714936882257462f;
  t224_n19: return -0.001019699382595718f;
  t224_n20: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t224_n140;
  t224_n21: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t224_n95;
  t224_n22: if (E.getFractionNameInContext() >= 3207044730 /*0.6547619104385376*/) goto t224_n82;
  t224_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t224_n79;
  t224_n24: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t224_n76;
  t224_n25: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t224_n29;
  t224_n26: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t224_n28;
  t224_n27: return -0.08831804245710373f;
  t224_n28: return -0.023140905424952507f;
  t224_n29: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t224_n71;
  t224_n30: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t224_n70;
  t224_n31: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t224_n43;
  t224_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t224_n36;
  t224_n33: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t224_n35;
  t224_n34: return 0.05896640568971634f;
  t224_n35: return -0.01397353783249855f;
  t224_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t224_n38;
  t224_n37: return -0.12430107593536377f;
  t224_n38: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t224_n40;
  t224_n39: return 0.0682600736618042f;
  t224_n40: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t224_n42;
  t224_n41: return 0.12079473584890366f;
  t224_n42: return 0.016750480979681015f;
  t224_n43: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t224_n49;
  t224_n44: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t224_n46;
  t224_n45: return -0.07626499980688095f;
  t224_n46: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t224_n48;
  t224_n47: return -0.14946678280830383f;
  t224_n48: return -0.2935170829296112f;
  t224_n49: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t224_n69;
  t224_n50: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t224_n68;
  t224_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t224_n53;
  t224_n52: return -0.15496239066123962f;
  t224_n53: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t224_n55;
  t224_n54: return 0.009722551330924034f;
  t224_n55: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t224_n59;
  t224_n56: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t224_n58;
  t224_n57: return 0.5213702321052551f;
  t224_n58: return 0.09587391465902328f;
  t224_n59: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t224_n65;
  t224_n60: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t224_n62;
  t224_n61: return -0.024887703359127045f;
  t224_n62: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t224_n64;
  t224_n63: return 0.09794525802135468f;
  t224_n64: return 0.127449169754982f;
  t224_n65: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t224_n67;
  t224_n66: return 0.03629850596189499f;
  t224_n67: return -0.7467078566551208f;
  t224_n68: return 0.10577988624572754f;
  t224_n69: return -0.03229739889502525f;
  t224_n70: return 0.038486775010824203f;
  t224_n71: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t224_n73;
  t224_n72: return -0.17040137946605682f;
  t224_n73: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t224_n75;
  t224_n74: return -0.013785510323941708f;
  t224_n75: return 0.20416218042373657f;
  t224_n76: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t224_n78;
  t224_n77: return -0.033869966864585876f;
  t224_n78: return 0.010209036991000175f;
  t224_n79: if (E.getNumReferences() >= 3281797120 /*312.5*/) goto t224_n81;
  t224_n80: return -0.0014328755205497146f;
  t224_n81: return -0.02982790581882f;
  t224_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t224_n84;
  t224_n83: return -0.07793281227350235f;
  t224_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t224_n86;
  t224_n85: return 0.014565002173185349f;
  t224_n86: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t224_n90;
  t224_n87: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t224_n89;
  t224_n88: return 0.033509526401758194f;
  t224_n89: return 0.10085099190473557f;
  t224_n90: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t224_n94;
  t224_n91: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t224_n93;
  t224_n92: return 0.07460363954305649f;
  t224_n93: return -0.08900344371795654f;
  t224_n94: return -0.061569586396217346f;
  t224_n95: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t224_n123;
  t224_n96: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t224_n120;
  t224_n97: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t224_n111;
  t224_n98: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t224_n110;
  t224_n99: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t224_n103;
  t224_n100: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t224_n102;
  t224_n101: return -0.0063917976804077625f;
  t224_n102: return 0.39085328578948975f;
  t224_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t224_n105;
  t224_n104: return -0.11243487149477005f;
  t224_n105: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t224_n107;
  t224_n106: return 0.07282735407352448f;
  t224_n107: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t224_n109;
  t224_n108: return -0.02822585590183735f;
  t224_n109: return 0.03305704519152641f;
  t224_n110: return -0.0231174286454916f;
  t224_n111: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t224_n113;
  t224_n112: return 0.009703650139272213f;
  t224_n113: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t224_n119;
  t224_n114: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t224_n116;
  t224_n115: return -0.10788820683956146f;
  t224_n116: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t224_n118;
  t224_n117: return 0.045679111033678055f;
  t224_n118: return 0.10301650315523148f;
  t224_n119: return 0.004123122431337833f;
  t224_n120: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t224_n122;
  t224_n121: return 0.03814820945262909f;
  t224_n122: return -0.19603314995765686f;
  t224_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t224_n135;
  t224_n124: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t224_n134;
  t224_n125: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t224_n127;
  t224_n126: return 0.003874777816236019f;
  t224_n127: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t224_n131;
  t224_n128: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t224_n130;
  t224_n129: return 0.01165751926600933f;
  t224_n130: return -0.02214215323328972f;
  t224_n131: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t224_n133;
  t224_n132: return 0.028608225286006927f;
  t224_n133: return 0.16036640107631683f;
  t224_n134: return -0.00527187017723918f;
  t224_n135: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t224_n139;
  t224_n136: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t224_n138;
  t224_n137: return 0.05540882423520088f;
  t224_n138: return 0.02155112475156784f;
  t224_n139: return 0.007870725356042385f;
  t224_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type))) goto t224_n250;
  t224_n141: if (E.getFractionNameInContext() >= 3172638464 /*0.037749290466308594*/) goto t224_n143;
  t224_n142: return 0.21723778545856476f;
  t224_n143: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t224_n151;
  t224_n144: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t224_n146;
  t224_n145: return 0.0008029927266761661f;
  t224_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t224_n148;
  t224_n147: return 0.002862925175577402f;
  t224_n148: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t224_n150;
  t224_n149: return 0.06770546734333038f;
  t224_n150: return 0.1421830952167511f;
  t224_n151: if (E.getNumReferences() >= 3290882048 /*667.5*/) goto t224_n249;
  t224_n152: if (E.getNumReferences() >= 3290144768 /*622.5*/) goto t224_n244;
  t224_n153: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t224_n157;
  t224_n154: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t224_n156;
  t224_n155: return -0.009816483594477177f;
  t224_n156: return 0.006406731903553009f;
  t224_n157: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t224_n241;
  t224_n158: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t224_n238;
  t224_n159: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t224_n193;
  t224_n160: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t224_n192;
  t224_n161: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t224_n177;
  t224_n162: if (E.getFractionNameInContext() >= 3197124868 /*0.2817460298538208*/) goto t224_n164;
  t224_n163: return -0.011535017751157284f;
  t224_n164: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t224_n166;
  t224_n165: return 0.004105531144887209f;
  t224_n166: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t224_n170;
  t224_n167: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t224_n169;
  t224_n168: return 0.024914337322115898f;
  t224_n169: return 0.08672945946455002f;
  t224_n170: if (E.getNumReferences() >= 3259760640 /*51.0*/) goto t224_n174;
  t224_n171: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t224_n173;
  t224_n172: return -0.05453520268201828f;
  t224_n173: return -0.4311443567276001f;
  t224_n174: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t224_n176;
  t224_n175: return -0.12575547397136688f;
  t224_n176: return 0.11592676490545273f;
  t224_n177: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t224_n179;
  t224_n178: return -0.05494866520166397f;
  t224_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t224_n191;
  t224_n180: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t224_n184;
  t224_n181: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t224_n183;
  t224_n182: return 0.09973997622728348f;
  t224_n183: return -0.017724711447954178f;
  t224_n184: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t224_n186;
  t224_n185: return -0.03334587439894676f;
  t224_n186: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t224_n188;
  t224_n187: return -0.1455744057893753f;
  t224_n188: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t224_n190;
  t224_n189: return 0.005517642945051193f;
  t224_n190: return 0.12533631920814514f;
  t224_n191: return -0.11812203377485275f;
  t224_n192: return 0.025535574182868004f;
  t224_n193: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t224_n195;
  t224_n194: return 0.02442621812224388f;
  t224_n195: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t224_n233;
  t224_n196: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t224_n198;
  t224_n197: return -0.0175428856164217f;
  t224_n198: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t224_n212;
  t224_n199: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t224_n209;
  t224_n200: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t224_n204;
  t224_n201: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t224_n203;
  t224_n202: return -0.105989009141922f;
  t224_n203: return 0.11995866149663925f;
  t224_n204: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t224_n206;
  t224_n205: return 0.034156545996665955f;
  t224_n206: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t224_n208;
  t224_n207: return 0.42081257700920105f;
  t224_n208: return -0.11114784330129623f;
  t224_n209: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t224_n211;
  t224_n210: return -0.03940271586179733f;
  t224_n211: return 0.03591422364115715f;
  t224_n212: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t224_n214;
  t224_n213: return -0.023977363482117653f;
  t224_n214: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t224_n232;
  t224_n215: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t224_n217;
  t224_n216: return 0.04917886480689049f;
  t224_n217: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t224_n231;
  t224_n218: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t224_n226;
  t224_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t224_n221;
  t224_n220: return -0.08288555592298508f;
  t224_n221: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t224_n225;
  t224_n222: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t224_n224;
  t224_n223: return 0.17270909249782562f;
  t224_n224: return 0.13783012330532074f;
  t224_n225: return 0.07660984247922897f;
  t224_n226: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t224_n230;
  t224_n227: if (E.getNumReferences() >= 3272474624 /*142.0*/) goto t224_n229;
  t224_n228: return 0.10659964382648468f;
  t224_n229: return 0.12516556680202484f;
  t224_n230: return 0.07381417602300644f;
  t224_n231: return -0.005276094656437635f;
  t224_n232: return 0.04567953199148178f;
  t224_n233: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t224_n235;
  t224_n234: return 0.11819697916507721f;
  t224_n235: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t224_n237;
  t224_n236: return -0.02092440240085125f;
  t224_n237: return 0.08592236787080765f;
  t224_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t224_n240;
  t224_n239: return -0.057858046144247055f;
  t224_n240: return 0.00596039230003953f;
  t224_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t224_n243;
  t224_n242: return -0.00016640395915601403f;
  t224_n243: return 0.06130484864115715f;
  t224_n244: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t224_n246;
  t224_n245: return 0.06677452474832535f;
  t224_n246: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t224_n248;
  t224_n247: return 0.1267353743314743f;
  t224_n248: return 0.12387009710073471f;
  t224_n249: return -0.025612445548176765f;
  t224_n250: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t224_n254;
  t224_n251: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t224_n253;
  t224_n252: return -0.10477439314126968f;
  t224_n253: return 0.008687841705977917f;
  t224_n254: return 0.008628133684396744f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree225(const Example& E) {
  t225_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t225_n2;
  t225_n1: return -0.0004026921233162284f;
  t225_n2: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t225_n252;
  t225_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t225_n97;
  t225_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t225_n54;
  t225_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t225_n7;
  t225_n6: return -0.05419348552823067f;
  t225_n7: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t225_n47;
  t225_n8: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t225_n10;
  t225_n9: return 0.08974724262952805f;
  t225_n10: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t225_n12;
  t225_n11: return -0.002026994712650776f;
  t225_n12: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t225_n46;
  t225_n13: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t225_n45;
  t225_n14: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t225_n16;
  t225_n15: return 0.006299291737377644f;
  t225_n16: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t225_n44;
  t225_n17: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t225_n35;
  t225_n18: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t225_n22;
  t225_n19: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t225_n21;
  t225_n20: return 0.09159080684185028f;
  t225_n21: return -0.30143803358078003f;
  t225_n22: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t225_n24;
  t225_n23: return -0.2411331832408905f;
  t225_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t225_n30;
  t225_n25: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t225_n27;
  t225_n26: return -0.009148654527962208f;
  t225_n27: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t225_n29;
  t225_n28: return 0.05124487727880478f;
  t225_n29: return 0.13129055500030518f;
  t225_n30: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t225_n34;
  t225_n31: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t225_n33;
  t225_n32: return 0.12185166776180267f;
  t225_n33: return 0.1453852355480194f;
  t225_n34: return 0.023513611406087875f;
  t225_n35: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t225_n43;
  t225_n36: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t225_n38;
  t225_n37: return 0.02930971048772335f;
  t225_n38: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t225_n42;
  t225_n39: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t225_n41;
  t225_n40: return 0.1335591971874237f;
  t225_n41: return 0.09847506135702133f;
  t225_n42: return 0.04724752530455589f;
  t225_n43: return 0.1462949961423874f;
  t225_n44: return 0.07843880355358124f;
  t225_n45: return -0.030225463211536407f;
  t225_n46: return 0.0705362856388092f;
  t225_n47: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t225_n53;
  t225_n48: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t225_n52;
  t225_n49: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t225_n51;
  t225_n50: return 0.037574127316474915f;
  t225_n51: return 0.12322952598333359f;
  t225_n52: return -0.030511481687426567f;
  t225_n53: return 0.11081081628799438f;
  t225_n54: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t225_n92;
  t225_n55: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t225_n57;
  t225_n56: return 0.0012423185398802161f;
  t225_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t225_n61;
  t225_n58: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t225_n60;
  t225_n59: return -0.0069032516330480576f;
  t225_n60: return -0.07575489580631256f;
  t225_n61: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t225_n67;
  t225_n62: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t225_n66;
  t225_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n65;
  t225_n64: return -0.0189290102571249f;
  t225_n65: return 0.014584622345864773f;
  t225_n66: return 0.01469219196587801f;
  t225_n67: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t225_n83;
  t225_n68: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t225_n76;
  t225_n69: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t225_n71;
  t225_n70: return -0.008489387109875679f;
  t225_n71: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t225_n75;
  t225_n72: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t225_n74;
  t225_n73: return 0.06619751453399658f;
  t225_n74: return -0.11817464232444763f;
  t225_n75: return 0.04662589728832245f;
  t225_n76: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n80;
  t225_n77: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t225_n79;
  t225_n78: return 0.023945381864905357f;
  t225_n79: return 0.001420447719283402f;
  t225_n80: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t225_n82;
  t225_n81: return -0.04997643083333969f;
  t225_n82: return 0.12323883920907974f;
  t225_n83: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t225_n89;
  t225_n84: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t225_n88;
  t225_n85: if (E.getFractionNameInContext() >= 3196898509 /*0.2750000059604645*/) goto t225_n87;
  t225_n86: return -0.182523712515831f;
  t225_n87: return 0.04582159221172333f;
  t225_n88: return -0.17415906488895416f;
  t225_n89: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t225_n91;
  t225_n90: return -0.15690653026103973f;
  t225_n91: return -0.014064404182136059f;
  t225_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n96;
  t225_n93: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t225_n95;
  t225_n94: return -0.28551414608955383f;
  t225_n95: return -0.005642486270517111f;
  t225_n96: return 0.0029483886901289225f;
  t225_n97: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t225_n103;
  t225_n98: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t225_n100;
  t225_n99: return 0.00621398538351059f;
  t225_n100: if (E.getNumReferences() >= 3298598912 /*1253.0*/) goto t225_n102;
  t225_n101: return 0.015283055603504181f;
  t225_n102: return 0.13580237329006195f;
  t225_n103: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t225_n105;
  t225_n104: return 0.0021664074156433344f;
  t225_n105: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t225_n161;
  t225_n106: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t225_n116;
  t225_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n109;
  t225_n108: return -0.06652910262346268f;
  t225_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t225_n113;
  t225_n110: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t225_n112;
  t225_n111: return -0.01734686829149723f;
  t225_n112: return 0.0658038780093193f;
  t225_n113: if (E.getFractionNameInContext() >= 3198755986 /*0.3303571343421936*/) goto t225_n115;
  t225_n114: return 0.14624309539794922f;
  t225_n115: return -0.10055892169475555f;
  t225_n116: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t225_n126;
  t225_n117: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n125;
  t225_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n122;
  t225_n119: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t225_n121;
  t225_n120: return -0.12821899354457855f;
  t225_n121: return 0.09197284281253815f;
  t225_n122: if (E.getFractionNameInContext() >= 3166318220 /*0.022727273404598236*/) goto t225_n124;
  t225_n123: return 0.05337502434849739f;
  t225_n124: return -0.05235147476196289f;
  t225_n125: return 0.06299327313899994f;
  t225_n126: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t225_n160;
  t225_n127: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t225_n159;
  t225_n128: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t225_n134;
  t225_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n131;
  t225_n130: return -0.05782346799969673f;
  t225_n131: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t225_n133;
  t225_n132: return 0.05216582491993904f;
  t225_n133: return 0.006277839187532663f;
  t225_n134: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t225_n140;
  t225_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n137;
  t225_n136: return -0.1052820011973381f;
  t225_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n139;
  t225_n138: return 0.038397010415792465f;
  t225_n139: return 0.047252681106328964f;
  t225_n140: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t225_n144;
  t225_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n143;
  t225_n142: return -0.06736090034246445f;
  t225_n143: return 0.03718452900648117f;
  t225_n144: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t225_n154;
  t225_n145: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t225_n149;
  t225_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t225_n148;
  t225_n147: return -0.12235038727521896f;
  t225_n148: return 0.037440139800310135f;
  t225_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n151;
  t225_n150: return 0.04506669193506241f;
  t225_n151: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t225_n153;
  t225_n152: return 0.12771882116794586f;
  t225_n153: return 0.10093294829130173f;
  t225_n154: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t225_n156;
  t225_n155: return -0.049859918653964996f;
  t225_n156: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n158;
  t225_n157: return -0.011174547486007214f;
  t225_n158: return 0.05914413183927536f;
  t225_n159: return -0.035946767777204514f;
  t225_n160: return -0.04108542203903198f;
  t225_n161: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t225_n251;
  t225_n162: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t225_n204;
  t225_n163: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t225_n169;
  t225_n164: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t225_n168;
  t225_n165: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t225_n167;
  t225_n166: return -0.42261528968811035f;
  t225_n167: return 0.021698979660868645f;
  t225_n168: return -0.00972256064414978f;
  t225_n169: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t225_n193;
  t225_n170: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t225_n176;
  t225_n171: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t225_n173;
  t225_n172: return -0.015981368720531464f;
  t225_n173: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t225_n175;
  t225_n174: return -0.11353745311498642f;
  t225_n175: return 1.5122425556182861f;
  t225_n176: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t225_n186;
  t225_n177: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t225_n183;
  t225_n178: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t225_n182;
  t225_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t225_n181;
  t225_n180: return 0.09243466705083847f;
  t225_n181: return 0.10547050833702087f;
  t225_n182: return 0.14240363240242004f;
  t225_n183: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t225_n185;
  t225_n184: return 0.21497860550880432f;
  t225_n185: return -0.019253332167863846f;
  t225_n186: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t225_n192;
  t225_n187: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t225_n191;
  t225_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t225_n190;
  t225_n189: return 0.1124483123421669f;
  t225_n190: return 0.03376861661672592f;
  t225_n191: return -0.5285942554473877f;
  t225_n192: return 0.004785196855664253f;
  t225_n193: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t225_n203;
  t225_n194: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t225_n196;
  t225_n195: return 0.04837749898433685f;
  t225_n196: if (E.getNumReferences() >= 3326905856 /*13091.5*/) goto t225_n198;
  t225_n197: return 0.03707685321569443f;
  t225_n198: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t225_n200;
  t225_n199: return -0.04793039336800575f;
  t225_n200: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t225_n202;
  t225_n201: return -1.0700196027755737f;
  t225_n202: return -0.2783159613609314f;
  t225_n203: return 0.11085280030965805f;
  t225_n204: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t225_n206;
  t225_n205: return -0.014703444205224514f;
  t225_n206: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t225_n212;
  t225_n207: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t225_n209;
  t225_n208: return 0.012376487255096436f;
  t225_n209: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t225_n211;
  t225_n210: return -0.9549548029899597f;
  t225_n211: return -0.004831764381378889f;
  t225_n212: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t225_n246;
  t225_n213: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t225_n215;
  t225_n214: return 0.013288545422255993f;
  t225_n215: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t225_n231;
  t225_n216: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t225_n222;
  t225_n217: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t225_n221;
  t225_n218: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t225_n220;
  t225_n219: return 0.007420097943395376f;
  t225_n220: return -0.03330317139625549f;
  t225_n221: return 0.01608213409781456f;
  t225_n222: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t225_n224;
  t225_n223: return 0.086362823843956f;
  t225_n224: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t225_n228;
  t225_n225: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t225_n227;
  t225_n226: return 0.024036569520831108f;
  t225_n227: return 0.06926006823778152f;
  t225_n228: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t225_n230;
  t225_n229: return -0.05156402289867401f;
  t225_n230: return 0.029264070093631744f;
  t225_n231: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t225_n239;
  t225_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t225_n234;
  t225_n233: return 0.0023590305354446173f;
  t225_n234: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t225_n236;
  t225_n235: return 0.032324157655239105f;
  t225_n236: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t225_n238;
  t225_n237: return 0.05897108465433121f;
  t225_n238: return 0.10718254745006561f;
  t225_n239: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t225_n245;
  t225_n240: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t225_n244;
  t225_n241: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t225_n243;
  t225_n242: return 0.017962459474802017f;
  t225_n243: return 0.08184181153774261f;
  t225_n244: return -0.023072047159075737f;
  t225_n245: return 0.04024558141827583f;
  t225_n246: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t225_n248;
  t225_n247: return -0.04436153918504715f;
  t225_n248: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t225_n250;
  t225_n249: return -0.0023185336031019688f;
  t225_n250: return 0.019441744312644005f;
  t225_n251: return -0.11089348793029785f;
  t225_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t225_n254;
  t225_n253: return 0.0006207784754224122f;
  t225_n254: return 0.006397797726094723f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree226(const Example& E) {
  t226_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t226_n2;
  t226_n1: return -0.0003844082239083946f;
  t226_n2: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t226_n228;
  t226_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t226_n133;
  t226_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t226_n6;
  t226_n5: return -0.014828518033027649f;
  t226_n6: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t226_n62;
  t226_n7: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t226_n61;
  t226_n8: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t226_n60;
  t226_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t226_n29;
  t226_n10: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t226_n14;
  t226_n11: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t226_n13;
  t226_n12: return 0.00996264535933733f;
  t226_n13: return -0.03686362877488136f;
  t226_n14: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t226_n18;
  t226_n15: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t226_n17;
  t226_n16: return -0.017573101446032524f;
  t226_n17: return 0.018097981810569763f;
  t226_n18: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t226_n28;
  t226_n19: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t226_n23;
  t226_n20: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t226_n22;
  t226_n21: return 0.112259142100811f;
  t226_n22: return -0.14147934317588806f;
  t226_n23: if (E.getNumReferences() >= 3345564672 /*59716.0*/) goto t226_n25;
  t226_n24: return 0.007909215986728668f;
  t226_n25: if (E.getNumReferences() >= 3350732864 /*94272.5*/) goto t226_n27;
  t226_n26: return -0.37371906638145447f;
  t226_n27: return -0.06566277891397476f;
  t226_n28: return -0.008212021552026272f;
  t226_n29: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t226_n33;
  t226_n30: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t226_n32;
  t226_n31: return -0.19553181529045105f;
  t226_n32: return -0.017773257568478584f;
  t226_n33: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t226_n39;
  t226_n34: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t226_n38;
  t226_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n37;
  t226_n36: return -0.015065222047269344f;
  t226_n37: return 0.031052691861987114f;
  t226_n38: return 0.012925445102155209f;
  t226_n39: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n55;
  t226_n40: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t226_n54;
  t226_n41: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t226_n47;
  t226_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t226_n44;
  t226_n43: return -0.05169041082262993f;
  t226_n44: if (E.getNumReferences() >= 3341024768 /*41982.0*/) goto t226_n46;
  t226_n45: return 0.027697179466485977f;
  t226_n46: return -0.3458908200263977f;
  t226_n47: if (E.getNumReferences() >= 3349178240 /*82127.0*/) goto t226_n53;
  t226_n48: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t226_n50;
  t226_n49: return 0.007242890540510416f;
  t226_n50: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t226_n52;
  t226_n51: return -0.047956209629774094f;
  t226_n52: return -0.0009248547139577568f;
  t226_n53: return 0.07293177396059036f;
  t226_n54: return -0.20185261964797974f;
  t226_n55: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t226_n57;
  t226_n56: return 0.09703968465328217f;
  t226_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n59;
  t226_n58: return -0.13609786331653595f;
  t226_n59: return 0.007597741205245256f;
  t226_n60: return -0.03731757029891014f;
  t226_n61: return 0.0006218644557520747f;
  t226_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n64;
  t226_n63: return -0.005244175437837839f;
  t226_n64: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t226_n74;
  t226_n65: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t226_n67;
  t226_n66: return 0.005082614254206419f;
  t226_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n73;
  t226_n68: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t226_n72;
  t226_n69: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t226_n71;
  t226_n70: return -0.03302653506398201f;
  t226_n71: return 0.027036089450120926f;
  t226_n72: return -0.20686286687850952f;
  t226_n73: return -0.0015791732585057616f;
  t226_n74: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t226_n132;
  t226_n75: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t226_n123;
  t226_n76: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t226_n86;
  t226_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n79;
  t226_n78: return -0.010806579142808914f;
  t226_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t226_n83;
  t226_n80: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t226_n82;
  t226_n81: return -0.0322943776845932f;
  t226_n82: return 0.04419367387890816f;
  t226_n83: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t226_n85;
  t226_n84: return 0.016022823750972748f;
  t226_n85: return -0.00453121168538928f;
  t226_n86: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t226_n98;
  t226_n87: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t226_n95;
  t226_n88: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t226_n94;
  t226_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n93;
  t226_n90: if (E.getNumReferences() >= 3337056768 /*29625.0*/) goto t226_n92;
  t226_n91: return -0.005721133667975664f;
  t226_n92: return 0.025190774351358414f;
  t226_n93: return 0.009437482804059982f;
  t226_n94: return -0.2506735920906067f;
  t226_n95: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t226_n97;
  t226_n96: return 0.004476940259337425f;
  t226_n97: return -0.010279630310833454f;
  t226_n98: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t226_n100;
  t226_n99: return 0.010387888178229332f;
  t226_n100: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t226_n118;
  t226_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t226_n103;
  t226_n102: return 0.0015395913505926728f;
  t226_n103: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t226_n105;
  t226_n104: return -0.11793915927410126f;
  t226_n105: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t226_n115;
  t226_n106: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t226_n112;
  t226_n107: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t226_n111;
  t226_n108: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t226_n110;
  t226_n109: return 0.018941134214401245f;
  t226_n110: return 0.0922691598534584f;
  t226_n111: return -0.020354948937892914f;
  t226_n112: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t226_n114;
  t226_n113: return 0.08293256163597107f;
  t226_n114: return -0.04007941111922264f;
  t226_n115: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t226_n117;
  t226_n116: return -0.1395096629858017f;
  t226_n117: return 0.04756103828549385f;
  t226_n118: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t226_n122;
  t226_n119: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t226_n121;
  t226_n120: return 0.046595681458711624f;
  t226_n121: return -0.11872202157974243f;
  t226_n122: return -0.3646431565284729f;
  t226_n123: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t226_n131;
  t226_n124: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t226_n130;
  t226_n125: if (E.getNumReferences() >= 3221225472 /*2.0*/) goto t226_n129;
  t226_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t226_n128;
  t226_n127: return 0.07130115479230881f;
  t226_n128: return 0.17549104988574982f;
  t226_n129: return -0.10076551884412766f;
  t226_n130: return -0.013312102295458317f;
  t226_n131: return 0.12071315944194794f;
  t226_n132: return -0.007529229857027531f;
  t226_n133: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t226_n141;
  t226_n134: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t226_n140;
  t226_n135: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n137;
  t226_n136: return 0.006665167864412069f;
  t226_n137: if (E.getNumReferences() >= 3343124992 /*50186.0*/) goto t226_n139;
  t226_n138: return -0.007549751549959183f;
  t226_n139: return -0.23927143216133118f;
  t226_n140: return 0.013759750872850418f;
  t226_n141: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t226_n165;
  t226_n142: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t226_n162;
  t226_n143: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t226_n161;
  t226_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n154;
  t226_n145: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t226_n147;
  t226_n146: return -0.057169198989868164f;
  t226_n147: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n153;
  t226_n148: if (E.getFractionNameInContext() >= 3192983825 /*0.20416666567325592*/) goto t226_n152;
  t226_n149: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t226_n151;
  t226_n150: return -0.08518145978450775f;
  t226_n151: return -0.11149214953184128f;
  t226_n152: return 0.04292137548327446f;
  t226_n153: return 0.020467346534132957f;
  t226_n154: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n156;
  t226_n155: return -0.10286352783441544f;
  t226_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t226_n160;
  t226_n157: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t226_n159;
  t226_n158: return 0.14980411529541016f;
  t226_n159: return -0.06324842572212219f;
  t226_n160: return 0.21578651666641235f;
  t226_n161: return 0.0023337232414633036f;
  t226_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t226_n164;
  t226_n163: return 0.030741270631551743f;
  t226_n164: return 0.5131773948669434f;
  t226_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t226_n219;
  t226_n166: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t226_n174;
  t226_n167: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t226_n173;
  t226_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n170;
  t226_n169: return -0.06608108431100845f;
  t226_n170: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t226_n172;
  t226_n171: return 0.057961784303188324f;
  t226_n172: return -0.0489460714161396f;
  t226_n173: return 0.05979117006063461f;
  t226_n174: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t226_n186;
  t226_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n183;
  t226_n176: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t226_n178;
  t226_n177: return -0.12117858976125717f;
  t226_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n180;
  t226_n179: return -0.21536073088645935f;
  t226_n180: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t226_n182;
  t226_n181: return 0.10235033929347992f;
  t226_n182: return 0.0011616817209869623f;
  t226_n183: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t226_n185;
  t226_n184: return 0.0593855157494545f;
  t226_n185: return 0.0020391647703945637f;
  t226_n186: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t226_n218;
  t226_n187: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t226_n191;
  t226_n188: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t226_n190;
  t226_n189: return -0.0068013472482562065f;
  t226_n190: return -0.04609369486570358f;
  t226_n191: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t226_n213;
  t226_n192: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t226_n194;
  t226_n193: return -0.05506644397974014f;
  t226_n194: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t226_n196;
  t226_n195: return 0.03944976255297661f;
  t226_n196: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t226_n198;
  t226_n197: return -0.09335817396640778f;
  t226_n198: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t226_n204;
  t226_n199: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t226_n201;
  t226_n200: return 0.03357818350195885f;
  t226_n201: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t226_n203;
  t226_n202: return 0.11998552829027176f;
  t226_n203: return 0.037519849836826324f;
  t226_n204: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t226_n206;
  t226_n205: return -0.07825111597776413f;
  t226_n206: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t226_n212;
  t226_n207: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t226_n211;
  t226_n208: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t226_n210;
  t226_n209: return 0.05082237720489502f;
  t226_n210: return 0.07714352011680603f;
  t226_n211: return -0.6048369407653809f;
  t226_n212: return -0.059246648102998734f;
  t226_n213: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t226_n215;
  t226_n214: return 0.05891428142786026f;
  t226_n215: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t226_n217;
  t226_n216: return 0.10532720386981964f;
  t226_n217: return 0.07268206775188446f;
  t226_n218: return -0.06290575861930847f;
  t226_n219: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t226_n225;
  t226_n220: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t226_n224;
  t226_n221: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t226_n223;
  t226_n222: return 0.005189301911741495f;
  t226_n223: return -0.0036586266942322254f;
  t226_n224: return -0.11465869098901749f;
  t226_n225: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t226_n227;
  t226_n226: return 0.011640165001153946f;
  t226_n227: return 0.04442106932401657f;
  t226_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t226_n240;
  t226_n229: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t226_n231;
  t226_n230: return 0.002774714957922697f;
  t226_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t226_n239;
  t226_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t226_n234;
  t226_n233: return -0.01743835210800171f;
  t226_n234: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n236;
  t226_n235: return 0.13009624183177948f;
  t226_n236: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t226_n238;
  t226_n237: return 0.0301705002784729f;
  t226_n238: return -0.21574869751930237f;
  t226_n239: return -0.0011781888315454125f;
  t226_n240: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t226_n246;
  t226_n241: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t226_n243;
  t226_n242: return 0.007185529451817274f;
  t226_n243: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n245;
  t226_n244: return -0.044625937938690186f;
  t226_n245: return -0.17843547463417053f;
  t226_n246: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t226_n254;
  t226_n247: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t226_n251;
  t226_n248: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t226_n250;
  t226_n249: return -0.20513255894184113f;
  t226_n250: return 0.03534277528524399f;
  t226_n251: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t226_n253;
  t226_n252: return -0.0500466413795948f;
  t226_n253: return 0.05585463345050812f;
  t226_n254: return 0.10630208998918533f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree227(const Example& E) {
  t227_n0: if (E.getNumReferences() >= 3325613056 /*11829.0*/) goto t227_n94;
  t227_n1: if (E.getNumReferences() >= 3324152832 /*10403.0*/) goto t227_n53;
  t227_n2: if (E.getNumReferences() >= 3324027392 /*10280.5*/) goto t227_n44;
  t227_n3: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t227_n13;
  t227_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t227_n6;
  t227_n5: return -0.000562851142603904f;
  t227_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t227_n8;
  t227_n7: return 0.0004778378352057189f;
  t227_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t227_n10;
  t227_n9: return 0.00014606033801101148f;
  t227_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t227_n12;
  t227_n11: return 0.005329979117959738f;
  t227_n12: return 0.013630050234496593f;
  t227_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n17;
  t227_n14: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t227_n16;
  t227_n15: return 0.09966189414262772f;
  t227_n16: return -0.02085857279598713f;
  t227_n17: if (E.getNumReferences() >= 3323953664 /*10208.5*/) goto t227_n43;
  t227_n18: if (E.getNumReferences() >= 3323924480 /*10180.0*/) goto t227_n36;
  t227_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n35;
  t227_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n26;
  t227_n21: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t227_n23;
  t227_n22: return 0.04234345629811287f;
  t227_n23: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t227_n25;
  t227_n24: return -0.03723917156457901f;
  t227_n25: return 0.016305653378367424f;
  t227_n26: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t227_n28;
  t227_n27: return 0.026592224836349487f;
  t227_n28: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t227_n30;
  t227_n29: return -0.11246966570615768f;
  t227_n30: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t227_n34;
  t227_n31: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t227_n33;
  t227_n32: return 0.10461011528968811f;
  t227_n33: return 0.13516803085803986f;
  t227_n34: return 0.06880433112382889f;
  t227_n35: return 0.01025488879531622f;
  t227_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t227_n40;
  t227_n37: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t227_n39;
  t227_n38: return 0.01708277501165867f;
  t227_n39: return 0.14800822734832764f;
  t227_n40: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n42;
  t227_n41: return 0.09677588194608688f;
  t227_n42: return 0.13786083459854126f;
  t227_n43: return -0.11283683776855469f;
  t227_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t227_n46;
  t227_n45: return 0.024316852912306786f;
  t227_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t227_n50;
  t227_n47: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t227_n49;
  t227_n48: return 0.13214798271656036f;
  t227_n49: return 0.10428135842084885f;
  t227_n50: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t227_n52;
  t227_n51: return -0.10666033625602722f;
  t227_n52: return 0.1900484561920166f;
  t227_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n57;
  t227_n54: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t227_n56;
  t227_n55: return -0.0667431503534317f;
  t227_n56: return -0.1188974380493164f;
  t227_n57: if (E.getNumReferences() >= 3325015040 /*11245.0*/) goto t227_n59;
  t227_n58: return -0.011962766759097576f;
  t227_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t227_n69;
  t227_n60: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t227_n64;
  t227_n61: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t227_n63;
  t227_n62: return 0.18050822615623474f;
  t227_n63: return -0.08459067344665527f;
  t227_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n66;
  t227_n65: return -0.09667528420686722f;
  t227_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n68;
  t227_n67: return -0.007781269960105419f;
  t227_n68: return 0.11727540194988251f;
  t227_n69: if (E.getNumReferences() >= 3325493760 /*11712.5*/) goto t227_n93;
  t227_n70: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t227_n76;
  t227_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t227_n75;
  t227_n72: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t227_n74;
  t227_n73: return 0.02136373333632946f;
  t227_n74: return 0.10309034585952759f;
  t227_n75: return 0.1126791462302208f;
  t227_n76: if (E.getNumReferences() >= 3325349888 /*11572.0*/) goto t227_n84;
  t227_n77: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t227_n79;
  t227_n78: return 0.036217376589775085f;
  t227_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n81;
  t227_n80: return -0.10629986226558685f;
  t227_n81: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t227_n83;
  t227_n82: return 0.14991801977157593f;
  t227_n83: return 0.10412554442882538f;
  t227_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t227_n90;
  t227_n85: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t227_n87;
  t227_n86: return 0.09173871576786041f;
  t227_n87: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t227_n89;
  t227_n88: return -0.13851654529571533f;
  t227_n89: return 0.10874887555837631f;
  t227_n90: if (E.getNumReferences() >= 3325383680 /*11605.0*/) goto t227_n92;
  t227_n91: return 0.1840714067220688f;
  t227_n92: return 0.022059589624404907f;
  t227_n93: return -0.1676694005727768f;
  t227_n94: if (E.getNumReferences() >= 3325976064 /*12183.5*/) goto t227_n126;
  t227_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t227_n97;
  t227_n96: return -0.03412257134914398f;
  t227_n97: if (E.getNumReferences() >= 3325867520 /*12077.5*/) goto t227_n111;
  t227_n98: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t227_n104;
  t227_n99: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t227_n101;
  t227_n100: return 0.06102193146944046f;
  t227_n101: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t227_n103;
  t227_n102: return 0.29048094153404236f;
  t227_n103: return -0.10164167732000351f;
  t227_n104: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n106;
  t227_n105: return 0.07971961051225662f;
  t227_n106: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t227_n110;
  t227_n107: if (E.getFractionNameInContext() >= 3193263446 /*0.2083333432674408*/) goto t227_n109;
  t227_n108: return 0.22772225737571716f;
  t227_n109: return 0.1461959183216095f;
  t227_n110: return 0.10154706984758377f;
  t227_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t227_n119;
  t227_n112: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t227_n118;
  t227_n113: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t227_n117;
  t227_n114: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t227_n116;
  t227_n115: return 0.06576003134250641f;
  t227_n116: return 0.12595565617084503f;
  t227_n117: return 0.14718671143054962f;
  t227_n118: return -0.10990709811449051f;
  t227_n119: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t227_n121;
  t227_n120: return -0.10214636474847794f;
  t227_n121: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t227_n123;
  t227_n122: return 0.12357817590236664f;
  t227_n123: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t227_n125;
  t227_n124: return 0.20198966562747955f;
  t227_n125: return 0.10640458762645721f;
  t227_n126: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t227_n230;
  t227_n127: if (E.getNumReferences() >= 3327362048 /*13537.0*/) goto t227_n151;
  t227_n128: if (E.getNumReferences() >= 3327202816 /*13381.5*/) goto t227_n146;
  t227_n129: if (E.getNumReferences() >= 3326945792 /*13130.5*/) goto t227_n143;
  t227_n130: if (E.getNumReferences() >= 3326745088 /*12934.5*/) goto t227_n134;
  t227_n131: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n133;
  t227_n132: return -0.205636665225029f;
  t227_n133: return 0.045043956488370895f;
  t227_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n140;
  t227_n135: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t227_n139;
  t227_n136: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t227_n138;
  t227_n137: return 0.09916865080595016f;
  t227_n138: return 0.2151649296283722f;
  t227_n139: return 0.048152074217796326f;
  t227_n140: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t227_n142;
  t227_n141: return 0.24445664882659912f;
  t227_n142: return 0.08764692395925522f;
  t227_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n145;
  t227_n144: return -0.23784422874450684f;
  t227_n145: return 0.010435453616082668f;
  t227_n146: if (E.getNumReferences() >= 3327246848 /*13424.5*/) goto t227_n150;
  t227_n147: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n149;
  t227_n148: return 0.11955372244119644f;
  t227_n149: return 0.04485267028212547f;
  t227_n150: return 0.054626379162073135f;
  t227_n151: if (E.getNumReferences() >= 3331353088 /*18485.0*/) goto t227_n185;
  t227_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t227_n184;
  t227_n153: if (E.getNumReferences() >= 3330547456 /*16911.5*/) goto t227_n183;
  t227_n154: if (E.getNumReferences() >= 3329439232 /*15565.5*/) goto t227_n176;
  t227_n155: if (E.getNumReferences() >= 3329227264 /*15358.5*/) goto t227_n175;
  t227_n156: if (E.getNumReferences() >= 3328655872 /*14800.5*/) goto t227_n158;
  t227_n157: return -0.04796050861477852f;
  t227_n158: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n160;
  t227_n159: return -0.08821194618940353f;
  t227_n160: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t227_n162;
  t227_n161: return -0.03552652895450592f;
  t227_n162: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t227_n174;
  t227_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n173;
  t227_n164: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t227_n166;
  t227_n165: return 0.14604178071022034f;
  t227_n166: if (E.getNumReferences() >= 3328821248 /*14962.0*/) goto t227_n170;
  t227_n167: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t227_n169;
  t227_n168: return 0.08180598169565201f;
  t227_n169: return 0.14925844967365265f;
  t227_n170: if (E.getNumReferences() >= 3329118720 /*15252.5*/) goto t227_n172;
  t227_n171: return 0.04298701882362366f;
  t227_n172: return 0.1959780901670456f;
  t227_n173: return -0.23280617594718933f;
  t227_n174: return -0.3225562870502472f;
  t227_n175: return -0.12152508646249771f;
  t227_n176: if (E.getNumReferences() >= 3329839104 /*15956.0*/) goto t227_n178;
  t227_n177: return 0.09024520963430405f;
  t227_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n180;
  t227_n179: return -0.055531956255435944f;
  t227_n180: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t227_n182;
  t227_n181: return -0.044487226754426956f;
  t227_n182: return 0.0792670026421547f;
  t227_n183: return -0.06550654023885727f;
  t227_n184: return 0.16324040293693542f;
  t227_n185: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t227_n189;
  t227_n186: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t227_n188;
  t227_n187: return 0.10242985188961029f;
  t227_n188: return 0.13493573665618896f;
  t227_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t227_n223;
  t227_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t227_n222;
  t227_n191: if (E.getNumReferences() >= 3331912192 /*19577.0*/) goto t227_n201;
  t227_n192: if (E.getNumReferences() >= 3331859712 /*19474.5*/) goto t227_n200;
  t227_n193: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t227_n197;
  t227_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t227_n196;
  t227_n195: return 0.02690250799059868f;
  t227_n196: return 0.07068586349487305f;
  t227_n197: if (E.getNumReferences() >= 3331418624 /*18613.0*/) goto t227_n199;
  t227_n198: return 0.0952613428235054f;
  t227_n199: return -0.08112822473049164f;
  t227_n200: return -0.1541067361831665f;
  t227_n201: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t227_n221;
  t227_n202: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t227_n212;
  t227_n203: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t227_n211;
  t227_n204: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t227_n208;
  t227_n205: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t227_n207;
  t227_n206: return 0.02616683393716812f;
  t227_n207: return 0.08375653624534607f;
  t227_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t227_n210;
  t227_n209: return -0.37798401713371277f;
  t227_n210: return 0.10033496469259262f;
  t227_n211: return -0.2009621262550354f;
  t227_n212: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n218;
  t227_n213: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t227_n217;
  t227_n214: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t227_n216;
  t227_n215: return 0.10407909750938416f;
  t227_n216: return 0.04152722656726837f;
  t227_n217: return -0.19046543538570404f;
  t227_n218: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t227_n220;
  t227_n219: return -0.02288665808737278f;
  t227_n220: return 0.0958697646856308f;
  t227_n221: return 0.014013686217367649f;
  t227_n222: return 0.11551924794912338f;
  t227_n223: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t227_n225;
  t227_n224: return 0.0013451224658638239f;
  t227_n225: if (E.getNumReferences() >= 3353955968 /*119453.0*/) goto t227_n229;
  t227_n226: if (E.getNumReferences() >= 3335819776 /*27209.0*/) goto t227_n228;
  t227_n227: return 0.03911489620804787f;
  t227_n228: return -2.2194712162017822f;
  t227_n229: return -0.06711427122354507f;
  t227_n230: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t227_n254;
  t227_n231: if (E.getNumReferences() >= 3338286080 /*32026.0*/) goto t227_n253;
  t227_n232: if (E.getNumReferences() >= 3338280960 /*32016.0*/) goto t227_n248;
  t227_n233: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t227_n243;
  t227_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t227_n236;
  t227_n235: return -0.010531439445912838f;
  t227_n236: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t227_n242;
  t227_n237: if (E.getNumReferences() >= 3331258112 /*18299.5*/) goto t227_n239;
  t227_n238: return 0.08174094557762146f;
  t227_n239: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t227_n241;
  t227_n240: return 0.1705499291419983f;
  t227_n241: return 0.1226547583937645f;
  t227_n242: return -0.026066891849040985f;
  t227_n243: if (E.getNumReferences() >= 3337447424 /*30388.0*/) goto t227_n247;
  t227_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n246;
  t227_n245: return -0.2730500102043152f;
  t227_n246: return 0.1415412724018097f;
  t227_n247: return 0.07234431803226471f;
  t227_n248: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t227_n252;
  t227_n249: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t227_n251;
  t227_n250: return -0.16279387474060059f;
  t227_n251: return 0.07831986993551254f;
  t227_n252: return 0.10812189429998398f;
  t227_n253: return -0.023171396926045418f;
  t227_n254: return 0.002320606494322419f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree228(const Example& E) {
  t228_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t228_n250;
  t228_n1: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t228_n223;
  t228_n2: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t228_n4;
  t228_n3: return -0.09230780601501465f;
  t228_n4: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t228_n6;
  t228_n5: return -5.868276275577955e-06f;
  t228_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t228_n64;
  t228_n7: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t228_n61;
  t228_n8: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t228_n58;
  t228_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t228_n47;
  t228_n10: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t228_n40;
  t228_n11: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t228_n35;
  t228_n12: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t228_n34;
  t228_n13: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t228_n33;
  t228_n14: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t228_n16;
  t228_n15: return 0.05609293282032013f;
  t228_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t228_n32;
  t228_n17: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n27;
  t228_n18: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t228_n22;
  t228_n19: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t228_n21;
  t228_n20: return -0.05808155611157417f;
  t228_n21: return 0.049990955740213394f;
  t228_n22: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t228_n26;
  t228_n23: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t228_n25;
  t228_n24: return -0.08089683949947357f;
  t228_n25: return -0.3008975088596344f;
  t228_n26: return 0.019137252122163773f;
  t228_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t228_n31;
  t228_n28: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t228_n30;
  t228_n29: return -0.1831478327512741f;
  t228_n30: return 0.04236863553524017f;
  t228_n31: return 0.006241426337510347f;
  t228_n32: return 0.0032784249633550644f;
  t228_n33: return -0.7652757167816162f;
  t228_n34: return 0.016242429614067078f;
  t228_n35: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t228_n39;
  t228_n36: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t228_n38;
  t228_n37: return 0.05499543249607086f;
  t228_n38: return 0.0933292806148529f;
  t228_n39: return 0.01270543783903122f;
  t228_n40: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t228_n44;
  t228_n41: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t228_n43;
  t228_n42: return 0.18459263443946838f;
  t228_n43: return -0.15632183849811554f;
  t228_n44: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t228_n46;
  t228_n45: return 0.20996323227882385f;
  t228_n46: return 0.1844148188829422f;
  t228_n47: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t228_n55;
  t228_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n50;
  t228_n49: return 0.019557679072022438f;
  t228_n50: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t228_n52;
  t228_n51: return 0.1143384799361229f;
  t228_n52: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t228_n54;
  t228_n53: return -0.08123756945133209f;
  t228_n54: return 0.13917312026023865f;
  t228_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t228_n57;
  t228_n56: return 0.026915227994322777f;
  t228_n57: return 0.12018032371997833f;
  t228_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t228_n60;
  t228_n59: return -0.012239141389727592f;
  t228_n60: return 0.01274009421467781f;
  t228_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t228_n63;
  t228_n62: return 0.008294661529362202f;
  t228_n63: return 0.22299246490001678f;
  t228_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t228_n66;
  t228_n65: return -0.004185959696769714f;
  t228_n66: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t228_n76;
  t228_n67: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t228_n69;
  t228_n68: return -0.009605873376131058f;
  t228_n69: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t228_n75;
  t228_n70: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t228_n74;
  t228_n71: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t228_n73;
  t228_n72: return -0.6110438704490662f;
  t228_n73: return 0.06048965826630592f;
  t228_n74: return 0.11146958917379379f;
  t228_n75: return -0.10266215354204178f;
  t228_n76: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t228_n206;
  t228_n77: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t228_n157;
  t228_n78: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t228_n152;
  t228_n79: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t228_n151;
  t228_n80: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t228_n84;
  t228_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t228_n83;
  t228_n82: return -0.0706610307097435f;
  t228_n83: return 0.051556456834077835f;
  t228_n84: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t228_n138;
  t228_n85: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n133;
  t228_n86: if (E.getNumReferences() >= 3266969600 /*93.0*/) goto t228_n128;
  t228_n87: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t228_n99;
  t228_n88: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t228_n92;
  t228_n89: if (E.getNumReferences() >= 3247964160 /*19.0*/) goto t228_n91;
  t228_n90: return -0.012233437970280647f;
  t228_n91: return 0.20915555953979492f;
  t228_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t228_n96;
  t228_n93: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t228_n95;
  t228_n94: return -0.07005397230386734f;
  t228_n95: return 0.10811993479728699f;
  t228_n96: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t228_n98;
  t228_n97: return 0.11516381800174713f;
  t228_n98: return -0.0445953868329525f;
  t228_n99: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t228_n107;
  t228_n100: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t228_n106;
  t228_n101: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t228_n105;
  t228_n102: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t228_n104;
  t228_n103: return -0.13764692842960358f;
  t228_n104: return 0.1718827337026596f;
  t228_n105: return 0.13995705544948578f;
  t228_n106: return -0.14348164200782776f;
  t228_n107: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t228_n115;
  t228_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t228_n110;
  t228_n109: return -0.037871070206165314f;
  t228_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t228_n114;
  t228_n111: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t228_n113;
  t228_n112: return 0.014562112279236317f;
  t228_n113: return 0.1404688060283661f;
  t228_n114: return 0.10324546694755554f;
  t228_n115: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t228_n127;
  t228_n116: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t228_n118;
  t228_n117: return 0.03769427165389061f;
  t228_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t228_n120;
  t228_n119: return 0.06336689740419388f;
  t228_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t228_n124;
  t228_n121: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t228_n123;
  t228_n122: return -0.10392060875892639f;
  t228_n123: return 0.09878014028072357f;
  t228_n124: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t228_n126;
  t228_n125: return 0.27226150035858154f;
  t228_n126: return -0.10008877515792847f;
  t228_n127: return 0.09407931566238403f;
  t228_n128: if (E.getNumReferences() >= 3278602240 /*235.5*/) goto t228_n132;
  t228_n129: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t228_n131;
  t228_n130: return 0.10313331335783005f;
  t228_n131: return 0.1609855741262436f;
  t228_n132: return -0.22934620082378387f;
  t228_n133: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t228_n137;
  t228_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t228_n136;
  t228_n135: return 0.09568671882152557f;
  t228_n136: return 0.12265372276306152f;
  t228_n137: return -0.04335731640458107f;
  t228_n138: if (E.getFractionNameInContext() >= 3207524079 /*0.6833333373069763*/) goto t228_n150;
  t228_n139: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t228_n149;
  t228_n140: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t228_n148;
  t228_n141: if (E.getNumReferences() >= 3225419776 /*3.0*/) goto t228_n147;
  t228_n142: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n144;
  t228_n143: return 0.12455827742815018f;
  t228_n144: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t228_n146;
  t228_n145: return 0.09846443682909012f;
  t228_n146: return 0.12500756978988647f;
  t228_n147: return 0.03864709287881851f;
  t228_n148: return 0.021504806354641914f;
  t228_n149: return 0.016250357031822205f;
  t228_n150: return 0.044816650450229645f;
  t228_n151: return 0.04953662306070328f;
  t228_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t228_n156;
  t228_n153: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t228_n155;
  t228_n154: return 0.0941428616642952f;
  t228_n155: return -0.019984759390354156f;
  t228_n156: return 0.05150469392538071f;
  t228_n157: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t228_n169;
  t228_n158: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t228_n160;
  t228_n159: return 0.006368972361087799f;
  t228_n160: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t228_n168;
  t228_n161: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t228_n163;
  t228_n162: return 0.11334756016731262f;
  t228_n163: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t228_n165;
  t228_n164: return 0.031706538051366806f;
  t228_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t228_n167;
  t228_n166: return 0.1447373777627945f;
  t228_n167: return 0.2561272084712982f;
  t228_n168: return 0.10782890021800995f;
  t228_n169: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t228_n171;
  t228_n170: return -0.008343439549207687f;
  t228_n171: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t228_n205;
  t228_n172: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t228_n204;
  t228_n173: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t228_n177;
  t228_n174: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t228_n176;
  t228_n175: return 0.05001479387283325f;
  t228_n176: return -0.08570617437362671f;
  t228_n177: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t228_n195;
  t228_n178: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t228_n184;
  t228_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t228_n181;
  t228_n180: return -0.025272460654377937f;
  t228_n181: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n183;
  t228_n182: return 0.11148198693990707f;
  t228_n183: return 0.03241681307554245f;
  t228_n184: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t228_n190;
  t228_n185: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t228_n189;
  t228_n186: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t228_n188;
  t228_n187: return 0.029701396822929382f;
  t228_n188: return 0.06271857768297195f;
  t228_n189: return 0.11800643056631088f;
  t228_n190: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t228_n194;
  t228_n191: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t228_n193;
  t228_n192: return 0.07939807325601578f;
  t228_n193: return 0.005496231839060783f;
  t228_n194: return -0.19789113104343414f;
  t228_n195: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t228_n203;
  t228_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t228_n198;
  t228_n197: return 0.07293606549501419f;
  t228_n198: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t228_n202;
  t228_n199: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t228_n201;
  t228_n200: return 0.1007954403758049f;
  t228_n201: return 0.10032417625188828f;
  t228_n202: return -0.037978075444698334f;
  t228_n203: return 0.03234534710645676f;
  t228_n204: return 0.10091125965118408f;
  t228_n205: return 0.01767336204648018f;
  t228_n206: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t228_n210;
  t228_n207: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t228_n209;
  t228_n208: return -0.41157230734825134f;
  t228_n209: return -0.04407310485839844f;
  t228_n210: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t228_n220;
  t228_n211: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t228_n213;
  t228_n212: return -0.010858495719730854f;
  t228_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t228_n217;
  t228_n214: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t228_n216;
  t228_n215: return 0.044177696108818054f;
  t228_n216: return -0.22693999111652374f;
  t228_n217: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t228_n219;
  t228_n218: return -0.10004817694425583f;
  t228_n219: return 0.15552693605422974f;
  t228_n220: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t228_n222;
  t228_n221: return 0.09047935903072357f;
  t228_n222: return -0.05320943519473076f;
  t228_n223: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t228_n227;
  t228_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t228_n226;
  t228_n225: return -0.026518087834119797f;
  t228_n226: return 0.011437127366662025f;
  t228_n227: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t228_n233;
  t228_n228: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t228_n230;
  t228_n229: return -0.11244067549705505f;
  t228_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t228_n232;
  t228_n231: return 0.07001253962516785f;
  t228_n232: return 0.1549959033727646f;
  t228_n233: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t228_n243;
  t228_n234: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t228_n236;
  t228_n235: return -0.07546154409646988f;
  t228_n236: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t228_n238;
  t228_n237: return -0.0827774852514267f;
  t228_n238: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t228_n240;
  t228_n239: return -0.15116335451602936f;
  t228_n240: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t228_n242;
  t228_n241: return 0.11679379642009735f;
  t228_n242: return 0.05192935839295387f;
  t228_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t228_n245;
  t228_n244: return -0.12831762433052063f;
  t228_n245: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t228_n247;
  t228_n246: return -0.23755531013011932f;
  t228_n247: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t228_n249;
  t228_n248: return 0.13384217023849487f;
  t228_n249: return -0.11223242431879044f;
  t228_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t228_n252;
  t228_n251: return 0.0013081288198009133f;
  t228_n252: if (E.getNumReferences() >= 3300212736 /*1450.0*/) goto t228_n254;
  t228_n253: return 0.0024874110240489244f;
  t228_n254: return 0.02431313320994377f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree229(const Example& E) {
  t229_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t229_n214;
  t229_n1: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t229_n89;
  t229_n2: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t229_n6;
  t229_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t229_n5;
  t229_n4: return -0.0020359845366328955f;
  t229_n5: return 0.00023910205345600843f;
  t229_n6: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t229_n50;
  t229_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t229_n15;
  t229_n8: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t229_n10;
  t229_n9: return 0.002171463333070278f;
  t229_n10: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t229_n12;
  t229_n11: return 0.020177777856588364f;
  t229_n12: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t229_n14;
  t229_n13: return -0.11506873369216919f;
  t229_n14: return 0.09199301898479462f;
  t229_n15: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t229_n49;
  t229_n16: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t229_n46;
  t229_n17: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t229_n33;
  t229_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t229_n32;
  t229_n19: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t229_n29;
  t229_n20: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t229_n28;
  t229_n21: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t229_n23;
  t229_n22: return 0.009273529052734375f;
  t229_n23: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t229_n25;
  t229_n24: return 0.041267815977334976f;
  t229_n25: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t229_n27;
  t229_n26: return -0.05306601524353027f;
  t229_n27: return 0.08394549041986465f;
  t229_n28: return 0.09447026997804642f;
  t229_n29: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t229_n31;
  t229_n30: return -0.06441076099872589f;
  t229_n31: return 0.052548885345458984f;
  t229_n32: return 0.04512321949005127f;
  t229_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t229_n41;
  t229_n34: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n36;
  t229_n35: return -0.0593760721385479f;
  t229_n36: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t229_n38;
  t229_n37: return 0.043365564197301865f;
  t229_n38: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t229_n40;
  t229_n39: return -0.029785729944705963f;
  t229_n40: return -0.7042045593261719f;
  t229_n41: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t229_n45;
  t229_n42: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t229_n44;
  t229_n43: return 0.012332971207797527f;
  t229_n44: return 0.08014637976884842f;
  t229_n45: return 0.11087024956941605f;
  t229_n46: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t229_n48;
  t229_n47: return 0.048518382012844086f;
  t229_n48: return 0.13512176275253296f;
  t229_n49: return -0.20016144216060638f;
  t229_n50: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t229_n88;
  t229_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t229_n75;
  t229_n52: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t229_n74;
  t229_n53: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t229_n61;
  t229_n54: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t229_n60;
  t229_n55: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t229_n57;
  t229_n56: return -0.05689764395356178f;
  t229_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t229_n59;
  t229_n58: return 0.03708070144057274f;
  t229_n59: return 0.13047316670417786f;
  t229_n60: return -0.0903095006942749f;
  t229_n61: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t229_n71;
  t229_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t229_n64;
  t229_n63: return -0.011067604646086693f;
  t229_n64: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t229_n66;
  t229_n65: return 0.03843434900045395f;
  t229_n66: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t229_n68;
  t229_n67: return 0.1557874083518982f;
  t229_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n70;
  t229_n69: return 0.0709933489561081f;
  t229_n70: return 0.10782479494810104f;
  t229_n71: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t229_n73;
  t229_n72: return -0.13645699620246887f;
  t229_n73: return 0.09760754555463791f;
  t229_n74: return -0.020863166078925133f;
  t229_n75: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t229_n79;
  t229_n76: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t229_n78;
  t229_n77: return -0.04661082848906517f;
  t229_n78: return -0.01520582102239132f;
  t229_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t229_n81;
  t229_n80: return -0.002476133406162262f;
  t229_n81: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n87;
  t229_n82: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t229_n84;
  t229_n83: return 0.010401537641882896f;
  t229_n84: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t229_n86;
  t229_n85: return 0.020490184426307678f;
  t229_n86: return 0.10518446564674377f;
  t229_n87: return -0.013091526925563812f;
  t229_n88: return -0.03718859702348709f;
  t229_n89: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t229_n143;
  t229_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t229_n142;
  t229_n91: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t229_n141;
  t229_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t229_n130;
  t229_n93: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t229_n113;
  t229_n94: if (E.getNumReferences() >= 3291627520 /*713.0*/) goto t229_n108;
  t229_n95: if (E.getNumReferences() >= 3281764352 /*311.5*/) goto t229_n97;
  t229_n96: return -0.01561479177325964f;
  t229_n97: if (E.getNumReferences() >= 3281895424 /*315.5*/) goto t229_n99;
  t229_n98: return 0.15866485238075256f;
  t229_n99: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t229_n107;
  t229_n100: if (E.getNumReferences() >= 3287859200 /*497.5*/) goto t229_n104;
  t229_n101: if (E.getNumReferences() >= 3285991424 /*440.5*/) goto t229_n103;
  t229_n102: return 0.007944485172629356f;
  t229_n103: return 0.07528991252183914f;
  t229_n104: if (E.getNumReferences() >= 3289522176 /*584.5*/) goto t229_n106;
  t229_n105: return -0.21129877865314484f;
  t229_n106: return 0.04632841795682907f;
  t229_n107: return 0.1130317822098732f;
  t229_n108: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t229_n110;
  t229_n109: return -0.06677159667015076f;
  t229_n110: if (E.getNumReferences() >= 3306340352 /*2348.0*/) goto t229_n112;
  t229_n111: return -0.015613981522619724f;
  t229_n112: return 0.11860755831003189f;
  t229_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t229_n129;
  t229_n114: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t229_n120;
  t229_n115: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t229_n119;
  t229_n116: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t229_n118;
  t229_n117: return 0.0201300997287035f;
  t229_n118: return 0.15193141996860504f;
  t229_n119: return 0.12290836870670319f;
  t229_n120: if (E.getNumReferences() >= 3266641920 /*90.5*/) goto t229_n128;
  t229_n121: if (E.getNumReferences() >= 3266379776 /*88.5*/) goto t229_n127;
  t229_n122: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t229_n126;
  t229_n123: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t229_n125;
  t229_n124: return -0.023893196135759354f;
  t229_n125: return -0.5710087418556213f;
  t229_n126: return 0.12701651453971863f;
  t229_n127: return -0.5197571516036987f;
  t229_n128: return 0.014966331422328949f;
  t229_n129: return 0.10801827162504196f;
  t229_n130: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n132;
  t229_n131: return 0.008727756328880787f;
  t229_n132: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t229_n134;
  t229_n133: return -5.228519057709491e-06f;
  t229_n134: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t229_n136;
  t229_n135: return 0.012606154195964336f;
  t229_n136: if (E.getNumReferences() >= 3273687040 /*160.5*/) goto t229_n138;
  t229_n137: return -0.16565477848052979f;
  t229_n138: if (E.getNumReferences() >= 3297529856 /*1122.5*/) goto t229_n140;
  t229_n139: return 0.18905463814735413f;
  t229_n140: return 0.091887928545475f;
  t229_n141: return -0.11745724827051163f;
  t229_n142: return 0.11964304745197296f;
  t229_n143: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t229_n197;
  t229_n144: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t229_n146;
  t229_n145: return 0.0020898596849292517f;
  t229_n146: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t229_n148;
  t229_n147: return -0.025987917557358742f;
  t229_n148: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t229_n184;
  t229_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t229_n165;
  t229_n150: if (E.getNumReferences() >= 3332358912 /*20449.5*/) goto t229_n158;
  t229_n151: if (E.getNumReferences() >= 3330841856 /*17486.5*/) goto t229_n153;
  t229_n152: return -0.02973240613937378f;
  t229_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t229_n157;
  t229_n154: if (E.getNumReferences() >= 3332121600 /*19986.0*/) goto t229_n156;
  t229_n155: return 0.10548528283834457f;
  t229_n156: return -0.3922637701034546f;
  t229_n157: return 0.15056146681308746f;
  t229_n158: if (E.getNumReferences() >= 3334465280 /*24563.5*/) goto t229_n162;
  t229_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t229_n161;
  t229_n160: return -0.5239505171775818f;
  t229_n161: return 0.10725543648004532f;
  t229_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t229_n164;
  t229_n163: return -0.2229522466659546f;
  t229_n164: return -0.02529648132622242f;
  t229_n165: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t229_n183;
  t229_n166: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t229_n176;
  t229_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t229_n169;
  t229_n168: return -0.0032353613059967756f;
  t229_n169: if (E.getNumReferences() >= 3274997760 /*180.5*/) goto t229_n171;
  t229_n170: return 0.005235161166638136f;
  t229_n171: if (E.getNumReferences() >= 3280060416 /*259.5*/) goto t229_n175;
  t229_n172: if (E.getNumReferences() >= 3276931072 /*210.0*/) goto t229_n174;
  t229_n173: return 0.16667868196964264f;
  t229_n174: return 0.23128582537174225f;
  t229_n175: return 0.015024389140307903f;
  t229_n176: if (E.getNumReferences() >= 3263627264 /*67.5*/) goto t229_n178;
  t229_n177: return -0.19509805738925934f;
  t229_n178: if (E.getNumReferences() >= 3279126528 /*243.5*/) goto t229_n182;
  t229_n179: if (E.getFileProximityDistanceCost() >= 3250585600 /*24.0*/) goto t229_n181;
  t229_n180: return 0.13061723113059998f;
  t229_n181: return 0.012969130650162697f;
  t229_n182: return 0.03365885466337204f;
  t229_n183: return -0.007824907079339027f;
  t229_n184: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t229_n190;
  t229_n185: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t229_n189;
  t229_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t229_n188;
  t229_n187: return 0.09078788757324219f;
  t229_n188: return 0.1317017823457718f;
  t229_n189: return -0.09058074653148651f;
  t229_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t229_n194;
  t229_n191: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n193;
  t229_n192: return 0.01048292312771082f;
  t229_n193: return -0.02023589052259922f;
  t229_n194: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t229_n196;
  t229_n195: return -0.5824673771858215f;
  t229_n196: return 0.0687517300248146f;
  t229_n197: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t229_n201;
  t229_n198: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t229_n200;
  t229_n199: return 0.12387582659721375f;
  t229_n200: return -0.14255845546722412f;
  t229_n201: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t229_n207;
  t229_n202: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t229_n204;
  t229_n203: return 0.03250380977988243f;
  t229_n204: if (E.getNumReferences() >= 3326113280 /*12317.5*/) goto t229_n206;
  t229_n205: return -0.10004398226737976f;
  t229_n206: return 0.16024072468280792f;
  t229_n207: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t229_n213;
  t229_n208: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t229_n212;
  t229_n209: if (E.getNumReferences() >= 3281223680 /*295.0*/) goto t229_n211;
  t229_n210: return 0.14774161577224731f;
  t229_n211: return 0.08443696051836014f;
  t229_n212: return -0.15379111468791962f;
  t229_n213: return 0.01744156703352928f;
  t229_n214: if (E.getNumReferences() >= 3325673472 /*11888.0*/) goto t229_n246;
  t229_n215: if (E.getNumReferences() >= 3325138432 /*11365.5*/) goto t229_n245;
  t229_n216: if (E.getNumReferences() >= 3323675648 /*9937.0*/) goto t229_n234;
  t229_n217: if (E.getNumReferences() >= 3317972992 /*6280.0*/) goto t229_n219;
  t229_n218: return 0.0014065635623410344f;
  t229_n219: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t229_n223;
  t229_n220: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t229_n222;
  t229_n221: return -0.09573472291231155f;
  t229_n222: return 0.06859893351793289f;
  t229_n223: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t229_n225;
  t229_n224: return -0.012103124521672726f;
  t229_n225: if (E.getNumReferences() >= 3320826880 /*7673.5*/) goto t229_n229;
  t229_n226: if (E.getNumReferences() >= 3320230912 /*7382.5*/) goto t229_n228;
  t229_n227: return 0.08823347836732864f;
  t229_n228: return -0.5038909912109375f;
  t229_n229: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t229_n231;
  t229_n230: return 0.18487204611301422f;
  t229_n231: if (E.getNumReferences() >= 3321914880 /*8217.5*/) goto t229_n233;
  t229_n232: return 0.1598455309867859f;
  t229_n233: return -0.09959990531206131f;
  t229_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t229_n238;
  t229_n235: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t229_n237;
  t229_n236: return -0.8939321041107178f;
  t229_n237: return -0.03815673291683197f;
  t229_n238: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t229_n240;
  t229_n239: return 0.007025180384516716f;
  t229_n240: if (E.getNumReferences() >= 3323745792 /*10005.5*/) goto t229_n244;
  t229_n241: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t229_n243;
  t229_n242: return 0.12921088933944702f;
  t229_n243: return 0.1300690621137619f;
  t229_n244: return 0.09365364164113998f;
  t229_n245: return -0.19039399921894073f;
  t229_n246: if (E.getNumReferences() >= 3325918208 /*12127.0*/) goto t229_n250;
  t229_n247: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t229_n249;
  t229_n248: return 0.12899631261825562f;
  t229_n249: return 0.09892572462558746f;
  t229_n250: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t229_n252;
  t229_n251: return 0.02575584314763546f;
  t229_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t229_n254;
  t229_n253: return -0.5952408909797668f;
  t229_n254: return -0.008458414115011692f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree230(const Example& E) {
  t230_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t230_n122;
  t230_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t230_n11;
  t230_n2: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t230_n10;
  t230_n3: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t230_n9;
  t230_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t230_n6;
  t230_n5: return -0.16115087270736694f;
  t230_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n8;
  t230_n7: return -0.027448736131191254f;
  t230_n8: return 0.07097196578979492f;
  t230_n9: return -0.16051852703094482f;
  t230_n10: return -0.0005199492443352938f;
  t230_n11: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t230_n17;
  t230_n12: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t230_n14;
  t230_n13: return -0.00626154663041234f;
  t230_n14: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t230_n16;
  t230_n15: return 0.10918352007865906f;
  t230_n16: return 0.00073273666203022f;
  t230_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t230_n55;
  t230_n18: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t230_n36;
  t230_n19: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t230_n35;
  t230_n20: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t230_n22;
  t230_n21: return 0.004977033007889986f;
  t230_n22: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t230_n32;
  t230_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n29;
  t230_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t230_n26;
  t230_n25: return -0.027219150215387344f;
  t230_n26: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t230_n28;
  t230_n27: return -0.005103521514683962f;
  t230_n28: return 0.10148826241493225f;
  t230_n29: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t230_n31;
  t230_n30: return -0.01399089302867651f;
  t230_n31: return 0.015706460922956467f;
  t230_n32: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t230_n34;
  t230_n33: return 0.19569063186645508f;
  t230_n34: return -0.13123025000095367f;
  t230_n35: return 0.03872660547494888f;
  t230_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n46;
  t230_n37: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t230_n39;
  t230_n38: return -0.015518827363848686f;
  t230_n39: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n43;
  t230_n40: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t230_n42;
  t230_n41: return 0.08171898871660233f;
  t230_n42: return 0.13173794746398926f;
  t230_n43: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t230_n45;
  t230_n44: return 0.05641449615359306f;
  t230_n45: return -0.08193577080965042f;
  t230_n46: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n50;
  t230_n47: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t230_n49;
  t230_n48: return -0.02722844108939171f;
  t230_n49: return 0.006039576139301062f;
  t230_n50: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t230_n54;
  t230_n51: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t230_n53;
  t230_n52: return 0.015183785930275917f;
  t230_n53: return -0.11312459409236908f;
  t230_n54: return -0.020315110683441162f;
  t230_n55: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t230_n81;
  t230_n56: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t230_n60;
  t230_n57: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t230_n59;
  t230_n58: return 0.005732627585530281f;
  t230_n59: return 0.8330191969871521f;
  t230_n60: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t230_n70;
  t230_n61: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t230_n65;
  t230_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t230_n64;
  t230_n63: return 0.0631115585565567f;
  t230_n64: return 0.09792331606149673f;
  t230_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n69;
  t230_n66: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t230_n68;
  t230_n67: return -0.007550915703177452f;
  t230_n68: return 0.4623119533061981f;
  t230_n69: return 0.034589651972055435f;
  t230_n70: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t230_n78;
  t230_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t230_n75;
  t230_n72: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n74;
  t230_n73: return 0.07169224321842194f;
  t230_n74: return -0.054769374430179596f;
  t230_n75: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n77;
  t230_n76: return 0.08473473787307739f;
  t230_n77: return 0.11507681012153625f;
  t230_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n80;
  t230_n79: return -0.2363905906677246f;
  t230_n80: return -0.008755495771765709f;
  t230_n81: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t230_n121;
  t230_n82: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t230_n96;
  t230_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n95;
  t230_n84: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t230_n94;
  t230_n85: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t230_n87;
  t230_n86: return -0.029167553409934044f;
  t230_n87: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n91;
  t230_n88: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t230_n90;
  t230_n89: return 0.03587481752038002f;
  t230_n90: return -0.024984678253531456f;
  t230_n91: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t230_n93;
  t230_n92: return -0.14894790947437286f;
  t230_n93: return 0.05269790440797806f;
  t230_n94: return -0.10176557302474976f;
  t230_n95: return 0.0017173168016597629f;
  t230_n96: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t230_n120;
  t230_n97: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t230_n101;
  t230_n98: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t230_n100;
  t230_n99: return 0.01814415119588375f;
  t230_n100: return -0.004135529976338148f;
  t230_n101: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t230_n107;
  t230_n102: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t230_n106;
  t230_n103: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n105;
  t230_n104: return -0.24455909430980682f;
  t230_n105: return 0.10880439728498459f;
  t230_n106: return -0.006573817227035761f;
  t230_n107: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t230_n117;
  t230_n108: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t230_n110;
  t230_n109: return 0.08802904933691025f;
  t230_n110: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n116;
  t230_n111: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t230_n113;
  t230_n112: return -0.01462502870708704f;
  t230_n113: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t230_n115;
  t230_n114: return 0.08018437772989273f;
  t230_n115: return 0.019365351647138596f;
  t230_n116: return -0.14499931037425995f;
  t230_n117: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t230_n119;
  t230_n118: return -0.009127918630838394f;
  t230_n119: return 0.1225014179944992f;
  t230_n120: return 0.022427309304475784f;
  t230_n121: return -0.16860851645469666f;
  t230_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t230_n254;
  t230_n123: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t230_n247;
  t230_n124: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t230_n246;
  t230_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t230_n127;
  t230_n126: return 0.01121558714658022f;
  t230_n127: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t230_n211;
  t230_n128: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t230_n130;
  t230_n129: return -0.033365398645401f;
  t230_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t230_n134;
  t230_n131: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t230_n133;
  t230_n132: return -0.06344912201166153f;
  t230_n133: return 0.08467652648687363f;
  t230_n134: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t230_n206;
  t230_n135: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t230_n177;
  t230_n136: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t230_n166;
  t230_n137: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t230_n141;
  t230_n138: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t230_n140;
  t230_n139: return 0.014265090227127075f;
  t230_n140: return -0.1352727860212326f;
  t230_n141: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t230_n155;
  t230_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t230_n152;
  t230_n143: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t230_n149;
  t230_n144: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t230_n146;
  t230_n145: return -0.11013541370630264f;
  t230_n146: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t230_n148;
  t230_n147: return 0.21885930001735687f;
  t230_n148: return -0.06377960741519928f;
  t230_n149: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t230_n151;
  t230_n150: return 0.09338504821062088f;
  t230_n151: return -0.10453271865844727f;
  t230_n152: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t230_n154;
  t230_n153: return -0.04530489817261696f;
  t230_n154: return 0.19522137939929962f;
  t230_n155: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t230_n165;
  t230_n156: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t230_n160;
  t230_n157: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t230_n159;
  t230_n158: return -0.21932336688041687f;
  t230_n159: return -0.048336803913116455f;
  t230_n160: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t230_n162;
  t230_n161: return -0.0010961109073832631f;
  t230_n162: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t230_n164;
  t230_n163: return -0.24286127090454102f;
  t230_n164: return -0.0499170757830143f;
  t230_n165: return 0.049428828060626984f;
  t230_n166: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t230_n168;
  t230_n167: return -0.020290561020374298f;
  t230_n168: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t230_n174;
  t230_n169: if (E.getFractionNameInContext() >= 3196898509 /*0.2750000059604645*/) goto t230_n171;
  t230_n170: return 0.11840711534023285f;
  t230_n171: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t230_n173;
  t230_n172: return 0.10582976043224335f;
  t230_n173: return -0.014548071660101414f;
  t230_n174: if (E.getFractionNameInContext() >= 3200953002 /*0.3958333134651184*/) goto t230_n176;
  t230_n175: return -0.14150339365005493f;
  t230_n176: return 0.12637081742286682f;
  t230_n177: if (E.getNumReferences() >= 3275096064 /*182.0*/) goto t230_n205;
  t230_n178: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t230_n202;
  t230_n179: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t230_n185;
  t230_n180: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t230_n182;
  t230_n181: return 0.016430871561169624f;
  t230_n182: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t230_n184;
  t230_n183: return 0.1129177138209343f;
  t230_n184: return -0.19992491602897644f;
  t230_n185: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t230_n199;
  t230_n186: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t230_n196;
  t230_n187: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t230_n193;
  t230_n188: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t230_n190;
  t230_n189: return -0.15384668111801147f;
  t230_n190: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t230_n192;
  t230_n191: return 0.07353518158197403f;
  t230_n192: return -0.02668176405131817f;
  t230_n193: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t230_n195;
  t230_n194: return 0.03660789877176285f;
  t230_n195: return -0.09331149607896805f;
  t230_n196: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t230_n198;
  t230_n197: return 0.11279445141553879f;
  t230_n198: return 0.07885763794183731f;
  t230_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t230_n201;
  t230_n200: return -0.5306468605995178f;
  t230_n201: return -0.06356561928987503f;
  t230_n202: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t230_n204;
  t230_n203: return 0.027021901682019234f;
  t230_n204: return 0.09850997477769852f;
  t230_n205: return 0.1271420270204544f;
  t230_n206: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t230_n208;
  t230_n207: return -0.058494046330451965f;
  t230_n208: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t230_n210;
  t230_n209: return 0.0950603038072586f;
  t230_n210: return 0.13801369071006775f;
  t230_n211: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t230_n213;
  t230_n212: return 0.11596296727657318f;
  t230_n213: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t230_n245;
  t230_n214: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t230_n244;
  t230_n215: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t230_n235;
  t230_n216: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t230_n234;
  t230_n217: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t230_n219;
  t230_n218: return 0.04844559729099274f;
  t230_n219: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t230_n221;
  t230_n220: return 0.01196671836078167f;
  t230_n221: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t230_n225;
  t230_n222: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t230_n224;
  t230_n223: return 0.11935374140739441f;
  t230_n224: return 0.07563180476427078f;
  t230_n225: if (E.getNumReferences() >= 3225419776 /*3.0*/) goto t230_n233;
  t230_n226: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t230_n230;
  t230_n227: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t230_n229;
  t230_n228: return -0.298422634601593f;
  t230_n229: return -0.10260973870754242f;
  t230_n230: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t230_n232;
  t230_n231: return 0.16853933036327362f;
  t230_n232: return 0.02457679621875286f;
  t230_n233: return 0.21737739443778992f;
  t230_n234: return -0.017327930778265f;
  t230_n235: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t230_n243;
  t230_n236: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t230_n238;
  t230_n237: return 0.1277584284543991f;
  t230_n238: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t230_n242;
  t230_n239: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t230_n241;
  t230_n240: return 0.10150455683469772f;
  t230_n241: return 0.0355822890996933f;
  t230_n242: return -0.10528761893510818f;
  t230_n243: return -0.07953860610723495f;
  t230_n244: return 0.19688071310520172f;
  t230_n245: return 0.006339645944535732f;
  t230_n246: return -0.000432012602686882f;
  t230_n247: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t230_n249;
  t230_n248: return 0.0061380756087601185f;
  t230_n249: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t230_n251;
  t230_n250: return 0.032940804958343506f;
  t230_n251: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t230_n253;
  t230_n252: return -0.20320695638656616f;
  t230_n253: return -0.013230588287115097f;
  t230_n254: return 0.003702468704432249f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree231(const Example& E) {
  t231_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t231_n2;
  t231_n1: return -0.00012495233386289328f;
  t231_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t231_n252;
  t231_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t231_n5;
  t231_n4: return -0.12303991615772247f;
  t231_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t231_n243;
  t231_n6: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t231_n216;
  t231_n7: if (E.getNumReferences() >= 3290570752 /*648.5*/) goto t231_n211;
  t231_n8: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t231_n206;
  t231_n9: if (E.getNumReferences() >= 3283271680 /*357.5*/) goto t231_n205;
  t231_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t231_n74;
  t231_n11: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t231_n21;
  t231_n12: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t231_n14;
  t231_n13: return 0.005860097240656614f;
  t231_n14: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t231_n20;
  t231_n15: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t231_n19;
  t231_n16: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t231_n18;
  t231_n17: return 0.14376099407672882f;
  t231_n18: return 0.04711627587676048f;
  t231_n19: return 0.12274657934904099f;
  t231_n20: return 0.016720060259103775f;
  t231_n21: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t231_n65;
  t231_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t231_n32;
  t231_n23: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t231_n27;
  t231_n24: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n26;
  t231_n25: return -0.056728385388851166f;
  t231_n26: return 0.09055795520544052f;
  t231_n27: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n31;
  t231_n28: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t231_n30;
  t231_n29: return 0.09447567164897919f;
  t231_n30: return -0.028676047921180725f;
  t231_n31: return -0.029523834586143494f;
  t231_n32: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t231_n56;
  t231_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t231_n43;
  t231_n34: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n36;
  t231_n35: return 0.025508182123303413f;
  t231_n36: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n40;
  t231_n37: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t231_n39;
  t231_n38: return 0.08734279125928879f;
  t231_n39: return 0.11372093856334686f;
  t231_n40: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t231_n42;
  t231_n41: return 0.07409121096134186f;
  t231_n42: return 0.10635150969028473f;
  t231_n43: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n45;
  t231_n44: return -0.007076118607074022f;
  t231_n45: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t231_n53;
  t231_n46: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t231_n52;
  t231_n47: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n49;
  t231_n48: return 0.279092013835907f;
  t231_n49: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t231_n51;
  t231_n50: return 0.002318008104339242f;
  t231_n51: return 0.10441885143518448f;
  t231_n52: return -0.15088605880737305f;
  t231_n53: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t231_n55;
  t231_n54: return -0.08229099214076996f;
  t231_n55: return 0.05523218587040901f;
  t231_n56: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t231_n60;
  t231_n57: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t231_n59;
  t231_n58: return 0.15045373141765594f;
  t231_n59: return -0.17747801542282104f;
  t231_n60: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t231_n64;
  t231_n61: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t231_n63;
  t231_n62: return -0.22581809759140015f;
  t231_n63: return 0.07773572951555252f;
  t231_n64: return 0.005655668210238218f;
  t231_n65: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t231_n67;
  t231_n66: return 0.07632278650999069f;
  t231_n67: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t231_n69;
  t231_n68: return -0.0012755568604916334f;
  t231_n69: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t231_n73;
  t231_n70: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t231_n72;
  t231_n71: return 0.042380768805742264f;
  t231_n72: return -0.12004092335700989f;
  t231_n73: return 0.10495787858963013f;
  t231_n74: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t231_n122;
  t231_n75: if (E.getNumReferences() >= 3277848576 /*224.0*/) goto t231_n119;
  t231_n76: if (E.getNumReferences() >= 3259891712 /*51.5*/) goto t231_n110;
  t231_n77: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t231_n109;
  t231_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t231_n80;
  t231_n79: return -0.04006645083427429f;
  t231_n80: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t231_n108;
  t231_n81: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t231_n83;
  t231_n82: return 0.0012842889409512281f;
  t231_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t231_n93;
  t231_n84: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t231_n90;
  t231_n85: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t231_n87;
  t231_n86: return -0.4840855002403259f;
  t231_n87: if (E.getNumReferences() >= 3253207040 /*29.0*/) goto t231_n89;
  t231_n88: return 0.08352640271186829f;
  t231_n89: return -0.2608788311481476f;
  t231_n90: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t231_n92;
  t231_n91: return 0.038035307079553604f;
  t231_n92: return -0.05605876073241234f;
  t231_n93: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t231_n95;
  t231_n94: return 0.032871026545763016f;
  t231_n95: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t231_n107;
  t231_n96: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t231_n106;
  t231_n97: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t231_n103;
  t231_n98: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t231_n102;
  t231_n99: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t231_n101;
  t231_n100: return 0.0974794551730156f;
  t231_n101: return 0.1567978411912918f;
  t231_n102: return 0.1337413787841797f;
  t231_n103: if (E.getNumReferences() >= 3252682752 /*28.0*/) goto t231_n105;
  t231_n104: return -0.21775545179843903f;
  t231_n105: return 0.10967177152633667f;
  t231_n106: return 0.035581089556217194f;
  t231_n107: return -0.10088792443275452f;
  t231_n108: return 0.02085915394127369f;
  t231_n109: return 0.11072696000337601f;
  t231_n110: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t231_n118;
  t231_n111: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t231_n113;
  t231_n112: return -0.19582270085811615f;
  t231_n113: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t231_n115;
  t231_n114: return -0.08286270499229431f;
  t231_n115: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t231_n117;
  t231_n116: return 0.16379746794700623f;
  t231_n117: return -0.21482223272323608f;
  t231_n118: return -0.014135662466287613f;
  t231_n119: if (E.getNumReferences() >= 3279847424 /*254.5*/) goto t231_n121;
  t231_n120: return 0.14660567045211792f;
  t231_n121: return 0.07382883876562119f;
  t231_n122: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t231_n160;
  t231_n123: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t231_n153;
  t231_n124: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t231_n128;
  t231_n125: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t231_n127;
  t231_n126: return -0.0956742987036705f;
  t231_n127: return -0.013787567615509033f;
  t231_n128: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t231_n144;
  t231_n129: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t231_n139;
  t231_n130: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t231_n132;
  t231_n131: return -0.06898624449968338f;
  t231_n132: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n136;
  t231_n133: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t231_n135;
  t231_n134: return -0.21116285026073456f;
  t231_n135: return 0.03586871922016144f;
  t231_n136: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t231_n138;
  t231_n137: return -0.043732233345508575f;
  t231_n138: return -0.2769940197467804f;
  t231_n139: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t231_n143;
  t231_n140: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n142;
  t231_n141: return -0.046503935009241104f;
  t231_n142: return -0.12301111966371536f;
  t231_n143: return 0.08796023577451706f;
  t231_n144: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t231_n152;
  t231_n145: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t231_n149;
  t231_n146: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t231_n148;
  t231_n147: return -0.3245692253112793f;
  t231_n148: return 0.0765245258808136f;
  t231_n149: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t231_n151;
  t231_n150: return -0.17985926568508148f;
  t231_n151: return 0.10576090961694717f;
  t231_n152: return 0.1395505666732788f;
  t231_n153: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t231_n159;
  t231_n154: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t231_n156;
  t231_n155: return 0.00802569929510355f;
  t231_n156: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t231_n158;
  t231_n157: return 0.03627493232488632f;
  t231_n158: return 0.1000649631023407f;
  t231_n159: return 0.08918511867523193f;
  t231_n160: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t231_n202;
  t231_n161: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t231_n187;
  t231_n162: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t231_n178;
  t231_n163: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n167;
  t231_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t231_n166;
  t231_n165: return 0.016982130706310272f;
  t231_n166: return 0.12713678181171417f;
  t231_n167: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t231_n177;
  t231_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t231_n170;
  t231_n169: return -0.12736444175243378f;
  t231_n170: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t231_n174;
  t231_n171: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t231_n173;
  t231_n172: return 0.028269855305552483f;
  t231_n173: return 0.09419343620538712f;
  t231_n174: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t231_n176;
  t231_n175: return -0.08791433274745941f;
  t231_n176: return 0.08734334260225296f;
  t231_n177: return 0.029768606647849083f;
  t231_n178: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t231_n180;
  t231_n179: return 0.11451483517885208f;
  t231_n180: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t231_n184;
  t231_n181: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t231_n183;
  t231_n182: return -0.08217267692089081f;
  t231_n183: return 0.07086877524852753f;
  t231_n184: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t231_n186;
  t231_n185: return 0.1481182724237442f;
  t231_n186: return 0.02291332744061947f;
  t231_n187: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t231_n191;
  t231_n188: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t231_n190;
  t231_n189: return -0.03393044322729111f;
  t231_n190: return 0.04166930168867111f;
  t231_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t231_n201;
  t231_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t231_n196;
  t231_n193: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t231_n195;
  t231_n194: return 0.06629760563373566f;
  t231_n195: return 0.12213046848773956f;
  t231_n196: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t231_n198;
  t231_n197: return 0.15480303764343262f;
  t231_n198: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t231_n200;
  t231_n199: return 0.1383848935365677f;
  t231_n200: return 0.007421695627272129f;
  t231_n201: return -0.009481096640229225f;
  t231_n202: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n204;
  t231_n203: return -0.39856618642807007f;
  t231_n204: return -0.15494877099990845f;
  t231_n205: return -0.6627768874168396f;
  t231_n206: if (E.getNumReferences() >= 3284549632 /*396.5*/) goto t231_n208;
  t231_n207: return 0.1190459132194519f;
  t231_n208: if (E.getNumReferences() >= 3287793664 /*495.5*/) goto t231_n210;
  t231_n209: return -0.018877733498811722f;
  t231_n210: return 0.11068783700466156f;
  t231_n211: if (E.getNumReferences() >= 3299483648 /*1361.0*/) goto t231_n215;
  t231_n212: if (E.getNumReferences() >= 3298816000 /*1279.5*/) goto t231_n214;
  t231_n213: return -0.19730369746685028f;
  t231_n214: return -0.6803434491157532f;
  t231_n215: return -0.03618226945400238f;
  t231_n216: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t231_n218;
  t231_n217: return -0.020047128200531006f;
  t231_n218: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t231_n226;
  t231_n219: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t231_n225;
  t231_n220: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t231_n222;
  t231_n221: return 0.006799724884331226f;
  t231_n222: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t231_n224;
  t231_n223: return 0.13805149495601654f;
  t231_n224: return 0.024627620354294777f;
  t231_n225: return -0.0008617429994046688f;
  t231_n226: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t231_n228;
  t231_n227: return 0.008870680816471577f;
  t231_n228: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t231_n236;
  t231_n229: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t231_n235;
  t231_n230: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t231_n232;
  t231_n231: return 0.11571932584047318f;
  t231_n232: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t231_n234;
  t231_n233: return 0.1148727610707283f;
  t231_n234: return 0.08729113638401031f;
  t231_n235: return -0.2586876451969147f;
  t231_n236: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t231_n240;
  t231_n237: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t231_n239;
  t231_n238: return 0.11721870303153992f;
  t231_n239: return -0.028943762183189392f;
  t231_n240: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t231_n242;
  t231_n241: return 0.12839961051940918f;
  t231_n242: return 0.03646324947476387f;
  t231_n243: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t231_n249;
  t231_n244: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t231_n246;
  t231_n245: return 0.00526876375079155f;
  t231_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t231_n248;
  t231_n247: return 0.037571027874946594f;
  t231_n248: return 0.09841771423816681f;
  t231_n249: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t231_n251;
  t231_n250: return 0.08767379075288773f;
  t231_n251: return -0.07292092591524124f;
  t231_n252: if (E.getNumReferences() >= 3300212736 /*1450.0*/) goto t231_n254;
  t231_n253: return 0.0022686882875859737f;
  t231_n254: return 0.023074399679899216f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree232(const Example& E) {
  t232_n0: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t232_n202;
  t232_n1: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t232_n151;
  t232_n2: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t232_n8;
  t232_n3: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t232_n5;
  t232_n4: return 0.0004088838759344071f;
  t232_n5: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t232_n7;
  t232_n6: return -0.0018896542023867369f;
  t232_n7: return -0.09088104218244553f;
  t232_n8: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t232_n134;
  t232_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t232_n11;
  t232_n10: return -0.034243226051330566f;
  t232_n11: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t232_n17;
  t232_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t232_n14;
  t232_n13: return 0.0007673390791751444f;
  t232_n14: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t232_n16;
  t232_n15: return 0.00326783605851233f;
  t232_n16: return -0.08371656388044357f;
  t232_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t232_n133;
  t232_n18: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t232_n56;
  t232_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n55;
  t232_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n38;
  t232_n21: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t232_n23;
  t232_n22: return -0.008320899680256844f;
  t232_n23: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n33;
  t232_n24: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t232_n28;
  t232_n25: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t232_n27;
  t232_n26: return 0.001015470246784389f;
  t232_n27: return 0.04128637537360191f;
  t232_n28: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t232_n32;
  t232_n29: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t232_n31;
  t232_n30: return 0.0930914357304573f;
  t232_n31: return 0.0674905776977539f;
  t232_n32: return -0.21576383709907532f;
  t232_n33: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t232_n35;
  t232_n34: return -0.007262249942868948f;
  t232_n35: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t232_n37;
  t232_n36: return -0.07156610488891602f;
  t232_n37: return -0.4182063341140747f;
  t232_n38: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t232_n54;
  t232_n39: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t232_n41;
  t232_n40: return 1.177601734525524e-05f;
  t232_n41: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t232_n51;
  t232_n42: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n50;
  t232_n43: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t232_n49;
  t232_n44: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t232_n46;
  t232_n45: return -0.08318889886140823f;
  t232_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n48;
  t232_n47: return -0.14443296194076538f;
  t232_n48: return 0.17008943855762482f;
  t232_n49: return 0.18488003313541412f;
  t232_n50: return -0.0827154591679573f;
  t232_n51: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t232_n53;
  t232_n52: return 0.14928379654884338f;
  t232_n53: return 0.01581507734954357f;
  t232_n54: return -0.004525091499090195f;
  t232_n55: return 0.01543132308870554f;
  t232_n56: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t232_n60;
  t232_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n59;
  t232_n58: return -0.2864881753921509f;
  t232_n59: return 0.07118688523769379f;
  t232_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n72;
  t232_n61: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t232_n71;
  t232_n62: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t232_n66;
  t232_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t232_n65;
  t232_n64: return -0.05725009739398956f;
  t232_n65: return -0.10154026746749878f;
  t232_n66: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t232_n70;
  t232_n67: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t232_n69;
  t232_n68: return 0.029370548203587532f;
  t232_n69: return 0.08105906844139099f;
  t232_n70: return -0.08486971259117126f;
  t232_n71: return 0.07236484438180923f;
  t232_n72: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t232_n74;
  t232_n73: return -0.2636924982070923f;
  t232_n74: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t232_n78;
  t232_n75: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t232_n77;
  t232_n76: return 0.0010763411410152912f;
  t232_n77: return 0.01010433305054903f;
  t232_n78: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t232_n100;
  t232_n79: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t232_n83;
  t232_n80: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n82;
  t232_n81: return -0.014077886939048767f;
  t232_n82: return -0.004460535477846861f;
  t232_n83: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t232_n97;
  t232_n84: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t232_n94;
  t232_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t232_n93;
  t232_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n88;
  t232_n87: return -0.023627379909157753f;
  t232_n88: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t232_n92;
  t232_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n91;
  t232_n90: return -0.10106798261404037f;
  t232_n91: return 0.07234644144773483f;
  t232_n92: return -0.017095422372221947f;
  t232_n93: return 0.019827693700790405f;
  t232_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n96;
  t232_n95: return -0.07452590763568878f;
  t232_n96: return 0.004858083091676235f;
  t232_n97: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n99;
  t232_n98: return 0.1062401831150055f;
  t232_n99: return -0.16821135580539703f;
  t232_n100: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t232_n108;
  t232_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n105;
  t232_n102: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t232_n104;
  t232_n103: return -0.1496574431657791f;
  t232_n104: return 0.0376690998673439f;
  t232_n105: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n107;
  t232_n106: return 0.021840572357177734f;
  t232_n107: return -0.03649406507611275f;
  t232_n108: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t232_n132;
  t232_n109: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t232_n129;
  t232_n110: if (E.getFractionNameInContext() >= 3207524079 /*0.6833333373069763*/) goto t232_n128;
  t232_n111: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t232_n121;
  t232_n112: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t232_n116;
  t232_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t232_n115;
  t232_n114: return 0.03563841059803963f;
  t232_n115: return 0.0593080036342144f;
  t232_n116: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t232_n118;
  t232_n117: return -0.042432840913534164f;
  t232_n118: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t232_n120;
  t232_n119: return 0.11755818873643875f;
  t232_n120: return -0.08748823404312134f;
  t232_n121: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t232_n123;
  t232_n122: return -0.0026246507186442614f;
  t232_n123: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t232_n125;
  t232_n124: return -0.19657978415489197f;
  t232_n125: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t232_n127;
  t232_n126: return 0.007709156256169081f;
  t232_n127: return 0.07856767624616623f;
  t232_n128: return 0.07494517415761948f;
  t232_n129: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t232_n131;
  t232_n130: return -0.35272884368896484f;
  t232_n131: return 0.004879368934780359f;
  t232_n132: return 0.005923907272517681f;
  t232_n133: return 0.0036059527192264795f;
  t232_n134: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t232_n150;
  t232_n135: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n147;
  t232_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t232_n144;
  t232_n137: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t232_n139;
  t232_n138: return -0.005921563133597374f;
  t232_n139: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t232_n143;
  t232_n140: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t232_n142;
  t232_n141: return 0.08055663853883743f;
  t232_n142: return 0.12646925449371338f;
  t232_n143: return -0.11375856399536133f;
  t232_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t232_n146;
  t232_n145: return -0.04469643905758858f;
  t232_n146: return 0.04770649969577789f;
  t232_n147: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t232_n149;
  t232_n148: return -0.0462484136223793f;
  t232_n149: return 0.044394511729478836f;
  t232_n150: return 0.10884400457143784f;
  t232_n151: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t232_n195;
  t232_n152: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t232_n172;
  t232_n153: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t232_n171;
  t232_n154: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t232_n168;
  t232_n155: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t232_n157;
  t232_n156: return -0.05275018885731697f;
  t232_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t232_n159;
  t232_n158: return -0.026781344786286354f;
  t232_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t232_n161;
  t232_n160: return -0.04506361111998558f;
  t232_n161: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t232_n163;
  t232_n162: return 0.060525666922330856f;
  t232_n163: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t232_n167;
  t232_n164: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t232_n166;
  t232_n165: return 0.1903803050518036f;
  t232_n166: return 0.39774417877197266f;
  t232_n167: return 0.14396962523460388f;
  t232_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n170;
  t232_n169: return -0.1554519683122635f;
  t232_n170: return -0.0604235902428627f;
  t232_n171: return 0.03460643067955971f;
  t232_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n184;
  t232_n173: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t232_n183;
  t232_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n178;
  t232_n175: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t232_n177;
  t232_n176: return -0.015747906640172005f;
  t232_n177: return -0.0786065012216568f;
  t232_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t232_n182;
  t232_n179: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t232_n181;
  t232_n180: return 0.007753151003271341f;
  t232_n181: return 0.07664073258638382f;
  t232_n182: return 0.05491660535335541f;
  t232_n183: return -0.014595955610275269f;
  t232_n184: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t232_n188;
  t232_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t232_n187;
  t232_n186: return -0.009898011572659016f;
  t232_n187: return 0.008991959504783154f;
  t232_n188: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t232_n192;
  t232_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n191;
  t232_n190: return -0.1715794950723648f;
  t232_n191: return 0.11706745624542236f;
  t232_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t232_n194;
  t232_n193: return -0.03532039746642113f;
  t232_n194: return 0.12033271044492722f;
  t232_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t232_n199;
  t232_n196: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t232_n198;
  t232_n197: return 0.022377945482730865f;
  t232_n198: return -0.09960882365703583f;
  t232_n199: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t232_n201;
  t232_n200: return 0.03189960867166519f;
  t232_n201: return 0.08621549606323242f;
  t232_n202: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t232_n218;
  t232_n203: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t232_n205;
  t232_n204: return 0.0808468610048294f;
  t232_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t232_n207;
  t232_n206: return -0.0025730284396559f;
  t232_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t232_n217;
  t232_n208: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t232_n214;
  t232_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t232_n213;
  t232_n210: if (E.getFractionNameInContext() >= 3212237678 /*0.9642857313156128*/) goto t232_n212;
  t232_n211: return -0.012085416354238987f;
  t232_n212: return 0.11826121807098389f;
  t232_n213: return 0.033820487558841705f;
  t232_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t232_n216;
  t232_n215: return -0.115547776222229f;
  t232_n216: return 0.044616300612688065f;
  t232_n217: return 0.035365089774131775f;
  t232_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t232_n252;
  t232_n219: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t232_n239;
  t232_n220: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t232_n238;
  t232_n221: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t232_n225;
  t232_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t232_n224;
  t232_n223: return 0.036080941557884216f;
  t232_n224: return 0.15727311372756958f;
  t232_n225: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t232_n237;
  t232_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n236;
  t232_n227: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t232_n231;
  t232_n228: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t232_n230;
  t232_n229: return -0.051464952528476715f;
  t232_n230: return 0.09315747767686844f;
  t232_n231: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t232_n235;
  t232_n232: if (E.getNumReferences() >= 3239051264 /*9.0*/) goto t232_n234;
  t232_n233: return -0.3393269181251526f;
  t232_n234: return -3.190878391265869f;
  t232_n235: return -0.1319388449192047f;
  t232_n236: return -0.01914958469569683f;
  t232_n237: return -0.0023001120425760746f;
  t232_n238: return 0.025548385456204414f;
  t232_n239: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t232_n249;
  t232_n240: if (E.getFractionNameInContext() >= 3177483234 /*0.05579746514558792*/) goto t232_n244;
  t232_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n243;
  t232_n242: return -0.031114069744944572f;
  t232_n243: return 0.03532892465591431f;
  t232_n244: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t232_n246;
  t232_n245: return -0.12694497406482697f;
  t232_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n248;
  t232_n247: return -0.08019766211509705f;
  t232_n248: return 0.03432459011673927f;
  t232_n249: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t232_n251;
  t232_n250: return -0.6021947860717773f;
  t232_n251: return 0.03403864800930023f;
  t232_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t232_n254;
  t232_n253: return -0.021281328052282333f;
  t232_n254: return 0.004012473858892918f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree233(const Example& E) {
  t233_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t233_n2;
  t233_n1: return -0.0003853313100989908f;
  t233_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t233_n254;
  t233_n3: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t233_n217;
  t233_n4: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t233_n68;
  t233_n5: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t233_n7;
  t233_n6: return -0.0007125415140762925f;
  t233_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t233_n51;
  t233_n8: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t233_n50;
  t233_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t233_n11;
  t233_n10: return -0.03442433848977089f;
  t233_n11: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t233_n41;
  t233_n12: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t233_n36;
  t233_n13: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t233_n35;
  t233_n14: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t233_n28;
  t233_n15: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t233_n19;
  t233_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t233_n18;
  t233_n17: return 0.07887981832027435f;
  t233_n18: return 0.2627253830432892f;
  t233_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t233_n21;
  t233_n20: return -0.03907783329486847f;
  t233_n21: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t233_n25;
  t233_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t233_n24;
  t233_n23: return -0.27455276250839233f;
  t233_n24: return 0.1371746063232422f;
  t233_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t233_n27;
  t233_n26: return -0.14891062676906586f;
  t233_n27: return 0.10625661164522171f;
  t233_n28: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t233_n32;
  t233_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t233_n31;
  t233_n30: return -0.00675092451274395f;
  t233_n31: return 0.08655235171318054f;
  t233_n32: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t233_n34;
  t233_n33: return -0.2578813433647156f;
  t233_n34: return -0.0429559089243412f;
  t233_n35: return 0.12025518715381622f;
  t233_n36: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t233_n40;
  t233_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t233_n39;
  t233_n38: return -0.18878336250782013f;
  t233_n39: return 0.07972007244825363f;
  t233_n40: return -0.005660921335220337f;
  t233_n41: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t233_n49;
  t233_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t233_n46;
  t233_n43: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t233_n45;
  t233_n44: return -0.18335780501365662f;
  t233_n45: return -0.23609966039657593f;
  t233_n46: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t233_n48;
  t233_n47: return -0.04523176699876785f;
  t233_n48: return -0.17005248367786407f;
  t233_n49: return -0.0022172476164996624f;
  t233_n50: return 0.00917547382414341f;
  t233_n51: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t233_n67;
  t233_n52: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t233_n54;
  t233_n53: return -0.2074476033449173f;
  t233_n54: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t233_n64;
  t233_n55: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t233_n63;
  t233_n56: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t233_n62;
  t233_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t233_n61;
  t233_n58: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t233_n60;
  t233_n59: return 0.01930234208703041f;
  t233_n60: return 0.19670052826404572f;
  t233_n61: return 0.04271992668509483f;
  t233_n62: return -0.08599353581666946f;
  t233_n63: return 0.008554128929972649f;
  t233_n64: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t233_n66;
  t233_n65: return -0.00543661555275321f;
  t233_n66: return 0.026477381587028503f;
  t233_n67: return -0.11656755208969116f;
  t233_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t233_n160;
  t233_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t233_n101;
  t233_n70: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t233_n84;
  t233_n71: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t233_n83;
  t233_n72: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t233_n74;
  t233_n73: return -0.0007381345494650304f;
  t233_n74: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t233_n76;
  t233_n75: return -0.014164800755679607f;
  t233_n76: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t233_n80;
  t233_n77: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t233_n79;
  t233_n78: return -0.002961815567687154f;
  t233_n79: return -0.028063781559467316f;
  t233_n80: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t233_n82;
  t233_n81: return 0.129124715924263f;
  t233_n82: return 0.018878119066357613f;
  t233_n83: return -0.08214855939149857f;
  t233_n84: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t233_n94;
  t233_n85: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t233_n89;
  t233_n86: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t233_n88;
  t233_n87: return 0.03146176412701607f;
  t233_n88: return -0.253724068403244f;
  t233_n89: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t233_n93;
  t233_n90: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t233_n92;
  t233_n91: return -0.13766108453273773f;
  t233_n92: return -0.6258599758148193f;
  t233_n93: return 0.09422796964645386f;
  t233_n94: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t233_n100;
  t233_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t233_n97;
  t233_n96: return -0.09982267767190933f;
  t233_n97: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t233_n99;
  t233_n98: return 0.06644442677497864f;
  t233_n99: return 0.016186203807592392f;
  t233_n100: return 0.005828942637890577f;
  t233_n101: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t233_n107;
  t233_n102: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t233_n106;
  t233_n103: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t233_n105;
  t233_n104: return -0.1007477343082428f;
  t233_n105: return 0.1411193311214447f;
  t233_n106: return -0.12765032052993774f;
  t233_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t233_n159;
  t233_n108: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t233_n118;
  t233_n109: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t233_n115;
  t233_n110: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t233_n112;
  t233_n111: return -0.013008114881813526f;
  t233_n112: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t233_n114;
  t233_n113: return 0.07401660829782486f;
  t233_n114: return -0.03461271524429321f;
  t233_n115: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t233_n117;
  t233_n116: return 0.01338502112776041f;
  t233_n117: return 0.10725896805524826f;
  t233_n118: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t233_n136;
  t233_n119: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t233_n123;
  t233_n120: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t233_n122;
  t233_n121: return -0.05207241699099541f;
  t233_n122: return 0.023314937949180603f;
  t233_n123: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t233_n129;
  t233_n124: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t233_n128;
  t233_n125: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t233_n127;
  t233_n126: return -0.06509532779455185f;
  t233_n127: return 0.15311163663864136f;
  t233_n128: return -0.030748669058084488f;
  t233_n129: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t233_n133;
  t233_n130: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t233_n132;
  t233_n131: return -0.383034884929657f;
  t233_n132: return -0.1378777176141739f;
  t233_n133: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t233_n135;
  t233_n134: return 0.017633311450481415f;
  t233_n135: return 0.10315870493650436f;
  t233_n136: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t233_n156;
  t233_n137: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t233_n145;
  t233_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t233_n140;
  t233_n139: return -0.19410340487957f;
  t233_n140: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t233_n144;
  t233_n141: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t233_n143;
  t233_n142: return 0.09417817741632462f;
  t233_n143: return -0.12900494039058685f;
  t233_n144: return 0.03910874202847481f;
  t233_n145: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t233_n147;
  t233_n146: return 0.005985877010971308f;
  t233_n147: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t233_n149;
  t233_n148: return -0.45594385266304016f;
  t233_n149: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t233_n155;
  t233_n150: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t233_n152;
  t233_n151: return 0.054382067173719406f;
  t233_n152: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t233_n154;
  t233_n153: return -0.12432394176721573f;
  t233_n154: return 0.03381642326712608f;
  t233_n155: return 0.0696600005030632f;
  t233_n156: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t233_n158;
  t233_n157: return -0.2626609206199646f;
  t233_n158: return 0.05090876296162605f;
  t233_n159: return -0.00303442170843482f;
  t233_n160: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t233_n186;
  t233_n161: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t233_n185;
  t233_n162: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t233_n176;
  t233_n163: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t233_n165;
  t233_n164: return -0.03478241711854935f;
  t233_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t233_n167;
  t233_n166: return -0.06010898947715759f;
  t233_n167: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t233_n175;
  t233_n168: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t233_n174;
  t233_n169: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t233_n171;
  t233_n170: return 0.0872524082660675f;
  t233_n171: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t233_n173;
  t233_n172: return 0.09797866642475128f;
  t233_n173: return 0.03675667196512222f;
  t233_n174: return 0.1518704742193222f;
  t233_n175: return 0.06020495668053627f;
  t233_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t233_n180;
  t233_n177: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t233_n179;
  t233_n178: return -0.0879228487610817f;
  t233_n179: return 0.09627348184585571f;
  t233_n180: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t233_n184;
  t233_n181: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t233_n183;
  t233_n182: return 0.11147577315568924f;
  t233_n183: return -0.09532208740711212f;
  t233_n184: return 0.004096746910363436f;
  t233_n185: return 0.14700408279895782f;
  t233_n186: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t233_n190;
  t233_n187: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t233_n189;
  t233_n188: return -0.2907702922821045f;
  t233_n189: return 0.24354028701782227f;
  t233_n190: if (E.getFractionNameInContext() >= 3211904796 /*0.9444444179534912*/) goto t233_n216;
  t233_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t233_n193;
  t233_n192: return 0.0015650630230084062f;
  t233_n193: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t233_n197;
  t233_n194: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t233_n196;
  t233_n195: return -0.1662372350692749f;
  t233_n196: return -0.33556801080703735f;
  t233_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t233_n199;
  t233_n198: return 0.009178970009088516f;
  t233_n199: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t233_n211;
  t233_n200: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t233_n210;
  t233_n201: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t233_n207;
  t233_n202: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t233_n204;
  t233_n203: return 0.08341947197914124f;
  t233_n204: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t233_n206;
  t233_n205: return -0.008152026683092117f;
  t233_n206: return 0.05203501507639885f;
  t233_n207: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t233_n209;
  t233_n208: return 0.01162432786077261f;
  t233_n209: return 0.07300399988889694f;
  t233_n210: return -0.004571117460727692f;
  t233_n211: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t233_n215;
  t233_n212: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t233_n214;
  t233_n213: return 0.1406765729188919f;
  t233_n214: return 0.11632600426673889f;
  t233_n215: return -0.09346214681863785f;
  t233_n216: return -0.02939712628722191f;
  t233_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t233_n249;
  t233_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t233_n220;
  t233_n219: return -0.07486077398061752f;
  t233_n220: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t233_n226;
  t233_n221: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t233_n223;
  t233_n222: return -0.01758730039000511f;
  t233_n223: if (E.getFractionNameInContext() >= 3211159142 /*0.8999999761581421*/) goto t233_n225;
  t233_n224: return -0.20043796300888062f;
  t233_n225: return 0.1108512431383133f;
  t233_n226: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t233_n246;
  t233_n227: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t233_n229;
  t233_n228: return 0.00962760578840971f;
  t233_n229: if (E.getNumReferences() >= 3294167040 /*868.0*/) goto t233_n245;
  t233_n230: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t233_n236;
  t233_n231: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t233_n233;
  t233_n232: return -0.12982399761676788f;
  t233_n233: if (E.getNumReferences() >= 3241148416 /*11.0*/) goto t233_n235;
  t233_n234: return 0.0707104280591011f;
  t233_n235: return 0.23487593233585358f;
  t233_n236: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t233_n238;
  t233_n237: return 0.007908961735665798f;
  t233_n238: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t233_n244;
  t233_n239: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t233_n243;
  t233_n240: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t233_n242;
  t233_n241: return 0.10467514395713806f;
  t233_n242: return 0.16958360373973846f;
  t233_n243: return -0.034621164202690125f;
  t233_n244: return -0.03598994016647339f;
  t233_n245: return 0.1412079632282257f;
  t233_n246: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t233_n248;
  t233_n247: return -0.1416633427143097f;
  t233_n248: return -0.006118709687143564f;
  t233_n249: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t233_n253;
  t233_n250: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t233_n252;
  t233_n251: return 0.0065057347528636456f;
  t233_n252: return 0.03853778913617134f;
  t233_n253: return -0.07672549039125443f;
  t233_n254: return 0.0017496675718575716f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree234(const Example& E) {
  t234_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t234_n250;
  t234_n1: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t234_n177;
  t234_n2: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t234_n176;
  t234_n3: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t234_n173;
  t234_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t234_n162;
  t234_n5: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t234_n143;
  t234_n6: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t234_n92;
  t234_n7: if (E.getNumReferences() >= 3289456640 /*580.5*/) goto t234_n71;
  t234_n8: if (E.getNumReferences() >= 3288719360 /*535.5*/) goto t234_n68;
  t234_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t234_n27;
  t234_n10: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t234_n24;
  t234_n11: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t234_n21;
  t234_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t234_n14;
  t234_n13: return -0.1446685492992401f;
  t234_n14: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t234_n18;
  t234_n15: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t234_n17;
  t234_n16: return 0.016134001314640045f;
  t234_n17: return -0.021441087126731873f;
  t234_n18: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t234_n20;
  t234_n19: return -0.14203451573848724f;
  t234_n20: return 0.07446866482496262f;
  t234_n21: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t234_n23;
  t234_n22: return -0.07761352509260178f;
  t234_n23: return 0.0634869635105133f;
  t234_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t234_n26;
  t234_n25: return -0.2615779638290405f;
  t234_n26: return 0.027179529890418053f;
  t234_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t234_n51;
  t234_n28: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t234_n50;
  t234_n29: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n49;
  t234_n30: if (E.getNumReferences() >= 3284451328 /*393.5*/) goto t234_n48;
  t234_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t234_n45;
  t234_n32: if (E.getNumReferences() >= 3284336640 /*390.0*/) goto t234_n44;
  t234_n33: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t234_n43;
  t234_n34: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t234_n40;
  t234_n35: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t234_n37;
  t234_n36: return -0.010529069229960442f;
  t234_n37: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t234_n39;
  t234_n38: return -0.3210557699203491f;
  t234_n39: return 0.21020197868347168f;
  t234_n40: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t234_n42;
  t234_n41: return 0.014387411996722221f;
  t234_n42: return -0.012157432734966278f;
  t234_n43: return 0.004108667839318514f;
  t234_n44: return -0.5606075525283813f;
  t234_n45: if (E.getNumReferences() >= 3270705152 /*121.5*/) goto t234_n47;
  t234_n46: return 0.005362443160265684f;
  t234_n47: return -0.02787509746849537f;
  t234_n48: return 0.03377389907836914f;
  t234_n49: return -0.011609993875026703f;
  t234_n50: return 0.10530268400907516f;
  t234_n51: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t234_n67;
  t234_n52: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t234_n56;
  t234_n53: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t234_n55;
  t234_n54: return 0.044945843517780304f;
  t234_n55: return 0.08749788999557495f;
  t234_n56: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n66;
  t234_n57: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t234_n59;
  t234_n58: return -0.014140366576611996f;
  t234_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t234_n65;
  t234_n60: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t234_n62;
  t234_n61: return -0.22706158459186554f;
  t234_n62: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t234_n64;
  t234_n63: return -0.5362542271614075f;
  t234_n64: return -0.11436574161052704f;
  t234_n65: return 0.049301888793706894f;
  t234_n66: return 0.030016478151082993f;
  t234_n67: return 0.002187741454690695f;
  t234_n68: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n70;
  t234_n69: return -0.13148446381092072f;
  t234_n70: return -0.05649390444159508f;
  t234_n71: if (E.getNumReferences() >= 3290914816 /*669.5*/) goto t234_n89;
  t234_n72: if (E.getNumReferences() >= 3290890240 /*668.0*/) goto t234_n84;
  t234_n73: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t234_n75;
  t234_n74: return 0.029715709388256073f;
  t234_n75: if (E.getNumReferences() >= 3290071040 /*618.0*/) goto t234_n77;
  t234_n76: return -0.0186233501881361f;
  t234_n77: if (E.getNumReferences() >= 3290128384 /*621.5*/) goto t234_n79;
  t234_n78: return 0.18873882293701172f;
  t234_n79: if (E.getNumReferences() >= 3290669056 /*654.5*/) goto t234_n81;
  t234_n80: return -0.0408974289894104f;
  t234_n81: if (E.getNumReferences() >= 3290767360 /*660.5*/) goto t234_n83;
  t234_n82: return 0.29479196667671204f;
  t234_n83: return -0.10047280043363571f;
  t234_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n88;
  t234_n85: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t234_n87;
  t234_n86: return 0.03537347540259361f;
  t234_n87: return 0.1560979187488556f;
  t234_n88: return 0.08322571218013763f;
  t234_n89: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t234_n91;
  t234_n90: return 0.004955622833222151f;
  t234_n91: return 0.022341767325997353f;
  t234_n92: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t234_n110;
  t234_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t234_n107;
  t234_n94: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t234_n102;
  t234_n95: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t234_n99;
  t234_n96: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t234_n98;
  t234_n97: return -0.042994823306798935f;
  t234_n98: return -0.10326321423053741f;
  t234_n99: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t234_n101;
  t234_n100: return -0.17323726415634155f;
  t234_n101: return -0.005525550339370966f;
  t234_n102: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t234_n106;
  t234_n103: if (E.getNumReferences() >= 3280388096 /*269.5*/) goto t234_n105;
  t234_n104: return 0.0481991320848465f;
  t234_n105: return 0.20592370629310608f;
  t234_n106: return -0.02070043236017227f;
  t234_n107: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t234_n109;
  t234_n108: return -0.007952232845127583f;
  t234_n109: return 0.021309807896614075f;
  t234_n110: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t234_n138;
  t234_n111: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t234_n119;
  t234_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t234_n118;
  t234_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t234_n115;
  t234_n114: return 0.02030777744948864f;
  t234_n115: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n117;
  t234_n116: return 0.13862670958042145f;
  t234_n117: return 0.021746618673205376f;
  t234_n118: return 0.07791082561016083f;
  t234_n119: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t234_n135;
  t234_n120: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t234_n134;
  t234_n121: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t234_n131;
  t234_n122: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t234_n130;
  t234_n123: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t234_n125;
  t234_n124: return -0.0337059386074543f;
  t234_n125: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n127;
  t234_n126: return 0.020754940807819366f;
  t234_n127: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t234_n129;
  t234_n128: return 0.06764135509729385f;
  t234_n129: return 0.13125687837600708f;
  t234_n130: return -0.2600895166397095f;
  t234_n131: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t234_n133;
  t234_n132: return 0.09630819410085678f;
  t234_n133: return -0.016084972769021988f;
  t234_n134: return 0.0713755339384079f;
  t234_n135: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t234_n137;
  t234_n136: return 0.06103716045618057f;
  t234_n137: return 0.1064574271440506f;
  t234_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t234_n140;
  t234_n139: return -0.03588121384382248f;
  t234_n140: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t234_n142;
  t234_n141: return 0.06495200842618942f;
  t234_n142: return -0.047679074108600616f;
  t234_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t234_n159;
  t234_n144: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t234_n146;
  t234_n145: return 0.0011117300018668175f;
  t234_n146: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t234_n148;
  t234_n147: return 0.018683496862649918f;
  t234_n148: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t234_n154;
  t234_n149: if (E.getNumReferences() >= 3265331200 /*80.5*/) goto t234_n151;
  t234_n150: return -0.002773136366158724f;
  t234_n151: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n153;
  t234_n152: return -0.10365785658359528f;
  t234_n153: return 0.23941954970359802f;
  t234_n154: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t234_n156;
  t234_n155: return 0.07295607775449753f;
  t234_n156: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t234_n158;
  t234_n157: return 0.0973920002579689f;
  t234_n158: return -0.6466081142425537f;
  t234_n159: if (E.getNumReferences() >= 3289464832 /*581.0*/) goto t234_n161;
  t234_n160: return 0.03372682258486748f;
  t234_n161: return -0.12146911770105362f;
  t234_n162: if (E.getNumReferences() >= 3323256320 /*9527.5*/) goto t234_n172;
  t234_n163: if (E.getNumReferences() >= 3322464256 /*8754.0*/) goto t234_n165;
  t234_n164: return 0.004773812834173441f;
  t234_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t234_n167;
  t234_n166: return 0.010272413492202759f;
  t234_n167: if (E.getNumReferences() >= 3322692096 /*8976.5*/) goto t234_n169;
  t234_n168: return 0.5219702124595642f;
  t234_n169: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t234_n171;
  t234_n170: return 0.19443324208259583f;
  t234_n171: return 0.10810165852308273f;
  t234_n172: return -0.0189491156488657f;
  t234_n173: if (E.getNumReferences() >= 3367700464 /*383039.5*/) goto t234_n175;
  t234_n174: return -0.005069971550256014f;
  t234_n175: return -0.04809238761663437f;
  t234_n176: return 0.0051068393513560295f;
  t234_n177: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t234_n249;
  t234_n178: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t234_n202;
  t234_n179: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t234_n193;
  t234_n180: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t234_n184;
  t234_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t234_n183;
  t234_n182: return -0.018527546897530556f;
  t234_n183: return -0.017123078927397728f;
  t234_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t234_n186;
  t234_n185: return -0.012830281630158424f;
  t234_n186: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t234_n188;
  t234_n187: return 0.03836053982377052f;
  t234_n188: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t234_n192;
  t234_n189: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t234_n191;
  t234_n190: return 0.033291302621364594f;
  t234_n191: return 0.0897272601723671f;
  t234_n192: return -0.07504977285861969f;
  t234_n193: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t234_n199;
  t234_n194: if (E.getNumReferences() >= 3314356224 /*4514.0*/) goto t234_n196;
  t234_n195: return 0.012236077338457108f;
  t234_n196: if (E.getNumReferences() >= 3327250432 /*13428.0*/) goto t234_n198;
  t234_n197: return 0.06587956845760345f;
  t234_n198: return 0.0161591786891222f;
  t234_n199: if (E.getNumReferences() >= 3361153472 /*220295.0*/) goto t234_n201;
  t234_n200: return -0.007108989637345076f;
  t234_n201: return -0.14056625962257385f;
  t234_n202: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t234_n246;
  t234_n203: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t234_n225;
  t234_n204: if (E.getNumReferences() >= 3326581248 /*12774.5*/) goto t234_n222;
  t234_n205: if (E.getNumReferences() >= 3325340672 /*11563.0*/) goto t234_n221;
  t234_n206: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t234_n220;
  t234_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t234_n209;
  t234_n208: return 0.015918442979454994f;
  t234_n209: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t234_n217;
  t234_n210: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t234_n216;
  t234_n211: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t234_n215;
  t234_n212: if (E.getFractionNameInContext() >= 3182156850 /*0.0839160829782486*/) goto t234_n214;
  t234_n213: return 0.0639728382229805f;
  t234_n214: return 0.11705132573843002f;
  t234_n215: return 0.07307405024766922f;
  t234_n216: return -0.18973274528980255f;
  t234_n217: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t234_n219;
  t234_n218: return 0.02559436485171318f;
  t234_n219: return -0.28603848814964294f;
  t234_n220: return 0.002725684316828847f;
  t234_n221: return 0.12290205806493759f;
  t234_n222: if (E.getNumReferences() >= 3336376576 /*28296.5*/) goto t234_n224;
  t234_n223: return -0.08550867438316345f;
  t234_n224: return 0.006121182814240456f;
  t234_n225: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t234_n245;
  t234_n226: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t234_n244;
  t234_n227: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t234_n235;
  t234_n228: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t234_n234;
  t234_n229: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t234_n231;
  t234_n230: return -1.0803390741348267f;
  t234_n231: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t234_n233;
  t234_n232: return -1.5804709196090698f;
  t234_n233: return -0.08811349421739578f;
  t234_n234: return -0.027056235820055008f;
  t234_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t234_n241;
  t234_n236: if (E.getNumReferences() >= 3346332288 /*62714.5*/) goto t234_n240;
  t234_n237: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t234_n239;
  t234_n238: return -0.07611553370952606f;
  t234_n239: return -0.7727490067481995f;
  t234_n240: return 0.06822828203439713f;
  t234_n241: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t234_n243;
  t234_n242: return 0.12718376517295837f;
  t234_n243: return -0.019224276766180992f;
  t234_n244: return 0.017749935388565063f;
  t234_n245: return -0.0007462198846042156f;
  t234_n246: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t234_n248;
  t234_n247: return 0.09444493055343628f;
  t234_n248: return 0.030839983373880386f;
  t234_n249: return -0.0005566305480897427f;
  t234_n250: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t234_n252;
  t234_n251: return 0.0005482733831740916f;
  t234_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t234_n254;
  t234_n253: return 0.0008736902382224798f;
  t234_n254: return 0.006080944091081619f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree235(const Example& E) {
  t235_n0: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t235_n2;
  t235_n1: return -0.0004815727879758924f;
  t235_n2: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t235_n252;
  t235_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n161;
  t235_n4: if (E.getNumReferences() >= 3294044160 /*860.5*/) goto t235_n160;
  t235_n5: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t235_n9;
  t235_n6: if (E.getNumReferences() >= 3280551936 /*274.5*/) goto t235_n8;
  t235_n7: return 0.0005023187259212136f;
  t235_n8: return 0.03779527172446251f;
  t235_n9: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t235_n153;
  t235_n10: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t235_n142;
  t235_n11: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t235_n131;
  t235_n12: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t235_n84;
  t235_n13: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t235_n63;
  t235_n14: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t235_n44;
  t235_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown))) goto t235_n43;
  t235_n16: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t235_n40;
  t235_n17: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t235_n31;
  t235_n18: if (E.getNumReferences() >= 3291013120 /*675.5*/) goto t235_n28;
  t235_n19: if (E.getNumReferences() >= 3290308608 /*632.5*/) goto t235_n27;
  t235_n20: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n22;
  t235_n21: return -0.01385472808033228f;
  t235_n22: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t235_n24;
  t235_n23: return -0.06444521248340607f;
  t235_n24: if (E.getNumReferences() >= 3281551360 /*305.0*/) goto t235_n26;
  t235_n25: return 0.018030844628810883f;
  t235_n26: return -0.3352019786834717f;
  t235_n27: return -0.7526670098304749f;
  t235_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n30;
  t235_n29: return 0.0542309544980526f;
  t235_n30: return 0.19960369169712067f;
  t235_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n39;
  t235_n32: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t235_n36;
  t235_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n35;
  t235_n34: return -0.11361410468816757f;
  t235_n35: return -0.02623804472386837f;
  t235_n36: if (E.getNumReferences() >= 3260416000 /*53.5*/) goto t235_n38;
  t235_n37: return -0.05090338736772537f;
  t235_n38: return 0.020846592262387276f;
  t235_n39: return -0.004163412377238274f;
  t235_n40: if (E.getNumReferences() >= 3293315072 /*816.0*/) goto t235_n42;
  t235_n41: return 0.0018066087504848838f;
  t235_n42: return -0.8798559308052063f;
  t235_n43: return -0.0021833789069205523f;
  t235_n44: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t235_n46;
  t235_n45: return 0.018997950479388237f;
  t235_n46: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t235_n54;
  t235_n47: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t235_n53;
  t235_n48: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t235_n52;
  t235_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n51;
  t235_n50: return -0.7390273213386536f;
  t235_n51: return 0.04891767352819443f;
  t235_n52: return -0.8514857888221741f;
  t235_n53: return -0.01975961960852146f;
  t235_n54: if (E.getNumReferences() >= 3273031680 /*150.5*/) goto t235_n60;
  t235_n55: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t235_n59;
  t235_n56: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t235_n58;
  t235_n57: return 0.12992353737354279f;
  t235_n58: return 0.0723607987165451f;
  t235_n59: return -0.04120840132236481f;
  t235_n60: if (E.getNumReferences() >= 3278340096 /*231.5*/) goto t235_n62;
  t235_n61: return -1.260669469833374f;
  t235_n62: return -0.09538737684488297f;
  t235_n63: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t235_n65;
  t235_n64: return -0.047878507524728775f;
  t235_n65: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n77;
  t235_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n70;
  t235_n67: if (E.getFractionNameInContext() >= 3191155539 /*0.17692308127880096*/) goto t235_n69;
  t235_n68: return -0.3453295826911926f;
  t235_n69: return -0.10045741498470306f;
  t235_n70: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t235_n74;
  t235_n71: if (E.getNumReferences() >= 3280683008 /*278.5*/) goto t235_n73;
  t235_n72: return 0.02373398095369339f;
  t235_n73: return -0.2780923843383789f;
  t235_n74: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t235_n76;
  t235_n75: return 0.002908359281718731f;
  t235_n76: return -0.6047995090484619f;
  t235_n77: if (E.getNumReferences() >= 3239051264 /*9.0*/) goto t235_n83;
  t235_n78: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t235_n80;
  t235_n79: return 0.06169416382908821f;
  t235_n80: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t235_n82;
  t235_n81: return 0.18310362100601196f;
  t235_n82: return 0.12508341670036316f;
  t235_n83: return -0.4905613660812378f;
  t235_n84: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t235_n110;
  t235_n85: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t235_n109;
  t235_n86: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t235_n108;
  t235_n87: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t235_n103;
  t235_n88: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t235_n102;
  t235_n89: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t235_n95;
  t235_n90: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t235_n94;
  t235_n91: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t235_n93;
  t235_n92: return 0.026089869439601898f;
  t235_n93: return 0.07534157484769821f;
  t235_n94: return -0.3341684937477112f;
  t235_n95: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n101;
  t235_n96: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t235_n98;
  t235_n97: return 0.01454906165599823f;
  t235_n98: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t235_n100;
  t235_n99: return 0.12888339161872864f;
  t235_n100: return 0.015183710493147373f;
  t235_n101: return -0.03282585367560387f;
  t235_n102: return 0.1440214365720749f;
  t235_n103: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t235_n105;
  t235_n104: return -0.04601944610476494f;
  t235_n105: if (E.getNumReferences() >= 3274571776 /*174.0*/) goto t235_n107;
  t235_n106: return -0.5282344222068787f;
  t235_n107: return -0.016690779477357864f;
  t235_n108: return -0.25352850556373596f;
  t235_n109: return 0.11046498268842697f;
  t235_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n128;
  t235_n111: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n127;
  t235_n112: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t235_n118;
  t235_n113: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n117;
  t235_n114: if (E.getNumReferences() >= 3257925632 /*44.0*/) goto t235_n116;
  t235_n115: return -0.28307172656059265f;
  t235_n116: return -0.1150425523519516f;
  t235_n117: return -0.016366146504878998f;
  t235_n118: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t235_n122;
  t235_n119: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n121;
  t235_n120: return -0.4465772211551666f;
  t235_n121: return -0.12682892382144928f;
  t235_n122: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n126;
  t235_n123: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t235_n125;
  t235_n124: return 0.061343852430582047f;
  t235_n125: return 0.10546383261680603f;
  t235_n126: return -0.02761111408472061f;
  t235_n127: return -0.018832022324204445f;
  t235_n128: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t235_n130;
  t235_n129: return -0.02481439709663391f;
  t235_n130: return 0.03947034105658531f;
  t235_n131: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t235_n141;
  t235_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n138;
  t235_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n137;
  t235_n134: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t235_n136;
  t235_n135: return -0.34226199984550476f;
  t235_n136: return -0.22354434430599213f;
  t235_n137: return -0.0609654039144516f;
  t235_n138: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t235_n140;
  t235_n139: return -0.05685662850737572f;
  t235_n140: return 0.14845478534698486f;
  t235_n141: return 0.04988927021622658f;
  t235_n142: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t235_n152;
  t235_n143: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t235_n149;
  t235_n144: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t235_n148;
  t235_n145: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t235_n147;
  t235_n146: return -0.19009092450141907f;
  t235_n147: return 0.002512378618121147f;
  t235_n148: return 0.0173481497913599f;
  t235_n149: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t235_n151;
  t235_n150: return 0.11679843068122864f;
  t235_n151: return -0.18359963595867157f;
  t235_n152: return -0.07907278090715408f;
  t235_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t235_n159;
  t235_n154: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t235_n156;
  t235_n155: return 0.018421003594994545f;
  t235_n156: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t235_n158;
  t235_n157: return -0.0595950186252594f;
  t235_n158: return 0.12717942893505096f;
  t235_n159: return 0.12488072365522385f;
  t235_n160: return 0.01045424398034811f;
  t235_n161: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t235_n227;
  t235_n162: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t235_n188;
  t235_n163: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t235_n179;
  t235_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t235_n178;
  t235_n165: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t235_n169;
  t235_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n168;
  t235_n167: return -0.20554488897323608f;
  t235_n168: return -0.0013474200386554003f;
  t235_n169: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t235_n173;
  t235_n170: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t235_n172;
  t235_n171: return 0.035977091640233994f;
  t235_n172: return 0.2534588575363159f;
  t235_n173: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t235_n175;
  t235_n174: return 0.12577876448631287f;
  t235_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n177;
  t235_n176: return -0.04981169477105141f;
  t235_n177: return 0.19619475305080414f;
  t235_n178: return 0.029160279780626297f;
  t235_n179: if (E.getFractionNameInContext() >= 3198076140 /*0.3100961446762085*/) goto t235_n183;
  t235_n180: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t235_n182;
  t235_n181: return 0.020465679466724396f;
  t235_n182: return 0.005947585683315992f;
  t235_n183: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t235_n187;
  t235_n184: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t235_n186;
  t235_n185: return -0.06895647197961807f;
  t235_n186: return 0.0018020633142441511f;
  t235_n187: return 0.0037946393713355064f;
  t235_n188: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t235_n194;
  t235_n189: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t235_n193;
  t235_n190: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n192;
  t235_n191: return -0.0046807024627923965f;
  t235_n192: return 0.004312308970838785f;
  t235_n193: return -0.026737239211797714f;
  t235_n194: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t235_n216;
  t235_n195: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t235_n215;
  t235_n196: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t235_n206;
  t235_n197: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t235_n205;
  t235_n198: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t235_n204;
  t235_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t235_n201;
  t235_n200: return 0.014549687504768372f;
  t235_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t235_n203;
  t235_n202: return 0.06026136502623558f;
  t235_n203: return 0.12899242341518402f;
  t235_n204: return 0.08948171138763428f;
  t235_n205: return 0.08813273906707764f;
  t235_n206: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t235_n214;
  t235_n207: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n213;
  t235_n208: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t235_n210;
  t235_n209: return 0.07065144181251526f;
  t235_n210: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t235_n212;
  t235_n211: return 0.08260753750801086f;
  t235_n212: return 0.1281125843524933f;
  t235_n213: return -0.04276003688573837f;
  t235_n214: return -0.14308132231235504f;
  t235_n215: return 0.011377382092177868f;
  t235_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n218;
  t235_n217: return -0.04593617841601372f;
  t235_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n226;
  t235_n219: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t235_n225;
  t235_n220: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t235_n224;
  t235_n221: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t235_n223;
  t235_n222: return -0.5508273839950562f;
  t235_n223: return -0.09266919642686844f;
  t235_n224: return -0.003216107375919819f;
  t235_n225: return -0.282368540763855f;
  t235_n226: return 0.0161000806838274f;
  t235_n227: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t235_n235;
  t235_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n234;
  t235_n229: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t235_n233;
  t235_n230: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t235_n232;
  t235_n231: return 0.08717963844537735f;
  t235_n232: return -0.9475957155227661f;
  t235_n233: return -0.06293074041604996f;
  t235_n234: return 0.00566534698009491f;
  t235_n235: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t235_n247;
  t235_n236: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t235_n246;
  t235_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t235_n245;
  t235_n238: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t235_n244;
  t235_n239: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t235_n241;
  t235_n240: return -0.05201488733291626f;
  t235_n241: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t235_n243;
  t235_n242: return 0.05214018374681473f;
  t235_n243: return -0.30042755603790283f;
  t235_n244: return 0.047261793166399f;
  t235_n245: return 0.06239712983369827f;
  t235_n246: return 0.007147838827222586f;
  t235_n247: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t235_n249;
  t235_n248: return 0.01226062886416912f;
  t235_n249: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t235_n251;
  t235_n250: return 0.5686025023460388f;
  t235_n251: return 0.02125779166817665f;
  t235_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t235_n254;
  t235_n253: return 0.00019040994811803102f;
  t235_n254: return 0.0044287145137786865f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree236(const Example& E) {
  t236_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t236_n2;
  t236_n1: return -0.0001645968877710402f;
  t236_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t236_n244;
  t236_n3: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t236_n235;
  t236_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n8;
  t236_n5: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t236_n7;
  t236_n6: return -0.016872186213731766f;
  t236_n7: return -0.10736044496297836f;
  t236_n8: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t236_n80;
  t236_n9: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t236_n61;
  t236_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n32;
  t236_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t236_n29;
  t236_n12: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t236_n14;
  t236_n13: return -0.0021849237382411957f;
  t236_n14: if (E.getSymbolScopeDistanceCost() >= 3248488448 /*20.0*/) goto t236_n18;
  t236_n15: if (E.getNumReferences() >= 3265265664 /*80.0*/) goto t236_n17;
  t236_n16: return -0.06497550010681152f;
  t236_n17: return -0.32137203216552734f;
  t236_n18: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t236_n20;
  t236_n19: return -0.02376948855817318f;
  t236_n20: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t236_n28;
  t236_n21: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t236_n23;
  t236_n22: return 0.033674631267786026f;
  t236_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t236_n25;
  t236_n24: return 0.06596869975328445f;
  t236_n25: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t236_n27;
  t236_n26: return 0.1623028814792633f;
  t236_n27: return 0.14201897382736206f;
  t236_n28: return -0.0529729500412941f;
  t236_n29: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t236_n31;
  t236_n30: return 0.029070686548948288f;
  t236_n31: return -0.016581695526838303f;
  t236_n32: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t236_n60;
  t236_n33: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t236_n53;
  t236_n34: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t236_n42;
  t236_n35: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t236_n41;
  t236_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n40;
  t236_n37: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t236_n39;
  t236_n38: return -0.040420934557914734f;
  t236_n39: return 0.09194854646921158f;
  t236_n40: return 0.020914671942591667f;
  t236_n41: return -0.02213161066174507f;
  t236_n42: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t236_n52;
  t236_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t236_n47;
  t236_n44: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t236_n46;
  t236_n45: return 0.014987083151936531f;
  t236_n46: return -0.9446399807929993f;
  t236_n47: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t236_n49;
  t236_n48: return 0.08678317815065384f;
  t236_n49: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t236_n51;
  t236_n50: return -0.28706130385398865f;
  t236_n51: return 0.03443703055381775f;
  t236_n52: return 0.33663415908813477f;
  t236_n53: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t236_n59;
  t236_n54: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t236_n58;
  t236_n55: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t236_n57;
  t236_n56: return 0.15433956682682037f;
  t236_n57: return -0.36579757928848267f;
  t236_n58: return 0.0280914343893528f;
  t236_n59: return -0.10282421112060547f;
  t236_n60: return 0.09173119813203812f;
  t236_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n71;
  t236_n62: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t236_n68;
  t236_n63: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t236_n65;
  t236_n64: return -0.010467610321938992f;
  t236_n65: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t236_n67;
  t236_n66: return 0.011053035967051983f;
  t236_n67: return -0.06243978068232536f;
  t236_n68: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t236_n70;
  t236_n69: return -0.05383343622088432f;
  t236_n70: return 0.05003568157553673f;
  t236_n71: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t236_n77;
  t236_n72: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t236_n74;
  t236_n73: return -0.07353000342845917f;
  t236_n74: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t236_n76;
  t236_n75: return -0.022397173568606377f;
  t236_n76: return 0.05515848845243454f;
  t236_n77: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t236_n79;
  t236_n78: return 0.036657217890024185f;
  t236_n79: return 0.004400012549012899f;
  t236_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t236_n138;
  t236_n81: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t236_n131;
  t236_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n92;
  t236_n83: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t236_n87;
  t236_n84: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t236_n86;
  t236_n85: return -0.01709551177918911f;
  t236_n86: return 0.07238324731588364f;
  t236_n87: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t236_n91;
  t236_n88: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t236_n90;
  t236_n89: return -1.7015914916992188f;
  t236_n90: return 0.0707327350974083f;
  t236_n91: return -0.09821166098117828f;
  t236_n92: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t236_n120;
  t236_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n111;
  t236_n94: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t236_n108;
  t236_n95: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t236_n107;
  t236_n96: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t236_n98;
  t236_n97: return -0.44199037551879883f;
  t236_n98: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t236_n102;
  t236_n99: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t236_n101;
  t236_n100: return -0.004559963941574097f;
  t236_n101: return -0.2138472944498062f;
  t236_n102: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t236_n106;
  t236_n103: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t236_n105;
  t236_n104: return -0.7866132855415344f;
  t236_n105: return -0.003217885969206691f;
  t236_n106: return -0.09841308742761612f;
  t236_n107: return 0.04450180381536484f;
  t236_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t236_n110;
  t236_n109: return -0.060588520020246506f;
  t236_n110: return 0.09746614098548889f;
  t236_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t236_n113;
  t236_n112: return -0.00411204295232892f;
  t236_n113: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t236_n119;
  t236_n114: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t236_n118;
  t236_n115: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t236_n117;
  t236_n116: return 0.1267922967672348f;
  t236_n117: return 0.15008045732975006f;
  t236_n118: return 0.054554082453250885f;
  t236_n119: return -0.3297276794910431f;
  t236_n120: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t236_n122;
  t236_n121: return 0.0063812923617661f;
  t236_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t236_n128;
  t236_n123: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t236_n125;
  t236_n124: return 0.041399795562028885f;
  t236_n125: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t236_n127;
  t236_n126: return -0.8918111324310303f;
  t236_n127: return -0.023509632796049118f;
  t236_n128: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t236_n130;
  t236_n129: return 0.08737153559923172f;
  t236_n130: return 0.03689263388514519f;
  t236_n131: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t236_n133;
  t236_n132: return 0.11550804227590561f;
  t236_n133: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t236_n135;
  t236_n134: return 0.12340857833623886f;
  t236_n135: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t236_n137;
  t236_n136: return -0.02261563576757908f;
  t236_n137: return 0.19088388979434967f;
  t236_n138: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t236_n204;
  t236_n139: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t236_n203;
  t236_n140: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t236_n150;
  t236_n141: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t236_n147;
  t236_n142: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t236_n144;
  t236_n143: return -0.039006706327199936f;
  t236_n144: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t236_n146;
  t236_n145: return -0.11459991335868835f;
  t236_n146: return 0.11152252554893494f;
  t236_n147: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t236_n149;
  t236_n148: return 0.10074569284915924f;
  t236_n149: return 0.07481749355792999f;
  t236_n150: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t236_n190;
  t236_n151: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t236_n157;
  t236_n152: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t236_n156;
  t236_n153: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t236_n155;
  t236_n154: return 0.02945656329393387f;
  t236_n155: return -0.027042482048273087f;
  t236_n156: return -0.25551971793174744f;
  t236_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t236_n171;
  t236_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n168;
  t236_n159: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t236_n167;
  t236_n160: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t236_n164;
  t236_n161: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t236_n163;
  t236_n162: return 0.0026221247389912605f;
  t236_n163: return -0.25282371044158936f;
  t236_n164: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t236_n166;
  t236_n165: return 0.1008729338645935f;
  t236_n166: return -0.00857471115887165f;
  t236_n167: return -0.20899836719036102f;
  t236_n168: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t236_n170;
  t236_n169: return 0.07995717227458954f;
  t236_n170: return -0.006394598633050919f;
  t236_n171: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t236_n173;
  t236_n172: return -0.007363622076809406f;
  t236_n173: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t236_n185;
  t236_n174: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t236_n178;
  t236_n175: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t236_n177;
  t236_n176: return 0.0788348987698555f;
  t236_n177: return 0.10978144407272339f;
  t236_n178: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t236_n180;
  t236_n179: return 0.13875806331634521f;
  t236_n180: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t236_n182;
  t236_n181: return -0.09309496730566025f;
  t236_n182: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t236_n184;
  t236_n183: return 0.1063784807920456f;
  t236_n184: return 0.12532754242420197f;
  t236_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t236_n189;
  t236_n186: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t236_n188;
  t236_n187: return 0.09721844643354416f;
  t236_n188: return -0.08377528190612793f;
  t236_n189: return 0.12111812084913254f;
  t236_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n194;
  t236_n191: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t236_n193;
  t236_n192: return -0.14127175509929657f;
  t236_n193: return 0.13845017552375793f;
  t236_n194: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t236_n196;
  t236_n195: return 0.007601628545671701f;
  t236_n196: if (E.getFractionNameInContext() >= 3200224622 /*0.3741258978843689*/) goto t236_n198;
  t236_n197: return 0.11672241240739822f;
  t236_n198: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t236_n200;
  t236_n199: return 0.02868110127747059f;
  t236_n200: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t236_n202;
  t236_n201: return 0.16785357892513275f;
  t236_n202: return -0.05088741332292557f;
  t236_n203: return 0.0070302607491612434f;
  t236_n204: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t236_n206;
  t236_n205: return 0.004409669898450375f;
  t236_n206: if (E.getFractionNameInContext() >= 3208152317 /*0.7207792401313782*/) goto t236_n234;
  t236_n207: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t236_n213;
  t236_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t236_n210;
  t236_n209: return -0.1511511206626892f;
  t236_n210: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t236_n212;
  t236_n211: return -0.0715346485376358f;
  t236_n212: return 0.15623797476291656f;
  t236_n213: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t236_n231;
  t236_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t236_n220;
  t236_n215: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t236_n217;
  t236_n216: return -0.4210331439971924f;
  t236_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t236_n219;
  t236_n218: return -0.05182953551411629f;
  t236_n219: return 0.10115519165992737f;
  t236_n220: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t236_n230;
  t236_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t236_n227;
  t236_n222: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t236_n226;
  t236_n223: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t236_n225;
  t236_n224: return -0.11366179585456848f;
  t236_n225: return 0.17921297252178192f;
  t236_n226: return 0.10403776913881302f;
  t236_n227: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t236_n229;
  t236_n228: return 0.09983980655670166f;
  t236_n229: return 0.059043679386377335f;
  t236_n230: return -0.14466899633407593f;
  t236_n231: if (E.getFractionNameInContext() >= 3206464748 /*0.620192289352417*/) goto t236_n233;
  t236_n232: return 0.09957122057676315f;
  t236_n233: return -0.14829163253307343f;
  t236_n234: return -0.03200972452759743f;
  t236_n235: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t236_n243;
  t236_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel))) goto t236_n242;
  t236_n237: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t236_n241;
  t236_n238: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t236_n240;
  t236_n239: return 0.007383853662759066f;
  t236_n240: return -0.05588741600513458f;
  t236_n241: return 0.0018617598107084632f;
  t236_n242: return 0.008611420169472694f;
  t236_n243: return 0.00041807268280535936f;
  t236_n244: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t236_n248;
  t236_n245: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t236_n247;
  t236_n246: return -0.0900302529335022f;
  t236_n247: return 0.046849533915519714f;
  t236_n248: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t236_n252;
  t236_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t236_n251;
  t236_n250: return 0.011825683526694775f;
  t236_n251: return 0.023974018171429634f;
  t236_n252: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t236_n254;
  t236_n253: return -0.07420989871025085f;
  t236_n254: return -0.8480251431465149f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree237(const Example& E) {
  t237_n0: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t237_n152;
  t237_n1: if (E.getNumReferences() >= 3350789120 /*94712.0*/) goto t237_n149;
  t237_n2: if (E.getNumReferences() >= 3350665472 /*93746.0*/) goto t237_n138;
  t237_n3: if (E.getNumReferences() >= 3350538496 /*92754.0*/) goto t237_n133;
  t237_n4: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t237_n66;
  t237_n5: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t237_n37;
  t237_n6: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t237_n28;
  t237_n7: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t237_n17;
  t237_n8: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t237_n16;
  t237_n9: if (E.getNumReferences() >= 3341597824 /*44220.5*/) goto t237_n11;
  t237_n10: return -5.74454061279539e-05f;
  t237_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n13;
  t237_n12: return -0.04853108152747154f;
  t237_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t237_n15;
  t237_n14: return 0.11359751969575882f;
  t237_n15: return 0.10169947147369385f;
  t237_n16: return -0.16471844911575317f;
  t237_n17: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t237_n23;
  t237_n18: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n20;
  t237_n19: return 0.0386219397187233f;
  t237_n20: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t237_n22;
  t237_n21: return -0.09323149919509888f;
  t237_n22: return 0.09911230951547623f;
  t237_n23: if (E.getNumReferences() >= 3341955328 /*45617.0*/) goto t237_n27;
  t237_n24: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t237_n26;
  t237_n25: return -0.1551995724439621f;
  t237_n26: return 0.07524415850639343f;
  t237_n27: return 0.022834520787000656f;
  t237_n28: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t237_n30;
  t237_n29: return -0.08767427504062653f;
  t237_n30: if (E.getNumReferences() >= 3343486720 /*51599.0*/) goto t237_n36;
  t237_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t237_n35;
  t237_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n34;
  t237_n33: return -0.21063075959682465f;
  t237_n34: return 0.08331090211868286f;
  t237_n35: return 0.13915467262268066f;
  t237_n36: return -0.022670641541481018f;
  t237_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t237_n55;
  t237_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n42;
  t237_n39: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t237_n41;
  t237_n40: return -0.05847952514886856f;
  t237_n41: return -0.3605172634124756f;
  t237_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n54;
  t237_n43: if (E.getNumReferences() >= 3344543232 /*55726.0*/) goto t237_n49;
  t237_n44: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t237_n48;
  t237_n45: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t237_n47;
  t237_n46: return 0.13518837094306946f;
  t237_n47: return 0.08956685662269592f;
  t237_n48: return -0.2954293191432953f;
  t237_n49: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t237_n51;
  t237_n50: return 0.08021238446235657f;
  t237_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n53;
  t237_n52: return -0.0051816049963235855f;
  t237_n53: return 0.08890662342309952f;
  t237_n54: return -0.12695707380771637f;
  t237_n55: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t237_n59;
  t237_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n58;
  t237_n57: return -0.12181192636489868f;
  t237_n58: return 0.09426884353160858f;
  t237_n59: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t237_n63;
  t237_n60: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t237_n62;
  t237_n61: return 0.058219972997903824f;
  t237_n62: return 0.12857066094875336f;
  t237_n63: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t237_n65;
  t237_n64: return 0.13839086890220642f;
  t237_n65: return 0.1978609263896942f;
  t237_n66: if (E.getNumReferences() >= 3344974336 /*57410.0*/) goto t237_n68;
  t237_n67: return -0.14845460653305054f;
  t237_n68: if (E.getNumReferences() >= 3345503360 /*59476.5*/) goto t237_n80;
  t237_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t237_n71;
  t237_n70: return 0.04011772200465202f;
  t237_n71: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t237_n77;
  t237_n72: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n74;
  t237_n73: return -0.03669889643788338f;
  t237_n74: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t237_n76;
  t237_n75: return 0.11594967544078827f;
  t237_n76: return 0.06225619092583656f;
  t237_n77: if (E.getNumReferences() >= 3345041664 /*57673.0*/) goto t237_n79;
  t237_n78: return -0.49872633814811707f;
  t237_n79: return 0.05895916000008583f;
  t237_n80: if (E.getNumReferences() >= 3345933184 /*61155.5*/) goto t237_n88;
  t237_n81: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n87;
  t237_n82: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t237_n84;
  t237_n83: return -0.15129606425762177f;
  t237_n84: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t237_n86;
  t237_n85: return -0.24157564342021942f;
  t237_n86: return 0.10410212725400925f;
  t237_n87: return 0.03332686051726341f;
  t237_n88: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t237_n120;
  t237_n89: if (E.getNumReferences() >= 3348120000 /*73859.5*/) goto t237_n113;
  t237_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n96;
  t237_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n95;
  t237_n92: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t237_n94;
  t237_n93: return -0.11908190697431564f;
  t237_n94: return -0.2930571734905243f;
  t237_n95: return -0.0662144273519516f;
  t237_n96: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t237_n100;
  t237_n97: if (E.getNumReferences() >= 3346034176 /*61550.0*/) goto t237_n99;
  t237_n98: return -0.06658333539962769f;
  t237_n99: return 0.00577909592539072f;
  t237_n100: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t237_n110;
  t237_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n103;
  t237_n102: return -0.1212863102555275f;
  t237_n103: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t237_n105;
  t237_n104: return 0.04734392464160919f;
  t237_n105: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n109;
  t237_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t237_n108;
  t237_n107: return 0.11233338713645935f;
  t237_n108: return 0.13794709742069244f;
  t237_n109: return -0.04904096946120262f;
  t237_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t237_n112;
  t237_n111: return -0.13405266404151917f;
  t237_n112: return 0.020942743867635727f;
  t237_n113: if (E.getNumReferences() >= 3348326400 /*75472.0*/) goto t237_n119;
  t237_n114: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t237_n116;
  t237_n115: return 0.017953040078282356f;
  t237_n116: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t237_n118;
  t237_n117: return 0.016405556350946426f;
  t237_n118: return 0.09933385998010635f;
  t237_n119: return -0.0033640838228166103f;
  t237_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t237_n132;
  t237_n121: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t237_n129;
  t237_n122: if (E.getNumReferences() >= 3350198912 /*90101.0*/) goto t237_n124;
  t237_n123: return 0.006534229498356581f;
  t237_n124: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n128;
  t237_n125: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t237_n127;
  t237_n126: return -0.33129560947418213f;
  t237_n127: return -0.1410001516342163f;
  t237_n128: return 0.09441298246383667f;
  t237_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t237_n131;
  t237_n130: return 0.08291058987379074f;
  t237_n131: return 0.14261756837368011f;
  t237_n132: return 0.040190499275922775f;
  t237_n133: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t237_n137;
  t237_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n136;
  t237_n135: return -0.13305401802062988f;
  t237_n136: return -0.17989878356456757f;
  t237_n137: return -0.30528727173805237f;
  t237_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n140;
  t237_n139: return -0.08987413346767426f;
  t237_n140: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t237_n144;
  t237_n141: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n143;
  t237_n142: return 0.08679547905921936f;
  t237_n143: return 0.16720332205295563f;
  t237_n144: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t237_n146;
  t237_n145: return 0.07820145040750504f;
  t237_n146: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t237_n148;
  t237_n147: return -0.31004083156585693f;
  t237_n148: return 0.009925659745931625f;
  t237_n149: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t237_n151;
  t237_n150: return -0.13809765875339508f;
  t237_n151: return -0.2408752143383026f;
  t237_n152: if (E.getNumReferences() >= 3351669952 /*101593.5*/) goto t237_n182;
  t237_n153: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t237_n159;
  t237_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n156;
  t237_n155: return -0.1638631373643875f;
  t237_n156: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t237_n158;
  t237_n157: return -0.047413263469934464f;
  t237_n158: return 0.08729731291532516f;
  t237_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n179;
  t237_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t237_n170;
  t237_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n163;
  t237_n162: return -0.12474633008241653f;
  t237_n163: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t237_n167;
  t237_n164: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t237_n166;
  t237_n165: return 0.03278364986181259f;
  t237_n166: return -0.09634023159742355f;
  t237_n167: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t237_n169;
  t237_n168: return -0.10270216315984726f;
  t237_n169: return 0.10527048259973526f;
  t237_n170: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t237_n172;
  t237_n171: return 0.02624540589749813f;
  t237_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t237_n174;
  t237_n173: return 0.06972058117389679f;
  t237_n174: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t237_n176;
  t237_n175: return 0.10295052081346512f;
  t237_n176: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t237_n178;
  t237_n177: return -0.0028369338251650333f;
  t237_n178: return 0.09122284501791f;
  t237_n179: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t237_n181;
  t237_n180: return -0.03028896078467369f;
  t237_n181: return -0.2892310321331024f;
  t237_n182: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t237_n230;
  t237_n183: if (E.getNumReferences() >= 3353778624 /*118067.5*/) goto t237_n187;
  t237_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n186;
  t237_n185: return -0.06217259168624878f;
  t237_n186: return -0.009968781843781471f;
  t237_n187: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t237_n229;
  t237_n188: if (E.getNumReferences() >= 3360067360 /*203324.5*/) goto t237_n224;
  t237_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t237_n207;
  t237_n190: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t237_n202;
  t237_n191: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n193;
  t237_n192: return -0.03732043504714966f;
  t237_n193: if (E.getNumReferences() >= 3357335872 /*160645.0*/) goto t237_n201;
  t237_n194: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t237_n200;
  t237_n195: if (E.getNumReferences() >= 3356822208 /*152619.0*/) goto t237_n197;
  t237_n196: return 0.011548863723874092f;
  t237_n197: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t237_n199;
  t237_n198: return 0.15251678228378296f;
  t237_n199: return 0.05594499781727791f;
  t237_n200: return 0.06601563841104507f;
  t237_n201: return -0.164509117603302f;
  t237_n202: if (E.getNumReferences() >= 3359621216 /*196353.5*/) goto t237_n206;
  t237_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n205;
  t237_n204: return -0.1531537026166916f;
  t237_n205: return 0.05171973630785942f;
  t237_n206: return -0.20246101915836334f;
  t237_n207: if (E.getNumReferences() >= 3355522464 /*132310.5*/) goto t237_n223;
  t237_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t237_n218;
  t237_n209: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t237_n211;
  t237_n210: return -0.12284400314092636f;
  t237_n211: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t237_n213;
  t237_n212: return 0.035342007875442505f;
  t237_n213: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n217;
  t237_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n216;
  t237_n215: return -0.10467062145471573f;
  t237_n216: return 0.11937430500984192f;
  t237_n217: return 0.059650495648384094f;
  t237_n218: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t237_n222;
  t237_n219: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t237_n221;
  t237_n220: return 0.0947800800204277f;
  t237_n221: return -0.17260314524173737f;
  t237_n222: return -0.0928109660744667f;
  t237_n223: return -0.06425553560256958f;
  t237_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n226;
  t237_n225: return -0.10023853182792664f;
  t237_n226: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t237_n228;
  t237_n227: return -0.0956331416964531f;
  t237_n228: return 0.02594994567334652f;
  t237_n229: return -0.010232306085526943f;
  t237_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t237_n248;
  t237_n231: if (E.getNumReferences() >= 3353301504 /*114340.0*/) goto t237_n241;
  t237_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n234;
  t237_n233: return -0.20436842739582062f;
  t237_n234: if (E.getNumReferences() >= 3352895680 /*111169.5*/) goto t237_n238;
  t237_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t237_n237;
  t237_n236: return 0.02336004376411438f;
  t237_n237: return 0.09470541030168533f;
  t237_n238: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t237_n240;
  t237_n239: return 0.02740597352385521f;
  t237_n240: return 0.10136868059635162f;
  t237_n241: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t237_n247;
  t237_n242: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t237_n246;
  t237_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t237_n245;
  t237_n244: return -0.010971518233418465f;
  t237_n245: return 0.06800809502601624f;
  t237_n246: return -0.12664592266082764f;
  t237_n247: return 0.0139124421402812f;
  t237_n248: if (E.getNumReferences() >= 3356396128 /*145961.5*/) goto t237_n252;
  t237_n249: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t237_n251;
  t237_n250: return -0.10929598659276962f;
  t237_n251: return -0.7297602891921997f;
  t237_n252: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t237_n254;
  t237_n253: return -0.036971963942050934f;
  t237_n254: return 0.016984935849905014f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree238(const Example& E) {
  t238_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t238_n106;
  t238_n1: if (E.getNumReferences() >= 3303133184 /*1806.5*/) goto t238_n7;
  t238_n2: if (E.getNumReferences() >= 3302985728 /*1788.5*/) goto t238_n6;
  t238_n3: if (E.getNumReferences() >= 3302977536 /*1787.5*/) goto t238_n5;
  t238_n4: return -0.0002448581508360803f;
  t238_n5: return 0.24600163102149963f;
  t238_n6: return -0.10766327381134033f;
  t238_n7: if (E.getNumReferences() >= 3303960576 /*1907.5*/) goto t238_n51;
  t238_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t238_n24;
  t238_n9: if (E.getNumReferences() >= 3303182336 /*1812.5*/) goto t238_n17;
  t238_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t238_n16;
  t238_n11: if (E.getNumReferences() >= 3303165952 /*1810.5*/) goto t238_n15;
  t238_n12: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t238_n14;
  t238_n13: return -0.027034761384129524f;
  t238_n14: return 0.2807612717151642f;
  t238_n15: return 0.1957664042711258f;
  t238_n16: return -0.18767884373664856f;
  t238_n17: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t238_n19;
  t238_n18: return -0.03651171922683716f;
  t238_n19: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t238_n21;
  t238_n20: return 0.03229799121618271f;
  t238_n21: if (E.getNumReferences() >= 3303702528 /*1876.0*/) goto t238_n23;
  t238_n22: return -0.11608214676380157f;
  t238_n23: return 0.148194819688797f;
  t238_n24: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t238_n42;
  t238_n25: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t238_n41;
  t238_n26: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t238_n28;
  t238_n27: return -0.050160206854343414f;
  t238_n28: if (E.getNumReferences() >= 3303288832 /*1825.5*/) goto t238_n32;
  t238_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t238_n31;
  t238_n30: return -0.028962397947907448f;
  t238_n31: return 0.18527939915657043f;
  t238_n32: if (E.getNumReferences() >= 3303305216 /*1827.5*/) goto t238_n34;
  t238_n33: return 0.41312453150749207f;
  t238_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t238_n36;
  t238_n35: return 0.0651034340262413f;
  t238_n36: if (E.getNumReferences() >= 3303448576 /*1845.0*/) goto t238_n40;
  t238_n37: if (E.getNumReferences() >= 3303419904 /*1841.5*/) goto t238_n39;
  t238_n38: return -0.07954077422618866f;
  t238_n39: return 0.13164298236370087f;
  t238_n40: return 0.04678219556808472f;
  t238_n41: return -0.049074072390794754f;
  t238_n42: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t238_n50;
  t238_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t238_n45;
  t238_n44: return 0.07241295278072357f;
  t238_n45: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t238_n49;
  t238_n46: if (E.getNumReferences() >= 3303301120 /*1827.0*/) goto t238_n48;
  t238_n47: return 0.1234966292977333f;
  t238_n48: return 0.07296479493379593f;
  t238_n49: return 0.13828814029693604f;
  t238_n50: return 0.20292730629444122f;
  t238_n51: if (E.getNumReferences() >= 3304742912 /*2003.0*/) goto t238_n61;
  t238_n52: if (E.getNumReferences() >= 3304624128 /*1988.5*/) goto t238_n56;
  t238_n53: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t238_n55;
  t238_n54: return -0.013322889804840088f;
  t238_n55: return 0.13762183487415314f;
  t238_n56: if (E.getNumReferences() >= 3304648704 /*1991.5*/) goto t238_n60;
  t238_n57: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t238_n59;
  t238_n58: return -0.23676824569702148f;
  t238_n59: return -1.638249397277832f;
  t238_n60: return -0.04058655723929405f;
  t238_n61: if (E.getNumReferences() >= 3305103360 /*2047.0*/) goto t238_n91;
  t238_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t238_n90;
  t238_n63: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t238_n65;
  t238_n64: return 0.1535630226135254f;
  t238_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t238_n67;
  t238_n66: return -0.08519325405359268f;
  t238_n67: if (E.getNumReferences() >= 3305033728 /*2038.5*/) goto t238_n77;
  t238_n68: if (E.getNumReferences() >= 3304804352 /*2010.5*/) goto t238_n76;
  t238_n69: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t238_n71;
  t238_n70: return 0.11423493921756744f;
  t238_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t238_n73;
  t238_n72: return -0.0549236498773098f;
  t238_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t238_n75;
  t238_n74: return 0.11636216938495636f;
  t238_n75: return 0.11770686507225037f;
  t238_n76: return 0.028027834370732307f;
  t238_n77: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t238_n79;
  t238_n78: return 0.16561497747898102f;
  t238_n79: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t238_n83;
  t238_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t238_n82;
  t238_n81: return -0.17059578001499176f;
  t238_n82: return 0.1490248292684555f;
  t238_n83: if (E.getFractionNameInContext() >= 3182766931 /*0.08846154063940048*/) goto t238_n89;
  t238_n84: if (E.getNumReferences() >= 3305078784 /*2044.0*/) goto t238_n88;
  t238_n85: if (E.getNumReferences() >= 3305050112 /*2040.5*/) goto t238_n87;
  t238_n86: return 0.12015338987112045f;
  t238_n87: return -0.026459552347660065f;
  t238_n88: return 0.2149071842432022f;
  t238_n89: return 0.09706849604845047f;
  t238_n90: return 0.12135050445795059f;
  t238_n91: if (E.getNumReferences() >= 3305279488 /*2089.0*/) goto t238_n97;
  t238_n92: if (E.getNumReferences() >= 3305220096 /*2074.5*/) goto t238_n96;
  t238_n93: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t238_n95;
  t238_n94: return -0.01585487090051174f;
  t238_n95: return 0.11731890588998795f;
  t238_n96: return -0.15761154890060425f;
  t238_n97: if (E.getNumReferences() >= 3305285632 /*2090.5*/) goto t238_n101;
  t238_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t238_n100;
  t238_n99: return 0.08947567641735077f;
  t238_n100: return 0.18162739276885986f;
  t238_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t238_n105;
  t238_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t238_n104;
  t238_n103: return -0.010926932096481323f;
  t238_n104: return 0.017366576939821243f;
  t238_n105: return 0.0028941307682543993f;
  t238_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t238_n108;
  t238_n107: return 0.0009908560896292329f;
  t238_n108: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t238_n248;
  t238_n109: if (E.getNumReferences() >= 3283714048 /*371.0*/) goto t238_n241;
  t238_n110: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t238_n170;
  t238_n111: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t238_n165;
  t238_n112: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t238_n164;
  t238_n113: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t238_n121;
  t238_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t238_n120;
  t238_n115: if (E.getNumReferences() >= 3276668928 /*206.0*/) goto t238_n117;
  t238_n116: return 0.005516953766345978f;
  t238_n117: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t238_n119;
  t238_n118: return -0.08736316114664078f;
  t238_n119: return 0.06992845237255096f;
  t238_n120: return -0.1462298482656479f;
  t238_n121: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t238_n159;
  t238_n122: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t238_n140;
  t238_n123: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t238_n127;
  t238_n124: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t238_n126;
  t238_n125: return -0.0013195981737226248f;
  t238_n126: return -0.06260286271572113f;
  t238_n127: if (E.getNumReferences() >= 3271163904 /*125.0*/) goto t238_n137;
  t238_n128: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t238_n136;
  t238_n129: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t238_n135;
  t238_n130: if (E.getNumReferences() >= 3235905536 /*7.0*/) goto t238_n134;
  t238_n131: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t238_n133;
  t238_n132: return 0.04763262718915939f;
  t238_n133: return 0.003945768810808659f;
  t238_n134: return -0.10973667353391647f;
  t238_n135: return 0.1941017508506775f;
  t238_n136: return -0.18496392667293549f;
  t238_n137: if (E.getNumReferences() >= 3278241792 /*230.0*/) goto t238_n139;
  t238_n138: return 0.09849271923303604f;
  t238_n139: return -0.0432833731174469f;
  t238_n140: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t238_n150;
  t238_n141: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t238_n143;
  t238_n142: return -1.2680813074111938f;
  t238_n143: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t238_n145;
  t238_n144: return -0.015458645299077034f;
  t238_n145: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t238_n149;
  t238_n146: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t238_n148;
  t238_n147: return -0.07513908296823502f;
  t238_n148: return -0.32367900013923645f;
  t238_n149: return -0.12988914549350739f;
  t238_n150: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t238_n158;
  t238_n151: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t238_n157;
  t238_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t238_n154;
  t238_n153: return -0.004889274947345257f;
  t238_n154: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t238_n156;
  t238_n155: return 0.11809361726045609f;
  t238_n156: return -0.1257522851228714f;
  t238_n157: return 0.1823238879442215f;
  t238_n158: return -0.017330987378954887f;
  t238_n159: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t238_n161;
  t238_n160: return -0.15297408401966095f;
  t238_n161: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t238_n163;
  t238_n162: return 0.08839388936758041f;
  t238_n163: return -0.0437094047665596f;
  t238_n164: return 0.1538870632648468f;
  t238_n165: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t238_n169;
  t238_n166: if (E.getFileProximityDistanceCost() >= 3254779904 /*32.0*/) goto t238_n168;
  t238_n167: return -0.05353802815079689f;
  t238_n168: return -0.33499011397361755f;
  t238_n169: return -0.19047518074512482f;
  t238_n170: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t238_n224;
  t238_n171: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t238_n181;
  t238_n172: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t238_n178;
  t238_n173: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t238_n177;
  t238_n174: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t238_n176;
  t238_n175: return 0.017963862046599388f;
  t238_n176: return 0.08711259067058563f;
  t238_n177: return 0.05651529133319855f;
  t238_n178: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t238_n180;
  t238_n179: return 8.37091647554189e-05f;
  t238_n180: return -0.13901248574256897f;
  t238_n181: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t238_n203;
  t238_n182: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t238_n190;
  t238_n183: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t238_n185;
  t238_n184: return 0.0017822090303525329f;
  t238_n185: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t238_n189;
  t238_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t238_n188;
  t238_n187: return 0.026703940704464912f;
  t238_n188: return 0.07070329040288925f;
  t238_n189: return 0.012846979312598705f;
  t238_n190: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t238_n194;
  t238_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t238_n193;
  t238_n192: return -0.04236571118235588f;
  t238_n193: return -0.010669343173503876f;
  t238_n194: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t238_n196;
  t238_n195: return -0.1174779012799263f;
  t238_n196: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t238_n202;
  t238_n197: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t238_n201;
  t238_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t238_n200;
  t238_n199: return 0.014219882898032665f;
  t238_n200: return 0.07457973808050156f;
  t238_n201: return -0.045933619141578674f;
  t238_n202: return 0.06736931949853897f;
  t238_n203: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t238_n223;
  t238_n204: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t238_n208;
  t238_n205: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t238_n207;
  t238_n206: return 0.021082943305373192f;
  t238_n207: return 0.040681350976228714f;
  t238_n208: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t238_n210;
  t238_n209: return -0.03736458718776703f;
  t238_n210: if (E.getFractionNameInContext() >= 3208451910 /*0.7386363744735718*/) goto t238_n220;
  t238_n211: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t238_n219;
  t238_n212: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t238_n216;
  t238_n213: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t238_n215;
  t238_n214: return 0.01637667790055275f;
  t238_n215: return -0.06408609449863434f;
  t238_n216: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t238_n218;
  t238_n217: return 0.014249339699745178f;
  t238_n218: return 0.11116103827953339f;
  t238_n219: return 0.05180668085813522f;
  t238_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t238_n222;
  t238_n221: return -0.16787216067314148f;
  t238_n222: return -0.029774963855743408f;
  t238_n223: return -0.13554616272449493f;
  t238_n224: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t238_n236;
  t238_n225: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t238_n231;
  t238_n226: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t238_n230;
  t238_n227: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t238_n229;
  t238_n228: return -0.03869837895035744f;
  t238_n229: return -0.1700061857700348f;
  t238_n230: return -0.010282697156071663f;
  t238_n231: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t238_n233;
  t238_n232: return -0.006607314571738243f;
  t238_n233: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t238_n235;
  t238_n234: return 0.11051172763109207f;
  t238_n235: return -0.05544697493314743f;
  t238_n236: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t238_n240;
  t238_n237: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t238_n239;
  t238_n238: return -0.038253482431173325f;
  t238_n239: return -0.1995193064212799f;
  t238_n240: return -0.06074336916208267f;
  t238_n241: if (E.getNumReferences() >= 3284238336 /*387.0*/) goto t238_n247;
  t238_n242: if (E.getFractionNameInContext() >= 3187804193 /*0.1269841343164444*/) goto t238_n246;
  t238_n243: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t238_n245;
  t238_n244: return 0.10479813069105148f;
  t238_n245: return 0.057739175856113434f;
  t238_n246: return 0.12656040489673615f;
  t238_n247: return 0.019158847630023956f;
  t238_n248: if (E.getNumReferences() >= 3292028928 /*737.5*/) goto t238_n252;
  t238_n249: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t238_n251;
  t238_n250: return -0.0935356393456459f;
  t238_n251: return 0.008806176483631134f;
  t238_n252: if (E.getNumReferences() >= 3296755712 /*1028.0*/) goto t238_n254;
  t238_n253: return -1.4482755661010742f;
  t238_n254: return -0.07324014604091644f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree239(const Example& E) {
  t239_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t239_n254;
  t239_n1: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t239_n51;
  t239_n2: if (E.getNumReferences() >= 3280257024 /*265.5*/) goto t239_n6;
  t239_n3: if (E.getNumReferences() >= 3280060416 /*259.5*/) goto t239_n5;
  t239_n4: return 0.0001316145935561508f;
  t239_n5: return -0.06735623627901077f;
  t239_n6: if (E.getNumReferences() >= 3281207296 /*294.5*/) goto t239_n24;
  t239_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t239_n15;
  t239_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t239_n10;
  t239_n9: return 0.020878342911601067f;
  t239_n10: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t239_n12;
  t239_n11: return 0.04351961612701416f;
  t239_n12: if (E.getNumReferences() >= 3280683008 /*278.5*/) goto t239_n14;
  t239_n13: return 0.23695318400859833f;
  t239_n14: return 0.07887908071279526f;
  t239_n15: if (E.getNumReferences() >= 3280846848 /*283.5*/) goto t239_n17;
  t239_n16: return 0.012345715425908566f;
  t239_n17: if (E.getNumReferences() >= 3280912384 /*285.5*/) goto t239_n19;
  t239_n18: return 0.1279437094926834f;
  t239_n19: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t239_n23;
  t239_n20: if (E.getNumReferences() >= 3281174528 /*293.5*/) goto t239_n22;
  t239_n21: return 0.07475852221250534f;
  t239_n22: return 0.12520429491996765f;
  t239_n23: return -0.020818883553147316f;
  t239_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t239_n40;
  t239_n25: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n35;
  t239_n26: if (E.getNumReferences() >= 3322050048 /*8349.5*/) goto t239_n34;
  t239_n27: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t239_n33;
  t239_n28: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t239_n30;
  t239_n29: return -0.029260240495204926f;
  t239_n30: if (E.getNumReferences() >= 3299446784 /*1356.5*/) goto t239_n32;
  t239_n31: return 0.021892299875617027f;
  t239_n32: return 0.14059916138648987f;
  t239_n33: return -0.0630941241979599f;
  t239_n34: return -0.18090219795703888f;
  t239_n35: if (E.getNumReferences() >= 3281829888 /*313.5*/) goto t239_n37;
  t239_n36: return -0.18924926221370697f;
  t239_n37: if (E.getNumReferences() >= 3281911808 /*316.0*/) goto t239_n39;
  t239_n38: return 0.1226949691772461f;
  t239_n39: return 0.0046508233062922955f;
  t239_n40: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t239_n50;
  t239_n41: if (E.getNumReferences() >= 3281829888 /*313.5*/) goto t239_n43;
  t239_n42: return -0.14122946560382843f;
  t239_n43: if (E.getNumReferences() >= 3282616320 /*337.5*/) goto t239_n47;
  t239_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t239_n46;
  t239_n45: return 0.0171833299100399f;
  t239_n46: return 0.1491941213607788f;
  t239_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t239_n49;
  t239_n48: return 0.029605761170387268f;
  t239_n49: return 0.12003464996814728f;
  t239_n50: return 0.006528304424136877f;
  t239_n51: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t239_n253;
  t239_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t239_n212;
  t239_n53: if (E.getNumReferences() >= 3292397568 /*760.0*/) goto t239_n205;
  t239_n54: if (E.getNumReferences() >= 3271983104 /*134.5*/) goto t239_n56;
  t239_n55: return 0.0008546325261704624f;
  t239_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t239_n76;
  t239_n57: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t239_n67;
  t239_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t239_n66;
  t239_n59: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t239_n65;
  t239_n60: if (E.getNumReferences() >= 3278340096 /*231.5*/) goto t239_n62;
  t239_n61: return -0.22440937161445618f;
  t239_n62: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t239_n64;
  t239_n63: return 0.05998386815190315f;
  t239_n64: return -0.1530153900384903f;
  t239_n65: return -0.01845581643283367f;
  t239_n66: return 0.006387524772435427f;
  t239_n67: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t239_n69;
  t239_n68: return 0.03960053250193596f;
  t239_n69: if (E.getNumReferences() >= 3284992000 /*410.0*/) goto t239_n71;
  t239_n70: return 0.11340384185314178f;
  t239_n71: if (E.getNumReferences() >= 3287793664 /*495.5*/) goto t239_n73;
  t239_n72: return -0.04634993523359299f;
  t239_n73: if (E.getNumReferences() >= 3289800704 /*601.5*/) goto t239_n75;
  t239_n74: return 0.11382109671831131f;
  t239_n75: return 0.018909532576799393f;
  t239_n76: if (E.getNumReferences() >= 3274145792 /*167.5*/) goto t239_n124;
  t239_n77: if (E.getNumReferences() >= 3273949184 /*164.5*/) goto t239_n115;
  t239_n78: if (E.getNumReferences() >= 3273424896 /*156.5*/) goto t239_n112;
  t239_n79: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t239_n95;
  t239_n80: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n94;
  t239_n81: if (E.getNumReferences() >= 3272179712 /*137.5*/) goto t239_n89;
  t239_n82: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t239_n84;
  t239_n83: return 0.009889262728393078f;
  t239_n84: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t239_n86;
  t239_n85: return 0.1622864305973053f;
  t239_n86: if (E.getNumReferences() >= 3272048640 /*135.5*/) goto t239_n88;
  t239_n87: return 0.26971495151519775f;
  t239_n88: return 0.3707387447357178f;
  t239_n89: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t239_n93;
  t239_n90: if (E.getNumReferences() >= 3272409088 /*141.0*/) goto t239_n92;
  t239_n91: return 0.1869104504585266f;
  t239_n92: return -0.10433634370565414f;
  t239_n93: return -0.004427281208336353f;
  t239_n94: return -0.15038427710533142f;
  t239_n95: if (E.getNumReferences() >= 3272966144 /*149.5*/) goto t239_n103;
  t239_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t239_n98;
  t239_n97: return 0.04176461696624756f;
  t239_n98: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t239_n100;
  t239_n99: return -0.05174562707543373f;
  t239_n100: if (E.getNumReferences() >= 3272900608 /*148.5*/) goto t239_n102;
  t239_n101: return 0.21922701597213745f;
  t239_n102: return 0.5494900345802307f;
  t239_n103: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t239_n111;
  t239_n104: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n108;
  t239_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t239_n107;
  t239_n106: return -0.1281939595937729f;
  t239_n107: return 0.23496590554714203f;
  t239_n108: if (E.getNumReferences() >= 3273228288 /*153.5*/) goto t239_n110;
  t239_n109: return 0.20531359314918518f;
  t239_n110: return 0.17955467104911804f;
  t239_n111: return -0.05717616155743599f;
  t239_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t239_n114;
  t239_n113: return -0.07172369956970215f;
  t239_n114: return 0.1250341683626175f;
  t239_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t239_n123;
  t239_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t239_n118;
  t239_n117: return -0.1558343768119812f;
  t239_n118: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t239_n122;
  t239_n119: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t239_n121;
  t239_n120: return 0.8300885558128357f;
  t239_n121: return 0.12076683342456818f;
  t239_n122: return -0.1656661033630371f;
  t239_n123: return 0.19070492684841156f;
  t239_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t239_n146;
  t239_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword))) goto t239_n127;
  t239_n126: return -0.01540881022810936f;
  t239_n127: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t239_n145;
  t239_n128: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n136;
  t239_n129: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t239_n135;
  t239_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t239_n132;
  t239_n131: return -0.01638510823249817f;
  t239_n132: if (E.getNumReferences() >= 3284303872 /*389.0*/) goto t239_n134;
  t239_n133: return -0.10110723972320557f;
  t239_n134: return 0.12695173919200897f;
  t239_n135: return 0.17527762055397034f;
  t239_n136: if (E.getNumReferences() >= 3285123072 /*414.0*/) goto t239_n144;
  t239_n137: if (E.getNumReferences() >= 3280109568 /*261.0*/) goto t239_n139;
  t239_n138: return -0.04243585467338562f;
  t239_n139: if (E.getNumReferences() >= 3280404480 /*270.0*/) goto t239_n141;
  t239_n140: return 0.23664309084415436f;
  t239_n141: if (E.getNumReferences() >= 3281436672 /*301.5*/) goto t239_n143;
  t239_n142: return -0.11943895369768143f;
  t239_n143: return 0.19537369906902313f;
  t239_n144: return -0.019910655915737152f;
  t239_n145: return -0.11118286103010178f;
  t239_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t239_n148;
  t239_n147: return -0.04143703728914261f;
  t239_n148: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t239_n200;
  t239_n149: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t239_n191;
  t239_n150: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t239_n186;
  t239_n151: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t239_n157;
  t239_n152: if (E.getNumReferences() >= 3286892544 /*468.0*/) goto t239_n156;
  t239_n153: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t239_n155;
  t239_n154: return 0.11885319650173187f;
  t239_n155: return -0.016935652121901512f;
  t239_n156: return 0.14307871460914612f;
  t239_n157: if (E.getNumReferences() >= 3287056384 /*473.0*/) goto t239_n185;
  t239_n158: if (E.getNumReferences() >= 3286450176 /*454.5*/) goto t239_n184;
  t239_n159: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n175;
  t239_n160: if (E.getNumReferences() >= 3282223104 /*325.5*/) goto t239_n174;
  t239_n161: if (E.getNumReferences() >= 3281813504 /*313.0*/) goto t239_n173;
  t239_n162: if (E.getNumReferences() >= 3279585280 /*250.5*/) goto t239_n172;
  t239_n163: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t239_n165;
  t239_n164: return -0.03755827993154526f;
  t239_n165: if (E.getNumReferences() >= 3278405632 /*232.5*/) goto t239_n171;
  t239_n166: if (E.getNumReferences() >= 3276013568 /*196.0*/) goto t239_n170;
  t239_n167: if (E.getNumReferences() >= 3275685888 /*191.0*/) goto t239_n169;
  t239_n168: return 0.0008865282288752496f;
  t239_n169: return 0.17538519203662872f;
  t239_n170: return -0.07654865086078644f;
  t239_n171: return 0.21569013595581055f;
  t239_n172: return -0.13117235898971558f;
  t239_n173: return 0.13784198462963104f;
  t239_n174: return -0.16335473954677582f;
  t239_n175: if (E.getNumReferences() >= 3283271680 /*357.5*/) goto t239_n183;
  t239_n176: if (E.getNumReferences() >= 3281616896 /*307.0*/) goto t239_n180;
  t239_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t239_n179;
  t239_n178: return 0.038181569427251816f;
  t239_n179: return 0.458452045917511f;
  t239_n180: if (E.getNumReferences() >= 3282075648 /*321.0*/) goto t239_n182;
  t239_n181: return 0.5633231401443481f;
  t239_n182: return 0.8199554681777954f;
  t239_n183: return -0.03857472911477089f;
  t239_n184: return 0.12882889807224274f;
  t239_n185: return -0.058196645230054855f;
  t239_n186: if (E.getNumReferences() >= 3289726976 /*597.0*/) goto t239_n190;
  t239_n187: if (E.getNumReferences() >= 3288948736 /*549.5*/) goto t239_n189;
  t239_n188: return 0.012117505073547363f;
  t239_n189: return -0.5144639611244202f;
  t239_n190: return 0.10524336248636246f;
  t239_n191: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t239_n193;
  t239_n192: return 0.2284148782491684f;
  t239_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t239_n199;
  t239_n194: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t239_n198;
  t239_n195: if (E.getNumReferences() >= 3278274560 /*230.5*/) goto t239_n197;
  t239_n196: return -0.10365477204322815f;
  t239_n197: return 0.11799488216638565f;
  t239_n198: return -0.1125461757183075f;
  t239_n199: return 0.18934430181980133f;
  t239_n200: if (E.getNumReferences() >= 3290554368 /*647.5*/) goto t239_n202;
  t239_n201: return 0.03916255757212639f;
  t239_n202: if (E.getNumReferences() >= 3291070464 /*679.0*/) goto t239_n204;
  t239_n203: return 0.6294713020324707f;
  t239_n204: return -0.1099928468465805f;
  t239_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t239_n207;
  t239_n206: return -0.03142178803682327f;
  t239_n207: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t239_n209;
  t239_n208: return 0.0698467269539833f;
  t239_n209: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t239_n211;
  t239_n210: return 0.034295447170734406f;
  t239_n211: return -0.10495191067457199f;
  t239_n212: if (E.getNumReferences() >= 3364595984 /*286024.5*/) goto t239_n252;
  t239_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t239_n219;
  t239_n214: if (E.getNumReferences() >= 3315998720 /*5316.0*/) goto t239_n218;
  t239_n215: if (E.getNumReferences() >= 3313971200 /*4326.0*/) goto t239_n217;
  t239_n216: return 0.00678632128983736f;
  t239_n217: return 0.28388121724128723f;
  t239_n218: return -0.11671200394630432f;
  t239_n219: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t239_n237;
  t239_n220: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t239_n222;
  t239_n221: return 0.043157149106264114f;
  t239_n222: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t239_n224;
  t239_n223: return 0.6897934079170227f;
  t239_n224: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n226;
  t239_n225: return 0.02333320491015911f;
  t239_n226: if (E.getNumReferences() >= 3296591872 /*1016.0*/) goto t239_n236;
  t239_n227: if (E.getNumReferences() >= 3269787648 /*114.5*/) goto t239_n229;
  t239_n228: return -0.08539354056119919f;
  t239_n229: if (E.getNumReferences() >= 3276767232 /*207.5*/) goto t239_n231;
  t239_n230: return 0.35796937346458435f;
  t239_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t239_n233;
  t239_n232: return -0.12653301656246185f;
  t239_n233: if (E.getNumReferences() >= 3281780736 /*312.0*/) goto t239_n235;
  t239_n234: return 0.12187745422124863f;
  t239_n235: return 0.13800999522209167f;
  t239_n236: return -0.05183624476194382f;
  t239_n237: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t239_n251;
  t239_n238: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t239_n242;
  t239_n239: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t239_n241;
  t239_n240: return 0.01201715599745512f;
  t239_n241: return -0.07275994122028351f;
  t239_n242: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t239_n246;
  t239_n243: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t239_n245;
  t239_n244: return -0.5033336877822876f;
  t239_n245: return 0.061453599482774734f;
  t239_n246: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t239_n250;
  t239_n247: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t239_n249;
  t239_n248: return 0.32624098658561707f;
  t239_n249: return -0.1492249071598053f;
  t239_n250: return 0.057514045387506485f;
  t239_n251: return -0.046392492949962616f;
  t239_n252: return 0.11697618663311005f;
  t239_n253: return -0.0018530839588493109f;
  t239_n254: return 0.0007593914051540196f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree240(const Example& E) {
  t240_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t240_n252;
  t240_n1: if (E.getNumReferences() >= 3303133184 /*1806.5*/) goto t240_n7;
  t240_n2: if (E.getNumReferences() >= 3302985728 /*1788.5*/) goto t240_n6;
  t240_n3: if (E.getNumReferences() >= 3302977536 /*1787.5*/) goto t240_n5;
  t240_n4: return -0.00024759332882240415f;
  t240_n5: return 0.24336527287960052f;
  t240_n6: return -0.10332146286964417f;
  t240_n7: if (E.getNumReferences() >= 3304108032 /*1925.5*/) goto t240_n63;
  t240_n8: if (E.getNumReferences() >= 3304099840 /*1924.5*/) goto t240_n62;
  t240_n9: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t240_n27;
  t240_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t240_n12;
  t240_n11: return 0.0015668092528358102f;
  t240_n12: if (E.getNumReferences() >= 3303202816 /*1815.0*/) goto t240_n14;
  t240_n13: return -0.07885904610157013f;
  t240_n14: if (E.getNumReferences() >= 3303440384 /*1844.0*/) goto t240_n18;
  t240_n15: if (E.getNumReferences() >= 3303419904 /*1841.5*/) goto t240_n17;
  t240_n16: return 0.0689583346247673f;
  t240_n17: return 0.12711429595947266f;
  t240_n18: if (E.getNumReferences() >= 3303559168 /*1858.5*/) goto t240_n20;
  t240_n19: return -0.10285172611474991f;
  t240_n20: if (E.getNumReferences() >= 3303796736 /*1887.5*/) goto t240_n26;
  t240_n21: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t240_n25;
  t240_n22: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t240_n24;
  t240_n23: return 0.022282762452960014f;
  t240_n24: return 0.15101467072963715f;
  t240_n25: return 0.0657612755894661f;
  t240_n26: return 0.054717767983675f;
  t240_n27: if (E.getNumReferences() >= 3303215104 /*1816.5*/) goto t240_n37;
  t240_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t240_n36;
  t240_n29: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t240_n35;
  t240_n30: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t240_n32;
  t240_n31: return 0.17156994342803955f;
  t240_n32: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t240_n34;
  t240_n33: return 0.10338478535413742f;
  t240_n34: return 0.26586493849754333f;
  t240_n35: return -0.05415476858615875f;
  t240_n36: return 0.232269287109375f;
  t240_n37: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t240_n39;
  t240_n38: return -0.04907439276576042f;
  t240_n39: if (E.getNumReferences() >= 3303247872 /*1820.5*/) goto t240_n41;
  t240_n40: return 0.1548745185136795f;
  t240_n41: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t240_n47;
  t240_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t240_n44;
  t240_n43: return 0.0038138222880661488f;
  t240_n44: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t240_n46;
  t240_n45: return 0.08012789487838745f;
  t240_n46: return 0.12902109324932098f;
  t240_n47: if (E.getNumReferences() >= 3303981056 /*1910.0*/) goto t240_n61;
  t240_n48: if (E.getNumReferences() >= 3303858176 /*1895.0*/) goto t240_n54;
  t240_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t240_n51;
  t240_n50: return -0.10037536174058914f;
  t240_n51: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t240_n53;
  t240_n52: return 0.16936609148979187f;
  t240_n53: return 0.07840003073215485f;
  t240_n54: if (E.getNumReferences() >= 3303936000 /*1904.5*/) goto t240_n58;
  t240_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t240_n57;
  t240_n56: return 0.09301072359085083f;
  t240_n57: return 0.14528922736644745f;
  t240_n58: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t240_n60;
  t240_n59: return 0.1467553973197937f;
  t240_n60: return -0.06050217151641846f;
  t240_n61: return -0.18493938446044922f;
  t240_n62: return 0.2957175374031067f;
  t240_n63: if (E.getNumReferences() >= 3304742912 /*2003.0*/) goto t240_n73;
  t240_n64: if (E.getNumReferences() >= 3304624128 /*1988.5*/) goto t240_n70;
  t240_n65: if (E.getNumReferences() >= 3304128512 /*1928.0*/) goto t240_n69;
  t240_n66: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t240_n68;
  t240_n67: return -0.5593701004981995f;
  t240_n68: return -0.08143611252307892f;
  t240_n69: return -0.008903113193809986f;
  t240_n70: if (E.getNumReferences() >= 3304648704 /*1991.5*/) goto t240_n72;
  t240_n71: return -0.21888850629329681f;
  t240_n72: return -0.036430563777685165f;
  t240_n73: if (E.getNumReferences() >= 3304804352 /*2010.5*/) goto t240_n89;
  t240_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t240_n86;
  t240_n75: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t240_n81;
  t240_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t240_n80;
  t240_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t240_n79;
  t240_n78: return -0.20504893362522125f;
  t240_n79: return 0.1673722267150879f;
  t240_n80: return 0.11069227755069733f;
  t240_n81: if (E.getNumReferences() >= 3304792064 /*2009.0*/) goto t240_n83;
  t240_n82: return -0.14900417625904083f;
  t240_n83: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t240_n85;
  t240_n84: return 0.019558532163500786f;
  t240_n85: return 0.13149796426296234f;
  t240_n86: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t240_n88;
  t240_n87: return 0.08585615456104279f;
  t240_n88: return 0.11453046649694443f;
  t240_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t240_n251;
  t240_n90: if (E.getNumReferences() >= 3304923136 /*2025.0*/) goto t240_n94;
  t240_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t240_n93;
  t240_n92: return -0.08988811820745468f;
  t240_n93: return 0.11940301209688187f;
  t240_n94: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t240_n98;
  t240_n95: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t240_n97;
  t240_n96: return -0.04187021404504776f;
  t240_n97: return 0.09230151772499084f;
  t240_n98: if (E.getNumReferences() >= 3372619744 /*549246.0*/) goto t240_n250;
  t240_n99: if (E.getNumReferences() >= 3369637072 /*443558.5*/) goto t240_n249;
  t240_n100: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t240_n102;
  t240_n101: return -0.07150048762559891f;
  t240_n102: if (E.getNumReferences() >= 3336160512 /*27874.5*/) goto t240_n182;
  t240_n103: if (E.getNumReferences() >= 3332305408 /*20345.0*/) goto t240_n181;
  t240_n104: if (E.getNumReferences() >= 3331498752 /*18769.5*/) goto t240_n168;
  t240_n105: if (E.getNumReferences() >= 3330532096 /*16881.5*/) goto t240_n167;
  t240_n106: if (E.getNumReferences() >= 3327053312 /*13235.5*/) goto t240_n148;
  t240_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t240_n117;
  t240_n108: if (E.getNumReferences() >= 3315083264 /*4869.0*/) goto t240_n116;
  t240_n109: if (E.getNumReferences() >= 3308431360 /*2858.5*/) goto t240_n111;
  t240_n110: return -0.08182121068239212f;
  t240_n111: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t240_n113;
  t240_n112: return -0.19063422083854675f;
  t240_n113: if (E.getNumReferences() >= 3308701696 /*2924.5*/) goto t240_n115;
  t240_n114: return 0.1214165911078453f;
  t240_n115: return 0.06678233295679092f;
  t240_n116: return -0.0556897409260273f;
  t240_n117: if (E.getNumReferences() >= 3323762688 /*10022.0*/) goto t240_n133;
  t240_n118: if (E.getNumReferences() >= 3320203264 /*7369.0*/) goto t240_n128;
  t240_n119: if (E.getNumReferences() >= 3319603200 /*7076.0*/) goto t240_n121;
  t240_n120: return -0.008976337499916553f;
  t240_n121: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t240_n123;
  t240_n122: return 0.04699351638555527f;
  t240_n123: if (E.getNumReferences() >= 3319744512 /*7145.0*/) goto t240_n127;
  t240_n124: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t240_n126;
  t240_n125: return 0.2445160299539566f;
  t240_n126: return 0.13038960099220276f;
  t240_n127: return 0.11420419067144394f;
  t240_n128: if (E.getNumReferences() >= 3322249216 /*8544.0*/) goto t240_n130;
  t240_n129: return -0.13374179601669312f;
  t240_n130: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t240_n132;
  t240_n131: return -0.0057174889370799065f;
  t240_n132: return 0.16157688200473785f;
  t240_n133: if (E.getNumReferences() >= 3323876864 /*10133.5*/) goto t240_n139;
  t240_n134: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t240_n136;
  t240_n135: return 0.09309312701225281f;
  t240_n136: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t240_n138;
  t240_n137: return 0.006722270976752043f;
  t240_n138: return 0.11743878573179245f;
  t240_n139: if (E.getFractionNameInContext() >= 3190786809 /*0.17142857611179352*/) goto t240_n147;
  t240_n140: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t240_n142;
  t240_n141: return 0.001343747484497726f;
  t240_n142: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t240_n146;
  t240_n143: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t240_n145;
  t240_n144: return 0.4273112416267395f;
  t240_n145: return -0.11652787029743195f;
  t240_n146: return -0.05647691711783409f;
  t240_n147: return 0.14281758666038513f;
  t240_n148: if (E.getNumReferences() >= 3327160832 /*13340.5*/) goto t240_n154;
  t240_n149: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t240_n151;
  t240_n150: return 0.08069583028554916f;
  t240_n151: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t240_n153;
  t240_n152: return 0.21151237189769745f;
  t240_n153: return 0.1684817224740982f;
  t240_n154: if (E.getNumReferences() >= 3327636992 /*13805.5*/) goto t240_n156;
  t240_n155: return -0.03692743927240372f;
  t240_n156: if (E.getNumReferences() >= 3327710208 /*13877.0*/) goto t240_n164;
  t240_n157: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t240_n159;
  t240_n158: return -0.0484975129365921f;
  t240_n159: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t240_n163;
  t240_n160: if (E.getNumReferences() >= 3327674368 /*13842.0*/) goto t240_n162;
  t240_n161: return 0.10197605192661285f;
  t240_n162: return 0.18620462715625763f;
  t240_n163: return 0.0318978875875473f;
  t240_n164: if (E.getNumReferences() >= 3327771648 /*13937.0*/) goto t240_n166;
  t240_n165: return -0.21631428599357605f;
  t240_n166: return 0.04631923884153366f;
  t240_n167: return -0.12402911484241486f;
  t240_n168: if (E.getNumReferences() >= 3331773952 /*19307.0*/) goto t240_n176;
  t240_n169: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t240_n175;
  t240_n170: if (E.getNumReferences() >= 3331557888 /*18885.0*/) goto t240_n174;
  t240_n171: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t240_n173;
  t240_n172: return -0.008456781506538391f;
  t240_n173: return 0.11092337965965271f;
  t240_n174: return -0.19428393244743347f;
  t240_n175: return 0.12511113286018372f;
  t240_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t240_n178;
  t240_n177: return 0.0006018199492245913f;
  t240_n178: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t240_n180;
  t240_n179: return 0.03924671933054924f;
  t240_n180: return 0.11849036067724228f;
  t240_n181: return -0.0535007119178772f;
  t240_n182: if (E.getNumReferences() >= 3336764160 /*29053.5*/) goto t240_n192;
  t240_n183: if (E.getNumReferences() >= 3336654336 /*28839.0*/) goto t240_n187;
  t240_n184: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t240_n186;
  t240_n185: return 0.07852382212877274f;
  t240_n186: return -0.37220579385757446f;
  t240_n187: if (E.getNumReferences() >= 3336705792 /*28939.5*/) goto t240_n191;
  t240_n188: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t240_n190;
  t240_n189: return 0.15050649642944336f;
  t240_n190: return 0.14543098211288452f;
  t240_n191: return 0.07924337685108185f;
  t240_n192: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t240_n194;
  t240_n193: return -0.35852137207984924f;
  t240_n194: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t240_n196;
  t240_n195: return 0.09147956967353821f;
  t240_n196: if (E.getNumReferences() >= 3338616320 /*32671.0*/) goto t240_n198;
  t240_n197: return -0.13022685050964355f;
  t240_n198: if (E.getNumReferences() >= 3340945792 /*41673.5*/) goto t240_n206;
  t240_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t240_n201;
  t240_n200: return -0.0698118582367897f;
  t240_n201: if (E.getNumReferences() >= 3340611840 /*40369.0*/) goto t240_n205;
  t240_n202: if (E.getNumReferences() >= 3339143808 /*34634.5*/) goto t240_n204;
  t240_n203: return 0.079783134162426f;
  t240_n204: return -0.16401799023151398f;
  t240_n205: return 0.08142751455307007f;
  t240_n206: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t240_n210;
  t240_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t240_n209;
  t240_n208: return -0.08953753858804703f;
  t240_n209: return 0.016059797257184982f;
  t240_n210: if (E.getNumReferences() >= 3345444992 /*59248.5*/) goto t240_n218;
  t240_n211: if (E.getNumReferences() >= 3345041664 /*57673.0*/) goto t240_n215;
  t240_n212: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t240_n214;
  t240_n213: return 0.09269391000270844f;
  t240_n214: return -0.04018368571996689f;
  t240_n215: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t240_n217;
  t240_n216: return 0.1288025975227356f;
  t240_n217: return 0.06703753769397736f;
  t240_n218: if (E.getNumReferences() >= 3347876160 /*71954.5*/) goto t240_n220;
  t240_n219: return -0.07196379452943802f;
  t240_n220: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t240_n224;
  t240_n221: if (E.getNumReferences() >= 3348227392 /*74698.5*/) goto t240_n223;
  t240_n222: return 0.005705477669835091f;
  t240_n223: return 0.06615137308835983f;
  t240_n224: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t240_n248;
  t240_n225: if (E.getNumReferences() >= 3350105536 /*89371.5*/) goto t240_n227;
  t240_n226: return -0.079233817756176f;
  t240_n227: if (E.getNumReferences() >= 3350198912 /*90101.0*/) goto t240_n229;
  t240_n228: return 0.0810222178697586f;
  t240_n229: if (E.getNumReferences() >= 3350665472 /*93746.0*/) goto t240_n231;
  t240_n230: return -0.15764111280441284f;
  t240_n231: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t240_n241;
  t240_n232: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t240_n236;
  t240_n233: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t240_n235;
  t240_n234: return -0.0841340571641922f;
  t240_n235: return 0.06599681079387665f;
  t240_n236: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t240_n238;
  t240_n237: return 0.10134374350309372f;
  t240_n238: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t240_n240;
  t240_n239: return 0.032725024968385696f;
  t240_n240: return 0.11409123986959457f;
  t240_n241: if (E.getNumReferences() >= 3354933440 /*127089.5*/) goto t240_n243;
  t240_n242: return -0.1848270446062088f;
  t240_n243: if (E.getNumReferences() >= 3355687776 /*134893.5*/) goto t240_n247;
  t240_n244: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t240_n246;
  t240_n245: return 0.12625554203987122f;
  t240_n246: return 0.032805703580379486f;
  t240_n247: return -0.008053619414567947f;
  t240_n248: return 0.2155979573726654f;
  t240_n249: return 0.07149479538202286f;
  t240_n250: return -0.07962210476398468f;
  t240_n251: return 0.0027671961579471827f;
  t240_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t240_n254;
  t240_n253: return 0.0010252677602693439f;
  t240_n254: return 0.003773104166612029f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree241(const Example& E) {
  t241_n0: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t241_n254;
  t241_n1: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t241_n77;
  t241_n2: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t241_n4;
  t241_n3: return -0.00033519809949211776f;
  t241_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t241_n18;
  t241_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t241_n7;
  t241_n6: return -0.044979408383369446f;
  t241_n7: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t241_n15;
  t241_n8: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t241_n10;
  t241_n9: return 0.0055434126406908035f;
  t241_n10: if (E.getFractionNameInContext() >= 3205753151 /*0.5777778029441833*/) goto t241_n12;
  t241_n11: return 0.05095645412802696f;
  t241_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t241_n14;
  t241_n13: return 0.1305941641330719f;
  t241_n14: return 0.04132694751024246f;
  t241_n15: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t241_n17;
  t241_n16: return -0.5865939259529114f;
  t241_n17: return -0.03702801838517189f;
  t241_n18: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t241_n62;
  t241_n19: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t241_n59;
  t241_n20: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t241_n48;
  t241_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t241_n31;
  t241_n22: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t241_n24;
  t241_n23: return -0.04540759697556496f;
  t241_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t241_n26;
  t241_n25: return -0.02847934328019619f;
  t241_n26: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t241_n28;
  t241_n27: return 0.0664873942732811f;
  t241_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t241_n30;
  t241_n29: return -0.06554994732141495f;
  t241_n30: return 0.0937180444598198f;
  t241_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t241_n33;
  t241_n32: return -0.005952467210590839f;
  t241_n33: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t241_n37;
  t241_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t241_n36;
  t241_n35: return 0.0568867102265358f;
  t241_n36: return 0.18132218718528748f;
  t241_n37: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t241_n47;
  t241_n38: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t241_n46;
  t241_n39: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t241_n45;
  t241_n40: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t241_n42;
  t241_n41: return -0.05250922590494156f;
  t241_n42: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t241_n44;
  t241_n43: return -0.2902040183544159f;
  t241_n44: return 0.09330621361732483f;
  t241_n45: return 0.07082303613424301f;
  t241_n46: return 0.09605759382247925f;
  t241_n47: return -0.07847398519515991f;
  t241_n48: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t241_n54;
  t241_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t241_n51;
  t241_n50: return 0.039963189512491226f;
  t241_n51: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t241_n53;
  t241_n52: return 0.11154621094465256f;
  t241_n53: return 0.032904002815485f;
  t241_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t241_n56;
  t241_n55: return 0.008322076871991158f;
  t241_n56: if (E.getNumReferences() >= 3241148416 /*11.0*/) goto t241_n58;
  t241_n57: return 0.13964664936065674f;
  t241_n58: return 0.008418443612754345f;
  t241_n59: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t241_n61;
  t241_n60: return 0.029022684320807457f;
  t241_n61: return 0.11292622983455658f;
  t241_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n64;
  t241_n63: return -0.04929042607545853f;
  t241_n64: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t241_n76;
  t241_n65: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t241_n75;
  t241_n66: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t241_n74;
  t241_n67: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t241_n71;
  t241_n68: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t241_n70;
  t241_n69: return 0.12640129029750824f;
  t241_n70: return -0.011156857013702393f;
  t241_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t241_n73;
  t241_n72: return 0.12296706438064575f;
  t241_n73: return -0.11887878179550171f;
  t241_n74: return -0.04605984315276146f;
  t241_n75: return 0.026689734309911728f;
  t241_n76: return 0.03396455571055412f;
  t241_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t241_n111;
  t241_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n100;
  t241_n79: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t241_n97;
  t241_n80: if (E.getNumReferences() >= 3342690432 /*48488.5*/) goto t241_n96;
  t241_n81: if (E.getNumReferences() >= 3342258560 /*46801.5*/) goto t241_n93;
  t241_n82: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t241_n88;
  t241_n83: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t241_n87;
  t241_n84: if (E.getNumReferences() >= 3341622016 /*44315.0*/) goto t241_n86;
  t241_n85: return 1.3617105878438451e-06f;
  t241_n86: return -0.22026462852954865f;
  t241_n87: return 0.08511441200971603f;
  t241_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n92;
  t241_n89: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t241_n91;
  t241_n90: return -0.12981510162353516f;
  t241_n91: return -0.361808717250824f;
  t241_n92: return 0.09474318474531174f;
  t241_n93: if (E.getNumReferences() >= 3342316800 /*47029.0*/) goto t241_n95;
  t241_n94: return 0.17001117765903473f;
  t241_n95: return 0.053502406924963f;
  t241_n96: return -0.09111648797988892f;
  t241_n97: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t241_n99;
  t241_n98: return 0.0815359577536583f;
  t241_n99: return 0.011469505727291107f;
  t241_n100: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t241_n102;
  t241_n101: return 0.014089643023908138f;
  t241_n102: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t241_n104;
  t241_n103: return -0.13998964428901672f;
  t241_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t241_n110;
  t241_n105: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t241_n109;
  t241_n106: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t241_n108;
  t241_n107: return 0.13784751296043396f;
  t241_n108: return -0.11111557483673096f;
  t241_n109: return -0.14728941023349762f;
  t241_n110: return 0.09866341948509216f;
  t241_n111: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t241_n251;
  t241_n112: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t241_n244;
  t241_n113: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t241_n241;
  t241_n114: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t241_n230;
  t241_n115: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t241_n225;
  t241_n116: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t241_n128;
  t241_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n127;
  t241_n118: if (E.getNumReferences() >= 3284221952 /*386.5*/) goto t241_n126;
  t241_n119: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t241_n123;
  t241_n120: if (E.getNumReferences() >= 3282452480 /*332.5*/) goto t241_n122;
  t241_n121: return -0.02592809870839119f;
  t241_n122: return -0.9933720827102661f;
  t241_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n125;
  t241_n124: return -1.48714280128479f;
  t241_n125: return -0.1375417858362198f;
  t241_n126: return 0.0211066622287035f;
  t241_n127: return 0.0020106055308133364f;
  t241_n128: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t241_n178;
  t241_n129: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t241_n151;
  t241_n130: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t241_n148;
  t241_n131: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t241_n147;
  t241_n132: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t241_n146;
  t241_n133: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t241_n137;
  t241_n134: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t241_n136;
  t241_n135: return 0.00013207846495788544f;
  t241_n136: return 0.0556768961250782f;
  t241_n137: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t241_n139;
  t241_n138: return 0.0713278204202652f;
  t241_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n141;
  t241_n140: return -0.033787406980991364f;
  t241_n141: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t241_n143;
  t241_n142: return 0.027178319171071053f;
  t241_n143: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t241_n145;
  t241_n144: return 0.1049361303448677f;
  t241_n145: return -0.32828810811042786f;
  t241_n146: return -0.006505084689706564f;
  t241_n147: return 0.22542992234230042f;
  t241_n148: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t241_n150;
  t241_n149: return -0.28409436345100403f;
  t241_n150: return -0.05009569972753525f;
  t241_n151: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t241_n159;
  t241_n152: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t241_n156;
  t241_n153: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t241_n155;
  t241_n154: return -0.2399439811706543f;
  t241_n155: return -1.44333815574646f;
  t241_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t241_n158;
  t241_n157: return -0.03974787890911102f;
  t241_n158: return 0.06977177411317825f;
  t241_n159: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t241_n177;
  t241_n160: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t241_n170;
  t241_n161: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t241_n163;
  t241_n162: return -0.3109188675880432f;
  t241_n163: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t241_n165;
  t241_n164: return -0.045197196304798126f;
  t241_n165: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t241_n169;
  t241_n166: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t241_n168;
  t241_n167: return 0.10282351076602936f;
  t241_n168: return 0.2717810273170471f;
  t241_n169: return 0.06862401962280273f;
  t241_n170: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t241_n176;
  t241_n171: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t241_n175;
  t241_n172: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t241_n174;
  t241_n173: return -0.020586425438523293f;
  t241_n174: return 0.07848884165287018f;
  t241_n175: return -0.037047725170850754f;
  t241_n176: return -0.17417152225971222f;
  t241_n177: return -0.020199883729219437f;
  t241_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t241_n224;
  t241_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t241_n181;
  t241_n180: return 0.015866916626691818f;
  t241_n181: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t241_n191;
  t241_n182: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t241_n188;
  t241_n183: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t241_n187;
  t241_n184: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t241_n186;
  t241_n185: return 0.08593539893627167f;
  t241_n186: return 0.11718116700649261f;
  t241_n187: return -0.05989497900009155f;
  t241_n188: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t241_n190;
  t241_n189: return -0.16083171963691711f;
  t241_n190: return 0.14641037583351135f;
  t241_n191: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t241_n221;
  t241_n192: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t241_n198;
  t241_n193: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t241_n195;
  t241_n194: return -0.036615561693906784f;
  t241_n195: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t241_n197;
  t241_n196: return 0.1597069501876831f;
  t241_n197: return 0.07167138904333115f;
  t241_n198: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t241_n220;
  t241_n199: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t241_n217;
  t241_n200: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t241_n212;
  t241_n201: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t241_n209;
  t241_n202: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t241_n204;
  t241_n203: return -0.06951358169317245f;
  t241_n204: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t241_n208;
  t241_n205: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t241_n207;
  t241_n206: return -0.18985192477703094f;
  t241_n207: return -0.7936413884162903f;
  t241_n208: return -0.07701554894447327f;
  t241_n209: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t241_n211;
  t241_n210: return 0.22904665768146515f;
  t241_n211: return -0.22738230228424072f;
  t241_n212: if (E.getNumReferences() >= 3294953472 /*916.0*/) goto t241_n214;
  t241_n213: return 0.042771194130182266f;
  t241_n214: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t241_n216;
  t241_n215: return 0.10016218572854996f;
  t241_n216: return -1.0134403705596924f;
  t241_n217: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t241_n219;
  t241_n218: return 0.002421804703772068f;
  t241_n219: return 0.06143476814031601f;
  t241_n220: return -0.22239187359809875f;
  t241_n221: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t241_n223;
  t241_n222: return 0.07556381821632385f;
  t241_n223: return -0.4718261957168579f;
  t241_n224: return 0.01655501499772072f;
  t241_n225: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t241_n229;
  t241_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n228;
  t241_n227: return -0.030283071100711823f;
  t241_n228: return 0.07215387374162674f;
  t241_n229: return 0.02462812140583992f;
  t241_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t241_n236;
  t241_n231: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t241_n235;
  t241_n232: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t241_n234;
  t241_n233: return -0.3261910378932953f;
  t241_n234: return -0.16766169667243958f;
  t241_n235: return 0.023066870868206024f;
  t241_n236: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t241_n238;
  t241_n237: return -0.024386078119277954f;
  t241_n238: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t241_n240;
  t241_n239: return 0.10776430368423462f;
  t241_n240: return -0.012424316257238388f;
  t241_n241: if (E.getNumReferences() >= 3296067584 /*984.0*/) goto t241_n243;
  t241_n242: return 0.057963624596595764f;
  t241_n243: return 0.01280209794640541f;
  t241_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t241_n246;
  t241_n245: return 0.000303310458548367f;
  t241_n246: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t241_n250;
  t241_n247: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t241_n249;
  t241_n248: return 0.007979095913469791f;
  t241_n249: return -0.02966386266052723f;
  t241_n250: return 0.04777643457055092f;
  t241_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t241_n253;
  t241_n252: return 0.0005112223443575203f;
  t241_n253: return 0.006007715128362179f;
  t241_n254: return -0.04599275812506676f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree242(const Example& E) {
  t242_n0: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t242_n254;
  t242_n1: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t242_n5;
  t242_n2: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t242_n4;
  t242_n3: return 0.00030198320746421814f;
  t242_n4: return 0.007119204383343458f;
  t242_n5: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t242_n249;
  t242_n6: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t242_n240;
  t242_n7: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t242_n235;
  t242_n8: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t242_n10;
  t242_n9: return -0.0007563820108771324f;
  t242_n10: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t242_n162;
  t242_n11: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t242_n47;
  t242_n12: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t242_n46;
  t242_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t242_n27;
  t242_n14: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t242_n24;
  t242_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n23;
  t242_n16: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t242_n22;
  t242_n17: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n19;
  t242_n18: return 0.006029767449945211f;
  t242_n19: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t242_n21;
  t242_n20: return -0.005206720437854528f;
  t242_n21: return -0.035724882036447525f;
  t242_n22: return -0.022370701655745506f;
  t242_n23: return 0.006663194391876459f;
  t242_n24: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t242_n26;
  t242_n25: return 0.0008205283083952963f;
  t242_n26: return 0.02878389321267605f;
  t242_n27: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t242_n29;
  t242_n28: return 0.011459177359938622f;
  t242_n29: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t242_n43;
  t242_n30: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t242_n42;
  t242_n31: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t242_n41;
  t242_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n38;
  t242_n33: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n35;
  t242_n34: return 0.003554390510544181f;
  t242_n35: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t242_n37;
  t242_n36: return -0.12407870590686798f;
  t242_n37: return -0.05135058984160423f;
  t242_n38: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t242_n40;
  t242_n39: return 0.20917846262454987f;
  t242_n40: return -0.00886567123234272f;
  t242_n41: return 0.05888255313038826f;
  t242_n42: return 0.020510388538241386f;
  t242_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n45;
  t242_n44: return -0.23364819586277008f;
  t242_n45: return 0.03137564659118652f;
  t242_n46: return -0.024519916623830795f;
  t242_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n49;
  t242_n48: return -0.008128311485052109f;
  t242_n49: if (E.getNumReferences() >= 3326554624 /*12748.5*/) goto t242_n159;
  t242_n50: if (E.getNumReferences() >= 3326458880 /*12655.0*/) goto t242_n154;
  t242_n51: if (E.getNumReferences() >= 3321914880 /*8217.5*/) goto t242_n145;
  t242_n52: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t242_n114;
  t242_n53: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t242_n85;
  t242_n54: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n60;
  t242_n55: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t242_n57;
  t242_n56: return 0.05420830845832825f;
  t242_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n59;
  t242_n58: return -0.23488467931747437f;
  t242_n59: return 0.07119862735271454f;
  t242_n60: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t242_n82;
  t242_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n67;
  t242_n62: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t242_n66;
  t242_n63: if (E.getNumReferences() >= 3287597056 /*489.5*/) goto t242_n65;
  t242_n64: return -0.09183064103126526f;
  t242_n65: return 0.02836933173239231f;
  t242_n66: return 0.023956764489412308f;
  t242_n67: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t242_n79;
  t242_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t242_n70;
  t242_n69: return -1.1897974014282227f;
  t242_n70: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t242_n74;
  t242_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n73;
  t242_n72: return -0.03439467400312424f;
  t242_n73: return 0.05696406587958336f;
  t242_n74: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t242_n78;
  t242_n75: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t242_n77;
  t242_n76: return 0.042177192866802216f;
  t242_n77: return 0.14336974918842316f;
  t242_n78: return -0.10635054856538773f;
  t242_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t242_n81;
  t242_n80: return -0.014573554508388042f;
  t242_n81: return 0.11570685356855392f;
  t242_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t242_n84;
  t242_n83: return -0.05806346610188484f;
  t242_n84: return 0.06200898066163063f;
  t242_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t242_n113;
  t242_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t242_n94;
  t242_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t242_n89;
  t242_n88: return 0.03972584381699562f;
  t242_n89: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n93;
  t242_n90: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t242_n92;
  t242_n91: return -0.010077067650854588f;
  t242_n92: return 0.0966048613190651f;
  t242_n93: return 0.0728989765048027f;
  t242_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t242_n96;
  t242_n95: return 0.025060037150979042f;
  t242_n96: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n98;
  t242_n97: return -0.12478908896446228f;
  t242_n98: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t242_n102;
  t242_n99: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t242_n101;
  t242_n100: return 0.11076297610998154f;
  t242_n101: return 0.04788186773657799f;
  t242_n102: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t242_n108;
  t242_n103: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t242_n105;
  t242_n104: return -0.3100817799568176f;
  t242_n105: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t242_n107;
  t242_n106: return 0.12728773057460785f;
  t242_n107: return 0.6446207165718079f;
  t242_n108: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t242_n110;
  t242_n109: return 0.05924580618739128f;
  t242_n110: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t242_n112;
  t242_n111: return 0.1398152858018875f;
  t242_n112: return -0.0017396471230313182f;
  t242_n113: return 0.0038334480486810207f;
  t242_n114: if (E.getNumReferences() >= 3319290880 /*6923.5*/) goto t242_n144;
  t242_n115: if (E.getNumReferences() >= 3318307840 /*6443.5*/) goto t242_n139;
  t242_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t242_n118;
  t242_n117: return 0.00243419804610312f;
  t242_n118: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t242_n136;
  t242_n119: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t242_n135;
  t242_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t242_n122;
  t242_n121: return -0.0018277441849932075f;
  t242_n122: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t242_n132;
  t242_n123: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t242_n125;
  t242_n124: return 0.04888268932700157f;
  t242_n125: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t242_n131;
  t242_n126: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t242_n130;
  t242_n127: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t242_n129;
  t242_n128: return 0.10204467177391052f;
  t242_n129: return -0.17320621013641357f;
  t242_n130: return 0.13066907227039337f;
  t242_n131: return 0.05187670886516571f;
  t242_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t242_n134;
  t242_n133: return -0.31921088695526123f;
  t242_n134: return 0.035988159477710724f;
  t242_n135: return 0.13618846237659454f;
  t242_n136: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t242_n138;
  t242_n137: return 0.004690547473728657f;
  t242_n138: return 0.023895055055618286f;
  t242_n139: if (E.getNumReferences() >= 3318860800 /*6713.5*/) goto t242_n143;
  t242_n140: if (E.getNumReferences() >= 3318737920 /*6653.5*/) goto t242_n142;
  t242_n141: return 0.10283499211072922f;
  t242_n142: return 0.11181309819221497f;
  t242_n143: return 0.011393357999622822f;
  t242_n144: return -0.08393634110689163f;
  t242_n145: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t242_n149;
  t242_n146: if (E.getNumReferences() >= 3324592128 /*10832.0*/) goto t242_n148;
  t242_n147: return 0.092991404235363f;
  t242_n148: return -0.0002984596067108214f;
  t242_n149: if (E.getNumReferences() >= 3325154304 /*11381.0*/) goto t242_n153;
  t242_n150: if (E.getNumReferences() >= 3322211840 /*8507.5*/) goto t242_n152;
  t242_n151: return 0.11636365950107574f;
  t242_n152: return -0.10451441258192062f;
  t242_n153: return 0.10195159912109375f;
  t242_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t242_n158;
  t242_n155: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t242_n157;
  t242_n156: return 0.14652703702449799f;
  t242_n157: return 0.09411461651325226f;
  t242_n158: return -0.09430630505084991f;
  t242_n159: if (E.getNumReferences() >= 3328117248 /*14274.5*/) goto t242_n161;
  t242_n160: return -0.10217908769845963f;
  t242_n161: return 0.001775600016117096f;
  t242_n162: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t242_n212;
  t242_n163: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t242_n199;
  t242_n164: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t242_n198;
  t242_n165: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t242_n195;
  t242_n166: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t242_n192;
  t242_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t242_n177;
  t242_n168: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t242_n174;
  t242_n169: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t242_n171;
  t242_n170: return 0.014034069143235683f;
  t242_n171: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t242_n173;
  t242_n172: return 0.046495210379362106f;
  t242_n173: return 0.10794896632432938f;
  t242_n174: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t242_n176;
  t242_n175: return 0.09072217345237732f;
  t242_n176: return -0.0796840712428093f;
  t242_n177: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t242_n191;
  t242_n178: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n182;
  t242_n179: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t242_n181;
  t242_n180: return 0.07987064123153687f;
  t242_n181: return 0.11016755551099777f;
  t242_n182: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t242_n188;
  t242_n183: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t242_n187;
  t242_n184: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t242_n186;
  t242_n185: return 0.07871860265731812f;
  t242_n186: return 0.1335214376449585f;
  t242_n187: return 0.06847217679023743f;
  t242_n188: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t242_n190;
  t242_n189: return -0.07675406336784363f;
  t242_n190: return 0.10010431706905365f;
  t242_n191: return -0.1947195827960968f;
  t242_n192: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t242_n194;
  t242_n193: return 0.13089653849601746f;
  t242_n194: return 0.0881873145699501f;
  t242_n195: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t242_n197;
  t242_n196: return 0.08222170919179916f;
  t242_n197: return -0.0061048297211527824f;
  t242_n198: return -0.012724081054329872f;
  t242_n199: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t242_n207;
  t242_n200: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t242_n206;
  t242_n201: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t242_n203;
  t242_n202: return 0.01494212169200182f;
  t242_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t242_n205;
  t242_n204: return 0.08027567714452744f;
  t242_n205: return 0.1307416409254074f;
  t242_n206: return -0.033170197159051895f;
  t242_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t242_n211;
  t242_n208: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t242_n210;
  t242_n209: return -0.21254408359527588f;
  t242_n210: return -0.020011812448501587f;
  t242_n211: return 0.021578442305326462f;
  t242_n212: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t242_n220;
  t242_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n219;
  t242_n214: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t242_n216;
  t242_n215: return -0.2451762706041336f;
  t242_n216: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t242_n218;
  t242_n217: return -0.721447229385376f;
  t242_n218: return 0.05013243481516838f;
  t242_n219: return -0.0074885012581944466f;
  t242_n220: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t242_n234;
  t242_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t242_n231;
  t242_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t242_n224;
  t242_n223: return -0.04201148450374603f;
  t242_n224: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t242_n230;
  t242_n225: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t242_n229;
  t242_n226: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t242_n228;
  t242_n227: return -0.04017296060919762f;
  t242_n228: return -0.7770768404006958f;
  t242_n229: return 0.04691823199391365f;
  t242_n230: return 0.049448732286691666f;
  t242_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t242_n233;
  t242_n232: return -0.0020491918548941612f;
  t242_n233: return 0.02545907348394394f;
  t242_n234: return -0.01563025452196598f;
  t242_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t242_n239;
  t242_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t242_n238;
  t242_n237: return -0.3012303411960602f;
  t242_n238: return -0.15642321109771729f;
  t242_n239: return -0.006618548650294542f;
  t242_n240: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t242_n242;
  t242_n241: return 0.0027340499218553305f;
  t242_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t242_n248;
  t242_n243: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t242_n245;
  t242_n244: return 0.003843048121780157f;
  t242_n245: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t242_n247;
  t242_n246: return 0.061190493404865265f;
  t242_n247: return -0.1189180389046669f;
  t242_n248: return 0.023896407335996628f;
  t242_n249: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t242_n253;
  t242_n250: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t242_n252;
  t242_n251: return -0.013090435415506363f;
  t242_n252: return -0.04692133516073227f;
  t242_n253: return 0.07518456131219864f;
  t242_n254: return -0.04397425055503845f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree243(const Example& E) {
  t243_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t243_n126;
  t243_n1: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t243_n9;
  t243_n2: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t243_n8;
  t243_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t243_n5;
  t243_n4: return -0.1557634472846985f;
  t243_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n7;
  t243_n6: return -0.019121289253234863f;
  t243_n7: return 0.06524057686328888f;
  t243_n8: return -0.14445939660072327f;
  t243_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t243_n83;
  t243_n10: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n12;
  t243_n11: return -0.00036328317946754396f;
  t243_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n16;
  t243_n13: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t243_n15;
  t243_n14: return -0.019588949158787727f;
  t243_n15: return 0.045968130230903625f;
  t243_n16: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t243_n20;
  t243_n17: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t243_n19;
  t243_n18: return -0.0033888835459947586f;
  t243_n19: return 0.00955816637724638f;
  t243_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n56;
  t243_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t243_n45;
  t243_n22: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t243_n40;
  t243_n23: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t243_n31;
  t243_n24: if (E.getSymbolScopeDistanceCost() >= 3246391296 /*16.0*/) goto t243_n30;
  t243_n25: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t243_n29;
  t243_n26: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t243_n28;
  t243_n27: return -0.013085062615573406f;
  t243_n28: return 0.14169159531593323f;
  t243_n29: return -0.10157272964715958f;
  t243_n30: return -0.06390992552042007f;
  t243_n31: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t243_n37;
  t243_n32: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t243_n34;
  t243_n33: return 0.016425393521785736f;
  t243_n34: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t243_n36;
  t243_n35: return 0.11530578136444092f;
  t243_n36: return -0.5911545157432556f;
  t243_n37: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t243_n39;
  t243_n38: return 0.06574659794569016f;
  t243_n39: return 0.15203127264976501f;
  t243_n40: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t243_n42;
  t243_n41: return -0.03313996270298958f;
  t243_n42: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t243_n44;
  t243_n43: return -0.19682848453521729f;
  t243_n44: return -0.041946154087781906f;
  t243_n45: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t243_n49;
  t243_n46: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t243_n48;
  t243_n47: return -0.017519468441605568f;
  t243_n48: return 0.009470180608332157f;
  t243_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n55;
  t243_n50: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t243_n54;
  t243_n51: if (E.getFractionNameInContext() >= 3202118087 /*0.4305555522441864*/) goto t243_n53;
  t243_n52: return -0.06262271851301193f;
  t243_n53: return 0.12139367312192917f;
  t243_n54: return -0.1444290578365326f;
  t243_n55: return -0.0434238463640213f;
  t243_n56: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t243_n80;
  t243_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t243_n59;
  t243_n58: return 0.01136066485196352f;
  t243_n59: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t243_n79;
  t243_n60: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t243_n70;
  t243_n61: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t243_n69;
  t243_n62: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t243_n68;
  t243_n63: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t243_n65;
  t243_n64: return 0.09590588510036469f;
  t243_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t243_n67;
  t243_n66: return 0.0028943114448338747f;
  t243_n67: return 0.09127345681190491f;
  t243_n68: return 0.012803379446268082f;
  t243_n69: return 0.09776391834020615f;
  t243_n70: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t243_n78;
  t243_n71: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t243_n73;
  t243_n72: return 0.0179961659014225f;
  t243_n73: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t243_n77;
  t243_n74: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t243_n76;
  t243_n75: return -0.0014066690346226096f;
  t243_n76: return -0.07463154941797256f;
  t243_n77: return 0.052137505263090134f;
  t243_n78: return 0.003795584663748741f;
  t243_n79: return -0.15874698758125305f;
  t243_n80: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t243_n82;
  t243_n81: return -0.021177293732762337f;
  t243_n82: return 0.015183218754827976f;
  t243_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t243_n107;
  t243_n84: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t243_n86;
  t243_n85: return -0.032036297023296356f;
  t243_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t243_n102;
  t243_n87: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t243_n93;
  t243_n88: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t243_n90;
  t243_n89: return 0.014876880683004856f;
  t243_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n92;
  t243_n91: return -0.10215220600366592f;
  t243_n92: return 0.15922768414020538f;
  t243_n93: if (E.getNumReferences() >= 3283959808 /*378.5*/) goto t243_n101;
  t243_n94: if (E.getNumReferences() >= 3279028224 /*242.0*/) goto t243_n100;
  t243_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t243_n97;
  t243_n96: return -0.008718143217265606f;
  t243_n97: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t243_n99;
  t243_n98: return 0.12032875418663025f;
  t243_n99: return -0.0024553898256272078f;
  t243_n100: return 0.17214980721473694f;
  t243_n101: return -0.094612717628479f;
  t243_n102: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t243_n106;
  t243_n103: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t243_n105;
  t243_n104: return 0.12589634954929352f;
  t243_n105: return 0.12582649290561676f;
  t243_n106: return 0.05081828311085701f;
  t243_n107: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t243_n115;
  t243_n108: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t243_n114;
  t243_n109: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t243_n113;
  t243_n110: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t243_n112;
  t243_n111: return 0.03218064084649086f;
  t243_n112: return -0.07240620255470276f;
  t243_n113: return 0.21612028777599335f;
  t243_n114: return 0.057303767651319504f;
  t243_n115: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t243_n119;
  t243_n116: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t243_n118;
  t243_n117: return -0.07291164994239807f;
  t243_n118: return -0.2409866899251938f;
  t243_n119: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t243_n125;
  t243_n120: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t243_n124;
  t243_n121: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t243_n123;
  t243_n122: return 0.08828592300415039f;
  t243_n123: return 0.16442586481571198f;
  t243_n124: return -0.12235421687364578f;
  t243_n125: return 0.01761723682284355f;
  t243_n126: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t243_n170;
  t243_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t243_n169;
  t243_n128: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t243_n132;
  t243_n129: if (E.getNumReferences() >= 3347831872 /*71608.5*/) goto t243_n131;
  t243_n130: return -0.04312409088015556f;
  t243_n131: return 0.16642402112483978f;
  t243_n132: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t243_n150;
  t243_n133: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t243_n137;
  t243_n134: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t243_n136;
  t243_n135: return 0.006744608748704195f;
  t243_n136: return 0.023714611306786537f;
  t243_n137: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t243_n143;
  t243_n138: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t243_n140;
  t243_n139: return -0.045983657240867615f;
  t243_n140: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t243_n142;
  t243_n141: return -2.900564193725586f;
  t243_n142: return -0.3127236068248749f;
  t243_n143: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t243_n145;
  t243_n144: return -0.14431418478488922f;
  t243_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t243_n149;
  t243_n146: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t243_n148;
  t243_n147: return 0.03405733406543732f;
  t243_n148: return -0.31548306345939636f;
  t243_n149: return 0.10642971098423004f;
  t243_n150: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t243_n166;
  t243_n151: if (E.getNumReferences() >= 3305873408 /*2234.0*/) goto t243_n153;
  t243_n152: return -0.014768716879189014f;
  t243_n153: if (E.getNumReferences() >= 3335566848 /*26715.0*/) goto t243_n155;
  t243_n154: return 0.04464159160852432f;
  t243_n155: if (E.getNumReferences() >= 3347359424 /*67917.5*/) goto t243_n165;
  t243_n156: if (E.getNumReferences() >= 3341518848 /*43912.0*/) goto t243_n162;
  t243_n157: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t243_n161;
  t243_n158: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t243_n160;
  t243_n159: return -0.014468868263065815f;
  t243_n160: return -0.5763194561004639f;
  t243_n161: return 0.08489295095205307f;
  t243_n162: if (E.getNumReferences() >= 3344570752 /*55833.5*/) goto t243_n164;
  t243_n163: return -0.7514475584030151f;
  t243_n164: return -0.1520465761423111f;
  t243_n165: return 0.033068131655454636f;
  t243_n166: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t243_n168;
  t243_n167: return 0.004087982699275017f;
  t243_n168: return -1.6582895517349243f;
  t243_n169: return 0.0007132327882573009f;
  t243_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t243_n172;
  t243_n171: return 0.0006068733637221158f;
  t243_n172: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t243_n208;
  t243_n173: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t243_n199;
  t243_n174: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t243_n178;
  t243_n175: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t243_n177;
  t243_n176: return 0.0058585116639733315f;
  t243_n177: return 0.30064788460731506f;
  t243_n178: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t243_n186;
  t243_n179: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t243_n181;
  t243_n180: return 0.08793234825134277f;
  t243_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n185;
  t243_n182: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t243_n184;
  t243_n183: return -0.009534258395433426f;
  t243_n184: return 0.3264893591403961f;
  t243_n185: return 0.02737496979534626f;
  t243_n186: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t243_n198;
  t243_n187: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t243_n195;
  t243_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t243_n192;
  t243_n189: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n191;
  t243_n190: return 0.07014698535203934f;
  t243_n191: return -0.05301621928811073f;
  t243_n192: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n194;
  t243_n193: return 0.07733791321516037f;
  t243_n194: return 0.11182378232479095f;
  t243_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n197;
  t243_n196: return -0.20586904883384705f;
  t243_n197: return -0.011054226197302341f;
  t243_n198: return -0.009372496046125889f;
  t243_n199: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t243_n203;
  t243_n200: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t243_n202;
  t243_n201: return -0.018291817978024483f;
  t243_n202: return 0.010861290618777275f;
  t243_n203: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t243_n207;
  t243_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n206;
  t243_n205: return -0.006551600061357021f;
  t243_n206: return 0.03762345388531685f;
  t243_n207: return -0.017944492399692535f;
  t243_n208: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t243_n234;
  t243_n209: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t243_n217;
  t243_n210: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t243_n216;
  t243_n211: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n213;
  t243_n212: return 0.016586292535066605f;
  t243_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n215;
  t243_n214: return -0.1215130165219307f;
  t243_n215: return 0.01780671626329422f;
  t243_n216: return 0.17577588558197021f;
  t243_n217: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t243_n227;
  t243_n218: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t243_n226;
  t243_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t243_n225;
  t243_n220: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t243_n222;
  t243_n221: return -0.06061296910047531f;
  t243_n222: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t243_n224;
  t243_n223: return 0.003617487847805023f;
  t243_n224: return 0.08925041556358337f;
  t243_n225: return 0.01329794991761446f;
  t243_n226: return -0.014876851812005043f;
  t243_n227: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t243_n233;
  t243_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t243_n230;
  t243_n229: return 0.02988303266465664f;
  t243_n230: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n232;
  t243_n231: return 0.07557956874370575f;
  t243_n232: return 0.0007872446440160275f;
  t243_n233: return 0.004802399780601263f;
  t243_n234: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t243_n254;
  t243_n235: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t243_n237;
  t243_n236: return 0.03751222416758537f;
  t243_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t243_n239;
  t243_n238: return 0.0043651568703353405f;
  t243_n239: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t243_n241;
  t243_n240: return -0.024135036394000053f;
  t243_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t243_n251;
  t243_n242: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t243_n250;
  t243_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t243_n249;
  t243_n244: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t243_n248;
  t243_n245: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t243_n247;
  t243_n246: return 0.05007283017039299f;
  t243_n247: return 0.13751628994941711f;
  t243_n248: return -0.057157985866069794f;
  t243_n249: return -0.007821504957973957f;
  t243_n250: return 0.021610522642731667f;
  t243_n251: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t243_n253;
  t243_n252: return 0.09990081191062927f;
  t243_n253: return -0.14280804991722107f;
  t243_n254: return 0.06684479117393494f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree244(const Example& E) {
  t244_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t244_n2;
  t244_n1: return -0.0003734116326086223f;
  t244_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t244_n28;
  t244_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t244_n13;
  t244_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t244_n6;
  t244_n5: return -0.0011404979741200805f;
  t244_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n12;
  t244_n7: if (E.getNumReferences() >= 3353598912 /*116663.5*/) goto t244_n9;
  t244_n8: return -0.0016161062521860003f;
  t244_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n11;
  t244_n10: return -0.08691240847110748f;
  t244_n11: return 0.005855195689946413f;
  t244_n12: return 0.006134577561169863f;
  t244_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n15;
  t244_n14: return -0.08457301557064056f;
  t244_n15: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t244_n17;
  t244_n16: return -0.013358678668737411f;
  t244_n17: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t244_n19;
  t244_n18: return 0.005987971555441618f;
  t244_n19: if (E.getFractionNameInContext() >= 3185414456 /*0.10818713903427124*/) goto t244_n21;
  t244_n20: return -0.014012845233082771f;
  t244_n21: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t244_n25;
  t244_n22: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n24;
  t244_n23: return 0.0023974289651960135f;
  t244_n24: return 0.015543129295110703f;
  t244_n25: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t244_n27;
  t244_n26: return -0.023615030571818352f;
  t244_n27: return 0.007321357261389494f;
  t244_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor))) goto t244_n252;
  t244_n29: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t244_n129;
  t244_n30: if (E.getFractionNameInContext() >= 3199584274 /*0.355042040348053*/) goto t244_n128;
  t244_n31: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t244_n37;
  t244_n32: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t244_n36;
  t244_n33: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t244_n35;
  t244_n34: return 0.005102019757032394f;
  t244_n35: return -0.007304941304028034f;
  t244_n36: return 0.026983611285686493f;
  t244_n37: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t244_n39;
  t244_n38: return -0.002863102126866579f;
  t244_n39: if (E.getFractionNameInContext() >= 3172460927 /*0.03708791360259056*/) goto t244_n41;
  t244_n40: return 0.17543141543865204f;
  t244_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t244_n43;
  t244_n42: return -0.12265900522470474f;
  t244_n43: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t244_n119;
  t244_n44: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t244_n64;
  t244_n45: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t244_n47;
  t244_n46: return 0.0001238161203218624f;
  t244_n47: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t244_n63;
  t244_n48: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t244_n52;
  t244_n49: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t244_n51;
  t244_n50: return 0.04931746795773506f;
  t244_n51: return 0.22445859014987946f;
  t244_n52: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t244_n62;
  t244_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t244_n55;
  t244_n54: return 0.048806820064783096f;
  t244_n55: if (E.getNumReferences() >= 3267166208 /*94.5*/) goto t244_n61;
  t244_n56: if (E.getNumReferences() >= 3261857792 /*59.0*/) goto t244_n58;
  t244_n57: return -0.10344471782445908f;
  t244_n58: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t244_n60;
  t244_n59: return -0.12993764877319336f;
  t244_n60: return 0.21186847984790802f;
  t244_n61: return 0.033329397439956665f;
  t244_n62: return 0.13684134185314178f;
  t244_n63: return 0.015779003500938416f;
  t244_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t244_n84;
  t244_n65: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t244_n83;
  t244_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t244_n70;
  t244_n67: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t244_n69;
  t244_n68: return 0.14010944962501526f;
  t244_n69: return -0.03321114182472229f;
  t244_n70: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n72;
  t244_n71: return 0.016777105629444122f;
  t244_n72: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t244_n74;
  t244_n73: return -0.033788591623306274f;
  t244_n74: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t244_n82;
  t244_n75: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t244_n79;
  t244_n76: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t244_n78;
  t244_n77: return 0.0922950878739357f;
  t244_n78: return -0.1944308876991272f;
  t244_n79: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t244_n81;
  t244_n80: return 0.12315405905246735f;
  t244_n81: return 0.014258692041039467f;
  t244_n82: return 0.02341039478778839f;
  t244_n83: return -0.004449930042028427f;
  t244_n84: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n102;
  t244_n85: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t244_n87;
  t244_n86: return 0.024870464578270912f;
  t244_n87: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t244_n101;
  t244_n88: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t244_n92;
  t244_n89: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t244_n91;
  t244_n90: return -0.36341145634651184f;
  t244_n91: return 0.3685491979122162f;
  t244_n92: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t244_n100;
  t244_n93: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t244_n99;
  t244_n94: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t244_n96;
  t244_n95: return 0.22911736369132996f;
  t244_n96: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t244_n98;
  t244_n97: return 0.062256306409835815f;
  t244_n98: return 0.34802696108818054f;
  t244_n99: return 0.06250318884849548f;
  t244_n100: return -0.10055302828550339f;
  t244_n101: return 0.02285951003432274f;
  t244_n102: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t244_n104;
  t244_n103: return 0.04480450227856636f;
  t244_n104: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t244_n118;
  t244_n105: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t244_n111;
  t244_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n108;
  t244_n107: return -0.1001005619764328f;
  t244_n108: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t244_n110;
  t244_n109: return 0.09318123757839203f;
  t244_n110: return 0.06189338117837906f;
  t244_n111: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t244_n117;
  t244_n112: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t244_n116;
  t244_n113: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t244_n115;
  t244_n114: return -0.18953928351402283f;
  t244_n115: return 0.09634818881750107f;
  t244_n116: return -0.40306785702705383f;
  t244_n117: return 0.06491498649120331f;
  t244_n118: return 0.02928606979548931f;
  t244_n119: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t244_n125;
  t244_n120: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t244_n122;
  t244_n121: return 0.0004124394617974758f;
  t244_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n124;
  t244_n123: return -0.06203049421310425f;
  t244_n124: return 0.12060758471488953f;
  t244_n125: if (E.getNumReferences() >= 3336926208 /*29370.0*/) goto t244_n127;
  t244_n126: return -0.2733103632926941f;
  t244_n127: return -0.021663723513484f;
  t244_n128: return 0.007063835393637419f;
  t244_n129: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t244_n249;
  t244_n130: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t244_n216;
  t244_n131: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t244_n181;
  t244_n132: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t244_n166;
  t244_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t244_n163;
  t244_n134: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t244_n162;
  t244_n135: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n137;
  t244_n136: return -0.005672896280884743f;
  t244_n137: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t244_n149;
  t244_n138: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t244_n140;
  t244_n139: return -0.08092265576124191f;
  t244_n140: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t244_n142;
  t244_n141: return 0.015272310003638268f;
  t244_n142: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t244_n148;
  t244_n143: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t244_n147;
  t244_n144: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t244_n146;
  t244_n145: return -0.11004878580570221f;
  t244_n146: return -0.18745329976081848f;
  t244_n147: return -0.004614067263901234f;
  t244_n148: return 0.030307039618492126f;
  t244_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t244_n155;
  t244_n150: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t244_n152;
  t244_n151: return -0.020798489451408386f;
  t244_n152: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t244_n154;
  t244_n153: return 0.06760963797569275f;
  t244_n154: return -0.05761341005563736f;
  t244_n155: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t244_n157;
  t244_n156: return -0.15802273154258728f;
  t244_n157: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t244_n159;
  t244_n158: return 0.13691499829292297f;
  t244_n159: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t244_n161;
  t244_n160: return -0.001445611473172903f;
  t244_n161: return 0.08047731965780258f;
  t244_n162: return 0.14621245861053467f;
  t244_n163: if (E.getFractionNameInContext() >= 3175607423 /*0.04880952462553978*/) goto t244_n165;
  t244_n164: return 0.5805178284645081f;
  t244_n165: return -0.008163738995790482f;
  t244_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t244_n168;
  t244_n167: return -0.008107432164251804f;
  t244_n168: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t244_n180;
  t244_n169: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t244_n171;
  t244_n170: return -0.03537306562066078f;
  t244_n171: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t244_n173;
  t244_n172: return -0.26284971833229065f;
  t244_n173: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t244_n179;
  t244_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t244_n178;
  t244_n175: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t244_n177;
  t244_n176: return 0.11065015941858292f;
  t244_n177: return -0.0036330975126475096f;
  t244_n178: return 0.09211015701293945f;
  t244_n179: return 0.0014468147419393063f;
  t244_n180: return 0.015400067903101444f;
  t244_n181: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n207;
  t244_n182: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t244_n204;
  t244_n183: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t244_n197;
  t244_n184: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t244_n192;
  t244_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n187;
  t244_n186: return -0.20324039459228516f;
  t244_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t244_n191;
  t244_n188: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t244_n190;
  t244_n189: return -0.06095843017101288f;
  t244_n190: return -0.2725050151348114f;
  t244_n191: return -0.027054326608777046f;
  t244_n192: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n194;
  t244_n193: return 0.12475544959306717f;
  t244_n194: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t244_n196;
  t244_n195: return -0.14437879621982574f;
  t244_n196: return 0.02245611883699894f;
  t244_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t244_n199;
  t244_n198: return -0.0464458167552948f;
  t244_n199: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n201;
  t244_n200: return 0.03040209785103798f;
  t244_n201: if (E.getFractionNameInContext() >= 3205428743 /*0.5584415793418884*/) goto t244_n203;
  t244_n202: return 0.0953674465417862f;
  t244_n203: return -0.005780932959169149f;
  t244_n204: if (E.getFractionNameInContext() >= 3206464748 /*0.620192289352417*/) goto t244_n206;
  t244_n205: return 0.06421516835689545f;
  t244_n206: return -0.010728883557021618f;
  t244_n207: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t244_n209;
  t244_n208: return -0.021981194615364075f;
  t244_n209: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t244_n213;
  t244_n210: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t244_n212;
  t244_n211: return 0.026366587728261948f;
  t244_n212: return 0.14682626724243164f;
  t244_n213: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t244_n215;
  t244_n214: return 0.11266595125198364f;
  t244_n215: return 0.0773918554186821f;
  t244_n216: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t244_n234;
  t244_n217: if (E.getFractionNameInContext() >= 3205753151 /*0.5777778029441833*/) goto t244_n231;
  t244_n218: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t244_n228;
  t244_n219: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t244_n227;
  t244_n220: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n224;
  t244_n221: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t244_n223;
  t244_n222: return 0.21144956350326538f;
  t244_n223: return 0.05977921560406685f;
  t244_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t244_n226;
  t244_n225: return 0.10136520862579346f;
  t244_n226: return 0.38007915019989014f;
  t244_n227: return -0.19125358760356903f;
  t244_n228: if (E.getFractionNameInContext() >= 3196618889 /*0.2666666805744171*/) goto t244_n230;
  t244_n229: return 0.11150640994310379f;
  t244_n230: return 0.15990425646305084f;
  t244_n231: if (E.getNumNameInContext() >= 3229614080 /*4.0*/) goto t244_n233;
  t244_n232: return 0.17447596788406372f;
  t244_n233: return 0.13089706003665924f;
  t244_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t244_n236;
  t244_n235: return -0.12987715005874634f;
  t244_n236: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t244_n242;
  t244_n237: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t244_n239;
  t244_n238: return 0.008757497183978558f;
  t244_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t244_n241;
  t244_n240: return -0.4375552833080292f;
  t244_n241: return 0.035454604774713516f;
  t244_n242: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t244_n246;
  t244_n243: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t244_n245;
  t244_n244: return 0.011629648506641388f;
  t244_n245: return 0.13299088180065155f;
  t244_n246: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t244_n248;
  t244_n247: return -0.029246363788843155f;
  t244_n248: return 0.11448647826910019f;
  t244_n249: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t244_n251;
  t244_n250: return 0.0011967628961429f;
  t244_n251: return -0.0022843708284199238f;
  t244_n252: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t244_n254;
  t244_n253: return 0.03926731273531914f;
  t244_n254: return -0.1069917157292366f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree245(const Example& E) {
  t245_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t245_n2;
  t245_n1: return -0.000359498051693663f;
  t245_n2: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t245_n6;
  t245_n3: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t245_n5;
  t245_n4: return -0.03430052474141121f;
  t245_n5: return 0.1605476588010788f;
  t245_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type))) goto t245_n254;
  t245_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t245_n29;
  t245_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t245_n10;
  t245_n9: return -0.001174329430796206f;
  t245_n10: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t245_n22;
  t245_n11: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t245_n15;
  t245_n12: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t245_n14;
  t245_n13: return -0.05982746183872223f;
  t245_n14: return -0.5039520263671875f;
  t245_n15: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t245_n19;
  t245_n16: if (E.getNumReferences() >= 3351182400 /*97784.5*/) goto t245_n18;
  t245_n17: return -0.027633022516965866f;
  t245_n18: return -0.17118039727210999f;
  t245_n19: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t245_n21;
  t245_n20: return -0.06944920867681503f;
  t245_n21: return 0.0914275199174881f;
  t245_n22: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t245_n28;
  t245_n23: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n25;
  t245_n24: return 0.008439654484391212f;
  t245_n25: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t245_n27;
  t245_n26: return 0.05218591168522835f;
  t245_n27: return 0.012222792953252792f;
  t245_n28: return -0.05196699500083923f;
  t245_n29: if (E.getFractionNameInContext() >= 3191965208 /*0.1889880895614624*/) goto t245_n131;
  t245_n30: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t245_n102;
  t245_n31: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t245_n51;
  t245_n32: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t245_n34;
  t245_n33: return 0.0018352136248722672f;
  t245_n34: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t245_n36;
  t245_n35: return -0.040054503828287125f;
  t245_n36: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t245_n48;
  t245_n37: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t245_n45;
  t245_n38: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t245_n44;
  t245_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n41;
  t245_n40: return -0.12034693360328674f;
  t245_n41: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t245_n43;
  t245_n42: return 0.09407611936330795f;
  t245_n43: return -0.03304416686296463f;
  t245_n44: return 0.01013240683823824f;
  t245_n45: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t245_n47;
  t245_n46: return 0.016532298177480698f;
  t245_n47: return 0.08707734197378159f;
  t245_n48: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t245_n50;
  t245_n49: return 0.22252990305423737f;
  t245_n50: return 0.03335978835821152f;
  t245_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t245_n53;
  t245_n52: return -0.010203605517745018f;
  t245_n53: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t245_n99;
  t245_n54: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t245_n96;
  t245_n55: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t245_n95;
  t245_n56: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t245_n58;
  t245_n57: return -0.07395073026418686f;
  t245_n58: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t245_n70;
  t245_n59: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t245_n69;
  t245_n60: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t245_n62;
  t245_n61: return -0.10947313904762268f;
  t245_n62: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t245_n66;
  t245_n63: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t245_n65;
  t245_n64: return 0.2987724244594574f;
  t245_n65: return -0.06006776541471481f;
  t245_n66: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t245_n68;
  t245_n67: return 0.0853387862443924f;
  t245_n68: return 0.07848887890577316f;
  t245_n69: return 0.01137124840170145f;
  t245_n70: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t245_n74;
  t245_n71: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t245_n73;
  t245_n72: return 0.01796008087694645f;
  t245_n73: return 0.07361278682947159f;
  t245_n74: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t245_n94;
  t245_n75: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t245_n77;
  t245_n76: return -0.009778098203241825f;
  t245_n77: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t245_n79;
  t245_n78: return 0.02057083509862423f;
  t245_n79: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t245_n87;
  t245_n80: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t245_n82;
  t245_n81: return 0.05637604370713234f;
  t245_n82: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t245_n84;
  t245_n83: return 0.1285371035337448f;
  t245_n84: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t245_n86;
  t245_n85: return 0.08991383016109467f;
  t245_n86: return -0.16594599187374115f;
  t245_n87: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t245_n93;
  t245_n88: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t245_n90;
  t245_n89: return 0.06698117405176163f;
  t245_n90: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n92;
  t245_n91: return -0.5987248420715332f;
  t245_n92: return -0.001597848953679204f;
  t245_n93: return 0.06972460448741913f;
  t245_n94: return 0.04690120741724968f;
  t245_n95: return 0.10837302356958389f;
  t245_n96: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t245_n98;
  t245_n97: return -0.11608116328716278f;
  t245_n98: return 0.17431141436100006f;
  t245_n99: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t245_n101;
  t245_n100: return 0.1281968653202057f;
  t245_n101: return 0.0016245378646999598f;
  t245_n102: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t245_n126;
  t245_n103: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t245_n107;
  t245_n104: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n106;
  t245_n105: return 0.05525490641593933f;
  t245_n106: return 0.08136346936225891f;
  t245_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t245_n109;
  t245_n108: return 0.0005805121036246419f;
  t245_n109: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t245_n111;
  t245_n110: return -0.02673196978867054f;
  t245_n111: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t245_n117;
  t245_n112: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t245_n116;
  t245_n113: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t245_n115;
  t245_n114: return 0.13442789018154144f;
  t245_n115: return -0.10105957835912704f;
  t245_n116: return 0.0246155746281147f;
  t245_n117: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t245_n119;
  t245_n118: return 0.0354086197912693f;
  t245_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t245_n123;
  t245_n120: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t245_n122;
  t245_n121: return 0.06305105984210968f;
  t245_n122: return 0.14864088594913483f;
  t245_n123: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t245_n125;
  t245_n124: return 0.10304757952690125f;
  t245_n125: return 0.01241969596594572f;
  t245_n126: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t245_n128;
  t245_n127: return -0.02506820484995842f;
  t245_n128: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t245_n130;
  t245_n129: return -0.0793084055185318f;
  t245_n130: return -0.24886710941791534f;
  t245_n131: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t245_n203;
  t245_n132: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t245_n154;
  t245_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n151;
  t245_n134: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t245_n150;
  t245_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t245_n147;
  t245_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n140;
  t245_n137: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t245_n139;
  t245_n138: return -0.01517487596720457f;
  t245_n139: return -0.11960815638303757f;
  t245_n140: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t245_n142;
  t245_n141: return -0.0984891727566719f;
  t245_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n144;
  t245_n143: return -0.018438931554555893f;
  t245_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t245_n146;
  t245_n145: return 0.023221777752041817f;
  t245_n146: return -0.3957124650478363f;
  t245_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n149;
  t245_n148: return -0.006924192421138287f;
  t245_n149: return 0.0033308141864836216f;
  t245_n150: return -0.001993641024455428f;
  t245_n151: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t245_n153;
  t245_n152: return -0.0687875896692276f;
  t245_n153: return 0.094387486577034f;
  t245_n154: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t245_n184;
  t245_n155: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t245_n159;
  t245_n156: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t245_n158;
  t245_n157: return -0.015143326483666897f;
  t245_n158: return -0.00269212294369936f;
  t245_n159: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t245_n183;
  t245_n160: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t245_n182;
  t245_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol))) goto t245_n173;
  t245_n162: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t245_n172;
  t245_n163: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t245_n165;
  t245_n164: return -0.03921310976147652f;
  t245_n165: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n167;
  t245_n166: return 0.006609865464270115f;
  t245_n167: if (E.getNumReferences() >= 3239051264 /*9.0*/) goto t245_n171;
  t245_n168: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t245_n170;
  t245_n169: return 0.07650740444660187f;
  t245_n170: return -0.07133954763412476f;
  t245_n171: return 0.09414809942245483f;
  t245_n172: return 0.10041917860507965f;
  t245_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t245_n175;
  t245_n174: return -0.010733540169894695f;
  t245_n175: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t245_n181;
  t245_n176: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t245_n180;
  t245_n177: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t245_n179;
  t245_n178: return -0.032200053334236145f;
  t245_n179: return 0.05582522600889206f;
  t245_n180: return -0.050710294395685196f;
  t245_n181: return 0.1262947916984558f;
  t245_n182: return -0.03276953473687172f;
  t245_n183: return 0.0009022139711305499f;
  t245_n184: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t245_n192;
  t245_n185: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t245_n189;
  t245_n186: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t245_n188;
  t245_n187: return 0.07894652336835861f;
  t245_n188: return 0.018066082149744034f;
  t245_n189: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t245_n191;
  t245_n190: return 0.06004077568650246f;
  t245_n191: return -0.059583358466625214f;
  t245_n192: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t245_n200;
  t245_n193: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t245_n195;
  t245_n194: return 0.015322484076023102f;
  t245_n195: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t245_n197;
  t245_n196: return 0.046841468662023544f;
  t245_n197: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t245_n199;
  t245_n198: return 0.12087278068065643f;
  t245_n199: return -0.0463976226747036f;
  t245_n200: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t245_n202;
  t245_n201: return -0.012702657841145992f;
  t245_n202: return 0.002808025572448969f;
  t245_n203: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t245_n235;
  t245_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t245_n206;
  t245_n205: return -0.038865502923727036f;
  t245_n206: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t245_n212;
  t245_n207: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t245_n209;
  t245_n208: return -0.01935555413365364f;
  t245_n209: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t245_n211;
  t245_n210: return 0.12474100291728973f;
  t245_n211: return 0.07006596028804779f;
  t245_n212: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t245_n230;
  t245_n213: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n229;
  t245_n214: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t245_n218;
  t245_n215: if (E.getNumReferences() >= 3256352768 /*38.0*/) goto t245_n217;
  t245_n216: return -0.13804088532924652f;
  t245_n217: return 0.10459090769290924f;
  t245_n218: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t245_n228;
  t245_n219: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t245_n225;
  t245_n220: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t245_n224;
  t245_n221: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t245_n223;
  t245_n222: return 0.15388786792755127f;
  t245_n223: return 0.09069070219993591f;
  t245_n224: return 0.05790194496512413f;
  t245_n225: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t245_n227;
  t245_n226: return -0.12159159779548645f;
  t245_n227: return 0.18850621581077576f;
  t245_n228: return -0.0018932400271296501f;
  t245_n229: return 0.00946254376322031f;
  t245_n230: if (E.getNumReferences() >= 3276111872 /*197.5*/) goto t245_n232;
  t245_n231: return 0.009953662753105164f;
  t245_n232: if (E.getNumReferences() >= 3276931072 /*210.0*/) goto t245_n234;
  t245_n233: return 0.1184777021408081f;
  t245_n234: return 0.09746058285236359f;
  t245_n235: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t245_n239;
  t245_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t245_n238;
  t245_n237: return -0.04204050451517105f;
  t245_n238: return 0.005682348273694515f;
  t245_n239: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t245_n253;
  t245_n240: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t245_n242;
  t245_n241: return 0.03534229099750519f;
  t245_n242: if (E.getNumReferences() >= 3315879936 /*5258.0*/) goto t245_n250;
  t245_n243: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t245_n249;
  t245_n244: if (E.getNumReferences() >= 3281240064 /*295.5*/) goto t245_n248;
  t245_n245: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t245_n247;
  t245_n246: return -0.6022817492485046f;
  t245_n247: return -0.05618803948163986f;
  t245_n248: return -0.044152967631816864f;
  t245_n249: return -0.027318088337779045f;
  t245_n250: if (E.getNumReferences() >= 3334648576 /*24921.5*/) goto t245_n252;
  t245_n251: return 0.08445850014686584f;
  t245_n252: return -0.08151470124721527f;
  t245_n253: return 8.250831888290122e-05f;
  t245_n254: return 0.0016179888043552637f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree246(const Example& E) {
  t246_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t246_n2;
  t246_n1: return -0.00011523839930305257f;
  t246_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t246_n254;
  t246_n3: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t246_n49;
  t246_n4: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n40;
  t246_n5: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t246_n27;
  t246_n6: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t246_n16;
  t246_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t246_n13;
  t246_n8: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t246_n12;
  t246_n9: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t246_n11;
  t246_n10: return -0.0015133520355448127f;
  t246_n11: return 0.08380194753408432f;
  t246_n12: return 0.13398505747318268f;
  t246_n13: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t246_n15;
  t246_n14: return 0.043134961277246475f;
  t246_n15: return 0.1157626137137413f;
  t246_n16: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t246_n22;
  t246_n17: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t246_n21;
  t246_n18: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t246_n20;
  t246_n19: return 0.09536483883857727f;
  t246_n20: return 0.01305522583425045f;
  t246_n21: return -0.10564026236534119f;
  t246_n22: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t246_n26;
  t246_n23: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t246_n25;
  t246_n24: return 0.07216598093509674f;
  t246_n25: return 0.03531912341713905f;
  t246_n26: return -0.08976440876722336f;
  t246_n27: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t246_n37;
  t246_n28: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t246_n34;
  t246_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t246_n31;
  t246_n30: return -0.06031467765569687f;
  t246_n31: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t246_n33;
  t246_n32: return 0.06236004829406738f;
  t246_n33: return -0.08827032148838043f;
  t246_n34: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t246_n36;
  t246_n35: return 0.059975817799568176f;
  t246_n36: return -0.035387713462114334f;
  t246_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t246_n39;
  t246_n38: return -0.054016388952732086f;
  t246_n39: return 0.0009212780860252678f;
  t246_n40: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t246_n48;
  t246_n41: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t246_n47;
  t246_n42: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t246_n44;
  t246_n43: return 0.03750592842698097f;
  t246_n44: if (E.getFractionNameInContext() >= 3173690027 /*0.0416666679084301*/) goto t246_n46;
  t246_n45: return 0.12985728681087494f;
  t246_n46: return -0.3919292390346527f;
  t246_n47: return 0.1416231095790863f;
  t246_n48: return 0.00706853112205863f;
  t246_n49: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t246_n249;
  t246_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t246_n72;
  t246_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t246_n69;
  t246_n52: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t246_n54;
  t246_n53: return -0.013044447638094425f;
  t246_n54: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t246_n68;
  t246_n55: if (E.getNumReferences() >= 3328351744 /*14503.5*/) goto t246_n65;
  t246_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t246_n58;
  t246_n57: return 0.02765112929046154f;
  t246_n58: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t246_n60;
  t246_n59: return 0.6406610608100891f;
  t246_n60: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t246_n62;
  t246_n61: return -0.3069174885749817f;
  t246_n62: if (E.getNumReferences() >= 3265134592 /*79.0*/) goto t246_n64;
  t246_n63: return 0.09656159579753876f;
  t246_n64: return 0.3093729019165039f;
  t246_n65: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t246_n67;
  t246_n66: return 0.03295198082923889f;
  t246_n67: return 0.24089457094669342f;
  t246_n68: return -0.04095122963190079f;
  t246_n69: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t246_n71;
  t246_n70: return 0.008640548214316368f;
  t246_n71: return 0.024276094511151314f;
  t246_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t246_n102;
  t246_n73: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t246_n85;
  t246_n74: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t246_n76;
  t246_n75: return -0.0023032163735479116f;
  t246_n76: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t246_n78;
  t246_n77: return -0.022722385823726654f;
  t246_n78: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t246_n84;
  t246_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t246_n81;
  t246_n80: return 0.040071725845336914f;
  t246_n81: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t246_n83;
  t246_n82: return 0.0986800342798233f;
  t246_n83: return 0.08276484161615372f;
  t246_n84: return -0.11667083948850632f;
  t246_n85: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t246_n95;
  t246_n86: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t246_n94;
  t246_n87: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t246_n89;
  t246_n88: return 0.07204824686050415f;
  t246_n89: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t246_n91;
  t246_n90: return 0.003226478351280093f;
  t246_n91: if (E.getFractionNameInContext() >= 3205668418 /*0.5727273225784302*/) goto t246_n93;
  t246_n92: return 0.07654830068349838f;
  t246_n93: return -0.06817270815372467f;
  t246_n94: return 0.06514681875705719f;
  t246_n95: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t246_n101;
  t246_n96: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t246_n98;
  t246_n97: return -0.11550026386976242f;
  t246_n98: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t246_n100;
  t246_n99: return 0.14275850355625153f;
  t246_n100: return 0.09465537220239639f;
  t246_n101: return -0.09375683963298798f;
  t246_n102: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t246_n228;
  t246_n103: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t246_n181;
  t246_n104: if (E.getNumReferences() >= 3349322368 /*83253.0*/) goto t246_n180;
  t246_n105: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t246_n107;
  t246_n106: return -0.009219329804182053f;
  t246_n107: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t246_n153;
  t246_n108: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t246_n122;
  t246_n109: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t246_n121;
  t246_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n120;
  t246_n111: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t246_n113;
  t246_n112: return -0.08804876357316971f;
  t246_n113: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t246_n115;
  t246_n114: return 0.10964379459619522f;
  t246_n115: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t246_n117;
  t246_n116: return 0.06373235583305359f;
  t246_n117: if (E.getNumReferences() >= 3250585600 /*24.0*/) goto t246_n119;
  t246_n118: return 0.14058281481266022f;
  t246_n119: return 0.1552269160747528f;
  t246_n120: return -0.11997294425964355f;
  t246_n121: return 0.0007257155375555158f;
  t246_n122: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t246_n126;
  t246_n123: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t246_n125;
  t246_n124: return 0.13422469794750214f;
  t246_n125: return 0.11207707226276398f;
  t246_n126: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t246_n152;
  t246_n127: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t246_n137;
  t246_n128: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t246_n134;
  t246_n129: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n133;
  t246_n130: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t246_n132;
  t246_n131: return 0.09628501534461975f;
  t246_n132: return -0.6268004179000854f;
  t246_n133: return 0.012927882373332977f;
  t246_n134: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n136;
  t246_n135: return -1.913978099822998f;
  t246_n136: return -0.05384018272161484f;
  t246_n137: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t246_n141;
  t246_n138: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n140;
  t246_n139: return 0.31779709458351135f;
  t246_n140: return 0.12666063010692596f;
  t246_n141: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t246_n151;
  t246_n142: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t246_n150;
  t246_n143: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t246_n145;
  t246_n144: return 0.09395194053649902f;
  t246_n145: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t246_n147;
  t246_n146: return -0.10975678265094757f;
  t246_n147: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t246_n149;
  t246_n148: return 0.17345863580703735f;
  t246_n149: return 0.12102869898080826f;
  t246_n150: return 0.010711407288908958f;
  t246_n151: return -0.08948691934347153f;
  t246_n152: return -0.0771520659327507f;
  t246_n153: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t246_n169;
  t246_n154: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t246_n156;
  t246_n155: return 0.05313868448138237f;
  t246_n156: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t246_n158;
  t246_n157: return 0.0026373101864010096f;
  t246_n158: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t246_n160;
  t246_n159: return -0.008200588636100292f;
  t246_n160: if (E.getNumReferences() >= 3269328896 /*111.0*/) goto t246_n168;
  t246_n161: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t246_n167;
  t246_n162: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t246_n166;
  t246_n163: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t246_n165;
  t246_n164: return 0.16393406689167023f;
  t246_n165: return 0.07837630808353424f;
  t246_n166: return 0.11529184132814407f;
  t246_n167: return 0.14654266834259033f;
  t246_n168: return 0.1428803950548172f;
  t246_n169: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t246_n175;
  t246_n170: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t246_n174;
  t246_n171: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t246_n173;
  t246_n172: return -0.04681200906634331f;
  t246_n173: return 0.15910644829273224f;
  t246_n174: return -0.17378166317939758f;
  t246_n175: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t246_n179;
  t246_n176: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t246_n178;
  t246_n177: return 0.04318653419613838f;
  t246_n178: return -0.5799321532249451f;
  t246_n179: return -0.8565914630889893f;
  t246_n180: return -0.430829256772995f;
  t246_n181: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t246_n215;
  t246_n182: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t246_n200;
  t246_n183: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t246_n185;
  t246_n184: return -0.0244259312748909f;
  t246_n185: if (E.getFractionNameInContext() >= 3171801682 /*0.034632034599781036*/) goto t246_n193;
  t246_n186: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t246_n192;
  t246_n187: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t246_n191;
  t246_n188: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t246_n190;
  t246_n189: return 0.00958226341754198f;
  t246_n190: return -0.02467808872461319f;
  t246_n191: return 0.04492644593119621f;
  t246_n192: return -0.008116927929222584f;
  t246_n193: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t246_n197;
  t246_n194: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t246_n196;
  t246_n195: return -0.3043850362300873f;
  t246_n196: return -0.04495767503976822f;
  t246_n197: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t246_n199;
  t246_n198: return -0.10033004730939865f;
  t246_n199: return 0.08888959139585495f;
  t246_n200: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t246_n212;
  t246_n201: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t246_n207;
  t246_n202: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t246_n206;
  t246_n203: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t246_n205;
  t246_n204: return 0.08548477292060852f;
  t246_n205: return -0.041197385638952255f;
  t246_n206: return 0.014027236960828304f;
  t246_n207: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t246_n211;
  t246_n208: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t246_n210;
  t246_n209: return 0.04979310184717178f;
  t246_n210: return 0.011309020221233368f;
  t246_n211: return 0.1618419885635376f;
  t246_n212: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t246_n214;
  t246_n213: return -0.075924351811409f;
  t246_n214: return 0.01908663846552372f;
  t246_n215: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t246_n217;
  t246_n216: return -0.0107919592410326f;
  t246_n217: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t246_n219;
  t246_n218: return -0.5204131007194519f;
  t246_n219: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t246_n225;
  t246_n220: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t246_n222;
  t246_n221: return 0.07790131121873856f;
  t246_n222: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t246_n224;
  t246_n223: return -0.043226778507232666f;
  t246_n224: return 0.08385462313890457f;
  t246_n225: if (E.getNumReferences() >= 3277848576 /*224.0*/) goto t246_n227;
  t246_n226: return -0.16638773679733276f;
  t246_n227: return 0.09370213001966476f;
  t246_n228: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t246_n238;
  t246_n229: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t246_n235;
  t246_n230: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t246_n234;
  t246_n231: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t246_n233;
  t246_n232: return 0.042792629450559616f;
  t246_n233: return 0.020345043390989304f;
  t246_n234: return 0.009567070752382278f;
  t246_n235: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t246_n237;
  t246_n236: return -0.10386711359024048f;
  t246_n237: return -0.13053405284881592f;
  t246_n238: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t246_n246;
  t246_n239: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t246_n243;
  t246_n240: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t246_n242;
  t246_n241: return -0.056045472621917725f;
  t246_n242: return -0.0030365614220499992f;
  t246_n243: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t246_n245;
  t246_n244: return -0.23875583708286285f;
  t246_n245: return 0.024884380400180817f;
  t246_n246: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t246_n248;
  t246_n247: return 0.06143707409501076f;
  t246_n248: return 0.33259856700897217f;
  t246_n249: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t246_n251;
  t246_n250: return 0.007819537073373795f;
  t246_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t246_n253;
  t246_n252: return -0.08541135489940643f;
  t246_n253: return -1.2962172031402588f;
  t246_n254: return 0.0035696011036634445f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree247(const Example& E) {
  t247_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t247_n244;
  t247_n1: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t247_n241;
  t247_n2: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t247_n54;
  t247_n3: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t247_n51;
  t247_n4: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t247_n6;
  t247_n5: return 0.000763066578656435f;
  t247_n6: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t247_n50;
  t247_n7: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t247_n29;
  t247_n8: if (E.getNumReferences() >= 3300950016 /*1540.0*/) goto t247_n12;
  t247_n9: if (E.getNumReferences() >= 3300929536 /*1537.5*/) goto t247_n11;
  t247_n10: return -0.012843016535043716f;
  t247_n11: return -0.2036678045988083f;
  t247_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t247_n28;
  t247_n13: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t247_n23;
  t247_n14: if (E.getNumReferences() >= 3347118592 /*66036.0*/) goto t247_n22;
  t247_n15: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t247_n17;
  t247_n16: return -0.0018620836781337857f;
  t247_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n19;
  t247_n18: return -0.02074679359793663f;
  t247_n19: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t247_n21;
  t247_n20: return 0.1354871690273285f;
  t247_n21: return 0.08369529992341995f;
  t247_n22: return -0.1035885140299797f;
  t247_n23: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t247_n27;
  t247_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t247_n26;
  t247_n25: return -0.059956930577754974f;
  t247_n26: return 0.0710921436548233f;
  t247_n27: return -0.011413118802011013f;
  t247_n28: return 0.030931947752833366f;
  t247_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t247_n45;
  t247_n30: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t247_n44;
  t247_n31: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t247_n37;
  t247_n32: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t247_n36;
  t247_n33: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t247_n35;
  t247_n34: return 0.026825372129678726f;
  t247_n35: return -0.33362752199172974f;
  t247_n36: return 0.004071472212672234f;
  t247_n37: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t247_n39;
  t247_n38: return -0.0683029517531395f;
  t247_n39: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t247_n41;
  t247_n40: return 0.0005293117719702423f;
  t247_n41: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t247_n43;
  t247_n42: return 0.27847322821617126f;
  t247_n43: return -0.1496015191078186f;
  t247_n44: return -0.27746549248695374f;
  t247_n45: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t247_n49;
  t247_n46: if (E.getNumReferences() >= 3323354112 /*9623.0*/) goto t247_n48;
  t247_n47: return 0.03846196085214615f;
  t247_n48: return -0.01989702321588993f;
  t247_n49: return 0.12798883020877838f;
  t247_n50: return -0.07632175832986832f;
  t247_n51: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t247_n53;
  t247_n52: return 0.0044795754365623f;
  t247_n53: return 0.07902766764163971f;
  t247_n54: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t247_n116;
  t247_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n57;
  t247_n56: return -0.018182042986154556f;
  t247_n57: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t247_n97;
  t247_n58: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t247_n70;
  t247_n59: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t247_n65;
  t247_n60: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t247_n64;
  t247_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n63;
  t247_n62: return -0.018447138369083405f;
  t247_n63: return -0.0011045638239011168f;
  t247_n64: return 0.0060295164585113525f;
  t247_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t247_n67;
  t247_n66: return 0.031670983880758286f;
  t247_n67: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t247_n69;
  t247_n68: return 0.18635743856430054f;
  t247_n69: return 0.40824535489082336f;
  t247_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t247_n86;
  t247_n71: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t247_n83;
  t247_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement))) goto t247_n76;
  t247_n73: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t247_n75;
  t247_n74: return -0.009897220879793167f;
  t247_n75: return 0.07469067722558975f;
  t247_n76: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t247_n78;
  t247_n77: return -0.031934984028339386f;
  t247_n78: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t247_n80;
  t247_n79: return 0.05041612684726715f;
  t247_n80: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t247_n82;
  t247_n81: return -0.43732061982154846f;
  t247_n82: return 0.015844276174902916f;
  t247_n83: if (E.getNumReferences() >= 3267821568 /*99.5*/) goto t247_n85;
  t247_n84: return 0.12819433212280273f;
  t247_n85: return 0.035185135900974274f;
  t247_n86: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t247_n96;
  t247_n87: if (E.getNumReferences() >= 3302043648 /*1673.5*/) goto t247_n95;
  t247_n88: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t247_n92;
  t247_n89: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t247_n91;
  t247_n90: return 0.0621437169611454f;
  t247_n91: return 0.11713245511054993f;
  t247_n92: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t247_n94;
  t247_n93: return 0.06949946284294128f;
  t247_n94: return 0.1177181676030159f;
  t247_n95: return 0.03264014795422554f;
  t247_n96: return -0.09955127537250519f;
  t247_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t247_n99;
  t247_n98: return 0.0026486089918762445f;
  t247_n99: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t247_n103;
  t247_n100: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t247_n102;
  t247_n101: return -0.06084069237112999f;
  t247_n102: return 0.03595393896102905f;
  t247_n103: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t247_n105;
  t247_n104: return 0.012044123373925686f;
  t247_n105: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t247_n115;
  t247_n106: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t247_n114;
  t247_n107: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t247_n109;
  t247_n108: return 0.08464700728654861f;
  t247_n109: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t247_n113;
  t247_n110: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n112;
  t247_n111: return -0.024508537724614143f;
  t247_n112: return 0.16218525171279907f;
  t247_n113: return -0.12558504939079285f;
  t247_n114: return 0.0011916211806237698f;
  t247_n115: return -0.027416205033659935f;
  t247_n116: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t247_n126;
  t247_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t247_n119;
  t247_n118: return -0.00029928633011877537f;
  t247_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n121;
  t247_n120: return -0.03664683923125267f;
  t247_n121: if (E.getNumReferences() >= 3294396416 /*882.0*/) goto t247_n123;
  t247_n122: return 0.0662761703133583f;
  t247_n123: if (E.getNumReferences() >= 3306844160 /*2471.0*/) goto t247_n125;
  t247_n124: return 0.13230985403060913f;
  t247_n125: return 0.07314497232437134f;
  t247_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n130;
  t247_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n129;
  t247_n128: return -0.04829917103052139f;
  t247_n129: return -0.0018999316962435842f;
  t247_n130: if (E.getNumReferences() >= 3363960080 /*266152.5*/) goto t247_n236;
  t247_n131: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t247_n235;
  t247_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t247_n226;
  t247_n133: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t247_n219;
  t247_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n150;
  t247_n135: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t247_n149;
  t247_n136: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t247_n142;
  t247_n137: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t247_n139;
  t247_n138: return 0.008461068384349346f;
  t247_n139: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t247_n141;
  t247_n140: return 0.06858716160058975f;
  t247_n141: return 0.030136993154883385f;
  t247_n142: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t247_n146;
  t247_n143: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t247_n145;
  t247_n144: return 0.02804373763501644f;
  t247_n145: return -0.12465889006853104f;
  t247_n146: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t247_n148;
  t247_n147: return 0.049030084162950516f;
  t247_n148: return -0.21036529541015625f;
  t247_n149: return -0.01506116148084402f;
  t247_n150: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t247_n186;
  t247_n151: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t247_n185;
  t247_n152: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t247_n176;
  t247_n153: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t247_n163;
  t247_n154: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t247_n162;
  t247_n155: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t247_n159;
  t247_n156: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t247_n158;
  t247_n157: return -0.40765243768692017f;
  t247_n158: return -0.018662571907043457f;
  t247_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t247_n161;
  t247_n160: return 0.01837405562400818f;
  t247_n161: return 0.04499569162726402f;
  t247_n162: return -0.028651287779211998f;
  t247_n163: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t247_n175;
  t247_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t247_n166;
  t247_n165: return 0.017365187406539917f;
  t247_n166: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t247_n170;
  t247_n167: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t247_n169;
  t247_n168: return 0.08252320438623428f;
  t247_n169: return 0.031934455037117004f;
  t247_n170: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t247_n172;
  t247_n171: return -0.03357852250337601f;
  t247_n172: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t247_n174;
  t247_n173: return 0.058509424328804016f;
  t247_n174: return -0.0930040180683136f;
  t247_n175: return -0.27690747380256653f;
  t247_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n178;
  t247_n177: return -0.008582857437431812f;
  t247_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t247_n180;
  t247_n179: return -0.10293981432914734f;
  t247_n180: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t247_n184;
  t247_n181: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t247_n183;
  t247_n182: return 0.026437899097800255f;
  t247_n183: return 0.10056915134191513f;
  t247_n184: return -0.10393135994672775f;
  t247_n185: return -0.028139842674136162f;
  t247_n186: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t247_n188;
  t247_n187: return -0.10407096892595291f;
  t247_n188: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t247_n218;
  t247_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t247_n215;
  t247_n190: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t247_n204;
  t247_n191: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t247_n203;
  t247_n192: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t247_n200;
  t247_n193: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t247_n199;
  t247_n194: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t247_n196;
  t247_n195: return 0.02460320107638836f;
  t247_n196: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t247_n198;
  t247_n197: return 0.11598843336105347f;
  t247_n198: return 0.05325670912861824f;
  t247_n199: return -0.04438911750912666f;
  t247_n200: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t247_n202;
  t247_n201: return 0.03180444613099098f;
  t247_n202: return 0.07928906381130219f;
  t247_n203: return -0.05348193645477295f;
  t247_n204: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t247_n212;
  t247_n205: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t247_n211;
  t247_n206: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t247_n208;
  t247_n207: return 0.07044526934623718f;
  t247_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n210;
  t247_n209: return -0.28422629833221436f;
  t247_n210: return -0.011632896959781647f;
  t247_n211: return 0.0751943588256836f;
  t247_n212: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t247_n214;
  t247_n213: return -0.15365983545780182f;
  t247_n214: return 0.045851144939661026f;
  t247_n215: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t247_n217;
  t247_n216: return 0.03295176103711128f;
  t247_n217: return -0.14747443795204163f;
  t247_n218: return -0.10229761898517609f;
  t247_n219: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t247_n223;
  t247_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t247_n222;
  t247_n221: return -0.14595608413219452f;
  t247_n222: return -0.030128810554742813f;
  t247_n223: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t247_n225;
  t247_n224: return 0.022208664566278458f;
  t247_n225: return 0.1223871260881424f;
  t247_n226: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t247_n232;
  t247_n227: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t247_n229;
  t247_n228: return 0.016331704333424568f;
  t247_n229: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t247_n231;
  t247_n230: return 0.019352590665221214f;
  t247_n231: return 0.05889981985092163f;
  t247_n232: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t247_n234;
  t247_n233: return 0.01541297696530819f;
  t247_n234: return -0.01199090015143156f;
  t247_n235: return -0.004253922495990992f;
  t247_n236: if (E.getNumReferences() >= 3364158064 /*272339.5*/) goto t247_n240;
  t247_n237: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t247_n239;
  t247_n238: return 0.09640154987573624f;
  t247_n239: return 0.05550083890557289f;
  t247_n240: return 0.011696959845721722f;
  t247_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t247_n243;
  t247_n242: return -0.0258359145373106f;
  t247_n243: return 0.03369399160146713f;
  t247_n244: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t247_n246;
  t247_n245: return 0.0005063940770924091f;
  t247_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t247_n248;
  t247_n247: return 0.0006823037983849645f;
  t247_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t247_n254;
  t247_n249: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t247_n251;
  t247_n250: return 0.006202578078955412f;
  t247_n251: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t247_n253;
  t247_n252: return -0.03719891607761383f;
  t247_n253: return 0.07734166085720062f;
  t247_n254: return -0.06189366802573204f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree248(const Example& E) {
  t248_n0: if (E.getNumReferences() >= 3329553408 /*15677.0*/) goto t248_n62;
  t248_n1: if (E.getNumReferences() >= 3329258496 /*15389.0*/) goto t248_n61;
  t248_n2: if (E.getNumReferences() >= 3328217600 /*14372.5*/) goto t248_n10;
  t248_n3: if (E.getNumReferences() >= 3328027648 /*14187.0*/) goto t248_n9;
  t248_n4: if (E.getNumReferences() >= 3327951360 /*14112.5*/) goto t248_n6;
  t248_n5: return -7.69677571952343e-05f;
  t248_n6: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t248_n8;
  t248_n7: return 0.1095048040151596f;
  t248_n8: return 0.10457184910774231f;
  t248_n9: return -0.13532164692878723f;
  t248_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t248_n22;
  t248_n11: if (E.getNumReferences() >= 3329101824 /*15236.0*/) goto t248_n21;
  t248_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t248_n14;
  t248_n13: return -0.008315442129969597f;
  t248_n14: if (E.getNumReferences() >= 3328893952 /*15033.0*/) goto t248_n16;
  t248_n15: return 0.023336302489042282f;
  t248_n16: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t248_n20;
  t248_n17: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t248_n19;
  t248_n18: return 0.1424151510000229f;
  t248_n19: return 0.08633359521627426f;
  t248_n20: return 0.15573012828826904f;
  t248_n21: return -0.08515160530805588f;
  t248_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t248_n32;
  t248_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t248_n25;
  t248_n24: return -0.12057686597108841f;
  t248_n25: if (E.getNumReferences() >= 3328450048 /*14599.5*/) goto t248_n31;
  t248_n26: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t248_n30;
  t248_n27: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t248_n29;
  t248_n28: return 0.19095003604888916f;
  t248_n29: return 0.10438977926969528f;
  t248_n30: return -0.11508127301931381f;
  t248_n31: return 0.0722145214676857f;
  t248_n32: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t248_n58;
  t248_n33: if (E.getNumReferences() >= 3328655872 /*14800.5*/) goto t248_n35;
  t248_n34: return 0.003526093438267708f;
  t248_n35: if (E.getNumReferences() >= 3329021952 /*15158.0*/) goto t248_n53;
  t248_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t248_n48;
  t248_n37: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t248_n41;
  t248_n38: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t248_n40;
  t248_n39: return -0.16148066520690918f;
  t248_n40: return 0.12190546840429306f;
  t248_n41: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t248_n43;
  t248_n42: return -0.46941545605659485f;
  t248_n43: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t248_n47;
  t248_n44: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t248_n46;
  t248_n45: return 0.18973037600517273f;
  t248_n46: return 0.2193271368741989f;
  t248_n47: return -0.06907968968153f;
  t248_n48: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t248_n50;
  t248_n49: return 0.07992283999919891f;
  t248_n50: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t248_n52;
  t248_n51: return 0.11561501026153564f;
  t248_n52: return -0.0075896866619586945f;
  t248_n53: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t248_n55;
  t248_n54: return 0.08012787252664566f;
  t248_n55: if (E.getNumReferences() >= 3329134080 /*15267.5*/) goto t248_n57;
  t248_n56: return -0.1117081567645073f;
  t248_n57: return 0.11339212954044342f;
  t248_n58: if (E.getNumReferences() >= 3328289792 /*14443.0*/) goto t248_n60;
  t248_n59: return 0.16446633636951447f;
  t248_n60: return -0.06469683349132538f;
  t248_n61: return -0.12649263441562653f;
  t248_n62: if (E.getNumReferences() >= 3330013184 /*16126.0*/) goto t248_n90;
  t248_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t248_n87;
  t248_n64: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t248_n86;
  t248_n65: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t248_n71;
  t248_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t248_n70;
  t248_n67: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t248_n69;
  t248_n68: return -0.013706396333873272f;
  t248_n69: return 0.1941167116165161f;
  t248_n70: return 0.07631704211235046f;
  t248_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t248_n73;
  t248_n72: return -0.01125249918550253f;
  t248_n73: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t248_n85;
  t248_n74: if (E.getFractionNameInContext() >= 3184451170 /*0.10101009905338287*/) goto t248_n82;
  t248_n75: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t248_n79;
  t248_n76: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t248_n78;
  t248_n77: return 0.17944516241550446f;
  t248_n78: return 0.0648629441857338f;
  t248_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t248_n81;
  t248_n80: return 0.11491748690605164f;
  t248_n81: return 0.11646219342947006f;
  t248_n82: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t248_n84;
  t248_n83: return 0.110450878739357f;
  t248_n84: return 0.1628478467464447f;
  t248_n85: return 0.018793512135744095f;
  t248_n86: return 0.15051497519016266f;
  t248_n87: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t248_n89;
  t248_n88: return 0.09342730045318604f;
  t248_n89: return 0.13317276537418365f;
  t248_n90: if (E.getNumReferences() >= 3330058752 /*16170.5*/) goto t248_n92;
  t248_n91: return -0.1930340677499771f;
  t248_n92: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t248_n98;
  t248_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t248_n97;
  t248_n94: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t248_n96;
  t248_n95: return -0.060027897357940674f;
  t248_n96: return 0.08660727739334106f;
  t248_n97: return -0.008282183669507504f;
  t248_n98: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t248_n212;
  t248_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t248_n185;
  t248_n100: if (E.getNumReferences() >= 3331353088 /*18485.0*/) goto t248_n102;
  t248_n101: return -0.05215192958712578f;
  t248_n102: if (E.getNumReferences() >= 3331374336 /*18526.5*/) goto t248_n106;
  t248_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t248_n105;
  t248_n104: return -0.040841493755578995f;
  t248_n105: return 0.10715487599372864f;
  t248_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t248_n130;
  t248_n107: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t248_n127;
  t248_n108: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t248_n120;
  t248_n109: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t248_n119;
  t248_n110: if (E.getNumReferences() >= 3332353792 /*20439.5*/) goto t248_n118;
  t248_n111: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t248_n117;
  t248_n112: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t248_n116;
  t248_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t248_n115;
  t248_n114: return 0.10907907783985138f;
  t248_n115: return 0.16757172346115112f;
  t248_n116: return 0.08832205832004547f;
  t248_n117: return 0.08245214074850082f;
  t248_n118: return -0.18505778908729553f;
  t248_n119: return -0.04119860753417015f;
  t248_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t248_n122;
  t248_n121: return 0.027847636491060257f;
  t248_n122: if (E.getNumReferences() >= 3334111488 /*23872.5*/) goto t248_n126;
  t248_n123: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t248_n125;
  t248_n124: return 0.17971430718898773f;
  t248_n125: return 0.09613870084285736f;
  t248_n126: return 0.09568852186203003f;
  t248_n127: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t248_n129;
  t248_n128: return 0.00661816680803895f;
  t248_n129: return -0.08190029859542847f;
  t248_n130: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t248_n144;
  t248_n131: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t248_n135;
  t248_n132: if (E.getNumReferences() >= 3333501440 /*22681.0*/) goto t248_n134;
  t248_n133: return -0.02416771464049816f;
  t248_n134: return -0.10683444142341614f;
  t248_n135: if (E.getNumReferences() >= 3331802112 /*19362.0*/) goto t248_n143;
  t248_n136: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t248_n138;
  t248_n137: return 0.08222652226686478f;
  t248_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t248_n140;
  t248_n139: return -0.021693743765354156f;
  t248_n140: if (E.getNumReferences() >= 3331773952 /*19307.0*/) goto t248_n142;
  t248_n141: return 0.13251833617687225f;
  t248_n142: return 0.10851931571960449f;
  t248_n143: return -0.0029424570966511965f;
  t248_n144: if (E.getNumReferences() >= 3334345728 /*24330.0*/) goto t248_n150;
  t248_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t248_n147;
  t248_n146: return -0.17999666929244995f;
  t248_n147: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t248_n149;
  t248_n148: return 0.01681874506175518f;
  t248_n149: return 0.08726903051137924f;
  t248_n150: if (E.getNumReferences() >= 3335283968 /*26162.5*/) goto t248_n152;
  t248_n151: return -0.08170419931411743f;
  t248_n152: if (E.getNumReferences() >= 3339162368 /*34707.0*/) goto t248_n184;
  t248_n153: if (E.getNumReferences() >= 3338968320 /*33949.0*/) goto t248_n183;
  t248_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t248_n166;
  t248_n155: if (E.getNumReferences() >= 3338478336 /*32401.5*/) goto t248_n159;
  t248_n156: if (E.getNumReferences() >= 3338286080 /*32026.0*/) goto t248_n158;
  t248_n157: return 0.030664965510368347f;
  t248_n158: return -0.1377706527709961f;
  t248_n159: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t248_n165;
  t248_n160: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t248_n162;
  t248_n161: return 0.08465773612260818f;
  t248_n162: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t248_n164;
  t248_n163: return 0.16406691074371338f;
  t248_n164: return 0.07218173146247864f;
  t248_n165: return 0.0027965730987489223f;
  t248_n166: if (E.getNumReferences() >= 3337362176 /*30221.5*/) goto t248_n180;
  t248_n167: if (E.getNumReferences() >= 3336347136 /*28239.0*/) goto t248_n169;
  t248_n168: return 0.01819685287773609f;
  t248_n169: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t248_n173;
  t248_n170: if (E.getNumReferences() >= 3336861696 /*29244.0*/) goto t248_n172;
  t248_n171: return -0.24466246366500854f;
  t248_n172: return 0.08252490311861038f;
  t248_n173: if (E.getNumReferences() >= 3336926208 /*29370.0*/) goto t248_n179;
  t248_n174: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t248_n176;
  t248_n175: return 0.15298707783222198f;
  t248_n176: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t248_n178;
  t248_n177: return 0.10249950736761093f;
  t248_n178: return 0.2167050689458847f;
  t248_n179: return 0.08392294496297836f;
  t248_n180: if (E.getNumReferences() >= 3338840960 /*33451.5*/) goto t248_n182;
  t248_n181: return -0.11433859169483185f;
  t248_n182: return 0.10214705020189285f;
  t248_n183: return 0.10155636072158813f;
  t248_n184: return 0.003347428049892187f;
  t248_n185: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t248_n191;
  t248_n186: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t248_n190;
  t248_n187: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t248_n189;
  t248_n188: return -0.01654960587620735f;
  t248_n189: return 0.044737156480550766f;
  t248_n190: return 0.15458574891090393f;
  t248_n191: if (E.getNumReferences() >= 3335188480 /*25976.0*/) goto t248_n193;
  t248_n192: return -0.017571711912751198f;
  t248_n193: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t248_n201;
  t248_n194: if (E.getNumReferences() >= 3350654080 /*93657.0*/) goto t248_n196;
  t248_n195: return 0.006871533580124378f;
  t248_n196: if (E.getNumReferences() >= 3352190592 /*105661.0*/) goto t248_n200;
  t248_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t248_n199;
  t248_n198: return 0.04713093116879463f;
  t248_n199: return -0.03309399634599686f;
  t248_n200: return 0.013013623654842377f;
  t248_n201: if (E.getNumReferences() >= 3364253712 /*275328.5*/) goto t248_n205;
  t248_n202: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t248_n204;
  t248_n203: return -0.10039646923542023f;
  t248_n204: return -0.22165046632289886f;
  t248_n205: if (E.getNumReferences() >= 3364286400 /*276350.0*/) goto t248_n211;
  t248_n206: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t248_n210;
  t248_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t248_n209;
  t248_n208: return 0.031403545290231705f;
  t248_n209: return 0.12164130061864853f;
  t248_n210: return 0.14190147817134857f;
  t248_n211: return 0.008308450691401958f;
  t248_n212: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t248_n214;
  t248_n213: return -0.0570656917989254f;
  t248_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t248_n216;
  t248_n215: return -0.012641041539609432f;
  t248_n216: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t248_n218;
  t248_n217: return 0.021702280268073082f;
  t248_n218: if (E.getNumReferences() >= 3332703232 /*21122.0*/) goto t248_n226;
  t248_n219: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t248_n221;
  t248_n220: return 0.006441863253712654f;
  t248_n221: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t248_n223;
  t248_n222: return 0.10938448458909988f;
  t248_n223: if (E.getNumReferences() >= 3332493312 /*20712.0*/) goto t248_n225;
  t248_n224: return -0.10081268101930618f;
  t248_n225: return 0.2496960461139679f;
  t248_n226: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t248_n230;
  t248_n227: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t248_n229;
  t248_n228: return -0.031108509749174118f;
  t248_n229: return -0.13893073797225952f;
  t248_n230: if (E.getNumReferences() >= 3339033984 /*34205.5*/) goto t248_n232;
  t248_n231: return 0.09329540282487869f;
  t248_n232: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t248_n246;
  t248_n233: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t248_n235;
  t248_n234: return -0.0498417392373085f;
  t248_n235: if (E.getNumReferences() >= 3342038016 /*45940.0*/) goto t248_n245;
  t248_n236: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t248_n244;
  t248_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t248_n239;
  t248_n238: return 0.054442983120679855f;
  t248_n239: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t248_n243;
  t248_n240: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t248_n242;
  t248_n241: return 0.13605956733226776f;
  t248_n242: return 0.06446623802185059f;
  t248_n243: return 0.09902752935886383f;
  t248_n244: return 0.1096353605389595f;
  t248_n245: return 0.04754034802317619f;
  t248_n246: if (E.getNumReferences() >= 3359911200 /*200884.5*/) goto t248_n252;
  t248_n247: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t248_n251;
  t248_n248: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t248_n250;
  t248_n249: return -0.05572620406746864f;
  t248_n250: return 0.02290835976600647f;
  t248_n251: return -0.15008467435836792f;
  t248_n252: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t248_n254;
  t248_n253: return 0.05780510604381561f;
  t248_n254: return -0.058975666761398315f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree249(const Example& E) {
  t249_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t249_n2;
  t249_n1: return -0.00012085591879440472f;
  t249_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t249_n4;
  t249_n3: return 0.0010348353534936905f;
  t249_n4: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t249_n206;
  t249_n5: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t249_n205;
  t249_n6: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t249_n170;
  t249_n7: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t249_n167;
  t249_n8: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t249_n60;
  t249_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t249_n17;
  t249_n10: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t249_n16;
  t249_n11: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t249_n13;
  t249_n12: return 0.058787696063518524f;
  t249_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n15;
  t249_n14: return -0.11605405062437057f;
  t249_n15: return 0.01715882122516632f;
  t249_n16: return 0.08386269211769104f;
  t249_n17: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t249_n59;
  t249_n18: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t249_n48;
  t249_n19: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t249_n41;
  t249_n20: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t249_n22;
  t249_n21: return 0.0015990616520866752f;
  t249_n22: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t249_n38;
  t249_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t249_n31;
  t249_n24: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t249_n26;
  t249_n25: return -0.13643133640289307f;
  t249_n26: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t249_n30;
  t249_n27: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t249_n29;
  t249_n28: return 0.044987037777900696f;
  t249_n29: return 0.28117677569389343f;
  t249_n30: return 0.000552432902622968f;
  t249_n31: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t249_n37;
  t249_n32: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t249_n34;
  t249_n33: return -0.1054949164390564f;
  t249_n34: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t249_n36;
  t249_n35: return 0.22491395473480225f;
  t249_n36: return 0.14582180976867676f;
  t249_n37: return -0.07740827649831772f;
  t249_n38: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t249_n40;
  t249_n39: return -0.020349182188510895f;
  t249_n40: return 0.17447873950004578f;
  t249_n41: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t249_n47;
  t249_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n46;
  t249_n43: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t249_n45;
  t249_n44: return -0.0971212387084961f;
  t249_n45: return -0.8067386746406555f;
  t249_n46: return -0.15219168365001678f;
  t249_n47: return -0.07658293098211288f;
  t249_n48: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t249_n58;
  t249_n49: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t249_n53;
  t249_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t249_n52;
  t249_n51: return 0.08869225531816483f;
  t249_n52: return 0.14732405543327332f;
  t249_n53: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t249_n55;
  t249_n54: return -0.0008703027851879597f;
  t249_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n57;
  t249_n56: return -0.0064478530548512936f;
  t249_n57: return 0.1103363186120987f;
  t249_n58: return -0.12898515164852142f;
  t249_n59: return 0.21183638274669647f;
  t249_n60: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t249_n166;
  t249_n61: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t249_n151;
  t249_n62: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t249_n82;
  t249_n63: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t249_n65;
  t249_n64: return -0.03854178637266159f;
  t249_n65: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t249_n73;
  t249_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n68;
  t249_n67: return -0.08038520067930222f;
  t249_n68: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t249_n72;
  t249_n69: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t249_n71;
  t249_n70: return 0.07440595328807831f;
  t249_n71: return -0.0006375951343216002f;
  t249_n72: return 0.11748211085796356f;
  t249_n73: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t249_n77;
  t249_n74: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t249_n76;
  t249_n75: return -0.0024286045227199793f;
  t249_n76: return -0.26023992896080017f;
  t249_n77: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t249_n81;
  t249_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t249_n80;
  t249_n79: return 0.08478452265262604f;
  t249_n80: return 0.09622263163328171f;
  t249_n81: return 0.03134883567690849f;
  t249_n82: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t249_n90;
  t249_n83: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t249_n85;
  t249_n84: return -0.3658659756183624f;
  t249_n85: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t249_n89;
  t249_n86: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t249_n88;
  t249_n87: return 0.07357488572597504f;
  t249_n88: return 0.23902475833892822f;
  t249_n89: return -0.09844443947076797f;
  t249_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n112;
  t249_n91: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t249_n103;
  t249_n92: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t249_n98;
  t249_n93: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t249_n95;
  t249_n94: return -0.020673541352152824f;
  t249_n95: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t249_n97;
  t249_n96: return 0.10171295702457428f;
  t249_n97: return -0.12320776283740997f;
  t249_n98: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t249_n100;
  t249_n99: return -0.04283928871154785f;
  t249_n100: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t249_n102;
  t249_n101: return 0.025080444291234016f;
  t249_n102: return -0.30416548252105713f;
  t249_n103: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t249_n107;
  t249_n104: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t249_n106;
  t249_n105: return -0.15336763858795166f;
  t249_n106: return -1.2617284059524536f;
  t249_n107: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t249_n111;
  t249_n108: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t249_n110;
  t249_n109: return 0.028020506724715233f;
  t249_n110: return -0.16592146456241608f;
  t249_n111: return -0.16067686676979065f;
  t249_n112: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t249_n124;
  t249_n113: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t249_n119;
  t249_n114: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t249_n116;
  t249_n115: return 0.16946083307266235f;
  t249_n116: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t249_n118;
  t249_n117: return -0.06109688803553581f;
  t249_n118: return 0.040636174380779266f;
  t249_n119: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t249_n123;
  t249_n120: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t249_n122;
  t249_n121: return -0.21576841175556183f;
  t249_n122: return -0.08144917339086533f;
  t249_n123: return 0.10079740732908249f;
  t249_n124: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t249_n150;
  t249_n125: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t249_n145;
  t249_n126: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t249_n138;
  t249_n127: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t249_n135;
  t249_n128: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t249_n134;
  t249_n129: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t249_n131;
  t249_n130: return -0.04554837942123413f;
  t249_n131: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t249_n133;
  t249_n132: return 0.094378761947155f;
  t249_n133: return 0.015332215465605259f;
  t249_n134: return -0.019061224535107613f;
  t249_n135: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t249_n137;
  t249_n136: return 0.10683972388505936f;
  t249_n137: return 0.032039035111665726f;
  t249_n138: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t249_n144;
  t249_n139: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t249_n143;
  t249_n140: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t249_n142;
  t249_n141: return -0.11733856797218323f;
  t249_n142: return 0.029814640060067177f;
  t249_n143: return -0.11498810350894928f;
  t249_n144: return 0.004564330447465181f;
  t249_n145: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t249_n149;
  t249_n146: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t249_n148;
  t249_n147: return 0.11157302558422089f;
  t249_n148: return -0.13542933762073517f;
  t249_n149: return 0.11963541805744171f;
  t249_n150: return -0.025725366547703743f;
  t249_n151: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t249_n165;
  t249_n152: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t249_n160;
  t249_n153: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t249_n159;
  t249_n154: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t249_n158;
  t249_n155: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t249_n157;
  t249_n156: return 0.07805066555738449f;
  t249_n157: return -0.10533341765403748f;
  t249_n158: return 0.02223992720246315f;
  t249_n159: return 0.11160359531641006f;
  t249_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t249_n162;
  t249_n161: return -0.010754132643342018f;
  t249_n162: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t249_n164;
  t249_n163: return -0.16680264472961426f;
  t249_n164: return 0.2672592103481293f;
  t249_n165: return -0.019414009526371956f;
  t249_n166: return 0.07983478903770447f;
  t249_n167: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t249_n169;
  t249_n168: return -0.16993781924247742f;
  t249_n169: return -0.1825026273727417f;
  t249_n170: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t249_n180;
  t249_n171: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t249_n177;
  t249_n172: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t249_n176;
  t249_n173: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t249_n175;
  t249_n174: return 0.018981389701366425f;
  t249_n175: return -0.02286849543452263f;
  t249_n176: return 0.06966809183359146f;
  t249_n177: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t249_n179;
  t249_n178: return -0.040280234068632126f;
  t249_n179: return 0.07979682087898254f;
  t249_n180: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t249_n182;
  t249_n181: return 0.0010294170351698995f;
  t249_n182: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t249_n186;
  t249_n183: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t249_n185;
  t249_n184: return 0.05526411160826683f;
  t249_n185: return 0.09237629175186157f;
  t249_n186: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t249_n204;
  t249_n187: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t249_n201;
  t249_n188: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t249_n196;
  t249_n189: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t249_n193;
  t249_n190: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t249_n192;
  t249_n191: return 0.012562853284180164f;
  t249_n192: return 0.09543825685977936f;
  t249_n193: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t249_n195;
  t249_n194: return -0.030518047511577606f;
  t249_n195: return -0.13163548707962036f;
  t249_n196: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t249_n198;
  t249_n197: return 0.03163977339863777f;
  t249_n198: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t249_n200;
  t249_n199: return -0.03533121570944786f;
  t249_n200: return 0.04988966882228851f;
  t249_n201: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t249_n203;
  t249_n202: return 0.010957224294543266f;
  t249_n203: return -0.6856123805046082f;
  t249_n204: return -0.10366316139698029f;
  t249_n205: return -0.10710521042346954f;
  t249_n206: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t249_n246;
  t249_n207: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t249_n229;
  t249_n208: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t249_n224;
  t249_n209: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t249_n223;
  t249_n210: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t249_n218;
  t249_n211: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t249_n217;
  t249_n212: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t249_n214;
  t249_n213: return 0.1440533548593521f;
  t249_n214: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t249_n216;
  t249_n215: return 0.11073219776153564f;
  t249_n216: return -0.17750601470470428f;
  t249_n217: return -0.013147346675395966f;
  t249_n218: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t249_n220;
  t249_n219: return -0.10353177785873413f;
  t249_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t249_n222;
  t249_n221: return 0.12720991671085358f;
  t249_n222: return 0.12931445240974426f;
  t249_n223: return 0.017171518877148628f;
  t249_n224: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t249_n226;
  t249_n225: return -0.055634137243032455f;
  t249_n226: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t249_n228;
  t249_n227: return 0.13254953920841217f;
  t249_n228: return -0.13879606127738953f;
  t249_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t249_n235;
  t249_n230: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t249_n232;
  t249_n231: return -0.05983280390501022f;
  t249_n232: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t249_n234;
  t249_n233: return 0.12750345468521118f;
  t249_n234: return 0.02581658773124218f;
  t249_n235: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t249_n241;
  t249_n236: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t249_n238;
  t249_n237: return 0.09713003039360046f;
  t249_n238: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t249_n240;
  t249_n239: return 0.14457669854164124f;
  t249_n240: return 0.05799271538853645f;
  t249_n241: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t249_n245;
  t249_n242: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t249_n244;
  t249_n243: return 0.05267026647925377f;
  t249_n244: return 0.11697559058666229f;
  t249_n245: return -0.06696378439664841f;
  t249_n246: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t249_n252;
  t249_n247: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t249_n251;
  t249_n248: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t249_n250;
  t249_n249: return -0.08906593918800354f;
  t249_n250: return 0.08571557700634003f;
  t249_n251: return -0.2492140233516693f;
  t249_n252: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t249_n254;
  t249_n253: return 0.011969049461185932f;
  t249_n254: return 0.10412155091762543f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree250(const Example& E) {
  t250_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t250_n176;
  t250_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t250_n3;
  t250_n2: return -7.444595394190401e-05f;
  t250_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t250_n7;
  t250_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t250_n6;
  t250_n5: return -0.12132454663515091f;
  t250_n6: return 0.0011048917658627033f;
  t250_n7: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t250_n175;
  t250_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t250_n110;
  t250_n9: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t250_n107;
  t250_n10: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t250_n106;
  t250_n11: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t250_n13;
  t250_n12: return 0.00612610811367631f;
  t250_n13: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t250_n61;
  t250_n14: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t250_n54;
  t250_n15: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t250_n53;
  t250_n16: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t250_n32;
  t250_n17: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t250_n25;
  t250_n18: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t250_n22;
  t250_n19: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t250_n21;
  t250_n20: return -0.00015368203457910568f;
  t250_n21: return -0.06640305370092392f;
  t250_n22: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t250_n24;
  t250_n23: return 0.038678672164678574f;
  t250_n24: return -0.007924535311758518f;
  t250_n25: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t250_n29;
  t250_n26: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t250_n28;
  t250_n27: return -0.007305561099201441f;
  t250_n28: return -0.04237757995724678f;
  t250_n29: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t250_n31;
  t250_n30: return -0.12097402662038803f;
  t250_n31: return -0.01434897817671299f;
  t250_n32: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t250_n52;
  t250_n33: if (E.getFractionNameInContext() >= 3181279720 /*0.0773809552192688*/) goto t250_n41;
  t250_n34: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t250_n36;
  t250_n35: return -0.11379829049110413f;
  t250_n36: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t250_n40;
  t250_n37: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t250_n39;
  t250_n38: return 0.2242230325937271f;
  t250_n39: return 0.030053265392780304f;
  t250_n40: return -0.08049901574850082f;
  t250_n41: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t250_n45;
  t250_n42: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t250_n44;
  t250_n43: return 0.0835973471403122f;
  t250_n44: return 0.2463977038860321f;
  t250_n45: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t250_n47;
  t250_n46: return -0.04109102860093117f;
  t250_n47: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t250_n49;
  t250_n48: return -0.25999513268470764f;
  t250_n49: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t250_n51;
  t250_n50: return 0.10345831513404846f;
  t250_n51: return -0.1505434364080429f;
  t250_n52: return 0.14507439732551575f;
  t250_n53: return -0.11815554648637772f;
  t250_n54: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t250_n58;
  t250_n55: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t250_n57;
  t250_n56: return -0.14063920080661774f;
  t250_n57: return -0.04923815280199051f;
  t250_n58: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t250_n60;
  t250_n59: return -0.06021460145711899f;
  t250_n60: return 0.07592976838350296f;
  t250_n61: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t250_n73;
  t250_n62: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t250_n70;
  t250_n63: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t250_n69;
  t250_n64: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t250_n66;
  t250_n65: return 0.10642364621162415f;
  t250_n66: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t250_n68;
  t250_n67: return 0.24519257247447968f;
  t250_n68: return -0.11280040442943573f;
  t250_n69: return -0.11872652173042297f;
  t250_n70: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t250_n72;
  t250_n71: return 0.11276467889547348f;
  t250_n72: return 0.04974352568387985f;
  t250_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t250_n75;
  t250_n74: return 0.0036132840905338526f;
  t250_n75: if (E.getNumReferences() >= 3313870848 /*4277.0*/) goto t250_n105;
  t250_n76: if (E.getNumReferences() >= 3286958080 /*470.0*/) goto t250_n102;
  t250_n77: if (E.getNumReferences() >= 3264086016 /*71.0*/) goto t250_n91;
  t250_n78: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t250_n80;
  t250_n79: return -0.025213081389665604f;
  t250_n80: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t250_n86;
  t250_n81: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t250_n83;
  t250_n82: return 0.0752643272280693f;
  t250_n83: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t250_n85;
  t250_n84: return -0.19308362901210785f;
  t250_n85: return 0.09045982360839844f;
  t250_n86: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t250_n90;
  t250_n87: if (E.getNumReferences() >= 3260022784 /*52.0*/) goto t250_n89;
  t250_n88: return 0.1193743646144867f;
  t250_n89: return 0.17496807873249054f;
  t250_n90: return -0.0032102770637720823f;
  t250_n91: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t250_n101;
  t250_n92: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t250_n100;
  t250_n93: if (E.getNumReferences() >= 3274440704 /*172.0*/) goto t250_n95;
  t250_n94: return -0.14166995882987976f;
  t250_n95: if (E.getNumReferences() >= 3275030528 /*181.0*/) goto t250_n97;
  t250_n96: return 0.14569386839866638f;
  t250_n97: if (E.getNumReferences() >= 3280093184 /*260.5*/) goto t250_n99;
  t250_n98: return -0.20066054165363312f;
  t250_n99: return 0.028544271364808083f;
  t250_n100: return 0.12611782550811768f;
  t250_n101: return 0.08565762639045715f;
  t250_n102: if (E.getFractionNameInContext() >= 3188603108 /*0.1388888955116272*/) goto t250_n104;
  t250_n103: return 0.09061522781848907f;
  t250_n104: return 0.10710134357213974f;
  t250_n105: return -0.18518251180648804f;
  t250_n106: return 0.13246114552021027f;
  t250_n107: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t250_n109;
  t250_n108: return -0.10332676768302917f;
  t250_n109: return -0.17737385630607605f;
  t250_n110: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t250_n150;
  t250_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t250_n145;
  t250_n112: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t250_n142;
  t250_n113: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t250_n139;
  t250_n114: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t250_n132;
  t250_n115: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t250_n121;
  t250_n116: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t250_n120;
  t250_n117: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t250_n119;
  t250_n118: return -0.10891043394804001f;
  t250_n119: return 0.08875743299722672f;
  t250_n120: return -0.1517820954322815f;
  t250_n121: if (E.getFractionNameInContext() >= 3181279720 /*0.0773809552192688*/) goto t250_n125;
  t250_n122: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t250_n124;
  t250_n123: return 0.06066865846514702f;
  t250_n124: return -0.0575505755841732f;
  t250_n125: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t250_n131;
  t250_n126: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t250_n128;
  t250_n127: return 0.08546362072229385f;
  t250_n128: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t250_n130;
  t250_n129: return -0.13137206435203552f;
  t250_n130: return 0.08248594403266907f;
  t250_n131: return -0.047235772013664246f;
  t250_n132: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t250_n134;
  t250_n133: return 0.01632385328412056f;
  t250_n134: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t250_n136;
  t250_n135: return -0.2412654459476471f;
  t250_n136: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t250_n138;
  t250_n137: return -0.019347146153450012f;
  t250_n138: return -0.2165651023387909f;
  t250_n139: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t250_n141;
  t250_n140: return -0.11886722594499588f;
  t250_n141: return -0.4045681953430176f;
  t250_n142: if (E.getFractionNameInContext() >= 3188638956 /*0.13942307233810425*/) goto t250_n144;
  t250_n143: return 0.037017542868852615f;
  t250_n144: return 0.1085054948925972f;
  t250_n145: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t250_n149;
  t250_n146: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t250_n148;
  t250_n147: return 0.02421683631837368f;
  t250_n148: return -0.016896545886993408f;
  t250_n149: return 0.07457376271486282f;
  t250_n150: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t250_n152;
  t250_n151: return 0.001787587534636259f;
  t250_n152: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t250_n172;
  t250_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t250_n155;
  t250_n154: return 0.0035410216078162193f;
  t250_n155: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t250_n157;
  t250_n156: return 0.013648148626089096f;
  t250_n157: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t250_n171;
  t250_n158: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t250_n164;
  t250_n159: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t250_n161;
  t250_n160: return 0.024660976603627205f;
  t250_n161: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t250_n163;
  t250_n162: return 0.019418522715568542f;
  t250_n163: return 0.07570170611143112f;
  t250_n164: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t250_n168;
  t250_n165: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t250_n167;
  t250_n166: return 0.06470697373151779f;
  t250_n167: return -0.04463667795062065f;
  t250_n168: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t250_n170;
  t250_n169: return -0.007403425872325897f;
  t250_n170: return 0.04598645120859146f;
  t250_n171: return 0.1325153261423111f;
  t250_n172: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t250_n174;
  t250_n173: return -0.023177823051810265f;
  t250_n174: return 0.011619794182479382f;
  t250_n175: return -0.045837387442588806f;
  t250_n176: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t250_n184;
  t250_n177: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t250_n179;
  t250_n178: return -0.01860186830163002f;
  t250_n179: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t250_n183;
  t250_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t250_n182;
  t250_n181: return 0.03897722437977791f;
  t250_n182: return 0.14235714077949524f;
  t250_n183: return 0.005075282882899046f;
  t250_n184: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t250_n186;
  t250_n185: return -0.11211700737476349f;
  t250_n186: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t250_n190;
  t250_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t250_n189;
  t250_n188: return -0.11847374588251114f;
  t250_n189: return -0.0008745655068196356f;
  t250_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t250_n230;
  t250_n191: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t250_n209;
  t250_n192: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t250_n200;
  t250_n193: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t250_n195;
  t250_n194: return -0.10462729632854462f;
  t250_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t250_n197;
  t250_n196: return 0.032681237906217575f;
  t250_n197: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t250_n199;
  t250_n198: return 0.06472878158092499f;
  t250_n199: return 0.17613214254379272f;
  t250_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t250_n208;
  t250_n201: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t250_n203;
  t250_n202: return 0.007206847425550222f;
  t250_n203: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t250_n207;
  t250_n204: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t250_n206;
  t250_n205: return 0.2085787057876587f;
  t250_n206: return -0.10209211707115173f;
  t250_n207: return -0.060276493430137634f;
  t250_n208: return 0.11032525449991226f;
  t250_n209: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t250_n217;
  t250_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t250_n216;
  t250_n211: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t250_n213;
  t250_n212: return -0.08439483493566513f;
  t250_n213: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t250_n215;
  t250_n214: return -0.6700765490531921f;
  t250_n215: return -0.21274343132972717f;
  t250_n216: return -0.05943842977285385f;
  t250_n217: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t250_n225;
  t250_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t250_n222;
  t250_n219: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t250_n221;
  t250_n220: return -0.12318670004606247f;
  t250_n221: return 0.1651105284690857f;
  t250_n222: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t250_n224;
  t250_n223: return -0.10851152241230011f;
  t250_n224: return 0.38597264885902405f;
  t250_n225: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t250_n227;
  t250_n226: return -0.11500362306833267f;
  t250_n227: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t250_n229;
  t250_n228: return 0.11168976873159409f;
  t250_n229: return 0.0016842068871483207f;
  t250_n230: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t250_n252;
  t250_n231: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t250_n249;
  t250_n232: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t250_n246;
  t250_n233: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t250_n245;
  t250_n234: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t250_n242;
  t250_n235: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t250_n241;
  t250_n236: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t250_n238;
  t250_n237: return -0.027512211352586746f;
  t250_n238: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t250_n240;
  t250_n239: return 0.1137901172041893f;
  t250_n240: return 0.09583625197410583f;
  t250_n241: return 0.292702853679657f;
  t250_n242: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t250_n244;
  t250_n243: return -0.26725876331329346f;
  t250_n244: return 0.05324305221438408f;
  t250_n245: return -0.04317072033882141f;
  t250_n246: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t250_n248;
  t250_n247: return 0.1916436403989792f;
  t250_n248: return -0.0644126757979393f;
  t250_n249: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t250_n251;
  t250_n250: return 0.34577444195747375f;
  t250_n251: return -0.1023002341389656f;
  t250_n252: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t250_n254;
  t250_n253: return 0.1174992173910141f;
  t250_n254: return -0.1449158489704132f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree251(const Example& E) {
  t251_n0: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t251_n116;
  t251_n1: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t251_n95;
  t251_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t251_n94;
  t251_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t251_n93;
  t251_n4: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t251_n6;
  t251_n5: return -0.0026479351799935102f;
  t251_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t251_n62;
  t251_n7: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t251_n45;
  t251_n8: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t251_n38;
  t251_n9: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t251_n31;
  t251_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t251_n18;
  t251_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n13;
  t251_n12: return 0.020172469317913055f;
  t251_n13: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t251_n15;
  t251_n14: return -0.019406944513320923f;
  t251_n15: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t251_n17;
  t251_n16: return 0.6383157968521118f;
  t251_n17: return 0.041137587279081345f;
  t251_n18: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t251_n20;
  t251_n19: return -0.14552322030067444f;
  t251_n20: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t251_n24;
  t251_n21: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t251_n23;
  t251_n22: return 0.13525277376174927f;
  t251_n23: return -0.5216602087020874f;
  t251_n24: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t251_n28;
  t251_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n27;
  t251_n26: return -0.2594146430492401f;
  t251_n27: return 0.09771163761615753f;
  t251_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n30;
  t251_n29: return 0.105148084461689f;
  t251_n30: return 0.07548469305038452f;
  t251_n31: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t251_n33;
  t251_n32: return -0.0012438463745638728f;
  t251_n33: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t251_n37;
  t251_n34: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t251_n36;
  t251_n35: return 0.0135762058198452f;
  t251_n36: return 0.06529653072357178f;
  t251_n37: return 0.007855551317334175f;
  t251_n38: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t251_n40;
  t251_n39: return -0.05711999163031578f;
  t251_n40: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t251_n42;
  t251_n41: return -0.16707104444503784f;
  t251_n42: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t251_n44;
  t251_n43: return 0.12507317960262299f;
  t251_n44: return -0.016374342143535614f;
  t251_n45: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t251_n47;
  t251_n46: return -0.023376494646072388f;
  t251_n47: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t251_n49;
  t251_n48: return -0.024633437395095825f;
  t251_n49: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n57;
  t251_n50: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t251_n54;
  t251_n51: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t251_n53;
  t251_n52: return 0.02364375814795494f;
  t251_n53: return 0.0852794200181961f;
  t251_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t251_n56;
  t251_n55: return -0.003644523210823536f;
  t251_n56: return 0.0345376692712307f;
  t251_n57: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t251_n61;
  t251_n58: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t251_n60;
  t251_n59: return -0.037787456065416336f;
  t251_n60: return 0.07890944182872772f;
  t251_n61: return 0.010564353317022324f;
  t251_n62: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t251_n78;
  t251_n63: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t251_n75;
  t251_n64: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t251_n74;
  t251_n65: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n69;
  t251_n66: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t251_n68;
  t251_n67: return -0.0015337354270741343f;
  t251_n68: return -0.3769509792327881f;
  t251_n69: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t251_n73;
  t251_n70: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t251_n72;
  t251_n71: return 0.0336727537214756f;
  t251_n72: return -0.016065942123532295f;
  t251_n73: return -0.0012630345299839973f;
  t251_n74: return 0.0177592895925045f;
  t251_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t251_n77;
  t251_n76: return 0.06268303096294403f;
  t251_n77: return 0.002985353348776698f;
  t251_n78: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n80;
  t251_n79: return 0.016024909913539886f;
  t251_n80: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t251_n82;
  t251_n81: return 0.013019423931837082f;
  t251_n82: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t251_n92;
  t251_n83: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t251_n89;
  t251_n84: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t251_n86;
  t251_n85: return 0.043342482298612595f;
  t251_n86: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t251_n88;
  t251_n87: return -0.036170948296785355f;
  t251_n88: return 0.06632306426763535f;
  t251_n89: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t251_n91;
  t251_n90: return 0.05788338929414749f;
  t251_n91: return -0.08788464218378067f;
  t251_n92: return -0.08317417651414871f;
  t251_n93: return -0.032546889036893845f;
  t251_n94: return 0.0003901041636709124f;
  t251_n95: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t251_n107;
  t251_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t251_n106;
  t251_n97: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t251_n103;
  t251_n98: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t251_n100;
  t251_n99: return 0.024508539587259293f;
  t251_n100: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t251_n102;
  t251_n101: return 0.08671071380376816f;
  t251_n102: return 0.014090278185904026f;
  t251_n103: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n105;
  t251_n104: return 0.0038975635543465614f;
  t251_n105: return -0.008648641407489777f;
  t251_n106: return 0.0154523104429245f;
  t251_n107: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t251_n111;
  t251_n108: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t251_n110;
  t251_n109: return -0.003128657117486f;
  t251_n110: return -0.03482880815863609f;
  t251_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t251_n113;
  t251_n112: return 0.042593590915203094f;
  t251_n113: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t251_n115;
  t251_n114: return 0.10896634310483932f;
  t251_n115: return -0.10762867331504822f;
  t251_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t251_n254;
  t251_n117: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t251_n119;
  t251_n118: return -0.005972077138721943f;
  t251_n119: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t251_n245;
  t251_n120: if (E.getNumReferences() >= 3291635712 /*713.5*/) goto t251_n242;
  t251_n121: if (E.getNumReferences() >= 3291471872 /*703.5*/) goto t251_n237;
  t251_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t251_n218;
  t251_n123: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t251_n201;
  t251_n124: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t251_n164;
  t251_n125: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t251_n163;
  t251_n126: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t251_n148;
  t251_n127: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t251_n129;
  t251_n128: return -0.009197017177939415f;
  t251_n129: if (E.getNumReferences() >= 3290718208 /*657.5*/) goto t251_n143;
  t251_n130: if (E.getNumReferences() >= 3290619904 /*651.5*/) goto t251_n140;
  t251_n131: if (E.getNumReferences() >= 3290177536 /*624.5*/) goto t251_n139;
  t251_n132: if (E.getNumReferences() >= 3290079232 /*618.5*/) goto t251_n134;
  t251_n133: return 0.10833583772182465f;
  t251_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t251_n136;
  t251_n135: return 0.07214350998401642f;
  t251_n136: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t251_n138;
  t251_n137: return 0.3625253140926361f;
  t251_n138: return 0.11400946229696274f;
  t251_n139: return 0.023080356419086456f;
  t251_n140: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t251_n142;
  t251_n141: return -0.12176434695720673f;
  t251_n142: return 0.2876319885253906f;
  t251_n143: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t251_n145;
  t251_n144: return 0.030949043110013008f;
  t251_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t251_n147;
  t251_n146: return -1.4080199003219604f;
  t251_n147: return 0.047158896923065186f;
  t251_n148: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t251_n162;
  t251_n149: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t251_n151;
  t251_n150: return 0.15324410796165466f;
  t251_n151: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t251_n153;
  t251_n152: return 0.017107179388403893f;
  t251_n153: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t251_n157;
  t251_n154: if (E.getNumReferences() >= 3269984256 /*116.0*/) goto t251_n156;
  t251_n155: return 0.1078404113650322f;
  t251_n156: return 0.2939662039279938f;
  t251_n157: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t251_n159;
  t251_n158: return 0.1508549302816391f;
  t251_n159: if (E.getNumReferences() >= 3280732160 /*280.0*/) goto t251_n161;
  t251_n160: return -0.16498465836048126f;
  t251_n161: return 0.19558805227279663f;
  t251_n162: return 0.03310782462358475f;
  t251_n163: return 0.02902437560260296f;
  t251_n164: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t251_n194;
  t251_n165: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t251_n193;
  t251_n166: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t251_n170;
  t251_n167: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t251_n169;
  t251_n168: return 0.018400128930807114f;
  t251_n169: return -0.3653605878353119f;
  t251_n170: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t251_n174;
  t251_n171: if (E.getFractionNameInContext() >= 3187804193 /*0.1269841343164444*/) goto t251_n173;
  t251_n172: return -0.17439715564250946f;
  t251_n173: return 0.11537493020296097f;
  t251_n174: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t251_n192;
  t251_n175: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t251_n183;
  t251_n176: if (E.getNumReferences() >= 3258318848 /*45.5*/) goto t251_n178;
  t251_n177: return -0.10285642743110657f;
  t251_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t251_n180;
  t251_n179: return 0.053130678832530975f;
  t251_n180: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t251_n182;
  t251_n181: return 0.18125957250595093f;
  t251_n182: return 0.15111680328845978f;
  t251_n183: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t251_n189;
  t251_n184: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t251_n188;
  t251_n185: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t251_n187;
  t251_n186: return -0.20260092616081238f;
  t251_n187: return 0.6406691074371338f;
  t251_n188: return 0.04122689366340637f;
  t251_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t251_n191;
  t251_n190: return 0.07279861718416214f;
  t251_n191: return 0.20392785966396332f;
  t251_n192: return -0.19396010041236877f;
  t251_n193: return 0.00792363379150629f;
  t251_n194: if (E.getNumReferences() >= 3284779008 /*403.5*/) goto t251_n196;
  t251_n195: return 0.11668990552425385f;
  t251_n196: if (E.getNumReferences() >= 3287826432 /*496.5*/) goto t251_n198;
  t251_n197: return -0.11278107762336731f;
  t251_n198: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t251_n200;
  t251_n199: return 0.1103859543800354f;
  t251_n200: return -0.10200589150190353f;
  t251_n201: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t251_n203;
  t251_n202: return 0.004757261369377375f;
  t251_n203: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t251_n217;
  t251_n204: if (E.getNumReferences() >= 3282255872 /*326.5*/) goto t251_n206;
  t251_n205: return 0.021927153691649437f;
  t251_n206: if (E.getNumReferences() >= 3284041728 /*381.0*/) goto t251_n214;
  t251_n207: if (E.getNumReferences() >= 3283501056 /*364.5*/) goto t251_n213;
  t251_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t251_n212;
  t251_n209: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t251_n211;
  t251_n210: return -0.2633465826511383f;
  t251_n211: return 0.14202260971069336f;
  t251_n212: return 0.13737672567367554f;
  t251_n213: return 0.13291983306407928f;
  t251_n214: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t251_n216;
  t251_n215: return 0.06460288166999817f;
  t251_n216: return 0.11530226469039917f;
  t251_n217: return -0.09381242841482162f;
  t251_n218: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t251_n222;
  t251_n219: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t251_n221;
  t251_n220: return 0.1375289410352707f;
  t251_n221: return 0.10326594859361649f;
  t251_n222: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t251_n224;
  t251_n223: return 0.0302895437926054f;
  t251_n224: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t251_n228;
  t251_n225: if (E.getNumReferences() >= 3290841088 /*665.0*/) goto t251_n227;
  t251_n226: return -0.03756100684404373f;
  t251_n227: return -0.5544900298118591f;
  t251_n228: if (E.getNumReferences() >= 3285401600 /*422.5*/) goto t251_n234;
  t251_n229: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t251_n233;
  t251_n230: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t251_n232;
  t251_n231: return 0.0436648465692997f;
  t251_n232: return 0.1275358945131302f;
  t251_n233: return -0.06404566019773483f;
  t251_n234: if (E.getNumReferences() >= 3286974464 /*470.5*/) goto t251_n236;
  t251_n235: return 0.3683401942253113f;
  t251_n236: return 0.1140264943242073f;
  t251_n237: if (E.getNumReferences() >= 3291488256 /*704.5*/) goto t251_n239;
  t251_n238: return 0.2963414490222931f;
  t251_n239: if (E.getNumReferences() >= 3291570176 /*709.5*/) goto t251_n241;
  t251_n240: return -0.05498979985713959f;
  t251_n241: return 0.14288917183876038f;
  t251_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t251_n244;
  t251_n243: return -0.02283237688243389f;
  t251_n244: return -0.0008517478127032518f;
  t251_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t251_n247;
  t251_n246: return -0.0022497500758618116f;
  t251_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t251_n249;
  t251_n248: return 0.011164126917719841f;
  t251_n249: if (E.getNumReferences() >= 3334782208 /*25182.5*/) goto t251_n253;
  t251_n250: if (E.getNumReferences() >= 3334246656 /*24136.5*/) goto t251_n252;
  t251_n251: return 0.08576512336730957f;
  t251_n252: return 0.10340968519449234f;
  t251_n253: return -0.15879981219768524f;
  t251_n254: return 0.005737107712775469f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree252(const Example& E) {
  t252_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t252_n252;
  t252_n1: if (E.getNumReferences() >= 3303133184 /*1806.5*/) goto t252_n7;
  t252_n2: if (E.getNumReferences() >= 3302985728 /*1788.5*/) goto t252_n6;
  t252_n3: if (E.getNumReferences() >= 3302977536 /*1787.5*/) goto t252_n5;
  t252_n4: return -0.00024253195442724973f;
  t252_n5: return 0.17133212089538574f;
  t252_n6: return -0.1004670262336731f;
  t252_n7: if (E.getNumReferences() >= 3303960576 /*1907.5*/) goto t252_n47;
  t252_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t252_n18;
  t252_n9: if (E.getNumReferences() >= 3303182336 /*1812.5*/) goto t252_n17;
  t252_n10: if (E.getNumReferences() >= 3303165952 /*1810.5*/) goto t252_n14;
  t252_n11: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t252_n13;
  t252_n12: return 0.2734017074108124f;
  t252_n13: return -0.14425432682037354f;
  t252_n14: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t252_n16;
  t252_n15: return 0.18852946162223816f;
  t252_n16: return 0.018454860895872116f;
  t252_n17: return 0.009236724115908146f;
  t252_n18: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t252_n44;
  t252_n19: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t252_n43;
  t252_n20: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t252_n22;
  t252_n21: return -0.044861651957035065f;
  t252_n22: if (E.getNumReferences() >= 3303202816 /*1815.0*/) goto t252_n24;
  t252_n23: return -0.07397418469190598f;
  t252_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t252_n32;
  t252_n25: if (E.getNumReferences() >= 3303673856 /*1872.5*/) goto t252_n27;
  t252_n26: return 0.0031670655589550734f;
  t252_n27: if (E.getNumReferences() >= 3303796736 /*1887.5*/) goto t252_n31;
  t252_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t252_n30;
  t252_n29: return 0.08441374450922012f;
  t252_n30: return 0.13288362324237823f;
  t252_n31: return 0.06958982348442078f;
  t252_n32: if (E.getNumReferences() >= 3303448576 /*1845.0*/) goto t252_n42;
  t252_n33: if (E.getNumReferences() >= 3303419904 /*1841.5*/) goto t252_n41;
  t252_n34: if (E.getNumReferences() >= 3303235584 /*1819.0*/) goto t252_n36;
  t252_n35: return 0.163064643740654f;
  t252_n36: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t252_n40;
  t252_n37: if (E.getNumReferences() >= 3303305216 /*1827.5*/) goto t252_n39;
  t252_n38: return 0.2992948591709137f;
  t252_n39: return 0.029540197923779488f;
  t252_n40: return -0.08404616266489029f;
  t252_n41: return 0.1226947084069252f;
  t252_n42: return 0.019951194524765015f;
  t252_n43: return -0.04341047629714012f;
  t252_n44: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t252_n46;
  t252_n45: return 0.0981530249118805f;
  t252_n46: return 0.19727984070777893f;
  t252_n47: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t252_n81;
  t252_n48: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t252_n74;
  t252_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t252_n55;
  t252_n50: if (E.getNumReferences() >= 3304783872 /*2008.0*/) goto t252_n54;
  t252_n51: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t252_n53;
  t252_n52: return 0.1729883849620819f;
  t252_n53: return -0.2318737953901291f;
  t252_n54: return -0.09539715945720673f;
  t252_n55: if (E.getNumReferences() >= 3303981056 /*1910.0*/) goto t252_n57;
  t252_n56: return 0.08736031502485275f;
  t252_n57: if (E.getNumReferences() >= 3304001536 /*1912.5*/) goto t252_n59;
  t252_n58: return -0.26035088300704956f;
  t252_n59: if (E.getNumReferences() >= 3304058880 /*1919.5*/) goto t252_n63;
  t252_n60: if (E.getNumReferences() >= 3304034304 /*1916.5*/) goto t252_n62;
  t252_n61: return -0.10890495777130127f;
  t252_n62: return 0.15689939260482788f;
  t252_n63: if (E.getNumReferences() >= 3304976384 /*2031.5*/) goto t252_n65;
  t252_n64: return -0.04279699921607971f;
  t252_n65: if (E.getNumReferences() >= 3305107456 /*2047.5*/) goto t252_n73;
  t252_n66: if (E.getFractionNameInContext() >= 3189069142 /*0.1458333432674408*/) goto t252_n72;
  t252_n67: if (E.getNumReferences() >= 3305033728 /*2038.5*/) goto t252_n69;
  t252_n68: return -0.017449136823415756f;
  t252_n69: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t252_n71;
  t252_n70: return 0.14874619245529175f;
  t252_n71: return 0.08650492876768112f;
  t252_n72: return 0.13143973052501678f;
  t252_n73: return 0.009342742152512074f;
  t252_n74: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t252_n78;
  t252_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t252_n77;
  t252_n76: return 0.024186206981539726f;
  t252_n77: return 0.1896066516637802f;
  t252_n78: if (E.getNumReferences() >= 3304001536 /*1912.5*/) goto t252_n80;
  t252_n79: return 0.14121337234973907f;
  t252_n80: return -0.007679438684135675f;
  t252_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t252_n251;
  t252_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t252_n244;
  t252_n83: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t252_n85;
  t252_n84: return -0.05391942337155342f;
  t252_n85: if (E.getNumReferences() >= 3372619744 /*549246.0*/) goto t252_n243;
  t252_n86: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t252_n88;
  t252_n87: return -0.10316995531320572f;
  t252_n88: if (E.getNumReferences() >= 3336160512 /*27874.5*/) goto t252_n176;
  t252_n89: if (E.getNumReferences() >= 3332305408 /*20345.0*/) goto t252_n175;
  t252_n90: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t252_n166;
  t252_n91: if (E.getNumReferences() >= 3330499328 /*16817.5*/) goto t252_n165;
  t252_n92: if (E.getNumReferences() >= 3327636992 /*13805.5*/) goto t252_n148;
  t252_n93: if (E.getNumReferences() >= 3327522304 /*13693.5*/) goto t252_n147;
  t252_n94: if (E.getNumReferences() >= 3327053312 /*13235.5*/) goto t252_n134;
  t252_n95: if (E.getNumReferences() >= 3316463616 /*5543.0*/) goto t252_n113;
  t252_n96: if (E.getNumReferences() >= 3314137088 /*4407.0*/) goto t252_n98;
  t252_n97: return 0.0001611550251254812f;
  t252_n98: if (E.getNumReferences() >= 3315159040 /*4906.0*/) goto t252_n108;
  t252_n99: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t252_n101;
  t252_n100: return 0.04633854702115059f;
  t252_n101: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t252_n105;
  t252_n102: if (E.getNumReferences() >= 3314482176 /*4575.5*/) goto t252_n104;
  t252_n103: return 0.14058445394039154f;
  t252_n104: return 0.13816478848457336f;
  t252_n105: if (E.getNumReferences() >= 3314320384 /*4496.5*/) goto t252_n107;
  t252_n106: return 0.1049342006444931f;
  t252_n107: return -0.9507569670677185f;
  t252_n108: if (E.getNumReferences() >= 3316164608 /*5397.0*/) goto t252_n112;
  t252_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t252_n111;
  t252_n110: return -0.04783402010798454f;
  t252_n111: return -0.7294478416442871f;
  t252_n112: return 0.07396332174539566f;
  t252_n113: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t252_n127;
  t252_n114: if (E.getNumReferences() >= 3317762048 /*6177.0*/) goto t252_n116;
  t252_n115: return -0.11502142250537872f;
  t252_n116: if (E.getNumReferences() >= 3324229632 /*10478.0*/) goto t252_n126;
  t252_n117: if (E.getNumReferences() >= 3323824640 /*10082.5*/) goto t252_n119;
  t252_n118: return -0.005073870532214642f;
  t252_n119: if (E.getNumReferences() >= 3323889152 /*10145.5*/) goto t252_n125;
  t252_n120: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t252_n122;
  t252_n121: return 0.1139787882566452f;
  t252_n122: if (E.getNumReferences() >= 3323854336 /*10111.5*/) goto t252_n124;
  t252_n123: return 0.08791010081768036f;
  t252_n124: return -0.13220572471618652f;
  t252_n125: return 0.007354799658060074f;
  t252_n126: return -0.057782549411058426f;
  t252_n127: if (E.getNumReferences() >= 3324623872 /*10863.0*/) goto t252_n129;
  t252_n128: return -0.01583576202392578f;
  t252_n129: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t252_n133;
  t252_n130: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t252_n132;
  t252_n131: return 0.19793888926506042f;
  t252_n132: return -0.10168366134166718f;
  t252_n133: return -0.013430512510240078f;
  t252_n134: if (E.getNumReferences() >= 3327144448 /*13324.5*/) goto t252_n140;
  t252_n135: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t252_n137;
  t252_n136: return -0.10897865146398544f;
  t252_n137: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t252_n139;
  t252_n138: return 0.1594296395778656f;
  t252_n139: return 0.16638007760047913f;
  t252_n140: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t252_n146;
  t252_n141: if (E.getNumReferences() >= 3327378944 /*13553.5*/) goto t252_n143;
  t252_n142: return -0.07133864611387253f;
  t252_n143: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t252_n145;
  t252_n144: return 0.11192436516284943f;
  t252_n145: return -0.10017330199480057f;
  t252_n146: return -0.11179501563310623f;
  t252_n147: return -0.33300670981407166f;
  t252_n148: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t252_n162;
  t252_n149: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t252_n157;
  t252_n150: if (E.getNumReferences() >= 3328016384 /*14176.0*/) goto t252_n156;
  t252_n151: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t252_n155;
  t252_n152: if (E.getNumReferences() >= 3327905792 /*14068.0*/) goto t252_n154;
  t252_n153: return -0.11217347532510757f;
  t252_n154: return 0.10765470564365387f;
  t252_n155: return 0.1054634153842926f;
  t252_n156: return -0.1588621288537979f;
  t252_n157: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t252_n161;
  t252_n158: if (E.getNumReferences() >= 3328821248 /*14962.0*/) goto t252_n160;
  t252_n159: return -0.1309400349855423f;
  t252_n160: return 0.09142516553401947f;
  t252_n161: return 0.114164337515831f;
  t252_n162: if (E.getNumReferences() >= 3330487040 /*16793.5*/) goto t252_n164;
  t252_n163: return 0.007172781508415937f;
  t252_n164: return 0.11197265982627869f;
  t252_n165: return -0.08418220281600952f;
  t252_n166: if (E.getNumReferences() >= 3331793408 /*19345.0*/) goto t252_n168;
  t252_n167: return 0.1214219406247139f;
  t252_n168: if (E.getNumReferences() >= 3331918592 /*19589.5*/) goto t252_n170;
  t252_n169: return -0.19375787675380707f;
  t252_n170: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t252_n174;
  t252_n171: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t252_n173;
  t252_n172: return 0.06940280646085739f;
  t252_n173: return -0.2505825161933899f;
  t252_n174: return -0.0014483650447800756f;
  t252_n175: return -0.05194682627916336f;
  t252_n176: if (E.getNumReferences() >= 3336764160 /*29053.5*/) goto t252_n190;
  t252_n177: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t252_n181;
  t252_n178: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t252_n180;
  t252_n179: return 0.11092915385961533f;
  t252_n180: return 0.12689079344272614f;
  t252_n181: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t252_n185;
  t252_n182: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t252_n184;
  t252_n183: return 0.07263519614934921f;
  t252_n184: return -0.4862916171550751f;
  t252_n185: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t252_n187;
  t252_n186: return -0.13240079581737518f;
  t252_n187: if (E.getNumReferences() >= 3336705792 /*28939.5*/) goto t252_n189;
  t252_n188: return 0.15324975550174713f;
  t252_n189: return 0.08905299752950668f;
  t252_n190: if (E.getNumReferences() >= 3336972544 /*29460.5*/) goto t252_n192;
  t252_n191: return -0.2919430434703827f;
  t252_n192: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t252_n196;
  t252_n193: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t252_n195;
  t252_n194: return 0.11661342531442642f;
  t252_n195: return 0.07448951154947281f;
  t252_n196: if (E.getNumReferences() >= 3338534656 /*32511.5*/) goto t252_n198;
  t252_n197: return -0.13839608430862427f;
  t252_n198: if (E.getNumReferences() >= 3340846976 /*41287.5*/) goto t252_n212;
  t252_n199: if (E.getNumReferences() >= 3340138112 /*38518.5*/) goto t252_n211;
  t252_n200: if (E.getNumReferences() >= 3339130368 /*34582.0*/) goto t252_n210;
  t252_n201: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t252_n209;
  t252_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t252_n204;
  t252_n203: return -0.005290532950311899f;
  t252_n204: if (E.getNumReferences() >= 3338840320 /*33449.0*/) goto t252_n206;
  t252_n205: return 0.04004734754562378f;
  t252_n206: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t252_n208;
  t252_n207: return 0.09961015731096268f;
  t252_n208: return 0.06892485171556473f;
  t252_n209: return -0.310783326625824f;
  t252_n210: return -0.16018353402614594f;
  t252_n211: return 0.09983105957508087f;
  t252_n212: if (E.getNumReferences() >= 3344616064 /*56010.5*/) goto t252_n216;
  t252_n213: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t252_n215;
  t252_n214: return -0.08407217264175415f;
  t252_n215: return 0.011164963245391846f;
  t252_n216: if (E.getNumReferences() >= 3345444992 /*59248.5*/) goto t252_n228;
  t252_n217: if (E.getNumReferences() >= 3345041664 /*57673.0*/) goto t252_n225;
  t252_n218: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t252_n220;
  t252_n219: return 0.09063626825809479f;
  t252_n220: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t252_n222;
  t252_n221: return -0.09989980608224869f;
  t252_n222: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t252_n224;
  t252_n223: return 0.10291720181703568f;
  t252_n224: return -0.1216035857796669f;
  t252_n225: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t252_n227;
  t252_n226: return 0.13887354731559753f;
  t252_n227: return 0.09220459312200546f;
  t252_n228: if (E.getNumReferences() >= 3348207616 /*74544.0*/) goto t252_n230;
  t252_n229: return -0.06867776066064835f;
  t252_n230: if (E.getNumReferences() >= 3348491712 /*76763.5*/) goto t252_n232;
  t252_n231: return 0.06279517710208893f;
  t252_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t252_n242;
  t252_n233: if (E.getNumReferences() >= 3350105536 /*89371.5*/) goto t252_n239;
  t252_n234: if (E.getNumReferences() >= 3349417664 /*83997.5*/) goto t252_n238;
  t252_n235: if (E.getNumReferences() >= 3348704768 /*78428.0*/) goto t252_n237;
  t252_n236: return -0.24131928384304047f;
  t252_n237: return 0.03155224025249481f;
  t252_n238: return -0.19684317708015442f;
  t252_n239: if (E.getNumReferences() >= 3350198912 /*90101.0*/) goto t252_n241;
  t252_n240: return 0.07768499851226807f;
  t252_n241: return -0.0013633514754474163f;
  t252_n242: return 0.21878241002559662f;
  t252_n243: return -0.1678907871246338f;
  t252_n244: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t252_n250;
  t252_n245: if (E.getNumReferences() >= 3307767808 /*2696.5*/) goto t252_n247;
  t252_n246: return -0.07641494274139404f;
  t252_n247: if (E.getNumReferences() >= 3308185600 /*2798.5*/) goto t252_n249;
  t252_n248: return 0.12158030271530151f;
  t252_n249: return 0.02586337737739086f;
  t252_n250: return -0.06810459494590759f;
  t252_n251: return 0.0017539974069222808f;
  t252_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t252_n254;
  t252_n253: return 0.000986419036053121f;
  t252_n254: return 0.0036198378074914217f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree253(const Example& E) {
  t253_n0: if (E.getNumReferences() >= 3325613056 /*11829.0*/) goto t253_n92;
  t253_n1: if (E.getNumReferences() >= 3324400128 /*10644.5*/) goto t253_n59;
  t253_n2: if (E.getNumReferences() >= 3324340736 /*10586.5*/) goto t253_n58;
  t253_n3: if (E.getNumReferences() >= 3321325568 /*7917.0*/) goto t253_n5;
  t253_n4: return -8.121519203996286e-05f;
  t253_n5: if (E.getNumReferences() >= 3322147840 /*8445.0*/) goto t253_n53;
  t253_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n14;
  t253_n7: if (E.getNumReferences() >= 3322118656 /*8416.5*/) goto t253_n13;
  t253_n8: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t253_n10;
  t253_n9: return 0.006778122391551733f;
  t253_n10: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t253_n12;
  t253_n11: return -0.056201204657554626f;
  t253_n12: return -0.4380287230014801f;
  t253_n13: return -0.22748856246471405f;
  t253_n14: if (E.getNumReferences() >= 3322037760 /*8337.5*/) goto t253_n36;
  t253_n15: if (E.getNumReferences() >= 3321819136 /*8158.0*/) goto t253_n35;
  t253_n16: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t253_n18;
  t253_n17: return -0.003758565057069063f;
  t253_n18: if (E.getNumReferences() >= 3321408512 /*7957.5*/) goto t253_n20;
  t253_n19: return -0.056698333472013474f;
  t253_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t253_n26;
  t253_n21: if (E.getNumReferences() >= 3321413632 /*7960.0*/) goto t253_n25;
  t253_n22: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t253_n24;
  t253_n23: return 0.1276753842830658f;
  t253_n24: return 0.02101626619696617f;
  t253_n25: return 0.19943132996559143f;
  t253_n26: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t253_n28;
  t253_n27: return 0.09776128828525543f;
  t253_n28: if (E.getNumReferences() >= 3321802752 /*8150.0*/) goto t253_n30;
  t253_n29: return 0.128321573138237f;
  t253_n30: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t253_n34;
  t253_n31: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t253_n33;
  t253_n32: return 0.1582752764225006f;
  t253_n33: return 0.13035304844379425f;
  t253_n34: return 0.0902610793709755f;
  t253_n35: return -0.0008727031527087092f;
  t253_n36: if (E.getNumReferences() >= 3322086912 /*8385.5*/) goto t253_n46;
  t253_n37: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t253_n41;
  t253_n38: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t253_n40;
  t253_n39: return 0.09607210010290146f;
  t253_n40: return 0.11713583767414093f;
  t253_n41: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t253_n45;
  t253_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t253_n44;
  t253_n43: return -0.10248871147632599f;
  t253_n44: return 0.2215220183134079f;
  t253_n45: return -0.1003078892827034f;
  t253_n46: if (E.getNumReferences() >= 3322118656 /*8416.5*/) goto t253_n48;
  t253_n47: return 0.19438932836055756f;
  t253_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t253_n50;
  t253_n49: return 0.10086923837661743f;
  t253_n50: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t253_n52;
  t253_n51: return 0.14865365624427795f;
  t253_n52: return -0.12372121959924698f;
  t253_n53: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t253_n57;
  t253_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t253_n56;
  t253_n55: return 0.013567850925028324f;
  t253_n56: return 0.09156593680381775f;
  t253_n57: return -0.018481815233826637f;
  t253_n58: return 0.17429175972938538f;
  t253_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t253_n61;
  t253_n60: return -0.06486909091472626f;
  t253_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t253_n63;
  t253_n62: return -0.06846821308135986f;
  t253_n63: if (E.getNumReferences() >= 3325015040 /*11245.0*/) goto t253_n65;
  t253_n64: return 0.0007847179658710957f;
  t253_n65: if (E.getNumReferences() >= 3325511680 /*11730.0*/) goto t253_n91;
  t253_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t253_n74;
  t253_n67: if (E.getNumReferences() >= 3325383680 /*11605.0*/) goto t253_n69;
  t253_n68: return 0.0571204274892807f;
  t253_n69: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t253_n71;
  t253_n70: return 0.09980429708957672f;
  t253_n71: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n73;
  t253_n72: return -0.11737806349992752f;
  t253_n73: return 0.11948325484991074f;
  t253_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t253_n76;
  t253_n75: return 0.034612901508808136f;
  t253_n76: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t253_n86;
  t253_n77: if (E.getNumReferences() >= 3325351424 /*11573.5*/) goto t253_n85;
  t253_n78: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t253_n82;
  t253_n79: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n81;
  t253_n80: return -0.23394566774368286f;
  t253_n81: return 0.10207793861627579f;
  t253_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t253_n84;
  t253_n83: return 0.1360122263431549f;
  t253_n84: return -0.10146766155958176f;
  t253_n85: return 0.11522858589887619f;
  t253_n86: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t253_n88;
  t253_n87: return -0.1116589605808258f;
  t253_n88: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t253_n90;
  t253_n89: return 0.12254016846418381f;
  t253_n90: return -0.10525618493556976f;
  t253_n91: return -0.15869306027889252f;
  t253_n92: if (E.getNumReferences() >= 3325976064 /*12183.5*/) goto t253_n96;
  t253_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t253_n95;
  t253_n94: return -0.22872860729694366f;
  t253_n95: return 0.08065535128116608f;
  t253_n96: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t253_n254;
  t253_n97: if (E.getNumReferences() >= 3327362048 /*13537.0*/) goto t253_n111;
  t253_n98: if (E.getNumReferences() >= 3326313984 /*12513.5*/) goto t253_n100;
  t253_n99: return -0.00999701488763094f;
  t253_n100: if (E.getNumReferences() >= 3326337536 /*12536.5*/) goto t253_n102;
  t253_n101: return 0.15675309300422668f;
  t253_n102: if (E.getNumReferences() >= 3326389760 /*12587.5*/) goto t253_n104;
  t253_n103: return -0.44049304723739624f;
  t253_n104: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t253_n110;
  t253_n105: if (E.getNumReferences() >= 3326745088 /*12934.5*/) goto t253_n107;
  t253_n106: return 0.02736271359026432f;
  t253_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t253_n109;
  t253_n108: return 0.028716782107949257f;
  t253_n109: return 0.15250112116336823f;
  t253_n110: return 0.03620478883385658f;
  t253_n111: if (E.getNumReferences() >= 3331498752 /*18769.5*/) goto t253_n115;
  t253_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t253_n114;
  t253_n113: return -0.020762117579579353f;
  t253_n114: return 0.1447726935148239f;
  t253_n115: if (E.getNumReferences() >= 3331859712 /*19474.5*/) goto t253_n129;
  t253_n116: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t253_n120;
  t253_n117: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t253_n119;
  t253_n118: return 0.031003009527921677f;
  t253_n119: return 0.18314632773399353f;
  t253_n120: if (E.getNumReferences() >= 3331773952 /*19307.0*/) goto t253_n126;
  t253_n121: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t253_n125;
  t253_n122: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n124;
  t253_n123: return 0.14624089002609253f;
  t253_n124: return 0.107379250228405f;
  t253_n125: return -0.01243122760206461f;
  t253_n126: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t253_n128;
  t253_n127: return 0.10816754400730133f;
  t253_n128: return 0.0027620841283351183f;
  t253_n129: if (E.getNumReferences() >= 3331912192 /*19577.0*/) goto t253_n131;
  t253_n130: return -0.14573360979557037f;
  t253_n131: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t253_n157;
  t253_n132: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t253_n154;
  t253_n133: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t253_n143;
  t253_n134: if (E.getNumReferences() >= 3332148992 /*20039.5*/) goto t253_n142;
  t253_n135: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t253_n137;
  t253_n136: return 0.012626209296286106f;
  t253_n137: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t253_n141;
  t253_n138: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t253_n140;
  t253_n139: return 0.09765913337469101f;
  t253_n140: return 0.0636298879981041f;
  t253_n141: return -0.3396815061569214f;
  t253_n142: return -0.18930327892303467f;
  t253_n143: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t253_n153;
  t253_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n150;
  t253_n145: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t253_n149;
  t253_n146: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t253_n148;
  t253_n147: return 0.09196925163269043f;
  t253_n148: return 0.03852362930774689f;
  t253_n149: return -0.16995620727539062f;
  t253_n150: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t253_n152;
  t253_n151: return -0.03857005015015602f;
  t253_n152: return 0.09637495130300522f;
  t253_n153: return -0.27314746379852295f;
  t253_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t253_n156;
  t253_n155: return 0.01274467445909977f;
  t253_n156: return 0.11202048510313034f;
  t253_n157: if (E.getNumReferences() >= 3332993536 /*21689.0*/) goto t253_n159;
  t253_n158: return -0.08500894904136658f;
  t253_n159: if (E.getNumReferences() >= 3334419968 /*24475.0*/) goto t253_n171;
  t253_n160: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n170;
  t253_n161: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t253_n167;
  t253_n162: if (E.getNumReferences() >= 3333992192 /*23639.5*/) goto t253_n166;
  t253_n163: if (E.getNumReferences() >= 3333620992 /*22914.5*/) goto t253_n165;
  t253_n164: return 0.06026998162269592f;
  t253_n165: return -0.05590580403804779f;
  t253_n166: return 0.08922752737998962f;
  t253_n167: if (E.getNumReferences() >= 3333934848 /*23527.5*/) goto t253_n169;
  t253_n168: return 0.04727589711546898f;
  t253_n169: return 0.13226276636123657f;
  t253_n170: return -0.0197854395955801f;
  t253_n171: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t253_n195;
  t253_n172: if (E.getNumReferences() >= 3337362176 /*30221.5*/) goto t253_n190;
  t253_n173: if (E.getNumReferences() >= 3336055296 /*27669.0*/) goto t253_n177;
  t253_n174: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t253_n176;
  t253_n175: return -0.11432745307683945f;
  t253_n176: return 0.07823356986045837f;
  t253_n177: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t253_n179;
  t253_n178: return 0.13380731642246246f;
  t253_n179: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t253_n187;
  t253_n180: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t253_n186;
  t253_n181: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t253_n185;
  t253_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n184;
  t253_n183: return 0.01546520460397005f;
  t253_n184: return 0.18403393030166626f;
  t253_n185: return 0.19045336544513702f;
  t253_n186: return 0.033323902636766434f;
  t253_n187: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n189;
  t253_n188: return 0.12706251442432404f;
  t253_n189: return -0.035923901945352554f;
  t253_n190: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t253_n194;
  t253_n191: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t253_n193;
  t253_n192: return -0.013871436938643456f;
  t253_n193: return 0.13648952543735504f;
  t253_n194: return -0.09547091275453568f;
  t253_n195: if (E.getNumReferences() >= 3339033984 /*34205.5*/) goto t253_n199;
  t253_n196: if (E.getNumReferences() >= 3338840960 /*33451.5*/) goto t253_n198;
  t253_n197: return 0.05901963263750076f;
  t253_n198: return 0.10188651829957962f;
  t253_n199: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t253_n203;
  t253_n200: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t253_n202;
  t253_n201: return -0.07920675724744797f;
  t253_n202: return 0.11904682964086533f;
  t253_n203: if (E.getNumReferences() >= 3342583296 /*48070.0*/) goto t253_n217;
  t253_n204: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t253_n212;
  t253_n205: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t253_n207;
  t253_n206: return 0.04188518971204758f;
  t253_n207: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t253_n211;
  t253_n208: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n210;
  t253_n209: return 0.1227310448884964f;
  t253_n210: return 0.09036314487457275f;
  t253_n211: return 0.10602591931819916f;
  t253_n212: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n214;
  t253_n213: return 0.08059763163328171f;
  t253_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t253_n216;
  t253_n215: return 0.1323566883802414f;
  t253_n216: return 0.022886764258146286f;
  t253_n217: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t253_n219;
  t253_n218: return -0.1217944473028183f;
  t253_n219: if (E.getNumReferences() >= 3344576000 /*55854.0*/) goto t253_n225;
  t253_n220: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t253_n224;
  t253_n221: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t253_n223;
  t253_n222: return 0.1304665058851242f;
  t253_n223: return 0.0828707292675972f;
  t253_n224: return -0.12864944338798523f;
  t253_n225: if (E.getNumReferences() >= 3345018368 /*57582.0*/) goto t253_n227;
  t253_n226: return -0.14304348826408386f;
  t253_n227: if (E.getNumReferences() >= 3352510912 /*108163.5*/) goto t253_n245;
  t253_n228: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t253_n238;
  t253_n229: if (E.getNumReferences() >= 3350315456 /*91011.5*/) goto t253_n237;
  t253_n230: if (E.getNumReferences() >= 3345512576 /*59512.5*/) goto t253_n232;
  t253_n231: return 0.09202045947313309f;
  t253_n232: if (E.getNumReferences() >= 3349646912 /*85788.5*/) goto t253_n234;
  t253_n233: return -0.04750330373644829f;
  t253_n234: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t253_n236;
  t253_n235: return 0.016895493492484093f;
  t253_n236: return 0.11574497818946838f;
  t253_n237: return -0.12751677632331848f;
  t253_n238: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t253_n242;
  t253_n239: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t253_n241;
  t253_n240: return 0.03408088907599449f;
  t253_n241: return 0.0946134403347969f;
  t253_n242: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t253_n244;
  t253_n243: return 0.10936596244573593f;
  t253_n244: return 0.03410328924655914f;
  t253_n245: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t253_n247;
  t253_n246: return -0.14772261679172516f;
  t253_n247: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t253_n253;
  t253_n248: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t253_n252;
  t253_n249: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t253_n251;
  t253_n250: return -0.035297639667987823f;
  t253_n251: return 0.08800239861011505f;
  t253_n252: return 0.06929202377796173f;
  t253_n253: return 0.040630489587783813f;
  t253_n254: return 0.0006156276795081794f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree254(const Example& E) {
  t254_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t254_n2;
  t254_n1: return -0.00015043704479467124f;
  t254_n2: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t254_n248;
  t254_n3: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t254_n101;
  t254_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n6;
  t254_n5: return -0.01768609695136547f;
  t254_n6: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t254_n74;
  t254_n7: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t254_n47;
  t254_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n14;
  t254_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t254_n13;
  t254_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t254_n12;
  t254_n11: return -0.0017074269708245993f;
  t254_n12: return -0.030829496681690216f;
  t254_n13: return 0.01829686388373375f;
  t254_n14: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t254_n22;
  t254_n15: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n17;
  t254_n16: return -0.00596721051260829f;
  t254_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t254_n21;
  t254_n18: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t254_n20;
  t254_n19: return 0.015217396430671215f;
  t254_n20: return 0.3103807866573334f;
  t254_n21: return 0.06352661550045013f;
  t254_n22: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t254_n40;
  t254_n23: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t254_n25;
  t254_n24: return 0.14880667626857758f;
  t254_n25: if (E.getSymbolScopeDistanceCost() >= 3242196992 /*12.0*/) goto t254_n39;
  t254_n26: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t254_n38;
  t254_n27: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t254_n29;
  t254_n28: return 0.025219637900590897f;
  t254_n29: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t254_n33;
  t254_n30: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t254_n32;
  t254_n31: return 0.04792246222496033f;
  t254_n32: return 0.10965732485055923f;
  t254_n33: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t254_n37;
  t254_n34: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t254_n36;
  t254_n35: return 0.12737378478050232f;
  t254_n36: return 0.009990357793867588f;
  t254_n37: return 0.12189282476902008f;
  t254_n38: return -0.08122441172599792f;
  t254_n39: return -0.04426044970750809f;
  t254_n40: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t254_n46;
  t254_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n45;
  t254_n42: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t254_n44;
  t254_n43: return -0.19915008544921875f;
  t254_n44: return 0.0009191466378979385f;
  t254_n45: return -0.00014146666217129678f;
  t254_n46: return -0.8320100903511047f;
  t254_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n53;
  t254_n48: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t254_n50;
  t254_n49: return -0.011335772462189198f;
  t254_n50: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t254_n52;
  t254_n51: return -0.04997266083955765f;
  t254_n52: return 0.05158375948667526f;
  t254_n53: if (E.getNumReferences() >= 3282206720 /*325.0*/) goto t254_n73;
  t254_n54: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t254_n58;
  t254_n55: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n57;
  t254_n56: return -0.06461445987224579f;
  t254_n57: return 0.019326256588101387f;
  t254_n58: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t254_n68;
  t254_n59: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n67;
  t254_n60: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t254_n66;
  t254_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t254_n63;
  t254_n62: return -0.13305503129959106f;
  t254_n63: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t254_n65;
  t254_n64: return 0.07934162765741348f;
  t254_n65: return 0.06201232969760895f;
  t254_n66: return 0.03261213004589081f;
  t254_n67: return 0.01117696613073349f;
  t254_n68: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t254_n70;
  t254_n69: return 0.011209148913621902f;
  t254_n70: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t254_n72;
  t254_n71: return -0.276410311460495f;
  t254_n72: return -0.010254988446831703f;
  t254_n73: return 0.10447166860103607f;
  t254_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t254_n100;
  t254_n75: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t254_n95;
  t254_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n82;
  t254_n77: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t254_n79;
  t254_n78: return -0.014080526307225227f;
  t254_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t254_n81;
  t254_n80: return -0.13138704001903534f;
  t254_n81: return 0.1260291486978531f;
  t254_n82: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t254_n94;
  t254_n83: if (E.getFractionNameInContext() >= 3197124868 /*0.2817460298538208*/) goto t254_n93;
  t254_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n90;
  t254_n85: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t254_n87;
  t254_n86: return -0.07451748847961426f;
  t254_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t254_n89;
  t254_n88: return -0.07994464039802551f;
  t254_n89: return 0.10959983617067337f;
  t254_n90: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t254_n92;
  t254_n91: return -0.006793798878788948f;
  t254_n92: return -0.12774251401424408f;
  t254_n93: return 0.005101995076984167f;
  t254_n94: return 0.04624377563595772f;
  t254_n95: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t254_n97;
  t254_n96: return 0.10781387984752655f;
  t254_n97: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t254_n99;
  t254_n98: return 0.1215500682592392f;
  t254_n99: return 0.14612190425395966f;
  t254_n100: return 0.015315339900553226f;
  t254_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t254_n223;
  t254_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t254_n172;
  t254_n103: if (E.getNumReferences() >= 3284533248 /*396.0*/) goto t254_n171;
  t254_n104: if (E.getNumReferences() >= 3284107264 /*383.0*/) goto t254_n168;
  t254_n105: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t254_n111;
  t254_n106: if (E.getNumReferences() >= 3278045184 /*227.0*/) goto t254_n108;
  t254_n107: return 0.004889004863798618f;
  t254_n108: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t254_n110;
  t254_n109: return 0.09681551158428192f;
  t254_n110: return 0.12620535492897034f;
  t254_n111: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t254_n155;
  t254_n112: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t254_n140;
  t254_n113: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t254_n115;
  t254_n114: return -0.006742340046912432f;
  t254_n115: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t254_n129;
  t254_n116: if (E.getNumReferences() >= 3273588736 /*159.0*/) goto t254_n128;
  t254_n117: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t254_n127;
  t254_n118: if (E.getNumReferences() >= 3250847744 /*24.5*/) goto t254_n126;
  t254_n119: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t254_n123;
  t254_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n122;
  t254_n121: return -0.428786039352417f;
  t254_n122: return -0.025682376697659492f;
  t254_n123: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t254_n125;
  t254_n124: return 0.19423364102840424f;
  t254_n125: return 0.03591807186603546f;
  t254_n126: return -0.1765713095664978f;
  t254_n127: return 0.050375182181596756f;
  t254_n128: return 0.15852268040180206f;
  t254_n129: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t254_n137;
  t254_n130: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t254_n136;
  t254_n131: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n133;
  t254_n132: return 0.0543585829436779f;
  t254_n133: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t254_n135;
  t254_n134: return 0.04088028892874718f;
  t254_n135: return -0.21902331709861755f;
  t254_n136: return 0.13284257054328918f;
  t254_n137: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t254_n139;
  t254_n138: return 0.04736846685409546f;
  t254_n139: return -0.3433738350868225f;
  t254_n140: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t254_n148;
  t254_n141: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n147;
  t254_n142: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t254_n146;
  t254_n143: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t254_n145;
  t254_n144: return -0.05108823627233505f;
  t254_n145: return -0.0041258311830461025f;
  t254_n146: return 0.04812847077846527f;
  t254_n147: return -0.004208981525152922f;
  t254_n148: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t254_n150;
  t254_n149: return 0.01040854025632143f;
  t254_n150: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t254_n152;
  t254_n151: return 0.07998667657375336f;
  t254_n152: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t254_n154;
  t254_n153: return -0.10066086798906326f;
  t254_n154: return 0.15716373920440674f;
  t254_n155: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t254_n161;
  t254_n156: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t254_n160;
  t254_n157: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t254_n159;
  t254_n158: return -0.004547253251075745f;
  t254_n159: return 0.11332742124795914f;
  t254_n160: return 0.06463377177715302f;
  t254_n161: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t254_n163;
  t254_n162: return -0.06200917065143585f;
  t254_n163: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t254_n167;
  t254_n164: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t254_n166;
  t254_n165: return -0.24224631488323212f;
  t254_n166: return 0.18934477865695953f;
  t254_n167: return -0.010915043763816357f;
  t254_n168: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t254_n170;
  t254_n169: return -0.26215848326683044f;
  t254_n170: return -0.004817558452486992f;
  t254_n171: return 0.013202695176005363f;
  t254_n172: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t254_n178;
  t254_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t254_n177;
  t254_n174: if (E.getNumReferences() >= 3233808384 /*6.0*/) goto t254_n176;
  t254_n175: return -0.21011514961719513f;
  t254_n176: return 0.09901975840330124f;
  t254_n177: return -0.2931327223777771f;
  t254_n178: if (E.getNumReferences() >= 3248226304 /*19.5*/) goto t254_n180;
  t254_n179: return 0.00937922578305006f;
  t254_n180: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t254_n184;
  t254_n181: if (E.getNumReferences() >= 3262906368 /*63.0*/) goto t254_n183;
  t254_n182: return 0.13894611597061157f;
  t254_n183: return 0.1013699471950531f;
  t254_n184: if (E.getNumReferences() >= 3266314240 /*88.0*/) goto t254_n222;
  t254_n185: if (E.getNumReferences() >= 3264806912 /*76.5*/) goto t254_n219;
  t254_n186: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t254_n198;
  t254_n187: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t254_n189;
  t254_n188: return -0.14095382392406464f;
  t254_n189: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t254_n191;
  t254_n190: return -0.03114437870681286f;
  t254_n191: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t254_n193;
  t254_n192: return 0.008131735026836395f;
  t254_n193: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t254_n195;
  t254_n194: return 0.139256551861763f;
  t254_n195: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t254_n197;
  t254_n196: return -0.02856401726603508f;
  t254_n197: return 0.10650280863046646f;
  t254_n198: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t254_n208;
  t254_n199: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t254_n205;
  t254_n200: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t254_n202;
  t254_n201: return 0.14336547255516052f;
  t254_n202: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t254_n204;
  t254_n203: return 0.20812183618545532f;
  t254_n204: return -0.12021069973707199f;
  t254_n205: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t254_n207;
  t254_n206: return 0.12997108697891235f;
  t254_n207: return -0.03667774051427841f;
  t254_n208: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t254_n218;
  t254_n209: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t254_n217;
  t254_n210: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t254_n214;
  t254_n211: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t254_n213;
  t254_n212: return 0.0847141370177269f;
  t254_n213: return -0.055715739727020264f;
  t254_n214: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t254_n216;
  t254_n215: return -0.014370478689670563f;
  t254_n216: return 0.13017207384109497f;
  t254_n217: return -0.011417328380048275f;
  t254_n218: return 0.21159453690052032f;
  t254_n219: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t254_n221;
  t254_n220: return 0.0595666840672493f;
  t254_n221: return 0.12257537990808487f;
  t254_n222: return 0.008503410965204239f;
  t254_n223: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t254_n245;
  t254_n224: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t254_n226;
  t254_n225: return 0.1487712413072586f;
  t254_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t254_n244;
  t254_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t254_n237;
  t254_n228: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t254_n234;
  t254_n229: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t254_n233;
  t254_n230: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t254_n232;
  t254_n231: return -0.11269066482782364f;
  t254_n232: return -0.20680804550647736f;
  t254_n233: return -0.024060439318418503f;
  t254_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t254_n236;
  t254_n235: return -0.029359914362430573f;
  t254_n236: return 0.027845485135912895f;
  t254_n237: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t254_n239;
  t254_n238: return 0.009974646382033825f;
  t254_n239: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t254_n241;
  t254_n240: return -0.017233097925782204f;
  t254_n241: if (E.getFractionNameInContext() >= 3211638492 /*0.9285714626312256*/) goto t254_n243;
  t254_n242: return 0.10701294243335724f;
  t254_n243: return 0.03862724453210831f;
  t254_n244: return 0.010891723446547985f;
  t254_n245: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t254_n247;
  t254_n246: return -0.022647840902209282f;
  t254_n247: return 0.0045783366076648235f;
  t254_n248: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t254_n254;
  t254_n249: if (E.getFractionNameInContext() >= 3184668802 /*0.1026315838098526*/) goto t254_n253;
  t254_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t254_n252;
  t254_n251: return -0.19512809813022614f;
  t254_n252: return 0.11743872612714767f;
  t254_n253: return -0.00495831947773695f;
  t254_n254: return 0.0013085210230201483f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree255(const Example& E) {
  t255_n0: if (E.getNumReferences() >= 3302551552 /*1735.5*/) goto t255_n22;
  t255_n1: if (E.getNumReferences() >= 3302199296 /*1692.5*/) goto t255_n3;
  t255_n2: return -0.00011076776718255132f;
  t255_n3: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t255_n15;
  t255_n4: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t255_n8;
  t255_n5: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t255_n7;
  t255_n6: return -0.13474318385124207f;
  t255_n7: return -0.039620209485292435f;
  t255_n8: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t255_n10;
  t255_n9: return 0.12408814579248428f;
  t255_n10: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t255_n14;
  t255_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t255_n13;
  t255_n12: return -0.14104978740215302f;
  t255_n13: return 0.46288424730300903f;
  t255_n14: return -0.04748566076159477f;
  t255_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t255_n21;
  t255_n16: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t255_n18;
  t255_n17: return -0.5101847052574158f;
  t255_n18: if (E.getNumReferences() >= 3302440960 /*1722.0*/) goto t255_n20;
  t255_n19: return -0.0982460007071495f;
  t255_n20: return -1.2210355997085571f;
  t255_n21: return -0.13979212939739227f;
  t255_n22: if (E.getNumReferences() >= 3302559744 /*1736.5*/) goto t255_n30;
  t255_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t255_n29;
  t255_n24: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t255_n26;
  t255_n25: return -0.31455281376838684f;
  t255_n26: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t255_n28;
  t255_n27: return 0.1903800070285797f;
  t255_n28: return 0.016395092010498047f;
  t255_n29: return 0.13363032042980194f;
  t255_n30: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t255_n36;
  t255_n31: if (E.getNumReferences() >= 3302662144 /*1749.0*/) goto t255_n35;
  t255_n32: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t255_n34;
  t255_n33: return -0.11641919612884521f;
  t255_n34: return 0.027483217418193817f;
  t255_n35: return 0.006579865701496601f;
  t255_n36: if (E.getNumReferences() >= 3302584320 /*1739.5*/) goto t255_n42;
  t255_n37: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t255_n41;
  t255_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t255_n40;
  t255_n39: return 0.0028157236520200968f;
  t255_n40: return 0.17991560697555542f;
  t255_n41: return -0.0995328426361084f;
  t255_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t255_n58;
  t255_n43: if (E.getNumReferences() >= 3302768640 /*1762.0*/) goto t255_n53;
  t255_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t255_n46;
  t255_n45: return -0.13681426644325256f;
  t255_n46: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t255_n52;
  t255_n47: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t255_n49;
  t255_n48: return 0.0003398528788238764f;
  t255_n49: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t255_n51;
  t255_n50: return 0.13383498787879944f;
  t255_n51: return 0.1276366114616394f;
  t255_n52: return -0.018417418003082275f;
  t255_n53: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t255_n55;
  t255_n54: return -0.013467930257320404f;
  t255_n55: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t255_n57;
  t255_n56: return 0.019282560795545578f;
  t255_n57: return -0.06683828681707382f;
  t255_n58: if (E.getNumReferences() >= 3318849536 /*6708.0*/) goto t255_n254;
  t255_n59: if (E.getNumReferences() >= 3318439936 /*6508.0*/) goto t255_n223;
  t255_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t255_n62;
  t255_n61: return 8.269980025943369e-05f;
  t255_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t255_n64;
  t255_n63: return -0.037322353571653366f;
  t255_n64: if (E.getNumReferences() >= 3317870592 /*6230.0*/) goto t255_n222;
  t255_n65: if (E.getNumReferences() >= 3317611520 /*6103.5*/) goto t255_n199;
  t255_n66: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t255_n196;
  t255_n67: if (E.getNumReferences() >= 3317043200 /*5826.0*/) goto t255_n195;
  t255_n68: if (E.getNumReferences() >= 3316915200 /*5763.5*/) goto t255_n186;
  t255_n69: if (E.getNumReferences() >= 3316045824 /*5339.0*/) goto t255_n185;
  t255_n70: if (E.getNumReferences() >= 3315954688 /*5294.5*/) goto t255_n178;
  t255_n71: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t255_n77;
  t255_n72: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t255_n76;
  t255_n73: if (E.getNumReferences() >= 3307511808 /*2634.0*/) goto t255_n75;
  t255_n74: return 0.12726083397865295f;
  t255_n75: return -0.10852296650409698f;
  t255_n76: return -0.02759283408522606f;
  t255_n77: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t255_n175;
  t255_n78: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t255_n128;
  t255_n79: if (E.getNumReferences() >= 3312912384 /*3952.5*/) goto t255_n125;
  t255_n80: if (E.getNumReferences() >= 3312891904 /*3947.5*/) goto t255_n124;
  t255_n81: if (E.getNumReferences() >= 3310979072 /*3480.5*/) goto t255_n115;
  t255_n82: if (E.getNumReferences() >= 3307874304 /*2722.5*/) goto t255_n114;
  t255_n83: if (E.getNumReferences() >= 3307841536 /*2714.5*/) goto t255_n113;
  t255_n84: if (E.getNumReferences() >= 3305695232 /*2190.5*/) goto t255_n86;
  t255_n85: return 0.012641890905797482f;
  t255_n86: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t255_n96;
  t255_n87: if (E.getNumReferences() >= 3306899456 /*2484.5*/) goto t255_n95;
  t255_n88: if (E.getNumReferences() >= 3306403840 /*2363.5*/) goto t255_n92;
  t255_n89: if (E.getNumReferences() >= 3306297344 /*2337.5*/) goto t255_n91;
  t255_n90: return 0.07847504317760468f;
  t255_n91: return -0.1108454018831253f;
  t255_n92: if (E.getNumReferences() >= 3306416128 /*2366.5*/) goto t255_n94;
  t255_n93: return 0.14572782814502716f;
  t255_n94: return 0.08261040598154068f;
  t255_n95: return -0.03391047194600105f;
  t255_n96: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t255_n98;
  t255_n97: return 0.038910865783691406f;
  t255_n98: if (E.getNumReferences() >= 3307112448 /*2536.5*/) goto t255_n112;
  t255_n99: if (E.getNumReferences() >= 3306958848 /*2499.0*/) goto t255_n111;
  t255_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t255_n102;
  t255_n101: return -0.10102817416191101f;
  t255_n102: if (E.getNumReferences() >= 3306240000 /*2323.5*/) goto t255_n106;
  t255_n103: if (E.getNumReferences() >= 3305916416 /*2244.5*/) goto t255_n105;
  t255_n104: return 0.14196011424064636f;
  t255_n105: return 0.12198911607265472f;
  t255_n106: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t255_n108;
  t255_n107: return 0.13981637358665466f;
  t255_n108: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t255_n110;
  t255_n109: return -0.10334575176239014f;
  t255_n110: return 0.1035681962966919f;
  t255_n111: return 0.11900884658098221f;
  t255_n112: return 0.05332908779382706f;
  t255_n113: return 0.2211056500673294f;
  t255_n114: return -0.010739848017692566f;
  t255_n115: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t255_n121;
  t255_n116: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t255_n118;
  t255_n117: return 0.005005742888897657f;
  t255_n118: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t255_n120;
  t255_n119: return 0.11874406784772873f;
  t255_n120: return 0.15006370842456818f;
  t255_n121: if (E.getNumReferences() >= 3311271936 /*3552.0*/) goto t255_n123;
  t255_n122: return 0.10998837649822235f;
  t255_n123: return -0.009904558770358562f;
  t255_n124: return 0.11323720961809158f;
  t255_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t255_n127;
  t255_n126: return -0.24087324738502502f;
  t255_n127: return 0.061093371361494064f;
  t255_n128: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t255_n130;
  t255_n129: return 0.14905089139938354f;
  t255_n130: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t255_n142;
  t255_n131: if (E.getNumReferences() >= 3313894400 /*4288.5*/) goto t255_n137;
  t255_n132: if (E.getNumReferences() >= 3313764352 /*4225.0*/) goto t255_n134;
  t255_n133: return 0.19417019188404083f;
  t255_n134: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t255_n136;
  t255_n135: return 0.1457972675561905f;
  t255_n136: return 0.12426846474409103f;
  t255_n137: if (E.getNumReferences() >= 3314725888 /*4694.5*/) goto t255_n141;
  t255_n138: if (E.getNumReferences() >= 3314442240 /*4556.0*/) goto t255_n140;
  t255_n139: return -0.05749588832259178f;
  t255_n140: return 0.1108761802315712f;
  t255_n141: return -0.03319510072469711f;
  t255_n142: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t255_n174;
  t255_n143: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t255_n173;
  t255_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t255_n146;
  t255_n145: return -0.027734726667404175f;
  t255_n146: if (E.getNumReferences() >= 3315533824 /*5089.0*/) goto t255_n172;
  t255_n147: if (E.getNumReferences() >= 3314929664 /*4794.0*/) goto t255_n167;
  t255_n148: if (E.getNumReferences() >= 3313659904 /*4174.0*/) goto t255_n152;
  t255_n149: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t255_n151;
  t255_n150: return 0.10239937901496887f;
  t255_n151: return 0.138527974486351f;
  t255_n152: if (E.getNumReferences() >= 3314120704 /*4399.0*/) goto t255_n154;
  t255_n153: return -0.15153071284294128f;
  t255_n154: if (E.getNumReferences() >= 3314141184 /*4409.0*/) goto t255_n156;
  t255_n155: return 0.15604616701602936f;
  t255_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t255_n158;
  t255_n157: return -0.060408033430576324f;
  t255_n158: if (E.getNumReferences() >= 3314375680 /*4523.5*/) goto t255_n160;
  t255_n159: return -0.13510867953300476f;
  t255_n160: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t255_n162;
  t255_n161: return 0.09270070493221283f;
  t255_n162: if (E.getNumReferences() >= 3314592768 /*4629.5*/) goto t255_n166;
  t255_n163: if (E.getNumReferences() >= 3314535424 /*4601.5*/) goto t255_n165;
  t255_n164: return 0.12293767184019089f;
  t255_n165: return 0.12829342484474182f;
  t255_n166: return 0.06763248890638351f;
  t255_n167: if (E.getNumReferences() >= 3315025920 /*4841.0*/) goto t255_n169;
  t255_n168: return 0.1326487958431244f;
  t255_n169: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t255_n171;
  t255_n170: return 0.08417579531669617f;
  t255_n171: return 0.947649359703064f;
  t255_n172: return -0.17996609210968018f;
  t255_n173: return 0.012718046084046364f;
  t255_n174: return -0.06837083399295807f;
  t255_n175: if (E.getNumReferences() >= 3315025920 /*4841.0*/) goto t255_n177;
  t255_n176: return -0.07918781787157059f;
  t255_n177: return -0.46755093336105347f;
  t255_n178: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t255_n184;
  t255_n179: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t255_n181;
  t255_n180: return 0.07882405817508698f;
  t255_n181: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t255_n183;
  t255_n182: return 0.10503843426704407f;
  t255_n183: return 0.12315382063388824f;
  t255_n184: return -0.10063910484313965f;
  t255_n185: return -0.03156474232673645f;
  t255_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t255_n190;
  t255_n187: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t255_n189;
  t255_n188: return 0.9052746295928955f;
  t255_n189: return 0.08224225044250488f;
  t255_n190: if (E.getNumReferences() >= 3316937728 /*5774.5*/) goto t255_n194;
  t255_n191: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t255_n193;
  t255_n192: return 0.22371451556682587f;
  t255_n193: return 0.13586242496967316f;
  t255_n194: return -0.10175739228725433f;
  t255_n195: return -0.047760382294654846f;
  t255_n196: if (E.getNumReferences() >= 3304009728 /*1913.5*/) goto t255_n198;
  t255_n197: return 0.1509207934141159f;
  t255_n198: return 0.0646524652838707f;
  t255_n199: if (E.getNumReferences() >= 3317616640 /*6106.0*/) goto t255_n217;
  t255_n200: if (E.getFractionNameInContext() >= 3182307328 /*0.0850372314453125*/) goto t255_n214;
  t255_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t255_n203;
  t255_n202: return -0.10389361530542374f;
  t255_n203: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t255_n213;
  t255_n204: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t255_n212;
  t255_n205: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t255_n209;
  t255_n206: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t255_n208;
  t255_n207: return -0.10362881422042847f;
  t255_n208: return 0.3627302348613739f;
  t255_n209: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t255_n211;
  t255_n210: return -0.10643711686134338f;
  t255_n211: return 0.1368415504693985f;
  t255_n212: return 0.26792091131210327f;
  t255_n213: return -0.1203993558883667f;
  t255_n214: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t255_n216;
  t255_n215: return 0.1155787855386734f;
  t255_n216: return 0.14260882139205933f;
  t255_n217: if (E.getFractionNameInContext() >= 3183942770 /*0.0972222238779068*/) goto t255_n219;
  t255_n218: return -0.03205994516611099f;
  t255_n219: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t255_n221;
  t255_n220: return 0.1346864253282547f;
  t255_n221: return -0.10371968895196915f;
  t255_n222: return -0.05411802604794502f;
  t255_n223: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t255_n247;
  t255_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t255_n230;
  t255_n225: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t255_n227;
  t255_n226: return -0.09373102337121964f;
  t255_n227: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t255_n229;
  t255_n228: return 0.9905678033828735f;
  t255_n229: return -0.12009607255458832f;
  t255_n230: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t255_n242;
  t255_n231: if (E.getNumReferences() >= 3318784000 /*6676.0*/) goto t255_n241;
  t255_n232: if (E.getNumReferences() >= 3318735872 /*6652.5*/) goto t255_n234;
  t255_n233: return 0.06350115686655045f;
  t255_n234: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t255_n240;
  t255_n235: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t255_n237;
  t255_n236: return 0.0026428618002682924f;
  t255_n237: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t255_n239;
  t255_n238: return 0.1285155862569809f;
  t255_n239: return -0.1582229882478714f;
  t255_n240: return 0.15263082087039948f;
  t255_n241: return -0.01992574706673622f;
  t255_n242: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t255_n246;
  t255_n243: if (E.getNumReferences() >= 3318611968 /*6592.0*/) goto t255_n245;
  t255_n244: return -0.10012417286634445f;
  t255_n245: return 0.25576451420783997f;
  t255_n246: return -0.10038573294878006f;
  t255_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t255_n251;
  t255_n248: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t255_n250;
  t255_n249: return 0.06093024089932442f;
  t255_n250: return 0.11064426600933075f;
  t255_n251: if (E.getNumReferences() >= 3318637568 /*6604.5*/) goto t255_n253;
  t255_n252: return 0.11272071301937103f;
  t255_n253: return 0.11062289774417877f;
  t255_n254: return 0.00012328811862971634f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree256(const Example& E) {
  t256_n0: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t256_n254;
  t256_n1: if (E.getNumReferences() >= 3265200128 /*79.5*/) goto t256_n241;
  t256_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n240;
  t256_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t256_n9;
  t256_n4: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t256_n6;
  t256_n5: return 0.03451596200466156f;
  t256_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n8;
  t256_n7: return -0.012315655127167702f;
  t256_n8: return 0.009091842919588089f;
  t256_n9: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t256_n229;
  t256_n10: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t256_n82;
  t256_n11: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t256_n81;
  t256_n12: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t256_n68;
  t256_n13: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t256_n47;
  t256_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n36;
  t256_n15: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t256_n27;
  t256_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n18;
  t256_n17: return -0.006416900083422661f;
  t256_n18: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t256_n22;
  t256_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n21;
  t256_n20: return -0.4484379291534424f;
  t256_n21: return 0.008348691277205944f;
  t256_n22: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t256_n26;
  t256_n23: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t256_n25;
  t256_n24: return 0.08359416574239731f;
  t256_n25: return 0.047029487788677216f;
  t256_n26: return 0.030073290690779686f;
  t256_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n29;
  t256_n28: return -0.0940265879034996f;
  t256_n29: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t256_n33;
  t256_n30: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t256_n32;
  t256_n31: return -0.06245467811822891f;
  t256_n32: return -0.010028764605522156f;
  t256_n33: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t256_n35;
  t256_n34: return -0.31586751341819763f;
  t256_n35: return 0.14264212548732758f;
  t256_n36: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t256_n46;
  t256_n37: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t256_n45;
  t256_n38: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t256_n40;
  t256_n39: return -0.005062886513769627f;
  t256_n40: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t256_n42;
  t256_n41: return 0.12203548103570938f;
  t256_n42: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t256_n44;
  t256_n43: return 0.02500372752547264f;
  t256_n44: return -0.0042070504277944565f;
  t256_n45: return -0.11278403550386429f;
  t256_n46: return 0.009789817966520786f;
  t256_n47: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t256_n49;
  t256_n48: return -0.0033556295093148947f;
  t256_n49: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t256_n67;
  t256_n50: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t256_n60;
  t256_n51: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t256_n53;
  t256_n52: return 0.04476054385304451f;
  t256_n53: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t256_n57;
  t256_n54: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t256_n56;
  t256_n55: return 0.060029324144124985f;
  t256_n56: return 0.13909314572811127f;
  t256_n57: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t256_n59;
  t256_n58: return 0.34511059522628784f;
  t256_n59: return 0.1313377171754837f;
  t256_n60: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t256_n62;
  t256_n61: return 0.0602974109351635f;
  t256_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n66;
  t256_n63: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t256_n65;
  t256_n64: return -0.12959615886211395f;
  t256_n65: return -1.9616578817367554f;
  t256_n66: return 0.0521966889500618f;
  t256_n67: return 0.006824877113103867f;
  t256_n68: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t256_n70;
  t256_n69: return -0.015989357605576515f;
  t256_n70: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t256_n78;
  t256_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t256_n75;
  t256_n72: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t256_n74;
  t256_n73: return 0.011512034572660923f;
  t256_n74: return 0.10330333560705185f;
  t256_n75: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t256_n77;
  t256_n76: return 0.09691859781742096f;
  t256_n77: return 0.16623452305793762f;
  t256_n78: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t256_n80;
  t256_n79: return -0.010800772346556187f;
  t256_n80: return -0.16559509932994843f;
  t256_n81: return -0.09390370547771454f;
  t256_n82: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t256_n144;
  t256_n83: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t256_n127;
  t256_n84: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t256_n112;
  t256_n85: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t256_n111;
  t256_n86: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t256_n96;
  t256_n87: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t256_n89;
  t256_n88: return 0.012001915834844112f;
  t256_n89: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t256_n95;
  t256_n90: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t256_n94;
  t256_n91: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t256_n93;
  t256_n92: return 0.18190321326255798f;
  t256_n93: return 0.05880139768123627f;
  t256_n94: return 0.02512104995548725f;
  t256_n95: return -0.05575799569487572f;
  t256_n96: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t256_n104;
  t256_n97: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t256_n99;
  t256_n98: return -0.004578745923936367f;
  t256_n99: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t256_n101;
  t256_n100: return -0.03580096736550331f;
  t256_n101: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t256_n103;
  t256_n102: return -0.17881989479064941f;
  t256_n103: return -0.06750784814357758f;
  t256_n104: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t256_n110;
  t256_n105: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t256_n107;
  t256_n106: return 0.08403728157281876f;
  t256_n107: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t256_n109;
  t256_n108: return -0.13062702119350433f;
  t256_n109: return 0.01488092914223671f;
  t256_n110: return 0.08532016724348068f;
  t256_n111: return -0.008188139647245407f;
  t256_n112: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t256_n126;
  t256_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t256_n119;
  t256_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n116;
  t256_n115: return -0.0034874831326305866f;
  t256_n116: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t256_n118;
  t256_n117: return 0.021456852555274963f;
  t256_n118: return -0.1714538186788559f;
  t256_n119: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n125;
  t256_n120: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t256_n122;
  t256_n121: return -0.03270149603486061f;
  t256_n122: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t256_n124;
  t256_n123: return -0.06677839159965515f;
  t256_n124: return 0.12586145102977753f;
  t256_n125: return 0.027066044509410858f;
  t256_n126: return 0.0021393171045929193f;
  t256_n127: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t256_n133;
  t256_n128: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n132;
  t256_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t256_n131;
  t256_n130: return -0.020033160224556923f;
  t256_n131: return 0.1071116030216217f;
  t256_n132: return 0.0411117859184742f;
  t256_n133: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t256_n143;
  t256_n134: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t256_n138;
  t256_n135: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t256_n137;
  t256_n136: return 0.07487130165100098f;
  t256_n137: return 0.00694099348038435f;
  t256_n138: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t256_n142;
  t256_n139: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t256_n141;
  t256_n140: return 0.11165471374988556f;
  t256_n141: return 0.059221722185611725f;
  t256_n142: return 0.0015161177143454552f;
  t256_n143: return -0.0239335335791111f;
  t256_n144: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t256_n174;
  t256_n145: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t256_n151;
  t256_n146: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t256_n148;
  t256_n147: return 0.0008587961783632636f;
  t256_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n150;
  t256_n149: return -0.06581904739141464f;
  t256_n150: return -0.00547413993626833f;
  t256_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t256_n173;
  t256_n152: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t256_n162;
  t256_n153: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t256_n161;
  t256_n154: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t256_n158;
  t256_n155: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t256_n157;
  t256_n156: return -0.05282102897763252f;
  t256_n157: return 0.04855336993932724f;
  t256_n158: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t256_n160;
  t256_n159: return 0.24533304572105408f;
  t256_n160: return 0.029796335846185684f;
  t256_n161: return 0.011397041380405426f;
  t256_n162: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t256_n172;
  t256_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t256_n169;
  t256_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n166;
  t256_n165: return -0.09214441478252411f;
  t256_n166: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t256_n168;
  t256_n167: return 0.07519262284040451f;
  t256_n168: return -0.10107722878456116f;
  t256_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t256_n171;
  t256_n170: return 0.038359202444553375f;
  t256_n171: return 0.09693747758865356f;
  t256_n172: return 0.11167633533477783f;
  t256_n173: return 0.1375259906053543f;
  t256_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n228;
  t256_n175: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t256_n177;
  t256_n176: return 0.00012213621812406927f;
  t256_n177: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t256_n213;
  t256_n178: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t256_n198;
  t256_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n195;
  t256_n180: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t256_n192;
  t256_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n189;
  t256_n182: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t256_n188;
  t256_n183: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t256_n187;
  t256_n184: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t256_n186;
  t256_n185: return -0.2642790973186493f;
  t256_n186: return 0.08150769025087357f;
  t256_n187: return -1.1137136220932007f;
  t256_n188: return 0.0658303052186966f;
  t256_n189: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t256_n191;
  t256_n190: return -0.41259753704071045f;
  t256_n191: return 0.014349951408803463f;
  t256_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t256_n194;
  t256_n193: return -0.04681152105331421f;
  t256_n194: return -0.007942584343254566f;
  t256_n195: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t256_n197;
  t256_n196: return 0.00010057917825179175f;
  t256_n197: return 0.014040899462997913f;
  t256_n198: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t256_n210;
  t256_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n201;
  t256_n200: return -0.021916547790169716f;
  t256_n201: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t256_n203;
  t256_n202: return -0.008468158543109894f;
  t256_n203: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t256_n207;
  t256_n204: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t256_n206;
  t256_n205: return -0.15035122632980347f;
  t256_n206: return 0.022366071119904518f;
  t256_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n209;
  t256_n208: return -0.30032095313072205f;
  t256_n209: return 0.0855996385216713f;
  t256_n210: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t256_n212;
  t256_n211: return 0.014844010584056377f;
  t256_n212: return 0.10359665751457214f;
  t256_n213: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t256_n215;
  t256_n214: return 0.005654971580952406f;
  t256_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n225;
  t256_n216: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t256_n222;
  t256_n217: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t256_n221;
  t256_n218: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t256_n220;
  t256_n219: return -0.08368518948554993f;
  t256_n220: return -0.25764477252960205f;
  t256_n221: return 0.013856029137969017f;
  t256_n222: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t256_n224;
  t256_n223: return -0.8938156962394714f;
  t256_n224: return -0.15079312026500702f;
  t256_n225: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t256_n227;
  t256_n226: return 0.11659891158342361f;
  t256_n227: return -0.2601226270198822f;
  t256_n228: return 0.0025486373342573643f;
  t256_n229: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t256_n239;
  t256_n230: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t256_n238;
  t256_n231: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t256_n233;
  t256_n232: return -0.008814084343612194f;
  t256_n233: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t256_n235;
  t256_n234: return 0.0186917744576931f;
  t256_n235: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t256_n237;
  t256_n236: return 0.045287854969501495f;
  t256_n237: return 0.09093855321407318f;
  t256_n238: return 0.0012802921701222658f;
  t256_n239: return -0.014184869825839996f;
  t256_n240: return 0.0015640847850590944f;
  t256_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t256_n243;
  t256_n242: return 0.005488248076289892f;
  t256_n243: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t256_n253;
  t256_n244: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t256_n250;
  t256_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t256_n247;
  t256_n246: return -0.08169574290513992f;
  t256_n247: if (E.getNumReferences() >= 3265462272 /*81.5*/) goto t256_n249;
  t256_n248: return -0.14054016768932343f;
  t256_n249: return 0.2848409414291382f;
  t256_n250: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t256_n252;
  t256_n251: return 0.08738625049591064f;
  t256_n252: return 0.1246897354722023f;
  t256_n253: return 0.018084779381752014f;
  t256_n254: return -0.0006235274486243725f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree257(const Example& E) {
  t257_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t257_n2;
  t257_n1: return -0.09089943021535873f;
  t257_n2: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t257_n254;
  t257_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t257_n7;
  t257_n4: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t257_n6;
  t257_n5: return 0.0003460862790234387f;
  t257_n6: return 0.006915526930242777f;
  t257_n7: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t257_n253;
  t257_n8: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t257_n12;
  t257_n9: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t257_n11;
  t257_n10: return -0.0005618362338282168f;
  t257_n11: return -0.12901118397712708f;
  t257_n12: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t257_n128;
  t257_n13: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t257_n123;
  t257_n14: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t257_n16;
  t257_n15: return -0.0003105397627223283f;
  t257_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t257_n28;
  t257_n17: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t257_n27;
  t257_n18: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t257_n24;
  t257_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t257_n21;
  t257_n20: return -0.07154642045497894f;
  t257_n21: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t257_n23;
  t257_n22: return 0.09931419044733047f;
  t257_n23: return -0.03991975635290146f;
  t257_n24: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t257_n26;
  t257_n25: return -0.08699189126491547f;
  t257_n26: return -1.0710848569869995f;
  t257_n27: return 0.05662418529391289f;
  t257_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t257_n70;
  t257_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t257_n33;
  t257_n30: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t257_n32;
  t257_n31: return -0.0028424658812582493f;
  t257_n32: return 0.004259520210325718f;
  t257_n33: if (E.getNumReferences() >= 3326554624 /*12748.5*/) goto t257_n65;
  t257_n34: if (E.getNumReferences() >= 3326458880 /*12655.0*/) goto t257_n56;
  t257_n35: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t257_n47;
  t257_n36: if (E.getNumReferences() >= 3319966720 /*7253.5*/) goto t257_n44;
  t257_n37: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t257_n39;
  t257_n38: return 0.004940217360854149f;
  t257_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t257_n41;
  t257_n40: return 0.005417935084551573f;
  t257_n41: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t257_n43;
  t257_n42: return 0.06029690429568291f;
  t257_n43: return -0.06623204052448273f;
  t257_n44: if (E.getNumReferences() >= 3320098816 /*7318.0*/) goto t257_n46;
  t257_n45: return 0.0981278344988823f;
  t257_n46: return 0.04935721307992935f;
  t257_n47: if (E.getNumReferences() >= 3293102080 /*803.0*/) goto t257_n55;
  t257_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t257_n50;
  t257_n49: return 0.040790826082229614f;
  t257_n50: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n52;
  t257_n51: return 0.10635156184434891f;
  t257_n52: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t257_n54;
  t257_n53: return -0.012675746344029903f;
  t257_n54: return 0.1161511242389679f;
  t257_n55: return 0.11073093116283417f;
  t257_n56: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t257_n64;
  t257_n57: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t257_n59;
  t257_n58: return 0.10026057064533234f;
  t257_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t257_n61;
  t257_n60: return 0.11310593038797379f;
  t257_n61: if (E.getFileProximityDistanceCost() >= 3247439872 /*18.0*/) goto t257_n63;
  t257_n62: return 0.12947750091552734f;
  t257_n63: return 0.18738430738449097f;
  t257_n64: return 0.064731664955616f;
  t257_n65: if (E.getNumReferences() >= 3346600320 /*63761.5*/) goto t257_n67;
  t257_n66: return -0.033183466643095016f;
  t257_n67: if (E.getNumReferences() >= 3368655888 /*412896.5*/) goto t257_n69;
  t257_n68: return 0.030687058344483376f;
  t257_n69: return -0.007675651926547289f;
  t257_n70: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t257_n110;
  t257_n71: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t257_n109;
  t257_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n100;
  t257_n73: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t257_n89;
  t257_n74: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t257_n78;
  t257_n75: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t257_n77;
  t257_n76: return -1.0112886428833008f;
  t257_n77: return 0.04923814535140991f;
  t257_n78: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t257_n84;
  t257_n79: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t257_n81;
  t257_n80: return -0.027467096224427223f;
  t257_n81: if (E.getNumReferences() >= 3327426048 /*13599.5*/) goto t257_n83;
  t257_n82: return -0.27067306637763977f;
  t257_n83: return 0.08135409653186798f;
  t257_n84: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n88;
  t257_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t257_n87;
  t257_n86: return 0.21705137193202972f;
  t257_n87: return 0.06534970551729202f;
  t257_n88: return -0.4141807556152344f;
  t257_n89: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t257_n99;
  t257_n90: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t257_n92;
  t257_n91: return 0.07848496735095978f;
  t257_n92: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t257_n98;
  t257_n93: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t257_n97;
  t257_n94: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t257_n96;
  t257_n95: return 0.7206336259841919f;
  t257_n96: return 0.11218754947185516f;
  t257_n97: return 0.24263393878936768f;
  t257_n98: return 0.10267425328493118f;
  t257_n99: return 0.020868362858891487f;
  t257_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n102;
  t257_n101: return 9.3999900855124e-05f;
  t257_n102: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t257_n108;
  t257_n103: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t257_n105;
  t257_n104: return -0.000568569463212043f;
  t257_n105: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t257_n107;
  t257_n106: return 0.03164307400584221f;
  t257_n107: return -0.007629256695508957f;
  t257_n108: return -7.69749385653995e-05f;
  t257_n109: return 0.022784795612096786f;
  t257_n110: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t257_n114;
  t257_n111: if (E.getFractionNameInContext() >= 3187804193 /*0.1269841343164444*/) goto t257_n113;
  t257_n112: return -0.035529881715774536f;
  t257_n113: return 0.6522120833396912f;
  t257_n114: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t257_n122;
  t257_n115: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t257_n117;
  t257_n116: return 0.06846142560243607f;
  t257_n117: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t257_n119;
  t257_n118: return 0.0073613631539046764f;
  t257_n119: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t257_n121;
  t257_n120: return -0.10278372466564178f;
  t257_n121: return 0.3137863874435425f;
  t257_n122: return -0.05418337881565094f;
  t257_n123: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t257_n127;
  t257_n124: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n126;
  t257_n125: return 0.007028700318187475f;
  t257_n126: return -0.07915341854095459f;
  t257_n127: return -0.0659526139497757f;
  t257_n128: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t257_n192;
  t257_n129: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t257_n181;
  t257_n130: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n180;
  t257_n131: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t257_n165;
  t257_n132: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t257_n162;
  t257_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t257_n147;
  t257_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n140;
  t257_n135: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t257_n137;
  t257_n136: return 0.014319192618131638f;
  t257_n137: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t257_n139;
  t257_n138: return 0.044357091188430786f;
  t257_n139: return 0.10646111518144608f;
  t257_n140: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t257_n146;
  t257_n141: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t257_n143;
  t257_n142: return 0.04682553932070732f;
  t257_n143: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t257_n145;
  t257_n144: return 0.12302924692630768f;
  t257_n145: return 0.08646604418754578f;
  t257_n146: return -0.06279534101486206f;
  t257_n147: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t257_n161;
  t257_n148: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n160;
  t257_n149: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t257_n153;
  t257_n150: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t257_n152;
  t257_n151: return 0.08169177174568176f;
  t257_n152: return 0.10518968850374222f;
  t257_n153: if (E.getFractionNameInContext() >= 3201152732 /*0.4017857313156128*/) goto t257_n157;
  t257_n154: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t257_n156;
  t257_n155: return 0.07378587126731873f;
  t257_n156: return 0.12716901302337646f;
  t257_n157: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t257_n159;
  t257_n158: return -0.14204268157482147f;
  t257_n159: return 0.10157617181539536f;
  t257_n160: return 0.05086611583828926f;
  t257_n161: return -0.16230861842632294f;
  t257_n162: if (E.getFractionNameInContext() >= 3202287554 /*0.43560606241226196*/) goto t257_n164;
  t257_n163: return 0.1270308792591095f;
  t257_n164: return 0.0825694277882576f;
  t257_n165: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n179;
  t257_n166: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t257_n174;
  t257_n167: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t257_n173;
  t257_n168: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t257_n172;
  t257_n169: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t257_n171;
  t257_n170: return 0.1201171725988388f;
  t257_n171: return -0.004086512140929699f;
  t257_n172: return 0.0034390916116535664f;
  t257_n173: return -0.45073845982551575f;
  t257_n174: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t257_n176;
  t257_n175: return -0.10590849816799164f;
  t257_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t257_n178;
  t257_n177: return 0.09461646527051926f;
  t257_n178: return 0.11630784720182419f;
  t257_n179: return -0.08948409557342529f;
  t257_n180: return -0.011631263419985771f;
  t257_n181: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t257_n191;
  t257_n182: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n190;
  t257_n183: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t257_n187;
  t257_n184: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t257_n186;
  t257_n185: return 0.007223767694085836f;
  t257_n186: return 0.08394119888544083f;
  t257_n187: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t257_n189;
  t257_n188: return 0.06678856164216995f;
  t257_n189: return 0.12304951995611191f;
  t257_n190: return -0.027739493176341057f;
  t257_n191: return -0.006453422363847494f;
  t257_n192: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t257_n198;
  t257_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t257_n197;
  t257_n194: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t257_n196;
  t257_n195: return -0.22701358795166016f;
  t257_n196: return 0.009011110290884972f;
  t257_n197: return -0.006128065288066864f;
  t257_n198: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t257_n252;
  t257_n199: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t257_n251;
  t257_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t257_n202;
  t257_n201: return 0.010290322825312614f;
  t257_n202: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t257_n238;
  t257_n203: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t257_n235;
  t257_n204: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t257_n216;
  t257_n205: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t257_n209;
  t257_n206: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t257_n208;
  t257_n207: return 0.12588703632354736f;
  t257_n208: return -0.03943401947617531f;
  t257_n209: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t257_n215;
  t257_n210: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t257_n214;
  t257_n211: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n213;
  t257_n212: return 0.10719364881515503f;
  t257_n213: return 0.026252098381519318f;
  t257_n214: return 0.07693447917699814f;
  t257_n215: return 0.031707439571619034f;
  t257_n216: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t257_n218;
  t257_n217: return -0.17533645033836365f;
  t257_n218: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n220;
  t257_n219: return 0.08540928363800049f;
  t257_n220: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t257_n222;
  t257_n221: return -0.2330666333436966f;
  t257_n222: if (E.getFractionNameInContext() >= 3198506326 /*0.3229166865348816*/) goto t257_n224;
  t257_n223: return -0.02193569950759411f;
  t257_n224: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t257_n228;
  t257_n225: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t257_n227;
  t257_n226: return -0.0423135943710804f;
  t257_n227: return 0.03943074494600296f;
  t257_n228: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t257_n230;
  t257_n229: return -0.19276553392410278f;
  t257_n230: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t257_n234;
  t257_n231: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t257_n233;
  t257_n232: return 0.09055398404598236f;
  t257_n233: return 0.044457364827394485f;
  t257_n234: return -0.1161186546087265f;
  t257_n235: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t257_n237;
  t257_n236: return 0.04267524555325508f;
  t257_n237: return 0.13780231773853302f;
  t257_n238: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t257_n244;
  t257_n239: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t257_n241;
  t257_n240: return 0.031981974840164185f;
  t257_n241: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t257_n243;
  t257_n242: return 0.13451321423053741f;
  t257_n243: return 0.06975522637367249f;
  t257_n244: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t257_n250;
  t257_n245: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t257_n247;
  t257_n246: return -0.01846211776137352f;
  t257_n247: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t257_n249;
  t257_n248: return 0.0757262110710144f;
  t257_n249: return -0.22107619047164917f;
  t257_n250: return 0.042584698647260666f;
  t257_n251: return 0.0030031870119273663f;
  t257_n252: return -0.01591220684349537f;
  t257_n253: return -0.05312570929527283f;
  t257_n254: return -0.041716594249010086f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree258(const Example& E) {
  t258_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t258_n2;
  t258_n1: return -0.00033020341652445495f;
  t258_n2: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t258_n170;
  t258_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t258_n107;
  t258_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t258_n24;
  t258_n5: if (E.getNumReferences() >= 3324623872 /*10863.0*/) goto t258_n7;
  t258_n6: return -0.013381152413785458f;
  t258_n7: if (E.getNumReferences() >= 3331047168 /*17887.5*/) goto t258_n23;
  t258_n8: if (E.getNumReferences() >= 3329239040 /*15370.0*/) goto t258_n16;
  t258_n9: if (E.getNumReferences() >= 3326229504 /*12431.0*/) goto t258_n15;
  t258_n10: if (E.getNumReferences() >= 3325309440 /*11532.5*/) goto t258_n14;
  t258_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n13;
  t258_n12: return -0.12618644535541534f;
  t258_n13: return 0.17979411780834198f;
  t258_n14: return 0.13134680688381195f;
  t258_n15: return -0.11539562046527863f;
  t258_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n18;
  t258_n17: return -0.11165100336074829f;
  t258_n18: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t258_n22;
  t258_n19: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t258_n21;
  t258_n20: return 0.13522617518901825f;
  t258_n21: return -0.13221649825572968f;
  t258_n22: return 0.13032813370227814f;
  t258_n23: return -0.1027209609746933f;
  t258_n24: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t258_n48;
  t258_n25: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t258_n35;
  t258_n26: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t258_n28;
  t258_n27: return 0.003007500898092985f;
  t258_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t258_n30;
  t258_n29: return -0.009161504916846752f;
  t258_n30: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t258_n32;
  t258_n31: return 0.04244643449783325f;
  t258_n32: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t258_n34;
  t258_n33: return -0.008139806799590588f;
  t258_n34: return 0.128583624958992f;
  t258_n35: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t258_n47;
  t258_n36: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t258_n38;
  t258_n37: return 0.02713767997920513f;
  t258_n38: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t258_n46;
  t258_n39: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t258_n45;
  t258_n40: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t258_n42;
  t258_n41: return -0.024575117975473404f;
  t258_n42: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t258_n44;
  t258_n43: return -0.15239304304122925f;
  t258_n44: return -0.04335467517375946f;
  t258_n45: return -0.10085643082857132f;
  t258_n46: return 0.006299011874943972f;
  t258_n47: return -0.0038442297372967005f;
  t258_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n50;
  t258_n49: return -0.00542040541768074f;
  t258_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t258_n52;
  t258_n51: return -0.00347530422732234f;
  t258_n52: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t258_n100;
  t258_n53: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t258_n95;
  t258_n54: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t258_n82;
  t258_n55: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t258_n57;
  t258_n56: return -0.023742806166410446f;
  t258_n57: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t258_n79;
  t258_n58: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t258_n66;
  t258_n59: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t258_n65;
  t258_n60: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t258_n64;
  t258_n61: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t258_n63;
  t258_n62: return 0.00881929136812687f;
  t258_n63: return -0.007614071946591139f;
  t258_n64: return 0.014339649118483067f;
  t258_n65: return -0.004240578040480614f;
  t258_n66: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t258_n70;
  t258_n67: if (E.getFractionNameInContext() >= 3169906929 /*0.029411764815449715*/) goto t258_n69;
  t258_n68: return 0.013613100163638592f;
  t258_n69: return -0.0028923412319272757f;
  t258_n70: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t258_n72;
  t258_n71: return 0.0021028893534094095f;
  t258_n72: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t258_n74;
  t258_n73: return 0.012602034024894238f;
  t258_n74: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t258_n78;
  t258_n75: if (E.getNumReferences() >= 3317173248 /*5889.5*/) goto t258_n77;
  t258_n76: return 0.1127319484949112f;
  t258_n77: return -0.04272134229540825f;
  t258_n78: return 0.03392729535698891f;
  t258_n79: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t258_n81;
  t258_n80: return -0.002247788244858384f;
  t258_n81: return 0.0658484473824501f;
  t258_n82: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t258_n92;
  t258_n83: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t258_n89;
  t258_n84: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t258_n86;
  t258_n85: return 0.028621409088373184f;
  t258_n86: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t258_n88;
  t258_n87: return 0.13468365371227264f;
  t258_n88: return 0.08360118418931961f;
  t258_n89: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t258_n91;
  t258_n90: return -0.009641232900321484f;
  t258_n91: return 0.03178038075566292f;
  t258_n92: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t258_n94;
  t258_n93: return -0.07215136289596558f;
  t258_n94: return 0.00698439497500658f;
  t258_n95: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t258_n97;
  t258_n96: return 0.020597543567419052f;
  t258_n97: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t258_n99;
  t258_n98: return -0.0203646719455719f;
  t258_n99: return -0.22970545291900635f;
  t258_n100: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t258_n104;
  t258_n101: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t258_n103;
  t258_n102: return -0.0005819007637910545f;
  t258_n103: return 0.08861159533262253f;
  t258_n104: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t258_n106;
  t258_n105: return -0.005079749505966902f;
  t258_n106: return -0.44938454031944275f;
  t258_n107: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t258_n169;
  t258_n108: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n132;
  t258_n109: if (E.getNumReferences() >= 3297095680 /*1069.5*/) goto t258_n129;
  t258_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t258_n128;
  t258_n111: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t258_n127;
  t258_n112: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t258_n122;
  t258_n113: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t258_n121;
  t258_n114: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t258_n120;
  t258_n115: if (E.getNumReferences() >= 3271950336 /*134.0*/) goto t258_n119;
  t258_n116: if (E.getNumReferences() >= 3268280320 /*103.0*/) goto t258_n118;
  t258_n117: return -0.03611159324645996f;
  t258_n118: return 0.10112950205802917f;
  t258_n119: return -0.34752485156059265f;
  t258_n120: return 0.008850467391312122f;
  t258_n121: return -0.08814448118209839f;
  t258_n122: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t258_n126;
  t258_n123: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t258_n125;
  t258_n124: return 0.06044718995690346f;
  t258_n125: return 0.2096593976020813f;
  t258_n126: return 0.006289428099989891f;
  t258_n127: return -0.17305883765220642f;
  t258_n128: return -0.008198805153369904f;
  t258_n129: if (E.getNumReferences() >= 3312803840 /*3926.0*/) goto t258_n131;
  t258_n130: return 0.08930227160453796f;
  t258_n131: return 0.0013428315287455916f;
  t258_n132: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t258_n134;
  t258_n133: return 0.005821967963129282f;
  t258_n134: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t258_n136;
  t258_n135: return 0.021837513893842697f;
  t258_n136: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t258_n162;
  t258_n137: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t258_n139;
  t258_n138: return 0.004780021961778402f;
  t258_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n159;
  t258_n140: if (E.getNumReferences() >= 3255042048 /*33.0*/) goto t258_n156;
  t258_n141: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t258_n147;
  t258_n142: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t258_n146;
  t258_n143: if (E.getNumReferences() >= 3240099840 /*10.0*/) goto t258_n145;
  t258_n144: return -0.11871728301048279f;
  t258_n145: return -0.20683163404464722f;
  t258_n146: return 0.127532958984375f;
  t258_n147: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t258_n153;
  t258_n148: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t258_n152;
  t258_n149: if (E.getFractionNameInContext() >= 3209960770 /*0.8285714387893677*/) goto t258_n151;
  t258_n150: return -0.10228125005960464f;
  t258_n151: return 0.07712467014789581f;
  t258_n152: return 0.12967905402183533f;
  t258_n153: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t258_n155;
  t258_n154: return 0.11798577755689621f;
  t258_n155: return -0.04724838212132454f;
  t258_n156: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t258_n158;
  t258_n157: return 0.1397639364004135f;
  t258_n158: return -0.09350723028182983f;
  t258_n159: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t258_n161;
  t258_n160: return 0.016771923750638962f;
  t258_n161: return 0.1297631412744522f;
  t258_n162: if (E.getNumReferences() >= 3225419776 /*3.0*/) goto t258_n168;
  t258_n163: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t258_n167;
  t258_n164: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t258_n166;
  t258_n165: return 0.010782798752188683f;
  t258_n166: return 0.10635218769311905f;
  t258_n167: return 0.037638742476701736f;
  t258_n168: return -0.17382030189037323f;
  t258_n169: return 0.0006402921862900257f;
  t258_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t258_n172;
  t258_n171: return 0.0006064096814952791f;
  t258_n172: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t258_n254;
  t258_n173: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t258_n195;
  t258_n174: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t258_n178;
  t258_n175: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t258_n177;
  t258_n176: return 0.00584126403555274f;
  t258_n177: return 0.24594800174236298f;
  t258_n178: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t258_n186;
  t258_n179: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t258_n183;
  t258_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t258_n182;
  t258_n181: return 0.04730523005127907f;
  t258_n182: return 0.11853129416704178f;
  t258_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n185;
  t258_n184: return -0.0060585238970816135f;
  t258_n185: return 0.031650375574827194f;
  t258_n186: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t258_n192;
  t258_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t258_n189;
  t258_n188: return 0.05917420983314514f;
  t258_n189: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t258_n191;
  t258_n190: return 0.07327045500278473f;
  t258_n191: return 0.1087176576256752f;
  t258_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n194;
  t258_n193: return -0.18928998708724976f;
  t258_n194: return -0.012261517345905304f;
  t258_n195: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t258_n249;
  t258_n196: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t258_n244;
  t258_n197: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t258_n207;
  t258_n198: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t258_n200;
  t258_n199: return -0.014221674762666225f;
  t258_n200: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t258_n206;
  t258_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n203;
  t258_n202: return -0.005841538775712252f;
  t258_n203: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t258_n205;
  t258_n204: return 0.04845726862549782f;
  t258_n205: return 0.007029456086456776f;
  t258_n206: return -0.028517944738268852f;
  t258_n207: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t258_n233;
  t258_n208: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t258_n210;
  t258_n209: return 0.010759647004306316f;
  t258_n210: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t258_n212;
  t258_n211: return -0.1586807668209076f;
  t258_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t258_n220;
  t258_n213: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t258_n215;
  t258_n214: return -0.01880170963704586f;
  t258_n215: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t258_n219;
  t258_n216: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t258_n218;
  t258_n217: return 0.004212627653032541f;
  t258_n218: return 0.05317150056362152f;
  t258_n219: return -0.16215643286705017f;
  t258_n220: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t258_n222;
  t258_n221: return 0.02903778851032257f;
  t258_n222: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t258_n230;
  t258_n223: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t258_n229;
  t258_n224: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t258_n226;
  t258_n225: return -0.00014879308582749218f;
  t258_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t258_n228;
  t258_n227: return 0.026809746399521828f;
  t258_n228: return 0.06262243539094925f;
  t258_n229: return -0.09627501666545868f;
  t258_n230: if (E.getFractionNameInContext() >= 3206445544 /*0.6190476417541504*/) goto t258_n232;
  t258_n231: return 0.022931858897209167f;
  t258_n232: return 0.06243975833058357f;
  t258_n233: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t258_n241;
  t258_n234: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t258_n240;
  t258_n235: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t258_n237;
  t258_n236: return 0.009987925179302692f;
  t258_n237: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t258_n239;
  t258_n238: return 0.07584115862846375f;
  t258_n239: return 0.030545109882950783f;
  t258_n240: return -0.16676628589630127f;
  t258_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t258_n243;
  t258_n242: return 0.04632548615336418f;
  t258_n243: return 0.13487093150615692f;
  t258_n244: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t258_n246;
  t258_n245: return -0.07213376462459564f;
  t258_n246: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t258_n248;
  t258_n247: return 0.1124921515583992f;
  t258_n248: return -0.24900391697883606f;
  t258_n249: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t258_n251;
  t258_n250: return -0.02217288687825203f;
  t258_n251: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t258_n253;
  t258_n252: return -0.0590139701962471f;
  t258_n253: return 0.0760764479637146f;
  t258_n254: return 0.01333978958427906f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree259(const Example& E) {
  t259_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t259_n198;
  t259_n1: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t259_n5;
  t259_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n4;
  t259_n3: return -0.05825084447860718f;
  t259_n4: return 0.04863428324460983f;
  t259_n5: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t259_n151;
  t259_n6: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t259_n8;
  t259_n7: return -0.0003310550528112799f;
  t259_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n10;
  t259_n9: return -0.01695038564503193f;
  t259_n10: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t259_n76;
  t259_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t259_n71;
  t259_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t259_n14;
  t259_n13: return -0.006401639897376299f;
  t259_n14: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t259_n56;
  t259_n15: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t259_n53;
  t259_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t259_n50;
  t259_n17: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t259_n23;
  t259_n18: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t259_n20;
  t259_n19: return -0.01555691473186016f;
  t259_n20: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t259_n22;
  t259_n21: return -0.06302788853645325f;
  t259_n22: return -0.4001947045326233f;
  t259_n23: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t259_n47;
  t259_n24: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t259_n28;
  t259_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t259_n27;
  t259_n26: return -0.10106255859136581f;
  t259_n27: return 0.10720852762460709f;
  t259_n28: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t259_n34;
  t259_n29: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t259_n31;
  t259_n30: return -0.010670403949916363f;
  t259_n31: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t259_n33;
  t259_n32: return 0.10452122241258621f;
  t259_n33: return 0.01939292997121811f;
  t259_n34: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t259_n46;
  t259_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t259_n39;
  t259_n36: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t259_n38;
  t259_n37: return -0.002242338377982378f;
  t259_n38: return 0.06885360181331635f;
  t259_n39: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t259_n45;
  t259_n40: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t259_n42;
  t259_n41: return 0.07094305008649826f;
  t259_n42: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t259_n44;
  t259_n43: return -0.008407507091760635f;
  t259_n44: return 0.09883609414100647f;
  t259_n45: return 0.016423417255282402f;
  t259_n46: return 0.029560478404164314f;
  t259_n47: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t259_n49;
  t259_n48: return -0.5430244207382202f;
  t259_n49: return -0.029292931780219078f;
  t259_n50: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t259_n52;
  t259_n51: return 0.08140189945697784f;
  t259_n52: return 0.006503264419734478f;
  t259_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n55;
  t259_n54: return -0.0929211750626564f;
  t259_n55: return 0.08412367850542068f;
  t259_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t259_n58;
  t259_n57: return -0.1013248935341835f;
  t259_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n60;
  t259_n59: return -0.0007405687938444316f;
  t259_n60: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t259_n70;
  t259_n61: if (E.getFractionNameInContext() >= 3178807750 /*0.0607316717505455*/) goto t259_n63;
  t259_n62: return 0.06660228967666626f;
  t259_n63: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t259_n65;
  t259_n64: return 0.06804630905389786f;
  t259_n65: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t259_n67;
  t259_n66: return 0.24003562331199646f;
  t259_n67: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t259_n69;
  t259_n68: return 0.031051507219672203f;
  t259_n69: return 0.6670101284980774f;
  t259_n70: return 0.02910751663148403f;
  t259_n71: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t259_n73;
  t259_n72: return -0.001490830210968852f;
  t259_n73: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t259_n75;
  t259_n74: return 0.06930581480264664f;
  t259_n75: return 0.017296496778726578f;
  t259_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n150;
  t259_n77: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t259_n143;
  t259_n78: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t259_n138;
  t259_n79: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t259_n121;
  t259_n80: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t259_n120;
  t259_n81: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t259_n119;
  t259_n82: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t259_n112;
  t259_n83: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t259_n101;
  t259_n84: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t259_n98;
  t259_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t259_n97;
  t259_n86: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t259_n94;
  t259_n87: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t259_n91;
  t259_n88: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t259_n90;
  t259_n89: return 0.1702299565076828f;
  t259_n90: return 0.08137832581996918f;
  t259_n91: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t259_n93;
  t259_n92: return -0.3145747482776642f;
  t259_n93: return 0.15173588693141937f;
  t259_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n96;
  t259_n95: return -0.11096924543380737f;
  t259_n96: return 0.07074019312858582f;
  t259_n97: return -0.01977641135454178f;
  t259_n98: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t259_n100;
  t259_n99: return 0.3133951425552368f;
  t259_n100: return -0.0008885439019650221f;
  t259_n101: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t259_n103;
  t259_n102: return 0.13284431397914886f;
  t259_n103: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t259_n109;
  t259_n104: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t259_n106;
  t259_n105: return -0.23983146250247955f;
  t259_n106: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t259_n108;
  t259_n107: return 0.003331075655296445f;
  t259_n108: return 0.16294749081134796f;
  t259_n109: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t259_n111;
  t259_n110: return 0.11610047519207001f;
  t259_n111: return 0.006968471221625805f;
  t259_n112: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t259_n114;
  t259_n113: return -0.026384197175502777f;
  t259_n114: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t259_n118;
  t259_n115: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t259_n117;
  t259_n116: return 0.09463092684745789f;
  t259_n117: return 0.1805531084537506f;
  t259_n118: return -0.05154518783092499f;
  t259_n119: return -0.15293173491954803f;
  t259_n120: return -0.025616999715566635f;
  t259_n121: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t259_n133;
  t259_n122: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t259_n128;
  t259_n123: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t259_n125;
  t259_n124: return -0.03965676948428154f;
  t259_n125: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t259_n127;
  t259_n126: return 0.0968761295080185f;
  t259_n127: return 0.028812166303396225f;
  t259_n128: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t259_n132;
  t259_n129: if (E.getFractionNameInContext() >= 3206445544 /*0.6190476417541504*/) goto t259_n131;
  t259_n130: return 0.04847332462668419f;
  t259_n131: return 0.12027788907289505f;
  t259_n132: return -0.5433186292648315f;
  t259_n133: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t259_n137;
  t259_n134: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t259_n136;
  t259_n135: return -0.32276251912117004f;
  t259_n136: return 0.1107073575258255f;
  t259_n137: return 0.14680156111717224f;
  t259_n138: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t259_n142;
  t259_n139: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t259_n141;
  t259_n140: return -0.18607909977436066f;
  t259_n141: return -0.03182249143719673f;
  t259_n142: return -0.04303227737545967f;
  t259_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t259_n145;
  t259_n144: return 0.0029026144184172153f;
  t259_n145: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t259_n149;
  t259_n146: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t259_n148;
  t259_n147: return 0.01396418921649456f;
  t259_n148: return 0.0020349284168332815f;
  t259_n149: return -0.0179850272834301f;
  t259_n150: return 0.008412652648985386f;
  t259_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Type))) goto t259_n169;
  t259_n152: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t259_n154;
  t259_n153: return -0.036470819264650345f;
  t259_n154: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t259_n164;
  t259_n155: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t259_n157;
  t259_n156: return 0.014105632901191711f;
  t259_n157: if (E.getNumReferences() >= 3283959808 /*378.5*/) goto t259_n163;
  t259_n158: if (E.getNumReferences() >= 3279028224 /*242.0*/) goto t259_n162;
  t259_n159: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t259_n161;
  t259_n160: return 0.0608319416642189f;
  t259_n161: return 0.1607009321451187f;
  t259_n162: return 0.1710006594657898f;
  t259_n163: return -0.08872105181217194f;
  t259_n164: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t259_n168;
  t259_n165: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t259_n167;
  t259_n166: return 0.1017267182469368f;
  t259_n167: return 0.11839602142572403f;
  t259_n168: return 0.07657654583454132f;
  t259_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t259_n185;
  t259_n170: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t259_n180;
  t259_n171: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t259_n175;
  t259_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t259_n174;
  t259_n173: return -0.10256323963403702f;
  t259_n174: return 0.08071714639663696f;
  t259_n175: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t259_n177;
  t259_n176: return 0.019978290423750877f;
  t259_n177: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t259_n179;
  t259_n178: return -0.1432504951953888f;
  t259_n179: return -0.1305059790611267f;
  t259_n180: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t259_n184;
  t259_n181: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t259_n183;
  t259_n182: return 0.03442057967185974f;
  t259_n183: return 0.11777611821889877f;
  t259_n184: return -0.0375942625105381f;
  t259_n185: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t259_n197;
  t259_n186: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t259_n192;
  t259_n187: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t259_n189;
  t259_n188: return -0.09700004011392593f;
  t259_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t259_n191;
  t259_n190: return 0.05721056088805199f;
  t259_n191: return 0.17794419825077057f;
  t259_n192: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t259_n196;
  t259_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n195;
  t259_n194: return -0.15422675013542175f;
  t259_n195: return -0.12600840628147125f;
  t259_n196: return -0.09569961577653885f;
  t259_n197: return 0.04620152339339256f;
  t259_n198: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t259_n202;
  t259_n199: if (E.getFileProximityDistanceCost() >= 3212836864 /*1.0*/) goto t259_n201;
  t259_n200: return 0.10607238858938217f;
  t259_n201: return 0.0004336679703556001f;
  t259_n202: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t259_n242;
  t259_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t259_n235;
  t259_n204: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t259_n232;
  t259_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n229;
  t259_n206: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t259_n224;
  t259_n207: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t259_n217;
  t259_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t259_n210;
  t259_n209: return -0.047167517244815826f;
  t259_n210: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t259_n214;
  t259_n211: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t259_n213;
  t259_n212: return -0.008299360983073711f;
  t259_n213: return 0.08704271167516708f;
  t259_n214: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t259_n216;
  t259_n215: return -0.027996741235256195f;
  t259_n216: return 0.06035498529672623f;
  t259_n217: if (E.getFractionNameInContext() >= 3196938455 /*0.27619048953056335*/) goto t259_n219;
  t259_n218: return -0.21393027901649475f;
  t259_n219: if (E.getFractionNameInContext() >= 3200482732 /*0.38181817531585693*/) goto t259_n223;
  t259_n220: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t259_n222;
  t259_n221: return -0.026563070714473724f;
  t259_n222: return 0.09475798159837723f;
  t259_n223: return -0.11934083700180054f;
  t259_n224: if (E.getFractionNameInContext() >= 3205962866 /*0.5902777910232544*/) goto t259_n226;
  t259_n225: return 0.004613263066858053f;
  t259_n226: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t259_n228;
  t259_n227: return 0.12792766094207764f;
  t259_n228: return 0.048695698380470276f;
  t259_n229: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t259_n231;
  t259_n230: return -0.009662264958024025f;
  t259_n231: return 0.01191246323287487f;
  t259_n232: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t259_n234;
  t259_n233: return 0.006845247931778431f;
  t259_n234: return -0.010770363733172417f;
  t259_n235: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t259_n239;
  t259_n236: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t259_n238;
  t259_n237: return 0.00735178217291832f;
  t259_n238: return 0.06016599014401436f;
  t259_n239: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t259_n241;
  t259_n240: return -0.001728787086904049f;
  t259_n241: return -0.049443136900663376f;
  t259_n242: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t259_n248;
  t259_n243: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t259_n247;
  t259_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t259_n246;
  t259_n245: return 0.0053690648637712f;
  t259_n246: return 0.05921519920229912f;
  t259_n247: return -0.05731617659330368f;
  t259_n248: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t259_n254;
  t259_n249: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t259_n253;
  t259_n250: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t259_n252;
  t259_n251: return 0.019235245883464813f;
  t259_n252: return -0.24213221669197083f;
  t259_n253: return 0.06866231560707092f;
  t259_n254: return -0.14122337102890015f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree260(const Example& E) {
  t260_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t260_n2;
  t260_n1: return -0.00015643158985767514f;
  t260_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t260_n242;
  t260_n3: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t260_n109;
  t260_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n8;
  t260_n5: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t260_n7;
  t260_n6: return -0.013941665180027485f;
  t260_n7: return -0.10370026528835297f;
  t260_n8: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t260_n64;
  t260_n9: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t260_n57;
  t260_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n40;
  t260_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t260_n35;
  t260_n12: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t260_n14;
  t260_n13: return -0.0009782543638721108f;
  t260_n14: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t260_n18;
  t260_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n17;
  t260_n16: return -0.054735779762268066f;
  t260_n17: return 0.0067720008082687855f;
  t260_n18: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t260_n30;
  t260_n19: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t260_n21;
  t260_n20: return 0.0016359920846298337f;
  t260_n21: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t260_n23;
  t260_n22: return 0.018424903973937035f;
  t260_n23: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t260_n27;
  t260_n24: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t260_n26;
  t260_n25: return 0.1046830415725708f;
  t260_n26: return -0.01460171490907669f;
  t260_n27: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t260_n29;
  t260_n28: return 0.11000493913888931f;
  t260_n29: return 0.14455507695674896f;
  t260_n30: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n32;
  t260_n31: return -0.03495439514517784f;
  t260_n32: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t260_n34;
  t260_n33: return -0.10165601968765259f;
  t260_n34: return -0.7600228190422058f;
  t260_n35: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n39;
  t260_n36: if (E.getSymbolScopeDistanceCost() >= 3238002688 /*8.0*/) goto t260_n38;
  t260_n37: return 0.033531997352838516f;
  t260_n38: return -0.11524420976638794f;
  t260_n39: return -0.015373732894659042f;
  t260_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t260_n42;
  t260_n41: return -0.022251373156905174f;
  t260_n42: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t260_n52;
  t260_n43: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t260_n45;
  t260_n44: return 0.011403867974877357f;
  t260_n45: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t260_n47;
  t260_n46: return 0.011920970864593983f;
  t260_n47: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t260_n49;
  t260_n48: return -0.11983273923397064f;
  t260_n49: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t260_n51;
  t260_n50: return 0.13178148865699768f;
  t260_n51: return -0.04964796453714371f;
  t260_n52: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t260_n56;
  t260_n53: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t260_n55;
  t260_n54: return 0.0344131663441658f;
  t260_n55: return -0.11964184790849686f;
  t260_n56: return -0.05318979173898697f;
  t260_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n61;
  t260_n58: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t260_n60;
  t260_n59: return -0.011263714171946049f;
  t260_n60: return -0.11009787768125534f;
  t260_n61: if (E.getNumReferences() >= 3282206720 /*325.0*/) goto t260_n63;
  t260_n62: return 0.007807421498000622f;
  t260_n63: return 0.10191763937473297f;
  t260_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n102;
  t260_n65: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t260_n95;
  t260_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t260_n88;
  t260_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t260_n69;
  t260_n68: return -0.008591155521571636f;
  t260_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t260_n87;
  t260_n70: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n72;
  t260_n71: return -0.0017105145379900932f;
  t260_n72: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t260_n84;
  t260_n73: if (E.getFractionNameInContext() >= 3211438762 /*0.9166666269302368*/) goto t260_n83;
  t260_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n76;
  t260_n75: return -0.019192645326256752f;
  t260_n76: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t260_n80;
  t260_n77: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t260_n79;
  t260_n78: return -0.0493890680372715f;
  t260_n79: return 0.06564965844154358f;
  t260_n80: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t260_n82;
  t260_n81: return 0.03464505076408386f;
  t260_n82: return -2.2526040077209473f;
  t260_n83: return 0.12789659202098846f;
  t260_n84: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t260_n86;
  t260_n85: return -0.15862329304218292f;
  t260_n86: return 0.10474996268749237f;
  t260_n87: return 0.08714766800403595f;
  t260_n88: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t260_n90;
  t260_n89: return 0.04732731729745865f;
  t260_n90: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t260_n94;
  t260_n91: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t260_n93;
  t260_n92: return -0.19306132197380066f;
  t260_n93: return -0.5625308752059937f;
  t260_n94: return -0.025117134675383568f;
  t260_n95: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t260_n97;
  t260_n96: return 0.11348898708820343f;
  t260_n97: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t260_n99;
  t260_n98: return 0.11994339525699615f;
  t260_n99: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t260_n101;
  t260_n100: return -0.030373718589544296f;
  t260_n101: return 0.17236238718032837f;
  t260_n102: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t260_n104;
  t260_n103: return 0.012208488769829273f;
  t260_n104: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t260_n108;
  t260_n105: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t260_n107;
  t260_n106: return -1.065373182296753f;
  t260_n107: return -0.1391877830028534f;
  t260_n108: return -0.01769859716296196f;
  t260_n109: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t260_n223;
  t260_n110: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t260_n114;
  t260_n111: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t260_n113;
  t260_n112: return 0.09525004774332047f;
  t260_n113: return 0.1435985118150711f;
  t260_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t260_n208;
  t260_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t260_n199;
  t260_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n144;
  t260_n117: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t260_n119;
  t260_n118: return -0.06019172817468643f;
  t260_n119: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t260_n135;
  t260_n120: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n128;
  t260_n121: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t260_n125;
  t260_n122: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t260_n124;
  t260_n123: return -0.13571949303150177f;
  t260_n124: return 0.0357222743332386f;
  t260_n125: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t260_n127;
  t260_n126: return 0.03664487600326538f;
  t260_n127: return -0.6397669315338135f;
  t260_n128: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t260_n132;
  t260_n129: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t260_n131;
  t260_n130: return 0.20550866425037384f;
  t260_n131: return 0.11894316226243973f;
  t260_n132: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t260_n134;
  t260_n133: return -0.12634973227977753f;
  t260_n134: return 0.09790155291557312f;
  t260_n135: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t260_n143;
  t260_n136: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t260_n142;
  t260_n137: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t260_n139;
  t260_n138: return -0.10529980063438416f;
  t260_n139: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t260_n141;
  t260_n140: return -0.11149302870035172f;
  t260_n141: return 0.03984522446990013f;
  t260_n142: return 0.000993890454992652f;
  t260_n143: return 0.07972359657287598f;
  t260_n144: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t260_n154;
  t260_n145: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t260_n147;
  t260_n146: return 0.006088665686547756f;
  t260_n147: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t260_n153;
  t260_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t260_n150;
  t260_n149: return -0.022619882598519325f;
  t260_n150: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t260_n152;
  t260_n151: return 0.1255931407213211f;
  t260_n152: return -0.15518823266029358f;
  t260_n153: return 0.021143361926078796f;
  t260_n154: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t260_n168;
  t260_n155: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t260_n167;
  t260_n156: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n160;
  t260_n157: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t260_n159;
  t260_n158: return -0.1972404569387436f;
  t260_n159: return 0.13781432807445526f;
  t260_n160: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t260_n162;
  t260_n161: return -0.022646315395832062f;
  t260_n162: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t260_n166;
  t260_n163: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t260_n165;
  t260_n164: return -0.08400948345661163f;
  t260_n165: return 0.08000044524669647f;
  t260_n166: return -0.22455960512161255f;
  t260_n167: return 0.008186894468963146f;
  t260_n168: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t260_n196;
  t260_n169: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t260_n187;
  t260_n170: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t260_n172;
  t260_n171: return 0.0017088082386180758f;
  t260_n172: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t260_n186;
  t260_n173: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t260_n175;
  t260_n174: return 0.035311006009578705f;
  t260_n175: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t260_n179;
  t260_n176: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t260_n178;
  t260_n177: return 0.012196850031614304f;
  t260_n178: return 0.31569933891296387f;
  t260_n179: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t260_n183;
  t260_n180: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t260_n182;
  t260_n181: return 0.2133762538433075f;
  t260_n182: return 0.0022437286097556353f;
  t260_n183: if (E.getNumReferences() >= 3221225472 /*2.0*/) goto t260_n185;
  t260_n184: return -0.215487539768219f;
  t260_n185: return 2.085143804550171f;
  t260_n186: return -0.1679169088602066f;
  t260_n187: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t260_n191;
  t260_n188: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t260_n190;
  t260_n189: return 0.11228173971176147f;
  t260_n190: return -0.11536779254674911f;
  t260_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n195;
  t260_n192: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t260_n194;
  t260_n193: return -0.5141041874885559f;
  t260_n194: return -0.17789269983768463f;
  t260_n195: return 0.018470043316483498f;
  t260_n196: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t260_n198;
  t260_n197: return -0.01585778407752514f;
  t260_n198: return 0.018517140299081802f;
  t260_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t260_n205;
  t260_n200: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t260_n204;
  t260_n201: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t260_n203;
  t260_n202: return -0.005324922036379576f;
  t260_n203: return -0.03387642651796341f;
  t260_n204: return 0.019699091091752052f;
  t260_n205: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t260_n207;
  t260_n206: return 0.011010621674358845f;
  t260_n207: return 0.0038183829747140408f;
  t260_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t260_n214;
  t260_n209: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t260_n213;
  t260_n210: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t260_n212;
  t260_n211: return -0.07536085695028305f;
  t260_n212: return 0.05195540562272072f;
  t260_n213: return 0.006042244378477335f;
  t260_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t260_n218;
  t260_n215: if (E.getNumReferences() >= 3305705472 /*2193.0*/) goto t260_n217;
  t260_n216: return 0.0244560819119215f;
  t260_n217: return -0.5864349603652954f;
  t260_n218: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t260_n222;
  t260_n219: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t260_n221;
  t260_n220: return 0.0932743102312088f;
  t260_n221: return -0.16742241382598877f;
  t260_n222: return 0.03173946961760521f;
  t260_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n241;
  t260_n224: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t260_n234;
  t260_n225: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t260_n231;
  t260_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n230;
  t260_n227: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t260_n229;
  t260_n228: return -0.7621148824691772f;
  t260_n229: return -0.1523962914943695f;
  t260_n230: return -0.02308301068842411f;
  t260_n231: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t260_n233;
  t260_n232: return -0.048641398549079895f;
  t260_n233: return -0.8568675518035889f;
  t260_n234: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t260_n240;
  t260_n235: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n239;
  t260_n236: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n238;
  t260_n237: return -0.020332539454102516f;
  t260_n238: return -0.17067183554172516f;
  t260_n239: return -0.0022035769652575254f;
  t260_n240: return -0.19498702883720398f;
  t260_n241: return 0.0007541659288108349f;
  t260_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t260_n246;
  t260_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t260_n245;
  t260_n244: return -0.08574992418289185f;
  t260_n245: return 0.05186716839671135f;
  t260_n246: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t260_n252;
  t260_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t260_n249;
  t260_n248: return 0.01478190440684557f;
  t260_n249: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t260_n251;
  t260_n250: return 0.01905595324933529f;
  t260_n251: return 0.045240577310323715f;
  t260_n252: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t260_n254;
  t260_n253: return -0.06960459798574448f;
  t260_n254: return -0.3610004484653473f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree261(const Example& E) {
  t261_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t261_n2;
  t261_n1: return -0.00035413887235336006f;
  t261_n2: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t261_n6;
  t261_n3: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t261_n5;
  t261_n4: return -0.03055974282324314f;
  t261_n5: return 0.1565963327884674f;
  t261_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t261_n254;
  t261_n7: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t261_n59;
  t261_n8: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t261_n58;
  t261_n9: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t261_n11;
  t261_n10: return -0.002796327928081155f;
  t261_n11: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t261_n17;
  t261_n12: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t261_n14;
  t261_n13: return -0.0016710031777620316f;
  t261_n14: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t261_n16;
  t261_n15: return 0.10512302070856094f;
  t261_n16: return 0.006738922093063593f;
  t261_n17: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t261_n19;
  t261_n18: return 0.10813067108392715f;
  t261_n19: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t261_n27;
  t261_n20: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t261_n24;
  t261_n21: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t261_n23;
  t261_n22: return 0.0222668144851923f;
  t261_n23: return -0.06221186742186546f;
  t261_n24: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t261_n26;
  t261_n25: return -0.0036100978031754494f;
  t261_n26: return 0.01702365279197693f;
  t261_n27: if (E.getFileProximityDistanceCost() >= 3254255616 /*31.0*/) goto t261_n31;
  t261_n28: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t261_n30;
  t261_n29: return 0.01475533377379179f;
  t261_n30: return 0.14826929569244385f;
  t261_n31: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t261_n51;
  t261_n32: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t261_n34;
  t261_n33: return -0.8708022832870483f;
  t261_n34: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t261_n42;
  t261_n35: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t261_n39;
  t261_n36: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t261_n38;
  t261_n37: return 0.00940156914293766f;
  t261_n38: return -0.13678909838199615f;
  t261_n39: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t261_n41;
  t261_n40: return 0.03324282541871071f;
  t261_n41: return -0.20140354335308075f;
  t261_n42: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t261_n44;
  t261_n43: return 0.014261558651924133f;
  t261_n44: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t261_n50;
  t261_n45: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t261_n47;
  t261_n46: return 0.072830930352211f;
  t261_n47: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t261_n49;
  t261_n48: return 0.014961973764002323f;
  t261_n49: return 0.06691808253526688f;
  t261_n50: return -0.09502469003200531f;
  t261_n51: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t261_n57;
  t261_n52: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t261_n56;
  t261_n53: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t261_n55;
  t261_n54: return -0.13356436789035797f;
  t261_n55: return 0.11487483978271484f;
  t261_n56: return 0.023719534277915955f;
  t261_n57: return 0.01083978358656168f;
  t261_n58: return -0.032219935208559036f;
  t261_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t261_n79;
  t261_n60: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t261_n72;
  t261_n61: if (E.getNumReferences() >= 3364262336 /*275598.0*/) goto t261_n67;
  t261_n62: if (E.getNumReferences() >= 3361153472 /*220295.0*/) goto t261_n64;
  t261_n63: return -0.009962293319404125f;
  t261_n64: if (E.getFractionNameInContext() >= 3196618889 /*0.2666666805744171*/) goto t261_n66;
  t261_n65: return -0.2774823009967804f;
  t261_n66: return 0.13035835325717926f;
  t261_n67: if (E.getNumReferences() >= 3370129424 /*458944.5*/) goto t261_n71;
  t261_n68: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t261_n70;
  t261_n69: return 0.11257898807525635f;
  t261_n70: return 0.062463901937007904f;
  t261_n71: return -0.027762748301029205f;
  t261_n72: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t261_n74;
  t261_n73: return 0.014104786328971386f;
  t261_n74: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t261_n78;
  t261_n75: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t261_n77;
  t261_n76: return -0.7692952156066895f;
  t261_n77: return -0.012550457380712032f;
  t261_n78: return 0.12722785770893097f;
  t261_n79: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t261_n87;
  t261_n80: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t261_n82;
  t261_n81: return 0.007352425251156092f;
  t261_n82: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t261_n84;
  t261_n83: return 0.021438930183649063f;
  t261_n84: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t261_n86;
  t261_n85: return 0.12236205488443375f;
  t261_n86: return 0.06566081196069717f;
  t261_n87: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t261_n129;
  t261_n88: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t261_n90;
  t261_n89: return -0.005825493950396776f;
  t261_n90: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t261_n108;
  t261_n91: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t261_n93;
  t261_n92: return 0.00701347179710865f;
  t261_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t261_n97;
  t261_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t261_n96;
  t261_n95: return -0.7934179902076721f;
  t261_n96: return 0.10769926011562347f;
  t261_n97: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t261_n107;
  t261_n98: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t261_n100;
  t261_n99: return -0.09359677881002426f;
  t261_n100: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t261_n106;
  t261_n101: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t261_n105;
  t261_n102: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t261_n104;
  t261_n103: return 0.0981127992272377f;
  t261_n104: return -0.23130545020103455f;
  t261_n105: return -0.16157636046409607f;
  t261_n106: return 0.06776534765958786f;
  t261_n107: return -0.45475533604621887f;
  t261_n108: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t261_n110;
  t261_n109: return 0.03402448073029518f;
  t261_n110: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t261_n112;
  t261_n111: return 0.08331488817930222f;
  t261_n112: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t261_n114;
  t261_n113: return -0.24959354102611542f;
  t261_n114: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t261_n116;
  t261_n115: return 0.013001542538404465f;
  t261_n116: if (E.getNumReferences() >= 3312979968 /*3969.0*/) goto t261_n128;
  t261_n117: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t261_n121;
  t261_n118: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t261_n120;
  t261_n119: return -0.7194625735282898f;
  t261_n120: return -0.005662471987307072f;
  t261_n121: if (E.getNumReferences() >= 3266772992 /*91.5*/) goto t261_n125;
  t261_n122: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t261_n124;
  t261_n123: return 0.1151425763964653f;
  t261_n124: return -0.19370301067829132f;
  t261_n125: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t261_n127;
  t261_n126: return -0.3664775490760803f;
  t261_n127: return 0.10574988275766373f;
  t261_n128: return 0.07570970058441162f;
  t261_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t261_n233;
  t261_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t261_n232;
  t261_n131: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t261_n181;
  t261_n132: if (E.getNumReferences() >= 3280437248 /*271.0*/) goto t261_n174;
  t261_n133: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t261_n149;
  t261_n134: if (E.getNumReferences() >= 3267428352 /*96.5*/) goto t261_n148;
  t261_n135: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t261_n147;
  t261_n136: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t261_n146;
  t261_n137: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t261_n143;
  t261_n138: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t261_n142;
  t261_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t261_n141;
  t261_n140: return -0.1336865872144699f;
  t261_n141: return 0.08373535424470901f;
  t261_n142: return 0.057085201144218445f;
  t261_n143: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t261_n145;
  t261_n144: return -0.16264985501766205f;
  t261_n145: return 0.08257577568292618f;
  t261_n146: return -0.30030444264411926f;
  t261_n147: return 0.08433090150356293f;
  t261_n148: return -0.3051929175853729f;
  t261_n149: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t261_n165;
  t261_n150: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t261_n154;
  t261_n151: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t261_n153;
  t261_n152: return 0.08368996530771255f;
  t261_n153: return 0.0802028626203537f;
  t261_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t261_n162;
  t261_n155: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t261_n161;
  t261_n156: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t261_n158;
  t261_n157: return -0.31499502062797546f;
  t261_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t261_n160;
  t261_n159: return 0.06608245521783829f;
  t261_n160: return 0.07060011476278305f;
  t261_n161: return -0.17203447222709656f;
  t261_n162: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t261_n164;
  t261_n163: return 0.05775947496294975f;
  t261_n164: return -0.19938182830810547f;
  t261_n165: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t261_n167;
  t261_n166: return -0.06356745958328247f;
  t261_n167: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t261_n171;
  t261_n168: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t261_n170;
  t261_n169: return -0.23825404047966003f;
  t261_n170: return 0.032510530203580856f;
  t261_n171: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t261_n173;
  t261_n172: return -0.23461532592773438f;
  t261_n173: return -0.028961731120944023f;
  t261_n174: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t261_n180;
  t261_n175: if (E.getNumReferences() >= 3282976768 /*348.5*/) goto t261_n177;
  t261_n176: return 0.13983742892742157f;
  t261_n177: if (E.getNumReferences() >= 3301646336 /*1625.0*/) goto t261_n179;
  t261_n178: return 0.11288133263587952f;
  t261_n179: return -0.009181240573525429f;
  t261_n180: return -0.09901779890060425f;
  t261_n181: if (E.getNumReferences() >= 3349220288 /*82455.5*/) goto t261_n231;
  t261_n182: if (E.getNumReferences() >= 3338694400 /*32879.0*/) goto t261_n228;
  t261_n183: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t261_n225;
  t261_n184: if (E.getNumReferences() >= 3273588736 /*159.0*/) goto t261_n216;
  t261_n185: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t261_n213;
  t261_n186: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t261_n208;
  t261_n187: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t261_n199;
  t261_n188: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t261_n194;
  t261_n189: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t261_n193;
  t261_n190: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t261_n192;
  t261_n191: return 0.1624363362789154f;
  t261_n192: return -0.25460588932037354f;
  t261_n193: return -0.3047841191291809f;
  t261_n194: if (E.getNumReferences() >= 3247964160 /*19.0*/) goto t261_n196;
  t261_n195: return 0.10319150239229202f;
  t261_n196: if (E.getNumReferences() >= 3255042048 /*33.0*/) goto t261_n198;
  t261_n197: return -0.4288650155067444f;
  t261_n198: return 0.02442270889878273f;
  t261_n199: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t261_n207;
  t261_n200: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t261_n204;
  t261_n201: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t261_n203;
  t261_n202: return 0.0941460132598877f;
  t261_n203: return -0.12523780763149261f;
  t261_n204: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t261_n206;
  t261_n205: return 0.11315039545297623f;
  t261_n206: return 0.11203529685735703f;
  t261_n207: return -0.004516097716987133f;
  t261_n208: if (E.getNumReferences() >= 3263692800 /*68.0*/) goto t261_n212;
  t261_n209: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t261_n211;
  t261_n210: return 0.11378739774227142f;
  t261_n211: return 0.019419290125370026f;
  t261_n212: return 0.0991358608007431f;
  t261_n213: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t261_n215;
  t261_n214: return 0.13566522300243378f;
  t261_n215: return -0.12500593066215515f;
  t261_n216: if (E.getNumReferences() >= 3316701184 /*5659.0*/) goto t261_n220;
  t261_n217: if (E.getNumReferences() >= 3315179520 /*4916.0*/) goto t261_n219;
  t261_n218: return -0.11933595687150955f;
  t261_n219: return -0.7370010018348694f;
  t261_n220: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t261_n224;
  t261_n221: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t261_n223;
  t261_n222: return -0.12784773111343384f;
  t261_n223: return 0.11879199743270874f;
  t261_n224: return -0.1595577895641327f;
  t261_n225: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t261_n227;
  t261_n226: return -0.12868578732013702f;
  t261_n227: return -0.05736614018678665f;
  t261_n228: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t261_n230;
  t261_n229: return -0.06961333006620407f;
  t261_n230: return -0.4432595670223236f;
  t261_n231: return 0.03313414379954338f;
  t261_n232: return 0.039958540350198746f;
  t261_n233: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t261_n245;
  t261_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t261_n238;
  t261_n235: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t261_n237;
  t261_n236: return 0.05544666200876236f;
  t261_n237: return -0.28564366698265076f;
  t261_n238: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t261_n244;
  t261_n239: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t261_n241;
  t261_n240: return 0.08616844564676285f;
  t261_n241: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t261_n243;
  t261_n242: return 0.09955912828445435f;
  t261_n243: return 0.12568072974681854f;
  t261_n244: return 0.07205747067928314f;
  t261_n245: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t261_n251;
  t261_n246: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t261_n250;
  t261_n247: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t261_n249;
  t261_n248: return -0.7703006267547607f;
  t261_n249: return -0.06720230728387833f;
  t261_n250: return 0.027908507734537125f;
  t261_n251: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t261_n253;
  t261_n252: return 0.010355936363339424f;
  t261_n253: return 0.109202541410923f;
  t261_n254: return 0.0010722011793404818f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree262(const Example& E) {
  t262_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t262_n254;
  t262_n1: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t262_n243;
  t262_n2: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t262_n148;
  t262_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n45;
  t262_n4: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t262_n44;
  t262_n5: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t262_n7;
  t262_n6: return -0.0023607518523931503f;
  t262_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t262_n13;
  t262_n8: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t262_n12;
  t262_n9: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n11;
  t262_n10: return -0.001273528323508799f;
  t262_n11: return -0.011595351621508598f;
  t262_n12: return 0.004406399559229612f;
  t262_n13: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t262_n33;
  t262_n14: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t262_n28;
  t262_n15: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t262_n25;
  t262_n16: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n20;
  t262_n17: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t262_n19;
  t262_n18: return -0.0019024793291464448f;
  t262_n19: return -0.3243067264556885f;
  t262_n20: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t262_n24;
  t262_n21: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t262_n23;
  t262_n22: return 0.029996130615472794f;
  t262_n23: return -0.009036761708557606f;
  t262_n24: return -0.00017763057257980108f;
  t262_n25: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t262_n27;
  t262_n26: return 0.052588582038879395f;
  t262_n27: return 0.0004713085072580725f;
  t262_n28: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n30;
  t262_n29: return 0.021841252222657204f;
  t262_n30: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t262_n32;
  t262_n31: return -0.0224851593375206f;
  t262_n32: return 0.02569383569061756f;
  t262_n33: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n35;
  t262_n34: return 0.01061207614839077f;
  t262_n35: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t262_n37;
  t262_n36: return 0.009555956348776817f;
  t262_n37: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t262_n43;
  t262_n38: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t262_n40;
  t262_n39: return 0.04607656970620155f;
  t262_n40: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t262_n42;
  t262_n41: return 0.04996083676815033f;
  t262_n42: return -0.06929594278335571f;
  t262_n43: return -0.1391420215368271f;
  t262_n44: return -0.030666593462228775f;
  t262_n45: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t262_n47;
  t262_n46: return -0.00024055314133875072f;
  t262_n47: if (E.getFractionNameInContext() >= 3173036146 /*0.039230771362781525*/) goto t262_n49;
  t262_n48: return 0.14974918961524963f;
  t262_n49: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t262_n79;
  t262_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n52;
  t262_n51: return -0.024293839931488037f;
  t262_n52: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t262_n54;
  t262_n53: return 0.005697482731193304f;
  t262_n54: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t262_n58;
  t262_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n57;
  t262_n56: return -0.012733865529298782f;
  t262_n57: return 0.010288363322615623f;
  t262_n58: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t262_n74;
  t262_n59: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t262_n61;
  t262_n60: return -0.01810193806886673f;
  t262_n61: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t262_n69;
  t262_n62: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t262_n64;
  t262_n63: return -0.03608047589659691f;
  t262_n64: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t262_n68;
  t262_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t262_n67;
  t262_n66: return 0.04298624023795128f;
  t262_n67: return 0.07841872423887253f;
  t262_n68: return 0.0033885869197547436f;
  t262_n69: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t262_n73;
  t262_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t262_n72;
  t262_n71: return 0.03142077848315239f;
  t262_n72: return 0.09898827224969864f;
  t262_n73: return 0.011316635645925999f;
  t262_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n76;
  t262_n75: return -0.0015701205702498555f;
  t262_n76: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t262_n78;
  t262_n77: return 0.002660900354385376f;
  t262_n78: return 0.010584604926407337f;
  t262_n79: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t262_n147;
  t262_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t262_n106;
  t262_n81: if (E.getFractionNameInContext() >= 3189742720 /*0.1558704376220703*/) goto t262_n103;
  t262_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n98;
  t262_n83: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t262_n85;
  t262_n84: return 0.01856665126979351f;
  t262_n85: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t262_n95;
  t262_n86: if (E.getFileProximityDistanceCost() >= 3254255616 /*31.0*/) goto t262_n90;
  t262_n87: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t262_n89;
  t262_n88: return -0.01224904228001833f;
  t262_n89: return -0.23621562123298645f;
  t262_n90: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t262_n94;
  t262_n91: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t262_n93;
  t262_n92: return 0.10240981727838516f;
  t262_n93: return 0.03275167569518089f;
  t262_n94: return -0.09661403298377991f;
  t262_n95: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t262_n97;
  t262_n96: return -0.06862006336450577f;
  t262_n97: return -0.27215465903282166f;
  t262_n98: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t262_n102;
  t262_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t262_n101;
  t262_n100: return -0.18372343480587006f;
  t262_n101: return 0.1315823495388031f;
  t262_n102: return -0.0017211450031027198f;
  t262_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t262_n105;
  t262_n104: return -0.001036793808452785f;
  t262_n105: return 0.00787798035889864f;
  t262_n106: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t262_n146;
  t262_n107: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t262_n111;
  t262_n108: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t262_n110;
  t262_n109: return 0.01078835129737854f;
  t262_n110: return 0.15614719688892365f;
  t262_n111: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t262_n137;
  t262_n112: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t262_n120;
  t262_n113: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t262_n117;
  t262_n114: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t262_n116;
  t262_n115: return 0.010486018843948841f;
  t262_n116: return -0.035964660346508026f;
  t262_n117: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t262_n119;
  t262_n118: return -0.09843596816062927f;
  t262_n119: return -0.023279543966054916f;
  t262_n120: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t262_n136;
  t262_n121: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t262_n123;
  t262_n122: return 0.07003697752952576f;
  t262_n123: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t262_n127;
  t262_n124: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t262_n126;
  t262_n125: return 0.09138932824134827f;
  t262_n126: return -0.03852396458387375f;
  t262_n127: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t262_n129;
  t262_n128: return 0.03932476416230202f;
  t262_n129: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t262_n131;
  t262_n130: return -0.019347483292222023f;
  t262_n131: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t262_n133;
  t262_n132: return 0.036577653139829636f;
  t262_n133: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t262_n135;
  t262_n134: return -0.030840333551168442f;
  t262_n135: return 0.026858873665332794f;
  t262_n136: return 0.006764498073607683f;
  t262_n137: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t262_n141;
  t262_n138: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t262_n140;
  t262_n139: return -0.02601313404738903f;
  t262_n140: return -0.14911498129367828f;
  t262_n141: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t262_n145;
  t262_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n144;
  t262_n143: return -0.1815851926803589f;
  t262_n144: return 0.021018482744693756f;
  t262_n145: return -0.006149402819573879f;
  t262_n146: return -0.0022833591792732477f;
  t262_n147: return 0.049461886286735535f;
  t262_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol))) goto t262_n200;
  t262_n149: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t262_n163;
  t262_n150: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t262_n156;
  t262_n151: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t262_n155;
  t262_n152: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t262_n154;
  t262_n153: return -0.14908438920974731f;
  t262_n154: return 0.09499246627092361f;
  t262_n155: return -0.03192492201924324f;
  t262_n156: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t262_n162;
  t262_n157: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t262_n159;
  t262_n158: return -0.07622656226158142f;
  t262_n159: if (E.getFractionNameInContext() >= 3185993318 /*0.11249999701976776*/) goto t262_n161;
  t262_n160: return 0.12290910631418228f;
  t262_n161: return 0.07298754155635834f;
  t262_n162: return 0.09796150028705597f;
  t262_n163: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t262_n165;
  t262_n164: return 0.0030818909872323275f;
  t262_n165: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t262_n167;
  t262_n166: return -0.010697540827095509f;
  t262_n167: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t262_n199;
  t262_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t262_n186;
  t262_n169: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t262_n175;
  t262_n170: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t262_n174;
  t262_n171: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t262_n173;
  t262_n172: return -0.08113288879394531f;
  t262_n173: return 0.03434880077838898f;
  t262_n174: return 0.10386651754379272f;
  t262_n175: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t262_n185;
  t262_n176: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t262_n180;
  t262_n177: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n179;
  t262_n178: return 0.013486754149198532f;
  t262_n179: return 0.15692394971847534f;
  t262_n180: if (E.getFractionNameInContext() >= 3199255308 /*0.3452380895614624*/) goto t262_n184;
  t262_n181: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t262_n183;
  t262_n182: return 0.1278836727142334f;
  t262_n183: return 0.12297340482473373f;
  t262_n184: return 0.06046309694647789f;
  t262_n185: return 0.07560541480779648f;
  t262_n186: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t262_n198;
  t262_n187: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t262_n195;
  t262_n188: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t262_n192;
  t262_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t262_n191;
  t262_n190: return -0.045126158744096756f;
  t262_n191: return 0.1114116758108139f;
  t262_n192: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t262_n194;
  t262_n193: return 0.12356936931610107f;
  t262_n194: return -0.011178099550306797f;
  t262_n195: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t262_n197;
  t262_n196: return 0.3079167902469635f;
  t262_n197: return 0.09953916072845459f;
  t262_n198: return 0.05524570867419243f;
  t262_n199: return 0.11803765594959259f;
  t262_n200: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t262_n202;
  t262_n201: return 0.013800320215523243f;
  t262_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t262_n238;
  t262_n203: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t262_n215;
  t262_n204: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t262_n206;
  t262_n205: return 0.004356069955974817f;
  t262_n206: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t262_n208;
  t262_n207: return 0.020167091861367226f;
  t262_n208: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t262_n212;
  t262_n209: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t262_n211;
  t262_n210: return 0.1018873006105423f;
  t262_n211: return -0.14548014104366302f;
  t262_n212: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t262_n214;
  t262_n213: return -0.02465050294995308f;
  t262_n214: return 0.07257521897554398f;
  t262_n215: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t262_n237;
  t262_n216: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t262_n236;
  t262_n217: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t262_n235;
  t262_n218: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t262_n232;
  t262_n219: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t262_n223;
  t262_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t262_n222;
  t262_n221: return -0.16759055852890015f;
  t262_n222: return 0.1240268349647522f;
  t262_n223: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t262_n225;
  t262_n224: return -0.01607595942914486f;
  t262_n225: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t262_n229;
  t262_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t262_n228;
  t262_n227: return -0.07440132647752762f;
  t262_n228: return 0.09499821811914444f;
  t262_n229: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t262_n231;
  t262_n230: return 0.11329200863838196f;
  t262_n231: return 0.04024846851825714f;
  t262_n232: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t262_n234;
  t262_n233: return 0.032875023782253265f;
  t262_n234: return 0.11409083008766174f;
  t262_n235: return -0.2969926595687866f;
  t262_n236: return -0.02099432982504368f;
  t262_n237: return 0.16362237930297852f;
  t262_n238: if (E.getFractionNameInContext() >= 3183942770 /*0.0972222238779068*/) goto t262_n240;
  t262_n239: return 0.08895602077245712f;
  t262_n240: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t262_n242;
  t262_n241: return -0.0806017592549324f;
  t262_n242: return 0.11369969695806503f;
  t262_n243: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t262_n253;
  t262_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t262_n248;
  t262_n245: if (E.getNumReferences() >= 3258056704 /*44.5*/) goto t262_n247;
  t262_n246: return -0.0237779151648283f;
  t262_n247: return -0.0026329203974455595f;
  t262_n248: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t262_n250;
  t262_n249: return 0.005486816167831421f;
  t262_n250: if (E.getNumReferences() >= 3258843136 /*47.5*/) goto t262_n252;
  t262_n251: return -0.10004192590713501f;
  t262_n252: return 0.11404836922883987f;
  t262_n253: return 0.00040012854151427746f;
  t262_n254: return 0.001743600587360561f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree263(const Example& E) {
  t263_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t263_n2;
  t263_n1: return -0.00011282770719844848f;
  t263_n2: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t263_n170;
  t263_n3: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t263_n169;
  t263_n4: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t263_n112;
  t263_n5: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t263_n111;
  t263_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_TopLevel))) goto t263_n110;
  t263_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t263_n91;
  t263_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t263_n88;
  t263_n9: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t263_n77;
  t263_n10: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t263_n70;
  t263_n11: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t263_n55;
  t263_n12: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t263_n50;
  t263_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n17;
  t263_n14: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n16;
  t263_n15: return -0.09744632244110107f;
  t263_n16: return 0.06256032735109329f;
  t263_n17: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t263_n39;
  t263_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t263_n38;
  t263_n19: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t263_n35;
  t263_n20: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t263_n34;
  t263_n21: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n25;
  t263_n22: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t263_n24;
  t263_n23: return -0.12431816011667252f;
  t263_n24: return -0.2005445808172226f;
  t263_n25: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t263_n33;
  t263_n26: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t263_n28;
  t263_n27: return -0.14223134517669678f;
  t263_n28: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t263_n32;
  t263_n29: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t263_n31;
  t263_n30: return 0.14647860825061798f;
  t263_n31: return -0.3937397301197052f;
  t263_n32: return -0.1828109622001648f;
  t263_n33: return -0.48719194531440735f;
  t263_n34: return -0.023801308125257492f;
  t263_n35: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t263_n37;
  t263_n36: return 0.016324980184435844f;
  t263_n37: return 0.08556336909532547f;
  t263_n38: return 0.009900417178869247f;
  t263_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n45;
  t263_n40: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t263_n44;
  t263_n41: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t263_n43;
  t263_n42: return -0.2598820924758911f;
  t263_n43: return 0.014747509732842445f;
  t263_n44: return -0.7681601643562317f;
  t263_n45: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t263_n47;
  t263_n46: return -0.037011098116636276f;
  t263_n47: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t263_n49;
  t263_n48: return 0.11204418540000916f;
  t263_n49: return -0.10450492054224014f;
  t263_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n52;
  t263_n51: return -0.06628250330686569f;
  t263_n52: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t263_n54;
  t263_n53: return -0.016526218503713608f;
  t263_n54: return 0.14642222225666046f;
  t263_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t263_n65;
  t263_n56: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t263_n64;
  t263_n57: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t263_n63;
  t263_n58: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t263_n60;
  t263_n59: return 0.004712248221039772f;
  t263_n60: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t263_n62;
  t263_n61: return -0.1841191202402115f;
  t263_n62: return 0.10418243706226349f;
  t263_n63: return -0.26782357692718506f;
  t263_n64: return 0.09297651797533035f;
  t263_n65: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t263_n67;
  t263_n66: return 0.12426618486642838f;
  t263_n67: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t263_n69;
  t263_n68: return -0.13582852482795715f;
  t263_n69: return 0.13492803275585175f;
  t263_n70: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t263_n74;
  t263_n71: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t263_n73;
  t263_n72: return 0.08164482563734055f;
  t263_n73: return 0.16357344388961792f;
  t263_n74: if (E.getFractionNameInContext() >= 3192331378 /*0.1944444477558136*/) goto t263_n76;
  t263_n75: return -0.2841401994228363f;
  t263_n76: return 0.09169796854257584f;
  t263_n77: if (E.getNumReferences() >= 3265069056 /*78.5*/) goto t263_n87;
  t263_n78: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t263_n82;
  t263_n79: if (E.getNumReferences() >= 3259891712 /*51.5*/) goto t263_n81;
  t263_n80: return -0.19306187331676483f;
  t263_n81: return 0.022798001766204834f;
  t263_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n86;
  t263_n83: if (E.getNumReferences() >= 3264151552 /*71.5*/) goto t263_n85;
  t263_n84: return -1.089254379272461f;
  t263_n85: return -0.460662841796875f;
  t263_n86: return -0.059738051146268845f;
  t263_n87: return -0.00570278475061059f;
  t263_n88: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t263_n90;
  t263_n89: return -0.012298418208956718f;
  t263_n90: return 0.15773777663707733f;
  t263_n91: if (E.getNumReferences() >= 3274473472 /*172.5*/) goto t263_n109;
  t263_n92: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t263_n106;
  t263_n93: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t263_n99;
  t263_n94: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t263_n96;
  t263_n95: return 0.03818855807185173f;
  t263_n96: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t263_n98;
  t263_n97: return -0.008545505814254284f;
  t263_n98: return -0.1479843705892563f;
  t263_n99: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t263_n105;
  t263_n100: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t263_n102;
  t263_n101: return 0.03586186468601227f;
  t263_n102: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t263_n104;
  t263_n103: return -0.08049960434436798f;
  t263_n104: return 0.0856826975941658f;
  t263_n105: return -0.003892738837748766f;
  t263_n106: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t263_n108;
  t263_n107: return 0.04641666263341904f;
  t263_n108: return -0.02582913637161255f;
  t263_n109: return 0.12071242928504944f;
  t263_n110: return 0.0234243031591177f;
  t263_n111: return 0.051085274666547775f;
  t263_n112: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t263_n140;
  t263_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n135;
  t263_n114: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t263_n128;
  t263_n115: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t263_n127;
  t263_n116: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t263_n124;
  t263_n117: if (E.getFractionNameInContext() >= 3200482732 /*0.38181817531585693*/) goto t263_n123;
  t263_n118: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n120;
  t263_n119: return -0.13477212190628052f;
  t263_n120: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t263_n122;
  t263_n121: return -0.08088700473308563f;
  t263_n122: return -0.6654525399208069f;
  t263_n123: return 0.05470235273241997f;
  t263_n124: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t263_n126;
  t263_n125: return -0.017985768616199493f;
  t263_n126: return -0.3254176676273346f;
  t263_n127: return 0.09085650742053986f;
  t263_n128: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t263_n134;
  t263_n129: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t263_n133;
  t263_n130: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t263_n132;
  t263_n131: return 0.12264148145914078f;
  t263_n132: return 0.0925246998667717f;
  t263_n133: return -0.013766392134130001f;
  t263_n134: return -0.21518473327159882f;
  t263_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n139;
  t263_n136: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t263_n138;
  t263_n137: return -0.0011606644839048386f;
  t263_n138: return -0.1415904462337494f;
  t263_n139: return 0.11268599331378937f;
  t263_n140: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t263_n168;
  t263_n141: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t263_n167;
  t263_n142: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t263_n158;
  t263_n143: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t263_n157;
  t263_n144: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t263_n154;
  t263_n145: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t263_n153;
  t263_n146: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t263_n148;
  t263_n147: return 0.005334124434739351f;
  t263_n148: if (E.getNumReferences() >= 3339744256 /*36980.0*/) goto t263_n150;
  t263_n149: return -0.05408273637294769f;
  t263_n150: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t263_n152;
  t263_n151: return -0.03017379343509674f;
  t263_n152: return 0.05504960939288139f;
  t263_n153: return 0.12009009718894958f;
  t263_n154: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t263_n156;
  t263_n155: return -0.030562300235033035f;
  t263_n156: return 0.005685066804289818f;
  t263_n157: return -0.3360721468925476f;
  t263_n158: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t263_n164;
  t263_n159: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t263_n163;
  t263_n160: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n162;
  t263_n161: return 0.09580536931753159f;
  t263_n162: return 0.12635710835456848f;
  t263_n163: return 0.06192624196410179f;
  t263_n164: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n166;
  t263_n165: return 0.016870856285095215f;
  t263_n166: return -4.674543380737305f;
  t263_n167: return 0.3881589472293854f;
  t263_n168: return 0.002436999697238207f;
  t263_n169: return 0.07016344368457794f;
  t263_n170: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n252;
  t263_n171: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t263_n215;
  t263_n172: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t263_n178;
  t263_n173: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t263_n177;
  t263_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t263_n176;
  t263_n175: return -0.0014465772546827793f;
  t263_n176: return 0.06887877732515335f;
  t263_n177: return -0.15387585759162903f;
  t263_n178: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t263_n210;
  t263_n179: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t263_n197;
  t263_n180: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t263_n182;
  t263_n181: return -0.3189833462238312f;
  t263_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t263_n184;
  t263_n183: return 0.03258739039301872f;
  t263_n184: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t263_n196;
  t263_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n193;
  t263_n186: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t263_n188;
  t263_n187: return 0.12969347834587097f;
  t263_n188: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n190;
  t263_n189: return 0.06972930580377579f;
  t263_n190: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t263_n192;
  t263_n191: return 0.19139109551906586f;
  t263_n192: return 0.09817677736282349f;
  t263_n193: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n195;
  t263_n194: return 0.2014501839876175f;
  t263_n195: return 0.0010361244203522801f;
  t263_n196: return -0.07578974217176437f;
  t263_n197: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t263_n201;
  t263_n198: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t263_n200;
  t263_n199: return -0.49239131808280945f;
  t263_n200: return -0.06353235244750977f;
  t263_n201: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t263_n209;
  t263_n202: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t263_n208;
  t263_n203: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n205;
  t263_n204: return -0.23839642107486725f;
  t263_n205: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t263_n207;
  t263_n206: return 0.134088933467865f;
  t263_n207: return -0.11489581316709518f;
  t263_n208: return 0.20520617067813873f;
  t263_n209: return 0.09581983089447021f;
  t263_n210: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t263_n212;
  t263_n211: return 0.040919892489910126f;
  t263_n212: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t263_n214;
  t263_n213: return 0.1637095808982849f;
  t263_n214: return 0.19748014211654663f;
  t263_n215: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t263_n231;
  t263_n216: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t263_n226;
  t263_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t263_n225;
  t263_n218: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t263_n220;
  t263_n219: return -0.07195708155632019f;
  t263_n220: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t263_n222;
  t263_n221: return -0.0461236834526062f;
  t263_n222: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n224;
  t263_n223: return 0.12623731791973114f;
  t263_n224: return 0.021493475884199142f;
  t263_n225: return -0.06350846588611603f;
  t263_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t263_n228;
  t263_n227: return -0.4083215892314911f;
  t263_n228: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t263_n230;
  t263_n229: return 0.059458017349243164f;
  t263_n230: return 0.13821496069431305f;
  t263_n231: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t263_n251;
  t263_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t263_n242;
  t263_n233: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t263_n235;
  t263_n234: return -0.05888419598340988f;
  t263_n235: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t263_n241;
  t263_n236: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t263_n240;
  t263_n237: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n239;
  t263_n238: return 0.18668989837169647f;
  t263_n239: return -0.19921882450580597f;
  t263_n240: return 2.300654172897339f;
  t263_n241: return 0.012529445812106133f;
  t263_n242: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t263_n250;
  t263_n243: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t263_n249;
  t263_n244: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t263_n248;
  t263_n245: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t263_n247;
  t263_n246: return 0.18087142705917358f;
  t263_n247: return -0.10256435722112656f;
  t263_n248: return 0.29618531465530396f;
  t263_n249: return -0.20433934032917023f;
  t263_n250: return 0.11193715780973434f;
  t263_n251: return 0.17939706146717072f;
  t263_n252: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t263_n254;
  t263_n253: return 0.06300811469554901f;
  t263_n254: return 0.1291343867778778f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree264(const Example& E) {
  t264_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t264_n232;
  t264_n1: if (E.getNumReferences() >= 3303133184 /*1806.5*/) goto t264_n5;
  t264_n2: if (E.getNumReferences() >= 3303108608 /*1803.5*/) goto t264_n4;
  t264_n3: return -0.00021546843345277011f;
  t264_n4: return -0.1225680336356163f;
  t264_n5: if (E.getNumReferences() >= 3303182336 /*1812.5*/) goto t264_n13;
  t264_n6: if (E.getNumReferences() >= 3303165952 /*1810.5*/) goto t264_n10;
  t264_n7: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t264_n9;
  t264_n8: return 0.25210291147232056f;
  t264_n9: return -0.05038895457983017f;
  t264_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t264_n12;
  t264_n11: return 0.11352844536304474f;
  t264_n12: return 0.17711129784584045f;
  t264_n13: if (E.getNumReferences() >= 3304108032 /*1925.5*/) goto t264_n69;
  t264_n14: if (E.getNumReferences() >= 3304099840 /*1924.5*/) goto t264_n68;
  t264_n15: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t264_n39;
  t264_n16: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t264_n20;
  t264_n17: if (E.getNumReferences() >= 3303247872 /*1820.5*/) goto t264_n19;
  t264_n18: return -0.4448087811470032f;
  t264_n19: return 0.012565139681100845f;
  t264_n20: if (E.getNumReferences() >= 3303190528 /*1813.5*/) goto t264_n22;
  t264_n21: return -0.14980411529541016f;
  t264_n22: if (E.getNumReferences() >= 3303440384 /*1844.0*/) goto t264_n30;
  t264_n23: if (E.getNumReferences() >= 3303419904 /*1841.5*/) goto t264_n29;
  t264_n24: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t264_n28;
  t264_n25: if (E.getNumReferences() >= 3303256064 /*1821.5*/) goto t264_n27;
  t264_n26: return 0.17544376850128174f;
  t264_n27: return -0.04498232528567314f;
  t264_n28: return 0.016113005578517914f;
  t264_n29: return 0.11855549365282059f;
  t264_n30: if (E.getNumReferences() >= 3303559168 /*1858.5*/) goto t264_n32;
  t264_n31: return -0.09756551682949066f;
  t264_n32: if (E.getNumReferences() >= 3303796736 /*1887.5*/) goto t264_n38;
  t264_n33: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t264_n37;
  t264_n34: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t264_n36;
  t264_n35: return 0.037133656442165375f;
  t264_n36: return 0.1439816802740097f;
  t264_n37: return 0.05667645111680031f;
  t264_n38: return 0.04824512451887131f;
  t264_n39: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t264_n41;
  t264_n40: return -0.04331805929541588f;
  t264_n41: if (E.getNumReferences() >= 3303247872 /*1820.5*/) goto t264_n47;
  t264_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t264_n46;
  t264_n43: if (E.getNumReferences() >= 3303227392 /*1818.0*/) goto t264_n45;
  t264_n44: return 0.0010912357829511166f;
  t264_n45: return 0.15215718746185303f;
  t264_n46: return 0.211378276348114f;
  t264_n47: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t264_n53;
  t264_n48: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t264_n50;
  t264_n49: return 0.004628479015082121f;
  t264_n50: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t264_n52;
  t264_n51: return 0.07277590036392212f;
  t264_n52: return 0.13414014875888824f;
  t264_n53: if (E.getNumReferences() >= 3303981056 /*1910.0*/) goto t264_n67;
  t264_n54: if (E.getNumReferences() >= 3303858176 /*1895.0*/) goto t264_n60;
  t264_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t264_n57;
  t264_n56: return -0.08771222829818726f;
  t264_n57: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t264_n59;
  t264_n58: return 0.1676609069108963f;
  t264_n59: return 0.06773702800273895f;
  t264_n60: if (E.getNumReferences() >= 3303936000 /*1904.5*/) goto t264_n64;
  t264_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t264_n63;
  t264_n62: return 0.08794928342103958f;
  t264_n63: return 0.14090873301029205f;
  t264_n64: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t264_n66;
  t264_n65: return 0.14092770218849182f;
  t264_n66: return -0.056439753621816635f;
  t264_n67: return -0.17584703862667084f;
  t264_n68: return 0.2852163016796112f;
  t264_n69: if (E.getNumReferences() >= 3304742912 /*2003.0*/) goto t264_n101;
  t264_n70: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t264_n100;
  t264_n71: if (E.getNumReferences() >= 3304624128 /*1988.5*/) goto t264_n91;
  t264_n72: if (E.getNumReferences() >= 3304128512 /*1928.0*/) goto t264_n76;
  t264_n73: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t264_n75;
  t264_n74: return -0.3332156538963318f;
  t264_n75: return -0.07754141837358475f;
  t264_n76: if (E.getNumReferences() >= 3304226816 /*1940.0*/) goto t264_n84;
  t264_n77: if (E.getNumReferences() >= 3304206336 /*1937.5*/) goto t264_n79;
  t264_n78: return -0.01726219616830349f;
  t264_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t264_n81;
  t264_n80: return -0.10062512755393982f;
  t264_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t264_n83;
  t264_n82: return 0.14570480585098267f;
  t264_n83: return 0.14337603747844696f;
  t264_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t264_n86;
  t264_n85: return -0.13481710851192474f;
  t264_n86: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t264_n88;
  t264_n87: return -0.018779784440994263f;
  t264_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t264_n90;
  t264_n89: return 0.10069186240434647f;
  t264_n90: return 0.13523466885089874f;
  t264_n91: if (E.getNumReferences() >= 3304648704 /*1991.5*/) goto t264_n93;
  t264_n92: return -0.1941969245672226f;
  t264_n93: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t264_n95;
  t264_n94: return -0.09235423803329468f;
  t264_n95: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t264_n99;
  t264_n96: if (E.getNumReferences() >= 3304685568 /*1996.0*/) goto t264_n98;
  t264_n97: return 0.14654424786567688f;
  t264_n98: return 0.09285056591033936f;
  t264_n99: return -0.07178331166505814f;
  t264_n100: return 0.13913317024707794f;
  t264_n101: if (E.getNumReferences() >= 3304804352 /*2010.5*/) goto t264_n115;
  t264_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t264_n112;
  t264_n103: if (E.getNumReferences() >= 3304792064 /*2009.0*/) goto t264_n109;
  t264_n104: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t264_n108;
  t264_n105: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t264_n107;
  t264_n106: return 0.13940951228141785f;
  t264_n107: return -0.1430128514766693f;
  t264_n108: return -0.1292477399110794f;
  t264_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t264_n111;
  t264_n110: return -0.06092749163508415f;
  t264_n111: return 0.11981268227100372f;
  t264_n112: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t264_n114;
  t264_n113: return 0.08269201964139938f;
  t264_n114: return 0.11280494928359985f;
  t264_n115: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t264_n151;
  t264_n116: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t264_n122;
  t264_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t264_n121;
  t264_n118: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t264_n120;
  t264_n119: return -0.1348135769367218f;
  t264_n120: return 0.06919434666633606f;
  t264_n121: return 0.008940663188695908f;
  t264_n122: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t264_n126;
  t264_n123: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t264_n125;
  t264_n124: return 0.019511668011546135f;
  t264_n125: return 0.17916099727153778f;
  t264_n126: if (E.getNumReferences() >= 3305140224 /*2055.0*/) goto t264_n132;
  t264_n127: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t264_n131;
  t264_n128: if (E.getFractionNameInContext() >= 3183143855 /*0.0912698432803154*/) goto t264_n130;
  t264_n129: return 0.29333963990211487f;
  t264_n130: return -0.06034204736351967f;
  t264_n131: return -0.009230878204107285f;
  t264_n132: if (E.getNumReferences() >= 3307939840 /*2738.5*/) goto t264_n136;
  t264_n133: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t264_n135;
  t264_n134: return -0.13858292996883392f;
  t264_n135: return -0.07700473070144653f;
  t264_n136: if (E.getNumReferences() >= 3328082944 /*14241.0*/) goto t264_n146;
  t264_n137: if (E.getNumReferences() >= 3326745088 /*12934.5*/) goto t264_n141;
  t264_n138: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t264_n140;
  t264_n139: return -0.10696611553430557f;
  t264_n140: return 0.004511157516390085f;
  t264_n141: if (E.getNumReferences() >= 3326970368 /*13154.5*/) goto t264_n145;
  t264_n142: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t264_n144;
  t264_n143: return 0.18134832382202148f;
  t264_n144: return 0.109281025826931f;
  t264_n145: return 0.01586247608065605f;
  t264_n146: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t264_n148;
  t264_n147: return -0.07665803283452988f;
  t264_n148: if (E.getNumReferences() >= 3332280320 /*20296.0*/) goto t264_n150;
  t264_n149: return 0.07563334703445435f;
  t264_n150: return -0.00778497289866209f;
  t264_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t264_n177;
  t264_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t264_n166;
  t264_n153: if (E.getNumReferences() >= 3359831936 /*199646.0*/) goto t264_n165;
  t264_n154: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t264_n164;
  t264_n155: if (E.getNumReferences() >= 3351450688 /*99880.5*/) goto t264_n163;
  t264_n156: if (E.getNumReferences() >= 3351221312 /*98088.5*/) goto t264_n158;
  t264_n157: return -0.010152717120945454f;
  t264_n158: if (E.getFileProximityDistanceCost() >= 3248488448 /*20.0*/) goto t264_n162;
  t264_n159: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t264_n161;
  t264_n160: return 0.09057797491550446f;
  t264_n161: return 0.02111547812819481f;
  t264_n162: return 0.10996975004673004f;
  t264_n163: return -0.10081915557384491f;
  t264_n164: return 0.0814322829246521f;
  t264_n165: return -0.05222681909799576f;
  t264_n166: if (E.getNumReferences() >= 3382462060 /*1280205.5*/) goto t264_n170;
  t264_n167: if (E.getNumReferences() >= 3307091968 /*2531.5*/) goto t264_n169;
  t264_n168: return -0.10034435242414474f;
  t264_n169: return 0.021374309435486794f;
  t264_n170: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t264_n176;
  t264_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t264_n173;
  t264_n172: return -0.18300999701023102f;
  t264_n173: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t264_n175;
  t264_n174: return -0.027567682787775993f;
  t264_n175: return 0.11361668258905411f;
  t264_n176: return -0.20779792964458466f;
  t264_n177: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t264_n179;
  t264_n178: return -0.017748765647411346f;
  t264_n179: if (E.getNumReferences() >= 3315335168 /*4992.0*/) goto t264_n231;
  t264_n180: if (E.getNumReferences() >= 3315314688 /*4982.0*/) goto t264_n224;
  t264_n181: if (E.getNumReferences() >= 3314332672 /*4502.5*/) goto t264_n203;
  t264_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t264_n184;
  t264_n183: return -0.0004068452399224043f;
  t264_n184: if (E.getNumReferences() >= 3313491968 /*4094.0*/) goto t264_n190;
  t264_n185: if (E.getNumReferences() >= 3313168384 /*4015.0*/) goto t264_n189;
  t264_n186: if (E.getNumReferences() >= 3313162240 /*4013.5*/) goto t264_n188;
  t264_n187: return 0.02155294455587864f;
  t264_n188: return 0.11394326388835907f;
  t264_n189: return -0.17227387428283691f;
  t264_n190: if (E.getNumReferences() >= 3313508352 /*4100.0*/) goto t264_n194;
  t264_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t264_n193;
  t264_n192: return 0.014797285199165344f;
  t264_n193: return 0.15049423277378082f;
  t264_n194: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t264_n202;
  t264_n195: if (E.getNumReferences() >= 3313894400 /*4288.5*/) goto t264_n201;
  t264_n196: if (E.getNumReferences() >= 3313715200 /*4201.0*/) goto t264_n198;
  t264_n197: return 0.18999901413917542f;
  t264_n198: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t264_n200;
  t264_n199: return 0.14387506246566772f;
  t264_n200: return 0.1248740628361702f;
  t264_n201: return -0.10840556770563126f;
  t264_n202: return 0.03601694107055664f;
  t264_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t264_n205;
  t264_n204: return 0.027261532843112946f;
  t264_n205: if (E.getNumReferences() >= 3315181568 /*4917.0*/) goto t264_n221;
  t264_n206: if (E.getNumReferences() >= 3314841600 /*4751.0*/) goto t264_n220;
  t264_n207: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t264_n217;
  t264_n208: if (E.getNumReferences() >= 3314535424 /*4601.5*/) goto t264_n216;
  t264_n209: if (E.getFractionNameInContext() >= 3177875683 /*0.057259451597929*/) goto t264_n211;
  t264_n210: return -0.011415696702897549f;
  t264_n211: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t264_n215;
  t264_n212: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t264_n214;
  t264_n213: return 0.3359915614128113f;
  t264_n214: return 0.10645415633916855f;
  t264_n215: return 0.06582222878932953f;
  t264_n216: return 0.10748492926359177f;
  t264_n217: if (E.getNumReferences() >= 3314396160 /*4533.5*/) goto t264_n219;
  t264_n218: return 0.14734682440757751f;
  t264_n219: return -0.032444488257169724f;
  t264_n220: return -0.048263005912303925f;
  t264_n221: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t264_n223;
  t264_n222: return 0.13103225827217102f;
  t264_n223: return 0.0953642874956131f;
  t264_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t264_n226;
  t264_n225: return -0.30567097663879395f;
  t264_n226: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t264_n228;
  t264_n227: return 0.1513882279396057f;
  t264_n228: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t264_n230;
  t264_n229: return 0.16577160358428955f;
  t264_n230: return 0.08938620984554291f;
  t264_n231: return 0.001385536277666688f;
  t264_n232: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t264_n234;
  t264_n233: return 0.0007822386105544865f;
  t264_n234: if (E.getNumReferences() >= 3306559488 /*2401.5*/) goto t264_n240;
  t264_n235: if (E.getNumReferences() >= 3305687040 /*2188.5*/) goto t264_n237;
  t264_n236: return 0.0027252701111137867f;
  t264_n237: if (E.getNumReferences() >= 3305932800 /*2248.5*/) goto t264_n239;
  t264_n238: return -0.31384536623954773f;
  t264_n239: return -0.06981304287910461f;
  t264_n240: if (E.getNumReferences() >= 3306696704 /*2435.0*/) goto t264_n242;
  t264_n241: return 0.10872606933116913f;
  t264_n242: if (E.getNumReferences() >= 3308023808 /*2759.0*/) goto t264_n254;
  t264_n243: if (E.getNumReferences() >= 3307098112 /*2533.0*/) goto t264_n245;
  t264_n244: return -0.05308517813682556f;
  t264_n245: if (E.getNumReferences() >= 3307692032 /*2678.0*/) goto t264_n251;
  t264_n246: if (E.getNumReferences() >= 3307503616 /*2632.0*/) goto t264_n248;
  t264_n247: return 0.09417887777090073f;
  t264_n248: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t264_n250;
  t264_n249: return 0.11453805863857269f;
  t264_n250: return 0.033542729914188385f;
  t264_n251: if (E.getNumReferences() >= 3307880448 /*2724.0*/) goto t264_n253;
  t264_n252: return -0.11449357122182846f;
  t264_n253: return 0.18424706161022186f;
  t264_n254: return 0.018205825239419937f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree265(const Example& E) {
  t265_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t265_n2;
  t265_n1: return -0.00034551462158560753f;
  t265_n2: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t265_n6;
  t265_n3: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t265_n5;
  t265_n4: return -0.030172273516654968f;
  t265_n5: return 0.14128297567367554f;
  t265_n6: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t265_n254;
  t265_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t265_n15;
  t265_n8: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t265_n10;
  t265_n9: return 0.0014120271662250161f;
  t265_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n12;
  t265_n11: return -0.06053333729505539f;
  t265_n12: if (E.getNumReferences() >= 3364597360 /*286067.5*/) goto t265_n14;
  t265_n13: return -0.08281359076499939f;
  t265_n14: return 0.030399437993764877f;
  t265_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t265_n23;
  t265_n16: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t265_n22;
  t265_n17: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t265_n19;
  t265_n18: return -0.0066990661434829235f;
  t265_n19: if (E.getNumReferences() >= 3283238912 /*356.5*/) goto t265_n21;
  t265_n20: return -0.0959552451968193f;
  t265_n21: return 0.10278939455747604f;
  t265_n22: return 0.019518274813890457f;
  t265_n23: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t265_n151;
  t265_n24: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t265_n140;
  t265_n25: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t265_n139;
  t265_n26: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t265_n66;
  t265_n27: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n53;
  t265_n28: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n42;
  t265_n29: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t265_n35;
  t265_n30: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t265_n34;
  t265_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n33;
  t265_n32: return -0.08223582804203033f;
  t265_n33: return 0.0212321188300848f;
  t265_n34: return -0.3013829290866852f;
  t265_n35: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t265_n41;
  t265_n36: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t265_n40;
  t265_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t265_n39;
  t265_n38: return -0.03893286734819412f;
  t265_n39: return 0.06856188923120499f;
  t265_n40: return 0.2715471684932709f;
  t265_n41: return -0.011336876079440117f;
  t265_n42: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t265_n44;
  t265_n43: return -0.004943088628351688f;
  t265_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t265_n46;
  t265_n45: return 0.023456381633877754f;
  t265_n46: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t265_n52;
  t265_n47: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t265_n49;
  t265_n48: return 0.05790785700082779f;
  t265_n49: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t265_n51;
  t265_n50: return 0.09362877160310745f;
  t265_n51: return 0.05590379610657692f;
  t265_n52: return -0.044759202748537064f;
  t265_n53: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t265_n57;
  t265_n54: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t265_n56;
  t265_n55: return 0.03139628469944f;
  t265_n56: return 0.0645182728767395f;
  t265_n57: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t265_n65;
  t265_n58: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t265_n62;
  t265_n59: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t265_n61;
  t265_n60: return -0.021041981875896454f;
  t265_n61: return 0.04773160442709923f;
  t265_n62: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t265_n64;
  t265_n63: return -0.17241835594177246f;
  t265_n64: return 0.034807782620191574f;
  t265_n65: return 0.09934873133897781f;
  t265_n66: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t265_n86;
  t265_n67: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n83;
  t265_n68: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t265_n82;
  t265_n69: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t265_n81;
  t265_n70: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t265_n72;
  t265_n71: return -0.05550771951675415f;
  t265_n72: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n80;
  t265_n73: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n77;
  t265_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n76;
  t265_n75: return -0.37890082597732544f;
  t265_n76: return -0.047910258173942566f;
  t265_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n79;
  t265_n78: return -0.20516334474086761f;
  t265_n79: return 0.12976229190826416f;
  t265_n80: return 0.09517018496990204f;
  t265_n81: return -0.11829746514558792f;
  t265_n82: return 0.028123727068305016f;
  t265_n83: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t265_n85;
  t265_n84: return 0.052692756056785583f;
  t265_n85: return -0.09138132631778717f;
  t265_n86: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t265_n122;
  t265_n87: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t265_n93;
  t265_n88: if (E.getFractionNameInContext() >= 3178788984 /*0.060661762952804565*/) goto t265_n92;
  t265_n89: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t265_n91;
  t265_n90: return 0.006489384453743696f;
  t265_n91: return -0.28353169560432434f;
  t265_n92: return -0.0013184752315282822f;
  t265_n93: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t265_n117;
  t265_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n100;
  t265_n95: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t265_n99;
  t265_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n98;
  t265_n97: return -0.2451322376728058f;
  t265_n98: return -0.007290103007107973f;
  t265_n99: return 0.004911534953862429f;
  t265_n100: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t265_n110;
  t265_n101: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t265_n109;
  t265_n102: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t265_n106;
  t265_n103: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n105;
  t265_n104: return 0.08298780769109726f;
  t265_n105: return -0.11107993870973587f;
  t265_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n108;
  t265_n107: return -0.13872230052947998f;
  t265_n108: return 0.13533388078212738f;
  t265_n109: return -0.015492808073759079f;
  t265_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n116;
  t265_n111: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t265_n115;
  t265_n112: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t265_n114;
  t265_n113: return 0.12848877906799316f;
  t265_n114: return -0.026495525613427162f;
  t265_n115: return 0.1270955502986908f;
  t265_n116: return 0.01340906135737896f;
  t265_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace))) goto t265_n119;
  t265_n118: return 0.031113281846046448f;
  t265_n119: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t265_n121;
  t265_n120: return 0.07333498448133469f;
  t265_n121: return -0.05169610306620598f;
  t265_n122: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t265_n128;
  t265_n123: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t265_n127;
  t265_n124: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n126;
  t265_n125: return 0.06118359416723251f;
  t265_n126: return 0.10787864774465561f;
  t265_n127: return -0.004588827956467867f;
  t265_n128: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t265_n138;
  t265_n129: if (E.getNumReferences() >= 3279650816 /*251.5*/) goto t265_n137;
  t265_n130: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t265_n132;
  t265_n131: return -0.20225483179092407f;
  t265_n132: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t265_n134;
  t265_n133: return -0.10825160145759583f;
  t265_n134: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t265_n136;
  t265_n135: return 0.08224368095397949f;
  t265_n136: return -0.22639217972755432f;
  t265_n137: return 0.1284208595752716f;
  t265_n138: return 0.027988681569695473f;
  t265_n139: return 0.025655221194028854f;
  t265_n140: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t265_n142;
  t265_n141: return 0.1166180893778801f;
  t265_n142: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t265_n144;
  t265_n143: return -0.05330458655953407f;
  t265_n144: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t265_n150;
  t265_n145: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t265_n147;
  t265_n146: return 0.02212965302169323f;
  t265_n147: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t265_n149;
  t265_n148: return 0.039896879345178604f;
  t265_n149: return 0.125589981675148f;
  t265_n150: return -0.5503409504890442f;
  t265_n151: if (E.getFractionNameInContext() >= 3176280194 /*0.0513157919049263*/) goto t265_n153;
  t265_n152: return 0.12113688886165619f;
  t265_n153: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t265_n253;
  t265_n154: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t265_n228;
  t265_n155: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t265_n167;
  t265_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement))) goto t265_n160;
  t265_n157: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t265_n159;
  t265_n158: return 0.08566375076770782f;
  t265_n159: return 0.0036993888206779957f;
  t265_n160: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t265_n166;
  t265_n161: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t265_n163;
  t265_n162: return 0.08104989677667618f;
  t265_n163: if (E.getNumReferences() >= 3356822208 /*152619.0*/) goto t265_n165;
  t265_n164: return 0.007898235693573952f;
  t265_n165: return 0.055048294365406036f;
  t265_n166: return -0.09612740576267242f;
  t265_n167: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t265_n191;
  t265_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t265_n180;
  t265_n169: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n173;
  t265_n170: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t265_n172;
  t265_n171: return -0.0072579155676066875f;
  t265_n172: return 0.0920913890004158f;
  t265_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t265_n179;
  t265_n174: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t265_n178;
  t265_n175: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t265_n177;
  t265_n176: return -0.10251732915639877f;
  t265_n177: return 0.11770213395357132f;
  t265_n178: return 0.011171447113156319f;
  t265_n179: return 0.11573608964681625f;
  t265_n180: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t265_n182;
  t265_n181: return -0.32362881302833557f;
  t265_n182: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n184;
  t265_n183: return -0.3129623234272003f;
  t265_n184: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n186;
  t265_n185: return -0.18637290596961975f;
  t265_n186: if (E.getNumReferences() >= 3317173248 /*5889.5*/) goto t265_n190;
  t265_n187: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t265_n189;
  t265_n188: return 0.1340571790933609f;
  t265_n189: return 0.1094973161816597f;
  t265_n190: return -0.03796956315636635f;
  t265_n191: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t265_n207;
  t265_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t265_n194;
  t265_n193: return -0.12109913676977158f;
  t265_n194: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t265_n196;
  t265_n195: return -0.05918502062559128f;
  t265_n196: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t265_n202;
  t265_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t265_n201;
  t265_n198: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n200;
  t265_n199: return 0.51908278465271f;
  t265_n200: return -0.10189130902290344f;
  t265_n201: return 0.1531650573015213f;
  t265_n202: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n204;
  t265_n203: return -0.1822981834411621f;
  t265_n204: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n206;
  t265_n205: return 0.06729834526777267f;
  t265_n206: return 0.14575877785682678f;
  t265_n207: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t265_n227;
  t265_n208: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t265_n214;
  t265_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n211;
  t265_n210: return -0.03719916567206383f;
  t265_n211: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n213;
  t265_n212: return -0.30243420600891113f;
  t265_n213: return 0.046782102435827255f;
  t265_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n226;
  t265_n215: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t265_n221;
  t265_n216: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t265_n218;
  t265_n217: return -0.027065835893154144f;
  t265_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n220;
  t265_n219: return -0.1096000149846077f;
  t265_n220: return -0.23994003236293793f;
  t265_n221: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t265_n225;
  t265_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n224;
  t265_n223: return -0.10049979388713837f;
  t265_n224: return 0.03382410109043121f;
  t265_n225: return -0.04312863200902939f;
  t265_n226: return 0.01899680681526661f;
  t265_n227: return 0.040277354419231415f;
  t265_n228: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t265_n252;
  t265_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t265_n251;
  t265_n230: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t265_n246;
  t265_n231: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t265_n233;
  t265_n232: return -0.04847463592886925f;
  t265_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t265_n235;
  t265_n234: return -0.0016985007096081972f;
  t265_n235: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t265_n241;
  t265_n236: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t265_n238;
  t265_n237: return -0.1629137396812439f;
  t265_n238: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t265_n240;
  t265_n239: return 0.11650554090738297f;
  t265_n240: return 0.12539373338222504f;
  t265_n241: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t265_n245;
  t265_n242: if (E.getNumReferences() >= 3266052096 /*86.0*/) goto t265_n244;
  t265_n243: return -0.10252116620540619f;
  t265_n244: return 0.11903852969408035f;
  t265_n245: return -0.3016655743122101f;
  t265_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t265_n250;
  t265_n247: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t265_n249;
  t265_n248: return -0.2332841455936432f;
  t265_n249: return -2.406282663345337f;
  t265_n250: return -0.06514030694961548f;
  t265_n251: return 0.0017798524349927902f;
  t265_n252: return 0.10398078709840775f;
  t265_n253: return 0.049116481095552444f;
  t265_n254: return -6.990712427068502e-05f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree266(const Example& E) {
  t266_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t266_n2;
  t266_n1: return -0.00035461224615573883f;
  t266_n2: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t266_n6;
  t266_n3: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t266_n5;
  t266_n4: return -0.02950967289507389f;
  t266_n5: return 0.13488785922527313f;
  t266_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t266_n254;
  t266_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t266_n77;
  t266_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t266_n32;
  t266_n9: if (E.getNumReferences() >= 3309334528 /*3079.0*/) goto t266_n13;
  t266_n10: if (E.getNumReferences() >= 3308642304 /*2910.0*/) goto t266_n12;
  t266_n11: return -0.0015018951380625367f;
  t266_n12: return -0.1706266552209854f;
  t266_n13: if (E.getNumReferences() >= 3312496640 /*3851.0*/) goto t266_n23;
  t266_n14: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t266_n18;
  t266_n15: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t266_n17;
  t266_n16: return 0.10176054388284683f;
  t266_n17: return 0.16961456835269928f;
  t266_n18: if (E.getNumReferences() >= 3312275456 /*3797.0*/) goto t266_n20;
  t266_n19: return 0.02467690221965313f;
  t266_n20: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t266_n22;
  t266_n21: return 0.049531422555446625f;
  t266_n22: return 0.1701415479183197f;
  t266_n23: if (E.getNumReferences() >= 3313147904 /*4010.0*/) goto t266_n31;
  t266_n24: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t266_n30;
  t266_n25: if (E.getNumReferences() >= 3312912384 /*3952.5*/) goto t266_n27;
  t266_n26: return -0.05275628715753555f;
  t266_n27: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t266_n29;
  t266_n28: return -0.15138867497444153f;
  t266_n29: return -0.9461108446121216f;
  t266_n30: return 0.05997980758547783f;
  t266_n31: return 0.010703977197408676f;
  t266_n32: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t266_n38;
  t266_n33: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t266_n37;
  t266_n34: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t266_n36;
  t266_n35: return -0.05091933533549309f;
  t266_n36: return -0.3695704936981201f;
  t266_n37: return -0.024200573563575745f;
  t266_n38: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t266_n76;
  t266_n39: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t266_n63;
  t266_n40: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t266_n48;
  t266_n41: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t266_n45;
  t266_n42: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t266_n44;
  t266_n43: return -0.18491162359714508f;
  t266_n44: return 0.17660711705684662f;
  t266_n45: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t266_n47;
  t266_n46: return 0.019336003810167313f;
  t266_n47: return 0.24475477635860443f;
  t266_n48: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t266_n56;
  t266_n49: if (E.getFractionNameInContext() >= 3208342967 /*0.7321428656578064*/) goto t266_n51;
  t266_n50: return -0.052012667059898376f;
  t266_n51: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t266_n53;
  t266_n52: return -0.0006372497882694006f;
  t266_n53: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t266_n55;
  t266_n54: return 0.13029515743255615f;
  t266_n55: return 0.034104492515325546f;
  t266_n56: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t266_n62;
  t266_n57: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t266_n61;
  t266_n58: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t266_n60;
  t266_n59: return 0.0657803937792778f;
  t266_n60: return 0.18491914868354797f;
  t266_n61: return 0.0423373319208622f;
  t266_n62: return -0.024196622893214226f;
  t266_n63: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t266_n75;
  t266_n64: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t266_n74;
  t266_n65: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t266_n71;
  t266_n66: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t266_n68;
  t266_n67: return 0.039640218019485474f;
  t266_n68: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t266_n70;
  t266_n69: return 0.047653503715991974f;
  t266_n70: return 0.11208966374397278f;
  t266_n71: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t266_n73;
  t266_n72: return -0.01973256841301918f;
  t266_n73: return -0.47009873390197754f;
  t266_n74: return 0.06478282064199448f;
  t266_n75: return 0.01781660132110119f;
  t266_n76: return -0.050646085292100906f;
  t266_n77: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t266_n93;
  t266_n78: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t266_n90;
  t266_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n89;
  t266_n80: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t266_n84;
  t266_n81: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n83;
  t266_n82: return -0.09411772340536118f;
  t266_n83: return -0.025984669104218483f;
  t266_n84: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n86;
  t266_n85: return -0.010445206426084042f;
  t266_n86: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t266_n88;
  t266_n87: return 0.059748902916908264f;
  t266_n88: return -0.03780249133706093f;
  t266_n89: return 0.006935302633792162f;
  t266_n90: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t266_n92;
  t266_n91: return 0.03191021457314491f;
  t266_n92: return -0.066706083714962f;
  t266_n93: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t266_n155;
  t266_n94: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n96;
  t266_n95: return -0.0004588330630213022f;
  t266_n96: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t266_n100;
  t266_n97: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t266_n99;
  t266_n98: return -0.007080096285790205f;
  t266_n99: return 0.004655259661376476f;
  t266_n100: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t266_n112;
  t266_n101: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t266_n105;
  t266_n102: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t266_n104;
  t266_n103: return 0.010340661741793156f;
  t266_n104: return -0.008635423146188259f;
  t266_n105: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t266_n107;
  t266_n106: return 0.00866064615547657f;
  t266_n107: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t266_n109;
  t266_n108: return 0.015368758700788021f;
  t266_n109: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t266_n111;
  t266_n110: return 0.7831994295120239f;
  t266_n111: return -0.10005566477775574f;
  t266_n112: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t266_n130;
  t266_n113: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t266_n119;
  t266_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t266_n118;
  t266_n115: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t266_n117;
  t266_n116: return -0.01753764972090721f;
  t266_n117: return 0.0010981193045154214f;
  t266_n118: return -0.00022965602693147957f;
  t266_n119: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t266_n123;
  t266_n120: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t266_n122;
  t266_n121: return 0.018279241397976875f;
  t266_n122: return -0.05996715649962425f;
  t266_n123: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t266_n125;
  t266_n124: return -0.011494789272546768f;
  t266_n125: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t266_n127;
  t266_n126: return 0.1555338352918625f;
  t266_n127: if (E.getFileProximityDistanceCost() >= 3254255616 /*31.0*/) goto t266_n129;
  t266_n128: return 0.07478144764900208f;
  t266_n129: return -0.20141687989234924f;
  t266_n130: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t266_n154;
  t266_n131: if (E.getNumReferences() >= 3294724096 /*902.0*/) goto t266_n151;
  t266_n132: if (E.getNumReferences() >= 3281125376 /*292.0*/) goto t266_n140;
  t266_n133: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t266_n139;
  t266_n134: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t266_n138;
  t266_n135: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t266_n137;
  t266_n136: return -0.1192231997847557f;
  t266_n137: return 0.10588720440864563f;
  t266_n138: return 0.08828114718198776f;
  t266_n139: return -0.051305707544088364f;
  t266_n140: if (E.getNumReferences() >= 3281747968 /*311.0*/) goto t266_n142;
  t266_n141: return 0.1341726928949356f;
  t266_n142: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t266_n144;
  t266_n143: return 0.08429230749607086f;
  t266_n144: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t266_n150;
  t266_n145: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t266_n149;
  t266_n146: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t266_n148;
  t266_n147: return 0.17334455251693726f;
  t266_n148: return 0.24439579248428345f;
  t266_n149: return 0.06876880675554276f;
  t266_n150: return 0.016885817050933838f;
  t266_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n153;
  t266_n152: return -0.5466465950012207f;
  t266_n153: return -0.028624454513192177f;
  t266_n154: return 0.018973566591739655f;
  t266_n155: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t266_n191;
  t266_n156: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t266_n166;
  t266_n157: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t266_n163;
  t266_n158: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t266_n160;
  t266_n159: return 0.24640949070453644f;
  t266_n160: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t266_n162;
  t266_n161: return -0.02905457466840744f;
  t266_n162: return 0.11137006431818008f;
  t266_n163: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t266_n165;
  t266_n164: return 0.900870144367218f;
  t266_n165: return 0.07667307555675507f;
  t266_n166: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t266_n168;
  t266_n167: return 0.01737036183476448f;
  t266_n168: if (E.getFractionNameInContext() >= 3208762398 /*0.7571429014205933*/) goto t266_n188;
  t266_n169: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t266_n175;
  t266_n170: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t266_n172;
  t266_n171: return -0.1073484942317009f;
  t266_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t266_n174;
  t266_n173: return 0.1368730217218399f;
  t266_n174: return 0.1495118886232376f;
  t266_n175: if (E.getFractionNameInContext() >= 3202118087 /*0.4305555522441864*/) goto t266_n179;
  t266_n176: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t266_n178;
  t266_n177: return -0.09625564515590668f;
  t266_n178: return 0.10613323003053665f;
  t266_n179: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t266_n185;
  t266_n180: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t266_n182;
  t266_n181: return 0.03582865372300148f;
  t266_n182: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t266_n184;
  t266_n183: return 0.13585002720355988f;
  t266_n184: return 0.10579074174165726f;
  t266_n185: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t266_n187;
  t266_n186: return 0.04883616417646408f;
  t266_n187: return -0.11028841882944107f;
  t266_n188: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t266_n190;
  t266_n189: return 0.06219375878572464f;
  t266_n190: return 0.1244945153594017f;
  t266_n191: if (E.getFractionNameInContext() >= 3194920923 /*0.23303167521953583*/) goto t266_n215;
  t266_n192: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t266_n198;
  t266_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t266_n197;
  t266_n194: if (E.getFractionNameInContext() >= 3192764124 /*0.2008928656578064*/) goto t266_n196;
  t266_n195: return 0.2867428660392761f;
  t266_n196: return 0.02070314809679985f;
  t266_n197: return 0.09512558579444885f;
  t266_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Other))) goto t266_n214;
  t266_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t266_n205;
  t266_n200: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t266_n204;
  t266_n201: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t266_n203;
  t266_n202: return -1.214383602142334f;
  t266_n203: return -0.21467748284339905f;
  t266_n204: return -0.03578803315758705f;
  t266_n205: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t266_n207;
  t266_n206: return -0.05156460031867027f;
  t266_n207: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t266_n213;
  t266_n208: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t266_n212;
  t266_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t266_n211;
  t266_n210: return 0.08809565007686615f;
  t266_n211: return 0.09867171198129654f;
  t266_n212: return -0.1644304245710373f;
  t266_n213: return -0.10650499910116196f;
  t266_n214: return 0.1403682380914688f;
  t266_n215: if (E.getFractionNameInContext() >= 3195166743 /*0.23669467866420746*/) goto t266_n217;
  t266_n216: return 0.11639553308486938f;
  t266_n217: if (E.getFractionNameInContext() >= 3202201308 /*0.4330357313156128*/) goto t266_n241;
  t266_n218: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t266_n236;
  t266_n219: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t266_n235;
  t266_n220: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t266_n224;
  t266_n221: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t266_n223;
  t266_n222: return -0.10251113772392273f;
  t266_n223: return 0.13143008947372437f;
  t266_n224: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t266_n232;
  t266_n225: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t266_n231;
  t266_n226: if (E.getNumReferences() >= 3251109888 /*25.0*/) goto t266_n230;
  t266_n227: if (E.getNumReferences() >= 3245342720 /*15.0*/) goto t266_n229;
  t266_n228: return -0.023592276498675346f;
  t266_n229: return -0.2532254755496979f;
  t266_n230: return 0.08462221920490265f;
  t266_n231: return -0.4317587912082672f;
  t266_n232: if (E.getNumReferences() >= 3263299584 /*65.0*/) goto t266_n234;
  t266_n233: return -0.3094911575317383f;
  t266_n234: return 0.13218152523040771f;
  t266_n235: return 0.007088583428412676f;
  t266_n236: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t266_n238;
  t266_n237: return 0.02583692967891693f;
  t266_n238: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t266_n240;
  t266_n239: return 0.011782958172261715f;
  t266_n240: return 0.08967070281505585f;
  t266_n241: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t266_n243;
  t266_n242: return -0.003697415813803673f;
  t266_n243: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t266_n245;
  t266_n244: return 0.00717447092756629f;
  t266_n245: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t266_n253;
  t266_n246: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t266_n250;
  t266_n247: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t266_n249;
  t266_n248: return -0.41408437490463257f;
  t266_n249: return 0.012441588565707207f;
  t266_n250: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t266_n252;
  t266_n251: return 0.04252662509679794f;
  t266_n252: return -0.22067514061927795f;
  t266_n253: return 0.07072515040636063f;
  t266_n254: return 0.0014662266476079822f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree267(const Example& E) {
  t267_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t267_n252;
  t267_n1: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t267_n251;
  t267_n2: if (E.getNumReferences() >= 3364818608 /*292981.5*/) goto t267_n240;
  t267_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t267_n167;
  t267_n4: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t267_n128;
  t267_n5: if (E.getNumReferences() >= 3292274688 /*752.5*/) goto t267_n127;
  t267_n6: if (E.getNumReferences() >= 3292110848 /*742.5*/) goto t267_n120;
  t267_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t267_n87;
  t267_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t267_n10;
  t267_n9: return -0.0278007909655571f;
  t267_n10: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t267_n86;
  t267_n11: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t267_n13;
  t267_n12: return -0.0008784137899056077f;
  t267_n13: if (E.getNumReferences() >= 3272638464 /*144.5*/) goto t267_n27;
  t267_n14: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t267_n24;
  t267_n15: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t267_n23;
  t267_n16: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t267_n18;
  t267_n17: return -0.009664658457040787f;
  t267_n18: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n20;
  t267_n19: return 0.008587577380239964f;
  t267_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t267_n22;
  t267_n21: return -0.046054087579250336f;
  t267_n22: return 0.03803030028939247f;
  t267_n23: return 0.009312614798545837f;
  t267_n24: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Statement))) goto t267_n26;
  t267_n25: return 0.00801964569836855f;
  t267_n26: return 0.022986093536019325f;
  t267_n27: if (E.getNumReferences() >= 3273359360 /*155.5*/) goto t267_n33;
  t267_n28: if (E.getNumReferences() >= 3273293824 /*154.5*/) goto t267_n30;
  t267_n29: return 0.0862959623336792f;
  t267_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t267_n32;
  t267_n31: return 0.15039373934268951f;
  t267_n32: return 0.15555457770824432f;
  t267_n33: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t267_n57;
  t267_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t267_n36;
  t267_n35: return -0.030943335965275764f;
  t267_n36: if (E.getNumReferences() >= 3281764352 /*311.5*/) goto t267_n56;
  t267_n37: if (E.getNumReferences() >= 3280224256 /*264.5*/) goto t267_n51;
  t267_n38: if (E.getNumReferences() >= 3277029376 /*211.5*/) goto t267_n50;
  t267_n39: if (E.getNumReferences() >= 3276963840 /*210.5*/) goto t267_n49;
  t267_n40: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t267_n48;
  t267_n41: if (E.getNumReferences() >= 3276210176 /*199.0*/) goto t267_n47;
  t267_n42: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t267_n46;
  t267_n43: if (E.getNumReferences() >= 3275554816 /*189.0*/) goto t267_n45;
  t267_n44: return -0.09699587523937225f;
  t267_n45: return 0.16974952816963196f;
  t267_n46: return 0.3613099753856659f;
  t267_n47: return 0.10526739805936813f;
  t267_n48: return -0.08777671307325363f;
  t267_n49: return 0.17137083411216736f;
  t267_n50: return -0.08078823983669281f;
  t267_n51: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t267_n53;
  t267_n52: return 0.17573219537734985f;
  t267_n53: if (E.getNumReferences() >= 3281207296 /*294.5*/) goto t267_n55;
  t267_n54: return -0.01550076249986887f;
  t267_n55: return 0.1885848045349121f;
  t267_n56: return -0.016269654035568237f;
  t267_n57: if (E.getNumReferences() >= 3291471872 /*703.5*/) goto t267_n83;
  t267_n58: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t267_n72;
  t267_n59: if (E.getNumReferences() >= 3288498176 /*522.0*/) goto t267_n69;
  t267_n60: if (E.getNumReferences() >= 3286974464 /*470.5*/) goto t267_n62;
  t267_n61: return 0.03873191773891449f;
  t267_n62: if (E.getNumReferences() >= 3287629824 /*490.5*/) goto t267_n68;
  t267_n63: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t267_n65;
  t267_n64: return 0.062703937292099f;
  t267_n65: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n67;
  t267_n66: return 0.2100578099489212f;
  t267_n67: return -0.10960228741168976f;
  t267_n68: return 0.02957426942884922f;
  t267_n69: if (E.getNumReferences() >= 3288997888 /*552.5*/) goto t267_n71;
  t267_n70: return -0.2525935471057892f;
  t267_n71: return 0.012475301511585712f;
  t267_n72: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t267_n82;
  t267_n73: if (E.getNumReferences() >= 3281682432 /*309.0*/) goto t267_n79;
  t267_n74: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n76;
  t267_n75: return -0.029747502878308296f;
  t267_n76: if (E.getNumReferences() >= 3278372864 /*232.0*/) goto t267_n78;
  t267_n77: return 0.11866056174039841f;
  t267_n78: return -0.10615997016429901f;
  t267_n79: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n81;
  t267_n80: return 0.21304866671562195f;
  t267_n81: return -0.08427340537309647f;
  t267_n82: return -0.02579346112906933f;
  t267_n83: if (E.getNumReferences() >= 3291586560 /*710.5*/) goto t267_n85;
  t267_n84: return 0.2644121050834656f;
  t267_n85: return -0.03867759555578232f;
  t267_n86: return -0.011052320711314678f;
  t267_n87: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t267_n89;
  t267_n88: return 0.0023741016630083323f;
  t267_n89: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t267_n91;
  t267_n90: return -0.014197162352502346f;
  t267_n91: if (E.getNumReferences() >= 3265462272 /*81.5*/) goto t267_n95;
  t267_n92: if (E.getFractionNameInContext() >= 3207923536 /*0.7071428298950195*/) goto t267_n94;
  t267_n93: return 0.016293605789542198f;
  t267_n94: return 0.03676450252532959f;
  t267_n95: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t267_n99;
  t267_n96: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t267_n98;
  t267_n97: return 0.24818198382854462f;
  t267_n98: return 0.0071958452463150024f;
  t267_n99: if (E.getNumReferences() >= 3272179712 /*137.5*/) goto t267_n117;
  t267_n100: if (E.getNumReferences() >= 3271786496 /*131.5*/) goto t267_n114;
  t267_n101: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n103;
  t267_n102: return 0.02986641228199005f;
  t267_n103: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t267_n105;
  t267_n104: return 0.10937419533729553f;
  t267_n105: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t267_n113;
  t267_n106: if (E.getNumReferences() >= 3267690496 /*98.5*/) goto t267_n112;
  t267_n107: if (E.getNumReferences() >= 3266969600 /*93.0*/) goto t267_n111;
  t267_n108: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t267_n110;
  t267_n109: return 0.5608999729156494f;
  t267_n110: return -0.3339700400829315f;
  t267_n111: return -0.1237817332148552f;
  t267_n112: return 0.5134268999099731f;
  t267_n113: return 0.05316148325800896f;
  t267_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t267_n116;
  t267_n115: return 0.21786512434482574f;
  t267_n116: return 0.07017318904399872f;
  t267_n117: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t267_n119;
  t267_n118: return -0.22660376131534576f;
  t267_n119: return 0.018511129543185234f;
  t267_n120: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t267_n126;
  t267_n121: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t267_n125;
  t267_n122: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t267_n124;
  t267_n123: return 0.027636218816041946f;
  t267_n124: return 0.2052839994430542f;
  t267_n125: return 0.017505386844277382f;
  t267_n126: return 0.37012264132499695f;
  t267_n127: return -0.0112198106944561f;
  t267_n128: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t267_n130;
  t267_n129: return -0.0036613030824810266f;
  t267_n130: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t267_n136;
  t267_n131: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t267_n133;
  t267_n132: return -0.004303271882236004f;
  t267_n133: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t267_n135;
  t267_n134: return 0.11010821908712387f;
  t267_n135: return -0.06747273355722427f;
  t267_n136: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t267_n166;
  t267_n137: if (E.getNumReferences() >= 3271819264 /*132.0*/) goto t267_n165;
  t267_n138: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t267_n162;
  t267_n139: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t267_n141;
  t267_n140: return -0.022897230461239815f;
  t267_n141: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t267_n161;
  t267_n142: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t267_n148;
  t267_n143: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n147;
  t267_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t267_n146;
  t267_n145: return -0.12220030277967453f;
  t267_n146: return 0.05935875326395035f;
  t267_n147: return -0.022675270214676857f;
  t267_n148: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t267_n150;
  t267_n149: return 0.033951643854379654f;
  t267_n150: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t267_n154;
  t267_n151: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t267_n153;
  t267_n152: return 0.1097591370344162f;
  t267_n153: return 0.06845619529485703f;
  t267_n154: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t267_n156;
  t267_n155: return 0.012099376879632473f;
  t267_n156: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t267_n160;
  t267_n157: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t267_n159;
  t267_n158: return 0.10858093947172165f;
  t267_n159: return 0.07169897854328156f;
  t267_n160: return 0.023655293509364128f;
  t267_n161: return -0.012412501499056816f;
  t267_n162: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t267_n164;
  t267_n163: return 0.08311536908149719f;
  t267_n164: return -0.046333663165569305f;
  t267_n165: return 0.12444180250167847f;
  t267_n166: return -0.002890028292313218f;
  t267_n167: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t267_n221;
  t267_n168: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t267_n208;
  t267_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t267_n201;
  t267_n170: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t267_n192;
  t267_n171: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t267_n181;
  t267_n172: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t267_n178;
  t267_n173: if (E.getNumReferences() >= 3289694208 /*595.0*/) goto t267_n177;
  t267_n174: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t267_n176;
  t267_n175: return -0.07029801607131958f;
  t267_n176: return 0.061345748603343964f;
  t267_n177: return 0.09891100227832794f;
  t267_n178: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t267_n180;
  t267_n179: return -0.00456261308863759f;
  t267_n180: return -0.6021850109100342f;
  t267_n181: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t267_n189;
  t267_n182: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t267_n186;
  t267_n183: if (E.getFractionNameInContext() >= 3179881618 /*0.0669642835855484*/) goto t267_n185;
  t267_n184: return 0.06496426463127136f;
  t267_n185: return -0.004520495422184467f;
  t267_n186: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t267_n188;
  t267_n187: return -0.053482286632061005f;
  t267_n188: return -1.3367421627044678f;
  t267_n189: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t267_n191;
  t267_n190: return -0.07965582609176636f;
  t267_n191: return -0.01190702710300684f;
  t267_n192: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t267_n200;
  t267_n193: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t267_n197;
  t267_n194: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t267_n196;
  t267_n195: return -0.016231611371040344f;
  t267_n196: return 0.04992561414837837f;
  t267_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t267_n199;
  t267_n198: return 0.10244908928871155f;
  t267_n199: return 0.15731379389762878f;
  t267_n200: return -0.009839381091296673f;
  t267_n201: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t267_n203;
  t267_n202: return 0.01109375897794962f;
  t267_n203: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t267_n207;
  t267_n204: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t267_n206;
  t267_n205: return -0.07696346193552017f;
  t267_n206: return 0.0015749548329040408f;
  t267_n207: return -0.006526482757180929f;
  t267_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t267_n216;
  t267_n209: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t267_n215;
  t267_n210: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t267_n214;
  t267_n211: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t267_n213;
  t267_n212: return -0.3320145905017853f;
  t267_n213: return 0.02894928865134716f;
  t267_n214: return 0.017299525439739227f;
  t267_n215: return 0.010136555880308151f;
  t267_n216: if (E.getNumReferences() >= 3308576768 /*2894.0*/) goto t267_n218;
  t267_n217: return 0.0377357117831707f;
  t267_n218: if (E.getNumReferences() >= 3322144768 /*8442.0*/) goto t267_n220;
  t267_n219: return 0.16791664063930511f;
  t267_n220: return -0.08932037651538849f;
  t267_n221: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t267_n223;
  t267_n222: return 0.038777098059654236f;
  t267_n223: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t267_n239;
  t267_n224: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t267_n226;
  t267_n225: return 0.01232698280364275f;
  t267_n226: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t267_n232;
  t267_n227: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t267_n231;
  t267_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t267_n230;
  t267_n229: return 0.0644262433052063f;
  t267_n230: return 0.11330650001764297f;
  t267_n231: return 0.02235167846083641f;
  t267_n232: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t267_n236;
  t267_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t267_n235;
  t267_n234: return -0.42853763699531555f;
  t267_n235: return 0.06794735789299011f;
  t267_n236: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t267_n238;
  t267_n237: return 0.09920833259820938f;
  t267_n238: return 0.3668237328529358f;
  t267_n239: return -0.014820441603660583f;
  t267_n240: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t267_n250;
  t267_n241: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t267_n243;
  t267_n242: return 0.0085873082280159f;
  t267_n243: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t267_n247;
  t267_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t267_n246;
  t267_n245: return 0.0937328189611435f;
  t267_n246: return 0.1360577642917633f;
  t267_n247: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t267_n249;
  t267_n248: return -0.41940632462501526f;
  t267_n249: return 0.1143195852637291f;
  t267_n250: return 0.006768862251192331f;
  t267_n251: return -0.00033240378252230585f;
  t267_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t267_n254;
  t267_n253: return 0.0008662227774038911f;
  t267_n254: return 0.0032579349353909492f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree268(const Example& E) {
  t268_n0: if (E.getNumReferences() >= 3328217600 /*14372.5*/) goto t268_n10;
  t268_n1: if (E.getNumReferences() >= 3328027648 /*14187.0*/) goto t268_n7;
  t268_n2: if (E.getNumReferences() >= 3327951360 /*14112.5*/) goto t268_n4;
  t268_n3: return -7.190542237367481e-05f;
  t268_n4: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t268_n6;
  t268_n5: return 0.10820979624986649f;
  t268_n6: return 0.10378208011388779f;
  t268_n7: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t268_n9;
  t268_n8: return -0.12321489304304123f;
  t268_n9: return -0.155925914645195f;
  t268_n10: if (E.getNumReferences() >= 3329021952 /*15158.0*/) goto t268_n64;
  t268_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t268_n51;
  t268_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t268_n32;
  t268_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t268_n25;
  t268_n14: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t268_n16;
  t268_n15: return -0.08399903774261475f;
  t268_n16: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t268_n24;
  t268_n17: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t268_n23;
  t268_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n20;
  t268_n19: return -0.0310211144387722f;
  t268_n20: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t268_n22;
  t268_n21: return 0.22651907801628113f;
  t268_n22: return 0.19047562777996063f;
  t268_n23: return -0.05554979294538498f;
  t268_n24: return -0.12646019458770752f;
  t268_n25: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t268_n31;
  t268_n26: if (E.getNumReferences() >= 3328450048 /*14599.5*/) goto t268_n30;
  t268_n27: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t268_n29;
  t268_n28: return 0.19175168871879578f;
  t268_n29: return 0.10416190326213837f;
  t268_n30: return -0.042418427765369415f;
  t268_n31: return -0.11345499008893967f;
  t268_n32: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t268_n42;
  t268_n33: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro))) goto t268_n35;
  t268_n34: return -0.021130334585905075f;
  t268_n35: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n41;
  t268_n36: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t268_n40;
  t268_n37: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t268_n39;
  t268_n38: return -0.0008033037302084267f;
  t268_n39: return 0.10722309350967407f;
  t268_n40: return 0.11175715178251266f;
  t268_n41: return 0.03138389065861702f;
  t268_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t268_n50;
  t268_n43: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t268_n45;
  t268_n44: return -0.3400067090988159f;
  t268_n45: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t268_n49;
  t268_n46: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t268_n48;
  t268_n47: return 0.19205662608146667f;
  t268_n48: return 0.18695250153541565f;
  t268_n49: return -0.06564934551715851f;
  t268_n50: return 0.038539811968803406f;
  t268_n51: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t268_n59;
  t268_n52: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t268_n54;
  t268_n53: return -0.1595916897058487f;
  t268_n54: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n56;
  t268_n55: return 0.017108600586652756f;
  t268_n56: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t268_n58;
  t268_n57: return 0.09670031815767288f;
  t268_n58: return 0.017626211047172546f;
  t268_n59: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t268_n61;
  t268_n60: return 0.053846586495637894f;
  t268_n61: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t268_n63;
  t268_n62: return 0.10791436582803726f;
  t268_n63: return 0.1118321344256401f;
  t268_n64: if (E.getNumReferences() >= 3329553408 /*15677.0*/) goto t268_n70;
  t268_n65: if (E.getNumReferences() >= 3329258496 /*15389.0*/) goto t268_n69;
  t268_n66: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t268_n68;
  t268_n67: return -0.025660252198576927f;
  t268_n68: return 0.07393573224544525f;
  t268_n69: return -0.11715395003557205f;
  t268_n70: if (E.getNumReferences() >= 3330013184 /*16126.0*/) goto t268_n104;
  t268_n71: if (E.getNumReferences() >= 3329833472 /*15950.5*/) goto t268_n79;
  t268_n72: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t268_n78;
  t268_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t268_n77;
  t268_n74: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t268_n76;
  t268_n75: return -0.011608892120420933f;
  t268_n76: return 0.1941222846508026f;
  t268_n77: return 0.07573805004358292f;
  t268_n78: return 0.1406518965959549f;
  t268_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t268_n101;
  t268_n80: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t268_n82;
  t268_n81: return -0.009490490891039371f;
  t268_n82: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t268_n100;
  t268_n83: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t268_n93;
  t268_n84: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t268_n90;
  t268_n85: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t268_n87;
  t268_n86: return 0.05990195274353027f;
  t268_n87: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t268_n89;
  t268_n88: return -0.10192440450191498f;
  t268_n89: return 0.23054258525371552f;
  t268_n90: if (E.getFractionNameInContext() >= 3189721589 /*0.15555556118488312*/) goto t268_n92;
  t268_n91: return 0.056854359805583954f;
  t268_n92: return 0.09985187649726868f;
  t268_n93: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t268_n97;
  t268_n94: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t268_n96;
  t268_n95: return 0.12578728795051575f;
  t268_n96: return 0.07482133060693741f;
  t268_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n99;
  t268_n98: return -0.10097275674343109f;
  t268_n99: return 0.11480019986629486f;
  t268_n100: return 0.01813952811062336f;
  t268_n101: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t268_n103;
  t268_n102: return 0.10554957389831543f;
  t268_n103: return 0.13020268082618713f;
  t268_n104: if (E.getNumReferences() >= 3330058752 /*16170.5*/) goto t268_n108;
  t268_n105: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t268_n107;
  t268_n106: return -0.17965173721313477f;
  t268_n107: return -0.31975406408309937f;
  t268_n108: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t268_n114;
  t268_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n113;
  t268_n110: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t268_n112;
  t268_n111: return -0.05261294171214104f;
  t268_n112: return 0.0853932648897171f;
  t268_n113: return -0.007669732440263033f;
  t268_n114: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t268_n186;
  t268_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t268_n167;
  t268_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n118;
  t268_n117: return -0.018538091331720352f;
  t268_n118: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t268_n122;
  t268_n119: if (E.getNumReferences() >= 3334015488 /*23685.0*/) goto t268_n121;
  t268_n120: return -0.010615679435431957f;
  t268_n121: return -0.20473426580429077f;
  t268_n122: if (E.getNumReferences() >= 3339162368 /*34707.0*/) goto t268_n166;
  t268_n123: if (E.getNumReferences() >= 3338968320 /*33949.0*/) goto t268_n165;
  t268_n124: if (E.getNumReferences() >= 3338478336 /*32401.5*/) goto t268_n156;
  t268_n125: if (E.getNumReferences() >= 3337362176 /*30221.5*/) goto t268_n153;
  t268_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n128;
  t268_n127: return -0.03367885574698448f;
  t268_n128: if (E.getNumReferences() >= 3336193792 /*27939.5*/) goto t268_n134;
  t268_n129: if (E.getNumReferences() >= 3334345728 /*24330.0*/) goto t268_n133;
  t268_n130: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t268_n132;
  t268_n131: return 0.0076113734394311905f;
  t268_n132: return 0.07735341042280197f;
  t268_n133: return -0.048079945147037506f;
  t268_n134: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t268_n140;
  t268_n135: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n137;
  t268_n136: return 0.1076442301273346f;
  t268_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t268_n139;
  t268_n138: return 0.17178265750408173f;
  t268_n139: return 0.12108982354402542f;
  t268_n140: if (E.getNumReferences() >= 3336316928 /*28180.0*/) goto t268_n142;
  t268_n141: return 0.1319737434387207f;
  t268_n142: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n144;
  t268_n143: return 0.0117184454575181f;
  t268_n144: if (E.getNumReferences() >= 3336347136 /*28239.0*/) goto t268_n146;
  t268_n145: return -0.10540648549795151f;
  t268_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t268_n150;
  t268_n147: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t268_n149;
  t268_n148: return 0.017528071999549866f;
  t268_n149: return 0.11095379292964935f;
  t268_n150: if (E.getNumReferences() >= 3336505344 /*28548.0*/) goto t268_n152;
  t268_n151: return 0.186709925532341f;
  t268_n152: return -0.10018864274024963f;
  t268_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t268_n155;
  t268_n154: return -0.12426794320344925f;
  t268_n155: return 0.020444542169570923f;
  t268_n156: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t268_n164;
  t268_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t268_n159;
  t268_n158: return 0.07299959659576416f;
  t268_n159: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t268_n163;
  t268_n160: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t268_n162;
  t268_n161: return 0.08251935988664627f;
  t268_n162: return 0.17645592987537384f;
  t268_n163: return 0.06724836677312851f;
  t268_n164: return 0.017646435648202896f;
  t268_n165: return 0.09605545550584793f;
  t268_n166: return 0.002795967971906066f;
  t268_n167: if (E.getNumReferences() >= 3331096576 /*17984.0*/) goto t268_n185;
  t268_n168: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n182;
  t268_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t268_n175;
  t268_n170: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t268_n174;
  t268_n171: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t268_n173;
  t268_n172: return 0.008427934721112251f;
  t268_n173: return 0.15574361383914948f;
  t268_n174: return 0.056879498064517975f;
  t268_n175: if (E.getNumReferences() >= 3330455552 /*16732.0*/) goto t268_n177;
  t268_n176: return 0.0018303488614037633f;
  t268_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t268_n179;
  t268_n178: return 0.1083819791674614f;
  t268_n179: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t268_n181;
  t268_n180: return 0.12028756737709045f;
  t268_n181: return 0.09224442392587662f;
  t268_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t268_n184;
  t268_n183: return 0.09362327307462692f;
  t268_n184: return 0.12791964411735535f;
  t268_n185: return 0.006967628840357065f;
  t268_n186: if (E.getNumReferences() >= 3332258560 /*20253.5*/) goto t268_n188;
  t268_n187: return -0.04982319474220276f;
  t268_n188: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t268_n236;
  t268_n189: if (E.getNumReferences() >= 3341362816 /*43302.5*/) goto t268_n203;
  t268_n190: if (E.getNumReferences() >= 3336316928 /*28180.0*/) goto t268_n202;
  t268_n191: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t268_n201;
  t268_n192: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t268_n194;
  t268_n193: return -0.029975933954119682f;
  t268_n194: if (E.getNumReferences() >= 3334782720 /*25183.5*/) goto t268_n196;
  t268_n195: return -0.0854903981089592f;
  t268_n196: if (E.getNumReferences() >= 3335478016 /*26541.5*/) goto t268_n200;
  t268_n197: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t268_n199;
  t268_n198: return 0.109901562333107f;
  t268_n199: return 0.07743952423334122f;
  t268_n200: return 0.020584821701049805f;
  t268_n201: return 0.027763957157731056f;
  t268_n202: return -0.0253477580845356f;
  t268_n203: if (E.getNumReferences() >= 3341797376 /*45000.0*/) goto t268_n223;
  t268_n204: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t268_n210;
  t268_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t268_n207;
  t268_n206: return 0.06167628616094589f;
  t268_n207: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t268_n209;
  t268_n208: return 0.1168675497174263f;
  t268_n209: return 0.06036591902375221f;
  t268_n210: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t268_n218;
  t268_n211: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t268_n217;
  t268_n212: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t268_n214;
  t268_n213: return 0.12983128428459167f;
  t268_n214: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t268_n216;
  t268_n215: return 0.11624347418546677f;
  t268_n216: return 0.034168317914009094f;
  t268_n217: return -0.369306355714798f;
  t268_n218: if (E.getNumReferences() >= 3341718784 /*44693.0*/) goto t268_n220;
  t268_n219: return -0.19705070555210114f;
  t268_n220: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t268_n222;
  t268_n221: return 0.0743224248290062f;
  t268_n222: return 0.10620104521512985f;
  t268_n223: if (E.getNumReferences() >= 3343638400 /*52191.5*/) goto t268_n235;
  t268_n224: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t268_n230;
  t268_n225: if (E.getNumReferences() >= 3342384384 /*47293.0*/) goto t268_n229;
  t268_n226: if (E.getNumReferences() >= 3342343040 /*47131.5*/) goto t268_n228;
  t268_n227: return 0.004769977182149887f;
  t268_n228: return 0.06950643658638f;
  t268_n229: return -0.08466453105211258f;
  t268_n230: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t268_n232;
  t268_n231: return -0.1828698366880417f;
  t268_n232: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t268_n234;
  t268_n233: return 0.15010987222194672f;
  t268_n234: return 0.13885067403316498f;
  t268_n235: return 0.0661521777510643f;
  t268_n236: if (E.getNumReferences() >= 3345144064 /*58073.0*/) goto t268_n238;
  t268_n237: return -0.14566488564014435f;
  t268_n238: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t268_n240;
  t268_n239: return -0.028098614886403084f;
  t268_n240: if (E.getNumReferences() >= 3347118592 /*66036.0*/) goto t268_n248;
  t268_n241: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t268_n245;
  t268_n242: if (E.getNumReferences() >= 3346297984 /*62580.5*/) goto t268_n244;
  t268_n243: return 0.020229298621416092f;
  t268_n244: return -0.07050791382789612f;
  t268_n245: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t268_n247;
  t268_n246: return 0.03579706326127052f;
  t268_n247: return 0.09244557470083237f;
  t268_n248: if (E.getNumReferences() >= 3360096736 /*203783.5*/) goto t268_n250;
  t268_n249: return -0.013734758831560612f;
  t268_n250: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t268_n254;
  t268_n251: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t268_n253;
  t268_n252: return 0.024662146344780922f;
  t268_n253: return 0.07604683190584183f;
  t268_n254: return 0.007683295290917158f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree269(const Example& E) {
  t269_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t269_n106;
  t269_n1: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t269_n3;
  t269_n2: return 0.0003309750172775239f;
  t269_n3: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t269_n105;
  t269_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n8;
  t269_n5: if (E.getNumReferences() >= 3382768696 /*1318535.0*/) goto t269_n7;
  t269_n6: return -0.010802354663610458f;
  t269_n7: return 0.1098007783293724f;
  t269_n8: if (E.getNumReferences() >= 3382768696 /*1318535.0*/) goto t269_n102;
  t269_n9: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t269_n101;
  t269_n10: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t269_n12;
  t269_n11: return -0.0008002194808796048f;
  t269_n12: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t269_n22;
  t269_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace))) goto t269_n21;
  t269_n14: if (E.getSymbolScopeDistanceCost() >= 3246915584 /*17.0*/) goto t269_n18;
  t269_n15: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t269_n17;
  t269_n16: return 0.1371854990720749f;
  t269_n17: return -0.11185935884714127f;
  t269_n18: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t269_n20;
  t269_n19: return -0.015236211009323597f;
  t269_n20: return -0.11567345261573792f;
  t269_n21: return 0.06656750291585922f;
  t269_n22: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t269_n32;
  t269_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t269_n25;
  t269_n24: return 0.007034088019281626f;
  t269_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t269_n27;
  t269_n26: return -0.07528083771467209f;
  t269_n27: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t269_n31;
  t269_n28: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t269_n30;
  t269_n29: return 0.033849600702524185f;
  t269_n30: return -0.10411010682582855f;
  t269_n31: return 0.007097084075212479f;
  t269_n32: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t269_n60;
  t269_n33: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t269_n43;
  t269_n34: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t269_n42;
  t269_n35: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t269_n41;
  t269_n36: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t269_n40;
  t269_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t269_n39;
  t269_n38: return -0.10349395126104355f;
  t269_n39: return 0.48389896750450134f;
  t269_n40: return -0.10002314299345016f;
  t269_n41: return -0.2211551070213318f;
  t269_n42: return 0.19737812876701355f;
  t269_n43: if (E.getFractionNameInContext() >= 3174997342 /*0.04653679579496384*/) goto t269_n45;
  t269_n44: return -0.02188705839216709f;
  t269_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t269_n59;
  t269_n46: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t269_n48;
  t269_n47: return -0.004287227522581816f;
  t269_n48: if (E.getFractionNameInContext() >= 3197626751 /*0.2967033088207245*/) goto t269_n50;
  t269_n49: return 0.14393271505832672f;
  t269_n50: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t269_n58;
  t269_n51: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t269_n57;
  t269_n52: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t269_n56;
  t269_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n55;
  t269_n54: return -0.10478320717811584f;
  t269_n55: return 0.11180570721626282f;
  t269_n56: return -0.1341480165719986f;
  t269_n57: return -0.10960186272859573f;
  t269_n58: return -0.12108323723077774f;
  t269_n59: return 0.027068575844168663f;
  t269_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n62;
  t269_n61: return -0.002405688865110278f;
  t269_n62: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t269_n98;
  t269_n63: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t269_n67;
  t269_n64: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t269_n66;
  t269_n65: return 0.007548295892775059f;
  t269_n66: return 0.0649118721485138f;
  t269_n67: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t269_n97;
  t269_n68: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t269_n88;
  t269_n69: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t269_n87;
  t269_n70: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t269_n86;
  t269_n71: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t269_n85;
  t269_n72: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t269_n82;
  t269_n73: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t269_n81;
  t269_n74: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t269_n80;
  t269_n75: if (E.getSymbolScopeDistanceCost() >= 3247964160 /*19.0*/) goto t269_n79;
  t269_n76: if (E.getFractionNameInContext() >= 3189348762 /*0.15000000596046448*/) goto t269_n78;
  t269_n77: return 0.20907603204250336f;
  t269_n78: return 0.22376617789268494f;
  t269_n79: return 0.02559938281774521f;
  t269_n80: return 0.09422367066144943f;
  t269_n81: return -0.03998567909002304f;
  t269_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t269_n84;
  t269_n83: return 0.02926800772547722f;
  t269_n84: return 0.03760063275694847f;
  t269_n85: return -0.05672251805663109f;
  t269_n86: return 0.03405416011810303f;
  t269_n87: return -0.10249985009431839f;
  t269_n88: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t269_n90;
  t269_n89: return -0.10971935093402863f;
  t269_n90: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t269_n94;
  t269_n91: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t269_n93;
  t269_n92: return -0.1043776348233223f;
  t269_n93: return 0.11609821021556854f;
  t269_n94: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t269_n96;
  t269_n95: return 0.040020521730184555f;
  t269_n96: return -0.12786021828651428f;
  t269_n97: return -0.013639305718243122f;
  t269_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator))) goto t269_n100;
  t269_n99: return -0.02249937132000923f;
  t269_n100: return 0.023512640967965126f;
  t269_n101: return -0.1136835440993309f;
  t269_n102: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t269_n104;
  t269_n103: return -0.5120421051979065f;
  t269_n104: return -0.002288749674335122f;
  t269_n105: return -0.014083604328334332f;
  t269_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n254;
  t269_n107: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t269_n241;
  t269_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t269_n230;
  t269_n109: if (E.getNumReferences() >= 3268214784 /*102.5*/) goto t269_n125;
  t269_n110: if (E.getNumReferences() >= 3266772992 /*91.5*/) goto t269_n112;
  t269_n111: return -0.0002242497430415824f;
  t269_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t269_n118;
  t269_n113: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t269_n117;
  t269_n114: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t269_n116;
  t269_n115: return -0.22557051479816437f;
  t269_n116: return -0.04664482921361923f;
  t269_n117: return 0.02576266974210739f;
  t269_n118: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t269_n122;
  t269_n119: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t269_n121;
  t269_n120: return 0.11044101417064667f;
  t269_n121: return -0.025117971003055573f;
  t269_n122: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t269_n124;
  t269_n123: return 0.12999209761619568f;
  t269_n124: return 0.07380350679159164f;
  t269_n125: if (E.getNumReferences() >= 3269918720 /*115.5*/) goto t269_n141;
  t269_n126: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t269_n130;
  t269_n127: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n129;
  t269_n128: return -0.3928969204425812f;
  t269_n129: return -0.017123332247138023f;
  t269_n130: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t269_n140;
  t269_n131: if (E.getNumReferences() >= 3268739072 /*106.5*/) goto t269_n139;
  t269_n132: if (E.getNumReferences() >= 3268476928 /*104.5*/) goto t269_n136;
  t269_n133: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t269_n135;
  t269_n134: return -1.2181775569915771f;
  t269_n135: return 0.10996689647436142f;
  t269_n136: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n138;
  t269_n137: return -0.1639670580625534f;
  t269_n138: return 0.10951811075210571f;
  t269_n139: return -0.32976678013801575f;
  t269_n140: return -0.0317179411649704f;
  t269_n141: if (E.getNumReferences() >= 3270180864 /*117.5*/) goto t269_n143;
  t269_n142: return 0.05523078143596649f;
  t269_n143: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t269_n187;
  t269_n144: if (E.getNumReferences() >= 3287826432 /*496.5*/) goto t269_n180;
  t269_n145: if (E.getNumReferences() >= 3285139456 /*414.5*/) goto t269_n175;
  t269_n146: if (E.getNumReferences() >= 3275194368 /*183.5*/) goto t269_n154;
  t269_n147: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t269_n149;
  t269_n148: return -0.027136793360114098f;
  t269_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n151;
  t269_n150: return -0.20632484555244446f;
  t269_n151: if (E.getFractionNameInContext() >= 3184451170 /*0.10101009905338287*/) goto t269_n153;
  t269_n152: return 0.1115151196718216f;
  t269_n153: return 0.015080546028912067f;
  t269_n154: if (E.getNumReferences() >= 3277160448 /*213.5*/) goto t269_n164;
  t269_n155: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t269_n163;
  t269_n156: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n160;
  t269_n157: if (E.getNumReferences() >= 3275882496 /*194.0*/) goto t269_n159;
  t269_n158: return -0.5511264801025391f;
  t269_n159: return 0.07073598355054855f;
  t269_n160: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t269_n162;
  t269_n161: return 0.10489661246538162f;
  t269_n162: return 0.055600814521312714f;
  t269_n163: return -0.04334450885653496f;
  t269_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t269_n174;
  t269_n165: if (E.getNumReferences() >= 3284484096 /*394.5*/) goto t269_n173;
  t269_n166: if (E.getNumReferences() >= 3283697664 /*370.5*/) goto t269_n170;
  t269_n167: if (E.getNumReferences() >= 3283566592 /*366.5*/) goto t269_n169;
  t269_n168: return -0.0214750524610281f;
  t269_n169: return 0.19435983896255493f;
  t269_n170: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t269_n172;
  t269_n171: return -0.21959106624126434f;
  t269_n172: return 0.0905425176024437f;
  t269_n173: return 0.06684733182191849f;
  t269_n174: return 0.03981500864028931f;
  t269_n175: if (E.getNumReferences() >= 3285204992 /*416.5*/) goto t269_n177;
  t269_n176: return -1.014699935913086f;
  t269_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t269_n179;
  t269_n178: return -1.3365211486816406f;
  t269_n179: return -0.04207262396812439f;
  t269_n180: if (E.getNumReferences() >= 3288702976 /*534.5*/) goto t269_n186;
  t269_n181: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t269_n183;
  t269_n182: return 0.06767022609710693f;
  t269_n183: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t269_n185;
  t269_n184: return 0.12583622336387634f;
  t269_n185: return 0.0973203256726265f;
  t269_n186: return -0.003930029459297657f;
  t269_n187: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t269_n209;
  t269_n188: if (E.getNumReferences() >= 3273228288 /*153.5*/) goto t269_n190;
  t269_n189: return 0.010758774355053902f;
  t269_n190: if (E.getNumReferences() >= 3283501056 /*364.5*/) goto t269_n204;
  t269_n191: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t269_n197;
  t269_n192: if (E.getNumReferences() >= 3283369984 /*360.5*/) goto t269_n196;
  t269_n193: if (E.getNumReferences() >= 3273359360 /*155.5*/) goto t269_n195;
  t269_n194: return -0.5838568210601807f;
  t269_n195: return -0.07465756684541702f;
  t269_n196: return -0.38245922327041626f;
  t269_n197: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t269_n199;
  t269_n198: return -0.009081565774977207f;
  t269_n199: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t269_n203;
  t269_n200: if (E.getNumReferences() >= 3279847424 /*254.5*/) goto t269_n202;
  t269_n201: return 0.1296849101781845f;
  t269_n202: return 0.18623369932174683f;
  t269_n203: return 0.11997878551483154f;
  t269_n204: if (E.getNumReferences() >= 3286876160 /*467.5*/) goto t269_n208;
  t269_n205: if (E.getNumReferences() >= 3286712320 /*462.5*/) goto t269_n207;
  t269_n206: return 0.031945083290338516f;
  t269_n207: return 0.11427243798971176f;
  t269_n208: return -0.0861969143152237f;
  t269_n209: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t269_n213;
  t269_n210: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t269_n212;
  t269_n211: return -0.010056725703179836f;
  t269_n212: return -0.1333499550819397f;
  t269_n213: if (E.getNumReferences() >= 3276898304 /*209.5*/) goto t269_n221;
  t269_n214: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t269_n220;
  t269_n215: if (E.getNumReferences() >= 3276374016 /*201.5*/) goto t269_n217;
  t269_n216: return 0.10823579877614975f;
  t269_n217: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t269_n219;
  t269_n218: return 0.19975174963474274f;
  t269_n219: return 0.1804409772157669f;
  t269_n220: return 0.126386359333992f;
  t269_n221: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t269_n227;
  t269_n222: if (E.getNumReferences() >= 3288539136 /*524.5*/) goto t269_n224;
  t269_n223: return -0.040699128061532974f;
  t269_n224: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t269_n226;
  t269_n225: return -0.15912814438343048f;
  t269_n226: return 0.1777482032775879f;
  t269_n227: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t269_n229;
  t269_n228: return 0.18601757287979126f;
  t269_n229: return 0.013413389213383198f;
  t269_n230: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t269_n234;
  t269_n231: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t269_n233;
  t269_n232: return 0.0006101387552917004f;
  t269_n233: return 0.008118373341858387f;
  t269_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t269_n236;
  t269_n235: return 0.0002474274660926312f;
  t269_n236: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t269_n238;
  t269_n237: return 0.00620337063446641f;
  t269_n238: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t269_n240;
  t269_n239: return -0.033017612993717194f;
  t269_n240: return 0.0011797514744102955f;
  t269_n241: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t269_n253;
  t269_n242: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t269_n250;
  t269_n243: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t269_n245;
  t269_n244: return 0.02268063649535179f;
  t269_n245: if (E.getNumReferences() >= 3337347584 /*30193.0*/) goto t269_n247;
  t269_n246: return 0.019893130287528038f;
  t269_n247: if (E.getNumReferences() >= 3338301440 /*32056.0*/) goto t269_n249;
  t269_n248: return 0.10754086077213287f;
  t269_n249: return 0.08208037912845612f;
  t269_n250: if (E.getNumReferences() >= 3290447872 /*641.0*/) goto t269_n252;
  t269_n251: return 0.10158269107341766f;
  t269_n252: return -0.045936468988657f;
  t269_n253: return 0.00685389619320631f;
  t269_n254: return 0.0035594291985034943f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree270(const Example& E) {
  t270_n0: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t270_n76;
  t270_n1: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t270_n5;
  t270_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n4;
  t270_n3: return -0.001738725695759058f;
  t270_n4: return 0.000350393820554018f;
  t270_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol))) goto t270_n67;
  t270_n6: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t270_n38;
  t270_n7: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t270_n9;
  t270_n8: return 0.0013991851592436433f;
  t270_n9: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t270_n31;
  t270_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t270_n28;
  t270_n11: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t270_n19;
  t270_n12: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t270_n16;
  t270_n13: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t270_n15;
  t270_n14: return 0.12728609144687653f;
  t270_n15: return -0.10161993652582169f;
  t270_n16: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t270_n18;
  t270_n17: return -0.09205379337072372f;
  t270_n18: return 0.08550523966550827f;
  t270_n19: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t270_n25;
  t270_n20: if (E.getFractionNameInContext() >= 3179324712 /*0.06281501054763794*/) goto t270_n22;
  t270_n21: return -0.1498962640762329f;
  t270_n22: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t270_n24;
  t270_n23: return 0.032845206558704376f;
  t270_n24: return -0.13021452724933624f;
  t270_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t270_n27;
  t270_n26: return -0.07637607306241989f;
  t270_n27: return 0.0963413342833519f;
  t270_n28: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t270_n30;
  t270_n29: return 0.02901962585747242f;
  t270_n30: return 0.08989080786705017f;
  t270_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t270_n37;
  t270_n32: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t270_n34;
  t270_n33: return 0.044744595885276794f;
  t270_n34: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t270_n36;
  t270_n35: return 0.1079963743686676f;
  t270_n36: return -0.18233714997768402f;
  t270_n37: return 0.15324607491493225f;
  t270_n38: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t270_n50;
  t270_n39: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t270_n47;
  t270_n40: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t270_n44;
  t270_n41: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t270_n43;
  t270_n42: return -0.04537227377295494f;
  t270_n43: return 0.018075324594974518f;
  t270_n44: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t270_n46;
  t270_n45: return -0.03288481757044792f;
  t270_n46: return -0.027291329577565193f;
  t270_n47: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t270_n49;
  t270_n48: return 0.004340766463428736f;
  t270_n49: return -0.02918752282857895f;
  t270_n50: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t270_n56;
  t270_n51: if (E.getFractionNameInContext() >= 3208362940 /*0.7333333492279053*/) goto t270_n53;
  t270_n52: return 0.018635286018252373f;
  t270_n53: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t270_n55;
  t270_n54: return 0.11821380257606506f;
  t270_n55: return -0.10142102092504501f;
  t270_n56: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t270_n58;
  t270_n57: return 0.1721367985010147f;
  t270_n58: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t270_n60;
  t270_n59: return 0.09269659966230392f;
  t270_n60: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t270_n62;
  t270_n61: return 0.08662862330675125f;
  t270_n62: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t270_n66;
  t270_n63: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t270_n65;
  t270_n64: return 0.12042906880378723f;
  t270_n65: return -0.13863855600357056f;
  t270_n66: return 0.10049468278884888f;
  t270_n67: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t270_n69;
  t270_n68: return 0.013076757080852985f;
  t270_n69: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t270_n71;
  t270_n70: return 0.0032811793498694897f;
  t270_n71: if (E.getFractionNameInContext() >= 3183942770 /*0.0972222238779068*/) goto t270_n73;
  t270_n72: return 0.08922728151082993f;
  t270_n73: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t270_n75;
  t270_n74: return -0.07490687817335129f;
  t270_n75: return 0.11022694408893585f;
  t270_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n252;
  t270_n77: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t270_n247;
  t270_n78: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t270_n102;
  t270_n79: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t270_n91;
  t270_n80: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t270_n90;
  t270_n81: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t270_n85;
  t270_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n84;
  t270_n83: return -0.019348153844475746f;
  t270_n84: return 0.06515216082334518f;
  t270_n85: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t270_n87;
  t270_n86: return -0.07183479517698288f;
  t270_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t270_n89;
  t270_n88: return 0.18959665298461914f;
  t270_n89: return 0.343011736869812f;
  t270_n90: return -0.04420216754078865f;
  t270_n91: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t270_n93;
  t270_n92: return -0.02923152968287468f;
  t270_n93: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t270_n95;
  t270_n94: return -0.021451961249113083f;
  t270_n95: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t270_n97;
  t270_n96: return -0.03228786960244179f;
  t270_n97: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t270_n101;
  t270_n98: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t270_n100;
  t270_n99: return 0.050921641290187836f;
  t270_n100: return 0.1300591230392456f;
  t270_n101: return 0.05917961150407791f;
  t270_n102: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t270_n232;
  t270_n103: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t270_n219;
  t270_n104: if (E.getNumReferences() >= 3260678144 /*54.5*/) goto t270_n114;
  t270_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n107;
  t270_n106: return -0.045216962695121765f;
  t270_n107: if (E.getNumReferences() >= 3259891712 /*51.5*/) goto t270_n113;
  t270_n108: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t270_n110;
  t270_n109: return -0.03869294375181198f;
  t270_n110: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t270_n112;
  t270_n111: return -0.05474628135561943f;
  t270_n112: return 0.01762492023408413f;
  t270_n113: return -0.09532858431339264f;
  t270_n114: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t270_n116;
  t270_n115: return -0.02285965159535408f;
  t270_n116: if (E.getNumReferences() >= 3296698368 /*1022.5*/) goto t270_n218;
  t270_n117: if (E.getNumReferences() >= 3290103808 /*620.0*/) goto t270_n207;
  t270_n118: if (E.getNumReferences() >= 3289808896 /*602.0*/) goto t270_n206;
  t270_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t270_n145;
  t270_n120: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t270_n142;
  t270_n121: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t270_n129;
  t270_n122: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t270_n124;
  t270_n123: return 0.05512005090713501f;
  t270_n124: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t270_n128;
  t270_n125: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t270_n127;
  t270_n126: return 0.07246336340904236f;
  t270_n127: return 0.14038363099098206f;
  t270_n128: return 0.05697322636842728f;
  t270_n129: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t270_n133;
  t270_n130: if (E.getNumReferences() >= 3281518592 /*304.0*/) goto t270_n132;
  t270_n131: return -0.006962424144148827f;
  t270_n132: return 0.06116804853081703f;
  t270_n133: if (E.getNumReferences() >= 3269001216 /*108.5*/) goto t270_n135;
  t270_n134: return 0.00015363974671345204f;
  t270_n135: if (E.getNumReferences() >= 3271622656 /*129.0*/) goto t270_n141;
  t270_n136: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t270_n140;
  t270_n137: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t270_n139;
  t270_n138: return 0.08235520869493484f;
  t270_n139: return 0.11619465798139572f;
  t270_n140: return 0.07394488155841827f;
  t270_n141: return 0.06609602272510529f;
  t270_n142: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t270_n144;
  t270_n143: return -0.22878508269786835f;
  t270_n144: return 0.003465122776106f;
  t270_n145: if (E.getNumReferences() >= 3260940288 /*55.5*/) goto t270_n149;
  t270_n146: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t270_n148;
  t270_n147: return -0.2187759280204773f;
  t270_n148: return 0.21273142099380493f;
  t270_n149: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t270_n185;
  t270_n150: if (E.getNumReferences() >= 3280633856 /*277.0*/) goto t270_n182;
  t270_n151: if (E.getNumReferences() >= 3278274560 /*230.5*/) goto t270_n177;
  t270_n152: if (E.getNumReferences() >= 3277324288 /*216.0*/) goto t270_n176;
  t270_n153: if (E.getNumReferences() >= 3276701696 /*206.5*/) goto t270_n175;
  t270_n154: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t270_n158;
  t270_n155: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t270_n157;
  t270_n156: return -0.03896215930581093f;
  t270_n157: return 0.12145281583070755f;
  t270_n158: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t270_n164;
  t270_n159: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t270_n163;
  t270_n160: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t270_n162;
  t270_n161: return 0.17549660801887512f;
  t270_n162: return 0.06668658554553986f;
  t270_n163: return -0.22143886983394623f;
  t270_n164: if (E.getNumReferences() >= 3265593344 /*82.5*/) goto t270_n166;
  t270_n165: return -0.024750523269176483f;
  t270_n166: if (E.getNumReferences() >= 3266772992 /*91.5*/) goto t270_n170;
  t270_n167: if (E.getNumReferences() >= 3266510848 /*89.5*/) goto t270_n169;
  t270_n168: return 0.0848580151796341f;
  t270_n169: return 0.1976948231458664f;
  t270_n170: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t270_n172;
  t270_n171: return 0.021460600197315216f;
  t270_n172: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t270_n174;
  t270_n173: return 0.1268816441297531f;
  t270_n174: return -0.04190914332866669f;
  t270_n175: return 0.12409542500972748f;
  t270_n176: return -0.3508942425251007f;
  t270_n177: if (E.getFractionNameInContext() >= 3177418297 /*0.0555555559694767*/) goto t270_n181;
  t270_n178: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t270_n180;
  t270_n179: return 0.15538741648197174f;
  t270_n180: return 0.1290074586868286f;
  t270_n181: return 0.05802232027053833f;
  t270_n182: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t270_n184;
  t270_n183: return -0.11743220686912537f;
  t270_n184: return 0.08425020426511765f;
  t270_n185: if (E.getNumReferences() >= 3275358208 /*186.0*/) goto t270_n199;
  t270_n186: if (E.getNumReferences() >= 3274801152 /*177.5*/) goto t270_n198;
  t270_n187: if (E.getNumReferences() >= 3267952640 /*100.5*/) goto t270_n193;
  t270_n188: if (E.getNumReferences() >= 3266379776 /*88.5*/) goto t270_n192;
  t270_n189: if (E.getNumReferences() >= 3265986560 /*85.5*/) goto t270_n191;
  t270_n190: return -0.07543142884969711f;
  t270_n191: return 0.13245852291584015f;
  t270_n192: return -0.23415076732635498f;
  t270_n193: if (E.getNumReferences() >= 3273129984 /*152.0*/) goto t270_n197;
  t270_n194: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n196;
  t270_n195: return -0.2803528606891632f;
  t270_n196: return 0.07959507405757904f;
  t270_n197: return -0.1732483208179474f;
  t270_n198: return -1.30049729347229f;
  t270_n199: if (E.getNumReferences() >= 3284467712 /*394.0*/) goto t270_n205;
  t270_n200: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t270_n202;
  t270_n201: return 0.11321723461151123f;
  t270_n202: if (E.getNumReferences() >= 3282059264 /*320.5*/) goto t270_n204;
  t270_n203: return -0.07170417159795761f;
  t270_n204: return 0.11349578946828842f;
  t270_n205: return -0.09374702721834183f;
  t270_n206: return -0.8555435538291931f;
  t270_n207: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t270_n211;
  t270_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t270_n210;
  t270_n209: return -1.0154190063476562f;
  t270_n210: return 0.05492454394698143f;
  t270_n211: if (E.getNumReferences() >= 3290947584 /*671.5*/) goto t270_n213;
  t270_n212: return 0.1731649488210678f;
  t270_n213: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t270_n217;
  t270_n214: if (E.getNumReferences() >= 3293052928 /*800.0*/) goto t270_n216;
  t270_n215: return 0.06330098956823349f;
  t270_n216: return 0.14743857085704803f;
  t270_n217: return -0.05052435025572777f;
  t270_n218: return -0.05933516100049019f;
  t270_n219: if (E.getFractionNameInContext() >= 3177111263 /*0.054411765187978745*/) goto t270_n227;
  t270_n220: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t270_n222;
  t270_n221: return -0.0223456472158432f;
  t270_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t270_n226;
  t270_n223: if (E.getNumReferences() >= 3304714240 /*1999.5*/) goto t270_n225;
  t270_n224: return 0.008960187435150146f;
  t270_n225: return -0.14811840653419495f;
  t270_n226: return 0.05918659642338753f;
  t270_n227: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t270_n229;
  t270_n228: return -0.019946392625570297f;
  t270_n229: if (E.getNumReferences() >= 3310553088 /*3376.5*/) goto t270_n231;
  t270_n230: return -0.08065947145223618f;
  t270_n231: return 0.11832238733768463f;
  t270_n232: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t270_n234;
  t270_n233: return -0.0016409491654485464f;
  t270_n234: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t270_n246;
  t270_n235: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t270_n243;
  t270_n236: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t270_n242;
  t270_n237: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t270_n239;
  t270_n238: return -0.13889165222644806f;
  t270_n239: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t270_n241;
  t270_n240: return -0.7101231217384338f;
  t270_n241: return -0.1046639084815979f;
  t270_n242: return 0.09359700232744217f;
  t270_n243: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t270_n245;
  t270_n244: return -0.04514506831765175f;
  t270_n245: return 0.1132287010550499f;
  t270_n246: return 0.03168109431862831f;
  t270_n247: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t270_n249;
  t270_n248: return 0.0027206195518374443f;
  t270_n249: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t270_n251;
  t270_n250: return -0.022701699286699295f;
  t270_n251: return -0.002007532399147749f;
  t270_n252: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t270_n254;
  t270_n253: return 0.02188258059322834f;
  t270_n254: return 0.005301021039485931f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree271(const Example& E) {
  t271_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t271_n2;
  t271_n1: return -0.00014576248941011727f;
  t271_n2: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t271_n30;
  t271_n3: if (E.getNumReferences() >= 3355726336 /*135496.0*/) goto t271_n27;
  t271_n4: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t271_n14;
  t271_n5: if (E.getNumReferences() >= 3322117632 /*8415.5*/) goto t271_n11;
  t271_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t271_n8;
  t271_n7: return -0.013276657089591026f;
  t271_n8: if (E.getFractionNameInContext() >= 3183996031 /*0.09761904925107956*/) goto t271_n10;
  t271_n9: return 0.08057219535112381f;
  t271_n10: return 0.027928262948989868f;
  t271_n11: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t271_n13;
  t271_n12: return 0.19941692054271698f;
  t271_n13: return -0.02521541714668274f;
  t271_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n24;
  t271_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n19;
  t271_n16: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t271_n18;
  t271_n17: return 0.08299461007118225f;
  t271_n18: return -0.11256316304206848f;
  t271_n19: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t271_n23;
  t271_n20: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t271_n22;
  t271_n21: return -0.1661480814218521f;
  t271_n22: return -0.6179695725440979f;
  t271_n23: return -0.02291194163262844f;
  t271_n24: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t271_n26;
  t271_n25: return 0.10653454065322876f;
  t271_n26: return 0.018967455253005028f;
  t271_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n29;
  t271_n28: return -0.45747697353363037f;
  t271_n29: return 0.08601552993059158f;
  t271_n30: if (E.getNumReferences() >= 3325673472 /*11888.0*/) goto t271_n248;
  t271_n31: if (E.getNumReferences() >= 3325138432 /*11365.5*/) goto t271_n247;
  t271_n32: if (E.getNumReferences() >= 3323262464 /*9533.5*/) goto t271_n244;
  t271_n33: if (E.getNumReferences() >= 3317972992 /*6280.0*/) goto t271_n233;
  t271_n34: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t271_n202;
  t271_n35: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t271_n199;
  t271_n36: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t271_n198;
  t271_n37: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t271_n83;
  t271_n38: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t271_n82;
  t271_n39: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t271_n47;
  t271_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n46;
  t271_n41: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t271_n45;
  t271_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n44;
  t271_n43: return -0.10170809179544449f;
  t271_n44: return 0.13752175867557526f;
  t271_n45: return -0.05871185287833214f;
  t271_n46: return -0.002029197756201029f;
  t271_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Type))) goto t271_n79;
  t271_n48: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t271_n50;
  t271_n49: return 0.0019255742663517594f;
  t271_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n64;
  t271_n51: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t271_n59;
  t271_n52: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t271_n56;
  t271_n53: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t271_n55;
  t271_n54: return -0.26597318053245544f;
  t271_n55: return -0.02763739600777626f;
  t271_n56: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t271_n58;
  t271_n57: return 0.12761035561561584f;
  t271_n58: return -0.06442447006702423f;
  t271_n59: if (E.getNumReferences() >= 3259367424 /*49.5*/) goto t271_n61;
  t271_n60: return -4.726995944976807f;
  t271_n61: if (E.getNumReferences() >= 3269525504 /*112.5*/) goto t271_n63;
  t271_n62: return -7.564284896943718e-05f;
  t271_n63: return -0.7036119103431702f;
  t271_n64: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t271_n68;
  t271_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t271_n67;
  t271_n66: return 0.04858868941664696f;
  t271_n67: return 0.10196654498577118f;
  t271_n68: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t271_n76;
  t271_n69: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t271_n73;
  t271_n70: if (E.getNumReferences() >= 3258056704 /*44.5*/) goto t271_n72;
  t271_n71: return -0.018036585301160812f;
  t271_n72: return 0.17632804811000824f;
  t271_n73: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t271_n75;
  t271_n74: return -0.6056565642356873f;
  t271_n75: return -0.058565352112054825f;
  t271_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t271_n78;
  t271_n77: return -0.03438090160489082f;
  t271_n78: return 0.07853154838085175f;
  t271_n79: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t271_n81;
  t271_n80: return 0.022014964371919632f;
  t271_n81: return 0.10752640664577484f;
  t271_n82: return -0.01315837912261486f;
  t271_n83: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t271_n117;
  t271_n84: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t271_n114;
  t271_n85: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t271_n105;
  t271_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t271_n104;
  t271_n87: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t271_n103;
  t271_n88: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t271_n102;
  t271_n89: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t271_n91;
  t271_n90: return -0.007840631529688835f;
  t271_n91: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t271_n97;
  t271_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n96;
  t271_n93: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t271_n95;
  t271_n94: return -0.0649782046675682f;
  t271_n95: return -0.46276402473449707f;
  t271_n96: return -0.048972852528095245f;
  t271_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n99;
  t271_n98: return -0.03628644347190857f;
  t271_n99: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t271_n101;
  t271_n100: return 0.0034146159887313843f;
  t271_n101: return 0.14594118297100067f;
  t271_n102: return -0.3523814082145691f;
  t271_n103: return 0.012526094913482666f;
  t271_n104: return 0.02722638100385666f;
  t271_n105: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n107;
  t271_n106: return -0.06748653948307037f;
  t271_n107: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t271_n109;
  t271_n108: return 0.028153197839856148f;
  t271_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t271_n111;
  t271_n110: return 0.05441557243466377f;
  t271_n111: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t271_n113;
  t271_n112: return 0.11363115906715393f;
  t271_n113: return 0.0660373866558075f;
  t271_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t271_n116;
  t271_n115: return 0.07624708116054535f;
  t271_n116: return 0.12003768235445023f;
  t271_n117: if (E.getNumReferences() >= 3278077952 /*227.5*/) goto t271_n197;
  t271_n118: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t271_n196;
  t271_n119: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t271_n195;
  t271_n120: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t271_n174;
  t271_n121: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t271_n159;
  t271_n122: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t271_n138;
  t271_n123: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t271_n129;
  t271_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n128;
  t271_n125: if (E.getNumReferences() >= 3225419776 /*3.0*/) goto t271_n127;
  t271_n126: return -0.10219011455774307f;
  t271_n127: return -0.38271281123161316f;
  t271_n128: return 0.024411844089627266f;
  t271_n129: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n131;
  t271_n130: return -0.09018781781196594f;
  t271_n131: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t271_n133;
  t271_n132: return -0.4130903482437134f;
  t271_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t271_n135;
  t271_n134: return 0.061517708003520966f;
  t271_n135: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t271_n137;
  t271_n136: return 0.11622750759124756f;
  t271_n137: return 0.09638061374425888f;
  t271_n138: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t271_n142;
  t271_n139: if (E.getNumReferences() >= 3231711232 /*5.0*/) goto t271_n141;
  t271_n140: return 0.07404352724552155f;
  t271_n141: return -0.20329806208610535f;
  t271_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n154;
  t271_n143: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t271_n147;
  t271_n144: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t271_n146;
  t271_n145: return -0.04600876569747925f;
  t271_n146: return -0.3010164797306061f;
  t271_n147: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t271_n151;
  t271_n148: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t271_n150;
  t271_n149: return -0.03932157903909683f;
  t271_n150: return 0.10074304789304733f;
  t271_n151: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t271_n153;
  t271_n152: return 0.10677506029605865f;
  t271_n153: return -0.002254464663565159f;
  t271_n154: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t271_n158;
  t271_n155: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t271_n157;
  t271_n156: return -0.027757516130805016f;
  t271_n157: return 0.0872441902756691f;
  t271_n158: return -0.09551811218261719f;
  t271_n159: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t271_n165;
  t271_n160: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t271_n162;
  t271_n161: return -0.034934330731630325f;
  t271_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n164;
  t271_n163: return -0.009399844333529472f;
  t271_n164: return 0.07073865830898285f;
  t271_n165: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t271_n167;
  t271_n166: return 0.014890623278915882f;
  t271_n167: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t271_n171;
  t271_n168: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t271_n170;
  t271_n169: return 0.08227811008691788f;
  t271_n170: return 0.10546255111694336f;
  t271_n171: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t271_n173;
  t271_n172: return 0.12201395630836487f;
  t271_n173: return 0.05877493694424629f;
  t271_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n178;
  t271_n175: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t271_n177;
  t271_n176: return -0.1171223521232605f;
  t271_n177: return 0.12674468755722046f;
  t271_n178: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t271_n188;
  t271_n179: if (E.getFractionNameInContext() >= 3208875577 /*0.7638888955116272*/) goto t271_n183;
  t271_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t271_n182;
  t271_n181: return -0.05957361310720444f;
  t271_n182: return 0.027004782110452652f;
  t271_n183: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t271_n185;
  t271_n184: return -0.023198217153549194f;
  t271_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t271_n187;
  t271_n186: return -0.6604711413383484f;
  t271_n187: return 0.10562597960233688f;
  t271_n188: if (E.getFractionNameInContext() >= 3200224622 /*0.3741258978843689*/) goto t271_n190;
  t271_n189: return 0.11363521218299866f;
  t271_n190: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t271_n192;
  t271_n191: return 0.014132020063698292f;
  t271_n192: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t271_n194;
  t271_n193: return 0.154534250497818f;
  t271_n194: return -0.04478763788938522f;
  t271_n195: return 0.002420699456706643f;
  t271_n196: return 0.006423818878829479f;
  t271_n197: return 0.12644924223423004f;
  t271_n198: return 0.004405850078910589f;
  t271_n199: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Namespace))) goto t271_n201;
  t271_n200: return -0.0004893142031505704f;
  t271_n201: return 0.02522488869726658f;
  t271_n202: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t271_n216;
  t271_n203: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Type))) goto t271_n205;
  t271_n204: return 0.004501497372984886f;
  t271_n205: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t271_n215;
  t271_n206: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t271_n210;
  t271_n207: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t271_n209;
  t271_n208: return 0.018983056768774986f;
  t271_n209: return 0.04729058966040611f;
  t271_n210: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t271_n214;
  t271_n211: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t271_n213;
  t271_n212: return 0.11180734634399414f;
  t271_n213: return 0.09185341745615005f;
  t271_n214: return -0.10039597749710083f;
  t271_n215: return -0.03555060923099518f;
  t271_n216: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t271_n218;
  t271_n217: return -0.018223131075501442f;
  t271_n218: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t271_n232;
  t271_n219: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t271_n221;
  t271_n220: return 0.11499452590942383f;
  t271_n221: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t271_n227;
  t271_n222: if (E.getNumReferences() >= 3271884800 /*133.0*/) goto t271_n226;
  t271_n223: if (E.getNumReferences() >= 3266183168 /*87.0*/) goto t271_n225;
  t271_n224: return -0.15481504797935486f;
  t271_n225: return -0.9167431592941284f;
  t271_n226: return -0.0636514276266098f;
  t271_n227: if (E.getNumReferences() >= 3283353600 /*360.0*/) goto t271_n229;
  t271_n228: return -0.05978022888302803f;
  t271_n229: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t271_n231;
  t271_n230: return 0.057604409754276276f;
  t271_n231: return -1.3913437128067017f;
  t271_n232: return 0.002546543488278985f;
  t271_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n243;
  t271_n234: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t271_n238;
  t271_n235: if (E.getNumReferences() >= 3320421376 /*7475.5*/) goto t271_n237;
  t271_n236: return -0.12576347589492798f;
  t271_n237: return -0.013466978445649147f;
  t271_n238: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t271_n242;
  t271_n239: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t271_n241;
  t271_n240: return -0.8494770526885986f;
  t271_n241: return -0.09075348824262619f;
  t271_n242: return 0.0003054952248930931f;
  t271_n243: return 0.013501114211976528f;
  t271_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t271_n246;
  t271_n245: return -0.07413031160831451f;
  t271_n246: return 0.08063910901546478f;
  t271_n247: return -0.17898491024971008f;
  t271_n248: if (E.getNumReferences() >= 3326510080 /*12705.0*/) goto t271_n250;
  t271_n249: return 0.08210089057683945f;
  t271_n250: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t271_n252;
  t271_n251: return 0.023091726005077362f;
  t271_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t271_n254;
  t271_n253: return -0.4729049801826477f;
  t271_n254: return -0.008934392593801022f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree272(const Example& E) {
  t272_n0: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t272_n2;
  t272_n1: return -0.0004076477198395878f;
  t272_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n254;
  t272_n3: if (E.getNumReferences() >= 3289784320 /*600.5*/) goto t272_n253;
  t272_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t272_n6;
  t272_n5: return -0.00021744145487900823f;
  t272_n6: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t272_n250;
  t272_n7: if (E.getScope() & (BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t272_n133;
  t272_n8: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t272_n76;
  t272_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n75;
  t272_n10: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t272_n30;
  t272_n11: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t272_n29;
  t272_n12: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t272_n16;
  t272_n13: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t272_n15;
  t272_n14: return -0.05166150629520416f;
  t272_n15: return -0.2856179475784302f;
  t272_n16: if (E.getFractionNameInContext() >= 3185414456 /*0.10818713903427124*/) goto t272_n18;
  t272_n17: return -0.004152842331677675f;
  t272_n18: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t272_n22;
  t272_n19: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t272_n21;
  t272_n20: return -0.023392200469970703f;
  t272_n21: return -0.12551249563694f;
  t272_n22: if (E.getNumReferences() >= 3257794560 /*43.5*/) goto t272_n28;
  t272_n23: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t272_n27;
  t272_n24: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t272_n26;
  t272_n25: return -0.06359412521123886f;
  t272_n26: return 0.09311527758836746f;
  t272_n27: return 0.12001483887434006f;
  t272_n28: return -0.013702026568353176f;
  t272_n29: return 0.04014884680509567f;
  t272_n30: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t272_n32;
  t272_n31: return -0.014020880684256554f;
  t272_n32: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t272_n34;
  t272_n33: return -0.0578555166721344f;
  t272_n34: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t272_n54;
  t272_n35: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t272_n53;
  t272_n36: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t272_n38;
  t272_n37: return 0.059144001454114914f;
  t272_n38: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t272_n42;
  t272_n39: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n41;
  t272_n40: return 0.01602390967309475f;
  t272_n41: return 0.07702288031578064f;
  t272_n42: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n46;
  t272_n43: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t272_n45;
  t272_n44: return 0.06299892067909241f;
  t272_n45: return 0.12135501205921173f;
  t272_n46: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t272_n52;
  t272_n47: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n49;
  t272_n48: return 0.1195487454533577f;
  t272_n49: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t272_n51;
  t272_n50: return -0.2281983494758606f;
  t272_n51: return 0.085516057908535f;
  t272_n52: return 0.049432575702667236f;
  t272_n53: return 0.00822471920400858f;
  t272_n54: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t272_n70;
  t272_n55: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t272_n67;
  t272_n56: if (E.getFractionNameInContext() >= 3204770895 /*0.5192307829856873*/) goto t272_n58;
  t272_n57: return 0.002772320993244648f;
  t272_n58: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n62;
  t272_n59: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t272_n61;
  t272_n60: return 0.079181969165802f;
  t272_n61: return 0.04380355030298233f;
  t272_n62: if (E.getFractionNameInContext() >= 3205152196 /*0.5419580936431885*/) goto t272_n64;
  t272_n63: return 0.147891104221344f;
  t272_n64: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n66;
  t272_n65: return 0.11563431471586227f;
  t272_n66: return -0.1626499891281128f;
  t272_n67: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n69;
  t272_n68: return -0.10351768136024475f;
  t272_n69: return -0.09937754273414612f;
  t272_n70: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n72;
  t272_n71: return -0.04231654852628708f;
  t272_n72: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n74;
  t272_n73: return 0.09171529859304428f;
  t272_n74: return -0.10181710124015808f;
  t272_n75: return -0.003558429656550288f;
  t272_n76: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t272_n78;
  t272_n77: return -0.0031096667516976595f;
  t272_n78: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t272_n112;
  t272_n79: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t272_n101;
  t272_n80: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t272_n82;
  t272_n81: return 0.017300115898251534f;
  t272_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t272_n86;
  t272_n83: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t272_n85;
  t272_n84: return 0.03644314408302307f;
  t272_n85: return 0.12071310728788376f;
  t272_n86: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t272_n92;
  t272_n87: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t272_n89;
  t272_n88: return -0.015425116755068302f;
  t272_n89: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t272_n91;
  t272_n90: return -0.2936517894268036f;
  t272_n91: return 0.0036645028740167618f;
  t272_n92: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t272_n98;
  t272_n93: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t272_n95;
  t272_n94: return 0.06779975444078445f;
  t272_n95: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t272_n97;
  t272_n96: return 0.15784955024719238f;
  t272_n97: return 0.2572380602359772f;
  t272_n98: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t272_n100;
  t272_n99: return 0.04811601713299751f;
  t272_n100: return -0.10966351628303528f;
  t272_n101: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t272_n103;
  t272_n102: return 0.02035270631313324f;
  t272_n103: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t272_n111;
  t272_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n110;
  t272_n105: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t272_n109;
  t272_n106: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t272_n108;
  t272_n107: return -0.07808836549520493f;
  t272_n108: return 0.10130243003368378f;
  t272_n109: return -0.3388873338699341f;
  t272_n110: return 0.06303153187036514f;
  t272_n111: return 0.003449592273682356f;
  t272_n112: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t272_n122;
  t272_n113: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t272_n119;
  t272_n114: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n116;
  t272_n115: return 0.0038611043710261583f;
  t272_n116: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n118;
  t272_n117: return 0.046848364174366f;
  t272_n118: return -0.016284707933664322f;
  t272_n119: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t272_n121;
  t272_n120: return -0.010429256595671177f;
  t272_n121: return -0.1113523542881012f;
  t272_n122: if (E.getFractionNameInContext() >= 3197848396 /*0.3033088445663452*/) goto t272_n124;
  t272_n123: return -0.12753742933273315f;
  t272_n124: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t272_n132;
  t272_n125: if (E.getFractionNameInContext() >= 3199399908 /*0.34954750537872314*/) goto t272_n127;
  t272_n126: return -0.06594208627939224f;
  t272_n127: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t272_n129;
  t272_n128: return 0.08053376525640488f;
  t272_n129: if (E.getFractionNameInContext() >= 3200515136 /*0.3827838897705078*/) goto t272_n131;
  t272_n130: return 0.10527556389570236f;
  t272_n131: return 0.03177713602781296f;
  t272_n132: return 0.038235854357481f;
  t272_n133: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t272_n135;
  t272_n134: return 0.0022830406669527292f;
  t272_n135: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t272_n229;
  t272_n136: if (E.getFractionNameInContext() >= 3198506326 /*0.3229166865348816*/) goto t272_n146;
  t272_n137: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n143;
  t272_n138: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t272_n142;
  t272_n139: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t272_n141;
  t272_n140: return -0.0031122544314712286f;
  t272_n141: return -0.07511632889509201f;
  t272_n142: return -0.026200823485851288f;
  t272_n143: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n145;
  t272_n144: return 0.009307477623224258f;
  t272_n145: return 0.27711865305900574f;
  t272_n146: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t272_n228;
  t272_n147: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t272_n157;
  t272_n148: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t272_n156;
  t272_n149: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t272_n153;
  t272_n150: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n152;
  t272_n151: return 0.03521082177758217f;
  t272_n152: return 0.061818379908800125f;
  t272_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t272_n155;
  t272_n154: return -0.0024991745594888926f;
  t272_n155: return 0.23885948956012726f;
  t272_n156: return -0.005747743416577578f;
  t272_n157: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t272_n197;
  t272_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n194;
  t272_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t272_n189;
  t272_n160: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t272_n174;
  t272_n161: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n171;
  t272_n162: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t272_n168;
  t272_n163: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t272_n165;
  t272_n164: return -0.06094209849834442f;
  t272_n165: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n167;
  t272_n166: return 0.17067968845367432f;
  t272_n167: return 0.12838992476463318f;
  t272_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n170;
  t272_n169: return -0.3282831013202667f;
  t272_n170: return -0.01606006920337677f;
  t272_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t272_n173;
  t272_n172: return 0.06784401088953018f;
  t272_n173: return 0.4502456784248352f;
  t272_n174: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t272_n180;
  t272_n175: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t272_n177;
  t272_n176: return -0.6293841600418091f;
  t272_n177: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n179;
  t272_n178: return 0.14469920098781586f;
  t272_n179: return -0.1167193055152893f;
  t272_n180: if (E.getFractionNameInContext() >= 3209913561 /*0.8257575631141663*/) goto t272_n188;
  t272_n181: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t272_n185;
  t272_n182: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t272_n184;
  t272_n183: return 0.049974311143159866f;
  t272_n184: return -0.24231784045696259f;
  t272_n185: if (E.getFractionNameInContext() >= 3206640733 /*0.6306818127632141*/) goto t272_n187;
  t272_n186: return -0.07330796867609024f;
  t272_n187: return 0.000652398681268096f;
  t272_n188: return -0.1939070075750351f;
  t272_n189: if (E.getNumReferences() >= 3245342720 /*15.0*/) goto t272_n191;
  t272_n190: return -0.0350937582552433f;
  t272_n191: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t272_n193;
  t272_n192: return -0.6102394461631775f;
  t272_n193: return -0.06780336052179337f;
  t272_n194: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t272_n196;
  t272_n195: return 0.2964102327823639f;
  t272_n196: return 0.10961715131998062f;
  t272_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t272_n201;
  t272_n198: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t272_n200;
  t272_n199: return -0.14842729270458221f;
  t272_n200: return -0.7498225569725037f;
  t272_n201: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n203;
  t272_n202: return -0.005525926593691111f;
  t272_n203: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t272_n213;
  t272_n204: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t272_n212;
  t272_n205: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t272_n209;
  t272_n206: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t272_n208;
  t272_n207: return 0.0853043869137764f;
  t272_n208: return -0.18474891781806946f;
  t272_n209: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t272_n211;
  t272_n210: return -0.013599694706499577f;
  t272_n211: return 0.0683710128068924f;
  t272_n212: return 0.017544696107506752f;
  t272_n213: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t272_n221;
  t272_n214: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t272_n220;
  t272_n215: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t272_n219;
  t272_n216: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t272_n218;
  t272_n217: return -15.621244430541992f;
  t272_n218: return 0.006548721808940172f;
  t272_n219: return 0.08284850418567657f;
  t272_n220: return -0.10806982219219208f;
  t272_n221: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t272_n227;
  t272_n222: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t272_n226;
  t272_n223: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t272_n225;
  t272_n224: return 0.0752536803483963f;
  t272_n225: return 0.12286292016506195f;
  t272_n226: return -0.200082927942276f;
  t272_n227: return 0.007968499325215816f;
  t272_n228: return 0.028723645955324173f;
  t272_n229: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t272_n231;
  t272_n230: return -0.0021294753532856703f;
  t272_n231: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t272_n249;
  t272_n232: if (E.getFractionNameInContext() >= 3211998003 /*0.949999988079071*/) goto t272_n240;
  t272_n233: if (E.getFractionNameInContext() >= 3206790530 /*0.6396104097366333*/) goto t272_n239;
  t272_n234: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t272_n238;
  t272_n235: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n237;
  t272_n236: return 0.002910449169576168f;
  t272_n237: return 0.022710874676704407f;
  t272_n238: return 0.09535907953977585f;
  t272_n239: return -0.03964105620980263f;
  t272_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t272_n248;
  t272_n241: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t272_n247;
  t272_n242: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t272_n246;
  t272_n243: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t272_n245;
  t272_n244: return 0.11623238027095795f;
  t272_n245: return 0.13185147941112518f;
  t272_n246: return -0.021191870793700218f;
  t272_n247: return -0.04039640724658966f;
  t272_n248: return 0.025113599374890327f;
  t272_n249: return -0.01095727737993002f;
  t272_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t272_n252;
  t272_n251: return 0.00014990216004662216f;
  t272_n252: return 0.005476553924381733f;
  t272_n253: return 0.00767519511282444f;
  t272_n254: return 0.0033338316716253757f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree273(const Example& E) {
  t273_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t273_n212;
  t273_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t273_n3;
  t273_n2: return -5.885892460355535e-05f;
  t273_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t273_n211;
  t273_n4: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t273_n204;
  t273_n5: if (E.getNumReferences() >= 3290570752 /*648.5*/) goto t273_n191;
  t273_n6: if (E.getNumReferences() >= 3283648512 /*369.0*/) goto t273_n182;
  t273_n7: if (E.getNumReferences() >= 3283271680 /*357.5*/) goto t273_n181;
  t273_n8: if (E.getNumReferences() >= 3277848576 /*224.0*/) goto t273_n172;
  t273_n9: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t273_n17;
  t273_n10: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t273_n12;
  t273_n11: return 0.00931217148900032f;
  t273_n12: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t273_n14;
  t273_n13: return 0.005571874789893627f;
  t273_n14: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t273_n16;
  t273_n15: return -0.1725580245256424f;
  t273_n16: return -0.5815138220787048f;
  t273_n17: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t273_n29;
  t273_n18: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n20;
  t273_n19: return 0.004338522907346487f;
  t273_n20: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t273_n24;
  t273_n21: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t273_n23;
  t273_n22: return -0.04159202054142952f;
  t273_n23: return 0.08678064495325089f;
  t273_n24: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t273_n28;
  t273_n25: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t273_n27;
  t273_n26: return 0.184225395321846f;
  t273_n27: return 0.1284501850605011f;
  t273_n28: return 0.021771930158138275f;
  t273_n29: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t273_n115;
  t273_n30: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t273_n84;
  t273_n31: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t273_n59;
  t273_n32: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t273_n44;
  t273_n33: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t273_n35;
  t273_n34: return -0.01707945391535759f;
  t273_n35: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t273_n41;
  t273_n36: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t273_n40;
  t273_n37: if (E.getFractionNameInContext() >= 3196898509 /*0.2750000059604645*/) goto t273_n39;
  t273_n38: return -0.07496210187673569f;
  t273_n39: return -0.15690121054649353f;
  t273_n40: return 0.1059802994132042f;
  t273_n41: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t273_n43;
  t273_n42: return 0.12197326868772507f;
  t273_n43: return -0.010523960925638676f;
  t273_n44: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t273_n58;
  t273_n45: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t273_n47;
  t273_n46: return -0.023689284920692444f;
  t273_n47: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t273_n57;
  t273_n48: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n56;
  t273_n49: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t273_n55;
  t273_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t273_n54;
  t273_n51: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t273_n53;
  t273_n52: return -0.7463006377220154f;
  t273_n53: return -0.32472899556159973f;
  t273_n54: return -0.005608633626252413f;
  t273_n55: return -0.13895483314990997f;
  t273_n56: return 0.11069735884666443f;
  t273_n57: return -0.48828479647636414f;
  t273_n58: return -0.1733531355857849f;
  t273_n59: if (E.getFractionNameInContext() >= 3200415272 /*0.379807710647583*/) goto t273_n81;
  t273_n60: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t273_n80;
  t273_n61: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t273_n71;
  t273_n62: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n68;
  t273_n63: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t273_n65;
  t273_n64: return -0.20240607857704163f;
  t273_n65: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t273_n67;
  t273_n66: return -0.15031908452510834f;
  t273_n67: return -0.3862011134624481f;
  t273_n68: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t273_n70;
  t273_n69: return -0.13651078939437866f;
  t273_n70: return 0.07591983675956726f;
  t273_n71: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t273_n79;
  t273_n72: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n76;
  t273_n73: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t273_n75;
  t273_n74: return -0.0014984289882704616f;
  t273_n75: return 0.09189345687627792f;
  t273_n76: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t273_n78;
  t273_n77: return -0.05833187699317932f;
  t273_n78: return -0.28311285376548767f;
  t273_n79: return 0.15379808843135834f;
  t273_n80: return -0.020027251914143562f;
  t273_n81: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n83;
  t273_n82: return 0.08072561025619507f;
  t273_n83: return -0.05430695042014122f;
  t273_n84: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t273_n90;
  t273_n85: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n89;
  t273_n86: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t273_n88;
  t273_n87: return -0.18206727504730225f;
  t273_n88: return -0.905667245388031f;
  t273_n89: return -0.007340544369071722f;
  t273_n90: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t273_n110;
  t273_n91: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t273_n95;
  t273_n92: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n94;
  t273_n93: return 0.09391529113054276f;
  t273_n94: return 0.018315142020583153f;
  t273_n95: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t273_n109;
  t273_n96: if (E.getNumReferences() >= 3243245568 /*13.0*/) goto t273_n106;
  t273_n97: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t273_n103;
  t273_n98: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t273_n100;
  t273_n99: return -0.08085771650075912f;
  t273_n100: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t273_n102;
  t273_n101: return -0.13085930049419403f;
  t273_n102: return -0.6752322912216187f;
  t273_n103: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t273_n105;
  t273_n104: return 0.10205813497304916f;
  t273_n105: return -0.11757905036211014f;
  t273_n106: if (E.getNumReferences() >= 3245342720 /*15.0*/) goto t273_n108;
  t273_n107: return 0.09757173806428909f;
  t273_n108: return -0.03412652760744095f;
  t273_n109: return 0.04385901987552643f;
  t273_n110: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n114;
  t273_n111: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t273_n113;
  t273_n112: return 0.11204966902732849f;
  t273_n113: return 0.04292033240199089f;
  t273_n114: return -0.15385660529136658f;
  t273_n115: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t273_n169;
  t273_n116: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t273_n146;
  t273_n117: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n135;
  t273_n118: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t273_n132;
  t273_n119: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n123;
  t273_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t273_n122;
  t273_n121: return 0.024469653144478798f;
  t273_n122: return 0.1385016143321991f;
  t273_n123: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t273_n127;
  t273_n124: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t273_n126;
  t273_n125: return -0.03648418188095093f;
  t273_n126: return -0.23428019881248474f;
  t273_n127: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t273_n129;
  t273_n128: return 0.10775469988584518f;
  t273_n129: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t273_n131;
  t273_n130: return -0.17204231023788452f;
  t273_n131: return 0.1163838654756546f;
  t273_n132: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t273_n134;
  t273_n133: return -0.08949190378189087f;
  t273_n134: return -0.39979252219200134f;
  t273_n135: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t273_n139;
  t273_n136: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t273_n138;
  t273_n137: return 0.12354511767625809f;
  t273_n138: return -0.15803831815719604f;
  t273_n139: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t273_n143;
  t273_n140: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t273_n142;
  t273_n141: return -0.12367759644985199f;
  t273_n142: return 0.040356799960136414f;
  t273_n143: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t273_n145;
  t273_n144: return 0.14271409809589386f;
  t273_n145: return 0.047449130564928055f;
  t273_n146: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t273_n152;
  t273_n147: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t273_n149;
  t273_n148: return -0.07878445833921432f;
  t273_n149: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t273_n151;
  t273_n150: return -0.3134682774543762f;
  t273_n151: return 0.03814122453331947f;
  t273_n152: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t273_n168;
  t273_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t273_n163;
  t273_n154: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t273_n162;
  t273_n155: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t273_n161;
  t273_n156: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t273_n160;
  t273_n157: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t273_n159;
  t273_n158: return 0.1172608733177185f;
  t273_n159: return 0.12540237605571747f;
  t273_n160: return 0.07805849611759186f;
  t273_n161: return 0.007815854623913765f;
  t273_n162: return -0.017589787021279335f;
  t273_n163: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t273_n165;
  t273_n164: return 0.14884710311889648f;
  t273_n165: if (E.getFractionNameInContext() >= 3203685656 /*0.47727274894714355*/) goto t273_n167;
  t273_n166: return 0.12992693483829498f;
  t273_n167: return 0.001955318031832576f;
  t273_n168: return -0.005254830699414015f;
  t273_n169: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t273_n171;
  t273_n170: return -0.2712676227092743f;
  t273_n171: return -0.08285515755414963f;
  t273_n172: if (E.getNumReferences() >= 3280420864 /*270.5*/) goto t273_n178;
  t273_n173: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t273_n175;
  t273_n174: return 0.020636023953557014f;
  t273_n175: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t273_n177;
  t273_n176: return 0.08492524921894073f;
  t273_n177: return 0.1594158411026001f;
  t273_n178: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t273_n180;
  t273_n179: return 0.06468790024518967f;
  t273_n180: return -0.47297433018684387f;
  t273_n181: return -0.3698512315750122f;
  t273_n182: if (E.getNumReferences() >= 3284549632 /*396.5*/) goto t273_n184;
  t273_n183: return 0.11186657845973969f;
  t273_n184: if (E.getNumReferences() >= 3287793664 /*495.5*/) goto t273_n190;
  t273_n185: if (E.getNumReferences() >= 3286188032 /*446.5*/) goto t273_n187;
  t273_n186: return 0.06841817498207092f;
  t273_n187: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t273_n189;
  t273_n188: return -0.571541965007782f;
  t273_n189: return 0.10724063217639923f;
  t273_n190: return 0.1023239865899086f;
  t273_n191: if (E.getFractionNameInContext() >= 3180951593 /*0.07493621855974197*/) goto t273_n203;
  t273_n192: if (E.getNumReferences() >= 3306201088 /*2314.0*/) goto t273_n200;
  t273_n193: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t273_n195;
  t273_n194: return -0.27136701345443726f;
  t273_n195: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t273_n197;
  t273_n196: return 0.07307474315166473f;
  t273_n197: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t273_n199;
  t273_n198: return -0.18739281594753265f;
  t273_n199: return -0.5773042440414429f;
  t273_n200: if (E.getNumReferences() >= 3307335680 /*2591.0*/) goto t273_n202;
  t273_n201: return 0.12172283232212067f;
  t273_n202: return -0.06663410365581512f;
  t273_n203: return 0.058372508734464645f;
  t273_n204: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t273_n208;
  t273_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t273_n207;
  t273_n206: return -0.024388806894421577f;
  t273_n207: return 0.05255052074790001f;
  t273_n208: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t273_n210;
  t273_n209: return -0.00025159845245070755f;
  t273_n210: return 0.006676039658486843f;
  t273_n211: return 0.003528944216668606f;
  t273_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t273_n226;
  t273_n213: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t273_n215;
  t273_n214: return -0.018739651888608932f;
  t273_n215: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t273_n225;
  t273_n216: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t273_n218;
  t273_n217: return -0.0867091491818428f;
  t273_n218: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t273_n220;
  t273_n219: return -0.1212640330195427f;
  t273_n220: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t273_n224;
  t273_n221: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t273_n223;
  t273_n222: return -0.1941245049238205f;
  t273_n223: return 0.11813176423311234f;
  t273_n224: return 0.05668352171778679f;
  t273_n225: return -0.02268357016146183f;
  t273_n226: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t273_n250;
  t273_n227: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t273_n235;
  t273_n228: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t273_n232;
  t273_n229: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t273_n231;
  t273_n230: return -0.11191446334123611f;
  t273_n231: return 0.3310299515724182f;
  t273_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t273_n234;
  t273_n233: return 0.023189377039670944f;
  t273_n234: return -0.16996966302394867f;
  t273_n235: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n237;
  t273_n236: return -0.02009790949523449f;
  t273_n237: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t273_n241;
  t273_n238: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t273_n240;
  t273_n239: return 0.30197542905807495f;
  t273_n240: return -0.11191517114639282f;
  t273_n241: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t273_n245;
  t273_n242: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t273_n244;
  t273_n243: return -0.39793476462364197f;
  t273_n244: return 0.15658017992973328f;
  t273_n245: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t273_n249;
  t273_n246: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t273_n248;
  t273_n247: return -0.1118631586432457f;
  t273_n248: return 0.010657158680260181f;
  t273_n249: return 0.08640711009502411f;
  t273_n250: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t273_n252;
  t273_n251: return -0.26014411449432373f;
  t273_n252: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t273_n254;
  t273_n253: return 0.1104288175702095f;
  t273_n254: return -0.14079271256923676f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree274(const Example& E) {
  t274_n0: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t274_n2;
  t274_n1: return -0.0003313148918095976f;
  t274_n2: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t274_n6;
  t274_n3: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t274_n5;
  t274_n4: return -0.026753833517432213f;
  t274_n5: return 0.12744779884815216f;
  t274_n6: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t274_n8;
  t274_n7: return 0.0018717256607487798f;
  t274_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag))) goto t274_n200;
  t274_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro))) goto t274_n21;
  t274_n10: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t274_n12;
  t274_n11: return -0.0008188149658963084f;
  t274_n12: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t274_n20;
  t274_n13: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t274_n15;
  t274_n14: return 0.033659473061561584f;
  t274_n15: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t274_n19;
  t274_n16: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t274_n18;
  t274_n17: return -0.037009622901678085f;
  t274_n18: return 0.16287311911582947f;
  t274_n19: return 0.08460716158151627f;
  t274_n20: return 0.0013969243736937642f;
  t274_n21: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t274_n141;
  t274_n22: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t274_n138;
  t274_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n25;
  t274_n24: return -0.11547267436981201f;
  t274_n25: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t274_n89;
  t274_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t274_n60;
  t274_n27: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t274_n59;
  t274_n28: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t274_n58;
  t274_n29: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t274_n51;
  t274_n30: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t274_n38;
  t274_n31: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t274_n35;
  t274_n32: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t274_n34;
  t274_n33: return -0.030637508258223534f;
  t274_n34: return 0.08216653764247894f;
  t274_n35: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t274_n37;
  t274_n36: return -0.059334490448236465f;
  t274_n37: return -0.6412124037742615f;
  t274_n38: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t274_n44;
  t274_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n43;
  t274_n40: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t274_n42;
  t274_n41: return 0.211892768740654f;
  t274_n42: return 0.3330828845500946f;
  t274_n43: return 0.06944242864847183f;
  t274_n44: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t274_n46;
  t274_n45: return -0.08282847702503204f;
  t274_n46: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t274_n50;
  t274_n47: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t274_n49;
  t274_n48: return 0.08416080474853516f;
  t274_n49: return 0.10369322448968887f;
  t274_n50: return -0.26153475046157837f;
  t274_n51: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t274_n53;
  t274_n52: return -0.07685066759586334f;
  t274_n53: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t274_n55;
  t274_n54: return -0.6646227836608887f;
  t274_n55: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t274_n57;
  t274_n56: return 0.11618729680776596f;
  t274_n57: return 0.031145166605710983f;
  t274_n58: return 0.13135065138339996f;
  t274_n59: return -0.02474166825413704f;
  t274_n60: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t274_n62;
  t274_n61: return -0.01057424582540989f;
  t274_n62: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t274_n88;
  t274_n63: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t274_n87;
  t274_n64: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t274_n82;
  t274_n65: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t274_n75;
  t274_n66: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t274_n74;
  t274_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n69;
  t274_n68: return -0.12447784096002579f;
  t274_n69: if (E.getNumReferences() >= 3332389888 /*20510.0*/) goto t274_n73;
  t274_n70: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t274_n72;
  t274_n71: return 0.0708022192120552f;
  t274_n72: return 0.1013222262263298f;
  t274_n73: return -0.16394731402397156f;
  t274_n74: return 0.018847787752747536f;
  t274_n75: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t274_n77;
  t274_n76: return -0.5126512050628662f;
  t274_n77: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t274_n79;
  t274_n78: return -0.004573109559714794f;
  t274_n79: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n81;
  t274_n80: return -0.0558333694934845f;
  t274_n81: return 0.020617762580513954f;
  t274_n82: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t274_n84;
  t274_n83: return -0.0603073351085186f;
  t274_n84: if (E.getFractionNameInContext() >= 3205846358 /*0.5833333730697632*/) goto t274_n86;
  t274_n85: return 0.003673472674563527f;
  t274_n86: return -0.6841458678245544f;
  t274_n87: return -0.046732090413570404f;
  t274_n88: return 0.009028811939060688f;
  t274_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n91;
  t274_n90: return -0.12271155416965485f;
  t274_n91: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t274_n111;
  t274_n92: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t274_n96;
  t274_n93: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n95;
  t274_n94: return -0.3345744013786316f;
  t274_n95: return 0.030561890453100204f;
  t274_n96: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t274_n110;
  t274_n97: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t274_n101;
  t274_n98: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n100;
  t274_n99: return -0.24566766619682312f;
  t274_n100: return 0.11420439183712006f;
  t274_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t274_n105;
  t274_n102: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n104;
  t274_n103: return 0.02081494778394699f;
  t274_n104: return 0.11537165194749832f;
  t274_n105: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t274_n107;
  t274_n106: return 0.13331516087055206f;
  t274_n107: if (E.getNumReferences() >= 3317173248 /*5889.5*/) goto t274_n109;
  t274_n108: return 0.08139026910066605f;
  t274_n109: return -0.0327271968126297f;
  t274_n110: return 0.03881407156586647f;
  t274_n111: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t274_n121;
  t274_n112: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t274_n116;
  t274_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n115;
  t274_n114: return 0.06522573530673981f;
  t274_n115: return 0.14206397533416748f;
  t274_n116: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t274_n118;
  t274_n117: return -0.16483858227729797f;
  t274_n118: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t274_n120;
  t274_n119: return 0.05303077772259712f;
  t274_n120: return 0.11431077867746353f;
  t274_n121: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t274_n123;
  t274_n122: return -0.08387959748506546f;
  t274_n123: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t274_n137;
  t274_n124: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t274_n126;
  t274_n125: return 0.03282612934708595f;
  t274_n126: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t274_n132;
  t274_n127: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t274_n129;
  t274_n128: return -0.028642170131206512f;
  t274_n129: if (E.getNumReferences() >= 3335769344 /*27110.5*/) goto t274_n131;
  t274_n130: return -1.2503443956375122f;
  t274_n131: return 0.11164231598377228f;
  t274_n132: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t274_n136;
  t274_n133: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n135;
  t274_n134: return -0.40767979621887207f;
  t274_n135: return 0.06831046938896179f;
  t274_n136: return -0.04070572927594185f;
  t274_n137: return 0.10302132368087769f;
  t274_n138: if (E.getFractionNameInContext() >= 3196618889 /*0.2666666805744171*/) goto t274_n140;
  t274_n139: return 0.5039477348327637f;
  t274_n140: return 0.09278362989425659f;
  t274_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n143;
  t274_n142: return -0.007576953154057264f;
  t274_n143: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t274_n159;
  t274_n144: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t274_n146;
  t274_n145: return -0.004261022433638573f;
  t274_n146: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t274_n158;
  t274_n147: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t274_n149;
  t274_n148: return 0.1291501373052597f;
  t274_n149: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n151;
  t274_n150: return -0.009403270669281483f;
  t274_n151: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t274_n153;
  t274_n152: return -0.10111163556575775f;
  t274_n153: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t274_n155;
  t274_n154: return 1.4179489612579346f;
  t274_n155: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t274_n157;
  t274_n156: return -0.10308888554573059f;
  t274_n157: return 0.3958979547023773f;
  t274_n158: return 4.898026943206787f;
  t274_n159: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t274_n199;
  t274_n160: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t274_n168;
  t274_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n163;
  t274_n162: return -0.12379114329814911f;
  t274_n163: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t274_n167;
  t274_n164: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t274_n166;
  t274_n165: return 0.0658605620265007f;
  t274_n166: return 0.010942035354673862f;
  t274_n167: return -0.19972550868988037f;
  t274_n168: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t274_n198;
  t274_n169: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t274_n195;
  t274_n170: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t274_n180;
  t274_n171: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t274_n175;
  t274_n172: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t274_n174;
  t274_n173: return 0.06515488773584366f;
  t274_n174: return 0.11707256734371185f;
  t274_n175: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t274_n177;
  t274_n176: return 0.1734626293182373f;
  t274_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t274_n179;
  t274_n178: return -0.2607700824737549f;
  t274_n179: return 0.1374674141407013f;
  t274_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t274_n186;
  t274_n181: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t274_n183;
  t274_n182: return -0.018641510978341103f;
  t274_n183: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t274_n185;
  t274_n184: return 0.055220700800418854f;
  t274_n185: return 0.099618099629879f;
  t274_n186: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t274_n188;
  t274_n187: return 0.050497427582740784f;
  t274_n188: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t274_n192;
  t274_n189: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t274_n191;
  t274_n190: return 0.10161642730236053f;
  t274_n191: return 0.1355130523443222f;
  t274_n192: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t274_n194;
  t274_n193: return 0.13191404938697815f;
  t274_n194: return -0.38492146134376526f;
  t274_n195: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t274_n197;
  t274_n196: return -0.030539153143763542f;
  t274_n197: return -0.23033185303211212f;
  t274_n198: return 0.025716962292790413f;
  t274_n199: return -0.10485170781612396f;
  t274_n200: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t274_n202;
  t274_n201: return -0.021763021126389503f;
  t274_n202: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t274_n254;
  t274_n203: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t274_n251;
  t274_n204: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t274_n214;
  t274_n205: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t274_n207;
  t274_n206: return -0.3678254783153534f;
  t274_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t274_n209;
  t274_n208: return -0.23325388133525848f;
  t274_n209: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t274_n211;
  t274_n210: return 0.10190503299236298f;
  t274_n211: if (E.getNumReferences() >= 3258712064 /*47.0*/) goto t274_n213;
  t274_n212: return 0.04986034706234932f;
  t274_n213: return -0.45892080664634705f;
  t274_n214: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t274_n216;
  t274_n215: return 0.012743436731398106f;
  t274_n216: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t274_n224;
  t274_n217: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t274_n221;
  t274_n218: if (E.getNumReferences() >= 3305279488 /*2089.0*/) goto t274_n220;
  t274_n219: return 0.03374401852488518f;
  t274_n220: return -0.12366662919521332f;
  t274_n221: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t274_n223;
  t274_n222: return -0.28405359387397766f;
  t274_n223: return -0.038659416139125824f;
  t274_n224: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t274_n250;
  t274_n225: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t274_n241;
  t274_n226: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t274_n240;
  t274_n227: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t274_n237;
  t274_n228: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t274_n234;
  t274_n229: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t274_n233;
  t274_n230: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t274_n232;
  t274_n231: return 0.13132597506046295f;
  t274_n232: return 0.12010353058576584f;
  t274_n233: return 0.1225099191069603f;
  t274_n234: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t274_n236;
  t274_n235: return -0.12501634657382965f;
  t274_n236: return 0.13529445230960846f;
  t274_n237: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t274_n239;
  t274_n238: return 0.15049293637275696f;
  t274_n239: return 0.09418672323226929f;
  t274_n240: return 0.03145745024085045f;
  t274_n241: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t274_n249;
  t274_n242: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t274_n244;
  t274_n243: return -0.1420121192932129f;
  t274_n244: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t274_n246;
  t274_n245: return -0.12609587609767914f;
  t274_n246: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t274_n248;
  t274_n247: return 0.3541596531867981f;
  t274_n248: return 0.12919671833515167f;
  t274_n249: return -0.13971373438835144f;
  t274_n250: return -0.015606977045536041f;
  t274_n251: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t274_n253;
  t274_n252: return 0.1354462057352066f;
  t274_n253: return 0.09366082400083542f;
  t274_n254: return -0.003420937340706587f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree275(const Example& E) {
  t275_n0: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t275_n248;
  t275_n1: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t275_n5;
  t275_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n4;
  t275_n3: return -0.002394032897427678f;
  t275_n4: return 0.0005623570177704096f;
  t275_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t275_n39;
  t275_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n18;
  t275_n7: if (E.getNumReferences() >= 3272900608 /*148.5*/) goto t275_n17;
  t275_n8: if (E.getNumNameInContext() >= 3234856960 /*6.5*/) goto t275_n16;
  t275_n9: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t275_n11;
  t275_n10: return -0.0035732234828174114f;
  t275_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t275_n15;
  t275_n12: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t275_n14;
  t275_n13: return -0.2154294103384018f;
  t275_n14: return 0.09793592244386673f;
  t275_n15: return 0.11361591517925262f;
  t275_n16: return 0.1283216029405594f;
  t275_n17: return -0.017326753586530685f;
  t275_n18: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n20;
  t275_n19: return 0.006511671468615532f;
  t275_n20: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t275_n34;
  t275_n21: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t275_n23;
  t275_n22: return -0.01832405850291252f;
  t275_n23: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t275_n27;
  t275_n24: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t275_n26;
  t275_n25: return 0.10539340227842331f;
  t275_n26: return 0.06890193372964859f;
  t275_n27: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t275_n31;
  t275_n28: if (E.getNumReferences() >= 3281453056 /*302.0*/) goto t275_n30;
  t275_n29: return -0.0354030504822731f;
  t275_n30: return -0.469855397939682f;
  t275_n31: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t275_n33;
  t275_n32: return 0.040550604462623596f;
  t275_n33: return 0.22794502973556519f;
  t275_n34: if (E.getNumReferences() >= 3292061696 /*739.5*/) goto t275_n36;
  t275_n35: return -0.10538287460803986f;
  t275_n36: if (E.getNumReferences() >= 3310876672 /*3455.5*/) goto t275_n38;
  t275_n37: return 0.147108793258667f;
  t275_n38: return 0.09185352176427841f;
  t275_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n41;
  t275_n40: return -0.0717831626534462f;
  t275_n41: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t275_n205;
  t275_n42: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t275_n94;
  t275_n43: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t275_n65;
  t275_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n46;
  t275_n45: return -0.1941809058189392f;
  t275_n46: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t275_n64;
  t275_n47: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t275_n49;
  t275_n48: return 0.008044593967497349f;
  t275_n49: if (E.getNumReferences() >= 3263758336 /*68.5*/) goto t275_n51;
  t275_n50: return 0.14779676496982574f;
  t275_n51: if (E.getNumReferences() >= 3280404480 /*270.0*/) goto t275_n63;
  t275_n52: if (E.getNumReferences() >= 3268345856 /*103.5*/) goto t275_n56;
  t275_n53: if (E.getNumReferences() >= 3265789952 /*84.0*/) goto t275_n55;
  t275_n54: return 0.09038704633712769f;
  t275_n55: return -0.34897729754447937f;
  t275_n56: if (E.getNumReferences() >= 3274866688 /*178.5*/) goto t275_n62;
  t275_n57: if (E.getNumReferences() >= 3274080256 /*166.5*/) goto t275_n61;
  t275_n58: if (E.getNumReferences() >= 3270574080 /*120.5*/) goto t275_n60;
  t275_n59: return 0.1424558013677597f;
  t275_n60: return 0.04315142706036568f;
  t275_n61: return 0.18508172035217285f;
  t275_n62: return 0.09694814682006836f;
  t275_n63: return -0.026091165840625763f;
  t275_n64: return -0.010039162822067738f;
  t275_n65: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t275_n77;
  t275_n66: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t275_n68;
  t275_n67: return 0.08337413519620895f;
  t275_n68: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t275_n72;
  t275_n69: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t275_n71;
  t275_n70: return 0.15369518101215363f;
  t275_n71: return 0.40806710720062256f;
  t275_n72: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t275_n76;
  t275_n73: if (E.getFileProximityDistanceCost() >= 3231711232 /*5.0*/) goto t275_n75;
  t275_n74: return 0.23810695111751556f;
  t275_n75: return -0.1432460993528366f;
  t275_n76: return 0.14960525929927826f;
  t275_n77: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t275_n85;
  t275_n78: if (E.getNumReferences() >= 3334042112 /*23737.0*/) goto t275_n82;
  t275_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t275_n81;
  t275_n80: return -0.015197151340544224f;
  t275_n81: return 0.11901114881038666f;
  t275_n82: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n84;
  t275_n83: return 0.0771266520023346f;
  t275_n84: return 0.04237362742424011f;
  t275_n85: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n87;
  t275_n86: return 0.037087179720401764f;
  t275_n87: if (E.getNumReferences() >= 3281977344 /*318.0*/) goto t275_n93;
  t275_n88: if (E.getNumReferences() >= 3279323136 /*246.5*/) goto t275_n92;
  t275_n89: if (E.getNumReferences() >= 3269263360 /*110.5*/) goto t275_n91;
  t275_n90: return 0.17676201462745667f;
  t275_n91: return -0.13872887194156647f;
  t275_n92: return 0.27397847175598145f;
  t275_n93: return -0.10159600526094437f;
  t275_n94: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t275_n114;
  t275_n95: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t275_n97;
  t275_n96: return 0.013540792278945446f;
  t275_n97: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t275_n109;
  t275_n98: if (E.getNumReferences() >= 3265069056 /*78.5*/) goto t275_n100;
  t275_n99: return -0.0029395839665085077f;
  t275_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t275_n108;
  t275_n101: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t275_n107;
  t275_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n104;
  t275_n103: return -0.2950431704521179f;
  t275_n104: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n106;
  t275_n105: return -0.027119427919387817f;
  t275_n106: return 0.06917565315961838f;
  t275_n107: return 0.06938831508159637f;
  t275_n108: return 0.10046844929456711f;
  t275_n109: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n113;
  t275_n110: if (E.getNumReferences() >= 3354462528 /*123410.5*/) goto t275_n112;
  t275_n111: return -0.05324548855423927f;
  t275_n112: return -3.3753154277801514f;
  t275_n113: return -0.004186372272670269f;
  t275_n114: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t275_n120;
  t275_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t275_n117;
  t275_n116: return 0.02741674892604351f;
  t275_n117: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t275_n119;
  t275_n118: return -0.10154261440038681f;
  t275_n119: return 0.3979248106479645f;
  t275_n120: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t275_n146;
  t275_n121: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t275_n127;
  t275_n122: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t275_n126;
  t275_n123: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t275_n125;
  t275_n124: return -0.26654210686683655f;
  t275_n125: return -0.013722638599574566f;
  t275_n126: return -0.11149422079324722f;
  t275_n127: if (E.getNumReferences() >= 3266707456 /*91.0*/) goto t275_n145;
  t275_n128: if (E.getNumReferences() >= 3260153856 /*52.5*/) goto t275_n132;
  t275_n129: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t275_n131;
  t275_n130: return 0.0028666530270129442f;
  t275_n131: return 0.08945433050394058f;
  t275_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t275_n138;
  t275_n133: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t275_n137;
  t275_n134: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t275_n136;
  t275_n135: return 0.13430672883987427f;
  t275_n136: return -0.00441723270341754f;
  t275_n137: return -0.06456397473812103f;
  t275_n138: if (E.getNumReferences() >= 3265134592 /*79.0*/) goto t275_n144;
  t275_n139: if (E.getNumReferences() >= 3262513152 /*61.5*/) goto t275_n143;
  t275_n140: if (E.getNumReferences() >= 3261726720 /*58.5*/) goto t275_n142;
  t275_n141: return 0.08075249940156937f;
  t275_n142: return 0.15850205719470978f;
  t275_n143: return -0.010470082983374596f;
  t275_n144: return 0.11617469042539597f;
  t275_n145: return -0.03945392742753029f;
  t275_n146: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t275_n204;
  t275_n147: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t275_n201;
  t275_n148: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t275_n178;
  t275_n149: if (E.getNumReferences() >= 3251372032 /*25.5*/) goto t275_n177;
  t275_n150: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t275_n162;
  t275_n151: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t275_n155;
  t275_n152: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t275_n154;
  t275_n153: return -0.021265344694256783f;
  t275_n154: return 0.04363810643553734f;
  t275_n155: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t275_n157;
  t275_n156: return -0.051516514271497726f;
  t275_n157: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t275_n161;
  t275_n158: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t275_n160;
  t275_n159: return 0.13041600584983826f;
  t275_n160: return 0.08967070281505585f;
  t275_n161: return 0.006769618950784206f;
  t275_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t275_n168;
  t275_n163: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t275_n165;
  t275_n164: return -0.04047468304634094f;
  t275_n165: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t275_n167;
  t275_n166: return 0.07138953357934952f;
  t275_n167: return -0.13155825436115265f;
  t275_n168: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t275_n176;
  t275_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t275_n171;
  t275_n170: return -0.06823164969682693f;
  t275_n171: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t275_n173;
  t275_n172: return 0.056120824068784714f;
  t275_n173: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t275_n175;
  t275_n174: return 0.12190036475658417f;
  t275_n175: return 0.046460844576358795f;
  t275_n176: return 0.11069852113723755f;
  t275_n177: return -0.048192135989665985f;
  t275_n178: if (E.getNumReferences() >= 3253469184 /*29.5*/) goto t275_n186;
  t275_n179: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t275_n183;
  t275_n180: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t275_n182;
  t275_n181: return -0.14156252145767212f;
  t275_n182: return 0.05959658324718475f;
  t275_n183: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t275_n185;
  t275_n184: return -0.032402005046606064f;
  t275_n185: return -0.18229830265045166f;
  t275_n186: if (E.getNumReferences() >= 3255435264 /*34.5*/) goto t275_n192;
  t275_n187: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t275_n191;
  t275_n188: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t275_n190;
  t275_n189: return 0.1444680094718933f;
  t275_n190: return -0.17791233956813812f;
  t275_n191: return 0.15622195601463318f;
  t275_n192: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t275_n196;
  t275_n193: if (E.getNumReferences() >= 3283599360 /*367.5*/) goto t275_n195;
  t275_n194: return 0.013386226259171963f;
  t275_n195: return -0.3707546591758728f;
  t275_n196: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t275_n198;
  t275_n197: return -0.1190553605556488f;
  t275_n198: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t275_n200;
  t275_n199: return 0.14888045191764832f;
  t275_n200: return 0.09634160995483398f;
  t275_n201: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t275_n203;
  t275_n202: return 0.1164303570985794f;
  t275_n203: return -0.061390869319438934f;
  t275_n204: return -0.014152809046208858f;
  t275_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t275_n209;
  t275_n206: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t275_n208;
  t275_n207: return -0.019234564155340195f;
  t275_n208: return 0.00794920139014721f;
  t275_n209: if (E.getNumReferences() >= 3331642112 /*19049.5*/) goto t275_n237;
  t275_n210: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t275_n212;
  t275_n211: return 0.03222213312983513f;
  t275_n212: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t275_n222;
  t275_n213: if (E.getFractionNameInContext() >= 3206445544 /*0.6190476417541504*/) goto t275_n215;
  t275_n214: return 0.07553237676620483f;
  t275_n215: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t275_n217;
  t275_n216: return -0.0013455427251756191f;
  t275_n217: if (E.getNumReferences() >= 3246391296 /*16.0*/) goto t275_n219;
  t275_n218: return -0.23085732758045197f;
  t275_n219: if (E.getNumReferences() >= 3256352768 /*38.0*/) goto t275_n221;
  t275_n220: return 0.275857150554657f;
  t275_n221: return -0.11422621458768845f;
  t275_n222: if (E.getNumReferences() >= 3271589888 /*128.5*/) goto t275_n224;
  t275_n223: return 0.1859825700521469f;
  t275_n224: if (E.getNumReferences() >= 3281125376 /*292.0*/) goto t275_n230;
  t275_n225: if (E.getNumReferences() >= 3272179712 /*137.5*/) goto t275_n229;
  t275_n226: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t275_n228;
  t275_n227: return 0.05207948014140129f;
  t275_n228: return 0.19814862310886383f;
  t275_n229: return -0.09669370204210281f;
  t275_n230: if (E.getNumReferences() >= 3282976768 /*348.5*/) goto t275_n232;
  t275_n231: return 0.15970081090927124f;
  t275_n232: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t275_n234;
  t275_n233: return -0.28047776222229004f;
  t275_n234: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t275_n236;
  t275_n235: return 0.15304826200008392f;
  t275_n236: return -0.09458060562610626f;
  t275_n237: if (E.getNumReferences() >= 3332281088 /*20297.5*/) goto t275_n245;
  t275_n238: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t275_n244;
  t275_n239: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t275_n243;
  t275_n240: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t275_n242;
  t275_n241: return 0.15929162502288818f;
  t275_n242: return 0.10764674097299576f;
  t275_n243: return 0.07977164536714554f;
  t275_n244: return 0.05231042206287384f;
  t275_n245: if (E.getNumReferences() >= 3352364672 /*107021.0*/) goto t275_n247;
  t275_n246: return -0.14153078198432922f;
  t275_n247: return 0.05921294167637825f;
  t275_n248: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t275_n254;
  t275_n249: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t275_n251;
  t275_n250: return 0.004357234574854374f;
  t275_n251: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t275_n253;
  t275_n252: return 0.040800515562295914f;
  t275_n253: return 0.00977100245654583f;
  t275_n254: return -0.0010340693406760693f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree276(const Example& E) {
  t276_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t276_n248;
  t276_n1: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t276_n91;
  t276_n2: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t276_n90;
  t276_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t276_n5;
  t276_n4: return -0.0009657945483922958f;
  t276_n5: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t276_n87;
  t276_n6: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t276_n22;
  t276_n7: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t276_n19;
  t276_n8: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t276_n18;
  t276_n9: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t276_n13;
  t276_n10: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t276_n12;
  t276_n11: return -0.019581977277994156f;
  t276_n12: return 0.06971387565135956f;
  t276_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t276_n15;
  t276_n14: return -0.0003372419159859419f;
  t276_n15: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t276_n17;
  t276_n16: return 0.00702510355040431f;
  t276_n17: return 0.13260358572006226f;
  t276_n18: return -0.008050629869103432f;
  t276_n19: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t276_n21;
  t276_n20: return 0.010450853034853935f;
  t276_n21: return 0.06528730690479279f;
  t276_n22: if (E.getNumReferences() >= 3332825344 /*21360.5*/) goto t276_n84;
  t276_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t276_n27;
  t276_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t276_n26;
  t276_n25: return -0.05250777304172516f;
  t276_n26: return 0.010286040604114532f;
  t276_n27: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t276_n57;
  t276_n28: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t276_n30;
  t276_n29: return 0.011682243086397648f;
  t276_n30: if (E.getNumReferences() >= 3331140352 /*18069.5*/) goto t276_n56;
  t276_n31: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t276_n47;
  t276_n32: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t276_n46;
  t276_n33: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t276_n41;
  t276_n34: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t276_n38;
  t276_n35: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t276_n37;
  t276_n36: return 0.019239775836467743f;
  t276_n37: return -0.20340481400489807f;
  t276_n38: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t276_n40;
  t276_n39: return -0.15112349390983582f;
  t276_n40: return -0.01793511211872101f;
  t276_n41: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t276_n45;
  t276_n42: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t276_n44;
  t276_n43: return -0.09127666056156158f;
  t276_n44: return -0.027783378958702087f;
  t276_n45: return -0.014791904017329216f;
  t276_n46: return 0.0220491960644722f;
  t276_n47: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t276_n53;
  t276_n48: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t276_n50;
  t276_n49: return -0.1375562697649002f;
  t276_n50: if (E.getNumReferences() >= 3297382400 /*1104.5*/) goto t276_n52;
  t276_n51: return 0.05086645111441612f;
  t276_n52: return -0.3149561285972595f;
  t276_n53: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t276_n55;
  t276_n54: return 0.054462894797325134f;
  t276_n55: return 0.003373744897544384f;
  t276_n56: return 0.10876457393169403f;
  t276_n57: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t276_n69;
  t276_n58: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t276_n68;
  t276_n59: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t276_n61;
  t276_n60: return 0.01759721338748932f;
  t276_n61: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t276_n63;
  t276_n62: return 0.021241411566734314f;
  t276_n63: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t276_n67;
  t276_n64: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t276_n66;
  t276_n65: return -0.0003596088499762118f;
  t276_n66: return 0.1876213252544403f;
  t276_n67: return 0.057460684329271317f;
  t276_n68: return 0.056397803127765656f;
  t276_n69: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t276_n71;
  t276_n70: return 0.023954574018716812f;
  t276_n71: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t276_n73;
  t276_n72: return -0.21319101750850677f;
  t276_n73: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t276_n81;
  t276_n74: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t276_n78;
  t276_n75: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t276_n77;
  t276_n76: return 0.09686118364334106f;
  t276_n77: return -0.11719121783971786f;
  t276_n78: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t276_n80;
  t276_n79: return 0.14439883828163147f;
  t276_n80: return 0.0133458711206913f;
  t276_n81: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t276_n83;
  t276_n82: return -2.294180393218994f;
  t276_n83: return 0.027468791231513023f;
  t276_n84: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t276_n86;
  t276_n85: return -0.035499271005392075f;
  t276_n86: return 0.04475614055991173f;
  t276_n87: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t276_n89;
  t276_n88: return 0.001360265421681106f;
  t276_n89: return -0.020973864942789078f;
  t276_n90: return 0.00461661396548152f;
  t276_n91: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t276_n161;
  t276_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t276_n96;
  t276_n93: if (E.getNumReferences() >= 3364188640 /*273295.0*/) goto t276_n95;
  t276_n94: return -0.016672169789671898f;
  t276_n95: return -0.1109352707862854f;
  t276_n96: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t276_n148;
  t276_n97: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t276_n129;
  t276_n98: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t276_n124;
  t276_n99: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t276_n105;
  t276_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t276_n102;
  t276_n101: return -0.01722591556608677f;
  t276_n102: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t276_n104;
  t276_n103: return -0.0073208617977797985f;
  t276_n104: return 0.04592491686344147f;
  t276_n105: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t276_n121;
  t276_n106: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t276_n120;
  t276_n107: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t276_n109;
  t276_n108: return 0.04449954628944397f;
  t276_n109: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t276_n111;
  t276_n110: return -0.008912676945328712f;
  t276_n111: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t276_n117;
  t276_n112: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t276_n116;
  t276_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t276_n115;
  t276_n114: return 0.1365889608860016f;
  t276_n115: return -0.10200362652540207f;
  t276_n116: return -0.028027920052409172f;
  t276_n117: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t276_n119;
  t276_n118: return -0.0780431479215622f;
  t276_n119: return 0.18714606761932373f;
  t276_n120: return -0.009427818469703197f;
  t276_n121: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t276_n123;
  t276_n122: return -0.000827132782433182f;
  t276_n123: return 0.08808711171150208f;
  t276_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t276_n126;
  t276_n125: return 0.018716897815465927f;
  t276_n126: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t276_n128;
  t276_n127: return 0.1718015968799591f;
  t276_n128: return 0.19981545209884644f;
  t276_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t276_n135;
  t276_n130: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t276_n132;
  t276_n131: return 0.0033772673923522234f;
  t276_n132: if (E.getNumReferences() >= 3267821568 /*99.5*/) goto t276_n134;
  t276_n133: return 0.12327919900417328f;
  t276_n134: return 0.045506253838539124f;
  t276_n135: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t276_n145;
  t276_n136: if (E.getNumReferences() >= 3302043648 /*1673.5*/) goto t276_n144;
  t276_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t276_n141;
  t276_n138: if (E.getFractionNameInContext() >= 3174706828 /*0.04545454680919647*/) goto t276_n140;
  t276_n139: return 0.06230400502681732f;
  t276_n140: return 0.1144193783402443f;
  t276_n141: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t276_n143;
  t276_n142: return 0.066253162920475f;
  t276_n143: return 0.1170412003993988f;
  t276_n144: return 0.030293811112642288f;
  t276_n145: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t276_n147;
  t276_n146: return -0.20260243117809296f;
  t276_n147: return -0.024177493527531624f;
  t276_n148: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t276_n150;
  t276_n149: return 0.0028794892132282257f;
  t276_n150: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t276_n154;
  t276_n151: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t276_n153;
  t276_n152: return -0.05726964771747589f;
  t276_n153: return 0.032826680690050125f;
  t276_n154: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t276_n156;
  t276_n155: return 0.007937203161418438f;
  t276_n156: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t276_n160;
  t276_n157: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t276_n159;
  t276_n158: return 0.08241478353738785f;
  t276_n159: return 0.006256320513784885f;
  t276_n160: return -0.02287282980978489f;
  t276_n161: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t276_n247;
  t276_n162: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t276_n240;
  t276_n163: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t276_n239;
  t276_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t276_n228;
  t276_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t276_n213;
  t276_n166: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t276_n202;
  t276_n167: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t276_n169;
  t276_n168: return -0.007550030946731567f;
  t276_n169: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t276_n201;
  t276_n170: if (E.getNumReferences() >= 3246653440 /*16.5*/) goto t276_n200;
  t276_n171: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t276_n187;
  t276_n172: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t276_n186;
  t276_n173: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t276_n183;
  t276_n174: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t276_n176;
  t276_n175: return 0.0357246957719326f;
  t276_n176: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t276_n182;
  t276_n177: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t276_n181;
  t276_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t276_n180;
  t276_n179: return -0.10222882032394409f;
  t276_n180: return 0.08905031532049179f;
  t276_n181: return -0.0707661435008049f;
  t276_n182: return 0.003360023256391287f;
  t276_n183: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t276_n185;
  t276_n184: return 0.4228033721446991f;
  t276_n185: return 0.06695858389139175f;
  t276_n186: return -0.030382169410586357f;
  t276_n187: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t276_n199;
  t276_n188: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t276_n196;
  t276_n189: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t276_n193;
  t276_n190: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t276_n192;
  t276_n191: return 0.06419746577739716f;
  t276_n192: return 0.12806519865989685f;
  t276_n193: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t276_n195;
  t276_n194: return -0.05007272958755493f;
  t276_n195: return 0.0883827805519104f;
  t276_n196: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t276_n198;
  t276_n197: return 0.07585714012384415f;
  t276_n198: return -0.012593698687851429f;
  t276_n199: return 0.24309226870536804f;
  t276_n200: return -0.0282040573656559f;
  t276_n201: return 0.007502318359911442f;
  t276_n202: if (E.getFractionNameInContext() >= 3166318220 /*0.022727273404598236*/) goto t276_n204;
  t276_n203: return -0.0341188870370388f;
  t276_n204: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t276_n206;
  t276_n205: return 0.055481743067502975f;
  t276_n206: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t276_n212;
  t276_n207: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t276_n211;
  t276_n208: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t276_n210;
  t276_n209: return -0.06946586072444916f;
  t276_n210: return -0.02117789350450039f;
  t276_n211: return -0.007429703138768673f;
  t276_n212: return 0.011408981867134571f;
  t276_n213: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t276_n217;
  t276_n214: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t276_n216;
  t276_n215: return 0.08720025420188904f;
  t276_n216: return 0.005422063637524843f;
  t276_n217: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t276_n219;
  t276_n218: return -0.05342138186097145f;
  t276_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t276_n227;
  t276_n220: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t276_n224;
  t276_n221: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t276_n223;
  t276_n222: return -0.01591833308339119f;
  t276_n223: return 0.0440199039876461f;
  t276_n224: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t276_n226;
  t276_n225: return -0.031062370166182518f;
  t276_n226: return -0.13076448440551758f;
  t276_n227: return 0.042297616600990295f;
  t276_n228: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t276_n238;
  t276_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t276_n231;
  t276_n230: return -0.07767973840236664f;
  t276_n231: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t276_n235;
  t276_n232: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t276_n234;
  t276_n233: return 0.01657867804169655f;
  t276_n234: return -0.0010097638005390763f;
  t276_n235: if (E.getNumReferences() >= 3364994640 /*298482.5*/) goto t276_n237;
  t276_n236: return 0.08556574583053589f;
  t276_n237: return -0.13942420482635498f;
  t276_n238: return -0.030940482392907143f;
  t276_n239: return 0.026018181815743446f;
  t276_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t276_n242;
  t276_n241: return 0.009110996499657631f;
  t276_n242: if (E.getNumReferences() >= 3367232976 /*368430.5*/) goto t276_n244;
  t276_n243: return 0.030491255223751068f;
  t276_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t276_n246;
  t276_n245: return 0.05943805351853371f;
  t276_n246: return 0.12645940482616425f;
  t276_n247: return -0.0009564811480231583f;
  t276_n248: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t276_n250;
  t276_n249: return 0.0004818516317754984f;
  t276_n250: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t276_n252;
  t276_n251: return 0.0003361884446348995f;
  t276_n252: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t276_n254;
  t276_n253: return 0.004616008140146732f;
  t276_n254: return 0.012828750535845757f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree277(const Example& E) {
  t277_n0: if (E.getNumReferences() >= 3328217600 /*14372.5*/) goto t277_n186;
  t277_n1: if (E.getNumReferences() >= 3327494144 /*13666.0*/) goto t277_n185;
  t277_n2: if (E.getNumReferences() >= 3326801920 /*12990.0*/) goto t277_n158;
  t277_n3: if (E.getNumReferences() >= 3326525952 /*12720.5*/) goto t277_n157;
  t277_n4: if (E.getNumReferences() >= 3326481408 /*12677.0*/) goto t277_n148;
  t277_n5: if (E.getNumReferences() >= 3324152832 /*10403.0*/) goto t277_n107;
  t277_n6: if (E.getNumReferences() >= 3324042752 /*10295.5*/) goto t277_n100;
  t277_n7: if (E.getNumReferences() >= 3321325568 /*7917.0*/) goto t277_n9;
  t277_n8: return -7.641374395461753e-05f;
  t277_n9: if (E.getNumReferences() >= 3322744320 /*9027.5*/) goto t277_n59;
  t277_n10: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n12;
  t277_n11: return 0.007331099361181259f;
  t277_n12: if (E.getNumReferences() >= 3322147840 /*8445.0*/) goto t277_n46;
  t277_n13: if (E.getNumReferences() >= 3322037760 /*8337.5*/) goto t277_n29;
  t277_n14: if (E.getNumReferences() >= 3321819136 /*8158.0*/) goto t277_n28;
  t277_n15: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t277_n17;
  t277_n16: return 0.008276225998997688f;
  t277_n17: if (E.getNumReferences() >= 3321408512 /*7957.5*/) goto t277_n19;
  t277_n18: return -0.05161791294813156f;
  t277_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t277_n27;
  t277_n20: if (E.getNumReferences() >= 3321413632 /*7960.0*/) goto t277_n26;
  t277_n21: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t277_n25;
  t277_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t277_n24;
  t277_n23: return 0.13776181638240814f;
  t277_n24: return 0.12564416229724884f;
  t277_n25: return 0.05188019201159477f;
  t277_n26: return 0.1445208489894867f;
  t277_n27: return 0.1471228003501892f;
  t277_n28: return -0.0014758111210539937f;
  t277_n29: if (E.getNumReferences() >= 3322086912 /*8385.5*/) goto t277_n39;
  t277_n30: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t277_n34;
  t277_n31: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t277_n33;
  t277_n32: return 0.09033394604921341f;
  t277_n33: return 0.11511650681495667f;
  t277_n34: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t277_n38;
  t277_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n37;
  t277_n36: return -0.1024213656783104f;
  t277_n37: return 0.21822252869606018f;
  t277_n38: return -0.10027886927127838f;
  t277_n39: if (E.getNumReferences() >= 3322118656 /*8416.5*/) goto t277_n41;
  t277_n40: return 0.17186565697193146f;
  t277_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t277_n43;
  t277_n42: return 0.08901721239089966f;
  t277_n43: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t277_n45;
  t277_n44: return 0.1437850445508957f;
  t277_n45: return -0.12094645202159882f;
  t277_n46: if (E.getNumReferences() >= 3322597376 /*8884.0*/) goto t277_n48;
  t277_n47: return -0.04752618074417114f;
  t277_n48: if (E.getNumReferences() >= 3322619392 /*8905.5*/) goto t277_n54;
  t277_n49: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t277_n53;
  t277_n50: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t277_n52;
  t277_n51: return 0.1857932060956955f;
  t277_n52: return 0.07472161948680878f;
  t277_n53: return 0.11819352954626083f;
  t277_n54: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t277_n56;
  t277_n55: return 0.004603848792612553f;
  t277_n56: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t277_n58;
  t277_n57: return 1.150951862335205f;
  t277_n58: return -0.11765310913324356f;
  t277_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t277_n67;
  t277_n60: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t277_n66;
  t277_n61: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n63;
  t277_n62: return -0.10733892023563385f;
  t277_n63: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t277_n65;
  t277_n64: return -0.06620621681213379f;
  t277_n65: return -0.1710485816001892f;
  t277_n66: return 0.0017099534161388874f;
  t277_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t277_n69;
  t277_n68: return 0.00418293010443449f;
  t277_n69: if (E.getNumReferences() >= 3323824640 /*10082.5*/) goto t277_n85;
  t277_n70: if (E.getNumReferences() >= 3323060224 /*9336.0*/) goto t277_n82;
  t277_n71: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t277_n77;
  t277_n72: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t277_n76;
  t277_n73: if (E.getNumReferences() >= 3322854400 /*9135.0*/) goto t277_n75;
  t277_n74: return -0.1419990062713623f;
  t277_n75: return 0.04004047438502312f;
  t277_n76: return 0.15143461525440216f;
  t277_n77: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t277_n81;
  t277_n78: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t277_n80;
  t277_n79: return 0.19883500039577484f;
  t277_n80: return 0.10588637739419937f;
  t277_n81: return 0.06823670864105225f;
  t277_n82: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t277_n84;
  t277_n83: return -0.015142427757382393f;
  t277_n84: return -0.2852044403553009f;
  t277_n85: if (E.getNumReferences() >= 3323889152 /*10145.5*/) goto t277_n97;
  t277_n86: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t277_n90;
  t277_n87: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t277_n89;
  t277_n88: return 0.16960546374320984f;
  t277_n89: return 0.12151312828063965f;
  t277_n90: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t277_n94;
  t277_n91: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t277_n93;
  t277_n92: return 0.1271021068096161f;
  t277_n93: return 0.09748747199773788f;
  t277_n94: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t277_n96;
  t277_n95: return 0.04344042018055916f;
  t277_n96: return 0.13015812635421753f;
  t277_n97: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n99;
  t277_n98: return 0.002714512636885047f;
  t277_n99: return 0.11181073635816574f;
  t277_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t277_n102;
  t277_n101: return 0.015644898638129234f;
  t277_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t277_n104;
  t277_n103: return 0.11703060567378998f;
  t277_n104: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t277_n106;
  t277_n105: return -0.10649281740188599f;
  t277_n106: return 0.16839613020420074f;
  t277_n107: if (E.getNumReferences() >= 3324729856 /*10966.5*/) goto t277_n109;
  t277_n108: return -0.05777952820062637f;
  t277_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t277_n131;
  t277_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t277_n124;
  t277_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t277_n113;
  t277_n112: return -0.020887773483991623f;
  t277_n113: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t277_n121;
  t277_n114: if (E.getNumReferences() >= 3325640192 /*11855.5*/) goto t277_n116;
  t277_n115: return -0.02136162295937538f;
  t277_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n118;
  t277_n117: return -0.11815718561410904f;
  t277_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t277_n120;
  t277_n119: return 0.06811664253473282f;
  t277_n120: return 0.12457503378391266f;
  t277_n121: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t277_n123;
  t277_n122: return 0.15392278134822845f;
  t277_n123: return -0.10172607749700546f;
  t277_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t277_n128;
  t277_n125: if (E.getNumReferences() >= 3324983296 /*11214.0*/) goto t277_n127;
  t277_n126: return -0.23405736684799194f;
  t277_n127: return 0.07426393777132034f;
  t277_n128: if (E.getNumReferences() >= 3325331968 /*11554.5*/) goto t277_n130;
  t277_n129: return 0.13949079811573029f;
  t277_n130: return 0.10338879376649857f;
  t277_n131: if (E.getNumReferences() >= 3324868096 /*11101.5*/) goto t277_n147;
  t277_n132: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t277_n134;
  t277_n133: return -0.123187355697155f;
  t277_n134: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t277_n146;
  t277_n135: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t277_n145;
  t277_n136: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t277_n142;
  t277_n137: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n139;
  t277_n138: return -0.0017585627501830459f;
  t277_n139: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t277_n141;
  t277_n140: return 0.41146108508110046f;
  t277_n141: return 0.09575733542442322f;
  t277_n142: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t277_n144;
  t277_n143: return -0.03970901668071747f;
  t277_n144: return 0.15885113179683685f;
  t277_n145: return 0.00617872504517436f;
  t277_n146: return -0.10045816749334335f;
  t277_n147: return 0.025091471150517464f;
  t277_n148: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t277_n152;
  t277_n149: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t277_n151;
  t277_n150: return 0.07442381978034973f;
  t277_n151: return 0.30916252732276917f;
  t277_n152: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t277_n156;
  t277_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t277_n155;
  t277_n154: return 0.10847757011651993f;
  t277_n155: return 0.14352937042713165f;
  t277_n156: return 0.07777668535709381f;
  t277_n157: return -0.08421371132135391f;
  t277_n158: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t277_n182;
  t277_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t277_n173;
  t277_n160: if (E.getNumReferences() >= 3326945792 /*13130.5*/) goto t277_n170;
  t277_n161: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n167;
  t277_n162: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t277_n166;
  t277_n163: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t277_n165;
  t277_n164: return 0.09331492334604263f;
  t277_n165: return 0.2079620063304901f;
  t277_n166: return 0.040442414581775665f;
  t277_n167: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t277_n169;
  t277_n168: return 0.1345331221818924f;
  t277_n169: return 0.081771120429039f;
  t277_n170: if (E.getNumReferences() >= 3326974976 /*13159.0*/) goto t277_n172;
  t277_n171: return -0.28136903047561646f;
  t277_n172: return 0.014225159771740437f;
  t277_n173: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n181;
  t277_n174: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t277_n176;
  t277_n175: return 0.10190887004137039f;
  t277_n176: if (E.getNumReferences() >= 3327426048 /*13599.5*/) goto t277_n180;
  t277_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n179;
  t277_n178: return -0.10101968050003052f;
  t277_n179: return 1.9708610773086548f;
  t277_n180: return 0.03745046257972717f;
  t277_n181: return 0.008417613804340363f;
  t277_n182: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t277_n184;
  t277_n183: return 0.032474979758262634f;
  t277_n184: return -0.055144187062978745f;
  t277_n185: return -0.05111447349190712f;
  t277_n186: if (E.getNumReferences() >= 3329021952 /*15158.0*/) goto t277_n218;
  t277_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t277_n211;
  t277_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n198;
  t277_n189: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t277_n191;
  t277_n190: return -0.03525437042117119f;
  t277_n191: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t277_n197;
  t277_n192: if (E.getNumReferences() >= 3328450048 /*14599.5*/) goto t277_n196;
  t277_n193: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t277_n195;
  t277_n194: return 0.2082744836807251f;
  t277_n195: return 0.10407532751560211f;
  t277_n196: return -0.039773374795913696f;
  t277_n197: return -0.11209006607532501f;
  t277_n198: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t277_n204;
  t277_n199: if (E.getNumReferences() >= 3328254464 /*14408.5*/) goto t277_n201;
  t277_n200: return 0.1280476599931717f;
  t277_n201: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t277_n203;
  t277_n202: return -0.03505285084247589f;
  t277_n203: return 0.0712692067027092f;
  t277_n204: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n210;
  t277_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t277_n207;
  t277_n206: return -0.3016008138656616f;
  t277_n207: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t277_n209;
  t277_n208: return 0.1609697788953781f;
  t277_n209: return -0.06047037988901138f;
  t277_n210: return 0.03122262842953205f;
  t277_n211: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t277_n213;
  t277_n212: return 0.03386608138680458f;
  t277_n213: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t277_n215;
  t277_n214: return 0.05028427019715309f;
  t277_n215: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t277_n217;
  t277_n216: return 0.10715392976999283f;
  t277_n217: return 0.10803373903036118f;
  t277_n218: if (E.getNumReferences() >= 3329553408 /*15677.0*/) goto t277_n224;
  t277_n219: if (E.getNumReferences() >= 3329258496 /*15389.0*/) goto t277_n223;
  t277_n220: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t277_n222;
  t277_n221: return -0.02457386627793312f;
  t277_n222: return 0.06888314336538315f;
  t277_n223: return -0.1088445857167244f;
  t277_n224: if (E.getNumReferences() >= 3330013184 /*16126.0*/) goto t277_n250;
  t277_n225: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t277_n249;
  t277_n226: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t277_n240;
  t277_n227: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t277_n231;
  t277_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t277_n230;
  t277_n229: return 0.02967831864953041f;
  t277_n230: return 0.12934301793575287f;
  t277_n231: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t277_n235;
  t277_n232: if (E.getNumReferences() >= 3329564160 /*15687.5*/) goto t277_n234;
  t277_n233: return 0.07383687049150467f;
  t277_n234: return 0.03594502434134483f;
  t277_n235: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t277_n237;
  t277_n236: return 0.10601187497377396f;
  t277_n237: if (E.getFractionNameInContext() >= 3183239395 /*0.0919816717505455*/) goto t277_n239;
  t277_n238: return 0.1334044486284256f;
  t277_n239: return 0.12129112333059311f;
  t277_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t277_n242;
  t277_n241: return -0.00017187716730404645f;
  t277_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t277_n248;
  t277_n243: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t277_n247;
  t277_n244: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t277_n246;
  t277_n245: return -0.10057473182678223f;
  t277_n246: return 0.1897999495267868f;
  t277_n247: return 0.09089646488428116f;
  t277_n248: return 0.12430448085069656f;
  t277_n249: return -0.07421792298555374f;
  t277_n250: if (E.getNumReferences() >= 3330058752 /*16170.5*/) goto t277_n252;
  t277_n251: return -0.1803659051656723f;
  t277_n252: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t277_n254;
  t277_n253: return -0.009365526027977467f;
  t277_n254: return 0.0035041123628616333f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree278(const Example& E) {
  t278_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t278_n248;
  t278_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t278_n3;
  t278_n2: return -0.0004272129153832793f;
  t278_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t278_n97;
  t278_n4: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t278_n6;
  t278_n5: return -0.006157553289085627f;
  t278_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t278_n26;
  t278_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t278_n15;
  t278_n8: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t278_n14;
  t278_n9: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n11;
  t278_n10: return -0.07745368033647537f;
  t278_n11: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t278_n13;
  t278_n12: return -0.0020220403093844652f;
  t278_n13: return -0.10507465153932571f;
  t278_n14: return -0.006678012665361166f;
  t278_n15: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t278_n19;
  t278_n16: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t278_n18;
  t278_n17: return -0.00209369626827538f;
  t278_n18: return 0.003001431468874216f;
  t278_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n21;
  t278_n20: return -0.005520411767065525f;
  t278_n21: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t278_n23;
  t278_n22: return 0.0021026807371526957f;
  t278_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t278_n25;
  t278_n24: return -0.007002220489084721f;
  t278_n25: return 0.006556342821568251f;
  t278_n26: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t278_n90;
  t278_n27: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t278_n31;
  t278_n28: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t278_n30;
  t278_n29: return -0.1316683292388916f;
  t278_n30: return 0.0017439215444028378f;
  t278_n31: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t278_n33;
  t278_n32: return 0.008498339913785458f;
  t278_n33: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t278_n87;
  t278_n34: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t278_n72;
  t278_n35: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t278_n55;
  t278_n36: if (E.getFractionNameInContext() >= 3206335693 /*0.612500011920929*/) goto t278_n40;
  t278_n37: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t278_n39;
  t278_n38: return 0.02901238575577736f;
  t278_n39: return 0.04323660582304001f;
  t278_n40: if (E.getNumReferences() >= 3250323456 /*23.5*/) goto t278_n54;
  t278_n41: if (E.getNumNameInContext() >= 3225419776 /*3.0*/) goto t278_n49;
  t278_n42: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t278_n44;
  t278_n43: return -0.25202617049217224f;
  t278_n44: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t278_n48;
  t278_n45: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n47;
  t278_n46: return 0.07323724031448364f;
  t278_n47: return -0.1961524933576584f;
  t278_n48: return -0.1177077442407608f;
  t278_n49: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t278_n53;
  t278_n50: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n52;
  t278_n51: return 0.03725242242217064f;
  t278_n52: return 0.07785462588071823f;
  t278_n53: return -0.67722487449646f;
  t278_n54: return 0.054804105311632156f;
  t278_n55: if (E.getNumReferences() >= 3247177728 /*17.5*/) goto t278_n65;
  t278_n56: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t278_n64;
  t278_n57: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t278_n61;
  t278_n58: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t278_n60;
  t278_n59: return -0.009406396187841892f;
  t278_n60: return -0.2155071198940277f;
  t278_n61: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t278_n63;
  t278_n62: return -0.046286292374134064f;
  t278_n63: return 0.07971157878637314f;
  t278_n64: return -0.01854240521788597f;
  t278_n65: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t278_n67;
  t278_n66: return 0.8656002283096313f;
  t278_n67: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t278_n69;
  t278_n68: return 0.016173573210835457f;
  t278_n69: if (E.getNumReferences() >= 3278503936 /*234.0*/) goto t278_n71;
  t278_n70: return 0.1494009792804718f;
  t278_n71: return -0.1623237282037735f;
  t278_n72: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t278_n82;
  t278_n73: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t278_n77;
  t278_n74: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t278_n76;
  t278_n75: return -0.20484600961208344f;
  t278_n76: return -0.008902235887944698f;
  t278_n77: if (E.getFractionNameInContext() >= 3210390186 /*0.8541666269302368*/) goto t278_n79;
  t278_n78: return 0.06188097968697548f;
  t278_n79: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t278_n81;
  t278_n80: return 0.04353208839893341f;
  t278_n81: return -0.24480144679546356f;
  t278_n82: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t278_n86;
  t278_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n85;
  t278_n84: return -0.10097780078649521f;
  t278_n85: return 0.1318020224571228f;
  t278_n86: return -0.1021943986415863f;
  t278_n87: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t278_n89;
  t278_n88: return 0.16861245036125183f;
  t278_n89: return 0.08526869118213654f;
  t278_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t278_n96;
  t278_n91: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t278_n95;
  t278_n92: if (E.getNumReferences() >= 3355241152 /*129493.5*/) goto t278_n94;
  t278_n93: return -0.008294439874589443f;
  t278_n94: return -0.21545323729515076f;
  t278_n95: return -0.11585478484630585f;
  t278_n96: return 0.0030462932772934437f;
  t278_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t278_n151;
  t278_n98: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t278_n134;
  t278_n99: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t278_n121;
  t278_n100: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t278_n114;
  t278_n101: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n113;
  t278_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n104;
  t278_n103: return -0.008128098212182522f;
  t278_n104: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n112;
  t278_n105: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t278_n111;
  t278_n106: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n108;
  t278_n107: return -0.0041506350971758366f;
  t278_n108: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t278_n110;
  t278_n109: return -0.44207850098609924f;
  t278_n110: return -0.05921999737620354f;
  t278_n111: return 0.028943508863449097f;
  t278_n112: return 0.02498052828013897f;
  t278_n113: return 0.010010361671447754f;
  t278_n114: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t278_n118;
  t278_n115: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t278_n117;
  t278_n116: return -0.01208247896283865f;
  t278_n117: return 0.016915250569581985f;
  t278_n118: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t278_n120;
  t278_n119: return 0.1251825988292694f;
  t278_n120: return 0.07172486931085587f;
  t278_n121: if (E.getFractionNameInContext() >= 3210240390 /*0.8452380895614624*/) goto t278_n133;
  t278_n122: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n132;
  t278_n123: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t278_n131;
  t278_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t278_n130;
  t278_n125: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t278_n129;
  t278_n126: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t278_n128;
  t278_n127: return -0.02965199388563633f;
  t278_n128: return 0.1001230925321579f;
  t278_n129: return -0.03401442617177963f;
  t278_n130: return 0.10992980748414993f;
  t278_n131: return 0.014307617209851742f;
  t278_n132: return -0.06906836479902267f;
  t278_n133: return 0.0806574672460556f;
  t278_n134: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n144;
  t278_n135: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t278_n137;
  t278_n136: return -0.01528547890484333f;
  t278_n137: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n141;
  t278_n138: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t278_n140;
  t278_n139: return 0.044688090682029724f;
  t278_n140: return 0.1309071183204651f;
  t278_n141: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t278_n143;
  t278_n142: return 0.023549752309918404f;
  t278_n143: return -0.16780488193035126f;
  t278_n144: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n148;
  t278_n145: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n147;
  t278_n146: return -0.02389409951865673f;
  t278_n147: return 0.004602217581123114f;
  t278_n148: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t278_n150;
  t278_n149: return 0.013414701446890831f;
  t278_n150: return -0.017237482592463493f;
  t278_n151: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t278_n199;
  t278_n152: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t278_n188;
  t278_n153: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t278_n179;
  t278_n154: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t278_n158;
  t278_n155: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t278_n157;
  t278_n156: return 0.004165898077189922f;
  t278_n157: return 0.43572309613227844f;
  t278_n158: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n168;
  t278_n159: if (E.getFractionNameInContext() >= 3188549847 /*0.13809524476528168*/) goto t278_n163;
  t278_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t278_n162;
  t278_n161: return 0.05114275589585304f;
  t278_n162: return 0.08627994358539581f;
  t278_n163: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n167;
  t278_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t278_n166;
  t278_n165: return 0.003306516446173191f;
  t278_n166: return 0.03167149797081947f;
  t278_n167: return -0.11376953125f;
  t278_n168: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n176;
  t278_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t278_n173;
  t278_n170: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n172;
  t278_n171: return 0.06038025766611099f;
  t278_n172: return -0.0454183854162693f;
  t278_n173: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n175;
  t278_n174: return 0.0866168811917305f;
  t278_n175: return 0.1058914065361023f;
  t278_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n178;
  t278_n177: return -0.16813194751739502f;
  t278_n178: return -0.00921063032001257f;
  t278_n179: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n181;
  t278_n180: return -0.0152140436694026f;
  t278_n181: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t278_n183;
  t278_n182: return 0.005698629654943943f;
  t278_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n185;
  t278_n184: return -0.02476922981441021f;
  t278_n185: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t278_n187;
  t278_n186: return 0.08033367246389389f;
  t278_n187: return 0.21207594871520996f;
  t278_n188: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n196;
  t278_n189: if (E.getFractionNameInContext() >= 3195014913 /*0.23443223536014557*/) goto t278_n193;
  t278_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n192;
  t278_n191: return -0.15892240405082703f;
  t278_n192: return 0.11997673660516739f;
  t278_n193: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n195;
  t278_n194: return 0.020421752706170082f;
  t278_n195: return 0.008169997483491898f;
  t278_n196: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t278_n198;
  t278_n197: return -0.02262839674949646f;
  t278_n198: return 0.09965476393699646f;
  t278_n199: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t278_n239;
  t278_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n206;
  t278_n201: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t278_n205;
  t278_n202: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n204;
  t278_n203: return -0.03350822255015373f;
  t278_n204: return -0.13774192333221436f;
  t278_n205: return 0.021092142909765244f;
  t278_n206: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t278_n228;
  t278_n207: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t278_n219;
  t278_n208: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t278_n218;
  t278_n209: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n217;
  t278_n210: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t278_n212;
  t278_n211: return -0.0351678766310215f;
  t278_n212: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n214;
  t278_n213: return -0.027605576440691948f;
  t278_n214: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t278_n216;
  t278_n215: return 0.04472828283905983f;
  t278_n216: return -0.06497775763273239f;
  t278_n217: return 0.035476088523864746f;
  t278_n218: return 0.05034945532679558f;
  t278_n219: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t278_n223;
  t278_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t278_n222;
  t278_n221: return -0.14714157581329346f;
  t278_n222: return -0.12764589488506317f;
  t278_n223: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n225;
  t278_n224: return -0.05298245698213577f;
  t278_n225: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n227;
  t278_n226: return -0.059588655829429626f;
  t278_n227: return 0.07438682019710541f;
  t278_n228: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t278_n238;
  t278_n229: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t278_n235;
  t278_n230: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t278_n232;
  t278_n231: return 0.052604783326387405f;
  t278_n232: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t278_n234;
  t278_n233: return 0.029003262519836426f;
  t278_n234: return -0.09560336917638779f;
  t278_n235: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t278_n237;
  t278_n236: return 0.1313956379890442f;
  t278_n237: return 0.008179529570043087f;
  t278_n238: return -0.12044935673475266f;
  t278_n239: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t278_n241;
  t278_n240: return -0.0766461119055748f;
  t278_n241: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t278_n243;
  t278_n242: return 0.020878272131085396f;
  t278_n243: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t278_n247;
  t278_n244: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t278_n246;
  t278_n245: return 0.11073604226112366f;
  t278_n246: return 0.12808075547218323f;
  t278_n247: return -0.17634665966033936f;
  t278_n248: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t278_n254;
  t278_n249: if (E.getScope() & (BIT(Scope_type::FunctionScope))) goto t278_n251;
  t278_n250: return 0.0002257624437334016f;
  t278_n251: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t278_n253;
  t278_n252: return 0.009375886991620064f;
  t278_n253: return -0.3506873548030853f;
  t278_n254: return 0.003264110069721937f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree279(const Example& E) {
  t279_n0: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t279_n2;
  t279_n1: return -0.00010324580216547474f;
  t279_n2: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t279_n60;
  t279_n3: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t279_n41;
  t279_n4: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t279_n34;
  t279_n5: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t279_n21;
  t279_n6: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t279_n16;
  t279_n7: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t279_n9;
  t279_n8: return 0.000605607230681926f;
  t279_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t279_n11;
  t279_n10: return 0.026550451293587685f;
  t279_n11: if (E.getFractionNameInContext() >= 3203516188 /*0.4722222089767456*/) goto t279_n13;
  t279_n12: return 0.12394938617944717f;
  t279_n13: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t279_n15;
  t279_n14: return -0.13332630693912506f;
  t279_n15: return 0.13177257776260376f;
  t279_n16: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t279_n18;
  t279_n17: return -0.28156018257141113f;
  t279_n18: if (E.getNumReferences() >= 3254910976 /*32.5*/) goto t279_n20;
  t279_n19: return 0.022890042513608932f;
  t279_n20: return -0.37597933411598206f;
  t279_n21: if (E.getFractionNameInContext() >= 3181561674 /*0.07948167622089386*/) goto t279_n27;
  t279_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t279_n24;
  t279_n23: return 0.10138405114412308f;
  t279_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t279_n26;
  t279_n25: return 0.07233163714408875f;
  t279_n26: return 0.13505318760871887f;
  t279_n27: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t279_n29;
  t279_n28: return 0.1037333607673645f;
  t279_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t279_n33;
  t279_n30: if (E.getFractionNameInContext() >= 3193263446 /*0.2083333432674408*/) goto t279_n32;
  t279_n31: return -0.5102033019065857f;
  t279_n32: return -0.005118700210005045f;
  t279_n33: return 0.09226870536804199f;
  t279_n34: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t279_n40;
  t279_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n37;
  t279_n36: return -0.10370370000600815f;
  t279_n37: if (E.getNumReferences() >= 3265331200 /*80.5*/) goto t279_n39;
  t279_n38: return 0.041051801294088364f;
  t279_n39: return -0.37423548102378845f;
  t279_n40: return -0.24921177327632904f;
  t279_n41: if (E.getNumReferences() >= 3268804608 /*107.0*/) goto t279_n47;
  t279_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t279_n46;
  t279_n43: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t279_n45;
  t279_n44: return 0.09924788773059845f;
  t279_n45: return -0.20009024441242218f;
  t279_n46: return 0.1227714866399765f;
  t279_n47: if (E.getNumReferences() >= 3270311936 /*118.5*/) goto t279_n49;
  t279_n48: return -0.16764762997627258f;
  t279_n49: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t279_n59;
  t279_n50: if (E.getNumReferences() >= 3271786496 /*131.5*/) goto t279_n52;
  t279_n51: return 0.11175883561372757f;
  t279_n52: if (E.getNumReferences() >= 3274539008 /*173.5*/) goto t279_n54;
  t279_n53: return -0.17780271172523499f;
  t279_n54: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n58;
  t279_n55: if (E.getNumReferences() >= 3274735616 /*176.5*/) goto t279_n57;
  t279_n56: return 0.10591867566108704f;
  t279_n57: return 0.016053898259997368f;
  t279_n58: return 0.0843639150261879f;
  t279_n59: return 0.11174428462982178f;
  t279_n60: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t279_n168;
  t279_n61: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t279_n167;
  t279_n62: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n64;
  t279_n63: return -0.01673804596066475f;
  t279_n64: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t279_n152;
  t279_n65: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t279_n151;
  t279_n66: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t279_n128;
  t279_n67: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t279_n111;
  t279_n68: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t279_n88;
  t279_n69: if (E.getNumReferences() >= 3324478464 /*10721.0*/) goto t279_n83;
  t279_n70: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t279_n74;
  t279_n71: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t279_n73;
  t279_n72: return -0.1382266730070114f;
  t279_n73: return 0.06553822755813599f;
  t279_n74: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t279_n76;
  t279_n75: return 0.023140033707022667f;
  t279_n76: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t279_n80;
  t279_n77: if (E.getNumReferences() >= 3264217088 /*72.0*/) goto t279_n79;
  t279_n78: return -0.030761003494262695f;
  t279_n79: return -0.6560274362564087f;
  t279_n80: if (E.getNumReferences() >= 3301154816 /*1565.0*/) goto t279_n82;
  t279_n81: return -0.02966233715415001f;
  t279_n82: return -0.41609495878219604f;
  t279_n83: if (E.getNumReferences() >= 3349292864 /*83022.5*/) goto t279_n87;
  t279_n84: if (E.getNumReferences() >= 3339744256 /*36980.0*/) goto t279_n86;
  t279_n85: return 0.11625964194536209f;
  t279_n86: return 0.18370972573757172f;
  t279_n87: return 0.04278842732310295f;
  t279_n88: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t279_n110;
  t279_n89: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t279_n103;
  t279_n90: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t279_n92;
  t279_n91: return -0.1769106537103653f;
  t279_n92: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t279_n96;
  t279_n93: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t279_n95;
  t279_n94: return 0.08212263882160187f;
  t279_n95: return 0.09442684799432755f;
  t279_n96: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t279_n98;
  t279_n97: return 0.11047646403312683f;
  t279_n98: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t279_n100;
  t279_n99: return -0.025656381621956825f;
  t279_n100: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t279_n102;
  t279_n101: return 0.10953187942504883f;
  t279_n102: return 0.008968859910964966f;
  t279_n103: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t279_n107;
  t279_n104: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n106;
  t279_n105: return 0.12587878108024597f;
  t279_n106: return 0.07895570993423462f;
  t279_n107: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t279_n109;
  t279_n108: return 0.03493119776248932f;
  t279_n109: return 0.1039828211069107f;
  t279_n110: return -0.028717007488012314f;
  t279_n111: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t279_n113;
  t279_n112: return -0.00019356637494638562f;
  t279_n113: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t279_n125;
  t279_n114: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t279_n116;
  t279_n115: return -0.535735011100769f;
  t279_n116: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t279_n122;
  t279_n117: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t279_n119;
  t279_n118: return -0.08296842873096466f;
  t279_n119: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t279_n121;
  t279_n120: return -0.1283629983663559f;
  t279_n121: return 0.16408449411392212f;
  t279_n122: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t279_n124;
  t279_n123: return -0.059037454426288605f;
  t279_n124: return -0.2264515608549118f;
  t279_n125: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t279_n127;
  t279_n126: return 0.1874329149723053f;
  t279_n127: return 0.02607531100511551f;
  t279_n128: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t279_n148;
  t279_n129: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t279_n131;
  t279_n130: return 0.08388136327266693f;
  t279_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n137;
  t279_n132: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t279_n136;
  t279_n133: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t279_n135;
  t279_n134: return -0.21908219158649445f;
  t279_n135: return -0.7067705988883972f;
  t279_n136: return -0.06541140377521515f;
  t279_n137: if (E.getNumReferences() >= 3244294144 /*14.0*/) goto t279_n147;
  t279_n138: if (E.getFractionNameInContext() >= 3202770534 /*0.44999998807907104*/) goto t279_n146;
  t279_n139: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n143;
  t279_n140: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t279_n142;
  t279_n141: return -0.05639920383691788f;
  t279_n142: return -0.20298460125923157f;
  t279_n143: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t279_n145;
  t279_n144: return -0.20188063383102417f;
  t279_n145: return -0.6463026404380798f;
  t279_n146: return 0.030213341116905212f;
  t279_n147: return 0.03434648737311363f;
  t279_n148: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t279_n150;
  t279_n149: return -0.04166421666741371f;
  t279_n150: return 0.049358487129211426f;
  t279_n151: return 0.03465850651264191f;
  t279_n152: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t279_n154;
  t279_n153: return 0.06414251774549484f;
  t279_n154: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t279_n158;
  t279_n155: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t279_n157;
  t279_n156: return -0.1114160567522049f;
  t279_n157: return -1.8925219774246216f;
  t279_n158: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t279_n164;
  t279_n159: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t279_n163;
  t279_n160: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t279_n162;
  t279_n161: return 0.1125035285949707f;
  t279_n162: return 0.02612427994608879f;
  t279_n163: return 0.02855616807937622f;
  t279_n164: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t279_n166;
  t279_n165: return -0.22718143463134766f;
  t279_n166: return 0.08502738177776337f;
  t279_n167: return 0.033956412225961685f;
  t279_n168: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t279_n226;
  t279_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t279_n195;
  t279_n170: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t279_n174;
  t279_n171: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t279_n173;
  t279_n172: return -0.0008150962530635297f;
  t279_n173: return 0.03815658390522003f;
  t279_n174: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n182;
  t279_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t279_n177;
  t279_n176: return 0.03323635831475258f;
  t279_n177: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t279_n181;
  t279_n178: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t279_n180;
  t279_n179: return -0.41596490144729614f;
  t279_n180: return -0.14362812042236328f;
  t279_n181: return -0.08083230257034302f;
  t279_n182: if (E.getNumReferences() >= 3310524416 /*3369.5*/) goto t279_n194;
  t279_n183: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t279_n185;
  t279_n184: return 0.06636658310890198f;
  t279_n185: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t279_n191;
  t279_n186: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t279_n190;
  t279_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n189;
  t279_n188: return -0.1592172086238861f;
  t279_n189: return 0.018507005646824837f;
  t279_n190: return 0.037728242576122284f;
  t279_n191: if (E.getNumReferences() >= 3281698816 /*309.5*/) goto t279_n193;
  t279_n192: return 0.1666877716779709f;
  t279_n193: return 0.11466530710458755f;
  t279_n194: return 0.018338466063141823f;
  t279_n195: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t279_n225;
  t279_n196: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t279_n222;
  t279_n197: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t279_n199;
  t279_n198: return 0.0287712961435318f;
  t279_n199: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t279_n219;
  t279_n200: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t279_n208;
  t279_n201: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t279_n203;
  t279_n202: return 0.043387629091739655f;
  t279_n203: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t279_n207;
  t279_n204: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t279_n206;
  t279_n205: return -0.019150827080011368f;
  t279_n206: return 0.046160921454429626f;
  t279_n207: return -0.6111412048339844f;
  t279_n208: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t279_n214;
  t279_n209: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t279_n213;
  t279_n210: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n212;
  t279_n211: return 0.10559382289648056f;
  t279_n212: return 0.0255274660885334f;
  t279_n213: return -0.3661811947822571f;
  t279_n214: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n216;
  t279_n215: return 0.1536083221435547f;
  t279_n216: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t279_n218;
  t279_n217: return -0.15497159957885742f;
  t279_n218: return 0.07760561257600784f;
  t279_n219: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n221;
  t279_n220: return 0.10522758960723877f;
  t279_n221: return -0.4203464090824127f;
  t279_n222: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n224;
  t279_n223: return 0.0975220799446106f;
  t279_n224: return 0.0722082182765007f;
  t279_n225: return 0.012021339498460293f;
  t279_n226: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t279_n238;
  t279_n227: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t279_n237;
  t279_n228: if (E.getFractionNameInContext() >= 3204914290 /*0.5277777910232544*/) goto t279_n234;
  t279_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n233;
  t279_n230: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t279_n232;
  t279_n231: return -0.15587377548217773f;
  t279_n232: return 0.10621203482151031f;
  t279_n233: return 0.04053334519267082f;
  t279_n234: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t279_n236;
  t279_n235: return 0.11119631677865982f;
  t279_n236: return 0.07457973062992096f;
  t279_n237: return 0.0011457462096586823f;
  t279_n238: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t279_n252;
  t279_n239: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t279_n245;
  t279_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t279_n242;
  t279_n241: return -0.09934351593255997f;
  t279_n242: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t279_n244;
  t279_n243: return -0.05482078343629837f;
  t279_n244: return 0.08968928456306458f;
  t279_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t279_n251;
  t279_n246: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t279_n248;
  t279_n247: return 0.011964847333729267f;
  t279_n248: if (E.getNumReferences() >= 3347859008 /*71820.5*/) goto t279_n250;
  t279_n249: return -0.04061729833483696f;
  t279_n250: return 0.04215262457728386f;
  t279_n251: return 0.057405222207307816f;
  t279_n252: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t279_n254;
  t279_n253: return 0.03715813159942627f;
  t279_n254: return 0.10901042819023132f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree280(const Example& E) {
  t280_n0: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t280_n48;
  t280_n1: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t280_n31;
  t280_n2: if (E.getNumReferences() >= 3272507392 /*142.5*/) goto t280_n30;
  t280_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n29;
  t280_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t280_n8;
  t280_n5: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t280_n7;
  t280_n6: return 0.029346052557229996f;
  t280_n7: return -0.009862770326435566f;
  t280_n8: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t280_n24;
  t280_n9: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t280_n11;
  t280_n10: return -0.002962651662528515f;
  t280_n11: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t280_n23;
  t280_n12: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t280_n20;
  t280_n13: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t280_n19;
  t280_n14: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t280_n16;
  t280_n15: return 0.007658597081899643f;
  t280_n16: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n18;
  t280_n17: return 0.028571274131536484f;
  t280_n18: return -0.008838364854454994f;
  t280_n19: return -0.011990376748144627f;
  t280_n20: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t280_n22;
  t280_n21: return -0.015267759561538696f;
  t280_n22: return 0.023311985656619072f;
  t280_n23: return 0.0007184527348726988f;
  t280_n24: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t280_n28;
  t280_n25: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n27;
  t280_n26: return -0.006547463126480579f;
  t280_n27: return 0.0011791492579504848f;
  t280_n28: return -0.0132070854306221f;
  t280_n29: return 0.001418603234924376f;
  t280_n30: return -0.08834530413150787f;
  t280_n31: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t280_n47;
  t280_n32: if (E.getNumReferences() >= 3272638464 /*144.5*/) goto t280_n46;
  t280_n33: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t280_n41;
  t280_n34: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t280_n36;
  t280_n35: return -0.02561151795089245f;
  t280_n36: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n40;
  t280_n37: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t280_n39;
  t280_n38: return 0.44094178080558777f;
  t280_n39: return 0.19512811303138733f;
  t280_n40: return 0.019455449655652046f;
  t280_n41: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t280_n43;
  t280_n42: return -0.018516600131988525f;
  t280_n43: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t280_n45;
  t280_n44: return 0.14440174400806427f;
  t280_n45: return -0.08429251611232758f;
  t280_n46: return 0.006182060576975346f;
  t280_n47: return -0.1566733866930008f;
  t280_n48: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t280_n50;
  t280_n49: return -0.00023631159274373204f;
  t280_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n212;
  t280_n51: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t280_n193;
  t280_n52: if (E.getFileProximityDistanceCost() >= 3255435264 /*34.5*/) goto t280_n186;
  t280_n53: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t280_n167;
  t280_n54: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t280_n74;
  t280_n55: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t280_n67;
  t280_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t280_n58;
  t280_n57: return -0.022549008950591087f;
  t280_n58: if (E.getFractionNameInContext() >= 3177875683 /*0.057259451597929*/) goto t280_n60;
  t280_n59: return -0.0027320277877151966f;
  t280_n60: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t280_n62;
  t280_n61: return -0.03983861953020096f;
  t280_n62: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t280_n64;
  t280_n63: return 0.035507019609212875f;
  t280_n64: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t280_n66;
  t280_n65: return 0.03671436756849289f;
  t280_n66: return 0.10763368010520935f;
  t280_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n71;
  t280_n68: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n70;
  t280_n69: return -0.0656089037656784f;
  t280_n70: return -0.055183835327625275f;
  t280_n71: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t280_n73;
  t280_n72: return -0.02477451227605343f;
  t280_n73: return -0.13653792440891266f;
  t280_n74: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t280_n100;
  t280_n75: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n77;
  t280_n76: return -0.015635721385478973f;
  t280_n77: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n99;
  t280_n78: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n98;
  t280_n79: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t280_n97;
  t280_n80: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t280_n84;
  t280_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n83;
  t280_n82: return -0.06994612514972687f;
  t280_n83: return 0.02844424545764923f;
  t280_n84: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t280_n96;
  t280_n85: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t280_n89;
  t280_n86: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t280_n88;
  t280_n87: return 0.025899328291416168f;
  t280_n88: return 0.04334588721394539f;
  t280_n89: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t280_n95;
  t280_n90: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t280_n92;
  t280_n91: return 0.05844544246792793f;
  t280_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t280_n94;
  t280_n93: return 0.06412531435489655f;
  t280_n94: return 0.12012193351984024f;
  t280_n95: return 0.02921346202492714f;
  t280_n96: return -0.13253523409366608f;
  t280_n97: return -0.10235980153083801f;
  t280_n98: return -0.032992664724588394f;
  t280_n99: return -0.02340225875377655f;
  t280_n100: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t280_n118;
  t280_n101: if (E.getSymbolScopeDistanceCost() >= 3251109888 /*25.0*/) goto t280_n107;
  t280_n102: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t280_n104;
  t280_n103: return -0.01564004458487034f;
  t280_n104: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n106;
  t280_n105: return -0.11096404492855072f;
  t280_n106: return 0.18709826469421387f;
  t280_n107: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n113;
  t280_n108: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n112;
  t280_n109: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t280_n111;
  t280_n110: return -0.07142744958400726f;
  t280_n111: return -0.13423661887645721f;
  t280_n112: return -0.06270391494035721f;
  t280_n113: if (E.getFractionNameInContext() >= 3198425666 /*0.3205128312110901*/) goto t280_n115;
  t280_n114: return -0.04416912421584129f;
  t280_n115: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t280_n117;
  t280_n116: return 0.1818467229604721f;
  t280_n117: return -0.3509986400604248f;
  t280_n118: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t280_n152;
  t280_n119: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t280_n143;
  t280_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n122;
  t280_n121: return -0.0657559186220169f;
  t280_n122: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t280_n142;
  t280_n123: if (E.getNumReferences() >= 3339744256 /*36980.0*/) goto t280_n139;
  t280_n124: if (E.getFractionNameInContext() >= 3179562052 /*0.06458333134651184*/) goto t280_n132;
  t280_n125: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n129;
  t280_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t280_n128;
  t280_n127: return -0.04172399267554283f;
  t280_n128: return 0.17061744630336761f;
  t280_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n131;
  t280_n130: return -0.12174183875322342f;
  t280_n131: return 0.12354326248168945f;
  t280_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t280_n138;
  t280_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n135;
  t280_n134: return -0.008161724545061588f;
  t280_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n137;
  t280_n136: return -0.10274595022201538f;
  t280_n137: return 0.12948040664196014f;
  t280_n138: return 0.14131379127502441f;
  t280_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t280_n141;
  t280_n140: return 0.09543780237436295f;
  t280_n141: return 0.19274787604808807f;
  t280_n142: return -0.046602584421634674f;
  t280_n143: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n151;
  t280_n144: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n146;
  t280_n145: return -0.0189192034304142f;
  t280_n146: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n150;
  t280_n147: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t280_n149;
  t280_n148: return 0.05311477929353714f;
  t280_n149: return -0.10238252580165863f;
  t280_n150: return -0.09612960368394852f;
  t280_n151: return -0.03918461874127388f;
  t280_n152: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t280_n166;
  t280_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t280_n155;
  t280_n154: return 0.02418966218829155f;
  t280_n155: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t280_n157;
  t280_n156: return -0.0195540189743042f;
  t280_n157: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t280_n165;
  t280_n158: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t280_n164;
  t280_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t280_n161;
  t280_n160: return 0.14088718593120575f;
  t280_n161: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n163;
  t280_n162: return -0.09333339333534241f;
  t280_n163: return 0.4325912892818451f;
  t280_n164: return 0.15147313475608826f;
  t280_n165: return -0.10914188623428345f;
  t280_n166: return -0.03028619848191738f;
  t280_n167: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro))) goto t280_n177;
  t280_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t280_n170;
  t280_n169: return 0.020078426226973534f;
  t280_n170: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t280_n172;
  t280_n171: return -0.030701443552970886f;
  t280_n172: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t280_n176;
  t280_n173: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t280_n175;
  t280_n174: return 0.15316647291183472f;
  t280_n175: return 0.10821136832237244f;
  t280_n176: return 0.00887535884976387f;
  t280_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t280_n183;
  t280_n178: if (E.getFractionNameInContext() >= 3210440119 /*0.8571428656578064*/) goto t280_n182;
  t280_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t280_n181;
  t280_n180: return -0.2171333283185959f;
  t280_n181: return 0.09813294559717178f;
  t280_n182: return -0.2670295536518097f;
  t280_n183: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t280_n185;
  t280_n184: return 0.15770238637924194f;
  t280_n185: return 0.046561527997255325f;
  t280_n186: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t280_n188;
  t280_n187: return -0.2814861536026001f;
  t280_n188: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t280_n190;
  t280_n189: return -0.12588457763195038f;
  t280_n190: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t280_n192;
  t280_n191: return -0.1602732241153717f;
  t280_n192: return -0.130020871758461f;
  t280_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t280_n209;
  t280_n194: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t280_n208;
  t280_n195: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t280_n205;
  t280_n196: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t280_n198;
  t280_n197: return -0.0015443336451426148f;
  t280_n198: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t280_n200;
  t280_n199: return 0.04284725710749626f;
  t280_n200: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t280_n204;
  t280_n201: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t280_n203;
  t280_n202: return 0.02029474452137947f;
  t280_n203: return -0.10138452798128128f;
  t280_n204: return 0.016914257779717445f;
  t280_n205: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t280_n207;
  t280_n206: return 0.08473264425992966f;
  t280_n207: return -0.0797424167394638f;
  t280_n208: return 0.051839519292116165f;
  t280_n209: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t280_n211;
  t280_n210: return 0.015594263561069965f;
  t280_n211: return -0.037954047322273254f;
  t280_n212: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t280_n252;
  t280_n213: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t280_n219;
  t280_n214: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t280_n216;
  t280_n215: return 0.003786474699154496f;
  t280_n216: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t280_n218;
  t280_n217: return 0.07669413089752197f;
  t280_n218: return -0.004955691285431385f;
  t280_n219: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t280_n251;
  t280_n220: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope))) goto t280_n226;
  t280_n221: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t280_n223;
  t280_n222: return -0.01723232865333557f;
  t280_n223: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t280_n225;
  t280_n224: return 0.09023912250995636f;
  t280_n225: return 0.01625225506722927f;
  t280_n226: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t280_n250;
  t280_n227: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t280_n249;
  t280_n228: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t280_n242;
  t280_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t280_n235;
  t280_n230: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t280_n232;
  t280_n231: return -0.010614694096148014f;
  t280_n232: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t280_n234;
  t280_n233: return 0.0064408485777676105f;
  t280_n234: return -0.10707096010446548f;
  t280_n235: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t280_n237;
  t280_n236: return 0.021566130220890045f;
  t280_n237: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t280_n239;
  t280_n238: return 0.0037024852354079485f;
  t280_n239: if (E.getFractionNameInContext() >= 3185414456 /*0.10818713903427124*/) goto t280_n241;
  t280_n240: return -0.06212626397609711f;
  t280_n241: return -0.0007960319635458291f;
  t280_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t280_n244;
  t280_n243: return -0.001635247841477394f;
  t280_n244: if (E.getFractionNameInContext() >= 3193183554 /*0.20714285969734192*/) goto t280_n248;
  t280_n245: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t280_n247;
  t280_n246: return 0.06451801210641861f;
  t280_n247: return 0.02301115356385708f;
  t280_n248: return 0.1343599557876587f;
  t280_n249: return -0.028112415224313736f;
  t280_n250: return 0.01580861024558544f;
  t280_n251: return -0.07908811420202255f;
  t280_n252: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t280_n254;
  t280_n253: return -0.07405544072389603f;
  t280_n254: return -0.07153547555208206f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree281(const Example& E) {
  t281_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t281_n32;
  t281_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t281_n3;
  t281_n2: return -0.0004520867660176009f;
  t281_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t281_n7;
  t281_n4: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t281_n6;
  t281_n5: return -0.006326609291136265f;
  t281_n6: return 0.0006786452140659094f;
  t281_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t281_n9;
  t281_n8: return 0.0002670983667485416f;
  t281_n9: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t281_n11;
  t281_n10: return 0.005481510888785124f;
  t281_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t281_n23;
  t281_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t281_n16;
  t281_n13: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t281_n15;
  t281_n14: return -0.037098463624715805f;
  t281_n15: return 0.09226275980472565f;
  t281_n16: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t281_n18;
  t281_n17: return -0.033590514212846756f;
  t281_n18: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t281_n20;
  t281_n19: return 0.01838085800409317f;
  t281_n20: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t281_n22;
  t281_n21: return -0.058129504323005676f;
  t281_n22: return 0.04684371128678322f;
  t281_n23: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t281_n31;
  t281_n24: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t281_n28;
  t281_n25: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t281_n27;
  t281_n26: return 0.009964835830032825f;
  t281_n27: return 0.11538996547460556f;
  t281_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t281_n30;
  t281_n29: return 0.015876689925789833f;
  t281_n30: return 0.1275458037853241f;
  t281_n31: return -0.14306698739528656f;
  t281_n32: if (E.getNumReferences() >= 3334924800 /*25461.0*/) goto t281_n246;
  t281_n33: if (E.getNumReferences() >= 3318231040 /*6406.0*/) goto t281_n229;
  t281_n34: if (E.getNumReferences() >= 3318137856 /*6360.5*/) goto t281_n220;
  t281_n35: if (E.getNumReferences() >= 3317133312 /*5870.0*/) goto t281_n219;
  t281_n36: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t281_n40;
  t281_n37: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t281_n39;
  t281_n38: return 0.0035141841508448124f;
  t281_n39: return 0.013199634850025177f;
  t281_n40: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t281_n62;
  t281_n41: if (E.getNumReferences() >= 3284090880 /*382.5*/) goto t281_n47;
  t281_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t281_n44;
  t281_n43: return -0.019819138571619987f;
  t281_n44: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t281_n46;
  t281_n45: return 0.05902260169386864f;
  t281_n46: return -0.1377071738243103f;
  t281_n47: if (E.getNumReferences() >= 3284680704 /*400.5*/) goto t281_n49;
  t281_n48: return 0.11481063812971115f;
  t281_n49: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t281_n61;
  t281_n50: if (E.getNumReferences() >= 3289268224 /*569.0*/) goto t281_n52;
  t281_n51: return -0.14639000594615936f;
  t281_n52: if (E.getNumReferences() >= 3293478912 /*826.0*/) goto t281_n56;
  t281_n53: if (E.getNumReferences() >= 3292160000 /*745.5*/) goto t281_n55;
  t281_n54: return 0.06439972668886185f;
  t281_n55: return 0.12015770375728607f;
  t281_n56: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t281_n58;
  t281_n57: return 0.03638811036944389f;
  t281_n58: if (E.getNumReferences() >= 3306743808 /*2446.5*/) goto t281_n60;
  t281_n59: return 0.11047214269638062f;
  t281_n60: return 0.04364597052335739f;
  t281_n61: return -0.022636961191892624f;
  t281_n62: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t281_n64;
  t281_n63: return 0.007135351188480854f;
  t281_n64: if (E.getNumReferences() >= 3316423680 /*5523.5*/) goto t281_n216;
  t281_n65: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t281_n163;
  t281_n66: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t281_n142;
  t281_n67: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t281_n121;
  t281_n68: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t281_n70;
  t281_n69: return -0.012277116999030113f;
  t281_n70: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t281_n98;
  t281_n71: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t281_n93;
  t281_n72: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t281_n78;
  t281_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t281_n77;
  t281_n74: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t281_n76;
  t281_n75: return 0.07754968851804733f;
  t281_n76: return -0.11889271438121796f;
  t281_n77: return 0.35455477237701416f;
  t281_n78: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t281_n80;
  t281_n79: return -0.010252668522298336f;
  t281_n80: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t281_n92;
  t281_n81: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t281_n87;
  t281_n82: if (E.getNumReferences() >= 3254517760 /*31.5*/) goto t281_n84;
  t281_n83: return -0.029427440837025642f;
  t281_n84: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t281_n86;
  t281_n85: return 0.16411222517490387f;
  t281_n86: return -0.11665932834148407f;
  t281_n87: if (E.getNumReferences() >= 3262382080 /*61.0*/) goto t281_n91;
  t281_n88: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t281_n90;
  t281_n89: return 0.080037422478199f;
  t281_n90: return -0.1390971690416336f;
  t281_n91: return 0.10573818534612656f;
  t281_n92: return -0.026854364201426506f;
  t281_n93: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t281_n95;
  t281_n94: return -0.0030586854554712772f;
  t281_n95: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t281_n97;
  t281_n96: return 0.13804353773593903f;
  t281_n97: return 0.16317041218280792f;
  t281_n98: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t281_n120;
  t281_n99: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t281_n113;
  t281_n100: if (E.getNumReferences() >= 3274670080 /*175.5*/) goto t281_n106;
  t281_n101: if (E.getNumReferences() >= 3274178560 /*168.0*/) goto t281_n103;
  t281_n102: return -0.06756269186735153f;
  t281_n103: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t281_n105;
  t281_n104: return -0.12161683291196823f;
  t281_n105: return -1.6422371864318848f;
  t281_n106: if (E.getNumReferences() >= 3274735616 /*176.5*/) goto t281_n108;
  t281_n107: return 0.13289998471736908f;
  t281_n108: if (E.getNumReferences() >= 3278536704 /*234.5*/) goto t281_n112;
  t281_n109: if (E.getNumReferences() >= 3277586432 /*220.0*/) goto t281_n111;
  t281_n110: return 0.05236091837286949f;
  t281_n111: return 0.12298669666051865f;
  t281_n112: return -0.2731361389160156f;
  t281_n113: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t281_n117;
  t281_n114: if (E.getNumReferences() >= 3274080256 /*166.5*/) goto t281_n116;
  t281_n115: return 0.12761476635932922f;
  t281_n116: return 0.0465000681579113f;
  t281_n117: if (E.getNumReferences() >= 3269066752 /*109.0*/) goto t281_n119;
  t281_n118: return -0.3967536389827728f;
  t281_n119: return 0.07615137845277786f;
  t281_n120: return -0.11214601248502731f;
  t281_n121: if (E.getNumReferences() >= 3284860928 /*406.0*/) goto t281_n129;
  t281_n122: if (E.getNumReferences() >= 3284484096 /*394.5*/) goto t281_n128;
  t281_n123: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t281_n127;
  t281_n124: if (E.getNumReferences() >= 3280928768 /*286.0*/) goto t281_n126;
  t281_n125: return 0.0986485481262207f;
  t281_n126: return 0.029629185795783997f;
  t281_n127: return 0.10493308305740356f;
  t281_n128: return 0.1280568391084671f;
  t281_n129: if (E.getNumReferences() >= 3287203840 /*477.5*/) goto t281_n131;
  t281_n130: return -0.1438577026128769f;
  t281_n131: if (E.getNumReferences() >= 3287891968 /*498.5*/) goto t281_n133;
  t281_n132: return 0.13374511897563934f;
  t281_n133: if (E.getNumReferences() >= 3289890816 /*607.0*/) goto t281_n139;
  t281_n134: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t281_n138;
  t281_n135: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t281_n137;
  t281_n136: return 0.09906323254108429f;
  t281_n137: return 0.08571306616067886f;
  t281_n138: return -0.27175623178482056f;
  t281_n139: if (E.getFractionNameInContext() >= 3181279720 /*0.0773809552192688*/) goto t281_n141;
  t281_n140: return 0.0011247785296291113f;
  t281_n141: return 0.08511801809072495f;
  t281_n142: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t281_n144;
  t281_n143: return 0.000106193678220734f;
  t281_n144: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t281_n154;
  t281_n145: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t281_n153;
  t281_n146: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t281_n148;
  t281_n147: return 0.020311495289206505f;
  t281_n148: if (E.getFractionNameInContext() >= 3201838467 /*0.42222222685813904*/) goto t281_n152;
  t281_n149: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t281_n151;
  t281_n150: return 0.04505274072289467f;
  t281_n151: return 0.15581092238426208f;
  t281_n152: return 0.032816868275403976f;
  t281_n153: return -0.2548324167728424f;
  t281_n154: if (E.getNumReferences() >= 3272867840 /*148.0*/) goto t281_n160;
  t281_n155: if (E.getNumReferences() >= 3270377472 /*119.0*/) goto t281_n159;
  t281_n156: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t281_n158;
  t281_n157: return -0.022590450942516327f;
  t281_n158: return 0.01548160333186388f;
  t281_n159: return -0.12603554129600525f;
  t281_n160: if (E.getNumReferences() >= 3275849728 /*193.5*/) goto t281_n162;
  t281_n161: return 0.12101747840642929f;
  t281_n162: return 0.02366901934146881f;
  t281_n163: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t281_n203;
  t281_n164: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t281_n166;
  t281_n165: return -0.0014247456565499306f;
  t281_n166: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t281_n180;
  t281_n167: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t281_n179;
  t281_n168: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t281_n176;
  t281_n169: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t281_n173;
  t281_n170: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t281_n172;
  t281_n171: return 0.032049939036369324f;
  t281_n172: return 0.08295999467372894f;
  t281_n173: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t281_n175;
  t281_n174: return -0.009732398204505444f;
  t281_n175: return 0.10059922933578491f;
  t281_n176: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t281_n178;
  t281_n177: return 0.022915182635188103f;
  t281_n178: return -0.02105698175728321f;
  t281_n179: return -0.1433647722005844f;
  t281_n180: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t281_n202;
  t281_n181: if (E.getFractionNameInContext() >= 3196720569 /*0.269696980714798*/) goto t281_n185;
  t281_n182: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t281_n184;
  t281_n183: return -0.01184885110706091f;
  t281_n184: return 0.009499596431851387f;
  t281_n185: if (E.getFractionNameInContext() >= 3209295008 /*0.7888889312744141*/) goto t281_n197;
  t281_n186: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t281_n192;
  t281_n187: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t281_n189;
  t281_n188: return -0.3235836327075958f;
  t281_n189: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t281_n191;
  t281_n190: return 0.039094477891922f;
  t281_n191: return -0.06351396441459656f;
  t281_n192: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t281_n196;
  t281_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t281_n195;
  t281_n194: return -0.08598622679710388f;
  t281_n195: return 0.0923716202378273f;
  t281_n196: return 0.032270647585392f;
  t281_n197: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t281_n201;
  t281_n198: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t281_n200;
  t281_n199: return -0.7167338132858276f;
  t281_n200: return 0.0010298644192516804f;
  t281_n201: return 0.1893010288476944f;
  t281_n202: return 0.04698091745376587f;
  t281_n203: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol))) goto t281_n205;
  t281_n204: return 0.005772641859948635f;
  t281_n205: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t281_n209;
  t281_n206: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t281_n208;
  t281_n207: return 0.007388540543615818f;
  t281_n208: return -0.06908457726240158f;
  t281_n209: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t281_n215;
  t281_n210: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t281_n212;
  t281_n211: return 0.015665587037801743f;
  t281_n212: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t281_n214;
  t281_n213: return 0.03736511617898941f;
  t281_n214: return 0.08672335743904114f;
  t281_n215: return 0.020966146141290665f;
  t281_n216: if (E.getNumReferences() >= 3316538368 /*5579.5*/) goto t281_n218;
  t281_n217: return 0.14461179077625275f;
  t281_n218: return 0.05020357668399811f;
  t281_n219: return -0.11787685751914978f;
  t281_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t281_n222;
  t281_n221: return 0.10507980734109879f;
  t281_n222: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t281_n228;
  t281_n223: if (E.getNumReferences() >= 3318195200 /*6388.5*/) goto t281_n227;
  t281_n224: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t281_n226;
  t281_n225: return 0.08455106616020203f;
  t281_n226: return 0.12542498111724854f;
  t281_n227: return 0.09617490321397781f;
  t281_n228: return 0.10702089220285416f;
  t281_n229: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t281_n239;
  t281_n230: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t281_n238;
  t281_n231: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t281_n237;
  t281_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t281_n236;
  t281_n233: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t281_n235;
  t281_n234: return -0.30878934264183044f;
  t281_n235: return -0.024050235748291016f;
  t281_n236: return 0.09436444193124771f;
  t281_n237: return -0.10050919651985168f;
  t281_n238: return 0.0814533531665802f;
  t281_n239: if (E.getNumReferences() >= 3331945984 /*19643.0*/) goto t281_n245;
  t281_n240: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t281_n242;
  t281_n241: return -0.003689663950353861f;
  t281_n242: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t281_n244;
  t281_n243: return 0.11939475685358047f;
  t281_n244: return 0.12895700335502625f;
  t281_n245: return -0.0685490071773529f;
  t281_n246: if (E.getNumReferences() >= 3338711808 /*32947.0*/) goto t281_n248;
  t281_n247: return 0.05105448141694069f;
  t281_n248: if (E.getNumReferences() >= 3341699072 /*44616.0*/) goto t281_n252;
  t281_n249: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t281_n251;
  t281_n250: return -0.04567360505461693f;
  t281_n251: return -0.2571876347064972f;
  t281_n252: if (E.getNumReferences() >= 3342374144 /*47253.0*/) goto t281_n254;
  t281_n253: return 0.06230619177222252f;
  t281_n254: return 0.02320082113146782f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree282(const Example& E) {
  t282_n0: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t282_n114;
  t282_n1: if (E.getIsReservedName() >= 3204448256 /*0.5*/) goto t282_n111;
  t282_n2: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t282_n108;
  t282_n3: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t282_n101;
  t282_n4: if (E.getNumReferences() >= 3359314816 /*191566.0*/) goto t282_n98;
  t282_n5: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t282_n97;
  t282_n6: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t282_n84;
  t282_n7: if (E.getNumReferences() >= 3350448512 /*92051.0*/) goto t282_n83;
  t282_n8: if (E.getNumReferences() >= 3350016768 /*88678.0*/) goto t282_n82;
  t282_n9: if (E.getNumReferences() >= 3344648832 /*56138.5*/) goto t282_n81;
  t282_n10: if (E.getNumReferences() >= 3344355328 /*54992.0*/) goto t282_n64;
  t282_n11: if (E.getNumReferences() >= 3342384640 /*47294.0*/) goto t282_n57;
  t282_n12: if (E.getNumReferences() >= 3338822784 /*33380.5*/) goto t282_n36;
  t282_n13: if (E.getNumReferences() >= 3338727808 /*33009.5*/) goto t282_n29;
  t282_n14: if (E.getNumReferences() >= 3338478336 /*32401.5*/) goto t282_n26;
  t282_n15: if (E.getNumReferences() >= 3338286080 /*32026.0*/) goto t282_n25;
  t282_n16: if (E.getNumReferences() >= 3337910528 /*31292.5*/) goto t282_n18;
  t282_n17: return -0.00027247637626715004f;
  t282_n18: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t282_n22;
  t282_n19: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n21;
  t282_n20: return -0.04391090199351311f;
  t282_n21: return 0.06958585977554321f;
  t282_n22: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n24;
  t282_n23: return 0.10575936734676361f;
  t282_n24: return 0.11682300269603729f;
  t282_n25: return -0.11526291817426682f;
  t282_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t282_n28;
  t282_n27: return -0.03942059725522995f;
  t282_n28: return 0.05479408800601959f;
  t282_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t282_n31;
  t282_n30: return -0.10652130842208862f;
  t282_n31: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n35;
  t282_n32: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t282_n34;
  t282_n33: return 0.19225098192691803f;
  t282_n34: return 0.1085558533668518f;
  t282_n35: return -0.10021448135375977f;
  t282_n36: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t282_n50;
  t282_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t282_n45;
  t282_n38: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t282_n40;
  t282_n39: return 0.04782579839229584f;
  t282_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n42;
  t282_n41: return -0.33143237233161926f;
  t282_n42: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t282_n44;
  t282_n43: return 0.11769956350326538f;
  t282_n44: return -0.19586089253425598f;
  t282_n45: if (E.getNumReferences() >= 3338855552 /*33508.5*/) goto t282_n47;
  t282_n46: return -0.1055046021938324f;
  t282_n47: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t282_n49;
  t282_n48: return 0.14583688974380493f;
  t282_n49: return 0.09106404334306717f;
  t282_n50: if (E.getNumReferences() >= 3338991488 /*34039.5*/) goto t282_n52;
  t282_n51: return -0.18268151581287384f;
  t282_n52: if (E.getNumReferences() >= 3339089536 /*34422.5*/) goto t282_n56;
  t282_n53: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t282_n55;
  t282_n54: return 0.11740434914827347f;
  t282_n55: return 0.04460281878709793f;
  t282_n56: return 0.01123026479035616f;
  t282_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n59;
  t282_n58: return -0.03479938581585884f;
  t282_n59: if (E.getNumReferences() >= 3343221504 /*50563.0*/) goto t282_n61;
  t282_n60: return -0.11604338139295578f;
  t282_n61: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t282_n63;
  t282_n62: return -0.10896789282560349f;
  t282_n63: return 0.13454294204711914f;
  t282_n64: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t282_n68;
  t282_n65: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t282_n67;
  t282_n66: return 0.03100857511162758f;
  t282_n67: return -0.16587990522384644f;
  t282_n68: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t282_n70;
  t282_n69: return -0.14650589227676392f;
  t282_n70: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n74;
  t282_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t282_n73;
  t282_n72: return -0.11918803304433823f;
  t282_n73: return 0.08964300900697708f;
  t282_n74: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t282_n78;
  t282_n75: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t282_n77;
  t282_n76: return 0.0583285428583622f;
  t282_n77: return 0.12697860598564148f;
  t282_n78: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t282_n80;
  t282_n79: return 0.13626278936862946f;
  t282_n80: return 0.1978272795677185f;
  t282_n81: return -0.01595141366124153f;
  t282_n82: return 0.024521280080080032f;
  t282_n83: return -0.03545009717345238f;
  t282_n84: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t282_n86;
  t282_n85: return 0.06953408569097519f;
  t282_n86: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t282_n96;
  t282_n87: if (E.getNumReferences() >= 3353301504 /*114340.0*/) goto t282_n91;
  t282_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t282_n90;
  t282_n89: return 0.012174658477306366f;
  t282_n90: return 0.048700615763664246f;
  t282_n91: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t282_n95;
  t282_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t282_n94;
  t282_n93: return 0.0023799347691237926f;
  t282_n94: return 0.061261385679244995f;
  t282_n95: return -0.03445158526301384f;
  t282_n96: return -0.03161797672510147f;
  t282_n97: return -0.1593134105205536f;
  t282_n98: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t282_n100;
  t282_n99: return -0.10656921565532684f;
  t282_n100: return 0.01226656325161457f;
  t282_n101: if (E.getNumReferences() >= 3363660160 /*259462.0*/) goto t282_n107;
  t282_n102: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t282_n106;
  t282_n103: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t282_n105;
  t282_n104: return 0.2116517275571823f;
  t282_n105: return -0.24294430017471313f;
  t282_n106: return -0.13959260284900665f;
  t282_n107: return -0.0008556596585549414f;
  t282_n108: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t282_n110;
  t282_n109: return -0.03325086832046509f;
  t282_n110: return 0.02845797874033451f;
  t282_n111: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t282_n113;
  t282_n112: return -0.06517060846090317f;
  t282_n113: return -0.2989616096019745f;
  t282_n114: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t282_n200;
  t282_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t282_n199;
  t282_n116: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t282_n120;
  t282_n117: if (E.getNumReferences() >= 3347831872 /*71608.5*/) goto t282_n119;
  t282_n118: return -0.03980913385748863f;
  t282_n119: return 0.16879531741142273f;
  t282_n120: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t282_n142;
  t282_n121: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t282_n123;
  t282_n122: return 0.003568721003830433f;
  t282_n123: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t282_n141;
  t282_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t282_n126;
  t282_n125: return 0.01374523714184761f;
  t282_n126: if (E.getNumReferences() >= 3303383040 /*1837.0*/) goto t282_n138;
  t282_n127: if (E.getNumReferences() >= 3291447296 /*702.0*/) goto t282_n129;
  t282_n128: return 0.02882395312190056f;
  t282_n129: if (E.getNumReferences() >= 3294527488 /*890.0*/) goto t282_n133;
  t282_n130: if (E.getNumReferences() >= 3294257152 /*873.5*/) goto t282_n132;
  t282_n131: return 0.10951138287782669f;
  t282_n132: return 0.12972481548786163f;
  t282_n133: if (E.getNumReferences() >= 3299950592 /*1418.0*/) goto t282_n135;
  t282_n134: return 0.019135430455207825f;
  t282_n135: if (E.getNumReferences() >= 3300896768 /*1533.5*/) goto t282_n137;
  t282_n136: return 0.1209132969379425f;
  t282_n137: return 0.08237457275390625f;
  t282_n138: if (E.getNumReferences() >= 3307515904 /*2635.0*/) goto t282_n140;
  t282_n139: return -0.7124608159065247f;
  t282_n140: return -0.019577674567699432f;
  t282_n141: return -0.0028219493106007576f;
  t282_n142: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t282_n196;
  t282_n143: if (E.getNumReferences() >= 3305873408 /*2234.0*/) goto t282_n183;
  t282_n144: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t282_n182;
  t282_n145: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t282_n161;
  t282_n146: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t282_n154;
  t282_n147: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t282_n151;
  t282_n148: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t282_n150;
  t282_n149: return -0.001897948794066906f;
  t282_n150: return -0.0518743172287941f;
  t282_n151: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t282_n153;
  t282_n152: return 0.004659051075577736f;
  t282_n153: return 0.07510744780302048f;
  t282_n154: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t282_n160;
  t282_n155: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t282_n157;
  t282_n156: return 0.061256133019924164f;
  t282_n157: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t282_n159;
  t282_n158: return 0.10785892605781555f;
  t282_n159: return 0.11897134780883789f;
  t282_n160: return -0.09165198355913162f;
  t282_n161: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t282_n175;
  t282_n162: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t282_n172;
  t282_n163: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t282_n165;
  t282_n164: return -0.013230015523731709f;
  t282_n165: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t282_n167;
  t282_n166: return 0.10057003051042557f;
  t282_n167: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t282_n169;
  t282_n168: return -0.24153678119182587f;
  t282_n169: if (E.getNumReferences() >= 3264413696 /*73.5*/) goto t282_n171;
  t282_n170: return 0.6536605954170227f;
  t282_n171: return 0.010364708490669727f;
  t282_n172: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t282_n174;
  t282_n173: return 0.0261981301009655f;
  t282_n174: return -0.06835383921861649f;
  t282_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n177;
  t282_n176: return -0.07948565483093262f;
  t282_n177: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t282_n181;
  t282_n178: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t282_n180;
  t282_n179: return 0.150568425655365f;
  t282_n180: return -0.14840492606163025f;
  t282_n181: return -0.12507034838199615f;
  t282_n182: return -0.06392182409763336f;
  t282_n183: if (E.getNumReferences() >= 3335566848 /*26715.0*/) goto t282_n185;
  t282_n184: return 0.043077122420072556f;
  t282_n185: if (E.getNumReferences() >= 3347359424 /*67917.5*/) goto t282_n195;
  t282_n186: if (E.getNumReferences() >= 3341518848 /*43912.0*/) goto t282_n192;
  t282_n187: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t282_n191;
  t282_n188: if (E.getNumReferences() >= 3340710656 /*40755.0*/) goto t282_n190;
  t282_n189: return -0.025200851261615753f;
  t282_n190: return -0.569713294506073f;
  t282_n191: return 0.07665212452411652f;
  t282_n192: if (E.getNumReferences() >= 3344570752 /*55833.5*/) goto t282_n194;
  t282_n193: return -0.5010406374931335f;
  t282_n194: return -0.14183305203914642f;
  t282_n195: return 0.030278395861387253f;
  t282_n196: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t282_n198;
  t282_n197: return 0.003168809460476041f;
  t282_n198: return -0.5856161117553711f;
  t282_n199: return 0.0006895459955558181f;
  t282_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t282_n202;
  t282_n201: return 0.0004050561983603984f;
  t282_n202: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t282_n228;
  t282_n203: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t282_n225;
  t282_n204: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t282_n206;
  t282_n205: return 0.005470613949000835f;
  t282_n206: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t282_n208;
  t282_n207: return 0.019407914951443672f;
  t282_n208: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n210;
  t282_n209: return -0.004218052141368389f;
  t282_n210: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t282_n212;
  t282_n211: return 0.006575756706297398f;
  t282_n212: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t282_n222;
  t282_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess))) goto t282_n215;
  t282_n214: return 0.027074508368968964f;
  t282_n215: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t282_n219;
  t282_n216: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t282_n218;
  t282_n217: return 0.07974226027727127f;
  t282_n218: return 0.034926023334264755f;
  t282_n219: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t282_n221;
  t282_n220: return -0.005261126905679703f;
  t282_n221: return 0.07218845933675766f;
  t282_n222: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n224;
  t282_n223: return -0.18249493837356567f;
  t282_n224: return 0.07564637064933777f;
  t282_n225: if (E.getFractionNameInContext() >= 3199974332 /*0.36666667461395264*/) goto t282_n227;
  t282_n226: return -0.07347175478935242f;
  t282_n227: return 0.15224669873714447f;
  t282_n228: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t282_n244;
  t282_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n243;
  t282_n230: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t282_n234;
  t282_n231: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n233;
  t282_n232: return -0.07218147814273834f;
  t282_n233: return -0.008433796465396881f;
  t282_n234: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t282_n238;
  t282_n235: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t282_n237;
  t282_n236: return -0.16592609882354736f;
  t282_n237: return 0.019231226295232773f;
  t282_n238: if (E.getFractionNameInContext() >= 3210739712 /*0.875*/) goto t282_n240;
  t282_n239: return -0.1751461625099182f;
  t282_n240: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t282_n242;
  t282_n241: return -0.060932859778404236f;
  t282_n242: return 0.10464941710233688f;
  t282_n243: return -0.0013598876539617777f;
  t282_n244: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t282_n254;
  t282_n245: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t282_n247;
  t282_n246: return 0.022740831598639488f;
  t282_n247: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t282_n253;
  t282_n248: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t282_n252;
  t282_n249: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t282_n251;
  t282_n250: return 0.06032702699303627f;
  t282_n251: return -0.2376089245080948f;
  t282_n252: return 0.1217019185423851f;
  t282_n253: return 0.04111652448773384f;
  t282_n254: return -0.011580000631511211f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree283(const Example& E) {
  t283_n0: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t283_n160;
  t283_n1: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t283_n101;
  t283_n2: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t283_n100;
  t283_n3: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t283_n99;
  t283_n4: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t283_n6;
  t283_n5: return -0.002493911422789097f;
  t283_n6: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Namespace))) goto t283_n68;
  t283_n7: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t283_n49;
  t283_n8: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t283_n48;
  t283_n9: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t283_n43;
  t283_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t283_n30;
  t283_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n19;
  t283_n12: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t283_n14;
  t283_n13: return 0.010210796259343624f;
  t283_n14: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t283_n16;
  t283_n15: return 0.0035899742506444454f;
  t283_n16: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t283_n18;
  t283_n17: return 0.06105421856045723f;
  t283_n18: return 0.01529089454561472f;
  t283_n19: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t283_n29;
  t283_n20: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t283_n24;
  t283_n21: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t283_n23;
  t283_n22: return -0.006354218348860741f;
  t283_n23: return -0.10097910463809967f;
  t283_n24: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t283_n28;
  t283_n25: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t283_n27;
  t283_n26: return -0.13362643122673035f;
  t283_n27: return 0.17139047384262085f;
  t283_n28: return 0.017085082828998566f;
  t283_n29: return 0.05293039232492447f;
  t283_n30: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t283_n32;
  t283_n31: return -0.1410866677761078f;
  t283_n32: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t283_n36;
  t283_n33: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t283_n35;
  t283_n34: return 0.12766902148723602f;
  t283_n35: return -0.3628818690776825f;
  t283_n36: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t283_n40;
  t283_n37: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n39;
  t283_n38: return -0.16412754356861115f;
  t283_n39: return 0.09578663855791092f;
  t283_n40: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t283_n42;
  t283_n41: return 0.10200157016515732f;
  t283_n42: return 0.07015518844127655f;
  t283_n43: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t283_n45;
  t283_n44: return -0.0022644877899438143f;
  t283_n45: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t283_n47;
  t283_n46: return 0.004195658955723047f;
  t283_n47: return 0.024965297430753708f;
  t283_n48: return -0.047654714435338974f;
  t283_n49: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t283_n51;
  t283_n50: return -0.019111253321170807f;
  t283_n51: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t283_n53;
  t283_n52: return -0.02033284865319729f;
  t283_n53: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n65;
  t283_n54: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t283_n58;
  t283_n55: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t283_n57;
  t283_n56: return 0.022270945832133293f;
  t283_n57: return 0.09490705281496048f;
  t283_n58: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t283_n64;
  t283_n59: if (E.getFractionNameInContext() >= 3199364252 /*0.3484848737716675*/) goto t283_n61;
  t283_n60: return -0.02160709910094738f;
  t283_n61: if (E.getFractionNameInContext() >= 3201372433 /*0.40833333134651184*/) goto t283_n63;
  t283_n62: return 0.043067481368780136f;
  t283_n63: return -0.024815520271658897f;
  t283_n64: return 0.030941996723413467f;
  t283_n65: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t283_n67;
  t283_n66: return -0.0671004131436348f;
  t283_n67: return 0.002276700222864747f;
  t283_n68: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t283_n80;
  t283_n69: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n71;
  t283_n70: return 0.01684088259935379f;
  t283_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t283_n73;
  t283_n72: return 0.013270038180053234f;
  t283_n73: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t283_n77;
  t283_n74: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t283_n76;
  t283_n75: return 0.03923942521214485f;
  t283_n76: return 0.06102827563881874f;
  t283_n77: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t283_n79;
  t283_n78: return 0.045240480452775955f;
  t283_n79: return -0.037613168358802795f;
  t283_n80: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t283_n94;
  t283_n81: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t283_n93;
  t283_n82: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t283_n92;
  t283_n83: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n87;
  t283_n84: if (E.getFractionNameInContext() >= 3194482303 /*0.22649572789669037*/) goto t283_n86;
  t283_n85: return -0.00019457688904367387f;
  t283_n86: return -0.2794130742549896f;
  t283_n87: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t283_n91;
  t283_n88: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t283_n90;
  t283_n89: return 0.02490333281457424f;
  t283_n90: return -0.006599785294383764f;
  t283_n91: return -0.0006534492131322622f;
  t283_n92: return 0.01624303311109543f;
  t283_n93: return -0.05458391085267067f;
  t283_n94: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t283_n96;
  t283_n95: return 0.04537765681743622f;
  t283_n96: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t283_n98;
  t283_n97: return -0.06471828371286392f;
  t283_n98: return 0.1021743044257164f;
  t283_n99: return -0.02865593135356903f;
  t283_n100: return 0.0003283754922449589f;
  t283_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Symbol))) goto t283_n141;
  t283_n102: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t283_n128;
  t283_n103: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t283_n127;
  t283_n104: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t283_n106;
  t283_n105: return -0.002508797449991107f;
  t283_n106: if (E.getNumReferences() >= 3248488448 /*20.0*/) goto t283_n126;
  t283_n107: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t283_n109;
  t283_n108: return 0.006147294770926237f;
  t283_n109: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t283_n111;
  t283_n110: return -0.11062265187501907f;
  t283_n111: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t283_n113;
  t283_n112: return -0.10232126712799072f;
  t283_n113: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t283_n117;
  t283_n114: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t283_n116;
  t283_n115: return -0.11535189300775528f;
  t283_n116: return 0.19419197738170624f;
  t283_n117: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t283_n119;
  t283_n118: return -0.04335863143205643f;
  t283_n119: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t283_n125;
  t283_n120: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t283_n122;
  t283_n121: return 0.2768239378929138f;
  t283_n122: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n124;
  t283_n123: return 0.09411739557981491f;
  t283_n124: return 0.13118146359920502f;
  t283_n125: return 0.04843294247984886f;
  t283_n126: return 0.11534325778484344f;
  t283_n127: return 0.008290820755064487f;
  t283_n128: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t283_n130;
  t283_n129: return -0.047432176768779755f;
  t283_n130: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t283_n132;
  t283_n131: return 0.005596037954092026f;
  t283_n132: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t283_n134;
  t283_n133: return -0.026904894039034843f;
  t283_n134: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t283_n136;
  t283_n135: return 0.14480134844779968f;
  t283_n136: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t283_n138;
  t283_n137: return -0.015200447291135788f;
  t283_n138: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t283_n140;
  t283_n139: return 0.12062855064868927f;
  t283_n140: return 0.0784568265080452f;
  t283_n141: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t283_n143;
  t283_n142: return 0.012293247506022453f;
  t283_n143: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t283_n159;
  t283_n144: if (E.getNumReferences() >= 3248750592 /*20.5*/) goto t283_n156;
  t283_n145: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t283_n147;
  t283_n146: return 0.003291820175945759f;
  t283_n147: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t283_n149;
  t283_n148: return 0.013950305990874767f;
  t283_n149: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t283_n155;
  t283_n150: if (E.getFractionNameInContext() >= 3191674694 /*0.18465909361839294*/) goto t283_n152;
  t283_n151: return -0.04054194316267967f;
  t283_n152: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t283_n154;
  t283_n153: return -0.021389326080679893f;
  t283_n154: return 0.042457785457372665f;
  t283_n155: return -0.4020061492919922f;
  t283_n156: if (E.getFileProximityDistanceCost() >= 3253993472 /*30.5*/) goto t283_n158;
  t283_n157: return -0.016746761277318f;
  t283_n158: return 0.09712986648082733f;
  t283_n159: return 0.0641360729932785f;
  t283_n160: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t283_n254;
  t283_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t283_n163;
  t283_n162: return -0.004516164772212505f;
  t283_n163: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t283_n229;
  t283_n164: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t283_n226;
  t283_n165: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t283_n171;
  t283_n166: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t283_n168;
  t283_n167: return -0.2569580674171448f;
  t283_n168: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t283_n170;
  t283_n169: return -0.165740504860878f;
  t283_n170: return 0.10020925849676132f;
  t283_n171: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t283_n173;
  t283_n172: return -0.013074085116386414f;
  t283_n173: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t283_n189;
  t283_n174: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t283_n188;
  t283_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t283_n181;
  t283_n176: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t283_n178;
  t283_n177: return 0.030366649851202965f;
  t283_n178: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t283_n180;
  t283_n179: return 0.03596092388033867f;
  t283_n180: return -0.330665647983551f;
  t283_n181: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t283_n183;
  t283_n182: return 0.08536576479673386f;
  t283_n183: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t283_n185;
  t283_n184: return 0.14242222905158997f;
  t283_n185: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t283_n187;
  t283_n186: return -0.10971590876579285f;
  t283_n187: return 0.12172307074069977f;
  t283_n188: return 0.16642700135707855f;
  t283_n189: if (E.getFractionNameInContext() >= 3182587036 /*0.08712121844291687*/) goto t283_n215;
  t283_n190: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t283_n208;
  t283_n191: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t283_n207;
  t283_n192: if (E.getNumReferences() >= 3263234048 /*64.5*/) goto t283_n206;
  t283_n193: if (E.getNumReferences() >= 3261988864 /*59.5*/) goto t283_n201;
  t283_n194: if (E.getNumReferences() >= 3259105280 /*48.5*/) goto t283_n196;
  t283_n195: return -0.06599827855825424f;
  t283_n196: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t283_n198;
  t283_n197: return -0.3668965697288513f;
  t283_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t283_n200;
  t283_n199: return 0.0343395359814167f;
  t283_n200: return 0.11457803100347519f;
  t283_n201: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t283_n205;
  t283_n202: if (E.getNumReferences() >= 3262513152 /*61.5*/) goto t283_n204;
  t283_n203: return 0.11914422363042831f;
  t283_n204: return 0.07970547676086426f;
  t283_n205: return 0.13420416414737701f;
  t283_n206: return 0.0074435132555663586f;
  t283_n207: return -0.10211606323719025f;
  t283_n208: if (E.getNumReferences() >= 3265069056 /*78.5*/) goto t283_n214;
  t283_n209: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t283_n211;
  t283_n210: return -0.2984766960144043f;
  t283_n211: if (E.getNumReferences() >= 3259629568 /*50.5*/) goto t283_n213;
  t283_n212: return 0.10282175242900848f;
  t283_n213: return -0.014033008366823196f;
  t283_n214: return -0.41528311371803284f;
  t283_n215: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t283_n217;
  t283_n216: return 0.025616394355893135f;
  t283_n217: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t283_n221;
  t283_n218: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t283_n220;
  t283_n219: return 0.07354053854942322f;
  t283_n220: return 0.10623906552791595f;
  t283_n221: if (E.getFractionNameInContext() >= 3196658834 /*0.2678571343421936*/) goto t283_n223;
  t283_n222: return -0.11064377427101135f;
  t283_n223: if (E.getFractionNameInContext() >= 3200063302 /*0.3693181872367859*/) goto t283_n225;
  t283_n224: return 0.14153917133808136f;
  t283_n225: return 0.08140766620635986f;
  t283_n226: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t283_n228;
  t283_n227: return 0.007166816387325525f;
  t283_n228: return 0.11164341866970062f;
  t283_n229: if (E.getNumReferences() >= 3252944896 /*28.5*/) goto t283_n245;
  t283_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t283_n232;
  t283_n231: return 0.028802959248423576f;
  t283_n232: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t283_n244;
  t283_n233: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t283_n235;
  t283_n234: return -0.0004195295041427016f;
  t283_n235: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t283_n243;
  t283_n236: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t283_n242;
  t283_n237: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t283_n241;
  t283_n238: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t283_n240;
  t283_n239: return 0.43785396218299866f;
  t283_n240: return -0.1032906323671341f;
  t283_n241: return 0.1646849364042282f;
  t283_n242: return 0.03278757259249687f;
  t283_n243: return -0.1778232455253601f;
  t283_n244: return 0.00033857920789159834f;
  t283_n245: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t283_n247;
  t283_n246: return -0.001102431328035891f;
  t283_n247: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t283_n249;
  t283_n248: return 0.0013341588201001287f;
  t283_n249: if (E.getNumReferences() >= 3273687040 /*160.5*/) goto t283_n251;
  t283_n250: return -0.15835566818714142f;
  t283_n251: if (E.getNumReferences() >= 3297529856 /*1122.5*/) goto t283_n253;
  t283_n252: return 0.1652044802904129f;
  t283_n253: return 0.0881490558385849f;
  t283_n254: return 0.00014864183322060853f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree284(const Example& E) {
  t284_n0: if (E.getFractionNameInContext() >= 3205295590 /*0.5505050420761108*/) goto t284_n82;
  t284_n1: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t284_n3;
  t284_n2: return -0.0001454634766560048f;
  t284_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t284_n81;
  t284_n4: if (E.getNumNameInContext() >= 3236954112 /*7.5*/) goto t284_n80;
  t284_n5: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t284_n7;
  t284_n6: return 0.00020031046005897224f;
  t284_n7: if (E.getFractionNameInContext() >= 3196661908 /*0.2679487466812134*/) goto t284_n37;
  t284_n8: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t284_n36;
  t284_n9: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t284_n21;
  t284_n10: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t284_n20;
  t284_n11: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t284_n19;
  t284_n12: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t284_n14;
  t284_n13: return 0.014573031105101109f;
  t284_n14: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t284_n18;
  t284_n15: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t284_n17;
  t284_n16: return 0.059211768209934235f;
  t284_n17: return 0.12375235557556152f;
  t284_n18: return -0.002020994434133172f;
  t284_n19: return 0.010512608103454113f;
  t284_n20: return -0.01418248750269413f;
  t284_n21: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t284_n29;
  t284_n22: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n28;
  t284_n23: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t284_n25;
  t284_n24: return 0.057517845183610916f;
  t284_n25: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t284_n27;
  t284_n26: return -0.13897189497947693f;
  t284_n27: return -0.03159727528691292f;
  t284_n28: return -0.011812795884907246f;
  t284_n29: if (E.getFractionNameInContext() >= 3195566200 /*0.24264705181121826*/) goto t284_n31;
  t284_n30: return 0.027827464044094086f;
  t284_n31: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t284_n35;
  t284_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess))) goto t284_n34;
  t284_n33: return 0.09801536053419113f;
  t284_n34: return 0.10470081865787506f;
  t284_n35: return -0.07078628987073898f;
  t284_n36: return -0.12088984251022339f;
  t284_n37: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t284_n77;
  t284_n38: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t284_n68;
  t284_n39: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t284_n65;
  t284_n40: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t284_n64;
  t284_n41: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t284_n57;
  t284_n42: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t284_n56;
  t284_n43: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t284_n55;
  t284_n44: if (E.getFractionNameInContext() >= 3196906982 /*0.2752525210380554*/) goto t284_n46;
  t284_n45: return -0.003590669482946396f;
  t284_n46: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t284_n54;
  t284_n47: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t284_n49;
  t284_n48: return 0.03780939429998398f;
  t284_n49: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t284_n53;
  t284_n50: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t284_n52;
  t284_n51: return 0.05625839903950691f;
  t284_n52: return 0.1419769823551178f;
  t284_n53: return 0.05589016526937485f;
  t284_n54: return 0.09058568626642227f;
  t284_n55: return 0.11615650355815887f;
  t284_n56: return -0.022599928081035614f;
  t284_n57: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t284_n63;
  t284_n58: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t284_n62;
  t284_n59: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t284_n61;
  t284_n60: return -0.18919523060321808f;
  t284_n61: return -0.03357359766960144f;
  t284_n62: return -0.011471418663859367f;
  t284_n63: return 0.04518374428153038f;
  t284_n64: return 0.009843437001109123f;
  t284_n65: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t284_n67;
  t284_n66: return 0.11924338340759277f;
  t284_n67: return 0.10586969554424286f;
  t284_n68: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t284_n70;
  t284_n69: return -0.021837454289197922f;
  t284_n70: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t284_n76;
  t284_n71: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t284_n73;
  t284_n72: return 0.06927590072154999f;
  t284_n73: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t284_n75;
  t284_n74: return -0.17501327395439148f;
  t284_n75: return -0.02909168414771557f;
  t284_n76: return 0.016895709559321404f;
  t284_n77: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n79;
  t284_n78: return -0.006122835911810398f;
  t284_n79: return 0.011419144459068775f;
  t284_n80: return 0.07221994549036026f;
  t284_n81: return 0.003407147480174899f;
  t284_n82: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t284_n192;
  t284_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n151;
  t284_n84: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t284_n86;
  t284_n85: return -0.008752009831368923f;
  t284_n86: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t284_n146;
  t284_n87: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t284_n145;
  t284_n88: if (E.getFractionNameInContext() >= 3206894934 /*0.6458333730697632*/) goto t284_n128;
  t284_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t284_n91;
  t284_n90: return 0.026648417115211487f;
  t284_n91: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t284_n109;
  t284_n92: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t284_n106;
  t284_n93: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t284_n103;
  t284_n94: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t284_n102;
  t284_n95: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t284_n101;
  t284_n96: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t284_n100;
  t284_n97: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t284_n99;
  t284_n98: return 0.09566418826580048f;
  t284_n99: return 0.0693674236536026f;
  t284_n100: return 0.10620760917663574f;
  t284_n101: return -0.1156231164932251f;
  t284_n102: return 0.04021063819527626f;
  t284_n103: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t284_n105;
  t284_n104: return 0.0910695418715477f;
  t284_n105: return 0.16309760510921478f;
  t284_n106: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t284_n108;
  t284_n107: return -0.01931682601571083f;
  t284_n108: return 0.11491642147302628f;
  t284_n109: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t284_n111;
  t284_n110: return -0.46004801988601685f;
  t284_n111: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t284_n113;
  t284_n112: return -0.23372401297092438f;
  t284_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t284_n125;
  t284_n114: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t284_n120;
  t284_n115: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t284_n119;
  t284_n116: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t284_n118;
  t284_n117: return 0.0954921618103981f;
  t284_n118: return -0.22712810337543488f;
  t284_n119: return -0.16167116165161133f;
  t284_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t284_n124;
  t284_n121: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t284_n123;
  t284_n122: return 0.11459980905056f;
  t284_n123: return -0.1804981678724289f;
  t284_n124: return 0.08581960201263428f;
  t284_n125: if (E.getFractionNameInContext() >= 3205886304 /*0.5857143402099609*/) goto t284_n127;
  t284_n126: return 0.09316626191139221f;
  t284_n127: return 0.06765687465667725f;
  t284_n128: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t284_n144;
  t284_n129: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t284_n137;
  t284_n130: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t284_n136;
  t284_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n133;
  t284_n132: return -0.09973003715276718f;
  t284_n133: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t284_n135;
  t284_n134: return 0.07285712659358978f;
  t284_n135: return 0.1760471761226654f;
  t284_n136: return -0.03708217665553093f;
  t284_n137: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t284_n139;
  t284_n138: return 0.05324793606996536f;
  t284_n139: if (E.getNumReferences() >= 3253731328 /*30.0*/) goto t284_n143;
  t284_n140: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t284_n142;
  t284_n141: return -0.1128784790635109f;
  t284_n142: return 0.030366284772753716f;
  t284_n143: return -0.2785165011882782f;
  t284_n144: return -0.0064353421330451965f;
  t284_n145: return -0.014811055734753609f;
  t284_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t284_n148;
  t284_n147: return 0.018495982512831688f;
  t284_n148: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t284_n150;
  t284_n149: return 0.11290279030799866f;
  t284_n150: return 0.09062197804450989f;
  t284_n151: if (E.getFileProximityDistanceCost() >= 3223322624 /*2.5*/) goto t284_n153;
  t284_n152: return -0.22260399162769318f;
  t284_n153: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t284_n177;
  t284_n154: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator))) goto t284_n158;
  t284_n155: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t284_n157;
  t284_n156: return 0.017613261938095093f;
  t284_n157: return -0.03228142485022545f;
  t284_n158: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t284_n162;
  t284_n159: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t284_n161;
  t284_n160: return 0.06437298655509949f;
  t284_n161: return -0.15524600446224213f;
  t284_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n166;
  t284_n163: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t284_n165;
  t284_n164: return 0.12083736807107925f;
  t284_n165: return -0.20705828070640564f;
  t284_n166: if (E.getFileProximityDistanceCost() >= 3249537024 /*22.0*/) goto t284_n174;
  t284_n167: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t284_n171;
  t284_n168: if (E.getNumReferences() >= 3321888768 /*8192.0*/) goto t284_n170;
  t284_n169: return 0.1328655332326889f;
  t284_n170: return 0.07144315540790558f;
  t284_n171: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t284_n173;
  t284_n172: return -0.18056468665599823f;
  t284_n173: return 0.11150423437356949f;
  t284_n174: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t284_n176;
  t284_n175: return 0.07426398992538452f;
  t284_n176: return -0.03574608638882637f;
  t284_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t284_n191;
  t284_n178: if (E.getNumReferences() >= 3268083712 /*101.5*/) goto t284_n180;
  t284_n179: return 0.0072935232892632484f;
  t284_n180: if (E.getNumReferences() >= 3269525504 /*112.5*/) goto t284_n182;
  t284_n181: return 0.14890284836292267f;
  t284_n182: if (E.getNumReferences() >= 3276963840 /*210.5*/) goto t284_n186;
  t284_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t284_n185;
  t284_n184: return 0.09682421386241913f;
  t284_n185: return 0.14388826489448547f;
  t284_n186: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n188;
  t284_n187: return -0.25881803035736084f;
  t284_n188: if (E.getNumReferences() >= 3324362240 /*10607.5*/) goto t284_n190;
  t284_n189: return -0.011325075291097164f;
  t284_n190: return 0.1193232461810112f;
  t284_n191: return 0.024277832359075546f;
  t284_n192: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement))) goto t284_n200;
  t284_n193: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t284_n195;
  t284_n194: return -0.003323411801829934f;
  t284_n195: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t284_n199;
  t284_n196: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t284_n198;
  t284_n197: return 0.08727813512086868f;
  t284_n198: return 0.06101454421877861f;
  t284_n199: return -0.053591251373291016f;
  t284_n200: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t284_n212;
  t284_n201: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t284_n203;
  t284_n202: return 0.016512295231223106f;
  t284_n203: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t284_n211;
  t284_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n210;
  t284_n205: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t284_n209;
  t284_n206: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t284_n208;
  t284_n207: return 0.11317108571529388f;
  t284_n208: return 0.044335126876831055f;
  t284_n209: return -0.03317493200302124f;
  t284_n210: return 0.03240743651986122f;
  t284_n211: return 0.16179245710372925f;
  t284_n212: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t284_n240;
  t284_n213: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t284_n217;
  t284_n214: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t284_n216;
  t284_n215: return -0.07425526529550552f;
  t284_n216: return 0.01622501201927662f;
  t284_n217: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t284_n229;
  t284_n218: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t284_n220;
  t284_n219: return -0.03137302026152611f;
  t284_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t284_n226;
  t284_n221: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t284_n223;
  t284_n222: return -0.07508248090744019f;
  t284_n223: if (E.getFractionNameInContext() >= 3207643916 /*0.6904761791229248*/) goto t284_n225;
  t284_n224: return 0.10794214904308319f;
  t284_n225: return -0.14847423136234283f;
  t284_n226: if (E.getScope() & (BIT(Scope_type::GlobalScope))) goto t284_n228;
  t284_n227: return 0.03308934345841408f;
  t284_n228: return 0.12138223648071289f;
  t284_n229: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t284_n239;
  t284_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t284_n232;
  t284_n231: return -0.12064848095178604f;
  t284_n232: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t284_n238;
  t284_n233: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t284_n235;
  t284_n234: return 0.05700526013970375f;
  t284_n235: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t284_n237;
  t284_n236: return 0.07169421017169952f;
  t284_n237: return 0.10316099226474762f;
  t284_n238: return 0.03979470580816269f;
  t284_n239: return 0.014295353554189205f;
  t284_n240: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t284_n246;
  t284_n241: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t284_n243;
  t284_n242: return -0.416925311088562f;
  t284_n243: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t284_n245;
  t284_n244: return 0.08484052121639252f;
  t284_n245: return -0.1801922768354416f;
  t284_n246: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t284_n252;
  t284_n247: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t284_n251;
  t284_n248: if (E.getNumReferences() >= 3229614080 /*4.0*/) goto t284_n250;
  t284_n249: return 0.1225217878818512f;
  t284_n250: return -0.19225914776325226f;
  t284_n251: return -0.06447757035493851f;
  t284_n252: if (E.getFractionNameInContext() >= 3205986168 /*0.5916666984558105*/) goto t284_n254;
  t284_n253: return 0.11160530894994736f;
  t284_n254: return -0.054471131414175034f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree285(const Example& E) {
  t285_n0: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t285_n252;
  t285_n1: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t285_n3;
  t285_n2: return -0.08924809098243713f;
  t285_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t285_n145;
  t285_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t285_n8;
  t285_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n7;
  t285_n6: return -0.0018201445927843451f;
  t285_n7: return 0.0006290103192441165f;
  t285_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t285_n104;
  t285_n9: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t285_n51;
  t285_n10: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t285_n28;
  t285_n11: if (E.getNumReferences() >= 3309969408 /*3234.0*/) goto t285_n25;
  t285_n12: if (E.getNumReferences() >= 3309606912 /*3145.5*/) goto t285_n24;
  t285_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n23;
  t285_n14: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t285_n18;
  t285_n15: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t285_n17;
  t285_n16: return -0.0055223372764885426f;
  t285_n17: return 0.14840231835842133f;
  t285_n18: if (E.getNumReferences() >= 3301462016 /*1602.5*/) goto t285_n22;
  t285_n19: if (E.getNumReferences() >= 3299647488 /*1381.0*/) goto t285_n21;
  t285_n20: return -0.0267350934445858f;
  t285_n21: return 0.10409433394670486f;
  t285_n22: return -0.2511843144893646f;
  t285_n23: return -0.0019132355228066444f;
  t285_n24: return 0.13516667485237122f;
  t285_n25: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t285_n27;
  t285_n26: return -0.07817521691322327f;
  t285_n27: return -0.4553963541984558f;
  t285_n28: if (E.getFractionNameInContext() >= 3185414456 /*0.10818713903427124*/) goto t285_n38;
  t285_n29: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t285_n31;
  t285_n30: return 0.016221530735492706f;
  t285_n31: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t285_n35;
  t285_n32: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t285_n34;
  t285_n33: return -0.026000086218118668f;
  t285_n34: return -2.4820363521575928f;
  t285_n35: if (E.getNumReferences() >= 3260284928 /*53.0*/) goto t285_n37;
  t285_n36: return 0.14237266778945923f;
  t285_n37: return 0.03625686839222908f;
  t285_n38: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t285_n48;
  t285_n39: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n47;
  t285_n40: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t285_n42;
  t285_n41: return -0.1407686471939087f;
  t285_n42: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t285_n46;
  t285_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t285_n45;
  t285_n44: return -0.015421172603964806f;
  t285_n45: return 0.10705236345529556f;
  t285_n46: return 0.05850023403763771f;
  t285_n47: return 0.0035237015690654516f;
  t285_n48: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t285_n50;
  t285_n49: return 0.07022646069526672f;
  t285_n50: return 0.19281306862831116f;
  t285_n51: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t285_n103;
  t285_n52: if (E.getNumReferences() >= 3275653120 /*190.5*/) goto t285_n100;
  t285_n53: if (E.getFileProximityDistanceCost() >= 3255566336 /*35.0*/) goto t285_n97;
  t285_n54: if (E.getNumReferences() >= 3255959552 /*36.5*/) goto t285_n82;
  t285_n55: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t285_n57;
  t285_n56: return 0.03266829624772072f;
  t285_n57: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t285_n81;
  t285_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t285_n72;
  t285_n59: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t285_n71;
  t285_n60: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t285_n70;
  t285_n61: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t285_n63;
  t285_n62: return 0.08533185720443726f;
  t285_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n65;
  t285_n64: return -0.07542635500431061f;
  t285_n65: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t285_n67;
  t285_n66: return 0.4113865792751312f;
  t285_n67: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t285_n69;
  t285_n68: return -0.012051818892359734f;
  t285_n69: return 0.1775113046169281f;
  t285_n70: return 0.09217248111963272f;
  t285_n71: return -0.08638297021389008f;
  t285_n72: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t285_n78;
  t285_n73: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t285_n75;
  t285_n74: return 0.0372600220143795f;
  t285_n75: if (E.getNumReferences() >= 3235905536 /*7.0*/) goto t285_n77;
  t285_n76: return 0.19021964073181152f;
  t285_n77: return 0.02544621005654335f;
  t285_n78: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t285_n80;
  t285_n79: return 0.20447179675102234f;
  t285_n80: return -0.43815314769744873f;
  t285_n81: return -0.14689026772975922f;
  t285_n82: if (E.getNumReferences() >= 3256745984 /*39.5*/) goto t285_n88;
  t285_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t285_n85;
  t285_n84: return -0.12993896007537842f;
  t285_n85: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t285_n87;
  t285_n86: return 0.19910970330238342f;
  t285_n87: return -0.048981864005327225f;
  t285_n88: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t285_n94;
  t285_n89: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t285_n93;
  t285_n90: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t285_n92;
  t285_n91: return 0.07696942239999771f;
  t285_n92: return -0.01795496605336666f;
  t285_n93: return 0.13163572549819946f;
  t285_n94: if (E.getFractionNameInContext() >= 3188230281 /*0.13333334028720856*/) goto t285_n96;
  t285_n95: return -0.12103817611932755f;
  t285_n96: return 0.14458103477954865f;
  t285_n97: if (E.getFractionNameInContext() >= 3182078635 /*0.0833333358168602*/) goto t285_n99;
  t285_n98: return -1.2377821207046509f;
  t285_n99: return -0.10547488927841187f;
  t285_n100: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t285_n102;
  t285_n101: return -0.08893584460020065f;
  t285_n102: return 0.12200689315795898f;
  t285_n103: return 0.10355112701654434f;
  t285_n104: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t285_n106;
  t285_n105: return 0.011452412232756615f;
  t285_n106: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t285_n112;
  t285_n107: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t285_n109;
  t285_n108: return 0.007907557301223278f;
  t285_n109: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t285_n111;
  t285_n110: return 0.1019468903541565f;
  t285_n111: return 0.06719762086868286f;
  t285_n112: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t285_n130;
  t285_n113: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t285_n125;
  t285_n114: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t285_n116;
  t285_n115: return -0.017701702192425728f;
  t285_n116: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t285_n122;
  t285_n117: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t285_n121;
  t285_n118: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t285_n120;
  t285_n119: return 0.14322613179683685f;
  t285_n120: return -0.19704721868038177f;
  t285_n121: return -0.12263986468315125f;
  t285_n122: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t285_n124;
  t285_n123: return 0.0038750567473471165f;
  t285_n124: return 0.08613144606351852f;
  t285_n125: if (E.getNumReferences() >= 3292061696 /*739.5*/) goto t285_n127;
  t285_n126: return -0.10023950040340424f;
  t285_n127: if (E.getNumReferences() >= 3315082240 /*4868.5*/) goto t285_n129;
  t285_n128: return 0.13025353848934174f;
  t285_n129: return 0.10212335735559464f;
  t285_n130: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t285_n138;
  t285_n131: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t285_n133;
  t285_n132: return 0.002231861697509885f;
  t285_n133: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t285_n137;
  t285_n134: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t285_n136;
  t285_n135: return 0.018679527565836906f;
  t285_n136: return 0.10717497766017914f;
  t285_n137: return 0.034071408212184906f;
  t285_n138: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t285_n144;
  t285_n139: if (E.getNumReferences() >= 3274407936 /*171.5*/) goto t285_n141;
  t285_n140: return 0.07433415204286575f;
  t285_n141: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t285_n143;
  t285_n142: return 0.14657673239707947f;
  t285_n143: return -0.10156191140413284f;
  t285_n144: return 0.11818061023950577f;
  t285_n145: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t285_n197;
  t285_n146: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t285_n148;
  t285_n147: return -0.0003088563389610499f;
  t285_n148: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t285_n150;
  t285_n149: return -0.008910941891372204f;
  t285_n150: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel))) goto t285_n164;
  t285_n151: if (E.getNumReferences() >= 3317586944 /*6091.5*/) goto t285_n159;
  t285_n152: if (E.getNumReferences() >= 3306360832 /*2353.0*/) goto t285_n154;
  t285_n153: return -0.016624774783849716f;
  t285_n154: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t285_n158;
  t285_n155: if (E.getNumReferences() >= 3308277760 /*2821.0*/) goto t285_n157;
  t285_n156: return 0.11388401687145233f;
  t285_n157: return 0.04952392727136612f;
  t285_n158: return -0.559287965297699f;
  t285_n159: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n163;
  t285_n160: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t285_n162;
  t285_n161: return -0.2214663028717041f;
  t285_n162: return -0.48665204644203186f;
  t285_n163: return 0.003483693115413189f;
  t285_n164: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t285_n182;
  t285_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t285_n177;
  t285_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t285_n170;
  t285_n167: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t285_n169;
  t285_n168: return 0.10843616724014282f;
  t285_n169: return 0.05911698192358017f;
  t285_n170: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t285_n176;
  t285_n171: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t285_n175;
  t285_n172: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t285_n174;
  t285_n173: return 0.0960698351264f;
  t285_n174: return 0.10520046204328537f;
  t285_n175: return 0.03686971589922905f;
  t285_n176: return -0.15821675956249237f;
  t285_n177: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t285_n179;
  t285_n178: return 0.07457812130451202f;
  t285_n179: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t285_n181;
  t285_n180: return 0.14098235964775085f;
  t285_n181: return 0.09683255851268768f;
  t285_n182: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t285_n196;
  t285_n183: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t285_n193;
  t285_n184: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t285_n186;
  t285_n185: return 0.0005711310659535229f;
  t285_n186: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t285_n192;
  t285_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t285_n189;
  t285_n188: return 0.10908263176679611f;
  t285_n189: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t285_n191;
  t285_n190: return 0.4494270384311676f;
  t285_n191: return 0.07424722611904144f;
  t285_n192: return -0.010418250225484371f;
  t285_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n195;
  t285_n194: return -0.1153671070933342f;
  t285_n195: return 0.13274551928043365f;
  t285_n196: return -0.1266002058982849f;
  t285_n197: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t285_n249;
  t285_n198: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t285_n248;
  t285_n199: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t285_n203;
  t285_n200: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t285_n202;
  t285_n201: return -0.001137568848207593f;
  t285_n202: return -0.029360005632042885f;
  t285_n203: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t285_n231;
  t285_n204: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t285_n228;
  t285_n205: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t285_n227;
  t285_n206: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t285_n210;
  t285_n207: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n209;
  t285_n208: return -0.06508219242095947f;
  t285_n209: return 0.028582138940691948f;
  t285_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t285_n220;
  t285_n211: if (E.getNumReferences() >= 3239575552 /*9.5*/) goto t285_n219;
  t285_n212: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t285_n214;
  t285_n213: return 0.012552897445857525f;
  t285_n214: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t285_n216;
  t285_n215: return 0.0854552760720253f;
  t285_n216: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t285_n218;
  t285_n217: return -0.14429804682731628f;
  t285_n218: return 0.050626400858163834f;
  t285_n219: return 0.01700952649116516f;
  t285_n220: if (E.getNumReferences() >= 3261595648 /*58.0*/) goto t285_n226;
  t285_n221: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t285_n223;
  t285_n222: return 0.13204322755336761f;
  t285_n223: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t285_n225;
  t285_n224: return 0.05233054608106613f;
  t285_n225: return 0.10637024790048599f;
  t285_n226: return -0.08479256182909012f;
  t285_n227: return 0.0024551956448704004f;
  t285_n228: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n230;
  t285_n229: return -0.10177964717149734f;
  t285_n230: return 0.037222035229206085f;
  t285_n231: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t285_n247;
  t285_n232: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t285_n240;
  t285_n233: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t285_n237;
  t285_n234: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n236;
  t285_n235: return -0.27304112911224365f;
  t285_n236: return -0.04135400429368019f;
  t285_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n239;
  t285_n238: return -0.03309830650687218f;
  t285_n239: return 0.011579221114516258f;
  t285_n240: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n244;
  t285_n241: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope))) goto t285_n243;
  t285_n242: return -0.025351259857416153f;
  t285_n243: return 0.0007001481135375798f;
  t285_n244: if (E.getNumReferences() >= 3364286400 /*276350.0*/) goto t285_n246;
  t285_n245: return 0.0073621817864477634f;
  t285_n246: return -0.04828207939863205f;
  t285_n247: return 0.003144486341625452f;
  t285_n248: return -0.0486026369035244f;
  t285_n249: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t285_n251;
  t285_n250: return 0.009205548092722893f;
  t285_n251: return -0.11460912972688675f;
  t285_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t285_n254;
  t285_n253: return -0.03057512268424034f;
  t285_n254: return -0.004311550408601761f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree286(const Example& E) {
  t286_n0: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t286_n16;
  t286_n1: if (E.getNumReferences() >= 3364818608 /*292981.5*/) goto t286_n5;
  t286_n2: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t286_n4;
  t286_n3: return 0.0009848774643614888f;
  t286_n4: return -0.17579662799835205f;
  t286_n5: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t286_n15;
  t286_n6: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t286_n8;
  t286_n7: return 0.006714810151606798f;
  t286_n8: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t286_n12;
  t286_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t286_n11;
  t286_n10: return 0.09033071249723434f;
  t286_n11: return 0.1277230978012085f;
  t286_n12: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t286_n14;
  t286_n13: return -0.2852419912815094f;
  t286_n14: return 0.11111972481012344f;
  t286_n15: return 0.004716977477073669f;
  t286_n16: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t286_n22;
  t286_n17: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t286_n19;
  t286_n18: return -2.8240128813195042e-06f;
  t286_n19: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t286_n21;
  t286_n20: return 0.00850109662860632f;
  t286_n21: return -0.002076759934425354f;
  t286_n22: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t286_n132;
  t286_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t286_n25;
  t286_n24: return -0.041544605046510696f;
  t286_n25: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t286_n113;
  t286_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag))) goto t286_n108;
  t286_n27: if (E.getNumReferences() >= 3364870464 /*294602.0*/) goto t286_n107;
  t286_n28: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t286_n44;
  t286_n29: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t286_n31;
  t286_n30: return -0.016951261088252068f;
  t286_n31: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n39;
  t286_n32: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t286_n34;
  t286_n33: return 0.026205414906144142f;
  t286_n34: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t286_n36;
  t286_n35: return 0.1241949200630188f;
  t286_n36: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t286_n38;
  t286_n37: return 0.037893783301115036f;
  t286_n38: return 0.11628150194883347f;
  t286_n39: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t286_n43;
  t286_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n42;
  t286_n41: return -0.051255401223897934f;
  t286_n42: return 0.0927814394235611f;
  t286_n43: return 0.1036534309387207f;
  t286_n44: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t286_n84;
  t286_n45: if (E.getNumReferences() >= 3352044608 /*104520.5*/) goto t286_n83;
  t286_n46: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t286_n58;
  t286_n47: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t286_n49;
  t286_n48: return -0.046476900577545166f;
  t286_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n55;
  t286_n50: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t286_n54;
  t286_n51: if (E.getNumReferences() >= 3339744256 /*36980.0*/) goto t286_n53;
  t286_n52: return -0.0607980340719223f;
  t286_n53: return 0.04825780168175697f;
  t286_n54: return 0.05586816743016243f;
  t286_n55: if (E.getNumReferences() >= 3347973248 /*72713.0*/) goto t286_n57;
  t286_n56: return 0.016895882785320282f;
  t286_n57: return -0.15076492726802826f;
  t286_n58: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n60;
  t286_n59: return -0.01652326062321663f;
  t286_n60: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t286_n74;
  t286_n61: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n63;
  t286_n62: return -0.005759338848292828f;
  t286_n63: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t286_n73;
  t286_n64: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t286_n72;
  t286_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t286_n67;
  t286_n66: return 0.0214782003313303f;
  t286_n67: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t286_n69;
  t286_n68: return 0.032249122858047485f;
  t286_n69: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t286_n71;
  t286_n70: return 0.09701702743768692f;
  t286_n71: return 0.03413897752761841f;
  t286_n72: return 0.1994190365076065f;
  t286_n73: return 0.00785101018846035f;
  t286_n74: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t286_n76;
  t286_n75: return -0.018665216863155365f;
  t286_n76: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t286_n82;
  t286_n77: if (E.getNumReferences() >= 3221225472 /*2.0*/) goto t286_n81;
  t286_n78: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t286_n80;
  t286_n79: return 0.12834300100803375f;
  t286_n80: return -0.12088781595230103f;
  t286_n81: return -0.010186317376792431f;
  t286_n82: return 0.0023065090645104647f;
  t286_n83: return 0.03432159870862961f;
  t286_n84: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t286_n100;
  t286_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t286_n99;
  t286_n86: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n88;
  t286_n87: return 0.010933182202279568f;
  t286_n88: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t286_n98;
  t286_n89: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t286_n93;
  t286_n90: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t286_n92;
  t286_n91: return 0.043044909834861755f;
  t286_n92: return 0.07081421464681625f;
  t286_n93: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t286_n95;
  t286_n94: return 0.06251420080661774f;
  t286_n95: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Symbol))) goto t286_n97;
  t286_n96: return 0.07507546991109848f;
  t286_n97: return 0.18413811922073364f;
  t286_n98: return -0.015504920855164528f;
  t286_n99: return -0.0011295386357232928f;
  t286_n100: if (E.getNumReferences() >= 3318737920 /*6653.5*/) goto t286_n104;
  t286_n101: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t286_n103;
  t286_n102: return -0.02773539535701275f;
  t286_n103: return 0.014029252342879772f;
  t286_n104: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t286_n106;
  t286_n105: return 0.1060967668890953f;
  t286_n106: return 0.10936740040779114f;
  t286_n107: return -0.031132616102695465f;
  t286_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t286_n110;
  t286_n109: return 0.08336839824914932f;
  t286_n110: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t286_n112;
  t286_n111: return 0.05713396519422531f;
  t286_n112: return 0.12763427197933197f;
  t286_n113: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t286_n131;
  t286_n114: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n116;
  t286_n115: return -0.005029371008276939f;
  t286_n116: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t286_n130;
  t286_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n125;
  t286_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t286_n124;
  t286_n119: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t286_n123;
  t286_n120: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n122;
  t286_n121: return -0.062200017273426056f;
  t286_n122: return -0.10532920807600021f;
  t286_n123: return -0.0023401177022606134f;
  t286_n124: return 0.00493338517844677f;
  t286_n125: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t286_n129;
  t286_n126: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t286_n128;
  t286_n127: return 0.020411629229784012f;
  t286_n128: return -0.004745498299598694f;
  t286_n129: return -0.024608075618743896f;
  t286_n130: return 0.0022884225472807884f;
  t286_n131: return -0.023558596149086952f;
  t286_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n214;
  t286_n133: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t286_n155;
  t286_n134: if (E.getNumReferences() >= 3344562560 /*55801.5*/) goto t286_n154;
  t286_n135: if (E.getNumReferences() >= 3344481024 /*55483.0*/) goto t286_n151;
  t286_n136: if (E.getNumReferences() >= 3334293248 /*24227.5*/) goto t286_n150;
  t286_n137: if (E.getNumReferences() >= 3334023936 /*23701.5*/) goto t286_n139;
  t286_n138: return -0.015606881119310856f;
  t286_n139: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t286_n141;
  t286_n140: return 0.022748274728655815f;
  t286_n141: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t286_n143;
  t286_n142: return -0.3171735107898712f;
  t286_n143: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t286_n149;
  t286_n144: if (E.getNumReferences() >= 3334224384 /*24093.0*/) goto t286_n146;
  t286_n145: return 0.16277997195720673f;
  t286_n146: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n148;
  t286_n147: return 0.12518151104450226f;
  t286_n148: return 0.07398053258657455f;
  t286_n149: return 0.05758035182952881f;
  t286_n150: return -0.10118549317121506f;
  t286_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n153;
  t286_n152: return -0.10834933072328568f;
  t286_n153: return 0.1282811015844345f;
  t286_n154: return -0.12577266991138458f;
  t286_n155: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n157;
  t286_n156: return -0.04588489979505539f;
  t286_n157: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t286_n203;
  t286_n158: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t286_n202;
  t286_n159: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t286_n185;
  t286_n160: if (E.getNumReferences() >= 3329480192 /*15605.5*/) goto t286_n176;
  t286_n161: if (E.getNumReferences() >= 3327780864 /*13946.0*/) goto t286_n175;
  t286_n162: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t286_n166;
  t286_n163: if (E.getSymbolScopeDistanceCost() >= 3250061312 /*23.0*/) goto t286_n165;
  t286_n164: return -0.04519061744213104f;
  t286_n165: return 0.00520961033180356f;
  t286_n166: if (E.getSymbolScopeDistanceCost() >= 3245342720 /*15.0*/) goto t286_n174;
  t286_n167: if (E.getNumReferences() >= 3295485952 /*948.5*/) goto t286_n169;
  t286_n168: return 0.06536819785833359f;
  t286_n169: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t286_n173;
  t286_n170: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t286_n172;
  t286_n171: return -0.09342202544212341f;
  t286_n172: return 0.13135851919651031f;
  t286_n173: return 0.4389839470386505f;
  t286_n174: return -0.014628218486905098f;
  t286_n175: return -0.1620010882616043f;
  t286_n176: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t286_n180;
  t286_n177: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t286_n179;
  t286_n178: return -0.19613780081272125f;
  t286_n179: return 0.11621938645839691f;
  t286_n180: if (E.getNumReferences() >= 3361098688 /*219439.0*/) goto t286_n182;
  t286_n181: return 0.02113214321434498f;
  t286_n182: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t286_n184;
  t286_n183: return -0.16449719667434692f;
  t286_n184: return -0.0031076185405254364f;
  t286_n185: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t286_n187;
  t286_n186: return -0.03839542344212532f;
  t286_n187: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t286_n189;
  t286_n188: return 0.013733488507568836f;
  t286_n189: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t286_n193;
  t286_n190: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t286_n192;
  t286_n191: return 0.09926760196685791f;
  t286_n192: return 0.18892046809196472f;
  t286_n193: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t286_n195;
  t286_n194: return -0.19453516602516174f;
  t286_n195: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t286_n201;
  t286_n196: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t286_n200;
  t286_n197: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Variable))) goto t286_n199;
  t286_n198: return -0.10000117123126984f;
  t286_n199: return 0.09026122093200684f;
  t286_n200: return -0.02602568455040455f;
  t286_n201: return -0.2071797400712967f;
  t286_n202: return 0.012798049487173557f;
  t286_n203: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t286_n205;
  t286_n204: return -0.058374568819999695f;
  t286_n205: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t286_n207;
  t286_n206: return 0.015099333599209785f;
  t286_n207: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t286_n211;
  t286_n208: if (E.getSymbolScopeDistanceCost() >= 3240099840 /*10.0*/) goto t286_n210;
  t286_n209: return 0.2058526575565338f;
  t286_n210: return -0.0838884562253952f;
  t286_n211: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t286_n213;
  t286_n212: return 0.021771855652332306f;
  t286_n213: return -0.04191283881664276f;
  t286_n214: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t286_n238;
  t286_n215: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t286_n227;
  t286_n216: if (E.getNumReferences() >= 3279847424 /*254.5*/) goto t286_n224;
  t286_n217: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t286_n219;
  t286_n218: return 0.024530835449695587f;
  t286_n219: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t286_n223;
  t286_n220: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t286_n222;
  t286_n221: return 0.09149549156427383f;
  t286_n222: return -0.016595540568232536f;
  t286_n223: return -1.2869962453842163f;
  t286_n224: if (E.getNumReferences() >= 3280027648 /*258.5*/) goto t286_n226;
  t286_n225: return 0.22136101126670837f;
  t286_n226: return 0.0616641491651535f;
  t286_n227: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t286_n229;
  t286_n228: return 0.04103521257638931f;
  t286_n229: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t286_n231;
  t286_n230: return 0.06838086992502213f;
  t286_n231: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t286_n237;
  t286_n232: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t286_n234;
  t286_n233: return 0.1962926983833313f;
  t286_n234: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t286_n236;
  t286_n235: return 0.16157403588294983f;
  t286_n236: return 0.09486077725887299f;
  t286_n237: return 0.08738276362419128f;
  t286_n238: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t286_n240;
  t286_n239: return -0.0016081552021205425f;
  t286_n240: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t286_n242;
  t286_n241: return 0.004619547165930271f;
  t286_n242: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t286_n246;
  t286_n243: if (E.getNumReferences() >= 3377570160 /*858647.0*/) goto t286_n245;
  t286_n244: return -0.009539306163787842f;
  t286_n245: return -0.10505779087543488f;
  t286_n246: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t286_n254;
  t286_n247: if (E.getNumReferences() >= 3363889440 /*263945.0*/) goto t286_n251;
  t286_n248: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t286_n250;
  t286_n249: return 0.010880551300942898f;
  t286_n250: return -0.012567941099405289f;
  t286_n251: if (E.getNumReferences() >= 3364539984 /*284274.5*/) goto t286_n253;
  t286_n252: return 0.08005902171134949f;
  t286_n253: return -0.03383629396557808f;
  t286_n254: return -0.07996450364589691f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree287(const Example& E) {
  t287_n0: if (E.getNumReferences() >= 3328217600 /*14372.5*/) goto t287_n170;
  t287_n1: if (E.getNumReferences() >= 3328027648 /*14187.0*/) goto t287_n169;
  t287_n2: if (E.getNumReferences() >= 3327951360 /*14112.5*/) goto t287_n166;
  t287_n3: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t287_n37;
  t287_n4: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t287_n8;
  t287_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n7;
  t287_n6: return -0.002238690387457609f;
  t287_n7: return 0.0004911479772999883f;
  t287_n8: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t287_n28;
  t287_n9: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n11;
  t287_n10: return -0.004249603021889925f;
  t287_n11: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n17;
  t287_n12: if (E.getNumReferences() >= 3327654912 /*13823.0*/) goto t287_n14;
  t287_n13: return 0.0049867648631334305f;
  t287_n14: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t287_n16;
  t287_n15: return -0.0713181421160698f;
  t287_n16: return -0.3091064393520355f;
  t287_n17: if (E.getNumReferences() >= 3325015040 /*11245.0*/) goto t287_n27;
  t287_n18: if (E.getFractionNameInContext() >= 3200834702 /*0.39230769872665405*/) goto t287_n20;
  t287_n19: return -0.017665443941950798f;
  t287_n20: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t287_n26;
  t287_n21: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t287_n23;
  t287_n22: return 0.047345831990242004f;
  t287_n23: if (E.getNumReferences() >= 3258056704 /*44.5*/) goto t287_n25;
  t287_n24: return 0.12330986559391022f;
  t287_n25: return 0.013875874690711498f;
  t287_n26: return 0.024914102628827095f;
  t287_n27: return 0.08141900599002838f;
  t287_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n30;
  t287_n29: return -0.09908746927976608f;
  t287_n30: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t287_n36;
  t287_n31: if (E.getNumReferences() >= 3326745088 /*12934.5*/) goto t287_n33;
  t287_n32: return 0.011487902142107487f;
  t287_n33: if (E.getNumReferences() >= 3326809600 /*12997.5*/) goto t287_n35;
  t287_n34: return 0.12626834213733673f;
  t287_n35: return 0.049631934612989426f;
  t287_n36: return 0.00794640090316534f;
  t287_n37: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t287_n39;
  t287_n38: return 0.004157901741564274f;
  t287_n39: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t287_n113;
  t287_n40: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t287_n42;
  t287_n41: return -0.000749573518987745f;
  t287_n42: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t287_n112;
  t287_n43: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t287_n65;
  t287_n44: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t287_n46;
  t287_n45: return 0.007525093853473663f;
  t287_n46: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t287_n60;
  t287_n47: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t287_n59;
  t287_n48: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t287_n58;
  t287_n49: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t287_n55;
  t287_n50: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t287_n54;
  t287_n51: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t287_n53;
  t287_n52: return 0.09654080867767334f;
  t287_n53: return 0.016135310754179955f;
  t287_n54: return 0.10515424609184265f;
  t287_n55: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t287_n57;
  t287_n56: return 0.17941200733184814f;
  t287_n57: return 0.18290702998638153f;
  t287_n58: return 0.02416215091943741f;
  t287_n59: return -0.00609182845801115f;
  t287_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t287_n64;
  t287_n61: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t287_n63;
  t287_n62: return 0.06027568504214287f;
  t287_n63: return 0.031172320246696472f;
  t287_n64: return 0.5196470618247986f;
  t287_n65: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t287_n91;
  t287_n66: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t287_n70;
  t287_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n69;
  t287_n68: return -0.2824150323867798f;
  t287_n69: return -0.013419484719634056f;
  t287_n70: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t287_n76;
  t287_n71: if (E.getFractionNameInContext() >= 3196339268 /*0.25833332538604736*/) goto t287_n75;
  t287_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n74;
  t287_n73: return -0.08410437405109406f;
  t287_n74: return -0.006734130904078484f;
  t287_n75: return 0.014212153851985931f;
  t287_n76: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t287_n86;
  t287_n77: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t287_n85;
  t287_n78: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t287_n82;
  t287_n79: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t287_n81;
  t287_n80: return -0.02129954844713211f;
  t287_n81: return -0.1011577770113945f;
  t287_n82: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t287_n84;
  t287_n83: return 0.08091912418603897f;
  t287_n84: return 0.00786390621215105f;
  t287_n85: return -0.08671228587627411f;
  t287_n86: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n88;
  t287_n87: return 0.022819215431809425f;
  t287_n88: if (E.getFractionNameInContext() >= 3192094124 /*0.19090908765792847*/) goto t287_n90;
  t287_n89: return 0.0954703614115715f;
  t287_n90: return -0.07064452767372131f;
  t287_n91: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t287_n107;
  t287_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t287_n94;
  t287_n93: return 0.0027213534340262413f;
  t287_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t287_n96;
  t287_n95: return 0.07402250170707703f;
  t287_n96: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t287_n106;
  t287_n97: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t287_n105;
  t287_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t287_n100;
  t287_n99: return 0.031575415283441544f;
  t287_n100: if (E.getFractionNameInContext() >= 3193736649 /*0.2153846174478531*/) goto t287_n104;
  t287_n101: if (E.getFractionNameInContext() >= 3190545458 /*0.1678321659564972*/) goto t287_n103;
  t287_n102: return -0.1719723492860794f;
  t287_n103: return 0.2578906714916229f;
  t287_n104: return -0.2538992464542389f;
  t287_n105: return 0.07120689749717712f;
  t287_n106: return 0.05285480618476868f;
  t287_n107: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t287_n109;
  t287_n108: return -0.009550091810524464f;
  t287_n109: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t287_n111;
  t287_n110: return -0.18406878411769867f;
  t287_n111: return 0.0771169438958168f;
  t287_n112: return 0.010866246186196804f;
  t287_n113: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t287_n155;
  t287_n114: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t287_n154;
  t287_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace))) goto t287_n117;
  t287_n116: return 0.006311174016445875f;
  t287_n117: if (E.getFractionNameInContext() >= 3166899249 /*0.02380952425301075*/) goto t287_n153;
  t287_n118: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t287_n122;
  t287_n119: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t287_n121;
  t287_n120: return -0.016475439071655273f;
  t287_n121: return 0.09016178548336029f;
  t287_n122: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n134;
  t287_n123: if (E.getNumReferences() >= 3301322752 /*1585.5*/) goto t287_n133;
  t287_n124: if (E.getNumReferences() >= 3297366016 /*1102.5*/) goto t287_n128;
  t287_n125: if (E.getNumReferences() >= 3296788480 /*1032.0*/) goto t287_n127;
  t287_n126: return 0.08076095581054688f;
  t287_n127: return -0.9011546969413757f;
  t287_n128: if (E.getNumReferences() >= 3299004416 /*1302.5*/) goto t287_n130;
  t287_n129: return 0.10733944177627563f;
  t287_n130: if (E.getNumReferences() >= 3300118528 /*1438.5*/) goto t287_n132;
  t287_n131: return -0.1402643322944641f;
  t287_n132: return 0.19504515826702118f;
  t287_n133: return -0.019258154556155205f;
  t287_n134: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t287_n136;
  t287_n135: return -0.11804169416427612f;
  t287_n136: if (E.getNumReferences() >= 3258449920 /*46.0*/) goto t287_n138;
  t287_n137: return 0.1771744042634964f;
  t287_n138: if (E.getNumReferences() >= 3274866688 /*178.5*/) goto t287_n140;
  t287_n139: return 0.040245942771434784f;
  t287_n140: if (E.getNumReferences() >= 3291889664 /*729.0*/) goto t287_n148;
  t287_n141: if (E.getNumReferences() >= 3283533824 /*365.5*/) goto t287_n145;
  t287_n142: if (E.getNumReferences() >= 3277520896 /*219.0*/) goto t287_n144;
  t287_n143: return 0.16579711437225342f;
  t287_n144: return -0.10125704109668732f;
  t287_n145: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t287_n147;
  t287_n146: return 0.03006824478507042f;
  t287_n147: return 0.16912853717803955f;
  t287_n148: if (E.getNumReferences() >= 3301023744 /*1549.0*/) goto t287_n150;
  t287_n149: return -0.2986131012439728f;
  t287_n150: if (E.getNumReferences() >= 3310979072 /*3480.5*/) goto t287_n152;
  t287_n151: return 0.12337666749954224f;
  t287_n152: return 0.042773108929395676f;
  t287_n153: return -0.023854410275816917f;
  t287_n154: return -0.014791684225201607f;
  t287_n155: if (E.getNumReferences() >= 3326554624 /*12748.5*/) goto t287_n165;
  t287_n156: if (E.getNumReferences() >= 3326481408 /*12677.0*/) goto t287_n158;
  t287_n157: return 0.014368527568876743f;
  t287_n158: if (E.getNumReferences() >= 3326525952 /*12720.5*/) goto t287_n164;
  t287_n159: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t287_n163;
  t287_n160: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Type))) goto t287_n162;
  t287_n161: return 0.10569652169942856f;
  t287_n162: return 0.14198987185955048f;
  t287_n163: return 0.09045524150133133f;
  t287_n164: return 0.042771317064762115f;
  t287_n165: return -0.08568226546049118f;
  t287_n166: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t287_n168;
  t287_n167: return 0.10755576193332672f;
  t287_n168: return 0.10361836850643158f;
  t287_n169: return -0.11677736043930054f;
  t287_n170: if (E.getNumReferences() >= 3329021952 /*15158.0*/) goto t287_n202;
  t287_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t287_n173;
  t287_n172: return 0.0019320267019793391f;
  t287_n173: if (E.getNumReferences() >= 3328941568 /*15079.5*/) goto t287_n193;
  t287_n174: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t287_n186;
  t287_n175: if (E.getNumReferences() >= 3328450048 /*14599.5*/) goto t287_n185;
  t287_n176: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t287_n184;
  t287_n177: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t287_n179;
  t287_n178: return -0.055069562047719955f;
  t287_n179: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n183;
  t287_n180: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t287_n182;
  t287_n181: return 0.03719789534807205f;
  t287_n182: return 0.17312155663967133f;
  t287_n183: return 0.14549696445465088f;
  t287_n184: return -0.5739210247993469f;
  t287_n185: return 0.011080422438681126f;
  t287_n186: if (E.getNumReferences() >= 3328701440 /*14845.0*/) goto t287_n188;
  t287_n187: return -0.02041461504995823f;
  t287_n188: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t287_n192;
  t287_n189: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t287_n191;
  t287_n190: return 0.09200970828533173f;
  t287_n191: return 0.003976177424192429f;
  t287_n192: return 0.10996665060520172f;
  t287_n193: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n201;
  t287_n194: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t287_n200;
  t287_n195: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n197;
  t287_n196: return -0.11666391789913177f;
  t287_n197: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t287_n199;
  t287_n198: return 0.157576322555542f;
  t287_n199: return 0.1577855944633484f;
  t287_n200: return -0.09188617765903473f;
  t287_n201: return 0.031658679246902466f;
  t287_n202: if (E.getNumReferences() >= 3329553408 /*15677.0*/) goto t287_n216;
  t287_n203: if (E.getNumReferences() >= 3329258496 /*15389.0*/) goto t287_n215;
  t287_n204: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t287_n214;
  t287_n205: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n207;
  t287_n206: return -0.050584230571985245f;
  t287_n207: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t287_n213;
  t287_n208: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t287_n210;
  t287_n209: return -0.06257608532905579f;
  t287_n210: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t287_n212;
  t287_n211: return 0.06931983679533005f;
  t287_n212: return 0.12386784702539444f;
  t287_n213: return 0.06315435469150543f;
  t287_n214: return 0.06703482568264008f;
  t287_n215: return -0.09925859421491623f;
  t287_n216: if (E.getNumReferences() >= 3330013184 /*16126.0*/) goto t287_n246;
  t287_n217: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t287_n245;
  t287_n218: if (E.getNumReferences() >= 3329918464 /*16033.5*/) goto t287_n236;
  t287_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t287_n223;
  t287_n220: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t287_n222;
  t287_n221: return 0.026115750893950462f;
  t287_n222: return 0.1198127344250679f;
  t287_n223: if (E.getNumReferences() >= 3329847808 /*15964.5*/) goto t287_n231;
  t287_n224: if (E.getNumReferences() >= 3329564160 /*15687.5*/) goto t287_n230;
  t287_n225: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t287_n229;
  t287_n226: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t287_n228;
  t287_n227: return 0.1083170473575592f;
  t287_n228: return -0.18368585407733917f;
  t287_n229: return 0.014867128804326057f;
  t287_n230: return 0.035667113959789276f;
  t287_n231: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t287_n233;
  t287_n232: return 0.10093075037002563f;
  t287_n233: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t287_n235;
  t287_n234: return 0.12910480797290802f;
  t287_n235: return 0.12057824432849884f;
  t287_n236: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t287_n240;
  t287_n237: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n239;
  t287_n238: return -0.07123707234859467f;
  t287_n239: return 0.08919265121221542f;
  t287_n240: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t287_n242;
  t287_n241: return 0.1642187535762787f;
  t287_n242: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t287_n244;
  t287_n243: return 0.08925667405128479f;
  t287_n244: return 0.15192297101020813f;
  t287_n245: return -0.07480168342590332f;
  t287_n246: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t287_n252;
  t287_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t287_n251;
  t287_n248: if (E.getNumReferences() >= 3380813364 /*1074118.5*/) goto t287_n250;
  t287_n249: return -0.04824540764093399f;
  t287_n250: return 0.08231926709413528f;
  t287_n251: return -0.007215278688818216f;
  t287_n252: if (E.getNumReferences() >= 3330058752 /*16170.5*/) goto t287_n254;
  t287_n253: return -0.17235681414604187f;
  t287_n254: return 0.003436499275267124f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree288(const Example& E) {
  t288_n0: if (E.getNumReferences() >= 3306539008 /*2396.5*/) goto t288_n38;
  t288_n1: if (E.getNumReferences() >= 3306424320 /*2368.5*/) goto t288_n25;
  t288_n2: if (E.getNumReferences() >= 3306391552 /*2360.5*/) goto t288_n6;
  t288_n3: if (E.getNumReferences() >= 3306264576 /*2329.5*/) goto t288_n5;
  t288_n4: return -0.0001065232718246989f;
  t288_n5: return -0.06746812164783478f;
  t288_n6: if (E.getNumReferences() >= 3306414080 /*2366.0*/) goto t288_n24;
  t288_n7: if (E.getNumReferences() >= 3306403840 /*2363.5*/) goto t288_n21;
  t288_n8: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n12;
  t288_n9: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t288_n11;
  t288_n10: return -0.43217960000038147f;
  t288_n11: return -0.10100223124027252f;
  t288_n12: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t288_n20;
  t288_n13: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t288_n17;
  t288_n14: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t288_n16;
  t288_n15: return 0.1226196140050888f;
  t288_n16: return -0.02745954506099224f;
  t288_n17: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t288_n19;
  t288_n18: return -0.25448375940322876f;
  t288_n19: return 0.20263274013996124f;
  t288_n20: return -0.21107201278209686f;
  t288_n21: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t288_n23;
  t288_n22: return 0.1333216279745102f;
  t288_n23: return 0.049389176070690155f;
  t288_n24: return -0.11890516430139542f;
  t288_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t288_n35;
  t288_n26: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t288_n28;
  t288_n27: return -0.08009068667888641f;
  t288_n28: if (E.getFractionNameInContext() >= 3181561674 /*0.07948167622089386*/) goto t288_n32;
  t288_n29: if (E.getNumReferences() >= 3306455040 /*2376.0*/) goto t288_n31;
  t288_n30: return 0.11853805929422379f;
  t288_n31: return -0.2866320013999939f;
  t288_n32: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t288_n34;
  t288_n33: return -1.7412958145141602f;
  t288_n34: return -0.07547356933355331f;
  t288_n35: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t288_n37;
  t288_n36: return 0.1148170605301857f;
  t288_n37: return -0.05448847636580467f;
  t288_n38: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t288_n52;
  t288_n39: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t288_n49;
  t288_n40: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n44;
  t288_n41: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t288_n43;
  t288_n42: return -0.058846160769462585f;
  t288_n43: return 0.04161196947097778f;
  t288_n44: if (E.getNumReferences() >= 3306688512 /*2433.0*/) goto t288_n48;
  t288_n45: if (E.getNumReferences() >= 3306606592 /*2413.0*/) goto t288_n47;
  t288_n46: return -0.04628428444266319f;
  t288_n47: return 0.23673371970653534f;
  t288_n48: return 0.020696060732007027f;
  t288_n49: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t288_n51;
  t288_n50: return 0.0052376133389770985f;
  t288_n51: return 0.10574854165315628f;
  t288_n52: if (E.getNumReferences() >= 3306543104 /*2397.5*/) goto t288_n58;
  t288_n53: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n55;
  t288_n54: return -0.1296103447675705f;
  t288_n55: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t288_n57;
  t288_n56: return 0.11655398458242416f;
  t288_n57: return 0.08375032991170883f;
  t288_n58: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t288_n68;
  t288_n59: if (E.getNumReferences() >= 3306870784 /*2477.5*/) goto t288_n67;
  t288_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t288_n62;
  t288_n61: return 0.06708860397338867f;
  t288_n62: if (E.getNumReferences() >= 3306856448 /*2474.0*/) goto t288_n66;
  t288_n63: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t288_n65;
  t288_n64: return 0.12268704175949097f;
  t288_n65: return 0.09302167594432831f;
  t288_n66: return 0.11258555203676224f;
  t288_n67: return -0.009966242127120495f;
  t288_n68: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t288_n116;
  t288_n69: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n115;
  t288_n70: if (E.getNumReferences() >= 3322564096 /*8851.5*/) goto t288_n100;
  t288_n71: if (E.getNumReferences() >= 3322543616 /*8831.5*/) goto t288_n91;
  t288_n72: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n74;
  t288_n73: return -0.029283693060278893f;
  t288_n74: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t288_n76;
  t288_n75: return -0.04221327602863312f;
  t288_n76: if (E.getNumReferences() >= 3306690560 /*2433.5*/) goto t288_n82;
  t288_n77: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t288_n81;
  t288_n78: if (E.getNumReferences() >= 3306643456 /*2422.0*/) goto t288_n80;
  t288_n79: return 0.10539853572845459f;
  t288_n80: return 0.15108948945999146f;
  t288_n81: return 0.07045062631368637f;
  t288_n82: if (E.getNumReferences() >= 3311417344 /*3587.5*/) goto t288_n84;
  t288_n83: return -0.07318396121263504f;
  t288_n84: if (E.getNumReferences() >= 3321224192 /*7867.5*/) goto t288_n90;
  t288_n85: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t288_n89;
  t288_n86: if (E.getNumReferences() >= 3311550464 /*3620.0*/) goto t288_n88;
  t288_n87: return 0.11801634728908539f;
  t288_n88: return 0.07516475766897202f;
  t288_n89: return -0.019558610394597054f;
  t288_n90: return -0.12955358624458313f;
  t288_n91: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t288_n99;
  t288_n92: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n94;
  t288_n93: return -0.1451665312051773f;
  t288_n94: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t288_n96;
  t288_n95: return -0.11033426225185394f;
  t288_n96: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t288_n98;
  t288_n97: return 0.14251470565795898f;
  t288_n98: return 0.23273968696594238f;
  t288_n99: return 0.17776896059513092f;
  t288_n100: if (E.getNumReferences() >= 3375821112 /*749331.5*/) goto t288_n114;
  t288_n101: if (E.getNumReferences() >= 3369273984 /*432212.0*/) goto t288_n103;
  t288_n102: return -0.03107127919793129f;
  t288_n103: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n111;
  t288_n104: if (E.getFileProximityDistanceCost() >= 3249012736 /*21.0*/) goto t288_n110;
  t288_n105: if (E.getFractionNameInContext() >= 3180443193 /*0.0711483433842659*/) goto t288_n109;
  t288_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n108;
  t288_n107: return -0.23354637622833252f;
  t288_n108: return 0.05436436086893082f;
  t288_n109: return -0.6644459962844849f;
  t288_n110: return -0.552549421787262f;
  t288_n111: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t288_n113;
  t288_n112: return 0.08590350300073624f;
  t288_n113: return -0.010871430858969688f;
  t288_n114: return 0.018654311075806618f;
  t288_n115: return 0.00523970415815711f;
  t288_n116: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t288_n152;
  t288_n117: if (E.getNumReferences() >= 3315294208 /*4972.0*/) goto t288_n151;
  t288_n118: if (E.getNumReferences() >= 3315181568 /*4917.0*/) goto t288_n146;
  t288_n119: if (E.getNumReferences() >= 3314535424 /*4601.5*/) goto t288_n131;
  t288_n120: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t288_n122;
  t288_n121: return 0.012891126796603203f;
  t288_n122: if (E.getNumReferences() >= 3307108352 /*2535.5*/) goto t288_n126;
  t288_n123: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t288_n125;
  t288_n124: return -0.022149287164211273f;
  t288_n125: return 0.1254696398973465f;
  t288_n126: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t288_n128;
  t288_n127: return 0.15096576511859894f;
  t288_n128: if (E.getNumReferences() >= 3314379776 /*4525.5*/) goto t288_n130;
  t288_n129: return 0.027319665998220444f;
  t288_n130: return 0.10079101473093033f;
  t288_n131: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t288_n139;
  t288_n132: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t288_n134;
  t288_n133: return 0.06627727299928665f;
  t288_n134: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t288_n138;
  t288_n135: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t288_n137;
  t288_n136: return 0.11353539675474167f;
  t288_n137: return 0.12773491442203522f;
  t288_n138: return 0.110075443983078f;
  t288_n139: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t288_n141;
  t288_n140: return 0.27360421419143677f;
  t288_n141: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t288_n145;
  t288_n142: if (E.getNumReferences() >= 3314929664 /*4794.0*/) goto t288_n144;
  t288_n143: return 0.01650853455066681f;
  t288_n144: return 0.11177250742912292f;
  t288_n145: return 0.12578043341636658f;
  t288_n146: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t288_n148;
  t288_n147: return 0.10991588979959488f;
  t288_n148: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t288_n150;
  t288_n149: return 0.12260854244232178f;
  t288_n150: return 0.1347561627626419f;
  t288_n151: return -0.0020825418177992105f;
  t288_n152: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t288_n154;
  t288_n153: return 0.10897723585367203f;
  t288_n154: if (E.getNumReferences() >= 3321896960 /*8200.0*/) goto t288_n156;
  t288_n155: return -0.0022441952023655176f;
  t288_n156: if (E.getNumReferences() >= 3321991168 /*8292.0*/) goto t288_n164;
  t288_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t288_n163;
  t288_n158: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t288_n162;
  t288_n159: if (E.getNumReferences() >= 3321944576 /*8246.5*/) goto t288_n161;
  t288_n160: return 0.15641610324382782f;
  t288_n161: return -0.10912522673606873f;
  t288_n162: return -0.11548933386802673f;
  t288_n163: return 0.2226387858390808f;
  t288_n164: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t288_n250;
  t288_n165: if (E.getNumReferences() >= 3362084992 /*234850.0*/) goto t288_n249;
  t288_n166: if (E.getNumReferences() >= 3361153472 /*220295.0*/) goto t288_n244;
  t288_n167: if (E.getNumReferences() >= 3322048000 /*8347.5*/) goto t288_n169;
  t288_n168: return -0.18368320167064667f;
  t288_n169: if (E.getNumReferences() >= 3323590144 /*9853.5*/) goto t288_n197;
  t288_n170: if (E.getNumReferences() >= 3323271680 /*9542.5*/) goto t288_n190;
  t288_n171: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t288_n189;
  t288_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t288_n178;
  t288_n173: if (E.getNumReferences() >= 3322279424 /*8573.5*/) goto t288_n177;
  t288_n174: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t288_n176;
  t288_n175: return 0.09932303428649902f;
  t288_n176: return 0.15849409997463226f;
  t288_n177: return -0.10369367152452469f;
  t288_n178: if (E.getNumReferences() >= 3322983936 /*9261.5*/) goto t288_n180;
  t288_n179: return 0.09508862346410751f;
  t288_n180: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t288_n182;
  t288_n181: return -0.10499357432126999f;
  t288_n182: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t288_n188;
  t288_n183: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t288_n187;
  t288_n184: if (E.getFileProximityDistanceCost() >= 3242196992 /*12.0*/) goto t288_n186;
  t288_n185: return 0.17723728716373444f;
  t288_n186: return 0.1676507145166397f;
  t288_n187: return 0.10530094802379608f;
  t288_n188: return -0.09202835708856583f;
  t288_n189: return -0.008884456939995289f;
  t288_n190: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t288_n194;
  t288_n191: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t288_n193;
  t288_n192: return 0.01409200020134449f;
  t288_n193: return 0.18665342032909393f;
  t288_n194: if (E.getNumReferences() >= 3323522048 /*9787.0*/) goto t288_n196;
  t288_n195: return 0.2001359462738037f;
  t288_n196: return -0.045203059911727905f;
  t288_n197: if (E.getNumReferences() >= 3349519808 /*84795.5*/) goto t288_n233;
  t288_n198: if (E.getNumReferences() >= 3341544320 /*44011.5*/) goto t288_n210;
  t288_n199: if (E.getNumReferences() >= 3339141760 /*34626.5*/) goto t288_n203;
  t288_n200: if (E.getFileProximityDistanceCost() >= 3254517760 /*31.5*/) goto t288_n202;
  t288_n201: return 0.01238479558378458f;
  t288_n202: return 0.15384957194328308f;
  t288_n203: if (E.getNumReferences() >= 3339463424 /*35883.0*/) goto t288_n207;
  t288_n204: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t288_n206;
  t288_n205: return 0.06238013878464699f;
  t288_n206: return 0.11288860440254211f;
  t288_n207: if (E.getNumReferences() >= 3340176384 /*38668.0*/) goto t288_n209;
  t288_n208: return -0.16824868321418762f;
  t288_n209: return 0.03111938387155533f;
  t288_n210: if (E.getNumReferences() >= 3341955328 /*45617.0*/) goto t288_n212;
  t288_n211: return -0.14278589189052582f;
  t288_n212: if (E.getNumReferences() >= 3342302208 /*46972.0*/) goto t288_n226;
  t288_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n217;
  t288_n214: if (E.getNumReferences() >= 3342140032 /*46338.5*/) goto t288_n216;
  t288_n215: return -0.1668405681848526f;
  t288_n216: return 0.06500621140003204f;
  t288_n217: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t288_n219;
  t288_n218: return 0.02812497317790985f;
  t288_n219: if (E.getSemaFileProximityScore() >= 3204448256 /*0.5*/) goto t288_n225;
  t288_n220: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t288_n224;
  t288_n221: if (E.getFileProximityDistanceCost() >= 3241148416 /*11.0*/) goto t288_n223;
  t288_n222: return -0.0469905249774456f;
  t288_n223: return 0.09920187294483185f;
  t288_n224: return -0.04961320012807846f;
  t288_n225: return 0.05633223056793213f;
  t288_n226: if (E.getNumReferences() >= 3342989312 /*49656.0*/) goto t288_n228;
  t288_n227: return -0.1608573943376541f;
  t288_n228: if (E.getNumReferences() >= 3345268096 /*58557.5*/) goto t288_n232;
  t288_n229: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n231;
  t288_n230: return -0.045288845896720886f;
  t288_n231: return 0.04963690787553787f;
  t288_n232: return -0.015273950062692165f;
  t288_n233: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t288_n243;
  t288_n234: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t288_n242;
  t288_n235: if (E.getNumReferences() >= 3352314880 /*106632.0*/) goto t288_n241;
  t288_n236: if (E.getNumReferences() >= 3352087680 /*104857.0*/) goto t288_n240;
  t288_n237: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t288_n239;
  t288_n238: return -0.02191457524895668f;
  t288_n239: return -0.1905660480260849f;
  t288_n240: return -1.3605698347091675f;
  t288_n241: return 0.003862885059788823f;
  t288_n242: return 0.011113401502370834f;
  t288_n243: return 0.027123363688588142f;
  t288_n244: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t288_n246;
  t288_n245: return -0.1677309274673462f;
  t288_n246: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t288_n248;
  t288_n247: return -0.21192941069602966f;
  t288_n248: return 0.09494727104902267f;
  t288_n249: return 0.03827682510018349f;
  t288_n250: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t288_n252;
  t288_n251: return -0.010008485987782478f;
  t288_n252: if (E.getNumReferences() >= 3364059232 /*269251.0*/) goto t288_n254;
  t288_n253: return -0.49215441942214966f;
  t288_n254: return 0.023905357345938683f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree289(const Example& E) {
  t289_n0: if (E.getNumReferences() >= 3284680704 /*400.5*/) goto t289_n48;
  t289_n1: if (E.getNumReferences() >= 3283042304 /*350.5*/) goto t289_n17;
  t289_n2: if (E.getNumReferences() >= 3283009536 /*349.5*/) goto t289_n4;
  t289_n3: return -0.00010970543371513486f;
  t289_n4: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t289_n16;
  t289_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t289_n11;
  t289_n6: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t289_n10;
  t289_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t289_n9;
  t289_n8: return -0.013839487917721272f;
  t289_n9: return 0.38737550377845764f;
  t289_n10: return -0.0032907375134527683f;
  t289_n11: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t289_n13;
  t289_n12: return 0.11506832391023636f;
  t289_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t289_n15;
  t289_n14: return -0.10172344744205475f;
  t289_n15: return 1.12201726436615f;
  t289_n16: return -0.11665608733892441f;
  t289_n17: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t289_n29;
  t289_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t289_n20;
  t289_n19: return -0.023937955498695374f;
  t289_n20: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t289_n24;
  t289_n21: if (E.getNumReferences() >= 3283943424 /*378.0*/) goto t289_n23;
  t289_n22: return 0.018631037324666977f;
  t289_n23: return 0.1182633638381958f;
  t289_n24: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t289_n26;
  t289_n25: return 0.022142842411994934f;
  t289_n26: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t289_n28;
  t289_n27: return 0.1413698047399521f;
  t289_n28: return 0.01888960972428322f;
  t289_n29: if (E.getNumReferences() >= 3283730432 /*371.5*/) goto t289_n39;
  t289_n30: if (E.getNumReferences() >= 3283517440 /*365.0*/) goto t289_n38;
  t289_n31: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t289_n33;
  t289_n32: return 0.10061187297105789f;
  t289_n33: if (E.getNumReferences() >= 3283435520 /*362.5*/) goto t289_n35;
  t289_n34: return -0.14869514107704163f;
  t289_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t289_n37;
  t289_n36: return 0.014549179002642632f;
  t289_n37: return 0.12964874505996704f;
  t289_n38: return -0.3220646381378174f;
  t289_n39: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t289_n43;
  t289_n40: if (E.getNumReferences() >= 3284484096 /*394.5*/) goto t289_n42;
  t289_n41: return 0.11556971818208694f;
  t289_n42: return 0.0337512232363224f;
  t289_n43: if (E.getFractionNameInContext() >= 3187804193 /*0.1269841343164444*/) goto t289_n45;
  t289_n44: return 0.037004999816417694f;
  t289_n45: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t289_n47;
  t289_n46: return 0.12246347218751907f;
  t289_n47: return 0.05957870930433273f;
  t289_n48: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t289_n92;
  t289_n49: if (E.getNumReferences() >= 3284746240 /*402.5*/) goto t289_n55;
  t289_n50: if (E.getFractionNameInContext() >= 3182398201 /*0.08571428805589676*/) goto t289_n54;
  t289_n51: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t289_n53;
  t289_n52: return 0.10938768088817596f;
  t289_n53: return -0.06739600747823715f;
  t289_n54: return 0.13121704757213593f;
  t289_n55: if (E.getNumReferences() >= 3285499904 /*425.5*/) goto t289_n91;
  t289_n56: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t289_n58;
  t289_n57: return -0.33670252561569214f;
  t289_n58: if (E.getNumReferences() >= 3285434368 /*423.5*/) goto t289_n84;
  t289_n59: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t289_n81;
  t289_n60: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t289_n74;
  t289_n61: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t289_n71;
  t289_n62: if (E.getFractionNameInContext() >= 3197457750 /*0.2916666865348816*/) goto t289_n70;
  t289_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t289_n65;
  t289_n64: return -0.009299448691308498f;
  t289_n65: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t289_n67;
  t289_n66: return 0.0045706890523433685f;
  t289_n67: if (E.getNumReferences() >= 3285057536 /*412.0*/) goto t289_n69;
  t289_n68: return 0.22048388421535492f;
  t289_n69: return 0.08883938193321228f;
  t289_n70: return -0.4024444818496704f;
  t289_n71: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t289_n73;
  t289_n72: return -0.29416176676750183f;
  t289_n73: return -0.09660042077302933f;
  t289_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t289_n76;
  t289_n75: return -0.22486236691474915f;
  t289_n76: if (E.getNumReferences() >= 3285336064 /*420.5*/) goto t289_n80;
  t289_n77: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t289_n79;
  t289_n78: return 0.08208002150058746f;
  t289_n79: return 0.11806301027536392f;
  t289_n80: return 0.010423921048641205f;
  t289_n81: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t289_n83;
  t289_n82: return 0.17850558459758759f;
  t289_n83: return -0.08711404353380203f;
  t289_n84: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t289_n90;
  t289_n85: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t289_n87;
  t289_n86: return -0.025980817154049873f;
  t289_n87: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t289_n89;
  t289_n88: return 0.11700215190649033f;
  t289_n89: return -0.1520233154296875f;
  t289_n90: return 0.3872886002063751f;
  t289_n91: return 0.006160333286970854f;
  t289_n92: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t289_n254;
  t289_n93: if (E.getNumReferences() >= 3288473600 /*520.5*/) goto t289_n167;
  t289_n94: if (E.getNumReferences() >= 3288391680 /*515.5*/) goto t289_n144;
  t289_n95: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t289_n131;
  t289_n96: if (E.getFileProximityDistanceCost() >= 3253469184 /*29.5*/) goto t289_n130;
  t289_n97: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t289_n117;
  t289_n98: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t289_n100;
  t289_n99: return -0.000949656474404037f;
  t289_n100: if (E.getNumReferences() >= 3284746240 /*402.5*/) goto t289_n104;
  t289_n101: if (E.getNumReferences() >= 3284713472 /*401.5*/) goto t289_n103;
  t289_n102: return -0.10076640546321869f;
  t289_n103: return 0.10420448333024979f;
  t289_n104: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t289_n106;
  t289_n105: return 0.027380799874663353f;
  t289_n106: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t289_n112;
  t289_n107: if (E.getNumReferences() >= 3286237184 /*448.0*/) goto t289_n111;
  t289_n108: if (E.getNumReferences() >= 3285008384 /*410.5*/) goto t289_n110;
  t289_n109: return -0.10029690712690353f;
  t289_n110: return 0.8630414605140686f;
  t289_n111: return -0.10079535096883774f;
  t289_n112: if (E.getNumReferences() >= 3285106688 /*413.5*/) goto t289_n116;
  t289_n113: if (E.getNumReferences() >= 3284910080 /*407.5*/) goto t289_n115;
  t289_n114: return 0.026734132319688797f;
  t289_n115: return 0.637587308883667f;
  t289_n116: return 0.007525804452598095f;
  t289_n117: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t289_n127;
  t289_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t289_n120;
  t289_n119: return 0.03782613202929497f;
  t289_n120: if (E.getNumReferences() >= 3286630400 /*460.0*/) goto t289_n126;
  t289_n121: if (E.getNumReferences() >= 3286302720 /*450.0*/) goto t289_n125;
  t289_n122: if (E.getFractionNameInContext() >= 3184451170 /*0.10101009905338287*/) goto t289_n124;
  t289_n123: return -0.2597801685333252f;
  t289_n124: return 0.29359865188598633f;
  t289_n125: return 0.3232956826686859f;
  t289_n126: return -0.3137515187263489f;
  t289_n127: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t289_n129;
  t289_n128: return 0.009609909728169441f;
  t289_n129: return 0.15904514491558075f;
  t289_n130: return 0.22740663588047028f;
  t289_n131: if (E.getNumReferences() >= 3288252416 /*509.5*/) goto t289_n137;
  t289_n132: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t289_n134;
  t289_n133: return 0.049396611750125885f;
  t289_n134: if (E.getNumReferences() >= 3286745088 /*463.5*/) goto t289_n136;
  t289_n135: return 0.016392644494771957f;
  t289_n136: return 0.13487206399440765f;
  t289_n137: if (E.getNumReferences() >= 3288334336 /*512.0*/) goto t289_n143;
  t289_n138: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t289_n142;
  t289_n139: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t289_n141;
  t289_n140: return 0.34508296847343445f;
  t289_n141: return 0.008547209203243256f;
  t289_n142: return 0.13900914788246155f;
  t289_n143: return 0.05087195336818695f;
  t289_n144: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t289_n150;
  t289_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t289_n149;
  t289_n146: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t289_n148;
  t289_n147: return 0.09103676676750183f;
  t289_n148: return -0.2997075319290161f;
  t289_n149: return 0.3348950445652008f;
  t289_n150: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t289_n152;
  t289_n151: return 0.050010599195957184f;
  t289_n152: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t289_n156;
  t289_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel))) goto t289_n155;
  t289_n154: return -0.012609558179974556f;
  t289_n155: return 0.18032318353652954f;
  t289_n156: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t289_n166;
  t289_n157: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t289_n163;
  t289_n158: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t289_n162;
  t289_n159: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t289_n161;
  t289_n160: return 0.11245977133512497f;
  t289_n161: return 0.04390859231352806f;
  t289_n162: return 0.12372361123561859f;
  t289_n163: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t289_n165;
  t289_n164: return -0.11122173815965652f;
  t289_n165: return 1.3732354640960693f;
  t289_n166: return 0.016651274636387825f;
  t289_n167: if (E.getNumReferences() >= 3302502400 /*1729.5*/) goto t289_n243;
  t289_n168: if (E.getNumReferences() >= 3302428672 /*1720.5*/) goto t289_n238;
  t289_n169: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t289_n171;
  t289_n170: return -0.01235773041844368f;
  t289_n171: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t289_n175;
  t289_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t289_n174;
  t289_n173: return -0.18533658981323242f;
  t289_n174: return 0.02680942602455616f;
  t289_n175: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t289_n185;
  t289_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t289_n178;
  t289_n177: return -0.008608760312199593f;
  t289_n178: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t289_n180;
  t289_n179: return 0.09008995443582535f;
  t289_n180: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t289_n182;
  t289_n181: return 0.06431320309638977f;
  t289_n182: if (E.getNumReferences() >= 3293855744 /*849.0*/) goto t289_n184;
  t289_n183: return 0.11499398201704025f;
  t289_n184: return 0.1351213902235031f;
  t289_n185: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t289_n229;
  t289_n186: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t289_n202;
  t289_n187: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t289_n201;
  t289_n188: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t289_n196;
  t289_n189: if (E.getNumReferences() >= 3299430400 /*1354.5*/) goto t289_n193;
  t289_n190: if (E.getNumReferences() >= 3296788480 /*1032.0*/) goto t289_n192;
  t289_n191: return -0.10422463715076447f;
  t289_n192: return -0.3551044762134552f;
  t289_n193: if (E.getNumReferences() >= 3300528128 /*1488.5*/) goto t289_n195;
  t289_n194: return 0.120048888027668f;
  t289_n195: return -0.26196372509002686f;
  t289_n196: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t289_n198;
  t289_n197: return 0.06149536743760109f;
  t289_n198: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t289_n200;
  t289_n199: return -0.1783074587583542f;
  t289_n200: return 0.08937860280275345f;
  t289_n201: return 0.04399266466498375f;
  t289_n202: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t289_n216;
  t289_n203: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t289_n211;
  t289_n204: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t289_n206;
  t289_n205: return 0.0520903654396534f;
  t289_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t289_n208;
  t289_n207: return -0.04713503643870354f;
  t289_n208: if (E.getNumReferences() >= 3295109120 /*925.5*/) goto t289_n210;
  t289_n209: return 0.10715515911579132f;
  t289_n210: return 0.10826091468334198f;
  t289_n211: if (E.getNumReferences() >= 3302039552 /*1673.0*/) goto t289_n213;
  t289_n212: return 0.04401693493127823f;
  t289_n213: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t289_n215;
  t289_n214: return -0.8385810852050781f;
  t289_n215: return -0.0639258399605751f;
  t289_n216: if (E.getSymbolScopeDistanceCost() >= 3241148416 /*11.0*/) goto t289_n218;
  t289_n217: return -0.03812374919652939f;
  t289_n218: if (E.getNumReferences() >= 3289071616 /*557.0*/) goto t289_n222;
  t289_n219: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t289_n221;
  t289_n220: return -0.14479847252368927f;
  t289_n221: return 0.10047905892133713f;
  t289_n222: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t289_n224;
  t289_n223: return 0.023806890472769737f;
  t289_n224: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t289_n228;
  t289_n225: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t289_n227;
  t289_n226: return 0.12477510422468185f;
  t289_n227: return -0.10556032508611679f;
  t289_n228: return 0.04781821742653847f;
  t289_n229: if (E.getNumReferences() >= 3292618752 /*773.5*/) goto t289_n237;
  t289_n230: if (E.getNumReferences() >= 3291545600 /*708.0*/) goto t289_n236;
  t289_n231: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t289_n233;
  t289_n232: return 0.104281485080719f;
  t289_n233: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t289_n235;
  t289_n234: return -0.3363729417324066f;
  t289_n235: return 0.08877469599246979f;
  t289_n236: return 0.11952941119670868f;
  t289_n237: return 0.023642759770154953f;
  t289_n238: if (E.getNumReferences() >= 3302465536 /*1725.0*/) goto t289_n242;
  t289_n239: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t289_n241;
  t289_n240: return -0.6841866970062256f;
  t289_n241: return -0.1440725177526474f;
  t289_n242: return -0.03302855044603348f;
  t289_n243: if (E.getNumReferences() >= 3302584320 /*1739.5*/) goto t289_n253;
  t289_n244: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t289_n246;
  t289_n245: return 0.1862826645374298f;
  t289_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t289_n250;
  t289_n247: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t289_n249;
  t289_n248: return -0.6737316250801086f;
  t289_n249: return 0.04977458342909813f;
  t289_n250: if (E.getNumReferences() >= 3302551552 /*1735.5*/) goto t289_n252;
  t289_n251: return -0.16222655773162842f;
  t289_n252: return 0.13004910945892334f;
  t289_n253: return 0.0008853296749293804f;
  t289_n254: return 0.12837475538253784f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree290(const Example& E) {
  t290_n0: if (E.getNumReferences() >= 3364464848 /*281926.5*/) goto t290_n198;
  t290_n1: if (E.getNumReferences() >= 3362537216 /*241916.0*/) goto t290_n177;
  t290_n2: if (E.getNumReferences() >= 3359104096 /*188273.5*/) goto t290_n134;
  t290_n3: if (E.getNumReferences() >= 3358365376 /*176731.0*/) goto t290_n133;
  t290_n4: if (E.getNumReferences() >= 3357792448 /*167779.0*/) goto t290_n90;
  t290_n5: if (E.getNumReferences() >= 3356972160 /*154962.0*/) goto t290_n89;
  t290_n6: if (E.getNumReferences() >= 3351031040 /*96602.0*/) goto t290_n28;
  t290_n7: if (E.getNumReferences() >= 3350789120 /*94712.0*/) goto t290_n25;
  t290_n8: if (E.getNumReferences() >= 3350665472 /*93746.0*/) goto t290_n14;
  t290_n9: if (E.getNumReferences() >= 3350538496 /*92754.0*/) goto t290_n11;
  t290_n10: return -2.647299515956547e-05f;
  t290_n11: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t290_n13;
  t290_n12: return -0.14650864899158478f;
  t290_n13: return -0.3005571961402893f;
  t290_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n18;
  t290_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n17;
  t290_n16: return -0.17759619653224945f;
  t290_n17: return -0.04701351001858711f;
  t290_n18: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t290_n22;
  t290_n19: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n21;
  t290_n20: return 0.0763019397854805f;
  t290_n21: return 0.16390511393547058f;
  t290_n22: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t290_n24;
  t290_n23: return 0.06903241574764252f;
  t290_n24: return -0.15768857300281525f;
  t290_n25: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t290_n27;
  t290_n26: return -0.136236771941185f;
  t290_n27: return -0.2329026460647583f;
  t290_n28: if (E.getNumReferences() >= 3351981568 /*104028.0*/) goto t290_n48;
  t290_n29: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t290_n33;
  t290_n30: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n32;
  t290_n31: return -0.07054077833890915f;
  t290_n32: return 0.08789926022291183f;
  t290_n33: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t290_n35;
  t290_n34: return -0.0953838974237442f;
  t290_n35: if (E.getNumReferences() >= 3351094208 /*97095.5*/) goto t290_n39;
  t290_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n38;
  t290_n37: return -0.1623714119195938f;
  t290_n38: return 0.07810285687446594f;
  t290_n39: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n45;
  t290_n40: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t290_n42;
  t290_n41: return -0.035577695816755295f;
  t290_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t290_n44;
  t290_n43: return 0.032139670103788376f;
  t290_n44: return 0.0704089030623436f;
  t290_n45: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t290_n47;
  t290_n46: return -0.030925245955586433f;
  t290_n47: return -0.25641778111457825f;
  t290_n48: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t290_n82;
  t290_n49: if (E.getNumReferences() >= 3353778624 /*118067.5*/) goto t290_n65;
  t290_n50: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t290_n52;
  t290_n51: return -0.05488481745123863f;
  t290_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n64;
  t290_n53: if (E.getNumReferences() >= 3352895680 /*111169.5*/) goto t290_n57;
  t290_n54: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n56;
  t290_n55: return -0.21622911095619202f;
  t290_n56: return 0.009848738089203835f;
  t290_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t290_n59;
  t290_n58: return 0.03239225223660469f;
  t290_n59: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t290_n61;
  t290_n60: return 0.1067977100610733f;
  t290_n61: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t290_n63;
  t290_n62: return -0.07616501301527023f;
  t290_n63: return 0.1003953292965889f;
  t290_n64: return -0.12122160196304321f;
  t290_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_TopLevel))) goto t290_n75;
  t290_n66: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t290_n74;
  t290_n67: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t290_n69;
  t290_n68: return 0.007915705442428589f;
  t290_n69: if (E.getNumReferences() >= 3354468032 /*123453.5*/) goto t290_n73;
  t290_n70: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t290_n72;
  t290_n71: return 0.0880635604262352f;
  t290_n72: return -0.14130137860774994f;
  t290_n73: return -0.020606009289622307f;
  t290_n74: return -0.018165621906518936f;
  t290_n75: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t290_n77;
  t290_n76: return -0.02259795181453228f;
  t290_n77: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t290_n81;
  t290_n78: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t290_n80;
  t290_n79: return -0.12088847160339355f;
  t290_n80: return 0.11581224203109741f;
  t290_n81: return 0.03801605477929115f;
  t290_n82: if (E.getNumReferences() >= 3353810176 /*118314.0*/) goto t290_n88;
  t290_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t290_n85;
  t290_n84: return 0.03892025724053383f;
  t290_n85: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t290_n87;
  t290_n86: return 0.13032416999340057f;
  t290_n87: return 0.07534519582986832f;
  t290_n88: return -0.01746280863881111f;
  t290_n89: return -0.1535951793193817f;
  t290_n90: if (E.getNumReferences() >= 3357836960 /*168474.5*/) goto t290_n118;
  t290_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t290_n97;
  t290_n92: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t290_n94;
  t290_n93: return -0.013960087671875954f;
  t290_n94: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n96;
  t290_n95: return 0.1198827251791954f;
  t290_n96: return 0.01054875273257494f;
  t290_n97: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t290_n103;
  t290_n98: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t290_n100;
  t290_n99: return -0.04299873113632202f;
  t290_n100: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n102;
  t290_n101: return 0.20148178935050964f;
  t290_n102: return -0.12873734533786774f;
  t290_n103: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t290_n113;
  t290_n104: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t290_n108;
  t290_n105: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n107;
  t290_n106: return -0.024576162919402122f;
  t290_n107: return 0.07668347656726837f;
  t290_n108: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t290_n110;
  t290_n109: return 0.12248382717370987f;
  t290_n110: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t290_n112;
  t290_n111: return 0.13557124137878418f;
  t290_n112: return 0.1309027373790741f;
  t290_n113: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n115;
  t290_n114: return 0.08199143409729004f;
  t290_n115: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t290_n117;
  t290_n116: return 0.03316781297326088f;
  t290_n117: return -0.26550793647766113f;
  t290_n118: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n120;
  t290_n119: return -0.1631610095500946f;
  t290_n120: if (E.getNumReferences() >= 3358250528 /*174936.5*/) goto t290_n122;
  t290_n121: return -0.14222627878189087f;
  t290_n122: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t290_n124;
  t290_n123: return 0.09944237023591995f;
  t290_n124: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t290_n130;
  t290_n125: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t290_n129;
  t290_n126: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t290_n128;
  t290_n127: return -0.006898452993482351f;
  t290_n128: return 0.23320621252059937f;
  t290_n129: return -0.11105549335479736f;
  t290_n130: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t290_n132;
  t290_n131: return 0.19685117900371552f;
  t290_n132: return -0.11676454544067383f;
  t290_n133: return -0.1624746173620224f;
  t290_n134: if (E.getNumReferences() >= 3360215360 /*205637.0*/) goto t290_n152;
  t290_n135: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t290_n139;
  t290_n136: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t290_n138;
  t290_n137: return 7.649235158169176e-06f;
  t290_n138: return -0.17195433378219604f;
  t290_n139: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t290_n143;
  t290_n140: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t290_n142;
  t290_n141: return -0.2284064143896103f;
  t290_n142: return 0.08588682115077972f;
  t290_n143: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t290_n151;
  t290_n144: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t290_n146;
  t290_n145: return -0.062398891896009445f;
  t290_n146: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t290_n148;
  t290_n147: return 0.09107688814401627f;
  t290_n148: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t290_n150;
  t290_n149: return -0.053720466792583466f;
  t290_n150: return 0.05256890133023262f;
  t290_n151: return 0.063362255692482f;
  t290_n152: if (E.getFractionNameInContext() >= 3178314303 /*0.058893438428640366*/) goto t290_n156;
  t290_n153: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t290_n155;
  t290_n154: return -0.020782532170414925f;
  t290_n155: return 0.006369562353938818f;
  t290_n156: if (E.getFractionNameInContext() >= 3180161239 /*0.06904762238264084*/) goto t290_n172;
  t290_n157: if (E.getNumReferences() >= 3360334912 /*207505.0*/) goto t290_n165;
  t290_n158: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t290_n164;
  t290_n159: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t290_n161;
  t290_n160: return 0.12906064093112946f;
  t290_n161: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t290_n163;
  t290_n162: return 0.054198555648326874f;
  t290_n163: return 0.11456799507141113f;
  t290_n164: return 0.02370517887175083f;
  t290_n165: if (E.getNumReferences() >= 3362180128 /*236336.5*/) goto t290_n167;
  t290_n166: return -0.1064625084400177f;
  t290_n167: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_TopLevel))) goto t290_n169;
  t290_n168: return 0.06301087141036987f;
  t290_n169: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t290_n171;
  t290_n170: return 0.0798070877790451f;
  t290_n171: return 0.14986813068389893f;
  t290_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Type))) goto t290_n176;
  t290_n173: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t290_n175;
  t290_n174: return -0.014854884706437588f;
  t290_n175: return -0.1426474004983902f;
  t290_n176: return 0.06888514012098312f;
  t290_n177: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t290_n183;
  t290_n178: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n182;
  t290_n179: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n181;
  t290_n180: return -0.3139163553714752f;
  t290_n181: return -0.005505293142050505f;
  t290_n182: return -0.35709208250045776f;
  t290_n183: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t290_n197;
  t290_n184: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t290_n186;
  t290_n185: return -0.08496607840061188f;
  t290_n186: if (E.getNumReferences() >= 3363672576 /*259656.0*/) goto t290_n192;
  t290_n187: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t290_n191;
  t290_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n190;
  t290_n189: return -0.11369666457176208f;
  t290_n190: return 0.18601109087467194f;
  t290_n191: return -0.13869446516036987f;
  t290_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t290_n194;
  t290_n193: return -0.050369199365377426f;
  t290_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n196;
  t290_n195: return 0.02574242278933525f;
  t290_n196: return 0.07166694849729538f;
  t290_n197: return 0.08089202642440796f;
  t290_n198: if (E.getNumReferences() >= 3364942528 /*296854.0*/) goto t290_n252;
  t290_n199: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_TopLevel))) goto t290_n221;
  t290_n200: if (E.getNumReferences() >= 3364686096 /*288840.5*/) goto t290_n206;
  t290_n201: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t290_n205;
  t290_n202: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t290_n204;
  t290_n203: return 0.10063745081424713f;
  t290_n204: return 0.0015706701669842005f;
  t290_n205: return 0.058304499834775925f;
  t290_n206: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t290_n214;
  t290_n207: if (E.getFileProximityDistanceCost() >= 3239051264 /*9.0*/) goto t290_n209;
  t290_n208: return 0.07043451070785522f;
  t290_n209: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t290_n211;
  t290_n210: return -1.51956045627594f;
  t290_n211: if (E.getFileProximityDistanceCost() >= 3245342720 /*15.0*/) goto t290_n213;
  t290_n212: return -0.31268030405044556f;
  t290_n213: return 0.12767192721366882f;
  t290_n214: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t290_n220;
  t290_n215: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t290_n217;
  t290_n216: return -0.06945838034152985f;
  t290_n217: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t290_n219;
  t290_n218: return 0.09432502090930939f;
  t290_n219: return 0.1679435521364212f;
  t290_n220: return -0.014831248670816422f;
  t290_n221: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t290_n239;
  t290_n222: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n230;
  t290_n223: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t290_n229;
  t290_n224: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t290_n228;
  t290_n225: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t290_n227;
  t290_n226: return -0.014151029288768768f;
  t290_n227: return 0.09859632700681686f;
  t290_n228: return 0.0359698086977005f;
  t290_n229: return 0.13269709050655365f;
  t290_n230: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n232;
  t290_n231: return -0.324260413646698f;
  t290_n232: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t290_n234;
  t290_n233: return 0.06495919078588486f;
  t290_n234: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t290_n236;
  t290_n235: return 0.11076274514198303f;
  t290_n236: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t290_n238;
  t290_n237: return -0.6528893113136292f;
  t290_n238: return 0.07075265794992447f;
  t290_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t290_n249;
  t290_n240: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t290_n244;
  t290_n241: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n243;
  t290_n242: return -0.053283508867025375f;
  t290_n243: return 0.08476655185222626f;
  t290_n244: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t290_n246;
  t290_n245: return -0.09621774405241013f;
  t290_n246: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t290_n248;
  t290_n247: return -0.1783163845539093f;
  t290_n248: return -0.5421934127807617f;
  t290_n249: if (E.getFractionNameInContext() >= 3180849535 /*0.07417582720518112*/) goto t290_n251;
  t290_n250: return -0.10432043671607971f;
  t290_n251: return 0.06873972713947296f;
  t290_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type))) goto t290_n254;
  t290_n253: return -0.004507748410105705f;
  t290_n254: return 0.031295426189899445f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree291(const Example& E) {
  t291_n0: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t291_n2;
  t291_n1: return -0.00017877877689898014f;
  t291_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t291_n12;
  t291_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t291_n11;
  t291_n4: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t291_n10;
  t291_n5: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t291_n7;
  t291_n6: return 0.004768875427544117f;
  t291_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t291_n9;
  t291_n8: return 0.06251801550388336f;
  t291_n9: return 0.18485905230045319f;
  t291_n10: return -0.0878027006983757f;
  t291_n11: return 0.1045546755194664f;
  t291_n12: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t291_n238;
  t291_n13: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t291_n55;
  t291_n14: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t291_n52;
  t291_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t291_n51;
  t291_n16: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t291_n50;
  t291_n17: if (E.getNumReferences() >= 3261202432 /*56.5*/) goto t291_n19;
  t291_n18: return -0.013184066861867905f;
  t291_n19: if (E.getNumReferences() >= 3261464576 /*57.5*/) goto t291_n49;
  t291_n20: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t291_n22;
  t291_n21: return 0.007488412316888571f;
  t291_n22: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t291_n48;
  t291_n23: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t291_n25;
  t291_n24: return -0.10942725092172623f;
  t291_n25: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t291_n31;
  t291_n26: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t291_n28;
  t291_n27: return -0.1410246342420578f;
  t291_n28: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t291_n30;
  t291_n29: return 0.08494766801595688f;
  t291_n30: return 0.15737693011760712f;
  t291_n31: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t291_n33;
  t291_n32: return 0.1590823531150818f;
  t291_n33: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t291_n35;
  t291_n34: return -0.10041482746601105f;
  t291_n35: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t291_n37;
  t291_n36: return 0.18287205696105957f;
  t291_n37: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t291_n39;
  t291_n38: return -0.018175886943936348f;
  t291_n39: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t291_n41;
  t291_n40: return 0.1981084793806076f;
  t291_n41: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t291_n45;
  t291_n42: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t291_n44;
  t291_n43: return -0.10020150989294052f;
  t291_n44: return 0.2282976508140564f;
  t291_n45: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t291_n47;
  t291_n46: return 0.21068528294563293f;
  t291_n47: return 0.145061656832695f;
  t291_n48: return -0.05038004741072655f;
  t291_n49: return -0.002633839612826705f;
  t291_n50: return -0.015082494355738163f;
  t291_n51: return 0.006773913279175758f;
  t291_n52: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement))) goto t291_n54;
  t291_n53: return 0.09386081248521805f;
  t291_n54: return 0.14740298688411713f;
  t291_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t291_n147;
  t291_n56: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t291_n104;
  t291_n57: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t291_n59;
  t291_n58: return -0.0313531793653965f;
  t291_n59: if (E.getNumReferences() >= 3304873984 /*2019.0*/) goto t291_n103;
  t291_n60: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t291_n62;
  t291_n61: return -0.07175672799348831f;
  t291_n62: if (E.getNumReferences() >= 3282321408 /*328.5*/) goto t291_n98;
  t291_n63: if (E.getNumReferences() >= 3280896000 /*285.0*/) goto t291_n97;
  t291_n64: if (E.getNumReferences() >= 3257270272 /*41.5*/) goto t291_n66;
  t291_n65: return -0.1801603138446808f;
  t291_n66: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t291_n96;
  t291_n67: if (E.getNumReferences() >= 3264675840 /*75.5*/) goto t291_n77;
  t291_n68: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t291_n74;
  t291_n69: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t291_n71;
  t291_n70: return 0.10337836295366287f;
  t291_n71: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t291_n73;
  t291_n72: return 0.1650341898202896f;
  t291_n73: return -0.07287029176950455f;
  t291_n74: if (E.getFractionNameInContext() >= 3179282432 /*0.0625*/) goto t291_n76;
  t291_n75: return 0.16155420243740082f;
  t291_n76: return 0.2901301383972168f;
  t291_n77: if (E.getNumReferences() >= 3271426048 /*127.0*/) goto t291_n79;
  t291_n78: return -0.03335162252187729f;
  t291_n79: if (E.getNumReferences() >= 3272835072 /*147.5*/) goto t291_n83;
  t291_n80: if (E.getNumReferences() >= 3272507392 /*142.5*/) goto t291_n82;
  t291_n81: return 0.0854564756155014f;
  t291_n82: return 0.14137662947177887f;
  t291_n83: if (E.getNumReferences() >= 3275128832 /*182.5*/) goto t291_n87;
  t291_n84: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t291_n86;
  t291_n85: return 0.10010304301977158f;
  t291_n86: return -0.2525129020214081f;
  t291_n87: if (E.getNumReferences() >= 3278503936 /*234.0*/) goto t291_n95;
  t291_n88: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t291_n94;
  t291_n89: if (E.getNumReferences() >= 3277619200 /*220.5*/) goto t291_n91;
  t291_n90: return -0.13970310986042023f;
  t291_n91: if (E.getNumReferences() >= 3277881344 /*224.5*/) goto t291_n93;
  t291_n92: return 0.18363142013549805f;
  t291_n93: return 0.11395042389631271f;
  t291_n94: return 0.20990408957004547f;
  t291_n95: return 0.07170310616493225f;
  t291_n96: return -0.12379911541938782f;
  t291_n97: return -0.4552386999130249f;
  t291_n98: if (E.getNumReferences() >= 3284074496 /*382.0*/) goto t291_n102;
  t291_n99: if (E.getNumReferences() >= 3282976768 /*348.5*/) goto t291_n101;
  t291_n100: return 0.12257234007120132f;
  t291_n101: return 0.1361626237630844f;
  t291_n102: return 0.11840621381998062f;
  t291_n103: return -0.09730472415685654f;
  t291_n104: if (E.getNumReferences() >= 3349481344 /*84495.0*/) goto t291_n146;
  t291_n105: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t291_n145;
  t291_n106: if (E.getNumReferences() >= 3318637568 /*6604.5*/) goto t291_n116;
  t291_n107: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t291_n109;
  t291_n108: return -0.026773614808917046f;
  t291_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t291_n113;
  t291_n110: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t291_n112;
  t291_n111: return -0.14085322618484497f;
  t291_n112: return 0.085993193089962f;
  t291_n113: if (E.getFractionNameInContext() >= 3190467243 /*0.1666666716337204*/) goto t291_n115;
  t291_n114: return -0.1082424744963646f;
  t291_n115: return -0.9428240656852722f;
  t291_n116: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t291_n140;
  t291_n117: if (E.getNumReferences() >= 3319820288 /*7182.0*/) goto t291_n123;
  t291_n118: if (E.getFractionNameInContext() >= 3183197116 /*0.09166666865348816*/) goto t291_n122;
  t291_n119: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t291_n121;
  t291_n120: return -0.016072113066911697f;
  t291_n121: return 0.0924689769744873f;
  t291_n122: return 0.13501079380512238f;
  t291_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t291_n131;
  t291_n124: if (E.getNumReferences() >= 3348232704 /*74740.0*/) goto t291_n130;
  t291_n125: if (E.getNumReferences() >= 3324483584 /*10726.0*/) goto t291_n129;
  t291_n126: if (E.getNumReferences() >= 3324265472 /*10513.0*/) goto t291_n128;
  t291_n127: return 0.04501963034272194f;
  t291_n128: return 0.17442716658115387f;
  t291_n129: return -0.04163006320595741f;
  t291_n130: return 0.07161418348550797f;
  t291_n131: if (E.getNumReferences() >= 3341092992 /*42248.5*/) goto t291_n139;
  t291_n132: if (E.getNumReferences() >= 3325240832 /*11465.5*/) goto t291_n136;
  t291_n133: if (E.getNumReferences() >= 3324701184 /*10938.5*/) goto t291_n135;
  t291_n134: return -0.23536594212055206f;
  t291_n135: return -0.8431922793388367f;
  t291_n136: if (E.getNumReferences() >= 3325425152 /*11645.5*/) goto t291_n138;
  t291_n137: return 0.1118076741695404f;
  t291_n138: return 0.06158895790576935f;
  t291_n139: return -0.1955985277891159f;
  t291_n140: if (E.getNumReferences() >= 3324282368 /*10529.5*/) goto t291_n142;
  t291_n141: return 0.10875492542982101f;
  t291_n142: if (E.getNumReferences() >= 3333771776 /*23209.0*/) goto t291_n144;
  t291_n143: return 0.09703059494495392f;
  t291_n144: return 0.1048809289932251f;
  t291_n145: return -0.009935219772160053f;
  t291_n146: return -0.11358609795570374f;
  t291_n147: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t291_n233;
  t291_n148: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t291_n150;
  t291_n149: return -0.02406109869480133f;
  t291_n150: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t291_n198;
  t291_n151: if (E.getNumReferences() >= 3331642112 /*19049.5*/) goto t291_n187;
  t291_n152: if (E.getNumReferences() >= 3291111424 /*681.5*/) goto t291_n184;
  t291_n153: if (E.getNumReferences() >= 3288039424 /*503.0*/) goto t291_n173;
  t291_n154: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t291_n166;
  t291_n155: if (E.getNumReferences() >= 3280224256 /*264.5*/) goto t291_n157;
  t291_n156: return 0.03519279137253761f;
  t291_n157: if (E.getNumReferences() >= 3280502784 /*273.0*/) goto t291_n159;
  t291_n158: return 0.16471584141254425f;
  t291_n159: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t291_n161;
  t291_n160: return 0.06127307564020157f;
  t291_n161: if (E.getNumReferences() >= 3285188608 /*416.0*/) goto t291_n165;
  t291_n162: if (E.getNumReferences() >= 3281436672 /*301.5*/) goto t291_n164;
  t291_n163: return -0.09325316548347473f;
  t291_n164: return 0.17573629319667816f;
  t291_n165: return -0.12471283227205276f;
  t291_n166: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t291_n172;
  t291_n167: if (E.getNumReferences() >= 3275718656 /*191.5*/) goto t291_n169;
  t291_n168: return 0.011716785840690136f;
  t291_n169: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t291_n171;
  t291_n170: return 0.013813690282404423f;
  t291_n171: return 0.1650196760892868f;
  t291_n172: return 0.03511384129524231f;
  t291_n173: if (E.getNumReferences() >= 3288416256 /*517.0*/) goto t291_n175;
  t291_n174: return 0.2386646270751953f;
  t291_n175: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t291_n183;
  t291_n176: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t291_n180;
  t291_n177: if (E.getNumReferences() >= 3290980352 /*673.5*/) goto t291_n179;
  t291_n178: return -0.12227217853069305f;
  t291_n179: return 0.21089406311511993f;
  t291_n180: if (E.getNumReferences() >= 3289481216 /*582.0*/) goto t291_n182;
  t291_n181: return 0.14653199911117554f;
  t291_n182: return 0.3205339014530182f;
  t291_n183: return -0.011962625198066235f;
  t291_n184: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t291_n186;
  t291_n185: return -0.09934641420841217f;
  t291_n186: return 0.007917309179902077f;
  t291_n187: if (E.getNumReferences() >= 3332281088 /*20297.5*/) goto t291_n195;
  t291_n188: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t291_n194;
  t291_n189: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t291_n193;
  t291_n190: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t291_n192;
  t291_n191: return 0.08672924339771271f;
  t291_n192: return 0.14248086512088776f;
  t291_n193: return 0.09138517081737518f;
  t291_n194: return 0.029374822974205017f;
  t291_n195: if (E.getNumReferences() >= 3352190016 /*105656.5*/) goto t291_n197;
  t291_n196: return -0.1760711967945099f;
  t291_n197: return 0.060098662972450256f;
  t291_n198: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t291_n232;
  t291_n199: if (E.getNumReferences() >= 3271360512 /*126.5*/) goto t291_n209;
  t291_n200: if (E.getFractionNameInContext() >= 3209061990 /*0.7749999761581421*/) goto t291_n202;
  t291_n201: return 0.05327622964978218f;
  t291_n202: if (E.getNumReferences() >= 3256483840 /*38.5*/) goto t291_n204;
  t291_n203: return 0.34869569540023804f;
  t291_n204: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t291_n206;
  t291_n205: return -0.060394518077373505f;
  t291_n206: if (E.getNumReferences() >= 3268935680 /*108.0*/) goto t291_n208;
  t291_n207: return 0.16991069912910461f;
  t291_n208: return -0.10042856633663177f;
  t291_n209: if (E.getNumReferences() >= 3271720960 /*130.5*/) goto t291_n211;
  t291_n210: return 0.16964471340179443f;
  t291_n211: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t291_n231;
  t291_n212: if (E.getNumReferences() >= 3281125376 /*292.0*/) goto t291_n220;
  t291_n213: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t291_n215;
  t291_n214: return 0.09640556573867798f;
  t291_n215: if (E.getNumReferences() >= 3272245248 /*138.5*/) goto t291_n219;
  t291_n216: if (E.getFractionNameInContext() >= 3201652054 /*0.4166666865348816*/) goto t291_n218;
  t291_n217: return -0.12167645245790482f;
  t291_n218: return 0.1992049217224121f;
  t291_n219: return -0.11959526687860489f;
  t291_n220: if (E.getNumReferences() >= 3282092032 /*321.5*/) goto t291_n222;
  t291_n221: return 0.1298217475414276f;
  t291_n222: if (E.getNumReferences() >= 3294986240 /*918.0*/) goto t291_n230;
  t291_n223: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t291_n225;
  t291_n224: return 0.08938392996788025f;
  t291_n225: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t291_n229;
  t291_n226: if (E.getFractionNameInContext() >= 3194381926 /*0.22499999403953552*/) goto t291_n228;
  t291_n227: return 0.11879018694162369f;
  t291_n228: return 0.21867568790912628f;
  t291_n229: return 0.1065666526556015f;
  t291_n230: return 0.02484084852039814f;
  t291_n231: return -0.05902503803372383f;
  t291_n232: return -0.009831731207668781f;
  t291_n233: if (E.getNumReferences() >= 3357594080 /*164679.5*/) goto t291_n235;
  t291_n234: return 0.0005075083463452756f;
  t291_n235: if (E.getNumReferences() >= 3359209472 /*189920.0*/) goto t291_n237;
  t291_n236: return 0.20942412316799164f;
  t291_n237: return 0.0209259781986475f;
  t291_n238: if (E.getSymbolScopeDistanceCost() >= 3250061312 /*23.0*/) goto t291_n254;
  t291_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_Type))) goto t291_n247;
  t291_n240: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t291_n246;
  t291_n241: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t291_n243;
  t291_n242: return -0.06694477051496506f;
  t291_n243: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t291_n245;
  t291_n244: return 0.23276428878307343f;
  t291_n245: return 0.040212482213974f;
  t291_n246: return 0.034857574850320816f;
  t291_n247: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t291_n253;
  t291_n248: if (E.getNumReferences() >= 3283697664 /*370.5*/) goto t291_n250;
  t291_n249: return 0.12061329931020737f;
  t291_n250: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t291_n252;
  t291_n251: return 0.17738579213619232f;
  t291_n252: return 0.5174714922904968f;
  t291_n253: return -0.09331991523504257f;
  t291_n254: return 0.0015429881168529391f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree292(const Example& E) {
  t292_n0: if (E.getNumReferences() >= 3301076992 /*1555.5*/) goto t292_n26;
  t292_n1: if (E.getNumReferences() >= 3300724736 /*1512.5*/) goto t292_n9;
  t292_n2: if (E.getNumReferences() >= 3300470784 /*1481.5*/) goto t292_n4;
  t292_n3: return -0.00010643956193234771f;
  t292_n4: if (E.getNumReferences() >= 3300478976 /*1482.5*/) goto t292_n8;
  t292_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t292_n7;
  t292_n6: return 0.09926671534776688f;
  t292_n7: return 0.15499138832092285f;
  t292_n8: return 0.03954660892486572f;
  t292_n9: if (E.getNumReferences() >= 3300794368 /*1521.0*/) goto t292_n11;
  t292_n10: return -0.14357411861419678f;
  t292_n11: if (E.getNumReferences() >= 3300814848 /*1523.5*/) goto t292_n17;
  t292_n12: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t292_n14;
  t292_n13: return -0.001876489957794547f;
  t292_n14: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t292_n16;
  t292_n15: return 0.448640912771225f;
  t292_n16: return 0.10507338494062424f;
  t292_n17: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n19;
  t292_n18: return -0.13184572756290436f;
  t292_n19: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t292_n21;
  t292_n20: return -0.04492493346333504f;
  t292_n21: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t292_n25;
  t292_n22: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t292_n24;
  t292_n23: return 0.025046277791261673f;
  t292_n24: return 0.14583592116832733f;
  t292_n25: return 0.03485521674156189f;
  t292_n26: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t292_n48;
  t292_n27: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n41;
  t292_n28: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t292_n40;
  t292_n29: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t292_n31;
  t292_n30: return -0.06192656606435776f;
  t292_n31: if (E.getNumReferences() >= 3322034688 /*8334.5*/) goto t292_n37;
  t292_n32: if (E.getNumReferences() >= 3314777088 /*4719.5*/) goto t292_n36;
  t292_n33: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t292_n35;
  t292_n34: return -0.07216344773769379f;
  t292_n35: return 0.10596461594104767f;
  t292_n36: return -0.2438717633485794f;
  t292_n37: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t292_n39;
  t292_n38: return 0.15098170936107635f;
  t292_n39: return 0.09526800364255905f;
  t292_n40: return 0.03284883126616478f;
  t292_n41: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t292_n47;
  t292_n42: if (E.getNumReferences() >= 3301531648 /*1611.0*/) goto t292_n46;
  t292_n43: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t292_n45;
  t292_n44: return 0.13932304084300995f;
  t292_n45: return 0.11895234137773514f;
  t292_n46: return 0.03930741548538208f;
  t292_n47: return 0.006084270775318146f;
  t292_n48: if (E.getNumReferences() >= 3301273600 /*1579.5*/) goto t292_n72;
  t292_n49: if (E.getNumReferences() >= 3301167104 /*1566.5*/) goto t292_n51;
  t292_n50: return -0.02157590351998806f;
  t292_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t292_n53;
  t292_n52: return -0.1751904934644699f;
  t292_n53: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t292_n71;
  t292_n54: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t292_n56;
  t292_n55: return -0.15220601856708527f;
  t292_n56: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t292_n58;
  t292_n57: return 0.3869872987270355f;
  t292_n58: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t292_n60;
  t292_n59: return 0.0460042729973793f;
  t292_n60: if (E.getNumReferences() >= 3301183488 /*1568.5*/) goto t292_n68;
  t292_n61: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t292_n67;
  t292_n62: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t292_n66;
  t292_n63: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t292_n65;
  t292_n64: return 0.13299186527729034f;
  t292_n65: return -0.10030864924192429f;
  t292_n66: return 0.13288740813732147f;
  t292_n67: return -0.1007649153470993f;
  t292_n68: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t292_n70;
  t292_n69: return 0.1558781862258911f;
  t292_n70: return 0.03439278155565262f;
  t292_n71: return 0.13215899467468262f;
  t292_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t292_n120;
  t292_n73: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n79;
  t292_n74: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown))) goto t292_n76;
  t292_n75: return -0.03673269972205162f;
  t292_n76: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t292_n78;
  t292_n77: return -0.027305129915475845f;
  t292_n78: return 0.0068029314279556274f;
  t292_n79: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t292_n115;
  t292_n80: if (E.getNumReferences() >= 3317105664 /*5856.5*/) goto t292_n110;
  t292_n81: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type))) goto t292_n83;
  t292_n82: return 0.01739048957824707f;
  t292_n83: if (E.getNumReferences() >= 3315510272 /*5077.5*/) goto t292_n101;
  t292_n84: if (E.getNumReferences() >= 3304468480 /*1969.5*/) goto t292_n88;
  t292_n85: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t292_n87;
  t292_n86: return -0.08674836158752441f;
  t292_n87: return 0.07893172651529312f;
  t292_n88: if (E.getNumReferences() >= 3312076800 /*3748.5*/) goto t292_n98;
  t292_n89: if (E.getNumReferences() >= 3311222784 /*3540.0*/) goto t292_n95;
  t292_n90: if (E.getNumReferences() >= 3307743232 /*2690.5*/) goto t292_n92;
  t292_n91: return 0.0835951417684555f;
  t292_n92: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t292_n94;
  t292_n93: return 0.10833698511123657f;
  t292_n94: return -0.014840011484920979f;
  t292_n95: if (E.getNumReferences() >= 3311788032 /*3678.0*/) goto t292_n97;
  t292_n96: return 0.24124786257743835f;
  t292_n97: return 0.10681584477424622f;
  t292_n98: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n100;
  t292_n99: return -0.17042304575443268f;
  t292_n100: return 0.1387716829776764f;
  t292_n101: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n103;
  t292_n102: return -0.1349417120218277f;
  t292_n103: if (E.getFileProximityDistanceCost() >= 3240099840 /*10.0*/) goto t292_n105;
  t292_n104: return -0.057568080723285675f;
  t292_n105: if (E.getNumReferences() >= 3315605504 /*5124.0*/) goto t292_n107;
  t292_n106: return 0.183158740401268f;
  t292_n107: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t292_n109;
  t292_n108: return 0.0889933854341507f;
  t292_n109: return 0.1868683099746704f;
  t292_n110: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t292_n112;
  t292_n111: return 0.0033309566788375378f;
  t292_n112: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n114;
  t292_n113: return -0.17925776541233063f;
  t292_n114: return 0.11345526576042175f;
  t292_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t292_n117;
  t292_n116: return 0.06791478395462036f;
  t292_n117: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t292_n119;
  t292_n118: return 0.19464711844921112f;
  t292_n119: return -0.03812247887253761f;
  t292_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t292_n220;
  t292_n121: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n219;
  t292_n122: if (E.getNumReferences() >= 3322564096 /*8851.5*/) goto t292_n180;
  t292_n123: if (E.getNumReferences() >= 3322543616 /*8831.5*/) goto t292_n171;
  t292_n124: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n126;
  t292_n125: return -0.025448402389883995f;
  t292_n126: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t292_n128;
  t292_n127: return -0.03133752942085266f;
  t292_n128: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t292_n170;
  t292_n129: if (E.getNumReferences() >= 3301384192 /*1593.0*/) goto t292_n131;
  t292_n130: return 0.43086761236190796f;
  t292_n131: if (E.getNumReferences() >= 3317616640 /*6106.0*/) goto t292_n165;
  t292_n132: if (E.getNumReferences() >= 3317563392 /*6080.0*/) goto t292_n156;
  t292_n133: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t292_n147;
  t292_n134: if (E.getNumReferences() >= 3317043200 /*5826.0*/) goto t292_n146;
  t292_n135: if (E.getNumReferences() >= 3316962304 /*5786.5*/) goto t292_n143;
  t292_n136: if (E.getNumReferences() >= 3316618240 /*5618.5*/) goto t292_n142;
  t292_n137: if (E.getNumReferences() >= 3314929664 /*4794.0*/) goto t292_n139;
  t292_n138: return 0.037621211260557175f;
  t292_n139: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t292_n141;
  t292_n140: return 0.08662839233875275f;
  t292_n141: return 0.17904675006866455f;
  t292_n142: return -0.07750601321458817f;
  t292_n143: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t292_n145;
  t292_n144: return 0.3289269506931305f;
  t292_n145: return 0.060486361384391785f;
  t292_n146: return -0.11170349270105362f;
  t292_n147: if (E.getNumReferences() >= 3314239488 /*4457.0*/) goto t292_n149;
  t292_n148: return 0.0029164047446101904f;
  t292_n149: if (E.getNumReferences() >= 3315017728 /*4837.0*/) goto t292_n155;
  t292_n150: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t292_n154;
  t292_n151: if (E.getNumReferences() >= 3314892800 /*4776.0*/) goto t292_n153;
  t292_n152: return -0.38295623660087585f;
  t292_n153: return -1.0065890550613403f;
  t292_n154: return -0.10155461728572845f;
  t292_n155: return -0.18516993522644043f;
  t292_n156: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t292_n158;
  t292_n157: return -0.10225098580121994f;
  t292_n158: if (E.getFractionNameInContext() >= 3182307328 /*0.0850372314453125*/) goto t292_n164;
  t292_n159: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t292_n161;
  t292_n160: return 0.35684776306152344f;
  t292_n161: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t292_n163;
  t292_n162: return -0.10490213334560394f;
  t292_n163: return 0.1387777030467987f;
  t292_n164: return 0.12847425043582916f;
  t292_n165: if (E.getNumReferences() >= 3319041024 /*6801.5*/) goto t292_n167;
  t292_n166: return -0.13215693831443787f;
  t292_n167: if (E.getNumReferences() >= 3319061504 /*6811.5*/) goto t292_n169;
  t292_n168: return 0.15200358629226685f;
  t292_n169: return 0.08134081959724426f;
  t292_n170: return 0.023149913176894188f;
  t292_n171: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t292_n179;
  t292_n172: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n174;
  t292_n173: return -0.1423344612121582f;
  t292_n174: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t292_n176;
  t292_n175: return -0.10974394530057907f;
  t292_n176: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t292_n178;
  t292_n177: return 0.13712479174137115f;
  t292_n178: return 0.25085654854774475f;
  t292_n179: return 0.16864913702011108f;
  t292_n180: if (E.getNumReferences() >= 3331470848 /*18715.0*/) goto t292_n188;
  t292_n181: if (E.getFractionNameInContext() >= 3198056936 /*0.3095238208770752*/) goto t292_n183;
  t292_n182: return -0.06259691715240479f;
  t292_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n187;
  t292_n184: if (E.getNumReferences() >= 3327690752 /*13858.0*/) goto t292_n186;
  t292_n185: return -0.06975921243429184f;
  t292_n186: return -0.47954511642456055f;
  t292_n187: return -0.1130913496017456f;
  t292_n188: if (E.getNumReferences() >= 3331802112 /*19362.0*/) goto t292_n212;
  t292_n189: if (E.getContextKind() & (BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t292_n201;
  t292_n190: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t292_n192;
  t292_n191: return -0.0021875486709177494f;
  t292_n192: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t292_n200;
  t292_n193: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n195;
  t292_n194: return -0.12370771914720535f;
  t292_n195: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t292_n199;
  t292_n196: if (E.getNumReferences() >= 3331634688 /*19035.0*/) goto t292_n198;
  t292_n197: return -0.13683603703975677f;
  t292_n198: return 0.5199753046035767f;
  t292_n199: return 0.06501589715480804f;
  t292_n200: return -0.10245569795370102f;
  t292_n201: if (E.getNumReferences() >= 3331720960 /*19203.5*/) goto t292_n207;
  t292_n202: if (E.getNumReferences() >= 3331557888 /*18885.0*/) goto t292_n206;
  t292_n203: if (E.getFileProximityDistanceCost() >= 3246391296 /*16.0*/) goto t292_n205;
  t292_n204: return -0.00015714671462774277f;
  t292_n205: return 0.10947050899267197f;
  t292_n206: return -0.17445337772369385f;
  t292_n207: if (E.getNumReferences() >= 3331773952 /*19307.0*/) goto t292_n211;
  t292_n208: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t292_n210;
  t292_n209: return 0.1683274507522583f;
  t292_n210: return 0.11592287570238113f;
  t292_n211: return 0.09143432974815369f;
  t292_n212: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t292_n216;
  t292_n213: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t292_n215;
  t292_n214: return 0.00663079833611846f;
  t292_n215: return 0.08878932148218155f;
  t292_n216: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t292_n218;
  t292_n217: return -0.02256106026470661f;
  t292_n218: return -0.16466939449310303f;
  t292_n219: return 0.003961900249123573f;
  t292_n220: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type))) goto t292_n228;
  t292_n221: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t292_n223;
  t292_n222: return 0.0012767926091328263f;
  t292_n223: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t292_n225;
  t292_n224: return 0.19629232585430145f;
  t292_n225: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t292_n227;
  t292_n226: return -0.036420900374650955f;
  t292_n227: return -2.905709981918335f;
  t292_n228: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t292_n230;
  t292_n229: return 0.10384159535169601f;
  t292_n230: if (E.getNumReferences() >= 3302625280 /*1744.5*/) goto t292_n254;
  t292_n231: if (E.getNumReferences() >= 3302445056 /*1722.5*/) goto t292_n253;
  t292_n232: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t292_n236;
  t292_n233: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t292_n235;
  t292_n234: return 0.08162941038608551f;
  t292_n235: return 0.1290973573923111f;
  t292_n236: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t292_n252;
  t292_n237: if (E.getNumReferences() >= 3302154240 /*1687.0*/) goto t292_n241;
  t292_n238: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t292_n240;
  t292_n239: return -0.06129513308405876f;
  t292_n240: return -1.2697627544403076f;
  t292_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t292_n243;
  t292_n242: return -0.03968014195561409f;
  t292_n243: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t292_n245;
  t292_n244: return -0.14816270768642426f;
  t292_n245: if (E.getNumReferences() >= 3302199296 /*1692.5*/) goto t292_n251;
  t292_n246: if (E.getFractionNameInContext() >= 3177418297 /*0.0555555559694767*/) goto t292_n250;
  t292_n247: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t292_n249;
  t292_n248: return 0.16385358572006226f;
  t292_n249: return 0.13178347051143646f;
  t292_n250: return 0.04433638975024223f;
  t292_n251: return 0.1314937323331833f;
  t292_n252: return 0.12125390022993088f;
  t292_n253: return -0.14188867807388306f;
  t292_n254: return 0.008687877096235752f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree293(const Example& E) {
  t293_n0: if (E.getNumReferences() >= 3278995456 /*241.5*/) goto t293_n2;
  t293_n1: return -0.0001333508116658777f;
  t293_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t293_n12;
  t293_n3: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t293_n5;
  t293_n4: return 0.008290800265967846f;
  t293_n5: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t293_n11;
  t293_n6: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t293_n8;
  t293_n7: return -0.0045789810828864574f;
  t293_n8: if (E.getNumReferences() >= 3297529856 /*1122.5*/) goto t293_n10;
  t293_n9: return 0.1316874772310257f;
  t293_n10: return 0.026329917833209038f;
  t293_n11: return 0.015830520540475845f;
  t293_n12: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t293_n252;
  t293_n13: if (E.getNumReferences() >= 3281010688 /*288.5*/) goto t293_n141;
  t293_n14: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t293_n114;
  t293_n15: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t293_n75;
  t293_n16: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t293_n74;
  t293_n17: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t293_n65;
  t293_n18: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t293_n50;
  t293_n19: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t293_n39;
  t293_n20: if (E.getNumReferences() >= 3279060992 /*242.5*/) goto t293_n24;
  t293_n21: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t293_n23;
  t293_n22: return -0.14295485615730286f;
  t293_n23: return -0.3750714361667633f;
  t293_n24: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t293_n36;
  t293_n25: if (E.getNumReferences() >= 3280863232 /*284.0*/) goto t293_n31;
  t293_n26: if (E.getNumReferences() >= 3280683008 /*278.5*/) goto t293_n28;
  t293_n27: return 0.06253645569086075f;
  t293_n28: if (E.getNumReferences() >= 3280715776 /*279.5*/) goto t293_n30;
  t293_n29: return -0.48539435863494873f;
  t293_n30: return 0.025092054158449173f;
  t293_n31: if (E.getNumReferences() >= 3280945152 /*286.5*/) goto t293_n35;
  t293_n32: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t293_n34;
  t293_n33: return 0.24437279999256134f;
  t293_n34: return 0.07622405886650085f;
  t293_n35: return -0.08933385461568832f;
  t293_n36: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t293_n38;
  t293_n37: return -0.4223736822605133f;
  t293_n38: return 0.06852363049983978f;
  t293_n39: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t293_n41;
  t293_n40: return -0.03910832479596138f;
  t293_n41: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t293_n49;
  t293_n42: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t293_n48;
  t293_n43: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t293_n45;
  t293_n44: return 0.03304189071059227f;
  t293_n45: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t293_n47;
  t293_n46: return 0.08223658800125122f;
  t293_n47: return 0.1163058653473854f;
  t293_n48: return -0.0624324306845665f;
  t293_n49: return 0.028771312907338142f;
  t293_n50: if (E.getNumReferences() >= 3280388096 /*269.5*/) goto t293_n64;
  t293_n51: if (E.getNumReferences() >= 3280125952 /*261.5*/) goto t293_n57;
  t293_n52: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t293_n56;
  t293_n53: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t293_n55;
  t293_n54: return -0.13398104906082153f;
  t293_n55: return 0.2747901678085327f;
  t293_n56: return 0.048389602452516556f;
  t293_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t293_n61;
  t293_n58: if (E.getNumReferences() >= 3280224256 /*264.5*/) goto t293_n60;
  t293_n59: return 0.2664181888103485f;
  t293_n60: return -0.08346358686685562f;
  t293_n61: if (E.getNumReferences() >= 3280224256 /*264.5*/) goto t293_n63;
  t293_n62: return -0.1001751720905304f;
  t293_n63: return 0.15595921874046326f;
  t293_n64: return -0.0015039937570691109f;
  t293_n65: if (E.getFractionNameInContext() >= 3210040662 /*0.8333333730697632*/) goto t293_n73;
  t293_n66: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t293_n70;
  t293_n67: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t293_n69;
  t293_n68: return -0.47610563039779663f;
  t293_n69: return 0.007838794961571693f;
  t293_n70: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t293_n72;
  t293_n71: return 0.030186593532562256f;
  t293_n72: return 0.16683749854564667f;
  t293_n73: return 0.12048441171646118f;
  t293_n74: return -0.010210234671831131f;
  t293_n75: if (E.getNumReferences() >= 3280879616 /*284.5*/) goto t293_n107;
  t293_n76: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t293_n78;
  t293_n77: return -0.05333375930786133f;
  t293_n78: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t293_n80;
  t293_n79: return -0.0021898173727095127f;
  t293_n80: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t293_n104;
  t293_n81: if (E.getNumReferences() >= 3279126528 /*243.5*/) goto t293_n83;
  t293_n82: return 0.14871732890605927f;
  t293_n83: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t293_n97;
  t293_n84: if (E.getNumReferences() >= 3279323136 /*246.5*/) goto t293_n88;
  t293_n85: if (E.getNumReferences() >= 3279257600 /*245.5*/) goto t293_n87;
  t293_n86: return -0.2880023717880249f;
  t293_n87: return 1.127785563468933f;
  t293_n88: if (E.getNumReferences() >= 3280617472 /*276.5*/) goto t293_n96;
  t293_n89: if (E.getNumReferences() >= 3280551936 /*274.5*/) goto t293_n93;
  t293_n90: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t293_n92;
  t293_n91: return 0.17115117609500885f;
  t293_n92: return 0.045807845890522f;
  t293_n93: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t293_n95;
  t293_n94: return -0.10507597774267197f;
  t293_n95: return 0.5199325680732727f;
  t293_n96: return -0.16721279919147491f;
  t293_n97: if (E.getNumReferences() >= 3280650240 /*277.5*/) goto t293_n103;
  t293_n98: if (E.getNumReferences() >= 3280076800 /*260.0*/) goto t293_n100;
  t293_n99: return -0.04171483963727951f;
  t293_n100: if (E.getNumReferences() >= 3280175104 /*263.0*/) goto t293_n102;
  t293_n101: return 0.12597782909870148f;
  t293_n102: return 0.07223904877901077f;
  t293_n103: return 0.41839155554771423f;
  t293_n104: if (E.getNumReferences() >= 3280748544 /*280.5*/) goto t293_n106;
  t293_n105: return 0.021879706531763077f;
  t293_n106: return -0.3233964145183563f;
  t293_n107: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t293_n113;
  t293_n108: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t293_n110;
  t293_n109: return 0.17964449524879456f;
  t293_n110: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t293_n112;
  t293_n111: return 0.15805412828922272f;
  t293_n112: return 0.1573547124862671f;
  t293_n113: return 0.002970025409013033f;
  t293_n114: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t293_n116;
  t293_n115: return -0.2410362958908081f;
  t293_n116: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t293_n134;
  t293_n117: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t293_n133;
  t293_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t293_n120;
  t293_n119: return -0.07736653089523315f;
  t293_n120: if (E.getNumReferences() >= 3280437248 /*271.0*/) goto t293_n132;
  t293_n121: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t293_n125;
  t293_n122: if (E.getNumReferences() >= 3279781888 /*253.5*/) goto t293_n124;
  t293_n123: return 0.1413295418024063f;
  t293_n124: return 0.1628420650959015f;
  t293_n125: if (E.getNumReferences() >= 3280289792 /*266.5*/) goto t293_n129;
  t293_n126: if (E.getNumReferences() >= 3279781888 /*253.5*/) goto t293_n128;
  t293_n127: return 0.10953360795974731f;
  t293_n128: return 0.054748326539993286f;
  t293_n129: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t293_n131;
  t293_n130: return 0.12669359147548676f;
  t293_n131: return 0.0666932687163353f;
  t293_n132: return -0.048667456954717636f;
  t293_n133: return -0.027746008709073067f;
  t293_n134: if (E.getNumReferences() >= 3280388096 /*269.5*/) goto t293_n140;
  t293_n135: if (E.getNumReferences() >= 3279994880 /*257.5*/) goto t293_n139;
  t293_n136: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t293_n138;
  t293_n137: return -0.10185285657644272f;
  t293_n138: return 0.22632081806659698f;
  t293_n139: return -0.292710542678833f;
  t293_n140: return 0.16448372602462769f;
  t293_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Variable))) goto t293_n143;
  t293_n142: return -0.0009520204039290547f;
  t293_n143: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t293_n249;
  t293_n144: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t293_n248;
  t293_n145: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_Expression))) goto t293_n183;
  t293_n146: if (E.getNumReferences() >= 3317499904 /*6049.0*/) goto t293_n148;
  t293_n147: return -0.10074985027313232f;
  t293_n148: if (E.getNumReferences() >= 3331047168 /*17887.5*/) goto t293_n164;
  t293_n149: if (E.getNumReferences() >= 3330417920 /*16658.5*/) goto t293_n157;
  t293_n150: if (E.getNumReferences() >= 3326208512 /*12410.5*/) goto t293_n156;
  t293_n151: if (E.getNumReferences() >= 3325800960 /*12012.5*/) goto t293_n153;
  t293_n152: return 0.036025553941726685f;
  t293_n153: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t293_n155;
  t293_n154: return -0.10743768513202667f;
  t293_n155: return 0.12065654247999191f;
  t293_n156: return -0.03348054736852646f;
  t293_n157: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t293_n159;
  t293_n158: return -0.11280521750450134f;
  t293_n159: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t293_n163;
  t293_n160: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t293_n162;
  t293_n161: return 0.12864433228969574f;
  t293_n162: return -0.12702807784080505f;
  t293_n163: return 0.1283947229385376f;
  t293_n164: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t293_n182;
  t293_n165: if (E.getNumReferences() >= 3334782208 /*25182.5*/) goto t293_n171;
  t293_n166: if (E.getNumReferences() >= 3332988672 /*21679.5*/) goto t293_n168;
  t293_n167: return 0.00394959282130003f;
  t293_n168: if (E.getNumReferences() >= 3333772544 /*23210.5*/) goto t293_n170;
  t293_n169: return -0.22291678190231323f;
  t293_n170: return -0.14458610117435455f;
  t293_n171: if (E.getNumReferences() >= 3335546112 /*26674.5*/) goto t293_n177;
  t293_n172: if (E.getFileProximityDistanceCost() >= 3247964160 /*19.0*/) goto t293_n176;
  t293_n173: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t293_n175;
  t293_n174: return -0.9530847668647766f;
  t293_n175: return 0.11735693365335464f;
  t293_n176: return -0.06815849244594574f;
  t293_n177: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t293_n179;
  t293_n178: return 0.012451714836061f;
  t293_n179: if (E.getNumReferences() >= 3337773824 /*31025.5*/) goto t293_n181;
  t293_n180: return -0.726635754108429f;
  t293_n181: return -0.005454115103930235f;
  t293_n182: return -0.10160122811794281f;
  t293_n183: if (E.getNumReferences() >= 3343394688 /*51239.5*/) goto t293_n245;
  t293_n184: if (E.getNumReferences() >= 3340578176 /*40237.5*/) goto t293_n244;
  t293_n185: if (E.getNumReferences() >= 3334782208 /*25182.5*/) goto t293_n243;
  t293_n186: if (E.getNumReferences() >= 3332605440 /*20931.0*/) goto t293_n234;
  t293_n187: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t293_n217;
  t293_n188: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t293_n190;
  t293_n189: return 0.0346715934574604f;
  t293_n190: if (E.getNumReferences() >= 3316737024 /*5676.5*/) goto t293_n216;
  t293_n191: if (E.getNumReferences() >= 3284598784 /*398.0*/) goto t293_n193;
  t293_n192: return -0.11996226012706757f;
  t293_n193: if (E.getNumReferences() >= 3285106688 /*413.5*/) goto t293_n195;
  t293_n194: return 0.16281571984291077f;
  t293_n195: if (E.getNumReferences() >= 3315036160 /*4846.0*/) goto t293_n213;
  t293_n196: if (E.getNumReferences() >= 3301523456 /*1610.0*/) goto t293_n212;
  t293_n197: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t293_n211;
  t293_n198: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t293_n202;
  t293_n199: if (E.getNumReferences() >= 3289571328 /*587.5*/) goto t293_n201;
  t293_n200: return 0.15812022984027863f;
  t293_n201: return -0.12310443818569183f;
  t293_n202: if (E.getNumReferences() >= 3287105536 /*474.5*/) goto t293_n206;
  t293_n203: if (E.getNumReferences() >= 3286941696 /*469.5*/) goto t293_n205;
  t293_n204: return -0.10066668689250946f;
  t293_n205: return 0.15219949185848236f;
  t293_n206: if (E.getNumReferences() >= 3291996160 /*735.5*/) goto t293_n208;
  t293_n207: return -0.12684877216815948f;
  t293_n208: if (E.getNumReferences() >= 3293077504 /*801.5*/) goto t293_n210;
  t293_n209: return 0.09665475785732269f;
  t293_n210: return 0.07922881841659546f;
  t293_n211: return 0.165486678481102f;
  t293_n212: return -0.0075683495961129665f;
  t293_n213: if (E.getNumReferences() >= 3316555776 /*5588.0*/) goto t293_n215;
  t293_n214: return 0.1271512508392334f;
  t293_n215: return 0.1504121571779251f;
  t293_n216: return 0.007184337358921766f;
  t293_n217: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t293_n219;
  t293_n218: return -0.04230573773384094f;
  t293_n219: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t293_n221;
  t293_n220: return 0.10290754586458206f;
  t293_n221: if (E.getNumReferences() >= 3308046336 /*2764.5*/) goto t293_n233;
  t293_n222: if (E.getNumReferences() >= 3307757568 /*2694.0*/) goto t293_n232;
  t293_n223: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t293_n229;
  t293_n224: if (E.getNumReferences() >= 3306885120 /*2481.0*/) goto t293_n226;
  t293_n225: return 0.0748426541686058f;
  t293_n226: if (E.getNumReferences() >= 3307317248 /*2586.5*/) goto t293_n228;
  t293_n227: return 0.12255379557609558f;
  t293_n228: return -0.056465648114681244f;
  t293_n229: if (E.getNumReferences() >= 3288858624 /*544.0*/) goto t293_n231;
  t293_n230: return -0.25998497009277344f;
  t293_n231: return 0.0660799965262413f;
  t293_n232: return 0.1404479295015335f;
  t293_n233: return 0.00044169023749418557f;
  t293_n234: if (E.getNumReferences() >= 3333403136 /*22489.0*/) goto t293_n236;
  t293_n235: return 0.11142141371965408f;
  t293_n236: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t293_n238;
  t293_n237: return 0.004128854721784592f;
  t293_n238: if (E.getNumReferences() >= 3333913600 /*23486.0*/) goto t293_n240;
  t293_n239: return 0.07829564809799194f;
  t293_n240: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t293_n242;
  t293_n241: return 0.11061322689056396f;
  t293_n242: return 0.06638634204864502f;
  t293_n243: return -0.09790457785129547f;
  t293_n244: return 0.07886287569999695f;
  t293_n245: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t293_n247;
  t293_n246: return -0.08541536331176758f;
  t293_n247: return 0.06449019908905029f;
  t293_n248: return -0.009385636076331139f;
  t293_n249: if (E.getNumReferences() >= 3282567168 /*336.0*/) goto t293_n251;
  t293_n250: return -0.3779848515987396f;
  t293_n251: return 0.07753023505210876f;
  t293_n252: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t293_n254;
  t293_n253: return -1.0331112146377563f;
  t293_n254: return 0.1268998086452484f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree294(const Example& E) {
  t294_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t294_n78;
  t294_n1: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t294_n3;
  t294_n2: return -0.0002726625825744122f;
  t294_n3: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t294_n17;
  t294_n4: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t294_n16;
  t294_n5: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t294_n15;
  t294_n6: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t294_n10;
  t294_n7: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t294_n9;
  t294_n8: return 0.006174413487315178f;
  t294_n9: return -0.10257120430469513f;
  t294_n10: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t294_n12;
  t294_n11: return -0.4684627950191498f;
  t294_n12: if (E.getNumReferences() >= 3273293824 /*154.5*/) goto t294_n14;
  t294_n13: return 0.11350394040346146f;
  t294_n14: return -0.11335763335227966f;
  t294_n15: return -0.002562563633546233f;
  t294_n16: return 0.10002081841230392f;
  t294_n17: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t294_n71;
  t294_n18: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t294_n70;
  t294_n19: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t294_n69;
  t294_n20: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t294_n68;
  t294_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t294_n53;
  t294_n22: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t294_n48;
  t294_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t294_n45;
  t294_n24: if (E.getNumReferences() >= 3266248704 /*87.5*/) goto t294_n34;
  t294_n25: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t294_n27;
  t294_n26: return -0.0941653847694397f;
  t294_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t294_n29;
  t294_n28: return -0.038465678691864014f;
  t294_n29: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement))) goto t294_n31;
  t294_n30: return -0.1260802000761032f;
  t294_n31: if (E.getFractionNameInContext() >= 3198296610 /*0.3166666626930237*/) goto t294_n33;
  t294_n32: return 0.13712915778160095f;
  t294_n33: return 0.12174160778522491f;
  t294_n34: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t294_n42;
  t294_n35: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t294_n41;
  t294_n36: if (E.getNumReferences() >= 3267493888 /*97.0*/) goto t294_n38;
  t294_n37: return 0.1294974386692047f;
  t294_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t294_n40;
  t294_n39: return -0.05648605525493622f;
  t294_n40: return 0.1348273605108261f;
  t294_n41: return 0.129031240940094f;
  t294_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t294_n44;
  t294_n43: return -0.10451851785182953f;
  t294_n44: return 0.025119217112660408f;
  t294_n45: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t294_n47;
  t294_n46: return 0.017575297504663467f;
  t294_n47: return 0.19724884629249573f;
  t294_n48: if (E.getNumReferences() >= 3283992576 /*379.5*/) goto t294_n50;
  t294_n49: return -0.026622993871569633f;
  t294_n50: if (E.getNumReferences() >= 3284025344 /*380.5*/) goto t294_n52;
  t294_n51: return 0.09569630771875381f;
  t294_n52: return -0.0007738377316854894f;
  t294_n53: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t294_n55;
  t294_n54: return -0.019381839781999588f;
  t294_n55: if (E.getNumReferences() >= 3305209856 /*2072.0*/) goto t294_n67;
  t294_n56: if (E.getNumReferences() >= 3282419712 /*331.5*/) goto t294_n58;
  t294_n57: return 0.06942299008369446f;
  t294_n58: if (E.getNumReferences() >= 3284074496 /*382.0*/) goto t294_n62;
  t294_n59: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t294_n61;
  t294_n60: return 0.11072345823049545f;
  t294_n61: return 0.13585598766803741f;
  t294_n62: if (E.getNumReferences() >= 3287629824 /*490.5*/) goto t294_n64;
  t294_n63: return -0.20709648728370667f;
  t294_n64: if (E.getNumReferences() >= 3291095040 /*680.5*/) goto t294_n66;
  t294_n65: return 0.21124474704265594f;
  t294_n66: return 0.07848618179559708f;
  t294_n67: return -0.17331591248512268f;
  t294_n68: return 0.052596356719732285f;
  t294_n69: return 0.002296584425494075f;
  t294_n70: return -0.006255114916712046f;
  t294_n71: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t294_n77;
  t294_n72: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t294_n76;
  t294_n73: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t294_n75;
  t294_n74: return 0.12598270177841187f;
  t294_n75: return 0.11546450853347778f;
  t294_n76: return 0.008826185017824173f;
  t294_n77: return 0.03401276469230652f;
  t294_n78: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t294_n82;
  t294_n79: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t294_n81;
  t294_n80: return -0.11964227259159088f;
  t294_n81: return 0.0009456570842303336f;
  t294_n82: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t294_n142;
  t294_n83: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t294_n117;
  t294_n84: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t294_n108;
  t294_n85: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t294_n87;
  t294_n86: return 0.040157873183488846f;
  t294_n87: if (E.getFractionNameInContext() >= 3190037058 /*0.16025641560554504*/) goto t294_n95;
  t294_n88: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t294_n92;
  t294_n89: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t294_n91;
  t294_n90: return -0.08202369511127472f;
  t294_n91: return -0.31623375415802f;
  t294_n92: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t294_n94;
  t294_n93: return -0.031039318069815636f;
  t294_n94: return 0.16593162715435028f;
  t294_n95: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t294_n99;
  t294_n96: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t294_n98;
  t294_n97: return 0.07268214225769043f;
  t294_n98: return 0.006718114018440247f;
  t294_n99: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t294_n105;
  t294_n100: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t294_n102;
  t294_n101: return 0.19877974689006805f;
  t294_n102: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t294_n104;
  t294_n103: return -0.4900897741317749f;
  t294_n104: return -0.16640357673168182f;
  t294_n105: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t294_n107;
  t294_n106: return 0.0212522204965353f;
  t294_n107: return -0.24469687044620514f;
  t294_n108: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t294_n112;
  t294_n109: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t294_n111;
  t294_n110: return -0.11986231058835983f;
  t294_n111: return 0.12138301879167557f;
  t294_n112: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t294_n116;
  t294_n113: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t294_n115;
  t294_n114: return 0.02692571096122265f;
  t294_n115: return -0.06103752553462982f;
  t294_n116: return -0.1399676650762558f;
  t294_n117: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t294_n127;
  t294_n118: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t294_n126;
  t294_n119: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t294_n125;
  t294_n120: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t294_n124;
  t294_n121: if (E.getNumNameInContext() >= 3204448256 /*0.5*/) goto t294_n123;
  t294_n122: return -0.15983733534812927f;
  t294_n123: return 0.09286628663539886f;
  t294_n124: return -0.2124442160129547f;
  t294_n125: return 0.13908299803733826f;
  t294_n126: return 0.10357282310724258f;
  t294_n127: if (E.getNumReferences() >= 3243769856 /*13.5*/) goto t294_n133;
  t294_n128: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t294_n132;
  t294_n129: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t294_n131;
  t294_n130: return -0.24387513101100922f;
  t294_n131: return 0.004966969136148691f;
  t294_n132: return 0.04941990226507187f;
  t294_n133: if (E.getNumReferences() >= 3273588736 /*159.0*/) goto t294_n141;
  t294_n134: if (E.getNumReferences() >= 3252420608 /*27.5*/) goto t294_n138;
  t294_n135: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t294_n137;
  t294_n136: return 0.10866162925958633f;
  t294_n137: return -0.041157979518175125f;
  t294_n138: if (E.getFractionNameInContext() >= 3185383238 /*0.10795454680919647*/) goto t294_n140;
  t294_n139: return 0.05958041176199913f;
  t294_n140: return 0.1275872141122818f;
  t294_n141: return -0.09945957362651825f;
  t294_n142: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t294_n234;
  t294_n143: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t294_n175;
  t294_n144: if (E.getFileProximityDistanceCost() >= 3251896320 /*26.5*/) goto t294_n174;
  t294_n145: if (E.getFileProximityDistanceCost() >= 3251372032 /*25.5*/) goto t294_n173;
  t294_n146: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t294_n148;
  t294_n147: return 0.004356464371085167f;
  t294_n148: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t294_n170;
  t294_n149: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t294_n157;
  t294_n150: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t294_n152;
  t294_n151: return -0.002143238903954625f;
  t294_n152: if (E.getNumReferences() >= 3240099840 /*10.0*/) goto t294_n154;
  t294_n153: return -0.12227557599544525f;
  t294_n154: if (E.getNumReferences() >= 3253993472 /*30.5*/) goto t294_n156;
  t294_n155: return 0.09970009326934814f;
  t294_n156: return 0.04571031406521797f;
  t294_n157: if (E.getNumReferences() >= 3258580992 /*46.5*/) goto t294_n159;
  t294_n158: return -0.06379066407680511f;
  t294_n159: if (E.getNumReferences() >= 3260022784 /*52.0*/) goto t294_n161;
  t294_n160: return 0.1025359183549881f;
  t294_n161: if (E.getNumReferences() >= 3270967296 /*123.5*/) goto t294_n165;
  t294_n162: if (E.getNumReferences() >= 3263889408 /*69.5*/) goto t294_n164;
  t294_n163: return -0.08400536328554153f;
  t294_n164: return 0.08157094568014145f;
  t294_n165: if (E.getNumReferences() >= 3288850432 /*543.5*/) goto t294_n167;
  t294_n166: return -0.13757385313510895f;
  t294_n167: if (E.getNumReferences() >= 3302936576 /*1782.5*/) goto t294_n169;
  t294_n168: return 0.10892301052808762f;
  t294_n169: return -0.05936073511838913f;
  t294_n170: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t294_n172;
  t294_n171: return 0.03196221590042114f;
  t294_n172: return -0.0027422751300036907f;
  t294_n173: return 0.12406844645738602f;
  t294_n174: return -0.13580410182476044f;
  t294_n175: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t294_n211;
  t294_n176: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t294_n206;
  t294_n177: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t294_n185;
  t294_n178: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t294_n180;
  t294_n179: return -0.004256896208971739f;
  t294_n180: if (E.getFractionNameInContext() >= 3195580300 /*0.2428571581840515*/) goto t294_n184;
  t294_n181: if (E.getNumReferences() >= 3231711232 /*5.0*/) goto t294_n183;
  t294_n182: return -0.14642837643623352f;
  t294_n183: return 0.02415599673986435f;
  t294_n184: return 0.06314531713724136f;
  t294_n185: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t294_n203;
  t294_n186: if (E.getFractionNameInContext() >= 3188270226 /*0.1339285671710968*/) goto t294_n200;
  t294_n187: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t294_n191;
  t294_n188: if (E.getNumReferences() >= 3212836864 /*1.0*/) goto t294_n190;
  t294_n189: return 0.07872215658426285f;
  t294_n190: return 0.0264888983219862f;
  t294_n191: if (E.getFractionNameInContext() >= 3169029689 /*0.02777777798473835*/) goto t294_n195;
  t294_n192: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t294_n194;
  t294_n193: return -0.4255882799625397f;
  t294_n194: return -0.034827493131160736f;
  t294_n195: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t294_n199;
  t294_n196: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t294_n198;
  t294_n197: return 0.07789546996355057f;
  t294_n198: return -0.0012803322169929743f;
  t294_n199: return -0.14996013045310974f;
  t294_n200: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t294_n202;
  t294_n201: return 0.0030060666613280773f;
  t294_n202: return 0.10599195957183838f;
  t294_n203: if (E.getNumReferences() >= 3279421440 /*248.0*/) goto t294_n205;
  t294_n204: return -0.08062668144702911f;
  t294_n205: return 0.08151387423276901f;
  t294_n206: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t294_n210;
  t294_n207: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t294_n209;
  t294_n208: return 0.024850964546203613f;
  t294_n209: return -0.01897972822189331f;
  t294_n210: return 0.06574668735265732f;
  t294_n211: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t294_n229;
  t294_n212: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t294_n226;
  t294_n213: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t294_n223;
  t294_n214: if (E.getFractionNameInContext() >= 3181648450 /*0.08012820780277252*/) goto t294_n216;
  t294_n215: return 0.0024300136137753725f;
  t294_n216: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t294_n218;
  t294_n217: return -0.04297032952308655f;
  t294_n218: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t294_n220;
  t294_n219: return 0.0011574423406273127f;
  t294_n220: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t294_n222;
  t294_n221: return -0.037206169217824936f;
  t294_n222: return -0.2598670721054077f;
  t294_n223: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t294_n225;
  t294_n224: return 0.03690030798316002f;
  t294_n225: return -0.09839562326669693f;
  t294_n226: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t294_n228;
  t294_n227: return -0.05329744517803192f;
  t294_n228: return -0.412372887134552f;
  t294_n229: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t294_n231;
  t294_n230: return 0.004829882178455591f;
  t294_n231: if (E.getFractionNameInContext() >= 3199763709 /*0.3603896200656891*/) goto t294_n233;
  t294_n232: return -0.15946657955646515f;
  t294_n233: return 0.04606055095791817f;
  t294_n234: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t294_n236;
  t294_n235: return -0.12284376472234726f;
  t294_n236: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t294_n238;
  t294_n237: return -0.08790931850671768f;
  t294_n238: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t294_n242;
  t294_n239: if (E.getNumReferences() >= 3282010112 /*319.0*/) goto t294_n241;
  t294_n240: return 0.18392427265644073f;
  t294_n241: return -0.16782313585281372f;
  t294_n242: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t294_n250;
  t294_n243: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t294_n249;
  t294_n244: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t294_n248;
  t294_n245: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t294_n247;
  t294_n246: return -0.10281042009592056f;
  t294_n247: return 0.013260139152407646f;
  t294_n248: return 0.07231831550598145f;
  t294_n249: return -0.12604033946990967f;
  t294_n250: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t294_n254;
  t294_n251: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t294_n253;
  t294_n252: return 0.14464855194091797f;
  t294_n253: return -0.10015922039747238f;
  t294_n254: return -0.06888487935066223f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree295(const Example& E) {
  t295_n0: if (E.getIsNameInContext() >= 3204448256 /*0.5*/) goto t295_n2;
  t295_n1: return -0.000307429232634604f;
  t295_n2: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t295_n254;
  t295_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t295_n249;
  t295_n4: if (E.getFractionNameInContext() >= 3203889016 /*0.4833333492279053*/) goto t295_n146;
  t295_n5: if (E.getFractionNameInContext() >= 3201851782 /*0.4226190447807312*/) goto t295_n7;
  t295_n6: return -0.0006625697133131325f;
  t295_n7: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Variable))) goto t295_n81;
  t295_n8: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t295_n56;
  t295_n9: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t295_n45;
  t295_n10: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t295_n16;
  t295_n11: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t295_n13;
  t295_n12: return -0.012654001824557781f;
  t295_n13: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t295_n15;
  t295_n14: return -0.2640877664089203f;
  t295_n15: return -0.03226972743868828f;
  t295_n16: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t295_n42;
  t295_n17: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t295_n25;
  t295_n18: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n20;
  t295_n19: return -0.1550494283437729f;
  t295_n20: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t295_n22;
  t295_n21: return 0.00531875342130661f;
  t295_n22: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t295_n24;
  t295_n23: return 0.12349960952997208f;
  t295_n24: return -0.05866634100675583f;
  t295_n25: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t295_n29;
  t295_n26: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t295_n28;
  t295_n27: return 0.09731168299913406f;
  t295_n28: return 0.14019875228405f;
  t295_n29: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t295_n41;
  t295_n30: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t295_n38;
  t295_n31: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t295_n37;
  t295_n32: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t295_n36;
  t295_n33: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t295_n35;
  t295_n34: return 0.10959303379058838f;
  t295_n35: return -0.24621346592903137f;
  t295_n36: return 0.11623165011405945f;
  t295_n37: return -0.000935611838940531f;
  t295_n38: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t295_n40;
  t295_n39: return 0.008401981554925442f;
  t295_n40: return 0.13638263940811157f;
  t295_n41: return -0.1604834645986557f;
  t295_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n44;
  t295_n43: return -0.8282231688499451f;
  t295_n44: return -0.03305497765541077f;
  t295_n45: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t295_n47;
  t295_n46: return 0.06733213365077972f;
  t295_n47: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t295_n55;
  t295_n48: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t295_n52;
  t295_n49: if (E.getFractionNameInContext() >= 3202317816 /*0.4365079402923584*/) goto t295_n51;
  t295_n50: return 0.05710090696811676f;
  t295_n51: return -0.03920721262693405f;
  t295_n52: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t295_n54;
  t295_n53: return 0.06720230728387833f;
  t295_n54: return -0.0020179220009595156f;
  t295_n55: return -0.21557879447937012f;
  t295_n56: if (E.getFileProximityDistanceCost() >= 3250323456 /*23.5*/) goto t295_n78;
  t295_n57: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t295_n71;
  t295_n58: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t295_n70;
  t295_n59: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n65;
  t295_n60: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t295_n62;
  t295_n61: return -0.05860811099410057f;
  t295_n62: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n64;
  t295_n63: return -0.3609648644924164f;
  t295_n64: return 0.08640354126691818f;
  t295_n65: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n69;
  t295_n66: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t295_n68;
  t295_n67: return 0.1278047412633896f;
  t295_n68: return -0.1553007960319519f;
  t295_n69: return -0.06481478363275528f;
  t295_n70: return -0.19566388428211212f;
  t295_n71: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n73;
  t295_n72: return -0.17133170366287231f;
  t295_n73: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t295_n77;
  t295_n74: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t295_n76;
  t295_n75: return 0.11721126735210419f;
  t295_n76: return 0.060570862144231796f;
  t295_n77: return -0.04187364876270294f;
  t295_n78: if (E.getFileProximityDistanceCost() >= 3250847744 /*24.5*/) goto t295_n80;
  t295_n79: return 0.09211702644824982f;
  t295_n80: return -0.03676316514611244f;
  t295_n81: if (E.getNumReferences() >= 3240624128 /*10.5*/) goto t295_n145;
  t295_n82: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t295_n102;
  t295_n83: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t295_n93;
  t295_n84: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t295_n92;
  t295_n85: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t295_n91;
  t295_n86: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t295_n90;
  t295_n87: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n89;
  t295_n88: return -0.21630682051181793f;
  t295_n89: return -0.11188769340515137f;
  t295_n90: return -0.1253291219472885f;
  t295_n91: return -0.12415684014558792f;
  t295_n92: return -0.018440399318933487f;
  t295_n93: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t295_n101;
  t295_n94: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t295_n98;
  t295_n95: if (E.getFileProximityDistanceCost() >= 3229614080 /*4.0*/) goto t295_n97;
  t295_n96: return 0.025073692202568054f;
  t295_n97: return 0.1115230992436409f;
  t295_n98: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t295_n100;
  t295_n99: return 0.09087934345006943f;
  t295_n100: return -0.03869011253118515f;
  t295_n101: return -0.052060581743717194f;
  t295_n102: if (E.getScope() & (BIT(Scope_type::FileScope)|BIT(Scope_type::GlobalScope))) goto t295_n112;
  t295_n103: if (E.getFractionNameInContext() >= 3203040378 /*0.4580419659614563*/) goto t295_n111;
  t295_n104: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n106;
  t295_n105: return 0.012290822342038155f;
  t295_n106: if (E.getFractionNameInContext() >= 3202753588 /*0.44949495792388916*/) goto t295_n108;
  t295_n107: return -0.026482488960027695f;
  t295_n108: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t295_n110;
  t295_n109: return 0.03994712978601456f;
  t295_n110: return 0.10160890966653824f;
  t295_n111: return 0.09466083347797394f;
  t295_n112: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t295_n126;
  t295_n113: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t295_n119;
  t295_n114: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t295_n118;
  t295_n115: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n117;
  t295_n116: return -0.1981862485408783f;
  t295_n117: return -0.021739741787314415f;
  t295_n118: return 0.11172333359718323f;
  t295_n119: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n123;
  t295_n120: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t295_n122;
  t295_n121: return -0.13695178925991058f;
  t295_n122: return 0.09152018278837204f;
  t295_n123: if (E.getFileProximityDistanceCost() >= 3250061312 /*23.0*/) goto t295_n125;
  t295_n124: return 0.11706038564443588f;
  t295_n125: return 0.0639403834939003f;
  t295_n126: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t295_n144;
  t295_n127: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t295_n143;
  t295_n128: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t295_n132;
  t295_n129: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t295_n131;
  t295_n130: return 0.10291269421577454f;
  t295_n131: return -0.13762104511260986f;
  t295_n132: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t295_n138;
  t295_n133: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t295_n135;
  t295_n134: return -0.08677633106708527f;
  t295_n135: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n137;
  t295_n136: return 0.024017345160245895f;
  t295_n137: return 0.09995020180940628f;
  t295_n138: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t295_n140;
  t295_n139: return 0.014941644854843616f;
  t295_n140: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n142;
  t295_n141: return 0.09949798136949539f;
  t295_n142: return 0.10376816987991333f;
  t295_n143: return -0.13502995669841766f;
  t295_n144: return 0.001933225430548191f;
  t295_n145: return 0.1689758002758026f;
  t295_n146: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t295_n162;
  t295_n147: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t295_n161;
  t295_n148: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t295_n150;
  t295_n149: return 0.0048643131740391254f;
  t295_n150: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t295_n158;
  t295_n151: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t295_n157;
  t295_n152: if (E.getFractionNameInContext() >= 3208642560 /*0.75*/) goto t295_n156;
  t295_n153: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t295_n155;
  t295_n154: return -0.16854827105998993f;
  t295_n155: return 0.09516482055187225f;
  t295_n156: return 0.1588035225868225f;
  t295_n157: return -0.25647881627082825f;
  t295_n158: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t295_n160;
  t295_n159: return 0.017297251150012016f;
  t295_n160: return 0.11959268152713776f;
  t295_n161: return -0.08530192822217941f;
  t295_n162: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t295_n164;
  t295_n163: return -0.00980293657630682f;
  t295_n164: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t295_n228;
  t295_n165: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n177;
  t295_n166: if (E.getFileProximityDistanceCost() >= 3249274880 /*21.5*/) goto t295_n176;
  t295_n167: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t295_n173;
  t295_n168: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t295_n170;
  t295_n169: return -0.37846654653549194f;
  t295_n170: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t295_n172;
  t295_n171: return -0.023147866129875183f;
  t295_n172: return -0.2622244656085968f;
  t295_n173: if (E.getFileProximityDistanceCost() >= 3246915584 /*17.0*/) goto t295_n175;
  t295_n174: return -0.021406371146440506f;
  t295_n175: return 0.14928047358989716f;
  t295_n176: return 0.08687737584114075f;
  t295_n177: if (E.getFractionNameInContext() >= 3207459552 /*0.6794872283935547*/) goto t295_n227;
  t295_n178: if (E.getFractionNameInContext() >= 3206990258 /*0.6515151262283325*/) goto t295_n220;
  t295_n179: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Type))) goto t295_n181;
  t295_n180: return -0.014042364433407784f;
  t295_n181: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t295_n193;
  t295_n182: if (E.getFractionNameInContext() >= 3204829556 /*0.5227272510528564*/) goto t295_n186;
  t295_n183: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t295_n185;
  t295_n184: return 0.2757861018180847f;
  t295_n185: return -0.11859754472970963f;
  t295_n186: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t295_n192;
  t295_n187: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t295_n191;
  t295_n188: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t295_n190;
  t295_n189: return -0.24034520983695984f;
  t295_n190: return 0.1129242479801178f;
  t295_n191: return 0.06556855142116547f;
  t295_n192: return 0.09529133886098862f;
  t295_n193: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t295_n201;
  t295_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t295_n200;
  t295_n195: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t295_n197;
  t295_n196: return 0.04850633069872856f;
  t295_n197: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t295_n199;
  t295_n198: return -0.3498709499835968f;
  t295_n199: return -0.054554253816604614f;
  t295_n200: return 0.1130017638206482f;
  t295_n201: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t295_n209;
  t295_n202: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t295_n206;
  t295_n203: if (E.getNumNameInContext() >= 3230662656 /*4.5*/) goto t295_n205;
  t295_n204: return -0.06617436558008194f;
  t295_n205: return 0.38530778884887695f;
  t295_n206: if (E.getFractionNameInContext() >= 3205746493 /*0.5773809552192688*/) goto t295_n208;
  t295_n207: return 0.14336660504341125f;
  t295_n208: return -0.10365808755159378f;
  t295_n209: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t295_n211;
  t295_n210: return -0.3350526988506317f;
  t295_n211: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n219;
  t295_n212: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t295_n218;
  t295_n213: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t295_n217;
  t295_n214: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n216;
  t295_n215: return -0.2363811433315277f;
  t295_n216: return 0.14460939168930054f;
  t295_n217: return 0.11127707362174988f;
  t295_n218: return -0.07436706125736237f;
  t295_n219: return 0.09676487743854523f;
  t295_n220: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n226;
  t295_n221: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t295_n223;
  t295_n222: return 0.014974238350987434f;
  t295_n223: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t295_n225;
  t295_n224: return 0.04542931169271469f;
  t295_n225: return 0.10399049520492554f;
  t295_n226: return 0.12573713064193726f;
  t295_n227: return -0.0445987731218338f;
  t295_n228: if (E.getFractionNameInContext() >= 3205513476 /*0.5634920597076416*/) goto t295_n246;
  t295_n229: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t295_n231;
  t295_n230: return -0.5574516654014587f;
  t295_n231: if (E.getScope() & (BIT(Scope_type::ClassScope))) goto t295_n239;
  t295_n232: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t295_n238;
  t295_n233: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t295_n237;
  t295_n234: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t295_n236;
  t295_n235: return -0.32188138365745544f;
  t295_n236: return -0.02264241874217987f;
  t295_n237: return 0.07032222300767899f;
  t295_n238: return 0.008437363430857658f;
  t295_n239: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t295_n241;
  t295_n240: return 0.01686001569032669f;
  t295_n241: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Variable))) goto t295_n243;
  t295_n242: return 0.03360579162836075f;
  t295_n243: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t295_n245;
  t295_n244: return 0.05428960919380188f;
  t295_n245: return -0.018798330798745155f;
  t295_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t295_n248;
  t295_n247: return -0.015596055425703526f;
  t295_n248: return 0.10657462477684021f;
  t295_n249: if (E.getFractionNameInContext() >= 3172269734 /*0.03637566417455673*/) goto t295_n253;
  t295_n250: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t295_n252;
  t295_n251: return -0.10013674944639206f;
  t295_n252: return 0.122661292552948f;
  t295_n253: return 0.0009560154867358506f;
  t295_n254: return 0.003941553179174662f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree296(const Example& E) {
  t296_n0: if (E.getNumReferences() >= 3301855232 /*1650.5*/) goto t296_n88;
  t296_n1: if (E.getNumReferences() >= 3301695488 /*1631.0*/) goto t296_n75;
  t296_n2: if (E.getFractionNameInContext() >= 3171926904 /*0.035098522901535034*/) goto t296_n72;
  t296_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t296_n23;
  t296_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n6;
  t296_n5: return -0.003792749485000968f;
  t296_n6: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t296_n20;
  t296_n7: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t296_n15;
  t296_n8: if (E.getNumReferences() >= 3266772992 /*91.5*/) goto t296_n12;
  t296_n9: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t296_n11;
  t296_n10: return 0.0186348594725132f;
  t296_n11: return -0.043958213180303574f;
  t296_n12: if (E.getNumReferences() >= 3267100672 /*94.0*/) goto t296_n14;
  t296_n13: return 0.10657303780317307f;
  t296_n14: return 0.055959876626729965f;
  t296_n15: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t296_n19;
  t296_n16: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t296_n18;
  t296_n17: return -0.03644685447216034f;
  t296_n18: return 0.031332019716501236f;
  t296_n19: return 0.06539852172136307f;
  t296_n20: if (E.getNumReferences() >= 3301306368 /*1583.5*/) goto t296_n22;
  t296_n21: return 0.005980568006634712f;
  t296_n22: return 0.09315266460180283f;
  t296_n23: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers))) goto t296_n51;
  t296_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t296_n26;
  t296_n25: return -0.01926361210644245f;
  t296_n26: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t296_n50;
  t296_n27: if (E.getNumReferences() >= 3296419840 /*1005.5*/) goto t296_n47;
  t296_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n30;
  t296_n29: return -0.021599426865577698f;
  t296_n30: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t296_n32;
  t296_n31: return -0.039887379854917526f;
  t296_n32: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t296_n40;
  t296_n33: if (E.getSymbolScopeDistanceCost() >= 3235905536 /*7.0*/) goto t296_n35;
  t296_n34: return 0.016259105876088142f;
  t296_n35: if (E.getNumReferences() >= 3204448256 /*0.5*/) goto t296_n39;
  t296_n36: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t296_n38;
  t296_n37: return 0.1415453851222992f;
  t296_n38: return 0.07881946116685867f;
  t296_n39: return -0.15154756605625153f;
  t296_n40: if (E.getNumReferences() >= 3275325440 /*185.5*/) goto t296_n44;
  t296_n41: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t296_n43;
  t296_n42: return 0.0986408144235611f;
  t296_n43: return -0.010855454951524734f;
  t296_n44: if (E.getNumReferences() >= 3276734464 /*207.0*/) goto t296_n46;
  t296_n45: return 0.2221812754869461f;
  t296_n46: return 0.05559378117322922f;
  t296_n47: if (E.getNumReferences() >= 3296763904 /*1029.0*/) goto t296_n49;
  t296_n48: return 0.22286051511764526f;
  t296_n49: return 0.05546315759420395f;
  t296_n50: return 0.008685861714184284f;
  t296_n51: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t296_n69;
  t296_n52: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t296_n54;
  t296_n53: return 0.00025683673447929323f;
  t296_n54: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t296_n66;
  t296_n55: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t296_n57;
  t296_n56: return 0.0044619617983698845f;
  t296_n57: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t296_n63;
  t296_n58: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t296_n62;
  t296_n59: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t296_n61;
  t296_n60: return 0.060925569385290146f;
  t296_n61: return 0.013857178390026093f;
  t296_n62: return -0.02066073566675186f;
  t296_n63: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t296_n65;
  t296_n64: return 0.015406932681798935f;
  t296_n65: return 0.03968654200434685f;
  t296_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Symbol))) goto t296_n68;
  t296_n67: return 0.00782020203769207f;
  t296_n68: return 0.03970589116215706f;
  t296_n69: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t296_n71;
  t296_n70: return 0.002321674255654216f;
  t296_n71: return -0.0032362155616283417f;
  t296_n72: if (E.getFractionNameInContext() >= 3172638464 /*0.037749290466308594*/) goto t296_n74;
  t296_n73: return 0.12454614043235779f;
  t296_n74: return 0.0005038258386775851f;
  t296_n75: if (E.getNumReferences() >= 3301715968 /*1633.5*/) goto t296_n81;
  t296_n76: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n80;
  t296_n77: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t296_n79;
  t296_n78: return -0.2580929100513458f;
  t296_n79: return -0.38063791394233704f;
  t296_n80: return -0.10820883512496948f;
  t296_n81: if (E.getNumReferences() >= 3301724160 /*1634.5*/) goto t296_n87;
  t296_n82: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t296_n84;
  t296_n83: return 0.1449742466211319f;
  t296_n84: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t296_n86;
  t296_n85: return -0.11247396469116211f;
  t296_n86: return 0.1002078577876091f;
  t296_n87: return -0.06895124167203903f;
  t296_n88: if (E.getNumReferences() >= 3301920768 /*1658.5*/) goto t296_n90;
  t296_n89: return 0.15479454398155212f;
  t296_n90: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t296_n150;
  t296_n91: if (E.getNumReferences() >= 3302162432 /*1688.0*/) goto t296_n103;
  t296_n92: if (E.getNumReferences() >= 3302121472 /*1683.0*/) goto t296_n102;
  t296_n93: if (E.getNumReferences() >= 3301961728 /*1663.5*/) goto t296_n97;
  t296_n94: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t296_n96;
  t296_n95: return -0.06404364854097366f;
  t296_n96: return 0.13379856944084167f;
  t296_n97: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t296_n101;
  t296_n98: if (E.getNumReferences() >= 3301982208 /*1666.0*/) goto t296_n100;
  t296_n99: return 0.10354941338300705f;
  t296_n100: return 0.08996894210577011f;
  t296_n101: return -0.24772214889526367f;
  t296_n102: return 0.15075752139091492f;
  t296_n103: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t296_n117;
  t296_n104: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n114;
  t296_n105: if (E.getNumReferences() >= 3302694912 /*1753.0*/) goto t296_n107;
  t296_n106: return -1.6892927885055542f;
  t296_n107: if (E.getNumReferences() >= 3308113920 /*2781.0*/) goto t296_n111;
  t296_n108: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t296_n110;
  t296_n109: return 0.10654772073030472f;
  t296_n110: return -0.004777395166456699f;
  t296_n111: if (E.getFractionNameInContext() >= 3191585724 /*0.18333333730697632*/) goto t296_n113;
  t296_n112: return -0.18352262675762177f;
  t296_n113: return 0.04837379604578018f;
  t296_n114: if (E.getNumReferences() >= 3302191104 /*1691.5*/) goto t296_n116;
  t296_n115: return -0.15382680296897888f;
  t296_n116: return 0.008313553407788277f;
  t296_n117: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t296_n125;
  t296_n118: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t296_n124;
  t296_n119: if (E.getNumReferences() >= 3303976960 /*1909.5*/) goto t296_n121;
  t296_n120: return -0.18376129865646362f;
  t296_n121: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Variable))) goto t296_n123;
  t296_n122: return 0.02351256273686886f;
  t296_n123: return 0.15767456591129303f;
  t296_n124: return -0.007267303299158812f;
  t296_n125: if (E.getNumReferences() >= 3303120896 /*1805.0*/) goto t296_n129;
  t296_n126: if (E.getNumReferences() >= 3302739968 /*1758.5*/) goto t296_n128;
  t296_n127: return -0.1110176369547844f;
  t296_n128: return 0.10070246458053589f;
  t296_n129: if (E.getNumReferences() >= 3306379264 /*2357.5*/) goto t296_n131;
  t296_n130: return -0.05872928351163864f;
  t296_n131: if (E.getNumReferences() >= 3309735936 /*3177.0*/) goto t296_n149;
  t296_n132: if (E.getNumReferences() >= 3309430784 /*3102.5*/) goto t296_n146;
  t296_n133: if (E.getNumReferences() >= 3308132352 /*2785.5*/) goto t296_n143;
  t296_n134: if (E.getNumReferences() >= 3307251712 /*2570.5*/) goto t296_n136;
  t296_n135: return 0.05037395656108856f;
  t296_n136: if (E.getNumReferences() >= 3307692032 /*2678.0*/) goto t296_n142;
  t296_n137: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t296_n141;
  t296_n138: if (E.getNumReferences() >= 3307503616 /*2632.0*/) goto t296_n140;
  t296_n139: return 0.08304157853126526f;
  t296_n140: return 0.11359670013189316f;
  t296_n141: return 0.030658887699246407f;
  t296_n142: return 0.14378581941127777f;
  t296_n143: if (E.getFractionNameInContext() >= 3181788120 /*0.08116883039474487*/) goto t296_n145;
  t296_n144: return -0.10849710553884506f;
  t296_n145: return 0.1457548588514328f;
  t296_n146: if (E.getNumReferences() >= 3309547520 /*3131.0*/) goto t296_n148;
  t296_n147: return 0.11989159137010574f;
  t296_n148: return 0.10930158942937851f;
  t296_n149: return 0.01509677991271019f;
  t296_n150: if (E.getNumReferences() >= 3302502400 /*1729.5*/) goto t296_n168;
  t296_n151: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n161;
  t296_n152: if (E.getNumReferences() >= 3301969920 /*1664.5*/) goto t296_n154;
  t296_n153: return 0.08533132821321487f;
  t296_n154: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t296_n156;
  t296_n155: return -0.11858794093132019f;
  t296_n156: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t296_n158;
  t296_n157: return -0.055958375334739685f;
  t296_n158: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t296_n160;
  t296_n159: return -0.5163357853889465f;
  t296_n160: return -0.1704103797674179f;
  t296_n161: if (E.getNumReferences() >= 3302445056 /*1722.5*/) goto t296_n163;
  t296_n162: return -0.01463487558066845f;
  t296_n163: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t296_n167;
  t296_n164: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t296_n166;
  t296_n165: return -0.17897403240203857f;
  t296_n166: return -0.13083557784557343f;
  t296_n167: return 0.13476771116256714f;
  t296_n168: if (E.getNumReferences() >= 3302584320 /*1739.5*/) goto t296_n180;
  t296_n169: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n171;
  t296_n170: return -0.4232894778251648f;
  t296_n171: if (E.getSymbolScopeDistanceCost() >= 3243245568 /*13.0*/) goto t296_n175;
  t296_n172: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t296_n174;
  t296_n173: return -0.042353902012109756f;
  t296_n174: return 0.2099180668592453f;
  t296_n175: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t296_n177;
  t296_n176: return 0.04703996703028679f;
  t296_n177: if (E.getNumReferences() >= 3302551552 /*1735.5*/) goto t296_n179;
  t296_n178: return -0.15542860329151154f;
  t296_n179: return 0.12402799725532532f;
  t296_n180: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t296_n254;
  t296_n181: if (E.getFractionNameInContext() >= 3197040135 /*0.2792207896709442*/) goto t296_n237;
  t296_n182: if (E.getNumReferences() >= 3338881664 /*33610.5*/) goto t296_n210;
  t296_n183: if (E.getNumReferences() >= 3337280256 /*30061.5*/) goto t296_n187;
  t296_n184: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n186;
  t296_n185: return -0.02735687792301178f;
  t296_n186: return 0.0012201581848785281f;
  t296_n187: if (E.getNumReferences() >= 3337447424 /*30388.0*/) goto t296_n191;
  t296_n188: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n190;
  t296_n189: return -0.26250189542770386f;
  t296_n190: return 0.13393740355968475f;
  t296_n191: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t296_n209;
  t296_n192: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t296_n200;
  t296_n193: if (E.getNumReferences() >= 3338715264 /*32960.5*/) goto t296_n199;
  t296_n194: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t296_n196;
  t296_n195: return -0.056298062205314636f;
  t296_n196: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t296_n198;
  t296_n197: return 0.0031348983757197857f;
  t296_n198: return 0.17455394566059113f;
  t296_n199: return 0.18958355486392975f;
  t296_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n202;
  t296_n201: return -0.1382000595331192f;
  t296_n202: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t296_n208;
  t296_n203: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t296_n205;
  t296_n204: return 0.08451903611421585f;
  t296_n205: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t296_n207;
  t296_n206: return -0.013845370151102543f;
  t296_n207: return 0.10723216086626053f;
  t296_n208: return 0.10501693934202194f;
  t296_n209: return 0.004088114481419325f;
  t296_n210: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t296_n234;
  t296_n211: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t296_n219;
  t296_n212: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t296_n214;
  t296_n213: return -0.035311974585056305f;
  t296_n214: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t296_n216;
  t296_n215: return 0.047436103224754333f;
  t296_n216: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t296_n218;
  t296_n217: return 0.12157131731510162f;
  t296_n218: return 0.15513156354427338f;
  t296_n219: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t296_n225;
  t296_n220: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t296_n224;
  t296_n221: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t296_n223;
  t296_n222: return -0.06026005744934082f;
  t296_n223: return -0.7315252423286438f;
  t296_n224: return 0.019418912008404732f;
  t296_n225: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t296_n233;
  t296_n226: if (E.getNumReferences() >= 3349739392 /*86511.0*/) goto t296_n228;
  t296_n227: return -0.07657524198293686f;
  t296_n228: if (E.getNumReferences() >= 3353027584 /*112200.0*/) goto t296_n230;
  t296_n229: return 0.24763749539852142f;
  t296_n230: if (E.getNumReferences() >= 3367011984 /*361524.5*/) goto t296_n232;
  t296_n231: return -0.053235385566949844f;
  t296_n232: return 0.10843934863805771f;
  t296_n233: return -0.0195704847574234f;
  t296_n234: if (E.getNumReferences() >= 3364768256 /*291408.0*/) goto t296_n236;
  t296_n235: return -0.03387778252363205f;
  t296_n236: return -0.08404858410358429f;
  t296_n237: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t296_n239;
  t296_n238: return 0.08236657828092575f;
  t296_n239: if (E.getNumReferences() >= 3304013824 /*1914.0*/) goto t296_n245;
  t296_n240: if (E.getNumReferences() >= 3303661568 /*1871.0*/) goto t296_n244;
  t296_n241: if (E.getNumReferences() >= 3302780928 /*1763.5*/) goto t296_n243;
  t296_n242: return 0.10647767782211304f;
  t296_n243: return -0.18207405507564545f;
  t296_n244: return 0.1608896404504776f;
  t296_n245: if (E.getNumReferences() >= 3346550272 /*63566.0*/) goto t296_n253;
  t296_n246: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t296_n252;
  t296_n247: if (E.getNumReferences() >= 3315738624 /*5189.0*/) goto t296_n249;
  t296_n248: return -0.06745928525924683f;
  t296_n249: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t296_n251;
  t296_n250: return -0.18824368715286255f;
  t296_n251: return -0.6234965324401855f;
  t296_n252: return -0.08751855045557022f;
  t296_n253: return 0.00636221282184124f;
  t296_n254: return 0.0019597213249653578f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree297(const Example& E) {
  t297_n0: if (E.getNumReferences() >= 3255173120 /*33.5*/) goto t297_n2;
  t297_n1: return -0.000162396056111902f;
  t297_n2: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t297_n18;
  t297_n3: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Unknown))) goto t297_n17;
  t297_n4: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t297_n12;
  t297_n5: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n11;
  t297_n6: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t297_n10;
  t297_n7: if (E.getNumReferences() >= 3310458880 /*3353.5*/) goto t297_n9;
  t297_n8: return -0.04646610468626022f;
  t297_n9: return 0.10676737129688263f;
  t297_n10: return 0.12825049459934235f;
  t297_n11: return 0.0044594802893698215f;
  t297_n12: if (E.getNumReferences() >= 3255697408 /*35.5*/) goto t297_n14;
  t297_n13: return 0.13219377398490906f;
  t297_n14: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_SymbolOrNewName))) goto t297_n16;
  t297_n15: return 0.054817404597997665f;
  t297_n16: return 0.17587308585643768f;
  t297_n17: return 0.09542221575975418f;
  t297_n18: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t297_n248;
  t297_n19: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t297_n237;
  t297_n20: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t297_n236;
  t297_n21: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro))) goto t297_n179;
  t297_n22: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t297_n166;
  t297_n23: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t297_n55;
  t297_n24: if (E.getFractionNameInContext() >= 3203249884 /*0.4642857313156128*/) goto t297_n54;
  t297_n25: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t297_n51;
  t297_n26: if (E.getNumReferences() >= 3266248704 /*87.5*/) goto t297_n36;
  t297_n27: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t297_n29;
  t297_n28: return -0.08434858173131943f;
  t297_n29: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown))) goto t297_n31;
  t297_n30: return -0.038784317672252655f;
  t297_n31: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t297_n35;
  t297_n32: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t297_n34;
  t297_n33: return 0.08834526687860489f;
  t297_n34: return -1.0223981142044067f;
  t297_n35: return 0.1262526959180832f;
  t297_n36: if (E.getNumReferences() >= 3272572928 /*143.5*/) goto t297_n44;
  t297_n37: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t297_n43;
  t297_n38: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t297_n42;
  t297_n39: if (E.getNumReferences() >= 3267559424 /*97.5*/) goto t297_n41;
  t297_n40: return 0.13970276713371277f;
  t297_n41: return -0.0444595031440258f;
  t297_n42: return 0.11563542485237122f;
  t297_n43: return 0.12244383990764618f;
  t297_n44: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t297_n46;
  t297_n45: return 0.03171764314174652f;
  t297_n46: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n50;
  t297_n47: if (E.getNumReferences() >= 3281485824 /*303.0*/) goto t297_n49;
  t297_n48: return -0.8457040786743164f;
  t297_n49: return -0.19370664656162262f;
  t297_n50: return -0.007869178429245949f;
  t297_n51: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t297_n53;
  t297_n52: return 0.01738603785634041f;
  t297_n53: return 0.19080039858818054f;
  t297_n54: return 0.11233960092067719f;
  t297_n55: if (E.getNumReferences() >= 3284418560 /*392.5*/) goto t297_n135;
  t297_n56: if (E.getFractionNameInContext() >= 3177025848 /*0.05409356951713562*/) goto t297_n58;
  t297_n57: return -0.017717301845550537f;
  t297_n58: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t297_n102;
  t297_n59: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t297_n81;
  t297_n60: if (E.getNumReferences() >= 3272638464 /*144.5*/) goto t297_n70;
  t297_n61: if (E.getNumReferences() >= 3272441856 /*141.5*/) goto t297_n69;
  t297_n62: if (E.getNumReferences() >= 3263758336 /*68.5*/) goto t297_n66;
  t297_n63: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n65;
  t297_n64: return -0.08594613522291183f;
  t297_n65: return 0.05254755914211273f;
  t297_n66: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n68;
  t297_n67: return -0.6564322710037231f;
  t297_n68: return -0.12111485749483109f;
  t297_n69: return 0.10383682698011398f;
  t297_n70: if (E.getFractionNameInContext() >= 3186245525 /*0.11437908560037613*/) goto t297_n80;
  t297_n71: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t297_n73;
  t297_n72: return -0.1878638118505478f;
  t297_n73: if (E.getNumReferences() >= 3282567168 /*336.0*/) goto t297_n79;
  t297_n74: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n78;
  t297_n75: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t297_n77;
  t297_n76: return -0.7003508806228638f;
  t297_n77: return -0.21790462732315063f;
  t297_n78: return 0.051654551178216934f;
  t297_n79: return -1.5273956060409546f;
  t297_n80: return -0.07271773368120193f;
  t297_n81: if (E.getNumReferences() >= 3279421440 /*248.0*/) goto t297_n99;
  t297_n82: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n90;
  t297_n83: if (E.getNumReferences() >= 3274833920 /*178.0*/) goto t297_n89;
  t297_n84: if (E.getNumReferences() >= 3265724416 /*83.5*/) goto t297_n88;
  t297_n85: if (E.getNumReferences() >= 3260547072 /*54.0*/) goto t297_n87;
  t297_n86: return -0.16768556833267212f;
  t297_n87: return -0.5788741111755371f;
  t297_n88: return -0.012046139687299728f;
  t297_n89: return -0.5685726404190063f;
  t297_n90: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t297_n94;
  t297_n91: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t297_n93;
  t297_n92: return -0.008047064766287804f;
  t297_n93: return 0.10731977224349976f;
  t297_n94: if (E.getNumReferences() >= 3267887104 /*100.0*/) goto t297_n96;
  t297_n95: return -0.043504584580659866f;
  t297_n96: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n98;
  t297_n97: return -0.3084180951118469f;
  t297_n98: return -0.08393162488937378f;
  t297_n99: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t297_n101;
  t297_n100: return -0.17402121424674988f;
  t297_n101: return 0.10589097440242767f;
  t297_n102: if (E.getNumReferences() >= 3273621504 /*159.5*/) goto t297_n126;
  t297_n103: if (E.getNumReferences() >= 3266117632 /*86.5*/) goto t297_n105;
  t297_n104: return -0.025028862059116364f;
  t297_n105: if (E.getNumReferences() >= 3268608000 /*105.5*/) goto t297_n121;
  t297_n106: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel))) goto t297_n112;
  t297_n107: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t297_n109;
  t297_n108: return -0.02750856801867485f;
  t297_n109: if (E.getNumReferences() >= 3267690496 /*98.5*/) goto t297_n111;
  t297_n110: return 0.06024286895990372f;
  t297_n111: return 0.17031744122505188f;
  t297_n112: if (E.getNumReferences() >= 3268476928 /*104.5*/) goto t297_n120;
  t297_n113: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t297_n119;
  t297_n114: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t297_n116;
  t297_n115: return 0.02146754041314125f;
  t297_n116: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t297_n118;
  t297_n117: return 0.10514731705188751f;
  t297_n118: return 0.13096722960472107f;
  t297_n119: return 0.22511735558509827f;
  t297_n120: return 0.1632101833820343f;
  t297_n121: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t297_n123;
  t297_n122: return -0.1166263073682785f;
  t297_n123: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t297_n125;
  t297_n124: return 0.031177712604403496f;
  t297_n125: return 0.13814754784107208f;
  t297_n126: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n130;
  t297_n127: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t297_n129;
  t297_n128: return -0.15846489369869232f;
  t297_n129: return 0.17483390867710114f;
  t297_n130: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t297_n132;
  t297_n131: return -0.012031342834234238f;
  t297_n132: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t297_n134;
  t297_n133: return -1.0010098218917847f;
  t297_n134: return -0.08454762399196625f;
  t297_n135: if (E.getNumReferences() >= 3284910080 /*407.5*/) goto t297_n145;
  t297_n136: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t297_n144;
  t297_n137: if (E.getNumReferences() >= 3284828160 /*405.0*/) goto t297_n141;
  t297_n138: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t297_n140;
  t297_n139: return 0.06181136518716812f;
  t297_n140: return 0.11487694829702377f;
  t297_n141: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type))) goto t297_n143;
  t297_n142: return -0.004677449353039265f;
  t297_n143: return 0.15810289978981018f;
  t297_n144: return 0.06728598475456238f;
  t297_n145: if (E.getFractionNameInContext() >= 3167538381 /*0.02500000037252903*/) goto t297_n151;
  t297_n146: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n150;
  t297_n147: if (E.getNumReferences() >= 3285499904 /*425.5*/) goto t297_n149;
  t297_n148: return 0.14779603481292725f;
  t297_n149: return -0.05738130956888199f;
  t297_n150: return -0.0030666403472423553f;
  t297_n151: if (E.getNumReferences() >= 3315510272 /*5077.5*/) goto t297_n159;
  t297_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Type))) goto t297_n156;
  t297_n153: if (E.getFractionNameInContext() >= 3190975644 /*0.17424243688583374*/) goto t297_n155;
  t297_n154: return 0.021170955151319504f;
  t297_n155: return -0.03910551592707634f;
  t297_n156: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t297_n158;
  t297_n157: return 0.11076421290636063f;
  t297_n158: return 0.10218308120965958f;
  t297_n159: if (E.getNumReferences() >= 3322851328 /*9132.0*/) goto t297_n163;
  t297_n160: if (E.getNumReferences() >= 3322442752 /*8733.0*/) goto t297_n162;
  t297_n161: return 0.09300302714109421f;
  t297_n162: return 0.1451912373304367f;
  t297_n163: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t297_n165;
  t297_n164: return 0.04138734191656113f;
  t297_n165: return -0.2789817154407501f;
  t297_n166: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t297_n176;
  t297_n167: if (E.getNumReferences() >= 3284221952 /*386.5*/) goto t297_n169;
  t297_n168: return 0.002764645731076598f;
  t297_n169: if (E.getNumReferences() >= 3287318528 /*481.0*/) goto t297_n175;
  t297_n170: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t297_n174;
  t297_n171: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t297_n173;
  t297_n172: return 0.11973293125629425f;
  t297_n173: return 0.16583319008350372f;
  t297_n174: return 0.03539631515741348f;
  t297_n175: return 0.029987385496497154f;
  t297_n176: if (E.getNumReferences() >= 3273359360 /*155.5*/) goto t297_n178;
  t297_n177: return 0.12631849944591522f;
  t297_n178: return -0.0073180124163627625f;
  t297_n179: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t297_n181;
  t297_n180: return -0.027746915817260742f;
  t297_n181: if (E.getNumReferences() >= 3305209856 /*2072.0*/) goto t297_n233;
  t297_n182: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol))) goto t297_n184;
  t297_n183: return -0.05914588272571564f;
  t297_n184: if (E.getNumReferences() >= 3282370560 /*330.0*/) goto t297_n218;
  t297_n185: if (E.getNumReferences() >= 3280896000 /*285.0*/) goto t297_n217;
  t297_n186: if (E.getNumReferences() >= 3257532416 /*42.5*/) goto t297_n188;
  t297_n187: return -0.11536994576454163f;
  t297_n188: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t297_n212;
  t297_n189: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t297_n195;
  t297_n190: if (E.getNumReferences() >= 3260547072 /*54.0*/) goto t297_n192;
  t297_n191: return 0.0988236516714096f;
  t297_n192: if (E.getNumReferences() >= 3280486400 /*272.5*/) goto t297_n194;
  t297_n193: return -0.02187265455722809f;
  t297_n194: return 0.17781947553157806f;
  t297_n195: if (E.getNumReferences() >= 3264020480 /*70.5*/) goto t297_n197;
  t297_n196: return 0.04295538738369942f;
  t297_n197: if (E.getNumReferences() >= 3265069056 /*78.5*/) goto t297_n201;
  t297_n198: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t297_n200;
  t297_n199: return 0.26144739985466003f;
  t297_n200: return -0.11110593378543854f;
  t297_n201: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t297_n203;
  t297_n202: return 0.04042457044124603f;
  t297_n203: if (E.getNumReferences() >= 3272376320 /*140.5*/) goto t297_n205;
  t297_n204: return 0.019264575093984604f;
  t297_n205: if (E.getNumReferences() >= 3277881344 /*224.5*/) goto t297_n211;
  t297_n206: if (E.getNumReferences() >= 3277684736 /*221.5*/) goto t297_n210;
  t297_n207: if (E.getFileProximityDistanceCost() >= 3244294144 /*14.0*/) goto t297_n209;
  t297_n208: return -0.015340505167841911f;
  t297_n209: return 0.14134478569030762f;
  t297_n210: return 0.2335490584373474f;
  t297_n211: return -0.10042140632867813f;
  t297_n212: if (E.getNumReferences() >= 3260678144 /*54.5*/) goto t297_n214;
  t297_n213: return 0.15121924877166748f;
  t297_n214: if (E.getNumReferences() >= 3270770688 /*122.0*/) goto t297_n216;
  t297_n215: return 0.030217887833714485f;
  t297_n216: return 0.11428221315145493f;
  t297_n217: return -0.26706916093826294f;
  t297_n218: if (E.getNumReferences() >= 3284074496 /*382.0*/) goto t297_n222;
  t297_n219: if (E.getNumReferences() >= 3283042304 /*350.5*/) goto t297_n221;
  t297_n220: return 0.11927229166030884f;
  t297_n221: return 0.13177724182605743f;
  t297_n222: if (E.getNumReferences() >= 3287629824 /*490.5*/) goto t297_n224;
  t297_n223: return -0.12774084508419037f;
  t297_n224: if (E.getNumReferences() >= 3287924736 /*499.5*/) goto t297_n226;
  t297_n225: return 0.2692658007144928f;
  t297_n226: if (E.getNumReferences() >= 3290832896 /*664.5*/) goto t297_n228;
  t297_n227: return 0.14379781484603882f;
  t297_n228: if (E.getNumReferences() >= 3296960512 /*1053.0*/) goto t297_n230;
  t297_n229: return 0.015248280018568039f;
  t297_n230: if (E.getNumReferences() >= 3300208640 /*1449.5*/) goto t297_n232;
  t297_n231: return 0.11032340675592422f;
  t297_n232: return 0.12127300351858139f;
  t297_n233: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t297_n235;
  t297_n234: return -0.059871431440114975f;
  t297_n235: return -0.30246877670288086f;
  t297_n236: return 0.0021668740082532167f;
  t297_n237: if (E.getNumReferences() >= 3338331904 /*32115.5*/) goto t297_n247;
  t297_n238: if (E.getNumReferences() >= 3334006272 /*23667.0*/) goto t297_n246;
  t297_n239: if (E.getNumReferences() >= 3333976064 /*23608.0*/) goto t297_n241;
  t297_n240: return -0.009582629427313805f;
  t297_n241: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t297_n243;
  t297_n242: return -0.24430054426193237f;
  t297_n243: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t297_n245;
  t297_n244: return 0.12562298774719238f;
  t297_n245: return 0.1064373180270195f;
  t297_n246: return -0.044380225241184235f;
  t297_n247: return 0.004720400087535381f;
  t297_n248: if (E.getNumReferences() >= 3257008128 /*40.5*/) goto t297_n254;
  t297_n249: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t297_n253;
  t297_n250: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t297_n252;
  t297_n251: return 0.06709620356559753f;
  t297_n252: return 0.11524037271738052f;
  t297_n253: return 0.01465646829456091f;
  t297_n254: return 0.031030455604195595f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree298(const Example& E) {
  t298_n0: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t298_n2;
  t298_n1: return -0.00012720409722533077f;
  t298_n2: if (E.getNumReferences() >= 3333012480 /*21726.0*/) goto t298_n252;
  t298_n3: if (E.getNumReferences() >= 3332287488 /*20310.0*/) goto t298_n249;
  t298_n4: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression))) goto t298_n230;
  t298_n5: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Variable))) goto t298_n229;
  t298_n6: if (E.getFileProximityDistanceCost() >= 3247702016 /*18.5*/) goto t298_n140;
  t298_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n21;
  t298_n8: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t298_n10;
  t298_n9: return 0.11651585251092911f;
  t298_n10: if (E.getNumReferences() >= 3295715328 /*962.5*/) goto t298_n16;
  t298_n11: if (E.getNumReferences() >= 3295363072 /*941.0*/) goto t298_n13;
  t298_n12: return -0.002286854898557067f;
  t298_n13: if (E.getFileProximityDistanceCost() >= 3245867008 /*15.5*/) goto t298_n15;
  t298_n14: return -0.2843558192253113f;
  t298_n15: return -1.361624836921692f;
  t298_n16: if (E.getNumReferences() >= 3296468992 /*1008.5*/) goto t298_n20;
  t298_n17: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t298_n19;
  t298_n18: return 0.1233784556388855f;
  t298_n19: return 0.06243206933140755f;
  t298_n20: return 0.011524008587002754f;
  t298_n21: if (E.getNumReferences() >= 3294879744 /*911.5*/) goto t298_n135;
  t298_n22: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t298_n128;
  t298_n23: if (E.getNumReferences() >= 3292749824 /*781.5*/) goto t298_n125;
  t298_n24: if (E.getNumReferences() >= 3292028928 /*737.5*/) goto t298_n124;
  t298_n25: if (E.getFractionNameInContext() >= 3210589916 /*0.8660714626312256*/) goto t298_n123;
  t298_n26: if (E.getFractionNameInContext() >= 3209761041 /*0.8166666626930237*/) goto t298_n116;
  t298_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Type))) goto t298_n31;
  t298_n28: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t298_n30;
  t298_n29: return -0.0019924561493098736f;
  t298_n30: return 0.016371915116906166f;
  t298_n31: if (E.getFileProximityDistanceCost() >= 3234856960 /*6.5*/) goto t298_n55;
  t298_n32: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Type))) goto t298_n50;
  t298_n33: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t298_n49;
  t298_n34: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t298_n38;
  t298_n35: if (E.getFractionNameInContext() >= 3197866426 /*0.30384618043899536*/) goto t298_n37;
  t298_n36: return 0.001981783425435424f;
  t298_n37: return -0.007265347056090832f;
  t298_n38: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t298_n48;
  t298_n39: if (E.getFractionNameInContext() >= 3201572162 /*0.41428571939468384*/) goto t298_n41;
  t298_n40: return -0.0019458995666354895f;
  t298_n41: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t298_n47;
  t298_n42: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t298_n44;
  t298_n43: return 0.05872517079114914f;
  t298_n44: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t298_n46;
  t298_n45: return 0.135545551776886f;
  t298_n46: return 0.004936794750392437f;
  t298_n47: return 0.0063592828810215f;
  t298_n48: return 0.0583641454577446f;
  t298_n49: return 0.12034869939088821f;
  t298_n50: if (E.getNumNameInContext() >= 3232759808 /*5.5*/) goto t298_n54;
  t298_n51: if (E.getNumReferences() >= 3263037440 /*63.5*/) goto t298_n53;
  t298_n52: return 0.010517234914004803f;
  t298_n53: return 0.08995116502046585f;
  t298_n54: return 0.1000923365354538f;
  t298_n55: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Expression))) goto t298_n91;
  t298_n56: if (E.getFractionNameInContext() >= 3189238143 /*0.14835165441036224*/) goto t298_n72;
  t298_n57: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t298_n65;
  t298_n58: if (E.getNumReferences() >= 3257401344 /*42.0*/) goto t298_n64;
  t298_n59: if (E.getSymbolScopeDistanceCost() >= 3239051264 /*9.0*/) goto t298_n63;
  t298_n60: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t298_n62;
  t298_n61: return 0.17658868432044983f;
  t298_n62: return -0.003906669560819864f;
  t298_n63: return 0.11050523817539215f;
  t298_n64: return 0.13039380311965942f;
  t298_n65: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t298_n67;
  t298_n66: return -0.035421814769506454f;
  t298_n67: if (E.getNumReferences() >= 3251896320 /*26.5*/) goto t298_n71;
  t298_n68: if (E.getFileProximityDistanceCost() >= 3243245568 /*13.0*/) goto t298_n70;
  t298_n69: return 0.12234776467084885f;
  t298_n70: return -0.06796687096357346f;
  t298_n71: return 0.34077930450439453f;
  t298_n72: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t298_n82;
  t298_n73: if (E.getNumReferences() >= 3272343552 /*140.0*/) goto t298_n75;
  t298_n74: return -0.007655052468180656f;
  t298_n75: if (E.getNumReferences() >= 3282649088 /*338.5*/) goto t298_n81;
  t298_n76: if (E.getFractionNameInContext() >= 3193449859 /*0.21111111342906952*/) goto t298_n78;
  t298_n77: return -0.6226409673690796f;
  t298_n78: if (E.getNumReferences() >= 3281895424 /*315.5*/) goto t298_n80;
  t298_n79: return -0.08847478032112122f;
  t298_n80: return -0.8452696800231934f;
  t298_n81: return 0.05218873172998428f;
  t298_n82: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t298_n90;
  t298_n83: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassStructUnion))) goto t298_n85;
  t298_n84: return 0.017302313819527626f;
  t298_n85: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t298_n87;
  t298_n86: return 0.007210799492895603f;
  t298_n87: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t298_n89;
  t298_n88: return 0.08910952508449554f;
  t298_n89: return 0.13023339211940765f;
  t298_n90: return 0.005966537166386843f;
  t298_n91: if (E.getNumReferences() >= 3284451328 /*393.5*/) goto t298_n109;
  t298_n92: if (E.getFractionNameInContext() >= 3193536758 /*0.2124060094356537*/) goto t298_n94;
  t298_n93: return 0.010405868291854858f;
  t298_n94: if (E.getFractionNameInContext() >= 3193929208 /*0.2182539701461792*/) goto t298_n108;
  t298_n95: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t298_n101;
  t298_n96: if (E.getNumReferences() >= 3230662656 /*4.5*/) goto t298_n100;
  t298_n97: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t298_n99;
  t298_n98: return -0.10972248762845993f;
  t298_n99: return 0.16622856259346008f;
  t298_n100: return 0.11200828105211258f;
  t298_n101: if (E.getFileProximityDistanceCost() >= 3242721280 /*12.5*/) goto t298_n107;
  t298_n102: if (E.getFileProximityDistanceCost() >= 3241672704 /*11.5*/) goto t298_n104;
  t298_n103: return 0.08774050325155258f;
  t298_n104: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t298_n106;
  t298_n105: return -0.10149089992046356f;
  t298_n106: return 0.12160718441009521f;
  t298_n107: return 0.07255212217569351f;
  t298_n108: return 0.019438592717051506f;
  t298_n109: if (E.getNumReferences() >= 3285303296 /*419.5*/) goto t298_n115;
  t298_n110: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t298_n114;
  t298_n111: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t298_n113;
  t298_n112: return 0.14084501564502716f;
  t298_n113: return 0.1212826743721962f;
  t298_n114: return 0.11496259272098541f;
  t298_n115: return 0.053612664341926575f;
  t298_n116: if (E.getFileProximityDistanceCost() >= 3235905536 /*7.0*/) goto t298_n120;
  t298_n117: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n119;
  t298_n118: return -0.007253664080053568f;
  t298_n119: return 0.1088143065571785f;
  t298_n120: if (E.getFileProximityDistanceCost() >= 3238526976 /*8.5*/) goto t298_n122;
  t298_n121: return 0.11890292167663574f;
  t298_n122: return -0.1045817956328392f;
  t298_n123: return -0.025143345817923546f;
  t298_n124: return -0.16059575974941254f;
  t298_n125: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t298_n127;
  t298_n126: return 0.102462999522686f;
  t298_n127: return 0.05644388124346733f;
  t298_n128: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t298_n132;
  t298_n129: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t298_n131;
  t298_n130: return -0.009201622568070889f;
  t298_n131: return 0.009336247108876705f;
  t298_n132: if (E.getFractionNameInContext() >= 3191304608 /*0.17914438247680664*/) goto t298_n134;
  t298_n133: return -0.02943468652665615f;
  t298_n134: return 0.015428273007273674f;
  t298_n135: if (E.getNumReferences() >= 3320421376 /*7475.5*/) goto t298_n139;
  t298_n136: if (E.getNumReferences() >= 3319669760 /*7108.5*/) goto t298_n138;
  t298_n137: return -0.03159508481621742f;
  t298_n138: return -0.1829959899187088f;
  t298_n139: return 0.02327469177544117f;
  t298_n140: if (E.getNumReferences() >= 3318651904 /*6611.5*/) goto t298_n226;
  t298_n141: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t298_n159;
  t298_n142: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n158;
  t298_n143: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t298_n157;
  t298_n144: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t298_n152;
  t298_n145: if (E.getFractionNameInContext() >= 3203802978 /*0.48076921701431274*/) goto t298_n151;
  t298_n146: if (E.getFileProximityDistanceCost() >= 3472883712 /*2147483648.0*/) goto t298_n150;
  t298_n147: if (E.getSymbolScopeDistanceCost() >= 3225419776 /*3.0*/) goto t298_n149;
  t298_n148: return 0.05635109171271324f;
  t298_n149: return -0.05850893259048462f;
  t298_n150: return -0.028697824105620384f;
  t298_n151: return 0.0035142838023602962f;
  t298_n152: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n154;
  t298_n153: return -0.1035672202706337f;
  t298_n154: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t298_n156;
  t298_n155: return 0.6840763688087463f;
  t298_n156: return -0.1002112627029419f;
  t298_n157: return -0.0005745849339291453f;
  t298_n158: return -0.005086633842438459f;
  t298_n159: if (E.getFractionNameInContext() >= 3183385950 /*0.09307359158992767*/) goto t298_n161;
  t298_n160: return 0.10574447363615036f;
  t298_n161: if (E.getFractionNameInContext() >= 3189109088 /*0.14642858505249023*/) goto t298_n195;
  t298_n162: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_Type))) goto t298_n188;
  t298_n163: if (E.getFractionNameInContext() >= 3188651527 /*0.1396103948354721*/) goto t298_n175;
  t298_n164: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t298_n170;
  t298_n165: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t298_n167;
  t298_n166: return -0.015284799970686436f;
  t298_n167: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t298_n169;
  t298_n168: return -0.022473420947790146f;
  t298_n169: return -0.11442946642637253f;
  t298_n170: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Statement))) goto t298_n172;
  t298_n171: return 0.012953737750649452f;
  t298_n172: if (E.getFractionNameInContext() >= 3187177592 /*0.12132352590560913*/) goto t298_n174;
  t298_n173: return 0.1136232390999794f;
  t298_n174: return 0.17776106297969818f;
  t298_n175: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t298_n187;
  t298_n176: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n178;
  t298_n177: return -0.042208172380924225f;
  t298_n178: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t298_n180;
  t298_n179: return 0.03744969144463539f;
  t298_n180: if (E.getScope() & (BIT(Scope_type::FileScope))) goto t298_n186;
  t298_n181: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t298_n183;
  t298_n182: return 0.031373415142297745f;
  t298_n183: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess))) goto t298_n185;
  t298_n184: return 0.04781370609998703f;
  t298_n185: return 0.11532542109489441f;
  t298_n186: return 0.12359613925218582f;
  t298_n187: return -0.006884464994072914f;
  t298_n188: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t298_n194;
  t298_n189: if (E.getFractionNameInContext() >= 3187950660 /*0.12916666269302368*/) goto t298_n191;
  t298_n190: return 0.00892975926399231f;
  t298_n191: if (E.getIsInstanceMember() >= 3204448256 /*0.5*/) goto t298_n193;
  t298_n192: return -0.10267515480518341f;
  t298_n193: return 0.06131017953157425f;
  t298_n194: return 0.023125408217310905f;
  t298_n195: if (E.getFractionNameInContext() >= 3190172906 /*0.16228070855140686*/) goto t298_n207;
  t298_n196: if (E.getIsDeprecated() >= 3204448256 /*0.5*/) goto t298_n200;
  t298_n197: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t298_n199;
  t298_n198: return -0.023560497909784317f;
  t298_n199: return 0.02347460761666298f;
  t298_n200: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n202;
  t298_n201: return -0.12577423453330994f;
  t298_n202: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t298_n204;
  t298_n203: return 0.07975086569786072f;
  t298_n204: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t298_n206;
  t298_n205: return 0.31292545795440674f;
  t298_n206: return -0.109196238219738f;
  t298_n207: if (E.getFractionNameInContext() >= 3190796208 /*0.1715686321258545*/) goto t298_n225;
  t298_n208: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t298_n212;
  t298_n209: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t298_n211;
  t298_n210: return -0.010752484202384949f;
  t298_n211: return -0.10484608262777328f;
  t298_n212: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t298_n224;
  t298_n213: if (E.getContextKind() & (BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type))) goto t298_n215;
  t298_n214: return 0.011973950080573559f;
  t298_n215: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t298_n221;
  t298_n216: if (E.getIsInBaseClass() >= 3204448256 /*0.5*/) goto t298_n218;
  t298_n217: return 0.0232253298163414f;
  t298_n218: if (E.getSymbolScopeDistanceCost() >= 3472883712 /*2147483648.0*/) goto t298_n220;
  t298_n219: return 0.04826255887746811f;
  t298_n220: return 0.12255699932575226f;
  t298_n221: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FunctionScope)|BIT(Scope_type::GlobalScope))) goto t298_n223;
  t298_n222: return -0.10074739158153534f;
  t298_n223: return 0.15019452571868896f;
  t298_n224: return 0.0018254911992698908f;
  t298_n225: return -0.0005896731745451689f;
  t298_n226: if (E.getNumReferences() >= 3319035904 /*6799.0*/) goto t298_n228;
  t298_n227: return -0.6611794829368591f;
  t298_n228: return -0.09756159037351608f;
  t298_n229: return 0.0026526569854468107f;
  t298_n230: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t298_n246;
  t298_n231: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t298_n239;
  t298_n232: if (E.getSymbolScopeDistanceCost() >= 3221225472 /*2.0*/) goto t298_n234;
  t298_n233: return -0.14459455013275146f;
  t298_n234: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t298_n236;
  t298_n235: return -0.06471586972475052f;
  t298_n236: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t298_n238;
  t298_n237: return 0.12346936762332916f;
  t298_n238: return -0.1001800149679184f;
  t298_n239: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t298_n243;
  t298_n240: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t298_n242;
  t298_n241: return 0.05811610445380211f;
  t298_n242: return 0.10521510988473892f;
  t298_n243: if (E.getFractionNameInContext() >= 3205047442 /*0.5357142686843872*/) goto t298_n245;
  t298_n244: return -0.12863685190677643f;
  t298_n245: return 0.06751266121864319f;
  t298_n246: if (E.getFractionNameInContext() >= 3206255034 /*0.6076923608779907*/) goto t298_n248;
  t298_n247: return 0.017428813502192497f;
  t298_n248: return -0.055605534464120865f;
  t298_n249: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t298_n251;
  t298_n250: return -0.1680881679058075f;
  t298_n251: return -0.8019477128982544f;
  t298_n252: if (E.getNumReferences() >= 3333101824 /*21900.5*/) goto t298_n254;
  t298_n253: return 0.11823438853025436f;
  t298_n254: return 0.015042753890156746f;
}

LLVM_ATTRIBUTE_NOINLINE float EvaluateTree299(const Example& E) {
  t299_n0: if (E.getTypeMatchesPreferred() >= 3204448256 /*0.5*/) goto t299_n252;
  t299_n1: if (E.getSymbolScopeDistanceCost() >= 3231711232 /*5.0*/) goto t299_n251;
  t299_n2: if (E.getNumReferences() >= 3217031168 /*1.5*/) goto t299_n6;
  t299_n3: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n5;
  t299_n4: return -0.0023861825466156006f;
  t299_n5: return 0.0005279325414448977f;
  t299_n6: if (E.getHadContextType() >= 3204448256 /*0.5*/) goto t299_n250;
  t299_n7: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n49;
  t299_n8: if (E.getFractionNameInContext() >= 3177856917 /*0.057189542800188065*/) goto t299_n48;
  t299_n9: if (E.getFileProximityDistanceCost() >= 3227516928 /*3.5*/) goto t299_n21;
  t299_n10: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type))) goto t299_n16;
  t299_n11: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n13;
  t299_n12: return -0.06488648056983948f;
  t299_n13: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t299_n15;
  t299_n14: return 0.1024707704782486f;
  t299_n15: return 0.006586769595742226f;
  t299_n16: if (E.getContextKind() & (BIT(ContextKind_type::CCC_TopLevel))) goto t299_n18;
  t299_n17: return 0.008908073417842388f;
  t299_n18: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t299_n20;
  t299_n19: return 0.029052870348095894f;
  t299_n20: return -0.02768993377685547f;
  t299_n21: if (E.getIsImplementationDetail() >= 3204448256 /*0.5*/) goto t299_n41;
  t299_n22: if (E.getSemaFileProximityScore() >= 3209481421 /*0.800000011920929*/) goto t299_n40;
  t299_n23: if (E.getNumReferences() >= 3245867008 /*15.5*/) goto t299_n27;
  t299_n24: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t299_n26;
  t299_n25: return -0.14242520928382874f;
  t299_n26: return -0.028134305030107498f;
  t299_n27: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t299_n39;
  t299_n28: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n30;
  t299_n29: return -0.021450268104672432f;
  t299_n30: if (E.getFileProximityDistanceCost() >= 3232759808 /*5.5*/) goto t299_n38;
  t299_n31: if (E.getNumReferences() >= 3249799168 /*22.5*/) goto t299_n35;
  t299_n32: if (E.getNumReferences() >= 3249274880 /*21.5*/) goto t299_n34;
  t299_n33: return 0.09126109629869461f;
  t299_n34: return 0.12633001804351807f;
  t299_n35: if (E.getNumReferences() >= 3256221696 /*37.5*/) goto t299_n37;
  t299_n36: return -0.07003024965524673f;
  t299_n37: return 0.04855829477310181f;
  t299_n38: return 0.007690619211643934f;
  t299_n39: return 0.024739732965826988f;
  t299_n40: return -0.07771000266075134f;
  t299_n41: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ParenthesizedExpression))) goto t299_n47;
  t299_n42: if (E.getNumReferences() >= 3315609600 /*5126.0*/) goto t299_n44;
  t299_n43: return -0.08703737705945969f;
  t299_n44: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Expression)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Statement)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n46;
  t299_n45: return -0.08404891937971115f;
  t299_n46: return 0.13552619516849518f;
  t299_n47: return 0.11743475496768951f;
  t299_n48: return -0.008180384524166584f;
  t299_n49: if (E.getNumNameInContext() >= 3227516928 /*3.5*/) goto t299_n243;
  t299_n50: if (E.getHadSymbolType() >= 3204448256 /*0.5*/) goto t299_n190;
  t299_n51: if (E.getContextKind() & (BIT(ContextKind_type::CCC_ArrowMemberAccess)|BIT(ContextKind_type::CCC_ClassOrStructTag)|BIT(ContextKind_type::CCC_ClassStructUnion)|BIT(ContextKind_type::CCC_DotMemberAccess)|BIT(ContextKind_type::CCC_EnumTag)|BIT(ContextKind_type::CCC_Namespace)|BIT(ContextKind_type::CCC_Other)|BIT(ContextKind_type::CCC_ParenthesizedExpression)|BIT(ContextKind_type::CCC_PreprocessorExpression)|BIT(ContextKind_type::CCC_Symbol)|BIT(ContextKind_type::CCC_SymbolOrNewName)|BIT(ContextKind_type::CCC_TopLevel)|BIT(ContextKind_type::CCC_Type)|BIT(ContextKind_type::CCC_TypeQualifiers)|BIT(ContextKind_type::CCC_UnionTag))) goto t299_n63;
  t299_n52: if (E.getNumReferences() >= 3274145792 /*167.5*/) goto t299_n62;
  t299_n53: if (E.getFileProximityDistanceCost() >= 3252420608 /*27.5*/) goto t299_n55;
  t299_n54: return 0.00022243650164455175f;
  t299_n55: if (E.getNumReferences() >= 3267166208 /*94.5*/) goto t299_n61;
  t299_n56: if (E.getNumReferences() >= 3238526976 /*8.5*/) goto t299_n58;
  t299_n57: return -0.2999655306339264f;
  t299_n58: if (E.getNumReferences() >= 3242721280 /*12.5*/) goto t299_n60;
  t299_n59: return 0.15266987681388855f;
  t299_n60: return 0.13541753590106964f;
  t299_n61: return 0.543752133846283f;
  t299_n62: return -0.014369897544384003f;
  t299_n63: if (E.getNumReferences() >= 3247702016 /*18.5*/) goto t299_n129;
  t299_n64: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t299_n68;
  t299_n65: if (E.getNumReferences() >= 3241672704 /*11.5*/) goto t299_n67;
  t299_n66: return 0.011975952424108982f;
  t299_n67: return 0.023076321929693222f;
  t299_n68: if (E.getFileProximityDistanceCost() >= 3247177728 /*17.5*/) goto t299_n114;
  t299_n69: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t299_n107;
  t299_n70: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t299_n106;
  t299_n71: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t299_n97;
  t299_n72: if (E.getFractionNameInContext() >= 3207943510 /*0.7083333730697632*/) goto t299_n96;
  t299_n73: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t299_n95;
  t299_n74: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t299_n92;
  t299_n75: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t299_n77;
  t299_n76: return -0.036624472588300705f;
  t299_n77: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t299_n83;
  t299_n78: if (E.getNumReferences() >= 3232759808 /*5.5*/) goto t299_n82;
  t299_n79: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t299_n81;
  t299_n80: return 0.0975385382771492f;
  t299_n81: return -0.0600242018699646f;
  t299_n82: return -0.8315944671630859f;
  t299_n83: if (E.getFractionNameInContext() >= 3183705516 /*0.09545454382896423*/) goto t299_n85;
  t299_n84: return 0.13921503722667694f;
  t299_n85: if (E.getFractionNameInContext() >= 3186738972 /*0.1180555522441864*/) goto t299_n87;
  t299_n86: return -0.49002164602279663f;
  t299_n87: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t299_n91;
  t299_n88: if (E.getFractionNameInContext() >= 3197497696 /*0.29285717010498047*/) goto t299_n90;
  t299_n89: return 0.062395840883255005f;
  t299_n90: return 0.10803288966417313f;
  t299_n91: return -0.016220178455114365f;
  t299_n92: if (E.getFileProximityDistanceCost() >= 3236954112 /*7.5*/) goto t299_n94;
  t299_n93: return 0.021440409123897552f;
  t299_n94: return -0.028124704957008362f;
  t299_n95: return -0.19034990668296814f;
  t299_n96: return 0.07715630531311035f;
  t299_n97: if (E.getNumReferences() >= 3223322624 /*2.5*/) goto t299_n103;
  t299_n98: if (E.getFractionNameInContext() >= 3185383238 /*0.10795454680919647*/) goto t299_n100;
  t299_n99: return -2.793297529220581f;
  t299_n100: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t299_n102;
  t299_n101: return 0.07931818068027496f;
  t299_n102: return -0.5115524530410767f;
  t299_n103: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t299_n105;
  t299_n104: return 0.06951530277729034f;
  t299_n105: return -0.04985569790005684f;
  t299_n106: return 0.05072949081659317f;
  t299_n107: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t299_n111;
  t299_n108: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t299_n110;
  t299_n109: return 0.09595001488924026f;
  t299_n110: return -0.12144060432910919f;
  t299_n111: if (E.getNumReferences() >= 3234856960 /*6.5*/) goto t299_n113;
  t299_n112: return 0.12285080552101135f;
  t299_n113: return 0.07340356707572937f;
  t299_n114: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t299_n124;
  t299_n115: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Function)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t299_n123;
  t299_n116: if (E.getFractionNameInContext() >= 3195127580 /*0.2361111044883728*/) goto t299_n120;
  t299_n117: if (E.getFileProximityDistanceCost() >= 3248750592 /*20.5*/) goto t299_n119;
  t299_n118: return -0.6132120490074158f;
  t299_n119: return -2.226938009262085f;
  t299_n120: if (E.getFractionNameInContext() >= 3199554902 /*0.3541666865348816*/) goto t299_n122;
  t299_n121: return 0.12390480935573578f;
  t299_n122: return -0.8184543251991272f;
  t299_n123: return 0.0f;
  t299_n124: if (E.getFractionNameInContext() >= 3200673382 /*0.38749998807907104*/) goto t299_n126;
  t299_n125: return -0.07405634969472885f;
  t299_n126: if (E.getNumReferences() >= 3241148416 /*11.0*/) goto t299_n128;
  t299_n127: return 0.08005435764789581f;
  t299_n128: return -0.053706567734479904f;
  t299_n129: if (E.getNumReferences() >= 3324873728 /*11107.0*/) goto t299_n189;
  t299_n130: if (E.getNumReferences() >= 3321837568 /*8167.0*/) goto t299_n170;
  t299_n131: if (E.getNumReferences() >= 3317905408 /*6247.0*/) goto t299_n145;
  t299_n132: if (E.getNumReferences() >= 3309791232 /*3190.5*/) goto t299_n144;
  t299_n133: if (E.getNumReferences() >= 3308636160 /*2908.5*/) goto t299_n135;
  t299_n134: return 0.0009790739277377725f;
  t299_n135: if (E.getNumReferences() >= 3308941312 /*2983.0*/) goto t299_n137;
  t299_n136: return 0.10414186865091324f;
  t299_n137: if (E.getSemaSaysInScope() >= 3204448256 /*0.5*/) goto t299_n141;
  t299_n138: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t299_n140;
  t299_n139: return 0.10860772430896759f;
  t299_n140: return 0.005822684150189161f;
  t299_n141: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Function))) goto t299_n143;
  t299_n142: return 0.07634374499320984f;
  t299_n143: return 0.11660820245742798f;
  t299_n144: return -0.02144465036690235f;
  t299_n145: if (E.getNumReferences() >= 3318672384 /*6621.5*/) goto t299_n165;
  t299_n146: if (E.getFileProximityDistanceCost() >= 3249799168 /*22.5*/) goto t299_n164;
  t299_n147: if (E.getNumReferences() >= 3318611968 /*6592.0*/) goto t299_n163;
  t299_n148: if (E.getFractionNameInContext() >= 3196618889 /*0.2666666805744171*/) goto t299_n162;
  t299_n149: if (E.getNumReferences() >= 3318157312 /*6370.0*/) goto t299_n153;
  t299_n150: if (E.getFileProximityDistanceCost() >= 3246653440 /*16.5*/) goto t299_n152;
  t299_n151: return 0.05634349212050438f;
  t299_n152: return 0.10359778255224228f;
  t299_n153: if (E.getNumReferences() >= 3318276096 /*6428.0*/) goto t299_n157;
  t299_n154: if (E.getFileProximityDistanceCost() >= 3243769856 /*13.5*/) goto t299_n156;
  t299_n155: return -0.8547974228858948f;
  t299_n156: return -0.14206628501415253f;
  t299_n157: if (E.getFileProximityDistanceCost() >= 3244818432 /*14.5*/) goto t299_n161;
  t299_n158: if (E.getNumReferences() >= 3318469632 /*6522.5*/) goto t299_n160;
  t299_n159: return 0.011830820702016354f;
  t299_n160: return 0.13008402287960052f;
  t299_n161: return -0.017511190846562386f;
  t299_n162: return 0.1540624350309372f;
  t299_n163: return 0.11245620250701904f;
  t299_n164: return 0.19629091024398804f;
  t299_n165: if (E.getFileProximityDistanceCost() >= 3239575552 /*9.5*/) goto t299_n169;
  t299_n166: if (E.getNumReferences() >= 3320292352 /*7412.5*/) goto t299_n168;
  t299_n167: return 0.051628630608320236f;
  t299_n168: return 0.1273808628320694f;
  t299_n169: return 0.005767251830548048f;
  t299_n170: if (E.getNumReferences() >= 3321896960 /*8200.0*/) goto t299_n172;
  t299_n171: return -0.3672906160354614f;
  t299_n172: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Operator)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t299_n180;
  t299_n173: if (E.getFractionNameInContext() >= 3185061251 /*0.10555555671453476*/) goto t299_n177;
  t299_n174: if (E.getNumReferences() >= 3322777088 /*9059.5*/) goto t299_n176;
  t299_n175: return 0.07229112088680267f;
  t299_n176: return -0.13464412093162537f;
  t299_n177: if (E.getFractionNameInContext() >= 3189668328 /*0.1547619104385376*/) goto t299_n179;
  t299_n178: return -0.5616565942764282f;
  t299_n179: return -0.14798685908317566f;
  t299_n180: if (E.getNumReferences() >= 3322210304 /*8506.0*/) goto t299_n182;
  t299_n181: return -0.12878279387950897f;
  t299_n182: if (E.getNumReferences() >= 3322376704 /*8668.5*/) goto t299_n188;
  t299_n183: if (E.getFileProximityDistanceCost() >= 3248226304 /*19.5*/) goto t299_n185;
  t299_n184: return 0.02388383261859417f;
  t299_n185: if (E.getFractionNameInContext() >= 3169944461 /*0.02948167361319065*/) goto t299_n187;
  t299_n186: return 0.10800489038228989f;
  t299_n187: return 0.11369214951992035f;
  t299_n188: return 0.005215323995798826f;
  t299_n189: return 0.011617377400398254f;
  t299_n190: if (E.getNumReferences() >= 3244818432 /*14.5*/) goto t299_n224;
  t299_n191: if (E.getNumReferences() >= 3227516928 /*3.5*/) goto t299_n211;
  t299_n192: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t299_n198;
  t299_n193: if (E.getFractionNameInContext() >= 3206685218 /*0.6333333253860474*/) goto t299_n195;
  t299_n194: return 0.021910659968852997f;
  t299_n195: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t299_n197;
  t299_n196: return 0.06733373552560806f;
  t299_n197: return -0.20189377665519714f;
  t299_n198: if (E.getFractionNameInContext() >= 3205287117 /*0.550000011920929*/) goto t299_n200;
  t299_n199: return 0.011755004525184631f;
  t299_n200: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t299_n202;
  t299_n201: return 0.1071215569972992f;
  t299_n202: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t299_n204;
  t299_n203: return -0.039280012249946594f;
  t299_n204: if (E.getFileProximityDistanceCost() >= 3240624128 /*10.5*/) goto t299_n210;
  t299_n205: if (E.getSemaFileProximityScore() >= 3197737370 /*0.30000001192092896*/) goto t299_n209;
  t299_n206: if (E.getFractionNameInContext() >= 3207384269 /*0.675000011920929*/) goto t299_n208;
  t299_n207: return 0.12001464515924454f;
  t299_n208: return -0.10242481529712677f;
  t299_n209: return -0.1301986128091812f;
  t299_n210: return 0.1144319400191307f;
  t299_n211: if (E.getFractionNameInContext() >= 3192284774 /*0.19374999403953552*/) goto t299_n223;
  t299_n212: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Constructor)|BIT(SymbolCategory_type::Destructor)|BIT(SymbolCategory_type::Keyword)|BIT(SymbolCategory_type::Macro)|BIT(SymbolCategory_type::Namespace)|BIT(SymbolCategory_type::Type)|BIT(SymbolCategory_type::Unknown)|BIT(SymbolCategory_type::Variable))) goto t299_n218;
  t299_n213: if (E.getScope() & (BIT(Scope_type::ClassScope)|BIT(Scope_type::FileScope)|BIT(Scope_type::FunctionScope))) goto t299_n215;
  t299_n214: return -0.019331078976392746f;
  t299_n215: if (E.getFileProximityDistanceCost() >= 3230662656 /*4.5*/) goto t299_n217;
  t299_n216: return 0.11024577915668488f;
  t299_n217: return -0.12810271978378296f;
  t299_n218: if (E.getFileProximityDistanceCost() >= 3252944896 /*28.5*/) goto t299_n220;
  t299_n219: return 0.014463216066360474f;
  t299_n220: if (E.getNumReferences() >= 3236954112 /*7.5*/) goto t299_n222;
  t299_n221: return 0.11567224562168121f;
  t299_n222: return -0.10011037439107895f;
  t299_n223: return -0.02554204687476158f;
  t299_n224: if (E.getFractionNameInContext() >= 3195414370 /*0.24038460850715637*/) goto t299_n226;
  t299_n225: return 0.016401227563619614f;
  t299_n226: if (E.getNumReferences() >= 3292528640 /*768.0*/) goto t299_n230;
  t299_n227: if (E.getNumReferences() >= 3292127232 /*743.5*/) goto t299_n229;
  t299_n228: return 0.03723776340484619f;
  t299_n229: return 0.14949820935726166f;
  t299_n230: if (E.getNumReferences() >= 3298484224 /*1239.0*/) goto t299_n240;
  t299_n231: if (E.getNumNameInContext() >= 3217031168 /*1.5*/) goto t299_n233;
  t299_n232: return -0.10958750545978546f;
  t299_n233: if (E.getSymbolScopeDistanceCost() >= 3212836864 /*1.0*/) goto t299_n239;
  t299_n234: if (E.getFractionNameInContext() >= 3196440948 /*0.2613636255264282*/) goto t299_n236;
  t299_n235: return -0.07419106364250183f;
  t299_n236: if (E.getNumReferences() >= 3295379456 /*942.0*/) goto t299_n238;
  t299_n237: return -0.31198373436927795f;
  t299_n238: return -1.0109971761703491f;
  t299_n239: return 0.08453667908906937f;
  t299_n240: if (E.getNumNameInContext() >= 3223322624 /*2.5*/) goto t299_n242;
  t299_n241: return -0.010659984312951565f;
  t299_n242: return 0.07497616112232208f;
  t299_n243: if (E.getNumReferences() >= 3300892672 /*1533.0*/) goto t299_n249;
  t299_n244: if (E.getNumReferences() >= 3292741632 /*781.0*/) goto t299_n248;
  t299_n245: if (E.getSymbolCategory() & (BIT(SymbolCategory_type::Namespace))) goto t299_n247;
  t299_n246: return 0.013447285629808903f;
  t299_n247: return 0.12509597837924957f;
  t299_n248: return 0.12423589825630188f;
  t299_n249: return -0.23005904257297516f;
  t299_n250: return -0.0051858569495379925f;
  t299_n251: return -0.0008623418398201466f;
  t299_n252: if (E.getContextKind() & (BIT(ContextKind_type::CCC_Symbol))) goto t299_n254;
  t299_n253: return 0.0008379556820727885f;
  t299_n254: return 0.0031091447453945875f;
}

} // namespace

float Evaluate(const Example& E) {
  float Score = 0;
  Score += EvaluateTree0(E);
  Score += EvaluateTree1(E);
  Score += EvaluateTree2(E);
  Score += EvaluateTree3(E);
  Score += EvaluateTree4(E);
  Score += EvaluateTree5(E);
  Score += EvaluateTree6(E);
  Score += EvaluateTree7(E);
  Score += EvaluateTree8(E);
  Score += EvaluateTree9(E);
  Score += EvaluateTree10(E);
  Score += EvaluateTree11(E);
  Score += EvaluateTree12(E);
  Score += EvaluateTree13(E);
  Score += EvaluateTree14(E);
  Score += EvaluateTree15(E);
  Score += EvaluateTree16(E);
  Score += EvaluateTree17(E);
  Score += EvaluateTree18(E);
  Score += EvaluateTree19(E);
  Score += EvaluateTree20(E);
  Score += EvaluateTree21(E);
  Score += EvaluateTree22(E);
  Score += EvaluateTree23(E);
  Score += EvaluateTree24(E);
  Score += EvaluateTree25(E);
  Score += EvaluateTree26(E);
  Score += EvaluateTree27(E);
  Score += EvaluateTree28(E);
  Score += EvaluateTree29(E);
  Score += EvaluateTree30(E);
  Score += EvaluateTree31(E);
  Score += EvaluateTree32(E);
  Score += EvaluateTree33(E);
  Score += EvaluateTree34(E);
  Score += EvaluateTree35(E);
  Score += EvaluateTree36(E);
  Score += EvaluateTree37(E);
  Score += EvaluateTree38(E);
  Score += EvaluateTree39(E);
  Score += EvaluateTree40(E);
  Score += EvaluateTree41(E);
  Score += EvaluateTree42(E);
  Score += EvaluateTree43(E);
  Score += EvaluateTree44(E);
  Score += EvaluateTree45(E);
  Score += EvaluateTree46(E);
  Score += EvaluateTree47(E);
  Score += EvaluateTree48(E);
  Score += EvaluateTree49(E);
  Score += EvaluateTree50(E);
  Score += EvaluateTree51(E);
  Score += EvaluateTree52(E);
  Score += EvaluateTree53(E);
  Score += EvaluateTree54(E);
  Score += EvaluateTree55(E);
  Score += EvaluateTree56(E);
  Score += EvaluateTree57(E);
  Score += EvaluateTree58(E);
  Score += EvaluateTree59(E);
  Score += EvaluateTree60(E);
  Score += EvaluateTree61(E);
  Score += EvaluateTree62(E);
  Score += EvaluateTree63(E);
  Score += EvaluateTree64(E);
  Score += EvaluateTree65(E);
  Score += EvaluateTree66(E);
  Score += EvaluateTree67(E);
  Score += EvaluateTree68(E);
  Score += EvaluateTree69(E);
  Score += EvaluateTree70(E);
  Score += EvaluateTree71(E);
  Score += EvaluateTree72(E);
  Score += EvaluateTree73(E);
  Score += EvaluateTree74(E);
  Score += EvaluateTree75(E);
  Score += EvaluateTree76(E);
  Score += EvaluateTree77(E);
  Score += EvaluateTree78(E);
  Score += EvaluateTree79(E);
  Score += EvaluateTree80(E);
  Score += EvaluateTree81(E);
  Score += EvaluateTree82(E);
  Score += EvaluateTree83(E);
  Score += EvaluateTree84(E);
  Score += EvaluateTree85(E);
  Score += EvaluateTree86(E);
  Score += EvaluateTree87(E);
  Score += EvaluateTree88(E);
  Score += EvaluateTree89(E);
  Score += EvaluateTree90(E);
  Score += EvaluateTree91(E);
  Score += EvaluateTree92(E);
  Score += EvaluateTree93(E);
  Score += EvaluateTree94(E);
  Score += EvaluateTree95(E);
  Score += EvaluateTree96(E);
  Score += EvaluateTree97(E);
  Score += EvaluateTree98(E);
  Score += EvaluateTree99(E);
  Score += EvaluateTree100(E);
  Score += EvaluateTree101(E);
  Score += EvaluateTree102(E);
  Score += EvaluateTree103(E);
  Score += EvaluateTree104(E);
  Score += EvaluateTree105(E);
  Score += EvaluateTree106(E);
  Score += EvaluateTree107(E);
  Score += EvaluateTree108(E);
  Score += EvaluateTree109(E);
  Score += EvaluateTree110(E);
  Score += EvaluateTree111(E);
  Score += EvaluateTree112(E);
  Score += EvaluateTree113(E);
  Score += EvaluateTree114(E);
  Score += EvaluateTree115(E);
  Score += EvaluateTree116(E);
  Score += EvaluateTree117(E);
  Score += EvaluateTree118(E);
  Score += EvaluateTree119(E);
  Score += EvaluateTree120(E);
  Score += EvaluateTree121(E);
  Score += EvaluateTree122(E);
  Score += EvaluateTree123(E);
  Score += EvaluateTree124(E);
  Score += EvaluateTree125(E);
  Score += EvaluateTree126(E);
  Score += EvaluateTree127(E);
  Score += EvaluateTree128(E);
  Score += EvaluateTree129(E);
  Score += EvaluateTree130(E);
  Score += EvaluateTree131(E);
  Score += EvaluateTree132(E);
  Score += EvaluateTree133(E);
  Score += EvaluateTree134(E);
  Score += EvaluateTree135(E);
  Score += EvaluateTree136(E);
  Score += EvaluateTree137(E);
  Score += EvaluateTree138(E);
  Score += EvaluateTree139(E);
  Score += EvaluateTree140(E);
  Score += EvaluateTree141(E);
  Score += EvaluateTree142(E);
  Score += EvaluateTree143(E);
  Score += EvaluateTree144(E);
  Score += EvaluateTree145(E);
  Score += EvaluateTree146(E);
  Score += EvaluateTree147(E);
  Score += EvaluateTree148(E);
  Score += EvaluateTree149(E);
  Score += EvaluateTree150(E);
  Score += EvaluateTree151(E);
  Score += EvaluateTree152(E);
  Score += EvaluateTree153(E);
  Score += EvaluateTree154(E);
  Score += EvaluateTree155(E);
  Score += EvaluateTree156(E);
  Score += EvaluateTree157(E);
  Score += EvaluateTree158(E);
  Score += EvaluateTree159(E);
  Score += EvaluateTree160(E);
  Score += EvaluateTree161(E);
  Score += EvaluateTree162(E);
  Score += EvaluateTree163(E);
  Score += EvaluateTree164(E);
  Score += EvaluateTree165(E);
  Score += EvaluateTree166(E);
  Score += EvaluateTree167(E);
  Score += EvaluateTree168(E);
  Score += EvaluateTree169(E);
  Score += EvaluateTree170(E);
  Score += EvaluateTree171(E);
  Score += EvaluateTree172(E);
  Score += EvaluateTree173(E);
  Score += EvaluateTree174(E);
  Score += EvaluateTree175(E);
  Score += EvaluateTree176(E);
  Score += EvaluateTree177(E);
  Score += EvaluateTree178(E);
  Score += EvaluateTree179(E);
  Score += EvaluateTree180(E);
  Score += EvaluateTree181(E);
  Score += EvaluateTree182(E);
  Score += EvaluateTree183(E);
  Score += EvaluateTree184(E);
  Score += EvaluateTree185(E);
  Score += EvaluateTree186(E);
  Score += EvaluateTree187(E);
  Score += EvaluateTree188(E);
  Score += EvaluateTree189(E);
  Score += EvaluateTree190(E);
  Score += EvaluateTree191(E);
  Score += EvaluateTree192(E);
  Score += EvaluateTree193(E);
  Score += EvaluateTree194(E);
  Score += EvaluateTree195(E);
  Score += EvaluateTree196(E);
  Score += EvaluateTree197(E);
  Score += EvaluateTree198(E);
  Score += EvaluateTree199(E);
  Score += EvaluateTree200(E);
  Score += EvaluateTree201(E);
  Score += EvaluateTree202(E);
  Score += EvaluateTree203(E);
  Score += EvaluateTree204(E);
  Score += EvaluateTree205(E);
  Score += EvaluateTree206(E);
  Score += EvaluateTree207(E);
  Score += EvaluateTree208(E);
  Score += EvaluateTree209(E);
  Score += EvaluateTree210(E);
  Score += EvaluateTree211(E);
  Score += EvaluateTree212(E);
  Score += EvaluateTree213(E);
  Score += EvaluateTree214(E);
  Score += EvaluateTree215(E);
  Score += EvaluateTree216(E);
  Score += EvaluateTree217(E);
  Score += EvaluateTree218(E);
  Score += EvaluateTree219(E);
  Score += EvaluateTree220(E);
  Score += EvaluateTree221(E);
  Score += EvaluateTree222(E);
  Score += EvaluateTree223(E);
  Score += EvaluateTree224(E);
  Score += EvaluateTree225(E);
  Score += EvaluateTree226(E);
  Score += EvaluateTree227(E);
  Score += EvaluateTree228(E);
  Score += EvaluateTree229(E);
  Score += EvaluateTree230(E);
  Score += EvaluateTree231(E);
  Score += EvaluateTree232(E);
  Score += EvaluateTree233(E);
  Score += EvaluateTree234(E);
  Score += EvaluateTree235(E);
  Score += EvaluateTree236(E);
  Score += EvaluateTree237(E);
  Score += EvaluateTree238(E);
  Score += EvaluateTree239(E);
  Score += EvaluateTree240(E);
  Score += EvaluateTree241(E);
  Score += EvaluateTree242(E);
  Score += EvaluateTree243(E);
  Score += EvaluateTree244(E);
  Score += EvaluateTree245(E);
  Score += EvaluateTree246(E);
  Score += EvaluateTree247(E);
  Score += EvaluateTree248(E);
  Score += EvaluateTree249(E);
  Score += EvaluateTree250(E);
  Score += EvaluateTree251(E);
  Score += EvaluateTree252(E);
  Score += EvaluateTree253(E);
  Score += EvaluateTree254(E);
  Score += EvaluateTree255(E);
  Score += EvaluateTree256(E);
  Score += EvaluateTree257(E);
  Score += EvaluateTree258(E);
  Score += EvaluateTree259(E);
  Score += EvaluateTree260(E);
  Score += EvaluateTree261(E);
  Score += EvaluateTree262(E);
  Score += EvaluateTree263(E);
  Score += EvaluateTree264(E);
  Score += EvaluateTree265(E);
  Score += EvaluateTree266(E);
  Score += EvaluateTree267(E);
  Score += EvaluateTree268(E);
  Score += EvaluateTree269(E);
  Score += EvaluateTree270(E);
  Score += EvaluateTree271(E);
  Score += EvaluateTree272(E);
  Score += EvaluateTree273(E);
  Score += EvaluateTree274(E);
  Score += EvaluateTree275(E);
  Score += EvaluateTree276(E);
  Score += EvaluateTree277(E);
  Score += EvaluateTree278(E);
  Score += EvaluateTree279(E);
  Score += EvaluateTree280(E);
  Score += EvaluateTree281(E);
  Score += EvaluateTree282(E);
  Score += EvaluateTree283(E);
  Score += EvaluateTree284(E);
  Score += EvaluateTree285(E);
  Score += EvaluateTree286(E);
  Score += EvaluateTree287(E);
  Score += EvaluateTree288(E);
  Score += EvaluateTree289(E);
  Score += EvaluateTree290(E);
  Score += EvaluateTree291(E);
  Score += EvaluateTree292(E);
  Score += EvaluateTree293(E);
  Score += EvaluateTree294(E);
  Score += EvaluateTree295(E);
  Score += EvaluateTree296(E);
  Score += EvaluateTree297(E);
  Score += EvaluateTree298(E);
  Score += EvaluateTree299(E);
  return Score;
}

} // namespace clangd
} // namespace clang
